
use strict;
use warnings;

# If run outside the normal test suite on an ASCII platform, you can
# just create a latin1_to_native() function that just returns its
# inputs, because that's the only function used from test.pl
require "test.pl";

# Test qr/\X/ and the \p{} regular expression constructs.  This file is
# constructed by mktables from the tables it generates, so if mktables is
# buggy, this won't necessarily catch those bugs.  Tests are generated for all
# feasible properties; a few aren't currently feasible; see
# is_code_point_usable() in mktables for details.

# Standard test packages are not used because this manipulates SIG_WARN.  It
# exits 0 if every non-skipped test succeeded; -1 if any failed.

my $Tests = 0;
my $Fails = 0;

sub Expect($$$$) {
    my $expected = shift;
    my $ord = shift;
    my $regex  = shift;
    my $warning_type = shift;   # Type of warning message, like 'deprecated'
                                # or empty if none
    my $line   = (caller)[2];

    # Convert the code point to hex form
    my $string = sprintf "\"\\x{%04X}\"", $ord;

    my @tests = "";

    # The first time through, use all warnings.  If the input should generate
    # a warning, add another time through with them turned off
    push @tests, "no warnings '$warning_type';" if $warning_type;

    foreach my $no_warnings (@tests) {

        # Store any warning messages instead of outputting them
        local $SIG{__WARN__} = $SIG{__WARN__};
        my $warning_message;
        $SIG{__WARN__} = sub { $warning_message = $_[0] };

        $Tests++;

        # A string eval is needed because of the 'no warnings'.
        # Assumes no parens in the regular expression
        my $result = eval "$no_warnings
                            my \$RegObj = qr($regex);
                            $string =~ \$RegObj ? 1 : 0";
        if (not defined $result) {
            print "not ok $Tests - couldn't compile /$regex/; line $line: $@\n";
            $Fails++;
        }
        elsif ($result ^ $expected) {
            print "not ok $Tests - expected $expected but got $result for $string =~ qr/$regex/; line $line\n";
            $Fails++;
        }
        elsif ($warning_message) {
            if (! $warning_type || ($warning_type && $no_warnings)) {
                print "not ok $Tests - for qr/$regex/ did not expect warning message '$warning_message'; line $line\n";
                $Fails++;
            }
            else {
                print "ok $Tests - expected and got a warning message for qr/$regex/; line $line\n";
            }
        }
        elsif ($warning_type && ! $no_warnings) {
            print "not ok $Tests - for qr/$regex/ expected a $warning_type warning message, but got none; line $line\n";
            $Fails++;
        }
        else {
            print "ok $Tests - got $result for $string =~ qr/$regex/; line $line\n";
        }
    }
    return;
}

sub Error($) {
    my $regex  = shift;
    $Tests++;
    if (eval { 'x' =~ qr/$regex/; 1 }) {
        $Fails++;
        my $line = (caller)[2];
        print "not ok $Tests - re compiled ok, but expected error for qr/$regex/; line $line: $@\n";
    }
    else {
        my $line = (caller)[2];
        print "ok $Tests - got and expected error for qr/$regex/; line $line\n";
    }
    return;
}

# GCBTest.txt character that separates grapheme clusters
my $breakable_utf8 = my $breakable = chr(utf8::unicode_to_native(0xF7));
utf8::upgrade($breakable_utf8);

# GCBTest.txt character that indicates that the adjoining code points are part
# of the same grapheme cluster
my $nobreak_utf8 = my $nobreak = chr(utf8::unicode_to_native(0xD7));
utf8::upgrade($nobreak_utf8);

sub Test_X($) {
    # Test qr/\X/ matches.  The input is a line from auxiliary/GCBTest.txt
    # Each such line is a sequence of code points given by their hex numbers,
    # separated by the two characters defined just before this subroutine that
    # indicate that either there can or cannot be a break between the adjacent
    # code points.  If there isn't a break, that means the sequence forms an
    # extended grapheme cluster, which means that \X should match the whole
    # thing.  If there is a break, \X should stop there.  This is all
    # converted by this routine into a match:
    #   $string =~ /(\X)/,
    # Each \X should match the next cluster; and that is what is checked.

    my $template = shift;

    my $line   = (caller)[2];

    # The line contains characters above the ASCII range, but in Latin1.  It
    # may or may not be in utf8, and if it is, it may or may not know it.  So,
    # convert these characters to 8 bits.  If knows is in utf8, simply
    # downgrade.
    if (utf8::is_utf8($template)) {
        utf8::downgrade($template);
    } else {

        # Otherwise, if it is in utf8, but doesn't know it, the next lines
        # convert the two problematic characters to their 8-bit equivalents.
        # If it isn't in utf8, they don't harm anything.
        use bytes;
        $template =~ s/$nobreak_utf8/$nobreak/g;
        $template =~ s/$breakable_utf8/$breakable/g;
    }

    # Get rid of the leading and trailing breakables
    $template =~ s/^ \s* $breakable \s* //x;
    $template =~ s/ \s* $breakable \s* $ //x;

    # And no-breaks become just a space.
    $template =~ s/ \s* $nobreak \s* / /xg;

    # Split the input into segments that are breakable between them.
    my @segments = split /\s*$breakable\s*/, $template;

    my $string = "";
    my $display_string = "";
    my @should_match;
    my @should_display;

    # Convert the code point sequence in each segment into a Perl string of
    # characters
    foreach my $segment (@segments) {
        my @code_points = split /\s+/, $segment;
        my $this_string = "";
        my $this_display = "";
        foreach my $code_point (@code_points) {
            $this_string .= latin1_to_native(chr(hex $code_point));
            $this_display .= "\\x{$code_point}";
        }

        # The next cluster should match the string in this segment.
        push @should_match, $this_string;
        push @should_display, $this_display;
        $string .= $this_string;
        $display_string .= $this_display;
    }

    # If a string can be represented in both non-ut8 and utf8, test both cases
    UPGRADE:
    for my $to_upgrade (0 .. 1) {

        if ($to_upgrade) {

            # If already in utf8, would just be a repeat
            next UPGRADE if utf8::is_utf8($string);

            utf8::upgrade($string);
        }

        # Finally, do the \X match.
        my @matches = $string =~ /(\X)/g;

        # Look through each matched cluster to verify that it matches what we
        # expect.
        my $min = (@matches < @should_match) ? @matches : @should_match;
        for my $i (0 .. $min - 1) {
            $Tests++;
            if ($matches[$i] eq $should_match[$i]) {
                print "ok $Tests - ";
                if ($i == 0) {
                    print "In \"$display_string\" =~ /(\\X)/g, \\X #1";
                } else {
                    print "And \\X #", $i + 1,
                }
                print " correctly matched $should_display[$i]; line $line\n";
            } else {
                $matches[$i] = join("", map { sprintf "\\x{%04X}", $_ }
                                                    unpack("U*", $matches[$i]));
                print "not ok $Tests - In \"$display_string\" =~ /(\\X)/g, \\X #",
                    $i + 1,
                    " should have matched $should_display[$i]",
                    " but instead matched $matches[$i]",
                    ".  Abandoning rest of line $line\n";
                next UPGRADE;
            }
        }

        # And the number of matches should equal the number of expected matches.
        $Tests++;
        if (@matches == @should_match) {
            print "ok $Tests - Nothing was left over; line $line\n";
        } else {
            print "not ok $Tests - There were ", scalar @should_match, " \\X matches expected, but got ", scalar @matches, " instead; line $line\n";
        }
    }

    return;
}

sub Finished() {
    print "1..$Tests\n";
    exit($Fails ? -1 : 0);
}

Error('\p{Script=InGreek}');    # Bug #69018
Test_X("1100 $nobreak 1161");  # Bug #70940
Expect(0, 0x2028, '\p{Print}', ""); # Bug # 71722
Expect(0, 0x2029, '\p{Print}', ""); # Bug # 71722
Expect(1, 0xFF10, '\p{XDigit}', ""); # Bug # 71726
Error('\p{ASCII_Hex_Digit= 	NO:=}');
Error('\P{ASCII_Hex_Digit= 	NO:=}');
Expect(1, 103, '\p{ASCII_Hex_Digit=no}', "");
Expect(0, 103, '\p{^ASCII_Hex_Digit=no}', "");
Expect(0, 103, '\P{ASCII_Hex_Digit=no}', "");
Expect(1, 103, '\P{^ASCII_Hex_Digit=no}', "");
Expect(0, 102, '\p{ASCII_Hex_Digit=no}', "");
Expect(1, 102, '\p{^ASCII_Hex_Digit=no}', "");
Expect(1, 102, '\P{ASCII_Hex_Digit=no}', "");
Expect(0, 102, '\P{^ASCII_Hex_Digit=no}', "");
Expect(1, 103, '\p{ASCII_Hex_Digit=-No}', "");
Expect(0, 103, '\p{^ASCII_Hex_Digit=-No}', "");
Expect(0, 103, '\P{ASCII_Hex_Digit=-No}', "");
Expect(1, 103, '\P{^ASCII_Hex_Digit=-No}', "");
Expect(0, 102, '\p{ASCII_Hex_Digit=-No}', "");
Expect(1, 102, '\p{^ASCII_Hex_Digit=-No}', "");
Expect(1, 102, '\P{ASCII_Hex_Digit=-No}', "");
Expect(0, 102, '\P{^ASCII_Hex_Digit=-No}', "");
Error('\p{AHex=-:=n}');
Error('\P{AHex=-:=n}');
Expect(1, 103, '\p{AHex=n}', "");
Expect(0, 103, '\p{^AHex=n}', "");
Expect(0, 103, '\P{AHex=n}', "");
Expect(1, 103, '\P{^AHex=n}', "");
Expect(0, 102, '\p{AHex=n}', "");
Expect(1, 102, '\p{^AHex=n}', "");
Expect(1, 102, '\P{AHex=n}', "");
Expect(0, 102, '\P{^AHex=n}', "");
Expect(1, 103, '\p{AHex=	-N}', "");
Expect(0, 103, '\p{^AHex=	-N}', "");
Expect(0, 103, '\P{AHex=	-N}', "");
Expect(1, 103, '\P{^AHex=	-N}', "");
Expect(0, 102, '\p{AHex=	-N}', "");
Expect(1, 102, '\p{^AHex=	-N}', "");
Expect(1, 102, '\P{AHex=	-N}', "");
Expect(0, 102, '\P{^AHex=	-N}', "");
Error('\p{Is_ASCII_Hex_Digit: _:=f}');
Error('\P{Is_ASCII_Hex_Digit: _:=f}');
Expect(1, 103, '\p{Is_ASCII_Hex_Digit=f}', "");
Expect(0, 103, '\p{^Is_ASCII_Hex_Digit=f}', "");
Expect(0, 103, '\P{Is_ASCII_Hex_Digit=f}', "");
Expect(1, 103, '\P{^Is_ASCII_Hex_Digit=f}', "");
Expect(0, 102, '\p{Is_ASCII_Hex_Digit=f}', "");
Expect(1, 102, '\p{^Is_ASCII_Hex_Digit=f}', "");
Expect(1, 102, '\P{Is_ASCII_Hex_Digit=f}', "");
Expect(0, 102, '\P{^Is_ASCII_Hex_Digit=f}', "");
Expect(1, 103, '\p{Is_ASCII_Hex_Digit=- F}', "");
Expect(0, 103, '\p{^Is_ASCII_Hex_Digit=- F}', "");
Expect(0, 103, '\P{Is_ASCII_Hex_Digit=- F}', "");
Expect(1, 103, '\P{^Is_ASCII_Hex_Digit=- F}', "");
Expect(0, 102, '\p{Is_ASCII_Hex_Digit=- F}', "");
Expect(1, 102, '\p{^Is_ASCII_Hex_Digit=- F}', "");
Expect(1, 102, '\P{Is_ASCII_Hex_Digit=- F}', "");
Expect(0, 102, '\P{^Is_ASCII_Hex_Digit=- F}', "");
Error('\p{Is_AHex=	 FALSE/a/}');
Error('\P{Is_AHex=	 FALSE/a/}');
Expect(1, 103, '\p{Is_AHex=false}', "");
Expect(0, 103, '\p{^Is_AHex=false}', "");
Expect(0, 103, '\P{Is_AHex=false}', "");
Expect(1, 103, '\P{^Is_AHex=false}', "");
Expect(0, 102, '\p{Is_AHex=false}', "");
Expect(1, 102, '\p{^Is_AHex=false}', "");
Expect(1, 102, '\P{Is_AHex=false}', "");
Expect(0, 102, '\P{^Is_AHex=false}', "");
Expect(1, 103, '\p{Is_AHex=	FALSE}', "");
Expect(0, 103, '\p{^Is_AHex=	FALSE}', "");
Expect(0, 103, '\P{Is_AHex=	FALSE}', "");
Expect(1, 103, '\P{^Is_AHex=	FALSE}', "");
Expect(0, 102, '\p{Is_AHex=	FALSE}', "");
Expect(1, 102, '\p{^Is_AHex=	FALSE}', "");
Expect(1, 102, '\P{Is_AHex=	FALSE}', "");
Expect(0, 102, '\P{^Is_AHex=	FALSE}', "");
Error('\p{ASCII_Hex_Digit=:=_-yes}');
Error('\P{ASCII_Hex_Digit=:=_-yes}');
Expect(1, 102, '\p{ASCII_Hex_Digit=yes}', "");
Expect(0, 102, '\p{^ASCII_Hex_Digit=yes}', "");
Expect(0, 102, '\P{ASCII_Hex_Digit=yes}', "");
Expect(1, 102, '\P{^ASCII_Hex_Digit=yes}', "");
Expect(0, 103, '\p{ASCII_Hex_Digit=yes}', "");
Expect(1, 103, '\p{^ASCII_Hex_Digit=yes}', "");
Expect(1, 103, '\P{ASCII_Hex_Digit=yes}', "");
Expect(0, 103, '\P{^ASCII_Hex_Digit=yes}', "");
Expect(1, 102, '\p{ASCII_Hex_Digit=_yes}', "");
Expect(0, 102, '\p{^ASCII_Hex_Digit=_yes}', "");
Expect(0, 102, '\P{ASCII_Hex_Digit=_yes}', "");
Expect(1, 102, '\P{^ASCII_Hex_Digit=_yes}', "");
Expect(0, 103, '\p{ASCII_Hex_Digit=_yes}', "");
Expect(1, 103, '\p{^ASCII_Hex_Digit=_yes}', "");
Expect(1, 103, '\P{ASCII_Hex_Digit=_yes}', "");
Expect(0, 103, '\P{^ASCII_Hex_Digit=_yes}', "");
Error('\p{AHex=_/a/Y}');
Error('\P{AHex=_/a/Y}');
Expect(1, 102, '\p{AHex=y}', "");
Expect(0, 102, '\p{^AHex=y}', "");
Expect(0, 102, '\P{AHex=y}', "");
Expect(1, 102, '\P{^AHex=y}', "");
Expect(0, 103, '\p{AHex=y}', "");
Expect(1, 103, '\p{^AHex=y}', "");
Expect(1, 103, '\P{AHex=y}', "");
Expect(0, 103, '\P{^AHex=y}', "");
Expect(1, 102, '\p{AHex=  Y}', "");
Expect(0, 102, '\p{^AHex=  Y}', "");
Expect(0, 102, '\P{AHex=  Y}', "");
Expect(1, 102, '\P{^AHex=  Y}', "");
Expect(0, 103, '\p{AHex=  Y}', "");
Expect(1, 103, '\p{^AHex=  Y}', "");
Expect(1, 103, '\P{AHex=  Y}', "");
Expect(0, 103, '\P{^AHex=  Y}', "");
Error('\p{Is_ASCII_Hex_Digit= T:=}');
Error('\P{Is_ASCII_Hex_Digit= T:=}');
Expect(1, 102, '\p{Is_ASCII_Hex_Digit=t}', "");
Expect(0, 102, '\p{^Is_ASCII_Hex_Digit=t}', "");
Expect(0, 102, '\P{Is_ASCII_Hex_Digit=t}', "");
Expect(1, 102, '\P{^Is_ASCII_Hex_Digit=t}', "");
Expect(0, 103, '\p{Is_ASCII_Hex_Digit=t}', "");
Expect(1, 103, '\p{^Is_ASCII_Hex_Digit=t}', "");
Expect(1, 103, '\P{Is_ASCII_Hex_Digit=t}', "");
Expect(0, 103, '\P{^Is_ASCII_Hex_Digit=t}', "");
Expect(1, 102, '\p{Is_ASCII_Hex_Digit:		T}', "");
Expect(0, 102, '\p{^Is_ASCII_Hex_Digit:		T}', "");
Expect(0, 102, '\P{Is_ASCII_Hex_Digit:		T}', "");
Expect(1, 102, '\P{^Is_ASCII_Hex_Digit:		T}', "");
Expect(0, 103, '\p{Is_ASCII_Hex_Digit:		T}', "");
Expect(1, 103, '\p{^Is_ASCII_Hex_Digit:		T}', "");
Expect(1, 103, '\P{Is_ASCII_Hex_Digit:		T}', "");
Expect(0, 103, '\P{^Is_ASCII_Hex_Digit:		T}', "");
Error('\p{Is_AHex:	/a/ true}');
Error('\P{Is_AHex:	/a/ true}');
Expect(1, 102, '\p{Is_AHex=true}', "");
Expect(0, 102, '\p{^Is_AHex=true}', "");
Expect(0, 102, '\P{Is_AHex=true}', "");
Expect(1, 102, '\P{^Is_AHex=true}', "");
Expect(0, 103, '\p{Is_AHex=true}', "");
Expect(1, 103, '\p{^Is_AHex=true}', "");
Expect(1, 103, '\P{Is_AHex=true}', "");
Expect(0, 103, '\P{^Is_AHex=true}', "");
Expect(1, 102, '\p{Is_AHex=	-True}', "");
Expect(0, 102, '\p{^Is_AHex=	-True}', "");
Expect(0, 102, '\P{Is_AHex=	-True}', "");
Expect(1, 102, '\P{^Is_AHex=	-True}', "");
Expect(0, 103, '\p{Is_AHex=	-True}', "");
Expect(1, 103, '\p{^Is_AHex=	-True}', "");
Expect(1, 103, '\P{Is_AHex=	-True}', "");
Expect(0, 103, '\P{^Is_AHex=	-True}', "");
Error('\p{Alphabetic=/a/_ NO}');
Error('\P{Alphabetic=/a/_ NO}');
Expect(1, 195102, '\p{Alphabetic=no}', "");
Expect(0, 195102, '\p{^Alphabetic=no}', "");
Expect(0, 195102, '\P{Alphabetic=no}', "");
Expect(1, 195102, '\P{^Alphabetic=no}', "");
Expect(0, 195101, '\p{Alphabetic=no}', "");
Expect(1, 195101, '\p{^Alphabetic=no}', "");
Expect(1, 195101, '\P{Alphabetic=no}', "");
Expect(0, 195101, '\P{^Alphabetic=no}', "");
Expect(1, 195102, '\p{Alphabetic=  No}', "");
Expect(0, 195102, '\p{^Alphabetic=  No}', "");
Expect(0, 195102, '\P{Alphabetic=  No}', "");
Expect(1, 195102, '\P{^Alphabetic=  No}', "");
Expect(0, 195101, '\p{Alphabetic=  No}', "");
Expect(1, 195101, '\p{^Alphabetic=  No}', "");
Expect(1, 195101, '\P{Alphabetic=  No}', "");
Expect(0, 195101, '\P{^Alphabetic=  No}', "");
Error('\p{Alpha=-	n:=}');
Error('\P{Alpha=-	n:=}');
Expect(1, 195102, '\p{Alpha=n}', "");
Expect(0, 195102, '\p{^Alpha=n}', "");
Expect(0, 195102, '\P{Alpha=n}', "");
Expect(1, 195102, '\P{^Alpha=n}', "");
Expect(0, 195101, '\p{Alpha=n}', "");
Expect(1, 195101, '\p{^Alpha=n}', "");
Expect(1, 195101, '\P{Alpha=n}', "");
Expect(0, 195101, '\P{^Alpha=n}', "");
Expect(1, 195102, '\p{Alpha=		n}', "");
Expect(0, 195102, '\p{^Alpha=		n}', "");
Expect(0, 195102, '\P{Alpha=		n}', "");
Expect(1, 195102, '\P{^Alpha=		n}', "");
Expect(0, 195101, '\p{Alpha=		n}', "");
Expect(1, 195101, '\p{^Alpha=		n}', "");
Expect(1, 195101, '\P{Alpha=		n}', "");
Expect(0, 195101, '\P{^Alpha=		n}', "");
Error('\p{Is_Alphabetic: 	 F:=}');
Error('\P{Is_Alphabetic: 	 F:=}');
Expect(1, 195102, '\p{Is_Alphabetic=f}', "");
Expect(0, 195102, '\p{^Is_Alphabetic=f}', "");
Expect(0, 195102, '\P{Is_Alphabetic=f}', "");
Expect(1, 195102, '\P{^Is_Alphabetic=f}', "");
Expect(0, 195101, '\p{Is_Alphabetic=f}', "");
Expect(1, 195101, '\p{^Is_Alphabetic=f}', "");
Expect(1, 195101, '\P{Is_Alphabetic=f}', "");
Expect(0, 195101, '\P{^Is_Alphabetic=f}', "");
Expect(1, 195102, '\p{Is_Alphabetic: _ F}', "");
Expect(0, 195102, '\p{^Is_Alphabetic: _ F}', "");
Expect(0, 195102, '\P{Is_Alphabetic: _ F}', "");
Expect(1, 195102, '\P{^Is_Alphabetic: _ F}', "");
Expect(0, 195101, '\p{Is_Alphabetic: _ F}', "");
Expect(1, 195101, '\p{^Is_Alphabetic: _ F}', "");
Expect(1, 195101, '\P{Is_Alphabetic: _ F}', "");
Expect(0, 195101, '\P{^Is_Alphabetic: _ F}', "");
Error('\p{Is_Alpha= /a/FALSE}');
Error('\P{Is_Alpha= /a/FALSE}');
Expect(1, 195102, '\p{Is_Alpha=false}', "");
Expect(0, 195102, '\p{^Is_Alpha=false}', "");
Expect(0, 195102, '\P{Is_Alpha=false}', "");
Expect(1, 195102, '\P{^Is_Alpha=false}', "");
Expect(0, 195101, '\p{Is_Alpha=false}', "");
Expect(1, 195101, '\p{^Is_Alpha=false}', "");
Expect(1, 195101, '\P{Is_Alpha=false}', "");
Expect(0, 195101, '\P{^Is_Alpha=false}', "");
Expect(1, 195102, '\p{Is_Alpha=-	FALSE}', "");
Expect(0, 195102, '\p{^Is_Alpha=-	FALSE}', "");
Expect(0, 195102, '\P{Is_Alpha=-	FALSE}', "");
Expect(1, 195102, '\P{^Is_Alpha=-	FALSE}', "");
Expect(0, 195101, '\p{Is_Alpha=-	FALSE}', "");
Expect(1, 195101, '\p{^Is_Alpha=-	FALSE}', "");
Expect(1, 195101, '\P{Is_Alpha=-	FALSE}', "");
Expect(0, 195101, '\P{^Is_Alpha=-	FALSE}', "");
Error('\p{Alphabetic=:=- Yes}');
Error('\P{Alphabetic=:=- Yes}');
Expect(1, 195101, '\p{Alphabetic=yes}', "");
Expect(0, 195101, '\p{^Alphabetic=yes}', "");
Expect(0, 195101, '\P{Alphabetic=yes}', "");
Expect(1, 195101, '\P{^Alphabetic=yes}', "");
Expect(0, 195102, '\p{Alphabetic=yes}', "");
Expect(1, 195102, '\p{^Alphabetic=yes}', "");
Expect(1, 195102, '\P{Alphabetic=yes}', "");
Expect(0, 195102, '\P{^Alphabetic=yes}', "");
Expect(1, 195101, '\p{Alphabetic=_	yes}', "");
Expect(0, 195101, '\p{^Alphabetic=_	yes}', "");
Expect(0, 195101, '\P{Alphabetic=_	yes}', "");
Expect(1, 195101, '\P{^Alphabetic=_	yes}', "");
Expect(0, 195102, '\p{Alphabetic=_	yes}', "");
Expect(1, 195102, '\p{^Alphabetic=_	yes}', "");
Expect(1, 195102, '\P{Alphabetic=_	yes}', "");
Expect(0, 195102, '\P{^Alphabetic=_	yes}', "");
Error('\p{Alpha=/a/Y}');
Error('\P{Alpha=/a/Y}');
Expect(1, 195101, '\p{Alpha=y}', "");
Expect(0, 195101, '\p{^Alpha=y}', "");
Expect(0, 195101, '\P{Alpha=y}', "");
Expect(1, 195101, '\P{^Alpha=y}', "");
Expect(0, 195102, '\p{Alpha=y}', "");
Expect(1, 195102, '\p{^Alpha=y}', "");
Expect(1, 195102, '\P{Alpha=y}', "");
Expect(0, 195102, '\P{^Alpha=y}', "");
Expect(1, 195101, '\p{Alpha=_-Y}', "");
Expect(0, 195101, '\p{^Alpha=_-Y}', "");
Expect(0, 195101, '\P{Alpha=_-Y}', "");
Expect(1, 195101, '\P{^Alpha=_-Y}', "");
Expect(0, 195102, '\p{Alpha=_-Y}', "");
Expect(1, 195102, '\p{^Alpha=_-Y}', "");
Expect(1, 195102, '\P{Alpha=_-Y}', "");
Expect(0, 195102, '\P{^Alpha=_-Y}', "");
Error('\p{Is_Alphabetic=/a/_T}');
Error('\P{Is_Alphabetic=/a/_T}');
Expect(1, 195101, '\p{Is_Alphabetic=t}', "");
Expect(0, 195101, '\p{^Is_Alphabetic=t}', "");
Expect(0, 195101, '\P{Is_Alphabetic=t}', "");
Expect(1, 195101, '\P{^Is_Alphabetic=t}', "");
Expect(0, 195102, '\p{Is_Alphabetic=t}', "");
Expect(1, 195102, '\p{^Is_Alphabetic=t}', "");
Expect(1, 195102, '\P{Is_Alphabetic=t}', "");
Expect(0, 195102, '\P{^Is_Alphabetic=t}', "");
Expect(1, 195101, '\p{Is_Alphabetic= T}', "");
Expect(0, 195101, '\p{^Is_Alphabetic= T}', "");
Expect(0, 195101, '\P{Is_Alphabetic= T}', "");
Expect(1, 195101, '\P{^Is_Alphabetic= T}', "");
Expect(0, 195102, '\p{Is_Alphabetic= T}', "");
Expect(1, 195102, '\p{^Is_Alphabetic= T}', "");
Expect(1, 195102, '\P{Is_Alphabetic= T}', "");
Expect(0, 195102, '\P{^Is_Alphabetic= T}', "");
Error('\p{Is_Alpha=:=-true}');
Error('\P{Is_Alpha=:=-true}');
Expect(1, 195101, '\p{Is_Alpha=true}', "");
Expect(0, 195101, '\p{^Is_Alpha=true}', "");
Expect(0, 195101, '\P{Is_Alpha=true}', "");
Expect(1, 195101, '\P{^Is_Alpha=true}', "");
Expect(0, 195102, '\p{Is_Alpha=true}', "");
Expect(1, 195102, '\p{^Is_Alpha=true}', "");
Expect(1, 195102, '\P{Is_Alpha=true}', "");
Expect(0, 195102, '\P{^Is_Alpha=true}', "");
Expect(1, 195101, '\p{Is_Alpha= TRUE}', "");
Expect(0, 195101, '\p{^Is_Alpha= TRUE}', "");
Expect(0, 195101, '\P{Is_Alpha= TRUE}', "");
Expect(1, 195101, '\P{^Is_Alpha= TRUE}', "");
Expect(0, 195102, '\p{Is_Alpha= TRUE}', "");
Expect(1, 195102, '\p{^Is_Alpha= TRUE}', "");
Expect(1, 195102, '\P{Is_Alpha= TRUE}', "");
Expect(0, 195102, '\P{^Is_Alpha= TRUE}', "");
Error('\p{Bidi_Control=		No/a/}');
Error('\P{Bidi_Control=		No/a/}');
Expect(1, 8298, '\p{Bidi_Control=no}', "");
Expect(0, 8298, '\p{^Bidi_Control=no}', "");
Expect(0, 8298, '\P{Bidi_Control=no}', "");
Expect(1, 8298, '\P{^Bidi_Control=no}', "");
Expect(0, 8297, '\p{Bidi_Control=no}', "");
Expect(1, 8297, '\p{^Bidi_Control=no}', "");
Expect(1, 8297, '\P{Bidi_Control=no}', "");
Expect(0, 8297, '\P{^Bidi_Control=no}', "");
Expect(1, 8298, '\p{Bidi_Control=	 no}', "");
Expect(0, 8298, '\p{^Bidi_Control=	 no}', "");
Expect(0, 8298, '\P{Bidi_Control=	 no}', "");
Expect(1, 8298, '\P{^Bidi_Control=	 no}', "");
Expect(0, 8297, '\p{Bidi_Control=	 no}', "");
Expect(1, 8297, '\p{^Bidi_Control=	 no}', "");
Expect(1, 8297, '\P{Bidi_Control=	 no}', "");
Expect(0, 8297, '\P{^Bidi_Control=	 no}', "");
Error('\p{Bidi_C=:=- N}');
Error('\P{Bidi_C=:=- N}');
Expect(1, 8298, '\p{Bidi_C=n}', "");
Expect(0, 8298, '\p{^Bidi_C=n}', "");
Expect(0, 8298, '\P{Bidi_C=n}', "");
Expect(1, 8298, '\P{^Bidi_C=n}', "");
Expect(0, 8297, '\p{Bidi_C=n}', "");
Expect(1, 8297, '\p{^Bidi_C=n}', "");
Expect(1, 8297, '\P{Bidi_C=n}', "");
Expect(0, 8297, '\P{^Bidi_C=n}', "");
Expect(1, 8298, '\p{Bidi_C=N}', "");
Expect(0, 8298, '\p{^Bidi_C=N}', "");
Expect(0, 8298, '\P{Bidi_C=N}', "");
Expect(1, 8298, '\P{^Bidi_C=N}', "");
Expect(0, 8297, '\p{Bidi_C=N}', "");
Expect(1, 8297, '\p{^Bidi_C=N}', "");
Expect(1, 8297, '\P{Bidi_C=N}', "");
Expect(0, 8297, '\P{^Bidi_C=N}', "");
Error('\p{Is_Bidi_Control=__F:=}');
Error('\P{Is_Bidi_Control=__F:=}');
Expect(1, 8298, '\p{Is_Bidi_Control=f}', "");
Expect(0, 8298, '\p{^Is_Bidi_Control=f}', "");
Expect(0, 8298, '\P{Is_Bidi_Control=f}', "");
Expect(1, 8298, '\P{^Is_Bidi_Control=f}', "");
Expect(0, 8297, '\p{Is_Bidi_Control=f}', "");
Expect(1, 8297, '\p{^Is_Bidi_Control=f}', "");
Expect(1, 8297, '\P{Is_Bidi_Control=f}', "");
Expect(0, 8297, '\P{^Is_Bidi_Control=f}', "");
Expect(1, 8298, '\p{Is_Bidi_Control: 	 F}', "");
Expect(0, 8298, '\p{^Is_Bidi_Control: 	 F}', "");
Expect(0, 8298, '\P{Is_Bidi_Control: 	 F}', "");
Expect(1, 8298, '\P{^Is_Bidi_Control: 	 F}', "");
Expect(0, 8297, '\p{Is_Bidi_Control: 	 F}', "");
Expect(1, 8297, '\p{^Is_Bidi_Control: 	 F}', "");
Expect(1, 8297, '\P{Is_Bidi_Control: 	 F}', "");
Expect(0, 8297, '\P{^Is_Bidi_Control: 	 F}', "");
Error('\p{Is_Bidi_C=:= False}');
Error('\P{Is_Bidi_C=:= False}');
Expect(1, 8298, '\p{Is_Bidi_C=false}', "");
Expect(0, 8298, '\p{^Is_Bidi_C=false}', "");
Expect(0, 8298, '\P{Is_Bidi_C=false}', "");
Expect(1, 8298, '\P{^Is_Bidi_C=false}', "");
Expect(0, 8297, '\p{Is_Bidi_C=false}', "");
Expect(1, 8297, '\p{^Is_Bidi_C=false}', "");
Expect(1, 8297, '\P{Is_Bidi_C=false}', "");
Expect(0, 8297, '\P{^Is_Bidi_C=false}', "");
Expect(1, 8298, '\p{Is_Bidi_C= -false}', "");
Expect(0, 8298, '\p{^Is_Bidi_C= -false}', "");
Expect(0, 8298, '\P{Is_Bidi_C= -false}', "");
Expect(1, 8298, '\P{^Is_Bidi_C= -false}', "");
Expect(0, 8297, '\p{Is_Bidi_C= -false}', "");
Expect(1, 8297, '\p{^Is_Bidi_C= -false}', "");
Expect(1, 8297, '\P{Is_Bidi_C= -false}', "");
Expect(0, 8297, '\P{^Is_Bidi_C= -false}', "");
Error('\p{Bidi_Control=:=--Yes}');
Error('\P{Bidi_Control=:=--Yes}');
Expect(1, 8297, '\p{Bidi_Control=yes}', "");
Expect(0, 8297, '\p{^Bidi_Control=yes}', "");
Expect(0, 8297, '\P{Bidi_Control=yes}', "");
Expect(1, 8297, '\P{^Bidi_Control=yes}', "");
Expect(0, 8298, '\p{Bidi_Control=yes}', "");
Expect(1, 8298, '\p{^Bidi_Control=yes}', "");
Expect(1, 8298, '\P{Bidi_Control=yes}', "");
Expect(0, 8298, '\P{^Bidi_Control=yes}', "");
Expect(1, 8297, '\p{Bidi_Control=-	Yes}', "");
Expect(0, 8297, '\p{^Bidi_Control=-	Yes}', "");
Expect(0, 8297, '\P{Bidi_Control=-	Yes}', "");
Expect(1, 8297, '\P{^Bidi_Control=-	Yes}', "");
Expect(0, 8298, '\p{Bidi_Control=-	Yes}', "");
Expect(1, 8298, '\p{^Bidi_Control=-	Yes}', "");
Expect(1, 8298, '\P{Bidi_Control=-	Yes}', "");
Expect(0, 8298, '\P{^Bidi_Control=-	Yes}', "");
Error('\p{Bidi_C=:=Y}');
Error('\P{Bidi_C=:=Y}');
Expect(1, 8297, '\p{Bidi_C=y}', "");
Expect(0, 8297, '\p{^Bidi_C=y}', "");
Expect(0, 8297, '\P{Bidi_C=y}', "");
Expect(1, 8297, '\P{^Bidi_C=y}', "");
Expect(0, 8298, '\p{Bidi_C=y}', "");
Expect(1, 8298, '\p{^Bidi_C=y}', "");
Expect(1, 8298, '\P{Bidi_C=y}', "");
Expect(0, 8298, '\P{^Bidi_C=y}', "");
Expect(1, 8297, '\p{Bidi_C=	Y}', "");
Expect(0, 8297, '\p{^Bidi_C=	Y}', "");
Expect(0, 8297, '\P{Bidi_C=	Y}', "");
Expect(1, 8297, '\P{^Bidi_C=	Y}', "");
Expect(0, 8298, '\p{Bidi_C=	Y}', "");
Expect(1, 8298, '\p{^Bidi_C=	Y}', "");
Expect(1, 8298, '\P{Bidi_C=	Y}', "");
Expect(0, 8298, '\P{^Bidi_C=	Y}', "");
Error('\p{Is_Bidi_Control=/a/-t}');
Error('\P{Is_Bidi_Control=/a/-t}');
Expect(1, 8297, '\p{Is_Bidi_Control=t}', "");
Expect(0, 8297, '\p{^Is_Bidi_Control=t}', "");
Expect(0, 8297, '\P{Is_Bidi_Control=t}', "");
Expect(1, 8297, '\P{^Is_Bidi_Control=t}', "");
Expect(0, 8298, '\p{Is_Bidi_Control=t}', "");
Expect(1, 8298, '\p{^Is_Bidi_Control=t}', "");
Expect(1, 8298, '\P{Is_Bidi_Control=t}', "");
Expect(0, 8298, '\P{^Is_Bidi_Control=t}', "");
Expect(1, 8297, '\p{Is_Bidi_Control= T}', "");
Expect(0, 8297, '\p{^Is_Bidi_Control= T}', "");
Expect(0, 8297, '\P{Is_Bidi_Control= T}', "");
Expect(1, 8297, '\P{^Is_Bidi_Control= T}', "");
Expect(0, 8298, '\p{Is_Bidi_Control= T}', "");
Expect(1, 8298, '\p{^Is_Bidi_Control= T}', "");
Expect(1, 8298, '\P{Is_Bidi_Control= T}', "");
Expect(0, 8298, '\P{^Is_Bidi_Control= T}', "");
Error('\p{Is_Bidi_C= 	True:=}');
Error('\P{Is_Bidi_C= 	True:=}');
Expect(1, 8297, '\p{Is_Bidi_C=true}', "");
Expect(0, 8297, '\p{^Is_Bidi_C=true}', "");
Expect(0, 8297, '\P{Is_Bidi_C=true}', "");
Expect(1, 8297, '\P{^Is_Bidi_C=true}', "");
Expect(0, 8298, '\p{Is_Bidi_C=true}', "");
Expect(1, 8298, '\p{^Is_Bidi_C=true}', "");
Expect(1, 8298, '\P{Is_Bidi_C=true}', "");
Expect(0, 8298, '\P{^Is_Bidi_C=true}', "");
Expect(1, 8297, '\p{Is_Bidi_C=  True}', "");
Expect(0, 8297, '\p{^Is_Bidi_C=  True}', "");
Expect(0, 8297, '\P{Is_Bidi_C=  True}', "");
Expect(1, 8297, '\P{^Is_Bidi_C=  True}', "");
Expect(0, 8298, '\p{Is_Bidi_C=  True}', "");
Expect(1, 8298, '\p{^Is_Bidi_C=  True}', "");
Expect(1, 8298, '\P{Is_Bidi_C=  True}', "");
Expect(0, 8298, '\P{^Is_Bidi_C=  True}', "");
Error('\p{Bidi_Mirrored=_NO:=}');
Error('\P{Bidi_Mirrored=_NO:=}');
Expect(1, 120772, '\p{Bidi_Mirrored:   no}', "");
Expect(0, 120772, '\p{^Bidi_Mirrored:   no}', "");
Expect(0, 120772, '\P{Bidi_Mirrored:   no}', "");
Expect(1, 120772, '\P{^Bidi_Mirrored:   no}', "");
Expect(0, 120771, '\p{Bidi_Mirrored:   no}', "");
Expect(1, 120771, '\p{^Bidi_Mirrored:   no}', "");
Expect(1, 120771, '\P{Bidi_Mirrored:   no}', "");
Expect(0, 120771, '\P{^Bidi_Mirrored:   no}', "");
Expect(1, 120772, '\p{Bidi_Mirrored=	-No}', "");
Expect(0, 120772, '\p{^Bidi_Mirrored=	-No}', "");
Expect(0, 120772, '\P{Bidi_Mirrored=	-No}', "");
Expect(1, 120772, '\P{^Bidi_Mirrored=	-No}', "");
Expect(0, 120771, '\p{Bidi_Mirrored=	-No}', "");
Expect(1, 120771, '\p{^Bidi_Mirrored=	-No}', "");
Expect(1, 120771, '\P{Bidi_Mirrored=	-No}', "");
Expect(0, 120771, '\P{^Bidi_Mirrored=	-No}', "");
Error('\p{Bidi_M=-N/a/}');
Error('\P{Bidi_M=-N/a/}');
Expect(1, 120772, '\p{Bidi_M=n}', "");
Expect(0, 120772, '\p{^Bidi_M=n}', "");
Expect(0, 120772, '\P{Bidi_M=n}', "");
Expect(1, 120772, '\P{^Bidi_M=n}', "");
Expect(0, 120771, '\p{Bidi_M=n}', "");
Expect(1, 120771, '\p{^Bidi_M=n}', "");
Expect(1, 120771, '\P{Bidi_M=n}', "");
Expect(0, 120771, '\P{^Bidi_M=n}', "");
Expect(1, 120772, '\p{Bidi_M=	_n}', "");
Expect(0, 120772, '\p{^Bidi_M=	_n}', "");
Expect(0, 120772, '\P{Bidi_M=	_n}', "");
Expect(1, 120772, '\P{^Bidi_M=	_n}', "");
Expect(0, 120771, '\p{Bidi_M=	_n}', "");
Expect(1, 120771, '\p{^Bidi_M=	_n}', "");
Expect(1, 120771, '\P{Bidi_M=	_n}', "");
Expect(0, 120771, '\P{^Bidi_M=	_n}', "");
Error('\p{Is_Bidi_Mirrored=/a/_ f}');
Error('\P{Is_Bidi_Mirrored=/a/_ f}');
Expect(1, 120772, '\p{Is_Bidi_Mirrored=f}', "");
Expect(0, 120772, '\p{^Is_Bidi_Mirrored=f}', "");
Expect(0, 120772, '\P{Is_Bidi_Mirrored=f}', "");
Expect(1, 120772, '\P{^Is_Bidi_Mirrored=f}', "");
Expect(0, 120771, '\p{Is_Bidi_Mirrored=f}', "");
Expect(1, 120771, '\p{^Is_Bidi_Mirrored=f}', "");
Expect(1, 120771, '\P{Is_Bidi_Mirrored=f}', "");
Expect(0, 120771, '\P{^Is_Bidi_Mirrored=f}', "");
Expect(1, 120772, '\p{Is_Bidi_Mirrored:	_	F}', "");
Expect(0, 120772, '\p{^Is_Bidi_Mirrored:	_	F}', "");
Expect(0, 120772, '\P{Is_Bidi_Mirrored:	_	F}', "");
Expect(1, 120772, '\P{^Is_Bidi_Mirrored:	_	F}', "");
Expect(0, 120771, '\p{Is_Bidi_Mirrored:	_	F}', "");
Expect(1, 120771, '\p{^Is_Bidi_Mirrored:	_	F}', "");
Expect(1, 120771, '\P{Is_Bidi_Mirrored:	_	F}', "");
Expect(0, 120771, '\P{^Is_Bidi_Mirrored:	_	F}', "");
Error('\p{Is_Bidi_M=-	False/a/}');
Error('\P{Is_Bidi_M=-	False/a/}');
Expect(1, 120772, '\p{Is_Bidi_M=false}', "");
Expect(0, 120772, '\p{^Is_Bidi_M=false}', "");
Expect(0, 120772, '\P{Is_Bidi_M=false}', "");
Expect(1, 120772, '\P{^Is_Bidi_M=false}', "");
Expect(0, 120771, '\p{Is_Bidi_M=false}', "");
Expect(1, 120771, '\p{^Is_Bidi_M=false}', "");
Expect(1, 120771, '\P{Is_Bidi_M=false}', "");
Expect(0, 120771, '\P{^Is_Bidi_M=false}', "");
Error('\p{Bidi_Mirrored=:= 	Yes}');
Error('\P{Bidi_Mirrored=:= 	Yes}');
Expect(1, 120771, '\p{Bidi_Mirrored=yes}', "");
Expect(0, 120771, '\p{^Bidi_Mirrored=yes}', "");
Expect(0, 120771, '\P{Bidi_Mirrored=yes}', "");
Expect(1, 120771, '\P{^Bidi_Mirrored=yes}', "");
Expect(0, 120772, '\p{Bidi_Mirrored=yes}', "");
Expect(1, 120772, '\p{^Bidi_Mirrored=yes}', "");
Expect(1, 120772, '\P{Bidi_Mirrored=yes}', "");
Expect(0, 120772, '\P{^Bidi_Mirrored=yes}', "");
Expect(1, 120771, '\p{Bidi_Mirrored=_ Yes}', "");
Expect(0, 120771, '\p{^Bidi_Mirrored=_ Yes}', "");
Expect(0, 120771, '\P{Bidi_Mirrored=_ Yes}', "");
Expect(1, 120771, '\P{^Bidi_Mirrored=_ Yes}', "");
Expect(0, 120772, '\p{Bidi_Mirrored=_ Yes}', "");
Expect(1, 120772, '\p{^Bidi_Mirrored=_ Yes}', "");
Expect(1, 120772, '\P{Bidi_Mirrored=_ Yes}', "");
Expect(0, 120772, '\P{^Bidi_Mirrored=_ Yes}', "");
Error('\p{Bidi_M=_Y/a/}');
Error('\P{Bidi_M=_Y/a/}');
Expect(1, 120771, '\p{Bidi_M=y}', "");
Expect(0, 120771, '\p{^Bidi_M=y}', "");
Expect(0, 120771, '\P{Bidi_M=y}', "");
Expect(1, 120771, '\P{^Bidi_M=y}', "");
Expect(0, 120772, '\p{Bidi_M=y}', "");
Expect(1, 120772, '\p{^Bidi_M=y}', "");
Expect(1, 120772, '\P{Bidi_M=y}', "");
Expect(0, 120772, '\P{^Bidi_M=y}', "");
Expect(1, 120771, '\p{Bidi_M= _y}', "");
Expect(0, 120771, '\p{^Bidi_M= _y}', "");
Expect(0, 120771, '\P{Bidi_M= _y}', "");
Expect(1, 120771, '\P{^Bidi_M= _y}', "");
Expect(0, 120772, '\p{Bidi_M= _y}', "");
Expect(1, 120772, '\p{^Bidi_M= _y}', "");
Expect(1, 120772, '\P{Bidi_M= _y}', "");
Expect(0, 120772, '\P{^Bidi_M= _y}', "");
Error('\p{Is_Bidi_Mirrored=:=_-T}');
Error('\P{Is_Bidi_Mirrored=:=_-T}');
Expect(1, 120771, '\p{Is_Bidi_Mirrored=t}', "");
Expect(0, 120771, '\p{^Is_Bidi_Mirrored=t}', "");
Expect(0, 120771, '\P{Is_Bidi_Mirrored=t}', "");
Expect(1, 120771, '\P{^Is_Bidi_Mirrored=t}', "");
Expect(0, 120772, '\p{Is_Bidi_Mirrored=t}', "");
Expect(1, 120772, '\p{^Is_Bidi_Mirrored=t}', "");
Expect(1, 120772, '\P{Is_Bidi_Mirrored=t}', "");
Expect(0, 120772, '\P{^Is_Bidi_Mirrored=t}', "");
Expect(1, 120771, '\p{Is_Bidi_Mirrored=_-T}', "");
Expect(0, 120771, '\p{^Is_Bidi_Mirrored=_-T}', "");
Expect(0, 120771, '\P{Is_Bidi_Mirrored=_-T}', "");
Expect(1, 120771, '\P{^Is_Bidi_Mirrored=_-T}', "");
Expect(0, 120772, '\p{Is_Bidi_Mirrored=_-T}', "");
Expect(1, 120772, '\p{^Is_Bidi_Mirrored=_-T}', "");
Expect(1, 120772, '\P{Is_Bidi_Mirrored=_-T}', "");
Expect(0, 120772, '\P{^Is_Bidi_Mirrored=_-T}', "");
Error('\p{Is_Bidi_M=:= _TRUE}');
Error('\P{Is_Bidi_M=:= _TRUE}');
Expect(1, 120771, '\p{Is_Bidi_M=true}', "");
Expect(0, 120771, '\p{^Is_Bidi_M=true}', "");
Expect(0, 120771, '\P{Is_Bidi_M=true}', "");
Expect(1, 120771, '\P{^Is_Bidi_M=true}', "");
Expect(0, 120772, '\p{Is_Bidi_M=true}', "");
Expect(1, 120772, '\p{^Is_Bidi_M=true}', "");
Expect(1, 120772, '\P{Is_Bidi_M=true}', "");
Expect(0, 120772, '\P{^Is_Bidi_M=true}', "");
Expect(1, 120771, '\p{Is_Bidi_M=-	True}', "");
Expect(0, 120771, '\p{^Is_Bidi_M=-	True}', "");
Expect(0, 120771, '\P{Is_Bidi_M=-	True}', "");
Expect(1, 120771, '\P{^Is_Bidi_M=-	True}', "");
Expect(0, 120772, '\p{Is_Bidi_M=-	True}', "");
Expect(1, 120772, '\p{^Is_Bidi_M=-	True}', "");
Expect(1, 120772, '\P{Is_Bidi_M=-	True}', "");
Expect(0, 120772, '\P{^Is_Bidi_M=-	True}', "");
Error('\p{Composition_Exclusion=	-No:=}');
Error('\P{Composition_Exclusion=	-No:=}');
Expect(1, 119233, '\p{Composition_Exclusion=no}', "");
Expect(0, 119233, '\p{^Composition_Exclusion=no}', "");
Expect(0, 119233, '\P{Composition_Exclusion=no}', "");
Expect(1, 119233, '\P{^Composition_Exclusion=no}', "");
Expect(0, 119232, '\p{Composition_Exclusion=no}', "");
Expect(1, 119232, '\p{^Composition_Exclusion=no}', "");
Expect(1, 119232, '\P{Composition_Exclusion=no}', "");
Expect(0, 119232, '\P{^Composition_Exclusion=no}', "");
Expect(1, 119233, '\p{Composition_Exclusion:_NO}', "");
Expect(0, 119233, '\p{^Composition_Exclusion:_NO}', "");
Expect(0, 119233, '\P{Composition_Exclusion:_NO}', "");
Expect(1, 119233, '\P{^Composition_Exclusion:_NO}', "");
Expect(0, 119232, '\p{Composition_Exclusion:_NO}', "");
Expect(1, 119232, '\p{^Composition_Exclusion:_NO}', "");
Expect(1, 119232, '\P{Composition_Exclusion:_NO}', "");
Expect(0, 119232, '\P{^Composition_Exclusion:_NO}', "");
Error('\p{CE=-_N:=}');
Error('\P{CE=-_N:=}');
Expect(1, 119233, '\p{CE=n}', "");
Expect(0, 119233, '\p{^CE=n}', "");
Expect(0, 119233, '\P{CE=n}', "");
Expect(1, 119233, '\P{^CE=n}', "");
Expect(0, 119232, '\p{CE=n}', "");
Expect(1, 119232, '\p{^CE=n}', "");
Expect(1, 119232, '\P{CE=n}', "");
Expect(0, 119232, '\P{^CE=n}', "");
Expect(1, 119233, '\p{CE= N}', "");
Expect(0, 119233, '\p{^CE= N}', "");
Expect(0, 119233, '\P{CE= N}', "");
Expect(1, 119233, '\P{^CE= N}', "");
Expect(0, 119232, '\p{CE= N}', "");
Expect(1, 119232, '\p{^CE= N}', "");
Expect(1, 119232, '\P{CE= N}', "");
Expect(0, 119232, '\P{^CE= N}', "");
Error('\p{Is_Composition_Exclusion=-F:=}');
Error('\P{Is_Composition_Exclusion=-F:=}');
Expect(1, 119233, '\p{Is_Composition_Exclusion=f}', "");
Expect(0, 119233, '\p{^Is_Composition_Exclusion=f}', "");
Expect(0, 119233, '\P{Is_Composition_Exclusion=f}', "");
Expect(1, 119233, '\P{^Is_Composition_Exclusion=f}', "");
Expect(0, 119232, '\p{Is_Composition_Exclusion=f}', "");
Expect(1, 119232, '\p{^Is_Composition_Exclusion=f}', "");
Expect(1, 119232, '\P{Is_Composition_Exclusion=f}', "");
Expect(0, 119232, '\P{^Is_Composition_Exclusion=f}', "");
Expect(1, 119233, '\p{Is_Composition_Exclusion=F}', "");
Expect(0, 119233, '\p{^Is_Composition_Exclusion=F}', "");
Expect(0, 119233, '\P{Is_Composition_Exclusion=F}', "");
Expect(1, 119233, '\P{^Is_Composition_Exclusion=F}', "");
Expect(0, 119232, '\p{Is_Composition_Exclusion=F}', "");
Expect(1, 119232, '\p{^Is_Composition_Exclusion=F}', "");
Expect(1, 119232, '\P{Is_Composition_Exclusion=F}', "");
Expect(0, 119232, '\P{^Is_Composition_Exclusion=F}', "");
Error('\p{Is_CE=:=  false}');
Error('\P{Is_CE=:=  false}');
Expect(1, 119233, '\p{Is_CE=false}', "");
Expect(0, 119233, '\p{^Is_CE=false}', "");
Expect(0, 119233, '\P{Is_CE=false}', "");
Expect(1, 119233, '\P{^Is_CE=false}', "");
Expect(0, 119232, '\p{Is_CE=false}', "");
Expect(1, 119232, '\p{^Is_CE=false}', "");
Expect(1, 119232, '\P{Is_CE=false}', "");
Expect(0, 119232, '\P{^Is_CE=false}', "");
Expect(1, 119233, '\p{Is_CE=False}', "");
Expect(0, 119233, '\p{^Is_CE=False}', "");
Expect(0, 119233, '\P{Is_CE=False}', "");
Expect(1, 119233, '\P{^Is_CE=False}', "");
Expect(0, 119232, '\p{Is_CE=False}', "");
Expect(1, 119232, '\p{^Is_CE=False}', "");
Expect(1, 119232, '\P{Is_CE=False}', "");
Expect(0, 119232, '\P{^Is_CE=False}', "");
Error('\p{Composition_Exclusion= :=yes}');
Error('\P{Composition_Exclusion= :=yes}');
Expect(1, 119232, '\p{Composition_Exclusion=yes}', "");
Expect(0, 119232, '\p{^Composition_Exclusion=yes}', "");
Expect(0, 119232, '\P{Composition_Exclusion=yes}', "");
Expect(1, 119232, '\P{^Composition_Exclusion=yes}', "");
Expect(0, 119233, '\p{Composition_Exclusion=yes}', "");
Expect(1, 119233, '\p{^Composition_Exclusion=yes}', "");
Expect(1, 119233, '\P{Composition_Exclusion=yes}', "");
Expect(0, 119233, '\P{^Composition_Exclusion=yes}', "");
Expect(1, 119232, '\p{Composition_Exclusion=-YES}', "");
Expect(0, 119232, '\p{^Composition_Exclusion=-YES}', "");
Expect(0, 119232, '\P{Composition_Exclusion=-YES}', "");
Expect(1, 119232, '\P{^Composition_Exclusion=-YES}', "");
Expect(0, 119233, '\p{Composition_Exclusion=-YES}', "");
Expect(1, 119233, '\p{^Composition_Exclusion=-YES}', "");
Expect(1, 119233, '\P{Composition_Exclusion=-YES}', "");
Expect(0, 119233, '\P{^Composition_Exclusion=-YES}', "");
Error('\p{CE=	_Y:=}');
Error('\P{CE=	_Y:=}');
Expect(1, 119232, '\p{CE=y}', "");
Expect(0, 119232, '\p{^CE=y}', "");
Expect(0, 119232, '\P{CE=y}', "");
Expect(1, 119232, '\P{^CE=y}', "");
Expect(0, 119233, '\p{CE=y}', "");
Expect(1, 119233, '\p{^CE=y}', "");
Expect(1, 119233, '\P{CE=y}', "");
Expect(0, 119233, '\P{^CE=y}', "");
Expect(1, 119232, '\p{CE=_-Y}', "");
Expect(0, 119232, '\p{^CE=_-Y}', "");
Expect(0, 119232, '\P{CE=_-Y}', "");
Expect(1, 119232, '\P{^CE=_-Y}', "");
Expect(0, 119233, '\p{CE=_-Y}', "");
Expect(1, 119233, '\p{^CE=_-Y}', "");
Expect(1, 119233, '\P{CE=_-Y}', "");
Expect(0, 119233, '\P{^CE=_-Y}', "");
Error('\p{Is_Composition_Exclusion:/a/T}');
Error('\P{Is_Composition_Exclusion:/a/T}');
Expect(1, 119232, '\p{Is_Composition_Exclusion=t}', "");
Expect(0, 119232, '\p{^Is_Composition_Exclusion=t}', "");
Expect(0, 119232, '\P{Is_Composition_Exclusion=t}', "");
Expect(1, 119232, '\P{^Is_Composition_Exclusion=t}', "");
Expect(0, 119233, '\p{Is_Composition_Exclusion=t}', "");
Expect(1, 119233, '\p{^Is_Composition_Exclusion=t}', "");
Expect(1, 119233, '\P{Is_Composition_Exclusion=t}', "");
Expect(0, 119233, '\P{^Is_Composition_Exclusion=t}', "");
Expect(1, 119232, '\p{Is_Composition_Exclusion=	_T}', "");
Expect(0, 119232, '\p{^Is_Composition_Exclusion=	_T}', "");
Expect(0, 119232, '\P{Is_Composition_Exclusion=	_T}', "");
Expect(1, 119232, '\P{^Is_Composition_Exclusion=	_T}', "");
Expect(0, 119233, '\p{Is_Composition_Exclusion=	_T}', "");
Expect(1, 119233, '\p{^Is_Composition_Exclusion=	_T}', "");
Expect(1, 119233, '\P{Is_Composition_Exclusion=	_T}', "");
Expect(0, 119233, '\P{^Is_Composition_Exclusion=	_T}', "");
Error('\p{Is_CE=:=-true}');
Error('\P{Is_CE=:=-true}');
Expect(1, 119232, '\p{Is_CE:true}', "");
Expect(0, 119232, '\p{^Is_CE:true}', "");
Expect(0, 119232, '\P{Is_CE:true}', "");
Expect(1, 119232, '\P{^Is_CE:true}', "");
Expect(0, 119233, '\p{Is_CE:true}', "");
Expect(1, 119233, '\p{^Is_CE:true}', "");
Expect(1, 119233, '\P{Is_CE:true}', "");
Expect(0, 119233, '\P{^Is_CE:true}', "");
Expect(1, 119232, '\p{Is_CE=  TRUE}', "");
Expect(0, 119232, '\p{^Is_CE=  TRUE}', "");
Expect(0, 119232, '\P{Is_CE=  TRUE}', "");
Expect(1, 119232, '\P{^Is_CE=  TRUE}', "");
Expect(0, 119233, '\p{Is_CE=  TRUE}', "");
Expect(1, 119233, '\p{^Is_CE=  TRUE}', "");
Expect(1, 119233, '\P{Is_CE=  TRUE}', "");
Expect(0, 119233, '\P{^Is_CE=  TRUE}', "");
Error('\p{Case_Ignorable=/a/	_No}');
Error('\P{Case_Ignorable=/a/	_No}');
Expect(1, 918000, '\p{Case_Ignorable=no}', "");
Expect(0, 918000, '\p{^Case_Ignorable=no}', "");
Expect(0, 918000, '\P{Case_Ignorable=no}', "");
Expect(1, 918000, '\P{^Case_Ignorable=no}', "");
Expect(0, 917999, '\p{Case_Ignorable=no}', "");
Expect(1, 917999, '\p{^Case_Ignorable=no}', "");
Expect(1, 917999, '\P{Case_Ignorable=no}', "");
Expect(0, 917999, '\P{^Case_Ignorable=no}', "");
Expect(1, 918000, '\p{Case_Ignorable=NO}', "");
Expect(0, 918000, '\p{^Case_Ignorable=NO}', "");
Expect(0, 918000, '\P{Case_Ignorable=NO}', "");
Expect(1, 918000, '\P{^Case_Ignorable=NO}', "");
Expect(0, 917999, '\p{Case_Ignorable=NO}', "");
Expect(1, 917999, '\p{^Case_Ignorable=NO}', "");
Expect(1, 917999, '\P{Case_Ignorable=NO}', "");
Expect(0, 917999, '\P{^Case_Ignorable=NO}', "");
Error('\p{CI=/a/_	N}');
Error('\P{CI=/a/_	N}');
Expect(1, 918000, '\p{CI=n}', "");
Expect(0, 918000, '\p{^CI=n}', "");
Expect(0, 918000, '\P{CI=n}', "");
Expect(1, 918000, '\P{^CI=n}', "");
Expect(0, 917999, '\p{CI=n}', "");
Expect(1, 917999, '\p{^CI=n}', "");
Expect(1, 917999, '\P{CI=n}', "");
Expect(0, 917999, '\P{^CI=n}', "");
Expect(1, 918000, '\p{CI=	N}', "");
Expect(0, 918000, '\p{^CI=	N}', "");
Expect(0, 918000, '\P{CI=	N}', "");
Expect(1, 918000, '\P{^CI=	N}', "");
Expect(0, 917999, '\p{CI=	N}', "");
Expect(1, 917999, '\p{^CI=	N}', "");
Expect(1, 917999, '\P{CI=	N}', "");
Expect(0, 917999, '\P{^CI=	N}', "");
Error('\p{Is_Case_Ignorable= 	F/a/}');
Error('\P{Is_Case_Ignorable= 	F/a/}');
Expect(1, 918000, '\p{Is_Case_Ignorable=f}', "");
Expect(0, 918000, '\p{^Is_Case_Ignorable=f}', "");
Expect(0, 918000, '\P{Is_Case_Ignorable=f}', "");
Expect(1, 918000, '\P{^Is_Case_Ignorable=f}', "");
Expect(0, 917999, '\p{Is_Case_Ignorable=f}', "");
Expect(1, 917999, '\p{^Is_Case_Ignorable=f}', "");
Expect(1, 917999, '\P{Is_Case_Ignorable=f}', "");
Expect(0, 917999, '\P{^Is_Case_Ignorable=f}', "");
Expect(1, 918000, '\p{Is_Case_Ignorable= -f}', "");
Expect(0, 918000, '\p{^Is_Case_Ignorable= -f}', "");
Expect(0, 918000, '\P{Is_Case_Ignorable= -f}', "");
Expect(1, 918000, '\P{^Is_Case_Ignorable= -f}', "");
Expect(0, 917999, '\p{Is_Case_Ignorable= -f}', "");
Expect(1, 917999, '\p{^Is_Case_Ignorable= -f}', "");
Expect(1, 917999, '\P{Is_Case_Ignorable= -f}', "");
Expect(0, 917999, '\P{^Is_Case_Ignorable= -f}', "");
Error('\p{Is_CI:	:=false}');
Error('\P{Is_CI:	:=false}');
Expect(1, 918000, '\p{Is_CI=false}', "");
Expect(0, 918000, '\p{^Is_CI=false}', "");
Expect(0, 918000, '\P{Is_CI=false}', "");
Expect(1, 918000, '\P{^Is_CI=false}', "");
Expect(0, 917999, '\p{Is_CI=false}', "");
Expect(1, 917999, '\p{^Is_CI=false}', "");
Expect(1, 917999, '\P{Is_CI=false}', "");
Expect(0, 917999, '\P{^Is_CI=false}', "");
Expect(1, 918000, '\p{Is_CI:_ FALSE}', "");
Expect(0, 918000, '\p{^Is_CI:_ FALSE}', "");
Expect(0, 918000, '\P{Is_CI:_ FALSE}', "");
Expect(1, 918000, '\P{^Is_CI:_ FALSE}', "");
Expect(0, 917999, '\p{Is_CI:_ FALSE}', "");
Expect(1, 917999, '\p{^Is_CI:_ FALSE}', "");
Expect(1, 917999, '\P{Is_CI:_ FALSE}', "");
Expect(0, 917999, '\P{^Is_CI:_ FALSE}', "");
Error('\p{Case_Ignorable:/a/ -yes}');
Error('\P{Case_Ignorable:/a/ -yes}');
Expect(1, 917999, '\p{Case_Ignorable:   yes}', "");
Expect(0, 917999, '\p{^Case_Ignorable:   yes}', "");
Expect(0, 917999, '\P{Case_Ignorable:   yes}', "");
Expect(1, 917999, '\P{^Case_Ignorable:   yes}', "");
Expect(0, 918000, '\p{Case_Ignorable:   yes}', "");
Expect(1, 918000, '\p{^Case_Ignorable:   yes}', "");
Expect(1, 918000, '\P{Case_Ignorable:   yes}', "");
Expect(0, 918000, '\P{^Case_Ignorable:   yes}', "");
Expect(1, 917999, '\p{Case_Ignorable:	_	Yes}', "");
Expect(0, 917999, '\p{^Case_Ignorable:	_	Yes}', "");
Expect(0, 917999, '\P{Case_Ignorable:	_	Yes}', "");
Expect(1, 917999, '\P{^Case_Ignorable:	_	Yes}', "");
Expect(0, 918000, '\p{Case_Ignorable:	_	Yes}', "");
Expect(1, 918000, '\p{^Case_Ignorable:	_	Yes}', "");
Expect(1, 918000, '\P{Case_Ignorable:	_	Yes}', "");
Expect(0, 918000, '\P{^Case_Ignorable:	_	Yes}', "");
Error('\p{CI=	:=Y}');
Error('\P{CI=	:=Y}');
Expect(1, 917999, '\p{CI=y}', "");
Expect(0, 917999, '\p{^CI=y}', "");
Expect(0, 917999, '\P{CI=y}', "");
Expect(1, 917999, '\P{^CI=y}', "");
Expect(0, 918000, '\p{CI=y}', "");
Expect(1, 918000, '\p{^CI=y}', "");
Expect(1, 918000, '\P{CI=y}', "");
Expect(0, 918000, '\P{^CI=y}', "");
Expect(1, 917999, '\p{CI= _Y}', "");
Expect(0, 917999, '\p{^CI= _Y}', "");
Expect(0, 917999, '\P{CI= _Y}', "");
Expect(1, 917999, '\P{^CI= _Y}', "");
Expect(0, 918000, '\p{CI= _Y}', "");
Expect(1, 918000, '\p{^CI= _Y}', "");
Expect(1, 918000, '\P{CI= _Y}', "");
Expect(0, 918000, '\P{^CI= _Y}', "");
Error('\p{Is_Case_Ignorable=_T:=}');
Error('\P{Is_Case_Ignorable=_T:=}');
Expect(1, 917999, '\p{Is_Case_Ignorable=t}', "");
Expect(0, 917999, '\p{^Is_Case_Ignorable=t}', "");
Expect(0, 917999, '\P{Is_Case_Ignorable=t}', "");
Expect(1, 917999, '\P{^Is_Case_Ignorable=t}', "");
Expect(0, 918000, '\p{Is_Case_Ignorable=t}', "");
Expect(1, 918000, '\p{^Is_Case_Ignorable=t}', "");
Expect(1, 918000, '\P{Is_Case_Ignorable=t}', "");
Expect(0, 918000, '\P{^Is_Case_Ignorable=t}', "");
Expect(1, 917999, '\p{Is_Case_Ignorable=T}', "");
Expect(0, 917999, '\p{^Is_Case_Ignorable=T}', "");
Expect(0, 917999, '\P{Is_Case_Ignorable=T}', "");
Expect(1, 917999, '\P{^Is_Case_Ignorable=T}', "");
Expect(0, 918000, '\p{Is_Case_Ignorable=T}', "");
Expect(1, 918000, '\p{^Is_Case_Ignorable=T}', "");
Expect(1, 918000, '\P{Is_Case_Ignorable=T}', "");
Expect(0, 918000, '\P{^Is_Case_Ignorable=T}', "");
Error('\p{Is_CI=/a/--true}');
Error('\P{Is_CI=/a/--true}');
Expect(1, 917999, '\p{Is_CI=true}', "");
Expect(0, 917999, '\p{^Is_CI=true}', "");
Expect(0, 917999, '\P{Is_CI=true}', "");
Expect(1, 917999, '\P{^Is_CI=true}', "");
Expect(0, 918000, '\p{Is_CI=true}', "");
Expect(1, 918000, '\p{^Is_CI=true}', "");
Expect(1, 918000, '\P{Is_CI=true}', "");
Expect(0, 918000, '\P{^Is_CI=true}', "");
Expect(1, 917999, '\p{Is_CI=	 True}', "");
Expect(0, 917999, '\p{^Is_CI=	 True}', "");
Expect(0, 917999, '\P{Is_CI=	 True}', "");
Expect(1, 917999, '\P{^Is_CI=	 True}', "");
Expect(0, 918000, '\p{Is_CI=	 True}', "");
Expect(1, 918000, '\p{^Is_CI=	 True}', "");
Expect(1, 918000, '\P{Is_CI=	 True}', "");
Expect(0, 918000, '\P{^Is_CI=	 True}', "");
Error('\p{Changes_When_Casefolded=	No:=}');
Error('\P{Changes_When_Casefolded=	No:=}');
Expect(1, 66600, '\p{Changes_When_Casefolded=no}', "");
Expect(0, 66600, '\p{^Changes_When_Casefolded=no}', "");
Expect(0, 66600, '\P{Changes_When_Casefolded=no}', "");
Expect(1, 66600, '\P{^Changes_When_Casefolded=no}', "");
Expect(0, 66599, '\p{Changes_When_Casefolded=no}', "");
Expect(1, 66599, '\p{^Changes_When_Casefolded=no}', "");
Expect(1, 66599, '\P{Changes_When_Casefolded=no}', "");
Expect(0, 66599, '\P{^Changes_When_Casefolded=no}', "");
Expect(1, 66600, '\p{Changes_When_Casefolded= _No}', "");
Expect(0, 66600, '\p{^Changes_When_Casefolded= _No}', "");
Expect(0, 66600, '\P{Changes_When_Casefolded= _No}', "");
Expect(1, 66600, '\P{^Changes_When_Casefolded= _No}', "");
Expect(0, 66599, '\p{Changes_When_Casefolded= _No}', "");
Expect(1, 66599, '\p{^Changes_When_Casefolded= _No}', "");
Expect(1, 66599, '\P{Changes_When_Casefolded= _No}', "");
Expect(0, 66599, '\P{^Changes_When_Casefolded= _No}', "");
Error('\p{CWCF:	n/a/}');
Error('\P{CWCF:	n/a/}');
Expect(1, 66600, '\p{CWCF=n}', "");
Expect(0, 66600, '\p{^CWCF=n}', "");
Expect(0, 66600, '\P{CWCF=n}', "");
Expect(1, 66600, '\P{^CWCF=n}', "");
Expect(0, 66599, '\p{CWCF=n}', "");
Expect(1, 66599, '\p{^CWCF=n}', "");
Expect(1, 66599, '\P{CWCF=n}', "");
Expect(0, 66599, '\P{^CWCF=n}', "");
Expect(1, 66600, '\p{CWCF=_-N}', "");
Expect(0, 66600, '\p{^CWCF=_-N}', "");
Expect(0, 66600, '\P{CWCF=_-N}', "");
Expect(1, 66600, '\P{^CWCF=_-N}', "");
Expect(0, 66599, '\p{CWCF=_-N}', "");
Expect(1, 66599, '\p{^CWCF=_-N}', "");
Expect(1, 66599, '\P{CWCF=_-N}', "");
Expect(0, 66599, '\P{^CWCF=_-N}', "");
Error('\p{Is_Changes_When_Casefolded=/a/_f}');
Error('\P{Is_Changes_When_Casefolded=/a/_f}');
Expect(1, 66600, '\p{Is_Changes_When_Casefolded=f}', "");
Expect(0, 66600, '\p{^Is_Changes_When_Casefolded=f}', "");
Expect(0, 66600, '\P{Is_Changes_When_Casefolded=f}', "");
Expect(1, 66600, '\P{^Is_Changes_When_Casefolded=f}', "");
Expect(0, 66599, '\p{Is_Changes_When_Casefolded=f}', "");
Expect(1, 66599, '\p{^Is_Changes_When_Casefolded=f}', "");
Expect(1, 66599, '\P{Is_Changes_When_Casefolded=f}', "");
Expect(0, 66599, '\P{^Is_Changes_When_Casefolded=f}', "");
Expect(1, 66600, '\p{Is_Changes_When_Casefolded=_-F}', "");
Expect(0, 66600, '\p{^Is_Changes_When_Casefolded=_-F}', "");
Expect(0, 66600, '\P{Is_Changes_When_Casefolded=_-F}', "");
Expect(1, 66600, '\P{^Is_Changes_When_Casefolded=_-F}', "");
Expect(0, 66599, '\p{Is_Changes_When_Casefolded=_-F}', "");
Expect(1, 66599, '\p{^Is_Changes_When_Casefolded=_-F}', "");
Expect(1, 66599, '\P{Is_Changes_When_Casefolded=_-F}', "");
Expect(0, 66599, '\P{^Is_Changes_When_Casefolded=_-F}', "");
Error('\p{Is_CWCF=	/a/False}');
Error('\P{Is_CWCF=	/a/False}');
Expect(1, 66600, '\p{Is_CWCF=false}', "");
Expect(0, 66600, '\p{^Is_CWCF=false}', "");
Expect(0, 66600, '\P{Is_CWCF=false}', "");
Expect(1, 66600, '\P{^Is_CWCF=false}', "");
Expect(0, 66599, '\p{Is_CWCF=false}', "");
Expect(1, 66599, '\p{^Is_CWCF=false}', "");
Expect(1, 66599, '\P{Is_CWCF=false}', "");
Expect(0, 66599, '\P{^Is_CWCF=false}', "");
Expect(1, 66600, '\p{Is_CWCF=--FALSE}', "");
Expect(0, 66600, '\p{^Is_CWCF=--FALSE}', "");
Expect(0, 66600, '\P{Is_CWCF=--FALSE}', "");
Expect(1, 66600, '\P{^Is_CWCF=--FALSE}', "");
Expect(0, 66599, '\p{Is_CWCF=--FALSE}', "");
Expect(1, 66599, '\p{^Is_CWCF=--FALSE}', "");
Expect(1, 66599, '\P{Is_CWCF=--FALSE}', "");
Expect(0, 66599, '\P{^Is_CWCF=--FALSE}', "");
Error('\p{Changes_When_Casefolded=_YES:=}');
Error('\P{Changes_When_Casefolded=_YES:=}');
Expect(1, 66599, '\p{Changes_When_Casefolded=yes}', "");
Expect(0, 66599, '\p{^Changes_When_Casefolded=yes}', "");
Expect(0, 66599, '\P{Changes_When_Casefolded=yes}', "");
Expect(1, 66599, '\P{^Changes_When_Casefolded=yes}', "");
Expect(0, 66600, '\p{Changes_When_Casefolded=yes}', "");
Expect(1, 66600, '\p{^Changes_When_Casefolded=yes}', "");
Expect(1, 66600, '\P{Changes_When_Casefolded=yes}', "");
Expect(0, 66600, '\P{^Changes_When_Casefolded=yes}', "");
Expect(1, 66599, '\p{Changes_When_Casefolded= _Yes}', "");
Expect(0, 66599, '\p{^Changes_When_Casefolded= _Yes}', "");
Expect(0, 66599, '\P{Changes_When_Casefolded= _Yes}', "");
Expect(1, 66599, '\P{^Changes_When_Casefolded= _Yes}', "");
Expect(0, 66600, '\p{Changes_When_Casefolded= _Yes}', "");
Expect(1, 66600, '\p{^Changes_When_Casefolded= _Yes}', "");
Expect(1, 66600, '\P{Changes_When_Casefolded= _Yes}', "");
Expect(0, 66600, '\P{^Changes_When_Casefolded= _Yes}', "");
Error('\p{CWCF= /a/Y}');
Error('\P{CWCF= /a/Y}');
Expect(1, 66599, '\p{CWCF=y}', "");
Expect(0, 66599, '\p{^CWCF=y}', "");
Expect(0, 66599, '\P{CWCF=y}', "");
Expect(1, 66599, '\P{^CWCF=y}', "");
Expect(0, 66600, '\p{CWCF=y}', "");
Expect(1, 66600, '\p{^CWCF=y}', "");
Expect(1, 66600, '\P{CWCF=y}', "");
Expect(0, 66600, '\P{^CWCF=y}', "");
Expect(1, 66599, '\p{CWCF= 	Y}', "");
Expect(0, 66599, '\p{^CWCF= 	Y}', "");
Expect(0, 66599, '\P{CWCF= 	Y}', "");
Expect(1, 66599, '\P{^CWCF= 	Y}', "");
Expect(0, 66600, '\p{CWCF= 	Y}', "");
Expect(1, 66600, '\p{^CWCF= 	Y}', "");
Expect(1, 66600, '\P{CWCF= 	Y}', "");
Expect(0, 66600, '\P{^CWCF= 	Y}', "");
Error('\p{Is_Changes_When_Casefolded=_-T/a/}');
Error('\P{Is_Changes_When_Casefolded=_-T/a/}');
Expect(1, 66599, '\p{Is_Changes_When_Casefolded=t}', "");
Expect(0, 66599, '\p{^Is_Changes_When_Casefolded=t}', "");
Expect(0, 66599, '\P{Is_Changes_When_Casefolded=t}', "");
Expect(1, 66599, '\P{^Is_Changes_When_Casefolded=t}', "");
Expect(0, 66600, '\p{Is_Changes_When_Casefolded=t}', "");
Expect(1, 66600, '\p{^Is_Changes_When_Casefolded=t}', "");
Expect(1, 66600, '\P{Is_Changes_When_Casefolded=t}', "");
Expect(0, 66600, '\P{^Is_Changes_When_Casefolded=t}', "");
Expect(1, 66599, '\p{Is_Changes_When_Casefolded=-T}', "");
Expect(0, 66599, '\p{^Is_Changes_When_Casefolded=-T}', "");
Expect(0, 66599, '\P{Is_Changes_When_Casefolded=-T}', "");
Expect(1, 66599, '\P{^Is_Changes_When_Casefolded=-T}', "");
Expect(0, 66600, '\p{Is_Changes_When_Casefolded=-T}', "");
Expect(1, 66600, '\p{^Is_Changes_When_Casefolded=-T}', "");
Expect(1, 66600, '\P{Is_Changes_When_Casefolded=-T}', "");
Expect(0, 66600, '\P{^Is_Changes_When_Casefolded=-T}', "");
Error('\p{Is_CWCF=	True/a/}');
Error('\P{Is_CWCF=	True/a/}');
Expect(1, 66599, '\p{Is_CWCF=true}', "");
Expect(0, 66599, '\p{^Is_CWCF=true}', "");
Expect(0, 66599, '\P{Is_CWCF=true}', "");
Expect(1, 66599, '\P{^Is_CWCF=true}', "");
Expect(0, 66600, '\p{Is_CWCF=true}', "");
Expect(1, 66600, '\p{^Is_CWCF=true}', "");
Expect(1, 66600, '\P{Is_CWCF=true}', "");
Expect(0, 66600, '\P{^Is_CWCF=true}', "");
Expect(1, 66599, '\p{Is_CWCF= True}', "");
Expect(0, 66599, '\p{^Is_CWCF= True}', "");
Expect(0, 66599, '\P{Is_CWCF= True}', "");
Expect(1, 66599, '\P{^Is_CWCF= True}', "");
Expect(0, 66600, '\p{Is_CWCF= True}', "");
Expect(1, 66600, '\p{^Is_CWCF= True}', "");
Expect(1, 66600, '\P{Is_CWCF= True}', "");
Expect(0, 66600, '\P{^Is_CWCF= True}', "");
Error('\p{Changes_When_Casemapped=	 no/a/}');
Error('\P{Changes_When_Casemapped=	 no/a/}');
Expect(1, 66640, '\p{Changes_When_Casemapped=no}', "");
Expect(0, 66640, '\p{^Changes_When_Casemapped=no}', "");
Expect(0, 66640, '\P{Changes_When_Casemapped=no}', "");
Expect(1, 66640, '\P{^Changes_When_Casemapped=no}', "");
Expect(0, 66639, '\p{Changes_When_Casemapped=no}', "");
Expect(1, 66639, '\p{^Changes_When_Casemapped=no}', "");
Expect(1, 66639, '\P{Changes_When_Casemapped=no}', "");
Expect(0, 66639, '\P{^Changes_When_Casemapped=no}', "");
Expect(1, 66640, '\p{Changes_When_Casemapped:   	_No}', "");
Expect(0, 66640, '\p{^Changes_When_Casemapped:   	_No}', "");
Expect(0, 66640, '\P{Changes_When_Casemapped:   	_No}', "");
Expect(1, 66640, '\P{^Changes_When_Casemapped:   	_No}', "");
Expect(0, 66639, '\p{Changes_When_Casemapped:   	_No}', "");
Expect(1, 66639, '\p{^Changes_When_Casemapped:   	_No}', "");
Expect(1, 66639, '\P{Changes_When_Casemapped:   	_No}', "");
Expect(0, 66639, '\P{^Changes_When_Casemapped:   	_No}', "");
Error('\p{CWCM=n:=}');
Error('\P{CWCM=n:=}');
Expect(1, 66640, '\p{CWCM=n}', "");
Expect(0, 66640, '\p{^CWCM=n}', "");
Expect(0, 66640, '\P{CWCM=n}', "");
Expect(1, 66640, '\P{^CWCM=n}', "");
Expect(0, 66639, '\p{CWCM=n}', "");
Expect(1, 66639, '\p{^CWCM=n}', "");
Expect(1, 66639, '\P{CWCM=n}', "");
Expect(0, 66639, '\P{^CWCM=n}', "");
Expect(1, 66640, '\p{CWCM=_N}', "");
Expect(0, 66640, '\p{^CWCM=_N}', "");
Expect(0, 66640, '\P{CWCM=_N}', "");
Expect(1, 66640, '\P{^CWCM=_N}', "");
Expect(0, 66639, '\p{CWCM=_N}', "");
Expect(1, 66639, '\p{^CWCM=_N}', "");
Expect(1, 66639, '\P{CWCM=_N}', "");
Expect(0, 66639, '\P{^CWCM=_N}', "");
Error('\p{Is_Changes_When_Casemapped:   /a/-_F}');
Error('\P{Is_Changes_When_Casemapped:   /a/-_F}');
Expect(1, 66640, '\p{Is_Changes_When_Casemapped=f}', "");
Expect(0, 66640, '\p{^Is_Changes_When_Casemapped=f}', "");
Expect(0, 66640, '\P{Is_Changes_When_Casemapped=f}', "");
Expect(1, 66640, '\P{^Is_Changes_When_Casemapped=f}', "");
Expect(0, 66639, '\p{Is_Changes_When_Casemapped=f}', "");
Expect(1, 66639, '\p{^Is_Changes_When_Casemapped=f}', "");
Expect(1, 66639, '\P{Is_Changes_When_Casemapped=f}', "");
Expect(0, 66639, '\P{^Is_Changes_When_Casemapped=f}', "");
Expect(1, 66640, '\p{Is_Changes_When_Casemapped=		F}', "");
Expect(0, 66640, '\p{^Is_Changes_When_Casemapped=		F}', "");
Expect(0, 66640, '\P{Is_Changes_When_Casemapped=		F}', "");
Expect(1, 66640, '\P{^Is_Changes_When_Casemapped=		F}', "");
Expect(0, 66639, '\p{Is_Changes_When_Casemapped=		F}', "");
Expect(1, 66639, '\p{^Is_Changes_When_Casemapped=		F}', "");
Expect(1, 66639, '\P{Is_Changes_When_Casemapped=		F}', "");
Expect(0, 66639, '\P{^Is_Changes_When_Casemapped=		F}', "");
Error('\p{Is_CWCM=/a/--false}');
Error('\P{Is_CWCM=/a/--false}');
Expect(1, 66640, '\p{Is_CWCM=false}', "");
Expect(0, 66640, '\p{^Is_CWCM=false}', "");
Expect(0, 66640, '\P{Is_CWCM=false}', "");
Expect(1, 66640, '\P{^Is_CWCM=false}', "");
Expect(0, 66639, '\p{Is_CWCM=false}', "");
Expect(1, 66639, '\p{^Is_CWCM=false}', "");
Expect(1, 66639, '\P{Is_CWCM=false}', "");
Expect(0, 66639, '\P{^Is_CWCM=false}', "");
Expect(1, 66640, '\p{Is_CWCM=_false}', "");
Expect(0, 66640, '\p{^Is_CWCM=_false}', "");
Expect(0, 66640, '\P{Is_CWCM=_false}', "");
Expect(1, 66640, '\P{^Is_CWCM=_false}', "");
Expect(0, 66639, '\p{Is_CWCM=_false}', "");
Expect(1, 66639, '\p{^Is_CWCM=_false}', "");
Expect(1, 66639, '\P{Is_CWCM=_false}', "");
Expect(0, 66639, '\P{^Is_CWCM=_false}', "");
Error('\p{Changes_When_Casemapped=--YES/a/}');
Error('\P{Changes_When_Casemapped=--YES/a/}');
Expect(1, 66639, '\p{Changes_When_Casemapped=yes}', "");
Expect(0, 66639, '\p{^Changes_When_Casemapped=yes}', "");
Expect(0, 66639, '\P{Changes_When_Casemapped=yes}', "");
Expect(1, 66639, '\P{^Changes_When_Casemapped=yes}', "");
Expect(0, 66640, '\p{Changes_When_Casemapped=yes}', "");
Expect(1, 66640, '\p{^Changes_When_Casemapped=yes}', "");
Expect(1, 66640, '\P{Changes_When_Casemapped=yes}', "");
Expect(0, 66640, '\P{^Changes_When_Casemapped=yes}', "");
Expect(1, 66639, '\p{Changes_When_Casemapped=	_YES}', "");
Expect(0, 66639, '\p{^Changes_When_Casemapped=	_YES}', "");
Expect(0, 66639, '\P{Changes_When_Casemapped=	_YES}', "");
Expect(1, 66639, '\P{^Changes_When_Casemapped=	_YES}', "");
Expect(0, 66640, '\p{Changes_When_Casemapped=	_YES}', "");
Expect(1, 66640, '\p{^Changes_When_Casemapped=	_YES}', "");
Expect(1, 66640, '\P{Changes_When_Casemapped=	_YES}', "");
Expect(0, 66640, '\P{^Changes_When_Casemapped=	_YES}', "");
Error('\p{CWCM=:= _Y}');
Error('\P{CWCM=:= _Y}');
Expect(1, 66639, '\p{CWCM=y}', "");
Expect(0, 66639, '\p{^CWCM=y}', "");
Expect(0, 66639, '\P{CWCM=y}', "");
Expect(1, 66639, '\P{^CWCM=y}', "");
Expect(0, 66640, '\p{CWCM=y}', "");
Expect(1, 66640, '\p{^CWCM=y}', "");
Expect(1, 66640, '\P{CWCM=y}', "");
Expect(0, 66640, '\P{^CWCM=y}', "");
Expect(1, 66639, '\p{CWCM=- y}', "");
Expect(0, 66639, '\p{^CWCM=- y}', "");
Expect(0, 66639, '\P{CWCM=- y}', "");
Expect(1, 66639, '\P{^CWCM=- y}', "");
Expect(0, 66640, '\p{CWCM=- y}', "");
Expect(1, 66640, '\p{^CWCM=- y}', "");
Expect(1, 66640, '\P{CWCM=- y}', "");
Expect(0, 66640, '\P{^CWCM=- y}', "");
Error('\p{Is_Changes_When_Casemapped=:=_T}');
Error('\P{Is_Changes_When_Casemapped=:=_T}');
Expect(1, 66639, '\p{Is_Changes_When_Casemapped: t}', "");
Expect(0, 66639, '\p{^Is_Changes_When_Casemapped: t}', "");
Expect(0, 66639, '\P{Is_Changes_When_Casemapped: t}', "");
Expect(1, 66639, '\P{^Is_Changes_When_Casemapped: t}', "");
Expect(0, 66640, '\p{Is_Changes_When_Casemapped: t}', "");
Expect(1, 66640, '\p{^Is_Changes_When_Casemapped: t}', "");
Expect(1, 66640, '\P{Is_Changes_When_Casemapped: t}', "");
Expect(0, 66640, '\P{^Is_Changes_When_Casemapped: t}', "");
Expect(1, 66639, '\p{Is_Changes_When_Casemapped:- T}', "");
Expect(0, 66639, '\p{^Is_Changes_When_Casemapped:- T}', "");
Expect(0, 66639, '\P{Is_Changes_When_Casemapped:- T}', "");
Expect(1, 66639, '\P{^Is_Changes_When_Casemapped:- T}', "");
Expect(0, 66640, '\p{Is_Changes_When_Casemapped:- T}', "");
Expect(1, 66640, '\p{^Is_Changes_When_Casemapped:- T}', "");
Expect(1, 66640, '\P{Is_Changes_When_Casemapped:- T}', "");
Expect(0, 66640, '\P{^Is_Changes_When_Casemapped:- T}', "");
Error('\p{Is_CWCM=	/a/true}');
Error('\P{Is_CWCM=	/a/true}');
Expect(1, 66639, '\p{Is_CWCM=true}', "");
Expect(0, 66639, '\p{^Is_CWCM=true}', "");
Expect(0, 66639, '\P{Is_CWCM=true}', "");
Expect(1, 66639, '\P{^Is_CWCM=true}', "");
Expect(0, 66640, '\p{Is_CWCM=true}', "");
Expect(1, 66640, '\p{^Is_CWCM=true}', "");
Expect(1, 66640, '\P{Is_CWCM=true}', "");
Expect(0, 66640, '\P{^Is_CWCM=true}', "");
Expect(1, 66639, '\p{Is_CWCM=	-TRUE}', "");
Expect(0, 66639, '\p{^Is_CWCM=	-TRUE}', "");
Expect(0, 66639, '\P{Is_CWCM=	-TRUE}', "");
Expect(1, 66639, '\P{^Is_CWCM=	-TRUE}', "");
Expect(0, 66640, '\p{Is_CWCM=	-TRUE}', "");
Expect(1, 66640, '\p{^Is_CWCM=	-TRUE}', "");
Expect(1, 66640, '\P{Is_CWCM=	-TRUE}', "");
Expect(0, 66640, '\P{^Is_CWCM=	-TRUE}', "");
Error('\p{Changes_When_NFKC_Casefolded=:=	NO}');
Error('\P{Changes_When_NFKC_Casefolded=:=	NO}');
Expect(1, 921600, '\p{Changes_When_NFKC_Casefolded:   no}', "");
Expect(0, 921600, '\p{^Changes_When_NFKC_Casefolded:   no}', "");
Expect(0, 921600, '\P{Changes_When_NFKC_Casefolded:   no}', "");
Expect(1, 921600, '\P{^Changes_When_NFKC_Casefolded:   no}', "");
Expect(0, 921599, '\p{Changes_When_NFKC_Casefolded:   no}', "");
Expect(1, 921599, '\p{^Changes_When_NFKC_Casefolded:   no}', "");
Expect(1, 921599, '\P{Changes_When_NFKC_Casefolded:   no}', "");
Expect(0, 921599, '\P{^Changes_When_NFKC_Casefolded:   no}', "");
Expect(1, 921600, '\p{Changes_When_NFKC_Casefolded= 	No}', "");
Expect(0, 921600, '\p{^Changes_When_NFKC_Casefolded= 	No}', "");
Expect(0, 921600, '\P{Changes_When_NFKC_Casefolded= 	No}', "");
Expect(1, 921600, '\P{^Changes_When_NFKC_Casefolded= 	No}', "");
Expect(0, 921599, '\p{Changes_When_NFKC_Casefolded= 	No}', "");
Expect(1, 921599, '\p{^Changes_When_NFKC_Casefolded= 	No}', "");
Expect(1, 921599, '\P{Changes_When_NFKC_Casefolded= 	No}', "");
Expect(0, 921599, '\P{^Changes_When_NFKC_Casefolded= 	No}', "");
Error('\p{CWKCF:	 /a/N}');
Error('\P{CWKCF:	 /a/N}');
Expect(1, 921600, '\p{CWKCF=n}', "");
Expect(0, 921600, '\p{^CWKCF=n}', "");
Expect(0, 921600, '\P{CWKCF=n}', "");
Expect(1, 921600, '\P{^CWKCF=n}', "");
Expect(0, 921599, '\p{CWKCF=n}', "");
Expect(1, 921599, '\p{^CWKCF=n}', "");
Expect(1, 921599, '\P{CWKCF=n}', "");
Expect(0, 921599, '\P{^CWKCF=n}', "");
Expect(1, 921600, '\p{CWKCF=		N}', "");
Expect(0, 921600, '\p{^CWKCF=		N}', "");
Expect(0, 921600, '\P{CWKCF=		N}', "");
Expect(1, 921600, '\P{^CWKCF=		N}', "");
Expect(0, 921599, '\p{CWKCF=		N}', "");
Expect(1, 921599, '\p{^CWKCF=		N}', "");
Expect(1, 921599, '\P{CWKCF=		N}', "");
Expect(0, 921599, '\P{^CWKCF=		N}', "");
Error('\p{Is_Changes_When_NFKC_Casefolded:   _ f:=}');
Error('\P{Is_Changes_When_NFKC_Casefolded:   _ f:=}');
Expect(1, 921600, '\p{Is_Changes_When_NFKC_Casefolded:	f}', "");
Expect(0, 921600, '\p{^Is_Changes_When_NFKC_Casefolded:	f}', "");
Expect(0, 921600, '\P{Is_Changes_When_NFKC_Casefolded:	f}', "");
Expect(1, 921600, '\P{^Is_Changes_When_NFKC_Casefolded:	f}', "");
Expect(0, 921599, '\p{Is_Changes_When_NFKC_Casefolded:	f}', "");
Expect(1, 921599, '\p{^Is_Changes_When_NFKC_Casefolded:	f}', "");
Expect(1, 921599, '\P{Is_Changes_When_NFKC_Casefolded:	f}', "");
Expect(0, 921599, '\P{^Is_Changes_When_NFKC_Casefolded:	f}', "");
Expect(1, 921600, '\p{Is_Changes_When_NFKC_Casefolded=_	f}', "");
Expect(0, 921600, '\p{^Is_Changes_When_NFKC_Casefolded=_	f}', "");
Expect(0, 921600, '\P{Is_Changes_When_NFKC_Casefolded=_	f}', "");
Expect(1, 921600, '\P{^Is_Changes_When_NFKC_Casefolded=_	f}', "");
Expect(0, 921599, '\p{Is_Changes_When_NFKC_Casefolded=_	f}', "");
Expect(1, 921599, '\p{^Is_Changes_When_NFKC_Casefolded=_	f}', "");
Expect(1, 921599, '\P{Is_Changes_When_NFKC_Casefolded=_	f}', "");
Expect(0, 921599, '\P{^Is_Changes_When_NFKC_Casefolded=_	f}', "");
Error('\p{Is_CWKCF=_-False:=}');
Error('\P{Is_CWKCF=_-False:=}');
Expect(1, 921600, '\p{Is_CWKCF=false}', "");
Expect(0, 921600, '\p{^Is_CWKCF=false}', "");
Expect(0, 921600, '\P{Is_CWKCF=false}', "");
Expect(1, 921600, '\P{^Is_CWKCF=false}', "");
Expect(0, 921599, '\p{Is_CWKCF=false}', "");
Expect(1, 921599, '\p{^Is_CWKCF=false}', "");
Expect(1, 921599, '\P{Is_CWKCF=false}', "");
Expect(0, 921599, '\P{^Is_CWKCF=false}', "");
Expect(1, 921600, '\p{Is_CWKCF=  FALSE}', "");
Expect(0, 921600, '\p{^Is_CWKCF=  FALSE}', "");
Expect(0, 921600, '\P{Is_CWKCF=  FALSE}', "");
Expect(1, 921600, '\P{^Is_CWKCF=  FALSE}', "");
Expect(0, 921599, '\p{Is_CWKCF=  FALSE}', "");
Expect(1, 921599, '\p{^Is_CWKCF=  FALSE}', "");
Expect(1, 921599, '\P{Is_CWKCF=  FALSE}', "");
Expect(0, 921599, '\P{^Is_CWKCF=  FALSE}', "");
Error('\p{Changes_When_NFKC_Casefolded=:= YES}');
Error('\P{Changes_When_NFKC_Casefolded=:= YES}');
Expect(1, 921599, '\p{Changes_When_NFKC_Casefolded=yes}', "");
Expect(0, 921599, '\p{^Changes_When_NFKC_Casefolded=yes}', "");
Expect(0, 921599, '\P{Changes_When_NFKC_Casefolded=yes}', "");
Expect(1, 921599, '\P{^Changes_When_NFKC_Casefolded=yes}', "");
Expect(0, 921600, '\p{Changes_When_NFKC_Casefolded=yes}', "");
Expect(1, 921600, '\p{^Changes_When_NFKC_Casefolded=yes}', "");
Expect(1, 921600, '\P{Changes_When_NFKC_Casefolded=yes}', "");
Expect(0, 921600, '\P{^Changes_When_NFKC_Casefolded=yes}', "");
Expect(1, 921599, '\p{Changes_When_NFKC_Casefolded=-Yes}', "");
Expect(0, 921599, '\p{^Changes_When_NFKC_Casefolded=-Yes}', "");
Expect(0, 921599, '\P{Changes_When_NFKC_Casefolded=-Yes}', "");
Expect(1, 921599, '\P{^Changes_When_NFKC_Casefolded=-Yes}', "");
Expect(0, 921600, '\p{Changes_When_NFKC_Casefolded=-Yes}', "");
Expect(1, 921600, '\p{^Changes_When_NFKC_Casefolded=-Yes}', "");
Expect(1, 921600, '\P{Changes_When_NFKC_Casefolded=-Yes}', "");
Expect(0, 921600, '\P{^Changes_When_NFKC_Casefolded=-Yes}', "");
Error('\p{CWKCF=_/a/Y}');
Error('\P{CWKCF=_/a/Y}');
Expect(1, 921599, '\p{CWKCF: y}', "");
Expect(0, 921599, '\p{^CWKCF: y}', "");
Expect(0, 921599, '\P{CWKCF: y}', "");
Expect(1, 921599, '\P{^CWKCF: y}', "");
Expect(0, 921600, '\p{CWKCF: y}', "");
Expect(1, 921600, '\p{^CWKCF: y}', "");
Expect(1, 921600, '\P{CWKCF: y}', "");
Expect(0, 921600, '\P{^CWKCF: y}', "");
Expect(1, 921599, '\p{CWKCF: _Y}', "");
Expect(0, 921599, '\p{^CWKCF: _Y}', "");
Expect(0, 921599, '\P{CWKCF: _Y}', "");
Expect(1, 921599, '\P{^CWKCF: _Y}', "");
Expect(0, 921600, '\p{CWKCF: _Y}', "");
Expect(1, 921600, '\p{^CWKCF: _Y}', "");
Expect(1, 921600, '\P{CWKCF: _Y}', "");
Expect(0, 921600, '\P{^CWKCF: _Y}', "");
Error('\p{Is_Changes_When_NFKC_Casefolded=	:=t}');
Error('\P{Is_Changes_When_NFKC_Casefolded=	:=t}');
Expect(1, 921599, '\p{Is_Changes_When_NFKC_Casefolded=t}', "");
Expect(0, 921599, '\p{^Is_Changes_When_NFKC_Casefolded=t}', "");
Expect(0, 921599, '\P{Is_Changes_When_NFKC_Casefolded=t}', "");
Expect(1, 921599, '\P{^Is_Changes_When_NFKC_Casefolded=t}', "");
Expect(0, 921600, '\p{Is_Changes_When_NFKC_Casefolded=t}', "");
Expect(1, 921600, '\p{^Is_Changes_When_NFKC_Casefolded=t}', "");
Expect(1, 921600, '\P{Is_Changes_When_NFKC_Casefolded=t}', "");
Expect(0, 921600, '\P{^Is_Changes_When_NFKC_Casefolded=t}', "");
Expect(1, 921599, '\p{Is_Changes_When_NFKC_Casefolded:  T}', "");
Expect(0, 921599, '\p{^Is_Changes_When_NFKC_Casefolded:  T}', "");
Expect(0, 921599, '\P{Is_Changes_When_NFKC_Casefolded:  T}', "");
Expect(1, 921599, '\P{^Is_Changes_When_NFKC_Casefolded:  T}', "");
Expect(0, 921600, '\p{Is_Changes_When_NFKC_Casefolded:  T}', "");
Expect(1, 921600, '\p{^Is_Changes_When_NFKC_Casefolded:  T}', "");
Expect(1, 921600, '\P{Is_Changes_When_NFKC_Casefolded:  T}', "");
Expect(0, 921600, '\P{^Is_Changes_When_NFKC_Casefolded:  T}', "");
Error('\p{Is_CWKCF=:=TRUE}');
Error('\P{Is_CWKCF=:=TRUE}');
Expect(1, 921599, '\p{Is_CWKCF:	true}', "");
Expect(0, 921599, '\p{^Is_CWKCF:	true}', "");
Expect(0, 921599, '\P{Is_CWKCF:	true}', "");
Expect(1, 921599, '\P{^Is_CWKCF:	true}', "");
Expect(0, 921600, '\p{Is_CWKCF:	true}', "");
Expect(1, 921600, '\p{^Is_CWKCF:	true}', "");
Expect(1, 921600, '\P{Is_CWKCF:	true}', "");
Expect(0, 921600, '\P{^Is_CWKCF:	true}', "");
Expect(1, 921599, '\p{Is_CWKCF=	True}', "");
Expect(0, 921599, '\p{^Is_CWKCF=	True}', "");
Expect(0, 921599, '\P{Is_CWKCF=	True}', "");
Expect(1, 921599, '\P{^Is_CWKCF=	True}', "");
Expect(0, 921600, '\p{Is_CWKCF=	True}', "");
Expect(1, 921600, '\p{^Is_CWKCF=	True}', "");
Expect(1, 921600, '\P{Is_CWKCF=	True}', "");
Expect(0, 921600, '\P{^Is_CWKCF=	True}', "");
Error('\p{Changes_When_Lowercased=-/a/No}');
Error('\P{Changes_When_Lowercased=-/a/No}');
Expect(1, 66600, '\p{Changes_When_Lowercased=no}', "");
Expect(0, 66600, '\p{^Changes_When_Lowercased=no}', "");
Expect(0, 66600, '\P{Changes_When_Lowercased=no}', "");
Expect(1, 66600, '\P{^Changes_When_Lowercased=no}', "");
Expect(0, 66599, '\p{Changes_When_Lowercased=no}', "");
Expect(1, 66599, '\p{^Changes_When_Lowercased=no}', "");
Expect(1, 66599, '\P{Changes_When_Lowercased=no}', "");
Expect(0, 66599, '\P{^Changes_When_Lowercased=no}', "");
Expect(1, 66600, '\p{Changes_When_Lowercased= 	No}', "");
Expect(0, 66600, '\p{^Changes_When_Lowercased= 	No}', "");
Expect(0, 66600, '\P{Changes_When_Lowercased= 	No}', "");
Expect(1, 66600, '\P{^Changes_When_Lowercased= 	No}', "");
Expect(0, 66599, '\p{Changes_When_Lowercased= 	No}', "");
Expect(1, 66599, '\p{^Changes_When_Lowercased= 	No}', "");
Expect(1, 66599, '\P{Changes_When_Lowercased= 	No}', "");
Expect(0, 66599, '\P{^Changes_When_Lowercased= 	No}', "");
Error('\p{CWL=:=- N}');
Error('\P{CWL=:=- N}');
Expect(1, 66600, '\p{CWL=n}', "");
Expect(0, 66600, '\p{^CWL=n}', "");
Expect(0, 66600, '\P{CWL=n}', "");
Expect(1, 66600, '\P{^CWL=n}', "");
Expect(0, 66599, '\p{CWL=n}', "");
Expect(1, 66599, '\p{^CWL=n}', "");
Expect(1, 66599, '\P{CWL=n}', "");
Expect(0, 66599, '\P{^CWL=n}', "");
Expect(1, 66600, '\p{CWL=--N}', "");
Expect(0, 66600, '\p{^CWL=--N}', "");
Expect(0, 66600, '\P{CWL=--N}', "");
Expect(1, 66600, '\P{^CWL=--N}', "");
Expect(0, 66599, '\p{CWL=--N}', "");
Expect(1, 66599, '\p{^CWL=--N}', "");
Expect(1, 66599, '\P{CWL=--N}', "");
Expect(0, 66599, '\P{^CWL=--N}', "");
Error('\p{Is_Changes_When_Lowercased: _/a/F}');
Error('\P{Is_Changes_When_Lowercased: _/a/F}');
Expect(1, 66600, '\p{Is_Changes_When_Lowercased=f}', "");
Expect(0, 66600, '\p{^Is_Changes_When_Lowercased=f}', "");
Expect(0, 66600, '\P{Is_Changes_When_Lowercased=f}', "");
Expect(1, 66600, '\P{^Is_Changes_When_Lowercased=f}', "");
Expect(0, 66599, '\p{Is_Changes_When_Lowercased=f}', "");
Expect(1, 66599, '\p{^Is_Changes_When_Lowercased=f}', "");
Expect(1, 66599, '\P{Is_Changes_When_Lowercased=f}', "");
Expect(0, 66599, '\P{^Is_Changes_When_Lowercased=f}', "");
Expect(1, 66600, '\p{Is_Changes_When_Lowercased= F}', "");
Expect(0, 66600, '\p{^Is_Changes_When_Lowercased= F}', "");
Expect(0, 66600, '\P{Is_Changes_When_Lowercased= F}', "");
Expect(1, 66600, '\P{^Is_Changes_When_Lowercased= F}', "");
Expect(0, 66599, '\p{Is_Changes_When_Lowercased= F}', "");
Expect(1, 66599, '\p{^Is_Changes_When_Lowercased= F}', "");
Expect(1, 66599, '\P{Is_Changes_When_Lowercased= F}', "");
Expect(0, 66599, '\P{^Is_Changes_When_Lowercased= F}', "");
Error('\p{Is_CWL=:= 	false}');
Error('\P{Is_CWL=:= 	false}');
Expect(1, 66600, '\p{Is_CWL=false}', "");
Expect(0, 66600, '\p{^Is_CWL=false}', "");
Expect(0, 66600, '\P{Is_CWL=false}', "");
Expect(1, 66600, '\P{^Is_CWL=false}', "");
Expect(0, 66599, '\p{Is_CWL=false}', "");
Expect(1, 66599, '\p{^Is_CWL=false}', "");
Expect(1, 66599, '\P{Is_CWL=false}', "");
Expect(0, 66599, '\P{^Is_CWL=false}', "");
Expect(1, 66600, '\p{Is_CWL= _FALSE}', "");
Expect(0, 66600, '\p{^Is_CWL= _FALSE}', "");
Expect(0, 66600, '\P{Is_CWL= _FALSE}', "");
Expect(1, 66600, '\P{^Is_CWL= _FALSE}', "");
Expect(0, 66599, '\p{Is_CWL= _FALSE}', "");
Expect(1, 66599, '\p{^Is_CWL= _FALSE}', "");
Expect(1, 66599, '\P{Is_CWL= _FALSE}', "");
Expect(0, 66599, '\P{^Is_CWL= _FALSE}', "");
Error('\p{Changes_When_Lowercased=-_Yes:=}');
Error('\P{Changes_When_Lowercased=-_Yes:=}');
Expect(1, 66599, '\p{Changes_When_Lowercased=yes}', "");
Expect(0, 66599, '\p{^Changes_When_Lowercased=yes}', "");
Expect(0, 66599, '\P{Changes_When_Lowercased=yes}', "");
Expect(1, 66599, '\P{^Changes_When_Lowercased=yes}', "");
Expect(0, 66600, '\p{Changes_When_Lowercased=yes}', "");
Expect(1, 66600, '\p{^Changes_When_Lowercased=yes}', "");
Expect(1, 66600, '\P{Changes_When_Lowercased=yes}', "");
Expect(0, 66600, '\P{^Changes_When_Lowercased=yes}', "");
Expect(1, 66599, '\p{Changes_When_Lowercased=	-yes}', "");
Expect(0, 66599, '\p{^Changes_When_Lowercased=	-yes}', "");
Expect(0, 66599, '\P{Changes_When_Lowercased=	-yes}', "");
Expect(1, 66599, '\P{^Changes_When_Lowercased=	-yes}', "");
Expect(0, 66600, '\p{Changes_When_Lowercased=	-yes}', "");
Expect(1, 66600, '\p{^Changes_When_Lowercased=	-yes}', "");
Expect(1, 66600, '\P{Changes_When_Lowercased=	-yes}', "");
Expect(0, 66600, '\P{^Changes_When_Lowercased=	-yes}', "");
Error('\p{CWL=_ Y:=}');
Error('\P{CWL=_ Y:=}');
Expect(1, 66599, '\p{CWL=y}', "");
Expect(0, 66599, '\p{^CWL=y}', "");
Expect(0, 66599, '\P{CWL=y}', "");
Expect(1, 66599, '\P{^CWL=y}', "");
Expect(0, 66600, '\p{CWL=y}', "");
Expect(1, 66600, '\p{^CWL=y}', "");
Expect(1, 66600, '\P{CWL=y}', "");
Expect(0, 66600, '\P{^CWL=y}', "");
Expect(1, 66599, '\p{CWL=__Y}', "");
Expect(0, 66599, '\p{^CWL=__Y}', "");
Expect(0, 66599, '\P{CWL=__Y}', "");
Expect(1, 66599, '\P{^CWL=__Y}', "");
Expect(0, 66600, '\p{CWL=__Y}', "");
Expect(1, 66600, '\p{^CWL=__Y}', "");
Expect(1, 66600, '\P{CWL=__Y}', "");
Expect(0, 66600, '\P{^CWL=__Y}', "");
Error('\p{Is_Changes_When_Lowercased=:= _T}');
Error('\P{Is_Changes_When_Lowercased=:= _T}');
Expect(1, 66599, '\p{Is_Changes_When_Lowercased:	t}', "");
Expect(0, 66599, '\p{^Is_Changes_When_Lowercased:	t}', "");
Expect(0, 66599, '\P{Is_Changes_When_Lowercased:	t}', "");
Expect(1, 66599, '\P{^Is_Changes_When_Lowercased:	t}', "");
Expect(0, 66600, '\p{Is_Changes_When_Lowercased:	t}', "");
Expect(1, 66600, '\p{^Is_Changes_When_Lowercased:	t}', "");
Expect(1, 66600, '\P{Is_Changes_When_Lowercased:	t}', "");
Expect(0, 66600, '\P{^Is_Changes_When_Lowercased:	t}', "");
Expect(1, 66599, '\p{Is_Changes_When_Lowercased= -T}', "");
Expect(0, 66599, '\p{^Is_Changes_When_Lowercased= -T}', "");
Expect(0, 66599, '\P{Is_Changes_When_Lowercased= -T}', "");
Expect(1, 66599, '\P{^Is_Changes_When_Lowercased= -T}', "");
Expect(0, 66600, '\p{Is_Changes_When_Lowercased= -T}', "");
Expect(1, 66600, '\p{^Is_Changes_When_Lowercased= -T}', "");
Expect(1, 66600, '\P{Is_Changes_When_Lowercased= -T}', "");
Expect(0, 66600, '\P{^Is_Changes_When_Lowercased= -T}', "");
Error('\p{Is_CWL= /a/True}');
Error('\P{Is_CWL= /a/True}');
Expect(1, 66599, '\p{Is_CWL=true}', "");
Expect(0, 66599, '\p{^Is_CWL=true}', "");
Expect(0, 66599, '\P{Is_CWL=true}', "");
Expect(1, 66599, '\P{^Is_CWL=true}', "");
Expect(0, 66600, '\p{Is_CWL=true}', "");
Expect(1, 66600, '\p{^Is_CWL=true}', "");
Expect(1, 66600, '\P{Is_CWL=true}', "");
Expect(0, 66600, '\P{^Is_CWL=true}', "");
Expect(1, 66599, '\p{Is_CWL=- True}', "");
Expect(0, 66599, '\p{^Is_CWL=- True}', "");
Expect(0, 66599, '\P{Is_CWL=- True}', "");
Expect(1, 66599, '\P{^Is_CWL=- True}', "");
Expect(0, 66600, '\p{Is_CWL=- True}', "");
Expect(1, 66600, '\p{^Is_CWL=- True}', "");
Expect(1, 66600, '\P{Is_CWL=- True}', "");
Expect(0, 66600, '\P{^Is_CWL=- True}', "");
Error('\p{Changes_When_Titlecased=	_No/a/}');
Error('\P{Changes_When_Titlecased=	_No/a/}');
Expect(1, 66640, '\p{Changes_When_Titlecased=no}', "");
Expect(0, 66640, '\p{^Changes_When_Titlecased=no}', "");
Expect(0, 66640, '\P{Changes_When_Titlecased=no}', "");
Expect(1, 66640, '\P{^Changes_When_Titlecased=no}', "");
Expect(0, 66639, '\p{Changes_When_Titlecased=no}', "");
Expect(1, 66639, '\p{^Changes_When_Titlecased=no}', "");
Expect(1, 66639, '\P{Changes_When_Titlecased=no}', "");
Expect(0, 66639, '\P{^Changes_When_Titlecased=no}', "");
Expect(1, 66640, '\p{Changes_When_Titlecased: -No}', "");
Expect(0, 66640, '\p{^Changes_When_Titlecased: -No}', "");
Expect(0, 66640, '\P{Changes_When_Titlecased: -No}', "");
Expect(1, 66640, '\P{^Changes_When_Titlecased: -No}', "");
Expect(0, 66639, '\p{Changes_When_Titlecased: -No}', "");
Expect(1, 66639, '\p{^Changes_When_Titlecased: -No}', "");
Expect(1, 66639, '\P{Changes_When_Titlecased: -No}', "");
Expect(0, 66639, '\P{^Changes_When_Titlecased: -No}', "");
Error('\p{CWT: /a/ N}');
Error('\P{CWT: /a/ N}');
Expect(1, 66640, '\p{CWT:n}', "");
Expect(0, 66640, '\p{^CWT:n}', "");
Expect(0, 66640, '\P{CWT:n}', "");
Expect(1, 66640, '\P{^CWT:n}', "");
Expect(0, 66639, '\p{CWT:n}', "");
Expect(1, 66639, '\p{^CWT:n}', "");
Expect(1, 66639, '\P{CWT:n}', "");
Expect(0, 66639, '\P{^CWT:n}', "");
Expect(1, 66640, '\p{CWT=	_N}', "");
Expect(0, 66640, '\p{^CWT=	_N}', "");
Expect(0, 66640, '\P{CWT=	_N}', "");
Expect(1, 66640, '\P{^CWT=	_N}', "");
Expect(0, 66639, '\p{CWT=	_N}', "");
Expect(1, 66639, '\p{^CWT=	_N}', "");
Expect(1, 66639, '\P{CWT=	_N}', "");
Expect(0, 66639, '\P{^CWT=	_N}', "");
Error('\p{Is_Changes_When_Titlecased=:= 	F}');
Error('\P{Is_Changes_When_Titlecased=:= 	F}');
Expect(1, 66640, '\p{Is_Changes_When_Titlecased=f}', "");
Expect(0, 66640, '\p{^Is_Changes_When_Titlecased=f}', "");
Expect(0, 66640, '\P{Is_Changes_When_Titlecased=f}', "");
Expect(1, 66640, '\P{^Is_Changes_When_Titlecased=f}', "");
Expect(0, 66639, '\p{Is_Changes_When_Titlecased=f}', "");
Expect(1, 66639, '\p{^Is_Changes_When_Titlecased=f}', "");
Expect(1, 66639, '\P{Is_Changes_When_Titlecased=f}', "");
Expect(0, 66639, '\P{^Is_Changes_When_Titlecased=f}', "");
Expect(1, 66640, '\p{Is_Changes_When_Titlecased:		F}', "");
Expect(0, 66640, '\p{^Is_Changes_When_Titlecased:		F}', "");
Expect(0, 66640, '\P{Is_Changes_When_Titlecased:		F}', "");
Expect(1, 66640, '\P{^Is_Changes_When_Titlecased:		F}', "");
Expect(0, 66639, '\p{Is_Changes_When_Titlecased:		F}', "");
Expect(1, 66639, '\p{^Is_Changes_When_Titlecased:		F}', "");
Expect(1, 66639, '\P{Is_Changes_When_Titlecased:		F}', "");
Expect(0, 66639, '\P{^Is_Changes_When_Titlecased:		F}', "");
Error('\p{Is_CWT= FALSE:=}');
Error('\P{Is_CWT= FALSE:=}');
Expect(1, 66640, '\p{Is_CWT=false}', "");
Expect(0, 66640, '\p{^Is_CWT=false}', "");
Expect(0, 66640, '\P{Is_CWT=false}', "");
Expect(1, 66640, '\P{^Is_CWT=false}', "");
Expect(0, 66639, '\p{Is_CWT=false}', "");
Expect(1, 66639, '\p{^Is_CWT=false}', "");
Expect(1, 66639, '\P{Is_CWT=false}', "");
Expect(0, 66639, '\P{^Is_CWT=false}', "");
Expect(1, 66640, '\p{Is_CWT: 	FALSE}', "");
Expect(0, 66640, '\p{^Is_CWT: 	FALSE}', "");
Expect(0, 66640, '\P{Is_CWT: 	FALSE}', "");
Expect(1, 66640, '\P{^Is_CWT: 	FALSE}', "");
Expect(0, 66639, '\p{Is_CWT: 	FALSE}', "");
Expect(1, 66639, '\p{^Is_CWT: 	FALSE}', "");
Expect(1, 66639, '\P{Is_CWT: 	FALSE}', "");
Expect(0, 66639, '\P{^Is_CWT: 	FALSE}', "");
Error('\p{Changes_When_Titlecased= /a/yes}');
Error('\P{Changes_When_Titlecased= /a/yes}');
Expect(1, 66639, '\p{Changes_When_Titlecased=yes}', "");
Expect(0, 66639, '\p{^Changes_When_Titlecased=yes}', "");
Expect(0, 66639, '\P{Changes_When_Titlecased=yes}', "");
Expect(1, 66639, '\P{^Changes_When_Titlecased=yes}', "");
Expect(0, 66640, '\p{Changes_When_Titlecased=yes}', "");
Expect(1, 66640, '\p{^Changes_When_Titlecased=yes}', "");
Expect(1, 66640, '\P{Changes_When_Titlecased=yes}', "");
Expect(0, 66640, '\P{^Changes_When_Titlecased=yes}', "");
Expect(1, 66639, '\p{Changes_When_Titlecased=  YES}', "");
Expect(0, 66639, '\p{^Changes_When_Titlecased=  YES}', "");
Expect(0, 66639, '\P{Changes_When_Titlecased=  YES}', "");
Expect(1, 66639, '\P{^Changes_When_Titlecased=  YES}', "");
Expect(0, 66640, '\p{Changes_When_Titlecased=  YES}', "");
Expect(1, 66640, '\p{^Changes_When_Titlecased=  YES}', "");
Expect(1, 66640, '\P{Changes_When_Titlecased=  YES}', "");
Expect(0, 66640, '\P{^Changes_When_Titlecased=  YES}', "");
Error('\p{CWT=_	Y:=}');
Error('\P{CWT=_	Y:=}');
Expect(1, 66639, '\p{CWT=y}', "");
Expect(0, 66639, '\p{^CWT=y}', "");
Expect(0, 66639, '\P{CWT=y}', "");
Expect(1, 66639, '\P{^CWT=y}', "");
Expect(0, 66640, '\p{CWT=y}', "");
Expect(1, 66640, '\p{^CWT=y}', "");
Expect(1, 66640, '\P{CWT=y}', "");
Expect(0, 66640, '\P{^CWT=y}', "");
Expect(1, 66639, '\p{CWT= -Y}', "");
Expect(0, 66639, '\p{^CWT= -Y}', "");
Expect(0, 66639, '\P{CWT= -Y}', "");
Expect(1, 66639, '\P{^CWT= -Y}', "");
Expect(0, 66640, '\p{CWT= -Y}', "");
Expect(1, 66640, '\p{^CWT= -Y}', "");
Expect(1, 66640, '\P{CWT= -Y}', "");
Expect(0, 66640, '\P{^CWT= -Y}', "");
Error('\p{Is_Changes_When_Titlecased=:=T}');
Error('\P{Is_Changes_When_Titlecased=:=T}');
Expect(1, 66639, '\p{Is_Changes_When_Titlecased:t}', "");
Expect(0, 66639, '\p{^Is_Changes_When_Titlecased:t}', "");
Expect(0, 66639, '\P{Is_Changes_When_Titlecased:t}', "");
Expect(1, 66639, '\P{^Is_Changes_When_Titlecased:t}', "");
Expect(0, 66640, '\p{Is_Changes_When_Titlecased:t}', "");
Expect(1, 66640, '\p{^Is_Changes_When_Titlecased:t}', "");
Expect(1, 66640, '\P{Is_Changes_When_Titlecased:t}', "");
Expect(0, 66640, '\P{^Is_Changes_When_Titlecased:t}', "");
Expect(1, 66639, '\p{Is_Changes_When_Titlecased=-	t}', "");
Expect(0, 66639, '\p{^Is_Changes_When_Titlecased=-	t}', "");
Expect(0, 66639, '\P{Is_Changes_When_Titlecased=-	t}', "");
Expect(1, 66639, '\P{^Is_Changes_When_Titlecased=-	t}', "");
Expect(0, 66640, '\p{Is_Changes_When_Titlecased=-	t}', "");
Expect(1, 66640, '\p{^Is_Changes_When_Titlecased=-	t}', "");
Expect(1, 66640, '\P{Is_Changes_When_Titlecased=-	t}', "");
Expect(0, 66640, '\P{^Is_Changes_When_Titlecased=-	t}', "");
Error('\p{Is_CWT=-/a/true}');
Error('\P{Is_CWT=-/a/true}');
Expect(1, 66639, '\p{Is_CWT=true}', "");
Expect(0, 66639, '\p{^Is_CWT=true}', "");
Expect(0, 66639, '\P{Is_CWT=true}', "");
Expect(1, 66639, '\P{^Is_CWT=true}', "");
Expect(0, 66640, '\p{Is_CWT=true}', "");
Expect(1, 66640, '\p{^Is_CWT=true}', "");
Expect(1, 66640, '\P{Is_CWT=true}', "");
Expect(0, 66640, '\P{^Is_CWT=true}', "");
Expect(1, 66639, '\p{Is_CWT=-True}', "");
Expect(0, 66639, '\p{^Is_CWT=-True}', "");
Expect(0, 66639, '\P{Is_CWT=-True}', "");
Expect(1, 66639, '\P{^Is_CWT=-True}', "");
Expect(0, 66640, '\p{Is_CWT=-True}', "");
Expect(1, 66640, '\p{^Is_CWT=-True}', "");
Expect(1, 66640, '\P{Is_CWT=-True}', "");
Expect(0, 66640, '\P{^Is_CWT=-True}', "");
Error('\p{Changes_When_Uppercased=	/a/NO}');
Error('\P{Changes_When_Uppercased=	/a/NO}');
Expect(1, 66640, '\p{Changes_When_Uppercased=no}', "");
Expect(0, 66640, '\p{^Changes_When_Uppercased=no}', "");
Expect(0, 66640, '\P{Changes_When_Uppercased=no}', "");
Expect(1, 66640, '\P{^Changes_When_Uppercased=no}', "");
Expect(0, 66639, '\p{Changes_When_Uppercased=no}', "");
Expect(1, 66639, '\p{^Changes_When_Uppercased=no}', "");
Expect(1, 66639, '\P{Changes_When_Uppercased=no}', "");
Expect(0, 66639, '\P{^Changes_When_Uppercased=no}', "");
Expect(1, 66640, '\p{Changes_When_Uppercased=--no}', "");
Expect(0, 66640, '\p{^Changes_When_Uppercased=--no}', "");
Expect(0, 66640, '\P{Changes_When_Uppercased=--no}', "");
Expect(1, 66640, '\P{^Changes_When_Uppercased=--no}', "");
Expect(0, 66639, '\p{Changes_When_Uppercased=--no}', "");
Expect(1, 66639, '\p{^Changes_When_Uppercased=--no}', "");
Expect(1, 66639, '\P{Changes_When_Uppercased=--no}', "");
Expect(0, 66639, '\P{^Changes_When_Uppercased=--no}', "");
Error('\p{CWU=N:=}');
Error('\P{CWU=N:=}');
Expect(1, 66640, '\p{CWU=n}', "");
Expect(0, 66640, '\p{^CWU=n}', "");
Expect(0, 66640, '\P{CWU=n}', "");
Expect(1, 66640, '\P{^CWU=n}', "");
Expect(0, 66639, '\p{CWU=n}', "");
Expect(1, 66639, '\p{^CWU=n}', "");
Expect(1, 66639, '\P{CWU=n}', "");
Expect(0, 66639, '\P{^CWU=n}', "");
Expect(1, 66640, '\p{CWU=N}', "");
Expect(0, 66640, '\p{^CWU=N}', "");
Expect(0, 66640, '\P{CWU=N}', "");
Expect(1, 66640, '\P{^CWU=N}', "");
Expect(0, 66639, '\p{CWU=N}', "");
Expect(1, 66639, '\p{^CWU=N}', "");
Expect(1, 66639, '\P{CWU=N}', "");
Expect(0, 66639, '\P{^CWU=N}', "");
Error('\p{Is_Changes_When_Uppercased=-:=F}');
Error('\P{Is_Changes_When_Uppercased=-:=F}');
Expect(1, 66640, '\p{Is_Changes_When_Uppercased=f}', "");
Expect(0, 66640, '\p{^Is_Changes_When_Uppercased=f}', "");
Expect(0, 66640, '\P{Is_Changes_When_Uppercased=f}', "");
Expect(1, 66640, '\P{^Is_Changes_When_Uppercased=f}', "");
Expect(0, 66639, '\p{Is_Changes_When_Uppercased=f}', "");
Expect(1, 66639, '\p{^Is_Changes_When_Uppercased=f}', "");
Expect(1, 66639, '\P{Is_Changes_When_Uppercased=f}', "");
Expect(0, 66639, '\P{^Is_Changes_When_Uppercased=f}', "");
Expect(1, 66640, '\p{Is_Changes_When_Uppercased= 	F}', "");
Expect(0, 66640, '\p{^Is_Changes_When_Uppercased= 	F}', "");
Expect(0, 66640, '\P{Is_Changes_When_Uppercased= 	F}', "");
Expect(1, 66640, '\P{^Is_Changes_When_Uppercased= 	F}', "");
Expect(0, 66639, '\p{Is_Changes_When_Uppercased= 	F}', "");
Expect(1, 66639, '\p{^Is_Changes_When_Uppercased= 	F}', "");
Expect(1, 66639, '\P{Is_Changes_When_Uppercased= 	F}', "");
Expect(0, 66639, '\P{^Is_Changes_When_Uppercased= 	F}', "");
Error('\p{Is_CWU=:=--False}');
Error('\P{Is_CWU=:=--False}');
Expect(1, 66640, '\p{Is_CWU=false}', "");
Expect(0, 66640, '\p{^Is_CWU=false}', "");
Expect(0, 66640, '\P{Is_CWU=false}', "");
Expect(1, 66640, '\P{^Is_CWU=false}', "");
Expect(0, 66639, '\p{Is_CWU=false}', "");
Expect(1, 66639, '\p{^Is_CWU=false}', "");
Expect(1, 66639, '\P{Is_CWU=false}', "");
Expect(0, 66639, '\P{^Is_CWU=false}', "");
Expect(1, 66640, '\p{Is_CWU=-	FALSE}', "");
Expect(0, 66640, '\p{^Is_CWU=-	FALSE}', "");
Expect(0, 66640, '\P{Is_CWU=-	FALSE}', "");
Expect(1, 66640, '\P{^Is_CWU=-	FALSE}', "");
Expect(0, 66639, '\p{Is_CWU=-	FALSE}', "");
Expect(1, 66639, '\p{^Is_CWU=-	FALSE}', "");
Expect(1, 66639, '\P{Is_CWU=-	FALSE}', "");
Expect(0, 66639, '\P{^Is_CWU=-	FALSE}', "");
Error('\p{Changes_When_Uppercased:	_:=YES}');
Error('\P{Changes_When_Uppercased:	_:=YES}');
Expect(1, 66639, '\p{Changes_When_Uppercased=yes}', "");
Expect(0, 66639, '\p{^Changes_When_Uppercased=yes}', "");
Expect(0, 66639, '\P{Changes_When_Uppercased=yes}', "");
Expect(1, 66639, '\P{^Changes_When_Uppercased=yes}', "");
Expect(0, 66640, '\p{Changes_When_Uppercased=yes}', "");
Expect(1, 66640, '\p{^Changes_When_Uppercased=yes}', "");
Expect(1, 66640, '\P{Changes_When_Uppercased=yes}', "");
Expect(0, 66640, '\P{^Changes_When_Uppercased=yes}', "");
Expect(1, 66639, '\p{Changes_When_Uppercased=_-Yes}', "");
Expect(0, 66639, '\p{^Changes_When_Uppercased=_-Yes}', "");
Expect(0, 66639, '\P{Changes_When_Uppercased=_-Yes}', "");
Expect(1, 66639, '\P{^Changes_When_Uppercased=_-Yes}', "");
Expect(0, 66640, '\p{Changes_When_Uppercased=_-Yes}', "");
Expect(1, 66640, '\p{^Changes_When_Uppercased=_-Yes}', "");
Expect(1, 66640, '\P{Changes_When_Uppercased=_-Yes}', "");
Expect(0, 66640, '\P{^Changes_When_Uppercased=_-Yes}', "");
Error('\p{CWU=	Y:=}');
Error('\P{CWU=	Y:=}');
Expect(1, 66639, '\p{CWU=y}', "");
Expect(0, 66639, '\p{^CWU=y}', "");
Expect(0, 66639, '\P{CWU=y}', "");
Expect(1, 66639, '\P{^CWU=y}', "");
Expect(0, 66640, '\p{CWU=y}', "");
Expect(1, 66640, '\p{^CWU=y}', "");
Expect(1, 66640, '\P{CWU=y}', "");
Expect(0, 66640, '\P{^CWU=y}', "");
Expect(1, 66639, '\p{CWU=-y}', "");
Expect(0, 66639, '\p{^CWU=-y}', "");
Expect(0, 66639, '\P{CWU=-y}', "");
Expect(1, 66639, '\P{^CWU=-y}', "");
Expect(0, 66640, '\p{CWU=-y}', "");
Expect(1, 66640, '\p{^CWU=-y}', "");
Expect(1, 66640, '\P{CWU=-y}', "");
Expect(0, 66640, '\P{^CWU=-y}', "");
Error('\p{Is_Changes_When_Uppercased=t/a/}');
Error('\P{Is_Changes_When_Uppercased=t/a/}');
Expect(1, 66639, '\p{Is_Changes_When_Uppercased=t}', "");
Expect(0, 66639, '\p{^Is_Changes_When_Uppercased=t}', "");
Expect(0, 66639, '\P{Is_Changes_When_Uppercased=t}', "");
Expect(1, 66639, '\P{^Is_Changes_When_Uppercased=t}', "");
Expect(0, 66640, '\p{Is_Changes_When_Uppercased=t}', "");
Expect(1, 66640, '\p{^Is_Changes_When_Uppercased=t}', "");
Expect(1, 66640, '\P{Is_Changes_When_Uppercased=t}', "");
Expect(0, 66640, '\P{^Is_Changes_When_Uppercased=t}', "");
Expect(1, 66639, '\p{Is_Changes_When_Uppercased:	  T}', "");
Expect(0, 66639, '\p{^Is_Changes_When_Uppercased:	  T}', "");
Expect(0, 66639, '\P{Is_Changes_When_Uppercased:	  T}', "");
Expect(1, 66639, '\P{^Is_Changes_When_Uppercased:	  T}', "");
Expect(0, 66640, '\p{Is_Changes_When_Uppercased:	  T}', "");
Expect(1, 66640, '\p{^Is_Changes_When_Uppercased:	  T}', "");
Expect(1, 66640, '\P{Is_Changes_When_Uppercased:	  T}', "");
Expect(0, 66640, '\P{^Is_Changes_When_Uppercased:	  T}', "");
Error('\p{Is_CWU=/a/ TRUE}');
Error('\P{Is_CWU=/a/ TRUE}');
Expect(1, 66639, '\p{Is_CWU=true}', "");
Expect(0, 66639, '\p{^Is_CWU=true}', "");
Expect(0, 66639, '\P{Is_CWU=true}', "");
Expect(1, 66639, '\P{^Is_CWU=true}', "");
Expect(0, 66640, '\p{Is_CWU=true}', "");
Expect(1, 66640, '\p{^Is_CWU=true}', "");
Expect(1, 66640, '\P{Is_CWU=true}', "");
Expect(0, 66640, '\P{^Is_CWU=true}', "");
Expect(1, 66639, '\p{Is_CWU= True}', "");
Expect(0, 66639, '\p{^Is_CWU= True}', "");
Expect(0, 66639, '\P{Is_CWU= True}', "");
Expect(1, 66639, '\P{^Is_CWU= True}', "");
Expect(0, 66640, '\p{Is_CWU= True}', "");
Expect(1, 66640, '\p{^Is_CWU= True}', "");
Expect(1, 66640, '\P{Is_CWU= True}', "");
Expect(0, 66640, '\P{^Is_CWU= True}', "");
Error('\p{Cased= _NO:=}');
Error('\P{Cased= _NO:=}');
Expect(1, 120780, '\p{Cased=no}', "");
Expect(0, 120780, '\p{^Cased=no}', "");
Expect(0, 120780, '\P{Cased=no}', "");
Expect(1, 120780, '\P{^Cased=no}', "");
Expect(0, 120779, '\p{Cased=no}', "");
Expect(1, 120779, '\p{^Cased=no}', "");
Expect(1, 120779, '\P{Cased=no}', "");
Expect(0, 120779, '\P{^Cased=no}', "");
Expect(1, 120780, '\p{Cased=-NO}', "");
Expect(0, 120780, '\p{^Cased=-NO}', "");
Expect(0, 120780, '\P{Cased=-NO}', "");
Expect(1, 120780, '\P{^Cased=-NO}', "");
Expect(0, 120779, '\p{Cased=-NO}', "");
Expect(1, 120779, '\p{^Cased=-NO}', "");
Expect(1, 120779, '\P{Cased=-NO}', "");
Expect(0, 120779, '\P{^Cased=-NO}', "");
Error('\p{Is_Cased=-:=N}');
Error('\P{Is_Cased=-:=N}');
Expect(1, 120780, '\p{Is_Cased=n}', "");
Expect(0, 120780, '\p{^Is_Cased=n}', "");
Expect(0, 120780, '\P{Is_Cased=n}', "");
Expect(1, 120780, '\P{^Is_Cased=n}', "");
Expect(0, 120779, '\p{Is_Cased=n}', "");
Expect(1, 120779, '\p{^Is_Cased=n}', "");
Expect(1, 120779, '\P{Is_Cased=n}', "");
Expect(0, 120779, '\P{^Is_Cased=n}', "");
Expect(1, 120780, '\p{Is_Cased=_N}', "");
Expect(0, 120780, '\p{^Is_Cased=_N}', "");
Expect(0, 120780, '\P{Is_Cased=_N}', "");
Expect(1, 120780, '\P{^Is_Cased=_N}', "");
Expect(0, 120779, '\p{Is_Cased=_N}', "");
Expect(1, 120779, '\p{^Is_Cased=_N}', "");
Expect(1, 120779, '\P{Is_Cased=_N}', "");
Expect(0, 120779, '\P{^Is_Cased=_N}', "");
Error('\p{Cased=	:=F}');
Error('\P{Cased=	:=F}');
Expect(1, 120780, '\p{Cased=f}', "");
Expect(0, 120780, '\p{^Cased=f}', "");
Expect(0, 120780, '\P{Cased=f}', "");
Expect(1, 120780, '\P{^Cased=f}', "");
Expect(0, 120779, '\p{Cased=f}', "");
Expect(1, 120779, '\p{^Cased=f}', "");
Expect(1, 120779, '\P{Cased=f}', "");
Expect(0, 120779, '\P{^Cased=f}', "");
Expect(1, 120780, '\p{Cased= F}', "");
Expect(0, 120780, '\p{^Cased= F}', "");
Expect(0, 120780, '\P{Cased= F}', "");
Expect(1, 120780, '\P{^Cased= F}', "");
Expect(0, 120779, '\p{Cased= F}', "");
Expect(1, 120779, '\p{^Cased= F}', "");
Expect(1, 120779, '\P{Cased= F}', "");
Expect(0, 120779, '\P{^Cased= F}', "");
Error('\p{Is_Cased=:=  False}');
Error('\P{Is_Cased=:=  False}');
Expect(1, 120780, '\p{Is_Cased=false}', "");
Expect(0, 120780, '\p{^Is_Cased=false}', "");
Expect(0, 120780, '\P{Is_Cased=false}', "");
Expect(1, 120780, '\P{^Is_Cased=false}', "");
Expect(0, 120779, '\p{Is_Cased=false}', "");
Expect(1, 120779, '\p{^Is_Cased=false}', "");
Expect(1, 120779, '\P{Is_Cased=false}', "");
Expect(0, 120779, '\P{^Is_Cased=false}', "");
Expect(1, 120780, '\p{Is_Cased=_ false}', "");
Expect(0, 120780, '\p{^Is_Cased=_ false}', "");
Expect(0, 120780, '\P{Is_Cased=_ false}', "");
Expect(1, 120780, '\P{^Is_Cased=_ false}', "");
Expect(0, 120779, '\p{Is_Cased=_ false}', "");
Expect(1, 120779, '\p{^Is_Cased=_ false}', "");
Expect(1, 120779, '\P{Is_Cased=_ false}', "");
Expect(0, 120779, '\P{^Is_Cased=_ false}', "");
Error('\p{Cased= yes/a/}');
Error('\P{Cased= yes/a/}');
Expect(1, 120779, '\p{Cased=yes}', "");
Expect(0, 120779, '\p{^Cased=yes}', "");
Expect(0, 120779, '\P{Cased=yes}', "");
Expect(1, 120779, '\P{^Cased=yes}', "");
Expect(0, 120780, '\p{Cased=yes}', "");
Expect(1, 120780, '\p{^Cased=yes}', "");
Expect(1, 120780, '\P{Cased=yes}', "");
Expect(0, 120780, '\P{^Cased=yes}', "");
Expect(1, 120779, '\p{Cased=	Yes}', "");
Expect(0, 120779, '\p{^Cased=	Yes}', "");
Expect(0, 120779, '\P{Cased=	Yes}', "");
Expect(1, 120779, '\P{^Cased=	Yes}', "");
Expect(0, 120780, '\p{Cased=	Yes}', "");
Expect(1, 120780, '\p{^Cased=	Yes}', "");
Expect(1, 120780, '\P{Cased=	Yes}', "");
Expect(0, 120780, '\P{^Cased=	Yes}', "");
Error('\p{Is_Cased=/a/y}');
Error('\P{Is_Cased=/a/y}');
Expect(1, 120779, '\p{Is_Cased:y}', "");
Expect(0, 120779, '\p{^Is_Cased:y}', "");
Expect(0, 120779, '\P{Is_Cased:y}', "");
Expect(1, 120779, '\P{^Is_Cased:y}', "");
Expect(0, 120780, '\p{Is_Cased:y}', "");
Expect(1, 120780, '\p{^Is_Cased:y}', "");
Expect(1, 120780, '\P{Is_Cased:y}', "");
Expect(0, 120780, '\P{^Is_Cased:y}', "");
Expect(1, 120779, '\p{Is_Cased=-	y}', "");
Expect(0, 120779, '\p{^Is_Cased=-	y}', "");
Expect(0, 120779, '\P{Is_Cased=-	y}', "");
Expect(1, 120779, '\P{^Is_Cased=-	y}', "");
Expect(0, 120780, '\p{Is_Cased=-	y}', "");
Expect(1, 120780, '\p{^Is_Cased=-	y}', "");
Expect(1, 120780, '\P{Is_Cased=-	y}', "");
Expect(0, 120780, '\P{^Is_Cased=-	y}', "");
Error('\p{Cased= /a/T}');
Error('\P{Cased= /a/T}');
Expect(1, 120779, '\p{Cased=t}', "");
Expect(0, 120779, '\p{^Cased=t}', "");
Expect(0, 120779, '\P{Cased=t}', "");
Expect(1, 120779, '\P{^Cased=t}', "");
Expect(0, 120780, '\p{Cased=t}', "");
Expect(1, 120780, '\p{^Cased=t}', "");
Expect(1, 120780, '\P{Cased=t}', "");
Expect(0, 120780, '\P{^Cased=t}', "");
Expect(1, 120779, '\p{Cased=  t}', "");
Expect(0, 120779, '\p{^Cased=  t}', "");
Expect(0, 120779, '\P{Cased=  t}', "");
Expect(1, 120779, '\P{^Cased=  t}', "");
Expect(0, 120780, '\p{Cased=  t}', "");
Expect(1, 120780, '\p{^Cased=  t}', "");
Expect(1, 120780, '\P{Cased=  t}', "");
Expect(0, 120780, '\P{^Cased=  t}', "");
Error('\p{Is_Cased= True/a/}');
Error('\P{Is_Cased= True/a/}');
Expect(1, 120779, '\p{Is_Cased=true}', "");
Expect(0, 120779, '\p{^Is_Cased=true}', "");
Expect(0, 120779, '\P{Is_Cased=true}', "");
Expect(1, 120779, '\P{^Is_Cased=true}', "");
Expect(0, 120780, '\p{Is_Cased=true}', "");
Expect(1, 120780, '\p{^Is_Cased=true}', "");
Expect(1, 120780, '\P{Is_Cased=true}', "");
Expect(0, 120780, '\P{^Is_Cased=true}', "");
Expect(1, 120779, '\p{Is_Cased=	_True}', "");
Expect(0, 120779, '\p{^Is_Cased=	_True}', "");
Expect(0, 120779, '\P{Is_Cased=	_True}', "");
Expect(1, 120779, '\P{^Is_Cased=	_True}', "");
Expect(0, 120780, '\p{Is_Cased=	_True}', "");
Expect(1, 120780, '\p{^Is_Cased=	_True}', "");
Expect(1, 120780, '\P{Is_Cased=	_True}', "");
Expect(0, 120780, '\P{^Is_Cased=	_True}', "");
Error('\p{Full_Composition_Exclusion:   :=_ no}');
Error('\P{Full_Composition_Exclusion:   :=_ no}');
Expect(1, 195102, '\p{Full_Composition_Exclusion:no}', "");
Expect(0, 195102, '\p{^Full_Composition_Exclusion:no}', "");
Expect(0, 195102, '\P{Full_Composition_Exclusion:no}', "");
Expect(1, 195102, '\P{^Full_Composition_Exclusion:no}', "");
Expect(0, 195101, '\p{Full_Composition_Exclusion:no}', "");
Expect(1, 195101, '\p{^Full_Composition_Exclusion:no}', "");
Expect(1, 195101, '\P{Full_Composition_Exclusion:no}', "");
Expect(0, 195101, '\P{^Full_Composition_Exclusion:no}', "");
Expect(1, 195102, '\p{Full_Composition_Exclusion= 	No}', "");
Expect(0, 195102, '\p{^Full_Composition_Exclusion= 	No}', "");
Expect(0, 195102, '\P{Full_Composition_Exclusion= 	No}', "");
Expect(1, 195102, '\P{^Full_Composition_Exclusion= 	No}', "");
Expect(0, 195101, '\p{Full_Composition_Exclusion= 	No}', "");
Expect(1, 195101, '\p{^Full_Composition_Exclusion= 	No}', "");
Expect(1, 195101, '\P{Full_Composition_Exclusion= 	No}', "");
Expect(0, 195101, '\P{^Full_Composition_Exclusion= 	No}', "");
Error('\p{Comp_Ex=/a/_N}');
Error('\P{Comp_Ex=/a/_N}');
Expect(1, 195102, '\p{Comp_Ex=n}', "");
Expect(0, 195102, '\p{^Comp_Ex=n}', "");
Expect(0, 195102, '\P{Comp_Ex=n}', "");
Expect(1, 195102, '\P{^Comp_Ex=n}', "");
Expect(0, 195101, '\p{Comp_Ex=n}', "");
Expect(1, 195101, '\p{^Comp_Ex=n}', "");
Expect(1, 195101, '\P{Comp_Ex=n}', "");
Expect(0, 195101, '\P{^Comp_Ex=n}', "");
Expect(1, 195102, '\p{Comp_Ex=--n}', "");
Expect(0, 195102, '\p{^Comp_Ex=--n}', "");
Expect(0, 195102, '\P{Comp_Ex=--n}', "");
Expect(1, 195102, '\P{^Comp_Ex=--n}', "");
Expect(0, 195101, '\p{Comp_Ex=--n}', "");
Expect(1, 195101, '\p{^Comp_Ex=--n}', "");
Expect(1, 195101, '\P{Comp_Ex=--n}', "");
Expect(0, 195101, '\P{^Comp_Ex=--n}', "");
Error('\p{Is_Full_Composition_Exclusion:	/a/f}');
Error('\P{Is_Full_Composition_Exclusion:	/a/f}');
Expect(1, 195102, '\p{Is_Full_Composition_Exclusion=f}', "");
Expect(0, 195102, '\p{^Is_Full_Composition_Exclusion=f}', "");
Expect(0, 195102, '\P{Is_Full_Composition_Exclusion=f}', "");
Expect(1, 195102, '\P{^Is_Full_Composition_Exclusion=f}', "");
Expect(0, 195101, '\p{Is_Full_Composition_Exclusion=f}', "");
Expect(1, 195101, '\p{^Is_Full_Composition_Exclusion=f}', "");
Expect(1, 195101, '\P{Is_Full_Composition_Exclusion=f}', "");
Expect(0, 195101, '\P{^Is_Full_Composition_Exclusion=f}', "");
Expect(1, 195102, '\p{Is_Full_Composition_Exclusion:			F}', "");
Expect(0, 195102, '\p{^Is_Full_Composition_Exclusion:			F}', "");
Expect(0, 195102, '\P{Is_Full_Composition_Exclusion:			F}', "");
Expect(1, 195102, '\P{^Is_Full_Composition_Exclusion:			F}', "");
Expect(0, 195101, '\p{Is_Full_Composition_Exclusion:			F}', "");
Expect(1, 195101, '\p{^Is_Full_Composition_Exclusion:			F}', "");
Expect(1, 195101, '\P{Is_Full_Composition_Exclusion:			F}', "");
Expect(0, 195101, '\P{^Is_Full_Composition_Exclusion:			F}', "");
Error('\p{Is_Comp_Ex= FALSE:=}');
Error('\P{Is_Comp_Ex= FALSE:=}');
Expect(1, 195102, '\p{Is_Comp_Ex=false}', "");
Expect(0, 195102, '\p{^Is_Comp_Ex=false}', "");
Expect(0, 195102, '\P{Is_Comp_Ex=false}', "");
Expect(1, 195102, '\P{^Is_Comp_Ex=false}', "");
Expect(0, 195101, '\p{Is_Comp_Ex=false}', "");
Expect(1, 195101, '\p{^Is_Comp_Ex=false}', "");
Expect(1, 195101, '\P{Is_Comp_Ex=false}', "");
Expect(0, 195101, '\P{^Is_Comp_Ex=false}', "");
Expect(1, 195102, '\p{Is_Comp_Ex=	-FALSE}', "");
Expect(0, 195102, '\p{^Is_Comp_Ex=	-FALSE}', "");
Expect(0, 195102, '\P{Is_Comp_Ex=	-FALSE}', "");
Expect(1, 195102, '\P{^Is_Comp_Ex=	-FALSE}', "");
Expect(0, 195101, '\p{Is_Comp_Ex=	-FALSE}', "");
Expect(1, 195101, '\p{^Is_Comp_Ex=	-FALSE}', "");
Expect(1, 195101, '\P{Is_Comp_Ex=	-FALSE}', "");
Expect(0, 195101, '\P{^Is_Comp_Ex=	-FALSE}', "");
Error('\p{Full_Composition_Exclusion:	-yes:=}');
Error('\P{Full_Composition_Exclusion:	-yes:=}');
Expect(1, 195101, '\p{Full_Composition_Exclusion=yes}', "");
Expect(0, 195101, '\p{^Full_Composition_Exclusion=yes}', "");
Expect(0, 195101, '\P{Full_Composition_Exclusion=yes}', "");
Expect(1, 195101, '\P{^Full_Composition_Exclusion=yes}', "");
Expect(0, 195102, '\p{Full_Composition_Exclusion=yes}', "");
Expect(1, 195102, '\p{^Full_Composition_Exclusion=yes}', "");
Expect(1, 195102, '\P{Full_Composition_Exclusion=yes}', "");
Expect(0, 195102, '\P{^Full_Composition_Exclusion=yes}', "");
Expect(1, 195101, '\p{Full_Composition_Exclusion=- Yes}', "");
Expect(0, 195101, '\p{^Full_Composition_Exclusion=- Yes}', "");
Expect(0, 195101, '\P{Full_Composition_Exclusion=- Yes}', "");
Expect(1, 195101, '\P{^Full_Composition_Exclusion=- Yes}', "");
Expect(0, 195102, '\p{Full_Composition_Exclusion=- Yes}', "");
Expect(1, 195102, '\p{^Full_Composition_Exclusion=- Yes}', "");
Expect(1, 195102, '\P{Full_Composition_Exclusion=- Yes}', "");
Expect(0, 195102, '\P{^Full_Composition_Exclusion=- Yes}', "");
Error('\p{Comp_Ex=:=	 Y}');
Error('\P{Comp_Ex=:=	 Y}');
Expect(1, 195101, '\p{Comp_Ex=y}', "");
Expect(0, 195101, '\p{^Comp_Ex=y}', "");
Expect(0, 195101, '\P{Comp_Ex=y}', "");
Expect(1, 195101, '\P{^Comp_Ex=y}', "");
Expect(0, 195102, '\p{Comp_Ex=y}', "");
Expect(1, 195102, '\p{^Comp_Ex=y}', "");
Expect(1, 195102, '\P{Comp_Ex=y}', "");
Expect(0, 195102, '\P{^Comp_Ex=y}', "");
Expect(1, 195101, '\p{Comp_Ex=-_Y}', "");
Expect(0, 195101, '\p{^Comp_Ex=-_Y}', "");
Expect(0, 195101, '\P{Comp_Ex=-_Y}', "");
Expect(1, 195101, '\P{^Comp_Ex=-_Y}', "");
Expect(0, 195102, '\p{Comp_Ex=-_Y}', "");
Expect(1, 195102, '\p{^Comp_Ex=-_Y}', "");
Expect(1, 195102, '\P{Comp_Ex=-_Y}', "");
Expect(0, 195102, '\P{^Comp_Ex=-_Y}', "");
Error('\p{Is_Full_Composition_Exclusion=	:=t}');
Error('\P{Is_Full_Composition_Exclusion=	:=t}');
Expect(1, 195101, '\p{Is_Full_Composition_Exclusion=t}', "");
Expect(0, 195101, '\p{^Is_Full_Composition_Exclusion=t}', "");
Expect(0, 195101, '\P{Is_Full_Composition_Exclusion=t}', "");
Expect(1, 195101, '\P{^Is_Full_Composition_Exclusion=t}', "");
Expect(0, 195102, '\p{Is_Full_Composition_Exclusion=t}', "");
Expect(1, 195102, '\p{^Is_Full_Composition_Exclusion=t}', "");
Expect(1, 195102, '\P{Is_Full_Composition_Exclusion=t}', "");
Expect(0, 195102, '\P{^Is_Full_Composition_Exclusion=t}', "");
Expect(1, 195101, '\p{Is_Full_Composition_Exclusion=	T}', "");
Expect(0, 195101, '\p{^Is_Full_Composition_Exclusion=	T}', "");
Expect(0, 195101, '\P{Is_Full_Composition_Exclusion=	T}', "");
Expect(1, 195101, '\P{^Is_Full_Composition_Exclusion=	T}', "");
Expect(0, 195102, '\p{Is_Full_Composition_Exclusion=	T}', "");
Expect(1, 195102, '\p{^Is_Full_Composition_Exclusion=	T}', "");
Expect(1, 195102, '\P{Is_Full_Composition_Exclusion=	T}', "");
Expect(0, 195102, '\P{^Is_Full_Composition_Exclusion=	T}', "");
Error('\p{Is_Comp_Ex=- True/a/}');
Error('\P{Is_Comp_Ex=- True/a/}');
Expect(1, 195101, '\p{Is_Comp_Ex=true}', "");
Expect(0, 195101, '\p{^Is_Comp_Ex=true}', "");
Expect(0, 195101, '\P{Is_Comp_Ex=true}', "");
Expect(1, 195101, '\P{^Is_Comp_Ex=true}', "");
Expect(0, 195102, '\p{Is_Comp_Ex=true}', "");
Expect(1, 195102, '\p{^Is_Comp_Ex=true}', "");
Expect(1, 195102, '\P{Is_Comp_Ex=true}', "");
Expect(0, 195102, '\P{^Is_Comp_Ex=true}', "");
Expect(1, 195101, '\p{Is_Comp_Ex:	_TRUE}', "");
Expect(0, 195101, '\p{^Is_Comp_Ex:	_TRUE}', "");
Expect(0, 195101, '\P{Is_Comp_Ex:	_TRUE}', "");
Expect(1, 195101, '\P{^Is_Comp_Ex:	_TRUE}', "");
Expect(0, 195102, '\p{Is_Comp_Ex:	_TRUE}', "");
Expect(1, 195102, '\p{^Is_Comp_Ex:	_TRUE}', "");
Expect(1, 195102, '\P{Is_Comp_Ex:	_TRUE}', "");
Expect(0, 195102, '\P{^Is_Comp_Ex:	_TRUE}', "");
Error('\p{Default_Ignorable_Code_Point=/a/No}');
Error('\P{Default_Ignorable_Code_Point=/a/No}');
Expect(1, 921600, '\p{Default_Ignorable_Code_Point=no}', "");
Expect(0, 921600, '\p{^Default_Ignorable_Code_Point=no}', "");
Expect(0, 921600, '\P{Default_Ignorable_Code_Point=no}', "");
Expect(1, 921600, '\P{^Default_Ignorable_Code_Point=no}', "");
Expect(0, 921599, '\p{Default_Ignorable_Code_Point=no}', "");
Expect(1, 921599, '\p{^Default_Ignorable_Code_Point=no}', "");
Expect(1, 921599, '\P{Default_Ignorable_Code_Point=no}', "");
Expect(0, 921599, '\P{^Default_Ignorable_Code_Point=no}', "");
Expect(1, 921600, '\p{Default_Ignorable_Code_Point=	No}', "");
Expect(0, 921600, '\p{^Default_Ignorable_Code_Point=	No}', "");
Expect(0, 921600, '\P{Default_Ignorable_Code_Point=	No}', "");
Expect(1, 921600, '\P{^Default_Ignorable_Code_Point=	No}', "");
Expect(0, 921599, '\p{Default_Ignorable_Code_Point=	No}', "");
Expect(1, 921599, '\p{^Default_Ignorable_Code_Point=	No}', "");
Expect(1, 921599, '\P{Default_Ignorable_Code_Point=	No}', "");
Expect(0, 921599, '\P{^Default_Ignorable_Code_Point=	No}', "");
Error('\p{DI:-n/a/}');
Error('\P{DI:-n/a/}');
Expect(1, 921600, '\p{DI=n}', "");
Expect(0, 921600, '\p{^DI=n}', "");
Expect(0, 921600, '\P{DI=n}', "");
Expect(1, 921600, '\P{^DI=n}', "");
Expect(0, 921599, '\p{DI=n}', "");
Expect(1, 921599, '\p{^DI=n}', "");
Expect(1, 921599, '\P{DI=n}', "");
Expect(0, 921599, '\P{^DI=n}', "");
Expect(1, 921600, '\p{DI= n}', "");
Expect(0, 921600, '\p{^DI= n}', "");
Expect(0, 921600, '\P{DI= n}', "");
Expect(1, 921600, '\P{^DI= n}', "");
Expect(0, 921599, '\p{DI= n}', "");
Expect(1, 921599, '\p{^DI= n}', "");
Expect(1, 921599, '\P{DI= n}', "");
Expect(0, 921599, '\P{^DI= n}', "");
Error('\p{Is_Default_Ignorable_Code_Point:	-:=F}');
Error('\P{Is_Default_Ignorable_Code_Point:	-:=F}');
Expect(1, 921600, '\p{Is_Default_Ignorable_Code_Point=f}', "");
Expect(0, 921600, '\p{^Is_Default_Ignorable_Code_Point=f}', "");
Expect(0, 921600, '\P{Is_Default_Ignorable_Code_Point=f}', "");
Expect(1, 921600, '\P{^Is_Default_Ignorable_Code_Point=f}', "");
Expect(0, 921599, '\p{Is_Default_Ignorable_Code_Point=f}', "");
Expect(1, 921599, '\p{^Is_Default_Ignorable_Code_Point=f}', "");
Expect(1, 921599, '\P{Is_Default_Ignorable_Code_Point=f}', "");
Expect(0, 921599, '\P{^Is_Default_Ignorable_Code_Point=f}', "");
Expect(1, 921600, '\p{Is_Default_Ignorable_Code_Point= f}', "");
Expect(0, 921600, '\p{^Is_Default_Ignorable_Code_Point= f}', "");
Expect(0, 921600, '\P{Is_Default_Ignorable_Code_Point= f}', "");
Expect(1, 921600, '\P{^Is_Default_Ignorable_Code_Point= f}', "");
Expect(0, 921599, '\p{Is_Default_Ignorable_Code_Point= f}', "");
Expect(1, 921599, '\p{^Is_Default_Ignorable_Code_Point= f}', "");
Expect(1, 921599, '\P{Is_Default_Ignorable_Code_Point= f}', "");
Expect(0, 921599, '\P{^Is_Default_Ignorable_Code_Point= f}', "");
Error('\p{Is_DI=	:=FALSE}');
Error('\P{Is_DI=	:=FALSE}');
Expect(1, 921600, '\p{Is_DI=false}', "");
Expect(0, 921600, '\p{^Is_DI=false}', "");
Expect(0, 921600, '\P{Is_DI=false}', "");
Expect(1, 921600, '\P{^Is_DI=false}', "");
Expect(0, 921599, '\p{Is_DI=false}', "");
Expect(1, 921599, '\p{^Is_DI=false}', "");
Expect(1, 921599, '\P{Is_DI=false}', "");
Expect(0, 921599, '\P{^Is_DI=false}', "");
Expect(1, 921600, '\p{Is_DI=	_False}', "");
Expect(0, 921600, '\p{^Is_DI=	_False}', "");
Expect(0, 921600, '\P{Is_DI=	_False}', "");
Expect(1, 921600, '\P{^Is_DI=	_False}', "");
Expect(0, 921599, '\p{Is_DI=	_False}', "");
Expect(1, 921599, '\p{^Is_DI=	_False}', "");
Expect(1, 921599, '\P{Is_DI=	_False}', "");
Expect(0, 921599, '\P{^Is_DI=	_False}', "");
Error('\p{Default_Ignorable_Code_Point=:= -yes}');
Error('\P{Default_Ignorable_Code_Point=:= -yes}');
Expect(1, 921599, '\p{Default_Ignorable_Code_Point=yes}', "");
Expect(0, 921599, '\p{^Default_Ignorable_Code_Point=yes}', "");
Expect(0, 921599, '\P{Default_Ignorable_Code_Point=yes}', "");
Expect(1, 921599, '\P{^Default_Ignorable_Code_Point=yes}', "");
Expect(0, 921600, '\p{Default_Ignorable_Code_Point=yes}', "");
Expect(1, 921600, '\p{^Default_Ignorable_Code_Point=yes}', "");
Expect(1, 921600, '\P{Default_Ignorable_Code_Point=yes}', "");
Expect(0, 921600, '\P{^Default_Ignorable_Code_Point=yes}', "");
Expect(1, 921599, '\p{Default_Ignorable_Code_Point=_YES}', "");
Expect(0, 921599, '\p{^Default_Ignorable_Code_Point=_YES}', "");
Expect(0, 921599, '\P{Default_Ignorable_Code_Point=_YES}', "");
Expect(1, 921599, '\P{^Default_Ignorable_Code_Point=_YES}', "");
Expect(0, 921600, '\p{Default_Ignorable_Code_Point=_YES}', "");
Expect(1, 921600, '\p{^Default_Ignorable_Code_Point=_YES}', "");
Expect(1, 921600, '\P{Default_Ignorable_Code_Point=_YES}', "");
Expect(0, 921600, '\P{^Default_Ignorable_Code_Point=_YES}', "");
Error('\p{DI=_ Y:=}');
Error('\P{DI=_ Y:=}');
Expect(1, 921599, '\p{DI=y}', "");
Expect(0, 921599, '\p{^DI=y}', "");
Expect(0, 921599, '\P{DI=y}', "");
Expect(1, 921599, '\P{^DI=y}', "");
Expect(0, 921600, '\p{DI=y}', "");
Expect(1, 921600, '\p{^DI=y}', "");
Expect(1, 921600, '\P{DI=y}', "");
Expect(0, 921600, '\P{^DI=y}', "");
Expect(1, 921599, '\p{DI=  Y}', "");
Expect(0, 921599, '\p{^DI=  Y}', "");
Expect(0, 921599, '\P{DI=  Y}', "");
Expect(1, 921599, '\P{^DI=  Y}', "");
Expect(0, 921600, '\p{DI=  Y}', "");
Expect(1, 921600, '\p{^DI=  Y}', "");
Expect(1, 921600, '\P{DI=  Y}', "");
Expect(0, 921600, '\P{^DI=  Y}', "");
Error('\p{Is_Default_Ignorable_Code_Point=:=	-T}');
Error('\P{Is_Default_Ignorable_Code_Point=:=	-T}');
Expect(1, 921599, '\p{Is_Default_Ignorable_Code_Point=t}', "");
Expect(0, 921599, '\p{^Is_Default_Ignorable_Code_Point=t}', "");
Expect(0, 921599, '\P{Is_Default_Ignorable_Code_Point=t}', "");
Expect(1, 921599, '\P{^Is_Default_Ignorable_Code_Point=t}', "");
Expect(0, 921600, '\p{Is_Default_Ignorable_Code_Point=t}', "");
Expect(1, 921600, '\p{^Is_Default_Ignorable_Code_Point=t}', "");
Expect(1, 921600, '\P{Is_Default_Ignorable_Code_Point=t}', "");
Expect(0, 921600, '\P{^Is_Default_Ignorable_Code_Point=t}', "");
Expect(1, 921599, '\p{Is_Default_Ignorable_Code_Point= T}', "");
Expect(0, 921599, '\p{^Is_Default_Ignorable_Code_Point= T}', "");
Expect(0, 921599, '\P{Is_Default_Ignorable_Code_Point= T}', "");
Expect(1, 921599, '\P{^Is_Default_Ignorable_Code_Point= T}', "");
Expect(0, 921600, '\p{Is_Default_Ignorable_Code_Point= T}', "");
Expect(1, 921600, '\p{^Is_Default_Ignorable_Code_Point= T}', "");
Expect(1, 921600, '\P{Is_Default_Ignorable_Code_Point= T}', "");
Expect(0, 921600, '\P{^Is_Default_Ignorable_Code_Point= T}', "");
Error('\p{Is_DI=/a/ 	true}');
Error('\P{Is_DI=/a/ 	true}');
Expect(1, 921599, '\p{Is_DI=true}', "");
Expect(0, 921599, '\p{^Is_DI=true}', "");
Expect(0, 921599, '\P{Is_DI=true}', "");
Expect(1, 921599, '\P{^Is_DI=true}', "");
Expect(0, 921600, '\p{Is_DI=true}', "");
Expect(1, 921600, '\p{^Is_DI=true}', "");
Expect(1, 921600, '\P{Is_DI=true}', "");
Expect(0, 921600, '\P{^Is_DI=true}', "");
Expect(1, 921599, '\p{Is_DI=	True}', "");
Expect(0, 921599, '\p{^Is_DI=	True}', "");
Expect(0, 921599, '\P{Is_DI=	True}', "");
Expect(1, 921599, '\P{^Is_DI=	True}', "");
Expect(0, 921600, '\p{Is_DI=	True}', "");
Expect(1, 921600, '\p{^Is_DI=	True}', "");
Expect(1, 921600, '\P{Is_DI=	True}', "");
Expect(0, 921600, '\P{^Is_DI=	True}', "");
Error('\p{Dash=:=_ No}');
Error('\P{Dash=:=_ No}');
Expect(1, 65294, '\p{Dash:no}', "");
Expect(0, 65294, '\p{^Dash:no}', "");
Expect(0, 65294, '\P{Dash:no}', "");
Expect(1, 65294, '\P{^Dash:no}', "");
Expect(0, 65293, '\p{Dash:no}', "");
Expect(1, 65293, '\p{^Dash:no}', "");
Expect(1, 65293, '\P{Dash:no}', "");
Expect(0, 65293, '\P{^Dash:no}', "");
Expect(1, 65294, '\p{Dash=_-No}', "");
Expect(0, 65294, '\p{^Dash=_-No}', "");
Expect(0, 65294, '\P{Dash=_-No}', "");
Expect(1, 65294, '\P{^Dash=_-No}', "");
Expect(0, 65293, '\p{Dash=_-No}', "");
Expect(1, 65293, '\p{^Dash=_-No}', "");
Expect(1, 65293, '\P{Dash=_-No}', "");
Expect(0, 65293, '\P{^Dash=_-No}', "");
Error('\p{Is_Dash=_N:=}');
Error('\P{Is_Dash=_N:=}');
Expect(1, 65294, '\p{Is_Dash=n}', "");
Expect(0, 65294, '\p{^Is_Dash=n}', "");
Expect(0, 65294, '\P{Is_Dash=n}', "");
Expect(1, 65294, '\P{^Is_Dash=n}', "");
Expect(0, 65293, '\p{Is_Dash=n}', "");
Expect(1, 65293, '\p{^Is_Dash=n}', "");
Expect(1, 65293, '\P{Is_Dash=n}', "");
Expect(0, 65293, '\P{^Is_Dash=n}', "");
Expect(1, 65294, '\p{Is_Dash=-	N}', "");
Expect(0, 65294, '\p{^Is_Dash=-	N}', "");
Expect(0, 65294, '\P{Is_Dash=-	N}', "");
Expect(1, 65294, '\P{^Is_Dash=-	N}', "");
Expect(0, 65293, '\p{Is_Dash=-	N}', "");
Expect(1, 65293, '\p{^Is_Dash=-	N}', "");
Expect(1, 65293, '\P{Is_Dash=-	N}', "");
Expect(0, 65293, '\P{^Is_Dash=-	N}', "");
Error('\p{Dash:   := -f}');
Error('\P{Dash:   := -f}');
Expect(1, 65294, '\p{Dash=f}', "");
Expect(0, 65294, '\p{^Dash=f}', "");
Expect(0, 65294, '\P{Dash=f}', "");
Expect(1, 65294, '\P{^Dash=f}', "");
Expect(0, 65293, '\p{Dash=f}', "");
Expect(1, 65293, '\p{^Dash=f}', "");
Expect(1, 65293, '\P{Dash=f}', "");
Expect(0, 65293, '\P{^Dash=f}', "");
Expect(1, 65294, '\p{Dash=_F}', "");
Expect(0, 65294, '\p{^Dash=_F}', "");
Expect(0, 65294, '\P{Dash=_F}', "");
Expect(1, 65294, '\P{^Dash=_F}', "");
Expect(0, 65293, '\p{Dash=_F}', "");
Expect(1, 65293, '\p{^Dash=_F}', "");
Expect(1, 65293, '\P{Dash=_F}', "");
Expect(0, 65293, '\P{^Dash=_F}', "");
Error('\p{Is_Dash=:=--False}');
Error('\P{Is_Dash=:=--False}');
Expect(1, 65294, '\p{Is_Dash=false}', "");
Expect(0, 65294, '\p{^Is_Dash=false}', "");
Expect(0, 65294, '\P{Is_Dash=false}', "");
Expect(1, 65294, '\P{^Is_Dash=false}', "");
Expect(0, 65293, '\p{Is_Dash=false}', "");
Expect(1, 65293, '\p{^Is_Dash=false}', "");
Expect(1, 65293, '\P{Is_Dash=false}', "");
Expect(0, 65293, '\P{^Is_Dash=false}', "");
Expect(1, 65294, '\p{Is_Dash=_ False}', "");
Expect(0, 65294, '\p{^Is_Dash=_ False}', "");
Expect(0, 65294, '\P{Is_Dash=_ False}', "");
Expect(1, 65294, '\P{^Is_Dash=_ False}', "");
Expect(0, 65293, '\p{Is_Dash=_ False}', "");
Expect(1, 65293, '\p{^Is_Dash=_ False}', "");
Expect(1, 65293, '\P{Is_Dash=_ False}', "");
Expect(0, 65293, '\P{^Is_Dash=_ False}', "");
Error('\p{Dash=/a/ 	Yes}');
Error('\P{Dash=/a/ 	Yes}');
Expect(1, 65293, '\p{Dash=yes}', "");
Expect(0, 65293, '\p{^Dash=yes}', "");
Expect(0, 65293, '\P{Dash=yes}', "");
Expect(1, 65293, '\P{^Dash=yes}', "");
Expect(0, 65294, '\p{Dash=yes}', "");
Expect(1, 65294, '\p{^Dash=yes}', "");
Expect(1, 65294, '\P{Dash=yes}', "");
Expect(0, 65294, '\P{^Dash=yes}', "");
Expect(1, 65293, '\p{Dash: - Yes}', "");
Expect(0, 65293, '\p{^Dash: - Yes}', "");
Expect(0, 65293, '\P{Dash: - Yes}', "");
Expect(1, 65293, '\P{^Dash: - Yes}', "");
Expect(0, 65294, '\p{Dash: - Yes}', "");
Expect(1, 65294, '\p{^Dash: - Yes}', "");
Expect(1, 65294, '\P{Dash: - Yes}', "");
Expect(0, 65294, '\P{^Dash: - Yes}', "");
Error('\p{Is_Dash=:=-y}');
Error('\P{Is_Dash=:=-y}');
Expect(1, 65293, '\p{Is_Dash=y}', "");
Expect(0, 65293, '\p{^Is_Dash=y}', "");
Expect(0, 65293, '\P{Is_Dash=y}', "");
Expect(1, 65293, '\P{^Is_Dash=y}', "");
Expect(0, 65294, '\p{Is_Dash=y}', "");
Expect(1, 65294, '\p{^Is_Dash=y}', "");
Expect(1, 65294, '\P{Is_Dash=y}', "");
Expect(0, 65294, '\P{^Is_Dash=y}', "");
Expect(1, 65293, '\p{Is_Dash=_Y}', "");
Expect(0, 65293, '\p{^Is_Dash=_Y}', "");
Expect(0, 65293, '\P{Is_Dash=_Y}', "");
Expect(1, 65293, '\P{^Is_Dash=_Y}', "");
Expect(0, 65294, '\p{Is_Dash=_Y}', "");
Expect(1, 65294, '\p{^Is_Dash=_Y}', "");
Expect(1, 65294, '\P{Is_Dash=_Y}', "");
Expect(0, 65294, '\P{^Is_Dash=_Y}', "");
Error('\p{Dash=	/a/T}');
Error('\P{Dash=	/a/T}');
Expect(1, 65293, '\p{Dash=t}', "");
Expect(0, 65293, '\p{^Dash=t}', "");
Expect(0, 65293, '\P{Dash=t}', "");
Expect(1, 65293, '\P{^Dash=t}', "");
Expect(0, 65294, '\p{Dash=t}', "");
Expect(1, 65294, '\p{^Dash=t}', "");
Expect(1, 65294, '\P{Dash=t}', "");
Expect(0, 65294, '\P{^Dash=t}', "");
Expect(1, 65293, '\p{Dash=-t}', "");
Expect(0, 65293, '\p{^Dash=-t}', "");
Expect(0, 65293, '\P{Dash=-t}', "");
Expect(1, 65293, '\P{^Dash=-t}', "");
Expect(0, 65294, '\p{Dash=-t}', "");
Expect(1, 65294, '\p{^Dash=-t}', "");
Expect(1, 65294, '\P{Dash=-t}', "");
Expect(0, 65294, '\P{^Dash=-t}', "");
Error('\p{Is_Dash=	/a/True}');
Error('\P{Is_Dash=	/a/True}');
Expect(1, 65293, '\p{Is_Dash=true}', "");
Expect(0, 65293, '\p{^Is_Dash=true}', "");
Expect(0, 65293, '\P{Is_Dash=true}', "");
Expect(1, 65293, '\P{^Is_Dash=true}', "");
Expect(0, 65294, '\p{Is_Dash=true}', "");
Expect(1, 65294, '\p{^Is_Dash=true}', "");
Expect(1, 65294, '\P{Is_Dash=true}', "");
Expect(0, 65294, '\P{^Is_Dash=true}', "");
Expect(1, 65293, '\p{Is_Dash= true}', "");
Expect(0, 65293, '\p{^Is_Dash= true}', "");
Expect(0, 65293, '\P{Is_Dash= true}', "");
Expect(1, 65293, '\P{^Is_Dash= true}', "");
Expect(0, 65294, '\p{Is_Dash= true}', "");
Expect(1, 65294, '\p{^Is_Dash= true}', "");
Expect(1, 65294, '\P{Is_Dash= true}', "");
Expect(0, 65294, '\P{^Is_Dash= true}', "");
Error('\p{Deprecated= /a/NO}');
Error('\P{Deprecated= /a/NO}');
Expect(1, 917632, '\p{Deprecated=no}', "");
Expect(0, 917632, '\p{^Deprecated=no}', "");
Expect(0, 917632, '\P{Deprecated=no}', "");
Expect(1, 917632, '\P{^Deprecated=no}', "");
Expect(0, 917631, '\p{Deprecated=no}', "");
Expect(1, 917631, '\p{^Deprecated=no}', "");
Expect(1, 917631, '\P{Deprecated=no}', "");
Expect(0, 917631, '\P{^Deprecated=no}', "");
Expect(1, 917632, '\p{Deprecated= 	NO}', "");
Expect(0, 917632, '\p{^Deprecated= 	NO}', "");
Expect(0, 917632, '\P{Deprecated= 	NO}', "");
Expect(1, 917632, '\P{^Deprecated= 	NO}', "");
Expect(0, 917631, '\p{Deprecated= 	NO}', "");
Expect(1, 917631, '\p{^Deprecated= 	NO}', "");
Expect(1, 917631, '\P{Deprecated= 	NO}', "");
Expect(0, 917631, '\P{^Deprecated= 	NO}', "");
Error('\p{Dep= 	N:=}');
Error('\P{Dep= 	N:=}');
Expect(1, 917632, '\p{Dep=n}', "");
Expect(0, 917632, '\p{^Dep=n}', "");
Expect(0, 917632, '\P{Dep=n}', "");
Expect(1, 917632, '\P{^Dep=n}', "");
Expect(0, 917631, '\p{Dep=n}', "");
Expect(1, 917631, '\p{^Dep=n}', "");
Expect(1, 917631, '\P{Dep=n}', "");
Expect(0, 917631, '\P{^Dep=n}', "");
Expect(1, 917632, '\p{Dep: _N}', "");
Expect(0, 917632, '\p{^Dep: _N}', "");
Expect(0, 917632, '\P{Dep: _N}', "");
Expect(1, 917632, '\P{^Dep: _N}', "");
Expect(0, 917631, '\p{Dep: _N}', "");
Expect(1, 917631, '\p{^Dep: _N}', "");
Expect(1, 917631, '\P{Dep: _N}', "");
Expect(0, 917631, '\P{^Dep: _N}', "");
Error('\p{Is_Deprecated=_F/a/}');
Error('\P{Is_Deprecated=_F/a/}');
Expect(1, 917632, '\p{Is_Deprecated=f}', "");
Expect(0, 917632, '\p{^Is_Deprecated=f}', "");
Expect(0, 917632, '\P{Is_Deprecated=f}', "");
Expect(1, 917632, '\P{^Is_Deprecated=f}', "");
Expect(0, 917631, '\p{Is_Deprecated=f}', "");
Expect(1, 917631, '\p{^Is_Deprecated=f}', "");
Expect(1, 917631, '\P{Is_Deprecated=f}', "");
Expect(0, 917631, '\P{^Is_Deprecated=f}', "");
Expect(1, 917632, '\p{Is_Deprecated=  F}', "");
Expect(0, 917632, '\p{^Is_Deprecated=  F}', "");
Expect(0, 917632, '\P{Is_Deprecated=  F}', "");
Expect(1, 917632, '\P{^Is_Deprecated=  F}', "");
Expect(0, 917631, '\p{Is_Deprecated=  F}', "");
Expect(1, 917631, '\p{^Is_Deprecated=  F}', "");
Expect(1, 917631, '\P{Is_Deprecated=  F}', "");
Expect(0, 917631, '\P{^Is_Deprecated=  F}', "");
Error('\p{Is_Dep=:=	-FALSE}');
Error('\P{Is_Dep=:=	-FALSE}');
Expect(1, 917632, '\p{Is_Dep=false}', "");
Expect(0, 917632, '\p{^Is_Dep=false}', "");
Expect(0, 917632, '\P{Is_Dep=false}', "");
Expect(1, 917632, '\P{^Is_Dep=false}', "");
Expect(0, 917631, '\p{Is_Dep=false}', "");
Expect(1, 917631, '\p{^Is_Dep=false}', "");
Expect(1, 917631, '\P{Is_Dep=false}', "");
Expect(0, 917631, '\P{^Is_Dep=false}', "");
Expect(1, 917632, '\p{Is_Dep=_-false}', "");
Expect(0, 917632, '\p{^Is_Dep=_-false}', "");
Expect(0, 917632, '\P{Is_Dep=_-false}', "");
Expect(1, 917632, '\P{^Is_Dep=_-false}', "");
Expect(0, 917631, '\p{Is_Dep=_-false}', "");
Expect(1, 917631, '\p{^Is_Dep=_-false}', "");
Expect(1, 917631, '\P{Is_Dep=_-false}', "");
Expect(0, 917631, '\P{^Is_Dep=_-false}', "");
Error('\p{Deprecated=_/a/yes}');
Error('\P{Deprecated=_/a/yes}');
Expect(1, 917631, '\p{Deprecated=yes}', "");
Expect(0, 917631, '\p{^Deprecated=yes}', "");
Expect(0, 917631, '\P{Deprecated=yes}', "");
Expect(1, 917631, '\P{^Deprecated=yes}', "");
Expect(0, 917632, '\p{Deprecated=yes}', "");
Expect(1, 917632, '\p{^Deprecated=yes}', "");
Expect(1, 917632, '\P{Deprecated=yes}', "");
Expect(0, 917632, '\P{^Deprecated=yes}', "");
Expect(1, 917631, '\p{Deprecated=_YES}', "");
Expect(0, 917631, '\p{^Deprecated=_YES}', "");
Expect(0, 917631, '\P{Deprecated=_YES}', "");
Expect(1, 917631, '\P{^Deprecated=_YES}', "");
Expect(0, 917632, '\p{Deprecated=_YES}', "");
Expect(1, 917632, '\p{^Deprecated=_YES}', "");
Expect(1, 917632, '\P{Deprecated=_YES}', "");
Expect(0, 917632, '\P{^Deprecated=_YES}', "");
Error('\p{Dep= Y/a/}');
Error('\P{Dep= Y/a/}');
Expect(1, 917631, '\p{Dep=y}', "");
Expect(0, 917631, '\p{^Dep=y}', "");
Expect(0, 917631, '\P{Dep=y}', "");
Expect(1, 917631, '\P{^Dep=y}', "");
Expect(0, 917632, '\p{Dep=y}', "");
Expect(1, 917632, '\p{^Dep=y}', "");
Expect(1, 917632, '\P{Dep=y}', "");
Expect(0, 917632, '\P{^Dep=y}', "");
Expect(1, 917631, '\p{Dep= -Y}', "");
Expect(0, 917631, '\p{^Dep= -Y}', "");
Expect(0, 917631, '\P{Dep= -Y}', "");
Expect(1, 917631, '\P{^Dep= -Y}', "");
Expect(0, 917632, '\p{Dep= -Y}', "");
Expect(1, 917632, '\p{^Dep= -Y}', "");
Expect(1, 917632, '\P{Dep= -Y}', "");
Expect(0, 917632, '\P{^Dep= -Y}', "");
Error('\p{Is_Deprecated=_t/a/}');
Error('\P{Is_Deprecated=_t/a/}');
Expect(1, 917631, '\p{Is_Deprecated=t}', "");
Expect(0, 917631, '\p{^Is_Deprecated=t}', "");
Expect(0, 917631, '\P{Is_Deprecated=t}', "");
Expect(1, 917631, '\P{^Is_Deprecated=t}', "");
Expect(0, 917632, '\p{Is_Deprecated=t}', "");
Expect(1, 917632, '\p{^Is_Deprecated=t}', "");
Expect(1, 917632, '\P{Is_Deprecated=t}', "");
Expect(0, 917632, '\P{^Is_Deprecated=t}', "");
Expect(1, 917631, '\p{Is_Deprecated=  t}', "");
Expect(0, 917631, '\p{^Is_Deprecated=  t}', "");
Expect(0, 917631, '\P{Is_Deprecated=  t}', "");
Expect(1, 917631, '\P{^Is_Deprecated=  t}', "");
Expect(0, 917632, '\p{Is_Deprecated=  t}', "");
Expect(1, 917632, '\p{^Is_Deprecated=  t}', "");
Expect(1, 917632, '\P{Is_Deprecated=  t}', "");
Expect(0, 917632, '\P{^Is_Deprecated=  t}', "");
Error('\p{Is_Dep=/a/  TRUE}');
Error('\P{Is_Dep=/a/  TRUE}');
Expect(1, 917631, '\p{Is_Dep: true}', "");
Expect(0, 917631, '\p{^Is_Dep: true}', "");
Expect(0, 917631, '\P{Is_Dep: true}', "");
Expect(1, 917631, '\P{^Is_Dep: true}', "");
Expect(0, 917632, '\p{Is_Dep: true}', "");
Expect(1, 917632, '\p{^Is_Dep: true}', "");
Expect(1, 917632, '\P{Is_Dep: true}', "");
Expect(0, 917632, '\P{^Is_Dep: true}', "");
Expect(1, 917631, '\p{Is_Dep=_	TRUE}', "");
Expect(0, 917631, '\p{^Is_Dep=_	TRUE}', "");
Expect(0, 917631, '\P{Is_Dep=_	TRUE}', "");
Expect(1, 917631, '\P{^Is_Dep=_	TRUE}', "");
Expect(0, 917632, '\p{Is_Dep=_	TRUE}', "");
Expect(1, 917632, '\p{^Is_Dep=_	TRUE}', "");
Expect(1, 917632, '\P{Is_Dep=_	TRUE}', "");
Expect(0, 917632, '\P{^Is_Dep=_	TRUE}', "");
Error('\p{Diacritic=:=No}');
Error('\P{Diacritic=:=No}');
Expect(1, 119214, '\p{Diacritic=no}', "");
Expect(0, 119214, '\p{^Diacritic=no}', "");
Expect(0, 119214, '\P{Diacritic=no}', "");
Expect(1, 119214, '\P{^Diacritic=no}', "");
Expect(0, 119213, '\p{Diacritic=no}', "");
Expect(1, 119213, '\p{^Diacritic=no}', "");
Expect(1, 119213, '\P{Diacritic=no}', "");
Expect(0, 119213, '\P{^Diacritic=no}', "");
Expect(1, 119214, '\p{Diacritic=	_no}', "");
Expect(0, 119214, '\p{^Diacritic=	_no}', "");
Expect(0, 119214, '\P{Diacritic=	_no}', "");
Expect(1, 119214, '\P{^Diacritic=	_no}', "");
Expect(0, 119213, '\p{Diacritic=	_no}', "");
Expect(1, 119213, '\p{^Diacritic=	_no}', "");
Expect(1, 119213, '\P{Diacritic=	_no}', "");
Expect(0, 119213, '\P{^Diacritic=	_no}', "");
Error('\p{Dia=__N/a/}');
Error('\P{Dia=__N/a/}');
Expect(1, 119214, '\p{Dia=n}', "");
Expect(0, 119214, '\p{^Dia=n}', "");
Expect(0, 119214, '\P{Dia=n}', "");
Expect(1, 119214, '\P{^Dia=n}', "");
Expect(0, 119213, '\p{Dia=n}', "");
Expect(1, 119213, '\p{^Dia=n}', "");
Expect(1, 119213, '\P{Dia=n}', "");
Expect(0, 119213, '\P{^Dia=n}', "");
Expect(1, 119214, '\p{Dia=_N}', "");
Expect(0, 119214, '\p{^Dia=_N}', "");
Expect(0, 119214, '\P{Dia=_N}', "");
Expect(1, 119214, '\P{^Dia=_N}', "");
Expect(0, 119213, '\p{Dia=_N}', "");
Expect(1, 119213, '\p{^Dia=_N}', "");
Expect(1, 119213, '\P{Dia=_N}', "");
Expect(0, 119213, '\P{^Dia=_N}', "");
Error('\p{Is_Diacritic: F/a/}');
Error('\P{Is_Diacritic: F/a/}');
Expect(1, 119214, '\p{Is_Diacritic=f}', "");
Expect(0, 119214, '\p{^Is_Diacritic=f}', "");
Expect(0, 119214, '\P{Is_Diacritic=f}', "");
Expect(1, 119214, '\P{^Is_Diacritic=f}', "");
Expect(0, 119213, '\p{Is_Diacritic=f}', "");
Expect(1, 119213, '\p{^Is_Diacritic=f}', "");
Expect(1, 119213, '\P{Is_Diacritic=f}', "");
Expect(0, 119213, '\P{^Is_Diacritic=f}', "");
Expect(1, 119214, '\p{Is_Diacritic= f}', "");
Expect(0, 119214, '\p{^Is_Diacritic= f}', "");
Expect(0, 119214, '\P{Is_Diacritic= f}', "");
Expect(1, 119214, '\P{^Is_Diacritic= f}', "");
Expect(0, 119213, '\p{Is_Diacritic= f}', "");
Expect(1, 119213, '\p{^Is_Diacritic= f}', "");
Expect(1, 119213, '\P{Is_Diacritic= f}', "");
Expect(0, 119213, '\P{^Is_Diacritic= f}', "");
Error('\p{Is_Dia=:=-False}');
Error('\P{Is_Dia=:=-False}');
Expect(1, 119214, '\p{Is_Dia=false}', "");
Expect(0, 119214, '\p{^Is_Dia=false}', "");
Expect(0, 119214, '\P{Is_Dia=false}', "");
Expect(1, 119214, '\P{^Is_Dia=false}', "");
Expect(0, 119213, '\p{Is_Dia=false}', "");
Expect(1, 119213, '\p{^Is_Dia=false}', "");
Expect(1, 119213, '\P{Is_Dia=false}', "");
Expect(0, 119213, '\P{^Is_Dia=false}', "");
Expect(1, 119214, '\p{Is_Dia:   _	False}', "");
Expect(0, 119214, '\p{^Is_Dia:   _	False}', "");
Expect(0, 119214, '\P{Is_Dia:   _	False}', "");
Expect(1, 119214, '\P{^Is_Dia:   _	False}', "");
Expect(0, 119213, '\p{Is_Dia:   _	False}', "");
Expect(1, 119213, '\p{^Is_Dia:   _	False}', "");
Expect(1, 119213, '\P{Is_Dia:   _	False}', "");
Expect(0, 119213, '\P{^Is_Dia:   _	False}', "");
Error('\p{Diacritic::=-	Yes}');
Error('\P{Diacritic::=-	Yes}');
Expect(1, 119213, '\p{Diacritic=yes}', "");
Expect(0, 119213, '\p{^Diacritic=yes}', "");
Expect(0, 119213, '\P{Diacritic=yes}', "");
Expect(1, 119213, '\P{^Diacritic=yes}', "");
Expect(0, 119214, '\p{Diacritic=yes}', "");
Expect(1, 119214, '\p{^Diacritic=yes}', "");
Expect(1, 119214, '\P{Diacritic=yes}', "");
Expect(0, 119214, '\P{^Diacritic=yes}', "");
Expect(1, 119213, '\p{Diacritic=-	Yes}', "");
Expect(0, 119213, '\p{^Diacritic=-	Yes}', "");
Expect(0, 119213, '\P{Diacritic=-	Yes}', "");
Expect(1, 119213, '\P{^Diacritic=-	Yes}', "");
Expect(0, 119214, '\p{Diacritic=-	Yes}', "");
Expect(1, 119214, '\p{^Diacritic=-	Yes}', "");
Expect(1, 119214, '\P{Diacritic=-	Yes}', "");
Expect(0, 119214, '\P{^Diacritic=-	Yes}', "");
Error('\p{Dia=:=	Y}');
Error('\P{Dia=:=	Y}');
Expect(1, 119213, '\p{Dia=y}', "");
Expect(0, 119213, '\p{^Dia=y}', "");
Expect(0, 119213, '\P{Dia=y}', "");
Expect(1, 119213, '\P{^Dia=y}', "");
Expect(0, 119214, '\p{Dia=y}', "");
Expect(1, 119214, '\p{^Dia=y}', "");
Expect(1, 119214, '\P{Dia=y}', "");
Expect(0, 119214, '\P{^Dia=y}', "");
Expect(1, 119213, '\p{Dia=-_Y}', "");
Expect(0, 119213, '\p{^Dia=-_Y}', "");
Expect(0, 119213, '\P{Dia=-_Y}', "");
Expect(1, 119213, '\P{^Dia=-_Y}', "");
Expect(0, 119214, '\p{Dia=-_Y}', "");
Expect(1, 119214, '\p{^Dia=-_Y}', "");
Expect(1, 119214, '\P{Dia=-_Y}', "");
Expect(0, 119214, '\P{^Dia=-_Y}', "");
Error('\p{Is_Diacritic=:= T}');
Error('\P{Is_Diacritic=:= T}');
Expect(1, 119213, '\p{Is_Diacritic=t}', "");
Expect(0, 119213, '\p{^Is_Diacritic=t}', "");
Expect(0, 119213, '\P{Is_Diacritic=t}', "");
Expect(1, 119213, '\P{^Is_Diacritic=t}', "");
Expect(0, 119214, '\p{Is_Diacritic=t}', "");
Expect(1, 119214, '\p{^Is_Diacritic=t}', "");
Expect(1, 119214, '\P{Is_Diacritic=t}', "");
Expect(0, 119214, '\P{^Is_Diacritic=t}', "");
Expect(1, 119213, '\p{Is_Diacritic=  T}', "");
Expect(0, 119213, '\p{^Is_Diacritic=  T}', "");
Expect(0, 119213, '\P{Is_Diacritic=  T}', "");
Expect(1, 119213, '\P{^Is_Diacritic=  T}', "");
Expect(0, 119214, '\p{Is_Diacritic=  T}', "");
Expect(1, 119214, '\p{^Is_Diacritic=  T}', "");
Expect(1, 119214, '\P{Is_Diacritic=  T}', "");
Expect(0, 119214, '\P{^Is_Diacritic=  T}', "");
Error('\p{Is_Dia=:=True}');
Error('\P{Is_Dia=:=True}');
Expect(1, 119213, '\p{Is_Dia:   true}', "");
Expect(0, 119213, '\p{^Is_Dia:   true}', "");
Expect(0, 119213, '\P{Is_Dia:   true}', "");
Expect(1, 119213, '\P{^Is_Dia:   true}', "");
Expect(0, 119214, '\p{Is_Dia:   true}', "");
Expect(1, 119214, '\p{^Is_Dia:   true}', "");
Expect(1, 119214, '\P{Is_Dia:   true}', "");
Expect(0, 119214, '\P{^Is_Dia:   true}', "");
Expect(1, 119213, '\p{Is_Dia=	_True}', "");
Expect(0, 119213, '\p{^Is_Dia=	_True}', "");
Expect(0, 119213, '\P{Is_Dia=	_True}', "");
Expect(1, 119213, '\P{^Is_Dia=	_True}', "");
Expect(0, 119214, '\p{Is_Dia=	_True}', "");
Expect(1, 119214, '\p{^Is_Dia=	_True}', "");
Expect(1, 119214, '\P{Is_Dia=	_True}', "");
Expect(0, 119214, '\P{^Is_Dia=	_True}', "");
Error('\p{Extender=/a/no}');
Error('\P{Extender=/a/no}');
Expect(1, 65393, '\p{Extender=no}', "");
Expect(0, 65393, '\p{^Extender=no}', "");
Expect(0, 65393, '\P{Extender=no}', "");
Expect(1, 65393, '\P{^Extender=no}', "");
Expect(0, 65392, '\p{Extender=no}', "");
Expect(1, 65392, '\p{^Extender=no}', "");
Expect(1, 65392, '\P{Extender=no}', "");
Expect(0, 65392, '\P{^Extender=no}', "");
Expect(1, 65393, '\p{Extender:	_no}', "");
Expect(0, 65393, '\p{^Extender:	_no}', "");
Expect(0, 65393, '\P{Extender:	_no}', "");
Expect(1, 65393, '\P{^Extender:	_no}', "");
Expect(0, 65392, '\p{Extender:	_no}', "");
Expect(1, 65392, '\p{^Extender:	_no}', "");
Expect(1, 65392, '\P{Extender:	_no}', "");
Expect(0, 65392, '\P{^Extender:	_no}', "");
Error('\p{Ext=_/a/N}');
Error('\P{Ext=_/a/N}');
Expect(1, 65393, '\p{Ext=n}', "");
Expect(0, 65393, '\p{^Ext=n}', "");
Expect(0, 65393, '\P{Ext=n}', "");
Expect(1, 65393, '\P{^Ext=n}', "");
Expect(0, 65392, '\p{Ext=n}', "");
Expect(1, 65392, '\p{^Ext=n}', "");
Expect(1, 65392, '\P{Ext=n}', "");
Expect(0, 65392, '\P{^Ext=n}', "");
Expect(1, 65393, '\p{Ext=_N}', "");
Expect(0, 65393, '\p{^Ext=_N}', "");
Expect(0, 65393, '\P{Ext=_N}', "");
Expect(1, 65393, '\P{^Ext=_N}', "");
Expect(0, 65392, '\p{Ext=_N}', "");
Expect(1, 65392, '\p{^Ext=_N}', "");
Expect(1, 65392, '\P{Ext=_N}', "");
Expect(0, 65392, '\P{^Ext=_N}', "");
Error('\p{Is_Extender=	 f/a/}');
Error('\P{Is_Extender=	 f/a/}');
Expect(1, 65393, '\p{Is_Extender:   f}', "");
Expect(0, 65393, '\p{^Is_Extender:   f}', "");
Expect(0, 65393, '\P{Is_Extender:   f}', "");
Expect(1, 65393, '\P{^Is_Extender:   f}', "");
Expect(0, 65392, '\p{Is_Extender:   f}', "");
Expect(1, 65392, '\p{^Is_Extender:   f}', "");
Expect(1, 65392, '\P{Is_Extender:   f}', "");
Expect(0, 65392, '\P{^Is_Extender:   f}', "");
Expect(1, 65393, '\p{Is_Extender=-_F}', "");
Expect(0, 65393, '\p{^Is_Extender=-_F}', "");
Expect(0, 65393, '\P{Is_Extender=-_F}', "");
Expect(1, 65393, '\P{^Is_Extender=-_F}', "");
Expect(0, 65392, '\p{Is_Extender=-_F}', "");
Expect(1, 65392, '\p{^Is_Extender=-_F}', "");
Expect(1, 65392, '\P{Is_Extender=-_F}', "");
Expect(0, 65392, '\P{^Is_Extender=-_F}', "");
Error('\p{Is_Ext=--false/a/}');
Error('\P{Is_Ext=--false/a/}');
Expect(1, 65393, '\p{Is_Ext=false}', "");
Expect(0, 65393, '\p{^Is_Ext=false}', "");
Expect(0, 65393, '\P{Is_Ext=false}', "");
Expect(1, 65393, '\P{^Is_Ext=false}', "");
Expect(0, 65392, '\p{Is_Ext=false}', "");
Expect(1, 65392, '\p{^Is_Ext=false}', "");
Expect(1, 65392, '\P{Is_Ext=false}', "");
Expect(0, 65392, '\P{^Is_Ext=false}', "");
Expect(1, 65393, '\p{Is_Ext= -False}', "");
Expect(0, 65393, '\p{^Is_Ext= -False}', "");
Expect(0, 65393, '\P{Is_Ext= -False}', "");
Expect(1, 65393, '\P{^Is_Ext= -False}', "");
Expect(0, 65392, '\p{Is_Ext= -False}', "");
Expect(1, 65392, '\p{^Is_Ext= -False}', "");
Expect(1, 65392, '\P{Is_Ext= -False}', "");
Expect(0, 65392, '\P{^Is_Ext= -False}', "");
Error('\p{Extender= /a/Yes}');
Error('\P{Extender= /a/Yes}');
Expect(1, 65392, '\p{Extender=yes}', "");
Expect(0, 65392, '\p{^Extender=yes}', "");
Expect(0, 65392, '\P{Extender=yes}', "");
Expect(1, 65392, '\P{^Extender=yes}', "");
Expect(0, 65393, '\p{Extender=yes}', "");
Expect(1, 65393, '\p{^Extender=yes}', "");
Expect(1, 65393, '\P{Extender=yes}', "");
Expect(0, 65393, '\P{^Extender=yes}', "");
Expect(1, 65392, '\p{Extender=  Yes}', "");
Expect(0, 65392, '\p{^Extender=  Yes}', "");
Expect(0, 65392, '\P{Extender=  Yes}', "");
Expect(1, 65392, '\P{^Extender=  Yes}', "");
Expect(0, 65393, '\p{Extender=  Yes}', "");
Expect(1, 65393, '\p{^Extender=  Yes}', "");
Expect(1, 65393, '\P{Extender=  Yes}', "");
Expect(0, 65393, '\P{^Extender=  Yes}', "");
Error('\p{Ext=/a/- Y}');
Error('\P{Ext=/a/- Y}');
Expect(1, 65392, '\p{Ext=y}', "");
Expect(0, 65392, '\p{^Ext=y}', "");
Expect(0, 65392, '\P{Ext=y}', "");
Expect(1, 65392, '\P{^Ext=y}', "");
Expect(0, 65393, '\p{Ext=y}', "");
Expect(1, 65393, '\p{^Ext=y}', "");
Expect(1, 65393, '\P{Ext=y}', "");
Expect(0, 65393, '\P{^Ext=y}', "");
Expect(1, 65392, '\p{Ext=  y}', "");
Expect(0, 65392, '\p{^Ext=  y}', "");
Expect(0, 65392, '\P{Ext=  y}', "");
Expect(1, 65392, '\P{^Ext=  y}', "");
Expect(0, 65393, '\p{Ext=  y}', "");
Expect(1, 65393, '\p{^Ext=  y}', "");
Expect(1, 65393, '\P{Ext=  y}', "");
Expect(0, 65393, '\P{^Ext=  y}', "");
Error('\p{Is_Extender=:=	T}');
Error('\P{Is_Extender=:=	T}');
Expect(1, 65392, '\p{Is_Extender=t}', "");
Expect(0, 65392, '\p{^Is_Extender=t}', "");
Expect(0, 65392, '\P{Is_Extender=t}', "");
Expect(1, 65392, '\P{^Is_Extender=t}', "");
Expect(0, 65393, '\p{Is_Extender=t}', "");
Expect(1, 65393, '\p{^Is_Extender=t}', "");
Expect(1, 65393, '\P{Is_Extender=t}', "");
Expect(0, 65393, '\P{^Is_Extender=t}', "");
Expect(1, 65392, '\p{Is_Extender= T}', "");
Expect(0, 65392, '\p{^Is_Extender= T}', "");
Expect(0, 65392, '\P{Is_Extender= T}', "");
Expect(1, 65392, '\P{^Is_Extender= T}', "");
Expect(0, 65393, '\p{Is_Extender= T}', "");
Expect(1, 65393, '\p{^Is_Extender= T}', "");
Expect(1, 65393, '\P{Is_Extender= T}', "");
Expect(0, 65393, '\P{^Is_Extender= T}', "");
Error('\p{Is_Ext=- true/a/}');
Error('\P{Is_Ext=- true/a/}');
Expect(1, 65392, '\p{Is_Ext=true}', "");
Expect(0, 65392, '\p{^Is_Ext=true}', "");
Expect(0, 65392, '\P{Is_Ext=true}', "");
Expect(1, 65392, '\P{^Is_Ext=true}', "");
Expect(0, 65393, '\p{Is_Ext=true}', "");
Expect(1, 65393, '\p{^Is_Ext=true}', "");
Expect(1, 65393, '\P{Is_Ext=true}', "");
Expect(0, 65393, '\P{^Is_Ext=true}', "");
Expect(1, 65392, '\p{Is_Ext=True}', "");
Expect(0, 65392, '\p{^Is_Ext=True}', "");
Expect(0, 65392, '\P{Is_Ext=True}', "");
Expect(1, 65392, '\P{^Is_Ext=True}', "");
Expect(0, 65393, '\p{Is_Ext=True}', "");
Expect(1, 65393, '\p{^Is_Ext=True}', "");
Expect(1, 65393, '\P{Is_Ext=True}', "");
Expect(0, 65393, '\P{^Is_Ext=True}', "");
Error('\p{Grapheme_Cluster_Break=/a/	_CONTROL}');
Error('\P{Grapheme_Cluster_Break=/a/	_CONTROL}');
Expect(1, 921599, '\p{Grapheme_Cluster_Break=control}', "");
Expect(0, 921599, '\p{^Grapheme_Cluster_Break=control}', "");
Expect(0, 921599, '\P{Grapheme_Cluster_Break=control}', "");
Expect(1, 921599, '\P{^Grapheme_Cluster_Break=control}', "");
Expect(0, 921600, '\p{Grapheme_Cluster_Break=control}', "");
Expect(1, 921600, '\p{^Grapheme_Cluster_Break=control}', "");
Expect(1, 921600, '\P{Grapheme_Cluster_Break=control}', "");
Expect(0, 921600, '\P{^Grapheme_Cluster_Break=control}', "");
Expect(1, 921599, '\p{Grapheme_Cluster_Break=Control}', "");
Expect(0, 921599, '\p{^Grapheme_Cluster_Break=Control}', "");
Expect(0, 921599, '\P{Grapheme_Cluster_Break=Control}', "");
Expect(1, 921599, '\P{^Grapheme_Cluster_Break=Control}', "");
Expect(0, 921600, '\p{Grapheme_Cluster_Break=Control}', "");
Expect(1, 921600, '\p{^Grapheme_Cluster_Break=Control}', "");
Expect(1, 921600, '\P{Grapheme_Cluster_Break=Control}', "");
Expect(0, 921600, '\P{^Grapheme_Cluster_Break=Control}', "");
Error('\p{GCB=  CN/a/}');
Error('\P{GCB=  CN/a/}');
Expect(1, 921599, '\p{GCB=cn}', "");
Expect(0, 921599, '\p{^GCB=cn}', "");
Expect(0, 921599, '\P{GCB=cn}', "");
Expect(1, 921599, '\P{^GCB=cn}', "");
Expect(0, 921600, '\p{GCB=cn}', "");
Expect(1, 921600, '\p{^GCB=cn}', "");
Expect(1, 921600, '\P{GCB=cn}', "");
Expect(0, 921600, '\P{^GCB=cn}', "");
Expect(1, 921599, '\p{GCB= CN}', "");
Expect(0, 921599, '\p{^GCB= CN}', "");
Expect(0, 921599, '\P{GCB= CN}', "");
Expect(1, 921599, '\P{^GCB= CN}', "");
Expect(0, 921600, '\p{GCB= CN}', "");
Expect(1, 921600, '\p{^GCB= CN}', "");
Expect(1, 921600, '\P{GCB= CN}', "");
Expect(0, 921600, '\P{^GCB= CN}', "");
Error('\p{Is_Grapheme_Cluster_Break=:=_CONTROL}');
Error('\P{Is_Grapheme_Cluster_Break=:=_CONTROL}');
Expect(1, 921599, '\p{Is_Grapheme_Cluster_Break=control}', "");
Expect(0, 921599, '\p{^Is_Grapheme_Cluster_Break=control}', "");
Expect(0, 921599, '\P{Is_Grapheme_Cluster_Break=control}', "");
Expect(1, 921599, '\P{^Is_Grapheme_Cluster_Break=control}', "");
Expect(0, 921600, '\p{Is_Grapheme_Cluster_Break=control}', "");
Expect(1, 921600, '\p{^Is_Grapheme_Cluster_Break=control}', "");
Expect(1, 921600, '\P{Is_Grapheme_Cluster_Break=control}', "");
Expect(0, 921600, '\P{^Is_Grapheme_Cluster_Break=control}', "");
Expect(1, 921599, '\p{Is_Grapheme_Cluster_Break=_Control}', "");
Expect(0, 921599, '\p{^Is_Grapheme_Cluster_Break=_Control}', "");
Expect(0, 921599, '\P{Is_Grapheme_Cluster_Break=_Control}', "");
Expect(1, 921599, '\P{^Is_Grapheme_Cluster_Break=_Control}', "");
Expect(0, 921600, '\p{Is_Grapheme_Cluster_Break=_Control}', "");
Expect(1, 921600, '\p{^Is_Grapheme_Cluster_Break=_Control}', "");
Expect(1, 921600, '\P{Is_Grapheme_Cluster_Break=_Control}', "");
Expect(0, 921600, '\P{^Is_Grapheme_Cluster_Break=_Control}', "");
Error('\p{Is_GCB=		cn:=}');
Error('\P{Is_GCB=		cn:=}');
Expect(1, 921599, '\p{Is_GCB=cn}', "");
Expect(0, 921599, '\p{^Is_GCB=cn}', "");
Expect(0, 921599, '\P{Is_GCB=cn}', "");
Expect(1, 921599, '\P{^Is_GCB=cn}', "");
Expect(0, 921600, '\p{Is_GCB=cn}', "");
Expect(1, 921600, '\p{^Is_GCB=cn}', "");
Expect(1, 921600, '\P{Is_GCB=cn}', "");
Expect(0, 921600, '\P{^Is_GCB=cn}', "");
Expect(1, 921599, '\p{Is_GCB=	 CN}', "");
Expect(0, 921599, '\p{^Is_GCB=	 CN}', "");
Expect(0, 921599, '\P{Is_GCB=	 CN}', "");
Expect(1, 921599, '\P{^Is_GCB=	 CN}', "");
Expect(0, 921600, '\p{Is_GCB=	 CN}', "");
Expect(1, 921600, '\p{^Is_GCB=	 CN}', "");
Expect(1, 921600, '\P{Is_GCB=	 CN}', "");
Expect(0, 921600, '\P{^Is_GCB=	 CN}', "");
Error('\p{Grapheme_Cluster_Break= :=CR}');
Error('\P{Grapheme_Cluster_Break= :=CR}');
Expect(1, 13, '\p{Grapheme_Cluster_Break=cr}', "");
Expect(0, 13, '\p{^Grapheme_Cluster_Break=cr}', "");
Expect(0, 13, '\P{Grapheme_Cluster_Break=cr}', "");
Expect(1, 13, '\P{^Grapheme_Cluster_Break=cr}', "");
Expect(0, 14, '\p{Grapheme_Cluster_Break=cr}', "");
Expect(1, 14, '\p{^Grapheme_Cluster_Break=cr}', "");
Expect(1, 14, '\P{Grapheme_Cluster_Break=cr}', "");
Expect(0, 14, '\P{^Grapheme_Cluster_Break=cr}', "");
Expect(1, 13, '\p{Grapheme_Cluster_Break= -CR}', "");
Expect(0, 13, '\p{^Grapheme_Cluster_Break= -CR}', "");
Expect(0, 13, '\P{Grapheme_Cluster_Break= -CR}', "");
Expect(1, 13, '\P{^Grapheme_Cluster_Break= -CR}', "");
Expect(0, 14, '\p{Grapheme_Cluster_Break= -CR}', "");
Expect(1, 14, '\p{^Grapheme_Cluster_Break= -CR}', "");
Expect(1, 14, '\P{Grapheme_Cluster_Break= -CR}', "");
Expect(0, 14, '\P{^Grapheme_Cluster_Break= -CR}', "");
Error('\p{GCB=	_CR:=}');
Error('\P{GCB=	_CR:=}');
Expect(1, 13, '\p{GCB=cr}', "");
Expect(0, 13, '\p{^GCB=cr}', "");
Expect(0, 13, '\P{GCB=cr}', "");
Expect(1, 13, '\P{^GCB=cr}', "");
Expect(0, 14, '\p{GCB=cr}', "");
Expect(1, 14, '\p{^GCB=cr}', "");
Expect(1, 14, '\P{GCB=cr}', "");
Expect(0, 14, '\P{^GCB=cr}', "");
Expect(1, 13, '\p{GCB=-CR}', "");
Expect(0, 13, '\p{^GCB=-CR}', "");
Expect(0, 13, '\P{GCB=-CR}', "");
Expect(1, 13, '\P{^GCB=-CR}', "");
Expect(0, 14, '\p{GCB=-CR}', "");
Expect(1, 14, '\p{^GCB=-CR}', "");
Expect(1, 14, '\P{GCB=-CR}', "");
Expect(0, 14, '\P{^GCB=-CR}', "");
Error('\p{Is_Grapheme_Cluster_Break=/a/	-cr}');
Error('\P{Is_Grapheme_Cluster_Break=/a/	-cr}');
Expect(1, 13, '\p{Is_Grapheme_Cluster_Break=cr}', "");
Expect(0, 13, '\p{^Is_Grapheme_Cluster_Break=cr}', "");
Expect(0, 13, '\P{Is_Grapheme_Cluster_Break=cr}', "");
Expect(1, 13, '\P{^Is_Grapheme_Cluster_Break=cr}', "");
Expect(0, 14, '\p{Is_Grapheme_Cluster_Break=cr}', "");
Expect(1, 14, '\p{^Is_Grapheme_Cluster_Break=cr}', "");
Expect(1, 14, '\P{Is_Grapheme_Cluster_Break=cr}', "");
Expect(0, 14, '\P{^Is_Grapheme_Cluster_Break=cr}', "");
Expect(1, 13, '\p{Is_Grapheme_Cluster_Break=_CR}', "");
Expect(0, 13, '\p{^Is_Grapheme_Cluster_Break=_CR}', "");
Expect(0, 13, '\P{Is_Grapheme_Cluster_Break=_CR}', "");
Expect(1, 13, '\P{^Is_Grapheme_Cluster_Break=_CR}', "");
Expect(0, 14, '\p{Is_Grapheme_Cluster_Break=_CR}', "");
Expect(1, 14, '\p{^Is_Grapheme_Cluster_Break=_CR}', "");
Expect(1, 14, '\P{Is_Grapheme_Cluster_Break=_CR}', "");
Expect(0, 14, '\P{^Is_Grapheme_Cluster_Break=_CR}', "");
Error('\p{Is_GCB=-CR:=}');
Error('\P{Is_GCB=-CR:=}');
Expect(1, 13, '\p{Is_GCB=cr}', "");
Expect(0, 13, '\p{^Is_GCB=cr}', "");
Expect(0, 13, '\P{Is_GCB=cr}', "");
Expect(1, 13, '\P{^Is_GCB=cr}', "");
Expect(0, 14, '\p{Is_GCB=cr}', "");
Expect(1, 14, '\p{^Is_GCB=cr}', "");
Expect(1, 14, '\P{Is_GCB=cr}', "");
Expect(0, 14, '\P{^Is_GCB=cr}', "");
Expect(1, 13, '\p{Is_GCB=	 CR}', "");
Expect(0, 13, '\p{^Is_GCB=	 CR}', "");
Expect(0, 13, '\P{Is_GCB=	 CR}', "");
Expect(1, 13, '\P{^Is_GCB=	 CR}', "");
Expect(0, 14, '\p{Is_GCB=	 CR}', "");
Expect(1, 14, '\p{^Is_GCB=	 CR}', "");
Expect(1, 14, '\P{Is_GCB=	 CR}', "");
Expect(0, 14, '\P{^Is_GCB=	 CR}', "");
Error('\p{Grapheme_Cluster_Break: -:=EXTEND}');
Error('\P{Grapheme_Cluster_Break: -:=EXTEND}');
Expect(1, 917999, '\p{Grapheme_Cluster_Break=extend}', "");
Expect(0, 917999, '\p{^Grapheme_Cluster_Break=extend}', "");
Expect(0, 917999, '\P{Grapheme_Cluster_Break=extend}', "");
Expect(1, 917999, '\P{^Grapheme_Cluster_Break=extend}', "");
Expect(0, 918000, '\p{Grapheme_Cluster_Break=extend}', "");
Expect(1, 918000, '\p{^Grapheme_Cluster_Break=extend}', "");
Expect(1, 918000, '\P{Grapheme_Cluster_Break=extend}', "");
Expect(0, 918000, '\P{^Grapheme_Cluster_Break=extend}', "");
Expect(1, 917999, '\p{Grapheme_Cluster_Break=_-Extend}', "");
Expect(0, 917999, '\p{^Grapheme_Cluster_Break=_-Extend}', "");
Expect(0, 917999, '\P{Grapheme_Cluster_Break=_-Extend}', "");
Expect(1, 917999, '\P{^Grapheme_Cluster_Break=_-Extend}', "");
Expect(0, 918000, '\p{Grapheme_Cluster_Break=_-Extend}', "");
Expect(1, 918000, '\p{^Grapheme_Cluster_Break=_-Extend}', "");
Expect(1, 918000, '\P{Grapheme_Cluster_Break=_-Extend}', "");
Expect(0, 918000, '\P{^Grapheme_Cluster_Break=_-Extend}', "");
Error('\p{GCB= :=EX}');
Error('\P{GCB= :=EX}');
Expect(1, 917999, '\p{GCB=ex}', "");
Expect(0, 917999, '\p{^GCB=ex}', "");
Expect(0, 917999, '\P{GCB=ex}', "");
Expect(1, 917999, '\P{^GCB=ex}', "");
Expect(0, 918000, '\p{GCB=ex}', "");
Expect(1, 918000, '\p{^GCB=ex}', "");
Expect(1, 918000, '\P{GCB=ex}', "");
Expect(0, 918000, '\P{^GCB=ex}', "");
Expect(1, 917999, '\p{GCB=EX}', "");
Expect(0, 917999, '\p{^GCB=EX}', "");
Expect(0, 917999, '\P{GCB=EX}', "");
Expect(1, 917999, '\P{^GCB=EX}', "");
Expect(0, 918000, '\p{GCB=EX}', "");
Expect(1, 918000, '\p{^GCB=EX}', "");
Expect(1, 918000, '\P{GCB=EX}', "");
Expect(0, 918000, '\P{^GCB=EX}', "");
Error('\p{Is_Grapheme_Cluster_Break=/a/-Extend}');
Error('\P{Is_Grapheme_Cluster_Break=/a/-Extend}');
Expect(1, 917999, '\p{Is_Grapheme_Cluster_Break=extend}', "");
Expect(0, 917999, '\p{^Is_Grapheme_Cluster_Break=extend}', "");
Expect(0, 917999, '\P{Is_Grapheme_Cluster_Break=extend}', "");
Expect(1, 917999, '\P{^Is_Grapheme_Cluster_Break=extend}', "");
Expect(0, 918000, '\p{Is_Grapheme_Cluster_Break=extend}', "");
Expect(1, 918000, '\p{^Is_Grapheme_Cluster_Break=extend}', "");
Expect(1, 918000, '\P{Is_Grapheme_Cluster_Break=extend}', "");
Expect(0, 918000, '\P{^Is_Grapheme_Cluster_Break=extend}', "");
Expect(1, 917999, '\p{Is_Grapheme_Cluster_Break=__Extend}', "");
Expect(0, 917999, '\p{^Is_Grapheme_Cluster_Break=__Extend}', "");
Expect(0, 917999, '\P{Is_Grapheme_Cluster_Break=__Extend}', "");
Expect(1, 917999, '\P{^Is_Grapheme_Cluster_Break=__Extend}', "");
Expect(0, 918000, '\p{Is_Grapheme_Cluster_Break=__Extend}', "");
Expect(1, 918000, '\p{^Is_Grapheme_Cluster_Break=__Extend}', "");
Expect(1, 918000, '\P{Is_Grapheme_Cluster_Break=__Extend}', "");
Expect(0, 918000, '\P{^Is_Grapheme_Cluster_Break=__Extend}', "");
Error('\p{Is_GCB: _ EX:=}');
Error('\P{Is_GCB: _ EX:=}');
Expect(1, 917999, '\p{Is_GCB=ex}', "");
Expect(0, 917999, '\p{^Is_GCB=ex}', "");
Expect(0, 917999, '\P{Is_GCB=ex}', "");
Expect(1, 917999, '\P{^Is_GCB=ex}', "");
Expect(0, 918000, '\p{Is_GCB=ex}', "");
Expect(1, 918000, '\p{^Is_GCB=ex}', "");
Expect(1, 918000, '\P{Is_GCB=ex}', "");
Expect(0, 918000, '\P{^Is_GCB=ex}', "");
Expect(1, 917999, '\p{Is_GCB= EX}', "");
Expect(0, 917999, '\p{^Is_GCB= EX}', "");
Expect(0, 917999, '\P{Is_GCB= EX}', "");
Expect(1, 917999, '\P{^Is_GCB= EX}', "");
Expect(0, 918000, '\p{Is_GCB= EX}', "");
Expect(1, 918000, '\p{^Is_GCB= EX}', "");
Expect(1, 918000, '\P{Is_GCB= EX}', "");
Expect(0, 918000, '\P{^Is_GCB= EX}', "");
Error('\p{Grapheme_Cluster_Break:	L/a/}');
Error('\P{Grapheme_Cluster_Break:	L/a/}');
Expect(1, 43388, '\p{Grapheme_Cluster_Break=l}', "");
Expect(0, 43388, '\p{^Grapheme_Cluster_Break=l}', "");
Expect(0, 43388, '\P{Grapheme_Cluster_Break=l}', "");
Expect(1, 43388, '\P{^Grapheme_Cluster_Break=l}', "");
Expect(0, 43389, '\p{Grapheme_Cluster_Break=l}', "");
Expect(1, 43389, '\p{^Grapheme_Cluster_Break=l}', "");
Expect(1, 43389, '\P{Grapheme_Cluster_Break=l}', "");
Expect(0, 43389, '\P{^Grapheme_Cluster_Break=l}', "");
Expect(1, 43388, '\p{Grapheme_Cluster_Break=__L}', "");
Expect(0, 43388, '\p{^Grapheme_Cluster_Break=__L}', "");
Expect(0, 43388, '\P{Grapheme_Cluster_Break=__L}', "");
Expect(1, 43388, '\P{^Grapheme_Cluster_Break=__L}', "");
Expect(0, 43389, '\p{Grapheme_Cluster_Break=__L}', "");
Expect(1, 43389, '\p{^Grapheme_Cluster_Break=__L}', "");
Expect(1, 43389, '\P{Grapheme_Cluster_Break=__L}', "");
Expect(0, 43389, '\P{^Grapheme_Cluster_Break=__L}', "");
Error('\p{GCB=__l:=}');
Error('\P{GCB=__l:=}');
Expect(1, 43388, '\p{GCB=l}', "");
Expect(0, 43388, '\p{^GCB=l}', "");
Expect(0, 43388, '\P{GCB=l}', "");
Expect(1, 43388, '\P{^GCB=l}', "");
Expect(0, 43389, '\p{GCB=l}', "");
Expect(1, 43389, '\p{^GCB=l}', "");
Expect(1, 43389, '\P{GCB=l}', "");
Expect(0, 43389, '\P{^GCB=l}', "");
Expect(1, 43388, '\p{GCB= -L}', "");
Expect(0, 43388, '\p{^GCB= -L}', "");
Expect(0, 43388, '\P{GCB= -L}', "");
Expect(1, 43388, '\P{^GCB= -L}', "");
Expect(0, 43389, '\p{GCB= -L}', "");
Expect(1, 43389, '\p{^GCB= -L}', "");
Expect(1, 43389, '\P{GCB= -L}', "");
Expect(0, 43389, '\P{^GCB= -L}', "");
Error('\p{Is_Grapheme_Cluster_Break=/a/	_l}');
Error('\P{Is_Grapheme_Cluster_Break=/a/	_l}');
Expect(1, 43388, '\p{Is_Grapheme_Cluster_Break=l}', "");
Expect(0, 43388, '\p{^Is_Grapheme_Cluster_Break=l}', "");
Expect(0, 43388, '\P{Is_Grapheme_Cluster_Break=l}', "");
Expect(1, 43388, '\P{^Is_Grapheme_Cluster_Break=l}', "");
Expect(0, 43389, '\p{Is_Grapheme_Cluster_Break=l}', "");
Expect(1, 43389, '\p{^Is_Grapheme_Cluster_Break=l}', "");
Expect(1, 43389, '\P{Is_Grapheme_Cluster_Break=l}', "");
Expect(0, 43389, '\P{^Is_Grapheme_Cluster_Break=l}', "");
Expect(1, 43388, '\p{Is_Grapheme_Cluster_Break=	L}', "");
Expect(0, 43388, '\p{^Is_Grapheme_Cluster_Break=	L}', "");
Expect(0, 43388, '\P{Is_Grapheme_Cluster_Break=	L}', "");
Expect(1, 43388, '\P{^Is_Grapheme_Cluster_Break=	L}', "");
Expect(0, 43389, '\p{Is_Grapheme_Cluster_Break=	L}', "");
Expect(1, 43389, '\p{^Is_Grapheme_Cluster_Break=	L}', "");
Expect(1, 43389, '\P{Is_Grapheme_Cluster_Break=	L}', "");
Expect(0, 43389, '\P{^Is_Grapheme_Cluster_Break=	L}', "");
Error('\p{Is_GCB= 	L:=}');
Error('\P{Is_GCB= 	L:=}');
Expect(1, 43388, '\p{Is_GCB=l}', "");
Expect(0, 43388, '\p{^Is_GCB=l}', "");
Expect(0, 43388, '\P{Is_GCB=l}', "");
Expect(1, 43388, '\P{^Is_GCB=l}', "");
Expect(0, 43389, '\p{Is_GCB=l}', "");
Expect(1, 43389, '\p{^Is_GCB=l}', "");
Expect(1, 43389, '\P{Is_GCB=l}', "");
Expect(0, 43389, '\P{^Is_GCB=l}', "");
Expect(1, 43388, '\p{Is_GCB=	-l}', "");
Expect(0, 43388, '\p{^Is_GCB=	-l}', "");
Expect(0, 43388, '\P{Is_GCB=	-l}', "");
Expect(1, 43388, '\P{^Is_GCB=	-l}', "");
Expect(0, 43389, '\p{Is_GCB=	-l}', "");
Expect(1, 43389, '\p{^Is_GCB=	-l}', "");
Expect(1, 43389, '\P{Is_GCB=	-l}', "");
Expect(0, 43389, '\P{^Is_GCB=	-l}', "");
Error('\p{Grapheme_Cluster_Break=:=LF}');
Error('\P{Grapheme_Cluster_Break=:=LF}');
Expect(1, 10, '\p{Grapheme_Cluster_Break=lf}', "");
Expect(0, 10, '\p{^Grapheme_Cluster_Break=lf}', "");
Expect(0, 10, '\P{Grapheme_Cluster_Break=lf}', "");
Expect(1, 10, '\P{^Grapheme_Cluster_Break=lf}', "");
Expect(0, 11, '\p{Grapheme_Cluster_Break=lf}', "");
Expect(1, 11, '\p{^Grapheme_Cluster_Break=lf}', "");
Expect(1, 11, '\P{Grapheme_Cluster_Break=lf}', "");
Expect(0, 11, '\P{^Grapheme_Cluster_Break=lf}', "");
Expect(1, 10, '\p{Grapheme_Cluster_Break= -LF}', "");
Expect(0, 10, '\p{^Grapheme_Cluster_Break= -LF}', "");
Expect(0, 10, '\P{Grapheme_Cluster_Break= -LF}', "");
Expect(1, 10, '\P{^Grapheme_Cluster_Break= -LF}', "");
Expect(0, 11, '\p{Grapheme_Cluster_Break= -LF}', "");
Expect(1, 11, '\p{^Grapheme_Cluster_Break= -LF}', "");
Expect(1, 11, '\P{Grapheme_Cluster_Break= -LF}', "");
Expect(0, 11, '\P{^Grapheme_Cluster_Break= -LF}', "");
Error('\p{GCB=- LF/a/}');
Error('\P{GCB=- LF/a/}');
Expect(1, 10, '\p{GCB=lf}', "");
Expect(0, 10, '\p{^GCB=lf}', "");
Expect(0, 10, '\P{GCB=lf}', "");
Expect(1, 10, '\P{^GCB=lf}', "");
Expect(0, 11, '\p{GCB=lf}', "");
Expect(1, 11, '\p{^GCB=lf}', "");
Expect(1, 11, '\P{GCB=lf}', "");
Expect(0, 11, '\P{^GCB=lf}', "");
Expect(1, 10, '\p{GCB=_LF}', "");
Expect(0, 10, '\p{^GCB=_LF}', "");
Expect(0, 10, '\P{GCB=_LF}', "");
Expect(1, 10, '\P{^GCB=_LF}', "");
Expect(0, 11, '\p{GCB=_LF}', "");
Expect(1, 11, '\p{^GCB=_LF}', "");
Expect(1, 11, '\P{GCB=_LF}', "");
Expect(0, 11, '\P{^GCB=_LF}', "");
Error('\p{Is_Grapheme_Cluster_Break=_	lf/a/}');
Error('\P{Is_Grapheme_Cluster_Break=_	lf/a/}');
Expect(1, 10, '\p{Is_Grapheme_Cluster_Break=lf}', "");
Expect(0, 10, '\p{^Is_Grapheme_Cluster_Break=lf}', "");
Expect(0, 10, '\P{Is_Grapheme_Cluster_Break=lf}', "");
Expect(1, 10, '\P{^Is_Grapheme_Cluster_Break=lf}', "");
Expect(0, 11, '\p{Is_Grapheme_Cluster_Break=lf}', "");
Expect(1, 11, '\p{^Is_Grapheme_Cluster_Break=lf}', "");
Expect(1, 11, '\P{Is_Grapheme_Cluster_Break=lf}', "");
Expect(0, 11, '\P{^Is_Grapheme_Cluster_Break=lf}', "");
Expect(1, 10, '\p{Is_Grapheme_Cluster_Break=-lf}', "");
Expect(0, 10, '\p{^Is_Grapheme_Cluster_Break=-lf}', "");
Expect(0, 10, '\P{Is_Grapheme_Cluster_Break=-lf}', "");
Expect(1, 10, '\P{^Is_Grapheme_Cluster_Break=-lf}', "");
Expect(0, 11, '\p{Is_Grapheme_Cluster_Break=-lf}', "");
Expect(1, 11, '\p{^Is_Grapheme_Cluster_Break=-lf}', "");
Expect(1, 11, '\P{Is_Grapheme_Cluster_Break=-lf}', "");
Expect(0, 11, '\P{^Is_Grapheme_Cluster_Break=-lf}', "");
Error('\p{Is_GCB=-:=LF}');
Error('\P{Is_GCB=-:=LF}');
Expect(1, 10, '\p{Is_GCB=lf}', "");
Expect(0, 10, '\p{^Is_GCB=lf}', "");
Expect(0, 10, '\P{Is_GCB=lf}', "");
Expect(1, 10, '\P{^Is_GCB=lf}', "");
Expect(0, 11, '\p{Is_GCB=lf}', "");
Expect(1, 11, '\p{^Is_GCB=lf}', "");
Expect(1, 11, '\P{Is_GCB=lf}', "");
Expect(0, 11, '\P{^Is_GCB=lf}', "");
Expect(1, 10, '\p{Is_GCB=	LF}', "");
Expect(0, 10, '\p{^Is_GCB=	LF}', "");
Expect(0, 10, '\P{Is_GCB=	LF}', "");
Expect(1, 10, '\P{^Is_GCB=	LF}', "");
Expect(0, 11, '\p{Is_GCB=	LF}', "");
Expect(1, 11, '\p{^Is_GCB=	LF}', "");
Expect(1, 11, '\P{Is_GCB=	LF}', "");
Expect(0, 11, '\P{^Is_GCB=	LF}', "");
Error('\p{Grapheme_Cluster_Break=	:=LV}');
Error('\P{Grapheme_Cluster_Break=	:=LV}');
Expect(1, 55176, '\p{Grapheme_Cluster_Break=lv}', "");
Expect(0, 55176, '\p{^Grapheme_Cluster_Break=lv}', "");
Expect(0, 55176, '\P{Grapheme_Cluster_Break=lv}', "");
Expect(1, 55176, '\P{^Grapheme_Cluster_Break=lv}', "");
Expect(0, 55177, '\p{Grapheme_Cluster_Break=lv}', "");
Expect(1, 55177, '\p{^Grapheme_Cluster_Break=lv}', "");
Expect(1, 55177, '\P{Grapheme_Cluster_Break=lv}', "");
Expect(0, 55177, '\P{^Grapheme_Cluster_Break=lv}', "");
Expect(1, 55176, '\p{Grapheme_Cluster_Break=_lv}', "");
Expect(0, 55176, '\p{^Grapheme_Cluster_Break=_lv}', "");
Expect(0, 55176, '\P{Grapheme_Cluster_Break=_lv}', "");
Expect(1, 55176, '\P{^Grapheme_Cluster_Break=_lv}', "");
Expect(0, 55177, '\p{Grapheme_Cluster_Break=_lv}', "");
Expect(1, 55177, '\p{^Grapheme_Cluster_Break=_lv}', "");
Expect(1, 55177, '\P{Grapheme_Cluster_Break=_lv}', "");
Expect(0, 55177, '\P{^Grapheme_Cluster_Break=_lv}', "");
Error('\p{GCB=/a/ -lv}');
Error('\P{GCB=/a/ -lv}');
Expect(1, 55176, '\p{GCB=lv}', "");
Expect(0, 55176, '\p{^GCB=lv}', "");
Expect(0, 55176, '\P{GCB=lv}', "");
Expect(1, 55176, '\P{^GCB=lv}', "");
Expect(0, 55177, '\p{GCB=lv}', "");
Expect(1, 55177, '\p{^GCB=lv}', "");
Expect(1, 55177, '\P{GCB=lv}', "");
Expect(0, 55177, '\P{^GCB=lv}', "");
Expect(1, 55176, '\p{GCB= 	LV}', "");
Expect(0, 55176, '\p{^GCB= 	LV}', "");
Expect(0, 55176, '\P{GCB= 	LV}', "");
Expect(1, 55176, '\P{^GCB= 	LV}', "");
Expect(0, 55177, '\p{GCB= 	LV}', "");
Expect(1, 55177, '\p{^GCB= 	LV}', "");
Expect(1, 55177, '\P{GCB= 	LV}', "");
Expect(0, 55177, '\P{^GCB= 	LV}', "");
Error('\p{Is_Grapheme_Cluster_Break: _/a/LV}');
Error('\P{Is_Grapheme_Cluster_Break: _/a/LV}');
Expect(1, 55176, '\p{Is_Grapheme_Cluster_Break=lv}', "");
Expect(0, 55176, '\p{^Is_Grapheme_Cluster_Break=lv}', "");
Expect(0, 55176, '\P{Is_Grapheme_Cluster_Break=lv}', "");
Expect(1, 55176, '\P{^Is_Grapheme_Cluster_Break=lv}', "");
Expect(0, 55177, '\p{Is_Grapheme_Cluster_Break=lv}', "");
Expect(1, 55177, '\p{^Is_Grapheme_Cluster_Break=lv}', "");
Expect(1, 55177, '\P{Is_Grapheme_Cluster_Break=lv}', "");
Expect(0, 55177, '\P{^Is_Grapheme_Cluster_Break=lv}', "");
Expect(1, 55176, '\p{Is_Grapheme_Cluster_Break=	LV}', "");
Expect(0, 55176, '\p{^Is_Grapheme_Cluster_Break=	LV}', "");
Expect(0, 55176, '\P{Is_Grapheme_Cluster_Break=	LV}', "");
Expect(1, 55176, '\P{^Is_Grapheme_Cluster_Break=	LV}', "");
Expect(0, 55177, '\p{Is_Grapheme_Cluster_Break=	LV}', "");
Expect(1, 55177, '\p{^Is_Grapheme_Cluster_Break=	LV}', "");
Expect(1, 55177, '\P{Is_Grapheme_Cluster_Break=	LV}', "");
Expect(0, 55177, '\P{^Is_Grapheme_Cluster_Break=	LV}', "");
Error('\p{Is_GCB:	_ lv/a/}');
Error('\P{Is_GCB:	_ lv/a/}');
Expect(1, 55176, '\p{Is_GCB=lv}', "");
Expect(0, 55176, '\p{^Is_GCB=lv}', "");
Expect(0, 55176, '\P{Is_GCB=lv}', "");
Expect(1, 55176, '\P{^Is_GCB=lv}', "");
Expect(0, 55177, '\p{Is_GCB=lv}', "");
Expect(1, 55177, '\p{^Is_GCB=lv}', "");
Expect(1, 55177, '\P{Is_GCB=lv}', "");
Expect(0, 55177, '\P{^Is_GCB=lv}', "");
Expect(1, 55176, '\p{Is_GCB:	 -LV}', "");
Expect(0, 55176, '\p{^Is_GCB:	 -LV}', "");
Expect(0, 55176, '\P{Is_GCB:	 -LV}', "");
Expect(1, 55176, '\P{^Is_GCB:	 -LV}', "");
Expect(0, 55177, '\p{Is_GCB:	 -LV}', "");
Expect(1, 55177, '\p{^Is_GCB:	 -LV}', "");
Expect(1, 55177, '\P{Is_GCB:	 -LV}', "");
Expect(0, 55177, '\P{^Is_GCB:	 -LV}', "");
Error('\p{Grapheme_Cluster_Break=	/a/LVT}');
Error('\P{Grapheme_Cluster_Break=	/a/LVT}');
Expect(1, 55203, '\p{Grapheme_Cluster_Break=lvt}', "");
Expect(0, 55203, '\p{^Grapheme_Cluster_Break=lvt}', "");
Expect(0, 55203, '\P{Grapheme_Cluster_Break=lvt}', "");
Expect(1, 55203, '\P{^Grapheme_Cluster_Break=lvt}', "");
Expect(0, 55204, '\p{Grapheme_Cluster_Break=lvt}', "");
Expect(1, 55204, '\p{^Grapheme_Cluster_Break=lvt}', "");
Expect(1, 55204, '\P{Grapheme_Cluster_Break=lvt}', "");
Expect(0, 55204, '\P{^Grapheme_Cluster_Break=lvt}', "");
Expect(1, 55203, '\p{Grapheme_Cluster_Break: 	LVT}', "");
Expect(0, 55203, '\p{^Grapheme_Cluster_Break: 	LVT}', "");
Expect(0, 55203, '\P{Grapheme_Cluster_Break: 	LVT}', "");
Expect(1, 55203, '\P{^Grapheme_Cluster_Break: 	LVT}', "");
Expect(0, 55204, '\p{Grapheme_Cluster_Break: 	LVT}', "");
Expect(1, 55204, '\p{^Grapheme_Cluster_Break: 	LVT}', "");
Expect(1, 55204, '\P{Grapheme_Cluster_Break: 	LVT}', "");
Expect(0, 55204, '\P{^Grapheme_Cluster_Break: 	LVT}', "");
Error('\p{GCB=:= LVT}');
Error('\P{GCB=:= LVT}');
Expect(1, 55203, '\p{GCB=lvt}', "");
Expect(0, 55203, '\p{^GCB=lvt}', "");
Expect(0, 55203, '\P{GCB=lvt}', "");
Expect(1, 55203, '\P{^GCB=lvt}', "");
Expect(0, 55204, '\p{GCB=lvt}', "");
Expect(1, 55204, '\p{^GCB=lvt}', "");
Expect(1, 55204, '\P{GCB=lvt}', "");
Expect(0, 55204, '\P{^GCB=lvt}', "");
Expect(1, 55203, '\p{GCB:		_LVT}', "");
Expect(0, 55203, '\p{^GCB:		_LVT}', "");
Expect(0, 55203, '\P{GCB:		_LVT}', "");
Expect(1, 55203, '\P{^GCB:		_LVT}', "");
Expect(0, 55204, '\p{GCB:		_LVT}', "");
Expect(1, 55204, '\p{^GCB:		_LVT}', "");
Expect(1, 55204, '\P{GCB:		_LVT}', "");
Expect(0, 55204, '\P{^GCB:		_LVT}', "");
Error('\p{Is_Grapheme_Cluster_Break=:=- LVT}');
Error('\P{Is_Grapheme_Cluster_Break=:=- LVT}');
Expect(1, 55203, '\p{Is_Grapheme_Cluster_Break:   lvt}', "");
Expect(0, 55203, '\p{^Is_Grapheme_Cluster_Break:   lvt}', "");
Expect(0, 55203, '\P{Is_Grapheme_Cluster_Break:   lvt}', "");
Expect(1, 55203, '\P{^Is_Grapheme_Cluster_Break:   lvt}', "");
Expect(0, 55204, '\p{Is_Grapheme_Cluster_Break:   lvt}', "");
Expect(1, 55204, '\p{^Is_Grapheme_Cluster_Break:   lvt}', "");
Expect(1, 55204, '\P{Is_Grapheme_Cluster_Break:   lvt}', "");
Expect(0, 55204, '\P{^Is_Grapheme_Cluster_Break:   lvt}', "");
Expect(1, 55203, '\p{Is_Grapheme_Cluster_Break=-LVT}', "");
Expect(0, 55203, '\p{^Is_Grapheme_Cluster_Break=-LVT}', "");
Expect(0, 55203, '\P{Is_Grapheme_Cluster_Break=-LVT}', "");
Expect(1, 55203, '\P{^Is_Grapheme_Cluster_Break=-LVT}', "");
Expect(0, 55204, '\p{Is_Grapheme_Cluster_Break=-LVT}', "");
Expect(1, 55204, '\p{^Is_Grapheme_Cluster_Break=-LVT}', "");
Expect(1, 55204, '\P{Is_Grapheme_Cluster_Break=-LVT}', "");
Expect(0, 55204, '\P{^Is_Grapheme_Cluster_Break=-LVT}', "");
Error('\p{Is_GCB=/a/	LVT}');
Error('\P{Is_GCB=/a/	LVT}');
Expect(1, 55203, '\p{Is_GCB: lvt}', "");
Expect(0, 55203, '\p{^Is_GCB: lvt}', "");
Expect(0, 55203, '\P{Is_GCB: lvt}', "");
Expect(1, 55203, '\P{^Is_GCB: lvt}', "");
Expect(0, 55204, '\p{Is_GCB: lvt}', "");
Expect(1, 55204, '\p{^Is_GCB: lvt}', "");
Expect(1, 55204, '\P{Is_GCB: lvt}', "");
Expect(0, 55204, '\P{^Is_GCB: lvt}', "");
Expect(1, 55203, '\p{Is_GCB=_ LVT}', "");
Expect(0, 55203, '\p{^Is_GCB=_ LVT}', "");
Expect(0, 55203, '\P{Is_GCB=_ LVT}', "");
Expect(1, 55203, '\P{^Is_GCB=_ LVT}', "");
Expect(0, 55204, '\p{Is_GCB=_ LVT}', "");
Expect(1, 55204, '\p{^Is_GCB=_ LVT}', "");
Expect(1, 55204, '\P{Is_GCB=_ LVT}', "");
Expect(0, 55204, '\P{^Is_GCB=_ LVT}', "");
Error('\p{Grapheme_Cluster_Break=	/a/Prepend}');
Error('\P{Grapheme_Cluster_Break=	/a/Prepend}');
Expect(0, 1, '\p{Grapheme_Cluster_Break=prepend}', "");
Expect(1, 1, '\p{^Grapheme_Cluster_Break=prepend}', "");
Expect(1, 1, '\P{Grapheme_Cluster_Break=prepend}', "");
Expect(0, 1, '\P{^Grapheme_Cluster_Break=prepend}', "");
Expect(0, 1, '\p{Grapheme_Cluster_Break= PREPEND}', "");
Expect(1, 1, '\p{^Grapheme_Cluster_Break= PREPEND}', "");
Expect(1, 1, '\P{Grapheme_Cluster_Break= PREPEND}', "");
Expect(0, 1, '\P{^Grapheme_Cluster_Break= PREPEND}', "");
Error('\p{GCB=:=	-PP}');
Error('\P{GCB=:=	-PP}');
Expect(0, 1, '\p{GCB=pp}', "");
Expect(1, 1, '\p{^GCB=pp}', "");
Expect(1, 1, '\P{GCB=pp}', "");
Expect(0, 1, '\P{^GCB=pp}', "");
Expect(0, 1, '\p{GCB=_PP}', "");
Expect(1, 1, '\p{^GCB=_PP}', "");
Expect(1, 1, '\P{GCB=_PP}', "");
Expect(0, 1, '\P{^GCB=_PP}', "");
Error('\p{Is_Grapheme_Cluster_Break=/a/_ prepend}');
Error('\P{Is_Grapheme_Cluster_Break=/a/_ prepend}');
Expect(0, 1, '\p{Is_Grapheme_Cluster_Break=prepend}', "");
Expect(1, 1, '\p{^Is_Grapheme_Cluster_Break=prepend}', "");
Expect(1, 1, '\P{Is_Grapheme_Cluster_Break=prepend}', "");
Expect(0, 1, '\P{^Is_Grapheme_Cluster_Break=prepend}', "");
Expect(0, 1, '\p{Is_Grapheme_Cluster_Break=_Prepend}', "");
Expect(1, 1, '\p{^Is_Grapheme_Cluster_Break=_Prepend}', "");
Expect(1, 1, '\P{Is_Grapheme_Cluster_Break=_Prepend}', "");
Expect(0, 1, '\P{^Is_Grapheme_Cluster_Break=_Prepend}', "");
Error('\p{Is_GCB=:= _PP}');
Error('\P{Is_GCB=:= _PP}');
Expect(0, 1, '\p{Is_GCB: pp}', "");
Expect(1, 1, '\p{^Is_GCB: pp}', "");
Expect(1, 1, '\P{Is_GCB: pp}', "");
Expect(0, 1, '\P{^Is_GCB: pp}', "");
Expect(0, 1, '\p{Is_GCB=-PP}', "");
Expect(1, 1, '\p{^Is_GCB=-PP}', "");
Expect(1, 1, '\P{Is_GCB=-PP}', "");
Expect(0, 1, '\P{^Is_GCB=-PP}', "");
Error('\p{Grapheme_Cluster_Break=_/a/regional_INDICATOR}');
Error('\P{Grapheme_Cluster_Break=_/a/regional_INDICATOR}');
Expect(1, 127487, '\p{Grapheme_Cluster_Break=regionalindicator}', "");
Expect(0, 127487, '\p{^Grapheme_Cluster_Break=regionalindicator}', "");
Expect(0, 127487, '\P{Grapheme_Cluster_Break=regionalindicator}', "");
Expect(1, 127487, '\P{^Grapheme_Cluster_Break=regionalindicator}', "");
Expect(0, 127488, '\p{Grapheme_Cluster_Break=regionalindicator}', "");
Expect(1, 127488, '\p{^Grapheme_Cluster_Break=regionalindicator}', "");
Expect(1, 127488, '\P{Grapheme_Cluster_Break=regionalindicator}', "");
Expect(0, 127488, '\P{^Grapheme_Cluster_Break=regionalindicator}', "");
Expect(1, 127487, '\p{Grapheme_Cluster_Break=--Regional_Indicator}', "");
Expect(0, 127487, '\p{^Grapheme_Cluster_Break=--Regional_Indicator}', "");
Expect(0, 127487, '\P{Grapheme_Cluster_Break=--Regional_Indicator}', "");
Expect(1, 127487, '\P{^Grapheme_Cluster_Break=--Regional_Indicator}', "");
Expect(0, 127488, '\p{Grapheme_Cluster_Break=--Regional_Indicator}', "");
Expect(1, 127488, '\p{^Grapheme_Cluster_Break=--Regional_Indicator}', "");
Expect(1, 127488, '\P{Grapheme_Cluster_Break=--Regional_Indicator}', "");
Expect(0, 127488, '\P{^Grapheme_Cluster_Break=--Regional_Indicator}', "");
Error('\p{GCB= 	RI/a/}');
Error('\P{GCB= 	RI/a/}');
Expect(1, 127487, '\p{GCB:	ri}', "");
Expect(0, 127487, '\p{^GCB:	ri}', "");
Expect(0, 127487, '\P{GCB:	ri}', "");
Expect(1, 127487, '\P{^GCB:	ri}', "");
Expect(0, 127488, '\p{GCB:	ri}', "");
Expect(1, 127488, '\p{^GCB:	ri}', "");
Expect(1, 127488, '\P{GCB:	ri}', "");
Expect(0, 127488, '\P{^GCB:	ri}', "");
Expect(1, 127487, '\p{GCB=	RI}', "");
Expect(0, 127487, '\p{^GCB=	RI}', "");
Expect(0, 127487, '\P{GCB=	RI}', "");
Expect(1, 127487, '\P{^GCB=	RI}', "");
Expect(0, 127488, '\p{GCB=	RI}', "");
Expect(1, 127488, '\p{^GCB=	RI}', "");
Expect(1, 127488, '\P{GCB=	RI}', "");
Expect(0, 127488, '\P{^GCB=	RI}', "");
Error('\p{Is_Grapheme_Cluster_Break=__REGIONAL_indicator/a/}');
Error('\P{Is_Grapheme_Cluster_Break=__REGIONAL_indicator/a/}');
Expect(1, 127487, '\p{Is_Grapheme_Cluster_Break=regionalindicator}', "");
Expect(0, 127487, '\p{^Is_Grapheme_Cluster_Break=regionalindicator}', "");
Expect(0, 127487, '\P{Is_Grapheme_Cluster_Break=regionalindicator}', "");
Expect(1, 127487, '\P{^Is_Grapheme_Cluster_Break=regionalindicator}', "");
Expect(0, 127488, '\p{Is_Grapheme_Cluster_Break=regionalindicator}', "");
Expect(1, 127488, '\p{^Is_Grapheme_Cluster_Break=regionalindicator}', "");
Expect(1, 127488, '\P{Is_Grapheme_Cluster_Break=regionalindicator}', "");
Expect(0, 127488, '\P{^Is_Grapheme_Cluster_Break=regionalindicator}', "");
Expect(1, 127487, '\p{Is_Grapheme_Cluster_Break=	 REGIONAL_INDICATOR}', "");
Expect(0, 127487, '\p{^Is_Grapheme_Cluster_Break=	 REGIONAL_INDICATOR}', "");
Expect(0, 127487, '\P{Is_Grapheme_Cluster_Break=	 REGIONAL_INDICATOR}', "");
Expect(1, 127487, '\P{^Is_Grapheme_Cluster_Break=	 REGIONAL_INDICATOR}', "");
Expect(0, 127488, '\p{Is_Grapheme_Cluster_Break=	 REGIONAL_INDICATOR}', "");
Expect(1, 127488, '\p{^Is_Grapheme_Cluster_Break=	 REGIONAL_INDICATOR}', "");
Expect(1, 127488, '\P{Is_Grapheme_Cluster_Break=	 REGIONAL_INDICATOR}', "");
Expect(0, 127488, '\P{^Is_Grapheme_Cluster_Break=	 REGIONAL_INDICATOR}', "");
Error('\p{Is_GCB=/a/ RI}');
Error('\P{Is_GCB=/a/ RI}');
Expect(1, 127487, '\p{Is_GCB: ri}', "");
Expect(0, 127487, '\p{^Is_GCB: ri}', "");
Expect(0, 127487, '\P{Is_GCB: ri}', "");
Expect(1, 127487, '\P{^Is_GCB: ri}', "");
Expect(0, 127488, '\p{Is_GCB: ri}', "");
Expect(1, 127488, '\p{^Is_GCB: ri}', "");
Expect(1, 127488, '\P{Is_GCB: ri}', "");
Expect(0, 127488, '\P{^Is_GCB: ri}', "");
Expect(1, 127487, '\p{Is_GCB=	_RI}', "");
Expect(0, 127487, '\p{^Is_GCB=	_RI}', "");
Expect(0, 127487, '\P{Is_GCB=	_RI}', "");
Expect(1, 127487, '\P{^Is_GCB=	_RI}', "");
Expect(0, 127488, '\p{Is_GCB=	_RI}', "");
Expect(1, 127488, '\p{^Is_GCB=	_RI}', "");
Expect(1, 127488, '\P{Is_GCB=	_RI}', "");
Expect(0, 127488, '\P{^Is_GCB=	_RI}', "");
Error('\p{Grapheme_Cluster_Break: - spacingmark/a/}');
Error('\P{Grapheme_Cluster_Break: - spacingmark/a/}');
Expect(1, 119149, '\p{Grapheme_Cluster_Break=spacingmark}', "");
Expect(0, 119149, '\p{^Grapheme_Cluster_Break=spacingmark}', "");
Expect(0, 119149, '\P{Grapheme_Cluster_Break=spacingmark}', "");
Expect(1, 119149, '\P{^Grapheme_Cluster_Break=spacingmark}', "");
Expect(0, 119150, '\p{Grapheme_Cluster_Break=spacingmark}', "");
Expect(1, 119150, '\p{^Grapheme_Cluster_Break=spacingmark}', "");
Expect(1, 119150, '\P{Grapheme_Cluster_Break=spacingmark}', "");
Expect(0, 119150, '\P{^Grapheme_Cluster_Break=spacingmark}', "");
Expect(1, 119149, '\p{Grapheme_Cluster_Break=-SpacingMark}', "");
Expect(0, 119149, '\p{^Grapheme_Cluster_Break=-SpacingMark}', "");
Expect(0, 119149, '\P{Grapheme_Cluster_Break=-SpacingMark}', "");
Expect(1, 119149, '\P{^Grapheme_Cluster_Break=-SpacingMark}', "");
Expect(0, 119150, '\p{Grapheme_Cluster_Break=-SpacingMark}', "");
Expect(1, 119150, '\p{^Grapheme_Cluster_Break=-SpacingMark}', "");
Expect(1, 119150, '\P{Grapheme_Cluster_Break=-SpacingMark}', "");
Expect(0, 119150, '\P{^Grapheme_Cluster_Break=-SpacingMark}', "");
Error('\p{GCB=:= _SM}');
Error('\P{GCB=:= _SM}');
Expect(1, 119149, '\p{GCB=sm}', "");
Expect(0, 119149, '\p{^GCB=sm}', "");
Expect(0, 119149, '\P{GCB=sm}', "");
Expect(1, 119149, '\P{^GCB=sm}', "");
Expect(0, 119150, '\p{GCB=sm}', "");
Expect(1, 119150, '\p{^GCB=sm}', "");
Expect(1, 119150, '\P{GCB=sm}', "");
Expect(0, 119150, '\P{^GCB=sm}', "");
Expect(1, 119149, '\p{GCB= _SM}', "");
Expect(0, 119149, '\p{^GCB= _SM}', "");
Expect(0, 119149, '\P{GCB= _SM}', "");
Expect(1, 119149, '\P{^GCB= _SM}', "");
Expect(0, 119150, '\p{GCB= _SM}', "");
Expect(1, 119150, '\p{^GCB= _SM}', "");
Expect(1, 119150, '\P{GCB= _SM}', "");
Expect(0, 119150, '\P{^GCB= _SM}', "");
Error('\p{Is_Grapheme_Cluster_Break=/a/  SpacingMark}');
Error('\P{Is_Grapheme_Cluster_Break=/a/  SpacingMark}');
Expect(1, 119149, '\p{Is_Grapheme_Cluster_Break=spacingmark}', "");
Expect(0, 119149, '\p{^Is_Grapheme_Cluster_Break=spacingmark}', "");
Expect(0, 119149, '\P{Is_Grapheme_Cluster_Break=spacingmark}', "");
Expect(1, 119149, '\P{^Is_Grapheme_Cluster_Break=spacingmark}', "");
Expect(0, 119150, '\p{Is_Grapheme_Cluster_Break=spacingmark}', "");
Expect(1, 119150, '\p{^Is_Grapheme_Cluster_Break=spacingmark}', "");
Expect(1, 119150, '\P{Is_Grapheme_Cluster_Break=spacingmark}', "");
Expect(0, 119150, '\P{^Is_Grapheme_Cluster_Break=spacingmark}', "");
Expect(1, 119149, '\p{Is_Grapheme_Cluster_Break=_spacingmark}', "");
Expect(0, 119149, '\p{^Is_Grapheme_Cluster_Break=_spacingmark}', "");
Expect(0, 119149, '\P{Is_Grapheme_Cluster_Break=_spacingmark}', "");
Expect(1, 119149, '\P{^Is_Grapheme_Cluster_Break=_spacingmark}', "");
Expect(0, 119150, '\p{Is_Grapheme_Cluster_Break=_spacingmark}', "");
Expect(1, 119150, '\p{^Is_Grapheme_Cluster_Break=_spacingmark}', "");
Expect(1, 119150, '\P{Is_Grapheme_Cluster_Break=_spacingmark}', "");
Expect(0, 119150, '\P{^Is_Grapheme_Cluster_Break=_spacingmark}', "");
Error('\p{Is_GCB=/a/_ sm}');
Error('\P{Is_GCB=/a/_ sm}');
Expect(1, 119149, '\p{Is_GCB=sm}', "");
Expect(0, 119149, '\p{^Is_GCB=sm}', "");
Expect(0, 119149, '\P{Is_GCB=sm}', "");
Expect(1, 119149, '\P{^Is_GCB=sm}', "");
Expect(0, 119150, '\p{Is_GCB=sm}', "");
Expect(1, 119150, '\p{^Is_GCB=sm}', "");
Expect(1, 119150, '\P{Is_GCB=sm}', "");
Expect(0, 119150, '\P{^Is_GCB=sm}', "");
Expect(1, 119149, '\p{Is_GCB=_-sm}', "");
Expect(0, 119149, '\p{^Is_GCB=_-sm}', "");
Expect(0, 119149, '\P{Is_GCB=_-sm}', "");
Expect(1, 119149, '\P{^Is_GCB=_-sm}', "");
Expect(0, 119150, '\p{Is_GCB=_-sm}', "");
Expect(1, 119150, '\p{^Is_GCB=_-sm}', "");
Expect(1, 119150, '\P{Is_GCB=_-sm}', "");
Expect(0, 119150, '\P{^Is_GCB=_-sm}', "");
Error('\p{Grapheme_Cluster_Break= /a/t}');
Error('\P{Grapheme_Cluster_Break= /a/t}');
Expect(1, 55291, '\p{Grapheme_Cluster_Break=t}', "");
Expect(0, 55291, '\p{^Grapheme_Cluster_Break=t}', "");
Expect(0, 55291, '\P{Grapheme_Cluster_Break=t}', "");
Expect(1, 55291, '\P{^Grapheme_Cluster_Break=t}', "");
Expect(0, 55292, '\p{Grapheme_Cluster_Break=t}', "");
Expect(1, 55292, '\p{^Grapheme_Cluster_Break=t}', "");
Expect(1, 55292, '\P{Grapheme_Cluster_Break=t}', "");
Expect(0, 55292, '\P{^Grapheme_Cluster_Break=t}', "");
Expect(1, 55291, '\p{Grapheme_Cluster_Break=  T}', "");
Expect(0, 55291, '\p{^Grapheme_Cluster_Break=  T}', "");
Expect(0, 55291, '\P{Grapheme_Cluster_Break=  T}', "");
Expect(1, 55291, '\P{^Grapheme_Cluster_Break=  T}', "");
Expect(0, 55292, '\p{Grapheme_Cluster_Break=  T}', "");
Expect(1, 55292, '\p{^Grapheme_Cluster_Break=  T}', "");
Expect(1, 55292, '\P{Grapheme_Cluster_Break=  T}', "");
Expect(0, 55292, '\P{^Grapheme_Cluster_Break=  T}', "");
Error('\p{GCB=/a/ _T}');
Error('\P{GCB=/a/ _T}');
Expect(1, 55291, '\p{GCB=t}', "");
Expect(0, 55291, '\p{^GCB=t}', "");
Expect(0, 55291, '\P{GCB=t}', "");
Expect(1, 55291, '\P{^GCB=t}', "");
Expect(0, 55292, '\p{GCB=t}', "");
Expect(1, 55292, '\p{^GCB=t}', "");
Expect(1, 55292, '\P{GCB=t}', "");
Expect(0, 55292, '\P{^GCB=t}', "");
Expect(1, 55291, '\p{GCB= 	T}', "");
Expect(0, 55291, '\p{^GCB= 	T}', "");
Expect(0, 55291, '\P{GCB= 	T}', "");
Expect(1, 55291, '\P{^GCB= 	T}', "");
Expect(0, 55292, '\p{GCB= 	T}', "");
Expect(1, 55292, '\p{^GCB= 	T}', "");
Expect(1, 55292, '\P{GCB= 	T}', "");
Expect(0, 55292, '\P{^GCB= 	T}', "");
Error('\p{Is_Grapheme_Cluster_Break=	/a/t}');
Error('\P{Is_Grapheme_Cluster_Break=	/a/t}');
Expect(1, 55291, '\p{Is_Grapheme_Cluster_Break=t}', "");
Expect(0, 55291, '\p{^Is_Grapheme_Cluster_Break=t}', "");
Expect(0, 55291, '\P{Is_Grapheme_Cluster_Break=t}', "");
Expect(1, 55291, '\P{^Is_Grapheme_Cluster_Break=t}', "");
Expect(0, 55292, '\p{Is_Grapheme_Cluster_Break=t}', "");
Expect(1, 55292, '\p{^Is_Grapheme_Cluster_Break=t}', "");
Expect(1, 55292, '\P{Is_Grapheme_Cluster_Break=t}', "");
Expect(0, 55292, '\P{^Is_Grapheme_Cluster_Break=t}', "");
Expect(1, 55291, '\p{Is_Grapheme_Cluster_Break=--t}', "");
Expect(0, 55291, '\p{^Is_Grapheme_Cluster_Break=--t}', "");
Expect(0, 55291, '\P{Is_Grapheme_Cluster_Break=--t}', "");
Expect(1, 55291, '\P{^Is_Grapheme_Cluster_Break=--t}', "");
Expect(0, 55292, '\p{Is_Grapheme_Cluster_Break=--t}', "");
Expect(1, 55292, '\p{^Is_Grapheme_Cluster_Break=--t}', "");
Expect(1, 55292, '\P{Is_Grapheme_Cluster_Break=--t}', "");
Expect(0, 55292, '\P{^Is_Grapheme_Cluster_Break=--t}', "");
Error('\p{Is_GCB=/a/	T}');
Error('\P{Is_GCB=/a/	T}');
Expect(1, 55291, '\p{Is_GCB=t}', "");
Expect(0, 55291, '\p{^Is_GCB=t}', "");
Expect(0, 55291, '\P{Is_GCB=t}', "");
Expect(1, 55291, '\P{^Is_GCB=t}', "");
Expect(0, 55292, '\p{Is_GCB=t}', "");
Expect(1, 55292, '\p{^Is_GCB=t}', "");
Expect(1, 55292, '\P{Is_GCB=t}', "");
Expect(0, 55292, '\P{^Is_GCB=t}', "");
Expect(1, 55291, '\p{Is_GCB=  T}', "");
Expect(0, 55291, '\p{^Is_GCB=  T}', "");
Expect(0, 55291, '\P{Is_GCB=  T}', "");
Expect(1, 55291, '\P{^Is_GCB=  T}', "");
Expect(0, 55292, '\p{Is_GCB=  T}', "");
Expect(1, 55292, '\p{^Is_GCB=  T}', "");
Expect(1, 55292, '\P{Is_GCB=  T}', "");
Expect(0, 55292, '\P{^Is_GCB=  T}', "");
Error('\p{Grapheme_Cluster_Break=:=-V}');
Error('\P{Grapheme_Cluster_Break=:=-V}');
Expect(1, 55238, '\p{Grapheme_Cluster_Break=v}', "");
Expect(0, 55238, '\p{^Grapheme_Cluster_Break=v}', "");
Expect(0, 55238, '\P{Grapheme_Cluster_Break=v}', "");
Expect(1, 55238, '\P{^Grapheme_Cluster_Break=v}', "");
Expect(0, 55239, '\p{Grapheme_Cluster_Break=v}', "");
Expect(1, 55239, '\p{^Grapheme_Cluster_Break=v}', "");
Expect(1, 55239, '\P{Grapheme_Cluster_Break=v}', "");
Expect(0, 55239, '\P{^Grapheme_Cluster_Break=v}', "");
Expect(1, 55238, '\p{Grapheme_Cluster_Break=	-V}', "");
Expect(0, 55238, '\p{^Grapheme_Cluster_Break=	-V}', "");
Expect(0, 55238, '\P{Grapheme_Cluster_Break=	-V}', "");
Expect(1, 55238, '\P{^Grapheme_Cluster_Break=	-V}', "");
Expect(0, 55239, '\p{Grapheme_Cluster_Break=	-V}', "");
Expect(1, 55239, '\p{^Grapheme_Cluster_Break=	-V}', "");
Expect(1, 55239, '\P{Grapheme_Cluster_Break=	-V}', "");
Expect(0, 55239, '\P{^Grapheme_Cluster_Break=	-V}', "");
Error('\p{GCB=	 V:=}');
Error('\P{GCB=	 V:=}');
Expect(1, 55238, '\p{GCB=v}', "");
Expect(0, 55238, '\p{^GCB=v}', "");
Expect(0, 55238, '\P{GCB=v}', "");
Expect(1, 55238, '\P{^GCB=v}', "");
Expect(0, 55239, '\p{GCB=v}', "");
Expect(1, 55239, '\p{^GCB=v}', "");
Expect(1, 55239, '\P{GCB=v}', "");
Expect(0, 55239, '\P{^GCB=v}', "");
Expect(1, 55238, '\p{GCB=_V}', "");
Expect(0, 55238, '\p{^GCB=_V}', "");
Expect(0, 55238, '\P{GCB=_V}', "");
Expect(1, 55238, '\P{^GCB=_V}', "");
Expect(0, 55239, '\p{GCB=_V}', "");
Expect(1, 55239, '\p{^GCB=_V}', "");
Expect(1, 55239, '\P{GCB=_V}', "");
Expect(0, 55239, '\P{^GCB=_V}', "");
Error('\p{Is_Grapheme_Cluster_Break=-/a/v}');
Error('\P{Is_Grapheme_Cluster_Break=-/a/v}');
Expect(1, 55238, '\p{Is_Grapheme_Cluster_Break=v}', "");
Expect(0, 55238, '\p{^Is_Grapheme_Cluster_Break=v}', "");
Expect(0, 55238, '\P{Is_Grapheme_Cluster_Break=v}', "");
Expect(1, 55238, '\P{^Is_Grapheme_Cluster_Break=v}', "");
Expect(0, 55239, '\p{Is_Grapheme_Cluster_Break=v}', "");
Expect(1, 55239, '\p{^Is_Grapheme_Cluster_Break=v}', "");
Expect(1, 55239, '\P{Is_Grapheme_Cluster_Break=v}', "");
Expect(0, 55239, '\P{^Is_Grapheme_Cluster_Break=v}', "");
Expect(1, 55238, '\p{Is_Grapheme_Cluster_Break:  	V}', "");
Expect(0, 55238, '\p{^Is_Grapheme_Cluster_Break:  	V}', "");
Expect(0, 55238, '\P{Is_Grapheme_Cluster_Break:  	V}', "");
Expect(1, 55238, '\P{^Is_Grapheme_Cluster_Break:  	V}', "");
Expect(0, 55239, '\p{Is_Grapheme_Cluster_Break:  	V}', "");
Expect(1, 55239, '\p{^Is_Grapheme_Cluster_Break:  	V}', "");
Expect(1, 55239, '\P{Is_Grapheme_Cluster_Break:  	V}', "");
Expect(0, 55239, '\P{^Is_Grapheme_Cluster_Break:  	V}', "");
Error('\p{Is_GCB=	V:=}');
Error('\P{Is_GCB=	V:=}');
Expect(1, 55238, '\p{Is_GCB=v}', "");
Expect(0, 55238, '\p{^Is_GCB=v}', "");
Expect(0, 55238, '\P{Is_GCB=v}', "");
Expect(1, 55238, '\P{^Is_GCB=v}', "");
Expect(0, 55239, '\p{Is_GCB=v}', "");
Expect(1, 55239, '\p{^Is_GCB=v}', "");
Expect(1, 55239, '\P{Is_GCB=v}', "");
Expect(0, 55239, '\P{^Is_GCB=v}', "");
Expect(1, 55238, '\p{Is_GCB= -V}', "");
Expect(0, 55238, '\p{^Is_GCB= -V}', "");
Expect(0, 55238, '\P{Is_GCB= -V}', "");
Expect(1, 55238, '\P{^Is_GCB= -V}', "");
Expect(0, 55239, '\p{Is_GCB= -V}', "");
Expect(1, 55239, '\p{^Is_GCB= -V}', "");
Expect(1, 55239, '\P{Is_GCB= -V}', "");
Expect(0, 55239, '\P{^Is_GCB= -V}', "");
Error('\p{Grapheme_Cluster_Break=	_OTHER:=}');
Error('\P{Grapheme_Cluster_Break=	_OTHER:=}');
Expect(1, 921600, '\p{Grapheme_Cluster_Break=other}', "");
Expect(0, 921600, '\p{^Grapheme_Cluster_Break=other}', "");
Expect(0, 921600, '\P{Grapheme_Cluster_Break=other}', "");
Expect(1, 921600, '\P{^Grapheme_Cluster_Break=other}', "");
Expect(0, 921599, '\p{Grapheme_Cluster_Break=other}', "");
Expect(1, 921599, '\p{^Grapheme_Cluster_Break=other}', "");
Expect(1, 921599, '\P{Grapheme_Cluster_Break=other}', "");
Expect(0, 921599, '\P{^Grapheme_Cluster_Break=other}', "");
Expect(1, 921600, '\p{Grapheme_Cluster_Break: Other}', "");
Expect(0, 921600, '\p{^Grapheme_Cluster_Break: Other}', "");
Expect(0, 921600, '\P{Grapheme_Cluster_Break: Other}', "");
Expect(1, 921600, '\P{^Grapheme_Cluster_Break: Other}', "");
Expect(0, 921599, '\p{Grapheme_Cluster_Break: Other}', "");
Expect(1, 921599, '\p{^Grapheme_Cluster_Break: Other}', "");
Expect(1, 921599, '\P{Grapheme_Cluster_Break: Other}', "");
Expect(0, 921599, '\P{^Grapheme_Cluster_Break: Other}', "");
Error('\p{GCB=_XX/a/}');
Error('\P{GCB=_XX/a/}');
Expect(1, 921600, '\p{GCB=xx}', "");
Expect(0, 921600, '\p{^GCB=xx}', "");
Expect(0, 921600, '\P{GCB=xx}', "");
Expect(1, 921600, '\P{^GCB=xx}', "");
Expect(0, 921599, '\p{GCB=xx}', "");
Expect(1, 921599, '\p{^GCB=xx}', "");
Expect(1, 921599, '\P{GCB=xx}', "");
Expect(0, 921599, '\P{^GCB=xx}', "");
Expect(1, 921600, '\p{GCB= 	XX}', "");
Expect(0, 921600, '\p{^GCB= 	XX}', "");
Expect(0, 921600, '\P{GCB= 	XX}', "");
Expect(1, 921600, '\P{^GCB= 	XX}', "");
Expect(0, 921599, '\p{GCB= 	XX}', "");
Expect(1, 921599, '\p{^GCB= 	XX}', "");
Expect(1, 921599, '\P{GCB= 	XX}', "");
Expect(0, 921599, '\P{^GCB= 	XX}', "");
Error('\p{Is_Grapheme_Cluster_Break=_/a/Other}');
Error('\P{Is_Grapheme_Cluster_Break=_/a/Other}');
Expect(1, 921600, '\p{Is_Grapheme_Cluster_Break=other}', "");
Expect(0, 921600, '\p{^Is_Grapheme_Cluster_Break=other}', "");
Expect(0, 921600, '\P{Is_Grapheme_Cluster_Break=other}', "");
Expect(1, 921600, '\P{^Is_Grapheme_Cluster_Break=other}', "");
Expect(0, 921599, '\p{Is_Grapheme_Cluster_Break=other}', "");
Expect(1, 921599, '\p{^Is_Grapheme_Cluster_Break=other}', "");
Expect(1, 921599, '\P{Is_Grapheme_Cluster_Break=other}', "");
Expect(0, 921599, '\P{^Is_Grapheme_Cluster_Break=other}', "");
Expect(1, 921600, '\p{Is_Grapheme_Cluster_Break=__Other}', "");
Expect(0, 921600, '\p{^Is_Grapheme_Cluster_Break=__Other}', "");
Expect(0, 921600, '\P{Is_Grapheme_Cluster_Break=__Other}', "");
Expect(1, 921600, '\P{^Is_Grapheme_Cluster_Break=__Other}', "");
Expect(0, 921599, '\p{Is_Grapheme_Cluster_Break=__Other}', "");
Expect(1, 921599, '\p{^Is_Grapheme_Cluster_Break=__Other}', "");
Expect(1, 921599, '\P{Is_Grapheme_Cluster_Break=__Other}', "");
Expect(0, 921599, '\P{^Is_Grapheme_Cluster_Break=__Other}', "");
Error('\p{Is_GCB=/a/-	XX}');
Error('\P{Is_GCB=/a/-	XX}');
Expect(1, 921600, '\p{Is_GCB=xx}', "");
Expect(0, 921600, '\p{^Is_GCB=xx}', "");
Expect(0, 921600, '\P{Is_GCB=xx}', "");
Expect(1, 921600, '\P{^Is_GCB=xx}', "");
Expect(0, 921599, '\p{Is_GCB=xx}', "");
Expect(1, 921599, '\p{^Is_GCB=xx}', "");
Expect(1, 921599, '\P{Is_GCB=xx}', "");
Expect(0, 921599, '\P{^Is_GCB=xx}', "");
Expect(1, 921600, '\p{Is_GCB= 	XX}', "");
Expect(0, 921600, '\p{^Is_GCB= 	XX}', "");
Expect(0, 921600, '\P{Is_GCB= 	XX}', "");
Expect(1, 921600, '\P{^Is_GCB= 	XX}', "");
Expect(0, 921599, '\p{Is_GCB= 	XX}', "");
Expect(1, 921599, '\p{^Is_GCB= 	XX}', "");
Expect(1, 921599, '\P{Is_GCB= 	XX}', "");
Expect(0, 921599, '\P{^Is_GCB= 	XX}', "");
Error('\p{Grapheme_Base= 	NO/a/}');
Error('\P{Grapheme_Base= 	NO/a/}');
Expect(1, 195102, '\p{Grapheme_Base=no}', "");
Expect(0, 195102, '\p{^Grapheme_Base=no}', "");
Expect(0, 195102, '\P{Grapheme_Base=no}', "");
Expect(1, 195102, '\P{^Grapheme_Base=no}', "");
Expect(0, 195101, '\p{Grapheme_Base=no}', "");
Expect(1, 195101, '\p{^Grapheme_Base=no}', "");
Expect(1, 195101, '\P{Grapheme_Base=no}', "");
Expect(0, 195101, '\P{^Grapheme_Base=no}', "");
Expect(1, 195102, '\p{Grapheme_Base:_	NO}', "");
Expect(0, 195102, '\p{^Grapheme_Base:_	NO}', "");
Expect(0, 195102, '\P{Grapheme_Base:_	NO}', "");
Expect(1, 195102, '\P{^Grapheme_Base:_	NO}', "");
Expect(0, 195101, '\p{Grapheme_Base:_	NO}', "");
Expect(1, 195101, '\p{^Grapheme_Base:_	NO}', "");
Expect(1, 195101, '\P{Grapheme_Base:_	NO}', "");
Expect(0, 195101, '\P{^Grapheme_Base:_	NO}', "");
Error('\p{Gr_Base=- N/a/}');
Error('\P{Gr_Base=- N/a/}');
Expect(1, 195102, '\p{Gr_Base=n}', "");
Expect(0, 195102, '\p{^Gr_Base=n}', "");
Expect(0, 195102, '\P{Gr_Base=n}', "");
Expect(1, 195102, '\P{^Gr_Base=n}', "");
Expect(0, 195101, '\p{Gr_Base=n}', "");
Expect(1, 195101, '\p{^Gr_Base=n}', "");
Expect(1, 195101, '\P{Gr_Base=n}', "");
Expect(0, 195101, '\P{^Gr_Base=n}', "");
Expect(1, 195102, '\p{Gr_Base=	 N}', "");
Expect(0, 195102, '\p{^Gr_Base=	 N}', "");
Expect(0, 195102, '\P{Gr_Base=	 N}', "");
Expect(1, 195102, '\P{^Gr_Base=	 N}', "");
Expect(0, 195101, '\p{Gr_Base=	 N}', "");
Expect(1, 195101, '\p{^Gr_Base=	 N}', "");
Expect(1, 195101, '\P{Gr_Base=	 N}', "");
Expect(0, 195101, '\P{^Gr_Base=	 N}', "");
Error('\p{Is_Grapheme_Base=_F/a/}');
Error('\P{Is_Grapheme_Base=_F/a/}');
Expect(1, 195102, '\p{Is_Grapheme_Base: f}', "");
Expect(0, 195102, '\p{^Is_Grapheme_Base: f}', "");
Expect(0, 195102, '\P{Is_Grapheme_Base: f}', "");
Expect(1, 195102, '\P{^Is_Grapheme_Base: f}', "");
Expect(0, 195101, '\p{Is_Grapheme_Base: f}', "");
Expect(1, 195101, '\p{^Is_Grapheme_Base: f}', "");
Expect(1, 195101, '\P{Is_Grapheme_Base: f}', "");
Expect(0, 195101, '\P{^Is_Grapheme_Base: f}', "");
Expect(1, 195102, '\p{Is_Grapheme_Base=-F}', "");
Expect(0, 195102, '\p{^Is_Grapheme_Base=-F}', "");
Expect(0, 195102, '\P{Is_Grapheme_Base=-F}', "");
Expect(1, 195102, '\P{^Is_Grapheme_Base=-F}', "");
Expect(0, 195101, '\p{Is_Grapheme_Base=-F}', "");
Expect(1, 195101, '\p{^Is_Grapheme_Base=-F}', "");
Expect(1, 195101, '\P{Is_Grapheme_Base=-F}', "");
Expect(0, 195101, '\P{^Is_Grapheme_Base=-F}', "");
Error('\p{Is_Gr_Base=	False:=}');
Error('\P{Is_Gr_Base=	False:=}');
Expect(1, 195102, '\p{Is_Gr_Base: false}', "");
Expect(0, 195102, '\p{^Is_Gr_Base: false}', "");
Expect(0, 195102, '\P{Is_Gr_Base: false}', "");
Expect(1, 195102, '\P{^Is_Gr_Base: false}', "");
Expect(0, 195101, '\p{Is_Gr_Base: false}', "");
Expect(1, 195101, '\p{^Is_Gr_Base: false}', "");
Expect(1, 195101, '\P{Is_Gr_Base: false}', "");
Expect(0, 195101, '\P{^Is_Gr_Base: false}', "");
Expect(1, 195102, '\p{Is_Gr_Base:   _ False}', "");
Expect(0, 195102, '\p{^Is_Gr_Base:   _ False}', "");
Expect(0, 195102, '\P{Is_Gr_Base:   _ False}', "");
Expect(1, 195102, '\P{^Is_Gr_Base:   _ False}', "");
Expect(0, 195101, '\p{Is_Gr_Base:   _ False}', "");
Expect(1, 195101, '\p{^Is_Gr_Base:   _ False}', "");
Expect(1, 195101, '\P{Is_Gr_Base:   _ False}', "");
Expect(0, 195101, '\P{^Is_Gr_Base:   _ False}', "");
Error('\p{Grapheme_Base=/a/YES}');
Error('\P{Grapheme_Base=/a/YES}');
Expect(1, 195101, '\p{Grapheme_Base=yes}', "");
Expect(0, 195101, '\p{^Grapheme_Base=yes}', "");
Expect(0, 195101, '\P{Grapheme_Base=yes}', "");
Expect(1, 195101, '\P{^Grapheme_Base=yes}', "");
Expect(0, 195102, '\p{Grapheme_Base=yes}', "");
Expect(1, 195102, '\p{^Grapheme_Base=yes}', "");
Expect(1, 195102, '\P{Grapheme_Base=yes}', "");
Expect(0, 195102, '\P{^Grapheme_Base=yes}', "");
Expect(1, 195101, '\p{Grapheme_Base=		Yes}', "");
Expect(0, 195101, '\p{^Grapheme_Base=		Yes}', "");
Expect(0, 195101, '\P{Grapheme_Base=		Yes}', "");
Expect(1, 195101, '\P{^Grapheme_Base=		Yes}', "");
Expect(0, 195102, '\p{Grapheme_Base=		Yes}', "");
Expect(1, 195102, '\p{^Grapheme_Base=		Yes}', "");
Expect(1, 195102, '\P{Grapheme_Base=		Yes}', "");
Expect(0, 195102, '\P{^Grapheme_Base=		Yes}', "");
Error('\p{Gr_Base=:=_	Y}');
Error('\P{Gr_Base=:=_	Y}');
Expect(1, 195101, '\p{Gr_Base=y}', "");
Expect(0, 195101, '\p{^Gr_Base=y}', "");
Expect(0, 195101, '\P{Gr_Base=y}', "");
Expect(1, 195101, '\P{^Gr_Base=y}', "");
Expect(0, 195102, '\p{Gr_Base=y}', "");
Expect(1, 195102, '\p{^Gr_Base=y}', "");
Expect(1, 195102, '\P{Gr_Base=y}', "");
Expect(0, 195102, '\P{^Gr_Base=y}', "");
Expect(1, 195101, '\p{Gr_Base=	 Y}', "");
Expect(0, 195101, '\p{^Gr_Base=	 Y}', "");
Expect(0, 195101, '\P{Gr_Base=	 Y}', "");
Expect(1, 195101, '\P{^Gr_Base=	 Y}', "");
Expect(0, 195102, '\p{Gr_Base=	 Y}', "");
Expect(1, 195102, '\p{^Gr_Base=	 Y}', "");
Expect(1, 195102, '\P{Gr_Base=	 Y}', "");
Expect(0, 195102, '\P{^Gr_Base=	 Y}', "");
Error('\p{Is_Grapheme_Base:    /a/T}');
Error('\P{Is_Grapheme_Base:    /a/T}');
Expect(1, 195101, '\p{Is_Grapheme_Base=t}', "");
Expect(0, 195101, '\p{^Is_Grapheme_Base=t}', "");
Expect(0, 195101, '\P{Is_Grapheme_Base=t}', "");
Expect(1, 195101, '\P{^Is_Grapheme_Base=t}', "");
Expect(0, 195102, '\p{Is_Grapheme_Base=t}', "");
Expect(1, 195102, '\p{^Is_Grapheme_Base=t}', "");
Expect(1, 195102, '\P{Is_Grapheme_Base=t}', "");
Expect(0, 195102, '\P{^Is_Grapheme_Base=t}', "");
Expect(1, 195101, '\p{Is_Grapheme_Base=		t}', "");
Expect(0, 195101, '\p{^Is_Grapheme_Base=		t}', "");
Expect(0, 195101, '\P{Is_Grapheme_Base=		t}', "");
Expect(1, 195101, '\P{^Is_Grapheme_Base=		t}', "");
Expect(0, 195102, '\p{Is_Grapheme_Base=		t}', "");
Expect(1, 195102, '\p{^Is_Grapheme_Base=		t}', "");
Expect(1, 195102, '\P{Is_Grapheme_Base=		t}', "");
Expect(0, 195102, '\P{^Is_Grapheme_Base=		t}', "");
Error('\p{Is_Gr_Base= TRUE/a/}');
Error('\P{Is_Gr_Base= TRUE/a/}');
Expect(1, 195101, '\p{Is_Gr_Base=true}', "");
Expect(0, 195101, '\p{^Is_Gr_Base=true}', "");
Expect(0, 195101, '\P{Is_Gr_Base=true}', "");
Expect(1, 195101, '\P{^Is_Gr_Base=true}', "");
Expect(0, 195102, '\p{Is_Gr_Base=true}', "");
Expect(1, 195102, '\p{^Is_Gr_Base=true}', "");
Expect(1, 195102, '\P{Is_Gr_Base=true}', "");
Expect(0, 195102, '\P{^Is_Gr_Base=true}', "");
Expect(1, 195101, '\p{Is_Gr_Base= -True}', "");
Expect(0, 195101, '\p{^Is_Gr_Base= -True}', "");
Expect(0, 195101, '\P{Is_Gr_Base= -True}', "");
Expect(1, 195101, '\P{^Is_Gr_Base= -True}', "");
Expect(0, 195102, '\p{Is_Gr_Base= -True}', "");
Expect(1, 195102, '\p{^Is_Gr_Base= -True}', "");
Expect(1, 195102, '\P{Is_Gr_Base= -True}', "");
Expect(0, 195102, '\P{^Is_Gr_Base= -True}', "");
Error('\p{Grapheme_Extend=	/a/No}');
Error('\P{Grapheme_Extend=	/a/No}');
Expect(1, 918000, '\p{Grapheme_Extend=no}', "");
Expect(0, 918000, '\p{^Grapheme_Extend=no}', "");
Expect(0, 918000, '\P{Grapheme_Extend=no}', "");
Expect(1, 918000, '\P{^Grapheme_Extend=no}', "");
Expect(0, 917999, '\p{Grapheme_Extend=no}', "");
Expect(1, 917999, '\p{^Grapheme_Extend=no}', "");
Expect(1, 917999, '\P{Grapheme_Extend=no}', "");
Expect(0, 917999, '\P{^Grapheme_Extend=no}', "");
Expect(1, 918000, '\p{Grapheme_Extend=	-No}', "");
Expect(0, 918000, '\p{^Grapheme_Extend=	-No}', "");
Expect(0, 918000, '\P{Grapheme_Extend=	-No}', "");
Expect(1, 918000, '\P{^Grapheme_Extend=	-No}', "");
Expect(0, 917999, '\p{Grapheme_Extend=	-No}', "");
Expect(1, 917999, '\p{^Grapheme_Extend=	-No}', "");
Expect(1, 917999, '\P{Grapheme_Extend=	-No}', "");
Expect(0, 917999, '\P{^Grapheme_Extend=	-No}', "");
Error('\p{Gr_Ext=	-N:=}');
Error('\P{Gr_Ext=	-N:=}');
Expect(1, 918000, '\p{Gr_Ext=n}', "");
Expect(0, 918000, '\p{^Gr_Ext=n}', "");
Expect(0, 918000, '\P{Gr_Ext=n}', "");
Expect(1, 918000, '\P{^Gr_Ext=n}', "");
Expect(0, 917999, '\p{Gr_Ext=n}', "");
Expect(1, 917999, '\p{^Gr_Ext=n}', "");
Expect(1, 917999, '\P{Gr_Ext=n}', "");
Expect(0, 917999, '\P{^Gr_Ext=n}', "");
Expect(1, 918000, '\p{Gr_Ext=- N}', "");
Expect(0, 918000, '\p{^Gr_Ext=- N}', "");
Expect(0, 918000, '\P{Gr_Ext=- N}', "");
Expect(1, 918000, '\P{^Gr_Ext=- N}', "");
Expect(0, 917999, '\p{Gr_Ext=- N}', "");
Expect(1, 917999, '\p{^Gr_Ext=- N}', "");
Expect(1, 917999, '\P{Gr_Ext=- N}', "");
Expect(0, 917999, '\P{^Gr_Ext=- N}', "");
Error('\p{Is_Grapheme_Extend=_f/a/}');
Error('\P{Is_Grapheme_Extend=_f/a/}');
Expect(1, 918000, '\p{Is_Grapheme_Extend:f}', "");
Expect(0, 918000, '\p{^Is_Grapheme_Extend:f}', "");
Expect(0, 918000, '\P{Is_Grapheme_Extend:f}', "");
Expect(1, 918000, '\P{^Is_Grapheme_Extend:f}', "");
Expect(0, 917999, '\p{Is_Grapheme_Extend:f}', "");
Expect(1, 917999, '\p{^Is_Grapheme_Extend:f}', "");
Expect(1, 917999, '\P{Is_Grapheme_Extend:f}', "");
Expect(0, 917999, '\P{^Is_Grapheme_Extend:f}', "");
Expect(1, 918000, '\p{Is_Grapheme_Extend=-	f}', "");
Expect(0, 918000, '\p{^Is_Grapheme_Extend=-	f}', "");
Expect(0, 918000, '\P{Is_Grapheme_Extend=-	f}', "");
Expect(1, 918000, '\P{^Is_Grapheme_Extend=-	f}', "");
Expect(0, 917999, '\p{Is_Grapheme_Extend=-	f}', "");
Expect(1, 917999, '\p{^Is_Grapheme_Extend=-	f}', "");
Expect(1, 917999, '\P{Is_Grapheme_Extend=-	f}', "");
Expect(0, 917999, '\P{^Is_Grapheme_Extend=-	f}', "");
Error('\p{Is_Gr_Ext: :=False}');
Error('\P{Is_Gr_Ext: :=False}');
Expect(1, 918000, '\p{Is_Gr_Ext=false}', "");
Expect(0, 918000, '\p{^Is_Gr_Ext=false}', "");
Expect(0, 918000, '\P{Is_Gr_Ext=false}', "");
Expect(1, 918000, '\P{^Is_Gr_Ext=false}', "");
Expect(0, 917999, '\p{Is_Gr_Ext=false}', "");
Expect(1, 917999, '\p{^Is_Gr_Ext=false}', "");
Expect(1, 917999, '\P{Is_Gr_Ext=false}', "");
Expect(0, 917999, '\P{^Is_Gr_Ext=false}', "");
Expect(1, 918000, '\p{Is_Gr_Ext=-false}', "");
Expect(0, 918000, '\p{^Is_Gr_Ext=-false}', "");
Expect(0, 918000, '\P{Is_Gr_Ext=-false}', "");
Expect(1, 918000, '\P{^Is_Gr_Ext=-false}', "");
Expect(0, 917999, '\p{Is_Gr_Ext=-false}', "");
Expect(1, 917999, '\p{^Is_Gr_Ext=-false}', "");
Expect(1, 917999, '\P{Is_Gr_Ext=-false}', "");
Expect(0, 917999, '\P{^Is_Gr_Ext=-false}', "");
Error('\p{Grapheme_Extend:   _:=Yes}');
Error('\P{Grapheme_Extend:   _:=Yes}');
Expect(1, 917999, '\p{Grapheme_Extend=yes}', "");
Expect(0, 917999, '\p{^Grapheme_Extend=yes}', "");
Expect(0, 917999, '\P{Grapheme_Extend=yes}', "");
Expect(1, 917999, '\P{^Grapheme_Extend=yes}', "");
Expect(0, 918000, '\p{Grapheme_Extend=yes}', "");
Expect(1, 918000, '\p{^Grapheme_Extend=yes}', "");
Expect(1, 918000, '\P{Grapheme_Extend=yes}', "");
Expect(0, 918000, '\P{^Grapheme_Extend=yes}', "");
Expect(1, 917999, '\p{Grapheme_Extend=__Yes}', "");
Expect(0, 917999, '\p{^Grapheme_Extend=__Yes}', "");
Expect(0, 917999, '\P{Grapheme_Extend=__Yes}', "");
Expect(1, 917999, '\P{^Grapheme_Extend=__Yes}', "");
Expect(0, 918000, '\p{Grapheme_Extend=__Yes}', "");
Expect(1, 918000, '\p{^Grapheme_Extend=__Yes}', "");
Expect(1, 918000, '\P{Grapheme_Extend=__Yes}', "");
Expect(0, 918000, '\P{^Grapheme_Extend=__Yes}', "");
Error('\p{Gr_Ext=:=_Y}');
Error('\P{Gr_Ext=:=_Y}');
Expect(1, 917999, '\p{Gr_Ext=y}', "");
Expect(0, 917999, '\p{^Gr_Ext=y}', "");
Expect(0, 917999, '\P{Gr_Ext=y}', "");
Expect(1, 917999, '\P{^Gr_Ext=y}', "");
Expect(0, 918000, '\p{Gr_Ext=y}', "");
Expect(1, 918000, '\p{^Gr_Ext=y}', "");
Expect(1, 918000, '\P{Gr_Ext=y}', "");
Expect(0, 918000, '\P{^Gr_Ext=y}', "");
Expect(1, 917999, '\p{Gr_Ext=_Y}', "");
Expect(0, 917999, '\p{^Gr_Ext=_Y}', "");
Expect(0, 917999, '\P{Gr_Ext=_Y}', "");
Expect(1, 917999, '\P{^Gr_Ext=_Y}', "");
Expect(0, 918000, '\p{Gr_Ext=_Y}', "");
Expect(1, 918000, '\p{^Gr_Ext=_Y}', "");
Expect(1, 918000, '\P{Gr_Ext=_Y}', "");
Expect(0, 918000, '\P{^Gr_Ext=_Y}', "");
Error('\p{Is_Grapheme_Extend=/a/	T}');
Error('\P{Is_Grapheme_Extend=/a/	T}');
Expect(1, 917999, '\p{Is_Grapheme_Extend=t}', "");
Expect(0, 917999, '\p{^Is_Grapheme_Extend=t}', "");
Expect(0, 917999, '\P{Is_Grapheme_Extend=t}', "");
Expect(1, 917999, '\P{^Is_Grapheme_Extend=t}', "");
Expect(0, 918000, '\p{Is_Grapheme_Extend=t}', "");
Expect(1, 918000, '\p{^Is_Grapheme_Extend=t}', "");
Expect(1, 918000, '\P{Is_Grapheme_Extend=t}', "");
Expect(0, 918000, '\P{^Is_Grapheme_Extend=t}', "");
Expect(1, 917999, '\p{Is_Grapheme_Extend=_T}', "");
Expect(0, 917999, '\p{^Is_Grapheme_Extend=_T}', "");
Expect(0, 917999, '\P{Is_Grapheme_Extend=_T}', "");
Expect(1, 917999, '\P{^Is_Grapheme_Extend=_T}', "");
Expect(0, 918000, '\p{Is_Grapheme_Extend=_T}', "");
Expect(1, 918000, '\p{^Is_Grapheme_Extend=_T}', "");
Expect(1, 918000, '\P{Is_Grapheme_Extend=_T}', "");
Expect(0, 918000, '\P{^Is_Grapheme_Extend=_T}', "");
Error('\p{Is_Gr_Ext=/a/_-TRUE}');
Error('\P{Is_Gr_Ext=/a/_-TRUE}');
Expect(1, 917999, '\p{Is_Gr_Ext=true}', "");
Expect(0, 917999, '\p{^Is_Gr_Ext=true}', "");
Expect(0, 917999, '\P{Is_Gr_Ext=true}', "");
Expect(1, 917999, '\P{^Is_Gr_Ext=true}', "");
Expect(0, 918000, '\p{Is_Gr_Ext=true}', "");
Expect(1, 918000, '\p{^Is_Gr_Ext=true}', "");
Expect(1, 918000, '\P{Is_Gr_Ext=true}', "");
Expect(0, 918000, '\P{^Is_Gr_Ext=true}', "");
Expect(1, 917999, '\p{Is_Gr_Ext=--true}', "");
Expect(0, 917999, '\p{^Is_Gr_Ext=--true}', "");
Expect(0, 917999, '\P{Is_Gr_Ext=--true}', "");
Expect(1, 917999, '\P{^Is_Gr_Ext=--true}', "");
Expect(0, 918000, '\p{Is_Gr_Ext=--true}', "");
Expect(1, 918000, '\p{^Is_Gr_Ext=--true}', "");
Expect(1, 918000, '\P{Is_Gr_Ext=--true}', "");
Expect(0, 918000, '\P{^Is_Gr_Ext=--true}', "");
Error('\p{Grapheme_Link=No}');
Error('\P{Grapheme_Link=No}');
Error('\p{Gr_Link=N}');
Error('\P{Gr_Link=N}');
Error('\p{Is_Grapheme_Link=F}');
Error('\P{Is_Grapheme_Link=F}');
Error('\p{Is_Gr_Link=False}');
Error('\P{Is_Gr_Link=False}');
Error('\p{Grapheme_Link=Yes}');
Error('\P{Grapheme_Link=Yes}');
Error('\p{Gr_Link=Y}');
Error('\P{Gr_Link=Y}');
Error('\p{Is_Grapheme_Link=T}');
Error('\P{Is_Grapheme_Link=T}');
Error('\p{Is_Gr_Link=True}');
Error('\P{Is_Gr_Link=True}');
Error('\p{Hex_Digit= /a/NO}');
Error('\P{Hex_Digit= /a/NO}');
Expect(1, 65351, '\p{Hex_Digit:no}', "");
Expect(0, 65351, '\p{^Hex_Digit:no}', "");
Expect(0, 65351, '\P{Hex_Digit:no}', "");
Expect(1, 65351, '\P{^Hex_Digit:no}', "");
Expect(0, 65350, '\p{Hex_Digit:no}', "");
Expect(1, 65350, '\p{^Hex_Digit:no}', "");
Expect(1, 65350, '\P{Hex_Digit:no}', "");
Expect(0, 65350, '\P{^Hex_Digit:no}', "");
Expect(1, 65351, '\p{Hex_Digit=__No}', "");
Expect(0, 65351, '\p{^Hex_Digit=__No}', "");
Expect(0, 65351, '\P{Hex_Digit=__No}', "");
Expect(1, 65351, '\P{^Hex_Digit=__No}', "");
Expect(0, 65350, '\p{Hex_Digit=__No}', "");
Expect(1, 65350, '\p{^Hex_Digit=__No}', "");
Expect(1, 65350, '\P{Hex_Digit=__No}', "");
Expect(0, 65350, '\P{^Hex_Digit=__No}', "");
Error('\p{Hex= /a/n}');
Error('\P{Hex= /a/n}');
Expect(1, 65351, '\p{Hex=n}', "");
Expect(0, 65351, '\p{^Hex=n}', "");
Expect(0, 65351, '\P{Hex=n}', "");
Expect(1, 65351, '\P{^Hex=n}', "");
Expect(0, 65350, '\p{Hex=n}', "");
Expect(1, 65350, '\p{^Hex=n}', "");
Expect(1, 65350, '\P{Hex=n}', "");
Expect(0, 65350, '\P{^Hex=n}', "");
Expect(1, 65351, '\p{Hex=		N}', "");
Expect(0, 65351, '\p{^Hex=		N}', "");
Expect(0, 65351, '\P{Hex=		N}', "");
Expect(1, 65351, '\P{^Hex=		N}', "");
Expect(0, 65350, '\p{Hex=		N}', "");
Expect(1, 65350, '\p{^Hex=		N}', "");
Expect(1, 65350, '\P{Hex=		N}', "");
Expect(0, 65350, '\P{^Hex=		N}', "");
Error('\p{Is_Hex_Digit=_:=F}');
Error('\P{Is_Hex_Digit=_:=F}');
Expect(1, 65351, '\p{Is_Hex_Digit=f}', "");
Expect(0, 65351, '\p{^Is_Hex_Digit=f}', "");
Expect(0, 65351, '\P{Is_Hex_Digit=f}', "");
Expect(1, 65351, '\P{^Is_Hex_Digit=f}', "");
Expect(0, 65350, '\p{Is_Hex_Digit=f}', "");
Expect(1, 65350, '\p{^Is_Hex_Digit=f}', "");
Expect(1, 65350, '\P{Is_Hex_Digit=f}', "");
Expect(0, 65350, '\P{^Is_Hex_Digit=f}', "");
Expect(1, 65351, '\p{Is_Hex_Digit=-F}', "");
Expect(0, 65351, '\p{^Is_Hex_Digit=-F}', "");
Expect(0, 65351, '\P{Is_Hex_Digit=-F}', "");
Expect(1, 65351, '\P{^Is_Hex_Digit=-F}', "");
Expect(0, 65350, '\p{Is_Hex_Digit=-F}', "");
Expect(1, 65350, '\p{^Is_Hex_Digit=-F}', "");
Expect(1, 65350, '\P{Is_Hex_Digit=-F}', "");
Expect(0, 65350, '\P{^Is_Hex_Digit=-F}', "");
Error('\p{Is_Hex=-:=False}');
Error('\P{Is_Hex=-:=False}');
Expect(1, 65351, '\p{Is_Hex=false}', "");
Expect(0, 65351, '\p{^Is_Hex=false}', "");
Expect(0, 65351, '\P{Is_Hex=false}', "");
Expect(1, 65351, '\P{^Is_Hex=false}', "");
Expect(0, 65350, '\p{Is_Hex=false}', "");
Expect(1, 65350, '\p{^Is_Hex=false}', "");
Expect(1, 65350, '\P{Is_Hex=false}', "");
Expect(0, 65350, '\P{^Is_Hex=false}', "");
Expect(1, 65351, '\p{Is_Hex=  false}', "");
Expect(0, 65351, '\p{^Is_Hex=  false}', "");
Expect(0, 65351, '\P{Is_Hex=  false}', "");
Expect(1, 65351, '\P{^Is_Hex=  false}', "");
Expect(0, 65350, '\p{Is_Hex=  false}', "");
Expect(1, 65350, '\p{^Is_Hex=  false}', "");
Expect(1, 65350, '\P{Is_Hex=  false}', "");
Expect(0, 65350, '\P{^Is_Hex=  false}', "");
Error('\p{Hex_Digit=:=-	Yes}');
Error('\P{Hex_Digit=:=-	Yes}');
Expect(1, 65350, '\p{Hex_Digit=yes}', "");
Expect(0, 65350, '\p{^Hex_Digit=yes}', "");
Expect(0, 65350, '\P{Hex_Digit=yes}', "");
Expect(1, 65350, '\P{^Hex_Digit=yes}', "");
Expect(0, 65351, '\p{Hex_Digit=yes}', "");
Expect(1, 65351, '\p{^Hex_Digit=yes}', "");
Expect(1, 65351, '\P{Hex_Digit=yes}', "");
Expect(0, 65351, '\P{^Hex_Digit=yes}', "");
Expect(1, 65350, '\p{Hex_Digit=		yes}', "");
Expect(0, 65350, '\p{^Hex_Digit=		yes}', "");
Expect(0, 65350, '\P{Hex_Digit=		yes}', "");
Expect(1, 65350, '\P{^Hex_Digit=		yes}', "");
Expect(0, 65351, '\p{Hex_Digit=		yes}', "");
Expect(1, 65351, '\p{^Hex_Digit=		yes}', "");
Expect(1, 65351, '\P{Hex_Digit=		yes}', "");
Expect(0, 65351, '\P{^Hex_Digit=		yes}', "");
Error('\p{Hex=:=	Y}');
Error('\P{Hex=:=	Y}');
Expect(1, 65350, '\p{Hex=y}', "");
Expect(0, 65350, '\p{^Hex=y}', "");
Expect(0, 65350, '\P{Hex=y}', "");
Expect(1, 65350, '\P{^Hex=y}', "");
Expect(0, 65351, '\p{Hex=y}', "");
Expect(1, 65351, '\p{^Hex=y}', "");
Expect(1, 65351, '\P{Hex=y}', "");
Expect(0, 65351, '\P{^Hex=y}', "");
Expect(1, 65350, '\p{Hex=_	y}', "");
Expect(0, 65350, '\p{^Hex=_	y}', "");
Expect(0, 65350, '\P{Hex=_	y}', "");
Expect(1, 65350, '\P{^Hex=_	y}', "");
Expect(0, 65351, '\p{Hex=_	y}', "");
Expect(1, 65351, '\p{^Hex=_	y}', "");
Expect(1, 65351, '\P{Hex=_	y}', "");
Expect(0, 65351, '\P{^Hex=_	y}', "");
Error('\p{Is_Hex_Digit=_:=t}');
Error('\P{Is_Hex_Digit=_:=t}');
Expect(1, 65350, '\p{Is_Hex_Digit=t}', "");
Expect(0, 65350, '\p{^Is_Hex_Digit=t}', "");
Expect(0, 65350, '\P{Is_Hex_Digit=t}', "");
Expect(1, 65350, '\P{^Is_Hex_Digit=t}', "");
Expect(0, 65351, '\p{Is_Hex_Digit=t}', "");
Expect(1, 65351, '\p{^Is_Hex_Digit=t}', "");
Expect(1, 65351, '\P{Is_Hex_Digit=t}', "");
Expect(0, 65351, '\P{^Is_Hex_Digit=t}', "");
Expect(1, 65350, '\p{Is_Hex_Digit= _T}', "");
Expect(0, 65350, '\p{^Is_Hex_Digit= _T}', "");
Expect(0, 65350, '\P{Is_Hex_Digit= _T}', "");
Expect(1, 65350, '\P{^Is_Hex_Digit= _T}', "");
Expect(0, 65351, '\p{Is_Hex_Digit= _T}', "");
Expect(1, 65351, '\p{^Is_Hex_Digit= _T}', "");
Expect(1, 65351, '\P{Is_Hex_Digit= _T}', "");
Expect(0, 65351, '\P{^Is_Hex_Digit= _T}', "");
Error('\p{Is_Hex=-/a/true}');
Error('\P{Is_Hex=-/a/true}');
Expect(1, 65350, '\p{Is_Hex=true}', "");
Expect(0, 65350, '\p{^Is_Hex=true}', "");
Expect(0, 65350, '\P{Is_Hex=true}', "");
Expect(1, 65350, '\P{^Is_Hex=true}', "");
Expect(0, 65351, '\p{Is_Hex=true}', "");
Expect(1, 65351, '\p{^Is_Hex=true}', "");
Expect(1, 65351, '\P{Is_Hex=true}', "");
Expect(0, 65351, '\P{^Is_Hex=true}', "");
Expect(1, 65350, '\p{Is_Hex=- true}', "");
Expect(0, 65350, '\p{^Is_Hex=- true}', "");
Expect(0, 65350, '\P{Is_Hex=- true}', "");
Expect(1, 65350, '\P{^Is_Hex=- true}', "");
Expect(0, 65351, '\p{Is_Hex=- true}', "");
Expect(1, 65351, '\p{^Is_Hex=- true}', "");
Expect(1, 65351, '\P{Is_Hex=- true}', "");
Expect(0, 65351, '\P{^Is_Hex=- true}', "");
Error('\p{Hyphen=/a/-_No}');
Error('\P{Hyphen=/a/-_No}');
Expect(1, 65382, '\p{Hyphen=no}', 'deprecated');
Expect(0, 65382, '\p{^Hyphen=no}', 'deprecated');
Expect(0, 65382, '\P{Hyphen=no}', 'deprecated');
Expect(1, 65382, '\P{^Hyphen=no}', 'deprecated');
Expect(0, 65381, '\p{Hyphen=no}', 'deprecated');
Expect(1, 65381, '\p{^Hyphen=no}', 'deprecated');
Expect(1, 65381, '\P{Hyphen=no}', 'deprecated');
Expect(0, 65381, '\P{^Hyphen=no}', 'deprecated');
Expect(1, 65382, '\p{Hyphen=	no}', 'deprecated');
Expect(0, 65382, '\p{^Hyphen=	no}', 'deprecated');
Expect(0, 65382, '\P{Hyphen=	no}', 'deprecated');
Expect(1, 65382, '\P{^Hyphen=	no}', 'deprecated');
Expect(0, 65381, '\p{Hyphen=	no}', 'deprecated');
Expect(1, 65381, '\p{^Hyphen=	no}', 'deprecated');
Expect(1, 65381, '\P{Hyphen=	no}', 'deprecated');
Expect(0, 65381, '\P{^Hyphen=	no}', 'deprecated');
Error('\p{Is_Hyphen=	:=N}');
Error('\P{Is_Hyphen=	:=N}');
Expect(1, 65382, '\p{Is_Hyphen=n}', 'deprecated');
Expect(0, 65382, '\p{^Is_Hyphen=n}', 'deprecated');
Expect(0, 65382, '\P{Is_Hyphen=n}', 'deprecated');
Expect(1, 65382, '\P{^Is_Hyphen=n}', 'deprecated');
Expect(0, 65381, '\p{Is_Hyphen=n}', 'deprecated');
Expect(1, 65381, '\p{^Is_Hyphen=n}', 'deprecated');
Expect(1, 65381, '\P{Is_Hyphen=n}', 'deprecated');
Expect(0, 65381, '\P{^Is_Hyphen=n}', 'deprecated');
Expect(1, 65382, '\p{Is_Hyphen=_-N}', 'deprecated');
Expect(0, 65382, '\p{^Is_Hyphen=_-N}', 'deprecated');
Expect(0, 65382, '\P{Is_Hyphen=_-N}', 'deprecated');
Expect(1, 65382, '\P{^Is_Hyphen=_-N}', 'deprecated');
Expect(0, 65381, '\p{Is_Hyphen=_-N}', 'deprecated');
Expect(1, 65381, '\p{^Is_Hyphen=_-N}', 'deprecated');
Expect(1, 65381, '\P{Is_Hyphen=_-N}', 'deprecated');
Expect(0, 65381, '\P{^Is_Hyphen=_-N}', 'deprecated');
Error('\p{Hyphen=f:=}');
Error('\P{Hyphen=f:=}');
Expect(1, 65382, '\p{Hyphen=f}', 'deprecated');
Expect(0, 65382, '\p{^Hyphen=f}', 'deprecated');
Expect(0, 65382, '\P{Hyphen=f}', 'deprecated');
Expect(1, 65382, '\P{^Hyphen=f}', 'deprecated');
Expect(0, 65381, '\p{Hyphen=f}', 'deprecated');
Expect(1, 65381, '\p{^Hyphen=f}', 'deprecated');
Expect(1, 65381, '\P{Hyphen=f}', 'deprecated');
Expect(0, 65381, '\P{^Hyphen=f}', 'deprecated');
Expect(1, 65382, '\p{Hyphen=_ F}', 'deprecated');
Expect(0, 65382, '\p{^Hyphen=_ F}', 'deprecated');
Expect(0, 65382, '\P{Hyphen=_ F}', 'deprecated');
Expect(1, 65382, '\P{^Hyphen=_ F}', 'deprecated');
Expect(0, 65381, '\p{Hyphen=_ F}', 'deprecated');
Expect(1, 65381, '\p{^Hyphen=_ F}', 'deprecated');
Expect(1, 65381, '\P{Hyphen=_ F}', 'deprecated');
Expect(0, 65381, '\P{^Hyphen=_ F}', 'deprecated');
Error('\p{Is_Hyphen=/a/- false}');
Error('\P{Is_Hyphen=/a/- false}');
Expect(1, 65382, '\p{Is_Hyphen=false}', 'deprecated');
Expect(0, 65382, '\p{^Is_Hyphen=false}', 'deprecated');
Expect(0, 65382, '\P{Is_Hyphen=false}', 'deprecated');
Expect(1, 65382, '\P{^Is_Hyphen=false}', 'deprecated');
Expect(0, 65381, '\p{Is_Hyphen=false}', 'deprecated');
Expect(1, 65381, '\p{^Is_Hyphen=false}', 'deprecated');
Expect(1, 65381, '\P{Is_Hyphen=false}', 'deprecated');
Expect(0, 65381, '\P{^Is_Hyphen=false}', 'deprecated');
Expect(1, 65382, '\p{Is_Hyphen=_ False}', 'deprecated');
Expect(0, 65382, '\p{^Is_Hyphen=_ False}', 'deprecated');
Expect(0, 65382, '\P{Is_Hyphen=_ False}', 'deprecated');
Expect(1, 65382, '\P{^Is_Hyphen=_ False}', 'deprecated');
Expect(0, 65381, '\p{Is_Hyphen=_ False}', 'deprecated');
Expect(1, 65381, '\p{^Is_Hyphen=_ False}', 'deprecated');
Expect(1, 65381, '\P{Is_Hyphen=_ False}', 'deprecated');
Expect(0, 65381, '\P{^Is_Hyphen=_ False}', 'deprecated');
Error('\p{Hyphen=/a/--Yes}');
Error('\P{Hyphen=/a/--Yes}');
Expect(1, 65381, '\p{Hyphen=yes}', 'deprecated');
Expect(0, 65381, '\p{^Hyphen=yes}', 'deprecated');
Expect(0, 65381, '\P{Hyphen=yes}', 'deprecated');
Expect(1, 65381, '\P{^Hyphen=yes}', 'deprecated');
Expect(0, 65382, '\p{Hyphen=yes}', 'deprecated');
Expect(1, 65382, '\p{^Hyphen=yes}', 'deprecated');
Expect(1, 65382, '\P{Hyphen=yes}', 'deprecated');
Expect(0, 65382, '\P{^Hyphen=yes}', 'deprecated');
Expect(1, 65381, '\p{Hyphen= _yes}', 'deprecated');
Expect(0, 65381, '\p{^Hyphen= _yes}', 'deprecated');
Expect(0, 65381, '\P{Hyphen= _yes}', 'deprecated');
Expect(1, 65381, '\P{^Hyphen= _yes}', 'deprecated');
Expect(0, 65382, '\p{Hyphen= _yes}', 'deprecated');
Expect(1, 65382, '\p{^Hyphen= _yes}', 'deprecated');
Expect(1, 65382, '\P{Hyphen= _yes}', 'deprecated');
Expect(0, 65382, '\P{^Hyphen= _yes}', 'deprecated');
Error('\p{Is_Hyphen=/a/  Y}');
Error('\P{Is_Hyphen=/a/  Y}');
Expect(1, 65381, '\p{Is_Hyphen=y}', 'deprecated');
Expect(0, 65381, '\p{^Is_Hyphen=y}', 'deprecated');
Expect(0, 65381, '\P{Is_Hyphen=y}', 'deprecated');
Expect(1, 65381, '\P{^Is_Hyphen=y}', 'deprecated');
Expect(0, 65382, '\p{Is_Hyphen=y}', 'deprecated');
Expect(1, 65382, '\p{^Is_Hyphen=y}', 'deprecated');
Expect(1, 65382, '\P{Is_Hyphen=y}', 'deprecated');
Expect(0, 65382, '\P{^Is_Hyphen=y}', 'deprecated');
Expect(1, 65381, '\p{Is_Hyphen=	Y}', 'deprecated');
Expect(0, 65381, '\p{^Is_Hyphen=	Y}', 'deprecated');
Expect(0, 65381, '\P{Is_Hyphen=	Y}', 'deprecated');
Expect(1, 65381, '\P{^Is_Hyphen=	Y}', 'deprecated');
Expect(0, 65382, '\p{Is_Hyphen=	Y}', 'deprecated');
Expect(1, 65382, '\p{^Is_Hyphen=	Y}', 'deprecated');
Expect(1, 65382, '\P{Is_Hyphen=	Y}', 'deprecated');
Expect(0, 65382, '\P{^Is_Hyphen=	Y}', 'deprecated');
Error('\p{Hyphen=_:=t}');
Error('\P{Hyphen=_:=t}');
Expect(1, 65381, '\p{Hyphen=t}', 'deprecated');
Expect(0, 65381, '\p{^Hyphen=t}', 'deprecated');
Expect(0, 65381, '\P{Hyphen=t}', 'deprecated');
Expect(1, 65381, '\P{^Hyphen=t}', 'deprecated');
Expect(0, 65382, '\p{Hyphen=t}', 'deprecated');
Expect(1, 65382, '\p{^Hyphen=t}', 'deprecated');
Expect(1, 65382, '\P{Hyphen=t}', 'deprecated');
Expect(0, 65382, '\P{^Hyphen=t}', 'deprecated');
Expect(1, 65381, '\p{Hyphen= 	t}', 'deprecated');
Expect(0, 65381, '\p{^Hyphen= 	t}', 'deprecated');
Expect(0, 65381, '\P{Hyphen= 	t}', 'deprecated');
Expect(1, 65381, '\P{^Hyphen= 	t}', 'deprecated');
Expect(0, 65382, '\p{Hyphen= 	t}', 'deprecated');
Expect(1, 65382, '\p{^Hyphen= 	t}', 'deprecated');
Expect(1, 65382, '\P{Hyphen= 	t}', 'deprecated');
Expect(0, 65382, '\P{^Hyphen= 	t}', 'deprecated');
Error('\p{Is_Hyphen=/a/	True}');
Error('\P{Is_Hyphen=/a/	True}');
Expect(1, 65381, '\p{Is_Hyphen=true}', 'deprecated');
Expect(0, 65381, '\p{^Is_Hyphen=true}', 'deprecated');
Expect(0, 65381, '\P{Is_Hyphen=true}', 'deprecated');
Expect(1, 65381, '\P{^Is_Hyphen=true}', 'deprecated');
Expect(0, 65382, '\p{Is_Hyphen=true}', 'deprecated');
Expect(1, 65382, '\p{^Is_Hyphen=true}', 'deprecated');
Expect(1, 65382, '\P{Is_Hyphen=true}', 'deprecated');
Expect(0, 65382, '\P{^Is_Hyphen=true}', 'deprecated');
Expect(1, 65381, '\p{Is_Hyphen=	_True}', 'deprecated');
Expect(0, 65381, '\p{^Is_Hyphen=	_True}', 'deprecated');
Expect(0, 65381, '\P{Is_Hyphen=	_True}', 'deprecated');
Expect(1, 65381, '\P{^Is_Hyphen=	_True}', 'deprecated');
Expect(0, 65382, '\p{Is_Hyphen=	_True}', 'deprecated');
Expect(1, 65382, '\p{^Is_Hyphen=	_True}', 'deprecated');
Expect(1, 65382, '\P{Is_Hyphen=	_True}', 'deprecated');
Expect(0, 65382, '\P{^Is_Hyphen=	_True}', 'deprecated');
Error('\p{ID_Continue=	No/a/}');
Error('\P{ID_Continue=	No/a/}');
Expect(1, 918000, '\p{ID_Continue=no}', "");
Expect(0, 918000, '\p{^ID_Continue=no}', "");
Expect(0, 918000, '\P{ID_Continue=no}', "");
Expect(1, 918000, '\P{^ID_Continue=no}', "");
Expect(0, 917999, '\p{ID_Continue=no}', "");
Expect(1, 917999, '\p{^ID_Continue=no}', "");
Expect(1, 917999, '\P{ID_Continue=no}', "");
Expect(0, 917999, '\P{^ID_Continue=no}', "");
Expect(1, 918000, '\p{ID_Continue=	_No}', "");
Expect(0, 918000, '\p{^ID_Continue=	_No}', "");
Expect(0, 918000, '\P{ID_Continue=	_No}', "");
Expect(1, 918000, '\P{^ID_Continue=	_No}', "");
Expect(0, 917999, '\p{ID_Continue=	_No}', "");
Expect(1, 917999, '\p{^ID_Continue=	_No}', "");
Expect(1, 917999, '\P{ID_Continue=	_No}', "");
Expect(0, 917999, '\P{^ID_Continue=	_No}', "");
Error('\p{IDC:_N/a/}');
Error('\P{IDC:_N/a/}');
Expect(1, 918000, '\p{IDC=n}', "");
Expect(0, 918000, '\p{^IDC=n}', "");
Expect(0, 918000, '\P{IDC=n}', "");
Expect(1, 918000, '\P{^IDC=n}', "");
Expect(0, 917999, '\p{IDC=n}', "");
Expect(1, 917999, '\p{^IDC=n}', "");
Expect(1, 917999, '\P{IDC=n}', "");
Expect(0, 917999, '\P{^IDC=n}', "");
Expect(1, 918000, '\p{IDC=	-N}', "");
Expect(0, 918000, '\p{^IDC=	-N}', "");
Expect(0, 918000, '\P{IDC=	-N}', "");
Expect(1, 918000, '\P{^IDC=	-N}', "");
Expect(0, 917999, '\p{IDC=	-N}', "");
Expect(1, 917999, '\p{^IDC=	-N}', "");
Expect(1, 917999, '\P{IDC=	-N}', "");
Expect(0, 917999, '\P{^IDC=	-N}', "");
Error('\p{Is_ID_Continue=/a/__F}');
Error('\P{Is_ID_Continue=/a/__F}');
Expect(1, 918000, '\p{Is_ID_Continue=f}', "");
Expect(0, 918000, '\p{^Is_ID_Continue=f}', "");
Expect(0, 918000, '\P{Is_ID_Continue=f}', "");
Expect(1, 918000, '\P{^Is_ID_Continue=f}', "");
Expect(0, 917999, '\p{Is_ID_Continue=f}', "");
Expect(1, 917999, '\p{^Is_ID_Continue=f}', "");
Expect(1, 917999, '\P{Is_ID_Continue=f}', "");
Expect(0, 917999, '\P{^Is_ID_Continue=f}', "");
Expect(1, 918000, '\p{Is_ID_Continue=	F}', "");
Expect(0, 918000, '\p{^Is_ID_Continue=	F}', "");
Expect(0, 918000, '\P{Is_ID_Continue=	F}', "");
Expect(1, 918000, '\P{^Is_ID_Continue=	F}', "");
Expect(0, 917999, '\p{Is_ID_Continue=	F}', "");
Expect(1, 917999, '\p{^Is_ID_Continue=	F}', "");
Expect(1, 917999, '\P{Is_ID_Continue=	F}', "");
Expect(0, 917999, '\P{^Is_ID_Continue=	F}', "");
Error('\p{Is_IDC=/a/--False}');
Error('\P{Is_IDC=/a/--False}');
Expect(1, 918000, '\p{Is_IDC: false}', "");
Expect(0, 918000, '\p{^Is_IDC: false}', "");
Expect(0, 918000, '\P{Is_IDC: false}', "");
Expect(1, 918000, '\P{^Is_IDC: false}', "");
Expect(0, 917999, '\p{Is_IDC: false}', "");
Expect(1, 917999, '\p{^Is_IDC: false}', "");
Expect(1, 917999, '\P{Is_IDC: false}', "");
Expect(0, 917999, '\P{^Is_IDC: false}', "");
Expect(1, 918000, '\p{Is_IDC= _false}', "");
Expect(0, 918000, '\p{^Is_IDC= _false}', "");
Expect(0, 918000, '\P{Is_IDC= _false}', "");
Expect(1, 918000, '\P{^Is_IDC= _false}', "");
Expect(0, 917999, '\p{Is_IDC= _false}', "");
Expect(1, 917999, '\p{^Is_IDC= _false}', "");
Expect(1, 917999, '\P{Is_IDC= _false}', "");
Expect(0, 917999, '\P{^Is_IDC= _false}', "");
Error('\p{ID_Continue=	:=Yes}');
Error('\P{ID_Continue=	:=Yes}');
Expect(1, 917999, '\p{ID_Continue=yes}', "");
Expect(0, 917999, '\p{^ID_Continue=yes}', "");
Expect(0, 917999, '\P{ID_Continue=yes}', "");
Expect(1, 917999, '\P{^ID_Continue=yes}', "");
Expect(0, 918000, '\p{ID_Continue=yes}', "");
Expect(1, 918000, '\p{^ID_Continue=yes}', "");
Expect(1, 918000, '\P{ID_Continue=yes}', "");
Expect(0, 918000, '\P{^ID_Continue=yes}', "");
Expect(1, 917999, '\p{ID_Continue=	-yes}', "");
Expect(0, 917999, '\p{^ID_Continue=	-yes}', "");
Expect(0, 917999, '\P{ID_Continue=	-yes}', "");
Expect(1, 917999, '\P{^ID_Continue=	-yes}', "");
Expect(0, 918000, '\p{ID_Continue=	-yes}', "");
Expect(1, 918000, '\p{^ID_Continue=	-yes}', "");
Expect(1, 918000, '\P{ID_Continue=	-yes}', "");
Expect(0, 918000, '\P{^ID_Continue=	-yes}', "");
Error('\p{IDC=:= Y}');
Error('\P{IDC=:= Y}');
Expect(1, 917999, '\p{IDC:y}', "");
Expect(0, 917999, '\p{^IDC:y}', "");
Expect(0, 917999, '\P{IDC:y}', "");
Expect(1, 917999, '\P{^IDC:y}', "");
Expect(0, 918000, '\p{IDC:y}', "");
Expect(1, 918000, '\p{^IDC:y}', "");
Expect(1, 918000, '\P{IDC:y}', "");
Expect(0, 918000, '\P{^IDC:y}', "");
Expect(1, 917999, '\p{IDC=_	Y}', "");
Expect(0, 917999, '\p{^IDC=_	Y}', "");
Expect(0, 917999, '\P{IDC=_	Y}', "");
Expect(1, 917999, '\P{^IDC=_	Y}', "");
Expect(0, 918000, '\p{IDC=_	Y}', "");
Expect(1, 918000, '\p{^IDC=_	Y}', "");
Expect(1, 918000, '\P{IDC=_	Y}', "");
Expect(0, 918000, '\P{^IDC=_	Y}', "");
Error('\p{Is_ID_Continue=--t/a/}');
Error('\P{Is_ID_Continue=--t/a/}');
Expect(1, 917999, '\p{Is_ID_Continue=t}', "");
Expect(0, 917999, '\p{^Is_ID_Continue=t}', "");
Expect(0, 917999, '\P{Is_ID_Continue=t}', "");
Expect(1, 917999, '\P{^Is_ID_Continue=t}', "");
Expect(0, 918000, '\p{Is_ID_Continue=t}', "");
Expect(1, 918000, '\p{^Is_ID_Continue=t}', "");
Expect(1, 918000, '\P{Is_ID_Continue=t}', "");
Expect(0, 918000, '\P{^Is_ID_Continue=t}', "");
Expect(1, 917999, '\p{Is_ID_Continue= _T}', "");
Expect(0, 917999, '\p{^Is_ID_Continue= _T}', "");
Expect(0, 917999, '\P{Is_ID_Continue= _T}', "");
Expect(1, 917999, '\P{^Is_ID_Continue= _T}', "");
Expect(0, 918000, '\p{Is_ID_Continue= _T}', "");
Expect(1, 918000, '\p{^Is_ID_Continue= _T}', "");
Expect(1, 918000, '\P{Is_ID_Continue= _T}', "");
Expect(0, 918000, '\P{^Is_ID_Continue= _T}', "");
Error('\p{Is_IDC= 	True/a/}');
Error('\P{Is_IDC= 	True/a/}');
Expect(1, 917999, '\p{Is_IDC=true}', "");
Expect(0, 917999, '\p{^Is_IDC=true}', "");
Expect(0, 917999, '\P{Is_IDC=true}', "");
Expect(1, 917999, '\P{^Is_IDC=true}', "");
Expect(0, 918000, '\p{Is_IDC=true}', "");
Expect(1, 918000, '\p{^Is_IDC=true}', "");
Expect(1, 918000, '\P{Is_IDC=true}', "");
Expect(0, 918000, '\P{^Is_IDC=true}', "");
Expect(1, 917999, '\p{Is_IDC=  True}', "");
Expect(0, 917999, '\p{^Is_IDC=  True}', "");
Expect(0, 917999, '\P{Is_IDC=  True}', "");
Expect(1, 917999, '\P{^Is_IDC=  True}', "");
Expect(0, 918000, '\p{Is_IDC=  True}', "");
Expect(1, 918000, '\p{^Is_IDC=  True}', "");
Expect(1, 918000, '\P{Is_IDC=  True}', "");
Expect(0, 918000, '\P{^Is_IDC=  True}', "");
Error('\p{ID_Start=/a/	no}');
Error('\P{ID_Start=/a/	no}');
Expect(1, 195102, '\p{ID_Start=no}', "");
Expect(0, 195102, '\p{^ID_Start=no}', "");
Expect(0, 195102, '\P{ID_Start=no}', "");
Expect(1, 195102, '\P{^ID_Start=no}', "");
Expect(0, 195101, '\p{ID_Start=no}', "");
Expect(1, 195101, '\p{^ID_Start=no}', "");
Expect(1, 195101, '\P{ID_Start=no}', "");
Expect(0, 195101, '\P{^ID_Start=no}', "");
Expect(1, 195102, '\p{ID_Start=- no}', "");
Expect(0, 195102, '\p{^ID_Start=- no}', "");
Expect(0, 195102, '\P{ID_Start=- no}', "");
Expect(1, 195102, '\P{^ID_Start=- no}', "");
Expect(0, 195101, '\p{ID_Start=- no}', "");
Expect(1, 195101, '\p{^ID_Start=- no}', "");
Expect(1, 195101, '\P{ID_Start=- no}', "");
Expect(0, 195101, '\P{^ID_Start=- no}', "");
Error('\p{IDS=-/a/N}');
Error('\P{IDS=-/a/N}');
Expect(1, 195102, '\p{IDS:	n}', "");
Expect(0, 195102, '\p{^IDS:	n}', "");
Expect(0, 195102, '\P{IDS:	n}', "");
Expect(1, 195102, '\P{^IDS:	n}', "");
Expect(0, 195101, '\p{IDS:	n}', "");
Expect(1, 195101, '\p{^IDS:	n}', "");
Expect(1, 195101, '\P{IDS:	n}', "");
Expect(0, 195101, '\P{^IDS:	n}', "");
Expect(1, 195102, '\p{IDS:  	n}', "");
Expect(0, 195102, '\p{^IDS:  	n}', "");
Expect(0, 195102, '\P{IDS:  	n}', "");
Expect(1, 195102, '\P{^IDS:  	n}', "");
Expect(0, 195101, '\p{IDS:  	n}', "");
Expect(1, 195101, '\p{^IDS:  	n}', "");
Expect(1, 195101, '\P{IDS:  	n}', "");
Expect(0, 195101, '\P{^IDS:  	n}', "");
Error('\p{Is_ID_Start=/a/	-F}');
Error('\P{Is_ID_Start=/a/	-F}');
Expect(1, 195102, '\p{Is_ID_Start=f}', "");
Expect(0, 195102, '\p{^Is_ID_Start=f}', "");
Expect(0, 195102, '\P{Is_ID_Start=f}', "");
Expect(1, 195102, '\P{^Is_ID_Start=f}', "");
Expect(0, 195101, '\p{Is_ID_Start=f}', "");
Expect(1, 195101, '\p{^Is_ID_Start=f}', "");
Expect(1, 195101, '\P{Is_ID_Start=f}', "");
Expect(0, 195101, '\P{^Is_ID_Start=f}', "");
Expect(1, 195102, '\p{Is_ID_Start= F}', "");
Expect(0, 195102, '\p{^Is_ID_Start= F}', "");
Expect(0, 195102, '\P{Is_ID_Start= F}', "");
Expect(1, 195102, '\P{^Is_ID_Start= F}', "");
Expect(0, 195101, '\p{Is_ID_Start= F}', "");
Expect(1, 195101, '\p{^Is_ID_Start= F}', "");
Expect(1, 195101, '\P{Is_ID_Start= F}', "");
Expect(0, 195101, '\P{^Is_ID_Start= F}', "");
Error('\p{Is_IDS=	:=False}');
Error('\P{Is_IDS=	:=False}');
Expect(1, 195102, '\p{Is_IDS=false}', "");
Expect(0, 195102, '\p{^Is_IDS=false}', "");
Expect(0, 195102, '\P{Is_IDS=false}', "");
Expect(1, 195102, '\P{^Is_IDS=false}', "");
Expect(0, 195101, '\p{Is_IDS=false}', "");
Expect(1, 195101, '\p{^Is_IDS=false}', "");
Expect(1, 195101, '\P{Is_IDS=false}', "");
Expect(0, 195101, '\P{^Is_IDS=false}', "");
Expect(1, 195102, '\p{Is_IDS= FALSE}', "");
Expect(0, 195102, '\p{^Is_IDS= FALSE}', "");
Expect(0, 195102, '\P{Is_IDS= FALSE}', "");
Expect(1, 195102, '\P{^Is_IDS= FALSE}', "");
Expect(0, 195101, '\p{Is_IDS= FALSE}', "");
Expect(1, 195101, '\p{^Is_IDS= FALSE}', "");
Expect(1, 195101, '\P{Is_IDS= FALSE}', "");
Expect(0, 195101, '\P{^Is_IDS= FALSE}', "");
Error('\p{ID_Start=/a/ -Yes}');
Error('\P{ID_Start=/a/ -Yes}');
Expect(1, 195101, '\p{ID_Start=yes}', "");
Expect(0, 195101, '\p{^ID_Start=yes}', "");
Expect(0, 195101, '\P{ID_Start=yes}', "");
Expect(1, 195101, '\P{^ID_Start=yes}', "");
Expect(0, 195102, '\p{ID_Start=yes}', "");
Expect(1, 195102, '\p{^ID_Start=yes}', "");
Expect(1, 195102, '\P{ID_Start=yes}', "");
Expect(0, 195102, '\P{^ID_Start=yes}', "");
Expect(1, 195101, '\p{ID_Start=_Yes}', "");
Expect(0, 195101, '\p{^ID_Start=_Yes}', "");
Expect(0, 195101, '\P{ID_Start=_Yes}', "");
Expect(1, 195101, '\P{^ID_Start=_Yes}', "");
Expect(0, 195102, '\p{ID_Start=_Yes}', "");
Expect(1, 195102, '\p{^ID_Start=_Yes}', "");
Expect(1, 195102, '\P{ID_Start=_Yes}', "");
Expect(0, 195102, '\P{^ID_Start=_Yes}', "");
Error('\p{IDS=:= Y}');
Error('\P{IDS=:= Y}');
Expect(1, 195101, '\p{IDS:y}', "");
Expect(0, 195101, '\p{^IDS:y}', "");
Expect(0, 195101, '\P{IDS:y}', "");
Expect(1, 195101, '\P{^IDS:y}', "");
Expect(0, 195102, '\p{IDS:y}', "");
Expect(1, 195102, '\p{^IDS:y}', "");
Expect(1, 195102, '\P{IDS:y}', "");
Expect(0, 195102, '\P{^IDS:y}', "");
Expect(1, 195101, '\p{IDS= 	Y}', "");
Expect(0, 195101, '\p{^IDS= 	Y}', "");
Expect(0, 195101, '\P{IDS= 	Y}', "");
Expect(1, 195101, '\P{^IDS= 	Y}', "");
Expect(0, 195102, '\p{IDS= 	Y}', "");
Expect(1, 195102, '\p{^IDS= 	Y}', "");
Expect(1, 195102, '\P{IDS= 	Y}', "");
Expect(0, 195102, '\P{^IDS= 	Y}', "");
Error('\p{Is_ID_Start=_T:=}');
Error('\P{Is_ID_Start=_T:=}');
Expect(1, 195101, '\p{Is_ID_Start=t}', "");
Expect(0, 195101, '\p{^Is_ID_Start=t}', "");
Expect(0, 195101, '\P{Is_ID_Start=t}', "");
Expect(1, 195101, '\P{^Is_ID_Start=t}', "");
Expect(0, 195102, '\p{Is_ID_Start=t}', "");
Expect(1, 195102, '\p{^Is_ID_Start=t}', "");
Expect(1, 195102, '\P{Is_ID_Start=t}', "");
Expect(0, 195102, '\P{^Is_ID_Start=t}', "");
Expect(1, 195101, '\p{Is_ID_Start=--t}', "");
Expect(0, 195101, '\p{^Is_ID_Start=--t}', "");
Expect(0, 195101, '\P{Is_ID_Start=--t}', "");
Expect(1, 195101, '\P{^Is_ID_Start=--t}', "");
Expect(0, 195102, '\p{Is_ID_Start=--t}', "");
Expect(1, 195102, '\p{^Is_ID_Start=--t}', "");
Expect(1, 195102, '\P{Is_ID_Start=--t}', "");
Expect(0, 195102, '\P{^Is_ID_Start=--t}', "");
Error('\p{Is_IDS=-_True/a/}');
Error('\P{Is_IDS=-_True/a/}');
Expect(1, 195101, '\p{Is_IDS=true}', "");
Expect(0, 195101, '\p{^Is_IDS=true}', "");
Expect(0, 195101, '\P{Is_IDS=true}', "");
Expect(1, 195101, '\P{^Is_IDS=true}', "");
Expect(0, 195102, '\p{Is_IDS=true}', "");
Expect(1, 195102, '\p{^Is_IDS=true}', "");
Expect(1, 195102, '\P{Is_IDS=true}', "");
Expect(0, 195102, '\P{^Is_IDS=true}', "");
Expect(1, 195101, '\p{Is_IDS=	 true}', "");
Expect(0, 195101, '\p{^Is_IDS=	 true}', "");
Expect(0, 195101, '\P{Is_IDS=	 true}', "");
Expect(1, 195101, '\P{^Is_IDS=	 true}', "");
Expect(0, 195102, '\p{Is_IDS=	 true}', "");
Expect(1, 195102, '\p{^Is_IDS=	 true}', "");
Expect(1, 195102, '\P{Is_IDS=	 true}', "");
Expect(0, 195102, '\P{^Is_IDS=	 true}', "");
Error('\p{IDS_Binary_Operator=/a/ NO}');
Error('\P{IDS_Binary_Operator=/a/ NO}');
Expect(1, 12284, '\p{IDS_Binary_Operator=no}', "");
Expect(0, 12284, '\p{^IDS_Binary_Operator=no}', "");
Expect(0, 12284, '\P{IDS_Binary_Operator=no}', "");
Expect(1, 12284, '\P{^IDS_Binary_Operator=no}', "");
Expect(0, 12283, '\p{IDS_Binary_Operator=no}', "");
Expect(1, 12283, '\p{^IDS_Binary_Operator=no}', "");
Expect(1, 12283, '\P{IDS_Binary_Operator=no}', "");
Expect(0, 12283, '\P{^IDS_Binary_Operator=no}', "");
Expect(1, 12284, '\p{IDS_Binary_Operator= _no}', "");
Expect(0, 12284, '\p{^IDS_Binary_Operator= _no}', "");
Expect(0, 12284, '\P{IDS_Binary_Operator= _no}', "");
Expect(1, 12284, '\P{^IDS_Binary_Operator= _no}', "");
Expect(0, 12283, '\p{IDS_Binary_Operator= _no}', "");
Expect(1, 12283, '\p{^IDS_Binary_Operator= _no}', "");
Expect(1, 12283, '\P{IDS_Binary_Operator= _no}', "");
Expect(0, 12283, '\P{^IDS_Binary_Operator= _no}', "");
Error('\p{IDSB= :=n}');
Error('\P{IDSB= :=n}');
Expect(1, 12284, '\p{IDSB=n}', "");
Expect(0, 12284, '\p{^IDSB=n}', "");
Expect(0, 12284, '\P{IDSB=n}', "");
Expect(1, 12284, '\P{^IDSB=n}', "");
Expect(0, 12283, '\p{IDSB=n}', "");
Expect(1, 12283, '\p{^IDSB=n}', "");
Expect(1, 12283, '\P{IDSB=n}', "");
Expect(0, 12283, '\P{^IDSB=n}', "");
Expect(1, 12284, '\p{IDSB= N}', "");
Expect(0, 12284, '\p{^IDSB= N}', "");
Expect(0, 12284, '\P{IDSB= N}', "");
Expect(1, 12284, '\P{^IDSB= N}', "");
Expect(0, 12283, '\p{IDSB= N}', "");
Expect(1, 12283, '\p{^IDSB= N}', "");
Expect(1, 12283, '\P{IDSB= N}', "");
Expect(0, 12283, '\P{^IDSB= N}', "");
Error('\p{Is_IDS_Binary_Operator=/a/f}');
Error('\P{Is_IDS_Binary_Operator=/a/f}');
Expect(1, 12284, '\p{Is_IDS_Binary_Operator=f}', "");
Expect(0, 12284, '\p{^Is_IDS_Binary_Operator=f}', "");
Expect(0, 12284, '\P{Is_IDS_Binary_Operator=f}', "");
Expect(1, 12284, '\P{^Is_IDS_Binary_Operator=f}', "");
Expect(0, 12283, '\p{Is_IDS_Binary_Operator=f}', "");
Expect(1, 12283, '\p{^Is_IDS_Binary_Operator=f}', "");
Expect(1, 12283, '\P{Is_IDS_Binary_Operator=f}', "");
Expect(0, 12283, '\P{^Is_IDS_Binary_Operator=f}', "");
Expect(1, 12284, '\p{Is_IDS_Binary_Operator=__F}', "");
Expect(0, 12284, '\p{^Is_IDS_Binary_Operator=__F}', "");
Expect(0, 12284, '\P{Is_IDS_Binary_Operator=__F}', "");
Expect(1, 12284, '\P{^Is_IDS_Binary_Operator=__F}', "");
Expect(0, 12283, '\p{Is_IDS_Binary_Operator=__F}', "");
Expect(1, 12283, '\p{^Is_IDS_Binary_Operator=__F}', "");
Expect(1, 12283, '\P{Is_IDS_Binary_Operator=__F}', "");
Expect(0, 12283, '\P{^Is_IDS_Binary_Operator=__F}', "");
Error('\p{Is_IDSB: /a/_-False}');
Error('\P{Is_IDSB: /a/_-False}');
Expect(1, 12284, '\p{Is_IDSB=false}', "");
Expect(0, 12284, '\p{^Is_IDSB=false}', "");
Expect(0, 12284, '\P{Is_IDSB=false}', "");
Expect(1, 12284, '\P{^Is_IDSB=false}', "");
Expect(0, 12283, '\p{Is_IDSB=false}', "");
Expect(1, 12283, '\p{^Is_IDSB=false}', "");
Expect(1, 12283, '\P{Is_IDSB=false}', "");
Expect(0, 12283, '\P{^Is_IDSB=false}', "");
Expect(1, 12284, '\p{Is_IDSB=FALSE}', "");
Expect(0, 12284, '\p{^Is_IDSB=FALSE}', "");
Expect(0, 12284, '\P{Is_IDSB=FALSE}', "");
Expect(1, 12284, '\P{^Is_IDSB=FALSE}', "");
Expect(0, 12283, '\p{Is_IDSB=FALSE}', "");
Expect(1, 12283, '\p{^Is_IDSB=FALSE}', "");
Expect(1, 12283, '\P{Is_IDSB=FALSE}', "");
Expect(0, 12283, '\P{^Is_IDSB=FALSE}', "");
Error('\p{IDS_Binary_Operator= /a/Yes}');
Error('\P{IDS_Binary_Operator= /a/Yes}');
Expect(1, 12283, '\p{IDS_Binary_Operator=yes}', "");
Expect(0, 12283, '\p{^IDS_Binary_Operator=yes}', "");
Expect(0, 12283, '\P{IDS_Binary_Operator=yes}', "");
Expect(1, 12283, '\P{^IDS_Binary_Operator=yes}', "");
Expect(0, 12284, '\p{IDS_Binary_Operator=yes}', "");
Expect(1, 12284, '\p{^IDS_Binary_Operator=yes}', "");
Expect(1, 12284, '\P{IDS_Binary_Operator=yes}', "");
Expect(0, 12284, '\P{^IDS_Binary_Operator=yes}', "");
Expect(1, 12283, '\p{IDS_Binary_Operator=	 Yes}', "");
Expect(0, 12283, '\p{^IDS_Binary_Operator=	 Yes}', "");
Expect(0, 12283, '\P{IDS_Binary_Operator=	 Yes}', "");
Expect(1, 12283, '\P{^IDS_Binary_Operator=	 Yes}', "");
Expect(0, 12284, '\p{IDS_Binary_Operator=	 Yes}', "");
Expect(1, 12284, '\p{^IDS_Binary_Operator=	 Yes}', "");
Expect(1, 12284, '\P{IDS_Binary_Operator=	 Yes}', "");
Expect(0, 12284, '\P{^IDS_Binary_Operator=	 Yes}', "");
Error('\p{IDSB=	y/a/}');
Error('\P{IDSB=	y/a/}');
Expect(1, 12283, '\p{IDSB=y}', "");
Expect(0, 12283, '\p{^IDSB=y}', "");
Expect(0, 12283, '\P{IDSB=y}', "");
Expect(1, 12283, '\P{^IDSB=y}', "");
Expect(0, 12284, '\p{IDSB=y}', "");
Expect(1, 12284, '\p{^IDSB=y}', "");
Expect(1, 12284, '\P{IDSB=y}', "");
Expect(0, 12284, '\P{^IDSB=y}', "");
Expect(1, 12283, '\p{IDSB=__y}', "");
Expect(0, 12283, '\p{^IDSB=__y}', "");
Expect(0, 12283, '\P{IDSB=__y}', "");
Expect(1, 12283, '\P{^IDSB=__y}', "");
Expect(0, 12284, '\p{IDSB=__y}', "");
Expect(1, 12284, '\p{^IDSB=__y}', "");
Expect(1, 12284, '\P{IDSB=__y}', "");
Expect(0, 12284, '\P{^IDSB=__y}', "");
Error('\p{Is_IDS_Binary_Operator=/a/t}');
Error('\P{Is_IDS_Binary_Operator=/a/t}');
Expect(1, 12283, '\p{Is_IDS_Binary_Operator:   t}', "");
Expect(0, 12283, '\p{^Is_IDS_Binary_Operator:   t}', "");
Expect(0, 12283, '\P{Is_IDS_Binary_Operator:   t}', "");
Expect(1, 12283, '\P{^Is_IDS_Binary_Operator:   t}', "");
Expect(0, 12284, '\p{Is_IDS_Binary_Operator:   t}', "");
Expect(1, 12284, '\p{^Is_IDS_Binary_Operator:   t}', "");
Expect(1, 12284, '\P{Is_IDS_Binary_Operator:   t}', "");
Expect(0, 12284, '\P{^Is_IDS_Binary_Operator:   t}', "");
Expect(1, 12283, '\p{Is_IDS_Binary_Operator= 	t}', "");
Expect(0, 12283, '\p{^Is_IDS_Binary_Operator= 	t}', "");
Expect(0, 12283, '\P{Is_IDS_Binary_Operator= 	t}', "");
Expect(1, 12283, '\P{^Is_IDS_Binary_Operator= 	t}', "");
Expect(0, 12284, '\p{Is_IDS_Binary_Operator= 	t}', "");
Expect(1, 12284, '\p{^Is_IDS_Binary_Operator= 	t}', "");
Expect(1, 12284, '\P{Is_IDS_Binary_Operator= 	t}', "");
Expect(0, 12284, '\P{^Is_IDS_Binary_Operator= 	t}', "");
Error('\p{Is_IDSB= true:=}');
Error('\P{Is_IDSB= true:=}');
Expect(1, 12283, '\p{Is_IDSB=true}', "");
Expect(0, 12283, '\p{^Is_IDSB=true}', "");
Expect(0, 12283, '\P{Is_IDSB=true}', "");
Expect(1, 12283, '\P{^Is_IDSB=true}', "");
Expect(0, 12284, '\p{Is_IDSB=true}', "");
Expect(1, 12284, '\p{^Is_IDSB=true}', "");
Expect(1, 12284, '\P{Is_IDSB=true}', "");
Expect(0, 12284, '\P{^Is_IDSB=true}', "");
Expect(1, 12283, '\p{Is_IDSB=_True}', "");
Expect(0, 12283, '\p{^Is_IDSB=_True}', "");
Expect(0, 12283, '\P{Is_IDSB=_True}', "");
Expect(1, 12283, '\P{^Is_IDSB=_True}', "");
Expect(0, 12284, '\p{Is_IDSB=_True}', "");
Expect(1, 12284, '\p{^Is_IDSB=_True}', "");
Expect(1, 12284, '\P{Is_IDSB=_True}', "");
Expect(0, 12284, '\P{^Is_IDSB=_True}', "");
Error('\p{IDS_Trinary_Operator:/a/- No}');
Error('\P{IDS_Trinary_Operator:/a/- No}');
Expect(1, 12276, '\p{IDS_Trinary_Operator=no}', "");
Expect(0, 12276, '\p{^IDS_Trinary_Operator=no}', "");
Expect(0, 12276, '\P{IDS_Trinary_Operator=no}', "");
Expect(1, 12276, '\P{^IDS_Trinary_Operator=no}', "");
Expect(0, 12275, '\p{IDS_Trinary_Operator=no}', "");
Expect(1, 12275, '\p{^IDS_Trinary_Operator=no}', "");
Expect(1, 12275, '\P{IDS_Trinary_Operator=no}', "");
Expect(0, 12275, '\P{^IDS_Trinary_Operator=no}', "");
Expect(1, 12276, '\p{IDS_Trinary_Operator=-	no}', "");
Expect(0, 12276, '\p{^IDS_Trinary_Operator=-	no}', "");
Expect(0, 12276, '\P{IDS_Trinary_Operator=-	no}', "");
Expect(1, 12276, '\P{^IDS_Trinary_Operator=-	no}', "");
Expect(0, 12275, '\p{IDS_Trinary_Operator=-	no}', "");
Expect(1, 12275, '\p{^IDS_Trinary_Operator=-	no}', "");
Expect(1, 12275, '\P{IDS_Trinary_Operator=-	no}', "");
Expect(0, 12275, '\P{^IDS_Trinary_Operator=-	no}', "");
Error('\p{IDST=/a/	N}');
Error('\P{IDST=/a/	N}');
Expect(1, 12276, '\p{IDST=n}', "");
Expect(0, 12276, '\p{^IDST=n}', "");
Expect(0, 12276, '\P{IDST=n}', "");
Expect(1, 12276, '\P{^IDST=n}', "");
Expect(0, 12275, '\p{IDST=n}', "");
Expect(1, 12275, '\p{^IDST=n}', "");
Expect(1, 12275, '\P{IDST=n}', "");
Expect(0, 12275, '\P{^IDST=n}', "");
Expect(1, 12276, '\p{IDST= n}', "");
Expect(0, 12276, '\p{^IDST= n}', "");
Expect(0, 12276, '\P{IDST= n}', "");
Expect(1, 12276, '\P{^IDST= n}', "");
Expect(0, 12275, '\p{IDST= n}', "");
Expect(1, 12275, '\p{^IDST= n}', "");
Expect(1, 12275, '\P{IDST= n}', "");
Expect(0, 12275, '\P{^IDST= n}', "");
Error('\p{Is_IDS_Trinary_Operator=/a/_-F}');
Error('\P{Is_IDS_Trinary_Operator=/a/_-F}');
Expect(1, 12276, '\p{Is_IDS_Trinary_Operator=f}', "");
Expect(0, 12276, '\p{^Is_IDS_Trinary_Operator=f}', "");
Expect(0, 12276, '\P{Is_IDS_Trinary_Operator=f}', "");
Expect(1, 12276, '\P{^Is_IDS_Trinary_Operator=f}', "");
Expect(0, 12275, '\p{Is_IDS_Trinary_Operator=f}', "");
Expect(1, 12275, '\p{^Is_IDS_Trinary_Operator=f}', "");
Expect(1, 12275, '\P{Is_IDS_Trinary_Operator=f}', "");
Expect(0, 12275, '\P{^Is_IDS_Trinary_Operator=f}', "");
Expect(1, 12276, '\p{Is_IDS_Trinary_Operator=_ F}', "");
Expect(0, 12276, '\p{^Is_IDS_Trinary_Operator=_ F}', "");
Expect(0, 12276, '\P{Is_IDS_Trinary_Operator=_ F}', "");
Expect(1, 12276, '\P{^Is_IDS_Trinary_Operator=_ F}', "");
Expect(0, 12275, '\p{Is_IDS_Trinary_Operator=_ F}', "");
Expect(1, 12275, '\p{^Is_IDS_Trinary_Operator=_ F}', "");
Expect(1, 12275, '\P{Is_IDS_Trinary_Operator=_ F}', "");
Expect(0, 12275, '\P{^Is_IDS_Trinary_Operator=_ F}', "");
Error('\p{Is_IDST=_/a/False}');
Error('\P{Is_IDST=_/a/False}');
Expect(1, 12276, '\p{Is_IDST=false}', "");
Expect(0, 12276, '\p{^Is_IDST=false}', "");
Expect(0, 12276, '\P{Is_IDST=false}', "");
Expect(1, 12276, '\P{^Is_IDST=false}', "");
Expect(0, 12275, '\p{Is_IDST=false}', "");
Expect(1, 12275, '\p{^Is_IDST=false}', "");
Expect(1, 12275, '\P{Is_IDST=false}', "");
Expect(0, 12275, '\P{^Is_IDST=false}', "");
Expect(1, 12276, '\p{Is_IDST= _False}', "");
Expect(0, 12276, '\p{^Is_IDST= _False}', "");
Expect(0, 12276, '\P{Is_IDST= _False}', "");
Expect(1, 12276, '\P{^Is_IDST= _False}', "");
Expect(0, 12275, '\p{Is_IDST= _False}', "");
Expect(1, 12275, '\p{^Is_IDST= _False}', "");
Expect(1, 12275, '\P{Is_IDST= _False}', "");
Expect(0, 12275, '\P{^Is_IDST= _False}', "");
Error('\p{IDS_Trinary_Operator:	 yes/a/}');
Error('\P{IDS_Trinary_Operator:	 yes/a/}');
Expect(1, 12275, '\p{IDS_Trinary_Operator=yes}', "");
Expect(0, 12275, '\p{^IDS_Trinary_Operator=yes}', "");
Expect(0, 12275, '\P{IDS_Trinary_Operator=yes}', "");
Expect(1, 12275, '\P{^IDS_Trinary_Operator=yes}', "");
Expect(0, 12276, '\p{IDS_Trinary_Operator=yes}', "");
Expect(1, 12276, '\p{^IDS_Trinary_Operator=yes}', "");
Expect(1, 12276, '\P{IDS_Trinary_Operator=yes}', "");
Expect(0, 12276, '\P{^IDS_Trinary_Operator=yes}', "");
Expect(1, 12275, '\p{IDS_Trinary_Operator:   --yes}', "");
Expect(0, 12275, '\p{^IDS_Trinary_Operator:   --yes}', "");
Expect(0, 12275, '\P{IDS_Trinary_Operator:   --yes}', "");
Expect(1, 12275, '\P{^IDS_Trinary_Operator:   --yes}', "");
Expect(0, 12276, '\p{IDS_Trinary_Operator:   --yes}', "");
Expect(1, 12276, '\p{^IDS_Trinary_Operator:   --yes}', "");
Expect(1, 12276, '\P{IDS_Trinary_Operator:   --yes}', "");
Expect(0, 12276, '\P{^IDS_Trinary_Operator:   --yes}', "");
Error('\p{IDST=:=-y}');
Error('\P{IDST=:=-y}');
Expect(1, 12275, '\p{IDST=y}', "");
Expect(0, 12275, '\p{^IDST=y}', "");
Expect(0, 12275, '\P{IDST=y}', "");
Expect(1, 12275, '\P{^IDST=y}', "");
Expect(0, 12276, '\p{IDST=y}', "");
Expect(1, 12276, '\p{^IDST=y}', "");
Expect(1, 12276, '\P{IDST=y}', "");
Expect(0, 12276, '\P{^IDST=y}', "");
Expect(1, 12275, '\p{IDST= Y}', "");
Expect(0, 12275, '\p{^IDST= Y}', "");
Expect(0, 12275, '\P{IDST= Y}', "");
Expect(1, 12275, '\P{^IDST= Y}', "");
Expect(0, 12276, '\p{IDST= Y}', "");
Expect(1, 12276, '\p{^IDST= Y}', "");
Expect(1, 12276, '\P{IDST= Y}', "");
Expect(0, 12276, '\P{^IDST= Y}', "");
Error('\p{Is_IDS_Trinary_Operator=/a/T}');
Error('\P{Is_IDS_Trinary_Operator=/a/T}');
Expect(1, 12275, '\p{Is_IDS_Trinary_Operator:   t}', "");
Expect(0, 12275, '\p{^Is_IDS_Trinary_Operator:   t}', "");
Expect(0, 12275, '\P{Is_IDS_Trinary_Operator:   t}', "");
Expect(1, 12275, '\P{^Is_IDS_Trinary_Operator:   t}', "");
Expect(0, 12276, '\p{Is_IDS_Trinary_Operator:   t}', "");
Expect(1, 12276, '\p{^Is_IDS_Trinary_Operator:   t}', "");
Expect(1, 12276, '\P{Is_IDS_Trinary_Operator:   t}', "");
Expect(0, 12276, '\P{^Is_IDS_Trinary_Operator:   t}', "");
Expect(1, 12275, '\p{Is_IDS_Trinary_Operator= T}', "");
Expect(0, 12275, '\p{^Is_IDS_Trinary_Operator= T}', "");
Expect(0, 12275, '\P{Is_IDS_Trinary_Operator= T}', "");
Expect(1, 12275, '\P{^Is_IDS_Trinary_Operator= T}', "");
Expect(0, 12276, '\p{Is_IDS_Trinary_Operator= T}', "");
Expect(1, 12276, '\p{^Is_IDS_Trinary_Operator= T}', "");
Expect(1, 12276, '\P{Is_IDS_Trinary_Operator= T}', "");
Expect(0, 12276, '\P{^Is_IDS_Trinary_Operator= T}', "");
Error('\p{Is_IDST=-true:=}');
Error('\P{Is_IDST=-true:=}');
Expect(1, 12275, '\p{Is_IDST=true}', "");
Expect(0, 12275, '\p{^Is_IDST=true}', "");
Expect(0, 12275, '\P{Is_IDST=true}', "");
Expect(1, 12275, '\P{^Is_IDST=true}', "");
Expect(0, 12276, '\p{Is_IDST=true}', "");
Expect(1, 12276, '\p{^Is_IDST=true}', "");
Expect(1, 12276, '\P{Is_IDST=true}', "");
Expect(0, 12276, '\P{^Is_IDST=true}', "");
Expect(1, 12275, '\p{Is_IDST=_	True}', "");
Expect(0, 12275, '\p{^Is_IDST=_	True}', "");
Expect(0, 12275, '\P{Is_IDST=_	True}', "");
Expect(1, 12275, '\P{^Is_IDST=_	True}', "");
Expect(0, 12276, '\p{Is_IDST=_	True}', "");
Expect(1, 12276, '\p{^Is_IDST=_	True}', "");
Expect(1, 12276, '\P{Is_IDST=_	True}', "");
Expect(0, 12276, '\P{^Is_IDST=_	True}', "");
Error('\p{Ideographic=:=_-no}');
Error('\P{Ideographic=:=_-no}');
Expect(1, 195102, '\p{Ideographic=no}', "");
Expect(0, 195102, '\p{^Ideographic=no}', "");
Expect(0, 195102, '\P{Ideographic=no}', "");
Expect(1, 195102, '\P{^Ideographic=no}', "");
Expect(0, 195101, '\p{Ideographic=no}', "");
Expect(1, 195101, '\p{^Ideographic=no}', "");
Expect(1, 195101, '\P{Ideographic=no}', "");
Expect(0, 195101, '\P{^Ideographic=no}', "");
Expect(1, 195102, '\p{Ideographic=__No}', "");
Expect(0, 195102, '\p{^Ideographic=__No}', "");
Expect(0, 195102, '\P{Ideographic=__No}', "");
Expect(1, 195102, '\P{^Ideographic=__No}', "");
Expect(0, 195101, '\p{Ideographic=__No}', "");
Expect(1, 195101, '\p{^Ideographic=__No}', "");
Expect(1, 195101, '\P{Ideographic=__No}', "");
Expect(0, 195101, '\P{^Ideographic=__No}', "");
Error('\p{Ideo=_N:=}');
Error('\P{Ideo=_N:=}');
Expect(1, 195102, '\p{Ideo=n}', "");
Expect(0, 195102, '\p{^Ideo=n}', "");
Expect(0, 195102, '\P{Ideo=n}', "");
Expect(1, 195102, '\P{^Ideo=n}', "");
Expect(0, 195101, '\p{Ideo=n}', "");
Expect(1, 195101, '\p{^Ideo=n}', "");
Expect(1, 195101, '\P{Ideo=n}', "");
Expect(0, 195101, '\P{^Ideo=n}', "");
Expect(1, 195102, '\p{Ideo=	N}', "");
Expect(0, 195102, '\p{^Ideo=	N}', "");
Expect(0, 195102, '\P{Ideo=	N}', "");
Expect(1, 195102, '\P{^Ideo=	N}', "");
Expect(0, 195101, '\p{Ideo=	N}', "");
Expect(1, 195101, '\p{^Ideo=	N}', "");
Expect(1, 195101, '\P{Ideo=	N}', "");
Expect(0, 195101, '\P{^Ideo=	N}', "");
Error('\p{Is_Ideographic=/a/F}');
Error('\P{Is_Ideographic=/a/F}');
Expect(1, 195102, '\p{Is_Ideographic=f}', "");
Expect(0, 195102, '\p{^Is_Ideographic=f}', "");
Expect(0, 195102, '\P{Is_Ideographic=f}', "");
Expect(1, 195102, '\P{^Is_Ideographic=f}', "");
Expect(0, 195101, '\p{Is_Ideographic=f}', "");
Expect(1, 195101, '\p{^Is_Ideographic=f}', "");
Expect(1, 195101, '\P{Is_Ideographic=f}', "");
Expect(0, 195101, '\P{^Is_Ideographic=f}', "");
Expect(1, 195102, '\p{Is_Ideographic=  F}', "");
Expect(0, 195102, '\p{^Is_Ideographic=  F}', "");
Expect(0, 195102, '\P{Is_Ideographic=  F}', "");
Expect(1, 195102, '\P{^Is_Ideographic=  F}', "");
Expect(0, 195101, '\p{Is_Ideographic=  F}', "");
Expect(1, 195101, '\p{^Is_Ideographic=  F}', "");
Expect(1, 195101, '\P{Is_Ideographic=  F}', "");
Expect(0, 195101, '\P{^Is_Ideographic=  F}', "");
Error('\p{Is_Ideo= /a/False}');
Error('\P{Is_Ideo= /a/False}');
Expect(1, 195102, '\p{Is_Ideo:   false}', "");
Expect(0, 195102, '\p{^Is_Ideo:   false}', "");
Expect(0, 195102, '\P{Is_Ideo:   false}', "");
Expect(1, 195102, '\P{^Is_Ideo:   false}', "");
Expect(0, 195101, '\p{Is_Ideo:   false}', "");
Expect(1, 195101, '\p{^Is_Ideo:   false}', "");
Expect(1, 195101, '\P{Is_Ideo:   false}', "");
Expect(0, 195101, '\P{^Is_Ideo:   false}', "");
Expect(1, 195102, '\p{Is_Ideo=  False}', "");
Expect(0, 195102, '\p{^Is_Ideo=  False}', "");
Expect(0, 195102, '\P{Is_Ideo=  False}', "");
Expect(1, 195102, '\P{^Is_Ideo=  False}', "");
Expect(0, 195101, '\p{Is_Ideo=  False}', "");
Expect(1, 195101, '\p{^Is_Ideo=  False}', "");
Expect(1, 195101, '\P{Is_Ideo=  False}', "");
Expect(0, 195101, '\P{^Is_Ideo=  False}', "");
Error('\p{Ideographic=	_Yes:=}');
Error('\P{Ideographic=	_Yes:=}');
Expect(1, 195101, '\p{Ideographic=yes}', "");
Expect(0, 195101, '\p{^Ideographic=yes}', "");
Expect(0, 195101, '\P{Ideographic=yes}', "");
Expect(1, 195101, '\P{^Ideographic=yes}', "");
Expect(0, 195102, '\p{Ideographic=yes}', "");
Expect(1, 195102, '\p{^Ideographic=yes}', "");
Expect(1, 195102, '\P{Ideographic=yes}', "");
Expect(0, 195102, '\P{^Ideographic=yes}', "");
Expect(1, 195101, '\p{Ideographic= _yes}', "");
Expect(0, 195101, '\p{^Ideographic= _yes}', "");
Expect(0, 195101, '\P{Ideographic= _yes}', "");
Expect(1, 195101, '\P{^Ideographic= _yes}', "");
Expect(0, 195102, '\p{Ideographic= _yes}', "");
Expect(1, 195102, '\p{^Ideographic= _yes}', "");
Expect(1, 195102, '\P{Ideographic= _yes}', "");
Expect(0, 195102, '\P{^Ideographic= _yes}', "");
Error('\p{Ideo=	/a/Y}');
Error('\P{Ideo=	/a/Y}');
Expect(1, 195101, '\p{Ideo=y}', "");
Expect(0, 195101, '\p{^Ideo=y}', "");
Expect(0, 195101, '\P{Ideo=y}', "");
Expect(1, 195101, '\P{^Ideo=y}', "");
Expect(0, 195102, '\p{Ideo=y}', "");
Expect(1, 195102, '\p{^Ideo=y}', "");
Expect(1, 195102, '\P{Ideo=y}', "");
Expect(0, 195102, '\P{^Ideo=y}', "");
Expect(1, 195101, '\p{Ideo=_-y}', "");
Expect(0, 195101, '\p{^Ideo=_-y}', "");
Expect(0, 195101, '\P{Ideo=_-y}', "");
Expect(1, 195101, '\P{^Ideo=_-y}', "");
Expect(0, 195102, '\p{Ideo=_-y}', "");
Expect(1, 195102, '\p{^Ideo=_-y}', "");
Expect(1, 195102, '\P{Ideo=_-y}', "");
Expect(0, 195102, '\P{^Ideo=_-y}', "");
Error('\p{Is_Ideographic= :=T}');
Error('\P{Is_Ideographic= :=T}');
Expect(1, 195101, '\p{Is_Ideographic=t}', "");
Expect(0, 195101, '\p{^Is_Ideographic=t}', "");
Expect(0, 195101, '\P{Is_Ideographic=t}', "");
Expect(1, 195101, '\P{^Is_Ideographic=t}', "");
Expect(0, 195102, '\p{Is_Ideographic=t}', "");
Expect(1, 195102, '\p{^Is_Ideographic=t}', "");
Expect(1, 195102, '\P{Is_Ideographic=t}', "");
Expect(0, 195102, '\P{^Is_Ideographic=t}', "");
Expect(1, 195101, '\p{Is_Ideographic=	 T}', "");
Expect(0, 195101, '\p{^Is_Ideographic=	 T}', "");
Expect(0, 195101, '\P{Is_Ideographic=	 T}', "");
Expect(1, 195101, '\P{^Is_Ideographic=	 T}', "");
Expect(0, 195102, '\p{Is_Ideographic=	 T}', "");
Expect(1, 195102, '\p{^Is_Ideographic=	 T}', "");
Expect(1, 195102, '\P{Is_Ideographic=	 T}', "");
Expect(0, 195102, '\P{^Is_Ideographic=	 T}', "");
Error('\p{Is_Ideo=/a/_True}');
Error('\P{Is_Ideo=/a/_True}');
Expect(1, 195101, '\p{Is_Ideo=true}', "");
Expect(0, 195101, '\p{^Is_Ideo=true}', "");
Expect(0, 195101, '\P{Is_Ideo=true}', "");
Expect(1, 195101, '\P{^Is_Ideo=true}', "");
Expect(0, 195102, '\p{Is_Ideo=true}', "");
Expect(1, 195102, '\p{^Is_Ideo=true}', "");
Expect(1, 195102, '\P{Is_Ideo=true}', "");
Expect(0, 195102, '\P{^Is_Ideo=true}', "");
Expect(1, 195101, '\p{Is_Ideo=  true}', "");
Expect(0, 195101, '\p{^Is_Ideo=  true}', "");
Expect(0, 195101, '\P{Is_Ideo=  true}', "");
Expect(1, 195101, '\P{^Is_Ideo=  true}', "");
Expect(0, 195102, '\p{Is_Ideo=  true}', "");
Expect(1, 195102, '\p{^Is_Ideo=  true}', "");
Expect(1, 195102, '\P{Is_Ideo=  true}', "");
Expect(0, 195102, '\P{^Is_Ideo=  true}', "");
Error('\p{Present_In=- +0_0_0_0_0_0_001.1/a/}');
Error('\P{Present_In=- +0_0_0_0_0_0_001.1/a/}');
Expect(1, 65533, '\p{Present_In=0000000001.1}', "");
Expect(0, 65533, '\p{^Present_In=0000000001.1}', "");
Expect(0, 65533, '\P{Present_In=0000000001.1}', "");
Expect(1, 65533, '\P{^Present_In=0000000001.1}', "");
Expect(0, 65536, '\p{Present_In=0000000001.1}', "");
Expect(1, 65536, '\p{^Present_In=0000000001.1}', "");
Expect(1, 65536, '\P{Present_In=0000000001.1}', "");
Expect(0, 65536, '\P{^Present_In=0000000001.1}', "");
Error('\p{In=:=-+0_0_01.1}');
Error('\P{In=:=-+0_0_01.1}');
Expect(1, 65533, '\p{In=00000001.1}', "");
Expect(0, 65533, '\p{^In=00000001.1}', "");
Expect(0, 65533, '\P{In=00000001.1}', "");
Expect(1, 65533, '\P{^In=00000001.1}', "");
Expect(0, 65536, '\p{In=00000001.1}', "");
Expect(1, 65536, '\p{^In=00000001.1}', "");
Expect(1, 65536, '\P{In=00000001.1}', "");
Expect(0, 65536, '\P{^In=00000001.1}', "");
Error('\p{Is_Present_In=/a/__000000001.1}');
Error('\P{Is_Present_In=/a/__000000001.1}');
Expect(1, 65533, '\p{Is_Present_In=000000001.1}', "");
Expect(0, 65533, '\p{^Is_Present_In=000000001.1}', "");
Expect(0, 65533, '\P{Is_Present_In=000000001.1}', "");
Expect(1, 65533, '\P{^Is_Present_In=000000001.1}', "");
Expect(0, 65536, '\p{Is_Present_In=000000001.1}', "");
Expect(1, 65536, '\p{^Is_Present_In=000000001.1}', "");
Expect(1, 65536, '\P{Is_Present_In=000000001.1}', "");
Expect(0, 65536, '\P{^Is_Present_In=000000001.1}', "");
Error('\p{Is_In= +00000001.1/a/}');
Error('\P{Is_In= +00000001.1/a/}');
Expect(1, 65533, '\p{Is_In=00_00_1.1}', "");
Expect(0, 65533, '\p{^Is_In=00_00_1.1}', "");
Expect(0, 65533, '\P{Is_In=00_00_1.1}', "");
Expect(1, 65533, '\P{^Is_In=00_00_1.1}', "");
Expect(0, 65536, '\p{Is_In=00_00_1.1}', "");
Expect(1, 65536, '\p{^Is_In=00_00_1.1}', "");
Expect(1, 65536, '\P{Is_In=00_00_1.1}', "");
Expect(0, 65536, '\P{^Is_In=00_00_1.1}', "");
Error('\p{Present_In=_/a/+2.0}');
Error('\P{Present_In=_/a/+2.0}');
Expect(1, 983040, '\p{Present_In:000000_2.0}', "");
Expect(0, 983040, '\p{^Present_In:000000_2.0}', "");
Expect(0, 983040, '\P{Present_In:000000_2.0}', "");
Expect(1, 983040, '\P{^Present_In:000000_2.0}', "");
Expect(0, 983037, '\p{Present_In:000000_2.0}', "");
Expect(1, 983037, '\p{^Present_In:000000_2.0}', "");
Expect(1, 983037, '\P{Present_In:000000_2.0}', "");
Expect(0, 983037, '\P{^Present_In:000000_2.0}', "");
Error('\p{In=:=-_2.0}');
Error('\P{In=:=-_2.0}');
Expect(1, 983040, '\p{In=000000002.0}', "");
Expect(0, 983040, '\p{^In=000000002.0}', "");
Expect(0, 983040, '\P{In=000000002.0}', "");
Expect(1, 983040, '\P{^In=000000002.0}', "");
Expect(0, 983037, '\p{In=000000002.0}', "");
Expect(1, 983037, '\p{^In=000000002.0}', "");
Expect(1, 983037, '\P{In=000000002.0}', "");
Expect(0, 983037, '\P{^In=000000002.0}', "");
Error('\p{Is_Present_In=	/a/0000002.0}');
Error('\P{Is_Present_In=	/a/0000002.0}');
Expect(1, 983040, '\p{Is_Present_In=+000000002.0}', "");
Expect(0, 983040, '\p{^Is_Present_In=+000000002.0}', "");
Expect(0, 983040, '\P{Is_Present_In=+000000002.0}', "");
Expect(1, 983040, '\P{^Is_Present_In=+000000002.0}', "");
Expect(0, 983037, '\p{Is_Present_In=+000000002.0}', "");
Expect(1, 983037, '\p{^Is_Present_In=+000000002.0}', "");
Expect(1, 983037, '\P{Is_Present_In=+000000002.0}', "");
Expect(0, 983037, '\P{^Is_Present_In=+000000002.0}', "");
Error('\p{Is_In=/a/-	00_00_02.0}');
Error('\P{Is_In=/a/-	00_00_02.0}');
Expect(1, 983040, '\p{Is_In=2.0}', "");
Expect(0, 983040, '\p{^Is_In=2.0}', "");
Expect(0, 983040, '\P{Is_In=2.0}', "");
Expect(1, 983040, '\P{^Is_In=2.0}', "");
Expect(0, 983037, '\p{Is_In=2.0}', "");
Expect(1, 983037, '\p{^Is_In=2.0}', "");
Expect(1, 983037, '\P{Is_In=2.0}', "");
Expect(0, 983037, '\P{^Is_In=2.0}', "");
Error('\p{Present_In=-0002.1:=}');
Error('\P{Present_In=-0002.1:=}');
Expect(1, 983040, '\p{Present_In=0_2.1}', "");
Expect(0, 983040, '\p{^Present_In=0_2.1}', "");
Expect(0, 983040, '\P{Present_In=0_2.1}', "");
Expect(1, 983040, '\P{^Present_In=0_2.1}', "");
Expect(0, 983037, '\p{Present_In=0_2.1}', "");
Expect(1, 983037, '\p{^Present_In=0_2.1}', "");
Expect(1, 983037, '\P{Present_In=0_2.1}', "");
Expect(0, 983037, '\P{^Present_In=0_2.1}', "");
Error('\p{In: 00000002.1:=}');
Error('\P{In: 00000002.1:=}');
Expect(1, 983040, '\p{In=00000002.1}', "");
Expect(0, 983040, '\p{^In=00000002.1}', "");
Expect(0, 983040, '\P{In=00000002.1}', "");
Expect(1, 983040, '\P{^In=00000002.1}', "");
Expect(0, 983037, '\p{In=00000002.1}', "");
Expect(1, 983037, '\p{^In=00000002.1}', "");
Expect(1, 983037, '\P{In=00000002.1}', "");
Expect(0, 983037, '\P{^In=00000002.1}', "");
Error('\p{Is_Present_In:	:=_-+002.1}');
Error('\P{Is_Present_In:	:=_-+002.1}');
Expect(1, 983040, '\p{Is_Present_In=+002.1}', "");
Expect(0, 983040, '\p{^Is_Present_In=+002.1}', "");
Expect(0, 983040, '\P{Is_Present_In=+002.1}', "");
Expect(1, 983040, '\P{^Is_Present_In=+002.1}', "");
Expect(0, 983037, '\p{Is_Present_In=+002.1}', "");
Expect(1, 983037, '\p{^Is_Present_In=+002.1}', "");
Expect(1, 983037, '\P{Is_Present_In=+002.1}', "");
Expect(0, 983037, '\P{^Is_Present_In=+002.1}', "");
Error('\p{Is_In=__+000_000_000_2.1/a/}');
Error('\P{Is_In=__+000_000_000_2.1/a/}');
Expect(1, 983040, '\p{Is_In=+00002.1}', "");
Expect(0, 983040, '\p{^Is_In=+00002.1}', "");
Expect(0, 983040, '\P{Is_In=+00002.1}', "");
Expect(1, 983040, '\P{^Is_In=+00002.1}', "");
Expect(0, 983037, '\p{Is_In=+00002.1}', "");
Expect(1, 983037, '\p{^Is_In=+00002.1}', "");
Expect(1, 983037, '\P{Is_In=+00002.1}', "");
Expect(0, 983037, '\P{^Is_In=+00002.1}', "");
Error('\p{Present_In=-:=+00_3.0}');
Error('\P{Present_In=-:=+00_3.0}');
Expect(1, 983040, '\p{Present_In=0000003.0}', "");
Expect(0, 983040, '\p{^Present_In=0000003.0}', "");
Expect(0, 983040, '\P{Present_In=0000003.0}', "");
Expect(1, 983040, '\P{^Present_In=0000003.0}', "");
Expect(0, 983037, '\p{Present_In=0000003.0}', "");
Expect(1, 983037, '\p{^Present_In=0000003.0}', "");
Expect(1, 983037, '\P{Present_In=0000003.0}', "");
Expect(0, 983037, '\P{^Present_In=0000003.0}', "");
Error('\p{In= /a/03.0}');
Error('\P{In= /a/03.0}');
Expect(1, 983040, '\p{In=0000003.0}', "");
Expect(0, 983040, '\p{^In=0000003.0}', "");
Expect(0, 983040, '\P{In=0000003.0}', "");
Expect(1, 983040, '\P{^In=0000003.0}', "");
Expect(0, 983037, '\p{In=0000003.0}', "");
Expect(1, 983037, '\p{^In=0000003.0}', "");
Expect(1, 983037, '\P{In=0000003.0}', "");
Expect(0, 983037, '\P{^In=0000003.0}', "");
Error('\p{Is_Present_In=	/a/0_0_03.0}');
Error('\P{Is_Present_In=	/a/0_0_03.0}');
Expect(1, 983040, '\p{Is_Present_In=00003.0}', "");
Expect(0, 983040, '\p{^Is_Present_In=00003.0}', "");
Expect(0, 983040, '\P{Is_Present_In=00003.0}', "");
Expect(1, 983040, '\P{^Is_Present_In=00003.0}', "");
Expect(0, 983037, '\p{Is_Present_In=00003.0}', "");
Expect(1, 983037, '\p{^Is_Present_In=00003.0}', "");
Expect(1, 983037, '\P{Is_Present_In=00003.0}', "");
Expect(0, 983037, '\P{^Is_Present_In=00003.0}', "");
Error('\p{Is_In=	:=003.0}');
Error('\P{Is_In=	:=003.0}');
Expect(1, 983040, '\p{Is_In=00003.0}', "");
Expect(0, 983040, '\p{^Is_In=00003.0}', "");
Expect(0, 983040, '\P{Is_In=00003.0}', "");
Expect(1, 983040, '\P{^Is_In=00003.0}', "");
Expect(0, 983037, '\p{Is_In=00003.0}', "");
Expect(1, 983037, '\p{^Is_In=00003.0}', "");
Expect(1, 983037, '\P{Is_In=00003.0}', "");
Expect(0, 983037, '\P{^Is_In=00003.0}', "");
Error('\p{Present_In:	-	00_00_00_3.1/a/}');
Error('\P{Present_In:	-	00_00_00_3.1/a/}');
Expect(1, 983040, '\p{Present_In=00_3.1}', "");
Expect(0, 983040, '\p{^Present_In=00_3.1}', "");
Expect(0, 983040, '\P{Present_In=00_3.1}', "");
Expect(1, 983040, '\P{^Present_In=00_3.1}', "");
Expect(0, 983037, '\p{Present_In=00_3.1}', "");
Expect(1, 983037, '\p{^Present_In=00_3.1}', "");
Expect(1, 983037, '\P{Present_In=00_3.1}', "");
Expect(0, 983037, '\P{^Present_In=00_3.1}', "");
Error('\p{In=0003.1:=}');
Error('\P{In=0003.1:=}');
Expect(1, 983040, '\p{In=0_0_03.1}', "");
Expect(0, 983040, '\p{^In=0_0_03.1}', "");
Expect(0, 983040, '\P{In=0_0_03.1}', "");
Expect(1, 983040, '\P{^In=0_0_03.1}', "");
Expect(0, 983037, '\p{In=0_0_03.1}', "");
Expect(1, 983037, '\p{^In=0_0_03.1}', "");
Expect(1, 983037, '\P{In=0_0_03.1}', "");
Expect(0, 983037, '\P{^In=0_0_03.1}', "");
Error('\p{Is_Present_In= 	0_3.1:=}');
Error('\P{Is_Present_In= 	0_3.1:=}');
Expect(1, 983040, '\p{Is_Present_In=+00000003.1}', "");
Expect(0, 983040, '\p{^Is_Present_In=+00000003.1}', "");
Expect(0, 983040, '\P{Is_Present_In=+00000003.1}', "");
Expect(1, 983040, '\P{^Is_Present_In=+00000003.1}', "");
Expect(0, 983037, '\p{Is_Present_In=+00000003.1}', "");
Expect(1, 983037, '\p{^Is_Present_In=+00000003.1}', "");
Expect(1, 983037, '\P{Is_Present_In=+00000003.1}', "");
Expect(0, 983037, '\P{^Is_Present_In=+00000003.1}', "");
Error('\p{Is_In=	_0000_3.1/a/}');
Error('\P{Is_In=	_0000_3.1/a/}');
Expect(1, 983040, '\p{Is_In=+00_3.1}', "");
Expect(0, 983040, '\p{^Is_In=+00_3.1}', "");
Expect(0, 983040, '\P{Is_In=+00_3.1}', "");
Expect(1, 983040, '\P{^Is_In=+00_3.1}', "");
Expect(0, 983037, '\p{Is_In=+00_3.1}', "");
Expect(1, 983037, '\p{^Is_In=+00_3.1}', "");
Expect(1, 983037, '\P{Is_In=+00_3.1}', "");
Expect(0, 983037, '\P{^Is_In=+00_3.1}', "");
Error('\p{Present_In= :=+0000003.2}');
Error('\P{Present_In= :=+0000003.2}');
Expect(1, 983040, '\p{Present_In=+0000003.2}', "");
Expect(0, 983040, '\p{^Present_In=+0000003.2}', "");
Expect(0, 983040, '\P{Present_In=+0000003.2}', "");
Expect(1, 983040, '\P{^Present_In=+0000003.2}', "");
Expect(0, 983037, '\p{Present_In=+0000003.2}', "");
Expect(1, 983037, '\p{^Present_In=+0000003.2}', "");
Expect(1, 983037, '\P{Present_In=+0000003.2}', "");
Expect(0, 983037, '\P{^Present_In=+0000003.2}', "");
Error('\p{In=-+0000000003.2/a/}');
Error('\P{In=-+0000000003.2/a/}');
Expect(1, 983040, '\p{In=+00000000_3.2}', "");
Expect(0, 983040, '\p{^In=+00000000_3.2}', "");
Expect(0, 983040, '\P{In=+00000000_3.2}', "");
Expect(1, 983040, '\P{^In=+00000000_3.2}', "");
Expect(0, 983037, '\p{In=+00000000_3.2}', "");
Expect(1, 983037, '\p{^In=+00000000_3.2}', "");
Expect(1, 983037, '\P{In=+00000000_3.2}', "");
Expect(0, 983037, '\P{^In=+00000000_3.2}', "");
Error('\p{Is_Present_In= 	0_0_03.2/a/}');
Error('\P{Is_Present_In= 	0_0_03.2/a/}');
Expect(1, 983040, '\p{Is_Present_In=000000003.2}', "");
Expect(0, 983040, '\p{^Is_Present_In=000000003.2}', "");
Expect(0, 983040, '\P{Is_Present_In=000000003.2}', "");
Expect(1, 983040, '\P{^Is_Present_In=000000003.2}', "");
Expect(0, 983037, '\p{Is_Present_In=000000003.2}', "");
Expect(1, 983037, '\p{^Is_Present_In=000000003.2}', "");
Expect(1, 983037, '\P{Is_Present_In=000000003.2}', "");
Expect(0, 983037, '\P{^Is_Present_In=000000003.2}', "");
Error('\p{Is_In= :=+0_0_0_0_03.2}');
Error('\P{Is_In= :=+0_0_0_0_03.2}');
Expect(1, 983040, '\p{Is_In=+000003.2}', "");
Expect(0, 983040, '\p{^Is_In=+000003.2}', "");
Expect(0, 983040, '\P{Is_In=+000003.2}', "");
Expect(1, 983040, '\P{^Is_In=+000003.2}', "");
Expect(0, 983037, '\p{Is_In=+000003.2}', "");
Expect(1, 983037, '\p{^Is_In=+000003.2}', "");
Expect(1, 983037, '\P{Is_In=+000003.2}', "");
Expect(0, 983037, '\P{^Is_In=+000003.2}', "");
Error('\p{Present_In=	_4.0:=}');
Error('\P{Present_In=	_4.0:=}');
Expect(1, 983040, '\p{Present_In=04.0}', "");
Expect(0, 983040, '\p{^Present_In=04.0}', "");
Expect(0, 983040, '\P{Present_In=04.0}', "");
Expect(1, 983040, '\P{^Present_In=04.0}', "");
Expect(0, 983037, '\p{Present_In=04.0}', "");
Expect(1, 983037, '\p{^Present_In=04.0}', "");
Expect(1, 983037, '\P{Present_In=04.0}', "");
Expect(0, 983037, '\P{^Present_In=04.0}', "");
Error('\p{In= _0000000004.0:=}');
Error('\P{In= _0000000004.0:=}');
Expect(1, 983040, '\p{In: 000_000_4.0}', "");
Expect(0, 983040, '\p{^In: 000_000_4.0}', "");
Expect(0, 983040, '\P{In: 000_000_4.0}', "");
Expect(1, 983040, '\P{^In: 000_000_4.0}', "");
Expect(0, 983037, '\p{In: 000_000_4.0}', "");
Expect(1, 983037, '\p{^In: 000_000_4.0}', "");
Expect(1, 983037, '\P{In: 000_000_4.0}', "");
Expect(0, 983037, '\P{^In: 000_000_4.0}', "");
Error('\p{Is_Present_In=:=-	0_0_0_0_0_0_004.0}');
Error('\P{Is_Present_In=:=-	0_0_0_0_0_0_004.0}');
Expect(1, 983040, '\p{Is_Present_In:0_0_0_0_004.0}', "");
Expect(0, 983040, '\p{^Is_Present_In:0_0_0_0_004.0}', "");
Expect(0, 983040, '\P{Is_Present_In:0_0_0_0_004.0}', "");
Expect(1, 983040, '\P{^Is_Present_In:0_0_0_0_004.0}', "");
Expect(0, 983037, '\p{Is_Present_In:0_0_0_0_004.0}', "");
Expect(1, 983037, '\p{^Is_Present_In:0_0_0_0_004.0}', "");
Expect(1, 983037, '\P{Is_Present_In:0_0_0_0_004.0}', "");
Expect(0, 983037, '\P{^Is_Present_In:0_0_0_0_004.0}', "");
Error('\p{Is_In=-:=4.0}');
Error('\P{Is_In=-:=4.0}');
Expect(1, 983040, '\p{Is_In=000000004.0}', "");
Expect(0, 983040, '\p{^Is_In=000000004.0}', "");
Expect(0, 983040, '\P{Is_In=000000004.0}', "");
Expect(1, 983040, '\P{^Is_In=000000004.0}', "");
Expect(0, 983037, '\p{Is_In=000000004.0}', "");
Expect(1, 983037, '\p{^Is_In=000000004.0}', "");
Expect(1, 983037, '\P{Is_In=000000004.0}', "");
Expect(0, 983037, '\P{^Is_In=000000004.0}', "");
Error('\p{Present_In:  :=04.1}');
Error('\P{Present_In:  :=04.1}');
Expect(1, 983040, '\p{Present_In=4.1}', "");
Expect(0, 983040, '\p{^Present_In=4.1}', "");
Expect(0, 983040, '\P{Present_In=4.1}', "");
Expect(1, 983040, '\P{^Present_In=4.1}', "");
Expect(0, 983037, '\p{Present_In=4.1}', "");
Expect(1, 983037, '\p{^Present_In=4.1}', "");
Expect(1, 983037, '\P{Present_In=4.1}', "");
Expect(0, 983037, '\P{^Present_In=4.1}', "");
Error('\p{In=_-0004.1/a/}');
Error('\P{In=_-0004.1/a/}');
Expect(1, 983040, '\p{In=004.1}', "");
Expect(0, 983040, '\p{^In=004.1}', "");
Expect(0, 983040, '\P{In=004.1}', "");
Expect(1, 983040, '\P{^In=004.1}', "");
Expect(0, 983037, '\p{In=004.1}', "");
Expect(1, 983037, '\p{^In=004.1}', "");
Expect(1, 983037, '\P{In=004.1}', "");
Expect(0, 983037, '\P{^In=004.1}', "");
Error('\p{Is_Present_In=:=_0_0_04.1}');
Error('\P{Is_Present_In=:=_0_0_04.1}');
Expect(1, 983040, '\p{Is_Present_In=+0000004.1}', "");
Expect(0, 983040, '\p{^Is_Present_In=+0000004.1}', "");
Expect(0, 983040, '\P{Is_Present_In=+0000004.1}', "");
Expect(1, 983040, '\P{^Is_Present_In=+0000004.1}', "");
Expect(0, 983037, '\p{Is_Present_In=+0000004.1}', "");
Expect(1, 983037, '\p{^Is_Present_In=+0000004.1}', "");
Expect(1, 983037, '\P{Is_Present_In=+0000004.1}', "");
Expect(0, 983037, '\P{^Is_Present_In=+0000004.1}', "");
Error('\p{Is_In=:=+000000000_4.1}');
Error('\P{Is_In=:=+000000000_4.1}');
Expect(1, 983040, '\p{Is_In=0000004.1}', "");
Expect(0, 983040, '\p{^Is_In=0000004.1}', "");
Expect(0, 983040, '\P{Is_In=0000004.1}', "");
Expect(1, 983040, '\P{^Is_In=0000004.1}', "");
Expect(0, 983037, '\p{Is_In=0000004.1}', "");
Expect(1, 983037, '\p{^Is_In=0000004.1}', "");
Expect(1, 983037, '\P{Is_In=0000004.1}', "");
Expect(0, 983037, '\P{^Is_In=0000004.1}', "");
Error('\p{Present_In= /a/5.0}');
Error('\P{Present_In= /a/5.0}');
Expect(1, 983040, '\p{Present_In=+005.0}', "");
Expect(0, 983040, '\p{^Present_In=+005.0}', "");
Expect(0, 983040, '\P{Present_In=+005.0}', "");
Expect(1, 983040, '\P{^Present_In=+005.0}', "");
Expect(0, 983037, '\p{Present_In=+005.0}', "");
Expect(1, 983037, '\p{^Present_In=+005.0}', "");
Expect(1, 983037, '\P{Present_In=+005.0}', "");
Expect(0, 983037, '\P{^Present_In=+005.0}', "");
Error('\p{In= :=+0005.0}');
Error('\P{In= :=+0005.0}');
Expect(1, 983040, '\p{In=+000_000_5.0}', "");
Expect(0, 983040, '\p{^In=+000_000_5.0}', "");
Expect(0, 983040, '\P{In=+000_000_5.0}', "");
Expect(1, 983040, '\P{^In=+000_000_5.0}', "");
Expect(0, 983037, '\p{In=+000_000_5.0}', "");
Expect(1, 983037, '\p{^In=+000_000_5.0}', "");
Expect(1, 983037, '\P{In=+000_000_5.0}', "");
Expect(0, 983037, '\P{^In=+000_000_5.0}', "");
Error('\p{Is_Present_In=	+005.0/a/}');
Error('\P{Is_Present_In=	+005.0/a/}');
Expect(1, 983040, '\p{Is_Present_In:   00_00_00_5.0}', "");
Expect(0, 983040, '\p{^Is_Present_In:   00_00_00_5.0}', "");
Expect(0, 983040, '\P{Is_Present_In:   00_00_00_5.0}', "");
Expect(1, 983040, '\P{^Is_Present_In:   00_00_00_5.0}', "");
Expect(0, 983037, '\p{Is_Present_In:   00_00_00_5.0}', "");
Expect(1, 983037, '\p{^Is_Present_In:   00_00_00_5.0}', "");
Expect(1, 983037, '\P{Is_Present_In:   00_00_00_5.0}', "");
Expect(0, 983037, '\P{^Is_Present_In:   00_00_00_5.0}', "");
Error('\p{Is_In:	 0005.0:=}');
Error('\P{Is_In:	 0005.0:=}');
Expect(1, 983040, '\p{Is_In=+00005.0}', "");
Expect(0, 983040, '\p{^Is_In=+00005.0}', "");
Expect(0, 983040, '\P{Is_In=+00005.0}', "");
Expect(1, 983040, '\P{^Is_In=+00005.0}', "");
Expect(0, 983037, '\p{Is_In=+00005.0}', "");
Expect(1, 983037, '\p{^Is_In=+00005.0}', "");
Expect(1, 983037, '\P{Is_In=+00005.0}', "");
Expect(0, 983037, '\P{^Is_In=+00005.0}', "");
Error('\p{Present_In:	/a/ +00000005.1}');
Error('\P{Present_In:	/a/ +00000005.1}');
Expect(1, 983040, '\p{Present_In=00_5.1}', "");
Expect(0, 983040, '\p{^Present_In=00_5.1}', "");
Expect(0, 983040, '\P{Present_In=00_5.1}', "");
Expect(1, 983040, '\P{^Present_In=00_5.1}', "");
Expect(0, 983037, '\p{Present_In=00_5.1}', "");
Expect(1, 983037, '\p{^Present_In=00_5.1}', "");
Expect(1, 983037, '\P{Present_In=00_5.1}', "");
Expect(0, 983037, '\P{^Present_In=00_5.1}', "");
Error('\p{In=	:=0000000005.1}');
Error('\P{In=	:=0000000005.1}');
Expect(1, 983040, '\p{In=00_5.1}', "");
Expect(0, 983040, '\p{^In=00_5.1}', "");
Expect(0, 983040, '\P{In=00_5.1}', "");
Expect(1, 983040, '\P{^In=00_5.1}', "");
Expect(0, 983037, '\p{In=00_5.1}', "");
Expect(1, 983037, '\p{^In=00_5.1}', "");
Expect(1, 983037, '\P{In=00_5.1}', "");
Expect(0, 983037, '\P{^In=00_5.1}', "");
Error('\p{Is_Present_In=_-+000_000_000_5.1:=}');
Error('\P{Is_Present_In=_-+000_000_000_5.1:=}');
Expect(1, 983040, '\p{Is_Present_In=00000_5.1}', "");
Expect(0, 983040, '\p{^Is_Present_In=00000_5.1}', "");
Expect(0, 983040, '\P{Is_Present_In=00000_5.1}', "");
Expect(1, 983040, '\P{^Is_Present_In=00000_5.1}', "");
Expect(0, 983037, '\p{Is_Present_In=00000_5.1}', "");
Expect(1, 983037, '\p{^Is_Present_In=00000_5.1}', "");
Expect(1, 983037, '\P{Is_Present_In=00000_5.1}', "");
Expect(0, 983037, '\P{^Is_Present_In=00000_5.1}', "");
Error('\p{Is_In=-/a/000_5.1}');
Error('\P{Is_In=-/a/000_5.1}');
Expect(1, 983040, '\p{Is_In=000005.1}', "");
Expect(0, 983040, '\p{^Is_In=000005.1}', "");
Expect(0, 983040, '\P{Is_In=000005.1}', "");
Expect(1, 983040, '\P{^Is_In=000005.1}', "");
Expect(0, 983037, '\p{Is_In=000005.1}', "");
Expect(1, 983037, '\p{^Is_In=000005.1}', "");
Expect(1, 983037, '\P{Is_In=000005.1}', "");
Expect(0, 983037, '\P{^Is_In=000005.1}', "");
Error('\p{Present_In:/a/0_5.2}');
Error('\P{Present_In:/a/0_5.2}');
Expect(1, 983040, '\p{Present_In=0000005.2}', "");
Expect(0, 983040, '\p{^Present_In=0000005.2}', "");
Expect(0, 983040, '\P{Present_In=0000005.2}', "");
Expect(1, 983040, '\P{^Present_In=0000005.2}', "");
Expect(0, 983037, '\p{Present_In=0000005.2}', "");
Expect(1, 983037, '\p{^Present_In=0000005.2}', "");
Expect(1, 983037, '\P{Present_In=0000005.2}', "");
Expect(0, 983037, '\P{^Present_In=0000005.2}', "");
Error('\p{In=_	5.2:=}');
Error('\P{In=_	5.2:=}');
Expect(1, 983040, '\p{In=000000005.2}', "");
Expect(0, 983040, '\p{^In=000000005.2}', "");
Expect(0, 983040, '\P{In=000000005.2}', "");
Expect(1, 983040, '\P{^In=000000005.2}', "");
Expect(0, 983037, '\p{In=000000005.2}', "");
Expect(1, 983037, '\p{^In=000000005.2}', "");
Expect(1, 983037, '\P{In=000000005.2}', "");
Expect(0, 983037, '\P{^In=000000005.2}', "");
Error('\p{Is_Present_In=:=	000005.2}');
Error('\P{Is_Present_In=:=	000005.2}');
Expect(1, 983040, '\p{Is_Present_In=0_0_05.2}', "");
Expect(0, 983040, '\p{^Is_Present_In=0_0_05.2}', "");
Expect(0, 983040, '\P{Is_Present_In=0_0_05.2}', "");
Expect(1, 983040, '\P{^Is_Present_In=0_0_05.2}', "");
Expect(0, 983037, '\p{Is_Present_In=0_0_05.2}', "");
Expect(1, 983037, '\p{^Is_Present_In=0_0_05.2}', "");
Expect(1, 983037, '\P{Is_Present_In=0_0_05.2}', "");
Expect(0, 983037, '\P{^Is_Present_In=0_0_05.2}', "");
Error('\p{Is_In=:= 	000_000_5.2}');
Error('\P{Is_In=:= 	000_000_5.2}');
Expect(1, 983040, '\p{Is_In=+0_0_0_0_0005.2}', "");
Expect(0, 983040, '\p{^Is_In=+0_0_0_0_0005.2}', "");
Expect(0, 983040, '\P{Is_In=+0_0_0_0_0005.2}', "");
Expect(1, 983040, '\P{^Is_In=+0_0_0_0_0005.2}', "");
Expect(0, 983037, '\p{Is_In=+0_0_0_0_0005.2}', "");
Expect(1, 983037, '\p{^Is_In=+0_0_0_0_0005.2}', "");
Expect(1, 983037, '\P{Is_In=+0_0_0_0_0005.2}', "");
Expect(0, 983037, '\P{^Is_In=+0_0_0_0_0005.2}', "");
Error('\p{Present_In=-006.0:=}');
Error('\P{Present_In=-006.0:=}');
Expect(1, 983040, '\p{Present_In=+000006.0}', "");
Expect(0, 983040, '\p{^Present_In=+000006.0}', "");
Expect(0, 983040, '\P{Present_In=+000006.0}', "");
Expect(1, 983040, '\P{^Present_In=+000006.0}', "");
Expect(0, 983037, '\p{Present_In=+000006.0}', "");
Expect(1, 983037, '\p{^Present_In=+000006.0}', "");
Expect(1, 983037, '\P{Present_In=+000006.0}', "");
Expect(0, 983037, '\P{^Present_In=+000006.0}', "");
Error('\p{In=	:=00_00_06.0}');
Error('\P{In=	:=00_00_06.0}');
Expect(1, 983040, '\p{In=06.0}', "");
Expect(0, 983040, '\p{^In=06.0}', "");
Expect(0, 983040, '\P{In=06.0}', "");
Expect(1, 983040, '\P{^In=06.0}', "");
Expect(0, 983037, '\p{In=06.0}', "");
Expect(1, 983037, '\p{^In=06.0}', "");
Expect(1, 983037, '\P{In=06.0}', "");
Expect(0, 983037, '\P{^In=06.0}', "");
Error('\p{Is_Present_In=/a/ -0_0_0_0_0_0_0006.0}');
Error('\P{Is_Present_In=/a/ -0_0_0_0_0_0_0006.0}');
Expect(1, 983040, '\p{Is_Present_In=+00_00_00_6.0}', "");
Expect(0, 983040, '\p{^Is_Present_In=+00_00_00_6.0}', "");
Expect(0, 983040, '\P{Is_Present_In=+00_00_00_6.0}', "");
Expect(1, 983040, '\P{^Is_Present_In=+00_00_00_6.0}', "");
Expect(0, 983037, '\p{Is_Present_In=+00_00_00_6.0}', "");
Expect(1, 983037, '\p{^Is_Present_In=+00_00_00_6.0}', "");
Expect(1, 983037, '\P{Is_Present_In=+00_00_00_6.0}', "");
Expect(0, 983037, '\P{^Is_Present_In=+00_00_00_6.0}', "");
Error('\p{Is_In=_/a/+0006.0}');
Error('\P{Is_In=_/a/+0006.0}');
Expect(1, 983040, '\p{Is_In=000000006.0}', "");
Expect(0, 983040, '\p{^Is_In=000000006.0}', "");
Expect(0, 983040, '\P{Is_In=000000006.0}', "");
Expect(1, 983040, '\P{^Is_In=000000006.0}', "");
Expect(0, 983037, '\p{Is_In=000000006.0}', "");
Expect(1, 983037, '\p{^Is_In=000000006.0}', "");
Expect(1, 983037, '\P{Is_In=000000006.0}', "");
Expect(0, 983037, '\P{^Is_In=000000006.0}', "");
Error('\p{Present_In= /a/+0_0_0_006.1}');
Error('\P{Present_In= /a/+0_0_0_006.1}');
Expect(1, 983040, '\p{Present_In=+0_6.1}', "");
Expect(0, 983040, '\p{^Present_In=+0_6.1}', "");
Expect(0, 983040, '\P{Present_In=+0_6.1}', "");
Expect(1, 983040, '\P{^Present_In=+0_6.1}', "");
Expect(0, 983037, '\p{Present_In=+0_6.1}', "");
Expect(1, 983037, '\p{^Present_In=+0_6.1}', "");
Expect(1, 983037, '\P{Present_In=+0_6.1}', "");
Expect(0, 983037, '\P{^Present_In=+0_6.1}', "");
Error('\p{In=/a/+00_00_00_006.1}');
Error('\P{In=/a/+00_00_00_006.1}');
Expect(1, 983040, '\p{In=6.1}', "");
Expect(0, 983040, '\p{^In=6.1}', "");
Expect(0, 983040, '\P{In=6.1}', "");
Expect(1, 983040, '\P{^In=6.1}', "");
Expect(0, 983037, '\p{In=6.1}', "");
Expect(1, 983037, '\p{^In=6.1}', "");
Expect(1, 983037, '\P{In=6.1}', "");
Expect(0, 983037, '\P{^In=6.1}', "");
Error('\p{Is_Present_In=:=000000006.1}');
Error('\P{Is_Present_In=:=000000006.1}');
Expect(1, 983040, '\p{Is_Present_In=000000006.1}', "");
Expect(0, 983040, '\p{^Is_Present_In=000000006.1}', "");
Expect(0, 983040, '\P{Is_Present_In=000000006.1}', "");
Expect(1, 983040, '\P{^Is_Present_In=000000006.1}', "");
Expect(0, 983037, '\p{Is_Present_In=000000006.1}', "");
Expect(1, 983037, '\p{^Is_Present_In=000000006.1}', "");
Expect(1, 983037, '\P{Is_Present_In=000000006.1}', "");
Expect(0, 983037, '\P{^Is_Present_In=000000006.1}', "");
Error('\p{Is_In=/a/ 	000000000_6.1}');
Error('\P{Is_In=/a/ 	000000000_6.1}');
Expect(1, 983040, '\p{Is_In:	00_00_00_06.1}', "");
Expect(0, 983040, '\p{^Is_In:	00_00_00_06.1}', "");
Expect(0, 983040, '\P{Is_In:	00_00_00_06.1}', "");
Expect(1, 983040, '\P{^Is_In:	00_00_00_06.1}', "");
Expect(0, 983037, '\p{Is_In:	00_00_00_06.1}', "");
Expect(1, 983037, '\p{^Is_In:	00_00_00_06.1}', "");
Expect(1, 983037, '\P{Is_In:	00_00_00_06.1}', "");
Expect(0, 983037, '\P{^Is_In:	00_00_00_06.1}', "");
Error('\p{Present_In=_:=00000006.2}');
Error('\P{Present_In=_:=00000006.2}');
Expect(1, 983040, '\p{Present_In=6.2}', "");
Expect(0, 983040, '\p{^Present_In=6.2}', "");
Expect(0, 983040, '\P{Present_In=6.2}', "");
Expect(1, 983040, '\P{^Present_In=6.2}', "");
Expect(0, 983037, '\p{Present_In=6.2}', "");
Expect(1, 983037, '\p{^Present_In=6.2}', "");
Expect(1, 983037, '\P{Present_In=6.2}', "");
Expect(0, 983037, '\P{^Present_In=6.2}', "");
Error('\p{In=:=	 000000006.2}');
Error('\P{In=:=	 000000006.2}');
Expect(1, 983040, '\p{In=+06.2}', "");
Expect(0, 983040, '\p{^In=+06.2}', "");
Expect(0, 983040, '\P{In=+06.2}', "");
Expect(1, 983040, '\P{^In=+06.2}', "");
Expect(0, 983037, '\p{In=+06.2}', "");
Expect(1, 983037, '\p{^In=+06.2}', "");
Expect(1, 983037, '\P{In=+06.2}', "");
Expect(0, 983037, '\P{^In=+06.2}', "");
Error('\p{Is_Present_In=__00000006.2:=}');
Error('\P{Is_Present_In=__00000006.2:=}');
Expect(1, 983040, '\p{Is_Present_In=00_00_00_6.2}', "");
Expect(0, 983040, '\p{^Is_Present_In=00_00_00_6.2}', "");
Expect(0, 983040, '\P{Is_Present_In=00_00_00_6.2}', "");
Expect(1, 983040, '\P{^Is_Present_In=00_00_00_6.2}', "");
Expect(0, 983037, '\p{Is_Present_In=00_00_00_6.2}', "");
Expect(1, 983037, '\p{^Is_Present_In=00_00_00_6.2}', "");
Expect(1, 983037, '\P{Is_Present_In=00_00_00_6.2}', "");
Expect(0, 983037, '\P{^Is_Present_In=00_00_00_6.2}', "");
Error('\p{Is_In=-0000006.2:=}');
Error('\P{Is_In=-0000006.2:=}');
Expect(1, 983040, '\p{Is_In=00000_6.2}', "");
Expect(0, 983040, '\p{^Is_In=00000_6.2}', "");
Expect(0, 983040, '\P{Is_In=00000_6.2}', "");
Expect(1, 983040, '\P{^Is_In=00000_6.2}', "");
Expect(0, 983037, '\p{Is_In=00000_6.2}', "");
Expect(1, 983037, '\p{^Is_In=00000_6.2}', "");
Expect(1, 983037, '\P{Is_In=00000_6.2}', "");
Expect(0, 983037, '\P{^Is_In=00000_6.2}', "");
Error('\p{Present_In=	/a/6.3}');
Error('\P{Present_In=	/a/6.3}');
Expect(1, 983040, '\p{Present_In=0000_6.3}', "");
Expect(0, 983040, '\p{^Present_In=0000_6.3}', "");
Expect(0, 983040, '\P{Present_In=0000_6.3}', "");
Expect(1, 983040, '\P{^Present_In=0000_6.3}', "");
Expect(0, 983037, '\p{Present_In=0000_6.3}', "");
Expect(1, 983037, '\p{^Present_In=0000_6.3}', "");
Expect(1, 983037, '\P{Present_In=0000_6.3}', "");
Expect(0, 983037, '\P{^Present_In=0000_6.3}', "");
Error('\p{In:	/a/0000006.3}');
Error('\P{In:	/a/0000006.3}');
Expect(1, 983040, '\p{In=00000006.3}', "");
Expect(0, 983040, '\p{^In=00000006.3}', "");
Expect(0, 983040, '\P{In=00000006.3}', "");
Expect(1, 983040, '\P{^In=00000006.3}', "");
Expect(0, 983037, '\p{In=00000006.3}', "");
Expect(1, 983037, '\p{^In=00000006.3}', "");
Expect(1, 983037, '\P{In=00000006.3}', "");
Expect(0, 983037, '\P{^In=00000006.3}', "");
Error('\p{Is_Present_In=/a/000_6.3}');
Error('\P{Is_Present_In=/a/000_6.3}');
Expect(1, 983040, '\p{Is_Present_In=000_000_000_6.3}', "");
Expect(0, 983040, '\p{^Is_Present_In=000_000_000_6.3}', "");
Expect(0, 983040, '\P{Is_Present_In=000_000_000_6.3}', "");
Expect(1, 983040, '\P{^Is_Present_In=000_000_000_6.3}', "");
Expect(0, 983037, '\p{Is_Present_In=000_000_000_6.3}', "");
Expect(1, 983037, '\p{^Is_Present_In=000_000_000_6.3}', "");
Expect(1, 983037, '\P{Is_Present_In=000_000_000_6.3}', "");
Expect(0, 983037, '\P{^Is_Present_In=000_000_000_6.3}', "");
Error('\p{Is_In=:= 00000006.3}');
Error('\P{Is_In=:= 00000006.3}');
Expect(1, 983040, '\p{Is_In=006.3}', "");
Expect(0, 983040, '\p{^Is_In=006.3}', "");
Expect(0, 983040, '\P{Is_In=006.3}', "");
Expect(1, 983040, '\P{^Is_In=006.3}', "");
Expect(0, 983037, '\p{Is_In=006.3}', "");
Expect(1, 983037, '\p{^Is_In=006.3}', "");
Expect(1, 983037, '\P{Is_In=006.3}', "");
Expect(0, 983037, '\P{^Is_In=006.3}', "");
Error('\p{Present_In=- UNASSIGNED/a/}');
Error('\P{Present_In=- UNASSIGNED/a/}');
Expect(1, 983037, '\p{Present_In=unassigned}', "");
Expect(0, 983037, '\p{^Present_In=unassigned}', "");
Expect(0, 983037, '\P{Present_In=unassigned}', "");
Expect(1, 983037, '\P{^Present_In=unassigned}', "");
Expect(0, 983040, '\p{Present_In=unassigned}', "");
Expect(1, 983040, '\p{^Present_In=unassigned}', "");
Expect(1, 983040, '\P{Present_In=unassigned}', "");
Expect(0, 983040, '\P{^Present_In=unassigned}', "");
Expect(1, 983037, '\p{Present_In=_Unassigned}', "");
Expect(0, 983037, '\p{^Present_In=_Unassigned}', "");
Expect(0, 983037, '\P{Present_In=_Unassigned}', "");
Expect(1, 983037, '\P{^Present_In=_Unassigned}', "");
Expect(0, 983040, '\p{Present_In=_Unassigned}', "");
Expect(1, 983040, '\p{^Present_In=_Unassigned}', "");
Expect(1, 983040, '\P{Present_In=_Unassigned}', "");
Expect(0, 983040, '\P{^Present_In=_Unassigned}', "");
Error('\p{In:   _UNASSIGNED:=}');
Error('\P{In:   _UNASSIGNED:=}');
Expect(1, 983037, '\p{In:   unassigned}', "");
Expect(0, 983037, '\p{^In:   unassigned}', "");
Expect(0, 983037, '\P{In:   unassigned}', "");
Expect(1, 983037, '\P{^In:   unassigned}', "");
Expect(0, 983040, '\p{In:   unassigned}', "");
Expect(1, 983040, '\p{^In:   unassigned}', "");
Expect(1, 983040, '\P{In:   unassigned}', "");
Expect(0, 983040, '\P{^In:   unassigned}', "");
Expect(1, 983037, '\p{In= 	Unassigned}', "");
Expect(0, 983037, '\p{^In= 	Unassigned}', "");
Expect(0, 983037, '\P{In= 	Unassigned}', "");
Expect(1, 983037, '\P{^In= 	Unassigned}', "");
Expect(0, 983040, '\p{In= 	Unassigned}', "");
Expect(1, 983040, '\p{^In= 	Unassigned}', "");
Expect(1, 983040, '\P{In= 	Unassigned}', "");
Expect(0, 983040, '\P{^In= 	Unassigned}', "");
Error('\p{Is_Present_In= _Unassigned:=}');
Error('\P{Is_Present_In= _Unassigned:=}');
Expect(1, 983037, '\p{Is_Present_In=unassigned}', "");
Expect(0, 983037, '\p{^Is_Present_In=unassigned}', "");
Expect(0, 983037, '\P{Is_Present_In=unassigned}', "");
Expect(1, 983037, '\P{^Is_Present_In=unassigned}', "");
Expect(0, 983040, '\p{Is_Present_In=unassigned}', "");
Expect(1, 983040, '\p{^Is_Present_In=unassigned}', "");
Expect(1, 983040, '\P{Is_Present_In=unassigned}', "");
Expect(0, 983040, '\P{^Is_Present_In=unassigned}', "");
Expect(1, 983037, '\p{Is_Present_In=_ Unassigned}', "");
Expect(0, 983037, '\p{^Is_Present_In=_ Unassigned}', "");
Expect(0, 983037, '\P{Is_Present_In=_ Unassigned}', "");
Expect(1, 983037, '\P{^Is_Present_In=_ Unassigned}', "");
Expect(0, 983040, '\p{Is_Present_In=_ Unassigned}', "");
Expect(1, 983040, '\p{^Is_Present_In=_ Unassigned}', "");
Expect(1, 983040, '\P{Is_Present_In=_ Unassigned}', "");
Expect(0, 983040, '\P{^Is_Present_In=_ Unassigned}', "");
Error('\p{Is_In=-/a/Unassigned}');
Error('\P{Is_In=-/a/Unassigned}');
Expect(1, 983037, '\p{Is_In:unassigned}', "");
Expect(0, 983037, '\p{^Is_In:unassigned}', "");
Expect(0, 983037, '\P{Is_In:unassigned}', "");
Expect(1, 983037, '\P{^Is_In:unassigned}', "");
Expect(0, 983040, '\p{Is_In:unassigned}', "");
Expect(1, 983040, '\p{^Is_In:unassigned}', "");
Expect(1, 983040, '\P{Is_In:unassigned}', "");
Expect(0, 983040, '\P{^Is_In:unassigned}', "");
Expect(1, 983037, '\p{Is_In= _UNASSIGNED}', "");
Expect(0, 983037, '\p{^Is_In= _UNASSIGNED}', "");
Expect(0, 983037, '\P{Is_In= _UNASSIGNED}', "");
Expect(1, 983037, '\P{^Is_In= _UNASSIGNED}', "");
Expect(0, 983040, '\p{Is_In= _UNASSIGNED}', "");
Expect(1, 983040, '\p{^Is_In= _UNASSIGNED}', "");
Expect(1, 983040, '\P{Is_In= _UNASSIGNED}', "");
Expect(0, 983040, '\P{^Is_In= _UNASSIGNED}', "");
Error('\p{Indic_Matra_Category=Bottom}');
Error('\P{Indic_Matra_Category=Bottom}');
Error('\p{InMC=Bottom}');
Error('\P{InMC=Bottom}');
Error('\p{Is_Indic_Matra_Category:Bottom}');
Error('\P{Is_Indic_Matra_Category:Bottom}');
Error('\p{Is_InMC=Bottom}');
Error('\P{Is_InMC=Bottom}');
Error('\p{Indic_Matra_Category=Bottom_And_Right}');
Error('\P{Indic_Matra_Category=Bottom_And_Right}');
Error('\p{InMC=Bottom_And_Right}');
Error('\P{InMC=Bottom_And_Right}');
Error('\p{Is_Indic_Matra_Category=Bottom_And_Right}');
Error('\P{Is_Indic_Matra_Category=Bottom_And_Right}');
Error('\p{Is_InMC=Bottom_And_Right}');
Error('\P{Is_InMC=Bottom_And_Right}');
Error('\p{Indic_Matra_Category=Invisible}');
Error('\P{Indic_Matra_Category=Invisible}');
Error('\p{InMC=Invisible}');
Error('\P{InMC=Invisible}');
Error('\p{Is_Indic_Matra_Category=Invisible}');
Error('\P{Is_Indic_Matra_Category=Invisible}');
Error('\p{Is_InMC:   Invisible}');
Error('\P{Is_InMC:   Invisible}');
Error('\p{Indic_Matra_Category=Left}');
Error('\P{Indic_Matra_Category=Left}');
Error('\p{InMC=Left}');
Error('\P{InMC=Left}');
Error('\p{Is_Indic_Matra_Category=Left}');
Error('\P{Is_Indic_Matra_Category=Left}');
Error('\p{Is_InMC=Left}');
Error('\P{Is_InMC=Left}');
Error('\p{Indic_Matra_Category: Left_And_Right}');
Error('\P{Indic_Matra_Category: Left_And_Right}');
Error('\p{InMC=Left_And_Right}');
Error('\P{InMC=Left_And_Right}');
Error('\p{Is_Indic_Matra_Category=Left_And_Right}');
Error('\P{Is_Indic_Matra_Category=Left_And_Right}');
Error('\p{Is_InMC=Left_And_Right}');
Error('\P{Is_InMC=Left_And_Right}');
Error('\p{Indic_Matra_Category=NA}');
Error('\P{Indic_Matra_Category=NA}');
Error('\p{InMC=NA}');
Error('\P{InMC=NA}');
Error('\p{Is_Indic_Matra_Category=NA}');
Error('\P{Is_Indic_Matra_Category=NA}');
Error('\p{Is_InMC=NA}');
Error('\P{Is_InMC=NA}');
Error('\p{Indic_Matra_Category=Overstruck}');
Error('\P{Indic_Matra_Category=Overstruck}');
Error('\p{InMC=Overstruck}');
Error('\P{InMC=Overstruck}');
Error('\p{Is_Indic_Matra_Category=Overstruck}');
Error('\P{Is_Indic_Matra_Category=Overstruck}');
Error('\p{Is_InMC: Overstruck}');
Error('\P{Is_InMC: Overstruck}');
Error('\p{Indic_Matra_Category=Right}');
Error('\P{Indic_Matra_Category=Right}');
Error('\p{InMC=Right}');
Error('\P{InMC=Right}');
Error('\p{Is_Indic_Matra_Category=Right}');
Error('\P{Is_Indic_Matra_Category=Right}');
Error('\p{Is_InMC=Right}');
Error('\P{Is_InMC=Right}');
Error('\p{Indic_Matra_Category=Top}');
Error('\P{Indic_Matra_Category=Top}');
Error('\p{InMC: Top}');
Error('\P{InMC: Top}');
Error('\p{Is_Indic_Matra_Category=Top}');
Error('\P{Is_Indic_Matra_Category=Top}');
Error('\p{Is_InMC=Top}');
Error('\P{Is_InMC=Top}');
Error('\p{Indic_Matra_Category=Top_And_Bottom}');
Error('\P{Indic_Matra_Category=Top_And_Bottom}');
Error('\p{InMC=Top_And_Bottom}');
Error('\P{InMC=Top_And_Bottom}');
Error('\p{Is_Indic_Matra_Category=Top_And_Bottom}');
Error('\P{Is_Indic_Matra_Category=Top_And_Bottom}');
Error('\p{Is_InMC=Top_And_Bottom}');
Error('\P{Is_InMC=Top_And_Bottom}');
Error('\p{Indic_Matra_Category=Top_And_Bottom_And_Right}');
Error('\P{Indic_Matra_Category=Top_And_Bottom_And_Right}');
Error('\p{InMC: Top_And_Bottom_And_Right}');
Error('\P{InMC: Top_And_Bottom_And_Right}');
Error('\p{Is_Indic_Matra_Category=Top_And_Bottom_And_Right}');
Error('\P{Is_Indic_Matra_Category=Top_And_Bottom_And_Right}');
Error('\p{Is_InMC:   Top_And_Bottom_And_Right}');
Error('\P{Is_InMC:   Top_And_Bottom_And_Right}');
Error('\p{Indic_Matra_Category=Top_And_Left}');
Error('\P{Indic_Matra_Category=Top_And_Left}');
Error('\p{InMC=Top_And_Left}');
Error('\P{InMC=Top_And_Left}');
Error('\p{Is_Indic_Matra_Category=Top_And_Left}');
Error('\P{Is_Indic_Matra_Category=Top_And_Left}');
Error('\p{Is_InMC=Top_And_Left}');
Error('\P{Is_InMC=Top_And_Left}');
Error('\p{Indic_Matra_Category=Top_And_Left_And_Right}');
Error('\P{Indic_Matra_Category=Top_And_Left_And_Right}');
Error('\p{InMC=Top_And_Left_And_Right}');
Error('\P{InMC=Top_And_Left_And_Right}');
Error('\p{Is_Indic_Matra_Category:   Top_And_Left_And_Right}');
Error('\P{Is_Indic_Matra_Category:   Top_And_Left_And_Right}');
Error('\p{Is_InMC=Top_And_Left_And_Right}');
Error('\P{Is_InMC=Top_And_Left_And_Right}');
Error('\p{Indic_Matra_Category=Top_And_Right}');
Error('\P{Indic_Matra_Category=Top_And_Right}');
Error('\p{InMC=Top_And_Right}');
Error('\P{InMC=Top_And_Right}');
Error('\p{Is_Indic_Matra_Category=Top_And_Right}');
Error('\P{Is_Indic_Matra_Category=Top_And_Right}');
Error('\p{Is_InMC: Top_And_Right}');
Error('\P{Is_InMC: Top_And_Right}');
Error('\p{Indic_Matra_Category=Visual_Order_Left}');
Error('\P{Indic_Matra_Category=Visual_Order_Left}');
Error('\p{InMC: Visual_Order_Left}');
Error('\P{InMC: Visual_Order_Left}');
Error('\p{Is_Indic_Matra_Category=Visual_Order_Left}');
Error('\P{Is_Indic_Matra_Category=Visual_Order_Left}');
Error('\p{Is_InMC=Visual_Order_Left}');
Error('\P{Is_InMC=Visual_Order_Left}');
Error('\p{Indic_Syllabic_Category=Avagraha}');
Error('\P{Indic_Syllabic_Category=Avagraha}');
Error('\p{InSC=Avagraha}');
Error('\P{InSC=Avagraha}');
Error('\p{Is_Indic_Syllabic_Category=Avagraha}');
Error('\P{Is_Indic_Syllabic_Category=Avagraha}');
Error('\p{Is_InSC:Avagraha}');
Error('\P{Is_InSC:Avagraha}');
Error('\p{Indic_Syllabic_Category=Bindu}');
Error('\P{Indic_Syllabic_Category=Bindu}');
Error('\p{InSC=Bindu}');
Error('\P{InSC=Bindu}');
Error('\p{Is_Indic_Syllabic_Category=Bindu}');
Error('\P{Is_Indic_Syllabic_Category=Bindu}');
Error('\p{Is_InSC=Bindu}');
Error('\P{Is_InSC=Bindu}');
Error('\p{Indic_Syllabic_Category=Consonant}');
Error('\P{Indic_Syllabic_Category=Consonant}');
Error('\p{InSC=Consonant}');
Error('\P{InSC=Consonant}');
Error('\p{Is_Indic_Syllabic_Category=Consonant}');
Error('\P{Is_Indic_Syllabic_Category=Consonant}');
Error('\p{Is_InSC=Consonant}');
Error('\P{Is_InSC=Consonant}');
Error('\p{Indic_Syllabic_Category=Consonant_Dead}');
Error('\P{Indic_Syllabic_Category=Consonant_Dead}');
Error('\p{InSC=Consonant_Dead}');
Error('\P{InSC=Consonant_Dead}');
Error('\p{Is_Indic_Syllabic_Category=Consonant_Dead}');
Error('\P{Is_Indic_Syllabic_Category=Consonant_Dead}');
Error('\p{Is_InSC=Consonant_Dead}');
Error('\P{Is_InSC=Consonant_Dead}');
Error('\p{Indic_Syllabic_Category=Consonant_Final}');
Error('\P{Indic_Syllabic_Category=Consonant_Final}');
Error('\p{InSC=Consonant_Final}');
Error('\P{InSC=Consonant_Final}');
Error('\p{Is_Indic_Syllabic_Category=Consonant_Final}');
Error('\P{Is_Indic_Syllabic_Category=Consonant_Final}');
Error('\p{Is_InSC=Consonant_Final}');
Error('\P{Is_InSC=Consonant_Final}');
Error('\p{Indic_Syllabic_Category=Consonant_Head_Letter}');
Error('\P{Indic_Syllabic_Category=Consonant_Head_Letter}');
Error('\p{InSC=Consonant_Head_Letter}');
Error('\P{InSC=Consonant_Head_Letter}');
Error('\p{Is_Indic_Syllabic_Category=Consonant_Head_Letter}');
Error('\P{Is_Indic_Syllabic_Category=Consonant_Head_Letter}');
Error('\p{Is_InSC:	Consonant_Head_Letter}');
Error('\P{Is_InSC:	Consonant_Head_Letter}');
Error('\p{Indic_Syllabic_Category=Consonant_Medial}');
Error('\P{Indic_Syllabic_Category=Consonant_Medial}');
Error('\p{InSC=Consonant_Medial}');
Error('\P{InSC=Consonant_Medial}');
Error('\p{Is_Indic_Syllabic_Category=Consonant_Medial}');
Error('\P{Is_Indic_Syllabic_Category=Consonant_Medial}');
Error('\p{Is_InSC=Consonant_Medial}');
Error('\P{Is_InSC=Consonant_Medial}');
Error('\p{Indic_Syllabic_Category=Consonant_Placeholder}');
Error('\P{Indic_Syllabic_Category=Consonant_Placeholder}');
Error('\p{InSC=Consonant_Placeholder}');
Error('\P{InSC=Consonant_Placeholder}');
Error('\p{Is_Indic_Syllabic_Category=Consonant_Placeholder}');
Error('\P{Is_Indic_Syllabic_Category=Consonant_Placeholder}');
Error('\p{Is_InSC=Consonant_Placeholder}');
Error('\P{Is_InSC=Consonant_Placeholder}');
Error('\p{Indic_Syllabic_Category=Consonant_Repha}');
Error('\P{Indic_Syllabic_Category=Consonant_Repha}');
Error('\p{InSC=Consonant_Repha}');
Error('\P{InSC=Consonant_Repha}');
Error('\p{Is_Indic_Syllabic_Category=Consonant_Repha}');
Error('\P{Is_Indic_Syllabic_Category=Consonant_Repha}');
Error('\p{Is_InSC:   Consonant_Repha}');
Error('\P{Is_InSC:   Consonant_Repha}');
Error('\p{Indic_Syllabic_Category=Consonant_Subjoined}');
Error('\P{Indic_Syllabic_Category=Consonant_Subjoined}');
Error('\p{InSC=Consonant_Subjoined}');
Error('\P{InSC=Consonant_Subjoined}');
Error('\p{Is_Indic_Syllabic_Category=Consonant_Subjoined}');
Error('\P{Is_Indic_Syllabic_Category=Consonant_Subjoined}');
Error('\p{Is_InSC=Consonant_Subjoined}');
Error('\P{Is_InSC=Consonant_Subjoined}');
Error('\p{Indic_Syllabic_Category=Modifying_Letter}');
Error('\P{Indic_Syllabic_Category=Modifying_Letter}');
Error('\p{InSC=Modifying_Letter}');
Error('\P{InSC=Modifying_Letter}');
Error('\p{Is_Indic_Syllabic_Category=Modifying_Letter}');
Error('\P{Is_Indic_Syllabic_Category=Modifying_Letter}');
Error('\p{Is_InSC=Modifying_Letter}');
Error('\P{Is_InSC=Modifying_Letter}');
Error('\p{Indic_Syllabic_Category=Nukta}');
Error('\P{Indic_Syllabic_Category=Nukta}');
Error('\p{InSC=Nukta}');
Error('\P{InSC=Nukta}');
Error('\p{Is_Indic_Syllabic_Category=Nukta}');
Error('\P{Is_Indic_Syllabic_Category=Nukta}');
Error('\p{Is_InSC=Nukta}');
Error('\P{Is_InSC=Nukta}');
Error('\p{Indic_Syllabic_Category=Other}');
Error('\P{Indic_Syllabic_Category=Other}');
Error('\p{InSC=Other}');
Error('\P{InSC=Other}');
Error('\p{Is_Indic_Syllabic_Category=Other}');
Error('\P{Is_Indic_Syllabic_Category=Other}');
Error('\p{Is_InSC=Other}');
Error('\P{Is_InSC=Other}');
Error('\p{Indic_Syllabic_Category=Register_Shifter}');
Error('\P{Indic_Syllabic_Category=Register_Shifter}');
Error('\p{InSC=Register_Shifter}');
Error('\P{InSC=Register_Shifter}');
Error('\p{Is_Indic_Syllabic_Category=Register_Shifter}');
Error('\P{Is_Indic_Syllabic_Category=Register_Shifter}');
Error('\p{Is_InSC=Register_Shifter}');
Error('\P{Is_InSC=Register_Shifter}');
Error('\p{Indic_Syllabic_Category=Tone_Letter}');
Error('\P{Indic_Syllabic_Category=Tone_Letter}');
Error('\p{InSC=Tone_Letter}');
Error('\P{InSC=Tone_Letter}');
Error('\p{Is_Indic_Syllabic_Category=Tone_Letter}');
Error('\P{Is_Indic_Syllabic_Category=Tone_Letter}');
Error('\p{Is_InSC=Tone_Letter}');
Error('\P{Is_InSC=Tone_Letter}');
Error('\p{Indic_Syllabic_Category=Tone_Mark}');
Error('\P{Indic_Syllabic_Category=Tone_Mark}');
Error('\p{InSC:   Tone_Mark}');
Error('\P{InSC:   Tone_Mark}');
Error('\p{Is_Indic_Syllabic_Category=Tone_Mark}');
Error('\P{Is_Indic_Syllabic_Category=Tone_Mark}');
Error('\p{Is_InSC=Tone_Mark}');
Error('\P{Is_InSC=Tone_Mark}');
Error('\p{Indic_Syllabic_Category: Virama}');
Error('\P{Indic_Syllabic_Category: Virama}');
Error('\p{InSC=Virama}');
Error('\P{InSC=Virama}');
Error('\p{Is_Indic_Syllabic_Category=Virama}');
Error('\P{Is_Indic_Syllabic_Category=Virama}');
Error('\p{Is_InSC:Virama}');
Error('\P{Is_InSC:Virama}');
Error('\p{Indic_Syllabic_Category=Visarga}');
Error('\P{Indic_Syllabic_Category=Visarga}');
Error('\p{InSC:   Visarga}');
Error('\P{InSC:   Visarga}');
Error('\p{Is_Indic_Syllabic_Category=Visarga}');
Error('\P{Is_Indic_Syllabic_Category=Visarga}');
Error('\p{Is_InSC=Visarga}');
Error('\P{Is_InSC=Visarga}');
Error('\p{Indic_Syllabic_Category=Vowel}');
Error('\P{Indic_Syllabic_Category=Vowel}');
Error('\p{InSC=Vowel}');
Error('\P{InSC=Vowel}');
Error('\p{Is_Indic_Syllabic_Category=Vowel}');
Error('\P{Is_Indic_Syllabic_Category=Vowel}');
Error('\p{Is_InSC=Vowel}');
Error('\P{Is_InSC=Vowel}');
Error('\p{Indic_Syllabic_Category=Vowel_Dependent}');
Error('\P{Indic_Syllabic_Category=Vowel_Dependent}');
Error('\p{InSC=Vowel_Dependent}');
Error('\P{InSC=Vowel_Dependent}');
Error('\p{Is_Indic_Syllabic_Category=Vowel_Dependent}');
Error('\P{Is_Indic_Syllabic_Category=Vowel_Dependent}');
Error('\p{Is_InSC=Vowel_Dependent}');
Error('\P{Is_InSC=Vowel_Dependent}');
Error('\p{Indic_Syllabic_Category=Vowel_Independent}');
Error('\P{Indic_Syllabic_Category=Vowel_Independent}');
Error('\p{InSC=Vowel_Independent}');
Error('\P{InSC=Vowel_Independent}');
Error('\p{Is_Indic_Syllabic_Category=Vowel_Independent}');
Error('\P{Is_Indic_Syllabic_Category=Vowel_Independent}');
Error('\p{Is_InSC=Vowel_Independent}');
Error('\P{Is_InSC=Vowel_Independent}');
Error('\p{Jamo_Short_Name=A}');
Error('\P{Jamo_Short_Name=A}');
Error('\p{JSN=A}');
Error('\P{JSN=A}');
Error('\p{Is_Jamo_Short_Name=A}');
Error('\P{Is_Jamo_Short_Name=A}');
Error('\p{Is_JSN: A}');
Error('\P{Is_JSN: A}');
Error('\p{Jamo_Short_Name=AE}');
Error('\P{Jamo_Short_Name=AE}');
Error('\p{JSN=AE}');
Error('\P{JSN=AE}');
Error('\p{Is_Jamo_Short_Name=AE}');
Error('\P{Is_Jamo_Short_Name=AE}');
Error('\p{Is_JSN:AE}');
Error('\P{Is_JSN:AE}');
Error('\p{Jamo_Short_Name=B}');
Error('\P{Jamo_Short_Name=B}');
Error('\p{JSN=B}');
Error('\P{JSN=B}');
Error('\p{Is_Jamo_Short_Name:	B}');
Error('\P{Is_Jamo_Short_Name:	B}');
Error('\p{Is_JSN=B}');
Error('\P{Is_JSN=B}');
Error('\p{Jamo_Short_Name=BB}');
Error('\P{Jamo_Short_Name=BB}');
Error('\p{JSN=BB}');
Error('\P{JSN=BB}');
Error('\p{Is_Jamo_Short_Name=BB}');
Error('\P{Is_Jamo_Short_Name=BB}');
Error('\p{Is_JSN=BB}');
Error('\P{Is_JSN=BB}');
Error('\p{Jamo_Short_Name=BS}');
Error('\P{Jamo_Short_Name=BS}');
Error('\p{JSN=BS}');
Error('\P{JSN=BS}');
Error('\p{Is_Jamo_Short_Name:BS}');
Error('\P{Is_Jamo_Short_Name:BS}');
Error('\p{Is_JSN=BS}');
Error('\P{Is_JSN=BS}');
Error('\p{Jamo_Short_Name=C}');
Error('\P{Jamo_Short_Name=C}');
Error('\p{JSN=C}');
Error('\P{JSN=C}');
Error('\p{Is_Jamo_Short_Name=C}');
Error('\P{Is_Jamo_Short_Name=C}');
Error('\p{Is_JSN=C}');
Error('\P{Is_JSN=C}');
Error('\p{Jamo_Short_Name: D}');
Error('\P{Jamo_Short_Name: D}');
Error('\p{JSN=D}');
Error('\P{JSN=D}');
Error('\p{Is_Jamo_Short_Name=D}');
Error('\P{Is_Jamo_Short_Name=D}');
Error('\p{Is_JSN=D}');
Error('\P{Is_JSN=D}');
Error('\p{Jamo_Short_Name:   DD}');
Error('\P{Jamo_Short_Name:   DD}');
Error('\p{JSN=DD}');
Error('\P{JSN=DD}');
Error('\p{Is_Jamo_Short_Name=DD}');
Error('\P{Is_Jamo_Short_Name=DD}');
Error('\p{Is_JSN=DD}');
Error('\P{Is_JSN=DD}');
Error('\p{Jamo_Short_Name=E}');
Error('\P{Jamo_Short_Name=E}');
Error('\p{JSN:	E}');
Error('\P{JSN:	E}');
Error('\p{Is_Jamo_Short_Name=E}');
Error('\P{Is_Jamo_Short_Name=E}');
Error('\p{Is_JSN=E}');
Error('\P{Is_JSN=E}');
Error('\p{Jamo_Short_Name:EO}');
Error('\P{Jamo_Short_Name:EO}');
Error('\p{JSN:   EO}');
Error('\P{JSN:   EO}');
Error('\p{Is_Jamo_Short_Name=EO}');
Error('\P{Is_Jamo_Short_Name=EO}');
Error('\p{Is_JSN=EO}');
Error('\P{Is_JSN=EO}');
Error('\p{Jamo_Short_Name=EU}');
Error('\P{Jamo_Short_Name=EU}');
Error('\p{JSN=EU}');
Error('\P{JSN=EU}');
Error('\p{Is_Jamo_Short_Name=EU}');
Error('\P{Is_Jamo_Short_Name=EU}');
Error('\p{Is_JSN=EU}');
Error('\P{Is_JSN=EU}');
Error('\p{Jamo_Short_Name=G}');
Error('\P{Jamo_Short_Name=G}');
Error('\p{JSN=G}');
Error('\P{JSN=G}');
Error('\p{Is_Jamo_Short_Name=G}');
Error('\P{Is_Jamo_Short_Name=G}');
Error('\p{Is_JSN=G}');
Error('\P{Is_JSN=G}');
Error('\p{Jamo_Short_Name=GG}');
Error('\P{Jamo_Short_Name=GG}');
Error('\p{JSN=GG}');
Error('\P{JSN=GG}');
Error('\p{Is_Jamo_Short_Name=GG}');
Error('\P{Is_Jamo_Short_Name=GG}');
Error('\p{Is_JSN=GG}');
Error('\P{Is_JSN=GG}');
Error('\p{Jamo_Short_Name: GS}');
Error('\P{Jamo_Short_Name: GS}');
Error('\p{JSN: GS}');
Error('\P{JSN: GS}');
Error('\p{Is_Jamo_Short_Name=GS}');
Error('\P{Is_Jamo_Short_Name=GS}');
Error('\p{Is_JSN=GS}');
Error('\P{Is_JSN=GS}');
Error('\p{Jamo_Short_Name=H}');
Error('\P{Jamo_Short_Name=H}');
Error('\p{JSN:H}');
Error('\P{JSN:H}');
Error('\p{Is_Jamo_Short_Name=H}');
Error('\P{Is_Jamo_Short_Name=H}');
Error('\p{Is_JSN=H}');
Error('\P{Is_JSN=H}');
Error('\p{Jamo_Short_Name=I}');
Error('\P{Jamo_Short_Name=I}');
Error('\p{JSN=I}');
Error('\P{JSN=I}');
Error('\p{Is_Jamo_Short_Name=I}');
Error('\P{Is_Jamo_Short_Name=I}');
Error('\p{Is_JSN:I}');
Error('\P{Is_JSN:I}');
Error('\p{Jamo_Short_Name=J}');
Error('\P{Jamo_Short_Name=J}');
Error('\p{JSN=J}');
Error('\P{JSN=J}');
Error('\p{Is_Jamo_Short_Name:   J}');
Error('\P{Is_Jamo_Short_Name:   J}');
Error('\p{Is_JSN=J}');
Error('\P{Is_JSN=J}');
Error('\p{Jamo_Short_Name=JJ}');
Error('\P{Jamo_Short_Name=JJ}');
Error('\p{JSN=JJ}');
Error('\P{JSN=JJ}');
Error('\p{Is_Jamo_Short_Name=JJ}');
Error('\P{Is_Jamo_Short_Name=JJ}');
Error('\p{Is_JSN=JJ}');
Error('\P{Is_JSN=JJ}');
Error('\p{Jamo_Short_Name:K}');
Error('\P{Jamo_Short_Name:K}');
Error('\p{JSN=K}');
Error('\P{JSN=K}');
Error('\p{Is_Jamo_Short_Name=K}');
Error('\P{Is_Jamo_Short_Name=K}');
Error('\p{Is_JSN=K}');
Error('\P{Is_JSN=K}');
Error('\p{Jamo_Short_Name:	L}');
Error('\P{Jamo_Short_Name:	L}');
Error('\p{JSN=L}');
Error('\P{JSN=L}');
Error('\p{Is_Jamo_Short_Name:	L}');
Error('\P{Is_Jamo_Short_Name:	L}');
Error('\p{Is_JSN=L}');
Error('\P{Is_JSN=L}');
Error('\p{Jamo_Short_Name=LB}');
Error('\P{Jamo_Short_Name=LB}');
Error('\p{JSN=LB}');
Error('\P{JSN=LB}');
Error('\p{Is_Jamo_Short_Name=LB}');
Error('\P{Is_Jamo_Short_Name=LB}');
Error('\p{Is_JSN=LB}');
Error('\P{Is_JSN=LB}');
Error('\p{Jamo_Short_Name=LG}');
Error('\P{Jamo_Short_Name=LG}');
Error('\p{JSN=LG}');
Error('\P{JSN=LG}');
Error('\p{Is_Jamo_Short_Name=LG}');
Error('\P{Is_Jamo_Short_Name=LG}');
Error('\p{Is_JSN=LG}');
Error('\P{Is_JSN=LG}');
Error('\p{Jamo_Short_Name=LH}');
Error('\P{Jamo_Short_Name=LH}');
Error('\p{JSN:LH}');
Error('\P{JSN:LH}');
Error('\p{Is_Jamo_Short_Name=LH}');
Error('\P{Is_Jamo_Short_Name=LH}');
Error('\p{Is_JSN=LH}');
Error('\P{Is_JSN=LH}');
Error('\p{Jamo_Short_Name=LM}');
Error('\P{Jamo_Short_Name=LM}');
Error('\p{JSN=LM}');
Error('\P{JSN=LM}');
Error('\p{Is_Jamo_Short_Name=LM}');
Error('\P{Is_Jamo_Short_Name=LM}');
Error('\p{Is_JSN=LM}');
Error('\P{Is_JSN=LM}');
Error('\p{Jamo_Short_Name=LP}');
Error('\P{Jamo_Short_Name=LP}');
Error('\p{JSN=LP}');
Error('\P{JSN=LP}');
Error('\p{Is_Jamo_Short_Name=LP}');
Error('\P{Is_Jamo_Short_Name=LP}');
Error('\p{Is_JSN=LP}');
Error('\P{Is_JSN=LP}');
Error('\p{Jamo_Short_Name=LS}');
Error('\P{Jamo_Short_Name=LS}');
Error('\p{JSN=LS}');
Error('\P{JSN=LS}');
Error('\p{Is_Jamo_Short_Name=LS}');
Error('\P{Is_Jamo_Short_Name=LS}');
Error('\p{Is_JSN=LS}');
Error('\P{Is_JSN=LS}');
Error('\p{Jamo_Short_Name=LT}');
Error('\P{Jamo_Short_Name=LT}');
Error('\p{JSN=LT}');
Error('\P{JSN=LT}');
Error('\p{Is_Jamo_Short_Name=LT}');
Error('\P{Is_Jamo_Short_Name=LT}');
Error('\p{Is_JSN=LT}');
Error('\P{Is_JSN=LT}');
Error('\p{Jamo_Short_Name=M}');
Error('\P{Jamo_Short_Name=M}');
Error('\p{JSN=M}');
Error('\P{JSN=M}');
Error('\p{Is_Jamo_Short_Name=M}');
Error('\P{Is_Jamo_Short_Name=M}');
Error('\p{Is_JSN: M}');
Error('\P{Is_JSN: M}');
Error('\p{Jamo_Short_Name=N}');
Error('\P{Jamo_Short_Name=N}');
Error('\p{JSN=N}');
Error('\P{JSN=N}');
Error('\p{Is_Jamo_Short_Name=N}');
Error('\P{Is_Jamo_Short_Name=N}');
Error('\p{Is_JSN=N}');
Error('\P{Is_JSN=N}');
Error('\p{Jamo_Short_Name=NG}');
Error('\P{Jamo_Short_Name=NG}');
Error('\p{JSN=NG}');
Error('\P{JSN=NG}');
Error('\p{Is_Jamo_Short_Name=NG}');
Error('\P{Is_Jamo_Short_Name=NG}');
Error('\p{Is_JSN=NG}');
Error('\P{Is_JSN=NG}');
Error('\p{Jamo_Short_Name=NH}');
Error('\P{Jamo_Short_Name=NH}');
Error('\p{JSN=NH}');
Error('\P{JSN=NH}');
Error('\p{Is_Jamo_Short_Name=NH}');
Error('\P{Is_Jamo_Short_Name=NH}');
Error('\p{Is_JSN=NH}');
Error('\P{Is_JSN=NH}');
Error('\p{Jamo_Short_Name:NJ}');
Error('\P{Jamo_Short_Name:NJ}');
Error('\p{JSN=NJ}');
Error('\P{JSN=NJ}');
Error('\p{Is_Jamo_Short_Name=NJ}');
Error('\P{Is_Jamo_Short_Name=NJ}');
Error('\p{Is_JSN=NJ}');
Error('\P{Is_JSN=NJ}');
Error('\p{Jamo_Short_Name: O}');
Error('\P{Jamo_Short_Name: O}');
Error('\p{JSN=O}');
Error('\P{JSN=O}');
Error('\p{Is_Jamo_Short_Name=O}');
Error('\P{Is_Jamo_Short_Name=O}');
Error('\p{Is_JSN:O}');
Error('\P{Is_JSN:O}');
Error('\p{Jamo_Short_Name=OE}');
Error('\P{Jamo_Short_Name=OE}');
Error('\p{JSN:   OE}');
Error('\P{JSN:   OE}');
Error('\p{Is_Jamo_Short_Name=OE}');
Error('\P{Is_Jamo_Short_Name=OE}');
Error('\p{Is_JSN=OE}');
Error('\P{Is_JSN=OE}');
Error('\p{Jamo_Short_Name=P}');
Error('\P{Jamo_Short_Name=P}');
Error('\p{JSN=P}');
Error('\P{JSN=P}');
Error('\p{Is_Jamo_Short_Name:   P}');
Error('\P{Is_Jamo_Short_Name:   P}');
Error('\p{Is_JSN=P}');
Error('\P{Is_JSN=P}');
Error('\p{Jamo_Short_Name=R}');
Error('\P{Jamo_Short_Name=R}');
Error('\p{JSN: R}');
Error('\P{JSN: R}');
Error('\p{Is_Jamo_Short_Name=R}');
Error('\P{Is_Jamo_Short_Name=R}');
Error('\p{Is_JSN=R}');
Error('\P{Is_JSN=R}');
Error('\p{Jamo_Short_Name=S}');
Error('\P{Jamo_Short_Name=S}');
Error('\p{JSN=S}');
Error('\P{JSN=S}');
Error('\p{Is_Jamo_Short_Name:   S}');
Error('\P{Is_Jamo_Short_Name:   S}');
Error('\p{Is_JSN=S}');
Error('\P{Is_JSN=S}');
Error('\p{Jamo_Short_Name=SS}');
Error('\P{Jamo_Short_Name=SS}');
Error('\p{JSN=SS}');
Error('\P{JSN=SS}');
Error('\p{Is_Jamo_Short_Name=SS}');
Error('\P{Is_Jamo_Short_Name=SS}');
Error('\p{Is_JSN=SS}');
Error('\P{Is_JSN=SS}');
Error('\p{Jamo_Short_Name=T}');
Error('\P{Jamo_Short_Name=T}');
Error('\p{JSN=T}');
Error('\P{JSN=T}');
Error('\p{Is_Jamo_Short_Name=T}');
Error('\P{Is_Jamo_Short_Name=T}');
Error('\p{Is_JSN=T}');
Error('\P{Is_JSN=T}');
Error('\p{Jamo_Short_Name=U}');
Error('\P{Jamo_Short_Name=U}');
Error('\p{JSN=U}');
Error('\P{JSN=U}');
Error('\p{Is_Jamo_Short_Name=U}');
Error('\P{Is_Jamo_Short_Name=U}');
Error('\p{Is_JSN=U}');
Error('\P{Is_JSN=U}');
Error('\p{Jamo_Short_Name=WA}');
Error('\P{Jamo_Short_Name=WA}');
Error('\p{JSN=WA}');
Error('\P{JSN=WA}');
Error('\p{Is_Jamo_Short_Name=WA}');
Error('\P{Is_Jamo_Short_Name=WA}');
Error('\p{Is_JSN=WA}');
Error('\P{Is_JSN=WA}');
Error('\p{Jamo_Short_Name=WAE}');
Error('\P{Jamo_Short_Name=WAE}');
Error('\p{JSN:   WAE}');
Error('\P{JSN:   WAE}');
Error('\p{Is_Jamo_Short_Name=WAE}');
Error('\P{Is_Jamo_Short_Name=WAE}');
Error('\p{Is_JSN=WAE}');
Error('\P{Is_JSN=WAE}');
Error('\p{Jamo_Short_Name=WE}');
Error('\P{Jamo_Short_Name=WE}');
Error('\p{JSN=WE}');
Error('\P{JSN=WE}');
Error('\p{Is_Jamo_Short_Name=WE}');
Error('\P{Is_Jamo_Short_Name=WE}');
Error('\p{Is_JSN=WE}');
Error('\P{Is_JSN=WE}');
Error('\p{Jamo_Short_Name:	WEO}');
Error('\P{Jamo_Short_Name:	WEO}');
Error('\p{JSN=WEO}');
Error('\P{JSN=WEO}');
Error('\p{Is_Jamo_Short_Name=WEO}');
Error('\P{Is_Jamo_Short_Name=WEO}');
Error('\p{Is_JSN=WEO}');
Error('\P{Is_JSN=WEO}');
Error('\p{Jamo_Short_Name=WI}');
Error('\P{Jamo_Short_Name=WI}');
Error('\p{JSN=WI}');
Error('\P{JSN=WI}');
Error('\p{Is_Jamo_Short_Name=WI}');
Error('\P{Is_Jamo_Short_Name=WI}');
Error('\p{Is_JSN=WI}');
Error('\P{Is_JSN=WI}');
Error('\p{Jamo_Short_Name=YA}');
Error('\P{Jamo_Short_Name=YA}');
Error('\p{JSN=YA}');
Error('\P{JSN=YA}');
Error('\p{Is_Jamo_Short_Name:   YA}');
Error('\P{Is_Jamo_Short_Name:   YA}');
Error('\p{Is_JSN=YA}');
Error('\P{Is_JSN=YA}');
Error('\p{Jamo_Short_Name=YAE}');
Error('\P{Jamo_Short_Name=YAE}');
Error('\p{JSN=YAE}');
Error('\P{JSN=YAE}');
Error('\p{Is_Jamo_Short_Name: YAE}');
Error('\P{Is_Jamo_Short_Name: YAE}');
Error('\p{Is_JSN=YAE}');
Error('\P{Is_JSN=YAE}');
Error('\p{Jamo_Short_Name=YE}');
Error('\P{Jamo_Short_Name=YE}');
Error('\p{JSN=YE}');
Error('\P{JSN=YE}');
Error('\p{Is_Jamo_Short_Name=YE}');
Error('\P{Is_Jamo_Short_Name=YE}');
Error('\p{Is_JSN=YE}');
Error('\P{Is_JSN=YE}');
Error('\p{Jamo_Short_Name:YEO}');
Error('\P{Jamo_Short_Name:YEO}');
Error('\p{JSN=YEO}');
Error('\P{JSN=YEO}');
Error('\p{Is_Jamo_Short_Name=YEO}');
Error('\P{Is_Jamo_Short_Name=YEO}');
Error('\p{Is_JSN=YEO}');
Error('\P{Is_JSN=YEO}');
Error('\p{Jamo_Short_Name=YI}');
Error('\P{Jamo_Short_Name=YI}');
Error('\p{JSN=YI}');
Error('\P{JSN=YI}');
Error('\p{Is_Jamo_Short_Name=YI}');
Error('\P{Is_Jamo_Short_Name=YI}');
Error('\p{Is_JSN=YI}');
Error('\P{Is_JSN=YI}');
Error('\p{Jamo_Short_Name=YO}');
Error('\P{Jamo_Short_Name=YO}');
Error('\p{JSN=YO}');
Error('\P{JSN=YO}');
Error('\p{Is_Jamo_Short_Name=YO}');
Error('\P{Is_Jamo_Short_Name=YO}');
Error('\p{Is_JSN=YO}');
Error('\P{Is_JSN=YO}');
Error('\p{Jamo_Short_Name=YU}');
Error('\P{Jamo_Short_Name=YU}');
Error('\p{JSN=YU}');
Error('\P{JSN=YU}');
Error('\p{Is_Jamo_Short_Name=YU}');
Error('\P{Is_Jamo_Short_Name=YU}');
Error('\p{Is_JSN=YU}');
Error('\P{Is_JSN=YU}');
Error('\p{Join_Control=:=No}');
Error('\P{Join_Control=:=No}');
Expect(1, 8206, '\p{Join_Control:	no}', "");
Expect(0, 8206, '\p{^Join_Control:	no}', "");
Expect(0, 8206, '\P{Join_Control:	no}', "");
Expect(1, 8206, '\P{^Join_Control:	no}', "");
Expect(0, 8205, '\p{Join_Control:	no}', "");
Expect(1, 8205, '\p{^Join_Control:	no}', "");
Expect(1, 8205, '\P{Join_Control:	no}', "");
Expect(0, 8205, '\P{^Join_Control:	no}', "");
Expect(1, 8206, '\p{Join_Control=		NO}', "");
Expect(0, 8206, '\p{^Join_Control=		NO}', "");
Expect(0, 8206, '\P{Join_Control=		NO}', "");
Expect(1, 8206, '\P{^Join_Control=		NO}', "");
Expect(0, 8205, '\p{Join_Control=		NO}', "");
Expect(1, 8205, '\p{^Join_Control=		NO}', "");
Expect(1, 8205, '\P{Join_Control=		NO}', "");
Expect(0, 8205, '\P{^Join_Control=		NO}', "");
Error('\p{Join_C=	N:=}');
Error('\P{Join_C=	N:=}');
Expect(1, 8206, '\p{Join_C=n}', "");
Expect(0, 8206, '\p{^Join_C=n}', "");
Expect(0, 8206, '\P{Join_C=n}', "");
Expect(1, 8206, '\P{^Join_C=n}', "");
Expect(0, 8205, '\p{Join_C=n}', "");
Expect(1, 8205, '\p{^Join_C=n}', "");
Expect(1, 8205, '\P{Join_C=n}', "");
Expect(0, 8205, '\P{^Join_C=n}', "");
Expect(1, 8206, '\p{Join_C: -	n}', "");
Expect(0, 8206, '\p{^Join_C: -	n}', "");
Expect(0, 8206, '\P{Join_C: -	n}', "");
Expect(1, 8206, '\P{^Join_C: -	n}', "");
Expect(0, 8205, '\p{Join_C: -	n}', "");
Expect(1, 8205, '\p{^Join_C: -	n}', "");
Expect(1, 8205, '\P{Join_C: -	n}', "");
Expect(0, 8205, '\P{^Join_C: -	n}', "");
Error('\p{Is_Join_Control=:= -F}');
Error('\P{Is_Join_Control=:= -F}');
Expect(1, 8206, '\p{Is_Join_Control=f}', "");
Expect(0, 8206, '\p{^Is_Join_Control=f}', "");
Expect(0, 8206, '\P{Is_Join_Control=f}', "");
Expect(1, 8206, '\P{^Is_Join_Control=f}', "");
Expect(0, 8205, '\p{Is_Join_Control=f}', "");
Expect(1, 8205, '\p{^Is_Join_Control=f}', "");
Expect(1, 8205, '\P{Is_Join_Control=f}', "");
Expect(0, 8205, '\P{^Is_Join_Control=f}', "");
Expect(1, 8206, '\p{Is_Join_Control=_ F}', "");
Expect(0, 8206, '\p{^Is_Join_Control=_ F}', "");
Expect(0, 8206, '\P{Is_Join_Control=_ F}', "");
Expect(1, 8206, '\P{^Is_Join_Control=_ F}', "");
Expect(0, 8205, '\p{Is_Join_Control=_ F}', "");
Expect(1, 8205, '\p{^Is_Join_Control=_ F}', "");
Expect(1, 8205, '\P{Is_Join_Control=_ F}', "");
Expect(0, 8205, '\P{^Is_Join_Control=_ F}', "");
Error('\p{Is_Join_C=_:=False}');
Error('\P{Is_Join_C=_:=False}');
Expect(1, 8206, '\p{Is_Join_C=false}', "");
Expect(0, 8206, '\p{^Is_Join_C=false}', "");
Expect(0, 8206, '\P{Is_Join_C=false}', "");
Expect(1, 8206, '\P{^Is_Join_C=false}', "");
Expect(0, 8205, '\p{Is_Join_C=false}', "");
Expect(1, 8205, '\p{^Is_Join_C=false}', "");
Expect(1, 8205, '\P{Is_Join_C=false}', "");
Expect(0, 8205, '\P{^Is_Join_C=false}', "");
Expect(1, 8206, '\p{Is_Join_C=-FALSE}', "");
Expect(0, 8206, '\p{^Is_Join_C=-FALSE}', "");
Expect(0, 8206, '\P{Is_Join_C=-FALSE}', "");
Expect(1, 8206, '\P{^Is_Join_C=-FALSE}', "");
Expect(0, 8205, '\p{Is_Join_C=-FALSE}', "");
Expect(1, 8205, '\p{^Is_Join_C=-FALSE}', "");
Expect(1, 8205, '\P{Is_Join_C=-FALSE}', "");
Expect(0, 8205, '\P{^Is_Join_C=-FALSE}', "");
Error('\p{Join_Control:_:=yes}');
Error('\P{Join_Control:_:=yes}');
Expect(1, 8205, '\p{Join_Control=yes}', "");
Expect(0, 8205, '\p{^Join_Control=yes}', "");
Expect(0, 8205, '\P{Join_Control=yes}', "");
Expect(1, 8205, '\P{^Join_Control=yes}', "");
Expect(0, 8206, '\p{Join_Control=yes}', "");
Expect(1, 8206, '\p{^Join_Control=yes}', "");
Expect(1, 8206, '\P{Join_Control=yes}', "");
Expect(0, 8206, '\P{^Join_Control=yes}', "");
Expect(1, 8205, '\p{Join_Control=_-YES}', "");
Expect(0, 8205, '\p{^Join_Control=_-YES}', "");
Expect(0, 8205, '\P{Join_Control=_-YES}', "");
Expect(1, 8205, '\P{^Join_Control=_-YES}', "");
Expect(0, 8206, '\p{Join_Control=_-YES}', "");
Expect(1, 8206, '\p{^Join_Control=_-YES}', "");
Expect(1, 8206, '\P{Join_Control=_-YES}', "");
Expect(0, 8206, '\P{^Join_Control=_-YES}', "");
Error('\p{Join_C:	:=- y}');
Error('\P{Join_C:	:=- y}');
Expect(1, 8205, '\p{Join_C=y}', "");
Expect(0, 8205, '\p{^Join_C=y}', "");
Expect(0, 8205, '\P{Join_C=y}', "");
Expect(1, 8205, '\P{^Join_C=y}', "");
Expect(0, 8206, '\p{Join_C=y}', "");
Expect(1, 8206, '\p{^Join_C=y}', "");
Expect(1, 8206, '\P{Join_C=y}', "");
Expect(0, 8206, '\P{^Join_C=y}', "");
Expect(1, 8205, '\p{Join_C:	 y}', "");
Expect(0, 8205, '\p{^Join_C:	 y}', "");
Expect(0, 8205, '\P{Join_C:	 y}', "");
Expect(1, 8205, '\P{^Join_C:	 y}', "");
Expect(0, 8206, '\p{Join_C:	 y}', "");
Expect(1, 8206, '\p{^Join_C:	 y}', "");
Expect(1, 8206, '\P{Join_C:	 y}', "");
Expect(0, 8206, '\P{^Join_C:	 y}', "");
Error('\p{Is_Join_Control=/a/ t}');
Error('\P{Is_Join_Control=/a/ t}');
Expect(1, 8205, '\p{Is_Join_Control=t}', "");
Expect(0, 8205, '\p{^Is_Join_Control=t}', "");
Expect(0, 8205, '\P{Is_Join_Control=t}', "");
Expect(1, 8205, '\P{^Is_Join_Control=t}', "");
Expect(0, 8206, '\p{Is_Join_Control=t}', "");
Expect(1, 8206, '\p{^Is_Join_Control=t}', "");
Expect(1, 8206, '\P{Is_Join_Control=t}', "");
Expect(0, 8206, '\P{^Is_Join_Control=t}', "");
Expect(1, 8205, '\p{Is_Join_Control= _T}', "");
Expect(0, 8205, '\p{^Is_Join_Control= _T}', "");
Expect(0, 8205, '\P{Is_Join_Control= _T}', "");
Expect(1, 8205, '\P{^Is_Join_Control= _T}', "");
Expect(0, 8206, '\p{Is_Join_Control= _T}', "");
Expect(1, 8206, '\p{^Is_Join_Control= _T}', "");
Expect(1, 8206, '\P{Is_Join_Control= _T}', "");
Expect(0, 8206, '\P{^Is_Join_Control= _T}', "");
Error('\p{Is_Join_C=	/a/True}');
Error('\P{Is_Join_C=	/a/True}');
Expect(1, 8205, '\p{Is_Join_C=true}', "");
Expect(0, 8205, '\p{^Is_Join_C=true}', "");
Expect(0, 8205, '\P{Is_Join_C=true}', "");
Expect(1, 8205, '\P{^Is_Join_C=true}', "");
Expect(0, 8206, '\p{Is_Join_C=true}', "");
Expect(1, 8206, '\p{^Is_Join_C=true}', "");
Expect(1, 8206, '\P{Is_Join_C=true}', "");
Expect(0, 8206, '\P{^Is_Join_C=true}', "");
Expect(1, 8205, '\p{Is_Join_C= _true}', "");
Expect(0, 8205, '\p{^Is_Join_C= _true}', "");
Expect(0, 8205, '\P{Is_Join_C= _true}', "");
Expect(1, 8205, '\P{^Is_Join_C= _true}', "");
Expect(0, 8206, '\p{Is_Join_C= _true}', "");
Expect(1, 8206, '\p{^Is_Join_C= _true}', "");
Expect(1, 8206, '\P{Is_Join_C= _true}', "");
Expect(0, 8206, '\P{^Is_Join_C= _true}', "");
Error('\p{Logical_Order_Exception=:= _NO}');
Error('\P{Logical_Order_Exception=:= _NO}');
Expect(1, 43709, '\p{Logical_Order_Exception:	no}', "");
Expect(0, 43709, '\p{^Logical_Order_Exception:	no}', "");
Expect(0, 43709, '\P{Logical_Order_Exception:	no}', "");
Expect(1, 43709, '\P{^Logical_Order_Exception:	no}', "");
Expect(0, 43708, '\p{Logical_Order_Exception:	no}', "");
Expect(1, 43708, '\p{^Logical_Order_Exception:	no}', "");
Expect(1, 43708, '\P{Logical_Order_Exception:	no}', "");
Expect(0, 43708, '\P{^Logical_Order_Exception:	no}', "");
Expect(1, 43709, '\p{Logical_Order_Exception=-NO}', "");
Expect(0, 43709, '\p{^Logical_Order_Exception=-NO}', "");
Expect(0, 43709, '\P{Logical_Order_Exception=-NO}', "");
Expect(1, 43709, '\P{^Logical_Order_Exception=-NO}', "");
Expect(0, 43708, '\p{Logical_Order_Exception=-NO}', "");
Expect(1, 43708, '\p{^Logical_Order_Exception=-NO}', "");
Expect(1, 43708, '\P{Logical_Order_Exception=-NO}', "");
Expect(0, 43708, '\P{^Logical_Order_Exception=-NO}', "");
Error('\p{LOE= _n:=}');
Error('\P{LOE= _n:=}');
Expect(1, 43709, '\p{LOE=n}', "");
Expect(0, 43709, '\p{^LOE=n}', "");
Expect(0, 43709, '\P{LOE=n}', "");
Expect(1, 43709, '\P{^LOE=n}', "");
Expect(0, 43708, '\p{LOE=n}', "");
Expect(1, 43708, '\p{^LOE=n}', "");
Expect(1, 43708, '\P{LOE=n}', "");
Expect(0, 43708, '\P{^LOE=n}', "");
Expect(1, 43709, '\p{LOE=-_N}', "");
Expect(0, 43709, '\p{^LOE=-_N}', "");
Expect(0, 43709, '\P{LOE=-_N}', "");
Expect(1, 43709, '\P{^LOE=-_N}', "");
Expect(0, 43708, '\p{LOE=-_N}', "");
Expect(1, 43708, '\p{^LOE=-_N}', "");
Expect(1, 43708, '\P{LOE=-_N}', "");
Expect(0, 43708, '\P{^LOE=-_N}', "");
Error('\p{Is_Logical_Order_Exception:    F:=}');
Error('\P{Is_Logical_Order_Exception:    F:=}');
Expect(1, 43709, '\p{Is_Logical_Order_Exception=f}', "");
Expect(0, 43709, '\p{^Is_Logical_Order_Exception=f}', "");
Expect(0, 43709, '\P{Is_Logical_Order_Exception=f}', "");
Expect(1, 43709, '\P{^Is_Logical_Order_Exception=f}', "");
Expect(0, 43708, '\p{Is_Logical_Order_Exception=f}', "");
Expect(1, 43708, '\p{^Is_Logical_Order_Exception=f}', "");
Expect(1, 43708, '\P{Is_Logical_Order_Exception=f}', "");
Expect(0, 43708, '\P{^Is_Logical_Order_Exception=f}', "");
Expect(1, 43709, '\p{Is_Logical_Order_Exception=--F}', "");
Expect(0, 43709, '\p{^Is_Logical_Order_Exception=--F}', "");
Expect(0, 43709, '\P{Is_Logical_Order_Exception=--F}', "");
Expect(1, 43709, '\P{^Is_Logical_Order_Exception=--F}', "");
Expect(0, 43708, '\p{Is_Logical_Order_Exception=--F}', "");
Expect(1, 43708, '\p{^Is_Logical_Order_Exception=--F}', "");
Expect(1, 43708, '\P{Is_Logical_Order_Exception=--F}', "");
Expect(0, 43708, '\P{^Is_Logical_Order_Exception=--F}', "");
Error('\p{Is_LOE= false/a/}');
Error('\P{Is_LOE= false/a/}');
Expect(1, 43709, '\p{Is_LOE=false}', "");
Expect(0, 43709, '\p{^Is_LOE=false}', "");
Expect(0, 43709, '\P{Is_LOE=false}', "");
Expect(1, 43709, '\P{^Is_LOE=false}', "");
Expect(0, 43708, '\p{Is_LOE=false}', "");
Expect(1, 43708, '\p{^Is_LOE=false}', "");
Expect(1, 43708, '\P{Is_LOE=false}', "");
Expect(0, 43708, '\P{^Is_LOE=false}', "");
Expect(1, 43709, '\p{Is_LOE=--false}', "");
Expect(0, 43709, '\p{^Is_LOE=--false}', "");
Expect(0, 43709, '\P{Is_LOE=--false}', "");
Expect(1, 43709, '\P{^Is_LOE=--false}', "");
Expect(0, 43708, '\p{Is_LOE=--false}', "");
Expect(1, 43708, '\p{^Is_LOE=--false}', "");
Expect(1, 43708, '\P{Is_LOE=--false}', "");
Expect(0, 43708, '\P{^Is_LOE=--false}', "");
Error('\p{Logical_Order_Exception::=_ Yes}');
Error('\P{Logical_Order_Exception::=_ Yes}');
Expect(1, 43708, '\p{Logical_Order_Exception: yes}', "");
Expect(0, 43708, '\p{^Logical_Order_Exception: yes}', "");
Expect(0, 43708, '\P{Logical_Order_Exception: yes}', "");
Expect(1, 43708, '\P{^Logical_Order_Exception: yes}', "");
Expect(0, 43709, '\p{Logical_Order_Exception: yes}', "");
Expect(1, 43709, '\p{^Logical_Order_Exception: yes}', "");
Expect(1, 43709, '\P{Logical_Order_Exception: yes}', "");
Expect(0, 43709, '\P{^Logical_Order_Exception: yes}', "");
Expect(1, 43708, '\p{Logical_Order_Exception: 	yes}', "");
Expect(0, 43708, '\p{^Logical_Order_Exception: 	yes}', "");
Expect(0, 43708, '\P{Logical_Order_Exception: 	yes}', "");
Expect(1, 43708, '\P{^Logical_Order_Exception: 	yes}', "");
Expect(0, 43709, '\p{Logical_Order_Exception: 	yes}', "");
Expect(1, 43709, '\p{^Logical_Order_Exception: 	yes}', "");
Expect(1, 43709, '\P{Logical_Order_Exception: 	yes}', "");
Expect(0, 43709, '\P{^Logical_Order_Exception: 	yes}', "");
Error('\p{LOE=:= _Y}');
Error('\P{LOE=:= _Y}');
Expect(1, 43708, '\p{LOE=y}', "");
Expect(0, 43708, '\p{^LOE=y}', "");
Expect(0, 43708, '\P{LOE=y}', "");
Expect(1, 43708, '\P{^LOE=y}', "");
Expect(0, 43709, '\p{LOE=y}', "");
Expect(1, 43709, '\p{^LOE=y}', "");
Expect(1, 43709, '\P{LOE=y}', "");
Expect(0, 43709, '\P{^LOE=y}', "");
Expect(1, 43708, '\p{LOE=	y}', "");
Expect(0, 43708, '\p{^LOE=	y}', "");
Expect(0, 43708, '\P{LOE=	y}', "");
Expect(1, 43708, '\P{^LOE=	y}', "");
Expect(0, 43709, '\p{LOE=	y}', "");
Expect(1, 43709, '\p{^LOE=	y}', "");
Expect(1, 43709, '\P{LOE=	y}', "");
Expect(0, 43709, '\P{^LOE=	y}', "");
Error('\p{Is_Logical_Order_Exception=__t/a/}');
Error('\P{Is_Logical_Order_Exception=__t/a/}');
Expect(1, 43708, '\p{Is_Logical_Order_Exception=t}', "");
Expect(0, 43708, '\p{^Is_Logical_Order_Exception=t}', "");
Expect(0, 43708, '\P{Is_Logical_Order_Exception=t}', "");
Expect(1, 43708, '\P{^Is_Logical_Order_Exception=t}', "");
Expect(0, 43709, '\p{Is_Logical_Order_Exception=t}', "");
Expect(1, 43709, '\p{^Is_Logical_Order_Exception=t}', "");
Expect(1, 43709, '\P{Is_Logical_Order_Exception=t}', "");
Expect(0, 43709, '\P{^Is_Logical_Order_Exception=t}', "");
Expect(1, 43708, '\p{Is_Logical_Order_Exception=		T}', "");
Expect(0, 43708, '\p{^Is_Logical_Order_Exception=		T}', "");
Expect(0, 43708, '\P{Is_Logical_Order_Exception=		T}', "");
Expect(1, 43708, '\P{^Is_Logical_Order_Exception=		T}', "");
Expect(0, 43709, '\p{Is_Logical_Order_Exception=		T}', "");
Expect(1, 43709, '\p{^Is_Logical_Order_Exception=		T}', "");
Expect(1, 43709, '\P{Is_Logical_Order_Exception=		T}', "");
Expect(0, 43709, '\P{^Is_Logical_Order_Exception=		T}', "");
Error('\p{Is_LOE: _-TRUE/a/}');
Error('\P{Is_LOE: _-TRUE/a/}');
Expect(1, 43708, '\p{Is_LOE=true}', "");
Expect(0, 43708, '\p{^Is_LOE=true}', "");
Expect(0, 43708, '\P{Is_LOE=true}', "");
Expect(1, 43708, '\P{^Is_LOE=true}', "");
Expect(0, 43709, '\p{Is_LOE=true}', "");
Expect(1, 43709, '\p{^Is_LOE=true}', "");
Expect(1, 43709, '\P{Is_LOE=true}', "");
Expect(0, 43709, '\P{^Is_LOE=true}', "");
Expect(1, 43708, '\p{Is_LOE:  TRUE}', "");
Expect(0, 43708, '\p{^Is_LOE:  TRUE}', "");
Expect(0, 43708, '\P{Is_LOE:  TRUE}', "");
Expect(1, 43708, '\P{^Is_LOE:  TRUE}', "");
Expect(0, 43709, '\p{Is_LOE:  TRUE}', "");
Expect(1, 43709, '\p{^Is_LOE:  TRUE}', "");
Expect(1, 43709, '\P{Is_LOE:  TRUE}', "");
Expect(0, 43709, '\P{^Is_LOE:  TRUE}', "");
Error('\p{Lowercase=NO:=}');
Error('\P{Lowercase=NO:=}');
Expect(1, 120780, '\p{Lowercase=no}', "");
Expect(0, 120780, '\p{^Lowercase=no}', "");
Expect(0, 120780, '\P{Lowercase=no}', "");
Expect(1, 120780, '\P{^Lowercase=no}', "");
Expect(0, 120779, '\p{Lowercase=no}', "");
Expect(1, 120779, '\p{^Lowercase=no}', "");
Expect(1, 120779, '\P{Lowercase=no}', "");
Expect(0, 120779, '\P{^Lowercase=no}', "");
Expect(1, 120780, '\p{Lowercase: _	no}', "");
Expect(0, 120780, '\p{^Lowercase: _	no}', "");
Expect(0, 120780, '\P{Lowercase: _	no}', "");
Expect(1, 120780, '\P{^Lowercase: _	no}', "");
Expect(0, 120779, '\p{Lowercase: _	no}', "");
Expect(1, 120779, '\p{^Lowercase: _	no}', "");
Expect(1, 120779, '\P{Lowercase: _	no}', "");
Expect(0, 120779, '\P{^Lowercase: _	no}', "");
Error('\p{Lower=	 n:=}');
Error('\P{Lower=	 n:=}');
Expect(1, 120780, '\p{Lower=n}', "");
Expect(0, 120780, '\p{^Lower=n}', "");
Expect(0, 120780, '\P{Lower=n}', "");
Expect(1, 120780, '\P{^Lower=n}', "");
Expect(0, 120779, '\p{Lower=n}', "");
Expect(1, 120779, '\p{^Lower=n}', "");
Expect(1, 120779, '\P{Lower=n}', "");
Expect(0, 120779, '\P{^Lower=n}', "");
Expect(1, 120780, '\p{Lower= N}', "");
Expect(0, 120780, '\p{^Lower= N}', "");
Expect(0, 120780, '\P{Lower= N}', "");
Expect(1, 120780, '\P{^Lower= N}', "");
Expect(0, 120779, '\p{Lower= N}', "");
Expect(1, 120779, '\p{^Lower= N}', "");
Expect(1, 120779, '\P{Lower= N}', "");
Expect(0, 120779, '\P{^Lower= N}', "");
Error('\p{Is_Lowercase:   _F:=}');
Error('\P{Is_Lowercase:   _F:=}');
Expect(1, 120780, '\p{Is_Lowercase=f}', "");
Expect(0, 120780, '\p{^Is_Lowercase=f}', "");
Expect(0, 120780, '\P{Is_Lowercase=f}', "");
Expect(1, 120780, '\P{^Is_Lowercase=f}', "");
Expect(0, 120779, '\p{Is_Lowercase=f}', "");
Expect(1, 120779, '\p{^Is_Lowercase=f}', "");
Expect(1, 120779, '\P{Is_Lowercase=f}', "");
Expect(0, 120779, '\P{^Is_Lowercase=f}', "");
Expect(1, 120780, '\p{Is_Lowercase=_ F}', "");
Expect(0, 120780, '\p{^Is_Lowercase=_ F}', "");
Expect(0, 120780, '\P{Is_Lowercase=_ F}', "");
Expect(1, 120780, '\P{^Is_Lowercase=_ F}', "");
Expect(0, 120779, '\p{Is_Lowercase=_ F}', "");
Expect(1, 120779, '\p{^Is_Lowercase=_ F}', "");
Expect(1, 120779, '\P{Is_Lowercase=_ F}', "");
Expect(0, 120779, '\P{^Is_Lowercase=_ F}', "");
Error('\p{Is_Lower=_/a/False}');
Error('\P{Is_Lower=_/a/False}');
Expect(1, 120780, '\p{Is_Lower:   false}', "");
Expect(0, 120780, '\p{^Is_Lower:   false}', "");
Expect(0, 120780, '\P{Is_Lower:   false}', "");
Expect(1, 120780, '\P{^Is_Lower:   false}', "");
Expect(0, 120779, '\p{Is_Lower:   false}', "");
Expect(1, 120779, '\p{^Is_Lower:   false}', "");
Expect(1, 120779, '\P{Is_Lower:   false}', "");
Expect(0, 120779, '\P{^Is_Lower:   false}', "");
Expect(1, 120780, '\p{Is_Lower=-False}', "");
Expect(0, 120780, '\p{^Is_Lower=-False}', "");
Expect(0, 120780, '\P{Is_Lower=-False}', "");
Expect(1, 120780, '\P{^Is_Lower=-False}', "");
Expect(0, 120779, '\p{Is_Lower=-False}', "");
Expect(1, 120779, '\p{^Is_Lower=-False}', "");
Expect(1, 120779, '\P{Is_Lower=-False}', "");
Expect(0, 120779, '\P{^Is_Lower=-False}', "");
Error('\p{Lowercase=:= -YES}');
Error('\P{Lowercase=:= -YES}');
Expect(1, 120779, '\p{Lowercase:yes}', "");
Expect(0, 120779, '\p{^Lowercase:yes}', "");
Expect(0, 120779, '\P{Lowercase:yes}', "");
Expect(1, 120779, '\P{^Lowercase:yes}', "");
Expect(0, 120780, '\p{Lowercase:yes}', "");
Expect(1, 120780, '\p{^Lowercase:yes}', "");
Expect(1, 120780, '\P{Lowercase:yes}', "");
Expect(0, 120780, '\P{^Lowercase:yes}', "");
Expect(1, 120779, '\p{Lowercase=		YES}', "");
Expect(0, 120779, '\p{^Lowercase=		YES}', "");
Expect(0, 120779, '\P{Lowercase=		YES}', "");
Expect(1, 120779, '\P{^Lowercase=		YES}', "");
Expect(0, 120780, '\p{Lowercase=		YES}', "");
Expect(1, 120780, '\p{^Lowercase=		YES}', "");
Expect(1, 120780, '\P{Lowercase=		YES}', "");
Expect(0, 120780, '\P{^Lowercase=		YES}', "");
Error('\p{Lower=:=-	Y}');
Error('\P{Lower=:=-	Y}');
Expect(1, 120779, '\p{Lower=y}', "");
Expect(0, 120779, '\p{^Lower=y}', "");
Expect(0, 120779, '\P{Lower=y}', "");
Expect(1, 120779, '\P{^Lower=y}', "");
Expect(0, 120780, '\p{Lower=y}', "");
Expect(1, 120780, '\p{^Lower=y}', "");
Expect(1, 120780, '\P{Lower=y}', "");
Expect(0, 120780, '\P{^Lower=y}', "");
Expect(1, 120779, '\p{Lower= -Y}', "");
Expect(0, 120779, '\p{^Lower= -Y}', "");
Expect(0, 120779, '\P{Lower= -Y}', "");
Expect(1, 120779, '\P{^Lower= -Y}', "");
Expect(0, 120780, '\p{Lower= -Y}', "");
Expect(1, 120780, '\p{^Lower= -Y}', "");
Expect(1, 120780, '\P{Lower= -Y}', "");
Expect(0, 120780, '\P{^Lower= -Y}', "");
Error('\p{Is_Lowercase=_ t/a/}');
Error('\P{Is_Lowercase=_ t/a/}');
Expect(1, 120779, '\p{Is_Lowercase=t}', "");
Expect(0, 120779, '\p{^Is_Lowercase=t}', "");
Expect(0, 120779, '\P{Is_Lowercase=t}', "");
Expect(1, 120779, '\P{^Is_Lowercase=t}', "");
Expect(0, 120780, '\p{Is_Lowercase=t}', "");
Expect(1, 120780, '\p{^Is_Lowercase=t}', "");
Expect(1, 120780, '\P{Is_Lowercase=t}', "");
Expect(0, 120780, '\P{^Is_Lowercase=t}', "");
Expect(1, 120779, '\p{Is_Lowercase= -T}', "");
Expect(0, 120779, '\p{^Is_Lowercase= -T}', "");
Expect(0, 120779, '\P{Is_Lowercase= -T}', "");
Expect(1, 120779, '\P{^Is_Lowercase= -T}', "");
Expect(0, 120780, '\p{Is_Lowercase= -T}', "");
Expect(1, 120780, '\p{^Is_Lowercase= -T}', "");
Expect(1, 120780, '\P{Is_Lowercase= -T}', "");
Expect(0, 120780, '\P{^Is_Lowercase= -T}', "");
Error('\p{Is_Lower=	:=true}');
Error('\P{Is_Lower=	:=true}');
Expect(1, 120779, '\p{Is_Lower=true}', "");
Expect(0, 120779, '\p{^Is_Lower=true}', "");
Expect(0, 120779, '\P{Is_Lower=true}', "");
Expect(1, 120779, '\P{^Is_Lower=true}', "");
Expect(0, 120780, '\p{Is_Lower=true}', "");
Expect(1, 120780, '\p{^Is_Lower=true}', "");
Expect(1, 120780, '\P{Is_Lower=true}', "");
Expect(0, 120780, '\P{^Is_Lower=true}', "");
Expect(1, 120779, '\p{Is_Lower=-TRUE}', "");
Expect(0, 120779, '\p{^Is_Lower=-TRUE}', "");
Expect(0, 120779, '\P{Is_Lower=-TRUE}', "");
Expect(1, 120779, '\P{^Is_Lower=-TRUE}', "");
Expect(0, 120780, '\p{Is_Lower=-TRUE}', "");
Expect(1, 120780, '\p{^Is_Lower=-TRUE}', "");
Expect(1, 120780, '\P{Is_Lower=-TRUE}', "");
Expect(0, 120780, '\P{^Is_Lower=-TRUE}', "");
Error('\p{Math=-/a/No}');
Error('\P{Math=-/a/No}');
Expect(1, 126706, '\p{Math=no}', "");
Expect(0, 126706, '\p{^Math=no}', "");
Expect(0, 126706, '\P{Math=no}', "");
Expect(1, 126706, '\P{^Math=no}', "");
Expect(0, 126705, '\p{Math=no}', "");
Expect(1, 126705, '\p{^Math=no}', "");
Expect(1, 126705, '\P{Math=no}', "");
Expect(0, 126705, '\P{^Math=no}', "");
Expect(1, 126706, '\p{Math=- No}', "");
Expect(0, 126706, '\p{^Math=- No}', "");
Expect(0, 126706, '\P{Math=- No}', "");
Expect(1, 126706, '\P{^Math=- No}', "");
Expect(0, 126705, '\p{Math=- No}', "");
Expect(1, 126705, '\p{^Math=- No}', "");
Expect(1, 126705, '\P{Math=- No}', "");
Expect(0, 126705, '\P{^Math=- No}', "");
Error('\p{Is_Math:   :=N}');
Error('\P{Is_Math:   :=N}');
Expect(1, 126706, '\p{Is_Math=n}', "");
Expect(0, 126706, '\p{^Is_Math=n}', "");
Expect(0, 126706, '\P{Is_Math=n}', "");
Expect(1, 126706, '\P{^Is_Math=n}', "");
Expect(0, 126705, '\p{Is_Math=n}', "");
Expect(1, 126705, '\p{^Is_Math=n}', "");
Expect(1, 126705, '\P{Is_Math=n}', "");
Expect(0, 126705, '\P{^Is_Math=n}', "");
Expect(1, 126706, '\p{Is_Math=__N}', "");
Expect(0, 126706, '\p{^Is_Math=__N}', "");
Expect(0, 126706, '\P{Is_Math=__N}', "");
Expect(1, 126706, '\P{^Is_Math=__N}', "");
Expect(0, 126705, '\p{Is_Math=__N}', "");
Expect(1, 126705, '\p{^Is_Math=__N}', "");
Expect(1, 126705, '\P{Is_Math=__N}', "");
Expect(0, 126705, '\P{^Is_Math=__N}', "");
Error('\p{Math= f:=}');
Error('\P{Math= f:=}');
Expect(1, 126706, '\p{Math=f}', "");
Expect(0, 126706, '\p{^Math=f}', "");
Expect(0, 126706, '\P{Math=f}', "");
Expect(1, 126706, '\P{^Math=f}', "");
Expect(0, 126705, '\p{Math=f}', "");
Expect(1, 126705, '\p{^Math=f}', "");
Expect(1, 126705, '\P{Math=f}', "");
Expect(0, 126705, '\P{^Math=f}', "");
Expect(1, 126706, '\p{Math:  F}', "");
Expect(0, 126706, '\p{^Math:  F}', "");
Expect(0, 126706, '\P{Math:  F}', "");
Expect(1, 126706, '\P{^Math:  F}', "");
Expect(0, 126705, '\p{Math:  F}', "");
Expect(1, 126705, '\p{^Math:  F}', "");
Expect(1, 126705, '\P{Math:  F}', "");
Expect(0, 126705, '\P{^Math:  F}', "");
Error('\p{Is_Math: /a/ false}');
Error('\P{Is_Math: /a/ false}');
Expect(1, 126706, '\p{Is_Math=false}', "");
Expect(0, 126706, '\p{^Is_Math=false}', "");
Expect(0, 126706, '\P{Is_Math=false}', "");
Expect(1, 126706, '\P{^Is_Math=false}', "");
Expect(0, 126705, '\p{Is_Math=false}', "");
Expect(1, 126705, '\p{^Is_Math=false}', "");
Expect(1, 126705, '\P{Is_Math=false}', "");
Expect(0, 126705, '\P{^Is_Math=false}', "");
Expect(1, 126706, '\p{Is_Math=	FALSE}', "");
Expect(0, 126706, '\p{^Is_Math=	FALSE}', "");
Expect(0, 126706, '\P{Is_Math=	FALSE}', "");
Expect(1, 126706, '\P{^Is_Math=	FALSE}', "");
Expect(0, 126705, '\p{Is_Math=	FALSE}', "");
Expect(1, 126705, '\p{^Is_Math=	FALSE}', "");
Expect(1, 126705, '\P{Is_Math=	FALSE}', "");
Expect(0, 126705, '\P{^Is_Math=	FALSE}', "");
Error('\p{Math::=  Yes}');
Error('\P{Math::=  Yes}');
Expect(1, 126705, '\p{Math=yes}', "");
Expect(0, 126705, '\p{^Math=yes}', "");
Expect(0, 126705, '\P{Math=yes}', "");
Expect(1, 126705, '\P{^Math=yes}', "");
Expect(0, 126706, '\p{Math=yes}', "");
Expect(1, 126706, '\p{^Math=yes}', "");
Expect(1, 126706, '\P{Math=yes}', "");
Expect(0, 126706, '\P{^Math=yes}', "");
Expect(1, 126705, '\p{Math= Yes}', "");
Expect(0, 126705, '\p{^Math= Yes}', "");
Expect(0, 126705, '\P{Math= Yes}', "");
Expect(1, 126705, '\P{^Math= Yes}', "");
Expect(0, 126706, '\p{Math= Yes}', "");
Expect(1, 126706, '\p{^Math= Yes}', "");
Expect(1, 126706, '\P{Math= Yes}', "");
Expect(0, 126706, '\P{^Math= Yes}', "");
Error('\p{Is_Math=/a/_Y}');
Error('\P{Is_Math=/a/_Y}');
Expect(1, 126705, '\p{Is_Math=y}', "");
Expect(0, 126705, '\p{^Is_Math=y}', "");
Expect(0, 126705, '\P{Is_Math=y}', "");
Expect(1, 126705, '\P{^Is_Math=y}', "");
Expect(0, 126706, '\p{Is_Math=y}', "");
Expect(1, 126706, '\p{^Is_Math=y}', "");
Expect(1, 126706, '\P{Is_Math=y}', "");
Expect(0, 126706, '\P{^Is_Math=y}', "");
Expect(1, 126705, '\p{Is_Math= 	Y}', "");
Expect(0, 126705, '\p{^Is_Math= 	Y}', "");
Expect(0, 126705, '\P{Is_Math= 	Y}', "");
Expect(1, 126705, '\P{^Is_Math= 	Y}', "");
Expect(0, 126706, '\p{Is_Math= 	Y}', "");
Expect(1, 126706, '\p{^Is_Math= 	Y}', "");
Expect(1, 126706, '\P{Is_Math= 	Y}', "");
Expect(0, 126706, '\P{^Is_Math= 	Y}', "");
Error('\p{Math: /a/T}');
Error('\P{Math: /a/T}');
Expect(1, 126705, '\p{Math=t}', "");
Expect(0, 126705, '\p{^Math=t}', "");
Expect(0, 126705, '\P{Math=t}', "");
Expect(1, 126705, '\P{^Math=t}', "");
Expect(0, 126706, '\p{Math=t}', "");
Expect(1, 126706, '\p{^Math=t}', "");
Expect(1, 126706, '\P{Math=t}', "");
Expect(0, 126706, '\P{^Math=t}', "");
Expect(1, 126705, '\p{Math=		t}', "");
Expect(0, 126705, '\p{^Math=		t}', "");
Expect(0, 126705, '\P{Math=		t}', "");
Expect(1, 126705, '\P{^Math=		t}', "");
Expect(0, 126706, '\p{Math=		t}', "");
Expect(1, 126706, '\p{^Math=		t}', "");
Expect(1, 126706, '\P{Math=		t}', "");
Expect(0, 126706, '\P{^Math=		t}', "");
Error('\p{Is_Math: /a/	_True}');
Error('\P{Is_Math: /a/	_True}');
Expect(1, 126705, '\p{Is_Math=true}', "");
Expect(0, 126705, '\p{^Is_Math=true}', "");
Expect(0, 126705, '\P{Is_Math=true}', "");
Expect(1, 126705, '\P{^Is_Math=true}', "");
Expect(0, 126706, '\p{Is_Math=true}', "");
Expect(1, 126706, '\p{^Is_Math=true}', "");
Expect(1, 126706, '\P{Is_Math=true}', "");
Expect(0, 126706, '\P{^Is_Math=true}', "");
Expect(1, 126705, '\p{Is_Math=_True}', "");
Expect(0, 126705, '\p{^Is_Math=_True}', "");
Expect(0, 126705, '\P{Is_Math=_True}', "");
Expect(1, 126705, '\P{^Is_Math=_True}', "");
Expect(0, 126706, '\p{Is_Math=_True}', "");
Expect(1, 126706, '\p{^Is_Math=_True}', "");
Expect(1, 126706, '\P{Is_Math=_True}', "");
Expect(0, 126706, '\P{^Is_Math=_True}', "");
Error('\p{Noncharacter_Code_Point=	:=No}');
Error('\P{Noncharacter_Code_Point=	:=No}');
Expect(1, 1114109, '\p{Noncharacter_Code_Point=no}', "");
Expect(0, 1114109, '\p{^Noncharacter_Code_Point=no}', "");
Expect(0, 1114109, '\P{Noncharacter_Code_Point=no}', "");
Expect(1, 1114109, '\P{^Noncharacter_Code_Point=no}', "");
Expect(0, 1114111, '\p{Noncharacter_Code_Point=no}', "");
Expect(1, 1114111, '\p{^Noncharacter_Code_Point=no}', "");
Expect(1, 1114111, '\P{Noncharacter_Code_Point=no}', "");
Expect(0, 1114111, '\P{^Noncharacter_Code_Point=no}', "");
Expect(1, 1114109, '\p{Noncharacter_Code_Point= NO}', "");
Expect(0, 1114109, '\p{^Noncharacter_Code_Point= NO}', "");
Expect(0, 1114109, '\P{Noncharacter_Code_Point= NO}', "");
Expect(1, 1114109, '\P{^Noncharacter_Code_Point= NO}', "");
Expect(0, 1114111, '\p{Noncharacter_Code_Point= NO}', "");
Expect(1, 1114111, '\p{^Noncharacter_Code_Point= NO}', "");
Expect(1, 1114111, '\P{Noncharacter_Code_Point= NO}', "");
Expect(0, 1114111, '\P{^Noncharacter_Code_Point= NO}', "");
Error('\p{NChar:	:=n}');
Error('\P{NChar:	:=n}');
Expect(1, 1114109, '\p{NChar:   n}', "");
Expect(0, 1114109, '\p{^NChar:   n}', "");
Expect(0, 1114109, '\P{NChar:   n}', "");
Expect(1, 1114109, '\P{^NChar:   n}', "");
Expect(0, 1114111, '\p{NChar:   n}', "");
Expect(1, 1114111, '\p{^NChar:   n}', "");
Expect(1, 1114111, '\P{NChar:   n}', "");
Expect(0, 1114111, '\P{^NChar:   n}', "");
Expect(1, 1114109, '\p{NChar=-N}', "");
Expect(0, 1114109, '\p{^NChar=-N}', "");
Expect(0, 1114109, '\P{NChar=-N}', "");
Expect(1, 1114109, '\P{^NChar=-N}', "");
Expect(0, 1114111, '\p{NChar=-N}', "");
Expect(1, 1114111, '\p{^NChar=-N}', "");
Expect(1, 1114111, '\P{NChar=-N}', "");
Expect(0, 1114111, '\P{^NChar=-N}', "");
Error('\p{Is_Noncharacter_Code_Point=/a/F}');
Error('\P{Is_Noncharacter_Code_Point=/a/F}');
Expect(1, 1114109, '\p{Is_Noncharacter_Code_Point=f}', "");
Expect(0, 1114109, '\p{^Is_Noncharacter_Code_Point=f}', "");
Expect(0, 1114109, '\P{Is_Noncharacter_Code_Point=f}', "");
Expect(1, 1114109, '\P{^Is_Noncharacter_Code_Point=f}', "");
Expect(0, 1114111, '\p{Is_Noncharacter_Code_Point=f}', "");
Expect(1, 1114111, '\p{^Is_Noncharacter_Code_Point=f}', "");
Expect(1, 1114111, '\P{Is_Noncharacter_Code_Point=f}', "");
Expect(0, 1114111, '\P{^Is_Noncharacter_Code_Point=f}', "");
Expect(1, 1114109, '\p{Is_Noncharacter_Code_Point:   _	F}', "");
Expect(0, 1114109, '\p{^Is_Noncharacter_Code_Point:   _	F}', "");
Expect(0, 1114109, '\P{Is_Noncharacter_Code_Point:   _	F}', "");
Expect(1, 1114109, '\P{^Is_Noncharacter_Code_Point:   _	F}', "");
Expect(0, 1114111, '\p{Is_Noncharacter_Code_Point:   _	F}', "");
Expect(1, 1114111, '\p{^Is_Noncharacter_Code_Point:   _	F}', "");
Expect(1, 1114111, '\P{Is_Noncharacter_Code_Point:   _	F}', "");
Expect(0, 1114111, '\P{^Is_Noncharacter_Code_Point:   _	F}', "");
Error('\p{Is_NChar=:=FALSE}');
Error('\P{Is_NChar=:=FALSE}');
Expect(1, 1114109, '\p{Is_NChar=false}', "");
Expect(0, 1114109, '\p{^Is_NChar=false}', "");
Expect(0, 1114109, '\P{Is_NChar=false}', "");
Expect(1, 1114109, '\P{^Is_NChar=false}', "");
Expect(0, 1114111, '\p{Is_NChar=false}', "");
Expect(1, 1114111, '\p{^Is_NChar=false}', "");
Expect(1, 1114111, '\P{Is_NChar=false}', "");
Expect(0, 1114111, '\P{^Is_NChar=false}', "");
Expect(1, 1114109, '\p{Is_NChar:False}', "");
Expect(0, 1114109, '\p{^Is_NChar:False}', "");
Expect(0, 1114109, '\P{Is_NChar:False}', "");
Expect(1, 1114109, '\P{^Is_NChar:False}', "");
Expect(0, 1114111, '\p{Is_NChar:False}', "");
Expect(1, 1114111, '\p{^Is_NChar:False}', "");
Expect(1, 1114111, '\P{Is_NChar:False}', "");
Expect(0, 1114111, '\P{^Is_NChar:False}', "");
Error('\p{Noncharacter_Code_Point=_/a/Yes}');
Error('\P{Noncharacter_Code_Point=_/a/Yes}');
Expect(1, 1114111, '\p{Noncharacter_Code_Point=yes}', "");
Expect(0, 1114111, '\p{^Noncharacter_Code_Point=yes}', "");
Expect(0, 1114111, '\P{Noncharacter_Code_Point=yes}', "");
Expect(1, 1114111, '\P{^Noncharacter_Code_Point=yes}', "");
Expect(0, 1114109, '\p{Noncharacter_Code_Point=yes}', "");
Expect(1, 1114109, '\p{^Noncharacter_Code_Point=yes}', "");
Expect(1, 1114109, '\P{Noncharacter_Code_Point=yes}', "");
Expect(0, 1114109, '\P{^Noncharacter_Code_Point=yes}', "");
Expect(1, 1114111, '\p{Noncharacter_Code_Point=	 YES}', "");
Expect(0, 1114111, '\p{^Noncharacter_Code_Point=	 YES}', "");
Expect(0, 1114111, '\P{Noncharacter_Code_Point=	 YES}', "");
Expect(1, 1114111, '\P{^Noncharacter_Code_Point=	 YES}', "");
Expect(0, 1114109, '\p{Noncharacter_Code_Point=	 YES}', "");
Expect(1, 1114109, '\p{^Noncharacter_Code_Point=	 YES}', "");
Expect(1, 1114109, '\P{Noncharacter_Code_Point=	 YES}', "");
Expect(0, 1114109, '\P{^Noncharacter_Code_Point=	 YES}', "");
Error('\p{NChar=:=--y}');
Error('\P{NChar=:=--y}');
Expect(1, 1114111, '\p{NChar=y}', "");
Expect(0, 1114111, '\p{^NChar=y}', "");
Expect(0, 1114111, '\P{NChar=y}', "");
Expect(1, 1114111, '\P{^NChar=y}', "");
Expect(0, 1114109, '\p{NChar=y}', "");
Expect(1, 1114109, '\p{^NChar=y}', "");
Expect(1, 1114109, '\P{NChar=y}', "");
Expect(0, 1114109, '\P{^NChar=y}', "");
Expect(1, 1114111, '\p{NChar=_	Y}', "");
Expect(0, 1114111, '\p{^NChar=_	Y}', "");
Expect(0, 1114111, '\P{NChar=_	Y}', "");
Expect(1, 1114111, '\P{^NChar=_	Y}', "");
Expect(0, 1114109, '\p{NChar=_	Y}', "");
Expect(1, 1114109, '\p{^NChar=_	Y}', "");
Expect(1, 1114109, '\P{NChar=_	Y}', "");
Expect(0, 1114109, '\P{^NChar=_	Y}', "");
Error('\p{Is_Noncharacter_Code_Point:	/a/-T}');
Error('\P{Is_Noncharacter_Code_Point:	/a/-T}');
Expect(1, 1114111, '\p{Is_Noncharacter_Code_Point:   t}', "");
Expect(0, 1114111, '\p{^Is_Noncharacter_Code_Point:   t}', "");
Expect(0, 1114111, '\P{Is_Noncharacter_Code_Point:   t}', "");
Expect(1, 1114111, '\P{^Is_Noncharacter_Code_Point:   t}', "");
Expect(0, 1114109, '\p{Is_Noncharacter_Code_Point:   t}', "");
Expect(1, 1114109, '\p{^Is_Noncharacter_Code_Point:   t}', "");
Expect(1, 1114109, '\P{Is_Noncharacter_Code_Point:   t}', "");
Expect(0, 1114109, '\P{^Is_Noncharacter_Code_Point:   t}', "");
Expect(1, 1114111, '\p{Is_Noncharacter_Code_Point=-T}', "");
Expect(0, 1114111, '\p{^Is_Noncharacter_Code_Point=-T}', "");
Expect(0, 1114111, '\P{Is_Noncharacter_Code_Point=-T}', "");
Expect(1, 1114111, '\P{^Is_Noncharacter_Code_Point=-T}', "");
Expect(0, 1114109, '\p{Is_Noncharacter_Code_Point=-T}', "");
Expect(1, 1114109, '\p{^Is_Noncharacter_Code_Point=-T}', "");
Expect(1, 1114109, '\P{Is_Noncharacter_Code_Point=-T}', "");
Expect(0, 1114109, '\P{^Is_Noncharacter_Code_Point=-T}', "");
Error('\p{Is_NChar=-:=True}');
Error('\P{Is_NChar=-:=True}');
Expect(1, 1114111, '\p{Is_NChar: true}', "");
Expect(0, 1114111, '\p{^Is_NChar: true}', "");
Expect(0, 1114111, '\P{Is_NChar: true}', "");
Expect(1, 1114111, '\P{^Is_NChar: true}', "");
Expect(0, 1114109, '\p{Is_NChar: true}', "");
Expect(1, 1114109, '\p{^Is_NChar: true}', "");
Expect(1, 1114109, '\P{Is_NChar: true}', "");
Expect(0, 1114109, '\P{^Is_NChar: true}', "");
Expect(1, 1114111, '\p{Is_NChar=_TRUE}', "");
Expect(0, 1114111, '\p{^Is_NChar=_TRUE}', "");
Expect(0, 1114111, '\P{Is_NChar=_TRUE}', "");
Expect(1, 1114111, '\P{^Is_NChar=_TRUE}', "");
Expect(0, 1114109, '\p{Is_NChar=_TRUE}', "");
Expect(1, 1114109, '\p{^Is_NChar=_TRUE}', "");
Expect(1, 1114109, '\P{Is_NChar=_TRUE}', "");
Expect(0, 1114109, '\P{^Is_NChar=_TRUE}', "");
Error('\p{NFC_Quick_Check=- maybe:=}');
Error('\P{NFC_Quick_Check=- maybe:=}');
Expect(1, 69927, '\p{NFC_Quick_Check=maybe}', "");
Expect(0, 69927, '\p{^NFC_Quick_Check=maybe}', "");
Expect(0, 69927, '\P{NFC_Quick_Check=maybe}', "");
Expect(1, 69927, '\P{^NFC_Quick_Check=maybe}', "");
Expect(0, 69928, '\p{NFC_Quick_Check=maybe}', "");
Expect(1, 69928, '\p{^NFC_Quick_Check=maybe}', "");
Expect(1, 69928, '\P{NFC_Quick_Check=maybe}', "");
Expect(0, 69928, '\P{^NFC_Quick_Check=maybe}', "");
Expect(1, 69927, '\p{NFC_Quick_Check=	maybe}', "");
Expect(0, 69927, '\p{^NFC_Quick_Check=	maybe}', "");
Expect(0, 69927, '\P{NFC_Quick_Check=	maybe}', "");
Expect(1, 69927, '\P{^NFC_Quick_Check=	maybe}', "");
Expect(0, 69928, '\p{NFC_Quick_Check=	maybe}', "");
Expect(1, 69928, '\p{^NFC_Quick_Check=	maybe}', "");
Expect(1, 69928, '\P{NFC_Quick_Check=	maybe}', "");
Expect(0, 69928, '\P{^NFC_Quick_Check=	maybe}', "");
Error('\p{NFC_QC=-M:=}');
Error('\P{NFC_QC=-M:=}');
Expect(1, 69927, '\p{NFC_QC=m}', "");
Expect(0, 69927, '\p{^NFC_QC=m}', "");
Expect(0, 69927, '\P{NFC_QC=m}', "");
Expect(1, 69927, '\P{^NFC_QC=m}', "");
Expect(0, 69928, '\p{NFC_QC=m}', "");
Expect(1, 69928, '\p{^NFC_QC=m}', "");
Expect(1, 69928, '\P{NFC_QC=m}', "");
Expect(0, 69928, '\P{^NFC_QC=m}', "");
Expect(1, 69927, '\p{NFC_QC=-_M}', "");
Expect(0, 69927, '\p{^NFC_QC=-_M}', "");
Expect(0, 69927, '\P{NFC_QC=-_M}', "");
Expect(1, 69927, '\P{^NFC_QC=-_M}', "");
Expect(0, 69928, '\p{NFC_QC=-_M}', "");
Expect(1, 69928, '\p{^NFC_QC=-_M}', "");
Expect(1, 69928, '\P{NFC_QC=-_M}', "");
Expect(0, 69928, '\P{^NFC_QC=-_M}', "");
Error('\p{Is_NFC_Quick_Check=:=_	maybe}');
Error('\P{Is_NFC_Quick_Check=:=_	maybe}');
Expect(1, 69927, '\p{Is_NFC_Quick_Check=maybe}', "");
Expect(0, 69927, '\p{^Is_NFC_Quick_Check=maybe}', "");
Expect(0, 69927, '\P{Is_NFC_Quick_Check=maybe}', "");
Expect(1, 69927, '\P{^Is_NFC_Quick_Check=maybe}', "");
Expect(0, 69928, '\p{Is_NFC_Quick_Check=maybe}', "");
Expect(1, 69928, '\p{^Is_NFC_Quick_Check=maybe}', "");
Expect(1, 69928, '\P{Is_NFC_Quick_Check=maybe}', "");
Expect(0, 69928, '\P{^Is_NFC_Quick_Check=maybe}', "");
Expect(1, 69927, '\p{Is_NFC_Quick_Check=	maybe}', "");
Expect(0, 69927, '\p{^Is_NFC_Quick_Check=	maybe}', "");
Expect(0, 69927, '\P{Is_NFC_Quick_Check=	maybe}', "");
Expect(1, 69927, '\P{^Is_NFC_Quick_Check=	maybe}', "");
Expect(0, 69928, '\p{Is_NFC_Quick_Check=	maybe}', "");
Expect(1, 69928, '\p{^Is_NFC_Quick_Check=	maybe}', "");
Expect(1, 69928, '\P{Is_NFC_Quick_Check=	maybe}', "");
Expect(0, 69928, '\P{^Is_NFC_Quick_Check=	maybe}', "");
Error('\p{Is_NFC_QC=_/a/m}');
Error('\P{Is_NFC_QC=_/a/m}');
Expect(1, 69927, '\p{Is_NFC_QC:   m}', "");
Expect(0, 69927, '\p{^Is_NFC_QC:   m}', "");
Expect(0, 69927, '\P{Is_NFC_QC:   m}', "");
Expect(1, 69927, '\P{^Is_NFC_QC:   m}', "");
Expect(0, 69928, '\p{Is_NFC_QC:   m}', "");
Expect(1, 69928, '\p{^Is_NFC_QC:   m}', "");
Expect(1, 69928, '\P{Is_NFC_QC:   m}', "");
Expect(0, 69928, '\P{^Is_NFC_QC:   m}', "");
Expect(1, 69927, '\p{Is_NFC_QC=_	m}', "");
Expect(0, 69927, '\p{^Is_NFC_QC=_	m}', "");
Expect(0, 69927, '\P{Is_NFC_QC=_	m}', "");
Expect(1, 69927, '\P{^Is_NFC_QC=_	m}', "");
Expect(0, 69928, '\p{Is_NFC_QC=_	m}', "");
Expect(1, 69928, '\p{^Is_NFC_QC=_	m}', "");
Expect(1, 69928, '\P{Is_NFC_QC=_	m}', "");
Expect(0, 69928, '\P{^Is_NFC_QC=_	m}', "");
Error('\p{NFC_Quick_Check=No/a/}');
Error('\P{NFC_Quick_Check=No/a/}');
Expect(1, 195101, '\p{NFC_Quick_Check:   no}', "");
Expect(0, 195101, '\p{^NFC_Quick_Check:   no}', "");
Expect(0, 195101, '\P{NFC_Quick_Check:   no}', "");
Expect(1, 195101, '\P{^NFC_Quick_Check:   no}', "");
Expect(0, 195102, '\p{NFC_Quick_Check:   no}', "");
Expect(1, 195102, '\p{^NFC_Quick_Check:   no}', "");
Expect(1, 195102, '\P{NFC_Quick_Check:   no}', "");
Expect(0, 195102, '\P{^NFC_Quick_Check:   no}', "");
Expect(1, 195101, '\p{NFC_Quick_Check=	 NO}', "");
Expect(0, 195101, '\p{^NFC_Quick_Check=	 NO}', "");
Expect(0, 195101, '\P{NFC_Quick_Check=	 NO}', "");
Expect(1, 195101, '\P{^NFC_Quick_Check=	 NO}', "");
Expect(0, 195102, '\p{NFC_Quick_Check=	 NO}', "");
Expect(1, 195102, '\p{^NFC_Quick_Check=	 NO}', "");
Expect(1, 195102, '\P{NFC_Quick_Check=	 NO}', "");
Expect(0, 195102, '\P{^NFC_Quick_Check=	 NO}', "");
Error('\p{NFC_QC=	-N/a/}');
Error('\P{NFC_QC=	-N/a/}');
Expect(1, 195101, '\p{NFC_QC=n}', "");
Expect(0, 195101, '\p{^NFC_QC=n}', "");
Expect(0, 195101, '\P{NFC_QC=n}', "");
Expect(1, 195101, '\P{^NFC_QC=n}', "");
Expect(0, 195102, '\p{NFC_QC=n}', "");
Expect(1, 195102, '\p{^NFC_QC=n}', "");
Expect(1, 195102, '\P{NFC_QC=n}', "");
Expect(0, 195102, '\P{^NFC_QC=n}', "");
Expect(1, 195101, '\p{NFC_QC=_n}', "");
Expect(0, 195101, '\p{^NFC_QC=_n}', "");
Expect(0, 195101, '\P{NFC_QC=_n}', "");
Expect(1, 195101, '\P{^NFC_QC=_n}', "");
Expect(0, 195102, '\p{NFC_QC=_n}', "");
Expect(1, 195102, '\p{^NFC_QC=_n}', "");
Expect(1, 195102, '\P{NFC_QC=_n}', "");
Expect(0, 195102, '\P{^NFC_QC=_n}', "");
Error('\p{Is_NFC_Quick_Check=:=NO}');
Error('\P{Is_NFC_Quick_Check=:=NO}');
Expect(1, 195101, '\p{Is_NFC_Quick_Check=no}', "");
Expect(0, 195101, '\p{^Is_NFC_Quick_Check=no}', "");
Expect(0, 195101, '\P{Is_NFC_Quick_Check=no}', "");
Expect(1, 195101, '\P{^Is_NFC_Quick_Check=no}', "");
Expect(0, 195102, '\p{Is_NFC_Quick_Check=no}', "");
Expect(1, 195102, '\p{^Is_NFC_Quick_Check=no}', "");
Expect(1, 195102, '\P{Is_NFC_Quick_Check=no}', "");
Expect(0, 195102, '\P{^Is_NFC_Quick_Check=no}', "");
Expect(1, 195101, '\p{Is_NFC_Quick_Check= No}', "");
Expect(0, 195101, '\p{^Is_NFC_Quick_Check= No}', "");
Expect(0, 195101, '\P{Is_NFC_Quick_Check= No}', "");
Expect(1, 195101, '\P{^Is_NFC_Quick_Check= No}', "");
Expect(0, 195102, '\p{Is_NFC_Quick_Check= No}', "");
Expect(1, 195102, '\p{^Is_NFC_Quick_Check= No}', "");
Expect(1, 195102, '\P{Is_NFC_Quick_Check= No}', "");
Expect(0, 195102, '\P{^Is_NFC_Quick_Check= No}', "");
Error('\p{Is_NFC_QC:   N/a/}');
Error('\P{Is_NFC_QC:   N/a/}');
Expect(1, 195101, '\p{Is_NFC_QC=n}', "");
Expect(0, 195101, '\p{^Is_NFC_QC=n}', "");
Expect(0, 195101, '\P{Is_NFC_QC=n}', "");
Expect(1, 195101, '\P{^Is_NFC_QC=n}', "");
Expect(0, 195102, '\p{Is_NFC_QC=n}', "");
Expect(1, 195102, '\p{^Is_NFC_QC=n}', "");
Expect(1, 195102, '\P{Is_NFC_QC=n}', "");
Expect(0, 195102, '\P{^Is_NFC_QC=n}', "");
Expect(1, 195101, '\p{Is_NFC_QC=--N}', "");
Expect(0, 195101, '\p{^Is_NFC_QC=--N}', "");
Expect(0, 195101, '\P{Is_NFC_QC=--N}', "");
Expect(1, 195101, '\P{^Is_NFC_QC=--N}', "");
Expect(0, 195102, '\p{Is_NFC_QC=--N}', "");
Expect(1, 195102, '\p{^Is_NFC_QC=--N}', "");
Expect(1, 195102, '\P{Is_NFC_QC=--N}', "");
Expect(0, 195102, '\P{^Is_NFC_QC=--N}', "");
Error('\p{NFC_Quick_Check=-YES:=}');
Error('\P{NFC_Quick_Check=-YES:=}');
Expect(1, 195102, '\p{NFC_Quick_Check=yes}', "");
Expect(0, 195102, '\p{^NFC_Quick_Check=yes}', "");
Expect(0, 195102, '\P{NFC_Quick_Check=yes}', "");
Expect(1, 195102, '\P{^NFC_Quick_Check=yes}', "");
Expect(0, 195101, '\p{NFC_Quick_Check=yes}', "");
Expect(1, 195101, '\p{^NFC_Quick_Check=yes}', "");
Expect(1, 195101, '\P{NFC_Quick_Check=yes}', "");
Expect(0, 195101, '\P{^NFC_Quick_Check=yes}', "");
Expect(1, 195102, '\p{NFC_Quick_Check=	_YES}', "");
Expect(0, 195102, '\p{^NFC_Quick_Check=	_YES}', "");
Expect(0, 195102, '\P{NFC_Quick_Check=	_YES}', "");
Expect(1, 195102, '\P{^NFC_Quick_Check=	_YES}', "");
Expect(0, 195101, '\p{NFC_Quick_Check=	_YES}', "");
Expect(1, 195101, '\p{^NFC_Quick_Check=	_YES}', "");
Expect(1, 195101, '\P{NFC_Quick_Check=	_YES}', "");
Expect(0, 195101, '\P{^NFC_Quick_Check=	_YES}', "");
Error('\p{NFC_QC:   --Y:=}');
Error('\P{NFC_QC:   --Y:=}');
Expect(1, 195102, '\p{NFC_QC=y}', "");
Expect(0, 195102, '\p{^NFC_QC=y}', "");
Expect(0, 195102, '\P{NFC_QC=y}', "");
Expect(1, 195102, '\P{^NFC_QC=y}', "");
Expect(0, 195101, '\p{NFC_QC=y}', "");
Expect(1, 195101, '\p{^NFC_QC=y}', "");
Expect(1, 195101, '\P{NFC_QC=y}', "");
Expect(0, 195101, '\P{^NFC_QC=y}', "");
Expect(1, 195102, '\p{NFC_QC=	Y}', "");
Expect(0, 195102, '\p{^NFC_QC=	Y}', "");
Expect(0, 195102, '\P{NFC_QC=	Y}', "");
Expect(1, 195102, '\P{^NFC_QC=	Y}', "");
Expect(0, 195101, '\p{NFC_QC=	Y}', "");
Expect(1, 195101, '\p{^NFC_QC=	Y}', "");
Expect(1, 195101, '\P{NFC_QC=	Y}', "");
Expect(0, 195101, '\P{^NFC_QC=	Y}', "");
Error('\p{Is_NFC_Quick_Check=	 Yes/a/}');
Error('\P{Is_NFC_Quick_Check=	 Yes/a/}');
Expect(1, 195102, '\p{Is_NFC_Quick_Check=yes}', "");
Expect(0, 195102, '\p{^Is_NFC_Quick_Check=yes}', "");
Expect(0, 195102, '\P{Is_NFC_Quick_Check=yes}', "");
Expect(1, 195102, '\P{^Is_NFC_Quick_Check=yes}', "");
Expect(0, 195101, '\p{Is_NFC_Quick_Check=yes}', "");
Expect(1, 195101, '\p{^Is_NFC_Quick_Check=yes}', "");
Expect(1, 195101, '\P{Is_NFC_Quick_Check=yes}', "");
Expect(0, 195101, '\P{^Is_NFC_Quick_Check=yes}', "");
Expect(1, 195102, '\p{Is_NFC_Quick_Check=_Yes}', "");
Expect(0, 195102, '\p{^Is_NFC_Quick_Check=_Yes}', "");
Expect(0, 195102, '\P{Is_NFC_Quick_Check=_Yes}', "");
Expect(1, 195102, '\P{^Is_NFC_Quick_Check=_Yes}', "");
Expect(0, 195101, '\p{Is_NFC_Quick_Check=_Yes}', "");
Expect(1, 195101, '\p{^Is_NFC_Quick_Check=_Yes}', "");
Expect(1, 195101, '\P{Is_NFC_Quick_Check=_Yes}', "");
Expect(0, 195101, '\P{^Is_NFC_Quick_Check=_Yes}', "");
Error('\p{Is_NFC_QC=_ Y:=}');
Error('\P{Is_NFC_QC=_ Y:=}');
Expect(1, 195102, '\p{Is_NFC_QC:	y}', "");
Expect(0, 195102, '\p{^Is_NFC_QC:	y}', "");
Expect(0, 195102, '\P{Is_NFC_QC:	y}', "");
Expect(1, 195102, '\P{^Is_NFC_QC:	y}', "");
Expect(0, 195101, '\p{Is_NFC_QC:	y}', "");
Expect(1, 195101, '\p{^Is_NFC_QC:	y}', "");
Expect(1, 195101, '\P{Is_NFC_QC:	y}', "");
Expect(0, 195101, '\P{^Is_NFC_QC:	y}', "");
Expect(1, 195102, '\p{Is_NFC_QC=-	Y}', "");
Expect(0, 195102, '\p{^Is_NFC_QC=-	Y}', "");
Expect(0, 195102, '\P{Is_NFC_QC=-	Y}', "");
Expect(1, 195102, '\P{^Is_NFC_QC=-	Y}', "");
Expect(0, 195101, '\p{Is_NFC_QC=-	Y}', "");
Expect(1, 195101, '\p{^Is_NFC_QC=-	Y}', "");
Expect(1, 195101, '\P{Is_NFC_QC=-	Y}', "");
Expect(0, 195101, '\P{^Is_NFC_QC=-	Y}', "");
Error('\p{NFD_Quick_Check: -no:=}');
Error('\P{NFD_Quick_Check: -no:=}');
Expect(1, 195101, '\p{NFD_Quick_Check=no}', "");
Expect(0, 195101, '\p{^NFD_Quick_Check=no}', "");
Expect(0, 195101, '\P{NFD_Quick_Check=no}', "");
Expect(1, 195101, '\P{^NFD_Quick_Check=no}', "");
Expect(0, 195102, '\p{NFD_Quick_Check=no}', "");
Expect(1, 195102, '\p{^NFD_Quick_Check=no}', "");
Expect(1, 195102, '\P{NFD_Quick_Check=no}', "");
Expect(0, 195102, '\P{^NFD_Quick_Check=no}', "");
Expect(1, 195101, '\p{NFD_Quick_Check=	No}', "");
Expect(0, 195101, '\p{^NFD_Quick_Check=	No}', "");
Expect(0, 195101, '\P{NFD_Quick_Check=	No}', "");
Expect(1, 195101, '\P{^NFD_Quick_Check=	No}', "");
Expect(0, 195102, '\p{NFD_Quick_Check=	No}', "");
Expect(1, 195102, '\p{^NFD_Quick_Check=	No}', "");
Expect(1, 195102, '\P{NFD_Quick_Check=	No}', "");
Expect(0, 195102, '\P{^NFD_Quick_Check=	No}', "");
Error('\p{NFD_QC=:=n}');
Error('\P{NFD_QC=:=n}');
Expect(1, 195101, '\p{NFD_QC:n}', "");
Expect(0, 195101, '\p{^NFD_QC:n}', "");
Expect(0, 195101, '\P{NFD_QC:n}', "");
Expect(1, 195101, '\P{^NFD_QC:n}', "");
Expect(0, 195102, '\p{NFD_QC:n}', "");
Expect(1, 195102, '\p{^NFD_QC:n}', "");
Expect(1, 195102, '\P{NFD_QC:n}', "");
Expect(0, 195102, '\P{^NFD_QC:n}', "");
Expect(1, 195101, '\p{NFD_QC=	N}', "");
Expect(0, 195101, '\p{^NFD_QC=	N}', "");
Expect(0, 195101, '\P{NFD_QC=	N}', "");
Expect(1, 195101, '\P{^NFD_QC=	N}', "");
Expect(0, 195102, '\p{NFD_QC=	N}', "");
Expect(1, 195102, '\p{^NFD_QC=	N}', "");
Expect(1, 195102, '\P{NFD_QC=	N}', "");
Expect(0, 195102, '\P{^NFD_QC=	N}', "");
Error('\p{Is_NFD_Quick_Check: No:=}');
Error('\P{Is_NFD_Quick_Check: No:=}');
Expect(1, 195101, '\p{Is_NFD_Quick_Check=no}', "");
Expect(0, 195101, '\p{^Is_NFD_Quick_Check=no}', "");
Expect(0, 195101, '\P{Is_NFD_Quick_Check=no}', "");
Expect(1, 195101, '\P{^Is_NFD_Quick_Check=no}', "");
Expect(0, 195102, '\p{Is_NFD_Quick_Check=no}', "");
Expect(1, 195102, '\p{^Is_NFD_Quick_Check=no}', "");
Expect(1, 195102, '\P{Is_NFD_Quick_Check=no}', "");
Expect(0, 195102, '\P{^Is_NFD_Quick_Check=no}', "");
Expect(1, 195101, '\p{Is_NFD_Quick_Check=_NO}', "");
Expect(0, 195101, '\p{^Is_NFD_Quick_Check=_NO}', "");
Expect(0, 195101, '\P{Is_NFD_Quick_Check=_NO}', "");
Expect(1, 195101, '\P{^Is_NFD_Quick_Check=_NO}', "");
Expect(0, 195102, '\p{Is_NFD_Quick_Check=_NO}', "");
Expect(1, 195102, '\p{^Is_NFD_Quick_Check=_NO}', "");
Expect(1, 195102, '\P{Is_NFD_Quick_Check=_NO}', "");
Expect(0, 195102, '\P{^Is_NFD_Quick_Check=_NO}', "");
Error('\p{Is_NFD_QC= :=n}');
Error('\P{Is_NFD_QC= :=n}');
Expect(1, 195101, '\p{Is_NFD_QC=n}', "");
Expect(0, 195101, '\p{^Is_NFD_QC=n}', "");
Expect(0, 195101, '\P{Is_NFD_QC=n}', "");
Expect(1, 195101, '\P{^Is_NFD_QC=n}', "");
Expect(0, 195102, '\p{Is_NFD_QC=n}', "");
Expect(1, 195102, '\p{^Is_NFD_QC=n}', "");
Expect(1, 195102, '\P{Is_NFD_QC=n}', "");
Expect(0, 195102, '\P{^Is_NFD_QC=n}', "");
Expect(1, 195101, '\p{Is_NFD_QC= 	n}', "");
Expect(0, 195101, '\p{^Is_NFD_QC= 	n}', "");
Expect(0, 195101, '\P{Is_NFD_QC= 	n}', "");
Expect(1, 195101, '\P{^Is_NFD_QC= 	n}', "");
Expect(0, 195102, '\p{Is_NFD_QC= 	n}', "");
Expect(1, 195102, '\p{^Is_NFD_QC= 	n}', "");
Expect(1, 195102, '\P{Is_NFD_QC= 	n}', "");
Expect(0, 195102, '\P{^Is_NFD_QC= 	n}', "");
Error('\p{NFD_Quick_Check=_-Yes:=}');
Error('\P{NFD_Quick_Check=_-Yes:=}');
Expect(1, 195102, '\p{NFD_Quick_Check=yes}', "");
Expect(0, 195102, '\p{^NFD_Quick_Check=yes}', "");
Expect(0, 195102, '\P{NFD_Quick_Check=yes}', "");
Expect(1, 195102, '\P{^NFD_Quick_Check=yes}', "");
Expect(0, 195101, '\p{NFD_Quick_Check=yes}', "");
Expect(1, 195101, '\p{^NFD_Quick_Check=yes}', "");
Expect(1, 195101, '\P{NFD_Quick_Check=yes}', "");
Expect(0, 195101, '\P{^NFD_Quick_Check=yes}', "");
Expect(1, 195102, '\p{NFD_Quick_Check=YES}', "");
Expect(0, 195102, '\p{^NFD_Quick_Check=YES}', "");
Expect(0, 195102, '\P{NFD_Quick_Check=YES}', "");
Expect(1, 195102, '\P{^NFD_Quick_Check=YES}', "");
Expect(0, 195101, '\p{NFD_Quick_Check=YES}', "");
Expect(1, 195101, '\p{^NFD_Quick_Check=YES}', "");
Expect(1, 195101, '\P{NFD_Quick_Check=YES}', "");
Expect(0, 195101, '\P{^NFD_Quick_Check=YES}', "");
Error('\p{NFD_QC=_ Y:=}');
Error('\P{NFD_QC=_ Y:=}');
Expect(1, 195102, '\p{NFD_QC=y}', "");
Expect(0, 195102, '\p{^NFD_QC=y}', "");
Expect(0, 195102, '\P{NFD_QC=y}', "");
Expect(1, 195102, '\P{^NFD_QC=y}', "");
Expect(0, 195101, '\p{NFD_QC=y}', "");
Expect(1, 195101, '\p{^NFD_QC=y}', "");
Expect(1, 195101, '\P{NFD_QC=y}', "");
Expect(0, 195101, '\P{^NFD_QC=y}', "");
Expect(1, 195102, '\p{NFD_QC= -Y}', "");
Expect(0, 195102, '\p{^NFD_QC= -Y}', "");
Expect(0, 195102, '\P{NFD_QC= -Y}', "");
Expect(1, 195102, '\P{^NFD_QC= -Y}', "");
Expect(0, 195101, '\p{NFD_QC= -Y}', "");
Expect(1, 195101, '\p{^NFD_QC= -Y}', "");
Expect(1, 195101, '\P{NFD_QC= -Y}', "");
Expect(0, 195101, '\P{^NFD_QC= -Y}', "");
Error('\p{Is_NFD_Quick_Check=-/a/YES}');
Error('\P{Is_NFD_Quick_Check=-/a/YES}');
Expect(1, 195102, '\p{Is_NFD_Quick_Check=yes}', "");
Expect(0, 195102, '\p{^Is_NFD_Quick_Check=yes}', "");
Expect(0, 195102, '\P{Is_NFD_Quick_Check=yes}', "");
Expect(1, 195102, '\P{^Is_NFD_Quick_Check=yes}', "");
Expect(0, 195101, '\p{Is_NFD_Quick_Check=yes}', "");
Expect(1, 195101, '\p{^Is_NFD_Quick_Check=yes}', "");
Expect(1, 195101, '\P{Is_NFD_Quick_Check=yes}', "");
Expect(0, 195101, '\P{^Is_NFD_Quick_Check=yes}', "");
Expect(1, 195102, '\p{Is_NFD_Quick_Check=-Yes}', "");
Expect(0, 195102, '\p{^Is_NFD_Quick_Check=-Yes}', "");
Expect(0, 195102, '\P{Is_NFD_Quick_Check=-Yes}', "");
Expect(1, 195102, '\P{^Is_NFD_Quick_Check=-Yes}', "");
Expect(0, 195101, '\p{Is_NFD_Quick_Check=-Yes}', "");
Expect(1, 195101, '\p{^Is_NFD_Quick_Check=-Yes}', "");
Expect(1, 195101, '\P{Is_NFD_Quick_Check=-Yes}', "");
Expect(0, 195101, '\P{^Is_NFD_Quick_Check=-Yes}', "");
Error('\p{Is_NFD_QC=_-Y:=}');
Error('\P{Is_NFD_QC=_-Y:=}');
Expect(1, 195102, '\p{Is_NFD_QC=y}', "");
Expect(0, 195102, '\p{^Is_NFD_QC=y}', "");
Expect(0, 195102, '\P{Is_NFD_QC=y}', "");
Expect(1, 195102, '\P{^Is_NFD_QC=y}', "");
Expect(0, 195101, '\p{Is_NFD_QC=y}', "");
Expect(1, 195101, '\p{^Is_NFD_QC=y}', "");
Expect(1, 195101, '\P{Is_NFD_QC=y}', "");
Expect(0, 195101, '\P{^Is_NFD_QC=y}', "");
Expect(1, 195102, '\p{Is_NFD_QC: -Y}', "");
Expect(0, 195102, '\p{^Is_NFD_QC: -Y}', "");
Expect(0, 195102, '\P{Is_NFD_QC: -Y}', "");
Expect(1, 195102, '\P{^Is_NFD_QC: -Y}', "");
Expect(0, 195101, '\p{Is_NFD_QC: -Y}', "");
Expect(1, 195101, '\p{^Is_NFD_QC: -Y}', "");
Expect(1, 195101, '\P{Is_NFD_QC: -Y}', "");
Expect(0, 195101, '\P{^Is_NFD_QC: -Y}', "");
Error('\p{NFKC_Quick_Check:   /a/_ maybe}');
Error('\P{NFKC_Quick_Check:   /a/_ maybe}');
Expect(1, 69927, '\p{NFKC_Quick_Check=maybe}', "");
Expect(0, 69927, '\p{^NFKC_Quick_Check=maybe}', "");
Expect(0, 69927, '\P{NFKC_Quick_Check=maybe}', "");
Expect(1, 69927, '\P{^NFKC_Quick_Check=maybe}', "");
Expect(0, 69928, '\p{NFKC_Quick_Check=maybe}', "");
Expect(1, 69928, '\p{^NFKC_Quick_Check=maybe}', "");
Expect(1, 69928, '\P{NFKC_Quick_Check=maybe}', "");
Expect(0, 69928, '\P{^NFKC_Quick_Check=maybe}', "");
Expect(1, 69927, '\p{NFKC_Quick_Check=  Maybe}', "");
Expect(0, 69927, '\p{^NFKC_Quick_Check=  Maybe}', "");
Expect(0, 69927, '\P{NFKC_Quick_Check=  Maybe}', "");
Expect(1, 69927, '\P{^NFKC_Quick_Check=  Maybe}', "");
Expect(0, 69928, '\p{NFKC_Quick_Check=  Maybe}', "");
Expect(1, 69928, '\p{^NFKC_Quick_Check=  Maybe}', "");
Expect(1, 69928, '\P{NFKC_Quick_Check=  Maybe}', "");
Expect(0, 69928, '\P{^NFKC_Quick_Check=  Maybe}', "");
Error('\p{NFKC_QC=:=M}');
Error('\P{NFKC_QC=:=M}');
Expect(1, 69927, '\p{NFKC_QC=m}', "");
Expect(0, 69927, '\p{^NFKC_QC=m}', "");
Expect(0, 69927, '\P{NFKC_QC=m}', "");
Expect(1, 69927, '\P{^NFKC_QC=m}', "");
Expect(0, 69928, '\p{NFKC_QC=m}', "");
Expect(1, 69928, '\p{^NFKC_QC=m}', "");
Expect(1, 69928, '\P{NFKC_QC=m}', "");
Expect(0, 69928, '\P{^NFKC_QC=m}', "");
Expect(1, 69927, '\p{NFKC_QC=  m}', "");
Expect(0, 69927, '\p{^NFKC_QC=  m}', "");
Expect(0, 69927, '\P{NFKC_QC=  m}', "");
Expect(1, 69927, '\P{^NFKC_QC=  m}', "");
Expect(0, 69928, '\p{NFKC_QC=  m}', "");
Expect(1, 69928, '\p{^NFKC_QC=  m}', "");
Expect(1, 69928, '\P{NFKC_QC=  m}', "");
Expect(0, 69928, '\P{^NFKC_QC=  m}', "");
Error('\p{Is_NFKC_Quick_Check:	 maybe:=}');
Error('\P{Is_NFKC_Quick_Check:	 maybe:=}');
Expect(1, 69927, '\p{Is_NFKC_Quick_Check=maybe}', "");
Expect(0, 69927, '\p{^Is_NFKC_Quick_Check=maybe}', "");
Expect(0, 69927, '\P{Is_NFKC_Quick_Check=maybe}', "");
Expect(1, 69927, '\P{^Is_NFKC_Quick_Check=maybe}', "");
Expect(0, 69928, '\p{Is_NFKC_Quick_Check=maybe}', "");
Expect(1, 69928, '\p{^Is_NFKC_Quick_Check=maybe}', "");
Expect(1, 69928, '\P{Is_NFKC_Quick_Check=maybe}', "");
Expect(0, 69928, '\P{^Is_NFKC_Quick_Check=maybe}', "");
Expect(1, 69927, '\p{Is_NFKC_Quick_Check=	_Maybe}', "");
Expect(0, 69927, '\p{^Is_NFKC_Quick_Check=	_Maybe}', "");
Expect(0, 69927, '\P{Is_NFKC_Quick_Check=	_Maybe}', "");
Expect(1, 69927, '\P{^Is_NFKC_Quick_Check=	_Maybe}', "");
Expect(0, 69928, '\p{Is_NFKC_Quick_Check=	_Maybe}', "");
Expect(1, 69928, '\p{^Is_NFKC_Quick_Check=	_Maybe}', "");
Expect(1, 69928, '\P{Is_NFKC_Quick_Check=	_Maybe}', "");
Expect(0, 69928, '\P{^Is_NFKC_Quick_Check=	_Maybe}', "");
Error('\p{Is_NFKC_QC=-	m:=}');
Error('\P{Is_NFKC_QC=-	m:=}');
Expect(1, 69927, '\p{Is_NFKC_QC=m}', "");
Expect(0, 69927, '\p{^Is_NFKC_QC=m}', "");
Expect(0, 69927, '\P{Is_NFKC_QC=m}', "");
Expect(1, 69927, '\P{^Is_NFKC_QC=m}', "");
Expect(0, 69928, '\p{Is_NFKC_QC=m}', "");
Expect(1, 69928, '\p{^Is_NFKC_QC=m}', "");
Expect(1, 69928, '\P{Is_NFKC_QC=m}', "");
Expect(0, 69928, '\P{^Is_NFKC_QC=m}', "");
Expect(1, 69927, '\p{Is_NFKC_QC:-M}', "");
Expect(0, 69927, '\p{^Is_NFKC_QC:-M}', "");
Expect(0, 69927, '\P{Is_NFKC_QC:-M}', "");
Expect(1, 69927, '\P{^Is_NFKC_QC:-M}', "");
Expect(0, 69928, '\p{Is_NFKC_QC:-M}', "");
Expect(1, 69928, '\p{^Is_NFKC_QC:-M}', "");
Expect(1, 69928, '\P{Is_NFKC_QC:-M}', "");
Expect(0, 69928, '\P{^Is_NFKC_QC:-M}', "");
Error('\p{NFKC_Quick_Check=:=-	NO}');
Error('\P{NFKC_Quick_Check=:=-	NO}');
Expect(1, 195101, '\p{NFKC_Quick_Check=no}', "");
Expect(0, 195101, '\p{^NFKC_Quick_Check=no}', "");
Expect(0, 195101, '\P{NFKC_Quick_Check=no}', "");
Expect(1, 195101, '\P{^NFKC_Quick_Check=no}', "");
Expect(0, 195102, '\p{NFKC_Quick_Check=no}', "");
Expect(1, 195102, '\p{^NFKC_Quick_Check=no}', "");
Expect(1, 195102, '\P{NFKC_Quick_Check=no}', "");
Expect(0, 195102, '\P{^NFKC_Quick_Check=no}', "");
Expect(1, 195101, '\p{NFKC_Quick_Check=_	NO}', "");
Expect(0, 195101, '\p{^NFKC_Quick_Check=_	NO}', "");
Expect(0, 195101, '\P{NFKC_Quick_Check=_	NO}', "");
Expect(1, 195101, '\P{^NFKC_Quick_Check=_	NO}', "");
Expect(0, 195102, '\p{NFKC_Quick_Check=_	NO}', "");
Expect(1, 195102, '\p{^NFKC_Quick_Check=_	NO}', "");
Expect(1, 195102, '\P{NFKC_Quick_Check=_	NO}', "");
Expect(0, 195102, '\P{^NFKC_Quick_Check=_	NO}', "");
Error('\p{NFKC_QC=_n/a/}');
Error('\P{NFKC_QC=_n/a/}');
Expect(1, 195101, '\p{NFKC_QC=n}', "");
Expect(0, 195101, '\p{^NFKC_QC=n}', "");
Expect(0, 195101, '\P{NFKC_QC=n}', "");
Expect(1, 195101, '\P{^NFKC_QC=n}', "");
Expect(0, 195102, '\p{NFKC_QC=n}', "");
Expect(1, 195102, '\p{^NFKC_QC=n}', "");
Expect(1, 195102, '\P{NFKC_QC=n}', "");
Expect(0, 195102, '\P{^NFKC_QC=n}', "");
Expect(1, 195101, '\p{NFKC_QC= 	N}', "");
Expect(0, 195101, '\p{^NFKC_QC= 	N}', "");
Expect(0, 195101, '\P{NFKC_QC= 	N}', "");
Expect(1, 195101, '\P{^NFKC_QC= 	N}', "");
Expect(0, 195102, '\p{NFKC_QC= 	N}', "");
Expect(1, 195102, '\p{^NFKC_QC= 	N}', "");
Expect(1, 195102, '\P{NFKC_QC= 	N}', "");
Expect(0, 195102, '\P{^NFKC_QC= 	N}', "");
Error('\p{Is_NFKC_Quick_Check=:=No}');
Error('\P{Is_NFKC_Quick_Check=:=No}');
Expect(1, 195101, '\p{Is_NFKC_Quick_Check=no}', "");
Expect(0, 195101, '\p{^Is_NFKC_Quick_Check=no}', "");
Expect(0, 195101, '\P{Is_NFKC_Quick_Check=no}', "");
Expect(1, 195101, '\P{^Is_NFKC_Quick_Check=no}', "");
Expect(0, 195102, '\p{Is_NFKC_Quick_Check=no}', "");
Expect(1, 195102, '\p{^Is_NFKC_Quick_Check=no}', "");
Expect(1, 195102, '\P{Is_NFKC_Quick_Check=no}', "");
Expect(0, 195102, '\P{^Is_NFKC_Quick_Check=no}', "");
Expect(1, 195101, '\p{Is_NFKC_Quick_Check=	No}', "");
Expect(0, 195101, '\p{^Is_NFKC_Quick_Check=	No}', "");
Expect(0, 195101, '\P{Is_NFKC_Quick_Check=	No}', "");
Expect(1, 195101, '\P{^Is_NFKC_Quick_Check=	No}', "");
Expect(0, 195102, '\p{Is_NFKC_Quick_Check=	No}', "");
Expect(1, 195102, '\p{^Is_NFKC_Quick_Check=	No}', "");
Expect(1, 195102, '\P{Is_NFKC_Quick_Check=	No}', "");
Expect(0, 195102, '\P{^Is_NFKC_Quick_Check=	No}', "");
Error('\p{Is_NFKC_QC=_N/a/}');
Error('\P{Is_NFKC_QC=_N/a/}');
Expect(1, 195101, '\p{Is_NFKC_QC=n}', "");
Expect(0, 195101, '\p{^Is_NFKC_QC=n}', "");
Expect(0, 195101, '\P{Is_NFKC_QC=n}', "");
Expect(1, 195101, '\P{^Is_NFKC_QC=n}', "");
Expect(0, 195102, '\p{Is_NFKC_QC=n}', "");
Expect(1, 195102, '\p{^Is_NFKC_QC=n}', "");
Expect(1, 195102, '\P{Is_NFKC_QC=n}', "");
Expect(0, 195102, '\P{^Is_NFKC_QC=n}', "");
Expect(1, 195101, '\p{Is_NFKC_QC=	 N}', "");
Expect(0, 195101, '\p{^Is_NFKC_QC=	 N}', "");
Expect(0, 195101, '\P{Is_NFKC_QC=	 N}', "");
Expect(1, 195101, '\P{^Is_NFKC_QC=	 N}', "");
Expect(0, 195102, '\p{Is_NFKC_QC=	 N}', "");
Expect(1, 195102, '\p{^Is_NFKC_QC=	 N}', "");
Expect(1, 195102, '\P{Is_NFKC_QC=	 N}', "");
Expect(0, 195102, '\P{^Is_NFKC_QC=	 N}', "");
Error('\p{NFKC_Quick_Check:	_/a/Yes}');
Error('\P{NFKC_Quick_Check:	_/a/Yes}');
Expect(1, 195102, '\p{NFKC_Quick_Check:   yes}', "");
Expect(0, 195102, '\p{^NFKC_Quick_Check:   yes}', "");
Expect(0, 195102, '\P{NFKC_Quick_Check:   yes}', "");
Expect(1, 195102, '\P{^NFKC_Quick_Check:   yes}', "");
Expect(0, 195101, '\p{NFKC_Quick_Check:   yes}', "");
Expect(1, 195101, '\p{^NFKC_Quick_Check:   yes}', "");
Expect(1, 195101, '\P{NFKC_Quick_Check:   yes}', "");
Expect(0, 195101, '\P{^NFKC_Quick_Check:   yes}', "");
Expect(1, 195102, '\p{NFKC_Quick_Check=_	Yes}', "");
Expect(0, 195102, '\p{^NFKC_Quick_Check=_	Yes}', "");
Expect(0, 195102, '\P{NFKC_Quick_Check=_	Yes}', "");
Expect(1, 195102, '\P{^NFKC_Quick_Check=_	Yes}', "");
Expect(0, 195101, '\p{NFKC_Quick_Check=_	Yes}', "");
Expect(1, 195101, '\p{^NFKC_Quick_Check=_	Yes}', "");
Expect(1, 195101, '\P{NFKC_Quick_Check=_	Yes}', "");
Expect(0, 195101, '\P{^NFKC_Quick_Check=_	Yes}', "");
Error('\p{NFKC_QC=_ Y:=}');
Error('\P{NFKC_QC=_ Y:=}');
Expect(1, 195102, '\p{NFKC_QC:	y}', "");
Expect(0, 195102, '\p{^NFKC_QC:	y}', "");
Expect(0, 195102, '\P{NFKC_QC:	y}', "");
Expect(1, 195102, '\P{^NFKC_QC:	y}', "");
Expect(0, 195101, '\p{NFKC_QC:	y}', "");
Expect(1, 195101, '\p{^NFKC_QC:	y}', "");
Expect(1, 195101, '\P{NFKC_QC:	y}', "");
Expect(0, 195101, '\P{^NFKC_QC:	y}', "");
Expect(1, 195102, '\p{NFKC_QC=		Y}', "");
Expect(0, 195102, '\p{^NFKC_QC=		Y}', "");
Expect(0, 195102, '\P{NFKC_QC=		Y}', "");
Expect(1, 195102, '\P{^NFKC_QC=		Y}', "");
Expect(0, 195101, '\p{NFKC_QC=		Y}', "");
Expect(1, 195101, '\p{^NFKC_QC=		Y}', "");
Expect(1, 195101, '\P{NFKC_QC=		Y}', "");
Expect(0, 195101, '\P{^NFKC_QC=		Y}', "");
Error('\p{Is_NFKC_Quick_Check=/a/  YES}');
Error('\P{Is_NFKC_Quick_Check=/a/  YES}');
Expect(1, 195102, '\p{Is_NFKC_Quick_Check=yes}', "");
Expect(0, 195102, '\p{^Is_NFKC_Quick_Check=yes}', "");
Expect(0, 195102, '\P{Is_NFKC_Quick_Check=yes}', "");
Expect(1, 195102, '\P{^Is_NFKC_Quick_Check=yes}', "");
Expect(0, 195101, '\p{Is_NFKC_Quick_Check=yes}', "");
Expect(1, 195101, '\p{^Is_NFKC_Quick_Check=yes}', "");
Expect(1, 195101, '\P{Is_NFKC_Quick_Check=yes}', "");
Expect(0, 195101, '\P{^Is_NFKC_Quick_Check=yes}', "");
Expect(1, 195102, '\p{Is_NFKC_Quick_Check=-_yes}', "");
Expect(0, 195102, '\p{^Is_NFKC_Quick_Check=-_yes}', "");
Expect(0, 195102, '\P{Is_NFKC_Quick_Check=-_yes}', "");
Expect(1, 195102, '\P{^Is_NFKC_Quick_Check=-_yes}', "");
Expect(0, 195101, '\p{Is_NFKC_Quick_Check=-_yes}', "");
Expect(1, 195101, '\p{^Is_NFKC_Quick_Check=-_yes}', "");
Expect(1, 195101, '\P{Is_NFKC_Quick_Check=-_yes}', "");
Expect(0, 195101, '\P{^Is_NFKC_Quick_Check=-_yes}', "");
Error('\p{Is_NFKC_QC=/a/Y}');
Error('\P{Is_NFKC_QC=/a/Y}');
Expect(1, 195102, '\p{Is_NFKC_QC=y}', "");
Expect(0, 195102, '\p{^Is_NFKC_QC=y}', "");
Expect(0, 195102, '\P{Is_NFKC_QC=y}', "");
Expect(1, 195102, '\P{^Is_NFKC_QC=y}', "");
Expect(0, 195101, '\p{Is_NFKC_QC=y}', "");
Expect(1, 195101, '\p{^Is_NFKC_QC=y}', "");
Expect(1, 195101, '\P{Is_NFKC_QC=y}', "");
Expect(0, 195101, '\P{^Is_NFKC_QC=y}', "");
Expect(1, 195102, '\p{Is_NFKC_QC=_ Y}', "");
Expect(0, 195102, '\p{^Is_NFKC_QC=_ Y}', "");
Expect(0, 195102, '\P{Is_NFKC_QC=_ Y}', "");
Expect(1, 195102, '\P{^Is_NFKC_QC=_ Y}', "");
Expect(0, 195101, '\p{Is_NFKC_QC=_ Y}', "");
Expect(1, 195101, '\p{^Is_NFKC_QC=_ Y}', "");
Expect(1, 195101, '\P{Is_NFKC_QC=_ Y}', "");
Expect(0, 195101, '\P{^Is_NFKC_QC=_ Y}', "");
Error('\p{NFKD_Quick_Check=/a/	 No}');
Error('\P{NFKD_Quick_Check=/a/	 No}');
Expect(1, 195101, '\p{NFKD_Quick_Check=no}', "");
Expect(0, 195101, '\p{^NFKD_Quick_Check=no}', "");
Expect(0, 195101, '\P{NFKD_Quick_Check=no}', "");
Expect(1, 195101, '\P{^NFKD_Quick_Check=no}', "");
Expect(0, 195102, '\p{NFKD_Quick_Check=no}', "");
Expect(1, 195102, '\p{^NFKD_Quick_Check=no}', "");
Expect(1, 195102, '\P{NFKD_Quick_Check=no}', "");
Expect(0, 195102, '\P{^NFKD_Quick_Check=no}', "");
Expect(1, 195101, '\p{NFKD_Quick_Check= 	No}', "");
Expect(0, 195101, '\p{^NFKD_Quick_Check= 	No}', "");
Expect(0, 195101, '\P{NFKD_Quick_Check= 	No}', "");
Expect(1, 195101, '\P{^NFKD_Quick_Check= 	No}', "");
Expect(0, 195102, '\p{NFKD_Quick_Check= 	No}', "");
Expect(1, 195102, '\p{^NFKD_Quick_Check= 	No}', "");
Expect(1, 195102, '\P{NFKD_Quick_Check= 	No}', "");
Expect(0, 195102, '\P{^NFKD_Quick_Check= 	No}', "");
Error('\p{NFKD_QC=_:=N}');
Error('\P{NFKD_QC=_:=N}');
Expect(1, 195101, '\p{NFKD_QC=n}', "");
Expect(0, 195101, '\p{^NFKD_QC=n}', "");
Expect(0, 195101, '\P{NFKD_QC=n}', "");
Expect(1, 195101, '\P{^NFKD_QC=n}', "");
Expect(0, 195102, '\p{NFKD_QC=n}', "");
Expect(1, 195102, '\p{^NFKD_QC=n}', "");
Expect(1, 195102, '\P{NFKD_QC=n}', "");
Expect(0, 195102, '\P{^NFKD_QC=n}', "");
Expect(1, 195101, '\p{NFKD_QC= 	N}', "");
Expect(0, 195101, '\p{^NFKD_QC= 	N}', "");
Expect(0, 195101, '\P{NFKD_QC= 	N}', "");
Expect(1, 195101, '\P{^NFKD_QC= 	N}', "");
Expect(0, 195102, '\p{NFKD_QC= 	N}', "");
Expect(1, 195102, '\p{^NFKD_QC= 	N}', "");
Expect(1, 195102, '\P{NFKD_QC= 	N}', "");
Expect(0, 195102, '\P{^NFKD_QC= 	N}', "");
Error('\p{Is_NFKD_Quick_Check::= _no}');
Error('\P{Is_NFKD_Quick_Check::= _no}');
Expect(1, 195101, '\p{Is_NFKD_Quick_Check=no}', "");
Expect(0, 195101, '\p{^Is_NFKD_Quick_Check=no}', "");
Expect(0, 195101, '\P{Is_NFKD_Quick_Check=no}', "");
Expect(1, 195101, '\P{^Is_NFKD_Quick_Check=no}', "");
Expect(0, 195102, '\p{Is_NFKD_Quick_Check=no}', "");
Expect(1, 195102, '\p{^Is_NFKD_Quick_Check=no}', "");
Expect(1, 195102, '\P{Is_NFKD_Quick_Check=no}', "");
Expect(0, 195102, '\P{^Is_NFKD_Quick_Check=no}', "");
Expect(1, 195101, '\p{Is_NFKD_Quick_Check= no}', "");
Expect(0, 195101, '\p{^Is_NFKD_Quick_Check= no}', "");
Expect(0, 195101, '\P{Is_NFKD_Quick_Check= no}', "");
Expect(1, 195101, '\P{^Is_NFKD_Quick_Check= no}', "");
Expect(0, 195102, '\p{Is_NFKD_Quick_Check= no}', "");
Expect(1, 195102, '\p{^Is_NFKD_Quick_Check= no}', "");
Expect(1, 195102, '\P{Is_NFKD_Quick_Check= no}', "");
Expect(0, 195102, '\P{^Is_NFKD_Quick_Check= no}', "");
Error('\p{Is_NFKD_QC=-/a/N}');
Error('\P{Is_NFKD_QC=-/a/N}');
Expect(1, 195101, '\p{Is_NFKD_QC=n}', "");
Expect(0, 195101, '\p{^Is_NFKD_QC=n}', "");
Expect(0, 195101, '\P{Is_NFKD_QC=n}', "");
Expect(1, 195101, '\P{^Is_NFKD_QC=n}', "");
Expect(0, 195102, '\p{Is_NFKD_QC=n}', "");
Expect(1, 195102, '\p{^Is_NFKD_QC=n}', "");
Expect(1, 195102, '\P{Is_NFKD_QC=n}', "");
Expect(0, 195102, '\P{^Is_NFKD_QC=n}', "");
Expect(1, 195101, '\p{Is_NFKD_QC=_ N}', "");
Expect(0, 195101, '\p{^Is_NFKD_QC=_ N}', "");
Expect(0, 195101, '\P{Is_NFKD_QC=_ N}', "");
Expect(1, 195101, '\P{^Is_NFKD_QC=_ N}', "");
Expect(0, 195102, '\p{Is_NFKD_QC=_ N}', "");
Expect(1, 195102, '\p{^Is_NFKD_QC=_ N}', "");
Expect(1, 195102, '\P{Is_NFKD_QC=_ N}', "");
Expect(0, 195102, '\P{^Is_NFKD_QC=_ N}', "");
Error('\p{NFKD_Quick_Check=-Yes/a/}');
Error('\P{NFKD_Quick_Check=-Yes/a/}');
Expect(1, 195102, '\p{NFKD_Quick_Check=yes}', "");
Expect(0, 195102, '\p{^NFKD_Quick_Check=yes}', "");
Expect(0, 195102, '\P{NFKD_Quick_Check=yes}', "");
Expect(1, 195102, '\P{^NFKD_Quick_Check=yes}', "");
Expect(0, 195101, '\p{NFKD_Quick_Check=yes}', "");
Expect(1, 195101, '\p{^NFKD_Quick_Check=yes}', "");
Expect(1, 195101, '\P{NFKD_Quick_Check=yes}', "");
Expect(0, 195101, '\P{^NFKD_Quick_Check=yes}', "");
Expect(1, 195102, '\p{NFKD_Quick_Check=	 Yes}', "");
Expect(0, 195102, '\p{^NFKD_Quick_Check=	 Yes}', "");
Expect(0, 195102, '\P{NFKD_Quick_Check=	 Yes}', "");
Expect(1, 195102, '\P{^NFKD_Quick_Check=	 Yes}', "");
Expect(0, 195101, '\p{NFKD_Quick_Check=	 Yes}', "");
Expect(1, 195101, '\p{^NFKD_Quick_Check=	 Yes}', "");
Expect(1, 195101, '\P{NFKD_Quick_Check=	 Yes}', "");
Expect(0, 195101, '\P{^NFKD_Quick_Check=	 Yes}', "");
Error('\p{NFKD_QC=--Y:=}');
Error('\P{NFKD_QC=--Y:=}');
Expect(1, 195102, '\p{NFKD_QC=y}', "");
Expect(0, 195102, '\p{^NFKD_QC=y}', "");
Expect(0, 195102, '\P{NFKD_QC=y}', "");
Expect(1, 195102, '\P{^NFKD_QC=y}', "");
Expect(0, 195101, '\p{NFKD_QC=y}', "");
Expect(1, 195101, '\p{^NFKD_QC=y}', "");
Expect(1, 195101, '\P{NFKD_QC=y}', "");
Expect(0, 195101, '\P{^NFKD_QC=y}', "");
Expect(1, 195102, '\p{NFKD_QC=-y}', "");
Expect(0, 195102, '\p{^NFKD_QC=-y}', "");
Expect(0, 195102, '\P{NFKD_QC=-y}', "");
Expect(1, 195102, '\P{^NFKD_QC=-y}', "");
Expect(0, 195101, '\p{NFKD_QC=-y}', "");
Expect(1, 195101, '\p{^NFKD_QC=-y}', "");
Expect(1, 195101, '\P{NFKD_QC=-y}', "");
Expect(0, 195101, '\P{^NFKD_QC=-y}', "");
Error('\p{Is_NFKD_Quick_Check=	:=Yes}');
Error('\P{Is_NFKD_Quick_Check=	:=Yes}');
Expect(1, 195102, '\p{Is_NFKD_Quick_Check=yes}', "");
Expect(0, 195102, '\p{^Is_NFKD_Quick_Check=yes}', "");
Expect(0, 195102, '\P{Is_NFKD_Quick_Check=yes}', "");
Expect(1, 195102, '\P{^Is_NFKD_Quick_Check=yes}', "");
Expect(0, 195101, '\p{Is_NFKD_Quick_Check=yes}', "");
Expect(1, 195101, '\p{^Is_NFKD_Quick_Check=yes}', "");
Expect(1, 195101, '\P{Is_NFKD_Quick_Check=yes}', "");
Expect(0, 195101, '\P{^Is_NFKD_Quick_Check=yes}', "");
Expect(1, 195102, '\p{Is_NFKD_Quick_Check=_YES}', "");
Expect(0, 195102, '\p{^Is_NFKD_Quick_Check=_YES}', "");
Expect(0, 195102, '\P{Is_NFKD_Quick_Check=_YES}', "");
Expect(1, 195102, '\P{^Is_NFKD_Quick_Check=_YES}', "");
Expect(0, 195101, '\p{Is_NFKD_Quick_Check=_YES}', "");
Expect(1, 195101, '\p{^Is_NFKD_Quick_Check=_YES}', "");
Expect(1, 195101, '\P{Is_NFKD_Quick_Check=_YES}', "");
Expect(0, 195101, '\P{^Is_NFKD_Quick_Check=_YES}', "");
Error('\p{Is_NFKD_QC=	:=Y}');
Error('\P{Is_NFKD_QC=	:=Y}');
Expect(1, 195102, '\p{Is_NFKD_QC=y}', "");
Expect(0, 195102, '\p{^Is_NFKD_QC=y}', "");
Expect(0, 195102, '\P{Is_NFKD_QC=y}', "");
Expect(1, 195102, '\P{^Is_NFKD_QC=y}', "");
Expect(0, 195101, '\p{Is_NFKD_QC=y}', "");
Expect(1, 195101, '\p{^Is_NFKD_QC=y}', "");
Expect(1, 195101, '\P{Is_NFKD_QC=y}', "");
Expect(0, 195101, '\P{^Is_NFKD_QC=y}', "");
Expect(1, 195102, '\p{Is_NFKD_QC=	-Y}', "");
Expect(0, 195102, '\p{^Is_NFKD_QC=	-Y}', "");
Expect(0, 195102, '\P{Is_NFKD_QC=	-Y}', "");
Expect(1, 195102, '\P{^Is_NFKD_QC=	-Y}', "");
Expect(0, 195101, '\p{Is_NFKD_QC=	-Y}', "");
Expect(1, 195101, '\p{^Is_NFKD_QC=	-Y}', "");
Expect(1, 195101, '\P{Is_NFKD_QC=	-Y}', "");
Expect(0, 195101, '\P{^Is_NFKD_QC=	-Y}', "");
Error('\p{Other_Alphabetic=No}');
Error('\P{Other_Alphabetic=No}');
Error('\p{OAlpha: N}');
Error('\P{OAlpha: N}');
Error('\p{Is_Other_Alphabetic=F}');
Error('\P{Is_Other_Alphabetic=F}');
Error('\p{Is_OAlpha=False}');
Error('\P{Is_OAlpha=False}');
Error('\p{Other_Alphabetic:   Yes}');
Error('\P{Other_Alphabetic:   Yes}');
Error('\p{OAlpha=Y}');
Error('\P{OAlpha=Y}');
Error('\p{Is_Other_Alphabetic=T}');
Error('\P{Is_Other_Alphabetic=T}');
Error('\p{Is_OAlpha=True}');
Error('\P{Is_OAlpha=True}');
Error('\p{Other_Default_Ignorable_Code_Point: No}');
Error('\P{Other_Default_Ignorable_Code_Point: No}');
Error('\p{ODI=N}');
Error('\P{ODI=N}');
Error('\p{Is_Other_Default_Ignorable_Code_Point=F}');
Error('\P{Is_Other_Default_Ignorable_Code_Point=F}');
Error('\p{Is_ODI=False}');
Error('\P{Is_ODI=False}');
Error('\p{Other_Default_Ignorable_Code_Point=Yes}');
Error('\P{Other_Default_Ignorable_Code_Point=Yes}');
Error('\p{ODI=Y}');
Error('\P{ODI=Y}');
Error('\p{Is_Other_Default_Ignorable_Code_Point: T}');
Error('\P{Is_Other_Default_Ignorable_Code_Point: T}');
Error('\p{Is_ODI=True}');
Error('\P{Is_ODI=True}');
Error('\p{Other_Grapheme_Extend=No}');
Error('\P{Other_Grapheme_Extend=No}');
Error('\p{OGr_Ext:	N}');
Error('\P{OGr_Ext:	N}');
Error('\p{Is_Other_Grapheme_Extend=F}');
Error('\P{Is_Other_Grapheme_Extend=F}');
Error('\p{Is_OGr_Ext=False}');
Error('\P{Is_OGr_Ext=False}');
Error('\p{Other_Grapheme_Extend:	Yes}');
Error('\P{Other_Grapheme_Extend:	Yes}');
Error('\p{OGr_Ext=Y}');
Error('\P{OGr_Ext=Y}');
Error('\p{Is_Other_Grapheme_Extend=T}');
Error('\P{Is_Other_Grapheme_Extend=T}');
Error('\p{Is_OGr_Ext=True}');
Error('\P{Is_OGr_Ext=True}');
Error('\p{Other_ID_Continue=No}');
Error('\P{Other_ID_Continue=No}');
Error('\p{OIDC=N}');
Error('\P{OIDC=N}');
Error('\p{Is_Other_ID_Continue=F}');
Error('\P{Is_Other_ID_Continue=F}');
Error('\p{Is_OIDC=False}');
Error('\P{Is_OIDC=False}');
Error('\p{Other_ID_Continue=Yes}');
Error('\P{Other_ID_Continue=Yes}');
Error('\p{OIDC=Y}');
Error('\P{OIDC=Y}');
Error('\p{Is_Other_ID_Continue=T}');
Error('\P{Is_Other_ID_Continue=T}');
Error('\p{Is_OIDC=True}');
Error('\P{Is_OIDC=True}');
Error('\p{Other_ID_Start:	No}');
Error('\P{Other_ID_Start:	No}');
Error('\p{OIDS=N}');
Error('\P{OIDS=N}');
Error('\p{Is_Other_ID_Start=F}');
Error('\P{Is_Other_ID_Start=F}');
Error('\p{Is_OIDS=False}');
Error('\P{Is_OIDS=False}');
Error('\p{Other_ID_Start=Yes}');
Error('\P{Other_ID_Start=Yes}');
Error('\p{OIDS=Y}');
Error('\P{OIDS=Y}');
Error('\p{Is_Other_ID_Start:T}');
Error('\P{Is_Other_ID_Start:T}');
Error('\p{Is_OIDS=True}');
Error('\P{Is_OIDS=True}');
Error('\p{Other_Lowercase=No}');
Error('\P{Other_Lowercase=No}');
Error('\p{OLower=N}');
Error('\P{OLower=N}');
Error('\p{Is_Other_Lowercase=F}');
Error('\P{Is_Other_Lowercase=F}');
Error('\p{Is_OLower:False}');
Error('\P{Is_OLower:False}');
Error('\p{Other_Lowercase=Yes}');
Error('\P{Other_Lowercase=Yes}');
Error('\p{OLower=Y}');
Error('\P{OLower=Y}');
Error('\p{Is_Other_Lowercase=T}');
Error('\P{Is_Other_Lowercase=T}');
Error('\p{Is_OLower=True}');
Error('\P{Is_OLower=True}');
Error('\p{Other_Math=No}');
Error('\P{Other_Math=No}');
Error('\p{OMath=N}');
Error('\P{OMath=N}');
Error('\p{Is_Other_Math=F}');
Error('\P{Is_Other_Math=F}');
Error('\p{Is_OMath=False}');
Error('\P{Is_OMath=False}');
Error('\p{Other_Math=Yes}');
Error('\P{Other_Math=Yes}');
Error('\p{OMath=Y}');
Error('\P{OMath=Y}');
Error('\p{Is_Other_Math=T}');
Error('\P{Is_Other_Math=T}');
Error('\p{Is_OMath:   True}');
Error('\P{Is_OMath:   True}');
Error('\p{Other_Uppercase=No}');
Error('\P{Other_Uppercase=No}');
Error('\p{OUpper=N}');
Error('\P{OUpper=N}');
Error('\p{Is_Other_Uppercase=F}');
Error('\P{Is_Other_Uppercase=F}');
Error('\p{Is_OUpper: False}');
Error('\P{Is_OUpper: False}');
Error('\p{Other_Uppercase: Yes}');
Error('\P{Other_Uppercase: Yes}');
Error('\p{OUpper=Y}');
Error('\P{OUpper=Y}');
Error('\p{Is_Other_Uppercase=T}');
Error('\P{Is_Other_Uppercase=T}');
Error('\p{Is_OUpper=True}');
Error('\P{Is_OUpper=True}');
Error('\p{Pattern_Syntax=	No/a/}');
Error('\P{Pattern_Syntax=	No/a/}');
Expect(1, 65095, '\p{Pattern_Syntax=no}', "");
Expect(0, 65095, '\p{^Pattern_Syntax=no}', "");
Expect(0, 65095, '\P{Pattern_Syntax=no}', "");
Expect(1, 65095, '\P{^Pattern_Syntax=no}', "");
Expect(0, 65094, '\p{Pattern_Syntax=no}', "");
Expect(1, 65094, '\p{^Pattern_Syntax=no}', "");
Expect(1, 65094, '\P{Pattern_Syntax=no}', "");
Expect(0, 65094, '\P{^Pattern_Syntax=no}', "");
Expect(1, 65095, '\p{Pattern_Syntax:  _No}', "");
Expect(0, 65095, '\p{^Pattern_Syntax:  _No}', "");
Expect(0, 65095, '\P{Pattern_Syntax:  _No}', "");
Expect(1, 65095, '\P{^Pattern_Syntax:  _No}', "");
Expect(0, 65094, '\p{Pattern_Syntax:  _No}', "");
Expect(1, 65094, '\p{^Pattern_Syntax:  _No}', "");
Expect(1, 65094, '\P{Pattern_Syntax:  _No}', "");
Expect(0, 65094, '\P{^Pattern_Syntax:  _No}', "");
Error('\p{Pat_Syn= /a/N}');
Error('\P{Pat_Syn= /a/N}');
Expect(1, 65095, '\p{Pat_Syn=n}', "");
Expect(0, 65095, '\p{^Pat_Syn=n}', "");
Expect(0, 65095, '\P{Pat_Syn=n}', "");
Expect(1, 65095, '\P{^Pat_Syn=n}', "");
Expect(0, 65094, '\p{Pat_Syn=n}', "");
Expect(1, 65094, '\p{^Pat_Syn=n}', "");
Expect(1, 65094, '\P{Pat_Syn=n}', "");
Expect(0, 65094, '\P{^Pat_Syn=n}', "");
Expect(1, 65095, '\p{Pat_Syn=- N}', "");
Expect(0, 65095, '\p{^Pat_Syn=- N}', "");
Expect(0, 65095, '\P{Pat_Syn=- N}', "");
Expect(1, 65095, '\P{^Pat_Syn=- N}', "");
Expect(0, 65094, '\p{Pat_Syn=- N}', "");
Expect(1, 65094, '\p{^Pat_Syn=- N}', "");
Expect(1, 65094, '\P{Pat_Syn=- N}', "");
Expect(0, 65094, '\P{^Pat_Syn=- N}', "");
Error('\p{Is_Pattern_Syntax=-_F/a/}');
Error('\P{Is_Pattern_Syntax=-_F/a/}');
Expect(1, 65095, '\p{Is_Pattern_Syntax:	f}', "");
Expect(0, 65095, '\p{^Is_Pattern_Syntax:	f}', "");
Expect(0, 65095, '\P{Is_Pattern_Syntax:	f}', "");
Expect(1, 65095, '\P{^Is_Pattern_Syntax:	f}', "");
Expect(0, 65094, '\p{Is_Pattern_Syntax:	f}', "");
Expect(1, 65094, '\p{^Is_Pattern_Syntax:	f}', "");
Expect(1, 65094, '\P{Is_Pattern_Syntax:	f}', "");
Expect(0, 65094, '\P{^Is_Pattern_Syntax:	f}', "");
Expect(1, 65095, '\p{Is_Pattern_Syntax= f}', "");
Expect(0, 65095, '\p{^Is_Pattern_Syntax= f}', "");
Expect(0, 65095, '\P{Is_Pattern_Syntax= f}', "");
Expect(1, 65095, '\P{^Is_Pattern_Syntax= f}', "");
Expect(0, 65094, '\p{Is_Pattern_Syntax= f}', "");
Expect(1, 65094, '\p{^Is_Pattern_Syntax= f}', "");
Expect(1, 65094, '\P{Is_Pattern_Syntax= f}', "");
Expect(0, 65094, '\P{^Is_Pattern_Syntax= f}', "");
Error('\p{Is_Pat_Syn:   := _False}');
Error('\P{Is_Pat_Syn:   := _False}');
Expect(1, 65095, '\p{Is_Pat_Syn=false}', "");
Expect(0, 65095, '\p{^Is_Pat_Syn=false}', "");
Expect(0, 65095, '\P{Is_Pat_Syn=false}', "");
Expect(1, 65095, '\P{^Is_Pat_Syn=false}', "");
Expect(0, 65094, '\p{Is_Pat_Syn=false}', "");
Expect(1, 65094, '\p{^Is_Pat_Syn=false}', "");
Expect(1, 65094, '\P{Is_Pat_Syn=false}', "");
Expect(0, 65094, '\P{^Is_Pat_Syn=false}', "");
Expect(1, 65095, '\p{Is_Pat_Syn= _False}', "");
Expect(0, 65095, '\p{^Is_Pat_Syn= _False}', "");
Expect(0, 65095, '\P{Is_Pat_Syn= _False}', "");
Expect(1, 65095, '\P{^Is_Pat_Syn= _False}', "");
Expect(0, 65094, '\p{Is_Pat_Syn= _False}', "");
Expect(1, 65094, '\p{^Is_Pat_Syn= _False}', "");
Expect(1, 65094, '\P{Is_Pat_Syn= _False}', "");
Expect(0, 65094, '\P{^Is_Pat_Syn= _False}', "");
Error('\p{Pattern_Syntax=_	Yes:=}');
Error('\P{Pattern_Syntax=_	Yes:=}');
Expect(1, 65094, '\p{Pattern_Syntax=yes}', "");
Expect(0, 65094, '\p{^Pattern_Syntax=yes}', "");
Expect(0, 65094, '\P{Pattern_Syntax=yes}', "");
Expect(1, 65094, '\P{^Pattern_Syntax=yes}', "");
Expect(0, 65095, '\p{Pattern_Syntax=yes}', "");
Expect(1, 65095, '\p{^Pattern_Syntax=yes}', "");
Expect(1, 65095, '\P{Pattern_Syntax=yes}', "");
Expect(0, 65095, '\P{^Pattern_Syntax=yes}', "");
Expect(1, 65094, '\p{Pattern_Syntax=		Yes}', "");
Expect(0, 65094, '\p{^Pattern_Syntax=		Yes}', "");
Expect(0, 65094, '\P{Pattern_Syntax=		Yes}', "");
Expect(1, 65094, '\P{^Pattern_Syntax=		Yes}', "");
Expect(0, 65095, '\p{Pattern_Syntax=		Yes}', "");
Expect(1, 65095, '\p{^Pattern_Syntax=		Yes}', "");
Expect(1, 65095, '\P{Pattern_Syntax=		Yes}', "");
Expect(0, 65095, '\P{^Pattern_Syntax=		Yes}', "");
Error('\p{Pat_Syn=:=_	Y}');
Error('\P{Pat_Syn=:=_	Y}');
Expect(1, 65094, '\p{Pat_Syn=y}', "");
Expect(0, 65094, '\p{^Pat_Syn=y}', "");
Expect(0, 65094, '\P{Pat_Syn=y}', "");
Expect(1, 65094, '\P{^Pat_Syn=y}', "");
Expect(0, 65095, '\p{Pat_Syn=y}', "");
Expect(1, 65095, '\p{^Pat_Syn=y}', "");
Expect(1, 65095, '\P{Pat_Syn=y}', "");
Expect(0, 65095, '\P{^Pat_Syn=y}', "");
Expect(1, 65094, '\p{Pat_Syn=__Y}', "");
Expect(0, 65094, '\p{^Pat_Syn=__Y}', "");
Expect(0, 65094, '\P{Pat_Syn=__Y}', "");
Expect(1, 65094, '\P{^Pat_Syn=__Y}', "");
Expect(0, 65095, '\p{Pat_Syn=__Y}', "");
Expect(1, 65095, '\p{^Pat_Syn=__Y}', "");
Expect(1, 65095, '\P{Pat_Syn=__Y}', "");
Expect(0, 65095, '\P{^Pat_Syn=__Y}', "");
Error('\p{Is_Pattern_Syntax=/a/T}');
Error('\P{Is_Pattern_Syntax=/a/T}');
Expect(1, 65094, '\p{Is_Pattern_Syntax=t}', "");
Expect(0, 65094, '\p{^Is_Pattern_Syntax=t}', "");
Expect(0, 65094, '\P{Is_Pattern_Syntax=t}', "");
Expect(1, 65094, '\P{^Is_Pattern_Syntax=t}', "");
Expect(0, 65095, '\p{Is_Pattern_Syntax=t}', "");
Expect(1, 65095, '\p{^Is_Pattern_Syntax=t}', "");
Expect(1, 65095, '\P{Is_Pattern_Syntax=t}', "");
Expect(0, 65095, '\P{^Is_Pattern_Syntax=t}', "");
Expect(1, 65094, '\p{Is_Pattern_Syntax=	 T}', "");
Expect(0, 65094, '\p{^Is_Pattern_Syntax=	 T}', "");
Expect(0, 65094, '\P{Is_Pattern_Syntax=	 T}', "");
Expect(1, 65094, '\P{^Is_Pattern_Syntax=	 T}', "");
Expect(0, 65095, '\p{Is_Pattern_Syntax=	 T}', "");
Expect(1, 65095, '\p{^Is_Pattern_Syntax=	 T}', "");
Expect(1, 65095, '\P{Is_Pattern_Syntax=	 T}', "");
Expect(0, 65095, '\P{^Is_Pattern_Syntax=	 T}', "");
Error('\p{Is_Pat_Syn=_True/a/}');
Error('\P{Is_Pat_Syn=_True/a/}');
Expect(1, 65094, '\p{Is_Pat_Syn=true}', "");
Expect(0, 65094, '\p{^Is_Pat_Syn=true}', "");
Expect(0, 65094, '\P{Is_Pat_Syn=true}', "");
Expect(1, 65094, '\P{^Is_Pat_Syn=true}', "");
Expect(0, 65095, '\p{Is_Pat_Syn=true}', "");
Expect(1, 65095, '\p{^Is_Pat_Syn=true}', "");
Expect(1, 65095, '\P{Is_Pat_Syn=true}', "");
Expect(0, 65095, '\P{^Is_Pat_Syn=true}', "");
Expect(1, 65094, '\p{Is_Pat_Syn=-	true}', "");
Expect(0, 65094, '\p{^Is_Pat_Syn=-	true}', "");
Expect(0, 65094, '\P{Is_Pat_Syn=-	true}', "");
Expect(1, 65094, '\P{^Is_Pat_Syn=-	true}', "");
Expect(0, 65095, '\p{Is_Pat_Syn=-	true}', "");
Expect(1, 65095, '\p{^Is_Pat_Syn=-	true}', "");
Expect(1, 65095, '\P{Is_Pat_Syn=-	true}', "");
Expect(0, 65095, '\P{^Is_Pat_Syn=-	true}', "");
Error('\p{Pattern_White_Space= No:=}');
Error('\P{Pattern_White_Space= No:=}');
Expect(1, 8234, '\p{Pattern_White_Space=no}', "");
Expect(0, 8234, '\p{^Pattern_White_Space=no}', "");
Expect(0, 8234, '\P{Pattern_White_Space=no}', "");
Expect(1, 8234, '\P{^Pattern_White_Space=no}', "");
Expect(0, 8233, '\p{Pattern_White_Space=no}', "");
Expect(1, 8233, '\p{^Pattern_White_Space=no}', "");
Expect(1, 8233, '\P{Pattern_White_Space=no}', "");
Expect(0, 8233, '\P{^Pattern_White_Space=no}', "");
Expect(1, 8234, '\p{Pattern_White_Space=- No}', "");
Expect(0, 8234, '\p{^Pattern_White_Space=- No}', "");
Expect(0, 8234, '\P{Pattern_White_Space=- No}', "");
Expect(1, 8234, '\P{^Pattern_White_Space=- No}', "");
Expect(0, 8233, '\p{Pattern_White_Space=- No}', "");
Expect(1, 8233, '\p{^Pattern_White_Space=- No}', "");
Expect(1, 8233, '\P{Pattern_White_Space=- No}', "");
Expect(0, 8233, '\P{^Pattern_White_Space=- No}', "");
Error('\p{Pat_WS=	N:=}');
Error('\P{Pat_WS=	N:=}');
Expect(1, 8234, '\p{Pat_WS=n}', "");
Expect(0, 8234, '\p{^Pat_WS=n}', "");
Expect(0, 8234, '\P{Pat_WS=n}', "");
Expect(1, 8234, '\P{^Pat_WS=n}', "");
Expect(0, 8233, '\p{Pat_WS=n}', "");
Expect(1, 8233, '\p{^Pat_WS=n}', "");
Expect(1, 8233, '\P{Pat_WS=n}', "");
Expect(0, 8233, '\P{^Pat_WS=n}', "");
Expect(1, 8234, '\p{Pat_WS:_n}', "");
Expect(0, 8234, '\p{^Pat_WS:_n}', "");
Expect(0, 8234, '\P{Pat_WS:_n}', "");
Expect(1, 8234, '\P{^Pat_WS:_n}', "");
Expect(0, 8233, '\p{Pat_WS:_n}', "");
Expect(1, 8233, '\p{^Pat_WS:_n}', "");
Expect(1, 8233, '\P{Pat_WS:_n}', "");
Expect(0, 8233, '\P{^Pat_WS:_n}', "");
Error('\p{Is_Pattern_White_Space=	F:=}');
Error('\P{Is_Pattern_White_Space=	F:=}');
Expect(1, 8234, '\p{Is_Pattern_White_Space=f}', "");
Expect(0, 8234, '\p{^Is_Pattern_White_Space=f}', "");
Expect(0, 8234, '\P{Is_Pattern_White_Space=f}', "");
Expect(1, 8234, '\P{^Is_Pattern_White_Space=f}', "");
Expect(0, 8233, '\p{Is_Pattern_White_Space=f}', "");
Expect(1, 8233, '\p{^Is_Pattern_White_Space=f}', "");
Expect(1, 8233, '\P{Is_Pattern_White_Space=f}', "");
Expect(0, 8233, '\P{^Is_Pattern_White_Space=f}', "");
Expect(1, 8234, '\p{Is_Pattern_White_Space=_ F}', "");
Expect(0, 8234, '\p{^Is_Pattern_White_Space=_ F}', "");
Expect(0, 8234, '\P{Is_Pattern_White_Space=_ F}', "");
Expect(1, 8234, '\P{^Is_Pattern_White_Space=_ F}', "");
Expect(0, 8233, '\p{Is_Pattern_White_Space=_ F}', "");
Expect(1, 8233, '\p{^Is_Pattern_White_Space=_ F}', "");
Expect(1, 8233, '\P{Is_Pattern_White_Space=_ F}', "");
Expect(0, 8233, '\P{^Is_Pattern_White_Space=_ F}', "");
Error('\p{Is_Pat_WS=:=--FALSE}');
Error('\P{Is_Pat_WS=:=--FALSE}');
Expect(1, 8234, '\p{Is_Pat_WS=false}', "");
Expect(0, 8234, '\p{^Is_Pat_WS=false}', "");
Expect(0, 8234, '\P{Is_Pat_WS=false}', "");
Expect(1, 8234, '\P{^Is_Pat_WS=false}', "");
Expect(0, 8233, '\p{Is_Pat_WS=false}', "");
Expect(1, 8233, '\p{^Is_Pat_WS=false}', "");
Expect(1, 8233, '\P{Is_Pat_WS=false}', "");
Expect(0, 8233, '\P{^Is_Pat_WS=false}', "");
Expect(1, 8234, '\p{Is_Pat_WS=  False}', "");
Expect(0, 8234, '\p{^Is_Pat_WS=  False}', "");
Expect(0, 8234, '\P{Is_Pat_WS=  False}', "");
Expect(1, 8234, '\P{^Is_Pat_WS=  False}', "");
Expect(0, 8233, '\p{Is_Pat_WS=  False}', "");
Expect(1, 8233, '\p{^Is_Pat_WS=  False}', "");
Expect(1, 8233, '\P{Is_Pat_WS=  False}', "");
Expect(0, 8233, '\P{^Is_Pat_WS=  False}', "");
Error('\p{Pattern_White_Space=/a/ -Yes}');
Error('\P{Pattern_White_Space=/a/ -Yes}');
Expect(1, 8233, '\p{Pattern_White_Space=yes}', "");
Expect(0, 8233, '\p{^Pattern_White_Space=yes}', "");
Expect(0, 8233, '\P{Pattern_White_Space=yes}', "");
Expect(1, 8233, '\P{^Pattern_White_Space=yes}', "");
Expect(0, 8234, '\p{Pattern_White_Space=yes}', "");
Expect(1, 8234, '\p{^Pattern_White_Space=yes}', "");
Expect(1, 8234, '\P{Pattern_White_Space=yes}', "");
Expect(0, 8234, '\P{^Pattern_White_Space=yes}', "");
Expect(1, 8233, '\p{Pattern_White_Space=-yes}', "");
Expect(0, 8233, '\p{^Pattern_White_Space=-yes}', "");
Expect(0, 8233, '\P{Pattern_White_Space=-yes}', "");
Expect(1, 8233, '\P{^Pattern_White_Space=-yes}', "");
Expect(0, 8234, '\p{Pattern_White_Space=-yes}', "");
Expect(1, 8234, '\p{^Pattern_White_Space=-yes}', "");
Expect(1, 8234, '\P{Pattern_White_Space=-yes}', "");
Expect(0, 8234, '\P{^Pattern_White_Space=-yes}', "");
Error('\p{Pat_WS=-/a/Y}');
Error('\P{Pat_WS=-/a/Y}');
Expect(1, 8233, '\p{Pat_WS=y}', "");
Expect(0, 8233, '\p{^Pat_WS=y}', "");
Expect(0, 8233, '\P{Pat_WS=y}', "");
Expect(1, 8233, '\P{^Pat_WS=y}', "");
Expect(0, 8234, '\p{Pat_WS=y}', "");
Expect(1, 8234, '\p{^Pat_WS=y}', "");
Expect(1, 8234, '\P{Pat_WS=y}', "");
Expect(0, 8234, '\P{^Pat_WS=y}', "");
Expect(1, 8233, '\p{Pat_WS=Y}', "");
Expect(0, 8233, '\p{^Pat_WS=Y}', "");
Expect(0, 8233, '\P{Pat_WS=Y}', "");
Expect(1, 8233, '\P{^Pat_WS=Y}', "");
Expect(0, 8234, '\p{Pat_WS=Y}', "");
Expect(1, 8234, '\p{^Pat_WS=Y}', "");
Expect(1, 8234, '\P{Pat_WS=Y}', "");
Expect(0, 8234, '\P{^Pat_WS=Y}', "");
Error('\p{Is_Pattern_White_Space=:=_-t}');
Error('\P{Is_Pattern_White_Space=:=_-t}');
Expect(1, 8233, '\p{Is_Pattern_White_Space=t}', "");
Expect(0, 8233, '\p{^Is_Pattern_White_Space=t}', "");
Expect(0, 8233, '\P{Is_Pattern_White_Space=t}', "");
Expect(1, 8233, '\P{^Is_Pattern_White_Space=t}', "");
Expect(0, 8234, '\p{Is_Pattern_White_Space=t}', "");
Expect(1, 8234, '\p{^Is_Pattern_White_Space=t}', "");
Expect(1, 8234, '\P{Is_Pattern_White_Space=t}', "");
Expect(0, 8234, '\P{^Is_Pattern_White_Space=t}', "");
Expect(1, 8233, '\p{Is_Pattern_White_Space=- T}', "");
Expect(0, 8233, '\p{^Is_Pattern_White_Space=- T}', "");
Expect(0, 8233, '\P{Is_Pattern_White_Space=- T}', "");
Expect(1, 8233, '\P{^Is_Pattern_White_Space=- T}', "");
Expect(0, 8234, '\p{Is_Pattern_White_Space=- T}', "");
Expect(1, 8234, '\p{^Is_Pattern_White_Space=- T}', "");
Expect(1, 8234, '\P{Is_Pattern_White_Space=- T}', "");
Expect(0, 8234, '\P{^Is_Pattern_White_Space=- T}', "");
Error('\p{Is_Pat_WS=:= True}');
Error('\P{Is_Pat_WS=:= True}');
Expect(1, 8233, '\p{Is_Pat_WS=true}', "");
Expect(0, 8233, '\p{^Is_Pat_WS=true}', "");
Expect(0, 8233, '\P{Is_Pat_WS=true}', "");
Expect(1, 8233, '\P{^Is_Pat_WS=true}', "");
Expect(0, 8234, '\p{Is_Pat_WS=true}', "");
Expect(1, 8234, '\p{^Is_Pat_WS=true}', "");
Expect(1, 8234, '\P{Is_Pat_WS=true}', "");
Expect(0, 8234, '\P{^Is_Pat_WS=true}', "");
Expect(1, 8233, '\p{Is_Pat_WS=	True}', "");
Expect(0, 8233, '\p{^Is_Pat_WS=	True}', "");
Expect(0, 8233, '\P{Is_Pat_WS=	True}', "");
Expect(1, 8233, '\P{^Is_Pat_WS=	True}', "");
Expect(0, 8234, '\p{Is_Pat_WS=	True}', "");
Expect(1, 8234, '\p{^Is_Pat_WS=	True}', "");
Expect(1, 8234, '\P{Is_Pat_WS=	True}', "");
Expect(0, 8234, '\P{^Is_Pat_WS=	True}', "");
Error('\p{Quotation_Mark=-:=NO}');
Error('\P{Quotation_Mark=-:=NO}');
Expect(1, 65380, '\p{Quotation_Mark=no}', "");
Expect(0, 65380, '\p{^Quotation_Mark=no}', "");
Expect(0, 65380, '\P{Quotation_Mark=no}', "");
Expect(1, 65380, '\P{^Quotation_Mark=no}', "");
Expect(0, 65379, '\p{Quotation_Mark=no}', "");
Expect(1, 65379, '\p{^Quotation_Mark=no}', "");
Expect(1, 65379, '\P{Quotation_Mark=no}', "");
Expect(0, 65379, '\P{^Quotation_Mark=no}', "");
Expect(1, 65380, '\p{Quotation_Mark=-	No}', "");
Expect(0, 65380, '\p{^Quotation_Mark=-	No}', "");
Expect(0, 65380, '\P{Quotation_Mark=-	No}', "");
Expect(1, 65380, '\P{^Quotation_Mark=-	No}', "");
Expect(0, 65379, '\p{Quotation_Mark=-	No}', "");
Expect(1, 65379, '\p{^Quotation_Mark=-	No}', "");
Expect(1, 65379, '\P{Quotation_Mark=-	No}', "");
Expect(0, 65379, '\P{^Quotation_Mark=-	No}', "");
Error('\p{QMark= :=N}');
Error('\P{QMark= :=N}');
Expect(1, 65380, '\p{QMark=n}', "");
Expect(0, 65380, '\p{^QMark=n}', "");
Expect(0, 65380, '\P{QMark=n}', "");
Expect(1, 65380, '\P{^QMark=n}', "");
Expect(0, 65379, '\p{QMark=n}', "");
Expect(1, 65379, '\p{^QMark=n}', "");
Expect(1, 65379, '\P{QMark=n}', "");
Expect(0, 65379, '\P{^QMark=n}', "");
Expect(1, 65380, '\p{QMark=  N}', "");
Expect(0, 65380, '\p{^QMark=  N}', "");
Expect(0, 65380, '\P{QMark=  N}', "");
Expect(1, 65380, '\P{^QMark=  N}', "");
Expect(0, 65379, '\p{QMark=  N}', "");
Expect(1, 65379, '\p{^QMark=  N}', "");
Expect(1, 65379, '\P{QMark=  N}', "");
Expect(0, 65379, '\P{^QMark=  N}', "");
Error('\p{Is_Quotation_Mark=:= 	F}');
Error('\P{Is_Quotation_Mark=:= 	F}');
Expect(1, 65380, '\p{Is_Quotation_Mark=f}', "");
Expect(0, 65380, '\p{^Is_Quotation_Mark=f}', "");
Expect(0, 65380, '\P{Is_Quotation_Mark=f}', "");
Expect(1, 65380, '\P{^Is_Quotation_Mark=f}', "");
Expect(0, 65379, '\p{Is_Quotation_Mark=f}', "");
Expect(1, 65379, '\p{^Is_Quotation_Mark=f}', "");
Expect(1, 65379, '\P{Is_Quotation_Mark=f}', "");
Expect(0, 65379, '\P{^Is_Quotation_Mark=f}', "");
Expect(1, 65380, '\p{Is_Quotation_Mark=-F}', "");
Expect(0, 65380, '\p{^Is_Quotation_Mark=-F}', "");
Expect(0, 65380, '\P{Is_Quotation_Mark=-F}', "");
Expect(1, 65380, '\P{^Is_Quotation_Mark=-F}', "");
Expect(0, 65379, '\p{Is_Quotation_Mark=-F}', "");
Expect(1, 65379, '\p{^Is_Quotation_Mark=-F}', "");
Expect(1, 65379, '\P{Is_Quotation_Mark=-F}', "");
Expect(0, 65379, '\P{^Is_Quotation_Mark=-F}', "");
Error('\p{Is_QMark=-/a/false}');
Error('\P{Is_QMark=-/a/false}');
Expect(1, 65380, '\p{Is_QMark=false}', "");
Expect(0, 65380, '\p{^Is_QMark=false}', "");
Expect(0, 65380, '\P{Is_QMark=false}', "");
Expect(1, 65380, '\P{^Is_QMark=false}', "");
Expect(0, 65379, '\p{Is_QMark=false}', "");
Expect(1, 65379, '\p{^Is_QMark=false}', "");
Expect(1, 65379, '\P{Is_QMark=false}', "");
Expect(0, 65379, '\P{^Is_QMark=false}', "");
Expect(1, 65380, '\p{Is_QMark=-False}', "");
Expect(0, 65380, '\p{^Is_QMark=-False}', "");
Expect(0, 65380, '\P{Is_QMark=-False}', "");
Expect(1, 65380, '\P{^Is_QMark=-False}', "");
Expect(0, 65379, '\p{Is_QMark=-False}', "");
Expect(1, 65379, '\p{^Is_QMark=-False}', "");
Expect(1, 65379, '\P{Is_QMark=-False}', "");
Expect(0, 65379, '\P{^Is_QMark=-False}', "");
Error('\p{Quotation_Mark=-/a/Yes}');
Error('\P{Quotation_Mark=-/a/Yes}');
Expect(1, 65379, '\p{Quotation_Mark=yes}', "");
Expect(0, 65379, '\p{^Quotation_Mark=yes}', "");
Expect(0, 65379, '\P{Quotation_Mark=yes}', "");
Expect(1, 65379, '\P{^Quotation_Mark=yes}', "");
Expect(0, 65380, '\p{Quotation_Mark=yes}', "");
Expect(1, 65380, '\p{^Quotation_Mark=yes}', "");
Expect(1, 65380, '\P{Quotation_Mark=yes}', "");
Expect(0, 65380, '\P{^Quotation_Mark=yes}', "");
Expect(1, 65379, '\p{Quotation_Mark=_-YES}', "");
Expect(0, 65379, '\p{^Quotation_Mark=_-YES}', "");
Expect(0, 65379, '\P{Quotation_Mark=_-YES}', "");
Expect(1, 65379, '\P{^Quotation_Mark=_-YES}', "");
Expect(0, 65380, '\p{Quotation_Mark=_-YES}', "");
Expect(1, 65380, '\p{^Quotation_Mark=_-YES}', "");
Expect(1, 65380, '\P{Quotation_Mark=_-YES}', "");
Expect(0, 65380, '\P{^Quotation_Mark=_-YES}', "");
Error('\p{QMark:	/a/ -Y}');
Error('\P{QMark:	/a/ -Y}');
Expect(1, 65379, '\p{QMark:	y}', "");
Expect(0, 65379, '\p{^QMark:	y}', "");
Expect(0, 65379, '\P{QMark:	y}', "");
Expect(1, 65379, '\P{^QMark:	y}', "");
Expect(0, 65380, '\p{QMark:	y}', "");
Expect(1, 65380, '\p{^QMark:	y}', "");
Expect(1, 65380, '\P{QMark:	y}', "");
Expect(0, 65380, '\P{^QMark:	y}', "");
Expect(1, 65379, '\p{QMark=Y}', "");
Expect(0, 65379, '\p{^QMark=Y}', "");
Expect(0, 65379, '\P{QMark=Y}', "");
Expect(1, 65379, '\P{^QMark=Y}', "");
Expect(0, 65380, '\p{QMark=Y}', "");
Expect(1, 65380, '\p{^QMark=Y}', "");
Expect(1, 65380, '\P{QMark=Y}', "");
Expect(0, 65380, '\P{^QMark=Y}', "");
Error('\p{Is_Quotation_Mark=	_T:=}');
Error('\P{Is_Quotation_Mark=	_T:=}');
Expect(1, 65379, '\p{Is_Quotation_Mark=t}', "");
Expect(0, 65379, '\p{^Is_Quotation_Mark=t}', "");
Expect(0, 65379, '\P{Is_Quotation_Mark=t}', "");
Expect(1, 65379, '\P{^Is_Quotation_Mark=t}', "");
Expect(0, 65380, '\p{Is_Quotation_Mark=t}', "");
Expect(1, 65380, '\p{^Is_Quotation_Mark=t}', "");
Expect(1, 65380, '\P{Is_Quotation_Mark=t}', "");
Expect(0, 65380, '\P{^Is_Quotation_Mark=t}', "");
Expect(1, 65379, '\p{Is_Quotation_Mark=-T}', "");
Expect(0, 65379, '\p{^Is_Quotation_Mark=-T}', "");
Expect(0, 65379, '\P{Is_Quotation_Mark=-T}', "");
Expect(1, 65379, '\P{^Is_Quotation_Mark=-T}', "");
Expect(0, 65380, '\p{Is_Quotation_Mark=-T}', "");
Expect(1, 65380, '\p{^Is_Quotation_Mark=-T}', "");
Expect(1, 65380, '\P{Is_Quotation_Mark=-T}', "");
Expect(0, 65380, '\P{^Is_Quotation_Mark=-T}', "");
Error('\p{Is_QMark=	:=TRUE}');
Error('\P{Is_QMark=	:=TRUE}');
Expect(1, 65379, '\p{Is_QMark=true}', "");
Expect(0, 65379, '\p{^Is_QMark=true}', "");
Expect(0, 65379, '\P{Is_QMark=true}', "");
Expect(1, 65379, '\P{^Is_QMark=true}', "");
Expect(0, 65380, '\p{Is_QMark=true}', "");
Expect(1, 65380, '\p{^Is_QMark=true}', "");
Expect(1, 65380, '\P{Is_QMark=true}', "");
Expect(0, 65380, '\P{^Is_QMark=true}', "");
Expect(1, 65379, '\p{Is_QMark=	True}', "");
Expect(0, 65379, '\p{^Is_QMark=	True}', "");
Expect(0, 65379, '\P{Is_QMark=	True}', "");
Expect(1, 65379, '\P{^Is_QMark=	True}', "");
Expect(0, 65380, '\p{Is_QMark=	True}', "");
Expect(1, 65380, '\p{^Is_QMark=	True}', "");
Expect(1, 65380, '\P{Is_QMark=	True}', "");
Expect(0, 65380, '\P{^Is_QMark=	True}', "");
Error('\p{Radical=:= _NO}');
Error('\P{Radical=:= _NO}');
Expect(1, 12246, '\p{Radical=no}', "");
Expect(0, 12246, '\p{^Radical=no}', "");
Expect(0, 12246, '\P{Radical=no}', "");
Expect(1, 12246, '\P{^Radical=no}', "");
Expect(0, 12245, '\p{Radical=no}', "");
Expect(1, 12245, '\p{^Radical=no}', "");
Expect(1, 12245, '\P{Radical=no}', "");
Expect(0, 12245, '\P{^Radical=no}', "");
Expect(1, 12246, '\p{Radical=- No}', "");
Expect(0, 12246, '\p{^Radical=- No}', "");
Expect(0, 12246, '\P{Radical=- No}', "");
Expect(1, 12246, '\P{^Radical=- No}', "");
Expect(0, 12245, '\p{Radical=- No}', "");
Expect(1, 12245, '\p{^Radical=- No}', "");
Expect(1, 12245, '\P{Radical=- No}', "");
Expect(0, 12245, '\P{^Radical=- No}', "");
Error('\p{Is_Radical=-_n/a/}');
Error('\P{Is_Radical=-_n/a/}');
Expect(1, 12246, '\p{Is_Radical=n}', "");
Expect(0, 12246, '\p{^Is_Radical=n}', "");
Expect(0, 12246, '\P{Is_Radical=n}', "");
Expect(1, 12246, '\P{^Is_Radical=n}', "");
Expect(0, 12245, '\p{Is_Radical=n}', "");
Expect(1, 12245, '\p{^Is_Radical=n}', "");
Expect(1, 12245, '\P{Is_Radical=n}', "");
Expect(0, 12245, '\P{^Is_Radical=n}', "");
Expect(1, 12246, '\p{Is_Radical= 	n}', "");
Expect(0, 12246, '\p{^Is_Radical= 	n}', "");
Expect(0, 12246, '\P{Is_Radical= 	n}', "");
Expect(1, 12246, '\P{^Is_Radical= 	n}', "");
Expect(0, 12245, '\p{Is_Radical= 	n}', "");
Expect(1, 12245, '\p{^Is_Radical= 	n}', "");
Expect(1, 12245, '\P{Is_Radical= 	n}', "");
Expect(0, 12245, '\P{^Is_Radical= 	n}', "");
Error('\p{Radical=:=	-f}');
Error('\P{Radical=:=	-f}');
Expect(1, 12246, '\p{Radical=f}', "");
Expect(0, 12246, '\p{^Radical=f}', "");
Expect(0, 12246, '\P{Radical=f}', "");
Expect(1, 12246, '\P{^Radical=f}', "");
Expect(0, 12245, '\p{Radical=f}', "");
Expect(1, 12245, '\p{^Radical=f}', "");
Expect(1, 12245, '\P{Radical=f}', "");
Expect(0, 12245, '\P{^Radical=f}', "");
Expect(1, 12246, '\p{Radical=-_F}', "");
Expect(0, 12246, '\p{^Radical=-_F}', "");
Expect(0, 12246, '\P{Radical=-_F}', "");
Expect(1, 12246, '\P{^Radical=-_F}', "");
Expect(0, 12245, '\p{Radical=-_F}', "");
Expect(1, 12245, '\p{^Radical=-_F}', "");
Expect(1, 12245, '\P{Radical=-_F}', "");
Expect(0, 12245, '\P{^Radical=-_F}', "");
Error('\p{Is_Radical=	-false:=}');
Error('\P{Is_Radical=	-false:=}');
Expect(1, 12246, '\p{Is_Radical=false}', "");
Expect(0, 12246, '\p{^Is_Radical=false}', "");
Expect(0, 12246, '\P{Is_Radical=false}', "");
Expect(1, 12246, '\P{^Is_Radical=false}', "");
Expect(0, 12245, '\p{Is_Radical=false}', "");
Expect(1, 12245, '\p{^Is_Radical=false}', "");
Expect(1, 12245, '\P{Is_Radical=false}', "");
Expect(0, 12245, '\P{^Is_Radical=false}', "");
Expect(1, 12246, '\p{Is_Radical=- False}', "");
Expect(0, 12246, '\p{^Is_Radical=- False}', "");
Expect(0, 12246, '\P{Is_Radical=- False}', "");
Expect(1, 12246, '\P{^Is_Radical=- False}', "");
Expect(0, 12245, '\p{Is_Radical=- False}', "");
Expect(1, 12245, '\p{^Is_Radical=- False}', "");
Expect(1, 12245, '\P{Is_Radical=- False}', "");
Expect(0, 12245, '\P{^Is_Radical=- False}', "");
Error('\p{Radical= /a/yes}');
Error('\P{Radical= /a/yes}');
Expect(1, 12245, '\p{Radical=yes}', "");
Expect(0, 12245, '\p{^Radical=yes}', "");
Expect(0, 12245, '\P{Radical=yes}', "");
Expect(1, 12245, '\P{^Radical=yes}', "");
Expect(0, 12246, '\p{Radical=yes}', "");
Expect(1, 12246, '\p{^Radical=yes}', "");
Expect(1, 12246, '\P{Radical=yes}', "");
Expect(0, 12246, '\P{^Radical=yes}', "");
Expect(1, 12245, '\p{Radical:  yes}', "");
Expect(0, 12245, '\p{^Radical:  yes}', "");
Expect(0, 12245, '\P{Radical:  yes}', "");
Expect(1, 12245, '\P{^Radical:  yes}', "");
Expect(0, 12246, '\p{Radical:  yes}', "");
Expect(1, 12246, '\p{^Radical:  yes}', "");
Expect(1, 12246, '\P{Radical:  yes}', "");
Expect(0, 12246, '\P{^Radical:  yes}', "");
Error('\p{Is_Radical=_	Y/a/}');
Error('\P{Is_Radical=_	Y/a/}');
Expect(1, 12245, '\p{Is_Radical=y}', "");
Expect(0, 12245, '\p{^Is_Radical=y}', "");
Expect(0, 12245, '\P{Is_Radical=y}', "");
Expect(1, 12245, '\P{^Is_Radical=y}', "");
Expect(0, 12246, '\p{Is_Radical=y}', "");
Expect(1, 12246, '\p{^Is_Radical=y}', "");
Expect(1, 12246, '\P{Is_Radical=y}', "");
Expect(0, 12246, '\P{^Is_Radical=y}', "");
Expect(1, 12245, '\p{Is_Radical= y}', "");
Expect(0, 12245, '\p{^Is_Radical= y}', "");
Expect(0, 12245, '\P{Is_Radical= y}', "");
Expect(1, 12245, '\P{^Is_Radical= y}', "");
Expect(0, 12246, '\p{Is_Radical= y}', "");
Expect(1, 12246, '\p{^Is_Radical= y}', "");
Expect(1, 12246, '\P{Is_Radical= y}', "");
Expect(0, 12246, '\P{^Is_Radical= y}', "");
Error('\p{Radical=_/a/T}');
Error('\P{Radical=_/a/T}');
Expect(1, 12245, '\p{Radical=t}', "");
Expect(0, 12245, '\p{^Radical=t}', "");
Expect(0, 12245, '\P{Radical=t}', "");
Expect(1, 12245, '\P{^Radical=t}', "");
Expect(0, 12246, '\p{Radical=t}', "");
Expect(1, 12246, '\p{^Radical=t}', "");
Expect(1, 12246, '\P{Radical=t}', "");
Expect(0, 12246, '\P{^Radical=t}', "");
Expect(1, 12245, '\p{Radical= 	T}', "");
Expect(0, 12245, '\p{^Radical= 	T}', "");
Expect(0, 12245, '\P{Radical= 	T}', "");
Expect(1, 12245, '\P{^Radical= 	T}', "");
Expect(0, 12246, '\p{Radical= 	T}', "");
Expect(1, 12246, '\p{^Radical= 	T}', "");
Expect(1, 12246, '\P{Radical= 	T}', "");
Expect(0, 12246, '\P{^Radical= 	T}', "");
Error('\p{Is_Radical: 	/a/TRUE}');
Error('\P{Is_Radical: 	/a/TRUE}');
Expect(1, 12245, '\p{Is_Radical:	true}', "");
Expect(0, 12245, '\p{^Is_Radical:	true}', "");
Expect(0, 12245, '\P{Is_Radical:	true}', "");
Expect(1, 12245, '\P{^Is_Radical:	true}', "");
Expect(0, 12246, '\p{Is_Radical:	true}', "");
Expect(1, 12246, '\p{^Is_Radical:	true}', "");
Expect(1, 12246, '\P{Is_Radical:	true}', "");
Expect(0, 12246, '\P{^Is_Radical:	true}', "");
Expect(1, 12245, '\p{Is_Radical=--True}', "");
Expect(0, 12245, '\p{^Is_Radical=--True}', "");
Expect(0, 12245, '\P{Is_Radical=--True}', "");
Expect(1, 12245, '\P{^Is_Radical=--True}', "");
Expect(0, 12246, '\p{Is_Radical=--True}', "");
Expect(1, 12246, '\p{^Is_Radical=--True}', "");
Expect(1, 12246, '\P{Is_Radical=--True}', "");
Expect(0, 12246, '\P{^Is_Radical=--True}', "");
Error('\p{Sentence_Break=:=_ATerm}');
Error('\P{Sentence_Break=:=_ATerm}');
Expect(1, 65294, '\p{Sentence_Break:   aterm}', "");
Expect(0, 65294, '\p{^Sentence_Break:   aterm}', "");
Expect(0, 65294, '\P{Sentence_Break:   aterm}', "");
Expect(1, 65294, '\P{^Sentence_Break:   aterm}', "");
Expect(0, 65295, '\p{Sentence_Break:   aterm}', "");
Expect(1, 65295, '\p{^Sentence_Break:   aterm}', "");
Expect(1, 65295, '\P{Sentence_Break:   aterm}', "");
Expect(0, 65295, '\P{^Sentence_Break:   aterm}', "");
Expect(1, 65294, '\p{Sentence_Break= 	aterm}', "");
Expect(0, 65294, '\p{^Sentence_Break= 	aterm}', "");
Expect(0, 65294, '\P{Sentence_Break= 	aterm}', "");
Expect(1, 65294, '\P{^Sentence_Break= 	aterm}', "");
Expect(0, 65295, '\p{Sentence_Break= 	aterm}', "");
Expect(1, 65295, '\p{^Sentence_Break= 	aterm}', "");
Expect(1, 65295, '\P{Sentence_Break= 	aterm}', "");
Expect(0, 65295, '\P{^Sentence_Break= 	aterm}', "");
Error('\p{SB=/a/ _AT}');
Error('\P{SB=/a/ _AT}');
Expect(1, 65294, '\p{SB=at}', "");
Expect(0, 65294, '\p{^SB=at}', "");
Expect(0, 65294, '\P{SB=at}', "");
Expect(1, 65294, '\P{^SB=at}', "");
Expect(0, 65295, '\p{SB=at}', "");
Expect(1, 65295, '\p{^SB=at}', "");
Expect(1, 65295, '\P{SB=at}', "");
Expect(0, 65295, '\P{^SB=at}', "");
Expect(1, 65294, '\p{SB=-AT}', "");
Expect(0, 65294, '\p{^SB=-AT}', "");
Expect(0, 65294, '\P{SB=-AT}', "");
Expect(1, 65294, '\P{^SB=-AT}', "");
Expect(0, 65295, '\p{SB=-AT}', "");
Expect(1, 65295, '\p{^SB=-AT}', "");
Expect(1, 65295, '\P{SB=-AT}', "");
Expect(0, 65295, '\P{^SB=-AT}', "");
Error('\p{Is_Sentence_Break=:=_ ATerm}');
Error('\P{Is_Sentence_Break=:=_ ATerm}');
Expect(1, 65294, '\p{Is_Sentence_Break:aterm}', "");
Expect(0, 65294, '\p{^Is_Sentence_Break:aterm}', "");
Expect(0, 65294, '\P{Is_Sentence_Break:aterm}', "");
Expect(1, 65294, '\P{^Is_Sentence_Break:aterm}', "");
Expect(0, 65295, '\p{Is_Sentence_Break:aterm}', "");
Expect(1, 65295, '\p{^Is_Sentence_Break:aterm}', "");
Expect(1, 65295, '\P{Is_Sentence_Break:aterm}', "");
Expect(0, 65295, '\P{^Is_Sentence_Break:aterm}', "");
Expect(1, 65294, '\p{Is_Sentence_Break=_	ATERM}', "");
Expect(0, 65294, '\p{^Is_Sentence_Break=_	ATERM}', "");
Expect(0, 65294, '\P{Is_Sentence_Break=_	ATERM}', "");
Expect(1, 65294, '\P{^Is_Sentence_Break=_	ATERM}', "");
Expect(0, 65295, '\p{Is_Sentence_Break=_	ATERM}', "");
Expect(1, 65295, '\p{^Is_Sentence_Break=_	ATERM}', "");
Expect(1, 65295, '\P{Is_Sentence_Break=_	ATERM}', "");
Expect(0, 65295, '\P{^Is_Sentence_Break=_	ATERM}', "");
Error('\p{Is_SB: /a/__AT}');
Error('\P{Is_SB: /a/__AT}');
Expect(1, 65294, '\p{Is_SB=at}', "");
Expect(0, 65294, '\p{^Is_SB=at}', "");
Expect(0, 65294, '\P{Is_SB=at}', "");
Expect(1, 65294, '\P{^Is_SB=at}', "");
Expect(0, 65295, '\p{Is_SB=at}', "");
Expect(1, 65295, '\p{^Is_SB=at}', "");
Expect(1, 65295, '\P{Is_SB=at}', "");
Expect(0, 65295, '\P{^Is_SB=at}', "");
Expect(1, 65294, '\p{Is_SB: 	AT}', "");
Expect(0, 65294, '\p{^Is_SB: 	AT}', "");
Expect(0, 65294, '\P{Is_SB: 	AT}', "");
Expect(1, 65294, '\P{^Is_SB: 	AT}', "");
Expect(0, 65295, '\p{Is_SB: 	AT}', "");
Expect(1, 65295, '\p{^Is_SB: 	AT}', "");
Expect(1, 65295, '\P{Is_SB: 	AT}', "");
Expect(0, 65295, '\P{^Is_SB: 	AT}', "");
Error('\p{Sentence_Break=/a/ Close}');
Error('\P{Sentence_Break=/a/ Close}');
Expect(1, 65379, '\p{Sentence_Break=close}', "");
Expect(0, 65379, '\p{^Sentence_Break=close}', "");
Expect(0, 65379, '\P{Sentence_Break=close}', "");
Expect(1, 65379, '\P{^Sentence_Break=close}', "");
Expect(0, 65380, '\p{Sentence_Break=close}', "");
Expect(1, 65380, '\p{^Sentence_Break=close}', "");
Expect(1, 65380, '\P{Sentence_Break=close}', "");
Expect(0, 65380, '\P{^Sentence_Break=close}', "");
Expect(1, 65379, '\p{Sentence_Break=	_CLOSE}', "");
Expect(0, 65379, '\p{^Sentence_Break=	_CLOSE}', "");
Expect(0, 65379, '\P{Sentence_Break=	_CLOSE}', "");
Expect(1, 65379, '\P{^Sentence_Break=	_CLOSE}', "");
Expect(0, 65380, '\p{Sentence_Break=	_CLOSE}', "");
Expect(1, 65380, '\p{^Sentence_Break=	_CLOSE}', "");
Expect(1, 65380, '\P{Sentence_Break=	_CLOSE}', "");
Expect(0, 65380, '\P{^Sentence_Break=	_CLOSE}', "");
Error('\p{SB=_cl/a/}');
Error('\P{SB=_cl/a/}');
Expect(1, 65379, '\p{SB=cl}', "");
Expect(0, 65379, '\p{^SB=cl}', "");
Expect(0, 65379, '\P{SB=cl}', "");
Expect(1, 65379, '\P{^SB=cl}', "");
Expect(0, 65380, '\p{SB=cl}', "");
Expect(1, 65380, '\p{^SB=cl}', "");
Expect(1, 65380, '\P{SB=cl}', "");
Expect(0, 65380, '\P{^SB=cl}', "");
Expect(1, 65379, '\p{SB=	CL}', "");
Expect(0, 65379, '\p{^SB=	CL}', "");
Expect(0, 65379, '\P{SB=	CL}', "");
Expect(1, 65379, '\P{^SB=	CL}', "");
Expect(0, 65380, '\p{SB=	CL}', "");
Expect(1, 65380, '\p{^SB=	CL}', "");
Expect(1, 65380, '\P{SB=	CL}', "");
Expect(0, 65380, '\P{^SB=	CL}', "");
Error('\p{Is_Sentence_Break= 	CLOSE/a/}');
Error('\P{Is_Sentence_Break= 	CLOSE/a/}');
Expect(1, 65379, '\p{Is_Sentence_Break=close}', "");
Expect(0, 65379, '\p{^Is_Sentence_Break=close}', "");
Expect(0, 65379, '\P{Is_Sentence_Break=close}', "");
Expect(1, 65379, '\P{^Is_Sentence_Break=close}', "");
Expect(0, 65380, '\p{Is_Sentence_Break=close}', "");
Expect(1, 65380, '\p{^Is_Sentence_Break=close}', "");
Expect(1, 65380, '\P{Is_Sentence_Break=close}', "");
Expect(0, 65380, '\P{^Is_Sentence_Break=close}', "");
Expect(1, 65379, '\p{Is_Sentence_Break=		Close}', "");
Expect(0, 65379, '\p{^Is_Sentence_Break=		Close}', "");
Expect(0, 65379, '\P{Is_Sentence_Break=		Close}', "");
Expect(1, 65379, '\P{^Is_Sentence_Break=		Close}', "");
Expect(0, 65380, '\p{Is_Sentence_Break=		Close}', "");
Expect(1, 65380, '\p{^Is_Sentence_Break=		Close}', "");
Expect(1, 65380, '\P{Is_Sentence_Break=		Close}', "");
Expect(0, 65380, '\P{^Is_Sentence_Break=		Close}', "");
Error('\p{Is_SB=_/a/CL}');
Error('\P{Is_SB=_/a/CL}');
Expect(1, 65379, '\p{Is_SB=cl}', "");
Expect(0, 65379, '\p{^Is_SB=cl}', "");
Expect(0, 65379, '\P{Is_SB=cl}', "");
Expect(1, 65379, '\P{^Is_SB=cl}', "");
Expect(0, 65380, '\p{Is_SB=cl}', "");
Expect(1, 65380, '\p{^Is_SB=cl}', "");
Expect(1, 65380, '\P{Is_SB=cl}', "");
Expect(0, 65380, '\P{^Is_SB=cl}', "");
Expect(1, 65379, '\p{Is_SB: -CL}', "");
Expect(0, 65379, '\p{^Is_SB: -CL}', "");
Expect(0, 65379, '\P{Is_SB: -CL}', "");
Expect(1, 65379, '\P{^Is_SB: -CL}', "");
Expect(0, 65380, '\p{Is_SB: -CL}', "");
Expect(1, 65380, '\p{^Is_SB: -CL}', "");
Expect(1, 65380, '\P{Is_SB: -CL}', "");
Expect(0, 65380, '\P{^Is_SB: -CL}', "");
Error('\p{Sentence_Break=/a/--CR}');
Error('\P{Sentence_Break=/a/--CR}');
Expect(1, 13, '\p{Sentence_Break=cr}', "");
Expect(0, 13, '\p{^Sentence_Break=cr}', "");
Expect(0, 13, '\P{Sentence_Break=cr}', "");
Expect(1, 13, '\P{^Sentence_Break=cr}', "");
Expect(0, 14, '\p{Sentence_Break=cr}', "");
Expect(1, 14, '\p{^Sentence_Break=cr}', "");
Expect(1, 14, '\P{Sentence_Break=cr}', "");
Expect(0, 14, '\P{^Sentence_Break=cr}', "");
Expect(1, 13, '\p{Sentence_Break=_ CR}', "");
Expect(0, 13, '\p{^Sentence_Break=_ CR}', "");
Expect(0, 13, '\P{Sentence_Break=_ CR}', "");
Expect(1, 13, '\P{^Sentence_Break=_ CR}', "");
Expect(0, 14, '\p{Sentence_Break=_ CR}', "");
Expect(1, 14, '\p{^Sentence_Break=_ CR}', "");
Expect(1, 14, '\P{Sentence_Break=_ CR}', "");
Expect(0, 14, '\P{^Sentence_Break=_ CR}', "");
Error('\p{SB=:=	CR}');
Error('\P{SB=:=	CR}');
Expect(1, 13, '\p{SB=cr}', "");
Expect(0, 13, '\p{^SB=cr}', "");
Expect(0, 13, '\P{SB=cr}', "");
Expect(1, 13, '\P{^SB=cr}', "");
Expect(0, 14, '\p{SB=cr}', "");
Expect(1, 14, '\p{^SB=cr}', "");
Expect(1, 14, '\P{SB=cr}', "");
Expect(0, 14, '\P{^SB=cr}', "");
Expect(1, 13, '\p{SB= -CR}', "");
Expect(0, 13, '\p{^SB= -CR}', "");
Expect(0, 13, '\P{SB= -CR}', "");
Expect(1, 13, '\P{^SB= -CR}', "");
Expect(0, 14, '\p{SB= -CR}', "");
Expect(1, 14, '\p{^SB= -CR}', "");
Expect(1, 14, '\P{SB= -CR}', "");
Expect(0, 14, '\P{^SB= -CR}', "");
Error('\p{Is_Sentence_Break=:=CR}');
Error('\P{Is_Sentence_Break=:=CR}');
Expect(1, 13, '\p{Is_Sentence_Break=cr}', "");
Expect(0, 13, '\p{^Is_Sentence_Break=cr}', "");
Expect(0, 13, '\P{Is_Sentence_Break=cr}', "");
Expect(1, 13, '\P{^Is_Sentence_Break=cr}', "");
Expect(0, 14, '\p{Is_Sentence_Break=cr}', "");
Expect(1, 14, '\p{^Is_Sentence_Break=cr}', "");
Expect(1, 14, '\P{Is_Sentence_Break=cr}', "");
Expect(0, 14, '\P{^Is_Sentence_Break=cr}', "");
Expect(1, 13, '\p{Is_Sentence_Break= CR}', "");
Expect(0, 13, '\p{^Is_Sentence_Break= CR}', "");
Expect(0, 13, '\P{Is_Sentence_Break= CR}', "");
Expect(1, 13, '\P{^Is_Sentence_Break= CR}', "");
Expect(0, 14, '\p{Is_Sentence_Break= CR}', "");
Expect(1, 14, '\p{^Is_Sentence_Break= CR}', "");
Expect(1, 14, '\P{Is_Sentence_Break= CR}', "");
Expect(0, 14, '\P{^Is_Sentence_Break= CR}', "");
Error('\p{Is_SB=_-cr/a/}');
Error('\P{Is_SB=_-cr/a/}');
Expect(1, 13, '\p{Is_SB=cr}', "");
Expect(0, 13, '\p{^Is_SB=cr}', "");
Expect(0, 13, '\P{Is_SB=cr}', "");
Expect(1, 13, '\P{^Is_SB=cr}', "");
Expect(0, 14, '\p{Is_SB=cr}', "");
Expect(1, 14, '\p{^Is_SB=cr}', "");
Expect(1, 14, '\P{Is_SB=cr}', "");
Expect(0, 14, '\P{^Is_SB=cr}', "");
Expect(1, 13, '\p{Is_SB=	CR}', "");
Expect(0, 13, '\p{^Is_SB=	CR}', "");
Expect(0, 13, '\P{Is_SB=	CR}', "");
Expect(1, 13, '\P{^Is_SB=	CR}', "");
Expect(0, 14, '\p{Is_SB=	CR}', "");
Expect(1, 14, '\p{^Is_SB=	CR}', "");
Expect(1, 14, '\P{Is_SB=	CR}', "");
Expect(0, 14, '\P{^Is_SB=	CR}', "");
Error('\p{Sentence_Break=/a/extend}');
Error('\P{Sentence_Break=/a/extend}');
Expect(1, 917999, '\p{Sentence_Break=extend}', "");
Expect(0, 917999, '\p{^Sentence_Break=extend}', "");
Expect(0, 917999, '\P{Sentence_Break=extend}', "");
Expect(1, 917999, '\P{^Sentence_Break=extend}', "");
Expect(0, 918000, '\p{Sentence_Break=extend}', "");
Expect(1, 918000, '\p{^Sentence_Break=extend}', "");
Expect(1, 918000, '\P{Sentence_Break=extend}', "");
Expect(0, 918000, '\P{^Sentence_Break=extend}', "");
Expect(1, 917999, '\p{Sentence_Break= 	EXTEND}', "");
Expect(0, 917999, '\p{^Sentence_Break= 	EXTEND}', "");
Expect(0, 917999, '\P{Sentence_Break= 	EXTEND}', "");
Expect(1, 917999, '\P{^Sentence_Break= 	EXTEND}', "");
Expect(0, 918000, '\p{Sentence_Break= 	EXTEND}', "");
Expect(1, 918000, '\p{^Sentence_Break= 	EXTEND}', "");
Expect(1, 918000, '\P{Sentence_Break= 	EXTEND}', "");
Expect(0, 918000, '\P{^Sentence_Break= 	EXTEND}', "");
Error('\p{SB=-ex/a/}');
Error('\P{SB=-ex/a/}');
Expect(1, 917999, '\p{SB=ex}', "");
Expect(0, 917999, '\p{^SB=ex}', "");
Expect(0, 917999, '\P{SB=ex}', "");
Expect(1, 917999, '\P{^SB=ex}', "");
Expect(0, 918000, '\p{SB=ex}', "");
Expect(1, 918000, '\p{^SB=ex}', "");
Expect(1, 918000, '\P{SB=ex}', "");
Expect(0, 918000, '\P{^SB=ex}', "");
Expect(1, 917999, '\p{SB= ex}', "");
Expect(0, 917999, '\p{^SB= ex}', "");
Expect(0, 917999, '\P{SB= ex}', "");
Expect(1, 917999, '\P{^SB= ex}', "");
Expect(0, 918000, '\p{SB= ex}', "");
Expect(1, 918000, '\p{^SB= ex}', "");
Expect(1, 918000, '\P{SB= ex}', "");
Expect(0, 918000, '\P{^SB= ex}', "");
Error('\p{Is_Sentence_Break=/a/Extend}');
Error('\P{Is_Sentence_Break=/a/Extend}');
Expect(1, 917999, '\p{Is_Sentence_Break=extend}', "");
Expect(0, 917999, '\p{^Is_Sentence_Break=extend}', "");
Expect(0, 917999, '\P{Is_Sentence_Break=extend}', "");
Expect(1, 917999, '\P{^Is_Sentence_Break=extend}', "");
Expect(0, 918000, '\p{Is_Sentence_Break=extend}', "");
Expect(1, 918000, '\p{^Is_Sentence_Break=extend}', "");
Expect(1, 918000, '\P{Is_Sentence_Break=extend}', "");
Expect(0, 918000, '\P{^Is_Sentence_Break=extend}', "");
Expect(1, 917999, '\p{Is_Sentence_Break=  EXTEND}', "");
Expect(0, 917999, '\p{^Is_Sentence_Break=  EXTEND}', "");
Expect(0, 917999, '\P{Is_Sentence_Break=  EXTEND}', "");
Expect(1, 917999, '\P{^Is_Sentence_Break=  EXTEND}', "");
Expect(0, 918000, '\p{Is_Sentence_Break=  EXTEND}', "");
Expect(1, 918000, '\p{^Is_Sentence_Break=  EXTEND}', "");
Expect(1, 918000, '\P{Is_Sentence_Break=  EXTEND}', "");
Expect(0, 918000, '\P{^Is_Sentence_Break=  EXTEND}', "");
Error('\p{Is_SB=_/a/ex}');
Error('\P{Is_SB=_/a/ex}');
Expect(1, 917999, '\p{Is_SB:   ex}', "");
Expect(0, 917999, '\p{^Is_SB:   ex}', "");
Expect(0, 917999, '\P{Is_SB:   ex}', "");
Expect(1, 917999, '\P{^Is_SB:   ex}', "");
Expect(0, 918000, '\p{Is_SB:   ex}', "");
Expect(1, 918000, '\p{^Is_SB:   ex}', "");
Expect(1, 918000, '\P{Is_SB:   ex}', "");
Expect(0, 918000, '\P{^Is_SB:   ex}', "");
Expect(1, 917999, '\p{Is_SB= 	EX}', "");
Expect(0, 917999, '\p{^Is_SB= 	EX}', "");
Expect(0, 917999, '\P{Is_SB= 	EX}', "");
Expect(1, 917999, '\P{^Is_SB= 	EX}', "");
Expect(0, 918000, '\p{Is_SB= 	EX}', "");
Expect(1, 918000, '\p{^Is_SB= 	EX}', "");
Expect(1, 918000, '\P{Is_SB= 	EX}', "");
Expect(0, 918000, '\P{^Is_SB= 	EX}', "");
Error('\p{Sentence_Break=:= Format}');
Error('\P{Sentence_Break=:= Format}');
Expect(1, 917631, '\p{Sentence_Break=format}', "");
Expect(0, 917631, '\p{^Sentence_Break=format}', "");
Expect(0, 917631, '\P{Sentence_Break=format}', "");
Expect(1, 917631, '\P{^Sentence_Break=format}', "");
Expect(0, 917632, '\p{Sentence_Break=format}', "");
Expect(1, 917632, '\p{^Sentence_Break=format}', "");
Expect(1, 917632, '\P{Sentence_Break=format}', "");
Expect(0, 917632, '\P{^Sentence_Break=format}', "");
Expect(1, 917631, '\p{Sentence_Break=	-Format}', "");
Expect(0, 917631, '\p{^Sentence_Break=	-Format}', "");
Expect(0, 917631, '\P{Sentence_Break=	-Format}', "");
Expect(1, 917631, '\P{^Sentence_Break=	-Format}', "");
Expect(0, 917632, '\p{Sentence_Break=	-Format}', "");
Expect(1, 917632, '\p{^Sentence_Break=	-Format}', "");
Expect(1, 917632, '\P{Sentence_Break=	-Format}', "");
Expect(0, 917632, '\P{^Sentence_Break=	-Format}', "");
Error('\p{SB=_	FO/a/}');
Error('\P{SB=_	FO/a/}');
Expect(1, 917631, '\p{SB:	fo}', "");
Expect(0, 917631, '\p{^SB:	fo}', "");
Expect(0, 917631, '\P{SB:	fo}', "");
Expect(1, 917631, '\P{^SB:	fo}', "");
Expect(0, 917632, '\p{SB:	fo}', "");
Expect(1, 917632, '\p{^SB:	fo}', "");
Expect(1, 917632, '\P{SB:	fo}', "");
Expect(0, 917632, '\P{^SB:	fo}', "");
Expect(1, 917631, '\p{SB=- FO}', "");
Expect(0, 917631, '\p{^SB=- FO}', "");
Expect(0, 917631, '\P{SB=- FO}', "");
Expect(1, 917631, '\P{^SB=- FO}', "");
Expect(0, 917632, '\p{SB=- FO}', "");
Expect(1, 917632, '\p{^SB=- FO}', "");
Expect(1, 917632, '\P{SB=- FO}', "");
Expect(0, 917632, '\P{^SB=- FO}', "");
Error('\p{Is_Sentence_Break=/a/ FORMAT}');
Error('\P{Is_Sentence_Break=/a/ FORMAT}');
Expect(1, 917631, '\p{Is_Sentence_Break=format}', "");
Expect(0, 917631, '\p{^Is_Sentence_Break=format}', "");
Expect(0, 917631, '\P{Is_Sentence_Break=format}', "");
Expect(1, 917631, '\P{^Is_Sentence_Break=format}', "");
Expect(0, 917632, '\p{Is_Sentence_Break=format}', "");
Expect(1, 917632, '\p{^Is_Sentence_Break=format}', "");
Expect(1, 917632, '\P{Is_Sentence_Break=format}', "");
Expect(0, 917632, '\P{^Is_Sentence_Break=format}', "");
Expect(1, 917631, '\p{Is_Sentence_Break=-	Format}', "");
Expect(0, 917631, '\p{^Is_Sentence_Break=-	Format}', "");
Expect(0, 917631, '\P{Is_Sentence_Break=-	Format}', "");
Expect(1, 917631, '\P{^Is_Sentence_Break=-	Format}', "");
Expect(0, 917632, '\p{Is_Sentence_Break=-	Format}', "");
Expect(1, 917632, '\p{^Is_Sentence_Break=-	Format}', "");
Expect(1, 917632, '\P{Is_Sentence_Break=-	Format}', "");
Expect(0, 917632, '\P{^Is_Sentence_Break=-	Format}', "");
Error('\p{Is_SB=/a/-_FO}');
Error('\P{Is_SB=/a/-_FO}');
Expect(1, 917631, '\p{Is_SB=fo}', "");
Expect(0, 917631, '\p{^Is_SB=fo}', "");
Expect(0, 917631, '\P{Is_SB=fo}', "");
Expect(1, 917631, '\P{^Is_SB=fo}', "");
Expect(0, 917632, '\p{Is_SB=fo}', "");
Expect(1, 917632, '\p{^Is_SB=fo}', "");
Expect(1, 917632, '\P{Is_SB=fo}', "");
Expect(0, 917632, '\P{^Is_SB=fo}', "");
Expect(1, 917631, '\p{Is_SB=	_FO}', "");
Expect(0, 917631, '\p{^Is_SB=	_FO}', "");
Expect(0, 917631, '\P{Is_SB=	_FO}', "");
Expect(1, 917631, '\P{^Is_SB=	_FO}', "");
Expect(0, 917632, '\p{Is_SB=	_FO}', "");
Expect(1, 917632, '\p{^Is_SB=	_FO}', "");
Expect(1, 917632, '\P{Is_SB=	_FO}', "");
Expect(0, 917632, '\P{^Is_SB=	_FO}', "");
Error('\p{Sentence_Break=/a/ OLetter}');
Error('\P{Sentence_Break=/a/ OLetter}');
Expect(1, 195101, '\p{Sentence_Break=oletter}', "");
Expect(0, 195101, '\p{^Sentence_Break=oletter}', "");
Expect(0, 195101, '\P{Sentence_Break=oletter}', "");
Expect(1, 195101, '\P{^Sentence_Break=oletter}', "");
Expect(0, 195102, '\p{Sentence_Break=oletter}', "");
Expect(1, 195102, '\p{^Sentence_Break=oletter}', "");
Expect(1, 195102, '\P{Sentence_Break=oletter}', "");
Expect(0, 195102, '\P{^Sentence_Break=oletter}', "");
Expect(1, 195101, '\p{Sentence_Break= oletter}', "");
Expect(0, 195101, '\p{^Sentence_Break= oletter}', "");
Expect(0, 195101, '\P{Sentence_Break= oletter}', "");
Expect(1, 195101, '\P{^Sentence_Break= oletter}', "");
Expect(0, 195102, '\p{Sentence_Break= oletter}', "");
Expect(1, 195102, '\p{^Sentence_Break= oletter}', "");
Expect(1, 195102, '\P{Sentence_Break= oletter}', "");
Expect(0, 195102, '\P{^Sentence_Break= oletter}', "");
Error('\p{SB=_LE/a/}');
Error('\P{SB=_LE/a/}');
Expect(1, 195101, '\p{SB=le}', "");
Expect(0, 195101, '\p{^SB=le}', "");
Expect(0, 195101, '\P{SB=le}', "");
Expect(1, 195101, '\P{^SB=le}', "");
Expect(0, 195102, '\p{SB=le}', "");
Expect(1, 195102, '\p{^SB=le}', "");
Expect(1, 195102, '\P{SB=le}', "");
Expect(0, 195102, '\P{^SB=le}', "");
Expect(1, 195101, '\p{SB=	le}', "");
Expect(0, 195101, '\p{^SB=	le}', "");
Expect(0, 195101, '\P{SB=	le}', "");
Expect(1, 195101, '\P{^SB=	le}', "");
Expect(0, 195102, '\p{SB=	le}', "");
Expect(1, 195102, '\p{^SB=	le}', "");
Expect(1, 195102, '\P{SB=	le}', "");
Expect(0, 195102, '\P{^SB=	le}', "");
Error('\p{Is_Sentence_Break=_	oletter:=}');
Error('\P{Is_Sentence_Break=_	oletter:=}');
Expect(1, 195101, '\p{Is_Sentence_Break=oletter}', "");
Expect(0, 195101, '\p{^Is_Sentence_Break=oletter}', "");
Expect(0, 195101, '\P{Is_Sentence_Break=oletter}', "");
Expect(1, 195101, '\P{^Is_Sentence_Break=oletter}', "");
Expect(0, 195102, '\p{Is_Sentence_Break=oletter}', "");
Expect(1, 195102, '\p{^Is_Sentence_Break=oletter}', "");
Expect(1, 195102, '\P{Is_Sentence_Break=oletter}', "");
Expect(0, 195102, '\P{^Is_Sentence_Break=oletter}', "");
Expect(1, 195101, '\p{Is_Sentence_Break= OLetter}', "");
Expect(0, 195101, '\p{^Is_Sentence_Break= OLetter}', "");
Expect(0, 195101, '\P{Is_Sentence_Break= OLetter}', "");
Expect(1, 195101, '\P{^Is_Sentence_Break= OLetter}', "");
Expect(0, 195102, '\p{Is_Sentence_Break= OLetter}', "");
Expect(1, 195102, '\p{^Is_Sentence_Break= OLetter}', "");
Expect(1, 195102, '\P{Is_Sentence_Break= OLetter}', "");
Expect(0, 195102, '\P{^Is_Sentence_Break= OLetter}', "");
Error('\p{Is_SB=:=-_LE}');
Error('\P{Is_SB=:=-_LE}');
Expect(1, 195101, '\p{Is_SB=le}', "");
Expect(0, 195101, '\p{^Is_SB=le}', "");
Expect(0, 195101, '\P{Is_SB=le}', "");
Expect(1, 195101, '\P{^Is_SB=le}', "");
Expect(0, 195102, '\p{Is_SB=le}', "");
Expect(1, 195102, '\p{^Is_SB=le}', "");
Expect(1, 195102, '\P{Is_SB=le}', "");
Expect(0, 195102, '\P{^Is_SB=le}', "");
Expect(1, 195101, '\p{Is_SB=_	LE}', "");
Expect(0, 195101, '\p{^Is_SB=_	LE}', "");
Expect(0, 195101, '\P{Is_SB=_	LE}', "");
Expect(1, 195101, '\P{^Is_SB=_	LE}', "");
Expect(0, 195102, '\p{Is_SB=_	LE}', "");
Expect(1, 195102, '\p{^Is_SB=_	LE}', "");
Expect(1, 195102, '\P{Is_SB=_	LE}', "");
Expect(0, 195102, '\P{^Is_SB=_	LE}', "");
Error('\p{Sentence_Break=	:=LF}');
Error('\P{Sentence_Break=	:=LF}');
Expect(1, 10, '\p{Sentence_Break=lf}', "");
Expect(0, 10, '\p{^Sentence_Break=lf}', "");
Expect(0, 10, '\P{Sentence_Break=lf}', "");
Expect(1, 10, '\P{^Sentence_Break=lf}', "");
Expect(0, 11, '\p{Sentence_Break=lf}', "");
Expect(1, 11, '\p{^Sentence_Break=lf}', "");
Expect(1, 11, '\P{Sentence_Break=lf}', "");
Expect(0, 11, '\P{^Sentence_Break=lf}', "");
Expect(1, 10, '\p{Sentence_Break:   -_lf}', "");
Expect(0, 10, '\p{^Sentence_Break:   -_lf}', "");
Expect(0, 10, '\P{Sentence_Break:   -_lf}', "");
Expect(1, 10, '\P{^Sentence_Break:   -_lf}', "");
Expect(0, 11, '\p{Sentence_Break:   -_lf}', "");
Expect(1, 11, '\p{^Sentence_Break:   -_lf}', "");
Expect(1, 11, '\P{Sentence_Break:   -_lf}', "");
Expect(0, 11, '\P{^Sentence_Break:   -_lf}', "");
Error('\p{SB= lf:=}');
Error('\P{SB= lf:=}');
Expect(1, 10, '\p{SB=lf}', "");
Expect(0, 10, '\p{^SB=lf}', "");
Expect(0, 10, '\P{SB=lf}', "");
Expect(1, 10, '\P{^SB=lf}', "");
Expect(0, 11, '\p{SB=lf}', "");
Expect(1, 11, '\p{^SB=lf}', "");
Expect(1, 11, '\P{SB=lf}', "");
Expect(0, 11, '\P{^SB=lf}', "");
Expect(1, 10, '\p{SB=	 LF}', "");
Expect(0, 10, '\p{^SB=	 LF}', "");
Expect(0, 10, '\P{SB=	 LF}', "");
Expect(1, 10, '\P{^SB=	 LF}', "");
Expect(0, 11, '\p{SB=	 LF}', "");
Expect(1, 11, '\p{^SB=	 LF}', "");
Expect(1, 11, '\P{SB=	 LF}', "");
Expect(0, 11, '\P{^SB=	 LF}', "");
Error('\p{Is_Sentence_Break=:=	-lf}');
Error('\P{Is_Sentence_Break=:=	-lf}');
Expect(1, 10, '\p{Is_Sentence_Break=lf}', "");
Expect(0, 10, '\p{^Is_Sentence_Break=lf}', "");
Expect(0, 10, '\P{Is_Sentence_Break=lf}', "");
Expect(1, 10, '\P{^Is_Sentence_Break=lf}', "");
Expect(0, 11, '\p{Is_Sentence_Break=lf}', "");
Expect(1, 11, '\p{^Is_Sentence_Break=lf}', "");
Expect(1, 11, '\P{Is_Sentence_Break=lf}', "");
Expect(0, 11, '\P{^Is_Sentence_Break=lf}', "");
Expect(1, 10, '\p{Is_Sentence_Break=	 LF}', "");
Expect(0, 10, '\p{^Is_Sentence_Break=	 LF}', "");
Expect(0, 10, '\P{Is_Sentence_Break=	 LF}', "");
Expect(1, 10, '\P{^Is_Sentence_Break=	 LF}', "");
Expect(0, 11, '\p{Is_Sentence_Break=	 LF}', "");
Expect(1, 11, '\p{^Is_Sentence_Break=	 LF}', "");
Expect(1, 11, '\P{Is_Sentence_Break=	 LF}', "");
Expect(0, 11, '\P{^Is_Sentence_Break=	 LF}', "");
Error('\p{Is_SB: /a/_-LF}');
Error('\P{Is_SB: /a/_-LF}');
Expect(1, 10, '\p{Is_SB:	lf}', "");
Expect(0, 10, '\p{^Is_SB:	lf}', "");
Expect(0, 10, '\P{Is_SB:	lf}', "");
Expect(1, 10, '\P{^Is_SB:	lf}', "");
Expect(0, 11, '\p{Is_SB:	lf}', "");
Expect(1, 11, '\p{^Is_SB:	lf}', "");
Expect(1, 11, '\P{Is_SB:	lf}', "");
Expect(0, 11, '\P{^Is_SB:	lf}', "");
Expect(1, 10, '\p{Is_SB: __LF}', "");
Expect(0, 10, '\p{^Is_SB: __LF}', "");
Expect(0, 10, '\P{Is_SB: __LF}', "");
Expect(1, 10, '\P{^Is_SB: __LF}', "");
Expect(0, 11, '\p{Is_SB: __LF}', "");
Expect(1, 11, '\p{^Is_SB: __LF}', "");
Expect(1, 11, '\P{Is_SB: __LF}', "");
Expect(0, 11, '\P{^Is_SB: __LF}', "");
Error('\p{Sentence_Break=-Lower:=}');
Error('\P{Sentence_Break=-Lower:=}');
Expect(1, 120779, '\p{Sentence_Break=lower}', "");
Expect(0, 120779, '\p{^Sentence_Break=lower}', "");
Expect(0, 120779, '\P{Sentence_Break=lower}', "");
Expect(1, 120779, '\P{^Sentence_Break=lower}', "");
Expect(0, 120780, '\p{Sentence_Break=lower}', "");
Expect(1, 120780, '\p{^Sentence_Break=lower}', "");
Expect(1, 120780, '\P{Sentence_Break=lower}', "");
Expect(0, 120780, '\P{^Sentence_Break=lower}', "");
Expect(1, 120779, '\p{Sentence_Break=-_Lower}', "");
Expect(0, 120779, '\p{^Sentence_Break=-_Lower}', "");
Expect(0, 120779, '\P{Sentence_Break=-_Lower}', "");
Expect(1, 120779, '\P{^Sentence_Break=-_Lower}', "");
Expect(0, 120780, '\p{Sentence_Break=-_Lower}', "");
Expect(1, 120780, '\p{^Sentence_Break=-_Lower}', "");
Expect(1, 120780, '\P{Sentence_Break=-_Lower}', "");
Expect(0, 120780, '\P{^Sentence_Break=-_Lower}', "");
Error('\p{SB=_/a/LO}');
Error('\P{SB=_/a/LO}');
Expect(1, 120779, '\p{SB=lo}', "");
Expect(0, 120779, '\p{^SB=lo}', "");
Expect(0, 120779, '\P{SB=lo}', "");
Expect(1, 120779, '\P{^SB=lo}', "");
Expect(0, 120780, '\p{SB=lo}', "");
Expect(1, 120780, '\p{^SB=lo}', "");
Expect(1, 120780, '\P{SB=lo}', "");
Expect(0, 120780, '\P{^SB=lo}', "");
Expect(1, 120779, '\p{SB=- LO}', "");
Expect(0, 120779, '\p{^SB=- LO}', "");
Expect(0, 120779, '\P{SB=- LO}', "");
Expect(1, 120779, '\P{^SB=- LO}', "");
Expect(0, 120780, '\p{SB=- LO}', "");
Expect(1, 120780, '\p{^SB=- LO}', "");
Expect(1, 120780, '\P{SB=- LO}', "");
Expect(0, 120780, '\P{^SB=- LO}', "");
Error('\p{Is_Sentence_Break=/a/	-Lower}');
Error('\P{Is_Sentence_Break=/a/	-Lower}');
Expect(1, 120779, '\p{Is_Sentence_Break=lower}', "");
Expect(0, 120779, '\p{^Is_Sentence_Break=lower}', "");
Expect(0, 120779, '\P{Is_Sentence_Break=lower}', "");
Expect(1, 120779, '\P{^Is_Sentence_Break=lower}', "");
Expect(0, 120780, '\p{Is_Sentence_Break=lower}', "");
Expect(1, 120780, '\p{^Is_Sentence_Break=lower}', "");
Expect(1, 120780, '\P{Is_Sentence_Break=lower}', "");
Expect(0, 120780, '\P{^Is_Sentence_Break=lower}', "");
Expect(1, 120779, '\p{Is_Sentence_Break=		lower}', "");
Expect(0, 120779, '\p{^Is_Sentence_Break=		lower}', "");
Expect(0, 120779, '\P{Is_Sentence_Break=		lower}', "");
Expect(1, 120779, '\P{^Is_Sentence_Break=		lower}', "");
Expect(0, 120780, '\p{Is_Sentence_Break=		lower}', "");
Expect(1, 120780, '\p{^Is_Sentence_Break=		lower}', "");
Expect(1, 120780, '\P{Is_Sentence_Break=		lower}', "");
Expect(0, 120780, '\P{^Is_Sentence_Break=		lower}', "");
Error('\p{Is_SB=/a/--lo}');
Error('\P{Is_SB=/a/--lo}');
Expect(1, 120779, '\p{Is_SB=lo}', "");
Expect(0, 120779, '\p{^Is_SB=lo}', "");
Expect(0, 120779, '\P{Is_SB=lo}', "");
Expect(1, 120779, '\P{^Is_SB=lo}', "");
Expect(0, 120780, '\p{Is_SB=lo}', "");
Expect(1, 120780, '\p{^Is_SB=lo}', "");
Expect(1, 120780, '\P{Is_SB=lo}', "");
Expect(0, 120780, '\P{^Is_SB=lo}', "");
Expect(1, 120779, '\p{Is_SB= LO}', "");
Expect(0, 120779, '\p{^Is_SB= LO}', "");
Expect(0, 120779, '\P{Is_SB= LO}', "");
Expect(1, 120779, '\P{^Is_SB= LO}', "");
Expect(0, 120780, '\p{Is_SB= LO}', "");
Expect(1, 120780, '\p{^Is_SB= LO}', "");
Expect(1, 120780, '\P{Is_SB= LO}', "");
Expect(0, 120780, '\P{^Is_SB= LO}', "");
Error('\p{Sentence_Break=:= _Numeric}');
Error('\P{Sentence_Break=:= _Numeric}');
Expect(1, 120831, '\p{Sentence_Break=numeric}', "");
Expect(0, 120831, '\p{^Sentence_Break=numeric}', "");
Expect(0, 120831, '\P{Sentence_Break=numeric}', "");
Expect(1, 120831, '\P{^Sentence_Break=numeric}', "");
Expect(0, 120832, '\p{Sentence_Break=numeric}', "");
Expect(1, 120832, '\p{^Sentence_Break=numeric}', "");
Expect(1, 120832, '\P{Sentence_Break=numeric}', "");
Expect(0, 120832, '\P{^Sentence_Break=numeric}', "");
Expect(1, 120831, '\p{Sentence_Break=_ Numeric}', "");
Expect(0, 120831, '\p{^Sentence_Break=_ Numeric}', "");
Expect(0, 120831, '\P{Sentence_Break=_ Numeric}', "");
Expect(1, 120831, '\P{^Sentence_Break=_ Numeric}', "");
Expect(0, 120832, '\p{Sentence_Break=_ Numeric}', "");
Expect(1, 120832, '\p{^Sentence_Break=_ Numeric}', "");
Expect(1, 120832, '\P{Sentence_Break=_ Numeric}', "");
Expect(0, 120832, '\P{^Sentence_Break=_ Numeric}', "");
Error('\p{SB=/a/-_NU}');
Error('\P{SB=/a/-_NU}');
Expect(1, 120831, '\p{SB=nu}', "");
Expect(0, 120831, '\p{^SB=nu}', "");
Expect(0, 120831, '\P{SB=nu}', "");
Expect(1, 120831, '\P{^SB=nu}', "");
Expect(0, 120832, '\p{SB=nu}', "");
Expect(1, 120832, '\p{^SB=nu}', "");
Expect(1, 120832, '\P{SB=nu}', "");
Expect(0, 120832, '\P{^SB=nu}', "");
Expect(1, 120831, '\p{SB=__NU}', "");
Expect(0, 120831, '\p{^SB=__NU}', "");
Expect(0, 120831, '\P{SB=__NU}', "");
Expect(1, 120831, '\P{^SB=__NU}', "");
Expect(0, 120832, '\p{SB=__NU}', "");
Expect(1, 120832, '\p{^SB=__NU}', "");
Expect(1, 120832, '\P{SB=__NU}', "");
Expect(0, 120832, '\P{^SB=__NU}', "");
Error('\p{Is_Sentence_Break=_/a/numeric}');
Error('\P{Is_Sentence_Break=_/a/numeric}');
Expect(1, 120831, '\p{Is_Sentence_Break=numeric}', "");
Expect(0, 120831, '\p{^Is_Sentence_Break=numeric}', "");
Expect(0, 120831, '\P{Is_Sentence_Break=numeric}', "");
Expect(1, 120831, '\P{^Is_Sentence_Break=numeric}', "");
Expect(0, 120832, '\p{Is_Sentence_Break=numeric}', "");
Expect(1, 120832, '\p{^Is_Sentence_Break=numeric}', "");
Expect(1, 120832, '\P{Is_Sentence_Break=numeric}', "");
Expect(0, 120832, '\P{^Is_Sentence_Break=numeric}', "");
Expect(1, 120831, '\p{Is_Sentence_Break= NUMERIC}', "");
Expect(0, 120831, '\p{^Is_Sentence_Break= NUMERIC}', "");
Expect(0, 120831, '\P{Is_Sentence_Break= NUMERIC}', "");
Expect(1, 120831, '\P{^Is_Sentence_Break= NUMERIC}', "");
Expect(0, 120832, '\p{Is_Sentence_Break= NUMERIC}', "");
Expect(1, 120832, '\p{^Is_Sentence_Break= NUMERIC}', "");
Expect(1, 120832, '\P{Is_Sentence_Break= NUMERIC}', "");
Expect(0, 120832, '\P{^Is_Sentence_Break= NUMERIC}', "");
Error('\p{Is_SB=-:=NU}');
Error('\P{Is_SB=-:=NU}');
Expect(1, 120831, '\p{Is_SB=nu}', "");
Expect(0, 120831, '\p{^Is_SB=nu}', "");
Expect(0, 120831, '\P{Is_SB=nu}', "");
Expect(1, 120831, '\P{^Is_SB=nu}', "");
Expect(0, 120832, '\p{Is_SB=nu}', "");
Expect(1, 120832, '\p{^Is_SB=nu}', "");
Expect(1, 120832, '\P{Is_SB=nu}', "");
Expect(0, 120832, '\P{^Is_SB=nu}', "");
Expect(1, 120831, '\p{Is_SB=_	nu}', "");
Expect(0, 120831, '\p{^Is_SB=_	nu}', "");
Expect(0, 120831, '\P{Is_SB=_	nu}', "");
Expect(1, 120831, '\P{^Is_SB=_	nu}', "");
Expect(0, 120832, '\p{Is_SB=_	nu}', "");
Expect(1, 120832, '\p{^Is_SB=_	nu}', "");
Expect(1, 120832, '\P{Is_SB=_	nu}', "");
Expect(0, 120832, '\P{^Is_SB=_	nu}', "");
Error('\p{Sentence_Break=	/a/SContinue}');
Error('\P{Sentence_Break=	/a/SContinue}');
Expect(1, 65380, '\p{Sentence_Break=scontinue}', "");
Expect(0, 65380, '\p{^Sentence_Break=scontinue}', "");
Expect(0, 65380, '\P{Sentence_Break=scontinue}', "");
Expect(1, 65380, '\P{^Sentence_Break=scontinue}', "");
Expect(0, 65381, '\p{Sentence_Break=scontinue}', "");
Expect(1, 65381, '\p{^Sentence_Break=scontinue}', "");
Expect(1, 65381, '\P{Sentence_Break=scontinue}', "");
Expect(0, 65381, '\P{^Sentence_Break=scontinue}', "");
Expect(1, 65380, '\p{Sentence_Break:			SContinue}', "");
Expect(0, 65380, '\p{^Sentence_Break:			SContinue}', "");
Expect(0, 65380, '\P{Sentence_Break:			SContinue}', "");
Expect(1, 65380, '\P{^Sentence_Break:			SContinue}', "");
Expect(0, 65381, '\p{Sentence_Break:			SContinue}', "");
Expect(1, 65381, '\p{^Sentence_Break:			SContinue}', "");
Expect(1, 65381, '\P{Sentence_Break:			SContinue}', "");
Expect(0, 65381, '\P{^Sentence_Break:			SContinue}', "");
Error('\p{SB=	sc/a/}');
Error('\P{SB=	sc/a/}');
Expect(1, 65380, '\p{SB:	sc}', "");
Expect(0, 65380, '\p{^SB:	sc}', "");
Expect(0, 65380, '\P{SB:	sc}', "");
Expect(1, 65380, '\P{^SB:	sc}', "");
Expect(0, 65381, '\p{SB:	sc}', "");
Expect(1, 65381, '\p{^SB:	sc}', "");
Expect(1, 65381, '\P{SB:	sc}', "");
Expect(0, 65381, '\P{^SB:	sc}', "");
Expect(1, 65380, '\p{SB=_ SC}', "");
Expect(0, 65380, '\p{^SB=_ SC}', "");
Expect(0, 65380, '\P{SB=_ SC}', "");
Expect(1, 65380, '\P{^SB=_ SC}', "");
Expect(0, 65381, '\p{SB=_ SC}', "");
Expect(1, 65381, '\p{^SB=_ SC}', "");
Expect(1, 65381, '\P{SB=_ SC}', "");
Expect(0, 65381, '\P{^SB=_ SC}', "");
Error('\p{Is_Sentence_Break=:=		scontinue}');
Error('\P{Is_Sentence_Break=:=		scontinue}');
Expect(1, 65380, '\p{Is_Sentence_Break=scontinue}', "");
Expect(0, 65380, '\p{^Is_Sentence_Break=scontinue}', "");
Expect(0, 65380, '\P{Is_Sentence_Break=scontinue}', "");
Expect(1, 65380, '\P{^Is_Sentence_Break=scontinue}', "");
Expect(0, 65381, '\p{Is_Sentence_Break=scontinue}', "");
Expect(1, 65381, '\p{^Is_Sentence_Break=scontinue}', "");
Expect(1, 65381, '\P{Is_Sentence_Break=scontinue}', "");
Expect(0, 65381, '\P{^Is_Sentence_Break=scontinue}', "");
Expect(1, 65380, '\p{Is_Sentence_Break: _-SContinue}', "");
Expect(0, 65380, '\p{^Is_Sentence_Break: _-SContinue}', "");
Expect(0, 65380, '\P{Is_Sentence_Break: _-SContinue}', "");
Expect(1, 65380, '\P{^Is_Sentence_Break: _-SContinue}', "");
Expect(0, 65381, '\p{Is_Sentence_Break: _-SContinue}', "");
Expect(1, 65381, '\p{^Is_Sentence_Break: _-SContinue}', "");
Expect(1, 65381, '\P{Is_Sentence_Break: _-SContinue}', "");
Expect(0, 65381, '\P{^Is_Sentence_Break: _-SContinue}', "");
Error('\p{Is_SB= /a/SC}');
Error('\P{Is_SB= /a/SC}');
Expect(1, 65380, '\p{Is_SB=sc}', "");
Expect(0, 65380, '\p{^Is_SB=sc}', "");
Expect(0, 65380, '\P{Is_SB=sc}', "");
Expect(1, 65380, '\P{^Is_SB=sc}', "");
Expect(0, 65381, '\p{Is_SB=sc}', "");
Expect(1, 65381, '\p{^Is_SB=sc}', "");
Expect(1, 65381, '\P{Is_SB=sc}', "");
Expect(0, 65381, '\P{^Is_SB=sc}', "");
Expect(1, 65380, '\p{Is_SB=		SC}', "");
Expect(0, 65380, '\p{^Is_SB=		SC}', "");
Expect(0, 65380, '\P{Is_SB=		SC}', "");
Expect(1, 65380, '\P{^Is_SB=		SC}', "");
Expect(0, 65381, '\p{Is_SB=		SC}', "");
Expect(1, 65381, '\p{^Is_SB=		SC}', "");
Expect(1, 65381, '\P{Is_SB=		SC}', "");
Expect(0, 65381, '\P{^Is_SB=		SC}', "");
Error('\p{Sentence_Break=_:=SEP}');
Error('\P{Sentence_Break=_:=SEP}');
Expect(1, 8233, '\p{Sentence_Break=sep}', "");
Expect(0, 8233, '\p{^Sentence_Break=sep}', "");
Expect(0, 8233, '\P{Sentence_Break=sep}', "");
Expect(1, 8233, '\P{^Sentence_Break=sep}', "");
Expect(0, 8234, '\p{Sentence_Break=sep}', "");
Expect(1, 8234, '\p{^Sentence_Break=sep}', "");
Expect(1, 8234, '\P{Sentence_Break=sep}', "");
Expect(0, 8234, '\P{^Sentence_Break=sep}', "");
Expect(1, 8233, '\p{Sentence_Break=-	sep}', "");
Expect(0, 8233, '\p{^Sentence_Break=-	sep}', "");
Expect(0, 8233, '\P{Sentence_Break=-	sep}', "");
Expect(1, 8233, '\P{^Sentence_Break=-	sep}', "");
Expect(0, 8234, '\p{Sentence_Break=-	sep}', "");
Expect(1, 8234, '\p{^Sentence_Break=-	sep}', "");
Expect(1, 8234, '\P{Sentence_Break=-	sep}', "");
Expect(0, 8234, '\P{^Sentence_Break=-	sep}', "");
Error('\p{SB= _se/a/}');
Error('\P{SB= _se/a/}');
Expect(1, 8233, '\p{SB=se}', "");
Expect(0, 8233, '\p{^SB=se}', "");
Expect(0, 8233, '\P{SB=se}', "");
Expect(1, 8233, '\P{^SB=se}', "");
Expect(0, 8234, '\p{SB=se}', "");
Expect(1, 8234, '\p{^SB=se}', "");
Expect(1, 8234, '\P{SB=se}', "");
Expect(0, 8234, '\P{^SB=se}', "");
Expect(1, 8233, '\p{SB= 	SE}', "");
Expect(0, 8233, '\p{^SB= 	SE}', "");
Expect(0, 8233, '\P{SB= 	SE}', "");
Expect(1, 8233, '\P{^SB= 	SE}', "");
Expect(0, 8234, '\p{SB= 	SE}', "");
Expect(1, 8234, '\p{^SB= 	SE}', "");
Expect(1, 8234, '\P{SB= 	SE}', "");
Expect(0, 8234, '\P{^SB= 	SE}', "");
Error('\p{Is_Sentence_Break=:=	SEP}');
Error('\P{Is_Sentence_Break=:=	SEP}');
Expect(1, 8233, '\p{Is_Sentence_Break=sep}', "");
Expect(0, 8233, '\p{^Is_Sentence_Break=sep}', "");
Expect(0, 8233, '\P{Is_Sentence_Break=sep}', "");
Expect(1, 8233, '\P{^Is_Sentence_Break=sep}', "");
Expect(0, 8234, '\p{Is_Sentence_Break=sep}', "");
Expect(1, 8234, '\p{^Is_Sentence_Break=sep}', "");
Expect(1, 8234, '\P{Is_Sentence_Break=sep}', "");
Expect(0, 8234, '\P{^Is_Sentence_Break=sep}', "");
Expect(1, 8233, '\p{Is_Sentence_Break=--SEP}', "");
Expect(0, 8233, '\p{^Is_Sentence_Break=--SEP}', "");
Expect(0, 8233, '\P{Is_Sentence_Break=--SEP}', "");
Expect(1, 8233, '\P{^Is_Sentence_Break=--SEP}', "");
Expect(0, 8234, '\p{Is_Sentence_Break=--SEP}', "");
Expect(1, 8234, '\p{^Is_Sentence_Break=--SEP}', "");
Expect(1, 8234, '\P{Is_Sentence_Break=--SEP}', "");
Expect(0, 8234, '\P{^Is_Sentence_Break=--SEP}', "");
Error('\p{Is_SB::=	 SE}');
Error('\P{Is_SB::=	 SE}');
Expect(1, 8233, '\p{Is_SB:se}', "");
Expect(0, 8233, '\p{^Is_SB:se}', "");
Expect(0, 8233, '\P{Is_SB:se}', "");
Expect(1, 8233, '\P{^Is_SB:se}', "");
Expect(0, 8234, '\p{Is_SB:se}', "");
Expect(1, 8234, '\p{^Is_SB:se}', "");
Expect(1, 8234, '\P{Is_SB:se}', "");
Expect(0, 8234, '\P{^Is_SB:se}', "");
Expect(1, 8233, '\p{Is_SB=- SE}', "");
Expect(0, 8233, '\p{^Is_SB=- SE}', "");
Expect(0, 8233, '\P{Is_SB=- SE}', "");
Expect(1, 8233, '\P{^Is_SB=- SE}', "");
Expect(0, 8234, '\p{Is_SB=- SE}', "");
Expect(1, 8234, '\p{^Is_SB=- SE}', "");
Expect(1, 8234, '\P{Is_SB=- SE}', "");
Expect(0, 8234, '\P{^Is_SB=- SE}', "");
Error('\p{Sentence_Break=SP:=}');
Error('\P{Sentence_Break=SP:=}');
Expect(1, 12288, '\p{Sentence_Break:	sp}', "");
Expect(0, 12288, '\p{^Sentence_Break:	sp}', "");
Expect(0, 12288, '\P{Sentence_Break:	sp}', "");
Expect(1, 12288, '\P{^Sentence_Break:	sp}', "");
Expect(0, 12289, '\p{Sentence_Break:	sp}', "");
Expect(1, 12289, '\p{^Sentence_Break:	sp}', "");
Expect(1, 12289, '\P{Sentence_Break:	sp}', "");
Expect(0, 12289, '\P{^Sentence_Break:	sp}', "");
Expect(1, 12288, '\p{Sentence_Break= SP}', "");
Expect(0, 12288, '\p{^Sentence_Break= SP}', "");
Expect(0, 12288, '\P{Sentence_Break= SP}', "");
Expect(1, 12288, '\P{^Sentence_Break= SP}', "");
Expect(0, 12289, '\p{Sentence_Break= SP}', "");
Expect(1, 12289, '\p{^Sentence_Break= SP}', "");
Expect(1, 12289, '\P{Sentence_Break= SP}', "");
Expect(0, 12289, '\P{^Sentence_Break= SP}', "");
Error('\p{SB=/a/ SP}');
Error('\P{SB=/a/ SP}');
Expect(1, 12288, '\p{SB=sp}', "");
Expect(0, 12288, '\p{^SB=sp}', "");
Expect(0, 12288, '\P{SB=sp}', "");
Expect(1, 12288, '\P{^SB=sp}', "");
Expect(0, 12289, '\p{SB=sp}', "");
Expect(1, 12289, '\p{^SB=sp}', "");
Expect(1, 12289, '\P{SB=sp}', "");
Expect(0, 12289, '\P{^SB=sp}', "");
Expect(1, 12288, '\p{SB= _Sp}', "");
Expect(0, 12288, '\p{^SB= _Sp}', "");
Expect(0, 12288, '\P{SB= _Sp}', "");
Expect(1, 12288, '\P{^SB= _Sp}', "");
Expect(0, 12289, '\p{SB= _Sp}', "");
Expect(1, 12289, '\p{^SB= _Sp}', "");
Expect(1, 12289, '\P{SB= _Sp}', "");
Expect(0, 12289, '\P{^SB= _Sp}', "");
Error('\p{Is_Sentence_Break=:=_ sp}');
Error('\P{Is_Sentence_Break=:=_ sp}');
Expect(1, 12288, '\p{Is_Sentence_Break=sp}', "");
Expect(0, 12288, '\p{^Is_Sentence_Break=sp}', "");
Expect(0, 12288, '\P{Is_Sentence_Break=sp}', "");
Expect(1, 12288, '\P{^Is_Sentence_Break=sp}', "");
Expect(0, 12289, '\p{Is_Sentence_Break=sp}', "");
Expect(1, 12289, '\p{^Is_Sentence_Break=sp}', "");
Expect(1, 12289, '\P{Is_Sentence_Break=sp}', "");
Expect(0, 12289, '\P{^Is_Sentence_Break=sp}', "");
Expect(1, 12288, '\p{Is_Sentence_Break= Sp}', "");
Expect(0, 12288, '\p{^Is_Sentence_Break= Sp}', "");
Expect(0, 12288, '\P{Is_Sentence_Break= Sp}', "");
Expect(1, 12288, '\P{^Is_Sentence_Break= Sp}', "");
Expect(0, 12289, '\p{Is_Sentence_Break= Sp}', "");
Expect(1, 12289, '\p{^Is_Sentence_Break= Sp}', "");
Expect(1, 12289, '\P{Is_Sentence_Break= Sp}', "");
Expect(0, 12289, '\P{^Is_Sentence_Break= Sp}', "");
Error('\p{Is_SB=/a/	Sp}');
Error('\P{Is_SB=/a/	Sp}');
Expect(1, 12288, '\p{Is_SB=sp}', "");
Expect(0, 12288, '\p{^Is_SB=sp}', "");
Expect(0, 12288, '\P{Is_SB=sp}', "");
Expect(1, 12288, '\P{^Is_SB=sp}', "");
Expect(0, 12289, '\p{Is_SB=sp}', "");
Expect(1, 12289, '\p{^Is_SB=sp}', "");
Expect(1, 12289, '\P{Is_SB=sp}', "");
Expect(0, 12289, '\P{^Is_SB=sp}', "");
Expect(1, 12288, '\p{Is_SB=_	SP}', "");
Expect(0, 12288, '\p{^Is_SB=_	SP}', "");
Expect(0, 12288, '\P{Is_SB=_	SP}', "");
Expect(1, 12288, '\P{^Is_SB=_	SP}', "");
Expect(0, 12289, '\p{Is_SB=_	SP}', "");
Expect(1, 12289, '\p{^Is_SB=_	SP}', "");
Expect(1, 12289, '\P{Is_SB=_	SP}', "");
Expect(0, 12289, '\P{^Is_SB=_	SP}', "");
Error('\p{Sentence_Break=		sterm/a/}');
Error('\P{Sentence_Break=		sterm/a/}');
Expect(1, 70086, '\p{Sentence_Break=sterm}', "");
Expect(0, 70086, '\p{^Sentence_Break=sterm}', "");
Expect(0, 70086, '\P{Sentence_Break=sterm}', "");
Expect(1, 70086, '\P{^Sentence_Break=sterm}', "");
Expect(0, 70087, '\p{Sentence_Break=sterm}', "");
Expect(1, 70087, '\p{^Sentence_Break=sterm}', "");
Expect(1, 70087, '\P{Sentence_Break=sterm}', "");
Expect(0, 70087, '\P{^Sentence_Break=sterm}', "");
Expect(1, 70086, '\p{Sentence_Break=-STerm}', "");
Expect(0, 70086, '\p{^Sentence_Break=-STerm}', "");
Expect(0, 70086, '\P{Sentence_Break=-STerm}', "");
Expect(1, 70086, '\P{^Sentence_Break=-STerm}', "");
Expect(0, 70087, '\p{Sentence_Break=-STerm}', "");
Expect(1, 70087, '\p{^Sentence_Break=-STerm}', "");
Expect(1, 70087, '\P{Sentence_Break=-STerm}', "");
Expect(0, 70087, '\P{^Sentence_Break=-STerm}', "");
Error('\p{SB=	:=ST}');
Error('\P{SB=	:=ST}');
Expect(1, 70086, '\p{SB=st}', "");
Expect(0, 70086, '\p{^SB=st}', "");
Expect(0, 70086, '\P{SB=st}', "");
Expect(1, 70086, '\P{^SB=st}', "");
Expect(0, 70087, '\p{SB=st}', "");
Expect(1, 70087, '\p{^SB=st}', "");
Expect(1, 70087, '\P{SB=st}', "");
Expect(0, 70087, '\P{^SB=st}', "");
Expect(1, 70086, '\p{SB= ST}', "");
Expect(0, 70086, '\p{^SB= ST}', "");
Expect(0, 70086, '\P{SB= ST}', "");
Expect(1, 70086, '\P{^SB= ST}', "");
Expect(0, 70087, '\p{SB= ST}', "");
Expect(1, 70087, '\p{^SB= ST}', "");
Expect(1, 70087, '\P{SB= ST}', "");
Expect(0, 70087, '\P{^SB= ST}', "");
Error('\p{Is_Sentence_Break=	-STerm:=}');
Error('\P{Is_Sentence_Break=	-STerm:=}');
Expect(1, 70086, '\p{Is_Sentence_Break=sterm}', "");
Expect(0, 70086, '\p{^Is_Sentence_Break=sterm}', "");
Expect(0, 70086, '\P{Is_Sentence_Break=sterm}', "");
Expect(1, 70086, '\P{^Is_Sentence_Break=sterm}', "");
Expect(0, 70087, '\p{Is_Sentence_Break=sterm}', "");
Expect(1, 70087, '\p{^Is_Sentence_Break=sterm}', "");
Expect(1, 70087, '\P{Is_Sentence_Break=sterm}', "");
Expect(0, 70087, '\P{^Is_Sentence_Break=sterm}', "");
Expect(1, 70086, '\p{Is_Sentence_Break=  STerm}', "");
Expect(0, 70086, '\p{^Is_Sentence_Break=  STerm}', "");
Expect(0, 70086, '\P{Is_Sentence_Break=  STerm}', "");
Expect(1, 70086, '\P{^Is_Sentence_Break=  STerm}', "");
Expect(0, 70087, '\p{Is_Sentence_Break=  STerm}', "");
Expect(1, 70087, '\p{^Is_Sentence_Break=  STerm}', "");
Expect(1, 70087, '\P{Is_Sentence_Break=  STerm}', "");
Expect(0, 70087, '\P{^Is_Sentence_Break=  STerm}', "");
Error('\p{Is_SB=/a/ st}');
Error('\P{Is_SB=/a/ st}');
Expect(1, 70086, '\p{Is_SB=st}', "");
Expect(0, 70086, '\p{^Is_SB=st}', "");
Expect(0, 70086, '\P{Is_SB=st}', "");
Expect(1, 70086, '\P{^Is_SB=st}', "");
Expect(0, 70087, '\p{Is_SB=st}', "");
Expect(1, 70087, '\p{^Is_SB=st}', "");
Expect(1, 70087, '\P{Is_SB=st}', "");
Expect(0, 70087, '\P{^Is_SB=st}', "");
Expect(1, 70086, '\p{Is_SB=_	ST}', "");
Expect(0, 70086, '\p{^Is_SB=_	ST}', "");
Expect(0, 70086, '\P{Is_SB=_	ST}', "");
Expect(1, 70086, '\P{^Is_SB=_	ST}', "");
Expect(0, 70087, '\p{Is_SB=_	ST}', "");
Expect(1, 70087, '\p{^Is_SB=_	ST}', "");
Expect(1, 70087, '\P{Is_SB=_	ST}', "");
Expect(0, 70087, '\P{^Is_SB=_	ST}', "");
Error('\p{Sentence_Break:		Upper:=}');
Error('\P{Sentence_Break:		Upper:=}');
Expect(1, 120778, '\p{Sentence_Break=upper}', "");
Expect(0, 120778, '\p{^Sentence_Break=upper}', "");
Expect(0, 120778, '\P{Sentence_Break=upper}', "");
Expect(1, 120778, '\P{^Sentence_Break=upper}', "");
Expect(0, 120779, '\p{Sentence_Break=upper}', "");
Expect(1, 120779, '\p{^Sentence_Break=upper}', "");
Expect(1, 120779, '\P{Sentence_Break=upper}', "");
Expect(0, 120779, '\P{^Sentence_Break=upper}', "");
Expect(1, 120778, '\p{Sentence_Break=	-Upper}', "");
Expect(0, 120778, '\p{^Sentence_Break=	-Upper}', "");
Expect(0, 120778, '\P{Sentence_Break=	-Upper}', "");
Expect(1, 120778, '\P{^Sentence_Break=	-Upper}', "");
Expect(0, 120779, '\p{Sentence_Break=	-Upper}', "");
Expect(1, 120779, '\p{^Sentence_Break=	-Upper}', "");
Expect(1, 120779, '\P{Sentence_Break=	-Upper}', "");
Expect(0, 120779, '\P{^Sentence_Break=	-Upper}', "");
Error('\p{SB=:=	UP}');
Error('\P{SB=:=	UP}');
Expect(1, 120778, '\p{SB=up}', "");
Expect(0, 120778, '\p{^SB=up}', "");
Expect(0, 120778, '\P{SB=up}', "");
Expect(1, 120778, '\P{^SB=up}', "");
Expect(0, 120779, '\p{SB=up}', "");
Expect(1, 120779, '\p{^SB=up}', "");
Expect(1, 120779, '\P{SB=up}', "");
Expect(0, 120779, '\P{^SB=up}', "");
Expect(1, 120778, '\p{SB=-up}', "");
Expect(0, 120778, '\p{^SB=-up}', "");
Expect(0, 120778, '\P{SB=-up}', "");
Expect(1, 120778, '\P{^SB=-up}', "");
Expect(0, 120779, '\p{SB=-up}', "");
Expect(1, 120779, '\p{^SB=-up}', "");
Expect(1, 120779, '\P{SB=-up}', "");
Expect(0, 120779, '\P{^SB=-up}', "");
Error('\p{Is_Sentence_Break=/a/UPPER}');
Error('\P{Is_Sentence_Break=/a/UPPER}');
Expect(1, 120778, '\p{Is_Sentence_Break=upper}', "");
Expect(0, 120778, '\p{^Is_Sentence_Break=upper}', "");
Expect(0, 120778, '\P{Is_Sentence_Break=upper}', "");
Expect(1, 120778, '\P{^Is_Sentence_Break=upper}', "");
Expect(0, 120779, '\p{Is_Sentence_Break=upper}', "");
Expect(1, 120779, '\p{^Is_Sentence_Break=upper}', "");
Expect(1, 120779, '\P{Is_Sentence_Break=upper}', "");
Expect(0, 120779, '\P{^Is_Sentence_Break=upper}', "");
Expect(1, 120778, '\p{Is_Sentence_Break=_-Upper}', "");
Expect(0, 120778, '\p{^Is_Sentence_Break=_-Upper}', "");
Expect(0, 120778, '\P{Is_Sentence_Break=_-Upper}', "");
Expect(1, 120778, '\P{^Is_Sentence_Break=_-Upper}', "");
Expect(0, 120779, '\p{Is_Sentence_Break=_-Upper}', "");
Expect(1, 120779, '\p{^Is_Sentence_Break=_-Upper}', "");
Expect(1, 120779, '\P{Is_Sentence_Break=_-Upper}', "");
Expect(0, 120779, '\P{^Is_Sentence_Break=_-Upper}', "");
Error('\p{Is_SB= UP/a/}');
Error('\P{Is_SB= UP/a/}');
Expect(1, 120778, '\p{Is_SB=up}', "");
Expect(0, 120778, '\p{^Is_SB=up}', "");
Expect(0, 120778, '\P{Is_SB=up}', "");
Expect(1, 120778, '\P{^Is_SB=up}', "");
Expect(0, 120779, '\p{Is_SB=up}', "");
Expect(1, 120779, '\p{^Is_SB=up}', "");
Expect(1, 120779, '\P{Is_SB=up}', "");
Expect(0, 120779, '\P{^Is_SB=up}', "");
Expect(1, 120778, '\p{Is_SB= 	up}', "");
Expect(0, 120778, '\p{^Is_SB= 	up}', "");
Expect(0, 120778, '\P{Is_SB= 	up}', "");
Expect(1, 120778, '\P{^Is_SB= 	up}', "");
Expect(0, 120779, '\p{Is_SB= 	up}', "");
Expect(1, 120779, '\p{^Is_SB= 	up}', "");
Expect(1, 120779, '\P{Is_SB= 	up}', "");
Expect(0, 120779, '\P{^Is_SB= 	up}', "");
Error('\p{Sentence_Break=/a/_Other}');
Error('\P{Sentence_Break=/a/_Other}');
Expect(1, 918000, '\p{Sentence_Break=other}', "");
Expect(0, 918000, '\p{^Sentence_Break=other}', "");
Expect(0, 918000, '\P{Sentence_Break=other}', "");
Expect(1, 918000, '\P{^Sentence_Break=other}', "");
Expect(0, 917999, '\p{Sentence_Break=other}', "");
Expect(1, 917999, '\p{^Sentence_Break=other}', "");
Expect(1, 917999, '\P{Sentence_Break=other}', "");
Expect(0, 917999, '\P{^Sentence_Break=other}', "");
Expect(1, 918000, '\p{Sentence_Break:  	Other}', "");
Expect(0, 918000, '\p{^Sentence_Break:  	Other}', "");
Expect(0, 918000, '\P{Sentence_Break:  	Other}', "");
Expect(1, 918000, '\P{^Sentence_Break:  	Other}', "");
Expect(0, 917999, '\p{Sentence_Break:  	Other}', "");
Expect(1, 917999, '\p{^Sentence_Break:  	Other}', "");
Expect(1, 917999, '\P{Sentence_Break:  	Other}', "");
Expect(0, 917999, '\P{^Sentence_Break:  	Other}', "");
Error('\p{SB:   /a/_ XX}');
Error('\P{SB:   /a/_ XX}');
Expect(1, 918000, '\p{SB=xx}', "");
Expect(0, 918000, '\p{^SB=xx}', "");
Expect(0, 918000, '\P{SB=xx}', "");
Expect(1, 918000, '\P{^SB=xx}', "");
Expect(0, 917999, '\p{SB=xx}', "");
Expect(1, 917999, '\p{^SB=xx}', "");
Expect(1, 917999, '\P{SB=xx}', "");
Expect(0, 917999, '\P{^SB=xx}', "");
Expect(1, 918000, '\p{SB=__XX}', "");
Expect(0, 918000, '\p{^SB=__XX}', "");
Expect(0, 918000, '\P{SB=__XX}', "");
Expect(1, 918000, '\P{^SB=__XX}', "");
Expect(0, 917999, '\p{SB=__XX}', "");
Expect(1, 917999, '\p{^SB=__XX}', "");
Expect(1, 917999, '\P{SB=__XX}', "");
Expect(0, 917999, '\P{^SB=__XX}', "");
Error('\p{Is_Sentence_Break= :=Other}');
Error('\P{Is_Sentence_Break= :=Other}');
Expect(1, 918000, '\p{Is_Sentence_Break=other}', "");
Expect(0, 918000, '\p{^Is_Sentence_Break=other}', "");
Expect(0, 918000, '\P{Is_Sentence_Break=other}', "");
Expect(1, 918000, '\P{^Is_Sentence_Break=other}', "");
Expect(0, 917999, '\p{Is_Sentence_Break=other}', "");
Expect(1, 917999, '\p{^Is_Sentence_Break=other}', "");
Expect(1, 917999, '\P{Is_Sentence_Break=other}', "");
Expect(0, 917999, '\P{^Is_Sentence_Break=other}', "");
Expect(1, 918000, '\p{Is_Sentence_Break=--other}', "");
Expect(0, 918000, '\p{^Is_Sentence_Break=--other}', "");
Expect(0, 918000, '\P{Is_Sentence_Break=--other}', "");
Expect(1, 918000, '\P{^Is_Sentence_Break=--other}', "");
Expect(0, 917999, '\p{Is_Sentence_Break=--other}', "");
Expect(1, 917999, '\p{^Is_Sentence_Break=--other}', "");
Expect(1, 917999, '\P{Is_Sentence_Break=--other}', "");
Expect(0, 917999, '\P{^Is_Sentence_Break=--other}', "");
Error('\p{Is_SB=		XX:=}');
Error('\P{Is_SB=		XX:=}');
Expect(1, 918000, '\p{Is_SB=xx}', "");
Expect(0, 918000, '\p{^Is_SB=xx}', "");
Expect(0, 918000, '\P{Is_SB=xx}', "");
Expect(1, 918000, '\P{^Is_SB=xx}', "");
Expect(0, 917999, '\p{Is_SB=xx}', "");
Expect(1, 917999, '\p{^Is_SB=xx}', "");
Expect(1, 917999, '\P{Is_SB=xx}', "");
Expect(0, 917999, '\P{^Is_SB=xx}', "");
Expect(1, 918000, '\p{Is_SB=-XX}', "");
Expect(0, 918000, '\p{^Is_SB=-XX}', "");
Expect(0, 918000, '\P{Is_SB=-XX}', "");
Expect(1, 918000, '\P{^Is_SB=-XX}', "");
Expect(0, 917999, '\p{Is_SB=-XX}', "");
Expect(1, 917999, '\p{^Is_SB=-XX}', "");
Expect(1, 917999, '\P{Is_SB=-XX}', "");
Expect(0, 917999, '\P{^Is_SB=-XX}', "");
Error('\p{Soft_Dotted=/a/-_NO}');
Error('\P{Soft_Dotted=/a/-_NO}');
Expect(1, 120468, '\p{Soft_Dotted=no}', "");
Expect(0, 120468, '\p{^Soft_Dotted=no}', "");
Expect(0, 120468, '\P{Soft_Dotted=no}', "");
Expect(1, 120468, '\P{^Soft_Dotted=no}', "");
Expect(0, 120467, '\p{Soft_Dotted=no}', "");
Expect(1, 120467, '\p{^Soft_Dotted=no}', "");
Expect(1, 120467, '\P{Soft_Dotted=no}', "");
Expect(0, 120467, '\P{^Soft_Dotted=no}', "");
Expect(1, 120468, '\p{Soft_Dotted=_-no}', "");
Expect(0, 120468, '\p{^Soft_Dotted=_-no}', "");
Expect(0, 120468, '\P{Soft_Dotted=_-no}', "");
Expect(1, 120468, '\P{^Soft_Dotted=_-no}', "");
Expect(0, 120467, '\p{Soft_Dotted=_-no}', "");
Expect(1, 120467, '\p{^Soft_Dotted=_-no}', "");
Expect(1, 120467, '\P{Soft_Dotted=_-no}', "");
Expect(0, 120467, '\P{^Soft_Dotted=_-no}', "");
Error('\p{SD=	:=N}');
Error('\P{SD=	:=N}');
Expect(1, 120468, '\p{SD=n}', "");
Expect(0, 120468, '\p{^SD=n}', "");
Expect(0, 120468, '\P{SD=n}', "");
Expect(1, 120468, '\P{^SD=n}', "");
Expect(0, 120467, '\p{SD=n}', "");
Expect(1, 120467, '\p{^SD=n}', "");
Expect(1, 120467, '\P{SD=n}', "");
Expect(0, 120467, '\P{^SD=n}', "");
Expect(1, 120468, '\p{SD=	-n}', "");
Expect(0, 120468, '\p{^SD=	-n}', "");
Expect(0, 120468, '\P{SD=	-n}', "");
Expect(1, 120468, '\P{^SD=	-n}', "");
Expect(0, 120467, '\p{SD=	-n}', "");
Expect(1, 120467, '\p{^SD=	-n}', "");
Expect(1, 120467, '\P{SD=	-n}', "");
Expect(0, 120467, '\P{^SD=	-n}', "");
Error('\p{Is_Soft_Dotted: :=F}');
Error('\P{Is_Soft_Dotted: :=F}');
Expect(1, 120468, '\p{Is_Soft_Dotted: f}', "");
Expect(0, 120468, '\p{^Is_Soft_Dotted: f}', "");
Expect(0, 120468, '\P{Is_Soft_Dotted: f}', "");
Expect(1, 120468, '\P{^Is_Soft_Dotted: f}', "");
Expect(0, 120467, '\p{Is_Soft_Dotted: f}', "");
Expect(1, 120467, '\p{^Is_Soft_Dotted: f}', "");
Expect(1, 120467, '\P{Is_Soft_Dotted: f}', "");
Expect(0, 120467, '\P{^Is_Soft_Dotted: f}', "");
Expect(1, 120468, '\p{Is_Soft_Dotted=_-F}', "");
Expect(0, 120468, '\p{^Is_Soft_Dotted=_-F}', "");
Expect(0, 120468, '\P{Is_Soft_Dotted=_-F}', "");
Expect(1, 120468, '\P{^Is_Soft_Dotted=_-F}', "");
Expect(0, 120467, '\p{Is_Soft_Dotted=_-F}', "");
Expect(1, 120467, '\p{^Is_Soft_Dotted=_-F}', "");
Expect(1, 120467, '\P{Is_Soft_Dotted=_-F}', "");
Expect(0, 120467, '\P{^Is_Soft_Dotted=_-F}', "");
Error('\p{Is_SD=:=_FALSE}');
Error('\P{Is_SD=:=_FALSE}');
Expect(1, 120468, '\p{Is_SD=false}', "");
Expect(0, 120468, '\p{^Is_SD=false}', "");
Expect(0, 120468, '\P{Is_SD=false}', "");
Expect(1, 120468, '\P{^Is_SD=false}', "");
Expect(0, 120467, '\p{Is_SD=false}', "");
Expect(1, 120467, '\p{^Is_SD=false}', "");
Expect(1, 120467, '\P{Is_SD=false}', "");
Expect(0, 120467, '\P{^Is_SD=false}', "");
Expect(1, 120468, '\p{Is_SD= False}', "");
Expect(0, 120468, '\p{^Is_SD= False}', "");
Expect(0, 120468, '\P{Is_SD= False}', "");
Expect(1, 120468, '\P{^Is_SD= False}', "");
Expect(0, 120467, '\p{Is_SD= False}', "");
Expect(1, 120467, '\p{^Is_SD= False}', "");
Expect(1, 120467, '\P{Is_SD= False}', "");
Expect(0, 120467, '\P{^Is_SD= False}', "");
Error('\p{Soft_Dotted=/a/- yes}');
Error('\P{Soft_Dotted=/a/- yes}');
Expect(1, 120467, '\p{Soft_Dotted=yes}', "");
Expect(0, 120467, '\p{^Soft_Dotted=yes}', "");
Expect(0, 120467, '\P{Soft_Dotted=yes}', "");
Expect(1, 120467, '\P{^Soft_Dotted=yes}', "");
Expect(0, 120468, '\p{Soft_Dotted=yes}', "");
Expect(1, 120468, '\p{^Soft_Dotted=yes}', "");
Expect(1, 120468, '\P{Soft_Dotted=yes}', "");
Expect(0, 120468, '\P{^Soft_Dotted=yes}', "");
Expect(1, 120467, '\p{Soft_Dotted=	Yes}', "");
Expect(0, 120467, '\p{^Soft_Dotted=	Yes}', "");
Expect(0, 120467, '\P{Soft_Dotted=	Yes}', "");
Expect(1, 120467, '\P{^Soft_Dotted=	Yes}', "");
Expect(0, 120468, '\p{Soft_Dotted=	Yes}', "");
Expect(1, 120468, '\p{^Soft_Dotted=	Yes}', "");
Expect(1, 120468, '\P{Soft_Dotted=	Yes}', "");
Expect(0, 120468, '\P{^Soft_Dotted=	Yes}', "");
Error('\p{SD: y/a/}');
Error('\P{SD: y/a/}');
Expect(1, 120467, '\p{SD:y}', "");
Expect(0, 120467, '\p{^SD:y}', "");
Expect(0, 120467, '\P{SD:y}', "");
Expect(1, 120467, '\P{^SD:y}', "");
Expect(0, 120468, '\p{SD:y}', "");
Expect(1, 120468, '\p{^SD:y}', "");
Expect(1, 120468, '\P{SD:y}', "");
Expect(0, 120468, '\P{^SD:y}', "");
Expect(1, 120467, '\p{SD= Y}', "");
Expect(0, 120467, '\p{^SD= Y}', "");
Expect(0, 120467, '\P{SD= Y}', "");
Expect(1, 120467, '\P{^SD= Y}', "");
Expect(0, 120468, '\p{SD= Y}', "");
Expect(1, 120468, '\p{^SD= Y}', "");
Expect(1, 120468, '\P{SD= Y}', "");
Expect(0, 120468, '\P{^SD= Y}', "");
Error('\p{Is_Soft_Dotted=/a/-T}');
Error('\P{Is_Soft_Dotted=/a/-T}');
Expect(1, 120467, '\p{Is_Soft_Dotted=t}', "");
Expect(0, 120467, '\p{^Is_Soft_Dotted=t}', "");
Expect(0, 120467, '\P{Is_Soft_Dotted=t}', "");
Expect(1, 120467, '\P{^Is_Soft_Dotted=t}', "");
Expect(0, 120468, '\p{Is_Soft_Dotted=t}', "");
Expect(1, 120468, '\p{^Is_Soft_Dotted=t}', "");
Expect(1, 120468, '\P{Is_Soft_Dotted=t}', "");
Expect(0, 120468, '\P{^Is_Soft_Dotted=t}', "");
Expect(1, 120467, '\p{Is_Soft_Dotted=  T}', "");
Expect(0, 120467, '\p{^Is_Soft_Dotted=  T}', "");
Expect(0, 120467, '\P{Is_Soft_Dotted=  T}', "");
Expect(1, 120467, '\P{^Is_Soft_Dotted=  T}', "");
Expect(0, 120468, '\p{Is_Soft_Dotted=  T}', "");
Expect(1, 120468, '\p{^Is_Soft_Dotted=  T}', "");
Expect(1, 120468, '\P{Is_Soft_Dotted=  T}', "");
Expect(0, 120468, '\P{^Is_Soft_Dotted=  T}', "");
Error('\p{Is_SD=-	true/a/}');
Error('\P{Is_SD=-	true/a/}');
Expect(1, 120467, '\p{Is_SD=true}', "");
Expect(0, 120467, '\p{^Is_SD=true}', "");
Expect(0, 120467, '\P{Is_SD=true}', "");
Expect(1, 120467, '\P{^Is_SD=true}', "");
Expect(0, 120468, '\p{Is_SD=true}', "");
Expect(1, 120468, '\p{^Is_SD=true}', "");
Expect(1, 120468, '\P{Is_SD=true}', "");
Expect(0, 120468, '\P{^Is_SD=true}', "");
Expect(1, 120467, '\p{Is_SD=_true}', "");
Expect(0, 120467, '\p{^Is_SD=_true}', "");
Expect(0, 120467, '\P{Is_SD=_true}', "");
Expect(1, 120467, '\P{^Is_SD=_true}', "");
Expect(0, 120468, '\p{Is_SD=_true}', "");
Expect(1, 120468, '\p{^Is_SD=_true}', "");
Expect(1, 120468, '\P{Is_SD=_true}', "");
Expect(0, 120468, '\P{^Is_SD=_true}', "");
Error('\p{STerm=	-No/a/}');
Error('\P{STerm=	-No/a/}');
Expect(1, 70087, '\p{STerm=no}', "");
Expect(0, 70087, '\p{^STerm=no}', "");
Expect(0, 70087, '\P{STerm=no}', "");
Expect(1, 70087, '\P{^STerm=no}', "");
Expect(0, 70086, '\p{STerm=no}', "");
Expect(1, 70086, '\p{^STerm=no}', "");
Expect(1, 70086, '\P{STerm=no}', "");
Expect(0, 70086, '\P{^STerm=no}', "");
Expect(1, 70087, '\p{STerm=		NO}', "");
Expect(0, 70087, '\p{^STerm=		NO}', "");
Expect(0, 70087, '\P{STerm=		NO}', "");
Expect(1, 70087, '\P{^STerm=		NO}', "");
Expect(0, 70086, '\p{STerm=		NO}', "");
Expect(1, 70086, '\p{^STerm=		NO}', "");
Expect(1, 70086, '\P{STerm=		NO}', "");
Expect(0, 70086, '\P{^STerm=		NO}', "");
Error('\p{Is_STerm=:= N}');
Error('\P{Is_STerm=:= N}');
Expect(1, 70087, '\p{Is_STerm=n}', "");
Expect(0, 70087, '\p{^Is_STerm=n}', "");
Expect(0, 70087, '\P{Is_STerm=n}', "");
Expect(1, 70087, '\P{^Is_STerm=n}', "");
Expect(0, 70086, '\p{Is_STerm=n}', "");
Expect(1, 70086, '\p{^Is_STerm=n}', "");
Expect(1, 70086, '\P{Is_STerm=n}', "");
Expect(0, 70086, '\P{^Is_STerm=n}', "");
Expect(1, 70087, '\p{Is_STerm=		N}', "");
Expect(0, 70087, '\p{^Is_STerm=		N}', "");
Expect(0, 70087, '\P{Is_STerm=		N}', "");
Expect(1, 70087, '\P{^Is_STerm=		N}', "");
Expect(0, 70086, '\p{Is_STerm=		N}', "");
Expect(1, 70086, '\p{^Is_STerm=		N}', "");
Expect(1, 70086, '\P{Is_STerm=		N}', "");
Expect(0, 70086, '\P{^Is_STerm=		N}', "");
Error('\p{STerm: :=- f}');
Error('\P{STerm: :=- f}');
Expect(1, 70087, '\p{STerm=f}', "");
Expect(0, 70087, '\p{^STerm=f}', "");
Expect(0, 70087, '\P{STerm=f}', "");
Expect(1, 70087, '\P{^STerm=f}', "");
Expect(0, 70086, '\p{STerm=f}', "");
Expect(1, 70086, '\p{^STerm=f}', "");
Expect(1, 70086, '\P{STerm=f}', "");
Expect(0, 70086, '\P{^STerm=f}', "");
Expect(1, 70087, '\p{STerm= 	F}', "");
Expect(0, 70087, '\p{^STerm= 	F}', "");
Expect(0, 70087, '\P{STerm= 	F}', "");
Expect(1, 70087, '\P{^STerm= 	F}', "");
Expect(0, 70086, '\p{STerm= 	F}', "");
Expect(1, 70086, '\p{^STerm= 	F}', "");
Expect(1, 70086, '\P{STerm= 	F}', "");
Expect(0, 70086, '\P{^STerm= 	F}', "");
Error('\p{Is_STerm=	:=false}');
Error('\P{Is_STerm=	:=false}');
Expect(1, 70087, '\p{Is_STerm=false}', "");
Expect(0, 70087, '\p{^Is_STerm=false}', "");
Expect(0, 70087, '\P{Is_STerm=false}', "");
Expect(1, 70087, '\P{^Is_STerm=false}', "");
Expect(0, 70086, '\p{Is_STerm=false}', "");
Expect(1, 70086, '\p{^Is_STerm=false}', "");
Expect(1, 70086, '\P{Is_STerm=false}', "");
Expect(0, 70086, '\P{^Is_STerm=false}', "");
Expect(1, 70087, '\p{Is_STerm=--false}', "");
Expect(0, 70087, '\p{^Is_STerm=--false}', "");
Expect(0, 70087, '\P{Is_STerm=--false}', "");
Expect(1, 70087, '\P{^Is_STerm=--false}', "");
Expect(0, 70086, '\p{Is_STerm=--false}', "");
Expect(1, 70086, '\p{^Is_STerm=--false}', "");
Expect(1, 70086, '\P{Is_STerm=--false}', "");
Expect(0, 70086, '\P{^Is_STerm=--false}', "");
Error('\p{STerm=:= Yes}');
Error('\P{STerm=:= Yes}');
Expect(1, 70086, '\p{STerm=yes}', "");
Expect(0, 70086, '\p{^STerm=yes}', "");
Expect(0, 70086, '\P{STerm=yes}', "");
Expect(1, 70086, '\P{^STerm=yes}', "");
Expect(0, 70087, '\p{STerm=yes}', "");
Expect(1, 70087, '\p{^STerm=yes}', "");
Expect(1, 70087, '\P{STerm=yes}', "");
Expect(0, 70087, '\P{^STerm=yes}', "");
Expect(1, 70086, '\p{STerm= Yes}', "");
Expect(0, 70086, '\p{^STerm= Yes}', "");
Expect(0, 70086, '\P{STerm= Yes}', "");
Expect(1, 70086, '\P{^STerm= Yes}', "");
Expect(0, 70087, '\p{STerm= Yes}', "");
Expect(1, 70087, '\p{^STerm= Yes}', "");
Expect(1, 70087, '\P{STerm= Yes}', "");
Expect(0, 70087, '\P{^STerm= Yes}', "");
Error('\p{Is_STerm=-/a/Y}');
Error('\P{Is_STerm=-/a/Y}');
Expect(1, 70086, '\p{Is_STerm=y}', "");
Expect(0, 70086, '\p{^Is_STerm=y}', "");
Expect(0, 70086, '\P{Is_STerm=y}', "");
Expect(1, 70086, '\P{^Is_STerm=y}', "");
Expect(0, 70087, '\p{Is_STerm=y}', "");
Expect(1, 70087, '\p{^Is_STerm=y}', "");
Expect(1, 70087, '\P{Is_STerm=y}', "");
Expect(0, 70087, '\P{^Is_STerm=y}', "");
Expect(1, 70086, '\p{Is_STerm=_y}', "");
Expect(0, 70086, '\p{^Is_STerm=_y}', "");
Expect(0, 70086, '\P{Is_STerm=_y}', "");
Expect(1, 70086, '\P{^Is_STerm=_y}', "");
Expect(0, 70087, '\p{Is_STerm=_y}', "");
Expect(1, 70087, '\p{^Is_STerm=_y}', "");
Expect(1, 70087, '\P{Is_STerm=_y}', "");
Expect(0, 70087, '\P{^Is_STerm=_y}', "");
Error('\p{STerm=:=T}');
Error('\P{STerm=:=T}');
Expect(1, 70086, '\p{STerm=t}', "");
Expect(0, 70086, '\p{^STerm=t}', "");
Expect(0, 70086, '\P{STerm=t}', "");
Expect(1, 70086, '\P{^STerm=t}', "");
Expect(0, 70087, '\p{STerm=t}', "");
Expect(1, 70087, '\p{^STerm=t}', "");
Expect(1, 70087, '\P{STerm=t}', "");
Expect(0, 70087, '\P{^STerm=t}', "");
Expect(1, 70086, '\p{STerm=	-T}', "");
Expect(0, 70086, '\p{^STerm=	-T}', "");
Expect(0, 70086, '\P{STerm=	-T}', "");
Expect(1, 70086, '\P{^STerm=	-T}', "");
Expect(0, 70087, '\p{STerm=	-T}', "");
Expect(1, 70087, '\p{^STerm=	-T}', "");
Expect(1, 70087, '\P{STerm=	-T}', "");
Expect(0, 70087, '\P{^STerm=	-T}', "");
Error('\p{Is_STerm=/a/_-TRUE}');
Error('\P{Is_STerm=/a/_-TRUE}');
Expect(1, 70086, '\p{Is_STerm=true}', "");
Expect(0, 70086, '\p{^Is_STerm=true}', "");
Expect(0, 70086, '\P{Is_STerm=true}', "");
Expect(1, 70086, '\P{^Is_STerm=true}', "");
Expect(0, 70087, '\p{Is_STerm=true}', "");
Expect(1, 70087, '\p{^Is_STerm=true}', "");
Expect(1, 70087, '\P{Is_STerm=true}', "");
Expect(0, 70087, '\P{^Is_STerm=true}', "");
Expect(1, 70086, '\p{Is_STerm=	True}', "");
Expect(0, 70086, '\p{^Is_STerm=	True}', "");
Expect(0, 70086, '\P{Is_STerm=	True}', "");
Expect(1, 70086, '\P{^Is_STerm=	True}', "");
Expect(0, 70087, '\p{Is_STerm=	True}', "");
Expect(1, 70087, '\p{^Is_STerm=	True}', "");
Expect(1, 70087, '\P{Is_STerm=	True}', "");
Expect(0, 70087, '\P{^Is_STerm=	True}', "");
Error('\p{Terminal_Punctuation=/a/no}');
Error('\P{Terminal_Punctuation=/a/no}');
Expect(1, 74868, '\p{Terminal_Punctuation=no}', "");
Expect(0, 74868, '\p{^Terminal_Punctuation=no}', "");
Expect(0, 74868, '\P{Terminal_Punctuation=no}', "");
Expect(1, 74868, '\P{^Terminal_Punctuation=no}', "");
Expect(0, 74867, '\p{Terminal_Punctuation=no}', "");
Expect(1, 74867, '\p{^Terminal_Punctuation=no}', "");
Expect(1, 74867, '\P{Terminal_Punctuation=no}', "");
Expect(0, 74867, '\P{^Terminal_Punctuation=no}', "");
Expect(1, 74868, '\p{Terminal_Punctuation=  NO}', "");
Expect(0, 74868, '\p{^Terminal_Punctuation=  NO}', "");
Expect(0, 74868, '\P{Terminal_Punctuation=  NO}', "");
Expect(1, 74868, '\P{^Terminal_Punctuation=  NO}', "");
Expect(0, 74867, '\p{Terminal_Punctuation=  NO}', "");
Expect(1, 74867, '\p{^Terminal_Punctuation=  NO}', "");
Expect(1, 74867, '\P{Terminal_Punctuation=  NO}', "");
Expect(0, 74867, '\P{^Terminal_Punctuation=  NO}', "");
Error('\p{Term= /a/N}');
Error('\P{Term= /a/N}');
Expect(1, 74868, '\p{Term=n}', "");
Expect(0, 74868, '\p{^Term=n}', "");
Expect(0, 74868, '\P{Term=n}', "");
Expect(1, 74868, '\P{^Term=n}', "");
Expect(0, 74867, '\p{Term=n}', "");
Expect(1, 74867, '\p{^Term=n}', "");
Expect(1, 74867, '\P{Term=n}', "");
Expect(0, 74867, '\P{^Term=n}', "");
Expect(1, 74868, '\p{Term= 	N}', "");
Expect(0, 74868, '\p{^Term= 	N}', "");
Expect(0, 74868, '\P{Term= 	N}', "");
Expect(1, 74868, '\P{^Term= 	N}', "");
Expect(0, 74867, '\p{Term= 	N}', "");
Expect(1, 74867, '\p{^Term= 	N}', "");
Expect(1, 74867, '\P{Term= 	N}', "");
Expect(0, 74867, '\P{^Term= 	N}', "");
Error('\p{Is_Terminal_Punctuation= _F/a/}');
Error('\P{Is_Terminal_Punctuation= _F/a/}');
Expect(1, 74868, '\p{Is_Terminal_Punctuation=f}', "");
Expect(0, 74868, '\p{^Is_Terminal_Punctuation=f}', "");
Expect(0, 74868, '\P{Is_Terminal_Punctuation=f}', "");
Expect(1, 74868, '\P{^Is_Terminal_Punctuation=f}', "");
Expect(0, 74867, '\p{Is_Terminal_Punctuation=f}', "");
Expect(1, 74867, '\p{^Is_Terminal_Punctuation=f}', "");
Expect(1, 74867, '\P{Is_Terminal_Punctuation=f}', "");
Expect(0, 74867, '\P{^Is_Terminal_Punctuation=f}', "");
Expect(1, 74868, '\p{Is_Terminal_Punctuation:-	F}', "");
Expect(0, 74868, '\p{^Is_Terminal_Punctuation:-	F}', "");
Expect(0, 74868, '\P{Is_Terminal_Punctuation:-	F}', "");
Expect(1, 74868, '\P{^Is_Terminal_Punctuation:-	F}', "");
Expect(0, 74867, '\p{Is_Terminal_Punctuation:-	F}', "");
Expect(1, 74867, '\p{^Is_Terminal_Punctuation:-	F}', "");
Expect(1, 74867, '\P{Is_Terminal_Punctuation:-	F}', "");
Expect(0, 74867, '\P{^Is_Terminal_Punctuation:-	F}', "");
Error('\p{Is_Term=-/a/FALSE}');
Error('\P{Is_Term=-/a/FALSE}');
Expect(1, 74868, '\p{Is_Term: false}', "");
Expect(0, 74868, '\p{^Is_Term: false}', "");
Expect(0, 74868, '\P{Is_Term: false}', "");
Expect(1, 74868, '\P{^Is_Term: false}', "");
Expect(0, 74867, '\p{Is_Term: false}', "");
Expect(1, 74867, '\p{^Is_Term: false}', "");
Expect(1, 74867, '\P{Is_Term: false}', "");
Expect(0, 74867, '\P{^Is_Term: false}', "");
Expect(1, 74868, '\p{Is_Term=-False}', "");
Expect(0, 74868, '\p{^Is_Term=-False}', "");
Expect(0, 74868, '\P{Is_Term=-False}', "");
Expect(1, 74868, '\P{^Is_Term=-False}', "");
Expect(0, 74867, '\p{Is_Term=-False}', "");
Expect(1, 74867, '\p{^Is_Term=-False}', "");
Expect(1, 74867, '\P{Is_Term=-False}', "");
Expect(0, 74867, '\P{^Is_Term=-False}', "");
Error('\p{Terminal_Punctuation=/a/	-yes}');
Error('\P{Terminal_Punctuation=/a/	-yes}');
Expect(1, 74867, '\p{Terminal_Punctuation=yes}', "");
Expect(0, 74867, '\p{^Terminal_Punctuation=yes}', "");
Expect(0, 74867, '\P{Terminal_Punctuation=yes}', "");
Expect(1, 74867, '\P{^Terminal_Punctuation=yes}', "");
Expect(0, 74868, '\p{Terminal_Punctuation=yes}', "");
Expect(1, 74868, '\p{^Terminal_Punctuation=yes}', "");
Expect(1, 74868, '\P{Terminal_Punctuation=yes}', "");
Expect(0, 74868, '\P{^Terminal_Punctuation=yes}', "");
Expect(1, 74867, '\p{Terminal_Punctuation= 	YES}', "");
Expect(0, 74867, '\p{^Terminal_Punctuation= 	YES}', "");
Expect(0, 74867, '\P{Terminal_Punctuation= 	YES}', "");
Expect(1, 74867, '\P{^Terminal_Punctuation= 	YES}', "");
Expect(0, 74868, '\p{Terminal_Punctuation= 	YES}', "");
Expect(1, 74868, '\p{^Terminal_Punctuation= 	YES}', "");
Expect(1, 74868, '\P{Terminal_Punctuation= 	YES}', "");
Expect(0, 74868, '\P{^Terminal_Punctuation= 	YES}', "");
Error('\p{Term=/a/-	y}');
Error('\P{Term=/a/-	y}');
Expect(1, 74867, '\p{Term=y}', "");
Expect(0, 74867, '\p{^Term=y}', "");
Expect(0, 74867, '\P{Term=y}', "");
Expect(1, 74867, '\P{^Term=y}', "");
Expect(0, 74868, '\p{Term=y}', "");
Expect(1, 74868, '\p{^Term=y}', "");
Expect(1, 74868, '\P{Term=y}', "");
Expect(0, 74868, '\P{^Term=y}', "");
Expect(1, 74867, '\p{Term=	 Y}', "");
Expect(0, 74867, '\p{^Term=	 Y}', "");
Expect(0, 74867, '\P{Term=	 Y}', "");
Expect(1, 74867, '\P{^Term=	 Y}', "");
Expect(0, 74868, '\p{Term=	 Y}', "");
Expect(1, 74868, '\p{^Term=	 Y}', "");
Expect(1, 74868, '\P{Term=	 Y}', "");
Expect(0, 74868, '\P{^Term=	 Y}', "");
Error('\p{Is_Terminal_Punctuation=:=		t}');
Error('\P{Is_Terminal_Punctuation=:=		t}');
Expect(1, 74867, '\p{Is_Terminal_Punctuation=t}', "");
Expect(0, 74867, '\p{^Is_Terminal_Punctuation=t}', "");
Expect(0, 74867, '\P{Is_Terminal_Punctuation=t}', "");
Expect(1, 74867, '\P{^Is_Terminal_Punctuation=t}', "");
Expect(0, 74868, '\p{Is_Terminal_Punctuation=t}', "");
Expect(1, 74868, '\p{^Is_Terminal_Punctuation=t}', "");
Expect(1, 74868, '\P{Is_Terminal_Punctuation=t}', "");
Expect(0, 74868, '\P{^Is_Terminal_Punctuation=t}', "");
Expect(1, 74867, '\p{Is_Terminal_Punctuation=  T}', "");
Expect(0, 74867, '\p{^Is_Terminal_Punctuation=  T}', "");
Expect(0, 74867, '\P{Is_Terminal_Punctuation=  T}', "");
Expect(1, 74867, '\P{^Is_Terminal_Punctuation=  T}', "");
Expect(0, 74868, '\p{Is_Terminal_Punctuation=  T}', "");
Expect(1, 74868, '\p{^Is_Terminal_Punctuation=  T}', "");
Expect(1, 74868, '\P{Is_Terminal_Punctuation=  T}', "");
Expect(0, 74868, '\P{^Is_Terminal_Punctuation=  T}', "");
Error('\p{Is_Term=/a/_ True}');
Error('\P{Is_Term=/a/_ True}');
Expect(1, 74867, '\p{Is_Term=true}', "");
Expect(0, 74867, '\p{^Is_Term=true}', "");
Expect(0, 74867, '\P{Is_Term=true}', "");
Expect(1, 74867, '\P{^Is_Term=true}', "");
Expect(0, 74868, '\p{Is_Term=true}', "");
Expect(1, 74868, '\p{^Is_Term=true}', "");
Expect(1, 74868, '\P{Is_Term=true}', "");
Expect(0, 74868, '\P{^Is_Term=true}', "");
Expect(1, 74867, '\p{Is_Term=-True}', "");
Expect(0, 74867, '\p{^Is_Term=-True}', "");
Expect(0, 74867, '\P{Is_Term=-True}', "");
Expect(1, 74867, '\P{^Is_Term=-True}', "");
Expect(0, 74868, '\p{Is_Term=-True}', "");
Expect(1, 74868, '\p{^Is_Term=-True}', "");
Expect(1, 74868, '\P{Is_Term=-True}', "");
Expect(0, 74868, '\P{^Is_Term=-True}', "");
Error('\p{Unified_Ideograph=:=__NO}');
Error('\P{Unified_Ideograph=:=__NO}');
Expect(1, 178206, '\p{Unified_Ideograph=no}', "");
Expect(0, 178206, '\p{^Unified_Ideograph=no}', "");
Expect(0, 178206, '\P{Unified_Ideograph=no}', "");
Expect(1, 178206, '\P{^Unified_Ideograph=no}', "");
Expect(0, 178205, '\p{Unified_Ideograph=no}', "");
Expect(1, 178205, '\p{^Unified_Ideograph=no}', "");
Expect(1, 178205, '\P{Unified_Ideograph=no}', "");
Expect(0, 178205, '\P{^Unified_Ideograph=no}', "");
Expect(1, 178206, '\p{Unified_Ideograph=_no}', "");
Expect(0, 178206, '\p{^Unified_Ideograph=_no}', "");
Expect(0, 178206, '\P{Unified_Ideograph=_no}', "");
Expect(1, 178206, '\P{^Unified_Ideograph=_no}', "");
Expect(0, 178205, '\p{Unified_Ideograph=_no}', "");
Expect(1, 178205, '\p{^Unified_Ideograph=_no}', "");
Expect(1, 178205, '\P{Unified_Ideograph=_no}', "");
Expect(0, 178205, '\P{^Unified_Ideograph=_no}', "");
Error('\p{UIdeo=/a/ _N}');
Error('\P{UIdeo=/a/ _N}');
Expect(1, 178206, '\p{UIdeo=n}', "");
Expect(0, 178206, '\p{^UIdeo=n}', "");
Expect(0, 178206, '\P{UIdeo=n}', "");
Expect(1, 178206, '\P{^UIdeo=n}', "");
Expect(0, 178205, '\p{UIdeo=n}', "");
Expect(1, 178205, '\p{^UIdeo=n}', "");
Expect(1, 178205, '\P{UIdeo=n}', "");
Expect(0, 178205, '\P{^UIdeo=n}', "");
Expect(1, 178206, '\p{UIdeo= N}', "");
Expect(0, 178206, '\p{^UIdeo= N}', "");
Expect(0, 178206, '\P{UIdeo= N}', "");
Expect(1, 178206, '\P{^UIdeo= N}', "");
Expect(0, 178205, '\p{UIdeo= N}', "");
Expect(1, 178205, '\p{^UIdeo= N}', "");
Expect(1, 178205, '\P{UIdeo= N}', "");
Expect(0, 178205, '\P{^UIdeo= N}', "");
Error('\p{Is_Unified_Ideograph=_f:=}');
Error('\P{Is_Unified_Ideograph=_f:=}');
Expect(1, 178206, '\p{Is_Unified_Ideograph=f}', "");
Expect(0, 178206, '\p{^Is_Unified_Ideograph=f}', "");
Expect(0, 178206, '\P{Is_Unified_Ideograph=f}', "");
Expect(1, 178206, '\P{^Is_Unified_Ideograph=f}', "");
Expect(0, 178205, '\p{Is_Unified_Ideograph=f}', "");
Expect(1, 178205, '\p{^Is_Unified_Ideograph=f}', "");
Expect(1, 178205, '\P{Is_Unified_Ideograph=f}', "");
Expect(0, 178205, '\P{^Is_Unified_Ideograph=f}', "");
Expect(1, 178206, '\p{Is_Unified_Ideograph:    F}', "");
Expect(0, 178206, '\p{^Is_Unified_Ideograph:    F}', "");
Expect(0, 178206, '\P{Is_Unified_Ideograph:    F}', "");
Expect(1, 178206, '\P{^Is_Unified_Ideograph:    F}', "");
Expect(0, 178205, '\p{Is_Unified_Ideograph:    F}', "");
Expect(1, 178205, '\p{^Is_Unified_Ideograph:    F}', "");
Expect(1, 178205, '\P{Is_Unified_Ideograph:    F}', "");
Expect(0, 178205, '\P{^Is_Unified_Ideograph:    F}', "");
Error('\p{Is_UIdeo= _False:=}');
Error('\P{Is_UIdeo= _False:=}');
Expect(1, 178206, '\p{Is_UIdeo=false}', "");
Expect(0, 178206, '\p{^Is_UIdeo=false}', "");
Expect(0, 178206, '\P{Is_UIdeo=false}', "");
Expect(1, 178206, '\P{^Is_UIdeo=false}', "");
Expect(0, 178205, '\p{Is_UIdeo=false}', "");
Expect(1, 178205, '\p{^Is_UIdeo=false}', "");
Expect(1, 178205, '\P{Is_UIdeo=false}', "");
Expect(0, 178205, '\P{^Is_UIdeo=false}', "");
Expect(1, 178206, '\p{Is_UIdeo=_	false}', "");
Expect(0, 178206, '\p{^Is_UIdeo=_	false}', "");
Expect(0, 178206, '\P{Is_UIdeo=_	false}', "");
Expect(1, 178206, '\P{^Is_UIdeo=_	false}', "");
Expect(0, 178205, '\p{Is_UIdeo=_	false}', "");
Expect(1, 178205, '\p{^Is_UIdeo=_	false}', "");
Expect(1, 178205, '\P{Is_UIdeo=_	false}', "");
Expect(0, 178205, '\P{^Is_UIdeo=_	false}', "");
Error('\p{Unified_Ideograph= Yes/a/}');
Error('\P{Unified_Ideograph= Yes/a/}');
Expect(1, 178205, '\p{Unified_Ideograph=yes}', "");
Expect(0, 178205, '\p{^Unified_Ideograph=yes}', "");
Expect(0, 178205, '\P{Unified_Ideograph=yes}', "");
Expect(1, 178205, '\P{^Unified_Ideograph=yes}', "");
Expect(0, 178206, '\p{Unified_Ideograph=yes}', "");
Expect(1, 178206, '\p{^Unified_Ideograph=yes}', "");
Expect(1, 178206, '\P{Unified_Ideograph=yes}', "");
Expect(0, 178206, '\P{^Unified_Ideograph=yes}', "");
Expect(1, 178205, '\p{Unified_Ideograph=_ yes}', "");
Expect(0, 178205, '\p{^Unified_Ideograph=_ yes}', "");
Expect(0, 178205, '\P{Unified_Ideograph=_ yes}', "");
Expect(1, 178205, '\P{^Unified_Ideograph=_ yes}', "");
Expect(0, 178206, '\p{Unified_Ideograph=_ yes}', "");
Expect(1, 178206, '\p{^Unified_Ideograph=_ yes}', "");
Expect(1, 178206, '\P{Unified_Ideograph=_ yes}', "");
Expect(0, 178206, '\P{^Unified_Ideograph=_ yes}', "");
Error('\p{UIdeo=:=y}');
Error('\P{UIdeo=:=y}');
Expect(1, 178205, '\p{UIdeo=y}', "");
Expect(0, 178205, '\p{^UIdeo=y}', "");
Expect(0, 178205, '\P{UIdeo=y}', "");
Expect(1, 178205, '\P{^UIdeo=y}', "");
Expect(0, 178206, '\p{UIdeo=y}', "");
Expect(1, 178206, '\p{^UIdeo=y}', "");
Expect(1, 178206, '\P{UIdeo=y}', "");
Expect(0, 178206, '\P{^UIdeo=y}', "");
Expect(1, 178205, '\p{UIdeo=_ Y}', "");
Expect(0, 178205, '\p{^UIdeo=_ Y}', "");
Expect(0, 178205, '\P{UIdeo=_ Y}', "");
Expect(1, 178205, '\P{^UIdeo=_ Y}', "");
Expect(0, 178206, '\p{UIdeo=_ Y}', "");
Expect(1, 178206, '\p{^UIdeo=_ Y}', "");
Expect(1, 178206, '\P{UIdeo=_ Y}', "");
Expect(0, 178206, '\P{^UIdeo=_ Y}', "");
Error('\p{Is_Unified_Ideograph=	T:=}');
Error('\P{Is_Unified_Ideograph=	T:=}');
Expect(1, 178205, '\p{Is_Unified_Ideograph=t}', "");
Expect(0, 178205, '\p{^Is_Unified_Ideograph=t}', "");
Expect(0, 178205, '\P{Is_Unified_Ideograph=t}', "");
Expect(1, 178205, '\P{^Is_Unified_Ideograph=t}', "");
Expect(0, 178206, '\p{Is_Unified_Ideograph=t}', "");
Expect(1, 178206, '\p{^Is_Unified_Ideograph=t}', "");
Expect(1, 178206, '\P{Is_Unified_Ideograph=t}', "");
Expect(0, 178206, '\P{^Is_Unified_Ideograph=t}', "");
Expect(1, 178205, '\p{Is_Unified_Ideograph= t}', "");
Expect(0, 178205, '\p{^Is_Unified_Ideograph= t}', "");
Expect(0, 178205, '\P{Is_Unified_Ideograph= t}', "");
Expect(1, 178205, '\P{^Is_Unified_Ideograph= t}', "");
Expect(0, 178206, '\p{Is_Unified_Ideograph= t}', "");
Expect(1, 178206, '\p{^Is_Unified_Ideograph= t}', "");
Expect(1, 178206, '\P{Is_Unified_Ideograph= t}', "");
Expect(0, 178206, '\P{^Is_Unified_Ideograph= t}', "");
Error('\p{Is_UIdeo=		true:=}');
Error('\P{Is_UIdeo=		true:=}');
Expect(1, 178205, '\p{Is_UIdeo=true}', "");
Expect(0, 178205, '\p{^Is_UIdeo=true}', "");
Expect(0, 178205, '\P{Is_UIdeo=true}', "");
Expect(1, 178205, '\P{^Is_UIdeo=true}', "");
Expect(0, 178206, '\p{Is_UIdeo=true}', "");
Expect(1, 178206, '\p{^Is_UIdeo=true}', "");
Expect(1, 178206, '\P{Is_UIdeo=true}', "");
Expect(0, 178206, '\P{^Is_UIdeo=true}', "");
Expect(1, 178205, '\p{Is_UIdeo=-True}', "");
Expect(0, 178205, '\p{^Is_UIdeo=-True}', "");
Expect(0, 178205, '\P{Is_UIdeo=-True}', "");
Expect(1, 178205, '\P{^Is_UIdeo=-True}', "");
Expect(0, 178206, '\p{Is_UIdeo=-True}', "");
Expect(1, 178206, '\p{^Is_UIdeo=-True}', "");
Expect(1, 178206, '\P{Is_UIdeo=-True}', "");
Expect(0, 178206, '\P{^Is_UIdeo=-True}', "");
Error('\p{Uppercase=-No:=}');
Error('\P{Uppercase=-No:=}');
Expect(1, 120779, '\p{Uppercase=no}', "");
Expect(0, 120779, '\p{^Uppercase=no}', "");
Expect(0, 120779, '\P{Uppercase=no}', "");
Expect(1, 120779, '\P{^Uppercase=no}', "");
Expect(0, 120778, '\p{Uppercase=no}', "");
Expect(1, 120778, '\p{^Uppercase=no}', "");
Expect(1, 120778, '\P{Uppercase=no}', "");
Expect(0, 120778, '\P{^Uppercase=no}', "");
Expect(1, 120779, '\p{Uppercase=-No}', "");
Expect(0, 120779, '\p{^Uppercase=-No}', "");
Expect(0, 120779, '\P{Uppercase=-No}', "");
Expect(1, 120779, '\P{^Uppercase=-No}', "");
Expect(0, 120778, '\p{Uppercase=-No}', "");
Expect(1, 120778, '\p{^Uppercase=-No}', "");
Expect(1, 120778, '\P{Uppercase=-No}', "");
Expect(0, 120778, '\P{^Uppercase=-No}', "");
Error('\p{Upper=:=--N}');
Error('\P{Upper=:=--N}');
Expect(1, 120779, '\p{Upper=n}', "");
Expect(0, 120779, '\p{^Upper=n}', "");
Expect(0, 120779, '\P{Upper=n}', "");
Expect(1, 120779, '\P{^Upper=n}', "");
Expect(0, 120778, '\p{Upper=n}', "");
Expect(1, 120778, '\p{^Upper=n}', "");
Expect(1, 120778, '\P{Upper=n}', "");
Expect(0, 120778, '\P{^Upper=n}', "");
Expect(1, 120779, '\p{Upper=-	N}', "");
Expect(0, 120779, '\p{^Upper=-	N}', "");
Expect(0, 120779, '\P{Upper=-	N}', "");
Expect(1, 120779, '\P{^Upper=-	N}', "");
Expect(0, 120778, '\p{Upper=-	N}', "");
Expect(1, 120778, '\p{^Upper=-	N}', "");
Expect(1, 120778, '\P{Upper=-	N}', "");
Expect(0, 120778, '\P{^Upper=-	N}', "");
Error('\p{Is_Uppercase=/a/  F}');
Error('\P{Is_Uppercase=/a/  F}');
Expect(1, 120779, '\p{Is_Uppercase=f}', "");
Expect(0, 120779, '\p{^Is_Uppercase=f}', "");
Expect(0, 120779, '\P{Is_Uppercase=f}', "");
Expect(1, 120779, '\P{^Is_Uppercase=f}', "");
Expect(0, 120778, '\p{Is_Uppercase=f}', "");
Expect(1, 120778, '\p{^Is_Uppercase=f}', "");
Expect(1, 120778, '\P{Is_Uppercase=f}', "");
Expect(0, 120778, '\P{^Is_Uppercase=f}', "");
Expect(1, 120779, '\p{Is_Uppercase:	- F}', "");
Expect(0, 120779, '\p{^Is_Uppercase:	- F}', "");
Expect(0, 120779, '\P{Is_Uppercase:	- F}', "");
Expect(1, 120779, '\P{^Is_Uppercase:	- F}', "");
Expect(0, 120778, '\p{Is_Uppercase:	- F}', "");
Expect(1, 120778, '\p{^Is_Uppercase:	- F}', "");
Expect(1, 120778, '\P{Is_Uppercase:	- F}', "");
Expect(0, 120778, '\P{^Is_Uppercase:	- F}', "");
Error('\p{Is_Upper=- FALSE/a/}');
Error('\P{Is_Upper=- FALSE/a/}');
Expect(1, 120779, '\p{Is_Upper=false}', "");
Expect(0, 120779, '\p{^Is_Upper=false}', "");
Expect(0, 120779, '\P{Is_Upper=false}', "");
Expect(1, 120779, '\P{^Is_Upper=false}', "");
Expect(0, 120778, '\p{Is_Upper=false}', "");
Expect(1, 120778, '\p{^Is_Upper=false}', "");
Expect(1, 120778, '\P{Is_Upper=false}', "");
Expect(0, 120778, '\P{^Is_Upper=false}', "");
Expect(1, 120779, '\p{Is_Upper=-FALSE}', "");
Expect(0, 120779, '\p{^Is_Upper=-FALSE}', "");
Expect(0, 120779, '\P{Is_Upper=-FALSE}', "");
Expect(1, 120779, '\P{^Is_Upper=-FALSE}', "");
Expect(0, 120778, '\p{Is_Upper=-FALSE}', "");
Expect(1, 120778, '\p{^Is_Upper=-FALSE}', "");
Expect(1, 120778, '\P{Is_Upper=-FALSE}', "");
Expect(0, 120778, '\P{^Is_Upper=-FALSE}', "");
Error('\p{Uppercase=__yes/a/}');
Error('\P{Uppercase=__yes/a/}');
Expect(1, 120778, '\p{Uppercase=yes}', "");
Expect(0, 120778, '\p{^Uppercase=yes}', "");
Expect(0, 120778, '\P{Uppercase=yes}', "");
Expect(1, 120778, '\P{^Uppercase=yes}', "");
Expect(0, 120779, '\p{Uppercase=yes}', "");
Expect(1, 120779, '\p{^Uppercase=yes}', "");
Expect(1, 120779, '\P{Uppercase=yes}', "");
Expect(0, 120779, '\P{^Uppercase=yes}', "");
Expect(1, 120778, '\p{Uppercase=_	yes}', "");
Expect(0, 120778, '\p{^Uppercase=_	yes}', "");
Expect(0, 120778, '\P{Uppercase=_	yes}', "");
Expect(1, 120778, '\P{^Uppercase=_	yes}', "");
Expect(0, 120779, '\p{Uppercase=_	yes}', "");
Expect(1, 120779, '\p{^Uppercase=_	yes}', "");
Expect(1, 120779, '\P{Uppercase=_	yes}', "");
Expect(0, 120779, '\P{^Uppercase=_	yes}', "");
Error('\p{Upper= :=Y}');
Error('\P{Upper= :=Y}');
Expect(1, 120778, '\p{Upper=y}', "");
Expect(0, 120778, '\p{^Upper=y}', "");
Expect(0, 120778, '\P{Upper=y}', "");
Expect(1, 120778, '\P{^Upper=y}', "");
Expect(0, 120779, '\p{Upper=y}', "");
Expect(1, 120779, '\p{^Upper=y}', "");
Expect(1, 120779, '\P{Upper=y}', "");
Expect(0, 120779, '\P{^Upper=y}', "");
Expect(1, 120778, '\p{Upper: _Y}', "");
Expect(0, 120778, '\p{^Upper: _Y}', "");
Expect(0, 120778, '\P{Upper: _Y}', "");
Expect(1, 120778, '\P{^Upper: _Y}', "");
Expect(0, 120779, '\p{Upper: _Y}', "");
Expect(1, 120779, '\p{^Upper: _Y}', "");
Expect(1, 120779, '\P{Upper: _Y}', "");
Expect(0, 120779, '\P{^Upper: _Y}', "");
Error('\p{Is_Uppercase=--T:=}');
Error('\P{Is_Uppercase=--T:=}');
Expect(1, 120778, '\p{Is_Uppercase=t}', "");
Expect(0, 120778, '\p{^Is_Uppercase=t}', "");
Expect(0, 120778, '\P{Is_Uppercase=t}', "");
Expect(1, 120778, '\P{^Is_Uppercase=t}', "");
Expect(0, 120779, '\p{Is_Uppercase=t}', "");
Expect(1, 120779, '\p{^Is_Uppercase=t}', "");
Expect(1, 120779, '\P{Is_Uppercase=t}', "");
Expect(0, 120779, '\P{^Is_Uppercase=t}', "");
Expect(1, 120778, '\p{Is_Uppercase= 	T}', "");
Expect(0, 120778, '\p{^Is_Uppercase= 	T}', "");
Expect(0, 120778, '\P{Is_Uppercase= 	T}', "");
Expect(1, 120778, '\P{^Is_Uppercase= 	T}', "");
Expect(0, 120779, '\p{Is_Uppercase= 	T}', "");
Expect(1, 120779, '\p{^Is_Uppercase= 	T}', "");
Expect(1, 120779, '\P{Is_Uppercase= 	T}', "");
Expect(0, 120779, '\P{^Is_Uppercase= 	T}', "");
Error('\p{Is_Upper:   :=  TRUE}');
Error('\P{Is_Upper:   :=  TRUE}');
Expect(1, 120778, '\p{Is_Upper=true}', "");
Expect(0, 120778, '\p{^Is_Upper=true}', "");
Expect(0, 120778, '\P{Is_Upper=true}', "");
Expect(1, 120778, '\P{^Is_Upper=true}', "");
Expect(0, 120779, '\p{Is_Upper=true}', "");
Expect(1, 120779, '\p{^Is_Upper=true}', "");
Expect(1, 120779, '\P{Is_Upper=true}', "");
Expect(0, 120779, '\P{^Is_Upper=true}', "");
Expect(1, 120778, '\p{Is_Upper=_true}', "");
Expect(0, 120778, '\p{^Is_Upper=_true}', "");
Expect(0, 120778, '\P{Is_Upper=_true}', "");
Expect(1, 120778, '\P{^Is_Upper=_true}', "");
Expect(0, 120779, '\p{Is_Upper=_true}', "");
Expect(1, 120779, '\p{^Is_Upper=_true}', "");
Expect(1, 120779, '\P{Is_Upper=_true}', "");
Expect(0, 120779, '\P{^Is_Upper=_true}', "");
Error('\p{Variation_Selector=	/a/no}');
Error('\P{Variation_Selector=	/a/no}');
Expect(1, 918000, '\p{Variation_Selector:   no}', "");
Expect(0, 918000, '\p{^Variation_Selector:   no}', "");
Expect(0, 918000, '\P{Variation_Selector:   no}', "");
Expect(1, 918000, '\P{^Variation_Selector:   no}', "");
Expect(0, 917999, '\p{Variation_Selector:   no}', "");
Expect(1, 917999, '\p{^Variation_Selector:   no}', "");
Expect(1, 917999, '\P{Variation_Selector:   no}', "");
Expect(0, 917999, '\P{^Variation_Selector:   no}', "");
Expect(1, 918000, '\p{Variation_Selector=	no}', "");
Expect(0, 918000, '\p{^Variation_Selector=	no}', "");
Expect(0, 918000, '\P{Variation_Selector=	no}', "");
Expect(1, 918000, '\P{^Variation_Selector=	no}', "");
Expect(0, 917999, '\p{Variation_Selector=	no}', "");
Expect(1, 917999, '\p{^Variation_Selector=	no}', "");
Expect(1, 917999, '\P{Variation_Selector=	no}', "");
Expect(0, 917999, '\P{^Variation_Selector=	no}', "");
Error('\p{VS=_	N/a/}');
Error('\P{VS=_	N/a/}');
Expect(1, 918000, '\p{VS:   n}', "");
Expect(0, 918000, '\p{^VS:   n}', "");
Expect(0, 918000, '\P{VS:   n}', "");
Expect(1, 918000, '\P{^VS:   n}', "");
Expect(0, 917999, '\p{VS:   n}', "");
Expect(1, 917999, '\p{^VS:   n}', "");
Expect(1, 917999, '\P{VS:   n}', "");
Expect(0, 917999, '\P{^VS:   n}', "");
Expect(1, 918000, '\p{VS= 	n}', "");
Expect(0, 918000, '\p{^VS= 	n}', "");
Expect(0, 918000, '\P{VS= 	n}', "");
Expect(1, 918000, '\P{^VS= 	n}', "");
Expect(0, 917999, '\p{VS= 	n}', "");
Expect(1, 917999, '\p{^VS= 	n}', "");
Expect(1, 917999, '\P{VS= 	n}', "");
Expect(0, 917999, '\P{^VS= 	n}', "");
Error('\p{Is_Variation_Selector:/a/__F}');
Error('\P{Is_Variation_Selector:/a/__F}');
Expect(1, 918000, '\p{Is_Variation_Selector=f}', "");
Expect(0, 918000, '\p{^Is_Variation_Selector=f}', "");
Expect(0, 918000, '\P{Is_Variation_Selector=f}', "");
Expect(1, 918000, '\P{^Is_Variation_Selector=f}', "");
Expect(0, 917999, '\p{Is_Variation_Selector=f}', "");
Expect(1, 917999, '\p{^Is_Variation_Selector=f}', "");
Expect(1, 917999, '\P{Is_Variation_Selector=f}', "");
Expect(0, 917999, '\P{^Is_Variation_Selector=f}', "");
Expect(1, 918000, '\p{Is_Variation_Selector:-_f}', "");
Expect(0, 918000, '\p{^Is_Variation_Selector:-_f}', "");
Expect(0, 918000, '\P{Is_Variation_Selector:-_f}', "");
Expect(1, 918000, '\P{^Is_Variation_Selector:-_f}', "");
Expect(0, 917999, '\p{Is_Variation_Selector:-_f}', "");
Expect(1, 917999, '\p{^Is_Variation_Selector:-_f}', "");
Expect(1, 917999, '\P{Is_Variation_Selector:-_f}', "");
Expect(0, 917999, '\P{^Is_Variation_Selector:-_f}', "");
Error('\p{Is_VS=_/a/False}');
Error('\P{Is_VS=_/a/False}');
Expect(1, 918000, '\p{Is_VS=false}', "");
Expect(0, 918000, '\p{^Is_VS=false}', "");
Expect(0, 918000, '\P{Is_VS=false}', "");
Expect(1, 918000, '\P{^Is_VS=false}', "");
Expect(0, 917999, '\p{Is_VS=false}', "");
Expect(1, 917999, '\p{^Is_VS=false}', "");
Expect(1, 917999, '\P{Is_VS=false}', "");
Expect(0, 917999, '\P{^Is_VS=false}', "");
Expect(1, 918000, '\p{Is_VS=- false}', "");
Expect(0, 918000, '\p{^Is_VS=- false}', "");
Expect(0, 918000, '\P{Is_VS=- false}', "");
Expect(1, 918000, '\P{^Is_VS=- false}', "");
Expect(0, 917999, '\p{Is_VS=- false}', "");
Expect(1, 917999, '\p{^Is_VS=- false}', "");
Expect(1, 917999, '\P{Is_VS=- false}', "");
Expect(0, 917999, '\P{^Is_VS=- false}', "");
Error('\p{Variation_Selector=/a/yes}');
Error('\P{Variation_Selector=/a/yes}');
Expect(1, 917999, '\p{Variation_Selector=yes}', "");
Expect(0, 917999, '\p{^Variation_Selector=yes}', "");
Expect(0, 917999, '\P{Variation_Selector=yes}', "");
Expect(1, 917999, '\P{^Variation_Selector=yes}', "");
Expect(0, 918000, '\p{Variation_Selector=yes}', "");
Expect(1, 918000, '\p{^Variation_Selector=yes}', "");
Expect(1, 918000, '\P{Variation_Selector=yes}', "");
Expect(0, 918000, '\P{^Variation_Selector=yes}', "");
Expect(1, 917999, '\p{Variation_Selector=		YES}', "");
Expect(0, 917999, '\p{^Variation_Selector=		YES}', "");
Expect(0, 917999, '\P{Variation_Selector=		YES}', "");
Expect(1, 917999, '\P{^Variation_Selector=		YES}', "");
Expect(0, 918000, '\p{Variation_Selector=		YES}', "");
Expect(1, 918000, '\p{^Variation_Selector=		YES}', "");
Expect(1, 918000, '\P{Variation_Selector=		YES}', "");
Expect(0, 918000, '\P{^Variation_Selector=		YES}', "");
Error('\p{VS=	_y:=}');
Error('\P{VS=	_y:=}');
Expect(1, 917999, '\p{VS=y}', "");
Expect(0, 917999, '\p{^VS=y}', "");
Expect(0, 917999, '\P{VS=y}', "");
Expect(1, 917999, '\P{^VS=y}', "");
Expect(0, 918000, '\p{VS=y}', "");
Expect(1, 918000, '\p{^VS=y}', "");
Expect(1, 918000, '\P{VS=y}', "");
Expect(0, 918000, '\P{^VS=y}', "");
Expect(1, 917999, '\p{VS=_y}', "");
Expect(0, 917999, '\p{^VS=_y}', "");
Expect(0, 917999, '\P{VS=_y}', "");
Expect(1, 917999, '\P{^VS=_y}', "");
Expect(0, 918000, '\p{VS=_y}', "");
Expect(1, 918000, '\p{^VS=_y}', "");
Expect(1, 918000, '\P{VS=_y}', "");
Expect(0, 918000, '\P{^VS=_y}', "");
Error('\p{Is_Variation_Selector=_/a/T}');
Error('\P{Is_Variation_Selector=_/a/T}');
Expect(1, 917999, '\p{Is_Variation_Selector:   t}', "");
Expect(0, 917999, '\p{^Is_Variation_Selector:   t}', "");
Expect(0, 917999, '\P{Is_Variation_Selector:   t}', "");
Expect(1, 917999, '\P{^Is_Variation_Selector:   t}', "");
Expect(0, 918000, '\p{Is_Variation_Selector:   t}', "");
Expect(1, 918000, '\p{^Is_Variation_Selector:   t}', "");
Expect(1, 918000, '\P{Is_Variation_Selector:   t}', "");
Expect(0, 918000, '\P{^Is_Variation_Selector:   t}', "");
Expect(1, 917999, '\p{Is_Variation_Selector=--T}', "");
Expect(0, 917999, '\p{^Is_Variation_Selector=--T}', "");
Expect(0, 917999, '\P{Is_Variation_Selector=--T}', "");
Expect(1, 917999, '\P{^Is_Variation_Selector=--T}', "");
Expect(0, 918000, '\p{Is_Variation_Selector=--T}', "");
Expect(1, 918000, '\p{^Is_Variation_Selector=--T}', "");
Expect(1, 918000, '\P{Is_Variation_Selector=--T}', "");
Expect(0, 918000, '\P{^Is_Variation_Selector=--T}', "");
Error('\p{Is_VS= /a/TRUE}');
Error('\P{Is_VS= /a/TRUE}');
Expect(1, 917999, '\p{Is_VS=true}', "");
Expect(0, 917999, '\p{^Is_VS=true}', "");
Expect(0, 917999, '\P{Is_VS=true}', "");
Expect(1, 917999, '\P{^Is_VS=true}', "");
Expect(0, 918000, '\p{Is_VS=true}', "");
Expect(1, 918000, '\p{^Is_VS=true}', "");
Expect(1, 918000, '\P{Is_VS=true}', "");
Expect(0, 918000, '\P{^Is_VS=true}', "");
Expect(1, 917999, '\p{Is_VS:			True}', "");
Expect(0, 917999, '\p{^Is_VS:			True}', "");
Expect(0, 917999, '\P{Is_VS:			True}', "");
Expect(1, 917999, '\P{^Is_VS:			True}', "");
Expect(0, 918000, '\p{Is_VS:			True}', "");
Expect(1, 918000, '\p{^Is_VS:			True}', "");
Expect(1, 918000, '\P{Is_VS:			True}', "");
Expect(0, 918000, '\P{^Is_VS:			True}', "");
Error('\p{Word_Break= :=CR}');
Error('\P{Word_Break= :=CR}');
Expect(1, 13, '\p{Word_Break=cr}', "");
Expect(0, 13, '\p{^Word_Break=cr}', "");
Expect(0, 13, '\P{Word_Break=cr}', "");
Expect(1, 13, '\P{^Word_Break=cr}', "");
Expect(0, 14, '\p{Word_Break=cr}', "");
Expect(1, 14, '\p{^Word_Break=cr}', "");
Expect(1, 14, '\P{Word_Break=cr}', "");
Expect(0, 14, '\P{^Word_Break=cr}', "");
Expect(1, 13, '\p{Word_Break=_-CR}', "");
Expect(0, 13, '\p{^Word_Break=_-CR}', "");
Expect(0, 13, '\P{Word_Break=_-CR}', "");
Expect(1, 13, '\P{^Word_Break=_-CR}', "");
Expect(0, 14, '\p{Word_Break=_-CR}', "");
Expect(1, 14, '\p{^Word_Break=_-CR}', "");
Expect(1, 14, '\P{Word_Break=_-CR}', "");
Expect(0, 14, '\P{^Word_Break=_-CR}', "");
Error('\p{WB=--cr:=}');
Error('\P{WB=--cr:=}');
Expect(1, 13, '\p{WB=cr}', "");
Expect(0, 13, '\p{^WB=cr}', "");
Expect(0, 13, '\P{WB=cr}', "");
Expect(1, 13, '\P{^WB=cr}', "");
Expect(0, 14, '\p{WB=cr}', "");
Expect(1, 14, '\p{^WB=cr}', "");
Expect(1, 14, '\P{WB=cr}', "");
Expect(0, 14, '\P{^WB=cr}', "");
Expect(1, 13, '\p{WB= CR}', "");
Expect(0, 13, '\p{^WB= CR}', "");
Expect(0, 13, '\P{WB= CR}', "");
Expect(1, 13, '\P{^WB= CR}', "");
Expect(0, 14, '\p{WB= CR}', "");
Expect(1, 14, '\p{^WB= CR}', "");
Expect(1, 14, '\P{WB= CR}', "");
Expect(0, 14, '\P{^WB= CR}', "");
Error('\p{Is_Word_Break=:=-	CR}');
Error('\P{Is_Word_Break=:=-	CR}');
Expect(1, 13, '\p{Is_Word_Break:	cr}', "");
Expect(0, 13, '\p{^Is_Word_Break:	cr}', "");
Expect(0, 13, '\P{Is_Word_Break:	cr}', "");
Expect(1, 13, '\P{^Is_Word_Break:	cr}', "");
Expect(0, 14, '\p{Is_Word_Break:	cr}', "");
Expect(1, 14, '\p{^Is_Word_Break:	cr}', "");
Expect(1, 14, '\P{Is_Word_Break:	cr}', "");
Expect(0, 14, '\P{^Is_Word_Break:	cr}', "");
Expect(1, 13, '\p{Is_Word_Break=CR}', "");
Expect(0, 13, '\p{^Is_Word_Break=CR}', "");
Expect(0, 13, '\P{Is_Word_Break=CR}', "");
Expect(1, 13, '\P{^Is_Word_Break=CR}', "");
Expect(0, 14, '\p{Is_Word_Break=CR}', "");
Expect(1, 14, '\p{^Is_Word_Break=CR}', "");
Expect(1, 14, '\P{Is_Word_Break=CR}', "");
Expect(0, 14, '\P{^Is_Word_Break=CR}', "");
Error('\p{Is_WB=/a/CR}');
Error('\P{Is_WB=/a/CR}');
Expect(1, 13, '\p{Is_WB=cr}', "");
Expect(0, 13, '\p{^Is_WB=cr}', "");
Expect(0, 13, '\P{Is_WB=cr}', "");
Expect(1, 13, '\P{^Is_WB=cr}', "");
Expect(0, 14, '\p{Is_WB=cr}', "");
Expect(1, 14, '\p{^Is_WB=cr}', "");
Expect(1, 14, '\P{Is_WB=cr}', "");
Expect(0, 14, '\P{^Is_WB=cr}', "");
Expect(1, 13, '\p{Is_WB=		CR}', "");
Expect(0, 13, '\p{^Is_WB=		CR}', "");
Expect(0, 13, '\P{Is_WB=		CR}', "");
Expect(1, 13, '\P{^Is_WB=		CR}', "");
Expect(0, 14, '\p{Is_WB=		CR}', "");
Expect(1, 14, '\p{^Is_WB=		CR}', "");
Expect(1, 14, '\P{Is_WB=		CR}', "");
Expect(0, 14, '\P{^Is_WB=		CR}', "");
Error('\p{Word_Break:/a/-DOUBLE_Quote}');
Error('\P{Word_Break:/a/-DOUBLE_Quote}');
Expect(1, 34, '\p{Word_Break=doublequote}', "");
Expect(0, 34, '\p{^Word_Break=doublequote}', "");
Expect(0, 34, '\P{Word_Break=doublequote}', "");
Expect(1, 34, '\P{^Word_Break=doublequote}', "");
Expect(0, 35, '\p{Word_Break=doublequote}', "");
Expect(1, 35, '\p{^Word_Break=doublequote}', "");
Expect(1, 35, '\P{Word_Break=doublequote}', "");
Expect(0, 35, '\P{^Word_Break=doublequote}', "");
Expect(1, 34, '\p{Word_Break:   - double_Quote}', "");
Expect(0, 34, '\p{^Word_Break:   - double_Quote}', "");
Expect(0, 34, '\P{Word_Break:   - double_Quote}', "");
Expect(1, 34, '\P{^Word_Break:   - double_Quote}', "");
Expect(0, 35, '\p{Word_Break:   - double_Quote}', "");
Expect(1, 35, '\p{^Word_Break:   - double_Quote}', "");
Expect(1, 35, '\P{Word_Break:   - double_Quote}', "");
Expect(0, 35, '\P{^Word_Break:   - double_Quote}', "");
Error('\p{WB=_	DQ/a/}');
Error('\P{WB=_	DQ/a/}');
Expect(1, 34, '\p{WB=dq}', "");
Expect(0, 34, '\p{^WB=dq}', "");
Expect(0, 34, '\P{WB=dq}', "");
Expect(1, 34, '\P{^WB=dq}', "");
Expect(0, 35, '\p{WB=dq}', "");
Expect(1, 35, '\p{^WB=dq}', "");
Expect(1, 35, '\P{WB=dq}', "");
Expect(0, 35, '\P{^WB=dq}', "");
Expect(1, 34, '\p{WB= -DQ}', "");
Expect(0, 34, '\p{^WB= -DQ}', "");
Expect(0, 34, '\P{WB= -DQ}', "");
Expect(1, 34, '\P{^WB= -DQ}', "");
Expect(0, 35, '\p{WB= -DQ}', "");
Expect(1, 35, '\p{^WB= -DQ}', "");
Expect(1, 35, '\P{WB= -DQ}', "");
Expect(0, 35, '\P{^WB= -DQ}', "");
Error('\p{Is_Word_Break=/a/_Double_Quote}');
Error('\P{Is_Word_Break=/a/_Double_Quote}');
Expect(1, 34, '\p{Is_Word_Break=doublequote}', "");
Expect(0, 34, '\p{^Is_Word_Break=doublequote}', "");
Expect(0, 34, '\P{Is_Word_Break=doublequote}', "");
Expect(1, 34, '\P{^Is_Word_Break=doublequote}', "");
Expect(0, 35, '\p{Is_Word_Break=doublequote}', "");
Expect(1, 35, '\p{^Is_Word_Break=doublequote}', "");
Expect(1, 35, '\P{Is_Word_Break=doublequote}', "");
Expect(0, 35, '\P{^Is_Word_Break=doublequote}', "");
Expect(1, 34, '\p{Is_Word_Break=	-Double_quote}', "");
Expect(0, 34, '\p{^Is_Word_Break=	-Double_quote}', "");
Expect(0, 34, '\P{Is_Word_Break=	-Double_quote}', "");
Expect(1, 34, '\P{^Is_Word_Break=	-Double_quote}', "");
Expect(0, 35, '\p{Is_Word_Break=	-Double_quote}', "");
Expect(1, 35, '\p{^Is_Word_Break=	-Double_quote}', "");
Expect(1, 35, '\P{Is_Word_Break=	-Double_quote}', "");
Expect(0, 35, '\P{^Is_Word_Break=	-Double_quote}', "");
Error('\p{Is_WB=	:=dq}');
Error('\P{Is_WB=	:=dq}');
Expect(1, 34, '\p{Is_WB=dq}', "");
Expect(0, 34, '\p{^Is_WB=dq}', "");
Expect(0, 34, '\P{Is_WB=dq}', "");
Expect(1, 34, '\P{^Is_WB=dq}', "");
Expect(0, 35, '\p{Is_WB=dq}', "");
Expect(1, 35, '\p{^Is_WB=dq}', "");
Expect(1, 35, '\P{Is_WB=dq}', "");
Expect(0, 35, '\P{^Is_WB=dq}', "");
Expect(1, 34, '\p{Is_WB=- DQ}', "");
Expect(0, 34, '\p{^Is_WB=- DQ}', "");
Expect(0, 34, '\P{Is_WB=- DQ}', "");
Expect(1, 34, '\P{^Is_WB=- DQ}', "");
Expect(0, 35, '\p{Is_WB=- DQ}', "");
Expect(1, 35, '\p{^Is_WB=- DQ}', "");
Expect(1, 35, '\P{Is_WB=- DQ}', "");
Expect(0, 35, '\P{^Is_WB=- DQ}', "");
Error('\p{Word_Break=-:=EXTENDNUMLET}');
Error('\P{Word_Break=-:=EXTENDNUMLET}');
Expect(1, 65343, '\p{Word_Break=extendnumlet}', "");
Expect(0, 65343, '\p{^Word_Break=extendnumlet}', "");
Expect(0, 65343, '\P{Word_Break=extendnumlet}', "");
Expect(1, 65343, '\P{^Word_Break=extendnumlet}', "");
Expect(0, 65344, '\p{Word_Break=extendnumlet}', "");
Expect(1, 65344, '\p{^Word_Break=extendnumlet}', "");
Expect(1, 65344, '\P{Word_Break=extendnumlet}', "");
Expect(0, 65344, '\P{^Word_Break=extendnumlet}', "");
Expect(1, 65343, '\p{Word_Break= _EXTENDNUMLET}', "");
Expect(0, 65343, '\p{^Word_Break= _EXTENDNUMLET}', "");
Expect(0, 65343, '\P{Word_Break= _EXTENDNUMLET}', "");
Expect(1, 65343, '\P{^Word_Break= _EXTENDNUMLET}', "");
Expect(0, 65344, '\p{Word_Break= _EXTENDNUMLET}', "");
Expect(1, 65344, '\p{^Word_Break= _EXTENDNUMLET}', "");
Expect(1, 65344, '\P{Word_Break= _EXTENDNUMLET}', "");
Expect(0, 65344, '\P{^Word_Break= _EXTENDNUMLET}', "");
Error('\p{WB=:=	_EX}');
Error('\P{WB=:=	_EX}');
Expect(1, 65343, '\p{WB=ex}', "");
Expect(0, 65343, '\p{^WB=ex}', "");
Expect(0, 65343, '\P{WB=ex}', "");
Expect(1, 65343, '\P{^WB=ex}', "");
Expect(0, 65344, '\p{WB=ex}', "");
Expect(1, 65344, '\p{^WB=ex}', "");
Expect(1, 65344, '\P{WB=ex}', "");
Expect(0, 65344, '\P{^WB=ex}', "");
Expect(1, 65343, '\p{WB= -EX}', "");
Expect(0, 65343, '\p{^WB= -EX}', "");
Expect(0, 65343, '\P{WB= -EX}', "");
Expect(1, 65343, '\P{^WB= -EX}', "");
Expect(0, 65344, '\p{WB= -EX}', "");
Expect(1, 65344, '\p{^WB= -EX}', "");
Expect(1, 65344, '\P{WB= -EX}', "");
Expect(0, 65344, '\P{^WB= -EX}', "");
Error('\p{Is_Word_Break= _ExtendNumLet:=}');
Error('\P{Is_Word_Break= _ExtendNumLet:=}');
Expect(1, 65343, '\p{Is_Word_Break=extendnumlet}', "");
Expect(0, 65343, '\p{^Is_Word_Break=extendnumlet}', "");
Expect(0, 65343, '\P{Is_Word_Break=extendnumlet}', "");
Expect(1, 65343, '\P{^Is_Word_Break=extendnumlet}', "");
Expect(0, 65344, '\p{Is_Word_Break=extendnumlet}', "");
Expect(1, 65344, '\p{^Is_Word_Break=extendnumlet}', "");
Expect(1, 65344, '\P{Is_Word_Break=extendnumlet}', "");
Expect(0, 65344, '\P{^Is_Word_Break=extendnumlet}', "");
Expect(1, 65343, '\p{Is_Word_Break=	 EXTENDNUMLET}', "");
Expect(0, 65343, '\p{^Is_Word_Break=	 EXTENDNUMLET}', "");
Expect(0, 65343, '\P{Is_Word_Break=	 EXTENDNUMLET}', "");
Expect(1, 65343, '\P{^Is_Word_Break=	 EXTENDNUMLET}', "");
Expect(0, 65344, '\p{Is_Word_Break=	 EXTENDNUMLET}', "");
Expect(1, 65344, '\p{^Is_Word_Break=	 EXTENDNUMLET}', "");
Expect(1, 65344, '\P{Is_Word_Break=	 EXTENDNUMLET}', "");
Expect(0, 65344, '\P{^Is_Word_Break=	 EXTENDNUMLET}', "");
Error('\p{Is_WB=_/a/EX}');
Error('\P{Is_WB=_/a/EX}');
Expect(1, 65343, '\p{Is_WB=ex}', "");
Expect(0, 65343, '\p{^Is_WB=ex}', "");
Expect(0, 65343, '\P{Is_WB=ex}', "");
Expect(1, 65343, '\P{^Is_WB=ex}', "");
Expect(0, 65344, '\p{Is_WB=ex}', "");
Expect(1, 65344, '\p{^Is_WB=ex}', "");
Expect(1, 65344, '\P{Is_WB=ex}', "");
Expect(0, 65344, '\P{^Is_WB=ex}', "");
Expect(1, 65343, '\p{Is_WB= EX}', "");
Expect(0, 65343, '\p{^Is_WB= EX}', "");
Expect(0, 65343, '\P{Is_WB= EX}', "");
Expect(1, 65343, '\P{^Is_WB= EX}', "");
Expect(0, 65344, '\p{Is_WB= EX}', "");
Expect(1, 65344, '\p{^Is_WB= EX}', "");
Expect(1, 65344, '\P{Is_WB= EX}', "");
Expect(0, 65344, '\P{^Is_WB= EX}', "");
Error('\p{Word_Break= EXTEND/a/}');
Error('\P{Word_Break= EXTEND/a/}');
Expect(1, 917999, '\p{Word_Break=extend}', "");
Expect(0, 917999, '\p{^Word_Break=extend}', "");
Expect(0, 917999, '\P{Word_Break=extend}', "");
Expect(1, 917999, '\P{^Word_Break=extend}', "");
Expect(0, 918000, '\p{Word_Break=extend}', "");
Expect(1, 918000, '\p{^Word_Break=extend}', "");
Expect(1, 918000, '\P{Word_Break=extend}', "");
Expect(0, 918000, '\P{^Word_Break=extend}', "");
Expect(1, 917999, '\p{Word_Break=	_extend}', "");
Expect(0, 917999, '\p{^Word_Break=	_extend}', "");
Expect(0, 917999, '\P{Word_Break=	_extend}', "");
Expect(1, 917999, '\P{^Word_Break=	_extend}', "");
Expect(0, 918000, '\p{Word_Break=	_extend}', "");
Expect(1, 918000, '\p{^Word_Break=	_extend}', "");
Expect(1, 918000, '\P{Word_Break=	_extend}', "");
Expect(0, 918000, '\P{^Word_Break=	_extend}', "");
Error('\p{WB:__Extend/a/}');
Error('\P{WB:__Extend/a/}');
Expect(1, 917999, '\p{WB=extend}', "");
Expect(0, 917999, '\p{^WB=extend}', "");
Expect(0, 917999, '\P{WB=extend}', "");
Expect(1, 917999, '\P{^WB=extend}', "");
Expect(0, 918000, '\p{WB=extend}', "");
Expect(1, 918000, '\p{^WB=extend}', "");
Expect(1, 918000, '\P{WB=extend}', "");
Expect(0, 918000, '\P{^WB=extend}', "");
Expect(1, 917999, '\p{WB:		Extend}', "");
Expect(0, 917999, '\p{^WB:		Extend}', "");
Expect(0, 917999, '\P{WB:		Extend}', "");
Expect(1, 917999, '\P{^WB:		Extend}', "");
Expect(0, 918000, '\p{WB:		Extend}', "");
Expect(1, 918000, '\p{^WB:		Extend}', "");
Expect(1, 918000, '\P{WB:		Extend}', "");
Expect(0, 918000, '\P{^WB:		Extend}', "");
Error('\p{Is_Word_Break=/a/EXTEND}');
Error('\P{Is_Word_Break=/a/EXTEND}');
Expect(1, 917999, '\p{Is_Word_Break=extend}', "");
Expect(0, 917999, '\p{^Is_Word_Break=extend}', "");
Expect(0, 917999, '\P{Is_Word_Break=extend}', "");
Expect(1, 917999, '\P{^Is_Word_Break=extend}', "");
Expect(0, 918000, '\p{Is_Word_Break=extend}', "");
Expect(1, 918000, '\p{^Is_Word_Break=extend}', "");
Expect(1, 918000, '\P{Is_Word_Break=extend}', "");
Expect(0, 918000, '\P{^Is_Word_Break=extend}', "");
Expect(1, 917999, '\p{Is_Word_Break=- extend}', "");
Expect(0, 917999, '\p{^Is_Word_Break=- extend}', "");
Expect(0, 917999, '\P{Is_Word_Break=- extend}', "");
Expect(1, 917999, '\P{^Is_Word_Break=- extend}', "");
Expect(0, 918000, '\p{Is_Word_Break=- extend}', "");
Expect(1, 918000, '\p{^Is_Word_Break=- extend}', "");
Expect(1, 918000, '\P{Is_Word_Break=- extend}', "");
Expect(0, 918000, '\P{^Is_Word_Break=- extend}', "");
Error('\p{Is_WB=:=_	Extend}');
Error('\P{Is_WB=:=_	Extend}');
Expect(1, 917999, '\p{Is_WB=extend}', "");
Expect(0, 917999, '\p{^Is_WB=extend}', "");
Expect(0, 917999, '\P{Is_WB=extend}', "");
Expect(1, 917999, '\P{^Is_WB=extend}', "");
Expect(0, 918000, '\p{Is_WB=extend}', "");
Expect(1, 918000, '\p{^Is_WB=extend}', "");
Expect(1, 918000, '\P{Is_WB=extend}', "");
Expect(0, 918000, '\P{^Is_WB=extend}', "");
Expect(1, 917999, '\p{Is_WB=- Extend}', "");
Expect(0, 917999, '\p{^Is_WB=- Extend}', "");
Expect(0, 917999, '\P{Is_WB=- Extend}', "");
Expect(1, 917999, '\P{^Is_WB=- Extend}', "");
Expect(0, 918000, '\p{Is_WB=- Extend}', "");
Expect(1, 918000, '\p{^Is_WB=- Extend}', "");
Expect(1, 918000, '\P{Is_WB=- Extend}', "");
Expect(0, 918000, '\P{^Is_WB=- Extend}', "");
Error('\p{Word_Break=-:=Format}');
Error('\P{Word_Break=-:=Format}');
Expect(1, 917631, '\p{Word_Break:	format}', "");
Expect(0, 917631, '\p{^Word_Break:	format}', "");
Expect(0, 917631, '\P{Word_Break:	format}', "");
Expect(1, 917631, '\P{^Word_Break:	format}', "");
Expect(0, 917632, '\p{Word_Break:	format}', "");
Expect(1, 917632, '\p{^Word_Break:	format}', "");
Expect(1, 917632, '\P{Word_Break:	format}', "");
Expect(0, 917632, '\P{^Word_Break:	format}', "");
Expect(1, 917631, '\p{Word_Break=_-Format}', "");
Expect(0, 917631, '\p{^Word_Break=_-Format}', "");
Expect(0, 917631, '\P{Word_Break=_-Format}', "");
Expect(1, 917631, '\P{^Word_Break=_-Format}', "");
Expect(0, 917632, '\p{Word_Break=_-Format}', "");
Expect(1, 917632, '\p{^Word_Break=_-Format}', "");
Expect(1, 917632, '\P{Word_Break=_-Format}', "");
Expect(0, 917632, '\P{^Word_Break=_-Format}', "");
Error('\p{WB= /a/fo}');
Error('\P{WB= /a/fo}');
Expect(1, 917631, '\p{WB=fo}', "");
Expect(0, 917631, '\p{^WB=fo}', "");
Expect(0, 917631, '\P{WB=fo}', "");
Expect(1, 917631, '\P{^WB=fo}', "");
Expect(0, 917632, '\p{WB=fo}', "");
Expect(1, 917632, '\p{^WB=fo}', "");
Expect(1, 917632, '\P{WB=fo}', "");
Expect(0, 917632, '\P{^WB=fo}', "");
Expect(1, 917631, '\p{WB=_FO}', "");
Expect(0, 917631, '\p{^WB=_FO}', "");
Expect(0, 917631, '\P{WB=_FO}', "");
Expect(1, 917631, '\P{^WB=_FO}', "");
Expect(0, 917632, '\p{WB=_FO}', "");
Expect(1, 917632, '\p{^WB=_FO}', "");
Expect(1, 917632, '\P{WB=_FO}', "");
Expect(0, 917632, '\P{^WB=_FO}', "");
Error('\p{Is_Word_Break=		Format:=}');
Error('\P{Is_Word_Break=		Format:=}');
Expect(1, 917631, '\p{Is_Word_Break=format}', "");
Expect(0, 917631, '\p{^Is_Word_Break=format}', "");
Expect(0, 917631, '\P{Is_Word_Break=format}', "");
Expect(1, 917631, '\P{^Is_Word_Break=format}', "");
Expect(0, 917632, '\p{Is_Word_Break=format}', "");
Expect(1, 917632, '\p{^Is_Word_Break=format}', "");
Expect(1, 917632, '\P{Is_Word_Break=format}', "");
Expect(0, 917632, '\P{^Is_Word_Break=format}', "");
Expect(1, 917631, '\p{Is_Word_Break=_FORMAT}', "");
Expect(0, 917631, '\p{^Is_Word_Break=_FORMAT}', "");
Expect(0, 917631, '\P{Is_Word_Break=_FORMAT}', "");
Expect(1, 917631, '\P{^Is_Word_Break=_FORMAT}', "");
Expect(0, 917632, '\p{Is_Word_Break=_FORMAT}', "");
Expect(1, 917632, '\p{^Is_Word_Break=_FORMAT}', "");
Expect(1, 917632, '\P{Is_Word_Break=_FORMAT}', "");
Expect(0, 917632, '\P{^Is_Word_Break=_FORMAT}', "");
Error('\p{Is_WB=-	FO/a/}');
Error('\P{Is_WB=-	FO/a/}');
Expect(1, 917631, '\p{Is_WB=fo}', "");
Expect(0, 917631, '\p{^Is_WB=fo}', "");
Expect(0, 917631, '\P{Is_WB=fo}', "");
Expect(1, 917631, '\P{^Is_WB=fo}', "");
Expect(0, 917632, '\p{Is_WB=fo}', "");
Expect(1, 917632, '\p{^Is_WB=fo}', "");
Expect(1, 917632, '\P{Is_WB=fo}', "");
Expect(0, 917632, '\P{^Is_WB=fo}', "");
Expect(1, 917631, '\p{Is_WB=_FO}', "");
Expect(0, 917631, '\p{^Is_WB=_FO}', "");
Expect(0, 917631, '\P{Is_WB=_FO}', "");
Expect(1, 917631, '\P{^Is_WB=_FO}', "");
Expect(0, 917632, '\p{Is_WB=_FO}', "");
Expect(1, 917632, '\p{^Is_WB=_FO}', "");
Expect(1, 917632, '\P{Is_WB=_FO}', "");
Expect(0, 917632, '\P{^Is_WB=_FO}', "");
Error('\p{Word_Break=/a/ -HEBREW_Letter}');
Error('\P{Word_Break=/a/ -HEBREW_Letter}');
Expect(1, 64335, '\p{Word_Break=hebrewletter}', "");
Expect(0, 64335, '\p{^Word_Break=hebrewletter}', "");
Expect(0, 64335, '\P{Word_Break=hebrewletter}', "");
Expect(1, 64335, '\P{^Word_Break=hebrewletter}', "");
Expect(0, 64336, '\p{Word_Break=hebrewletter}', "");
Expect(1, 64336, '\p{^Word_Break=hebrewletter}', "");
Expect(1, 64336, '\P{Word_Break=hebrewletter}', "");
Expect(0, 64336, '\P{^Word_Break=hebrewletter}', "");
Expect(1, 64335, '\p{Word_Break=		HEBREW_LETTER}', "");
Expect(0, 64335, '\p{^Word_Break=		HEBREW_LETTER}', "");
Expect(0, 64335, '\P{Word_Break=		HEBREW_LETTER}', "");
Expect(1, 64335, '\P{^Word_Break=		HEBREW_LETTER}', "");
Expect(0, 64336, '\p{Word_Break=		HEBREW_LETTER}', "");
Expect(1, 64336, '\p{^Word_Break=		HEBREW_LETTER}', "");
Expect(1, 64336, '\P{Word_Break=		HEBREW_LETTER}', "");
Expect(0, 64336, '\P{^Word_Break=		HEBREW_LETTER}', "");
Error('\p{WB=:=HL}');
Error('\P{WB=:=HL}');
Expect(1, 64335, '\p{WB=hl}', "");
Expect(0, 64335, '\p{^WB=hl}', "");
Expect(0, 64335, '\P{WB=hl}', "");
Expect(1, 64335, '\P{^WB=hl}', "");
Expect(0, 64336, '\p{WB=hl}', "");
Expect(1, 64336, '\p{^WB=hl}', "");
Expect(1, 64336, '\P{WB=hl}', "");
Expect(0, 64336, '\P{^WB=hl}', "");
Expect(1, 64335, '\p{WB=	-hl}', "");
Expect(0, 64335, '\p{^WB=	-hl}', "");
Expect(0, 64335, '\P{WB=	-hl}', "");
Expect(1, 64335, '\P{^WB=	-hl}', "");
Expect(0, 64336, '\p{WB=	-hl}', "");
Expect(1, 64336, '\p{^WB=	-hl}', "");
Expect(1, 64336, '\P{WB=	-hl}', "");
Expect(0, 64336, '\P{^WB=	-hl}', "");
Error('\p{Is_Word_Break= /a/Hebrew_Letter}');
Error('\P{Is_Word_Break= /a/Hebrew_Letter}');
Expect(1, 64335, '\p{Is_Word_Break=hebrewletter}', "");
Expect(0, 64335, '\p{^Is_Word_Break=hebrewletter}', "");
Expect(0, 64335, '\P{Is_Word_Break=hebrewletter}', "");
Expect(1, 64335, '\P{^Is_Word_Break=hebrewletter}', "");
Expect(0, 64336, '\p{Is_Word_Break=hebrewletter}', "");
Expect(1, 64336, '\p{^Is_Word_Break=hebrewletter}', "");
Expect(1, 64336, '\P{Is_Word_Break=hebrewletter}', "");
Expect(0, 64336, '\P{^Is_Word_Break=hebrewletter}', "");
Expect(1, 64335, '\p{Is_Word_Break=	HEBREW_Letter}', "");
Expect(0, 64335, '\p{^Is_Word_Break=	HEBREW_Letter}', "");
Expect(0, 64335, '\P{Is_Word_Break=	HEBREW_Letter}', "");
Expect(1, 64335, '\P{^Is_Word_Break=	HEBREW_Letter}', "");
Expect(0, 64336, '\p{Is_Word_Break=	HEBREW_Letter}', "");
Expect(1, 64336, '\p{^Is_Word_Break=	HEBREW_Letter}', "");
Expect(1, 64336, '\P{Is_Word_Break=	HEBREW_Letter}', "");
Expect(0, 64336, '\P{^Is_Word_Break=	HEBREW_Letter}', "");
Error('\p{Is_WB=--HL/a/}');
Error('\P{Is_WB=--HL/a/}');
Expect(1, 64335, '\p{Is_WB:   hl}', "");
Expect(0, 64335, '\p{^Is_WB:   hl}', "");
Expect(0, 64335, '\P{Is_WB:   hl}', "");
Expect(1, 64335, '\P{^Is_WB:   hl}', "");
Expect(0, 64336, '\p{Is_WB:   hl}', "");
Expect(1, 64336, '\p{^Is_WB:   hl}', "");
Expect(1, 64336, '\P{Is_WB:   hl}', "");
Expect(0, 64336, '\P{^Is_WB:   hl}', "");
Expect(1, 64335, '\p{Is_WB=- HL}', "");
Expect(0, 64335, '\p{^Is_WB=- HL}', "");
Expect(0, 64335, '\P{Is_WB=- HL}', "");
Expect(1, 64335, '\P{^Is_WB=- HL}', "");
Expect(0, 64336, '\p{Is_WB=- HL}', "");
Expect(1, 64336, '\p{^Is_WB=- HL}', "");
Expect(1, 64336, '\P{Is_WB=- HL}', "");
Expect(0, 64336, '\P{^Is_WB=- HL}', "");
Error('\p{Word_Break=/a/_-Katakana}');
Error('\P{Word_Break=/a/_-Katakana}');
Expect(1, 110592, '\p{Word_Break=katakana}', "");
Expect(0, 110592, '\p{^Word_Break=katakana}', "");
Expect(0, 110592, '\P{Word_Break=katakana}', "");
Expect(1, 110592, '\P{^Word_Break=katakana}', "");
Expect(0, 110593, '\p{Word_Break=katakana}', "");
Expect(1, 110593, '\p{^Word_Break=katakana}', "");
Expect(1, 110593, '\P{Word_Break=katakana}', "");
Expect(0, 110593, '\P{^Word_Break=katakana}', "");
Expect(1, 110592, '\p{Word_Break=	KATAKANA}', "");
Expect(0, 110592, '\p{^Word_Break=	KATAKANA}', "");
Expect(0, 110592, '\P{Word_Break=	KATAKANA}', "");
Expect(1, 110592, '\P{^Word_Break=	KATAKANA}', "");
Expect(0, 110593, '\p{Word_Break=	KATAKANA}', "");
Expect(1, 110593, '\p{^Word_Break=	KATAKANA}', "");
Expect(1, 110593, '\P{Word_Break=	KATAKANA}', "");
Expect(0, 110593, '\P{^Word_Break=	KATAKANA}', "");
Error('\p{WB=/a/	KA}');
Error('\P{WB=/a/	KA}');
Expect(1, 110592, '\p{WB=ka}', "");
Expect(0, 110592, '\p{^WB=ka}', "");
Expect(0, 110592, '\P{WB=ka}', "");
Expect(1, 110592, '\P{^WB=ka}', "");
Expect(0, 110593, '\p{WB=ka}', "");
Expect(1, 110593, '\p{^WB=ka}', "");
Expect(1, 110593, '\P{WB=ka}', "");
Expect(0, 110593, '\P{^WB=ka}', "");
Expect(1, 110592, '\p{WB= KA}', "");
Expect(0, 110592, '\p{^WB= KA}', "");
Expect(0, 110592, '\P{WB= KA}', "");
Expect(1, 110592, '\P{^WB= KA}', "");
Expect(0, 110593, '\p{WB= KA}', "");
Expect(1, 110593, '\p{^WB= KA}', "");
Expect(1, 110593, '\P{WB= KA}', "");
Expect(0, 110593, '\P{^WB= KA}', "");
Error('\p{Is_Word_Break=:= 	KATAKANA}');
Error('\P{Is_Word_Break=:= 	KATAKANA}');
Expect(1, 110592, '\p{Is_Word_Break=katakana}', "");
Expect(0, 110592, '\p{^Is_Word_Break=katakana}', "");
Expect(0, 110592, '\P{Is_Word_Break=katakana}', "");
Expect(1, 110592, '\P{^Is_Word_Break=katakana}', "");
Expect(0, 110593, '\p{Is_Word_Break=katakana}', "");
Expect(1, 110593, '\p{^Is_Word_Break=katakana}', "");
Expect(1, 110593, '\P{Is_Word_Break=katakana}', "");
Expect(0, 110593, '\P{^Is_Word_Break=katakana}', "");
Expect(1, 110592, '\p{Is_Word_Break=	-Katakana}', "");
Expect(0, 110592, '\p{^Is_Word_Break=	-Katakana}', "");
Expect(0, 110592, '\P{Is_Word_Break=	-Katakana}', "");
Expect(1, 110592, '\P{^Is_Word_Break=	-Katakana}', "");
Expect(0, 110593, '\p{Is_Word_Break=	-Katakana}', "");
Expect(1, 110593, '\p{^Is_Word_Break=	-Katakana}', "");
Expect(1, 110593, '\P{Is_Word_Break=	-Katakana}', "");
Expect(0, 110593, '\P{^Is_Word_Break=	-Katakana}', "");
Error('\p{Is_WB=:= 	ka}');
Error('\P{Is_WB=:= 	ka}');
Expect(1, 110592, '\p{Is_WB=ka}', "");
Expect(0, 110592, '\p{^Is_WB=ka}', "");
Expect(0, 110592, '\P{Is_WB=ka}', "");
Expect(1, 110592, '\P{^Is_WB=ka}', "");
Expect(0, 110593, '\p{Is_WB=ka}', "");
Expect(1, 110593, '\p{^Is_WB=ka}', "");
Expect(1, 110593, '\P{Is_WB=ka}', "");
Expect(0, 110593, '\P{^Is_WB=ka}', "");
Expect(1, 110592, '\p{Is_WB=- KA}', "");
Expect(0, 110592, '\p{^Is_WB=- KA}', "");
Expect(0, 110592, '\P{Is_WB=- KA}', "");
Expect(1, 110592, '\P{^Is_WB=- KA}', "");
Expect(0, 110593, '\p{Is_WB=- KA}', "");
Expect(1, 110593, '\p{^Is_WB=- KA}', "");
Expect(1, 110593, '\P{Is_WB=- KA}', "");
Expect(0, 110593, '\P{^Is_WB=- KA}', "");
Error('\p{Word_Break=ALetter:=}');
Error('\P{Word_Break=ALetter:=}');
Expect(1, 126651, '\p{Word_Break=aletter}', "");
Expect(0, 126651, '\p{^Word_Break=aletter}', "");
Expect(0, 126651, '\P{Word_Break=aletter}', "");
Expect(1, 126651, '\P{^Word_Break=aletter}', "");
Expect(0, 126652, '\p{Word_Break=aletter}', "");
Expect(1, 126652, '\p{^Word_Break=aletter}', "");
Expect(1, 126652, '\P{Word_Break=aletter}', "");
Expect(0, 126652, '\P{^Word_Break=aletter}', "");
Expect(1, 126651, '\p{Word_Break=	_aletter}', "");
Expect(0, 126651, '\p{^Word_Break=	_aletter}', "");
Expect(0, 126651, '\P{Word_Break=	_aletter}', "");
Expect(1, 126651, '\P{^Word_Break=	_aletter}', "");
Expect(0, 126652, '\p{Word_Break=	_aletter}', "");
Expect(1, 126652, '\p{^Word_Break=	_aletter}', "");
Expect(1, 126652, '\P{Word_Break=	_aletter}', "");
Expect(0, 126652, '\P{^Word_Break=	_aletter}', "");
Error('\p{WB= :=le}');
Error('\P{WB= :=le}');
Expect(1, 126651, '\p{WB=le}', "");
Expect(0, 126651, '\p{^WB=le}', "");
Expect(0, 126651, '\P{WB=le}', "");
Expect(1, 126651, '\P{^WB=le}', "");
Expect(0, 126652, '\p{WB=le}', "");
Expect(1, 126652, '\p{^WB=le}', "");
Expect(1, 126652, '\P{WB=le}', "");
Expect(0, 126652, '\P{^WB=le}', "");
Expect(1, 126651, '\p{WB=- LE}', "");
Expect(0, 126651, '\p{^WB=- LE}', "");
Expect(0, 126651, '\P{WB=- LE}', "");
Expect(1, 126651, '\P{^WB=- LE}', "");
Expect(0, 126652, '\p{WB=- LE}', "");
Expect(1, 126652, '\p{^WB=- LE}', "");
Expect(1, 126652, '\P{WB=- LE}', "");
Expect(0, 126652, '\P{^WB=- LE}', "");
Error('\p{Is_Word_Break=:=ALetter}');
Error('\P{Is_Word_Break=:=ALetter}');
Expect(1, 126651, '\p{Is_Word_Break=aletter}', "");
Expect(0, 126651, '\p{^Is_Word_Break=aletter}', "");
Expect(0, 126651, '\P{Is_Word_Break=aletter}', "");
Expect(1, 126651, '\P{^Is_Word_Break=aletter}', "");
Expect(0, 126652, '\p{Is_Word_Break=aletter}', "");
Expect(1, 126652, '\p{^Is_Word_Break=aletter}', "");
Expect(1, 126652, '\P{Is_Word_Break=aletter}', "");
Expect(0, 126652, '\P{^Is_Word_Break=aletter}', "");
Expect(1, 126651, '\p{Is_Word_Break:  -ALetter}', "");
Expect(0, 126651, '\p{^Is_Word_Break:  -ALetter}', "");
Expect(0, 126651, '\P{Is_Word_Break:  -ALetter}', "");
Expect(1, 126651, '\P{^Is_Word_Break:  -ALetter}', "");
Expect(0, 126652, '\p{Is_Word_Break:  -ALetter}', "");
Expect(1, 126652, '\p{^Is_Word_Break:  -ALetter}', "");
Expect(1, 126652, '\P{Is_Word_Break:  -ALetter}', "");
Expect(0, 126652, '\P{^Is_Word_Break:  -ALetter}', "");
Error('\p{Is_WB= LE/a/}');
Error('\P{Is_WB= LE/a/}');
Expect(1, 126651, '\p{Is_WB=le}', "");
Expect(0, 126651, '\p{^Is_WB=le}', "");
Expect(0, 126651, '\P{Is_WB=le}', "");
Expect(1, 126651, '\P{^Is_WB=le}', "");
Expect(0, 126652, '\p{Is_WB=le}', "");
Expect(1, 126652, '\p{^Is_WB=le}', "");
Expect(1, 126652, '\P{Is_WB=le}', "");
Expect(0, 126652, '\P{^Is_WB=le}', "");
Expect(1, 126651, '\p{Is_WB=_le}', "");
Expect(0, 126651, '\p{^Is_WB=_le}', "");
Expect(0, 126651, '\P{Is_WB=_le}', "");
Expect(1, 126651, '\P{^Is_WB=_le}', "");
Expect(0, 126652, '\p{Is_WB=_le}', "");
Expect(1, 126652, '\p{^Is_WB=_le}', "");
Expect(1, 126652, '\P{Is_WB=_le}', "");
Expect(0, 126652, '\P{^Is_WB=_le}', "");
Error('\p{Word_Break=/a/LF}');
Error('\P{Word_Break=/a/LF}');
Expect(1, 10, '\p{Word_Break: lf}', "");
Expect(0, 10, '\p{^Word_Break: lf}', "");
Expect(0, 10, '\P{Word_Break: lf}', "");
Expect(1, 10, '\P{^Word_Break: lf}', "");
Expect(0, 11, '\p{Word_Break: lf}', "");
Expect(1, 11, '\p{^Word_Break: lf}', "");
Expect(1, 11, '\P{Word_Break: lf}', "");
Expect(0, 11, '\P{^Word_Break: lf}', "");
Expect(1, 10, '\p{Word_Break=_LF}', "");
Expect(0, 10, '\p{^Word_Break=_LF}', "");
Expect(0, 10, '\P{Word_Break=_LF}', "");
Expect(1, 10, '\P{^Word_Break=_LF}', "");
Expect(0, 11, '\p{Word_Break=_LF}', "");
Expect(1, 11, '\p{^Word_Break=_LF}', "");
Expect(1, 11, '\P{Word_Break=_LF}', "");
Expect(0, 11, '\P{^Word_Break=_LF}', "");
Error('\p{WB=-LF:=}');
Error('\P{WB=-LF:=}');
Expect(1, 10, '\p{WB=lf}', "");
Expect(0, 10, '\p{^WB=lf}', "");
Expect(0, 10, '\P{WB=lf}', "");
Expect(1, 10, '\P{^WB=lf}', "");
Expect(0, 11, '\p{WB=lf}', "");
Expect(1, 11, '\p{^WB=lf}', "");
Expect(1, 11, '\P{WB=lf}', "");
Expect(0, 11, '\P{^WB=lf}', "");
Expect(1, 10, '\p{WB=-lf}', "");
Expect(0, 10, '\p{^WB=-lf}', "");
Expect(0, 10, '\P{WB=-lf}', "");
Expect(1, 10, '\P{^WB=-lf}', "");
Expect(0, 11, '\p{WB=-lf}', "");
Expect(1, 11, '\p{^WB=-lf}', "");
Expect(1, 11, '\P{WB=-lf}', "");
Expect(0, 11, '\P{^WB=-lf}', "");
Error('\p{Is_Word_Break=/a/--LF}');
Error('\P{Is_Word_Break=/a/--LF}');
Expect(1, 10, '\p{Is_Word_Break=lf}', "");
Expect(0, 10, '\p{^Is_Word_Break=lf}', "");
Expect(0, 10, '\P{Is_Word_Break=lf}', "");
Expect(1, 10, '\P{^Is_Word_Break=lf}', "");
Expect(0, 11, '\p{Is_Word_Break=lf}', "");
Expect(1, 11, '\p{^Is_Word_Break=lf}', "");
Expect(1, 11, '\P{Is_Word_Break=lf}', "");
Expect(0, 11, '\P{^Is_Word_Break=lf}', "");
Expect(1, 10, '\p{Is_Word_Break=_lf}', "");
Expect(0, 10, '\p{^Is_Word_Break=_lf}', "");
Expect(0, 10, '\P{Is_Word_Break=_lf}', "");
Expect(1, 10, '\P{^Is_Word_Break=_lf}', "");
Expect(0, 11, '\p{Is_Word_Break=_lf}', "");
Expect(1, 11, '\p{^Is_Word_Break=_lf}', "");
Expect(1, 11, '\P{Is_Word_Break=_lf}', "");
Expect(0, 11, '\P{^Is_Word_Break=_lf}', "");
Error('\p{Is_WB:   :=LF}');
Error('\P{Is_WB:   :=LF}');
Expect(1, 10, '\p{Is_WB:	lf}', "");
Expect(0, 10, '\p{^Is_WB:	lf}', "");
Expect(0, 10, '\P{Is_WB:	lf}', "");
Expect(1, 10, '\P{^Is_WB:	lf}', "");
Expect(0, 11, '\p{Is_WB:	lf}', "");
Expect(1, 11, '\p{^Is_WB:	lf}', "");
Expect(1, 11, '\P{Is_WB:	lf}', "");
Expect(0, 11, '\P{^Is_WB:	lf}', "");
Expect(1, 10, '\p{Is_WB= LF}', "");
Expect(0, 10, '\p{^Is_WB= LF}', "");
Expect(0, 10, '\P{Is_WB= LF}', "");
Expect(1, 10, '\P{^Is_WB= LF}', "");
Expect(0, 11, '\p{Is_WB= LF}', "");
Expect(1, 11, '\p{^Is_WB= LF}', "");
Expect(1, 11, '\P{Is_WB= LF}', "");
Expect(0, 11, '\P{^Is_WB= LF}', "");
Error('\p{Word_Break=	:=MIDNUMLET}');
Error('\P{Word_Break=	:=MIDNUMLET}');
Expect(1, 65294, '\p{Word_Break=midnumlet}', "");
Expect(0, 65294, '\p{^Word_Break=midnumlet}', "");
Expect(0, 65294, '\P{Word_Break=midnumlet}', "");
Expect(1, 65294, '\P{^Word_Break=midnumlet}', "");
Expect(0, 65295, '\p{Word_Break=midnumlet}', "");
Expect(1, 65295, '\p{^Word_Break=midnumlet}', "");
Expect(1, 65295, '\P{Word_Break=midnumlet}', "");
Expect(0, 65295, '\P{^Word_Break=midnumlet}', "");
Expect(1, 65294, '\p{Word_Break=-_MidNumLet}', "");
Expect(0, 65294, '\p{^Word_Break=-_MidNumLet}', "");
Expect(0, 65294, '\P{Word_Break=-_MidNumLet}', "");
Expect(1, 65294, '\P{^Word_Break=-_MidNumLet}', "");
Expect(0, 65295, '\p{Word_Break=-_MidNumLet}', "");
Expect(1, 65295, '\p{^Word_Break=-_MidNumLet}', "");
Expect(1, 65295, '\P{Word_Break=-_MidNumLet}', "");
Expect(0, 65295, '\P{^Word_Break=-_MidNumLet}', "");
Error('\p{WB=_/a/MB}');
Error('\P{WB=_/a/MB}');
Expect(1, 65294, '\p{WB=mb}', "");
Expect(0, 65294, '\p{^WB=mb}', "");
Expect(0, 65294, '\P{WB=mb}', "");
Expect(1, 65294, '\P{^WB=mb}', "");
Expect(0, 65295, '\p{WB=mb}', "");
Expect(1, 65295, '\p{^WB=mb}', "");
Expect(1, 65295, '\P{WB=mb}', "");
Expect(0, 65295, '\P{^WB=mb}', "");
Expect(1, 65294, '\p{WB=	_MB}', "");
Expect(0, 65294, '\p{^WB=	_MB}', "");
Expect(0, 65294, '\P{WB=	_MB}', "");
Expect(1, 65294, '\P{^WB=	_MB}', "");
Expect(0, 65295, '\p{WB=	_MB}', "");
Expect(1, 65295, '\p{^WB=	_MB}', "");
Expect(1, 65295, '\P{WB=	_MB}', "");
Expect(0, 65295, '\P{^WB=	_MB}', "");
Error('\p{Is_Word_Break=- MIDNUMLET/a/}');
Error('\P{Is_Word_Break=- MIDNUMLET/a/}');
Expect(1, 65294, '\p{Is_Word_Break=midnumlet}', "");
Expect(0, 65294, '\p{^Is_Word_Break=midnumlet}', "");
Expect(0, 65294, '\P{Is_Word_Break=midnumlet}', "");
Expect(1, 65294, '\P{^Is_Word_Break=midnumlet}', "");
Expect(0, 65295, '\p{Is_Word_Break=midnumlet}', "");
Expect(1, 65295, '\p{^Is_Word_Break=midnumlet}', "");
Expect(1, 65295, '\P{Is_Word_Break=midnumlet}', "");
Expect(0, 65295, '\P{^Is_Word_Break=midnumlet}', "");
Expect(1, 65294, '\p{Is_Word_Break=-_MidNumLet}', "");
Expect(0, 65294, '\p{^Is_Word_Break=-_MidNumLet}', "");
Expect(0, 65294, '\P{Is_Word_Break=-_MidNumLet}', "");
Expect(1, 65294, '\P{^Is_Word_Break=-_MidNumLet}', "");
Expect(0, 65295, '\p{Is_Word_Break=-_MidNumLet}', "");
Expect(1, 65295, '\p{^Is_Word_Break=-_MidNumLet}', "");
Expect(1, 65295, '\P{Is_Word_Break=-_MidNumLet}', "");
Expect(0, 65295, '\P{^Is_Word_Break=-_MidNumLet}', "");
Error('\p{Is_WB=MB/a/}');
Error('\P{Is_WB=MB/a/}');
Expect(1, 65294, '\p{Is_WB=mb}', "");
Expect(0, 65294, '\p{^Is_WB=mb}', "");
Expect(0, 65294, '\P{Is_WB=mb}', "");
Expect(1, 65294, '\P{^Is_WB=mb}', "");
Expect(0, 65295, '\p{Is_WB=mb}', "");
Expect(1, 65295, '\p{^Is_WB=mb}', "");
Expect(1, 65295, '\P{Is_WB=mb}', "");
Expect(0, 65295, '\P{^Is_WB=mb}', "");
Expect(1, 65294, '\p{Is_WB=MB}', "");
Expect(0, 65294, '\p{^Is_WB=MB}', "");
Expect(0, 65294, '\P{Is_WB=MB}', "");
Expect(1, 65294, '\P{^Is_WB=MB}', "");
Expect(0, 65295, '\p{Is_WB=MB}', "");
Expect(1, 65295, '\p{^Is_WB=MB}', "");
Expect(1, 65295, '\P{Is_WB=MB}', "");
Expect(0, 65295, '\P{^Is_WB=MB}', "");
Error('\p{Word_Break= 	midletter/a/}');
Error('\P{Word_Break= 	midletter/a/}');
Expect(1, 65306, '\p{Word_Break=midletter}', "");
Expect(0, 65306, '\p{^Word_Break=midletter}', "");
Expect(0, 65306, '\P{Word_Break=midletter}', "");
Expect(1, 65306, '\P{^Word_Break=midletter}', "");
Expect(0, 65307, '\p{Word_Break=midletter}', "");
Expect(1, 65307, '\p{^Word_Break=midletter}', "");
Expect(1, 65307, '\P{Word_Break=midletter}', "");
Expect(0, 65307, '\P{^Word_Break=midletter}', "");
Expect(1, 65306, '\p{Word_Break=-_midletter}', "");
Expect(0, 65306, '\p{^Word_Break=-_midletter}', "");
Expect(0, 65306, '\P{Word_Break=-_midletter}', "");
Expect(1, 65306, '\P{^Word_Break=-_midletter}', "");
Expect(0, 65307, '\p{Word_Break=-_midletter}', "");
Expect(1, 65307, '\p{^Word_Break=-_midletter}', "");
Expect(1, 65307, '\P{Word_Break=-_midletter}', "");
Expect(0, 65307, '\P{^Word_Break=-_midletter}', "");
Error('\p{WB=	/a/ML}');
Error('\P{WB=	/a/ML}');
Expect(1, 65306, '\p{WB=ml}', "");
Expect(0, 65306, '\p{^WB=ml}', "");
Expect(0, 65306, '\P{WB=ml}', "");
Expect(1, 65306, '\P{^WB=ml}', "");
Expect(0, 65307, '\p{WB=ml}', "");
Expect(1, 65307, '\p{^WB=ml}', "");
Expect(1, 65307, '\P{WB=ml}', "");
Expect(0, 65307, '\P{^WB=ml}', "");
Expect(1, 65306, '\p{WB= ml}', "");
Expect(0, 65306, '\p{^WB= ml}', "");
Expect(0, 65306, '\P{WB= ml}', "");
Expect(1, 65306, '\P{^WB= ml}', "");
Expect(0, 65307, '\p{WB= ml}', "");
Expect(1, 65307, '\p{^WB= ml}', "");
Expect(1, 65307, '\P{WB= ml}', "");
Expect(0, 65307, '\P{^WB= ml}', "");
Error('\p{Is_Word_Break=	/a/MidLetter}');
Error('\P{Is_Word_Break=	/a/MidLetter}');
Expect(1, 65306, '\p{Is_Word_Break=midletter}', "");
Expect(0, 65306, '\p{^Is_Word_Break=midletter}', "");
Expect(0, 65306, '\P{Is_Word_Break=midletter}', "");
Expect(1, 65306, '\P{^Is_Word_Break=midletter}', "");
Expect(0, 65307, '\p{Is_Word_Break=midletter}', "");
Expect(1, 65307, '\p{^Is_Word_Break=midletter}', "");
Expect(1, 65307, '\P{Is_Word_Break=midletter}', "");
Expect(0, 65307, '\P{^Is_Word_Break=midletter}', "");
Expect(1, 65306, '\p{Is_Word_Break=-	midletter}', "");
Expect(0, 65306, '\p{^Is_Word_Break=-	midletter}', "");
Expect(0, 65306, '\P{Is_Word_Break=-	midletter}', "");
Expect(1, 65306, '\P{^Is_Word_Break=-	midletter}', "");
Expect(0, 65307, '\p{Is_Word_Break=-	midletter}', "");
Expect(1, 65307, '\p{^Is_Word_Break=-	midletter}', "");
Expect(1, 65307, '\P{Is_Word_Break=-	midletter}', "");
Expect(0, 65307, '\P{^Is_Word_Break=-	midletter}', "");
Error('\p{Is_WB=- ML/a/}');
Error('\P{Is_WB=- ML/a/}');
Expect(1, 65306, '\p{Is_WB=ml}', "");
Expect(0, 65306, '\p{^Is_WB=ml}', "");
Expect(0, 65306, '\P{Is_WB=ml}', "");
Expect(1, 65306, '\P{^Is_WB=ml}', "");
Expect(0, 65307, '\p{Is_WB=ml}', "");
Expect(1, 65307, '\p{^Is_WB=ml}', "");
Expect(1, 65307, '\P{Is_WB=ml}', "");
Expect(0, 65307, '\P{^Is_WB=ml}', "");
Expect(1, 65306, '\p{Is_WB=_-ML}', "");
Expect(0, 65306, '\p{^Is_WB=_-ML}', "");
Expect(0, 65306, '\P{Is_WB=_-ML}', "");
Expect(1, 65306, '\P{^Is_WB=_-ML}', "");
Expect(0, 65307, '\p{Is_WB=_-ML}', "");
Expect(1, 65307, '\p{^Is_WB=_-ML}', "");
Expect(1, 65307, '\P{Is_WB=_-ML}', "");
Expect(0, 65307, '\P{^Is_WB=_-ML}', "");
Error('\p{Word_Break= 	MidNum/a/}');
Error('\P{Word_Break= 	MidNum/a/}');
Expect(1, 65307, '\p{Word_Break=midnum}', "");
Expect(0, 65307, '\p{^Word_Break=midnum}', "");
Expect(0, 65307, '\P{Word_Break=midnum}', "");
Expect(1, 65307, '\P{^Word_Break=midnum}', "");
Expect(0, 65308, '\p{Word_Break=midnum}', "");
Expect(1, 65308, '\p{^Word_Break=midnum}', "");
Expect(1, 65308, '\P{Word_Break=midnum}', "");
Expect(0, 65308, '\P{^Word_Break=midnum}', "");
Expect(1, 65307, '\p{Word_Break=_ MIDNUM}', "");
Expect(0, 65307, '\p{^Word_Break=_ MIDNUM}', "");
Expect(0, 65307, '\P{Word_Break=_ MIDNUM}', "");
Expect(1, 65307, '\P{^Word_Break=_ MIDNUM}', "");
Expect(0, 65308, '\p{Word_Break=_ MIDNUM}', "");
Expect(1, 65308, '\p{^Word_Break=_ MIDNUM}', "");
Expect(1, 65308, '\P{Word_Break=_ MIDNUM}', "");
Expect(0, 65308, '\P{^Word_Break=_ MIDNUM}', "");
Error('\p{WB=_	MN:=}');
Error('\P{WB=_	MN:=}');
Expect(1, 65307, '\p{WB=mn}', "");
Expect(0, 65307, '\p{^WB=mn}', "");
Expect(0, 65307, '\P{WB=mn}', "");
Expect(1, 65307, '\P{^WB=mn}', "");
Expect(0, 65308, '\p{WB=mn}', "");
Expect(1, 65308, '\p{^WB=mn}', "");
Expect(1, 65308, '\P{WB=mn}', "");
Expect(0, 65308, '\P{^WB=mn}', "");
Expect(1, 65307, '\p{WB: _MN}', "");
Expect(0, 65307, '\p{^WB: _MN}', "");
Expect(0, 65307, '\P{WB: _MN}', "");
Expect(1, 65307, '\P{^WB: _MN}', "");
Expect(0, 65308, '\p{WB: _MN}', "");
Expect(1, 65308, '\p{^WB: _MN}', "");
Expect(1, 65308, '\P{WB: _MN}', "");
Expect(0, 65308, '\P{^WB: _MN}', "");
Error('\p{Is_Word_Break=/a/midnum}');
Error('\P{Is_Word_Break=/a/midnum}');
Expect(1, 65307, '\p{Is_Word_Break=midnum}', "");
Expect(0, 65307, '\p{^Is_Word_Break=midnum}', "");
Expect(0, 65307, '\P{Is_Word_Break=midnum}', "");
Expect(1, 65307, '\P{^Is_Word_Break=midnum}', "");
Expect(0, 65308, '\p{Is_Word_Break=midnum}', "");
Expect(1, 65308, '\p{^Is_Word_Break=midnum}', "");
Expect(1, 65308, '\P{Is_Word_Break=midnum}', "");
Expect(0, 65308, '\P{^Is_Word_Break=midnum}', "");
Expect(1, 65307, '\p{Is_Word_Break=-	MidNum}', "");
Expect(0, 65307, '\p{^Is_Word_Break=-	MidNum}', "");
Expect(0, 65307, '\P{Is_Word_Break=-	MidNum}', "");
Expect(1, 65307, '\P{^Is_Word_Break=-	MidNum}', "");
Expect(0, 65308, '\p{Is_Word_Break=-	MidNum}', "");
Expect(1, 65308, '\p{^Is_Word_Break=-	MidNum}', "");
Expect(1, 65308, '\P{Is_Word_Break=-	MidNum}', "");
Expect(0, 65308, '\P{^Is_Word_Break=-	MidNum}', "");
Error('\p{Is_WB=:=-MN}');
Error('\P{Is_WB=:=-MN}');
Expect(1, 65307, '\p{Is_WB: mn}', "");
Expect(0, 65307, '\p{^Is_WB: mn}', "");
Expect(0, 65307, '\P{Is_WB: mn}', "");
Expect(1, 65307, '\P{^Is_WB: mn}', "");
Expect(0, 65308, '\p{Is_WB: mn}', "");
Expect(1, 65308, '\p{^Is_WB: mn}', "");
Expect(1, 65308, '\P{Is_WB: mn}', "");
Expect(0, 65308, '\P{^Is_WB: mn}', "");
Expect(1, 65307, '\p{Is_WB=	-MN}', "");
Expect(0, 65307, '\p{^Is_WB=	-MN}', "");
Expect(0, 65307, '\P{Is_WB=	-MN}', "");
Expect(1, 65307, '\P{^Is_WB=	-MN}', "");
Expect(0, 65308, '\p{Is_WB=	-MN}', "");
Expect(1, 65308, '\p{^Is_WB=	-MN}', "");
Expect(1, 65308, '\P{Is_WB=	-MN}', "");
Expect(0, 65308, '\P{^Is_WB=	-MN}', "");
Error('\p{Word_Break=	:=Newline}');
Error('\P{Word_Break=	:=Newline}');
Expect(1, 8233, '\p{Word_Break=newline}', "");
Expect(0, 8233, '\p{^Word_Break=newline}', "");
Expect(0, 8233, '\P{Word_Break=newline}', "");
Expect(1, 8233, '\P{^Word_Break=newline}', "");
Expect(0, 8234, '\p{Word_Break=newline}', "");
Expect(1, 8234, '\p{^Word_Break=newline}', "");
Expect(1, 8234, '\P{Word_Break=newline}', "");
Expect(0, 8234, '\P{^Word_Break=newline}', "");
Expect(1, 8233, '\p{Word_Break=-Newline}', "");
Expect(0, 8233, '\p{^Word_Break=-Newline}', "");
Expect(0, 8233, '\P{Word_Break=-Newline}', "");
Expect(1, 8233, '\P{^Word_Break=-Newline}', "");
Expect(0, 8234, '\p{Word_Break=-Newline}', "");
Expect(1, 8234, '\p{^Word_Break=-Newline}', "");
Expect(1, 8234, '\P{Word_Break=-Newline}', "");
Expect(0, 8234, '\P{^Word_Break=-Newline}', "");
Error('\p{WB:   /a/-_nl}');
Error('\P{WB:   /a/-_nl}');
Expect(1, 8233, '\p{WB=nl}', "");
Expect(0, 8233, '\p{^WB=nl}', "");
Expect(0, 8233, '\P{WB=nl}', "");
Expect(1, 8233, '\P{^WB=nl}', "");
Expect(0, 8234, '\p{WB=nl}', "");
Expect(1, 8234, '\p{^WB=nl}', "");
Expect(1, 8234, '\P{WB=nl}', "");
Expect(0, 8234, '\P{^WB=nl}', "");
Expect(1, 8233, '\p{WB=  NL}', "");
Expect(0, 8233, '\p{^WB=  NL}', "");
Expect(0, 8233, '\P{WB=  NL}', "");
Expect(1, 8233, '\P{^WB=  NL}', "");
Expect(0, 8234, '\p{WB=  NL}', "");
Expect(1, 8234, '\p{^WB=  NL}', "");
Expect(1, 8234, '\P{WB=  NL}', "");
Expect(0, 8234, '\P{^WB=  NL}', "");
Error('\p{Is_Word_Break=/a/Newline}');
Error('\P{Is_Word_Break=/a/Newline}');
Expect(1, 8233, '\p{Is_Word_Break=newline}', "");
Expect(0, 8233, '\p{^Is_Word_Break=newline}', "");
Expect(0, 8233, '\P{Is_Word_Break=newline}', "");
Expect(1, 8233, '\P{^Is_Word_Break=newline}', "");
Expect(0, 8234, '\p{Is_Word_Break=newline}', "");
Expect(1, 8234, '\p{^Is_Word_Break=newline}', "");
Expect(1, 8234, '\P{Is_Word_Break=newline}', "");
Expect(0, 8234, '\P{^Is_Word_Break=newline}', "");
Expect(1, 8233, '\p{Is_Word_Break= newline}', "");
Expect(0, 8233, '\p{^Is_Word_Break= newline}', "");
Expect(0, 8233, '\P{Is_Word_Break= newline}', "");
Expect(1, 8233, '\P{^Is_Word_Break= newline}', "");
Expect(0, 8234, '\p{Is_Word_Break= newline}', "");
Expect(1, 8234, '\p{^Is_Word_Break= newline}', "");
Expect(1, 8234, '\P{Is_Word_Break= newline}', "");
Expect(0, 8234, '\P{^Is_Word_Break= newline}', "");
Error('\p{Is_WB:	/a/-_NL}');
Error('\P{Is_WB:	/a/-_NL}');
Expect(1, 8233, '\p{Is_WB=nl}', "");
Expect(0, 8233, '\p{^Is_WB=nl}', "");
Expect(0, 8233, '\P{Is_WB=nl}', "");
Expect(1, 8233, '\P{^Is_WB=nl}', "");
Expect(0, 8234, '\p{Is_WB=nl}', "");
Expect(1, 8234, '\p{^Is_WB=nl}', "");
Expect(1, 8234, '\P{Is_WB=nl}', "");
Expect(0, 8234, '\P{^Is_WB=nl}', "");
Expect(1, 8233, '\p{Is_WB=  nl}', "");
Expect(0, 8233, '\p{^Is_WB=  nl}', "");
Expect(0, 8233, '\P{Is_WB=  nl}', "");
Expect(1, 8233, '\P{^Is_WB=  nl}', "");
Expect(0, 8234, '\p{Is_WB=  nl}', "");
Expect(1, 8234, '\p{^Is_WB=  nl}', "");
Expect(1, 8234, '\P{Is_WB=  nl}', "");
Expect(0, 8234, '\P{^Is_WB=  nl}', "");
Error('\p{Word_Break= /a/numeric}');
Error('\P{Word_Break= /a/numeric}');
Expect(1, 120831, '\p{Word_Break=numeric}', "");
Expect(0, 120831, '\p{^Word_Break=numeric}', "");
Expect(0, 120831, '\P{Word_Break=numeric}', "");
Expect(1, 120831, '\P{^Word_Break=numeric}', "");
Expect(0, 120832, '\p{Word_Break=numeric}', "");
Expect(1, 120832, '\p{^Word_Break=numeric}', "");
Expect(1, 120832, '\P{Word_Break=numeric}', "");
Expect(0, 120832, '\P{^Word_Break=numeric}', "");
Expect(1, 120831, '\p{Word_Break=	Numeric}', "");
Expect(0, 120831, '\p{^Word_Break=	Numeric}', "");
Expect(0, 120831, '\P{Word_Break=	Numeric}', "");
Expect(1, 120831, '\P{^Word_Break=	Numeric}', "");
Expect(0, 120832, '\p{Word_Break=	Numeric}', "");
Expect(1, 120832, '\p{^Word_Break=	Numeric}', "");
Expect(1, 120832, '\P{Word_Break=	Numeric}', "");
Expect(0, 120832, '\P{^Word_Break=	Numeric}', "");
Error('\p{WB=_:=NU}');
Error('\P{WB=_:=NU}');
Expect(1, 120831, '\p{WB=nu}', "");
Expect(0, 120831, '\p{^WB=nu}', "");
Expect(0, 120831, '\P{WB=nu}', "");
Expect(1, 120831, '\P{^WB=nu}', "");
Expect(0, 120832, '\p{WB=nu}', "");
Expect(1, 120832, '\p{^WB=nu}', "");
Expect(1, 120832, '\P{WB=nu}', "");
Expect(0, 120832, '\P{^WB=nu}', "");
Expect(1, 120831, '\p{WB=		NU}', "");
Expect(0, 120831, '\p{^WB=		NU}', "");
Expect(0, 120831, '\P{WB=		NU}', "");
Expect(1, 120831, '\P{^WB=		NU}', "");
Expect(0, 120832, '\p{WB=		NU}', "");
Expect(1, 120832, '\p{^WB=		NU}', "");
Expect(1, 120832, '\P{WB=		NU}', "");
Expect(0, 120832, '\P{^WB=		NU}', "");
Error('\p{Is_Word_Break=/a/ NUMERIC}');
Error('\P{Is_Word_Break=/a/ NUMERIC}');
Expect(1, 120831, '\p{Is_Word_Break=numeric}', "");
Expect(0, 120831, '\p{^Is_Word_Break=numeric}', "");
Expect(0, 120831, '\P{Is_Word_Break=numeric}', "");
Expect(1, 120831, '\P{^Is_Word_Break=numeric}', "");
Expect(0, 120832, '\p{Is_Word_Break=numeric}', "");
Expect(1, 120832, '\p{^Is_Word_Break=numeric}', "");
Expect(1, 120832, '\P{Is_Word_Break=numeric}', "");
Expect(0, 120832, '\P{^Is_Word_Break=numeric}', "");
Expect(1, 120831, '\p{Is_Word_Break=NUMERIC}', "");
Expect(0, 120831, '\p{^Is_Word_Break=NUMERIC}', "");
Expect(0, 120831, '\P{Is_Word_Break=NUMERIC}', "");
Expect(1, 120831, '\P{^Is_Word_Break=NUMERIC}', "");
Expect(0, 120832, '\p{Is_Word_Break=NUMERIC}', "");
Expect(1, 120832, '\p{^Is_Word_Break=NUMERIC}', "");
Expect(1, 120832, '\P{Is_Word_Break=NUMERIC}', "");
Expect(0, 120832, '\P{^Is_Word_Break=NUMERIC}', "");
Error('\p{Is_WB=:=- NU}');
Error('\P{Is_WB=:=- NU}');
Expect(1, 120831, '\p{Is_WB=nu}', "");
Expect(0, 120831, '\p{^Is_WB=nu}', "");
Expect(0, 120831, '\P{Is_WB=nu}', "");
Expect(1, 120831, '\P{^Is_WB=nu}', "");
Expect(0, 120832, '\p{Is_WB=nu}', "");
Expect(1, 120832, '\p{^Is_WB=nu}', "");
Expect(1, 120832, '\P{Is_WB=nu}', "");
Expect(0, 120832, '\P{^Is_WB=nu}', "");
Expect(1, 120831, '\p{Is_WB=		NU}', "");
Expect(0, 120831, '\p{^Is_WB=		NU}', "");
Expect(0, 120831, '\P{Is_WB=		NU}', "");
Expect(1, 120831, '\P{^Is_WB=		NU}', "");
Expect(0, 120832, '\p{Is_WB=		NU}', "");
Expect(1, 120832, '\p{^Is_WB=		NU}', "");
Expect(1, 120832, '\P{Is_WB=		NU}', "");
Expect(0, 120832, '\P{^Is_WB=		NU}', "");
Error('\p{Word_Break=:=		REGIONAL_indicator}');
Error('\P{Word_Break=:=		REGIONAL_indicator}');
Expect(1, 127487, '\p{Word_Break=regionalindicator}', "");
Expect(0, 127487, '\p{^Word_Break=regionalindicator}', "");
Expect(0, 127487, '\P{Word_Break=regionalindicator}', "");
Expect(1, 127487, '\P{^Word_Break=regionalindicator}', "");
Expect(0, 127488, '\p{Word_Break=regionalindicator}', "");
Expect(1, 127488, '\p{^Word_Break=regionalindicator}', "");
Expect(1, 127488, '\P{Word_Break=regionalindicator}', "");
Expect(0, 127488, '\P{^Word_Break=regionalindicator}', "");
Expect(1, 127487, '\p{Word_Break= _Regional_Indicator}', "");
Expect(0, 127487, '\p{^Word_Break= _Regional_Indicator}', "");
Expect(0, 127487, '\P{Word_Break= _Regional_Indicator}', "");
Expect(1, 127487, '\P{^Word_Break= _Regional_Indicator}', "");
Expect(0, 127488, '\p{Word_Break= _Regional_Indicator}', "");
Expect(1, 127488, '\p{^Word_Break= _Regional_Indicator}', "");
Expect(1, 127488, '\P{Word_Break= _Regional_Indicator}', "");
Expect(0, 127488, '\P{^Word_Break= _Regional_Indicator}', "");
Error('\p{WB=- RI/a/}');
Error('\P{WB=- RI/a/}');
Expect(1, 127487, '\p{WB=ri}', "");
Expect(0, 127487, '\p{^WB=ri}', "");
Expect(0, 127487, '\P{WB=ri}', "");
Expect(1, 127487, '\P{^WB=ri}', "");
Expect(0, 127488, '\p{WB=ri}', "");
Expect(1, 127488, '\p{^WB=ri}', "");
Expect(1, 127488, '\P{WB=ri}', "");
Expect(0, 127488, '\P{^WB=ri}', "");
Expect(1, 127487, '\p{WB=	 RI}', "");
Expect(0, 127487, '\p{^WB=	 RI}', "");
Expect(0, 127487, '\P{WB=	 RI}', "");
Expect(1, 127487, '\P{^WB=	 RI}', "");
Expect(0, 127488, '\p{WB=	 RI}', "");
Expect(1, 127488, '\p{^WB=	 RI}', "");
Expect(1, 127488, '\P{WB=	 RI}', "");
Expect(0, 127488, '\P{^WB=	 RI}', "");
Error('\p{Is_Word_Break=-Regional_Indicator/a/}');
Error('\P{Is_Word_Break=-Regional_Indicator/a/}');
Expect(1, 127487, '\p{Is_Word_Break=regionalindicator}', "");
Expect(0, 127487, '\p{^Is_Word_Break=regionalindicator}', "");
Expect(0, 127487, '\P{Is_Word_Break=regionalindicator}', "");
Expect(1, 127487, '\P{^Is_Word_Break=regionalindicator}', "");
Expect(0, 127488, '\p{Is_Word_Break=regionalindicator}', "");
Expect(1, 127488, '\p{^Is_Word_Break=regionalindicator}', "");
Expect(1, 127488, '\P{Is_Word_Break=regionalindicator}', "");
Expect(0, 127488, '\P{^Is_Word_Break=regionalindicator}', "");
Expect(1, 127487, '\p{Is_Word_Break=-_regional_indicator}', "");
Expect(0, 127487, '\p{^Is_Word_Break=-_regional_indicator}', "");
Expect(0, 127487, '\P{Is_Word_Break=-_regional_indicator}', "");
Expect(1, 127487, '\P{^Is_Word_Break=-_regional_indicator}', "");
Expect(0, 127488, '\p{Is_Word_Break=-_regional_indicator}', "");
Expect(1, 127488, '\p{^Is_Word_Break=-_regional_indicator}', "");
Expect(1, 127488, '\P{Is_Word_Break=-_regional_indicator}', "");
Expect(0, 127488, '\P{^Is_Word_Break=-_regional_indicator}', "");
Error('\p{Is_WB=RI/a/}');
Error('\P{Is_WB=RI/a/}');
Expect(1, 127487, '\p{Is_WB=ri}', "");
Expect(0, 127487, '\p{^Is_WB=ri}', "");
Expect(0, 127487, '\P{Is_WB=ri}', "");
Expect(1, 127487, '\P{^Is_WB=ri}', "");
Expect(0, 127488, '\p{Is_WB=ri}', "");
Expect(1, 127488, '\p{^Is_WB=ri}', "");
Expect(1, 127488, '\P{Is_WB=ri}', "");
Expect(0, 127488, '\P{^Is_WB=ri}', "");
Expect(1, 127487, '\p{Is_WB=_ RI}', "");
Expect(0, 127487, '\p{^Is_WB=_ RI}', "");
Expect(0, 127487, '\P{Is_WB=_ RI}', "");
Expect(1, 127487, '\P{^Is_WB=_ RI}', "");
Expect(0, 127488, '\p{Is_WB=_ RI}', "");
Expect(1, 127488, '\p{^Is_WB=_ RI}', "");
Expect(1, 127488, '\P{Is_WB=_ RI}', "");
Expect(0, 127488, '\P{^Is_WB=_ RI}', "");
Error('\p{Word_Break=		Single_Quote/a/}');
Error('\P{Word_Break=		Single_Quote/a/}');
Expect(1, 39, '\p{Word_Break=singlequote}', "");
Expect(0, 39, '\p{^Word_Break=singlequote}', "");
Expect(0, 39, '\P{Word_Break=singlequote}', "");
Expect(1, 39, '\P{^Word_Break=singlequote}', "");
Expect(0, 40, '\p{Word_Break=singlequote}', "");
Expect(1, 40, '\p{^Word_Break=singlequote}', "");
Expect(1, 40, '\P{Word_Break=singlequote}', "");
Expect(0, 40, '\P{^Word_Break=singlequote}', "");
Expect(1, 39, '\p{Word_Break=__single_QUOTE}', "");
Expect(0, 39, '\p{^Word_Break=__single_QUOTE}', "");
Expect(0, 39, '\P{Word_Break=__single_QUOTE}', "");
Expect(1, 39, '\P{^Word_Break=__single_QUOTE}', "");
Expect(0, 40, '\p{Word_Break=__single_QUOTE}', "");
Expect(1, 40, '\p{^Word_Break=__single_QUOTE}', "");
Expect(1, 40, '\P{Word_Break=__single_QUOTE}', "");
Expect(0, 40, '\P{^Word_Break=__single_QUOTE}', "");
Error('\p{WB=_SQ:=}');
Error('\P{WB=_SQ:=}');
Expect(1, 39, '\p{WB=sq}', "");
Expect(0, 39, '\p{^WB=sq}', "");
Expect(0, 39, '\P{WB=sq}', "");
Expect(1, 39, '\P{^WB=sq}', "");
Expect(0, 40, '\p{WB=sq}', "");
Expect(1, 40, '\p{^WB=sq}', "");
Expect(1, 40, '\P{WB=sq}', "");
Expect(0, 40, '\P{^WB=sq}', "");
Expect(1, 39, '\p{WB: 	SQ}', "");
Expect(0, 39, '\p{^WB: 	SQ}', "");
Expect(0, 39, '\P{WB: 	SQ}', "");
Expect(1, 39, '\P{^WB: 	SQ}', "");
Expect(0, 40, '\p{WB: 	SQ}', "");
Expect(1, 40, '\p{^WB: 	SQ}', "");
Expect(1, 40, '\P{WB: 	SQ}', "");
Expect(0, 40, '\P{^WB: 	SQ}', "");
Error('\p{Is_Word_Break=-:=Single_Quote}');
Error('\P{Is_Word_Break=-:=Single_Quote}');
Expect(1, 39, '\p{Is_Word_Break=singlequote}', "");
Expect(0, 39, '\p{^Is_Word_Break=singlequote}', "");
Expect(0, 39, '\P{Is_Word_Break=singlequote}', "");
Expect(1, 39, '\P{^Is_Word_Break=singlequote}', "");
Expect(0, 40, '\p{Is_Word_Break=singlequote}', "");
Expect(1, 40, '\p{^Is_Word_Break=singlequote}', "");
Expect(1, 40, '\P{Is_Word_Break=singlequote}', "");
Expect(0, 40, '\P{^Is_Word_Break=singlequote}', "");
Expect(1, 39, '\p{Is_Word_Break=_ Single_Quote}', "");
Expect(0, 39, '\p{^Is_Word_Break=_ Single_Quote}', "");
Expect(0, 39, '\P{Is_Word_Break=_ Single_Quote}', "");
Expect(1, 39, '\P{^Is_Word_Break=_ Single_Quote}', "");
Expect(0, 40, '\p{Is_Word_Break=_ Single_Quote}', "");
Expect(1, 40, '\p{^Is_Word_Break=_ Single_Quote}', "");
Expect(1, 40, '\P{Is_Word_Break=_ Single_Quote}', "");
Expect(0, 40, '\P{^Is_Word_Break=_ Single_Quote}', "");
Error('\p{Is_WB= :=SQ}');
Error('\P{Is_WB= :=SQ}');
Expect(1, 39, '\p{Is_WB:	sq}', "");
Expect(0, 39, '\p{^Is_WB:	sq}', "");
Expect(0, 39, '\P{Is_WB:	sq}', "");
Expect(1, 39, '\P{^Is_WB:	sq}', "");
Expect(0, 40, '\p{Is_WB:	sq}', "");
Expect(1, 40, '\p{^Is_WB:	sq}', "");
Expect(1, 40, '\P{Is_WB:	sq}', "");
Expect(0, 40, '\P{^Is_WB:	sq}', "");
Expect(1, 39, '\p{Is_WB= -SQ}', "");
Expect(0, 39, '\p{^Is_WB= -SQ}', "");
Expect(0, 39, '\P{Is_WB= -SQ}', "");
Expect(1, 39, '\P{^Is_WB= -SQ}', "");
Expect(0, 40, '\p{Is_WB= -SQ}', "");
Expect(1, 40, '\p{^Is_WB= -SQ}', "");
Expect(1, 40, '\P{Is_WB= -SQ}', "");
Expect(0, 40, '\P{^Is_WB= -SQ}', "");
Error('\p{Word_Break=__Other/a/}');
Error('\P{Word_Break=__Other/a/}');
Expect(1, 918000, '\p{Word_Break=other}', "");
Expect(0, 918000, '\p{^Word_Break=other}', "");
Expect(0, 918000, '\P{Word_Break=other}', "");
Expect(1, 918000, '\P{^Word_Break=other}', "");
Expect(0, 917999, '\p{Word_Break=other}', "");
Expect(1, 917999, '\p{^Word_Break=other}', "");
Expect(1, 917999, '\P{Word_Break=other}', "");
Expect(0, 917999, '\P{^Word_Break=other}', "");
Expect(1, 918000, '\p{Word_Break= Other}', "");
Expect(0, 918000, '\p{^Word_Break= Other}', "");
Expect(0, 918000, '\P{Word_Break= Other}', "");
Expect(1, 918000, '\P{^Word_Break= Other}', "");
Expect(0, 917999, '\p{Word_Break= Other}', "");
Expect(1, 917999, '\p{^Word_Break= Other}', "");
Expect(1, 917999, '\P{Word_Break= Other}', "");
Expect(0, 917999, '\P{^Word_Break= Other}', "");
Error('\p{WB=/a/	XX}');
Error('\P{WB=/a/	XX}');
Expect(1, 918000, '\p{WB:	xx}', "");
Expect(0, 918000, '\p{^WB:	xx}', "");
Expect(0, 918000, '\P{WB:	xx}', "");
Expect(1, 918000, '\P{^WB:	xx}', "");
Expect(0, 917999, '\p{WB:	xx}', "");
Expect(1, 917999, '\p{^WB:	xx}', "");
Expect(1, 917999, '\P{WB:	xx}', "");
Expect(0, 917999, '\P{^WB:	xx}', "");
Expect(1, 918000, '\p{WB=	_XX}', "");
Expect(0, 918000, '\p{^WB=	_XX}', "");
Expect(0, 918000, '\P{WB=	_XX}', "");
Expect(1, 918000, '\P{^WB=	_XX}', "");
Expect(0, 917999, '\p{WB=	_XX}', "");
Expect(1, 917999, '\p{^WB=	_XX}', "");
Expect(1, 917999, '\P{WB=	_XX}', "");
Expect(0, 917999, '\P{^WB=	_XX}', "");
Error('\p{Is_Word_Break:    :=Other}');
Error('\P{Is_Word_Break:    :=Other}');
Expect(1, 918000, '\p{Is_Word_Break=other}', "");
Expect(0, 918000, '\p{^Is_Word_Break=other}', "");
Expect(0, 918000, '\P{Is_Word_Break=other}', "");
Expect(1, 918000, '\P{^Is_Word_Break=other}', "");
Expect(0, 917999, '\p{Is_Word_Break=other}', "");
Expect(1, 917999, '\p{^Is_Word_Break=other}', "");
Expect(1, 917999, '\P{Is_Word_Break=other}', "");
Expect(0, 917999, '\P{^Is_Word_Break=other}', "");
Expect(1, 918000, '\p{Is_Word_Break=-_Other}', "");
Expect(0, 918000, '\p{^Is_Word_Break=-_Other}', "");
Expect(0, 918000, '\P{Is_Word_Break=-_Other}', "");
Expect(1, 918000, '\P{^Is_Word_Break=-_Other}', "");
Expect(0, 917999, '\p{Is_Word_Break=-_Other}', "");
Expect(1, 917999, '\p{^Is_Word_Break=-_Other}', "");
Expect(1, 917999, '\P{Is_Word_Break=-_Other}', "");
Expect(0, 917999, '\P{^Is_Word_Break=-_Other}', "");
Error('\p{Is_WB: 	:=xx}');
Error('\P{Is_WB: 	:=xx}');
Expect(1, 918000, '\p{Is_WB:   xx}', "");
Expect(0, 918000, '\p{^Is_WB:   xx}', "");
Expect(0, 918000, '\P{Is_WB:   xx}', "");
Expect(1, 918000, '\P{^Is_WB:   xx}', "");
Expect(0, 917999, '\p{Is_WB:   xx}', "");
Expect(1, 917999, '\p{^Is_WB:   xx}', "");
Expect(1, 917999, '\P{Is_WB:   xx}', "");
Expect(0, 917999, '\P{^Is_WB:   xx}', "");
Expect(1, 918000, '\p{Is_WB=	XX}', "");
Expect(0, 918000, '\p{^Is_WB=	XX}', "");
Expect(0, 918000, '\P{Is_WB=	XX}', "");
Expect(1, 918000, '\P{^Is_WB=	XX}', "");
Expect(0, 917999, '\p{Is_WB=	XX}', "");
Expect(1, 917999, '\p{^Is_WB=	XX}', "");
Expect(1, 917999, '\P{Is_WB=	XX}', "");
Expect(0, 917999, '\P{^Is_WB=	XX}', "");
Error('\p{White_Space=:=no}');
Error('\P{White_Space=:=no}');
Expect(1, 12289, '\p{White_Space=no}', "");
Expect(0, 12289, '\p{^White_Space=no}', "");
Expect(0, 12289, '\P{White_Space=no}', "");
Expect(1, 12289, '\P{^White_Space=no}', "");
Expect(0, 12288, '\p{White_Space=no}', "");
Expect(1, 12288, '\p{^White_Space=no}', "");
Expect(1, 12288, '\P{White_Space=no}', "");
Expect(0, 12288, '\P{^White_Space=no}', "");
Expect(1, 12289, '\p{White_Space= no}', "");
Expect(0, 12289, '\p{^White_Space= no}', "");
Expect(0, 12289, '\P{White_Space= no}', "");
Expect(1, 12289, '\P{^White_Space= no}', "");
Expect(0, 12288, '\p{White_Space= no}', "");
Expect(1, 12288, '\p{^White_Space= no}', "");
Expect(1, 12288, '\P{White_Space= no}', "");
Expect(0, 12288, '\P{^White_Space= no}', "");
Error('\p{WSpace=_/a/N}');
Error('\P{WSpace=_/a/N}');
Expect(1, 12289, '\p{WSpace=n}', "");
Expect(0, 12289, '\p{^WSpace=n}', "");
Expect(0, 12289, '\P{WSpace=n}', "");
Expect(1, 12289, '\P{^WSpace=n}', "");
Expect(0, 12288, '\p{WSpace=n}', "");
Expect(1, 12288, '\p{^WSpace=n}', "");
Expect(1, 12288, '\P{WSpace=n}', "");
Expect(0, 12288, '\P{^WSpace=n}', "");
Expect(1, 12289, '\p{WSpace: 	_N}', "");
Expect(0, 12289, '\p{^WSpace: 	_N}', "");
Expect(0, 12289, '\P{WSpace: 	_N}', "");
Expect(1, 12289, '\P{^WSpace: 	_N}', "");
Expect(0, 12288, '\p{WSpace: 	_N}', "");
Expect(1, 12288, '\p{^WSpace: 	_N}', "");
Expect(1, 12288, '\P{WSpace: 	_N}', "");
Expect(0, 12288, '\P{^WSpace: 	_N}', "");
Error('\p{Space= -F/a/}');
Error('\P{Space= -F/a/}');
Expect(1, 12289, '\p{Space=f}', "");
Expect(0, 12289, '\p{^Space=f}', "");
Expect(0, 12289, '\P{Space=f}', "");
Expect(1, 12289, '\P{^Space=f}', "");
Expect(0, 12288, '\p{Space=f}', "");
Expect(1, 12288, '\p{^Space=f}', "");
Expect(1, 12288, '\P{Space=f}', "");
Expect(0, 12288, '\P{^Space=f}', "");
Expect(1, 12289, '\p{Space=--f}', "");
Expect(0, 12289, '\p{^Space=--f}', "");
Expect(0, 12289, '\P{Space=--f}', "");
Expect(1, 12289, '\P{^Space=--f}', "");
Expect(0, 12288, '\p{Space=--f}', "");
Expect(1, 12288, '\p{^Space=--f}', "");
Expect(1, 12288, '\P{Space=--f}', "");
Expect(0, 12288, '\P{^Space=--f}', "");
Error('\p{Is_White_Space=-:=False}');
Error('\P{Is_White_Space=-:=False}');
Expect(1, 12289, '\p{Is_White_Space=false}', "");
Expect(0, 12289, '\p{^Is_White_Space=false}', "");
Expect(0, 12289, '\P{Is_White_Space=false}', "");
Expect(1, 12289, '\P{^Is_White_Space=false}', "");
Expect(0, 12288, '\p{Is_White_Space=false}', "");
Expect(1, 12288, '\p{^Is_White_Space=false}', "");
Expect(1, 12288, '\P{Is_White_Space=false}', "");
Expect(0, 12288, '\P{^Is_White_Space=false}', "");
Expect(1, 12289, '\p{Is_White_Space=	-false}', "");
Expect(0, 12289, '\p{^Is_White_Space=	-false}', "");
Expect(0, 12289, '\P{Is_White_Space=	-false}', "");
Expect(1, 12289, '\P{^Is_White_Space=	-false}', "");
Expect(0, 12288, '\p{Is_White_Space=	-false}', "");
Expect(1, 12288, '\p{^Is_White_Space=	-false}', "");
Expect(1, 12288, '\P{Is_White_Space=	-false}', "");
Expect(0, 12288, '\P{^Is_White_Space=	-false}', "");
Error('\p{Is_WSpace= /a/NO}');
Error('\P{Is_WSpace= /a/NO}');
Expect(1, 12289, '\p{Is_WSpace:   no}', "");
Expect(0, 12289, '\p{^Is_WSpace:   no}', "");
Expect(0, 12289, '\P{Is_WSpace:   no}', "");
Expect(1, 12289, '\P{^Is_WSpace:   no}', "");
Expect(0, 12288, '\p{Is_WSpace:   no}', "");
Expect(1, 12288, '\p{^Is_WSpace:   no}', "");
Expect(1, 12288, '\P{Is_WSpace:   no}', "");
Expect(0, 12288, '\P{^Is_WSpace:   no}', "");
Expect(1, 12289, '\p{Is_WSpace=	No}', "");
Expect(0, 12289, '\p{^Is_WSpace=	No}', "");
Expect(0, 12289, '\P{Is_WSpace=	No}', "");
Expect(1, 12289, '\P{^Is_WSpace=	No}', "");
Expect(0, 12288, '\p{Is_WSpace=	No}', "");
Expect(1, 12288, '\p{^Is_WSpace=	No}', "");
Expect(1, 12288, '\P{Is_WSpace=	No}', "");
Expect(0, 12288, '\P{^Is_WSpace=	No}', "");
Error('\p{Is_Space=/a/N}');
Error('\P{Is_Space=/a/N}');
Expect(1, 12289, '\p{Is_Space=n}', "");
Expect(0, 12289, '\p{^Is_Space=n}', "");
Expect(0, 12289, '\P{Is_Space=n}', "");
Expect(1, 12289, '\P{^Is_Space=n}', "");
Expect(0, 12288, '\p{Is_Space=n}', "");
Expect(1, 12288, '\p{^Is_Space=n}', "");
Expect(1, 12288, '\P{Is_Space=n}', "");
Expect(0, 12288, '\P{^Is_Space=n}', "");
Expect(1, 12289, '\p{Is_Space=	_n}', "");
Expect(0, 12289, '\p{^Is_Space=	_n}', "");
Expect(0, 12289, '\P{Is_Space=	_n}', "");
Expect(1, 12289, '\P{^Is_Space=	_n}', "");
Expect(0, 12288, '\p{Is_Space=	_n}', "");
Expect(1, 12288, '\p{^Is_Space=	_n}', "");
Expect(1, 12288, '\P{Is_Space=	_n}', "");
Expect(0, 12288, '\P{^Is_Space=	_n}', "");
Error('\p{White_Space= :=Yes}');
Error('\P{White_Space= :=Yes}');
Expect(1, 12288, '\p{White_Space=yes}', "");
Expect(0, 12288, '\p{^White_Space=yes}', "");
Expect(0, 12288, '\P{White_Space=yes}', "");
Expect(1, 12288, '\P{^White_Space=yes}', "");
Expect(0, 12289, '\p{White_Space=yes}', "");
Expect(1, 12289, '\p{^White_Space=yes}', "");
Expect(1, 12289, '\P{White_Space=yes}', "");
Expect(0, 12289, '\P{^White_Space=yes}', "");
Expect(1, 12288, '\p{White_Space:   _yes}', "");
Expect(0, 12288, '\p{^White_Space:   _yes}', "");
Expect(0, 12288, '\P{White_Space:   _yes}', "");
Expect(1, 12288, '\P{^White_Space:   _yes}', "");
Expect(0, 12289, '\p{White_Space:   _yes}', "");
Expect(1, 12289, '\p{^White_Space:   _yes}', "");
Expect(1, 12289, '\P{White_Space:   _yes}', "");
Expect(0, 12289, '\P{^White_Space:   _yes}', "");
Error('\p{WSpace:   _ Y/a/}');
Error('\P{WSpace:   _ Y/a/}');
Expect(1, 12288, '\p{WSpace=y}', "");
Expect(0, 12288, '\p{^WSpace=y}', "");
Expect(0, 12288, '\P{WSpace=y}', "");
Expect(1, 12288, '\P{^WSpace=y}', "");
Expect(0, 12289, '\p{WSpace=y}', "");
Expect(1, 12289, '\p{^WSpace=y}', "");
Expect(1, 12289, '\P{WSpace=y}', "");
Expect(0, 12289, '\P{^WSpace=y}', "");
Expect(1, 12288, '\p{WSpace= -Y}', "");
Expect(0, 12288, '\p{^WSpace= -Y}', "");
Expect(0, 12288, '\P{WSpace= -Y}', "");
Expect(1, 12288, '\P{^WSpace= -Y}', "");
Expect(0, 12289, '\p{WSpace= -Y}', "");
Expect(1, 12289, '\p{^WSpace= -Y}', "");
Expect(1, 12289, '\P{WSpace= -Y}', "");
Expect(0, 12289, '\P{^WSpace= -Y}', "");
Error('\p{Space: 	_T:=}');
Error('\P{Space: 	_T:=}');
Expect(1, 12288, '\p{Space=t}', "");
Expect(0, 12288, '\p{^Space=t}', "");
Expect(0, 12288, '\P{Space=t}', "");
Expect(1, 12288, '\P{^Space=t}', "");
Expect(0, 12289, '\p{Space=t}', "");
Expect(1, 12289, '\p{^Space=t}', "");
Expect(1, 12289, '\P{Space=t}', "");
Expect(0, 12289, '\P{^Space=t}', "");
Expect(1, 12288, '\p{Space=_	t}', "");
Expect(0, 12288, '\p{^Space=_	t}', "");
Expect(0, 12288, '\P{Space=_	t}', "");
Expect(1, 12288, '\P{^Space=_	t}', "");
Expect(0, 12289, '\p{Space=_	t}', "");
Expect(1, 12289, '\p{^Space=_	t}', "");
Expect(1, 12289, '\P{Space=_	t}', "");
Expect(0, 12289, '\P{^Space=_	t}', "");
Error('\p{Is_White_Space=-/a/True}');
Error('\P{Is_White_Space=-/a/True}');
Expect(1, 12288, '\p{Is_White_Space=true}', "");
Expect(0, 12288, '\p{^Is_White_Space=true}', "");
Expect(0, 12288, '\P{Is_White_Space=true}', "");
Expect(1, 12288, '\P{^Is_White_Space=true}', "");
Expect(0, 12289, '\p{Is_White_Space=true}', "");
Expect(1, 12289, '\p{^Is_White_Space=true}', "");
Expect(1, 12289, '\P{Is_White_Space=true}', "");
Expect(0, 12289, '\P{^Is_White_Space=true}', "");
Expect(1, 12288, '\p{Is_White_Space=  True}', "");
Expect(0, 12288, '\p{^Is_White_Space=  True}', "");
Expect(0, 12288, '\P{Is_White_Space=  True}', "");
Expect(1, 12288, '\P{^Is_White_Space=  True}', "");
Expect(0, 12289, '\p{Is_White_Space=  True}', "");
Expect(1, 12289, '\p{^Is_White_Space=  True}', "");
Expect(1, 12289, '\P{Is_White_Space=  True}', "");
Expect(0, 12289, '\P{^Is_White_Space=  True}', "");
Error('\p{Is_WSpace:	:=	 YES}');
Error('\P{Is_WSpace:	:=	 YES}');
Expect(1, 12288, '\p{Is_WSpace=yes}', "");
Expect(0, 12288, '\p{^Is_WSpace=yes}', "");
Expect(0, 12288, '\P{Is_WSpace=yes}', "");
Expect(1, 12288, '\P{^Is_WSpace=yes}', "");
Expect(0, 12289, '\p{Is_WSpace=yes}', "");
Expect(1, 12289, '\p{^Is_WSpace=yes}', "");
Expect(1, 12289, '\P{Is_WSpace=yes}', "");
Expect(0, 12289, '\P{^Is_WSpace=yes}', "");
Expect(1, 12288, '\p{Is_WSpace=	yes}', "");
Expect(0, 12288, '\p{^Is_WSpace=	yes}', "");
Expect(0, 12288, '\P{Is_WSpace=	yes}', "");
Expect(1, 12288, '\P{^Is_WSpace=	yes}', "");
Expect(0, 12289, '\p{Is_WSpace=	yes}', "");
Expect(1, 12289, '\p{^Is_WSpace=	yes}', "");
Expect(1, 12289, '\P{Is_WSpace=	yes}', "");
Expect(0, 12289, '\P{^Is_WSpace=	yes}', "");
Error('\p{Is_Space=	:=Y}');
Error('\P{Is_Space=	:=Y}');
Expect(1, 12288, '\p{Is_Space=y}', "");
Expect(0, 12288, '\p{^Is_Space=y}', "");
Expect(0, 12288, '\P{Is_Space=y}', "");
Expect(1, 12288, '\P{^Is_Space=y}', "");
Expect(0, 12289, '\p{Is_Space=y}', "");
Expect(1, 12289, '\p{^Is_Space=y}', "");
Expect(1, 12289, '\P{Is_Space=y}', "");
Expect(0, 12289, '\P{^Is_Space=y}', "");
Expect(1, 12288, '\p{Is_Space=-_Y}', "");
Expect(0, 12288, '\p{^Is_Space=-_Y}', "");
Expect(0, 12288, '\P{Is_Space=-_Y}', "");
Expect(1, 12288, '\P{^Is_Space=-_Y}', "");
Expect(0, 12289, '\p{Is_Space=-_Y}', "");
Expect(1, 12289, '\p{^Is_Space=-_Y}', "");
Expect(1, 12289, '\P{Is_Space=-_Y}', "");
Expect(0, 12289, '\P{^Is_Space=-_Y}', "");
Error('\p{XID_Continue=-:=NO}');
Error('\P{XID_Continue=-:=NO}');
Expect(1, 918000, '\p{XID_Continue=no}', "");
Expect(0, 918000, '\p{^XID_Continue=no}', "");
Expect(0, 918000, '\P{XID_Continue=no}', "");
Expect(1, 918000, '\P{^XID_Continue=no}', "");
Expect(0, 917999, '\p{XID_Continue=no}', "");
Expect(1, 917999, '\p{^XID_Continue=no}', "");
Expect(1, 917999, '\P{XID_Continue=no}', "");
Expect(0, 917999, '\P{^XID_Continue=no}', "");
Expect(1, 918000, '\p{XID_Continue= _no}', "");
Expect(0, 918000, '\p{^XID_Continue= _no}', "");
Expect(0, 918000, '\P{XID_Continue= _no}', "");
Expect(1, 918000, '\P{^XID_Continue= _no}', "");
Expect(0, 917999, '\p{XID_Continue= _no}', "");
Expect(1, 917999, '\p{^XID_Continue= _no}', "");
Expect(1, 917999, '\P{XID_Continue= _no}', "");
Expect(0, 917999, '\P{^XID_Continue= _no}', "");
Error('\p{XIDC=:=-_n}');
Error('\P{XIDC=:=-_n}');
Expect(1, 918000, '\p{XIDC=n}', "");
Expect(0, 918000, '\p{^XIDC=n}', "");
Expect(0, 918000, '\P{XIDC=n}', "");
Expect(1, 918000, '\P{^XIDC=n}', "");
Expect(0, 917999, '\p{XIDC=n}', "");
Expect(1, 917999, '\p{^XIDC=n}', "");
Expect(1, 917999, '\P{XIDC=n}', "");
Expect(0, 917999, '\P{^XIDC=n}', "");
Expect(1, 918000, '\p{XIDC=		N}', "");
Expect(0, 918000, '\p{^XIDC=		N}', "");
Expect(0, 918000, '\P{XIDC=		N}', "");
Expect(1, 918000, '\P{^XIDC=		N}', "");
Expect(0, 917999, '\p{XIDC=		N}', "");
Expect(1, 917999, '\p{^XIDC=		N}', "");
Expect(1, 917999, '\P{XIDC=		N}', "");
Expect(0, 917999, '\P{^XIDC=		N}', "");
Error('\p{Is_XID_Continue:	/a/_F}');
Error('\P{Is_XID_Continue:	/a/_F}');
Expect(1, 918000, '\p{Is_XID_Continue=f}', "");
Expect(0, 918000, '\p{^Is_XID_Continue=f}', "");
Expect(0, 918000, '\P{Is_XID_Continue=f}', "");
Expect(1, 918000, '\P{^Is_XID_Continue=f}', "");
Expect(0, 917999, '\p{Is_XID_Continue=f}', "");
Expect(1, 917999, '\p{^Is_XID_Continue=f}', "");
Expect(1, 917999, '\P{Is_XID_Continue=f}', "");
Expect(0, 917999, '\P{^Is_XID_Continue=f}', "");
Expect(1, 918000, '\p{Is_XID_Continue=	F}', "");
Expect(0, 918000, '\p{^Is_XID_Continue=	F}', "");
Expect(0, 918000, '\P{Is_XID_Continue=	F}', "");
Expect(1, 918000, '\P{^Is_XID_Continue=	F}', "");
Expect(0, 917999, '\p{Is_XID_Continue=	F}', "");
Expect(1, 917999, '\p{^Is_XID_Continue=	F}', "");
Expect(1, 917999, '\P{Is_XID_Continue=	F}', "");
Expect(0, 917999, '\P{^Is_XID_Continue=	F}', "");
Error('\p{Is_XIDC=:=	false}');
Error('\P{Is_XIDC=:=	false}');
Expect(1, 918000, '\p{Is_XIDC=false}', "");
Expect(0, 918000, '\p{^Is_XIDC=false}', "");
Expect(0, 918000, '\P{Is_XIDC=false}', "");
Expect(1, 918000, '\P{^Is_XIDC=false}', "");
Expect(0, 917999, '\p{Is_XIDC=false}', "");
Expect(1, 917999, '\p{^Is_XIDC=false}', "");
Expect(1, 917999, '\P{Is_XIDC=false}', "");
Expect(0, 917999, '\P{^Is_XIDC=false}', "");
Expect(1, 918000, '\p{Is_XIDC= false}', "");
Expect(0, 918000, '\p{^Is_XIDC= false}', "");
Expect(0, 918000, '\P{Is_XIDC= false}', "");
Expect(1, 918000, '\P{^Is_XIDC= false}', "");
Expect(0, 917999, '\p{Is_XIDC= false}', "");
Expect(1, 917999, '\p{^Is_XIDC= false}', "");
Expect(1, 917999, '\P{Is_XIDC= false}', "");
Expect(0, 917999, '\P{^Is_XIDC= false}', "");
Error('\p{XID_Continue=-	Yes/a/}');
Error('\P{XID_Continue=-	Yes/a/}');
Expect(1, 917999, '\p{XID_Continue=yes}', "");
Expect(0, 917999, '\p{^XID_Continue=yes}', "");
Expect(0, 917999, '\P{XID_Continue=yes}', "");
Expect(1, 917999, '\P{^XID_Continue=yes}', "");
Expect(0, 918000, '\p{XID_Continue=yes}', "");
Expect(1, 918000, '\p{^XID_Continue=yes}', "");
Expect(1, 918000, '\P{XID_Continue=yes}', "");
Expect(0, 918000, '\P{^XID_Continue=yes}', "");
Expect(1, 917999, '\p{XID_Continue=_Yes}', "");
Expect(0, 917999, '\p{^XID_Continue=_Yes}', "");
Expect(0, 917999, '\P{XID_Continue=_Yes}', "");
Expect(1, 917999, '\P{^XID_Continue=_Yes}', "");
Expect(0, 918000, '\p{XID_Continue=_Yes}', "");
Expect(1, 918000, '\p{^XID_Continue=_Yes}', "");
Expect(1, 918000, '\P{XID_Continue=_Yes}', "");
Expect(0, 918000, '\P{^XID_Continue=_Yes}', "");
Error('\p{XIDC=-/a/y}');
Error('\P{XIDC=-/a/y}');
Expect(1, 917999, '\p{XIDC:   y}', "");
Expect(0, 917999, '\p{^XIDC:   y}', "");
Expect(0, 917999, '\P{XIDC:   y}', "");
Expect(1, 917999, '\P{^XIDC:   y}', "");
Expect(0, 918000, '\p{XIDC:   y}', "");
Expect(1, 918000, '\p{^XIDC:   y}', "");
Expect(1, 918000, '\P{XIDC:   y}', "");
Expect(0, 918000, '\P{^XIDC:   y}', "");
Expect(1, 917999, '\p{XIDC: 	_y}', "");
Expect(0, 917999, '\p{^XIDC: 	_y}', "");
Expect(0, 917999, '\P{XIDC: 	_y}', "");
Expect(1, 917999, '\P{^XIDC: 	_y}', "");
Expect(0, 918000, '\p{XIDC: 	_y}', "");
Expect(1, 918000, '\p{^XIDC: 	_y}', "");
Expect(1, 918000, '\P{XIDC: 	_y}', "");
Expect(0, 918000, '\P{^XIDC: 	_y}', "");
Error('\p{Is_XID_Continue=- T:=}');
Error('\P{Is_XID_Continue=- T:=}');
Expect(1, 917999, '\p{Is_XID_Continue=t}', "");
Expect(0, 917999, '\p{^Is_XID_Continue=t}', "");
Expect(0, 917999, '\P{Is_XID_Continue=t}', "");
Expect(1, 917999, '\P{^Is_XID_Continue=t}', "");
Expect(0, 918000, '\p{Is_XID_Continue=t}', "");
Expect(1, 918000, '\p{^Is_XID_Continue=t}', "");
Expect(1, 918000, '\P{Is_XID_Continue=t}', "");
Expect(0, 918000, '\P{^Is_XID_Continue=t}', "");
Expect(1, 917999, '\p{Is_XID_Continue:T}', "");
Expect(0, 917999, '\p{^Is_XID_Continue:T}', "");
Expect(0, 917999, '\P{Is_XID_Continue:T}', "");
Expect(1, 917999, '\P{^Is_XID_Continue:T}', "");
Expect(0, 918000, '\p{Is_XID_Continue:T}', "");
Expect(1, 918000, '\p{^Is_XID_Continue:T}', "");
Expect(1, 918000, '\P{Is_XID_Continue:T}', "");
Expect(0, 918000, '\P{^Is_XID_Continue:T}', "");
Error('\p{Is_XIDC= 	True/a/}');
Error('\P{Is_XIDC= 	True/a/}');
Expect(1, 917999, '\p{Is_XIDC=true}', "");
Expect(0, 917999, '\p{^Is_XIDC=true}', "");
Expect(0, 917999, '\P{Is_XIDC=true}', "");
Expect(1, 917999, '\P{^Is_XIDC=true}', "");
Expect(0, 918000, '\p{Is_XIDC=true}', "");
Expect(1, 918000, '\p{^Is_XIDC=true}', "");
Expect(1, 918000, '\P{Is_XIDC=true}', "");
Expect(0, 918000, '\P{^Is_XIDC=true}', "");
Expect(1, 917999, '\p{Is_XIDC=_True}', "");
Expect(0, 917999, '\p{^Is_XIDC=_True}', "");
Expect(0, 917999, '\P{Is_XIDC=_True}', "");
Expect(1, 917999, '\P{^Is_XIDC=_True}', "");
Expect(0, 918000, '\p{Is_XIDC=_True}', "");
Expect(1, 918000, '\p{^Is_XIDC=_True}', "");
Expect(1, 918000, '\P{Is_XIDC=_True}', "");
Expect(0, 918000, '\P{^Is_XIDC=_True}', "");
Error('\p{XID_Start=:= NO}');
Error('\P{XID_Start=:= NO}');
Expect(1, 195102, '\p{XID_Start=no}', "");
Expect(0, 195102, '\p{^XID_Start=no}', "");
Expect(0, 195102, '\P{XID_Start=no}', "");
Expect(1, 195102, '\P{^XID_Start=no}', "");
Expect(0, 195101, '\p{XID_Start=no}', "");
Expect(1, 195101, '\p{^XID_Start=no}', "");
Expect(1, 195101, '\P{XID_Start=no}', "");
Expect(0, 195101, '\P{^XID_Start=no}', "");
Expect(1, 195102, '\p{XID_Start= No}', "");
Expect(0, 195102, '\p{^XID_Start= No}', "");
Expect(0, 195102, '\P{XID_Start= No}', "");
Expect(1, 195102, '\P{^XID_Start= No}', "");
Expect(0, 195101, '\p{XID_Start= No}', "");
Expect(1, 195101, '\p{^XID_Start= No}', "");
Expect(1, 195101, '\P{XID_Start= No}', "");
Expect(0, 195101, '\P{^XID_Start= No}', "");
Error('\p{XIDS=/a/_-N}');
Error('\P{XIDS=/a/_-N}');
Expect(1, 195102, '\p{XIDS=n}', "");
Expect(0, 195102, '\p{^XIDS=n}', "");
Expect(0, 195102, '\P{XIDS=n}', "");
Expect(1, 195102, '\P{^XIDS=n}', "");
Expect(0, 195101, '\p{XIDS=n}', "");
Expect(1, 195101, '\p{^XIDS=n}', "");
Expect(1, 195101, '\P{XIDS=n}', "");
Expect(0, 195101, '\P{^XIDS=n}', "");
Expect(1, 195102, '\p{XIDS=- N}', "");
Expect(0, 195102, '\p{^XIDS=- N}', "");
Expect(0, 195102, '\P{XIDS=- N}', "");
Expect(1, 195102, '\P{^XIDS=- N}', "");
Expect(0, 195101, '\p{XIDS=- N}', "");
Expect(1, 195101, '\p{^XIDS=- N}', "");
Expect(1, 195101, '\P{XIDS=- N}', "");
Expect(0, 195101, '\P{^XIDS=- N}', "");
Error('\p{Is_XID_Start=:=-_F}');
Error('\P{Is_XID_Start=:=-_F}');
Expect(1, 195102, '\p{Is_XID_Start=f}', "");
Expect(0, 195102, '\p{^Is_XID_Start=f}', "");
Expect(0, 195102, '\P{Is_XID_Start=f}', "");
Expect(1, 195102, '\P{^Is_XID_Start=f}', "");
Expect(0, 195101, '\p{Is_XID_Start=f}', "");
Expect(1, 195101, '\p{^Is_XID_Start=f}', "");
Expect(1, 195101, '\P{Is_XID_Start=f}', "");
Expect(0, 195101, '\P{^Is_XID_Start=f}', "");
Expect(1, 195102, '\p{Is_XID_Start=_F}', "");
Expect(0, 195102, '\p{^Is_XID_Start=_F}', "");
Expect(0, 195102, '\P{Is_XID_Start=_F}', "");
Expect(1, 195102, '\P{^Is_XID_Start=_F}', "");
Expect(0, 195101, '\p{Is_XID_Start=_F}', "");
Expect(1, 195101, '\p{^Is_XID_Start=_F}', "");
Expect(1, 195101, '\P{Is_XID_Start=_F}', "");
Expect(0, 195101, '\P{^Is_XID_Start=_F}', "");
Error('\p{Is_XIDS= :=FALSE}');
Error('\P{Is_XIDS= :=FALSE}');
Expect(1, 195102, '\p{Is_XIDS=false}', "");
Expect(0, 195102, '\p{^Is_XIDS=false}', "");
Expect(0, 195102, '\P{Is_XIDS=false}', "");
Expect(1, 195102, '\P{^Is_XIDS=false}', "");
Expect(0, 195101, '\p{Is_XIDS=false}', "");
Expect(1, 195101, '\p{^Is_XIDS=false}', "");
Expect(1, 195101, '\P{Is_XIDS=false}', "");
Expect(0, 195101, '\P{^Is_XIDS=false}', "");
Expect(1, 195102, '\p{Is_XIDS=	FALSE}', "");
Expect(0, 195102, '\p{^Is_XIDS=	FALSE}', "");
Expect(0, 195102, '\P{Is_XIDS=	FALSE}', "");
Expect(1, 195102, '\P{^Is_XIDS=	FALSE}', "");
Expect(0, 195101, '\p{Is_XIDS=	FALSE}', "");
Expect(1, 195101, '\p{^Is_XIDS=	FALSE}', "");
Expect(1, 195101, '\P{Is_XIDS=	FALSE}', "");
Expect(0, 195101, '\P{^Is_XIDS=	FALSE}', "");
Error('\p{XID_Start:	_/a/Yes}');
Error('\P{XID_Start:	_/a/Yes}');
Expect(1, 195101, '\p{XID_Start=yes}', "");
Expect(0, 195101, '\p{^XID_Start=yes}', "");
Expect(0, 195101, '\P{XID_Start=yes}', "");
Expect(1, 195101, '\P{^XID_Start=yes}', "");
Expect(0, 195102, '\p{XID_Start=yes}', "");
Expect(1, 195102, '\p{^XID_Start=yes}', "");
Expect(1, 195102, '\P{XID_Start=yes}', "");
Expect(0, 195102, '\P{^XID_Start=yes}', "");
Expect(1, 195101, '\p{XID_Start=		YES}', "");
Expect(0, 195101, '\p{^XID_Start=		YES}', "");
Expect(0, 195101, '\P{XID_Start=		YES}', "");
Expect(1, 195101, '\P{^XID_Start=		YES}', "");
Expect(0, 195102, '\p{XID_Start=		YES}', "");
Expect(1, 195102, '\p{^XID_Start=		YES}', "");
Expect(1, 195102, '\P{XID_Start=		YES}', "");
Expect(0, 195102, '\P{^XID_Start=		YES}', "");
Error('\p{XIDS=/a/_Y}');
Error('\P{XIDS=/a/_Y}');
Expect(1, 195101, '\p{XIDS=y}', "");
Expect(0, 195101, '\p{^XIDS=y}', "");
Expect(0, 195101, '\P{XIDS=y}', "");
Expect(1, 195101, '\P{^XIDS=y}', "");
Expect(0, 195102, '\p{XIDS=y}', "");
Expect(1, 195102, '\p{^XIDS=y}', "");
Expect(1, 195102, '\P{XIDS=y}', "");
Expect(0, 195102, '\P{^XIDS=y}', "");
Expect(1, 195101, '\p{XIDS=_	y}', "");
Expect(0, 195101, '\p{^XIDS=_	y}', "");
Expect(0, 195101, '\P{XIDS=_	y}', "");
Expect(1, 195101, '\P{^XIDS=_	y}', "");
Expect(0, 195102, '\p{XIDS=_	y}', "");
Expect(1, 195102, '\p{^XIDS=_	y}', "");
Expect(1, 195102, '\P{XIDS=_	y}', "");
Expect(0, 195102, '\P{^XIDS=_	y}', "");
Error('\p{Is_XID_Start=-t/a/}');
Error('\P{Is_XID_Start=-t/a/}');
Expect(1, 195101, '\p{Is_XID_Start=t}', "");
Expect(0, 195101, '\p{^Is_XID_Start=t}', "");
Expect(0, 195101, '\P{Is_XID_Start=t}', "");
Expect(1, 195101, '\P{^Is_XID_Start=t}', "");
Expect(0, 195102, '\p{Is_XID_Start=t}', "");
Expect(1, 195102, '\p{^Is_XID_Start=t}', "");
Expect(1, 195102, '\P{Is_XID_Start=t}', "");
Expect(0, 195102, '\P{^Is_XID_Start=t}', "");
Expect(1, 195101, '\p{Is_XID_Start=_T}', "");
Expect(0, 195101, '\p{^Is_XID_Start=_T}', "");
Expect(0, 195101, '\P{Is_XID_Start=_T}', "");
Expect(1, 195101, '\P{^Is_XID_Start=_T}', "");
Expect(0, 195102, '\p{Is_XID_Start=_T}', "");
Expect(1, 195102, '\p{^Is_XID_Start=_T}', "");
Expect(1, 195102, '\P{Is_XID_Start=_T}', "");
Expect(0, 195102, '\P{^Is_XID_Start=_T}', "");
Error('\p{Is_XIDS= :=True}');
Error('\P{Is_XIDS= :=True}');
Expect(1, 195101, '\p{Is_XIDS=true}', "");
Expect(0, 195101, '\p{^Is_XIDS=true}', "");
Expect(0, 195101, '\P{Is_XIDS=true}', "");
Expect(1, 195101, '\P{^Is_XIDS=true}', "");
Expect(0, 195102, '\p{Is_XIDS=true}', "");
Expect(1, 195102, '\p{^Is_XIDS=true}', "");
Expect(1, 195102, '\P{Is_XIDS=true}', "");
Expect(0, 195102, '\P{^Is_XIDS=true}', "");
Expect(1, 195101, '\p{Is_XIDS=_	True}', "");
Expect(0, 195101, '\p{^Is_XIDS=_	True}', "");
Expect(0, 195101, '\P{Is_XIDS=_	True}', "");
Expect(1, 195101, '\P{^Is_XIDS=_	True}', "");
Expect(0, 195102, '\p{Is_XIDS=_	True}', "");
Expect(1, 195102, '\p{^Is_XIDS=_	True}', "");
Expect(1, 195102, '\P{Is_XIDS=_	True}', "");
Expect(0, 195102, '\P{^Is_XIDS=_	True}', "");
Error('\p{Expands_On_NFC=No}');
Error('\P{Expands_On_NFC=No}');
Error('\p{XO_NFC=N}');
Error('\P{XO_NFC=N}');
Error('\p{Is_Expands_On_NFC=F}');
Error('\P{Is_Expands_On_NFC=F}');
Error('\p{Is_XO_NFC:	False}');
Error('\P{Is_XO_NFC:	False}');
Error('\p{Expands_On_NFC=Yes}');
Error('\P{Expands_On_NFC=Yes}');
Error('\p{XO_NFC=Y}');
Error('\P{XO_NFC=Y}');
Error('\p{Is_Expands_On_NFC=T}');
Error('\P{Is_Expands_On_NFC=T}');
Error('\p{Is_XO_NFC:	True}');
Error('\P{Is_XO_NFC:	True}');
Error('\p{Expands_On_NFD:	No}');
Error('\P{Expands_On_NFD:	No}');
Error('\p{XO_NFD=N}');
Error('\P{XO_NFD=N}');
Error('\p{Is_Expands_On_NFD=F}');
Error('\P{Is_Expands_On_NFD=F}');
Error('\p{Is_XO_NFD=False}');
Error('\P{Is_XO_NFD=False}');
Error('\p{Expands_On_NFD=Yes}');
Error('\P{Expands_On_NFD=Yes}');
Error('\p{XO_NFD=Y}');
Error('\P{XO_NFD=Y}');
Error('\p{Is_Expands_On_NFD=T}');
Error('\P{Is_Expands_On_NFD=T}');
Error('\p{Is_XO_NFD=True}');
Error('\P{Is_XO_NFD=True}');
Error('\p{Expands_On_NFKC: No}');
Error('\P{Expands_On_NFKC: No}');
Error('\p{XO_NFKC: N}');
Error('\P{XO_NFKC: N}');
Error('\p{Is_Expands_On_NFKC=F}');
Error('\P{Is_Expands_On_NFKC=F}');
Error('\p{Is_XO_NFKC=False}');
Error('\P{Is_XO_NFKC=False}');
Error('\p{Expands_On_NFKC:   Yes}');
Error('\P{Expands_On_NFKC:   Yes}');
Error('\p{XO_NFKC=Y}');
Error('\P{XO_NFKC=Y}');
Error('\p{Is_Expands_On_NFKC=T}');
Error('\P{Is_Expands_On_NFKC=T}');
Error('\p{Is_XO_NFKC=True}');
Error('\P{Is_XO_NFKC=True}');
Error('\p{Expands_On_NFKD=No}');
Error('\P{Expands_On_NFKD=No}');
Error('\p{XO_NFKD=N}');
Error('\P{XO_NFKD=N}');
Error('\p{Is_Expands_On_NFKD=F}');
Error('\P{Is_Expands_On_NFKD=F}');
Error('\p{Is_XO_NFKD=False}');
Error('\P{Is_XO_NFKD=False}');
Error('\p{Expands_On_NFKD:   Yes}');
Error('\P{Expands_On_NFKD:   Yes}');
Error('\p{XO_NFKD=Y}');
Error('\P{XO_NFKD=Y}');
Error('\p{Is_Expands_On_NFKD:   T}');
Error('\P{Is_Expands_On_NFKD:   T}');
Error('\p{Is_XO_NFKD=True}');
Error('\P{Is_XO_NFKD=True}');
Error('\p{Age= :=V1_1}');
Error('\P{Age= :=V1_1}');
Expect(1, 65533, '\p{Age=v11}', "");
Expect(0, 65533, '\p{^Age=v11}', "");
Expect(0, 65533, '\P{Age=v11}', "");
Expect(1, 65533, '\P{^Age=v11}', "");
Expect(0, 65536, '\p{Age=v11}', "");
Expect(1, 65536, '\p{^Age=v11}', "");
Expect(1, 65536, '\P{Age=v11}', "");
Expect(0, 65536, '\P{^Age=v11}', "");
Expect(1, 65533, '\p{Age=_V1_1}', "");
Expect(0, 65533, '\p{^Age=_V1_1}', "");
Expect(0, 65533, '\P{Age=_V1_1}', "");
Expect(1, 65533, '\P{^Age=_V1_1}', "");
Expect(0, 65536, '\p{Age=_V1_1}', "");
Expect(1, 65536, '\p{^Age=_V1_1}', "");
Expect(1, 65536, '\P{Age=_V1_1}', "");
Expect(0, 65536, '\P{^Age=_V1_1}', "");
Error('\p{Is_Age=:=__+0001.1}');
Error('\P{Is_Age=:=__+0001.1}');
Expect(1, 65533, '\p{Is_Age=+1.1}', "");
Expect(0, 65533, '\p{^Is_Age=+1.1}', "");
Expect(0, 65533, '\P{Is_Age=+1.1}', "");
Expect(1, 65533, '\P{^Is_Age=+1.1}', "");
Expect(0, 65536, '\p{Is_Age=+1.1}', "");
Expect(1, 65536, '\p{^Is_Age=+1.1}', "");
Expect(1, 65536, '\P{Is_Age=+1.1}', "");
Expect(0, 65536, '\P{^Is_Age=+1.1}', "");
Error('\p{Age=/a/V2_0}');
Error('\P{Age=/a/V2_0}');
Expect(1, 983040, '\p{Age=v20}', "");
Expect(0, 983040, '\p{^Age=v20}', "");
Expect(0, 983040, '\P{Age=v20}', "");
Expect(1, 983040, '\P{^Age=v20}', "");
Expect(0, 983037, '\p{Age=v20}', "");
Expect(1, 983037, '\p{^Age=v20}', "");
Expect(1, 983037, '\P{Age=v20}', "");
Expect(0, 983037, '\P{^Age=v20}', "");
Expect(1, 983040, '\p{Age=__V2_0}', "");
Expect(0, 983040, '\p{^Age=__V2_0}', "");
Expect(0, 983040, '\P{Age=__V2_0}', "");
Expect(1, 983040, '\P{^Age=__V2_0}', "");
Expect(0, 983037, '\p{Age=__V2_0}', "");
Expect(1, 983037, '\p{^Age=__V2_0}', "");
Expect(1, 983037, '\P{Age=__V2_0}', "");
Expect(0, 983037, '\P{^Age=__V2_0}', "");
Error('\p{Is_Age=-:=+2.0}');
Error('\P{Is_Age=-:=+2.0}');
Expect(1, 983040, '\p{Is_Age=+0002.0}', "");
Expect(0, 983040, '\p{^Is_Age=+0002.0}', "");
Expect(0, 983040, '\P{Is_Age=+0002.0}', "");
Expect(1, 983040, '\P{^Is_Age=+0002.0}', "");
Expect(0, 983037, '\p{Is_Age=+0002.0}', "");
Expect(1, 983037, '\p{^Is_Age=+0002.0}', "");
Expect(1, 983037, '\P{Is_Age=+0002.0}', "");
Expect(0, 983037, '\P{^Is_Age=+0002.0}', "");
Error('\p{Age=_v2_1/a/}');
Error('\P{Age=_v2_1/a/}');
Expect(1, 65532, '\p{Age=v21}', "");
Expect(0, 65532, '\p{^Age=v21}', "");
Expect(0, 65532, '\P{Age=v21}', "");
Expect(1, 65532, '\P{^Age=v21}', "");
Expect(0, 65533, '\p{Age=v21}', "");
Expect(1, 65533, '\p{^Age=v21}', "");
Expect(1, 65533, '\P{Age=v21}', "");
Expect(0, 65533, '\P{^Age=v21}', "");
Expect(1, 65532, '\p{Age=-V2_1}', "");
Expect(0, 65532, '\p{^Age=-V2_1}', "");
Expect(0, 65532, '\P{Age=-V2_1}', "");
Expect(1, 65532, '\P{^Age=-V2_1}', "");
Expect(0, 65533, '\p{Age=-V2_1}', "");
Expect(1, 65533, '\p{^Age=-V2_1}', "");
Expect(1, 65533, '\P{Age=-V2_1}', "");
Expect(0, 65533, '\P{^Age=-V2_1}', "");
Error('\p{Is_Age: __2.1/a/}');
Error('\P{Is_Age: __2.1/a/}');
Expect(1, 65532, '\p{Is_Age=02.1}', "");
Expect(0, 65532, '\p{^Is_Age=02.1}', "");
Expect(0, 65532, '\P{Is_Age=02.1}', "");
Expect(1, 65532, '\P{^Is_Age=02.1}', "");
Expect(0, 65533, '\p{Is_Age=02.1}', "");
Expect(1, 65533, '\p{^Is_Age=02.1}', "");
Expect(1, 65533, '\P{Is_Age=02.1}', "");
Expect(0, 65533, '\P{^Is_Age=02.1}', "");
Error('\p{Age=-:=v3_0}');
Error('\P{Age=-:=v3_0}');
Expect(1, 65531, '\p{Age=v30}', "");
Expect(0, 65531, '\p{^Age=v30}', "");
Expect(0, 65531, '\P{Age=v30}', "");
Expect(1, 65531, '\P{^Age=v30}', "");
Expect(0, 65532, '\p{Age=v30}', "");
Expect(1, 65532, '\p{^Age=v30}', "");
Expect(1, 65532, '\P{Age=v30}', "");
Expect(0, 65532, '\P{^Age=v30}', "");
Expect(1, 65531, '\p{Age: --v3_0}', "");
Expect(0, 65531, '\p{^Age: --v3_0}', "");
Expect(0, 65531, '\P{Age: --v3_0}', "");
Expect(1, 65531, '\P{^Age: --v3_0}', "");
Expect(0, 65532, '\p{Age: --v3_0}', "");
Expect(1, 65532, '\p{^Age: --v3_0}', "");
Expect(1, 65532, '\P{Age: --v3_0}', "");
Expect(0, 65532, '\P{^Age: --v3_0}', "");
Error('\p{Is_Age=_:=+3.0}');
Error('\P{Is_Age=_:=+3.0}');
Expect(1, 65531, '\p{Is_Age=0000003.0}', "");
Expect(0, 65531, '\p{^Is_Age=0000003.0}', "");
Expect(0, 65531, '\P{Is_Age=0000003.0}', "");
Expect(1, 65531, '\P{^Is_Age=0000003.0}', "");
Expect(0, 65532, '\p{Is_Age=0000003.0}', "");
Expect(1, 65532, '\p{^Is_Age=0000003.0}', "");
Expect(1, 65532, '\P{Is_Age=0000003.0}', "");
Expect(0, 65532, '\P{^Is_Age=0000003.0}', "");
Error('\p{Age=:=	 V3_1}');
Error('\P{Age=:=	 V3_1}');
Expect(1, 917631, '\p{Age:v31}', "");
Expect(0, 917631, '\p{^Age:v31}', "");
Expect(0, 917631, '\P{Age:v31}', "");
Expect(1, 917631, '\P{^Age:v31}', "");
Expect(0, 917632, '\p{Age:v31}', "");
Expect(1, 917632, '\p{^Age:v31}', "");
Expect(1, 917632, '\P{Age:v31}', "");
Expect(0, 917632, '\P{^Age:v31}', "");
Expect(1, 917631, '\p{Age=V3_1}', "");
Expect(0, 917631, '\p{^Age=V3_1}', "");
Expect(0, 917631, '\P{Age=V3_1}', "");
Expect(1, 917631, '\P{^Age=V3_1}', "");
Expect(0, 917632, '\p{Age=V3_1}', "");
Expect(1, 917632, '\p{^Age=V3_1}', "");
Expect(1, 917632, '\P{Age=V3_1}', "");
Expect(0, 917632, '\P{^Age=V3_1}', "");
Error('\p{Is_Age=/a/  000_3.1}');
Error('\P{Is_Age=/a/  000_3.1}');
Expect(1, 917631, '\p{Is_Age=0000003.1}', "");
Expect(0, 917631, '\p{^Is_Age=0000003.1}', "");
Expect(0, 917631, '\P{Is_Age=0000003.1}', "");
Expect(1, 917631, '\P{^Is_Age=0000003.1}', "");
Expect(0, 917632, '\p{Is_Age=0000003.1}', "");
Expect(1, 917632, '\p{^Is_Age=0000003.1}', "");
Expect(1, 917632, '\P{Is_Age=0000003.1}', "");
Expect(0, 917632, '\P{^Is_Age=0000003.1}', "");
Error('\p{Age=V3_2:=}');
Error('\P{Age=V3_2:=}');
Expect(1, 65376, '\p{Age: v32}', "");
Expect(0, 65376, '\p{^Age: v32}', "");
Expect(0, 65376, '\P{Age: v32}', "");
Expect(1, 65376, '\P{^Age: v32}', "");
Expect(0, 65377, '\p{Age: v32}', "");
Expect(1, 65377, '\p{^Age: v32}', "");
Expect(1, 65377, '\P{Age: v32}', "");
Expect(0, 65377, '\P{^Age: v32}', "");
Expect(1, 65376, '\p{Age=v3_2}', "");
Expect(0, 65376, '\p{^Age=v3_2}', "");
Expect(0, 65376, '\P{Age=v3_2}', "");
Expect(1, 65376, '\P{^Age=v3_2}', "");
Expect(0, 65377, '\p{Age=v3_2}', "");
Expect(1, 65377, '\p{^Age=v3_2}', "");
Expect(1, 65377, '\P{Age=v3_2}', "");
Expect(0, 65377, '\P{^Age=v3_2}', "");
Error('\p{Is_Age=_-3.2/a/}');
Error('\P{Is_Age=_-3.2/a/}');
Expect(1, 65376, '\p{Is_Age=+0000_3.2}', "");
Expect(0, 65376, '\p{^Is_Age=+0000_3.2}', "");
Expect(0, 65376, '\P{Is_Age=+0000_3.2}', "");
Expect(1, 65376, '\P{^Is_Age=+0000_3.2}', "");
Expect(0, 65377, '\p{Is_Age=+0000_3.2}', "");
Expect(1, 65377, '\p{^Is_Age=+0000_3.2}', "");
Expect(1, 65377, '\P{Is_Age=+0000_3.2}', "");
Expect(0, 65377, '\P{^Is_Age=+0000_3.2}', "");
Error('\p{Age=_	V4_0/a/}');
Error('\P{Age=_	V4_0/a/}');
Expect(1, 917999, '\p{Age:	v40}', "");
Expect(0, 917999, '\p{^Age:	v40}', "");
Expect(0, 917999, '\P{Age:	v40}', "");
Expect(1, 917999, '\P{^Age:	v40}', "");
Expect(0, 918000, '\p{Age:	v40}', "");
Expect(1, 918000, '\p{^Age:	v40}', "");
Expect(1, 918000, '\P{Age:	v40}', "");
Expect(0, 918000, '\P{^Age:	v40}', "");
Expect(1, 917999, '\p{Age=	V4_0}', "");
Expect(0, 917999, '\p{^Age=	V4_0}', "");
Expect(0, 917999, '\P{Age=	V4_0}', "");
Expect(1, 917999, '\P{^Age=	V4_0}', "");
Expect(0, 918000, '\p{Age=	V4_0}', "");
Expect(1, 918000, '\p{^Age=	V4_0}', "");
Expect(1, 918000, '\P{Age=	V4_0}', "");
Expect(0, 918000, '\P{^Age=	V4_0}', "");
Error('\p{Is_Age= 	+00000004.0/a/}');
Error('\P{Is_Age= 	+00000004.0/a/}');
Expect(1, 917999, '\p{Is_Age=00004.0}', "");
Expect(0, 917999, '\p{^Is_Age=00004.0}', "");
Expect(0, 917999, '\P{Is_Age=00004.0}', "");
Expect(1, 917999, '\P{^Is_Age=00004.0}', "");
Expect(0, 918000, '\p{Is_Age=00004.0}', "");
Expect(1, 918000, '\p{^Is_Age=00004.0}', "");
Expect(1, 918000, '\P{Is_Age=00004.0}', "");
Expect(0, 918000, '\P{^Is_Age=00004.0}', "");
Error('\p{Age=  V4_1:=}');
Error('\P{Age=  V4_1:=}');
Expect(1, 120485, '\p{Age=v41}', "");
Expect(0, 120485, '\p{^Age=v41}', "");
Expect(0, 120485, '\P{Age=v41}', "");
Expect(1, 120485, '\P{^Age=v41}', "");
Expect(0, 120486, '\p{Age=v41}', "");
Expect(1, 120486, '\p{^Age=v41}', "");
Expect(1, 120486, '\P{Age=v41}', "");
Expect(0, 120486, '\P{^Age=v41}', "");
Expect(1, 120485, '\p{Age:_	V4_1}', "");
Expect(0, 120485, '\p{^Age:_	V4_1}', "");
Expect(0, 120485, '\P{Age:_	V4_1}', "");
Expect(1, 120485, '\P{^Age:_	V4_1}', "");
Expect(0, 120486, '\p{Age:_	V4_1}', "");
Expect(1, 120486, '\p{^Age:_	V4_1}', "");
Expect(1, 120486, '\P{Age:_	V4_1}', "");
Expect(0, 120486, '\P{^Age:_	V4_1}', "");
Error('\p{Is_Age: :=_000000004.1}');
Error('\P{Is_Age: :=_000000004.1}');
Expect(1, 120485, '\p{Is_Age=+0000000004.1}', "");
Expect(0, 120485, '\p{^Is_Age=+0000000004.1}', "");
Expect(0, 120485, '\P{Is_Age=+0000000004.1}', "");
Expect(1, 120485, '\P{^Is_Age=+0000000004.1}', "");
Expect(0, 120486, '\p{Is_Age=+0000000004.1}', "");
Expect(1, 120486, '\p{^Is_Age=+0000000004.1}', "");
Expect(1, 120486, '\P{Is_Age=+0000000004.1}', "");
Expect(0, 120486, '\P{^Is_Age=+0000000004.1}', "");
Error('\p{Age::= V5_0}');
Error('\P{Age::= V5_0}');
Expect(1, 120779, '\p{Age=v50}', "");
Expect(0, 120779, '\p{^Age=v50}', "");
Expect(0, 120779, '\P{Age=v50}', "");
Expect(1, 120779, '\P{^Age=v50}', "");
Expect(0, 120780, '\p{Age=v50}', "");
Expect(1, 120780, '\p{^Age=v50}', "");
Expect(1, 120780, '\P{Age=v50}', "");
Expect(0, 120780, '\P{^Age=v50}', "");
Expect(1, 120779, '\p{Age=__v5_0}', "");
Expect(0, 120779, '\p{^Age=__v5_0}', "");
Expect(0, 120779, '\P{Age=__v5_0}', "");
Expect(1, 120779, '\P{^Age=__v5_0}', "");
Expect(0, 120780, '\p{Age=__v5_0}', "");
Expect(1, 120780, '\p{^Age=__v5_0}', "");
Expect(1, 120780, '\P{Age=__v5_0}', "");
Expect(0, 120780, '\P{^Age=__v5_0}', "");
Error('\p{Is_Age:	 /a/+00000_5.0}');
Error('\P{Is_Age:	 /a/+00000_5.0}');
Expect(1, 120779, '\p{Is_Age=000000005.0}', "");
Expect(0, 120779, '\p{^Is_Age=000000005.0}', "");
Expect(0, 120779, '\P{Is_Age=000000005.0}', "");
Expect(1, 120779, '\P{^Is_Age=000000005.0}', "");
Expect(0, 120780, '\p{Is_Age=000000005.0}', "");
Expect(1, 120780, '\p{^Is_Age=000000005.0}', "");
Expect(1, 120780, '\P{Is_Age=000000005.0}', "");
Expect(0, 120780, '\P{^Is_Age=000000005.0}', "");
Error('\p{Age: /a/v5_1}');
Error('\P{Age: /a/v5_1}');
Expect(1, 127123, '\p{Age=v51}', "");
Expect(0, 127123, '\p{^Age=v51}', "");
Expect(0, 127123, '\P{Age=v51}', "");
Expect(1, 127123, '\P{^Age=v51}', "");
Expect(0, 127124, '\p{Age=v51}', "");
Expect(1, 127124, '\p{^Age=v51}', "");
Expect(1, 127124, '\P{Age=v51}', "");
Expect(0, 127124, '\P{^Age=v51}', "");
Expect(1, 127123, '\p{Age=-	V5_1}', "");
Expect(0, 127123, '\p{^Age=-	V5_1}', "");
Expect(0, 127123, '\P{Age=-	V5_1}', "");
Expect(1, 127123, '\P{^Age=-	V5_1}', "");
Expect(0, 127124, '\p{Age=-	V5_1}', "");
Expect(1, 127124, '\p{^Age=-	V5_1}', "");
Expect(1, 127124, '\P{Age=-	V5_1}', "");
Expect(0, 127124, '\P{^Age=-	V5_1}', "");
Error('\p{Is_Age=:=0_0_0_05.1}');
Error('\P{Is_Age=:=0_0_0_05.1}');
Expect(1, 127123, '\p{Is_Age=0000000005.1}', "");
Expect(0, 127123, '\p{^Is_Age=0000000005.1}', "");
Expect(0, 127123, '\P{Is_Age=0000000005.1}', "");
Expect(1, 127123, '\P{^Is_Age=0000000005.1}', "");
Expect(0, 127124, '\p{Is_Age=0000000005.1}', "");
Expect(1, 127124, '\p{^Is_Age=0000000005.1}', "");
Expect(1, 127124, '\P{Is_Age=0000000005.1}', "");
Expect(0, 127124, '\P{^Is_Age=0000000005.1}', "");
Error('\p{Age=:=-_v5_2}');
Error('\P{Age=:=-_v5_2}');
Expect(1, 177972, '\p{Age=v52}', "");
Expect(0, 177972, '\p{^Age=v52}', "");
Expect(0, 177972, '\P{Age=v52}', "");
Expect(1, 177972, '\P{^Age=v52}', "");
Expect(0, 177973, '\p{Age=v52}', "");
Expect(1, 177973, '\p{^Age=v52}', "");
Expect(1, 177973, '\P{Age=v52}', "");
Expect(0, 177973, '\P{^Age=v52}', "");
Expect(1, 177972, '\p{Age=		V5_2}', "");
Expect(0, 177972, '\p{^Age=		V5_2}', "");
Expect(0, 177972, '\P{Age=		V5_2}', "");
Expect(1, 177972, '\P{^Age=		V5_2}', "");
Expect(0, 177973, '\p{Age=		V5_2}', "");
Expect(1, 177973, '\p{^Age=		V5_2}', "");
Expect(1, 177973, '\P{Age=		V5_2}', "");
Expect(0, 177973, '\P{^Age=		V5_2}', "");
Error('\p{Is_Age= 0_0_0_0_0_00005.2:=}');
Error('\P{Is_Age= 0_0_0_0_0_00005.2:=}');
Expect(1, 177972, '\p{Is_Age=0000000005.2}', "");
Expect(0, 177972, '\p{^Is_Age=0000000005.2}', "");
Expect(0, 177972, '\P{Is_Age=0000000005.2}', "");
Expect(1, 177972, '\P{^Is_Age=0000000005.2}', "");
Expect(0, 177973, '\p{Is_Age=0000000005.2}', "");
Expect(1, 177973, '\p{^Is_Age=0000000005.2}', "");
Expect(1, 177973, '\P{Is_Age=0000000005.2}', "");
Expect(0, 177973, '\P{^Is_Age=0000000005.2}', "");
Error('\p{Age= -V6_0:=}');
Error('\P{Age= -V6_0:=}');
Expect(1, 178205, '\p{Age: v60}', "");
Expect(0, 178205, '\p{^Age: v60}', "");
Expect(0, 178205, '\P{Age: v60}', "");
Expect(1, 178205, '\P{^Age: v60}', "");
Expect(0, 178206, '\p{Age: v60}', "");
Expect(1, 178206, '\p{^Age: v60}', "");
Expect(1, 178206, '\P{Age: v60}', "");
Expect(0, 178206, '\P{^Age: v60}', "");
Expect(1, 178205, '\p{Age=- V6_0}', "");
Expect(0, 178205, '\p{^Age=- V6_0}', "");
Expect(0, 178205, '\P{Age=- V6_0}', "");
Expect(1, 178205, '\P{^Age=- V6_0}', "");
Expect(0, 178206, '\p{Age=- V6_0}', "");
Expect(1, 178206, '\p{^Age=- V6_0}', "");
Expect(1, 178206, '\P{Age=- V6_0}', "");
Expect(0, 178206, '\P{^Age=- V6_0}', "");
Error('\p{Is_Age=:= _+0_0_0_06.0}');
Error('\P{Is_Age=:= _+0_0_0_06.0}');
Expect(1, 178205, '\p{Is_Age=0_0_0_0_0_006.0}', "");
Expect(0, 178205, '\p{^Is_Age=0_0_0_0_0_006.0}', "");
Expect(0, 178205, '\P{Is_Age=0_0_0_0_0_006.0}', "");
Expect(1, 178205, '\P{^Is_Age=0_0_0_0_0_006.0}', "");
Expect(0, 178206, '\p{Is_Age=0_0_0_0_0_006.0}', "");
Expect(1, 178206, '\p{^Is_Age=0_0_0_0_0_006.0}', "");
Expect(1, 178206, '\P{Is_Age=0_0_0_0_0_006.0}', "");
Expect(0, 178206, '\P{^Is_Age=0_0_0_0_0_006.0}', "");
Error('\p{Age=-V6_1/a/}');
Error('\P{Age=-V6_1/a/}');
Expect(1, 128564, '\p{Age:   v61}', "");
Expect(0, 128564, '\p{^Age:   v61}', "");
Expect(0, 128564, '\P{Age:   v61}', "");
Expect(1, 128564, '\P{^Age:   v61}', "");
Expect(0, 128565, '\p{Age:   v61}', "");
Expect(1, 128565, '\p{^Age:   v61}', "");
Expect(1, 128565, '\P{Age:   v61}', "");
Expect(0, 128565, '\P{^Age:   v61}', "");
Expect(1, 128564, '\p{Age=- v6_1}', "");
Expect(0, 128564, '\p{^Age=- v6_1}', "");
Expect(0, 128564, '\P{Age=- v6_1}', "");
Expect(1, 128564, '\P{^Age=- v6_1}', "");
Expect(0, 128565, '\p{Age=- v6_1}', "");
Expect(1, 128565, '\p{^Age=- v6_1}', "");
Expect(1, 128565, '\P{Age=- v6_1}', "");
Expect(0, 128565, '\P{^Age=- v6_1}', "");
Error('\p{Is_Age=_ +006.1:=}');
Error('\P{Is_Age=_ +006.1:=}');
Expect(1, 128564, '\p{Is_Age=0_0_0_0_06.1}', "");
Expect(0, 128564, '\p{^Is_Age=0_0_0_0_06.1}', "");
Expect(0, 128564, '\P{Is_Age=0_0_0_0_06.1}', "");
Expect(1, 128564, '\P{^Is_Age=0_0_0_0_06.1}', "");
Expect(0, 128565, '\p{Is_Age=0_0_0_0_06.1}', "");
Expect(1, 128565, '\p{^Is_Age=0_0_0_0_06.1}', "");
Expect(1, 128565, '\P{Is_Age=0_0_0_0_06.1}', "");
Expect(0, 128565, '\P{^Is_Age=0_0_0_0_06.1}', "");
Error('\p{Age=/a/	v6_2}');
Error('\P{Age=/a/	v6_2}');
Expect(1, 8378, '\p{Age=v62}', "");
Expect(0, 8378, '\p{^Age=v62}', "");
Expect(0, 8378, '\P{Age=v62}', "");
Expect(1, 8378, '\P{^Age=v62}', "");
Expect(0, 8379, '\p{Age=v62}', "");
Expect(1, 8379, '\p{^Age=v62}', "");
Expect(1, 8379, '\P{Age=v62}', "");
Expect(0, 8379, '\P{^Age=v62}', "");
Expect(1, 8378, '\p{Age:   _	V6_2}', "");
Expect(0, 8378, '\p{^Age:   _	V6_2}', "");
Expect(0, 8378, '\P{Age:   _	V6_2}', "");
Expect(1, 8378, '\P{^Age:   _	V6_2}', "");
Expect(0, 8379, '\p{Age:   _	V6_2}', "");
Expect(1, 8379, '\p{^Age:   _	V6_2}', "");
Expect(1, 8379, '\P{Age:   _	V6_2}', "");
Expect(0, 8379, '\P{^Age:   _	V6_2}', "");
Error('\p{Is_Age=-00000006.2/a/}');
Error('\P{Is_Age=-00000006.2/a/}');
Expect(1, 8378, '\p{Is_Age:   06.2}', "");
Expect(0, 8378, '\p{^Is_Age:   06.2}', "");
Expect(0, 8378, '\P{Is_Age:   06.2}', "");
Expect(1, 8378, '\P{^Is_Age:   06.2}', "");
Expect(0, 8379, '\p{Is_Age:   06.2}', "");
Expect(1, 8379, '\p{^Is_Age:   06.2}', "");
Expect(1, 8379, '\P{Is_Age:   06.2}', "");
Expect(0, 8379, '\P{^Is_Age:   06.2}', "");
Error('\p{Age=--V6_3/a/}');
Error('\P{Age=--V6_3/a/}');
Expect(1, 8297, '\p{Age=v63}', "");
Expect(0, 8297, '\p{^Age=v63}', "");
Expect(0, 8297, '\P{Age=v63}', "");
Expect(1, 8297, '\P{^Age=v63}', "");
Expect(0, 8298, '\p{Age=v63}', "");
Expect(1, 8298, '\p{^Age=v63}', "");
Expect(1, 8298, '\P{Age=v63}', "");
Expect(0, 8298, '\P{^Age=v63}', "");
Expect(1, 8297, '\p{Age=--V6_3}', "");
Expect(0, 8297, '\p{^Age=--V6_3}', "");
Expect(0, 8297, '\P{Age=--V6_3}', "");
Expect(1, 8297, '\P{^Age=--V6_3}', "");
Expect(0, 8298, '\p{Age=--V6_3}', "");
Expect(1, 8298, '\p{^Age=--V6_3}', "");
Expect(1, 8298, '\P{Age=--V6_3}', "");
Expect(0, 8298, '\P{^Age=--V6_3}', "");
Error('\p{Is_Age=:= 06.3}');
Error('\P{Is_Age=:= 06.3}');
Expect(1, 8297, '\p{Is_Age=+000006.3}', "");
Expect(0, 8297, '\p{^Is_Age=+000006.3}', "");
Expect(0, 8297, '\P{Is_Age=+000006.3}', "");
Expect(1, 8297, '\P{^Is_Age=+000006.3}', "");
Expect(0, 8298, '\p{Is_Age=+000006.3}', "");
Expect(1, 8298, '\p{^Is_Age=+000006.3}', "");
Expect(1, 8298, '\P{Is_Age=+000006.3}', "");
Expect(0, 8298, '\P{^Is_Age=+000006.3}', "");
Error('\p{Age=:=Unassigned}');
Error('\P{Age=:=Unassigned}');
Expect(1, 983037, '\p{Age=unassigned}', "");
Expect(0, 983037, '\p{^Age=unassigned}', "");
Expect(0, 983037, '\P{Age=unassigned}', "");
Expect(1, 983037, '\P{^Age=unassigned}', "");
Expect(0, 983040, '\p{Age=unassigned}', "");
Expect(1, 983040, '\p{^Age=unassigned}', "");
Expect(1, 983040, '\P{Age=unassigned}', "");
Expect(0, 983040, '\P{^Age=unassigned}', "");
Expect(1, 983037, '\p{Age=	Unassigned}', "");
Expect(0, 983037, '\p{^Age=	Unassigned}', "");
Expect(0, 983037, '\P{Age=	Unassigned}', "");
Expect(1, 983037, '\P{^Age=	Unassigned}', "");
Expect(0, 983040, '\p{Age=	Unassigned}', "");
Expect(1, 983040, '\p{^Age=	Unassigned}', "");
Expect(1, 983040, '\P{Age=	Unassigned}', "");
Expect(0, 983040, '\P{^Age=	Unassigned}', "");
Error('\p{Is_Age=	:=NA}');
Error('\P{Is_Age=	:=NA}');
Expect(1, 983037, '\p{Is_Age=na}', "");
Expect(0, 983037, '\p{^Is_Age=na}', "");
Expect(0, 983037, '\P{Is_Age=na}', "");
Expect(1, 983037, '\P{^Is_Age=na}', "");
Expect(0, 983040, '\p{Is_Age=na}', "");
Expect(1, 983040, '\p{^Is_Age=na}', "");
Expect(1, 983040, '\P{Is_Age=na}', "");
Expect(0, 983040, '\P{^Is_Age=na}', "");
Expect(1, 983037, '\p{Is_Age:   	NA}', "");
Expect(0, 983037, '\p{^Is_Age:   	NA}', "");
Expect(0, 983037, '\P{Is_Age:   	NA}', "");
Expect(1, 983037, '\P{^Is_Age:   	NA}', "");
Expect(0, 983040, '\p{Is_Age:   	NA}', "");
Expect(1, 983040, '\p{^Is_Age:   	NA}', "");
Expect(1, 983040, '\P{Is_Age:   	NA}', "");
Expect(0, 983040, '\P{^Is_Age:   	NA}', "");
Error('\p{Bidi_Class=:= arabic_Letter}');
Error('\P{Bidi_Class=:= arabic_Letter}');
Expect(1, 126719, '\p{Bidi_Class=arabicletter}', "");
Expect(0, 126719, '\p{^Bidi_Class=arabicletter}', "");
Expect(0, 126719, '\P{Bidi_Class=arabicletter}', "");
Expect(1, 126719, '\P{^Bidi_Class=arabicletter}', "");
Expect(0, 126720, '\p{Bidi_Class=arabicletter}', "");
Expect(1, 126720, '\p{^Bidi_Class=arabicletter}', "");
Expect(1, 126720, '\P{Bidi_Class=arabicletter}', "");
Expect(0, 126720, '\P{^Bidi_Class=arabicletter}', "");
Expect(1, 126719, '\p{Bidi_Class=__Arabic_letter}', "");
Expect(0, 126719, '\p{^Bidi_Class=__Arabic_letter}', "");
Expect(0, 126719, '\P{Bidi_Class=__Arabic_letter}', "");
Expect(1, 126719, '\P{^Bidi_Class=__Arabic_letter}', "");
Expect(0, 126720, '\p{Bidi_Class=__Arabic_letter}', "");
Expect(1, 126720, '\p{^Bidi_Class=__Arabic_letter}', "");
Expect(1, 126720, '\P{Bidi_Class=__Arabic_letter}', "");
Expect(0, 126720, '\P{^Bidi_Class=__Arabic_letter}', "");
Error('\p{Bc=:=	 AL}');
Error('\P{Bc=:=	 AL}');
Expect(1, 126719, '\p{Bc:   al}', "");
Expect(0, 126719, '\p{^Bc:   al}', "");
Expect(0, 126719, '\P{Bc:   al}', "");
Expect(1, 126719, '\P{^Bc:   al}', "");
Expect(0, 126720, '\p{Bc:   al}', "");
Expect(1, 126720, '\p{^Bc:   al}', "");
Expect(1, 126720, '\P{Bc:   al}', "");
Expect(0, 126720, '\P{^Bc:   al}', "");
Expect(1, 126719, '\p{Bc=__AL}', "");
Expect(0, 126719, '\p{^Bc=__AL}', "");
Expect(0, 126719, '\P{Bc=__AL}', "");
Expect(1, 126719, '\P{^Bc=__AL}', "");
Expect(0, 126720, '\p{Bc=__AL}', "");
Expect(1, 126720, '\p{^Bc=__AL}', "");
Expect(1, 126720, '\P{Bc=__AL}', "");
Expect(0, 126720, '\P{^Bc=__AL}', "");
Error('\p{Is_Bidi_Class=_/a/Arabic_Letter}');
Error('\P{Is_Bidi_Class=_/a/Arabic_Letter}');
Expect(1, 126719, '\p{Is_Bidi_Class=arabicletter}', "");
Expect(0, 126719, '\p{^Is_Bidi_Class=arabicletter}', "");
Expect(0, 126719, '\P{Is_Bidi_Class=arabicletter}', "");
Expect(1, 126719, '\P{^Is_Bidi_Class=arabicletter}', "");
Expect(0, 126720, '\p{Is_Bidi_Class=arabicletter}', "");
Expect(1, 126720, '\p{^Is_Bidi_Class=arabicletter}', "");
Expect(1, 126720, '\P{Is_Bidi_Class=arabicletter}', "");
Expect(0, 126720, '\P{^Is_Bidi_Class=arabicletter}', "");
Expect(1, 126719, '\p{Is_Bidi_Class=	_Arabic_Letter}', "");
Expect(0, 126719, '\p{^Is_Bidi_Class=	_Arabic_Letter}', "");
Expect(0, 126719, '\P{Is_Bidi_Class=	_Arabic_Letter}', "");
Expect(1, 126719, '\P{^Is_Bidi_Class=	_Arabic_Letter}', "");
Expect(0, 126720, '\p{Is_Bidi_Class=	_Arabic_Letter}', "");
Expect(1, 126720, '\p{^Is_Bidi_Class=	_Arabic_Letter}', "");
Expect(1, 126720, '\P{Is_Bidi_Class=	_Arabic_Letter}', "");
Expect(0, 126720, '\P{^Is_Bidi_Class=	_Arabic_Letter}', "");
Error('\p{Is_Bc= /a/AL}');
Error('\P{Is_Bc= /a/AL}');
Expect(1, 126719, '\p{Is_Bc=al}', "");
Expect(0, 126719, '\p{^Is_Bc=al}', "");
Expect(0, 126719, '\P{Is_Bc=al}', "");
Expect(1, 126719, '\P{^Is_Bc=al}', "");
Expect(0, 126720, '\p{Is_Bc=al}', "");
Expect(1, 126720, '\p{^Is_Bc=al}', "");
Expect(1, 126720, '\P{Is_Bc=al}', "");
Expect(0, 126720, '\P{^Is_Bc=al}', "");
Expect(1, 126719, '\p{Is_Bc=_-AL}', "");
Expect(0, 126719, '\p{^Is_Bc=_-AL}', "");
Expect(0, 126719, '\P{Is_Bc=_-AL}', "");
Expect(1, 126719, '\P{^Is_Bc=_-AL}', "");
Expect(0, 126720, '\p{Is_Bc=_-AL}', "");
Expect(1, 126720, '\p{^Is_Bc=_-AL}', "");
Expect(1, 126720, '\P{Is_Bc=_-AL}', "");
Expect(0, 126720, '\P{^Is_Bc=_-AL}', "");
Error('\p{Bidi_Class=-:=Arabic_number}');
Error('\P{Bidi_Class=-:=Arabic_number}');
Expect(1, 69246, '\p{Bidi_Class=arabicnumber}', "");
Expect(0, 69246, '\p{^Bidi_Class=arabicnumber}', "");
Expect(0, 69246, '\P{Bidi_Class=arabicnumber}', "");
Expect(1, 69246, '\P{^Bidi_Class=arabicnumber}', "");
Expect(0, 69247, '\p{Bidi_Class=arabicnumber}', "");
Expect(1, 69247, '\p{^Bidi_Class=arabicnumber}', "");
Expect(1, 69247, '\P{Bidi_Class=arabicnumber}', "");
Expect(0, 69247, '\P{^Bidi_Class=arabicnumber}', "");
Expect(1, 69246, '\p{Bidi_Class= _arabic_number}', "");
Expect(0, 69246, '\p{^Bidi_Class= _arabic_number}', "");
Expect(0, 69246, '\P{Bidi_Class= _arabic_number}', "");
Expect(1, 69246, '\P{^Bidi_Class= _arabic_number}', "");
Expect(0, 69247, '\p{Bidi_Class= _arabic_number}', "");
Expect(1, 69247, '\p{^Bidi_Class= _arabic_number}', "");
Expect(1, 69247, '\P{Bidi_Class= _arabic_number}', "");
Expect(0, 69247, '\P{^Bidi_Class= _arabic_number}', "");
Error('\p{Bc=	/a/AN}');
Error('\P{Bc=	/a/AN}');
Expect(1, 69246, '\p{Bc=an}', "");
Expect(0, 69246, '\p{^Bc=an}', "");
Expect(0, 69246, '\P{Bc=an}', "");
Expect(1, 69246, '\P{^Bc=an}', "");
Expect(0, 69247, '\p{Bc=an}', "");
Expect(1, 69247, '\p{^Bc=an}', "");
Expect(1, 69247, '\P{Bc=an}', "");
Expect(0, 69247, '\P{^Bc=an}', "");
Expect(1, 69246, '\p{Bc=--AN}', "");
Expect(0, 69246, '\p{^Bc=--AN}', "");
Expect(0, 69246, '\P{Bc=--AN}', "");
Expect(1, 69246, '\P{^Bc=--AN}', "");
Expect(0, 69247, '\p{Bc=--AN}', "");
Expect(1, 69247, '\p{^Bc=--AN}', "");
Expect(1, 69247, '\P{Bc=--AN}', "");
Expect(0, 69247, '\P{^Bc=--AN}', "");
Error('\p{Is_Bidi_Class=:= ARABIC_NUMBER}');
Error('\P{Is_Bidi_Class=:= ARABIC_NUMBER}');
Expect(1, 69246, '\p{Is_Bidi_Class=arabicnumber}', "");
Expect(0, 69246, '\p{^Is_Bidi_Class=arabicnumber}', "");
Expect(0, 69246, '\P{Is_Bidi_Class=arabicnumber}', "");
Expect(1, 69246, '\P{^Is_Bidi_Class=arabicnumber}', "");
Expect(0, 69247, '\p{Is_Bidi_Class=arabicnumber}', "");
Expect(1, 69247, '\p{^Is_Bidi_Class=arabicnumber}', "");
Expect(1, 69247, '\P{Is_Bidi_Class=arabicnumber}', "");
Expect(0, 69247, '\P{^Is_Bidi_Class=arabicnumber}', "");
Expect(1, 69246, '\p{Is_Bidi_Class= _ARABIC_NUMBER}', "");
Expect(0, 69246, '\p{^Is_Bidi_Class= _ARABIC_NUMBER}', "");
Expect(0, 69246, '\P{Is_Bidi_Class= _ARABIC_NUMBER}', "");
Expect(1, 69246, '\P{^Is_Bidi_Class= _ARABIC_NUMBER}', "");
Expect(0, 69247, '\p{Is_Bidi_Class= _ARABIC_NUMBER}', "");
Expect(1, 69247, '\p{^Is_Bidi_Class= _ARABIC_NUMBER}', "");
Expect(1, 69247, '\P{Is_Bidi_Class= _ARABIC_NUMBER}', "");
Expect(0, 69247, '\P{^Is_Bidi_Class= _ARABIC_NUMBER}', "");
Error('\p{Is_Bc:/a/_an}');
Error('\P{Is_Bc:/a/_an}');
Expect(1, 69246, '\p{Is_Bc=an}', "");
Expect(0, 69246, '\p{^Is_Bc=an}', "");
Expect(0, 69246, '\P{Is_Bc=an}', "");
Expect(1, 69246, '\P{^Is_Bc=an}', "");
Expect(0, 69247, '\p{Is_Bc=an}', "");
Expect(1, 69247, '\p{^Is_Bc=an}', "");
Expect(1, 69247, '\P{Is_Bc=an}', "");
Expect(0, 69247, '\P{^Is_Bc=an}', "");
Expect(1, 69246, '\p{Is_Bc=_	AN}', "");
Expect(0, 69246, '\p{^Is_Bc=_	AN}', "");
Expect(0, 69246, '\P{Is_Bc=_	AN}', "");
Expect(1, 69246, '\P{^Is_Bc=_	AN}', "");
Expect(0, 69247, '\p{Is_Bc=_	AN}', "");
Expect(1, 69247, '\p{^Is_Bc=_	AN}', "");
Expect(1, 69247, '\P{Is_Bc=_	AN}', "");
Expect(0, 69247, '\P{^Is_Bc=_	AN}', "");
Error('\p{Bidi_Class= :=Paragraph_SEPARATOR}');
Error('\P{Bidi_Class= :=Paragraph_SEPARATOR}');
Expect(1, 8233, '\p{Bidi_Class=paragraphseparator}', "");
Expect(0, 8233, '\p{^Bidi_Class=paragraphseparator}', "");
Expect(0, 8233, '\P{Bidi_Class=paragraphseparator}', "");
Expect(1, 8233, '\P{^Bidi_Class=paragraphseparator}', "");
Expect(0, 8234, '\p{Bidi_Class=paragraphseparator}', "");
Expect(1, 8234, '\p{^Bidi_Class=paragraphseparator}', "");
Expect(1, 8234, '\P{Bidi_Class=paragraphseparator}', "");
Expect(0, 8234, '\P{^Bidi_Class=paragraphseparator}', "");
Expect(1, 8233, '\p{Bidi_Class=-	Paragraph_Separator}', "");
Expect(0, 8233, '\p{^Bidi_Class=-	Paragraph_Separator}', "");
Expect(0, 8233, '\P{Bidi_Class=-	Paragraph_Separator}', "");
Expect(1, 8233, '\P{^Bidi_Class=-	Paragraph_Separator}', "");
Expect(0, 8234, '\p{Bidi_Class=-	Paragraph_Separator}', "");
Expect(1, 8234, '\p{^Bidi_Class=-	Paragraph_Separator}', "");
Expect(1, 8234, '\P{Bidi_Class=-	Paragraph_Separator}', "");
Expect(0, 8234, '\P{^Bidi_Class=-	Paragraph_Separator}', "");
Error('\p{Bc=	:=B}');
Error('\P{Bc=	:=B}');
Expect(1, 8233, '\p{Bc:	b}', "");
Expect(0, 8233, '\p{^Bc:	b}', "");
Expect(0, 8233, '\P{Bc:	b}', "");
Expect(1, 8233, '\P{^Bc:	b}', "");
Expect(0, 8234, '\p{Bc:	b}', "");
Expect(1, 8234, '\p{^Bc:	b}', "");
Expect(1, 8234, '\P{Bc:	b}', "");
Expect(0, 8234, '\P{^Bc:	b}', "");
Expect(1, 8233, '\p{Bc=	 B}', "");
Expect(0, 8233, '\p{^Bc=	 B}', "");
Expect(0, 8233, '\P{Bc=	 B}', "");
Expect(1, 8233, '\P{^Bc=	 B}', "");
Expect(0, 8234, '\p{Bc=	 B}', "");
Expect(1, 8234, '\p{^Bc=	 B}', "");
Expect(1, 8234, '\P{Bc=	 B}', "");
Expect(0, 8234, '\P{^Bc=	 B}', "");
Error('\p{Is_Bidi_Class=/a/-	PARAGRAPH_separator}');
Error('\P{Is_Bidi_Class=/a/-	PARAGRAPH_separator}');
Expect(1, 8233, '\p{Is_Bidi_Class=paragraphseparator}', "");
Expect(0, 8233, '\p{^Is_Bidi_Class=paragraphseparator}', "");
Expect(0, 8233, '\P{Is_Bidi_Class=paragraphseparator}', "");
Expect(1, 8233, '\P{^Is_Bidi_Class=paragraphseparator}', "");
Expect(0, 8234, '\p{Is_Bidi_Class=paragraphseparator}', "");
Expect(1, 8234, '\p{^Is_Bidi_Class=paragraphseparator}', "");
Expect(1, 8234, '\P{Is_Bidi_Class=paragraphseparator}', "");
Expect(0, 8234, '\P{^Is_Bidi_Class=paragraphseparator}', "");
Expect(1, 8233, '\p{Is_Bidi_Class:  	Paragraph_Separator}', "");
Expect(0, 8233, '\p{^Is_Bidi_Class:  	Paragraph_Separator}', "");
Expect(0, 8233, '\P{Is_Bidi_Class:  	Paragraph_Separator}', "");
Expect(1, 8233, '\P{^Is_Bidi_Class:  	Paragraph_Separator}', "");
Expect(0, 8234, '\p{Is_Bidi_Class:  	Paragraph_Separator}', "");
Expect(1, 8234, '\p{^Is_Bidi_Class:  	Paragraph_Separator}', "");
Expect(1, 8234, '\P{Is_Bidi_Class:  	Paragraph_Separator}', "");
Expect(0, 8234, '\P{^Is_Bidi_Class:  	Paragraph_Separator}', "");
Error('\p{Is_Bc=_:=b}');
Error('\P{Is_Bc=_:=b}');
Expect(1, 8233, '\p{Is_Bc:b}', "");
Expect(0, 8233, '\p{^Is_Bc:b}', "");
Expect(0, 8233, '\P{Is_Bc:b}', "");
Expect(1, 8233, '\P{^Is_Bc:b}', "");
Expect(0, 8234, '\p{Is_Bc:b}', "");
Expect(1, 8234, '\p{^Is_Bc:b}', "");
Expect(1, 8234, '\P{Is_Bc:b}', "");
Expect(0, 8234, '\P{^Is_Bc:b}', "");
Expect(1, 8233, '\p{Is_Bc:   B}', "");
Expect(0, 8233, '\p{^Is_Bc:   B}', "");
Expect(0, 8233, '\P{Is_Bc:   B}', "");
Expect(1, 8233, '\P{^Is_Bc:   B}', "");
Expect(0, 8234, '\p{Is_Bc:   B}', "");
Expect(1, 8234, '\p{^Is_Bc:   B}', "");
Expect(1, 8234, '\P{Is_Bc:   B}', "");
Expect(0, 8234, '\P{^Is_Bc:   B}', "");
Error('\p{Bidi_Class=:=	boundary_NEUTRAL}');
Error('\P{Bidi_Class=:=	boundary_NEUTRAL}');
Expect(1, 921599, '\p{Bidi_Class=boundaryneutral}', "");
Expect(0, 921599, '\p{^Bidi_Class=boundaryneutral}', "");
Expect(0, 921599, '\P{Bidi_Class=boundaryneutral}', "");
Expect(1, 921599, '\P{^Bidi_Class=boundaryneutral}', "");
Expect(0, 1114109, '\p{Bidi_Class=boundaryneutral}', "");
Expect(1, 1114109, '\p{^Bidi_Class=boundaryneutral}', "");
Expect(1, 1114109, '\P{Bidi_Class=boundaryneutral}', "");
Expect(0, 1114109, '\P{^Bidi_Class=boundaryneutral}', "");
Expect(1, 921599, '\p{Bidi_Class= _boundary_Neutral}', "");
Expect(0, 921599, '\p{^Bidi_Class= _boundary_Neutral}', "");
Expect(0, 921599, '\P{Bidi_Class= _boundary_Neutral}', "");
Expect(1, 921599, '\P{^Bidi_Class= _boundary_Neutral}', "");
Expect(0, 1114109, '\p{Bidi_Class= _boundary_Neutral}', "");
Expect(1, 1114109, '\p{^Bidi_Class= _boundary_Neutral}', "");
Expect(1, 1114109, '\P{Bidi_Class= _boundary_Neutral}', "");
Expect(0, 1114109, '\P{^Bidi_Class= _boundary_Neutral}', "");
Error('\p{Bc::=_BN}');
Error('\P{Bc::=_BN}');
Expect(1, 921599, '\p{Bc=bn}', "");
Expect(0, 921599, '\p{^Bc=bn}', "");
Expect(0, 921599, '\P{Bc=bn}', "");
Expect(1, 921599, '\P{^Bc=bn}', "");
Expect(0, 1114109, '\p{Bc=bn}', "");
Expect(1, 1114109, '\p{^Bc=bn}', "");
Expect(1, 1114109, '\P{Bc=bn}', "");
Expect(0, 1114109, '\P{^Bc=bn}', "");
Expect(1, 921599, '\p{Bc= _BN}', "");
Expect(0, 921599, '\p{^Bc= _BN}', "");
Expect(0, 921599, '\P{Bc= _BN}', "");
Expect(1, 921599, '\P{^Bc= _BN}', "");
Expect(0, 1114109, '\p{Bc= _BN}', "");
Expect(1, 1114109, '\p{^Bc= _BN}', "");
Expect(1, 1114109, '\P{Bc= _BN}', "");
Expect(0, 1114109, '\P{^Bc= _BN}', "");
Error('\p{Is_Bidi_Class=_ Boundary_NEUTRAL/a/}');
Error('\P{Is_Bidi_Class=_ Boundary_NEUTRAL/a/}');
Expect(1, 921599, '\p{Is_Bidi_Class=boundaryneutral}', "");
Expect(0, 921599, '\p{^Is_Bidi_Class=boundaryneutral}', "");
Expect(0, 921599, '\P{Is_Bidi_Class=boundaryneutral}', "");
Expect(1, 921599, '\P{^Is_Bidi_Class=boundaryneutral}', "");
Expect(0, 1114109, '\p{Is_Bidi_Class=boundaryneutral}', "");
Expect(1, 1114109, '\p{^Is_Bidi_Class=boundaryneutral}', "");
Expect(1, 1114109, '\P{Is_Bidi_Class=boundaryneutral}', "");
Expect(0, 1114109, '\P{^Is_Bidi_Class=boundaryneutral}', "");
Expect(1, 921599, '\p{Is_Bidi_Class=	boundary_neutral}', "");
Expect(0, 921599, '\p{^Is_Bidi_Class=	boundary_neutral}', "");
Expect(0, 921599, '\P{Is_Bidi_Class=	boundary_neutral}', "");
Expect(1, 921599, '\P{^Is_Bidi_Class=	boundary_neutral}', "");
Expect(0, 1114109, '\p{Is_Bidi_Class=	boundary_neutral}', "");
Expect(1, 1114109, '\p{^Is_Bidi_Class=	boundary_neutral}', "");
Expect(1, 1114109, '\P{Is_Bidi_Class=	boundary_neutral}', "");
Expect(0, 1114109, '\P{^Is_Bidi_Class=	boundary_neutral}', "");
Error('\p{Is_Bc=-BN/a/}');
Error('\P{Is_Bc=-BN/a/}');
Expect(1, 921599, '\p{Is_Bc=bn}', "");
Expect(0, 921599, '\p{^Is_Bc=bn}', "");
Expect(0, 921599, '\P{Is_Bc=bn}', "");
Expect(1, 921599, '\P{^Is_Bc=bn}', "");
Expect(0, 1114109, '\p{Is_Bc=bn}', "");
Expect(1, 1114109, '\p{^Is_Bc=bn}', "");
Expect(1, 1114109, '\P{Is_Bc=bn}', "");
Expect(0, 1114109, '\P{^Is_Bc=bn}', "");
Expect(1, 921599, '\p{Is_Bc= -BN}', "");
Expect(0, 921599, '\p{^Is_Bc= -BN}', "");
Expect(0, 921599, '\P{Is_Bc= -BN}', "");
Expect(1, 921599, '\P{^Is_Bc= -BN}', "");
Expect(0, 1114109, '\p{Is_Bc= -BN}', "");
Expect(1, 1114109, '\p{^Is_Bc= -BN}', "");
Expect(1, 1114109, '\P{Is_Bc= -BN}', "");
Expect(0, 1114109, '\P{^Is_Bc= -BN}', "");
Error('\p{Bidi_Class=_:=Common_Separator}');
Error('\P{Bidi_Class=_:=Common_Separator}');
Expect(1, 65306, '\p{Bidi_Class=commonseparator}', "");
Expect(0, 65306, '\p{^Bidi_Class=commonseparator}', "");
Expect(0, 65306, '\P{Bidi_Class=commonseparator}', "");
Expect(1, 65306, '\P{^Bidi_Class=commonseparator}', "");
Expect(0, 65307, '\p{Bidi_Class=commonseparator}', "");
Expect(1, 65307, '\p{^Bidi_Class=commonseparator}', "");
Expect(1, 65307, '\P{Bidi_Class=commonseparator}', "");
Expect(0, 65307, '\P{^Bidi_Class=commonseparator}', "");
Expect(1, 65306, '\p{Bidi_Class=_COMMON_Separator}', "");
Expect(0, 65306, '\p{^Bidi_Class=_COMMON_Separator}', "");
Expect(0, 65306, '\P{Bidi_Class=_COMMON_Separator}', "");
Expect(1, 65306, '\P{^Bidi_Class=_COMMON_Separator}', "");
Expect(0, 65307, '\p{Bidi_Class=_COMMON_Separator}', "");
Expect(1, 65307, '\p{^Bidi_Class=_COMMON_Separator}', "");
Expect(1, 65307, '\P{Bidi_Class=_COMMON_Separator}', "");
Expect(0, 65307, '\P{^Bidi_Class=_COMMON_Separator}', "");
Error('\p{Bc= 	cs/a/}');
Error('\P{Bc= 	cs/a/}');
Expect(1, 65306, '\p{Bc=cs}', "");
Expect(0, 65306, '\p{^Bc=cs}', "");
Expect(0, 65306, '\P{Bc=cs}', "");
Expect(1, 65306, '\P{^Bc=cs}', "");
Expect(0, 65307, '\p{Bc=cs}', "");
Expect(1, 65307, '\p{^Bc=cs}', "");
Expect(1, 65307, '\P{Bc=cs}', "");
Expect(0, 65307, '\P{^Bc=cs}', "");
Expect(1, 65306, '\p{Bc=_cs}', "");
Expect(0, 65306, '\p{^Bc=_cs}', "");
Expect(0, 65306, '\P{Bc=_cs}', "");
Expect(1, 65306, '\P{^Bc=_cs}', "");
Expect(0, 65307, '\p{Bc=_cs}', "");
Expect(1, 65307, '\p{^Bc=_cs}', "");
Expect(1, 65307, '\P{Bc=_cs}', "");
Expect(0, 65307, '\P{^Bc=_cs}', "");
Error('\p{Is_Bidi_Class=/a/-	common_Separator}');
Error('\P{Is_Bidi_Class=/a/-	common_Separator}');
Expect(1, 65306, '\p{Is_Bidi_Class=commonseparator}', "");
Expect(0, 65306, '\p{^Is_Bidi_Class=commonseparator}', "");
Expect(0, 65306, '\P{Is_Bidi_Class=commonseparator}', "");
Expect(1, 65306, '\P{^Is_Bidi_Class=commonseparator}', "");
Expect(0, 65307, '\p{Is_Bidi_Class=commonseparator}', "");
Expect(1, 65307, '\p{^Is_Bidi_Class=commonseparator}', "");
Expect(1, 65307, '\P{Is_Bidi_Class=commonseparator}', "");
Expect(0, 65307, '\P{^Is_Bidi_Class=commonseparator}', "");
Expect(1, 65306, '\p{Is_Bidi_Class=_ COMMON_SEPARATOR}', "");
Expect(0, 65306, '\p{^Is_Bidi_Class=_ COMMON_SEPARATOR}', "");
Expect(0, 65306, '\P{Is_Bidi_Class=_ COMMON_SEPARATOR}', "");
Expect(1, 65306, '\P{^Is_Bidi_Class=_ COMMON_SEPARATOR}', "");
Expect(0, 65307, '\p{Is_Bidi_Class=_ COMMON_SEPARATOR}', "");
Expect(1, 65307, '\p{^Is_Bidi_Class=_ COMMON_SEPARATOR}', "");
Expect(1, 65307, '\P{Is_Bidi_Class=_ COMMON_SEPARATOR}', "");
Expect(0, 65307, '\P{^Is_Bidi_Class=_ COMMON_SEPARATOR}', "");
Error('\p{Is_Bc= :=CS}');
Error('\P{Is_Bc= :=CS}');
Expect(1, 65306, '\p{Is_Bc=cs}', "");
Expect(0, 65306, '\p{^Is_Bc=cs}', "");
Expect(0, 65306, '\P{Is_Bc=cs}', "");
Expect(1, 65306, '\P{^Is_Bc=cs}', "");
Expect(0, 65307, '\p{Is_Bc=cs}', "");
Expect(1, 65307, '\p{^Is_Bc=cs}', "");
Expect(1, 65307, '\P{Is_Bc=cs}', "");
Expect(0, 65307, '\P{^Is_Bc=cs}', "");
Expect(1, 65306, '\p{Is_Bc= 	CS}', "");
Expect(0, 65306, '\p{^Is_Bc= 	CS}', "");
Expect(0, 65306, '\P{Is_Bc= 	CS}', "");
Expect(1, 65306, '\P{^Is_Bc= 	CS}', "");
Expect(0, 65307, '\p{Is_Bc= 	CS}', "");
Expect(1, 65307, '\p{^Is_Bc= 	CS}', "");
Expect(1, 65307, '\P{Is_Bc= 	CS}', "");
Expect(0, 65307, '\P{^Is_Bc= 	CS}', "");
Error('\p{Bidi_Class=	 european_number:=}');
Error('\P{Bidi_Class=	 european_number:=}');
Expect(1, 127242, '\p{Bidi_Class=europeannumber}', "");
Expect(0, 127242, '\p{^Bidi_Class=europeannumber}', "");
Expect(0, 127242, '\P{Bidi_Class=europeannumber}', "");
Expect(1, 127242, '\P{^Bidi_Class=europeannumber}', "");
Expect(0, 127243, '\p{Bidi_Class=europeannumber}', "");
Expect(1, 127243, '\p{^Bidi_Class=europeannumber}', "");
Expect(1, 127243, '\P{Bidi_Class=europeannumber}', "");
Expect(0, 127243, '\P{^Bidi_Class=europeannumber}', "");
Expect(1, 127242, '\p{Bidi_Class=European_number}', "");
Expect(0, 127242, '\p{^Bidi_Class=European_number}', "");
Expect(0, 127242, '\P{Bidi_Class=European_number}', "");
Expect(1, 127242, '\P{^Bidi_Class=European_number}', "");
Expect(0, 127243, '\p{Bidi_Class=European_number}', "");
Expect(1, 127243, '\p{^Bidi_Class=European_number}', "");
Expect(1, 127243, '\P{Bidi_Class=European_number}', "");
Expect(0, 127243, '\P{^Bidi_Class=European_number}', "");
Error('\p{Bc=/a/  en}');
Error('\P{Bc=/a/  en}');
Expect(1, 127242, '\p{Bc=en}', "");
Expect(0, 127242, '\p{^Bc=en}', "");
Expect(0, 127242, '\P{Bc=en}', "");
Expect(1, 127242, '\P{^Bc=en}', "");
Expect(0, 127243, '\p{Bc=en}', "");
Expect(1, 127243, '\p{^Bc=en}', "");
Expect(1, 127243, '\P{Bc=en}', "");
Expect(0, 127243, '\P{^Bc=en}', "");
Expect(1, 127242, '\p{Bc=		EN}', "");
Expect(0, 127242, '\p{^Bc=		EN}', "");
Expect(0, 127242, '\P{Bc=		EN}', "");
Expect(1, 127242, '\P{^Bc=		EN}', "");
Expect(0, 127243, '\p{Bc=		EN}', "");
Expect(1, 127243, '\p{^Bc=		EN}', "");
Expect(1, 127243, '\P{Bc=		EN}', "");
Expect(0, 127243, '\P{^Bc=		EN}', "");
Error('\p{Is_Bidi_Class=:=EUROPEAN_NUMBER}');
Error('\P{Is_Bidi_Class=:=EUROPEAN_NUMBER}');
Expect(1, 127242, '\p{Is_Bidi_Class:   europeannumber}', "");
Expect(0, 127242, '\p{^Is_Bidi_Class:   europeannumber}', "");
Expect(0, 127242, '\P{Is_Bidi_Class:   europeannumber}', "");
Expect(1, 127242, '\P{^Is_Bidi_Class:   europeannumber}', "");
Expect(0, 127243, '\p{Is_Bidi_Class:   europeannumber}', "");
Expect(1, 127243, '\p{^Is_Bidi_Class:   europeannumber}', "");
Expect(1, 127243, '\P{Is_Bidi_Class:   europeannumber}', "");
Expect(0, 127243, '\P{^Is_Bidi_Class:   europeannumber}', "");
Expect(1, 127242, '\p{Is_Bidi_Class=_	European_Number}', "");
Expect(0, 127242, '\p{^Is_Bidi_Class=_	European_Number}', "");
Expect(0, 127242, '\P{Is_Bidi_Class=_	European_Number}', "");
Expect(1, 127242, '\P{^Is_Bidi_Class=_	European_Number}', "");
Expect(0, 127243, '\p{Is_Bidi_Class=_	European_Number}', "");
Expect(1, 127243, '\p{^Is_Bidi_Class=_	European_Number}', "");
Expect(1, 127243, '\P{Is_Bidi_Class=_	European_Number}', "");
Expect(0, 127243, '\P{^Is_Bidi_Class=_	European_Number}', "");
Error('\p{Is_Bc=_:=EN}');
Error('\P{Is_Bc=_:=EN}');
Expect(1, 127242, '\p{Is_Bc=en}', "");
Expect(0, 127242, '\p{^Is_Bc=en}', "");
Expect(0, 127242, '\P{Is_Bc=en}', "");
Expect(1, 127242, '\P{^Is_Bc=en}', "");
Expect(0, 127243, '\p{Is_Bc=en}', "");
Expect(1, 127243, '\p{^Is_Bc=en}', "");
Expect(1, 127243, '\P{Is_Bc=en}', "");
Expect(0, 127243, '\P{^Is_Bc=en}', "");
Expect(1, 127242, '\p{Is_Bc=_en}', "");
Expect(0, 127242, '\p{^Is_Bc=_en}', "");
Expect(0, 127242, '\P{Is_Bc=_en}', "");
Expect(1, 127242, '\P{^Is_Bc=_en}', "");
Expect(0, 127243, '\p{Is_Bc=_en}', "");
Expect(1, 127243, '\p{^Is_Bc=_en}', "");
Expect(1, 127243, '\P{Is_Bc=_en}', "");
Expect(0, 127243, '\P{^Is_Bc=_en}', "");
Error('\p{Bidi_Class= :=European_SEPARATOR}');
Error('\P{Bidi_Class= :=European_SEPARATOR}');
Expect(1, 65293, '\p{Bidi_Class:   europeanseparator}', "");
Expect(0, 65293, '\p{^Bidi_Class:   europeanseparator}', "");
Expect(0, 65293, '\P{Bidi_Class:   europeanseparator}', "");
Expect(1, 65293, '\P{^Bidi_Class:   europeanseparator}', "");
Expect(0, 65294, '\p{Bidi_Class:   europeanseparator}', "");
Expect(1, 65294, '\p{^Bidi_Class:   europeanseparator}', "");
Expect(1, 65294, '\P{Bidi_Class:   europeanseparator}', "");
Expect(0, 65294, '\P{^Bidi_Class:   europeanseparator}', "");
Expect(1, 65293, '\p{Bidi_Class= -European_Separator}', "");
Expect(0, 65293, '\p{^Bidi_Class= -European_Separator}', "");
Expect(0, 65293, '\P{Bidi_Class= -European_Separator}', "");
Expect(1, 65293, '\P{^Bidi_Class= -European_Separator}', "");
Expect(0, 65294, '\p{Bidi_Class= -European_Separator}', "");
Expect(1, 65294, '\p{^Bidi_Class= -European_Separator}', "");
Expect(1, 65294, '\P{Bidi_Class= -European_Separator}', "");
Expect(0, 65294, '\P{^Bidi_Class= -European_Separator}', "");
Error('\p{Bc=:= ES}');
Error('\P{Bc=:= ES}');
Expect(1, 65293, '\p{Bc:es}', "");
Expect(0, 65293, '\p{^Bc:es}', "");
Expect(0, 65293, '\P{Bc:es}', "");
Expect(1, 65293, '\P{^Bc:es}', "");
Expect(0, 65294, '\p{Bc:es}', "");
Expect(1, 65294, '\p{^Bc:es}', "");
Expect(1, 65294, '\P{Bc:es}', "");
Expect(0, 65294, '\P{^Bc:es}', "");
Expect(1, 65293, '\p{Bc=_ES}', "");
Expect(0, 65293, '\p{^Bc=_ES}', "");
Expect(0, 65293, '\P{Bc=_ES}', "");
Expect(1, 65293, '\P{^Bc=_ES}', "");
Expect(0, 65294, '\p{Bc=_ES}', "");
Expect(1, 65294, '\p{^Bc=_ES}', "");
Expect(1, 65294, '\P{Bc=_ES}', "");
Expect(0, 65294, '\P{^Bc=_ES}', "");
Error('\p{Is_Bidi_Class= -european_SEPARATOR:=}');
Error('\P{Is_Bidi_Class= -european_SEPARATOR:=}');
Expect(1, 65293, '\p{Is_Bidi_Class=europeanseparator}', "");
Expect(0, 65293, '\p{^Is_Bidi_Class=europeanseparator}', "");
Expect(0, 65293, '\P{Is_Bidi_Class=europeanseparator}', "");
Expect(1, 65293, '\P{^Is_Bidi_Class=europeanseparator}', "");
Expect(0, 65294, '\p{Is_Bidi_Class=europeanseparator}', "");
Expect(1, 65294, '\p{^Is_Bidi_Class=europeanseparator}', "");
Expect(1, 65294, '\P{Is_Bidi_Class=europeanseparator}', "");
Expect(0, 65294, '\P{^Is_Bidi_Class=europeanseparator}', "");
Expect(1, 65293, '\p{Is_Bidi_Class=- European_separator}', "");
Expect(0, 65293, '\p{^Is_Bidi_Class=- European_separator}', "");
Expect(0, 65293, '\P{Is_Bidi_Class=- European_separator}', "");
Expect(1, 65293, '\P{^Is_Bidi_Class=- European_separator}', "");
Expect(0, 65294, '\p{Is_Bidi_Class=- European_separator}', "");
Expect(1, 65294, '\p{^Is_Bidi_Class=- European_separator}', "");
Expect(1, 65294, '\P{Is_Bidi_Class=- European_separator}', "");
Expect(0, 65294, '\P{^Is_Bidi_Class=- European_separator}', "");
Error('\p{Is_Bc=	/a/ES}');
Error('\P{Is_Bc=	/a/ES}');
Expect(1, 65293, '\p{Is_Bc=es}', "");
Expect(0, 65293, '\p{^Is_Bc=es}', "");
Expect(0, 65293, '\P{Is_Bc=es}', "");
Expect(1, 65293, '\P{^Is_Bc=es}', "");
Expect(0, 65294, '\p{Is_Bc=es}', "");
Expect(1, 65294, '\p{^Is_Bc=es}', "");
Expect(1, 65294, '\P{Is_Bc=es}', "");
Expect(0, 65294, '\P{^Is_Bc=es}', "");
Expect(1, 65293, '\p{Is_Bc:-ES}', "");
Expect(0, 65293, '\p{^Is_Bc:-ES}', "");
Expect(0, 65293, '\P{Is_Bc:-ES}', "");
Expect(1, 65293, '\P{^Is_Bc:-ES}', "");
Expect(0, 65294, '\p{Is_Bc:-ES}', "");
Expect(1, 65294, '\p{^Is_Bc:-ES}', "");
Expect(1, 65294, '\P{Is_Bc:-ES}', "");
Expect(0, 65294, '\P{^Is_Bc:-ES}', "");
Error('\p{Bidi_Class:   _:=EUROPEAN_TERMINATOR}');
Error('\P{Bidi_Class:   _:=EUROPEAN_TERMINATOR}');
Expect(1, 65510, '\p{Bidi_Class:europeanterminator}', "");
Expect(0, 65510, '\p{^Bidi_Class:europeanterminator}', "");
Expect(0, 65510, '\P{Bidi_Class:europeanterminator}', "");
Expect(1, 65510, '\P{^Bidi_Class:europeanterminator}', "");
Expect(0, 65511, '\p{Bidi_Class:europeanterminator}', "");
Expect(1, 65511, '\p{^Bidi_Class:europeanterminator}', "");
Expect(1, 65511, '\P{Bidi_Class:europeanterminator}', "");
Expect(0, 65511, '\P{^Bidi_Class:europeanterminator}', "");
Expect(1, 65510, '\p{Bidi_Class=-European_TERMINATOR}', "");
Expect(0, 65510, '\p{^Bidi_Class=-European_TERMINATOR}', "");
Expect(0, 65510, '\P{Bidi_Class=-European_TERMINATOR}', "");
Expect(1, 65510, '\P{^Bidi_Class=-European_TERMINATOR}', "");
Expect(0, 65511, '\p{Bidi_Class=-European_TERMINATOR}', "");
Expect(1, 65511, '\p{^Bidi_Class=-European_TERMINATOR}', "");
Expect(1, 65511, '\P{Bidi_Class=-European_TERMINATOR}', "");
Expect(0, 65511, '\P{^Bidi_Class=-European_TERMINATOR}', "");
Error('\p{Bc=ET/a/}');
Error('\P{Bc=ET/a/}');
Expect(1, 65510, '\p{Bc=et}', "");
Expect(0, 65510, '\p{^Bc=et}', "");
Expect(0, 65510, '\P{Bc=et}', "");
Expect(1, 65510, '\P{^Bc=et}', "");
Expect(0, 65511, '\p{Bc=et}', "");
Expect(1, 65511, '\p{^Bc=et}', "");
Expect(1, 65511, '\P{Bc=et}', "");
Expect(0, 65511, '\P{^Bc=et}', "");
Expect(1, 65510, '\p{Bc= ET}', "");
Expect(0, 65510, '\p{^Bc= ET}', "");
Expect(0, 65510, '\P{Bc= ET}', "");
Expect(1, 65510, '\P{^Bc= ET}', "");
Expect(0, 65511, '\p{Bc= ET}', "");
Expect(1, 65511, '\p{^Bc= ET}', "");
Expect(1, 65511, '\P{Bc= ET}', "");
Expect(0, 65511, '\P{^Bc= ET}', "");
Error('\p{Is_Bidi_Class= /a/european_Terminator}');
Error('\P{Is_Bidi_Class= /a/european_Terminator}');
Expect(1, 65510, '\p{Is_Bidi_Class=europeanterminator}', "");
Expect(0, 65510, '\p{^Is_Bidi_Class=europeanterminator}', "");
Expect(0, 65510, '\P{Is_Bidi_Class=europeanterminator}', "");
Expect(1, 65510, '\P{^Is_Bidi_Class=europeanterminator}', "");
Expect(0, 65511, '\p{Is_Bidi_Class=europeanterminator}', "");
Expect(1, 65511, '\p{^Is_Bidi_Class=europeanterminator}', "");
Expect(1, 65511, '\P{Is_Bidi_Class=europeanterminator}', "");
Expect(0, 65511, '\P{^Is_Bidi_Class=europeanterminator}', "");
Expect(1, 65510, '\p{Is_Bidi_Class=-	european_Terminator}', "");
Expect(0, 65510, '\p{^Is_Bidi_Class=-	european_Terminator}', "");
Expect(0, 65510, '\P{Is_Bidi_Class=-	european_Terminator}', "");
Expect(1, 65510, '\P{^Is_Bidi_Class=-	european_Terminator}', "");
Expect(0, 65511, '\p{Is_Bidi_Class=-	european_Terminator}', "");
Expect(1, 65511, '\p{^Is_Bidi_Class=-	european_Terminator}', "");
Expect(1, 65511, '\P{Is_Bidi_Class=-	european_Terminator}', "");
Expect(0, 65511, '\P{^Is_Bidi_Class=-	european_Terminator}', "");
Error('\p{Is_Bc=:=et}');
Error('\P{Is_Bc=:=et}');
Expect(1, 65510, '\p{Is_Bc=et}', "");
Expect(0, 65510, '\p{^Is_Bc=et}', "");
Expect(0, 65510, '\P{Is_Bc=et}', "");
Expect(1, 65510, '\P{^Is_Bc=et}', "");
Expect(0, 65511, '\p{Is_Bc=et}', "");
Expect(1, 65511, '\p{^Is_Bc=et}', "");
Expect(1, 65511, '\P{Is_Bc=et}', "");
Expect(0, 65511, '\P{^Is_Bc=et}', "");
Expect(1, 65510, '\p{Is_Bc=	_ET}', "");
Expect(0, 65510, '\p{^Is_Bc=	_ET}', "");
Expect(0, 65510, '\P{Is_Bc=	_ET}', "");
Expect(1, 65510, '\P{^Is_Bc=	_ET}', "");
Expect(0, 65511, '\p{Is_Bc=	_ET}', "");
Expect(1, 65511, '\p{^Is_Bc=	_ET}', "");
Expect(1, 65511, '\P{Is_Bc=	_ET}', "");
Expect(0, 65511, '\P{^Is_Bc=	_ET}', "");
Error('\p{Bidi_Class=/a/	_First_Strong_isolate}');
Error('\P{Bidi_Class=/a/	_First_Strong_isolate}');
Expect(1, 8296, '\p{Bidi_Class=firststrongisolate}', "");
Expect(0, 8296, '\p{^Bidi_Class=firststrongisolate}', "");
Expect(0, 8296, '\P{Bidi_Class=firststrongisolate}', "");
Expect(1, 8296, '\P{^Bidi_Class=firststrongisolate}', "");
Expect(0, 8297, '\p{Bidi_Class=firststrongisolate}', "");
Expect(1, 8297, '\p{^Bidi_Class=firststrongisolate}', "");
Expect(1, 8297, '\P{Bidi_Class=firststrongisolate}', "");
Expect(0, 8297, '\P{^Bidi_Class=firststrongisolate}', "");
Expect(1, 8296, '\p{Bidi_Class=	FIRST_Strong_ISOLATE}', "");
Expect(0, 8296, '\p{^Bidi_Class=	FIRST_Strong_ISOLATE}', "");
Expect(0, 8296, '\P{Bidi_Class=	FIRST_Strong_ISOLATE}', "");
Expect(1, 8296, '\P{^Bidi_Class=	FIRST_Strong_ISOLATE}', "");
Expect(0, 8297, '\p{Bidi_Class=	FIRST_Strong_ISOLATE}', "");
Expect(1, 8297, '\p{^Bidi_Class=	FIRST_Strong_ISOLATE}', "");
Expect(1, 8297, '\P{Bidi_Class=	FIRST_Strong_ISOLATE}', "");
Expect(0, 8297, '\P{^Bidi_Class=	FIRST_Strong_ISOLATE}', "");
Error('\p{Bc:_:=FSI}');
Error('\P{Bc:_:=FSI}');
Expect(1, 8296, '\p{Bc=fsi}', "");
Expect(0, 8296, '\p{^Bc=fsi}', "");
Expect(0, 8296, '\P{Bc=fsi}', "");
Expect(1, 8296, '\P{^Bc=fsi}', "");
Expect(0, 8297, '\p{Bc=fsi}', "");
Expect(1, 8297, '\p{^Bc=fsi}', "");
Expect(1, 8297, '\P{Bc=fsi}', "");
Expect(0, 8297, '\P{^Bc=fsi}', "");
Expect(1, 8296, '\p{Bc=  FSI}', "");
Expect(0, 8296, '\p{^Bc=  FSI}', "");
Expect(0, 8296, '\P{Bc=  FSI}', "");
Expect(1, 8296, '\P{^Bc=  FSI}', "");
Expect(0, 8297, '\p{Bc=  FSI}', "");
Expect(1, 8297, '\p{^Bc=  FSI}', "");
Expect(1, 8297, '\P{Bc=  FSI}', "");
Expect(0, 8297, '\P{^Bc=  FSI}', "");
Error('\p{Is_Bidi_Class= /a/first_STRONG_ISOLATE}');
Error('\P{Is_Bidi_Class= /a/first_STRONG_ISOLATE}');
Expect(1, 8296, '\p{Is_Bidi_Class=firststrongisolate}', "");
Expect(0, 8296, '\p{^Is_Bidi_Class=firststrongisolate}', "");
Expect(0, 8296, '\P{Is_Bidi_Class=firststrongisolate}', "");
Expect(1, 8296, '\P{^Is_Bidi_Class=firststrongisolate}', "");
Expect(0, 8297, '\p{Is_Bidi_Class=firststrongisolate}', "");
Expect(1, 8297, '\p{^Is_Bidi_Class=firststrongisolate}', "");
Expect(1, 8297, '\P{Is_Bidi_Class=firststrongisolate}', "");
Expect(0, 8297, '\P{^Is_Bidi_Class=firststrongisolate}', "");
Expect(1, 8296, '\p{Is_Bidi_Class=	_First_STRONG_Isolate}', "");
Expect(0, 8296, '\p{^Is_Bidi_Class=	_First_STRONG_Isolate}', "");
Expect(0, 8296, '\P{Is_Bidi_Class=	_First_STRONG_Isolate}', "");
Expect(1, 8296, '\P{^Is_Bidi_Class=	_First_STRONG_Isolate}', "");
Expect(0, 8297, '\p{Is_Bidi_Class=	_First_STRONG_Isolate}', "");
Expect(1, 8297, '\p{^Is_Bidi_Class=	_First_STRONG_Isolate}', "");
Expect(1, 8297, '\P{Is_Bidi_Class=	_First_STRONG_Isolate}', "");
Expect(0, 8297, '\P{^Is_Bidi_Class=	_First_STRONG_Isolate}', "");
Error('\p{Is_Bc=	:=fsi}');
Error('\P{Is_Bc=	:=fsi}');
Expect(1, 8296, '\p{Is_Bc=fsi}', "");
Expect(0, 8296, '\p{^Is_Bc=fsi}', "");
Expect(0, 8296, '\P{Is_Bc=fsi}', "");
Expect(1, 8296, '\P{^Is_Bc=fsi}', "");
Expect(0, 8297, '\p{Is_Bc=fsi}', "");
Expect(1, 8297, '\p{^Is_Bc=fsi}', "");
Expect(1, 8297, '\P{Is_Bc=fsi}', "");
Expect(0, 8297, '\P{^Is_Bc=fsi}', "");
Expect(1, 8296, '\p{Is_Bc=--fsi}', "");
Expect(0, 8296, '\p{^Is_Bc=--fsi}', "");
Expect(0, 8296, '\P{Is_Bc=--fsi}', "");
Expect(1, 8296, '\P{^Is_Bc=--fsi}', "");
Expect(0, 8297, '\p{Is_Bc=--fsi}', "");
Expect(1, 8297, '\p{^Is_Bc=--fsi}', "");
Expect(1, 8297, '\P{Is_Bc=--fsi}', "");
Expect(0, 8297, '\P{^Is_Bc=--fsi}', "");
Error('\p{Bidi_Class=/a/	-LEFT_To_Right}');
Error('\P{Bidi_Class=/a/	-LEFT_To_Right}');
Expect(1, 1114109, '\p{Bidi_Class=lefttoright}', "");
Expect(0, 1114109, '\p{^Bidi_Class=lefttoright}', "");
Expect(0, 1114109, '\P{Bidi_Class=lefttoright}', "");
Expect(1, 1114109, '\P{^Bidi_Class=lefttoright}', "");
Expect(0, 921599, '\p{Bidi_Class=lefttoright}', "");
Expect(1, 921599, '\p{^Bidi_Class=lefttoright}', "");
Expect(1, 921599, '\P{Bidi_Class=lefttoright}', "");
Expect(0, 921599, '\P{^Bidi_Class=lefttoright}', "");
Expect(1, 1114109, '\p{Bidi_Class=	_left_to_Right}', "");
Expect(0, 1114109, '\p{^Bidi_Class=	_left_to_Right}', "");
Expect(0, 1114109, '\P{Bidi_Class=	_left_to_Right}', "");
Expect(1, 1114109, '\P{^Bidi_Class=	_left_to_Right}', "");
Expect(0, 921599, '\p{Bidi_Class=	_left_to_Right}', "");
Expect(1, 921599, '\p{^Bidi_Class=	_left_to_Right}', "");
Expect(1, 921599, '\P{Bidi_Class=	_left_to_Right}', "");
Expect(0, 921599, '\P{^Bidi_Class=	_left_to_Right}', "");
Error('\p{Bc=_l/a/}');
Error('\P{Bc=_l/a/}');
Expect(1, 1114109, '\p{Bc:l}', "");
Expect(0, 1114109, '\p{^Bc:l}', "");
Expect(0, 1114109, '\P{Bc:l}', "");
Expect(1, 1114109, '\P{^Bc:l}', "");
Expect(0, 921599, '\p{Bc:l}', "");
Expect(1, 921599, '\p{^Bc:l}', "");
Expect(1, 921599, '\P{Bc:l}', "");
Expect(0, 921599, '\P{^Bc:l}', "");
Expect(1, 1114109, '\p{Bc=__L}', "");
Expect(0, 1114109, '\p{^Bc=__L}', "");
Expect(0, 1114109, '\P{Bc=__L}', "");
Expect(1, 1114109, '\P{^Bc=__L}', "");
Expect(0, 921599, '\p{Bc=__L}', "");
Expect(1, 921599, '\p{^Bc=__L}', "");
Expect(1, 921599, '\P{Bc=__L}', "");
Expect(0, 921599, '\P{^Bc=__L}', "");
Error('\p{Is_Bidi_Class=:=-Left_to_right}');
Error('\P{Is_Bidi_Class=:=-Left_to_right}');
Expect(1, 1114109, '\p{Is_Bidi_Class=lefttoright}', "");
Expect(0, 1114109, '\p{^Is_Bidi_Class=lefttoright}', "");
Expect(0, 1114109, '\P{Is_Bidi_Class=lefttoright}', "");
Expect(1, 1114109, '\P{^Is_Bidi_Class=lefttoright}', "");
Expect(0, 921599, '\p{Is_Bidi_Class=lefttoright}', "");
Expect(1, 921599, '\p{^Is_Bidi_Class=lefttoright}', "");
Expect(1, 921599, '\P{Is_Bidi_Class=lefttoright}', "");
Expect(0, 921599, '\P{^Is_Bidi_Class=lefttoright}', "");
Expect(1, 1114109, '\p{Is_Bidi_Class=	_LEFT_To_right}', "");
Expect(0, 1114109, '\p{^Is_Bidi_Class=	_LEFT_To_right}', "");
Expect(0, 1114109, '\P{Is_Bidi_Class=	_LEFT_To_right}', "");
Expect(1, 1114109, '\P{^Is_Bidi_Class=	_LEFT_To_right}', "");
Expect(0, 921599, '\p{Is_Bidi_Class=	_LEFT_To_right}', "");
Expect(1, 921599, '\p{^Is_Bidi_Class=	_LEFT_To_right}', "");
Expect(1, 921599, '\P{Is_Bidi_Class=	_LEFT_To_right}', "");
Expect(0, 921599, '\P{^Is_Bidi_Class=	_LEFT_To_right}', "");
Error('\p{Is_Bc:	:=_L}');
Error('\P{Is_Bc:	:=_L}');
Expect(1, 1114109, '\p{Is_Bc=l}', "");
Expect(0, 1114109, '\p{^Is_Bc=l}', "");
Expect(0, 1114109, '\P{Is_Bc=l}', "");
Expect(1, 1114109, '\P{^Is_Bc=l}', "");
Expect(0, 921599, '\p{Is_Bc=l}', "");
Expect(1, 921599, '\p{^Is_Bc=l}', "");
Expect(1, 921599, '\P{Is_Bc=l}', "");
Expect(0, 921599, '\P{^Is_Bc=l}', "");
Expect(1, 1114109, '\p{Is_Bc=	 l}', "");
Expect(0, 1114109, '\p{^Is_Bc=	 l}', "");
Expect(0, 1114109, '\P{Is_Bc=	 l}', "");
Expect(1, 1114109, '\P{^Is_Bc=	 l}', "");
Expect(0, 921599, '\p{Is_Bc=	 l}', "");
Expect(1, 921599, '\p{^Is_Bc=	 l}', "");
Expect(1, 921599, '\P{Is_Bc=	 l}', "");
Expect(0, 921599, '\P{^Is_Bc=	 l}', "");
Error('\p{Bidi_Class=:= Left_To_Right_embedding}');
Error('\P{Bidi_Class=:= Left_To_Right_embedding}');
Expect(1, 8234, '\p{Bidi_Class=lefttorightembedding}', "");
Expect(0, 8234, '\p{^Bidi_Class=lefttorightembedding}', "");
Expect(0, 8234, '\P{Bidi_Class=lefttorightembedding}', "");
Expect(1, 8234, '\P{^Bidi_Class=lefttorightembedding}', "");
Expect(0, 8235, '\p{Bidi_Class=lefttorightembedding}', "");
Expect(1, 8235, '\p{^Bidi_Class=lefttorightembedding}', "");
Expect(1, 8235, '\P{Bidi_Class=lefttorightembedding}', "");
Expect(0, 8235, '\P{^Bidi_Class=lefttorightembedding}', "");
Expect(1, 8234, '\p{Bidi_Class= Left_To_RIGHT_Embedding}', "");
Expect(0, 8234, '\p{^Bidi_Class= Left_To_RIGHT_Embedding}', "");
Expect(0, 8234, '\P{Bidi_Class= Left_To_RIGHT_Embedding}', "");
Expect(1, 8234, '\P{^Bidi_Class= Left_To_RIGHT_Embedding}', "");
Expect(0, 8235, '\p{Bidi_Class= Left_To_RIGHT_Embedding}', "");
Expect(1, 8235, '\p{^Bidi_Class= Left_To_RIGHT_Embedding}', "");
Expect(1, 8235, '\P{Bidi_Class= Left_To_RIGHT_Embedding}', "");
Expect(0, 8235, '\P{^Bidi_Class= Left_To_RIGHT_Embedding}', "");
Error('\p{Bc=:=_LRE}');
Error('\P{Bc=:=_LRE}');
Expect(1, 8234, '\p{Bc=lre}', "");
Expect(0, 8234, '\p{^Bc=lre}', "");
Expect(0, 8234, '\P{Bc=lre}', "");
Expect(1, 8234, '\P{^Bc=lre}', "");
Expect(0, 8235, '\p{Bc=lre}', "");
Expect(1, 8235, '\p{^Bc=lre}', "");
Expect(1, 8235, '\P{Bc=lre}', "");
Expect(0, 8235, '\P{^Bc=lre}', "");
Expect(1, 8234, '\p{Bc: 	lre}', "");
Expect(0, 8234, '\p{^Bc: 	lre}', "");
Expect(0, 8234, '\P{Bc: 	lre}', "");
Expect(1, 8234, '\P{^Bc: 	lre}', "");
Expect(0, 8235, '\p{Bc: 	lre}', "");
Expect(1, 8235, '\p{^Bc: 	lre}', "");
Expect(1, 8235, '\P{Bc: 	lre}', "");
Expect(0, 8235, '\P{^Bc: 	lre}', "");
Error('\p{Is_Bidi_Class=  Left_To_Right_embedding:=}');
Error('\P{Is_Bidi_Class=  Left_To_Right_embedding:=}');
Expect(1, 8234, '\p{Is_Bidi_Class=lefttorightembedding}', "");
Expect(0, 8234, '\p{^Is_Bidi_Class=lefttorightembedding}', "");
Expect(0, 8234, '\P{Is_Bidi_Class=lefttorightembedding}', "");
Expect(1, 8234, '\P{^Is_Bidi_Class=lefttorightembedding}', "");
Expect(0, 8235, '\p{Is_Bidi_Class=lefttorightembedding}', "");
Expect(1, 8235, '\p{^Is_Bidi_Class=lefttorightembedding}', "");
Expect(1, 8235, '\P{Is_Bidi_Class=lefttorightembedding}', "");
Expect(0, 8235, '\P{^Is_Bidi_Class=lefttorightembedding}', "");
Expect(1, 8234, '\p{Is_Bidi_Class=-left_To_Right_Embedding}', "");
Expect(0, 8234, '\p{^Is_Bidi_Class=-left_To_Right_Embedding}', "");
Expect(0, 8234, '\P{Is_Bidi_Class=-left_To_Right_Embedding}', "");
Expect(1, 8234, '\P{^Is_Bidi_Class=-left_To_Right_Embedding}', "");
Expect(0, 8235, '\p{Is_Bidi_Class=-left_To_Right_Embedding}', "");
Expect(1, 8235, '\p{^Is_Bidi_Class=-left_To_Right_Embedding}', "");
Expect(1, 8235, '\P{Is_Bidi_Class=-left_To_Right_Embedding}', "");
Expect(0, 8235, '\P{^Is_Bidi_Class=-left_To_Right_Embedding}', "");
Error('\p{Is_Bc=_/a/lre}');
Error('\P{Is_Bc=_/a/lre}');
Expect(1, 8234, '\p{Is_Bc=lre}', "");
Expect(0, 8234, '\p{^Is_Bc=lre}', "");
Expect(0, 8234, '\P{Is_Bc=lre}', "");
Expect(1, 8234, '\P{^Is_Bc=lre}', "");
Expect(0, 8235, '\p{Is_Bc=lre}', "");
Expect(1, 8235, '\p{^Is_Bc=lre}', "");
Expect(1, 8235, '\P{Is_Bc=lre}', "");
Expect(0, 8235, '\P{^Is_Bc=lre}', "");
Expect(1, 8234, '\p{Is_Bc= lre}', "");
Expect(0, 8234, '\p{^Is_Bc= lre}', "");
Expect(0, 8234, '\P{Is_Bc= lre}', "");
Expect(1, 8234, '\P{^Is_Bc= lre}', "");
Expect(0, 8235, '\p{Is_Bc= lre}', "");
Expect(1, 8235, '\p{^Is_Bc= lre}', "");
Expect(1, 8235, '\P{Is_Bc= lre}', "");
Expect(0, 8235, '\P{^Is_Bc= lre}', "");
Error('\p{Bidi_Class=-/a/Left_To_RIGHT_ISOLATE}');
Error('\P{Bidi_Class=-/a/Left_To_RIGHT_ISOLATE}');
Expect(1, 8294, '\p{Bidi_Class=lefttorightisolate}', "");
Expect(0, 8294, '\p{^Bidi_Class=lefttorightisolate}', "");
Expect(0, 8294, '\P{Bidi_Class=lefttorightisolate}', "");
Expect(1, 8294, '\P{^Bidi_Class=lefttorightisolate}', "");
Expect(0, 8295, '\p{Bidi_Class=lefttorightisolate}', "");
Expect(1, 8295, '\p{^Bidi_Class=lefttorightisolate}', "");
Expect(1, 8295, '\P{Bidi_Class=lefttorightisolate}', "");
Expect(0, 8295, '\P{^Bidi_Class=lefttorightisolate}', "");
Expect(1, 8294, '\p{Bidi_Class=-Left_To_Right_Isolate}', "");
Expect(0, 8294, '\p{^Bidi_Class=-Left_To_Right_Isolate}', "");
Expect(0, 8294, '\P{Bidi_Class=-Left_To_Right_Isolate}', "");
Expect(1, 8294, '\P{^Bidi_Class=-Left_To_Right_Isolate}', "");
Expect(0, 8295, '\p{Bidi_Class=-Left_To_Right_Isolate}', "");
Expect(1, 8295, '\p{^Bidi_Class=-Left_To_Right_Isolate}', "");
Expect(1, 8295, '\P{Bidi_Class=-Left_To_Right_Isolate}', "");
Expect(0, 8295, '\P{^Bidi_Class=-Left_To_Right_Isolate}', "");
Error('\p{Bc:   __LRI/a/}');
Error('\P{Bc:   __LRI/a/}');
Expect(1, 8294, '\p{Bc=lri}', "");
Expect(0, 8294, '\p{^Bc=lri}', "");
Expect(0, 8294, '\P{Bc=lri}', "");
Expect(1, 8294, '\P{^Bc=lri}', "");
Expect(0, 8295, '\p{Bc=lri}', "");
Expect(1, 8295, '\p{^Bc=lri}', "");
Expect(1, 8295, '\P{Bc=lri}', "");
Expect(0, 8295, '\P{^Bc=lri}', "");
Expect(1, 8294, '\p{Bc=LRI}', "");
Expect(0, 8294, '\p{^Bc=LRI}', "");
Expect(0, 8294, '\P{Bc=LRI}', "");
Expect(1, 8294, '\P{^Bc=LRI}', "");
Expect(0, 8295, '\p{Bc=LRI}', "");
Expect(1, 8295, '\p{^Bc=LRI}', "");
Expect(1, 8295, '\P{Bc=LRI}', "");
Expect(0, 8295, '\P{^Bc=LRI}', "");
Error('\p{Is_Bidi_Class=/a/Left_To_right_isolate}');
Error('\P{Is_Bidi_Class=/a/Left_To_right_isolate}');
Expect(1, 8294, '\p{Is_Bidi_Class=lefttorightisolate}', "");
Expect(0, 8294, '\p{^Is_Bidi_Class=lefttorightisolate}', "");
Expect(0, 8294, '\P{Is_Bidi_Class=lefttorightisolate}', "");
Expect(1, 8294, '\P{^Is_Bidi_Class=lefttorightisolate}', "");
Expect(0, 8295, '\p{Is_Bidi_Class=lefttorightisolate}', "");
Expect(1, 8295, '\p{^Is_Bidi_Class=lefttorightisolate}', "");
Expect(1, 8295, '\P{Is_Bidi_Class=lefttorightisolate}', "");
Expect(0, 8295, '\P{^Is_Bidi_Class=lefttorightisolate}', "");
Expect(1, 8294, '\p{Is_Bidi_Class=-Left_To_right_Isolate}', "");
Expect(0, 8294, '\p{^Is_Bidi_Class=-Left_To_right_Isolate}', "");
Expect(0, 8294, '\P{Is_Bidi_Class=-Left_To_right_Isolate}', "");
Expect(1, 8294, '\P{^Is_Bidi_Class=-Left_To_right_Isolate}', "");
Expect(0, 8295, '\p{Is_Bidi_Class=-Left_To_right_Isolate}', "");
Expect(1, 8295, '\p{^Is_Bidi_Class=-Left_To_right_Isolate}', "");
Expect(1, 8295, '\P{Is_Bidi_Class=-Left_To_right_Isolate}', "");
Expect(0, 8295, '\P{^Is_Bidi_Class=-Left_To_right_Isolate}', "");
Error('\p{Is_Bc=:=-LRI}');
Error('\P{Is_Bc=:=-LRI}');
Expect(1, 8294, '\p{Is_Bc:	lri}', "");
Expect(0, 8294, '\p{^Is_Bc:	lri}', "");
Expect(0, 8294, '\P{Is_Bc:	lri}', "");
Expect(1, 8294, '\P{^Is_Bc:	lri}', "");
Expect(0, 8295, '\p{Is_Bc:	lri}', "");
Expect(1, 8295, '\p{^Is_Bc:	lri}', "");
Expect(1, 8295, '\P{Is_Bc:	lri}', "");
Expect(0, 8295, '\P{^Is_Bc:	lri}', "");
Expect(1, 8294, '\p{Is_Bc=-LRI}', "");
Expect(0, 8294, '\p{^Is_Bc=-LRI}', "");
Expect(0, 8294, '\P{Is_Bc=-LRI}', "");
Expect(1, 8294, '\P{^Is_Bc=-LRI}', "");
Expect(0, 8295, '\p{Is_Bc=-LRI}', "");
Expect(1, 8295, '\p{^Is_Bc=-LRI}', "");
Expect(1, 8295, '\P{Is_Bc=-LRI}', "");
Expect(0, 8295, '\P{^Is_Bc=-LRI}', "");
Error('\p{Bidi_Class=/a/Left_TO_RIGHT_Override}');
Error('\P{Bidi_Class=/a/Left_TO_RIGHT_Override}');
Expect(1, 8237, '\p{Bidi_Class=lefttorightoverride}', "");
Expect(0, 8237, '\p{^Bidi_Class=lefttorightoverride}', "");
Expect(0, 8237, '\P{Bidi_Class=lefttorightoverride}', "");
Expect(1, 8237, '\P{^Bidi_Class=lefttorightoverride}', "");
Expect(0, 8238, '\p{Bidi_Class=lefttorightoverride}', "");
Expect(1, 8238, '\p{^Bidi_Class=lefttorightoverride}', "");
Expect(1, 8238, '\P{Bidi_Class=lefttorightoverride}', "");
Expect(0, 8238, '\P{^Bidi_Class=lefttorightoverride}', "");
Expect(1, 8237, '\p{Bidi_Class=	_left_To_Right_Override}', "");
Expect(0, 8237, '\p{^Bidi_Class=	_left_To_Right_Override}', "");
Expect(0, 8237, '\P{Bidi_Class=	_left_To_Right_Override}', "");
Expect(1, 8237, '\P{^Bidi_Class=	_left_To_Right_Override}', "");
Expect(0, 8238, '\p{Bidi_Class=	_left_To_Right_Override}', "");
Expect(1, 8238, '\p{^Bidi_Class=	_left_To_Right_Override}', "");
Expect(1, 8238, '\P{Bidi_Class=	_left_To_Right_Override}', "");
Expect(0, 8238, '\P{^Bidi_Class=	_left_To_Right_Override}', "");
Error('\p{Bc=:= LRO}');
Error('\P{Bc=:= LRO}');
Expect(1, 8237, '\p{Bc=lro}', "");
Expect(0, 8237, '\p{^Bc=lro}', "");
Expect(0, 8237, '\P{Bc=lro}', "");
Expect(1, 8237, '\P{^Bc=lro}', "");
Expect(0, 8238, '\p{Bc=lro}', "");
Expect(1, 8238, '\p{^Bc=lro}', "");
Expect(1, 8238, '\P{Bc=lro}', "");
Expect(0, 8238, '\P{^Bc=lro}', "");
Expect(1, 8237, '\p{Bc= _lro}', "");
Expect(0, 8237, '\p{^Bc= _lro}', "");
Expect(0, 8237, '\P{Bc= _lro}', "");
Expect(1, 8237, '\P{^Bc= _lro}', "");
Expect(0, 8238, '\p{Bc= _lro}', "");
Expect(1, 8238, '\p{^Bc= _lro}', "");
Expect(1, 8238, '\P{Bc= _lro}', "");
Expect(0, 8238, '\P{^Bc= _lro}', "");
Error('\p{Is_Bidi_Class=  Left_To_Right_OVERRIDE:=}');
Error('\P{Is_Bidi_Class=  Left_To_Right_OVERRIDE:=}');
Expect(1, 8237, '\p{Is_Bidi_Class=lefttorightoverride}', "");
Expect(0, 8237, '\p{^Is_Bidi_Class=lefttorightoverride}', "");
Expect(0, 8237, '\P{Is_Bidi_Class=lefttorightoverride}', "");
Expect(1, 8237, '\P{^Is_Bidi_Class=lefttorightoverride}', "");
Expect(0, 8238, '\p{Is_Bidi_Class=lefttorightoverride}', "");
Expect(1, 8238, '\p{^Is_Bidi_Class=lefttorightoverride}', "");
Expect(1, 8238, '\P{Is_Bidi_Class=lefttorightoverride}', "");
Expect(0, 8238, '\P{^Is_Bidi_Class=lefttorightoverride}', "");
Expect(1, 8237, '\p{Is_Bidi_Class=  Left_TO_RIGHT_Override}', "");
Expect(0, 8237, '\p{^Is_Bidi_Class=  Left_TO_RIGHT_Override}', "");
Expect(0, 8237, '\P{Is_Bidi_Class=  Left_TO_RIGHT_Override}', "");
Expect(1, 8237, '\P{^Is_Bidi_Class=  Left_TO_RIGHT_Override}', "");
Expect(0, 8238, '\p{Is_Bidi_Class=  Left_TO_RIGHT_Override}', "");
Expect(1, 8238, '\p{^Is_Bidi_Class=  Left_TO_RIGHT_Override}', "");
Expect(1, 8238, '\P{Is_Bidi_Class=  Left_TO_RIGHT_Override}', "");
Expect(0, 8238, '\P{^Is_Bidi_Class=  Left_TO_RIGHT_Override}', "");
Error('\p{Is_Bc=:=-lro}');
Error('\P{Is_Bc=:=-lro}');
Expect(1, 8237, '\p{Is_Bc=lro}', "");
Expect(0, 8237, '\p{^Is_Bc=lro}', "");
Expect(0, 8237, '\P{Is_Bc=lro}', "");
Expect(1, 8237, '\P{^Is_Bc=lro}', "");
Expect(0, 8238, '\p{Is_Bc=lro}', "");
Expect(1, 8238, '\p{^Is_Bc=lro}', "");
Expect(1, 8238, '\P{Is_Bc=lro}', "");
Expect(0, 8238, '\P{^Is_Bc=lro}', "");
Expect(1, 8237, '\p{Is_Bc=_LRO}', "");
Expect(0, 8237, '\p{^Is_Bc=_LRO}', "");
Expect(0, 8237, '\P{Is_Bc=_LRO}', "");
Expect(1, 8237, '\P{^Is_Bc=_LRO}', "");
Expect(0, 8238, '\p{Is_Bc=_LRO}', "");
Expect(1, 8238, '\p{^Is_Bc=_LRO}', "");
Expect(1, 8238, '\P{Is_Bc=_LRO}', "");
Expect(0, 8238, '\P{^Is_Bc=_LRO}', "");
Error('\p{Bidi_Class=/a/_	Nonspacing_MARK}');
Error('\P{Bidi_Class=/a/_	Nonspacing_MARK}');
Expect(1, 917999, '\p{Bidi_Class=nonspacingmark}', "");
Expect(0, 917999, '\p{^Bidi_Class=nonspacingmark}', "");
Expect(0, 917999, '\P{Bidi_Class=nonspacingmark}', "");
Expect(1, 917999, '\P{^Bidi_Class=nonspacingmark}', "");
Expect(0, 918000, '\p{Bidi_Class=nonspacingmark}', "");
Expect(1, 918000, '\p{^Bidi_Class=nonspacingmark}', "");
Expect(1, 918000, '\P{Bidi_Class=nonspacingmark}', "");
Expect(0, 918000, '\P{^Bidi_Class=nonspacingmark}', "");
Expect(1, 917999, '\p{Bidi_Class=_Nonspacing_Mark}', "");
Expect(0, 917999, '\p{^Bidi_Class=_Nonspacing_Mark}', "");
Expect(0, 917999, '\P{Bidi_Class=_Nonspacing_Mark}', "");
Expect(1, 917999, '\P{^Bidi_Class=_Nonspacing_Mark}', "");
Expect(0, 918000, '\p{Bidi_Class=_Nonspacing_Mark}', "");
Expect(1, 918000, '\p{^Bidi_Class=_Nonspacing_Mark}', "");
Expect(1, 918000, '\P{Bidi_Class=_Nonspacing_Mark}', "");
Expect(0, 918000, '\P{^Bidi_Class=_Nonspacing_Mark}', "");
Error('\p{Bc=:=-_nsm}');
Error('\P{Bc=:=-_nsm}');
Expect(1, 917999, '\p{Bc=nsm}', "");
Expect(0, 917999, '\p{^Bc=nsm}', "");
Expect(0, 917999, '\P{Bc=nsm}', "");
Expect(1, 917999, '\P{^Bc=nsm}', "");
Expect(0, 918000, '\p{Bc=nsm}', "");
Expect(1, 918000, '\p{^Bc=nsm}', "");
Expect(1, 918000, '\P{Bc=nsm}', "");
Expect(0, 918000, '\P{^Bc=nsm}', "");
Expect(1, 917999, '\p{Bc=- NSM}', "");
Expect(0, 917999, '\p{^Bc=- NSM}', "");
Expect(0, 917999, '\P{Bc=- NSM}', "");
Expect(1, 917999, '\P{^Bc=- NSM}', "");
Expect(0, 918000, '\p{Bc=- NSM}', "");
Expect(1, 918000, '\p{^Bc=- NSM}', "");
Expect(1, 918000, '\P{Bc=- NSM}', "");
Expect(0, 918000, '\P{^Bc=- NSM}', "");
Error('\p{Is_Bidi_Class=	 Nonspacing_Mark/a/}');
Error('\P{Is_Bidi_Class=	 Nonspacing_Mark/a/}');
Expect(1, 917999, '\p{Is_Bidi_Class=nonspacingmark}', "");
Expect(0, 917999, '\p{^Is_Bidi_Class=nonspacingmark}', "");
Expect(0, 917999, '\P{Is_Bidi_Class=nonspacingmark}', "");
Expect(1, 917999, '\P{^Is_Bidi_Class=nonspacingmark}', "");
Expect(0, 918000, '\p{Is_Bidi_Class=nonspacingmark}', "");
Expect(1, 918000, '\p{^Is_Bidi_Class=nonspacingmark}', "");
Expect(1, 918000, '\P{Is_Bidi_Class=nonspacingmark}', "");
Expect(0, 918000, '\P{^Is_Bidi_Class=nonspacingmark}', "");
Expect(1, 917999, '\p{Is_Bidi_Class=-	NONSPACING_mark}', "");
Expect(0, 917999, '\p{^Is_Bidi_Class=-	NONSPACING_mark}', "");
Expect(0, 917999, '\P{Is_Bidi_Class=-	NONSPACING_mark}', "");
Expect(1, 917999, '\P{^Is_Bidi_Class=-	NONSPACING_mark}', "");
Expect(0, 918000, '\p{Is_Bidi_Class=-	NONSPACING_mark}', "");
Expect(1, 918000, '\p{^Is_Bidi_Class=-	NONSPACING_mark}', "");
Expect(1, 918000, '\P{Is_Bidi_Class=-	NONSPACING_mark}', "");
Expect(0, 918000, '\P{^Is_Bidi_Class=-	NONSPACING_mark}', "");
Error('\p{Is_Bc:			NSM/a/}');
Error('\P{Is_Bc:			NSM/a/}');
Expect(1, 917999, '\p{Is_Bc=nsm}', "");
Expect(0, 917999, '\p{^Is_Bc=nsm}', "");
Expect(0, 917999, '\P{Is_Bc=nsm}', "");
Expect(1, 917999, '\P{^Is_Bc=nsm}', "");
Expect(0, 918000, '\p{Is_Bc=nsm}', "");
Expect(1, 918000, '\p{^Is_Bc=nsm}', "");
Expect(1, 918000, '\P{Is_Bc=nsm}', "");
Expect(0, 918000, '\P{^Is_Bc=nsm}', "");
Expect(1, 917999, '\p{Is_Bc= NSM}', "");
Expect(0, 917999, '\p{^Is_Bc= NSM}', "");
Expect(0, 917999, '\P{Is_Bc= NSM}', "");
Expect(1, 917999, '\P{^Is_Bc= NSM}', "");
Expect(0, 918000, '\p{Is_Bc= NSM}', "");
Expect(1, 918000, '\p{^Is_Bc= NSM}', "");
Expect(1, 918000, '\P{Is_Bc= NSM}', "");
Expect(0, 918000, '\P{^Is_Bc= NSM}', "");
Error('\p{Bidi_Class= Other_Neutral:=}');
Error('\P{Bidi_Class= Other_Neutral:=}');
Expect(1, 128883, '\p{Bidi_Class=otherneutral}', "");
Expect(0, 128883, '\p{^Bidi_Class=otherneutral}', "");
Expect(0, 128883, '\P{Bidi_Class=otherneutral}', "");
Expect(1, 128883, '\P{^Bidi_Class=otherneutral}', "");
Expect(0, 128884, '\p{Bidi_Class=otherneutral}', "");
Expect(1, 128884, '\p{^Bidi_Class=otherneutral}', "");
Expect(1, 128884, '\P{Bidi_Class=otherneutral}', "");
Expect(0, 128884, '\P{^Bidi_Class=otherneutral}', "");
Expect(1, 128883, '\p{Bidi_Class=- OTHER_Neutral}', "");
Expect(0, 128883, '\p{^Bidi_Class=- OTHER_Neutral}', "");
Expect(0, 128883, '\P{Bidi_Class=- OTHER_Neutral}', "");
Expect(1, 128883, '\P{^Bidi_Class=- OTHER_Neutral}', "");
Expect(0, 128884, '\p{Bidi_Class=- OTHER_Neutral}', "");
Expect(1, 128884, '\p{^Bidi_Class=- OTHER_Neutral}', "");
Expect(1, 128884, '\P{Bidi_Class=- OTHER_Neutral}', "");
Expect(0, 128884, '\P{^Bidi_Class=- OTHER_Neutral}', "");
Error('\p{Bc:   /a/ -ON}');
Error('\P{Bc:   /a/ -ON}');
Expect(1, 128883, '\p{Bc=on}', "");
Expect(0, 128883, '\p{^Bc=on}', "");
Expect(0, 128883, '\P{Bc=on}', "");
Expect(1, 128883, '\P{^Bc=on}', "");
Expect(0, 128884, '\p{Bc=on}', "");
Expect(1, 128884, '\p{^Bc=on}', "");
Expect(1, 128884, '\P{Bc=on}', "");
Expect(0, 128884, '\P{^Bc=on}', "");
Expect(1, 128883, '\p{Bc=	_ON}', "");
Expect(0, 128883, '\p{^Bc=	_ON}', "");
Expect(0, 128883, '\P{Bc=	_ON}', "");
Expect(1, 128883, '\P{^Bc=	_ON}', "");
Expect(0, 128884, '\p{Bc=	_ON}', "");
Expect(1, 128884, '\p{^Bc=	_ON}', "");
Expect(1, 128884, '\P{Bc=	_ON}', "");
Expect(0, 128884, '\P{^Bc=	_ON}', "");
Error('\p{Is_Bidi_Class=-OTHER_Neutral/a/}');
Error('\P{Is_Bidi_Class=-OTHER_Neutral/a/}');
Expect(1, 128883, '\p{Is_Bidi_Class=otherneutral}', "");
Expect(0, 128883, '\p{^Is_Bidi_Class=otherneutral}', "");
Expect(0, 128883, '\P{Is_Bidi_Class=otherneutral}', "");
Expect(1, 128883, '\P{^Is_Bidi_Class=otherneutral}', "");
Expect(0, 128884, '\p{Is_Bidi_Class=otherneutral}', "");
Expect(1, 128884, '\p{^Is_Bidi_Class=otherneutral}', "");
Expect(1, 128884, '\P{Is_Bidi_Class=otherneutral}', "");
Expect(0, 128884, '\P{^Is_Bidi_Class=otherneutral}', "");
Expect(1, 128883, '\p{Is_Bidi_Class=	Other_Neutral}', "");
Expect(0, 128883, '\p{^Is_Bidi_Class=	Other_Neutral}', "");
Expect(0, 128883, '\P{Is_Bidi_Class=	Other_Neutral}', "");
Expect(1, 128883, '\P{^Is_Bidi_Class=	Other_Neutral}', "");
Expect(0, 128884, '\p{Is_Bidi_Class=	Other_Neutral}', "");
Expect(1, 128884, '\p{^Is_Bidi_Class=	Other_Neutral}', "");
Expect(1, 128884, '\P{Is_Bidi_Class=	Other_Neutral}', "");
Expect(0, 128884, '\P{^Is_Bidi_Class=	Other_Neutral}', "");
Error('\p{Is_Bc=:=__ON}');
Error('\P{Is_Bc=:=__ON}');
Expect(1, 128883, '\p{Is_Bc=on}', "");
Expect(0, 128883, '\p{^Is_Bc=on}', "");
Expect(0, 128883, '\P{Is_Bc=on}', "");
Expect(1, 128883, '\P{^Is_Bc=on}', "");
Expect(0, 128884, '\p{Is_Bc=on}', "");
Expect(1, 128884, '\p{^Is_Bc=on}', "");
Expect(1, 128884, '\P{Is_Bc=on}', "");
Expect(0, 128884, '\P{^Is_Bc=on}', "");
Expect(1, 128883, '\p{Is_Bc:   	_ON}', "");
Expect(0, 128883, '\p{^Is_Bc:   	_ON}', "");
Expect(0, 128883, '\P{Is_Bc:   	_ON}', "");
Expect(1, 128883, '\P{^Is_Bc:   	_ON}', "");
Expect(0, 128884, '\p{Is_Bc:   	_ON}', "");
Expect(1, 128884, '\p{^Is_Bc:   	_ON}', "");
Expect(1, 128884, '\P{Is_Bc:   	_ON}', "");
Expect(0, 128884, '\P{^Is_Bc:   	_ON}', "");
Error('\p{Bidi_Class=-/a/POP_directional_Format}');
Error('\P{Bidi_Class=-/a/POP_directional_Format}');
Expect(1, 8236, '\p{Bidi_Class=popdirectionalformat}', "");
Expect(0, 8236, '\p{^Bidi_Class=popdirectionalformat}', "");
Expect(0, 8236, '\P{Bidi_Class=popdirectionalformat}', "");
Expect(1, 8236, '\P{^Bidi_Class=popdirectionalformat}', "");
Expect(0, 8237, '\p{Bidi_Class=popdirectionalformat}', "");
Expect(1, 8237, '\p{^Bidi_Class=popdirectionalformat}', "");
Expect(1, 8237, '\P{Bidi_Class=popdirectionalformat}', "");
Expect(0, 8237, '\P{^Bidi_Class=popdirectionalformat}', "");
Expect(1, 8236, '\p{Bidi_Class=	POP_DIRECTIONAL_Format}', "");
Expect(0, 8236, '\p{^Bidi_Class=	POP_DIRECTIONAL_Format}', "");
Expect(0, 8236, '\P{Bidi_Class=	POP_DIRECTIONAL_Format}', "");
Expect(1, 8236, '\P{^Bidi_Class=	POP_DIRECTIONAL_Format}', "");
Expect(0, 8237, '\p{Bidi_Class=	POP_DIRECTIONAL_Format}', "");
Expect(1, 8237, '\p{^Bidi_Class=	POP_DIRECTIONAL_Format}', "");
Expect(1, 8237, '\P{Bidi_Class=	POP_DIRECTIONAL_Format}', "");
Expect(0, 8237, '\P{^Bidi_Class=	POP_DIRECTIONAL_Format}', "");
Error('\p{Bc:	 /a/PDF}');
Error('\P{Bc:	 /a/PDF}');
Expect(1, 8236, '\p{Bc=pdf}', "");
Expect(0, 8236, '\p{^Bc=pdf}', "");
Expect(0, 8236, '\P{Bc=pdf}', "");
Expect(1, 8236, '\P{^Bc=pdf}', "");
Expect(0, 8237, '\p{Bc=pdf}', "");
Expect(1, 8237, '\p{^Bc=pdf}', "");
Expect(1, 8237, '\P{Bc=pdf}', "");
Expect(0, 8237, '\P{^Bc=pdf}', "");
Expect(1, 8236, '\p{Bc= PDF}', "");
Expect(0, 8236, '\p{^Bc= PDF}', "");
Expect(0, 8236, '\P{Bc= PDF}', "");
Expect(1, 8236, '\P{^Bc= PDF}', "");
Expect(0, 8237, '\p{Bc= PDF}', "");
Expect(1, 8237, '\p{^Bc= PDF}', "");
Expect(1, 8237, '\P{Bc= PDF}', "");
Expect(0, 8237, '\P{^Bc= PDF}', "");
Error('\p{Is_Bidi_Class=_Pop_Directional_Format/a/}');
Error('\P{Is_Bidi_Class=_Pop_Directional_Format/a/}');
Expect(1, 8236, '\p{Is_Bidi_Class=popdirectionalformat}', "");
Expect(0, 8236, '\p{^Is_Bidi_Class=popdirectionalformat}', "");
Expect(0, 8236, '\P{Is_Bidi_Class=popdirectionalformat}', "");
Expect(1, 8236, '\P{^Is_Bidi_Class=popdirectionalformat}', "");
Expect(0, 8237, '\p{Is_Bidi_Class=popdirectionalformat}', "");
Expect(1, 8237, '\p{^Is_Bidi_Class=popdirectionalformat}', "");
Expect(1, 8237, '\P{Is_Bidi_Class=popdirectionalformat}', "");
Expect(0, 8237, '\P{^Is_Bidi_Class=popdirectionalformat}', "");
Expect(1, 8236, '\p{Is_Bidi_Class=-_pop_DIRECTIONAL_Format}', "");
Expect(0, 8236, '\p{^Is_Bidi_Class=-_pop_DIRECTIONAL_Format}', "");
Expect(0, 8236, '\P{Is_Bidi_Class=-_pop_DIRECTIONAL_Format}', "");
Expect(1, 8236, '\P{^Is_Bidi_Class=-_pop_DIRECTIONAL_Format}', "");
Expect(0, 8237, '\p{Is_Bidi_Class=-_pop_DIRECTIONAL_Format}', "");
Expect(1, 8237, '\p{^Is_Bidi_Class=-_pop_DIRECTIONAL_Format}', "");
Expect(1, 8237, '\P{Is_Bidi_Class=-_pop_DIRECTIONAL_Format}', "");
Expect(0, 8237, '\P{^Is_Bidi_Class=-_pop_DIRECTIONAL_Format}', "");
Error('\p{Is_Bc=	pdf:=}');
Error('\P{Is_Bc=	pdf:=}');
Expect(1, 8236, '\p{Is_Bc=pdf}', "");
Expect(0, 8236, '\p{^Is_Bc=pdf}', "");
Expect(0, 8236, '\P{Is_Bc=pdf}', "");
Expect(1, 8236, '\P{^Is_Bc=pdf}', "");
Expect(0, 8237, '\p{Is_Bc=pdf}', "");
Expect(1, 8237, '\p{^Is_Bc=pdf}', "");
Expect(1, 8237, '\P{Is_Bc=pdf}', "");
Expect(0, 8237, '\P{^Is_Bc=pdf}', "");
Expect(1, 8236, '\p{Is_Bc=__PDF}', "");
Expect(0, 8236, '\p{^Is_Bc=__PDF}', "");
Expect(0, 8236, '\P{Is_Bc=__PDF}', "");
Expect(1, 8236, '\P{^Is_Bc=__PDF}', "");
Expect(0, 8237, '\p{Is_Bc=__PDF}', "");
Expect(1, 8237, '\p{^Is_Bc=__PDF}', "");
Expect(1, 8237, '\P{Is_Bc=__PDF}', "");
Expect(0, 8237, '\P{^Is_Bc=__PDF}', "");
Error('\p{Bidi_Class=	-Pop_Directional_Isolate/a/}');
Error('\P{Bidi_Class=	-Pop_Directional_Isolate/a/}');
Expect(1, 8297, '\p{Bidi_Class=popdirectionalisolate}', "");
Expect(0, 8297, '\p{^Bidi_Class=popdirectionalisolate}', "");
Expect(0, 8297, '\P{Bidi_Class=popdirectionalisolate}', "");
Expect(1, 8297, '\P{^Bidi_Class=popdirectionalisolate}', "");
Expect(0, 8298, '\p{Bidi_Class=popdirectionalisolate}', "");
Expect(1, 8298, '\p{^Bidi_Class=popdirectionalisolate}', "");
Expect(1, 8298, '\P{Bidi_Class=popdirectionalisolate}', "");
Expect(0, 8298, '\P{^Bidi_Class=popdirectionalisolate}', "");
Expect(1, 8297, '\p{Bidi_Class=	POP_Directional_Isolate}', "");
Expect(0, 8297, '\p{^Bidi_Class=	POP_Directional_Isolate}', "");
Expect(0, 8297, '\P{Bidi_Class=	POP_Directional_Isolate}', "");
Expect(1, 8297, '\P{^Bidi_Class=	POP_Directional_Isolate}', "");
Expect(0, 8298, '\p{Bidi_Class=	POP_Directional_Isolate}', "");
Expect(1, 8298, '\p{^Bidi_Class=	POP_Directional_Isolate}', "");
Expect(1, 8298, '\P{Bidi_Class=	POP_Directional_Isolate}', "");
Expect(0, 8298, '\P{^Bidi_Class=	POP_Directional_Isolate}', "");
Error('\p{Bc=-:=PDI}');
Error('\P{Bc=-:=PDI}');
Expect(1, 8297, '\p{Bc=pdi}', "");
Expect(0, 8297, '\p{^Bc=pdi}', "");
Expect(0, 8297, '\P{Bc=pdi}', "");
Expect(1, 8297, '\P{^Bc=pdi}', "");
Expect(0, 8298, '\p{Bc=pdi}', "");
Expect(1, 8298, '\p{^Bc=pdi}', "");
Expect(1, 8298, '\P{Bc=pdi}', "");
Expect(0, 8298, '\P{^Bc=pdi}', "");
Expect(1, 8297, '\p{Bc=- PDI}', "");
Expect(0, 8297, '\p{^Bc=- PDI}', "");
Expect(0, 8297, '\P{Bc=- PDI}', "");
Expect(1, 8297, '\P{^Bc=- PDI}', "");
Expect(0, 8298, '\p{Bc=- PDI}', "");
Expect(1, 8298, '\p{^Bc=- PDI}', "");
Expect(1, 8298, '\P{Bc=- PDI}', "");
Expect(0, 8298, '\P{^Bc=- PDI}', "");
Error('\p{Is_Bidi_Class=:=-Pop_Directional_isolate}');
Error('\P{Is_Bidi_Class=:=-Pop_Directional_isolate}');
Expect(1, 8297, '\p{Is_Bidi_Class:   popdirectionalisolate}', "");
Expect(0, 8297, '\p{^Is_Bidi_Class:   popdirectionalisolate}', "");
Expect(0, 8297, '\P{Is_Bidi_Class:   popdirectionalisolate}', "");
Expect(1, 8297, '\P{^Is_Bidi_Class:   popdirectionalisolate}', "");
Expect(0, 8298, '\p{Is_Bidi_Class:   popdirectionalisolate}', "");
Expect(1, 8298, '\p{^Is_Bidi_Class:   popdirectionalisolate}', "");
Expect(1, 8298, '\P{Is_Bidi_Class:   popdirectionalisolate}', "");
Expect(0, 8298, '\P{^Is_Bidi_Class:   popdirectionalisolate}', "");
Expect(1, 8297, '\p{Is_Bidi_Class=- pop_DIRECTIONAL_Isolate}', "");
Expect(0, 8297, '\p{^Is_Bidi_Class=- pop_DIRECTIONAL_Isolate}', "");
Expect(0, 8297, '\P{Is_Bidi_Class=- pop_DIRECTIONAL_Isolate}', "");
Expect(1, 8297, '\P{^Is_Bidi_Class=- pop_DIRECTIONAL_Isolate}', "");
Expect(0, 8298, '\p{Is_Bidi_Class=- pop_DIRECTIONAL_Isolate}', "");
Expect(1, 8298, '\p{^Is_Bidi_Class=- pop_DIRECTIONAL_Isolate}', "");
Expect(1, 8298, '\P{Is_Bidi_Class=- pop_DIRECTIONAL_Isolate}', "");
Expect(0, 8298, '\P{^Is_Bidi_Class=- pop_DIRECTIONAL_Isolate}', "");
Error('\p{Is_Bc:	/a/ _PDI}');
Error('\P{Is_Bc:	/a/ _PDI}');
Expect(1, 8297, '\p{Is_Bc=pdi}', "");
Expect(0, 8297, '\p{^Is_Bc=pdi}', "");
Expect(0, 8297, '\P{Is_Bc=pdi}', "");
Expect(1, 8297, '\P{^Is_Bc=pdi}', "");
Expect(0, 8298, '\p{Is_Bc=pdi}', "");
Expect(1, 8298, '\p{^Is_Bc=pdi}', "");
Expect(1, 8298, '\P{Is_Bc=pdi}', "");
Expect(0, 8298, '\P{^Is_Bc=pdi}', "");
Expect(1, 8297, '\p{Is_Bc= _PDI}', "");
Expect(0, 8297, '\p{^Is_Bc= _PDI}', "");
Expect(0, 8297, '\P{Is_Bc= _PDI}', "");
Expect(1, 8297, '\P{^Is_Bc= _PDI}', "");
Expect(0, 8298, '\p{Is_Bc= _PDI}', "");
Expect(1, 8298, '\p{^Is_Bc= _PDI}', "");
Expect(1, 8298, '\P{Is_Bc= _PDI}', "");
Expect(0, 8298, '\P{^Is_Bc= _PDI}', "");
Error('\p{Bidi_Class=:= 	right_To_left}');
Error('\P{Bidi_Class=:= 	right_To_left}');
Expect(1, 126975, '\p{Bidi_Class=righttoleft}', "");
Expect(0, 126975, '\p{^Bidi_Class=righttoleft}', "");
Expect(0, 126975, '\P{Bidi_Class=righttoleft}', "");
Expect(1, 126975, '\P{^Bidi_Class=righttoleft}', "");
Expect(0, 126976, '\p{Bidi_Class=righttoleft}', "");
Expect(1, 126976, '\p{^Bidi_Class=righttoleft}', "");
Expect(1, 126976, '\P{Bidi_Class=righttoleft}', "");
Expect(0, 126976, '\P{^Bidi_Class=righttoleft}', "");
Expect(1, 126975, '\p{Bidi_Class=-Right_to_Left}', "");
Expect(0, 126975, '\p{^Bidi_Class=-Right_to_Left}', "");
Expect(0, 126975, '\P{Bidi_Class=-Right_to_Left}', "");
Expect(1, 126975, '\P{^Bidi_Class=-Right_to_Left}', "");
Expect(0, 126976, '\p{Bidi_Class=-Right_to_Left}', "");
Expect(1, 126976, '\p{^Bidi_Class=-Right_to_Left}', "");
Expect(1, 126976, '\P{Bidi_Class=-Right_to_Left}', "");
Expect(0, 126976, '\P{^Bidi_Class=-Right_to_Left}', "");
Error('\p{Bc:  R/a/}');
Error('\P{Bc:  R/a/}');
Expect(1, 126975, '\p{Bc=r}', "");
Expect(0, 126975, '\p{^Bc=r}', "");
Expect(0, 126975, '\P{Bc=r}', "");
Expect(1, 126975, '\P{^Bc=r}', "");
Expect(0, 126976, '\p{Bc=r}', "");
Expect(1, 126976, '\p{^Bc=r}', "");
Expect(1, 126976, '\P{Bc=r}', "");
Expect(0, 126976, '\P{^Bc=r}', "");
Expect(1, 126975, '\p{Bc=_R}', "");
Expect(0, 126975, '\p{^Bc=_R}', "");
Expect(0, 126975, '\P{Bc=_R}', "");
Expect(1, 126975, '\P{^Bc=_R}', "");
Expect(0, 126976, '\p{Bc=_R}', "");
Expect(1, 126976, '\p{^Bc=_R}', "");
Expect(1, 126976, '\P{Bc=_R}', "");
Expect(0, 126976, '\P{^Bc=_R}', "");
Error('\p{Is_Bidi_Class=__RIGHT_To_left/a/}');
Error('\P{Is_Bidi_Class=__RIGHT_To_left/a/}');
Expect(1, 126975, '\p{Is_Bidi_Class=righttoleft}', "");
Expect(0, 126975, '\p{^Is_Bidi_Class=righttoleft}', "");
Expect(0, 126975, '\P{Is_Bidi_Class=righttoleft}', "");
Expect(1, 126975, '\P{^Is_Bidi_Class=righttoleft}', "");
Expect(0, 126976, '\p{Is_Bidi_Class=righttoleft}', "");
Expect(1, 126976, '\p{^Is_Bidi_Class=righttoleft}', "");
Expect(1, 126976, '\P{Is_Bidi_Class=righttoleft}', "");
Expect(0, 126976, '\P{^Is_Bidi_Class=righttoleft}', "");
Expect(1, 126975, '\p{Is_Bidi_Class=		Right_to_left}', "");
Expect(0, 126975, '\p{^Is_Bidi_Class=		Right_to_left}', "");
Expect(0, 126975, '\P{Is_Bidi_Class=		Right_to_left}', "");
Expect(1, 126975, '\P{^Is_Bidi_Class=		Right_to_left}', "");
Expect(0, 126976, '\p{Is_Bidi_Class=		Right_to_left}', "");
Expect(1, 126976, '\p{^Is_Bidi_Class=		Right_to_left}', "");
Expect(1, 126976, '\P{Is_Bidi_Class=		Right_to_left}', "");
Expect(0, 126976, '\P{^Is_Bidi_Class=		Right_to_left}', "");
Error('\p{Is_Bc=	/a/R}');
Error('\P{Is_Bc=	/a/R}');
Expect(1, 126975, '\p{Is_Bc=r}', "");
Expect(0, 126975, '\p{^Is_Bc=r}', "");
Expect(0, 126975, '\P{Is_Bc=r}', "");
Expect(1, 126975, '\P{^Is_Bc=r}', "");
Expect(0, 126976, '\p{Is_Bc=r}', "");
Expect(1, 126976, '\p{^Is_Bc=r}', "");
Expect(1, 126976, '\P{Is_Bc=r}', "");
Expect(0, 126976, '\P{^Is_Bc=r}', "");
Expect(1, 126975, '\p{Is_Bc=-	R}', "");
Expect(0, 126975, '\p{^Is_Bc=-	R}', "");
Expect(0, 126975, '\P{Is_Bc=-	R}', "");
Expect(1, 126975, '\P{^Is_Bc=-	R}', "");
Expect(0, 126976, '\p{Is_Bc=-	R}', "");
Expect(1, 126976, '\p{^Is_Bc=-	R}', "");
Expect(1, 126976, '\P{Is_Bc=-	R}', "");
Expect(0, 126976, '\P{^Is_Bc=-	R}', "");
Error('\p{Bidi_Class=  Right_To_Left_Embedding/a/}');
Error('\P{Bidi_Class=  Right_To_Left_Embedding/a/}');
Expect(1, 8235, '\p{Bidi_Class=righttoleftembedding}', "");
Expect(0, 8235, '\p{^Bidi_Class=righttoleftembedding}', "");
Expect(0, 8235, '\P{Bidi_Class=righttoleftembedding}', "");
Expect(1, 8235, '\P{^Bidi_Class=righttoleftembedding}', "");
Expect(0, 8236, '\p{Bidi_Class=righttoleftembedding}', "");
Expect(1, 8236, '\p{^Bidi_Class=righttoleftembedding}', "");
Expect(1, 8236, '\P{Bidi_Class=righttoleftembedding}', "");
Expect(0, 8236, '\P{^Bidi_Class=righttoleftembedding}', "");
Expect(1, 8235, '\p{Bidi_Class=_ Right_TO_left_Embedding}', "");
Expect(0, 8235, '\p{^Bidi_Class=_ Right_TO_left_Embedding}', "");
Expect(0, 8235, '\P{Bidi_Class=_ Right_TO_left_Embedding}', "");
Expect(1, 8235, '\P{^Bidi_Class=_ Right_TO_left_Embedding}', "");
Expect(0, 8236, '\p{Bidi_Class=_ Right_TO_left_Embedding}', "");
Expect(1, 8236, '\p{^Bidi_Class=_ Right_TO_left_Embedding}', "");
Expect(1, 8236, '\P{Bidi_Class=_ Right_TO_left_Embedding}', "");
Expect(0, 8236, '\P{^Bidi_Class=_ Right_TO_left_Embedding}', "");
Error('\p{Bc=/a/__RLE}');
Error('\P{Bc=/a/__RLE}');
Expect(1, 8235, '\p{Bc=rle}', "");
Expect(0, 8235, '\p{^Bc=rle}', "");
Expect(0, 8235, '\P{Bc=rle}', "");
Expect(1, 8235, '\P{^Bc=rle}', "");
Expect(0, 8236, '\p{Bc=rle}', "");
Expect(1, 8236, '\p{^Bc=rle}', "");
Expect(1, 8236, '\P{Bc=rle}', "");
Expect(0, 8236, '\P{^Bc=rle}', "");
Expect(1, 8235, '\p{Bc=__RLE}', "");
Expect(0, 8235, '\p{^Bc=__RLE}', "");
Expect(0, 8235, '\P{Bc=__RLE}', "");
Expect(1, 8235, '\P{^Bc=__RLE}', "");
Expect(0, 8236, '\p{Bc=__RLE}', "");
Expect(1, 8236, '\p{^Bc=__RLE}', "");
Expect(1, 8236, '\P{Bc=__RLE}', "");
Expect(0, 8236, '\P{^Bc=__RLE}', "");
Error('\p{Is_Bidi_Class=:=-	right_TO_left_embedding}');
Error('\P{Is_Bidi_Class=:=-	right_TO_left_embedding}');
Expect(1, 8235, '\p{Is_Bidi_Class=righttoleftembedding}', "");
Expect(0, 8235, '\p{^Is_Bidi_Class=righttoleftembedding}', "");
Expect(0, 8235, '\P{Is_Bidi_Class=righttoleftembedding}', "");
Expect(1, 8235, '\P{^Is_Bidi_Class=righttoleftembedding}', "");
Expect(0, 8236, '\p{Is_Bidi_Class=righttoleftembedding}', "");
Expect(1, 8236, '\p{^Is_Bidi_Class=righttoleftembedding}', "");
Expect(1, 8236, '\P{Is_Bidi_Class=righttoleftembedding}', "");
Expect(0, 8236, '\P{^Is_Bidi_Class=righttoleftembedding}', "");
Expect(1, 8235, '\p{Is_Bidi_Class= RIGHT_to_Left_Embedding}', "");
Expect(0, 8235, '\p{^Is_Bidi_Class= RIGHT_to_Left_Embedding}', "");
Expect(0, 8235, '\P{Is_Bidi_Class= RIGHT_to_Left_Embedding}', "");
Expect(1, 8235, '\P{^Is_Bidi_Class= RIGHT_to_Left_Embedding}', "");
Expect(0, 8236, '\p{Is_Bidi_Class= RIGHT_to_Left_Embedding}', "");
Expect(1, 8236, '\p{^Is_Bidi_Class= RIGHT_to_Left_Embedding}', "");
Expect(1, 8236, '\P{Is_Bidi_Class= RIGHT_to_Left_Embedding}', "");
Expect(0, 8236, '\P{^Is_Bidi_Class= RIGHT_to_Left_Embedding}', "");
Error('\p{Is_Bc=_:=RLE}');
Error('\P{Is_Bc=_:=RLE}');
Expect(1, 8235, '\p{Is_Bc=rle}', "");
Expect(0, 8235, '\p{^Is_Bc=rle}', "");
Expect(0, 8235, '\P{Is_Bc=rle}', "");
Expect(1, 8235, '\P{^Is_Bc=rle}', "");
Expect(0, 8236, '\p{Is_Bc=rle}', "");
Expect(1, 8236, '\p{^Is_Bc=rle}', "");
Expect(1, 8236, '\P{Is_Bc=rle}', "");
Expect(0, 8236, '\P{^Is_Bc=rle}', "");
Expect(1, 8235, '\p{Is_Bc: -rle}', "");
Expect(0, 8235, '\p{^Is_Bc: -rle}', "");
Expect(0, 8235, '\P{Is_Bc: -rle}', "");
Expect(1, 8235, '\P{^Is_Bc: -rle}', "");
Expect(0, 8236, '\p{Is_Bc: -rle}', "");
Expect(1, 8236, '\p{^Is_Bc: -rle}', "");
Expect(1, 8236, '\P{Is_Bc: -rle}', "");
Expect(0, 8236, '\P{^Is_Bc: -rle}', "");
Error('\p{Bidi_Class=	:=Right_to_LEFT_Isolate}');
Error('\P{Bidi_Class=	:=Right_to_LEFT_Isolate}');
Expect(1, 8295, '\p{Bidi_Class=righttoleftisolate}', "");
Expect(0, 8295, '\p{^Bidi_Class=righttoleftisolate}', "");
Expect(0, 8295, '\P{Bidi_Class=righttoleftisolate}', "");
Expect(1, 8295, '\P{^Bidi_Class=righttoleftisolate}', "");
Expect(0, 8296, '\p{Bidi_Class=righttoleftisolate}', "");
Expect(1, 8296, '\p{^Bidi_Class=righttoleftisolate}', "");
Expect(1, 8296, '\P{Bidi_Class=righttoleftisolate}', "");
Expect(0, 8296, '\P{^Bidi_Class=righttoleftisolate}', "");
Expect(1, 8295, '\p{Bidi_Class= -right_to_left_Isolate}', "");
Expect(0, 8295, '\p{^Bidi_Class= -right_to_left_Isolate}', "");
Expect(0, 8295, '\P{Bidi_Class= -right_to_left_Isolate}', "");
Expect(1, 8295, '\P{^Bidi_Class= -right_to_left_Isolate}', "");
Expect(0, 8296, '\p{Bidi_Class= -right_to_left_Isolate}', "");
Expect(1, 8296, '\p{^Bidi_Class= -right_to_left_Isolate}', "");
Expect(1, 8296, '\P{Bidi_Class= -right_to_left_Isolate}', "");
Expect(0, 8296, '\P{^Bidi_Class= -right_to_left_Isolate}', "");
Error('\p{Bc=/a/ _RLI}');
Error('\P{Bc=/a/ _RLI}');
Expect(1, 8295, '\p{Bc=rli}', "");
Expect(0, 8295, '\p{^Bc=rli}', "");
Expect(0, 8295, '\P{Bc=rli}', "");
Expect(1, 8295, '\P{^Bc=rli}', "");
Expect(0, 8296, '\p{Bc=rli}', "");
Expect(1, 8296, '\p{^Bc=rli}', "");
Expect(1, 8296, '\P{Bc=rli}', "");
Expect(0, 8296, '\P{^Bc=rli}', "");
Expect(1, 8295, '\p{Bc=_ rli}', "");
Expect(0, 8295, '\p{^Bc=_ rli}', "");
Expect(0, 8295, '\P{Bc=_ rli}', "");
Expect(1, 8295, '\P{^Bc=_ rli}', "");
Expect(0, 8296, '\p{Bc=_ rli}', "");
Expect(1, 8296, '\p{^Bc=_ rli}', "");
Expect(1, 8296, '\P{Bc=_ rli}', "");
Expect(0, 8296, '\P{^Bc=_ rli}', "");
Error('\p{Is_Bidi_Class:   :=-Right_TO_left_ISOLATE}');
Error('\P{Is_Bidi_Class:   :=-Right_TO_left_ISOLATE}');
Expect(1, 8295, '\p{Is_Bidi_Class: righttoleftisolate}', "");
Expect(0, 8295, '\p{^Is_Bidi_Class: righttoleftisolate}', "");
Expect(0, 8295, '\P{Is_Bidi_Class: righttoleftisolate}', "");
Expect(1, 8295, '\P{^Is_Bidi_Class: righttoleftisolate}', "");
Expect(0, 8296, '\p{Is_Bidi_Class: righttoleftisolate}', "");
Expect(1, 8296, '\p{^Is_Bidi_Class: righttoleftisolate}', "");
Expect(1, 8296, '\P{Is_Bidi_Class: righttoleftisolate}', "");
Expect(0, 8296, '\P{^Is_Bidi_Class: righttoleftisolate}', "");
Expect(1, 8295, '\p{Is_Bidi_Class:   Right_TO_Left_ISOLATE}', "");
Expect(0, 8295, '\p{^Is_Bidi_Class:   Right_TO_Left_ISOLATE}', "");
Expect(0, 8295, '\P{Is_Bidi_Class:   Right_TO_Left_ISOLATE}', "");
Expect(1, 8295, '\P{^Is_Bidi_Class:   Right_TO_Left_ISOLATE}', "");
Expect(0, 8296, '\p{Is_Bidi_Class:   Right_TO_Left_ISOLATE}', "");
Expect(1, 8296, '\p{^Is_Bidi_Class:   Right_TO_Left_ISOLATE}', "");
Expect(1, 8296, '\P{Is_Bidi_Class:   Right_TO_Left_ISOLATE}', "");
Expect(0, 8296, '\P{^Is_Bidi_Class:   Right_TO_Left_ISOLATE}', "");
Error('\p{Is_Bc=/a/ rli}');
Error('\P{Is_Bc=/a/ rli}');
Expect(1, 8295, '\p{Is_Bc=rli}', "");
Expect(0, 8295, '\p{^Is_Bc=rli}', "");
Expect(0, 8295, '\P{Is_Bc=rli}', "");
Expect(1, 8295, '\P{^Is_Bc=rli}', "");
Expect(0, 8296, '\p{Is_Bc=rli}', "");
Expect(1, 8296, '\p{^Is_Bc=rli}', "");
Expect(1, 8296, '\P{Is_Bc=rli}', "");
Expect(0, 8296, '\P{^Is_Bc=rli}', "");
Expect(1, 8295, '\p{Is_Bc=_-RLI}', "");
Expect(0, 8295, '\p{^Is_Bc=_-RLI}', "");
Expect(0, 8295, '\P{Is_Bc=_-RLI}', "");
Expect(1, 8295, '\P{^Is_Bc=_-RLI}', "");
Expect(0, 8296, '\p{Is_Bc=_-RLI}', "");
Expect(1, 8296, '\p{^Is_Bc=_-RLI}', "");
Expect(1, 8296, '\P{Is_Bc=_-RLI}', "");
Expect(0, 8296, '\P{^Is_Bc=_-RLI}', "");
Error('\p{Bidi_Class=-Right_To_left_override/a/}');
Error('\P{Bidi_Class=-Right_To_left_override/a/}');
Expect(1, 8238, '\p{Bidi_Class=righttoleftoverride}', "");
Expect(0, 8238, '\p{^Bidi_Class=righttoleftoverride}', "");
Expect(0, 8238, '\P{Bidi_Class=righttoleftoverride}', "");
Expect(1, 8238, '\P{^Bidi_Class=righttoleftoverride}', "");
Expect(0, 8239, '\p{Bidi_Class=righttoleftoverride}', "");
Expect(1, 8239, '\p{^Bidi_Class=righttoleftoverride}', "");
Expect(1, 8239, '\P{Bidi_Class=righttoleftoverride}', "");
Expect(0, 8239, '\P{^Bidi_Class=righttoleftoverride}', "");
Expect(1, 8238, '\p{Bidi_Class=_	right_To_Left_Override}', "");
Expect(0, 8238, '\p{^Bidi_Class=_	right_To_Left_Override}', "");
Expect(0, 8238, '\P{Bidi_Class=_	right_To_Left_Override}', "");
Expect(1, 8238, '\P{^Bidi_Class=_	right_To_Left_Override}', "");
Expect(0, 8239, '\p{Bidi_Class=_	right_To_Left_Override}', "");
Expect(1, 8239, '\p{^Bidi_Class=_	right_To_Left_Override}', "");
Expect(1, 8239, '\P{Bidi_Class=_	right_To_Left_Override}', "");
Expect(0, 8239, '\P{^Bidi_Class=_	right_To_Left_Override}', "");
Error('\p{Bc=-	RLO/a/}');
Error('\P{Bc=-	RLO/a/}');
Expect(1, 8238, '\p{Bc=rlo}', "");
Expect(0, 8238, '\p{^Bc=rlo}', "");
Expect(0, 8238, '\P{Bc=rlo}', "");
Expect(1, 8238, '\P{^Bc=rlo}', "");
Expect(0, 8239, '\p{Bc=rlo}', "");
Expect(1, 8239, '\p{^Bc=rlo}', "");
Expect(1, 8239, '\P{Bc=rlo}', "");
Expect(0, 8239, '\P{^Bc=rlo}', "");
Expect(1, 8238, '\p{Bc=__RLO}', "");
Expect(0, 8238, '\p{^Bc=__RLO}', "");
Expect(0, 8238, '\P{Bc=__RLO}', "");
Expect(1, 8238, '\P{^Bc=__RLO}', "");
Expect(0, 8239, '\p{Bc=__RLO}', "");
Expect(1, 8239, '\p{^Bc=__RLO}', "");
Expect(1, 8239, '\P{Bc=__RLO}', "");
Expect(0, 8239, '\P{^Bc=__RLO}', "");
Error('\p{Is_Bidi_Class=/a/--Right_To_Left_override}');
Error('\P{Is_Bidi_Class=/a/--Right_To_Left_override}');
Expect(1, 8238, '\p{Is_Bidi_Class=righttoleftoverride}', "");
Expect(0, 8238, '\p{^Is_Bidi_Class=righttoleftoverride}', "");
Expect(0, 8238, '\P{Is_Bidi_Class=righttoleftoverride}', "");
Expect(1, 8238, '\P{^Is_Bidi_Class=righttoleftoverride}', "");
Expect(0, 8239, '\p{Is_Bidi_Class=righttoleftoverride}', "");
Expect(1, 8239, '\p{^Is_Bidi_Class=righttoleftoverride}', "");
Expect(1, 8239, '\P{Is_Bidi_Class=righttoleftoverride}', "");
Expect(0, 8239, '\P{^Is_Bidi_Class=righttoleftoverride}', "");
Expect(1, 8238, '\p{Is_Bidi_Class= _Right_to_left_override}', "");
Expect(0, 8238, '\p{^Is_Bidi_Class= _Right_to_left_override}', "");
Expect(0, 8238, '\P{Is_Bidi_Class= _Right_to_left_override}', "");
Expect(1, 8238, '\P{^Is_Bidi_Class= _Right_to_left_override}', "");
Expect(0, 8239, '\p{Is_Bidi_Class= _Right_to_left_override}', "");
Expect(1, 8239, '\p{^Is_Bidi_Class= _Right_to_left_override}', "");
Expect(1, 8239, '\P{Is_Bidi_Class= _Right_to_left_override}', "");
Expect(0, 8239, '\P{^Is_Bidi_Class= _Right_to_left_override}', "");
Error('\p{Is_Bc=:=_-RLO}');
Error('\P{Is_Bc=:=_-RLO}');
Expect(1, 8238, '\p{Is_Bc=rlo}', "");
Expect(0, 8238, '\p{^Is_Bc=rlo}', "");
Expect(0, 8238, '\P{Is_Bc=rlo}', "");
Expect(1, 8238, '\P{^Is_Bc=rlo}', "");
Expect(0, 8239, '\p{Is_Bc=rlo}', "");
Expect(1, 8239, '\p{^Is_Bc=rlo}', "");
Expect(1, 8239, '\P{Is_Bc=rlo}', "");
Expect(0, 8239, '\P{^Is_Bc=rlo}', "");
Expect(1, 8238, '\p{Is_Bc= 	rlo}', "");
Expect(0, 8238, '\p{^Is_Bc= 	rlo}', "");
Expect(0, 8238, '\P{Is_Bc= 	rlo}', "");
Expect(1, 8238, '\P{^Is_Bc= 	rlo}', "");
Expect(0, 8239, '\p{Is_Bc= 	rlo}', "");
Expect(1, 8239, '\p{^Is_Bc= 	rlo}', "");
Expect(1, 8239, '\P{Is_Bc= 	rlo}', "");
Expect(0, 8239, '\P{^Is_Bc= 	rlo}', "");
Error('\p{Bidi_Class=	segment_Separator:=}');
Error('\P{Bidi_Class=	segment_Separator:=}');
Expect(1, 31, '\p{Bidi_Class: segmentseparator}', "");
Expect(0, 31, '\p{^Bidi_Class: segmentseparator}', "");
Expect(0, 31, '\P{Bidi_Class: segmentseparator}', "");
Expect(1, 31, '\P{^Bidi_Class: segmentseparator}', "");
Expect(0, 32, '\p{Bidi_Class: segmentseparator}', "");
Expect(1, 32, '\p{^Bidi_Class: segmentseparator}', "");
Expect(1, 32, '\P{Bidi_Class: segmentseparator}', "");
Expect(0, 32, '\P{^Bidi_Class: segmentseparator}', "");
Expect(1, 31, '\p{Bidi_Class=	Segment_Separator}', "");
Expect(0, 31, '\p{^Bidi_Class=	Segment_Separator}', "");
Expect(0, 31, '\P{Bidi_Class=	Segment_Separator}', "");
Expect(1, 31, '\P{^Bidi_Class=	Segment_Separator}', "");
Expect(0, 32, '\p{Bidi_Class=	Segment_Separator}', "");
Expect(1, 32, '\p{^Bidi_Class=	Segment_Separator}', "");
Expect(1, 32, '\P{Bidi_Class=	Segment_Separator}', "");
Expect(0, 32, '\P{^Bidi_Class=	Segment_Separator}', "");
Error('\p{Bc=	-S/a/}');
Error('\P{Bc=	-S/a/}');
Expect(1, 31, '\p{Bc=s}', "");
Expect(0, 31, '\p{^Bc=s}', "");
Expect(0, 31, '\P{Bc=s}', "");
Expect(1, 31, '\P{^Bc=s}', "");
Expect(0, 32, '\p{Bc=s}', "");
Expect(1, 32, '\p{^Bc=s}', "");
Expect(1, 32, '\P{Bc=s}', "");
Expect(0, 32, '\P{^Bc=s}', "");
Expect(1, 31, '\p{Bc:__s}', "");
Expect(0, 31, '\p{^Bc:__s}', "");
Expect(0, 31, '\P{Bc:__s}', "");
Expect(1, 31, '\P{^Bc:__s}', "");
Expect(0, 32, '\p{Bc:__s}', "");
Expect(1, 32, '\p{^Bc:__s}', "");
Expect(1, 32, '\P{Bc:__s}', "");
Expect(0, 32, '\P{^Bc:__s}', "");
Error('\p{Is_Bidi_Class=Segment_Separator/a/}');
Error('\P{Is_Bidi_Class=Segment_Separator/a/}');
Expect(1, 31, '\p{Is_Bidi_Class=segmentseparator}', "");
Expect(0, 31, '\p{^Is_Bidi_Class=segmentseparator}', "");
Expect(0, 31, '\P{Is_Bidi_Class=segmentseparator}', "");
Expect(1, 31, '\P{^Is_Bidi_Class=segmentseparator}', "");
Expect(0, 32, '\p{Is_Bidi_Class=segmentseparator}', "");
Expect(1, 32, '\p{^Is_Bidi_Class=segmentseparator}', "");
Expect(1, 32, '\P{Is_Bidi_Class=segmentseparator}', "");
Expect(0, 32, '\P{^Is_Bidi_Class=segmentseparator}', "");
Expect(1, 31, '\p{Is_Bidi_Class=  Segment_Separator}', "");
Expect(0, 31, '\p{^Is_Bidi_Class=  Segment_Separator}', "");
Expect(0, 31, '\P{Is_Bidi_Class=  Segment_Separator}', "");
Expect(1, 31, '\P{^Is_Bidi_Class=  Segment_Separator}', "");
Expect(0, 32, '\p{Is_Bidi_Class=  Segment_Separator}', "");
Expect(1, 32, '\p{^Is_Bidi_Class=  Segment_Separator}', "");
Expect(1, 32, '\P{Is_Bidi_Class=  Segment_Separator}', "");
Expect(0, 32, '\P{^Is_Bidi_Class=  Segment_Separator}', "");
Error('\p{Is_Bc=-/a/S}');
Error('\P{Is_Bc=-/a/S}');
Expect(1, 31, '\p{Is_Bc=s}', "");
Expect(0, 31, '\p{^Is_Bc=s}', "");
Expect(0, 31, '\P{Is_Bc=s}', "");
Expect(1, 31, '\P{^Is_Bc=s}', "");
Expect(0, 32, '\p{Is_Bc=s}', "");
Expect(1, 32, '\p{^Is_Bc=s}', "");
Expect(1, 32, '\P{Is_Bc=s}', "");
Expect(0, 32, '\P{^Is_Bc=s}', "");
Expect(1, 31, '\p{Is_Bc=-	S}', "");
Expect(0, 31, '\p{^Is_Bc=-	S}', "");
Expect(0, 31, '\P{Is_Bc=-	S}', "");
Expect(1, 31, '\P{^Is_Bc=-	S}', "");
Expect(0, 32, '\p{Is_Bc=-	S}', "");
Expect(1, 32, '\p{^Is_Bc=-	S}', "");
Expect(1, 32, '\P{Is_Bc=-	S}', "");
Expect(0, 32, '\P{^Is_Bc=-	S}', "");
Error('\p{Bidi_Class= white_Space/a/}');
Error('\P{Bidi_Class= white_Space/a/}');
Expect(1, 12288, '\p{Bidi_Class=whitespace}', "");
Expect(0, 12288, '\p{^Bidi_Class=whitespace}', "");
Expect(0, 12288, '\P{Bidi_Class=whitespace}', "");
Expect(1, 12288, '\P{^Bidi_Class=whitespace}', "");
Expect(0, 12289, '\p{Bidi_Class=whitespace}', "");
Expect(1, 12289, '\p{^Bidi_Class=whitespace}', "");
Expect(1, 12289, '\P{Bidi_Class=whitespace}', "");
Expect(0, 12289, '\P{^Bidi_Class=whitespace}', "");
Expect(1, 12288, '\p{Bidi_Class:		_white_SPACE}', "");
Expect(0, 12288, '\p{^Bidi_Class:		_white_SPACE}', "");
Expect(0, 12288, '\P{Bidi_Class:		_white_SPACE}', "");
Expect(1, 12288, '\P{^Bidi_Class:		_white_SPACE}', "");
Expect(0, 12289, '\p{Bidi_Class:		_white_SPACE}', "");
Expect(1, 12289, '\p{^Bidi_Class:		_white_SPACE}', "");
Expect(1, 12289, '\P{Bidi_Class:		_white_SPACE}', "");
Expect(0, 12289, '\P{^Bidi_Class:		_white_SPACE}', "");
Error('\p{Bc=-:=WS}');
Error('\P{Bc=-:=WS}');
Expect(1, 12288, '\p{Bc=ws}', "");
Expect(0, 12288, '\p{^Bc=ws}', "");
Expect(0, 12288, '\P{Bc=ws}', "");
Expect(1, 12288, '\P{^Bc=ws}', "");
Expect(0, 12289, '\p{Bc=ws}', "");
Expect(1, 12289, '\p{^Bc=ws}', "");
Expect(1, 12289, '\P{Bc=ws}', "");
Expect(0, 12289, '\P{^Bc=ws}', "");
Expect(1, 12288, '\p{Bc=	_WS}', "");
Expect(0, 12288, '\p{^Bc=	_WS}', "");
Expect(0, 12288, '\P{Bc=	_WS}', "");
Expect(1, 12288, '\P{^Bc=	_WS}', "");
Expect(0, 12289, '\p{Bc=	_WS}', "");
Expect(1, 12289, '\p{^Bc=	_WS}', "");
Expect(1, 12289, '\P{Bc=	_WS}', "");
Expect(0, 12289, '\P{^Bc=	_WS}', "");
Error('\p{Is_Bidi_Class= 	White_SPACE/a/}');
Error('\P{Is_Bidi_Class= 	White_SPACE/a/}');
Expect(1, 12288, '\p{Is_Bidi_Class:	whitespace}', "");
Expect(0, 12288, '\p{^Is_Bidi_Class:	whitespace}', "");
Expect(0, 12288, '\P{Is_Bidi_Class:	whitespace}', "");
Expect(1, 12288, '\P{^Is_Bidi_Class:	whitespace}', "");
Expect(0, 12289, '\p{Is_Bidi_Class:	whitespace}', "");
Expect(1, 12289, '\p{^Is_Bidi_Class:	whitespace}', "");
Expect(1, 12289, '\P{Is_Bidi_Class:	whitespace}', "");
Expect(0, 12289, '\P{^Is_Bidi_Class:	whitespace}', "");
Expect(1, 12288, '\p{Is_Bidi_Class=- white_space}', "");
Expect(0, 12288, '\p{^Is_Bidi_Class=- white_space}', "");
Expect(0, 12288, '\P{Is_Bidi_Class=- white_space}', "");
Expect(1, 12288, '\P{^Is_Bidi_Class=- white_space}', "");
Expect(0, 12289, '\p{Is_Bidi_Class=- white_space}', "");
Expect(1, 12289, '\p{^Is_Bidi_Class=- white_space}', "");
Expect(1, 12289, '\P{Is_Bidi_Class=- white_space}', "");
Expect(0, 12289, '\P{^Is_Bidi_Class=- white_space}', "");
Error('\p{Is_Bc=  WS/a/}');
Error('\P{Is_Bc=  WS/a/}');
Expect(1, 12288, '\p{Is_Bc=ws}', "");
Expect(0, 12288, '\p{^Is_Bc=ws}', "");
Expect(0, 12288, '\P{Is_Bc=ws}', "");
Expect(1, 12288, '\P{^Is_Bc=ws}', "");
Expect(0, 12289, '\p{Is_Bc=ws}', "");
Expect(1, 12289, '\p{^Is_Bc=ws}', "");
Expect(1, 12289, '\P{Is_Bc=ws}', "");
Expect(0, 12289, '\P{^Is_Bc=ws}', "");
Expect(1, 12288, '\p{Is_Bc: WS}', "");
Expect(0, 12288, '\p{^Is_Bc: WS}', "");
Expect(0, 12288, '\P{Is_Bc: WS}', "");
Expect(1, 12288, '\P{^Is_Bc: WS}', "");
Expect(0, 12289, '\p{Is_Bc: WS}', "");
Expect(1, 12289, '\p{^Is_Bc: WS}', "");
Expect(1, 12289, '\P{Is_Bc: WS}', "");
Expect(0, 12289, '\P{^Is_Bc: WS}', "");
Error('\p{Block= 	Basic_latin:=}');
Error('\P{Block= 	Basic_latin:=}');
Expect(1, 127, '\p{Block=basiclatin}', "");
Expect(0, 127, '\p{^Block=basiclatin}', "");
Expect(0, 127, '\P{Block=basiclatin}', "");
Expect(1, 127, '\P{^Block=basiclatin}', "");
Expect(0, 128, '\p{Block=basiclatin}', "");
Expect(1, 128, '\p{^Block=basiclatin}', "");
Expect(1, 128, '\P{Block=basiclatin}', "");
Expect(0, 128, '\P{^Block=basiclatin}', "");
Expect(1, 127, '\p{Block= BASIC_Latin}', "");
Expect(0, 127, '\p{^Block= BASIC_Latin}', "");
Expect(0, 127, '\P{Block= BASIC_Latin}', "");
Expect(1, 127, '\P{^Block= BASIC_Latin}', "");
Expect(0, 128, '\p{Block= BASIC_Latin}', "");
Expect(1, 128, '\p{^Block= BASIC_Latin}', "");
Expect(1, 128, '\P{Block= BASIC_Latin}', "");
Expect(0, 128, '\P{^Block= BASIC_Latin}', "");
Error('\p{Blk=_/a/ASCII}');
Error('\P{Blk=_/a/ASCII}');
Expect(1, 127, '\p{Blk=ascii}', "");
Expect(0, 127, '\p{^Blk=ascii}', "");
Expect(0, 127, '\P{Blk=ascii}', "");
Expect(1, 127, '\P{^Blk=ascii}', "");
Expect(0, 128, '\p{Blk=ascii}', "");
Expect(1, 128, '\p{^Blk=ascii}', "");
Expect(1, 128, '\P{Blk=ascii}', "");
Expect(0, 128, '\P{^Blk=ascii}', "");
Expect(1, 127, '\p{Blk:	_-ascii}', "");
Expect(0, 127, '\p{^Blk:	_-ascii}', "");
Expect(0, 127, '\P{Blk:	_-ascii}', "");
Expect(1, 127, '\P{^Blk:	_-ascii}', "");
Expect(0, 128, '\p{Blk:	_-ascii}', "");
Expect(1, 128, '\p{^Blk:	_-ascii}', "");
Expect(1, 128, '\P{Blk:	_-ascii}', "");
Expect(0, 128, '\P{^Blk:	_-ascii}', "");
Error('\p{Is_Block=_/a/Basic_Latin}');
Error('\P{Is_Block=_/a/Basic_Latin}');
Expect(1, 127, '\p{Is_Block=basiclatin}', "");
Expect(0, 127, '\p{^Is_Block=basiclatin}', "");
Expect(0, 127, '\P{Is_Block=basiclatin}', "");
Expect(1, 127, '\P{^Is_Block=basiclatin}', "");
Expect(0, 128, '\p{Is_Block=basiclatin}', "");
Expect(1, 128, '\p{^Is_Block=basiclatin}', "");
Expect(1, 128, '\P{Is_Block=basiclatin}', "");
Expect(0, 128, '\P{^Is_Block=basiclatin}', "");
Expect(1, 127, '\p{Is_Block=_basic_latin}', "");
Expect(0, 127, '\p{^Is_Block=_basic_latin}', "");
Expect(0, 127, '\P{Is_Block=_basic_latin}', "");
Expect(1, 127, '\P{^Is_Block=_basic_latin}', "");
Expect(0, 128, '\p{Is_Block=_basic_latin}', "");
Expect(1, 128, '\p{^Is_Block=_basic_latin}', "");
Expect(1, 128, '\P{Is_Block=_basic_latin}', "");
Expect(0, 128, '\P{^Is_Block=_basic_latin}', "");
Error('\p{Is_Blk=-	ASCII:=}');
Error('\P{Is_Blk=-	ASCII:=}');
Expect(1, 127, '\p{Is_Blk:   ascii}', "");
Expect(0, 127, '\p{^Is_Blk:   ascii}', "");
Expect(0, 127, '\P{Is_Blk:   ascii}', "");
Expect(1, 127, '\P{^Is_Blk:   ascii}', "");
Expect(0, 128, '\p{Is_Blk:   ascii}', "");
Expect(1, 128, '\p{^Is_Blk:   ascii}', "");
Expect(1, 128, '\P{Is_Blk:   ascii}', "");
Expect(0, 128, '\P{^Is_Blk:   ascii}', "");
Expect(1, 127, '\p{Is_Blk= _ascii}', "");
Expect(0, 127, '\p{^Is_Blk= _ascii}', "");
Expect(0, 127, '\P{Is_Blk= _ascii}', "");
Expect(1, 127, '\P{^Is_Blk= _ascii}', "");
Expect(0, 128, '\p{Is_Blk= _ascii}', "");
Expect(1, 128, '\p{^Is_Blk= _ascii}', "");
Expect(1, 128, '\P{Is_Blk= _ascii}', "");
Expect(0, 128, '\P{^Is_Blk= _ascii}', "");
Error('\p{Block=	-Aegean_numbers/a/}');
Error('\P{Block=	-Aegean_numbers/a/}');
Expect(1, 65855, '\p{Block=aegeannumbers}', "");
Expect(0, 65855, '\p{^Block=aegeannumbers}', "");
Expect(0, 65855, '\P{Block=aegeannumbers}', "");
Expect(1, 65855, '\P{^Block=aegeannumbers}', "");
Expect(0, 65856, '\p{Block=aegeannumbers}', "");
Expect(1, 65856, '\p{^Block=aegeannumbers}', "");
Expect(1, 65856, '\P{Block=aegeannumbers}', "");
Expect(0, 65856, '\P{^Block=aegeannumbers}', "");
Expect(1, 65855, '\p{Block:	-Aegean_Numbers}', "");
Expect(0, 65855, '\p{^Block:	-Aegean_Numbers}', "");
Expect(0, 65855, '\P{Block:	-Aegean_Numbers}', "");
Expect(1, 65855, '\P{^Block:	-Aegean_Numbers}', "");
Expect(0, 65856, '\p{Block:	-Aegean_Numbers}', "");
Expect(1, 65856, '\p{^Block:	-Aegean_Numbers}', "");
Expect(1, 65856, '\P{Block:	-Aegean_Numbers}', "");
Expect(0, 65856, '\P{^Block:	-Aegean_Numbers}', "");
Error('\p{Blk=_	AEGEAN_NUMBERS/a/}');
Error('\P{Blk=_	AEGEAN_NUMBERS/a/}');
Expect(1, 65855, '\p{Blk:aegeannumbers}', "");
Expect(0, 65855, '\p{^Blk:aegeannumbers}', "");
Expect(0, 65855, '\P{Blk:aegeannumbers}', "");
Expect(1, 65855, '\P{^Blk:aegeannumbers}', "");
Expect(0, 65856, '\p{Blk:aegeannumbers}', "");
Expect(1, 65856, '\p{^Blk:aegeannumbers}', "");
Expect(1, 65856, '\P{Blk:aegeannumbers}', "");
Expect(0, 65856, '\P{^Blk:aegeannumbers}', "");
Expect(1, 65855, '\p{Blk= Aegean_numbers}', "");
Expect(0, 65855, '\p{^Blk= Aegean_numbers}', "");
Expect(0, 65855, '\P{Blk= Aegean_numbers}', "");
Expect(1, 65855, '\P{^Blk= Aegean_numbers}', "");
Expect(0, 65856, '\p{Blk= Aegean_numbers}', "");
Expect(1, 65856, '\p{^Blk= Aegean_numbers}', "");
Expect(1, 65856, '\P{Blk= Aegean_numbers}', "");
Expect(0, 65856, '\P{^Blk= Aegean_numbers}', "");
Error('\p{Is_Block=/a/_AEGEAN_numbers}');
Error('\P{Is_Block=/a/_AEGEAN_numbers}');
Expect(1, 65855, '\p{Is_Block=aegeannumbers}', "");
Expect(0, 65855, '\p{^Is_Block=aegeannumbers}', "");
Expect(0, 65855, '\P{Is_Block=aegeannumbers}', "");
Expect(1, 65855, '\P{^Is_Block=aegeannumbers}', "");
Expect(0, 65856, '\p{Is_Block=aegeannumbers}', "");
Expect(1, 65856, '\p{^Is_Block=aegeannumbers}', "");
Expect(1, 65856, '\P{Is_Block=aegeannumbers}', "");
Expect(0, 65856, '\P{^Is_Block=aegeannumbers}', "");
Expect(1, 65855, '\p{Is_Block=--AEGEAN_NUMBERS}', "");
Expect(0, 65855, '\p{^Is_Block=--AEGEAN_NUMBERS}', "");
Expect(0, 65855, '\P{Is_Block=--AEGEAN_NUMBERS}', "");
Expect(1, 65855, '\P{^Is_Block=--AEGEAN_NUMBERS}', "");
Expect(0, 65856, '\p{Is_Block=--AEGEAN_NUMBERS}', "");
Expect(1, 65856, '\p{^Is_Block=--AEGEAN_NUMBERS}', "");
Expect(1, 65856, '\P{Is_Block=--AEGEAN_NUMBERS}', "");
Expect(0, 65856, '\P{^Is_Block=--AEGEAN_NUMBERS}', "");
Error('\p{Is_Blk= Aegean_numbers/a/}');
Error('\P{Is_Blk= Aegean_numbers/a/}');
Expect(1, 65855, '\p{Is_Blk=aegeannumbers}', "");
Expect(0, 65855, '\p{^Is_Blk=aegeannumbers}', "");
Expect(0, 65855, '\P{Is_Blk=aegeannumbers}', "");
Expect(1, 65855, '\P{^Is_Blk=aegeannumbers}', "");
Expect(0, 65856, '\p{Is_Blk=aegeannumbers}', "");
Expect(1, 65856, '\p{^Is_Blk=aegeannumbers}', "");
Expect(1, 65856, '\P{Is_Blk=aegeannumbers}', "");
Expect(0, 65856, '\P{^Is_Blk=aegeannumbers}', "");
Expect(1, 65855, '\p{Is_Blk=- aegean_NUMBERS}', "");
Expect(0, 65855, '\p{^Is_Blk=- aegean_NUMBERS}', "");
Expect(0, 65855, '\P{Is_Blk=- aegean_NUMBERS}', "");
Expect(1, 65855, '\P{^Is_Blk=- aegean_NUMBERS}', "");
Expect(0, 65856, '\p{Is_Blk=- aegean_NUMBERS}', "");
Expect(1, 65856, '\p{^Is_Blk=- aegean_NUMBERS}', "");
Expect(1, 65856, '\P{Is_Blk=- aegean_NUMBERS}', "");
Expect(0, 65856, '\P{^Is_Blk=- aegean_NUMBERS}', "");
Error('\p{Block:   alchemical_Symbols:=}');
Error('\P{Block:   alchemical_Symbols:=}');
Expect(1, 128895, '\p{Block=alchemicalsymbols}', "");
Expect(0, 128895, '\p{^Block=alchemicalsymbols}', "");
Expect(0, 128895, '\P{Block=alchemicalsymbols}', "");
Expect(1, 128895, '\P{^Block=alchemicalsymbols}', "");
Expect(0, 128896, '\p{Block=alchemicalsymbols}', "");
Expect(1, 128896, '\p{^Block=alchemicalsymbols}', "");
Expect(1, 128896, '\P{Block=alchemicalsymbols}', "");
Expect(0, 128896, '\P{^Block=alchemicalsymbols}', "");
Expect(1, 128895, '\p{Block= ALCHEMICAL_SYMBOLS}', "");
Expect(0, 128895, '\p{^Block= ALCHEMICAL_SYMBOLS}', "");
Expect(0, 128895, '\P{Block= ALCHEMICAL_SYMBOLS}', "");
Expect(1, 128895, '\P{^Block= ALCHEMICAL_SYMBOLS}', "");
Expect(0, 128896, '\p{Block= ALCHEMICAL_SYMBOLS}', "");
Expect(1, 128896, '\p{^Block= ALCHEMICAL_SYMBOLS}', "");
Expect(1, 128896, '\P{Block= ALCHEMICAL_SYMBOLS}', "");
Expect(0, 128896, '\P{^Block= ALCHEMICAL_SYMBOLS}', "");
Error('\p{Blk=/a/-Alchemical}');
Error('\P{Blk=/a/-Alchemical}');
Expect(1, 128895, '\p{Blk=alchemical}', "");
Expect(0, 128895, '\p{^Blk=alchemical}', "");
Expect(0, 128895, '\P{Blk=alchemical}', "");
Expect(1, 128895, '\P{^Blk=alchemical}', "");
Expect(0, 128896, '\p{Blk=alchemical}', "");
Expect(1, 128896, '\p{^Blk=alchemical}', "");
Expect(1, 128896, '\P{Blk=alchemical}', "");
Expect(0, 128896, '\P{^Blk=alchemical}', "");
Expect(1, 128895, '\p{Blk=__Alchemical}', "");
Expect(0, 128895, '\p{^Blk=__Alchemical}', "");
Expect(0, 128895, '\P{Blk=__Alchemical}', "");
Expect(1, 128895, '\P{^Blk=__Alchemical}', "");
Expect(0, 128896, '\p{Blk=__Alchemical}', "");
Expect(1, 128896, '\p{^Blk=__Alchemical}', "");
Expect(1, 128896, '\P{Blk=__Alchemical}', "");
Expect(0, 128896, '\P{^Blk=__Alchemical}', "");
Error('\p{Is_Block:	/a/--alchemical_Symbols}');
Error('\P{Is_Block:	/a/--alchemical_Symbols}');
Expect(1, 128895, '\p{Is_Block=alchemicalsymbols}', "");
Expect(0, 128895, '\p{^Is_Block=alchemicalsymbols}', "");
Expect(0, 128895, '\P{Is_Block=alchemicalsymbols}', "");
Expect(1, 128895, '\P{^Is_Block=alchemicalsymbols}', "");
Expect(0, 128896, '\p{Is_Block=alchemicalsymbols}', "");
Expect(1, 128896, '\p{^Is_Block=alchemicalsymbols}', "");
Expect(1, 128896, '\P{Is_Block=alchemicalsymbols}', "");
Expect(0, 128896, '\P{^Is_Block=alchemicalsymbols}', "");
Expect(1, 128895, '\p{Is_Block=-Alchemical_Symbols}', "");
Expect(0, 128895, '\p{^Is_Block=-Alchemical_Symbols}', "");
Expect(0, 128895, '\P{Is_Block=-Alchemical_Symbols}', "");
Expect(1, 128895, '\P{^Is_Block=-Alchemical_Symbols}', "");
Expect(0, 128896, '\p{Is_Block=-Alchemical_Symbols}', "");
Expect(1, 128896, '\p{^Is_Block=-Alchemical_Symbols}', "");
Expect(1, 128896, '\P{Is_Block=-Alchemical_Symbols}', "");
Expect(0, 128896, '\P{^Is_Block=-Alchemical_Symbols}', "");
Error('\p{Is_Blk=--Alchemical/a/}');
Error('\P{Is_Blk=--Alchemical/a/}');
Expect(1, 128895, '\p{Is_Blk=alchemical}', "");
Expect(0, 128895, '\p{^Is_Blk=alchemical}', "");
Expect(0, 128895, '\P{Is_Blk=alchemical}', "");
Expect(1, 128895, '\P{^Is_Blk=alchemical}', "");
Expect(0, 128896, '\p{Is_Blk=alchemical}', "");
Expect(1, 128896, '\p{^Is_Blk=alchemical}', "");
Expect(1, 128896, '\P{Is_Blk=alchemical}', "");
Expect(0, 128896, '\P{^Is_Blk=alchemical}', "");
Expect(1, 128895, '\p{Is_Blk=	 Alchemical}', "");
Expect(0, 128895, '\p{^Is_Blk=	 Alchemical}', "");
Expect(0, 128895, '\P{Is_Blk=	 Alchemical}', "");
Expect(1, 128895, '\P{^Is_Blk=	 Alchemical}', "");
Expect(0, 128896, '\p{Is_Blk=	 Alchemical}', "");
Expect(1, 128896, '\p{^Is_Blk=	 Alchemical}', "");
Expect(1, 128896, '\P{Is_Blk=	 Alchemical}', "");
Expect(0, 128896, '\P{^Is_Blk=	 Alchemical}', "");
Error('\p{Block=/a/--Alphabetic_Presentation_FORMS}');
Error('\P{Block=/a/--Alphabetic_Presentation_FORMS}');
Expect(1, 64335, '\p{Block=alphabeticpresentationforms}', "");
Expect(0, 64335, '\p{^Block=alphabeticpresentationforms}', "");
Expect(0, 64335, '\P{Block=alphabeticpresentationforms}', "");
Expect(1, 64335, '\P{^Block=alphabeticpresentationforms}', "");
Expect(0, 64336, '\p{Block=alphabeticpresentationforms}', "");
Expect(1, 64336, '\p{^Block=alphabeticpresentationforms}', "");
Expect(1, 64336, '\P{Block=alphabeticpresentationforms}', "");
Expect(0, 64336, '\P{^Block=alphabeticpresentationforms}', "");
Expect(1, 64335, '\p{Block=-ALPHABETIC_presentation_FORMS}', "");
Expect(0, 64335, '\p{^Block=-ALPHABETIC_presentation_FORMS}', "");
Expect(0, 64335, '\P{Block=-ALPHABETIC_presentation_FORMS}', "");
Expect(1, 64335, '\P{^Block=-ALPHABETIC_presentation_FORMS}', "");
Expect(0, 64336, '\p{Block=-ALPHABETIC_presentation_FORMS}', "");
Expect(1, 64336, '\p{^Block=-ALPHABETIC_presentation_FORMS}', "");
Expect(1, 64336, '\P{Block=-ALPHABETIC_presentation_FORMS}', "");
Expect(0, 64336, '\P{^Block=-ALPHABETIC_presentation_FORMS}', "");
Error('\p{Blk=/a/_Alphabetic_PF}');
Error('\P{Blk=/a/_Alphabetic_PF}');
Expect(1, 64335, '\p{Blk=alphabeticpf}', "");
Expect(0, 64335, '\p{^Blk=alphabeticpf}', "");
Expect(0, 64335, '\P{Blk=alphabeticpf}', "");
Expect(1, 64335, '\P{^Blk=alphabeticpf}', "");
Expect(0, 64336, '\p{Blk=alphabeticpf}', "");
Expect(1, 64336, '\p{^Blk=alphabeticpf}', "");
Expect(1, 64336, '\P{Blk=alphabeticpf}', "");
Expect(0, 64336, '\P{^Blk=alphabeticpf}', "");
Expect(1, 64335, '\p{Blk=_alphabetic_PF}', "");
Expect(0, 64335, '\p{^Blk=_alphabetic_PF}', "");
Expect(0, 64335, '\P{Blk=_alphabetic_PF}', "");
Expect(1, 64335, '\P{^Blk=_alphabetic_PF}', "");
Expect(0, 64336, '\p{Blk=_alphabetic_PF}', "");
Expect(1, 64336, '\p{^Blk=_alphabetic_PF}', "");
Expect(1, 64336, '\P{Blk=_alphabetic_PF}', "");
Expect(0, 64336, '\P{^Blk=_alphabetic_PF}', "");
Error('\p{Is_Block=--Alphabetic_PRESENTATION_Forms/a/}');
Error('\P{Is_Block=--Alphabetic_PRESENTATION_Forms/a/}');
Expect(1, 64335, '\p{Is_Block=alphabeticpresentationforms}', "");
Expect(0, 64335, '\p{^Is_Block=alphabeticpresentationforms}', "");
Expect(0, 64335, '\P{Is_Block=alphabeticpresentationforms}', "");
Expect(1, 64335, '\P{^Is_Block=alphabeticpresentationforms}', "");
Expect(0, 64336, '\p{Is_Block=alphabeticpresentationforms}', "");
Expect(1, 64336, '\p{^Is_Block=alphabeticpresentationforms}', "");
Expect(1, 64336, '\P{Is_Block=alphabeticpresentationforms}', "");
Expect(0, 64336, '\P{^Is_Block=alphabeticpresentationforms}', "");
Expect(1, 64335, '\p{Is_Block:   		Alphabetic_presentation_Forms}', "");
Expect(0, 64335, '\p{^Is_Block:   		Alphabetic_presentation_Forms}', "");
Expect(0, 64335, '\P{Is_Block:   		Alphabetic_presentation_Forms}', "");
Expect(1, 64335, '\P{^Is_Block:   		Alphabetic_presentation_Forms}', "");
Expect(0, 64336, '\p{Is_Block:   		Alphabetic_presentation_Forms}', "");
Expect(1, 64336, '\p{^Is_Block:   		Alphabetic_presentation_Forms}', "");
Expect(1, 64336, '\P{Is_Block:   		Alphabetic_presentation_Forms}', "");
Expect(0, 64336, '\P{^Is_Block:   		Alphabetic_presentation_Forms}', "");
Error('\p{Is_Blk=:=	_Alphabetic_pf}');
Error('\P{Is_Blk=:=	_Alphabetic_pf}');
Expect(1, 64335, '\p{Is_Blk=alphabeticpf}', "");
Expect(0, 64335, '\p{^Is_Blk=alphabeticpf}', "");
Expect(0, 64335, '\P{Is_Blk=alphabeticpf}', "");
Expect(1, 64335, '\P{^Is_Blk=alphabeticpf}', "");
Expect(0, 64336, '\p{Is_Blk=alphabeticpf}', "");
Expect(1, 64336, '\p{^Is_Blk=alphabeticpf}', "");
Expect(1, 64336, '\P{Is_Blk=alphabeticpf}', "");
Expect(0, 64336, '\P{^Is_Blk=alphabeticpf}', "");
Expect(1, 64335, '\p{Is_Blk:	__ALPHABETIC_pf}', "");
Expect(0, 64335, '\p{^Is_Blk:	__ALPHABETIC_pf}', "");
Expect(0, 64335, '\P{Is_Blk:	__ALPHABETIC_pf}', "");
Expect(1, 64335, '\P{^Is_Blk:	__ALPHABETIC_pf}', "");
Expect(0, 64336, '\p{Is_Blk:	__ALPHABETIC_pf}', "");
Expect(1, 64336, '\p{^Is_Blk:	__ALPHABETIC_pf}', "");
Expect(1, 64336, '\P{Is_Blk:	__ALPHABETIC_pf}', "");
Expect(0, 64336, '\P{^Is_Blk:	__ALPHABETIC_pf}', "");
Error('\p{Block=:=__Ancient_Greek_Musical_Notation}');
Error('\P{Block=:=__Ancient_Greek_Musical_Notation}');
Expect(1, 119375, '\p{Block=ancientgreekmusicalnotation}', "");
Expect(0, 119375, '\p{^Block=ancientgreekmusicalnotation}', "");
Expect(0, 119375, '\P{Block=ancientgreekmusicalnotation}', "");
Expect(1, 119375, '\P{^Block=ancientgreekmusicalnotation}', "");
Expect(0, 119376, '\p{Block=ancientgreekmusicalnotation}', "");
Expect(1, 119376, '\p{^Block=ancientgreekmusicalnotation}', "");
Expect(1, 119376, '\P{Block=ancientgreekmusicalnotation}', "");
Expect(0, 119376, '\P{^Block=ancientgreekmusicalnotation}', "");
Expect(1, 119375, '\p{Block=		Ancient_Greek_Musical_Notation}', "");
Expect(0, 119375, '\p{^Block=		Ancient_Greek_Musical_Notation}', "");
Expect(0, 119375, '\P{Block=		Ancient_Greek_Musical_Notation}', "");
Expect(1, 119375, '\P{^Block=		Ancient_Greek_Musical_Notation}', "");
Expect(0, 119376, '\p{Block=		Ancient_Greek_Musical_Notation}', "");
Expect(1, 119376, '\p{^Block=		Ancient_Greek_Musical_Notation}', "");
Expect(1, 119376, '\P{Block=		Ancient_Greek_Musical_Notation}', "");
Expect(0, 119376, '\P{^Block=		Ancient_Greek_Musical_Notation}', "");
Error('\p{Blk=	/a/ancient_Greek_MUSIC}');
Error('\P{Blk=	/a/ancient_Greek_MUSIC}');
Expect(1, 119375, '\p{Blk=ancientgreekmusic}', "");
Expect(0, 119375, '\p{^Blk=ancientgreekmusic}', "");
Expect(0, 119375, '\P{Blk=ancientgreekmusic}', "");
Expect(1, 119375, '\P{^Blk=ancientgreekmusic}', "");
Expect(0, 119376, '\p{Blk=ancientgreekmusic}', "");
Expect(1, 119376, '\p{^Blk=ancientgreekmusic}', "");
Expect(1, 119376, '\P{Blk=ancientgreekmusic}', "");
Expect(0, 119376, '\P{^Blk=ancientgreekmusic}', "");
Expect(1, 119375, '\p{Blk=_	Ancient_greek_music}', "");
Expect(0, 119375, '\p{^Blk=_	Ancient_greek_music}', "");
Expect(0, 119375, '\P{Blk=_	Ancient_greek_music}', "");
Expect(1, 119375, '\P{^Blk=_	Ancient_greek_music}', "");
Expect(0, 119376, '\p{Blk=_	Ancient_greek_music}', "");
Expect(1, 119376, '\p{^Blk=_	Ancient_greek_music}', "");
Expect(1, 119376, '\P{Blk=_	Ancient_greek_music}', "");
Expect(0, 119376, '\P{^Blk=_	Ancient_greek_music}', "");
Error('\p{Is_Block: 	ANCIENT_GREEK_MUSICAL_Notation:=}');
Error('\P{Is_Block: 	ANCIENT_GREEK_MUSICAL_Notation:=}');
Expect(1, 119375, '\p{Is_Block=ancientgreekmusicalnotation}', "");
Expect(0, 119375, '\p{^Is_Block=ancientgreekmusicalnotation}', "");
Expect(0, 119375, '\P{Is_Block=ancientgreekmusicalnotation}', "");
Expect(1, 119375, '\P{^Is_Block=ancientgreekmusicalnotation}', "");
Expect(0, 119376, '\p{Is_Block=ancientgreekmusicalnotation}', "");
Expect(1, 119376, '\p{^Is_Block=ancientgreekmusicalnotation}', "");
Expect(1, 119376, '\P{Is_Block=ancientgreekmusicalnotation}', "");
Expect(0, 119376, '\P{^Is_Block=ancientgreekmusicalnotation}', "");
Expect(1, 119375, '\p{Is_Block=		ANCIENT_Greek_MUSICAL_NOTATION}', "");
Expect(0, 119375, '\p{^Is_Block=		ANCIENT_Greek_MUSICAL_NOTATION}', "");
Expect(0, 119375, '\P{Is_Block=		ANCIENT_Greek_MUSICAL_NOTATION}', "");
Expect(1, 119375, '\P{^Is_Block=		ANCIENT_Greek_MUSICAL_NOTATION}', "");
Expect(0, 119376, '\p{Is_Block=		ANCIENT_Greek_MUSICAL_NOTATION}', "");
Expect(1, 119376, '\p{^Is_Block=		ANCIENT_Greek_MUSICAL_NOTATION}', "");
Expect(1, 119376, '\P{Is_Block=		ANCIENT_Greek_MUSICAL_NOTATION}', "");
Expect(0, 119376, '\P{^Is_Block=		ANCIENT_Greek_MUSICAL_NOTATION}', "");
Error('\p{Is_Blk=__Ancient_Greek_Music:=}');
Error('\P{Is_Blk=__Ancient_Greek_Music:=}');
Expect(1, 119375, '\p{Is_Blk:	ancientgreekmusic}', "");
Expect(0, 119375, '\p{^Is_Blk:	ancientgreekmusic}', "");
Expect(0, 119375, '\P{Is_Blk:	ancientgreekmusic}', "");
Expect(1, 119375, '\P{^Is_Blk:	ancientgreekmusic}', "");
Expect(0, 119376, '\p{Is_Blk:	ancientgreekmusic}', "");
Expect(1, 119376, '\p{^Is_Blk:	ancientgreekmusic}', "");
Expect(1, 119376, '\P{Is_Blk:	ancientgreekmusic}', "");
Expect(0, 119376, '\P{^Is_Blk:	ancientgreekmusic}', "");
Expect(1, 119375, '\p{Is_Blk=_	Ancient_greek_Music}', "");
Expect(0, 119375, '\p{^Is_Blk=_	Ancient_greek_Music}', "");
Expect(0, 119375, '\P{Is_Blk=_	Ancient_greek_Music}', "");
Expect(1, 119375, '\P{^Is_Blk=_	Ancient_greek_Music}', "");
Expect(0, 119376, '\p{Is_Blk=_	Ancient_greek_Music}', "");
Expect(1, 119376, '\p{^Is_Blk=_	Ancient_greek_Music}', "");
Expect(1, 119376, '\P{Is_Blk=_	Ancient_greek_Music}', "");
Expect(0, 119376, '\P{^Is_Blk=_	Ancient_greek_Music}', "");
Error('\p{Block= ancient_Greek_NUMBERS/a/}');
Error('\P{Block= ancient_Greek_NUMBERS/a/}');
Expect(1, 65935, '\p{Block=ancientgreeknumbers}', "");
Expect(0, 65935, '\p{^Block=ancientgreeknumbers}', "");
Expect(0, 65935, '\P{Block=ancientgreeknumbers}', "");
Expect(1, 65935, '\P{^Block=ancientgreeknumbers}', "");
Expect(0, 65936, '\p{Block=ancientgreeknumbers}', "");
Expect(1, 65936, '\p{^Block=ancientgreeknumbers}', "");
Expect(1, 65936, '\P{Block=ancientgreeknumbers}', "");
Expect(0, 65936, '\P{^Block=ancientgreeknumbers}', "");
Expect(1, 65935, '\p{Block=		ANCIENT_GREEK_Numbers}', "");
Expect(0, 65935, '\p{^Block=		ANCIENT_GREEK_Numbers}', "");
Expect(0, 65935, '\P{Block=		ANCIENT_GREEK_Numbers}', "");
Expect(1, 65935, '\P{^Block=		ANCIENT_GREEK_Numbers}', "");
Expect(0, 65936, '\p{Block=		ANCIENT_GREEK_Numbers}', "");
Expect(1, 65936, '\p{^Block=		ANCIENT_GREEK_Numbers}', "");
Expect(1, 65936, '\P{Block=		ANCIENT_GREEK_Numbers}', "");
Expect(0, 65936, '\P{^Block=		ANCIENT_GREEK_Numbers}', "");
Error('\p{Blk=/a/_ ancient_Greek_Numbers}');
Error('\P{Blk=/a/_ ancient_Greek_Numbers}');
Expect(1, 65935, '\p{Blk=ancientgreeknumbers}', "");
Expect(0, 65935, '\p{^Blk=ancientgreeknumbers}', "");
Expect(0, 65935, '\P{Blk=ancientgreeknumbers}', "");
Expect(1, 65935, '\P{^Blk=ancientgreeknumbers}', "");
Expect(0, 65936, '\p{Blk=ancientgreeknumbers}', "");
Expect(1, 65936, '\p{^Blk=ancientgreeknumbers}', "");
Expect(1, 65936, '\P{Blk=ancientgreeknumbers}', "");
Expect(0, 65936, '\P{^Blk=ancientgreeknumbers}', "");
Expect(1, 65935, '\p{Blk=-ANCIENT_greek_Numbers}', "");
Expect(0, 65935, '\p{^Blk=-ANCIENT_greek_Numbers}', "");
Expect(0, 65935, '\P{Blk=-ANCIENT_greek_Numbers}', "");
Expect(1, 65935, '\P{^Blk=-ANCIENT_greek_Numbers}', "");
Expect(0, 65936, '\p{Blk=-ANCIENT_greek_Numbers}', "");
Expect(1, 65936, '\p{^Blk=-ANCIENT_greek_Numbers}', "");
Expect(1, 65936, '\P{Blk=-ANCIENT_greek_Numbers}', "");
Expect(0, 65936, '\P{^Blk=-ANCIENT_greek_Numbers}', "");
Error('\p{Is_Block:-:=ancient_GREEK_Numbers}');
Error('\P{Is_Block:-:=ancient_GREEK_Numbers}');
Expect(1, 65935, '\p{Is_Block=ancientgreeknumbers}', "");
Expect(0, 65935, '\p{^Is_Block=ancientgreeknumbers}', "");
Expect(0, 65935, '\P{Is_Block=ancientgreeknumbers}', "");
Expect(1, 65935, '\P{^Is_Block=ancientgreeknumbers}', "");
Expect(0, 65936, '\p{Is_Block=ancientgreeknumbers}', "");
Expect(1, 65936, '\p{^Is_Block=ancientgreeknumbers}', "");
Expect(1, 65936, '\P{Is_Block=ancientgreeknumbers}', "");
Expect(0, 65936, '\P{^Is_Block=ancientgreeknumbers}', "");
Expect(1, 65935, '\p{Is_Block= ANCIENT_greek_NUMBERS}', "");
Expect(0, 65935, '\p{^Is_Block= ANCIENT_greek_NUMBERS}', "");
Expect(0, 65935, '\P{Is_Block= ANCIENT_greek_NUMBERS}', "");
Expect(1, 65935, '\P{^Is_Block= ANCIENT_greek_NUMBERS}', "");
Expect(0, 65936, '\p{Is_Block= ANCIENT_greek_NUMBERS}', "");
Expect(1, 65936, '\p{^Is_Block= ANCIENT_greek_NUMBERS}', "");
Expect(1, 65936, '\P{Is_Block= ANCIENT_greek_NUMBERS}', "");
Expect(0, 65936, '\P{^Is_Block= ANCIENT_greek_NUMBERS}', "");
Error('\p{Is_Blk=-/a/ANCIENT_Greek_numbers}');
Error('\P{Is_Blk=-/a/ANCIENT_Greek_numbers}');
Expect(1, 65935, '\p{Is_Blk=ancientgreeknumbers}', "");
Expect(0, 65935, '\p{^Is_Blk=ancientgreeknumbers}', "");
Expect(0, 65935, '\P{Is_Blk=ancientgreeknumbers}', "");
Expect(1, 65935, '\P{^Is_Blk=ancientgreeknumbers}', "");
Expect(0, 65936, '\p{Is_Blk=ancientgreeknumbers}', "");
Expect(1, 65936, '\p{^Is_Blk=ancientgreeknumbers}', "");
Expect(1, 65936, '\P{Is_Blk=ancientgreeknumbers}', "");
Expect(0, 65936, '\P{^Is_Blk=ancientgreeknumbers}', "");
Expect(1, 65935, '\p{Is_Blk= Ancient_GREEK_NUMBERS}', "");
Expect(0, 65935, '\p{^Is_Blk= Ancient_GREEK_NUMBERS}', "");
Expect(0, 65935, '\P{Is_Blk= Ancient_GREEK_NUMBERS}', "");
Expect(1, 65935, '\P{^Is_Blk= Ancient_GREEK_NUMBERS}', "");
Expect(0, 65936, '\p{Is_Blk= Ancient_GREEK_NUMBERS}', "");
Expect(1, 65936, '\p{^Is_Blk= Ancient_GREEK_NUMBERS}', "");
Expect(1, 65936, '\P{Is_Blk= Ancient_GREEK_NUMBERS}', "");
Expect(0, 65936, '\P{^Is_Blk= Ancient_GREEK_NUMBERS}', "");
Error('\p{Block=:=_	ANCIENT_SYMBOLS}');
Error('\P{Block=:=_	ANCIENT_SYMBOLS}');
Expect(1, 65999, '\p{Block=ancientsymbols}', "");
Expect(0, 65999, '\p{^Block=ancientsymbols}', "");
Expect(0, 65999, '\P{Block=ancientsymbols}', "");
Expect(1, 65999, '\P{^Block=ancientsymbols}', "");
Expect(0, 66000, '\p{Block=ancientsymbols}', "");
Expect(1, 66000, '\p{^Block=ancientsymbols}', "");
Expect(1, 66000, '\P{Block=ancientsymbols}', "");
Expect(0, 66000, '\P{^Block=ancientsymbols}', "");
Expect(1, 65999, '\p{Block= -Ancient_Symbols}', "");
Expect(0, 65999, '\p{^Block= -Ancient_Symbols}', "");
Expect(0, 65999, '\P{Block= -Ancient_Symbols}', "");
Expect(1, 65999, '\P{^Block= -Ancient_Symbols}', "");
Expect(0, 66000, '\p{Block= -Ancient_Symbols}', "");
Expect(1, 66000, '\p{^Block= -Ancient_Symbols}', "");
Expect(1, 66000, '\P{Block= -Ancient_Symbols}', "");
Expect(0, 66000, '\P{^Block= -Ancient_Symbols}', "");
Error('\p{Blk:	_Ancient_symbols:=}');
Error('\P{Blk:	_Ancient_symbols:=}');
Expect(1, 65999, '\p{Blk=ancientsymbols}', "");
Expect(0, 65999, '\p{^Blk=ancientsymbols}', "");
Expect(0, 65999, '\P{Blk=ancientsymbols}', "");
Expect(1, 65999, '\P{^Blk=ancientsymbols}', "");
Expect(0, 66000, '\p{Blk=ancientsymbols}', "");
Expect(1, 66000, '\p{^Blk=ancientsymbols}', "");
Expect(1, 66000, '\P{Blk=ancientsymbols}', "");
Expect(0, 66000, '\P{^Blk=ancientsymbols}', "");
Expect(1, 65999, '\p{Blk= -ancient_symbols}', "");
Expect(0, 65999, '\p{^Blk= -ancient_symbols}', "");
Expect(0, 65999, '\P{Blk= -ancient_symbols}', "");
Expect(1, 65999, '\P{^Blk= -ancient_symbols}', "");
Expect(0, 66000, '\p{Blk= -ancient_symbols}', "");
Expect(1, 66000, '\p{^Blk= -ancient_symbols}', "");
Expect(1, 66000, '\P{Blk= -ancient_symbols}', "");
Expect(0, 66000, '\P{^Blk= -ancient_symbols}', "");
Error('\p{Is_Block= Ancient_Symbols:=}');
Error('\P{Is_Block= Ancient_Symbols:=}');
Expect(1, 65999, '\p{Is_Block=ancientsymbols}', "");
Expect(0, 65999, '\p{^Is_Block=ancientsymbols}', "");
Expect(0, 65999, '\P{Is_Block=ancientsymbols}', "");
Expect(1, 65999, '\P{^Is_Block=ancientsymbols}', "");
Expect(0, 66000, '\p{Is_Block=ancientsymbols}', "");
Expect(1, 66000, '\p{^Is_Block=ancientsymbols}', "");
Expect(1, 66000, '\P{Is_Block=ancientsymbols}', "");
Expect(0, 66000, '\P{^Is_Block=ancientsymbols}', "");
Expect(1, 65999, '\p{Is_Block= _Ancient_Symbols}', "");
Expect(0, 65999, '\p{^Is_Block= _Ancient_Symbols}', "");
Expect(0, 65999, '\P{Is_Block= _Ancient_Symbols}', "");
Expect(1, 65999, '\P{^Is_Block= _Ancient_Symbols}', "");
Expect(0, 66000, '\p{Is_Block= _Ancient_Symbols}', "");
Expect(1, 66000, '\p{^Is_Block= _Ancient_Symbols}', "");
Expect(1, 66000, '\P{Is_Block= _Ancient_Symbols}', "");
Expect(0, 66000, '\P{^Is_Block= _Ancient_Symbols}', "");
Error('\p{Is_Blk=	/a/ancient_Symbols}');
Error('\P{Is_Blk=	/a/ancient_Symbols}');
Expect(1, 65999, '\p{Is_Blk=ancientsymbols}', "");
Expect(0, 65999, '\p{^Is_Blk=ancientsymbols}', "");
Expect(0, 65999, '\P{Is_Blk=ancientsymbols}', "");
Expect(1, 65999, '\P{^Is_Blk=ancientsymbols}', "");
Expect(0, 66000, '\p{Is_Blk=ancientsymbols}', "");
Expect(1, 66000, '\p{^Is_Blk=ancientsymbols}', "");
Expect(1, 66000, '\P{Is_Blk=ancientsymbols}', "");
Expect(0, 66000, '\P{^Is_Blk=ancientsymbols}', "");
Expect(1, 65999, '\p{Is_Blk=	-ancient_SYMBOLS}', "");
Expect(0, 65999, '\p{^Is_Blk=	-ancient_SYMBOLS}', "");
Expect(0, 65999, '\P{Is_Blk=	-ancient_SYMBOLS}', "");
Expect(1, 65999, '\P{^Is_Blk=	-ancient_SYMBOLS}', "");
Expect(0, 66000, '\p{Is_Blk=	-ancient_SYMBOLS}', "");
Expect(1, 66000, '\p{^Is_Blk=	-ancient_SYMBOLS}', "");
Expect(1, 66000, '\P{Is_Blk=	-ancient_SYMBOLS}', "");
Expect(0, 66000, '\P{^Is_Blk=	-ancient_SYMBOLS}', "");
Error('\p{Block= -Arabic:=}');
Error('\P{Block= -Arabic:=}');
Expect(1, 1791, '\p{Block=arabic}', "");
Expect(0, 1791, '\p{^Block=arabic}', "");
Expect(0, 1791, '\P{Block=arabic}', "");
Expect(1, 1791, '\P{^Block=arabic}', "");
Expect(0, 1792, '\p{Block=arabic}', "");
Expect(1, 1792, '\p{^Block=arabic}', "");
Expect(1, 1792, '\P{Block=arabic}', "");
Expect(0, 1792, '\P{^Block=arabic}', "");
Expect(1, 1791, '\p{Block=	ARABIC}', "");
Expect(0, 1791, '\p{^Block=	ARABIC}', "");
Expect(0, 1791, '\P{Block=	ARABIC}', "");
Expect(1, 1791, '\P{^Block=	ARABIC}', "");
Expect(0, 1792, '\p{Block=	ARABIC}', "");
Expect(1, 1792, '\p{^Block=	ARABIC}', "");
Expect(1, 1792, '\P{Block=	ARABIC}', "");
Expect(0, 1792, '\P{^Block=	ARABIC}', "");
Error('\p{Blk= ARABIC/a/}');
Error('\P{Blk= ARABIC/a/}');
Expect(1, 1791, '\p{Blk=arabic}', "");
Expect(0, 1791, '\p{^Blk=arabic}', "");
Expect(0, 1791, '\P{Blk=arabic}', "");
Expect(1, 1791, '\P{^Blk=arabic}', "");
Expect(0, 1792, '\p{Blk=arabic}', "");
Expect(1, 1792, '\p{^Blk=arabic}', "");
Expect(1, 1792, '\P{Blk=arabic}', "");
Expect(0, 1792, '\P{^Blk=arabic}', "");
Expect(1, 1791, '\p{Blk=	arabic}', "");
Expect(0, 1791, '\p{^Blk=	arabic}', "");
Expect(0, 1791, '\P{Blk=	arabic}', "");
Expect(1, 1791, '\P{^Blk=	arabic}', "");
Expect(0, 1792, '\p{Blk=	arabic}', "");
Expect(1, 1792, '\p{^Blk=	arabic}', "");
Expect(1, 1792, '\P{Blk=	arabic}', "");
Expect(0, 1792, '\P{^Blk=	arabic}', "");
Error('\p{Is_Block: /a/--arabic}');
Error('\P{Is_Block: /a/--arabic}');
Expect(1, 1791, '\p{Is_Block=arabic}', "");
Expect(0, 1791, '\p{^Is_Block=arabic}', "");
Expect(0, 1791, '\P{Is_Block=arabic}', "");
Expect(1, 1791, '\P{^Is_Block=arabic}', "");
Expect(0, 1792, '\p{Is_Block=arabic}', "");
Expect(1, 1792, '\p{^Is_Block=arabic}', "");
Expect(1, 1792, '\P{Is_Block=arabic}', "");
Expect(0, 1792, '\P{^Is_Block=arabic}', "");
Expect(1, 1791, '\p{Is_Block=_	Arabic}', "");
Expect(0, 1791, '\p{^Is_Block=_	Arabic}', "");
Expect(0, 1791, '\P{Is_Block=_	Arabic}', "");
Expect(1, 1791, '\P{^Is_Block=_	Arabic}', "");
Expect(0, 1792, '\p{Is_Block=_	Arabic}', "");
Expect(1, 1792, '\p{^Is_Block=_	Arabic}', "");
Expect(1, 1792, '\P{Is_Block=_	Arabic}', "");
Expect(0, 1792, '\P{^Is_Block=_	Arabic}', "");
Error('\p{Is_Blk=-/a/ARABIC}');
Error('\P{Is_Blk=-/a/ARABIC}');
Expect(1, 1791, '\p{Is_Blk=arabic}', "");
Expect(0, 1791, '\p{^Is_Blk=arabic}', "");
Expect(0, 1791, '\P{Is_Blk=arabic}', "");
Expect(1, 1791, '\P{^Is_Blk=arabic}', "");
Expect(0, 1792, '\p{Is_Blk=arabic}', "");
Expect(1, 1792, '\p{^Is_Blk=arabic}', "");
Expect(1, 1792, '\P{Is_Blk=arabic}', "");
Expect(0, 1792, '\P{^Is_Blk=arabic}', "");
Error('\p{Block=:=_-ARABIC_Extended_A}');
Error('\P{Block=:=_-ARABIC_Extended_A}');
Expect(1, 2303, '\p{Block:   arabicextendeda}', "");
Expect(0, 2303, '\p{^Block:   arabicextendeda}', "");
Expect(0, 2303, '\P{Block:   arabicextendeda}', "");
Expect(1, 2303, '\P{^Block:   arabicextendeda}', "");
Expect(0, 2304, '\p{Block:   arabicextendeda}', "");
Expect(1, 2304, '\p{^Block:   arabicextendeda}', "");
Expect(1, 2304, '\P{Block:   arabicextendeda}', "");
Expect(0, 2304, '\P{^Block:   arabicextendeda}', "");
Expect(1, 2303, '\p{Block= -ARABIC_EXTENDED_A}', "");
Expect(0, 2303, '\p{^Block= -ARABIC_EXTENDED_A}', "");
Expect(0, 2303, '\P{Block= -ARABIC_EXTENDED_A}', "");
Expect(1, 2303, '\P{^Block= -ARABIC_EXTENDED_A}', "");
Expect(0, 2304, '\p{Block= -ARABIC_EXTENDED_A}', "");
Expect(1, 2304, '\p{^Block= -ARABIC_EXTENDED_A}', "");
Expect(1, 2304, '\P{Block= -ARABIC_EXTENDED_A}', "");
Expect(0, 2304, '\P{^Block= -ARABIC_EXTENDED_A}', "");
Error('\p{Blk=_/a/Arabic_Ext_A}');
Error('\P{Blk=_/a/Arabic_Ext_A}');
Expect(1, 2303, '\p{Blk=arabicexta}', "");
Expect(0, 2303, '\p{^Blk=arabicexta}', "");
Expect(0, 2303, '\P{Blk=arabicexta}', "");
Expect(1, 2303, '\P{^Blk=arabicexta}', "");
Expect(0, 2304, '\p{Blk=arabicexta}', "");
Expect(1, 2304, '\p{^Blk=arabicexta}', "");
Expect(1, 2304, '\P{Blk=arabicexta}', "");
Expect(0, 2304, '\P{^Blk=arabicexta}', "");
Expect(1, 2303, '\p{Blk=  Arabic_Ext_a}', "");
Expect(0, 2303, '\p{^Blk=  Arabic_Ext_a}', "");
Expect(0, 2303, '\P{Blk=  Arabic_Ext_a}', "");
Expect(1, 2303, '\P{^Blk=  Arabic_Ext_a}', "");
Expect(0, 2304, '\p{Blk=  Arabic_Ext_a}', "");
Expect(1, 2304, '\p{^Blk=  Arabic_Ext_a}', "");
Expect(1, 2304, '\P{Blk=  Arabic_Ext_a}', "");
Expect(0, 2304, '\P{^Blk=  Arabic_Ext_a}', "");
Error('\p{Is_Block=_:=Arabic_EXTENDED_A}');
Error('\P{Is_Block=_:=Arabic_EXTENDED_A}');
Expect(1, 2303, '\p{Is_Block=arabicextendeda}', "");
Expect(0, 2303, '\p{^Is_Block=arabicextendeda}', "");
Expect(0, 2303, '\P{Is_Block=arabicextendeda}', "");
Expect(1, 2303, '\P{^Is_Block=arabicextendeda}', "");
Expect(0, 2304, '\p{Is_Block=arabicextendeda}', "");
Expect(1, 2304, '\p{^Is_Block=arabicextendeda}', "");
Expect(1, 2304, '\P{Is_Block=arabicextendeda}', "");
Expect(0, 2304, '\P{^Is_Block=arabicextendeda}', "");
Expect(1, 2303, '\p{Is_Block:	_ARABIC_EXTENDED_A}', "");
Expect(0, 2303, '\p{^Is_Block:	_ARABIC_EXTENDED_A}', "");
Expect(0, 2303, '\P{Is_Block:	_ARABIC_EXTENDED_A}', "");
Expect(1, 2303, '\P{^Is_Block:	_ARABIC_EXTENDED_A}', "");
Expect(0, 2304, '\p{Is_Block:	_ARABIC_EXTENDED_A}', "");
Expect(1, 2304, '\p{^Is_Block:	_ARABIC_EXTENDED_A}', "");
Expect(1, 2304, '\P{Is_Block:	_ARABIC_EXTENDED_A}', "");
Expect(0, 2304, '\P{^Is_Block:	_ARABIC_EXTENDED_A}', "");
Error('\p{Is_Blk=  Arabic_EXT_A/a/}');
Error('\P{Is_Blk=  Arabic_EXT_A/a/}');
Expect(1, 2303, '\p{Is_Blk=arabicexta}', "");
Expect(0, 2303, '\p{^Is_Blk=arabicexta}', "");
Expect(0, 2303, '\P{Is_Blk=arabicexta}', "");
Expect(1, 2303, '\P{^Is_Blk=arabicexta}', "");
Expect(0, 2304, '\p{Is_Blk=arabicexta}', "");
Expect(1, 2304, '\p{^Is_Blk=arabicexta}', "");
Expect(1, 2304, '\P{Is_Blk=arabicexta}', "");
Expect(0, 2304, '\P{^Is_Blk=arabicexta}', "");
Expect(1, 2303, '\p{Is_Blk= 	arabic_Ext_A}', "");
Expect(0, 2303, '\p{^Is_Blk= 	arabic_Ext_A}', "");
Expect(0, 2303, '\P{Is_Blk= 	arabic_Ext_A}', "");
Expect(1, 2303, '\P{^Is_Blk= 	arabic_Ext_A}', "");
Expect(0, 2304, '\p{Is_Blk= 	arabic_Ext_A}', "");
Expect(1, 2304, '\p{^Is_Blk= 	arabic_Ext_A}', "");
Expect(1, 2304, '\P{Is_Blk= 	arabic_Ext_A}', "");
Expect(0, 2304, '\P{^Is_Blk= 	arabic_Ext_A}', "");
Error('\p{Block=_:=arabic_MATHEMATICAL_Alphabetic_Symbols}');
Error('\P{Block=_:=arabic_MATHEMATICAL_Alphabetic_Symbols}');
Expect(1, 126719, '\p{Block=arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\p{^Block=arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\P{Block=arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\P{^Block=arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\p{Block=arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\p{^Block=arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\P{Block=arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\P{^Block=arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\p{Block= 	Arabic_MATHEMATICAL_alphabetic_symbols}', "");
Expect(0, 126719, '\p{^Block= 	Arabic_MATHEMATICAL_alphabetic_symbols}', "");
Expect(0, 126719, '\P{Block= 	Arabic_MATHEMATICAL_alphabetic_symbols}', "");
Expect(1, 126719, '\P{^Block= 	Arabic_MATHEMATICAL_alphabetic_symbols}', "");
Expect(0, 126720, '\p{Block= 	Arabic_MATHEMATICAL_alphabetic_symbols}', "");
Expect(1, 126720, '\p{^Block= 	Arabic_MATHEMATICAL_alphabetic_symbols}', "");
Expect(1, 126720, '\P{Block= 	Arabic_MATHEMATICAL_alphabetic_symbols}', "");
Expect(0, 126720, '\P{^Block= 	Arabic_MATHEMATICAL_alphabetic_symbols}', "");
Error('\p{Blk= 	ARABIC_Math/a/}');
Error('\P{Blk= 	ARABIC_Math/a/}');
Expect(1, 126719, '\p{Blk=arabicmath}', "");
Expect(0, 126719, '\p{^Blk=arabicmath}', "");
Expect(0, 126719, '\P{Blk=arabicmath}', "");
Expect(1, 126719, '\P{^Blk=arabicmath}', "");
Expect(0, 126720, '\p{Blk=arabicmath}', "");
Expect(1, 126720, '\p{^Blk=arabicmath}', "");
Expect(1, 126720, '\P{Blk=arabicmath}', "");
Expect(0, 126720, '\P{^Blk=arabicmath}', "");
Expect(1, 126719, '\p{Blk=		arabic_Math}', "");
Expect(0, 126719, '\p{^Blk=		arabic_Math}', "");
Expect(0, 126719, '\P{Blk=		arabic_Math}', "");
Expect(1, 126719, '\P{^Blk=		arabic_Math}', "");
Expect(0, 126720, '\p{Blk=		arabic_Math}', "");
Expect(1, 126720, '\p{^Blk=		arabic_Math}', "");
Expect(1, 126720, '\P{Blk=		arabic_Math}', "");
Expect(0, 126720, '\P{^Blk=		arabic_Math}', "");
Error('\p{Is_Block=	Arabic_Mathematical_Alphabetic_Symbols:=}');
Error('\P{Is_Block=	Arabic_Mathematical_Alphabetic_Symbols:=}');
Expect(1, 126719, '\p{Is_Block=arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\p{^Is_Block=arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\P{Is_Block=arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\P{^Is_Block=arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\p{Is_Block=arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\p{^Is_Block=arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\P{Is_Block=arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\P{^Is_Block=arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\p{Is_Block=_ Arabic_mathematical_ALPHABETIC_symbols}', "");
Expect(0, 126719, '\p{^Is_Block=_ Arabic_mathematical_ALPHABETIC_symbols}', "");
Expect(0, 126719, '\P{Is_Block=_ Arabic_mathematical_ALPHABETIC_symbols}', "");
Expect(1, 126719, '\P{^Is_Block=_ Arabic_mathematical_ALPHABETIC_symbols}', "");
Expect(0, 126720, '\p{Is_Block=_ Arabic_mathematical_ALPHABETIC_symbols}', "");
Expect(1, 126720, '\p{^Is_Block=_ Arabic_mathematical_ALPHABETIC_symbols}', "");
Expect(1, 126720, '\P{Is_Block=_ Arabic_mathematical_ALPHABETIC_symbols}', "");
Expect(0, 126720, '\P{^Is_Block=_ Arabic_mathematical_ALPHABETIC_symbols}', "");
Error('\p{Is_Blk:	/a/-ARABIC_MATH}');
Error('\P{Is_Blk:	/a/-ARABIC_MATH}');
Expect(1, 126719, '\p{Is_Blk=arabicmath}', "");
Expect(0, 126719, '\p{^Is_Blk=arabicmath}', "");
Expect(0, 126719, '\P{Is_Blk=arabicmath}', "");
Expect(1, 126719, '\P{^Is_Blk=arabicmath}', "");
Expect(0, 126720, '\p{Is_Blk=arabicmath}', "");
Expect(1, 126720, '\p{^Is_Blk=arabicmath}', "");
Expect(1, 126720, '\P{Is_Blk=arabicmath}', "");
Expect(0, 126720, '\P{^Is_Blk=arabicmath}', "");
Expect(1, 126719, '\p{Is_Blk=--Arabic_Math}', "");
Expect(0, 126719, '\p{^Is_Blk=--Arabic_Math}', "");
Expect(0, 126719, '\P{Is_Blk=--Arabic_Math}', "");
Expect(1, 126719, '\P{^Is_Blk=--Arabic_Math}', "");
Expect(0, 126720, '\p{Is_Blk=--Arabic_Math}', "");
Expect(1, 126720, '\p{^Is_Blk=--Arabic_Math}', "");
Expect(1, 126720, '\P{Is_Blk=--Arabic_Math}', "");
Expect(0, 126720, '\P{^Is_Blk=--Arabic_Math}', "");
Error('\p{Block=		ARABIC_Presentation_FORMS_A:=}');
Error('\P{Block=		ARABIC_Presentation_FORMS_A:=}');
Expect(1, 65023, '\p{Block=arabicpresentationformsa}', "");
Expect(0, 65023, '\p{^Block=arabicpresentationformsa}', "");
Expect(0, 65023, '\P{Block=arabicpresentationformsa}', "");
Expect(1, 65023, '\P{^Block=arabicpresentationformsa}', "");
Expect(0, 65024, '\p{Block=arabicpresentationformsa}', "");
Expect(1, 65024, '\p{^Block=arabicpresentationformsa}', "");
Expect(1, 65024, '\P{Block=arabicpresentationformsa}', "");
Expect(0, 65024, '\P{^Block=arabicpresentationformsa}', "");
Expect(1, 65023, '\p{Block=-_arabic_presentation_Forms_A}', "");
Expect(0, 65023, '\p{^Block=-_arabic_presentation_Forms_A}', "");
Expect(0, 65023, '\P{Block=-_arabic_presentation_Forms_A}', "");
Expect(1, 65023, '\P{^Block=-_arabic_presentation_Forms_A}', "");
Expect(0, 65024, '\p{Block=-_arabic_presentation_Forms_A}', "");
Expect(1, 65024, '\p{^Block=-_arabic_presentation_Forms_A}', "");
Expect(1, 65024, '\P{Block=-_arabic_presentation_Forms_A}', "");
Expect(0, 65024, '\P{^Block=-_arabic_presentation_Forms_A}', "");
Error('\p{Blk=		ARABIC_pf_A/a/}');
Error('\P{Blk=		ARABIC_pf_A/a/}');
Expect(1, 65023, '\p{Blk=arabicpfa}', "");
Expect(0, 65023, '\p{^Blk=arabicpfa}', "");
Expect(0, 65023, '\P{Blk=arabicpfa}', "");
Expect(1, 65023, '\P{^Blk=arabicpfa}', "");
Expect(0, 65024, '\p{Blk=arabicpfa}', "");
Expect(1, 65024, '\p{^Blk=arabicpfa}', "");
Expect(1, 65024, '\P{Blk=arabicpfa}', "");
Expect(0, 65024, '\P{^Blk=arabicpfa}', "");
Expect(1, 65023, '\p{Blk=_arabic_PF_A}', "");
Expect(0, 65023, '\p{^Blk=_arabic_PF_A}', "");
Expect(0, 65023, '\P{Blk=_arabic_PF_A}', "");
Expect(1, 65023, '\P{^Blk=_arabic_PF_A}', "");
Expect(0, 65024, '\p{Blk=_arabic_PF_A}', "");
Expect(1, 65024, '\p{^Blk=_arabic_PF_A}', "");
Expect(1, 65024, '\P{Blk=_arabic_PF_A}', "");
Expect(0, 65024, '\P{^Blk=_arabic_PF_A}', "");
Error('\p{Is_Block=- arabic_presentation_Forms_A:=}');
Error('\P{Is_Block=- arabic_presentation_Forms_A:=}');
Expect(1, 65023, '\p{Is_Block=arabicpresentationformsa}', "");
Expect(0, 65023, '\p{^Is_Block=arabicpresentationformsa}', "");
Expect(0, 65023, '\P{Is_Block=arabicpresentationformsa}', "");
Expect(1, 65023, '\P{^Is_Block=arabicpresentationformsa}', "");
Expect(0, 65024, '\p{Is_Block=arabicpresentationformsa}', "");
Expect(1, 65024, '\p{^Is_Block=arabicpresentationformsa}', "");
Expect(1, 65024, '\P{Is_Block=arabicpresentationformsa}', "");
Expect(0, 65024, '\P{^Is_Block=arabicpresentationformsa}', "");
Expect(1, 65023, '\p{Is_Block=- Arabic_PRESENTATION_forms_A}', "");
Expect(0, 65023, '\p{^Is_Block=- Arabic_PRESENTATION_forms_A}', "");
Expect(0, 65023, '\P{Is_Block=- Arabic_PRESENTATION_forms_A}', "");
Expect(1, 65023, '\P{^Is_Block=- Arabic_PRESENTATION_forms_A}', "");
Expect(0, 65024, '\p{Is_Block=- Arabic_PRESENTATION_forms_A}', "");
Expect(1, 65024, '\p{^Is_Block=- Arabic_PRESENTATION_forms_A}', "");
Expect(1, 65024, '\P{Is_Block=- Arabic_PRESENTATION_forms_A}', "");
Expect(0, 65024, '\P{^Is_Block=- Arabic_PRESENTATION_forms_A}', "");
Error('\p{Is_Blk= _arabic_pf_A/a/}');
Error('\P{Is_Blk= _arabic_pf_A/a/}');
Expect(1, 65023, '\p{Is_Blk=arabicpfa}', "");
Expect(0, 65023, '\p{^Is_Blk=arabicpfa}', "");
Expect(0, 65023, '\P{Is_Blk=arabicpfa}', "");
Expect(1, 65023, '\P{^Is_Blk=arabicpfa}', "");
Expect(0, 65024, '\p{Is_Blk=arabicpfa}', "");
Expect(1, 65024, '\p{^Is_Blk=arabicpfa}', "");
Expect(1, 65024, '\P{Is_Blk=arabicpfa}', "");
Expect(0, 65024, '\P{^Is_Blk=arabicpfa}', "");
Expect(1, 65023, '\p{Is_Blk=-_ARABIC_PF_A}', "");
Expect(0, 65023, '\p{^Is_Blk=-_ARABIC_PF_A}', "");
Expect(0, 65023, '\P{Is_Blk=-_ARABIC_PF_A}', "");
Expect(1, 65023, '\P{^Is_Blk=-_ARABIC_PF_A}', "");
Expect(0, 65024, '\p{Is_Blk=-_ARABIC_PF_A}', "");
Expect(1, 65024, '\p{^Is_Blk=-_ARABIC_PF_A}', "");
Expect(1, 65024, '\P{Is_Blk=-_ARABIC_PF_A}', "");
Expect(0, 65024, '\P{^Is_Blk=-_ARABIC_PF_A}', "");
Error('\p{Block=-/a/ARABIC_presentation_FORMS_B}');
Error('\P{Block=-/a/ARABIC_presentation_FORMS_B}');
Expect(1, 65279, '\p{Block=arabicpresentationformsb}', "");
Expect(0, 65279, '\p{^Block=arabicpresentationformsb}', "");
Expect(0, 65279, '\P{Block=arabicpresentationformsb}', "");
Expect(1, 65279, '\P{^Block=arabicpresentationformsb}', "");
Expect(0, 65280, '\p{Block=arabicpresentationformsb}', "");
Expect(1, 65280, '\p{^Block=arabicpresentationformsb}', "");
Expect(1, 65280, '\P{Block=arabicpresentationformsb}', "");
Expect(0, 65280, '\P{^Block=arabicpresentationformsb}', "");
Expect(1, 65279, '\p{Block=		ARABIC_Presentation_Forms_B}', "");
Expect(0, 65279, '\p{^Block=		ARABIC_Presentation_Forms_B}', "");
Expect(0, 65279, '\P{Block=		ARABIC_Presentation_Forms_B}', "");
Expect(1, 65279, '\P{^Block=		ARABIC_Presentation_Forms_B}', "");
Expect(0, 65280, '\p{Block=		ARABIC_Presentation_Forms_B}', "");
Expect(1, 65280, '\p{^Block=		ARABIC_Presentation_Forms_B}', "");
Expect(1, 65280, '\P{Block=		ARABIC_Presentation_Forms_B}', "");
Expect(0, 65280, '\P{^Block=		ARABIC_Presentation_Forms_B}', "");
Error('\p{Blk= :=ARABIC_PF_B}');
Error('\P{Blk= :=ARABIC_PF_B}');
Expect(1, 65279, '\p{Blk=arabicpfb}', "");
Expect(0, 65279, '\p{^Blk=arabicpfb}', "");
Expect(0, 65279, '\P{Blk=arabicpfb}', "");
Expect(1, 65279, '\P{^Blk=arabicpfb}', "");
Expect(0, 65280, '\p{Blk=arabicpfb}', "");
Expect(1, 65280, '\p{^Blk=arabicpfb}', "");
Expect(1, 65280, '\P{Blk=arabicpfb}', "");
Expect(0, 65280, '\P{^Blk=arabicpfb}', "");
Expect(1, 65279, '\p{Blk=	-Arabic_PF_b}', "");
Expect(0, 65279, '\p{^Blk=	-Arabic_PF_b}', "");
Expect(0, 65279, '\P{Blk=	-Arabic_PF_b}', "");
Expect(1, 65279, '\P{^Blk=	-Arabic_PF_b}', "");
Expect(0, 65280, '\p{Blk=	-Arabic_PF_b}', "");
Expect(1, 65280, '\p{^Blk=	-Arabic_PF_b}', "");
Expect(1, 65280, '\P{Blk=	-Arabic_PF_b}', "");
Expect(0, 65280, '\P{^Blk=	-Arabic_PF_b}', "");
Error('\p{Is_Block=-	arabic_presentation_forms_B/a/}');
Error('\P{Is_Block=-	arabic_presentation_forms_B/a/}');
Expect(1, 65279, '\p{Is_Block=arabicpresentationformsb}', "");
Expect(0, 65279, '\p{^Is_Block=arabicpresentationformsb}', "");
Expect(0, 65279, '\P{Is_Block=arabicpresentationformsb}', "");
Expect(1, 65279, '\P{^Is_Block=arabicpresentationformsb}', "");
Expect(0, 65280, '\p{Is_Block=arabicpresentationformsb}', "");
Expect(1, 65280, '\p{^Is_Block=arabicpresentationformsb}', "");
Expect(1, 65280, '\P{Is_Block=arabicpresentationformsb}', "");
Expect(0, 65280, '\P{^Is_Block=arabicpresentationformsb}', "");
Expect(1, 65279, '\p{Is_Block=	Arabic_Presentation_Forms_B}', "");
Expect(0, 65279, '\p{^Is_Block=	Arabic_Presentation_Forms_B}', "");
Expect(0, 65279, '\P{Is_Block=	Arabic_Presentation_Forms_B}', "");
Expect(1, 65279, '\P{^Is_Block=	Arabic_Presentation_Forms_B}', "");
Expect(0, 65280, '\p{Is_Block=	Arabic_Presentation_Forms_B}', "");
Expect(1, 65280, '\p{^Is_Block=	Arabic_Presentation_Forms_B}', "");
Expect(1, 65280, '\P{Is_Block=	Arabic_Presentation_Forms_B}', "");
Expect(0, 65280, '\P{^Is_Block=	Arabic_Presentation_Forms_B}', "");
Error('\p{Is_Blk=__arabic_PF_B/a/}');
Error('\P{Is_Blk=__arabic_PF_B/a/}');
Expect(1, 65279, '\p{Is_Blk=arabicpfb}', "");
Expect(0, 65279, '\p{^Is_Blk=arabicpfb}', "");
Expect(0, 65279, '\P{Is_Blk=arabicpfb}', "");
Expect(1, 65279, '\P{^Is_Blk=arabicpfb}', "");
Expect(0, 65280, '\p{Is_Blk=arabicpfb}', "");
Expect(1, 65280, '\p{^Is_Blk=arabicpfb}', "");
Expect(1, 65280, '\P{Is_Blk=arabicpfb}', "");
Expect(0, 65280, '\P{^Is_Blk=arabicpfb}', "");
Expect(1, 65279, '\p{Is_Blk= -Arabic_pf_b}', "");
Expect(0, 65279, '\p{^Is_Blk= -Arabic_pf_b}', "");
Expect(0, 65279, '\P{Is_Blk= -Arabic_pf_b}', "");
Expect(1, 65279, '\P{^Is_Blk= -Arabic_pf_b}', "");
Expect(0, 65280, '\p{Is_Blk= -Arabic_pf_b}', "");
Expect(1, 65280, '\p{^Is_Blk= -Arabic_pf_b}', "");
Expect(1, 65280, '\P{Is_Blk= -Arabic_pf_b}', "");
Expect(0, 65280, '\P{^Is_Blk= -Arabic_pf_b}', "");
Error('\p{Block=-:=Arabic_supplement}');
Error('\P{Block=-:=Arabic_supplement}');
Expect(1, 1919, '\p{Block=arabicsupplement}', "");
Expect(0, 1919, '\p{^Block=arabicsupplement}', "");
Expect(0, 1919, '\P{Block=arabicsupplement}', "");
Expect(1, 1919, '\P{^Block=arabicsupplement}', "");
Expect(0, 1920, '\p{Block=arabicsupplement}', "");
Expect(1, 1920, '\p{^Block=arabicsupplement}', "");
Expect(1, 1920, '\P{Block=arabicsupplement}', "");
Expect(0, 1920, '\P{^Block=arabicsupplement}', "");
Expect(1, 1919, '\p{Block= arabic_SUPPLEMENT}', "");
Expect(0, 1919, '\p{^Block= arabic_SUPPLEMENT}', "");
Expect(0, 1919, '\P{Block= arabic_SUPPLEMENT}', "");
Expect(1, 1919, '\P{^Block= arabic_SUPPLEMENT}', "");
Expect(0, 1920, '\p{Block= arabic_SUPPLEMENT}', "");
Expect(1, 1920, '\p{^Block= arabic_SUPPLEMENT}', "");
Expect(1, 1920, '\P{Block= arabic_SUPPLEMENT}', "");
Expect(0, 1920, '\P{^Block= arabic_SUPPLEMENT}', "");
Error('\p{Blk=:= Arabic_Sup}');
Error('\P{Blk=:= Arabic_Sup}');
Expect(1, 1919, '\p{Blk=arabicsup}', "");
Expect(0, 1919, '\p{^Blk=arabicsup}', "");
Expect(0, 1919, '\P{Blk=arabicsup}', "");
Expect(1, 1919, '\P{^Blk=arabicsup}', "");
Expect(0, 1920, '\p{Blk=arabicsup}', "");
Expect(1, 1920, '\p{^Blk=arabicsup}', "");
Expect(1, 1920, '\P{Blk=arabicsup}', "");
Expect(0, 1920, '\P{^Blk=arabicsup}', "");
Expect(1, 1919, '\p{Blk:	_ Arabic_SUP}', "");
Expect(0, 1919, '\p{^Blk:	_ Arabic_SUP}', "");
Expect(0, 1919, '\P{Blk:	_ Arabic_SUP}', "");
Expect(1, 1919, '\P{^Blk:	_ Arabic_SUP}', "");
Expect(0, 1920, '\p{Blk:	_ Arabic_SUP}', "");
Expect(1, 1920, '\p{^Blk:	_ Arabic_SUP}', "");
Expect(1, 1920, '\P{Blk:	_ Arabic_SUP}', "");
Expect(0, 1920, '\P{^Blk:	_ Arabic_SUP}', "");
Error('\p{Is_Block=:= 	Arabic_Supplement}');
Error('\P{Is_Block=:= 	Arabic_Supplement}');
Expect(1, 1919, '\p{Is_Block=arabicsupplement}', "");
Expect(0, 1919, '\p{^Is_Block=arabicsupplement}', "");
Expect(0, 1919, '\P{Is_Block=arabicsupplement}', "");
Expect(1, 1919, '\P{^Is_Block=arabicsupplement}', "");
Expect(0, 1920, '\p{Is_Block=arabicsupplement}', "");
Expect(1, 1920, '\p{^Is_Block=arabicsupplement}', "");
Expect(1, 1920, '\P{Is_Block=arabicsupplement}', "");
Expect(0, 1920, '\P{^Is_Block=arabicsupplement}', "");
Expect(1, 1919, '\p{Is_Block=_	Arabic_Supplement}', "");
Expect(0, 1919, '\p{^Is_Block=_	Arabic_Supplement}', "");
Expect(0, 1919, '\P{Is_Block=_	Arabic_Supplement}', "");
Expect(1, 1919, '\P{^Is_Block=_	Arabic_Supplement}', "");
Expect(0, 1920, '\p{Is_Block=_	Arabic_Supplement}', "");
Expect(1, 1920, '\p{^Is_Block=_	Arabic_Supplement}', "");
Expect(1, 1920, '\P{Is_Block=_	Arabic_Supplement}', "");
Expect(0, 1920, '\P{^Is_Block=_	Arabic_Supplement}', "");
Error('\p{Is_Blk=/a/arabic_Sup}');
Error('\P{Is_Blk=/a/arabic_Sup}');
Expect(1, 1919, '\p{Is_Blk=arabicsup}', "");
Expect(0, 1919, '\p{^Is_Blk=arabicsup}', "");
Expect(0, 1919, '\P{Is_Blk=arabicsup}', "");
Expect(1, 1919, '\P{^Is_Blk=arabicsup}', "");
Expect(0, 1920, '\p{Is_Blk=arabicsup}', "");
Expect(1, 1920, '\p{^Is_Blk=arabicsup}', "");
Expect(1, 1920, '\P{Is_Blk=arabicsup}', "");
Expect(0, 1920, '\P{^Is_Blk=arabicsup}', "");
Expect(1, 1919, '\p{Is_Blk=		Arabic_Sup}', "");
Expect(0, 1919, '\p{^Is_Blk=		Arabic_Sup}', "");
Expect(0, 1919, '\P{Is_Blk=		Arabic_Sup}', "");
Expect(1, 1919, '\P{^Is_Blk=		Arabic_Sup}', "");
Expect(0, 1920, '\p{Is_Blk=		Arabic_Sup}', "");
Expect(1, 1920, '\p{^Is_Blk=		Arabic_Sup}', "");
Expect(1, 1920, '\P{Is_Blk=		Arabic_Sup}', "");
Expect(0, 1920, '\P{^Is_Blk=		Arabic_Sup}', "");
Error('\p{Block=:=ARMENIAN}');
Error('\P{Block=:=ARMENIAN}');
Expect(1, 1423, '\p{Block=armenian}', "");
Expect(0, 1423, '\p{^Block=armenian}', "");
Expect(0, 1423, '\P{Block=armenian}', "");
Expect(1, 1423, '\P{^Block=armenian}', "");
Expect(0, 1424, '\p{Block=armenian}', "");
Expect(1, 1424, '\p{^Block=armenian}', "");
Expect(1, 1424, '\P{Block=armenian}', "");
Expect(0, 1424, '\P{^Block=armenian}', "");
Expect(1, 1423, '\p{Block=_	Armenian}', "");
Expect(0, 1423, '\p{^Block=_	Armenian}', "");
Expect(0, 1423, '\P{Block=_	Armenian}', "");
Expect(1, 1423, '\P{^Block=_	Armenian}', "");
Expect(0, 1424, '\p{Block=_	Armenian}', "");
Expect(1, 1424, '\p{^Block=_	Armenian}', "");
Expect(1, 1424, '\P{Block=_	Armenian}', "");
Expect(0, 1424, '\P{^Block=_	Armenian}', "");
Error('\p{Blk= 	armenian/a/}');
Error('\P{Blk= 	armenian/a/}');
Expect(1, 1423, '\p{Blk=armenian}', "");
Expect(0, 1423, '\p{^Blk=armenian}', "");
Expect(0, 1423, '\P{Blk=armenian}', "");
Expect(1, 1423, '\P{^Blk=armenian}', "");
Expect(0, 1424, '\p{Blk=armenian}', "");
Expect(1, 1424, '\p{^Blk=armenian}', "");
Expect(1, 1424, '\P{Blk=armenian}', "");
Expect(0, 1424, '\P{^Blk=armenian}', "");
Expect(1, 1423, '\p{Blk=	Armenian}', "");
Expect(0, 1423, '\p{^Blk=	Armenian}', "");
Expect(0, 1423, '\P{Blk=	Armenian}', "");
Expect(1, 1423, '\P{^Blk=	Armenian}', "");
Expect(0, 1424, '\p{Blk=	Armenian}', "");
Expect(1, 1424, '\p{^Blk=	Armenian}', "");
Expect(1, 1424, '\P{Blk=	Armenian}', "");
Expect(0, 1424, '\P{^Blk=	Armenian}', "");
Error('\p{Is_Block=_-armenian:=}');
Error('\P{Is_Block=_-armenian:=}');
Expect(1, 1423, '\p{Is_Block=armenian}', "");
Expect(0, 1423, '\p{^Is_Block=armenian}', "");
Expect(0, 1423, '\P{Is_Block=armenian}', "");
Expect(1, 1423, '\P{^Is_Block=armenian}', "");
Expect(0, 1424, '\p{Is_Block=armenian}', "");
Expect(1, 1424, '\p{^Is_Block=armenian}', "");
Expect(1, 1424, '\P{Is_Block=armenian}', "");
Expect(0, 1424, '\P{^Is_Block=armenian}', "");
Expect(1, 1423, '\p{Is_Block= ARMENIAN}', "");
Expect(0, 1423, '\p{^Is_Block= ARMENIAN}', "");
Expect(0, 1423, '\P{Is_Block= ARMENIAN}', "");
Expect(1, 1423, '\P{^Is_Block= ARMENIAN}', "");
Expect(0, 1424, '\p{Is_Block= ARMENIAN}', "");
Expect(1, 1424, '\p{^Is_Block= ARMENIAN}', "");
Expect(1, 1424, '\P{Is_Block= ARMENIAN}', "");
Expect(0, 1424, '\P{^Is_Block= ARMENIAN}', "");
Error('\p{Is_Blk=:=	-Armenian}');
Error('\P{Is_Blk=:=	-Armenian}');
Expect(1, 1423, '\p{Is_Blk=armenian}', "");
Expect(0, 1423, '\p{^Is_Blk=armenian}', "");
Expect(0, 1423, '\P{Is_Blk=armenian}', "");
Expect(1, 1423, '\P{^Is_Blk=armenian}', "");
Expect(0, 1424, '\p{Is_Blk=armenian}', "");
Expect(1, 1424, '\p{^Is_Blk=armenian}', "");
Expect(1, 1424, '\P{Is_Blk=armenian}', "");
Expect(0, 1424, '\P{^Is_Blk=armenian}', "");
Expect(1, 1423, '\p{Is_Blk=_Armenian}', "");
Expect(0, 1423, '\p{^Is_Blk=_Armenian}', "");
Expect(0, 1423, '\P{Is_Blk=_Armenian}', "");
Expect(1, 1423, '\P{^Is_Blk=_Armenian}', "");
Expect(0, 1424, '\p{Is_Blk=_Armenian}', "");
Expect(1, 1424, '\p{^Is_Blk=_Armenian}', "");
Expect(1, 1424, '\P{Is_Blk=_Armenian}', "");
Expect(0, 1424, '\P{^Is_Blk=_Armenian}', "");
Error('\p{Block=-_ARROWS:=}');
Error('\P{Block=-_ARROWS:=}');
Expect(1, 8703, '\p{Block=arrows}', "");
Expect(0, 8703, '\p{^Block=arrows}', "");
Expect(0, 8703, '\P{Block=arrows}', "");
Expect(1, 8703, '\P{^Block=arrows}', "");
Expect(0, 8704, '\p{Block=arrows}', "");
Expect(1, 8704, '\p{^Block=arrows}', "");
Expect(1, 8704, '\P{Block=arrows}', "");
Expect(0, 8704, '\P{^Block=arrows}', "");
Expect(1, 8703, '\p{Block=-ARROWS}', "");
Expect(0, 8703, '\p{^Block=-ARROWS}', "");
Expect(0, 8703, '\P{Block=-ARROWS}', "");
Expect(1, 8703, '\P{^Block=-ARROWS}', "");
Expect(0, 8704, '\p{Block=-ARROWS}', "");
Expect(1, 8704, '\p{^Block=-ARROWS}', "");
Expect(1, 8704, '\P{Block=-ARROWS}', "");
Expect(0, 8704, '\P{^Block=-ARROWS}', "");
Error('\p{Blk:/a/--ARROWS}');
Error('\P{Blk:/a/--ARROWS}');
Expect(1, 8703, '\p{Blk=arrows}', "");
Expect(0, 8703, '\p{^Blk=arrows}', "");
Expect(0, 8703, '\P{Blk=arrows}', "");
Expect(1, 8703, '\P{^Blk=arrows}', "");
Expect(0, 8704, '\p{Blk=arrows}', "");
Expect(1, 8704, '\p{^Blk=arrows}', "");
Expect(1, 8704, '\P{Blk=arrows}', "");
Expect(0, 8704, '\P{^Blk=arrows}', "");
Expect(1, 8703, '\p{Blk=-	ARROWS}', "");
Expect(0, 8703, '\p{^Blk=-	ARROWS}', "");
Expect(0, 8703, '\P{Blk=-	ARROWS}', "");
Expect(1, 8703, '\P{^Blk=-	ARROWS}', "");
Expect(0, 8704, '\p{Blk=-	ARROWS}', "");
Expect(1, 8704, '\p{^Blk=-	ARROWS}', "");
Expect(1, 8704, '\P{Blk=-	ARROWS}', "");
Expect(0, 8704, '\P{^Blk=-	ARROWS}', "");
Error('\p{Is_Block=	:=Arrows}');
Error('\P{Is_Block=	:=Arrows}');
Expect(1, 8703, '\p{Is_Block: arrows}', "");
Expect(0, 8703, '\p{^Is_Block: arrows}', "");
Expect(0, 8703, '\P{Is_Block: arrows}', "");
Expect(1, 8703, '\P{^Is_Block: arrows}', "");
Expect(0, 8704, '\p{Is_Block: arrows}', "");
Expect(1, 8704, '\p{^Is_Block: arrows}', "");
Expect(1, 8704, '\P{Is_Block: arrows}', "");
Expect(0, 8704, '\P{^Is_Block: arrows}', "");
Expect(1, 8703, '\p{Is_Block= -arrows}', "");
Expect(0, 8703, '\p{^Is_Block= -arrows}', "");
Expect(0, 8703, '\P{Is_Block= -arrows}', "");
Expect(1, 8703, '\P{^Is_Block= -arrows}', "");
Expect(0, 8704, '\p{Is_Block= -arrows}', "");
Expect(1, 8704, '\p{^Is_Block= -arrows}', "");
Expect(1, 8704, '\P{Is_Block= -arrows}', "");
Expect(0, 8704, '\P{^Is_Block= -arrows}', "");
Error('\p{Is_Blk=/a/-_Arrows}');
Error('\P{Is_Blk=/a/-_Arrows}');
Expect(1, 8703, '\p{Is_Blk=arrows}', "");
Expect(0, 8703, '\p{^Is_Blk=arrows}', "");
Expect(0, 8703, '\P{Is_Blk=arrows}', "");
Expect(1, 8703, '\P{^Is_Blk=arrows}', "");
Expect(0, 8704, '\p{Is_Blk=arrows}', "");
Expect(1, 8704, '\p{^Is_Blk=arrows}', "");
Expect(1, 8704, '\P{Is_Blk=arrows}', "");
Expect(0, 8704, '\P{^Is_Blk=arrows}', "");
Expect(1, 8703, '\p{Is_Blk=-_Arrows}', "");
Expect(0, 8703, '\p{^Is_Blk=-_Arrows}', "");
Expect(0, 8703, '\P{Is_Blk=-_Arrows}', "");
Expect(1, 8703, '\P{^Is_Blk=-_Arrows}', "");
Expect(0, 8704, '\p{Is_Blk=-_Arrows}', "");
Expect(1, 8704, '\p{^Is_Blk=-_Arrows}', "");
Expect(1, 8704, '\P{Is_Blk=-_Arrows}', "");
Expect(0, 8704, '\P{^Is_Blk=-_Arrows}', "");
Error('\p{Block:_Avestan:=}');
Error('\P{Block:_Avestan:=}');
Expect(1, 68415, '\p{Block=avestan}', "");
Expect(0, 68415, '\p{^Block=avestan}', "");
Expect(0, 68415, '\P{Block=avestan}', "");
Expect(1, 68415, '\P{^Block=avestan}', "");
Expect(0, 68416, '\p{Block=avestan}', "");
Expect(1, 68416, '\p{^Block=avestan}', "");
Expect(1, 68416, '\P{Block=avestan}', "");
Expect(0, 68416, '\P{^Block=avestan}', "");
Expect(1, 68415, '\p{Block=	 avestan}', "");
Expect(0, 68415, '\p{^Block=	 avestan}', "");
Expect(0, 68415, '\P{Block=	 avestan}', "");
Expect(1, 68415, '\P{^Block=	 avestan}', "");
Expect(0, 68416, '\p{Block=	 avestan}', "");
Expect(1, 68416, '\p{^Block=	 avestan}', "");
Expect(1, 68416, '\P{Block=	 avestan}', "");
Expect(0, 68416, '\P{^Block=	 avestan}', "");
Error('\p{Blk=/a/-	Avestan}');
Error('\P{Blk=/a/-	Avestan}');
Expect(1, 68415, '\p{Blk=avestan}', "");
Expect(0, 68415, '\p{^Blk=avestan}', "");
Expect(0, 68415, '\P{Blk=avestan}', "");
Expect(1, 68415, '\P{^Blk=avestan}', "");
Expect(0, 68416, '\p{Blk=avestan}', "");
Expect(1, 68416, '\p{^Blk=avestan}', "");
Expect(1, 68416, '\P{Blk=avestan}', "");
Expect(0, 68416, '\P{^Blk=avestan}', "");
Expect(1, 68415, '\p{Blk: -AVESTAN}', "");
Expect(0, 68415, '\p{^Blk: -AVESTAN}', "");
Expect(0, 68415, '\P{Blk: -AVESTAN}', "");
Expect(1, 68415, '\P{^Blk: -AVESTAN}', "");
Expect(0, 68416, '\p{Blk: -AVESTAN}', "");
Expect(1, 68416, '\p{^Blk: -AVESTAN}', "");
Expect(1, 68416, '\P{Blk: -AVESTAN}', "");
Expect(0, 68416, '\P{^Blk: -AVESTAN}', "");
Error('\p{Is_Block=-	avestan/a/}');
Error('\P{Is_Block=-	avestan/a/}');
Expect(1, 68415, '\p{Is_Block=avestan}', "");
Expect(0, 68415, '\p{^Is_Block=avestan}', "");
Expect(0, 68415, '\P{Is_Block=avestan}', "");
Expect(1, 68415, '\P{^Is_Block=avestan}', "");
Expect(0, 68416, '\p{Is_Block=avestan}', "");
Expect(1, 68416, '\p{^Is_Block=avestan}', "");
Expect(1, 68416, '\P{Is_Block=avestan}', "");
Expect(0, 68416, '\P{^Is_Block=avestan}', "");
Expect(1, 68415, '\p{Is_Block=--Avestan}', "");
Expect(0, 68415, '\p{^Is_Block=--Avestan}', "");
Expect(0, 68415, '\P{Is_Block=--Avestan}', "");
Expect(1, 68415, '\P{^Is_Block=--Avestan}', "");
Expect(0, 68416, '\p{Is_Block=--Avestan}', "");
Expect(1, 68416, '\p{^Is_Block=--Avestan}', "");
Expect(1, 68416, '\P{Is_Block=--Avestan}', "");
Expect(0, 68416, '\P{^Is_Block=--Avestan}', "");
Error('\p{Is_Blk=-/a/avestan}');
Error('\P{Is_Blk=-/a/avestan}');
Expect(1, 68415, '\p{Is_Blk:avestan}', "");
Expect(0, 68415, '\p{^Is_Blk:avestan}', "");
Expect(0, 68415, '\P{Is_Blk:avestan}', "");
Expect(1, 68415, '\P{^Is_Blk:avestan}', "");
Expect(0, 68416, '\p{Is_Blk:avestan}', "");
Expect(1, 68416, '\p{^Is_Blk:avestan}', "");
Expect(1, 68416, '\P{Is_Blk:avestan}', "");
Expect(0, 68416, '\P{^Is_Blk:avestan}', "");
Expect(1, 68415, '\p{Is_Blk=_avestan}', "");
Expect(0, 68415, '\p{^Is_Blk=_avestan}', "");
Expect(0, 68415, '\P{Is_Blk=_avestan}', "");
Expect(1, 68415, '\P{^Is_Blk=_avestan}', "");
Expect(0, 68416, '\p{Is_Blk=_avestan}', "");
Expect(1, 68416, '\p{^Is_Blk=_avestan}', "");
Expect(1, 68416, '\P{Is_Blk=_avestan}', "");
Expect(0, 68416, '\P{^Is_Blk=_avestan}', "");
Error('\p{Block=:=	_BALINESE}');
Error('\P{Block=:=	_BALINESE}');
Expect(1, 7039, '\p{Block=balinese}', "");
Expect(0, 7039, '\p{^Block=balinese}', "");
Expect(0, 7039, '\P{Block=balinese}', "");
Expect(1, 7039, '\P{^Block=balinese}', "");
Expect(0, 7040, '\p{Block=balinese}', "");
Expect(1, 7040, '\p{^Block=balinese}', "");
Expect(1, 7040, '\P{Block=balinese}', "");
Expect(0, 7040, '\P{^Block=balinese}', "");
Expect(1, 7039, '\p{Block=-	Balinese}', "");
Expect(0, 7039, '\p{^Block=-	Balinese}', "");
Expect(0, 7039, '\P{Block=-	Balinese}', "");
Expect(1, 7039, '\P{^Block=-	Balinese}', "");
Expect(0, 7040, '\p{Block=-	Balinese}', "");
Expect(1, 7040, '\p{^Block=-	Balinese}', "");
Expect(1, 7040, '\P{Block=-	Balinese}', "");
Expect(0, 7040, '\P{^Block=-	Balinese}', "");
Error('\p{Blk=:= 	Balinese}');
Error('\P{Blk=:= 	Balinese}');
Expect(1, 7039, '\p{Blk=balinese}', "");
Expect(0, 7039, '\p{^Blk=balinese}', "");
Expect(0, 7039, '\P{Blk=balinese}', "");
Expect(1, 7039, '\P{^Blk=balinese}', "");
Expect(0, 7040, '\p{Blk=balinese}', "");
Expect(1, 7040, '\p{^Blk=balinese}', "");
Expect(1, 7040, '\P{Blk=balinese}', "");
Expect(0, 7040, '\P{^Blk=balinese}', "");
Expect(1, 7039, '\p{Blk=	Balinese}', "");
Expect(0, 7039, '\p{^Blk=	Balinese}', "");
Expect(0, 7039, '\P{Blk=	Balinese}', "");
Expect(1, 7039, '\P{^Blk=	Balinese}', "");
Expect(0, 7040, '\p{Blk=	Balinese}', "");
Expect(1, 7040, '\p{^Blk=	Balinese}', "");
Expect(1, 7040, '\P{Blk=	Balinese}', "");
Expect(0, 7040, '\P{^Blk=	Balinese}', "");
Error('\p{Is_Block=_/a/BALINESE}');
Error('\P{Is_Block=_/a/BALINESE}');
Expect(1, 7039, '\p{Is_Block=balinese}', "");
Expect(0, 7039, '\p{^Is_Block=balinese}', "");
Expect(0, 7039, '\P{Is_Block=balinese}', "");
Expect(1, 7039, '\P{^Is_Block=balinese}', "");
Expect(0, 7040, '\p{Is_Block=balinese}', "");
Expect(1, 7040, '\p{^Is_Block=balinese}', "");
Expect(1, 7040, '\P{Is_Block=balinese}', "");
Expect(0, 7040, '\P{^Is_Block=balinese}', "");
Expect(1, 7039, '\p{Is_Block= _Balinese}', "");
Expect(0, 7039, '\p{^Is_Block= _Balinese}', "");
Expect(0, 7039, '\P{Is_Block= _Balinese}', "");
Expect(1, 7039, '\P{^Is_Block= _Balinese}', "");
Expect(0, 7040, '\p{Is_Block= _Balinese}', "");
Expect(1, 7040, '\p{^Is_Block= _Balinese}', "");
Expect(1, 7040, '\P{Is_Block= _Balinese}', "");
Expect(0, 7040, '\P{^Is_Block= _Balinese}', "");
Error('\p{Is_Blk=:=_balinese}');
Error('\P{Is_Blk=:=_balinese}');
Expect(1, 7039, '\p{Is_Blk=balinese}', "");
Expect(0, 7039, '\p{^Is_Blk=balinese}', "");
Expect(0, 7039, '\P{Is_Blk=balinese}', "");
Expect(1, 7039, '\P{^Is_Blk=balinese}', "");
Expect(0, 7040, '\p{Is_Blk=balinese}', "");
Expect(1, 7040, '\p{^Is_Blk=balinese}', "");
Expect(1, 7040, '\P{Is_Blk=balinese}', "");
Expect(0, 7040, '\P{^Is_Blk=balinese}', "");
Expect(1, 7039, '\p{Is_Blk=_Balinese}', "");
Expect(0, 7039, '\p{^Is_Blk=_Balinese}', "");
Expect(0, 7039, '\P{Is_Blk=_Balinese}', "");
Expect(1, 7039, '\P{^Is_Blk=_Balinese}', "");
Expect(0, 7040, '\p{Is_Blk=_Balinese}', "");
Expect(1, 7040, '\p{^Is_Blk=_Balinese}', "");
Expect(1, 7040, '\P{Is_Blk=_Balinese}', "");
Expect(0, 7040, '\P{^Is_Blk=_Balinese}', "");
Error('\p{Block= /a/Bamum}');
Error('\P{Block= /a/Bamum}');
Expect(1, 42751, '\p{Block=bamum}', "");
Expect(0, 42751, '\p{^Block=bamum}', "");
Expect(0, 42751, '\P{Block=bamum}', "");
Expect(1, 42751, '\P{^Block=bamum}', "");
Expect(0, 42752, '\p{Block=bamum}', "");
Expect(1, 42752, '\p{^Block=bamum}', "");
Expect(1, 42752, '\P{Block=bamum}', "");
Expect(0, 42752, '\P{^Block=bamum}', "");
Expect(1, 42751, '\p{Block= Bamum}', "");
Expect(0, 42751, '\p{^Block= Bamum}', "");
Expect(0, 42751, '\P{Block= Bamum}', "");
Expect(1, 42751, '\P{^Block= Bamum}', "");
Expect(0, 42752, '\p{Block= Bamum}', "");
Expect(1, 42752, '\p{^Block= Bamum}', "");
Expect(1, 42752, '\P{Block= Bamum}', "");
Expect(0, 42752, '\P{^Block= Bamum}', "");
Error('\p{Blk=_/a/BAMUM}');
Error('\P{Blk=_/a/BAMUM}');
Expect(1, 42751, '\p{Blk=bamum}', "");
Expect(0, 42751, '\p{^Blk=bamum}', "");
Expect(0, 42751, '\P{Blk=bamum}', "");
Expect(1, 42751, '\P{^Blk=bamum}', "");
Expect(0, 42752, '\p{Blk=bamum}', "");
Expect(1, 42752, '\p{^Blk=bamum}', "");
Expect(1, 42752, '\P{Blk=bamum}', "");
Expect(0, 42752, '\P{^Blk=bamum}', "");
Expect(1, 42751, '\p{Blk:   _-Bamum}', "");
Expect(0, 42751, '\p{^Blk:   _-Bamum}', "");
Expect(0, 42751, '\P{Blk:   _-Bamum}', "");
Expect(1, 42751, '\P{^Blk:   _-Bamum}', "");
Expect(0, 42752, '\p{Blk:   _-Bamum}', "");
Expect(1, 42752, '\p{^Blk:   _-Bamum}', "");
Expect(1, 42752, '\P{Blk:   _-Bamum}', "");
Expect(0, 42752, '\P{^Blk:   _-Bamum}', "");
Error('\p{Is_Block:		Bamum:=}');
Error('\P{Is_Block:		Bamum:=}');
Expect(1, 42751, '\p{Is_Block=bamum}', "");
Expect(0, 42751, '\p{^Is_Block=bamum}', "");
Expect(0, 42751, '\P{Is_Block=bamum}', "");
Expect(1, 42751, '\P{^Is_Block=bamum}', "");
Expect(0, 42752, '\p{Is_Block=bamum}', "");
Expect(1, 42752, '\p{^Is_Block=bamum}', "");
Expect(1, 42752, '\P{Is_Block=bamum}', "");
Expect(0, 42752, '\P{^Is_Block=bamum}', "");
Expect(1, 42751, '\p{Is_Block=_Bamum}', "");
Expect(0, 42751, '\p{^Is_Block=_Bamum}', "");
Expect(0, 42751, '\P{Is_Block=_Bamum}', "");
Expect(1, 42751, '\P{^Is_Block=_Bamum}', "");
Expect(0, 42752, '\p{Is_Block=_Bamum}', "");
Expect(1, 42752, '\p{^Is_Block=_Bamum}', "");
Expect(1, 42752, '\P{Is_Block=_Bamum}', "");
Expect(0, 42752, '\P{^Is_Block=_Bamum}', "");
Error('\p{Is_Blk=/a/BAMUM}');
Error('\P{Is_Blk=/a/BAMUM}');
Expect(1, 42751, '\p{Is_Blk=bamum}', "");
Expect(0, 42751, '\p{^Is_Blk=bamum}', "");
Expect(0, 42751, '\P{Is_Blk=bamum}', "");
Expect(1, 42751, '\P{^Is_Blk=bamum}', "");
Expect(0, 42752, '\p{Is_Blk=bamum}', "");
Expect(1, 42752, '\p{^Is_Blk=bamum}', "");
Expect(1, 42752, '\P{Is_Blk=bamum}', "");
Expect(0, 42752, '\P{^Is_Blk=bamum}', "");
Expect(1, 42751, '\p{Is_Blk= Bamum}', "");
Expect(0, 42751, '\p{^Is_Blk= Bamum}', "");
Expect(0, 42751, '\P{Is_Blk= Bamum}', "");
Expect(1, 42751, '\P{^Is_Blk= Bamum}', "");
Expect(0, 42752, '\p{Is_Blk= Bamum}', "");
Expect(1, 42752, '\p{^Is_Blk= Bamum}', "");
Expect(1, 42752, '\P{Is_Blk= Bamum}', "");
Expect(0, 42752, '\P{^Is_Blk= Bamum}', "");
Error('\p{Block=/a/	 Bamum_SUPPLEMENT}');
Error('\P{Block=/a/	 Bamum_SUPPLEMENT}');
Expect(1, 92735, '\p{Block=bamumsupplement}', "");
Expect(0, 92735, '\p{^Block=bamumsupplement}', "");
Expect(0, 92735, '\P{Block=bamumsupplement}', "");
Expect(1, 92735, '\P{^Block=bamumsupplement}', "");
Expect(0, 92736, '\p{Block=bamumsupplement}', "");
Expect(1, 92736, '\p{^Block=bamumsupplement}', "");
Expect(1, 92736, '\P{Block=bamumsupplement}', "");
Expect(0, 92736, '\P{^Block=bamumsupplement}', "");
Expect(1, 92735, '\p{Block=-Bamum_supplement}', "");
Expect(0, 92735, '\p{^Block=-Bamum_supplement}', "");
Expect(0, 92735, '\P{Block=-Bamum_supplement}', "");
Expect(1, 92735, '\P{^Block=-Bamum_supplement}', "");
Expect(0, 92736, '\p{Block=-Bamum_supplement}', "");
Expect(1, 92736, '\p{^Block=-Bamum_supplement}', "");
Expect(1, 92736, '\P{Block=-Bamum_supplement}', "");
Expect(0, 92736, '\P{^Block=-Bamum_supplement}', "");
Error('\p{Blk=_Bamum_sup:=}');
Error('\P{Blk=_Bamum_sup:=}');
Expect(1, 92735, '\p{Blk=bamumsup}', "");
Expect(0, 92735, '\p{^Blk=bamumsup}', "");
Expect(0, 92735, '\P{Blk=bamumsup}', "");
Expect(1, 92735, '\P{^Blk=bamumsup}', "");
Expect(0, 92736, '\p{Blk=bamumsup}', "");
Expect(1, 92736, '\p{^Blk=bamumsup}', "");
Expect(1, 92736, '\P{Blk=bamumsup}', "");
Expect(0, 92736, '\P{^Blk=bamumsup}', "");
Expect(1, 92735, '\p{Blk=_-Bamum_SUP}', "");
Expect(0, 92735, '\p{^Blk=_-Bamum_SUP}', "");
Expect(0, 92735, '\P{Blk=_-Bamum_SUP}', "");
Expect(1, 92735, '\P{^Blk=_-Bamum_SUP}', "");
Expect(0, 92736, '\p{Blk=_-Bamum_SUP}', "");
Expect(1, 92736, '\p{^Blk=_-Bamum_SUP}', "");
Expect(1, 92736, '\P{Blk=_-Bamum_SUP}', "");
Expect(0, 92736, '\P{^Blk=_-Bamum_SUP}', "");
Error('\p{Is_Block=-BAMUM_supplement/a/}');
Error('\P{Is_Block=-BAMUM_supplement/a/}');
Expect(1, 92735, '\p{Is_Block=bamumsupplement}', "");
Expect(0, 92735, '\p{^Is_Block=bamumsupplement}', "");
Expect(0, 92735, '\P{Is_Block=bamumsupplement}', "");
Expect(1, 92735, '\P{^Is_Block=bamumsupplement}', "");
Expect(0, 92736, '\p{Is_Block=bamumsupplement}', "");
Expect(1, 92736, '\p{^Is_Block=bamumsupplement}', "");
Expect(1, 92736, '\P{Is_Block=bamumsupplement}', "");
Expect(0, 92736, '\P{^Is_Block=bamumsupplement}', "");
Expect(1, 92735, '\p{Is_Block=	-bamum_Supplement}', "");
Expect(0, 92735, '\p{^Is_Block=	-bamum_Supplement}', "");
Expect(0, 92735, '\P{Is_Block=	-bamum_Supplement}', "");
Expect(1, 92735, '\P{^Is_Block=	-bamum_Supplement}', "");
Expect(0, 92736, '\p{Is_Block=	-bamum_Supplement}', "");
Expect(1, 92736, '\p{^Is_Block=	-bamum_Supplement}', "");
Expect(1, 92736, '\P{Is_Block=	-bamum_Supplement}', "");
Expect(0, 92736, '\P{^Is_Block=	-bamum_Supplement}', "");
Error('\p{Is_Blk= :=Bamum_SUP}');
Error('\P{Is_Blk= :=Bamum_SUP}');
Expect(1, 92735, '\p{Is_Blk=bamumsup}', "");
Expect(0, 92735, '\p{^Is_Blk=bamumsup}', "");
Expect(0, 92735, '\P{Is_Blk=bamumsup}', "");
Expect(1, 92735, '\P{^Is_Blk=bamumsup}', "");
Expect(0, 92736, '\p{Is_Blk=bamumsup}', "");
Expect(1, 92736, '\p{^Is_Blk=bamumsup}', "");
Expect(1, 92736, '\P{Is_Blk=bamumsup}', "");
Expect(0, 92736, '\P{^Is_Blk=bamumsup}', "");
Expect(1, 92735, '\p{Is_Blk= 	bamum_Sup}', "");
Expect(0, 92735, '\p{^Is_Blk= 	bamum_Sup}', "");
Expect(0, 92735, '\P{Is_Blk= 	bamum_Sup}', "");
Expect(1, 92735, '\P{^Is_Blk= 	bamum_Sup}', "");
Expect(0, 92736, '\p{Is_Blk= 	bamum_Sup}', "");
Expect(1, 92736, '\p{^Is_Blk= 	bamum_Sup}', "");
Expect(1, 92736, '\P{Is_Blk= 	bamum_Sup}', "");
Expect(0, 92736, '\P{^Is_Blk= 	bamum_Sup}', "");
Error('\p{Block=_:=BATAK}');
Error('\P{Block=_:=BATAK}');
Expect(1, 7167, '\p{Block=batak}', "");
Expect(0, 7167, '\p{^Block=batak}', "");
Expect(0, 7167, '\P{Block=batak}', "");
Expect(1, 7167, '\P{^Block=batak}', "");
Expect(0, 7168, '\p{Block=batak}', "");
Expect(1, 7168, '\p{^Block=batak}', "");
Expect(1, 7168, '\P{Block=batak}', "");
Expect(0, 7168, '\P{^Block=batak}', "");
Expect(1, 7167, '\p{Block= 	Batak}', "");
Expect(0, 7167, '\p{^Block= 	Batak}', "");
Expect(0, 7167, '\P{Block= 	Batak}', "");
Expect(1, 7167, '\P{^Block= 	Batak}', "");
Expect(0, 7168, '\p{Block= 	Batak}', "");
Expect(1, 7168, '\p{^Block= 	Batak}', "");
Expect(1, 7168, '\P{Block= 	Batak}', "");
Expect(0, 7168, '\P{^Block= 	Batak}', "");
Error('\p{Blk= 	Batak/a/}');
Error('\P{Blk= 	Batak/a/}');
Expect(1, 7167, '\p{Blk=batak}', "");
Expect(0, 7167, '\p{^Blk=batak}', "");
Expect(0, 7167, '\P{Blk=batak}', "");
Expect(1, 7167, '\P{^Blk=batak}', "");
Expect(0, 7168, '\p{Blk=batak}', "");
Expect(1, 7168, '\p{^Blk=batak}', "");
Expect(1, 7168, '\P{Blk=batak}', "");
Expect(0, 7168, '\P{^Blk=batak}', "");
Expect(1, 7167, '\p{Blk=_-BATAK}', "");
Expect(0, 7167, '\p{^Blk=_-BATAK}', "");
Expect(0, 7167, '\P{Blk=_-BATAK}', "");
Expect(1, 7167, '\P{^Blk=_-BATAK}', "");
Expect(0, 7168, '\p{Blk=_-BATAK}', "");
Expect(1, 7168, '\p{^Blk=_-BATAK}', "");
Expect(1, 7168, '\P{Blk=_-BATAK}', "");
Expect(0, 7168, '\P{^Blk=_-BATAK}', "");
Error('\p{Is_Block=	/a/Batak}');
Error('\P{Is_Block=	/a/Batak}');
Expect(1, 7167, '\p{Is_Block: batak}', "");
Expect(0, 7167, '\p{^Is_Block: batak}', "");
Expect(0, 7167, '\P{Is_Block: batak}', "");
Expect(1, 7167, '\P{^Is_Block: batak}', "");
Expect(0, 7168, '\p{Is_Block: batak}', "");
Expect(1, 7168, '\p{^Is_Block: batak}', "");
Expect(1, 7168, '\P{Is_Block: batak}', "");
Expect(0, 7168, '\P{^Is_Block: batak}', "");
Expect(1, 7167, '\p{Is_Block=_BATAK}', "");
Expect(0, 7167, '\p{^Is_Block=_BATAK}', "");
Expect(0, 7167, '\P{Is_Block=_BATAK}', "");
Expect(1, 7167, '\P{^Is_Block=_BATAK}', "");
Expect(0, 7168, '\p{Is_Block=_BATAK}', "");
Expect(1, 7168, '\p{^Is_Block=_BATAK}', "");
Expect(1, 7168, '\P{Is_Block=_BATAK}', "");
Expect(0, 7168, '\P{^Is_Block=_BATAK}', "");
Error('\p{Is_Blk=_ Batak/a/}');
Error('\P{Is_Blk=_ Batak/a/}');
Expect(1, 7167, '\p{Is_Blk=batak}', "");
Expect(0, 7167, '\p{^Is_Blk=batak}', "");
Expect(0, 7167, '\P{Is_Blk=batak}', "");
Expect(1, 7167, '\P{^Is_Blk=batak}', "");
Expect(0, 7168, '\p{Is_Blk=batak}', "");
Expect(1, 7168, '\p{^Is_Blk=batak}', "");
Expect(1, 7168, '\P{Is_Blk=batak}', "");
Expect(0, 7168, '\P{^Is_Blk=batak}', "");
Expect(1, 7167, '\p{Is_Blk=	-batak}', "");
Expect(0, 7167, '\p{^Is_Blk=	-batak}', "");
Expect(0, 7167, '\P{Is_Blk=	-batak}', "");
Expect(1, 7167, '\P{^Is_Blk=	-batak}', "");
Expect(0, 7168, '\p{Is_Blk=	-batak}', "");
Expect(1, 7168, '\p{^Is_Blk=	-batak}', "");
Expect(1, 7168, '\P{Is_Blk=	-batak}', "");
Expect(0, 7168, '\P{^Is_Blk=	-batak}', "");
Error('\p{Block=_-BENGALI/a/}');
Error('\P{Block=_-BENGALI/a/}');
Expect(1, 2559, '\p{Block=bengali}', "");
Expect(0, 2559, '\p{^Block=bengali}', "");
Expect(0, 2559, '\P{Block=bengali}', "");
Expect(1, 2559, '\P{^Block=bengali}', "");
Expect(0, 2560, '\p{Block=bengali}', "");
Expect(1, 2560, '\p{^Block=bengali}', "");
Expect(1, 2560, '\P{Block=bengali}', "");
Expect(0, 2560, '\P{^Block=bengali}', "");
Expect(1, 2559, '\p{Block=--bengali}', "");
Expect(0, 2559, '\p{^Block=--bengali}', "");
Expect(0, 2559, '\P{Block=--bengali}', "");
Expect(1, 2559, '\P{^Block=--bengali}', "");
Expect(0, 2560, '\p{Block=--bengali}', "");
Expect(1, 2560, '\p{^Block=--bengali}', "");
Expect(1, 2560, '\P{Block=--bengali}', "");
Expect(0, 2560, '\P{^Block=--bengali}', "");
Error('\p{Blk=:=_ Bengali}');
Error('\P{Blk=:=_ Bengali}');
Expect(1, 2559, '\p{Blk:   bengali}', "");
Expect(0, 2559, '\p{^Blk:   bengali}', "");
Expect(0, 2559, '\P{Blk:   bengali}', "");
Expect(1, 2559, '\P{^Blk:   bengali}', "");
Expect(0, 2560, '\p{Blk:   bengali}', "");
Expect(1, 2560, '\p{^Blk:   bengali}', "");
Expect(1, 2560, '\P{Blk:   bengali}', "");
Expect(0, 2560, '\P{^Blk:   bengali}', "");
Expect(1, 2559, '\p{Blk=	-BENGALI}', "");
Expect(0, 2559, '\p{^Blk=	-BENGALI}', "");
Expect(0, 2559, '\P{Blk=	-BENGALI}', "");
Expect(1, 2559, '\P{^Blk=	-BENGALI}', "");
Expect(0, 2560, '\p{Blk=	-BENGALI}', "");
Expect(1, 2560, '\p{^Blk=	-BENGALI}', "");
Expect(1, 2560, '\P{Blk=	-BENGALI}', "");
Expect(0, 2560, '\P{^Blk=	-BENGALI}', "");
Error('\p{Is_Block= _BENGALI/a/}');
Error('\P{Is_Block= _BENGALI/a/}');
Expect(1, 2559, '\p{Is_Block=bengali}', "");
Expect(0, 2559, '\p{^Is_Block=bengali}', "");
Expect(0, 2559, '\P{Is_Block=bengali}', "");
Expect(1, 2559, '\P{^Is_Block=bengali}', "");
Expect(0, 2560, '\p{Is_Block=bengali}', "");
Expect(1, 2560, '\p{^Is_Block=bengali}', "");
Expect(1, 2560, '\P{Is_Block=bengali}', "");
Expect(0, 2560, '\P{^Is_Block=bengali}', "");
Expect(1, 2559, '\p{Is_Block: Bengali}', "");
Expect(0, 2559, '\p{^Is_Block: Bengali}', "");
Expect(0, 2559, '\P{Is_Block: Bengali}', "");
Expect(1, 2559, '\P{^Is_Block: Bengali}', "");
Expect(0, 2560, '\p{Is_Block: Bengali}', "");
Expect(1, 2560, '\p{^Is_Block: Bengali}', "");
Expect(1, 2560, '\P{Is_Block: Bengali}', "");
Expect(0, 2560, '\P{^Is_Block: Bengali}', "");
Error('\p{Is_Blk:	/a/ BENGALI}');
Error('\P{Is_Blk:	/a/ BENGALI}');
Expect(1, 2559, '\p{Is_Blk:	bengali}', "");
Expect(0, 2559, '\p{^Is_Blk:	bengali}', "");
Expect(0, 2559, '\P{Is_Blk:	bengali}', "");
Expect(1, 2559, '\P{^Is_Blk:	bengali}', "");
Expect(0, 2560, '\p{Is_Blk:	bengali}', "");
Expect(1, 2560, '\p{^Is_Blk:	bengali}', "");
Expect(1, 2560, '\P{Is_Blk:	bengali}', "");
Expect(0, 2560, '\P{^Is_Blk:	bengali}', "");
Expect(1, 2559, '\p{Is_Blk=-bengali}', "");
Expect(0, 2559, '\p{^Is_Blk=-bengali}', "");
Expect(0, 2559, '\P{Is_Blk=-bengali}', "");
Expect(1, 2559, '\P{^Is_Blk=-bengali}', "");
Expect(0, 2560, '\p{Is_Blk=-bengali}', "");
Expect(1, 2560, '\p{^Is_Blk=-bengali}', "");
Expect(1, 2560, '\P{Is_Blk=-bengali}', "");
Expect(0, 2560, '\P{^Is_Blk=-bengali}', "");
Error('\p{Block=_:=BLOCK_Elements}');
Error('\P{Block=_:=BLOCK_Elements}');
Expect(1, 9631, '\p{Block=blockelements}', "");
Expect(0, 9631, '\p{^Block=blockelements}', "");
Expect(0, 9631, '\P{Block=blockelements}', "");
Expect(1, 9631, '\P{^Block=blockelements}', "");
Expect(0, 9632, '\p{Block=blockelements}', "");
Expect(1, 9632, '\p{^Block=blockelements}', "");
Expect(1, 9632, '\P{Block=blockelements}', "");
Expect(0, 9632, '\P{^Block=blockelements}', "");
Expect(1, 9631, '\p{Block= _Block_elements}', "");
Expect(0, 9631, '\p{^Block= _Block_elements}', "");
Expect(0, 9631, '\P{Block= _Block_elements}', "");
Expect(1, 9631, '\P{^Block= _Block_elements}', "");
Expect(0, 9632, '\p{Block= _Block_elements}', "");
Expect(1, 9632, '\p{^Block= _Block_elements}', "");
Expect(1, 9632, '\P{Block= _Block_elements}', "");
Expect(0, 9632, '\P{^Block= _Block_elements}', "");
Error('\p{Blk=Block_Elements:=}');
Error('\P{Blk=Block_Elements:=}');
Expect(1, 9631, '\p{Blk=blockelements}', "");
Expect(0, 9631, '\p{^Blk=blockelements}', "");
Expect(0, 9631, '\P{Blk=blockelements}', "");
Expect(1, 9631, '\P{^Blk=blockelements}', "");
Expect(0, 9632, '\p{Blk=blockelements}', "");
Expect(1, 9632, '\p{^Blk=blockelements}', "");
Expect(1, 9632, '\P{Blk=blockelements}', "");
Expect(0, 9632, '\P{^Blk=blockelements}', "");
Expect(1, 9631, '\p{Blk=	BLOCK_ELEMENTS}', "");
Expect(0, 9631, '\p{^Blk=	BLOCK_ELEMENTS}', "");
Expect(0, 9631, '\P{Blk=	BLOCK_ELEMENTS}', "");
Expect(1, 9631, '\P{^Blk=	BLOCK_ELEMENTS}', "");
Expect(0, 9632, '\p{Blk=	BLOCK_ELEMENTS}', "");
Expect(1, 9632, '\p{^Blk=	BLOCK_ELEMENTS}', "");
Expect(1, 9632, '\P{Blk=	BLOCK_ELEMENTS}', "");
Expect(0, 9632, '\P{^Blk=	BLOCK_ELEMENTS}', "");
Error('\p{Is_Block=/a/_-Block_elements}');
Error('\P{Is_Block=/a/_-Block_elements}');
Expect(1, 9631, '\p{Is_Block=blockelements}', "");
Expect(0, 9631, '\p{^Is_Block=blockelements}', "");
Expect(0, 9631, '\P{Is_Block=blockelements}', "");
Expect(1, 9631, '\P{^Is_Block=blockelements}', "");
Expect(0, 9632, '\p{Is_Block=blockelements}', "");
Expect(1, 9632, '\p{^Is_Block=blockelements}', "");
Expect(1, 9632, '\P{Is_Block=blockelements}', "");
Expect(0, 9632, '\P{^Is_Block=blockelements}', "");
Expect(1, 9631, '\p{Is_Block=  Block_Elements}', "");
Expect(0, 9631, '\p{^Is_Block=  Block_Elements}', "");
Expect(0, 9631, '\P{Is_Block=  Block_Elements}', "");
Expect(1, 9631, '\P{^Is_Block=  Block_Elements}', "");
Expect(0, 9632, '\p{Is_Block=  Block_Elements}', "");
Expect(1, 9632, '\p{^Is_Block=  Block_Elements}', "");
Expect(1, 9632, '\P{Is_Block=  Block_Elements}', "");
Expect(0, 9632, '\P{^Is_Block=  Block_Elements}', "");
Error('\p{Is_Blk= block_Elements/a/}');
Error('\P{Is_Blk= block_Elements/a/}');
Expect(1, 9631, '\p{Is_Blk=blockelements}', "");
Expect(0, 9631, '\p{^Is_Blk=blockelements}', "");
Expect(0, 9631, '\P{Is_Blk=blockelements}', "");
Expect(1, 9631, '\P{^Is_Blk=blockelements}', "");
Expect(0, 9632, '\p{Is_Blk=blockelements}', "");
Expect(1, 9632, '\p{^Is_Blk=blockelements}', "");
Expect(1, 9632, '\P{Is_Blk=blockelements}', "");
Expect(0, 9632, '\P{^Is_Blk=blockelements}', "");
Expect(1, 9631, '\p{Is_Blk=_-BLOCK_ELEMENTS}', "");
Expect(0, 9631, '\p{^Is_Blk=_-BLOCK_ELEMENTS}', "");
Expect(0, 9631, '\P{Is_Blk=_-BLOCK_ELEMENTS}', "");
Expect(1, 9631, '\P{^Is_Blk=_-BLOCK_ELEMENTS}', "");
Expect(0, 9632, '\p{Is_Blk=_-BLOCK_ELEMENTS}', "");
Expect(1, 9632, '\p{^Is_Blk=_-BLOCK_ELEMENTS}', "");
Expect(1, 9632, '\P{Is_Blk=_-BLOCK_ELEMENTS}', "");
Expect(0, 9632, '\P{^Is_Blk=_-BLOCK_ELEMENTS}', "");
Error('\p{Block=:=	-Bopomofo}');
Error('\P{Block=:=	-Bopomofo}');
Expect(1, 12591, '\p{Block=bopomofo}', "");
Expect(0, 12591, '\p{^Block=bopomofo}', "");
Expect(0, 12591, '\P{Block=bopomofo}', "");
Expect(1, 12591, '\P{^Block=bopomofo}', "");
Expect(0, 12592, '\p{Block=bopomofo}', "");
Expect(1, 12592, '\p{^Block=bopomofo}', "");
Expect(1, 12592, '\P{Block=bopomofo}', "");
Expect(0, 12592, '\P{^Block=bopomofo}', "");
Expect(1, 12591, '\p{Block=_	BOPOMOFO}', "");
Expect(0, 12591, '\p{^Block=_	BOPOMOFO}', "");
Expect(0, 12591, '\P{Block=_	BOPOMOFO}', "");
Expect(1, 12591, '\P{^Block=_	BOPOMOFO}', "");
Expect(0, 12592, '\p{Block=_	BOPOMOFO}', "");
Expect(1, 12592, '\p{^Block=_	BOPOMOFO}', "");
Expect(1, 12592, '\P{Block=_	BOPOMOFO}', "");
Expect(0, 12592, '\P{^Block=_	BOPOMOFO}', "");
Error('\p{Blk:   :=_Bopomofo}');
Error('\P{Blk:   :=_Bopomofo}');
Expect(1, 12591, '\p{Blk=bopomofo}', "");
Expect(0, 12591, '\p{^Blk=bopomofo}', "");
Expect(0, 12591, '\P{Blk=bopomofo}', "");
Expect(1, 12591, '\P{^Blk=bopomofo}', "");
Expect(0, 12592, '\p{Blk=bopomofo}', "");
Expect(1, 12592, '\p{^Blk=bopomofo}', "");
Expect(1, 12592, '\P{Blk=bopomofo}', "");
Expect(0, 12592, '\P{^Blk=bopomofo}', "");
Expect(1, 12591, '\p{Blk=  Bopomofo}', "");
Expect(0, 12591, '\p{^Blk=  Bopomofo}', "");
Expect(0, 12591, '\P{Blk=  Bopomofo}', "");
Expect(1, 12591, '\P{^Blk=  Bopomofo}', "");
Expect(0, 12592, '\p{Blk=  Bopomofo}', "");
Expect(1, 12592, '\p{^Blk=  Bopomofo}', "");
Expect(1, 12592, '\P{Blk=  Bopomofo}', "");
Expect(0, 12592, '\P{^Blk=  Bopomofo}', "");
Error('\p{Is_Block= /a/bopomofo}');
Error('\P{Is_Block= /a/bopomofo}');
Expect(1, 12591, '\p{Is_Block:	bopomofo}', "");
Expect(0, 12591, '\p{^Is_Block:	bopomofo}', "");
Expect(0, 12591, '\P{Is_Block:	bopomofo}', "");
Expect(1, 12591, '\P{^Is_Block:	bopomofo}', "");
Expect(0, 12592, '\p{Is_Block:	bopomofo}', "");
Expect(1, 12592, '\p{^Is_Block:	bopomofo}', "");
Expect(1, 12592, '\P{Is_Block:	bopomofo}', "");
Expect(0, 12592, '\P{^Is_Block:	bopomofo}', "");
Expect(1, 12591, '\p{Is_Block=-	BOPOMOFO}', "");
Expect(0, 12591, '\p{^Is_Block=-	BOPOMOFO}', "");
Expect(0, 12591, '\P{Is_Block=-	BOPOMOFO}', "");
Expect(1, 12591, '\P{^Is_Block=-	BOPOMOFO}', "");
Expect(0, 12592, '\p{Is_Block=-	BOPOMOFO}', "");
Expect(1, 12592, '\p{^Is_Block=-	BOPOMOFO}', "");
Expect(1, 12592, '\P{Is_Block=-	BOPOMOFO}', "");
Expect(0, 12592, '\P{^Is_Block=-	BOPOMOFO}', "");
Error('\p{Is_Blk= :=BOPOMOFO}');
Error('\P{Is_Blk= :=BOPOMOFO}');
Expect(1, 12591, '\p{Is_Blk=bopomofo}', "");
Expect(0, 12591, '\p{^Is_Blk=bopomofo}', "");
Expect(0, 12591, '\P{Is_Blk=bopomofo}', "");
Expect(1, 12591, '\P{^Is_Blk=bopomofo}', "");
Expect(0, 12592, '\p{Is_Blk=bopomofo}', "");
Expect(1, 12592, '\p{^Is_Blk=bopomofo}', "");
Expect(1, 12592, '\P{Is_Blk=bopomofo}', "");
Expect(0, 12592, '\P{^Is_Blk=bopomofo}', "");
Expect(1, 12591, '\p{Is_Blk=		Bopomofo}', "");
Expect(0, 12591, '\p{^Is_Blk=		Bopomofo}', "");
Expect(0, 12591, '\P{Is_Blk=		Bopomofo}', "");
Expect(1, 12591, '\P{^Is_Blk=		Bopomofo}', "");
Expect(0, 12592, '\p{Is_Blk=		Bopomofo}', "");
Expect(1, 12592, '\p{^Is_Blk=		Bopomofo}', "");
Expect(1, 12592, '\P{Is_Blk=		Bopomofo}', "");
Expect(0, 12592, '\P{^Is_Blk=		Bopomofo}', "");
Error('\p{Block=	:=BOPOMOFO_Extended}');
Error('\P{Block=	:=BOPOMOFO_Extended}');
Expect(1, 12735, '\p{Block=bopomofoextended}', "");
Expect(0, 12735, '\p{^Block=bopomofoextended}', "");
Expect(0, 12735, '\P{Block=bopomofoextended}', "");
Expect(1, 12735, '\P{^Block=bopomofoextended}', "");
Expect(0, 12736, '\p{Block=bopomofoextended}', "");
Expect(1, 12736, '\p{^Block=bopomofoextended}', "");
Expect(1, 12736, '\P{Block=bopomofoextended}', "");
Expect(0, 12736, '\P{^Block=bopomofoextended}', "");
Expect(1, 12735, '\p{Block=	Bopomofo_Extended}', "");
Expect(0, 12735, '\p{^Block=	Bopomofo_Extended}', "");
Expect(0, 12735, '\P{Block=	Bopomofo_Extended}', "");
Expect(1, 12735, '\P{^Block=	Bopomofo_Extended}', "");
Expect(0, 12736, '\p{Block=	Bopomofo_Extended}', "");
Expect(1, 12736, '\p{^Block=	Bopomofo_Extended}', "");
Expect(1, 12736, '\P{Block=	Bopomofo_Extended}', "");
Expect(0, 12736, '\P{^Block=	Bopomofo_Extended}', "");
Error('\p{Blk= /a/Bopomofo_EXT}');
Error('\P{Blk= /a/Bopomofo_EXT}');
Expect(1, 12735, '\p{Blk=bopomofoext}', "");
Expect(0, 12735, '\p{^Blk=bopomofoext}', "");
Expect(0, 12735, '\P{Blk=bopomofoext}', "");
Expect(1, 12735, '\P{^Blk=bopomofoext}', "");
Expect(0, 12736, '\p{Blk=bopomofoext}', "");
Expect(1, 12736, '\p{^Blk=bopomofoext}', "");
Expect(1, 12736, '\P{Blk=bopomofoext}', "");
Expect(0, 12736, '\P{^Blk=bopomofoext}', "");
Expect(1, 12735, '\p{Blk=_Bopomofo_ext}', "");
Expect(0, 12735, '\p{^Blk=_Bopomofo_ext}', "");
Expect(0, 12735, '\P{Blk=_Bopomofo_ext}', "");
Expect(1, 12735, '\P{^Blk=_Bopomofo_ext}', "");
Expect(0, 12736, '\p{Blk=_Bopomofo_ext}', "");
Expect(1, 12736, '\p{^Blk=_Bopomofo_ext}', "");
Expect(1, 12736, '\P{Blk=_Bopomofo_ext}', "");
Expect(0, 12736, '\P{^Blk=_Bopomofo_ext}', "");
Error('\p{Is_Block=		BOPOMOFO_extended:=}');
Error('\P{Is_Block=		BOPOMOFO_extended:=}');
Expect(1, 12735, '\p{Is_Block=bopomofoextended}', "");
Expect(0, 12735, '\p{^Is_Block=bopomofoextended}', "");
Expect(0, 12735, '\P{Is_Block=bopomofoextended}', "");
Expect(1, 12735, '\P{^Is_Block=bopomofoextended}', "");
Expect(0, 12736, '\p{Is_Block=bopomofoextended}', "");
Expect(1, 12736, '\p{^Is_Block=bopomofoextended}', "");
Expect(1, 12736, '\P{Is_Block=bopomofoextended}', "");
Expect(0, 12736, '\P{^Is_Block=bopomofoextended}', "");
Expect(1, 12735, '\p{Is_Block=_Bopomofo_EXTENDED}', "");
Expect(0, 12735, '\p{^Is_Block=_Bopomofo_EXTENDED}', "");
Expect(0, 12735, '\P{Is_Block=_Bopomofo_EXTENDED}', "");
Expect(1, 12735, '\P{^Is_Block=_Bopomofo_EXTENDED}', "");
Expect(0, 12736, '\p{Is_Block=_Bopomofo_EXTENDED}', "");
Expect(1, 12736, '\p{^Is_Block=_Bopomofo_EXTENDED}', "");
Expect(1, 12736, '\P{Is_Block=_Bopomofo_EXTENDED}', "");
Expect(0, 12736, '\P{^Is_Block=_Bopomofo_EXTENDED}', "");
Error('\p{Is_Blk=:= 	BOPOMOFO_ext}');
Error('\P{Is_Blk=:= 	BOPOMOFO_ext}');
Expect(1, 12735, '\p{Is_Blk=bopomofoext}', "");
Expect(0, 12735, '\p{^Is_Blk=bopomofoext}', "");
Expect(0, 12735, '\P{Is_Blk=bopomofoext}', "");
Expect(1, 12735, '\P{^Is_Blk=bopomofoext}', "");
Expect(0, 12736, '\p{Is_Blk=bopomofoext}', "");
Expect(1, 12736, '\p{^Is_Blk=bopomofoext}', "");
Expect(1, 12736, '\P{Is_Blk=bopomofoext}', "");
Expect(0, 12736, '\P{^Is_Blk=bopomofoext}', "");
Expect(1, 12735, '\p{Is_Blk=BOPOMOFO_EXT}', "");
Expect(0, 12735, '\p{^Is_Blk=BOPOMOFO_EXT}', "");
Expect(0, 12735, '\P{Is_Blk=BOPOMOFO_EXT}', "");
Expect(1, 12735, '\P{^Is_Blk=BOPOMOFO_EXT}', "");
Expect(0, 12736, '\p{Is_Blk=BOPOMOFO_EXT}', "");
Expect(1, 12736, '\p{^Is_Blk=BOPOMOFO_EXT}', "");
Expect(1, 12736, '\P{Is_Blk=BOPOMOFO_EXT}', "");
Expect(0, 12736, '\P{^Is_Blk=BOPOMOFO_EXT}', "");
Error('\p{Block=/a/Box_Drawing}');
Error('\P{Block=/a/Box_Drawing}');
Expect(1, 9599, '\p{Block=boxdrawing}', "");
Expect(0, 9599, '\p{^Block=boxdrawing}', "");
Expect(0, 9599, '\P{Block=boxdrawing}', "");
Expect(1, 9599, '\P{^Block=boxdrawing}', "");
Expect(0, 9600, '\p{Block=boxdrawing}', "");
Expect(1, 9600, '\p{^Block=boxdrawing}', "");
Expect(1, 9600, '\P{Block=boxdrawing}', "");
Expect(0, 9600, '\P{^Block=boxdrawing}', "");
Expect(1, 9599, '\p{Block=_ box_Drawing}', "");
Expect(0, 9599, '\p{^Block=_ box_Drawing}', "");
Expect(0, 9599, '\P{Block=_ box_Drawing}', "");
Expect(1, 9599, '\P{^Block=_ box_Drawing}', "");
Expect(0, 9600, '\p{Block=_ box_Drawing}', "");
Expect(1, 9600, '\p{^Block=_ box_Drawing}', "");
Expect(1, 9600, '\P{Block=_ box_Drawing}', "");
Expect(0, 9600, '\P{^Block=_ box_Drawing}', "");
Error('\p{Blk= 	box_Drawing/a/}');
Error('\P{Blk= 	box_Drawing/a/}');
Expect(1, 9599, '\p{Blk=boxdrawing}', "");
Expect(0, 9599, '\p{^Blk=boxdrawing}', "");
Expect(0, 9599, '\P{Blk=boxdrawing}', "");
Expect(1, 9599, '\P{^Blk=boxdrawing}', "");
Expect(0, 9600, '\p{Blk=boxdrawing}', "");
Expect(1, 9600, '\p{^Blk=boxdrawing}', "");
Expect(1, 9600, '\P{Blk=boxdrawing}', "");
Expect(0, 9600, '\P{^Blk=boxdrawing}', "");
Expect(1, 9599, '\p{Blk=--Box_DRAWING}', "");
Expect(0, 9599, '\p{^Blk=--Box_DRAWING}', "");
Expect(0, 9599, '\P{Blk=--Box_DRAWING}', "");
Expect(1, 9599, '\P{^Blk=--Box_DRAWING}', "");
Expect(0, 9600, '\p{Blk=--Box_DRAWING}', "");
Expect(1, 9600, '\p{^Blk=--Box_DRAWING}', "");
Expect(1, 9600, '\P{Blk=--Box_DRAWING}', "");
Expect(0, 9600, '\P{^Blk=--Box_DRAWING}', "");
Error('\p{Is_Block=:= Box_Drawing}');
Error('\P{Is_Block=:= Box_Drawing}');
Expect(1, 9599, '\p{Is_Block=boxdrawing}', "");
Expect(0, 9599, '\p{^Is_Block=boxdrawing}', "");
Expect(0, 9599, '\P{Is_Block=boxdrawing}', "");
Expect(1, 9599, '\P{^Is_Block=boxdrawing}', "");
Expect(0, 9600, '\p{Is_Block=boxdrawing}', "");
Expect(1, 9600, '\p{^Is_Block=boxdrawing}', "");
Expect(1, 9600, '\P{Is_Block=boxdrawing}', "");
Expect(0, 9600, '\P{^Is_Block=boxdrawing}', "");
Expect(1, 9599, '\p{Is_Block=-box_Drawing}', "");
Expect(0, 9599, '\p{^Is_Block=-box_Drawing}', "");
Expect(0, 9599, '\P{Is_Block=-box_Drawing}', "");
Expect(1, 9599, '\P{^Is_Block=-box_Drawing}', "");
Expect(0, 9600, '\p{Is_Block=-box_Drawing}', "");
Expect(1, 9600, '\p{^Is_Block=-box_Drawing}', "");
Expect(1, 9600, '\P{Is_Block=-box_Drawing}', "");
Expect(0, 9600, '\P{^Is_Block=-box_Drawing}', "");
Error('\p{Is_Blk=/a/-Box_Drawing}');
Error('\P{Is_Blk=/a/-Box_Drawing}');
Expect(1, 9599, '\p{Is_Blk: boxdrawing}', "");
Expect(0, 9599, '\p{^Is_Blk: boxdrawing}', "");
Expect(0, 9599, '\P{Is_Blk: boxdrawing}', "");
Expect(1, 9599, '\P{^Is_Blk: boxdrawing}', "");
Expect(0, 9600, '\p{Is_Blk: boxdrawing}', "");
Expect(1, 9600, '\p{^Is_Blk: boxdrawing}', "");
Expect(1, 9600, '\P{Is_Blk: boxdrawing}', "");
Expect(0, 9600, '\P{^Is_Blk: boxdrawing}', "");
Expect(1, 9599, '\p{Is_Blk: 		Box_Drawing}', "");
Expect(0, 9599, '\p{^Is_Blk: 		Box_Drawing}', "");
Expect(0, 9599, '\P{Is_Blk: 		Box_Drawing}', "");
Expect(1, 9599, '\P{^Is_Blk: 		Box_Drawing}', "");
Expect(0, 9600, '\p{Is_Blk: 		Box_Drawing}', "");
Expect(1, 9600, '\p{^Is_Blk: 		Box_Drawing}', "");
Expect(1, 9600, '\P{Is_Blk: 		Box_Drawing}', "");
Expect(0, 9600, '\P{^Is_Blk: 		Box_Drawing}', "");
Error('\p{Block=/a/ -brahmi}');
Error('\P{Block=/a/ -brahmi}');
Expect(1, 69759, '\p{Block=brahmi}', "");
Expect(0, 69759, '\p{^Block=brahmi}', "");
Expect(0, 69759, '\P{Block=brahmi}', "");
Expect(1, 69759, '\P{^Block=brahmi}', "");
Expect(0, 69760, '\p{Block=brahmi}', "");
Expect(1, 69760, '\p{^Block=brahmi}', "");
Expect(1, 69760, '\P{Block=brahmi}', "");
Expect(0, 69760, '\P{^Block=brahmi}', "");
Expect(1, 69759, '\p{Block= Brahmi}', "");
Expect(0, 69759, '\p{^Block= Brahmi}', "");
Expect(0, 69759, '\P{Block= Brahmi}', "");
Expect(1, 69759, '\P{^Block= Brahmi}', "");
Expect(0, 69760, '\p{Block= Brahmi}', "");
Expect(1, 69760, '\p{^Block= Brahmi}', "");
Expect(1, 69760, '\P{Block= Brahmi}', "");
Expect(0, 69760, '\P{^Block= Brahmi}', "");
Error('\p{Blk=/a/Brahmi}');
Error('\P{Blk=/a/Brahmi}');
Expect(1, 69759, '\p{Blk:   brahmi}', "");
Expect(0, 69759, '\p{^Blk:   brahmi}', "");
Expect(0, 69759, '\P{Blk:   brahmi}', "");
Expect(1, 69759, '\P{^Blk:   brahmi}', "");
Expect(0, 69760, '\p{Blk:   brahmi}', "");
Expect(1, 69760, '\p{^Blk:   brahmi}', "");
Expect(1, 69760, '\P{Blk:   brahmi}', "");
Expect(0, 69760, '\P{^Blk:   brahmi}', "");
Expect(1, 69759, '\p{Blk=--brahmi}', "");
Expect(0, 69759, '\p{^Blk=--brahmi}', "");
Expect(0, 69759, '\P{Blk=--brahmi}', "");
Expect(1, 69759, '\P{^Blk=--brahmi}', "");
Expect(0, 69760, '\p{Blk=--brahmi}', "");
Expect(1, 69760, '\p{^Blk=--brahmi}', "");
Expect(1, 69760, '\P{Blk=--brahmi}', "");
Expect(0, 69760, '\P{^Blk=--brahmi}', "");
Error('\p{Is_Block=		Brahmi:=}');
Error('\P{Is_Block=		Brahmi:=}');
Expect(1, 69759, '\p{Is_Block=brahmi}', "");
Expect(0, 69759, '\p{^Is_Block=brahmi}', "");
Expect(0, 69759, '\P{Is_Block=brahmi}', "");
Expect(1, 69759, '\P{^Is_Block=brahmi}', "");
Expect(0, 69760, '\p{Is_Block=brahmi}', "");
Expect(1, 69760, '\p{^Is_Block=brahmi}', "");
Expect(1, 69760, '\P{Is_Block=brahmi}', "");
Expect(0, 69760, '\P{^Is_Block=brahmi}', "");
Expect(1, 69759, '\p{Is_Block=_Brahmi}', "");
Expect(0, 69759, '\p{^Is_Block=_Brahmi}', "");
Expect(0, 69759, '\P{Is_Block=_Brahmi}', "");
Expect(1, 69759, '\P{^Is_Block=_Brahmi}', "");
Expect(0, 69760, '\p{Is_Block=_Brahmi}', "");
Expect(1, 69760, '\p{^Is_Block=_Brahmi}', "");
Expect(1, 69760, '\P{Is_Block=_Brahmi}', "");
Expect(0, 69760, '\P{^Is_Block=_Brahmi}', "");
Error('\p{Is_Blk=:= _Brahmi}');
Error('\P{Is_Blk=:= _Brahmi}');
Expect(1, 69759, '\p{Is_Blk=brahmi}', "");
Expect(0, 69759, '\p{^Is_Blk=brahmi}', "");
Expect(0, 69759, '\P{Is_Blk=brahmi}', "");
Expect(1, 69759, '\P{^Is_Blk=brahmi}', "");
Expect(0, 69760, '\p{Is_Blk=brahmi}', "");
Expect(1, 69760, '\p{^Is_Blk=brahmi}', "");
Expect(1, 69760, '\P{Is_Blk=brahmi}', "");
Expect(0, 69760, '\P{^Is_Blk=brahmi}', "");
Expect(1, 69759, '\p{Is_Blk=--Brahmi}', "");
Expect(0, 69759, '\p{^Is_Blk=--Brahmi}', "");
Expect(0, 69759, '\P{Is_Blk=--Brahmi}', "");
Expect(1, 69759, '\P{^Is_Blk=--Brahmi}', "");
Expect(0, 69760, '\p{Is_Blk=--Brahmi}', "");
Expect(1, 69760, '\p{^Is_Blk=--Brahmi}', "");
Expect(1, 69760, '\P{Is_Blk=--Brahmi}', "");
Expect(0, 69760, '\P{^Is_Blk=--Brahmi}', "");
Error('\p{Block:	:=_-Braille_patterns}');
Error('\P{Block:	:=_-Braille_patterns}');
Expect(1, 10495, '\p{Block:braillepatterns}', "");
Expect(0, 10495, '\p{^Block:braillepatterns}', "");
Expect(0, 10495, '\P{Block:braillepatterns}', "");
Expect(1, 10495, '\P{^Block:braillepatterns}', "");
Expect(0, 10496, '\p{Block:braillepatterns}', "");
Expect(1, 10496, '\p{^Block:braillepatterns}', "");
Expect(1, 10496, '\P{Block:braillepatterns}', "");
Expect(0, 10496, '\P{^Block:braillepatterns}', "");
Expect(1, 10495, '\p{Block=_Braille_Patterns}', "");
Expect(0, 10495, '\p{^Block=_Braille_Patterns}', "");
Expect(0, 10495, '\P{Block=_Braille_Patterns}', "");
Expect(1, 10495, '\P{^Block=_Braille_Patterns}', "");
Expect(0, 10496, '\p{Block=_Braille_Patterns}', "");
Expect(1, 10496, '\p{^Block=_Braille_Patterns}', "");
Expect(1, 10496, '\P{Block=_Braille_Patterns}', "");
Expect(0, 10496, '\P{^Block=_Braille_Patterns}', "");
Error('\p{Blk=_:=Braille}');
Error('\P{Blk=_:=Braille}');
Expect(1, 10495, '\p{Blk:	braille}', "");
Expect(0, 10495, '\p{^Blk:	braille}', "");
Expect(0, 10495, '\P{Blk:	braille}', "");
Expect(1, 10495, '\P{^Blk:	braille}', "");
Expect(0, 10496, '\p{Blk:	braille}', "");
Expect(1, 10496, '\p{^Blk:	braille}', "");
Expect(1, 10496, '\P{Blk:	braille}', "");
Expect(0, 10496, '\P{^Blk:	braille}', "");
Expect(1, 10495, '\p{Blk:   Braille}', "");
Expect(0, 10495, '\p{^Blk:   Braille}', "");
Expect(0, 10495, '\P{Blk:   Braille}', "");
Expect(1, 10495, '\P{^Blk:   Braille}', "");
Expect(0, 10496, '\p{Blk:   Braille}', "");
Expect(1, 10496, '\p{^Blk:   Braille}', "");
Expect(1, 10496, '\P{Blk:   Braille}', "");
Expect(0, 10496, '\P{^Blk:   Braille}', "");
Error('\p{Is_Block=_-Braille_PATTERNS:=}');
Error('\P{Is_Block=_-Braille_PATTERNS:=}');
Expect(1, 10495, '\p{Is_Block=braillepatterns}', "");
Expect(0, 10495, '\p{^Is_Block=braillepatterns}', "");
Expect(0, 10495, '\P{Is_Block=braillepatterns}', "");
Expect(1, 10495, '\P{^Is_Block=braillepatterns}', "");
Expect(0, 10496, '\p{Is_Block=braillepatterns}', "");
Expect(1, 10496, '\p{^Is_Block=braillepatterns}', "");
Expect(1, 10496, '\P{Is_Block=braillepatterns}', "");
Expect(0, 10496, '\P{^Is_Block=braillepatterns}', "");
Expect(1, 10495, '\p{Is_Block=	 Braille_patterns}', "");
Expect(0, 10495, '\p{^Is_Block=	 Braille_patterns}', "");
Expect(0, 10495, '\P{Is_Block=	 Braille_patterns}', "");
Expect(1, 10495, '\P{^Is_Block=	 Braille_patterns}', "");
Expect(0, 10496, '\p{Is_Block=	 Braille_patterns}', "");
Expect(1, 10496, '\p{^Is_Block=	 Braille_patterns}', "");
Expect(1, 10496, '\P{Is_Block=	 Braille_patterns}', "");
Expect(0, 10496, '\P{^Is_Block=	 Braille_patterns}', "");
Error('\p{Is_Blk=-braille:=}');
Error('\P{Is_Blk=-braille:=}');
Expect(1, 10495, '\p{Is_Blk: braille}', "");
Expect(0, 10495, '\p{^Is_Blk: braille}', "");
Expect(0, 10495, '\P{Is_Blk: braille}', "");
Expect(1, 10495, '\P{^Is_Blk: braille}', "");
Expect(0, 10496, '\p{Is_Blk: braille}', "");
Expect(1, 10496, '\p{^Is_Blk: braille}', "");
Expect(1, 10496, '\P{Is_Blk: braille}', "");
Expect(0, 10496, '\P{^Is_Blk: braille}', "");
Expect(1, 10495, '\p{Is_Blk=	-braille}', "");
Expect(0, 10495, '\p{^Is_Blk=	-braille}', "");
Expect(0, 10495, '\P{Is_Blk=	-braille}', "");
Expect(1, 10495, '\P{^Is_Blk=	-braille}', "");
Expect(0, 10496, '\p{Is_Blk=	-braille}', "");
Expect(1, 10496, '\p{^Is_Blk=	-braille}', "");
Expect(1, 10496, '\P{Is_Blk=	-braille}', "");
Expect(0, 10496, '\P{^Is_Blk=	-braille}', "");
Error('\p{Block:	_:=Buginese}');
Error('\P{Block:	_:=Buginese}');
Expect(1, 6687, '\p{Block=buginese}', "");
Expect(0, 6687, '\p{^Block=buginese}', "");
Expect(0, 6687, '\P{Block=buginese}', "");
Expect(1, 6687, '\P{^Block=buginese}', "");
Expect(0, 6688, '\p{Block=buginese}', "");
Expect(1, 6688, '\p{^Block=buginese}', "");
Expect(1, 6688, '\P{Block=buginese}', "");
Expect(0, 6688, '\P{^Block=buginese}', "");
Expect(1, 6687, '\p{Block:    -Buginese}', "");
Expect(0, 6687, '\p{^Block:    -Buginese}', "");
Expect(0, 6687, '\P{Block:    -Buginese}', "");
Expect(1, 6687, '\P{^Block:    -Buginese}', "");
Expect(0, 6688, '\p{Block:    -Buginese}', "");
Expect(1, 6688, '\p{^Block:    -Buginese}', "");
Expect(1, 6688, '\P{Block:    -Buginese}', "");
Expect(0, 6688, '\P{^Block:    -Buginese}', "");
Error('\p{Blk=/a/-Buginese}');
Error('\P{Blk=/a/-Buginese}');
Expect(1, 6687, '\p{Blk:	buginese}', "");
Expect(0, 6687, '\p{^Blk:	buginese}', "");
Expect(0, 6687, '\P{Blk:	buginese}', "");
Expect(1, 6687, '\P{^Blk:	buginese}', "");
Expect(0, 6688, '\p{Blk:	buginese}', "");
Expect(1, 6688, '\p{^Blk:	buginese}', "");
Expect(1, 6688, '\P{Blk:	buginese}', "");
Expect(0, 6688, '\P{^Blk:	buginese}', "");
Expect(1, 6687, '\p{Blk=--Buginese}', "");
Expect(0, 6687, '\p{^Blk=--Buginese}', "");
Expect(0, 6687, '\P{Blk=--Buginese}', "");
Expect(1, 6687, '\P{^Blk=--Buginese}', "");
Expect(0, 6688, '\p{Blk=--Buginese}', "");
Expect(1, 6688, '\p{^Blk=--Buginese}', "");
Expect(1, 6688, '\P{Blk=--Buginese}', "");
Expect(0, 6688, '\P{^Blk=--Buginese}', "");
Error('\p{Is_Block= /a/buginese}');
Error('\P{Is_Block= /a/buginese}');
Expect(1, 6687, '\p{Is_Block=buginese}', "");
Expect(0, 6687, '\p{^Is_Block=buginese}', "");
Expect(0, 6687, '\P{Is_Block=buginese}', "");
Expect(1, 6687, '\P{^Is_Block=buginese}', "");
Expect(0, 6688, '\p{Is_Block=buginese}', "");
Expect(1, 6688, '\p{^Is_Block=buginese}', "");
Expect(1, 6688, '\P{Is_Block=buginese}', "");
Expect(0, 6688, '\P{^Is_Block=buginese}', "");
Expect(1, 6687, '\p{Is_Block=	_Buginese}', "");
Expect(0, 6687, '\p{^Is_Block=	_Buginese}', "");
Expect(0, 6687, '\P{Is_Block=	_Buginese}', "");
Expect(1, 6687, '\P{^Is_Block=	_Buginese}', "");
Expect(0, 6688, '\p{Is_Block=	_Buginese}', "");
Expect(1, 6688, '\p{^Is_Block=	_Buginese}', "");
Expect(1, 6688, '\P{Is_Block=	_Buginese}', "");
Expect(0, 6688, '\P{^Is_Block=	_Buginese}', "");
Error('\p{Is_Blk=	 Buginese/a/}');
Error('\P{Is_Blk=	 Buginese/a/}');
Expect(1, 6687, '\p{Is_Blk=buginese}', "");
Expect(0, 6687, '\p{^Is_Blk=buginese}', "");
Expect(0, 6687, '\P{Is_Blk=buginese}', "");
Expect(1, 6687, '\P{^Is_Blk=buginese}', "");
Expect(0, 6688, '\p{Is_Blk=buginese}', "");
Expect(1, 6688, '\p{^Is_Blk=buginese}', "");
Expect(1, 6688, '\P{Is_Blk=buginese}', "");
Expect(0, 6688, '\P{^Is_Blk=buginese}', "");
Expect(1, 6687, '\p{Is_Blk= Buginese}', "");
Expect(0, 6687, '\p{^Is_Blk= Buginese}', "");
Expect(0, 6687, '\P{Is_Blk= Buginese}', "");
Expect(1, 6687, '\P{^Is_Blk= Buginese}', "");
Expect(0, 6688, '\p{Is_Blk= Buginese}', "");
Expect(1, 6688, '\p{^Is_Blk= Buginese}', "");
Expect(1, 6688, '\P{Is_Blk= Buginese}', "");
Expect(0, 6688, '\P{^Is_Blk= Buginese}', "");
Error('\p{Block= :=BUHID}');
Error('\P{Block= :=BUHID}');
Expect(1, 5983, '\p{Block=buhid}', "");
Expect(0, 5983, '\p{^Block=buhid}', "");
Expect(0, 5983, '\P{Block=buhid}', "");
Expect(1, 5983, '\P{^Block=buhid}', "");
Expect(0, 5984, '\p{Block=buhid}', "");
Expect(1, 5984, '\p{^Block=buhid}', "");
Expect(1, 5984, '\P{Block=buhid}', "");
Expect(0, 5984, '\P{^Block=buhid}', "");
Expect(1, 5983, '\p{Block=-BUHID}', "");
Expect(0, 5983, '\p{^Block=-BUHID}', "");
Expect(0, 5983, '\P{Block=-BUHID}', "");
Expect(1, 5983, '\P{^Block=-BUHID}', "");
Expect(0, 5984, '\p{Block=-BUHID}', "");
Expect(1, 5984, '\p{^Block=-BUHID}', "");
Expect(1, 5984, '\P{Block=-BUHID}', "");
Expect(0, 5984, '\P{^Block=-BUHID}', "");
Error('\p{Blk= :=Buhid}');
Error('\P{Blk= :=Buhid}');
Expect(1, 5983, '\p{Blk=buhid}', "");
Expect(0, 5983, '\p{^Blk=buhid}', "");
Expect(0, 5983, '\P{Blk=buhid}', "");
Expect(1, 5983, '\P{^Blk=buhid}', "");
Expect(0, 5984, '\p{Blk=buhid}', "");
Expect(1, 5984, '\p{^Blk=buhid}', "");
Expect(1, 5984, '\P{Blk=buhid}', "");
Expect(0, 5984, '\P{^Blk=buhid}', "");
Expect(1, 5983, '\p{Blk=	_buhid}', "");
Expect(0, 5983, '\p{^Blk=	_buhid}', "");
Expect(0, 5983, '\P{Blk=	_buhid}', "");
Expect(1, 5983, '\P{^Blk=	_buhid}', "");
Expect(0, 5984, '\p{Blk=	_buhid}', "");
Expect(1, 5984, '\p{^Blk=	_buhid}', "");
Expect(1, 5984, '\P{Blk=	_buhid}', "");
Expect(0, 5984, '\P{^Blk=	_buhid}', "");
Error('\p{Is_Block:	 -Buhid/a/}');
Error('\P{Is_Block:	 -Buhid/a/}');
Expect(1, 5983, '\p{Is_Block=buhid}', "");
Expect(0, 5983, '\p{^Is_Block=buhid}', "");
Expect(0, 5983, '\P{Is_Block=buhid}', "");
Expect(1, 5983, '\P{^Is_Block=buhid}', "");
Expect(0, 5984, '\p{Is_Block=buhid}', "");
Expect(1, 5984, '\p{^Is_Block=buhid}', "");
Expect(1, 5984, '\P{Is_Block=buhid}', "");
Expect(0, 5984, '\P{^Is_Block=buhid}', "");
Expect(1, 5983, '\p{Is_Block=	-BUHID}', "");
Expect(0, 5983, '\p{^Is_Block=	-BUHID}', "");
Expect(0, 5983, '\P{Is_Block=	-BUHID}', "");
Expect(1, 5983, '\P{^Is_Block=	-BUHID}', "");
Expect(0, 5984, '\p{Is_Block=	-BUHID}', "");
Expect(1, 5984, '\p{^Is_Block=	-BUHID}', "");
Expect(1, 5984, '\P{Is_Block=	-BUHID}', "");
Expect(0, 5984, '\P{^Is_Block=	-BUHID}', "");
Error('\p{Is_Blk=-BUHID/a/}');
Error('\P{Is_Blk=-BUHID/a/}');
Expect(1, 5983, '\p{Is_Blk=buhid}', "");
Expect(0, 5983, '\p{^Is_Blk=buhid}', "");
Expect(0, 5983, '\P{Is_Blk=buhid}', "");
Expect(1, 5983, '\P{^Is_Blk=buhid}', "");
Expect(0, 5984, '\p{Is_Blk=buhid}', "");
Expect(1, 5984, '\p{^Is_Blk=buhid}', "");
Expect(1, 5984, '\P{Is_Blk=buhid}', "");
Expect(0, 5984, '\P{^Is_Blk=buhid}', "");
Expect(1, 5983, '\p{Is_Blk= -Buhid}', "");
Expect(0, 5983, '\p{^Is_Blk= -Buhid}', "");
Expect(0, 5983, '\P{Is_Blk= -Buhid}', "");
Expect(1, 5983, '\P{^Is_Blk= -Buhid}', "");
Expect(0, 5984, '\p{Is_Blk= -Buhid}', "");
Expect(1, 5984, '\p{^Is_Blk= -Buhid}', "");
Expect(1, 5984, '\P{Is_Blk= -Buhid}', "");
Expect(0, 5984, '\P{^Is_Blk= -Buhid}', "");
Error('\p{Block=_ BYZANTINE_musical_symbols:=}');
Error('\P{Block=_ BYZANTINE_musical_symbols:=}');
Expect(1, 119039, '\p{Block=byzantinemusicalsymbols}', "");
Expect(0, 119039, '\p{^Block=byzantinemusicalsymbols}', "");
Expect(0, 119039, '\P{Block=byzantinemusicalsymbols}', "");
Expect(1, 119039, '\P{^Block=byzantinemusicalsymbols}', "");
Expect(0, 119040, '\p{Block=byzantinemusicalsymbols}', "");
Expect(1, 119040, '\p{^Block=byzantinemusicalsymbols}', "");
Expect(1, 119040, '\P{Block=byzantinemusicalsymbols}', "");
Expect(0, 119040, '\P{^Block=byzantinemusicalsymbols}', "");
Expect(1, 119039, '\p{Block= _byzantine_musical_SYMBOLS}', "");
Expect(0, 119039, '\p{^Block= _byzantine_musical_SYMBOLS}', "");
Expect(0, 119039, '\P{Block= _byzantine_musical_SYMBOLS}', "");
Expect(1, 119039, '\P{^Block= _byzantine_musical_SYMBOLS}', "");
Expect(0, 119040, '\p{Block= _byzantine_musical_SYMBOLS}', "");
Expect(1, 119040, '\p{^Block= _byzantine_musical_SYMBOLS}', "");
Expect(1, 119040, '\P{Block= _byzantine_musical_SYMBOLS}', "");
Expect(0, 119040, '\P{^Block= _byzantine_musical_SYMBOLS}', "");
Error('\p{Blk: 	byzantine_Music:=}');
Error('\P{Blk: 	byzantine_Music:=}');
Expect(1, 119039, '\p{Blk=byzantinemusic}', "");
Expect(0, 119039, '\p{^Blk=byzantinemusic}', "");
Expect(0, 119039, '\P{Blk=byzantinemusic}', "");
Expect(1, 119039, '\P{^Blk=byzantinemusic}', "");
Expect(0, 119040, '\p{Blk=byzantinemusic}', "");
Expect(1, 119040, '\p{^Blk=byzantinemusic}', "");
Expect(1, 119040, '\P{Blk=byzantinemusic}', "");
Expect(0, 119040, '\P{^Blk=byzantinemusic}', "");
Expect(1, 119039, '\p{Blk=_	Byzantine_Music}', "");
Expect(0, 119039, '\p{^Blk=_	Byzantine_Music}', "");
Expect(0, 119039, '\P{Blk=_	Byzantine_Music}', "");
Expect(1, 119039, '\P{^Blk=_	Byzantine_Music}', "");
Expect(0, 119040, '\p{Blk=_	Byzantine_Music}', "");
Expect(1, 119040, '\p{^Blk=_	Byzantine_Music}', "");
Expect(1, 119040, '\P{Blk=_	Byzantine_Music}', "");
Expect(0, 119040, '\P{^Blk=_	Byzantine_Music}', "");
Error('\p{Is_Block=	_BYZANTINE_Musical_Symbols:=}');
Error('\P{Is_Block=	_BYZANTINE_Musical_Symbols:=}');
Expect(1, 119039, '\p{Is_Block=byzantinemusicalsymbols}', "");
Expect(0, 119039, '\p{^Is_Block=byzantinemusicalsymbols}', "");
Expect(0, 119039, '\P{Is_Block=byzantinemusicalsymbols}', "");
Expect(1, 119039, '\P{^Is_Block=byzantinemusicalsymbols}', "");
Expect(0, 119040, '\p{Is_Block=byzantinemusicalsymbols}', "");
Expect(1, 119040, '\p{^Is_Block=byzantinemusicalsymbols}', "");
Expect(1, 119040, '\P{Is_Block=byzantinemusicalsymbols}', "");
Expect(0, 119040, '\P{^Is_Block=byzantinemusicalsymbols}', "");
Expect(1, 119039, '\p{Is_Block=	 BYZANTINE_Musical_Symbols}', "");
Expect(0, 119039, '\p{^Is_Block=	 BYZANTINE_Musical_Symbols}', "");
Expect(0, 119039, '\P{Is_Block=	 BYZANTINE_Musical_Symbols}', "");
Expect(1, 119039, '\P{^Is_Block=	 BYZANTINE_Musical_Symbols}', "");
Expect(0, 119040, '\p{Is_Block=	 BYZANTINE_Musical_Symbols}', "");
Expect(1, 119040, '\p{^Is_Block=	 BYZANTINE_Musical_Symbols}', "");
Expect(1, 119040, '\P{Is_Block=	 BYZANTINE_Musical_Symbols}', "");
Expect(0, 119040, '\P{^Is_Block=	 BYZANTINE_Musical_Symbols}', "");
Error('\p{Is_Blk=-	byzantine_Music:=}');
Error('\P{Is_Blk=-	byzantine_Music:=}');
Expect(1, 119039, '\p{Is_Blk: byzantinemusic}', "");
Expect(0, 119039, '\p{^Is_Blk: byzantinemusic}', "");
Expect(0, 119039, '\P{Is_Blk: byzantinemusic}', "");
Expect(1, 119039, '\P{^Is_Blk: byzantinemusic}', "");
Expect(0, 119040, '\p{Is_Blk: byzantinemusic}', "");
Expect(1, 119040, '\p{^Is_Blk: byzantinemusic}', "");
Expect(1, 119040, '\P{Is_Blk: byzantinemusic}', "");
Expect(0, 119040, '\P{^Is_Blk: byzantinemusic}', "");
Expect(1, 119039, '\p{Is_Blk= -BYZANTINE_MUSIC}', "");
Expect(0, 119039, '\p{^Is_Blk= -BYZANTINE_MUSIC}', "");
Expect(0, 119039, '\P{Is_Blk= -BYZANTINE_MUSIC}', "");
Expect(1, 119039, '\P{^Is_Blk= -BYZANTINE_MUSIC}', "");
Expect(0, 119040, '\p{Is_Blk= -BYZANTINE_MUSIC}', "");
Expect(1, 119040, '\p{^Is_Blk= -BYZANTINE_MUSIC}', "");
Expect(1, 119040, '\P{Is_Blk= -BYZANTINE_MUSIC}', "");
Expect(0, 119040, '\P{^Is_Blk= -BYZANTINE_MUSIC}', "");
Error('\p{Block=-:=CJK_Unified_Ideographs}');
Error('\P{Block=-:=CJK_Unified_Ideographs}');
Expect(1, 40959, '\p{Block=cjkunifiedideographs}', "");
Expect(0, 40959, '\p{^Block=cjkunifiedideographs}', "");
Expect(0, 40959, '\P{Block=cjkunifiedideographs}', "");
Expect(1, 40959, '\P{^Block=cjkunifiedideographs}', "");
Expect(0, 40960, '\p{Block=cjkunifiedideographs}', "");
Expect(1, 40960, '\p{^Block=cjkunifiedideographs}', "");
Expect(1, 40960, '\P{Block=cjkunifiedideographs}', "");
Expect(0, 40960, '\P{^Block=cjkunifiedideographs}', "");
Expect(1, 40959, '\p{Block=- CJK_Unified_IDEOGRAPHS}', "");
Expect(0, 40959, '\p{^Block=- CJK_Unified_IDEOGRAPHS}', "");
Expect(0, 40959, '\P{Block=- CJK_Unified_IDEOGRAPHS}', "");
Expect(1, 40959, '\P{^Block=- CJK_Unified_IDEOGRAPHS}', "");
Expect(0, 40960, '\p{Block=- CJK_Unified_IDEOGRAPHS}', "");
Expect(1, 40960, '\p{^Block=- CJK_Unified_IDEOGRAPHS}', "");
Expect(1, 40960, '\P{Block=- CJK_Unified_IDEOGRAPHS}', "");
Expect(0, 40960, '\P{^Block=- CJK_Unified_IDEOGRAPHS}', "");
Error('\p{Blk=	CJK:=}');
Error('\P{Blk=	CJK:=}');
Expect(1, 40959, '\p{Blk=cjk}', "");
Expect(0, 40959, '\p{^Blk=cjk}', "");
Expect(0, 40959, '\P{Blk=cjk}', "");
Expect(1, 40959, '\P{^Blk=cjk}', "");
Expect(0, 40960, '\p{Blk=cjk}', "");
Expect(1, 40960, '\p{^Blk=cjk}', "");
Expect(1, 40960, '\P{Blk=cjk}', "");
Expect(0, 40960, '\P{^Blk=cjk}', "");
Expect(1, 40959, '\p{Blk= CJK}', "");
Expect(0, 40959, '\p{^Blk= CJK}', "");
Expect(0, 40959, '\P{Blk= CJK}', "");
Expect(1, 40959, '\P{^Blk= CJK}', "");
Expect(0, 40960, '\p{Blk= CJK}', "");
Expect(1, 40960, '\p{^Blk= CJK}', "");
Expect(1, 40960, '\P{Blk= CJK}', "");
Expect(0, 40960, '\P{^Blk= CJK}', "");
Error('\p{Is_Block=	 cjk_unified_Ideographs/a/}');
Error('\P{Is_Block=	 cjk_unified_Ideographs/a/}');
Expect(1, 40959, '\p{Is_Block:	cjkunifiedideographs}', "");
Expect(0, 40959, '\p{^Is_Block:	cjkunifiedideographs}', "");
Expect(0, 40959, '\P{Is_Block:	cjkunifiedideographs}', "");
Expect(1, 40959, '\P{^Is_Block:	cjkunifiedideographs}', "");
Expect(0, 40960, '\p{Is_Block:	cjkunifiedideographs}', "");
Expect(1, 40960, '\p{^Is_Block:	cjkunifiedideographs}', "");
Expect(1, 40960, '\P{Is_Block:	cjkunifiedideographs}', "");
Expect(0, 40960, '\P{^Is_Block:	cjkunifiedideographs}', "");
Expect(1, 40959, '\p{Is_Block:   CJK_UNIFIED_Ideographs}', "");
Expect(0, 40959, '\p{^Is_Block:   CJK_UNIFIED_Ideographs}', "");
Expect(0, 40959, '\P{Is_Block:   CJK_UNIFIED_Ideographs}', "");
Expect(1, 40959, '\P{^Is_Block:   CJK_UNIFIED_Ideographs}', "");
Expect(0, 40960, '\p{Is_Block:   CJK_UNIFIED_Ideographs}', "");
Expect(1, 40960, '\p{^Is_Block:   CJK_UNIFIED_Ideographs}', "");
Expect(1, 40960, '\P{Is_Block:   CJK_UNIFIED_Ideographs}', "");
Expect(0, 40960, '\P{^Is_Block:   CJK_UNIFIED_Ideographs}', "");
Error('\p{Is_Blk:	/a/__CJK}');
Error('\P{Is_Blk:	/a/__CJK}');
Expect(1, 40959, '\p{Is_Blk=cjk}', "");
Expect(0, 40959, '\p{^Is_Blk=cjk}', "");
Expect(0, 40959, '\P{Is_Blk=cjk}', "");
Expect(1, 40959, '\P{^Is_Blk=cjk}', "");
Expect(0, 40960, '\p{Is_Blk=cjk}', "");
Expect(1, 40960, '\p{^Is_Blk=cjk}', "");
Expect(1, 40960, '\P{Is_Blk=cjk}', "");
Expect(0, 40960, '\P{^Is_Blk=cjk}', "");
Expect(1, 40959, '\p{Is_Blk= cjk}', "");
Expect(0, 40959, '\p{^Is_Blk= cjk}', "");
Expect(0, 40959, '\P{Is_Blk= cjk}', "");
Expect(1, 40959, '\P{^Is_Blk= cjk}', "");
Expect(0, 40960, '\p{Is_Blk= cjk}', "");
Expect(1, 40960, '\p{^Is_Blk= cjk}', "");
Expect(1, 40960, '\P{Is_Blk= cjk}', "");
Expect(0, 40960, '\P{^Is_Blk= cjk}', "");
Error('\p{Block=_CJK_compatibility/a/}');
Error('\P{Block=_CJK_compatibility/a/}');
Expect(1, 13311, '\p{Block=cjkcompatibility}', "");
Expect(0, 13311, '\p{^Block=cjkcompatibility}', "");
Expect(0, 13311, '\P{Block=cjkcompatibility}', "");
Expect(1, 13311, '\P{^Block=cjkcompatibility}', "");
Expect(0, 13312, '\p{Block=cjkcompatibility}', "");
Expect(1, 13312, '\p{^Block=cjkcompatibility}', "");
Expect(1, 13312, '\P{Block=cjkcompatibility}', "");
Expect(0, 13312, '\P{^Block=cjkcompatibility}', "");
Expect(1, 13311, '\p{Block= -CJK_Compatibility}', "");
Expect(0, 13311, '\p{^Block= -CJK_Compatibility}', "");
Expect(0, 13311, '\P{Block= -CJK_Compatibility}', "");
Expect(1, 13311, '\P{^Block= -CJK_Compatibility}', "");
Expect(0, 13312, '\p{Block= -CJK_Compatibility}', "");
Expect(1, 13312, '\p{^Block= -CJK_Compatibility}', "");
Expect(1, 13312, '\P{Block= -CJK_Compatibility}', "");
Expect(0, 13312, '\P{^Block= -CJK_Compatibility}', "");
Error('\p{Blk=	/a/cjk_compat}');
Error('\P{Blk=	/a/cjk_compat}');
Expect(1, 13311, '\p{Blk=cjkcompat}', "");
Expect(0, 13311, '\p{^Blk=cjkcompat}', "");
Expect(0, 13311, '\P{Blk=cjkcompat}', "");
Expect(1, 13311, '\P{^Blk=cjkcompat}', "");
Expect(0, 13312, '\p{Blk=cjkcompat}', "");
Expect(1, 13312, '\p{^Blk=cjkcompat}', "");
Expect(1, 13312, '\P{Blk=cjkcompat}', "");
Expect(0, 13312, '\P{^Blk=cjkcompat}', "");
Expect(1, 13311, '\p{Blk=	-CJK_compat}', "");
Expect(0, 13311, '\p{^Blk=	-CJK_compat}', "");
Expect(0, 13311, '\P{Blk=	-CJK_compat}', "");
Expect(1, 13311, '\P{^Blk=	-CJK_compat}', "");
Expect(0, 13312, '\p{Blk=	-CJK_compat}', "");
Expect(1, 13312, '\p{^Blk=	-CJK_compat}', "");
Expect(1, 13312, '\P{Blk=	-CJK_compat}', "");
Expect(0, 13312, '\P{^Blk=	-CJK_compat}', "");
Error('\p{Is_Block:   /a/-	CJK_COMPATIBILITY}');
Error('\P{Is_Block:   /a/-	CJK_COMPATIBILITY}');
Expect(1, 13311, '\p{Is_Block=cjkcompatibility}', "");
Expect(0, 13311, '\p{^Is_Block=cjkcompatibility}', "");
Expect(0, 13311, '\P{Is_Block=cjkcompatibility}', "");
Expect(1, 13311, '\P{^Is_Block=cjkcompatibility}', "");
Expect(0, 13312, '\p{Is_Block=cjkcompatibility}', "");
Expect(1, 13312, '\p{^Is_Block=cjkcompatibility}', "");
Expect(1, 13312, '\P{Is_Block=cjkcompatibility}', "");
Expect(0, 13312, '\P{^Is_Block=cjkcompatibility}', "");
Expect(1, 13311, '\p{Is_Block=	-CJK_Compatibility}', "");
Expect(0, 13311, '\p{^Is_Block=	-CJK_Compatibility}', "");
Expect(0, 13311, '\P{Is_Block=	-CJK_Compatibility}', "");
Expect(1, 13311, '\P{^Is_Block=	-CJK_Compatibility}', "");
Expect(0, 13312, '\p{Is_Block=	-CJK_Compatibility}', "");
Expect(1, 13312, '\p{^Is_Block=	-CJK_Compatibility}', "");
Expect(1, 13312, '\P{Is_Block=	-CJK_Compatibility}', "");
Expect(0, 13312, '\P{^Is_Block=	-CJK_Compatibility}', "");
Error('\p{Is_Blk= CJK_Compat:=}');
Error('\P{Is_Blk= CJK_Compat:=}');
Expect(1, 13311, '\p{Is_Blk=cjkcompat}', "");
Expect(0, 13311, '\p{^Is_Blk=cjkcompat}', "");
Expect(0, 13311, '\P{Is_Blk=cjkcompat}', "");
Expect(1, 13311, '\P{^Is_Blk=cjkcompat}', "");
Expect(0, 13312, '\p{Is_Blk=cjkcompat}', "");
Expect(1, 13312, '\p{^Is_Blk=cjkcompat}', "");
Expect(1, 13312, '\P{Is_Blk=cjkcompat}', "");
Expect(0, 13312, '\P{^Is_Blk=cjkcompat}', "");
Expect(1, 13311, '\p{Is_Blk=-CJK_compat}', "");
Expect(0, 13311, '\p{^Is_Blk=-CJK_compat}', "");
Expect(0, 13311, '\P{Is_Blk=-CJK_compat}', "");
Expect(1, 13311, '\P{^Is_Blk=-CJK_compat}', "");
Expect(0, 13312, '\p{Is_Blk=-CJK_compat}', "");
Expect(1, 13312, '\p{^Is_Blk=-CJK_compat}', "");
Expect(1, 13312, '\P{Is_Blk=-CJK_compat}', "");
Expect(0, 13312, '\P{^Is_Blk=-CJK_compat}', "");
Error('\p{Block=- CJK_Compatibility_forms:=}');
Error('\P{Block=- CJK_Compatibility_forms:=}');
Expect(1, 65103, '\p{Block=cjkcompatibilityforms}', "");
Expect(0, 65103, '\p{^Block=cjkcompatibilityforms}', "");
Expect(0, 65103, '\P{Block=cjkcompatibilityforms}', "");
Expect(1, 65103, '\P{^Block=cjkcompatibilityforms}', "");
Expect(0, 65104, '\p{Block=cjkcompatibilityforms}', "");
Expect(1, 65104, '\p{^Block=cjkcompatibilityforms}', "");
Expect(1, 65104, '\P{Block=cjkcompatibilityforms}', "");
Expect(0, 65104, '\P{^Block=cjkcompatibilityforms}', "");
Expect(1, 65103, '\p{Block=_CJK_Compatibility_Forms}', "");
Expect(0, 65103, '\p{^Block=_CJK_Compatibility_Forms}', "");
Expect(0, 65103, '\P{Block=_CJK_Compatibility_Forms}', "");
Expect(1, 65103, '\P{^Block=_CJK_Compatibility_Forms}', "");
Expect(0, 65104, '\p{Block=_CJK_Compatibility_Forms}', "");
Expect(1, 65104, '\p{^Block=_CJK_Compatibility_Forms}', "");
Expect(1, 65104, '\P{Block=_CJK_Compatibility_Forms}', "");
Expect(0, 65104, '\P{^Block=_CJK_Compatibility_Forms}', "");
Error('\p{Blk:	/a/CJK_COMPAT_FORMS}');
Error('\P{Blk:	/a/CJK_COMPAT_FORMS}');
Expect(1, 65103, '\p{Blk=cjkcompatforms}', "");
Expect(0, 65103, '\p{^Blk=cjkcompatforms}', "");
Expect(0, 65103, '\P{Blk=cjkcompatforms}', "");
Expect(1, 65103, '\P{^Blk=cjkcompatforms}', "");
Expect(0, 65104, '\p{Blk=cjkcompatforms}', "");
Expect(1, 65104, '\p{^Blk=cjkcompatforms}', "");
Expect(1, 65104, '\P{Blk=cjkcompatforms}', "");
Expect(0, 65104, '\P{^Blk=cjkcompatforms}', "");
Expect(1, 65103, '\p{Blk=--CJK_Compat_Forms}', "");
Expect(0, 65103, '\p{^Blk=--CJK_Compat_Forms}', "");
Expect(0, 65103, '\P{Blk=--CJK_Compat_Forms}', "");
Expect(1, 65103, '\P{^Blk=--CJK_Compat_Forms}', "");
Expect(0, 65104, '\p{Blk=--CJK_Compat_Forms}', "");
Expect(1, 65104, '\p{^Blk=--CJK_Compat_Forms}', "");
Expect(1, 65104, '\P{Blk=--CJK_Compat_Forms}', "");
Expect(0, 65104, '\P{^Blk=--CJK_Compat_Forms}', "");
Error('\p{Is_Block=-:=CJK_compatibility_Forms}');
Error('\P{Is_Block=-:=CJK_compatibility_Forms}');
Expect(1, 65103, '\p{Is_Block=cjkcompatibilityforms}', "");
Expect(0, 65103, '\p{^Is_Block=cjkcompatibilityforms}', "");
Expect(0, 65103, '\P{Is_Block=cjkcompatibilityforms}', "");
Expect(1, 65103, '\P{^Is_Block=cjkcompatibilityforms}', "");
Expect(0, 65104, '\p{Is_Block=cjkcompatibilityforms}', "");
Expect(1, 65104, '\p{^Is_Block=cjkcompatibilityforms}', "");
Expect(1, 65104, '\P{Is_Block=cjkcompatibilityforms}', "");
Expect(0, 65104, '\P{^Is_Block=cjkcompatibilityforms}', "");
Expect(1, 65103, '\p{Is_Block=_cjk_Compatibility_Forms}', "");
Expect(0, 65103, '\p{^Is_Block=_cjk_Compatibility_Forms}', "");
Expect(0, 65103, '\P{Is_Block=_cjk_Compatibility_Forms}', "");
Expect(1, 65103, '\P{^Is_Block=_cjk_Compatibility_Forms}', "");
Expect(0, 65104, '\p{Is_Block=_cjk_Compatibility_Forms}', "");
Expect(1, 65104, '\p{^Is_Block=_cjk_Compatibility_Forms}', "");
Expect(1, 65104, '\P{Is_Block=_cjk_Compatibility_Forms}', "");
Expect(0, 65104, '\P{^Is_Block=_cjk_Compatibility_Forms}', "");
Error('\p{Is_Blk=_/a/cjk_COMPAT_Forms}');
Error('\P{Is_Blk=_/a/cjk_COMPAT_Forms}');
Expect(1, 65103, '\p{Is_Blk=cjkcompatforms}', "");
Expect(0, 65103, '\p{^Is_Blk=cjkcompatforms}', "");
Expect(0, 65103, '\P{Is_Blk=cjkcompatforms}', "");
Expect(1, 65103, '\P{^Is_Blk=cjkcompatforms}', "");
Expect(0, 65104, '\p{Is_Blk=cjkcompatforms}', "");
Expect(1, 65104, '\p{^Is_Blk=cjkcompatforms}', "");
Expect(1, 65104, '\P{Is_Blk=cjkcompatforms}', "");
Expect(0, 65104, '\P{^Is_Blk=cjkcompatforms}', "");
Expect(1, 65103, '\p{Is_Blk= 	CJK_Compat_FORMS}', "");
Expect(0, 65103, '\p{^Is_Blk= 	CJK_Compat_FORMS}', "");
Expect(0, 65103, '\P{Is_Blk= 	CJK_Compat_FORMS}', "");
Expect(1, 65103, '\P{^Is_Blk= 	CJK_Compat_FORMS}', "");
Expect(0, 65104, '\p{Is_Blk= 	CJK_Compat_FORMS}', "");
Expect(1, 65104, '\p{^Is_Blk= 	CJK_Compat_FORMS}', "");
Expect(1, 65104, '\P{Is_Blk= 	CJK_Compat_FORMS}', "");
Expect(0, 65104, '\P{^Is_Blk= 	CJK_Compat_FORMS}', "");
Error('\p{Block=:=_CJK_Compatibility_ideographs}');
Error('\P{Block=:=_CJK_Compatibility_ideographs}');
Expect(1, 64255, '\p{Block=cjkcompatibilityideographs}', "");
Expect(0, 64255, '\p{^Block=cjkcompatibilityideographs}', "");
Expect(0, 64255, '\P{Block=cjkcompatibilityideographs}', "");
Expect(1, 64255, '\P{^Block=cjkcompatibilityideographs}', "");
Expect(0, 64256, '\p{Block=cjkcompatibilityideographs}', "");
Expect(1, 64256, '\p{^Block=cjkcompatibilityideographs}', "");
Expect(1, 64256, '\P{Block=cjkcompatibilityideographs}', "");
Expect(0, 64256, '\P{^Block=cjkcompatibilityideographs}', "");
Expect(1, 64255, '\p{Block= 	cjk_Compatibility_IDEOGRAPHS}', "");
Expect(0, 64255, '\p{^Block= 	cjk_Compatibility_IDEOGRAPHS}', "");
Expect(0, 64255, '\P{Block= 	cjk_Compatibility_IDEOGRAPHS}', "");
Expect(1, 64255, '\P{^Block= 	cjk_Compatibility_IDEOGRAPHS}', "");
Expect(0, 64256, '\p{Block= 	cjk_Compatibility_IDEOGRAPHS}', "");
Expect(1, 64256, '\p{^Block= 	cjk_Compatibility_IDEOGRAPHS}', "");
Expect(1, 64256, '\P{Block= 	cjk_Compatibility_IDEOGRAPHS}', "");
Expect(0, 64256, '\P{^Block= 	cjk_Compatibility_IDEOGRAPHS}', "");
Error('\p{Blk=_:=CJK_compat_ideographs}');
Error('\P{Blk=_:=CJK_compat_ideographs}');
Expect(1, 64255, '\p{Blk=cjkcompatideographs}', "");
Expect(0, 64255, '\p{^Blk=cjkcompatideographs}', "");
Expect(0, 64255, '\P{Blk=cjkcompatideographs}', "");
Expect(1, 64255, '\P{^Blk=cjkcompatideographs}', "");
Expect(0, 64256, '\p{Blk=cjkcompatideographs}', "");
Expect(1, 64256, '\p{^Blk=cjkcompatideographs}', "");
Expect(1, 64256, '\P{Blk=cjkcompatideographs}', "");
Expect(0, 64256, '\P{^Blk=cjkcompatideographs}', "");
Expect(1, 64255, '\p{Blk=--CJK_compat_Ideographs}', "");
Expect(0, 64255, '\p{^Blk=--CJK_compat_Ideographs}', "");
Expect(0, 64255, '\P{Blk=--CJK_compat_Ideographs}', "");
Expect(1, 64255, '\P{^Blk=--CJK_compat_Ideographs}', "");
Expect(0, 64256, '\p{Blk=--CJK_compat_Ideographs}', "");
Expect(1, 64256, '\p{^Blk=--CJK_compat_Ideographs}', "");
Expect(1, 64256, '\P{Blk=--CJK_compat_Ideographs}', "");
Expect(0, 64256, '\P{^Blk=--CJK_compat_Ideographs}', "");
Error('\p{Is_Block= -CJK_Compatibility_ideographs/a/}');
Error('\P{Is_Block= -CJK_Compatibility_ideographs/a/}');
Expect(1, 64255, '\p{Is_Block=cjkcompatibilityideographs}', "");
Expect(0, 64255, '\p{^Is_Block=cjkcompatibilityideographs}', "");
Expect(0, 64255, '\P{Is_Block=cjkcompatibilityideographs}', "");
Expect(1, 64255, '\P{^Is_Block=cjkcompatibilityideographs}', "");
Expect(0, 64256, '\p{Is_Block=cjkcompatibilityideographs}', "");
Expect(1, 64256, '\p{^Is_Block=cjkcompatibilityideographs}', "");
Expect(1, 64256, '\P{Is_Block=cjkcompatibilityideographs}', "");
Expect(0, 64256, '\P{^Is_Block=cjkcompatibilityideographs}', "");
Expect(1, 64255, '\p{Is_Block=--cjk_Compatibility_ideographs}', "");
Expect(0, 64255, '\p{^Is_Block=--cjk_Compatibility_ideographs}', "");
Expect(0, 64255, '\P{Is_Block=--cjk_Compatibility_ideographs}', "");
Expect(1, 64255, '\P{^Is_Block=--cjk_Compatibility_ideographs}', "");
Expect(0, 64256, '\p{Is_Block=--cjk_Compatibility_ideographs}', "");
Expect(1, 64256, '\p{^Is_Block=--cjk_Compatibility_ideographs}', "");
Expect(1, 64256, '\P{Is_Block=--cjk_Compatibility_ideographs}', "");
Expect(0, 64256, '\P{^Is_Block=--cjk_Compatibility_ideographs}', "");
Error('\p{Is_Blk= _CJK_COMPAT_Ideographs:=}');
Error('\P{Is_Blk= _CJK_COMPAT_Ideographs:=}');
Expect(1, 64255, '\p{Is_Blk=cjkcompatideographs}', "");
Expect(0, 64255, '\p{^Is_Blk=cjkcompatideographs}', "");
Expect(0, 64255, '\P{Is_Blk=cjkcompatideographs}', "");
Expect(1, 64255, '\P{^Is_Blk=cjkcompatideographs}', "");
Expect(0, 64256, '\p{Is_Blk=cjkcompatideographs}', "");
Expect(1, 64256, '\p{^Is_Blk=cjkcompatideographs}', "");
Expect(1, 64256, '\P{Is_Blk=cjkcompatideographs}', "");
Expect(0, 64256, '\P{^Is_Blk=cjkcompatideographs}', "");
Expect(1, 64255, '\p{Is_Blk=__CJK_Compat_Ideographs}', "");
Expect(0, 64255, '\p{^Is_Blk=__CJK_Compat_Ideographs}', "");
Expect(0, 64255, '\P{Is_Blk=__CJK_Compat_Ideographs}', "");
Expect(1, 64255, '\P{^Is_Blk=__CJK_Compat_Ideographs}', "");
Expect(0, 64256, '\p{Is_Blk=__CJK_Compat_Ideographs}', "");
Expect(1, 64256, '\p{^Is_Blk=__CJK_Compat_Ideographs}', "");
Expect(1, 64256, '\P{Is_Blk=__CJK_Compat_Ideographs}', "");
Expect(0, 64256, '\P{^Is_Blk=__CJK_Compat_Ideographs}', "");
Error('\p{Block:_:=CJK_Compatibility_Ideographs_Supplement}');
Error('\P{Block:_:=CJK_Compatibility_Ideographs_Supplement}');
Expect(1, 195103, '\p{Block=cjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\p{^Block=cjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\P{Block=cjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\P{^Block=cjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\p{Block=cjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\p{^Block=cjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\P{Block=cjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\P{^Block=cjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\p{Block:    CJK_Compatibility_Ideographs_SUPPLEMENT}', "");
Expect(0, 195103, '\p{^Block:    CJK_Compatibility_Ideographs_SUPPLEMENT}', "");
Expect(0, 195103, '\P{Block:    CJK_Compatibility_Ideographs_SUPPLEMENT}', "");
Expect(1, 195103, '\P{^Block:    CJK_Compatibility_Ideographs_SUPPLEMENT}', "");
Expect(0, 195104, '\p{Block:    CJK_Compatibility_Ideographs_SUPPLEMENT}', "");
Expect(1, 195104, '\p{^Block:    CJK_Compatibility_Ideographs_SUPPLEMENT}', "");
Expect(1, 195104, '\P{Block:    CJK_Compatibility_Ideographs_SUPPLEMENT}', "");
Expect(0, 195104, '\P{^Block:    CJK_Compatibility_Ideographs_SUPPLEMENT}', "");
Error('\p{Blk=:= _CJK_Compat_IDEOGRAPHS_sup}');
Error('\P{Blk=:= _CJK_Compat_IDEOGRAPHS_sup}');
Expect(1, 195103, '\p{Blk=cjkcompatideographssup}', "");
Expect(0, 195103, '\p{^Blk=cjkcompatideographssup}', "");
Expect(0, 195103, '\P{Blk=cjkcompatideographssup}', "");
Expect(1, 195103, '\P{^Blk=cjkcompatideographssup}', "");
Expect(0, 195104, '\p{Blk=cjkcompatideographssup}', "");
Expect(1, 195104, '\p{^Blk=cjkcompatideographssup}', "");
Expect(1, 195104, '\P{Blk=cjkcompatideographssup}', "");
Expect(0, 195104, '\P{^Blk=cjkcompatideographssup}', "");
Expect(1, 195103, '\p{Blk=_-CJK_compat_ideographs_Sup}', "");
Expect(0, 195103, '\p{^Blk=_-CJK_compat_ideographs_Sup}', "");
Expect(0, 195103, '\P{Blk=_-CJK_compat_ideographs_Sup}', "");
Expect(1, 195103, '\P{^Blk=_-CJK_compat_ideographs_Sup}', "");
Expect(0, 195104, '\p{Blk=_-CJK_compat_ideographs_Sup}', "");
Expect(1, 195104, '\p{^Blk=_-CJK_compat_ideographs_Sup}', "");
Expect(1, 195104, '\P{Blk=_-CJK_compat_ideographs_Sup}', "");
Expect(0, 195104, '\P{^Blk=_-CJK_compat_ideographs_Sup}', "");
Error('\p{Is_Block=		cjk_compatibility_ideographs_supplement/a/}');
Error('\P{Is_Block=		cjk_compatibility_ideographs_supplement/a/}');
Expect(1, 195103, '\p{Is_Block=cjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\p{^Is_Block=cjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\P{Is_Block=cjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\P{^Is_Block=cjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\p{Is_Block=cjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\p{^Is_Block=cjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\P{Is_Block=cjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\P{^Is_Block=cjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\p{Is_Block=		CJK_Compatibility_ideographs_supplement}', "");
Expect(0, 195103, '\p{^Is_Block=		CJK_Compatibility_ideographs_supplement}', "");
Expect(0, 195103, '\P{Is_Block=		CJK_Compatibility_ideographs_supplement}', "");
Expect(1, 195103, '\P{^Is_Block=		CJK_Compatibility_ideographs_supplement}', "");
Expect(0, 195104, '\p{Is_Block=		CJK_Compatibility_ideographs_supplement}', "");
Expect(1, 195104, '\p{^Is_Block=		CJK_Compatibility_ideographs_supplement}', "");
Expect(1, 195104, '\P{Is_Block=		CJK_Compatibility_ideographs_supplement}', "");
Expect(0, 195104, '\P{^Is_Block=		CJK_Compatibility_ideographs_supplement}', "");
Error('\p{Is_Blk:/a/ -CJK_Compat_ideographs_Sup}');
Error('\P{Is_Blk:/a/ -CJK_Compat_ideographs_Sup}');
Expect(1, 195103, '\p{Is_Blk=cjkcompatideographssup}', "");
Expect(0, 195103, '\p{^Is_Blk=cjkcompatideographssup}', "");
Expect(0, 195103, '\P{Is_Blk=cjkcompatideographssup}', "");
Expect(1, 195103, '\P{^Is_Blk=cjkcompatideographssup}', "");
Expect(0, 195104, '\p{Is_Blk=cjkcompatideographssup}', "");
Expect(1, 195104, '\p{^Is_Blk=cjkcompatideographssup}', "");
Expect(1, 195104, '\P{Is_Blk=cjkcompatideographssup}', "");
Expect(0, 195104, '\P{^Is_Blk=cjkcompatideographssup}', "");
Expect(1, 195103, '\p{Is_Blk=	CJK_compat_Ideographs_sup}', "");
Expect(0, 195103, '\p{^Is_Blk=	CJK_compat_Ideographs_sup}', "");
Expect(0, 195103, '\P{Is_Blk=	CJK_compat_Ideographs_sup}', "");
Expect(1, 195103, '\P{^Is_Blk=	CJK_compat_Ideographs_sup}', "");
Expect(0, 195104, '\p{Is_Blk=	CJK_compat_Ideographs_sup}', "");
Expect(1, 195104, '\p{^Is_Blk=	CJK_compat_Ideographs_sup}', "");
Expect(1, 195104, '\P{Is_Blk=	CJK_compat_Ideographs_sup}', "");
Expect(0, 195104, '\P{^Is_Blk=	CJK_compat_Ideographs_sup}', "");
Error('\p{Block= -CJK_unified_ideographs_Extension_A:=}');
Error('\P{Block= -CJK_unified_ideographs_Extension_A:=}');
Expect(1, 19903, '\p{Block=cjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\p{^Block=cjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\P{Block=cjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\P{^Block=cjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\p{Block=cjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\p{^Block=cjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\P{Block=cjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\P{^Block=cjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\p{Block=-_CJK_unified_ideographs_extension_A}', "");
Expect(0, 19903, '\p{^Block=-_CJK_unified_ideographs_extension_A}', "");
Expect(0, 19903, '\P{Block=-_CJK_unified_ideographs_extension_A}', "");
Expect(1, 19903, '\P{^Block=-_CJK_unified_ideographs_extension_A}', "");
Expect(0, 19904, '\p{Block=-_CJK_unified_ideographs_extension_A}', "");
Expect(1, 19904, '\p{^Block=-_CJK_unified_ideographs_extension_A}', "");
Expect(1, 19904, '\P{Block=-_CJK_unified_ideographs_extension_A}', "");
Expect(0, 19904, '\P{^Block=-_CJK_unified_ideographs_extension_A}', "");
Error('\p{Blk= :=CJK_ext_A}');
Error('\P{Blk= :=CJK_ext_A}');
Expect(1, 19903, '\p{Blk=cjkexta}', "");
Expect(0, 19903, '\p{^Blk=cjkexta}', "");
Expect(0, 19903, '\P{Blk=cjkexta}', "");
Expect(1, 19903, '\P{^Blk=cjkexta}', "");
Expect(0, 19904, '\p{Blk=cjkexta}', "");
Expect(1, 19904, '\p{^Blk=cjkexta}', "");
Expect(1, 19904, '\P{Blk=cjkexta}', "");
Expect(0, 19904, '\P{^Blk=cjkexta}', "");
Expect(1, 19903, '\p{Blk=cjk_Ext_a}', "");
Expect(0, 19903, '\p{^Blk=cjk_Ext_a}', "");
Expect(0, 19903, '\P{Blk=cjk_Ext_a}', "");
Expect(1, 19903, '\P{^Blk=cjk_Ext_a}', "");
Expect(0, 19904, '\p{Blk=cjk_Ext_a}', "");
Expect(1, 19904, '\p{^Blk=cjk_Ext_a}', "");
Expect(1, 19904, '\P{Blk=cjk_Ext_a}', "");
Expect(0, 19904, '\P{^Blk=cjk_Ext_a}', "");
Error('\p{Is_Block:	:= CJK_unified_Ideographs_extension_A}');
Error('\P{Is_Block:	:= CJK_unified_Ideographs_extension_A}');
Expect(1, 19903, '\p{Is_Block=cjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\p{^Is_Block=cjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\P{Is_Block=cjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\P{^Is_Block=cjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\p{Is_Block=cjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\p{^Is_Block=cjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\P{Is_Block=cjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\P{^Is_Block=cjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\p{Is_Block= -CJK_unified_Ideographs_extension_A}', "");
Expect(0, 19903, '\p{^Is_Block= -CJK_unified_Ideographs_extension_A}', "");
Expect(0, 19903, '\P{Is_Block= -CJK_unified_Ideographs_extension_A}', "");
Expect(1, 19903, '\P{^Is_Block= -CJK_unified_Ideographs_extension_A}', "");
Expect(0, 19904, '\p{Is_Block= -CJK_unified_Ideographs_extension_A}', "");
Expect(1, 19904, '\p{^Is_Block= -CJK_unified_Ideographs_extension_A}', "");
Expect(1, 19904, '\P{Is_Block= -CJK_unified_Ideographs_extension_A}', "");
Expect(0, 19904, '\P{^Is_Block= -CJK_unified_Ideographs_extension_A}', "");
Error('\p{Is_Blk=	:=CJK_ext_A}');
Error('\P{Is_Blk=	:=CJK_ext_A}');
Expect(1, 19903, '\p{Is_Blk=cjkexta}', "");
Expect(0, 19903, '\p{^Is_Blk=cjkexta}', "");
Expect(0, 19903, '\P{Is_Blk=cjkexta}', "");
Expect(1, 19903, '\P{^Is_Blk=cjkexta}', "");
Expect(0, 19904, '\p{Is_Blk=cjkexta}', "");
Expect(1, 19904, '\p{^Is_Blk=cjkexta}', "");
Expect(1, 19904, '\P{Is_Blk=cjkexta}', "");
Expect(0, 19904, '\P{^Is_Blk=cjkexta}', "");
Expect(1, 19903, '\p{Is_Blk=_CJK_ext_A}', "");
Expect(0, 19903, '\p{^Is_Blk=_CJK_ext_A}', "");
Expect(0, 19903, '\P{Is_Blk=_CJK_ext_A}', "");
Expect(1, 19903, '\P{^Is_Blk=_CJK_ext_A}', "");
Expect(0, 19904, '\p{Is_Blk=_CJK_ext_A}', "");
Expect(1, 19904, '\p{^Is_Blk=_CJK_ext_A}', "");
Expect(1, 19904, '\P{Is_Blk=_CJK_ext_A}', "");
Expect(0, 19904, '\P{^Is_Blk=_CJK_ext_A}', "");
Error('\p{Block:   -CJK_unified_ideographs_Extension_B:=}');
Error('\P{Block:   -CJK_unified_ideographs_Extension_B:=}');
Expect(1, 173791, '\p{Block=cjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\p{^Block=cjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\P{Block=cjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\P{^Block=cjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\p{Block=cjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\p{^Block=cjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\P{Block=cjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\P{^Block=cjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\p{Block=-	CJK_Unified_Ideographs_extension_B}', "");
Expect(0, 173791, '\p{^Block=-	CJK_Unified_Ideographs_extension_B}', "");
Expect(0, 173791, '\P{Block=-	CJK_Unified_Ideographs_extension_B}', "");
Expect(1, 173791, '\P{^Block=-	CJK_Unified_Ideographs_extension_B}', "");
Expect(0, 173792, '\p{Block=-	CJK_Unified_Ideographs_extension_B}', "");
Expect(1, 173792, '\p{^Block=-	CJK_Unified_Ideographs_extension_B}', "");
Expect(1, 173792, '\P{Block=-	CJK_Unified_Ideographs_extension_B}', "");
Expect(0, 173792, '\P{^Block=-	CJK_Unified_Ideographs_extension_B}', "");
Error('\p{Blk=/a/ 	cjk_EXT_b}');
Error('\P{Blk=/a/ 	cjk_EXT_b}');
Expect(1, 173791, '\p{Blk=cjkextb}', "");
Expect(0, 173791, '\p{^Blk=cjkextb}', "");
Expect(0, 173791, '\P{Blk=cjkextb}', "");
Expect(1, 173791, '\P{^Blk=cjkextb}', "");
Expect(0, 173792, '\p{Blk=cjkextb}', "");
Expect(1, 173792, '\p{^Blk=cjkextb}', "");
Expect(1, 173792, '\P{Blk=cjkextb}', "");
Expect(0, 173792, '\P{^Blk=cjkextb}', "");
Expect(1, 173791, '\p{Blk=_ CJK_ext_b}', "");
Expect(0, 173791, '\p{^Blk=_ CJK_ext_b}', "");
Expect(0, 173791, '\P{Blk=_ CJK_ext_b}', "");
Expect(1, 173791, '\P{^Blk=_ CJK_ext_b}', "");
Expect(0, 173792, '\p{Blk=_ CJK_ext_b}', "");
Expect(1, 173792, '\p{^Blk=_ CJK_ext_b}', "");
Expect(1, 173792, '\P{Blk=_ CJK_ext_b}', "");
Expect(0, 173792, '\P{^Blk=_ CJK_ext_b}', "");
Error('\p{Is_Block=		CJK_Unified_Ideographs_Extension_b:=}');
Error('\P{Is_Block=		CJK_Unified_Ideographs_Extension_b:=}');
Expect(1, 173791, '\p{Is_Block=cjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\p{^Is_Block=cjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\P{Is_Block=cjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\P{^Is_Block=cjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\p{Is_Block=cjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\p{^Is_Block=cjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\P{Is_Block=cjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\P{^Is_Block=cjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\p{Is_Block=  cjk_Unified_Ideographs_Extension_B}', "");
Expect(0, 173791, '\p{^Is_Block=  cjk_Unified_Ideographs_Extension_B}', "");
Expect(0, 173791, '\P{Is_Block=  cjk_Unified_Ideographs_Extension_B}', "");
Expect(1, 173791, '\P{^Is_Block=  cjk_Unified_Ideographs_Extension_B}', "");
Expect(0, 173792, '\p{Is_Block=  cjk_Unified_Ideographs_Extension_B}', "");
Expect(1, 173792, '\p{^Is_Block=  cjk_Unified_Ideographs_Extension_B}', "");
Expect(1, 173792, '\P{Is_Block=  cjk_Unified_Ideographs_Extension_B}', "");
Expect(0, 173792, '\P{^Is_Block=  cjk_Unified_Ideographs_Extension_B}', "");
Error('\p{Is_Blk=/a/	CJK_Ext_B}');
Error('\P{Is_Blk=/a/	CJK_Ext_B}');
Expect(1, 173791, '\p{Is_Blk=cjkextb}', "");
Expect(0, 173791, '\p{^Is_Blk=cjkextb}', "");
Expect(0, 173791, '\P{Is_Blk=cjkextb}', "");
Expect(1, 173791, '\P{^Is_Blk=cjkextb}', "");
Expect(0, 173792, '\p{Is_Blk=cjkextb}', "");
Expect(1, 173792, '\p{^Is_Blk=cjkextb}', "");
Expect(1, 173792, '\P{Is_Blk=cjkextb}', "");
Expect(0, 173792, '\P{^Is_Blk=cjkextb}', "");
Expect(1, 173791, '\p{Is_Blk= 	cjk_Ext_b}', "");
Expect(0, 173791, '\p{^Is_Blk= 	cjk_Ext_b}', "");
Expect(0, 173791, '\P{Is_Blk= 	cjk_Ext_b}', "");
Expect(1, 173791, '\P{^Is_Blk= 	cjk_Ext_b}', "");
Expect(0, 173792, '\p{Is_Blk= 	cjk_Ext_b}', "");
Expect(1, 173792, '\p{^Is_Blk= 	cjk_Ext_b}', "");
Expect(1, 173792, '\P{Is_Blk= 	cjk_Ext_b}', "");
Expect(0, 173792, '\P{^Is_Blk= 	cjk_Ext_b}', "");
Error('\p{Block:   :=	_cjk_UNIFIED_IDEOGRAPHS_Extension_C}');
Error('\P{Block:   :=	_cjk_UNIFIED_IDEOGRAPHS_Extension_C}');
Expect(1, 177983, '\p{Block=cjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\p{^Block=cjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\P{Block=cjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\P{^Block=cjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\p{Block=cjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\p{^Block=cjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\P{Block=cjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\P{^Block=cjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\p{Block=--CJK_Unified_Ideographs_Extension_C}', "");
Expect(0, 177983, '\p{^Block=--CJK_Unified_Ideographs_Extension_C}', "");
Expect(0, 177983, '\P{Block=--CJK_Unified_Ideographs_Extension_C}', "");
Expect(1, 177983, '\P{^Block=--CJK_Unified_Ideographs_Extension_C}', "");
Expect(0, 177984, '\p{Block=--CJK_Unified_Ideographs_Extension_C}', "");
Expect(1, 177984, '\p{^Block=--CJK_Unified_Ideographs_Extension_C}', "");
Expect(1, 177984, '\P{Block=--CJK_Unified_Ideographs_Extension_C}', "");
Expect(0, 177984, '\P{^Block=--CJK_Unified_Ideographs_Extension_C}', "");
Error('\p{Blk=-CJK_EXT_c:=}');
Error('\P{Blk=-CJK_EXT_c:=}');
Expect(1, 177983, '\p{Blk=cjkextc}', "");
Expect(0, 177983, '\p{^Blk=cjkextc}', "");
Expect(0, 177983, '\P{Blk=cjkextc}', "");
Expect(1, 177983, '\P{^Blk=cjkextc}', "");
Expect(0, 177984, '\p{Blk=cjkextc}', "");
Expect(1, 177984, '\p{^Blk=cjkextc}', "");
Expect(1, 177984, '\P{Blk=cjkextc}', "");
Expect(0, 177984, '\P{^Blk=cjkextc}', "");
Expect(1, 177983, '\p{Blk:		CJK_Ext_C}', "");
Expect(0, 177983, '\p{^Blk:		CJK_Ext_C}', "");
Expect(0, 177983, '\P{Blk:		CJK_Ext_C}', "");
Expect(1, 177983, '\P{^Blk:		CJK_Ext_C}', "");
Expect(0, 177984, '\p{Blk:		CJK_Ext_C}', "");
Expect(1, 177984, '\p{^Blk:		CJK_Ext_C}', "");
Expect(1, 177984, '\P{Blk:		CJK_Ext_C}', "");
Expect(0, 177984, '\P{^Blk:		CJK_Ext_C}', "");
Error('\p{Is_Block=- CJK_Unified_ideographs_Extension_C/a/}');
Error('\P{Is_Block=- CJK_Unified_ideographs_Extension_C/a/}');
Expect(1, 177983, '\p{Is_Block=cjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\p{^Is_Block=cjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\P{Is_Block=cjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\P{^Is_Block=cjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\p{Is_Block=cjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\p{^Is_Block=cjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\P{Is_Block=cjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\P{^Is_Block=cjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\p{Is_Block:    CJK_UNIFIED_IDEOGRAPHS_EXTENSION_C}', "");
Expect(0, 177983, '\p{^Is_Block:    CJK_UNIFIED_IDEOGRAPHS_EXTENSION_C}', "");
Expect(0, 177983, '\P{Is_Block:    CJK_UNIFIED_IDEOGRAPHS_EXTENSION_C}', "");
Expect(1, 177983, '\P{^Is_Block:    CJK_UNIFIED_IDEOGRAPHS_EXTENSION_C}', "");
Expect(0, 177984, '\p{Is_Block:    CJK_UNIFIED_IDEOGRAPHS_EXTENSION_C}', "");
Expect(1, 177984, '\p{^Is_Block:    CJK_UNIFIED_IDEOGRAPHS_EXTENSION_C}', "");
Expect(1, 177984, '\P{Is_Block:    CJK_UNIFIED_IDEOGRAPHS_EXTENSION_C}', "");
Expect(0, 177984, '\P{^Is_Block:    CJK_UNIFIED_IDEOGRAPHS_EXTENSION_C}', "");
Error('\p{Is_Blk=		CJK_Ext_c:=}');
Error('\P{Is_Blk=		CJK_Ext_c:=}');
Expect(1, 177983, '\p{Is_Blk=cjkextc}', "");
Expect(0, 177983, '\p{^Is_Blk=cjkextc}', "");
Expect(0, 177983, '\P{Is_Blk=cjkextc}', "");
Expect(1, 177983, '\P{^Is_Blk=cjkextc}', "");
Expect(0, 177984, '\p{Is_Blk=cjkextc}', "");
Expect(1, 177984, '\p{^Is_Blk=cjkextc}', "");
Expect(1, 177984, '\P{Is_Blk=cjkextc}', "");
Expect(0, 177984, '\P{^Is_Blk=cjkextc}', "");
Expect(1, 177983, '\p{Is_Blk=	cjk_ext_c}', "");
Expect(0, 177983, '\p{^Is_Blk=	cjk_ext_c}', "");
Expect(0, 177983, '\P{Is_Blk=	cjk_ext_c}', "");
Expect(1, 177983, '\P{^Is_Blk=	cjk_ext_c}', "");
Expect(0, 177984, '\p{Is_Blk=	cjk_ext_c}', "");
Expect(1, 177984, '\p{^Is_Blk=	cjk_ext_c}', "");
Expect(1, 177984, '\P{Is_Blk=	cjk_ext_c}', "");
Expect(0, 177984, '\P{^Is_Blk=	cjk_ext_c}', "");
Error('\p{Block=-/a/CJK_Unified_IDEOGRAPHS_Extension_D}');
Error('\P{Block=-/a/CJK_Unified_IDEOGRAPHS_Extension_D}');
Expect(1, 178207, '\p{Block=cjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\p{^Block=cjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\P{Block=cjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\P{^Block=cjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\p{Block=cjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\p{^Block=cjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\P{Block=cjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\P{^Block=cjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\p{Block=	_CJK_UNIFIED_Ideographs_extension_D}', "");
Expect(0, 178207, '\p{^Block=	_CJK_UNIFIED_Ideographs_extension_D}', "");
Expect(0, 178207, '\P{Block=	_CJK_UNIFIED_Ideographs_extension_D}', "");
Expect(1, 178207, '\P{^Block=	_CJK_UNIFIED_Ideographs_extension_D}', "");
Expect(0, 178208, '\p{Block=	_CJK_UNIFIED_Ideographs_extension_D}', "");
Expect(1, 178208, '\p{^Block=	_CJK_UNIFIED_Ideographs_extension_D}', "");
Expect(1, 178208, '\P{Block=	_CJK_UNIFIED_Ideographs_extension_D}', "");
Expect(0, 178208, '\P{^Block=	_CJK_UNIFIED_Ideographs_extension_D}', "");
Error('\p{Blk=__CJK_EXT_d:=}');
Error('\P{Blk=__CJK_EXT_d:=}');
Expect(1, 178207, '\p{Blk=cjkextd}', "");
Expect(0, 178207, '\p{^Blk=cjkextd}', "");
Expect(0, 178207, '\P{Blk=cjkextd}', "");
Expect(1, 178207, '\P{^Blk=cjkextd}', "");
Expect(0, 178208, '\p{Blk=cjkextd}', "");
Expect(1, 178208, '\p{^Blk=cjkextd}', "");
Expect(1, 178208, '\P{Blk=cjkextd}', "");
Expect(0, 178208, '\P{^Blk=cjkextd}', "");
Expect(1, 178207, '\p{Blk=--CJK_Ext_D}', "");
Expect(0, 178207, '\p{^Blk=--CJK_Ext_D}', "");
Expect(0, 178207, '\P{Blk=--CJK_Ext_D}', "");
Expect(1, 178207, '\P{^Blk=--CJK_Ext_D}', "");
Expect(0, 178208, '\p{Blk=--CJK_Ext_D}', "");
Expect(1, 178208, '\p{^Blk=--CJK_Ext_D}', "");
Expect(1, 178208, '\P{Blk=--CJK_Ext_D}', "");
Expect(0, 178208, '\P{^Blk=--CJK_Ext_D}', "");
Error('\p{Is_Block:	:=CJK_unified_Ideographs_Extension_D}');
Error('\P{Is_Block:	:=CJK_unified_Ideographs_Extension_D}');
Expect(1, 178207, '\p{Is_Block=cjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\p{^Is_Block=cjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\P{Is_Block=cjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\P{^Is_Block=cjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\p{Is_Block=cjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\p{^Is_Block=cjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\P{Is_Block=cjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\P{^Is_Block=cjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\p{Is_Block=CJK_Unified_Ideographs_Extension_D}', "");
Expect(0, 178207, '\p{^Is_Block=CJK_Unified_Ideographs_Extension_D}', "");
Expect(0, 178207, '\P{Is_Block=CJK_Unified_Ideographs_Extension_D}', "");
Expect(1, 178207, '\P{^Is_Block=CJK_Unified_Ideographs_Extension_D}', "");
Expect(0, 178208, '\p{Is_Block=CJK_Unified_Ideographs_Extension_D}', "");
Expect(1, 178208, '\p{^Is_Block=CJK_Unified_Ideographs_Extension_D}', "");
Expect(1, 178208, '\P{Is_Block=CJK_Unified_Ideographs_Extension_D}', "");
Expect(0, 178208, '\P{^Is_Block=CJK_Unified_Ideographs_Extension_D}', "");
Error('\p{Is_Blk=:= -CJK_Ext_D}');
Error('\P{Is_Blk=:= -CJK_Ext_D}');
Expect(1, 178207, '\p{Is_Blk=cjkextd}', "");
Expect(0, 178207, '\p{^Is_Blk=cjkextd}', "");
Expect(0, 178207, '\P{Is_Blk=cjkextd}', "");
Expect(1, 178207, '\P{^Is_Blk=cjkextd}', "");
Expect(0, 178208, '\p{Is_Blk=cjkextd}', "");
Expect(1, 178208, '\p{^Is_Blk=cjkextd}', "");
Expect(1, 178208, '\P{Is_Blk=cjkextd}', "");
Expect(0, 178208, '\P{^Is_Blk=cjkextd}', "");
Expect(1, 178207, '\p{Is_Blk=  CJK_Ext_D}', "");
Expect(0, 178207, '\p{^Is_Blk=  CJK_Ext_D}', "");
Expect(0, 178207, '\P{Is_Blk=  CJK_Ext_D}', "");
Expect(1, 178207, '\P{^Is_Blk=  CJK_Ext_D}', "");
Expect(0, 178208, '\p{Is_Blk=  CJK_Ext_D}', "");
Expect(1, 178208, '\p{^Is_Blk=  CJK_Ext_D}', "");
Expect(1, 178208, '\P{Is_Blk=  CJK_Ext_D}', "");
Expect(0, 178208, '\P{^Is_Blk=  CJK_Ext_D}', "");
Error('\p{Block=cjk_RADICALS_Supplement/a/}');
Error('\P{Block=cjk_RADICALS_Supplement/a/}');
Expect(1, 12031, '\p{Block=cjkradicalssupplement}', "");
Expect(0, 12031, '\p{^Block=cjkradicalssupplement}', "");
Expect(0, 12031, '\P{Block=cjkradicalssupplement}', "");
Expect(1, 12031, '\P{^Block=cjkradicalssupplement}', "");
Expect(0, 12032, '\p{Block=cjkradicalssupplement}', "");
Expect(1, 12032, '\p{^Block=cjkradicalssupplement}', "");
Expect(1, 12032, '\P{Block=cjkradicalssupplement}', "");
Expect(0, 12032, '\P{^Block=cjkradicalssupplement}', "");
Expect(1, 12031, '\p{Block=--cjk_Radicals_Supplement}', "");
Expect(0, 12031, '\p{^Block=--cjk_Radicals_Supplement}', "");
Expect(0, 12031, '\P{Block=--cjk_Radicals_Supplement}', "");
Expect(1, 12031, '\P{^Block=--cjk_Radicals_Supplement}', "");
Expect(0, 12032, '\p{Block=--cjk_Radicals_Supplement}', "");
Expect(1, 12032, '\p{^Block=--cjk_Radicals_Supplement}', "");
Expect(1, 12032, '\P{Block=--cjk_Radicals_Supplement}', "");
Expect(0, 12032, '\P{^Block=--cjk_Radicals_Supplement}', "");
Error('\p{Blk=	:=CJK_RADICALS_Sup}');
Error('\P{Blk=	:=CJK_RADICALS_Sup}');
Expect(1, 12031, '\p{Blk=cjkradicalssup}', "");
Expect(0, 12031, '\p{^Blk=cjkradicalssup}', "");
Expect(0, 12031, '\P{Blk=cjkradicalssup}', "");
Expect(1, 12031, '\P{^Blk=cjkradicalssup}', "");
Expect(0, 12032, '\p{Blk=cjkradicalssup}', "");
Expect(1, 12032, '\p{^Blk=cjkradicalssup}', "");
Expect(1, 12032, '\P{Blk=cjkradicalssup}', "");
Expect(0, 12032, '\P{^Blk=cjkradicalssup}', "");
Expect(1, 12031, '\p{Blk=  CJK_radicals_sup}', "");
Expect(0, 12031, '\p{^Blk=  CJK_radicals_sup}', "");
Expect(0, 12031, '\P{Blk=  CJK_radicals_sup}', "");
Expect(1, 12031, '\P{^Blk=  CJK_radicals_sup}', "");
Expect(0, 12032, '\p{Blk=  CJK_radicals_sup}', "");
Expect(1, 12032, '\p{^Blk=  CJK_radicals_sup}', "");
Expect(1, 12032, '\P{Blk=  CJK_radicals_sup}', "");
Expect(0, 12032, '\P{^Blk=  CJK_radicals_sup}', "");
Error('\p{Is_Block=:=	CJK_radicals_SUPPLEMENT}');
Error('\P{Is_Block=:=	CJK_radicals_SUPPLEMENT}');
Expect(1, 12031, '\p{Is_Block=cjkradicalssupplement}', "");
Expect(0, 12031, '\p{^Is_Block=cjkradicalssupplement}', "");
Expect(0, 12031, '\P{Is_Block=cjkradicalssupplement}', "");
Expect(1, 12031, '\P{^Is_Block=cjkradicalssupplement}', "");
Expect(0, 12032, '\p{Is_Block=cjkradicalssupplement}', "");
Expect(1, 12032, '\p{^Is_Block=cjkradicalssupplement}', "");
Expect(1, 12032, '\P{Is_Block=cjkradicalssupplement}', "");
Expect(0, 12032, '\P{^Is_Block=cjkradicalssupplement}', "");
Expect(1, 12031, '\p{Is_Block=	CJK_Radicals_Supplement}', "");
Expect(0, 12031, '\p{^Is_Block=	CJK_Radicals_Supplement}', "");
Expect(0, 12031, '\P{Is_Block=	CJK_Radicals_Supplement}', "");
Expect(1, 12031, '\P{^Is_Block=	CJK_Radicals_Supplement}', "");
Expect(0, 12032, '\p{Is_Block=	CJK_Radicals_Supplement}', "");
Expect(1, 12032, '\p{^Is_Block=	CJK_Radicals_Supplement}', "");
Expect(1, 12032, '\P{Is_Block=	CJK_Radicals_Supplement}', "");
Expect(0, 12032, '\P{^Is_Block=	CJK_Radicals_Supplement}', "");
Error('\p{Is_Blk=:=--CJK_Radicals_sup}');
Error('\P{Is_Blk=:=--CJK_Radicals_sup}');
Expect(1, 12031, '\p{Is_Blk=cjkradicalssup}', "");
Expect(0, 12031, '\p{^Is_Blk=cjkradicalssup}', "");
Expect(0, 12031, '\P{Is_Blk=cjkradicalssup}', "");
Expect(1, 12031, '\P{^Is_Blk=cjkradicalssup}', "");
Expect(0, 12032, '\p{Is_Blk=cjkradicalssup}', "");
Expect(1, 12032, '\p{^Is_Blk=cjkradicalssup}', "");
Expect(1, 12032, '\P{Is_Blk=cjkradicalssup}', "");
Expect(0, 12032, '\P{^Is_Blk=cjkradicalssup}', "");
Expect(1, 12031, '\p{Is_Blk=_ CJK_radicals_sup}', "");
Expect(0, 12031, '\p{^Is_Blk=_ CJK_radicals_sup}', "");
Expect(0, 12031, '\P{Is_Blk=_ CJK_radicals_sup}', "");
Expect(1, 12031, '\P{^Is_Blk=_ CJK_radicals_sup}', "");
Expect(0, 12032, '\p{Is_Blk=_ CJK_radicals_sup}', "");
Expect(1, 12032, '\p{^Is_Blk=_ CJK_radicals_sup}', "");
Expect(1, 12032, '\P{Is_Blk=_ CJK_radicals_sup}', "");
Expect(0, 12032, '\P{^Is_Blk=_ CJK_radicals_sup}', "");
Error('\p{Block=-	cjk_Strokes/a/}');
Error('\P{Block=-	cjk_Strokes/a/}');
Expect(1, 12783, '\p{Block:   cjkstrokes}', "");
Expect(0, 12783, '\p{^Block:   cjkstrokes}', "");
Expect(0, 12783, '\P{Block:   cjkstrokes}', "");
Expect(1, 12783, '\P{^Block:   cjkstrokes}', "");
Expect(0, 12784, '\p{Block:   cjkstrokes}', "");
Expect(1, 12784, '\p{^Block:   cjkstrokes}', "");
Expect(1, 12784, '\P{Block:   cjkstrokes}', "");
Expect(0, 12784, '\P{^Block:   cjkstrokes}', "");
Expect(1, 12783, '\p{Block= -CJK_Strokes}', "");
Expect(0, 12783, '\p{^Block= -CJK_Strokes}', "");
Expect(0, 12783, '\P{Block= -CJK_Strokes}', "");
Expect(1, 12783, '\P{^Block= -CJK_Strokes}', "");
Expect(0, 12784, '\p{Block= -CJK_Strokes}', "");
Expect(1, 12784, '\p{^Block= -CJK_Strokes}', "");
Expect(1, 12784, '\P{Block= -CJK_Strokes}', "");
Expect(0, 12784, '\P{^Block= -CJK_Strokes}', "");
Error('\p{Blk=- CJK_strokes:=}');
Error('\P{Blk=- CJK_strokes:=}');
Expect(1, 12783, '\p{Blk=cjkstrokes}', "");
Expect(0, 12783, '\p{^Blk=cjkstrokes}', "");
Expect(0, 12783, '\P{Blk=cjkstrokes}', "");
Expect(1, 12783, '\P{^Blk=cjkstrokes}', "");
Expect(0, 12784, '\p{Blk=cjkstrokes}', "");
Expect(1, 12784, '\p{^Blk=cjkstrokes}', "");
Expect(1, 12784, '\P{Blk=cjkstrokes}', "");
Expect(0, 12784, '\P{^Blk=cjkstrokes}', "");
Expect(1, 12783, '\p{Blk=_CJK_Strokes}', "");
Expect(0, 12783, '\p{^Blk=_CJK_Strokes}', "");
Expect(0, 12783, '\P{Blk=_CJK_Strokes}', "");
Expect(1, 12783, '\P{^Blk=_CJK_Strokes}', "");
Expect(0, 12784, '\p{Blk=_CJK_Strokes}', "");
Expect(1, 12784, '\p{^Blk=_CJK_Strokes}', "");
Expect(1, 12784, '\P{Blk=_CJK_Strokes}', "");
Expect(0, 12784, '\P{^Blk=_CJK_Strokes}', "");
Error('\p{Is_Block=_-CJK_STROKES:=}');
Error('\P{Is_Block=_-CJK_STROKES:=}');
Expect(1, 12783, '\p{Is_Block=cjkstrokes}', "");
Expect(0, 12783, '\p{^Is_Block=cjkstrokes}', "");
Expect(0, 12783, '\P{Is_Block=cjkstrokes}', "");
Expect(1, 12783, '\P{^Is_Block=cjkstrokes}', "");
Expect(0, 12784, '\p{Is_Block=cjkstrokes}', "");
Expect(1, 12784, '\p{^Is_Block=cjkstrokes}', "");
Expect(1, 12784, '\P{Is_Block=cjkstrokes}', "");
Expect(0, 12784, '\P{^Is_Block=cjkstrokes}', "");
Expect(1, 12783, '\p{Is_Block= _CJK_Strokes}', "");
Expect(0, 12783, '\p{^Is_Block= _CJK_Strokes}', "");
Expect(0, 12783, '\P{Is_Block= _CJK_Strokes}', "");
Expect(1, 12783, '\P{^Is_Block= _CJK_Strokes}', "");
Expect(0, 12784, '\p{Is_Block= _CJK_Strokes}', "");
Expect(1, 12784, '\p{^Is_Block= _CJK_Strokes}', "");
Expect(1, 12784, '\P{Is_Block= _CJK_Strokes}', "");
Expect(0, 12784, '\P{^Is_Block= _CJK_Strokes}', "");
Error('\p{Is_Blk=/a/CJK_strokes}');
Error('\P{Is_Blk=/a/CJK_strokes}');
Expect(1, 12783, '\p{Is_Blk=cjkstrokes}', "");
Expect(0, 12783, '\p{^Is_Blk=cjkstrokes}', "");
Expect(0, 12783, '\P{Is_Blk=cjkstrokes}', "");
Expect(1, 12783, '\P{^Is_Blk=cjkstrokes}', "");
Expect(0, 12784, '\p{Is_Blk=cjkstrokes}', "");
Expect(1, 12784, '\p{^Is_Blk=cjkstrokes}', "");
Expect(1, 12784, '\P{Is_Blk=cjkstrokes}', "");
Expect(0, 12784, '\P{^Is_Blk=cjkstrokes}', "");
Expect(1, 12783, '\p{Is_Blk:	-	CJK_STROKES}', "");
Expect(0, 12783, '\p{^Is_Blk:	-	CJK_STROKES}', "");
Expect(0, 12783, '\P{Is_Blk:	-	CJK_STROKES}', "");
Expect(1, 12783, '\P{^Is_Blk:	-	CJK_STROKES}', "");
Expect(0, 12784, '\p{Is_Blk:	-	CJK_STROKES}', "");
Expect(1, 12784, '\p{^Is_Blk:	-	CJK_STROKES}', "");
Expect(1, 12784, '\P{Is_Blk:	-	CJK_STROKES}', "");
Expect(0, 12784, '\P{^Is_Blk:	-	CJK_STROKES}', "");
Error('\p{Block=/a/__CJK_symbols_and_punctuation}');
Error('\P{Block=/a/__CJK_symbols_and_punctuation}');
Expect(1, 12351, '\p{Block=cjksymbolsandpunctuation}', "");
Expect(0, 12351, '\p{^Block=cjksymbolsandpunctuation}', "");
Expect(0, 12351, '\P{Block=cjksymbolsandpunctuation}', "");
Expect(1, 12351, '\P{^Block=cjksymbolsandpunctuation}', "");
Expect(0, 12352, '\p{Block=cjksymbolsandpunctuation}', "");
Expect(1, 12352, '\p{^Block=cjksymbolsandpunctuation}', "");
Expect(1, 12352, '\P{Block=cjksymbolsandpunctuation}', "");
Expect(0, 12352, '\P{^Block=cjksymbolsandpunctuation}', "");
Expect(1, 12351, '\p{Block:	 CJK_Symbols_AND_punctuation}', "");
Expect(0, 12351, '\p{^Block:	 CJK_Symbols_AND_punctuation}', "");
Expect(0, 12351, '\P{Block:	 CJK_Symbols_AND_punctuation}', "");
Expect(1, 12351, '\P{^Block:	 CJK_Symbols_AND_punctuation}', "");
Expect(0, 12352, '\p{Block:	 CJK_Symbols_AND_punctuation}', "");
Expect(1, 12352, '\p{^Block:	 CJK_Symbols_AND_punctuation}', "");
Expect(1, 12352, '\P{Block:	 CJK_Symbols_AND_punctuation}', "");
Expect(0, 12352, '\P{^Block:	 CJK_Symbols_AND_punctuation}', "");
Error('\p{Blk=	CJK_SYMBOLS:=}');
Error('\P{Blk=	CJK_SYMBOLS:=}');
Expect(1, 12351, '\p{Blk=cjksymbols}', "");
Expect(0, 12351, '\p{^Blk=cjksymbols}', "");
Expect(0, 12351, '\P{Blk=cjksymbols}', "");
Expect(1, 12351, '\P{^Blk=cjksymbols}', "");
Expect(0, 12352, '\p{Blk=cjksymbols}', "");
Expect(1, 12352, '\p{^Blk=cjksymbols}', "");
Expect(1, 12352, '\P{Blk=cjksymbols}', "");
Expect(0, 12352, '\P{^Blk=cjksymbols}', "");
Expect(1, 12351, '\p{Blk=	 cjk_symbols}', "");
Expect(0, 12351, '\p{^Blk=	 cjk_symbols}', "");
Expect(0, 12351, '\P{Blk=	 cjk_symbols}', "");
Expect(1, 12351, '\P{^Blk=	 cjk_symbols}', "");
Expect(0, 12352, '\p{Blk=	 cjk_symbols}', "");
Expect(1, 12352, '\p{^Blk=	 cjk_symbols}', "");
Expect(1, 12352, '\P{Blk=	 cjk_symbols}', "");
Expect(0, 12352, '\P{^Blk=	 cjk_symbols}', "");
Error('\p{Is_Block=/a/	 CJK_Symbols_And_PUNCTUATION}');
Error('\P{Is_Block=/a/	 CJK_Symbols_And_PUNCTUATION}');
Expect(1, 12351, '\p{Is_Block=cjksymbolsandpunctuation}', "");
Expect(0, 12351, '\p{^Is_Block=cjksymbolsandpunctuation}', "");
Expect(0, 12351, '\P{Is_Block=cjksymbolsandpunctuation}', "");
Expect(1, 12351, '\P{^Is_Block=cjksymbolsandpunctuation}', "");
Expect(0, 12352, '\p{Is_Block=cjksymbolsandpunctuation}', "");
Expect(1, 12352, '\p{^Is_Block=cjksymbolsandpunctuation}', "");
Expect(1, 12352, '\P{Is_Block=cjksymbolsandpunctuation}', "");
Expect(0, 12352, '\P{^Is_Block=cjksymbolsandpunctuation}', "");
Expect(1, 12351, '\p{Is_Block=_CJK_Symbols_And_punctuation}', "");
Expect(0, 12351, '\p{^Is_Block=_CJK_Symbols_And_punctuation}', "");
Expect(0, 12351, '\P{Is_Block=_CJK_Symbols_And_punctuation}', "");
Expect(1, 12351, '\P{^Is_Block=_CJK_Symbols_And_punctuation}', "");
Expect(0, 12352, '\p{Is_Block=_CJK_Symbols_And_punctuation}', "");
Expect(1, 12352, '\p{^Is_Block=_CJK_Symbols_And_punctuation}', "");
Expect(1, 12352, '\P{Is_Block=_CJK_Symbols_And_punctuation}', "");
Expect(0, 12352, '\P{^Is_Block=_CJK_Symbols_And_punctuation}', "");
Error('\p{Is_Blk=/a/  cjk_symbols}');
Error('\P{Is_Blk=/a/  cjk_symbols}');
Expect(1, 12351, '\p{Is_Blk=cjksymbols}', "");
Expect(0, 12351, '\p{^Is_Blk=cjksymbols}', "");
Expect(0, 12351, '\P{Is_Blk=cjksymbols}', "");
Expect(1, 12351, '\P{^Is_Blk=cjksymbols}', "");
Expect(0, 12352, '\p{Is_Blk=cjksymbols}', "");
Expect(1, 12352, '\p{^Is_Blk=cjksymbols}', "");
Expect(1, 12352, '\P{Is_Blk=cjksymbols}', "");
Expect(0, 12352, '\P{^Is_Blk=cjksymbols}', "");
Expect(1, 12351, '\p{Is_Blk=__cjk_symbols}', "");
Expect(0, 12351, '\p{^Is_Blk=__cjk_symbols}', "");
Expect(0, 12351, '\P{Is_Blk=__cjk_symbols}', "");
Expect(1, 12351, '\P{^Is_Blk=__cjk_symbols}', "");
Expect(0, 12352, '\p{Is_Blk=__cjk_symbols}', "");
Expect(1, 12352, '\p{^Is_Blk=__cjk_symbols}', "");
Expect(1, 12352, '\P{Is_Blk=__cjk_symbols}', "");
Expect(0, 12352, '\P{^Is_Blk=__cjk_symbols}', "");
Error('\p{Block=:=--Carian}');
Error('\P{Block=:=--Carian}');
Expect(1, 66271, '\p{Block=carian}', "");
Expect(0, 66271, '\p{^Block=carian}', "");
Expect(0, 66271, '\P{Block=carian}', "");
Expect(1, 66271, '\P{^Block=carian}', "");
Expect(0, 66272, '\p{Block=carian}', "");
Expect(1, 66272, '\p{^Block=carian}', "");
Expect(1, 66272, '\P{Block=carian}', "");
Expect(0, 66272, '\P{^Block=carian}', "");
Expect(1, 66271, '\p{Block=	_Carian}', "");
Expect(0, 66271, '\p{^Block=	_Carian}', "");
Expect(0, 66271, '\P{Block=	_Carian}', "");
Expect(1, 66271, '\P{^Block=	_Carian}', "");
Expect(0, 66272, '\p{Block=	_Carian}', "");
Expect(1, 66272, '\p{^Block=	_Carian}', "");
Expect(1, 66272, '\P{Block=	_Carian}', "");
Expect(0, 66272, '\P{^Block=	_Carian}', "");
Error('\p{Blk=/a/_	Carian}');
Error('\P{Blk=/a/_	Carian}');
Expect(1, 66271, '\p{Blk=carian}', "");
Expect(0, 66271, '\p{^Blk=carian}', "");
Expect(0, 66271, '\P{Blk=carian}', "");
Expect(1, 66271, '\P{^Blk=carian}', "");
Expect(0, 66272, '\p{Blk=carian}', "");
Expect(1, 66272, '\p{^Blk=carian}', "");
Expect(1, 66272, '\P{Blk=carian}', "");
Expect(0, 66272, '\P{^Blk=carian}', "");
Expect(1, 66271, '\p{Blk: _	carian}', "");
Expect(0, 66271, '\p{^Blk: _	carian}', "");
Expect(0, 66271, '\P{Blk: _	carian}', "");
Expect(1, 66271, '\P{^Blk: _	carian}', "");
Expect(0, 66272, '\p{Blk: _	carian}', "");
Expect(1, 66272, '\p{^Blk: _	carian}', "");
Expect(1, 66272, '\P{Blk: _	carian}', "");
Expect(0, 66272, '\P{^Blk: _	carian}', "");
Error('\p{Is_Block= CARIAN/a/}');
Error('\P{Is_Block= CARIAN/a/}');
Expect(1, 66271, '\p{Is_Block=carian}', "");
Expect(0, 66271, '\p{^Is_Block=carian}', "");
Expect(0, 66271, '\P{Is_Block=carian}', "");
Expect(1, 66271, '\P{^Is_Block=carian}', "");
Expect(0, 66272, '\p{Is_Block=carian}', "");
Expect(1, 66272, '\p{^Is_Block=carian}', "");
Expect(1, 66272, '\P{Is_Block=carian}', "");
Expect(0, 66272, '\P{^Is_Block=carian}', "");
Expect(1, 66271, '\p{Is_Block=-_carian}', "");
Expect(0, 66271, '\p{^Is_Block=-_carian}', "");
Expect(0, 66271, '\P{Is_Block=-_carian}', "");
Expect(1, 66271, '\P{^Is_Block=-_carian}', "");
Expect(0, 66272, '\p{Is_Block=-_carian}', "");
Expect(1, 66272, '\p{^Is_Block=-_carian}', "");
Expect(1, 66272, '\P{Is_Block=-_carian}', "");
Expect(0, 66272, '\P{^Is_Block=-_carian}', "");
Error('\p{Is_Blk=:=	-Carian}');
Error('\P{Is_Blk=:=	-Carian}');
Expect(1, 66271, '\p{Is_Blk=carian}', "");
Expect(0, 66271, '\p{^Is_Blk=carian}', "");
Expect(0, 66271, '\P{Is_Blk=carian}', "");
Expect(1, 66271, '\P{^Is_Blk=carian}', "");
Expect(0, 66272, '\p{Is_Blk=carian}', "");
Expect(1, 66272, '\p{^Is_Blk=carian}', "");
Expect(1, 66272, '\P{Is_Blk=carian}', "");
Expect(0, 66272, '\P{^Is_Blk=carian}', "");
Expect(1, 66271, '\p{Is_Blk=		Carian}', "");
Expect(0, 66271, '\p{^Is_Blk=		Carian}', "");
Expect(0, 66271, '\P{Is_Blk=		Carian}', "");
Expect(1, 66271, '\P{^Is_Blk=		Carian}', "");
Expect(0, 66272, '\p{Is_Blk=		Carian}', "");
Expect(1, 66272, '\p{^Is_Blk=		Carian}', "");
Expect(1, 66272, '\P{Is_Blk=		Carian}', "");
Expect(0, 66272, '\P{^Is_Blk=		Carian}', "");
Error('\p{Block=-/a/CHAKMA}');
Error('\P{Block=-/a/CHAKMA}');
Expect(1, 69967, '\p{Block=chakma}', "");
Expect(0, 69967, '\p{^Block=chakma}', "");
Expect(0, 69967, '\P{Block=chakma}', "");
Expect(1, 69967, '\P{^Block=chakma}', "");
Expect(0, 69968, '\p{Block=chakma}', "");
Expect(1, 69968, '\p{^Block=chakma}', "");
Expect(1, 69968, '\P{Block=chakma}', "");
Expect(0, 69968, '\P{^Block=chakma}', "");
Expect(1, 69967, '\p{Block=--chakma}', "");
Expect(0, 69967, '\p{^Block=--chakma}', "");
Expect(0, 69967, '\P{Block=--chakma}', "");
Expect(1, 69967, '\P{^Block=--chakma}', "");
Expect(0, 69968, '\p{Block=--chakma}', "");
Expect(1, 69968, '\p{^Block=--chakma}', "");
Expect(1, 69968, '\P{Block=--chakma}', "");
Expect(0, 69968, '\P{^Block=--chakma}', "");
Error('\p{Blk:	/a/chakma}');
Error('\P{Blk:	/a/chakma}');
Expect(1, 69967, '\p{Blk=chakma}', "");
Expect(0, 69967, '\p{^Blk=chakma}', "");
Expect(0, 69967, '\P{Blk=chakma}', "");
Expect(1, 69967, '\P{^Blk=chakma}', "");
Expect(0, 69968, '\p{Blk=chakma}', "");
Expect(1, 69968, '\p{^Blk=chakma}', "");
Expect(1, 69968, '\P{Blk=chakma}', "");
Expect(0, 69968, '\P{^Blk=chakma}', "");
Expect(1, 69967, '\p{Blk=--chakma}', "");
Expect(0, 69967, '\p{^Blk=--chakma}', "");
Expect(0, 69967, '\P{Blk=--chakma}', "");
Expect(1, 69967, '\P{^Blk=--chakma}', "");
Expect(0, 69968, '\p{Blk=--chakma}', "");
Expect(1, 69968, '\p{^Blk=--chakma}', "");
Expect(1, 69968, '\P{Blk=--chakma}', "");
Expect(0, 69968, '\P{^Blk=--chakma}', "");
Error('\p{Is_Block:	:= -chakma}');
Error('\P{Is_Block:	:= -chakma}');
Expect(1, 69967, '\p{Is_Block=chakma}', "");
Expect(0, 69967, '\p{^Is_Block=chakma}', "");
Expect(0, 69967, '\P{Is_Block=chakma}', "");
Expect(1, 69967, '\P{^Is_Block=chakma}', "");
Expect(0, 69968, '\p{Is_Block=chakma}', "");
Expect(1, 69968, '\p{^Is_Block=chakma}', "");
Expect(1, 69968, '\P{Is_Block=chakma}', "");
Expect(0, 69968, '\P{^Is_Block=chakma}', "");
Expect(1, 69967, '\p{Is_Block= -Chakma}', "");
Expect(0, 69967, '\p{^Is_Block= -Chakma}', "");
Expect(0, 69967, '\P{Is_Block= -Chakma}', "");
Expect(1, 69967, '\P{^Is_Block= -Chakma}', "");
Expect(0, 69968, '\p{Is_Block= -Chakma}', "");
Expect(1, 69968, '\p{^Is_Block= -Chakma}', "");
Expect(1, 69968, '\P{Is_Block= -Chakma}', "");
Expect(0, 69968, '\P{^Is_Block= -Chakma}', "");
Error('\p{Is_Blk=:=-Chakma}');
Error('\P{Is_Blk=:=-Chakma}');
Expect(1, 69967, '\p{Is_Blk=chakma}', "");
Expect(0, 69967, '\p{^Is_Blk=chakma}', "");
Expect(0, 69967, '\P{Is_Blk=chakma}', "");
Expect(1, 69967, '\P{^Is_Blk=chakma}', "");
Expect(0, 69968, '\p{Is_Blk=chakma}', "");
Expect(1, 69968, '\p{^Is_Blk=chakma}', "");
Expect(1, 69968, '\P{Is_Blk=chakma}', "");
Expect(0, 69968, '\P{^Is_Blk=chakma}', "");
Expect(1, 69967, '\p{Is_Blk=		chakma}', "");
Expect(0, 69967, '\p{^Is_Blk=		chakma}', "");
Expect(0, 69967, '\P{Is_Blk=		chakma}', "");
Expect(1, 69967, '\P{^Is_Blk=		chakma}', "");
Expect(0, 69968, '\p{Is_Blk=		chakma}', "");
Expect(1, 69968, '\p{^Is_Blk=		chakma}', "");
Expect(1, 69968, '\P{Is_Blk=		chakma}', "");
Expect(0, 69968, '\P{^Is_Blk=		chakma}', "");
Error('\p{Block= _Cham:=}');
Error('\P{Block= _Cham:=}');
Expect(1, 43615, '\p{Block=cham}', "");
Expect(0, 43615, '\p{^Block=cham}', "");
Expect(0, 43615, '\P{Block=cham}', "");
Expect(1, 43615, '\P{^Block=cham}', "");
Expect(0, 43616, '\p{Block=cham}', "");
Expect(1, 43616, '\p{^Block=cham}', "");
Expect(1, 43616, '\P{Block=cham}', "");
Expect(0, 43616, '\P{^Block=cham}', "");
Expect(1, 43615, '\p{Block:-cham}', "");
Expect(0, 43615, '\p{^Block:-cham}', "");
Expect(0, 43615, '\P{Block:-cham}', "");
Expect(1, 43615, '\P{^Block:-cham}', "");
Expect(0, 43616, '\p{Block:-cham}', "");
Expect(1, 43616, '\p{^Block:-cham}', "");
Expect(1, 43616, '\P{Block:-cham}', "");
Expect(0, 43616, '\P{^Block:-cham}', "");
Error('\p{Blk=	/a/cham}');
Error('\P{Blk=	/a/cham}');
Expect(1, 43615, '\p{Blk=cham}', "");
Expect(0, 43615, '\p{^Blk=cham}', "");
Expect(0, 43615, '\P{Blk=cham}', "");
Expect(1, 43615, '\P{^Blk=cham}', "");
Expect(0, 43616, '\p{Blk=cham}', "");
Expect(1, 43616, '\p{^Blk=cham}', "");
Expect(1, 43616, '\P{Blk=cham}', "");
Expect(0, 43616, '\P{^Blk=cham}', "");
Expect(1, 43615, '\p{Blk=_CHAM}', "");
Expect(0, 43615, '\p{^Blk=_CHAM}', "");
Expect(0, 43615, '\P{Blk=_CHAM}', "");
Expect(1, 43615, '\P{^Blk=_CHAM}', "");
Expect(0, 43616, '\p{Blk=_CHAM}', "");
Expect(1, 43616, '\p{^Blk=_CHAM}', "");
Expect(1, 43616, '\P{Blk=_CHAM}', "");
Expect(0, 43616, '\P{^Blk=_CHAM}', "");
Error('\p{Is_Block: :=Cham}');
Error('\P{Is_Block: :=Cham}');
Expect(1, 43615, '\p{Is_Block=cham}', "");
Expect(0, 43615, '\p{^Is_Block=cham}', "");
Expect(0, 43615, '\P{Is_Block=cham}', "");
Expect(1, 43615, '\P{^Is_Block=cham}', "");
Expect(0, 43616, '\p{Is_Block=cham}', "");
Expect(1, 43616, '\p{^Is_Block=cham}', "");
Expect(1, 43616, '\P{Is_Block=cham}', "");
Expect(0, 43616, '\P{^Is_Block=cham}', "");
Expect(1, 43615, '\p{Is_Block=-cham}', "");
Expect(0, 43615, '\p{^Is_Block=-cham}', "");
Expect(0, 43615, '\P{Is_Block=-cham}', "");
Expect(1, 43615, '\P{^Is_Block=-cham}', "");
Expect(0, 43616, '\p{Is_Block=-cham}', "");
Expect(1, 43616, '\p{^Is_Block=-cham}', "");
Expect(1, 43616, '\P{Is_Block=-cham}', "");
Expect(0, 43616, '\P{^Is_Block=-cham}', "");
Error('\p{Is_Blk=--cham:=}');
Error('\P{Is_Blk=--cham:=}');
Expect(1, 43615, '\p{Is_Blk=cham}', "");
Expect(0, 43615, '\p{^Is_Blk=cham}', "");
Expect(0, 43615, '\P{Is_Blk=cham}', "");
Expect(1, 43615, '\P{^Is_Blk=cham}', "");
Expect(0, 43616, '\p{Is_Blk=cham}', "");
Expect(1, 43616, '\p{^Is_Blk=cham}', "");
Expect(1, 43616, '\P{Is_Blk=cham}', "");
Expect(0, 43616, '\P{^Is_Blk=cham}', "");
Expect(1, 43615, '\p{Is_Blk:	 CHAM}', "");
Expect(0, 43615, '\p{^Is_Blk:	 CHAM}', "");
Expect(0, 43615, '\P{Is_Blk:	 CHAM}', "");
Expect(1, 43615, '\P{^Is_Blk:	 CHAM}', "");
Expect(0, 43616, '\p{Is_Blk:	 CHAM}', "");
Expect(1, 43616, '\p{^Is_Blk:	 CHAM}', "");
Expect(1, 43616, '\P{Is_Blk:	 CHAM}', "");
Expect(0, 43616, '\P{^Is_Blk:	 CHAM}', "");
Error('\p{Block=--Cherokee:=}');
Error('\P{Block=--Cherokee:=}');
Expect(1, 5119, '\p{Block=cherokee}', "");
Expect(0, 5119, '\p{^Block=cherokee}', "");
Expect(0, 5119, '\P{Block=cherokee}', "");
Expect(1, 5119, '\P{^Block=cherokee}', "");
Expect(0, 5120, '\p{Block=cherokee}', "");
Expect(1, 5120, '\p{^Block=cherokee}', "");
Expect(1, 5120, '\P{Block=cherokee}', "");
Expect(0, 5120, '\P{^Block=cherokee}', "");
Expect(1, 5119, '\p{Block=_CHEROKEE}', "");
Expect(0, 5119, '\p{^Block=_CHEROKEE}', "");
Expect(0, 5119, '\P{Block=_CHEROKEE}', "");
Expect(1, 5119, '\P{^Block=_CHEROKEE}', "");
Expect(0, 5120, '\p{Block=_CHEROKEE}', "");
Expect(1, 5120, '\p{^Block=_CHEROKEE}', "");
Expect(1, 5120, '\P{Block=_CHEROKEE}', "");
Expect(0, 5120, '\P{^Block=_CHEROKEE}', "");
Error('\p{Blk:	:=	 Cherokee}');
Error('\P{Blk:	:=	 Cherokee}');
Expect(1, 5119, '\p{Blk=cherokee}', "");
Expect(0, 5119, '\p{^Blk=cherokee}', "");
Expect(0, 5119, '\P{Blk=cherokee}', "");
Expect(1, 5119, '\P{^Blk=cherokee}', "");
Expect(0, 5120, '\p{Blk=cherokee}', "");
Expect(1, 5120, '\p{^Blk=cherokee}', "");
Expect(1, 5120, '\P{Blk=cherokee}', "");
Expect(0, 5120, '\P{^Blk=cherokee}', "");
Expect(1, 5119, '\p{Blk=		cherokee}', "");
Expect(0, 5119, '\p{^Blk=		cherokee}', "");
Expect(0, 5119, '\P{Blk=		cherokee}', "");
Expect(1, 5119, '\P{^Blk=		cherokee}', "");
Expect(0, 5120, '\p{Blk=		cherokee}', "");
Expect(1, 5120, '\p{^Blk=		cherokee}', "");
Expect(1, 5120, '\P{Blk=		cherokee}', "");
Expect(0, 5120, '\P{^Blk=		cherokee}', "");
Error('\p{Is_Block=:=		Cherokee}');
Error('\P{Is_Block=:=		Cherokee}');
Expect(1, 5119, '\p{Is_Block=cherokee}', "");
Expect(0, 5119, '\p{^Is_Block=cherokee}', "");
Expect(0, 5119, '\P{Is_Block=cherokee}', "");
Expect(1, 5119, '\P{^Is_Block=cherokee}', "");
Expect(0, 5120, '\p{Is_Block=cherokee}', "");
Expect(1, 5120, '\p{^Is_Block=cherokee}', "");
Expect(1, 5120, '\P{Is_Block=cherokee}', "");
Expect(0, 5120, '\P{^Is_Block=cherokee}', "");
Expect(1, 5119, '\p{Is_Block=_	Cherokee}', "");
Expect(0, 5119, '\p{^Is_Block=_	Cherokee}', "");
Expect(0, 5119, '\P{Is_Block=_	Cherokee}', "");
Expect(1, 5119, '\P{^Is_Block=_	Cherokee}', "");
Expect(0, 5120, '\p{Is_Block=_	Cherokee}', "");
Expect(1, 5120, '\p{^Is_Block=_	Cherokee}', "");
Expect(1, 5120, '\P{Is_Block=_	Cherokee}', "");
Expect(0, 5120, '\P{^Is_Block=_	Cherokee}', "");
Error('\p{Is_Blk=-Cherokee/a/}');
Error('\P{Is_Blk=-Cherokee/a/}');
Expect(1, 5119, '\p{Is_Blk=cherokee}', "");
Expect(0, 5119, '\p{^Is_Blk=cherokee}', "");
Expect(0, 5119, '\P{Is_Blk=cherokee}', "");
Expect(1, 5119, '\P{^Is_Blk=cherokee}', "");
Expect(0, 5120, '\p{Is_Blk=cherokee}', "");
Expect(1, 5120, '\p{^Is_Blk=cherokee}', "");
Expect(1, 5120, '\P{Is_Blk=cherokee}', "");
Expect(0, 5120, '\P{^Is_Blk=cherokee}', "");
Expect(1, 5119, '\p{Is_Blk=	 Cherokee}', "");
Expect(0, 5119, '\p{^Is_Blk=	 Cherokee}', "");
Expect(0, 5119, '\P{Is_Blk=	 Cherokee}', "");
Expect(1, 5119, '\P{^Is_Blk=	 Cherokee}', "");
Expect(0, 5120, '\p{Is_Blk=	 Cherokee}', "");
Expect(1, 5120, '\p{^Is_Blk=	 Cherokee}', "");
Expect(1, 5120, '\P{Is_Blk=	 Cherokee}', "");
Expect(0, 5120, '\P{^Is_Blk=	 Cherokee}', "");
Error('\p{Block=	/a/Hangul_compatibility_Jamo}');
Error('\P{Block=	/a/Hangul_compatibility_Jamo}');
Expect(1, 12687, '\p{Block=hangulcompatibilityjamo}', "");
Expect(0, 12687, '\p{^Block=hangulcompatibilityjamo}', "");
Expect(0, 12687, '\P{Block=hangulcompatibilityjamo}', "");
Expect(1, 12687, '\P{^Block=hangulcompatibilityjamo}', "");
Expect(0, 12688, '\p{Block=hangulcompatibilityjamo}', "");
Expect(1, 12688, '\p{^Block=hangulcompatibilityjamo}', "");
Expect(1, 12688, '\P{Block=hangulcompatibilityjamo}', "");
Expect(0, 12688, '\P{^Block=hangulcompatibilityjamo}', "");
Expect(1, 12687, '\p{Block:	 	HANGUL_COMPATIBILITY_jamo}', "");
Expect(0, 12687, '\p{^Block:	 	HANGUL_COMPATIBILITY_jamo}', "");
Expect(0, 12687, '\P{Block:	 	HANGUL_COMPATIBILITY_jamo}', "");
Expect(1, 12687, '\P{^Block:	 	HANGUL_COMPATIBILITY_jamo}', "");
Expect(0, 12688, '\p{Block:	 	HANGUL_COMPATIBILITY_jamo}', "");
Expect(1, 12688, '\p{^Block:	 	HANGUL_COMPATIBILITY_jamo}', "");
Expect(1, 12688, '\P{Block:	 	HANGUL_COMPATIBILITY_jamo}', "");
Expect(0, 12688, '\P{^Block:	 	HANGUL_COMPATIBILITY_jamo}', "");
Error('\p{Blk=_-COMPAT_Jamo:=}');
Error('\P{Blk=_-COMPAT_Jamo:=}');
Expect(1, 12687, '\p{Blk=compatjamo}', "");
Expect(0, 12687, '\p{^Blk=compatjamo}', "");
Expect(0, 12687, '\P{Blk=compatjamo}', "");
Expect(1, 12687, '\P{^Blk=compatjamo}', "");
Expect(0, 12688, '\p{Blk=compatjamo}', "");
Expect(1, 12688, '\p{^Blk=compatjamo}', "");
Expect(1, 12688, '\P{Blk=compatjamo}', "");
Expect(0, 12688, '\P{^Blk=compatjamo}', "");
Expect(1, 12687, '\p{Blk= _Compat_jamo}', "");
Expect(0, 12687, '\p{^Blk= _Compat_jamo}', "");
Expect(0, 12687, '\P{Blk= _Compat_jamo}', "");
Expect(1, 12687, '\P{^Blk= _Compat_jamo}', "");
Expect(0, 12688, '\p{Blk= _Compat_jamo}', "");
Expect(1, 12688, '\p{^Blk= _Compat_jamo}', "");
Expect(1, 12688, '\P{Blk= _Compat_jamo}', "");
Expect(0, 12688, '\P{^Blk= _Compat_jamo}', "");
Error('\p{Is_Block= Hangul_compatibility_Jamo:=}');
Error('\P{Is_Block= Hangul_compatibility_Jamo:=}');
Expect(1, 12687, '\p{Is_Block=hangulcompatibilityjamo}', "");
Expect(0, 12687, '\p{^Is_Block=hangulcompatibilityjamo}', "");
Expect(0, 12687, '\P{Is_Block=hangulcompatibilityjamo}', "");
Expect(1, 12687, '\P{^Is_Block=hangulcompatibilityjamo}', "");
Expect(0, 12688, '\p{Is_Block=hangulcompatibilityjamo}', "");
Expect(1, 12688, '\p{^Is_Block=hangulcompatibilityjamo}', "");
Expect(1, 12688, '\P{Is_Block=hangulcompatibilityjamo}', "");
Expect(0, 12688, '\P{^Is_Block=hangulcompatibilityjamo}', "");
Expect(1, 12687, '\p{Is_Block=	-Hangul_compatibility_Jamo}', "");
Expect(0, 12687, '\p{^Is_Block=	-Hangul_compatibility_Jamo}', "");
Expect(0, 12687, '\P{Is_Block=	-Hangul_compatibility_Jamo}', "");
Expect(1, 12687, '\P{^Is_Block=	-Hangul_compatibility_Jamo}', "");
Expect(0, 12688, '\p{Is_Block=	-Hangul_compatibility_Jamo}', "");
Expect(1, 12688, '\p{^Is_Block=	-Hangul_compatibility_Jamo}', "");
Expect(1, 12688, '\P{Is_Block=	-Hangul_compatibility_Jamo}', "");
Expect(0, 12688, '\P{^Is_Block=	-Hangul_compatibility_Jamo}', "");
Error('\p{Is_Blk=:=  Compat_jamo}');
Error('\P{Is_Blk=:=  Compat_jamo}');
Expect(1, 12687, '\p{Is_Blk=compatjamo}', "");
Expect(0, 12687, '\p{^Is_Blk=compatjamo}', "");
Expect(0, 12687, '\P{Is_Blk=compatjamo}', "");
Expect(1, 12687, '\P{^Is_Blk=compatjamo}', "");
Expect(0, 12688, '\p{Is_Blk=compatjamo}', "");
Expect(1, 12688, '\p{^Is_Blk=compatjamo}', "");
Expect(1, 12688, '\P{Is_Blk=compatjamo}', "");
Expect(0, 12688, '\P{^Is_Blk=compatjamo}', "");
Expect(1, 12687, '\p{Is_Blk=- Compat_jamo}', "");
Expect(0, 12687, '\p{^Is_Blk=- Compat_jamo}', "");
Expect(0, 12687, '\P{Is_Blk=- Compat_jamo}', "");
Expect(1, 12687, '\P{^Is_Blk=- Compat_jamo}', "");
Expect(0, 12688, '\p{Is_Blk=- Compat_jamo}', "");
Expect(1, 12688, '\p{^Is_Blk=- Compat_jamo}', "");
Expect(1, 12688, '\P{Is_Blk=- Compat_jamo}', "");
Expect(0, 12688, '\P{^Is_Blk=- Compat_jamo}', "");
Error('\p{Block:   := control_PICTURES}');
Error('\P{Block:   := control_PICTURES}');
Expect(1, 9279, '\p{Block:   controlpictures}', "");
Expect(0, 9279, '\p{^Block:   controlpictures}', "");
Expect(0, 9279, '\P{Block:   controlpictures}', "");
Expect(1, 9279, '\P{^Block:   controlpictures}', "");
Expect(0, 9280, '\p{Block:   controlpictures}', "");
Expect(1, 9280, '\p{^Block:   controlpictures}', "");
Expect(1, 9280, '\P{Block:   controlpictures}', "");
Expect(0, 9280, '\P{^Block:   controlpictures}', "");
Expect(1, 9279, '\p{Block=- control_PICTURES}', "");
Expect(0, 9279, '\p{^Block=- control_PICTURES}', "");
Expect(0, 9279, '\P{Block=- control_PICTURES}', "");
Expect(1, 9279, '\P{^Block=- control_PICTURES}', "");
Expect(0, 9280, '\p{Block=- control_PICTURES}', "");
Expect(1, 9280, '\p{^Block=- control_PICTURES}', "");
Expect(1, 9280, '\P{Block=- control_PICTURES}', "");
Expect(0, 9280, '\P{^Block=- control_PICTURES}', "");
Error('\p{Blk= 	CONTROL_PICTURES:=}');
Error('\P{Blk= 	CONTROL_PICTURES:=}');
Expect(1, 9279, '\p{Blk: controlpictures}', "");
Expect(0, 9279, '\p{^Blk: controlpictures}', "");
Expect(0, 9279, '\P{Blk: controlpictures}', "");
Expect(1, 9279, '\P{^Blk: controlpictures}', "");
Expect(0, 9280, '\p{Blk: controlpictures}', "");
Expect(1, 9280, '\p{^Blk: controlpictures}', "");
Expect(1, 9280, '\P{Blk: controlpictures}', "");
Expect(0, 9280, '\P{^Blk: controlpictures}', "");
Expect(1, 9279, '\p{Blk=_Control_pictures}', "");
Expect(0, 9279, '\p{^Blk=_Control_pictures}', "");
Expect(0, 9279, '\P{Blk=_Control_pictures}', "");
Expect(1, 9279, '\P{^Blk=_Control_pictures}', "");
Expect(0, 9280, '\p{Blk=_Control_pictures}', "");
Expect(1, 9280, '\p{^Blk=_Control_pictures}', "");
Expect(1, 9280, '\P{Blk=_Control_pictures}', "");
Expect(0, 9280, '\P{^Blk=_Control_pictures}', "");
Error('\p{Is_Block=-:=CONTROL_Pictures}');
Error('\P{Is_Block=-:=CONTROL_Pictures}');
Expect(1, 9279, '\p{Is_Block=controlpictures}', "");
Expect(0, 9279, '\p{^Is_Block=controlpictures}', "");
Expect(0, 9279, '\P{Is_Block=controlpictures}', "");
Expect(1, 9279, '\P{^Is_Block=controlpictures}', "");
Expect(0, 9280, '\p{Is_Block=controlpictures}', "");
Expect(1, 9280, '\p{^Is_Block=controlpictures}', "");
Expect(1, 9280, '\P{Is_Block=controlpictures}', "");
Expect(0, 9280, '\P{^Is_Block=controlpictures}', "");
Expect(1, 9279, '\p{Is_Block:_control_PICTURES}', "");
Expect(0, 9279, '\p{^Is_Block:_control_PICTURES}', "");
Expect(0, 9279, '\P{Is_Block:_control_PICTURES}', "");
Expect(1, 9279, '\P{^Is_Block:_control_PICTURES}', "");
Expect(0, 9280, '\p{Is_Block:_control_PICTURES}', "");
Expect(1, 9280, '\p{^Is_Block:_control_PICTURES}', "");
Expect(1, 9280, '\P{Is_Block:_control_PICTURES}', "");
Expect(0, 9280, '\P{^Is_Block:_control_PICTURES}', "");
Error('\p{Is_Blk=	/a/CONTROL_PICTURES}');
Error('\P{Is_Blk=	/a/CONTROL_PICTURES}');
Expect(1, 9279, '\p{Is_Blk=controlpictures}', "");
Expect(0, 9279, '\p{^Is_Blk=controlpictures}', "");
Expect(0, 9279, '\P{Is_Blk=controlpictures}', "");
Expect(1, 9279, '\P{^Is_Blk=controlpictures}', "");
Expect(0, 9280, '\p{Is_Blk=controlpictures}', "");
Expect(1, 9280, '\p{^Is_Blk=controlpictures}', "");
Expect(1, 9280, '\P{Is_Blk=controlpictures}', "");
Expect(0, 9280, '\P{^Is_Blk=controlpictures}', "");
Expect(1, 9279, '\p{Is_Blk=_	Control_PICTURES}', "");
Expect(0, 9279, '\p{^Is_Blk=_	Control_PICTURES}', "");
Expect(0, 9279, '\P{Is_Blk=_	Control_PICTURES}', "");
Expect(1, 9279, '\P{^Is_Blk=_	Control_PICTURES}', "");
Expect(0, 9280, '\p{Is_Blk=_	Control_PICTURES}', "");
Expect(1, 9280, '\p{^Is_Blk=_	Control_PICTURES}', "");
Expect(1, 9280, '\P{Is_Blk=_	Control_PICTURES}', "");
Expect(0, 9280, '\P{^Is_Blk=_	Control_PICTURES}', "");
Error('\p{Block= :=Coptic}');
Error('\P{Block= :=Coptic}');
Expect(1, 11519, '\p{Block:	coptic}', "");
Expect(0, 11519, '\p{^Block:	coptic}', "");
Expect(0, 11519, '\P{Block:	coptic}', "");
Expect(1, 11519, '\P{^Block:	coptic}', "");
Expect(0, 11520, '\p{Block:	coptic}', "");
Expect(1, 11520, '\p{^Block:	coptic}', "");
Expect(1, 11520, '\P{Block:	coptic}', "");
Expect(0, 11520, '\P{^Block:	coptic}', "");
Expect(1, 11519, '\p{Block=	 Coptic}', "");
Expect(0, 11519, '\p{^Block=	 Coptic}', "");
Expect(0, 11519, '\P{Block=	 Coptic}', "");
Expect(1, 11519, '\P{^Block=	 Coptic}', "");
Expect(0, 11520, '\p{Block=	 Coptic}', "");
Expect(1, 11520, '\p{^Block=	 Coptic}', "");
Expect(1, 11520, '\P{Block=	 Coptic}', "");
Expect(0, 11520, '\P{^Block=	 Coptic}', "");
Error('\p{Blk=_:=Coptic}');
Error('\P{Blk=_:=Coptic}');
Expect(1, 11519, '\p{Blk=coptic}', "");
Expect(0, 11519, '\p{^Blk=coptic}', "");
Expect(0, 11519, '\P{Blk=coptic}', "");
Expect(1, 11519, '\P{^Blk=coptic}', "");
Expect(0, 11520, '\p{Blk=coptic}', "");
Expect(1, 11520, '\p{^Blk=coptic}', "");
Expect(1, 11520, '\P{Blk=coptic}', "");
Expect(0, 11520, '\P{^Blk=coptic}', "");
Expect(1, 11519, '\p{Blk:	-_Coptic}', "");
Expect(0, 11519, '\p{^Blk:	-_Coptic}', "");
Expect(0, 11519, '\P{Blk:	-_Coptic}', "");
Expect(1, 11519, '\P{^Blk:	-_Coptic}', "");
Expect(0, 11520, '\p{Blk:	-_Coptic}', "");
Expect(1, 11520, '\p{^Blk:	-_Coptic}', "");
Expect(1, 11520, '\P{Blk:	-_Coptic}', "");
Expect(0, 11520, '\P{^Blk:	-_Coptic}', "");
Error('\p{Is_Block=_/a/Coptic}');
Error('\P{Is_Block=_/a/Coptic}');
Expect(1, 11519, '\p{Is_Block=coptic}', "");
Expect(0, 11519, '\p{^Is_Block=coptic}', "");
Expect(0, 11519, '\P{Is_Block=coptic}', "");
Expect(1, 11519, '\P{^Is_Block=coptic}', "");
Expect(0, 11520, '\p{Is_Block=coptic}', "");
Expect(1, 11520, '\p{^Is_Block=coptic}', "");
Expect(1, 11520, '\P{Is_Block=coptic}', "");
Expect(0, 11520, '\P{^Is_Block=coptic}', "");
Expect(1, 11519, '\p{Is_Block=	-Coptic}', "");
Expect(0, 11519, '\p{^Is_Block=	-Coptic}', "");
Expect(0, 11519, '\P{Is_Block=	-Coptic}', "");
Expect(1, 11519, '\P{^Is_Block=	-Coptic}', "");
Expect(0, 11520, '\p{Is_Block=	-Coptic}', "");
Expect(1, 11520, '\p{^Is_Block=	-Coptic}', "");
Expect(1, 11520, '\P{Is_Block=	-Coptic}', "");
Expect(0, 11520, '\P{^Is_Block=	-Coptic}', "");
Error('\p{Is_Blk= /a/COPTIC}');
Error('\P{Is_Blk= /a/COPTIC}');
Expect(1, 11519, '\p{Is_Blk=coptic}', "");
Expect(0, 11519, '\p{^Is_Blk=coptic}', "");
Expect(0, 11519, '\P{Is_Blk=coptic}', "");
Expect(1, 11519, '\P{^Is_Blk=coptic}', "");
Expect(0, 11520, '\p{Is_Blk=coptic}', "");
Expect(1, 11520, '\p{^Is_Blk=coptic}', "");
Expect(1, 11520, '\P{Is_Blk=coptic}', "");
Expect(0, 11520, '\P{^Is_Blk=coptic}', "");
Expect(1, 11519, '\p{Is_Blk=COPTIC}', "");
Expect(0, 11519, '\p{^Is_Blk=COPTIC}', "");
Expect(0, 11519, '\P{Is_Blk=COPTIC}', "");
Expect(1, 11519, '\P{^Is_Blk=COPTIC}', "");
Expect(0, 11520, '\p{Is_Blk=COPTIC}', "");
Expect(1, 11520, '\p{^Is_Blk=COPTIC}', "");
Expect(1, 11520, '\P{Is_Blk=COPTIC}', "");
Expect(0, 11520, '\P{^Is_Blk=COPTIC}', "");
Error('\p{Block= _COUNTING_ROD_Numerals:=}');
Error('\P{Block= _COUNTING_ROD_Numerals:=}');
Expect(1, 119679, '\p{Block=countingrodnumerals}', "");
Expect(0, 119679, '\p{^Block=countingrodnumerals}', "");
Expect(0, 119679, '\P{Block=countingrodnumerals}', "");
Expect(1, 119679, '\P{^Block=countingrodnumerals}', "");
Expect(0, 119680, '\p{Block=countingrodnumerals}', "");
Expect(1, 119680, '\p{^Block=countingrodnumerals}', "");
Expect(1, 119680, '\P{Block=countingrodnumerals}', "");
Expect(0, 119680, '\P{^Block=countingrodnumerals}', "");
Expect(1, 119679, '\p{Block=	-counting_Rod_numerals}', "");
Expect(0, 119679, '\p{^Block=	-counting_Rod_numerals}', "");
Expect(0, 119679, '\P{Block=	-counting_Rod_numerals}', "");
Expect(1, 119679, '\P{^Block=	-counting_Rod_numerals}', "");
Expect(0, 119680, '\p{Block=	-counting_Rod_numerals}', "");
Expect(1, 119680, '\p{^Block=	-counting_Rod_numerals}', "");
Expect(1, 119680, '\P{Block=	-counting_Rod_numerals}', "");
Expect(0, 119680, '\P{^Block=	-counting_Rod_numerals}', "");
Error('\p{Blk= Counting_rod:=}');
Error('\P{Blk= Counting_rod:=}');
Expect(1, 119679, '\p{Blk=countingrod}', "");
Expect(0, 119679, '\p{^Blk=countingrod}', "");
Expect(0, 119679, '\P{Blk=countingrod}', "");
Expect(1, 119679, '\P{^Blk=countingrod}', "");
Expect(0, 119680, '\p{Blk=countingrod}', "");
Expect(1, 119680, '\p{^Blk=countingrod}', "");
Expect(1, 119680, '\P{Blk=countingrod}', "");
Expect(0, 119680, '\P{^Blk=countingrod}', "");
Expect(1, 119679, '\p{Blk=_	counting_Rod}', "");
Expect(0, 119679, '\p{^Blk=_	counting_Rod}', "");
Expect(0, 119679, '\P{Blk=_	counting_Rod}', "");
Expect(1, 119679, '\P{^Blk=_	counting_Rod}', "");
Expect(0, 119680, '\p{Blk=_	counting_Rod}', "");
Expect(1, 119680, '\p{^Blk=_	counting_Rod}', "");
Expect(1, 119680, '\P{Blk=_	counting_Rod}', "");
Expect(0, 119680, '\P{^Blk=_	counting_Rod}', "");
Error('\p{Is_Block=:=	-counting_Rod_numerals}');
Error('\P{Is_Block=:=	-counting_Rod_numerals}');
Expect(1, 119679, '\p{Is_Block=countingrodnumerals}', "");
Expect(0, 119679, '\p{^Is_Block=countingrodnumerals}', "");
Expect(0, 119679, '\P{Is_Block=countingrodnumerals}', "");
Expect(1, 119679, '\P{^Is_Block=countingrodnumerals}', "");
Expect(0, 119680, '\p{Is_Block=countingrodnumerals}', "");
Expect(1, 119680, '\p{^Is_Block=countingrodnumerals}', "");
Expect(1, 119680, '\P{Is_Block=countingrodnumerals}', "");
Expect(0, 119680, '\P{^Is_Block=countingrodnumerals}', "");
Expect(1, 119679, '\p{Is_Block=	_Counting_Rod_NUMERALS}', "");
Expect(0, 119679, '\p{^Is_Block=	_Counting_Rod_NUMERALS}', "");
Expect(0, 119679, '\P{Is_Block=	_Counting_Rod_NUMERALS}', "");
Expect(1, 119679, '\P{^Is_Block=	_Counting_Rod_NUMERALS}', "");
Expect(0, 119680, '\p{Is_Block=	_Counting_Rod_NUMERALS}', "");
Expect(1, 119680, '\p{^Is_Block=	_Counting_Rod_NUMERALS}', "");
Expect(1, 119680, '\P{Is_Block=	_Counting_Rod_NUMERALS}', "");
Expect(0, 119680, '\P{^Is_Block=	_Counting_Rod_NUMERALS}', "");
Error('\p{Is_Blk=/a/-_Counting_Rod}');
Error('\P{Is_Blk=/a/-_Counting_Rod}');
Expect(1, 119679, '\p{Is_Blk=countingrod}', "");
Expect(0, 119679, '\p{^Is_Blk=countingrod}', "");
Expect(0, 119679, '\P{Is_Blk=countingrod}', "");
Expect(1, 119679, '\P{^Is_Blk=countingrod}', "");
Expect(0, 119680, '\p{Is_Blk=countingrod}', "");
Expect(1, 119680, '\p{^Is_Blk=countingrod}', "");
Expect(1, 119680, '\P{Is_Blk=countingrod}', "");
Expect(0, 119680, '\P{^Is_Blk=countingrod}', "");
Expect(1, 119679, '\p{Is_Blk=	Counting_Rod}', "");
Expect(0, 119679, '\p{^Is_Blk=	Counting_Rod}', "");
Expect(0, 119679, '\P{Is_Blk=	Counting_Rod}', "");
Expect(1, 119679, '\P{^Is_Blk=	Counting_Rod}', "");
Expect(0, 119680, '\p{Is_Blk=	Counting_Rod}', "");
Expect(1, 119680, '\p{^Is_Blk=	Counting_Rod}', "");
Expect(1, 119680, '\P{Is_Blk=	Counting_Rod}', "");
Expect(0, 119680, '\P{^Is_Blk=	Counting_Rod}', "");
Error('\p{Block=Cuneiform:=}');
Error('\P{Block=Cuneiform:=}');
Expect(1, 74751, '\p{Block=cuneiform}', "");
Expect(0, 74751, '\p{^Block=cuneiform}', "");
Expect(0, 74751, '\P{Block=cuneiform}', "");
Expect(1, 74751, '\P{^Block=cuneiform}', "");
Expect(0, 74752, '\p{Block=cuneiform}', "");
Expect(1, 74752, '\p{^Block=cuneiform}', "");
Expect(1, 74752, '\P{Block=cuneiform}', "");
Expect(0, 74752, '\P{^Block=cuneiform}', "");
Expect(1, 74751, '\p{Block= CUNEIFORM}', "");
Expect(0, 74751, '\p{^Block= CUNEIFORM}', "");
Expect(0, 74751, '\P{Block= CUNEIFORM}', "");
Expect(1, 74751, '\P{^Block= CUNEIFORM}', "");
Expect(0, 74752, '\p{Block= CUNEIFORM}', "");
Expect(1, 74752, '\p{^Block= CUNEIFORM}', "");
Expect(1, 74752, '\P{Block= CUNEIFORM}', "");
Expect(0, 74752, '\P{^Block= CUNEIFORM}', "");
Error('\p{Blk=:=	Cuneiform}');
Error('\P{Blk=:=	Cuneiform}');
Expect(1, 74751, '\p{Blk=cuneiform}', "");
Expect(0, 74751, '\p{^Blk=cuneiform}', "");
Expect(0, 74751, '\P{Blk=cuneiform}', "");
Expect(1, 74751, '\P{^Blk=cuneiform}', "");
Expect(0, 74752, '\p{Blk=cuneiform}', "");
Expect(1, 74752, '\p{^Blk=cuneiform}', "");
Expect(1, 74752, '\P{Blk=cuneiform}', "");
Expect(0, 74752, '\P{^Blk=cuneiform}', "");
Expect(1, 74751, '\p{Blk= CUNEIFORM}', "");
Expect(0, 74751, '\p{^Blk= CUNEIFORM}', "");
Expect(0, 74751, '\P{Blk= CUNEIFORM}', "");
Expect(1, 74751, '\P{^Blk= CUNEIFORM}', "");
Expect(0, 74752, '\p{Blk= CUNEIFORM}', "");
Expect(1, 74752, '\p{^Blk= CUNEIFORM}', "");
Expect(1, 74752, '\P{Blk= CUNEIFORM}', "");
Expect(0, 74752, '\P{^Blk= CUNEIFORM}', "");
Error('\p{Is_Block=/a/	Cuneiform}');
Error('\P{Is_Block=/a/	Cuneiform}');
Expect(1, 74751, '\p{Is_Block=cuneiform}', "");
Expect(0, 74751, '\p{^Is_Block=cuneiform}', "");
Expect(0, 74751, '\P{Is_Block=cuneiform}', "");
Expect(1, 74751, '\P{^Is_Block=cuneiform}', "");
Expect(0, 74752, '\p{Is_Block=cuneiform}', "");
Expect(1, 74752, '\p{^Is_Block=cuneiform}', "");
Expect(1, 74752, '\P{Is_Block=cuneiform}', "");
Expect(0, 74752, '\P{^Is_Block=cuneiform}', "");
Expect(1, 74751, '\p{Is_Block=_CUNEIFORM}', "");
Expect(0, 74751, '\p{^Is_Block=_CUNEIFORM}', "");
Expect(0, 74751, '\P{Is_Block=_CUNEIFORM}', "");
Expect(1, 74751, '\P{^Is_Block=_CUNEIFORM}', "");
Expect(0, 74752, '\p{Is_Block=_CUNEIFORM}', "");
Expect(1, 74752, '\p{^Is_Block=_CUNEIFORM}', "");
Expect(1, 74752, '\P{Is_Block=_CUNEIFORM}', "");
Expect(0, 74752, '\P{^Is_Block=_CUNEIFORM}', "");
Error('\p{Is_Blk:    _CUNEIFORM/a/}');
Error('\P{Is_Blk:    _CUNEIFORM/a/}');
Expect(1, 74751, '\p{Is_Blk=cuneiform}', "");
Expect(0, 74751, '\p{^Is_Blk=cuneiform}', "");
Expect(0, 74751, '\P{Is_Blk=cuneiform}', "");
Expect(1, 74751, '\P{^Is_Blk=cuneiform}', "");
Expect(0, 74752, '\p{Is_Blk=cuneiform}', "");
Expect(1, 74752, '\p{^Is_Blk=cuneiform}', "");
Expect(1, 74752, '\P{Is_Blk=cuneiform}', "");
Expect(0, 74752, '\P{^Is_Blk=cuneiform}', "");
Expect(1, 74751, '\p{Is_Blk=_ CUNEIFORM}', "");
Expect(0, 74751, '\p{^Is_Blk=_ CUNEIFORM}', "");
Expect(0, 74751, '\P{Is_Blk=_ CUNEIFORM}', "");
Expect(1, 74751, '\P{^Is_Blk=_ CUNEIFORM}', "");
Expect(0, 74752, '\p{Is_Blk=_ CUNEIFORM}', "");
Expect(1, 74752, '\p{^Is_Blk=_ CUNEIFORM}', "");
Expect(1, 74752, '\P{Is_Blk=_ CUNEIFORM}', "");
Expect(0, 74752, '\P{^Is_Blk=_ CUNEIFORM}', "");
Error('\p{Block=	:=Cuneiform_Numbers_And_Punctuation}');
Error('\P{Block=	:=Cuneiform_Numbers_And_Punctuation}');
Expect(1, 74879, '\p{Block:	cuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\p{^Block:	cuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\P{Block:	cuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\P{^Block:	cuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\p{Block:	cuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\p{^Block:	cuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\P{Block:	cuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\P{^Block:	cuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\p{Block=--Cuneiform_NUMBERS_AND_Punctuation}', "");
Expect(0, 74879, '\p{^Block=--Cuneiform_NUMBERS_AND_Punctuation}', "");
Expect(0, 74879, '\P{Block=--Cuneiform_NUMBERS_AND_Punctuation}', "");
Expect(1, 74879, '\P{^Block=--Cuneiform_NUMBERS_AND_Punctuation}', "");
Expect(0, 74880, '\p{Block=--Cuneiform_NUMBERS_AND_Punctuation}', "");
Expect(1, 74880, '\p{^Block=--Cuneiform_NUMBERS_AND_Punctuation}', "");
Expect(1, 74880, '\P{Block=--Cuneiform_NUMBERS_AND_Punctuation}', "");
Expect(0, 74880, '\P{^Block=--Cuneiform_NUMBERS_AND_Punctuation}', "");
Error('\p{Blk=:= _Cuneiform_numbers}');
Error('\P{Blk=:= _Cuneiform_numbers}');
Expect(1, 74879, '\p{Blk=cuneiformnumbers}', "");
Expect(0, 74879, '\p{^Blk=cuneiformnumbers}', "");
Expect(0, 74879, '\P{Blk=cuneiformnumbers}', "");
Expect(1, 74879, '\P{^Blk=cuneiformnumbers}', "");
Expect(0, 74880, '\p{Blk=cuneiformnumbers}', "");
Expect(1, 74880, '\p{^Blk=cuneiformnumbers}', "");
Expect(1, 74880, '\P{Blk=cuneiformnumbers}', "");
Expect(0, 74880, '\P{^Blk=cuneiformnumbers}', "");
Expect(1, 74879, '\p{Blk=-Cuneiform_numbers}', "");
Expect(0, 74879, '\p{^Blk=-Cuneiform_numbers}', "");
Expect(0, 74879, '\P{Blk=-Cuneiform_numbers}', "");
Expect(1, 74879, '\P{^Blk=-Cuneiform_numbers}', "");
Expect(0, 74880, '\p{Blk=-Cuneiform_numbers}', "");
Expect(1, 74880, '\p{^Blk=-Cuneiform_numbers}', "");
Expect(1, 74880, '\P{Blk=-Cuneiform_numbers}', "");
Expect(0, 74880, '\P{^Blk=-Cuneiform_numbers}', "");
Error('\p{Is_Block=:=- Cuneiform_NUMBERS_and_Punctuation}');
Error('\P{Is_Block=:=- Cuneiform_NUMBERS_and_Punctuation}');
Expect(1, 74879, '\p{Is_Block=cuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\p{^Is_Block=cuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\P{Is_Block=cuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\P{^Is_Block=cuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\p{Is_Block=cuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\p{^Is_Block=cuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\P{Is_Block=cuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\P{^Is_Block=cuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\p{Is_Block=- Cuneiform_Numbers_AND_PUNCTUATION}', "");
Expect(0, 74879, '\p{^Is_Block=- Cuneiform_Numbers_AND_PUNCTUATION}', "");
Expect(0, 74879, '\P{Is_Block=- Cuneiform_Numbers_AND_PUNCTUATION}', "");
Expect(1, 74879, '\P{^Is_Block=- Cuneiform_Numbers_AND_PUNCTUATION}', "");
Expect(0, 74880, '\p{Is_Block=- Cuneiform_Numbers_AND_PUNCTUATION}', "");
Expect(1, 74880, '\p{^Is_Block=- Cuneiform_Numbers_AND_PUNCTUATION}', "");
Expect(1, 74880, '\P{Is_Block=- Cuneiform_Numbers_AND_PUNCTUATION}', "");
Expect(0, 74880, '\P{^Is_Block=- Cuneiform_Numbers_AND_PUNCTUATION}', "");
Error('\p{Is_Blk=	 cuneiform_Numbers:=}');
Error('\P{Is_Blk=	 cuneiform_Numbers:=}');
Expect(1, 74879, '\p{Is_Blk=cuneiformnumbers}', "");
Expect(0, 74879, '\p{^Is_Blk=cuneiformnumbers}', "");
Expect(0, 74879, '\P{Is_Blk=cuneiformnumbers}', "");
Expect(1, 74879, '\P{^Is_Blk=cuneiformnumbers}', "");
Expect(0, 74880, '\p{Is_Blk=cuneiformnumbers}', "");
Expect(1, 74880, '\p{^Is_Blk=cuneiformnumbers}', "");
Expect(1, 74880, '\P{Is_Blk=cuneiformnumbers}', "");
Expect(0, 74880, '\P{^Is_Blk=cuneiformnumbers}', "");
Expect(1, 74879, '\p{Is_Blk= _cuneiform_Numbers}', "");
Expect(0, 74879, '\p{^Is_Blk= _cuneiform_Numbers}', "");
Expect(0, 74879, '\P{Is_Blk= _cuneiform_Numbers}', "");
Expect(1, 74879, '\P{^Is_Blk= _cuneiform_Numbers}', "");
Expect(0, 74880, '\p{Is_Blk= _cuneiform_Numbers}', "");
Expect(1, 74880, '\p{^Is_Blk= _cuneiform_Numbers}', "");
Expect(1, 74880, '\P{Is_Blk= _cuneiform_Numbers}', "");
Expect(0, 74880, '\P{^Is_Blk= _cuneiform_Numbers}', "");
Error('\p{Block= Currency_Symbols:=}');
Error('\P{Block= Currency_Symbols:=}');
Expect(1, 8399, '\p{Block=currencysymbols}', "");
Expect(0, 8399, '\p{^Block=currencysymbols}', "");
Expect(0, 8399, '\P{Block=currencysymbols}', "");
Expect(1, 8399, '\P{^Block=currencysymbols}', "");
Expect(0, 8400, '\p{Block=currencysymbols}', "");
Expect(1, 8400, '\p{^Block=currencysymbols}', "");
Expect(1, 8400, '\P{Block=currencysymbols}', "");
Expect(0, 8400, '\P{^Block=currencysymbols}', "");
Expect(1, 8399, '\p{Block:__Currency_Symbols}', "");
Expect(0, 8399, '\p{^Block:__Currency_Symbols}', "");
Expect(0, 8399, '\P{Block:__Currency_Symbols}', "");
Expect(1, 8399, '\P{^Block:__Currency_Symbols}', "");
Expect(0, 8400, '\p{Block:__Currency_Symbols}', "");
Expect(1, 8400, '\p{^Block:__Currency_Symbols}', "");
Expect(1, 8400, '\P{Block:__Currency_Symbols}', "");
Expect(0, 8400, '\P{^Block:__Currency_Symbols}', "");
Error('\p{Blk=_currency_Symbols:=}');
Error('\P{Blk=_currency_Symbols:=}');
Expect(1, 8399, '\p{Blk=currencysymbols}', "");
Expect(0, 8399, '\p{^Blk=currencysymbols}', "");
Expect(0, 8399, '\P{Blk=currencysymbols}', "");
Expect(1, 8399, '\P{^Blk=currencysymbols}', "");
Expect(0, 8400, '\p{Blk=currencysymbols}', "");
Expect(1, 8400, '\p{^Blk=currencysymbols}', "");
Expect(1, 8400, '\P{Blk=currencysymbols}', "");
Expect(0, 8400, '\P{^Blk=currencysymbols}', "");
Expect(1, 8399, '\p{Blk=-Currency_symbols}', "");
Expect(0, 8399, '\p{^Blk=-Currency_symbols}', "");
Expect(0, 8399, '\P{Blk=-Currency_symbols}', "");
Expect(1, 8399, '\P{^Blk=-Currency_symbols}', "");
Expect(0, 8400, '\p{Blk=-Currency_symbols}', "");
Expect(1, 8400, '\p{^Blk=-Currency_symbols}', "");
Expect(1, 8400, '\P{Blk=-Currency_symbols}', "");
Expect(0, 8400, '\P{^Blk=-Currency_symbols}', "");
Error('\p{Is_Block=_/a/CURRENCY_symbols}');
Error('\P{Is_Block=_/a/CURRENCY_symbols}');
Expect(1, 8399, '\p{Is_Block=currencysymbols}', "");
Expect(0, 8399, '\p{^Is_Block=currencysymbols}', "");
Expect(0, 8399, '\P{Is_Block=currencysymbols}', "");
Expect(1, 8399, '\P{^Is_Block=currencysymbols}', "");
Expect(0, 8400, '\p{Is_Block=currencysymbols}', "");
Expect(1, 8400, '\p{^Is_Block=currencysymbols}', "");
Expect(1, 8400, '\P{Is_Block=currencysymbols}', "");
Expect(0, 8400, '\P{^Is_Block=currencysymbols}', "");
Expect(1, 8399, '\p{Is_Block: - currency_symbols}', "");
Expect(0, 8399, '\p{^Is_Block: - currency_symbols}', "");
Expect(0, 8399, '\P{Is_Block: - currency_symbols}', "");
Expect(1, 8399, '\P{^Is_Block: - currency_symbols}', "");
Expect(0, 8400, '\p{Is_Block: - currency_symbols}', "");
Expect(1, 8400, '\p{^Is_Block: - currency_symbols}', "");
Expect(1, 8400, '\P{Is_Block: - currency_symbols}', "");
Expect(0, 8400, '\P{^Is_Block: - currency_symbols}', "");
Error('\p{Is_Blk=/a/-	Currency_Symbols}');
Error('\P{Is_Blk=/a/-	Currency_Symbols}');
Expect(1, 8399, '\p{Is_Blk=currencysymbols}', "");
Expect(0, 8399, '\p{^Is_Blk=currencysymbols}', "");
Expect(0, 8399, '\P{Is_Blk=currencysymbols}', "");
Expect(1, 8399, '\P{^Is_Blk=currencysymbols}', "");
Expect(0, 8400, '\p{Is_Blk=currencysymbols}', "");
Expect(1, 8400, '\p{^Is_Blk=currencysymbols}', "");
Expect(1, 8400, '\P{Is_Blk=currencysymbols}', "");
Expect(0, 8400, '\P{^Is_Blk=currencysymbols}', "");
Expect(1, 8399, '\p{Is_Blk=_	currency_symbols}', "");
Expect(0, 8399, '\p{^Is_Blk=_	currency_symbols}', "");
Expect(0, 8399, '\P{Is_Blk=_	currency_symbols}', "");
Expect(1, 8399, '\P{^Is_Blk=_	currency_symbols}', "");
Expect(0, 8400, '\p{Is_Blk=_	currency_symbols}', "");
Expect(1, 8400, '\p{^Is_Blk=_	currency_symbols}', "");
Expect(1, 8400, '\P{Is_Blk=_	currency_symbols}', "");
Expect(0, 8400, '\P{^Is_Blk=_	currency_symbols}', "");
Error('\p{Block=:=-	CYPRIOT_Syllabary}');
Error('\P{Block=:=-	CYPRIOT_Syllabary}');
Expect(1, 67647, '\p{Block=cypriotsyllabary}', "");
Expect(0, 67647, '\p{^Block=cypriotsyllabary}', "");
Expect(0, 67647, '\P{Block=cypriotsyllabary}', "");
Expect(1, 67647, '\P{^Block=cypriotsyllabary}', "");
Expect(0, 67648, '\p{Block=cypriotsyllabary}', "");
Expect(1, 67648, '\p{^Block=cypriotsyllabary}', "");
Expect(1, 67648, '\P{Block=cypriotsyllabary}', "");
Expect(0, 67648, '\P{^Block=cypriotsyllabary}', "");
Expect(1, 67647, '\p{Block=_Cypriot_syllabary}', "");
Expect(0, 67647, '\p{^Block=_Cypriot_syllabary}', "");
Expect(0, 67647, '\P{Block=_Cypriot_syllabary}', "");
Expect(1, 67647, '\P{^Block=_Cypriot_syllabary}', "");
Expect(0, 67648, '\p{Block=_Cypriot_syllabary}', "");
Expect(1, 67648, '\p{^Block=_Cypriot_syllabary}', "");
Expect(1, 67648, '\P{Block=_Cypriot_syllabary}', "");
Expect(0, 67648, '\P{^Block=_Cypriot_syllabary}', "");
Error('\p{Blk=/a/cypriot_syllabary}');
Error('\P{Blk=/a/cypriot_syllabary}');
Expect(1, 67647, '\p{Blk=cypriotsyllabary}', "");
Expect(0, 67647, '\p{^Blk=cypriotsyllabary}', "");
Expect(0, 67647, '\P{Blk=cypriotsyllabary}', "");
Expect(1, 67647, '\P{^Blk=cypriotsyllabary}', "");
Expect(0, 67648, '\p{Blk=cypriotsyllabary}', "");
Expect(1, 67648, '\p{^Blk=cypriotsyllabary}', "");
Expect(1, 67648, '\P{Blk=cypriotsyllabary}', "");
Expect(0, 67648, '\P{^Blk=cypriotsyllabary}', "");
Expect(1, 67647, '\p{Blk=_	Cypriot_Syllabary}', "");
Expect(0, 67647, '\p{^Blk=_	Cypriot_Syllabary}', "");
Expect(0, 67647, '\P{Blk=_	Cypriot_Syllabary}', "");
Expect(1, 67647, '\P{^Blk=_	Cypriot_Syllabary}', "");
Expect(0, 67648, '\p{Blk=_	Cypriot_Syllabary}', "");
Expect(1, 67648, '\p{^Blk=_	Cypriot_Syllabary}', "");
Expect(1, 67648, '\P{Blk=_	Cypriot_Syllabary}', "");
Expect(0, 67648, '\P{^Blk=_	Cypriot_Syllabary}', "");
Error('\p{Is_Block= -Cypriot_SYLLABARY/a/}');
Error('\P{Is_Block= -Cypriot_SYLLABARY/a/}');
Expect(1, 67647, '\p{Is_Block=cypriotsyllabary}', "");
Expect(0, 67647, '\p{^Is_Block=cypriotsyllabary}', "");
Expect(0, 67647, '\P{Is_Block=cypriotsyllabary}', "");
Expect(1, 67647, '\P{^Is_Block=cypriotsyllabary}', "");
Expect(0, 67648, '\p{Is_Block=cypriotsyllabary}', "");
Expect(1, 67648, '\p{^Is_Block=cypriotsyllabary}', "");
Expect(1, 67648, '\P{Is_Block=cypriotsyllabary}', "");
Expect(0, 67648, '\P{^Is_Block=cypriotsyllabary}', "");
Expect(1, 67647, '\p{Is_Block= Cypriot_SYLLABARY}', "");
Expect(0, 67647, '\p{^Is_Block= Cypriot_SYLLABARY}', "");
Expect(0, 67647, '\P{Is_Block= Cypriot_SYLLABARY}', "");
Expect(1, 67647, '\P{^Is_Block= Cypriot_SYLLABARY}', "");
Expect(0, 67648, '\p{Is_Block= Cypriot_SYLLABARY}', "");
Expect(1, 67648, '\p{^Is_Block= Cypriot_SYLLABARY}', "");
Expect(1, 67648, '\P{Is_Block= Cypriot_SYLLABARY}', "");
Expect(0, 67648, '\P{^Is_Block= Cypriot_SYLLABARY}', "");
Error('\p{Is_Blk: :=	-CYPRIOT_syllabary}');
Error('\P{Is_Blk: :=	-CYPRIOT_syllabary}');
Expect(1, 67647, '\p{Is_Blk:   cypriotsyllabary}', "");
Expect(0, 67647, '\p{^Is_Blk:   cypriotsyllabary}', "");
Expect(0, 67647, '\P{Is_Blk:   cypriotsyllabary}', "");
Expect(1, 67647, '\P{^Is_Blk:   cypriotsyllabary}', "");
Expect(0, 67648, '\p{Is_Blk:   cypriotsyllabary}', "");
Expect(1, 67648, '\p{^Is_Blk:   cypriotsyllabary}', "");
Expect(1, 67648, '\P{Is_Blk:   cypriotsyllabary}', "");
Expect(0, 67648, '\P{^Is_Blk:   cypriotsyllabary}', "");
Expect(1, 67647, '\p{Is_Blk=_cypriot_Syllabary}', "");
Expect(0, 67647, '\p{^Is_Blk=_cypriot_Syllabary}', "");
Expect(0, 67647, '\P{Is_Blk=_cypriot_Syllabary}', "");
Expect(1, 67647, '\P{^Is_Blk=_cypriot_Syllabary}', "");
Expect(0, 67648, '\p{Is_Blk=_cypriot_Syllabary}', "");
Expect(1, 67648, '\p{^Is_Blk=_cypriot_Syllabary}', "");
Expect(1, 67648, '\P{Is_Blk=_cypriot_Syllabary}', "");
Expect(0, 67648, '\P{^Is_Blk=_cypriot_Syllabary}', "");
Error('\p{Block= :=Cyrillic}');
Error('\P{Block= :=Cyrillic}');
Expect(1, 1279, '\p{Block=cyrillic}', "");
Expect(0, 1279, '\p{^Block=cyrillic}', "");
Expect(0, 1279, '\P{Block=cyrillic}', "");
Expect(1, 1279, '\P{^Block=cyrillic}', "");
Expect(0, 1280, '\p{Block=cyrillic}', "");
Expect(1, 1280, '\p{^Block=cyrillic}', "");
Expect(1, 1280, '\P{Block=cyrillic}', "");
Expect(0, 1280, '\P{^Block=cyrillic}', "");
Expect(1, 1279, '\p{Block=	Cyrillic}', "");
Expect(0, 1279, '\p{^Block=	Cyrillic}', "");
Expect(0, 1279, '\P{Block=	Cyrillic}', "");
Expect(1, 1279, '\P{^Block=	Cyrillic}', "");
Expect(0, 1280, '\p{Block=	Cyrillic}', "");
Expect(1, 1280, '\p{^Block=	Cyrillic}', "");
Expect(1, 1280, '\P{Block=	Cyrillic}', "");
Expect(0, 1280, '\P{^Block=	Cyrillic}', "");
Error('\p{Blk=- Cyrillic/a/}');
Error('\P{Blk=- Cyrillic/a/}');
Expect(1, 1279, '\p{Blk:	cyrillic}', "");
Expect(0, 1279, '\p{^Blk:	cyrillic}', "");
Expect(0, 1279, '\P{Blk:	cyrillic}', "");
Expect(1, 1279, '\P{^Blk:	cyrillic}', "");
Expect(0, 1280, '\p{Blk:	cyrillic}', "");
Expect(1, 1280, '\p{^Blk:	cyrillic}', "");
Expect(1, 1280, '\P{Blk:	cyrillic}', "");
Expect(0, 1280, '\P{^Blk:	cyrillic}', "");
Expect(1, 1279, '\p{Blk=	_Cyrillic}', "");
Expect(0, 1279, '\p{^Blk=	_Cyrillic}', "");
Expect(0, 1279, '\P{Blk=	_Cyrillic}', "");
Expect(1, 1279, '\P{^Blk=	_Cyrillic}', "");
Expect(0, 1280, '\p{Blk=	_Cyrillic}', "");
Expect(1, 1280, '\p{^Blk=	_Cyrillic}', "");
Expect(1, 1280, '\P{Blk=	_Cyrillic}', "");
Expect(0, 1280, '\P{^Blk=	_Cyrillic}', "");
Error('\p{Is_Block=:=	_CYRILLIC}');
Error('\P{Is_Block=:=	_CYRILLIC}');
Expect(1, 1279, '\p{Is_Block=cyrillic}', "");
Expect(0, 1279, '\p{^Is_Block=cyrillic}', "");
Expect(0, 1279, '\P{Is_Block=cyrillic}', "");
Expect(1, 1279, '\P{^Is_Block=cyrillic}', "");
Expect(0, 1280, '\p{Is_Block=cyrillic}', "");
Expect(1, 1280, '\p{^Is_Block=cyrillic}', "");
Expect(1, 1280, '\P{Is_Block=cyrillic}', "");
Expect(0, 1280, '\P{^Is_Block=cyrillic}', "");
Expect(1, 1279, '\p{Is_Block:	Cyrillic}', "");
Expect(0, 1279, '\p{^Is_Block:	Cyrillic}', "");
Expect(0, 1279, '\P{Is_Block:	Cyrillic}', "");
Expect(1, 1279, '\P{^Is_Block:	Cyrillic}', "");
Expect(0, 1280, '\p{Is_Block:	Cyrillic}', "");
Expect(1, 1280, '\p{^Is_Block:	Cyrillic}', "");
Expect(1, 1280, '\P{Is_Block:	Cyrillic}', "");
Expect(0, 1280, '\P{^Is_Block:	Cyrillic}', "");
Error('\p{Is_Blk:	/a/cyrillic}');
Error('\P{Is_Blk:	/a/cyrillic}');
Expect(1, 1279, '\p{Is_Blk=cyrillic}', "");
Expect(0, 1279, '\p{^Is_Blk=cyrillic}', "");
Expect(0, 1279, '\P{Is_Blk=cyrillic}', "");
Expect(1, 1279, '\P{^Is_Blk=cyrillic}', "");
Expect(0, 1280, '\p{Is_Blk=cyrillic}', "");
Expect(1, 1280, '\p{^Is_Blk=cyrillic}', "");
Expect(1, 1280, '\P{Is_Blk=cyrillic}', "");
Expect(0, 1280, '\P{^Is_Blk=cyrillic}', "");
Expect(1, 1279, '\p{Is_Blk= CYRILLIC}', "");
Expect(0, 1279, '\p{^Is_Blk= CYRILLIC}', "");
Expect(0, 1279, '\P{Is_Blk= CYRILLIC}', "");
Expect(1, 1279, '\P{^Is_Blk= CYRILLIC}', "");
Expect(0, 1280, '\p{Is_Blk= CYRILLIC}', "");
Expect(1, 1280, '\p{^Is_Blk= CYRILLIC}', "");
Expect(1, 1280, '\P{Is_Blk= CYRILLIC}', "");
Expect(0, 1280, '\P{^Is_Blk= CYRILLIC}', "");
Error('\p{Block:		Cyrillic_EXTENDED_A:=}');
Error('\P{Block:		Cyrillic_EXTENDED_A:=}');
Expect(1, 11775, '\p{Block:cyrillicextendeda}', "");
Expect(0, 11775, '\p{^Block:cyrillicextendeda}', "");
Expect(0, 11775, '\P{Block:cyrillicextendeda}', "");
Expect(1, 11775, '\P{^Block:cyrillicextendeda}', "");
Expect(0, 11776, '\p{Block:cyrillicextendeda}', "");
Expect(1, 11776, '\p{^Block:cyrillicextendeda}', "");
Expect(1, 11776, '\P{Block:cyrillicextendeda}', "");
Expect(0, 11776, '\P{^Block:cyrillicextendeda}', "");
Expect(1, 11775, '\p{Block=_	cyrillic_Extended_A}', "");
Expect(0, 11775, '\p{^Block=_	cyrillic_Extended_A}', "");
Expect(0, 11775, '\P{Block=_	cyrillic_Extended_A}', "");
Expect(1, 11775, '\P{^Block=_	cyrillic_Extended_A}', "");
Expect(0, 11776, '\p{Block=_	cyrillic_Extended_A}', "");
Expect(1, 11776, '\p{^Block=_	cyrillic_Extended_A}', "");
Expect(1, 11776, '\P{Block=_	cyrillic_Extended_A}', "");
Expect(0, 11776, '\P{^Block=_	cyrillic_Extended_A}', "");
Error('\p{Blk=-_Cyrillic_ext_A/a/}');
Error('\P{Blk=-_Cyrillic_ext_A/a/}');
Expect(1, 11775, '\p{Blk:	cyrillicexta}', "");
Expect(0, 11775, '\p{^Blk:	cyrillicexta}', "");
Expect(0, 11775, '\P{Blk:	cyrillicexta}', "");
Expect(1, 11775, '\P{^Blk:	cyrillicexta}', "");
Expect(0, 11776, '\p{Blk:	cyrillicexta}', "");
Expect(1, 11776, '\p{^Blk:	cyrillicexta}', "");
Expect(1, 11776, '\P{Blk:	cyrillicexta}', "");
Expect(0, 11776, '\P{^Blk:	cyrillicexta}', "");
Expect(1, 11775, '\p{Blk:  -Cyrillic_Ext_a}', "");
Expect(0, 11775, '\p{^Blk:  -Cyrillic_Ext_a}', "");
Expect(0, 11775, '\P{Blk:  -Cyrillic_Ext_a}', "");
Expect(1, 11775, '\P{^Blk:  -Cyrillic_Ext_a}', "");
Expect(0, 11776, '\p{Blk:  -Cyrillic_Ext_a}', "");
Expect(1, 11776, '\p{^Blk:  -Cyrillic_Ext_a}', "");
Expect(1, 11776, '\P{Blk:  -Cyrillic_Ext_a}', "");
Expect(0, 11776, '\P{^Blk:  -Cyrillic_Ext_a}', "");
Error('\p{Is_Block=_/a/Cyrillic_extended_A}');
Error('\P{Is_Block=_/a/Cyrillic_extended_A}');
Expect(1, 11775, '\p{Is_Block:   cyrillicextendeda}', "");
Expect(0, 11775, '\p{^Is_Block:   cyrillicextendeda}', "");
Expect(0, 11775, '\P{Is_Block:   cyrillicextendeda}', "");
Expect(1, 11775, '\P{^Is_Block:   cyrillicextendeda}', "");
Expect(0, 11776, '\p{Is_Block:   cyrillicextendeda}', "");
Expect(1, 11776, '\p{^Is_Block:   cyrillicextendeda}', "");
Expect(1, 11776, '\P{Is_Block:   cyrillicextendeda}', "");
Expect(0, 11776, '\P{^Is_Block:   cyrillicextendeda}', "");
Expect(1, 11775, '\p{Is_Block=_-Cyrillic_Extended_A}', "");
Expect(0, 11775, '\p{^Is_Block=_-Cyrillic_Extended_A}', "");
Expect(0, 11775, '\P{Is_Block=_-Cyrillic_Extended_A}', "");
Expect(1, 11775, '\P{^Is_Block=_-Cyrillic_Extended_A}', "");
Expect(0, 11776, '\p{Is_Block=_-Cyrillic_Extended_A}', "");
Expect(1, 11776, '\p{^Is_Block=_-Cyrillic_Extended_A}', "");
Expect(1, 11776, '\P{Is_Block=_-Cyrillic_Extended_A}', "");
Expect(0, 11776, '\P{^Is_Block=_-Cyrillic_Extended_A}', "");
Error('\p{Is_Blk: :=  Cyrillic_ext_a}');
Error('\P{Is_Blk: :=  Cyrillic_ext_a}');
Expect(1, 11775, '\p{Is_Blk=cyrillicexta}', "");
Expect(0, 11775, '\p{^Is_Blk=cyrillicexta}', "");
Expect(0, 11775, '\P{Is_Blk=cyrillicexta}', "");
Expect(1, 11775, '\P{^Is_Blk=cyrillicexta}', "");
Expect(0, 11776, '\p{Is_Blk=cyrillicexta}', "");
Expect(1, 11776, '\p{^Is_Blk=cyrillicexta}', "");
Expect(1, 11776, '\P{Is_Blk=cyrillicexta}', "");
Expect(0, 11776, '\P{^Is_Blk=cyrillicexta}', "");
Expect(1, 11775, '\p{Is_Blk=CYRILLIC_Ext_A}', "");
Expect(0, 11775, '\p{^Is_Blk=CYRILLIC_Ext_A}', "");
Expect(0, 11775, '\P{Is_Blk=CYRILLIC_Ext_A}', "");
Expect(1, 11775, '\P{^Is_Blk=CYRILLIC_Ext_A}', "");
Expect(0, 11776, '\p{Is_Blk=CYRILLIC_Ext_A}', "");
Expect(1, 11776, '\p{^Is_Blk=CYRILLIC_Ext_A}', "");
Expect(1, 11776, '\P{Is_Blk=CYRILLIC_Ext_A}', "");
Expect(0, 11776, '\P{^Is_Blk=CYRILLIC_Ext_A}', "");
Error('\p{Block= /a/Cyrillic_Extended_B}');
Error('\P{Block= /a/Cyrillic_Extended_B}');
Expect(1, 42655, '\p{Block=cyrillicextendedb}', "");
Expect(0, 42655, '\p{^Block=cyrillicextendedb}', "");
Expect(0, 42655, '\P{Block=cyrillicextendedb}', "");
Expect(1, 42655, '\P{^Block=cyrillicextendedb}', "");
Expect(0, 42656, '\p{Block=cyrillicextendedb}', "");
Expect(1, 42656, '\p{^Block=cyrillicextendedb}', "");
Expect(1, 42656, '\P{Block=cyrillicextendedb}', "");
Expect(0, 42656, '\P{^Block=cyrillicextendedb}', "");
Expect(1, 42655, '\p{Block= Cyrillic_Extended_b}', "");
Expect(0, 42655, '\p{^Block= Cyrillic_Extended_b}', "");
Expect(0, 42655, '\P{Block= Cyrillic_Extended_b}', "");
Expect(1, 42655, '\P{^Block= Cyrillic_Extended_b}', "");
Expect(0, 42656, '\p{Block= Cyrillic_Extended_b}', "");
Expect(1, 42656, '\p{^Block= Cyrillic_Extended_b}', "");
Expect(1, 42656, '\P{Block= Cyrillic_Extended_b}', "");
Expect(0, 42656, '\P{^Block= Cyrillic_Extended_b}', "");
Error('\p{Blk=:=Cyrillic_Ext_B}');
Error('\P{Blk=:=Cyrillic_Ext_B}');
Expect(1, 42655, '\p{Blk=cyrillicextb}', "");
Expect(0, 42655, '\p{^Blk=cyrillicextb}', "");
Expect(0, 42655, '\P{Blk=cyrillicextb}', "");
Expect(1, 42655, '\P{^Blk=cyrillicextb}', "");
Expect(0, 42656, '\p{Blk=cyrillicextb}', "");
Expect(1, 42656, '\p{^Blk=cyrillicextb}', "");
Expect(1, 42656, '\P{Blk=cyrillicextb}', "");
Expect(0, 42656, '\P{^Blk=cyrillicextb}', "");
Expect(1, 42655, '\p{Blk=_-Cyrillic_ext_b}', "");
Expect(0, 42655, '\p{^Blk=_-Cyrillic_ext_b}', "");
Expect(0, 42655, '\P{Blk=_-Cyrillic_ext_b}', "");
Expect(1, 42655, '\P{^Blk=_-Cyrillic_ext_b}', "");
Expect(0, 42656, '\p{Blk=_-Cyrillic_ext_b}', "");
Expect(1, 42656, '\p{^Blk=_-Cyrillic_ext_b}', "");
Expect(1, 42656, '\P{Blk=_-Cyrillic_ext_b}', "");
Expect(0, 42656, '\P{^Blk=_-Cyrillic_ext_b}', "");
Error('\p{Is_Block=_CYRILLIC_EXTENDED_B/a/}');
Error('\P{Is_Block=_CYRILLIC_EXTENDED_B/a/}');
Expect(1, 42655, '\p{Is_Block=cyrillicextendedb}', "");
Expect(0, 42655, '\p{^Is_Block=cyrillicextendedb}', "");
Expect(0, 42655, '\P{Is_Block=cyrillicextendedb}', "");
Expect(1, 42655, '\P{^Is_Block=cyrillicextendedb}', "");
Expect(0, 42656, '\p{Is_Block=cyrillicextendedb}', "");
Expect(1, 42656, '\p{^Is_Block=cyrillicextendedb}', "");
Expect(1, 42656, '\P{Is_Block=cyrillicextendedb}', "");
Expect(0, 42656, '\P{^Is_Block=cyrillicextendedb}', "");
Expect(1, 42655, '\p{Is_Block=-CYRILLIC_extended_B}', "");
Expect(0, 42655, '\p{^Is_Block=-CYRILLIC_extended_B}', "");
Expect(0, 42655, '\P{Is_Block=-CYRILLIC_extended_B}', "");
Expect(1, 42655, '\P{^Is_Block=-CYRILLIC_extended_B}', "");
Expect(0, 42656, '\p{Is_Block=-CYRILLIC_extended_B}', "");
Expect(1, 42656, '\p{^Is_Block=-CYRILLIC_extended_B}', "");
Expect(1, 42656, '\P{Is_Block=-CYRILLIC_extended_B}', "");
Expect(0, 42656, '\P{^Is_Block=-CYRILLIC_extended_B}', "");
Error('\p{Is_Blk=	:=cyrillic_ext_B}');
Error('\P{Is_Blk=	:=cyrillic_ext_B}');
Expect(1, 42655, '\p{Is_Blk=cyrillicextb}', "");
Expect(0, 42655, '\p{^Is_Blk=cyrillicextb}', "");
Expect(0, 42655, '\P{Is_Blk=cyrillicextb}', "");
Expect(1, 42655, '\P{^Is_Blk=cyrillicextb}', "");
Expect(0, 42656, '\p{Is_Blk=cyrillicextb}', "");
Expect(1, 42656, '\p{^Is_Blk=cyrillicextb}', "");
Expect(1, 42656, '\P{Is_Blk=cyrillicextb}', "");
Expect(0, 42656, '\P{^Is_Blk=cyrillicextb}', "");
Expect(1, 42655, '\p{Is_Blk=__Cyrillic_EXT_B}', "");
Expect(0, 42655, '\p{^Is_Blk=__Cyrillic_EXT_B}', "");
Expect(0, 42655, '\P{Is_Blk=__Cyrillic_EXT_B}', "");
Expect(1, 42655, '\P{^Is_Blk=__Cyrillic_EXT_B}', "");
Expect(0, 42656, '\p{Is_Blk=__Cyrillic_EXT_B}', "");
Expect(1, 42656, '\p{^Is_Blk=__Cyrillic_EXT_B}', "");
Expect(1, 42656, '\P{Is_Blk=__Cyrillic_EXT_B}', "");
Expect(0, 42656, '\P{^Is_Blk=__Cyrillic_EXT_B}', "");
Error('\p{Block=:=	-cyrillic_supplement}');
Error('\P{Block=:=	-cyrillic_supplement}');
Expect(1, 1327, '\p{Block: cyrillicsupplement}', "");
Expect(0, 1327, '\p{^Block: cyrillicsupplement}', "");
Expect(0, 1327, '\P{Block: cyrillicsupplement}', "");
Expect(1, 1327, '\P{^Block: cyrillicsupplement}', "");
Expect(0, 1328, '\p{Block: cyrillicsupplement}', "");
Expect(1, 1328, '\p{^Block: cyrillicsupplement}', "");
Expect(1, 1328, '\P{Block: cyrillicsupplement}', "");
Expect(0, 1328, '\P{^Block: cyrillicsupplement}', "");
Expect(1, 1327, '\p{Block=	CYRILLIC_Supplement}', "");
Expect(0, 1327, '\p{^Block=	CYRILLIC_Supplement}', "");
Expect(0, 1327, '\P{Block=	CYRILLIC_Supplement}', "");
Expect(1, 1327, '\P{^Block=	CYRILLIC_Supplement}', "");
Expect(0, 1328, '\p{Block=	CYRILLIC_Supplement}', "");
Expect(1, 1328, '\p{^Block=	CYRILLIC_Supplement}', "");
Expect(1, 1328, '\P{Block=	CYRILLIC_Supplement}', "");
Expect(0, 1328, '\P{^Block=	CYRILLIC_Supplement}', "");
Error('\p{Blk:	 Cyrillic_SUP:=}');
Error('\P{Blk:	 Cyrillic_SUP:=}');
Expect(1, 1327, '\p{Blk:cyrillicsup}', "");
Expect(0, 1327, '\p{^Blk:cyrillicsup}', "");
Expect(0, 1327, '\P{Blk:cyrillicsup}', "");
Expect(1, 1327, '\P{^Blk:cyrillicsup}', "");
Expect(0, 1328, '\p{Blk:cyrillicsup}', "");
Expect(1, 1328, '\p{^Blk:cyrillicsup}', "");
Expect(1, 1328, '\P{Blk:cyrillicsup}', "");
Expect(0, 1328, '\P{^Blk:cyrillicsup}', "");
Expect(1, 1327, '\p{Blk= _Cyrillic_Sup}', "");
Expect(0, 1327, '\p{^Blk= _Cyrillic_Sup}', "");
Expect(0, 1327, '\P{Blk= _Cyrillic_Sup}', "");
Expect(1, 1327, '\P{^Blk= _Cyrillic_Sup}', "");
Expect(0, 1328, '\p{Blk= _Cyrillic_Sup}', "");
Expect(1, 1328, '\p{^Blk= _Cyrillic_Sup}', "");
Expect(1, 1328, '\P{Blk= _Cyrillic_Sup}', "");
Expect(0, 1328, '\P{^Blk= _Cyrillic_Sup}', "");
Error('\p{Is_Block=:=-	Cyrillic_Supplementary}');
Error('\P{Is_Block=:=-	Cyrillic_Supplementary}');
Expect(1, 1327, '\p{Is_Block=cyrillicsupplementary}', "");
Expect(0, 1327, '\p{^Is_Block=cyrillicsupplementary}', "");
Expect(0, 1327, '\P{Is_Block=cyrillicsupplementary}', "");
Expect(1, 1327, '\P{^Is_Block=cyrillicsupplementary}', "");
Expect(0, 1328, '\p{Is_Block=cyrillicsupplementary}', "");
Expect(1, 1328, '\p{^Is_Block=cyrillicsupplementary}', "");
Expect(1, 1328, '\P{Is_Block=cyrillicsupplementary}', "");
Expect(0, 1328, '\P{^Is_Block=cyrillicsupplementary}', "");
Expect(1, 1327, '\p{Is_Block=_Cyrillic_supplementary}', "");
Expect(0, 1327, '\p{^Is_Block=_Cyrillic_supplementary}', "");
Expect(0, 1327, '\P{Is_Block=_Cyrillic_supplementary}', "");
Expect(1, 1327, '\P{^Is_Block=_Cyrillic_supplementary}', "");
Expect(0, 1328, '\p{Is_Block=_Cyrillic_supplementary}', "");
Expect(1, 1328, '\p{^Is_Block=_Cyrillic_supplementary}', "");
Expect(1, 1328, '\P{Is_Block=_Cyrillic_supplementary}', "");
Expect(0, 1328, '\P{^Is_Block=_Cyrillic_supplementary}', "");
Error('\p{Is_Blk=_/a/Cyrillic_supplement}');
Error('\P{Is_Blk=_/a/Cyrillic_supplement}');
Expect(1, 1327, '\p{Is_Blk=cyrillicsupplement}', "");
Expect(0, 1327, '\p{^Is_Blk=cyrillicsupplement}', "");
Expect(0, 1327, '\P{Is_Blk=cyrillicsupplement}', "");
Expect(1, 1327, '\P{^Is_Blk=cyrillicsupplement}', "");
Expect(0, 1328, '\p{Is_Blk=cyrillicsupplement}', "");
Expect(1, 1328, '\p{^Is_Blk=cyrillicsupplement}', "");
Expect(1, 1328, '\P{Is_Blk=cyrillicsupplement}', "");
Expect(0, 1328, '\P{^Is_Blk=cyrillicsupplement}', "");
Expect(1, 1327, '\p{Is_Blk=_Cyrillic_supplement}', "");
Expect(0, 1327, '\p{^Is_Blk=_Cyrillic_supplement}', "");
Expect(0, 1327, '\P{Is_Blk=_Cyrillic_supplement}', "");
Expect(1, 1327, '\P{^Is_Blk=_Cyrillic_supplement}', "");
Expect(0, 1328, '\p{Is_Blk=_Cyrillic_supplement}', "");
Expect(1, 1328, '\p{^Is_Blk=_Cyrillic_supplement}', "");
Expect(1, 1328, '\P{Is_Blk=_Cyrillic_supplement}', "");
Expect(0, 1328, '\P{^Is_Blk=_Cyrillic_supplement}', "");
Error('\p{Block:   :=--Deseret}');
Error('\P{Block:   :=--Deseret}');
Expect(1, 66639, '\p{Block=deseret}', "");
Expect(0, 66639, '\p{^Block=deseret}', "");
Expect(0, 66639, '\P{Block=deseret}', "");
Expect(1, 66639, '\P{^Block=deseret}', "");
Expect(0, 66640, '\p{Block=deseret}', "");
Expect(1, 66640, '\p{^Block=deseret}', "");
Expect(1, 66640, '\P{Block=deseret}', "");
Expect(0, 66640, '\P{^Block=deseret}', "");
Expect(1, 66639, '\p{Block= DESERET}', "");
Expect(0, 66639, '\p{^Block= DESERET}', "");
Expect(0, 66639, '\P{Block= DESERET}', "");
Expect(1, 66639, '\P{^Block= DESERET}', "");
Expect(0, 66640, '\p{Block= DESERET}', "");
Expect(1, 66640, '\p{^Block= DESERET}', "");
Expect(1, 66640, '\P{Block= DESERET}', "");
Expect(0, 66640, '\P{^Block= DESERET}', "");
Error('\p{Blk= _Deseret:=}');
Error('\P{Blk= _Deseret:=}');
Expect(1, 66639, '\p{Blk=deseret}', "");
Expect(0, 66639, '\p{^Blk=deseret}', "");
Expect(0, 66639, '\P{Blk=deseret}', "");
Expect(1, 66639, '\P{^Blk=deseret}', "");
Expect(0, 66640, '\p{Blk=deseret}', "");
Expect(1, 66640, '\p{^Blk=deseret}', "");
Expect(1, 66640, '\P{Blk=deseret}', "");
Expect(0, 66640, '\P{^Blk=deseret}', "");
Expect(1, 66639, '\p{Blk=	 Deseret}', "");
Expect(0, 66639, '\p{^Blk=	 Deseret}', "");
Expect(0, 66639, '\P{Blk=	 Deseret}', "");
Expect(1, 66639, '\P{^Blk=	 Deseret}', "");
Expect(0, 66640, '\p{Blk=	 Deseret}', "");
Expect(1, 66640, '\p{^Blk=	 Deseret}', "");
Expect(1, 66640, '\P{Blk=	 Deseret}', "");
Expect(0, 66640, '\P{^Blk=	 Deseret}', "");
Error('\p{Is_Block=/a/Deseret}');
Error('\P{Is_Block=/a/Deseret}');
Expect(1, 66639, '\p{Is_Block=deseret}', "");
Expect(0, 66639, '\p{^Is_Block=deseret}', "");
Expect(0, 66639, '\P{Is_Block=deseret}', "");
Expect(1, 66639, '\P{^Is_Block=deseret}', "");
Expect(0, 66640, '\p{Is_Block=deseret}', "");
Expect(1, 66640, '\p{^Is_Block=deseret}', "");
Expect(1, 66640, '\P{Is_Block=deseret}', "");
Expect(0, 66640, '\P{^Is_Block=deseret}', "");
Expect(1, 66639, '\p{Is_Block=_-Deseret}', "");
Expect(0, 66639, '\p{^Is_Block=_-Deseret}', "");
Expect(0, 66639, '\P{Is_Block=_-Deseret}', "");
Expect(1, 66639, '\P{^Is_Block=_-Deseret}', "");
Expect(0, 66640, '\p{Is_Block=_-Deseret}', "");
Expect(1, 66640, '\p{^Is_Block=_-Deseret}', "");
Expect(1, 66640, '\P{Is_Block=_-Deseret}', "");
Expect(0, 66640, '\P{^Is_Block=_-Deseret}', "");
Error('\p{Is_Blk= :=Deseret}');
Error('\P{Is_Blk= :=Deseret}');
Expect(1, 66639, '\p{Is_Blk=deseret}', "");
Expect(0, 66639, '\p{^Is_Blk=deseret}', "");
Expect(0, 66639, '\P{Is_Blk=deseret}', "");
Expect(1, 66639, '\P{^Is_Blk=deseret}', "");
Expect(0, 66640, '\p{Is_Blk=deseret}', "");
Expect(1, 66640, '\p{^Is_Blk=deseret}', "");
Expect(1, 66640, '\P{Is_Blk=deseret}', "");
Expect(0, 66640, '\P{^Is_Blk=deseret}', "");
Expect(1, 66639, '\p{Is_Blk=  deseret}', "");
Expect(0, 66639, '\p{^Is_Blk=  deseret}', "");
Expect(0, 66639, '\P{Is_Blk=  deseret}', "");
Expect(1, 66639, '\P{^Is_Blk=  deseret}', "");
Expect(0, 66640, '\p{Is_Blk=  deseret}', "");
Expect(1, 66640, '\p{^Is_Blk=  deseret}', "");
Expect(1, 66640, '\P{Is_Blk=  deseret}', "");
Expect(0, 66640, '\P{^Is_Blk=  deseret}', "");
Error('\p{Block=_Devanagari/a/}');
Error('\P{Block=_Devanagari/a/}');
Expect(1, 2431, '\p{Block=devanagari}', "");
Expect(0, 2431, '\p{^Block=devanagari}', "");
Expect(0, 2431, '\P{Block=devanagari}', "");
Expect(1, 2431, '\P{^Block=devanagari}', "");
Expect(0, 2432, '\p{Block=devanagari}', "");
Expect(1, 2432, '\p{^Block=devanagari}', "");
Expect(1, 2432, '\P{Block=devanagari}', "");
Expect(0, 2432, '\P{^Block=devanagari}', "");
Expect(1, 2431, '\p{Block= -Devanagari}', "");
Expect(0, 2431, '\p{^Block= -Devanagari}', "");
Expect(0, 2431, '\P{Block= -Devanagari}', "");
Expect(1, 2431, '\P{^Block= -Devanagari}', "");
Expect(0, 2432, '\p{Block= -Devanagari}', "");
Expect(1, 2432, '\p{^Block= -Devanagari}', "");
Expect(1, 2432, '\P{Block= -Devanagari}', "");
Expect(0, 2432, '\P{^Block= -Devanagari}', "");
Error('\p{Blk=_Devanagari:=}');
Error('\P{Blk=_Devanagari:=}');
Expect(1, 2431, '\p{Blk=devanagari}', "");
Expect(0, 2431, '\p{^Blk=devanagari}', "");
Expect(0, 2431, '\P{Blk=devanagari}', "");
Expect(1, 2431, '\P{^Blk=devanagari}', "");
Expect(0, 2432, '\p{Blk=devanagari}', "");
Expect(1, 2432, '\p{^Blk=devanagari}', "");
Expect(1, 2432, '\P{Blk=devanagari}', "");
Expect(0, 2432, '\P{^Blk=devanagari}', "");
Expect(1, 2431, '\p{Blk= DEVANAGARI}', "");
Expect(0, 2431, '\p{^Blk= DEVANAGARI}', "");
Expect(0, 2431, '\P{Blk= DEVANAGARI}', "");
Expect(1, 2431, '\P{^Blk= DEVANAGARI}', "");
Expect(0, 2432, '\p{Blk= DEVANAGARI}', "");
Expect(1, 2432, '\p{^Blk= DEVANAGARI}', "");
Expect(1, 2432, '\P{Blk= DEVANAGARI}', "");
Expect(0, 2432, '\P{^Blk= DEVANAGARI}', "");
Error('\p{Is_Block= :=devanagari}');
Error('\P{Is_Block= :=devanagari}');
Expect(1, 2431, '\p{Is_Block=devanagari}', "");
Expect(0, 2431, '\p{^Is_Block=devanagari}', "");
Expect(0, 2431, '\P{Is_Block=devanagari}', "");
Expect(1, 2431, '\P{^Is_Block=devanagari}', "");
Expect(0, 2432, '\p{Is_Block=devanagari}', "");
Expect(1, 2432, '\p{^Is_Block=devanagari}', "");
Expect(1, 2432, '\P{Is_Block=devanagari}', "");
Expect(0, 2432, '\P{^Is_Block=devanagari}', "");
Expect(1, 2431, '\p{Is_Block=_	DEVANAGARI}', "");
Expect(0, 2431, '\p{^Is_Block=_	DEVANAGARI}', "");
Expect(0, 2431, '\P{Is_Block=_	DEVANAGARI}', "");
Expect(1, 2431, '\P{^Is_Block=_	DEVANAGARI}', "");
Expect(0, 2432, '\p{Is_Block=_	DEVANAGARI}', "");
Expect(1, 2432, '\p{^Is_Block=_	DEVANAGARI}', "");
Expect(1, 2432, '\P{Is_Block=_	DEVANAGARI}', "");
Expect(0, 2432, '\P{^Is_Block=_	DEVANAGARI}', "");
Error('\p{Is_Blk=/a/-DEVANAGARI}');
Error('\P{Is_Blk=/a/-DEVANAGARI}');
Expect(1, 2431, '\p{Is_Blk=devanagari}', "");
Expect(0, 2431, '\p{^Is_Blk=devanagari}', "");
Expect(0, 2431, '\P{Is_Blk=devanagari}', "");
Expect(1, 2431, '\P{^Is_Blk=devanagari}', "");
Expect(0, 2432, '\p{Is_Blk=devanagari}', "");
Expect(1, 2432, '\p{^Is_Blk=devanagari}', "");
Expect(1, 2432, '\P{Is_Blk=devanagari}', "");
Expect(0, 2432, '\P{^Is_Blk=devanagari}', "");
Expect(1, 2431, '\p{Is_Blk=		DEVANAGARI}', "");
Expect(0, 2431, '\p{^Is_Blk=		DEVANAGARI}', "");
Expect(0, 2431, '\P{Is_Blk=		DEVANAGARI}', "");
Expect(1, 2431, '\P{^Is_Blk=		DEVANAGARI}', "");
Expect(0, 2432, '\p{Is_Blk=		DEVANAGARI}', "");
Expect(1, 2432, '\p{^Is_Blk=		DEVANAGARI}', "");
Expect(1, 2432, '\P{Is_Blk=		DEVANAGARI}', "");
Expect(0, 2432, '\P{^Is_Blk=		DEVANAGARI}', "");
Error('\p{Block=:= _DEVANAGARI_extended}');
Error('\P{Block=:= _DEVANAGARI_extended}');
Expect(1, 43263, '\p{Block=devanagariextended}', "");
Expect(0, 43263, '\p{^Block=devanagariextended}', "");
Expect(0, 43263, '\P{Block=devanagariextended}', "");
Expect(1, 43263, '\P{^Block=devanagariextended}', "");
Expect(0, 43264, '\p{Block=devanagariextended}', "");
Expect(1, 43264, '\p{^Block=devanagariextended}', "");
Expect(1, 43264, '\P{Block=devanagariextended}', "");
Expect(0, 43264, '\P{^Block=devanagariextended}', "");
Expect(1, 43263, '\p{Block=_Devanagari_Extended}', "");
Expect(0, 43263, '\p{^Block=_Devanagari_Extended}', "");
Expect(0, 43263, '\P{Block=_Devanagari_Extended}', "");
Expect(1, 43263, '\P{^Block=_Devanagari_Extended}', "");
Expect(0, 43264, '\p{Block=_Devanagari_Extended}', "");
Expect(1, 43264, '\p{^Block=_Devanagari_Extended}', "");
Expect(1, 43264, '\P{Block=_Devanagari_Extended}', "");
Expect(0, 43264, '\P{^Block=_Devanagari_Extended}', "");
Error('\p{Blk=- Devanagari_ext:=}');
Error('\P{Blk=- Devanagari_ext:=}');
Expect(1, 43263, '\p{Blk=devanagariext}', "");
Expect(0, 43263, '\p{^Blk=devanagariext}', "");
Expect(0, 43263, '\P{Blk=devanagariext}', "");
Expect(1, 43263, '\P{^Blk=devanagariext}', "");
Expect(0, 43264, '\p{Blk=devanagariext}', "");
Expect(1, 43264, '\p{^Blk=devanagariext}', "");
Expect(1, 43264, '\P{Blk=devanagariext}', "");
Expect(0, 43264, '\P{^Blk=devanagariext}', "");
Expect(1, 43263, '\p{Blk:		_devanagari_ext}', "");
Expect(0, 43263, '\p{^Blk:		_devanagari_ext}', "");
Expect(0, 43263, '\P{Blk:		_devanagari_ext}', "");
Expect(1, 43263, '\P{^Blk:		_devanagari_ext}', "");
Expect(0, 43264, '\p{Blk:		_devanagari_ext}', "");
Expect(1, 43264, '\p{^Blk:		_devanagari_ext}', "");
Expect(1, 43264, '\P{Blk:		_devanagari_ext}', "");
Expect(0, 43264, '\P{^Blk:		_devanagari_ext}', "");
Error('\p{Is_Block=:= devanagari_Extended}');
Error('\P{Is_Block=:= devanagari_Extended}');
Expect(1, 43263, '\p{Is_Block=devanagariextended}', "");
Expect(0, 43263, '\p{^Is_Block=devanagariextended}', "");
Expect(0, 43263, '\P{Is_Block=devanagariextended}', "");
Expect(1, 43263, '\P{^Is_Block=devanagariextended}', "");
Expect(0, 43264, '\p{Is_Block=devanagariextended}', "");
Expect(1, 43264, '\p{^Is_Block=devanagariextended}', "");
Expect(1, 43264, '\P{Is_Block=devanagariextended}', "");
Expect(0, 43264, '\P{^Is_Block=devanagariextended}', "");
Expect(1, 43263, '\p{Is_Block:	-DEVANAGARI_Extended}', "");
Expect(0, 43263, '\p{^Is_Block:	-DEVANAGARI_Extended}', "");
Expect(0, 43263, '\P{Is_Block:	-DEVANAGARI_Extended}', "");
Expect(1, 43263, '\P{^Is_Block:	-DEVANAGARI_Extended}', "");
Expect(0, 43264, '\p{Is_Block:	-DEVANAGARI_Extended}', "");
Expect(1, 43264, '\p{^Is_Block:	-DEVANAGARI_Extended}', "");
Expect(1, 43264, '\P{Is_Block:	-DEVANAGARI_Extended}', "");
Expect(0, 43264, '\P{^Is_Block:	-DEVANAGARI_Extended}', "");
Error('\p{Is_Blk=/a/  devanagari_ext}');
Error('\P{Is_Blk=/a/  devanagari_ext}');
Expect(1, 43263, '\p{Is_Blk=devanagariext}', "");
Expect(0, 43263, '\p{^Is_Blk=devanagariext}', "");
Expect(0, 43263, '\P{Is_Blk=devanagariext}', "");
Expect(1, 43263, '\P{^Is_Blk=devanagariext}', "");
Expect(0, 43264, '\p{Is_Blk=devanagariext}', "");
Expect(1, 43264, '\p{^Is_Blk=devanagariext}', "");
Expect(1, 43264, '\P{Is_Blk=devanagariext}', "");
Expect(0, 43264, '\P{^Is_Blk=devanagariext}', "");
Expect(1, 43263, '\p{Is_Blk=	_devanagari_EXT}', "");
Expect(0, 43263, '\p{^Is_Blk=	_devanagari_EXT}', "");
Expect(0, 43263, '\P{Is_Blk=	_devanagari_EXT}', "");
Expect(1, 43263, '\P{^Is_Blk=	_devanagari_EXT}', "");
Expect(0, 43264, '\p{Is_Blk=	_devanagari_EXT}', "");
Expect(1, 43264, '\p{^Is_Blk=	_devanagari_EXT}', "");
Expect(1, 43264, '\P{Is_Blk=	_devanagari_EXT}', "");
Expect(0, 43264, '\P{^Is_Blk=	_devanagari_EXT}', "");
Error('\p{Block= /a/combining_diacritical_MARKS}');
Error('\P{Block= /a/combining_diacritical_MARKS}');
Expect(1, 879, '\p{Block=combiningdiacriticalmarks}', "");
Expect(0, 879, '\p{^Block=combiningdiacriticalmarks}', "");
Expect(0, 879, '\P{Block=combiningdiacriticalmarks}', "");
Expect(1, 879, '\P{^Block=combiningdiacriticalmarks}', "");
Expect(0, 880, '\p{Block=combiningdiacriticalmarks}', "");
Expect(1, 880, '\p{^Block=combiningdiacriticalmarks}', "");
Expect(1, 880, '\P{Block=combiningdiacriticalmarks}', "");
Expect(0, 880, '\P{^Block=combiningdiacriticalmarks}', "");
Expect(1, 879, '\p{Block=-	Combining_Diacritical_Marks}', "");
Expect(0, 879, '\p{^Block=-	Combining_Diacritical_Marks}', "");
Expect(0, 879, '\P{Block=-	Combining_Diacritical_Marks}', "");
Expect(1, 879, '\P{^Block=-	Combining_Diacritical_Marks}', "");
Expect(0, 880, '\p{Block=-	Combining_Diacritical_Marks}', "");
Expect(1, 880, '\p{^Block=-	Combining_Diacritical_Marks}', "");
Expect(1, 880, '\P{Block=-	Combining_Diacritical_Marks}', "");
Expect(0, 880, '\P{^Block=-	Combining_Diacritical_Marks}', "");
Error('\p{Blk=:=	_diacriticals}');
Error('\P{Blk=:=	_diacriticals}');
Expect(1, 879, '\p{Blk=diacriticals}', "");
Expect(0, 879, '\p{^Blk=diacriticals}', "");
Expect(0, 879, '\P{Blk=diacriticals}', "");
Expect(1, 879, '\P{^Blk=diacriticals}', "");
Expect(0, 880, '\p{Blk=diacriticals}', "");
Expect(1, 880, '\p{^Blk=diacriticals}', "");
Expect(1, 880, '\P{Blk=diacriticals}', "");
Expect(0, 880, '\P{^Blk=diacriticals}', "");
Expect(1, 879, '\p{Blk=--DIACRITICALS}', "");
Expect(0, 879, '\p{^Blk=--DIACRITICALS}', "");
Expect(0, 879, '\P{Blk=--DIACRITICALS}', "");
Expect(1, 879, '\P{^Blk=--DIACRITICALS}', "");
Expect(0, 880, '\p{Blk=--DIACRITICALS}', "");
Expect(1, 880, '\p{^Blk=--DIACRITICALS}', "");
Expect(1, 880, '\P{Blk=--DIACRITICALS}', "");
Expect(0, 880, '\P{^Blk=--DIACRITICALS}', "");
Error('\p{Is_Block=-/a/combining_DIACRITICAL_MARKS}');
Error('\P{Is_Block=-/a/combining_DIACRITICAL_MARKS}');
Expect(1, 879, '\p{Is_Block=combiningdiacriticalmarks}', "");
Expect(0, 879, '\p{^Is_Block=combiningdiacriticalmarks}', "");
Expect(0, 879, '\P{Is_Block=combiningdiacriticalmarks}', "");
Expect(1, 879, '\P{^Is_Block=combiningdiacriticalmarks}', "");
Expect(0, 880, '\p{Is_Block=combiningdiacriticalmarks}', "");
Expect(1, 880, '\p{^Is_Block=combiningdiacriticalmarks}', "");
Expect(1, 880, '\P{Is_Block=combiningdiacriticalmarks}', "");
Expect(0, 880, '\P{^Is_Block=combiningdiacriticalmarks}', "");
Expect(1, 879, '\p{Is_Block=_	combining_diacritical_MARKS}', "");
Expect(0, 879, '\p{^Is_Block=_	combining_diacritical_MARKS}', "");
Expect(0, 879, '\P{Is_Block=_	combining_diacritical_MARKS}', "");
Expect(1, 879, '\P{^Is_Block=_	combining_diacritical_MARKS}', "");
Expect(0, 880, '\p{Is_Block=_	combining_diacritical_MARKS}', "");
Expect(1, 880, '\p{^Is_Block=_	combining_diacritical_MARKS}', "");
Expect(1, 880, '\P{Is_Block=_	combining_diacritical_MARKS}', "");
Expect(0, 880, '\P{^Is_Block=_	combining_diacritical_MARKS}', "");
Error('\p{Is_Blk=:=_	diacriticals}');
Error('\P{Is_Blk=:=_	diacriticals}');
Expect(1, 879, '\p{Is_Blk=diacriticals}', "");
Expect(0, 879, '\p{^Is_Blk=diacriticals}', "");
Expect(0, 879, '\P{Is_Blk=diacriticals}', "");
Expect(1, 879, '\P{^Is_Blk=diacriticals}', "");
Expect(0, 880, '\p{Is_Blk=diacriticals}', "");
Expect(1, 880, '\p{^Is_Blk=diacriticals}', "");
Expect(1, 880, '\P{Is_Blk=diacriticals}', "");
Expect(0, 880, '\P{^Is_Blk=diacriticals}', "");
Expect(1, 879, '\p{Is_Blk:   -	Diacriticals}', "");
Expect(0, 879, '\p{^Is_Blk:   -	Diacriticals}', "");
Expect(0, 879, '\P{Is_Blk:   -	Diacriticals}', "");
Expect(1, 879, '\P{^Is_Blk:   -	Diacriticals}', "");
Expect(0, 880, '\p{Is_Blk:   -	Diacriticals}', "");
Expect(1, 880, '\p{^Is_Blk:   -	Diacriticals}', "");
Expect(1, 880, '\P{Is_Blk:   -	Diacriticals}', "");
Expect(0, 880, '\P{^Is_Blk:   -	Diacriticals}', "");
Error('\p{Block=	_Combining_Diacritical_Marks_for_SYMBOLS/a/}');
Error('\P{Block=	_Combining_Diacritical_Marks_for_SYMBOLS/a/}');
Expect(1, 8447, '\p{Block:combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\p{^Block:combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\P{Block:combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\P{^Block:combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\p{Block:combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\p{^Block:combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\P{Block:combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\P{^Block:combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\p{Block:   	_Combining_Diacritical_Marks_For_Symbols}', "");
Expect(0, 8447, '\p{^Block:   	_Combining_Diacritical_Marks_For_Symbols}', "");
Expect(0, 8447, '\P{Block:   	_Combining_Diacritical_Marks_For_Symbols}', "");
Expect(1, 8447, '\P{^Block:   	_Combining_Diacritical_Marks_For_Symbols}', "");
Expect(0, 8448, '\p{Block:   	_Combining_Diacritical_Marks_For_Symbols}', "");
Expect(1, 8448, '\p{^Block:   	_Combining_Diacritical_Marks_For_Symbols}', "");
Expect(1, 8448, '\P{Block:   	_Combining_Diacritical_Marks_For_Symbols}', "");
Expect(0, 8448, '\P{^Block:   	_Combining_Diacritical_Marks_For_Symbols}', "");
Error('\p{Blk=/a/_DIACRITICALS_for_Symbols}');
Error('\P{Blk=/a/_DIACRITICALS_for_Symbols}');
Expect(1, 8447, '\p{Blk=diacriticalsforsymbols}', "");
Expect(0, 8447, '\p{^Blk=diacriticalsforsymbols}', "");
Expect(0, 8447, '\P{Blk=diacriticalsforsymbols}', "");
Expect(1, 8447, '\P{^Blk=diacriticalsforsymbols}', "");
Expect(0, 8448, '\p{Blk=diacriticalsforsymbols}', "");
Expect(1, 8448, '\p{^Blk=diacriticalsforsymbols}', "");
Expect(1, 8448, '\P{Blk=diacriticalsforsymbols}', "");
Expect(0, 8448, '\P{^Blk=diacriticalsforsymbols}', "");
Expect(1, 8447, '\p{Blk=- DIACRITICALS_For_SYMBOLS}', "");
Expect(0, 8447, '\p{^Blk=- DIACRITICALS_For_SYMBOLS}', "");
Expect(0, 8447, '\P{Blk=- DIACRITICALS_For_SYMBOLS}', "");
Expect(1, 8447, '\P{^Blk=- DIACRITICALS_For_SYMBOLS}', "");
Expect(0, 8448, '\p{Blk=- DIACRITICALS_For_SYMBOLS}', "");
Expect(1, 8448, '\p{^Blk=- DIACRITICALS_For_SYMBOLS}', "");
Expect(1, 8448, '\P{Blk=- DIACRITICALS_For_SYMBOLS}', "");
Expect(0, 8448, '\P{^Blk=- DIACRITICALS_For_SYMBOLS}', "");
Error('\p{Is_Block=/a/_Combining_marks_FOR_Symbols}');
Error('\P{Is_Block=/a/_Combining_marks_FOR_Symbols}');
Expect(1, 8447, '\p{Is_Block=combiningmarksforsymbols}', "");
Expect(0, 8447, '\p{^Is_Block=combiningmarksforsymbols}', "");
Expect(0, 8447, '\P{Is_Block=combiningmarksforsymbols}', "");
Expect(1, 8447, '\P{^Is_Block=combiningmarksforsymbols}', "");
Expect(0, 8448, '\p{Is_Block=combiningmarksforsymbols}', "");
Expect(1, 8448, '\p{^Is_Block=combiningmarksforsymbols}', "");
Expect(1, 8448, '\P{Is_Block=combiningmarksforsymbols}', "");
Expect(0, 8448, '\P{^Is_Block=combiningmarksforsymbols}', "");
Expect(1, 8447, '\p{Is_Block=	combining_marks_For_Symbols}', "");
Expect(0, 8447, '\p{^Is_Block=	combining_marks_For_Symbols}', "");
Expect(0, 8447, '\P{Is_Block=	combining_marks_For_Symbols}', "");
Expect(1, 8447, '\P{^Is_Block=	combining_marks_For_Symbols}', "");
Expect(0, 8448, '\p{Is_Block=	combining_marks_For_Symbols}', "");
Expect(1, 8448, '\p{^Is_Block=	combining_marks_For_Symbols}', "");
Expect(1, 8448, '\P{Is_Block=	combining_marks_For_Symbols}', "");
Expect(0, 8448, '\P{^Is_Block=	combining_marks_For_Symbols}', "");
Error('\p{Is_Blk=-COMBINING_Diacritical_Marks_For_Symbols/a/}');
Error('\P{Is_Blk=-COMBINING_Diacritical_Marks_For_Symbols/a/}');
Expect(1, 8447, '\p{Is_Blk=combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\p{^Is_Blk=combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\P{Is_Blk=combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\P{^Is_Blk=combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\p{Is_Blk=combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\p{^Is_Blk=combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\P{Is_Blk=combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\P{^Is_Blk=combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\p{Is_Blk=-COMBINING_Diacritical_MARKS_FOR_SYMBOLS}', "");
Expect(0, 8447, '\p{^Is_Blk=-COMBINING_Diacritical_MARKS_FOR_SYMBOLS}', "");
Expect(0, 8447, '\P{Is_Blk=-COMBINING_Diacritical_MARKS_FOR_SYMBOLS}', "");
Expect(1, 8447, '\P{^Is_Blk=-COMBINING_Diacritical_MARKS_FOR_SYMBOLS}', "");
Expect(0, 8448, '\p{Is_Blk=-COMBINING_Diacritical_MARKS_FOR_SYMBOLS}', "");
Expect(1, 8448, '\p{^Is_Blk=-COMBINING_Diacritical_MARKS_FOR_SYMBOLS}', "");
Expect(1, 8448, '\P{Is_Blk=-COMBINING_Diacritical_MARKS_FOR_SYMBOLS}', "");
Expect(0, 8448, '\P{^Is_Blk=-COMBINING_Diacritical_MARKS_FOR_SYMBOLS}', "");
Error('\p{Block=:= -COMBINING_diacritical_Marks_Supplement}');
Error('\P{Block=:= -COMBINING_diacritical_Marks_Supplement}');
Expect(1, 7679, '\p{Block=combiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\p{^Block=combiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\P{Block=combiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\P{^Block=combiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\p{Block=combiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\p{^Block=combiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\P{Block=combiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\P{^Block=combiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\p{Block=	_Combining_DIACRITICAL_Marks_SUPPLEMENT}', "");
Expect(0, 7679, '\p{^Block=	_Combining_DIACRITICAL_Marks_SUPPLEMENT}', "");
Expect(0, 7679, '\P{Block=	_Combining_DIACRITICAL_Marks_SUPPLEMENT}', "");
Expect(1, 7679, '\P{^Block=	_Combining_DIACRITICAL_Marks_SUPPLEMENT}', "");
Expect(0, 7680, '\p{Block=	_Combining_DIACRITICAL_Marks_SUPPLEMENT}', "");
Expect(1, 7680, '\p{^Block=	_Combining_DIACRITICAL_Marks_SUPPLEMENT}', "");
Expect(1, 7680, '\P{Block=	_Combining_DIACRITICAL_Marks_SUPPLEMENT}', "");
Expect(0, 7680, '\P{^Block=	_Combining_DIACRITICAL_Marks_SUPPLEMENT}', "");
Error('\p{Blk=:= Diacriticals_SUP}');
Error('\P{Blk=:= Diacriticals_SUP}');
Expect(1, 7679, '\p{Blk=diacriticalssup}', "");
Expect(0, 7679, '\p{^Blk=diacriticalssup}', "");
Expect(0, 7679, '\P{Blk=diacriticalssup}', "");
Expect(1, 7679, '\P{^Blk=diacriticalssup}', "");
Expect(0, 7680, '\p{Blk=diacriticalssup}', "");
Expect(1, 7680, '\p{^Blk=diacriticalssup}', "");
Expect(1, 7680, '\P{Blk=diacriticalssup}', "");
Expect(0, 7680, '\P{^Blk=diacriticalssup}', "");
Expect(1, 7679, '\p{Blk:- diacriticals_Sup}', "");
Expect(0, 7679, '\p{^Blk:- diacriticals_Sup}', "");
Expect(0, 7679, '\P{Blk:- diacriticals_Sup}', "");
Expect(1, 7679, '\P{^Blk:- diacriticals_Sup}', "");
Expect(0, 7680, '\p{Blk:- diacriticals_Sup}', "");
Expect(1, 7680, '\p{^Blk:- diacriticals_Sup}', "");
Expect(1, 7680, '\P{Blk:- diacriticals_Sup}', "");
Expect(0, 7680, '\P{^Blk:- diacriticals_Sup}', "");
Error('\p{Is_Block=:=-_combining_Diacritical_MARKS_Supplement}');
Error('\P{Is_Block=:=-_combining_Diacritical_MARKS_Supplement}');
Expect(1, 7679, '\p{Is_Block=combiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\p{^Is_Block=combiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\P{Is_Block=combiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\P{^Is_Block=combiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\p{Is_Block=combiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\p{^Is_Block=combiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\P{Is_Block=combiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\P{^Is_Block=combiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\p{Is_Block=-combining_Diacritical_Marks_Supplement}', "");
Expect(0, 7679, '\p{^Is_Block=-combining_Diacritical_Marks_Supplement}', "");
Expect(0, 7679, '\P{Is_Block=-combining_Diacritical_Marks_Supplement}', "");
Expect(1, 7679, '\P{^Is_Block=-combining_Diacritical_Marks_Supplement}', "");
Expect(0, 7680, '\p{Is_Block=-combining_Diacritical_Marks_Supplement}', "");
Expect(1, 7680, '\p{^Is_Block=-combining_Diacritical_Marks_Supplement}', "");
Expect(1, 7680, '\P{Is_Block=-combining_Diacritical_Marks_Supplement}', "");
Expect(0, 7680, '\P{^Is_Block=-combining_Diacritical_Marks_Supplement}', "");
Error('\p{Is_Blk::=_Diacriticals_sup}');
Error('\P{Is_Blk::=_Diacriticals_sup}');
Expect(1, 7679, '\p{Is_Blk=diacriticalssup}', "");
Expect(0, 7679, '\p{^Is_Blk=diacriticalssup}', "");
Expect(0, 7679, '\P{Is_Blk=diacriticalssup}', "");
Expect(1, 7679, '\P{^Is_Blk=diacriticalssup}', "");
Expect(0, 7680, '\p{Is_Blk=diacriticalssup}', "");
Expect(1, 7680, '\p{^Is_Blk=diacriticalssup}', "");
Expect(1, 7680, '\P{Is_Blk=diacriticalssup}', "");
Expect(0, 7680, '\P{^Is_Blk=diacriticalssup}', "");
Expect(1, 7679, '\p{Is_Blk= Diacriticals_SUP}', "");
Expect(0, 7679, '\p{^Is_Blk= Diacriticals_SUP}', "");
Expect(0, 7679, '\P{Is_Blk= Diacriticals_SUP}', "");
Expect(1, 7679, '\P{^Is_Blk= Diacriticals_SUP}', "");
Expect(0, 7680, '\p{Is_Blk= Diacriticals_SUP}', "");
Expect(1, 7680, '\p{^Is_Blk= Diacriticals_SUP}', "");
Expect(1, 7680, '\P{Is_Blk= Diacriticals_SUP}', "");
Expect(0, 7680, '\P{^Is_Blk= Diacriticals_SUP}', "");
Error('\p{Block=-:=dingbats}');
Error('\P{Block=-:=dingbats}');
Expect(1, 10175, '\p{Block=dingbats}', "");
Expect(0, 10175, '\p{^Block=dingbats}', "");
Expect(0, 10175, '\P{Block=dingbats}', "");
Expect(1, 10175, '\P{^Block=dingbats}', "");
Expect(0, 10176, '\p{Block=dingbats}', "");
Expect(1, 10176, '\p{^Block=dingbats}', "");
Expect(1, 10176, '\P{Block=dingbats}', "");
Expect(0, 10176, '\P{^Block=dingbats}', "");
Expect(1, 10175, '\p{Block=	-Dingbats}', "");
Expect(0, 10175, '\p{^Block=	-Dingbats}', "");
Expect(0, 10175, '\P{Block=	-Dingbats}', "");
Expect(1, 10175, '\P{^Block=	-Dingbats}', "");
Expect(0, 10176, '\p{Block=	-Dingbats}', "");
Expect(1, 10176, '\p{^Block=	-Dingbats}', "");
Expect(1, 10176, '\P{Block=	-Dingbats}', "");
Expect(0, 10176, '\P{^Block=	-Dingbats}', "");
Error('\p{Blk=_ DINGBATS/a/}');
Error('\P{Blk=_ DINGBATS/a/}');
Expect(1, 10175, '\p{Blk=dingbats}', "");
Expect(0, 10175, '\p{^Blk=dingbats}', "");
Expect(0, 10175, '\P{Blk=dingbats}', "");
Expect(1, 10175, '\P{^Blk=dingbats}', "");
Expect(0, 10176, '\p{Blk=dingbats}', "");
Expect(1, 10176, '\p{^Blk=dingbats}', "");
Expect(1, 10176, '\P{Blk=dingbats}', "");
Expect(0, 10176, '\P{^Blk=dingbats}', "");
Expect(1, 10175, '\p{Blk=_dingbats}', "");
Expect(0, 10175, '\p{^Blk=_dingbats}', "");
Expect(0, 10175, '\P{Blk=_dingbats}', "");
Expect(1, 10175, '\P{^Blk=_dingbats}', "");
Expect(0, 10176, '\p{Blk=_dingbats}', "");
Expect(1, 10176, '\p{^Blk=_dingbats}', "");
Expect(1, 10176, '\P{Blk=_dingbats}', "");
Expect(0, 10176, '\P{^Blk=_dingbats}', "");
Error('\p{Is_Block=-:=Dingbats}');
Error('\P{Is_Block=-:=Dingbats}');
Expect(1, 10175, '\p{Is_Block: dingbats}', "");
Expect(0, 10175, '\p{^Is_Block: dingbats}', "");
Expect(0, 10175, '\P{Is_Block: dingbats}', "");
Expect(1, 10175, '\P{^Is_Block: dingbats}', "");
Expect(0, 10176, '\p{Is_Block: dingbats}', "");
Expect(1, 10176, '\p{^Is_Block: dingbats}', "");
Expect(1, 10176, '\P{Is_Block: dingbats}', "");
Expect(0, 10176, '\P{^Is_Block: dingbats}', "");
Expect(1, 10175, '\p{Is_Block=		Dingbats}', "");
Expect(0, 10175, '\p{^Is_Block=		Dingbats}', "");
Expect(0, 10175, '\P{Is_Block=		Dingbats}', "");
Expect(1, 10175, '\P{^Is_Block=		Dingbats}', "");
Expect(0, 10176, '\p{Is_Block=		Dingbats}', "");
Expect(1, 10176, '\p{^Is_Block=		Dingbats}', "");
Expect(1, 10176, '\P{Is_Block=		Dingbats}', "");
Expect(0, 10176, '\P{^Is_Block=		Dingbats}', "");
Error('\p{Is_Blk=-:=Dingbats}');
Error('\P{Is_Blk=-:=Dingbats}');
Expect(1, 10175, '\p{Is_Blk=dingbats}', "");
Expect(0, 10175, '\p{^Is_Blk=dingbats}', "");
Expect(0, 10175, '\P{Is_Blk=dingbats}', "");
Expect(1, 10175, '\P{^Is_Blk=dingbats}', "");
Expect(0, 10176, '\p{Is_Blk=dingbats}', "");
Expect(1, 10176, '\p{^Is_Blk=dingbats}', "");
Expect(1, 10176, '\P{Is_Blk=dingbats}', "");
Expect(0, 10176, '\P{^Is_Blk=dingbats}', "");
Expect(1, 10175, '\p{Is_Blk=-_dingbats}', "");
Expect(0, 10175, '\p{^Is_Blk=-_dingbats}', "");
Expect(0, 10175, '\P{Is_Blk=-_dingbats}', "");
Expect(1, 10175, '\P{^Is_Blk=-_dingbats}', "");
Expect(0, 10176, '\p{Is_Blk=-_dingbats}', "");
Expect(1, 10176, '\p{^Is_Blk=-_dingbats}', "");
Expect(1, 10176, '\P{Is_Blk=-_dingbats}', "");
Expect(0, 10176, '\P{^Is_Blk=-_dingbats}', "");
Error('\p{Block= -DOMINO_Tiles/a/}');
Error('\P{Block= -DOMINO_Tiles/a/}');
Expect(1, 127135, '\p{Block=dominotiles}', "");
Expect(0, 127135, '\p{^Block=dominotiles}', "");
Expect(0, 127135, '\P{Block=dominotiles}', "");
Expect(1, 127135, '\P{^Block=dominotiles}', "");
Expect(0, 127136, '\p{Block=dominotiles}', "");
Expect(1, 127136, '\p{^Block=dominotiles}', "");
Expect(1, 127136, '\P{Block=dominotiles}', "");
Expect(0, 127136, '\P{^Block=dominotiles}', "");
Expect(1, 127135, '\p{Block=		domino_Tiles}', "");
Expect(0, 127135, '\p{^Block=		domino_Tiles}', "");
Expect(0, 127135, '\P{Block=		domino_Tiles}', "");
Expect(1, 127135, '\P{^Block=		domino_Tiles}', "");
Expect(0, 127136, '\p{Block=		domino_Tiles}', "");
Expect(1, 127136, '\p{^Block=		domino_Tiles}', "");
Expect(1, 127136, '\P{Block=		domino_Tiles}', "");
Expect(0, 127136, '\P{^Block=		domino_Tiles}', "");
Error('\p{Blk=	-Domino:=}');
Error('\P{Blk=	-Domino:=}');
Expect(1, 127135, '\p{Blk=domino}', "");
Expect(0, 127135, '\p{^Blk=domino}', "");
Expect(0, 127135, '\P{Blk=domino}', "");
Expect(1, 127135, '\P{^Blk=domino}', "");
Expect(0, 127136, '\p{Blk=domino}', "");
Expect(1, 127136, '\p{^Blk=domino}', "");
Expect(1, 127136, '\P{Blk=domino}', "");
Expect(0, 127136, '\P{^Blk=domino}', "");
Expect(1, 127135, '\p{Blk= 	domino}', "");
Expect(0, 127135, '\p{^Blk= 	domino}', "");
Expect(0, 127135, '\P{Blk= 	domino}', "");
Expect(1, 127135, '\P{^Blk= 	domino}', "");
Expect(0, 127136, '\p{Blk= 	domino}', "");
Expect(1, 127136, '\p{^Blk= 	domino}', "");
Expect(1, 127136, '\P{Blk= 	domino}', "");
Expect(0, 127136, '\P{^Blk= 	domino}', "");
Error('\p{Is_Block=/a/	Domino_Tiles}');
Error('\P{Is_Block=/a/	Domino_Tiles}');
Expect(1, 127135, '\p{Is_Block:dominotiles}', "");
Expect(0, 127135, '\p{^Is_Block:dominotiles}', "");
Expect(0, 127135, '\P{Is_Block:dominotiles}', "");
Expect(1, 127135, '\P{^Is_Block:dominotiles}', "");
Expect(0, 127136, '\p{Is_Block:dominotiles}', "");
Expect(1, 127136, '\p{^Is_Block:dominotiles}', "");
Expect(1, 127136, '\P{Is_Block:dominotiles}', "");
Expect(0, 127136, '\P{^Is_Block:dominotiles}', "");
Expect(1, 127135, '\p{Is_Block= _Domino_Tiles}', "");
Expect(0, 127135, '\p{^Is_Block= _Domino_Tiles}', "");
Expect(0, 127135, '\P{Is_Block= _Domino_Tiles}', "");
Expect(1, 127135, '\P{^Is_Block= _Domino_Tiles}', "");
Expect(0, 127136, '\p{Is_Block= _Domino_Tiles}', "");
Expect(1, 127136, '\p{^Is_Block= _Domino_Tiles}', "");
Expect(1, 127136, '\P{Is_Block= _Domino_Tiles}', "");
Expect(0, 127136, '\P{^Is_Block= _Domino_Tiles}', "");
Error('\p{Is_Blk=:=-Domino}');
Error('\P{Is_Blk=:=-Domino}');
Expect(1, 127135, '\p{Is_Blk=domino}', "");
Expect(0, 127135, '\p{^Is_Blk=domino}', "");
Expect(0, 127135, '\P{Is_Blk=domino}', "");
Expect(1, 127135, '\P{^Is_Blk=domino}', "");
Expect(0, 127136, '\p{Is_Blk=domino}', "");
Expect(1, 127136, '\p{^Is_Blk=domino}', "");
Expect(1, 127136, '\P{Is_Blk=domino}', "");
Expect(0, 127136, '\P{^Is_Blk=domino}', "");
Expect(1, 127135, '\p{Is_Blk=-	DOMINO}', "");
Expect(0, 127135, '\p{^Is_Blk=-	DOMINO}', "");
Expect(0, 127135, '\P{Is_Blk=-	DOMINO}', "");
Expect(1, 127135, '\P{^Is_Blk=-	DOMINO}', "");
Expect(0, 127136, '\p{Is_Blk=-	DOMINO}', "");
Expect(1, 127136, '\p{^Is_Blk=-	DOMINO}', "");
Expect(1, 127136, '\P{Is_Blk=-	DOMINO}', "");
Expect(0, 127136, '\P{^Is_Blk=-	DOMINO}', "");
Error('\p{Block= 	egyptian_hieroglyphs:=}');
Error('\P{Block= 	egyptian_hieroglyphs:=}');
Expect(1, 78895, '\p{Block=egyptianhieroglyphs}', "");
Expect(0, 78895, '\p{^Block=egyptianhieroglyphs}', "");
Expect(0, 78895, '\P{Block=egyptianhieroglyphs}', "");
Expect(1, 78895, '\P{^Block=egyptianhieroglyphs}', "");
Expect(0, 78896, '\p{Block=egyptianhieroglyphs}', "");
Expect(1, 78896, '\p{^Block=egyptianhieroglyphs}', "");
Expect(1, 78896, '\P{Block=egyptianhieroglyphs}', "");
Expect(0, 78896, '\P{^Block=egyptianhieroglyphs}', "");
Expect(1, 78895, '\p{Block:  Egyptian_HIEROGLYPHS}', "");
Expect(0, 78895, '\p{^Block:  Egyptian_HIEROGLYPHS}', "");
Expect(0, 78895, '\P{Block:  Egyptian_HIEROGLYPHS}', "");
Expect(1, 78895, '\P{^Block:  Egyptian_HIEROGLYPHS}', "");
Expect(0, 78896, '\p{Block:  Egyptian_HIEROGLYPHS}', "");
Expect(1, 78896, '\p{^Block:  Egyptian_HIEROGLYPHS}', "");
Expect(1, 78896, '\P{Block:  Egyptian_HIEROGLYPHS}', "");
Expect(0, 78896, '\P{^Block:  Egyptian_HIEROGLYPHS}', "");
Error('\p{Blk:-/a/egyptian_Hieroglyphs}');
Error('\P{Blk:-/a/egyptian_Hieroglyphs}');
Expect(1, 78895, '\p{Blk:   egyptianhieroglyphs}', "");
Expect(0, 78895, '\p{^Blk:   egyptianhieroglyphs}', "");
Expect(0, 78895, '\P{Blk:   egyptianhieroglyphs}', "");
Expect(1, 78895, '\P{^Blk:   egyptianhieroglyphs}', "");
Expect(0, 78896, '\p{Blk:   egyptianhieroglyphs}', "");
Expect(1, 78896, '\p{^Blk:   egyptianhieroglyphs}', "");
Expect(1, 78896, '\P{Blk:   egyptianhieroglyphs}', "");
Expect(0, 78896, '\P{^Blk:   egyptianhieroglyphs}', "");
Expect(1, 78895, '\p{Blk=EGYPTIAN_HIEROGLYPHS}', "");
Expect(0, 78895, '\p{^Blk=EGYPTIAN_HIEROGLYPHS}', "");
Expect(0, 78895, '\P{Blk=EGYPTIAN_HIEROGLYPHS}', "");
Expect(1, 78895, '\P{^Blk=EGYPTIAN_HIEROGLYPHS}', "");
Expect(0, 78896, '\p{Blk=EGYPTIAN_HIEROGLYPHS}', "");
Expect(1, 78896, '\p{^Blk=EGYPTIAN_HIEROGLYPHS}', "");
Expect(1, 78896, '\P{Blk=EGYPTIAN_HIEROGLYPHS}', "");
Expect(0, 78896, '\P{^Blk=EGYPTIAN_HIEROGLYPHS}', "");
Error('\p{Is_Block=/a/ Egyptian_hieroglyphs}');
Error('\P{Is_Block=/a/ Egyptian_hieroglyphs}');
Expect(1, 78895, '\p{Is_Block=egyptianhieroglyphs}', "");
Expect(0, 78895, '\p{^Is_Block=egyptianhieroglyphs}', "");
Expect(0, 78895, '\P{Is_Block=egyptianhieroglyphs}', "");
Expect(1, 78895, '\P{^Is_Block=egyptianhieroglyphs}', "");
Expect(0, 78896, '\p{Is_Block=egyptianhieroglyphs}', "");
Expect(1, 78896, '\p{^Is_Block=egyptianhieroglyphs}', "");
Expect(1, 78896, '\P{Is_Block=egyptianhieroglyphs}', "");
Expect(0, 78896, '\P{^Is_Block=egyptianhieroglyphs}', "");
Expect(1, 78895, '\p{Is_Block=- EGYPTIAN_hieroglyphs}', "");
Expect(0, 78895, '\p{^Is_Block=- EGYPTIAN_hieroglyphs}', "");
Expect(0, 78895, '\P{Is_Block=- EGYPTIAN_hieroglyphs}', "");
Expect(1, 78895, '\P{^Is_Block=- EGYPTIAN_hieroglyphs}', "");
Expect(0, 78896, '\p{Is_Block=- EGYPTIAN_hieroglyphs}', "");
Expect(1, 78896, '\p{^Is_Block=- EGYPTIAN_hieroglyphs}', "");
Expect(1, 78896, '\P{Is_Block=- EGYPTIAN_hieroglyphs}', "");
Expect(0, 78896, '\P{^Is_Block=- EGYPTIAN_hieroglyphs}', "");
Error('\p{Is_Blk=_egyptian_Hieroglyphs/a/}');
Error('\P{Is_Blk=_egyptian_Hieroglyphs/a/}');
Expect(1, 78895, '\p{Is_Blk=egyptianhieroglyphs}', "");
Expect(0, 78895, '\p{^Is_Blk=egyptianhieroglyphs}', "");
Expect(0, 78895, '\P{Is_Blk=egyptianhieroglyphs}', "");
Expect(1, 78895, '\P{^Is_Blk=egyptianhieroglyphs}', "");
Expect(0, 78896, '\p{Is_Blk=egyptianhieroglyphs}', "");
Expect(1, 78896, '\p{^Is_Blk=egyptianhieroglyphs}', "");
Expect(1, 78896, '\P{Is_Blk=egyptianhieroglyphs}', "");
Expect(0, 78896, '\P{^Is_Blk=egyptianhieroglyphs}', "");
Expect(1, 78895, '\p{Is_Blk=- Egyptian_hieroglyphs}', "");
Expect(0, 78895, '\p{^Is_Blk=- Egyptian_hieroglyphs}', "");
Expect(0, 78895, '\P{Is_Blk=- Egyptian_hieroglyphs}', "");
Expect(1, 78895, '\P{^Is_Blk=- Egyptian_hieroglyphs}', "");
Expect(0, 78896, '\p{Is_Blk=- Egyptian_hieroglyphs}', "");
Expect(1, 78896, '\p{^Is_Blk=- Egyptian_hieroglyphs}', "");
Expect(1, 78896, '\P{Is_Blk=- Egyptian_hieroglyphs}', "");
Expect(0, 78896, '\P{^Is_Blk=- Egyptian_hieroglyphs}', "");
Error('\p{Block= :=Emoticons}');
Error('\P{Block= :=Emoticons}');
Expect(1, 128591, '\p{Block=emoticons}', "");
Expect(0, 128591, '\p{^Block=emoticons}', "");
Expect(0, 128591, '\P{Block=emoticons}', "");
Expect(1, 128591, '\P{^Block=emoticons}', "");
Expect(0, 128592, '\p{Block=emoticons}', "");
Expect(1, 128592, '\p{^Block=emoticons}', "");
Expect(1, 128592, '\P{Block=emoticons}', "");
Expect(0, 128592, '\P{^Block=emoticons}', "");
Expect(1, 128591, '\p{Block=	_EMOTICONS}', "");
Expect(0, 128591, '\p{^Block=	_EMOTICONS}', "");
Expect(0, 128591, '\P{Block=	_EMOTICONS}', "");
Expect(1, 128591, '\P{^Block=	_EMOTICONS}', "");
Expect(0, 128592, '\p{Block=	_EMOTICONS}', "");
Expect(1, 128592, '\p{^Block=	_EMOTICONS}', "");
Expect(1, 128592, '\P{Block=	_EMOTICONS}', "");
Expect(0, 128592, '\P{^Block=	_EMOTICONS}', "");
Error('\p{Blk=:=  EMOTICONS}');
Error('\P{Blk=:=  EMOTICONS}');
Expect(1, 128591, '\p{Blk=emoticons}', "");
Expect(0, 128591, '\p{^Blk=emoticons}', "");
Expect(0, 128591, '\P{Blk=emoticons}', "");
Expect(1, 128591, '\P{^Blk=emoticons}', "");
Expect(0, 128592, '\p{Blk=emoticons}', "");
Expect(1, 128592, '\p{^Blk=emoticons}', "");
Expect(1, 128592, '\P{Blk=emoticons}', "");
Expect(0, 128592, '\P{^Blk=emoticons}', "");
Expect(1, 128591, '\p{Blk=-_EMOTICONS}', "");
Expect(0, 128591, '\p{^Blk=-_EMOTICONS}', "");
Expect(0, 128591, '\P{Blk=-_EMOTICONS}', "");
Expect(1, 128591, '\P{^Blk=-_EMOTICONS}', "");
Expect(0, 128592, '\p{Blk=-_EMOTICONS}', "");
Expect(1, 128592, '\p{^Blk=-_EMOTICONS}', "");
Expect(1, 128592, '\P{Blk=-_EMOTICONS}', "");
Expect(0, 128592, '\P{^Blk=-_EMOTICONS}', "");
Error('\p{Is_Block=-:=Emoticons}');
Error('\P{Is_Block=-:=Emoticons}');
Expect(1, 128591, '\p{Is_Block=emoticons}', "");
Expect(0, 128591, '\p{^Is_Block=emoticons}', "");
Expect(0, 128591, '\P{Is_Block=emoticons}', "");
Expect(1, 128591, '\P{^Is_Block=emoticons}', "");
Expect(0, 128592, '\p{Is_Block=emoticons}', "");
Expect(1, 128592, '\p{^Is_Block=emoticons}', "");
Expect(1, 128592, '\P{Is_Block=emoticons}', "");
Expect(0, 128592, '\P{^Is_Block=emoticons}', "");
Expect(1, 128591, '\p{Is_Block=_	emoticons}', "");
Expect(0, 128591, '\p{^Is_Block=_	emoticons}', "");
Expect(0, 128591, '\P{Is_Block=_	emoticons}', "");
Expect(1, 128591, '\P{^Is_Block=_	emoticons}', "");
Expect(0, 128592, '\p{Is_Block=_	emoticons}', "");
Expect(1, 128592, '\p{^Is_Block=_	emoticons}', "");
Expect(1, 128592, '\P{Is_Block=_	emoticons}', "");
Expect(0, 128592, '\P{^Is_Block=_	emoticons}', "");
Error('\p{Is_Blk=/a/ 	EMOTICONS}');
Error('\P{Is_Blk=/a/ 	EMOTICONS}');
Expect(1, 128591, '\p{Is_Blk=emoticons}', "");
Expect(0, 128591, '\p{^Is_Blk=emoticons}', "");
Expect(0, 128591, '\P{Is_Blk=emoticons}', "");
Expect(1, 128591, '\P{^Is_Blk=emoticons}', "");
Expect(0, 128592, '\p{Is_Blk=emoticons}', "");
Expect(1, 128592, '\p{^Is_Blk=emoticons}', "");
Expect(1, 128592, '\P{Is_Blk=emoticons}', "");
Expect(0, 128592, '\P{^Is_Blk=emoticons}', "");
Expect(1, 128591, '\p{Is_Blk=	_Emoticons}', "");
Expect(0, 128591, '\p{^Is_Blk=	_Emoticons}', "");
Expect(0, 128591, '\P{Is_Blk=	_Emoticons}', "");
Expect(1, 128591, '\P{^Is_Blk=	_Emoticons}', "");
Expect(0, 128592, '\p{Is_Blk=	_Emoticons}', "");
Expect(1, 128592, '\p{^Is_Blk=	_Emoticons}', "");
Expect(1, 128592, '\P{Is_Blk=	_Emoticons}', "");
Expect(0, 128592, '\P{^Is_Blk=	_Emoticons}', "");
Error('\p{Block=-	enclosed_Alphanumerics:=}');
Error('\P{Block=-	enclosed_Alphanumerics:=}');
Expect(1, 9471, '\p{Block=enclosedalphanumerics}', "");
Expect(0, 9471, '\p{^Block=enclosedalphanumerics}', "");
Expect(0, 9471, '\P{Block=enclosedalphanumerics}', "");
Expect(1, 9471, '\P{^Block=enclosedalphanumerics}', "");
Expect(0, 9472, '\p{Block=enclosedalphanumerics}', "");
Expect(1, 9472, '\p{^Block=enclosedalphanumerics}', "");
Expect(1, 9472, '\P{Block=enclosedalphanumerics}', "");
Expect(0, 9472, '\P{^Block=enclosedalphanumerics}', "");
Expect(1, 9471, '\p{Block=		ENCLOSED_alphanumerics}', "");
Expect(0, 9471, '\p{^Block=		ENCLOSED_alphanumerics}', "");
Expect(0, 9471, '\P{Block=		ENCLOSED_alphanumerics}', "");
Expect(1, 9471, '\P{^Block=		ENCLOSED_alphanumerics}', "");
Expect(0, 9472, '\p{Block=		ENCLOSED_alphanumerics}', "");
Expect(1, 9472, '\p{^Block=		ENCLOSED_alphanumerics}', "");
Expect(1, 9472, '\P{Block=		ENCLOSED_alphanumerics}', "");
Expect(0, 9472, '\P{^Block=		ENCLOSED_alphanumerics}', "");
Error('\p{Blk:	:=__Enclosed_alphanum}');
Error('\P{Blk:	:=__Enclosed_alphanum}');
Expect(1, 9471, '\p{Blk=enclosedalphanum}', "");
Expect(0, 9471, '\p{^Blk=enclosedalphanum}', "");
Expect(0, 9471, '\P{Blk=enclosedalphanum}', "");
Expect(1, 9471, '\P{^Blk=enclosedalphanum}', "");
Expect(0, 9472, '\p{Blk=enclosedalphanum}', "");
Expect(1, 9472, '\p{^Blk=enclosedalphanum}', "");
Expect(1, 9472, '\P{Blk=enclosedalphanum}', "");
Expect(0, 9472, '\P{^Blk=enclosedalphanum}', "");
Expect(1, 9471, '\p{Blk=_enclosed_ALPHANUM}', "");
Expect(0, 9471, '\p{^Blk=_enclosed_ALPHANUM}', "");
Expect(0, 9471, '\P{Blk=_enclosed_ALPHANUM}', "");
Expect(1, 9471, '\P{^Blk=_enclosed_ALPHANUM}', "");
Expect(0, 9472, '\p{Blk=_enclosed_ALPHANUM}', "");
Expect(1, 9472, '\p{^Blk=_enclosed_ALPHANUM}', "");
Expect(1, 9472, '\P{Blk=_enclosed_ALPHANUM}', "");
Expect(0, 9472, '\P{^Blk=_enclosed_ALPHANUM}', "");
Error('\p{Is_Block=/a/ 	Enclosed_ALPHANUMERICS}');
Error('\P{Is_Block=/a/ 	Enclosed_ALPHANUMERICS}');
Expect(1, 9471, '\p{Is_Block=enclosedalphanumerics}', "");
Expect(0, 9471, '\p{^Is_Block=enclosedalphanumerics}', "");
Expect(0, 9471, '\P{Is_Block=enclosedalphanumerics}', "");
Expect(1, 9471, '\P{^Is_Block=enclosedalphanumerics}', "");
Expect(0, 9472, '\p{Is_Block=enclosedalphanumerics}', "");
Expect(1, 9472, '\p{^Is_Block=enclosedalphanumerics}', "");
Expect(1, 9472, '\P{Is_Block=enclosedalphanumerics}', "");
Expect(0, 9472, '\P{^Is_Block=enclosedalphanumerics}', "");
Expect(1, 9471, '\p{Is_Block= 	enclosed_ALPHANUMERICS}', "");
Expect(0, 9471, '\p{^Is_Block= 	enclosed_ALPHANUMERICS}', "");
Expect(0, 9471, '\P{Is_Block= 	enclosed_ALPHANUMERICS}', "");
Expect(1, 9471, '\P{^Is_Block= 	enclosed_ALPHANUMERICS}', "");
Expect(0, 9472, '\p{Is_Block= 	enclosed_ALPHANUMERICS}', "");
Expect(1, 9472, '\p{^Is_Block= 	enclosed_ALPHANUMERICS}', "");
Expect(1, 9472, '\P{Is_Block= 	enclosed_ALPHANUMERICS}', "");
Expect(0, 9472, '\P{^Is_Block= 	enclosed_ALPHANUMERICS}', "");
Error('\p{Is_Blk= /a/Enclosed_alphanum}');
Error('\P{Is_Blk= /a/Enclosed_alphanum}');
Expect(1, 9471, '\p{Is_Blk=enclosedalphanum}', "");
Expect(0, 9471, '\p{^Is_Blk=enclosedalphanum}', "");
Expect(0, 9471, '\P{Is_Blk=enclosedalphanum}', "");
Expect(1, 9471, '\P{^Is_Blk=enclosedalphanum}', "");
Expect(0, 9472, '\p{Is_Blk=enclosedalphanum}', "");
Expect(1, 9472, '\p{^Is_Blk=enclosedalphanum}', "");
Expect(1, 9472, '\P{Is_Blk=enclosedalphanum}', "");
Expect(0, 9472, '\P{^Is_Blk=enclosedalphanum}', "");
Expect(1, 9471, '\p{Is_Blk= -Enclosed_alphanum}', "");
Expect(0, 9471, '\p{^Is_Blk= -Enclosed_alphanum}', "");
Expect(0, 9471, '\P{Is_Blk= -Enclosed_alphanum}', "");
Expect(1, 9471, '\P{^Is_Blk= -Enclosed_alphanum}', "");
Expect(0, 9472, '\p{Is_Blk= -Enclosed_alphanum}', "");
Expect(1, 9472, '\p{^Is_Blk= -Enclosed_alphanum}', "");
Expect(1, 9472, '\P{Is_Blk= -Enclosed_alphanum}', "");
Expect(0, 9472, '\P{^Is_Blk= -Enclosed_alphanum}', "");
Error('\p{Block=/a/--Enclosed_Alphanumeric_Supplement}');
Error('\P{Block=/a/--Enclosed_Alphanumeric_Supplement}');
Expect(1, 127487, '\p{Block=enclosedalphanumericsupplement}', "");
Expect(0, 127487, '\p{^Block=enclosedalphanumericsupplement}', "");
Expect(0, 127487, '\P{Block=enclosedalphanumericsupplement}', "");
Expect(1, 127487, '\P{^Block=enclosedalphanumericsupplement}', "");
Expect(0, 127488, '\p{Block=enclosedalphanumericsupplement}', "");
Expect(1, 127488, '\p{^Block=enclosedalphanumericsupplement}', "");
Expect(1, 127488, '\P{Block=enclosedalphanumericsupplement}', "");
Expect(0, 127488, '\P{^Block=enclosedalphanumericsupplement}', "");
Expect(1, 127487, '\p{Block=-	Enclosed_alphanumeric_SUPPLEMENT}', "");
Expect(0, 127487, '\p{^Block=-	Enclosed_alphanumeric_SUPPLEMENT}', "");
Expect(0, 127487, '\P{Block=-	Enclosed_alphanumeric_SUPPLEMENT}', "");
Expect(1, 127487, '\P{^Block=-	Enclosed_alphanumeric_SUPPLEMENT}', "");
Expect(0, 127488, '\p{Block=-	Enclosed_alphanumeric_SUPPLEMENT}', "");
Expect(1, 127488, '\p{^Block=-	Enclosed_alphanumeric_SUPPLEMENT}', "");
Expect(1, 127488, '\P{Block=-	Enclosed_alphanumeric_SUPPLEMENT}', "");
Expect(0, 127488, '\P{^Block=-	Enclosed_alphanumeric_SUPPLEMENT}', "");
Error('\p{Blk=/a/	enclosed_ALPHANUM_Sup}');
Error('\P{Blk=/a/	enclosed_ALPHANUM_Sup}');
Expect(1, 127487, '\p{Blk=enclosedalphanumsup}', "");
Expect(0, 127487, '\p{^Blk=enclosedalphanumsup}', "");
Expect(0, 127487, '\P{Blk=enclosedalphanumsup}', "");
Expect(1, 127487, '\P{^Blk=enclosedalphanumsup}', "");
Expect(0, 127488, '\p{Blk=enclosedalphanumsup}', "");
Expect(1, 127488, '\p{^Blk=enclosedalphanumsup}', "");
Expect(1, 127488, '\P{Blk=enclosedalphanumsup}', "");
Expect(0, 127488, '\P{^Blk=enclosedalphanumsup}', "");
Expect(1, 127487, '\p{Blk=-_ENCLOSED_alphanum_Sup}', "");
Expect(0, 127487, '\p{^Blk=-_ENCLOSED_alphanum_Sup}', "");
Expect(0, 127487, '\P{Blk=-_ENCLOSED_alphanum_Sup}', "");
Expect(1, 127487, '\P{^Blk=-_ENCLOSED_alphanum_Sup}', "");
Expect(0, 127488, '\p{Blk=-_ENCLOSED_alphanum_Sup}', "");
Expect(1, 127488, '\p{^Blk=-_ENCLOSED_alphanum_Sup}', "");
Expect(1, 127488, '\P{Blk=-_ENCLOSED_alphanum_Sup}', "");
Expect(0, 127488, '\P{^Blk=-_ENCLOSED_alphanum_Sup}', "");
Error('\p{Is_Block=:=_	Enclosed_Alphanumeric_Supplement}');
Error('\P{Is_Block=:=_	Enclosed_Alphanumeric_Supplement}');
Expect(1, 127487, '\p{Is_Block=enclosedalphanumericsupplement}', "");
Expect(0, 127487, '\p{^Is_Block=enclosedalphanumericsupplement}', "");
Expect(0, 127487, '\P{Is_Block=enclosedalphanumericsupplement}', "");
Expect(1, 127487, '\P{^Is_Block=enclosedalphanumericsupplement}', "");
Expect(0, 127488, '\p{Is_Block=enclosedalphanumericsupplement}', "");
Expect(1, 127488, '\p{^Is_Block=enclosedalphanumericsupplement}', "");
Expect(1, 127488, '\P{Is_Block=enclosedalphanumericsupplement}', "");
Expect(0, 127488, '\P{^Is_Block=enclosedalphanumericsupplement}', "");
Expect(1, 127487, '\p{Is_Block=		Enclosed_Alphanumeric_supplement}', "");
Expect(0, 127487, '\p{^Is_Block=		Enclosed_Alphanumeric_supplement}', "");
Expect(0, 127487, '\P{Is_Block=		Enclosed_Alphanumeric_supplement}', "");
Expect(1, 127487, '\P{^Is_Block=		Enclosed_Alphanumeric_supplement}', "");
Expect(0, 127488, '\p{Is_Block=		Enclosed_Alphanumeric_supplement}', "");
Expect(1, 127488, '\p{^Is_Block=		Enclosed_Alphanumeric_supplement}', "");
Expect(1, 127488, '\P{Is_Block=		Enclosed_Alphanumeric_supplement}', "");
Expect(0, 127488, '\P{^Is_Block=		Enclosed_Alphanumeric_supplement}', "");
Error('\p{Is_Blk=	/a/Enclosed_ALPHANUM_sup}');
Error('\P{Is_Blk=	/a/Enclosed_ALPHANUM_sup}');
Expect(1, 127487, '\p{Is_Blk=enclosedalphanumsup}', "");
Expect(0, 127487, '\p{^Is_Blk=enclosedalphanumsup}', "");
Expect(0, 127487, '\P{Is_Blk=enclosedalphanumsup}', "");
Expect(1, 127487, '\P{^Is_Blk=enclosedalphanumsup}', "");
Expect(0, 127488, '\p{Is_Blk=enclosedalphanumsup}', "");
Expect(1, 127488, '\p{^Is_Blk=enclosedalphanumsup}', "");
Expect(1, 127488, '\P{Is_Blk=enclosedalphanumsup}', "");
Expect(0, 127488, '\P{^Is_Blk=enclosedalphanumsup}', "");
Expect(1, 127487, '\p{Is_Blk=-Enclosed_Alphanum_Sup}', "");
Expect(0, 127487, '\p{^Is_Blk=-Enclosed_Alphanum_Sup}', "");
Expect(0, 127487, '\P{Is_Blk=-Enclosed_Alphanum_Sup}', "");
Expect(1, 127487, '\P{^Is_Blk=-Enclosed_Alphanum_Sup}', "");
Expect(0, 127488, '\p{Is_Blk=-Enclosed_Alphanum_Sup}', "");
Expect(1, 127488, '\p{^Is_Blk=-Enclosed_Alphanum_Sup}', "");
Expect(1, 127488, '\P{Is_Blk=-Enclosed_Alphanum_Sup}', "");
Expect(0, 127488, '\P{^Is_Blk=-Enclosed_Alphanum_Sup}', "");
Error('\p{Block= /a/enclosed_CJK_Letters_AND_months}');
Error('\P{Block= /a/enclosed_CJK_Letters_AND_months}');
Expect(1, 13055, '\p{Block=enclosedcjklettersandmonths}', "");
Expect(0, 13055, '\p{^Block=enclosedcjklettersandmonths}', "");
Expect(0, 13055, '\P{Block=enclosedcjklettersandmonths}', "");
Expect(1, 13055, '\P{^Block=enclosedcjklettersandmonths}', "");
Expect(0, 13056, '\p{Block=enclosedcjklettersandmonths}', "");
Expect(1, 13056, '\p{^Block=enclosedcjklettersandmonths}', "");
Expect(1, 13056, '\P{Block=enclosedcjklettersandmonths}', "");
Expect(0, 13056, '\P{^Block=enclosedcjklettersandmonths}', "");
Expect(1, 13055, '\p{Block= Enclosed_cjk_Letters_AND_Months}', "");
Expect(0, 13055, '\p{^Block= Enclosed_cjk_Letters_AND_Months}', "");
Expect(0, 13055, '\P{Block= Enclosed_cjk_Letters_AND_Months}', "");
Expect(1, 13055, '\P{^Block= Enclosed_cjk_Letters_AND_Months}', "");
Expect(0, 13056, '\p{Block= Enclosed_cjk_Letters_AND_Months}', "");
Expect(1, 13056, '\p{^Block= Enclosed_cjk_Letters_AND_Months}', "");
Expect(1, 13056, '\P{Block= Enclosed_cjk_Letters_AND_Months}', "");
Expect(0, 13056, '\P{^Block= Enclosed_cjk_Letters_AND_Months}', "");
Error('\p{Blk=:= 	Enclosed_cjk}');
Error('\P{Blk=:= 	Enclosed_cjk}');
Expect(1, 13055, '\p{Blk=enclosedcjk}', "");
Expect(0, 13055, '\p{^Blk=enclosedcjk}', "");
Expect(0, 13055, '\P{Blk=enclosedcjk}', "");
Expect(1, 13055, '\P{^Blk=enclosedcjk}', "");
Expect(0, 13056, '\p{Blk=enclosedcjk}', "");
Expect(1, 13056, '\p{^Blk=enclosedcjk}', "");
Expect(1, 13056, '\P{Blk=enclosedcjk}', "");
Expect(0, 13056, '\P{^Blk=enclosedcjk}', "");
Expect(1, 13055, '\p{Blk=_-enclosed_CJK}', "");
Expect(0, 13055, '\p{^Blk=_-enclosed_CJK}', "");
Expect(0, 13055, '\P{Blk=_-enclosed_CJK}', "");
Expect(1, 13055, '\P{^Blk=_-enclosed_CJK}', "");
Expect(0, 13056, '\p{Blk=_-enclosed_CJK}', "");
Expect(1, 13056, '\p{^Blk=_-enclosed_CJK}', "");
Expect(1, 13056, '\P{Blk=_-enclosed_CJK}', "");
Expect(0, 13056, '\P{^Blk=_-enclosed_CJK}', "");
Error('\p{Is_Block=-:=Enclosed_CJK_letters_AND_Months}');
Error('\P{Is_Block=-:=Enclosed_CJK_letters_AND_Months}');
Expect(1, 13055, '\p{Is_Block=enclosedcjklettersandmonths}', "");
Expect(0, 13055, '\p{^Is_Block=enclosedcjklettersandmonths}', "");
Expect(0, 13055, '\P{Is_Block=enclosedcjklettersandmonths}', "");
Expect(1, 13055, '\P{^Is_Block=enclosedcjklettersandmonths}', "");
Expect(0, 13056, '\p{Is_Block=enclosedcjklettersandmonths}', "");
Expect(1, 13056, '\p{^Is_Block=enclosedcjklettersandmonths}', "");
Expect(1, 13056, '\P{Is_Block=enclosedcjklettersandmonths}', "");
Expect(0, 13056, '\P{^Is_Block=enclosedcjklettersandmonths}', "");
Expect(1, 13055, '\p{Is_Block= _ENCLOSED_CJK_letters_and_MONTHS}', "");
Expect(0, 13055, '\p{^Is_Block= _ENCLOSED_CJK_letters_and_MONTHS}', "");
Expect(0, 13055, '\P{Is_Block= _ENCLOSED_CJK_letters_and_MONTHS}', "");
Expect(1, 13055, '\P{^Is_Block= _ENCLOSED_CJK_letters_and_MONTHS}', "");
Expect(0, 13056, '\p{Is_Block= _ENCLOSED_CJK_letters_and_MONTHS}', "");
Expect(1, 13056, '\p{^Is_Block= _ENCLOSED_CJK_letters_and_MONTHS}', "");
Expect(1, 13056, '\P{Is_Block= _ENCLOSED_CJK_letters_and_MONTHS}', "");
Expect(0, 13056, '\P{^Is_Block= _ENCLOSED_CJK_letters_and_MONTHS}', "");
Error('\p{Is_Blk=_Enclosed_CJK:=}');
Error('\P{Is_Blk=_Enclosed_CJK:=}');
Expect(1, 13055, '\p{Is_Blk=enclosedcjk}', "");
Expect(0, 13055, '\p{^Is_Blk=enclosedcjk}', "");
Expect(0, 13055, '\P{Is_Blk=enclosedcjk}', "");
Expect(1, 13055, '\P{^Is_Blk=enclosedcjk}', "");
Expect(0, 13056, '\p{Is_Blk=enclosedcjk}', "");
Expect(1, 13056, '\p{^Is_Blk=enclosedcjk}', "");
Expect(1, 13056, '\P{Is_Blk=enclosedcjk}', "");
Expect(0, 13056, '\P{^Is_Blk=enclosedcjk}', "");
Expect(1, 13055, '\p{Is_Blk=	 Enclosed_CJK}', "");
Expect(0, 13055, '\p{^Is_Blk=	 Enclosed_CJK}', "");
Expect(0, 13055, '\P{Is_Blk=	 Enclosed_CJK}', "");
Expect(1, 13055, '\P{^Is_Blk=	 Enclosed_CJK}', "");
Expect(0, 13056, '\p{Is_Blk=	 Enclosed_CJK}', "");
Expect(1, 13056, '\p{^Is_Blk=	 Enclosed_CJK}', "");
Expect(1, 13056, '\P{Is_Blk=	 Enclosed_CJK}', "");
Expect(0, 13056, '\P{^Is_Blk=	 Enclosed_CJK}', "");
Error('\p{Block= ENCLOSED_Ideographic_Supplement:=}');
Error('\P{Block= ENCLOSED_Ideographic_Supplement:=}');
Expect(1, 127743, '\p{Block=enclosedideographicsupplement}', "");
Expect(0, 127743, '\p{^Block=enclosedideographicsupplement}', "");
Expect(0, 127743, '\P{Block=enclosedideographicsupplement}', "");
Expect(1, 127743, '\P{^Block=enclosedideographicsupplement}', "");
Expect(0, 127744, '\p{Block=enclosedideographicsupplement}', "");
Expect(1, 127744, '\p{^Block=enclosedideographicsupplement}', "");
Expect(1, 127744, '\P{Block=enclosedideographicsupplement}', "");
Expect(0, 127744, '\P{^Block=enclosedideographicsupplement}', "");
Expect(1, 127743, '\p{Block=_-ENCLOSED_IDEOGRAPHIC_Supplement}', "");
Expect(0, 127743, '\p{^Block=_-ENCLOSED_IDEOGRAPHIC_Supplement}', "");
Expect(0, 127743, '\P{Block=_-ENCLOSED_IDEOGRAPHIC_Supplement}', "");
Expect(1, 127743, '\P{^Block=_-ENCLOSED_IDEOGRAPHIC_Supplement}', "");
Expect(0, 127744, '\p{Block=_-ENCLOSED_IDEOGRAPHIC_Supplement}', "");
Expect(1, 127744, '\p{^Block=_-ENCLOSED_IDEOGRAPHIC_Supplement}', "");
Expect(1, 127744, '\P{Block=_-ENCLOSED_IDEOGRAPHIC_Supplement}', "");
Expect(0, 127744, '\P{^Block=_-ENCLOSED_IDEOGRAPHIC_Supplement}', "");
Error('\p{Blk: /a/	 Enclosed_ideographic_sup}');
Error('\P{Blk: /a/	 Enclosed_ideographic_sup}');
Expect(1, 127743, '\p{Blk=enclosedideographicsup}', "");
Expect(0, 127743, '\p{^Blk=enclosedideographicsup}', "");
Expect(0, 127743, '\P{Blk=enclosedideographicsup}', "");
Expect(1, 127743, '\P{^Blk=enclosedideographicsup}', "");
Expect(0, 127744, '\p{Blk=enclosedideographicsup}', "");
Expect(1, 127744, '\p{^Blk=enclosedideographicsup}', "");
Expect(1, 127744, '\P{Blk=enclosedideographicsup}', "");
Expect(0, 127744, '\P{^Blk=enclosedideographicsup}', "");
Expect(1, 127743, '\p{Blk= ENCLOSED_IDEOGRAPHIC_SUP}', "");
Expect(0, 127743, '\p{^Blk= ENCLOSED_IDEOGRAPHIC_SUP}', "");
Expect(0, 127743, '\P{Blk= ENCLOSED_IDEOGRAPHIC_SUP}', "");
Expect(1, 127743, '\P{^Blk= ENCLOSED_IDEOGRAPHIC_SUP}', "");
Expect(0, 127744, '\p{Blk= ENCLOSED_IDEOGRAPHIC_SUP}', "");
Expect(1, 127744, '\p{^Blk= ENCLOSED_IDEOGRAPHIC_SUP}', "");
Expect(1, 127744, '\P{Blk= ENCLOSED_IDEOGRAPHIC_SUP}', "");
Expect(0, 127744, '\P{^Blk= ENCLOSED_IDEOGRAPHIC_SUP}', "");
Error('\p{Is_Block=_-Enclosed_Ideographic_SUPPLEMENT:=}');
Error('\P{Is_Block=_-Enclosed_Ideographic_SUPPLEMENT:=}');
Expect(1, 127743, '\p{Is_Block=enclosedideographicsupplement}', "");
Expect(0, 127743, '\p{^Is_Block=enclosedideographicsupplement}', "");
Expect(0, 127743, '\P{Is_Block=enclosedideographicsupplement}', "");
Expect(1, 127743, '\P{^Is_Block=enclosedideographicsupplement}', "");
Expect(0, 127744, '\p{Is_Block=enclosedideographicsupplement}', "");
Expect(1, 127744, '\p{^Is_Block=enclosedideographicsupplement}', "");
Expect(1, 127744, '\P{Is_Block=enclosedideographicsupplement}', "");
Expect(0, 127744, '\P{^Is_Block=enclosedideographicsupplement}', "");
Expect(1, 127743, '\p{Is_Block=_Enclosed_IDEOGRAPHIC_supplement}', "");
Expect(0, 127743, '\p{^Is_Block=_Enclosed_IDEOGRAPHIC_supplement}', "");
Expect(0, 127743, '\P{Is_Block=_Enclosed_IDEOGRAPHIC_supplement}', "");
Expect(1, 127743, '\P{^Is_Block=_Enclosed_IDEOGRAPHIC_supplement}', "");
Expect(0, 127744, '\p{Is_Block=_Enclosed_IDEOGRAPHIC_supplement}', "");
Expect(1, 127744, '\p{^Is_Block=_Enclosed_IDEOGRAPHIC_supplement}', "");
Expect(1, 127744, '\P{Is_Block=_Enclosed_IDEOGRAPHIC_supplement}', "");
Expect(0, 127744, '\P{^Is_Block=_Enclosed_IDEOGRAPHIC_supplement}', "");
Error('\p{Is_Blk=	:=Enclosed_Ideographic_SUP}');
Error('\P{Is_Blk=	:=Enclosed_Ideographic_SUP}');
Expect(1, 127743, '\p{Is_Blk=enclosedideographicsup}', "");
Expect(0, 127743, '\p{^Is_Blk=enclosedideographicsup}', "");
Expect(0, 127743, '\P{Is_Blk=enclosedideographicsup}', "");
Expect(1, 127743, '\P{^Is_Blk=enclosedideographicsup}', "");
Expect(0, 127744, '\p{Is_Blk=enclosedideographicsup}', "");
Expect(1, 127744, '\p{^Is_Blk=enclosedideographicsup}', "");
Expect(1, 127744, '\P{Is_Blk=enclosedideographicsup}', "");
Expect(0, 127744, '\P{^Is_Blk=enclosedideographicsup}', "");
Expect(1, 127743, '\p{Is_Blk=-enclosed_Ideographic_Sup}', "");
Expect(0, 127743, '\p{^Is_Blk=-enclosed_Ideographic_Sup}', "");
Expect(0, 127743, '\P{Is_Blk=-enclosed_Ideographic_Sup}', "");
Expect(1, 127743, '\P{^Is_Blk=-enclosed_Ideographic_Sup}', "");
Expect(0, 127744, '\p{Is_Blk=-enclosed_Ideographic_Sup}', "");
Expect(1, 127744, '\p{^Is_Blk=-enclosed_Ideographic_Sup}', "");
Expect(1, 127744, '\P{Is_Blk=-enclosed_Ideographic_Sup}', "");
Expect(0, 127744, '\P{^Is_Blk=-enclosed_Ideographic_Sup}', "");
Error('\p{Block=/a/ethiopic}');
Error('\P{Block=/a/ethiopic}');
Expect(1, 4991, '\p{Block=ethiopic}', "");
Expect(0, 4991, '\p{^Block=ethiopic}', "");
Expect(0, 4991, '\P{Block=ethiopic}', "");
Expect(1, 4991, '\P{^Block=ethiopic}', "");
Expect(0, 4992, '\p{Block=ethiopic}', "");
Expect(1, 4992, '\p{^Block=ethiopic}', "");
Expect(1, 4992, '\P{Block=ethiopic}', "");
Expect(0, 4992, '\P{^Block=ethiopic}', "");
Expect(1, 4991, '\p{Block=_	ethiopic}', "");
Expect(0, 4991, '\p{^Block=_	ethiopic}', "");
Expect(0, 4991, '\P{Block=_	ethiopic}', "");
Expect(1, 4991, '\P{^Block=_	ethiopic}', "");
Expect(0, 4992, '\p{Block=_	ethiopic}', "");
Expect(1, 4992, '\p{^Block=_	ethiopic}', "");
Expect(1, 4992, '\P{Block=_	ethiopic}', "");
Expect(0, 4992, '\P{^Block=_	ethiopic}', "");
Error('\p{Blk=	 Ethiopic/a/}');
Error('\P{Blk=	 Ethiopic/a/}');
Expect(1, 4991, '\p{Blk=ethiopic}', "");
Expect(0, 4991, '\p{^Blk=ethiopic}', "");
Expect(0, 4991, '\P{Blk=ethiopic}', "");
Expect(1, 4991, '\P{^Blk=ethiopic}', "");
Expect(0, 4992, '\p{Blk=ethiopic}', "");
Expect(1, 4992, '\p{^Blk=ethiopic}', "");
Expect(1, 4992, '\P{Blk=ethiopic}', "");
Expect(0, 4992, '\P{^Blk=ethiopic}', "");
Expect(1, 4991, '\p{Blk=_ Ethiopic}', "");
Expect(0, 4991, '\p{^Blk=_ Ethiopic}', "");
Expect(0, 4991, '\P{Blk=_ Ethiopic}', "");
Expect(1, 4991, '\P{^Blk=_ Ethiopic}', "");
Expect(0, 4992, '\p{Blk=_ Ethiopic}', "");
Expect(1, 4992, '\p{^Blk=_ Ethiopic}', "");
Expect(1, 4992, '\P{Blk=_ Ethiopic}', "");
Expect(0, 4992, '\P{^Blk=_ Ethiopic}', "");
Error('\p{Is_Block=/a/-_ETHIOPIC}');
Error('\P{Is_Block=/a/-_ETHIOPIC}');
Expect(1, 4991, '\p{Is_Block=ethiopic}', "");
Expect(0, 4991, '\p{^Is_Block=ethiopic}', "");
Expect(0, 4991, '\P{Is_Block=ethiopic}', "");
Expect(1, 4991, '\P{^Is_Block=ethiopic}', "");
Expect(0, 4992, '\p{Is_Block=ethiopic}', "");
Expect(1, 4992, '\p{^Is_Block=ethiopic}', "");
Expect(1, 4992, '\P{Is_Block=ethiopic}', "");
Expect(0, 4992, '\P{^Is_Block=ethiopic}', "");
Expect(1, 4991, '\p{Is_Block=_Ethiopic}', "");
Expect(0, 4991, '\p{^Is_Block=_Ethiopic}', "");
Expect(0, 4991, '\P{Is_Block=_Ethiopic}', "");
Expect(1, 4991, '\P{^Is_Block=_Ethiopic}', "");
Expect(0, 4992, '\p{Is_Block=_Ethiopic}', "");
Expect(1, 4992, '\p{^Is_Block=_Ethiopic}', "");
Expect(1, 4992, '\P{Is_Block=_Ethiopic}', "");
Expect(0, 4992, '\P{^Is_Block=_Ethiopic}', "");
Error('\p{Is_Blk=-/a/ETHIOPIC}');
Error('\P{Is_Blk=-/a/ETHIOPIC}');
Expect(1, 4991, '\p{Is_Blk=ethiopic}', "");
Expect(0, 4991, '\p{^Is_Blk=ethiopic}', "");
Expect(0, 4991, '\P{Is_Blk=ethiopic}', "");
Expect(1, 4991, '\P{^Is_Blk=ethiopic}', "");
Expect(0, 4992, '\p{Is_Blk=ethiopic}', "");
Expect(1, 4992, '\p{^Is_Blk=ethiopic}', "");
Expect(1, 4992, '\P{Is_Blk=ethiopic}', "");
Expect(0, 4992, '\P{^Is_Blk=ethiopic}', "");
Expect(1, 4991, '\p{Is_Blk: 	Ethiopic}', "");
Expect(0, 4991, '\p{^Is_Blk: 	Ethiopic}', "");
Expect(0, 4991, '\P{Is_Blk: 	Ethiopic}', "");
Expect(1, 4991, '\P{^Is_Blk: 	Ethiopic}', "");
Expect(0, 4992, '\p{Is_Blk: 	Ethiopic}', "");
Expect(1, 4992, '\p{^Is_Blk: 	Ethiopic}', "");
Expect(1, 4992, '\P{Is_Blk: 	Ethiopic}', "");
Expect(0, 4992, '\P{^Is_Blk: 	Ethiopic}', "");
Error('\p{Block=/a/-_Ethiopic_Extended}');
Error('\P{Block=/a/-_Ethiopic_Extended}');
Expect(1, 11743, '\p{Block=ethiopicextended}', "");
Expect(0, 11743, '\p{^Block=ethiopicextended}', "");
Expect(0, 11743, '\P{Block=ethiopicextended}', "");
Expect(1, 11743, '\P{^Block=ethiopicextended}', "");
Expect(0, 11744, '\p{Block=ethiopicextended}', "");
Expect(1, 11744, '\p{^Block=ethiopicextended}', "");
Expect(1, 11744, '\P{Block=ethiopicextended}', "");
Expect(0, 11744, '\P{^Block=ethiopicextended}', "");
Expect(1, 11743, '\p{Block=ethiopic_Extended}', "");
Expect(0, 11743, '\p{^Block=ethiopic_Extended}', "");
Expect(0, 11743, '\P{Block=ethiopic_Extended}', "");
Expect(1, 11743, '\P{^Block=ethiopic_Extended}', "");
Expect(0, 11744, '\p{Block=ethiopic_Extended}', "");
Expect(1, 11744, '\p{^Block=ethiopic_Extended}', "");
Expect(1, 11744, '\P{Block=ethiopic_Extended}', "");
Expect(0, 11744, '\P{^Block=ethiopic_Extended}', "");
Error('\p{Blk=__ethiopic_Ext:=}');
Error('\P{Blk=__ethiopic_Ext:=}');
Expect(1, 11743, '\p{Blk=ethiopicext}', "");
Expect(0, 11743, '\p{^Blk=ethiopicext}', "");
Expect(0, 11743, '\P{Blk=ethiopicext}', "");
Expect(1, 11743, '\P{^Blk=ethiopicext}', "");
Expect(0, 11744, '\p{Blk=ethiopicext}', "");
Expect(1, 11744, '\p{^Blk=ethiopicext}', "");
Expect(1, 11744, '\P{Blk=ethiopicext}', "");
Expect(0, 11744, '\P{^Blk=ethiopicext}', "");
Expect(1, 11743, '\p{Blk=  ethiopic_EXT}', "");
Expect(0, 11743, '\p{^Blk=  ethiopic_EXT}', "");
Expect(0, 11743, '\P{Blk=  ethiopic_EXT}', "");
Expect(1, 11743, '\P{^Blk=  ethiopic_EXT}', "");
Expect(0, 11744, '\p{Blk=  ethiopic_EXT}', "");
Expect(1, 11744, '\p{^Blk=  ethiopic_EXT}', "");
Expect(1, 11744, '\P{Blk=  ethiopic_EXT}', "");
Expect(0, 11744, '\P{^Blk=  ethiopic_EXT}', "");
Error('\p{Is_Block=:=-	Ethiopic_EXTENDED}');
Error('\P{Is_Block=:=-	Ethiopic_EXTENDED}');
Expect(1, 11743, '\p{Is_Block=ethiopicextended}', "");
Expect(0, 11743, '\p{^Is_Block=ethiopicextended}', "");
Expect(0, 11743, '\P{Is_Block=ethiopicextended}', "");
Expect(1, 11743, '\P{^Is_Block=ethiopicextended}', "");
Expect(0, 11744, '\p{Is_Block=ethiopicextended}', "");
Expect(1, 11744, '\p{^Is_Block=ethiopicextended}', "");
Expect(1, 11744, '\P{Is_Block=ethiopicextended}', "");
Expect(0, 11744, '\P{^Is_Block=ethiopicextended}', "");
Expect(1, 11743, '\p{Is_Block=_Ethiopic_extended}', "");
Expect(0, 11743, '\p{^Is_Block=_Ethiopic_extended}', "");
Expect(0, 11743, '\P{Is_Block=_Ethiopic_extended}', "");
Expect(1, 11743, '\P{^Is_Block=_Ethiopic_extended}', "");
Expect(0, 11744, '\p{Is_Block=_Ethiopic_extended}', "");
Expect(1, 11744, '\p{^Is_Block=_Ethiopic_extended}', "");
Expect(1, 11744, '\P{Is_Block=_Ethiopic_extended}', "");
Expect(0, 11744, '\P{^Is_Block=_Ethiopic_extended}', "");
Error('\p{Is_Blk=	 ethiopic_Ext/a/}');
Error('\P{Is_Blk=	 ethiopic_Ext/a/}');
Expect(1, 11743, '\p{Is_Blk=ethiopicext}', "");
Expect(0, 11743, '\p{^Is_Blk=ethiopicext}', "");
Expect(0, 11743, '\P{Is_Blk=ethiopicext}', "");
Expect(1, 11743, '\P{^Is_Blk=ethiopicext}', "");
Expect(0, 11744, '\p{Is_Blk=ethiopicext}', "");
Expect(1, 11744, '\p{^Is_Blk=ethiopicext}', "");
Expect(1, 11744, '\P{Is_Blk=ethiopicext}', "");
Expect(0, 11744, '\P{^Is_Blk=ethiopicext}', "");
Expect(1, 11743, '\p{Is_Blk=	ethiopic_ext}', "");
Expect(0, 11743, '\p{^Is_Blk=	ethiopic_ext}', "");
Expect(0, 11743, '\P{Is_Blk=	ethiopic_ext}', "");
Expect(1, 11743, '\P{^Is_Blk=	ethiopic_ext}', "");
Expect(0, 11744, '\p{Is_Blk=	ethiopic_ext}', "");
Expect(1, 11744, '\p{^Is_Blk=	ethiopic_ext}', "");
Expect(1, 11744, '\P{Is_Blk=	ethiopic_ext}', "");
Expect(0, 11744, '\P{^Is_Blk=	ethiopic_ext}', "");
Error('\p{Block:- ETHIOPIC_EXTENDED_a/a/}');
Error('\P{Block:- ETHIOPIC_EXTENDED_a/a/}');
Expect(1, 43823, '\p{Block=ethiopicextendeda}', "");
Expect(0, 43823, '\p{^Block=ethiopicextendeda}', "");
Expect(0, 43823, '\P{Block=ethiopicextendeda}', "");
Expect(1, 43823, '\P{^Block=ethiopicextendeda}', "");
Expect(0, 43824, '\p{Block=ethiopicextendeda}', "");
Expect(1, 43824, '\p{^Block=ethiopicextendeda}', "");
Expect(1, 43824, '\P{Block=ethiopicextendeda}', "");
Expect(0, 43824, '\P{^Block=ethiopicextendeda}', "");
Expect(1, 43823, '\p{Block=-Ethiopic_extended_A}', "");
Expect(0, 43823, '\p{^Block=-Ethiopic_extended_A}', "");
Expect(0, 43823, '\P{Block=-Ethiopic_extended_A}', "");
Expect(1, 43823, '\P{^Block=-Ethiopic_extended_A}', "");
Expect(0, 43824, '\p{Block=-Ethiopic_extended_A}', "");
Expect(1, 43824, '\p{^Block=-Ethiopic_extended_A}', "");
Expect(1, 43824, '\P{Block=-Ethiopic_extended_A}', "");
Expect(0, 43824, '\P{^Block=-Ethiopic_extended_A}', "");
Error('\p{Blk=:=-ETHIOPIC_Ext_a}');
Error('\P{Blk=:=-ETHIOPIC_Ext_a}');
Expect(1, 43823, '\p{Blk=ethiopicexta}', "");
Expect(0, 43823, '\p{^Blk=ethiopicexta}', "");
Expect(0, 43823, '\P{Blk=ethiopicexta}', "");
Expect(1, 43823, '\P{^Blk=ethiopicexta}', "");
Expect(0, 43824, '\p{Blk=ethiopicexta}', "");
Expect(1, 43824, '\p{^Blk=ethiopicexta}', "");
Expect(1, 43824, '\P{Blk=ethiopicexta}', "");
Expect(0, 43824, '\P{^Blk=ethiopicexta}', "");
Expect(1, 43823, '\p{Blk= -ethiopic_Ext_a}', "");
Expect(0, 43823, '\p{^Blk= -ethiopic_Ext_a}', "");
Expect(0, 43823, '\P{Blk= -ethiopic_Ext_a}', "");
Expect(1, 43823, '\P{^Blk= -ethiopic_Ext_a}', "");
Expect(0, 43824, '\p{Blk= -ethiopic_Ext_a}', "");
Expect(1, 43824, '\p{^Blk= -ethiopic_Ext_a}', "");
Expect(1, 43824, '\P{Blk= -ethiopic_Ext_a}', "");
Expect(0, 43824, '\P{^Blk= -ethiopic_Ext_a}', "");
Error('\p{Is_Block=:=		Ethiopic_Extended_A}');
Error('\P{Is_Block=:=		Ethiopic_Extended_A}');
Expect(1, 43823, '\p{Is_Block=ethiopicextendeda}', "");
Expect(0, 43823, '\p{^Is_Block=ethiopicextendeda}', "");
Expect(0, 43823, '\P{Is_Block=ethiopicextendeda}', "");
Expect(1, 43823, '\P{^Is_Block=ethiopicextendeda}', "");
Expect(0, 43824, '\p{Is_Block=ethiopicextendeda}', "");
Expect(1, 43824, '\p{^Is_Block=ethiopicextendeda}', "");
Expect(1, 43824, '\P{Is_Block=ethiopicextendeda}', "");
Expect(0, 43824, '\P{^Is_Block=ethiopicextendeda}', "");
Expect(1, 43823, '\p{Is_Block: 	_Ethiopic_extended_a}', "");
Expect(0, 43823, '\p{^Is_Block: 	_Ethiopic_extended_a}', "");
Expect(0, 43823, '\P{Is_Block: 	_Ethiopic_extended_a}', "");
Expect(1, 43823, '\P{^Is_Block: 	_Ethiopic_extended_a}', "");
Expect(0, 43824, '\p{Is_Block: 	_Ethiopic_extended_a}', "");
Expect(1, 43824, '\p{^Is_Block: 	_Ethiopic_extended_a}', "");
Expect(1, 43824, '\P{Is_Block: 	_Ethiopic_extended_a}', "");
Expect(0, 43824, '\P{^Is_Block: 	_Ethiopic_extended_a}', "");
Error('\p{Is_Blk:  Ethiopic_Ext_A:=}');
Error('\P{Is_Blk:  Ethiopic_Ext_A:=}');
Expect(1, 43823, '\p{Is_Blk=ethiopicexta}', "");
Expect(0, 43823, '\p{^Is_Blk=ethiopicexta}', "");
Expect(0, 43823, '\P{Is_Blk=ethiopicexta}', "");
Expect(1, 43823, '\P{^Is_Blk=ethiopicexta}', "");
Expect(0, 43824, '\p{Is_Blk=ethiopicexta}', "");
Expect(1, 43824, '\p{^Is_Blk=ethiopicexta}', "");
Expect(1, 43824, '\P{Is_Blk=ethiopicexta}', "");
Expect(0, 43824, '\P{^Is_Blk=ethiopicexta}', "");
Expect(1, 43823, '\p{Is_Blk=_ Ethiopic_Ext_A}', "");
Expect(0, 43823, '\p{^Is_Blk=_ Ethiopic_Ext_A}', "");
Expect(0, 43823, '\P{Is_Blk=_ Ethiopic_Ext_A}', "");
Expect(1, 43823, '\P{^Is_Blk=_ Ethiopic_Ext_A}', "");
Expect(0, 43824, '\p{Is_Blk=_ Ethiopic_Ext_A}', "");
Expect(1, 43824, '\p{^Is_Blk=_ Ethiopic_Ext_A}', "");
Expect(1, 43824, '\P{Is_Blk=_ Ethiopic_Ext_A}', "");
Expect(0, 43824, '\P{^Is_Blk=_ Ethiopic_Ext_A}', "");
Error('\p{Block=_:=Ethiopic_SUPPLEMENT}');
Error('\P{Block=_:=Ethiopic_SUPPLEMENT}');
Expect(1, 5023, '\p{Block=ethiopicsupplement}', "");
Expect(0, 5023, '\p{^Block=ethiopicsupplement}', "");
Expect(0, 5023, '\P{Block=ethiopicsupplement}', "");
Expect(1, 5023, '\P{^Block=ethiopicsupplement}', "");
Expect(0, 5024, '\p{Block=ethiopicsupplement}', "");
Expect(1, 5024, '\p{^Block=ethiopicsupplement}', "");
Expect(1, 5024, '\P{Block=ethiopicsupplement}', "");
Expect(0, 5024, '\P{^Block=ethiopicsupplement}', "");
Expect(1, 5023, '\p{Block=_	Ethiopic_SUPPLEMENT}', "");
Expect(0, 5023, '\p{^Block=_	Ethiopic_SUPPLEMENT}', "");
Expect(0, 5023, '\P{Block=_	Ethiopic_SUPPLEMENT}', "");
Expect(1, 5023, '\P{^Block=_	Ethiopic_SUPPLEMENT}', "");
Expect(0, 5024, '\p{Block=_	Ethiopic_SUPPLEMENT}', "");
Expect(1, 5024, '\p{^Block=_	Ethiopic_SUPPLEMENT}', "");
Expect(1, 5024, '\P{Block=_	Ethiopic_SUPPLEMENT}', "");
Expect(0, 5024, '\P{^Block=_	Ethiopic_SUPPLEMENT}', "");
Error('\p{Blk=/a/ Ethiopic_sup}');
Error('\P{Blk=/a/ Ethiopic_sup}');
Expect(1, 5023, '\p{Blk=ethiopicsup}', "");
Expect(0, 5023, '\p{^Blk=ethiopicsup}', "");
Expect(0, 5023, '\P{Blk=ethiopicsup}', "");
Expect(1, 5023, '\P{^Blk=ethiopicsup}', "");
Expect(0, 5024, '\p{Blk=ethiopicsup}', "");
Expect(1, 5024, '\p{^Blk=ethiopicsup}', "");
Expect(1, 5024, '\P{Blk=ethiopicsup}', "");
Expect(0, 5024, '\P{^Blk=ethiopicsup}', "");
Expect(1, 5023, '\p{Blk= ETHIOPIC_Sup}', "");
Expect(0, 5023, '\p{^Blk= ETHIOPIC_Sup}', "");
Expect(0, 5023, '\P{Blk= ETHIOPIC_Sup}', "");
Expect(1, 5023, '\P{^Blk= ETHIOPIC_Sup}', "");
Expect(0, 5024, '\p{Blk= ETHIOPIC_Sup}', "");
Expect(1, 5024, '\p{^Blk= ETHIOPIC_Sup}', "");
Expect(1, 5024, '\P{Blk= ETHIOPIC_Sup}', "");
Expect(0, 5024, '\P{^Blk= ETHIOPIC_Sup}', "");
Error('\p{Is_Block=	Ethiopic_Supplement:=}');
Error('\P{Is_Block=	Ethiopic_Supplement:=}');
Expect(1, 5023, '\p{Is_Block=ethiopicsupplement}', "");
Expect(0, 5023, '\p{^Is_Block=ethiopicsupplement}', "");
Expect(0, 5023, '\P{Is_Block=ethiopicsupplement}', "");
Expect(1, 5023, '\P{^Is_Block=ethiopicsupplement}', "");
Expect(0, 5024, '\p{Is_Block=ethiopicsupplement}', "");
Expect(1, 5024, '\p{^Is_Block=ethiopicsupplement}', "");
Expect(1, 5024, '\P{Is_Block=ethiopicsupplement}', "");
Expect(0, 5024, '\P{^Is_Block=ethiopicsupplement}', "");
Expect(1, 5023, '\p{Is_Block=_	Ethiopic_Supplement}', "");
Expect(0, 5023, '\p{^Is_Block=_	Ethiopic_Supplement}', "");
Expect(0, 5023, '\P{Is_Block=_	Ethiopic_Supplement}', "");
Expect(1, 5023, '\P{^Is_Block=_	Ethiopic_Supplement}', "");
Expect(0, 5024, '\p{Is_Block=_	Ethiopic_Supplement}', "");
Expect(1, 5024, '\p{^Is_Block=_	Ethiopic_Supplement}', "");
Expect(1, 5024, '\P{Is_Block=_	Ethiopic_Supplement}', "");
Expect(0, 5024, '\P{^Is_Block=_	Ethiopic_Supplement}', "");
Error('\p{Is_Blk=/a/_	Ethiopic_Sup}');
Error('\P{Is_Blk=/a/_	Ethiopic_Sup}');
Expect(1, 5023, '\p{Is_Blk=ethiopicsup}', "");
Expect(0, 5023, '\p{^Is_Blk=ethiopicsup}', "");
Expect(0, 5023, '\P{Is_Blk=ethiopicsup}', "");
Expect(1, 5023, '\P{^Is_Blk=ethiopicsup}', "");
Expect(0, 5024, '\p{Is_Blk=ethiopicsup}', "");
Expect(1, 5024, '\p{^Is_Blk=ethiopicsup}', "");
Expect(1, 5024, '\P{Is_Blk=ethiopicsup}', "");
Expect(0, 5024, '\P{^Is_Blk=ethiopicsup}', "");
Expect(1, 5023, '\p{Is_Blk=-ethiopic_Sup}', "");
Expect(0, 5023, '\p{^Is_Blk=-ethiopic_Sup}', "");
Expect(0, 5023, '\P{Is_Blk=-ethiopic_Sup}', "");
Expect(1, 5023, '\P{^Is_Blk=-ethiopic_Sup}', "");
Expect(0, 5024, '\p{Is_Blk=-ethiopic_Sup}', "");
Expect(1, 5024, '\p{^Is_Blk=-ethiopic_Sup}', "");
Expect(1, 5024, '\P{Is_Blk=-ethiopic_Sup}', "");
Expect(0, 5024, '\P{^Is_Blk=-ethiopic_Sup}', "");
Error('\p{Block=/a/--Geometric_shapes}');
Error('\P{Block=/a/--Geometric_shapes}');
Expect(1, 9727, '\p{Block=geometricshapes}', "");
Expect(0, 9727, '\p{^Block=geometricshapes}', "");
Expect(0, 9727, '\P{Block=geometricshapes}', "");
Expect(1, 9727, '\P{^Block=geometricshapes}', "");
Expect(0, 9728, '\p{Block=geometricshapes}', "");
Expect(1, 9728, '\p{^Block=geometricshapes}', "");
Expect(1, 9728, '\P{Block=geometricshapes}', "");
Expect(0, 9728, '\P{^Block=geometricshapes}', "");
Expect(1, 9727, '\p{Block=-geometric_SHAPES}', "");
Expect(0, 9727, '\p{^Block=-geometric_SHAPES}', "");
Expect(0, 9727, '\P{Block=-geometric_SHAPES}', "");
Expect(1, 9727, '\P{^Block=-geometric_SHAPES}', "");
Expect(0, 9728, '\p{Block=-geometric_SHAPES}', "");
Expect(1, 9728, '\p{^Block=-geometric_SHAPES}', "");
Expect(1, 9728, '\P{Block=-geometric_SHAPES}', "");
Expect(0, 9728, '\P{^Block=-geometric_SHAPES}', "");
Error('\p{Blk:    _GEOMETRIC_shapes:=}');
Error('\P{Blk:    _GEOMETRIC_shapes:=}');
Expect(1, 9727, '\p{Blk=geometricshapes}', "");
Expect(0, 9727, '\p{^Blk=geometricshapes}', "");
Expect(0, 9727, '\P{Blk=geometricshapes}', "");
Expect(1, 9727, '\P{^Blk=geometricshapes}', "");
Expect(0, 9728, '\p{Blk=geometricshapes}', "");
Expect(1, 9728, '\p{^Blk=geometricshapes}', "");
Expect(1, 9728, '\P{Blk=geometricshapes}', "");
Expect(0, 9728, '\P{^Blk=geometricshapes}', "");
Expect(1, 9727, '\p{Blk=--Geometric_SHAPES}', "");
Expect(0, 9727, '\p{^Blk=--Geometric_SHAPES}', "");
Expect(0, 9727, '\P{Blk=--Geometric_SHAPES}', "");
Expect(1, 9727, '\P{^Blk=--Geometric_SHAPES}', "");
Expect(0, 9728, '\p{Blk=--Geometric_SHAPES}', "");
Expect(1, 9728, '\p{^Blk=--Geometric_SHAPES}', "");
Expect(1, 9728, '\P{Blk=--Geometric_SHAPES}', "");
Expect(0, 9728, '\P{^Blk=--Geometric_SHAPES}', "");
Error('\p{Is_Block=-/a/Geometric_Shapes}');
Error('\P{Is_Block=-/a/Geometric_Shapes}');
Expect(1, 9727, '\p{Is_Block=geometricshapes}', "");
Expect(0, 9727, '\p{^Is_Block=geometricshapes}', "");
Expect(0, 9727, '\P{Is_Block=geometricshapes}', "");
Expect(1, 9727, '\P{^Is_Block=geometricshapes}', "");
Expect(0, 9728, '\p{Is_Block=geometricshapes}', "");
Expect(1, 9728, '\p{^Is_Block=geometricshapes}', "");
Expect(1, 9728, '\P{Is_Block=geometricshapes}', "");
Expect(0, 9728, '\P{^Is_Block=geometricshapes}', "");
Expect(1, 9727, '\p{Is_Block= _Geometric_shapes}', "");
Expect(0, 9727, '\p{^Is_Block= _Geometric_shapes}', "");
Expect(0, 9727, '\P{Is_Block= _Geometric_shapes}', "");
Expect(1, 9727, '\P{^Is_Block= _Geometric_shapes}', "");
Expect(0, 9728, '\p{Is_Block= _Geometric_shapes}', "");
Expect(1, 9728, '\p{^Is_Block= _Geometric_shapes}', "");
Expect(1, 9728, '\P{Is_Block= _Geometric_shapes}', "");
Expect(0, 9728, '\P{^Is_Block= _Geometric_shapes}', "");
Error('\p{Is_Blk=_/a/Geometric_shapes}');
Error('\P{Is_Blk=_/a/Geometric_shapes}');
Expect(1, 9727, '\p{Is_Blk=geometricshapes}', "");
Expect(0, 9727, '\p{^Is_Blk=geometricshapes}', "");
Expect(0, 9727, '\P{Is_Blk=geometricshapes}', "");
Expect(1, 9727, '\P{^Is_Blk=geometricshapes}', "");
Expect(0, 9728, '\p{Is_Blk=geometricshapes}', "");
Expect(1, 9728, '\p{^Is_Blk=geometricshapes}', "");
Expect(1, 9728, '\P{Is_Blk=geometricshapes}', "");
Expect(0, 9728, '\P{^Is_Blk=geometricshapes}', "");
Expect(1, 9727, '\p{Is_Blk=	_GEOMETRIC_shapes}', "");
Expect(0, 9727, '\p{^Is_Blk=	_GEOMETRIC_shapes}', "");
Expect(0, 9727, '\P{Is_Blk=	_GEOMETRIC_shapes}', "");
Expect(1, 9727, '\P{^Is_Blk=	_GEOMETRIC_shapes}', "");
Expect(0, 9728, '\p{Is_Blk=	_GEOMETRIC_shapes}', "");
Expect(1, 9728, '\p{^Is_Blk=	_GEOMETRIC_shapes}', "");
Expect(1, 9728, '\P{Is_Blk=	_GEOMETRIC_shapes}', "");
Expect(0, 9728, '\P{^Is_Blk=	_GEOMETRIC_shapes}', "");
Error('\p{Block=:=  Georgian}');
Error('\P{Block=:=  Georgian}');
Expect(1, 4351, '\p{Block=georgian}', "");
Expect(0, 4351, '\p{^Block=georgian}', "");
Expect(0, 4351, '\P{Block=georgian}', "");
Expect(1, 4351, '\P{^Block=georgian}', "");
Expect(0, 4352, '\p{Block=georgian}', "");
Expect(1, 4352, '\p{^Block=georgian}', "");
Expect(1, 4352, '\P{Block=georgian}', "");
Expect(0, 4352, '\P{^Block=georgian}', "");
Expect(1, 4351, '\p{Block=	_georgian}', "");
Expect(0, 4351, '\p{^Block=	_georgian}', "");
Expect(0, 4351, '\P{Block=	_georgian}', "");
Expect(1, 4351, '\P{^Block=	_georgian}', "");
Expect(0, 4352, '\p{Block=	_georgian}', "");
Expect(1, 4352, '\p{^Block=	_georgian}', "");
Expect(1, 4352, '\P{Block=	_georgian}', "");
Expect(0, 4352, '\P{^Block=	_georgian}', "");
Error('\p{Blk=:=_Georgian}');
Error('\P{Blk=:=_Georgian}');
Expect(1, 4351, '\p{Blk=georgian}', "");
Expect(0, 4351, '\p{^Blk=georgian}', "");
Expect(0, 4351, '\P{Blk=georgian}', "");
Expect(1, 4351, '\P{^Blk=georgian}', "");
Expect(0, 4352, '\p{Blk=georgian}', "");
Expect(1, 4352, '\p{^Blk=georgian}', "");
Expect(1, 4352, '\P{Blk=georgian}', "");
Expect(0, 4352, '\P{^Blk=georgian}', "");
Expect(1, 4351, '\p{Blk=_	Georgian}', "");
Expect(0, 4351, '\p{^Blk=_	Georgian}', "");
Expect(0, 4351, '\P{Blk=_	Georgian}', "");
Expect(1, 4351, '\P{^Blk=_	Georgian}', "");
Expect(0, 4352, '\p{Blk=_	Georgian}', "");
Expect(1, 4352, '\p{^Blk=_	Georgian}', "");
Expect(1, 4352, '\P{Blk=_	Georgian}', "");
Expect(0, 4352, '\P{^Blk=_	Georgian}', "");
Error('\p{Is_Block=/a/Georgian}');
Error('\P{Is_Block=/a/Georgian}');
Expect(1, 4351, '\p{Is_Block=georgian}', "");
Expect(0, 4351, '\p{^Is_Block=georgian}', "");
Expect(0, 4351, '\P{Is_Block=georgian}', "");
Expect(1, 4351, '\P{^Is_Block=georgian}', "");
Expect(0, 4352, '\p{Is_Block=georgian}', "");
Expect(1, 4352, '\p{^Is_Block=georgian}', "");
Expect(1, 4352, '\P{Is_Block=georgian}', "");
Expect(0, 4352, '\P{^Is_Block=georgian}', "");
Expect(1, 4351, '\p{Is_Block=-Georgian}', "");
Expect(0, 4351, '\p{^Is_Block=-Georgian}', "");
Expect(0, 4351, '\P{Is_Block=-Georgian}', "");
Expect(1, 4351, '\P{^Is_Block=-Georgian}', "");
Expect(0, 4352, '\p{Is_Block=-Georgian}', "");
Expect(1, 4352, '\p{^Is_Block=-Georgian}', "");
Expect(1, 4352, '\P{Is_Block=-Georgian}', "");
Expect(0, 4352, '\P{^Is_Block=-Georgian}', "");
Error('\p{Is_Blk=/a/		Georgian}');
Error('\P{Is_Blk=/a/		Georgian}');
Expect(1, 4351, '\p{Is_Blk=georgian}', "");
Expect(0, 4351, '\p{^Is_Blk=georgian}', "");
Expect(0, 4351, '\P{Is_Blk=georgian}', "");
Expect(1, 4351, '\P{^Is_Blk=georgian}', "");
Expect(0, 4352, '\p{Is_Blk=georgian}', "");
Expect(1, 4352, '\p{^Is_Blk=georgian}', "");
Expect(1, 4352, '\P{Is_Blk=georgian}', "");
Expect(0, 4352, '\P{^Is_Blk=georgian}', "");
Expect(1, 4351, '\p{Is_Blk=_-Georgian}', "");
Expect(0, 4351, '\p{^Is_Blk=_-Georgian}', "");
Expect(0, 4351, '\P{Is_Blk=_-Georgian}', "");
Expect(1, 4351, '\P{^Is_Blk=_-Georgian}', "");
Expect(0, 4352, '\p{Is_Blk=_-Georgian}', "");
Expect(1, 4352, '\p{^Is_Blk=_-Georgian}', "");
Expect(1, 4352, '\P{Is_Blk=_-Georgian}', "");
Expect(0, 4352, '\P{^Is_Blk=_-Georgian}', "");
Error('\p{Block=:= -georgian_Supplement}');
Error('\P{Block=:= -georgian_Supplement}');
Expect(1, 11567, '\p{Block: georgiansupplement}', "");
Expect(0, 11567, '\p{^Block: georgiansupplement}', "");
Expect(0, 11567, '\P{Block: georgiansupplement}', "");
Expect(1, 11567, '\P{^Block: georgiansupplement}', "");
Expect(0, 11568, '\p{Block: georgiansupplement}', "");
Expect(1, 11568, '\p{^Block: georgiansupplement}', "");
Expect(1, 11568, '\P{Block: georgiansupplement}', "");
Expect(0, 11568, '\P{^Block: georgiansupplement}', "");
Expect(1, 11567, '\p{Block=	-Georgian_Supplement}', "");
Expect(0, 11567, '\p{^Block=	-Georgian_Supplement}', "");
Expect(0, 11567, '\P{Block=	-Georgian_Supplement}', "");
Expect(1, 11567, '\P{^Block=	-Georgian_Supplement}', "");
Expect(0, 11568, '\p{Block=	-Georgian_Supplement}', "");
Expect(1, 11568, '\p{^Block=	-Georgian_Supplement}', "");
Expect(1, 11568, '\P{Block=	-Georgian_Supplement}', "");
Expect(0, 11568, '\P{^Block=	-Georgian_Supplement}', "");
Error('\p{Blk=  GEORGIAN_Sup:=}');
Error('\P{Blk=  GEORGIAN_Sup:=}');
Expect(1, 11567, '\p{Blk=georgiansup}', "");
Expect(0, 11567, '\p{^Blk=georgiansup}', "");
Expect(0, 11567, '\P{Blk=georgiansup}', "");
Expect(1, 11567, '\P{^Blk=georgiansup}', "");
Expect(0, 11568, '\p{Blk=georgiansup}', "");
Expect(1, 11568, '\p{^Blk=georgiansup}', "");
Expect(1, 11568, '\P{Blk=georgiansup}', "");
Expect(0, 11568, '\P{^Blk=georgiansup}', "");
Expect(1, 11567, '\p{Blk=--Georgian_Sup}', "");
Expect(0, 11567, '\p{^Blk=--Georgian_Sup}', "");
Expect(0, 11567, '\P{Blk=--Georgian_Sup}', "");
Expect(1, 11567, '\P{^Blk=--Georgian_Sup}', "");
Expect(0, 11568, '\p{Blk=--Georgian_Sup}', "");
Expect(1, 11568, '\p{^Blk=--Georgian_Sup}', "");
Expect(1, 11568, '\P{Blk=--Georgian_Sup}', "");
Expect(0, 11568, '\P{^Blk=--Georgian_Sup}', "");
Error('\p{Is_Block= /a/Georgian_Supplement}');
Error('\P{Is_Block= /a/Georgian_Supplement}');
Expect(1, 11567, '\p{Is_Block=georgiansupplement}', "");
Expect(0, 11567, '\p{^Is_Block=georgiansupplement}', "");
Expect(0, 11567, '\P{Is_Block=georgiansupplement}', "");
Expect(1, 11567, '\P{^Is_Block=georgiansupplement}', "");
Expect(0, 11568, '\p{Is_Block=georgiansupplement}', "");
Expect(1, 11568, '\p{^Is_Block=georgiansupplement}', "");
Expect(1, 11568, '\P{Is_Block=georgiansupplement}', "");
Expect(0, 11568, '\P{^Is_Block=georgiansupplement}', "");
Expect(1, 11567, '\p{Is_Block=Georgian_Supplement}', "");
Expect(0, 11567, '\p{^Is_Block=Georgian_Supplement}', "");
Expect(0, 11567, '\P{Is_Block=Georgian_Supplement}', "");
Expect(1, 11567, '\P{^Is_Block=Georgian_Supplement}', "");
Expect(0, 11568, '\p{Is_Block=Georgian_Supplement}', "");
Expect(1, 11568, '\p{^Is_Block=Georgian_Supplement}', "");
Expect(1, 11568, '\P{Is_Block=Georgian_Supplement}', "");
Expect(0, 11568, '\P{^Is_Block=Georgian_Supplement}', "");
Error('\p{Is_Blk=:=_GEORGIAN_sup}');
Error('\P{Is_Blk=:=_GEORGIAN_sup}');
Expect(1, 11567, '\p{Is_Blk=georgiansup}', "");
Expect(0, 11567, '\p{^Is_Blk=georgiansup}', "");
Expect(0, 11567, '\P{Is_Blk=georgiansup}', "");
Expect(1, 11567, '\P{^Is_Blk=georgiansup}', "");
Expect(0, 11568, '\p{Is_Blk=georgiansup}', "");
Expect(1, 11568, '\p{^Is_Blk=georgiansup}', "");
Expect(1, 11568, '\P{Is_Blk=georgiansup}', "");
Expect(0, 11568, '\P{^Is_Blk=georgiansup}', "");
Expect(1, 11567, '\p{Is_Blk= 	Georgian_SUP}', "");
Expect(0, 11567, '\p{^Is_Blk= 	Georgian_SUP}', "");
Expect(0, 11567, '\P{Is_Blk= 	Georgian_SUP}', "");
Expect(1, 11567, '\P{^Is_Blk= 	Georgian_SUP}', "");
Expect(0, 11568, '\p{Is_Blk= 	Georgian_SUP}', "");
Expect(1, 11568, '\p{^Is_Blk= 	Georgian_SUP}', "");
Expect(1, 11568, '\P{Is_Blk= 	Georgian_SUP}', "");
Expect(0, 11568, '\P{^Is_Blk= 	Georgian_SUP}', "");
Error('\p{Block=glagolitic/a/}');
Error('\P{Block=glagolitic/a/}');
Expect(1, 11359, '\p{Block=glagolitic}', "");
Expect(0, 11359, '\p{^Block=glagolitic}', "");
Expect(0, 11359, '\P{Block=glagolitic}', "");
Expect(1, 11359, '\P{^Block=glagolitic}', "");
Expect(0, 11360, '\p{Block=glagolitic}', "");
Expect(1, 11360, '\p{^Block=glagolitic}', "");
Expect(1, 11360, '\P{Block=glagolitic}', "");
Expect(0, 11360, '\P{^Block=glagolitic}', "");
Expect(1, 11359, '\p{Block= 	GLAGOLITIC}', "");
Expect(0, 11359, '\p{^Block= 	GLAGOLITIC}', "");
Expect(0, 11359, '\P{Block= 	GLAGOLITIC}', "");
Expect(1, 11359, '\P{^Block= 	GLAGOLITIC}', "");
Expect(0, 11360, '\p{Block= 	GLAGOLITIC}', "");
Expect(1, 11360, '\p{^Block= 	GLAGOLITIC}', "");
Expect(1, 11360, '\P{Block= 	GLAGOLITIC}', "");
Expect(0, 11360, '\P{^Block= 	GLAGOLITIC}', "");
Error('\p{Blk=/a/glagolitic}');
Error('\P{Blk=/a/glagolitic}');
Expect(1, 11359, '\p{Blk:glagolitic}', "");
Expect(0, 11359, '\p{^Blk:glagolitic}', "");
Expect(0, 11359, '\P{Blk:glagolitic}', "");
Expect(1, 11359, '\P{^Blk:glagolitic}', "");
Expect(0, 11360, '\p{Blk:glagolitic}', "");
Expect(1, 11360, '\p{^Blk:glagolitic}', "");
Expect(1, 11360, '\P{Blk:glagolitic}', "");
Expect(0, 11360, '\P{^Blk:glagolitic}', "");
Expect(1, 11359, '\p{Blk=-GLAGOLITIC}', "");
Expect(0, 11359, '\p{^Blk=-GLAGOLITIC}', "");
Expect(0, 11359, '\P{Blk=-GLAGOLITIC}', "");
Expect(1, 11359, '\P{^Blk=-GLAGOLITIC}', "");
Expect(0, 11360, '\p{Blk=-GLAGOLITIC}', "");
Expect(1, 11360, '\p{^Blk=-GLAGOLITIC}', "");
Expect(1, 11360, '\P{Blk=-GLAGOLITIC}', "");
Expect(0, 11360, '\P{^Blk=-GLAGOLITIC}', "");
Error('\p{Is_Block:			Glagolitic:=}');
Error('\P{Is_Block:			Glagolitic:=}');
Expect(1, 11359, '\p{Is_Block=glagolitic}', "");
Expect(0, 11359, '\p{^Is_Block=glagolitic}', "");
Expect(0, 11359, '\P{Is_Block=glagolitic}', "");
Expect(1, 11359, '\P{^Is_Block=glagolitic}', "");
Expect(0, 11360, '\p{Is_Block=glagolitic}', "");
Expect(1, 11360, '\p{^Is_Block=glagolitic}', "");
Expect(1, 11360, '\P{Is_Block=glagolitic}', "");
Expect(0, 11360, '\P{^Is_Block=glagolitic}', "");
Expect(1, 11359, '\p{Is_Block=-Glagolitic}', "");
Expect(0, 11359, '\p{^Is_Block=-Glagolitic}', "");
Expect(0, 11359, '\P{Is_Block=-Glagolitic}', "");
Expect(1, 11359, '\P{^Is_Block=-Glagolitic}', "");
Expect(0, 11360, '\p{Is_Block=-Glagolitic}', "");
Expect(1, 11360, '\p{^Is_Block=-Glagolitic}', "");
Expect(1, 11360, '\P{Is_Block=-Glagolitic}', "");
Expect(0, 11360, '\P{^Is_Block=-Glagolitic}', "");
Error('\p{Is_Blk=_:=GLAGOLITIC}');
Error('\P{Is_Blk=_:=GLAGOLITIC}');
Expect(1, 11359, '\p{Is_Blk=glagolitic}', "");
Expect(0, 11359, '\p{^Is_Blk=glagolitic}', "");
Expect(0, 11359, '\P{Is_Blk=glagolitic}', "");
Expect(1, 11359, '\P{^Is_Blk=glagolitic}', "");
Expect(0, 11360, '\p{Is_Blk=glagolitic}', "");
Expect(1, 11360, '\p{^Is_Blk=glagolitic}', "");
Expect(1, 11360, '\P{Is_Blk=glagolitic}', "");
Expect(0, 11360, '\P{^Is_Blk=glagolitic}', "");
Expect(1, 11359, '\p{Is_Blk=__GLAGOLITIC}', "");
Expect(0, 11359, '\p{^Is_Blk=__GLAGOLITIC}', "");
Expect(0, 11359, '\P{Is_Blk=__GLAGOLITIC}', "");
Expect(1, 11359, '\P{^Is_Blk=__GLAGOLITIC}', "");
Expect(0, 11360, '\p{Is_Blk=__GLAGOLITIC}', "");
Expect(1, 11360, '\p{^Is_Blk=__GLAGOLITIC}', "");
Expect(1, 11360, '\P{Is_Blk=__GLAGOLITIC}', "");
Expect(0, 11360, '\P{^Is_Blk=__GLAGOLITIC}', "");
Error('\p{Block=-_Gothic/a/}');
Error('\P{Block=-_Gothic/a/}');
Expect(1, 66383, '\p{Block=gothic}', "");
Expect(0, 66383, '\p{^Block=gothic}', "");
Expect(0, 66383, '\P{Block=gothic}', "");
Expect(1, 66383, '\P{^Block=gothic}', "");
Expect(0, 66384, '\p{Block=gothic}', "");
Expect(1, 66384, '\p{^Block=gothic}', "");
Expect(1, 66384, '\P{Block=gothic}', "");
Expect(0, 66384, '\P{^Block=gothic}', "");
Expect(1, 66383, '\p{Block= GOTHIC}', "");
Expect(0, 66383, '\p{^Block= GOTHIC}', "");
Expect(0, 66383, '\P{Block= GOTHIC}', "");
Expect(1, 66383, '\P{^Block= GOTHIC}', "");
Expect(0, 66384, '\p{Block= GOTHIC}', "");
Expect(1, 66384, '\p{^Block= GOTHIC}', "");
Expect(1, 66384, '\P{Block= GOTHIC}', "");
Expect(0, 66384, '\P{^Block= GOTHIC}', "");
Error('\p{Blk:  GOTHIC/a/}');
Error('\P{Blk:  GOTHIC/a/}');
Expect(1, 66383, '\p{Blk=gothic}', "");
Expect(0, 66383, '\p{^Blk=gothic}', "");
Expect(0, 66383, '\P{Blk=gothic}', "");
Expect(1, 66383, '\P{^Blk=gothic}', "");
Expect(0, 66384, '\p{Blk=gothic}', "");
Expect(1, 66384, '\p{^Blk=gothic}', "");
Expect(1, 66384, '\P{Blk=gothic}', "");
Expect(0, 66384, '\P{^Blk=gothic}', "");
Expect(1, 66383, '\p{Blk=  Gothic}', "");
Expect(0, 66383, '\p{^Blk=  Gothic}', "");
Expect(0, 66383, '\P{Blk=  Gothic}', "");
Expect(1, 66383, '\P{^Blk=  Gothic}', "");
Expect(0, 66384, '\p{Blk=  Gothic}', "");
Expect(1, 66384, '\p{^Blk=  Gothic}', "");
Expect(1, 66384, '\P{Blk=  Gothic}', "");
Expect(0, 66384, '\P{^Blk=  Gothic}', "");
Error('\p{Is_Block=__gothic:=}');
Error('\P{Is_Block=__gothic:=}');
Expect(1, 66383, '\p{Is_Block=gothic}', "");
Expect(0, 66383, '\p{^Is_Block=gothic}', "");
Expect(0, 66383, '\P{Is_Block=gothic}', "");
Expect(1, 66383, '\P{^Is_Block=gothic}', "");
Expect(0, 66384, '\p{Is_Block=gothic}', "");
Expect(1, 66384, '\p{^Is_Block=gothic}', "");
Expect(1, 66384, '\P{Is_Block=gothic}', "");
Expect(0, 66384, '\P{^Is_Block=gothic}', "");
Expect(1, 66383, '\p{Is_Block=-_gothic}', "");
Expect(0, 66383, '\p{^Is_Block=-_gothic}', "");
Expect(0, 66383, '\P{Is_Block=-_gothic}', "");
Expect(1, 66383, '\P{^Is_Block=-_gothic}', "");
Expect(0, 66384, '\p{Is_Block=-_gothic}', "");
Expect(1, 66384, '\p{^Is_Block=-_gothic}', "");
Expect(1, 66384, '\P{Is_Block=-_gothic}', "");
Expect(0, 66384, '\P{^Is_Block=-_gothic}', "");
Error('\p{Is_Blk=_:=Gothic}');
Error('\P{Is_Blk=_:=Gothic}');
Expect(1, 66383, '\p{Is_Blk=gothic}', "");
Expect(0, 66383, '\p{^Is_Blk=gothic}', "");
Expect(0, 66383, '\P{Is_Blk=gothic}', "");
Expect(1, 66383, '\P{^Is_Blk=gothic}', "");
Expect(0, 66384, '\p{Is_Blk=gothic}', "");
Expect(1, 66384, '\p{^Is_Blk=gothic}', "");
Expect(1, 66384, '\P{Is_Blk=gothic}', "");
Expect(0, 66384, '\P{^Is_Blk=gothic}', "");
Expect(1, 66383, '\p{Is_Blk=	Gothic}', "");
Expect(0, 66383, '\p{^Is_Blk=	Gothic}', "");
Expect(0, 66383, '\P{Is_Blk=	Gothic}', "");
Expect(1, 66383, '\P{^Is_Blk=	Gothic}', "");
Expect(0, 66384, '\p{Is_Blk=	Gothic}', "");
Expect(1, 66384, '\p{^Is_Blk=	Gothic}', "");
Expect(1, 66384, '\P{Is_Blk=	Gothic}', "");
Expect(0, 66384, '\P{^Is_Blk=	Gothic}', "");
Error('\p{Block= _greek_AND_coptic:=}');
Error('\P{Block= _greek_AND_coptic:=}');
Expect(1, 1023, '\p{Block=greekandcoptic}', "");
Expect(0, 1023, '\p{^Block=greekandcoptic}', "");
Expect(0, 1023, '\P{Block=greekandcoptic}', "");
Expect(1, 1023, '\P{^Block=greekandcoptic}', "");
Expect(0, 1024, '\p{Block=greekandcoptic}', "");
Expect(1, 1024, '\p{^Block=greekandcoptic}', "");
Expect(1, 1024, '\P{Block=greekandcoptic}', "");
Expect(0, 1024, '\P{^Block=greekandcoptic}', "");
Expect(1, 1023, '\p{Block=		GREEK_And_COPTIC}', "");
Expect(0, 1023, '\p{^Block=		GREEK_And_COPTIC}', "");
Expect(0, 1023, '\P{Block=		GREEK_And_COPTIC}', "");
Expect(1, 1023, '\P{^Block=		GREEK_And_COPTIC}', "");
Expect(0, 1024, '\p{Block=		GREEK_And_COPTIC}', "");
Expect(1, 1024, '\p{^Block=		GREEK_And_COPTIC}', "");
Expect(1, 1024, '\P{Block=		GREEK_And_COPTIC}', "");
Expect(0, 1024, '\P{^Block=		GREEK_And_COPTIC}', "");
Error('\p{Blk=	:=Greek}');
Error('\P{Blk=	:=Greek}');
Expect(1, 1023, '\p{Blk=greek}', "");
Expect(0, 1023, '\p{^Blk=greek}', "");
Expect(0, 1023, '\P{Blk=greek}', "");
Expect(1, 1023, '\P{^Blk=greek}', "");
Expect(0, 1024, '\p{Blk=greek}', "");
Expect(1, 1024, '\p{^Blk=greek}', "");
Expect(1, 1024, '\P{Blk=greek}', "");
Expect(0, 1024, '\P{^Blk=greek}', "");
Expect(1, 1023, '\p{Blk=--greek}', "");
Expect(0, 1023, '\p{^Blk=--greek}', "");
Expect(0, 1023, '\P{Blk=--greek}', "");
Expect(1, 1023, '\P{^Blk=--greek}', "");
Expect(0, 1024, '\p{Blk=--greek}', "");
Expect(1, 1024, '\p{^Blk=--greek}', "");
Expect(1, 1024, '\P{Blk=--greek}', "");
Expect(0, 1024, '\P{^Blk=--greek}', "");
Error('\p{Is_Block:		:=Greek_and_coptic}');
Error('\P{Is_Block:		:=Greek_and_coptic}');
Expect(1, 1023, '\p{Is_Block=greekandcoptic}', "");
Expect(0, 1023, '\p{^Is_Block=greekandcoptic}', "");
Expect(0, 1023, '\P{Is_Block=greekandcoptic}', "");
Expect(1, 1023, '\P{^Is_Block=greekandcoptic}', "");
Expect(0, 1024, '\p{Is_Block=greekandcoptic}', "");
Expect(1, 1024, '\p{^Is_Block=greekandcoptic}', "");
Expect(1, 1024, '\P{Is_Block=greekandcoptic}', "");
Expect(0, 1024, '\P{^Is_Block=greekandcoptic}', "");
Expect(1, 1023, '\p{Is_Block=__Greek_And_Coptic}', "");
Expect(0, 1023, '\p{^Is_Block=__Greek_And_Coptic}', "");
Expect(0, 1023, '\P{Is_Block=__Greek_And_Coptic}', "");
Expect(1, 1023, '\P{^Is_Block=__Greek_And_Coptic}', "");
Expect(0, 1024, '\p{Is_Block=__Greek_And_Coptic}', "");
Expect(1, 1024, '\p{^Is_Block=__Greek_And_Coptic}', "");
Expect(1, 1024, '\P{Is_Block=__Greek_And_Coptic}', "");
Expect(0, 1024, '\P{^Is_Block=__Greek_And_Coptic}', "");
Error('\p{Is_Blk= -Greek:=}');
Error('\P{Is_Blk= -Greek:=}');
Expect(1, 1023, '\p{Is_Blk:   greek}', "");
Expect(0, 1023, '\p{^Is_Blk:   greek}', "");
Expect(0, 1023, '\P{Is_Blk:   greek}', "");
Expect(1, 1023, '\P{^Is_Blk:   greek}', "");
Expect(0, 1024, '\p{Is_Blk:   greek}', "");
Expect(1, 1024, '\p{^Is_Blk:   greek}', "");
Expect(1, 1024, '\P{Is_Blk:   greek}', "");
Expect(0, 1024, '\P{^Is_Blk:   greek}', "");
Expect(1, 1023, '\p{Is_Blk=-greek}', "");
Expect(0, 1023, '\p{^Is_Blk=-greek}', "");
Expect(0, 1023, '\P{Is_Blk=-greek}', "");
Expect(1, 1023, '\P{^Is_Blk=-greek}', "");
Expect(0, 1024, '\p{Is_Blk=-greek}', "");
Expect(1, 1024, '\p{^Is_Blk=-greek}', "");
Expect(1, 1024, '\P{Is_Blk=-greek}', "");
Expect(0, 1024, '\P{^Is_Blk=-greek}', "");
Error('\p{Block=-:=Greek_EXTENDED}');
Error('\P{Block=-:=Greek_EXTENDED}');
Expect(1, 8191, '\p{Block=greekextended}', "");
Expect(0, 8191, '\p{^Block=greekextended}', "");
Expect(0, 8191, '\P{Block=greekextended}', "");
Expect(1, 8191, '\P{^Block=greekextended}', "");
Expect(0, 8192, '\p{Block=greekextended}', "");
Expect(1, 8192, '\p{^Block=greekextended}', "");
Expect(1, 8192, '\P{Block=greekextended}', "");
Expect(0, 8192, '\P{^Block=greekextended}', "");
Expect(1, 8191, '\p{Block=  Greek_EXTENDED}', "");
Expect(0, 8191, '\p{^Block=  Greek_EXTENDED}', "");
Expect(0, 8191, '\P{Block=  Greek_EXTENDED}', "");
Expect(1, 8191, '\P{^Block=  Greek_EXTENDED}', "");
Expect(0, 8192, '\p{Block=  Greek_EXTENDED}', "");
Expect(1, 8192, '\p{^Block=  Greek_EXTENDED}', "");
Expect(1, 8192, '\P{Block=  Greek_EXTENDED}', "");
Expect(0, 8192, '\P{^Block=  Greek_EXTENDED}', "");
Error('\p{Blk=:=--Greek_EXT}');
Error('\P{Blk=:=--Greek_EXT}');
Expect(1, 8191, '\p{Blk=greekext}', "");
Expect(0, 8191, '\p{^Blk=greekext}', "");
Expect(0, 8191, '\P{Blk=greekext}', "");
Expect(1, 8191, '\P{^Blk=greekext}', "");
Expect(0, 8192, '\p{Blk=greekext}', "");
Expect(1, 8192, '\p{^Blk=greekext}', "");
Expect(1, 8192, '\P{Blk=greekext}', "");
Expect(0, 8192, '\P{^Blk=greekext}', "");
Expect(1, 8191, '\p{Blk=_ GREEK_EXT}', "");
Expect(0, 8191, '\p{^Blk=_ GREEK_EXT}', "");
Expect(0, 8191, '\P{Blk=_ GREEK_EXT}', "");
Expect(1, 8191, '\P{^Blk=_ GREEK_EXT}', "");
Expect(0, 8192, '\p{Blk=_ GREEK_EXT}', "");
Expect(1, 8192, '\p{^Blk=_ GREEK_EXT}', "");
Expect(1, 8192, '\P{Blk=_ GREEK_EXT}', "");
Expect(0, 8192, '\P{^Blk=_ GREEK_EXT}', "");
Error('\p{Is_Block:	:= greek_Extended}');
Error('\P{Is_Block:	:= greek_Extended}');
Expect(1, 8191, '\p{Is_Block=greekextended}', "");
Expect(0, 8191, '\p{^Is_Block=greekextended}', "");
Expect(0, 8191, '\P{Is_Block=greekextended}', "");
Expect(1, 8191, '\P{^Is_Block=greekextended}', "");
Expect(0, 8192, '\p{Is_Block=greekextended}', "");
Expect(1, 8192, '\p{^Is_Block=greekextended}', "");
Expect(1, 8192, '\P{Is_Block=greekextended}', "");
Expect(0, 8192, '\P{^Is_Block=greekextended}', "");
Expect(1, 8191, '\p{Is_Block:		greek_Extended}', "");
Expect(0, 8191, '\p{^Is_Block:		greek_Extended}', "");
Expect(0, 8191, '\P{Is_Block:		greek_Extended}', "");
Expect(1, 8191, '\P{^Is_Block:		greek_Extended}', "");
Expect(0, 8192, '\p{Is_Block:		greek_Extended}', "");
Expect(1, 8192, '\p{^Is_Block:		greek_Extended}', "");
Expect(1, 8192, '\P{Is_Block:		greek_Extended}', "");
Expect(0, 8192, '\P{^Is_Block:		greek_Extended}', "");
Error('\p{Is_Blk=	/a/Greek_Ext}');
Error('\P{Is_Blk=	/a/Greek_Ext}');
Expect(1, 8191, '\p{Is_Blk=greekext}', "");
Expect(0, 8191, '\p{^Is_Blk=greekext}', "");
Expect(0, 8191, '\P{Is_Blk=greekext}', "");
Expect(1, 8191, '\P{^Is_Blk=greekext}', "");
Expect(0, 8192, '\p{Is_Blk=greekext}', "");
Expect(1, 8192, '\p{^Is_Blk=greekext}', "");
Expect(1, 8192, '\P{Is_Blk=greekext}', "");
Expect(0, 8192, '\P{^Is_Blk=greekext}', "");
Expect(1, 8191, '\p{Is_Blk=	greek_EXT}', "");
Expect(0, 8191, '\p{^Is_Blk=	greek_EXT}', "");
Expect(0, 8191, '\P{Is_Blk=	greek_EXT}', "");
Expect(1, 8191, '\P{^Is_Blk=	greek_EXT}', "");
Expect(0, 8192, '\p{Is_Blk=	greek_EXT}', "");
Expect(1, 8192, '\p{^Is_Blk=	greek_EXT}', "");
Expect(1, 8192, '\P{Is_Blk=	greek_EXT}', "");
Expect(0, 8192, '\P{^Is_Blk=	greek_EXT}', "");
Error('\p{Block=- Gujarati/a/}');
Error('\P{Block=- Gujarati/a/}');
Expect(1, 2815, '\p{Block=gujarati}', "");
Expect(0, 2815, '\p{^Block=gujarati}', "");
Expect(0, 2815, '\P{Block=gujarati}', "");
Expect(1, 2815, '\P{^Block=gujarati}', "");
Expect(0, 2816, '\p{Block=gujarati}', "");
Expect(1, 2816, '\p{^Block=gujarati}', "");
Expect(1, 2816, '\P{Block=gujarati}', "");
Expect(0, 2816, '\P{^Block=gujarati}', "");
Expect(1, 2815, '\p{Block:	_-GUJARATI}', "");
Expect(0, 2815, '\p{^Block:	_-GUJARATI}', "");
Expect(0, 2815, '\P{Block:	_-GUJARATI}', "");
Expect(1, 2815, '\P{^Block:	_-GUJARATI}', "");
Expect(0, 2816, '\p{Block:	_-GUJARATI}', "");
Expect(1, 2816, '\p{^Block:	_-GUJARATI}', "");
Expect(1, 2816, '\P{Block:	_-GUJARATI}', "");
Expect(0, 2816, '\P{^Block:	_-GUJARATI}', "");
Error('\p{Blk=/a/Gujarati}');
Error('\P{Blk=/a/Gujarati}');
Expect(1, 2815, '\p{Blk=gujarati}', "");
Expect(0, 2815, '\p{^Blk=gujarati}', "");
Expect(0, 2815, '\P{Blk=gujarati}', "");
Expect(1, 2815, '\P{^Blk=gujarati}', "");
Expect(0, 2816, '\p{Blk=gujarati}', "");
Expect(1, 2816, '\p{^Blk=gujarati}', "");
Expect(1, 2816, '\P{Blk=gujarati}', "");
Expect(0, 2816, '\P{^Blk=gujarati}', "");
Expect(1, 2815, '\p{Blk=	-gujarati}', "");
Expect(0, 2815, '\p{^Blk=	-gujarati}', "");
Expect(0, 2815, '\P{Blk=	-gujarati}', "");
Expect(1, 2815, '\P{^Blk=	-gujarati}', "");
Expect(0, 2816, '\p{Blk=	-gujarati}', "");
Expect(1, 2816, '\p{^Blk=	-gujarati}', "");
Expect(1, 2816, '\P{Blk=	-gujarati}', "");
Expect(0, 2816, '\P{^Blk=	-gujarati}', "");
Error('\p{Is_Block=_GUJARATI:=}');
Error('\P{Is_Block=_GUJARATI:=}');
Expect(1, 2815, '\p{Is_Block=gujarati}', "");
Expect(0, 2815, '\p{^Is_Block=gujarati}', "");
Expect(0, 2815, '\P{Is_Block=gujarati}', "");
Expect(1, 2815, '\P{^Is_Block=gujarati}', "");
Expect(0, 2816, '\p{Is_Block=gujarati}', "");
Expect(1, 2816, '\p{^Is_Block=gujarati}', "");
Expect(1, 2816, '\P{Is_Block=gujarati}', "");
Expect(0, 2816, '\P{^Is_Block=gujarati}', "");
Expect(1, 2815, '\p{Is_Block=_gujarati}', "");
Expect(0, 2815, '\p{^Is_Block=_gujarati}', "");
Expect(0, 2815, '\P{Is_Block=_gujarati}', "");
Expect(1, 2815, '\P{^Is_Block=_gujarati}', "");
Expect(0, 2816, '\p{Is_Block=_gujarati}', "");
Expect(1, 2816, '\p{^Is_Block=_gujarati}', "");
Expect(1, 2816, '\P{Is_Block=_gujarati}', "");
Expect(0, 2816, '\P{^Is_Block=_gujarati}', "");
Error('\p{Is_Blk=/a/_ Gujarati}');
Error('\P{Is_Blk=/a/_ Gujarati}');
Expect(1, 2815, '\p{Is_Blk=gujarati}', "");
Expect(0, 2815, '\p{^Is_Blk=gujarati}', "");
Expect(0, 2815, '\P{Is_Blk=gujarati}', "");
Expect(1, 2815, '\P{^Is_Blk=gujarati}', "");
Expect(0, 2816, '\p{Is_Blk=gujarati}', "");
Expect(1, 2816, '\p{^Is_Blk=gujarati}', "");
Expect(1, 2816, '\P{Is_Blk=gujarati}', "");
Expect(0, 2816, '\P{^Is_Blk=gujarati}', "");
Expect(1, 2815, '\p{Is_Blk=	-gujarati}', "");
Expect(0, 2815, '\p{^Is_Blk=	-gujarati}', "");
Expect(0, 2815, '\P{Is_Blk=	-gujarati}', "");
Expect(1, 2815, '\P{^Is_Blk=	-gujarati}', "");
Expect(0, 2816, '\p{Is_Blk=	-gujarati}', "");
Expect(1, 2816, '\p{^Is_Blk=	-gujarati}', "");
Expect(1, 2816, '\P{Is_Blk=	-gujarati}', "");
Expect(0, 2816, '\P{^Is_Blk=	-gujarati}', "");
Error('\p{Block=:=GURMUKHI}');
Error('\P{Block=:=GURMUKHI}');
Expect(1, 2687, '\p{Block=gurmukhi}', "");
Expect(0, 2687, '\p{^Block=gurmukhi}', "");
Expect(0, 2687, '\P{Block=gurmukhi}', "");
Expect(1, 2687, '\P{^Block=gurmukhi}', "");
Expect(0, 2688, '\p{Block=gurmukhi}', "");
Expect(1, 2688, '\p{^Block=gurmukhi}', "");
Expect(1, 2688, '\P{Block=gurmukhi}', "");
Expect(0, 2688, '\P{^Block=gurmukhi}', "");
Expect(1, 2687, '\p{Block=- Gurmukhi}', "");
Expect(0, 2687, '\p{^Block=- Gurmukhi}', "");
Expect(0, 2687, '\P{Block=- Gurmukhi}', "");
Expect(1, 2687, '\P{^Block=- Gurmukhi}', "");
Expect(0, 2688, '\p{Block=- Gurmukhi}', "");
Expect(1, 2688, '\p{^Block=- Gurmukhi}', "");
Expect(1, 2688, '\P{Block=- Gurmukhi}', "");
Expect(0, 2688, '\P{^Block=- Gurmukhi}', "");
Error('\p{Blk=:=-_Gurmukhi}');
Error('\P{Blk=:=-_Gurmukhi}');
Expect(1, 2687, '\p{Blk=gurmukhi}', "");
Expect(0, 2687, '\p{^Blk=gurmukhi}', "");
Expect(0, 2687, '\P{Blk=gurmukhi}', "");
Expect(1, 2687, '\P{^Blk=gurmukhi}', "");
Expect(0, 2688, '\p{Blk=gurmukhi}', "");
Expect(1, 2688, '\p{^Blk=gurmukhi}', "");
Expect(1, 2688, '\P{Blk=gurmukhi}', "");
Expect(0, 2688, '\P{^Blk=gurmukhi}', "");
Expect(1, 2687, '\p{Blk= Gurmukhi}', "");
Expect(0, 2687, '\p{^Blk= Gurmukhi}', "");
Expect(0, 2687, '\P{Blk= Gurmukhi}', "");
Expect(1, 2687, '\P{^Blk= Gurmukhi}', "");
Expect(0, 2688, '\p{Blk= Gurmukhi}', "");
Expect(1, 2688, '\p{^Blk= Gurmukhi}', "");
Expect(1, 2688, '\P{Blk= Gurmukhi}', "");
Expect(0, 2688, '\P{^Blk= Gurmukhi}', "");
Error('\p{Is_Block=_-gurmukhi:=}');
Error('\P{Is_Block=_-gurmukhi:=}');
Expect(1, 2687, '\p{Is_Block=gurmukhi}', "");
Expect(0, 2687, '\p{^Is_Block=gurmukhi}', "");
Expect(0, 2687, '\P{Is_Block=gurmukhi}', "");
Expect(1, 2687, '\P{^Is_Block=gurmukhi}', "");
Expect(0, 2688, '\p{Is_Block=gurmukhi}', "");
Expect(1, 2688, '\p{^Is_Block=gurmukhi}', "");
Expect(1, 2688, '\P{Is_Block=gurmukhi}', "");
Expect(0, 2688, '\P{^Is_Block=gurmukhi}', "");
Expect(1, 2687, '\p{Is_Block= gurmukhi}', "");
Expect(0, 2687, '\p{^Is_Block= gurmukhi}', "");
Expect(0, 2687, '\P{Is_Block= gurmukhi}', "");
Expect(1, 2687, '\P{^Is_Block= gurmukhi}', "");
Expect(0, 2688, '\p{Is_Block= gurmukhi}', "");
Expect(1, 2688, '\p{^Is_Block= gurmukhi}', "");
Expect(1, 2688, '\P{Is_Block= gurmukhi}', "");
Expect(0, 2688, '\P{^Is_Block= gurmukhi}', "");
Error('\p{Is_Blk:   :=--gurmukhi}');
Error('\P{Is_Blk:   :=--gurmukhi}');
Expect(1, 2687, '\p{Is_Blk=gurmukhi}', "");
Expect(0, 2687, '\p{^Is_Blk=gurmukhi}', "");
Expect(0, 2687, '\P{Is_Blk=gurmukhi}', "");
Expect(1, 2687, '\P{^Is_Blk=gurmukhi}', "");
Expect(0, 2688, '\p{Is_Blk=gurmukhi}', "");
Expect(1, 2688, '\p{^Is_Blk=gurmukhi}', "");
Expect(1, 2688, '\P{Is_Blk=gurmukhi}', "");
Expect(0, 2688, '\P{^Is_Blk=gurmukhi}', "");
Expect(1, 2687, '\p{Is_Blk=	_Gurmukhi}', "");
Expect(0, 2687, '\p{^Is_Blk=	_Gurmukhi}', "");
Expect(0, 2687, '\P{Is_Blk=	_Gurmukhi}', "");
Expect(1, 2687, '\P{^Is_Blk=	_Gurmukhi}', "");
Expect(0, 2688, '\p{Is_Blk=	_Gurmukhi}', "");
Expect(1, 2688, '\p{^Is_Blk=	_Gurmukhi}', "");
Expect(1, 2688, '\P{Is_Blk=	_Gurmukhi}', "");
Expect(0, 2688, '\P{^Is_Blk=	_Gurmukhi}', "");
Error('\p{Block:	/a/-_Halfwidth_And_fullwidth_forms}');
Error('\P{Block:	/a/-_Halfwidth_And_fullwidth_forms}');
Expect(1, 65519, '\p{Block=halfwidthandfullwidthforms}', "");
Expect(0, 65519, '\p{^Block=halfwidthandfullwidthforms}', "");
Expect(0, 65519, '\P{Block=halfwidthandfullwidthforms}', "");
Expect(1, 65519, '\P{^Block=halfwidthandfullwidthforms}', "");
Expect(0, 65520, '\p{Block=halfwidthandfullwidthforms}', "");
Expect(1, 65520, '\p{^Block=halfwidthandfullwidthforms}', "");
Expect(1, 65520, '\P{Block=halfwidthandfullwidthforms}', "");
Expect(0, 65520, '\P{^Block=halfwidthandfullwidthforms}', "");
Expect(1, 65519, '\p{Block=-	halfwidth_And_Fullwidth_forms}', "");
Expect(0, 65519, '\p{^Block=-	halfwidth_And_Fullwidth_forms}', "");
Expect(0, 65519, '\P{Block=-	halfwidth_And_Fullwidth_forms}', "");
Expect(1, 65519, '\P{^Block=-	halfwidth_And_Fullwidth_forms}', "");
Expect(0, 65520, '\p{Block=-	halfwidth_And_Fullwidth_forms}', "");
Expect(1, 65520, '\p{^Block=-	halfwidth_And_Fullwidth_forms}', "");
Expect(1, 65520, '\P{Block=-	halfwidth_And_Fullwidth_forms}', "");
Expect(0, 65520, '\P{^Block=-	halfwidth_And_Fullwidth_forms}', "");
Error('\p{Blk=:=		Half_and_Full_Forms}');
Error('\P{Blk=:=		Half_and_Full_Forms}');
Expect(1, 65519, '\p{Blk=halfandfullforms}', "");
Expect(0, 65519, '\p{^Blk=halfandfullforms}', "");
Expect(0, 65519, '\P{Blk=halfandfullforms}', "");
Expect(1, 65519, '\P{^Blk=halfandfullforms}', "");
Expect(0, 65520, '\p{Blk=halfandfullforms}', "");
Expect(1, 65520, '\p{^Blk=halfandfullforms}', "");
Expect(1, 65520, '\P{Blk=halfandfullforms}', "");
Expect(0, 65520, '\P{^Blk=halfandfullforms}', "");
Expect(1, 65519, '\p{Blk=	 HALF_and_Full_forms}', "");
Expect(0, 65519, '\p{^Blk=	 HALF_and_Full_forms}', "");
Expect(0, 65519, '\P{Blk=	 HALF_and_Full_forms}', "");
Expect(1, 65519, '\P{^Blk=	 HALF_and_Full_forms}', "");
Expect(0, 65520, '\p{Blk=	 HALF_and_Full_forms}', "");
Expect(1, 65520, '\p{^Blk=	 HALF_and_Full_forms}', "");
Expect(1, 65520, '\P{Blk=	 HALF_and_Full_forms}', "");
Expect(0, 65520, '\P{^Blk=	 HALF_and_Full_forms}', "");
Error('\p{Is_Block=:=	-Halfwidth_And_Fullwidth_Forms}');
Error('\P{Is_Block=:=	-Halfwidth_And_Fullwidth_Forms}');
Expect(1, 65519, '\p{Is_Block=halfwidthandfullwidthforms}', "");
Expect(0, 65519, '\p{^Is_Block=halfwidthandfullwidthforms}', "");
Expect(0, 65519, '\P{Is_Block=halfwidthandfullwidthforms}', "");
Expect(1, 65519, '\P{^Is_Block=halfwidthandfullwidthforms}', "");
Expect(0, 65520, '\p{Is_Block=halfwidthandfullwidthforms}', "");
Expect(1, 65520, '\p{^Is_Block=halfwidthandfullwidthforms}', "");
Expect(1, 65520, '\P{Is_Block=halfwidthandfullwidthforms}', "");
Expect(0, 65520, '\P{^Is_Block=halfwidthandfullwidthforms}', "");
Expect(1, 65519, '\p{Is_Block:  Halfwidth_and_fullwidth_Forms}', "");
Expect(0, 65519, '\p{^Is_Block:  Halfwidth_and_fullwidth_Forms}', "");
Expect(0, 65519, '\P{Is_Block:  Halfwidth_and_fullwidth_Forms}', "");
Expect(1, 65519, '\P{^Is_Block:  Halfwidth_and_fullwidth_Forms}', "");
Expect(0, 65520, '\p{Is_Block:  Halfwidth_and_fullwidth_Forms}', "");
Expect(1, 65520, '\p{^Is_Block:  Halfwidth_and_fullwidth_Forms}', "");
Expect(1, 65520, '\P{Is_Block:  Halfwidth_and_fullwidth_Forms}', "");
Expect(0, 65520, '\P{^Is_Block:  Halfwidth_and_fullwidth_Forms}', "");
Error('\p{Is_Blk=:=-Half_and_full_FORMS}');
Error('\P{Is_Blk=:=-Half_and_full_FORMS}');
Expect(1, 65519, '\p{Is_Blk=halfandfullforms}', "");
Expect(0, 65519, '\p{^Is_Blk=halfandfullforms}', "");
Expect(0, 65519, '\P{Is_Blk=halfandfullforms}', "");
Expect(1, 65519, '\P{^Is_Blk=halfandfullforms}', "");
Expect(0, 65520, '\p{Is_Blk=halfandfullforms}', "");
Expect(1, 65520, '\p{^Is_Blk=halfandfullforms}', "");
Expect(1, 65520, '\P{Is_Blk=halfandfullforms}', "");
Expect(0, 65520, '\P{^Is_Blk=halfandfullforms}', "");
Expect(1, 65519, '\p{Is_Blk=	 HALF_And_full_Forms}', "");
Expect(0, 65519, '\p{^Is_Blk=	 HALF_And_full_Forms}', "");
Expect(0, 65519, '\P{Is_Blk=	 HALF_And_full_Forms}', "");
Expect(1, 65519, '\P{^Is_Blk=	 HALF_And_full_Forms}', "");
Expect(0, 65520, '\p{Is_Blk=	 HALF_And_full_Forms}', "");
Expect(1, 65520, '\p{^Is_Blk=	 HALF_And_full_Forms}', "");
Expect(1, 65520, '\P{Is_Blk=	 HALF_And_full_Forms}', "");
Expect(0, 65520, '\P{^Is_Blk=	 HALF_And_full_Forms}', "");
Error('\p{Block: 		COMBINING_Half_Marks:=}');
Error('\P{Block: 		COMBINING_Half_Marks:=}');
Expect(1, 65071, '\p{Block:   combininghalfmarks}', "");
Expect(0, 65071, '\p{^Block:   combininghalfmarks}', "");
Expect(0, 65071, '\P{Block:   combininghalfmarks}', "");
Expect(1, 65071, '\P{^Block:   combininghalfmarks}', "");
Expect(0, 65072, '\p{Block:   combininghalfmarks}', "");
Expect(1, 65072, '\p{^Block:   combininghalfmarks}', "");
Expect(1, 65072, '\P{Block:   combininghalfmarks}', "");
Expect(0, 65072, '\P{^Block:   combininghalfmarks}', "");
Expect(1, 65071, '\p{Block=_Combining_HALF_MARKS}', "");
Expect(0, 65071, '\p{^Block=_Combining_HALF_MARKS}', "");
Expect(0, 65071, '\P{Block=_Combining_HALF_MARKS}', "");
Expect(1, 65071, '\P{^Block=_Combining_HALF_MARKS}', "");
Expect(0, 65072, '\p{Block=_Combining_HALF_MARKS}', "");
Expect(1, 65072, '\p{^Block=_Combining_HALF_MARKS}', "");
Expect(1, 65072, '\P{Block=_Combining_HALF_MARKS}', "");
Expect(0, 65072, '\P{^Block=_Combining_HALF_MARKS}', "");
Error('\p{Blk=/a/	-HALF_Marks}');
Error('\P{Blk=/a/	-HALF_Marks}');
Expect(1, 65071, '\p{Blk=halfmarks}', "");
Expect(0, 65071, '\p{^Blk=halfmarks}', "");
Expect(0, 65071, '\P{Blk=halfmarks}', "");
Expect(1, 65071, '\P{^Blk=halfmarks}', "");
Expect(0, 65072, '\p{Blk=halfmarks}', "");
Expect(1, 65072, '\p{^Blk=halfmarks}', "");
Expect(1, 65072, '\P{Blk=halfmarks}', "");
Expect(0, 65072, '\P{^Blk=halfmarks}', "");
Expect(1, 65071, '\p{Blk=- half_marks}', "");
Expect(0, 65071, '\p{^Blk=- half_marks}', "");
Expect(0, 65071, '\P{Blk=- half_marks}', "");
Expect(1, 65071, '\P{^Blk=- half_marks}', "");
Expect(0, 65072, '\p{Blk=- half_marks}', "");
Expect(1, 65072, '\p{^Blk=- half_marks}', "");
Expect(1, 65072, '\P{Blk=- half_marks}', "");
Expect(0, 65072, '\P{^Blk=- half_marks}', "");
Error('\p{Is_Block=-/a/COMBINING_Half_Marks}');
Error('\P{Is_Block=-/a/COMBINING_Half_Marks}');
Expect(1, 65071, '\p{Is_Block=combininghalfmarks}', "");
Expect(0, 65071, '\p{^Is_Block=combininghalfmarks}', "");
Expect(0, 65071, '\P{Is_Block=combininghalfmarks}', "");
Expect(1, 65071, '\P{^Is_Block=combininghalfmarks}', "");
Expect(0, 65072, '\p{Is_Block=combininghalfmarks}', "");
Expect(1, 65072, '\p{^Is_Block=combininghalfmarks}', "");
Expect(1, 65072, '\P{Is_Block=combininghalfmarks}', "");
Expect(0, 65072, '\P{^Is_Block=combininghalfmarks}', "");
Expect(1, 65071, '\p{Is_Block=--Combining_HALF_Marks}', "");
Expect(0, 65071, '\p{^Is_Block=--Combining_HALF_Marks}', "");
Expect(0, 65071, '\P{Is_Block=--Combining_HALF_Marks}', "");
Expect(1, 65071, '\P{^Is_Block=--Combining_HALF_Marks}', "");
Expect(0, 65072, '\p{Is_Block=--Combining_HALF_Marks}', "");
Expect(1, 65072, '\p{^Is_Block=--Combining_HALF_Marks}', "");
Expect(1, 65072, '\P{Is_Block=--Combining_HALF_Marks}', "");
Expect(0, 65072, '\P{^Is_Block=--Combining_HALF_Marks}', "");
Error('\p{Is_Blk=_ half_marks:=}');
Error('\P{Is_Blk=_ half_marks:=}');
Expect(1, 65071, '\p{Is_Blk=halfmarks}', "");
Expect(0, 65071, '\p{^Is_Blk=halfmarks}', "");
Expect(0, 65071, '\P{Is_Blk=halfmarks}', "");
Expect(1, 65071, '\P{^Is_Blk=halfmarks}', "");
Expect(0, 65072, '\p{Is_Blk=halfmarks}', "");
Expect(1, 65072, '\p{^Is_Blk=halfmarks}', "");
Expect(1, 65072, '\P{Is_Blk=halfmarks}', "");
Expect(0, 65072, '\P{^Is_Blk=halfmarks}', "");
Expect(1, 65071, '\p{Is_Blk=--Half_marks}', "");
Expect(0, 65071, '\p{^Is_Blk=--Half_marks}', "");
Expect(0, 65071, '\P{Is_Blk=--Half_marks}', "");
Expect(1, 65071, '\P{^Is_Blk=--Half_marks}', "");
Expect(0, 65072, '\p{Is_Blk=--Half_marks}', "");
Expect(1, 65072, '\p{^Is_Blk=--Half_marks}', "");
Expect(1, 65072, '\P{Is_Blk=--Half_marks}', "");
Expect(0, 65072, '\P{^Is_Blk=--Half_marks}', "");
Error('\p{Block=/a/_-Hangul_Syllables}');
Error('\P{Block=/a/_-Hangul_Syllables}');
Expect(1, 55215, '\p{Block=hangulsyllables}', "");
Expect(0, 55215, '\p{^Block=hangulsyllables}', "");
Expect(0, 55215, '\P{Block=hangulsyllables}', "");
Expect(1, 55215, '\P{^Block=hangulsyllables}', "");
Expect(0, 55216, '\p{Block=hangulsyllables}', "");
Expect(1, 55216, '\p{^Block=hangulsyllables}', "");
Expect(1, 55216, '\P{Block=hangulsyllables}', "");
Expect(0, 55216, '\P{^Block=hangulsyllables}', "");
Expect(1, 55215, '\p{Block= hangul_syllables}', "");
Expect(0, 55215, '\p{^Block= hangul_syllables}', "");
Expect(0, 55215, '\P{Block= hangul_syllables}', "");
Expect(1, 55215, '\P{^Block= hangul_syllables}', "");
Expect(0, 55216, '\p{Block= hangul_syllables}', "");
Expect(1, 55216, '\p{^Block= hangul_syllables}', "");
Expect(1, 55216, '\P{Block= hangul_syllables}', "");
Expect(0, 55216, '\P{^Block= hangul_syllables}', "");
Error('\p{Blk=	_hangul/a/}');
Error('\P{Blk=	_hangul/a/}');
Expect(1, 55215, '\p{Blk=hangul}', "");
Expect(0, 55215, '\p{^Blk=hangul}', "");
Expect(0, 55215, '\P{Blk=hangul}', "");
Expect(1, 55215, '\P{^Blk=hangul}', "");
Expect(0, 55216, '\p{Blk=hangul}', "");
Expect(1, 55216, '\p{^Blk=hangul}', "");
Expect(1, 55216, '\P{Blk=hangul}', "");
Expect(0, 55216, '\P{^Blk=hangul}', "");
Expect(1, 55215, '\p{Blk=		hangul}', "");
Expect(0, 55215, '\p{^Blk=		hangul}', "");
Expect(0, 55215, '\P{Blk=		hangul}', "");
Expect(1, 55215, '\P{^Blk=		hangul}', "");
Expect(0, 55216, '\p{Blk=		hangul}', "");
Expect(1, 55216, '\p{^Blk=		hangul}', "");
Expect(1, 55216, '\P{Blk=		hangul}', "");
Expect(0, 55216, '\P{^Blk=		hangul}', "");
Error('\p{Is_Block=_:=hangul_syllables}');
Error('\P{Is_Block=_:=hangul_syllables}');
Expect(1, 55215, '\p{Is_Block=hangulsyllables}', "");
Expect(0, 55215, '\p{^Is_Block=hangulsyllables}', "");
Expect(0, 55215, '\P{Is_Block=hangulsyllables}', "");
Expect(1, 55215, '\P{^Is_Block=hangulsyllables}', "");
Expect(0, 55216, '\p{Is_Block=hangulsyllables}', "");
Expect(1, 55216, '\p{^Is_Block=hangulsyllables}', "");
Expect(1, 55216, '\P{Is_Block=hangulsyllables}', "");
Expect(0, 55216, '\P{^Is_Block=hangulsyllables}', "");
Expect(1, 55215, '\p{Is_Block:	_	HANGUL_Syllables}', "");
Expect(0, 55215, '\p{^Is_Block:	_	HANGUL_Syllables}', "");
Expect(0, 55215, '\P{Is_Block:	_	HANGUL_Syllables}', "");
Expect(1, 55215, '\P{^Is_Block:	_	HANGUL_Syllables}', "");
Expect(0, 55216, '\p{Is_Block:	_	HANGUL_Syllables}', "");
Expect(1, 55216, '\p{^Is_Block:	_	HANGUL_Syllables}', "");
Expect(1, 55216, '\P{Is_Block:	_	HANGUL_Syllables}', "");
Expect(0, 55216, '\P{^Is_Block:	_	HANGUL_Syllables}', "");
Error('\p{Is_Blk=-	Hangul/a/}');
Error('\P{Is_Blk=-	Hangul/a/}');
Expect(1, 55215, '\p{Is_Blk=hangul}', "");
Expect(0, 55215, '\p{^Is_Blk=hangul}', "");
Expect(0, 55215, '\P{Is_Blk=hangul}', "");
Expect(1, 55215, '\P{^Is_Blk=hangul}', "");
Expect(0, 55216, '\p{Is_Blk=hangul}', "");
Expect(1, 55216, '\p{^Is_Blk=hangul}', "");
Expect(1, 55216, '\P{Is_Blk=hangul}', "");
Expect(0, 55216, '\P{^Is_Blk=hangul}', "");
Expect(1, 55215, '\p{Is_Blk=-HANGUL}', "");
Expect(0, 55215, '\p{^Is_Blk=-HANGUL}', "");
Expect(0, 55215, '\P{Is_Blk=-HANGUL}', "");
Expect(1, 55215, '\P{^Is_Blk=-HANGUL}', "");
Expect(0, 55216, '\p{Is_Blk=-HANGUL}', "");
Expect(1, 55216, '\p{^Is_Blk=-HANGUL}', "");
Expect(1, 55216, '\P{Is_Blk=-HANGUL}', "");
Expect(0, 55216, '\P{^Is_Blk=-HANGUL}', "");
Error('\p{Block=-:=Hanunoo}');
Error('\P{Block=-:=Hanunoo}');
Expect(1, 5951, '\p{Block=hanunoo}', "");
Expect(0, 5951, '\p{^Block=hanunoo}', "");
Expect(0, 5951, '\P{Block=hanunoo}', "");
Expect(1, 5951, '\P{^Block=hanunoo}', "");
Expect(0, 5952, '\p{Block=hanunoo}', "");
Expect(1, 5952, '\p{^Block=hanunoo}', "");
Expect(1, 5952, '\P{Block=hanunoo}', "");
Expect(0, 5952, '\P{^Block=hanunoo}', "");
Expect(1, 5951, '\p{Block=_Hanunoo}', "");
Expect(0, 5951, '\p{^Block=_Hanunoo}', "");
Expect(0, 5951, '\P{Block=_Hanunoo}', "");
Expect(1, 5951, '\P{^Block=_Hanunoo}', "");
Expect(0, 5952, '\p{Block=_Hanunoo}', "");
Expect(1, 5952, '\p{^Block=_Hanunoo}', "");
Expect(1, 5952, '\P{Block=_Hanunoo}', "");
Expect(0, 5952, '\P{^Block=_Hanunoo}', "");
Error('\p{Blk=	:=HANUNOO}');
Error('\P{Blk=	:=HANUNOO}');
Expect(1, 5951, '\p{Blk=hanunoo}', "");
Expect(0, 5951, '\p{^Blk=hanunoo}', "");
Expect(0, 5951, '\P{Blk=hanunoo}', "");
Expect(1, 5951, '\P{^Blk=hanunoo}', "");
Expect(0, 5952, '\p{Blk=hanunoo}', "");
Expect(1, 5952, '\p{^Blk=hanunoo}', "");
Expect(1, 5952, '\P{Blk=hanunoo}', "");
Expect(0, 5952, '\P{^Blk=hanunoo}', "");
Expect(1, 5951, '\p{Blk=-Hanunoo}', "");
Expect(0, 5951, '\p{^Blk=-Hanunoo}', "");
Expect(0, 5951, '\P{Blk=-Hanunoo}', "");
Expect(1, 5951, '\P{^Blk=-Hanunoo}', "");
Expect(0, 5952, '\p{Blk=-Hanunoo}', "");
Expect(1, 5952, '\p{^Blk=-Hanunoo}', "");
Expect(1, 5952, '\P{Blk=-Hanunoo}', "");
Expect(0, 5952, '\P{^Blk=-Hanunoo}', "");
Error('\p{Is_Block=-hanunoo/a/}');
Error('\P{Is_Block=-hanunoo/a/}');
Expect(1, 5951, '\p{Is_Block=hanunoo}', "");
Expect(0, 5951, '\p{^Is_Block=hanunoo}', "");
Expect(0, 5951, '\P{Is_Block=hanunoo}', "");
Expect(1, 5951, '\P{^Is_Block=hanunoo}', "");
Expect(0, 5952, '\p{Is_Block=hanunoo}', "");
Expect(1, 5952, '\p{^Is_Block=hanunoo}', "");
Expect(1, 5952, '\P{Is_Block=hanunoo}', "");
Expect(0, 5952, '\P{^Is_Block=hanunoo}', "");
Expect(1, 5951, '\p{Is_Block=	 HANUNOO}', "");
Expect(0, 5951, '\p{^Is_Block=	 HANUNOO}', "");
Expect(0, 5951, '\P{Is_Block=	 HANUNOO}', "");
Expect(1, 5951, '\P{^Is_Block=	 HANUNOO}', "");
Expect(0, 5952, '\p{Is_Block=	 HANUNOO}', "");
Expect(1, 5952, '\p{^Is_Block=	 HANUNOO}', "");
Expect(1, 5952, '\P{Is_Block=	 HANUNOO}', "");
Expect(0, 5952, '\P{^Is_Block=	 HANUNOO}', "");
Error('\p{Is_Blk=:=_	HANUNOO}');
Error('\P{Is_Blk=:=_	HANUNOO}');
Expect(1, 5951, '\p{Is_Blk=hanunoo}', "");
Expect(0, 5951, '\p{^Is_Blk=hanunoo}', "");
Expect(0, 5951, '\P{Is_Blk=hanunoo}', "");
Expect(1, 5951, '\P{^Is_Blk=hanunoo}', "");
Expect(0, 5952, '\p{Is_Blk=hanunoo}', "");
Expect(1, 5952, '\p{^Is_Blk=hanunoo}', "");
Expect(1, 5952, '\P{Is_Blk=hanunoo}', "");
Expect(0, 5952, '\P{^Is_Blk=hanunoo}', "");
Expect(1, 5951, '\p{Is_Blk=- Hanunoo}', "");
Expect(0, 5951, '\p{^Is_Blk=- Hanunoo}', "");
Expect(0, 5951, '\P{Is_Blk=- Hanunoo}', "");
Expect(1, 5951, '\P{^Is_Blk=- Hanunoo}', "");
Expect(0, 5952, '\p{Is_Blk=- Hanunoo}', "");
Expect(1, 5952, '\p{^Is_Blk=- Hanunoo}', "");
Expect(1, 5952, '\P{Is_Blk=- Hanunoo}', "");
Expect(0, 5952, '\P{^Is_Blk=- Hanunoo}', "");
Error('\p{Block=_HEBREW/a/}');
Error('\P{Block=_HEBREW/a/}');
Expect(1, 1535, '\p{Block=hebrew}', "");
Expect(0, 1535, '\p{^Block=hebrew}', "");
Expect(0, 1535, '\P{Block=hebrew}', "");
Expect(1, 1535, '\P{^Block=hebrew}', "");
Expect(0, 1536, '\p{Block=hebrew}', "");
Expect(1, 1536, '\p{^Block=hebrew}', "");
Expect(1, 1536, '\P{Block=hebrew}', "");
Expect(0, 1536, '\P{^Block=hebrew}', "");
Expect(1, 1535, '\p{Block=_-HEBREW}', "");
Expect(0, 1535, '\p{^Block=_-HEBREW}', "");
Expect(0, 1535, '\P{Block=_-HEBREW}', "");
Expect(1, 1535, '\P{^Block=_-HEBREW}', "");
Expect(0, 1536, '\p{Block=_-HEBREW}', "");
Expect(1, 1536, '\p{^Block=_-HEBREW}', "");
Expect(1, 1536, '\P{Block=_-HEBREW}', "");
Expect(0, 1536, '\P{^Block=_-HEBREW}', "");
Error('\p{Blk: _-Hebrew/a/}');
Error('\P{Blk: _-Hebrew/a/}');
Expect(1, 1535, '\p{Blk=hebrew}', "");
Expect(0, 1535, '\p{^Blk=hebrew}', "");
Expect(0, 1535, '\P{Blk=hebrew}', "");
Expect(1, 1535, '\P{^Blk=hebrew}', "");
Expect(0, 1536, '\p{Blk=hebrew}', "");
Expect(1, 1536, '\p{^Blk=hebrew}', "");
Expect(1, 1536, '\P{Blk=hebrew}', "");
Expect(0, 1536, '\P{^Blk=hebrew}', "");
Expect(1, 1535, '\p{Blk=_ hebrew}', "");
Expect(0, 1535, '\p{^Blk=_ hebrew}', "");
Expect(0, 1535, '\P{Blk=_ hebrew}', "");
Expect(1, 1535, '\P{^Blk=_ hebrew}', "");
Expect(0, 1536, '\p{Blk=_ hebrew}', "");
Expect(1, 1536, '\p{^Blk=_ hebrew}', "");
Expect(1, 1536, '\P{Blk=_ hebrew}', "");
Expect(0, 1536, '\P{^Blk=_ hebrew}', "");
Error('\p{Is_Block=/a/_	Hebrew}');
Error('\P{Is_Block=/a/_	Hebrew}');
Expect(1, 1535, '\p{Is_Block: hebrew}', "");
Expect(0, 1535, '\p{^Is_Block: hebrew}', "");
Expect(0, 1535, '\P{Is_Block: hebrew}', "");
Expect(1, 1535, '\P{^Is_Block: hebrew}', "");
Expect(0, 1536, '\p{Is_Block: hebrew}', "");
Expect(1, 1536, '\p{^Is_Block: hebrew}', "");
Expect(1, 1536, '\P{Is_Block: hebrew}', "");
Expect(0, 1536, '\P{^Is_Block: hebrew}', "");
Expect(1, 1535, '\p{Is_Block=_Hebrew}', "");
Expect(0, 1535, '\p{^Is_Block=_Hebrew}', "");
Expect(0, 1535, '\P{Is_Block=_Hebrew}', "");
Expect(1, 1535, '\P{^Is_Block=_Hebrew}', "");
Expect(0, 1536, '\p{Is_Block=_Hebrew}', "");
Expect(1, 1536, '\p{^Is_Block=_Hebrew}', "");
Expect(1, 1536, '\P{Is_Block=_Hebrew}', "");
Expect(0, 1536, '\P{^Is_Block=_Hebrew}', "");
Error('\p{Is_Blk=/a/ -hebrew}');
Error('\P{Is_Blk=/a/ -hebrew}');
Expect(1, 1535, '\p{Is_Blk=hebrew}', "");
Expect(0, 1535, '\p{^Is_Blk=hebrew}', "");
Expect(0, 1535, '\P{Is_Blk=hebrew}', "");
Expect(1, 1535, '\P{^Is_Blk=hebrew}', "");
Expect(0, 1536, '\p{Is_Blk=hebrew}', "");
Expect(1, 1536, '\p{^Is_Blk=hebrew}', "");
Expect(1, 1536, '\P{Is_Blk=hebrew}', "");
Expect(0, 1536, '\P{^Is_Blk=hebrew}', "");
Expect(1, 1535, '\p{Is_Blk=	 hebrew}', "");
Expect(0, 1535, '\p{^Is_Blk=	 hebrew}', "");
Expect(0, 1535, '\P{Is_Blk=	 hebrew}', "");
Expect(1, 1535, '\P{^Is_Blk=	 hebrew}', "");
Expect(0, 1536, '\p{Is_Blk=	 hebrew}', "");
Expect(1, 1536, '\p{^Is_Blk=	 hebrew}', "");
Expect(1, 1536, '\P{Is_Blk=	 hebrew}', "");
Expect(0, 1536, '\P{^Is_Blk=	 hebrew}', "");
Error('\p{Block=_ HIGH_Private_use_Surrogates:=}');
Error('\P{Block=_ HIGH_Private_use_Surrogates:=}');
Expect(1, 56319, '\p{Block=highprivateusesurrogates}', "");
Expect(0, 56319, '\p{^Block=highprivateusesurrogates}', "");
Expect(0, 56319, '\P{Block=highprivateusesurrogates}', "");
Expect(1, 56319, '\P{^Block=highprivateusesurrogates}', "");
Expect(0, 57344, '\p{Block=highprivateusesurrogates}', "");
Expect(1, 57344, '\p{^Block=highprivateusesurrogates}', "");
Expect(1, 57344, '\P{Block=highprivateusesurrogates}', "");
Expect(0, 57344, '\P{^Block=highprivateusesurrogates}', "");
Expect(1, 56319, '\p{Block= 	High_PRIVATE_USE_surrogates}', "");
Expect(0, 56319, '\p{^Block= 	High_PRIVATE_USE_surrogates}', "");
Expect(0, 56319, '\P{Block= 	High_PRIVATE_USE_surrogates}', "");
Expect(1, 56319, '\P{^Block= 	High_PRIVATE_USE_surrogates}', "");
Expect(0, 57344, '\p{Block= 	High_PRIVATE_USE_surrogates}', "");
Expect(1, 57344, '\p{^Block= 	High_PRIVATE_USE_surrogates}', "");
Expect(1, 57344, '\P{Block= 	High_PRIVATE_USE_surrogates}', "");
Expect(0, 57344, '\P{^Block= 	High_PRIVATE_USE_surrogates}', "");
Error('\p{Blk=/a/-_High_PU_Surrogates}');
Error('\P{Blk=/a/-_High_PU_Surrogates}');
Expect(1, 56319, '\p{Blk=highpusurrogates}', "");
Expect(0, 56319, '\p{^Blk=highpusurrogates}', "");
Expect(0, 56319, '\P{Blk=highpusurrogates}', "");
Expect(1, 56319, '\P{^Blk=highpusurrogates}', "");
Expect(0, 57344, '\p{Blk=highpusurrogates}', "");
Expect(1, 57344, '\p{^Blk=highpusurrogates}', "");
Expect(1, 57344, '\P{Blk=highpusurrogates}', "");
Expect(0, 57344, '\P{^Blk=highpusurrogates}', "");
Expect(1, 56319, '\p{Blk=- High_PU_surrogates}', "");
Expect(0, 56319, '\p{^Blk=- High_PU_surrogates}', "");
Expect(0, 56319, '\P{Blk=- High_PU_surrogates}', "");
Expect(1, 56319, '\P{^Blk=- High_PU_surrogates}', "");
Expect(0, 57344, '\p{Blk=- High_PU_surrogates}', "");
Expect(1, 57344, '\p{^Blk=- High_PU_surrogates}', "");
Expect(1, 57344, '\P{Blk=- High_PU_surrogates}', "");
Expect(0, 57344, '\P{^Blk=- High_PU_surrogates}', "");
Error('\p{Is_Block=/a/high_Private_USE_SURROGATES}');
Error('\P{Is_Block=/a/high_Private_USE_SURROGATES}');
Expect(1, 56319, '\p{Is_Block=highprivateusesurrogates}', "");
Expect(0, 56319, '\p{^Is_Block=highprivateusesurrogates}', "");
Expect(0, 56319, '\P{Is_Block=highprivateusesurrogates}', "");
Expect(1, 56319, '\P{^Is_Block=highprivateusesurrogates}', "");
Expect(0, 57344, '\p{Is_Block=highprivateusesurrogates}', "");
Expect(1, 57344, '\p{^Is_Block=highprivateusesurrogates}', "");
Expect(1, 57344, '\P{Is_Block=highprivateusesurrogates}', "");
Expect(0, 57344, '\P{^Is_Block=highprivateusesurrogates}', "");
Expect(1, 56319, '\p{Is_Block=	_High_private_USE_SURROGATES}', "");
Expect(0, 56319, '\p{^Is_Block=	_High_private_USE_SURROGATES}', "");
Expect(0, 56319, '\P{Is_Block=	_High_private_USE_SURROGATES}', "");
Expect(1, 56319, '\P{^Is_Block=	_High_private_USE_SURROGATES}', "");
Expect(0, 57344, '\p{Is_Block=	_High_private_USE_SURROGATES}', "");
Expect(1, 57344, '\p{^Is_Block=	_High_private_USE_SURROGATES}', "");
Expect(1, 57344, '\P{Is_Block=	_High_private_USE_SURROGATES}', "");
Expect(0, 57344, '\P{^Is_Block=	_High_private_USE_SURROGATES}', "");
Error('\p{Is_Blk=_/a/High_PU_Surrogates}');
Error('\P{Is_Blk=_/a/High_PU_Surrogates}');
Expect(1, 56319, '\p{Is_Blk=highpusurrogates}', "");
Expect(0, 56319, '\p{^Is_Blk=highpusurrogates}', "");
Expect(0, 56319, '\P{Is_Blk=highpusurrogates}', "");
Expect(1, 56319, '\P{^Is_Blk=highpusurrogates}', "");
Expect(0, 57344, '\p{Is_Blk=highpusurrogates}', "");
Expect(1, 57344, '\p{^Is_Blk=highpusurrogates}', "");
Expect(1, 57344, '\P{Is_Blk=highpusurrogates}', "");
Expect(0, 57344, '\P{^Is_Blk=highpusurrogates}', "");
Expect(1, 56319, '\p{Is_Blk:	 High_PU_Surrogates}', "");
Expect(0, 56319, '\p{^Is_Blk:	 High_PU_Surrogates}', "");
Expect(0, 56319, '\P{Is_Blk:	 High_PU_Surrogates}', "");
Expect(1, 56319, '\P{^Is_Blk:	 High_PU_Surrogates}', "");
Expect(0, 57344, '\p{Is_Blk:	 High_PU_Surrogates}', "");
Expect(1, 57344, '\p{^Is_Blk:	 High_PU_Surrogates}', "");
Expect(1, 57344, '\P{Is_Blk:	 High_PU_Surrogates}', "");
Expect(0, 57344, '\P{^Is_Blk:	 High_PU_Surrogates}', "");
Error('\p{Block=-/a/HIGH_Surrogates}');
Error('\P{Block=-/a/HIGH_Surrogates}');
Expect(1, 56191, '\p{Block=highsurrogates}', "");
Expect(0, 56191, '\p{^Block=highsurrogates}', "");
Expect(0, 56191, '\P{Block=highsurrogates}', "");
Expect(1, 56191, '\P{^Block=highsurrogates}', "");
Expect(0, 57344, '\p{Block=highsurrogates}', "");
Expect(1, 57344, '\p{^Block=highsurrogates}', "");
Expect(1, 57344, '\P{Block=highsurrogates}', "");
Expect(0, 57344, '\P{^Block=highsurrogates}', "");
Expect(1, 56191, '\p{Block=HIGH_Surrogates}', "");
Expect(0, 56191, '\p{^Block=HIGH_Surrogates}', "");
Expect(0, 56191, '\P{Block=HIGH_Surrogates}', "");
Expect(1, 56191, '\P{^Block=HIGH_Surrogates}', "");
Expect(0, 57344, '\p{Block=HIGH_Surrogates}', "");
Expect(1, 57344, '\p{^Block=HIGH_Surrogates}', "");
Expect(1, 57344, '\P{Block=HIGH_Surrogates}', "");
Expect(0, 57344, '\P{^Block=HIGH_Surrogates}', "");
Error('\p{Blk=_ HIGH_Surrogates:=}');
Error('\P{Blk=_ HIGH_Surrogates:=}');
Expect(1, 56191, '\p{Blk=highsurrogates}', "");
Expect(0, 56191, '\p{^Blk=highsurrogates}', "");
Expect(0, 56191, '\P{Blk=highsurrogates}', "");
Expect(1, 56191, '\P{^Blk=highsurrogates}', "");
Expect(0, 57344, '\p{Blk=highsurrogates}', "");
Expect(1, 57344, '\p{^Blk=highsurrogates}', "");
Expect(1, 57344, '\P{Blk=highsurrogates}', "");
Expect(0, 57344, '\P{^Blk=highsurrogates}', "");
Expect(1, 56191, '\p{Blk=_ HIGH_Surrogates}', "");
Expect(0, 56191, '\p{^Blk=_ HIGH_Surrogates}', "");
Expect(0, 56191, '\P{Blk=_ HIGH_Surrogates}', "");
Expect(1, 56191, '\P{^Blk=_ HIGH_Surrogates}', "");
Expect(0, 57344, '\p{Blk=_ HIGH_Surrogates}', "");
Expect(1, 57344, '\p{^Blk=_ HIGH_Surrogates}', "");
Expect(1, 57344, '\P{Blk=_ HIGH_Surrogates}', "");
Expect(0, 57344, '\P{^Blk=_ HIGH_Surrogates}', "");
Error('\p{Is_Block= /a/High_SURROGATES}');
Error('\P{Is_Block= /a/High_SURROGATES}');
Expect(1, 56191, '\p{Is_Block=highsurrogates}', "");
Expect(0, 56191, '\p{^Is_Block=highsurrogates}', "");
Expect(0, 56191, '\P{Is_Block=highsurrogates}', "");
Expect(1, 56191, '\P{^Is_Block=highsurrogates}', "");
Expect(0, 57344, '\p{Is_Block=highsurrogates}', "");
Expect(1, 57344, '\p{^Is_Block=highsurrogates}', "");
Expect(1, 57344, '\P{Is_Block=highsurrogates}', "");
Expect(0, 57344, '\P{^Is_Block=highsurrogates}', "");
Expect(1, 56191, '\p{Is_Block=  High_SURROGATES}', "");
Expect(0, 56191, '\p{^Is_Block=  High_SURROGATES}', "");
Expect(0, 56191, '\P{Is_Block=  High_SURROGATES}', "");
Expect(1, 56191, '\P{^Is_Block=  High_SURROGATES}', "");
Expect(0, 57344, '\p{Is_Block=  High_SURROGATES}', "");
Expect(1, 57344, '\p{^Is_Block=  High_SURROGATES}', "");
Expect(1, 57344, '\P{Is_Block=  High_SURROGATES}', "");
Expect(0, 57344, '\P{^Is_Block=  High_SURROGATES}', "");
Error('\p{Is_Blk: 	High_SURROGATES:=}');
Error('\P{Is_Blk: 	High_SURROGATES:=}');
Expect(1, 56191, '\p{Is_Blk:   highsurrogates}', "");
Expect(0, 56191, '\p{^Is_Blk:   highsurrogates}', "");
Expect(0, 56191, '\P{Is_Blk:   highsurrogates}', "");
Expect(1, 56191, '\P{^Is_Blk:   highsurrogates}', "");
Expect(0, 57344, '\p{Is_Blk:   highsurrogates}', "");
Expect(1, 57344, '\p{^Is_Blk:   highsurrogates}', "");
Expect(1, 57344, '\P{Is_Blk:   highsurrogates}', "");
Expect(0, 57344, '\P{^Is_Blk:   highsurrogates}', "");
Expect(1, 56191, '\p{Is_Blk=High_SURROGATES}', "");
Expect(0, 56191, '\p{^Is_Blk=High_SURROGATES}', "");
Expect(0, 56191, '\P{Is_Blk=High_SURROGATES}', "");
Expect(1, 56191, '\P{^Is_Blk=High_SURROGATES}', "");
Expect(0, 57344, '\p{Is_Blk=High_SURROGATES}', "");
Expect(1, 57344, '\p{^Is_Blk=High_SURROGATES}', "");
Expect(1, 57344, '\P{Is_Blk=High_SURROGATES}', "");
Expect(0, 57344, '\P{^Is_Blk=High_SURROGATES}', "");
Error('\p{Block:   /a/-Hiragana}');
Error('\P{Block:   /a/-Hiragana}');
Expect(1, 12447, '\p{Block:	hiragana}', "");
Expect(0, 12447, '\p{^Block:	hiragana}', "");
Expect(0, 12447, '\P{Block:	hiragana}', "");
Expect(1, 12447, '\P{^Block:	hiragana}', "");
Expect(0, 12448, '\p{Block:	hiragana}', "");
Expect(1, 12448, '\p{^Block:	hiragana}', "");
Expect(1, 12448, '\P{Block:	hiragana}', "");
Expect(0, 12448, '\P{^Block:	hiragana}', "");
Expect(1, 12447, '\p{Block: HIRAGANA}', "");
Expect(0, 12447, '\p{^Block: HIRAGANA}', "");
Expect(0, 12447, '\P{Block: HIRAGANA}', "");
Expect(1, 12447, '\P{^Block: HIRAGANA}', "");
Expect(0, 12448, '\p{Block: HIRAGANA}', "");
Expect(1, 12448, '\p{^Block: HIRAGANA}', "");
Expect(1, 12448, '\P{Block: HIRAGANA}', "");
Expect(0, 12448, '\P{^Block: HIRAGANA}', "");
Error('\p{Blk=/a/	_Hiragana}');
Error('\P{Blk=/a/	_Hiragana}');
Expect(1, 12447, '\p{Blk: hiragana}', "");
Expect(0, 12447, '\p{^Blk: hiragana}', "");
Expect(0, 12447, '\P{Blk: hiragana}', "");
Expect(1, 12447, '\P{^Blk: hiragana}', "");
Expect(0, 12448, '\p{Blk: hiragana}', "");
Expect(1, 12448, '\p{^Blk: hiragana}', "");
Expect(1, 12448, '\P{Blk: hiragana}', "");
Expect(0, 12448, '\P{^Blk: hiragana}', "");
Expect(1, 12447, '\p{Blk= -Hiragana}', "");
Expect(0, 12447, '\p{^Blk= -Hiragana}', "");
Expect(0, 12447, '\P{Blk= -Hiragana}', "");
Expect(1, 12447, '\P{^Blk= -Hiragana}', "");
Expect(0, 12448, '\p{Blk= -Hiragana}', "");
Expect(1, 12448, '\p{^Blk= -Hiragana}', "");
Expect(1, 12448, '\P{Blk= -Hiragana}', "");
Expect(0, 12448, '\P{^Blk= -Hiragana}', "");
Error('\p{Is_Block::=_ Hiragana}');
Error('\P{Is_Block::=_ Hiragana}');
Expect(1, 12447, '\p{Is_Block: hiragana}', "");
Expect(0, 12447, '\p{^Is_Block: hiragana}', "");
Expect(0, 12447, '\P{Is_Block: hiragana}', "");
Expect(1, 12447, '\P{^Is_Block: hiragana}', "");
Expect(0, 12448, '\p{Is_Block: hiragana}', "");
Expect(1, 12448, '\p{^Is_Block: hiragana}', "");
Expect(1, 12448, '\P{Is_Block: hiragana}', "");
Expect(0, 12448, '\P{^Is_Block: hiragana}', "");
Expect(1, 12447, '\p{Is_Block=-Hiragana}', "");
Expect(0, 12447, '\p{^Is_Block=-Hiragana}', "");
Expect(0, 12447, '\P{Is_Block=-Hiragana}', "");
Expect(1, 12447, '\P{^Is_Block=-Hiragana}', "");
Expect(0, 12448, '\p{Is_Block=-Hiragana}', "");
Expect(1, 12448, '\p{^Is_Block=-Hiragana}', "");
Expect(1, 12448, '\P{Is_Block=-Hiragana}', "");
Expect(0, 12448, '\P{^Is_Block=-Hiragana}', "");
Error('\p{Is_Blk=:=	_Hiragana}');
Error('\P{Is_Blk=:=	_Hiragana}');
Expect(1, 12447, '\p{Is_Blk=hiragana}', "");
Expect(0, 12447, '\p{^Is_Blk=hiragana}', "");
Expect(0, 12447, '\P{Is_Blk=hiragana}', "");
Expect(1, 12447, '\P{^Is_Blk=hiragana}', "");
Expect(0, 12448, '\p{Is_Blk=hiragana}', "");
Expect(1, 12448, '\p{^Is_Blk=hiragana}', "");
Expect(1, 12448, '\P{Is_Blk=hiragana}', "");
Expect(0, 12448, '\P{^Is_Blk=hiragana}', "");
Expect(1, 12447, '\p{Is_Blk: hiragana}', "");
Expect(0, 12447, '\p{^Is_Blk: hiragana}', "");
Expect(0, 12447, '\P{Is_Blk: hiragana}', "");
Expect(1, 12447, '\P{^Is_Blk: hiragana}', "");
Expect(0, 12448, '\p{Is_Blk: hiragana}', "");
Expect(1, 12448, '\p{^Is_Blk: hiragana}', "");
Expect(1, 12448, '\P{Is_Blk: hiragana}', "");
Expect(0, 12448, '\P{^Is_Blk: hiragana}', "");
Error('\p{Block=:= Ideographic_Description_characters}');
Error('\P{Block=:= Ideographic_Description_characters}');
Expect(1, 12287, '\p{Block=ideographicdescriptioncharacters}', "");
Expect(0, 12287, '\p{^Block=ideographicdescriptioncharacters}', "");
Expect(0, 12287, '\P{Block=ideographicdescriptioncharacters}', "");
Expect(1, 12287, '\P{^Block=ideographicdescriptioncharacters}', "");
Expect(0, 12288, '\p{Block=ideographicdescriptioncharacters}', "");
Expect(1, 12288, '\p{^Block=ideographicdescriptioncharacters}', "");
Expect(1, 12288, '\P{Block=ideographicdescriptioncharacters}', "");
Expect(0, 12288, '\P{^Block=ideographicdescriptioncharacters}', "");
Expect(1, 12287, '\p{Block=	 ideographic_description_Characters}', "");
Expect(0, 12287, '\p{^Block=	 ideographic_description_Characters}', "");
Expect(0, 12287, '\P{Block=	 ideographic_description_Characters}', "");
Expect(1, 12287, '\P{^Block=	 ideographic_description_Characters}', "");
Expect(0, 12288, '\p{Block=	 ideographic_description_Characters}', "");
Expect(1, 12288, '\p{^Block=	 ideographic_description_Characters}', "");
Expect(1, 12288, '\P{Block=	 ideographic_description_Characters}', "");
Expect(0, 12288, '\P{^Block=	 ideographic_description_Characters}', "");
Error('\p{Blk=-_IDC/a/}');
Error('\P{Blk=-_IDC/a/}');
Expect(1, 12287, '\p{Blk=idc}', "");
Expect(0, 12287, '\p{^Blk=idc}', "");
Expect(0, 12287, '\P{Blk=idc}', "");
Expect(1, 12287, '\P{^Blk=idc}', "");
Expect(0, 12288, '\p{Blk=idc}', "");
Expect(1, 12288, '\p{^Blk=idc}', "");
Expect(1, 12288, '\P{Blk=idc}', "");
Expect(0, 12288, '\P{^Blk=idc}', "");
Expect(1, 12287, '\p{Blk=-IDC}', "");
Expect(0, 12287, '\p{^Blk=-IDC}', "");
Expect(0, 12287, '\P{Blk=-IDC}', "");
Expect(1, 12287, '\P{^Blk=-IDC}', "");
Expect(0, 12288, '\p{Blk=-IDC}', "");
Expect(1, 12288, '\p{^Blk=-IDC}', "");
Expect(1, 12288, '\P{Blk=-IDC}', "");
Expect(0, 12288, '\P{^Blk=-IDC}', "");
Error('\p{Is_Block=-:=Ideographic_Description_characters}');
Error('\P{Is_Block=-:=Ideographic_Description_characters}');
Expect(1, 12287, '\p{Is_Block=ideographicdescriptioncharacters}', "");
Expect(0, 12287, '\p{^Is_Block=ideographicdescriptioncharacters}', "");
Expect(0, 12287, '\P{Is_Block=ideographicdescriptioncharacters}', "");
Expect(1, 12287, '\P{^Is_Block=ideographicdescriptioncharacters}', "");
Expect(0, 12288, '\p{Is_Block=ideographicdescriptioncharacters}', "");
Expect(1, 12288, '\p{^Is_Block=ideographicdescriptioncharacters}', "");
Expect(1, 12288, '\P{Is_Block=ideographicdescriptioncharacters}', "");
Expect(0, 12288, '\P{^Is_Block=ideographicdescriptioncharacters}', "");
Expect(1, 12287, '\p{Is_Block= 	ideographic_Description_Characters}', "");
Expect(0, 12287, '\p{^Is_Block= 	ideographic_Description_Characters}', "");
Expect(0, 12287, '\P{Is_Block= 	ideographic_Description_Characters}', "");
Expect(1, 12287, '\P{^Is_Block= 	ideographic_Description_Characters}', "");
Expect(0, 12288, '\p{Is_Block= 	ideographic_Description_Characters}', "");
Expect(1, 12288, '\p{^Is_Block= 	ideographic_Description_Characters}', "");
Expect(1, 12288, '\P{Is_Block= 	ideographic_Description_Characters}', "");
Expect(0, 12288, '\P{^Is_Block= 	ideographic_Description_Characters}', "");
Error('\p{Is_Blk= :=idc}');
Error('\P{Is_Blk= :=idc}');
Expect(1, 12287, '\p{Is_Blk=idc}', "");
Expect(0, 12287, '\p{^Is_Blk=idc}', "");
Expect(0, 12287, '\P{Is_Blk=idc}', "");
Expect(1, 12287, '\P{^Is_Blk=idc}', "");
Expect(0, 12288, '\p{Is_Blk=idc}', "");
Expect(1, 12288, '\p{^Is_Blk=idc}', "");
Expect(1, 12288, '\P{Is_Blk=idc}', "");
Expect(0, 12288, '\P{^Is_Blk=idc}', "");
Expect(1, 12287, '\p{Is_Blk=-_IDC}', "");
Expect(0, 12287, '\p{^Is_Blk=-_IDC}', "");
Expect(0, 12287, '\P{Is_Blk=-_IDC}', "");
Expect(1, 12287, '\P{^Is_Blk=-_IDC}', "");
Expect(0, 12288, '\p{Is_Blk=-_IDC}', "");
Expect(1, 12288, '\p{^Is_Blk=-_IDC}', "");
Expect(1, 12288, '\P{Is_Blk=-_IDC}', "");
Expect(0, 12288, '\P{^Is_Blk=-_IDC}', "");
Error('\p{Block=	_ipa_extensions:=}');
Error('\P{Block=	_ipa_extensions:=}');
Expect(1, 687, '\p{Block=ipaextensions}', "");
Expect(0, 687, '\p{^Block=ipaextensions}', "");
Expect(0, 687, '\P{Block=ipaextensions}', "");
Expect(1, 687, '\P{^Block=ipaextensions}', "");
Expect(0, 688, '\p{Block=ipaextensions}', "");
Expect(1, 688, '\p{^Block=ipaextensions}', "");
Expect(1, 688, '\P{Block=ipaextensions}', "");
Expect(0, 688, '\P{^Block=ipaextensions}', "");
Expect(1, 687, '\p{Block:   IPA_EXTENSIONS}', "");
Expect(0, 687, '\p{^Block:   IPA_EXTENSIONS}', "");
Expect(0, 687, '\P{Block:   IPA_EXTENSIONS}', "");
Expect(1, 687, '\P{^Block:   IPA_EXTENSIONS}', "");
Expect(0, 688, '\p{Block:   IPA_EXTENSIONS}', "");
Expect(1, 688, '\p{^Block:   IPA_EXTENSIONS}', "");
Expect(1, 688, '\P{Block:   IPA_EXTENSIONS}', "");
Expect(0, 688, '\P{^Block:   IPA_EXTENSIONS}', "");
Error('\p{Blk=-	IPA_EXT:=}');
Error('\P{Blk=-	IPA_EXT:=}');
Expect(1, 687, '\p{Blk=ipaext}', "");
Expect(0, 687, '\p{^Blk=ipaext}', "");
Expect(0, 687, '\P{Blk=ipaext}', "");
Expect(1, 687, '\P{^Blk=ipaext}', "");
Expect(0, 688, '\p{Blk=ipaext}', "");
Expect(1, 688, '\p{^Blk=ipaext}', "");
Expect(1, 688, '\P{Blk=ipaext}', "");
Expect(0, 688, '\P{^Blk=ipaext}', "");
Expect(1, 687, '\p{Blk= _IPA_Ext}', "");
Expect(0, 687, '\p{^Blk= _IPA_Ext}', "");
Expect(0, 687, '\P{Blk= _IPA_Ext}', "");
Expect(1, 687, '\P{^Blk= _IPA_Ext}', "");
Expect(0, 688, '\p{Blk= _IPA_Ext}', "");
Expect(1, 688, '\p{^Blk= _IPA_Ext}', "");
Expect(1, 688, '\P{Blk= _IPA_Ext}', "");
Expect(0, 688, '\P{^Blk= _IPA_Ext}', "");
Error('\p{Is_Block=-:=IPA_extensions}');
Error('\P{Is_Block=-:=IPA_extensions}');
Expect(1, 687, '\p{Is_Block=ipaextensions}', "");
Expect(0, 687, '\p{^Is_Block=ipaextensions}', "");
Expect(0, 687, '\P{Is_Block=ipaextensions}', "");
Expect(1, 687, '\P{^Is_Block=ipaextensions}', "");
Expect(0, 688, '\p{Is_Block=ipaextensions}', "");
Expect(1, 688, '\p{^Is_Block=ipaextensions}', "");
Expect(1, 688, '\P{Is_Block=ipaextensions}', "");
Expect(0, 688, '\P{^Is_Block=ipaextensions}', "");
Expect(1, 687, '\p{Is_Block=	IPA_EXTENSIONS}', "");
Expect(0, 687, '\p{^Is_Block=	IPA_EXTENSIONS}', "");
Expect(0, 687, '\P{Is_Block=	IPA_EXTENSIONS}', "");
Expect(1, 687, '\P{^Is_Block=	IPA_EXTENSIONS}', "");
Expect(0, 688, '\p{Is_Block=	IPA_EXTENSIONS}', "");
Expect(1, 688, '\p{^Is_Block=	IPA_EXTENSIONS}', "");
Expect(1, 688, '\P{Is_Block=	IPA_EXTENSIONS}', "");
Expect(0, 688, '\P{^Is_Block=	IPA_EXTENSIONS}', "");
Error('\p{Is_Blk=	/a/IPA_Ext}');
Error('\P{Is_Blk=	/a/IPA_Ext}');
Expect(1, 687, '\p{Is_Blk=ipaext}', "");
Expect(0, 687, '\p{^Is_Blk=ipaext}', "");
Expect(0, 687, '\P{Is_Blk=ipaext}', "");
Expect(1, 687, '\P{^Is_Blk=ipaext}', "");
Expect(0, 688, '\p{Is_Blk=ipaext}', "");
Expect(1, 688, '\p{^Is_Blk=ipaext}', "");
Expect(1, 688, '\P{Is_Blk=ipaext}', "");
Expect(0, 688, '\P{^Is_Blk=ipaext}', "");
Expect(1, 687, '\p{Is_Blk= _IPA_EXT}', "");
Expect(0, 687, '\p{^Is_Blk= _IPA_EXT}', "");
Expect(0, 687, '\P{Is_Blk= _IPA_EXT}', "");
Expect(1, 687, '\P{^Is_Blk= _IPA_EXT}', "");
Expect(0, 688, '\p{Is_Blk= _IPA_EXT}', "");
Expect(1, 688, '\p{^Is_Blk= _IPA_EXT}', "");
Expect(1, 688, '\P{Is_Blk= _IPA_EXT}', "");
Expect(0, 688, '\P{^Is_Blk= _IPA_EXT}', "");
Error('\p{Block=	/a/Imperial_Aramaic}');
Error('\P{Block=	/a/Imperial_Aramaic}');
Expect(1, 67679, '\p{Block=imperialaramaic}', "");
Expect(0, 67679, '\p{^Block=imperialaramaic}', "");
Expect(0, 67679, '\P{Block=imperialaramaic}', "");
Expect(1, 67679, '\P{^Block=imperialaramaic}', "");
Expect(0, 67680, '\p{Block=imperialaramaic}', "");
Expect(1, 67680, '\p{^Block=imperialaramaic}', "");
Expect(1, 67680, '\P{Block=imperialaramaic}', "");
Expect(0, 67680, '\P{^Block=imperialaramaic}', "");
Expect(1, 67679, '\p{Block=_IMPERIAL_Aramaic}', "");
Expect(0, 67679, '\p{^Block=_IMPERIAL_Aramaic}', "");
Expect(0, 67679, '\P{Block=_IMPERIAL_Aramaic}', "");
Expect(1, 67679, '\P{^Block=_IMPERIAL_Aramaic}', "");
Expect(0, 67680, '\p{Block=_IMPERIAL_Aramaic}', "");
Expect(1, 67680, '\p{^Block=_IMPERIAL_Aramaic}', "");
Expect(1, 67680, '\P{Block=_IMPERIAL_Aramaic}', "");
Expect(0, 67680, '\P{^Block=_IMPERIAL_Aramaic}', "");
Error('\p{Blk=:=_	imperial_Aramaic}');
Error('\P{Blk=:=_	imperial_Aramaic}');
Expect(1, 67679, '\p{Blk=imperialaramaic}', "");
Expect(0, 67679, '\p{^Blk=imperialaramaic}', "");
Expect(0, 67679, '\P{Blk=imperialaramaic}', "");
Expect(1, 67679, '\P{^Blk=imperialaramaic}', "");
Expect(0, 67680, '\p{Blk=imperialaramaic}', "");
Expect(1, 67680, '\p{^Blk=imperialaramaic}', "");
Expect(1, 67680, '\P{Blk=imperialaramaic}', "");
Expect(0, 67680, '\P{^Blk=imperialaramaic}', "");
Expect(1, 67679, '\p{Blk= Imperial_aramaic}', "");
Expect(0, 67679, '\p{^Blk= Imperial_aramaic}', "");
Expect(0, 67679, '\P{Blk= Imperial_aramaic}', "");
Expect(1, 67679, '\P{^Blk= Imperial_aramaic}', "");
Expect(0, 67680, '\p{Blk= Imperial_aramaic}', "");
Expect(1, 67680, '\p{^Blk= Imperial_aramaic}', "");
Expect(1, 67680, '\P{Blk= Imperial_aramaic}', "");
Expect(0, 67680, '\P{^Blk= Imperial_aramaic}', "");
Error('\p{Is_Block=:= -Imperial_ARAMAIC}');
Error('\P{Is_Block=:= -Imperial_ARAMAIC}');
Expect(1, 67679, '\p{Is_Block=imperialaramaic}', "");
Expect(0, 67679, '\p{^Is_Block=imperialaramaic}', "");
Expect(0, 67679, '\P{Is_Block=imperialaramaic}', "");
Expect(1, 67679, '\P{^Is_Block=imperialaramaic}', "");
Expect(0, 67680, '\p{Is_Block=imperialaramaic}', "");
Expect(1, 67680, '\p{^Is_Block=imperialaramaic}', "");
Expect(1, 67680, '\P{Is_Block=imperialaramaic}', "");
Expect(0, 67680, '\P{^Is_Block=imperialaramaic}', "");
Expect(1, 67679, '\p{Is_Block:   --imperial_Aramaic}', "");
Expect(0, 67679, '\p{^Is_Block:   --imperial_Aramaic}', "");
Expect(0, 67679, '\P{Is_Block:   --imperial_Aramaic}', "");
Expect(1, 67679, '\P{^Is_Block:   --imperial_Aramaic}', "");
Expect(0, 67680, '\p{Is_Block:   --imperial_Aramaic}', "");
Expect(1, 67680, '\p{^Is_Block:   --imperial_Aramaic}', "");
Expect(1, 67680, '\P{Is_Block:   --imperial_Aramaic}', "");
Expect(0, 67680, '\P{^Is_Block:   --imperial_Aramaic}', "");
Error('\p{Is_Blk=:=-imperial_Aramaic}');
Error('\P{Is_Blk=:=-imperial_Aramaic}');
Expect(1, 67679, '\p{Is_Blk=imperialaramaic}', "");
Expect(0, 67679, '\p{^Is_Blk=imperialaramaic}', "");
Expect(0, 67679, '\P{Is_Blk=imperialaramaic}', "");
Expect(1, 67679, '\P{^Is_Blk=imperialaramaic}', "");
Expect(0, 67680, '\p{Is_Blk=imperialaramaic}', "");
Expect(1, 67680, '\p{^Is_Blk=imperialaramaic}', "");
Expect(1, 67680, '\P{Is_Blk=imperialaramaic}', "");
Expect(0, 67680, '\P{^Is_Blk=imperialaramaic}', "");
Expect(1, 67679, '\p{Is_Blk=--IMPERIAL_ARAMAIC}', "");
Expect(0, 67679, '\p{^Is_Blk=--IMPERIAL_ARAMAIC}', "");
Expect(0, 67679, '\P{Is_Blk=--IMPERIAL_ARAMAIC}', "");
Expect(1, 67679, '\P{^Is_Blk=--IMPERIAL_ARAMAIC}', "");
Expect(0, 67680, '\p{Is_Blk=--IMPERIAL_ARAMAIC}', "");
Expect(1, 67680, '\p{^Is_Blk=--IMPERIAL_ARAMAIC}', "");
Expect(1, 67680, '\P{Is_Blk=--IMPERIAL_ARAMAIC}', "");
Expect(0, 67680, '\P{^Is_Blk=--IMPERIAL_ARAMAIC}', "");
Error('\p{Block=	_Common_Indic_Number_Forms/a/}');
Error('\P{Block=	_Common_Indic_Number_Forms/a/}');
Expect(1, 43071, '\p{Block=commonindicnumberforms}', "");
Expect(0, 43071, '\p{^Block=commonindicnumberforms}', "");
Expect(0, 43071, '\P{Block=commonindicnumberforms}', "");
Expect(1, 43071, '\P{^Block=commonindicnumberforms}', "");
Expect(0, 43072, '\p{Block=commonindicnumberforms}', "");
Expect(1, 43072, '\p{^Block=commonindicnumberforms}', "");
Expect(1, 43072, '\P{Block=commonindicnumberforms}', "");
Expect(0, 43072, '\P{^Block=commonindicnumberforms}', "");
Expect(1, 43071, '\p{Block=		COMMON_INDIC_number_forms}', "");
Expect(0, 43071, '\p{^Block=		COMMON_INDIC_number_forms}', "");
Expect(0, 43071, '\P{Block=		COMMON_INDIC_number_forms}', "");
Expect(1, 43071, '\P{^Block=		COMMON_INDIC_number_forms}', "");
Expect(0, 43072, '\p{Block=		COMMON_INDIC_number_forms}', "");
Expect(1, 43072, '\p{^Block=		COMMON_INDIC_number_forms}', "");
Expect(1, 43072, '\P{Block=		COMMON_INDIC_number_forms}', "");
Expect(0, 43072, '\P{^Block=		COMMON_INDIC_number_forms}', "");
Error('\p{Blk=/a/indic_Number_Forms}');
Error('\P{Blk=/a/indic_Number_Forms}');
Expect(1, 43071, '\p{Blk=indicnumberforms}', "");
Expect(0, 43071, '\p{^Blk=indicnumberforms}', "");
Expect(0, 43071, '\P{Blk=indicnumberforms}', "");
Expect(1, 43071, '\P{^Blk=indicnumberforms}', "");
Expect(0, 43072, '\p{Blk=indicnumberforms}', "");
Expect(1, 43072, '\p{^Blk=indicnumberforms}', "");
Expect(1, 43072, '\P{Blk=indicnumberforms}', "");
Expect(0, 43072, '\P{^Blk=indicnumberforms}', "");
Expect(1, 43071, '\p{Blk=--Indic_NUMBER_Forms}', "");
Expect(0, 43071, '\p{^Blk=--Indic_NUMBER_Forms}', "");
Expect(0, 43071, '\P{Blk=--Indic_NUMBER_Forms}', "");
Expect(1, 43071, '\P{^Blk=--Indic_NUMBER_Forms}', "");
Expect(0, 43072, '\p{Blk=--Indic_NUMBER_Forms}', "");
Expect(1, 43072, '\p{^Blk=--Indic_NUMBER_Forms}', "");
Expect(1, 43072, '\P{Blk=--Indic_NUMBER_Forms}', "");
Expect(0, 43072, '\P{^Blk=--Indic_NUMBER_Forms}', "");
Error('\p{Is_Block=:=common_Indic_Number_forms}');
Error('\P{Is_Block=:=common_Indic_Number_forms}');
Expect(1, 43071, '\p{Is_Block=commonindicnumberforms}', "");
Expect(0, 43071, '\p{^Is_Block=commonindicnumberforms}', "");
Expect(0, 43071, '\P{Is_Block=commonindicnumberforms}', "");
Expect(1, 43071, '\P{^Is_Block=commonindicnumberforms}', "");
Expect(0, 43072, '\p{Is_Block=commonindicnumberforms}', "");
Expect(1, 43072, '\p{^Is_Block=commonindicnumberforms}', "");
Expect(1, 43072, '\P{Is_Block=commonindicnumberforms}', "");
Expect(0, 43072, '\P{^Is_Block=commonindicnumberforms}', "");
Expect(1, 43071, '\p{Is_Block=-_Common_Indic_NUMBER_FORMS}', "");
Expect(0, 43071, '\p{^Is_Block=-_Common_Indic_NUMBER_FORMS}', "");
Expect(0, 43071, '\P{Is_Block=-_Common_Indic_NUMBER_FORMS}', "");
Expect(1, 43071, '\P{^Is_Block=-_Common_Indic_NUMBER_FORMS}', "");
Expect(0, 43072, '\p{Is_Block=-_Common_Indic_NUMBER_FORMS}', "");
Expect(1, 43072, '\p{^Is_Block=-_Common_Indic_NUMBER_FORMS}', "");
Expect(1, 43072, '\P{Is_Block=-_Common_Indic_NUMBER_FORMS}', "");
Expect(0, 43072, '\P{^Is_Block=-_Common_Indic_NUMBER_FORMS}', "");
Error('\p{Is_Blk=:=--indic_number_FORMS}');
Error('\P{Is_Blk=:=--indic_number_FORMS}');
Expect(1, 43071, '\p{Is_Blk:	indicnumberforms}', "");
Expect(0, 43071, '\p{^Is_Blk:	indicnumberforms}', "");
Expect(0, 43071, '\P{Is_Blk:	indicnumberforms}', "");
Expect(1, 43071, '\P{^Is_Blk:	indicnumberforms}', "");
Expect(0, 43072, '\p{Is_Blk:	indicnumberforms}', "");
Expect(1, 43072, '\p{^Is_Blk:	indicnumberforms}', "");
Expect(1, 43072, '\P{Is_Blk:	indicnumberforms}', "");
Expect(0, 43072, '\P{^Is_Blk:	indicnumberforms}', "");
Expect(1, 43071, '\p{Is_Blk:   -	INDIC_NUMBER_Forms}', "");
Expect(0, 43071, '\p{^Is_Blk:   -	INDIC_NUMBER_Forms}', "");
Expect(0, 43071, '\P{Is_Blk:   -	INDIC_NUMBER_Forms}', "");
Expect(1, 43071, '\P{^Is_Blk:   -	INDIC_NUMBER_Forms}', "");
Expect(0, 43072, '\p{Is_Blk:   -	INDIC_NUMBER_Forms}', "");
Expect(1, 43072, '\p{^Is_Blk:   -	INDIC_NUMBER_Forms}', "");
Expect(1, 43072, '\P{Is_Blk:   -	INDIC_NUMBER_Forms}', "");
Expect(0, 43072, '\P{^Is_Blk:   -	INDIC_NUMBER_Forms}', "");
Error('\p{Block=-_Inscriptional_PAHLAVI/a/}');
Error('\P{Block=-_Inscriptional_PAHLAVI/a/}');
Expect(1, 68479, '\p{Block=inscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^Block=inscriptionalpahlavi}', "");
Expect(0, 68479, '\P{Block=inscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^Block=inscriptionalpahlavi}', "");
Expect(0, 68480, '\p{Block=inscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^Block=inscriptionalpahlavi}', "");
Expect(1, 68480, '\P{Block=inscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^Block=inscriptionalpahlavi}', "");
Expect(1, 68479, '\p{Block= Inscriptional_Pahlavi}', "");
Expect(0, 68479, '\p{^Block= Inscriptional_Pahlavi}', "");
Expect(0, 68479, '\P{Block= Inscriptional_Pahlavi}', "");
Expect(1, 68479, '\P{^Block= Inscriptional_Pahlavi}', "");
Expect(0, 68480, '\p{Block= Inscriptional_Pahlavi}', "");
Expect(1, 68480, '\p{^Block= Inscriptional_Pahlavi}', "");
Expect(1, 68480, '\P{Block= Inscriptional_Pahlavi}', "");
Expect(0, 68480, '\P{^Block= Inscriptional_Pahlavi}', "");
Error('\p{Blk=_INSCRIPTIONAL_Pahlavi/a/}');
Error('\P{Blk=_INSCRIPTIONAL_Pahlavi/a/}');
Expect(1, 68479, '\p{Blk=inscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^Blk=inscriptionalpahlavi}', "");
Expect(0, 68479, '\P{Blk=inscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^Blk=inscriptionalpahlavi}', "");
Expect(0, 68480, '\p{Blk=inscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^Blk=inscriptionalpahlavi}', "");
Expect(1, 68480, '\P{Blk=inscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^Blk=inscriptionalpahlavi}', "");
Expect(1, 68479, '\p{Blk=__Inscriptional_PAHLAVI}', "");
Expect(0, 68479, '\p{^Blk=__Inscriptional_PAHLAVI}', "");
Expect(0, 68479, '\P{Blk=__Inscriptional_PAHLAVI}', "");
Expect(1, 68479, '\P{^Blk=__Inscriptional_PAHLAVI}', "");
Expect(0, 68480, '\p{Blk=__Inscriptional_PAHLAVI}', "");
Expect(1, 68480, '\p{^Blk=__Inscriptional_PAHLAVI}', "");
Expect(1, 68480, '\P{Blk=__Inscriptional_PAHLAVI}', "");
Expect(0, 68480, '\P{^Blk=__Inscriptional_PAHLAVI}', "");
Error('\p{Is_Block=_/a/inscriptional_Pahlavi}');
Error('\P{Is_Block=_/a/inscriptional_Pahlavi}');
Expect(1, 68479, '\p{Is_Block=inscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^Is_Block=inscriptionalpahlavi}', "");
Expect(0, 68479, '\P{Is_Block=inscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^Is_Block=inscriptionalpahlavi}', "");
Expect(0, 68480, '\p{Is_Block=inscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^Is_Block=inscriptionalpahlavi}', "");
Expect(1, 68480, '\P{Is_Block=inscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^Is_Block=inscriptionalpahlavi}', "");
Expect(1, 68479, '\p{Is_Block:   	_Inscriptional_PAHLAVI}', "");
Expect(0, 68479, '\p{^Is_Block:   	_Inscriptional_PAHLAVI}', "");
Expect(0, 68479, '\P{Is_Block:   	_Inscriptional_PAHLAVI}', "");
Expect(1, 68479, '\P{^Is_Block:   	_Inscriptional_PAHLAVI}', "");
Expect(0, 68480, '\p{Is_Block:   	_Inscriptional_PAHLAVI}', "");
Expect(1, 68480, '\p{^Is_Block:   	_Inscriptional_PAHLAVI}', "");
Expect(1, 68480, '\P{Is_Block:   	_Inscriptional_PAHLAVI}', "");
Expect(0, 68480, '\P{^Is_Block:   	_Inscriptional_PAHLAVI}', "");
Error('\p{Is_Blk=_ inscriptional_Pahlavi/a/}');
Error('\P{Is_Blk=_ inscriptional_Pahlavi/a/}');
Expect(1, 68479, '\p{Is_Blk=inscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^Is_Blk=inscriptionalpahlavi}', "");
Expect(0, 68479, '\P{Is_Blk=inscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^Is_Blk=inscriptionalpahlavi}', "");
Expect(0, 68480, '\p{Is_Blk=inscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^Is_Blk=inscriptionalpahlavi}', "");
Expect(1, 68480, '\P{Is_Blk=inscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^Is_Blk=inscriptionalpahlavi}', "");
Expect(1, 68479, '\p{Is_Blk=-inscriptional_Pahlavi}', "");
Expect(0, 68479, '\p{^Is_Blk=-inscriptional_Pahlavi}', "");
Expect(0, 68479, '\P{Is_Blk=-inscriptional_Pahlavi}', "");
Expect(1, 68479, '\P{^Is_Blk=-inscriptional_Pahlavi}', "");
Expect(0, 68480, '\p{Is_Blk=-inscriptional_Pahlavi}', "");
Expect(1, 68480, '\p{^Is_Blk=-inscriptional_Pahlavi}', "");
Expect(1, 68480, '\P{Is_Blk=-inscriptional_Pahlavi}', "");
Expect(0, 68480, '\P{^Is_Blk=-inscriptional_Pahlavi}', "");
Error('\p{Block=-:=inscriptional_parthian}');
Error('\P{Block=-:=inscriptional_parthian}');
Expect(1, 68447, '\p{Block=inscriptionalparthian}', "");
Expect(0, 68447, '\p{^Block=inscriptionalparthian}', "");
Expect(0, 68447, '\P{Block=inscriptionalparthian}', "");
Expect(1, 68447, '\P{^Block=inscriptionalparthian}', "");
Expect(0, 68448, '\p{Block=inscriptionalparthian}', "");
Expect(1, 68448, '\p{^Block=inscriptionalparthian}', "");
Expect(1, 68448, '\P{Block=inscriptionalparthian}', "");
Expect(0, 68448, '\P{^Block=inscriptionalparthian}', "");
Expect(1, 68447, '\p{Block=_ inscriptional_PARTHIAN}', "");
Expect(0, 68447, '\p{^Block=_ inscriptional_PARTHIAN}', "");
Expect(0, 68447, '\P{Block=_ inscriptional_PARTHIAN}', "");
Expect(1, 68447, '\P{^Block=_ inscriptional_PARTHIAN}', "");
Expect(0, 68448, '\p{Block=_ inscriptional_PARTHIAN}', "");
Expect(1, 68448, '\p{^Block=_ inscriptional_PARTHIAN}', "");
Expect(1, 68448, '\P{Block=_ inscriptional_PARTHIAN}', "");
Expect(0, 68448, '\P{^Block=_ inscriptional_PARTHIAN}', "");
Error('\p{Blk=-/a/Inscriptional_parthian}');
Error('\P{Blk=-/a/Inscriptional_parthian}');
Expect(1, 68447, '\p{Blk=inscriptionalparthian}', "");
Expect(0, 68447, '\p{^Blk=inscriptionalparthian}', "");
Expect(0, 68447, '\P{Blk=inscriptionalparthian}', "");
Expect(1, 68447, '\P{^Blk=inscriptionalparthian}', "");
Expect(0, 68448, '\p{Blk=inscriptionalparthian}', "");
Expect(1, 68448, '\p{^Blk=inscriptionalparthian}', "");
Expect(1, 68448, '\P{Blk=inscriptionalparthian}', "");
Expect(0, 68448, '\P{^Blk=inscriptionalparthian}', "");
Expect(1, 68447, '\p{Blk=__INSCRIPTIONAL_parthian}', "");
Expect(0, 68447, '\p{^Blk=__INSCRIPTIONAL_parthian}', "");
Expect(0, 68447, '\P{Blk=__INSCRIPTIONAL_parthian}', "");
Expect(1, 68447, '\P{^Blk=__INSCRIPTIONAL_parthian}', "");
Expect(0, 68448, '\p{Blk=__INSCRIPTIONAL_parthian}', "");
Expect(1, 68448, '\p{^Blk=__INSCRIPTIONAL_parthian}', "");
Expect(1, 68448, '\P{Blk=__INSCRIPTIONAL_parthian}', "");
Expect(0, 68448, '\P{^Blk=__INSCRIPTIONAL_parthian}', "");
Error('\p{Is_Block=		inscriptional_Parthian:=}');
Error('\P{Is_Block=		inscriptional_Parthian:=}');
Expect(1, 68447, '\p{Is_Block=inscriptionalparthian}', "");
Expect(0, 68447, '\p{^Is_Block=inscriptionalparthian}', "");
Expect(0, 68447, '\P{Is_Block=inscriptionalparthian}', "");
Expect(1, 68447, '\P{^Is_Block=inscriptionalparthian}', "");
Expect(0, 68448, '\p{Is_Block=inscriptionalparthian}', "");
Expect(1, 68448, '\p{^Is_Block=inscriptionalparthian}', "");
Expect(1, 68448, '\P{Is_Block=inscriptionalparthian}', "");
Expect(0, 68448, '\P{^Is_Block=inscriptionalparthian}', "");
Expect(1, 68447, '\p{Is_Block=		inscriptional_Parthian}', "");
Expect(0, 68447, '\p{^Is_Block=		inscriptional_Parthian}', "");
Expect(0, 68447, '\P{Is_Block=		inscriptional_Parthian}', "");
Expect(1, 68447, '\P{^Is_Block=		inscriptional_Parthian}', "");
Expect(0, 68448, '\p{Is_Block=		inscriptional_Parthian}', "");
Expect(1, 68448, '\p{^Is_Block=		inscriptional_Parthian}', "");
Expect(1, 68448, '\P{Is_Block=		inscriptional_Parthian}', "");
Expect(0, 68448, '\P{^Is_Block=		inscriptional_Parthian}', "");
Error('\p{Is_Blk=-:=INSCRIPTIONAL_Parthian}');
Error('\P{Is_Blk=-:=INSCRIPTIONAL_Parthian}');
Expect(1, 68447, '\p{Is_Blk=inscriptionalparthian}', "");
Expect(0, 68447, '\p{^Is_Blk=inscriptionalparthian}', "");
Expect(0, 68447, '\P{Is_Blk=inscriptionalparthian}', "");
Expect(1, 68447, '\P{^Is_Blk=inscriptionalparthian}', "");
Expect(0, 68448, '\p{Is_Blk=inscriptionalparthian}', "");
Expect(1, 68448, '\p{^Is_Blk=inscriptionalparthian}', "");
Expect(1, 68448, '\P{Is_Blk=inscriptionalparthian}', "");
Expect(0, 68448, '\P{^Is_Blk=inscriptionalparthian}', "");
Expect(1, 68447, '\p{Is_Blk= 	INSCRIPTIONAL_Parthian}', "");
Expect(0, 68447, '\p{^Is_Blk= 	INSCRIPTIONAL_Parthian}', "");
Expect(0, 68447, '\P{Is_Blk= 	INSCRIPTIONAL_Parthian}', "");
Expect(1, 68447, '\P{^Is_Blk= 	INSCRIPTIONAL_Parthian}', "");
Expect(0, 68448, '\p{Is_Blk= 	INSCRIPTIONAL_Parthian}', "");
Expect(1, 68448, '\p{^Is_Blk= 	INSCRIPTIONAL_Parthian}', "");
Expect(1, 68448, '\P{Is_Blk= 	INSCRIPTIONAL_Parthian}', "");
Expect(0, 68448, '\P{^Is_Blk= 	INSCRIPTIONAL_Parthian}', "");
Error('\p{Block::=_ HANGUL_Jamo}');
Error('\P{Block::=_ HANGUL_Jamo}');
Expect(1, 4607, '\p{Block:   hanguljamo}', "");
Expect(0, 4607, '\p{^Block:   hanguljamo}', "");
Expect(0, 4607, '\P{Block:   hanguljamo}', "");
Expect(1, 4607, '\P{^Block:   hanguljamo}', "");
Expect(0, 4608, '\p{Block:   hanguljamo}', "");
Expect(1, 4608, '\p{^Block:   hanguljamo}', "");
Expect(1, 4608, '\P{Block:   hanguljamo}', "");
Expect(0, 4608, '\P{^Block:   hanguljamo}', "");
Expect(1, 4607, '\p{Block= _Hangul_Jamo}', "");
Expect(0, 4607, '\p{^Block= _Hangul_Jamo}', "");
Expect(0, 4607, '\P{Block= _Hangul_Jamo}', "");
Expect(1, 4607, '\P{^Block= _Hangul_Jamo}', "");
Expect(0, 4608, '\p{Block= _Hangul_Jamo}', "");
Expect(1, 4608, '\p{^Block= _Hangul_Jamo}', "");
Expect(1, 4608, '\P{Block= _Hangul_Jamo}', "");
Expect(0, 4608, '\P{^Block= _Hangul_Jamo}', "");
Error('\p{Blk=	/a/jamo}');
Error('\P{Blk=	/a/jamo}');
Expect(1, 4607, '\p{Blk=jamo}', "");
Expect(0, 4607, '\p{^Blk=jamo}', "");
Expect(0, 4607, '\P{Blk=jamo}', "");
Expect(1, 4607, '\P{^Blk=jamo}', "");
Expect(0, 4608, '\p{Blk=jamo}', "");
Expect(1, 4608, '\p{^Blk=jamo}', "");
Expect(1, 4608, '\P{Blk=jamo}', "");
Expect(0, 4608, '\P{^Blk=jamo}', "");
Expect(1, 4607, '\p{Blk=_-Jamo}', "");
Expect(0, 4607, '\p{^Blk=_-Jamo}', "");
Expect(0, 4607, '\P{Blk=_-Jamo}', "");
Expect(1, 4607, '\P{^Blk=_-Jamo}', "");
Expect(0, 4608, '\p{Blk=_-Jamo}', "");
Expect(1, 4608, '\p{^Blk=_-Jamo}', "");
Expect(1, 4608, '\P{Blk=_-Jamo}', "");
Expect(0, 4608, '\P{^Blk=_-Jamo}', "");
Error('\p{Is_Block=/a/	_HANGUL_JAMO}');
Error('\P{Is_Block=/a/	_HANGUL_JAMO}');
Expect(1, 4607, '\p{Is_Block=hanguljamo}', "");
Expect(0, 4607, '\p{^Is_Block=hanguljamo}', "");
Expect(0, 4607, '\P{Is_Block=hanguljamo}', "");
Expect(1, 4607, '\P{^Is_Block=hanguljamo}', "");
Expect(0, 4608, '\p{Is_Block=hanguljamo}', "");
Expect(1, 4608, '\p{^Is_Block=hanguljamo}', "");
Expect(1, 4608, '\P{Is_Block=hanguljamo}', "");
Expect(0, 4608, '\P{^Is_Block=hanguljamo}', "");
Expect(1, 4607, '\p{Is_Block=_hangul_JAMO}', "");
Expect(0, 4607, '\p{^Is_Block=_hangul_JAMO}', "");
Expect(0, 4607, '\P{Is_Block=_hangul_JAMO}', "");
Expect(1, 4607, '\P{^Is_Block=_hangul_JAMO}', "");
Expect(0, 4608, '\p{Is_Block=_hangul_JAMO}', "");
Expect(1, 4608, '\p{^Is_Block=_hangul_JAMO}', "");
Expect(1, 4608, '\P{Is_Block=_hangul_JAMO}', "");
Expect(0, 4608, '\P{^Is_Block=_hangul_JAMO}', "");
Error('\p{Is_Blk=:=Jamo}');
Error('\P{Is_Blk=:=Jamo}');
Expect(1, 4607, '\p{Is_Blk=jamo}', "");
Expect(0, 4607, '\p{^Is_Blk=jamo}', "");
Expect(0, 4607, '\P{Is_Blk=jamo}', "");
Expect(1, 4607, '\P{^Is_Blk=jamo}', "");
Expect(0, 4608, '\p{Is_Blk=jamo}', "");
Expect(1, 4608, '\p{^Is_Blk=jamo}', "");
Expect(1, 4608, '\P{Is_Blk=jamo}', "");
Expect(0, 4608, '\P{^Is_Blk=jamo}', "");
Expect(1, 4607, '\p{Is_Blk=-	jamo}', "");
Expect(0, 4607, '\p{^Is_Blk=-	jamo}', "");
Expect(0, 4607, '\P{Is_Blk=-	jamo}', "");
Expect(1, 4607, '\P{^Is_Blk=-	jamo}', "");
Expect(0, 4608, '\p{Is_Blk=-	jamo}', "");
Expect(1, 4608, '\p{^Is_Blk=-	jamo}', "");
Expect(1, 4608, '\P{Is_Blk=-	jamo}', "");
Expect(0, 4608, '\P{^Is_Blk=-	jamo}', "");
Error('\p{Block=-/a/Hangul_Jamo_Extended_A}');
Error('\P{Block=-/a/Hangul_Jamo_Extended_A}');
Expect(1, 43391, '\p{Block=hanguljamoextendeda}', "");
Expect(0, 43391, '\p{^Block=hanguljamoextendeda}', "");
Expect(0, 43391, '\P{Block=hanguljamoextendeda}', "");
Expect(1, 43391, '\P{^Block=hanguljamoextendeda}', "");
Expect(0, 43392, '\p{Block=hanguljamoextendeda}', "");
Expect(1, 43392, '\p{^Block=hanguljamoextendeda}', "");
Expect(1, 43392, '\P{Block=hanguljamoextendeda}', "");
Expect(0, 43392, '\P{^Block=hanguljamoextendeda}', "");
Expect(1, 43391, '\p{Block=_hangul_Jamo_EXTENDED_A}', "");
Expect(0, 43391, '\p{^Block=_hangul_Jamo_EXTENDED_A}', "");
Expect(0, 43391, '\P{Block=_hangul_Jamo_EXTENDED_A}', "");
Expect(1, 43391, '\P{^Block=_hangul_Jamo_EXTENDED_A}', "");
Expect(0, 43392, '\p{Block=_hangul_Jamo_EXTENDED_A}', "");
Expect(1, 43392, '\p{^Block=_hangul_Jamo_EXTENDED_A}', "");
Expect(1, 43392, '\P{Block=_hangul_Jamo_EXTENDED_A}', "");
Expect(0, 43392, '\P{^Block=_hangul_Jamo_EXTENDED_A}', "");
Error('\p{Blk=_:=jamo_ext_A}');
Error('\P{Blk=_:=jamo_ext_A}');
Expect(1, 43391, '\p{Blk=jamoexta}', "");
Expect(0, 43391, '\p{^Blk=jamoexta}', "");
Expect(0, 43391, '\P{Blk=jamoexta}', "");
Expect(1, 43391, '\P{^Blk=jamoexta}', "");
Expect(0, 43392, '\p{Blk=jamoexta}', "");
Expect(1, 43392, '\p{^Blk=jamoexta}', "");
Expect(1, 43392, '\P{Blk=jamoexta}', "");
Expect(0, 43392, '\P{^Blk=jamoexta}', "");
Expect(1, 43391, '\p{Blk=	 jamo_Ext_A}', "");
Expect(0, 43391, '\p{^Blk=	 jamo_Ext_A}', "");
Expect(0, 43391, '\P{Blk=	 jamo_Ext_A}', "");
Expect(1, 43391, '\P{^Blk=	 jamo_Ext_A}', "");
Expect(0, 43392, '\p{Blk=	 jamo_Ext_A}', "");
Expect(1, 43392, '\p{^Blk=	 jamo_Ext_A}', "");
Expect(1, 43392, '\P{Blk=	 jamo_Ext_A}', "");
Expect(0, 43392, '\P{^Blk=	 jamo_Ext_A}', "");
Error('\p{Is_Block=/a/HANGUL_Jamo_Extended_A}');
Error('\P{Is_Block=/a/HANGUL_Jamo_Extended_A}');
Expect(1, 43391, '\p{Is_Block:	hanguljamoextendeda}', "");
Expect(0, 43391, '\p{^Is_Block:	hanguljamoextendeda}', "");
Expect(0, 43391, '\P{Is_Block:	hanguljamoextendeda}', "");
Expect(1, 43391, '\P{^Is_Block:	hanguljamoextendeda}', "");
Expect(0, 43392, '\p{Is_Block:	hanguljamoextendeda}', "");
Expect(1, 43392, '\p{^Is_Block:	hanguljamoextendeda}', "");
Expect(1, 43392, '\P{Is_Block:	hanguljamoextendeda}', "");
Expect(0, 43392, '\P{^Is_Block:	hanguljamoextendeda}', "");
Expect(1, 43391, '\p{Is_Block=_HANGUL_Jamo_EXTENDED_a}', "");
Expect(0, 43391, '\p{^Is_Block=_HANGUL_Jamo_EXTENDED_a}', "");
Expect(0, 43391, '\P{Is_Block=_HANGUL_Jamo_EXTENDED_a}', "");
Expect(1, 43391, '\P{^Is_Block=_HANGUL_Jamo_EXTENDED_a}', "");
Expect(0, 43392, '\p{Is_Block=_HANGUL_Jamo_EXTENDED_a}', "");
Expect(1, 43392, '\p{^Is_Block=_HANGUL_Jamo_EXTENDED_a}', "");
Expect(1, 43392, '\P{Is_Block=_HANGUL_Jamo_EXTENDED_a}', "");
Expect(0, 43392, '\P{^Is_Block=_HANGUL_Jamo_EXTENDED_a}', "");
Error('\p{Is_Blk:   		JAMO_ext_a/a/}');
Error('\P{Is_Blk:   		JAMO_ext_a/a/}');
Expect(1, 43391, '\p{Is_Blk=jamoexta}', "");
Expect(0, 43391, '\p{^Is_Blk=jamoexta}', "");
Expect(0, 43391, '\P{Is_Blk=jamoexta}', "");
Expect(1, 43391, '\P{^Is_Blk=jamoexta}', "");
Expect(0, 43392, '\p{Is_Blk=jamoexta}', "");
Expect(1, 43392, '\p{^Is_Blk=jamoexta}', "");
Expect(1, 43392, '\P{Is_Blk=jamoexta}', "");
Expect(0, 43392, '\P{^Is_Blk=jamoexta}', "");
Expect(1, 43391, '\p{Is_Blk=- Jamo_EXT_A}', "");
Expect(0, 43391, '\p{^Is_Blk=- Jamo_EXT_A}', "");
Expect(0, 43391, '\P{Is_Blk=- Jamo_EXT_A}', "");
Expect(1, 43391, '\P{^Is_Blk=- Jamo_EXT_A}', "");
Expect(0, 43392, '\p{Is_Blk=- Jamo_EXT_A}', "");
Expect(1, 43392, '\p{^Is_Blk=- Jamo_EXT_A}', "");
Expect(1, 43392, '\P{Is_Blk=- Jamo_EXT_A}', "");
Expect(0, 43392, '\P{^Is_Blk=- Jamo_EXT_A}', "");
Error('\p{Block=:=_Hangul_Jamo_EXTENDED_b}');
Error('\P{Block=:=_Hangul_Jamo_EXTENDED_b}');
Expect(1, 55295, '\p{Block=hanguljamoextendedb}', "");
Expect(0, 55295, '\p{^Block=hanguljamoextendedb}', "");
Expect(0, 55295, '\P{Block=hanguljamoextendedb}', "");
Expect(1, 55295, '\P{^Block=hanguljamoextendedb}', "");
Expect(0, 57344, '\p{Block=hanguljamoextendedb}', "");
Expect(1, 57344, '\p{^Block=hanguljamoextendedb}', "");
Expect(1, 57344, '\P{Block=hanguljamoextendedb}', "");
Expect(0, 57344, '\P{^Block=hanguljamoextendedb}', "");
Expect(1, 55295, '\p{Block=_Hangul_Jamo_EXTENDED_B}', "");
Expect(0, 55295, '\p{^Block=_Hangul_Jamo_EXTENDED_B}', "");
Expect(0, 55295, '\P{Block=_Hangul_Jamo_EXTENDED_B}', "");
Expect(1, 55295, '\P{^Block=_Hangul_Jamo_EXTENDED_B}', "");
Expect(0, 57344, '\p{Block=_Hangul_Jamo_EXTENDED_B}', "");
Expect(1, 57344, '\p{^Block=_Hangul_Jamo_EXTENDED_B}', "");
Expect(1, 57344, '\P{Block=_Hangul_Jamo_EXTENDED_B}', "");
Expect(0, 57344, '\P{^Block=_Hangul_Jamo_EXTENDED_B}', "");
Error('\p{Blk=-/a/Jamo_EXT_b}');
Error('\P{Blk=-/a/Jamo_EXT_b}');
Expect(1, 55295, '\p{Blk=jamoextb}', "");
Expect(0, 55295, '\p{^Blk=jamoextb}', "");
Expect(0, 55295, '\P{Blk=jamoextb}', "");
Expect(1, 55295, '\P{^Blk=jamoextb}', "");
Expect(0, 57344, '\p{Blk=jamoextb}', "");
Expect(1, 57344, '\p{^Blk=jamoextb}', "");
Expect(1, 57344, '\P{Blk=jamoextb}', "");
Expect(0, 57344, '\P{^Blk=jamoextb}', "");
Expect(1, 55295, '\p{Blk=	_JAMO_ext_B}', "");
Expect(0, 55295, '\p{^Blk=	_JAMO_ext_B}', "");
Expect(0, 55295, '\P{Blk=	_JAMO_ext_B}', "");
Expect(1, 55295, '\P{^Blk=	_JAMO_ext_B}', "");
Expect(0, 57344, '\p{Blk=	_JAMO_ext_B}', "");
Expect(1, 57344, '\p{^Blk=	_JAMO_ext_B}', "");
Expect(1, 57344, '\P{Blk=	_JAMO_ext_B}', "");
Expect(0, 57344, '\P{^Blk=	_JAMO_ext_B}', "");
Error('\p{Is_Block=:= _HANGUL_JAMO_Extended_B}');
Error('\P{Is_Block=:= _HANGUL_JAMO_Extended_B}');
Expect(1, 55295, '\p{Is_Block=hanguljamoextendedb}', "");
Expect(0, 55295, '\p{^Is_Block=hanguljamoextendedb}', "");
Expect(0, 55295, '\P{Is_Block=hanguljamoextendedb}', "");
Expect(1, 55295, '\P{^Is_Block=hanguljamoextendedb}', "");
Expect(0, 57344, '\p{Is_Block=hanguljamoextendedb}', "");
Expect(1, 57344, '\p{^Is_Block=hanguljamoextendedb}', "");
Expect(1, 57344, '\P{Is_Block=hanguljamoextendedb}', "");
Expect(0, 57344, '\P{^Is_Block=hanguljamoextendedb}', "");
Expect(1, 55295, '\p{Is_Block=	hangul_JAMO_extended_B}', "");
Expect(0, 55295, '\p{^Is_Block=	hangul_JAMO_extended_B}', "");
Expect(0, 55295, '\P{Is_Block=	hangul_JAMO_extended_B}', "");
Expect(1, 55295, '\P{^Is_Block=	hangul_JAMO_extended_B}', "");
Expect(0, 57344, '\p{Is_Block=	hangul_JAMO_extended_B}', "");
Expect(1, 57344, '\p{^Is_Block=	hangul_JAMO_extended_B}', "");
Expect(1, 57344, '\P{Is_Block=	hangul_JAMO_extended_B}', "");
Expect(0, 57344, '\P{^Is_Block=	hangul_JAMO_extended_B}', "");
Error('\p{Is_Blk=/a/jamo_Ext_B}');
Error('\P{Is_Blk=/a/jamo_Ext_B}');
Expect(1, 55295, '\p{Is_Blk=jamoextb}', "");
Expect(0, 55295, '\p{^Is_Blk=jamoextb}', "");
Expect(0, 55295, '\P{Is_Blk=jamoextb}', "");
Expect(1, 55295, '\P{^Is_Blk=jamoextb}', "");
Expect(0, 57344, '\p{Is_Blk=jamoextb}', "");
Expect(1, 57344, '\p{^Is_Blk=jamoextb}', "");
Expect(1, 57344, '\P{Is_Blk=jamoextb}', "");
Expect(0, 57344, '\P{^Is_Blk=jamoextb}', "");
Expect(1, 55295, '\p{Is_Blk= JAMO_EXT_B}', "");
Expect(0, 55295, '\p{^Is_Blk= JAMO_EXT_B}', "");
Expect(0, 55295, '\P{Is_Blk= JAMO_EXT_B}', "");
Expect(1, 55295, '\P{^Is_Blk= JAMO_EXT_B}', "");
Expect(0, 57344, '\p{Is_Blk= JAMO_EXT_B}', "");
Expect(1, 57344, '\p{^Is_Blk= JAMO_EXT_B}', "");
Expect(1, 57344, '\P{Is_Blk= JAMO_EXT_B}', "");
Expect(0, 57344, '\P{^Is_Blk= JAMO_EXT_B}', "");
Error('\p{Block=-Javanese/a/}');
Error('\P{Block=-Javanese/a/}');
Expect(1, 43487, '\p{Block=javanese}', "");
Expect(0, 43487, '\p{^Block=javanese}', "");
Expect(0, 43487, '\P{Block=javanese}', "");
Expect(1, 43487, '\P{^Block=javanese}', "");
Expect(0, 43488, '\p{Block=javanese}', "");
Expect(1, 43488, '\p{^Block=javanese}', "");
Expect(1, 43488, '\P{Block=javanese}', "");
Expect(0, 43488, '\P{^Block=javanese}', "");
Expect(1, 43487, '\p{Block:-JAVANESE}', "");
Expect(0, 43487, '\p{^Block:-JAVANESE}', "");
Expect(0, 43487, '\P{Block:-JAVANESE}', "");
Expect(1, 43487, '\P{^Block:-JAVANESE}', "");
Expect(0, 43488, '\p{Block:-JAVANESE}', "");
Expect(1, 43488, '\p{^Block:-JAVANESE}', "");
Expect(1, 43488, '\P{Block:-JAVANESE}', "");
Expect(0, 43488, '\P{^Block:-JAVANESE}', "");
Error('\p{Blk= :=JAVANESE}');
Error('\P{Blk= :=JAVANESE}');
Expect(1, 43487, '\p{Blk=javanese}', "");
Expect(0, 43487, '\p{^Blk=javanese}', "");
Expect(0, 43487, '\P{Blk=javanese}', "");
Expect(1, 43487, '\P{^Blk=javanese}', "");
Expect(0, 43488, '\p{Blk=javanese}', "");
Expect(1, 43488, '\p{^Blk=javanese}', "");
Expect(1, 43488, '\P{Blk=javanese}', "");
Expect(0, 43488, '\P{^Blk=javanese}', "");
Expect(1, 43487, '\p{Blk=-Javanese}', "");
Expect(0, 43487, '\p{^Blk=-Javanese}', "");
Expect(0, 43487, '\P{Blk=-Javanese}', "");
Expect(1, 43487, '\P{^Blk=-Javanese}', "");
Expect(0, 43488, '\p{Blk=-Javanese}', "");
Expect(1, 43488, '\p{^Blk=-Javanese}', "");
Expect(1, 43488, '\P{Blk=-Javanese}', "");
Expect(0, 43488, '\P{^Blk=-Javanese}', "");
Error('\p{Is_Block=/a/	_javanese}');
Error('\P{Is_Block=/a/	_javanese}');
Expect(1, 43487, '\p{Is_Block:javanese}', "");
Expect(0, 43487, '\p{^Is_Block:javanese}', "");
Expect(0, 43487, '\P{Is_Block:javanese}', "");
Expect(1, 43487, '\P{^Is_Block:javanese}', "");
Expect(0, 43488, '\p{Is_Block:javanese}', "");
Expect(1, 43488, '\p{^Is_Block:javanese}', "");
Expect(1, 43488, '\P{Is_Block:javanese}', "");
Expect(0, 43488, '\P{^Is_Block:javanese}', "");
Expect(1, 43487, '\p{Is_Block: _javanese}', "");
Expect(0, 43487, '\p{^Is_Block: _javanese}', "");
Expect(0, 43487, '\P{Is_Block: _javanese}', "");
Expect(1, 43487, '\P{^Is_Block: _javanese}', "");
Expect(0, 43488, '\p{Is_Block: _javanese}', "");
Expect(1, 43488, '\p{^Is_Block: _javanese}', "");
Expect(1, 43488, '\P{Is_Block: _javanese}', "");
Expect(0, 43488, '\P{^Is_Block: _javanese}', "");
Error('\p{Is_Blk= /a/javanese}');
Error('\P{Is_Blk= /a/javanese}');
Expect(1, 43487, '\p{Is_Blk:	javanese}', "");
Expect(0, 43487, '\p{^Is_Blk:	javanese}', "");
Expect(0, 43487, '\P{Is_Blk:	javanese}', "");
Expect(1, 43487, '\P{^Is_Blk:	javanese}', "");
Expect(0, 43488, '\p{Is_Blk:	javanese}', "");
Expect(1, 43488, '\p{^Is_Blk:	javanese}', "");
Expect(1, 43488, '\P{Is_Blk:	javanese}', "");
Expect(0, 43488, '\P{^Is_Blk:	javanese}', "");
Expect(1, 43487, '\p{Is_Blk=-Javanese}', "");
Expect(0, 43487, '\p{^Is_Blk=-Javanese}', "");
Expect(0, 43487, '\P{Is_Blk=-Javanese}', "");
Expect(1, 43487, '\P{^Is_Blk=-Javanese}', "");
Expect(0, 43488, '\p{Is_Blk=-Javanese}', "");
Expect(1, 43488, '\p{^Is_Blk=-Javanese}', "");
Expect(1, 43488, '\P{Is_Blk=-Javanese}', "");
Expect(0, 43488, '\P{^Is_Blk=-Javanese}', "");
Error('\p{Block=:= KAITHI}');
Error('\P{Block=:= KAITHI}');
Expect(1, 69839, '\p{Block=kaithi}', "");
Expect(0, 69839, '\p{^Block=kaithi}', "");
Expect(0, 69839, '\P{Block=kaithi}', "");
Expect(1, 69839, '\P{^Block=kaithi}', "");
Expect(0, 69840, '\p{Block=kaithi}', "");
Expect(1, 69840, '\p{^Block=kaithi}', "");
Expect(1, 69840, '\P{Block=kaithi}', "");
Expect(0, 69840, '\P{^Block=kaithi}', "");
Expect(1, 69839, '\p{Block=- Kaithi}', "");
Expect(0, 69839, '\p{^Block=- Kaithi}', "");
Expect(0, 69839, '\P{Block=- Kaithi}', "");
Expect(1, 69839, '\P{^Block=- Kaithi}', "");
Expect(0, 69840, '\p{Block=- Kaithi}', "");
Expect(1, 69840, '\p{^Block=- Kaithi}', "");
Expect(1, 69840, '\P{Block=- Kaithi}', "");
Expect(0, 69840, '\P{^Block=- Kaithi}', "");
Error('\p{Blk=:=_	Kaithi}');
Error('\P{Blk=:=_	Kaithi}');
Expect(1, 69839, '\p{Blk=kaithi}', "");
Expect(0, 69839, '\p{^Blk=kaithi}', "");
Expect(0, 69839, '\P{Blk=kaithi}', "");
Expect(1, 69839, '\P{^Blk=kaithi}', "");
Expect(0, 69840, '\p{Blk=kaithi}', "");
Expect(1, 69840, '\p{^Blk=kaithi}', "");
Expect(1, 69840, '\P{Blk=kaithi}', "");
Expect(0, 69840, '\P{^Blk=kaithi}', "");
Expect(1, 69839, '\p{Blk=-_Kaithi}', "");
Expect(0, 69839, '\p{^Blk=-_Kaithi}', "");
Expect(0, 69839, '\P{Blk=-_Kaithi}', "");
Expect(1, 69839, '\P{^Blk=-_Kaithi}', "");
Expect(0, 69840, '\p{Blk=-_Kaithi}', "");
Expect(1, 69840, '\p{^Blk=-_Kaithi}', "");
Expect(1, 69840, '\P{Blk=-_Kaithi}', "");
Expect(0, 69840, '\P{^Blk=-_Kaithi}', "");
Error('\p{Is_Block= Kaithi:=}');
Error('\P{Is_Block= Kaithi:=}');
Expect(1, 69839, '\p{Is_Block=kaithi}', "");
Expect(0, 69839, '\p{^Is_Block=kaithi}', "");
Expect(0, 69839, '\P{Is_Block=kaithi}', "");
Expect(1, 69839, '\P{^Is_Block=kaithi}', "");
Expect(0, 69840, '\p{Is_Block=kaithi}', "");
Expect(1, 69840, '\p{^Is_Block=kaithi}', "");
Expect(1, 69840, '\P{Is_Block=kaithi}', "");
Expect(0, 69840, '\P{^Is_Block=kaithi}', "");
Expect(1, 69839, '\p{Is_Block=-kaithi}', "");
Expect(0, 69839, '\p{^Is_Block=-kaithi}', "");
Expect(0, 69839, '\P{Is_Block=-kaithi}', "");
Expect(1, 69839, '\P{^Is_Block=-kaithi}', "");
Expect(0, 69840, '\p{Is_Block=-kaithi}', "");
Expect(1, 69840, '\p{^Is_Block=-kaithi}', "");
Expect(1, 69840, '\P{Is_Block=-kaithi}', "");
Expect(0, 69840, '\P{^Is_Block=-kaithi}', "");
Error('\p{Is_Blk= :=KAITHI}');
Error('\P{Is_Blk= :=KAITHI}');
Expect(1, 69839, '\p{Is_Blk=kaithi}', "");
Expect(0, 69839, '\p{^Is_Blk=kaithi}', "");
Expect(0, 69839, '\P{Is_Blk=kaithi}', "");
Expect(1, 69839, '\P{^Is_Blk=kaithi}', "");
Expect(0, 69840, '\p{Is_Blk=kaithi}', "");
Expect(1, 69840, '\p{^Is_Blk=kaithi}', "");
Expect(1, 69840, '\P{Is_Blk=kaithi}', "");
Expect(0, 69840, '\P{^Is_Blk=kaithi}', "");
Expect(1, 69839, '\p{Is_Blk=_kaithi}', "");
Expect(0, 69839, '\p{^Is_Blk=_kaithi}', "");
Expect(0, 69839, '\P{Is_Blk=_kaithi}', "");
Expect(1, 69839, '\P{^Is_Blk=_kaithi}', "");
Expect(0, 69840, '\p{Is_Blk=_kaithi}', "");
Expect(1, 69840, '\p{^Is_Blk=_kaithi}', "");
Expect(1, 69840, '\P{Is_Blk=_kaithi}', "");
Expect(0, 69840, '\P{^Is_Blk=_kaithi}', "");
Error('\p{Block:	/a/_-kana_supplement}');
Error('\P{Block:	/a/_-kana_supplement}');
Expect(1, 110847, '\p{Block=kanasupplement}', "");
Expect(0, 110847, '\p{^Block=kanasupplement}', "");
Expect(0, 110847, '\P{Block=kanasupplement}', "");
Expect(1, 110847, '\P{^Block=kanasupplement}', "");
Expect(0, 110848, '\p{Block=kanasupplement}', "");
Expect(1, 110848, '\p{^Block=kanasupplement}', "");
Expect(1, 110848, '\P{Block=kanasupplement}', "");
Expect(0, 110848, '\P{^Block=kanasupplement}', "");
Expect(1, 110847, '\p{Block=-KANA_supplement}', "");
Expect(0, 110847, '\p{^Block=-KANA_supplement}', "");
Expect(0, 110847, '\P{Block=-KANA_supplement}', "");
Expect(1, 110847, '\P{^Block=-KANA_supplement}', "");
Expect(0, 110848, '\p{Block=-KANA_supplement}', "");
Expect(1, 110848, '\p{^Block=-KANA_supplement}', "");
Expect(1, 110848, '\P{Block=-KANA_supplement}', "");
Expect(0, 110848, '\P{^Block=-KANA_supplement}', "");
Error('\p{Blk=/a/	-Kana_Sup}');
Error('\P{Blk=/a/	-Kana_Sup}');
Expect(1, 110847, '\p{Blk=kanasup}', "");
Expect(0, 110847, '\p{^Blk=kanasup}', "");
Expect(0, 110847, '\P{Blk=kanasup}', "");
Expect(1, 110847, '\P{^Blk=kanasup}', "");
Expect(0, 110848, '\p{Blk=kanasup}', "");
Expect(1, 110848, '\p{^Blk=kanasup}', "");
Expect(1, 110848, '\P{Blk=kanasup}', "");
Expect(0, 110848, '\P{^Blk=kanasup}', "");
Expect(1, 110847, '\p{Blk:_Kana_Sup}', "");
Expect(0, 110847, '\p{^Blk:_Kana_Sup}', "");
Expect(0, 110847, '\P{Blk:_Kana_Sup}', "");
Expect(1, 110847, '\P{^Blk:_Kana_Sup}', "");
Expect(0, 110848, '\p{Blk:_Kana_Sup}', "");
Expect(1, 110848, '\p{^Blk:_Kana_Sup}', "");
Expect(1, 110848, '\P{Blk:_Kana_Sup}', "");
Expect(0, 110848, '\P{^Blk:_Kana_Sup}', "");
Error('\p{Is_Block=:=__KANA_SUPPLEMENT}');
Error('\P{Is_Block=:=__KANA_SUPPLEMENT}');
Expect(1, 110847, '\p{Is_Block=kanasupplement}', "");
Expect(0, 110847, '\p{^Is_Block=kanasupplement}', "");
Expect(0, 110847, '\P{Is_Block=kanasupplement}', "");
Expect(1, 110847, '\P{^Is_Block=kanasupplement}', "");
Expect(0, 110848, '\p{Is_Block=kanasupplement}', "");
Expect(1, 110848, '\p{^Is_Block=kanasupplement}', "");
Expect(1, 110848, '\P{Is_Block=kanasupplement}', "");
Expect(0, 110848, '\P{^Is_Block=kanasupplement}', "");
Expect(1, 110847, '\p{Is_Block:_Kana_SUPPLEMENT}', "");
Expect(0, 110847, '\p{^Is_Block:_Kana_SUPPLEMENT}', "");
Expect(0, 110847, '\P{Is_Block:_Kana_SUPPLEMENT}', "");
Expect(1, 110847, '\P{^Is_Block:_Kana_SUPPLEMENT}', "");
Expect(0, 110848, '\p{Is_Block:_Kana_SUPPLEMENT}', "");
Expect(1, 110848, '\p{^Is_Block:_Kana_SUPPLEMENT}', "");
Expect(1, 110848, '\P{Is_Block:_Kana_SUPPLEMENT}', "");
Expect(0, 110848, '\P{^Is_Block:_Kana_SUPPLEMENT}', "");
Error('\p{Is_Blk=/a/	kana_Sup}');
Error('\P{Is_Blk=/a/	kana_Sup}');
Expect(1, 110847, '\p{Is_Blk=kanasup}', "");
Expect(0, 110847, '\p{^Is_Blk=kanasup}', "");
Expect(0, 110847, '\P{Is_Blk=kanasup}', "");
Expect(1, 110847, '\P{^Is_Blk=kanasup}', "");
Expect(0, 110848, '\p{Is_Blk=kanasup}', "");
Expect(1, 110848, '\p{^Is_Blk=kanasup}', "");
Expect(1, 110848, '\P{Is_Blk=kanasup}', "");
Expect(0, 110848, '\P{^Is_Blk=kanasup}', "");
Expect(1, 110847, '\p{Is_Blk= _Kana_Sup}', "");
Expect(0, 110847, '\p{^Is_Blk= _Kana_Sup}', "");
Expect(0, 110847, '\P{Is_Blk= _Kana_Sup}', "");
Expect(1, 110847, '\P{^Is_Blk= _Kana_Sup}', "");
Expect(0, 110848, '\p{Is_Blk= _Kana_Sup}', "");
Expect(1, 110848, '\p{^Is_Blk= _Kana_Sup}', "");
Expect(1, 110848, '\P{Is_Blk= _Kana_Sup}', "");
Expect(0, 110848, '\P{^Is_Blk= _Kana_Sup}', "");
Error('\p{Block:		/a/KANBUN}');
Error('\P{Block:		/a/KANBUN}');
Expect(1, 12703, '\p{Block=kanbun}', "");
Expect(0, 12703, '\p{^Block=kanbun}', "");
Expect(0, 12703, '\P{Block=kanbun}', "");
Expect(1, 12703, '\P{^Block=kanbun}', "");
Expect(0, 12704, '\p{Block=kanbun}', "");
Expect(1, 12704, '\p{^Block=kanbun}', "");
Expect(1, 12704, '\P{Block=kanbun}', "");
Expect(0, 12704, '\P{^Block=kanbun}', "");
Expect(1, 12703, '\p{Block=__KANBUN}', "");
Expect(0, 12703, '\p{^Block=__KANBUN}', "");
Expect(0, 12703, '\P{Block=__KANBUN}', "");
Expect(1, 12703, '\P{^Block=__KANBUN}', "");
Expect(0, 12704, '\p{Block=__KANBUN}', "");
Expect(1, 12704, '\p{^Block=__KANBUN}', "");
Expect(1, 12704, '\P{Block=__KANBUN}', "");
Expect(0, 12704, '\P{^Block=__KANBUN}', "");
Error('\p{Blk: /a/	 KANBUN}');
Error('\P{Blk: /a/	 KANBUN}');
Expect(1, 12703, '\p{Blk=kanbun}', "");
Expect(0, 12703, '\p{^Blk=kanbun}', "");
Expect(0, 12703, '\P{Blk=kanbun}', "");
Expect(1, 12703, '\P{^Blk=kanbun}', "");
Expect(0, 12704, '\p{Blk=kanbun}', "");
Expect(1, 12704, '\p{^Blk=kanbun}', "");
Expect(1, 12704, '\P{Blk=kanbun}', "");
Expect(0, 12704, '\P{^Blk=kanbun}', "");
Expect(1, 12703, '\p{Blk=-	KANBUN}', "");
Expect(0, 12703, '\p{^Blk=-	KANBUN}', "");
Expect(0, 12703, '\P{Blk=-	KANBUN}', "");
Expect(1, 12703, '\P{^Blk=-	KANBUN}', "");
Expect(0, 12704, '\p{Blk=-	KANBUN}', "");
Expect(1, 12704, '\p{^Blk=-	KANBUN}', "");
Expect(1, 12704, '\P{Blk=-	KANBUN}', "");
Expect(0, 12704, '\P{^Blk=-	KANBUN}', "");
Error('\p{Is_Block=:=-_kanbun}');
Error('\P{Is_Block=:=-_kanbun}');
Expect(1, 12703, '\p{Is_Block=kanbun}', "");
Expect(0, 12703, '\p{^Is_Block=kanbun}', "");
Expect(0, 12703, '\P{Is_Block=kanbun}', "");
Expect(1, 12703, '\P{^Is_Block=kanbun}', "");
Expect(0, 12704, '\p{Is_Block=kanbun}', "");
Expect(1, 12704, '\p{^Is_Block=kanbun}', "");
Expect(1, 12704, '\P{Is_Block=kanbun}', "");
Expect(0, 12704, '\P{^Is_Block=kanbun}', "");
Expect(1, 12703, '\p{Is_Block: -	Kanbun}', "");
Expect(0, 12703, '\p{^Is_Block: -	Kanbun}', "");
Expect(0, 12703, '\P{Is_Block: -	Kanbun}', "");
Expect(1, 12703, '\P{^Is_Block: -	Kanbun}', "");
Expect(0, 12704, '\p{Is_Block: -	Kanbun}', "");
Expect(1, 12704, '\p{^Is_Block: -	Kanbun}', "");
Expect(1, 12704, '\P{Is_Block: -	Kanbun}', "");
Expect(0, 12704, '\P{^Is_Block: -	Kanbun}', "");
Error('\p{Is_Blk=:=-_KANBUN}');
Error('\P{Is_Blk=:=-_KANBUN}');
Expect(1, 12703, '\p{Is_Blk=kanbun}', "");
Expect(0, 12703, '\p{^Is_Blk=kanbun}', "");
Expect(0, 12703, '\P{Is_Blk=kanbun}', "");
Expect(1, 12703, '\P{^Is_Blk=kanbun}', "");
Expect(0, 12704, '\p{Is_Blk=kanbun}', "");
Expect(1, 12704, '\p{^Is_Blk=kanbun}', "");
Expect(1, 12704, '\P{Is_Blk=kanbun}', "");
Expect(0, 12704, '\P{^Is_Blk=kanbun}', "");
Expect(1, 12703, '\p{Is_Blk=	kanbun}', "");
Expect(0, 12703, '\p{^Is_Blk=	kanbun}', "");
Expect(0, 12703, '\P{Is_Blk=	kanbun}', "");
Expect(1, 12703, '\P{^Is_Blk=	kanbun}', "");
Expect(0, 12704, '\p{Is_Blk=	kanbun}', "");
Expect(1, 12704, '\p{^Is_Blk=	kanbun}', "");
Expect(1, 12704, '\P{Is_Blk=	kanbun}', "");
Expect(0, 12704, '\P{^Is_Blk=	kanbun}', "");
Error('\p{Block=/a/- KANGXI_Radicals}');
Error('\P{Block=/a/- KANGXI_Radicals}');
Expect(1, 12255, '\p{Block=kangxiradicals}', "");
Expect(0, 12255, '\p{^Block=kangxiradicals}', "");
Expect(0, 12255, '\P{Block=kangxiradicals}', "");
Expect(1, 12255, '\P{^Block=kangxiradicals}', "");
Expect(0, 12256, '\p{Block=kangxiradicals}', "");
Expect(1, 12256, '\p{^Block=kangxiradicals}', "");
Expect(1, 12256, '\P{Block=kangxiradicals}', "");
Expect(0, 12256, '\P{^Block=kangxiradicals}', "");
Expect(1, 12255, '\p{Block=_	kangxi_Radicals}', "");
Expect(0, 12255, '\p{^Block=_	kangxi_Radicals}', "");
Expect(0, 12255, '\P{Block=_	kangxi_Radicals}', "");
Expect(1, 12255, '\P{^Block=_	kangxi_Radicals}', "");
Expect(0, 12256, '\p{Block=_	kangxi_Radicals}', "");
Expect(1, 12256, '\p{^Block=_	kangxi_Radicals}', "");
Expect(1, 12256, '\P{Block=_	kangxi_Radicals}', "");
Expect(0, 12256, '\P{^Block=_	kangxi_Radicals}', "");
Error('\p{Blk=:=Kangxi}');
Error('\P{Blk=:=Kangxi}');
Expect(1, 12255, '\p{Blk=kangxi}', "");
Expect(0, 12255, '\p{^Blk=kangxi}', "");
Expect(0, 12255, '\P{Blk=kangxi}', "");
Expect(1, 12255, '\P{^Blk=kangxi}', "");
Expect(0, 12256, '\p{Blk=kangxi}', "");
Expect(1, 12256, '\p{^Blk=kangxi}', "");
Expect(1, 12256, '\P{Blk=kangxi}', "");
Expect(0, 12256, '\P{^Blk=kangxi}', "");
Expect(1, 12255, '\p{Blk=	Kangxi}', "");
Expect(0, 12255, '\p{^Blk=	Kangxi}', "");
Expect(0, 12255, '\P{Blk=	Kangxi}', "");
Expect(1, 12255, '\P{^Blk=	Kangxi}', "");
Expect(0, 12256, '\p{Blk=	Kangxi}', "");
Expect(1, 12256, '\p{^Blk=	Kangxi}', "");
Expect(1, 12256, '\P{Blk=	Kangxi}', "");
Expect(0, 12256, '\P{^Blk=	Kangxi}', "");
Error('\p{Is_Block=_/a/KANGXI_Radicals}');
Error('\P{Is_Block=_/a/KANGXI_Radicals}');
Expect(1, 12255, '\p{Is_Block=kangxiradicals}', "");
Expect(0, 12255, '\p{^Is_Block=kangxiradicals}', "");
Expect(0, 12255, '\P{Is_Block=kangxiradicals}', "");
Expect(1, 12255, '\P{^Is_Block=kangxiradicals}', "");
Expect(0, 12256, '\p{Is_Block=kangxiradicals}', "");
Expect(1, 12256, '\p{^Is_Block=kangxiradicals}', "");
Expect(1, 12256, '\P{Is_Block=kangxiradicals}', "");
Expect(0, 12256, '\P{^Is_Block=kangxiradicals}', "");
Expect(1, 12255, '\p{Is_Block=--Kangxi_radicals}', "");
Expect(0, 12255, '\p{^Is_Block=--Kangxi_radicals}', "");
Expect(0, 12255, '\P{Is_Block=--Kangxi_radicals}', "");
Expect(1, 12255, '\P{^Is_Block=--Kangxi_radicals}', "");
Expect(0, 12256, '\p{Is_Block=--Kangxi_radicals}', "");
Expect(1, 12256, '\p{^Is_Block=--Kangxi_radicals}', "");
Expect(1, 12256, '\P{Is_Block=--Kangxi_radicals}', "");
Expect(0, 12256, '\P{^Is_Block=--Kangxi_radicals}', "");
Error('\p{Is_Blk=  kangxi:=}');
Error('\P{Is_Blk=  kangxi:=}');
Expect(1, 12255, '\p{Is_Blk=kangxi}', "");
Expect(0, 12255, '\p{^Is_Blk=kangxi}', "");
Expect(0, 12255, '\P{Is_Blk=kangxi}', "");
Expect(1, 12255, '\P{^Is_Blk=kangxi}', "");
Expect(0, 12256, '\p{Is_Blk=kangxi}', "");
Expect(1, 12256, '\p{^Is_Blk=kangxi}', "");
Expect(1, 12256, '\P{Is_Blk=kangxi}', "");
Expect(0, 12256, '\P{^Is_Blk=kangxi}', "");
Expect(1, 12255, '\p{Is_Blk=	kangxi}', "");
Expect(0, 12255, '\p{^Is_Blk=	kangxi}', "");
Expect(0, 12255, '\P{Is_Blk=	kangxi}', "");
Expect(1, 12255, '\P{^Is_Blk=	kangxi}', "");
Expect(0, 12256, '\p{Is_Blk=	kangxi}', "");
Expect(1, 12256, '\p{^Is_Blk=	kangxi}', "");
Expect(1, 12256, '\P{Is_Blk=	kangxi}', "");
Expect(0, 12256, '\P{^Is_Blk=	kangxi}', "");
Error('\p{Block=_	Kannada/a/}');
Error('\P{Block=_	Kannada/a/}');
Expect(1, 3327, '\p{Block=kannada}', "");
Expect(0, 3327, '\p{^Block=kannada}', "");
Expect(0, 3327, '\P{Block=kannada}', "");
Expect(1, 3327, '\P{^Block=kannada}', "");
Expect(0, 3328, '\p{Block=kannada}', "");
Expect(1, 3328, '\p{^Block=kannada}', "");
Expect(1, 3328, '\P{Block=kannada}', "");
Expect(0, 3328, '\P{^Block=kannada}', "");
Expect(1, 3327, '\p{Block=__kannada}', "");
Expect(0, 3327, '\p{^Block=__kannada}', "");
Expect(0, 3327, '\P{Block=__kannada}', "");
Expect(1, 3327, '\P{^Block=__kannada}', "");
Expect(0, 3328, '\p{Block=__kannada}', "");
Expect(1, 3328, '\p{^Block=__kannada}', "");
Expect(1, 3328, '\P{Block=__kannada}', "");
Expect(0, 3328, '\P{^Block=__kannada}', "");
Error('\p{Blk= -kannada/a/}');
Error('\P{Blk= -kannada/a/}');
Expect(1, 3327, '\p{Blk=kannada}', "");
Expect(0, 3327, '\p{^Blk=kannada}', "");
Expect(0, 3327, '\P{Blk=kannada}', "");
Expect(1, 3327, '\P{^Blk=kannada}', "");
Expect(0, 3328, '\p{Blk=kannada}', "");
Expect(1, 3328, '\p{^Blk=kannada}', "");
Expect(1, 3328, '\P{Blk=kannada}', "");
Expect(0, 3328, '\P{^Blk=kannada}', "");
Expect(1, 3327, '\p{Blk: -KANNADA}', "");
Expect(0, 3327, '\p{^Blk: -KANNADA}', "");
Expect(0, 3327, '\P{Blk: -KANNADA}', "");
Expect(1, 3327, '\P{^Blk: -KANNADA}', "");
Expect(0, 3328, '\p{Blk: -KANNADA}', "");
Expect(1, 3328, '\p{^Blk: -KANNADA}', "");
Expect(1, 3328, '\P{Blk: -KANNADA}', "");
Expect(0, 3328, '\P{^Blk: -KANNADA}', "");
Error('\p{Is_Block=	-Kannada/a/}');
Error('\P{Is_Block=	-Kannada/a/}');
Expect(1, 3327, '\p{Is_Block=kannada}', "");
Expect(0, 3327, '\p{^Is_Block=kannada}', "");
Expect(0, 3327, '\P{Is_Block=kannada}', "");
Expect(1, 3327, '\P{^Is_Block=kannada}', "");
Expect(0, 3328, '\p{Is_Block=kannada}', "");
Expect(1, 3328, '\p{^Is_Block=kannada}', "");
Expect(1, 3328, '\P{Is_Block=kannada}', "");
Expect(0, 3328, '\P{^Is_Block=kannada}', "");
Expect(1, 3327, '\p{Is_Block=		Kannada}', "");
Expect(0, 3327, '\p{^Is_Block=		Kannada}', "");
Expect(0, 3327, '\P{Is_Block=		Kannada}', "");
Expect(1, 3327, '\P{^Is_Block=		Kannada}', "");
Expect(0, 3328, '\p{Is_Block=		Kannada}', "");
Expect(1, 3328, '\p{^Is_Block=		Kannada}', "");
Expect(1, 3328, '\P{Is_Block=		Kannada}', "");
Expect(0, 3328, '\P{^Is_Block=		Kannada}', "");
Error('\p{Is_Blk=/a/ -kannada}');
Error('\P{Is_Blk=/a/ -kannada}');
Expect(1, 3327, '\p{Is_Blk=kannada}', "");
Expect(0, 3327, '\p{^Is_Blk=kannada}', "");
Expect(0, 3327, '\P{Is_Blk=kannada}', "");
Expect(1, 3327, '\P{^Is_Blk=kannada}', "");
Expect(0, 3328, '\p{Is_Blk=kannada}', "");
Expect(1, 3328, '\p{^Is_Blk=kannada}', "");
Expect(1, 3328, '\P{Is_Blk=kannada}', "");
Expect(0, 3328, '\P{^Is_Blk=kannada}', "");
Expect(1, 3327, '\p{Is_Blk= kannada}', "");
Expect(0, 3327, '\p{^Is_Blk= kannada}', "");
Expect(0, 3327, '\P{Is_Blk= kannada}', "");
Expect(1, 3327, '\P{^Is_Blk= kannada}', "");
Expect(0, 3328, '\p{Is_Blk= kannada}', "");
Expect(1, 3328, '\p{^Is_Blk= kannada}', "");
Expect(1, 3328, '\P{Is_Blk= kannada}', "");
Expect(0, 3328, '\P{^Is_Blk= kannada}', "");
Error('\p{Block= :=KATAKANA}');
Error('\P{Block= :=KATAKANA}');
Expect(1, 12543, '\p{Block=katakana}', "");
Expect(0, 12543, '\p{^Block=katakana}', "");
Expect(0, 12543, '\P{Block=katakana}', "");
Expect(1, 12543, '\P{^Block=katakana}', "");
Expect(0, 12544, '\p{Block=katakana}', "");
Expect(1, 12544, '\p{^Block=katakana}', "");
Expect(1, 12544, '\P{Block=katakana}', "");
Expect(0, 12544, '\P{^Block=katakana}', "");
Expect(1, 12543, '\p{Block=-Katakana}', "");
Expect(0, 12543, '\p{^Block=-Katakana}', "");
Expect(0, 12543, '\P{Block=-Katakana}', "");
Expect(1, 12543, '\P{^Block=-Katakana}', "");
Expect(0, 12544, '\p{Block=-Katakana}', "");
Expect(1, 12544, '\p{^Block=-Katakana}', "");
Expect(1, 12544, '\P{Block=-Katakana}', "");
Expect(0, 12544, '\P{^Block=-Katakana}', "");
Error('\p{Blk=/a/-	Katakana}');
Error('\P{Blk=/a/-	Katakana}');
Expect(1, 12543, '\p{Blk=katakana}', "");
Expect(0, 12543, '\p{^Blk=katakana}', "");
Expect(0, 12543, '\P{Blk=katakana}', "");
Expect(1, 12543, '\P{^Blk=katakana}', "");
Expect(0, 12544, '\p{Blk=katakana}', "");
Expect(1, 12544, '\p{^Blk=katakana}', "");
Expect(1, 12544, '\P{Blk=katakana}', "");
Expect(0, 12544, '\P{^Blk=katakana}', "");
Expect(1, 12543, '\p{Blk= 	Katakana}', "");
Expect(0, 12543, '\p{^Blk= 	Katakana}', "");
Expect(0, 12543, '\P{Blk= 	Katakana}', "");
Expect(1, 12543, '\P{^Blk= 	Katakana}', "");
Expect(0, 12544, '\p{Blk= 	Katakana}', "");
Expect(1, 12544, '\p{^Blk= 	Katakana}', "");
Expect(1, 12544, '\P{Blk= 	Katakana}', "");
Expect(0, 12544, '\P{^Blk= 	Katakana}', "");
Error('\p{Is_Block=:=katakana}');
Error('\P{Is_Block=:=katakana}');
Expect(1, 12543, '\p{Is_Block=katakana}', "");
Expect(0, 12543, '\p{^Is_Block=katakana}', "");
Expect(0, 12543, '\P{Is_Block=katakana}', "");
Expect(1, 12543, '\P{^Is_Block=katakana}', "");
Expect(0, 12544, '\p{Is_Block=katakana}', "");
Expect(1, 12544, '\p{^Is_Block=katakana}', "");
Expect(1, 12544, '\P{Is_Block=katakana}', "");
Expect(0, 12544, '\P{^Is_Block=katakana}', "");
Expect(1, 12543, '\p{Is_Block=	-KATAKANA}', "");
Expect(0, 12543, '\p{^Is_Block=	-KATAKANA}', "");
Expect(0, 12543, '\P{Is_Block=	-KATAKANA}', "");
Expect(1, 12543, '\P{^Is_Block=	-KATAKANA}', "");
Expect(0, 12544, '\p{Is_Block=	-KATAKANA}', "");
Expect(1, 12544, '\p{^Is_Block=	-KATAKANA}', "");
Expect(1, 12544, '\P{Is_Block=	-KATAKANA}', "");
Expect(0, 12544, '\P{^Is_Block=	-KATAKANA}', "");
Error('\p{Is_Blk= _katakana:=}');
Error('\P{Is_Blk= _katakana:=}');
Expect(1, 12543, '\p{Is_Blk=katakana}', "");
Expect(0, 12543, '\p{^Is_Blk=katakana}', "");
Expect(0, 12543, '\P{Is_Blk=katakana}', "");
Expect(1, 12543, '\P{^Is_Blk=katakana}', "");
Expect(0, 12544, '\p{Is_Blk=katakana}', "");
Expect(1, 12544, '\p{^Is_Blk=katakana}', "");
Expect(1, 12544, '\P{Is_Blk=katakana}', "");
Expect(0, 12544, '\P{^Is_Blk=katakana}', "");
Expect(1, 12543, '\p{Is_Blk=	Katakana}', "");
Expect(0, 12543, '\p{^Is_Blk=	Katakana}', "");
Expect(0, 12543, '\P{Is_Blk=	Katakana}', "");
Expect(1, 12543, '\P{^Is_Blk=	Katakana}', "");
Expect(0, 12544, '\p{Is_Blk=	Katakana}', "");
Expect(1, 12544, '\p{^Is_Blk=	Katakana}', "");
Expect(1, 12544, '\P{Is_Blk=	Katakana}', "");
Expect(0, 12544, '\P{^Is_Blk=	Katakana}', "");
Error('\p{Block=:=  KATAKANA_Phonetic_Extensions}');
Error('\P{Block=:=  KATAKANA_Phonetic_Extensions}');
Expect(1, 12799, '\p{Block=katakanaphoneticextensions}', "");
Expect(0, 12799, '\p{^Block=katakanaphoneticextensions}', "");
Expect(0, 12799, '\P{Block=katakanaphoneticextensions}', "");
Expect(1, 12799, '\P{^Block=katakanaphoneticextensions}', "");
Expect(0, 12800, '\p{Block=katakanaphoneticextensions}', "");
Expect(1, 12800, '\p{^Block=katakanaphoneticextensions}', "");
Expect(1, 12800, '\P{Block=katakanaphoneticextensions}', "");
Expect(0, 12800, '\P{^Block=katakanaphoneticextensions}', "");
Expect(1, 12799, '\p{Block: Katakana_PHONETIC_EXTENSIONS}', "");
Expect(0, 12799, '\p{^Block: Katakana_PHONETIC_EXTENSIONS}', "");
Expect(0, 12799, '\P{Block: Katakana_PHONETIC_EXTENSIONS}', "");
Expect(1, 12799, '\P{^Block: Katakana_PHONETIC_EXTENSIONS}', "");
Expect(0, 12800, '\p{Block: Katakana_PHONETIC_EXTENSIONS}', "");
Expect(1, 12800, '\p{^Block: Katakana_PHONETIC_EXTENSIONS}', "");
Expect(1, 12800, '\P{Block: Katakana_PHONETIC_EXTENSIONS}', "");
Expect(0, 12800, '\P{^Block: Katakana_PHONETIC_EXTENSIONS}', "");
Error('\p{Blk=-:=Katakana_ext}');
Error('\P{Blk=-:=Katakana_ext}');
Expect(1, 12799, '\p{Blk=katakanaext}', "");
Expect(0, 12799, '\p{^Blk=katakanaext}', "");
Expect(0, 12799, '\P{Blk=katakanaext}', "");
Expect(1, 12799, '\P{^Blk=katakanaext}', "");
Expect(0, 12800, '\p{Blk=katakanaext}', "");
Expect(1, 12800, '\p{^Blk=katakanaext}', "");
Expect(1, 12800, '\P{Blk=katakanaext}', "");
Expect(0, 12800, '\P{^Blk=katakanaext}', "");
Expect(1, 12799, '\p{Blk:  -KATAKANA_Ext}', "");
Expect(0, 12799, '\p{^Blk:  -KATAKANA_Ext}', "");
Expect(0, 12799, '\P{Blk:  -KATAKANA_Ext}', "");
Expect(1, 12799, '\P{^Blk:  -KATAKANA_Ext}', "");
Expect(0, 12800, '\p{Blk:  -KATAKANA_Ext}', "");
Expect(1, 12800, '\p{^Blk:  -KATAKANA_Ext}', "");
Expect(1, 12800, '\P{Blk:  -KATAKANA_Ext}', "");
Expect(0, 12800, '\P{^Blk:  -KATAKANA_Ext}', "");
Error('\p{Is_Block=_:=katakana_phonetic_Extensions}');
Error('\P{Is_Block=_:=katakana_phonetic_Extensions}');
Expect(1, 12799, '\p{Is_Block=katakanaphoneticextensions}', "");
Expect(0, 12799, '\p{^Is_Block=katakanaphoneticextensions}', "");
Expect(0, 12799, '\P{Is_Block=katakanaphoneticextensions}', "");
Expect(1, 12799, '\P{^Is_Block=katakanaphoneticextensions}', "");
Expect(0, 12800, '\p{Is_Block=katakanaphoneticextensions}', "");
Expect(1, 12800, '\p{^Is_Block=katakanaphoneticextensions}', "");
Expect(1, 12800, '\P{Is_Block=katakanaphoneticextensions}', "");
Expect(0, 12800, '\P{^Is_Block=katakanaphoneticextensions}', "");
Expect(1, 12799, '\p{Is_Block=	Katakana_PHONETIC_extensions}', "");
Expect(0, 12799, '\p{^Is_Block=	Katakana_PHONETIC_extensions}', "");
Expect(0, 12799, '\P{Is_Block=	Katakana_PHONETIC_extensions}', "");
Expect(1, 12799, '\P{^Is_Block=	Katakana_PHONETIC_extensions}', "");
Expect(0, 12800, '\p{Is_Block=	Katakana_PHONETIC_extensions}', "");
Expect(1, 12800, '\p{^Is_Block=	Katakana_PHONETIC_extensions}', "");
Expect(1, 12800, '\P{Is_Block=	Katakana_PHONETIC_extensions}', "");
Expect(0, 12800, '\P{^Is_Block=	Katakana_PHONETIC_extensions}', "");
Error('\p{Is_Blk=_	katakana_Ext/a/}');
Error('\P{Is_Blk=_	katakana_Ext/a/}');
Expect(1, 12799, '\p{Is_Blk=katakanaext}', "");
Expect(0, 12799, '\p{^Is_Blk=katakanaext}', "");
Expect(0, 12799, '\P{Is_Blk=katakanaext}', "");
Expect(1, 12799, '\P{^Is_Blk=katakanaext}', "");
Expect(0, 12800, '\p{Is_Blk=katakanaext}', "");
Expect(1, 12800, '\p{^Is_Blk=katakanaext}', "");
Expect(1, 12800, '\P{Is_Blk=katakanaext}', "");
Expect(0, 12800, '\P{^Is_Blk=katakanaext}', "");
Expect(1, 12799, '\p{Is_Blk=-katakana_EXT}', "");
Expect(0, 12799, '\p{^Is_Blk=-katakana_EXT}', "");
Expect(0, 12799, '\P{Is_Blk=-katakana_EXT}', "");
Expect(1, 12799, '\P{^Is_Blk=-katakana_EXT}', "");
Expect(0, 12800, '\p{Is_Blk=-katakana_EXT}', "");
Expect(1, 12800, '\p{^Is_Blk=-katakana_EXT}', "");
Expect(1, 12800, '\P{Is_Blk=-katakana_EXT}', "");
Expect(0, 12800, '\P{^Is_Blk=-katakana_EXT}', "");
Error('\p{Block=/a/ KAYAH_Li}');
Error('\P{Block=/a/ KAYAH_Li}');
Expect(1, 43311, '\p{Block=kayahli}', "");
Expect(0, 43311, '\p{^Block=kayahli}', "");
Expect(0, 43311, '\P{Block=kayahli}', "");
Expect(1, 43311, '\P{^Block=kayahli}', "");
Expect(0, 43312, '\p{Block=kayahli}', "");
Expect(1, 43312, '\p{^Block=kayahli}', "");
Expect(1, 43312, '\P{Block=kayahli}', "");
Expect(0, 43312, '\P{^Block=kayahli}', "");
Expect(1, 43311, '\p{Block=	kayah_li}', "");
Expect(0, 43311, '\p{^Block=	kayah_li}', "");
Expect(0, 43311, '\P{Block=	kayah_li}', "");
Expect(1, 43311, '\P{^Block=	kayah_li}', "");
Expect(0, 43312, '\p{Block=	kayah_li}', "");
Expect(1, 43312, '\p{^Block=	kayah_li}', "");
Expect(1, 43312, '\P{Block=	kayah_li}', "");
Expect(0, 43312, '\P{^Block=	kayah_li}', "");
Error('\p{Blk=Kayah_li/a/}');
Error('\P{Blk=Kayah_li/a/}');
Expect(1, 43311, '\p{Blk=kayahli}', "");
Expect(0, 43311, '\p{^Blk=kayahli}', "");
Expect(0, 43311, '\P{Blk=kayahli}', "");
Expect(1, 43311, '\P{^Blk=kayahli}', "");
Expect(0, 43312, '\p{Blk=kayahli}', "");
Expect(1, 43312, '\p{^Blk=kayahli}', "");
Expect(1, 43312, '\P{Blk=kayahli}', "");
Expect(0, 43312, '\P{^Blk=kayahli}', "");
Expect(1, 43311, '\p{Blk=_	Kayah_Li}', "");
Expect(0, 43311, '\p{^Blk=_	Kayah_Li}', "");
Expect(0, 43311, '\P{Blk=_	Kayah_Li}', "");
Expect(1, 43311, '\P{^Blk=_	Kayah_Li}', "");
Expect(0, 43312, '\p{Blk=_	Kayah_Li}', "");
Expect(1, 43312, '\p{^Blk=_	Kayah_Li}', "");
Expect(1, 43312, '\P{Blk=_	Kayah_Li}', "");
Expect(0, 43312, '\P{^Blk=_	Kayah_Li}', "");
Error('\p{Is_Block=-:=Kayah_Li}');
Error('\P{Is_Block=-:=Kayah_Li}');
Expect(1, 43311, '\p{Is_Block=kayahli}', "");
Expect(0, 43311, '\p{^Is_Block=kayahli}', "");
Expect(0, 43311, '\P{Is_Block=kayahli}', "");
Expect(1, 43311, '\P{^Is_Block=kayahli}', "");
Expect(0, 43312, '\p{Is_Block=kayahli}', "");
Expect(1, 43312, '\p{^Is_Block=kayahli}', "");
Expect(1, 43312, '\P{Is_Block=kayahli}', "");
Expect(0, 43312, '\P{^Is_Block=kayahli}', "");
Expect(1, 43311, '\p{Is_Block=- kayah_Li}', "");
Expect(0, 43311, '\p{^Is_Block=- kayah_Li}', "");
Expect(0, 43311, '\P{Is_Block=- kayah_Li}', "");
Expect(1, 43311, '\P{^Is_Block=- kayah_Li}', "");
Expect(0, 43312, '\p{Is_Block=- kayah_Li}', "");
Expect(1, 43312, '\p{^Is_Block=- kayah_Li}', "");
Expect(1, 43312, '\P{Is_Block=- kayah_Li}', "");
Expect(0, 43312, '\P{^Is_Block=- kayah_Li}', "");
Error('\p{Is_Blk: 	:=kayah_Li}');
Error('\P{Is_Blk: 	:=kayah_Li}');
Expect(1, 43311, '\p{Is_Blk: kayahli}', "");
Expect(0, 43311, '\p{^Is_Blk: kayahli}', "");
Expect(0, 43311, '\P{Is_Blk: kayahli}', "");
Expect(1, 43311, '\P{^Is_Blk: kayahli}', "");
Expect(0, 43312, '\p{Is_Blk: kayahli}', "");
Expect(1, 43312, '\p{^Is_Blk: kayahli}', "");
Expect(1, 43312, '\P{Is_Blk: kayahli}', "");
Expect(0, 43312, '\P{^Is_Blk: kayahli}', "");
Expect(1, 43311, '\p{Is_Blk=-_KAYAH_Li}', "");
Expect(0, 43311, '\p{^Is_Blk=-_KAYAH_Li}', "");
Expect(0, 43311, '\P{Is_Blk=-_KAYAH_Li}', "");
Expect(1, 43311, '\P{^Is_Blk=-_KAYAH_Li}', "");
Expect(0, 43312, '\p{Is_Blk=-_KAYAH_Li}', "");
Expect(1, 43312, '\p{^Is_Blk=-_KAYAH_Li}', "");
Expect(1, 43312, '\P{Is_Blk=-_KAYAH_Li}', "");
Expect(0, 43312, '\P{^Is_Blk=-_KAYAH_Li}', "");
Error('\p{Block= kharoshthi/a/}');
Error('\P{Block= kharoshthi/a/}');
Expect(1, 68191, '\p{Block:	kharoshthi}', "");
Expect(0, 68191, '\p{^Block:	kharoshthi}', "");
Expect(0, 68191, '\P{Block:	kharoshthi}', "");
Expect(1, 68191, '\P{^Block:	kharoshthi}', "");
Expect(0, 68192, '\p{Block:	kharoshthi}', "");
Expect(1, 68192, '\p{^Block:	kharoshthi}', "");
Expect(1, 68192, '\P{Block:	kharoshthi}', "");
Expect(0, 68192, '\P{^Block:	kharoshthi}', "");
Expect(1, 68191, '\p{Block=__Kharoshthi}', "");
Expect(0, 68191, '\p{^Block=__Kharoshthi}', "");
Expect(0, 68191, '\P{Block=__Kharoshthi}', "");
Expect(1, 68191, '\P{^Block=__Kharoshthi}', "");
Expect(0, 68192, '\p{Block=__Kharoshthi}', "");
Expect(1, 68192, '\p{^Block=__Kharoshthi}', "");
Expect(1, 68192, '\P{Block=__Kharoshthi}', "");
Expect(0, 68192, '\P{^Block=__Kharoshthi}', "");
Error('\p{Blk= Kharoshthi:=}');
Error('\P{Blk= Kharoshthi:=}');
Expect(1, 68191, '\p{Blk=kharoshthi}', "");
Expect(0, 68191, '\p{^Blk=kharoshthi}', "");
Expect(0, 68191, '\P{Blk=kharoshthi}', "");
Expect(1, 68191, '\P{^Blk=kharoshthi}', "");
Expect(0, 68192, '\p{Blk=kharoshthi}', "");
Expect(1, 68192, '\p{^Blk=kharoshthi}', "");
Expect(1, 68192, '\P{Blk=kharoshthi}', "");
Expect(0, 68192, '\P{^Blk=kharoshthi}', "");
Expect(1, 68191, '\p{Blk=		Kharoshthi}', "");
Expect(0, 68191, '\p{^Blk=		Kharoshthi}', "");
Expect(0, 68191, '\P{Blk=		Kharoshthi}', "");
Expect(1, 68191, '\P{^Blk=		Kharoshthi}', "");
Expect(0, 68192, '\p{Blk=		Kharoshthi}', "");
Expect(1, 68192, '\p{^Blk=		Kharoshthi}', "");
Expect(1, 68192, '\P{Blk=		Kharoshthi}', "");
Expect(0, 68192, '\P{^Blk=		Kharoshthi}', "");
Error('\p{Is_Block=-/a/KHAROSHTHI}');
Error('\P{Is_Block=-/a/KHAROSHTHI}');
Expect(1, 68191, '\p{Is_Block=kharoshthi}', "");
Expect(0, 68191, '\p{^Is_Block=kharoshthi}', "");
Expect(0, 68191, '\P{Is_Block=kharoshthi}', "");
Expect(1, 68191, '\P{^Is_Block=kharoshthi}', "");
Expect(0, 68192, '\p{Is_Block=kharoshthi}', "");
Expect(1, 68192, '\p{^Is_Block=kharoshthi}', "");
Expect(1, 68192, '\P{Is_Block=kharoshthi}', "");
Expect(0, 68192, '\P{^Is_Block=kharoshthi}', "");
Expect(1, 68191, '\p{Is_Block=_-Kharoshthi}', "");
Expect(0, 68191, '\p{^Is_Block=_-Kharoshthi}', "");
Expect(0, 68191, '\P{Is_Block=_-Kharoshthi}', "");
Expect(1, 68191, '\P{^Is_Block=_-Kharoshthi}', "");
Expect(0, 68192, '\p{Is_Block=_-Kharoshthi}', "");
Expect(1, 68192, '\p{^Is_Block=_-Kharoshthi}', "");
Expect(1, 68192, '\P{Is_Block=_-Kharoshthi}', "");
Expect(0, 68192, '\P{^Is_Block=_-Kharoshthi}', "");
Error('\p{Is_Blk=	KHAROSHTHI:=}');
Error('\P{Is_Blk=	KHAROSHTHI:=}');
Expect(1, 68191, '\p{Is_Blk=kharoshthi}', "");
Expect(0, 68191, '\p{^Is_Blk=kharoshthi}', "");
Expect(0, 68191, '\P{Is_Blk=kharoshthi}', "");
Expect(1, 68191, '\P{^Is_Blk=kharoshthi}', "");
Expect(0, 68192, '\p{Is_Blk=kharoshthi}', "");
Expect(1, 68192, '\p{^Is_Blk=kharoshthi}', "");
Expect(1, 68192, '\P{Is_Blk=kharoshthi}', "");
Expect(0, 68192, '\P{^Is_Blk=kharoshthi}', "");
Expect(1, 68191, '\p{Is_Blk: 	_Kharoshthi}', "");
Expect(0, 68191, '\p{^Is_Blk: 	_Kharoshthi}', "");
Expect(0, 68191, '\P{Is_Blk: 	_Kharoshthi}', "");
Expect(1, 68191, '\P{^Is_Blk: 	_Kharoshthi}', "");
Expect(0, 68192, '\p{Is_Blk: 	_Kharoshthi}', "");
Expect(1, 68192, '\p{^Is_Blk: 	_Kharoshthi}', "");
Expect(1, 68192, '\P{Is_Blk: 	_Kharoshthi}', "");
Expect(0, 68192, '\P{^Is_Blk: 	_Kharoshthi}', "");
Error('\p{Block=_ KHMER:=}');
Error('\P{Block=_ KHMER:=}');
Expect(1, 6143, '\p{Block=khmer}', "");
Expect(0, 6143, '\p{^Block=khmer}', "");
Expect(0, 6143, '\P{Block=khmer}', "");
Expect(1, 6143, '\P{^Block=khmer}', "");
Expect(0, 6144, '\p{Block=khmer}', "");
Expect(1, 6144, '\p{^Block=khmer}', "");
Expect(1, 6144, '\P{Block=khmer}', "");
Expect(0, 6144, '\P{^Block=khmer}', "");
Expect(1, 6143, '\p{Block=_-Khmer}', "");
Expect(0, 6143, '\p{^Block=_-Khmer}', "");
Expect(0, 6143, '\P{Block=_-Khmer}', "");
Expect(1, 6143, '\P{^Block=_-Khmer}', "");
Expect(0, 6144, '\p{Block=_-Khmer}', "");
Expect(1, 6144, '\p{^Block=_-Khmer}', "");
Expect(1, 6144, '\P{Block=_-Khmer}', "");
Expect(0, 6144, '\P{^Block=_-Khmer}', "");
Error('\p{Blk= Khmer/a/}');
Error('\P{Blk= Khmer/a/}');
Expect(1, 6143, '\p{Blk=khmer}', "");
Expect(0, 6143, '\p{^Blk=khmer}', "");
Expect(0, 6143, '\P{Blk=khmer}', "");
Expect(1, 6143, '\P{^Blk=khmer}', "");
Expect(0, 6144, '\p{Blk=khmer}', "");
Expect(1, 6144, '\p{^Blk=khmer}', "");
Expect(1, 6144, '\P{Blk=khmer}', "");
Expect(0, 6144, '\P{^Blk=khmer}', "");
Expect(1, 6143, '\p{Blk=  KHMER}', "");
Expect(0, 6143, '\p{^Blk=  KHMER}', "");
Expect(0, 6143, '\P{Blk=  KHMER}', "");
Expect(1, 6143, '\P{^Blk=  KHMER}', "");
Expect(0, 6144, '\p{Blk=  KHMER}', "");
Expect(1, 6144, '\p{^Blk=  KHMER}', "");
Expect(1, 6144, '\P{Blk=  KHMER}', "");
Expect(0, 6144, '\P{^Blk=  KHMER}', "");
Error('\p{Is_Block=/a/		Khmer}');
Error('\P{Is_Block=/a/		Khmer}');
Expect(1, 6143, '\p{Is_Block=khmer}', "");
Expect(0, 6143, '\p{^Is_Block=khmer}', "");
Expect(0, 6143, '\P{Is_Block=khmer}', "");
Expect(1, 6143, '\P{^Is_Block=khmer}', "");
Expect(0, 6144, '\p{Is_Block=khmer}', "");
Expect(1, 6144, '\p{^Is_Block=khmer}', "");
Expect(1, 6144, '\P{Is_Block=khmer}', "");
Expect(0, 6144, '\P{^Is_Block=khmer}', "");
Expect(1, 6143, '\p{Is_Block=-_khmer}', "");
Expect(0, 6143, '\p{^Is_Block=-_khmer}', "");
Expect(0, 6143, '\P{Is_Block=-_khmer}', "");
Expect(1, 6143, '\P{^Is_Block=-_khmer}', "");
Expect(0, 6144, '\p{Is_Block=-_khmer}', "");
Expect(1, 6144, '\p{^Is_Block=-_khmer}', "");
Expect(1, 6144, '\P{Is_Block=-_khmer}', "");
Expect(0, 6144, '\P{^Is_Block=-_khmer}', "");
Error('\p{Is_Blk=:=khmer}');
Error('\P{Is_Blk=:=khmer}');
Expect(1, 6143, '\p{Is_Blk=khmer}', "");
Expect(0, 6143, '\p{^Is_Blk=khmer}', "");
Expect(0, 6143, '\P{Is_Blk=khmer}', "");
Expect(1, 6143, '\P{^Is_Blk=khmer}', "");
Expect(0, 6144, '\p{Is_Blk=khmer}', "");
Expect(1, 6144, '\p{^Is_Blk=khmer}', "");
Expect(1, 6144, '\P{Is_Blk=khmer}', "");
Expect(0, 6144, '\P{^Is_Blk=khmer}', "");
Expect(1, 6143, '\p{Is_Blk=_-Khmer}', "");
Expect(0, 6143, '\p{^Is_Blk=_-Khmer}', "");
Expect(0, 6143, '\P{Is_Blk=_-Khmer}', "");
Expect(1, 6143, '\P{^Is_Blk=_-Khmer}', "");
Expect(0, 6144, '\p{Is_Blk=_-Khmer}', "");
Expect(1, 6144, '\p{^Is_Blk=_-Khmer}', "");
Expect(1, 6144, '\P{Is_Blk=_-Khmer}', "");
Expect(0, 6144, '\P{^Is_Blk=_-Khmer}', "");
Error('\p{Block=:= -KHMER_Symbols}');
Error('\P{Block=:= -KHMER_Symbols}');
Expect(1, 6655, '\p{Block=khmersymbols}', "");
Expect(0, 6655, '\p{^Block=khmersymbols}', "");
Expect(0, 6655, '\P{Block=khmersymbols}', "");
Expect(1, 6655, '\P{^Block=khmersymbols}', "");
Expect(0, 6656, '\p{Block=khmersymbols}', "");
Expect(1, 6656, '\p{^Block=khmersymbols}', "");
Expect(1, 6656, '\P{Block=khmersymbols}', "");
Expect(0, 6656, '\P{^Block=khmersymbols}', "");
Expect(1, 6655, '\p{Block=  khmer_SYMBOLS}', "");
Expect(0, 6655, '\p{^Block=  khmer_SYMBOLS}', "");
Expect(0, 6655, '\P{Block=  khmer_SYMBOLS}', "");
Expect(1, 6655, '\P{^Block=  khmer_SYMBOLS}', "");
Expect(0, 6656, '\p{Block=  khmer_SYMBOLS}', "");
Expect(1, 6656, '\p{^Block=  khmer_SYMBOLS}', "");
Expect(1, 6656, '\P{Block=  khmer_SYMBOLS}', "");
Expect(0, 6656, '\P{^Block=  khmer_SYMBOLS}', "");
Error('\p{Blk=	:=khmer_symbols}');
Error('\P{Blk=	:=khmer_symbols}');
Expect(1, 6655, '\p{Blk=khmersymbols}', "");
Expect(0, 6655, '\p{^Blk=khmersymbols}', "");
Expect(0, 6655, '\P{Blk=khmersymbols}', "");
Expect(1, 6655, '\P{^Blk=khmersymbols}', "");
Expect(0, 6656, '\p{Blk=khmersymbols}', "");
Expect(1, 6656, '\p{^Blk=khmersymbols}', "");
Expect(1, 6656, '\P{Blk=khmersymbols}', "");
Expect(0, 6656, '\P{^Blk=khmersymbols}', "");
Expect(1, 6655, '\p{Blk=-KHMER_SYMBOLS}', "");
Expect(0, 6655, '\p{^Blk=-KHMER_SYMBOLS}', "");
Expect(0, 6655, '\P{Blk=-KHMER_SYMBOLS}', "");
Expect(1, 6655, '\P{^Blk=-KHMER_SYMBOLS}', "");
Expect(0, 6656, '\p{Blk=-KHMER_SYMBOLS}', "");
Expect(1, 6656, '\p{^Blk=-KHMER_SYMBOLS}', "");
Expect(1, 6656, '\P{Blk=-KHMER_SYMBOLS}', "");
Expect(0, 6656, '\P{^Blk=-KHMER_SYMBOLS}', "");
Error('\p{Is_Block=/a/-_KHMER_Symbols}');
Error('\P{Is_Block=/a/-_KHMER_Symbols}');
Expect(1, 6655, '\p{Is_Block=khmersymbols}', "");
Expect(0, 6655, '\p{^Is_Block=khmersymbols}', "");
Expect(0, 6655, '\P{Is_Block=khmersymbols}', "");
Expect(1, 6655, '\P{^Is_Block=khmersymbols}', "");
Expect(0, 6656, '\p{Is_Block=khmersymbols}', "");
Expect(1, 6656, '\p{^Is_Block=khmersymbols}', "");
Expect(1, 6656, '\P{Is_Block=khmersymbols}', "");
Expect(0, 6656, '\P{^Is_Block=khmersymbols}', "");
Expect(1, 6655, '\p{Is_Block= _KHMER_SYMBOLS}', "");
Expect(0, 6655, '\p{^Is_Block= _KHMER_SYMBOLS}', "");
Expect(0, 6655, '\P{Is_Block= _KHMER_SYMBOLS}', "");
Expect(1, 6655, '\P{^Is_Block= _KHMER_SYMBOLS}', "");
Expect(0, 6656, '\p{Is_Block= _KHMER_SYMBOLS}', "");
Expect(1, 6656, '\p{^Is_Block= _KHMER_SYMBOLS}', "");
Expect(1, 6656, '\P{Is_Block= _KHMER_SYMBOLS}', "");
Expect(0, 6656, '\P{^Is_Block= _KHMER_SYMBOLS}', "");
Error('\p{Is_Blk=:= -khmer_Symbols}');
Error('\P{Is_Blk=:= -khmer_Symbols}');
Expect(1, 6655, '\p{Is_Blk=khmersymbols}', "");
Expect(0, 6655, '\p{^Is_Blk=khmersymbols}', "");
Expect(0, 6655, '\P{Is_Blk=khmersymbols}', "");
Expect(1, 6655, '\P{^Is_Blk=khmersymbols}', "");
Expect(0, 6656, '\p{Is_Blk=khmersymbols}', "");
Expect(1, 6656, '\p{^Is_Blk=khmersymbols}', "");
Expect(1, 6656, '\P{Is_Blk=khmersymbols}', "");
Expect(0, 6656, '\P{^Is_Blk=khmersymbols}', "");
Expect(1, 6655, '\p{Is_Blk=-Khmer_Symbols}', "");
Expect(0, 6655, '\p{^Is_Blk=-Khmer_Symbols}', "");
Expect(0, 6655, '\P{Is_Blk=-Khmer_Symbols}', "");
Expect(1, 6655, '\P{^Is_Blk=-Khmer_Symbols}', "");
Expect(0, 6656, '\p{Is_Blk=-Khmer_Symbols}', "");
Expect(1, 6656, '\p{^Is_Blk=-Khmer_Symbols}', "");
Expect(1, 6656, '\P{Is_Blk=-Khmer_Symbols}', "");
Expect(0, 6656, '\P{^Is_Blk=-Khmer_Symbols}', "");
Error('\p{Block=_:=LAO}');
Error('\P{Block=_:=LAO}');
Expect(1, 3839, '\p{Block=lao}', "");
Expect(0, 3839, '\p{^Block=lao}', "");
Expect(0, 3839, '\P{Block=lao}', "");
Expect(1, 3839, '\P{^Block=lao}', "");
Expect(0, 3840, '\p{Block=lao}', "");
Expect(1, 3840, '\p{^Block=lao}', "");
Expect(1, 3840, '\P{Block=lao}', "");
Expect(0, 3840, '\P{^Block=lao}', "");
Expect(1, 3839, '\p{Block=  Lao}', "");
Expect(0, 3839, '\p{^Block=  Lao}', "");
Expect(0, 3839, '\P{Block=  Lao}', "");
Expect(1, 3839, '\P{^Block=  Lao}', "");
Expect(0, 3840, '\p{Block=  Lao}', "");
Expect(1, 3840, '\p{^Block=  Lao}', "");
Expect(1, 3840, '\P{Block=  Lao}', "");
Expect(0, 3840, '\P{^Block=  Lao}', "");
Error('\p{Blk=	LAO/a/}');
Error('\P{Blk=	LAO/a/}');
Expect(1, 3839, '\p{Blk=lao}', "");
Expect(0, 3839, '\p{^Blk=lao}', "");
Expect(0, 3839, '\P{Blk=lao}', "");
Expect(1, 3839, '\P{^Blk=lao}', "");
Expect(0, 3840, '\p{Blk=lao}', "");
Expect(1, 3840, '\p{^Blk=lao}', "");
Expect(1, 3840, '\P{Blk=lao}', "");
Expect(0, 3840, '\P{^Blk=lao}', "");
Expect(1, 3839, '\p{Blk=LAO}', "");
Expect(0, 3839, '\p{^Blk=LAO}', "");
Expect(0, 3839, '\P{Blk=LAO}', "");
Expect(1, 3839, '\P{^Blk=LAO}', "");
Expect(0, 3840, '\p{Blk=LAO}', "");
Expect(1, 3840, '\p{^Blk=LAO}', "");
Expect(1, 3840, '\P{Blk=LAO}', "");
Expect(0, 3840, '\P{^Blk=LAO}', "");
Error('\p{Is_Block=-:=Lao}');
Error('\P{Is_Block=-:=Lao}');
Expect(1, 3839, '\p{Is_Block=lao}', "");
Expect(0, 3839, '\p{^Is_Block=lao}', "");
Expect(0, 3839, '\P{Is_Block=lao}', "");
Expect(1, 3839, '\P{^Is_Block=lao}', "");
Expect(0, 3840, '\p{Is_Block=lao}', "");
Expect(1, 3840, '\p{^Is_Block=lao}', "");
Expect(1, 3840, '\P{Is_Block=lao}', "");
Expect(0, 3840, '\P{^Is_Block=lao}', "");
Expect(1, 3839, '\p{Is_Block= Lao}', "");
Expect(0, 3839, '\p{^Is_Block= Lao}', "");
Expect(0, 3839, '\P{Is_Block= Lao}', "");
Expect(1, 3839, '\P{^Is_Block= Lao}', "");
Expect(0, 3840, '\p{Is_Block= Lao}', "");
Expect(1, 3840, '\p{^Is_Block= Lao}', "");
Expect(1, 3840, '\P{Is_Block= Lao}', "");
Expect(0, 3840, '\P{^Is_Block= Lao}', "");
Error('\p{Is_Blk=_-Lao:=}');
Error('\P{Is_Blk=_-Lao:=}');
Expect(1, 3839, '\p{Is_Blk:   lao}', "");
Expect(0, 3839, '\p{^Is_Blk:   lao}', "");
Expect(0, 3839, '\P{Is_Blk:   lao}', "");
Expect(1, 3839, '\P{^Is_Blk:   lao}', "");
Expect(0, 3840, '\p{Is_Blk:   lao}', "");
Expect(1, 3840, '\p{^Is_Blk:   lao}', "");
Expect(1, 3840, '\P{Is_Blk:   lao}', "");
Expect(0, 3840, '\P{^Is_Blk:   lao}', "");
Expect(1, 3839, '\p{Is_Blk=	 LAO}', "");
Expect(0, 3839, '\p{^Is_Blk=	 LAO}', "");
Expect(0, 3839, '\P{Is_Blk=	 LAO}', "");
Expect(1, 3839, '\P{^Is_Blk=	 LAO}', "");
Expect(0, 3840, '\p{Is_Blk=	 LAO}', "");
Expect(1, 3840, '\p{^Is_Blk=	 LAO}', "");
Expect(1, 3840, '\P{Is_Blk=	 LAO}', "");
Expect(0, 3840, '\P{^Is_Blk=	 LAO}', "");
Error('\p{Block:	_	Latin_1_Supplement/a/}');
Error('\P{Block:	_	Latin_1_Supplement/a/}');
Expect(1, 255, '\p{Block:latin1supplement}', "");
Expect(0, 255, '\p{^Block:latin1supplement}', "");
Expect(0, 255, '\P{Block:latin1supplement}', "");
Expect(1, 255, '\P{^Block:latin1supplement}', "");
Expect(0, 256, '\p{Block:latin1supplement}', "");
Expect(1, 256, '\p{^Block:latin1supplement}', "");
Expect(1, 256, '\P{Block:latin1supplement}', "");
Expect(0, 256, '\P{^Block:latin1supplement}', "");
Expect(1, 255, '\p{Block=	 Latin_1_Supplement}', "");
Expect(0, 255, '\p{^Block=	 Latin_1_Supplement}', "");
Expect(0, 255, '\P{Block=	 Latin_1_Supplement}', "");
Expect(1, 255, '\P{^Block=	 Latin_1_Supplement}', "");
Expect(0, 256, '\p{Block=	 Latin_1_Supplement}', "");
Expect(1, 256, '\p{^Block=	 Latin_1_Supplement}', "");
Expect(1, 256, '\P{Block=	 Latin_1_Supplement}', "");
Expect(0, 256, '\P{^Block=	 Latin_1_Supplement}', "");
Error('\p{Blk=/a/ -Latin_1_sup}');
Error('\P{Blk=/a/ -Latin_1_sup}');
Expect(1, 255, '\p{Blk=latin1sup}', "");
Expect(0, 255, '\p{^Blk=latin1sup}', "");
Expect(0, 255, '\P{Blk=latin1sup}', "");
Expect(1, 255, '\P{^Blk=latin1sup}', "");
Expect(0, 256, '\p{Blk=latin1sup}', "");
Expect(1, 256, '\p{^Blk=latin1sup}', "");
Expect(1, 256, '\P{Blk=latin1sup}', "");
Expect(0, 256, '\P{^Blk=latin1sup}', "");
Expect(1, 255, '\p{Blk=_	Latin_1_SUP}', "");
Expect(0, 255, '\p{^Blk=_	Latin_1_SUP}', "");
Expect(0, 255, '\P{Blk=_	Latin_1_SUP}', "");
Expect(1, 255, '\P{^Blk=_	Latin_1_SUP}', "");
Expect(0, 256, '\p{Blk=_	Latin_1_SUP}', "");
Expect(1, 256, '\p{^Blk=_	Latin_1_SUP}', "");
Expect(1, 256, '\P{Blk=_	Latin_1_SUP}', "");
Expect(0, 256, '\P{^Blk=_	Latin_1_SUP}', "");
Error('\p{Is_Block= -LATIN_1/a/}');
Error('\P{Is_Block= -LATIN_1/a/}');
Expect(1, 255, '\p{Is_Block=latin1}', "");
Expect(0, 255, '\p{^Is_Block=latin1}', "");
Expect(0, 255, '\P{Is_Block=latin1}', "");
Expect(1, 255, '\P{^Is_Block=latin1}', "");
Expect(0, 256, '\p{Is_Block=latin1}', "");
Expect(1, 256, '\p{^Is_Block=latin1}', "");
Expect(1, 256, '\P{Is_Block=latin1}', "");
Expect(0, 256, '\P{^Is_Block=latin1}', "");
Expect(1, 255, '\p{Is_Block=_Latin_1}', "");
Expect(0, 255, '\p{^Is_Block=_Latin_1}', "");
Expect(0, 255, '\P{Is_Block=_Latin_1}', "");
Expect(1, 255, '\P{^Is_Block=_Latin_1}', "");
Expect(0, 256, '\p{Is_Block=_Latin_1}', "");
Expect(1, 256, '\p{^Is_Block=_Latin_1}', "");
Expect(1, 256, '\P{Is_Block=_Latin_1}', "");
Expect(0, 256, '\P{^Is_Block=_Latin_1}', "");
Error('\p{Is_Blk= -LATIN_1_Supplement/a/}');
Error('\P{Is_Blk= -LATIN_1_Supplement/a/}');
Expect(1, 255, '\p{Is_Blk=latin1supplement}', "");
Expect(0, 255, '\p{^Is_Blk=latin1supplement}', "");
Expect(0, 255, '\P{Is_Blk=latin1supplement}', "");
Expect(1, 255, '\P{^Is_Blk=latin1supplement}', "");
Expect(0, 256, '\p{Is_Blk=latin1supplement}', "");
Expect(1, 256, '\p{^Is_Blk=latin1supplement}', "");
Expect(1, 256, '\P{Is_Blk=latin1supplement}', "");
Expect(0, 256, '\P{^Is_Blk=latin1supplement}', "");
Expect(1, 255, '\p{Is_Blk=		LATIN_1_Supplement}', "");
Expect(0, 255, '\p{^Is_Blk=		LATIN_1_Supplement}', "");
Expect(0, 255, '\P{Is_Blk=		LATIN_1_Supplement}', "");
Expect(1, 255, '\P{^Is_Blk=		LATIN_1_Supplement}', "");
Expect(0, 256, '\p{Is_Blk=		LATIN_1_Supplement}', "");
Expect(1, 256, '\p{^Is_Blk=		LATIN_1_Supplement}', "");
Expect(1, 256, '\P{Is_Blk=		LATIN_1_Supplement}', "");
Expect(0, 256, '\P{^Is_Blk=		LATIN_1_Supplement}', "");
Error('\p{Block=/a/latin_extended_A}');
Error('\P{Block=/a/latin_extended_A}');
Expect(1, 383, '\p{Block=latinextendeda}', "");
Expect(0, 383, '\p{^Block=latinextendeda}', "");
Expect(0, 383, '\P{Block=latinextendeda}', "");
Expect(1, 383, '\P{^Block=latinextendeda}', "");
Expect(0, 384, '\p{Block=latinextendeda}', "");
Expect(1, 384, '\p{^Block=latinextendeda}', "");
Expect(1, 384, '\P{Block=latinextendeda}', "");
Expect(0, 384, '\P{^Block=latinextendeda}', "");
Expect(1, 383, '\p{Block= 	LATIN_Extended_a}', "");
Expect(0, 383, '\p{^Block= 	LATIN_Extended_a}', "");
Expect(0, 383, '\P{Block= 	LATIN_Extended_a}', "");
Expect(1, 383, '\P{^Block= 	LATIN_Extended_a}', "");
Expect(0, 384, '\p{Block= 	LATIN_Extended_a}', "");
Expect(1, 384, '\p{^Block= 	LATIN_Extended_a}', "");
Expect(1, 384, '\P{Block= 	LATIN_Extended_a}', "");
Expect(0, 384, '\P{^Block= 	LATIN_Extended_a}', "");
Error('\p{Blk= _Latin_EXT_A:=}');
Error('\P{Blk= _Latin_EXT_A:=}');
Expect(1, 383, '\p{Blk=latinexta}', "");
Expect(0, 383, '\p{^Blk=latinexta}', "");
Expect(0, 383, '\P{Blk=latinexta}', "");
Expect(1, 383, '\P{^Blk=latinexta}', "");
Expect(0, 384, '\p{Blk=latinexta}', "");
Expect(1, 384, '\p{^Blk=latinexta}', "");
Expect(1, 384, '\P{Blk=latinexta}', "");
Expect(0, 384, '\P{^Blk=latinexta}', "");
Expect(1, 383, '\p{Blk:-Latin_Ext_A}', "");
Expect(0, 383, '\p{^Blk:-Latin_Ext_A}', "");
Expect(0, 383, '\P{Blk:-Latin_Ext_A}', "");
Expect(1, 383, '\P{^Blk:-Latin_Ext_A}', "");
Expect(0, 384, '\p{Blk:-Latin_Ext_A}', "");
Expect(1, 384, '\p{^Blk:-Latin_Ext_A}', "");
Expect(1, 384, '\P{Blk:-Latin_Ext_A}', "");
Expect(0, 384, '\P{^Blk:-Latin_Ext_A}', "");
Error('\p{Is_Block=-_LATIN_Extended_A:=}');
Error('\P{Is_Block=-_LATIN_Extended_A:=}');
Expect(1, 383, '\p{Is_Block=latinextendeda}', "");
Expect(0, 383, '\p{^Is_Block=latinextendeda}', "");
Expect(0, 383, '\P{Is_Block=latinextendeda}', "");
Expect(1, 383, '\P{^Is_Block=latinextendeda}', "");
Expect(0, 384, '\p{Is_Block=latinextendeda}', "");
Expect(1, 384, '\p{^Is_Block=latinextendeda}', "");
Expect(1, 384, '\P{Is_Block=latinextendeda}', "");
Expect(0, 384, '\P{^Is_Block=latinextendeda}', "");
Expect(1, 383, '\p{Is_Block:    Latin_extended_A}', "");
Expect(0, 383, '\p{^Is_Block:    Latin_extended_A}', "");
Expect(0, 383, '\P{Is_Block:    Latin_extended_A}', "");
Expect(1, 383, '\P{^Is_Block:    Latin_extended_A}', "");
Expect(0, 384, '\p{Is_Block:    Latin_extended_A}', "");
Expect(1, 384, '\p{^Is_Block:    Latin_extended_A}', "");
Expect(1, 384, '\P{Is_Block:    Latin_extended_A}', "");
Expect(0, 384, '\P{^Is_Block:    Latin_extended_A}', "");
Error('\p{Is_Blk=/a/	-Latin_Ext_A}');
Error('\P{Is_Blk=/a/	-Latin_Ext_A}');
Expect(1, 383, '\p{Is_Blk=latinexta}', "");
Expect(0, 383, '\p{^Is_Blk=latinexta}', "");
Expect(0, 383, '\P{Is_Blk=latinexta}', "");
Expect(1, 383, '\P{^Is_Blk=latinexta}', "");
Expect(0, 384, '\p{Is_Blk=latinexta}', "");
Expect(1, 384, '\p{^Is_Blk=latinexta}', "");
Expect(1, 384, '\P{Is_Blk=latinexta}', "");
Expect(0, 384, '\P{^Is_Blk=latinexta}', "");
Expect(1, 383, '\p{Is_Blk=_latin_EXT_A}', "");
Expect(0, 383, '\p{^Is_Blk=_latin_EXT_A}', "");
Expect(0, 383, '\P{Is_Blk=_latin_EXT_A}', "");
Expect(1, 383, '\P{^Is_Blk=_latin_EXT_A}', "");
Expect(0, 384, '\p{Is_Blk=_latin_EXT_A}', "");
Expect(1, 384, '\p{^Is_Blk=_latin_EXT_A}', "");
Expect(1, 384, '\P{Is_Blk=_latin_EXT_A}', "");
Expect(0, 384, '\P{^Is_Blk=_latin_EXT_A}', "");
Error('\p{Block=/a/ -Latin_extended_Additional}');
Error('\P{Block=/a/ -Latin_extended_Additional}');
Expect(1, 7935, '\p{Block=latinextendedadditional}', "");
Expect(0, 7935, '\p{^Block=latinextendedadditional}', "");
Expect(0, 7935, '\P{Block=latinextendedadditional}', "");
Expect(1, 7935, '\P{^Block=latinextendedadditional}', "");
Expect(0, 7936, '\p{Block=latinextendedadditional}', "");
Expect(1, 7936, '\p{^Block=latinextendedadditional}', "");
Expect(1, 7936, '\P{Block=latinextendedadditional}', "");
Expect(0, 7936, '\P{^Block=latinextendedadditional}', "");
Expect(1, 7935, '\p{Block:	--LATIN_Extended_additional}', "");
Expect(0, 7935, '\p{^Block:	--LATIN_Extended_additional}', "");
Expect(0, 7935, '\P{Block:	--LATIN_Extended_additional}', "");
Expect(1, 7935, '\P{^Block:	--LATIN_Extended_additional}', "");
Expect(0, 7936, '\p{Block:	--LATIN_Extended_additional}', "");
Expect(1, 7936, '\p{^Block:	--LATIN_Extended_additional}', "");
Expect(1, 7936, '\P{Block:	--LATIN_Extended_additional}', "");
Expect(0, 7936, '\P{^Block:	--LATIN_Extended_additional}', "");
Error('\p{Blk:	:=Latin_ext_ADDITIONAL}');
Error('\P{Blk:	:=Latin_ext_ADDITIONAL}');
Expect(1, 7935, '\p{Blk=latinextadditional}', "");
Expect(0, 7935, '\p{^Blk=latinextadditional}', "");
Expect(0, 7935, '\P{Blk=latinextadditional}', "");
Expect(1, 7935, '\P{^Blk=latinextadditional}', "");
Expect(0, 7936, '\p{Blk=latinextadditional}', "");
Expect(1, 7936, '\p{^Blk=latinextadditional}', "");
Expect(1, 7936, '\P{Blk=latinextadditional}', "");
Expect(0, 7936, '\P{^Blk=latinextadditional}', "");
Expect(1, 7935, '\p{Blk=LATIN_Ext_additional}', "");
Expect(0, 7935, '\p{^Blk=LATIN_Ext_additional}', "");
Expect(0, 7935, '\P{Blk=LATIN_Ext_additional}', "");
Expect(1, 7935, '\P{^Blk=LATIN_Ext_additional}', "");
Expect(0, 7936, '\p{Blk=LATIN_Ext_additional}', "");
Expect(1, 7936, '\p{^Blk=LATIN_Ext_additional}', "");
Expect(1, 7936, '\P{Blk=LATIN_Ext_additional}', "");
Expect(0, 7936, '\P{^Blk=LATIN_Ext_additional}', "");
Error('\p{Is_Block= latin_Extended_ADDITIONAL/a/}');
Error('\P{Is_Block= latin_Extended_ADDITIONAL/a/}');
Expect(1, 7935, '\p{Is_Block=latinextendedadditional}', "");
Expect(0, 7935, '\p{^Is_Block=latinextendedadditional}', "");
Expect(0, 7935, '\P{Is_Block=latinextendedadditional}', "");
Expect(1, 7935, '\P{^Is_Block=latinextendedadditional}', "");
Expect(0, 7936, '\p{Is_Block=latinextendedadditional}', "");
Expect(1, 7936, '\p{^Is_Block=latinextendedadditional}', "");
Expect(1, 7936, '\P{Is_Block=latinextendedadditional}', "");
Expect(0, 7936, '\P{^Is_Block=latinextendedadditional}', "");
Expect(1, 7935, '\p{Is_Block=-	LATIN_extended_Additional}', "");
Expect(0, 7935, '\p{^Is_Block=-	LATIN_extended_Additional}', "");
Expect(0, 7935, '\P{Is_Block=-	LATIN_extended_Additional}', "");
Expect(1, 7935, '\P{^Is_Block=-	LATIN_extended_Additional}', "");
Expect(0, 7936, '\p{Is_Block=-	LATIN_extended_Additional}', "");
Expect(1, 7936, '\p{^Is_Block=-	LATIN_extended_Additional}', "");
Expect(1, 7936, '\P{Is_Block=-	LATIN_extended_Additional}', "");
Expect(0, 7936, '\P{^Is_Block=-	LATIN_extended_Additional}', "");
Error('\p{Is_Blk:	 LATIN_EXT_Additional/a/}');
Error('\P{Is_Blk:	 LATIN_EXT_Additional/a/}');
Expect(1, 7935, '\p{Is_Blk=latinextadditional}', "");
Expect(0, 7935, '\p{^Is_Blk=latinextadditional}', "");
Expect(0, 7935, '\P{Is_Blk=latinextadditional}', "");
Expect(1, 7935, '\P{^Is_Blk=latinextadditional}', "");
Expect(0, 7936, '\p{Is_Blk=latinextadditional}', "");
Expect(1, 7936, '\p{^Is_Blk=latinextadditional}', "");
Expect(1, 7936, '\P{Is_Blk=latinextadditional}', "");
Expect(0, 7936, '\P{^Is_Blk=latinextadditional}', "");
Expect(1, 7935, '\p{Is_Blk=  Latin_ext_Additional}', "");
Expect(0, 7935, '\p{^Is_Blk=  Latin_ext_Additional}', "");
Expect(0, 7935, '\P{Is_Blk=  Latin_ext_Additional}', "");
Expect(1, 7935, '\P{^Is_Blk=  Latin_ext_Additional}', "");
Expect(0, 7936, '\p{Is_Blk=  Latin_ext_Additional}', "");
Expect(1, 7936, '\p{^Is_Blk=  Latin_ext_Additional}', "");
Expect(1, 7936, '\P{Is_Blk=  Latin_ext_Additional}', "");
Expect(0, 7936, '\P{^Is_Blk=  Latin_ext_Additional}', "");
Error('\p{Block=-:=Latin_EXTENDED_B}');
Error('\P{Block=-:=Latin_EXTENDED_B}');
Expect(1, 591, '\p{Block=latinextendedb}', "");
Expect(0, 591, '\p{^Block=latinextendedb}', "");
Expect(0, 591, '\P{Block=latinextendedb}', "");
Expect(1, 591, '\P{^Block=latinextendedb}', "");
Expect(0, 592, '\p{Block=latinextendedb}', "");
Expect(1, 592, '\p{^Block=latinextendedb}', "");
Expect(1, 592, '\P{Block=latinextendedb}', "");
Expect(0, 592, '\P{^Block=latinextendedb}', "");
Expect(1, 591, '\p{Block=  Latin_Extended_B}', "");
Expect(0, 591, '\p{^Block=  Latin_Extended_B}', "");
Expect(0, 591, '\P{Block=  Latin_Extended_B}', "");
Expect(1, 591, '\P{^Block=  Latin_Extended_B}', "");
Expect(0, 592, '\p{Block=  Latin_Extended_B}', "");
Expect(1, 592, '\p{^Block=  Latin_Extended_B}', "");
Expect(1, 592, '\P{Block=  Latin_Extended_B}', "");
Expect(0, 592, '\P{^Block=  Latin_Extended_B}', "");
Error('\p{Blk= /a/Latin_ext_B}');
Error('\P{Blk= /a/Latin_ext_B}');
Expect(1, 591, '\p{Blk: latinextb}', "");
Expect(0, 591, '\p{^Blk: latinextb}', "");
Expect(0, 591, '\P{Blk: latinextb}', "");
Expect(1, 591, '\P{^Blk: latinextb}', "");
Expect(0, 592, '\p{Blk: latinextb}', "");
Expect(1, 592, '\p{^Blk: latinextb}', "");
Expect(1, 592, '\P{Blk: latinextb}', "");
Expect(0, 592, '\P{^Blk: latinextb}', "");
Expect(1, 591, '\p{Blk=		LATIN_ext_B}', "");
Expect(0, 591, '\p{^Blk=		LATIN_ext_B}', "");
Expect(0, 591, '\P{Blk=		LATIN_ext_B}', "");
Expect(1, 591, '\P{^Blk=		LATIN_ext_B}', "");
Expect(0, 592, '\p{Blk=		LATIN_ext_B}', "");
Expect(1, 592, '\p{^Blk=		LATIN_ext_B}', "");
Expect(1, 592, '\P{Blk=		LATIN_ext_B}', "");
Expect(0, 592, '\P{^Blk=		LATIN_ext_B}', "");
Error('\p{Is_Block=/a/	_Latin_extended_B}');
Error('\P{Is_Block=/a/	_Latin_extended_B}');
Expect(1, 591, '\p{Is_Block=latinextendedb}', "");
Expect(0, 591, '\p{^Is_Block=latinextendedb}', "");
Expect(0, 591, '\P{Is_Block=latinextendedb}', "");
Expect(1, 591, '\P{^Is_Block=latinextendedb}', "");
Expect(0, 592, '\p{Is_Block=latinextendedb}', "");
Expect(1, 592, '\p{^Is_Block=latinextendedb}', "");
Expect(1, 592, '\P{Is_Block=latinextendedb}', "");
Expect(0, 592, '\P{^Is_Block=latinextendedb}', "");
Expect(1, 591, '\p{Is_Block=-	latin_EXTENDED_b}', "");
Expect(0, 591, '\p{^Is_Block=-	latin_EXTENDED_b}', "");
Expect(0, 591, '\P{Is_Block=-	latin_EXTENDED_b}', "");
Expect(1, 591, '\P{^Is_Block=-	latin_EXTENDED_b}', "");
Expect(0, 592, '\p{Is_Block=-	latin_EXTENDED_b}', "");
Expect(1, 592, '\p{^Is_Block=-	latin_EXTENDED_b}', "");
Expect(1, 592, '\P{Is_Block=-	latin_EXTENDED_b}', "");
Expect(0, 592, '\P{^Is_Block=-	latin_EXTENDED_b}', "");
Error('\p{Is_Blk=	:=Latin_Ext_B}');
Error('\P{Is_Blk=	:=Latin_Ext_B}');
Expect(1, 591, '\p{Is_Blk=latinextb}', "");
Expect(0, 591, '\p{^Is_Blk=latinextb}', "");
Expect(0, 591, '\P{Is_Blk=latinextb}', "");
Expect(1, 591, '\P{^Is_Blk=latinextb}', "");
Expect(0, 592, '\p{Is_Blk=latinextb}', "");
Expect(1, 592, '\p{^Is_Blk=latinextb}', "");
Expect(1, 592, '\P{Is_Blk=latinextb}', "");
Expect(0, 592, '\P{^Is_Blk=latinextb}', "");
Expect(1, 591, '\p{Is_Blk=_-LATIN_ext_B}', "");
Expect(0, 591, '\p{^Is_Blk=_-LATIN_ext_B}', "");
Expect(0, 591, '\P{Is_Blk=_-LATIN_ext_B}', "");
Expect(1, 591, '\P{^Is_Blk=_-LATIN_ext_B}', "");
Expect(0, 592, '\p{Is_Blk=_-LATIN_ext_B}', "");
Expect(1, 592, '\p{^Is_Blk=_-LATIN_ext_B}', "");
Expect(1, 592, '\P{Is_Blk=_-LATIN_ext_B}', "");
Expect(0, 592, '\P{^Is_Blk=_-LATIN_ext_B}', "");
Error('\p{Block= 	LATIN_EXTENDED_C:=}');
Error('\P{Block= 	LATIN_EXTENDED_C:=}');
Expect(1, 11391, '\p{Block=latinextendedc}', "");
Expect(0, 11391, '\p{^Block=latinextendedc}', "");
Expect(0, 11391, '\P{Block=latinextendedc}', "");
Expect(1, 11391, '\P{^Block=latinextendedc}', "");
Expect(0, 11392, '\p{Block=latinextendedc}', "");
Expect(1, 11392, '\p{^Block=latinextendedc}', "");
Expect(1, 11392, '\P{Block=latinextendedc}', "");
Expect(0, 11392, '\P{^Block=latinextendedc}', "");
Expect(1, 11391, '\p{Block=	_Latin_Extended_C}', "");
Expect(0, 11391, '\p{^Block=	_Latin_Extended_C}', "");
Expect(0, 11391, '\P{Block=	_Latin_Extended_C}', "");
Expect(1, 11391, '\P{^Block=	_Latin_Extended_C}', "");
Expect(0, 11392, '\p{Block=	_Latin_Extended_C}', "");
Expect(1, 11392, '\p{^Block=	_Latin_Extended_C}', "");
Expect(1, 11392, '\P{Block=	_Latin_Extended_C}', "");
Expect(0, 11392, '\P{^Block=	_Latin_Extended_C}', "");
Error('\p{Blk=/a/_	latin_ext_C}');
Error('\P{Blk=/a/_	latin_ext_C}');
Expect(1, 11391, '\p{Blk: latinextc}', "");
Expect(0, 11391, '\p{^Blk: latinextc}', "");
Expect(0, 11391, '\P{Blk: latinextc}', "");
Expect(1, 11391, '\P{^Blk: latinextc}', "");
Expect(0, 11392, '\p{Blk: latinextc}', "");
Expect(1, 11392, '\p{^Blk: latinextc}', "");
Expect(1, 11392, '\P{Blk: latinextc}', "");
Expect(0, 11392, '\P{^Blk: latinextc}', "");
Expect(1, 11391, '\p{Blk=  Latin_Ext_C}', "");
Expect(0, 11391, '\p{^Blk=  Latin_Ext_C}', "");
Expect(0, 11391, '\P{Blk=  Latin_Ext_C}', "");
Expect(1, 11391, '\P{^Blk=  Latin_Ext_C}', "");
Expect(0, 11392, '\p{Blk=  Latin_Ext_C}', "");
Expect(1, 11392, '\p{^Blk=  Latin_Ext_C}', "");
Expect(1, 11392, '\P{Blk=  Latin_Ext_C}', "");
Expect(0, 11392, '\P{^Blk=  Latin_Ext_C}', "");
Error('\p{Is_Block:	_:=Latin_Extended_C}');
Error('\P{Is_Block:	_:=Latin_Extended_C}');
Expect(1, 11391, '\p{Is_Block=latinextendedc}', "");
Expect(0, 11391, '\p{^Is_Block=latinextendedc}', "");
Expect(0, 11391, '\P{Is_Block=latinextendedc}', "");
Expect(1, 11391, '\P{^Is_Block=latinextendedc}', "");
Expect(0, 11392, '\p{Is_Block=latinextendedc}', "");
Expect(1, 11392, '\p{^Is_Block=latinextendedc}', "");
Expect(1, 11392, '\P{Is_Block=latinextendedc}', "");
Expect(0, 11392, '\P{^Is_Block=latinextendedc}', "");
Expect(1, 11391, '\p{Is_Block=	Latin_extended_c}', "");
Expect(0, 11391, '\p{^Is_Block=	Latin_extended_c}', "");
Expect(0, 11391, '\P{Is_Block=	Latin_extended_c}', "");
Expect(1, 11391, '\P{^Is_Block=	Latin_extended_c}', "");
Expect(0, 11392, '\p{Is_Block=	Latin_extended_c}', "");
Expect(1, 11392, '\p{^Is_Block=	Latin_extended_c}', "");
Expect(1, 11392, '\P{Is_Block=	Latin_extended_c}', "");
Expect(0, 11392, '\P{^Is_Block=	Latin_extended_c}', "");
Error('\p{Is_Blk=:=latin_Ext_C}');
Error('\P{Is_Blk=:=latin_Ext_C}');
Expect(1, 11391, '\p{Is_Blk=latinextc}', "");
Expect(0, 11391, '\p{^Is_Blk=latinextc}', "");
Expect(0, 11391, '\P{Is_Blk=latinextc}', "");
Expect(1, 11391, '\P{^Is_Blk=latinextc}', "");
Expect(0, 11392, '\p{Is_Blk=latinextc}', "");
Expect(1, 11392, '\p{^Is_Blk=latinextc}', "");
Expect(1, 11392, '\P{Is_Blk=latinextc}', "");
Expect(0, 11392, '\P{^Is_Blk=latinextc}', "");
Expect(1, 11391, '\p{Is_Blk=Latin_EXT_C}', "");
Expect(0, 11391, '\p{^Is_Blk=Latin_EXT_C}', "");
Expect(0, 11391, '\P{Is_Blk=Latin_EXT_C}', "");
Expect(1, 11391, '\P{^Is_Blk=Latin_EXT_C}', "");
Expect(0, 11392, '\p{Is_Blk=Latin_EXT_C}', "");
Expect(1, 11392, '\p{^Is_Blk=Latin_EXT_C}', "");
Expect(1, 11392, '\P{Is_Blk=Latin_EXT_C}', "");
Expect(0, 11392, '\P{^Is_Blk=Latin_EXT_C}', "");
Error('\p{Block=/a/_-LATIN_Extended_D}');
Error('\P{Block=/a/_-LATIN_Extended_D}');
Expect(1, 43007, '\p{Block=latinextendedd}', "");
Expect(0, 43007, '\p{^Block=latinextendedd}', "");
Expect(0, 43007, '\P{Block=latinextendedd}', "");
Expect(1, 43007, '\P{^Block=latinextendedd}', "");
Expect(0, 43008, '\p{Block=latinextendedd}', "");
Expect(1, 43008, '\p{^Block=latinextendedd}', "");
Expect(1, 43008, '\P{Block=latinextendedd}', "");
Expect(0, 43008, '\P{^Block=latinextendedd}', "");
Expect(1, 43007, '\p{Block=LATIN_EXTENDED_D}', "");
Expect(0, 43007, '\p{^Block=LATIN_EXTENDED_D}', "");
Expect(0, 43007, '\P{Block=LATIN_EXTENDED_D}', "");
Expect(1, 43007, '\P{^Block=LATIN_EXTENDED_D}', "");
Expect(0, 43008, '\p{Block=LATIN_EXTENDED_D}', "");
Expect(1, 43008, '\p{^Block=LATIN_EXTENDED_D}', "");
Expect(1, 43008, '\P{Block=LATIN_EXTENDED_D}', "");
Expect(0, 43008, '\P{^Block=LATIN_EXTENDED_D}', "");
Error('\p{Blk=- latin_Ext_D:=}');
Error('\P{Blk=- latin_Ext_D:=}');
Expect(1, 43007, '\p{Blk=latinextd}', "");
Expect(0, 43007, '\p{^Blk=latinextd}', "");
Expect(0, 43007, '\P{Blk=latinextd}', "");
Expect(1, 43007, '\P{^Blk=latinextd}', "");
Expect(0, 43008, '\p{Blk=latinextd}', "");
Expect(1, 43008, '\p{^Blk=latinextd}', "");
Expect(1, 43008, '\P{Blk=latinextd}', "");
Expect(0, 43008, '\P{^Blk=latinextd}', "");
Expect(1, 43007, '\p{Blk=-_latin_EXT_D}', "");
Expect(0, 43007, '\p{^Blk=-_latin_EXT_D}', "");
Expect(0, 43007, '\P{Blk=-_latin_EXT_D}', "");
Expect(1, 43007, '\P{^Blk=-_latin_EXT_D}', "");
Expect(0, 43008, '\p{Blk=-_latin_EXT_D}', "");
Expect(1, 43008, '\p{^Blk=-_latin_EXT_D}', "");
Expect(1, 43008, '\P{Blk=-_latin_EXT_D}', "");
Expect(0, 43008, '\P{^Blk=-_latin_EXT_D}', "");
Error('\p{Is_Block=/a/LATIN_extended_d}');
Error('\P{Is_Block=/a/LATIN_extended_d}');
Expect(1, 43007, '\p{Is_Block=latinextendedd}', "");
Expect(0, 43007, '\p{^Is_Block=latinextendedd}', "");
Expect(0, 43007, '\P{Is_Block=latinextendedd}', "");
Expect(1, 43007, '\P{^Is_Block=latinextendedd}', "");
Expect(0, 43008, '\p{Is_Block=latinextendedd}', "");
Expect(1, 43008, '\p{^Is_Block=latinextendedd}', "");
Expect(1, 43008, '\P{Is_Block=latinextendedd}', "");
Expect(0, 43008, '\P{^Is_Block=latinextendedd}', "");
Expect(1, 43007, '\p{Is_Block=_ LATIN_extended_D}', "");
Expect(0, 43007, '\p{^Is_Block=_ LATIN_extended_D}', "");
Expect(0, 43007, '\P{Is_Block=_ LATIN_extended_D}', "");
Expect(1, 43007, '\P{^Is_Block=_ LATIN_extended_D}', "");
Expect(0, 43008, '\p{Is_Block=_ LATIN_extended_D}', "");
Expect(1, 43008, '\p{^Is_Block=_ LATIN_extended_D}', "");
Expect(1, 43008, '\P{Is_Block=_ LATIN_extended_D}', "");
Expect(0, 43008, '\P{^Is_Block=_ LATIN_extended_D}', "");
Error('\p{Is_Blk= /a/Latin_Ext_D}');
Error('\P{Is_Blk= /a/Latin_Ext_D}');
Expect(1, 43007, '\p{Is_Blk=latinextd}', "");
Expect(0, 43007, '\p{^Is_Blk=latinextd}', "");
Expect(0, 43007, '\P{Is_Blk=latinextd}', "");
Expect(1, 43007, '\P{^Is_Blk=latinextd}', "");
Expect(0, 43008, '\p{Is_Blk=latinextd}', "");
Expect(1, 43008, '\p{^Is_Blk=latinextd}', "");
Expect(1, 43008, '\P{Is_Blk=latinextd}', "");
Expect(0, 43008, '\P{^Is_Blk=latinextd}', "");
Expect(1, 43007, '\p{Is_Blk=--latin_ext_d}', "");
Expect(0, 43007, '\p{^Is_Blk=--latin_ext_d}', "");
Expect(0, 43007, '\P{Is_Blk=--latin_ext_d}', "");
Expect(1, 43007, '\P{^Is_Blk=--latin_ext_d}', "");
Expect(0, 43008, '\p{Is_Blk=--latin_ext_d}', "");
Expect(1, 43008, '\p{^Is_Blk=--latin_ext_d}', "");
Expect(1, 43008, '\P{Is_Blk=--latin_ext_d}', "");
Expect(0, 43008, '\P{^Is_Blk=--latin_ext_d}', "");
Error('\p{Block=_	Lepcha/a/}');
Error('\P{Block=_	Lepcha/a/}');
Expect(1, 7247, '\p{Block=lepcha}', "");
Expect(0, 7247, '\p{^Block=lepcha}', "");
Expect(0, 7247, '\P{Block=lepcha}', "");
Expect(1, 7247, '\P{^Block=lepcha}', "");
Expect(0, 7248, '\p{Block=lepcha}', "");
Expect(1, 7248, '\p{^Block=lepcha}', "");
Expect(1, 7248, '\P{Block=lepcha}', "");
Expect(0, 7248, '\P{^Block=lepcha}', "");
Expect(1, 7247, '\p{Block= -Lepcha}', "");
Expect(0, 7247, '\p{^Block= -Lepcha}', "");
Expect(0, 7247, '\P{Block= -Lepcha}', "");
Expect(1, 7247, '\P{^Block= -Lepcha}', "");
Expect(0, 7248, '\p{Block= -Lepcha}', "");
Expect(1, 7248, '\p{^Block= -Lepcha}', "");
Expect(1, 7248, '\P{Block= -Lepcha}', "");
Expect(0, 7248, '\P{^Block= -Lepcha}', "");
Error('\p{Blk=-LEPCHA/a/}');
Error('\P{Blk=-LEPCHA/a/}');
Expect(1, 7247, '\p{Blk=lepcha}', "");
Expect(0, 7247, '\p{^Blk=lepcha}', "");
Expect(0, 7247, '\P{Blk=lepcha}', "");
Expect(1, 7247, '\P{^Blk=lepcha}', "");
Expect(0, 7248, '\p{Blk=lepcha}', "");
Expect(1, 7248, '\p{^Blk=lepcha}', "");
Expect(1, 7248, '\P{Blk=lepcha}', "");
Expect(0, 7248, '\P{^Blk=lepcha}', "");
Expect(1, 7247, '\p{Blk:	_LEPCHA}', "");
Expect(0, 7247, '\p{^Blk:	_LEPCHA}', "");
Expect(0, 7247, '\P{Blk:	_LEPCHA}', "");
Expect(1, 7247, '\P{^Blk:	_LEPCHA}', "");
Expect(0, 7248, '\p{Blk:	_LEPCHA}', "");
Expect(1, 7248, '\p{^Blk:	_LEPCHA}', "");
Expect(1, 7248, '\P{Blk:	_LEPCHA}', "");
Expect(0, 7248, '\P{^Blk:	_LEPCHA}', "");
Error('\p{Is_Block=:= Lepcha}');
Error('\P{Is_Block=:= Lepcha}');
Expect(1, 7247, '\p{Is_Block=lepcha}', "");
Expect(0, 7247, '\p{^Is_Block=lepcha}', "");
Expect(0, 7247, '\P{Is_Block=lepcha}', "");
Expect(1, 7247, '\P{^Is_Block=lepcha}', "");
Expect(0, 7248, '\p{Is_Block=lepcha}', "");
Expect(1, 7248, '\p{^Is_Block=lepcha}', "");
Expect(1, 7248, '\P{Is_Block=lepcha}', "");
Expect(0, 7248, '\P{^Is_Block=lepcha}', "");
Expect(1, 7247, '\p{Is_Block=_lepcha}', "");
Expect(0, 7247, '\p{^Is_Block=_lepcha}', "");
Expect(0, 7247, '\P{Is_Block=_lepcha}', "");
Expect(1, 7247, '\P{^Is_Block=_lepcha}', "");
Expect(0, 7248, '\p{Is_Block=_lepcha}', "");
Expect(1, 7248, '\p{^Is_Block=_lepcha}', "");
Expect(1, 7248, '\P{Is_Block=_lepcha}', "");
Expect(0, 7248, '\P{^Is_Block=_lepcha}', "");
Error('\p{Is_Blk=	:=LEPCHA}');
Error('\P{Is_Blk=	:=LEPCHA}');
Expect(1, 7247, '\p{Is_Blk=lepcha}', "");
Expect(0, 7247, '\p{^Is_Blk=lepcha}', "");
Expect(0, 7247, '\P{Is_Blk=lepcha}', "");
Expect(1, 7247, '\P{^Is_Blk=lepcha}', "");
Expect(0, 7248, '\p{Is_Blk=lepcha}', "");
Expect(1, 7248, '\p{^Is_Blk=lepcha}', "");
Expect(1, 7248, '\P{Is_Blk=lepcha}', "");
Expect(0, 7248, '\P{^Is_Blk=lepcha}', "");
Expect(1, 7247, '\p{Is_Blk=-Lepcha}', "");
Expect(0, 7247, '\p{^Is_Blk=-Lepcha}', "");
Expect(0, 7247, '\P{Is_Blk=-Lepcha}', "");
Expect(1, 7247, '\P{^Is_Blk=-Lepcha}', "");
Expect(0, 7248, '\p{Is_Blk=-Lepcha}', "");
Expect(1, 7248, '\p{^Is_Blk=-Lepcha}', "");
Expect(1, 7248, '\P{Is_Blk=-Lepcha}', "");
Expect(0, 7248, '\P{^Is_Blk=-Lepcha}', "");
Error('\p{Block= /a/Letterlike_symbols}');
Error('\P{Block= /a/Letterlike_symbols}');
Expect(1, 8527, '\p{Block=letterlikesymbols}', "");
Expect(0, 8527, '\p{^Block=letterlikesymbols}', "");
Expect(0, 8527, '\P{Block=letterlikesymbols}', "");
Expect(1, 8527, '\P{^Block=letterlikesymbols}', "");
Expect(0, 8528, '\p{Block=letterlikesymbols}', "");
Expect(1, 8528, '\p{^Block=letterlikesymbols}', "");
Expect(1, 8528, '\P{Block=letterlikesymbols}', "");
Expect(0, 8528, '\P{^Block=letterlikesymbols}', "");
Expect(1, 8527, '\p{Block:   __LETTERLIKE_SYMBOLS}', "");
Expect(0, 8527, '\p{^Block:   __LETTERLIKE_SYMBOLS}', "");
Expect(0, 8527, '\P{Block:   __LETTERLIKE_SYMBOLS}', "");
Expect(1, 8527, '\P{^Block:   __LETTERLIKE_SYMBOLS}', "");
Expect(0, 8528, '\p{Block:   __LETTERLIKE_SYMBOLS}', "");
Expect(1, 8528, '\p{^Block:   __LETTERLIKE_SYMBOLS}', "");
Expect(1, 8528, '\P{Block:   __LETTERLIKE_SYMBOLS}', "");
Expect(0, 8528, '\P{^Block:   __LETTERLIKE_SYMBOLS}', "");
Error('\p{Blk=_Letterlike_SYMBOLS:=}');
Error('\P{Blk=_Letterlike_SYMBOLS:=}');
Expect(1, 8527, '\p{Blk=letterlikesymbols}', "");
Expect(0, 8527, '\p{^Blk=letterlikesymbols}', "");
Expect(0, 8527, '\P{Blk=letterlikesymbols}', "");
Expect(1, 8527, '\P{^Blk=letterlikesymbols}', "");
Expect(0, 8528, '\p{Blk=letterlikesymbols}', "");
Expect(1, 8528, '\p{^Blk=letterlikesymbols}', "");
Expect(1, 8528, '\P{Blk=letterlikesymbols}', "");
Expect(0, 8528, '\P{^Blk=letterlikesymbols}', "");
Expect(1, 8527, '\p{Blk=__LETTERLIKE_Symbols}', "");
Expect(0, 8527, '\p{^Blk=__LETTERLIKE_Symbols}', "");
Expect(0, 8527, '\P{Blk=__LETTERLIKE_Symbols}', "");
Expect(1, 8527, '\P{^Blk=__LETTERLIKE_Symbols}', "");
Expect(0, 8528, '\p{Blk=__LETTERLIKE_Symbols}', "");
Expect(1, 8528, '\p{^Blk=__LETTERLIKE_Symbols}', "");
Expect(1, 8528, '\P{Blk=__LETTERLIKE_Symbols}', "");
Expect(0, 8528, '\P{^Blk=__LETTERLIKE_Symbols}', "");
Error('\p{Is_Block=  Letterlike_symbols/a/}');
Error('\P{Is_Block=  Letterlike_symbols/a/}');
Expect(1, 8527, '\p{Is_Block=letterlikesymbols}', "");
Expect(0, 8527, '\p{^Is_Block=letterlikesymbols}', "");
Expect(0, 8527, '\P{Is_Block=letterlikesymbols}', "");
Expect(1, 8527, '\P{^Is_Block=letterlikesymbols}', "");
Expect(0, 8528, '\p{Is_Block=letterlikesymbols}', "");
Expect(1, 8528, '\p{^Is_Block=letterlikesymbols}', "");
Expect(1, 8528, '\P{Is_Block=letterlikesymbols}', "");
Expect(0, 8528, '\P{^Is_Block=letterlikesymbols}', "");
Expect(1, 8527, '\p{Is_Block= -LETTERLIKE_Symbols}', "");
Expect(0, 8527, '\p{^Is_Block= -LETTERLIKE_Symbols}', "");
Expect(0, 8527, '\P{Is_Block= -LETTERLIKE_Symbols}', "");
Expect(1, 8527, '\P{^Is_Block= -LETTERLIKE_Symbols}', "");
Expect(0, 8528, '\p{Is_Block= -LETTERLIKE_Symbols}', "");
Expect(1, 8528, '\p{^Is_Block= -LETTERLIKE_Symbols}', "");
Expect(1, 8528, '\P{Is_Block= -LETTERLIKE_Symbols}', "");
Expect(0, 8528, '\P{^Is_Block= -LETTERLIKE_Symbols}', "");
Error('\p{Is_Blk=__Letterlike_Symbols/a/}');
Error('\P{Is_Blk=__Letterlike_Symbols/a/}');
Expect(1, 8527, '\p{Is_Blk=letterlikesymbols}', "");
Expect(0, 8527, '\p{^Is_Blk=letterlikesymbols}', "");
Expect(0, 8527, '\P{Is_Blk=letterlikesymbols}', "");
Expect(1, 8527, '\P{^Is_Blk=letterlikesymbols}', "");
Expect(0, 8528, '\p{Is_Blk=letterlikesymbols}', "");
Expect(1, 8528, '\p{^Is_Blk=letterlikesymbols}', "");
Expect(1, 8528, '\P{Is_Blk=letterlikesymbols}', "");
Expect(0, 8528, '\P{^Is_Blk=letterlikesymbols}', "");
Expect(1, 8527, '\p{Is_Blk=-_Letterlike_symbols}', "");
Expect(0, 8527, '\p{^Is_Blk=-_Letterlike_symbols}', "");
Expect(0, 8527, '\P{Is_Blk=-_Letterlike_symbols}', "");
Expect(1, 8527, '\P{^Is_Blk=-_Letterlike_symbols}', "");
Expect(0, 8528, '\p{Is_Blk=-_Letterlike_symbols}', "");
Expect(1, 8528, '\p{^Is_Blk=-_Letterlike_symbols}', "");
Expect(1, 8528, '\P{Is_Blk=-_Letterlike_symbols}', "");
Expect(0, 8528, '\P{^Is_Blk=-_Letterlike_symbols}', "");
Error('\p{Block=:= Limbu}');
Error('\P{Block=:= Limbu}');
Expect(1, 6479, '\p{Block=limbu}', "");
Expect(0, 6479, '\p{^Block=limbu}', "");
Expect(0, 6479, '\P{Block=limbu}', "");
Expect(1, 6479, '\P{^Block=limbu}', "");
Expect(0, 6480, '\p{Block=limbu}', "");
Expect(1, 6480, '\p{^Block=limbu}', "");
Expect(1, 6480, '\P{Block=limbu}', "");
Expect(0, 6480, '\P{^Block=limbu}', "");
Error('\p{Blk=	limbu/a/}');
Error('\P{Blk=	limbu/a/}');
Expect(1, 6479, '\p{Blk=limbu}', "");
Expect(0, 6479, '\p{^Blk=limbu}', "");
Expect(0, 6479, '\P{Blk=limbu}', "");
Expect(1, 6479, '\P{^Blk=limbu}', "");
Expect(0, 6480, '\p{Blk=limbu}', "");
Expect(1, 6480, '\p{^Blk=limbu}', "");
Expect(1, 6480, '\P{Blk=limbu}', "");
Expect(0, 6480, '\P{^Blk=limbu}', "");
Expect(1, 6479, '\p{Blk=-Limbu}', "");
Expect(0, 6479, '\p{^Blk=-Limbu}', "");
Expect(0, 6479, '\P{Blk=-Limbu}', "");
Expect(1, 6479, '\P{^Blk=-Limbu}', "");
Expect(0, 6480, '\p{Blk=-Limbu}', "");
Expect(1, 6480, '\p{^Blk=-Limbu}', "");
Expect(1, 6480, '\P{Blk=-Limbu}', "");
Expect(0, 6480, '\P{^Blk=-Limbu}', "");
Error('\p{Is_Block=:=-Limbu}');
Error('\P{Is_Block=:=-Limbu}');
Expect(1, 6479, '\p{Is_Block=limbu}', "");
Expect(0, 6479, '\p{^Is_Block=limbu}', "");
Expect(0, 6479, '\P{Is_Block=limbu}', "");
Expect(1, 6479, '\P{^Is_Block=limbu}', "");
Expect(0, 6480, '\p{Is_Block=limbu}', "");
Expect(1, 6480, '\p{^Is_Block=limbu}', "");
Expect(1, 6480, '\P{Is_Block=limbu}', "");
Expect(0, 6480, '\P{^Is_Block=limbu}', "");
Expect(1, 6479, '\p{Is_Block:   __LIMBU}', "");
Expect(0, 6479, '\p{^Is_Block:   __LIMBU}', "");
Expect(0, 6479, '\P{Is_Block:   __LIMBU}', "");
Expect(1, 6479, '\P{^Is_Block:   __LIMBU}', "");
Expect(0, 6480, '\p{Is_Block:   __LIMBU}', "");
Expect(1, 6480, '\p{^Is_Block:   __LIMBU}', "");
Expect(1, 6480, '\P{Is_Block:   __LIMBU}', "");
Expect(0, 6480, '\P{^Is_Block:   __LIMBU}', "");
Error('\p{Is_Blk=_	Limbu:=}');
Error('\P{Is_Blk=_	Limbu:=}');
Expect(1, 6479, '\p{Is_Blk=limbu}', "");
Expect(0, 6479, '\p{^Is_Blk=limbu}', "");
Expect(0, 6479, '\P{Is_Blk=limbu}', "");
Expect(1, 6479, '\P{^Is_Blk=limbu}', "");
Expect(0, 6480, '\p{Is_Blk=limbu}', "");
Expect(1, 6480, '\p{^Is_Blk=limbu}', "");
Expect(1, 6480, '\P{Is_Blk=limbu}', "");
Expect(0, 6480, '\P{^Is_Blk=limbu}', "");
Expect(1, 6479, '\p{Is_Blk=- Limbu}', "");
Expect(0, 6479, '\p{^Is_Blk=- Limbu}', "");
Expect(0, 6479, '\P{Is_Blk=- Limbu}', "");
Expect(1, 6479, '\P{^Is_Blk=- Limbu}', "");
Expect(0, 6480, '\p{Is_Blk=- Limbu}', "");
Expect(1, 6480, '\p{^Is_Blk=- Limbu}', "");
Expect(1, 6480, '\P{Is_Blk=- Limbu}', "");
Expect(0, 6480, '\P{^Is_Blk=- Limbu}', "");
Error('\p{Block=/a/- LINEAR_b_ideograms}');
Error('\P{Block=/a/- LINEAR_b_ideograms}');
Expect(1, 65791, '\p{Block:linearbideograms}', "");
Expect(0, 65791, '\p{^Block:linearbideograms}', "");
Expect(0, 65791, '\P{Block:linearbideograms}', "");
Expect(1, 65791, '\P{^Block:linearbideograms}', "");
Expect(0, 65792, '\p{Block:linearbideograms}', "");
Expect(1, 65792, '\p{^Block:linearbideograms}', "");
Expect(1, 65792, '\P{Block:linearbideograms}', "");
Expect(0, 65792, '\P{^Block:linearbideograms}', "");
Expect(1, 65791, '\p{Block=	-LINEAR_B_ideograms}', "");
Expect(0, 65791, '\p{^Block=	-LINEAR_B_ideograms}', "");
Expect(0, 65791, '\P{Block=	-LINEAR_B_ideograms}', "");
Expect(1, 65791, '\P{^Block=	-LINEAR_B_ideograms}', "");
Expect(0, 65792, '\p{Block=	-LINEAR_B_ideograms}', "");
Expect(1, 65792, '\p{^Block=	-LINEAR_B_ideograms}', "");
Expect(1, 65792, '\P{Block=	-LINEAR_B_ideograms}', "");
Expect(0, 65792, '\P{^Block=	-LINEAR_B_ideograms}', "");
Error('\p{Blk=-/a/linear_b_ideograms}');
Error('\P{Blk=-/a/linear_b_ideograms}');
Expect(1, 65791, '\p{Blk=linearbideograms}', "");
Expect(0, 65791, '\p{^Blk=linearbideograms}', "");
Expect(0, 65791, '\P{Blk=linearbideograms}', "");
Expect(1, 65791, '\P{^Blk=linearbideograms}', "");
Expect(0, 65792, '\p{Blk=linearbideograms}', "");
Expect(1, 65792, '\p{^Blk=linearbideograms}', "");
Expect(1, 65792, '\P{Blk=linearbideograms}', "");
Expect(0, 65792, '\P{^Blk=linearbideograms}', "");
Expect(1, 65791, '\p{Blk=	 Linear_b_Ideograms}', "");
Expect(0, 65791, '\p{^Blk=	 Linear_b_Ideograms}', "");
Expect(0, 65791, '\P{Blk=	 Linear_b_Ideograms}', "");
Expect(1, 65791, '\P{^Blk=	 Linear_b_Ideograms}', "");
Expect(0, 65792, '\p{Blk=	 Linear_b_Ideograms}', "");
Expect(1, 65792, '\p{^Blk=	 Linear_b_Ideograms}', "");
Expect(1, 65792, '\P{Blk=	 Linear_b_Ideograms}', "");
Expect(0, 65792, '\P{^Blk=	 Linear_b_Ideograms}', "");
Error('\p{Is_Block=_ Linear_b_Ideograms/a/}');
Error('\P{Is_Block=_ Linear_b_Ideograms/a/}');
Expect(1, 65791, '\p{Is_Block=linearbideograms}', "");
Expect(0, 65791, '\p{^Is_Block=linearbideograms}', "");
Expect(0, 65791, '\P{Is_Block=linearbideograms}', "");
Expect(1, 65791, '\P{^Is_Block=linearbideograms}', "");
Expect(0, 65792, '\p{Is_Block=linearbideograms}', "");
Expect(1, 65792, '\p{^Is_Block=linearbideograms}', "");
Expect(1, 65792, '\P{Is_Block=linearbideograms}', "");
Expect(0, 65792, '\P{^Is_Block=linearbideograms}', "");
Expect(1, 65791, '\p{Is_Block=-Linear_B_IDEOGRAMS}', "");
Expect(0, 65791, '\p{^Is_Block=-Linear_B_IDEOGRAMS}', "");
Expect(0, 65791, '\P{Is_Block=-Linear_B_IDEOGRAMS}', "");
Expect(1, 65791, '\P{^Is_Block=-Linear_B_IDEOGRAMS}', "");
Expect(0, 65792, '\p{Is_Block=-Linear_B_IDEOGRAMS}', "");
Expect(1, 65792, '\p{^Is_Block=-Linear_B_IDEOGRAMS}', "");
Expect(1, 65792, '\P{Is_Block=-Linear_B_IDEOGRAMS}', "");
Expect(0, 65792, '\P{^Is_Block=-Linear_B_IDEOGRAMS}', "");
Error('\p{Is_Blk:  	Linear_B_ideograms/a/}');
Error('\P{Is_Blk:  	Linear_B_ideograms/a/}');
Expect(1, 65791, '\p{Is_Blk=linearbideograms}', "");
Expect(0, 65791, '\p{^Is_Blk=linearbideograms}', "");
Expect(0, 65791, '\P{Is_Blk=linearbideograms}', "");
Expect(1, 65791, '\P{^Is_Blk=linearbideograms}', "");
Expect(0, 65792, '\p{Is_Blk=linearbideograms}', "");
Expect(1, 65792, '\p{^Is_Blk=linearbideograms}', "");
Expect(1, 65792, '\P{Is_Blk=linearbideograms}', "");
Expect(0, 65792, '\P{^Is_Blk=linearbideograms}', "");
Expect(1, 65791, '\p{Is_Blk=_LINEAR_B_Ideograms}', "");
Expect(0, 65791, '\p{^Is_Blk=_LINEAR_B_Ideograms}', "");
Expect(0, 65791, '\P{Is_Blk=_LINEAR_B_Ideograms}', "");
Expect(1, 65791, '\P{^Is_Blk=_LINEAR_B_Ideograms}', "");
Expect(0, 65792, '\p{Is_Blk=_LINEAR_B_Ideograms}', "");
Expect(1, 65792, '\p{^Is_Blk=_LINEAR_B_Ideograms}', "");
Expect(1, 65792, '\P{Is_Blk=_LINEAR_B_Ideograms}', "");
Expect(0, 65792, '\P{^Is_Blk=_LINEAR_B_Ideograms}', "");
Error('\p{Block=/a/_	Linear_b_syllabary}');
Error('\P{Block=/a/_	Linear_b_syllabary}');
Expect(1, 65663, '\p{Block=linearbsyllabary}', "");
Expect(0, 65663, '\p{^Block=linearbsyllabary}', "");
Expect(0, 65663, '\P{Block=linearbsyllabary}', "");
Expect(1, 65663, '\P{^Block=linearbsyllabary}', "");
Expect(0, 65664, '\p{Block=linearbsyllabary}', "");
Expect(1, 65664, '\p{^Block=linearbsyllabary}', "");
Expect(1, 65664, '\P{Block=linearbsyllabary}', "");
Expect(0, 65664, '\P{^Block=linearbsyllabary}', "");
Expect(1, 65663, '\p{Block=_linear_B_syllabary}', "");
Expect(0, 65663, '\p{^Block=_linear_B_syllabary}', "");
Expect(0, 65663, '\P{Block=_linear_B_syllabary}', "");
Expect(1, 65663, '\P{^Block=_linear_B_syllabary}', "");
Expect(0, 65664, '\p{Block=_linear_B_syllabary}', "");
Expect(1, 65664, '\p{^Block=_linear_B_syllabary}', "");
Expect(1, 65664, '\P{Block=_linear_B_syllabary}', "");
Expect(0, 65664, '\P{^Block=_linear_B_syllabary}', "");
Error('\p{Blk=/a/ linear_B_Syllabary}');
Error('\P{Blk=/a/ linear_B_Syllabary}');
Expect(1, 65663, '\p{Blk=linearbsyllabary}', "");
Expect(0, 65663, '\p{^Blk=linearbsyllabary}', "");
Expect(0, 65663, '\P{Blk=linearbsyllabary}', "");
Expect(1, 65663, '\P{^Blk=linearbsyllabary}', "");
Expect(0, 65664, '\p{Blk=linearbsyllabary}', "");
Expect(1, 65664, '\p{^Blk=linearbsyllabary}', "");
Expect(1, 65664, '\P{Blk=linearbsyllabary}', "");
Expect(0, 65664, '\P{^Blk=linearbsyllabary}', "");
Expect(1, 65663, '\p{Blk=_Linear_B_Syllabary}', "");
Expect(0, 65663, '\p{^Blk=_Linear_B_Syllabary}', "");
Expect(0, 65663, '\P{Blk=_Linear_B_Syllabary}', "");
Expect(1, 65663, '\P{^Blk=_Linear_B_Syllabary}', "");
Expect(0, 65664, '\p{Blk=_Linear_B_Syllabary}', "");
Expect(1, 65664, '\p{^Blk=_Linear_B_Syllabary}', "");
Expect(1, 65664, '\P{Blk=_Linear_B_Syllabary}', "");
Expect(0, 65664, '\P{^Blk=_Linear_B_Syllabary}', "");
Error('\p{Is_Block=- Linear_B_syllabary/a/}');
Error('\P{Is_Block=- Linear_B_syllabary/a/}');
Expect(1, 65663, '\p{Is_Block=linearbsyllabary}', "");
Expect(0, 65663, '\p{^Is_Block=linearbsyllabary}', "");
Expect(0, 65663, '\P{Is_Block=linearbsyllabary}', "");
Expect(1, 65663, '\P{^Is_Block=linearbsyllabary}', "");
Expect(0, 65664, '\p{Is_Block=linearbsyllabary}', "");
Expect(1, 65664, '\p{^Is_Block=linearbsyllabary}', "");
Expect(1, 65664, '\P{Is_Block=linearbsyllabary}', "");
Expect(0, 65664, '\P{^Is_Block=linearbsyllabary}', "");
Expect(1, 65663, '\p{Is_Block=	LINEAR_B_syllabary}', "");
Expect(0, 65663, '\p{^Is_Block=	LINEAR_B_syllabary}', "");
Expect(0, 65663, '\P{Is_Block=	LINEAR_B_syllabary}', "");
Expect(1, 65663, '\P{^Is_Block=	LINEAR_B_syllabary}', "");
Expect(0, 65664, '\p{Is_Block=	LINEAR_B_syllabary}', "");
Expect(1, 65664, '\p{^Is_Block=	LINEAR_B_syllabary}', "");
Expect(1, 65664, '\P{Is_Block=	LINEAR_B_syllabary}', "");
Expect(0, 65664, '\P{^Is_Block=	LINEAR_B_syllabary}', "");
Error('\p{Is_Blk=/a/ _Linear_B_SYLLABARY}');
Error('\P{Is_Blk=/a/ _Linear_B_SYLLABARY}');
Expect(1, 65663, '\p{Is_Blk:	linearbsyllabary}', "");
Expect(0, 65663, '\p{^Is_Blk:	linearbsyllabary}', "");
Expect(0, 65663, '\P{Is_Blk:	linearbsyllabary}', "");
Expect(1, 65663, '\P{^Is_Blk:	linearbsyllabary}', "");
Expect(0, 65664, '\p{Is_Blk:	linearbsyllabary}', "");
Expect(1, 65664, '\p{^Is_Blk:	linearbsyllabary}', "");
Expect(1, 65664, '\P{Is_Blk:	linearbsyllabary}', "");
Expect(0, 65664, '\P{^Is_Blk:	linearbsyllabary}', "");
Expect(1, 65663, '\p{Is_Blk=-Linear_B_syllabary}', "");
Expect(0, 65663, '\p{^Is_Blk=-Linear_B_syllabary}', "");
Expect(0, 65663, '\P{Is_Blk=-Linear_B_syllabary}', "");
Expect(1, 65663, '\P{^Is_Blk=-Linear_B_syllabary}', "");
Expect(0, 65664, '\p{Is_Blk=-Linear_B_syllabary}', "");
Expect(1, 65664, '\p{^Is_Blk=-Linear_B_syllabary}', "");
Expect(1, 65664, '\P{Is_Blk=-Linear_B_syllabary}', "");
Expect(0, 65664, '\P{^Is_Blk=-Linear_B_syllabary}', "");
Error('\p{Block=	-lisu/a/}');
Error('\P{Block=	-lisu/a/}');
Expect(1, 42239, '\p{Block=lisu}', "");
Expect(0, 42239, '\p{^Block=lisu}', "");
Expect(0, 42239, '\P{Block=lisu}', "");
Expect(1, 42239, '\P{^Block=lisu}', "");
Expect(0, 42240, '\p{Block=lisu}', "");
Expect(1, 42240, '\p{^Block=lisu}', "");
Expect(1, 42240, '\P{Block=lisu}', "");
Expect(0, 42240, '\P{^Block=lisu}', "");
Expect(1, 42239, '\p{Block=-Lisu}', "");
Expect(0, 42239, '\p{^Block=-Lisu}', "");
Expect(0, 42239, '\P{Block=-Lisu}', "");
Expect(1, 42239, '\P{^Block=-Lisu}', "");
Expect(0, 42240, '\p{Block=-Lisu}', "");
Expect(1, 42240, '\p{^Block=-Lisu}', "");
Expect(1, 42240, '\P{Block=-Lisu}', "");
Expect(0, 42240, '\P{^Block=-Lisu}', "");
Error('\p{Blk=-:=lisu}');
Error('\P{Blk=-:=lisu}');
Expect(1, 42239, '\p{Blk:lisu}', "");
Expect(0, 42239, '\p{^Blk:lisu}', "");
Expect(0, 42239, '\P{Blk:lisu}', "");
Expect(1, 42239, '\P{^Blk:lisu}', "");
Expect(0, 42240, '\p{Blk:lisu}', "");
Expect(1, 42240, '\p{^Blk:lisu}', "");
Expect(1, 42240, '\P{Blk:lisu}', "");
Expect(0, 42240, '\P{^Blk:lisu}', "");
Expect(1, 42239, '\p{Blk= -LISU}', "");
Expect(0, 42239, '\p{^Blk= -LISU}', "");
Expect(0, 42239, '\P{Blk= -LISU}', "");
Expect(1, 42239, '\P{^Blk= -LISU}', "");
Expect(0, 42240, '\p{Blk= -LISU}', "");
Expect(1, 42240, '\p{^Blk= -LISU}', "");
Expect(1, 42240, '\P{Blk= -LISU}', "");
Expect(0, 42240, '\P{^Blk= -LISU}', "");
Error('\p{Is_Block=:= _Lisu}');
Error('\P{Is_Block=:= _Lisu}');
Expect(1, 42239, '\p{Is_Block:lisu}', "");
Expect(0, 42239, '\p{^Is_Block:lisu}', "");
Expect(0, 42239, '\P{Is_Block:lisu}', "");
Expect(1, 42239, '\P{^Is_Block:lisu}', "");
Expect(0, 42240, '\p{Is_Block:lisu}', "");
Expect(1, 42240, '\p{^Is_Block:lisu}', "");
Expect(1, 42240, '\P{Is_Block:lisu}', "");
Expect(0, 42240, '\P{^Is_Block:lisu}', "");
Expect(1, 42239, '\p{Is_Block=_-Lisu}', "");
Expect(0, 42239, '\p{^Is_Block=_-Lisu}', "");
Expect(0, 42239, '\P{Is_Block=_-Lisu}', "");
Expect(1, 42239, '\P{^Is_Block=_-Lisu}', "");
Expect(0, 42240, '\p{Is_Block=_-Lisu}', "");
Expect(1, 42240, '\p{^Is_Block=_-Lisu}', "");
Expect(1, 42240, '\P{Is_Block=_-Lisu}', "");
Expect(0, 42240, '\P{^Is_Block=_-Lisu}', "");
Error('\p{Is_Blk=:=_	Lisu}');
Error('\P{Is_Blk=:=_	Lisu}');
Expect(1, 42239, '\p{Is_Blk=lisu}', "");
Expect(0, 42239, '\p{^Is_Blk=lisu}', "");
Expect(0, 42239, '\P{Is_Blk=lisu}', "");
Expect(1, 42239, '\P{^Is_Blk=lisu}', "");
Expect(0, 42240, '\p{Is_Blk=lisu}', "");
Expect(1, 42240, '\p{^Is_Blk=lisu}', "");
Expect(1, 42240, '\P{Is_Blk=lisu}', "");
Expect(0, 42240, '\P{^Is_Blk=lisu}', "");
Expect(1, 42239, '\p{Is_Blk:   _ Lisu}', "");
Expect(0, 42239, '\p{^Is_Blk:   _ Lisu}', "");
Expect(0, 42239, '\P{Is_Blk:   _ Lisu}', "");
Expect(1, 42239, '\P{^Is_Blk:   _ Lisu}', "");
Expect(0, 42240, '\p{Is_Blk:   _ Lisu}', "");
Expect(1, 42240, '\p{^Is_Blk:   _ Lisu}', "");
Expect(1, 42240, '\P{Is_Blk:   _ Lisu}', "");
Expect(0, 42240, '\P{^Is_Blk:   _ Lisu}', "");
Error('\p{Block=-	Low_Surrogates:=}');
Error('\P{Block=-	Low_Surrogates:=}');
Expect(1, 57343, '\p{Block:   lowsurrogates}', "");
Expect(0, 57343, '\p{^Block:   lowsurrogates}', "");
Expect(0, 57343, '\P{Block:   lowsurrogates}', "");
Expect(1, 57343, '\P{^Block:   lowsurrogates}', "");
Expect(0, 57344, '\p{Block:   lowsurrogates}', "");
Expect(1, 57344, '\p{^Block:   lowsurrogates}', "");
Expect(1, 57344, '\P{Block:   lowsurrogates}', "");
Expect(0, 57344, '\P{^Block:   lowsurrogates}', "");
Expect(1, 57343, '\p{Block=	-Low_SURROGATES}', "");
Expect(0, 57343, '\p{^Block=	-Low_SURROGATES}', "");
Expect(0, 57343, '\P{Block=	-Low_SURROGATES}', "");
Expect(1, 57343, '\P{^Block=	-Low_SURROGATES}', "");
Expect(0, 57344, '\p{Block=	-Low_SURROGATES}', "");
Expect(1, 57344, '\p{^Block=	-Low_SURROGATES}', "");
Expect(1, 57344, '\P{Block=	-Low_SURROGATES}', "");
Expect(0, 57344, '\P{^Block=	-Low_SURROGATES}', "");
Error('\p{Blk= :=LOW_Surrogates}');
Error('\P{Blk= :=LOW_Surrogates}');
Expect(1, 57343, '\p{Blk=lowsurrogates}', "");
Expect(0, 57343, '\p{^Blk=lowsurrogates}', "");
Expect(0, 57343, '\P{Blk=lowsurrogates}', "");
Expect(1, 57343, '\P{^Blk=lowsurrogates}', "");
Expect(0, 57344, '\p{Blk=lowsurrogates}', "");
Expect(1, 57344, '\p{^Blk=lowsurrogates}', "");
Expect(1, 57344, '\P{Blk=lowsurrogates}', "");
Expect(0, 57344, '\P{^Blk=lowsurrogates}', "");
Expect(1, 57343, '\p{Blk=-_Low_surrogates}', "");
Expect(0, 57343, '\p{^Blk=-_Low_surrogates}', "");
Expect(0, 57343, '\P{Blk=-_Low_surrogates}', "");
Expect(1, 57343, '\P{^Blk=-_Low_surrogates}', "");
Expect(0, 57344, '\p{Blk=-_Low_surrogates}', "");
Expect(1, 57344, '\p{^Blk=-_Low_surrogates}', "");
Expect(1, 57344, '\P{Blk=-_Low_surrogates}', "");
Expect(0, 57344, '\P{^Blk=-_Low_surrogates}', "");
Error('\p{Is_Block=	Low_Surrogates/a/}');
Error('\P{Is_Block=	Low_Surrogates/a/}');
Expect(1, 57343, '\p{Is_Block=lowsurrogates}', "");
Expect(0, 57343, '\p{^Is_Block=lowsurrogates}', "");
Expect(0, 57343, '\P{Is_Block=lowsurrogates}', "");
Expect(1, 57343, '\P{^Is_Block=lowsurrogates}', "");
Expect(0, 57344, '\p{Is_Block=lowsurrogates}', "");
Expect(1, 57344, '\p{^Is_Block=lowsurrogates}', "");
Expect(1, 57344, '\P{Is_Block=lowsurrogates}', "");
Expect(0, 57344, '\P{^Is_Block=lowsurrogates}', "");
Expect(1, 57343, '\p{Is_Block=-low_Surrogates}', "");
Expect(0, 57343, '\p{^Is_Block=-low_Surrogates}', "");
Expect(0, 57343, '\P{Is_Block=-low_Surrogates}', "");
Expect(1, 57343, '\P{^Is_Block=-low_Surrogates}', "");
Expect(0, 57344, '\p{Is_Block=-low_Surrogates}', "");
Expect(1, 57344, '\p{^Is_Block=-low_Surrogates}', "");
Expect(1, 57344, '\P{Is_Block=-low_Surrogates}', "");
Expect(0, 57344, '\P{^Is_Block=-low_Surrogates}', "");
Error('\p{Is_Blk=-/a/low_Surrogates}');
Error('\P{Is_Blk=-/a/low_Surrogates}');
Expect(1, 57343, '\p{Is_Blk=lowsurrogates}', "");
Expect(0, 57343, '\p{^Is_Blk=lowsurrogates}', "");
Expect(0, 57343, '\P{Is_Blk=lowsurrogates}', "");
Expect(1, 57343, '\P{^Is_Blk=lowsurrogates}', "");
Expect(0, 57344, '\p{Is_Blk=lowsurrogates}', "");
Expect(1, 57344, '\p{^Is_Blk=lowsurrogates}', "");
Expect(1, 57344, '\P{Is_Blk=lowsurrogates}', "");
Expect(0, 57344, '\P{^Is_Blk=lowsurrogates}', "");
Expect(1, 57343, '\p{Is_Blk:   _ Low_Surrogates}', "");
Expect(0, 57343, '\p{^Is_Blk:   _ Low_Surrogates}', "");
Expect(0, 57343, '\P{Is_Blk:   _ Low_Surrogates}', "");
Expect(1, 57343, '\P{^Is_Blk:   _ Low_Surrogates}', "");
Expect(0, 57344, '\p{Is_Blk:   _ Low_Surrogates}', "");
Expect(1, 57344, '\p{^Is_Blk:   _ Low_Surrogates}', "");
Expect(1, 57344, '\P{Is_Blk:   _ Low_Surrogates}', "");
Expect(0, 57344, '\P{^Is_Blk:   _ Low_Surrogates}', "");
Error('\p{Block=_lycian:=}');
Error('\P{Block=_lycian:=}');
Expect(1, 66207, '\p{Block=lycian}', "");
Expect(0, 66207, '\p{^Block=lycian}', "");
Expect(0, 66207, '\P{Block=lycian}', "");
Expect(1, 66207, '\P{^Block=lycian}', "");
Expect(0, 66208, '\p{Block=lycian}', "");
Expect(1, 66208, '\p{^Block=lycian}', "");
Expect(1, 66208, '\P{Block=lycian}', "");
Expect(0, 66208, '\P{^Block=lycian}', "");
Expect(1, 66207, '\p{Block=_	LYCIAN}', "");
Expect(0, 66207, '\p{^Block=_	LYCIAN}', "");
Expect(0, 66207, '\P{Block=_	LYCIAN}', "");
Expect(1, 66207, '\P{^Block=_	LYCIAN}', "");
Expect(0, 66208, '\p{Block=_	LYCIAN}', "");
Expect(1, 66208, '\p{^Block=_	LYCIAN}', "");
Expect(1, 66208, '\P{Block=_	LYCIAN}', "");
Expect(0, 66208, '\P{^Block=_	LYCIAN}', "");
Error('\p{Blk=lycian/a/}');
Error('\P{Blk=lycian/a/}');
Expect(1, 66207, '\p{Blk=lycian}', "");
Expect(0, 66207, '\p{^Blk=lycian}', "");
Expect(0, 66207, '\P{Blk=lycian}', "");
Expect(1, 66207, '\P{^Blk=lycian}', "");
Expect(0, 66208, '\p{Blk=lycian}', "");
Expect(1, 66208, '\p{^Blk=lycian}', "");
Expect(1, 66208, '\P{Blk=lycian}', "");
Expect(0, 66208, '\P{^Blk=lycian}', "");
Expect(1, 66207, '\p{Blk=--lycian}', "");
Expect(0, 66207, '\p{^Blk=--lycian}', "");
Expect(0, 66207, '\P{Blk=--lycian}', "");
Expect(1, 66207, '\P{^Blk=--lycian}', "");
Expect(0, 66208, '\p{Blk=--lycian}', "");
Expect(1, 66208, '\p{^Blk=--lycian}', "");
Expect(1, 66208, '\P{Blk=--lycian}', "");
Expect(0, 66208, '\P{^Blk=--lycian}', "");
Error('\p{Is_Block=__Lycian:=}');
Error('\P{Is_Block=__Lycian:=}');
Expect(1, 66207, '\p{Is_Block=lycian}', "");
Expect(0, 66207, '\p{^Is_Block=lycian}', "");
Expect(0, 66207, '\P{Is_Block=lycian}', "");
Expect(1, 66207, '\P{^Is_Block=lycian}', "");
Expect(0, 66208, '\p{Is_Block=lycian}', "");
Expect(1, 66208, '\p{^Is_Block=lycian}', "");
Expect(1, 66208, '\P{Is_Block=lycian}', "");
Expect(0, 66208, '\P{^Is_Block=lycian}', "");
Expect(1, 66207, '\p{Is_Block= Lycian}', "");
Expect(0, 66207, '\p{^Is_Block= Lycian}', "");
Expect(0, 66207, '\P{Is_Block= Lycian}', "");
Expect(1, 66207, '\P{^Is_Block= Lycian}', "");
Expect(0, 66208, '\p{Is_Block= Lycian}', "");
Expect(1, 66208, '\p{^Is_Block= Lycian}', "");
Expect(1, 66208, '\P{Is_Block= Lycian}', "");
Expect(0, 66208, '\P{^Is_Block= Lycian}', "");
Error('\p{Is_Blk=:=_Lycian}');
Error('\P{Is_Blk=:=_Lycian}');
Expect(1, 66207, '\p{Is_Blk=lycian}', "");
Expect(0, 66207, '\p{^Is_Blk=lycian}', "");
Expect(0, 66207, '\P{Is_Blk=lycian}', "");
Expect(1, 66207, '\P{^Is_Blk=lycian}', "");
Expect(0, 66208, '\p{Is_Blk=lycian}', "");
Expect(1, 66208, '\p{^Is_Blk=lycian}', "");
Expect(1, 66208, '\P{Is_Blk=lycian}', "");
Expect(0, 66208, '\P{^Is_Blk=lycian}', "");
Expect(1, 66207, '\p{Is_Blk=-	Lycian}', "");
Expect(0, 66207, '\p{^Is_Blk=-	Lycian}', "");
Expect(0, 66207, '\P{Is_Blk=-	Lycian}', "");
Expect(1, 66207, '\P{^Is_Blk=-	Lycian}', "");
Expect(0, 66208, '\p{Is_Blk=-	Lycian}', "");
Expect(1, 66208, '\p{^Is_Blk=-	Lycian}', "");
Expect(1, 66208, '\P{Is_Blk=-	Lycian}', "");
Expect(0, 66208, '\P{^Is_Blk=-	Lycian}', "");
Error('\p{Block= :=lydian}');
Error('\P{Block= :=lydian}');
Expect(1, 67903, '\p{Block=lydian}', "");
Expect(0, 67903, '\p{^Block=lydian}', "");
Expect(0, 67903, '\P{Block=lydian}', "");
Expect(1, 67903, '\P{^Block=lydian}', "");
Expect(0, 67904, '\p{Block=lydian}', "");
Expect(1, 67904, '\p{^Block=lydian}', "");
Expect(1, 67904, '\P{Block=lydian}', "");
Expect(0, 67904, '\P{^Block=lydian}', "");
Expect(1, 67903, '\p{Block=	Lydian}', "");
Expect(0, 67903, '\p{^Block=	Lydian}', "");
Expect(0, 67903, '\P{Block=	Lydian}', "");
Expect(1, 67903, '\P{^Block=	Lydian}', "");
Expect(0, 67904, '\p{Block=	Lydian}', "");
Expect(1, 67904, '\p{^Block=	Lydian}', "");
Expect(1, 67904, '\P{Block=	Lydian}', "");
Expect(0, 67904, '\P{^Block=	Lydian}', "");
Error('\p{Blk=:=_-Lydian}');
Error('\P{Blk=:=_-Lydian}');
Expect(1, 67903, '\p{Blk: lydian}', "");
Expect(0, 67903, '\p{^Blk: lydian}', "");
Expect(0, 67903, '\P{Blk: lydian}', "");
Expect(1, 67903, '\P{^Blk: lydian}', "");
Expect(0, 67904, '\p{Blk: lydian}', "");
Expect(1, 67904, '\p{^Blk: lydian}', "");
Expect(1, 67904, '\P{Blk: lydian}', "");
Expect(0, 67904, '\P{^Blk: lydian}', "");
Expect(1, 67903, '\p{Blk=	-Lydian}', "");
Expect(0, 67903, '\p{^Blk=	-Lydian}', "");
Expect(0, 67903, '\P{Blk=	-Lydian}', "");
Expect(1, 67903, '\P{^Blk=	-Lydian}', "");
Expect(0, 67904, '\p{Blk=	-Lydian}', "");
Expect(1, 67904, '\p{^Blk=	-Lydian}', "");
Expect(1, 67904, '\P{Blk=	-Lydian}', "");
Expect(0, 67904, '\P{^Blk=	-Lydian}', "");
Error('\p{Is_Block=	/a/Lydian}');
Error('\P{Is_Block=	/a/Lydian}');
Expect(1, 67903, '\p{Is_Block=lydian}', "");
Expect(0, 67903, '\p{^Is_Block=lydian}', "");
Expect(0, 67903, '\P{Is_Block=lydian}', "");
Expect(1, 67903, '\P{^Is_Block=lydian}', "");
Expect(0, 67904, '\p{Is_Block=lydian}', "");
Expect(1, 67904, '\p{^Is_Block=lydian}', "");
Expect(1, 67904, '\P{Is_Block=lydian}', "");
Expect(0, 67904, '\P{^Is_Block=lydian}', "");
Expect(1, 67903, '\p{Is_Block=-	lydian}', "");
Expect(0, 67903, '\p{^Is_Block=-	lydian}', "");
Expect(0, 67903, '\P{Is_Block=-	lydian}', "");
Expect(1, 67903, '\P{^Is_Block=-	lydian}', "");
Expect(0, 67904, '\p{Is_Block=-	lydian}', "");
Expect(1, 67904, '\p{^Is_Block=-	lydian}', "");
Expect(1, 67904, '\P{Is_Block=-	lydian}', "");
Expect(0, 67904, '\P{^Is_Block=-	lydian}', "");
Error('\p{Is_Blk=/a/_Lydian}');
Error('\P{Is_Blk=/a/_Lydian}');
Expect(1, 67903, '\p{Is_Blk:	lydian}', "");
Expect(0, 67903, '\p{^Is_Blk:	lydian}', "");
Expect(0, 67903, '\P{Is_Blk:	lydian}', "");
Expect(1, 67903, '\P{^Is_Blk:	lydian}', "");
Expect(0, 67904, '\p{Is_Blk:	lydian}', "");
Expect(1, 67904, '\p{^Is_Blk:	lydian}', "");
Expect(1, 67904, '\P{Is_Blk:	lydian}', "");
Expect(0, 67904, '\P{^Is_Blk:	lydian}', "");
Expect(1, 67903, '\p{Is_Blk=	_Lydian}', "");
Expect(0, 67903, '\p{^Is_Blk=	_Lydian}', "");
Expect(0, 67903, '\P{Is_Blk=	_Lydian}', "");
Expect(1, 67903, '\P{^Is_Blk=	_Lydian}', "");
Expect(0, 67904, '\p{Is_Blk=	_Lydian}', "");
Expect(1, 67904, '\p{^Is_Blk=	_Lydian}', "");
Expect(1, 67904, '\P{Is_Blk=	_Lydian}', "");
Expect(0, 67904, '\P{^Is_Blk=	_Lydian}', "");
Error('\p{Block=:=-Mahjong_tiles}');
Error('\P{Block=:=-Mahjong_tiles}');
Expect(1, 127023, '\p{Block=mahjongtiles}', "");
Expect(0, 127023, '\p{^Block=mahjongtiles}', "");
Expect(0, 127023, '\P{Block=mahjongtiles}', "");
Expect(1, 127023, '\P{^Block=mahjongtiles}', "");
Expect(0, 127024, '\p{Block=mahjongtiles}', "");
Expect(1, 127024, '\p{^Block=mahjongtiles}', "");
Expect(1, 127024, '\P{Block=mahjongtiles}', "");
Expect(0, 127024, '\P{^Block=mahjongtiles}', "");
Expect(1, 127023, '\p{Block=--Mahjong_Tiles}', "");
Expect(0, 127023, '\p{^Block=--Mahjong_Tiles}', "");
Expect(0, 127023, '\P{Block=--Mahjong_Tiles}', "");
Expect(1, 127023, '\P{^Block=--Mahjong_Tiles}', "");
Expect(0, 127024, '\p{Block=--Mahjong_Tiles}', "");
Expect(1, 127024, '\p{^Block=--Mahjong_Tiles}', "");
Expect(1, 127024, '\P{Block=--Mahjong_Tiles}', "");
Expect(0, 127024, '\P{^Block=--Mahjong_Tiles}', "");
Error('\p{Blk= -mahjong/a/}');
Error('\P{Blk= -mahjong/a/}');
Expect(1, 127023, '\p{Blk=mahjong}', "");
Expect(0, 127023, '\p{^Blk=mahjong}', "");
Expect(0, 127023, '\P{Blk=mahjong}', "");
Expect(1, 127023, '\P{^Blk=mahjong}', "");
Expect(0, 127024, '\p{Blk=mahjong}', "");
Expect(1, 127024, '\p{^Blk=mahjong}', "");
Expect(1, 127024, '\P{Blk=mahjong}', "");
Expect(0, 127024, '\P{^Blk=mahjong}', "");
Expect(1, 127023, '\p{Blk=_Mahjong}', "");
Expect(0, 127023, '\p{^Blk=_Mahjong}', "");
Expect(0, 127023, '\P{Blk=_Mahjong}', "");
Expect(1, 127023, '\P{^Blk=_Mahjong}', "");
Expect(0, 127024, '\p{Blk=_Mahjong}', "");
Expect(1, 127024, '\p{^Blk=_Mahjong}', "");
Expect(1, 127024, '\P{Blk=_Mahjong}', "");
Expect(0, 127024, '\P{^Blk=_Mahjong}', "");
Error('\p{Is_Block=:=_	Mahjong_Tiles}');
Error('\P{Is_Block=:=_	Mahjong_Tiles}');
Expect(1, 127023, '\p{Is_Block=mahjongtiles}', "");
Expect(0, 127023, '\p{^Is_Block=mahjongtiles}', "");
Expect(0, 127023, '\P{Is_Block=mahjongtiles}', "");
Expect(1, 127023, '\P{^Is_Block=mahjongtiles}', "");
Expect(0, 127024, '\p{Is_Block=mahjongtiles}', "");
Expect(1, 127024, '\p{^Is_Block=mahjongtiles}', "");
Expect(1, 127024, '\P{Is_Block=mahjongtiles}', "");
Expect(0, 127024, '\P{^Is_Block=mahjongtiles}', "");
Expect(1, 127023, '\p{Is_Block=-mahjong_tiles}', "");
Expect(0, 127023, '\p{^Is_Block=-mahjong_tiles}', "");
Expect(0, 127023, '\P{Is_Block=-mahjong_tiles}', "");
Expect(1, 127023, '\P{^Is_Block=-mahjong_tiles}', "");
Expect(0, 127024, '\p{Is_Block=-mahjong_tiles}', "");
Expect(1, 127024, '\p{^Is_Block=-mahjong_tiles}', "");
Expect(1, 127024, '\P{Is_Block=-mahjong_tiles}', "");
Expect(0, 127024, '\P{^Is_Block=-mahjong_tiles}', "");
Error('\p{Is_Blk=-/a/Mahjong}');
Error('\P{Is_Blk=-/a/Mahjong}');
Expect(1, 127023, '\p{Is_Blk=mahjong}', "");
Expect(0, 127023, '\p{^Is_Blk=mahjong}', "");
Expect(0, 127023, '\P{Is_Blk=mahjong}', "");
Expect(1, 127023, '\P{^Is_Blk=mahjong}', "");
Expect(0, 127024, '\p{Is_Blk=mahjong}', "");
Expect(1, 127024, '\p{^Is_Blk=mahjong}', "");
Expect(1, 127024, '\P{Is_Blk=mahjong}', "");
Expect(0, 127024, '\P{^Is_Blk=mahjong}', "");
Expect(1, 127023, '\p{Is_Blk= -Mahjong}', "");
Expect(0, 127023, '\p{^Is_Blk= -Mahjong}', "");
Expect(0, 127023, '\P{Is_Blk= -Mahjong}', "");
Expect(1, 127023, '\P{^Is_Blk= -Mahjong}', "");
Expect(0, 127024, '\p{Is_Blk= -Mahjong}', "");
Expect(1, 127024, '\p{^Is_Blk= -Mahjong}', "");
Expect(1, 127024, '\P{Is_Blk= -Mahjong}', "");
Expect(0, 127024, '\P{^Is_Blk= -Mahjong}', "");
Error('\p{Block=_-Malayalam:=}');
Error('\P{Block=_-Malayalam:=}');
Expect(1, 3455, '\p{Block=malayalam}', "");
Expect(0, 3455, '\p{^Block=malayalam}', "");
Expect(0, 3455, '\P{Block=malayalam}', "");
Expect(1, 3455, '\P{^Block=malayalam}', "");
Expect(0, 3456, '\p{Block=malayalam}', "");
Expect(1, 3456, '\p{^Block=malayalam}', "");
Expect(1, 3456, '\P{Block=malayalam}', "");
Expect(0, 3456, '\P{^Block=malayalam}', "");
Expect(1, 3455, '\p{Block=-Malayalam}', "");
Expect(0, 3455, '\p{^Block=-Malayalam}', "");
Expect(0, 3455, '\P{Block=-Malayalam}', "");
Expect(1, 3455, '\P{^Block=-Malayalam}', "");
Expect(0, 3456, '\p{Block=-Malayalam}', "");
Expect(1, 3456, '\p{^Block=-Malayalam}', "");
Expect(1, 3456, '\P{Block=-Malayalam}', "");
Expect(0, 3456, '\P{^Block=-Malayalam}', "");
Error('\p{Blk=:=_-Malayalam}');
Error('\P{Blk=:=_-Malayalam}');
Expect(1, 3455, '\p{Blk=malayalam}', "");
Expect(0, 3455, '\p{^Blk=malayalam}', "");
Expect(0, 3455, '\P{Blk=malayalam}', "");
Expect(1, 3455, '\P{^Blk=malayalam}', "");
Expect(0, 3456, '\p{Blk=malayalam}', "");
Expect(1, 3456, '\p{^Blk=malayalam}', "");
Expect(1, 3456, '\P{Blk=malayalam}', "");
Expect(0, 3456, '\P{^Blk=malayalam}', "");
Expect(1, 3455, '\p{Blk=	-Malayalam}', "");
Expect(0, 3455, '\p{^Blk=	-Malayalam}', "");
Expect(0, 3455, '\P{Blk=	-Malayalam}', "");
Expect(1, 3455, '\P{^Blk=	-Malayalam}', "");
Expect(0, 3456, '\p{Blk=	-Malayalam}', "");
Expect(1, 3456, '\p{^Blk=	-Malayalam}', "");
Expect(1, 3456, '\P{Blk=	-Malayalam}', "");
Expect(0, 3456, '\P{^Blk=	-Malayalam}', "");
Error('\p{Is_Block= 	MALAYALAM:=}');
Error('\P{Is_Block= 	MALAYALAM:=}');
Expect(1, 3455, '\p{Is_Block=malayalam}', "");
Expect(0, 3455, '\p{^Is_Block=malayalam}', "");
Expect(0, 3455, '\P{Is_Block=malayalam}', "");
Expect(1, 3455, '\P{^Is_Block=malayalam}', "");
Expect(0, 3456, '\p{Is_Block=malayalam}', "");
Expect(1, 3456, '\p{^Is_Block=malayalam}', "");
Expect(1, 3456, '\P{Is_Block=malayalam}', "");
Expect(0, 3456, '\P{^Is_Block=malayalam}', "");
Expect(1, 3455, '\p{Is_Block=MALAYALAM}', "");
Expect(0, 3455, '\p{^Is_Block=MALAYALAM}', "");
Expect(0, 3455, '\P{Is_Block=MALAYALAM}', "");
Expect(1, 3455, '\P{^Is_Block=MALAYALAM}', "");
Expect(0, 3456, '\p{Is_Block=MALAYALAM}', "");
Expect(1, 3456, '\p{^Is_Block=MALAYALAM}', "");
Expect(1, 3456, '\P{Is_Block=MALAYALAM}', "");
Expect(0, 3456, '\P{^Is_Block=MALAYALAM}', "");
Error('\p{Is_Blk=-/a/malayalam}');
Error('\P{Is_Blk=-/a/malayalam}');
Expect(1, 3455, '\p{Is_Blk=malayalam}', "");
Expect(0, 3455, '\p{^Is_Blk=malayalam}', "");
Expect(0, 3455, '\P{Is_Blk=malayalam}', "");
Expect(1, 3455, '\P{^Is_Blk=malayalam}', "");
Expect(0, 3456, '\p{Is_Blk=malayalam}', "");
Expect(1, 3456, '\p{^Is_Blk=malayalam}', "");
Expect(1, 3456, '\P{Is_Blk=malayalam}', "");
Expect(0, 3456, '\P{^Is_Blk=malayalam}', "");
Expect(1, 3455, '\p{Is_Blk= Malayalam}', "");
Expect(0, 3455, '\p{^Is_Blk= Malayalam}', "");
Expect(0, 3455, '\P{Is_Blk= Malayalam}', "");
Expect(1, 3455, '\P{^Is_Blk= Malayalam}', "");
Expect(0, 3456, '\p{Is_Blk= Malayalam}', "");
Expect(1, 3456, '\p{^Is_Blk= Malayalam}', "");
Expect(1, 3456, '\P{Is_Blk= Malayalam}', "");
Expect(0, 3456, '\P{^Is_Blk= Malayalam}', "");
Error('\p{Block=/a/-_MANDAIC}');
Error('\P{Block=/a/-_MANDAIC}');
Expect(1, 2143, '\p{Block=mandaic}', "");
Expect(0, 2143, '\p{^Block=mandaic}', "");
Expect(0, 2143, '\P{Block=mandaic}', "");
Expect(1, 2143, '\P{^Block=mandaic}', "");
Expect(0, 2144, '\p{Block=mandaic}', "");
Expect(1, 2144, '\p{^Block=mandaic}', "");
Expect(1, 2144, '\P{Block=mandaic}', "");
Expect(0, 2144, '\P{^Block=mandaic}', "");
Expect(1, 2143, '\p{Block= mandaic}', "");
Expect(0, 2143, '\p{^Block= mandaic}', "");
Expect(0, 2143, '\P{Block= mandaic}', "");
Expect(1, 2143, '\P{^Block= mandaic}', "");
Expect(0, 2144, '\p{Block= mandaic}', "");
Expect(1, 2144, '\p{^Block= mandaic}', "");
Expect(1, 2144, '\P{Block= mandaic}', "");
Expect(0, 2144, '\P{^Block= mandaic}', "");
Error('\p{Blk=	/a/Mandaic}');
Error('\P{Blk=	/a/Mandaic}');
Expect(1, 2143, '\p{Blk:mandaic}', "");
Expect(0, 2143, '\p{^Blk:mandaic}', "");
Expect(0, 2143, '\P{Blk:mandaic}', "");
Expect(1, 2143, '\P{^Blk:mandaic}', "");
Expect(0, 2144, '\p{Blk:mandaic}', "");
Expect(1, 2144, '\p{^Blk:mandaic}', "");
Expect(1, 2144, '\P{Blk:mandaic}', "");
Expect(0, 2144, '\P{^Blk:mandaic}', "");
Expect(1, 2143, '\p{Blk= Mandaic}', "");
Expect(0, 2143, '\p{^Blk= Mandaic}', "");
Expect(0, 2143, '\P{Blk= Mandaic}', "");
Expect(1, 2143, '\P{^Blk= Mandaic}', "");
Expect(0, 2144, '\p{Blk= Mandaic}', "");
Expect(1, 2144, '\p{^Blk= Mandaic}', "");
Expect(1, 2144, '\P{Blk= Mandaic}', "");
Expect(0, 2144, '\P{^Blk= Mandaic}', "");
Error('\p{Is_Block=/a/ Mandaic}');
Error('\P{Is_Block=/a/ Mandaic}');
Expect(1, 2143, '\p{Is_Block=mandaic}', "");
Expect(0, 2143, '\p{^Is_Block=mandaic}', "");
Expect(0, 2143, '\P{Is_Block=mandaic}', "");
Expect(1, 2143, '\P{^Is_Block=mandaic}', "");
Expect(0, 2144, '\p{Is_Block=mandaic}', "");
Expect(1, 2144, '\p{^Is_Block=mandaic}', "");
Expect(1, 2144, '\P{Is_Block=mandaic}', "");
Expect(0, 2144, '\P{^Is_Block=mandaic}', "");
Expect(1, 2143, '\p{Is_Block=	 MANDAIC}', "");
Expect(0, 2143, '\p{^Is_Block=	 MANDAIC}', "");
Expect(0, 2143, '\P{Is_Block=	 MANDAIC}', "");
Expect(1, 2143, '\P{^Is_Block=	 MANDAIC}', "");
Expect(0, 2144, '\p{Is_Block=	 MANDAIC}', "");
Expect(1, 2144, '\p{^Is_Block=	 MANDAIC}', "");
Expect(1, 2144, '\P{Is_Block=	 MANDAIC}', "");
Expect(0, 2144, '\P{^Is_Block=	 MANDAIC}', "");
Error('\p{Is_Blk=_:=mandaic}');
Error('\P{Is_Blk=_:=mandaic}');
Expect(1, 2143, '\p{Is_Blk:	mandaic}', "");
Expect(0, 2143, '\p{^Is_Blk:	mandaic}', "");
Expect(0, 2143, '\P{Is_Blk:	mandaic}', "");
Expect(1, 2143, '\P{^Is_Blk:	mandaic}', "");
Expect(0, 2144, '\p{Is_Blk:	mandaic}', "");
Expect(1, 2144, '\p{^Is_Blk:	mandaic}', "");
Expect(1, 2144, '\P{Is_Blk:	mandaic}', "");
Expect(0, 2144, '\P{^Is_Blk:	mandaic}', "");
Expect(1, 2143, '\p{Is_Blk=	mandaic}', "");
Expect(0, 2143, '\p{^Is_Blk=	mandaic}', "");
Expect(0, 2143, '\P{Is_Blk=	mandaic}', "");
Expect(1, 2143, '\P{^Is_Blk=	mandaic}', "");
Expect(0, 2144, '\p{Is_Blk=	mandaic}', "");
Expect(1, 2144, '\p{^Is_Blk=	mandaic}', "");
Expect(1, 2144, '\P{Is_Blk=	mandaic}', "");
Expect(0, 2144, '\P{^Is_Blk=	mandaic}', "");
Error('\p{Block=:=--mathematical_Alphanumeric_Symbols}');
Error('\P{Block=:=--mathematical_Alphanumeric_Symbols}');
Expect(1, 120831, '\p{Block=mathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\p{^Block=mathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\P{Block=mathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\P{^Block=mathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\p{Block=mathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\p{^Block=mathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\P{Block=mathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\P{^Block=mathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\p{Block=	-MATHEMATICAL_ALPHANUMERIC_SYMBOLS}', "");
Expect(0, 120831, '\p{^Block=	-MATHEMATICAL_ALPHANUMERIC_SYMBOLS}', "");
Expect(0, 120831, '\P{Block=	-MATHEMATICAL_ALPHANUMERIC_SYMBOLS}', "");
Expect(1, 120831, '\P{^Block=	-MATHEMATICAL_ALPHANUMERIC_SYMBOLS}', "");
Expect(0, 120832, '\p{Block=	-MATHEMATICAL_ALPHANUMERIC_SYMBOLS}', "");
Expect(1, 120832, '\p{^Block=	-MATHEMATICAL_ALPHANUMERIC_SYMBOLS}', "");
Expect(1, 120832, '\P{Block=	-MATHEMATICAL_ALPHANUMERIC_SYMBOLS}', "");
Expect(0, 120832, '\P{^Block=	-MATHEMATICAL_ALPHANUMERIC_SYMBOLS}', "");
Error('\p{Blk= :=math_Alphanum}');
Error('\P{Blk= :=math_Alphanum}');
Expect(1, 120831, '\p{Blk=mathalphanum}', "");
Expect(0, 120831, '\p{^Blk=mathalphanum}', "");
Expect(0, 120831, '\P{Blk=mathalphanum}', "");
Expect(1, 120831, '\P{^Blk=mathalphanum}', "");
Expect(0, 120832, '\p{Blk=mathalphanum}', "");
Expect(1, 120832, '\p{^Blk=mathalphanum}', "");
Expect(1, 120832, '\P{Blk=mathalphanum}', "");
Expect(0, 120832, '\P{^Blk=mathalphanum}', "");
Expect(1, 120831, '\p{Blk=__MATH_Alphanum}', "");
Expect(0, 120831, '\p{^Blk=__MATH_Alphanum}', "");
Expect(0, 120831, '\P{Blk=__MATH_Alphanum}', "");
Expect(1, 120831, '\P{^Blk=__MATH_Alphanum}', "");
Expect(0, 120832, '\p{Blk=__MATH_Alphanum}', "");
Expect(1, 120832, '\p{^Blk=__MATH_Alphanum}', "");
Expect(1, 120832, '\P{Blk=__MATH_Alphanum}', "");
Expect(0, 120832, '\P{^Blk=__MATH_Alphanum}', "");
Error('\p{Is_Block=/a/_-mathematical_ALPHANUMERIC_Symbols}');
Error('\P{Is_Block=/a/_-mathematical_ALPHANUMERIC_Symbols}');
Expect(1, 120831, '\p{Is_Block=mathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\p{^Is_Block=mathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\P{Is_Block=mathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\P{^Is_Block=mathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\p{Is_Block=mathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\p{^Is_Block=mathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\P{Is_Block=mathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\P{^Is_Block=mathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\p{Is_Block= mathematical_Alphanumeric_Symbols}', "");
Expect(0, 120831, '\p{^Is_Block= mathematical_Alphanumeric_Symbols}', "");
Expect(0, 120831, '\P{Is_Block= mathematical_Alphanumeric_Symbols}', "");
Expect(1, 120831, '\P{^Is_Block= mathematical_Alphanumeric_Symbols}', "");
Expect(0, 120832, '\p{Is_Block= mathematical_Alphanumeric_Symbols}', "");
Expect(1, 120832, '\p{^Is_Block= mathematical_Alphanumeric_Symbols}', "");
Expect(1, 120832, '\P{Is_Block= mathematical_Alphanumeric_Symbols}', "");
Expect(0, 120832, '\P{^Is_Block= mathematical_Alphanumeric_Symbols}', "");
Error('\p{Is_Blk=/a/--math_ALPHANUM}');
Error('\P{Is_Blk=/a/--math_ALPHANUM}');
Expect(1, 120831, '\p{Is_Blk:mathalphanum}', "");
Expect(0, 120831, '\p{^Is_Blk:mathalphanum}', "");
Expect(0, 120831, '\P{Is_Blk:mathalphanum}', "");
Expect(1, 120831, '\P{^Is_Blk:mathalphanum}', "");
Expect(0, 120832, '\p{Is_Blk:mathalphanum}', "");
Expect(1, 120832, '\p{^Is_Blk:mathalphanum}', "");
Expect(1, 120832, '\P{Is_Blk:mathalphanum}', "");
Expect(0, 120832, '\P{^Is_Blk:mathalphanum}', "");
Expect(1, 120831, '\p{Is_Blk=	MATH_ALPHANUM}', "");
Expect(0, 120831, '\p{^Is_Blk=	MATH_ALPHANUM}', "");
Expect(0, 120831, '\P{Is_Blk=	MATH_ALPHANUM}', "");
Expect(1, 120831, '\P{^Is_Blk=	MATH_ALPHANUM}', "");
Expect(0, 120832, '\p{Is_Blk=	MATH_ALPHANUM}', "");
Expect(1, 120832, '\p{^Is_Blk=	MATH_ALPHANUM}', "");
Expect(1, 120832, '\P{Is_Blk=	MATH_ALPHANUM}', "");
Expect(0, 120832, '\P{^Is_Blk=	MATH_ALPHANUM}', "");
Error('\p{Block= :=MATHEMATICAL_operators}');
Error('\P{Block= :=MATHEMATICAL_operators}');
Expect(1, 8959, '\p{Block: mathematicaloperators}', "");
Expect(0, 8959, '\p{^Block: mathematicaloperators}', "");
Expect(0, 8959, '\P{Block: mathematicaloperators}', "");
Expect(1, 8959, '\P{^Block: mathematicaloperators}', "");
Expect(0, 8960, '\p{Block: mathematicaloperators}', "");
Expect(1, 8960, '\p{^Block: mathematicaloperators}', "");
Expect(1, 8960, '\P{Block: mathematicaloperators}', "");
Expect(0, 8960, '\P{^Block: mathematicaloperators}', "");
Expect(1, 8959, '\p{Block=_ MATHEMATICAL_Operators}', "");
Expect(0, 8959, '\p{^Block=_ MATHEMATICAL_Operators}', "");
Expect(0, 8959, '\P{Block=_ MATHEMATICAL_Operators}', "");
Expect(1, 8959, '\P{^Block=_ MATHEMATICAL_Operators}', "");
Expect(0, 8960, '\p{Block=_ MATHEMATICAL_Operators}', "");
Expect(1, 8960, '\p{^Block=_ MATHEMATICAL_Operators}', "");
Expect(1, 8960, '\P{Block=_ MATHEMATICAL_Operators}', "");
Expect(0, 8960, '\P{^Block=_ MATHEMATICAL_Operators}', "");
Error('\p{Blk=:=	 MATH_Operators}');
Error('\P{Blk=:=	 MATH_Operators}');
Expect(1, 8959, '\p{Blk:	mathoperators}', "");
Expect(0, 8959, '\p{^Blk:	mathoperators}', "");
Expect(0, 8959, '\P{Blk:	mathoperators}', "");
Expect(1, 8959, '\P{^Blk:	mathoperators}', "");
Expect(0, 8960, '\p{Blk:	mathoperators}', "");
Expect(1, 8960, '\p{^Blk:	mathoperators}', "");
Expect(1, 8960, '\P{Blk:	mathoperators}', "");
Expect(0, 8960, '\P{^Blk:	mathoperators}', "");
Expect(1, 8959, '\p{Blk=__Math_Operators}', "");
Expect(0, 8959, '\p{^Blk=__Math_Operators}', "");
Expect(0, 8959, '\P{Blk=__Math_Operators}', "");
Expect(1, 8959, '\P{^Blk=__Math_Operators}', "");
Expect(0, 8960, '\p{Blk=__Math_Operators}', "");
Expect(1, 8960, '\p{^Blk=__Math_Operators}', "");
Expect(1, 8960, '\P{Blk=__Math_Operators}', "");
Expect(0, 8960, '\P{^Blk=__Math_Operators}', "");
Error('\p{Is_Block=	/a/Mathematical_Operators}');
Error('\P{Is_Block=	/a/Mathematical_Operators}');
Expect(1, 8959, '\p{Is_Block:	mathematicaloperators}', "");
Expect(0, 8959, '\p{^Is_Block:	mathematicaloperators}', "");
Expect(0, 8959, '\P{Is_Block:	mathematicaloperators}', "");
Expect(1, 8959, '\P{^Is_Block:	mathematicaloperators}', "");
Expect(0, 8960, '\p{Is_Block:	mathematicaloperators}', "");
Expect(1, 8960, '\p{^Is_Block:	mathematicaloperators}', "");
Expect(1, 8960, '\P{Is_Block:	mathematicaloperators}', "");
Expect(0, 8960, '\P{^Is_Block:	mathematicaloperators}', "");
Expect(1, 8959, '\p{Is_Block:_ mathematical_Operators}', "");
Expect(0, 8959, '\p{^Is_Block:_ mathematical_Operators}', "");
Expect(0, 8959, '\P{Is_Block:_ mathematical_Operators}', "");
Expect(1, 8959, '\P{^Is_Block:_ mathematical_Operators}', "");
Expect(0, 8960, '\p{Is_Block:_ mathematical_Operators}', "");
Expect(1, 8960, '\p{^Is_Block:_ mathematical_Operators}', "");
Expect(1, 8960, '\P{Is_Block:_ mathematical_Operators}', "");
Expect(0, 8960, '\P{^Is_Block:_ mathematical_Operators}', "");
Error('\p{Is_Blk=/a/-_Math_Operators}');
Error('\P{Is_Blk=/a/-_Math_Operators}');
Expect(1, 8959, '\p{Is_Blk=mathoperators}', "");
Expect(0, 8959, '\p{^Is_Blk=mathoperators}', "");
Expect(0, 8959, '\P{Is_Blk=mathoperators}', "");
Expect(1, 8959, '\P{^Is_Blk=mathoperators}', "");
Expect(0, 8960, '\p{Is_Blk=mathoperators}', "");
Expect(1, 8960, '\p{^Is_Blk=mathoperators}', "");
Expect(1, 8960, '\P{Is_Blk=mathoperators}', "");
Expect(0, 8960, '\P{^Is_Blk=mathoperators}', "");
Expect(1, 8959, '\p{Is_Blk=_-Math_Operators}', "");
Expect(0, 8959, '\p{^Is_Blk=_-Math_Operators}', "");
Expect(0, 8959, '\P{Is_Blk=_-Math_Operators}', "");
Expect(1, 8959, '\P{^Is_Blk=_-Math_Operators}', "");
Expect(0, 8960, '\p{Is_Blk=_-Math_Operators}', "");
Expect(1, 8960, '\p{^Is_Blk=_-Math_Operators}', "");
Expect(1, 8960, '\P{Is_Blk=_-Math_Operators}', "");
Expect(0, 8960, '\P{^Is_Blk=_-Math_Operators}', "");
Error('\p{Block=/a/	 MEETEI_mayek}');
Error('\P{Block=/a/	 MEETEI_mayek}');
Expect(1, 44031, '\p{Block=meeteimayek}', "");
Expect(0, 44031, '\p{^Block=meeteimayek}', "");
Expect(0, 44031, '\P{Block=meeteimayek}', "");
Expect(1, 44031, '\P{^Block=meeteimayek}', "");
Expect(0, 44032, '\p{Block=meeteimayek}', "");
Expect(1, 44032, '\p{^Block=meeteimayek}', "");
Expect(1, 44032, '\P{Block=meeteimayek}', "");
Expect(0, 44032, '\P{^Block=meeteimayek}', "");
Expect(1, 44031, '\p{Block:		_Meetei_Mayek}', "");
Expect(0, 44031, '\p{^Block:		_Meetei_Mayek}', "");
Expect(0, 44031, '\P{Block:		_Meetei_Mayek}', "");
Expect(1, 44031, '\P{^Block:		_Meetei_Mayek}', "");
Expect(0, 44032, '\p{Block:		_Meetei_Mayek}', "");
Expect(1, 44032, '\p{^Block:		_Meetei_Mayek}', "");
Expect(1, 44032, '\P{Block:		_Meetei_Mayek}', "");
Expect(0, 44032, '\P{^Block:		_Meetei_Mayek}', "");
Error('\p{Blk=/a/_-MEETEI_Mayek}');
Error('\P{Blk=/a/_-MEETEI_Mayek}');
Expect(1, 44031, '\p{Blk=meeteimayek}', "");
Expect(0, 44031, '\p{^Blk=meeteimayek}', "");
Expect(0, 44031, '\P{Blk=meeteimayek}', "");
Expect(1, 44031, '\P{^Blk=meeteimayek}', "");
Expect(0, 44032, '\p{Blk=meeteimayek}', "");
Expect(1, 44032, '\p{^Blk=meeteimayek}', "");
Expect(1, 44032, '\P{Blk=meeteimayek}', "");
Expect(0, 44032, '\P{^Blk=meeteimayek}', "");
Expect(1, 44031, '\p{Blk= _MEETEI_mayek}', "");
Expect(0, 44031, '\p{^Blk= _MEETEI_mayek}', "");
Expect(0, 44031, '\P{Blk= _MEETEI_mayek}', "");
Expect(1, 44031, '\P{^Blk= _MEETEI_mayek}', "");
Expect(0, 44032, '\p{Blk= _MEETEI_mayek}', "");
Expect(1, 44032, '\p{^Blk= _MEETEI_mayek}', "");
Expect(1, 44032, '\P{Blk= _MEETEI_mayek}', "");
Expect(0, 44032, '\P{^Blk= _MEETEI_mayek}', "");
Error('\p{Is_Block=	 Meetei_Mayek/a/}');
Error('\P{Is_Block=	 Meetei_Mayek/a/}');
Expect(1, 44031, '\p{Is_Block=meeteimayek}', "");
Expect(0, 44031, '\p{^Is_Block=meeteimayek}', "");
Expect(0, 44031, '\P{Is_Block=meeteimayek}', "");
Expect(1, 44031, '\P{^Is_Block=meeteimayek}', "");
Expect(0, 44032, '\p{Is_Block=meeteimayek}', "");
Expect(1, 44032, '\p{^Is_Block=meeteimayek}', "");
Expect(1, 44032, '\P{Is_Block=meeteimayek}', "");
Expect(0, 44032, '\P{^Is_Block=meeteimayek}', "");
Expect(1, 44031, '\p{Is_Block=_MEETEI_Mayek}', "");
Expect(0, 44031, '\p{^Is_Block=_MEETEI_Mayek}', "");
Expect(0, 44031, '\P{Is_Block=_MEETEI_Mayek}', "");
Expect(1, 44031, '\P{^Is_Block=_MEETEI_Mayek}', "");
Expect(0, 44032, '\p{Is_Block=_MEETEI_Mayek}', "");
Expect(1, 44032, '\p{^Is_Block=_MEETEI_Mayek}', "");
Expect(1, 44032, '\P{Is_Block=_MEETEI_Mayek}', "");
Expect(0, 44032, '\P{^Is_Block=_MEETEI_Mayek}', "");
Error('\p{Is_Blk:  _Meetei_Mayek/a/}');
Error('\P{Is_Blk:  _Meetei_Mayek/a/}');
Expect(1, 44031, '\p{Is_Blk=meeteimayek}', "");
Expect(0, 44031, '\p{^Is_Blk=meeteimayek}', "");
Expect(0, 44031, '\P{Is_Blk=meeteimayek}', "");
Expect(1, 44031, '\P{^Is_Blk=meeteimayek}', "");
Expect(0, 44032, '\p{Is_Blk=meeteimayek}', "");
Expect(1, 44032, '\p{^Is_Blk=meeteimayek}', "");
Expect(1, 44032, '\P{Is_Blk=meeteimayek}', "");
Expect(0, 44032, '\P{^Is_Blk=meeteimayek}', "");
Expect(1, 44031, '\p{Is_Blk=	-MEETEI_Mayek}', "");
Expect(0, 44031, '\p{^Is_Blk=	-MEETEI_Mayek}', "");
Expect(0, 44031, '\P{Is_Blk=	-MEETEI_Mayek}', "");
Expect(1, 44031, '\P{^Is_Blk=	-MEETEI_Mayek}', "");
Expect(0, 44032, '\p{Is_Blk=	-MEETEI_Mayek}', "");
Expect(1, 44032, '\p{^Is_Blk=	-MEETEI_Mayek}', "");
Expect(1, 44032, '\P{Is_Blk=	-MEETEI_Mayek}', "");
Expect(0, 44032, '\P{^Is_Blk=	-MEETEI_Mayek}', "");
Error('\p{Block::=	Meetei_Mayek_Extensions}');
Error('\P{Block::=	Meetei_Mayek_Extensions}');
Expect(1, 43775, '\p{Block=meeteimayekextensions}', "");
Expect(0, 43775, '\p{^Block=meeteimayekextensions}', "");
Expect(0, 43775, '\P{Block=meeteimayekextensions}', "");
Expect(1, 43775, '\P{^Block=meeteimayekextensions}', "");
Expect(0, 43776, '\p{Block=meeteimayekextensions}', "");
Expect(1, 43776, '\p{^Block=meeteimayekextensions}', "");
Expect(1, 43776, '\P{Block=meeteimayekextensions}', "");
Expect(0, 43776, '\P{^Block=meeteimayekextensions}', "");
Expect(1, 43775, '\p{Block=Meetei_Mayek_Extensions}', "");
Expect(0, 43775, '\p{^Block=Meetei_Mayek_Extensions}', "");
Expect(0, 43775, '\P{Block=Meetei_Mayek_Extensions}', "");
Expect(1, 43775, '\P{^Block=Meetei_Mayek_Extensions}', "");
Expect(0, 43776, '\p{Block=Meetei_Mayek_Extensions}', "");
Expect(1, 43776, '\p{^Block=Meetei_Mayek_Extensions}', "");
Expect(1, 43776, '\P{Block=Meetei_Mayek_Extensions}', "");
Expect(0, 43776, '\P{^Block=Meetei_Mayek_Extensions}', "");
Error('\p{Blk:   /a/ 	Meetei_mayek_Ext}');
Error('\P{Blk:   /a/ 	Meetei_mayek_Ext}');
Expect(1, 43775, '\p{Blk=meeteimayekext}', "");
Expect(0, 43775, '\p{^Blk=meeteimayekext}', "");
Expect(0, 43775, '\P{Blk=meeteimayekext}', "");
Expect(1, 43775, '\P{^Blk=meeteimayekext}', "");
Expect(0, 43776, '\p{Blk=meeteimayekext}', "");
Expect(1, 43776, '\p{^Blk=meeteimayekext}', "");
Expect(1, 43776, '\P{Blk=meeteimayekext}', "");
Expect(0, 43776, '\P{^Blk=meeteimayekext}', "");
Expect(1, 43775, '\p{Blk= 	meetei_Mayek_Ext}', "");
Expect(0, 43775, '\p{^Blk= 	meetei_Mayek_Ext}', "");
Expect(0, 43775, '\P{Blk= 	meetei_Mayek_Ext}', "");
Expect(1, 43775, '\P{^Blk= 	meetei_Mayek_Ext}', "");
Expect(0, 43776, '\p{Blk= 	meetei_Mayek_Ext}', "");
Expect(1, 43776, '\p{^Blk= 	meetei_Mayek_Ext}', "");
Expect(1, 43776, '\P{Blk= 	meetei_Mayek_Ext}', "");
Expect(0, 43776, '\P{^Blk= 	meetei_Mayek_Ext}', "");
Error('\p{Is_Block=:=_Meetei_MAYEK_EXTENSIONS}');
Error('\P{Is_Block=:=_Meetei_MAYEK_EXTENSIONS}');
Expect(1, 43775, '\p{Is_Block=meeteimayekextensions}', "");
Expect(0, 43775, '\p{^Is_Block=meeteimayekextensions}', "");
Expect(0, 43775, '\P{Is_Block=meeteimayekextensions}', "");
Expect(1, 43775, '\P{^Is_Block=meeteimayekextensions}', "");
Expect(0, 43776, '\p{Is_Block=meeteimayekextensions}', "");
Expect(1, 43776, '\p{^Is_Block=meeteimayekextensions}', "");
Expect(1, 43776, '\P{Is_Block=meeteimayekextensions}', "");
Expect(0, 43776, '\P{^Is_Block=meeteimayekextensions}', "");
Expect(1, 43775, '\p{Is_Block=  Meetei_mayek_Extensions}', "");
Expect(0, 43775, '\p{^Is_Block=  Meetei_mayek_Extensions}', "");
Expect(0, 43775, '\P{Is_Block=  Meetei_mayek_Extensions}', "");
Expect(1, 43775, '\P{^Is_Block=  Meetei_mayek_Extensions}', "");
Expect(0, 43776, '\p{Is_Block=  Meetei_mayek_Extensions}', "");
Expect(1, 43776, '\p{^Is_Block=  Meetei_mayek_Extensions}', "");
Expect(1, 43776, '\P{Is_Block=  Meetei_mayek_Extensions}', "");
Expect(0, 43776, '\P{^Is_Block=  Meetei_mayek_Extensions}', "");
Error('\p{Is_Blk:    /a/meetei_Mayek_Ext}');
Error('\P{Is_Blk:    /a/meetei_Mayek_Ext}');
Expect(1, 43775, '\p{Is_Blk:   meeteimayekext}', "");
Expect(0, 43775, '\p{^Is_Blk:   meeteimayekext}', "");
Expect(0, 43775, '\P{Is_Blk:   meeteimayekext}', "");
Expect(1, 43775, '\P{^Is_Blk:   meeteimayekext}', "");
Expect(0, 43776, '\p{Is_Blk:   meeteimayekext}', "");
Expect(1, 43776, '\p{^Is_Blk:   meeteimayekext}', "");
Expect(1, 43776, '\P{Is_Blk:   meeteimayekext}', "");
Expect(0, 43776, '\P{^Is_Blk:   meeteimayekext}', "");
Expect(1, 43775, '\p{Is_Blk=	-Meetei_MAYEK_Ext}', "");
Expect(0, 43775, '\p{^Is_Blk=	-Meetei_MAYEK_Ext}', "");
Expect(0, 43775, '\P{Is_Blk=	-Meetei_MAYEK_Ext}', "");
Expect(1, 43775, '\P{^Is_Blk=	-Meetei_MAYEK_Ext}', "");
Expect(0, 43776, '\p{Is_Blk=	-Meetei_MAYEK_Ext}', "");
Expect(1, 43776, '\p{^Is_Blk=	-Meetei_MAYEK_Ext}', "");
Expect(1, 43776, '\P{Is_Blk=	-Meetei_MAYEK_Ext}', "");
Expect(0, 43776, '\P{^Is_Blk=	-Meetei_MAYEK_Ext}', "");
Error('\p{Block=--meroitic_Cursive:=}');
Error('\P{Block=--meroitic_Cursive:=}');
Expect(1, 68095, '\p{Block=meroiticcursive}', "");
Expect(0, 68095, '\p{^Block=meroiticcursive}', "");
Expect(0, 68095, '\P{Block=meroiticcursive}', "");
Expect(1, 68095, '\P{^Block=meroiticcursive}', "");
Expect(0, 68096, '\p{Block=meroiticcursive}', "");
Expect(1, 68096, '\p{^Block=meroiticcursive}', "");
Expect(1, 68096, '\P{Block=meroiticcursive}', "");
Expect(0, 68096, '\P{^Block=meroiticcursive}', "");
Expect(1, 68095, '\p{Block=	-meroitic_CURSIVE}', "");
Expect(0, 68095, '\p{^Block=	-meroitic_CURSIVE}', "");
Expect(0, 68095, '\P{Block=	-meroitic_CURSIVE}', "");
Expect(1, 68095, '\P{^Block=	-meroitic_CURSIVE}', "");
Expect(0, 68096, '\p{Block=	-meroitic_CURSIVE}', "");
Expect(1, 68096, '\p{^Block=	-meroitic_CURSIVE}', "");
Expect(1, 68096, '\P{Block=	-meroitic_CURSIVE}', "");
Expect(0, 68096, '\P{^Block=	-meroitic_CURSIVE}', "");
Error('\p{Blk=-:=MEROITIC_Cursive}');
Error('\P{Blk=-:=MEROITIC_Cursive}');
Expect(1, 68095, '\p{Blk:meroiticcursive}', "");
Expect(0, 68095, '\p{^Blk:meroiticcursive}', "");
Expect(0, 68095, '\P{Blk:meroiticcursive}', "");
Expect(1, 68095, '\P{^Blk:meroiticcursive}', "");
Expect(0, 68096, '\p{Blk:meroiticcursive}', "");
Expect(1, 68096, '\p{^Blk:meroiticcursive}', "");
Expect(1, 68096, '\P{Blk:meroiticcursive}', "");
Expect(0, 68096, '\P{^Blk:meroiticcursive}', "");
Expect(1, 68095, '\p{Blk:	meroitic_CURSIVE}', "");
Expect(0, 68095, '\p{^Blk:	meroitic_CURSIVE}', "");
Expect(0, 68095, '\P{Blk:	meroitic_CURSIVE}', "");
Expect(1, 68095, '\P{^Blk:	meroitic_CURSIVE}', "");
Expect(0, 68096, '\p{Blk:	meroitic_CURSIVE}', "");
Expect(1, 68096, '\p{^Blk:	meroitic_CURSIVE}', "");
Expect(1, 68096, '\P{Blk:	meroitic_CURSIVE}', "");
Expect(0, 68096, '\P{^Blk:	meroitic_CURSIVE}', "");
Error('\p{Is_Block=/a/meroitic_cursive}');
Error('\P{Is_Block=/a/meroitic_cursive}');
Expect(1, 68095, '\p{Is_Block=meroiticcursive}', "");
Expect(0, 68095, '\p{^Is_Block=meroiticcursive}', "");
Expect(0, 68095, '\P{Is_Block=meroiticcursive}', "");
Expect(1, 68095, '\P{^Is_Block=meroiticcursive}', "");
Expect(0, 68096, '\p{Is_Block=meroiticcursive}', "");
Expect(1, 68096, '\p{^Is_Block=meroiticcursive}', "");
Expect(1, 68096, '\P{Is_Block=meroiticcursive}', "");
Expect(0, 68096, '\P{^Is_Block=meroiticcursive}', "");
Expect(1, 68095, '\p{Is_Block=	 MEROITIC_Cursive}', "");
Expect(0, 68095, '\p{^Is_Block=	 MEROITIC_Cursive}', "");
Expect(0, 68095, '\P{Is_Block=	 MEROITIC_Cursive}', "");
Expect(1, 68095, '\P{^Is_Block=	 MEROITIC_Cursive}', "");
Expect(0, 68096, '\p{Is_Block=	 MEROITIC_Cursive}', "");
Expect(1, 68096, '\p{^Is_Block=	 MEROITIC_Cursive}', "");
Expect(1, 68096, '\P{Is_Block=	 MEROITIC_Cursive}', "");
Expect(0, 68096, '\P{^Is_Block=	 MEROITIC_Cursive}', "");
Error('\p{Is_Blk=	 meroitic_cursive/a/}');
Error('\P{Is_Blk=	 meroitic_cursive/a/}');
Expect(1, 68095, '\p{Is_Blk=meroiticcursive}', "");
Expect(0, 68095, '\p{^Is_Blk=meroiticcursive}', "");
Expect(0, 68095, '\P{Is_Blk=meroiticcursive}', "");
Expect(1, 68095, '\P{^Is_Blk=meroiticcursive}', "");
Expect(0, 68096, '\p{Is_Blk=meroiticcursive}', "");
Expect(1, 68096, '\p{^Is_Blk=meroiticcursive}', "");
Expect(1, 68096, '\P{Is_Blk=meroiticcursive}', "");
Expect(0, 68096, '\P{^Is_Blk=meroiticcursive}', "");
Expect(1, 68095, '\p{Is_Blk=	Meroitic_Cursive}', "");
Expect(0, 68095, '\p{^Is_Blk=	Meroitic_Cursive}', "");
Expect(0, 68095, '\P{Is_Blk=	Meroitic_Cursive}', "");
Expect(1, 68095, '\P{^Is_Blk=	Meroitic_Cursive}', "");
Expect(0, 68096, '\p{Is_Blk=	Meroitic_Cursive}', "");
Expect(1, 68096, '\p{^Is_Blk=	Meroitic_Cursive}', "");
Expect(1, 68096, '\P{Is_Blk=	Meroitic_Cursive}', "");
Expect(0, 68096, '\P{^Is_Blk=	Meroitic_Cursive}', "");
Error('\p{Block=_:=Meroitic_Hieroglyphs}');
Error('\P{Block=_:=Meroitic_Hieroglyphs}');
Expect(1, 67999, '\p{Block=meroitichieroglyphs}', "");
Expect(0, 67999, '\p{^Block=meroitichieroglyphs}', "");
Expect(0, 67999, '\P{Block=meroitichieroglyphs}', "");
Expect(1, 67999, '\P{^Block=meroitichieroglyphs}', "");
Expect(0, 68000, '\p{Block=meroitichieroglyphs}', "");
Expect(1, 68000, '\p{^Block=meroitichieroglyphs}', "");
Expect(1, 68000, '\P{Block=meroitichieroglyphs}', "");
Expect(0, 68000, '\P{^Block=meroitichieroglyphs}', "");
Expect(1, 67999, '\p{Block=	 MEROITIC_Hieroglyphs}', "");
Expect(0, 67999, '\p{^Block=	 MEROITIC_Hieroglyphs}', "");
Expect(0, 67999, '\P{Block=	 MEROITIC_Hieroglyphs}', "");
Expect(1, 67999, '\P{^Block=	 MEROITIC_Hieroglyphs}', "");
Expect(0, 68000, '\p{Block=	 MEROITIC_Hieroglyphs}', "");
Expect(1, 68000, '\p{^Block=	 MEROITIC_Hieroglyphs}', "");
Expect(1, 68000, '\P{Block=	 MEROITIC_Hieroglyphs}', "");
Expect(0, 68000, '\P{^Block=	 MEROITIC_Hieroglyphs}', "");
Error('\p{Blk= /a/Meroitic_Hieroglyphs}');
Error('\P{Blk= /a/Meroitic_Hieroglyphs}');
Expect(1, 67999, '\p{Blk=meroitichieroglyphs}', "");
Expect(0, 67999, '\p{^Blk=meroitichieroglyphs}', "");
Expect(0, 67999, '\P{Blk=meroitichieroglyphs}', "");
Expect(1, 67999, '\P{^Blk=meroitichieroglyphs}', "");
Expect(0, 68000, '\p{Blk=meroitichieroglyphs}', "");
Expect(1, 68000, '\p{^Blk=meroitichieroglyphs}', "");
Expect(1, 68000, '\P{Blk=meroitichieroglyphs}', "");
Expect(0, 68000, '\P{^Blk=meroitichieroglyphs}', "");
Expect(1, 67999, '\p{Blk=	_meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\p{^Blk=	_meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\P{Blk=	_meroitic_Hieroglyphs}', "");
Expect(1, 67999, '\P{^Blk=	_meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\p{Blk=	_meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\p{^Blk=	_meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\P{Blk=	_meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\P{^Blk=	_meroitic_Hieroglyphs}', "");
Error('\p{Is_Block=:=-Meroitic_Hieroglyphs}');
Error('\P{Is_Block=:=-Meroitic_Hieroglyphs}');
Expect(1, 67999, '\p{Is_Block: meroitichieroglyphs}', "");
Expect(0, 67999, '\p{^Is_Block: meroitichieroglyphs}', "");
Expect(0, 67999, '\P{Is_Block: meroitichieroglyphs}', "");
Expect(1, 67999, '\P{^Is_Block: meroitichieroglyphs}', "");
Expect(0, 68000, '\p{Is_Block: meroitichieroglyphs}', "");
Expect(1, 68000, '\p{^Is_Block: meroitichieroglyphs}', "");
Expect(1, 68000, '\P{Is_Block: meroitichieroglyphs}', "");
Expect(0, 68000, '\P{^Is_Block: meroitichieroglyphs}', "");
Expect(1, 67999, '\p{Is_Block=_Meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\p{^Is_Block=_Meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\P{Is_Block=_Meroitic_Hieroglyphs}', "");
Expect(1, 67999, '\P{^Is_Block=_Meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\p{Is_Block=_Meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\p{^Is_Block=_Meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\P{Is_Block=_Meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\P{^Is_Block=_Meroitic_Hieroglyphs}', "");
Error('\p{Is_Blk=_meroitic_Hieroglyphs:=}');
Error('\P{Is_Blk=_meroitic_Hieroglyphs:=}');
Expect(1, 67999, '\p{Is_Blk=meroitichieroglyphs}', "");
Expect(0, 67999, '\p{^Is_Blk=meroitichieroglyphs}', "");
Expect(0, 67999, '\P{Is_Blk=meroitichieroglyphs}', "");
Expect(1, 67999, '\P{^Is_Blk=meroitichieroglyphs}', "");
Expect(0, 68000, '\p{Is_Blk=meroitichieroglyphs}', "");
Expect(1, 68000, '\p{^Is_Blk=meroitichieroglyphs}', "");
Expect(1, 68000, '\P{Is_Blk=meroitichieroglyphs}', "");
Expect(0, 68000, '\P{^Is_Blk=meroitichieroglyphs}', "");
Expect(1, 67999, '\p{Is_Blk=_ Meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\p{^Is_Blk=_ Meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\P{Is_Blk=_ Meroitic_Hieroglyphs}', "");
Expect(1, 67999, '\P{^Is_Blk=_ Meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\p{Is_Blk=_ Meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\p{^Is_Blk=_ Meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\P{Is_Blk=_ Meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\P{^Is_Blk=_ Meroitic_Hieroglyphs}', "");
Error('\p{Block:   _/a/Miao}');
Error('\P{Block:   _/a/Miao}');
Expect(1, 94111, '\p{Block=miao}', "");
Expect(0, 94111, '\p{^Block=miao}', "");
Expect(0, 94111, '\P{Block=miao}', "");
Expect(1, 94111, '\P{^Block=miao}', "");
Expect(0, 94112, '\p{Block=miao}', "");
Expect(1, 94112, '\p{^Block=miao}', "");
Expect(1, 94112, '\P{Block=miao}', "");
Expect(0, 94112, '\P{^Block=miao}', "");
Expect(1, 94111, '\p{Block=-miao}', "");
Expect(0, 94111, '\p{^Block=-miao}', "");
Expect(0, 94111, '\P{Block=-miao}', "");
Expect(1, 94111, '\P{^Block=-miao}', "");
Expect(0, 94112, '\p{Block=-miao}', "");
Expect(1, 94112, '\p{^Block=-miao}', "");
Expect(1, 94112, '\P{Block=-miao}', "");
Expect(0, 94112, '\P{^Block=-miao}', "");
Error('\p{Blk= _Miao:=}');
Error('\P{Blk= _Miao:=}');
Expect(1, 94111, '\p{Blk=miao}', "");
Expect(0, 94111, '\p{^Blk=miao}', "");
Expect(0, 94111, '\P{Blk=miao}', "");
Expect(1, 94111, '\P{^Blk=miao}', "");
Expect(0, 94112, '\p{Blk=miao}', "");
Expect(1, 94112, '\p{^Blk=miao}', "");
Expect(1, 94112, '\P{Blk=miao}', "");
Expect(0, 94112, '\P{^Blk=miao}', "");
Expect(1, 94111, '\p{Blk=-miao}', "");
Expect(0, 94111, '\p{^Blk=-miao}', "");
Expect(0, 94111, '\P{Blk=-miao}', "");
Expect(1, 94111, '\P{^Blk=-miao}', "");
Expect(0, 94112, '\p{Blk=-miao}', "");
Expect(1, 94112, '\p{^Blk=-miao}', "");
Expect(1, 94112, '\P{Blk=-miao}', "");
Expect(0, 94112, '\P{^Blk=-miao}', "");
Error('\p{Is_Block=-:=Miao}');
Error('\P{Is_Block=-:=Miao}');
Expect(1, 94111, '\p{Is_Block=miao}', "");
Expect(0, 94111, '\p{^Is_Block=miao}', "");
Expect(0, 94111, '\P{Is_Block=miao}', "");
Expect(1, 94111, '\P{^Is_Block=miao}', "");
Expect(0, 94112, '\p{Is_Block=miao}', "");
Expect(1, 94112, '\p{^Is_Block=miao}', "");
Expect(1, 94112, '\P{Is_Block=miao}', "");
Expect(0, 94112, '\P{^Is_Block=miao}', "");
Expect(1, 94111, '\p{Is_Block=--Miao}', "");
Expect(0, 94111, '\p{^Is_Block=--Miao}', "");
Expect(0, 94111, '\P{Is_Block=--Miao}', "");
Expect(1, 94111, '\P{^Is_Block=--Miao}', "");
Expect(0, 94112, '\p{Is_Block=--Miao}', "");
Expect(1, 94112, '\p{^Is_Block=--Miao}', "");
Expect(1, 94112, '\P{Is_Block=--Miao}', "");
Expect(0, 94112, '\P{^Is_Block=--Miao}', "");
Error('\p{Is_Blk=  Miao:=}');
Error('\P{Is_Blk=  Miao:=}');
Expect(1, 94111, '\p{Is_Blk=miao}', "");
Expect(0, 94111, '\p{^Is_Blk=miao}', "");
Expect(0, 94111, '\P{Is_Blk=miao}', "");
Expect(1, 94111, '\P{^Is_Blk=miao}', "");
Expect(0, 94112, '\p{Is_Blk=miao}', "");
Expect(1, 94112, '\p{^Is_Blk=miao}', "");
Expect(1, 94112, '\P{Is_Blk=miao}', "");
Expect(0, 94112, '\P{^Is_Blk=miao}', "");
Expect(1, 94111, '\p{Is_Blk= MIAO}', "");
Expect(0, 94111, '\p{^Is_Blk= MIAO}', "");
Expect(0, 94111, '\P{Is_Blk= MIAO}', "");
Expect(1, 94111, '\P{^Is_Blk= MIAO}', "");
Expect(0, 94112, '\p{Is_Blk= MIAO}', "");
Expect(1, 94112, '\p{^Is_Blk= MIAO}', "");
Expect(1, 94112, '\P{Is_Blk= MIAO}', "");
Expect(0, 94112, '\P{^Is_Blk= MIAO}', "");
Error('\p{Block=_ miscellaneous_symbols_And_arrows:=}');
Error('\P{Block=_ miscellaneous_symbols_And_arrows:=}');
Expect(1, 11263, '\p{Block=miscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\p{^Block=miscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\P{Block=miscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\P{^Block=miscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\p{Block=miscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\p{^Block=miscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\P{Block=miscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\P{^Block=miscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\p{Block=		Miscellaneous_Symbols_AND_Arrows}', "");
Expect(0, 11263, '\p{^Block=		Miscellaneous_Symbols_AND_Arrows}', "");
Expect(0, 11263, '\P{Block=		Miscellaneous_Symbols_AND_Arrows}', "");
Expect(1, 11263, '\P{^Block=		Miscellaneous_Symbols_AND_Arrows}', "");
Expect(0, 11264, '\p{Block=		Miscellaneous_Symbols_AND_Arrows}', "");
Expect(1, 11264, '\p{^Block=		Miscellaneous_Symbols_AND_Arrows}', "");
Expect(1, 11264, '\P{Block=		Miscellaneous_Symbols_AND_Arrows}', "");
Expect(0, 11264, '\P{^Block=		Miscellaneous_Symbols_AND_Arrows}', "");
Error('\p{Blk= :=Misc_Arrows}');
Error('\P{Blk= :=Misc_Arrows}');
Expect(1, 11263, '\p{Blk=miscarrows}', "");
Expect(0, 11263, '\p{^Blk=miscarrows}', "");
Expect(0, 11263, '\P{Blk=miscarrows}', "");
Expect(1, 11263, '\P{^Blk=miscarrows}', "");
Expect(0, 11264, '\p{Blk=miscarrows}', "");
Expect(1, 11264, '\p{^Blk=miscarrows}', "");
Expect(1, 11264, '\P{Blk=miscarrows}', "");
Expect(0, 11264, '\P{^Blk=miscarrows}', "");
Expect(1, 11263, '\p{Blk=_	MISC_Arrows}', "");
Expect(0, 11263, '\p{^Blk=_	MISC_Arrows}', "");
Expect(0, 11263, '\P{Blk=_	MISC_Arrows}', "");
Expect(1, 11263, '\P{^Blk=_	MISC_Arrows}', "");
Expect(0, 11264, '\p{Blk=_	MISC_Arrows}', "");
Expect(1, 11264, '\p{^Blk=_	MISC_Arrows}', "");
Expect(1, 11264, '\P{Blk=_	MISC_Arrows}', "");
Expect(0, 11264, '\P{^Blk=_	MISC_Arrows}', "");
Error('\p{Is_Block:	_-MISCELLANEOUS_symbols_AND_ARROWS/a/}');
Error('\P{Is_Block:	_-MISCELLANEOUS_symbols_AND_ARROWS/a/}');
Expect(1, 11263, '\p{Is_Block=miscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\p{^Is_Block=miscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\P{Is_Block=miscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\P{^Is_Block=miscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\p{Is_Block=miscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\p{^Is_Block=miscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\P{Is_Block=miscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\P{^Is_Block=miscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\p{Is_Block=  miscellaneous_Symbols_And_ARROWS}', "");
Expect(0, 11263, '\p{^Is_Block=  miscellaneous_Symbols_And_ARROWS}', "");
Expect(0, 11263, '\P{Is_Block=  miscellaneous_Symbols_And_ARROWS}', "");
Expect(1, 11263, '\P{^Is_Block=  miscellaneous_Symbols_And_ARROWS}', "");
Expect(0, 11264, '\p{Is_Block=  miscellaneous_Symbols_And_ARROWS}', "");
Expect(1, 11264, '\p{^Is_Block=  miscellaneous_Symbols_And_ARROWS}', "");
Expect(1, 11264, '\P{Is_Block=  miscellaneous_Symbols_And_ARROWS}', "");
Expect(0, 11264, '\P{^Is_Block=  miscellaneous_Symbols_And_ARROWS}', "");
Error('\p{Is_Blk= :=Misc_Arrows}');
Error('\P{Is_Blk= :=Misc_Arrows}');
Expect(1, 11263, '\p{Is_Blk=miscarrows}', "");
Expect(0, 11263, '\p{^Is_Blk=miscarrows}', "");
Expect(0, 11263, '\P{Is_Blk=miscarrows}', "");
Expect(1, 11263, '\P{^Is_Blk=miscarrows}', "");
Expect(0, 11264, '\p{Is_Blk=miscarrows}', "");
Expect(1, 11264, '\p{^Is_Blk=miscarrows}', "");
Expect(1, 11264, '\P{Is_Blk=miscarrows}', "");
Expect(0, 11264, '\P{^Is_Blk=miscarrows}', "");
Expect(1, 11263, '\p{Is_Blk= -Misc_arrows}', "");
Expect(0, 11263, '\p{^Is_Blk= -Misc_arrows}', "");
Expect(0, 11263, '\P{Is_Blk= -Misc_arrows}', "");
Expect(1, 11263, '\P{^Is_Blk= -Misc_arrows}', "");
Expect(0, 11264, '\p{Is_Blk= -Misc_arrows}', "");
Expect(1, 11264, '\p{^Is_Blk= -Misc_arrows}', "");
Expect(1, 11264, '\P{Is_Blk= -Misc_arrows}', "");
Expect(0, 11264, '\P{^Is_Blk= -Misc_arrows}', "");
Error('\p{Block=- Miscellaneous_mathematical_SYMBOLS_a:=}');
Error('\P{Block=- Miscellaneous_mathematical_SYMBOLS_a:=}');
Expect(1, 10223, '\p{Block=miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\p{^Block=miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\P{Block=miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\P{^Block=miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\p{Block=miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\p{^Block=miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\P{Block=miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\P{^Block=miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\p{Block=_-MISCELLANEOUS_Mathematical_Symbols_A}', "");
Expect(0, 10223, '\p{^Block=_-MISCELLANEOUS_Mathematical_Symbols_A}', "");
Expect(0, 10223, '\P{Block=_-MISCELLANEOUS_Mathematical_Symbols_A}', "");
Expect(1, 10223, '\P{^Block=_-MISCELLANEOUS_Mathematical_Symbols_A}', "");
Expect(0, 10224, '\p{Block=_-MISCELLANEOUS_Mathematical_Symbols_A}', "");
Expect(1, 10224, '\p{^Block=_-MISCELLANEOUS_Mathematical_Symbols_A}', "");
Expect(1, 10224, '\P{Block=_-MISCELLANEOUS_Mathematical_Symbols_A}', "");
Expect(0, 10224, '\P{^Block=_-MISCELLANEOUS_Mathematical_Symbols_A}', "");
Error('\p{Blk=--misc_math_Symbols_a/a/}');
Error('\P{Blk=--misc_math_Symbols_a/a/}');
Expect(1, 10223, '\p{Blk=miscmathsymbolsa}', "");
Expect(0, 10223, '\p{^Blk=miscmathsymbolsa}', "");
Expect(0, 10223, '\P{Blk=miscmathsymbolsa}', "");
Expect(1, 10223, '\P{^Blk=miscmathsymbolsa}', "");
Expect(0, 10224, '\p{Blk=miscmathsymbolsa}', "");
Expect(1, 10224, '\p{^Blk=miscmathsymbolsa}', "");
Expect(1, 10224, '\P{Blk=miscmathsymbolsa}', "");
Expect(0, 10224, '\P{^Blk=miscmathsymbolsa}', "");
Expect(1, 10223, '\p{Blk=- MISC_math_SYMBOLS_a}', "");
Expect(0, 10223, '\p{^Blk=- MISC_math_SYMBOLS_a}', "");
Expect(0, 10223, '\P{Blk=- MISC_math_SYMBOLS_a}', "");
Expect(1, 10223, '\P{^Blk=- MISC_math_SYMBOLS_a}', "");
Expect(0, 10224, '\p{Blk=- MISC_math_SYMBOLS_a}', "");
Expect(1, 10224, '\p{^Blk=- MISC_math_SYMBOLS_a}', "");
Expect(1, 10224, '\P{Blk=- MISC_math_SYMBOLS_a}', "");
Expect(0, 10224, '\P{^Blk=- MISC_math_SYMBOLS_a}', "");
Error('\p{Is_Block=	/a/Miscellaneous_Mathematical_symbols_A}');
Error('\P{Is_Block=	/a/Miscellaneous_Mathematical_symbols_A}');
Expect(1, 10223, '\p{Is_Block=miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\p{^Is_Block=miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\P{Is_Block=miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\P{^Is_Block=miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\p{Is_Block=miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\p{^Is_Block=miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\P{Is_Block=miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\P{^Is_Block=miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\p{Is_Block=_miscellaneous_mathematical_SYMBOLS_A}', "");
Expect(0, 10223, '\p{^Is_Block=_miscellaneous_mathematical_SYMBOLS_A}', "");
Expect(0, 10223, '\P{Is_Block=_miscellaneous_mathematical_SYMBOLS_A}', "");
Expect(1, 10223, '\P{^Is_Block=_miscellaneous_mathematical_SYMBOLS_A}', "");
Expect(0, 10224, '\p{Is_Block=_miscellaneous_mathematical_SYMBOLS_A}', "");
Expect(1, 10224, '\p{^Is_Block=_miscellaneous_mathematical_SYMBOLS_A}', "");
Expect(1, 10224, '\P{Is_Block=_miscellaneous_mathematical_SYMBOLS_A}', "");
Expect(0, 10224, '\P{^Is_Block=_miscellaneous_mathematical_SYMBOLS_A}', "");
Error('\p{Is_Blk: /a/Misc_math_Symbols_A}');
Error('\P{Is_Blk: /a/Misc_math_Symbols_A}');
Expect(1, 10223, '\p{Is_Blk=miscmathsymbolsa}', "");
Expect(0, 10223, '\p{^Is_Blk=miscmathsymbolsa}', "");
Expect(0, 10223, '\P{Is_Blk=miscmathsymbolsa}', "");
Expect(1, 10223, '\P{^Is_Blk=miscmathsymbolsa}', "");
Expect(0, 10224, '\p{Is_Blk=miscmathsymbolsa}', "");
Expect(1, 10224, '\p{^Is_Blk=miscmathsymbolsa}', "");
Expect(1, 10224, '\P{Is_Blk=miscmathsymbolsa}', "");
Expect(0, 10224, '\P{^Is_Blk=miscmathsymbolsa}', "");
Expect(1, 10223, '\p{Is_Blk=		Misc_Math_Symbols_A}', "");
Expect(0, 10223, '\p{^Is_Blk=		Misc_Math_Symbols_A}', "");
Expect(0, 10223, '\P{Is_Blk=		Misc_Math_Symbols_A}', "");
Expect(1, 10223, '\P{^Is_Blk=		Misc_Math_Symbols_A}', "");
Expect(0, 10224, '\p{Is_Blk=		Misc_Math_Symbols_A}', "");
Expect(1, 10224, '\p{^Is_Blk=		Misc_Math_Symbols_A}', "");
Expect(1, 10224, '\P{Is_Blk=		Misc_Math_Symbols_A}', "");
Expect(0, 10224, '\P{^Is_Blk=		Misc_Math_Symbols_A}', "");
Error('\p{Block=--Miscellaneous_Mathematical_SYMBOLS_B:=}');
Error('\P{Block=--Miscellaneous_Mathematical_SYMBOLS_B:=}');
Expect(1, 10751, '\p{Block=miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\p{^Block=miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\P{Block=miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\P{^Block=miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\p{Block=miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\p{^Block=miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\P{Block=miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\P{^Block=miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\p{Block=  miscellaneous_MATHEMATICAL_SYMBOLS_B}', "");
Expect(0, 10751, '\p{^Block=  miscellaneous_MATHEMATICAL_SYMBOLS_B}', "");
Expect(0, 10751, '\P{Block=  miscellaneous_MATHEMATICAL_SYMBOLS_B}', "");
Expect(1, 10751, '\P{^Block=  miscellaneous_MATHEMATICAL_SYMBOLS_B}', "");
Expect(0, 10752, '\p{Block=  miscellaneous_MATHEMATICAL_SYMBOLS_B}', "");
Expect(1, 10752, '\p{^Block=  miscellaneous_MATHEMATICAL_SYMBOLS_B}', "");
Expect(1, 10752, '\P{Block=  miscellaneous_MATHEMATICAL_SYMBOLS_B}', "");
Expect(0, 10752, '\P{^Block=  miscellaneous_MATHEMATICAL_SYMBOLS_B}', "");
Error('\p{Blk=_/a/MISC_Math_Symbols_B}');
Error('\P{Blk=_/a/MISC_Math_Symbols_B}');
Expect(1, 10751, '\p{Blk=miscmathsymbolsb}', "");
Expect(0, 10751, '\p{^Blk=miscmathsymbolsb}', "");
Expect(0, 10751, '\P{Blk=miscmathsymbolsb}', "");
Expect(1, 10751, '\P{^Blk=miscmathsymbolsb}', "");
Expect(0, 10752, '\p{Blk=miscmathsymbolsb}', "");
Expect(1, 10752, '\p{^Blk=miscmathsymbolsb}', "");
Expect(1, 10752, '\P{Blk=miscmathsymbolsb}', "");
Expect(0, 10752, '\P{^Blk=miscmathsymbolsb}', "");
Expect(1, 10751, '\p{Blk=-Misc_MATH_Symbols_B}', "");
Expect(0, 10751, '\p{^Blk=-Misc_MATH_Symbols_B}', "");
Expect(0, 10751, '\P{Blk=-Misc_MATH_Symbols_B}', "");
Expect(1, 10751, '\P{^Blk=-Misc_MATH_Symbols_B}', "");
Expect(0, 10752, '\p{Blk=-Misc_MATH_Symbols_B}', "");
Expect(1, 10752, '\p{^Blk=-Misc_MATH_Symbols_B}', "");
Expect(1, 10752, '\P{Blk=-Misc_MATH_Symbols_B}', "");
Expect(0, 10752, '\P{^Blk=-Misc_MATH_Symbols_B}', "");
Error('\p{Is_Block=:=Miscellaneous_MATHEMATICAL_Symbols_b}');
Error('\P{Is_Block=:=Miscellaneous_MATHEMATICAL_Symbols_b}');
Expect(1, 10751, '\p{Is_Block: miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\p{^Is_Block: miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\P{Is_Block: miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\P{^Is_Block: miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\p{Is_Block: miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\p{^Is_Block: miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\P{Is_Block: miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\P{^Is_Block: miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\p{Is_Block:-	Miscellaneous_mathematical_SYMBOLS_B}', "");
Expect(0, 10751, '\p{^Is_Block:-	Miscellaneous_mathematical_SYMBOLS_B}', "");
Expect(0, 10751, '\P{Is_Block:-	Miscellaneous_mathematical_SYMBOLS_B}', "");
Expect(1, 10751, '\P{^Is_Block:-	Miscellaneous_mathematical_SYMBOLS_B}', "");
Expect(0, 10752, '\p{Is_Block:-	Miscellaneous_mathematical_SYMBOLS_B}', "");
Expect(1, 10752, '\p{^Is_Block:-	Miscellaneous_mathematical_SYMBOLS_B}', "");
Expect(1, 10752, '\P{Is_Block:-	Miscellaneous_mathematical_SYMBOLS_B}', "");
Expect(0, 10752, '\P{^Is_Block:-	Miscellaneous_mathematical_SYMBOLS_B}', "");
Error('\p{Is_Blk=		Misc_MATH_Symbols_B/a/}');
Error('\P{Is_Blk=		Misc_MATH_Symbols_B/a/}');
Expect(1, 10751, '\p{Is_Blk=miscmathsymbolsb}', "");
Expect(0, 10751, '\p{^Is_Blk=miscmathsymbolsb}', "");
Expect(0, 10751, '\P{Is_Blk=miscmathsymbolsb}', "");
Expect(1, 10751, '\P{^Is_Blk=miscmathsymbolsb}', "");
Expect(0, 10752, '\p{Is_Blk=miscmathsymbolsb}', "");
Expect(1, 10752, '\p{^Is_Blk=miscmathsymbolsb}', "");
Expect(1, 10752, '\P{Is_Blk=miscmathsymbolsb}', "");
Expect(0, 10752, '\P{^Is_Blk=miscmathsymbolsb}', "");
Expect(1, 10751, '\p{Is_Blk= misc_MATH_Symbols_B}', "");
Expect(0, 10751, '\p{^Is_Blk= misc_MATH_Symbols_B}', "");
Expect(0, 10751, '\P{Is_Blk= misc_MATH_Symbols_B}', "");
Expect(1, 10751, '\P{^Is_Blk= misc_MATH_Symbols_B}', "");
Expect(0, 10752, '\p{Is_Blk= misc_MATH_Symbols_B}', "");
Expect(1, 10752, '\p{^Is_Blk= misc_MATH_Symbols_B}', "");
Expect(1, 10752, '\P{Is_Blk= misc_MATH_Symbols_B}', "");
Expect(0, 10752, '\P{^Is_Blk= misc_MATH_Symbols_B}', "");
Error('\p{Block=/a/MISCELLANEOUS_SYMBOLS_AND_Pictographs}');
Error('\P{Block=/a/MISCELLANEOUS_SYMBOLS_AND_Pictographs}');
Expect(1, 128511, '\p{Block=miscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\p{^Block=miscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\P{Block=miscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\P{^Block=miscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\p{Block=miscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\p{^Block=miscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\P{Block=miscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\P{^Block=miscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\p{Block=_ miscellaneous_SYMBOLS_AND_Pictographs}', "");
Expect(0, 128511, '\p{^Block=_ miscellaneous_SYMBOLS_AND_Pictographs}', "");
Expect(0, 128511, '\P{Block=_ miscellaneous_SYMBOLS_AND_Pictographs}', "");
Expect(1, 128511, '\P{^Block=_ miscellaneous_SYMBOLS_AND_Pictographs}', "");
Expect(0, 128512, '\p{Block=_ miscellaneous_SYMBOLS_AND_Pictographs}', "");
Expect(1, 128512, '\p{^Block=_ miscellaneous_SYMBOLS_AND_Pictographs}', "");
Expect(1, 128512, '\P{Block=_ miscellaneous_SYMBOLS_AND_Pictographs}', "");
Expect(0, 128512, '\P{^Block=_ miscellaneous_SYMBOLS_AND_Pictographs}', "");
Error('\p{Blk=	 Misc_pictographs:=}');
Error('\P{Blk=	 Misc_pictographs:=}');
Expect(1, 128511, '\p{Blk=miscpictographs}', "");
Expect(0, 128511, '\p{^Blk=miscpictographs}', "");
Expect(0, 128511, '\P{Blk=miscpictographs}', "");
Expect(1, 128511, '\P{^Blk=miscpictographs}', "");
Expect(0, 128512, '\p{Blk=miscpictographs}', "");
Expect(1, 128512, '\p{^Blk=miscpictographs}', "");
Expect(1, 128512, '\P{Blk=miscpictographs}', "");
Expect(0, 128512, '\P{^Blk=miscpictographs}', "");
Expect(1, 128511, '\p{Blk: _Misc_pictographs}', "");
Expect(0, 128511, '\p{^Blk: _Misc_pictographs}', "");
Expect(0, 128511, '\P{Blk: _Misc_pictographs}', "");
Expect(1, 128511, '\P{^Blk: _Misc_pictographs}', "");
Expect(0, 128512, '\p{Blk: _Misc_pictographs}', "");
Expect(1, 128512, '\p{^Blk: _Misc_pictographs}', "");
Expect(1, 128512, '\P{Blk: _Misc_pictographs}', "");
Expect(0, 128512, '\P{^Blk: _Misc_pictographs}', "");
Error('\p{Is_Block=  Miscellaneous_symbols_And_Pictographs:=}');
Error('\P{Is_Block=  Miscellaneous_symbols_And_Pictographs:=}');
Expect(1, 128511, '\p{Is_Block=miscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\p{^Is_Block=miscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\P{Is_Block=miscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\P{^Is_Block=miscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\p{Is_Block=miscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\p{^Is_Block=miscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\P{Is_Block=miscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\P{^Is_Block=miscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\p{Is_Block=_Miscellaneous_Symbols_and_pictographs}', "");
Expect(0, 128511, '\p{^Is_Block=_Miscellaneous_Symbols_and_pictographs}', "");
Expect(0, 128511, '\P{Is_Block=_Miscellaneous_Symbols_and_pictographs}', "");
Expect(1, 128511, '\P{^Is_Block=_Miscellaneous_Symbols_and_pictographs}', "");
Expect(0, 128512, '\p{Is_Block=_Miscellaneous_Symbols_and_pictographs}', "");
Expect(1, 128512, '\p{^Is_Block=_Miscellaneous_Symbols_and_pictographs}', "");
Expect(1, 128512, '\P{Is_Block=_Miscellaneous_Symbols_and_pictographs}', "");
Expect(0, 128512, '\P{^Is_Block=_Miscellaneous_Symbols_and_pictographs}', "");
Error('\p{Is_Blk=-Misc_Pictographs/a/}');
Error('\P{Is_Blk=-Misc_Pictographs/a/}');
Expect(1, 128511, '\p{Is_Blk=miscpictographs}', "");
Expect(0, 128511, '\p{^Is_Blk=miscpictographs}', "");
Expect(0, 128511, '\P{Is_Blk=miscpictographs}', "");
Expect(1, 128511, '\P{^Is_Blk=miscpictographs}', "");
Expect(0, 128512, '\p{Is_Blk=miscpictographs}', "");
Expect(1, 128512, '\p{^Is_Blk=miscpictographs}', "");
Expect(1, 128512, '\P{Is_Blk=miscpictographs}', "");
Expect(0, 128512, '\P{^Is_Blk=miscpictographs}', "");
Expect(1, 128511, '\p{Is_Blk=	_Misc_pictographs}', "");
Expect(0, 128511, '\p{^Is_Blk=	_Misc_pictographs}', "");
Expect(0, 128511, '\P{Is_Blk=	_Misc_pictographs}', "");
Expect(1, 128511, '\P{^Is_Blk=	_Misc_pictographs}', "");
Expect(0, 128512, '\p{Is_Blk=	_Misc_pictographs}', "");
Expect(1, 128512, '\p{^Is_Blk=	_Misc_pictographs}', "");
Expect(1, 128512, '\P{Is_Blk=	_Misc_pictographs}', "");
Expect(0, 128512, '\P{^Is_Blk=	_Misc_pictographs}', "");
Error('\p{Block=/a/_-miscellaneous_SYMBOLS}');
Error('\P{Block=/a/_-miscellaneous_SYMBOLS}');
Expect(1, 9983, '\p{Block:	miscellaneoussymbols}', "");
Expect(0, 9983, '\p{^Block:	miscellaneoussymbols}', "");
Expect(0, 9983, '\P{Block:	miscellaneoussymbols}', "");
Expect(1, 9983, '\P{^Block:	miscellaneoussymbols}', "");
Expect(0, 9984, '\p{Block:	miscellaneoussymbols}', "");
Expect(1, 9984, '\p{^Block:	miscellaneoussymbols}', "");
Expect(1, 9984, '\P{Block:	miscellaneoussymbols}', "");
Expect(0, 9984, '\P{^Block:	miscellaneoussymbols}', "");
Expect(1, 9983, '\p{Block= _MISCELLANEOUS_Symbols}', "");
Expect(0, 9983, '\p{^Block= _MISCELLANEOUS_Symbols}', "");
Expect(0, 9983, '\P{Block= _MISCELLANEOUS_Symbols}', "");
Expect(1, 9983, '\P{^Block= _MISCELLANEOUS_Symbols}', "");
Expect(0, 9984, '\p{Block= _MISCELLANEOUS_Symbols}', "");
Expect(1, 9984, '\p{^Block= _MISCELLANEOUS_Symbols}', "");
Expect(1, 9984, '\P{Block= _MISCELLANEOUS_Symbols}', "");
Expect(0, 9984, '\P{^Block= _MISCELLANEOUS_Symbols}', "");
Error('\p{Blk=-_Misc_Symbols:=}');
Error('\P{Blk=-_Misc_Symbols:=}');
Expect(1, 9983, '\p{Blk=miscsymbols}', "");
Expect(0, 9983, '\p{^Blk=miscsymbols}', "");
Expect(0, 9983, '\P{Blk=miscsymbols}', "");
Expect(1, 9983, '\P{^Blk=miscsymbols}', "");
Expect(0, 9984, '\p{Blk=miscsymbols}', "");
Expect(1, 9984, '\p{^Blk=miscsymbols}', "");
Expect(1, 9984, '\P{Blk=miscsymbols}', "");
Expect(0, 9984, '\P{^Blk=miscsymbols}', "");
Expect(1, 9983, '\p{Blk=-Misc_symbols}', "");
Expect(0, 9983, '\p{^Blk=-Misc_symbols}', "");
Expect(0, 9983, '\P{Blk=-Misc_symbols}', "");
Expect(1, 9983, '\P{^Blk=-Misc_symbols}', "");
Expect(0, 9984, '\p{Blk=-Misc_symbols}', "");
Expect(1, 9984, '\p{^Blk=-Misc_symbols}', "");
Expect(1, 9984, '\P{Blk=-Misc_symbols}', "");
Expect(0, 9984, '\P{^Blk=-Misc_symbols}', "");
Error('\p{Is_Block= _miscellaneous_SYMBOLS:=}');
Error('\P{Is_Block= _miscellaneous_SYMBOLS:=}');
Expect(1, 9983, '\p{Is_Block=miscellaneoussymbols}', "");
Expect(0, 9983, '\p{^Is_Block=miscellaneoussymbols}', "");
Expect(0, 9983, '\P{Is_Block=miscellaneoussymbols}', "");
Expect(1, 9983, '\P{^Is_Block=miscellaneoussymbols}', "");
Expect(0, 9984, '\p{Is_Block=miscellaneoussymbols}', "");
Expect(1, 9984, '\p{^Is_Block=miscellaneoussymbols}', "");
Expect(1, 9984, '\P{Is_Block=miscellaneoussymbols}', "");
Expect(0, 9984, '\P{^Is_Block=miscellaneoussymbols}', "");
Expect(1, 9983, '\p{Is_Block=_Miscellaneous_Symbols}', "");
Expect(0, 9983, '\p{^Is_Block=_Miscellaneous_Symbols}', "");
Expect(0, 9983, '\P{Is_Block=_Miscellaneous_Symbols}', "");
Expect(1, 9983, '\P{^Is_Block=_Miscellaneous_Symbols}', "");
Expect(0, 9984, '\p{Is_Block=_Miscellaneous_Symbols}', "");
Expect(1, 9984, '\p{^Is_Block=_Miscellaneous_Symbols}', "");
Expect(1, 9984, '\P{Is_Block=_Miscellaneous_Symbols}', "");
Expect(0, 9984, '\P{^Is_Block=_Miscellaneous_Symbols}', "");
Error('\p{Is_Blk=:= misc_symbols}');
Error('\P{Is_Blk=:= misc_symbols}');
Expect(1, 9983, '\p{Is_Blk=miscsymbols}', "");
Expect(0, 9983, '\p{^Is_Blk=miscsymbols}', "");
Expect(0, 9983, '\P{Is_Blk=miscsymbols}', "");
Expect(1, 9983, '\P{^Is_Blk=miscsymbols}', "");
Expect(0, 9984, '\p{Is_Blk=miscsymbols}', "");
Expect(1, 9984, '\p{^Is_Blk=miscsymbols}', "");
Expect(1, 9984, '\P{Is_Blk=miscsymbols}', "");
Expect(0, 9984, '\P{^Is_Blk=miscsymbols}', "");
Expect(1, 9983, '\p{Is_Blk=__MISC_Symbols}', "");
Expect(0, 9983, '\p{^Is_Blk=__MISC_Symbols}', "");
Expect(0, 9983, '\P{Is_Blk=__MISC_Symbols}', "");
Expect(1, 9983, '\P{^Is_Blk=__MISC_Symbols}', "");
Expect(0, 9984, '\p{Is_Blk=__MISC_Symbols}', "");
Expect(1, 9984, '\p{^Is_Blk=__MISC_Symbols}', "");
Expect(1, 9984, '\P{Is_Blk=__MISC_Symbols}', "");
Expect(0, 9984, '\P{^Is_Blk=__MISC_Symbols}', "");
Error('\p{Block=-:=Miscellaneous_Technical}');
Error('\P{Block=-:=Miscellaneous_Technical}');
Expect(1, 9215, '\p{Block=miscellaneoustechnical}', "");
Expect(0, 9215, '\p{^Block=miscellaneoustechnical}', "");
Expect(0, 9215, '\P{Block=miscellaneoustechnical}', "");
Expect(1, 9215, '\P{^Block=miscellaneoustechnical}', "");
Expect(0, 9216, '\p{Block=miscellaneoustechnical}', "");
Expect(1, 9216, '\p{^Block=miscellaneoustechnical}', "");
Expect(1, 9216, '\P{Block=miscellaneoustechnical}', "");
Expect(0, 9216, '\P{^Block=miscellaneoustechnical}', "");
Expect(1, 9215, '\p{Block=_	miscellaneous_Technical}', "");
Expect(0, 9215, '\p{^Block=_	miscellaneous_Technical}', "");
Expect(0, 9215, '\P{Block=_	miscellaneous_Technical}', "");
Expect(1, 9215, '\P{^Block=_	miscellaneous_Technical}', "");
Expect(0, 9216, '\p{Block=_	miscellaneous_Technical}', "");
Expect(1, 9216, '\p{^Block=_	miscellaneous_Technical}', "");
Expect(1, 9216, '\P{Block=_	miscellaneous_Technical}', "");
Expect(0, 9216, '\P{^Block=_	miscellaneous_Technical}', "");
Error('\p{Blk=/a/_Misc_Technical}');
Error('\P{Blk=/a/_Misc_Technical}');
Expect(1, 9215, '\p{Blk=misctechnical}', "");
Expect(0, 9215, '\p{^Blk=misctechnical}', "");
Expect(0, 9215, '\P{Blk=misctechnical}', "");
Expect(1, 9215, '\P{^Blk=misctechnical}', "");
Expect(0, 9216, '\p{Blk=misctechnical}', "");
Expect(1, 9216, '\p{^Blk=misctechnical}', "");
Expect(1, 9216, '\P{Blk=misctechnical}', "");
Expect(0, 9216, '\P{^Blk=misctechnical}', "");
Expect(1, 9215, '\p{Blk=- MISC_Technical}', "");
Expect(0, 9215, '\p{^Blk=- MISC_Technical}', "");
Expect(0, 9215, '\P{Blk=- MISC_Technical}', "");
Expect(1, 9215, '\P{^Blk=- MISC_Technical}', "");
Expect(0, 9216, '\p{Blk=- MISC_Technical}', "");
Expect(1, 9216, '\p{^Blk=- MISC_Technical}', "");
Expect(1, 9216, '\P{Blk=- MISC_Technical}', "");
Expect(0, 9216, '\P{^Blk=- MISC_Technical}', "");
Error('\p{Is_Block=:=_Miscellaneous_TECHNICAL}');
Error('\P{Is_Block=:=_Miscellaneous_TECHNICAL}');
Expect(1, 9215, '\p{Is_Block=miscellaneoustechnical}', "");
Expect(0, 9215, '\p{^Is_Block=miscellaneoustechnical}', "");
Expect(0, 9215, '\P{Is_Block=miscellaneoustechnical}', "");
Expect(1, 9215, '\P{^Is_Block=miscellaneoustechnical}', "");
Expect(0, 9216, '\p{Is_Block=miscellaneoustechnical}', "");
Expect(1, 9216, '\p{^Is_Block=miscellaneoustechnical}', "");
Expect(1, 9216, '\P{Is_Block=miscellaneoustechnical}', "");
Expect(0, 9216, '\P{^Is_Block=miscellaneoustechnical}', "");
Expect(1, 9215, '\p{Is_Block=	 Miscellaneous_Technical}', "");
Expect(0, 9215, '\p{^Is_Block=	 Miscellaneous_Technical}', "");
Expect(0, 9215, '\P{Is_Block=	 Miscellaneous_Technical}', "");
Expect(1, 9215, '\P{^Is_Block=	 Miscellaneous_Technical}', "");
Expect(0, 9216, '\p{Is_Block=	 Miscellaneous_Technical}', "");
Expect(1, 9216, '\p{^Is_Block=	 Miscellaneous_Technical}', "");
Expect(1, 9216, '\P{Is_Block=	 Miscellaneous_Technical}', "");
Expect(0, 9216, '\P{^Is_Block=	 Miscellaneous_Technical}', "");
Error('\p{Is_Blk=/a/	misc_Technical}');
Error('\P{Is_Blk=/a/	misc_Technical}');
Expect(1, 9215, '\p{Is_Blk=misctechnical}', "");
Expect(0, 9215, '\p{^Is_Blk=misctechnical}', "");
Expect(0, 9215, '\P{Is_Blk=misctechnical}', "");
Expect(1, 9215, '\P{^Is_Blk=misctechnical}', "");
Expect(0, 9216, '\p{Is_Blk=misctechnical}', "");
Expect(1, 9216, '\p{^Is_Blk=misctechnical}', "");
Expect(1, 9216, '\P{Is_Blk=misctechnical}', "");
Expect(0, 9216, '\P{^Is_Blk=misctechnical}', "");
Expect(1, 9215, '\p{Is_Blk= -MISC_technical}', "");
Expect(0, 9215, '\p{^Is_Blk= -MISC_technical}', "");
Expect(0, 9215, '\P{Is_Blk= -MISC_technical}', "");
Expect(1, 9215, '\P{^Is_Blk= -MISC_technical}', "");
Expect(0, 9216, '\p{Is_Blk= -MISC_technical}', "");
Expect(1, 9216, '\p{^Is_Blk= -MISC_technical}', "");
Expect(1, 9216, '\P{Is_Blk= -MISC_technical}', "");
Expect(0, 9216, '\P{^Is_Blk= -MISC_technical}', "");
Error('\p{Block=__Spacing_MODIFIER_LETTERS:=}');
Error('\P{Block=__Spacing_MODIFIER_LETTERS:=}');
Expect(1, 767, '\p{Block:   spacingmodifierletters}', "");
Expect(0, 767, '\p{^Block:   spacingmodifierletters}', "");
Expect(0, 767, '\P{Block:   spacingmodifierletters}', "");
Expect(1, 767, '\P{^Block:   spacingmodifierletters}', "");
Expect(0, 768, '\p{Block:   spacingmodifierletters}', "");
Expect(1, 768, '\p{^Block:   spacingmodifierletters}', "");
Expect(1, 768, '\P{Block:   spacingmodifierletters}', "");
Expect(0, 768, '\P{^Block:   spacingmodifierletters}', "");
Expect(1, 767, '\p{Block= Spacing_Modifier_letters}', "");
Expect(0, 767, '\p{^Block= Spacing_Modifier_letters}', "");
Expect(0, 767, '\P{Block= Spacing_Modifier_letters}', "");
Expect(1, 767, '\P{^Block= Spacing_Modifier_letters}', "");
Expect(0, 768, '\p{Block= Spacing_Modifier_letters}', "");
Expect(1, 768, '\p{^Block= Spacing_Modifier_letters}', "");
Expect(1, 768, '\P{Block= Spacing_Modifier_letters}', "");
Expect(0, 768, '\P{^Block= Spacing_Modifier_letters}', "");
Error('\p{Blk=Modifier_Letters:=}');
Error('\P{Blk=Modifier_Letters:=}');
Expect(1, 767, '\p{Blk=modifierletters}', "");
Expect(0, 767, '\p{^Blk=modifierletters}', "");
Expect(0, 767, '\P{Blk=modifierletters}', "");
Expect(1, 767, '\P{^Blk=modifierletters}', "");
Expect(0, 768, '\p{Blk=modifierletters}', "");
Expect(1, 768, '\p{^Blk=modifierletters}', "");
Expect(1, 768, '\P{Blk=modifierletters}', "");
Expect(0, 768, '\P{^Blk=modifierletters}', "");
Expect(1, 767, '\p{Blk=-Modifier_Letters}', "");
Expect(0, 767, '\p{^Blk=-Modifier_Letters}', "");
Expect(0, 767, '\P{Blk=-Modifier_Letters}', "");
Expect(1, 767, '\P{^Blk=-Modifier_Letters}', "");
Expect(0, 768, '\p{Blk=-Modifier_Letters}', "");
Expect(1, 768, '\p{^Blk=-Modifier_Letters}', "");
Expect(1, 768, '\P{Blk=-Modifier_Letters}', "");
Expect(0, 768, '\P{^Blk=-Modifier_Letters}', "");
Error('\p{Is_Block=-/a/SPACING_Modifier_letters}');
Error('\P{Is_Block=-/a/SPACING_Modifier_letters}');
Expect(1, 767, '\p{Is_Block=spacingmodifierletters}', "");
Expect(0, 767, '\p{^Is_Block=spacingmodifierletters}', "");
Expect(0, 767, '\P{Is_Block=spacingmodifierletters}', "");
Expect(1, 767, '\P{^Is_Block=spacingmodifierletters}', "");
Expect(0, 768, '\p{Is_Block=spacingmodifierletters}', "");
Expect(1, 768, '\p{^Is_Block=spacingmodifierletters}', "");
Expect(1, 768, '\P{Is_Block=spacingmodifierletters}', "");
Expect(0, 768, '\P{^Is_Block=spacingmodifierletters}', "");
Expect(1, 767, '\p{Is_Block= -SPACING_Modifier_LETTERS}', "");
Expect(0, 767, '\p{^Is_Block= -SPACING_Modifier_LETTERS}', "");
Expect(0, 767, '\P{Is_Block= -SPACING_Modifier_LETTERS}', "");
Expect(1, 767, '\P{^Is_Block= -SPACING_Modifier_LETTERS}', "");
Expect(0, 768, '\p{Is_Block= -SPACING_Modifier_LETTERS}', "");
Expect(1, 768, '\p{^Is_Block= -SPACING_Modifier_LETTERS}', "");
Expect(1, 768, '\P{Is_Block= -SPACING_Modifier_LETTERS}', "");
Expect(0, 768, '\P{^Is_Block= -SPACING_Modifier_LETTERS}', "");
Error('\p{Is_Blk=:=	_modifier_letters}');
Error('\P{Is_Blk=:=	_modifier_letters}');
Expect(1, 767, '\p{Is_Blk=modifierletters}', "");
Expect(0, 767, '\p{^Is_Blk=modifierletters}', "");
Expect(0, 767, '\P{Is_Blk=modifierletters}', "");
Expect(1, 767, '\P{^Is_Blk=modifierletters}', "");
Expect(0, 768, '\p{Is_Blk=modifierletters}', "");
Expect(1, 768, '\p{^Is_Blk=modifierletters}', "");
Expect(1, 768, '\P{Is_Blk=modifierletters}', "");
Expect(0, 768, '\P{^Is_Blk=modifierletters}', "");
Expect(1, 767, '\p{Is_Blk=-_Modifier_Letters}', "");
Expect(0, 767, '\p{^Is_Blk=-_Modifier_Letters}', "");
Expect(0, 767, '\P{Is_Blk=-_Modifier_Letters}', "");
Expect(1, 767, '\P{^Is_Blk=-_Modifier_Letters}', "");
Expect(0, 768, '\p{Is_Blk=-_Modifier_Letters}', "");
Expect(1, 768, '\p{^Is_Blk=-_Modifier_Letters}', "");
Expect(1, 768, '\P{Is_Blk=-_Modifier_Letters}', "");
Expect(0, 768, '\P{^Is_Blk=-_Modifier_Letters}', "");
Error('\p{Block=		Modifier_Tone_Letters:=}');
Error('\P{Block=		Modifier_Tone_Letters:=}');
Expect(1, 42783, '\p{Block=modifiertoneletters}', "");
Expect(0, 42783, '\p{^Block=modifiertoneletters}', "");
Expect(0, 42783, '\P{Block=modifiertoneletters}', "");
Expect(1, 42783, '\P{^Block=modifiertoneletters}', "");
Expect(0, 42784, '\p{Block=modifiertoneletters}', "");
Expect(1, 42784, '\p{^Block=modifiertoneletters}', "");
Expect(1, 42784, '\P{Block=modifiertoneletters}', "");
Expect(0, 42784, '\P{^Block=modifiertoneletters}', "");
Expect(1, 42783, '\p{Block= _MODIFIER_TONE_Letters}', "");
Expect(0, 42783, '\p{^Block= _MODIFIER_TONE_Letters}', "");
Expect(0, 42783, '\P{Block= _MODIFIER_TONE_Letters}', "");
Expect(1, 42783, '\P{^Block= _MODIFIER_TONE_Letters}', "");
Expect(0, 42784, '\p{Block= _MODIFIER_TONE_Letters}', "");
Expect(1, 42784, '\p{^Block= _MODIFIER_TONE_Letters}', "");
Expect(1, 42784, '\P{Block= _MODIFIER_TONE_Letters}', "");
Expect(0, 42784, '\P{^Block= _MODIFIER_TONE_Letters}', "");
Error('\p{Blk: 		modifier_Tone_Letters:=}');
Error('\P{Blk: 		modifier_Tone_Letters:=}');
Expect(1, 42783, '\p{Blk=modifiertoneletters}', "");
Expect(0, 42783, '\p{^Blk=modifiertoneletters}', "");
Expect(0, 42783, '\P{Blk=modifiertoneletters}', "");
Expect(1, 42783, '\P{^Blk=modifiertoneletters}', "");
Expect(0, 42784, '\p{Blk=modifiertoneletters}', "");
Expect(1, 42784, '\p{^Blk=modifiertoneletters}', "");
Expect(1, 42784, '\P{Blk=modifiertoneletters}', "");
Expect(0, 42784, '\P{^Blk=modifiertoneletters}', "");
Expect(1, 42783, '\p{Blk=	 modifier_tone_Letters}', "");
Expect(0, 42783, '\p{^Blk=	 modifier_tone_Letters}', "");
Expect(0, 42783, '\P{Blk=	 modifier_tone_Letters}', "");
Expect(1, 42783, '\P{^Blk=	 modifier_tone_Letters}', "");
Expect(0, 42784, '\p{Blk=	 modifier_tone_Letters}', "");
Expect(1, 42784, '\p{^Blk=	 modifier_tone_Letters}', "");
Expect(1, 42784, '\P{Blk=	 modifier_tone_Letters}', "");
Expect(0, 42784, '\P{^Blk=	 modifier_tone_Letters}', "");
Error('\p{Is_Block=	Modifier_tone_LETTERS/a/}');
Error('\P{Is_Block=	Modifier_tone_LETTERS/a/}');
Expect(1, 42783, '\p{Is_Block=modifiertoneletters}', "");
Expect(0, 42783, '\p{^Is_Block=modifiertoneletters}', "");
Expect(0, 42783, '\P{Is_Block=modifiertoneletters}', "");
Expect(1, 42783, '\P{^Is_Block=modifiertoneletters}', "");
Expect(0, 42784, '\p{Is_Block=modifiertoneletters}', "");
Expect(1, 42784, '\p{^Is_Block=modifiertoneletters}', "");
Expect(1, 42784, '\P{Is_Block=modifiertoneletters}', "");
Expect(0, 42784, '\P{^Is_Block=modifiertoneletters}', "");
Expect(1, 42783, '\p{Is_Block=	 MODIFIER_Tone_letters}', "");
Expect(0, 42783, '\p{^Is_Block=	 MODIFIER_Tone_letters}', "");
Expect(0, 42783, '\P{Is_Block=	 MODIFIER_Tone_letters}', "");
Expect(1, 42783, '\P{^Is_Block=	 MODIFIER_Tone_letters}', "");
Expect(0, 42784, '\p{Is_Block=	 MODIFIER_Tone_letters}', "");
Expect(1, 42784, '\p{^Is_Block=	 MODIFIER_Tone_letters}', "");
Expect(1, 42784, '\P{Is_Block=	 MODIFIER_Tone_letters}', "");
Expect(0, 42784, '\P{^Is_Block=	 MODIFIER_Tone_letters}', "");
Error('\p{Is_Blk=/a/_-modifier_tone_LETTERS}');
Error('\P{Is_Blk=/a/_-modifier_tone_LETTERS}');
Expect(1, 42783, '\p{Is_Blk=modifiertoneletters}', "");
Expect(0, 42783, '\p{^Is_Blk=modifiertoneletters}', "");
Expect(0, 42783, '\P{Is_Blk=modifiertoneletters}', "");
Expect(1, 42783, '\P{^Is_Blk=modifiertoneletters}', "");
Expect(0, 42784, '\p{Is_Blk=modifiertoneletters}', "");
Expect(1, 42784, '\p{^Is_Blk=modifiertoneletters}', "");
Expect(1, 42784, '\P{Is_Blk=modifiertoneletters}', "");
Expect(0, 42784, '\P{^Is_Blk=modifiertoneletters}', "");
Expect(1, 42783, '\p{Is_Blk:	-	Modifier_Tone_Letters}', "");
Expect(0, 42783, '\p{^Is_Blk:	-	Modifier_Tone_Letters}', "");
Expect(0, 42783, '\P{Is_Blk:	-	Modifier_Tone_Letters}', "");
Expect(1, 42783, '\P{^Is_Blk:	-	Modifier_Tone_Letters}', "");
Expect(0, 42784, '\p{Is_Blk:	-	Modifier_Tone_Letters}', "");
Expect(1, 42784, '\p{^Is_Blk:	-	Modifier_Tone_Letters}', "");
Expect(1, 42784, '\P{Is_Blk:	-	Modifier_Tone_Letters}', "");
Expect(0, 42784, '\P{^Is_Blk:	-	Modifier_Tone_Letters}', "");
Error('\p{Block=__Mongolian:=}');
Error('\P{Block=__Mongolian:=}');
Expect(1, 6319, '\p{Block=mongolian}', "");
Expect(0, 6319, '\p{^Block=mongolian}', "");
Expect(0, 6319, '\P{Block=mongolian}', "");
Expect(1, 6319, '\P{^Block=mongolian}', "");
Expect(0, 6320, '\p{Block=mongolian}', "");
Expect(1, 6320, '\p{^Block=mongolian}', "");
Expect(1, 6320, '\P{Block=mongolian}', "");
Expect(0, 6320, '\P{^Block=mongolian}', "");
Expect(1, 6319, '\p{Block=-_Mongolian}', "");
Expect(0, 6319, '\p{^Block=-_Mongolian}', "");
Expect(0, 6319, '\P{Block=-_Mongolian}', "");
Expect(1, 6319, '\P{^Block=-_Mongolian}', "");
Expect(0, 6320, '\p{Block=-_Mongolian}', "");
Expect(1, 6320, '\p{^Block=-_Mongolian}', "");
Expect(1, 6320, '\P{Block=-_Mongolian}', "");
Expect(0, 6320, '\P{^Block=-_Mongolian}', "");
Error('\p{Blk=		Mongolian/a/}');
Error('\P{Blk=		Mongolian/a/}');
Expect(1, 6319, '\p{Blk=mongolian}', "");
Expect(0, 6319, '\p{^Blk=mongolian}', "");
Expect(0, 6319, '\P{Blk=mongolian}', "");
Expect(1, 6319, '\P{^Blk=mongolian}', "");
Expect(0, 6320, '\p{Blk=mongolian}', "");
Expect(1, 6320, '\p{^Blk=mongolian}', "");
Expect(1, 6320, '\P{Blk=mongolian}', "");
Expect(0, 6320, '\P{^Blk=mongolian}', "");
Expect(1, 6319, '\p{Blk=-MONGOLIAN}', "");
Expect(0, 6319, '\p{^Blk=-MONGOLIAN}', "");
Expect(0, 6319, '\P{Blk=-MONGOLIAN}', "");
Expect(1, 6319, '\P{^Blk=-MONGOLIAN}', "");
Expect(0, 6320, '\p{Blk=-MONGOLIAN}', "");
Expect(1, 6320, '\p{^Blk=-MONGOLIAN}', "");
Expect(1, 6320, '\P{Blk=-MONGOLIAN}', "");
Expect(0, 6320, '\P{^Blk=-MONGOLIAN}', "");
Error('\p{Is_Block= :=mongolian}');
Error('\P{Is_Block= :=mongolian}');
Expect(1, 6319, '\p{Is_Block: mongolian}', "");
Expect(0, 6319, '\p{^Is_Block: mongolian}', "");
Expect(0, 6319, '\P{Is_Block: mongolian}', "");
Expect(1, 6319, '\P{^Is_Block: mongolian}', "");
Expect(0, 6320, '\p{Is_Block: mongolian}', "");
Expect(1, 6320, '\p{^Is_Block: mongolian}', "");
Expect(1, 6320, '\P{Is_Block: mongolian}', "");
Expect(0, 6320, '\P{^Is_Block: mongolian}', "");
Expect(1, 6319, '\p{Is_Block=_mongolian}', "");
Expect(0, 6319, '\p{^Is_Block=_mongolian}', "");
Expect(0, 6319, '\P{Is_Block=_mongolian}', "");
Expect(1, 6319, '\P{^Is_Block=_mongolian}', "");
Expect(0, 6320, '\p{Is_Block=_mongolian}', "");
Expect(1, 6320, '\p{^Is_Block=_mongolian}', "");
Expect(1, 6320, '\P{Is_Block=_mongolian}', "");
Expect(0, 6320, '\P{^Is_Block=_mongolian}', "");
Error('\p{Is_Blk= :=MONGOLIAN}');
Error('\P{Is_Blk= :=MONGOLIAN}');
Expect(1, 6319, '\p{Is_Blk=mongolian}', "");
Expect(0, 6319, '\p{^Is_Blk=mongolian}', "");
Expect(0, 6319, '\P{Is_Blk=mongolian}', "");
Expect(1, 6319, '\P{^Is_Blk=mongolian}', "");
Expect(0, 6320, '\p{Is_Blk=mongolian}', "");
Expect(1, 6320, '\p{^Is_Blk=mongolian}', "");
Expect(1, 6320, '\P{Is_Blk=mongolian}', "");
Expect(0, 6320, '\P{^Is_Blk=mongolian}', "");
Expect(1, 6319, '\p{Is_Blk= _Mongolian}', "");
Expect(0, 6319, '\p{^Is_Blk= _Mongolian}', "");
Expect(0, 6319, '\P{Is_Blk= _Mongolian}', "");
Expect(1, 6319, '\P{^Is_Blk= _Mongolian}', "");
Expect(0, 6320, '\p{Is_Blk= _Mongolian}', "");
Expect(1, 6320, '\p{^Is_Blk= _Mongolian}', "");
Expect(1, 6320, '\P{Is_Blk= _Mongolian}', "");
Expect(0, 6320, '\P{^Is_Blk= _Mongolian}', "");
Error('\p{Block:	-:=Musical_symbols}');
Error('\P{Block:	-:=Musical_symbols}');
Expect(1, 119295, '\p{Block=musicalsymbols}', "");
Expect(0, 119295, '\p{^Block=musicalsymbols}', "");
Expect(0, 119295, '\P{Block=musicalsymbols}', "");
Expect(1, 119295, '\P{^Block=musicalsymbols}', "");
Expect(0, 119296, '\p{Block=musicalsymbols}', "");
Expect(1, 119296, '\p{^Block=musicalsymbols}', "");
Expect(1, 119296, '\P{Block=musicalsymbols}', "");
Expect(0, 119296, '\P{^Block=musicalsymbols}', "");
Expect(1, 119295, '\p{Block= -musical_Symbols}', "");
Expect(0, 119295, '\p{^Block= -musical_Symbols}', "");
Expect(0, 119295, '\P{Block= -musical_Symbols}', "");
Expect(1, 119295, '\P{^Block= -musical_Symbols}', "");
Expect(0, 119296, '\p{Block= -musical_Symbols}', "");
Expect(1, 119296, '\p{^Block= -musical_Symbols}', "");
Expect(1, 119296, '\P{Block= -musical_Symbols}', "");
Expect(0, 119296, '\P{^Block= -musical_Symbols}', "");
Error('\p{Blk=	MUSIC:=}');
Error('\P{Blk=	MUSIC:=}');
Expect(1, 119295, '\p{Blk=music}', "");
Expect(0, 119295, '\p{^Blk=music}', "");
Expect(0, 119295, '\P{Blk=music}', "");
Expect(1, 119295, '\P{^Blk=music}', "");
Expect(0, 119296, '\p{Blk=music}', "");
Expect(1, 119296, '\p{^Blk=music}', "");
Expect(1, 119296, '\P{Blk=music}', "");
Expect(0, 119296, '\P{^Blk=music}', "");
Expect(1, 119295, '\p{Blk=	-Music}', "");
Expect(0, 119295, '\p{^Blk=	-Music}', "");
Expect(0, 119295, '\P{Blk=	-Music}', "");
Expect(1, 119295, '\P{^Blk=	-Music}', "");
Expect(0, 119296, '\p{Blk=	-Music}', "");
Expect(1, 119296, '\p{^Blk=	-Music}', "");
Expect(1, 119296, '\P{Blk=	-Music}', "");
Expect(0, 119296, '\P{^Blk=	-Music}', "");
Error('\p{Is_Block:/a/Musical_SYMBOLS}');
Error('\P{Is_Block:/a/Musical_SYMBOLS}');
Expect(1, 119295, '\p{Is_Block: musicalsymbols}', "");
Expect(0, 119295, '\p{^Is_Block: musicalsymbols}', "");
Expect(0, 119295, '\P{Is_Block: musicalsymbols}', "");
Expect(1, 119295, '\P{^Is_Block: musicalsymbols}', "");
Expect(0, 119296, '\p{Is_Block: musicalsymbols}', "");
Expect(1, 119296, '\p{^Is_Block: musicalsymbols}', "");
Expect(1, 119296, '\P{Is_Block: musicalsymbols}', "");
Expect(0, 119296, '\P{^Is_Block: musicalsymbols}', "");
Expect(1, 119295, '\p{Is_Block=_Musical_symbols}', "");
Expect(0, 119295, '\p{^Is_Block=_Musical_symbols}', "");
Expect(0, 119295, '\P{Is_Block=_Musical_symbols}', "");
Expect(1, 119295, '\P{^Is_Block=_Musical_symbols}', "");
Expect(0, 119296, '\p{Is_Block=_Musical_symbols}', "");
Expect(1, 119296, '\p{^Is_Block=_Musical_symbols}', "");
Expect(1, 119296, '\P{Is_Block=_Musical_symbols}', "");
Expect(0, 119296, '\P{^Is_Block=_Musical_symbols}', "");
Error('\p{Is_Blk=_Music:=}');
Error('\P{Is_Blk=_Music:=}');
Expect(1, 119295, '\p{Is_Blk=music}', "");
Expect(0, 119295, '\p{^Is_Blk=music}', "");
Expect(0, 119295, '\P{Is_Blk=music}', "");
Expect(1, 119295, '\P{^Is_Blk=music}', "");
Expect(0, 119296, '\p{Is_Blk=music}', "");
Expect(1, 119296, '\p{^Is_Blk=music}', "");
Expect(1, 119296, '\P{Is_Blk=music}', "");
Expect(0, 119296, '\P{^Is_Blk=music}', "");
Expect(1, 119295, '\p{Is_Blk=- Music}', "");
Expect(0, 119295, '\p{^Is_Blk=- Music}', "");
Expect(0, 119295, '\P{Is_Blk=- Music}', "");
Expect(1, 119295, '\P{^Is_Blk=- Music}', "");
Expect(0, 119296, '\p{Is_Blk=- Music}', "");
Expect(1, 119296, '\p{^Is_Blk=- Music}', "");
Expect(1, 119296, '\P{Is_Blk=- Music}', "");
Expect(0, 119296, '\P{^Is_Blk=- Music}', "");
Error('\p{Block=_:=Myanmar}');
Error('\P{Block=_:=Myanmar}');
Expect(1, 4255, '\p{Block=myanmar}', "");
Expect(0, 4255, '\p{^Block=myanmar}', "");
Expect(0, 4255, '\P{Block=myanmar}', "");
Expect(1, 4255, '\P{^Block=myanmar}', "");
Expect(0, 4256, '\p{Block=myanmar}', "");
Expect(1, 4256, '\p{^Block=myanmar}', "");
Expect(1, 4256, '\P{Block=myanmar}', "");
Expect(0, 4256, '\P{^Block=myanmar}', "");
Expect(1, 4255, '\p{Block=_Myanmar}', "");
Expect(0, 4255, '\p{^Block=_Myanmar}', "");
Expect(0, 4255, '\P{Block=_Myanmar}', "");
Expect(1, 4255, '\P{^Block=_Myanmar}', "");
Expect(0, 4256, '\p{Block=_Myanmar}', "");
Expect(1, 4256, '\p{^Block=_Myanmar}', "");
Expect(1, 4256, '\P{Block=_Myanmar}', "");
Expect(0, 4256, '\P{^Block=_Myanmar}', "");
Error('\p{Blk= /a/Myanmar}');
Error('\P{Blk= /a/Myanmar}');
Expect(1, 4255, '\p{Blk=myanmar}', "");
Expect(0, 4255, '\p{^Blk=myanmar}', "");
Expect(0, 4255, '\P{Blk=myanmar}', "");
Expect(1, 4255, '\P{^Blk=myanmar}', "");
Expect(0, 4256, '\p{Blk=myanmar}', "");
Expect(1, 4256, '\p{^Blk=myanmar}', "");
Expect(1, 4256, '\P{Blk=myanmar}', "");
Expect(0, 4256, '\P{^Blk=myanmar}', "");
Expect(1, 4255, '\p{Blk:	_myanmar}', "");
Expect(0, 4255, '\p{^Blk:	_myanmar}', "");
Expect(0, 4255, '\P{Blk:	_myanmar}', "");
Expect(1, 4255, '\P{^Blk:	_myanmar}', "");
Expect(0, 4256, '\p{Blk:	_myanmar}', "");
Expect(1, 4256, '\p{^Blk:	_myanmar}', "");
Expect(1, 4256, '\P{Blk:	_myanmar}', "");
Expect(0, 4256, '\P{^Blk:	_myanmar}', "");
Error('\p{Is_Block=-/a/Myanmar}');
Error('\P{Is_Block=-/a/Myanmar}');
Expect(1, 4255, '\p{Is_Block=myanmar}', "");
Expect(0, 4255, '\p{^Is_Block=myanmar}', "");
Expect(0, 4255, '\P{Is_Block=myanmar}', "");
Expect(1, 4255, '\P{^Is_Block=myanmar}', "");
Expect(0, 4256, '\p{Is_Block=myanmar}', "");
Expect(1, 4256, '\p{^Is_Block=myanmar}', "");
Expect(1, 4256, '\P{Is_Block=myanmar}', "");
Expect(0, 4256, '\P{^Is_Block=myanmar}', "");
Expect(1, 4255, '\p{Is_Block=-_myanmar}', "");
Expect(0, 4255, '\p{^Is_Block=-_myanmar}', "");
Expect(0, 4255, '\P{Is_Block=-_myanmar}', "");
Expect(1, 4255, '\P{^Is_Block=-_myanmar}', "");
Expect(0, 4256, '\p{Is_Block=-_myanmar}', "");
Expect(1, 4256, '\p{^Is_Block=-_myanmar}', "");
Expect(1, 4256, '\P{Is_Block=-_myanmar}', "");
Expect(0, 4256, '\P{^Is_Block=-_myanmar}', "");
Error('\p{Is_Blk=-Myanmar:=}');
Error('\P{Is_Blk=-Myanmar:=}');
Expect(1, 4255, '\p{Is_Blk=myanmar}', "");
Expect(0, 4255, '\p{^Is_Blk=myanmar}', "");
Expect(0, 4255, '\P{Is_Blk=myanmar}', "");
Expect(1, 4255, '\P{^Is_Blk=myanmar}', "");
Expect(0, 4256, '\p{Is_Blk=myanmar}', "");
Expect(1, 4256, '\p{^Is_Blk=myanmar}', "");
Expect(1, 4256, '\P{Is_Blk=myanmar}', "");
Expect(0, 4256, '\P{^Is_Blk=myanmar}', "");
Expect(1, 4255, '\p{Is_Blk=-	myanmar}', "");
Expect(0, 4255, '\p{^Is_Blk=-	myanmar}', "");
Expect(0, 4255, '\P{Is_Blk=-	myanmar}', "");
Expect(1, 4255, '\P{^Is_Blk=-	myanmar}', "");
Expect(0, 4256, '\p{Is_Blk=-	myanmar}', "");
Expect(1, 4256, '\p{^Is_Blk=-	myanmar}', "");
Expect(1, 4256, '\P{Is_Blk=-	myanmar}', "");
Expect(0, 4256, '\P{^Is_Blk=-	myanmar}', "");
Error('\p{Block=:=-myanmar_EXTENDED_a}');
Error('\P{Block=:=-myanmar_EXTENDED_a}');
Expect(1, 43647, '\p{Block: myanmarextendeda}', "");
Expect(0, 43647, '\p{^Block: myanmarextendeda}', "");
Expect(0, 43647, '\P{Block: myanmarextendeda}', "");
Expect(1, 43647, '\P{^Block: myanmarextendeda}', "");
Expect(0, 43648, '\p{Block: myanmarextendeda}', "");
Expect(1, 43648, '\p{^Block: myanmarextendeda}', "");
Expect(1, 43648, '\P{Block: myanmarextendeda}', "");
Expect(0, 43648, '\P{^Block: myanmarextendeda}', "");
Expect(1, 43647, '\p{Block=-	myanmar_extended_A}', "");
Expect(0, 43647, '\p{^Block=-	myanmar_extended_A}', "");
Expect(0, 43647, '\P{Block=-	myanmar_extended_A}', "");
Expect(1, 43647, '\P{^Block=-	myanmar_extended_A}', "");
Expect(0, 43648, '\p{Block=-	myanmar_extended_A}', "");
Expect(1, 43648, '\p{^Block=-	myanmar_extended_A}', "");
Expect(1, 43648, '\P{Block=-	myanmar_extended_A}', "");
Expect(0, 43648, '\P{^Block=-	myanmar_extended_A}', "");
Error('\p{Blk=:=	 myanmar_EXT_A}');
Error('\P{Blk=:=	 myanmar_EXT_A}');
Expect(1, 43647, '\p{Blk=myanmarexta}', "");
Expect(0, 43647, '\p{^Blk=myanmarexta}', "");
Expect(0, 43647, '\P{Blk=myanmarexta}', "");
Expect(1, 43647, '\P{^Blk=myanmarexta}', "");
Expect(0, 43648, '\p{Blk=myanmarexta}', "");
Expect(1, 43648, '\p{^Blk=myanmarexta}', "");
Expect(1, 43648, '\P{Blk=myanmarexta}', "");
Expect(0, 43648, '\P{^Blk=myanmarexta}', "");
Expect(1, 43647, '\p{Blk=-myanmar_Ext_a}', "");
Expect(0, 43647, '\p{^Blk=-myanmar_Ext_a}', "");
Expect(0, 43647, '\P{Blk=-myanmar_Ext_a}', "");
Expect(1, 43647, '\P{^Blk=-myanmar_Ext_a}', "");
Expect(0, 43648, '\p{Blk=-myanmar_Ext_a}', "");
Expect(1, 43648, '\p{^Blk=-myanmar_Ext_a}', "");
Expect(1, 43648, '\P{Blk=-myanmar_Ext_a}', "");
Expect(0, 43648, '\P{^Blk=-myanmar_Ext_a}', "");
Error('\p{Is_Block=/a/_myanmar_Extended_a}');
Error('\P{Is_Block=/a/_myanmar_Extended_a}');
Expect(1, 43647, '\p{Is_Block=myanmarextendeda}', "");
Expect(0, 43647, '\p{^Is_Block=myanmarextendeda}', "");
Expect(0, 43647, '\P{Is_Block=myanmarextendeda}', "");
Expect(1, 43647, '\P{^Is_Block=myanmarextendeda}', "");
Expect(0, 43648, '\p{Is_Block=myanmarextendeda}', "");
Expect(1, 43648, '\p{^Is_Block=myanmarextendeda}', "");
Expect(1, 43648, '\P{Is_Block=myanmarextendeda}', "");
Expect(0, 43648, '\P{^Is_Block=myanmarextendeda}', "");
Expect(1, 43647, '\p{Is_Block= Myanmar_Extended_a}', "");
Expect(0, 43647, '\p{^Is_Block= Myanmar_Extended_a}', "");
Expect(0, 43647, '\P{Is_Block= Myanmar_Extended_a}', "");
Expect(1, 43647, '\P{^Is_Block= Myanmar_Extended_a}', "");
Expect(0, 43648, '\p{Is_Block= Myanmar_Extended_a}', "");
Expect(1, 43648, '\p{^Is_Block= Myanmar_Extended_a}', "");
Expect(1, 43648, '\P{Is_Block= Myanmar_Extended_a}', "");
Expect(0, 43648, '\P{^Is_Block= Myanmar_Extended_a}', "");
Error('\p{Is_Blk=/a/Myanmar_EXT_a}');
Error('\P{Is_Blk=/a/Myanmar_EXT_a}');
Expect(1, 43647, '\p{Is_Blk=myanmarexta}', "");
Expect(0, 43647, '\p{^Is_Blk=myanmarexta}', "");
Expect(0, 43647, '\P{Is_Blk=myanmarexta}', "");
Expect(1, 43647, '\P{^Is_Blk=myanmarexta}', "");
Expect(0, 43648, '\p{Is_Blk=myanmarexta}', "");
Expect(1, 43648, '\p{^Is_Blk=myanmarexta}', "");
Expect(1, 43648, '\P{Is_Blk=myanmarexta}', "");
Expect(0, 43648, '\P{^Is_Blk=myanmarexta}', "");
Expect(1, 43647, '\p{Is_Blk=	-Myanmar_Ext_A}', "");
Expect(0, 43647, '\p{^Is_Blk=	-Myanmar_Ext_A}', "");
Expect(0, 43647, '\P{Is_Blk=	-Myanmar_Ext_A}', "");
Expect(1, 43647, '\P{^Is_Blk=	-Myanmar_Ext_A}', "");
Expect(0, 43648, '\p{Is_Blk=	-Myanmar_Ext_A}', "");
Expect(1, 43648, '\p{^Is_Blk=	-Myanmar_Ext_A}', "");
Expect(1, 43648, '\P{Is_Blk=	-Myanmar_Ext_A}', "");
Expect(0, 43648, '\P{^Is_Blk=	-Myanmar_Ext_A}', "");
Error('\p{Block=-_No_Block:=}');
Error('\P{Block=-_No_Block:=}');
Expect(1, 918000, '\p{Block=noblock}', "");
Expect(0, 918000, '\p{^Block=noblock}', "");
Expect(0, 918000, '\P{Block=noblock}', "");
Expect(1, 918000, '\P{^Block=noblock}', "");
Expect(0, 983040, '\p{Block=noblock}', "");
Expect(1, 983040, '\p{^Block=noblock}', "");
Expect(1, 983040, '\P{Block=noblock}', "");
Expect(0, 983040, '\P{^Block=noblock}', "");
Expect(1, 918000, '\p{Block=	_no_Block}', "");
Expect(0, 918000, '\p{^Block=	_no_Block}', "");
Expect(0, 918000, '\P{Block=	_no_Block}', "");
Expect(1, 918000, '\P{^Block=	_no_Block}', "");
Expect(0, 983040, '\p{Block=	_no_Block}', "");
Expect(1, 983040, '\p{^Block=	_no_Block}', "");
Expect(1, 983040, '\P{Block=	_no_Block}', "");
Expect(0, 983040, '\P{^Block=	_no_Block}', "");
Error('\p{Blk=nb:=}');
Error('\P{Blk=nb:=}');
Expect(1, 918000, '\p{Blk=nb}', "");
Expect(0, 918000, '\p{^Blk=nb}', "");
Expect(0, 918000, '\P{Blk=nb}', "");
Expect(1, 918000, '\P{^Blk=nb}', "");
Expect(0, 983040, '\p{Blk=nb}', "");
Expect(1, 983040, '\p{^Blk=nb}', "");
Expect(1, 983040, '\P{Blk=nb}', "");
Expect(0, 983040, '\P{^Blk=nb}', "");
Expect(1, 918000, '\p{Blk=_NB}', "");
Expect(0, 918000, '\p{^Blk=_NB}', "");
Expect(0, 918000, '\P{Blk=_NB}', "");
Expect(1, 918000, '\P{^Blk=_NB}', "");
Expect(0, 983040, '\p{Blk=_NB}', "");
Expect(1, 983040, '\p{^Blk=_NB}', "");
Expect(1, 983040, '\P{Blk=_NB}', "");
Expect(0, 983040, '\P{^Blk=_NB}', "");
Error('\p{Is_Block= /a/No_block}');
Error('\P{Is_Block= /a/No_block}');
Expect(1, 918000, '\p{Is_Block=noblock}', "");
Expect(0, 918000, '\p{^Is_Block=noblock}', "");
Expect(0, 918000, '\P{Is_Block=noblock}', "");
Expect(1, 918000, '\P{^Is_Block=noblock}', "");
Expect(0, 983040, '\p{Is_Block=noblock}', "");
Expect(1, 983040, '\p{^Is_Block=noblock}', "");
Expect(1, 983040, '\P{Is_Block=noblock}', "");
Expect(0, 983040, '\P{^Is_Block=noblock}', "");
Expect(1, 918000, '\p{Is_Block=_NO_BLOCK}', "");
Expect(0, 918000, '\p{^Is_Block=_NO_BLOCK}', "");
Expect(0, 918000, '\P{Is_Block=_NO_BLOCK}', "");
Expect(1, 918000, '\P{^Is_Block=_NO_BLOCK}', "");
Expect(0, 983040, '\p{Is_Block=_NO_BLOCK}', "");
Expect(1, 983040, '\p{^Is_Block=_NO_BLOCK}', "");
Expect(1, 983040, '\P{Is_Block=_NO_BLOCK}', "");
Expect(0, 983040, '\P{^Is_Block=_NO_BLOCK}', "");
Error('\p{Is_Blk=/a/	 NB}');
Error('\P{Is_Blk=/a/	 NB}');
Expect(1, 918000, '\p{Is_Blk=nb}', "");
Expect(0, 918000, '\p{^Is_Blk=nb}', "");
Expect(0, 918000, '\P{Is_Blk=nb}', "");
Expect(1, 918000, '\P{^Is_Blk=nb}', "");
Expect(0, 983040, '\p{Is_Blk=nb}', "");
Expect(1, 983040, '\p{^Is_Blk=nb}', "");
Expect(1, 983040, '\P{Is_Blk=nb}', "");
Expect(0, 983040, '\P{^Is_Blk=nb}', "");
Expect(1, 918000, '\p{Is_Blk=	_NB}', "");
Expect(0, 918000, '\p{^Is_Blk=	_NB}', "");
Expect(0, 918000, '\P{Is_Blk=	_NB}', "");
Expect(1, 918000, '\P{^Is_Blk=	_NB}', "");
Expect(0, 983040, '\p{Is_Blk=	_NB}', "");
Expect(1, 983040, '\p{^Is_Blk=	_NB}', "");
Expect(1, 983040, '\P{Is_Blk=	_NB}', "");
Expect(0, 983040, '\P{^Is_Blk=	_NB}', "");
Error('\p{Block:_ nko:=}');
Error('\P{Block:_ nko:=}');
Expect(1, 2047, '\p{Block=nko}', "");
Expect(0, 2047, '\p{^Block=nko}', "");
Expect(0, 2047, '\P{Block=nko}', "");
Expect(1, 2047, '\P{^Block=nko}', "");
Expect(0, 2048, '\p{Block=nko}', "");
Expect(1, 2048, '\p{^Block=nko}', "");
Expect(1, 2048, '\P{Block=nko}', "");
Expect(0, 2048, '\P{^Block=nko}', "");
Expect(1, 2047, '\p{Block= nko}', "");
Expect(0, 2047, '\p{^Block= nko}', "");
Expect(0, 2047, '\P{Block= nko}', "");
Expect(1, 2047, '\P{^Block= nko}', "");
Expect(0, 2048, '\p{Block= nko}', "");
Expect(1, 2048, '\p{^Block= nko}', "");
Expect(1, 2048, '\P{Block= nko}', "");
Expect(0, 2048, '\P{^Block= nko}', "");
Error('\p{Blk:   -_nko:=}');
Error('\P{Blk:   -_nko:=}');
Expect(1, 2047, '\p{Blk=nko}', "");
Expect(0, 2047, '\p{^Blk=nko}', "");
Expect(0, 2047, '\P{Blk=nko}', "");
Expect(1, 2047, '\P{^Blk=nko}', "");
Expect(0, 2048, '\p{Blk=nko}', "");
Expect(1, 2048, '\p{^Blk=nko}', "");
Expect(1, 2048, '\P{Blk=nko}', "");
Expect(0, 2048, '\P{^Blk=nko}', "");
Expect(1, 2047, '\p{Blk= NKO}', "");
Expect(0, 2047, '\p{^Blk= NKO}', "");
Expect(0, 2047, '\P{Blk= NKO}', "");
Expect(1, 2047, '\P{^Blk= NKO}', "");
Expect(0, 2048, '\p{Blk= NKO}', "");
Expect(1, 2048, '\p{^Blk= NKO}', "");
Expect(1, 2048, '\P{Blk= NKO}', "");
Expect(0, 2048, '\P{^Blk= NKO}', "");
Error('\p{Is_Block=__NKO:=}');
Error('\P{Is_Block=__NKO:=}');
Expect(1, 2047, '\p{Is_Block=nko}', "");
Expect(0, 2047, '\p{^Is_Block=nko}', "");
Expect(0, 2047, '\P{Is_Block=nko}', "");
Expect(1, 2047, '\P{^Is_Block=nko}', "");
Expect(0, 2048, '\p{Is_Block=nko}', "");
Expect(1, 2048, '\p{^Is_Block=nko}', "");
Expect(1, 2048, '\P{Is_Block=nko}', "");
Expect(0, 2048, '\P{^Is_Block=nko}', "");
Expect(1, 2047, '\p{Is_Block=_-NKo}', "");
Expect(0, 2047, '\p{^Is_Block=_-NKo}', "");
Expect(0, 2047, '\P{Is_Block=_-NKo}', "");
Expect(1, 2047, '\P{^Is_Block=_-NKo}', "");
Expect(0, 2048, '\p{Is_Block=_-NKo}', "");
Expect(1, 2048, '\p{^Is_Block=_-NKo}', "");
Expect(1, 2048, '\P{Is_Block=_-NKo}', "");
Expect(0, 2048, '\P{^Is_Block=_-NKo}', "");
Error('\p{Is_Blk=_/a/NKo}');
Error('\P{Is_Blk=_/a/NKo}');
Expect(1, 2047, '\p{Is_Blk=nko}', "");
Expect(0, 2047, '\p{^Is_Blk=nko}', "");
Expect(0, 2047, '\P{Is_Blk=nko}', "");
Expect(1, 2047, '\P{^Is_Blk=nko}', "");
Expect(0, 2048, '\p{Is_Blk=nko}', "");
Expect(1, 2048, '\p{^Is_Blk=nko}', "");
Expect(1, 2048, '\P{Is_Blk=nko}', "");
Expect(0, 2048, '\P{^Is_Blk=nko}', "");
Expect(1, 2047, '\p{Is_Blk=NKo}', "");
Expect(0, 2047, '\p{^Is_Blk=NKo}', "");
Expect(0, 2047, '\P{Is_Blk=NKo}', "");
Expect(1, 2047, '\P{^Is_Blk=NKo}', "");
Expect(0, 2048, '\p{Is_Blk=NKo}', "");
Expect(1, 2048, '\p{^Is_Blk=NKo}', "");
Expect(1, 2048, '\P{Is_Blk=NKo}', "");
Expect(0, 2048, '\P{^Is_Blk=NKo}', "");
Error('\p{Block= New_tai_Lue:=}');
Error('\P{Block= New_tai_Lue:=}');
Expect(1, 6623, '\p{Block: newtailue}', "");
Expect(0, 6623, '\p{^Block: newtailue}', "");
Expect(0, 6623, '\P{Block: newtailue}', "");
Expect(1, 6623, '\P{^Block: newtailue}', "");
Expect(0, 6624, '\p{Block: newtailue}', "");
Expect(1, 6624, '\p{^Block: newtailue}', "");
Expect(1, 6624, '\P{Block: newtailue}', "");
Expect(0, 6624, '\P{^Block: newtailue}', "");
Expect(1, 6623, '\p{Block= -NEW_Tai_Lue}', "");
Expect(0, 6623, '\p{^Block= -NEW_Tai_Lue}', "");
Expect(0, 6623, '\P{Block= -NEW_Tai_Lue}', "");
Expect(1, 6623, '\P{^Block= -NEW_Tai_Lue}', "");
Expect(0, 6624, '\p{Block= -NEW_Tai_Lue}', "");
Expect(1, 6624, '\p{^Block= -NEW_Tai_Lue}', "");
Expect(1, 6624, '\P{Block= -NEW_Tai_Lue}', "");
Expect(0, 6624, '\P{^Block= -NEW_Tai_Lue}', "");
Error('\p{Blk:   /a/_-NEW_Tai_LUE}');
Error('\P{Blk:   /a/_-NEW_Tai_LUE}');
Expect(1, 6623, '\p{Blk=newtailue}', "");
Expect(0, 6623, '\p{^Blk=newtailue}', "");
Expect(0, 6623, '\P{Blk=newtailue}', "");
Expect(1, 6623, '\P{^Blk=newtailue}', "");
Expect(0, 6624, '\p{Blk=newtailue}', "");
Expect(1, 6624, '\p{^Blk=newtailue}', "");
Expect(1, 6624, '\P{Blk=newtailue}', "");
Expect(0, 6624, '\P{^Blk=newtailue}', "");
Expect(1, 6623, '\p{Blk= 	new_tai_lue}', "");
Expect(0, 6623, '\p{^Blk= 	new_tai_lue}', "");
Expect(0, 6623, '\P{Blk= 	new_tai_lue}', "");
Expect(1, 6623, '\P{^Blk= 	new_tai_lue}', "");
Expect(0, 6624, '\p{Blk= 	new_tai_lue}', "");
Expect(1, 6624, '\p{^Blk= 	new_tai_lue}', "");
Expect(1, 6624, '\P{Blk= 	new_tai_lue}', "");
Expect(0, 6624, '\P{^Blk= 	new_tai_lue}', "");
Error('\p{Is_Block=/a/ new_Tai_lue}');
Error('\P{Is_Block=/a/ new_Tai_lue}');
Expect(1, 6623, '\p{Is_Block=newtailue}', "");
Expect(0, 6623, '\p{^Is_Block=newtailue}', "");
Expect(0, 6623, '\P{Is_Block=newtailue}', "");
Expect(1, 6623, '\P{^Is_Block=newtailue}', "");
Expect(0, 6624, '\p{Is_Block=newtailue}', "");
Expect(1, 6624, '\p{^Is_Block=newtailue}', "");
Expect(1, 6624, '\P{Is_Block=newtailue}', "");
Expect(0, 6624, '\P{^Is_Block=newtailue}', "");
Expect(1, 6623, '\p{Is_Block=_NEW_Tai_lue}', "");
Expect(0, 6623, '\p{^Is_Block=_NEW_Tai_lue}', "");
Expect(0, 6623, '\P{Is_Block=_NEW_Tai_lue}', "");
Expect(1, 6623, '\P{^Is_Block=_NEW_Tai_lue}', "");
Expect(0, 6624, '\p{Is_Block=_NEW_Tai_lue}', "");
Expect(1, 6624, '\p{^Is_Block=_NEW_Tai_lue}', "");
Expect(1, 6624, '\P{Is_Block=_NEW_Tai_lue}', "");
Expect(0, 6624, '\P{^Is_Block=_NEW_Tai_lue}', "");
Error('\p{Is_Blk=	/a/NEW_Tai_Lue}');
Error('\P{Is_Blk=	/a/NEW_Tai_Lue}');
Expect(1, 6623, '\p{Is_Blk=newtailue}', "");
Expect(0, 6623, '\p{^Is_Blk=newtailue}', "");
Expect(0, 6623, '\P{Is_Blk=newtailue}', "");
Expect(1, 6623, '\P{^Is_Blk=newtailue}', "");
Expect(0, 6624, '\p{Is_Blk=newtailue}', "");
Expect(1, 6624, '\p{^Is_Blk=newtailue}', "");
Expect(1, 6624, '\P{Is_Blk=newtailue}', "");
Expect(0, 6624, '\P{^Is_Blk=newtailue}', "");
Expect(1, 6623, '\p{Is_Blk= -new_TAI_Lue}', "");
Expect(0, 6623, '\p{^Is_Blk= -new_TAI_Lue}', "");
Expect(0, 6623, '\P{Is_Blk= -new_TAI_Lue}', "");
Expect(1, 6623, '\P{^Is_Blk= -new_TAI_Lue}', "");
Expect(0, 6624, '\p{Is_Blk= -new_TAI_Lue}', "");
Expect(1, 6624, '\p{^Is_Blk= -new_TAI_Lue}', "");
Expect(1, 6624, '\P{Is_Blk= -new_TAI_Lue}', "");
Expect(0, 6624, '\P{^Is_Blk= -new_TAI_Lue}', "");
Error('\p{Block=-/a/Number_forms}');
Error('\P{Block=-/a/Number_forms}');
Expect(1, 8591, '\p{Block=numberforms}', "");
Expect(0, 8591, '\p{^Block=numberforms}', "");
Expect(0, 8591, '\P{Block=numberforms}', "");
Expect(1, 8591, '\P{^Block=numberforms}', "");
Expect(0, 8592, '\p{Block=numberforms}', "");
Expect(1, 8592, '\p{^Block=numberforms}', "");
Expect(1, 8592, '\P{Block=numberforms}', "");
Expect(0, 8592, '\P{^Block=numberforms}', "");
Expect(1, 8591, '\p{Block:	_NUMBER_FORMS}', "");
Expect(0, 8591, '\p{^Block:	_NUMBER_FORMS}', "");
Expect(0, 8591, '\P{Block:	_NUMBER_FORMS}', "");
Expect(1, 8591, '\P{^Block:	_NUMBER_FORMS}', "");
Expect(0, 8592, '\p{Block:	_NUMBER_FORMS}', "");
Expect(1, 8592, '\p{^Block:	_NUMBER_FORMS}', "");
Expect(1, 8592, '\P{Block:	_NUMBER_FORMS}', "");
Expect(0, 8592, '\P{^Block:	_NUMBER_FORMS}', "");
Error('\p{Blk=/a/	number_Forms}');
Error('\P{Blk=/a/	number_Forms}');
Expect(1, 8591, '\p{Blk=numberforms}', "");
Expect(0, 8591, '\p{^Blk=numberforms}', "");
Expect(0, 8591, '\P{Blk=numberforms}', "");
Expect(1, 8591, '\P{^Blk=numberforms}', "");
Expect(0, 8592, '\p{Blk=numberforms}', "");
Expect(1, 8592, '\p{^Blk=numberforms}', "");
Expect(1, 8592, '\P{Blk=numberforms}', "");
Expect(0, 8592, '\P{^Blk=numberforms}', "");
Expect(1, 8591, '\p{Blk=	Number_Forms}', "");
Expect(0, 8591, '\p{^Blk=	Number_Forms}', "");
Expect(0, 8591, '\P{Blk=	Number_Forms}', "");
Expect(1, 8591, '\P{^Blk=	Number_Forms}', "");
Expect(0, 8592, '\p{Blk=	Number_Forms}', "");
Expect(1, 8592, '\p{^Blk=	Number_Forms}', "");
Expect(1, 8592, '\P{Blk=	Number_Forms}', "");
Expect(0, 8592, '\P{^Blk=	Number_Forms}', "");
Error('\p{Is_Block=-number_Forms:=}');
Error('\P{Is_Block=-number_Forms:=}');
Expect(1, 8591, '\p{Is_Block=numberforms}', "");
Expect(0, 8591, '\p{^Is_Block=numberforms}', "");
Expect(0, 8591, '\P{Is_Block=numberforms}', "");
Expect(1, 8591, '\P{^Is_Block=numberforms}', "");
Expect(0, 8592, '\p{Is_Block=numberforms}', "");
Expect(1, 8592, '\p{^Is_Block=numberforms}', "");
Expect(1, 8592, '\P{Is_Block=numberforms}', "");
Expect(0, 8592, '\P{^Is_Block=numberforms}', "");
Expect(1, 8591, '\p{Is_Block=-	Number_FORMS}', "");
Expect(0, 8591, '\p{^Is_Block=-	Number_FORMS}', "");
Expect(0, 8591, '\P{Is_Block=-	Number_FORMS}', "");
Expect(1, 8591, '\P{^Is_Block=-	Number_FORMS}', "");
Expect(0, 8592, '\p{Is_Block=-	Number_FORMS}', "");
Expect(1, 8592, '\p{^Is_Block=-	Number_FORMS}', "");
Expect(1, 8592, '\P{Is_Block=-	Number_FORMS}', "");
Expect(0, 8592, '\P{^Is_Block=-	Number_FORMS}', "");
Error('\p{Is_Blk:	/a/- Number_Forms}');
Error('\P{Is_Blk:	/a/- Number_Forms}');
Expect(1, 8591, '\p{Is_Blk=numberforms}', "");
Expect(0, 8591, '\p{^Is_Blk=numberforms}', "");
Expect(0, 8591, '\P{Is_Blk=numberforms}', "");
Expect(1, 8591, '\P{^Is_Blk=numberforms}', "");
Expect(0, 8592, '\p{Is_Blk=numberforms}', "");
Expect(1, 8592, '\p{^Is_Blk=numberforms}', "");
Expect(1, 8592, '\P{Is_Blk=numberforms}', "");
Expect(0, 8592, '\P{^Is_Blk=numberforms}', "");
Expect(1, 8591, '\p{Is_Blk= 	NUMBER_FORMS}', "");
Expect(0, 8591, '\p{^Is_Blk= 	NUMBER_FORMS}', "");
Expect(0, 8591, '\P{Is_Blk= 	NUMBER_FORMS}', "");
Expect(1, 8591, '\P{^Is_Blk= 	NUMBER_FORMS}', "");
Expect(0, 8592, '\p{Is_Blk= 	NUMBER_FORMS}', "");
Expect(1, 8592, '\p{^Is_Blk= 	NUMBER_FORMS}', "");
Expect(1, 8592, '\P{Is_Blk= 	NUMBER_FORMS}', "");
Expect(0, 8592, '\P{^Is_Blk= 	NUMBER_FORMS}', "");
Error('\p{Block= /a/Optical_CHARACTER_Recognition}');
Error('\P{Block= /a/Optical_CHARACTER_Recognition}');
Expect(1, 9311, '\p{Block=opticalcharacterrecognition}', "");
Expect(0, 9311, '\p{^Block=opticalcharacterrecognition}', "");
Expect(0, 9311, '\P{Block=opticalcharacterrecognition}', "");
Expect(1, 9311, '\P{^Block=opticalcharacterrecognition}', "");
Expect(0, 9312, '\p{Block=opticalcharacterrecognition}', "");
Expect(1, 9312, '\p{^Block=opticalcharacterrecognition}', "");
Expect(1, 9312, '\P{Block=opticalcharacterrecognition}', "");
Expect(0, 9312, '\P{^Block=opticalcharacterrecognition}', "");
Expect(1, 9311, '\p{Block= _optical_Character_Recognition}', "");
Expect(0, 9311, '\p{^Block= _optical_Character_Recognition}', "");
Expect(0, 9311, '\P{Block= _optical_Character_Recognition}', "");
Expect(1, 9311, '\P{^Block= _optical_Character_Recognition}', "");
Expect(0, 9312, '\p{Block= _optical_Character_Recognition}', "");
Expect(1, 9312, '\p{^Block= _optical_Character_Recognition}', "");
Expect(1, 9312, '\P{Block= _optical_Character_Recognition}', "");
Expect(0, 9312, '\P{^Block= _optical_Character_Recognition}', "");
Error('\p{Blk=	ocr:=}');
Error('\P{Blk=	ocr:=}');
Expect(1, 9311, '\p{Blk:   ocr}', "");
Expect(0, 9311, '\p{^Blk:   ocr}', "");
Expect(0, 9311, '\P{Blk:   ocr}', "");
Expect(1, 9311, '\P{^Blk:   ocr}', "");
Expect(0, 9312, '\p{Blk:   ocr}', "");
Expect(1, 9312, '\p{^Blk:   ocr}', "");
Expect(1, 9312, '\P{Blk:   ocr}', "");
Expect(0, 9312, '\P{^Blk:   ocr}', "");
Expect(1, 9311, '\p{Blk=	-OCR}', "");
Expect(0, 9311, '\p{^Blk=	-OCR}', "");
Expect(0, 9311, '\P{Blk=	-OCR}', "");
Expect(1, 9311, '\P{^Blk=	-OCR}', "");
Expect(0, 9312, '\p{Blk=	-OCR}', "");
Expect(1, 9312, '\p{^Blk=	-OCR}', "");
Expect(1, 9312, '\P{Blk=	-OCR}', "");
Expect(0, 9312, '\P{^Blk=	-OCR}', "");
Error('\p{Is_Block=/a/optical_CHARACTER_Recognition}');
Error('\P{Is_Block=/a/optical_CHARACTER_Recognition}');
Expect(1, 9311, '\p{Is_Block:	opticalcharacterrecognition}', "");
Expect(0, 9311, '\p{^Is_Block:	opticalcharacterrecognition}', "");
Expect(0, 9311, '\P{Is_Block:	opticalcharacterrecognition}', "");
Expect(1, 9311, '\P{^Is_Block:	opticalcharacterrecognition}', "");
Expect(0, 9312, '\p{Is_Block:	opticalcharacterrecognition}', "");
Expect(1, 9312, '\p{^Is_Block:	opticalcharacterrecognition}', "");
Expect(1, 9312, '\P{Is_Block:	opticalcharacterrecognition}', "");
Expect(0, 9312, '\P{^Is_Block:	opticalcharacterrecognition}', "");
Expect(1, 9311, '\p{Is_Block=--OPTICAL_Character_RECOGNITION}', "");
Expect(0, 9311, '\p{^Is_Block=--OPTICAL_Character_RECOGNITION}', "");
Expect(0, 9311, '\P{Is_Block=--OPTICAL_Character_RECOGNITION}', "");
Expect(1, 9311, '\P{^Is_Block=--OPTICAL_Character_RECOGNITION}', "");
Expect(0, 9312, '\p{Is_Block=--OPTICAL_Character_RECOGNITION}', "");
Expect(1, 9312, '\p{^Is_Block=--OPTICAL_Character_RECOGNITION}', "");
Expect(1, 9312, '\P{Is_Block=--OPTICAL_Character_RECOGNITION}', "");
Expect(0, 9312, '\P{^Is_Block=--OPTICAL_Character_RECOGNITION}', "");
Error('\p{Is_Blk=OCR/a/}');
Error('\P{Is_Blk=OCR/a/}');
Expect(1, 9311, '\p{Is_Blk=ocr}', "");
Expect(0, 9311, '\p{^Is_Blk=ocr}', "");
Expect(0, 9311, '\P{Is_Blk=ocr}', "");
Expect(1, 9311, '\P{^Is_Blk=ocr}', "");
Expect(0, 9312, '\p{Is_Blk=ocr}', "");
Expect(1, 9312, '\p{^Is_Blk=ocr}', "");
Expect(1, 9312, '\P{Is_Blk=ocr}', "");
Expect(0, 9312, '\P{^Is_Blk=ocr}', "");
Expect(1, 9311, '\p{Is_Blk= -OCR}', "");
Expect(0, 9311, '\p{^Is_Blk= -OCR}', "");
Expect(0, 9311, '\P{Is_Blk= -OCR}', "");
Expect(1, 9311, '\P{^Is_Blk= -OCR}', "");
Expect(0, 9312, '\p{Is_Blk= -OCR}', "");
Expect(1, 9312, '\p{^Is_Blk= -OCR}', "");
Expect(1, 9312, '\P{Is_Blk= -OCR}', "");
Expect(0, 9312, '\P{^Is_Blk= -OCR}', "");
Error('\p{Block=:=	-OGHAM}');
Error('\P{Block=:=	-OGHAM}');
Expect(1, 5791, '\p{Block=ogham}', "");
Expect(0, 5791, '\p{^Block=ogham}', "");
Expect(0, 5791, '\P{Block=ogham}', "");
Expect(1, 5791, '\P{^Block=ogham}', "");
Expect(0, 5792, '\p{Block=ogham}', "");
Expect(1, 5792, '\p{^Block=ogham}', "");
Expect(1, 5792, '\P{Block=ogham}', "");
Expect(0, 5792, '\P{^Block=ogham}', "");
Expect(1, 5791, '\p{Block= 	ogham}', "");
Expect(0, 5791, '\p{^Block= 	ogham}', "");
Expect(0, 5791, '\P{Block= 	ogham}', "");
Expect(1, 5791, '\P{^Block= 	ogham}', "");
Expect(0, 5792, '\p{Block= 	ogham}', "");
Expect(1, 5792, '\p{^Block= 	ogham}', "");
Expect(1, 5792, '\P{Block= 	ogham}', "");
Expect(0, 5792, '\P{^Block= 	ogham}', "");
Error('\p{Blk=-:=ogham}');
Error('\P{Blk=-:=ogham}');
Expect(1, 5791, '\p{Blk=ogham}', "");
Expect(0, 5791, '\p{^Blk=ogham}', "");
Expect(0, 5791, '\P{Blk=ogham}', "");
Expect(1, 5791, '\P{^Blk=ogham}', "");
Expect(0, 5792, '\p{Blk=ogham}', "");
Expect(1, 5792, '\p{^Blk=ogham}', "");
Expect(1, 5792, '\P{Blk=ogham}', "");
Expect(0, 5792, '\P{^Blk=ogham}', "");
Expect(1, 5791, '\p{Blk=_OGHAM}', "");
Expect(0, 5791, '\p{^Blk=_OGHAM}', "");
Expect(0, 5791, '\P{Blk=_OGHAM}', "");
Expect(1, 5791, '\P{^Blk=_OGHAM}', "");
Expect(0, 5792, '\p{Blk=_OGHAM}', "");
Expect(1, 5792, '\p{^Blk=_OGHAM}', "");
Expect(1, 5792, '\P{Blk=_OGHAM}', "");
Expect(0, 5792, '\P{^Blk=_OGHAM}', "");
Error('\p{Is_Block= -ogham/a/}');
Error('\P{Is_Block= -ogham/a/}');
Expect(1, 5791, '\p{Is_Block:	ogham}', "");
Expect(0, 5791, '\p{^Is_Block:	ogham}', "");
Expect(0, 5791, '\P{Is_Block:	ogham}', "");
Expect(1, 5791, '\P{^Is_Block:	ogham}', "");
Expect(0, 5792, '\p{Is_Block:	ogham}', "");
Expect(1, 5792, '\p{^Is_Block:	ogham}', "");
Expect(1, 5792, '\P{Is_Block:	ogham}', "");
Expect(0, 5792, '\P{^Is_Block:	ogham}', "");
Expect(1, 5791, '\p{Is_Block=_ OGHAM}', "");
Expect(0, 5791, '\p{^Is_Block=_ OGHAM}', "");
Expect(0, 5791, '\P{Is_Block=_ OGHAM}', "");
Expect(1, 5791, '\P{^Is_Block=_ OGHAM}', "");
Expect(0, 5792, '\p{Is_Block=_ OGHAM}', "");
Expect(1, 5792, '\p{^Is_Block=_ OGHAM}', "");
Expect(1, 5792, '\P{Is_Block=_ OGHAM}', "");
Expect(0, 5792, '\P{^Is_Block=_ OGHAM}', "");
Error('\p{Is_Blk=/a/_OGHAM}');
Error('\P{Is_Blk=/a/_OGHAM}');
Expect(1, 5791, '\p{Is_Blk=ogham}', "");
Expect(0, 5791, '\p{^Is_Blk=ogham}', "");
Expect(0, 5791, '\P{Is_Blk=ogham}', "");
Expect(1, 5791, '\P{^Is_Blk=ogham}', "");
Expect(0, 5792, '\p{Is_Blk=ogham}', "");
Expect(1, 5792, '\p{^Is_Blk=ogham}', "");
Expect(1, 5792, '\P{Is_Blk=ogham}', "");
Expect(0, 5792, '\P{^Is_Blk=ogham}', "");
Expect(1, 5791, '\p{Is_Blk=__OGHAM}', "");
Expect(0, 5791, '\p{^Is_Blk=__OGHAM}', "");
Expect(0, 5791, '\P{Is_Blk=__OGHAM}', "");
Expect(1, 5791, '\P{^Is_Blk=__OGHAM}', "");
Expect(0, 5792, '\p{Is_Blk=__OGHAM}', "");
Expect(1, 5792, '\p{^Is_Blk=__OGHAM}', "");
Expect(1, 5792, '\P{Is_Blk=__OGHAM}', "");
Expect(0, 5792, '\P{^Is_Blk=__OGHAM}', "");
Error('\p{Block=:=	Ol_chiki}');
Error('\P{Block=:=	Ol_chiki}');
Expect(1, 7295, '\p{Block=olchiki}', "");
Expect(0, 7295, '\p{^Block=olchiki}', "");
Expect(0, 7295, '\P{Block=olchiki}', "");
Expect(1, 7295, '\P{^Block=olchiki}', "");
Expect(0, 7296, '\p{Block=olchiki}', "");
Expect(1, 7296, '\p{^Block=olchiki}', "");
Expect(1, 7296, '\P{Block=olchiki}', "");
Expect(0, 7296, '\P{^Block=olchiki}', "");
Expect(1, 7295, '\p{Block:  OL_chiki}', "");
Expect(0, 7295, '\p{^Block:  OL_chiki}', "");
Expect(0, 7295, '\P{Block:  OL_chiki}', "");
Expect(1, 7295, '\P{^Block:  OL_chiki}', "");
Expect(0, 7296, '\p{Block:  OL_chiki}', "");
Expect(1, 7296, '\p{^Block:  OL_chiki}', "");
Expect(1, 7296, '\P{Block:  OL_chiki}', "");
Expect(0, 7296, '\P{^Block:  OL_chiki}', "");
Error('\p{Blk=:=OL_Chiki}');
Error('\P{Blk=:=OL_Chiki}');
Expect(1, 7295, '\p{Blk=olchiki}', "");
Expect(0, 7295, '\p{^Blk=olchiki}', "");
Expect(0, 7295, '\P{Blk=olchiki}', "");
Expect(1, 7295, '\P{^Blk=olchiki}', "");
Expect(0, 7296, '\p{Blk=olchiki}', "");
Expect(1, 7296, '\p{^Blk=olchiki}', "");
Expect(1, 7296, '\P{Blk=olchiki}', "");
Expect(0, 7296, '\P{^Blk=olchiki}', "");
Expect(1, 7295, '\p{Blk= Ol_Chiki}', "");
Expect(0, 7295, '\p{^Blk= Ol_Chiki}', "");
Expect(0, 7295, '\P{Blk= Ol_Chiki}', "");
Expect(1, 7295, '\P{^Blk= Ol_Chiki}', "");
Expect(0, 7296, '\p{Blk= Ol_Chiki}', "");
Expect(1, 7296, '\p{^Blk= Ol_Chiki}', "");
Expect(1, 7296, '\P{Blk= Ol_Chiki}', "");
Expect(0, 7296, '\P{^Blk= Ol_Chiki}', "");
Error('\p{Is_Block=	 ol_Chiki:=}');
Error('\P{Is_Block=	 ol_Chiki:=}');
Expect(1, 7295, '\p{Is_Block=olchiki}', "");
Expect(0, 7295, '\p{^Is_Block=olchiki}', "");
Expect(0, 7295, '\P{Is_Block=olchiki}', "");
Expect(1, 7295, '\P{^Is_Block=olchiki}', "");
Expect(0, 7296, '\p{Is_Block=olchiki}', "");
Expect(1, 7296, '\p{^Is_Block=olchiki}', "");
Expect(1, 7296, '\P{Is_Block=olchiki}', "");
Expect(0, 7296, '\P{^Is_Block=olchiki}', "");
Expect(1, 7295, '\p{Is_Block= ol_Chiki}', "");
Expect(0, 7295, '\p{^Is_Block= ol_Chiki}', "");
Expect(0, 7295, '\P{Is_Block= ol_Chiki}', "");
Expect(1, 7295, '\P{^Is_Block= ol_Chiki}', "");
Expect(0, 7296, '\p{Is_Block= ol_Chiki}', "");
Expect(1, 7296, '\p{^Is_Block= ol_Chiki}', "");
Expect(1, 7296, '\P{Is_Block= ol_Chiki}', "");
Expect(0, 7296, '\P{^Is_Block= ol_Chiki}', "");
Error('\p{Is_Blk=/a/	 OL_Chiki}');
Error('\P{Is_Blk=/a/	 OL_Chiki}');
Expect(1, 7295, '\p{Is_Blk=olchiki}', "");
Expect(0, 7295, '\p{^Is_Blk=olchiki}', "");
Expect(0, 7295, '\P{Is_Blk=olchiki}', "");
Expect(1, 7295, '\P{^Is_Blk=olchiki}', "");
Expect(0, 7296, '\p{Is_Blk=olchiki}', "");
Expect(1, 7296, '\p{^Is_Blk=olchiki}', "");
Expect(1, 7296, '\P{Is_Blk=olchiki}', "");
Expect(0, 7296, '\P{^Is_Blk=olchiki}', "");
Expect(1, 7295, '\p{Is_Blk=-Ol_Chiki}', "");
Expect(0, 7295, '\p{^Is_Blk=-Ol_Chiki}', "");
Expect(0, 7295, '\P{Is_Blk=-Ol_Chiki}', "");
Expect(1, 7295, '\P{^Is_Blk=-Ol_Chiki}', "");
Expect(0, 7296, '\p{Is_Blk=-Ol_Chiki}', "");
Expect(1, 7296, '\p{^Is_Blk=-Ol_Chiki}', "");
Expect(1, 7296, '\P{Is_Blk=-Ol_Chiki}', "");
Expect(0, 7296, '\P{^Is_Blk=-Ol_Chiki}', "");
Error('\p{Block=_old_ITALIC/a/}');
Error('\P{Block=_old_ITALIC/a/}');
Expect(1, 66351, '\p{Block=olditalic}', "");
Expect(0, 66351, '\p{^Block=olditalic}', "");
Expect(0, 66351, '\P{Block=olditalic}', "");
Expect(1, 66351, '\P{^Block=olditalic}', "");
Expect(0, 66352, '\p{Block=olditalic}', "");
Expect(1, 66352, '\p{^Block=olditalic}', "");
Expect(1, 66352, '\P{Block=olditalic}', "");
Expect(0, 66352, '\P{^Block=olditalic}', "");
Expect(1, 66351, '\p{Block=_	Old_Italic}', "");
Expect(0, 66351, '\p{^Block=_	Old_Italic}', "");
Expect(0, 66351, '\P{Block=_	Old_Italic}', "");
Expect(1, 66351, '\P{^Block=_	Old_Italic}', "");
Expect(0, 66352, '\p{Block=_	Old_Italic}', "");
Expect(1, 66352, '\p{^Block=_	Old_Italic}', "");
Expect(1, 66352, '\P{Block=_	Old_Italic}', "");
Expect(0, 66352, '\P{^Block=_	Old_Italic}', "");
Error('\p{Blk=_ old_ITALIC/a/}');
Error('\P{Blk=_ old_ITALIC/a/}');
Expect(1, 66351, '\p{Blk=olditalic}', "");
Expect(0, 66351, '\p{^Blk=olditalic}', "");
Expect(0, 66351, '\P{Blk=olditalic}', "");
Expect(1, 66351, '\P{^Blk=olditalic}', "");
Expect(0, 66352, '\p{Blk=olditalic}', "");
Expect(1, 66352, '\p{^Blk=olditalic}', "");
Expect(1, 66352, '\P{Blk=olditalic}', "");
Expect(0, 66352, '\P{^Blk=olditalic}', "");
Expect(1, 66351, '\p{Blk= 	Old_italic}', "");
Expect(0, 66351, '\p{^Blk= 	Old_italic}', "");
Expect(0, 66351, '\P{Blk= 	Old_italic}', "");
Expect(1, 66351, '\P{^Blk= 	Old_italic}', "");
Expect(0, 66352, '\p{Blk= 	Old_italic}', "");
Expect(1, 66352, '\p{^Blk= 	Old_italic}', "");
Expect(1, 66352, '\P{Blk= 	Old_italic}', "");
Expect(0, 66352, '\P{^Blk= 	Old_italic}', "");
Error('\p{Is_Block= _Old_ITALIC/a/}');
Error('\P{Is_Block= _Old_ITALIC/a/}');
Expect(1, 66351, '\p{Is_Block=olditalic}', "");
Expect(0, 66351, '\p{^Is_Block=olditalic}', "");
Expect(0, 66351, '\P{Is_Block=olditalic}', "");
Expect(1, 66351, '\P{^Is_Block=olditalic}', "");
Expect(0, 66352, '\p{Is_Block=olditalic}', "");
Expect(1, 66352, '\p{^Is_Block=olditalic}', "");
Expect(1, 66352, '\P{Is_Block=olditalic}', "");
Expect(0, 66352, '\P{^Is_Block=olditalic}', "");
Expect(1, 66351, '\p{Is_Block=_ Old_ITALIC}', "");
Expect(0, 66351, '\p{^Is_Block=_ Old_ITALIC}', "");
Expect(0, 66351, '\P{Is_Block=_ Old_ITALIC}', "");
Expect(1, 66351, '\P{^Is_Block=_ Old_ITALIC}', "");
Expect(0, 66352, '\p{Is_Block=_ Old_ITALIC}', "");
Expect(1, 66352, '\p{^Is_Block=_ Old_ITALIC}', "");
Expect(1, 66352, '\P{Is_Block=_ Old_ITALIC}', "");
Expect(0, 66352, '\P{^Is_Block=_ Old_ITALIC}', "");
Error('\p{Is_Blk:	:=Old_ITALIC}');
Error('\P{Is_Blk:	:=Old_ITALIC}');
Expect(1, 66351, '\p{Is_Blk:olditalic}', "");
Expect(0, 66351, '\p{^Is_Blk:olditalic}', "");
Expect(0, 66351, '\P{Is_Blk:olditalic}', "");
Expect(1, 66351, '\P{^Is_Blk:olditalic}', "");
Expect(0, 66352, '\p{Is_Blk:olditalic}', "");
Expect(1, 66352, '\p{^Is_Blk:olditalic}', "");
Expect(1, 66352, '\P{Is_Blk:olditalic}', "");
Expect(0, 66352, '\P{^Is_Blk:olditalic}', "");
Expect(1, 66351, '\p{Is_Blk=-	old_Italic}', "");
Expect(0, 66351, '\p{^Is_Blk=-	old_Italic}', "");
Expect(0, 66351, '\P{Is_Blk=-	old_Italic}', "");
Expect(1, 66351, '\P{^Is_Blk=-	old_Italic}', "");
Expect(0, 66352, '\p{Is_Blk=-	old_Italic}', "");
Expect(1, 66352, '\p{^Is_Blk=-	old_Italic}', "");
Expect(1, 66352, '\P{Is_Blk=-	old_Italic}', "");
Expect(0, 66352, '\P{^Is_Blk=-	old_Italic}', "");
Error('\p{Block=__OLD_Persian:=}');
Error('\P{Block=__OLD_Persian:=}');
Expect(1, 66527, '\p{Block=oldpersian}', "");
Expect(0, 66527, '\p{^Block=oldpersian}', "");
Expect(0, 66527, '\P{Block=oldpersian}', "");
Expect(1, 66527, '\P{^Block=oldpersian}', "");
Expect(0, 66528, '\p{Block=oldpersian}', "");
Expect(1, 66528, '\p{^Block=oldpersian}', "");
Expect(1, 66528, '\P{Block=oldpersian}', "");
Expect(0, 66528, '\P{^Block=oldpersian}', "");
Expect(1, 66527, '\p{Block=	 old_Persian}', "");
Expect(0, 66527, '\p{^Block=	 old_Persian}', "");
Expect(0, 66527, '\P{Block=	 old_Persian}', "");
Expect(1, 66527, '\P{^Block=	 old_Persian}', "");
Expect(0, 66528, '\p{Block=	 old_Persian}', "");
Expect(1, 66528, '\p{^Block=	 old_Persian}', "");
Expect(1, 66528, '\P{Block=	 old_Persian}', "");
Expect(0, 66528, '\P{^Block=	 old_Persian}', "");
Error('\p{Blk=-/a/old_PERSIAN}');
Error('\P{Blk=-/a/old_PERSIAN}');
Expect(1, 66527, '\p{Blk=oldpersian}', "");
Expect(0, 66527, '\p{^Blk=oldpersian}', "");
Expect(0, 66527, '\P{Blk=oldpersian}', "");
Expect(1, 66527, '\P{^Blk=oldpersian}', "");
Expect(0, 66528, '\p{Blk=oldpersian}', "");
Expect(1, 66528, '\p{^Blk=oldpersian}', "");
Expect(1, 66528, '\P{Blk=oldpersian}', "");
Expect(0, 66528, '\P{^Blk=oldpersian}', "");
Expect(1, 66527, '\p{Blk=__old_persian}', "");
Expect(0, 66527, '\p{^Blk=__old_persian}', "");
Expect(0, 66527, '\P{Blk=__old_persian}', "");
Expect(1, 66527, '\P{^Blk=__old_persian}', "");
Expect(0, 66528, '\p{Blk=__old_persian}', "");
Expect(1, 66528, '\p{^Blk=__old_persian}', "");
Expect(1, 66528, '\P{Blk=__old_persian}', "");
Expect(0, 66528, '\P{^Blk=__old_persian}', "");
Error('\p{Is_Block=:=-OLD_Persian}');
Error('\P{Is_Block=:=-OLD_Persian}');
Expect(1, 66527, '\p{Is_Block=oldpersian}', "");
Expect(0, 66527, '\p{^Is_Block=oldpersian}', "");
Expect(0, 66527, '\P{Is_Block=oldpersian}', "");
Expect(1, 66527, '\P{^Is_Block=oldpersian}', "");
Expect(0, 66528, '\p{Is_Block=oldpersian}', "");
Expect(1, 66528, '\p{^Is_Block=oldpersian}', "");
Expect(1, 66528, '\P{Is_Block=oldpersian}', "");
Expect(0, 66528, '\P{^Is_Block=oldpersian}', "");
Expect(1, 66527, '\p{Is_Block=	Old_PERSIAN}', "");
Expect(0, 66527, '\p{^Is_Block=	Old_PERSIAN}', "");
Expect(0, 66527, '\P{Is_Block=	Old_PERSIAN}', "");
Expect(1, 66527, '\P{^Is_Block=	Old_PERSIAN}', "");
Expect(0, 66528, '\p{Is_Block=	Old_PERSIAN}', "");
Expect(1, 66528, '\p{^Is_Block=	Old_PERSIAN}', "");
Expect(1, 66528, '\P{Is_Block=	Old_PERSIAN}', "");
Expect(0, 66528, '\P{^Is_Block=	Old_PERSIAN}', "");
Error('\p{Is_Blk=:=	Old_persian}');
Error('\P{Is_Blk=:=	Old_persian}');
Expect(1, 66527, '\p{Is_Blk=oldpersian}', "");
Expect(0, 66527, '\p{^Is_Blk=oldpersian}', "");
Expect(0, 66527, '\P{Is_Blk=oldpersian}', "");
Expect(1, 66527, '\P{^Is_Blk=oldpersian}', "");
Expect(0, 66528, '\p{Is_Blk=oldpersian}', "");
Expect(1, 66528, '\p{^Is_Blk=oldpersian}', "");
Expect(1, 66528, '\P{Is_Blk=oldpersian}', "");
Expect(0, 66528, '\P{^Is_Blk=oldpersian}', "");
Expect(1, 66527, '\p{Is_Blk:   		Old_Persian}', "");
Expect(0, 66527, '\p{^Is_Blk:   		Old_Persian}', "");
Expect(0, 66527, '\P{Is_Blk:   		Old_Persian}', "");
Expect(1, 66527, '\P{^Is_Blk:   		Old_Persian}', "");
Expect(0, 66528, '\p{Is_Blk:   		Old_Persian}', "");
Expect(1, 66528, '\p{^Is_Blk:   		Old_Persian}', "");
Expect(1, 66528, '\P{Is_Blk:   		Old_Persian}', "");
Expect(0, 66528, '\P{^Is_Blk:   		Old_Persian}', "");
Error('\p{Block=-:=Old_SOUTH_arabian}');
Error('\P{Block=-:=Old_SOUTH_arabian}');
Expect(1, 68223, '\p{Block=oldsoutharabian}', "");
Expect(0, 68223, '\p{^Block=oldsoutharabian}', "");
Expect(0, 68223, '\P{Block=oldsoutharabian}', "");
Expect(1, 68223, '\P{^Block=oldsoutharabian}', "");
Expect(0, 68224, '\p{Block=oldsoutharabian}', "");
Expect(1, 68224, '\p{^Block=oldsoutharabian}', "");
Expect(1, 68224, '\P{Block=oldsoutharabian}', "");
Expect(0, 68224, '\P{^Block=oldsoutharabian}', "");
Expect(1, 68223, '\p{Block=OLD_south_Arabian}', "");
Expect(0, 68223, '\p{^Block=OLD_south_Arabian}', "");
Expect(0, 68223, '\P{Block=OLD_south_Arabian}', "");
Expect(1, 68223, '\P{^Block=OLD_south_Arabian}', "");
Expect(0, 68224, '\p{Block=OLD_south_Arabian}', "");
Expect(1, 68224, '\p{^Block=OLD_south_Arabian}', "");
Expect(1, 68224, '\P{Block=OLD_south_Arabian}', "");
Expect(0, 68224, '\P{^Block=OLD_south_Arabian}', "");
Error('\p{Blk=/a/  old_South_Arabian}');
Error('\P{Blk=/a/  old_South_Arabian}');
Expect(1, 68223, '\p{Blk=oldsoutharabian}', "");
Expect(0, 68223, '\p{^Blk=oldsoutharabian}', "");
Expect(0, 68223, '\P{Blk=oldsoutharabian}', "");
Expect(1, 68223, '\P{^Blk=oldsoutharabian}', "");
Expect(0, 68224, '\p{Blk=oldsoutharabian}', "");
Expect(1, 68224, '\p{^Blk=oldsoutharabian}', "");
Expect(1, 68224, '\P{Blk=oldsoutharabian}', "");
Expect(0, 68224, '\P{^Blk=oldsoutharabian}', "");
Expect(1, 68223, '\p{Blk= Old_south_Arabian}', "");
Expect(0, 68223, '\p{^Blk= Old_south_Arabian}', "");
Expect(0, 68223, '\P{Blk= Old_south_Arabian}', "");
Expect(1, 68223, '\P{^Blk= Old_south_Arabian}', "");
Expect(0, 68224, '\p{Blk= Old_south_Arabian}', "");
Expect(1, 68224, '\p{^Blk= Old_south_Arabian}', "");
Expect(1, 68224, '\P{Blk= Old_south_Arabian}', "");
Expect(0, 68224, '\P{^Blk= Old_south_Arabian}', "");
Error('\p{Is_Block=-/a/old_South_arabian}');
Error('\P{Is_Block=-/a/old_South_arabian}');
Expect(1, 68223, '\p{Is_Block=oldsoutharabian}', "");
Expect(0, 68223, '\p{^Is_Block=oldsoutharabian}', "");
Expect(0, 68223, '\P{Is_Block=oldsoutharabian}', "");
Expect(1, 68223, '\P{^Is_Block=oldsoutharabian}', "");
Expect(0, 68224, '\p{Is_Block=oldsoutharabian}', "");
Expect(1, 68224, '\p{^Is_Block=oldsoutharabian}', "");
Expect(1, 68224, '\P{Is_Block=oldsoutharabian}', "");
Expect(0, 68224, '\P{^Is_Block=oldsoutharabian}', "");
Expect(1, 68223, '\p{Is_Block=_	old_SOUTH_ARABIAN}', "");
Expect(0, 68223, '\p{^Is_Block=_	old_SOUTH_ARABIAN}', "");
Expect(0, 68223, '\P{Is_Block=_	old_SOUTH_ARABIAN}', "");
Expect(1, 68223, '\P{^Is_Block=_	old_SOUTH_ARABIAN}', "");
Expect(0, 68224, '\p{Is_Block=_	old_SOUTH_ARABIAN}', "");
Expect(1, 68224, '\p{^Is_Block=_	old_SOUTH_ARABIAN}', "");
Expect(1, 68224, '\P{Is_Block=_	old_SOUTH_ARABIAN}', "");
Expect(0, 68224, '\P{^Is_Block=_	old_SOUTH_ARABIAN}', "");
Error('\p{Is_Blk=	Old_SOUTH_Arabian/a/}');
Error('\P{Is_Blk=	Old_SOUTH_Arabian/a/}');
Expect(1, 68223, '\p{Is_Blk=oldsoutharabian}', "");
Expect(0, 68223, '\p{^Is_Blk=oldsoutharabian}', "");
Expect(0, 68223, '\P{Is_Blk=oldsoutharabian}', "");
Expect(1, 68223, '\P{^Is_Blk=oldsoutharabian}', "");
Expect(0, 68224, '\p{Is_Blk=oldsoutharabian}', "");
Expect(1, 68224, '\p{^Is_Blk=oldsoutharabian}', "");
Expect(1, 68224, '\P{Is_Blk=oldsoutharabian}', "");
Expect(0, 68224, '\P{^Is_Blk=oldsoutharabian}', "");
Expect(1, 68223, '\p{Is_Blk= Old_SOUTH_ARABIAN}', "");
Expect(0, 68223, '\p{^Is_Blk= Old_SOUTH_ARABIAN}', "");
Expect(0, 68223, '\P{Is_Blk= Old_SOUTH_ARABIAN}', "");
Expect(1, 68223, '\P{^Is_Blk= Old_SOUTH_ARABIAN}', "");
Expect(0, 68224, '\p{Is_Blk= Old_SOUTH_ARABIAN}', "");
Expect(1, 68224, '\p{^Is_Blk= Old_SOUTH_ARABIAN}', "");
Expect(1, 68224, '\P{Is_Blk= Old_SOUTH_ARABIAN}', "");
Expect(0, 68224, '\P{^Is_Blk= Old_SOUTH_ARABIAN}', "");
Error('\p{Block=/a/ Old_Turkic}');
Error('\P{Block=/a/ Old_Turkic}');
Expect(1, 68687, '\p{Block=oldturkic}', "");
Expect(0, 68687, '\p{^Block=oldturkic}', "");
Expect(0, 68687, '\P{Block=oldturkic}', "");
Expect(1, 68687, '\P{^Block=oldturkic}', "");
Expect(0, 68688, '\p{Block=oldturkic}', "");
Expect(1, 68688, '\p{^Block=oldturkic}', "");
Expect(1, 68688, '\P{Block=oldturkic}', "");
Expect(0, 68688, '\P{^Block=oldturkic}', "");
Expect(1, 68687, '\p{Block= -OLD_turkic}', "");
Expect(0, 68687, '\p{^Block= -OLD_turkic}', "");
Expect(0, 68687, '\P{Block= -OLD_turkic}', "");
Expect(1, 68687, '\P{^Block= -OLD_turkic}', "");
Expect(0, 68688, '\p{Block= -OLD_turkic}', "");
Expect(1, 68688, '\p{^Block= -OLD_turkic}', "");
Expect(1, 68688, '\P{Block= -OLD_turkic}', "");
Expect(0, 68688, '\P{^Block= -OLD_turkic}', "");
Error('\p{Blk=	old_Turkic/a/}');
Error('\P{Blk=	old_Turkic/a/}');
Expect(1, 68687, '\p{Blk:oldturkic}', "");
Expect(0, 68687, '\p{^Blk:oldturkic}', "");
Expect(0, 68687, '\P{Blk:oldturkic}', "");
Expect(1, 68687, '\P{^Blk:oldturkic}', "");
Expect(0, 68688, '\p{Blk:oldturkic}', "");
Expect(1, 68688, '\p{^Blk:oldturkic}', "");
Expect(1, 68688, '\P{Blk:oldturkic}', "");
Expect(0, 68688, '\P{^Blk:oldturkic}', "");
Expect(1, 68687, '\p{Blk=  OLD_turkic}', "");
Expect(0, 68687, '\p{^Blk=  OLD_turkic}', "");
Expect(0, 68687, '\P{Blk=  OLD_turkic}', "");
Expect(1, 68687, '\P{^Blk=  OLD_turkic}', "");
Expect(0, 68688, '\p{Blk=  OLD_turkic}', "");
Expect(1, 68688, '\p{^Blk=  OLD_turkic}', "");
Expect(1, 68688, '\P{Blk=  OLD_turkic}', "");
Expect(0, 68688, '\P{^Blk=  OLD_turkic}', "");
Error('\p{Is_Block= :=Old_Turkic}');
Error('\P{Is_Block= :=Old_Turkic}');
Expect(1, 68687, '\p{Is_Block=oldturkic}', "");
Expect(0, 68687, '\p{^Is_Block=oldturkic}', "");
Expect(0, 68687, '\P{Is_Block=oldturkic}', "");
Expect(1, 68687, '\P{^Is_Block=oldturkic}', "");
Expect(0, 68688, '\p{Is_Block=oldturkic}', "");
Expect(1, 68688, '\p{^Is_Block=oldturkic}', "");
Expect(1, 68688, '\P{Is_Block=oldturkic}', "");
Expect(0, 68688, '\P{^Is_Block=oldturkic}', "");
Expect(1, 68687, '\p{Is_Block=	_old_turkic}', "");
Expect(0, 68687, '\p{^Is_Block=	_old_turkic}', "");
Expect(0, 68687, '\P{Is_Block=	_old_turkic}', "");
Expect(1, 68687, '\P{^Is_Block=	_old_turkic}', "");
Expect(0, 68688, '\p{Is_Block=	_old_turkic}', "");
Expect(1, 68688, '\p{^Is_Block=	_old_turkic}', "");
Expect(1, 68688, '\P{Is_Block=	_old_turkic}', "");
Expect(0, 68688, '\P{^Is_Block=	_old_turkic}', "");
Error('\p{Is_Blk=/a/ 	old_Turkic}');
Error('\P{Is_Blk=/a/ 	old_Turkic}');
Expect(1, 68687, '\p{Is_Blk=oldturkic}', "");
Expect(0, 68687, '\p{^Is_Blk=oldturkic}', "");
Expect(0, 68687, '\P{Is_Blk=oldturkic}', "");
Expect(1, 68687, '\P{^Is_Blk=oldturkic}', "");
Expect(0, 68688, '\p{Is_Blk=oldturkic}', "");
Expect(1, 68688, '\p{^Is_Blk=oldturkic}', "");
Expect(1, 68688, '\P{Is_Blk=oldturkic}', "");
Expect(0, 68688, '\P{^Is_Blk=oldturkic}', "");
Expect(1, 68687, '\p{Is_Blk=-Old_turkic}', "");
Expect(0, 68687, '\p{^Is_Blk=-Old_turkic}', "");
Expect(0, 68687, '\P{Is_Blk=-Old_turkic}', "");
Expect(1, 68687, '\P{^Is_Blk=-Old_turkic}', "");
Expect(0, 68688, '\p{Is_Blk=-Old_turkic}', "");
Expect(1, 68688, '\p{^Is_Blk=-Old_turkic}', "");
Expect(1, 68688, '\P{Is_Blk=-Old_turkic}', "");
Expect(0, 68688, '\P{^Is_Blk=-Old_turkic}', "");
Error('\p{Block=-:=oriya}');
Error('\P{Block=-:=oriya}');
Expect(1, 2943, '\p{Block=oriya}', "");
Expect(0, 2943, '\p{^Block=oriya}', "");
Expect(0, 2943, '\P{Block=oriya}', "");
Expect(1, 2943, '\P{^Block=oriya}', "");
Expect(0, 2944, '\p{Block=oriya}', "");
Expect(1, 2944, '\p{^Block=oriya}', "");
Expect(1, 2944, '\P{Block=oriya}', "");
Expect(0, 2944, '\P{^Block=oriya}', "");
Expect(1, 2943, '\p{Block=__Oriya}', "");
Expect(0, 2943, '\p{^Block=__Oriya}', "");
Expect(0, 2943, '\P{Block=__Oriya}', "");
Expect(1, 2943, '\P{^Block=__Oriya}', "");
Expect(0, 2944, '\p{Block=__Oriya}', "");
Expect(1, 2944, '\p{^Block=__Oriya}', "");
Expect(1, 2944, '\P{Block=__Oriya}', "");
Expect(0, 2944, '\P{^Block=__Oriya}', "");
Error('\p{Blk=/a/_	Oriya}');
Error('\P{Blk=/a/_	Oriya}');
Expect(1, 2943, '\p{Blk=oriya}', "");
Expect(0, 2943, '\p{^Blk=oriya}', "");
Expect(0, 2943, '\P{Blk=oriya}', "");
Expect(1, 2943, '\P{^Blk=oriya}', "");
Expect(0, 2944, '\p{Blk=oriya}', "");
Expect(1, 2944, '\p{^Blk=oriya}', "");
Expect(1, 2944, '\P{Blk=oriya}', "");
Expect(0, 2944, '\P{^Blk=oriya}', "");
Expect(1, 2943, '\p{Blk=	Oriya}', "");
Expect(0, 2943, '\p{^Blk=	Oriya}', "");
Expect(0, 2943, '\P{Blk=	Oriya}', "");
Expect(1, 2943, '\P{^Blk=	Oriya}', "");
Expect(0, 2944, '\p{Blk=	Oriya}', "");
Expect(1, 2944, '\p{^Blk=	Oriya}', "");
Expect(1, 2944, '\P{Blk=	Oriya}', "");
Expect(0, 2944, '\P{^Blk=	Oriya}', "");
Error('\p{Is_Block=_	oriya/a/}');
Error('\P{Is_Block=_	oriya/a/}');
Expect(1, 2943, '\p{Is_Block=oriya}', "");
Expect(0, 2943, '\p{^Is_Block=oriya}', "");
Expect(0, 2943, '\P{Is_Block=oriya}', "");
Expect(1, 2943, '\P{^Is_Block=oriya}', "");
Expect(0, 2944, '\p{Is_Block=oriya}', "");
Expect(1, 2944, '\p{^Is_Block=oriya}', "");
Expect(1, 2944, '\P{Is_Block=oriya}', "");
Expect(0, 2944, '\P{^Is_Block=oriya}', "");
Expect(1, 2943, '\p{Is_Block=-_Oriya}', "");
Expect(0, 2943, '\p{^Is_Block=-_Oriya}', "");
Expect(0, 2943, '\P{Is_Block=-_Oriya}', "");
Expect(1, 2943, '\P{^Is_Block=-_Oriya}', "");
Expect(0, 2944, '\p{Is_Block=-_Oriya}', "");
Expect(1, 2944, '\p{^Is_Block=-_Oriya}', "");
Expect(1, 2944, '\P{Is_Block=-_Oriya}', "");
Expect(0, 2944, '\P{^Is_Block=-_Oriya}', "");
Error('\p{Is_Blk=/a/_ Oriya}');
Error('\P{Is_Blk=/a/_ Oriya}');
Expect(1, 2943, '\p{Is_Blk=oriya}', "");
Expect(0, 2943, '\p{^Is_Blk=oriya}', "");
Expect(0, 2943, '\P{Is_Blk=oriya}', "");
Expect(1, 2943, '\P{^Is_Blk=oriya}', "");
Expect(0, 2944, '\p{Is_Blk=oriya}', "");
Expect(1, 2944, '\p{^Is_Blk=oriya}', "");
Expect(1, 2944, '\P{Is_Blk=oriya}', "");
Expect(0, 2944, '\P{^Is_Blk=oriya}', "");
Expect(1, 2943, '\p{Is_Blk=_ORIYA}', "");
Expect(0, 2943, '\p{^Is_Blk=_ORIYA}', "");
Expect(0, 2943, '\P{Is_Blk=_ORIYA}', "");
Expect(1, 2943, '\P{^Is_Blk=_ORIYA}', "");
Expect(0, 2944, '\p{Is_Blk=_ORIYA}', "");
Expect(1, 2944, '\p{^Is_Blk=_ORIYA}', "");
Expect(1, 2944, '\P{Is_Blk=_ORIYA}', "");
Expect(0, 2944, '\P{^Is_Blk=_ORIYA}', "");
Error('\p{Block=:=- Osmanya}');
Error('\P{Block=:=- Osmanya}');
Expect(1, 66735, '\p{Block=osmanya}', "");
Expect(0, 66735, '\p{^Block=osmanya}', "");
Expect(0, 66735, '\P{Block=osmanya}', "");
Expect(1, 66735, '\P{^Block=osmanya}', "");
Expect(0, 66736, '\p{Block=osmanya}', "");
Expect(1, 66736, '\p{^Block=osmanya}', "");
Expect(1, 66736, '\P{Block=osmanya}', "");
Expect(0, 66736, '\P{^Block=osmanya}', "");
Expect(1, 66735, '\p{Block=	 Osmanya}', "");
Expect(0, 66735, '\p{^Block=	 Osmanya}', "");
Expect(0, 66735, '\P{Block=	 Osmanya}', "");
Expect(1, 66735, '\P{^Block=	 Osmanya}', "");
Expect(0, 66736, '\p{Block=	 Osmanya}', "");
Expect(1, 66736, '\p{^Block=	 Osmanya}', "");
Expect(1, 66736, '\P{Block=	 Osmanya}', "");
Expect(0, 66736, '\P{^Block=	 Osmanya}', "");
Error('\p{Blk=_ OSMANYA:=}');
Error('\P{Blk=_ OSMANYA:=}');
Expect(1, 66735, '\p{Blk=osmanya}', "");
Expect(0, 66735, '\p{^Blk=osmanya}', "");
Expect(0, 66735, '\P{Blk=osmanya}', "");
Expect(1, 66735, '\P{^Blk=osmanya}', "");
Expect(0, 66736, '\p{Blk=osmanya}', "");
Expect(1, 66736, '\p{^Blk=osmanya}', "");
Expect(1, 66736, '\P{Blk=osmanya}', "");
Expect(0, 66736, '\P{^Blk=osmanya}', "");
Expect(1, 66735, '\p{Blk=	_Osmanya}', "");
Expect(0, 66735, '\p{^Blk=	_Osmanya}', "");
Expect(0, 66735, '\P{Blk=	_Osmanya}', "");
Expect(1, 66735, '\P{^Blk=	_Osmanya}', "");
Expect(0, 66736, '\p{Blk=	_Osmanya}', "");
Expect(1, 66736, '\p{^Blk=	_Osmanya}', "");
Expect(1, 66736, '\P{Blk=	_Osmanya}', "");
Expect(0, 66736, '\P{^Blk=	_Osmanya}', "");
Error('\p{Is_Block=	:=osmanya}');
Error('\P{Is_Block=	:=osmanya}');
Expect(1, 66735, '\p{Is_Block=osmanya}', "");
Expect(0, 66735, '\p{^Is_Block=osmanya}', "");
Expect(0, 66735, '\P{Is_Block=osmanya}', "");
Expect(1, 66735, '\P{^Is_Block=osmanya}', "");
Expect(0, 66736, '\p{Is_Block=osmanya}', "");
Expect(1, 66736, '\p{^Is_Block=osmanya}', "");
Expect(1, 66736, '\P{Is_Block=osmanya}', "");
Expect(0, 66736, '\P{^Is_Block=osmanya}', "");
Expect(1, 66735, '\p{Is_Block=_-Osmanya}', "");
Expect(0, 66735, '\p{^Is_Block=_-Osmanya}', "");
Expect(0, 66735, '\P{Is_Block=_-Osmanya}', "");
Expect(1, 66735, '\P{^Is_Block=_-Osmanya}', "");
Expect(0, 66736, '\p{Is_Block=_-Osmanya}', "");
Expect(1, 66736, '\p{^Is_Block=_-Osmanya}', "");
Expect(1, 66736, '\P{Is_Block=_-Osmanya}', "");
Expect(0, 66736, '\P{^Is_Block=_-Osmanya}', "");
Error('\p{Is_Blk=-/a/Osmanya}');
Error('\P{Is_Blk=-/a/Osmanya}');
Expect(1, 66735, '\p{Is_Blk=osmanya}', "");
Expect(0, 66735, '\p{^Is_Blk=osmanya}', "");
Expect(0, 66735, '\P{Is_Blk=osmanya}', "");
Expect(1, 66735, '\P{^Is_Blk=osmanya}', "");
Expect(0, 66736, '\p{Is_Blk=osmanya}', "");
Expect(1, 66736, '\p{^Is_Blk=osmanya}', "");
Expect(1, 66736, '\P{Is_Blk=osmanya}', "");
Expect(0, 66736, '\P{^Is_Blk=osmanya}', "");
Expect(1, 66735, '\p{Is_Blk=	Osmanya}', "");
Expect(0, 66735, '\p{^Is_Blk=	Osmanya}', "");
Expect(0, 66735, '\P{Is_Blk=	Osmanya}', "");
Expect(1, 66735, '\P{^Is_Blk=	Osmanya}', "");
Expect(0, 66736, '\p{Is_Blk=	Osmanya}', "");
Expect(1, 66736, '\p{^Is_Blk=	Osmanya}', "");
Expect(1, 66736, '\P{Is_Blk=	Osmanya}', "");
Expect(0, 66736, '\P{^Is_Blk=	Osmanya}', "");
Error('\p{Block=__PRIVATE_use_area/a/}');
Error('\P{Block=__PRIVATE_use_area/a/}');
Expect(1, 63743, '\p{Block=privateusearea}', "");
Expect(0, 63743, '\p{^Block=privateusearea}', "");
Expect(0, 63743, '\P{Block=privateusearea}', "");
Expect(1, 63743, '\P{^Block=privateusearea}', "");
Expect(0, 63744, '\p{Block=privateusearea}', "");
Expect(1, 63744, '\p{^Block=privateusearea}', "");
Expect(1, 63744, '\P{Block=privateusearea}', "");
Expect(0, 63744, '\P{^Block=privateusearea}', "");
Expect(1, 63743, '\p{Block=		PRIVATE_use_AREA}', "");
Expect(0, 63743, '\p{^Block=		PRIVATE_use_AREA}', "");
Expect(0, 63743, '\P{Block=		PRIVATE_use_AREA}', "");
Expect(1, 63743, '\P{^Block=		PRIVATE_use_AREA}', "");
Expect(0, 63744, '\p{Block=		PRIVATE_use_AREA}', "");
Expect(1, 63744, '\p{^Block=		PRIVATE_use_AREA}', "");
Expect(1, 63744, '\P{Block=		PRIVATE_use_AREA}', "");
Expect(0, 63744, '\P{^Block=		PRIVATE_use_AREA}', "");
Error('\p{Blk=_/a/pua}');
Error('\P{Blk=_/a/pua}');
Expect(1, 63743, '\p{Blk=pua}', "");
Expect(0, 63743, '\p{^Blk=pua}', "");
Expect(0, 63743, '\P{Blk=pua}', "");
Expect(1, 63743, '\P{^Blk=pua}', "");
Expect(0, 63744, '\p{Blk=pua}', "");
Expect(1, 63744, '\p{^Blk=pua}', "");
Expect(1, 63744, '\P{Blk=pua}', "");
Expect(0, 63744, '\P{^Blk=pua}', "");
Expect(1, 63743, '\p{Blk:	 pua}', "");
Expect(0, 63743, '\p{^Blk:	 pua}', "");
Expect(0, 63743, '\P{Blk:	 pua}', "");
Expect(1, 63743, '\P{^Blk:	 pua}', "");
Expect(0, 63744, '\p{Blk:	 pua}', "");
Expect(1, 63744, '\p{^Blk:	 pua}', "");
Expect(1, 63744, '\P{Blk:	 pua}', "");
Expect(0, 63744, '\P{^Blk:	 pua}', "");
Error('\p{Is_Block=/a/-Private_USE}');
Error('\P{Is_Block=/a/-Private_USE}');
Expect(1, 63743, '\p{Is_Block=privateuse}', "");
Expect(0, 63743, '\p{^Is_Block=privateuse}', "");
Expect(0, 63743, '\P{Is_Block=privateuse}', "");
Expect(1, 63743, '\P{^Is_Block=privateuse}', "");
Expect(0, 63744, '\p{Is_Block=privateuse}', "");
Expect(1, 63744, '\p{^Is_Block=privateuse}', "");
Expect(1, 63744, '\P{Is_Block=privateuse}', "");
Expect(0, 63744, '\P{^Is_Block=privateuse}', "");
Expect(1, 63743, '\p{Is_Block= private_use}', "");
Expect(0, 63743, '\p{^Is_Block= private_use}', "");
Expect(0, 63743, '\P{Is_Block= private_use}', "");
Expect(1, 63743, '\P{^Is_Block= private_use}', "");
Expect(0, 63744, '\p{Is_Block= private_use}', "");
Expect(1, 63744, '\p{^Is_Block= private_use}', "");
Expect(1, 63744, '\P{Is_Block= private_use}', "");
Expect(0, 63744, '\P{^Is_Block= private_use}', "");
Error('\p{Is_Blk:	-Private_USE_Area/a/}');
Error('\P{Is_Blk:	-Private_USE_Area/a/}');
Expect(1, 63743, '\p{Is_Blk=privateusearea}', "");
Expect(0, 63743, '\p{^Is_Blk=privateusearea}', "");
Expect(0, 63743, '\P{Is_Blk=privateusearea}', "");
Expect(1, 63743, '\P{^Is_Blk=privateusearea}', "");
Expect(0, 63744, '\p{Is_Blk=privateusearea}', "");
Expect(1, 63744, '\p{^Is_Blk=privateusearea}', "");
Expect(1, 63744, '\P{Is_Blk=privateusearea}', "");
Expect(0, 63744, '\P{^Is_Blk=privateusearea}', "");
Expect(1, 63743, '\p{Is_Blk=_private_USE_Area}', "");
Expect(0, 63743, '\p{^Is_Blk=_private_USE_Area}', "");
Expect(0, 63743, '\P{Is_Blk=_private_USE_Area}', "");
Expect(1, 63743, '\P{^Is_Blk=_private_USE_Area}', "");
Expect(0, 63744, '\p{Is_Blk=_private_USE_Area}', "");
Expect(1, 63744, '\p{^Is_Blk=_private_USE_Area}', "");
Expect(1, 63744, '\P{Is_Blk=_private_USE_Area}', "");
Expect(0, 63744, '\P{^Is_Blk=_private_USE_Area}', "");
Error('\p{Block=__phags_Pa/a/}');
Error('\P{Block=__phags_Pa/a/}');
Expect(1, 43135, '\p{Block=phagspa}', "");
Expect(0, 43135, '\p{^Block=phagspa}', "");
Expect(0, 43135, '\P{Block=phagspa}', "");
Expect(1, 43135, '\P{^Block=phagspa}', "");
Expect(0, 43136, '\p{Block=phagspa}', "");
Expect(1, 43136, '\p{^Block=phagspa}', "");
Expect(1, 43136, '\P{Block=phagspa}', "");
Expect(0, 43136, '\P{^Block=phagspa}', "");
Expect(1, 43135, '\p{Block=_PHAGS_pa}', "");
Expect(0, 43135, '\p{^Block=_PHAGS_pa}', "");
Expect(0, 43135, '\P{Block=_PHAGS_pa}', "");
Expect(1, 43135, '\P{^Block=_PHAGS_pa}', "");
Expect(0, 43136, '\p{Block=_PHAGS_pa}', "");
Expect(1, 43136, '\p{^Block=_PHAGS_pa}', "");
Expect(1, 43136, '\P{Block=_PHAGS_pa}', "");
Expect(0, 43136, '\P{^Block=_PHAGS_pa}', "");
Error('\p{Blk=/a/--phags_Pa}');
Error('\P{Blk=/a/--phags_Pa}');
Expect(1, 43135, '\p{Blk=phagspa}', "");
Expect(0, 43135, '\p{^Blk=phagspa}', "");
Expect(0, 43135, '\P{Blk=phagspa}', "");
Expect(1, 43135, '\P{^Blk=phagspa}', "");
Expect(0, 43136, '\p{Blk=phagspa}', "");
Expect(1, 43136, '\p{^Blk=phagspa}', "");
Expect(1, 43136, '\P{Blk=phagspa}', "");
Expect(0, 43136, '\P{^Blk=phagspa}', "");
Expect(1, 43135, '\p{Blk=  phags_Pa}', "");
Expect(0, 43135, '\p{^Blk=  phags_Pa}', "");
Expect(0, 43135, '\P{Blk=  phags_Pa}', "");
Expect(1, 43135, '\P{^Blk=  phags_Pa}', "");
Expect(0, 43136, '\p{Blk=  phags_Pa}', "");
Expect(1, 43136, '\p{^Blk=  phags_Pa}', "");
Expect(1, 43136, '\P{Blk=  phags_Pa}', "");
Expect(0, 43136, '\P{^Blk=  phags_Pa}', "");
Error('\p{Is_Block=/a/--Phags_PA}');
Error('\P{Is_Block=/a/--Phags_PA}');
Expect(1, 43135, '\p{Is_Block=phagspa}', "");
Expect(0, 43135, '\p{^Is_Block=phagspa}', "");
Expect(0, 43135, '\P{Is_Block=phagspa}', "");
Expect(1, 43135, '\P{^Is_Block=phagspa}', "");
Expect(0, 43136, '\p{Is_Block=phagspa}', "");
Expect(1, 43136, '\p{^Is_Block=phagspa}', "");
Expect(1, 43136, '\P{Is_Block=phagspa}', "");
Expect(0, 43136, '\P{^Is_Block=phagspa}', "");
Expect(1, 43135, '\p{Is_Block= 	phags_PA}', "");
Expect(0, 43135, '\p{^Is_Block= 	phags_PA}', "");
Expect(0, 43135, '\P{Is_Block= 	phags_PA}', "");
Expect(1, 43135, '\P{^Is_Block= 	phags_PA}', "");
Expect(0, 43136, '\p{Is_Block= 	phags_PA}', "");
Expect(1, 43136, '\p{^Is_Block= 	phags_PA}', "");
Expect(1, 43136, '\P{Is_Block= 	phags_PA}', "");
Expect(0, 43136, '\P{^Is_Block= 	phags_PA}', "");
Error('\p{Is_Blk=--phags_pa/a/}');
Error('\P{Is_Blk=--phags_pa/a/}');
Expect(1, 43135, '\p{Is_Blk=phagspa}', "");
Expect(0, 43135, '\p{^Is_Blk=phagspa}', "");
Expect(0, 43135, '\P{Is_Blk=phagspa}', "");
Expect(1, 43135, '\P{^Is_Blk=phagspa}', "");
Expect(0, 43136, '\p{Is_Blk=phagspa}', "");
Expect(1, 43136, '\p{^Is_Blk=phagspa}', "");
Expect(1, 43136, '\P{Is_Blk=phagspa}', "");
Expect(0, 43136, '\P{^Is_Blk=phagspa}', "");
Expect(1, 43135, '\p{Is_Blk=	_PHAGS_pa}', "");
Expect(0, 43135, '\p{^Is_Blk=	_PHAGS_pa}', "");
Expect(0, 43135, '\P{Is_Blk=	_PHAGS_pa}', "");
Expect(1, 43135, '\P{^Is_Blk=	_PHAGS_pa}', "");
Expect(0, 43136, '\p{Is_Blk=	_PHAGS_pa}', "");
Expect(1, 43136, '\p{^Is_Blk=	_PHAGS_pa}', "");
Expect(1, 43136, '\P{Is_Blk=	_PHAGS_pa}', "");
Expect(0, 43136, '\P{^Is_Blk=	_PHAGS_pa}', "");
Error('\p{Block=  PHAISTOS_Disc:=}');
Error('\P{Block=  PHAISTOS_Disc:=}');
Expect(1, 66047, '\p{Block=phaistosdisc}', "");
Expect(0, 66047, '\p{^Block=phaistosdisc}', "");
Expect(0, 66047, '\P{Block=phaistosdisc}', "");
Expect(1, 66047, '\P{^Block=phaistosdisc}', "");
Expect(0, 66048, '\p{Block=phaistosdisc}', "");
Expect(1, 66048, '\p{^Block=phaistosdisc}', "");
Expect(1, 66048, '\P{Block=phaistosdisc}', "");
Expect(0, 66048, '\P{^Block=phaistosdisc}', "");
Expect(1, 66047, '\p{Block= Phaistos_Disc}', "");
Expect(0, 66047, '\p{^Block= Phaistos_Disc}', "");
Expect(0, 66047, '\P{Block= Phaistos_Disc}', "");
Expect(1, 66047, '\P{^Block= Phaistos_Disc}', "");
Expect(0, 66048, '\p{Block= Phaistos_Disc}', "");
Expect(1, 66048, '\p{^Block= Phaistos_Disc}', "");
Expect(1, 66048, '\P{Block= Phaistos_Disc}', "");
Expect(0, 66048, '\P{^Block= Phaistos_Disc}', "");
Error('\p{Blk=/a/ _Phaistos}');
Error('\P{Blk=/a/ _Phaistos}');
Expect(1, 66047, '\p{Blk=phaistos}', "");
Expect(0, 66047, '\p{^Blk=phaistos}', "");
Expect(0, 66047, '\P{Blk=phaistos}', "");
Expect(1, 66047, '\P{^Blk=phaistos}', "");
Expect(0, 66048, '\p{Blk=phaistos}', "");
Expect(1, 66048, '\p{^Blk=phaistos}', "");
Expect(1, 66048, '\P{Blk=phaistos}', "");
Expect(0, 66048, '\P{^Blk=phaistos}', "");
Expect(1, 66047, '\p{Blk= _Phaistos}', "");
Expect(0, 66047, '\p{^Blk= _Phaistos}', "");
Expect(0, 66047, '\P{Blk= _Phaistos}', "");
Expect(1, 66047, '\P{^Blk= _Phaistos}', "");
Expect(0, 66048, '\p{Blk= _Phaistos}', "");
Expect(1, 66048, '\p{^Blk= _Phaistos}', "");
Expect(1, 66048, '\P{Blk= _Phaistos}', "");
Expect(0, 66048, '\P{^Blk= _Phaistos}', "");
Error('\p{Is_Block: :=_	Phaistos_disc}');
Error('\P{Is_Block: :=_	Phaistos_disc}');
Expect(1, 66047, '\p{Is_Block=phaistosdisc}', "");
Expect(0, 66047, '\p{^Is_Block=phaistosdisc}', "");
Expect(0, 66047, '\P{Is_Block=phaistosdisc}', "");
Expect(1, 66047, '\P{^Is_Block=phaistosdisc}', "");
Expect(0, 66048, '\p{Is_Block=phaistosdisc}', "");
Expect(1, 66048, '\p{^Is_Block=phaistosdisc}', "");
Expect(1, 66048, '\P{Is_Block=phaistosdisc}', "");
Expect(0, 66048, '\P{^Is_Block=phaistosdisc}', "");
Expect(1, 66047, '\p{Is_Block=_-Phaistos_Disc}', "");
Expect(0, 66047, '\p{^Is_Block=_-Phaistos_Disc}', "");
Expect(0, 66047, '\P{Is_Block=_-Phaistos_Disc}', "");
Expect(1, 66047, '\P{^Is_Block=_-Phaistos_Disc}', "");
Expect(0, 66048, '\p{Is_Block=_-Phaistos_Disc}', "");
Expect(1, 66048, '\p{^Is_Block=_-Phaistos_Disc}', "");
Expect(1, 66048, '\P{Is_Block=_-Phaistos_Disc}', "");
Expect(0, 66048, '\P{^Is_Block=_-Phaistos_Disc}', "");
Error('\p{Is_Blk=:=--Phaistos}');
Error('\P{Is_Blk=:=--Phaistos}');
Expect(1, 66047, '\p{Is_Blk=phaistos}', "");
Expect(0, 66047, '\p{^Is_Blk=phaistos}', "");
Expect(0, 66047, '\P{Is_Blk=phaistos}', "");
Expect(1, 66047, '\P{^Is_Blk=phaistos}', "");
Expect(0, 66048, '\p{Is_Blk=phaistos}', "");
Expect(1, 66048, '\p{^Is_Blk=phaistos}', "");
Expect(1, 66048, '\P{Is_Blk=phaistos}', "");
Expect(0, 66048, '\P{^Is_Blk=phaistos}', "");
Expect(1, 66047, '\p{Is_Blk=  Phaistos}', "");
Expect(0, 66047, '\p{^Is_Blk=  Phaistos}', "");
Expect(0, 66047, '\P{Is_Blk=  Phaistos}', "");
Expect(1, 66047, '\P{^Is_Blk=  Phaistos}', "");
Expect(0, 66048, '\p{Is_Blk=  Phaistos}', "");
Expect(1, 66048, '\p{^Is_Blk=  Phaistos}', "");
Expect(1, 66048, '\P{Is_Blk=  Phaistos}', "");
Expect(0, 66048, '\P{^Is_Blk=  Phaistos}', "");
Error('\p{Block= /a/Phoenician}');
Error('\P{Block= /a/Phoenician}');
Expect(1, 67871, '\p{Block=phoenician}', "");
Expect(0, 67871, '\p{^Block=phoenician}', "");
Expect(0, 67871, '\P{Block=phoenician}', "");
Expect(1, 67871, '\P{^Block=phoenician}', "");
Expect(0, 67872, '\p{Block=phoenician}', "");
Expect(1, 67872, '\p{^Block=phoenician}', "");
Expect(1, 67872, '\P{Block=phoenician}', "");
Expect(0, 67872, '\P{^Block=phoenician}', "");
Expect(1, 67871, '\p{Block=-	phoenician}', "");
Expect(0, 67871, '\p{^Block=-	phoenician}', "");
Expect(0, 67871, '\P{Block=-	phoenician}', "");
Expect(1, 67871, '\P{^Block=-	phoenician}', "");
Expect(0, 67872, '\p{Block=-	phoenician}', "");
Expect(1, 67872, '\p{^Block=-	phoenician}', "");
Expect(1, 67872, '\P{Block=-	phoenician}', "");
Expect(0, 67872, '\P{^Block=-	phoenician}', "");
Error('\p{Blk=/a/_-Phoenician}');
Error('\P{Blk=/a/_-Phoenician}');
Expect(1, 67871, '\p{Blk:	phoenician}', "");
Expect(0, 67871, '\p{^Blk:	phoenician}', "");
Expect(0, 67871, '\P{Blk:	phoenician}', "");
Expect(1, 67871, '\P{^Blk:	phoenician}', "");
Expect(0, 67872, '\p{Blk:	phoenician}', "");
Expect(1, 67872, '\p{^Blk:	phoenician}', "");
Expect(1, 67872, '\P{Blk:	phoenician}', "");
Expect(0, 67872, '\P{^Blk:	phoenician}', "");
Expect(1, 67871, '\p{Blk=--PHOENICIAN}', "");
Expect(0, 67871, '\p{^Blk=--PHOENICIAN}', "");
Expect(0, 67871, '\P{Blk=--PHOENICIAN}', "");
Expect(1, 67871, '\P{^Blk=--PHOENICIAN}', "");
Expect(0, 67872, '\p{Blk=--PHOENICIAN}', "");
Expect(1, 67872, '\p{^Blk=--PHOENICIAN}', "");
Expect(1, 67872, '\P{Blk=--PHOENICIAN}', "");
Expect(0, 67872, '\P{^Blk=--PHOENICIAN}', "");
Error('\p{Is_Block: :=Phoenician}');
Error('\P{Is_Block: :=Phoenician}');
Expect(1, 67871, '\p{Is_Block=phoenician}', "");
Expect(0, 67871, '\p{^Is_Block=phoenician}', "");
Expect(0, 67871, '\P{Is_Block=phoenician}', "");
Expect(1, 67871, '\P{^Is_Block=phoenician}', "");
Expect(0, 67872, '\p{Is_Block=phoenician}', "");
Expect(1, 67872, '\p{^Is_Block=phoenician}', "");
Expect(1, 67872, '\P{Is_Block=phoenician}', "");
Expect(0, 67872, '\P{^Is_Block=phoenician}', "");
Expect(1, 67871, '\p{Is_Block= Phoenician}', "");
Expect(0, 67871, '\p{^Is_Block= Phoenician}', "");
Expect(0, 67871, '\P{Is_Block= Phoenician}', "");
Expect(1, 67871, '\P{^Is_Block= Phoenician}', "");
Expect(0, 67872, '\p{Is_Block= Phoenician}', "");
Expect(1, 67872, '\p{^Is_Block= Phoenician}', "");
Expect(1, 67872, '\P{Is_Block= Phoenician}', "");
Expect(0, 67872, '\P{^Is_Block= Phoenician}', "");
Error('\p{Is_Blk= Phoenician:=}');
Error('\P{Is_Blk= Phoenician:=}');
Expect(1, 67871, '\p{Is_Blk=phoenician}', "");
Expect(0, 67871, '\p{^Is_Blk=phoenician}', "");
Expect(0, 67871, '\P{Is_Blk=phoenician}', "");
Expect(1, 67871, '\P{^Is_Blk=phoenician}', "");
Expect(0, 67872, '\p{Is_Blk=phoenician}', "");
Expect(1, 67872, '\p{^Is_Blk=phoenician}', "");
Expect(1, 67872, '\P{Is_Blk=phoenician}', "");
Expect(0, 67872, '\P{^Is_Blk=phoenician}', "");
Expect(1, 67871, '\p{Is_Blk=	Phoenician}', "");
Expect(0, 67871, '\p{^Is_Blk=	Phoenician}', "");
Expect(0, 67871, '\P{Is_Blk=	Phoenician}', "");
Expect(1, 67871, '\P{^Is_Blk=	Phoenician}', "");
Expect(0, 67872, '\p{Is_Blk=	Phoenician}', "");
Expect(1, 67872, '\p{^Is_Blk=	Phoenician}', "");
Expect(1, 67872, '\P{Is_Blk=	Phoenician}', "");
Expect(0, 67872, '\P{^Is_Blk=	Phoenician}', "");
Error('\p{Block=:=_phonetic_Extensions}');
Error('\P{Block=:=_phonetic_Extensions}');
Expect(1, 7551, '\p{Block=phoneticextensions}', "");
Expect(0, 7551, '\p{^Block=phoneticextensions}', "");
Expect(0, 7551, '\P{Block=phoneticextensions}', "");
Expect(1, 7551, '\P{^Block=phoneticextensions}', "");
Expect(0, 7552, '\p{Block=phoneticextensions}', "");
Expect(1, 7552, '\p{^Block=phoneticextensions}', "");
Expect(1, 7552, '\P{Block=phoneticextensions}', "");
Expect(0, 7552, '\P{^Block=phoneticextensions}', "");
Expect(1, 7551, '\p{Block=_	phonetic_extensions}', "");
Expect(0, 7551, '\p{^Block=_	phonetic_extensions}', "");
Expect(0, 7551, '\P{Block=_	phonetic_extensions}', "");
Expect(1, 7551, '\P{^Block=_	phonetic_extensions}', "");
Expect(0, 7552, '\p{Block=_	phonetic_extensions}', "");
Expect(1, 7552, '\p{^Block=_	phonetic_extensions}', "");
Expect(1, 7552, '\P{Block=_	phonetic_extensions}', "");
Expect(0, 7552, '\P{^Block=_	phonetic_extensions}', "");
Error('\p{Blk=/a/Phonetic_Ext}');
Error('\P{Blk=/a/Phonetic_Ext}');
Expect(1, 7551, '\p{Blk=phoneticext}', "");
Expect(0, 7551, '\p{^Blk=phoneticext}', "");
Expect(0, 7551, '\P{Blk=phoneticext}', "");
Expect(1, 7551, '\P{^Blk=phoneticext}', "");
Expect(0, 7552, '\p{Blk=phoneticext}', "");
Expect(1, 7552, '\p{^Blk=phoneticext}', "");
Expect(1, 7552, '\P{Blk=phoneticext}', "");
Expect(0, 7552, '\P{^Blk=phoneticext}', "");
Expect(1, 7551, '\p{Blk= 	Phonetic_Ext}', "");
Expect(0, 7551, '\p{^Blk= 	Phonetic_Ext}', "");
Expect(0, 7551, '\P{Blk= 	Phonetic_Ext}', "");
Expect(1, 7551, '\P{^Blk= 	Phonetic_Ext}', "");
Expect(0, 7552, '\p{Blk= 	Phonetic_Ext}', "");
Expect(1, 7552, '\p{^Blk= 	Phonetic_Ext}', "");
Expect(1, 7552, '\P{Blk= 	Phonetic_Ext}', "");
Expect(0, 7552, '\P{^Blk= 	Phonetic_Ext}', "");
Error('\p{Is_Block=_:=PHONETIC_extensions}');
Error('\P{Is_Block=_:=PHONETIC_extensions}');
Expect(1, 7551, '\p{Is_Block=phoneticextensions}', "");
Expect(0, 7551, '\p{^Is_Block=phoneticextensions}', "");
Expect(0, 7551, '\P{Is_Block=phoneticextensions}', "");
Expect(1, 7551, '\P{^Is_Block=phoneticextensions}', "");
Expect(0, 7552, '\p{Is_Block=phoneticextensions}', "");
Expect(1, 7552, '\p{^Is_Block=phoneticextensions}', "");
Expect(1, 7552, '\P{Is_Block=phoneticextensions}', "");
Expect(0, 7552, '\P{^Is_Block=phoneticextensions}', "");
Expect(1, 7551, '\p{Is_Block=Phonetic_Extensions}', "");
Expect(0, 7551, '\p{^Is_Block=Phonetic_Extensions}', "");
Expect(0, 7551, '\P{Is_Block=Phonetic_Extensions}', "");
Expect(1, 7551, '\P{^Is_Block=Phonetic_Extensions}', "");
Expect(0, 7552, '\p{Is_Block=Phonetic_Extensions}', "");
Expect(1, 7552, '\p{^Is_Block=Phonetic_Extensions}', "");
Expect(1, 7552, '\P{Is_Block=Phonetic_Extensions}', "");
Expect(0, 7552, '\P{^Is_Block=Phonetic_Extensions}', "");
Error('\p{Is_Blk=_ PHONETIC_Ext/a/}');
Error('\P{Is_Blk=_ PHONETIC_Ext/a/}');
Expect(1, 7551, '\p{Is_Blk=phoneticext}', "");
Expect(0, 7551, '\p{^Is_Blk=phoneticext}', "");
Expect(0, 7551, '\P{Is_Blk=phoneticext}', "");
Expect(1, 7551, '\P{^Is_Blk=phoneticext}', "");
Expect(0, 7552, '\p{Is_Blk=phoneticext}', "");
Expect(1, 7552, '\p{^Is_Blk=phoneticext}', "");
Expect(1, 7552, '\P{Is_Blk=phoneticext}', "");
Expect(0, 7552, '\P{^Is_Blk=phoneticext}', "");
Expect(1, 7551, '\p{Is_Blk=	PHONETIC_EXT}', "");
Expect(0, 7551, '\p{^Is_Blk=	PHONETIC_EXT}', "");
Expect(0, 7551, '\P{Is_Blk=	PHONETIC_EXT}', "");
Expect(1, 7551, '\P{^Is_Blk=	PHONETIC_EXT}', "");
Expect(0, 7552, '\p{Is_Blk=	PHONETIC_EXT}', "");
Expect(1, 7552, '\p{^Is_Blk=	PHONETIC_EXT}', "");
Expect(1, 7552, '\P{Is_Blk=	PHONETIC_EXT}', "");
Expect(0, 7552, '\P{^Is_Blk=	PHONETIC_EXT}', "");
Error('\p{Block=:=	_PHONETIC_Extensions_SUPPLEMENT}');
Error('\P{Block=:=	_PHONETIC_Extensions_SUPPLEMENT}');
Expect(1, 7615, '\p{Block=phoneticextensionssupplement}', "");
Expect(0, 7615, '\p{^Block=phoneticextensionssupplement}', "");
Expect(0, 7615, '\P{Block=phoneticextensionssupplement}', "");
Expect(1, 7615, '\P{^Block=phoneticextensionssupplement}', "");
Expect(0, 7616, '\p{Block=phoneticextensionssupplement}', "");
Expect(1, 7616, '\p{^Block=phoneticextensionssupplement}', "");
Expect(1, 7616, '\P{Block=phoneticextensionssupplement}', "");
Expect(0, 7616, '\P{^Block=phoneticextensionssupplement}', "");
Expect(1, 7615, '\p{Block=_-Phonetic_EXTENSIONS_Supplement}', "");
Expect(0, 7615, '\p{^Block=_-Phonetic_EXTENSIONS_Supplement}', "");
Expect(0, 7615, '\P{Block=_-Phonetic_EXTENSIONS_Supplement}', "");
Expect(1, 7615, '\P{^Block=_-Phonetic_EXTENSIONS_Supplement}', "");
Expect(0, 7616, '\p{Block=_-Phonetic_EXTENSIONS_Supplement}', "");
Expect(1, 7616, '\p{^Block=_-Phonetic_EXTENSIONS_Supplement}', "");
Expect(1, 7616, '\P{Block=_-Phonetic_EXTENSIONS_Supplement}', "");
Expect(0, 7616, '\P{^Block=_-Phonetic_EXTENSIONS_Supplement}', "");
Error('\p{Blk=_/a/phonetic_EXT_Sup}');
Error('\P{Blk=_/a/phonetic_EXT_Sup}');
Expect(1, 7615, '\p{Blk=phoneticextsup}', "");
Expect(0, 7615, '\p{^Blk=phoneticextsup}', "");
Expect(0, 7615, '\P{Blk=phoneticextsup}', "");
Expect(1, 7615, '\P{^Blk=phoneticextsup}', "");
Expect(0, 7616, '\p{Blk=phoneticextsup}', "");
Expect(1, 7616, '\p{^Blk=phoneticextsup}', "");
Expect(1, 7616, '\P{Blk=phoneticextsup}', "");
Expect(0, 7616, '\P{^Blk=phoneticextsup}', "");
Expect(1, 7615, '\p{Blk:  -Phonetic_Ext_sup}', "");
Expect(0, 7615, '\p{^Blk:  -Phonetic_Ext_sup}', "");
Expect(0, 7615, '\P{Blk:  -Phonetic_Ext_sup}', "");
Expect(1, 7615, '\P{^Blk:  -Phonetic_Ext_sup}', "");
Expect(0, 7616, '\p{Blk:  -Phonetic_Ext_sup}', "");
Expect(1, 7616, '\p{^Blk:  -Phonetic_Ext_sup}', "");
Expect(1, 7616, '\P{Blk:  -Phonetic_Ext_sup}', "");
Expect(0, 7616, '\P{^Blk:  -Phonetic_Ext_sup}', "");
Error('\p{Is_Block=/a/ -phonetic_EXTENSIONS_Supplement}');
Error('\P{Is_Block=/a/ -phonetic_EXTENSIONS_Supplement}');
Expect(1, 7615, '\p{Is_Block=phoneticextensionssupplement}', "");
Expect(0, 7615, '\p{^Is_Block=phoneticextensionssupplement}', "");
Expect(0, 7615, '\P{Is_Block=phoneticextensionssupplement}', "");
Expect(1, 7615, '\P{^Is_Block=phoneticextensionssupplement}', "");
Expect(0, 7616, '\p{Is_Block=phoneticextensionssupplement}', "");
Expect(1, 7616, '\p{^Is_Block=phoneticextensionssupplement}', "");
Expect(1, 7616, '\P{Is_Block=phoneticextensionssupplement}', "");
Expect(0, 7616, '\P{^Is_Block=phoneticextensionssupplement}', "");
Expect(1, 7615, '\p{Is_Block=-	PHONETIC_extensions_SUPPLEMENT}', "");
Expect(0, 7615, '\p{^Is_Block=-	PHONETIC_extensions_SUPPLEMENT}', "");
Expect(0, 7615, '\P{Is_Block=-	PHONETIC_extensions_SUPPLEMENT}', "");
Expect(1, 7615, '\P{^Is_Block=-	PHONETIC_extensions_SUPPLEMENT}', "");
Expect(0, 7616, '\p{Is_Block=-	PHONETIC_extensions_SUPPLEMENT}', "");
Expect(1, 7616, '\p{^Is_Block=-	PHONETIC_extensions_SUPPLEMENT}', "");
Expect(1, 7616, '\P{Is_Block=-	PHONETIC_extensions_SUPPLEMENT}', "");
Expect(0, 7616, '\P{^Is_Block=-	PHONETIC_extensions_SUPPLEMENT}', "");
Error('\p{Is_Blk:	 :=PHONETIC_EXT_Sup}');
Error('\P{Is_Blk:	 :=PHONETIC_EXT_Sup}');
Expect(1, 7615, '\p{Is_Blk=phoneticextsup}', "");
Expect(0, 7615, '\p{^Is_Blk=phoneticextsup}', "");
Expect(0, 7615, '\P{Is_Blk=phoneticextsup}', "");
Expect(1, 7615, '\P{^Is_Blk=phoneticextsup}', "");
Expect(0, 7616, '\p{Is_Blk=phoneticextsup}', "");
Expect(1, 7616, '\p{^Is_Blk=phoneticextsup}', "");
Expect(1, 7616, '\P{Is_Blk=phoneticextsup}', "");
Expect(0, 7616, '\P{^Is_Blk=phoneticextsup}', "");
Expect(1, 7615, '\p{Is_Blk=-	PHONETIC_EXT_sup}', "");
Expect(0, 7615, '\p{^Is_Blk=-	PHONETIC_EXT_sup}', "");
Expect(0, 7615, '\P{Is_Blk=-	PHONETIC_EXT_sup}', "");
Expect(1, 7615, '\P{^Is_Blk=-	PHONETIC_EXT_sup}', "");
Expect(0, 7616, '\p{Is_Blk=-	PHONETIC_EXT_sup}', "");
Expect(1, 7616, '\p{^Is_Blk=-	PHONETIC_EXT_sup}', "");
Expect(1, 7616, '\P{Is_Blk=-	PHONETIC_EXT_sup}', "");
Expect(0, 7616, '\P{^Is_Blk=-	PHONETIC_EXT_sup}', "");
Error('\p{Block=_ playing_CARDS:=}');
Error('\P{Block=_ playing_CARDS:=}');
Expect(1, 127231, '\p{Block=playingcards}', "");
Expect(0, 127231, '\p{^Block=playingcards}', "");
Expect(0, 127231, '\P{Block=playingcards}', "");
Expect(1, 127231, '\P{^Block=playingcards}', "");
Expect(0, 127232, '\p{Block=playingcards}', "");
Expect(1, 127232, '\p{^Block=playingcards}', "");
Expect(1, 127232, '\P{Block=playingcards}', "");
Expect(0, 127232, '\P{^Block=playingcards}', "");
Expect(1, 127231, '\p{Block=	-PLAYING_Cards}', "");
Expect(0, 127231, '\p{^Block=	-PLAYING_Cards}', "");
Expect(0, 127231, '\P{Block=	-PLAYING_Cards}', "");
Expect(1, 127231, '\P{^Block=	-PLAYING_Cards}', "");
Expect(0, 127232, '\p{Block=	-PLAYING_Cards}', "");
Expect(1, 127232, '\p{^Block=	-PLAYING_Cards}', "");
Expect(1, 127232, '\P{Block=	-PLAYING_Cards}', "");
Expect(0, 127232, '\P{^Block=	-PLAYING_Cards}', "");
Error('\p{Blk=	-Playing_Cards:=}');
Error('\P{Blk=	-Playing_Cards:=}');
Expect(1, 127231, '\p{Blk:playingcards}', "");
Expect(0, 127231, '\p{^Blk:playingcards}', "");
Expect(0, 127231, '\P{Blk:playingcards}', "");
Expect(1, 127231, '\P{^Blk:playingcards}', "");
Expect(0, 127232, '\p{Blk:playingcards}', "");
Expect(1, 127232, '\p{^Blk:playingcards}', "");
Expect(1, 127232, '\P{Blk:playingcards}', "");
Expect(0, 127232, '\P{^Blk:playingcards}', "");
Expect(1, 127231, '\p{Blk:	 PLAYING_CARDS}', "");
Expect(0, 127231, '\p{^Blk:	 PLAYING_CARDS}', "");
Expect(0, 127231, '\P{Blk:	 PLAYING_CARDS}', "");
Expect(1, 127231, '\P{^Blk:	 PLAYING_CARDS}', "");
Expect(0, 127232, '\p{Blk:	 PLAYING_CARDS}', "");
Expect(1, 127232, '\p{^Blk:	 PLAYING_CARDS}', "");
Expect(1, 127232, '\P{Blk:	 PLAYING_CARDS}', "");
Expect(0, 127232, '\P{^Blk:	 PLAYING_CARDS}', "");
Error('\p{Is_Block=	 PLAYING_CARDS:=}');
Error('\P{Is_Block=	 PLAYING_CARDS:=}');
Expect(1, 127231, '\p{Is_Block=playingcards}', "");
Expect(0, 127231, '\p{^Is_Block=playingcards}', "");
Expect(0, 127231, '\P{Is_Block=playingcards}', "");
Expect(1, 127231, '\P{^Is_Block=playingcards}', "");
Expect(0, 127232, '\p{Is_Block=playingcards}', "");
Expect(1, 127232, '\p{^Is_Block=playingcards}', "");
Expect(1, 127232, '\P{Is_Block=playingcards}', "");
Expect(0, 127232, '\P{^Is_Block=playingcards}', "");
Expect(1, 127231, '\p{Is_Block= 	playing_Cards}', "");
Expect(0, 127231, '\p{^Is_Block= 	playing_Cards}', "");
Expect(0, 127231, '\P{Is_Block= 	playing_Cards}', "");
Expect(1, 127231, '\P{^Is_Block= 	playing_Cards}', "");
Expect(0, 127232, '\p{Is_Block= 	playing_Cards}', "");
Expect(1, 127232, '\p{^Is_Block= 	playing_Cards}', "");
Expect(1, 127232, '\P{Is_Block= 	playing_Cards}', "");
Expect(0, 127232, '\P{^Is_Block= 	playing_Cards}', "");
Error('\p{Is_Blk= :=playing_Cards}');
Error('\P{Is_Blk= :=playing_Cards}');
Expect(1, 127231, '\p{Is_Blk=playingcards}', "");
Expect(0, 127231, '\p{^Is_Blk=playingcards}', "");
Expect(0, 127231, '\P{Is_Blk=playingcards}', "");
Expect(1, 127231, '\P{^Is_Blk=playingcards}', "");
Expect(0, 127232, '\p{Is_Blk=playingcards}', "");
Expect(1, 127232, '\p{^Is_Blk=playingcards}', "");
Expect(1, 127232, '\P{Is_Blk=playingcards}', "");
Expect(0, 127232, '\P{^Is_Blk=playingcards}', "");
Expect(1, 127231, '\p{Is_Blk=-_PLAYING_CARDS}', "");
Expect(0, 127231, '\p{^Is_Blk=-_PLAYING_CARDS}', "");
Expect(0, 127231, '\P{Is_Blk=-_PLAYING_CARDS}', "");
Expect(1, 127231, '\P{^Is_Blk=-_PLAYING_CARDS}', "");
Expect(0, 127232, '\p{Is_Blk=-_PLAYING_CARDS}', "");
Expect(1, 127232, '\p{^Is_Blk=-_PLAYING_CARDS}', "");
Expect(1, 127232, '\P{Is_Blk=-_PLAYING_CARDS}', "");
Expect(0, 127232, '\P{^Is_Blk=-_PLAYING_CARDS}', "");
Error('\p{Block=_/a/general_Punctuation}');
Error('\P{Block=_/a/general_Punctuation}');
Expect(1, 8303, '\p{Block=generalpunctuation}', "");
Expect(0, 8303, '\p{^Block=generalpunctuation}', "");
Expect(0, 8303, '\P{Block=generalpunctuation}', "");
Expect(1, 8303, '\P{^Block=generalpunctuation}', "");
Expect(0, 8304, '\p{Block=generalpunctuation}', "");
Expect(1, 8304, '\p{^Block=generalpunctuation}', "");
Expect(1, 8304, '\P{Block=generalpunctuation}', "");
Expect(0, 8304, '\P{^Block=generalpunctuation}', "");
Expect(1, 8303, '\p{Block=	_General_Punctuation}', "");
Expect(0, 8303, '\p{^Block=	_General_Punctuation}', "");
Expect(0, 8303, '\P{Block=	_General_Punctuation}', "");
Expect(1, 8303, '\P{^Block=	_General_Punctuation}', "");
Expect(0, 8304, '\p{Block=	_General_Punctuation}', "");
Expect(1, 8304, '\p{^Block=	_General_Punctuation}', "");
Expect(1, 8304, '\P{Block=	_General_Punctuation}', "");
Expect(0, 8304, '\P{^Block=	_General_Punctuation}', "");
Error('\p{Blk=- Punctuation/a/}');
Error('\P{Blk=- Punctuation/a/}');
Expect(1, 8303, '\p{Blk=punctuation}', "");
Expect(0, 8303, '\p{^Blk=punctuation}', "");
Expect(0, 8303, '\P{Blk=punctuation}', "");
Expect(1, 8303, '\P{^Blk=punctuation}', "");
Expect(0, 8304, '\p{Blk=punctuation}', "");
Expect(1, 8304, '\p{^Blk=punctuation}', "");
Expect(1, 8304, '\P{Blk=punctuation}', "");
Expect(0, 8304, '\P{^Blk=punctuation}', "");
Expect(1, 8303, '\p{Blk=-_Punctuation}', "");
Expect(0, 8303, '\p{^Blk=-_Punctuation}', "");
Expect(0, 8303, '\P{Blk=-_Punctuation}', "");
Expect(1, 8303, '\P{^Blk=-_Punctuation}', "");
Expect(0, 8304, '\p{Blk=-_Punctuation}', "");
Expect(1, 8304, '\p{^Blk=-_Punctuation}', "");
Expect(1, 8304, '\P{Blk=-_Punctuation}', "");
Expect(0, 8304, '\P{^Blk=-_Punctuation}', "");
Error('\p{Is_Block:			GENERAL_Punctuation/a/}');
Error('\P{Is_Block:			GENERAL_Punctuation/a/}');
Expect(1, 8303, '\p{Is_Block=generalpunctuation}', "");
Expect(0, 8303, '\p{^Is_Block=generalpunctuation}', "");
Expect(0, 8303, '\P{Is_Block=generalpunctuation}', "");
Expect(1, 8303, '\P{^Is_Block=generalpunctuation}', "");
Expect(0, 8304, '\p{Is_Block=generalpunctuation}', "");
Expect(1, 8304, '\p{^Is_Block=generalpunctuation}', "");
Expect(1, 8304, '\P{Is_Block=generalpunctuation}', "");
Expect(0, 8304, '\P{^Is_Block=generalpunctuation}', "");
Expect(1, 8303, '\p{Is_Block=-general_Punctuation}', "");
Expect(0, 8303, '\p{^Is_Block=-general_Punctuation}', "");
Expect(0, 8303, '\P{Is_Block=-general_Punctuation}', "");
Expect(1, 8303, '\P{^Is_Block=-general_Punctuation}', "");
Expect(0, 8304, '\p{Is_Block=-general_Punctuation}', "");
Expect(1, 8304, '\p{^Is_Block=-general_Punctuation}', "");
Expect(1, 8304, '\P{Is_Block=-general_Punctuation}', "");
Expect(0, 8304, '\P{^Is_Block=-general_Punctuation}', "");
Error('\p{Is_Blk:_ PUNCTUATION/a/}');
Error('\P{Is_Blk:_ PUNCTUATION/a/}');
Expect(1, 8303, '\p{Is_Blk=punctuation}', "");
Expect(0, 8303, '\p{^Is_Blk=punctuation}', "");
Expect(0, 8303, '\P{Is_Blk=punctuation}', "");
Expect(1, 8303, '\P{^Is_Blk=punctuation}', "");
Expect(0, 8304, '\p{Is_Blk=punctuation}', "");
Expect(1, 8304, '\p{^Is_Blk=punctuation}', "");
Expect(1, 8304, '\P{Is_Blk=punctuation}', "");
Expect(0, 8304, '\P{^Is_Blk=punctuation}', "");
Expect(1, 8303, '\p{Is_Blk=_Punctuation}', "");
Expect(0, 8303, '\p{^Is_Blk=_Punctuation}', "");
Expect(0, 8303, '\P{Is_Blk=_Punctuation}', "");
Expect(1, 8303, '\P{^Is_Blk=_Punctuation}', "");
Expect(0, 8304, '\p{Is_Blk=_Punctuation}', "");
Expect(1, 8304, '\p{^Is_Blk=_Punctuation}', "");
Expect(1, 8304, '\P{Is_Blk=_Punctuation}', "");
Expect(0, 8304, '\P{^Is_Blk=_Punctuation}', "");
Error('\p{Block=REJANG:=}');
Error('\P{Block=REJANG:=}');
Expect(1, 43359, '\p{Block=rejang}', "");
Expect(0, 43359, '\p{^Block=rejang}', "");
Expect(0, 43359, '\P{Block=rejang}', "");
Expect(1, 43359, '\P{^Block=rejang}', "");
Expect(0, 43360, '\p{Block=rejang}', "");
Expect(1, 43360, '\p{^Block=rejang}', "");
Expect(1, 43360, '\P{Block=rejang}', "");
Expect(0, 43360, '\P{^Block=rejang}', "");
Expect(1, 43359, '\p{Block=	Rejang}', "");
Expect(0, 43359, '\p{^Block=	Rejang}', "");
Expect(0, 43359, '\P{Block=	Rejang}', "");
Expect(1, 43359, '\P{^Block=	Rejang}', "");
Expect(0, 43360, '\p{Block=	Rejang}', "");
Expect(1, 43360, '\p{^Block=	Rejang}', "");
Expect(1, 43360, '\P{Block=	Rejang}', "");
Expect(0, 43360, '\P{^Block=	Rejang}', "");
Error('\p{Blk= :=rejang}');
Error('\P{Blk= :=rejang}');
Expect(1, 43359, '\p{Blk:	rejang}', "");
Expect(0, 43359, '\p{^Blk:	rejang}', "");
Expect(0, 43359, '\P{Blk:	rejang}', "");
Expect(1, 43359, '\P{^Blk:	rejang}', "");
Expect(0, 43360, '\p{Blk:	rejang}', "");
Expect(1, 43360, '\p{^Blk:	rejang}', "");
Expect(1, 43360, '\P{Blk:	rejang}', "");
Expect(0, 43360, '\P{^Blk:	rejang}', "");
Expect(1, 43359, '\p{Blk:  	REJANG}', "");
Expect(0, 43359, '\p{^Blk:  	REJANG}', "");
Expect(0, 43359, '\P{Blk:  	REJANG}', "");
Expect(1, 43359, '\P{^Blk:  	REJANG}', "");
Expect(0, 43360, '\p{Blk:  	REJANG}', "");
Expect(1, 43360, '\p{^Blk:  	REJANG}', "");
Expect(1, 43360, '\P{Blk:  	REJANG}', "");
Expect(0, 43360, '\P{^Blk:  	REJANG}', "");
Error('\p{Is_Block=_-Rejang/a/}');
Error('\P{Is_Block=_-Rejang/a/}');
Expect(1, 43359, '\p{Is_Block=rejang}', "");
Expect(0, 43359, '\p{^Is_Block=rejang}', "");
Expect(0, 43359, '\P{Is_Block=rejang}', "");
Expect(1, 43359, '\P{^Is_Block=rejang}', "");
Expect(0, 43360, '\p{Is_Block=rejang}', "");
Expect(1, 43360, '\p{^Is_Block=rejang}', "");
Expect(1, 43360, '\P{Is_Block=rejang}', "");
Expect(0, 43360, '\P{^Is_Block=rejang}', "");
Expect(1, 43359, '\p{Is_Block=_	REJANG}', "");
Expect(0, 43359, '\p{^Is_Block=_	REJANG}', "");
Expect(0, 43359, '\P{Is_Block=_	REJANG}', "");
Expect(1, 43359, '\P{^Is_Block=_	REJANG}', "");
Expect(0, 43360, '\p{Is_Block=_	REJANG}', "");
Expect(1, 43360, '\p{^Is_Block=_	REJANG}', "");
Expect(1, 43360, '\P{Is_Block=_	REJANG}', "");
Expect(0, 43360, '\P{^Is_Block=_	REJANG}', "");
Error('\p{Is_Blk=:= Rejang}');
Error('\P{Is_Blk=:= Rejang}');
Expect(1, 43359, '\p{Is_Blk:   rejang}', "");
Expect(0, 43359, '\p{^Is_Blk:   rejang}', "");
Expect(0, 43359, '\P{Is_Blk:   rejang}', "");
Expect(1, 43359, '\P{^Is_Blk:   rejang}', "");
Expect(0, 43360, '\p{Is_Blk:   rejang}', "");
Expect(1, 43360, '\p{^Is_Blk:   rejang}', "");
Expect(1, 43360, '\P{Is_Blk:   rejang}', "");
Expect(0, 43360, '\P{^Is_Blk:   rejang}', "");
Expect(1, 43359, '\p{Is_Blk=  Rejang}', "");
Expect(0, 43359, '\p{^Is_Blk=  Rejang}', "");
Expect(0, 43359, '\P{Is_Blk=  Rejang}', "");
Expect(1, 43359, '\P{^Is_Blk=  Rejang}', "");
Expect(0, 43360, '\p{Is_Blk=  Rejang}', "");
Expect(1, 43360, '\p{^Is_Blk=  Rejang}', "");
Expect(1, 43360, '\P{Is_Blk=  Rejang}', "");
Expect(0, 43360, '\P{^Is_Blk=  Rejang}', "");
Error('\p{Block=:= _RUMI_Numeral_symbols}');
Error('\P{Block=:= _RUMI_Numeral_symbols}');
Expect(1, 69247, '\p{Block=ruminumeralsymbols}', "");
Expect(0, 69247, '\p{^Block=ruminumeralsymbols}', "");
Expect(0, 69247, '\P{Block=ruminumeralsymbols}', "");
Expect(1, 69247, '\P{^Block=ruminumeralsymbols}', "");
Expect(0, 69248, '\p{Block=ruminumeralsymbols}', "");
Expect(1, 69248, '\p{^Block=ruminumeralsymbols}', "");
Expect(1, 69248, '\P{Block=ruminumeralsymbols}', "");
Expect(0, 69248, '\P{^Block=ruminumeralsymbols}', "");
Expect(1, 69247, '\p{Block=	_RUMI_NUMERAL_symbols}', "");
Expect(0, 69247, '\p{^Block=	_RUMI_NUMERAL_symbols}', "");
Expect(0, 69247, '\P{Block=	_RUMI_NUMERAL_symbols}', "");
Expect(1, 69247, '\P{^Block=	_RUMI_NUMERAL_symbols}', "");
Expect(0, 69248, '\p{Block=	_RUMI_NUMERAL_symbols}', "");
Expect(1, 69248, '\p{^Block=	_RUMI_NUMERAL_symbols}', "");
Expect(1, 69248, '\P{Block=	_RUMI_NUMERAL_symbols}', "");
Expect(0, 69248, '\P{^Block=	_RUMI_NUMERAL_symbols}', "");
Error('\p{Blk=_	RUMI/a/}');
Error('\P{Blk=_	RUMI/a/}');
Expect(1, 69247, '\p{Blk=rumi}', "");
Expect(0, 69247, '\p{^Blk=rumi}', "");
Expect(0, 69247, '\P{Blk=rumi}', "");
Expect(1, 69247, '\P{^Blk=rumi}', "");
Expect(0, 69248, '\p{Blk=rumi}', "");
Expect(1, 69248, '\p{^Blk=rumi}', "");
Expect(1, 69248, '\P{Blk=rumi}', "");
Expect(0, 69248, '\P{^Blk=rumi}', "");
Expect(1, 69247, '\p{Blk=-_Rumi}', "");
Expect(0, 69247, '\p{^Blk=-_Rumi}', "");
Expect(0, 69247, '\P{Blk=-_Rumi}', "");
Expect(1, 69247, '\P{^Blk=-_Rumi}', "");
Expect(0, 69248, '\p{Blk=-_Rumi}', "");
Expect(1, 69248, '\p{^Blk=-_Rumi}', "");
Expect(1, 69248, '\P{Blk=-_Rumi}', "");
Expect(0, 69248, '\P{^Blk=-_Rumi}', "");
Error('\p{Is_Block=/a/	_Rumi_Numeral_SYMBOLS}');
Error('\P{Is_Block=/a/	_Rumi_Numeral_SYMBOLS}');
Expect(1, 69247, '\p{Is_Block=ruminumeralsymbols}', "");
Expect(0, 69247, '\p{^Is_Block=ruminumeralsymbols}', "");
Expect(0, 69247, '\P{Is_Block=ruminumeralsymbols}', "");
Expect(1, 69247, '\P{^Is_Block=ruminumeralsymbols}', "");
Expect(0, 69248, '\p{Is_Block=ruminumeralsymbols}', "");
Expect(1, 69248, '\p{^Is_Block=ruminumeralsymbols}', "");
Expect(1, 69248, '\P{Is_Block=ruminumeralsymbols}', "");
Expect(0, 69248, '\P{^Is_Block=ruminumeralsymbols}', "");
Expect(1, 69247, '\p{Is_Block=_Rumi_Numeral_SYMBOLS}', "");
Expect(0, 69247, '\p{^Is_Block=_Rumi_Numeral_SYMBOLS}', "");
Expect(0, 69247, '\P{Is_Block=_Rumi_Numeral_SYMBOLS}', "");
Expect(1, 69247, '\P{^Is_Block=_Rumi_Numeral_SYMBOLS}', "");
Expect(0, 69248, '\p{Is_Block=_Rumi_Numeral_SYMBOLS}', "");
Expect(1, 69248, '\p{^Is_Block=_Rumi_Numeral_SYMBOLS}', "");
Expect(1, 69248, '\P{Is_Block=_Rumi_Numeral_SYMBOLS}', "");
Expect(0, 69248, '\P{^Is_Block=_Rumi_Numeral_SYMBOLS}', "");
Error('\p{Is_Blk: /a/ Rumi}');
Error('\P{Is_Blk: /a/ Rumi}');
Expect(1, 69247, '\p{Is_Blk=rumi}', "");
Expect(0, 69247, '\p{^Is_Blk=rumi}', "");
Expect(0, 69247, '\P{Is_Blk=rumi}', "");
Expect(1, 69247, '\P{^Is_Blk=rumi}', "");
Expect(0, 69248, '\p{Is_Blk=rumi}', "");
Expect(1, 69248, '\p{^Is_Blk=rumi}', "");
Expect(1, 69248, '\P{Is_Blk=rumi}', "");
Expect(0, 69248, '\P{^Is_Blk=rumi}', "");
Expect(1, 69247, '\p{Is_Blk=- RUMI}', "");
Expect(0, 69247, '\p{^Is_Blk=- RUMI}', "");
Expect(0, 69247, '\P{Is_Blk=- RUMI}', "");
Expect(1, 69247, '\P{^Is_Blk=- RUMI}', "");
Expect(0, 69248, '\p{Is_Blk=- RUMI}', "");
Expect(1, 69248, '\p{^Is_Blk=- RUMI}', "");
Expect(1, 69248, '\P{Is_Blk=- RUMI}', "");
Expect(0, 69248, '\P{^Is_Blk=- RUMI}', "");
Error('\p{Block=/a/_runic}');
Error('\P{Block=/a/_runic}');
Expect(1, 5887, '\p{Block=runic}', "");
Expect(0, 5887, '\p{^Block=runic}', "");
Expect(0, 5887, '\P{Block=runic}', "");
Expect(1, 5887, '\P{^Block=runic}', "");
Expect(0, 5888, '\p{Block=runic}', "");
Expect(1, 5888, '\p{^Block=runic}', "");
Expect(1, 5888, '\P{Block=runic}', "");
Expect(0, 5888, '\P{^Block=runic}', "");
Expect(1, 5887, '\p{Block=-RUNIC}', "");
Expect(0, 5887, '\p{^Block=-RUNIC}', "");
Expect(0, 5887, '\P{Block=-RUNIC}', "");
Expect(1, 5887, '\P{^Block=-RUNIC}', "");
Expect(0, 5888, '\p{Block=-RUNIC}', "");
Expect(1, 5888, '\p{^Block=-RUNIC}', "");
Expect(1, 5888, '\P{Block=-RUNIC}', "");
Expect(0, 5888, '\P{^Block=-RUNIC}', "");
Error('\p{Blk:   /a/ _RUNIC}');
Error('\P{Blk:   /a/ _RUNIC}');
Expect(1, 5887, '\p{Blk=runic}', "");
Expect(0, 5887, '\p{^Blk=runic}', "");
Expect(0, 5887, '\P{Blk=runic}', "");
Expect(1, 5887, '\P{^Blk=runic}', "");
Expect(0, 5888, '\p{Blk=runic}', "");
Expect(1, 5888, '\p{^Blk=runic}', "");
Expect(1, 5888, '\P{Blk=runic}', "");
Expect(0, 5888, '\P{^Blk=runic}', "");
Expect(1, 5887, '\p{Blk=-_Runic}', "");
Expect(0, 5887, '\p{^Blk=-_Runic}', "");
Expect(0, 5887, '\P{Blk=-_Runic}', "");
Expect(1, 5887, '\P{^Blk=-_Runic}', "");
Expect(0, 5888, '\p{Blk=-_Runic}', "");
Expect(1, 5888, '\p{^Blk=-_Runic}', "");
Expect(1, 5888, '\P{Blk=-_Runic}', "");
Expect(0, 5888, '\P{^Blk=-_Runic}', "");
Error('\p{Is_Block=-:=RUNIC}');
Error('\P{Is_Block=-:=RUNIC}');
Expect(1, 5887, '\p{Is_Block=runic}', "");
Expect(0, 5887, '\p{^Is_Block=runic}', "");
Expect(0, 5887, '\P{Is_Block=runic}', "");
Expect(1, 5887, '\P{^Is_Block=runic}', "");
Expect(0, 5888, '\p{Is_Block=runic}', "");
Expect(1, 5888, '\p{^Is_Block=runic}', "");
Expect(1, 5888, '\P{Is_Block=runic}', "");
Expect(0, 5888, '\P{^Is_Block=runic}', "");
Expect(1, 5887, '\p{Is_Block= 	Runic}', "");
Expect(0, 5887, '\p{^Is_Block= 	Runic}', "");
Expect(0, 5887, '\P{Is_Block= 	Runic}', "");
Expect(1, 5887, '\P{^Is_Block= 	Runic}', "");
Expect(0, 5888, '\p{Is_Block= 	Runic}', "");
Expect(1, 5888, '\p{^Is_Block= 	Runic}', "");
Expect(1, 5888, '\P{Is_Block= 	Runic}', "");
Expect(0, 5888, '\P{^Is_Block= 	Runic}', "");
Error('\p{Is_Blk=	 runic:=}');
Error('\P{Is_Blk=	 runic:=}');
Expect(1, 5887, '\p{Is_Blk=runic}', "");
Expect(0, 5887, '\p{^Is_Blk=runic}', "");
Expect(0, 5887, '\P{Is_Blk=runic}', "");
Expect(1, 5887, '\P{^Is_Blk=runic}', "");
Expect(0, 5888, '\p{Is_Blk=runic}', "");
Expect(1, 5888, '\p{^Is_Blk=runic}', "");
Expect(1, 5888, '\P{Is_Blk=runic}', "");
Expect(0, 5888, '\P{^Is_Blk=runic}', "");
Expect(1, 5887, '\p{Is_Blk=	 runic}', "");
Expect(0, 5887, '\p{^Is_Blk=	 runic}', "");
Expect(0, 5887, '\P{Is_Blk=	 runic}', "");
Expect(1, 5887, '\P{^Is_Blk=	 runic}', "");
Expect(0, 5888, '\p{Is_Blk=	 runic}', "");
Expect(1, 5888, '\p{^Is_Blk=	 runic}', "");
Expect(1, 5888, '\P{Is_Blk=	 runic}', "");
Expect(0, 5888, '\P{^Is_Blk=	 runic}', "");
Error('\p{Block=/a/	SAMARITAN}');
Error('\P{Block=/a/	SAMARITAN}');
Expect(1, 2111, '\p{Block=samaritan}', "");
Expect(0, 2111, '\p{^Block=samaritan}', "");
Expect(0, 2111, '\P{Block=samaritan}', "");
Expect(1, 2111, '\P{^Block=samaritan}', "");
Expect(0, 2112, '\p{Block=samaritan}', "");
Expect(1, 2112, '\p{^Block=samaritan}', "");
Expect(1, 2112, '\P{Block=samaritan}', "");
Expect(0, 2112, '\P{^Block=samaritan}', "");
Expect(1, 2111, '\p{Block=_	samaritan}', "");
Expect(0, 2111, '\p{^Block=_	samaritan}', "");
Expect(0, 2111, '\P{Block=_	samaritan}', "");
Expect(1, 2111, '\P{^Block=_	samaritan}', "");
Expect(0, 2112, '\p{Block=_	samaritan}', "");
Expect(1, 2112, '\p{^Block=_	samaritan}', "");
Expect(1, 2112, '\P{Block=_	samaritan}', "");
Expect(0, 2112, '\P{^Block=_	samaritan}', "");
Error('\p{Blk=-SAMARITAN:=}');
Error('\P{Blk=-SAMARITAN:=}');
Expect(1, 2111, '\p{Blk=samaritan}', "");
Expect(0, 2111, '\p{^Blk=samaritan}', "");
Expect(0, 2111, '\P{Blk=samaritan}', "");
Expect(1, 2111, '\P{^Blk=samaritan}', "");
Expect(0, 2112, '\p{Blk=samaritan}', "");
Expect(1, 2112, '\p{^Blk=samaritan}', "");
Expect(1, 2112, '\P{Blk=samaritan}', "");
Expect(0, 2112, '\P{^Blk=samaritan}', "");
Expect(1, 2111, '\p{Blk=-_Samaritan}', "");
Expect(0, 2111, '\p{^Blk=-_Samaritan}', "");
Expect(0, 2111, '\P{Blk=-_Samaritan}', "");
Expect(1, 2111, '\P{^Blk=-_Samaritan}', "");
Expect(0, 2112, '\p{Blk=-_Samaritan}', "");
Expect(1, 2112, '\p{^Blk=-_Samaritan}', "");
Expect(1, 2112, '\P{Blk=-_Samaritan}', "");
Expect(0, 2112, '\P{^Blk=-_Samaritan}', "");
Error('\p{Is_Block=--Samaritan:=}');
Error('\P{Is_Block=--Samaritan:=}');
Expect(1, 2111, '\p{Is_Block=samaritan}', "");
Expect(0, 2111, '\p{^Is_Block=samaritan}', "");
Expect(0, 2111, '\P{Is_Block=samaritan}', "");
Expect(1, 2111, '\P{^Is_Block=samaritan}', "");
Expect(0, 2112, '\p{Is_Block=samaritan}', "");
Expect(1, 2112, '\p{^Is_Block=samaritan}', "");
Expect(1, 2112, '\P{Is_Block=samaritan}', "");
Expect(0, 2112, '\P{^Is_Block=samaritan}', "");
Expect(1, 2111, '\p{Is_Block=  Samaritan}', "");
Expect(0, 2111, '\p{^Is_Block=  Samaritan}', "");
Expect(0, 2111, '\P{Is_Block=  Samaritan}', "");
Expect(1, 2111, '\P{^Is_Block=  Samaritan}', "");
Expect(0, 2112, '\p{Is_Block=  Samaritan}', "");
Expect(1, 2112, '\p{^Is_Block=  Samaritan}', "");
Expect(1, 2112, '\P{Is_Block=  Samaritan}', "");
Expect(0, 2112, '\P{^Is_Block=  Samaritan}', "");
Error('\p{Is_Blk= 	SAMARITAN/a/}');
Error('\P{Is_Blk= 	SAMARITAN/a/}');
Expect(1, 2111, '\p{Is_Blk=samaritan}', "");
Expect(0, 2111, '\p{^Is_Blk=samaritan}', "");
Expect(0, 2111, '\P{Is_Blk=samaritan}', "");
Expect(1, 2111, '\P{^Is_Blk=samaritan}', "");
Expect(0, 2112, '\p{Is_Blk=samaritan}', "");
Expect(1, 2112, '\p{^Is_Blk=samaritan}', "");
Expect(1, 2112, '\P{Is_Blk=samaritan}', "");
Expect(0, 2112, '\P{^Is_Blk=samaritan}', "");
Expect(1, 2111, '\p{Is_Blk: samaritan}', "");
Expect(0, 2111, '\p{^Is_Blk: samaritan}', "");
Expect(0, 2111, '\P{Is_Blk: samaritan}', "");
Expect(1, 2111, '\P{^Is_Blk: samaritan}', "");
Expect(0, 2112, '\p{Is_Blk: samaritan}', "");
Expect(1, 2112, '\p{^Is_Blk: samaritan}', "");
Expect(1, 2112, '\P{Is_Blk: samaritan}', "");
Expect(0, 2112, '\P{^Is_Blk: samaritan}', "");
Error('\p{Block= :=saurashtra}');
Error('\P{Block= :=saurashtra}');
Expect(1, 43231, '\p{Block:saurashtra}', "");
Expect(0, 43231, '\p{^Block:saurashtra}', "");
Expect(0, 43231, '\P{Block:saurashtra}', "");
Expect(1, 43231, '\P{^Block:saurashtra}', "");
Expect(0, 43232, '\p{Block:saurashtra}', "");
Expect(1, 43232, '\p{^Block:saurashtra}', "");
Expect(1, 43232, '\P{Block:saurashtra}', "");
Expect(0, 43232, '\P{^Block:saurashtra}', "");
Expect(1, 43231, '\p{Block= Saurashtra}', "");
Expect(0, 43231, '\p{^Block= Saurashtra}', "");
Expect(0, 43231, '\P{Block= Saurashtra}', "");
Expect(1, 43231, '\P{^Block= Saurashtra}', "");
Expect(0, 43232, '\p{Block= Saurashtra}', "");
Expect(1, 43232, '\p{^Block= Saurashtra}', "");
Expect(1, 43232, '\P{Block= Saurashtra}', "");
Expect(0, 43232, '\P{^Block= Saurashtra}', "");
Error('\p{Blk=	_Saurashtra:=}');
Error('\P{Blk=	_Saurashtra:=}');
Expect(1, 43231, '\p{Blk=saurashtra}', "");
Expect(0, 43231, '\p{^Blk=saurashtra}', "");
Expect(0, 43231, '\P{Blk=saurashtra}', "");
Expect(1, 43231, '\P{^Blk=saurashtra}', "");
Expect(0, 43232, '\p{Blk=saurashtra}', "");
Expect(1, 43232, '\p{^Blk=saurashtra}', "");
Expect(1, 43232, '\P{Blk=saurashtra}', "");
Expect(0, 43232, '\P{^Blk=saurashtra}', "");
Expect(1, 43231, '\p{Blk=	-Saurashtra}', "");
Expect(0, 43231, '\p{^Blk=	-Saurashtra}', "");
Expect(0, 43231, '\P{Blk=	-Saurashtra}', "");
Expect(1, 43231, '\P{^Blk=	-Saurashtra}', "");
Expect(0, 43232, '\p{Blk=	-Saurashtra}', "");
Expect(1, 43232, '\p{^Blk=	-Saurashtra}', "");
Expect(1, 43232, '\P{Blk=	-Saurashtra}', "");
Expect(0, 43232, '\P{^Blk=	-Saurashtra}', "");
Error('\p{Is_Block= Saurashtra/a/}');
Error('\P{Is_Block= Saurashtra/a/}');
Expect(1, 43231, '\p{Is_Block=saurashtra}', "");
Expect(0, 43231, '\p{^Is_Block=saurashtra}', "");
Expect(0, 43231, '\P{Is_Block=saurashtra}', "");
Expect(1, 43231, '\P{^Is_Block=saurashtra}', "");
Expect(0, 43232, '\p{Is_Block=saurashtra}', "");
Expect(1, 43232, '\p{^Is_Block=saurashtra}', "");
Expect(1, 43232, '\P{Is_Block=saurashtra}', "");
Expect(0, 43232, '\P{^Is_Block=saurashtra}', "");
Expect(1, 43231, '\p{Is_Block=-_Saurashtra}', "");
Expect(0, 43231, '\p{^Is_Block=-_Saurashtra}', "");
Expect(0, 43231, '\P{Is_Block=-_Saurashtra}', "");
Expect(1, 43231, '\P{^Is_Block=-_Saurashtra}', "");
Expect(0, 43232, '\p{Is_Block=-_Saurashtra}', "");
Expect(1, 43232, '\p{^Is_Block=-_Saurashtra}', "");
Expect(1, 43232, '\P{Is_Block=-_Saurashtra}', "");
Expect(0, 43232, '\P{^Is_Block=-_Saurashtra}', "");
Error('\p{Is_Blk= 	Saurashtra:=}');
Error('\P{Is_Blk= 	Saurashtra:=}');
Expect(1, 43231, '\p{Is_Blk=saurashtra}', "");
Expect(0, 43231, '\p{^Is_Blk=saurashtra}', "");
Expect(0, 43231, '\P{Is_Blk=saurashtra}', "");
Expect(1, 43231, '\P{^Is_Blk=saurashtra}', "");
Expect(0, 43232, '\p{Is_Blk=saurashtra}', "");
Expect(1, 43232, '\p{^Is_Blk=saurashtra}', "");
Expect(1, 43232, '\P{Is_Blk=saurashtra}', "");
Expect(0, 43232, '\P{^Is_Blk=saurashtra}', "");
Expect(1, 43231, '\p{Is_Blk=SAURASHTRA}', "");
Expect(0, 43231, '\p{^Is_Blk=SAURASHTRA}', "");
Expect(0, 43231, '\P{Is_Blk=SAURASHTRA}', "");
Expect(1, 43231, '\P{^Is_Blk=SAURASHTRA}', "");
Expect(0, 43232, '\p{Is_Blk=SAURASHTRA}', "");
Expect(1, 43232, '\p{^Is_Blk=SAURASHTRA}', "");
Expect(1, 43232, '\P{Is_Blk=SAURASHTRA}', "");
Expect(0, 43232, '\P{^Is_Blk=SAURASHTRA}', "");
Error('\p{Block=:=Sharada}');
Error('\P{Block=:=Sharada}');
Expect(1, 70111, '\p{Block=sharada}', "");
Expect(0, 70111, '\p{^Block=sharada}', "");
Expect(0, 70111, '\P{Block=sharada}', "");
Expect(1, 70111, '\P{^Block=sharada}', "");
Expect(0, 70112, '\p{Block=sharada}', "");
Expect(1, 70112, '\p{^Block=sharada}', "");
Expect(1, 70112, '\P{Block=sharada}', "");
Expect(0, 70112, '\P{^Block=sharada}', "");
Expect(1, 70111, '\p{Block= 	SHARADA}', "");
Expect(0, 70111, '\p{^Block= 	SHARADA}', "");
Expect(0, 70111, '\P{Block= 	SHARADA}', "");
Expect(1, 70111, '\P{^Block= 	SHARADA}', "");
Expect(0, 70112, '\p{Block= 	SHARADA}', "");
Expect(1, 70112, '\p{^Block= 	SHARADA}', "");
Expect(1, 70112, '\P{Block= 	SHARADA}', "");
Expect(0, 70112, '\P{^Block= 	SHARADA}', "");
Error('\p{Blk=	:=Sharada}');
Error('\P{Blk=	:=Sharada}');
Expect(1, 70111, '\p{Blk=sharada}', "");
Expect(0, 70111, '\p{^Blk=sharada}', "");
Expect(0, 70111, '\P{Blk=sharada}', "");
Expect(1, 70111, '\P{^Blk=sharada}', "");
Expect(0, 70112, '\p{Blk=sharada}', "");
Expect(1, 70112, '\p{^Blk=sharada}', "");
Expect(1, 70112, '\P{Blk=sharada}', "");
Expect(0, 70112, '\P{^Blk=sharada}', "");
Expect(1, 70111, '\p{Blk:		 sharada}', "");
Expect(0, 70111, '\p{^Blk:		 sharada}', "");
Expect(0, 70111, '\P{Blk:		 sharada}', "");
Expect(1, 70111, '\P{^Blk:		 sharada}', "");
Expect(0, 70112, '\p{Blk:		 sharada}', "");
Expect(1, 70112, '\p{^Blk:		 sharada}', "");
Expect(1, 70112, '\P{Blk:		 sharada}', "");
Expect(0, 70112, '\P{^Blk:		 sharada}', "");
Error('\p{Is_Block:  -SHARADA/a/}');
Error('\P{Is_Block:  -SHARADA/a/}');
Expect(1, 70111, '\p{Is_Block=sharada}', "");
Expect(0, 70111, '\p{^Is_Block=sharada}', "");
Expect(0, 70111, '\P{Is_Block=sharada}', "");
Expect(1, 70111, '\P{^Is_Block=sharada}', "");
Expect(0, 70112, '\p{Is_Block=sharada}', "");
Expect(1, 70112, '\p{^Is_Block=sharada}', "");
Expect(1, 70112, '\P{Is_Block=sharada}', "");
Expect(0, 70112, '\P{^Is_Block=sharada}', "");
Expect(1, 70111, '\p{Is_Block= -SHARADA}', "");
Expect(0, 70111, '\p{^Is_Block= -SHARADA}', "");
Expect(0, 70111, '\P{Is_Block= -SHARADA}', "");
Expect(1, 70111, '\P{^Is_Block= -SHARADA}', "");
Expect(0, 70112, '\p{Is_Block= -SHARADA}', "");
Expect(1, 70112, '\p{^Is_Block= -SHARADA}', "");
Expect(1, 70112, '\P{Is_Block= -SHARADA}', "");
Expect(0, 70112, '\P{^Is_Block= -SHARADA}', "");
Error('\p{Is_Blk=/a/_SHARADA}');
Error('\P{Is_Blk=/a/_SHARADA}');
Expect(1, 70111, '\p{Is_Blk=sharada}', "");
Expect(0, 70111, '\p{^Is_Blk=sharada}', "");
Expect(0, 70111, '\P{Is_Blk=sharada}', "");
Expect(1, 70111, '\P{^Is_Blk=sharada}', "");
Expect(0, 70112, '\p{Is_Blk=sharada}', "");
Expect(1, 70112, '\p{^Is_Blk=sharada}', "");
Expect(1, 70112, '\P{Is_Blk=sharada}', "");
Expect(0, 70112, '\P{^Is_Blk=sharada}', "");
Expect(1, 70111, '\p{Is_Blk=  sharada}', "");
Expect(0, 70111, '\p{^Is_Blk=  sharada}', "");
Expect(0, 70111, '\P{Is_Blk=  sharada}', "");
Expect(1, 70111, '\P{^Is_Blk=  sharada}', "");
Expect(0, 70112, '\p{Is_Blk=  sharada}', "");
Expect(1, 70112, '\p{^Is_Blk=  sharada}', "");
Expect(1, 70112, '\P{Is_Blk=  sharada}', "");
Expect(0, 70112, '\P{^Is_Blk=  sharada}', "");
Error('\p{Block=/a/--Shavian}');
Error('\P{Block=/a/--Shavian}');
Expect(1, 66687, '\p{Block=shavian}', "");
Expect(0, 66687, '\p{^Block=shavian}', "");
Expect(0, 66687, '\P{Block=shavian}', "");
Expect(1, 66687, '\P{^Block=shavian}', "");
Expect(0, 66688, '\p{Block=shavian}', "");
Expect(1, 66688, '\p{^Block=shavian}', "");
Expect(1, 66688, '\P{Block=shavian}', "");
Expect(0, 66688, '\P{^Block=shavian}', "");
Expect(1, 66687, '\p{Block=-	Shavian}', "");
Expect(0, 66687, '\p{^Block=-	Shavian}', "");
Expect(0, 66687, '\P{Block=-	Shavian}', "");
Expect(1, 66687, '\P{^Block=-	Shavian}', "");
Expect(0, 66688, '\p{Block=-	Shavian}', "");
Expect(1, 66688, '\p{^Block=-	Shavian}', "");
Expect(1, 66688, '\P{Block=-	Shavian}', "");
Expect(0, 66688, '\P{^Block=-	Shavian}', "");
Error('\p{Blk= -shavian/a/}');
Error('\P{Blk= -shavian/a/}');
Expect(1, 66687, '\p{Blk=shavian}', "");
Expect(0, 66687, '\p{^Blk=shavian}', "");
Expect(0, 66687, '\P{Blk=shavian}', "");
Expect(1, 66687, '\P{^Blk=shavian}', "");
Expect(0, 66688, '\p{Blk=shavian}', "");
Expect(1, 66688, '\p{^Blk=shavian}', "");
Expect(1, 66688, '\P{Blk=shavian}', "");
Expect(0, 66688, '\P{^Blk=shavian}', "");
Expect(1, 66687, '\p{Blk= 	shavian}', "");
Expect(0, 66687, '\p{^Blk= 	shavian}', "");
Expect(0, 66687, '\P{Blk= 	shavian}', "");
Expect(1, 66687, '\P{^Blk= 	shavian}', "");
Expect(0, 66688, '\p{Blk= 	shavian}', "");
Expect(1, 66688, '\p{^Blk= 	shavian}', "");
Expect(1, 66688, '\P{Blk= 	shavian}', "");
Expect(0, 66688, '\P{^Blk= 	shavian}', "");
Error('\p{Is_Block=_ SHAVIAN/a/}');
Error('\P{Is_Block=_ SHAVIAN/a/}');
Expect(1, 66687, '\p{Is_Block=shavian}', "");
Expect(0, 66687, '\p{^Is_Block=shavian}', "");
Expect(0, 66687, '\P{Is_Block=shavian}', "");
Expect(1, 66687, '\P{^Is_Block=shavian}', "");
Expect(0, 66688, '\p{Is_Block=shavian}', "");
Expect(1, 66688, '\p{^Is_Block=shavian}', "");
Expect(1, 66688, '\P{Is_Block=shavian}', "");
Expect(0, 66688, '\P{^Is_Block=shavian}', "");
Expect(1, 66687, '\p{Is_Block=	_shavian}', "");
Expect(0, 66687, '\p{^Is_Block=	_shavian}', "");
Expect(0, 66687, '\P{Is_Block=	_shavian}', "");
Expect(1, 66687, '\P{^Is_Block=	_shavian}', "");
Expect(0, 66688, '\p{Is_Block=	_shavian}', "");
Expect(1, 66688, '\p{^Is_Block=	_shavian}', "");
Expect(1, 66688, '\P{Is_Block=	_shavian}', "");
Expect(0, 66688, '\P{^Is_Block=	_shavian}', "");
Error('\p{Is_Blk=	 SHAVIAN:=}');
Error('\P{Is_Blk=	 SHAVIAN:=}');
Expect(1, 66687, '\p{Is_Blk=shavian}', "");
Expect(0, 66687, '\p{^Is_Blk=shavian}', "");
Expect(0, 66687, '\P{Is_Blk=shavian}', "");
Expect(1, 66687, '\P{^Is_Blk=shavian}', "");
Expect(0, 66688, '\p{Is_Blk=shavian}', "");
Expect(1, 66688, '\p{^Is_Blk=shavian}', "");
Expect(1, 66688, '\P{Is_Blk=shavian}', "");
Expect(0, 66688, '\P{^Is_Blk=shavian}', "");
Expect(1, 66687, '\p{Is_Blk=_-Shavian}', "");
Expect(0, 66687, '\p{^Is_Blk=_-Shavian}', "");
Expect(0, 66687, '\P{Is_Blk=_-Shavian}', "");
Expect(1, 66687, '\P{^Is_Blk=_-Shavian}', "");
Expect(0, 66688, '\p{Is_Blk=_-Shavian}', "");
Expect(1, 66688, '\p{^Is_Blk=_-Shavian}', "");
Expect(1, 66688, '\P{Is_Blk=_-Shavian}', "");
Expect(0, 66688, '\P{^Is_Blk=_-Shavian}', "");
Error('\p{Block=	:=SINHALA}');
Error('\P{Block=	:=SINHALA}');
Expect(1, 3583, '\p{Block=sinhala}', "");
Expect(0, 3583, '\p{^Block=sinhala}', "");
Expect(0, 3583, '\P{Block=sinhala}', "");
Expect(1, 3583, '\P{^Block=sinhala}', "");
Expect(0, 3584, '\p{Block=sinhala}', "");
Expect(1, 3584, '\p{^Block=sinhala}', "");
Expect(1, 3584, '\P{Block=sinhala}', "");
Expect(0, 3584, '\P{^Block=sinhala}', "");
Expect(1, 3583, '\p{Block=__SINHALA}', "");
Expect(0, 3583, '\p{^Block=__SINHALA}', "");
Expect(0, 3583, '\P{Block=__SINHALA}', "");
Expect(1, 3583, '\P{^Block=__SINHALA}', "");
Expect(0, 3584, '\p{Block=__SINHALA}', "");
Expect(1, 3584, '\p{^Block=__SINHALA}', "");
Expect(1, 3584, '\P{Block=__SINHALA}', "");
Expect(0, 3584, '\P{^Block=__SINHALA}', "");
Error('\p{Blk=-sinhala/a/}');
Error('\P{Blk=-sinhala/a/}');
Expect(1, 3583, '\p{Blk:sinhala}', "");
Expect(0, 3583, '\p{^Blk:sinhala}', "");
Expect(0, 3583, '\P{Blk:sinhala}', "");
Expect(1, 3583, '\P{^Blk:sinhala}', "");
Expect(0, 3584, '\p{Blk:sinhala}', "");
Expect(1, 3584, '\p{^Blk:sinhala}', "");
Expect(1, 3584, '\P{Blk:sinhala}', "");
Expect(0, 3584, '\P{^Blk:sinhala}', "");
Expect(1, 3583, '\p{Blk=	-SINHALA}', "");
Expect(0, 3583, '\p{^Blk=	-SINHALA}', "");
Expect(0, 3583, '\P{Blk=	-SINHALA}', "");
Expect(1, 3583, '\P{^Blk=	-SINHALA}', "");
Expect(0, 3584, '\p{Blk=	-SINHALA}', "");
Expect(1, 3584, '\p{^Blk=	-SINHALA}', "");
Expect(1, 3584, '\P{Blk=	-SINHALA}', "");
Expect(0, 3584, '\P{^Blk=	-SINHALA}', "");
Error('\p{Is_Block= :=SINHALA}');
Error('\P{Is_Block= :=SINHALA}');
Expect(1, 3583, '\p{Is_Block=sinhala}', "");
Expect(0, 3583, '\p{^Is_Block=sinhala}', "");
Expect(0, 3583, '\P{Is_Block=sinhala}', "");
Expect(1, 3583, '\P{^Is_Block=sinhala}', "");
Expect(0, 3584, '\p{Is_Block=sinhala}', "");
Expect(1, 3584, '\p{^Is_Block=sinhala}', "");
Expect(1, 3584, '\P{Is_Block=sinhala}', "");
Expect(0, 3584, '\P{^Is_Block=sinhala}', "");
Expect(1, 3583, '\p{Is_Block: 	SINHALA}', "");
Expect(0, 3583, '\p{^Is_Block: 	SINHALA}', "");
Expect(0, 3583, '\P{Is_Block: 	SINHALA}', "");
Expect(1, 3583, '\P{^Is_Block: 	SINHALA}', "");
Expect(0, 3584, '\p{Is_Block: 	SINHALA}', "");
Expect(1, 3584, '\p{^Is_Block: 	SINHALA}', "");
Expect(1, 3584, '\P{Is_Block: 	SINHALA}', "");
Expect(0, 3584, '\P{^Is_Block: 	SINHALA}', "");
Error('\p{Is_Blk=- Sinhala/a/}');
Error('\P{Is_Blk=- Sinhala/a/}');
Expect(1, 3583, '\p{Is_Blk=sinhala}', "");
Expect(0, 3583, '\p{^Is_Blk=sinhala}', "");
Expect(0, 3583, '\P{Is_Blk=sinhala}', "");
Expect(1, 3583, '\P{^Is_Blk=sinhala}', "");
Expect(0, 3584, '\p{Is_Blk=sinhala}', "");
Expect(1, 3584, '\p{^Is_Blk=sinhala}', "");
Expect(1, 3584, '\P{Is_Blk=sinhala}', "");
Expect(0, 3584, '\P{^Is_Blk=sinhala}', "");
Expect(1, 3583, '\p{Is_Blk= SINHALA}', "");
Expect(0, 3583, '\p{^Is_Blk= SINHALA}', "");
Expect(0, 3583, '\P{Is_Blk= SINHALA}', "");
Expect(1, 3583, '\P{^Is_Blk= SINHALA}', "");
Expect(0, 3584, '\p{Is_Blk= SINHALA}', "");
Expect(1, 3584, '\p{^Is_Blk= SINHALA}', "");
Expect(1, 3584, '\P{Is_Blk= SINHALA}', "");
Expect(0, 3584, '\P{^Is_Blk= SINHALA}', "");
Error('\p{Block=-/a/SMALL_Form_VARIANTS}');
Error('\P{Block=-/a/SMALL_Form_VARIANTS}');
Expect(1, 65135, '\p{Block=smallformvariants}', "");
Expect(0, 65135, '\p{^Block=smallformvariants}', "");
Expect(0, 65135, '\P{Block=smallformvariants}', "");
Expect(1, 65135, '\P{^Block=smallformvariants}', "");
Expect(0, 65136, '\p{Block=smallformvariants}', "");
Expect(1, 65136, '\p{^Block=smallformvariants}', "");
Expect(1, 65136, '\P{Block=smallformvariants}', "");
Expect(0, 65136, '\P{^Block=smallformvariants}', "");
Expect(1, 65135, '\p{Block=_Small_FORM_variants}', "");
Expect(0, 65135, '\p{^Block=_Small_FORM_variants}', "");
Expect(0, 65135, '\P{Block=_Small_FORM_variants}', "");
Expect(1, 65135, '\P{^Block=_Small_FORM_variants}', "");
Expect(0, 65136, '\p{Block=_Small_FORM_variants}', "");
Expect(1, 65136, '\p{^Block=_Small_FORM_variants}', "");
Expect(1, 65136, '\P{Block=_Small_FORM_variants}', "");
Expect(0, 65136, '\P{^Block=_Small_FORM_variants}', "");
Error('\p{Blk=_small_forms:=}');
Error('\P{Blk=_small_forms:=}');
Expect(1, 65135, '\p{Blk=smallforms}', "");
Expect(0, 65135, '\p{^Blk=smallforms}', "");
Expect(0, 65135, '\P{Blk=smallforms}', "");
Expect(1, 65135, '\P{^Blk=smallforms}', "");
Expect(0, 65136, '\p{Blk=smallforms}', "");
Expect(1, 65136, '\p{^Blk=smallforms}', "");
Expect(1, 65136, '\P{Blk=smallforms}', "");
Expect(0, 65136, '\P{^Blk=smallforms}', "");
Expect(1, 65135, '\p{Blk=	Small_forms}', "");
Expect(0, 65135, '\p{^Blk=	Small_forms}', "");
Expect(0, 65135, '\P{Blk=	Small_forms}', "");
Expect(1, 65135, '\P{^Blk=	Small_forms}', "");
Expect(0, 65136, '\p{Blk=	Small_forms}', "");
Expect(1, 65136, '\p{^Blk=	Small_forms}', "");
Expect(1, 65136, '\P{Blk=	Small_forms}', "");
Expect(0, 65136, '\P{^Blk=	Small_forms}', "");
Error('\p{Is_Block= small_form_Variants/a/}');
Error('\P{Is_Block= small_form_Variants/a/}');
Expect(1, 65135, '\p{Is_Block=smallformvariants}', "");
Expect(0, 65135, '\p{^Is_Block=smallformvariants}', "");
Expect(0, 65135, '\P{Is_Block=smallformvariants}', "");
Expect(1, 65135, '\P{^Is_Block=smallformvariants}', "");
Expect(0, 65136, '\p{Is_Block=smallformvariants}', "");
Expect(1, 65136, '\p{^Is_Block=smallformvariants}', "");
Expect(1, 65136, '\P{Is_Block=smallformvariants}', "");
Expect(0, 65136, '\P{^Is_Block=smallformvariants}', "");
Expect(1, 65135, '\p{Is_Block= Small_FORM_VARIANTS}', "");
Expect(0, 65135, '\p{^Is_Block= Small_FORM_VARIANTS}', "");
Expect(0, 65135, '\P{Is_Block= Small_FORM_VARIANTS}', "");
Expect(1, 65135, '\P{^Is_Block= Small_FORM_VARIANTS}', "");
Expect(0, 65136, '\p{Is_Block= Small_FORM_VARIANTS}', "");
Expect(1, 65136, '\p{^Is_Block= Small_FORM_VARIANTS}', "");
Expect(1, 65136, '\P{Is_Block= Small_FORM_VARIANTS}', "");
Expect(0, 65136, '\P{^Is_Block= Small_FORM_VARIANTS}', "");
Error('\p{Is_Blk=_:=SMALL_FORMS}');
Error('\P{Is_Blk=_:=SMALL_FORMS}');
Expect(1, 65135, '\p{Is_Blk=smallforms}', "");
Expect(0, 65135, '\p{^Is_Blk=smallforms}', "");
Expect(0, 65135, '\P{Is_Blk=smallforms}', "");
Expect(1, 65135, '\P{^Is_Blk=smallforms}', "");
Expect(0, 65136, '\p{Is_Blk=smallforms}', "");
Expect(1, 65136, '\p{^Is_Blk=smallforms}', "");
Expect(1, 65136, '\P{Is_Blk=smallforms}', "");
Expect(0, 65136, '\P{^Is_Blk=smallforms}', "");
Expect(1, 65135, '\p{Is_Blk: 	-Small_Forms}', "");
Expect(0, 65135, '\p{^Is_Blk: 	-Small_Forms}', "");
Expect(0, 65135, '\P{Is_Blk: 	-Small_Forms}', "");
Expect(1, 65135, '\P{^Is_Blk: 	-Small_Forms}', "");
Expect(0, 65136, '\p{Is_Blk: 	-Small_Forms}', "");
Expect(1, 65136, '\p{^Is_Blk: 	-Small_Forms}', "");
Expect(1, 65136, '\P{Is_Blk: 	-Small_Forms}', "");
Expect(0, 65136, '\P{^Is_Blk: 	-Small_Forms}', "");
Error('\p{Block: -_SORA_Sompeng:=}');
Error('\P{Block: -_SORA_Sompeng:=}');
Expect(1, 69887, '\p{Block=sorasompeng}', "");
Expect(0, 69887, '\p{^Block=sorasompeng}', "");
Expect(0, 69887, '\P{Block=sorasompeng}', "");
Expect(1, 69887, '\P{^Block=sorasompeng}', "");
Expect(0, 69888, '\p{Block=sorasompeng}', "");
Expect(1, 69888, '\p{^Block=sorasompeng}', "");
Expect(1, 69888, '\P{Block=sorasompeng}', "");
Expect(0, 69888, '\P{^Block=sorasompeng}', "");
Expect(1, 69887, '\p{Block=	sora_SOMPENG}', "");
Expect(0, 69887, '\p{^Block=	sora_SOMPENG}', "");
Expect(0, 69887, '\P{Block=	sora_SOMPENG}', "");
Expect(1, 69887, '\P{^Block=	sora_SOMPENG}', "");
Expect(0, 69888, '\p{Block=	sora_SOMPENG}', "");
Expect(1, 69888, '\p{^Block=	sora_SOMPENG}', "");
Expect(1, 69888, '\P{Block=	sora_SOMPENG}', "");
Expect(0, 69888, '\P{^Block=	sora_SOMPENG}', "");
Error('\p{Blk=/a/ 	Sora_Sompeng}');
Error('\P{Blk=/a/ 	Sora_Sompeng}');
Expect(1, 69887, '\p{Blk=sorasompeng}', "");
Expect(0, 69887, '\p{^Blk=sorasompeng}', "");
Expect(0, 69887, '\P{Blk=sorasompeng}', "");
Expect(1, 69887, '\P{^Blk=sorasompeng}', "");
Expect(0, 69888, '\p{Blk=sorasompeng}', "");
Expect(1, 69888, '\p{^Blk=sorasompeng}', "");
Expect(1, 69888, '\P{Blk=sorasompeng}', "");
Expect(0, 69888, '\P{^Blk=sorasompeng}', "");
Expect(1, 69887, '\p{Blk=	-Sora_Sompeng}', "");
Expect(0, 69887, '\p{^Blk=	-Sora_Sompeng}', "");
Expect(0, 69887, '\P{Blk=	-Sora_Sompeng}', "");
Expect(1, 69887, '\P{^Blk=	-Sora_Sompeng}', "");
Expect(0, 69888, '\p{Blk=	-Sora_Sompeng}', "");
Expect(1, 69888, '\p{^Blk=	-Sora_Sompeng}', "");
Expect(1, 69888, '\P{Blk=	-Sora_Sompeng}', "");
Expect(0, 69888, '\P{^Blk=	-Sora_Sompeng}', "");
Error('\p{Is_Block=/a/_	SORA_Sompeng}');
Error('\P{Is_Block=/a/_	SORA_Sompeng}');
Expect(1, 69887, '\p{Is_Block=sorasompeng}', "");
Expect(0, 69887, '\p{^Is_Block=sorasompeng}', "");
Expect(0, 69887, '\P{Is_Block=sorasompeng}', "");
Expect(1, 69887, '\P{^Is_Block=sorasompeng}', "");
Expect(0, 69888, '\p{Is_Block=sorasompeng}', "");
Expect(1, 69888, '\p{^Is_Block=sorasompeng}', "");
Expect(1, 69888, '\P{Is_Block=sorasompeng}', "");
Expect(0, 69888, '\P{^Is_Block=sorasompeng}', "");
Expect(1, 69887, '\p{Is_Block= SORA_Sompeng}', "");
Expect(0, 69887, '\p{^Is_Block= SORA_Sompeng}', "");
Expect(0, 69887, '\P{Is_Block= SORA_Sompeng}', "");
Expect(1, 69887, '\P{^Is_Block= SORA_Sompeng}', "");
Expect(0, 69888, '\p{Is_Block= SORA_Sompeng}', "");
Expect(1, 69888, '\p{^Is_Block= SORA_Sompeng}', "");
Expect(1, 69888, '\P{Is_Block= SORA_Sompeng}', "");
Expect(0, 69888, '\P{^Is_Block= SORA_Sompeng}', "");
Error('\p{Is_Blk= /a/Sora_sompeng}');
Error('\P{Is_Blk= /a/Sora_sompeng}');
Expect(1, 69887, '\p{Is_Blk=sorasompeng}', "");
Expect(0, 69887, '\p{^Is_Blk=sorasompeng}', "");
Expect(0, 69887, '\P{Is_Blk=sorasompeng}', "");
Expect(1, 69887, '\P{^Is_Blk=sorasompeng}', "");
Expect(0, 69888, '\p{Is_Blk=sorasompeng}', "");
Expect(1, 69888, '\p{^Is_Blk=sorasompeng}', "");
Expect(1, 69888, '\P{Is_Blk=sorasompeng}', "");
Expect(0, 69888, '\P{^Is_Blk=sorasompeng}', "");
Expect(1, 69887, '\p{Is_Blk=-SORA_SOMPENG}', "");
Expect(0, 69887, '\p{^Is_Blk=-SORA_SOMPENG}', "");
Expect(0, 69887, '\P{Is_Blk=-SORA_SOMPENG}', "");
Expect(1, 69887, '\P{^Is_Blk=-SORA_SOMPENG}', "");
Expect(0, 69888, '\p{Is_Blk=-SORA_SOMPENG}', "");
Expect(1, 69888, '\p{^Is_Blk=-SORA_SOMPENG}', "");
Expect(1, 69888, '\P{Is_Blk=-SORA_SOMPENG}', "");
Expect(0, 69888, '\P{^Is_Blk=-SORA_SOMPENG}', "");
Error('\p{Block:   :=specials}');
Error('\P{Block:   :=specials}');
Expect(1, 65520, '\p{Block=specials}', "");
Expect(0, 65520, '\p{^Block=specials}', "");
Expect(0, 65520, '\P{Block=specials}', "");
Expect(1, 65520, '\P{^Block=specials}', "");
Expect(0, 65536, '\p{Block=specials}', "");
Expect(1, 65536, '\p{^Block=specials}', "");
Expect(1, 65536, '\P{Block=specials}', "");
Expect(0, 65536, '\P{^Block=specials}', "");
Expect(1, 65520, '\p{Block=	 specials}', "");
Expect(0, 65520, '\p{^Block=	 specials}', "");
Expect(0, 65520, '\P{Block=	 specials}', "");
Expect(1, 65520, '\P{^Block=	 specials}', "");
Expect(0, 65536, '\p{Block=	 specials}', "");
Expect(1, 65536, '\p{^Block=	 specials}', "");
Expect(1, 65536, '\P{Block=	 specials}', "");
Expect(0, 65536, '\P{^Block=	 specials}', "");
Error('\p{Blk=/a/_ Specials}');
Error('\P{Blk=/a/_ Specials}');
Expect(1, 65520, '\p{Blk=specials}', "");
Expect(0, 65520, '\p{^Blk=specials}', "");
Expect(0, 65520, '\P{Blk=specials}', "");
Expect(1, 65520, '\P{^Blk=specials}', "");
Expect(0, 65536, '\p{Blk=specials}', "");
Expect(1, 65536, '\p{^Blk=specials}', "");
Expect(1, 65536, '\P{Blk=specials}', "");
Expect(0, 65536, '\P{^Blk=specials}', "");
Expect(1, 65520, '\p{Blk= 	Specials}', "");
Expect(0, 65520, '\p{^Blk= 	Specials}', "");
Expect(0, 65520, '\P{Blk= 	Specials}', "");
Expect(1, 65520, '\P{^Blk= 	Specials}', "");
Expect(0, 65536, '\p{Blk= 	Specials}', "");
Expect(1, 65536, '\p{^Blk= 	Specials}', "");
Expect(1, 65536, '\P{Blk= 	Specials}', "");
Expect(0, 65536, '\P{^Blk= 	Specials}', "");
Error('\p{Is_Block=:=		Specials}');
Error('\P{Is_Block=:=		Specials}');
Expect(1, 65520, '\p{Is_Block=specials}', "");
Expect(0, 65520, '\p{^Is_Block=specials}', "");
Expect(0, 65520, '\P{Is_Block=specials}', "");
Expect(1, 65520, '\P{^Is_Block=specials}', "");
Expect(0, 65536, '\p{Is_Block=specials}', "");
Expect(1, 65536, '\p{^Is_Block=specials}', "");
Expect(1, 65536, '\P{Is_Block=specials}', "");
Expect(0, 65536, '\P{^Is_Block=specials}', "");
Expect(1, 65520, '\p{Is_Block= 	specials}', "");
Expect(0, 65520, '\p{^Is_Block= 	specials}', "");
Expect(0, 65520, '\P{Is_Block= 	specials}', "");
Expect(1, 65520, '\P{^Is_Block= 	specials}', "");
Expect(0, 65536, '\p{Is_Block= 	specials}', "");
Expect(1, 65536, '\p{^Is_Block= 	specials}', "");
Expect(1, 65536, '\P{Is_Block= 	specials}', "");
Expect(0, 65536, '\P{^Is_Block= 	specials}', "");
Error('\p{Is_Blk=/a/	Specials}');
Error('\P{Is_Blk=/a/	Specials}');
Expect(1, 65520, '\p{Is_Blk=specials}', "");
Expect(0, 65520, '\p{^Is_Blk=specials}', "");
Expect(0, 65520, '\P{Is_Blk=specials}', "");
Expect(1, 65520, '\P{^Is_Blk=specials}', "");
Expect(0, 65536, '\p{Is_Blk=specials}', "");
Expect(1, 65536, '\p{^Is_Blk=specials}', "");
Expect(1, 65536, '\P{Is_Blk=specials}', "");
Expect(0, 65536, '\P{^Is_Blk=specials}', "");
Expect(1, 65520, '\p{Is_Blk=-	specials}', "");
Expect(0, 65520, '\p{^Is_Blk=-	specials}', "");
Expect(0, 65520, '\P{Is_Blk=-	specials}', "");
Expect(1, 65520, '\P{^Is_Blk=-	specials}', "");
Expect(0, 65536, '\p{Is_Blk=-	specials}', "");
Expect(1, 65536, '\p{^Is_Blk=-	specials}', "");
Expect(1, 65536, '\P{Is_Blk=-	specials}', "");
Expect(0, 65536, '\P{^Is_Blk=-	specials}', "");
Error('\p{Block=:=_-Sundanese}');
Error('\P{Block=:=_-Sundanese}');
Expect(1, 7103, '\p{Block=sundanese}', "");
Expect(0, 7103, '\p{^Block=sundanese}', "");
Expect(0, 7103, '\P{Block=sundanese}', "");
Expect(1, 7103, '\P{^Block=sundanese}', "");
Expect(0, 7104, '\p{Block=sundanese}', "");
Expect(1, 7104, '\p{^Block=sundanese}', "");
Expect(1, 7104, '\P{Block=sundanese}', "");
Expect(0, 7104, '\P{^Block=sundanese}', "");
Expect(1, 7103, '\p{Block:	--Sundanese}', "");
Expect(0, 7103, '\p{^Block:	--Sundanese}', "");
Expect(0, 7103, '\P{Block:	--Sundanese}', "");
Expect(1, 7103, '\P{^Block:	--Sundanese}', "");
Expect(0, 7104, '\p{Block:	--Sundanese}', "");
Expect(1, 7104, '\p{^Block:	--Sundanese}', "");
Expect(1, 7104, '\P{Block:	--Sundanese}', "");
Expect(0, 7104, '\P{^Block:	--Sundanese}', "");
Error('\p{Blk= -SUNDANESE:=}');
Error('\P{Blk= -SUNDANESE:=}');
Expect(1, 7103, '\p{Blk=sundanese}', "");
Expect(0, 7103, '\p{^Blk=sundanese}', "");
Expect(0, 7103, '\P{Blk=sundanese}', "");
Expect(1, 7103, '\P{^Blk=sundanese}', "");
Expect(0, 7104, '\p{Blk=sundanese}', "");
Expect(1, 7104, '\p{^Blk=sundanese}', "");
Expect(1, 7104, '\P{Blk=sundanese}', "");
Expect(0, 7104, '\P{^Blk=sundanese}', "");
Expect(1, 7103, '\p{Blk=- Sundanese}', "");
Expect(0, 7103, '\p{^Blk=- Sundanese}', "");
Expect(0, 7103, '\P{Blk=- Sundanese}', "");
Expect(1, 7103, '\P{^Blk=- Sundanese}', "");
Expect(0, 7104, '\p{Blk=- Sundanese}', "");
Expect(1, 7104, '\p{^Blk=- Sundanese}', "");
Expect(1, 7104, '\P{Blk=- Sundanese}', "");
Expect(0, 7104, '\P{^Blk=- Sundanese}', "");
Error('\p{Is_Block=- SUNDANESE/a/}');
Error('\P{Is_Block=- SUNDANESE/a/}');
Expect(1, 7103, '\p{Is_Block=sundanese}', "");
Expect(0, 7103, '\p{^Is_Block=sundanese}', "");
Expect(0, 7103, '\P{Is_Block=sundanese}', "");
Expect(1, 7103, '\P{^Is_Block=sundanese}', "");
Expect(0, 7104, '\p{Is_Block=sundanese}', "");
Expect(1, 7104, '\p{^Is_Block=sundanese}', "");
Expect(1, 7104, '\P{Is_Block=sundanese}', "");
Expect(0, 7104, '\P{^Is_Block=sundanese}', "");
Expect(1, 7103, '\p{Is_Block=_ Sundanese}', "");
Expect(0, 7103, '\p{^Is_Block=_ Sundanese}', "");
Expect(0, 7103, '\P{Is_Block=_ Sundanese}', "");
Expect(1, 7103, '\P{^Is_Block=_ Sundanese}', "");
Expect(0, 7104, '\p{Is_Block=_ Sundanese}', "");
Expect(1, 7104, '\p{^Is_Block=_ Sundanese}', "");
Expect(1, 7104, '\P{Is_Block=_ Sundanese}', "");
Expect(0, 7104, '\P{^Is_Block=_ Sundanese}', "");
Error('\p{Is_Blk=-sundanese:=}');
Error('\P{Is_Blk=-sundanese:=}');
Expect(1, 7103, '\p{Is_Blk=sundanese}', "");
Expect(0, 7103, '\p{^Is_Blk=sundanese}', "");
Expect(0, 7103, '\P{Is_Blk=sundanese}', "");
Expect(1, 7103, '\P{^Is_Blk=sundanese}', "");
Expect(0, 7104, '\p{Is_Blk=sundanese}', "");
Expect(1, 7104, '\p{^Is_Blk=sundanese}', "");
Expect(1, 7104, '\P{Is_Blk=sundanese}', "");
Expect(0, 7104, '\P{^Is_Blk=sundanese}', "");
Expect(1, 7103, '\p{Is_Blk=_	sundanese}', "");
Expect(0, 7103, '\p{^Is_Blk=_	sundanese}', "");
Expect(0, 7103, '\P{Is_Blk=_	sundanese}', "");
Expect(1, 7103, '\P{^Is_Blk=_	sundanese}', "");
Expect(0, 7104, '\p{Is_Blk=_	sundanese}', "");
Expect(1, 7104, '\p{^Is_Blk=_	sundanese}', "");
Expect(1, 7104, '\P{Is_Blk=_	sundanese}', "");
Expect(0, 7104, '\P{^Is_Blk=_	sundanese}', "");
Error('\p{Block:_:=SUNDANESE_SUPPLEMENT}');
Error('\P{Block:_:=SUNDANESE_SUPPLEMENT}');
Expect(1, 7375, '\p{Block=sundanesesupplement}', "");
Expect(0, 7375, '\p{^Block=sundanesesupplement}', "");
Expect(0, 7375, '\P{Block=sundanesesupplement}', "");
Expect(1, 7375, '\P{^Block=sundanesesupplement}', "");
Expect(0, 7376, '\p{Block=sundanesesupplement}', "");
Expect(1, 7376, '\p{^Block=sundanesesupplement}', "");
Expect(1, 7376, '\P{Block=sundanesesupplement}', "");
Expect(0, 7376, '\P{^Block=sundanesesupplement}', "");
Expect(1, 7375, '\p{Block=	sundanese_SUPPLEMENT}', "");
Expect(0, 7375, '\p{^Block=	sundanese_SUPPLEMENT}', "");
Expect(0, 7375, '\P{Block=	sundanese_SUPPLEMENT}', "");
Expect(1, 7375, '\P{^Block=	sundanese_SUPPLEMENT}', "");
Expect(0, 7376, '\p{Block=	sundanese_SUPPLEMENT}', "");
Expect(1, 7376, '\p{^Block=	sundanese_SUPPLEMENT}', "");
Expect(1, 7376, '\P{Block=	sundanese_SUPPLEMENT}', "");
Expect(0, 7376, '\P{^Block=	sundanese_SUPPLEMENT}', "");
Error('\p{Blk=:=_SUNDANESE_SUP}');
Error('\P{Blk=:=_SUNDANESE_SUP}');
Expect(1, 7375, '\p{Blk=sundanesesup}', "");
Expect(0, 7375, '\p{^Blk=sundanesesup}', "");
Expect(0, 7375, '\P{Blk=sundanesesup}', "");
Expect(1, 7375, '\P{^Blk=sundanesesup}', "");
Expect(0, 7376, '\p{Blk=sundanesesup}', "");
Expect(1, 7376, '\p{^Blk=sundanesesup}', "");
Expect(1, 7376, '\P{Blk=sundanesesup}', "");
Expect(0, 7376, '\P{^Blk=sundanesesup}', "");
Expect(1, 7375, '\p{Blk= 	SUNDANESE_SUP}', "");
Expect(0, 7375, '\p{^Blk= 	SUNDANESE_SUP}', "");
Expect(0, 7375, '\P{Blk= 	SUNDANESE_SUP}', "");
Expect(1, 7375, '\P{^Blk= 	SUNDANESE_SUP}', "");
Expect(0, 7376, '\p{Blk= 	SUNDANESE_SUP}', "");
Expect(1, 7376, '\p{^Blk= 	SUNDANESE_SUP}', "");
Expect(1, 7376, '\P{Blk= 	SUNDANESE_SUP}', "");
Expect(0, 7376, '\P{^Blk= 	SUNDANESE_SUP}', "");
Error('\p{Is_Block=/a/- SUNDANESE_Supplement}');
Error('\P{Is_Block=/a/- SUNDANESE_Supplement}');
Expect(1, 7375, '\p{Is_Block=sundanesesupplement}', "");
Expect(0, 7375, '\p{^Is_Block=sundanesesupplement}', "");
Expect(0, 7375, '\P{Is_Block=sundanesesupplement}', "");
Expect(1, 7375, '\P{^Is_Block=sundanesesupplement}', "");
Expect(0, 7376, '\p{Is_Block=sundanesesupplement}', "");
Expect(1, 7376, '\p{^Is_Block=sundanesesupplement}', "");
Expect(1, 7376, '\P{Is_Block=sundanesesupplement}', "");
Expect(0, 7376, '\P{^Is_Block=sundanesesupplement}', "");
Expect(1, 7375, '\p{Is_Block= -SUNDANESE_Supplement}', "");
Expect(0, 7375, '\p{^Is_Block= -SUNDANESE_Supplement}', "");
Expect(0, 7375, '\P{Is_Block= -SUNDANESE_Supplement}', "");
Expect(1, 7375, '\P{^Is_Block= -SUNDANESE_Supplement}', "");
Expect(0, 7376, '\p{Is_Block= -SUNDANESE_Supplement}', "");
Expect(1, 7376, '\p{^Is_Block= -SUNDANESE_Supplement}', "");
Expect(1, 7376, '\P{Is_Block= -SUNDANESE_Supplement}', "");
Expect(0, 7376, '\P{^Is_Block= -SUNDANESE_Supplement}', "");
Error('\p{Is_Blk=/a/	sundanese_SUP}');
Error('\P{Is_Blk=/a/	sundanese_SUP}');
Expect(1, 7375, '\p{Is_Blk:	sundanesesup}', "");
Expect(0, 7375, '\p{^Is_Blk:	sundanesesup}', "");
Expect(0, 7375, '\P{Is_Blk:	sundanesesup}', "");
Expect(1, 7375, '\P{^Is_Blk:	sundanesesup}', "");
Expect(0, 7376, '\p{Is_Blk:	sundanesesup}', "");
Expect(1, 7376, '\p{^Is_Blk:	sundanesesup}', "");
Expect(1, 7376, '\P{Is_Blk:	sundanesesup}', "");
Expect(0, 7376, '\P{^Is_Blk:	sundanesesup}', "");
Expect(1, 7375, '\p{Is_Blk=-Sundanese_Sup}', "");
Expect(0, 7375, '\p{^Is_Blk=-Sundanese_Sup}', "");
Expect(0, 7375, '\P{Is_Blk=-Sundanese_Sup}', "");
Expect(1, 7375, '\P{^Is_Blk=-Sundanese_Sup}', "");
Expect(0, 7376, '\p{Is_Blk=-Sundanese_Sup}', "");
Expect(1, 7376, '\p{^Is_Blk=-Sundanese_Sup}', "");
Expect(1, 7376, '\P{Is_Blk=-Sundanese_Sup}', "");
Expect(0, 7376, '\P{^Is_Blk=-Sundanese_Sup}', "");
Error('\p{Block=	 supplemental_arrows_A/a/}');
Error('\P{Block=	 supplemental_arrows_A/a/}');
Expect(1, 10239, '\p{Block=supplementalarrowsa}', "");
Expect(0, 10239, '\p{^Block=supplementalarrowsa}', "");
Expect(0, 10239, '\P{Block=supplementalarrowsa}', "");
Expect(1, 10239, '\P{^Block=supplementalarrowsa}', "");
Expect(0, 10240, '\p{Block=supplementalarrowsa}', "");
Expect(1, 10240, '\p{^Block=supplementalarrowsa}', "");
Expect(1, 10240, '\P{Block=supplementalarrowsa}', "");
Expect(0, 10240, '\P{^Block=supplementalarrowsa}', "");
Expect(1, 10239, '\p{Block=-_supplemental_Arrows_A}', "");
Expect(0, 10239, '\p{^Block=-_supplemental_Arrows_A}', "");
Expect(0, 10239, '\P{Block=-_supplemental_Arrows_A}', "");
Expect(1, 10239, '\P{^Block=-_supplemental_Arrows_A}', "");
Expect(0, 10240, '\p{Block=-_supplemental_Arrows_A}', "");
Expect(1, 10240, '\p{^Block=-_supplemental_Arrows_A}', "");
Expect(1, 10240, '\P{Block=-_supplemental_Arrows_A}', "");
Expect(0, 10240, '\P{^Block=-_supplemental_Arrows_A}', "");
Error('\p{Blk:			Sup_Arrows_A/a/}');
Error('\P{Blk:			Sup_Arrows_A/a/}');
Expect(1, 10239, '\p{Blk=suparrowsa}', "");
Expect(0, 10239, '\p{^Blk=suparrowsa}', "");
Expect(0, 10239, '\P{Blk=suparrowsa}', "");
Expect(1, 10239, '\P{^Blk=suparrowsa}', "");
Expect(0, 10240, '\p{Blk=suparrowsa}', "");
Expect(1, 10240, '\p{^Blk=suparrowsa}', "");
Expect(1, 10240, '\P{Blk=suparrowsa}', "");
Expect(0, 10240, '\P{^Blk=suparrowsa}', "");
Expect(1, 10239, '\p{Blk=-Sup_arrows_a}', "");
Expect(0, 10239, '\p{^Blk=-Sup_arrows_a}', "");
Expect(0, 10239, '\P{Blk=-Sup_arrows_a}', "");
Expect(1, 10239, '\P{^Blk=-Sup_arrows_a}', "");
Expect(0, 10240, '\p{Blk=-Sup_arrows_a}', "");
Expect(1, 10240, '\p{^Blk=-Sup_arrows_a}', "");
Expect(1, 10240, '\P{Blk=-Sup_arrows_a}', "");
Expect(0, 10240, '\P{^Blk=-Sup_arrows_a}', "");
Error('\p{Is_Block=_:=Supplemental_ARROWS_A}');
Error('\P{Is_Block=_:=Supplemental_ARROWS_A}');
Expect(1, 10239, '\p{Is_Block=supplementalarrowsa}', "");
Expect(0, 10239, '\p{^Is_Block=supplementalarrowsa}', "");
Expect(0, 10239, '\P{Is_Block=supplementalarrowsa}', "");
Expect(1, 10239, '\P{^Is_Block=supplementalarrowsa}', "");
Expect(0, 10240, '\p{Is_Block=supplementalarrowsa}', "");
Expect(1, 10240, '\p{^Is_Block=supplementalarrowsa}', "");
Expect(1, 10240, '\P{Is_Block=supplementalarrowsa}', "");
Expect(0, 10240, '\P{^Is_Block=supplementalarrowsa}', "");
Expect(1, 10239, '\p{Is_Block= -SUPPLEMENTAL_arrows_A}', "");
Expect(0, 10239, '\p{^Is_Block= -SUPPLEMENTAL_arrows_A}', "");
Expect(0, 10239, '\P{Is_Block= -SUPPLEMENTAL_arrows_A}', "");
Expect(1, 10239, '\P{^Is_Block= -SUPPLEMENTAL_arrows_A}', "");
Expect(0, 10240, '\p{Is_Block= -SUPPLEMENTAL_arrows_A}', "");
Expect(1, 10240, '\p{^Is_Block= -SUPPLEMENTAL_arrows_A}', "");
Expect(1, 10240, '\P{Is_Block= -SUPPLEMENTAL_arrows_A}', "");
Expect(0, 10240, '\P{^Is_Block= -SUPPLEMENTAL_arrows_A}', "");
Error('\p{Is_Blk=_:=sup_ARROWS_a}');
Error('\P{Is_Blk=_:=sup_ARROWS_a}');
Expect(1, 10239, '\p{Is_Blk=suparrowsa}', "");
Expect(0, 10239, '\p{^Is_Blk=suparrowsa}', "");
Expect(0, 10239, '\P{Is_Blk=suparrowsa}', "");
Expect(1, 10239, '\P{^Is_Blk=suparrowsa}', "");
Expect(0, 10240, '\p{Is_Blk=suparrowsa}', "");
Expect(1, 10240, '\p{^Is_Blk=suparrowsa}', "");
Expect(1, 10240, '\P{Is_Blk=suparrowsa}', "");
Expect(0, 10240, '\P{^Is_Blk=suparrowsa}', "");
Expect(1, 10239, '\p{Is_Blk=Sup_ARROWS_A}', "");
Expect(0, 10239, '\p{^Is_Blk=Sup_ARROWS_A}', "");
Expect(0, 10239, '\P{Is_Blk=Sup_ARROWS_A}', "");
Expect(1, 10239, '\P{^Is_Blk=Sup_ARROWS_A}', "");
Expect(0, 10240, '\p{Is_Blk=Sup_ARROWS_A}', "");
Expect(1, 10240, '\p{^Is_Blk=Sup_ARROWS_A}', "");
Expect(1, 10240, '\P{Is_Blk=Sup_ARROWS_A}', "");
Expect(0, 10240, '\P{^Is_Blk=Sup_ARROWS_A}', "");
Error('\p{Block:   /a/Supplemental_arrows_B}');
Error('\P{Block:   /a/Supplemental_arrows_B}');
Expect(1, 10623, '\p{Block=supplementalarrowsb}', "");
Expect(0, 10623, '\p{^Block=supplementalarrowsb}', "");
Expect(0, 10623, '\P{Block=supplementalarrowsb}', "");
Expect(1, 10623, '\P{^Block=supplementalarrowsb}', "");
Expect(0, 10624, '\p{Block=supplementalarrowsb}', "");
Expect(1, 10624, '\p{^Block=supplementalarrowsb}', "");
Expect(1, 10624, '\P{Block=supplementalarrowsb}', "");
Expect(0, 10624, '\P{^Block=supplementalarrowsb}', "");
Expect(1, 10623, '\p{Block=SUPPLEMENTAL_Arrows_B}', "");
Expect(0, 10623, '\p{^Block=SUPPLEMENTAL_Arrows_B}', "");
Expect(0, 10623, '\P{Block=SUPPLEMENTAL_Arrows_B}', "");
Expect(1, 10623, '\P{^Block=SUPPLEMENTAL_Arrows_B}', "");
Expect(0, 10624, '\p{Block=SUPPLEMENTAL_Arrows_B}', "");
Expect(1, 10624, '\p{^Block=SUPPLEMENTAL_Arrows_B}', "");
Expect(1, 10624, '\P{Block=SUPPLEMENTAL_Arrows_B}', "");
Expect(0, 10624, '\P{^Block=SUPPLEMENTAL_Arrows_B}', "");
Error('\p{Blk=:=		Sup_Arrows_B}');
Error('\P{Blk=:=		Sup_Arrows_B}');
Expect(1, 10623, '\p{Blk=suparrowsb}', "");
Expect(0, 10623, '\p{^Blk=suparrowsb}', "");
Expect(0, 10623, '\P{Blk=suparrowsb}', "");
Expect(1, 10623, '\P{^Blk=suparrowsb}', "");
Expect(0, 10624, '\p{Blk=suparrowsb}', "");
Expect(1, 10624, '\p{^Blk=suparrowsb}', "");
Expect(1, 10624, '\P{Blk=suparrowsb}', "");
Expect(0, 10624, '\P{^Blk=suparrowsb}', "");
Expect(1, 10623, '\p{Blk=_ SUP_Arrows_b}', "");
Expect(0, 10623, '\p{^Blk=_ SUP_Arrows_b}', "");
Expect(0, 10623, '\P{Blk=_ SUP_Arrows_b}', "");
Expect(1, 10623, '\P{^Blk=_ SUP_Arrows_b}', "");
Expect(0, 10624, '\p{Blk=_ SUP_Arrows_b}', "");
Expect(1, 10624, '\p{^Blk=_ SUP_Arrows_b}', "");
Expect(1, 10624, '\P{Blk=_ SUP_Arrows_b}', "");
Expect(0, 10624, '\P{^Blk=_ SUP_Arrows_b}', "");
Error('\p{Is_Block= Supplemental_arrows_B/a/}');
Error('\P{Is_Block= Supplemental_arrows_B/a/}');
Expect(1, 10623, '\p{Is_Block=supplementalarrowsb}', "");
Expect(0, 10623, '\p{^Is_Block=supplementalarrowsb}', "");
Expect(0, 10623, '\P{Is_Block=supplementalarrowsb}', "");
Expect(1, 10623, '\P{^Is_Block=supplementalarrowsb}', "");
Expect(0, 10624, '\p{Is_Block=supplementalarrowsb}', "");
Expect(1, 10624, '\p{^Is_Block=supplementalarrowsb}', "");
Expect(1, 10624, '\P{Is_Block=supplementalarrowsb}', "");
Expect(0, 10624, '\P{^Is_Block=supplementalarrowsb}', "");
Expect(1, 10623, '\p{Is_Block=_Supplemental_arrows_B}', "");
Expect(0, 10623, '\p{^Is_Block=_Supplemental_arrows_B}', "");
Expect(0, 10623, '\P{Is_Block=_Supplemental_arrows_B}', "");
Expect(1, 10623, '\P{^Is_Block=_Supplemental_arrows_B}', "");
Expect(0, 10624, '\p{Is_Block=_Supplemental_arrows_B}', "");
Expect(1, 10624, '\p{^Is_Block=_Supplemental_arrows_B}', "");
Expect(1, 10624, '\P{Is_Block=_Supplemental_arrows_B}', "");
Expect(0, 10624, '\P{^Is_Block=_Supplemental_arrows_B}', "");
Error('\p{Is_Blk=_/a/sup_arrows_B}');
Error('\P{Is_Blk=_/a/sup_arrows_B}');
Expect(1, 10623, '\p{Is_Blk=suparrowsb}', "");
Expect(0, 10623, '\p{^Is_Blk=suparrowsb}', "");
Expect(0, 10623, '\P{Is_Blk=suparrowsb}', "");
Expect(1, 10623, '\P{^Is_Blk=suparrowsb}', "");
Expect(0, 10624, '\p{Is_Blk=suparrowsb}', "");
Expect(1, 10624, '\p{^Is_Blk=suparrowsb}', "");
Expect(1, 10624, '\P{Is_Blk=suparrowsb}', "");
Expect(0, 10624, '\P{^Is_Blk=suparrowsb}', "");
Expect(1, 10623, '\p{Is_Blk=- Sup_ARROWS_B}', "");
Expect(0, 10623, '\p{^Is_Blk=- Sup_ARROWS_B}', "");
Expect(0, 10623, '\P{Is_Blk=- Sup_ARROWS_B}', "");
Expect(1, 10623, '\P{^Is_Blk=- Sup_ARROWS_B}', "");
Expect(0, 10624, '\p{Is_Blk=- Sup_ARROWS_B}', "");
Expect(1, 10624, '\p{^Is_Blk=- Sup_ARROWS_B}', "");
Expect(1, 10624, '\P{Is_Blk=- Sup_ARROWS_B}', "");
Expect(0, 10624, '\P{^Is_Blk=- Sup_ARROWS_B}', "");
Error('\p{Block=/a/_ Supplemental_Mathematical_Operators}');
Error('\P{Block=/a/_ Supplemental_Mathematical_Operators}');
Expect(1, 11007, '\p{Block=supplementalmathematicaloperators}', "");
Expect(0, 11007, '\p{^Block=supplementalmathematicaloperators}', "");
Expect(0, 11007, '\P{Block=supplementalmathematicaloperators}', "");
Expect(1, 11007, '\P{^Block=supplementalmathematicaloperators}', "");
Expect(0, 11008, '\p{Block=supplementalmathematicaloperators}', "");
Expect(1, 11008, '\p{^Block=supplementalmathematicaloperators}', "");
Expect(1, 11008, '\P{Block=supplementalmathematicaloperators}', "");
Expect(0, 11008, '\P{^Block=supplementalmathematicaloperators}', "");
Expect(1, 11007, '\p{Block=SUPPLEMENTAL_Mathematical_OPERATORS}', "");
Expect(0, 11007, '\p{^Block=SUPPLEMENTAL_Mathematical_OPERATORS}', "");
Expect(0, 11007, '\P{Block=SUPPLEMENTAL_Mathematical_OPERATORS}', "");
Expect(1, 11007, '\P{^Block=SUPPLEMENTAL_Mathematical_OPERATORS}', "");
Expect(0, 11008, '\p{Block=SUPPLEMENTAL_Mathematical_OPERATORS}', "");
Expect(1, 11008, '\p{^Block=SUPPLEMENTAL_Mathematical_OPERATORS}', "");
Expect(1, 11008, '\P{Block=SUPPLEMENTAL_Mathematical_OPERATORS}', "");
Expect(0, 11008, '\P{^Block=SUPPLEMENTAL_Mathematical_OPERATORS}', "");
Error('\p{Blk=--Sup_Math_Operators:=}');
Error('\P{Blk=--Sup_Math_Operators:=}');
Expect(1, 11007, '\p{Blk=supmathoperators}', "");
Expect(0, 11007, '\p{^Blk=supmathoperators}', "");
Expect(0, 11007, '\P{Blk=supmathoperators}', "");
Expect(1, 11007, '\P{^Blk=supmathoperators}', "");
Expect(0, 11008, '\p{Blk=supmathoperators}', "");
Expect(1, 11008, '\p{^Blk=supmathoperators}', "");
Expect(1, 11008, '\P{Blk=supmathoperators}', "");
Expect(0, 11008, '\P{^Blk=supmathoperators}', "");
Expect(1, 11007, '\p{Blk=		Sup_MATH_operators}', "");
Expect(0, 11007, '\p{^Blk=		Sup_MATH_operators}', "");
Expect(0, 11007, '\P{Blk=		Sup_MATH_operators}', "");
Expect(1, 11007, '\P{^Blk=		Sup_MATH_operators}', "");
Expect(0, 11008, '\p{Blk=		Sup_MATH_operators}', "");
Expect(1, 11008, '\p{^Blk=		Sup_MATH_operators}', "");
Expect(1, 11008, '\P{Blk=		Sup_MATH_operators}', "");
Expect(0, 11008, '\P{^Blk=		Sup_MATH_operators}', "");
Error('\p{Is_Block:	 /a/Supplemental_Mathematical_Operators}');
Error('\P{Is_Block:	 /a/Supplemental_Mathematical_Operators}');
Expect(1, 11007, '\p{Is_Block=supplementalmathematicaloperators}', "");
Expect(0, 11007, '\p{^Is_Block=supplementalmathematicaloperators}', "");
Expect(0, 11007, '\P{Is_Block=supplementalmathematicaloperators}', "");
Expect(1, 11007, '\P{^Is_Block=supplementalmathematicaloperators}', "");
Expect(0, 11008, '\p{Is_Block=supplementalmathematicaloperators}', "");
Expect(1, 11008, '\p{^Is_Block=supplementalmathematicaloperators}', "");
Expect(1, 11008, '\P{Is_Block=supplementalmathematicaloperators}', "");
Expect(0, 11008, '\P{^Is_Block=supplementalmathematicaloperators}', "");
Expect(1, 11007, '\p{Is_Block=_-supplemental_mathematical_OPERATORS}', "");
Expect(0, 11007, '\p{^Is_Block=_-supplemental_mathematical_OPERATORS}', "");
Expect(0, 11007, '\P{Is_Block=_-supplemental_mathematical_OPERATORS}', "");
Expect(1, 11007, '\P{^Is_Block=_-supplemental_mathematical_OPERATORS}', "");
Expect(0, 11008, '\p{Is_Block=_-supplemental_mathematical_OPERATORS}', "");
Expect(1, 11008, '\p{^Is_Block=_-supplemental_mathematical_OPERATORS}', "");
Expect(1, 11008, '\P{Is_Block=_-supplemental_mathematical_OPERATORS}', "");
Expect(0, 11008, '\P{^Is_Block=_-supplemental_mathematical_OPERATORS}', "");
Error('\p{Is_Blk: /a/_-sup_Math_Operators}');
Error('\P{Is_Blk: /a/_-sup_Math_Operators}');
Expect(1, 11007, '\p{Is_Blk=supmathoperators}', "");
Expect(0, 11007, '\p{^Is_Blk=supmathoperators}', "");
Expect(0, 11007, '\P{Is_Blk=supmathoperators}', "");
Expect(1, 11007, '\P{^Is_Blk=supmathoperators}', "");
Expect(0, 11008, '\p{Is_Blk=supmathoperators}', "");
Expect(1, 11008, '\p{^Is_Blk=supmathoperators}', "");
Expect(1, 11008, '\P{Is_Blk=supmathoperators}', "");
Expect(0, 11008, '\P{^Is_Blk=supmathoperators}', "");
Expect(1, 11007, '\p{Is_Blk=-SUP_MATH_operators}', "");
Expect(0, 11007, '\p{^Is_Blk=-SUP_MATH_operators}', "");
Expect(0, 11007, '\P{Is_Blk=-SUP_MATH_operators}', "");
Expect(1, 11007, '\P{^Is_Blk=-SUP_MATH_operators}', "");
Expect(0, 11008, '\p{Is_Blk=-SUP_MATH_operators}', "");
Expect(1, 11008, '\p{^Is_Blk=-SUP_MATH_operators}', "");
Expect(1, 11008, '\P{Is_Blk=-SUP_MATH_operators}', "");
Expect(0, 11008, '\P{^Is_Blk=-SUP_MATH_operators}', "");
Error('\p{Block=	_Supplementary_PRIVATE_USE_Area_A/a/}');
Error('\P{Block=	_Supplementary_PRIVATE_USE_Area_A/a/}');
Expect(1, 983040, '\p{Block=supplementaryprivateuseareaa}', "");
Expect(0, 983040, '\p{^Block=supplementaryprivateuseareaa}', "");
Expect(0, 983040, '\P{Block=supplementaryprivateuseareaa}', "");
Expect(1, 983040, '\P{^Block=supplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\p{Block=supplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\p{^Block=supplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\P{Block=supplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\P{^Block=supplementaryprivateuseareaa}', "");
Expect(1, 983040, '\p{Block=_-supplementary_Private_use_area_a}', "");
Expect(0, 983040, '\p{^Block=_-supplementary_Private_use_area_a}', "");
Expect(0, 983040, '\P{Block=_-supplementary_Private_use_area_a}', "");
Expect(1, 983040, '\P{^Block=_-supplementary_Private_use_area_a}', "");
Expect(0, 1048576, '\p{Block=_-supplementary_Private_use_area_a}', "");
Expect(1, 1048576, '\p{^Block=_-supplementary_Private_use_area_a}', "");
Expect(1, 1048576, '\P{Block=_-supplementary_Private_use_area_a}', "");
Expect(0, 1048576, '\P{^Block=_-supplementary_Private_use_area_a}', "");
Error('\p{Blk=-Sup_pua_A:=}');
Error('\P{Blk=-Sup_pua_A:=}');
Expect(1, 983040, '\p{Blk=suppuaa}', "");
Expect(0, 983040, '\p{^Blk=suppuaa}', "");
Expect(0, 983040, '\P{Blk=suppuaa}', "");
Expect(1, 983040, '\P{^Blk=suppuaa}', "");
Expect(0, 1048576, '\p{Blk=suppuaa}', "");
Expect(1, 1048576, '\p{^Blk=suppuaa}', "");
Expect(1, 1048576, '\P{Blk=suppuaa}', "");
Expect(0, 1048576, '\P{^Blk=suppuaa}', "");
Expect(1, 983040, '\p{Blk=	_Sup_PUA_A}', "");
Expect(0, 983040, '\p{^Blk=	_Sup_PUA_A}', "");
Expect(0, 983040, '\P{Blk=	_Sup_PUA_A}', "");
Expect(1, 983040, '\P{^Blk=	_Sup_PUA_A}', "");
Expect(0, 1048576, '\p{Blk=	_Sup_PUA_A}', "");
Expect(1, 1048576, '\p{^Blk=	_Sup_PUA_A}', "");
Expect(1, 1048576, '\P{Blk=	_Sup_PUA_A}', "");
Expect(0, 1048576, '\P{^Blk=	_Sup_PUA_A}', "");
Error('\p{Is_Block=_/a/Supplementary_PRIVATE_USE_Area_A}');
Error('\P{Is_Block=_/a/Supplementary_PRIVATE_USE_Area_A}');
Expect(1, 983040, '\p{Is_Block=supplementaryprivateuseareaa}', "");
Expect(0, 983040, '\p{^Is_Block=supplementaryprivateuseareaa}', "");
Expect(0, 983040, '\P{Is_Block=supplementaryprivateuseareaa}', "");
Expect(1, 983040, '\P{^Is_Block=supplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\p{Is_Block=supplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\p{^Is_Block=supplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\P{Is_Block=supplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\P{^Is_Block=supplementaryprivateuseareaa}', "");
Expect(1, 983040, '\p{Is_Block=	_Supplementary_Private_Use_area_A}', "");
Expect(0, 983040, '\p{^Is_Block=	_Supplementary_Private_Use_area_A}', "");
Expect(0, 983040, '\P{Is_Block=	_Supplementary_Private_Use_area_A}', "");
Expect(1, 983040, '\P{^Is_Block=	_Supplementary_Private_Use_area_A}', "");
Expect(0, 1048576, '\p{Is_Block=	_Supplementary_Private_Use_area_A}', "");
Expect(1, 1048576, '\p{^Is_Block=	_Supplementary_Private_Use_area_A}', "");
Expect(1, 1048576, '\P{Is_Block=	_Supplementary_Private_Use_area_A}', "");
Expect(0, 1048576, '\P{^Is_Block=	_Supplementary_Private_Use_area_A}', "");
Error('\p{Is_Blk=:=	_SUP_PUA_A}');
Error('\P{Is_Blk=:=	_SUP_PUA_A}');
Expect(1, 983040, '\p{Is_Blk=suppuaa}', "");
Expect(0, 983040, '\p{^Is_Blk=suppuaa}', "");
Expect(0, 983040, '\P{Is_Blk=suppuaa}', "");
Expect(1, 983040, '\P{^Is_Blk=suppuaa}', "");
Expect(0, 1048576, '\p{Is_Blk=suppuaa}', "");
Expect(1, 1048576, '\p{^Is_Blk=suppuaa}', "");
Expect(1, 1048576, '\P{Is_Blk=suppuaa}', "");
Expect(0, 1048576, '\P{^Is_Blk=suppuaa}', "");
Expect(1, 983040, '\p{Is_Blk:	-sup_PUA_A}', "");
Expect(0, 983040, '\p{^Is_Blk:	-sup_PUA_A}', "");
Expect(0, 983040, '\P{Is_Blk:	-sup_PUA_A}', "");
Expect(1, 983040, '\P{^Is_Blk:	-sup_PUA_A}', "");
Expect(0, 1048576, '\p{Is_Blk:	-sup_PUA_A}', "");
Expect(1, 1048576, '\p{^Is_Blk:	-sup_PUA_A}', "");
Expect(1, 1048576, '\P{Is_Blk:	-sup_PUA_A}', "");
Expect(0, 1048576, '\P{^Is_Blk:	-sup_PUA_A}', "");
Error('\p{Block=/a/-Supplementary_private_USE_area_b}');
Error('\P{Block=/a/-Supplementary_private_USE_area_b}');
Expect(1, 1048576, '\p{Block=supplementaryprivateuseareab}', "");
Expect(0, 1048576, '\p{^Block=supplementaryprivateuseareab}', "");
Expect(0, 1048576, '\P{Block=supplementaryprivateuseareab}', "");
Expect(1, 1048576, '\P{^Block=supplementaryprivateuseareab}', "");
Expect(0, 1, '\p{Block=supplementaryprivateuseareab}', "");
Expect(1, 1, '\p{^Block=supplementaryprivateuseareab}', "");
Expect(1, 1, '\P{Block=supplementaryprivateuseareab}', "");
Expect(0, 1, '\P{^Block=supplementaryprivateuseareab}', "");
Expect(1, 1048576, '\p{Block=-Supplementary_Private_USE_AREA_b}', "");
Expect(0, 1048576, '\p{^Block=-Supplementary_Private_USE_AREA_b}', "");
Expect(0, 1048576, '\P{Block=-Supplementary_Private_USE_AREA_b}', "");
Expect(1, 1048576, '\P{^Block=-Supplementary_Private_USE_AREA_b}', "");
Expect(0, 1, '\p{Block=-Supplementary_Private_USE_AREA_b}', "");
Expect(1, 1, '\p{^Block=-Supplementary_Private_USE_AREA_b}', "");
Expect(1, 1, '\P{Block=-Supplementary_Private_USE_AREA_b}', "");
Expect(0, 1, '\P{^Block=-Supplementary_Private_USE_AREA_b}', "");
Error('\p{Blk=:= Sup_pua_b}');
Error('\P{Blk=:= Sup_pua_b}');
Expect(1, 1048576, '\p{Blk:	suppuab}', "");
Expect(0, 1048576, '\p{^Blk:	suppuab}', "");
Expect(0, 1048576, '\P{Blk:	suppuab}', "");
Expect(1, 1048576, '\P{^Blk:	suppuab}', "");
Expect(0, 1, '\p{Blk:	suppuab}', "");
Expect(1, 1, '\p{^Blk:	suppuab}', "");
Expect(1, 1, '\P{Blk:	suppuab}', "");
Expect(0, 1, '\P{^Blk:	suppuab}', "");
Expect(1, 1048576, '\p{Blk=		sup_PUA_B}', "");
Expect(0, 1048576, '\p{^Blk=		sup_PUA_B}', "");
Expect(0, 1048576, '\P{Blk=		sup_PUA_B}', "");
Expect(1, 1048576, '\P{^Blk=		sup_PUA_B}', "");
Expect(0, 1, '\p{Blk=		sup_PUA_B}', "");
Expect(1, 1, '\p{^Blk=		sup_PUA_B}', "");
Expect(1, 1, '\P{Blk=		sup_PUA_B}', "");
Expect(0, 1, '\P{^Blk=		sup_PUA_B}', "");
Error('\p{Is_Block= /a/Supplementary_Private_Use_area_B}');
Error('\P{Is_Block= /a/Supplementary_Private_Use_area_B}');
Expect(1, 1048576, '\p{Is_Block=supplementaryprivateuseareab}', "");
Expect(0, 1048576, '\p{^Is_Block=supplementaryprivateuseareab}', "");
Expect(0, 1048576, '\P{Is_Block=supplementaryprivateuseareab}', "");
Expect(1, 1048576, '\P{^Is_Block=supplementaryprivateuseareab}', "");
Expect(0, 1, '\p{Is_Block=supplementaryprivateuseareab}', "");
Expect(1, 1, '\p{^Is_Block=supplementaryprivateuseareab}', "");
Expect(1, 1, '\P{Is_Block=supplementaryprivateuseareab}', "");
Expect(0, 1, '\P{^Is_Block=supplementaryprivateuseareab}', "");
Expect(1, 1048576, '\p{Is_Block=	Supplementary_Private_USE_area_b}', "");
Expect(0, 1048576, '\p{^Is_Block=	Supplementary_Private_USE_area_b}', "");
Expect(0, 1048576, '\P{Is_Block=	Supplementary_Private_USE_area_b}', "");
Expect(1, 1048576, '\P{^Is_Block=	Supplementary_Private_USE_area_b}', "");
Expect(0, 1, '\p{Is_Block=	Supplementary_Private_USE_area_b}', "");
Expect(1, 1, '\p{^Is_Block=	Supplementary_Private_USE_area_b}', "");
Expect(1, 1, '\P{Is_Block=	Supplementary_Private_USE_area_b}', "");
Expect(0, 1, '\P{^Is_Block=	Supplementary_Private_USE_area_b}', "");
Error('\p{Is_Blk=--sup_PUA_B:=}');
Error('\P{Is_Blk=--sup_PUA_B:=}');
Expect(1, 1048576, '\p{Is_Blk=suppuab}', "");
Expect(0, 1048576, '\p{^Is_Blk=suppuab}', "");
Expect(0, 1048576, '\P{Is_Blk=suppuab}', "");
Expect(1, 1048576, '\P{^Is_Blk=suppuab}', "");
Expect(0, 1, '\p{Is_Blk=suppuab}', "");
Expect(1, 1, '\p{^Is_Blk=suppuab}', "");
Expect(1, 1, '\P{Is_Blk=suppuab}', "");
Expect(0, 1, '\P{^Is_Blk=suppuab}', "");
Expect(1, 1048576, '\p{Is_Blk= -sup_pua_B}', "");
Expect(0, 1048576, '\p{^Is_Blk= -sup_pua_B}', "");
Expect(0, 1048576, '\P{Is_Blk= -sup_pua_B}', "");
Expect(1, 1048576, '\P{^Is_Blk= -sup_pua_B}', "");
Expect(0, 1, '\p{Is_Blk= -sup_pua_B}', "");
Expect(1, 1, '\p{^Is_Blk= -sup_pua_B}', "");
Expect(1, 1, '\P{Is_Blk= -sup_pua_B}', "");
Expect(0, 1, '\P{^Is_Blk= -sup_pua_B}', "");
Error('\p{Block=:= -supplemental_Punctuation}');
Error('\P{Block=:= -supplemental_Punctuation}');
Expect(1, 11903, '\p{Block=supplementalpunctuation}', "");
Expect(0, 11903, '\p{^Block=supplementalpunctuation}', "");
Expect(0, 11903, '\P{Block=supplementalpunctuation}', "");
Expect(1, 11903, '\P{^Block=supplementalpunctuation}', "");
Expect(0, 11904, '\p{Block=supplementalpunctuation}', "");
Expect(1, 11904, '\p{^Block=supplementalpunctuation}', "");
Expect(1, 11904, '\P{Block=supplementalpunctuation}', "");
Expect(0, 11904, '\P{^Block=supplementalpunctuation}', "");
Expect(1, 11903, '\p{Block=	Supplemental_Punctuation}', "");
Expect(0, 11903, '\p{^Block=	Supplemental_Punctuation}', "");
Expect(0, 11903, '\P{Block=	Supplemental_Punctuation}', "");
Expect(1, 11903, '\P{^Block=	Supplemental_Punctuation}', "");
Expect(0, 11904, '\p{Block=	Supplemental_Punctuation}', "");
Expect(1, 11904, '\p{^Block=	Supplemental_Punctuation}', "");
Expect(1, 11904, '\P{Block=	Supplemental_Punctuation}', "");
Expect(0, 11904, '\P{^Block=	Supplemental_Punctuation}', "");
Error('\p{Blk=__sup_PUNCTUATION:=}');
Error('\P{Blk=__sup_PUNCTUATION:=}');
Expect(1, 11903, '\p{Blk=suppunctuation}', "");
Expect(0, 11903, '\p{^Blk=suppunctuation}', "");
Expect(0, 11903, '\P{Blk=suppunctuation}', "");
Expect(1, 11903, '\P{^Blk=suppunctuation}', "");
Expect(0, 11904, '\p{Blk=suppunctuation}', "");
Expect(1, 11904, '\p{^Blk=suppunctuation}', "");
Expect(1, 11904, '\P{Blk=suppunctuation}', "");
Expect(0, 11904, '\P{^Blk=suppunctuation}', "");
Expect(1, 11903, '\p{Blk=		Sup_Punctuation}', "");
Expect(0, 11903, '\p{^Blk=		Sup_Punctuation}', "");
Expect(0, 11903, '\P{Blk=		Sup_Punctuation}', "");
Expect(1, 11903, '\P{^Blk=		Sup_Punctuation}', "");
Expect(0, 11904, '\p{Blk=		Sup_Punctuation}', "");
Expect(1, 11904, '\p{^Blk=		Sup_Punctuation}', "");
Expect(1, 11904, '\P{Blk=		Sup_Punctuation}', "");
Expect(0, 11904, '\P{^Blk=		Sup_Punctuation}', "");
Error('\p{Is_Block=-:=Supplemental_Punctuation}');
Error('\P{Is_Block=-:=Supplemental_Punctuation}');
Expect(1, 11903, '\p{Is_Block=supplementalpunctuation}', "");
Expect(0, 11903, '\p{^Is_Block=supplementalpunctuation}', "");
Expect(0, 11903, '\P{Is_Block=supplementalpunctuation}', "");
Expect(1, 11903, '\P{^Is_Block=supplementalpunctuation}', "");
Expect(0, 11904, '\p{Is_Block=supplementalpunctuation}', "");
Expect(1, 11904, '\p{^Is_Block=supplementalpunctuation}', "");
Expect(1, 11904, '\P{Is_Block=supplementalpunctuation}', "");
Expect(0, 11904, '\P{^Is_Block=supplementalpunctuation}', "");
Expect(1, 11903, '\p{Is_Block=-_SUPPLEMENTAL_Punctuation}', "");
Expect(0, 11903, '\p{^Is_Block=-_SUPPLEMENTAL_Punctuation}', "");
Expect(0, 11903, '\P{Is_Block=-_SUPPLEMENTAL_Punctuation}', "");
Expect(1, 11903, '\P{^Is_Block=-_SUPPLEMENTAL_Punctuation}', "");
Expect(0, 11904, '\p{Is_Block=-_SUPPLEMENTAL_Punctuation}', "");
Expect(1, 11904, '\p{^Is_Block=-_SUPPLEMENTAL_Punctuation}', "");
Expect(1, 11904, '\P{Is_Block=-_SUPPLEMENTAL_Punctuation}', "");
Expect(0, 11904, '\P{^Is_Block=-_SUPPLEMENTAL_Punctuation}', "");
Error('\p{Is_Blk=- sup_punctuation:=}');
Error('\P{Is_Blk=- sup_punctuation:=}');
Expect(1, 11903, '\p{Is_Blk=suppunctuation}', "");
Expect(0, 11903, '\p{^Is_Blk=suppunctuation}', "");
Expect(0, 11903, '\P{Is_Blk=suppunctuation}', "");
Expect(1, 11903, '\P{^Is_Blk=suppunctuation}', "");
Expect(0, 11904, '\p{Is_Blk=suppunctuation}', "");
Expect(1, 11904, '\p{^Is_Blk=suppunctuation}', "");
Expect(1, 11904, '\P{Is_Blk=suppunctuation}', "");
Expect(0, 11904, '\P{^Is_Blk=suppunctuation}', "");
Expect(1, 11903, '\p{Is_Blk=-Sup_Punctuation}', "");
Expect(0, 11903, '\p{^Is_Blk=-Sup_Punctuation}', "");
Expect(0, 11903, '\P{Is_Blk=-Sup_Punctuation}', "");
Expect(1, 11903, '\P{^Is_Blk=-Sup_Punctuation}', "");
Expect(0, 11904, '\p{Is_Blk=-Sup_Punctuation}', "");
Expect(1, 11904, '\p{^Is_Blk=-Sup_Punctuation}', "");
Expect(1, 11904, '\P{Is_Blk=-Sup_Punctuation}', "");
Expect(0, 11904, '\P{^Is_Blk=-Sup_Punctuation}', "");
Error('\p{Block=/a/-	SUPERSCRIPTS_And_Subscripts}');
Error('\P{Block=/a/-	SUPERSCRIPTS_And_Subscripts}');
Expect(1, 8351, '\p{Block=superscriptsandsubscripts}', "");
Expect(0, 8351, '\p{^Block=superscriptsandsubscripts}', "");
Expect(0, 8351, '\P{Block=superscriptsandsubscripts}', "");
Expect(1, 8351, '\P{^Block=superscriptsandsubscripts}', "");
Expect(0, 8352, '\p{Block=superscriptsandsubscripts}', "");
Expect(1, 8352, '\p{^Block=superscriptsandsubscripts}', "");
Expect(1, 8352, '\P{Block=superscriptsandsubscripts}', "");
Expect(0, 8352, '\P{^Block=superscriptsandsubscripts}', "");
Expect(1, 8351, '\p{Block=_SUPERSCRIPTS_And_SUBSCRIPTS}', "");
Expect(0, 8351, '\p{^Block=_SUPERSCRIPTS_And_SUBSCRIPTS}', "");
Expect(0, 8351, '\P{Block=_SUPERSCRIPTS_And_SUBSCRIPTS}', "");
Expect(1, 8351, '\P{^Block=_SUPERSCRIPTS_And_SUBSCRIPTS}', "");
Expect(0, 8352, '\p{Block=_SUPERSCRIPTS_And_SUBSCRIPTS}', "");
Expect(1, 8352, '\p{^Block=_SUPERSCRIPTS_And_SUBSCRIPTS}', "");
Expect(1, 8352, '\P{Block=_SUPERSCRIPTS_And_SUBSCRIPTS}', "");
Expect(0, 8352, '\P{^Block=_SUPERSCRIPTS_And_SUBSCRIPTS}', "");
Error('\p{Blk=-:=Super_AND_Sub}');
Error('\P{Blk=-:=Super_AND_Sub}');
Expect(1, 8351, '\p{Blk=superandsub}', "");
Expect(0, 8351, '\p{^Blk=superandsub}', "");
Expect(0, 8351, '\P{Blk=superandsub}', "");
Expect(1, 8351, '\P{^Blk=superandsub}', "");
Expect(0, 8352, '\p{Blk=superandsub}', "");
Expect(1, 8352, '\p{^Blk=superandsub}', "");
Expect(1, 8352, '\P{Blk=superandsub}', "");
Expect(0, 8352, '\P{^Blk=superandsub}', "");
Expect(1, 8351, '\p{Blk=-Super_And_sub}', "");
Expect(0, 8351, '\p{^Blk=-Super_And_sub}', "");
Expect(0, 8351, '\P{Blk=-Super_And_sub}', "");
Expect(1, 8351, '\P{^Blk=-Super_And_sub}', "");
Expect(0, 8352, '\p{Blk=-Super_And_sub}', "");
Expect(1, 8352, '\p{^Blk=-Super_And_sub}', "");
Expect(1, 8352, '\P{Blk=-Super_And_sub}', "");
Expect(0, 8352, '\P{^Blk=-Super_And_sub}', "");
Error('\p{Is_Block=/a/	_superscripts_AND_subscripts}');
Error('\P{Is_Block=/a/	_superscripts_AND_subscripts}');
Expect(1, 8351, '\p{Is_Block=superscriptsandsubscripts}', "");
Expect(0, 8351, '\p{^Is_Block=superscriptsandsubscripts}', "");
Expect(0, 8351, '\P{Is_Block=superscriptsandsubscripts}', "");
Expect(1, 8351, '\P{^Is_Block=superscriptsandsubscripts}', "");
Expect(0, 8352, '\p{Is_Block=superscriptsandsubscripts}', "");
Expect(1, 8352, '\p{^Is_Block=superscriptsandsubscripts}', "");
Expect(1, 8352, '\P{Is_Block=superscriptsandsubscripts}', "");
Expect(0, 8352, '\P{^Is_Block=superscriptsandsubscripts}', "");
Expect(1, 8351, '\p{Is_Block= 	Superscripts_And_subscripts}', "");
Expect(0, 8351, '\p{^Is_Block= 	Superscripts_And_subscripts}', "");
Expect(0, 8351, '\P{Is_Block= 	Superscripts_And_subscripts}', "");
Expect(1, 8351, '\P{^Is_Block= 	Superscripts_And_subscripts}', "");
Expect(0, 8352, '\p{Is_Block= 	Superscripts_And_subscripts}', "");
Expect(1, 8352, '\p{^Is_Block= 	Superscripts_And_subscripts}', "");
Expect(1, 8352, '\P{Is_Block= 	Superscripts_And_subscripts}', "");
Expect(0, 8352, '\P{^Is_Block= 	Superscripts_And_subscripts}', "");
Error('\p{Is_Blk=_ SUPER_and_SUB/a/}');
Error('\P{Is_Blk=_ SUPER_and_SUB/a/}');
Expect(1, 8351, '\p{Is_Blk=superandsub}', "");
Expect(0, 8351, '\p{^Is_Blk=superandsub}', "");
Expect(0, 8351, '\P{Is_Blk=superandsub}', "");
Expect(1, 8351, '\P{^Is_Blk=superandsub}', "");
Expect(0, 8352, '\p{Is_Blk=superandsub}', "");
Expect(1, 8352, '\p{^Is_Blk=superandsub}', "");
Expect(1, 8352, '\P{Is_Blk=superandsub}', "");
Expect(0, 8352, '\P{^Is_Blk=superandsub}', "");
Expect(1, 8351, '\p{Is_Blk=	 Super_and_SUB}', "");
Expect(0, 8351, '\p{^Is_Blk=	 Super_and_SUB}', "");
Expect(0, 8351, '\P{Is_Blk=	 Super_and_SUB}', "");
Expect(1, 8351, '\P{^Is_Blk=	 Super_and_SUB}', "");
Expect(0, 8352, '\p{Is_Blk=	 Super_and_SUB}', "");
Expect(1, 8352, '\p{^Is_Blk=	 Super_and_SUB}', "");
Expect(1, 8352, '\P{Is_Blk=	 Super_and_SUB}', "");
Expect(0, 8352, '\P{^Is_Blk=	 Super_and_SUB}', "");
Error('\p{Block=/a/--SYLOTI_Nagri}');
Error('\P{Block=/a/--SYLOTI_Nagri}');
Expect(1, 43055, '\p{Block=sylotinagri}', "");
Expect(0, 43055, '\p{^Block=sylotinagri}', "");
Expect(0, 43055, '\P{Block=sylotinagri}', "");
Expect(1, 43055, '\P{^Block=sylotinagri}', "");
Expect(0, 43056, '\p{Block=sylotinagri}', "");
Expect(1, 43056, '\p{^Block=sylotinagri}', "");
Expect(1, 43056, '\P{Block=sylotinagri}', "");
Expect(0, 43056, '\P{^Block=sylotinagri}', "");
Expect(1, 43055, '\p{Block=-	Syloti_Nagri}', "");
Expect(0, 43055, '\p{^Block=-	Syloti_Nagri}', "");
Expect(0, 43055, '\P{Block=-	Syloti_Nagri}', "");
Expect(1, 43055, '\P{^Block=-	Syloti_Nagri}', "");
Expect(0, 43056, '\p{Block=-	Syloti_Nagri}', "");
Expect(1, 43056, '\p{^Block=-	Syloti_Nagri}', "");
Expect(1, 43056, '\P{Block=-	Syloti_Nagri}', "");
Expect(0, 43056, '\P{^Block=-	Syloti_Nagri}', "");
Error('\p{Blk=	_syloti_NAGRI/a/}');
Error('\P{Blk=	_syloti_NAGRI/a/}');
Expect(1, 43055, '\p{Blk=sylotinagri}', "");
Expect(0, 43055, '\p{^Blk=sylotinagri}', "");
Expect(0, 43055, '\P{Blk=sylotinagri}', "");
Expect(1, 43055, '\P{^Blk=sylotinagri}', "");
Expect(0, 43056, '\p{Blk=sylotinagri}', "");
Expect(1, 43056, '\p{^Blk=sylotinagri}', "");
Expect(1, 43056, '\P{Blk=sylotinagri}', "");
Expect(0, 43056, '\P{^Blk=sylotinagri}', "");
Expect(1, 43055, '\p{Blk:    SYLOTI_Nagri}', "");
Expect(0, 43055, '\p{^Blk:    SYLOTI_Nagri}', "");
Expect(0, 43055, '\P{Blk:    SYLOTI_Nagri}', "");
Expect(1, 43055, '\P{^Blk:    SYLOTI_Nagri}', "");
Expect(0, 43056, '\p{Blk:    SYLOTI_Nagri}', "");
Expect(1, 43056, '\p{^Blk:    SYLOTI_Nagri}', "");
Expect(1, 43056, '\P{Blk:    SYLOTI_Nagri}', "");
Expect(0, 43056, '\P{^Blk:    SYLOTI_Nagri}', "");
Error('\p{Is_Block=-syloti_Nagri:=}');
Error('\P{Is_Block=-syloti_Nagri:=}');
Expect(1, 43055, '\p{Is_Block=sylotinagri}', "");
Expect(0, 43055, '\p{^Is_Block=sylotinagri}', "");
Expect(0, 43055, '\P{Is_Block=sylotinagri}', "");
Expect(1, 43055, '\P{^Is_Block=sylotinagri}', "");
Expect(0, 43056, '\p{Is_Block=sylotinagri}', "");
Expect(1, 43056, '\p{^Is_Block=sylotinagri}', "");
Expect(1, 43056, '\P{Is_Block=sylotinagri}', "");
Expect(0, 43056, '\P{^Is_Block=sylotinagri}', "");
Expect(1, 43055, '\p{Is_Block=-	Syloti_Nagri}', "");
Expect(0, 43055, '\p{^Is_Block=-	Syloti_Nagri}', "");
Expect(0, 43055, '\P{Is_Block=-	Syloti_Nagri}', "");
Expect(1, 43055, '\P{^Is_Block=-	Syloti_Nagri}', "");
Expect(0, 43056, '\p{Is_Block=-	Syloti_Nagri}', "");
Expect(1, 43056, '\p{^Is_Block=-	Syloti_Nagri}', "");
Expect(1, 43056, '\P{Is_Block=-	Syloti_Nagri}', "");
Expect(0, 43056, '\P{^Is_Block=-	Syloti_Nagri}', "");
Error('\p{Is_Blk=:=- Syloti_Nagri}');
Error('\P{Is_Blk=:=- Syloti_Nagri}');
Expect(1, 43055, '\p{Is_Blk=sylotinagri}', "");
Expect(0, 43055, '\p{^Is_Blk=sylotinagri}', "");
Expect(0, 43055, '\P{Is_Blk=sylotinagri}', "");
Expect(1, 43055, '\P{^Is_Blk=sylotinagri}', "");
Expect(0, 43056, '\p{Is_Blk=sylotinagri}', "");
Expect(1, 43056, '\p{^Is_Blk=sylotinagri}', "");
Expect(1, 43056, '\P{Is_Blk=sylotinagri}', "");
Expect(0, 43056, '\P{^Is_Blk=sylotinagri}', "");
Expect(1, 43055, '\p{Is_Blk=_ SYLOTI_NAGRI}', "");
Expect(0, 43055, '\p{^Is_Blk=_ SYLOTI_NAGRI}', "");
Expect(0, 43055, '\P{Is_Blk=_ SYLOTI_NAGRI}', "");
Expect(1, 43055, '\P{^Is_Blk=_ SYLOTI_NAGRI}', "");
Expect(0, 43056, '\p{Is_Blk=_ SYLOTI_NAGRI}', "");
Expect(1, 43056, '\p{^Is_Blk=_ SYLOTI_NAGRI}', "");
Expect(1, 43056, '\P{Is_Blk=_ SYLOTI_NAGRI}', "");
Expect(0, 43056, '\P{^Is_Blk=_ SYLOTI_NAGRI}', "");
Error('\p{Block=	:=Syriac}');
Error('\P{Block=	:=Syriac}');
Expect(1, 1871, '\p{Block=syriac}', "");
Expect(0, 1871, '\p{^Block=syriac}', "");
Expect(0, 1871, '\P{Block=syriac}', "");
Expect(1, 1871, '\P{^Block=syriac}', "");
Expect(0, 1872, '\p{Block=syriac}', "");
Expect(1, 1872, '\p{^Block=syriac}', "");
Expect(1, 1872, '\P{Block=syriac}', "");
Expect(0, 1872, '\P{^Block=syriac}', "");
Expect(1, 1871, '\p{Block=_syriac}', "");
Expect(0, 1871, '\p{^Block=_syriac}', "");
Expect(0, 1871, '\P{Block=_syriac}', "");
Expect(1, 1871, '\P{^Block=_syriac}', "");
Expect(0, 1872, '\p{Block=_syriac}', "");
Expect(1, 1872, '\p{^Block=_syriac}', "");
Expect(1, 1872, '\P{Block=_syriac}', "");
Expect(0, 1872, '\P{^Block=_syriac}', "");
Error('\p{Blk= 	SYRIAC:=}');
Error('\P{Blk= 	SYRIAC:=}');
Expect(1, 1871, '\p{Blk=syriac}', "");
Expect(0, 1871, '\p{^Blk=syriac}', "");
Expect(0, 1871, '\P{Blk=syriac}', "");
Expect(1, 1871, '\P{^Blk=syriac}', "");
Expect(0, 1872, '\p{Blk=syriac}', "");
Expect(1, 1872, '\p{^Blk=syriac}', "");
Expect(1, 1872, '\P{Blk=syriac}', "");
Expect(0, 1872, '\P{^Blk=syriac}', "");
Expect(1, 1871, '\p{Blk=- SYRIAC}', "");
Expect(0, 1871, '\p{^Blk=- SYRIAC}', "");
Expect(0, 1871, '\P{Blk=- SYRIAC}', "");
Expect(1, 1871, '\P{^Blk=- SYRIAC}', "");
Expect(0, 1872, '\p{Blk=- SYRIAC}', "");
Expect(1, 1872, '\p{^Blk=- SYRIAC}', "");
Expect(1, 1872, '\P{Blk=- SYRIAC}', "");
Expect(0, 1872, '\P{^Blk=- SYRIAC}', "");
Error('\p{Is_Block= _syriac:=}');
Error('\P{Is_Block= _syriac:=}');
Expect(1, 1871, '\p{Is_Block=syriac}', "");
Expect(0, 1871, '\p{^Is_Block=syriac}', "");
Expect(0, 1871, '\P{Is_Block=syriac}', "");
Expect(1, 1871, '\P{^Is_Block=syriac}', "");
Expect(0, 1872, '\p{Is_Block=syriac}', "");
Expect(1, 1872, '\p{^Is_Block=syriac}', "");
Expect(1, 1872, '\P{Is_Block=syriac}', "");
Expect(0, 1872, '\P{^Is_Block=syriac}', "");
Expect(1, 1871, '\p{Is_Block=-Syriac}', "");
Expect(0, 1871, '\p{^Is_Block=-Syriac}', "");
Expect(0, 1871, '\P{Is_Block=-Syriac}', "");
Expect(1, 1871, '\P{^Is_Block=-Syriac}', "");
Expect(0, 1872, '\p{Is_Block=-Syriac}', "");
Expect(1, 1872, '\p{^Is_Block=-Syriac}', "");
Expect(1, 1872, '\P{Is_Block=-Syriac}', "");
Expect(0, 1872, '\P{^Is_Block=-Syriac}', "");
Error('\p{Is_Blk=_Syriac/a/}');
Error('\P{Is_Blk=_Syriac/a/}');
Expect(1, 1871, '\p{Is_Blk=syriac}', "");
Expect(0, 1871, '\p{^Is_Blk=syriac}', "");
Expect(0, 1871, '\P{Is_Blk=syriac}', "");
Expect(1, 1871, '\P{^Is_Blk=syriac}', "");
Expect(0, 1872, '\p{Is_Blk=syriac}', "");
Expect(1, 1872, '\p{^Is_Blk=syriac}', "");
Expect(1, 1872, '\P{Is_Blk=syriac}', "");
Expect(0, 1872, '\P{^Is_Blk=syriac}', "");
Expect(1, 1871, '\p{Is_Blk=_-syriac}', "");
Expect(0, 1871, '\p{^Is_Blk=_-syriac}', "");
Expect(0, 1871, '\P{Is_Blk=_-syriac}', "");
Expect(1, 1871, '\P{^Is_Blk=_-syriac}', "");
Expect(0, 1872, '\p{Is_Blk=_-syriac}', "");
Expect(1, 1872, '\p{^Is_Blk=_-syriac}', "");
Expect(1, 1872, '\P{Is_Blk=_-syriac}', "");
Expect(0, 1872, '\P{^Is_Blk=_-syriac}', "");
Error('\p{Block=:=	TAGALOG}');
Error('\P{Block=:=	TAGALOG}');
Expect(1, 5919, '\p{Block=tagalog}', "");
Expect(0, 5919, '\p{^Block=tagalog}', "");
Expect(0, 5919, '\P{Block=tagalog}', "");
Expect(1, 5919, '\P{^Block=tagalog}', "");
Expect(0, 5920, '\p{Block=tagalog}', "");
Expect(1, 5920, '\p{^Block=tagalog}', "");
Expect(1, 5920, '\P{Block=tagalog}', "");
Expect(0, 5920, '\P{^Block=tagalog}', "");
Expect(1, 5919, '\p{Block=  Tagalog}', "");
Expect(0, 5919, '\p{^Block=  Tagalog}', "");
Expect(0, 5919, '\P{Block=  Tagalog}', "");
Expect(1, 5919, '\P{^Block=  Tagalog}', "");
Expect(0, 5920, '\p{Block=  Tagalog}', "");
Expect(1, 5920, '\p{^Block=  Tagalog}', "");
Expect(1, 5920, '\P{Block=  Tagalog}', "");
Expect(0, 5920, '\P{^Block=  Tagalog}', "");
Error('\p{Blk=_ tagalog:=}');
Error('\P{Blk=_ tagalog:=}');
Expect(1, 5919, '\p{Blk=tagalog}', "");
Expect(0, 5919, '\p{^Blk=tagalog}', "");
Expect(0, 5919, '\P{Blk=tagalog}', "");
Expect(1, 5919, '\P{^Blk=tagalog}', "");
Expect(0, 5920, '\p{Blk=tagalog}', "");
Expect(1, 5920, '\p{^Blk=tagalog}', "");
Expect(1, 5920, '\P{Blk=tagalog}', "");
Expect(0, 5920, '\P{^Blk=tagalog}', "");
Expect(1, 5919, '\p{Blk= 	tagalog}', "");
Expect(0, 5919, '\p{^Blk= 	tagalog}', "");
Expect(0, 5919, '\P{Blk= 	tagalog}', "");
Expect(1, 5919, '\P{^Blk= 	tagalog}', "");
Expect(0, 5920, '\p{Blk= 	tagalog}', "");
Expect(1, 5920, '\p{^Blk= 	tagalog}', "");
Expect(1, 5920, '\P{Blk= 	tagalog}', "");
Expect(0, 5920, '\P{^Blk= 	tagalog}', "");
Error('\p{Is_Block=	/a/Tagalog}');
Error('\P{Is_Block=	/a/Tagalog}');
Expect(1, 5919, '\p{Is_Block=tagalog}', "");
Expect(0, 5919, '\p{^Is_Block=tagalog}', "");
Expect(0, 5919, '\P{Is_Block=tagalog}', "");
Expect(1, 5919, '\P{^Is_Block=tagalog}', "");
Expect(0, 5920, '\p{Is_Block=tagalog}', "");
Expect(1, 5920, '\p{^Is_Block=tagalog}', "");
Expect(1, 5920, '\P{Is_Block=tagalog}', "");
Expect(0, 5920, '\P{^Is_Block=tagalog}', "");
Expect(1, 5919, '\p{Is_Block=	_Tagalog}', "");
Expect(0, 5919, '\p{^Is_Block=	_Tagalog}', "");
Expect(0, 5919, '\P{Is_Block=	_Tagalog}', "");
Expect(1, 5919, '\P{^Is_Block=	_Tagalog}', "");
Expect(0, 5920, '\p{Is_Block=	_Tagalog}', "");
Expect(1, 5920, '\p{^Is_Block=	_Tagalog}', "");
Expect(1, 5920, '\P{Is_Block=	_Tagalog}', "");
Expect(0, 5920, '\P{^Is_Block=	_Tagalog}', "");
Error('\p{Is_Blk=/a/_ Tagalog}');
Error('\P{Is_Blk=/a/_ Tagalog}');
Expect(1, 5919, '\p{Is_Blk=tagalog}', "");
Expect(0, 5919, '\p{^Is_Blk=tagalog}', "");
Expect(0, 5919, '\P{Is_Blk=tagalog}', "");
Expect(1, 5919, '\P{^Is_Blk=tagalog}', "");
Expect(0, 5920, '\p{Is_Blk=tagalog}', "");
Expect(1, 5920, '\p{^Is_Blk=tagalog}', "");
Expect(1, 5920, '\P{Is_Blk=tagalog}', "");
Expect(0, 5920, '\P{^Is_Blk=tagalog}', "");
Expect(1, 5919, '\p{Is_Blk= 	Tagalog}', "");
Expect(0, 5919, '\p{^Is_Blk= 	Tagalog}', "");
Expect(0, 5919, '\P{Is_Blk= 	Tagalog}', "");
Expect(1, 5919, '\P{^Is_Blk= 	Tagalog}', "");
Expect(0, 5920, '\p{Is_Blk= 	Tagalog}', "");
Expect(1, 5920, '\p{^Is_Blk= 	Tagalog}', "");
Expect(1, 5920, '\P{Is_Blk= 	Tagalog}', "");
Expect(0, 5920, '\P{^Is_Blk= 	Tagalog}', "");
Error('\p{Block= -TAGBANWA/a/}');
Error('\P{Block= -TAGBANWA/a/}');
Expect(1, 6015, '\p{Block=tagbanwa}', "");
Expect(0, 6015, '\p{^Block=tagbanwa}', "");
Expect(0, 6015, '\P{Block=tagbanwa}', "");
Expect(1, 6015, '\P{^Block=tagbanwa}', "");
Expect(0, 6016, '\p{Block=tagbanwa}', "");
Expect(1, 6016, '\p{^Block=tagbanwa}', "");
Expect(1, 6016, '\P{Block=tagbanwa}', "");
Expect(0, 6016, '\P{^Block=tagbanwa}', "");
Expect(1, 6015, '\p{Block=Tagbanwa}', "");
Expect(0, 6015, '\p{^Block=Tagbanwa}', "");
Expect(0, 6015, '\P{Block=Tagbanwa}', "");
Expect(1, 6015, '\P{^Block=Tagbanwa}', "");
Expect(0, 6016, '\p{Block=Tagbanwa}', "");
Expect(1, 6016, '\p{^Block=Tagbanwa}', "");
Expect(1, 6016, '\P{Block=Tagbanwa}', "");
Expect(0, 6016, '\P{^Block=Tagbanwa}', "");
Error('\p{Blk=__Tagbanwa:=}');
Error('\P{Blk=__Tagbanwa:=}');
Expect(1, 6015, '\p{Blk=tagbanwa}', "");
Expect(0, 6015, '\p{^Blk=tagbanwa}', "");
Expect(0, 6015, '\P{Blk=tagbanwa}', "");
Expect(1, 6015, '\P{^Blk=tagbanwa}', "");
Expect(0, 6016, '\p{Blk=tagbanwa}', "");
Expect(1, 6016, '\p{^Blk=tagbanwa}', "");
Expect(1, 6016, '\P{Blk=tagbanwa}', "");
Expect(0, 6016, '\P{^Blk=tagbanwa}', "");
Expect(1, 6015, '\p{Blk:-TAGBANWA}', "");
Expect(0, 6015, '\p{^Blk:-TAGBANWA}', "");
Expect(0, 6015, '\P{Blk:-TAGBANWA}', "");
Expect(1, 6015, '\P{^Blk:-TAGBANWA}', "");
Expect(0, 6016, '\p{Blk:-TAGBANWA}', "");
Expect(1, 6016, '\p{^Blk:-TAGBANWA}', "");
Expect(1, 6016, '\P{Blk:-TAGBANWA}', "");
Expect(0, 6016, '\P{^Blk:-TAGBANWA}', "");
Error('\p{Is_Block=/a/tagbanwa}');
Error('\P{Is_Block=/a/tagbanwa}');
Expect(1, 6015, '\p{Is_Block=tagbanwa}', "");
Expect(0, 6015, '\p{^Is_Block=tagbanwa}', "");
Expect(0, 6015, '\P{Is_Block=tagbanwa}', "");
Expect(1, 6015, '\P{^Is_Block=tagbanwa}', "");
Expect(0, 6016, '\p{Is_Block=tagbanwa}', "");
Expect(1, 6016, '\p{^Is_Block=tagbanwa}', "");
Expect(1, 6016, '\P{Is_Block=tagbanwa}', "");
Expect(0, 6016, '\P{^Is_Block=tagbanwa}', "");
Expect(1, 6015, '\p{Is_Block= -TAGBANWA}', "");
Expect(0, 6015, '\p{^Is_Block= -TAGBANWA}', "");
Expect(0, 6015, '\P{Is_Block= -TAGBANWA}', "");
Expect(1, 6015, '\P{^Is_Block= -TAGBANWA}', "");
Expect(0, 6016, '\p{Is_Block= -TAGBANWA}', "");
Expect(1, 6016, '\p{^Is_Block= -TAGBANWA}', "");
Expect(1, 6016, '\P{Is_Block= -TAGBANWA}', "");
Expect(0, 6016, '\P{^Is_Block= -TAGBANWA}', "");
Error('\p{Is_Blk=-:=Tagbanwa}');
Error('\P{Is_Blk=-:=Tagbanwa}');
Expect(1, 6015, '\p{Is_Blk=tagbanwa}', "");
Expect(0, 6015, '\p{^Is_Blk=tagbanwa}', "");
Expect(0, 6015, '\P{Is_Blk=tagbanwa}', "");
Expect(1, 6015, '\P{^Is_Blk=tagbanwa}', "");
Expect(0, 6016, '\p{Is_Blk=tagbanwa}', "");
Expect(1, 6016, '\p{^Is_Blk=tagbanwa}', "");
Expect(1, 6016, '\P{Is_Blk=tagbanwa}', "");
Expect(0, 6016, '\P{^Is_Blk=tagbanwa}', "");
Expect(1, 6015, '\p{Is_Blk= 	Tagbanwa}', "");
Expect(0, 6015, '\p{^Is_Blk= 	Tagbanwa}', "");
Expect(0, 6015, '\P{Is_Blk= 	Tagbanwa}', "");
Expect(1, 6015, '\P{^Is_Blk= 	Tagbanwa}', "");
Expect(0, 6016, '\p{Is_Blk= 	Tagbanwa}', "");
Expect(1, 6016, '\p{^Is_Blk= 	Tagbanwa}', "");
Expect(1, 6016, '\P{Is_Blk= 	Tagbanwa}', "");
Expect(0, 6016, '\P{^Is_Blk= 	Tagbanwa}', "");
Error('\p{Block=-:=Tags}');
Error('\P{Block=-:=Tags}');
Expect(1, 917631, '\p{Block: tags}', "");
Expect(0, 917631, '\p{^Block: tags}', "");
Expect(0, 917631, '\P{Block: tags}', "");
Expect(1, 917631, '\P{^Block: tags}', "");
Expect(0, 917632, '\p{Block: tags}', "");
Expect(1, 917632, '\p{^Block: tags}', "");
Expect(1, 917632, '\P{Block: tags}', "");
Expect(0, 917632, '\P{^Block: tags}', "");
Expect(1, 917631, '\p{Block=_-Tags}', "");
Expect(0, 917631, '\p{^Block=_-Tags}', "");
Expect(0, 917631, '\P{Block=_-Tags}', "");
Expect(1, 917631, '\P{^Block=_-Tags}', "");
Expect(0, 917632, '\p{Block=_-Tags}', "");
Expect(1, 917632, '\p{^Block=_-Tags}', "");
Expect(1, 917632, '\P{Block=_-Tags}', "");
Expect(0, 917632, '\P{^Block=_-Tags}', "");
Error('\p{Blk=_:=tags}');
Error('\P{Blk=_:=tags}');
Expect(1, 917631, '\p{Blk:	tags}', "");
Expect(0, 917631, '\p{^Blk:	tags}', "");
Expect(0, 917631, '\P{Blk:	tags}', "");
Expect(1, 917631, '\P{^Blk:	tags}', "");
Expect(0, 917632, '\p{Blk:	tags}', "");
Expect(1, 917632, '\p{^Blk:	tags}', "");
Expect(1, 917632, '\P{Blk:	tags}', "");
Expect(0, 917632, '\P{^Blk:	tags}', "");
Expect(1, 917631, '\p{Blk:	--TAGS}', "");
Expect(0, 917631, '\p{^Blk:	--TAGS}', "");
Expect(0, 917631, '\P{Blk:	--TAGS}', "");
Expect(1, 917631, '\P{^Blk:	--TAGS}', "");
Expect(0, 917632, '\p{Blk:	--TAGS}', "");
Expect(1, 917632, '\p{^Blk:	--TAGS}', "");
Expect(1, 917632, '\P{Blk:	--TAGS}', "");
Expect(0, 917632, '\P{^Blk:	--TAGS}', "");
Error('\p{Is_Block= -Tags/a/}');
Error('\P{Is_Block= -Tags/a/}');
Expect(1, 917631, '\p{Is_Block=tags}', "");
Expect(0, 917631, '\p{^Is_Block=tags}', "");
Expect(0, 917631, '\P{Is_Block=tags}', "");
Expect(1, 917631, '\P{^Is_Block=tags}', "");
Expect(0, 917632, '\p{Is_Block=tags}', "");
Expect(1, 917632, '\p{^Is_Block=tags}', "");
Expect(1, 917632, '\P{Is_Block=tags}', "");
Expect(0, 917632, '\P{^Is_Block=tags}', "");
Expect(1, 917631, '\p{Is_Block=-_Tags}', "");
Expect(0, 917631, '\p{^Is_Block=-_Tags}', "");
Expect(0, 917631, '\P{Is_Block=-_Tags}', "");
Expect(1, 917631, '\P{^Is_Block=-_Tags}', "");
Expect(0, 917632, '\p{Is_Block=-_Tags}', "");
Expect(1, 917632, '\p{^Is_Block=-_Tags}', "");
Expect(1, 917632, '\P{Is_Block=-_Tags}', "");
Expect(0, 917632, '\P{^Is_Block=-_Tags}', "");
Error('\p{Is_Blk=_ Tags/a/}');
Error('\P{Is_Blk=_ Tags/a/}');
Expect(1, 917631, '\p{Is_Blk=tags}', "");
Expect(0, 917631, '\p{^Is_Blk=tags}', "");
Expect(0, 917631, '\P{Is_Blk=tags}', "");
Expect(1, 917631, '\P{^Is_Blk=tags}', "");
Expect(0, 917632, '\p{Is_Blk=tags}', "");
Expect(1, 917632, '\p{^Is_Blk=tags}', "");
Expect(1, 917632, '\P{Is_Blk=tags}', "");
Expect(0, 917632, '\P{^Is_Blk=tags}', "");
Expect(1, 917631, '\p{Is_Blk=	TAGS}', "");
Expect(0, 917631, '\p{^Is_Blk=	TAGS}', "");
Expect(0, 917631, '\P{Is_Blk=	TAGS}', "");
Expect(1, 917631, '\P{^Is_Blk=	TAGS}', "");
Expect(0, 917632, '\p{Is_Blk=	TAGS}', "");
Expect(1, 917632, '\p{^Is_Blk=	TAGS}', "");
Expect(1, 917632, '\P{Is_Blk=	TAGS}', "");
Expect(0, 917632, '\P{^Is_Blk=	TAGS}', "");
Error('\p{Block=/a/	Tai_LE}');
Error('\P{Block=/a/	Tai_LE}');
Expect(1, 6527, '\p{Block=taile}', "");
Expect(0, 6527, '\p{^Block=taile}', "");
Expect(0, 6527, '\P{Block=taile}', "");
Expect(1, 6527, '\P{^Block=taile}', "");
Expect(0, 6528, '\p{Block=taile}', "");
Expect(1, 6528, '\p{^Block=taile}', "");
Expect(1, 6528, '\P{Block=taile}', "");
Expect(0, 6528, '\P{^Block=taile}', "");
Expect(1, 6527, '\p{Block=-_Tai_Le}', "");
Expect(0, 6527, '\p{^Block=-_Tai_Le}', "");
Expect(0, 6527, '\P{Block=-_Tai_Le}', "");
Expect(1, 6527, '\P{^Block=-_Tai_Le}', "");
Expect(0, 6528, '\p{Block=-_Tai_Le}', "");
Expect(1, 6528, '\p{^Block=-_Tai_Le}', "");
Expect(1, 6528, '\P{Block=-_Tai_Le}', "");
Expect(0, 6528, '\P{^Block=-_Tai_Le}', "");
Error('\p{Blk=:=-tai_le}');
Error('\P{Blk=:=-tai_le}');
Expect(1, 6527, '\p{Blk=taile}', "");
Expect(0, 6527, '\p{^Blk=taile}', "");
Expect(0, 6527, '\P{Blk=taile}', "");
Expect(1, 6527, '\P{^Blk=taile}', "");
Expect(0, 6528, '\p{Blk=taile}', "");
Expect(1, 6528, '\p{^Blk=taile}', "");
Expect(1, 6528, '\P{Blk=taile}', "");
Expect(0, 6528, '\P{^Blk=taile}', "");
Expect(1, 6527, '\p{Blk=__TAI_Le}', "");
Expect(0, 6527, '\p{^Blk=__TAI_Le}', "");
Expect(0, 6527, '\P{Blk=__TAI_Le}', "");
Expect(1, 6527, '\P{^Blk=__TAI_Le}', "");
Expect(0, 6528, '\p{Blk=__TAI_Le}', "");
Expect(1, 6528, '\p{^Blk=__TAI_Le}', "");
Expect(1, 6528, '\P{Blk=__TAI_Le}', "");
Expect(0, 6528, '\P{^Blk=__TAI_Le}', "");
Error('\p{Is_Block=	-TAI_LE/a/}');
Error('\P{Is_Block=	-TAI_LE/a/}');
Expect(1, 6527, '\p{Is_Block=taile}', "");
Expect(0, 6527, '\p{^Is_Block=taile}', "");
Expect(0, 6527, '\P{Is_Block=taile}', "");
Expect(1, 6527, '\P{^Is_Block=taile}', "");
Expect(0, 6528, '\p{Is_Block=taile}', "");
Expect(1, 6528, '\p{^Is_Block=taile}', "");
Expect(1, 6528, '\P{Is_Block=taile}', "");
Expect(0, 6528, '\P{^Is_Block=taile}', "");
Expect(1, 6527, '\p{Is_Block= -Tai_le}', "");
Expect(0, 6527, '\p{^Is_Block= -Tai_le}', "");
Expect(0, 6527, '\P{Is_Block= -Tai_le}', "");
Expect(1, 6527, '\P{^Is_Block= -Tai_le}', "");
Expect(0, 6528, '\p{Is_Block= -Tai_le}', "");
Expect(1, 6528, '\p{^Is_Block= -Tai_le}', "");
Expect(1, 6528, '\P{Is_Block= -Tai_le}', "");
Expect(0, 6528, '\P{^Is_Block= -Tai_le}', "");
Error('\p{Is_Blk=:=-	Tai_Le}');
Error('\P{Is_Blk=:=-	Tai_Le}');
Expect(1, 6527, '\p{Is_Blk=taile}', "");
Expect(0, 6527, '\p{^Is_Blk=taile}', "");
Expect(0, 6527, '\P{Is_Blk=taile}', "");
Expect(1, 6527, '\P{^Is_Blk=taile}', "");
Expect(0, 6528, '\p{Is_Blk=taile}', "");
Expect(1, 6528, '\p{^Is_Blk=taile}', "");
Expect(1, 6528, '\P{Is_Blk=taile}', "");
Expect(0, 6528, '\P{^Is_Blk=taile}', "");
Expect(1, 6527, '\p{Is_Blk= _tai_le}', "");
Expect(0, 6527, '\p{^Is_Blk= _tai_le}', "");
Expect(0, 6527, '\P{Is_Blk= _tai_le}', "");
Expect(1, 6527, '\P{^Is_Blk= _tai_le}', "");
Expect(0, 6528, '\p{Is_Blk= _tai_le}', "");
Expect(1, 6528, '\p{^Is_Blk= _tai_le}', "");
Expect(1, 6528, '\P{Is_Blk= _tai_le}', "");
Expect(0, 6528, '\P{^Is_Blk= _tai_le}', "");
Error('\p{Block=	Tai_tham:=}');
Error('\P{Block=	Tai_tham:=}');
Expect(1, 6831, '\p{Block=taitham}', "");
Expect(0, 6831, '\p{^Block=taitham}', "");
Expect(0, 6831, '\P{Block=taitham}', "");
Expect(1, 6831, '\P{^Block=taitham}', "");
Expect(0, 6832, '\p{Block=taitham}', "");
Expect(1, 6832, '\p{^Block=taitham}', "");
Expect(1, 6832, '\P{Block=taitham}', "");
Expect(0, 6832, '\P{^Block=taitham}', "");
Expect(1, 6831, '\p{Block= 	Tai_Tham}', "");
Expect(0, 6831, '\p{^Block= 	Tai_Tham}', "");
Expect(0, 6831, '\P{Block= 	Tai_Tham}', "");
Expect(1, 6831, '\P{^Block= 	Tai_Tham}', "");
Expect(0, 6832, '\p{Block= 	Tai_Tham}', "");
Expect(1, 6832, '\p{^Block= 	Tai_Tham}', "");
Expect(1, 6832, '\P{Block= 	Tai_Tham}', "");
Expect(0, 6832, '\P{^Block= 	Tai_Tham}', "");
Error('\p{Blk= :=Tai_Tham}');
Error('\P{Blk= :=Tai_Tham}');
Expect(1, 6831, '\p{Blk=taitham}', "");
Expect(0, 6831, '\p{^Blk=taitham}', "");
Expect(0, 6831, '\P{Blk=taitham}', "");
Expect(1, 6831, '\P{^Blk=taitham}', "");
Expect(0, 6832, '\p{Blk=taitham}', "");
Expect(1, 6832, '\p{^Blk=taitham}', "");
Expect(1, 6832, '\P{Blk=taitham}', "");
Expect(0, 6832, '\P{^Blk=taitham}', "");
Expect(1, 6831, '\p{Blk:	_-tai_THAM}', "");
Expect(0, 6831, '\p{^Blk:	_-tai_THAM}', "");
Expect(0, 6831, '\P{Blk:	_-tai_THAM}', "");
Expect(1, 6831, '\P{^Blk:	_-tai_THAM}', "");
Expect(0, 6832, '\p{Blk:	_-tai_THAM}', "");
Expect(1, 6832, '\p{^Blk:	_-tai_THAM}', "");
Expect(1, 6832, '\P{Blk:	_-tai_THAM}', "");
Expect(0, 6832, '\P{^Blk:	_-tai_THAM}', "");
Error('\p{Is_Block:	-:=tai_tham}');
Error('\P{Is_Block:	-:=tai_tham}');
Expect(1, 6831, '\p{Is_Block=taitham}', "");
Expect(0, 6831, '\p{^Is_Block=taitham}', "");
Expect(0, 6831, '\P{Is_Block=taitham}', "");
Expect(1, 6831, '\P{^Is_Block=taitham}', "");
Expect(0, 6832, '\p{Is_Block=taitham}', "");
Expect(1, 6832, '\p{^Is_Block=taitham}', "");
Expect(1, 6832, '\P{Is_Block=taitham}', "");
Expect(0, 6832, '\P{^Is_Block=taitham}', "");
Expect(1, 6831, '\p{Is_Block=-tai_Tham}', "");
Expect(0, 6831, '\p{^Is_Block=-tai_Tham}', "");
Expect(0, 6831, '\P{Is_Block=-tai_Tham}', "");
Expect(1, 6831, '\P{^Is_Block=-tai_Tham}', "");
Expect(0, 6832, '\p{Is_Block=-tai_Tham}', "");
Expect(1, 6832, '\p{^Is_Block=-tai_Tham}', "");
Expect(1, 6832, '\P{Is_Block=-tai_Tham}', "");
Expect(0, 6832, '\P{^Is_Block=-tai_Tham}', "");
Error('\p{Is_Blk=:=_ Tai_THAM}');
Error('\P{Is_Blk=:=_ Tai_THAM}');
Expect(1, 6831, '\p{Is_Blk=taitham}', "");
Expect(0, 6831, '\p{^Is_Blk=taitham}', "");
Expect(0, 6831, '\P{Is_Blk=taitham}', "");
Expect(1, 6831, '\P{^Is_Blk=taitham}', "");
Expect(0, 6832, '\p{Is_Blk=taitham}', "");
Expect(1, 6832, '\p{^Is_Blk=taitham}', "");
Expect(1, 6832, '\P{Is_Blk=taitham}', "");
Expect(0, 6832, '\P{^Is_Blk=taitham}', "");
Expect(1, 6831, '\p{Is_Blk=	 tai_Tham}', "");
Expect(0, 6831, '\p{^Is_Blk=	 tai_Tham}', "");
Expect(0, 6831, '\P{Is_Blk=	 tai_Tham}', "");
Expect(1, 6831, '\P{^Is_Blk=	 tai_Tham}', "");
Expect(0, 6832, '\p{Is_Blk=	 tai_Tham}', "");
Expect(1, 6832, '\p{^Is_Blk=	 tai_Tham}', "");
Expect(1, 6832, '\P{Is_Blk=	 tai_Tham}', "");
Expect(0, 6832, '\P{^Is_Blk=	 tai_Tham}', "");
Error('\p{Block=	 tai_Viet:=}');
Error('\P{Block=	 tai_Viet:=}');
Expect(1, 43743, '\p{Block=taiviet}', "");
Expect(0, 43743, '\p{^Block=taiviet}', "");
Expect(0, 43743, '\P{Block=taiviet}', "");
Expect(1, 43743, '\P{^Block=taiviet}', "");
Expect(0, 43744, '\p{Block=taiviet}', "");
Expect(1, 43744, '\p{^Block=taiviet}', "");
Expect(1, 43744, '\P{Block=taiviet}', "");
Expect(0, 43744, '\P{^Block=taiviet}', "");
Expect(1, 43743, '\p{Block=__tai_viet}', "");
Expect(0, 43743, '\p{^Block=__tai_viet}', "");
Expect(0, 43743, '\P{Block=__tai_viet}', "");
Expect(1, 43743, '\P{^Block=__tai_viet}', "");
Expect(0, 43744, '\p{Block=__tai_viet}', "");
Expect(1, 43744, '\p{^Block=__tai_viet}', "");
Expect(1, 43744, '\P{Block=__tai_viet}', "");
Expect(0, 43744, '\P{^Block=__tai_viet}', "");
Error('\p{Blk=:=tai_viet}');
Error('\P{Blk=:=tai_viet}');
Expect(1, 43743, '\p{Blk:taiviet}', "");
Expect(0, 43743, '\p{^Blk:taiviet}', "");
Expect(0, 43743, '\P{Blk:taiviet}', "");
Expect(1, 43743, '\P{^Blk:taiviet}', "");
Expect(0, 43744, '\p{Blk:taiviet}', "");
Expect(1, 43744, '\p{^Blk:taiviet}', "");
Expect(1, 43744, '\P{Blk:taiviet}', "");
Expect(0, 43744, '\P{^Blk:taiviet}', "");
Expect(1, 43743, '\p{Blk=  Tai_Viet}', "");
Expect(0, 43743, '\p{^Blk=  Tai_Viet}', "");
Expect(0, 43743, '\P{Blk=  Tai_Viet}', "");
Expect(1, 43743, '\P{^Blk=  Tai_Viet}', "");
Expect(0, 43744, '\p{Blk=  Tai_Viet}', "");
Expect(1, 43744, '\p{^Blk=  Tai_Viet}', "");
Expect(1, 43744, '\P{Blk=  Tai_Viet}', "");
Expect(0, 43744, '\P{^Blk=  Tai_Viet}', "");
Error('\p{Is_Block=	Tai_Viet:=}');
Error('\P{Is_Block=	Tai_Viet:=}');
Expect(1, 43743, '\p{Is_Block=taiviet}', "");
Expect(0, 43743, '\p{^Is_Block=taiviet}', "");
Expect(0, 43743, '\P{Is_Block=taiviet}', "");
Expect(1, 43743, '\P{^Is_Block=taiviet}', "");
Expect(0, 43744, '\p{Is_Block=taiviet}', "");
Expect(1, 43744, '\p{^Is_Block=taiviet}', "");
Expect(1, 43744, '\P{Is_Block=taiviet}', "");
Expect(0, 43744, '\P{^Is_Block=taiviet}', "");
Expect(1, 43743, '\p{Is_Block=_-tai_VIET}', "");
Expect(0, 43743, '\p{^Is_Block=_-tai_VIET}', "");
Expect(0, 43743, '\P{Is_Block=_-tai_VIET}', "");
Expect(1, 43743, '\P{^Is_Block=_-tai_VIET}', "");
Expect(0, 43744, '\p{Is_Block=_-tai_VIET}', "");
Expect(1, 43744, '\p{^Is_Block=_-tai_VIET}', "");
Expect(1, 43744, '\P{Is_Block=_-tai_VIET}', "");
Expect(0, 43744, '\P{^Is_Block=_-tai_VIET}', "");
Error('\p{Is_Blk=:=	TAI_Viet}');
Error('\P{Is_Blk=:=	TAI_Viet}');
Expect(1, 43743, '\p{Is_Blk=taiviet}', "");
Expect(0, 43743, '\p{^Is_Blk=taiviet}', "");
Expect(0, 43743, '\P{Is_Blk=taiviet}', "");
Expect(1, 43743, '\P{^Is_Blk=taiviet}', "");
Expect(0, 43744, '\p{Is_Blk=taiviet}', "");
Expect(1, 43744, '\p{^Is_Blk=taiviet}', "");
Expect(1, 43744, '\P{Is_Blk=taiviet}', "");
Expect(0, 43744, '\P{^Is_Blk=taiviet}', "");
Expect(1, 43743, '\p{Is_Blk=-Tai_VIET}', "");
Expect(0, 43743, '\p{^Is_Blk=-Tai_VIET}', "");
Expect(0, 43743, '\P{Is_Blk=-Tai_VIET}', "");
Expect(1, 43743, '\P{^Is_Blk=-Tai_VIET}', "");
Expect(0, 43744, '\p{Is_Blk=-Tai_VIET}', "");
Expect(1, 43744, '\p{^Is_Blk=-Tai_VIET}', "");
Expect(1, 43744, '\P{Is_Blk=-Tai_VIET}', "");
Expect(0, 43744, '\P{^Is_Blk=-Tai_VIET}', "");
Error('\p{Block=:=  tai_XUAN_JING_SYMBOLS}');
Error('\P{Block=:=  tai_XUAN_JING_SYMBOLS}');
Expect(1, 119647, '\p{Block=taixuanjingsymbols}', "");
Expect(0, 119647, '\p{^Block=taixuanjingsymbols}', "");
Expect(0, 119647, '\P{Block=taixuanjingsymbols}', "");
Expect(1, 119647, '\P{^Block=taixuanjingsymbols}', "");
Expect(0, 119648, '\p{Block=taixuanjingsymbols}', "");
Expect(1, 119648, '\p{^Block=taixuanjingsymbols}', "");
Expect(1, 119648, '\P{Block=taixuanjingsymbols}', "");
Expect(0, 119648, '\P{^Block=taixuanjingsymbols}', "");
Expect(1, 119647, '\p{Block=_ Tai_XUAN_JING_Symbols}', "");
Expect(0, 119647, '\p{^Block=_ Tai_XUAN_JING_Symbols}', "");
Expect(0, 119647, '\P{Block=_ Tai_XUAN_JING_Symbols}', "");
Expect(1, 119647, '\P{^Block=_ Tai_XUAN_JING_Symbols}', "");
Expect(0, 119648, '\p{Block=_ Tai_XUAN_JING_Symbols}', "");
Expect(1, 119648, '\p{^Block=_ Tai_XUAN_JING_Symbols}', "");
Expect(1, 119648, '\P{Block=_ Tai_XUAN_JING_Symbols}', "");
Expect(0, 119648, '\P{^Block=_ Tai_XUAN_JING_Symbols}', "");
Error('\p{Blk=_TAI_Xuan_jing/a/}');
Error('\P{Blk=_TAI_Xuan_jing/a/}');
Expect(1, 119647, '\p{Blk:	taixuanjing}', "");
Expect(0, 119647, '\p{^Blk:	taixuanjing}', "");
Expect(0, 119647, '\P{Blk:	taixuanjing}', "");
Expect(1, 119647, '\P{^Blk:	taixuanjing}', "");
Expect(0, 119648, '\p{Blk:	taixuanjing}', "");
Expect(1, 119648, '\p{^Blk:	taixuanjing}', "");
Expect(1, 119648, '\P{Blk:	taixuanjing}', "");
Expect(0, 119648, '\P{^Blk:	taixuanjing}', "");
Expect(1, 119647, '\p{Blk=-	Tai_Xuan_Jing}', "");
Expect(0, 119647, '\p{^Blk=-	Tai_Xuan_Jing}', "");
Expect(0, 119647, '\P{Blk=-	Tai_Xuan_Jing}', "");
Expect(1, 119647, '\P{^Blk=-	Tai_Xuan_Jing}', "");
Expect(0, 119648, '\p{Blk=-	Tai_Xuan_Jing}', "");
Expect(1, 119648, '\p{^Blk=-	Tai_Xuan_Jing}', "");
Expect(1, 119648, '\P{Blk=-	Tai_Xuan_Jing}', "");
Expect(0, 119648, '\P{^Blk=-	Tai_Xuan_Jing}', "");
Error('\p{Is_Block=/a/_-Tai_Xuan_Jing_SYMBOLS}');
Error('\P{Is_Block=/a/_-Tai_Xuan_Jing_SYMBOLS}');
Expect(1, 119647, '\p{Is_Block:	taixuanjingsymbols}', "");
Expect(0, 119647, '\p{^Is_Block:	taixuanjingsymbols}', "");
Expect(0, 119647, '\P{Is_Block:	taixuanjingsymbols}', "");
Expect(1, 119647, '\P{^Is_Block:	taixuanjingsymbols}', "");
Expect(0, 119648, '\p{Is_Block:	taixuanjingsymbols}', "");
Expect(1, 119648, '\p{^Is_Block:	taixuanjingsymbols}', "");
Expect(1, 119648, '\P{Is_Block:	taixuanjingsymbols}', "");
Expect(0, 119648, '\P{^Is_Block:	taixuanjingsymbols}', "");
Expect(1, 119647, '\p{Is_Block= -Tai_Xuan_Jing_Symbols}', "");
Expect(0, 119647, '\p{^Is_Block= -Tai_Xuan_Jing_Symbols}', "");
Expect(0, 119647, '\P{Is_Block= -Tai_Xuan_Jing_Symbols}', "");
Expect(1, 119647, '\P{^Is_Block= -Tai_Xuan_Jing_Symbols}', "");
Expect(0, 119648, '\p{Is_Block= -Tai_Xuan_Jing_Symbols}', "");
Expect(1, 119648, '\p{^Is_Block= -Tai_Xuan_Jing_Symbols}', "");
Expect(1, 119648, '\P{Is_Block= -Tai_Xuan_Jing_Symbols}', "");
Expect(0, 119648, '\P{^Is_Block= -Tai_Xuan_Jing_Symbols}', "");
Error('\p{Is_Blk=:=__TAI_Xuan_JING}');
Error('\P{Is_Blk=:=__TAI_Xuan_JING}');
Expect(1, 119647, '\p{Is_Blk=taixuanjing}', "");
Expect(0, 119647, '\p{^Is_Blk=taixuanjing}', "");
Expect(0, 119647, '\P{Is_Blk=taixuanjing}', "");
Expect(1, 119647, '\P{^Is_Blk=taixuanjing}', "");
Expect(0, 119648, '\p{Is_Blk=taixuanjing}', "");
Expect(1, 119648, '\p{^Is_Blk=taixuanjing}', "");
Expect(1, 119648, '\P{Is_Blk=taixuanjing}', "");
Expect(0, 119648, '\P{^Is_Blk=taixuanjing}', "");
Expect(1, 119647, '\p{Is_Blk=	_Tai_Xuan_Jing}', "");
Expect(0, 119647, '\p{^Is_Blk=	_Tai_Xuan_Jing}', "");
Expect(0, 119647, '\P{Is_Blk=	_Tai_Xuan_Jing}', "");
Expect(1, 119647, '\P{^Is_Blk=	_Tai_Xuan_Jing}', "");
Expect(0, 119648, '\p{Is_Blk=	_Tai_Xuan_Jing}', "");
Expect(1, 119648, '\p{^Is_Blk=	_Tai_Xuan_Jing}', "");
Expect(1, 119648, '\P{Is_Blk=	_Tai_Xuan_Jing}', "");
Expect(0, 119648, '\P{^Is_Blk=	_Tai_Xuan_Jing}', "");
Error('\p{Block=-:=takri}');
Error('\P{Block=-:=takri}');
Expect(1, 71375, '\p{Block=takri}', "");
Expect(0, 71375, '\p{^Block=takri}', "");
Expect(0, 71375, '\P{Block=takri}', "");
Expect(1, 71375, '\P{^Block=takri}', "");
Expect(0, 71376, '\p{Block=takri}', "");
Expect(1, 71376, '\p{^Block=takri}', "");
Expect(1, 71376, '\P{Block=takri}', "");
Expect(0, 71376, '\P{^Block=takri}', "");
Expect(1, 71375, '\p{Block=	_takri}', "");
Expect(0, 71375, '\p{^Block=	_takri}', "");
Expect(0, 71375, '\P{Block=	_takri}', "");
Expect(1, 71375, '\P{^Block=	_takri}', "");
Expect(0, 71376, '\p{Block=	_takri}', "");
Expect(1, 71376, '\p{^Block=	_takri}', "");
Expect(1, 71376, '\P{Block=	_takri}', "");
Expect(0, 71376, '\P{^Block=	_takri}', "");
Error('\p{Blk=/a/_TAKRI}');
Error('\P{Blk=/a/_TAKRI}');
Expect(1, 71375, '\p{Blk:	takri}', "");
Expect(0, 71375, '\p{^Blk:	takri}', "");
Expect(0, 71375, '\P{Blk:	takri}', "");
Expect(1, 71375, '\P{^Blk:	takri}', "");
Expect(0, 71376, '\p{Blk:	takri}', "");
Expect(1, 71376, '\p{^Blk:	takri}', "");
Expect(1, 71376, '\P{Blk:	takri}', "");
Expect(0, 71376, '\P{^Blk:	takri}', "");
Expect(1, 71375, '\p{Blk=_Takri}', "");
Expect(0, 71375, '\p{^Blk=_Takri}', "");
Expect(0, 71375, '\P{Blk=_Takri}', "");
Expect(1, 71375, '\P{^Blk=_Takri}', "");
Expect(0, 71376, '\p{Blk=_Takri}', "");
Expect(1, 71376, '\p{^Blk=_Takri}', "");
Expect(1, 71376, '\P{Blk=_Takri}', "");
Expect(0, 71376, '\P{^Blk=_Takri}', "");
Error('\p{Is_Block=_:=Takri}');
Error('\P{Is_Block=_:=Takri}');
Expect(1, 71375, '\p{Is_Block=takri}', "");
Expect(0, 71375, '\p{^Is_Block=takri}', "");
Expect(0, 71375, '\P{Is_Block=takri}', "");
Expect(1, 71375, '\P{^Is_Block=takri}', "");
Expect(0, 71376, '\p{Is_Block=takri}', "");
Expect(1, 71376, '\p{^Is_Block=takri}', "");
Expect(1, 71376, '\P{Is_Block=takri}', "");
Expect(0, 71376, '\P{^Is_Block=takri}', "");
Expect(1, 71375, '\p{Is_Block=	Takri}', "");
Expect(0, 71375, '\p{^Is_Block=	Takri}', "");
Expect(0, 71375, '\P{Is_Block=	Takri}', "");
Expect(1, 71375, '\P{^Is_Block=	Takri}', "");
Expect(0, 71376, '\p{Is_Block=	Takri}', "");
Expect(1, 71376, '\p{^Is_Block=	Takri}', "");
Expect(1, 71376, '\P{Is_Block=	Takri}', "");
Expect(0, 71376, '\P{^Is_Block=	Takri}', "");
Error('\p{Is_Blk= /a/TAKRI}');
Error('\P{Is_Blk= /a/TAKRI}');
Expect(1, 71375, '\p{Is_Blk=takri}', "");
Expect(0, 71375, '\p{^Is_Blk=takri}', "");
Expect(0, 71375, '\P{Is_Blk=takri}', "");
Expect(1, 71375, '\P{^Is_Blk=takri}', "");
Expect(0, 71376, '\p{Is_Blk=takri}', "");
Expect(1, 71376, '\p{^Is_Blk=takri}', "");
Expect(1, 71376, '\P{Is_Blk=takri}', "");
Expect(0, 71376, '\P{^Is_Blk=takri}', "");
Expect(1, 71375, '\p{Is_Blk=- Takri}', "");
Expect(0, 71375, '\p{^Is_Blk=- Takri}', "");
Expect(0, 71375, '\P{Is_Blk=- Takri}', "");
Expect(1, 71375, '\P{^Is_Blk=- Takri}', "");
Expect(0, 71376, '\p{Is_Blk=- Takri}', "");
Expect(1, 71376, '\p{^Is_Blk=- Takri}', "");
Expect(1, 71376, '\P{Is_Blk=- Takri}', "");
Expect(0, 71376, '\P{^Is_Blk=- Takri}', "");
Error('\p{Block:		Tamil/a/}');
Error('\P{Block:		Tamil/a/}');
Expect(1, 3071, '\p{Block=tamil}', "");
Expect(0, 3071, '\p{^Block=tamil}', "");
Expect(0, 3071, '\P{Block=tamil}', "");
Expect(1, 3071, '\P{^Block=tamil}', "");
Expect(0, 3072, '\p{Block=tamil}', "");
Expect(1, 3072, '\p{^Block=tamil}', "");
Expect(1, 3072, '\P{Block=tamil}', "");
Expect(0, 3072, '\P{^Block=tamil}', "");
Expect(1, 3071, '\p{Block=-Tamil}', "");
Expect(0, 3071, '\p{^Block=-Tamil}', "");
Expect(0, 3071, '\P{Block=-Tamil}', "");
Expect(1, 3071, '\P{^Block=-Tamil}', "");
Expect(0, 3072, '\p{Block=-Tamil}', "");
Expect(1, 3072, '\p{^Block=-Tamil}', "");
Expect(1, 3072, '\P{Block=-Tamil}', "");
Expect(0, 3072, '\P{^Block=-Tamil}', "");
Error('\p{Blk= 	TAMIL/a/}');
Error('\P{Blk= 	TAMIL/a/}');
Expect(1, 3071, '\p{Blk=tamil}', "");
Expect(0, 3071, '\p{^Blk=tamil}', "");
Expect(0, 3071, '\P{Blk=tamil}', "");
Expect(1, 3071, '\P{^Blk=tamil}', "");
Expect(0, 3072, '\p{Blk=tamil}', "");
Expect(1, 3072, '\p{^Blk=tamil}', "");
Expect(1, 3072, '\P{Blk=tamil}', "");
Expect(0, 3072, '\P{^Blk=tamil}', "");
Expect(1, 3071, '\p{Blk=	_Tamil}', "");
Expect(0, 3071, '\p{^Blk=	_Tamil}', "");
Expect(0, 3071, '\P{Blk=	_Tamil}', "");
Expect(1, 3071, '\P{^Blk=	_Tamil}', "");
Expect(0, 3072, '\p{Blk=	_Tamil}', "");
Expect(1, 3072, '\p{^Blk=	_Tamil}', "");
Expect(1, 3072, '\P{Blk=	_Tamil}', "");
Expect(0, 3072, '\P{^Blk=	_Tamil}', "");
Error('\p{Is_Block=:=-	TAMIL}');
Error('\P{Is_Block=:=-	TAMIL}');
Expect(1, 3071, '\p{Is_Block=tamil}', "");
Expect(0, 3071, '\p{^Is_Block=tamil}', "");
Expect(0, 3071, '\P{Is_Block=tamil}', "");
Expect(1, 3071, '\P{^Is_Block=tamil}', "");
Expect(0, 3072, '\p{Is_Block=tamil}', "");
Expect(1, 3072, '\p{^Is_Block=tamil}', "");
Expect(1, 3072, '\P{Is_Block=tamil}', "");
Expect(0, 3072, '\P{^Is_Block=tamil}', "");
Expect(1, 3071, '\p{Is_Block=	Tamil}', "");
Expect(0, 3071, '\p{^Is_Block=	Tamil}', "");
Expect(0, 3071, '\P{Is_Block=	Tamil}', "");
Expect(1, 3071, '\P{^Is_Block=	Tamil}', "");
Expect(0, 3072, '\p{Is_Block=	Tamil}', "");
Expect(1, 3072, '\p{^Is_Block=	Tamil}', "");
Expect(1, 3072, '\P{Is_Block=	Tamil}', "");
Expect(0, 3072, '\P{^Is_Block=	Tamil}', "");
Error('\p{Is_Blk=:=Tamil}');
Error('\P{Is_Blk=:=Tamil}');
Expect(1, 3071, '\p{Is_Blk: tamil}', "");
Expect(0, 3071, '\p{^Is_Blk: tamil}', "");
Expect(0, 3071, '\P{Is_Blk: tamil}', "");
Expect(1, 3071, '\P{^Is_Blk: tamil}', "");
Expect(0, 3072, '\p{Is_Blk: tamil}', "");
Expect(1, 3072, '\p{^Is_Blk: tamil}', "");
Expect(1, 3072, '\P{Is_Blk: tamil}', "");
Expect(0, 3072, '\P{^Is_Blk: tamil}', "");
Expect(1, 3071, '\p{Is_Blk=Tamil}', "");
Expect(0, 3071, '\p{^Is_Blk=Tamil}', "");
Expect(0, 3071, '\P{Is_Blk=Tamil}', "");
Expect(1, 3071, '\P{^Is_Blk=Tamil}', "");
Expect(0, 3072, '\p{Is_Blk=Tamil}', "");
Expect(1, 3072, '\p{^Is_Blk=Tamil}', "");
Expect(1, 3072, '\P{Is_Blk=Tamil}', "");
Expect(0, 3072, '\P{^Is_Blk=Tamil}', "");
Error('\p{Block=-/a/Telugu}');
Error('\P{Block=-/a/Telugu}');
Expect(1, 3199, '\p{Block=telugu}', "");
Expect(0, 3199, '\p{^Block=telugu}', "");
Expect(0, 3199, '\P{Block=telugu}', "");
Expect(1, 3199, '\P{^Block=telugu}', "");
Expect(0, 3200, '\p{Block=telugu}', "");
Expect(1, 3200, '\p{^Block=telugu}', "");
Expect(1, 3200, '\P{Block=telugu}', "");
Expect(0, 3200, '\P{^Block=telugu}', "");
Expect(1, 3199, '\p{Block=_TELUGU}', "");
Expect(0, 3199, '\p{^Block=_TELUGU}', "");
Expect(0, 3199, '\P{Block=_TELUGU}', "");
Expect(1, 3199, '\P{^Block=_TELUGU}', "");
Expect(0, 3200, '\p{Block=_TELUGU}', "");
Expect(1, 3200, '\p{^Block=_TELUGU}', "");
Expect(1, 3200, '\P{Block=_TELUGU}', "");
Expect(0, 3200, '\P{^Block=_TELUGU}', "");
Error('\p{Blk=:=TELUGU}');
Error('\P{Blk=:=TELUGU}');
Expect(1, 3199, '\p{Blk=telugu}', "");
Expect(0, 3199, '\p{^Blk=telugu}', "");
Expect(0, 3199, '\P{Blk=telugu}', "");
Expect(1, 3199, '\P{^Blk=telugu}', "");
Expect(0, 3200, '\p{Blk=telugu}', "");
Expect(1, 3200, '\p{^Blk=telugu}', "");
Expect(1, 3200, '\P{Blk=telugu}', "");
Expect(0, 3200, '\P{^Blk=telugu}', "");
Expect(1, 3199, '\p{Blk=_ TELUGU}', "");
Expect(0, 3199, '\p{^Blk=_ TELUGU}', "");
Expect(0, 3199, '\P{Blk=_ TELUGU}', "");
Expect(1, 3199, '\P{^Blk=_ TELUGU}', "");
Expect(0, 3200, '\p{Blk=_ TELUGU}', "");
Expect(1, 3200, '\p{^Blk=_ TELUGU}', "");
Expect(1, 3200, '\P{Blk=_ TELUGU}', "");
Expect(0, 3200, '\P{^Blk=_ TELUGU}', "");
Error('\p{Is_Block=_/a/Telugu}');
Error('\P{Is_Block=_/a/Telugu}');
Expect(1, 3199, '\p{Is_Block=telugu}', "");
Expect(0, 3199, '\p{^Is_Block=telugu}', "");
Expect(0, 3199, '\P{Is_Block=telugu}', "");
Expect(1, 3199, '\P{^Is_Block=telugu}', "");
Expect(0, 3200, '\p{Is_Block=telugu}', "");
Expect(1, 3200, '\p{^Is_Block=telugu}', "");
Expect(1, 3200, '\P{Is_Block=telugu}', "");
Expect(0, 3200, '\P{^Is_Block=telugu}', "");
Expect(1, 3199, '\p{Is_Block=- Telugu}', "");
Expect(0, 3199, '\p{^Is_Block=- Telugu}', "");
Expect(0, 3199, '\P{Is_Block=- Telugu}', "");
Expect(1, 3199, '\P{^Is_Block=- Telugu}', "");
Expect(0, 3200, '\p{Is_Block=- Telugu}', "");
Expect(1, 3200, '\p{^Is_Block=- Telugu}', "");
Expect(1, 3200, '\P{Is_Block=- Telugu}', "");
Expect(0, 3200, '\P{^Is_Block=- Telugu}', "");
Error('\p{Is_Blk=-telugu/a/}');
Error('\P{Is_Blk=-telugu/a/}');
Expect(1, 3199, '\p{Is_Blk=telugu}', "");
Expect(0, 3199, '\p{^Is_Blk=telugu}', "");
Expect(0, 3199, '\P{Is_Blk=telugu}', "");
Expect(1, 3199, '\P{^Is_Blk=telugu}', "");
Expect(0, 3200, '\p{Is_Blk=telugu}', "");
Expect(1, 3200, '\p{^Is_Blk=telugu}', "");
Expect(1, 3200, '\P{Is_Blk=telugu}', "");
Expect(0, 3200, '\P{^Is_Blk=telugu}', "");
Expect(1, 3199, '\p{Is_Blk=  TELUGU}', "");
Expect(0, 3199, '\p{^Is_Blk=  TELUGU}', "");
Expect(0, 3199, '\P{Is_Blk=  TELUGU}', "");
Expect(1, 3199, '\P{^Is_Blk=  TELUGU}', "");
Expect(0, 3200, '\p{Is_Blk=  TELUGU}', "");
Expect(1, 3200, '\p{^Is_Blk=  TELUGU}', "");
Expect(1, 3200, '\P{Is_Blk=  TELUGU}', "");
Expect(0, 3200, '\P{^Is_Blk=  TELUGU}', "");
Error('\p{Block=:= -THAANA}');
Error('\P{Block=:= -THAANA}');
Expect(1, 1983, '\p{Block=thaana}', "");
Expect(0, 1983, '\p{^Block=thaana}', "");
Expect(0, 1983, '\P{Block=thaana}', "");
Expect(1, 1983, '\P{^Block=thaana}', "");
Expect(0, 1984, '\p{Block=thaana}', "");
Expect(1, 1984, '\p{^Block=thaana}', "");
Expect(1, 1984, '\P{Block=thaana}', "");
Expect(0, 1984, '\P{^Block=thaana}', "");
Expect(1, 1983, '\p{Block:	Thaana}', "");
Expect(0, 1983, '\p{^Block:	Thaana}', "");
Expect(0, 1983, '\P{Block:	Thaana}', "");
Expect(1, 1983, '\P{^Block:	Thaana}', "");
Expect(0, 1984, '\p{Block:	Thaana}', "");
Expect(1, 1984, '\p{^Block:	Thaana}', "");
Expect(1, 1984, '\P{Block:	Thaana}', "");
Expect(0, 1984, '\P{^Block:	Thaana}', "");
Error('\p{Blk=/a/	Thaana}');
Error('\P{Blk=/a/	Thaana}');
Expect(1, 1983, '\p{Blk: thaana}', "");
Expect(0, 1983, '\p{^Blk: thaana}', "");
Expect(0, 1983, '\P{Blk: thaana}', "");
Expect(1, 1983, '\P{^Blk: thaana}', "");
Expect(0, 1984, '\p{Blk: thaana}', "");
Expect(1, 1984, '\p{^Blk: thaana}', "");
Expect(1, 1984, '\P{Blk: thaana}', "");
Expect(0, 1984, '\P{^Blk: thaana}', "");
Expect(1, 1983, '\p{Blk=_THAANA}', "");
Expect(0, 1983, '\p{^Blk=_THAANA}', "");
Expect(0, 1983, '\P{Blk=_THAANA}', "");
Expect(1, 1983, '\P{^Blk=_THAANA}', "");
Expect(0, 1984, '\p{Blk=_THAANA}', "");
Expect(1, 1984, '\p{^Blk=_THAANA}', "");
Expect(1, 1984, '\P{Blk=_THAANA}', "");
Expect(0, 1984, '\P{^Blk=_THAANA}', "");
Error('\p{Is_Block::=Thaana}');
Error('\P{Is_Block::=Thaana}');
Expect(1, 1983, '\p{Is_Block=thaana}', "");
Expect(0, 1983, '\p{^Is_Block=thaana}', "");
Expect(0, 1983, '\P{Is_Block=thaana}', "");
Expect(1, 1983, '\P{^Is_Block=thaana}', "");
Expect(0, 1984, '\p{Is_Block=thaana}', "");
Expect(1, 1984, '\p{^Is_Block=thaana}', "");
Expect(1, 1984, '\P{Is_Block=thaana}', "");
Expect(0, 1984, '\P{^Is_Block=thaana}', "");
Expect(1, 1983, '\p{Is_Block=-THAANA}', "");
Expect(0, 1983, '\p{^Is_Block=-THAANA}', "");
Expect(0, 1983, '\P{Is_Block=-THAANA}', "");
Expect(1, 1983, '\P{^Is_Block=-THAANA}', "");
Expect(0, 1984, '\p{Is_Block=-THAANA}', "");
Expect(1, 1984, '\p{^Is_Block=-THAANA}', "");
Expect(1, 1984, '\P{Is_Block=-THAANA}', "");
Expect(0, 1984, '\P{^Is_Block=-THAANA}', "");
Error('\p{Is_Blk=_Thaana:=}');
Error('\P{Is_Blk=_Thaana:=}');
Expect(1, 1983, '\p{Is_Blk=thaana}', "");
Expect(0, 1983, '\p{^Is_Blk=thaana}', "");
Expect(0, 1983, '\P{Is_Blk=thaana}', "");
Expect(1, 1983, '\P{^Is_Blk=thaana}', "");
Expect(0, 1984, '\p{Is_Blk=thaana}', "");
Expect(1, 1984, '\p{^Is_Blk=thaana}', "");
Expect(1, 1984, '\P{Is_Blk=thaana}', "");
Expect(0, 1984, '\P{^Is_Blk=thaana}', "");
Expect(1, 1983, '\p{Is_Blk: 		Thaana}', "");
Expect(0, 1983, '\p{^Is_Blk: 		Thaana}', "");
Expect(0, 1983, '\P{Is_Blk: 		Thaana}', "");
Expect(1, 1983, '\P{^Is_Blk: 		Thaana}', "");
Expect(0, 1984, '\p{Is_Blk: 		Thaana}', "");
Expect(1, 1984, '\p{^Is_Blk: 		Thaana}', "");
Expect(1, 1984, '\P{Is_Blk: 		Thaana}', "");
Expect(0, 1984, '\P{^Is_Blk: 		Thaana}', "");
Error('\p{Block=:=	-thai}');
Error('\P{Block=:=	-thai}');
Expect(1, 3711, '\p{Block=thai}', "");
Expect(0, 3711, '\p{^Block=thai}', "");
Expect(0, 3711, '\P{Block=thai}', "");
Expect(1, 3711, '\P{^Block=thai}', "");
Expect(0, 3712, '\p{Block=thai}', "");
Expect(1, 3712, '\p{^Block=thai}', "");
Expect(1, 3712, '\P{Block=thai}', "");
Expect(0, 3712, '\P{^Block=thai}', "");
Expect(1, 3711, '\p{Block=-THAI}', "");
Expect(0, 3711, '\p{^Block=-THAI}', "");
Expect(0, 3711, '\P{Block=-THAI}', "");
Expect(1, 3711, '\P{^Block=-THAI}', "");
Expect(0, 3712, '\p{Block=-THAI}', "");
Expect(1, 3712, '\p{^Block=-THAI}', "");
Expect(1, 3712, '\P{Block=-THAI}', "");
Expect(0, 3712, '\P{^Block=-THAI}', "");
Error('\p{Blk= /a/Thai}');
Error('\P{Blk= /a/Thai}');
Expect(1, 3711, '\p{Blk=thai}', "");
Expect(0, 3711, '\p{^Blk=thai}', "");
Expect(0, 3711, '\P{Blk=thai}', "");
Expect(1, 3711, '\P{^Blk=thai}', "");
Expect(0, 3712, '\p{Blk=thai}', "");
Expect(1, 3712, '\p{^Blk=thai}', "");
Expect(1, 3712, '\P{Blk=thai}', "");
Expect(0, 3712, '\P{^Blk=thai}', "");
Expect(1, 3711, '\p{Blk=-THAI}', "");
Expect(0, 3711, '\p{^Blk=-THAI}', "");
Expect(0, 3711, '\P{Blk=-THAI}', "");
Expect(1, 3711, '\P{^Blk=-THAI}', "");
Expect(0, 3712, '\p{Blk=-THAI}', "");
Expect(1, 3712, '\p{^Blk=-THAI}', "");
Expect(1, 3712, '\P{Blk=-THAI}', "");
Expect(0, 3712, '\P{^Blk=-THAI}', "");
Error('\p{Is_Block: :=	Thai}');
Error('\P{Is_Block: :=	Thai}');
Expect(1, 3711, '\p{Is_Block:thai}', "");
Expect(0, 3711, '\p{^Is_Block:thai}', "");
Expect(0, 3711, '\P{Is_Block:thai}', "");
Expect(1, 3711, '\P{^Is_Block:thai}', "");
Expect(0, 3712, '\p{Is_Block:thai}', "");
Expect(1, 3712, '\p{^Is_Block:thai}', "");
Expect(1, 3712, '\P{Is_Block:thai}', "");
Expect(0, 3712, '\P{^Is_Block:thai}', "");
Expect(1, 3711, '\p{Is_Block= _Thai}', "");
Expect(0, 3711, '\p{^Is_Block= _Thai}', "");
Expect(0, 3711, '\P{Is_Block= _Thai}', "");
Expect(1, 3711, '\P{^Is_Block= _Thai}', "");
Expect(0, 3712, '\p{Is_Block= _Thai}', "");
Expect(1, 3712, '\p{^Is_Block= _Thai}', "");
Expect(1, 3712, '\P{Is_Block= _Thai}', "");
Expect(0, 3712, '\P{^Is_Block= _Thai}', "");
Error('\p{Is_Blk: :=Thai}');
Error('\P{Is_Blk: :=Thai}');
Expect(1, 3711, '\p{Is_Blk:   thai}', "");
Expect(0, 3711, '\p{^Is_Blk:   thai}', "");
Expect(0, 3711, '\P{Is_Blk:   thai}', "");
Expect(1, 3711, '\P{^Is_Blk:   thai}', "");
Expect(0, 3712, '\p{Is_Blk:   thai}', "");
Expect(1, 3712, '\p{^Is_Blk:   thai}', "");
Expect(1, 3712, '\P{Is_Blk:   thai}', "");
Expect(0, 3712, '\P{^Is_Blk:   thai}', "");
Expect(1, 3711, '\p{Is_Blk=	_THAI}', "");
Expect(0, 3711, '\p{^Is_Blk=	_THAI}', "");
Expect(0, 3711, '\P{Is_Blk=	_THAI}', "");
Expect(1, 3711, '\P{^Is_Blk=	_THAI}', "");
Expect(0, 3712, '\p{Is_Blk=	_THAI}', "");
Expect(1, 3712, '\p{^Is_Blk=	_THAI}', "");
Expect(1, 3712, '\P{Is_Blk=	_THAI}', "");
Expect(0, 3712, '\P{^Is_Blk=	_THAI}', "");
Error('\p{Block=:=-	tibetan}');
Error('\P{Block=:=-	tibetan}');
Expect(1, 4095, '\p{Block=tibetan}', "");
Expect(0, 4095, '\p{^Block=tibetan}', "");
Expect(0, 4095, '\P{Block=tibetan}', "");
Expect(1, 4095, '\P{^Block=tibetan}', "");
Expect(0, 4096, '\p{Block=tibetan}', "");
Expect(1, 4096, '\p{^Block=tibetan}', "");
Expect(1, 4096, '\P{Block=tibetan}', "");
Expect(0, 4096, '\P{^Block=tibetan}', "");
Expect(1, 4095, '\p{Block=	Tibetan}', "");
Expect(0, 4095, '\p{^Block=	Tibetan}', "");
Expect(0, 4095, '\P{Block=	Tibetan}', "");
Expect(1, 4095, '\P{^Block=	Tibetan}', "");
Expect(0, 4096, '\p{Block=	Tibetan}', "");
Expect(1, 4096, '\p{^Block=	Tibetan}', "");
Expect(1, 4096, '\P{Block=	Tibetan}', "");
Expect(0, 4096, '\P{^Block=	Tibetan}', "");
Error('\p{Blk=/a/	-Tibetan}');
Error('\P{Blk=/a/	-Tibetan}');
Expect(1, 4095, '\p{Blk=tibetan}', "");
Expect(0, 4095, '\p{^Blk=tibetan}', "");
Expect(0, 4095, '\P{Blk=tibetan}', "");
Expect(1, 4095, '\P{^Blk=tibetan}', "");
Expect(0, 4096, '\p{Blk=tibetan}', "");
Expect(1, 4096, '\p{^Blk=tibetan}', "");
Expect(1, 4096, '\P{Blk=tibetan}', "");
Expect(0, 4096, '\P{^Blk=tibetan}', "");
Expect(1, 4095, '\p{Blk=_ Tibetan}', "");
Expect(0, 4095, '\p{^Blk=_ Tibetan}', "");
Expect(0, 4095, '\P{Blk=_ Tibetan}', "");
Expect(1, 4095, '\P{^Blk=_ Tibetan}', "");
Expect(0, 4096, '\p{Blk=_ Tibetan}', "");
Expect(1, 4096, '\p{^Blk=_ Tibetan}', "");
Expect(1, 4096, '\P{Blk=_ Tibetan}', "");
Expect(0, 4096, '\P{^Blk=_ Tibetan}', "");
Error('\p{Is_Block= /a/tibetan}');
Error('\P{Is_Block= /a/tibetan}');
Expect(1, 4095, '\p{Is_Block:	tibetan}', "");
Expect(0, 4095, '\p{^Is_Block:	tibetan}', "");
Expect(0, 4095, '\P{Is_Block:	tibetan}', "");
Expect(1, 4095, '\P{^Is_Block:	tibetan}', "");
Expect(0, 4096, '\p{Is_Block:	tibetan}', "");
Expect(1, 4096, '\p{^Is_Block:	tibetan}', "");
Expect(1, 4096, '\P{Is_Block:	tibetan}', "");
Expect(0, 4096, '\P{^Is_Block:	tibetan}', "");
Expect(1, 4095, '\p{Is_Block:  _tibetan}', "");
Expect(0, 4095, '\p{^Is_Block:  _tibetan}', "");
Expect(0, 4095, '\P{Is_Block:  _tibetan}', "");
Expect(1, 4095, '\P{^Is_Block:  _tibetan}', "");
Expect(0, 4096, '\p{Is_Block:  _tibetan}', "");
Expect(1, 4096, '\p{^Is_Block:  _tibetan}', "");
Expect(1, 4096, '\P{Is_Block:  _tibetan}', "");
Expect(0, 4096, '\P{^Is_Block:  _tibetan}', "");
Error('\p{Is_Blk=/a/	Tibetan}');
Error('\P{Is_Blk=/a/	Tibetan}');
Expect(1, 4095, '\p{Is_Blk=tibetan}', "");
Expect(0, 4095, '\p{^Is_Blk=tibetan}', "");
Expect(0, 4095, '\P{Is_Blk=tibetan}', "");
Expect(1, 4095, '\P{^Is_Blk=tibetan}', "");
Expect(0, 4096, '\p{Is_Blk=tibetan}', "");
Expect(1, 4096, '\p{^Is_Blk=tibetan}', "");
Expect(1, 4096, '\P{Is_Blk=tibetan}', "");
Expect(0, 4096, '\P{^Is_Blk=tibetan}', "");
Expect(1, 4095, '\p{Is_Blk=-tibetan}', "");
Expect(0, 4095, '\p{^Is_Blk=-tibetan}', "");
Expect(0, 4095, '\P{Is_Blk=-tibetan}', "");
Expect(1, 4095, '\P{^Is_Blk=-tibetan}', "");
Expect(0, 4096, '\p{Is_Blk=-tibetan}', "");
Expect(1, 4096, '\p{^Is_Blk=-tibetan}', "");
Expect(1, 4096, '\P{Is_Blk=-tibetan}', "");
Expect(0, 4096, '\P{^Is_Blk=-tibetan}', "");
Error('\p{Block:   /a/ tifinagh}');
Error('\P{Block:   /a/ tifinagh}');
Expect(1, 11647, '\p{Block=tifinagh}', "");
Expect(0, 11647, '\p{^Block=tifinagh}', "");
Expect(0, 11647, '\P{Block=tifinagh}', "");
Expect(1, 11647, '\P{^Block=tifinagh}', "");
Expect(0, 11648, '\p{Block=tifinagh}', "");
Expect(1, 11648, '\p{^Block=tifinagh}', "");
Expect(1, 11648, '\P{Block=tifinagh}', "");
Expect(0, 11648, '\P{^Block=tifinagh}', "");
Expect(1, 11647, '\p{Block=-	Tifinagh}', "");
Expect(0, 11647, '\p{^Block=-	Tifinagh}', "");
Expect(0, 11647, '\P{Block=-	Tifinagh}', "");
Expect(1, 11647, '\P{^Block=-	Tifinagh}', "");
Expect(0, 11648, '\p{Block=-	Tifinagh}', "");
Expect(1, 11648, '\p{^Block=-	Tifinagh}', "");
Expect(1, 11648, '\P{Block=-	Tifinagh}', "");
Expect(0, 11648, '\P{^Block=-	Tifinagh}', "");
Error('\p{Blk=:=	Tifinagh}');
Error('\P{Blk=:=	Tifinagh}');
Expect(1, 11647, '\p{Blk:tifinagh}', "");
Expect(0, 11647, '\p{^Blk:tifinagh}', "");
Expect(0, 11647, '\P{Blk:tifinagh}', "");
Expect(1, 11647, '\P{^Blk:tifinagh}', "");
Expect(0, 11648, '\p{Blk:tifinagh}', "");
Expect(1, 11648, '\p{^Blk:tifinagh}', "");
Expect(1, 11648, '\P{Blk:tifinagh}', "");
Expect(0, 11648, '\P{^Blk:tifinagh}', "");
Expect(1, 11647, '\p{Blk=	tifinagh}', "");
Expect(0, 11647, '\p{^Blk=	tifinagh}', "");
Expect(0, 11647, '\P{Blk=	tifinagh}', "");
Expect(1, 11647, '\P{^Blk=	tifinagh}', "");
Expect(0, 11648, '\p{Blk=	tifinagh}', "");
Expect(1, 11648, '\p{^Blk=	tifinagh}', "");
Expect(1, 11648, '\P{Blk=	tifinagh}', "");
Expect(0, 11648, '\P{^Blk=	tifinagh}', "");
Error('\p{Is_Block=/a/_ TIFINAGH}');
Error('\P{Is_Block=/a/_ TIFINAGH}');
Expect(1, 11647, '\p{Is_Block=tifinagh}', "");
Expect(0, 11647, '\p{^Is_Block=tifinagh}', "");
Expect(0, 11647, '\P{Is_Block=tifinagh}', "");
Expect(1, 11647, '\P{^Is_Block=tifinagh}', "");
Expect(0, 11648, '\p{Is_Block=tifinagh}', "");
Expect(1, 11648, '\p{^Is_Block=tifinagh}', "");
Expect(1, 11648, '\P{Is_Block=tifinagh}', "");
Expect(0, 11648, '\P{^Is_Block=tifinagh}', "");
Expect(1, 11647, '\p{Is_Block= _Tifinagh}', "");
Expect(0, 11647, '\p{^Is_Block= _Tifinagh}', "");
Expect(0, 11647, '\P{Is_Block= _Tifinagh}', "");
Expect(1, 11647, '\P{^Is_Block= _Tifinagh}', "");
Expect(0, 11648, '\p{Is_Block= _Tifinagh}', "");
Expect(1, 11648, '\p{^Is_Block= _Tifinagh}', "");
Expect(1, 11648, '\P{Is_Block= _Tifinagh}', "");
Expect(0, 11648, '\P{^Is_Block= _Tifinagh}', "");
Error('\p{Is_Blk=/a/Tifinagh}');
Error('\P{Is_Blk=/a/Tifinagh}');
Expect(1, 11647, '\p{Is_Blk: tifinagh}', "");
Expect(0, 11647, '\p{^Is_Blk: tifinagh}', "");
Expect(0, 11647, '\P{Is_Blk: tifinagh}', "");
Expect(1, 11647, '\P{^Is_Blk: tifinagh}', "");
Expect(0, 11648, '\p{Is_Blk: tifinagh}', "");
Expect(1, 11648, '\p{^Is_Blk: tifinagh}', "");
Expect(1, 11648, '\P{Is_Blk: tifinagh}', "");
Expect(0, 11648, '\P{^Is_Blk: tifinagh}', "");
Expect(1, 11647, '\p{Is_Blk=--Tifinagh}', "");
Expect(0, 11647, '\p{^Is_Blk=--Tifinagh}', "");
Expect(0, 11647, '\P{Is_Blk=--Tifinagh}', "");
Expect(1, 11647, '\P{^Is_Blk=--Tifinagh}', "");
Expect(0, 11648, '\p{Is_Blk=--Tifinagh}', "");
Expect(1, 11648, '\p{^Is_Blk=--Tifinagh}', "");
Expect(1, 11648, '\P{Is_Blk=--Tifinagh}', "");
Expect(0, 11648, '\P{^Is_Blk=--Tifinagh}', "");
Error('\p{Block:	/a/- transport_AND_Map_SYMBOLS}');
Error('\P{Block:	/a/- transport_AND_Map_SYMBOLS}');
Expect(1, 128767, '\p{Block=transportandmapsymbols}', "");
Expect(0, 128767, '\p{^Block=transportandmapsymbols}', "");
Expect(0, 128767, '\P{Block=transportandmapsymbols}', "");
Expect(1, 128767, '\P{^Block=transportandmapsymbols}', "");
Expect(0, 128768, '\p{Block=transportandmapsymbols}', "");
Expect(1, 128768, '\p{^Block=transportandmapsymbols}', "");
Expect(1, 128768, '\P{Block=transportandmapsymbols}', "");
Expect(0, 128768, '\P{^Block=transportandmapsymbols}', "");
Expect(1, 128767, '\p{Block=transport_and_Map_Symbols}', "");
Expect(0, 128767, '\p{^Block=transport_and_Map_Symbols}', "");
Expect(0, 128767, '\P{Block=transport_and_Map_Symbols}', "");
Expect(1, 128767, '\P{^Block=transport_and_Map_Symbols}', "");
Expect(0, 128768, '\p{Block=transport_and_Map_Symbols}', "");
Expect(1, 128768, '\p{^Block=transport_and_Map_Symbols}', "");
Expect(1, 128768, '\P{Block=transport_and_Map_Symbols}', "");
Expect(0, 128768, '\P{^Block=transport_and_Map_Symbols}', "");
Error('\p{Blk= _Transport_And_Map:=}');
Error('\P{Blk= _Transport_And_Map:=}');
Expect(1, 128767, '\p{Blk=transportandmap}', "");
Expect(0, 128767, '\p{^Blk=transportandmap}', "");
Expect(0, 128767, '\P{Blk=transportandmap}', "");
Expect(1, 128767, '\P{^Blk=transportandmap}', "");
Expect(0, 128768, '\p{Blk=transportandmap}', "");
Expect(1, 128768, '\p{^Blk=transportandmap}', "");
Expect(1, 128768, '\P{Blk=transportandmap}', "");
Expect(0, 128768, '\P{^Blk=transportandmap}', "");
Expect(1, 128767, '\p{Blk=_	TRANSPORT_And_Map}', "");
Expect(0, 128767, '\p{^Blk=_	TRANSPORT_And_Map}', "");
Expect(0, 128767, '\P{Blk=_	TRANSPORT_And_Map}', "");
Expect(1, 128767, '\P{^Blk=_	TRANSPORT_And_Map}', "");
Expect(0, 128768, '\p{Blk=_	TRANSPORT_And_Map}', "");
Expect(1, 128768, '\p{^Blk=_	TRANSPORT_And_Map}', "");
Expect(1, 128768, '\P{Blk=_	TRANSPORT_And_Map}', "");
Expect(0, 128768, '\P{^Blk=_	TRANSPORT_And_Map}', "");
Error('\p{Is_Block= transport_AND_Map_Symbols:=}');
Error('\P{Is_Block= transport_AND_Map_Symbols:=}');
Expect(1, 128767, '\p{Is_Block:   transportandmapsymbols}', "");
Expect(0, 128767, '\p{^Is_Block:   transportandmapsymbols}', "");
Expect(0, 128767, '\P{Is_Block:   transportandmapsymbols}', "");
Expect(1, 128767, '\P{^Is_Block:   transportandmapsymbols}', "");
Expect(0, 128768, '\p{Is_Block:   transportandmapsymbols}', "");
Expect(1, 128768, '\p{^Is_Block:   transportandmapsymbols}', "");
Expect(1, 128768, '\P{Is_Block:   transportandmapsymbols}', "");
Expect(0, 128768, '\P{^Is_Block:   transportandmapsymbols}', "");
Expect(1, 128767, '\p{Is_Block=-Transport_And_Map_Symbols}', "");
Expect(0, 128767, '\p{^Is_Block=-Transport_And_Map_Symbols}', "");
Expect(0, 128767, '\P{Is_Block=-Transport_And_Map_Symbols}', "");
Expect(1, 128767, '\P{^Is_Block=-Transport_And_Map_Symbols}', "");
Expect(0, 128768, '\p{Is_Block=-Transport_And_Map_Symbols}', "");
Expect(1, 128768, '\p{^Is_Block=-Transport_And_Map_Symbols}', "");
Expect(1, 128768, '\P{Is_Block=-Transport_And_Map_Symbols}', "");
Expect(0, 128768, '\P{^Is_Block=-Transport_And_Map_Symbols}', "");
Error('\p{Is_Blk:   	:=Transport_And_Map}');
Error('\P{Is_Blk:   	:=Transport_And_Map}');
Expect(1, 128767, '\p{Is_Blk=transportandmap}', "");
Expect(0, 128767, '\p{^Is_Blk=transportandmap}', "");
Expect(0, 128767, '\P{Is_Blk=transportandmap}', "");
Expect(1, 128767, '\P{^Is_Blk=transportandmap}', "");
Expect(0, 128768, '\p{Is_Blk=transportandmap}', "");
Expect(1, 128768, '\p{^Is_Blk=transportandmap}', "");
Expect(1, 128768, '\P{Is_Blk=transportandmap}', "");
Expect(0, 128768, '\P{^Is_Blk=transportandmap}', "");
Expect(1, 128767, '\p{Is_Blk=	Transport_and_Map}', "");
Expect(0, 128767, '\p{^Is_Blk=	Transport_and_Map}', "");
Expect(0, 128767, '\P{Is_Blk=	Transport_and_Map}', "");
Expect(1, 128767, '\P{^Is_Blk=	Transport_and_Map}', "");
Expect(0, 128768, '\p{Is_Blk=	Transport_and_Map}', "");
Expect(1, 128768, '\p{^Is_Blk=	Transport_and_Map}', "");
Expect(1, 128768, '\P{Is_Blk=	Transport_and_Map}', "");
Expect(0, 128768, '\P{^Is_Blk=	Transport_and_Map}', "");
Error('\p{Block= :=Unified_canadian_ABORIGINAL_syllabics}');
Error('\P{Block= :=Unified_canadian_ABORIGINAL_syllabics}');
Expect(1, 5759, '\p{Block=unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\p{^Block=unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\P{Block=unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\P{^Block=unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\p{Block=unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\p{^Block=unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\P{Block=unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\P{^Block=unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\p{Block=- Unified_Canadian_Aboriginal_Syllabics}', "");
Expect(0, 5759, '\p{^Block=- Unified_Canadian_Aboriginal_Syllabics}', "");
Expect(0, 5759, '\P{Block=- Unified_Canadian_Aboriginal_Syllabics}', "");
Expect(1, 5759, '\P{^Block=- Unified_Canadian_Aboriginal_Syllabics}', "");
Expect(0, 5760, '\p{Block=- Unified_Canadian_Aboriginal_Syllabics}', "");
Expect(1, 5760, '\p{^Block=- Unified_Canadian_Aboriginal_Syllabics}', "");
Expect(1, 5760, '\P{Block=- Unified_Canadian_Aboriginal_Syllabics}', "");
Expect(0, 5760, '\P{^Block=- Unified_Canadian_Aboriginal_Syllabics}', "");
Error('\p{Blk:/a/_ucas}');
Error('\P{Blk:/a/_ucas}');
Expect(1, 5759, '\p{Blk=ucas}', "");
Expect(0, 5759, '\p{^Blk=ucas}', "");
Expect(0, 5759, '\P{Blk=ucas}', "");
Expect(1, 5759, '\P{^Blk=ucas}', "");
Expect(0, 5760, '\p{Blk=ucas}', "");
Expect(1, 5760, '\p{^Blk=ucas}', "");
Expect(1, 5760, '\P{Blk=ucas}', "");
Expect(0, 5760, '\P{^Blk=ucas}', "");
Expect(1, 5759, '\p{Blk= _ucas}', "");
Expect(0, 5759, '\p{^Blk= _ucas}', "");
Expect(0, 5759, '\P{Blk= _ucas}', "");
Expect(1, 5759, '\P{^Blk= _ucas}', "");
Expect(0, 5760, '\p{Blk= _ucas}', "");
Expect(1, 5760, '\p{^Blk= _ucas}', "");
Expect(1, 5760, '\P{Blk= _ucas}', "");
Expect(0, 5760, '\P{^Blk= _ucas}', "");
Error('\p{Is_Block=-/a/canadian_Syllabics}');
Error('\P{Is_Block=-/a/canadian_Syllabics}');
Expect(1, 5759, '\p{Is_Block=canadiansyllabics}', "");
Expect(0, 5759, '\p{^Is_Block=canadiansyllabics}', "");
Expect(0, 5759, '\P{Is_Block=canadiansyllabics}', "");
Expect(1, 5759, '\P{^Is_Block=canadiansyllabics}', "");
Expect(0, 5760, '\p{Is_Block=canadiansyllabics}', "");
Expect(1, 5760, '\p{^Is_Block=canadiansyllabics}', "");
Expect(1, 5760, '\P{Is_Block=canadiansyllabics}', "");
Expect(0, 5760, '\P{^Is_Block=canadiansyllabics}', "");
Expect(1, 5759, '\p{Is_Block=- CANADIAN_syllabics}', "");
Expect(0, 5759, '\p{^Is_Block=- CANADIAN_syllabics}', "");
Expect(0, 5759, '\P{Is_Block=- CANADIAN_syllabics}', "");
Expect(1, 5759, '\P{^Is_Block=- CANADIAN_syllabics}', "");
Expect(0, 5760, '\p{Is_Block=- CANADIAN_syllabics}', "");
Expect(1, 5760, '\p{^Is_Block=- CANADIAN_syllabics}', "");
Expect(1, 5760, '\P{Is_Block=- CANADIAN_syllabics}', "");
Expect(0, 5760, '\P{^Is_Block=- CANADIAN_syllabics}', "");
Error('\p{Is_Blk=-Unified_Canadian_ABORIGINAL_Syllabics/a/}');
Error('\P{Is_Blk=-Unified_Canadian_ABORIGINAL_Syllabics/a/}');
Expect(1, 5759, '\p{Is_Blk=unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\p{^Is_Blk=unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\P{Is_Blk=unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\P{^Is_Blk=unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\p{Is_Blk=unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\p{^Is_Blk=unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\P{Is_Blk=unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\P{^Is_Blk=unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\p{Is_Blk=_unified_Canadian_Aboriginal_syllabics}', "");
Expect(0, 5759, '\p{^Is_Blk=_unified_Canadian_Aboriginal_syllabics}', "");
Expect(0, 5759, '\P{Is_Blk=_unified_Canadian_Aboriginal_syllabics}', "");
Expect(1, 5759, '\P{^Is_Blk=_unified_Canadian_Aboriginal_syllabics}', "");
Expect(0, 5760, '\p{Is_Blk=_unified_Canadian_Aboriginal_syllabics}', "");
Expect(1, 5760, '\p{^Is_Blk=_unified_Canadian_Aboriginal_syllabics}', "");
Expect(1, 5760, '\P{Is_Blk=_unified_Canadian_Aboriginal_syllabics}', "");
Expect(0, 5760, '\P{^Is_Blk=_unified_Canadian_Aboriginal_syllabics}', "");
Error('\p{Block=-	UNIFIED_Canadian_ABORIGINAL_Syllabics_Extended/a/}');
Error('\P{Block=-	UNIFIED_Canadian_ABORIGINAL_Syllabics_Extended/a/}');
Expect(1, 6399, '\p{Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\p{^Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\P{Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\P{^Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\p{Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\p{^Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\P{Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\P{^Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\p{Block=_	Unified_Canadian_ABORIGINAL_Syllabics_Extended}', "");
Expect(0, 6399, '\p{^Block=_	Unified_Canadian_ABORIGINAL_Syllabics_Extended}', "");
Expect(0, 6399, '\P{Block=_	Unified_Canadian_ABORIGINAL_Syllabics_Extended}', "");
Expect(1, 6399, '\P{^Block=_	Unified_Canadian_ABORIGINAL_Syllabics_Extended}', "");
Expect(0, 6400, '\p{Block=_	Unified_Canadian_ABORIGINAL_Syllabics_Extended}', "");
Expect(1, 6400, '\p{^Block=_	Unified_Canadian_ABORIGINAL_Syllabics_Extended}', "");
Expect(1, 6400, '\P{Block=_	Unified_Canadian_ABORIGINAL_Syllabics_Extended}', "");
Expect(0, 6400, '\P{^Block=_	Unified_Canadian_ABORIGINAL_Syllabics_Extended}', "");
Error('\p{Blk=_ ucas_EXT:=}');
Error('\P{Blk=_ ucas_EXT:=}');
Expect(1, 6399, '\p{Blk=ucasext}', "");
Expect(0, 6399, '\p{^Blk=ucasext}', "");
Expect(0, 6399, '\P{Blk=ucasext}', "");
Expect(1, 6399, '\P{^Blk=ucasext}', "");
Expect(0, 6400, '\p{Blk=ucasext}', "");
Expect(1, 6400, '\p{^Blk=ucasext}', "");
Expect(1, 6400, '\P{Blk=ucasext}', "");
Expect(0, 6400, '\P{^Blk=ucasext}', "");
Expect(1, 6399, '\p{Blk= _UCAS_Ext}', "");
Expect(0, 6399, '\p{^Blk= _UCAS_Ext}', "");
Expect(0, 6399, '\P{Blk= _UCAS_Ext}', "");
Expect(1, 6399, '\P{^Blk= _UCAS_Ext}', "");
Expect(0, 6400, '\p{Blk= _UCAS_Ext}', "");
Expect(1, 6400, '\p{^Blk= _UCAS_Ext}', "");
Expect(1, 6400, '\P{Blk= _UCAS_Ext}', "");
Expect(0, 6400, '\P{^Blk= _UCAS_Ext}', "");
Error('\p{Is_Block=	-Unified_Canadian_Aboriginal_syllabics_EXTENDED:=}');
Error('\P{Is_Block=	-Unified_Canadian_Aboriginal_syllabics_EXTENDED:=}');
Expect(1, 6399, '\p{Is_Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\p{^Is_Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\P{Is_Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\P{^Is_Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\p{Is_Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\p{^Is_Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\P{Is_Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\P{^Is_Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\p{Is_Block= unified_CANADIAN_aboriginal_syllabics_Extended}', "");
Expect(0, 6399, '\p{^Is_Block= unified_CANADIAN_aboriginal_syllabics_Extended}', "");
Expect(0, 6399, '\P{Is_Block= unified_CANADIAN_aboriginal_syllabics_Extended}', "");
Expect(1, 6399, '\P{^Is_Block= unified_CANADIAN_aboriginal_syllabics_Extended}', "");
Expect(0, 6400, '\p{Is_Block= unified_CANADIAN_aboriginal_syllabics_Extended}', "");
Expect(1, 6400, '\p{^Is_Block= unified_CANADIAN_aboriginal_syllabics_Extended}', "");
Expect(1, 6400, '\P{Is_Block= unified_CANADIAN_aboriginal_syllabics_Extended}', "");
Expect(0, 6400, '\P{^Is_Block= unified_CANADIAN_aboriginal_syllabics_Extended}', "");
Error('\p{Is_Blk: :=UCAS_Ext}');
Error('\P{Is_Blk: :=UCAS_Ext}');
Expect(1, 6399, '\p{Is_Blk=ucasext}', "");
Expect(0, 6399, '\p{^Is_Blk=ucasext}', "");
Expect(0, 6399, '\P{Is_Blk=ucasext}', "");
Expect(1, 6399, '\P{^Is_Blk=ucasext}', "");
Expect(0, 6400, '\p{Is_Blk=ucasext}', "");
Expect(1, 6400, '\p{^Is_Blk=ucasext}', "");
Expect(1, 6400, '\P{Is_Blk=ucasext}', "");
Expect(0, 6400, '\P{^Is_Blk=ucasext}', "");
Expect(1, 6399, '\p{Is_Blk=-ucas_EXT}', "");
Expect(0, 6399, '\p{^Is_Blk=-ucas_EXT}', "");
Expect(0, 6399, '\P{Is_Blk=-ucas_EXT}', "");
Expect(1, 6399, '\P{^Is_Blk=-ucas_EXT}', "");
Expect(0, 6400, '\p{Is_Blk=-ucas_EXT}', "");
Expect(1, 6400, '\p{^Is_Blk=-ucas_EXT}', "");
Expect(1, 6400, '\P{Is_Blk=-ucas_EXT}', "");
Expect(0, 6400, '\P{^Is_Blk=-ucas_EXT}', "");
Error('\p{Block=	:=ugaritic}');
Error('\P{Block=	:=ugaritic}');
Expect(1, 66463, '\p{Block:	ugaritic}', "");
Expect(0, 66463, '\p{^Block:	ugaritic}', "");
Expect(0, 66463, '\P{Block:	ugaritic}', "");
Expect(1, 66463, '\P{^Block:	ugaritic}', "");
Expect(0, 66464, '\p{Block:	ugaritic}', "");
Expect(1, 66464, '\p{^Block:	ugaritic}', "");
Expect(1, 66464, '\P{Block:	ugaritic}', "");
Expect(0, 66464, '\P{^Block:	ugaritic}', "");
Expect(1, 66463, '\p{Block=_ UGARITIC}', "");
Expect(0, 66463, '\p{^Block=_ UGARITIC}', "");
Expect(0, 66463, '\P{Block=_ UGARITIC}', "");
Expect(1, 66463, '\P{^Block=_ UGARITIC}', "");
Expect(0, 66464, '\p{Block=_ UGARITIC}', "");
Expect(1, 66464, '\p{^Block=_ UGARITIC}', "");
Expect(1, 66464, '\P{Block=_ UGARITIC}', "");
Expect(0, 66464, '\P{^Block=_ UGARITIC}', "");
Error('\p{Blk=-Ugaritic:=}');
Error('\P{Blk=-Ugaritic:=}');
Expect(1, 66463, '\p{Blk=ugaritic}', "");
Expect(0, 66463, '\p{^Blk=ugaritic}', "");
Expect(0, 66463, '\P{Blk=ugaritic}', "");
Expect(1, 66463, '\P{^Blk=ugaritic}', "");
Expect(0, 66464, '\p{Blk=ugaritic}', "");
Expect(1, 66464, '\p{^Blk=ugaritic}', "");
Expect(1, 66464, '\P{Blk=ugaritic}', "");
Expect(0, 66464, '\P{^Blk=ugaritic}', "");
Expect(1, 66463, '\p{Blk=  UGARITIC}', "");
Expect(0, 66463, '\p{^Blk=  UGARITIC}', "");
Expect(0, 66463, '\P{Blk=  UGARITIC}', "");
Expect(1, 66463, '\P{^Blk=  UGARITIC}', "");
Expect(0, 66464, '\p{Blk=  UGARITIC}', "");
Expect(1, 66464, '\p{^Blk=  UGARITIC}', "");
Expect(1, 66464, '\P{Blk=  UGARITIC}', "");
Expect(0, 66464, '\P{^Blk=  UGARITIC}', "");
Error('\p{Is_Block=_-UGARITIC/a/}');
Error('\P{Is_Block=_-UGARITIC/a/}');
Expect(1, 66463, '\p{Is_Block=ugaritic}', "");
Expect(0, 66463, '\p{^Is_Block=ugaritic}', "");
Expect(0, 66463, '\P{Is_Block=ugaritic}', "");
Expect(1, 66463, '\P{^Is_Block=ugaritic}', "");
Expect(0, 66464, '\p{Is_Block=ugaritic}', "");
Expect(1, 66464, '\p{^Is_Block=ugaritic}', "");
Expect(1, 66464, '\P{Is_Block=ugaritic}', "");
Expect(0, 66464, '\P{^Is_Block=ugaritic}', "");
Expect(1, 66463, '\p{Is_Block= -ugaritic}', "");
Expect(0, 66463, '\p{^Is_Block= -ugaritic}', "");
Expect(0, 66463, '\P{Is_Block= -ugaritic}', "");
Expect(1, 66463, '\P{^Is_Block= -ugaritic}', "");
Expect(0, 66464, '\p{Is_Block= -ugaritic}', "");
Expect(1, 66464, '\p{^Is_Block= -ugaritic}', "");
Expect(1, 66464, '\P{Is_Block= -ugaritic}', "");
Expect(0, 66464, '\P{^Is_Block= -ugaritic}', "");
Error('\p{Is_Blk=:= Ugaritic}');
Error('\P{Is_Blk=:= Ugaritic}');
Expect(1, 66463, '\p{Is_Blk:ugaritic}', "");
Expect(0, 66463, '\p{^Is_Blk:ugaritic}', "");
Expect(0, 66463, '\P{Is_Blk:ugaritic}', "");
Expect(1, 66463, '\P{^Is_Blk:ugaritic}', "");
Expect(0, 66464, '\p{Is_Blk:ugaritic}', "");
Expect(1, 66464, '\p{^Is_Blk:ugaritic}', "");
Expect(1, 66464, '\P{Is_Blk:ugaritic}', "");
Expect(0, 66464, '\P{^Is_Blk:ugaritic}', "");
Expect(1, 66463, '\p{Is_Blk=-Ugaritic}', "");
Expect(0, 66463, '\p{^Is_Blk=-Ugaritic}', "");
Expect(0, 66463, '\P{Is_Blk=-Ugaritic}', "");
Expect(1, 66463, '\P{^Is_Blk=-Ugaritic}', "");
Expect(0, 66464, '\p{Is_Blk=-Ugaritic}', "");
Expect(1, 66464, '\p{^Is_Blk=-Ugaritic}', "");
Expect(1, 66464, '\P{Is_Blk=-Ugaritic}', "");
Expect(0, 66464, '\P{^Is_Blk=-Ugaritic}', "");
Error('\p{Block=/a/variation_SELECTORS}');
Error('\P{Block=/a/variation_SELECTORS}');
Expect(1, 65039, '\p{Block=variationselectors}', "");
Expect(0, 65039, '\p{^Block=variationselectors}', "");
Expect(0, 65039, '\P{Block=variationselectors}', "");
Expect(1, 65039, '\P{^Block=variationselectors}', "");
Expect(0, 65040, '\p{Block=variationselectors}', "");
Expect(1, 65040, '\p{^Block=variationselectors}', "");
Expect(1, 65040, '\P{Block=variationselectors}', "");
Expect(0, 65040, '\P{^Block=variationselectors}', "");
Expect(1, 65039, '\p{Block= 	Variation_SELECTORS}', "");
Expect(0, 65039, '\p{^Block= 	Variation_SELECTORS}', "");
Expect(0, 65039, '\P{Block= 	Variation_SELECTORS}', "");
Expect(1, 65039, '\P{^Block= 	Variation_SELECTORS}', "");
Expect(0, 65040, '\p{Block= 	Variation_SELECTORS}', "");
Expect(1, 65040, '\p{^Block= 	Variation_SELECTORS}', "");
Expect(1, 65040, '\P{Block= 	Variation_SELECTORS}', "");
Expect(0, 65040, '\P{^Block= 	Variation_SELECTORS}', "");
Error('\p{Blk=:=_VS}');
Error('\P{Blk=:=_VS}');
Expect(1, 65039, '\p{Blk:vs}', "");
Expect(0, 65039, '\p{^Blk:vs}', "");
Expect(0, 65039, '\P{Blk:vs}', "");
Expect(1, 65039, '\P{^Blk:vs}', "");
Expect(0, 65040, '\p{Blk:vs}', "");
Expect(1, 65040, '\p{^Blk:vs}', "");
Expect(1, 65040, '\P{Blk:vs}', "");
Expect(0, 65040, '\P{^Blk:vs}', "");
Expect(1, 65039, '\p{Blk=_ VS}', "");
Expect(0, 65039, '\p{^Blk=_ VS}', "");
Expect(0, 65039, '\P{Blk=_ VS}', "");
Expect(1, 65039, '\P{^Blk=_ VS}', "");
Expect(0, 65040, '\p{Blk=_ VS}', "");
Expect(1, 65040, '\p{^Blk=_ VS}', "");
Expect(1, 65040, '\P{Blk=_ VS}', "");
Expect(0, 65040, '\P{^Blk=_ VS}', "");
Error('\p{Is_Block=- Variation_SELECTORS:=}');
Error('\P{Is_Block=- Variation_SELECTORS:=}');
Expect(1, 65039, '\p{Is_Block=variationselectors}', "");
Expect(0, 65039, '\p{^Is_Block=variationselectors}', "");
Expect(0, 65039, '\P{Is_Block=variationselectors}', "");
Expect(1, 65039, '\P{^Is_Block=variationselectors}', "");
Expect(0, 65040, '\p{Is_Block=variationselectors}', "");
Expect(1, 65040, '\p{^Is_Block=variationselectors}', "");
Expect(1, 65040, '\P{Is_Block=variationselectors}', "");
Expect(0, 65040, '\P{^Is_Block=variationselectors}', "");
Expect(1, 65039, '\p{Is_Block=_	Variation_Selectors}', "");
Expect(0, 65039, '\p{^Is_Block=_	Variation_Selectors}', "");
Expect(0, 65039, '\P{Is_Block=_	Variation_Selectors}', "");
Expect(1, 65039, '\P{^Is_Block=_	Variation_Selectors}', "");
Expect(0, 65040, '\p{Is_Block=_	Variation_Selectors}', "");
Expect(1, 65040, '\p{^Is_Block=_	Variation_Selectors}', "");
Expect(1, 65040, '\P{Is_Block=_	Variation_Selectors}', "");
Expect(0, 65040, '\P{^Is_Block=_	Variation_Selectors}', "");
Error('\p{Is_Blk=__VS/a/}');
Error('\P{Is_Blk=__VS/a/}');
Expect(1, 65039, '\p{Is_Blk=vs}', "");
Expect(0, 65039, '\p{^Is_Blk=vs}', "");
Expect(0, 65039, '\P{Is_Blk=vs}', "");
Expect(1, 65039, '\P{^Is_Blk=vs}', "");
Expect(0, 65040, '\p{Is_Blk=vs}', "");
Expect(1, 65040, '\p{^Is_Blk=vs}', "");
Expect(1, 65040, '\P{Is_Blk=vs}', "");
Expect(0, 65040, '\P{^Is_Blk=vs}', "");
Expect(1, 65039, '\p{Is_Blk:   _ VS}', "");
Expect(0, 65039, '\p{^Is_Blk:   _ VS}', "");
Expect(0, 65039, '\P{Is_Blk:   _ VS}', "");
Expect(1, 65039, '\P{^Is_Blk:   _ VS}', "");
Expect(0, 65040, '\p{Is_Blk:   _ VS}', "");
Expect(1, 65040, '\p{^Is_Blk:   _ VS}', "");
Expect(1, 65040, '\P{Is_Blk:   _ VS}', "");
Expect(0, 65040, '\P{^Is_Blk:   _ VS}', "");
Error('\p{Block=/a/	VARIATION_Selectors_Supplement}');
Error('\P{Block=/a/	VARIATION_Selectors_Supplement}');
Expect(1, 917999, '\p{Block=variationselectorssupplement}', "");
Expect(0, 917999, '\p{^Block=variationselectorssupplement}', "");
Expect(0, 917999, '\P{Block=variationselectorssupplement}', "");
Expect(1, 917999, '\P{^Block=variationselectorssupplement}', "");
Expect(0, 918000, '\p{Block=variationselectorssupplement}', "");
Expect(1, 918000, '\p{^Block=variationselectorssupplement}', "");
Expect(1, 918000, '\P{Block=variationselectorssupplement}', "");
Expect(0, 918000, '\P{^Block=variationselectorssupplement}', "");
Expect(1, 917999, '\p{Block=- Variation_Selectors_Supplement}', "");
Expect(0, 917999, '\p{^Block=- Variation_Selectors_Supplement}', "");
Expect(0, 917999, '\P{Block=- Variation_Selectors_Supplement}', "");
Expect(1, 917999, '\P{^Block=- Variation_Selectors_Supplement}', "");
Expect(0, 918000, '\p{Block=- Variation_Selectors_Supplement}', "");
Expect(1, 918000, '\p{^Block=- Variation_Selectors_Supplement}', "");
Expect(1, 918000, '\P{Block=- Variation_Selectors_Supplement}', "");
Expect(0, 918000, '\P{^Block=- Variation_Selectors_Supplement}', "");
Error('\p{Blk= /a/vs_Sup}');
Error('\P{Blk= /a/vs_Sup}');
Expect(1, 917999, '\p{Blk=vssup}', "");
Expect(0, 917999, '\p{^Blk=vssup}', "");
Expect(0, 917999, '\P{Blk=vssup}', "");
Expect(1, 917999, '\P{^Blk=vssup}', "");
Expect(0, 918000, '\p{Blk=vssup}', "");
Expect(1, 918000, '\p{^Blk=vssup}', "");
Expect(1, 918000, '\P{Blk=vssup}', "");
Expect(0, 918000, '\P{^Blk=vssup}', "");
Expect(1, 917999, '\p{Blk= VS_Sup}', "");
Expect(0, 917999, '\p{^Blk= VS_Sup}', "");
Expect(0, 917999, '\P{Blk= VS_Sup}', "");
Expect(1, 917999, '\P{^Blk= VS_Sup}', "");
Expect(0, 918000, '\p{Blk= VS_Sup}', "");
Expect(1, 918000, '\p{^Blk= VS_Sup}', "");
Expect(1, 918000, '\P{Blk= VS_Sup}', "");
Expect(0, 918000, '\P{^Blk= VS_Sup}', "");
Error('\p{Is_Block=/a/		Variation_Selectors_supplement}');
Error('\P{Is_Block=/a/		Variation_Selectors_supplement}');
Expect(1, 917999, '\p{Is_Block=variationselectorssupplement}', "");
Expect(0, 917999, '\p{^Is_Block=variationselectorssupplement}', "");
Expect(0, 917999, '\P{Is_Block=variationselectorssupplement}', "");
Expect(1, 917999, '\P{^Is_Block=variationselectorssupplement}', "");
Expect(0, 918000, '\p{Is_Block=variationselectorssupplement}', "");
Expect(1, 918000, '\p{^Is_Block=variationselectorssupplement}', "");
Expect(1, 918000, '\P{Is_Block=variationselectorssupplement}', "");
Expect(0, 918000, '\P{^Is_Block=variationselectorssupplement}', "");
Expect(1, 917999, '\p{Is_Block=--VARIATION_Selectors_Supplement}', "");
Expect(0, 917999, '\p{^Is_Block=--VARIATION_Selectors_Supplement}', "");
Expect(0, 917999, '\P{Is_Block=--VARIATION_Selectors_Supplement}', "");
Expect(1, 917999, '\P{^Is_Block=--VARIATION_Selectors_Supplement}', "");
Expect(0, 918000, '\p{Is_Block=--VARIATION_Selectors_Supplement}', "");
Expect(1, 918000, '\p{^Is_Block=--VARIATION_Selectors_Supplement}', "");
Expect(1, 918000, '\P{Is_Block=--VARIATION_Selectors_Supplement}', "");
Expect(0, 918000, '\P{^Is_Block=--VARIATION_Selectors_Supplement}', "");
Error('\p{Is_Blk=:=VS_sup}');
Error('\P{Is_Blk=:=VS_sup}');
Expect(1, 917999, '\p{Is_Blk=vssup}', "");
Expect(0, 917999, '\p{^Is_Blk=vssup}', "");
Expect(0, 917999, '\P{Is_Blk=vssup}', "");
Expect(1, 917999, '\P{^Is_Blk=vssup}', "");
Expect(0, 918000, '\p{Is_Blk=vssup}', "");
Expect(1, 918000, '\p{^Is_Blk=vssup}', "");
Expect(1, 918000, '\P{Is_Blk=vssup}', "");
Expect(0, 918000, '\P{^Is_Blk=vssup}', "");
Expect(1, 917999, '\p{Is_Blk=-	VS_Sup}', "");
Expect(0, 917999, '\p{^Is_Blk=-	VS_Sup}', "");
Expect(0, 917999, '\P{Is_Blk=-	VS_Sup}', "");
Expect(1, 917999, '\P{^Is_Blk=-	VS_Sup}', "");
Expect(0, 918000, '\p{Is_Blk=-	VS_Sup}', "");
Expect(1, 918000, '\p{^Is_Blk=-	VS_Sup}', "");
Expect(1, 918000, '\P{Is_Blk=-	VS_Sup}', "");
Expect(0, 918000, '\P{^Is_Blk=-	VS_Sup}', "");
Error('\p{Block= _vai/a/}');
Error('\P{Block= _vai/a/}');
Expect(1, 42559, '\p{Block=vai}', "");
Expect(0, 42559, '\p{^Block=vai}', "");
Expect(0, 42559, '\P{Block=vai}', "");
Expect(1, 42559, '\P{^Block=vai}', "");
Expect(0, 42560, '\p{Block=vai}', "");
Expect(1, 42560, '\p{^Block=vai}', "");
Expect(1, 42560, '\P{Block=vai}', "");
Expect(0, 42560, '\P{^Block=vai}', "");
Expect(1, 42559, '\p{Block=	_vai}', "");
Expect(0, 42559, '\p{^Block=	_vai}', "");
Expect(0, 42559, '\P{Block=	_vai}', "");
Expect(1, 42559, '\P{^Block=	_vai}', "");
Expect(0, 42560, '\p{Block=	_vai}', "");
Expect(1, 42560, '\p{^Block=	_vai}', "");
Expect(1, 42560, '\P{Block=	_vai}', "");
Expect(0, 42560, '\P{^Block=	_vai}', "");
Error('\p{Blk=-	Vai:=}');
Error('\P{Blk=-	Vai:=}');
Expect(1, 42559, '\p{Blk=vai}', "");
Expect(0, 42559, '\p{^Blk=vai}', "");
Expect(0, 42559, '\P{Blk=vai}', "");
Expect(1, 42559, '\P{^Blk=vai}', "");
Expect(0, 42560, '\p{Blk=vai}', "");
Expect(1, 42560, '\p{^Blk=vai}', "");
Expect(1, 42560, '\P{Blk=vai}', "");
Expect(0, 42560, '\P{^Blk=vai}', "");
Expect(1, 42559, '\p{Blk= -vai}', "");
Expect(0, 42559, '\p{^Blk= -vai}', "");
Expect(0, 42559, '\P{Blk= -vai}', "");
Expect(1, 42559, '\P{^Blk= -vai}', "");
Expect(0, 42560, '\p{Blk= -vai}', "");
Expect(1, 42560, '\p{^Blk= -vai}', "");
Expect(1, 42560, '\P{Blk= -vai}', "");
Expect(0, 42560, '\P{^Blk= -vai}', "");
Error('\p{Is_Block= /a/Vai}');
Error('\P{Is_Block= /a/Vai}');
Expect(1, 42559, '\p{Is_Block=vai}', "");
Expect(0, 42559, '\p{^Is_Block=vai}', "");
Expect(0, 42559, '\P{Is_Block=vai}', "");
Expect(1, 42559, '\P{^Is_Block=vai}', "");
Expect(0, 42560, '\p{Is_Block=vai}', "");
Expect(1, 42560, '\p{^Is_Block=vai}', "");
Expect(1, 42560, '\P{Is_Block=vai}', "");
Expect(0, 42560, '\P{^Is_Block=vai}', "");
Expect(1, 42559, '\p{Is_Block=__Vai}', "");
Expect(0, 42559, '\p{^Is_Block=__Vai}', "");
Expect(0, 42559, '\P{Is_Block=__Vai}', "");
Expect(1, 42559, '\P{^Is_Block=__Vai}', "");
Expect(0, 42560, '\p{Is_Block=__Vai}', "");
Expect(1, 42560, '\p{^Is_Block=__Vai}', "");
Expect(1, 42560, '\P{Is_Block=__Vai}', "");
Expect(0, 42560, '\P{^Is_Block=__Vai}', "");
Error('\p{Is_Blk=_:=vai}');
Error('\P{Is_Blk=_:=vai}');
Expect(1, 42559, '\p{Is_Blk=vai}', "");
Expect(0, 42559, '\p{^Is_Blk=vai}', "");
Expect(0, 42559, '\P{Is_Blk=vai}', "");
Expect(1, 42559, '\P{^Is_Blk=vai}', "");
Expect(0, 42560, '\p{Is_Blk=vai}', "");
Expect(1, 42560, '\p{^Is_Blk=vai}', "");
Expect(1, 42560, '\P{Is_Blk=vai}', "");
Expect(0, 42560, '\P{^Is_Blk=vai}', "");
Expect(1, 42559, '\p{Is_Blk=_-VAI}', "");
Expect(0, 42559, '\p{^Is_Blk=_-VAI}', "");
Expect(0, 42559, '\P{Is_Blk=_-VAI}', "");
Expect(1, 42559, '\P{^Is_Blk=_-VAI}', "");
Expect(0, 42560, '\p{Is_Blk=_-VAI}', "");
Expect(1, 42560, '\p{^Is_Blk=_-VAI}', "");
Expect(1, 42560, '\P{Is_Blk=_-VAI}', "");
Expect(0, 42560, '\P{^Is_Blk=_-VAI}', "");
Error('\p{Block=_/a/Vedic_EXTENSIONS}');
Error('\P{Block=_/a/Vedic_EXTENSIONS}');
Expect(1, 7423, '\p{Block=vedicextensions}', "");
Expect(0, 7423, '\p{^Block=vedicextensions}', "");
Expect(0, 7423, '\P{Block=vedicextensions}', "");
Expect(1, 7423, '\P{^Block=vedicextensions}', "");
Expect(0, 7424, '\p{Block=vedicextensions}', "");
Expect(1, 7424, '\p{^Block=vedicextensions}', "");
Expect(1, 7424, '\P{Block=vedicextensions}', "");
Expect(0, 7424, '\P{^Block=vedicextensions}', "");
Expect(1, 7423, '\p{Block= _Vedic_Extensions}', "");
Expect(0, 7423, '\p{^Block= _Vedic_Extensions}', "");
Expect(0, 7423, '\P{Block= _Vedic_Extensions}', "");
Expect(1, 7423, '\P{^Block= _Vedic_Extensions}', "");
Expect(0, 7424, '\p{Block= _Vedic_Extensions}', "");
Expect(1, 7424, '\p{^Block= _Vedic_Extensions}', "");
Expect(1, 7424, '\P{Block= _Vedic_Extensions}', "");
Expect(0, 7424, '\P{^Block= _Vedic_Extensions}', "");
Error('\p{Blk= :=VEDIC_EXT}');
Error('\P{Blk= :=VEDIC_EXT}');
Expect(1, 7423, '\p{Blk=vedicext}', "");
Expect(0, 7423, '\p{^Blk=vedicext}', "");
Expect(0, 7423, '\P{Blk=vedicext}', "");
Expect(1, 7423, '\P{^Blk=vedicext}', "");
Expect(0, 7424, '\p{Blk=vedicext}', "");
Expect(1, 7424, '\p{^Blk=vedicext}', "");
Expect(1, 7424, '\P{Blk=vedicext}', "");
Expect(0, 7424, '\P{^Blk=vedicext}', "");
Expect(1, 7423, '\p{Blk=_Vedic_EXT}', "");
Expect(0, 7423, '\p{^Blk=_Vedic_EXT}', "");
Expect(0, 7423, '\P{Blk=_Vedic_EXT}', "");
Expect(1, 7423, '\P{^Blk=_Vedic_EXT}', "");
Expect(0, 7424, '\p{Blk=_Vedic_EXT}', "");
Expect(1, 7424, '\p{^Blk=_Vedic_EXT}', "");
Expect(1, 7424, '\P{Blk=_Vedic_EXT}', "");
Expect(0, 7424, '\P{^Blk=_Vedic_EXT}', "");
Error('\p{Is_Block=:=	VEDIC_EXTENSIONS}');
Error('\P{Is_Block=:=	VEDIC_EXTENSIONS}');
Expect(1, 7423, '\p{Is_Block=vedicextensions}', "");
Expect(0, 7423, '\p{^Is_Block=vedicextensions}', "");
Expect(0, 7423, '\P{Is_Block=vedicextensions}', "");
Expect(1, 7423, '\P{^Is_Block=vedicextensions}', "");
Expect(0, 7424, '\p{Is_Block=vedicextensions}', "");
Expect(1, 7424, '\p{^Is_Block=vedicextensions}', "");
Expect(1, 7424, '\P{Is_Block=vedicextensions}', "");
Expect(0, 7424, '\P{^Is_Block=vedicextensions}', "");
Expect(1, 7423, '\p{Is_Block=--VEDIC_EXTENSIONS}', "");
Expect(0, 7423, '\p{^Is_Block=--VEDIC_EXTENSIONS}', "");
Expect(0, 7423, '\P{Is_Block=--VEDIC_EXTENSIONS}', "");
Expect(1, 7423, '\P{^Is_Block=--VEDIC_EXTENSIONS}', "");
Expect(0, 7424, '\p{Is_Block=--VEDIC_EXTENSIONS}', "");
Expect(1, 7424, '\p{^Is_Block=--VEDIC_EXTENSIONS}', "");
Expect(1, 7424, '\P{Is_Block=--VEDIC_EXTENSIONS}', "");
Expect(0, 7424, '\P{^Is_Block=--VEDIC_EXTENSIONS}', "");
Error('\p{Is_Blk= Vedic_ext/a/}');
Error('\P{Is_Blk= Vedic_ext/a/}');
Expect(1, 7423, '\p{Is_Blk=vedicext}', "");
Expect(0, 7423, '\p{^Is_Blk=vedicext}', "");
Expect(0, 7423, '\P{Is_Blk=vedicext}', "");
Expect(1, 7423, '\P{^Is_Blk=vedicext}', "");
Expect(0, 7424, '\p{Is_Blk=vedicext}', "");
Expect(1, 7424, '\p{^Is_Blk=vedicext}', "");
Expect(1, 7424, '\P{Is_Blk=vedicext}', "");
Expect(0, 7424, '\P{^Is_Blk=vedicext}', "");
Expect(1, 7423, '\p{Is_Blk=-VEDIC_Ext}', "");
Expect(0, 7423, '\p{^Is_Blk=-VEDIC_Ext}', "");
Expect(0, 7423, '\P{Is_Blk=-VEDIC_Ext}', "");
Expect(1, 7423, '\P{^Is_Blk=-VEDIC_Ext}', "");
Expect(0, 7424, '\p{Is_Blk=-VEDIC_Ext}', "");
Expect(1, 7424, '\p{^Is_Blk=-VEDIC_Ext}', "");
Expect(1, 7424, '\P{Is_Blk=-VEDIC_Ext}', "");
Expect(0, 7424, '\P{^Is_Blk=-VEDIC_Ext}', "");
Error('\p{Block:  	vertical_forms/a/}');
Error('\P{Block:  	vertical_forms/a/}');
Expect(1, 65055, '\p{Block=verticalforms}', "");
Expect(0, 65055, '\p{^Block=verticalforms}', "");
Expect(0, 65055, '\P{Block=verticalforms}', "");
Expect(1, 65055, '\P{^Block=verticalforms}', "");
Expect(0, 65056, '\p{Block=verticalforms}', "");
Expect(1, 65056, '\p{^Block=verticalforms}', "");
Expect(1, 65056, '\P{Block=verticalforms}', "");
Expect(0, 65056, '\P{^Block=verticalforms}', "");
Expect(1, 65055, '\p{Block=_ VERTICAL_Forms}', "");
Expect(0, 65055, '\p{^Block=_ VERTICAL_Forms}', "");
Expect(0, 65055, '\P{Block=_ VERTICAL_Forms}', "");
Expect(1, 65055, '\P{^Block=_ VERTICAL_Forms}', "");
Expect(0, 65056, '\p{Block=_ VERTICAL_Forms}', "");
Expect(1, 65056, '\p{^Block=_ VERTICAL_Forms}', "");
Expect(1, 65056, '\P{Block=_ VERTICAL_Forms}', "");
Expect(0, 65056, '\P{^Block=_ VERTICAL_Forms}', "");
Error('\p{Blk= :=vertical_Forms}');
Error('\P{Blk= :=vertical_Forms}');
Expect(1, 65055, '\p{Blk=verticalforms}', "");
Expect(0, 65055, '\p{^Blk=verticalforms}', "");
Expect(0, 65055, '\P{Blk=verticalforms}', "");
Expect(1, 65055, '\P{^Blk=verticalforms}', "");
Expect(0, 65056, '\p{Blk=verticalforms}', "");
Expect(1, 65056, '\p{^Blk=verticalforms}', "");
Expect(1, 65056, '\P{Blk=verticalforms}', "");
Expect(0, 65056, '\P{^Blk=verticalforms}', "");
Expect(1, 65055, '\p{Blk= 	Vertical_forms}', "");
Expect(0, 65055, '\p{^Blk= 	Vertical_forms}', "");
Expect(0, 65055, '\P{Blk= 	Vertical_forms}', "");
Expect(1, 65055, '\P{^Blk= 	Vertical_forms}', "");
Expect(0, 65056, '\p{Blk= 	Vertical_forms}', "");
Expect(1, 65056, '\p{^Blk= 	Vertical_forms}', "");
Expect(1, 65056, '\P{Blk= 	Vertical_forms}', "");
Expect(0, 65056, '\P{^Blk= 	Vertical_forms}', "");
Error('\p{Is_Block=:=vertical_Forms}');
Error('\P{Is_Block=:=vertical_Forms}');
Expect(1, 65055, '\p{Is_Block=verticalforms}', "");
Expect(0, 65055, '\p{^Is_Block=verticalforms}', "");
Expect(0, 65055, '\P{Is_Block=verticalforms}', "");
Expect(1, 65055, '\P{^Is_Block=verticalforms}', "");
Expect(0, 65056, '\p{Is_Block=verticalforms}', "");
Expect(1, 65056, '\p{^Is_Block=verticalforms}', "");
Expect(1, 65056, '\P{Is_Block=verticalforms}', "");
Expect(0, 65056, '\P{^Is_Block=verticalforms}', "");
Expect(1, 65055, '\p{Is_Block=_-Vertical_forms}', "");
Expect(0, 65055, '\p{^Is_Block=_-Vertical_forms}', "");
Expect(0, 65055, '\P{Is_Block=_-Vertical_forms}', "");
Expect(1, 65055, '\P{^Is_Block=_-Vertical_forms}', "");
Expect(0, 65056, '\p{Is_Block=_-Vertical_forms}', "");
Expect(1, 65056, '\p{^Is_Block=_-Vertical_forms}', "");
Expect(1, 65056, '\P{Is_Block=_-Vertical_forms}', "");
Expect(0, 65056, '\P{^Is_Block=_-Vertical_forms}', "");
Error('\p{Is_Blk=:=_vertical_Forms}');
Error('\P{Is_Blk=:=_vertical_Forms}');
Expect(1, 65055, '\p{Is_Blk=verticalforms}', "");
Expect(0, 65055, '\p{^Is_Blk=verticalforms}', "");
Expect(0, 65055, '\P{Is_Blk=verticalforms}', "");
Expect(1, 65055, '\P{^Is_Blk=verticalforms}', "");
Expect(0, 65056, '\p{Is_Blk=verticalforms}', "");
Expect(1, 65056, '\p{^Is_Blk=verticalforms}', "");
Expect(1, 65056, '\P{Is_Blk=verticalforms}', "");
Expect(0, 65056, '\P{^Is_Blk=verticalforms}', "");
Expect(1, 65055, '\p{Is_Blk=_VERTICAL_Forms}', "");
Expect(0, 65055, '\p{^Is_Blk=_VERTICAL_Forms}', "");
Expect(0, 65055, '\P{Is_Blk=_VERTICAL_Forms}', "");
Expect(1, 65055, '\P{^Is_Blk=_VERTICAL_Forms}', "");
Expect(0, 65056, '\p{Is_Blk=_VERTICAL_Forms}', "");
Expect(1, 65056, '\p{^Is_Blk=_VERTICAL_Forms}', "");
Expect(1, 65056, '\P{Is_Blk=_VERTICAL_Forms}', "");
Expect(0, 65056, '\P{^Is_Blk=_VERTICAL_Forms}', "");
Error('\p{Block=:=yi_RADICALS}');
Error('\P{Block=:=yi_RADICALS}');
Expect(1, 42191, '\p{Block:   yiradicals}', "");
Expect(0, 42191, '\p{^Block:   yiradicals}', "");
Expect(0, 42191, '\P{Block:   yiradicals}', "");
Expect(1, 42191, '\P{^Block:   yiradicals}', "");
Expect(0, 42192, '\p{Block:   yiradicals}', "");
Expect(1, 42192, '\p{^Block:   yiradicals}', "");
Expect(1, 42192, '\P{Block:   yiradicals}', "");
Expect(0, 42192, '\P{^Block:   yiradicals}', "");
Expect(1, 42191, '\p{Block=--yi_radicals}', "");
Expect(0, 42191, '\p{^Block=--yi_radicals}', "");
Expect(0, 42191, '\P{Block=--yi_radicals}', "");
Expect(1, 42191, '\P{^Block=--yi_radicals}', "");
Expect(0, 42192, '\p{Block=--yi_radicals}', "");
Expect(1, 42192, '\p{^Block=--yi_radicals}', "");
Expect(1, 42192, '\P{Block=--yi_radicals}', "");
Expect(0, 42192, '\P{^Block=--yi_radicals}', "");
Error('\p{Blk=	/a/yi_Radicals}');
Error('\P{Blk=	/a/yi_Radicals}');
Expect(1, 42191, '\p{Blk=yiradicals}', "");
Expect(0, 42191, '\p{^Blk=yiradicals}', "");
Expect(0, 42191, '\P{Blk=yiradicals}', "");
Expect(1, 42191, '\P{^Blk=yiradicals}', "");
Expect(0, 42192, '\p{Blk=yiradicals}', "");
Expect(1, 42192, '\p{^Blk=yiradicals}', "");
Expect(1, 42192, '\P{Blk=yiradicals}', "");
Expect(0, 42192, '\P{^Blk=yiradicals}', "");
Expect(1, 42191, '\p{Blk= -Yi_radicals}', "");
Expect(0, 42191, '\p{^Blk= -Yi_radicals}', "");
Expect(0, 42191, '\P{Blk= -Yi_radicals}', "");
Expect(1, 42191, '\P{^Blk= -Yi_radicals}', "");
Expect(0, 42192, '\p{Blk= -Yi_radicals}', "");
Expect(1, 42192, '\p{^Blk= -Yi_radicals}', "");
Expect(1, 42192, '\P{Blk= -Yi_radicals}', "");
Expect(0, 42192, '\P{^Blk= -Yi_radicals}', "");
Error('\p{Is_Block= /a/Yi_Radicals}');
Error('\P{Is_Block= /a/Yi_Radicals}');
Expect(1, 42191, '\p{Is_Block:yiradicals}', "");
Expect(0, 42191, '\p{^Is_Block:yiradicals}', "");
Expect(0, 42191, '\P{Is_Block:yiradicals}', "");
Expect(1, 42191, '\P{^Is_Block:yiradicals}', "");
Expect(0, 42192, '\p{Is_Block:yiradicals}', "");
Expect(1, 42192, '\p{^Is_Block:yiradicals}', "");
Expect(1, 42192, '\P{Is_Block:yiradicals}', "");
Expect(0, 42192, '\P{^Is_Block:yiradicals}', "");
Expect(1, 42191, '\p{Is_Block=	YI_Radicals}', "");
Expect(0, 42191, '\p{^Is_Block=	YI_Radicals}', "");
Expect(0, 42191, '\P{Is_Block=	YI_Radicals}', "");
Expect(1, 42191, '\P{^Is_Block=	YI_Radicals}', "");
Expect(0, 42192, '\p{Is_Block=	YI_Radicals}', "");
Expect(1, 42192, '\p{^Is_Block=	YI_Radicals}', "");
Expect(1, 42192, '\P{Is_Block=	YI_Radicals}', "");
Expect(0, 42192, '\P{^Is_Block=	YI_Radicals}', "");
Error('\p{Is_Blk=:=Yi_radicals}');
Error('\P{Is_Blk=:=Yi_radicals}');
Expect(1, 42191, '\p{Is_Blk=yiradicals}', "");
Expect(0, 42191, '\p{^Is_Blk=yiradicals}', "");
Expect(0, 42191, '\P{Is_Blk=yiradicals}', "");
Expect(1, 42191, '\P{^Is_Blk=yiradicals}', "");
Expect(0, 42192, '\p{Is_Blk=yiradicals}', "");
Expect(1, 42192, '\p{^Is_Blk=yiradicals}', "");
Expect(1, 42192, '\P{Is_Blk=yiradicals}', "");
Expect(0, 42192, '\P{^Is_Blk=yiradicals}', "");
Expect(1, 42191, '\p{Is_Blk=__yi_Radicals}', "");
Expect(0, 42191, '\p{^Is_Blk=__yi_Radicals}', "");
Expect(0, 42191, '\P{Is_Blk=__yi_Radicals}', "");
Expect(1, 42191, '\P{^Is_Blk=__yi_Radicals}', "");
Expect(0, 42192, '\p{Is_Blk=__yi_Radicals}', "");
Expect(1, 42192, '\p{^Is_Blk=__yi_Radicals}', "");
Expect(1, 42192, '\P{Is_Blk=__yi_Radicals}', "");
Expect(0, 42192, '\P{^Is_Blk=__yi_Radicals}', "");
Error('\p{Block= yi_Syllables/a/}');
Error('\P{Block= yi_Syllables/a/}');
Expect(1, 42127, '\p{Block=yisyllables}', "");
Expect(0, 42127, '\p{^Block=yisyllables}', "");
Expect(0, 42127, '\P{Block=yisyllables}', "");
Expect(1, 42127, '\P{^Block=yisyllables}', "");
Expect(0, 42128, '\p{Block=yisyllables}', "");
Expect(1, 42128, '\p{^Block=yisyllables}', "");
Expect(1, 42128, '\P{Block=yisyllables}', "");
Expect(0, 42128, '\P{^Block=yisyllables}', "");
Expect(1, 42127, '\p{Block:	-YI_SYLLABLES}', "");
Expect(0, 42127, '\p{^Block:	-YI_SYLLABLES}', "");
Expect(0, 42127, '\P{Block:	-YI_SYLLABLES}', "");
Expect(1, 42127, '\P{^Block:	-YI_SYLLABLES}', "");
Expect(0, 42128, '\p{Block:	-YI_SYLLABLES}', "");
Expect(1, 42128, '\p{^Block:	-YI_SYLLABLES}', "");
Expect(1, 42128, '\P{Block:	-YI_SYLLABLES}', "");
Expect(0, 42128, '\P{^Block:	-YI_SYLLABLES}', "");
Error('\p{Blk=:=_	YI_Syllables}');
Error('\P{Blk=:=_	YI_Syllables}');
Expect(1, 42127, '\p{Blk:	yisyllables}', "");
Expect(0, 42127, '\p{^Blk:	yisyllables}', "");
Expect(0, 42127, '\P{Blk:	yisyllables}', "");
Expect(1, 42127, '\P{^Blk:	yisyllables}', "");
Expect(0, 42128, '\p{Blk:	yisyllables}', "");
Expect(1, 42128, '\p{^Blk:	yisyllables}', "");
Expect(1, 42128, '\P{Blk:	yisyllables}', "");
Expect(0, 42128, '\P{^Blk:	yisyllables}', "");
Expect(1, 42127, '\p{Blk=-Yi_SYLLABLES}', "");
Expect(0, 42127, '\p{^Blk=-Yi_SYLLABLES}', "");
Expect(0, 42127, '\P{Blk=-Yi_SYLLABLES}', "");
Expect(1, 42127, '\P{^Blk=-Yi_SYLLABLES}', "");
Expect(0, 42128, '\p{Blk=-Yi_SYLLABLES}', "");
Expect(1, 42128, '\p{^Blk=-Yi_SYLLABLES}', "");
Expect(1, 42128, '\P{Blk=-Yi_SYLLABLES}', "");
Expect(0, 42128, '\P{^Blk=-Yi_SYLLABLES}', "");
Error('\p{Is_Block=:=--Yi_SYLLABLES}');
Error('\P{Is_Block=:=--Yi_SYLLABLES}');
Expect(1, 42127, '\p{Is_Block=yisyllables}', "");
Expect(0, 42127, '\p{^Is_Block=yisyllables}', "");
Expect(0, 42127, '\P{Is_Block=yisyllables}', "");
Expect(1, 42127, '\P{^Is_Block=yisyllables}', "");
Expect(0, 42128, '\p{Is_Block=yisyllables}', "");
Expect(1, 42128, '\p{^Is_Block=yisyllables}', "");
Expect(1, 42128, '\P{Is_Block=yisyllables}', "");
Expect(0, 42128, '\P{^Is_Block=yisyllables}', "");
Expect(1, 42127, '\p{Is_Block=__yi_Syllables}', "");
Expect(0, 42127, '\p{^Is_Block=__yi_Syllables}', "");
Expect(0, 42127, '\P{Is_Block=__yi_Syllables}', "");
Expect(1, 42127, '\P{^Is_Block=__yi_Syllables}', "");
Expect(0, 42128, '\p{Is_Block=__yi_Syllables}', "");
Expect(1, 42128, '\p{^Is_Block=__yi_Syllables}', "");
Expect(1, 42128, '\P{Is_Block=__yi_Syllables}', "");
Expect(0, 42128, '\P{^Is_Block=__yi_Syllables}', "");
Error('\p{Is_Blk: :=yi_SYLLABLES}');
Error('\P{Is_Blk: :=yi_SYLLABLES}');
Expect(1, 42127, '\p{Is_Blk=yisyllables}', "");
Expect(0, 42127, '\p{^Is_Blk=yisyllables}', "");
Expect(0, 42127, '\P{Is_Blk=yisyllables}', "");
Expect(1, 42127, '\P{^Is_Blk=yisyllables}', "");
Expect(0, 42128, '\p{Is_Blk=yisyllables}', "");
Expect(1, 42128, '\p{^Is_Blk=yisyllables}', "");
Expect(1, 42128, '\P{Is_Blk=yisyllables}', "");
Expect(0, 42128, '\P{^Is_Blk=yisyllables}', "");
Expect(1, 42127, '\p{Is_Blk=_yi_Syllables}', "");
Expect(0, 42127, '\p{^Is_Blk=_yi_Syllables}', "");
Expect(0, 42127, '\P{Is_Blk=_yi_Syllables}', "");
Expect(1, 42127, '\P{^Is_Blk=_yi_Syllables}', "");
Expect(0, 42128, '\p{Is_Blk=_yi_Syllables}', "");
Expect(1, 42128, '\p{^Is_Blk=_yi_Syllables}', "");
Expect(1, 42128, '\P{Is_Blk=_yi_Syllables}', "");
Expect(0, 42128, '\P{^Is_Blk=_yi_Syllables}', "");
Error('\p{Block=-	Yijing_HEXAGRAM_symbols/a/}');
Error('\P{Block=-	Yijing_HEXAGRAM_symbols/a/}');
Expect(1, 19967, '\p{Block:yijinghexagramsymbols}', "");
Expect(0, 19967, '\p{^Block:yijinghexagramsymbols}', "");
Expect(0, 19967, '\P{Block:yijinghexagramsymbols}', "");
Expect(1, 19967, '\P{^Block:yijinghexagramsymbols}', "");
Expect(0, 19968, '\p{Block:yijinghexagramsymbols}', "");
Expect(1, 19968, '\p{^Block:yijinghexagramsymbols}', "");
Expect(1, 19968, '\P{Block:yijinghexagramsymbols}', "");
Expect(0, 19968, '\P{^Block:yijinghexagramsymbols}', "");
Expect(1, 19967, '\p{Block=_Yijing_Hexagram_Symbols}', "");
Expect(0, 19967, '\p{^Block=_Yijing_Hexagram_Symbols}', "");
Expect(0, 19967, '\P{Block=_Yijing_Hexagram_Symbols}', "");
Expect(1, 19967, '\P{^Block=_Yijing_Hexagram_Symbols}', "");
Expect(0, 19968, '\p{Block=_Yijing_Hexagram_Symbols}', "");
Expect(1, 19968, '\p{^Block=_Yijing_Hexagram_Symbols}', "");
Expect(1, 19968, '\P{Block=_Yijing_Hexagram_Symbols}', "");
Expect(0, 19968, '\P{^Block=_Yijing_Hexagram_Symbols}', "");
Error('\p{Blk=:=	_Yijing}');
Error('\P{Blk=:=	_Yijing}');
Expect(1, 19967, '\p{Blk=yijing}', "");
Expect(0, 19967, '\p{^Blk=yijing}', "");
Expect(0, 19967, '\P{Blk=yijing}', "");
Expect(1, 19967, '\P{^Blk=yijing}', "");
Expect(0, 19968, '\p{Blk=yijing}', "");
Expect(1, 19968, '\p{^Blk=yijing}', "");
Expect(1, 19968, '\P{Blk=yijing}', "");
Expect(0, 19968, '\P{^Blk=yijing}', "");
Expect(1, 19967, '\p{Blk=--YIJING}', "");
Expect(0, 19967, '\p{^Blk=--YIJING}', "");
Expect(0, 19967, '\P{Blk=--YIJING}', "");
Expect(1, 19967, '\P{^Blk=--YIJING}', "");
Expect(0, 19968, '\p{Blk=--YIJING}', "");
Expect(1, 19968, '\p{^Blk=--YIJING}', "");
Expect(1, 19968, '\P{Blk=--YIJING}', "");
Expect(0, 19968, '\P{^Blk=--YIJING}', "");
Error('\p{Is_Block= :=yijing_Hexagram_Symbols}');
Error('\P{Is_Block= :=yijing_Hexagram_Symbols}');
Expect(1, 19967, '\p{Is_Block=yijinghexagramsymbols}', "");
Expect(0, 19967, '\p{^Is_Block=yijinghexagramsymbols}', "");
Expect(0, 19967, '\P{Is_Block=yijinghexagramsymbols}', "");
Expect(1, 19967, '\P{^Is_Block=yijinghexagramsymbols}', "");
Expect(0, 19968, '\p{Is_Block=yijinghexagramsymbols}', "");
Expect(1, 19968, '\p{^Is_Block=yijinghexagramsymbols}', "");
Expect(1, 19968, '\P{Is_Block=yijinghexagramsymbols}', "");
Expect(0, 19968, '\P{^Is_Block=yijinghexagramsymbols}', "");
Expect(1, 19967, '\p{Is_Block=--yijing_HEXAGRAM_Symbols}', "");
Expect(0, 19967, '\p{^Is_Block=--yijing_HEXAGRAM_Symbols}', "");
Expect(0, 19967, '\P{Is_Block=--yijing_HEXAGRAM_Symbols}', "");
Expect(1, 19967, '\P{^Is_Block=--yijing_HEXAGRAM_Symbols}', "");
Expect(0, 19968, '\p{Is_Block=--yijing_HEXAGRAM_Symbols}', "");
Expect(1, 19968, '\p{^Is_Block=--yijing_HEXAGRAM_Symbols}', "");
Expect(1, 19968, '\P{Is_Block=--yijing_HEXAGRAM_Symbols}', "");
Expect(0, 19968, '\P{^Is_Block=--yijing_HEXAGRAM_Symbols}', "");
Error('\p{Is_Blk=__Yijing/a/}');
Error('\P{Is_Blk=__Yijing/a/}');
Expect(1, 19967, '\p{Is_Blk=yijing}', "");
Expect(0, 19967, '\p{^Is_Blk=yijing}', "");
Expect(0, 19967, '\P{Is_Blk=yijing}', "");
Expect(1, 19967, '\P{^Is_Blk=yijing}', "");
Expect(0, 19968, '\p{Is_Blk=yijing}', "");
Expect(1, 19968, '\p{^Is_Blk=yijing}', "");
Expect(1, 19968, '\P{Is_Blk=yijing}', "");
Expect(0, 19968, '\P{^Is_Blk=yijing}', "");
Expect(1, 19967, '\p{Is_Blk=	YIJING}', "");
Expect(0, 19967, '\p{^Is_Blk=	YIJING}', "");
Expect(0, 19967, '\P{Is_Blk=	YIJING}', "");
Expect(1, 19967, '\P{^Is_Blk=	YIJING}', "");
Expect(0, 19968, '\p{Is_Blk=	YIJING}', "");
Expect(1, 19968, '\p{^Is_Blk=	YIJING}', "");
Expect(1, 19968, '\P{Is_Blk=	YIJING}', "");
Expect(0, 19968, '\P{^Is_Blk=	YIJING}', "");
Error('\p{Bidi_Paired_Bracket_Type=	/a/Close}');
Error('\P{Bidi_Paired_Bracket_Type=	/a/Close}');
Expect(1, 65379, '\p{Bidi_Paired_Bracket_Type:   close}', "");
Expect(0, 65379, '\p{^Bidi_Paired_Bracket_Type:   close}', "");
Expect(0, 65379, '\P{Bidi_Paired_Bracket_Type:   close}', "");
Expect(1, 65379, '\P{^Bidi_Paired_Bracket_Type:   close}', "");
Expect(0, 65380, '\p{Bidi_Paired_Bracket_Type:   close}', "");
Expect(1, 65380, '\p{^Bidi_Paired_Bracket_Type:   close}', "");
Expect(1, 65380, '\P{Bidi_Paired_Bracket_Type:   close}', "");
Expect(0, 65380, '\P{^Bidi_Paired_Bracket_Type:   close}', "");
Expect(1, 65379, '\p{Bidi_Paired_Bracket_Type=-Close}', "");
Expect(0, 65379, '\p{^Bidi_Paired_Bracket_Type=-Close}', "");
Expect(0, 65379, '\P{Bidi_Paired_Bracket_Type=-Close}', "");
Expect(1, 65379, '\P{^Bidi_Paired_Bracket_Type=-Close}', "");
Expect(0, 65380, '\p{Bidi_Paired_Bracket_Type=-Close}', "");
Expect(1, 65380, '\p{^Bidi_Paired_Bracket_Type=-Close}', "");
Expect(1, 65380, '\P{Bidi_Paired_Bracket_Type=-Close}', "");
Expect(0, 65380, '\P{^Bidi_Paired_Bracket_Type=-Close}', "");
Error('\p{Bpt=__C:=}');
Error('\P{Bpt=__C:=}');
Expect(1, 65379, '\p{Bpt=c}', "");
Expect(0, 65379, '\p{^Bpt=c}', "");
Expect(0, 65379, '\P{Bpt=c}', "");
Expect(1, 65379, '\P{^Bpt=c}', "");
Expect(0, 65380, '\p{Bpt=c}', "");
Expect(1, 65380, '\p{^Bpt=c}', "");
Expect(1, 65380, '\P{Bpt=c}', "");
Expect(0, 65380, '\P{^Bpt=c}', "");
Expect(1, 65379, '\p{Bpt=-_C}', "");
Expect(0, 65379, '\p{^Bpt=-_C}', "");
Expect(0, 65379, '\P{Bpt=-_C}', "");
Expect(1, 65379, '\P{^Bpt=-_C}', "");
Expect(0, 65380, '\p{Bpt=-_C}', "");
Expect(1, 65380, '\p{^Bpt=-_C}', "");
Expect(1, 65380, '\P{Bpt=-_C}', "");
Expect(0, 65380, '\P{^Bpt=-_C}', "");
Error('\p{Is_Bidi_Paired_Bracket_Type=:=Close}');
Error('\P{Is_Bidi_Paired_Bracket_Type=:=Close}');
Expect(1, 65379, '\p{Is_Bidi_Paired_Bracket_Type=close}', "");
Expect(0, 65379, '\p{^Is_Bidi_Paired_Bracket_Type=close}', "");
Expect(0, 65379, '\P{Is_Bidi_Paired_Bracket_Type=close}', "");
Expect(1, 65379, '\P{^Is_Bidi_Paired_Bracket_Type=close}', "");
Expect(0, 65380, '\p{Is_Bidi_Paired_Bracket_Type=close}', "");
Expect(1, 65380, '\p{^Is_Bidi_Paired_Bracket_Type=close}', "");
Expect(1, 65380, '\P{Is_Bidi_Paired_Bracket_Type=close}', "");
Expect(0, 65380, '\P{^Is_Bidi_Paired_Bracket_Type=close}', "");
Expect(1, 65379, '\p{Is_Bidi_Paired_Bracket_Type=Close}', "");
Expect(0, 65379, '\p{^Is_Bidi_Paired_Bracket_Type=Close}', "");
Expect(0, 65379, '\P{Is_Bidi_Paired_Bracket_Type=Close}', "");
Expect(1, 65379, '\P{^Is_Bidi_Paired_Bracket_Type=Close}', "");
Expect(0, 65380, '\p{Is_Bidi_Paired_Bracket_Type=Close}', "");
Expect(1, 65380, '\p{^Is_Bidi_Paired_Bracket_Type=Close}', "");
Expect(1, 65380, '\P{Is_Bidi_Paired_Bracket_Type=Close}', "");
Expect(0, 65380, '\P{^Is_Bidi_Paired_Bracket_Type=Close}', "");
Error('\p{Is_Bpt=/a/  C}');
Error('\P{Is_Bpt=/a/  C}');
Expect(1, 65379, '\p{Is_Bpt=c}', "");
Expect(0, 65379, '\p{^Is_Bpt=c}', "");
Expect(0, 65379, '\P{Is_Bpt=c}', "");
Expect(1, 65379, '\P{^Is_Bpt=c}', "");
Expect(0, 65380, '\p{Is_Bpt=c}', "");
Expect(1, 65380, '\p{^Is_Bpt=c}', "");
Expect(1, 65380, '\P{Is_Bpt=c}', "");
Expect(0, 65380, '\P{^Is_Bpt=c}', "");
Expect(1, 65379, '\p{Is_Bpt=		c}', "");
Expect(0, 65379, '\p{^Is_Bpt=		c}', "");
Expect(0, 65379, '\P{Is_Bpt=		c}', "");
Expect(1, 65379, '\P{^Is_Bpt=		c}', "");
Expect(0, 65380, '\p{Is_Bpt=		c}', "");
Expect(1, 65380, '\p{^Is_Bpt=		c}', "");
Expect(1, 65380, '\P{Is_Bpt=		c}', "");
Expect(0, 65380, '\P{^Is_Bpt=		c}', "");
Error('\p{Bidi_Paired_Bracket_Type=_/a/None}');
Error('\P{Bidi_Paired_Bracket_Type=_/a/None}');
Expect(1, 65380, '\p{Bidi_Paired_Bracket_Type=none}', "");
Expect(0, 65380, '\p{^Bidi_Paired_Bracket_Type=none}', "");
Expect(0, 65380, '\P{Bidi_Paired_Bracket_Type=none}', "");
Expect(1, 65380, '\P{^Bidi_Paired_Bracket_Type=none}', "");
Expect(0, 65379, '\p{Bidi_Paired_Bracket_Type=none}', "");
Expect(1, 65379, '\p{^Bidi_Paired_Bracket_Type=none}', "");
Expect(1, 65379, '\P{Bidi_Paired_Bracket_Type=none}', "");
Expect(0, 65379, '\P{^Bidi_Paired_Bracket_Type=none}', "");
Expect(1, 65380, '\p{Bidi_Paired_Bracket_Type=-	NONE}', "");
Expect(0, 65380, '\p{^Bidi_Paired_Bracket_Type=-	NONE}', "");
Expect(0, 65380, '\P{Bidi_Paired_Bracket_Type=-	NONE}', "");
Expect(1, 65380, '\P{^Bidi_Paired_Bracket_Type=-	NONE}', "");
Expect(0, 65379, '\p{Bidi_Paired_Bracket_Type=-	NONE}', "");
Expect(1, 65379, '\p{^Bidi_Paired_Bracket_Type=-	NONE}', "");
Expect(1, 65379, '\P{Bidi_Paired_Bracket_Type=-	NONE}', "");
Expect(0, 65379, '\P{^Bidi_Paired_Bracket_Type=-	NONE}', "");
Error('\p{Bpt=:=_-n}');
Error('\P{Bpt=:=_-n}');
Expect(1, 65380, '\p{Bpt=n}', "");
Expect(0, 65380, '\p{^Bpt=n}', "");
Expect(0, 65380, '\P{Bpt=n}', "");
Expect(1, 65380, '\P{^Bpt=n}', "");
Expect(0, 65379, '\p{Bpt=n}', "");
Expect(1, 65379, '\p{^Bpt=n}', "");
Expect(1, 65379, '\P{Bpt=n}', "");
Expect(0, 65379, '\P{^Bpt=n}', "");
Expect(1, 65380, '\p{Bpt=_-N}', "");
Expect(0, 65380, '\p{^Bpt=_-N}', "");
Expect(0, 65380, '\P{Bpt=_-N}', "");
Expect(1, 65380, '\P{^Bpt=_-N}', "");
Expect(0, 65379, '\p{Bpt=_-N}', "");
Expect(1, 65379, '\p{^Bpt=_-N}', "");
Expect(1, 65379, '\P{Bpt=_-N}', "");
Expect(0, 65379, '\P{^Bpt=_-N}', "");
Error('\p{Is_Bidi_Paired_Bracket_Type=:=NONE}');
Error('\P{Is_Bidi_Paired_Bracket_Type=:=NONE}');
Expect(1, 65380, '\p{Is_Bidi_Paired_Bracket_Type=none}', "");
Expect(0, 65380, '\p{^Is_Bidi_Paired_Bracket_Type=none}', "");
Expect(0, 65380, '\P{Is_Bidi_Paired_Bracket_Type=none}', "");
Expect(1, 65380, '\P{^Is_Bidi_Paired_Bracket_Type=none}', "");
Expect(0, 65379, '\p{Is_Bidi_Paired_Bracket_Type=none}', "");
Expect(1, 65379, '\p{^Is_Bidi_Paired_Bracket_Type=none}', "");
Expect(1, 65379, '\P{Is_Bidi_Paired_Bracket_Type=none}', "");
Expect(0, 65379, '\P{^Is_Bidi_Paired_Bracket_Type=none}', "");
Expect(1, 65380, '\p{Is_Bidi_Paired_Bracket_Type=_NONE}', "");
Expect(0, 65380, '\p{^Is_Bidi_Paired_Bracket_Type=_NONE}', "");
Expect(0, 65380, '\P{Is_Bidi_Paired_Bracket_Type=_NONE}', "");
Expect(1, 65380, '\P{^Is_Bidi_Paired_Bracket_Type=_NONE}', "");
Expect(0, 65379, '\p{Is_Bidi_Paired_Bracket_Type=_NONE}', "");
Expect(1, 65379, '\p{^Is_Bidi_Paired_Bracket_Type=_NONE}', "");
Expect(1, 65379, '\P{Is_Bidi_Paired_Bracket_Type=_NONE}', "");
Expect(0, 65379, '\P{^Is_Bidi_Paired_Bracket_Type=_NONE}', "");
Error('\p{Is_Bpt=-:=n}');
Error('\P{Is_Bpt=-:=n}');
Expect(1, 65380, '\p{Is_Bpt=n}', "");
Expect(0, 65380, '\p{^Is_Bpt=n}', "");
Expect(0, 65380, '\P{Is_Bpt=n}', "");
Expect(1, 65380, '\P{^Is_Bpt=n}', "");
Expect(0, 65379, '\p{Is_Bpt=n}', "");
Expect(1, 65379, '\p{^Is_Bpt=n}', "");
Expect(1, 65379, '\P{Is_Bpt=n}', "");
Expect(0, 65379, '\P{^Is_Bpt=n}', "");
Expect(1, 65380, '\p{Is_Bpt=_-N}', "");
Expect(0, 65380, '\p{^Is_Bpt=_-N}', "");
Expect(0, 65380, '\P{Is_Bpt=_-N}', "");
Expect(1, 65380, '\P{^Is_Bpt=_-N}', "");
Expect(0, 65379, '\p{Is_Bpt=_-N}', "");
Expect(1, 65379, '\p{^Is_Bpt=_-N}', "");
Expect(1, 65379, '\P{Is_Bpt=_-N}', "");
Expect(0, 65379, '\P{^Is_Bpt=_-N}', "");
Error('\p{Bidi_Paired_Bracket_Type=/a/-open}');
Error('\P{Bidi_Paired_Bracket_Type=/a/-open}');
Expect(1, 65378, '\p{Bidi_Paired_Bracket_Type: open}', "");
Expect(0, 65378, '\p{^Bidi_Paired_Bracket_Type: open}', "");
Expect(0, 65378, '\P{Bidi_Paired_Bracket_Type: open}', "");
Expect(1, 65378, '\P{^Bidi_Paired_Bracket_Type: open}', "");
Expect(0, 65379, '\p{Bidi_Paired_Bracket_Type: open}', "");
Expect(1, 65379, '\p{^Bidi_Paired_Bracket_Type: open}', "");
Expect(1, 65379, '\P{Bidi_Paired_Bracket_Type: open}', "");
Expect(0, 65379, '\P{^Bidi_Paired_Bracket_Type: open}', "");
Expect(1, 65378, '\p{Bidi_Paired_Bracket_Type= Open}', "");
Expect(0, 65378, '\p{^Bidi_Paired_Bracket_Type= Open}', "");
Expect(0, 65378, '\P{Bidi_Paired_Bracket_Type= Open}', "");
Expect(1, 65378, '\P{^Bidi_Paired_Bracket_Type= Open}', "");
Expect(0, 65379, '\p{Bidi_Paired_Bracket_Type= Open}', "");
Expect(1, 65379, '\p{^Bidi_Paired_Bracket_Type= Open}', "");
Expect(1, 65379, '\P{Bidi_Paired_Bracket_Type= Open}', "");
Expect(0, 65379, '\P{^Bidi_Paired_Bracket_Type= Open}', "");
Error('\p{Bpt=/a/_	O}');
Error('\P{Bpt=/a/_	O}');
Expect(1, 65378, '\p{Bpt=o}', "");
Expect(0, 65378, '\p{^Bpt=o}', "");
Expect(0, 65378, '\P{Bpt=o}', "");
Expect(1, 65378, '\P{^Bpt=o}', "");
Expect(0, 65379, '\p{Bpt=o}', "");
Expect(1, 65379, '\p{^Bpt=o}', "");
Expect(1, 65379, '\P{Bpt=o}', "");
Expect(0, 65379, '\P{^Bpt=o}', "");
Expect(1, 65378, '\p{Bpt=--O}', "");
Expect(0, 65378, '\p{^Bpt=--O}', "");
Expect(0, 65378, '\P{Bpt=--O}', "");
Expect(1, 65378, '\P{^Bpt=--O}', "");
Expect(0, 65379, '\p{Bpt=--O}', "");
Expect(1, 65379, '\p{^Bpt=--O}', "");
Expect(1, 65379, '\P{Bpt=--O}', "");
Expect(0, 65379, '\P{^Bpt=--O}', "");
Error('\p{Is_Bidi_Paired_Bracket_Type=/a/	-OPEN}');
Error('\P{Is_Bidi_Paired_Bracket_Type=/a/	-OPEN}');
Expect(1, 65378, '\p{Is_Bidi_Paired_Bracket_Type=open}', "");
Expect(0, 65378, '\p{^Is_Bidi_Paired_Bracket_Type=open}', "");
Expect(0, 65378, '\P{Is_Bidi_Paired_Bracket_Type=open}', "");
Expect(1, 65378, '\P{^Is_Bidi_Paired_Bracket_Type=open}', "");
Expect(0, 65379, '\p{Is_Bidi_Paired_Bracket_Type=open}', "");
Expect(1, 65379, '\p{^Is_Bidi_Paired_Bracket_Type=open}', "");
Expect(1, 65379, '\P{Is_Bidi_Paired_Bracket_Type=open}', "");
Expect(0, 65379, '\P{^Is_Bidi_Paired_Bracket_Type=open}', "");
Expect(1, 65378, '\p{Is_Bidi_Paired_Bracket_Type=	Open}', "");
Expect(0, 65378, '\p{^Is_Bidi_Paired_Bracket_Type=	Open}', "");
Expect(0, 65378, '\P{Is_Bidi_Paired_Bracket_Type=	Open}', "");
Expect(1, 65378, '\P{^Is_Bidi_Paired_Bracket_Type=	Open}', "");
Expect(0, 65379, '\p{Is_Bidi_Paired_Bracket_Type=	Open}', "");
Expect(1, 65379, '\p{^Is_Bidi_Paired_Bracket_Type=	Open}', "");
Expect(1, 65379, '\P{Is_Bidi_Paired_Bracket_Type=	Open}', "");
Expect(0, 65379, '\P{^Is_Bidi_Paired_Bracket_Type=	Open}', "");
Error('\p{Is_Bpt=:= _o}');
Error('\P{Is_Bpt=:= _o}');
Expect(1, 65378, '\p{Is_Bpt:   o}', "");
Expect(0, 65378, '\p{^Is_Bpt:   o}', "");
Expect(0, 65378, '\P{Is_Bpt:   o}', "");
Expect(1, 65378, '\P{^Is_Bpt:   o}', "");
Expect(0, 65379, '\p{Is_Bpt:   o}', "");
Expect(1, 65379, '\p{^Is_Bpt:   o}', "");
Expect(1, 65379, '\P{Is_Bpt:   o}', "");
Expect(0, 65379, '\P{^Is_Bpt:   o}', "");
Expect(1, 65378, '\p{Is_Bpt=-_O}', "");
Expect(0, 65378, '\p{^Is_Bpt=-_O}', "");
Expect(0, 65378, '\P{Is_Bpt=-_O}', "");
Expect(1, 65378, '\P{^Is_Bpt=-_O}', "");
Expect(0, 65379, '\p{Is_Bpt=-_O}', "");
Expect(1, 65379, '\p{^Is_Bpt=-_O}', "");
Expect(1, 65379, '\P{Is_Bpt=-_O}', "");
Expect(0, 65379, '\P{^Is_Bpt=-_O}', "");
Error('\p{Canonical_Combining_Class: -Above:=}');
Error('\P{Canonical_Combining_Class: -Above:=}');
Expect(1, 119364, '\p{Canonical_Combining_Class=above}', "");
Expect(0, 119364, '\p{^Canonical_Combining_Class=above}', "");
Expect(0, 119364, '\P{Canonical_Combining_Class=above}', "");
Expect(1, 119364, '\P{^Canonical_Combining_Class=above}', "");
Expect(0, 119365, '\p{Canonical_Combining_Class=above}', "");
Expect(1, 119365, '\p{^Canonical_Combining_Class=above}', "");
Expect(1, 119365, '\P{Canonical_Combining_Class=above}', "");
Expect(0, 119365, '\P{^Canonical_Combining_Class=above}', "");
Expect(1, 119364, '\p{Canonical_Combining_Class=- above}', "");
Expect(0, 119364, '\p{^Canonical_Combining_Class=- above}', "");
Expect(0, 119364, '\P{Canonical_Combining_Class=- above}', "");
Expect(1, 119364, '\P{^Canonical_Combining_Class=- above}', "");
Expect(0, 119365, '\p{Canonical_Combining_Class=- above}', "");
Expect(1, 119365, '\p{^Canonical_Combining_Class=- above}', "");
Expect(1, 119365, '\P{Canonical_Combining_Class=- above}', "");
Expect(0, 119365, '\P{^Canonical_Combining_Class=- above}', "");
Error('\p{Ccc=	A:=}');
Error('\P{Ccc=	A:=}');
Expect(1, 119364, '\p{Ccc=a}', "");
Expect(0, 119364, '\p{^Ccc=a}', "");
Expect(0, 119364, '\P{Ccc=a}', "");
Expect(1, 119364, '\P{^Ccc=a}', "");
Expect(0, 119365, '\p{Ccc=a}', "");
Expect(1, 119365, '\p{^Ccc=a}', "");
Expect(1, 119365, '\P{Ccc=a}', "");
Expect(0, 119365, '\P{^Ccc=a}', "");
Expect(1, 119364, '\p{Ccc:   	A}', "");
Expect(0, 119364, '\p{^Ccc:   	A}', "");
Expect(0, 119364, '\P{Ccc:   	A}', "");
Expect(1, 119364, '\P{^Ccc:   	A}', "");
Expect(0, 119365, '\p{Ccc:   	A}', "");
Expect(1, 119365, '\p{^Ccc:   	A}', "");
Expect(1, 119365, '\P{Ccc:   	A}', "");
Expect(0, 119365, '\P{^Ccc:   	A}', "");
Error('\p{Is_Canonical_Combining_Class=	/a/00_00_00_02_30}');
Error('\P{Is_Canonical_Combining_Class=	/a/00_00_00_02_30}');
Expect(1, 119364, '\p{Is_Canonical_Combining_Class=+0_0_0_0_0_0230}', "");
Expect(0, 119364, '\p{^Is_Canonical_Combining_Class=+0_0_0_0_0_0230}', "");
Expect(0, 119364, '\P{Is_Canonical_Combining_Class=+0_0_0_0_0_0230}', "");
Expect(1, 119364, '\P{^Is_Canonical_Combining_Class=+0_0_0_0_0_0230}', "");
Expect(0, 119365, '\p{Is_Canonical_Combining_Class=+0_0_0_0_0_0230}', "");
Expect(1, 119365, '\p{^Is_Canonical_Combining_Class=+0_0_0_0_0_0230}', "");
Expect(1, 119365, '\P{Is_Canonical_Combining_Class=+0_0_0_0_0_0230}', "");
Expect(0, 119365, '\P{^Is_Canonical_Combining_Class=+0_0_0_0_0_0230}', "");
Error('\p{Is_Ccc=/a/-above}');
Error('\P{Is_Ccc=/a/-above}');
Expect(1, 119364, '\p{Is_Ccc:	above}', "");
Expect(0, 119364, '\p{^Is_Ccc:	above}', "");
Expect(0, 119364, '\P{Is_Ccc:	above}', "");
Expect(1, 119364, '\P{^Is_Ccc:	above}', "");
Expect(0, 119365, '\p{Is_Ccc:	above}', "");
Expect(1, 119365, '\p{^Is_Ccc:	above}', "");
Expect(1, 119365, '\P{Is_Ccc:	above}', "");
Expect(0, 119365, '\P{^Is_Ccc:	above}', "");
Expect(1, 119364, '\p{Is_Ccc=-Above}', "");
Expect(0, 119364, '\p{^Is_Ccc=-Above}', "");
Expect(0, 119364, '\P{Is_Ccc=-Above}', "");
Expect(1, 119364, '\P{^Is_Ccc=-Above}', "");
Expect(0, 119365, '\p{Is_Ccc=-Above}', "");
Expect(1, 119365, '\p{^Is_Ccc=-Above}', "");
Expect(1, 119365, '\P{Is_Ccc=-Above}', "");
Expect(0, 119365, '\P{^Is_Ccc=-Above}', "");
Error('\p{Canonical_Combining_Class:	/a/_ ABOVE_left}');
Error('\P{Canonical_Combining_Class:	/a/_ ABOVE_left}');
Expect(1, 12331, '\p{Canonical_Combining_Class=aboveleft}', "");
Expect(0, 12331, '\p{^Canonical_Combining_Class=aboveleft}', "");
Expect(0, 12331, '\P{Canonical_Combining_Class=aboveleft}', "");
Expect(1, 12331, '\P{^Canonical_Combining_Class=aboveleft}', "");
Expect(0, 12332, '\p{Canonical_Combining_Class=aboveleft}', "");
Expect(1, 12332, '\p{^Canonical_Combining_Class=aboveleft}', "");
Expect(1, 12332, '\P{Canonical_Combining_Class=aboveleft}', "");
Expect(0, 12332, '\P{^Canonical_Combining_Class=aboveleft}', "");
Expect(1, 12331, '\p{Canonical_Combining_Class=_ above_Left}', "");
Expect(0, 12331, '\p{^Canonical_Combining_Class=_ above_Left}', "");
Expect(0, 12331, '\P{Canonical_Combining_Class=_ above_Left}', "");
Expect(1, 12331, '\P{^Canonical_Combining_Class=_ above_Left}', "");
Expect(0, 12332, '\p{Canonical_Combining_Class=_ above_Left}', "");
Expect(1, 12332, '\p{^Canonical_Combining_Class=_ above_Left}', "");
Expect(1, 12332, '\P{Canonical_Combining_Class=_ above_Left}', "");
Expect(0, 12332, '\P{^Canonical_Combining_Class=_ above_Left}', "");
Error('\p{Ccc=AL/a/}');
Error('\P{Ccc=AL/a/}');
Expect(1, 12331, '\p{Ccc=al}', "");
Expect(0, 12331, '\p{^Ccc=al}', "");
Expect(0, 12331, '\P{Ccc=al}', "");
Expect(1, 12331, '\P{^Ccc=al}', "");
Expect(0, 12332, '\p{Ccc=al}', "");
Expect(1, 12332, '\p{^Ccc=al}', "");
Expect(1, 12332, '\P{Ccc=al}', "");
Expect(0, 12332, '\P{^Ccc=al}', "");
Expect(1, 12331, '\p{Ccc=_-AL}', "");
Expect(0, 12331, '\p{^Ccc=_-AL}', "");
Expect(0, 12331, '\P{Ccc=_-AL}', "");
Expect(1, 12331, '\P{^Ccc=_-AL}', "");
Expect(0, 12332, '\p{Ccc=_-AL}', "");
Expect(1, 12332, '\p{^Ccc=_-AL}', "");
Expect(1, 12332, '\P{Ccc=_-AL}', "");
Expect(0, 12332, '\P{^Ccc=_-AL}', "");
Error('\p{Is_Canonical_Combining_Class: :=--000000000228}');
Error('\P{Is_Canonical_Combining_Class: :=--000000000228}');
Expect(1, 12331, '\p{Is_Canonical_Combining_Class=228}', "");
Expect(0, 12331, '\p{^Is_Canonical_Combining_Class=228}', "");
Expect(0, 12331, '\P{Is_Canonical_Combining_Class=228}', "");
Expect(1, 12331, '\P{^Is_Canonical_Combining_Class=228}', "");
Expect(0, 12332, '\p{Is_Canonical_Combining_Class=228}', "");
Expect(1, 12332, '\p{^Is_Canonical_Combining_Class=228}', "");
Expect(1, 12332, '\P{Is_Canonical_Combining_Class=228}', "");
Expect(0, 12332, '\P{^Is_Canonical_Combining_Class=228}', "");
Error('\p{Is_Ccc=	_Above_LEFT/a/}');
Error('\P{Is_Ccc=	_Above_LEFT/a/}');
Expect(1, 12331, '\p{Is_Ccc=aboveleft}', "");
Expect(0, 12331, '\p{^Is_Ccc=aboveleft}', "");
Expect(0, 12331, '\P{Is_Ccc=aboveleft}', "");
Expect(1, 12331, '\P{^Is_Ccc=aboveleft}', "");
Expect(0, 12332, '\p{Is_Ccc=aboveleft}', "");
Expect(1, 12332, '\p{^Is_Ccc=aboveleft}', "");
Expect(1, 12332, '\P{Is_Ccc=aboveleft}', "");
Expect(0, 12332, '\P{^Is_Ccc=aboveleft}', "");
Expect(1, 12331, '\p{Is_Ccc= _ABOVE_LEFT}', "");
Expect(0, 12331, '\p{^Is_Ccc= _ABOVE_LEFT}', "");
Expect(0, 12331, '\P{Is_Ccc= _ABOVE_LEFT}', "");
Expect(1, 12331, '\P{^Is_Ccc= _ABOVE_LEFT}', "");
Expect(0, 12332, '\p{Is_Ccc= _ABOVE_LEFT}', "");
Expect(1, 12332, '\p{^Is_Ccc= _ABOVE_LEFT}', "");
Expect(1, 12332, '\P{Is_Ccc= _ABOVE_LEFT}', "");
Expect(0, 12332, '\P{^Is_Ccc= _ABOVE_LEFT}', "");
Error('\p{Canonical_Combining_Class=__Above_Right/a/}');
Error('\P{Canonical_Combining_Class=__Above_Right/a/}');
Expect(1, 12332, '\p{Canonical_Combining_Class=aboveright}', "");
Expect(0, 12332, '\p{^Canonical_Combining_Class=aboveright}', "");
Expect(0, 12332, '\P{Canonical_Combining_Class=aboveright}', "");
Expect(1, 12332, '\P{^Canonical_Combining_Class=aboveright}', "");
Expect(0, 12333, '\p{Canonical_Combining_Class=aboveright}', "");
Expect(1, 12333, '\p{^Canonical_Combining_Class=aboveright}', "");
Expect(1, 12333, '\P{Canonical_Combining_Class=aboveright}', "");
Expect(0, 12333, '\P{^Canonical_Combining_Class=aboveright}', "");
Expect(1, 12332, '\p{Canonical_Combining_Class=	-ABOVE_right}', "");
Expect(0, 12332, '\p{^Canonical_Combining_Class=	-ABOVE_right}', "");
Expect(0, 12332, '\P{Canonical_Combining_Class=	-ABOVE_right}', "");
Expect(1, 12332, '\P{^Canonical_Combining_Class=	-ABOVE_right}', "");
Expect(0, 12333, '\p{Canonical_Combining_Class=	-ABOVE_right}', "");
Expect(1, 12333, '\p{^Canonical_Combining_Class=	-ABOVE_right}', "");
Expect(1, 12333, '\P{Canonical_Combining_Class=	-ABOVE_right}', "");
Expect(0, 12333, '\P{^Canonical_Combining_Class=	-ABOVE_right}', "");
Error('\p{Ccc= :=AR}');
Error('\P{Ccc= :=AR}');
Expect(1, 12332, '\p{Ccc=ar}', "");
Expect(0, 12332, '\p{^Ccc=ar}', "");
Expect(0, 12332, '\P{Ccc=ar}', "");
Expect(1, 12332, '\P{^Ccc=ar}', "");
Expect(0, 12333, '\p{Ccc=ar}', "");
Expect(1, 12333, '\p{^Ccc=ar}', "");
Expect(1, 12333, '\P{Ccc=ar}', "");
Expect(0, 12333, '\P{^Ccc=ar}', "");
Expect(1, 12332, '\p{Ccc= -ar}', "");
Expect(0, 12332, '\p{^Ccc= -ar}', "");
Expect(0, 12332, '\P{Ccc= -ar}', "");
Expect(1, 12332, '\P{^Ccc= -ar}', "");
Expect(0, 12333, '\p{Ccc= -ar}', "");
Expect(1, 12333, '\p{^Ccc= -ar}', "");
Expect(1, 12333, '\P{Ccc= -ar}', "");
Expect(0, 12333, '\P{^Ccc= -ar}', "");
Error('\p{Is_Canonical_Combining_Class=_:=000000232}');
Error('\P{Is_Canonical_Combining_Class=_:=000000232}');
Expect(1, 12332, '\p{Is_Canonical_Combining_Class=000_000_023_2}', "");
Expect(0, 12332, '\p{^Is_Canonical_Combining_Class=000_000_023_2}', "");
Expect(0, 12332, '\P{Is_Canonical_Combining_Class=000_000_023_2}', "");
Expect(1, 12332, '\P{^Is_Canonical_Combining_Class=000_000_023_2}', "");
Expect(0, 12333, '\p{Is_Canonical_Combining_Class=000_000_023_2}', "");
Expect(1, 12333, '\p{^Is_Canonical_Combining_Class=000_000_023_2}', "");
Expect(1, 12333, '\P{Is_Canonical_Combining_Class=000_000_023_2}', "");
Expect(0, 12333, '\P{^Is_Canonical_Combining_Class=000_000_023_2}', "");
Error('\p{Is_Ccc=/a/_	ABOVE_Right}');
Error('\P{Is_Ccc=/a/_	ABOVE_Right}');
Expect(1, 12332, '\p{Is_Ccc=aboveright}', "");
Expect(0, 12332, '\p{^Is_Ccc=aboveright}', "");
Expect(0, 12332, '\P{Is_Ccc=aboveright}', "");
Expect(1, 12332, '\P{^Is_Ccc=aboveright}', "");
Expect(0, 12333, '\p{Is_Ccc=aboveright}', "");
Expect(1, 12333, '\p{^Is_Ccc=aboveright}', "");
Expect(1, 12333, '\P{Is_Ccc=aboveright}', "");
Expect(0, 12333, '\P{^Is_Ccc=aboveright}', "");
Expect(1, 12332, '\p{Is_Ccc=_ Above_Right}', "");
Expect(0, 12332, '\p{^Is_Ccc=_ Above_Right}', "");
Expect(0, 12332, '\P{Is_Ccc=_ Above_Right}', "");
Expect(1, 12332, '\P{^Is_Ccc=_ Above_Right}', "");
Expect(0, 12333, '\p{Is_Ccc=_ Above_Right}', "");
Expect(1, 12333, '\p{^Is_Ccc=_ Above_Right}', "");
Expect(1, 12333, '\P{Is_Ccc=_ Above_Right}', "");
Expect(0, 12333, '\P{^Is_Ccc=_ Above_Right}', "");
Error('\p{Canonical_Combining_Class=:= ATTACHED_Above}');
Error('\P{Canonical_Combining_Class=:= ATTACHED_Above}');
Expect(1, 7630, '\p{Canonical_Combining_Class:	attachedabove}', "");
Expect(0, 7630, '\p{^Canonical_Combining_Class:	attachedabove}', "");
Expect(0, 7630, '\P{Canonical_Combining_Class:	attachedabove}', "");
Expect(1, 7630, '\P{^Canonical_Combining_Class:	attachedabove}', "");
Expect(0, 7631, '\p{Canonical_Combining_Class:	attachedabove}', "");
Expect(1, 7631, '\p{^Canonical_Combining_Class:	attachedabove}', "");
Expect(1, 7631, '\P{Canonical_Combining_Class:	attachedabove}', "");
Expect(0, 7631, '\P{^Canonical_Combining_Class:	attachedabove}', "");
Expect(1, 7630, '\p{Canonical_Combining_Class: Attached_Above}', "");
Expect(0, 7630, '\p{^Canonical_Combining_Class: Attached_Above}', "");
Expect(0, 7630, '\P{Canonical_Combining_Class: Attached_Above}', "");
Expect(1, 7630, '\P{^Canonical_Combining_Class: Attached_Above}', "");
Expect(0, 7631, '\p{Canonical_Combining_Class: Attached_Above}', "");
Expect(1, 7631, '\p{^Canonical_Combining_Class: Attached_Above}', "");
Expect(1, 7631, '\P{Canonical_Combining_Class: Attached_Above}', "");
Expect(0, 7631, '\P{^Canonical_Combining_Class: Attached_Above}', "");
Error('\p{Ccc= -ATA/a/}');
Error('\P{Ccc= -ATA/a/}');
Expect(1, 7630, '\p{Ccc=ata}', "");
Expect(0, 7630, '\p{^Ccc=ata}', "");
Expect(0, 7630, '\P{Ccc=ata}', "");
Expect(1, 7630, '\P{^Ccc=ata}', "");
Expect(0, 7631, '\p{Ccc=ata}', "");
Expect(1, 7631, '\p{^Ccc=ata}', "");
Expect(1, 7631, '\P{Ccc=ata}', "");
Expect(0, 7631, '\P{^Ccc=ata}', "");
Expect(1, 7630, '\p{Ccc=- ata}', "");
Expect(0, 7630, '\p{^Ccc=- ata}', "");
Expect(0, 7630, '\P{Ccc=- ata}', "");
Expect(1, 7630, '\P{^Ccc=- ata}', "");
Expect(0, 7631, '\p{Ccc=- ata}', "");
Expect(1, 7631, '\p{^Ccc=- ata}', "");
Expect(1, 7631, '\P{Ccc=- ata}', "");
Expect(0, 7631, '\P{^Ccc=- ata}', "");
Error('\p{Is_Canonical_Combining_Class=--+021_4:=}');
Error('\P{Is_Canonical_Combining_Class=--+021_4:=}');
Expect(1, 7630, '\p{Is_Canonical_Combining_Class=00000214}', "");
Expect(0, 7630, '\p{^Is_Canonical_Combining_Class=00000214}', "");
Expect(0, 7630, '\P{Is_Canonical_Combining_Class=00000214}', "");
Expect(1, 7630, '\P{^Is_Canonical_Combining_Class=00000214}', "");
Expect(0, 7631, '\p{Is_Canonical_Combining_Class=00000214}', "");
Expect(1, 7631, '\p{^Is_Canonical_Combining_Class=00000214}', "");
Expect(1, 7631, '\P{Is_Canonical_Combining_Class=00000214}', "");
Expect(0, 7631, '\P{^Is_Canonical_Combining_Class=00000214}', "");
Error('\p{Is_Ccc= /a/attached_above}');
Error('\P{Is_Ccc= /a/attached_above}');
Expect(1, 7630, '\p{Is_Ccc=attachedabove}', "");
Expect(0, 7630, '\p{^Is_Ccc=attachedabove}', "");
Expect(0, 7630, '\P{Is_Ccc=attachedabove}', "");
Expect(1, 7630, '\P{^Is_Ccc=attachedabove}', "");
Expect(0, 7631, '\p{Is_Ccc=attachedabove}', "");
Expect(1, 7631, '\p{^Is_Ccc=attachedabove}', "");
Expect(1, 7631, '\P{Is_Ccc=attachedabove}', "");
Expect(0, 7631, '\P{^Is_Ccc=attachedabove}', "");
Expect(1, 7630, '\p{Is_Ccc=	-Attached_Above}', "");
Expect(0, 7630, '\p{^Is_Ccc=	-Attached_Above}', "");
Expect(0, 7630, '\P{Is_Ccc=	-Attached_Above}', "");
Expect(1, 7630, '\P{^Is_Ccc=	-Attached_Above}', "");
Expect(0, 7631, '\p{Is_Ccc=	-Attached_Above}', "");
Expect(1, 7631, '\p{^Is_Ccc=	-Attached_Above}', "");
Expect(1, 7631, '\P{Is_Ccc=	-Attached_Above}', "");
Expect(0, 7631, '\P{^Is_Ccc=	-Attached_Above}', "");
Error('\p{Canonical_Combining_Class=/a/_	Attached_above_Right}');
Error('\P{Canonical_Combining_Class=/a/_	Attached_above_Right}');
Expect(1, 119154, '\p{Canonical_Combining_Class=attachedaboveright}', "");
Expect(0, 119154, '\p{^Canonical_Combining_Class=attachedaboveright}', "");
Expect(0, 119154, '\P{Canonical_Combining_Class=attachedaboveright}', "");
Expect(1, 119154, '\P{^Canonical_Combining_Class=attachedaboveright}', "");
Expect(0, 119155, '\p{Canonical_Combining_Class=attachedaboveright}', "");
Expect(1, 119155, '\p{^Canonical_Combining_Class=attachedaboveright}', "");
Expect(1, 119155, '\P{Canonical_Combining_Class=attachedaboveright}', "");
Expect(0, 119155, '\P{^Canonical_Combining_Class=attachedaboveright}', "");
Expect(1, 119154, '\p{Canonical_Combining_Class= 	attached_ABOVE_Right}', "");
Expect(0, 119154, '\p{^Canonical_Combining_Class= 	attached_ABOVE_Right}', "");
Expect(0, 119154, '\P{Canonical_Combining_Class= 	attached_ABOVE_Right}', "");
Expect(1, 119154, '\P{^Canonical_Combining_Class= 	attached_ABOVE_Right}', "");
Expect(0, 119155, '\p{Canonical_Combining_Class= 	attached_ABOVE_Right}', "");
Expect(1, 119155, '\p{^Canonical_Combining_Class= 	attached_ABOVE_Right}', "");
Expect(1, 119155, '\P{Canonical_Combining_Class= 	attached_ABOVE_Right}', "");
Expect(0, 119155, '\P{^Canonical_Combining_Class= 	attached_ABOVE_Right}', "");
Error('\p{Ccc= -atar/a/}');
Error('\P{Ccc= -atar/a/}');
Expect(1, 119154, '\p{Ccc=atar}', "");
Expect(0, 119154, '\p{^Ccc=atar}', "");
Expect(0, 119154, '\P{Ccc=atar}', "");
Expect(1, 119154, '\P{^Ccc=atar}', "");
Expect(0, 119155, '\p{Ccc=atar}', "");
Expect(1, 119155, '\p{^Ccc=atar}', "");
Expect(1, 119155, '\P{Ccc=atar}', "");
Expect(0, 119155, '\P{^Ccc=atar}', "");
Expect(1, 119154, '\p{Ccc=-atar}', "");
Expect(0, 119154, '\p{^Ccc=-atar}', "");
Expect(0, 119154, '\P{Ccc=-atar}', "");
Expect(1, 119154, '\P{^Ccc=-atar}', "");
Expect(0, 119155, '\p{Ccc=-atar}', "");
Expect(1, 119155, '\p{^Ccc=-atar}', "");
Expect(1, 119155, '\P{Ccc=-atar}', "");
Expect(0, 119155, '\P{^Ccc=-atar}', "");
Error('\p{Is_Canonical_Combining_Class=/a/-0000000216}');
Error('\P{Is_Canonical_Combining_Class=/a/-0000000216}');
Expect(1, 119154, '\p{Is_Canonical_Combining_Class:   21_6}', "");
Expect(0, 119154, '\p{^Is_Canonical_Combining_Class:   21_6}', "");
Expect(0, 119154, '\P{Is_Canonical_Combining_Class:   21_6}', "");
Expect(1, 119154, '\P{^Is_Canonical_Combining_Class:   21_6}', "");
Expect(0, 119155, '\p{Is_Canonical_Combining_Class:   21_6}', "");
Expect(1, 119155, '\p{^Is_Canonical_Combining_Class:   21_6}', "");
Expect(1, 119155, '\P{Is_Canonical_Combining_Class:   21_6}', "");
Expect(0, 119155, '\P{^Is_Canonical_Combining_Class:   21_6}', "");
Error('\p{Is_Ccc=/a/Attached_ABOVE_right}');
Error('\P{Is_Ccc=/a/Attached_ABOVE_right}');
Expect(1, 119154, '\p{Is_Ccc=attachedaboveright}', "");
Expect(0, 119154, '\p{^Is_Ccc=attachedaboveright}', "");
Expect(0, 119154, '\P{Is_Ccc=attachedaboveright}', "");
Expect(1, 119154, '\P{^Is_Ccc=attachedaboveright}', "");
Expect(0, 119155, '\p{Is_Ccc=attachedaboveright}', "");
Expect(1, 119155, '\p{^Is_Ccc=attachedaboveright}', "");
Expect(1, 119155, '\P{Is_Ccc=attachedaboveright}', "");
Expect(0, 119155, '\P{^Is_Ccc=attachedaboveright}', "");
Expect(1, 119154, '\p{Is_Ccc=-Attached_above_right}', "");
Expect(0, 119154, '\p{^Is_Ccc=-Attached_above_right}', "");
Expect(0, 119154, '\P{Is_Ccc=-Attached_above_right}', "");
Expect(1, 119154, '\P{^Is_Ccc=-Attached_above_right}', "");
Expect(0, 119155, '\p{Is_Ccc=-Attached_above_right}', "");
Expect(1, 119155, '\p{^Is_Ccc=-Attached_above_right}', "");
Expect(1, 119155, '\P{Is_Ccc=-Attached_above_right}', "");
Expect(0, 119155, '\P{^Is_Ccc=-Attached_above_right}', "");
Error('\p{Canonical_Combining_Class= :=ATTACHED_BELOW}');
Error('\P{Canonical_Combining_Class= :=ATTACHED_BELOW}');
Expect(1, 7632, '\p{Canonical_Combining_Class=attachedbelow}', "");
Expect(0, 7632, '\p{^Canonical_Combining_Class=attachedbelow}', "");
Expect(0, 7632, '\P{Canonical_Combining_Class=attachedbelow}', "");
Expect(1, 7632, '\P{^Canonical_Combining_Class=attachedbelow}', "");
Expect(0, 7633, '\p{Canonical_Combining_Class=attachedbelow}', "");
Expect(1, 7633, '\p{^Canonical_Combining_Class=attachedbelow}', "");
Expect(1, 7633, '\P{Canonical_Combining_Class=attachedbelow}', "");
Expect(0, 7633, '\P{^Canonical_Combining_Class=attachedbelow}', "");
Expect(1, 7632, '\p{Canonical_Combining_Class= -Attached_BELOW}', "");
Expect(0, 7632, '\p{^Canonical_Combining_Class= -Attached_BELOW}', "");
Expect(0, 7632, '\P{Canonical_Combining_Class= -Attached_BELOW}', "");
Expect(1, 7632, '\P{^Canonical_Combining_Class= -Attached_BELOW}', "");
Expect(0, 7633, '\p{Canonical_Combining_Class= -Attached_BELOW}', "");
Expect(1, 7633, '\p{^Canonical_Combining_Class= -Attached_BELOW}', "");
Expect(1, 7633, '\P{Canonical_Combining_Class= -Attached_BELOW}', "");
Expect(0, 7633, '\P{^Canonical_Combining_Class= -Attached_BELOW}', "");
Error('\p{Ccc=__atb:=}');
Error('\P{Ccc=__atb:=}');
Expect(1, 7632, '\p{Ccc=atb}', "");
Expect(0, 7632, '\p{^Ccc=atb}', "");
Expect(0, 7632, '\P{Ccc=atb}', "");
Expect(1, 7632, '\P{^Ccc=atb}', "");
Expect(0, 7633, '\p{Ccc=atb}', "");
Expect(1, 7633, '\p{^Ccc=atb}', "");
Expect(1, 7633, '\P{Ccc=atb}', "");
Expect(0, 7633, '\P{^Ccc=atb}', "");
Expect(1, 7632, '\p{Ccc=- ATB}', "");
Expect(0, 7632, '\p{^Ccc=- ATB}', "");
Expect(0, 7632, '\P{Ccc=- ATB}', "");
Expect(1, 7632, '\P{^Ccc=- ATB}', "");
Expect(0, 7633, '\p{Ccc=- ATB}', "");
Expect(1, 7633, '\p{^Ccc=- ATB}', "");
Expect(1, 7633, '\P{Ccc=- ATB}', "");
Expect(0, 7633, '\P{^Ccc=- ATB}', "");
Error('\p{Is_Canonical_Combining_Class=:=_+00000000202}');
Error('\P{Is_Canonical_Combining_Class=:=_+00000000202}');
Expect(1, 7632, '\p{Is_Canonical_Combining_Class=+20_2}', "");
Expect(0, 7632, '\p{^Is_Canonical_Combining_Class=+20_2}', "");
Expect(0, 7632, '\P{Is_Canonical_Combining_Class=+20_2}', "");
Expect(1, 7632, '\P{^Is_Canonical_Combining_Class=+20_2}', "");
Expect(0, 7633, '\p{Is_Canonical_Combining_Class=+20_2}', "");
Expect(1, 7633, '\p{^Is_Canonical_Combining_Class=+20_2}', "");
Expect(1, 7633, '\P{Is_Canonical_Combining_Class=+20_2}', "");
Expect(0, 7633, '\P{^Is_Canonical_Combining_Class=+20_2}', "");
Error('\p{Is_Ccc=	/a/ATTACHED_below}');
Error('\P{Is_Ccc=	/a/ATTACHED_below}');
Expect(1, 7632, '\p{Is_Ccc:   attachedbelow}', "");
Expect(0, 7632, '\p{^Is_Ccc:   attachedbelow}', "");
Expect(0, 7632, '\P{Is_Ccc:   attachedbelow}', "");
Expect(1, 7632, '\P{^Is_Ccc:   attachedbelow}', "");
Expect(0, 7633, '\p{Is_Ccc:   attachedbelow}', "");
Expect(1, 7633, '\p{^Is_Ccc:   attachedbelow}', "");
Expect(1, 7633, '\P{Is_Ccc:   attachedbelow}', "");
Expect(0, 7633, '\P{^Is_Ccc:   attachedbelow}', "");
Expect(1, 7632, '\p{Is_Ccc:-Attached_BELOW}', "");
Expect(0, 7632, '\p{^Is_Ccc:-Attached_BELOW}', "");
Expect(0, 7632, '\P{Is_Ccc:-Attached_BELOW}', "");
Expect(1, 7632, '\P{^Is_Ccc:-Attached_BELOW}', "");
Expect(0, 7633, '\p{Is_Ccc:-Attached_BELOW}', "");
Expect(1, 7633, '\p{^Is_Ccc:-Attached_BELOW}', "");
Expect(1, 7633, '\P{Is_Ccc:-Attached_BELOW}', "");
Expect(0, 7633, '\P{^Is_Ccc:-Attached_BELOW}', "");
Error('\p{Canonical_Combining_Class=	-Attached_BELOW_LEFT/a/}');
Error('\P{Canonical_Combining_Class=	-Attached_BELOW_LEFT/a/}');
Expect(0, 1, '\p{Canonical_Combining_Class=attachedbelowleft}', "");
Expect(1, 1, '\p{^Canonical_Combining_Class=attachedbelowleft}', "");
Expect(1, 1, '\P{Canonical_Combining_Class=attachedbelowleft}', "");
Expect(0, 1, '\P{^Canonical_Combining_Class=attachedbelowleft}', "");
Expect(0, 1, '\p{Canonical_Combining_Class=ATTACHED_Below_Left}', "");
Expect(1, 1, '\p{^Canonical_Combining_Class=ATTACHED_Below_Left}', "");
Expect(1, 1, '\P{Canonical_Combining_Class=ATTACHED_Below_Left}', "");
Expect(0, 1, '\P{^Canonical_Combining_Class=ATTACHED_Below_Left}', "");
Error('\p{Ccc=/a/_-ATBL}');
Error('\P{Ccc=/a/_-ATBL}');
Expect(0, 1, '\p{Ccc=atbl}', "");
Expect(1, 1, '\p{^Ccc=atbl}', "");
Expect(1, 1, '\P{Ccc=atbl}', "");
Expect(0, 1, '\P{^Ccc=atbl}', "");
Expect(0, 1, '\p{Ccc=-ATBL}', "");
Expect(1, 1, '\p{^Ccc=-ATBL}', "");
Expect(1, 1, '\P{Ccc=-ATBL}', "");
Expect(0, 1, '\P{^Ccc=-ATBL}', "");
Error('\p{Is_Canonical_Combining_Class=:=0000000200}');
Error('\P{Is_Canonical_Combining_Class=:=0000000200}');
Expect(0, 1, '\p{Is_Canonical_Combining_Class=0200}', "");
Expect(1, 1, '\p{^Is_Canonical_Combining_Class=0200}', "");
Expect(1, 1, '\P{Is_Canonical_Combining_Class=0200}', "");
Expect(0, 1, '\P{^Is_Canonical_Combining_Class=0200}', "");
Error('\p{Is_Ccc= -Attached_Below_Left/a/}');
Error('\P{Is_Ccc= -Attached_Below_Left/a/}');
Expect(0, 1, '\p{Is_Ccc=attachedbelowleft}', "");
Expect(1, 1, '\p{^Is_Ccc=attachedbelowleft}', "");
Expect(1, 1, '\P{Is_Ccc=attachedbelowleft}', "");
Expect(0, 1, '\P{^Is_Ccc=attachedbelowleft}', "");
Expect(0, 1, '\p{Is_Ccc= Attached_Below_left}', "");
Expect(1, 1, '\p{^Is_Ccc= Attached_Below_left}', "");
Expect(1, 1, '\P{Is_Ccc= Attached_Below_left}', "");
Expect(0, 1, '\P{^Is_Ccc= Attached_Below_left}', "");
Error('\p{Canonical_Combining_Class=:= BELOW}');
Error('\P{Canonical_Combining_Class=:= BELOW}');
Expect(1, 119179, '\p{Canonical_Combining_Class=below}', "");
Expect(0, 119179, '\p{^Canonical_Combining_Class=below}', "");
Expect(0, 119179, '\P{Canonical_Combining_Class=below}', "");
Expect(1, 119179, '\P{^Canonical_Combining_Class=below}', "");
Expect(0, 119180, '\p{Canonical_Combining_Class=below}', "");
Expect(1, 119180, '\p{^Canonical_Combining_Class=below}', "");
Expect(1, 119180, '\P{Canonical_Combining_Class=below}', "");
Expect(0, 119180, '\P{^Canonical_Combining_Class=below}', "");
Expect(1, 119179, '\p{Canonical_Combining_Class= -below}', "");
Expect(0, 119179, '\p{^Canonical_Combining_Class= -below}', "");
Expect(0, 119179, '\P{Canonical_Combining_Class= -below}', "");
Expect(1, 119179, '\P{^Canonical_Combining_Class= -below}', "");
Expect(0, 119180, '\p{Canonical_Combining_Class= -below}', "");
Expect(1, 119180, '\p{^Canonical_Combining_Class= -below}', "");
Expect(1, 119180, '\P{Canonical_Combining_Class= -below}', "");
Expect(0, 119180, '\P{^Canonical_Combining_Class= -below}', "");
Error('\p{Ccc=_:=B}');
Error('\P{Ccc=_:=B}');
Expect(1, 119179, '\p{Ccc=b}', "");
Expect(0, 119179, '\p{^Ccc=b}', "");
Expect(0, 119179, '\P{Ccc=b}', "");
Expect(1, 119179, '\P{^Ccc=b}', "");
Expect(0, 119180, '\p{Ccc=b}', "");
Expect(1, 119180, '\p{^Ccc=b}', "");
Expect(1, 119180, '\P{Ccc=b}', "");
Expect(0, 119180, '\P{^Ccc=b}', "");
Expect(1, 119179, '\p{Ccc: _B}', "");
Expect(0, 119179, '\p{^Ccc: _B}', "");
Expect(0, 119179, '\P{Ccc: _B}', "");
Expect(1, 119179, '\P{^Ccc: _B}', "");
Expect(0, 119180, '\p{Ccc: _B}', "");
Expect(1, 119180, '\p{^Ccc: _B}', "");
Expect(1, 119180, '\P{Ccc: _B}', "");
Expect(0, 119180, '\P{^Ccc: _B}', "");
Error('\p{Is_Canonical_Combining_Class=	:=22_0}');
Error('\P{Is_Canonical_Combining_Class=	:=22_0}');
Expect(1, 119179, '\p{Is_Canonical_Combining_Class=+00_02_20}', "");
Expect(0, 119179, '\p{^Is_Canonical_Combining_Class=+00_02_20}', "");
Expect(0, 119179, '\P{Is_Canonical_Combining_Class=+00_02_20}', "");
Expect(1, 119179, '\P{^Is_Canonical_Combining_Class=+00_02_20}', "");
Expect(0, 119180, '\p{Is_Canonical_Combining_Class=+00_02_20}', "");
Expect(1, 119180, '\p{^Is_Canonical_Combining_Class=+00_02_20}', "");
Expect(1, 119180, '\P{Is_Canonical_Combining_Class=+00_02_20}', "");
Expect(0, 119180, '\P{^Is_Canonical_Combining_Class=+00_02_20}', "");
Error('\p{Is_Ccc=-	Below/a/}');
Error('\P{Is_Ccc=-	Below/a/}');
Expect(1, 119179, '\p{Is_Ccc=below}', "");
Expect(0, 119179, '\p{^Is_Ccc=below}', "");
Expect(0, 119179, '\P{Is_Ccc=below}', "");
Expect(1, 119179, '\P{^Is_Ccc=below}', "");
Expect(0, 119180, '\p{Is_Ccc=below}', "");
Expect(1, 119180, '\p{^Is_Ccc=below}', "");
Expect(1, 119180, '\P{Is_Ccc=below}', "");
Expect(0, 119180, '\P{^Is_Ccc=below}', "");
Expect(1, 119179, '\p{Is_Ccc= 	Below}', "");
Expect(0, 119179, '\p{^Is_Ccc= 	Below}', "");
Expect(0, 119179, '\P{Is_Ccc= 	Below}', "");
Expect(1, 119179, '\P{^Is_Ccc= 	Below}', "");
Expect(0, 119180, '\p{Is_Ccc= 	Below}', "");
Expect(1, 119180, '\p{^Is_Ccc= 	Below}', "");
Expect(1, 119180, '\P{Is_Ccc= 	Below}', "");
Expect(0, 119180, '\P{^Is_Ccc= 	Below}', "");
Error('\p{Canonical_Combining_Class=:=_-Below_left}');
Error('\P{Canonical_Combining_Class=:=_-Below_left}');
Expect(1, 12330, '\p{Canonical_Combining_Class=belowleft}', "");
Expect(0, 12330, '\p{^Canonical_Combining_Class=belowleft}', "");
Expect(0, 12330, '\P{Canonical_Combining_Class=belowleft}', "");
Expect(1, 12330, '\P{^Canonical_Combining_Class=belowleft}', "");
Expect(0, 12331, '\p{Canonical_Combining_Class=belowleft}', "");
Expect(1, 12331, '\p{^Canonical_Combining_Class=belowleft}', "");
Expect(1, 12331, '\P{Canonical_Combining_Class=belowleft}', "");
Expect(0, 12331, '\P{^Canonical_Combining_Class=belowleft}', "");
Expect(1, 12330, '\p{Canonical_Combining_Class= below_left}', "");
Expect(0, 12330, '\p{^Canonical_Combining_Class= below_left}', "");
Expect(0, 12330, '\P{Canonical_Combining_Class= below_left}', "");
Expect(1, 12330, '\P{^Canonical_Combining_Class= below_left}', "");
Expect(0, 12331, '\p{Canonical_Combining_Class= below_left}', "");
Expect(1, 12331, '\p{^Canonical_Combining_Class= below_left}', "");
Expect(1, 12331, '\P{Canonical_Combining_Class= below_left}', "");
Expect(0, 12331, '\P{^Canonical_Combining_Class= below_left}', "");
Error('\p{Ccc=_	BL:=}');
Error('\P{Ccc=_	BL:=}');
Expect(1, 12330, '\p{Ccc=bl}', "");
Expect(0, 12330, '\p{^Ccc=bl}', "");
Expect(0, 12330, '\P{Ccc=bl}', "");
Expect(1, 12330, '\P{^Ccc=bl}', "");
Expect(0, 12331, '\p{Ccc=bl}', "");
Expect(1, 12331, '\p{^Ccc=bl}', "");
Expect(1, 12331, '\P{Ccc=bl}', "");
Expect(0, 12331, '\P{^Ccc=bl}', "");
Expect(1, 12330, '\p{Ccc=	_BL}', "");
Expect(0, 12330, '\p{^Ccc=	_BL}', "");
Expect(0, 12330, '\P{Ccc=	_BL}', "");
Expect(1, 12330, '\P{^Ccc=	_BL}', "");
Expect(0, 12331, '\p{Ccc=	_BL}', "");
Expect(1, 12331, '\p{^Ccc=	_BL}', "");
Expect(1, 12331, '\P{Ccc=	_BL}', "");
Expect(0, 12331, '\P{^Ccc=	_BL}', "");
Error('\p{Is_Canonical_Combining_Class=:=00218}');
Error('\P{Is_Canonical_Combining_Class=:=00218}');
Expect(1, 12330, '\p{Is_Canonical_Combining_Class=0_0_0_0_0_0_0_00218}', "");
Expect(0, 12330, '\p{^Is_Canonical_Combining_Class=0_0_0_0_0_0_0_00218}', "");
Expect(0, 12330, '\P{Is_Canonical_Combining_Class=0_0_0_0_0_0_0_00218}', "");
Expect(1, 12330, '\P{^Is_Canonical_Combining_Class=0_0_0_0_0_0_0_00218}', "");
Expect(0, 12331, '\p{Is_Canonical_Combining_Class=0_0_0_0_0_0_0_00218}', "");
Expect(1, 12331, '\p{^Is_Canonical_Combining_Class=0_0_0_0_0_0_0_00218}', "");
Expect(1, 12331, '\P{Is_Canonical_Combining_Class=0_0_0_0_0_0_0_00218}', "");
Expect(0, 12331, '\P{^Is_Canonical_Combining_Class=0_0_0_0_0_0_0_00218}', "");
Error('\p{Is_Ccc=	/a/Below_LEFT}');
Error('\P{Is_Ccc=	/a/Below_LEFT}');
Expect(1, 12330, '\p{Is_Ccc=belowleft}', "");
Expect(0, 12330, '\p{^Is_Ccc=belowleft}', "");
Expect(0, 12330, '\P{Is_Ccc=belowleft}', "");
Expect(1, 12330, '\P{^Is_Ccc=belowleft}', "");
Expect(0, 12331, '\p{Is_Ccc=belowleft}', "");
Expect(1, 12331, '\p{^Is_Ccc=belowleft}', "");
Expect(1, 12331, '\P{Is_Ccc=belowleft}', "");
Expect(0, 12331, '\P{^Is_Ccc=belowleft}', "");
Expect(1, 12330, '\p{Is_Ccc:    Below_Left}', "");
Expect(0, 12330, '\p{^Is_Ccc:    Below_Left}', "");
Expect(0, 12330, '\P{Is_Ccc:    Below_Left}', "");
Expect(1, 12330, '\P{^Is_Ccc:    Below_Left}', "");
Expect(0, 12331, '\p{Is_Ccc:    Below_Left}', "");
Expect(1, 12331, '\p{^Is_Ccc:    Below_Left}', "");
Expect(1, 12331, '\P{Is_Ccc:    Below_Left}', "");
Expect(0, 12331, '\P{^Is_Ccc:    Below_Left}', "");
Error('\p{Canonical_Combining_Class=	:=BELOW_right}');
Error('\P{Canonical_Combining_Class=	:=BELOW_right}');
Expect(1, 12333, '\p{Canonical_Combining_Class=belowright}', "");
Expect(0, 12333, '\p{^Canonical_Combining_Class=belowright}', "");
Expect(0, 12333, '\P{Canonical_Combining_Class=belowright}', "");
Expect(1, 12333, '\P{^Canonical_Combining_Class=belowright}', "");
Expect(0, 12334, '\p{Canonical_Combining_Class=belowright}', "");
Expect(1, 12334, '\p{^Canonical_Combining_Class=belowright}', "");
Expect(1, 12334, '\P{Canonical_Combining_Class=belowright}', "");
Expect(0, 12334, '\P{^Canonical_Combining_Class=belowright}', "");
Expect(1, 12333, '\p{Canonical_Combining_Class= BELOW_Right}', "");
Expect(0, 12333, '\p{^Canonical_Combining_Class= BELOW_Right}', "");
Expect(0, 12333, '\P{Canonical_Combining_Class= BELOW_Right}', "");
Expect(1, 12333, '\P{^Canonical_Combining_Class= BELOW_Right}', "");
Expect(0, 12334, '\p{Canonical_Combining_Class= BELOW_Right}', "");
Expect(1, 12334, '\p{^Canonical_Combining_Class= BELOW_Right}', "");
Expect(1, 12334, '\P{Canonical_Combining_Class= BELOW_Right}', "");
Expect(0, 12334, '\P{^Canonical_Combining_Class= BELOW_Right}', "");
Error('\p{Ccc=_/a/BR}');
Error('\P{Ccc=_/a/BR}');
Expect(1, 12333, '\p{Ccc: br}', "");
Expect(0, 12333, '\p{^Ccc: br}', "");
Expect(0, 12333, '\P{Ccc: br}', "");
Expect(1, 12333, '\P{^Ccc: br}', "");
Expect(0, 12334, '\p{Ccc: br}', "");
Expect(1, 12334, '\p{^Ccc: br}', "");
Expect(1, 12334, '\P{Ccc: br}', "");
Expect(0, 12334, '\P{^Ccc: br}', "");
Expect(1, 12333, '\p{Ccc=-br}', "");
Expect(0, 12333, '\p{^Ccc=-br}', "");
Expect(0, 12333, '\P{Ccc=-br}', "");
Expect(1, 12333, '\P{^Ccc=-br}', "");
Expect(0, 12334, '\p{Ccc=-br}', "");
Expect(1, 12334, '\p{^Ccc=-br}', "");
Expect(1, 12334, '\P{Ccc=-br}', "");
Expect(0, 12334, '\P{^Ccc=-br}', "");
Error('\p{Is_Canonical_Combining_Class= _+00_22_2:=}');
Error('\P{Is_Canonical_Combining_Class= _+00_22_2:=}');
Expect(1, 12333, '\p{Is_Canonical_Combining_Class=+00_22_2}', "");
Expect(0, 12333, '\p{^Is_Canonical_Combining_Class=+00_22_2}', "");
Expect(0, 12333, '\P{Is_Canonical_Combining_Class=+00_22_2}', "");
Expect(1, 12333, '\P{^Is_Canonical_Combining_Class=+00_22_2}', "");
Expect(0, 12334, '\p{Is_Canonical_Combining_Class=+00_22_2}', "");
Expect(1, 12334, '\p{^Is_Canonical_Combining_Class=+00_22_2}', "");
Expect(1, 12334, '\P{Is_Canonical_Combining_Class=+00_22_2}', "");
Expect(0, 12334, '\P{^Is_Canonical_Combining_Class=+00_22_2}', "");
Error('\p{Is_Ccc=-_Below_Right/a/}');
Error('\P{Is_Ccc=-_Below_Right/a/}');
Expect(1, 12333, '\p{Is_Ccc=belowright}', "");
Expect(0, 12333, '\p{^Is_Ccc=belowright}', "");
Expect(0, 12333, '\P{Is_Ccc=belowright}', "");
Expect(1, 12333, '\P{^Is_Ccc=belowright}', "");
Expect(0, 12334, '\p{Is_Ccc=belowright}', "");
Expect(1, 12334, '\p{^Is_Ccc=belowright}', "");
Expect(1, 12334, '\P{Is_Ccc=belowright}', "");
Expect(0, 12334, '\P{^Is_Ccc=belowright}', "");
Expect(1, 12333, '\p{Is_Ccc=-	BELOW_Right}', "");
Expect(0, 12333, '\p{^Is_Ccc=-	BELOW_Right}', "");
Expect(0, 12333, '\P{Is_Ccc=-	BELOW_Right}', "");
Expect(1, 12333, '\P{^Is_Ccc=-	BELOW_Right}', "");
Expect(0, 12334, '\p{Is_Ccc=-	BELOW_Right}', "");
Expect(1, 12334, '\p{^Is_Ccc=-	BELOW_Right}', "");
Expect(1, 12334, '\P{Is_Ccc=-	BELOW_Right}', "");
Expect(0, 12334, '\P{^Is_Ccc=-	BELOW_Right}', "");
Error('\p{Canonical_Combining_Class=_ ccc10/a/}');
Error('\P{Canonical_Combining_Class=_ ccc10/a/}');
Expect(1, 1456, '\p{Canonical_Combining_Class=ccc10}', "");
Expect(0, 1456, '\p{^Canonical_Combining_Class=ccc10}', "");
Expect(0, 1456, '\P{Canonical_Combining_Class=ccc10}', "");
Expect(1, 1456, '\P{^Canonical_Combining_Class=ccc10}', "");
Expect(0, 1457, '\p{Canonical_Combining_Class=ccc10}', "");
Expect(1, 1457, '\p{^Canonical_Combining_Class=ccc10}', "");
Expect(1, 1457, '\P{Canonical_Combining_Class=ccc10}', "");
Expect(0, 1457, '\P{^Canonical_Combining_Class=ccc10}', "");
Expect(1, 1456, '\p{Canonical_Combining_Class=- ccc10}', "");
Expect(0, 1456, '\p{^Canonical_Combining_Class=- ccc10}', "");
Expect(0, 1456, '\P{Canonical_Combining_Class=- ccc10}', "");
Expect(1, 1456, '\P{^Canonical_Combining_Class=- ccc10}', "");
Expect(0, 1457, '\p{Canonical_Combining_Class=- ccc10}', "");
Expect(1, 1457, '\p{^Canonical_Combining_Class=- ccc10}', "");
Expect(1, 1457, '\P{Canonical_Combining_Class=- ccc10}', "");
Expect(0, 1457, '\P{^Canonical_Combining_Class=- ccc10}', "");
Error('\p{Ccc= _+0_0_0_0_10:=}');
Error('\P{Ccc= _+0_0_0_0_10:=}');
Expect(1, 1456, '\p{Ccc=0000000010}', "");
Expect(0, 1456, '\p{^Ccc=0000000010}', "");
Expect(0, 1456, '\P{Ccc=0000000010}', "");
Expect(1, 1456, '\P{^Ccc=0000000010}', "");
Expect(0, 1457, '\p{Ccc=0000000010}', "");
Expect(1, 1457, '\p{^Ccc=0000000010}', "");
Expect(1, 1457, '\P{Ccc=0000000010}', "");
Expect(0, 1457, '\P{^Ccc=0000000010}', "");
Error('\p{Is_Canonical_Combining_Class=:=ccc10}');
Error('\P{Is_Canonical_Combining_Class=:=ccc10}');
Expect(1, 1456, '\p{Is_Canonical_Combining_Class=ccc10}', "");
Expect(0, 1456, '\p{^Is_Canonical_Combining_Class=ccc10}', "");
Expect(0, 1456, '\P{Is_Canonical_Combining_Class=ccc10}', "");
Expect(1, 1456, '\P{^Is_Canonical_Combining_Class=ccc10}', "");
Expect(0, 1457, '\p{Is_Canonical_Combining_Class=ccc10}', "");
Expect(1, 1457, '\p{^Is_Canonical_Combining_Class=ccc10}', "");
Expect(1, 1457, '\P{Is_Canonical_Combining_Class=ccc10}', "");
Expect(0, 1457, '\P{^Is_Canonical_Combining_Class=ccc10}', "");
Expect(1, 1456, '\p{Is_Canonical_Combining_Class=_CCC10}', "");
Expect(0, 1456, '\p{^Is_Canonical_Combining_Class=_CCC10}', "");
Expect(0, 1456, '\P{Is_Canonical_Combining_Class=_CCC10}', "");
Expect(1, 1456, '\P{^Is_Canonical_Combining_Class=_CCC10}', "");
Expect(0, 1457, '\p{Is_Canonical_Combining_Class=_CCC10}', "");
Expect(1, 1457, '\p{^Is_Canonical_Combining_Class=_CCC10}', "");
Expect(1, 1457, '\P{Is_Canonical_Combining_Class=_CCC10}', "");
Expect(0, 1457, '\P{^Is_Canonical_Combining_Class=_CCC10}', "");
Error('\p{Is_Ccc=/a/	0_0_0_010}');
Error('\P{Is_Ccc=/a/	0_0_0_010}');
Expect(1, 1456, '\p{Is_Ccc=+00000000010}', "");
Expect(0, 1456, '\p{^Is_Ccc=+00000000010}', "");
Expect(0, 1456, '\P{Is_Ccc=+00000000010}', "");
Expect(1, 1456, '\P{^Is_Ccc=+00000000010}', "");
Expect(0, 1457, '\p{Is_Ccc=+00000000010}', "");
Expect(1, 1457, '\p{^Is_Ccc=+00000000010}', "");
Expect(1, 1457, '\P{Is_Ccc=+00000000010}', "");
Expect(0, 1457, '\P{^Is_Ccc=+00000000010}', "");
Error('\p{Canonical_Combining_Class=/a/	CCC103}');
Error('\P{Canonical_Combining_Class=/a/	CCC103}');
Expect(1, 3641, '\p{Canonical_Combining_Class=ccc103}', "");
Expect(0, 3641, '\p{^Canonical_Combining_Class=ccc103}', "");
Expect(0, 3641, '\P{Canonical_Combining_Class=ccc103}', "");
Expect(1, 3641, '\P{^Canonical_Combining_Class=ccc103}', "");
Expect(0, 3642, '\p{Canonical_Combining_Class=ccc103}', "");
Expect(1, 3642, '\p{^Canonical_Combining_Class=ccc103}', "");
Expect(1, 3642, '\P{Canonical_Combining_Class=ccc103}', "");
Expect(0, 3642, '\P{^Canonical_Combining_Class=ccc103}', "");
Expect(1, 3641, '\p{Canonical_Combining_Class:_	CCC103}', "");
Expect(0, 3641, '\p{^Canonical_Combining_Class:_	CCC103}', "");
Expect(0, 3641, '\P{Canonical_Combining_Class:_	CCC103}', "");
Expect(1, 3641, '\P{^Canonical_Combining_Class:_	CCC103}', "");
Expect(0, 3642, '\p{Canonical_Combining_Class:_	CCC103}', "");
Expect(1, 3642, '\p{^Canonical_Combining_Class:_	CCC103}', "");
Expect(1, 3642, '\P{Canonical_Combining_Class:_	CCC103}', "");
Expect(0, 3642, '\P{^Canonical_Combining_Class:_	CCC103}', "");
Error('\p{Ccc=:=  +00_00_01_03}');
Error('\P{Ccc=:=  +00_00_01_03}');
Expect(1, 3641, '\p{Ccc=0_0_0_0_0_0_0_1_03}', "");
Expect(0, 3641, '\p{^Ccc=0_0_0_0_0_0_0_1_03}', "");
Expect(0, 3641, '\P{Ccc=0_0_0_0_0_0_0_1_03}', "");
Expect(1, 3641, '\P{^Ccc=0_0_0_0_0_0_0_1_03}', "");
Expect(0, 3642, '\p{Ccc=0_0_0_0_0_0_0_1_03}', "");
Expect(1, 3642, '\p{^Ccc=0_0_0_0_0_0_0_1_03}', "");
Expect(1, 3642, '\P{Ccc=0_0_0_0_0_0_0_1_03}', "");
Expect(0, 3642, '\P{^Ccc=0_0_0_0_0_0_0_1_03}', "");
Error('\p{Is_Canonical_Combining_Class=-:=ccc103}');
Error('\P{Is_Canonical_Combining_Class=-:=ccc103}');
Expect(1, 3641, '\p{Is_Canonical_Combining_Class=ccc103}', "");
Expect(0, 3641, '\p{^Is_Canonical_Combining_Class=ccc103}', "");
Expect(0, 3641, '\P{Is_Canonical_Combining_Class=ccc103}', "");
Expect(1, 3641, '\P{^Is_Canonical_Combining_Class=ccc103}', "");
Expect(0, 3642, '\p{Is_Canonical_Combining_Class=ccc103}', "");
Expect(1, 3642, '\p{^Is_Canonical_Combining_Class=ccc103}', "");
Expect(1, 3642, '\P{Is_Canonical_Combining_Class=ccc103}', "");
Expect(0, 3642, '\P{^Is_Canonical_Combining_Class=ccc103}', "");
Expect(1, 3641, '\p{Is_Canonical_Combining_Class=-_ccc103}', "");
Expect(0, 3641, '\p{^Is_Canonical_Combining_Class=-_ccc103}', "");
Expect(0, 3641, '\P{Is_Canonical_Combining_Class=-_ccc103}', "");
Expect(1, 3641, '\P{^Is_Canonical_Combining_Class=-_ccc103}', "");
Expect(0, 3642, '\p{Is_Canonical_Combining_Class=-_ccc103}', "");
Expect(1, 3642, '\p{^Is_Canonical_Combining_Class=-_ccc103}', "");
Expect(1, 3642, '\P{Is_Canonical_Combining_Class=-_ccc103}', "");
Expect(0, 3642, '\P{^Is_Canonical_Combining_Class=-_ccc103}', "");
Error('\p{Is_Ccc=	-0_0_0_0_0_0_0_1_03/a/}');
Error('\P{Is_Ccc=	-0_0_0_0_0_0_0_1_03/a/}');
Expect(1, 3641, '\p{Is_Ccc=103}', "");
Expect(0, 3641, '\p{^Is_Ccc=103}', "");
Expect(0, 3641, '\P{Is_Ccc=103}', "");
Expect(1, 3641, '\P{^Is_Ccc=103}', "");
Expect(0, 3642, '\p{Is_Ccc=103}', "");
Expect(1, 3642, '\p{^Is_Ccc=103}', "");
Expect(1, 3642, '\P{Is_Ccc=103}', "");
Expect(0, 3642, '\P{^Is_Ccc=103}', "");
Error('\p{Canonical_Combining_Class=	/a/ccc107}');
Error('\P{Canonical_Combining_Class=	/a/ccc107}');
Expect(1, 3659, '\p{Canonical_Combining_Class=ccc107}', "");
Expect(0, 3659, '\p{^Canonical_Combining_Class=ccc107}', "");
Expect(0, 3659, '\P{Canonical_Combining_Class=ccc107}', "");
Expect(1, 3659, '\P{^Canonical_Combining_Class=ccc107}', "");
Expect(0, 3660, '\p{Canonical_Combining_Class=ccc107}', "");
Expect(1, 3660, '\p{^Canonical_Combining_Class=ccc107}', "");
Expect(1, 3660, '\P{Canonical_Combining_Class=ccc107}', "");
Expect(0, 3660, '\P{^Canonical_Combining_Class=ccc107}', "");
Expect(1, 3659, '\p{Canonical_Combining_Class=	-CCC107}', "");
Expect(0, 3659, '\p{^Canonical_Combining_Class=	-CCC107}', "");
Expect(0, 3659, '\P{Canonical_Combining_Class=	-CCC107}', "");
Expect(1, 3659, '\P{^Canonical_Combining_Class=	-CCC107}', "");
Expect(0, 3660, '\p{Canonical_Combining_Class=	-CCC107}', "");
Expect(1, 3660, '\p{^Canonical_Combining_Class=	-CCC107}', "");
Expect(1, 3660, '\P{Canonical_Combining_Class=	-CCC107}', "");
Expect(0, 3660, '\P{^Canonical_Combining_Class=	-CCC107}', "");
Error('\p{Ccc=	:=0_1_07}');
Error('\P{Ccc=	:=0_1_07}');
Expect(1, 3659, '\p{Ccc=+0_0_0_0_0_0_0_0_0_107}', "");
Expect(0, 3659, '\p{^Ccc=+0_0_0_0_0_0_0_0_0_107}', "");
Expect(0, 3659, '\P{Ccc=+0_0_0_0_0_0_0_0_0_107}', "");
Expect(1, 3659, '\P{^Ccc=+0_0_0_0_0_0_0_0_0_107}', "");
Expect(0, 3660, '\p{Ccc=+0_0_0_0_0_0_0_0_0_107}', "");
Expect(1, 3660, '\p{^Ccc=+0_0_0_0_0_0_0_0_0_107}', "");
Expect(1, 3660, '\P{Ccc=+0_0_0_0_0_0_0_0_0_107}', "");
Expect(0, 3660, '\P{^Ccc=+0_0_0_0_0_0_0_0_0_107}', "");
Error('\p{Is_Canonical_Combining_Class=-CCC107:=}');
Error('\P{Is_Canonical_Combining_Class=-CCC107:=}');
Expect(1, 3659, '\p{Is_Canonical_Combining_Class=ccc107}', "");
Expect(0, 3659, '\p{^Is_Canonical_Combining_Class=ccc107}', "");
Expect(0, 3659, '\P{Is_Canonical_Combining_Class=ccc107}', "");
Expect(1, 3659, '\P{^Is_Canonical_Combining_Class=ccc107}', "");
Expect(0, 3660, '\p{Is_Canonical_Combining_Class=ccc107}', "");
Expect(1, 3660, '\p{^Is_Canonical_Combining_Class=ccc107}', "");
Expect(1, 3660, '\P{Is_Canonical_Combining_Class=ccc107}', "");
Expect(0, 3660, '\P{^Is_Canonical_Combining_Class=ccc107}', "");
Expect(1, 3659, '\p{Is_Canonical_Combining_Class=_CCC107}', "");
Expect(0, 3659, '\p{^Is_Canonical_Combining_Class=_CCC107}', "");
Expect(0, 3659, '\P{Is_Canonical_Combining_Class=_CCC107}', "");
Expect(1, 3659, '\P{^Is_Canonical_Combining_Class=_CCC107}', "");
Expect(0, 3660, '\p{Is_Canonical_Combining_Class=_CCC107}', "");
Expect(1, 3660, '\p{^Is_Canonical_Combining_Class=_CCC107}', "");
Expect(1, 3660, '\P{Is_Canonical_Combining_Class=_CCC107}', "");
Expect(0, 3660, '\P{^Is_Canonical_Combining_Class=_CCC107}', "");
Error('\p{Is_Ccc= 	00_00_10_7:=}');
Error('\P{Is_Ccc= 	00_00_10_7:=}');
Expect(1, 3659, '\p{Is_Ccc:	+000000000107}', "");
Expect(0, 3659, '\p{^Is_Ccc:	+000000000107}', "");
Expect(0, 3659, '\P{Is_Ccc:	+000000000107}', "");
Expect(1, 3659, '\P{^Is_Ccc:	+000000000107}', "");
Expect(0, 3660, '\p{Is_Ccc:	+000000000107}', "");
Expect(1, 3660, '\p{^Is_Ccc:	+000000000107}', "");
Expect(1, 3660, '\P{Is_Ccc:	+000000000107}', "");
Expect(0, 3660, '\P{^Is_Ccc:	+000000000107}', "");
Error('\p{Canonical_Combining_Class=_ccc11/a/}');
Error('\P{Canonical_Combining_Class=_ccc11/a/}');
Expect(1, 1457, '\p{Canonical_Combining_Class=ccc11}', "");
Expect(0, 1457, '\p{^Canonical_Combining_Class=ccc11}', "");
Expect(0, 1457, '\P{Canonical_Combining_Class=ccc11}', "");
Expect(1, 1457, '\P{^Canonical_Combining_Class=ccc11}', "");
Expect(0, 1458, '\p{Canonical_Combining_Class=ccc11}', "");
Expect(1, 1458, '\p{^Canonical_Combining_Class=ccc11}', "");
Expect(1, 1458, '\P{Canonical_Combining_Class=ccc11}', "");
Expect(0, 1458, '\P{^Canonical_Combining_Class=ccc11}', "");
Expect(1, 1457, '\p{Canonical_Combining_Class=- ccc11}', "");
Expect(0, 1457, '\p{^Canonical_Combining_Class=- ccc11}', "");
Expect(0, 1457, '\P{Canonical_Combining_Class=- ccc11}', "");
Expect(1, 1457, '\P{^Canonical_Combining_Class=- ccc11}', "");
Expect(0, 1458, '\p{Canonical_Combining_Class=- ccc11}', "");
Expect(1, 1458, '\p{^Canonical_Combining_Class=- ccc11}', "");
Expect(1, 1458, '\P{Canonical_Combining_Class=- ccc11}', "");
Expect(0, 1458, '\P{^Canonical_Combining_Class=- ccc11}', "");
Error('\p{Ccc=00_01_1:=}');
Error('\P{Ccc=00_01_1:=}');
Expect(1, 1457, '\p{Ccc=000000011}', "");
Expect(0, 1457, '\p{^Ccc=000000011}', "");
Expect(0, 1457, '\P{Ccc=000000011}', "");
Expect(1, 1457, '\P{^Ccc=000000011}', "");
Expect(0, 1458, '\p{Ccc=000000011}', "");
Expect(1, 1458, '\p{^Ccc=000000011}', "");
Expect(1, 1458, '\P{Ccc=000000011}', "");
Expect(0, 1458, '\P{^Ccc=000000011}', "");
Error('\p{Is_Canonical_Combining_Class=/a/_ CCC11}');
Error('\P{Is_Canonical_Combining_Class=/a/_ CCC11}');
Expect(1, 1457, '\p{Is_Canonical_Combining_Class=ccc11}', "");
Expect(0, 1457, '\p{^Is_Canonical_Combining_Class=ccc11}', "");
Expect(0, 1457, '\P{Is_Canonical_Combining_Class=ccc11}', "");
Expect(1, 1457, '\P{^Is_Canonical_Combining_Class=ccc11}', "");
Expect(0, 1458, '\p{Is_Canonical_Combining_Class=ccc11}', "");
Expect(1, 1458, '\p{^Is_Canonical_Combining_Class=ccc11}', "");
Expect(1, 1458, '\P{Is_Canonical_Combining_Class=ccc11}', "");
Expect(0, 1458, '\P{^Is_Canonical_Combining_Class=ccc11}', "");
Expect(1, 1457, '\p{Is_Canonical_Combining_Class:		-CCC11}', "");
Expect(0, 1457, '\p{^Is_Canonical_Combining_Class:		-CCC11}', "");
Expect(0, 1457, '\P{Is_Canonical_Combining_Class:		-CCC11}', "");
Expect(1, 1457, '\P{^Is_Canonical_Combining_Class:		-CCC11}', "");
Expect(0, 1458, '\p{Is_Canonical_Combining_Class:		-CCC11}', "");
Expect(1, 1458, '\p{^Is_Canonical_Combining_Class:		-CCC11}', "");
Expect(1, 1458, '\P{Is_Canonical_Combining_Class:		-CCC11}', "");
Expect(0, 1458, '\P{^Is_Canonical_Combining_Class:		-CCC11}', "");
Error('\p{Is_Ccc= -00001_1:=}');
Error('\P{Is_Ccc= -00001_1:=}');
Expect(1, 1457, '\p{Is_Ccc=+000001_1}', "");
Expect(0, 1457, '\p{^Is_Ccc=+000001_1}', "");
Expect(0, 1457, '\P{Is_Ccc=+000001_1}', "");
Expect(1, 1457, '\P{^Is_Ccc=+000001_1}', "");
Expect(0, 1458, '\p{Is_Ccc=+000001_1}', "");
Expect(1, 1458, '\p{^Is_Ccc=+000001_1}', "");
Expect(1, 1458, '\P{Is_Ccc=+000001_1}', "");
Expect(0, 1458, '\P{^Is_Ccc=+000001_1}', "");
Error('\p{Canonical_Combining_Class=-	CCC118:=}');
Error('\P{Canonical_Combining_Class=-	CCC118:=}');
Expect(1, 3769, '\p{Canonical_Combining_Class=ccc118}', "");
Expect(0, 3769, '\p{^Canonical_Combining_Class=ccc118}', "");
Expect(0, 3769, '\P{Canonical_Combining_Class=ccc118}', "");
Expect(1, 3769, '\P{^Canonical_Combining_Class=ccc118}', "");
Expect(0, 3770, '\p{Canonical_Combining_Class=ccc118}', "");
Expect(1, 3770, '\p{^Canonical_Combining_Class=ccc118}', "");
Expect(1, 3770, '\P{Canonical_Combining_Class=ccc118}', "");
Expect(0, 3770, '\P{^Canonical_Combining_Class=ccc118}', "");
Expect(1, 3769, '\p{Canonical_Combining_Class=	_CCC118}', "");
Expect(0, 3769, '\p{^Canonical_Combining_Class=	_CCC118}', "");
Expect(0, 3769, '\P{Canonical_Combining_Class=	_CCC118}', "");
Expect(1, 3769, '\P{^Canonical_Combining_Class=	_CCC118}', "");
Expect(0, 3770, '\p{Canonical_Combining_Class=	_CCC118}', "");
Expect(1, 3770, '\p{^Canonical_Combining_Class=	_CCC118}', "");
Expect(1, 3770, '\P{Canonical_Combining_Class=	_CCC118}', "");
Expect(0, 3770, '\P{^Canonical_Combining_Class=	_CCC118}', "");
Error('\p{Ccc=- 00000000118/a/}');
Error('\P{Ccc=- 00000000118/a/}');
Expect(1, 3769, '\p{Ccc: 00000011_8}', "");
Expect(0, 3769, '\p{^Ccc: 00000011_8}', "");
Expect(0, 3769, '\P{Ccc: 00000011_8}', "");
Expect(1, 3769, '\P{^Ccc: 00000011_8}', "");
Expect(0, 3770, '\p{Ccc: 00000011_8}', "");
Expect(1, 3770, '\p{^Ccc: 00000011_8}', "");
Expect(1, 3770, '\P{Ccc: 00000011_8}', "");
Expect(0, 3770, '\P{^Ccc: 00000011_8}', "");
Error('\p{Is_Canonical_Combining_Class=:=- CCC118}');
Error('\P{Is_Canonical_Combining_Class=:=- CCC118}');
Expect(1, 3769, '\p{Is_Canonical_Combining_Class=ccc118}', "");
Expect(0, 3769, '\p{^Is_Canonical_Combining_Class=ccc118}', "");
Expect(0, 3769, '\P{Is_Canonical_Combining_Class=ccc118}', "");
Expect(1, 3769, '\P{^Is_Canonical_Combining_Class=ccc118}', "");
Expect(0, 3770, '\p{Is_Canonical_Combining_Class=ccc118}', "");
Expect(1, 3770, '\p{^Is_Canonical_Combining_Class=ccc118}', "");
Expect(1, 3770, '\P{Is_Canonical_Combining_Class=ccc118}', "");
Expect(0, 3770, '\P{^Is_Canonical_Combining_Class=ccc118}', "");
Expect(1, 3769, '\p{Is_Canonical_Combining_Class=	-ccc118}', "");
Expect(0, 3769, '\p{^Is_Canonical_Combining_Class=	-ccc118}', "");
Expect(0, 3769, '\P{Is_Canonical_Combining_Class=	-ccc118}', "");
Expect(1, 3769, '\P{^Is_Canonical_Combining_Class=	-ccc118}', "");
Expect(0, 3770, '\p{Is_Canonical_Combining_Class=	-ccc118}', "");
Expect(1, 3770, '\p{^Is_Canonical_Combining_Class=	-ccc118}', "");
Expect(1, 3770, '\P{Is_Canonical_Combining_Class=	-ccc118}', "");
Expect(0, 3770, '\P{^Is_Canonical_Combining_Class=	-ccc118}', "");
Error('\p{Is_Ccc:_:=00_00_00_11_8}');
Error('\P{Is_Ccc:_:=00_00_00_11_8}');
Expect(1, 3769, '\p{Is_Ccc=000000118}', "");
Expect(0, 3769, '\p{^Is_Ccc=000000118}', "");
Expect(0, 3769, '\P{Is_Ccc=000000118}', "");
Expect(1, 3769, '\P{^Is_Ccc=000000118}', "");
Expect(0, 3770, '\p{Is_Ccc=000000118}', "");
Expect(1, 3770, '\p{^Is_Ccc=000000118}', "");
Expect(1, 3770, '\P{Is_Ccc=000000118}', "");
Expect(0, 3770, '\P{^Is_Ccc=000000118}', "");
Error('\p{Canonical_Combining_Class=/a/--CCC12}');
Error('\P{Canonical_Combining_Class=/a/--CCC12}');
Expect(1, 1458, '\p{Canonical_Combining_Class=ccc12}', "");
Expect(0, 1458, '\p{^Canonical_Combining_Class=ccc12}', "");
Expect(0, 1458, '\P{Canonical_Combining_Class=ccc12}', "");
Expect(1, 1458, '\P{^Canonical_Combining_Class=ccc12}', "");
Expect(0, 1459, '\p{Canonical_Combining_Class=ccc12}', "");
Expect(1, 1459, '\p{^Canonical_Combining_Class=ccc12}', "");
Expect(1, 1459, '\P{Canonical_Combining_Class=ccc12}', "");
Expect(0, 1459, '\P{^Canonical_Combining_Class=ccc12}', "");
Expect(1, 1458, '\p{Canonical_Combining_Class=-CCC12}', "");
Expect(0, 1458, '\p{^Canonical_Combining_Class=-CCC12}', "");
Expect(0, 1458, '\P{Canonical_Combining_Class=-CCC12}', "");
Expect(1, 1458, '\P{^Canonical_Combining_Class=-CCC12}', "");
Expect(0, 1459, '\p{Canonical_Combining_Class=-CCC12}', "");
Expect(1, 1459, '\p{^Canonical_Combining_Class=-CCC12}', "");
Expect(1, 1459, '\P{Canonical_Combining_Class=-CCC12}', "");
Expect(0, 1459, '\P{^Canonical_Combining_Class=-CCC12}', "");
Error('\p{Ccc=:=0000_0001_2}');
Error('\P{Ccc=:=0000_0001_2}');
Expect(1, 1458, '\p{Ccc:	01_2}', "");
Expect(0, 1458, '\p{^Ccc:	01_2}', "");
Expect(0, 1458, '\P{Ccc:	01_2}', "");
Expect(1, 1458, '\P{^Ccc:	01_2}', "");
Expect(0, 1459, '\p{Ccc:	01_2}', "");
Expect(1, 1459, '\p{^Ccc:	01_2}', "");
Expect(1, 1459, '\P{Ccc:	01_2}', "");
Expect(0, 1459, '\P{^Ccc:	01_2}', "");
Error('\p{Is_Canonical_Combining_Class=	 CCC12/a/}');
Error('\P{Is_Canonical_Combining_Class=	 CCC12/a/}');
Expect(1, 1458, '\p{Is_Canonical_Combining_Class=ccc12}', "");
Expect(0, 1458, '\p{^Is_Canonical_Combining_Class=ccc12}', "");
Expect(0, 1458, '\P{Is_Canonical_Combining_Class=ccc12}', "");
Expect(1, 1458, '\P{^Is_Canonical_Combining_Class=ccc12}', "");
Expect(0, 1459, '\p{Is_Canonical_Combining_Class=ccc12}', "");
Expect(1, 1459, '\p{^Is_Canonical_Combining_Class=ccc12}', "");
Expect(1, 1459, '\P{Is_Canonical_Combining_Class=ccc12}', "");
Expect(0, 1459, '\P{^Is_Canonical_Combining_Class=ccc12}', "");
Expect(1, 1458, '\p{Is_Canonical_Combining_Class=_CCC12}', "");
Expect(0, 1458, '\p{^Is_Canonical_Combining_Class=_CCC12}', "");
Expect(0, 1458, '\P{Is_Canonical_Combining_Class=_CCC12}', "");
Expect(1, 1458, '\P{^Is_Canonical_Combining_Class=_CCC12}', "");
Expect(0, 1459, '\p{Is_Canonical_Combining_Class=_CCC12}', "");
Expect(1, 1459, '\p{^Is_Canonical_Combining_Class=_CCC12}', "");
Expect(1, 1459, '\P{Is_Canonical_Combining_Class=_CCC12}', "");
Expect(0, 1459, '\P{^Is_Canonical_Combining_Class=_CCC12}', "");
Error('\p{Is_Ccc=/a/ _+00_00_00_012}');
Error('\P{Is_Ccc=/a/ _+00_00_00_012}');
Expect(1, 1458, '\p{Is_Ccc=01_2}', "");
Expect(0, 1458, '\p{^Is_Ccc=01_2}', "");
Expect(0, 1458, '\P{Is_Ccc=01_2}', "");
Expect(1, 1458, '\P{^Is_Ccc=01_2}', "");
Expect(0, 1459, '\p{Is_Ccc=01_2}', "");
Expect(1, 1459, '\p{^Is_Ccc=01_2}', "");
Expect(1, 1459, '\P{Is_Ccc=01_2}', "");
Expect(0, 1459, '\P{^Is_Ccc=01_2}', "");
Error('\p{Canonical_Combining_Class=/a/-	ccc122}');
Error('\P{Canonical_Combining_Class=/a/-	ccc122}');
Expect(1, 3787, '\p{Canonical_Combining_Class=ccc122}', "");
Expect(0, 3787, '\p{^Canonical_Combining_Class=ccc122}', "");
Expect(0, 3787, '\P{Canonical_Combining_Class=ccc122}', "");
Expect(1, 3787, '\P{^Canonical_Combining_Class=ccc122}', "");
Expect(0, 3788, '\p{Canonical_Combining_Class=ccc122}', "");
Expect(1, 3788, '\p{^Canonical_Combining_Class=ccc122}', "");
Expect(1, 3788, '\P{Canonical_Combining_Class=ccc122}', "");
Expect(0, 3788, '\P{^Canonical_Combining_Class=ccc122}', "");
Expect(1, 3787, '\p{Canonical_Combining_Class:	-ccc122}', "");
Expect(0, 3787, '\p{^Canonical_Combining_Class:	-ccc122}', "");
Expect(0, 3787, '\P{Canonical_Combining_Class:	-ccc122}', "");
Expect(1, 3787, '\P{^Canonical_Combining_Class:	-ccc122}', "");
Expect(0, 3788, '\p{Canonical_Combining_Class:	-ccc122}', "");
Expect(1, 3788, '\p{^Canonical_Combining_Class:	-ccc122}', "");
Expect(1, 3788, '\P{Canonical_Combining_Class:	-ccc122}', "");
Expect(0, 3788, '\P{^Canonical_Combining_Class:	-ccc122}', "");
Error('\p{Ccc=-+0_0_0_0_1_22:=}');
Error('\P{Ccc=-+0_0_0_0_1_22:=}');
Expect(1, 3787, '\p{Ccc=0_0_0_122}', "");
Expect(0, 3787, '\p{^Ccc=0_0_0_122}', "");
Expect(0, 3787, '\P{Ccc=0_0_0_122}', "");
Expect(1, 3787, '\P{^Ccc=0_0_0_122}', "");
Expect(0, 3788, '\p{Ccc=0_0_0_122}', "");
Expect(1, 3788, '\p{^Ccc=0_0_0_122}', "");
Expect(1, 3788, '\P{Ccc=0_0_0_122}', "");
Expect(0, 3788, '\P{^Ccc=0_0_0_122}', "");
Error('\p{Is_Canonical_Combining_Class=:=	_CCC122}');
Error('\P{Is_Canonical_Combining_Class=:=	_CCC122}');
Expect(1, 3787, '\p{Is_Canonical_Combining_Class=ccc122}', "");
Expect(0, 3787, '\p{^Is_Canonical_Combining_Class=ccc122}', "");
Expect(0, 3787, '\P{Is_Canonical_Combining_Class=ccc122}', "");
Expect(1, 3787, '\P{^Is_Canonical_Combining_Class=ccc122}', "");
Expect(0, 3788, '\p{Is_Canonical_Combining_Class=ccc122}', "");
Expect(1, 3788, '\p{^Is_Canonical_Combining_Class=ccc122}', "");
Expect(1, 3788, '\P{Is_Canonical_Combining_Class=ccc122}', "");
Expect(0, 3788, '\P{^Is_Canonical_Combining_Class=ccc122}', "");
Expect(1, 3787, '\p{Is_Canonical_Combining_Class=_CCC122}', "");
Expect(0, 3787, '\p{^Is_Canonical_Combining_Class=_CCC122}', "");
Expect(0, 3787, '\P{Is_Canonical_Combining_Class=_CCC122}', "");
Expect(1, 3787, '\P{^Is_Canonical_Combining_Class=_CCC122}', "");
Expect(0, 3788, '\p{Is_Canonical_Combining_Class=_CCC122}', "");
Expect(1, 3788, '\p{^Is_Canonical_Combining_Class=_CCC122}', "");
Expect(1, 3788, '\P{Is_Canonical_Combining_Class=_CCC122}', "");
Expect(0, 3788, '\P{^Is_Canonical_Combining_Class=_CCC122}', "");
Error('\p{Is_Ccc=-:=00000000122}');
Error('\P{Is_Ccc=-:=00000000122}');
Expect(1, 3787, '\p{Is_Ccc=+122}', "");
Expect(0, 3787, '\p{^Is_Ccc=+122}', "");
Expect(0, 3787, '\P{Is_Ccc=+122}', "");
Expect(1, 3787, '\P{^Is_Ccc=+122}', "");
Expect(0, 3788, '\p{Is_Ccc=+122}', "");
Expect(1, 3788, '\p{^Is_Ccc=+122}', "");
Expect(1, 3788, '\P{Is_Ccc=+122}', "");
Expect(0, 3788, '\P{^Is_Ccc=+122}', "");
Error('\p{Canonical_Combining_Class:	:=CCC129}');
Error('\P{Canonical_Combining_Class:	:=CCC129}');
Expect(1, 3953, '\p{Canonical_Combining_Class:ccc129}', "");
Expect(0, 3953, '\p{^Canonical_Combining_Class:ccc129}', "");
Expect(0, 3953, '\P{Canonical_Combining_Class:ccc129}', "");
Expect(1, 3953, '\P{^Canonical_Combining_Class:ccc129}', "");
Expect(0, 3954, '\p{Canonical_Combining_Class:ccc129}', "");
Expect(1, 3954, '\p{^Canonical_Combining_Class:ccc129}', "");
Expect(1, 3954, '\P{Canonical_Combining_Class:ccc129}', "");
Expect(0, 3954, '\P{^Canonical_Combining_Class:ccc129}', "");
Expect(1, 3953, '\p{Canonical_Combining_Class=	-CCC129}', "");
Expect(0, 3953, '\p{^Canonical_Combining_Class=	-CCC129}', "");
Expect(0, 3953, '\P{Canonical_Combining_Class=	-CCC129}', "");
Expect(1, 3953, '\P{^Canonical_Combining_Class=	-CCC129}', "");
Expect(0, 3954, '\p{Canonical_Combining_Class=	-CCC129}', "");
Expect(1, 3954, '\p{^Canonical_Combining_Class=	-CCC129}', "");
Expect(1, 3954, '\P{Canonical_Combining_Class=	-CCC129}', "");
Expect(0, 3954, '\P{^Canonical_Combining_Class=	-CCC129}', "");
Error('\p{Ccc=_	00012_9:=}');
Error('\P{Ccc=_	00012_9:=}');
Expect(1, 3953, '\p{Ccc=00_00_00_129}', "");
Expect(0, 3953, '\p{^Ccc=00_00_00_129}', "");
Expect(0, 3953, '\P{Ccc=00_00_00_129}', "");
Expect(1, 3953, '\P{^Ccc=00_00_00_129}', "");
Expect(0, 3954, '\p{Ccc=00_00_00_129}', "");
Expect(1, 3954, '\p{^Ccc=00_00_00_129}', "");
Expect(1, 3954, '\P{Ccc=00_00_00_129}', "");
Expect(0, 3954, '\P{^Ccc=00_00_00_129}', "");
Error('\p{Is_Canonical_Combining_Class=-:=CCC129}');
Error('\P{Is_Canonical_Combining_Class=-:=CCC129}');
Expect(1, 3953, '\p{Is_Canonical_Combining_Class=ccc129}', "");
Expect(0, 3953, '\p{^Is_Canonical_Combining_Class=ccc129}', "");
Expect(0, 3953, '\P{Is_Canonical_Combining_Class=ccc129}', "");
Expect(1, 3953, '\P{^Is_Canonical_Combining_Class=ccc129}', "");
Expect(0, 3954, '\p{Is_Canonical_Combining_Class=ccc129}', "");
Expect(1, 3954, '\p{^Is_Canonical_Combining_Class=ccc129}', "");
Expect(1, 3954, '\P{Is_Canonical_Combining_Class=ccc129}', "");
Expect(0, 3954, '\P{^Is_Canonical_Combining_Class=ccc129}', "");
Expect(1, 3953, '\p{Is_Canonical_Combining_Class=- CCC129}', "");
Expect(0, 3953, '\p{^Is_Canonical_Combining_Class=- CCC129}', "");
Expect(0, 3953, '\P{Is_Canonical_Combining_Class=- CCC129}', "");
Expect(1, 3953, '\P{^Is_Canonical_Combining_Class=- CCC129}', "");
Expect(0, 3954, '\p{Is_Canonical_Combining_Class=- CCC129}', "");
Expect(1, 3954, '\p{^Is_Canonical_Combining_Class=- CCC129}', "");
Expect(1, 3954, '\P{Is_Canonical_Combining_Class=- CCC129}', "");
Expect(0, 3954, '\P{^Is_Canonical_Combining_Class=- CCC129}', "");
Error('\p{Is_Ccc=	/a/0129}');
Error('\P{Is_Ccc=	/a/0129}');
Expect(1, 3953, '\p{Is_Ccc=+0_0_1_29}', "");
Expect(0, 3953, '\p{^Is_Ccc=+0_0_1_29}', "");
Expect(0, 3953, '\P{Is_Ccc=+0_0_1_29}', "");
Expect(1, 3953, '\P{^Is_Ccc=+0_0_1_29}', "");
Expect(0, 3954, '\p{Is_Ccc=+0_0_1_29}', "");
Expect(1, 3954, '\p{^Is_Ccc=+0_0_1_29}', "");
Expect(1, 3954, '\P{Is_Ccc=+0_0_1_29}', "");
Expect(0, 3954, '\P{^Is_Ccc=+0_0_1_29}', "");
Error('\p{Canonical_Combining_Class=CCC13/a/}');
Error('\P{Canonical_Combining_Class=CCC13/a/}');
Expect(1, 1459, '\p{Canonical_Combining_Class=ccc13}', "");
Expect(0, 1459, '\p{^Canonical_Combining_Class=ccc13}', "");
Expect(0, 1459, '\P{Canonical_Combining_Class=ccc13}', "");
Expect(1, 1459, '\P{^Canonical_Combining_Class=ccc13}', "");
Expect(0, 1460, '\p{Canonical_Combining_Class=ccc13}', "");
Expect(1, 1460, '\p{^Canonical_Combining_Class=ccc13}', "");
Expect(1, 1460, '\P{Canonical_Combining_Class=ccc13}', "");
Expect(0, 1460, '\P{^Canonical_Combining_Class=ccc13}', "");
Expect(1, 1459, '\p{Canonical_Combining_Class=_	CCC13}', "");
Expect(0, 1459, '\p{^Canonical_Combining_Class=_	CCC13}', "");
Expect(0, 1459, '\P{Canonical_Combining_Class=_	CCC13}', "");
Expect(1, 1459, '\P{^Canonical_Combining_Class=_	CCC13}', "");
Expect(0, 1460, '\p{Canonical_Combining_Class=_	CCC13}', "");
Expect(1, 1460, '\p{^Canonical_Combining_Class=_	CCC13}', "");
Expect(1, 1460, '\P{Canonical_Combining_Class=_	CCC13}', "");
Expect(0, 1460, '\P{^Canonical_Combining_Class=_	CCC13}', "");
Error('\p{Ccc= 	+000000013:=}');
Error('\P{Ccc= 	+000000013:=}');
Expect(1, 1459, '\p{Ccc=000013}', "");
Expect(0, 1459, '\p{^Ccc=000013}', "");
Expect(0, 1459, '\P{Ccc=000013}', "");
Expect(1, 1459, '\P{^Ccc=000013}', "");
Expect(0, 1460, '\p{Ccc=000013}', "");
Expect(1, 1460, '\p{^Ccc=000013}', "");
Expect(1, 1460, '\P{Ccc=000013}', "");
Expect(0, 1460, '\P{^Ccc=000013}', "");
Error('\p{Is_Canonical_Combining_Class:	 CCC13:=}');
Error('\P{Is_Canonical_Combining_Class:	 CCC13:=}');
Expect(1, 1459, '\p{Is_Canonical_Combining_Class: ccc13}', "");
Expect(0, 1459, '\p{^Is_Canonical_Combining_Class: ccc13}', "");
Expect(0, 1459, '\P{Is_Canonical_Combining_Class: ccc13}', "");
Expect(1, 1459, '\P{^Is_Canonical_Combining_Class: ccc13}', "");
Expect(0, 1460, '\p{Is_Canonical_Combining_Class: ccc13}', "");
Expect(1, 1460, '\p{^Is_Canonical_Combining_Class: ccc13}', "");
Expect(1, 1460, '\P{Is_Canonical_Combining_Class: ccc13}', "");
Expect(0, 1460, '\P{^Is_Canonical_Combining_Class: ccc13}', "");
Expect(1, 1459, '\p{Is_Canonical_Combining_Class=-CCC13}', "");
Expect(0, 1459, '\p{^Is_Canonical_Combining_Class=-CCC13}', "");
Expect(0, 1459, '\P{Is_Canonical_Combining_Class=-CCC13}', "");
Expect(1, 1459, '\P{^Is_Canonical_Combining_Class=-CCC13}', "");
Expect(0, 1460, '\p{Is_Canonical_Combining_Class=-CCC13}', "");
Expect(1, 1460, '\p{^Is_Canonical_Combining_Class=-CCC13}', "");
Expect(1, 1460, '\P{Is_Canonical_Combining_Class=-CCC13}', "");
Expect(0, 1460, '\P{^Is_Canonical_Combining_Class=-CCC13}', "");
Error('\p{Is_Ccc:	/a/0_0_0_0_0_13}');
Error('\P{Is_Ccc:	/a/0_0_0_0_0_13}');
Expect(1, 1459, '\p{Is_Ccc:   00000000013}', "");
Expect(0, 1459, '\p{^Is_Ccc:   00000000013}', "");
Expect(0, 1459, '\P{Is_Ccc:   00000000013}', "");
Expect(1, 1459, '\P{^Is_Ccc:   00000000013}', "");
Expect(0, 1460, '\p{Is_Ccc:   00000000013}', "");
Expect(1, 1460, '\p{^Is_Ccc:   00000000013}', "");
Expect(1, 1460, '\P{Is_Ccc:   00000000013}', "");
Expect(0, 1460, '\P{^Is_Ccc:   00000000013}', "");
Error('\p{Canonical_Combining_Class=/a/_-CCC130}');
Error('\P{Canonical_Combining_Class=/a/_-CCC130}');
Expect(1, 3968, '\p{Canonical_Combining_Class=ccc130}', "");
Expect(0, 3968, '\p{^Canonical_Combining_Class=ccc130}', "");
Expect(0, 3968, '\P{Canonical_Combining_Class=ccc130}', "");
Expect(1, 3968, '\P{^Canonical_Combining_Class=ccc130}', "");
Expect(0, 3969, '\p{Canonical_Combining_Class=ccc130}', "");
Expect(1, 3969, '\p{^Canonical_Combining_Class=ccc130}', "");
Expect(1, 3969, '\P{Canonical_Combining_Class=ccc130}', "");
Expect(0, 3969, '\P{^Canonical_Combining_Class=ccc130}', "");
Expect(1, 3968, '\p{Canonical_Combining_Class= -CCC130}', "");
Expect(0, 3968, '\p{^Canonical_Combining_Class= -CCC130}', "");
Expect(0, 3968, '\P{Canonical_Combining_Class= -CCC130}', "");
Expect(1, 3968, '\P{^Canonical_Combining_Class= -CCC130}', "");
Expect(0, 3969, '\p{Canonical_Combining_Class= -CCC130}', "");
Expect(1, 3969, '\p{^Canonical_Combining_Class= -CCC130}', "");
Expect(1, 3969, '\P{Canonical_Combining_Class= -CCC130}', "");
Expect(0, 3969, '\P{^Canonical_Combining_Class= -CCC130}', "");
Error('\p{Ccc=	:=130}');
Error('\P{Ccc=	:=130}');
Expect(1, 3968, '\p{Ccc=+0_0_0_1_30}', "");
Expect(0, 3968, '\p{^Ccc=+0_0_0_1_30}', "");
Expect(0, 3968, '\P{Ccc=+0_0_0_1_30}', "");
Expect(1, 3968, '\P{^Ccc=+0_0_0_1_30}', "");
Expect(0, 3969, '\p{Ccc=+0_0_0_1_30}', "");
Expect(1, 3969, '\p{^Ccc=+0_0_0_1_30}', "");
Expect(1, 3969, '\P{Ccc=+0_0_0_1_30}', "");
Expect(0, 3969, '\P{^Ccc=+0_0_0_1_30}', "");
Error('\p{Is_Canonical_Combining_Class= :=ccc130}');
Error('\P{Is_Canonical_Combining_Class= :=ccc130}');
Expect(1, 3968, '\p{Is_Canonical_Combining_Class=ccc130}', "");
Expect(0, 3968, '\p{^Is_Canonical_Combining_Class=ccc130}', "");
Expect(0, 3968, '\P{Is_Canonical_Combining_Class=ccc130}', "");
Expect(1, 3968, '\P{^Is_Canonical_Combining_Class=ccc130}', "");
Expect(0, 3969, '\p{Is_Canonical_Combining_Class=ccc130}', "");
Expect(1, 3969, '\p{^Is_Canonical_Combining_Class=ccc130}', "");
Expect(1, 3969, '\P{Is_Canonical_Combining_Class=ccc130}', "");
Expect(0, 3969, '\P{^Is_Canonical_Combining_Class=ccc130}', "");
Expect(1, 3968, '\p{Is_Canonical_Combining_Class=	CCC130}', "");
Expect(0, 3968, '\p{^Is_Canonical_Combining_Class=	CCC130}', "");
Expect(0, 3968, '\P{Is_Canonical_Combining_Class=	CCC130}', "");
Expect(1, 3968, '\P{^Is_Canonical_Combining_Class=	CCC130}', "");
Expect(0, 3969, '\p{Is_Canonical_Combining_Class=	CCC130}', "");
Expect(1, 3969, '\p{^Is_Canonical_Combining_Class=	CCC130}', "");
Expect(1, 3969, '\P{Is_Canonical_Combining_Class=	CCC130}', "");
Expect(0, 3969, '\P{^Is_Canonical_Combining_Class=	CCC130}', "");
Error('\p{Is_Ccc: --000000000130:=}');
Error('\P{Is_Ccc: --000000000130:=}');
Expect(1, 3968, '\p{Is_Ccc=130}', "");
Expect(0, 3968, '\p{^Is_Ccc=130}', "");
Expect(0, 3968, '\P{Is_Ccc=130}', "");
Expect(1, 3968, '\P{^Is_Ccc=130}', "");
Expect(0, 3969, '\p{Is_Ccc=130}', "");
Expect(1, 3969, '\p{^Is_Ccc=130}', "");
Expect(1, 3969, '\P{Is_Ccc=130}', "");
Expect(0, 3969, '\P{^Is_Ccc=130}', "");
Error('\p{Canonical_Combining_Class=:=--CCC132}');
Error('\P{Canonical_Combining_Class=:=--CCC132}');
Expect(1, 3956, '\p{Canonical_Combining_Class=ccc132}', "");
Expect(0, 3956, '\p{^Canonical_Combining_Class=ccc132}', "");
Expect(0, 3956, '\P{Canonical_Combining_Class=ccc132}', "");
Expect(1, 3956, '\P{^Canonical_Combining_Class=ccc132}', "");
Expect(0, 3957, '\p{Canonical_Combining_Class=ccc132}', "");
Expect(1, 3957, '\p{^Canonical_Combining_Class=ccc132}', "");
Expect(1, 3957, '\P{Canonical_Combining_Class=ccc132}', "");
Expect(0, 3957, '\P{^Canonical_Combining_Class=ccc132}', "");
Expect(1, 3956, '\p{Canonical_Combining_Class=_ CCC132}', "");
Expect(0, 3956, '\p{^Canonical_Combining_Class=_ CCC132}', "");
Expect(0, 3956, '\P{Canonical_Combining_Class=_ CCC132}', "");
Expect(1, 3956, '\P{^Canonical_Combining_Class=_ CCC132}', "");
Expect(0, 3957, '\p{Canonical_Combining_Class=_ CCC132}', "");
Expect(1, 3957, '\p{^Canonical_Combining_Class=_ CCC132}', "");
Expect(1, 3957, '\P{Canonical_Combining_Class=_ CCC132}', "");
Expect(0, 3957, '\P{^Canonical_Combining_Class=_ CCC132}', "");
Error('\p{Ccc= +0132:=}');
Error('\P{Ccc= +0132:=}');
Expect(1, 3956, '\p{Ccc=00_00_13_2}', "");
Expect(0, 3956, '\p{^Ccc=00_00_13_2}', "");
Expect(0, 3956, '\P{Ccc=00_00_13_2}', "");
Expect(1, 3956, '\P{^Ccc=00_00_13_2}', "");
Expect(0, 3957, '\p{Ccc=00_00_13_2}', "");
Expect(1, 3957, '\p{^Ccc=00_00_13_2}', "");
Expect(1, 3957, '\P{Ccc=00_00_13_2}', "");
Expect(0, 3957, '\P{^Ccc=00_00_13_2}', "");
Error('\p{Is_Canonical_Combining_Class=:= CCC132}');
Error('\P{Is_Canonical_Combining_Class=:= CCC132}');
Expect(1, 3956, '\p{Is_Canonical_Combining_Class:	ccc132}', "");
Expect(0, 3956, '\p{^Is_Canonical_Combining_Class:	ccc132}', "");
Expect(0, 3956, '\P{Is_Canonical_Combining_Class:	ccc132}', "");
Expect(1, 3956, '\P{^Is_Canonical_Combining_Class:	ccc132}', "");
Expect(0, 3957, '\p{Is_Canonical_Combining_Class:	ccc132}', "");
Expect(1, 3957, '\p{^Is_Canonical_Combining_Class:	ccc132}', "");
Expect(1, 3957, '\P{Is_Canonical_Combining_Class:	ccc132}', "");
Expect(0, 3957, '\P{^Is_Canonical_Combining_Class:	ccc132}', "");
Expect(1, 3956, '\p{Is_Canonical_Combining_Class:	_CCC132}', "");
Expect(0, 3956, '\p{^Is_Canonical_Combining_Class:	_CCC132}', "");
Expect(0, 3956, '\P{Is_Canonical_Combining_Class:	_CCC132}', "");
Expect(1, 3956, '\P{^Is_Canonical_Combining_Class:	_CCC132}', "");
Expect(0, 3957, '\p{Is_Canonical_Combining_Class:	_CCC132}', "");
Expect(1, 3957, '\p{^Is_Canonical_Combining_Class:	_CCC132}', "");
Expect(1, 3957, '\P{Is_Canonical_Combining_Class:	_CCC132}', "");
Expect(0, 3957, '\P{^Is_Canonical_Combining_Class:	_CCC132}', "");
Error('\p{Is_Ccc=	/a/0_0_0_0_0_1_32}');
Error('\P{Is_Ccc=	/a/0_0_0_0_0_1_32}');
Expect(1, 3956, '\p{Is_Ccc=13_2}', "");
Expect(0, 3956, '\p{^Is_Ccc=13_2}', "");
Expect(0, 3956, '\P{Is_Ccc=13_2}', "");
Expect(1, 3956, '\P{^Is_Ccc=13_2}', "");
Expect(0, 3957, '\p{Is_Ccc=13_2}', "");
Expect(1, 3957, '\p{^Is_Ccc=13_2}', "");
Expect(1, 3957, '\P{Is_Ccc=13_2}', "");
Expect(0, 3957, '\P{^Is_Ccc=13_2}', "");
Error('\p{Canonical_Combining_Class=	_CCC133:=}');
Error('\P{Canonical_Combining_Class=	_CCC133:=}');
Expect(0, 1, '\p{Canonical_Combining_Class=ccc133}', "");
Expect(1, 1, '\p{^Canonical_Combining_Class=ccc133}', "");
Expect(1, 1, '\P{Canonical_Combining_Class=ccc133}', "");
Expect(0, 1, '\P{^Canonical_Combining_Class=ccc133}', "");
Expect(0, 1, '\p{Canonical_Combining_Class=__CCC133}', "");
Expect(1, 1, '\p{^Canonical_Combining_Class=__CCC133}', "");
Expect(1, 1, '\P{Canonical_Combining_Class=__CCC133}', "");
Expect(0, 1, '\P{^Canonical_Combining_Class=__CCC133}', "");
Error('\p{Ccc=- +00_00_01_33:=}');
Error('\P{Ccc=- +00_00_01_33:=}');
Expect(0, 1, '\p{Ccc=00_00_00_00_133}', "");
Expect(1, 1, '\p{^Ccc=00_00_00_00_133}', "");
Expect(1, 1, '\P{Ccc=00_00_00_00_133}', "");
Expect(0, 1, '\P{^Ccc=00_00_00_00_133}', "");
Error('\p{Is_Canonical_Combining_Class:  ccc133:=}');
Error('\P{Is_Canonical_Combining_Class:  ccc133:=}');
Expect(0, 1, '\p{Is_Canonical_Combining_Class=ccc133}', "");
Expect(1, 1, '\p{^Is_Canonical_Combining_Class=ccc133}', "");
Expect(1, 1, '\P{Is_Canonical_Combining_Class=ccc133}', "");
Expect(0, 1, '\P{^Is_Canonical_Combining_Class=ccc133}', "");
Expect(0, 1, '\p{Is_Canonical_Combining_Class=-CCC133}', "");
Expect(1, 1, '\p{^Is_Canonical_Combining_Class=-CCC133}', "");
Expect(1, 1, '\P{Is_Canonical_Combining_Class=-CCC133}', "");
Expect(0, 1, '\P{^Is_Canonical_Combining_Class=-CCC133}', "");
Error('\p{Is_Ccc=:=_000000133}');
Error('\P{Is_Ccc=:=_000000133}');
Expect(0, 1, '\p{Is_Ccc=+000133}', "");
Expect(1, 1, '\p{^Is_Ccc=+000133}', "");
Expect(1, 1, '\P{Is_Ccc=+000133}', "");
Expect(0, 1, '\P{^Is_Ccc=+000133}', "");
Error('\p{Canonical_Combining_Class=:=__CCC14}');
Error('\P{Canonical_Combining_Class=:=__CCC14}');
Expect(1, 1460, '\p{Canonical_Combining_Class=ccc14}', "");
Expect(0, 1460, '\p{^Canonical_Combining_Class=ccc14}', "");
Expect(0, 1460, '\P{Canonical_Combining_Class=ccc14}', "");
Expect(1, 1460, '\P{^Canonical_Combining_Class=ccc14}', "");
Expect(0, 1461, '\p{Canonical_Combining_Class=ccc14}', "");
Expect(1, 1461, '\p{^Canonical_Combining_Class=ccc14}', "");
Expect(1, 1461, '\P{Canonical_Combining_Class=ccc14}', "");
Expect(0, 1461, '\P{^Canonical_Combining_Class=ccc14}', "");
Expect(1, 1460, '\p{Canonical_Combining_Class= CCC14}', "");
Expect(0, 1460, '\p{^Canonical_Combining_Class= CCC14}', "");
Expect(0, 1460, '\P{Canonical_Combining_Class= CCC14}', "");
Expect(1, 1460, '\P{^Canonical_Combining_Class= CCC14}', "");
Expect(0, 1461, '\p{Canonical_Combining_Class= CCC14}', "");
Expect(1, 1461, '\p{^Canonical_Combining_Class= CCC14}', "");
Expect(1, 1461, '\P{Canonical_Combining_Class= CCC14}', "");
Expect(0, 1461, '\P{^Canonical_Combining_Class= CCC14}', "");
Error('\p{Ccc: _ +14:=}');
Error('\P{Ccc: _ +14:=}');
Expect(1, 1460, '\p{Ccc=000001_4}', "");
Expect(0, 1460, '\p{^Ccc=000001_4}', "");
Expect(0, 1460, '\P{Ccc=000001_4}', "");
Expect(1, 1460, '\P{^Ccc=000001_4}', "");
Expect(0, 1461, '\p{Ccc=000001_4}', "");
Expect(1, 1461, '\p{^Ccc=000001_4}', "");
Expect(1, 1461, '\P{Ccc=000001_4}', "");
Expect(0, 1461, '\P{^Ccc=000001_4}', "");
Error('\p{Is_Canonical_Combining_Class=-_CCC14/a/}');
Error('\P{Is_Canonical_Combining_Class=-_CCC14/a/}');
Expect(1, 1460, '\p{Is_Canonical_Combining_Class=ccc14}', "");
Expect(0, 1460, '\p{^Is_Canonical_Combining_Class=ccc14}', "");
Expect(0, 1460, '\P{Is_Canonical_Combining_Class=ccc14}', "");
Expect(1, 1460, '\P{^Is_Canonical_Combining_Class=ccc14}', "");
Expect(0, 1461, '\p{Is_Canonical_Combining_Class=ccc14}', "");
Expect(1, 1461, '\p{^Is_Canonical_Combining_Class=ccc14}', "");
Expect(1, 1461, '\P{Is_Canonical_Combining_Class=ccc14}', "");
Expect(0, 1461, '\P{^Is_Canonical_Combining_Class=ccc14}', "");
Expect(1, 1460, '\p{Is_Canonical_Combining_Class:   _	CCC14}', "");
Expect(0, 1460, '\p{^Is_Canonical_Combining_Class:   _	CCC14}', "");
Expect(0, 1460, '\P{Is_Canonical_Combining_Class:   _	CCC14}', "");
Expect(1, 1460, '\P{^Is_Canonical_Combining_Class:   _	CCC14}', "");
Expect(0, 1461, '\p{Is_Canonical_Combining_Class:   _	CCC14}', "");
Expect(1, 1461, '\p{^Is_Canonical_Combining_Class:   _	CCC14}', "");
Expect(1, 1461, '\P{Is_Canonical_Combining_Class:   _	CCC14}', "");
Expect(0, 1461, '\P{^Is_Canonical_Combining_Class:   _	CCC14}', "");
Error('\p{Is_Ccc=		+1_4:=}');
Error('\P{Is_Ccc=		+1_4:=}');
Expect(1, 1460, '\p{Is_Ccc=0_0_0_0_0_014}', "");
Expect(0, 1460, '\p{^Is_Ccc=0_0_0_0_0_014}', "");
Expect(0, 1460, '\P{Is_Ccc=0_0_0_0_0_014}', "");
Expect(1, 1460, '\P{^Is_Ccc=0_0_0_0_0_014}', "");
Expect(0, 1461, '\p{Is_Ccc=0_0_0_0_0_014}', "");
Expect(1, 1461, '\p{^Is_Ccc=0_0_0_0_0_014}', "");
Expect(1, 1461, '\P{Is_Ccc=0_0_0_0_0_014}', "");
Expect(0, 1461, '\P{^Is_Ccc=0_0_0_0_0_014}', "");
Error('\p{Canonical_Combining_Class=:=CCC15}');
Error('\P{Canonical_Combining_Class=:=CCC15}');
Expect(1, 1461, '\p{Canonical_Combining_Class=ccc15}', "");
Expect(0, 1461, '\p{^Canonical_Combining_Class=ccc15}', "");
Expect(0, 1461, '\P{Canonical_Combining_Class=ccc15}', "");
Expect(1, 1461, '\P{^Canonical_Combining_Class=ccc15}', "");
Expect(0, 1462, '\p{Canonical_Combining_Class=ccc15}', "");
Expect(1, 1462, '\p{^Canonical_Combining_Class=ccc15}', "");
Expect(1, 1462, '\P{Canonical_Combining_Class=ccc15}', "");
Expect(0, 1462, '\P{^Canonical_Combining_Class=ccc15}', "");
Expect(1, 1461, '\p{Canonical_Combining_Class:--CCC15}', "");
Expect(0, 1461, '\p{^Canonical_Combining_Class:--CCC15}', "");
Expect(0, 1461, '\P{Canonical_Combining_Class:--CCC15}', "");
Expect(1, 1461, '\P{^Canonical_Combining_Class:--CCC15}', "");
Expect(0, 1462, '\p{Canonical_Combining_Class:--CCC15}', "");
Expect(1, 1462, '\p{^Canonical_Combining_Class:--CCC15}', "");
Expect(1, 1462, '\P{Canonical_Combining_Class:--CCC15}', "");
Expect(0, 1462, '\P{^Canonical_Combining_Class:--CCC15}', "");
Error('\p{Ccc=_	000_000_15/a/}');
Error('\P{Ccc=_	000_000_15/a/}');
Expect(1, 1461, '\p{Ccc=+0_0_15}', "");
Expect(0, 1461, '\p{^Ccc=+0_0_15}', "");
Expect(0, 1461, '\P{Ccc=+0_0_15}', "");
Expect(1, 1461, '\P{^Ccc=+0_0_15}', "");
Expect(0, 1462, '\p{Ccc=+0_0_15}', "");
Expect(1, 1462, '\p{^Ccc=+0_0_15}', "");
Expect(1, 1462, '\P{Ccc=+0_0_15}', "");
Expect(0, 1462, '\P{^Ccc=+0_0_15}', "");
Error('\p{Is_Canonical_Combining_Class=:= -CCC15}');
Error('\P{Is_Canonical_Combining_Class=:= -CCC15}');
Expect(1, 1461, '\p{Is_Canonical_Combining_Class=ccc15}', "");
Expect(0, 1461, '\p{^Is_Canonical_Combining_Class=ccc15}', "");
Expect(0, 1461, '\P{Is_Canonical_Combining_Class=ccc15}', "");
Expect(1, 1461, '\P{^Is_Canonical_Combining_Class=ccc15}', "");
Expect(0, 1462, '\p{Is_Canonical_Combining_Class=ccc15}', "");
Expect(1, 1462, '\p{^Is_Canonical_Combining_Class=ccc15}', "");
Expect(1, 1462, '\P{Is_Canonical_Combining_Class=ccc15}', "");
Expect(0, 1462, '\P{^Is_Canonical_Combining_Class=ccc15}', "");
Expect(1, 1461, '\p{Is_Canonical_Combining_Class=-ccc15}', "");
Expect(0, 1461, '\p{^Is_Canonical_Combining_Class=-ccc15}', "");
Expect(0, 1461, '\P{Is_Canonical_Combining_Class=-ccc15}', "");
Expect(1, 1461, '\P{^Is_Canonical_Combining_Class=-ccc15}', "");
Expect(0, 1462, '\p{Is_Canonical_Combining_Class=-ccc15}', "");
Expect(1, 1462, '\p{^Is_Canonical_Combining_Class=-ccc15}', "");
Expect(1, 1462, '\P{Is_Canonical_Combining_Class=-ccc15}', "");
Expect(0, 1462, '\P{^Is_Canonical_Combining_Class=-ccc15}', "");
Error('\p{Is_Ccc=	:=00000_00001_5}');
Error('\P{Is_Ccc=	:=00000_00001_5}');
Expect(1, 1461, '\p{Is_Ccc=+0_0_15}', "");
Expect(0, 1461, '\p{^Is_Ccc=+0_0_15}', "");
Expect(0, 1461, '\P{Is_Ccc=+0_0_15}', "");
Expect(1, 1461, '\P{^Is_Ccc=+0_0_15}', "");
Expect(0, 1462, '\p{Is_Ccc=+0_0_15}', "");
Expect(1, 1462, '\p{^Is_Ccc=+0_0_15}', "");
Expect(1, 1462, '\P{Is_Ccc=+0_0_15}', "");
Expect(0, 1462, '\P{^Is_Ccc=+0_0_15}', "");
Error('\p{Canonical_Combining_Class=:=-	CCC16}');
Error('\P{Canonical_Combining_Class=:=-	CCC16}');
Expect(1, 1462, '\p{Canonical_Combining_Class=ccc16}', "");
Expect(0, 1462, '\p{^Canonical_Combining_Class=ccc16}', "");
Expect(0, 1462, '\P{Canonical_Combining_Class=ccc16}', "");
Expect(1, 1462, '\P{^Canonical_Combining_Class=ccc16}', "");
Expect(0, 1463, '\p{Canonical_Combining_Class=ccc16}', "");
Expect(1, 1463, '\p{^Canonical_Combining_Class=ccc16}', "");
Expect(1, 1463, '\P{Canonical_Combining_Class=ccc16}', "");
Expect(0, 1463, '\P{^Canonical_Combining_Class=ccc16}', "");
Expect(1, 1462, '\p{Canonical_Combining_Class= _CCC16}', "");
Expect(0, 1462, '\p{^Canonical_Combining_Class= _CCC16}', "");
Expect(0, 1462, '\P{Canonical_Combining_Class= _CCC16}', "");
Expect(1, 1462, '\P{^Canonical_Combining_Class= _CCC16}', "");
Expect(0, 1463, '\p{Canonical_Combining_Class= _CCC16}', "");
Expect(1, 1463, '\p{^Canonical_Combining_Class= _CCC16}', "");
Expect(1, 1463, '\P{Canonical_Combining_Class= _CCC16}', "");
Expect(0, 1463, '\P{^Canonical_Combining_Class= _CCC16}', "");
Error('\p{Ccc= :=0016}');
Error('\P{Ccc= :=0016}');
Expect(1, 1462, '\p{Ccc=0_0_0_0_016}', "");
Expect(0, 1462, '\p{^Ccc=0_0_0_0_016}', "");
Expect(0, 1462, '\P{Ccc=0_0_0_0_016}', "");
Expect(1, 1462, '\P{^Ccc=0_0_0_0_016}', "");
Expect(0, 1463, '\p{Ccc=0_0_0_0_016}', "");
Expect(1, 1463, '\p{^Ccc=0_0_0_0_016}', "");
Expect(1, 1463, '\P{Ccc=0_0_0_0_016}', "");
Expect(0, 1463, '\P{^Ccc=0_0_0_0_016}', "");
Error('\p{Is_Canonical_Combining_Class=/a/_	CCC16}');
Error('\P{Is_Canonical_Combining_Class=/a/_	CCC16}');
Expect(1, 1462, '\p{Is_Canonical_Combining_Class=ccc16}', "");
Expect(0, 1462, '\p{^Is_Canonical_Combining_Class=ccc16}', "");
Expect(0, 1462, '\P{Is_Canonical_Combining_Class=ccc16}', "");
Expect(1, 1462, '\P{^Is_Canonical_Combining_Class=ccc16}', "");
Expect(0, 1463, '\p{Is_Canonical_Combining_Class=ccc16}', "");
Expect(1, 1463, '\p{^Is_Canonical_Combining_Class=ccc16}', "");
Expect(1, 1463, '\P{Is_Canonical_Combining_Class=ccc16}', "");
Expect(0, 1463, '\P{^Is_Canonical_Combining_Class=ccc16}', "");
Expect(1, 1462, '\p{Is_Canonical_Combining_Class=-ccc16}', "");
Expect(0, 1462, '\p{^Is_Canonical_Combining_Class=-ccc16}', "");
Expect(0, 1462, '\P{Is_Canonical_Combining_Class=-ccc16}', "");
Expect(1, 1462, '\P{^Is_Canonical_Combining_Class=-ccc16}', "");
Expect(0, 1463, '\p{Is_Canonical_Combining_Class=-ccc16}', "");
Expect(1, 1463, '\p{^Is_Canonical_Combining_Class=-ccc16}', "");
Expect(1, 1463, '\P{Is_Canonical_Combining_Class=-ccc16}', "");
Expect(0, 1463, '\P{^Is_Canonical_Combining_Class=-ccc16}', "");
Error('\p{Is_Ccc=_	+00_00_16:=}');
Error('\P{Is_Ccc=_	+00_00_16:=}');
Expect(1, 1462, '\p{Is_Ccc=0000000016}', "");
Expect(0, 1462, '\p{^Is_Ccc=0000000016}', "");
Expect(0, 1462, '\P{Is_Ccc=0000000016}', "");
Expect(1, 1462, '\P{^Is_Ccc=0000000016}', "");
Expect(0, 1463, '\p{Is_Ccc=0000000016}', "");
Expect(1, 1463, '\p{^Is_Ccc=0000000016}', "");
Expect(1, 1463, '\P{Is_Ccc=0000000016}', "");
Expect(0, 1463, '\P{^Is_Ccc=0000000016}', "");
Error('\p{Canonical_Combining_Class=/a/_CCC17}');
Error('\P{Canonical_Combining_Class=/a/_CCC17}');
Expect(1, 1463, '\p{Canonical_Combining_Class=ccc17}', "");
Expect(0, 1463, '\p{^Canonical_Combining_Class=ccc17}', "");
Expect(0, 1463, '\P{Canonical_Combining_Class=ccc17}', "");
Expect(1, 1463, '\P{^Canonical_Combining_Class=ccc17}', "");
Expect(0, 1464, '\p{Canonical_Combining_Class=ccc17}', "");
Expect(1, 1464, '\p{^Canonical_Combining_Class=ccc17}', "");
Expect(1, 1464, '\P{Canonical_Combining_Class=ccc17}', "");
Expect(0, 1464, '\P{^Canonical_Combining_Class=ccc17}', "");
Expect(1, 1463, '\p{Canonical_Combining_Class=-CCC17}', "");
Expect(0, 1463, '\p{^Canonical_Combining_Class=-CCC17}', "");
Expect(0, 1463, '\P{Canonical_Combining_Class=-CCC17}', "");
Expect(1, 1463, '\P{^Canonical_Combining_Class=-CCC17}', "");
Expect(0, 1464, '\p{Canonical_Combining_Class=-CCC17}', "");
Expect(1, 1464, '\p{^Canonical_Combining_Class=-CCC17}', "");
Expect(1, 1464, '\P{Canonical_Combining_Class=-CCC17}', "");
Expect(0, 1464, '\P{^Canonical_Combining_Class=-CCC17}', "");
Error('\p{Ccc=:=+000000017}');
Error('\P{Ccc=:=+000000017}');
Expect(1, 1463, '\p{Ccc:1_7}', "");
Expect(0, 1463, '\p{^Ccc:1_7}', "");
Expect(0, 1463, '\P{Ccc:1_7}', "");
Expect(1, 1463, '\P{^Ccc:1_7}', "");
Expect(0, 1464, '\p{Ccc:1_7}', "");
Expect(1, 1464, '\p{^Ccc:1_7}', "");
Expect(1, 1464, '\P{Ccc:1_7}', "");
Expect(0, 1464, '\P{^Ccc:1_7}', "");
Error('\p{Is_Canonical_Combining_Class=_	CCC17:=}');
Error('\P{Is_Canonical_Combining_Class=_	CCC17:=}');
Expect(1, 1463, '\p{Is_Canonical_Combining_Class: ccc17}', "");
Expect(0, 1463, '\p{^Is_Canonical_Combining_Class: ccc17}', "");
Expect(0, 1463, '\P{Is_Canonical_Combining_Class: ccc17}', "");
Expect(1, 1463, '\P{^Is_Canonical_Combining_Class: ccc17}', "");
Expect(0, 1464, '\p{Is_Canonical_Combining_Class: ccc17}', "");
Expect(1, 1464, '\p{^Is_Canonical_Combining_Class: ccc17}', "");
Expect(1, 1464, '\P{Is_Canonical_Combining_Class: ccc17}', "");
Expect(0, 1464, '\P{^Is_Canonical_Combining_Class: ccc17}', "");
Expect(1, 1463, '\p{Is_Canonical_Combining_Class=--ccc17}', "");
Expect(0, 1463, '\p{^Is_Canonical_Combining_Class=--ccc17}', "");
Expect(0, 1463, '\P{Is_Canonical_Combining_Class=--ccc17}', "");
Expect(1, 1463, '\P{^Is_Canonical_Combining_Class=--ccc17}', "");
Expect(0, 1464, '\p{Is_Canonical_Combining_Class=--ccc17}', "");
Expect(1, 1464, '\p{^Is_Canonical_Combining_Class=--ccc17}', "");
Expect(1, 1464, '\P{Is_Canonical_Combining_Class=--ccc17}', "");
Expect(0, 1464, '\P{^Is_Canonical_Combining_Class=--ccc17}', "");
Error('\p{Is_Ccc=_0000000017/a/}');
Error('\P{Is_Ccc=_0000000017/a/}');
Expect(1, 1463, '\p{Is_Ccc=0000_0001_7}', "");
Expect(0, 1463, '\p{^Is_Ccc=0000_0001_7}', "");
Expect(0, 1463, '\P{Is_Ccc=0000_0001_7}', "");
Expect(1, 1463, '\P{^Is_Ccc=0000_0001_7}', "");
Expect(0, 1464, '\p{Is_Ccc=0000_0001_7}', "");
Expect(1, 1464, '\p{^Is_Ccc=0000_0001_7}', "");
Expect(1, 1464, '\P{Is_Ccc=0000_0001_7}', "");
Expect(0, 1464, '\P{^Is_Ccc=0000_0001_7}', "");
Error('\p{Canonical_Combining_Class:   _/a/CCC18}');
Error('\P{Canonical_Combining_Class:   _/a/CCC18}');
Expect(1, 1479, '\p{Canonical_Combining_Class=ccc18}', "");
Expect(0, 1479, '\p{^Canonical_Combining_Class=ccc18}', "");
Expect(0, 1479, '\P{Canonical_Combining_Class=ccc18}', "");
Expect(1, 1479, '\P{^Canonical_Combining_Class=ccc18}', "");
Expect(0, 1480, '\p{Canonical_Combining_Class=ccc18}', "");
Expect(1, 1480, '\p{^Canonical_Combining_Class=ccc18}', "");
Expect(1, 1480, '\P{Canonical_Combining_Class=ccc18}', "");
Expect(0, 1480, '\P{^Canonical_Combining_Class=ccc18}', "");
Expect(1, 1479, '\p{Canonical_Combining_Class=		ccc18}', "");
Expect(0, 1479, '\p{^Canonical_Combining_Class=		ccc18}', "");
Expect(0, 1479, '\P{Canonical_Combining_Class=		ccc18}', "");
Expect(1, 1479, '\P{^Canonical_Combining_Class=		ccc18}', "");
Expect(0, 1480, '\p{Canonical_Combining_Class=		ccc18}', "");
Expect(1, 1480, '\p{^Canonical_Combining_Class=		ccc18}', "");
Expect(1, 1480, '\P{Canonical_Combining_Class=		ccc18}', "");
Expect(0, 1480, '\P{^Canonical_Combining_Class=		ccc18}', "");
Error('\p{Ccc=/a/0000000018}');
Error('\P{Ccc=/a/0000000018}');
Expect(1, 1479, '\p{Ccc=1_8}', "");
Expect(0, 1479, '\p{^Ccc=1_8}', "");
Expect(0, 1479, '\P{Ccc=1_8}', "");
Expect(1, 1479, '\P{^Ccc=1_8}', "");
Expect(0, 1480, '\p{Ccc=1_8}', "");
Expect(1, 1480, '\p{^Ccc=1_8}', "");
Expect(1, 1480, '\P{Ccc=1_8}', "");
Expect(0, 1480, '\P{^Ccc=1_8}', "");
Error('\p{Is_Canonical_Combining_Class=:=-CCC18}');
Error('\P{Is_Canonical_Combining_Class=:=-CCC18}');
Expect(1, 1479, '\p{Is_Canonical_Combining_Class=ccc18}', "");
Expect(0, 1479, '\p{^Is_Canonical_Combining_Class=ccc18}', "");
Expect(0, 1479, '\P{Is_Canonical_Combining_Class=ccc18}', "");
Expect(1, 1479, '\P{^Is_Canonical_Combining_Class=ccc18}', "");
Expect(0, 1480, '\p{Is_Canonical_Combining_Class=ccc18}', "");
Expect(1, 1480, '\p{^Is_Canonical_Combining_Class=ccc18}', "");
Expect(1, 1480, '\P{Is_Canonical_Combining_Class=ccc18}', "");
Expect(0, 1480, '\P{^Is_Canonical_Combining_Class=ccc18}', "");
Expect(1, 1479, '\p{Is_Canonical_Combining_Class:			CCC18}', "");
Expect(0, 1479, '\p{^Is_Canonical_Combining_Class:			CCC18}', "");
Expect(0, 1479, '\P{Is_Canonical_Combining_Class:			CCC18}', "");
Expect(1, 1479, '\P{^Is_Canonical_Combining_Class:			CCC18}', "");
Expect(0, 1480, '\p{Is_Canonical_Combining_Class:			CCC18}', "");
Expect(1, 1480, '\p{^Is_Canonical_Combining_Class:			CCC18}', "");
Expect(1, 1480, '\P{Is_Canonical_Combining_Class:			CCC18}', "");
Expect(0, 1480, '\P{^Is_Canonical_Combining_Class:			CCC18}', "");
Error('\p{Is_Ccc=-/a/01_8}');
Error('\P{Is_Ccc=-/a/01_8}');
Expect(1, 1479, '\p{Is_Ccc:   00_00_01_8}', "");
Expect(0, 1479, '\p{^Is_Ccc:   00_00_01_8}', "");
Expect(0, 1479, '\P{Is_Ccc:   00_00_01_8}', "");
Expect(1, 1479, '\P{^Is_Ccc:   00_00_01_8}', "");
Expect(0, 1480, '\p{Is_Ccc:   00_00_01_8}', "");
Expect(1, 1480, '\p{^Is_Ccc:   00_00_01_8}', "");
Expect(1, 1480, '\P{Is_Ccc:   00_00_01_8}', "");
Expect(0, 1480, '\P{^Is_Ccc:   00_00_01_8}', "");
Error('\p{Canonical_Combining_Class= CCC19/a/}');
Error('\P{Canonical_Combining_Class= CCC19/a/}');
Expect(1, 1466, '\p{Canonical_Combining_Class=ccc19}', "");
Expect(0, 1466, '\p{^Canonical_Combining_Class=ccc19}', "");
Expect(0, 1466, '\P{Canonical_Combining_Class=ccc19}', "");
Expect(1, 1466, '\P{^Canonical_Combining_Class=ccc19}', "");
Expect(0, 1467, '\p{Canonical_Combining_Class=ccc19}', "");
Expect(1, 1467, '\p{^Canonical_Combining_Class=ccc19}', "");
Expect(1, 1467, '\P{Canonical_Combining_Class=ccc19}', "");
Expect(0, 1467, '\P{^Canonical_Combining_Class=ccc19}', "");
Expect(1, 1466, '\p{Canonical_Combining_Class=CCC19}', "");
Expect(0, 1466, '\p{^Canonical_Combining_Class=CCC19}', "");
Expect(0, 1466, '\P{Canonical_Combining_Class=CCC19}', "");
Expect(1, 1466, '\P{^Canonical_Combining_Class=CCC19}', "");
Expect(0, 1467, '\p{Canonical_Combining_Class=CCC19}', "");
Expect(1, 1467, '\p{^Canonical_Combining_Class=CCC19}', "");
Expect(1, 1467, '\P{Canonical_Combining_Class=CCC19}', "");
Expect(0, 1467, '\P{^Canonical_Combining_Class=CCC19}', "");
Error('\p{Ccc= 019/a/}');
Error('\P{Ccc= 019/a/}');
Expect(1, 1466, '\p{Ccc:	0019}', "");
Expect(0, 1466, '\p{^Ccc:	0019}', "");
Expect(0, 1466, '\P{Ccc:	0019}', "");
Expect(1, 1466, '\P{^Ccc:	0019}', "");
Expect(0, 1467, '\p{Ccc:	0019}', "");
Expect(1, 1467, '\p{^Ccc:	0019}', "");
Expect(1, 1467, '\P{Ccc:	0019}', "");
Expect(0, 1467, '\P{^Ccc:	0019}', "");
Error('\p{Is_Canonical_Combining_Class=/a/_-CCC19}');
Error('\P{Is_Canonical_Combining_Class=/a/_-CCC19}');
Expect(1, 1466, '\p{Is_Canonical_Combining_Class=ccc19}', "");
Expect(0, 1466, '\p{^Is_Canonical_Combining_Class=ccc19}', "");
Expect(0, 1466, '\P{Is_Canonical_Combining_Class=ccc19}', "");
Expect(1, 1466, '\P{^Is_Canonical_Combining_Class=ccc19}', "");
Expect(0, 1467, '\p{Is_Canonical_Combining_Class=ccc19}', "");
Expect(1, 1467, '\p{^Is_Canonical_Combining_Class=ccc19}', "");
Expect(1, 1467, '\P{Is_Canonical_Combining_Class=ccc19}', "");
Expect(0, 1467, '\P{^Is_Canonical_Combining_Class=ccc19}', "");
Expect(1, 1466, '\p{Is_Canonical_Combining_Class=-	ccc19}', "");
Expect(0, 1466, '\p{^Is_Canonical_Combining_Class=-	ccc19}', "");
Expect(0, 1466, '\P{Is_Canonical_Combining_Class=-	ccc19}', "");
Expect(1, 1466, '\P{^Is_Canonical_Combining_Class=-	ccc19}', "");
Expect(0, 1467, '\p{Is_Canonical_Combining_Class=-	ccc19}', "");
Expect(1, 1467, '\p{^Is_Canonical_Combining_Class=-	ccc19}', "");
Expect(1, 1467, '\P{Is_Canonical_Combining_Class=-	ccc19}', "");
Expect(0, 1467, '\P{^Is_Canonical_Combining_Class=-	ccc19}', "");
Error('\p{Is_Ccc= /a/0000_0001_9}');
Error('\P{Is_Ccc= /a/0000_0001_9}');
Expect(1, 1466, '\p{Is_Ccc=+000000019}', "");
Expect(0, 1466, '\p{^Is_Ccc=+000000019}', "");
Expect(0, 1466, '\P{Is_Ccc=+000000019}', "");
Expect(1, 1466, '\P{^Is_Ccc=+000000019}', "");
Expect(0, 1467, '\p{Is_Ccc=+000000019}', "");
Expect(1, 1467, '\p{^Is_Ccc=+000000019}', "");
Expect(1, 1467, '\P{Is_Ccc=+000000019}', "");
Expect(0, 1467, '\P{^Is_Ccc=+000000019}', "");
Error('\p{Canonical_Combining_Class=/a/_ccc20}');
Error('\P{Canonical_Combining_Class=/a/_ccc20}');
Expect(1, 1467, '\p{Canonical_Combining_Class=ccc20}', "");
Expect(0, 1467, '\p{^Canonical_Combining_Class=ccc20}', "");
Expect(0, 1467, '\P{Canonical_Combining_Class=ccc20}', "");
Expect(1, 1467, '\P{^Canonical_Combining_Class=ccc20}', "");
Expect(0, 1468, '\p{Canonical_Combining_Class=ccc20}', "");
Expect(1, 1468, '\p{^Canonical_Combining_Class=ccc20}', "");
Expect(1, 1468, '\P{Canonical_Combining_Class=ccc20}', "");
Expect(0, 1468, '\P{^Canonical_Combining_Class=ccc20}', "");
Expect(1, 1467, '\p{Canonical_Combining_Class=_	CCC20}', "");
Expect(0, 1467, '\p{^Canonical_Combining_Class=_	CCC20}', "");
Expect(0, 1467, '\P{Canonical_Combining_Class=_	CCC20}', "");
Expect(1, 1467, '\P{^Canonical_Combining_Class=_	CCC20}', "");
Expect(0, 1468, '\p{Canonical_Combining_Class=_	CCC20}', "");
Expect(1, 1468, '\p{^Canonical_Combining_Class=_	CCC20}', "");
Expect(1, 1468, '\P{Canonical_Combining_Class=_	CCC20}', "");
Expect(0, 1468, '\P{^Canonical_Combining_Class=_	CCC20}', "");
Error('\p{Ccc= -+00000000020:=}');
Error('\P{Ccc= -+00000000020:=}');
Expect(1, 1467, '\p{Ccc=020}', "");
Expect(0, 1467, '\p{^Ccc=020}', "");
Expect(0, 1467, '\P{Ccc=020}', "");
Expect(1, 1467, '\P{^Ccc=020}', "");
Expect(0, 1468, '\p{Ccc=020}', "");
Expect(1, 1468, '\p{^Ccc=020}', "");
Expect(1, 1468, '\P{Ccc=020}', "");
Expect(0, 1468, '\P{^Ccc=020}', "");
Error('\p{Is_Canonical_Combining_Class=:=		CCC20}');
Error('\P{Is_Canonical_Combining_Class=:=		CCC20}');
Expect(1, 1467, '\p{Is_Canonical_Combining_Class=ccc20}', "");
Expect(0, 1467, '\p{^Is_Canonical_Combining_Class=ccc20}', "");
Expect(0, 1467, '\P{Is_Canonical_Combining_Class=ccc20}', "");
Expect(1, 1467, '\P{^Is_Canonical_Combining_Class=ccc20}', "");
Expect(0, 1468, '\p{Is_Canonical_Combining_Class=ccc20}', "");
Expect(1, 1468, '\p{^Is_Canonical_Combining_Class=ccc20}', "");
Expect(1, 1468, '\P{Is_Canonical_Combining_Class=ccc20}', "");
Expect(0, 1468, '\P{^Is_Canonical_Combining_Class=ccc20}', "");
Expect(1, 1467, '\p{Is_Canonical_Combining_Class=- CCC20}', "");
Expect(0, 1467, '\p{^Is_Canonical_Combining_Class=- CCC20}', "");
Expect(0, 1467, '\P{Is_Canonical_Combining_Class=- CCC20}', "");
Expect(1, 1467, '\P{^Is_Canonical_Combining_Class=- CCC20}', "");
Expect(0, 1468, '\p{Is_Canonical_Combining_Class=- CCC20}', "");
Expect(1, 1468, '\p{^Is_Canonical_Combining_Class=- CCC20}', "");
Expect(1, 1468, '\P{Is_Canonical_Combining_Class=- CCC20}', "");
Expect(0, 1468, '\P{^Is_Canonical_Combining_Class=- CCC20}', "");
Error('\p{Is_Ccc=/a/	00_00_00_020}');
Error('\P{Is_Ccc=/a/	00_00_00_020}');
Expect(1, 1467, '\p{Is_Ccc=+0_0_0_0_020}', "");
Expect(0, 1467, '\p{^Is_Ccc=+0_0_0_0_020}', "");
Expect(0, 1467, '\P{Is_Ccc=+0_0_0_0_020}', "");
Expect(1, 1467, '\P{^Is_Ccc=+0_0_0_0_020}', "");
Expect(0, 1468, '\p{Is_Ccc=+0_0_0_0_020}', "");
Expect(1, 1468, '\p{^Is_Ccc=+0_0_0_0_020}', "");
Expect(1, 1468, '\P{Is_Ccc=+0_0_0_0_020}', "");
Expect(0, 1468, '\P{^Is_Ccc=+0_0_0_0_020}', "");
Error('\p{Canonical_Combining_Class= :=CCC21}');
Error('\P{Canonical_Combining_Class= :=CCC21}');
Expect(1, 1468, '\p{Canonical_Combining_Class=ccc21}', "");
Expect(0, 1468, '\p{^Canonical_Combining_Class=ccc21}', "");
Expect(0, 1468, '\P{Canonical_Combining_Class=ccc21}', "");
Expect(1, 1468, '\P{^Canonical_Combining_Class=ccc21}', "");
Expect(0, 1469, '\p{Canonical_Combining_Class=ccc21}', "");
Expect(1, 1469, '\p{^Canonical_Combining_Class=ccc21}', "");
Expect(1, 1469, '\P{Canonical_Combining_Class=ccc21}', "");
Expect(0, 1469, '\P{^Canonical_Combining_Class=ccc21}', "");
Expect(1, 1468, '\p{Canonical_Combining_Class=CCC21}', "");
Expect(0, 1468, '\p{^Canonical_Combining_Class=CCC21}', "");
Expect(0, 1468, '\P{Canonical_Combining_Class=CCC21}', "");
Expect(1, 1468, '\P{^Canonical_Combining_Class=CCC21}', "");
Expect(0, 1469, '\p{Canonical_Combining_Class=CCC21}', "");
Expect(1, 1469, '\p{^Canonical_Combining_Class=CCC21}', "");
Expect(1, 1469, '\P{Canonical_Combining_Class=CCC21}', "");
Expect(0, 1469, '\P{^Canonical_Combining_Class=CCC21}', "");
Error('\p{Ccc=:=__+000021}');
Error('\P{Ccc=:=__+000021}');
Expect(1, 1468, '\p{Ccc=0_0_0_0_0_0_0_0021}', "");
Expect(0, 1468, '\p{^Ccc=0_0_0_0_0_0_0_0021}', "");
Expect(0, 1468, '\P{Ccc=0_0_0_0_0_0_0_0021}', "");
Expect(1, 1468, '\P{^Ccc=0_0_0_0_0_0_0_0021}', "");
Expect(0, 1469, '\p{Ccc=0_0_0_0_0_0_0_0021}', "");
Expect(1, 1469, '\p{^Ccc=0_0_0_0_0_0_0_0021}', "");
Expect(1, 1469, '\P{Ccc=0_0_0_0_0_0_0_0021}', "");
Expect(0, 1469, '\P{^Ccc=0_0_0_0_0_0_0_0021}', "");
Error('\p{Is_Canonical_Combining_Class=	:=CCC21}');
Error('\P{Is_Canonical_Combining_Class=	:=CCC21}');
Expect(1, 1468, '\p{Is_Canonical_Combining_Class=ccc21}', "");
Expect(0, 1468, '\p{^Is_Canonical_Combining_Class=ccc21}', "");
Expect(0, 1468, '\P{Is_Canonical_Combining_Class=ccc21}', "");
Expect(1, 1468, '\P{^Is_Canonical_Combining_Class=ccc21}', "");
Expect(0, 1469, '\p{Is_Canonical_Combining_Class=ccc21}', "");
Expect(1, 1469, '\p{^Is_Canonical_Combining_Class=ccc21}', "");
Expect(1, 1469, '\P{Is_Canonical_Combining_Class=ccc21}', "");
Expect(0, 1469, '\P{^Is_Canonical_Combining_Class=ccc21}', "");
Expect(1, 1468, '\p{Is_Canonical_Combining_Class=_ CCC21}', "");
Expect(0, 1468, '\p{^Is_Canonical_Combining_Class=_ CCC21}', "");
Expect(0, 1468, '\P{Is_Canonical_Combining_Class=_ CCC21}', "");
Expect(1, 1468, '\P{^Is_Canonical_Combining_Class=_ CCC21}', "");
Expect(0, 1469, '\p{Is_Canonical_Combining_Class=_ CCC21}', "");
Expect(1, 1469, '\p{^Is_Canonical_Combining_Class=_ CCC21}', "");
Expect(1, 1469, '\P{Is_Canonical_Combining_Class=_ CCC21}', "");
Expect(0, 1469, '\P{^Is_Canonical_Combining_Class=_ CCC21}', "");
Error('\p{Is_Ccc=_ +00000021:=}');
Error('\P{Is_Ccc=_ +00000021:=}');
Expect(1, 1468, '\p{Is_Ccc=+000_000_21}', "");
Expect(0, 1468, '\p{^Is_Ccc=+000_000_21}', "");
Expect(0, 1468, '\P{Is_Ccc=+000_000_21}', "");
Expect(1, 1468, '\P{^Is_Ccc=+000_000_21}', "");
Expect(0, 1469, '\p{Is_Ccc=+000_000_21}', "");
Expect(1, 1469, '\p{^Is_Ccc=+000_000_21}', "");
Expect(1, 1469, '\P{Is_Ccc=+000_000_21}', "");
Expect(0, 1469, '\P{^Is_Ccc=+000_000_21}', "");
Error('\p{Canonical_Combining_Class=_	CCC22/a/}');
Error('\P{Canonical_Combining_Class=_	CCC22/a/}');
Expect(1, 1469, '\p{Canonical_Combining_Class=ccc22}', "");
Expect(0, 1469, '\p{^Canonical_Combining_Class=ccc22}', "");
Expect(0, 1469, '\P{Canonical_Combining_Class=ccc22}', "");
Expect(1, 1469, '\P{^Canonical_Combining_Class=ccc22}', "");
Expect(0, 1470, '\p{Canonical_Combining_Class=ccc22}', "");
Expect(1, 1470, '\p{^Canonical_Combining_Class=ccc22}', "");
Expect(1, 1470, '\P{Canonical_Combining_Class=ccc22}', "");
Expect(0, 1470, '\P{^Canonical_Combining_Class=ccc22}', "");
Expect(1, 1469, '\p{Canonical_Combining_Class=-_CCC22}', "");
Expect(0, 1469, '\p{^Canonical_Combining_Class=-_CCC22}', "");
Expect(0, 1469, '\P{Canonical_Combining_Class=-_CCC22}', "");
Expect(1, 1469, '\P{^Canonical_Combining_Class=-_CCC22}', "");
Expect(0, 1470, '\p{Canonical_Combining_Class=-_CCC22}', "");
Expect(1, 1470, '\p{^Canonical_Combining_Class=-_CCC22}', "");
Expect(1, 1470, '\P{Canonical_Combining_Class=-_CCC22}', "");
Expect(0, 1470, '\P{^Canonical_Combining_Class=-_CCC22}', "");
Error('\p{Ccc=-	000000022/a/}');
Error('\P{Ccc=-	000000022/a/}');
Expect(1, 1469, '\p{Ccc=0_0_0_0_0022}', "");
Expect(0, 1469, '\p{^Ccc=0_0_0_0_0022}', "");
Expect(0, 1469, '\P{Ccc=0_0_0_0_0022}', "");
Expect(1, 1469, '\P{^Ccc=0_0_0_0_0022}', "");
Expect(0, 1470, '\p{Ccc=0_0_0_0_0022}', "");
Expect(1, 1470, '\p{^Ccc=0_0_0_0_0022}', "");
Expect(1, 1470, '\P{Ccc=0_0_0_0_0022}', "");
Expect(0, 1470, '\P{^Ccc=0_0_0_0_0022}', "");
Error('\p{Is_Canonical_Combining_Class=		ccc22:=}');
Error('\P{Is_Canonical_Combining_Class=		ccc22:=}');
Expect(1, 1469, '\p{Is_Canonical_Combining_Class=ccc22}', "");
Expect(0, 1469, '\p{^Is_Canonical_Combining_Class=ccc22}', "");
Expect(0, 1469, '\P{Is_Canonical_Combining_Class=ccc22}', "");
Expect(1, 1469, '\P{^Is_Canonical_Combining_Class=ccc22}', "");
Expect(0, 1470, '\p{Is_Canonical_Combining_Class=ccc22}', "");
Expect(1, 1470, '\p{^Is_Canonical_Combining_Class=ccc22}', "");
Expect(1, 1470, '\P{Is_Canonical_Combining_Class=ccc22}', "");
Expect(0, 1470, '\P{^Is_Canonical_Combining_Class=ccc22}', "");
Expect(1, 1469, '\p{Is_Canonical_Combining_Class:	_CCC22}', "");
Expect(0, 1469, '\p{^Is_Canonical_Combining_Class:	_CCC22}', "");
Expect(0, 1469, '\P{Is_Canonical_Combining_Class:	_CCC22}', "");
Expect(1, 1469, '\P{^Is_Canonical_Combining_Class:	_CCC22}', "");
Expect(0, 1470, '\p{Is_Canonical_Combining_Class:	_CCC22}', "");
Expect(1, 1470, '\p{^Is_Canonical_Combining_Class:	_CCC22}', "");
Expect(1, 1470, '\P{Is_Canonical_Combining_Class:	_CCC22}', "");
Expect(0, 1470, '\P{^Is_Canonical_Combining_Class:	_CCC22}', "");
Error('\p{Is_Ccc=-00_00_00_22:=}');
Error('\P{Is_Ccc=-00_00_00_22:=}');
Expect(1, 1469, '\p{Is_Ccc=0_0_0_0_0_0_0_22}', "");
Expect(0, 1469, '\p{^Is_Ccc=0_0_0_0_0_0_0_22}', "");
Expect(0, 1469, '\P{Is_Ccc=0_0_0_0_0_0_0_22}', "");
Expect(1, 1469, '\P{^Is_Ccc=0_0_0_0_0_0_0_22}', "");
Expect(0, 1470, '\p{Is_Ccc=0_0_0_0_0_0_0_22}', "");
Expect(1, 1470, '\p{^Is_Ccc=0_0_0_0_0_0_0_22}', "");
Expect(1, 1470, '\P{Is_Ccc=0_0_0_0_0_0_0_22}', "");
Expect(0, 1470, '\P{^Is_Ccc=0_0_0_0_0_0_0_22}', "");
Error('\p{Canonical_Combining_Class=/a/	_CCC23}');
Error('\P{Canonical_Combining_Class=/a/	_CCC23}');
Expect(1, 1471, '\p{Canonical_Combining_Class=ccc23}', "");
Expect(0, 1471, '\p{^Canonical_Combining_Class=ccc23}', "");
Expect(0, 1471, '\P{Canonical_Combining_Class=ccc23}', "");
Expect(1, 1471, '\P{^Canonical_Combining_Class=ccc23}', "");
Expect(0, 1472, '\p{Canonical_Combining_Class=ccc23}', "");
Expect(1, 1472, '\p{^Canonical_Combining_Class=ccc23}', "");
Expect(1, 1472, '\P{Canonical_Combining_Class=ccc23}', "");
Expect(0, 1472, '\P{^Canonical_Combining_Class=ccc23}', "");
Expect(1, 1471, '\p{Canonical_Combining_Class:   __CCC23}', "");
Expect(0, 1471, '\p{^Canonical_Combining_Class:   __CCC23}', "");
Expect(0, 1471, '\P{Canonical_Combining_Class:   __CCC23}', "");
Expect(1, 1471, '\P{^Canonical_Combining_Class:   __CCC23}', "");
Expect(0, 1472, '\p{Canonical_Combining_Class:   __CCC23}', "");
Expect(1, 1472, '\p{^Canonical_Combining_Class:   __CCC23}', "");
Expect(1, 1472, '\P{Canonical_Combining_Class:   __CCC23}', "");
Expect(0, 1472, '\P{^Canonical_Combining_Class:   __CCC23}', "");
Error('\p{Ccc=__00000023/a/}');
Error('\P{Ccc=__00000023/a/}');
Expect(1, 1471, '\p{Ccc=00023}', "");
Expect(0, 1471, '\p{^Ccc=00023}', "");
Expect(0, 1471, '\P{Ccc=00023}', "");
Expect(1, 1471, '\P{^Ccc=00023}', "");
Expect(0, 1472, '\p{Ccc=00023}', "");
Expect(1, 1472, '\p{^Ccc=00023}', "");
Expect(1, 1472, '\P{Ccc=00023}', "");
Expect(0, 1472, '\P{^Ccc=00023}', "");
Error('\p{Is_Canonical_Combining_Class=	 CCC23/a/}');
Error('\P{Is_Canonical_Combining_Class=	 CCC23/a/}');
Expect(1, 1471, '\p{Is_Canonical_Combining_Class=ccc23}', "");
Expect(0, 1471, '\p{^Is_Canonical_Combining_Class=ccc23}', "");
Expect(0, 1471, '\P{Is_Canonical_Combining_Class=ccc23}', "");
Expect(1, 1471, '\P{^Is_Canonical_Combining_Class=ccc23}', "");
Expect(0, 1472, '\p{Is_Canonical_Combining_Class=ccc23}', "");
Expect(1, 1472, '\p{^Is_Canonical_Combining_Class=ccc23}', "");
Expect(1, 1472, '\P{Is_Canonical_Combining_Class=ccc23}', "");
Expect(0, 1472, '\P{^Is_Canonical_Combining_Class=ccc23}', "");
Expect(1, 1471, '\p{Is_Canonical_Combining_Class=	 CCC23}', "");
Expect(0, 1471, '\p{^Is_Canonical_Combining_Class=	 CCC23}', "");
Expect(0, 1471, '\P{Is_Canonical_Combining_Class=	 CCC23}', "");
Expect(1, 1471, '\P{^Is_Canonical_Combining_Class=	 CCC23}', "");
Expect(0, 1472, '\p{Is_Canonical_Combining_Class=	 CCC23}', "");
Expect(1, 1472, '\p{^Is_Canonical_Combining_Class=	 CCC23}', "");
Expect(1, 1472, '\P{Is_Canonical_Combining_Class=	 CCC23}', "");
Expect(0, 1472, '\P{^Is_Canonical_Combining_Class=	 CCC23}', "");
Error('\p{Is_Ccc=-:=23}');
Error('\P{Is_Ccc=-:=23}');
Expect(1, 1471, '\p{Is_Ccc=023}', "");
Expect(0, 1471, '\p{^Is_Ccc=023}', "");
Expect(0, 1471, '\P{Is_Ccc=023}', "");
Expect(1, 1471, '\P{^Is_Ccc=023}', "");
Expect(0, 1472, '\p{Is_Ccc=023}', "");
Expect(1, 1472, '\p{^Is_Ccc=023}', "");
Expect(1, 1472, '\P{Is_Ccc=023}', "");
Expect(0, 1472, '\P{^Is_Ccc=023}', "");
Error('\p{Canonical_Combining_Class:	-CCC24/a/}');
Error('\P{Canonical_Combining_Class:	-CCC24/a/}');
Expect(1, 1473, '\p{Canonical_Combining_Class=ccc24}', "");
Expect(0, 1473, '\p{^Canonical_Combining_Class=ccc24}', "");
Expect(0, 1473, '\P{Canonical_Combining_Class=ccc24}', "");
Expect(1, 1473, '\P{^Canonical_Combining_Class=ccc24}', "");
Expect(0, 1474, '\p{Canonical_Combining_Class=ccc24}', "");
Expect(1, 1474, '\p{^Canonical_Combining_Class=ccc24}', "");
Expect(1, 1474, '\P{Canonical_Combining_Class=ccc24}', "");
Expect(0, 1474, '\P{^Canonical_Combining_Class=ccc24}', "");
Expect(1, 1473, '\p{Canonical_Combining_Class=	CCC24}', "");
Expect(0, 1473, '\p{^Canonical_Combining_Class=	CCC24}', "");
Expect(0, 1473, '\P{Canonical_Combining_Class=	CCC24}', "");
Expect(1, 1473, '\P{^Canonical_Combining_Class=	CCC24}', "");
Expect(0, 1474, '\p{Canonical_Combining_Class=	CCC24}', "");
Expect(1, 1474, '\p{^Canonical_Combining_Class=	CCC24}', "");
Expect(1, 1474, '\P{Canonical_Combining_Class=	CCC24}', "");
Expect(0, 1474, '\P{^Canonical_Combining_Class=	CCC24}', "");
Error('\p{Ccc=:= -0000000024}');
Error('\P{Ccc=:= -0000000024}');
Expect(1, 1473, '\p{Ccc=024}', "");
Expect(0, 1473, '\p{^Ccc=024}', "");
Expect(0, 1473, '\P{Ccc=024}', "");
Expect(1, 1473, '\P{^Ccc=024}', "");
Expect(0, 1474, '\p{Ccc=024}', "");
Expect(1, 1474, '\p{^Ccc=024}', "");
Expect(1, 1474, '\P{Ccc=024}', "");
Expect(0, 1474, '\P{^Ccc=024}', "");
Error('\p{Is_Canonical_Combining_Class=_-ccc24:=}');
Error('\P{Is_Canonical_Combining_Class=_-ccc24:=}');
Expect(1, 1473, '\p{Is_Canonical_Combining_Class=ccc24}', "");
Expect(0, 1473, '\p{^Is_Canonical_Combining_Class=ccc24}', "");
Expect(0, 1473, '\P{Is_Canonical_Combining_Class=ccc24}', "");
Expect(1, 1473, '\P{^Is_Canonical_Combining_Class=ccc24}', "");
Expect(0, 1474, '\p{Is_Canonical_Combining_Class=ccc24}', "");
Expect(1, 1474, '\p{^Is_Canonical_Combining_Class=ccc24}', "");
Expect(1, 1474, '\P{Is_Canonical_Combining_Class=ccc24}', "");
Expect(0, 1474, '\P{^Is_Canonical_Combining_Class=ccc24}', "");
Expect(1, 1473, '\p{Is_Canonical_Combining_Class=	_CCC24}', "");
Expect(0, 1473, '\p{^Is_Canonical_Combining_Class=	_CCC24}', "");
Expect(0, 1473, '\P{Is_Canonical_Combining_Class=	_CCC24}', "");
Expect(1, 1473, '\P{^Is_Canonical_Combining_Class=	_CCC24}', "");
Expect(0, 1474, '\p{Is_Canonical_Combining_Class=	_CCC24}', "");
Expect(1, 1474, '\p{^Is_Canonical_Combining_Class=	_CCC24}', "");
Expect(1, 1474, '\P{Is_Canonical_Combining_Class=	_CCC24}', "");
Expect(0, 1474, '\P{^Is_Canonical_Combining_Class=	_CCC24}', "");
Error('\p{Is_Ccc: /a/--+0000024}');
Error('\P{Is_Ccc: /a/--+0000024}');
Expect(1, 1473, '\p{Is_Ccc=2_4}', "");
Expect(0, 1473, '\p{^Is_Ccc=2_4}', "");
Expect(0, 1473, '\P{Is_Ccc=2_4}', "");
Expect(1, 1473, '\P{^Is_Ccc=2_4}', "");
Expect(0, 1474, '\p{Is_Ccc=2_4}', "");
Expect(1, 1474, '\p{^Is_Ccc=2_4}', "");
Expect(1, 1474, '\P{Is_Ccc=2_4}', "");
Expect(0, 1474, '\P{^Is_Ccc=2_4}', "");
Error('\p{Canonical_Combining_Class=:=_ ccc25}');
Error('\P{Canonical_Combining_Class=:=_ ccc25}');
Expect(1, 1474, '\p{Canonical_Combining_Class=ccc25}', "");
Expect(0, 1474, '\p{^Canonical_Combining_Class=ccc25}', "");
Expect(0, 1474, '\P{Canonical_Combining_Class=ccc25}', "");
Expect(1, 1474, '\P{^Canonical_Combining_Class=ccc25}', "");
Expect(0, 1475, '\p{Canonical_Combining_Class=ccc25}', "");
Expect(1, 1475, '\p{^Canonical_Combining_Class=ccc25}', "");
Expect(1, 1475, '\P{Canonical_Combining_Class=ccc25}', "");
Expect(0, 1475, '\P{^Canonical_Combining_Class=ccc25}', "");
Expect(1, 1474, '\p{Canonical_Combining_Class=	ccc25}', "");
Expect(0, 1474, '\p{^Canonical_Combining_Class=	ccc25}', "");
Expect(0, 1474, '\P{Canonical_Combining_Class=	ccc25}', "");
Expect(1, 1474, '\P{^Canonical_Combining_Class=	ccc25}', "");
Expect(0, 1475, '\p{Canonical_Combining_Class=	ccc25}', "");
Expect(1, 1475, '\p{^Canonical_Combining_Class=	ccc25}', "");
Expect(1, 1475, '\P{Canonical_Combining_Class=	ccc25}', "");
Expect(0, 1475, '\P{^Canonical_Combining_Class=	ccc25}', "");
Error('\p{Ccc=	+00000000025:=}');
Error('\P{Ccc=	+00000000025:=}');
Expect(1, 1474, '\p{Ccc=00025}', "");
Expect(0, 1474, '\p{^Ccc=00025}', "");
Expect(0, 1474, '\P{Ccc=00025}', "");
Expect(1, 1474, '\P{^Ccc=00025}', "");
Expect(0, 1475, '\p{Ccc=00025}', "");
Expect(1, 1475, '\p{^Ccc=00025}', "");
Expect(1, 1475, '\P{Ccc=00025}', "");
Expect(0, 1475, '\P{^Ccc=00025}', "");
Error('\p{Is_Canonical_Combining_Class=	CCC25/a/}');
Error('\P{Is_Canonical_Combining_Class=	CCC25/a/}');
Expect(1, 1474, '\p{Is_Canonical_Combining_Class:   ccc25}', "");
Expect(0, 1474, '\p{^Is_Canonical_Combining_Class:   ccc25}', "");
Expect(0, 1474, '\P{Is_Canonical_Combining_Class:   ccc25}', "");
Expect(1, 1474, '\P{^Is_Canonical_Combining_Class:   ccc25}', "");
Expect(0, 1475, '\p{Is_Canonical_Combining_Class:   ccc25}', "");
Expect(1, 1475, '\p{^Is_Canonical_Combining_Class:   ccc25}', "");
Expect(1, 1475, '\P{Is_Canonical_Combining_Class:   ccc25}', "");
Expect(0, 1475, '\P{^Is_Canonical_Combining_Class:   ccc25}', "");
Expect(1, 1474, '\p{Is_Canonical_Combining_Class=	 CCC25}', "");
Expect(0, 1474, '\p{^Is_Canonical_Combining_Class=	 CCC25}', "");
Expect(0, 1474, '\P{Is_Canonical_Combining_Class=	 CCC25}', "");
Expect(1, 1474, '\P{^Is_Canonical_Combining_Class=	 CCC25}', "");
Expect(0, 1475, '\p{Is_Canonical_Combining_Class=	 CCC25}', "");
Expect(1, 1475, '\p{^Is_Canonical_Combining_Class=	 CCC25}', "");
Expect(1, 1475, '\P{Is_Canonical_Combining_Class=	 CCC25}', "");
Expect(0, 1475, '\P{^Is_Canonical_Combining_Class=	 CCC25}', "");
Error('\p{Is_Ccc= /a/2_5}');
Error('\P{Is_Ccc= /a/2_5}');
Expect(1, 1474, '\p{Is_Ccc=00000000025}', "");
Expect(0, 1474, '\p{^Is_Ccc=00000000025}', "");
Expect(0, 1474, '\P{Is_Ccc=00000000025}', "");
Expect(1, 1474, '\P{^Is_Ccc=00000000025}', "");
Expect(0, 1475, '\p{Is_Ccc=00000000025}', "");
Expect(1, 1475, '\p{^Is_Ccc=00000000025}', "");
Expect(1, 1475, '\P{Is_Ccc=00000000025}', "");
Expect(0, 1475, '\P{^Is_Ccc=00000000025}', "");
Error('\p{Canonical_Combining_Class=/a/ 	CCC26}');
Error('\P{Canonical_Combining_Class=/a/ 	CCC26}');
Expect(1, 64286, '\p{Canonical_Combining_Class=ccc26}', "");
Expect(0, 64286, '\p{^Canonical_Combining_Class=ccc26}', "");
Expect(0, 64286, '\P{Canonical_Combining_Class=ccc26}', "");
Expect(1, 64286, '\P{^Canonical_Combining_Class=ccc26}', "");
Expect(0, 64287, '\p{Canonical_Combining_Class=ccc26}', "");
Expect(1, 64287, '\p{^Canonical_Combining_Class=ccc26}', "");
Expect(1, 64287, '\P{Canonical_Combining_Class=ccc26}', "");
Expect(0, 64287, '\P{^Canonical_Combining_Class=ccc26}', "");
Expect(1, 64286, '\p{Canonical_Combining_Class=-	CCC26}', "");
Expect(0, 64286, '\p{^Canonical_Combining_Class=-	CCC26}', "");
Expect(0, 64286, '\P{Canonical_Combining_Class=-	CCC26}', "");
Expect(1, 64286, '\P{^Canonical_Combining_Class=-	CCC26}', "");
Expect(0, 64287, '\p{Canonical_Combining_Class=-	CCC26}', "");
Expect(1, 64287, '\p{^Canonical_Combining_Class=-	CCC26}', "");
Expect(1, 64287, '\P{Canonical_Combining_Class=-	CCC26}', "");
Expect(0, 64287, '\P{^Canonical_Combining_Class=-	CCC26}', "");
Error('\p{Ccc=_+00002_6:=}');
Error('\P{Ccc=_+00002_6:=}');
Expect(1, 64286, '\p{Ccc=000026}', "");
Expect(0, 64286, '\p{^Ccc=000026}', "");
Expect(0, 64286, '\P{Ccc=000026}', "");
Expect(1, 64286, '\P{^Ccc=000026}', "");
Expect(0, 64287, '\p{Ccc=000026}', "");
Expect(1, 64287, '\p{^Ccc=000026}', "");
Expect(1, 64287, '\P{Ccc=000026}', "");
Expect(0, 64287, '\P{^Ccc=000026}', "");
Error('\p{Is_Canonical_Combining_Class= /a/CCC26}');
Error('\P{Is_Canonical_Combining_Class= /a/CCC26}');
Expect(1, 64286, '\p{Is_Canonical_Combining_Class=ccc26}', "");
Expect(0, 64286, '\p{^Is_Canonical_Combining_Class=ccc26}', "");
Expect(0, 64286, '\P{Is_Canonical_Combining_Class=ccc26}', "");
Expect(1, 64286, '\P{^Is_Canonical_Combining_Class=ccc26}', "");
Expect(0, 64287, '\p{Is_Canonical_Combining_Class=ccc26}', "");
Expect(1, 64287, '\p{^Is_Canonical_Combining_Class=ccc26}', "");
Expect(1, 64287, '\P{Is_Canonical_Combining_Class=ccc26}', "");
Expect(0, 64287, '\P{^Is_Canonical_Combining_Class=ccc26}', "");
Expect(1, 64286, '\p{Is_Canonical_Combining_Class:   _ CCC26}', "");
Expect(0, 64286, '\p{^Is_Canonical_Combining_Class:   _ CCC26}', "");
Expect(0, 64286, '\P{Is_Canonical_Combining_Class:   _ CCC26}', "");
Expect(1, 64286, '\P{^Is_Canonical_Combining_Class:   _ CCC26}', "");
Expect(0, 64287, '\p{Is_Canonical_Combining_Class:   _ CCC26}', "");
Expect(1, 64287, '\p{^Is_Canonical_Combining_Class:   _ CCC26}', "");
Expect(1, 64287, '\P{Is_Canonical_Combining_Class:   _ CCC26}', "");
Expect(0, 64287, '\P{^Is_Canonical_Combining_Class:   _ CCC26}', "");
Error('\p{Is_Ccc=0000_0002_6:=}');
Error('\P{Is_Ccc=0000_0002_6:=}');
Expect(1, 64286, '\p{Is_Ccc=0000_0002_6}', "");
Expect(0, 64286, '\p{^Is_Ccc=0000_0002_6}', "");
Expect(0, 64286, '\P{Is_Ccc=0000_0002_6}', "");
Expect(1, 64286, '\P{^Is_Ccc=0000_0002_6}', "");
Expect(0, 64287, '\p{Is_Ccc=0000_0002_6}', "");
Expect(1, 64287, '\p{^Is_Ccc=0000_0002_6}', "");
Expect(1, 64287, '\P{Is_Ccc=0000_0002_6}', "");
Expect(0, 64287, '\P{^Is_Ccc=0000_0002_6}', "");
Error('\p{Canonical_Combining_Class=/a/CCC27}');
Error('\P{Canonical_Combining_Class=/a/CCC27}');
Expect(1, 2288, '\p{Canonical_Combining_Class=ccc27}', "");
Expect(0, 2288, '\p{^Canonical_Combining_Class=ccc27}', "");
Expect(0, 2288, '\P{Canonical_Combining_Class=ccc27}', "");
Expect(1, 2288, '\P{^Canonical_Combining_Class=ccc27}', "");
Expect(0, 2289, '\p{Canonical_Combining_Class=ccc27}', "");
Expect(1, 2289, '\p{^Canonical_Combining_Class=ccc27}', "");
Expect(1, 2289, '\P{Canonical_Combining_Class=ccc27}', "");
Expect(0, 2289, '\P{^Canonical_Combining_Class=ccc27}', "");
Expect(1, 2288, '\p{Canonical_Combining_Class=-	CCC27}', "");
Expect(0, 2288, '\p{^Canonical_Combining_Class=-	CCC27}', "");
Expect(0, 2288, '\P{Canonical_Combining_Class=-	CCC27}', "");
Expect(1, 2288, '\P{^Canonical_Combining_Class=-	CCC27}', "");
Expect(0, 2289, '\p{Canonical_Combining_Class=-	CCC27}', "");
Expect(1, 2289, '\p{^Canonical_Combining_Class=-	CCC27}', "");
Expect(1, 2289, '\P{Canonical_Combining_Class=-	CCC27}', "");
Expect(0, 2289, '\P{^Canonical_Combining_Class=-	CCC27}', "");
Error('\p{Ccc=	/a/+0000000002_7}');
Error('\P{Ccc=	/a/+0000000002_7}');
Expect(1, 2288, '\p{Ccc=27}', "");
Expect(0, 2288, '\p{^Ccc=27}', "");
Expect(0, 2288, '\P{Ccc=27}', "");
Expect(1, 2288, '\P{^Ccc=27}', "");
Expect(0, 2289, '\p{Ccc=27}', "");
Expect(1, 2289, '\p{^Ccc=27}', "");
Expect(1, 2289, '\P{Ccc=27}', "");
Expect(0, 2289, '\P{^Ccc=27}', "");
Error('\p{Is_Canonical_Combining_Class=/a/ccc27}');
Error('\P{Is_Canonical_Combining_Class=/a/ccc27}');
Expect(1, 2288, '\p{Is_Canonical_Combining_Class=ccc27}', "");
Expect(0, 2288, '\p{^Is_Canonical_Combining_Class=ccc27}', "");
Expect(0, 2288, '\P{Is_Canonical_Combining_Class=ccc27}', "");
Expect(1, 2288, '\P{^Is_Canonical_Combining_Class=ccc27}', "");
Expect(0, 2289, '\p{Is_Canonical_Combining_Class=ccc27}', "");
Expect(1, 2289, '\p{^Is_Canonical_Combining_Class=ccc27}', "");
Expect(1, 2289, '\P{Is_Canonical_Combining_Class=ccc27}', "");
Expect(0, 2289, '\P{^Is_Canonical_Combining_Class=ccc27}', "");
Expect(1, 2288, '\p{Is_Canonical_Combining_Class=	-CCC27}', "");
Expect(0, 2288, '\p{^Is_Canonical_Combining_Class=	-CCC27}', "");
Expect(0, 2288, '\P{Is_Canonical_Combining_Class=	-CCC27}', "");
Expect(1, 2288, '\P{^Is_Canonical_Combining_Class=	-CCC27}', "");
Expect(0, 2289, '\p{Is_Canonical_Combining_Class=	-CCC27}', "");
Expect(1, 2289, '\p{^Is_Canonical_Combining_Class=	-CCC27}', "");
Expect(1, 2289, '\P{Is_Canonical_Combining_Class=	-CCC27}', "");
Expect(0, 2289, '\P{^Is_Canonical_Combining_Class=	-CCC27}', "");
Error('\p{Is_Ccc=--00000027:=}');
Error('\P{Is_Ccc=--00000027:=}');
Expect(1, 2288, '\p{Is_Ccc=02_7}', "");
Expect(0, 2288, '\p{^Is_Ccc=02_7}', "");
Expect(0, 2288, '\P{Is_Ccc=02_7}', "");
Expect(1, 2288, '\P{^Is_Ccc=02_7}', "");
Expect(0, 2289, '\p{Is_Ccc=02_7}', "");
Expect(1, 2289, '\p{^Is_Ccc=02_7}', "");
Expect(1, 2289, '\P{Is_Ccc=02_7}', "");
Expect(0, 2289, '\P{^Is_Ccc=02_7}', "");
Error('\p{Canonical_Combining_Class=:=-CCC28}');
Error('\P{Canonical_Combining_Class=:=-CCC28}');
Expect(1, 2289, '\p{Canonical_Combining_Class=ccc28}', "");
Expect(0, 2289, '\p{^Canonical_Combining_Class=ccc28}', "");
Expect(0, 2289, '\P{Canonical_Combining_Class=ccc28}', "");
Expect(1, 2289, '\P{^Canonical_Combining_Class=ccc28}', "");
Expect(0, 2290, '\p{Canonical_Combining_Class=ccc28}', "");
Expect(1, 2290, '\p{^Canonical_Combining_Class=ccc28}', "");
Expect(1, 2290, '\P{Canonical_Combining_Class=ccc28}', "");
Expect(0, 2290, '\P{^Canonical_Combining_Class=ccc28}', "");
Expect(1, 2289, '\p{Canonical_Combining_Class=-ccc28}', "");
Expect(0, 2289, '\p{^Canonical_Combining_Class=-ccc28}', "");
Expect(0, 2289, '\P{Canonical_Combining_Class=-ccc28}', "");
Expect(1, 2289, '\P{^Canonical_Combining_Class=-ccc28}', "");
Expect(0, 2290, '\p{Canonical_Combining_Class=-ccc28}', "");
Expect(1, 2290, '\p{^Canonical_Combining_Class=-ccc28}', "");
Expect(1, 2290, '\P{Canonical_Combining_Class=-ccc28}', "");
Expect(0, 2290, '\P{^Canonical_Combining_Class=-ccc28}', "");
Error('\p{Ccc=_	+0028/a/}');
Error('\P{Ccc=_	+0028/a/}');
Expect(1, 2289, '\p{Ccc:0_0_0_0_0_0_0_0_0_28}', "");
Expect(0, 2289, '\p{^Ccc:0_0_0_0_0_0_0_0_0_28}', "");
Expect(0, 2289, '\P{Ccc:0_0_0_0_0_0_0_0_0_28}', "");
Expect(1, 2289, '\P{^Ccc:0_0_0_0_0_0_0_0_0_28}', "");
Expect(0, 2290, '\p{Ccc:0_0_0_0_0_0_0_0_0_28}', "");
Expect(1, 2290, '\p{^Ccc:0_0_0_0_0_0_0_0_0_28}', "");
Expect(1, 2290, '\P{Ccc:0_0_0_0_0_0_0_0_0_28}', "");
Expect(0, 2290, '\P{^Ccc:0_0_0_0_0_0_0_0_0_28}', "");
Error('\p{Is_Canonical_Combining_Class=/a/_	CCC28}');
Error('\P{Is_Canonical_Combining_Class=/a/_	CCC28}');
Expect(1, 2289, '\p{Is_Canonical_Combining_Class=ccc28}', "");
Expect(0, 2289, '\p{^Is_Canonical_Combining_Class=ccc28}', "");
Expect(0, 2289, '\P{Is_Canonical_Combining_Class=ccc28}', "");
Expect(1, 2289, '\P{^Is_Canonical_Combining_Class=ccc28}', "");
Expect(0, 2290, '\p{Is_Canonical_Combining_Class=ccc28}', "");
Expect(1, 2290, '\p{^Is_Canonical_Combining_Class=ccc28}', "");
Expect(1, 2290, '\P{Is_Canonical_Combining_Class=ccc28}', "");
Expect(0, 2290, '\P{^Is_Canonical_Combining_Class=ccc28}', "");
Expect(1, 2289, '\p{Is_Canonical_Combining_Class= CCC28}', "");
Expect(0, 2289, '\p{^Is_Canonical_Combining_Class= CCC28}', "");
Expect(0, 2289, '\P{Is_Canonical_Combining_Class= CCC28}', "");
Expect(1, 2289, '\P{^Is_Canonical_Combining_Class= CCC28}', "");
Expect(0, 2290, '\p{Is_Canonical_Combining_Class= CCC28}', "");
Expect(1, 2290, '\p{^Is_Canonical_Combining_Class= CCC28}', "");
Expect(1, 2290, '\P{Is_Canonical_Combining_Class= CCC28}', "");
Expect(0, 2290, '\P{^Is_Canonical_Combining_Class= CCC28}', "");
Error('\p{Is_Ccc=	/a/28}');
Error('\P{Is_Ccc=	/a/28}');
Expect(1, 2289, '\p{Is_Ccc=+000000028}', "");
Expect(0, 2289, '\p{^Is_Ccc=+000000028}', "");
Expect(0, 2289, '\P{Is_Ccc=+000000028}', "");
Expect(1, 2289, '\P{^Is_Ccc=+000000028}', "");
Expect(0, 2290, '\p{Is_Ccc=+000000028}', "");
Expect(1, 2290, '\p{^Is_Ccc=+000000028}', "");
Expect(1, 2290, '\P{Is_Ccc=+000000028}', "");
Expect(0, 2290, '\P{^Is_Ccc=+000000028}', "");
Error('\p{Canonical_Combining_Class=/a/-CCC29}');
Error('\P{Canonical_Combining_Class=/a/-CCC29}');
Expect(1, 2290, '\p{Canonical_Combining_Class:ccc29}', "");
Expect(0, 2290, '\p{^Canonical_Combining_Class:ccc29}', "");
Expect(0, 2290, '\P{Canonical_Combining_Class:ccc29}', "");
Expect(1, 2290, '\P{^Canonical_Combining_Class:ccc29}', "");
Expect(0, 2291, '\p{Canonical_Combining_Class:ccc29}', "");
Expect(1, 2291, '\p{^Canonical_Combining_Class:ccc29}', "");
Expect(1, 2291, '\P{Canonical_Combining_Class:ccc29}', "");
Expect(0, 2291, '\P{^Canonical_Combining_Class:ccc29}', "");
Expect(1, 2290, '\p{Canonical_Combining_Class=	_CCC29}', "");
Expect(0, 2290, '\p{^Canonical_Combining_Class=	_CCC29}', "");
Expect(0, 2290, '\P{Canonical_Combining_Class=	_CCC29}', "");
Expect(1, 2290, '\P{^Canonical_Combining_Class=	_CCC29}', "");
Expect(0, 2291, '\p{Canonical_Combining_Class=	_CCC29}', "");
Expect(1, 2291, '\p{^Canonical_Combining_Class=	_CCC29}', "");
Expect(1, 2291, '\P{Canonical_Combining_Class=	_CCC29}', "");
Expect(0, 2291, '\P{^Canonical_Combining_Class=	_CCC29}', "");
Error('\p{Ccc=/a/0_0_0_0_0_029}');
Error('\P{Ccc=/a/0_0_0_0_0_029}');
Expect(1, 2290, '\p{Ccc=0_0_0_0_29}', "");
Expect(0, 2290, '\p{^Ccc=0_0_0_0_29}', "");
Expect(0, 2290, '\P{Ccc=0_0_0_0_29}', "");
Expect(1, 2290, '\P{^Ccc=0_0_0_0_29}', "");
Expect(0, 2291, '\p{Ccc=0_0_0_0_29}', "");
Expect(1, 2291, '\p{^Ccc=0_0_0_0_29}', "");
Expect(1, 2291, '\P{Ccc=0_0_0_0_29}', "");
Expect(0, 2291, '\P{^Ccc=0_0_0_0_29}', "");
Error('\p{Is_Canonical_Combining_Class=:= CCC29}');
Error('\P{Is_Canonical_Combining_Class=:= CCC29}');
Expect(1, 2290, '\p{Is_Canonical_Combining_Class: ccc29}', "");
Expect(0, 2290, '\p{^Is_Canonical_Combining_Class: ccc29}', "");
Expect(0, 2290, '\P{Is_Canonical_Combining_Class: ccc29}', "");
Expect(1, 2290, '\P{^Is_Canonical_Combining_Class: ccc29}', "");
Expect(0, 2291, '\p{Is_Canonical_Combining_Class: ccc29}', "");
Expect(1, 2291, '\p{^Is_Canonical_Combining_Class: ccc29}', "");
Expect(1, 2291, '\P{Is_Canonical_Combining_Class: ccc29}', "");
Expect(0, 2291, '\P{^Is_Canonical_Combining_Class: ccc29}', "");
Expect(1, 2290, '\p{Is_Canonical_Combining_Class=	 CCC29}', "");
Expect(0, 2290, '\p{^Is_Canonical_Combining_Class=	 CCC29}', "");
Expect(0, 2290, '\P{Is_Canonical_Combining_Class=	 CCC29}', "");
Expect(1, 2290, '\P{^Is_Canonical_Combining_Class=	 CCC29}', "");
Expect(0, 2291, '\p{Is_Canonical_Combining_Class=	 CCC29}', "");
Expect(1, 2291, '\p{^Is_Canonical_Combining_Class=	 CCC29}', "");
Expect(1, 2291, '\P{Is_Canonical_Combining_Class=	 CCC29}', "");
Expect(0, 2291, '\P{^Is_Canonical_Combining_Class=	 CCC29}', "");
Error('\p{Is_Ccc=_000029:=}');
Error('\P{Is_Ccc=_000029:=}');
Expect(1, 2290, '\p{Is_Ccc=0002_9}', "");
Expect(0, 2290, '\p{^Is_Ccc=0002_9}', "");
Expect(0, 2290, '\P{Is_Ccc=0002_9}', "");
Expect(1, 2290, '\P{^Is_Ccc=0002_9}', "");
Expect(0, 2291, '\p{Is_Ccc=0002_9}', "");
Expect(1, 2291, '\p{^Is_Ccc=0002_9}', "");
Expect(1, 2291, '\P{Is_Ccc=0002_9}', "");
Expect(0, 2291, '\P{^Is_Ccc=0002_9}', "");
Error('\p{Canonical_Combining_Class=:=ccc30}');
Error('\P{Canonical_Combining_Class=:=ccc30}');
Expect(1, 1614, '\p{Canonical_Combining_Class:	ccc30}', "");
Expect(0, 1614, '\p{^Canonical_Combining_Class:	ccc30}', "");
Expect(0, 1614, '\P{Canonical_Combining_Class:	ccc30}', "");
Expect(1, 1614, '\P{^Canonical_Combining_Class:	ccc30}', "");
Expect(0, 1615, '\p{Canonical_Combining_Class:	ccc30}', "");
Expect(1, 1615, '\p{^Canonical_Combining_Class:	ccc30}', "");
Expect(1, 1615, '\P{Canonical_Combining_Class:	ccc30}', "");
Expect(0, 1615, '\P{^Canonical_Combining_Class:	ccc30}', "");
Expect(1, 1614, '\p{Canonical_Combining_Class=	_CCC30}', "");
Expect(0, 1614, '\p{^Canonical_Combining_Class=	_CCC30}', "");
Expect(0, 1614, '\P{Canonical_Combining_Class=	_CCC30}', "");
Expect(1, 1614, '\P{^Canonical_Combining_Class=	_CCC30}', "");
Expect(0, 1615, '\p{Canonical_Combining_Class=	_CCC30}', "");
Expect(1, 1615, '\p{^Canonical_Combining_Class=	_CCC30}', "");
Expect(1, 1615, '\P{Canonical_Combining_Class=	_CCC30}', "");
Expect(0, 1615, '\P{^Canonical_Combining_Class=	_CCC30}', "");
Error('\p{Ccc= -0000000030:=}');
Error('\P{Ccc= -0000000030:=}');
Expect(1, 1614, '\p{Ccc=0000000030}', "");
Expect(0, 1614, '\p{^Ccc=0000000030}', "");
Expect(0, 1614, '\P{Ccc=0000000030}', "");
Expect(1, 1614, '\P{^Ccc=0000000030}', "");
Expect(0, 1615, '\p{Ccc=0000000030}', "");
Expect(1, 1615, '\p{^Ccc=0000000030}', "");
Expect(1, 1615, '\P{Ccc=0000000030}', "");
Expect(0, 1615, '\P{^Ccc=0000000030}', "");
Error('\p{Is_Canonical_Combining_Class=-/a/CCC30}');
Error('\P{Is_Canonical_Combining_Class=-/a/CCC30}');
Expect(1, 1614, '\p{Is_Canonical_Combining_Class:   ccc30}', "");
Expect(0, 1614, '\p{^Is_Canonical_Combining_Class:   ccc30}', "");
Expect(0, 1614, '\P{Is_Canonical_Combining_Class:   ccc30}', "");
Expect(1, 1614, '\P{^Is_Canonical_Combining_Class:   ccc30}', "");
Expect(0, 1615, '\p{Is_Canonical_Combining_Class:   ccc30}', "");
Expect(1, 1615, '\p{^Is_Canonical_Combining_Class:   ccc30}', "");
Expect(1, 1615, '\P{Is_Canonical_Combining_Class:   ccc30}', "");
Expect(0, 1615, '\P{^Is_Canonical_Combining_Class:   ccc30}', "");
Expect(1, 1614, '\p{Is_Canonical_Combining_Class= _ccc30}', "");
Expect(0, 1614, '\p{^Is_Canonical_Combining_Class= _ccc30}', "");
Expect(0, 1614, '\P{Is_Canonical_Combining_Class= _ccc30}', "");
Expect(1, 1614, '\P{^Is_Canonical_Combining_Class= _ccc30}', "");
Expect(0, 1615, '\p{Is_Canonical_Combining_Class= _ccc30}', "");
Expect(1, 1615, '\p{^Is_Canonical_Combining_Class= _ccc30}', "");
Expect(1, 1615, '\P{Is_Canonical_Combining_Class= _ccc30}', "");
Expect(0, 1615, '\P{^Is_Canonical_Combining_Class= _ccc30}', "");
Error('\p{Is_Ccc=-/a/000000030}');
Error('\P{Is_Ccc=-/a/000000030}');
Expect(1, 1614, '\p{Is_Ccc:03_0}', "");
Expect(0, 1614, '\p{^Is_Ccc:03_0}', "");
Expect(0, 1614, '\P{Is_Ccc:03_0}', "");
Expect(1, 1614, '\P{^Is_Ccc:03_0}', "");
Expect(0, 1615, '\p{Is_Ccc:03_0}', "");
Expect(1, 1615, '\p{^Is_Ccc:03_0}', "");
Expect(1, 1615, '\P{Is_Ccc:03_0}', "");
Expect(0, 1615, '\P{^Is_Ccc:03_0}', "");
Error('\p{Canonical_Combining_Class=-CCC31/a/}');
Error('\P{Canonical_Combining_Class=-CCC31/a/}');
Expect(1, 1615, '\p{Canonical_Combining_Class=ccc31}', "");
Expect(0, 1615, '\p{^Canonical_Combining_Class=ccc31}', "");
Expect(0, 1615, '\P{Canonical_Combining_Class=ccc31}', "");
Expect(1, 1615, '\P{^Canonical_Combining_Class=ccc31}', "");
Expect(0, 1616, '\p{Canonical_Combining_Class=ccc31}', "");
Expect(1, 1616, '\p{^Canonical_Combining_Class=ccc31}', "");
Expect(1, 1616, '\P{Canonical_Combining_Class=ccc31}', "");
Expect(0, 1616, '\P{^Canonical_Combining_Class=ccc31}', "");
Expect(1, 1615, '\p{Canonical_Combining_Class=-	CCC31}', "");
Expect(0, 1615, '\p{^Canonical_Combining_Class=-	CCC31}', "");
Expect(0, 1615, '\P{Canonical_Combining_Class=-	CCC31}', "");
Expect(1, 1615, '\P{^Canonical_Combining_Class=-	CCC31}', "");
Expect(0, 1616, '\p{Canonical_Combining_Class=-	CCC31}', "");
Expect(1, 1616, '\p{^Canonical_Combining_Class=-	CCC31}', "");
Expect(1, 1616, '\P{Canonical_Combining_Class=-	CCC31}', "");
Expect(0, 1616, '\P{^Canonical_Combining_Class=-	CCC31}', "");
Error('\p{Ccc: -0000031:=}');
Error('\P{Ccc: -0000031:=}');
Expect(1, 1615, '\p{Ccc=0031}', "");
Expect(0, 1615, '\p{^Ccc=0031}', "");
Expect(0, 1615, '\P{Ccc=0031}', "");
Expect(1, 1615, '\P{^Ccc=0031}', "");
Expect(0, 1616, '\p{Ccc=0031}', "");
Expect(1, 1616, '\p{^Ccc=0031}', "");
Expect(1, 1616, '\P{Ccc=0031}', "");
Expect(0, 1616, '\P{^Ccc=0031}', "");
Error('\p{Is_Canonical_Combining_Class=-	ccc31:=}');
Error('\P{Is_Canonical_Combining_Class=-	ccc31:=}');
Expect(1, 1615, '\p{Is_Canonical_Combining_Class=ccc31}', "");
Expect(0, 1615, '\p{^Is_Canonical_Combining_Class=ccc31}', "");
Expect(0, 1615, '\P{Is_Canonical_Combining_Class=ccc31}', "");
Expect(1, 1615, '\P{^Is_Canonical_Combining_Class=ccc31}', "");
Expect(0, 1616, '\p{Is_Canonical_Combining_Class=ccc31}', "");
Expect(1, 1616, '\p{^Is_Canonical_Combining_Class=ccc31}', "");
Expect(1, 1616, '\P{Is_Canonical_Combining_Class=ccc31}', "");
Expect(0, 1616, '\P{^Is_Canonical_Combining_Class=ccc31}', "");
Expect(1, 1615, '\p{Is_Canonical_Combining_Class=- CCC31}', "");
Expect(0, 1615, '\p{^Is_Canonical_Combining_Class=- CCC31}', "");
Expect(0, 1615, '\P{Is_Canonical_Combining_Class=- CCC31}', "");
Expect(1, 1615, '\P{^Is_Canonical_Combining_Class=- CCC31}', "");
Expect(0, 1616, '\p{Is_Canonical_Combining_Class=- CCC31}', "");
Expect(1, 1616, '\p{^Is_Canonical_Combining_Class=- CCC31}', "");
Expect(1, 1616, '\P{Is_Canonical_Combining_Class=- CCC31}', "");
Expect(0, 1616, '\P{^Is_Canonical_Combining_Class=- CCC31}', "");
Error('\p{Is_Ccc=:=0000031}');
Error('\P{Is_Ccc=:=0000031}');
Expect(1, 1615, '\p{Is_Ccc=000000031}', "");
Expect(0, 1615, '\p{^Is_Ccc=000000031}', "");
Expect(0, 1615, '\P{Is_Ccc=000000031}', "");
Expect(1, 1615, '\P{^Is_Ccc=000000031}', "");
Expect(0, 1616, '\p{Is_Ccc=000000031}', "");
Expect(1, 1616, '\p{^Is_Ccc=000000031}', "");
Expect(1, 1616, '\P{Is_Ccc=000000031}', "");
Expect(0, 1616, '\P{^Is_Ccc=000000031}', "");
Error('\p{Canonical_Combining_Class=		ccc32:=}');
Error('\P{Canonical_Combining_Class=		ccc32:=}');
Expect(1, 1616, '\p{Canonical_Combining_Class=ccc32}', "");
Expect(0, 1616, '\p{^Canonical_Combining_Class=ccc32}', "");
Expect(0, 1616, '\P{Canonical_Combining_Class=ccc32}', "");
Expect(1, 1616, '\P{^Canonical_Combining_Class=ccc32}', "");
Expect(0, 1617, '\p{Canonical_Combining_Class=ccc32}', "");
Expect(1, 1617, '\p{^Canonical_Combining_Class=ccc32}', "");
Expect(1, 1617, '\P{Canonical_Combining_Class=ccc32}', "");
Expect(0, 1617, '\P{^Canonical_Combining_Class=ccc32}', "");
Expect(1, 1616, '\p{Canonical_Combining_Class:     CCC32}', "");
Expect(0, 1616, '\p{^Canonical_Combining_Class:     CCC32}', "");
Expect(0, 1616, '\P{Canonical_Combining_Class:     CCC32}', "");
Expect(1, 1616, '\P{^Canonical_Combining_Class:     CCC32}', "");
Expect(0, 1617, '\p{Canonical_Combining_Class:     CCC32}', "");
Expect(1, 1617, '\p{^Canonical_Combining_Class:     CCC32}', "");
Expect(1, 1617, '\P{Canonical_Combining_Class:     CCC32}', "");
Expect(0, 1617, '\P{^Canonical_Combining_Class:     CCC32}', "");
Error('\p{Ccc=00000032:=}');
Error('\P{Ccc=00000032:=}');
Expect(1, 1616, '\p{Ccc=0_0_0_0_0_0_0032}', "");
Expect(0, 1616, '\p{^Ccc=0_0_0_0_0_0_0032}', "");
Expect(0, 1616, '\P{Ccc=0_0_0_0_0_0_0032}', "");
Expect(1, 1616, '\P{^Ccc=0_0_0_0_0_0_0032}', "");
Expect(0, 1617, '\p{Ccc=0_0_0_0_0_0_0032}', "");
Expect(1, 1617, '\p{^Ccc=0_0_0_0_0_0_0032}', "");
Expect(1, 1617, '\P{Ccc=0_0_0_0_0_0_0032}', "");
Expect(0, 1617, '\P{^Ccc=0_0_0_0_0_0_0032}', "");
Error('\p{Is_Canonical_Combining_Class= 	CCC32:=}');
Error('\P{Is_Canonical_Combining_Class= 	CCC32:=}');
Expect(1, 1616, '\p{Is_Canonical_Combining_Class=ccc32}', "");
Expect(0, 1616, '\p{^Is_Canonical_Combining_Class=ccc32}', "");
Expect(0, 1616, '\P{Is_Canonical_Combining_Class=ccc32}', "");
Expect(1, 1616, '\P{^Is_Canonical_Combining_Class=ccc32}', "");
Expect(0, 1617, '\p{Is_Canonical_Combining_Class=ccc32}', "");
Expect(1, 1617, '\p{^Is_Canonical_Combining_Class=ccc32}', "");
Expect(1, 1617, '\P{Is_Canonical_Combining_Class=ccc32}', "");
Expect(0, 1617, '\P{^Is_Canonical_Combining_Class=ccc32}', "");
Expect(1, 1616, '\p{Is_Canonical_Combining_Class= CCC32}', "");
Expect(0, 1616, '\p{^Is_Canonical_Combining_Class= CCC32}', "");
Expect(0, 1616, '\P{Is_Canonical_Combining_Class= CCC32}', "");
Expect(1, 1616, '\P{^Is_Canonical_Combining_Class= CCC32}', "");
Expect(0, 1617, '\p{Is_Canonical_Combining_Class= CCC32}', "");
Expect(1, 1617, '\p{^Is_Canonical_Combining_Class= CCC32}', "");
Expect(1, 1617, '\P{Is_Canonical_Combining_Class= CCC32}', "");
Expect(0, 1617, '\P{^Is_Canonical_Combining_Class= CCC32}', "");
Error('\p{Is_Ccc=	-0000032/a/}');
Error('\P{Is_Ccc=	-0000032/a/}');
Expect(1, 1616, '\p{Is_Ccc=00000032}', "");
Expect(0, 1616, '\p{^Is_Ccc=00000032}', "");
Expect(0, 1616, '\P{Is_Ccc=00000032}', "");
Expect(1, 1616, '\P{^Is_Ccc=00000032}', "");
Expect(0, 1617, '\p{Is_Ccc=00000032}', "");
Expect(1, 1617, '\p{^Is_Ccc=00000032}', "");
Expect(1, 1617, '\P{Is_Ccc=00000032}', "");
Expect(0, 1617, '\P{^Is_Ccc=00000032}', "");
Error('\p{Canonical_Combining_Class= /a/CCC33}');
Error('\P{Canonical_Combining_Class= /a/CCC33}');
Expect(1, 1617, '\p{Canonical_Combining_Class=ccc33}', "");
Expect(0, 1617, '\p{^Canonical_Combining_Class=ccc33}', "");
Expect(0, 1617, '\P{Canonical_Combining_Class=ccc33}', "");
Expect(1, 1617, '\P{^Canonical_Combining_Class=ccc33}', "");
Expect(0, 1618, '\p{Canonical_Combining_Class=ccc33}', "");
Expect(1, 1618, '\p{^Canonical_Combining_Class=ccc33}', "");
Expect(1, 1618, '\P{Canonical_Combining_Class=ccc33}', "");
Expect(0, 1618, '\P{^Canonical_Combining_Class=ccc33}', "");
Expect(1, 1617, '\p{Canonical_Combining_Class=-CCC33}', "");
Expect(0, 1617, '\p{^Canonical_Combining_Class=-CCC33}', "");
Expect(0, 1617, '\P{Canonical_Combining_Class=-CCC33}', "");
Expect(1, 1617, '\P{^Canonical_Combining_Class=-CCC33}', "");
Expect(0, 1618, '\p{Canonical_Combining_Class=-CCC33}', "");
Expect(1, 1618, '\p{^Canonical_Combining_Class=-CCC33}', "");
Expect(1, 1618, '\P{Canonical_Combining_Class=-CCC33}', "");
Expect(0, 1618, '\P{^Canonical_Combining_Class=-CCC33}', "");
Error('\p{Ccc=	 00000033:=}');
Error('\P{Ccc=	 00000033:=}');
Expect(1, 1617, '\p{Ccc=000000033}', "");
Expect(0, 1617, '\p{^Ccc=000000033}', "");
Expect(0, 1617, '\P{Ccc=000000033}', "");
Expect(1, 1617, '\P{^Ccc=000000033}', "");
Expect(0, 1618, '\p{Ccc=000000033}', "");
Expect(1, 1618, '\p{^Ccc=000000033}', "");
Expect(1, 1618, '\P{Ccc=000000033}', "");
Expect(0, 1618, '\P{^Ccc=000000033}', "");
Error('\p{Is_Canonical_Combining_Class: --CCC33:=}');
Error('\P{Is_Canonical_Combining_Class: --CCC33:=}');
Expect(1, 1617, '\p{Is_Canonical_Combining_Class=ccc33}', "");
Expect(0, 1617, '\p{^Is_Canonical_Combining_Class=ccc33}', "");
Expect(0, 1617, '\P{Is_Canonical_Combining_Class=ccc33}', "");
Expect(1, 1617, '\P{^Is_Canonical_Combining_Class=ccc33}', "");
Expect(0, 1618, '\p{Is_Canonical_Combining_Class=ccc33}', "");
Expect(1, 1618, '\p{^Is_Canonical_Combining_Class=ccc33}', "");
Expect(1, 1618, '\P{Is_Canonical_Combining_Class=ccc33}', "");
Expect(0, 1618, '\P{^Is_Canonical_Combining_Class=ccc33}', "");
Expect(1, 1617, '\p{Is_Canonical_Combining_Class=__CCC33}', "");
Expect(0, 1617, '\p{^Is_Canonical_Combining_Class=__CCC33}', "");
Expect(0, 1617, '\P{Is_Canonical_Combining_Class=__CCC33}', "");
Expect(1, 1617, '\P{^Is_Canonical_Combining_Class=__CCC33}', "");
Expect(0, 1618, '\p{Is_Canonical_Combining_Class=__CCC33}', "");
Expect(1, 1618, '\p{^Is_Canonical_Combining_Class=__CCC33}', "");
Expect(1, 1618, '\P{Is_Canonical_Combining_Class=__CCC33}', "");
Expect(0, 1618, '\P{^Is_Canonical_Combining_Class=__CCC33}', "");
Error('\p{Is_Ccc:   _/a/00033}');
Error('\P{Is_Ccc:   _/a/00033}');
Expect(1, 1617, '\p{Is_Ccc=0_0_0_0_0_00033}', "");
Expect(0, 1617, '\p{^Is_Ccc=0_0_0_0_0_00033}', "");
Expect(0, 1617, '\P{Is_Ccc=0_0_0_0_0_00033}', "");
Expect(1, 1617, '\P{^Is_Ccc=0_0_0_0_0_00033}', "");
Expect(0, 1618, '\p{Is_Ccc=0_0_0_0_0_00033}', "");
Expect(1, 1618, '\p{^Is_Ccc=0_0_0_0_0_00033}', "");
Expect(1, 1618, '\P{Is_Ccc=0_0_0_0_0_00033}', "");
Expect(0, 1618, '\P{^Is_Ccc=0_0_0_0_0_00033}', "");
Error('\p{Canonical_Combining_Class=_/a/ccc34}');
Error('\P{Canonical_Combining_Class=_/a/ccc34}');
Expect(1, 1618, '\p{Canonical_Combining_Class=ccc34}', "");
Expect(0, 1618, '\p{^Canonical_Combining_Class=ccc34}', "");
Expect(0, 1618, '\P{Canonical_Combining_Class=ccc34}', "");
Expect(1, 1618, '\P{^Canonical_Combining_Class=ccc34}', "");
Expect(0, 1619, '\p{Canonical_Combining_Class=ccc34}', "");
Expect(1, 1619, '\p{^Canonical_Combining_Class=ccc34}', "");
Expect(1, 1619, '\P{Canonical_Combining_Class=ccc34}', "");
Expect(0, 1619, '\P{^Canonical_Combining_Class=ccc34}', "");
Expect(1, 1618, '\p{Canonical_Combining_Class=	ccc34}', "");
Expect(0, 1618, '\p{^Canonical_Combining_Class=	ccc34}', "");
Expect(0, 1618, '\P{Canonical_Combining_Class=	ccc34}', "");
Expect(1, 1618, '\P{^Canonical_Combining_Class=	ccc34}', "");
Expect(0, 1619, '\p{Canonical_Combining_Class=	ccc34}', "");
Expect(1, 1619, '\p{^Canonical_Combining_Class=	ccc34}', "");
Expect(1, 1619, '\P{Canonical_Combining_Class=	ccc34}', "");
Expect(0, 1619, '\P{^Canonical_Combining_Class=	ccc34}', "");
Error('\p{Ccc=/a/00000_00003_4}');
Error('\P{Ccc=/a/00000_00003_4}');
Expect(1, 1618, '\p{Ccc=3_4}', "");
Expect(0, 1618, '\p{^Ccc=3_4}', "");
Expect(0, 1618, '\P{Ccc=3_4}', "");
Expect(1, 1618, '\P{^Ccc=3_4}', "");
Expect(0, 1619, '\p{Ccc=3_4}', "");
Expect(1, 1619, '\p{^Ccc=3_4}', "");
Expect(1, 1619, '\P{Ccc=3_4}', "");
Expect(0, 1619, '\P{^Ccc=3_4}', "");
Error('\p{Is_Canonical_Combining_Class=/a/-_ccc34}');
Error('\P{Is_Canonical_Combining_Class=/a/-_ccc34}');
Expect(1, 1618, '\p{Is_Canonical_Combining_Class=ccc34}', "");
Expect(0, 1618, '\p{^Is_Canonical_Combining_Class=ccc34}', "");
Expect(0, 1618, '\P{Is_Canonical_Combining_Class=ccc34}', "");
Expect(1, 1618, '\P{^Is_Canonical_Combining_Class=ccc34}', "");
Expect(0, 1619, '\p{Is_Canonical_Combining_Class=ccc34}', "");
Expect(1, 1619, '\p{^Is_Canonical_Combining_Class=ccc34}', "");
Expect(1, 1619, '\P{Is_Canonical_Combining_Class=ccc34}', "");
Expect(0, 1619, '\P{^Is_Canonical_Combining_Class=ccc34}', "");
Expect(1, 1618, '\p{Is_Canonical_Combining_Class=-ccc34}', "");
Expect(0, 1618, '\p{^Is_Canonical_Combining_Class=-ccc34}', "");
Expect(0, 1618, '\P{Is_Canonical_Combining_Class=-ccc34}', "");
Expect(1, 1618, '\P{^Is_Canonical_Combining_Class=-ccc34}', "");
Expect(0, 1619, '\p{Is_Canonical_Combining_Class=-ccc34}', "");
Expect(1, 1619, '\p{^Is_Canonical_Combining_Class=-ccc34}', "");
Expect(1, 1619, '\P{Is_Canonical_Combining_Class=-ccc34}', "");
Expect(0, 1619, '\P{^Is_Canonical_Combining_Class=-ccc34}', "");
Error('\p{Is_Ccc=-_03_4:=}');
Error('\P{Is_Ccc=-_03_4:=}');
Expect(1, 1618, '\p{Is_Ccc=+00034}', "");
Expect(0, 1618, '\p{^Is_Ccc=+00034}', "");
Expect(0, 1618, '\P{Is_Ccc=+00034}', "");
Expect(1, 1618, '\P{^Is_Ccc=+00034}', "");
Expect(0, 1619, '\p{Is_Ccc=+00034}', "");
Expect(1, 1619, '\p{^Is_Ccc=+00034}', "");
Expect(1, 1619, '\P{Is_Ccc=+00034}', "");
Expect(0, 1619, '\P{^Is_Ccc=+00034}', "");
Error('\p{Canonical_Combining_Class=-CCC35:=}');
Error('\P{Canonical_Combining_Class=-CCC35:=}');
Expect(1, 1648, '\p{Canonical_Combining_Class=ccc35}', "");
Expect(0, 1648, '\p{^Canonical_Combining_Class=ccc35}', "");
Expect(0, 1648, '\P{Canonical_Combining_Class=ccc35}', "");
Expect(1, 1648, '\P{^Canonical_Combining_Class=ccc35}', "");
Expect(0, 1649, '\p{Canonical_Combining_Class=ccc35}', "");
Expect(1, 1649, '\p{^Canonical_Combining_Class=ccc35}', "");
Expect(1, 1649, '\P{Canonical_Combining_Class=ccc35}', "");
Expect(0, 1649, '\P{^Canonical_Combining_Class=ccc35}', "");
Expect(1, 1648, '\p{Canonical_Combining_Class=-CCC35}', "");
Expect(0, 1648, '\p{^Canonical_Combining_Class=-CCC35}', "");
Expect(0, 1648, '\P{Canonical_Combining_Class=-CCC35}', "");
Expect(1, 1648, '\P{^Canonical_Combining_Class=-CCC35}', "");
Expect(0, 1649, '\p{Canonical_Combining_Class=-CCC35}', "");
Expect(1, 1649, '\p{^Canonical_Combining_Class=-CCC35}', "");
Expect(1, 1649, '\P{Canonical_Combining_Class=-CCC35}', "");
Expect(0, 1649, '\P{^Canonical_Combining_Class=-CCC35}', "");
Error('\p{Ccc= 00000035/a/}');
Error('\P{Ccc= 00000035/a/}');
Expect(1, 1648, '\p{Ccc=0_0_0_0_0_0_035}', "");
Expect(0, 1648, '\p{^Ccc=0_0_0_0_0_0_035}', "");
Expect(0, 1648, '\P{Ccc=0_0_0_0_0_0_035}', "");
Expect(1, 1648, '\P{^Ccc=0_0_0_0_0_0_035}', "");
Expect(0, 1649, '\p{Ccc=0_0_0_0_0_0_035}', "");
Expect(1, 1649, '\p{^Ccc=0_0_0_0_0_0_035}', "");
Expect(1, 1649, '\P{Ccc=0_0_0_0_0_0_035}', "");
Expect(0, 1649, '\P{^Ccc=0_0_0_0_0_0_035}', "");
Error('\p{Is_Canonical_Combining_Class:   -/a/ccc35}');
Error('\P{Is_Canonical_Combining_Class:   -/a/ccc35}');
Expect(1, 1648, '\p{Is_Canonical_Combining_Class=ccc35}', "");
Expect(0, 1648, '\p{^Is_Canonical_Combining_Class=ccc35}', "");
Expect(0, 1648, '\P{Is_Canonical_Combining_Class=ccc35}', "");
Expect(1, 1648, '\P{^Is_Canonical_Combining_Class=ccc35}', "");
Expect(0, 1649, '\p{Is_Canonical_Combining_Class=ccc35}', "");
Expect(1, 1649, '\p{^Is_Canonical_Combining_Class=ccc35}', "");
Expect(1, 1649, '\P{Is_Canonical_Combining_Class=ccc35}', "");
Expect(0, 1649, '\P{^Is_Canonical_Combining_Class=ccc35}', "");
Expect(1, 1648, '\p{Is_Canonical_Combining_Class= -CCC35}', "");
Expect(0, 1648, '\p{^Is_Canonical_Combining_Class= -CCC35}', "");
Expect(0, 1648, '\P{Is_Canonical_Combining_Class= -CCC35}', "");
Expect(1, 1648, '\P{^Is_Canonical_Combining_Class= -CCC35}', "");
Expect(0, 1649, '\p{Is_Canonical_Combining_Class= -CCC35}', "");
Expect(1, 1649, '\p{^Is_Canonical_Combining_Class= -CCC35}', "");
Expect(1, 1649, '\P{Is_Canonical_Combining_Class= -CCC35}', "");
Expect(0, 1649, '\P{^Is_Canonical_Combining_Class= -CCC35}', "");
Error('\p{Is_Ccc:  000_000_35/a/}');
Error('\P{Is_Ccc:  000_000_35/a/}');
Expect(1, 1648, '\p{Is_Ccc=0000035}', "");
Expect(0, 1648, '\p{^Is_Ccc=0000035}', "");
Expect(0, 1648, '\P{Is_Ccc=0000035}', "");
Expect(1, 1648, '\P{^Is_Ccc=0000035}', "");
Expect(0, 1649, '\p{Is_Ccc=0000035}', "");
Expect(1, 1649, '\p{^Is_Ccc=0000035}', "");
Expect(1, 1649, '\P{Is_Ccc=0000035}', "");
Expect(0, 1649, '\P{^Is_Ccc=0000035}', "");
Error('\p{Canonical_Combining_Class=:=ccc36}');
Error('\P{Canonical_Combining_Class=:=ccc36}');
Expect(1, 1809, '\p{Canonical_Combining_Class=ccc36}', "");
Expect(0, 1809, '\p{^Canonical_Combining_Class=ccc36}', "");
Expect(0, 1809, '\P{Canonical_Combining_Class=ccc36}', "");
Expect(1, 1809, '\P{^Canonical_Combining_Class=ccc36}', "");
Expect(0, 1810, '\p{Canonical_Combining_Class=ccc36}', "");
Expect(1, 1810, '\p{^Canonical_Combining_Class=ccc36}', "");
Expect(1, 1810, '\P{Canonical_Combining_Class=ccc36}', "");
Expect(0, 1810, '\P{^Canonical_Combining_Class=ccc36}', "");
Expect(1, 1809, '\p{Canonical_Combining_Class=-CCC36}', "");
Expect(0, 1809, '\p{^Canonical_Combining_Class=-CCC36}', "");
Expect(0, 1809, '\P{Canonical_Combining_Class=-CCC36}', "");
Expect(1, 1809, '\P{^Canonical_Combining_Class=-CCC36}', "");
Expect(0, 1810, '\p{Canonical_Combining_Class=-CCC36}', "");
Expect(1, 1810, '\p{^Canonical_Combining_Class=-CCC36}', "");
Expect(1, 1810, '\P{Canonical_Combining_Class=-CCC36}', "");
Expect(0, 1810, '\P{^Canonical_Combining_Class=-CCC36}', "");
Error('\p{Ccc=	/a/00036}');
Error('\P{Ccc=	/a/00036}');
Expect(1, 1809, '\p{Ccc=+0_0_0_0_0_0_0_036}', "");
Expect(0, 1809, '\p{^Ccc=+0_0_0_0_0_0_0_036}', "");
Expect(0, 1809, '\P{Ccc=+0_0_0_0_0_0_0_036}', "");
Expect(1, 1809, '\P{^Ccc=+0_0_0_0_0_0_0_036}', "");
Expect(0, 1810, '\p{Ccc=+0_0_0_0_0_0_0_036}', "");
Expect(1, 1810, '\p{^Ccc=+0_0_0_0_0_0_0_036}', "");
Expect(1, 1810, '\P{Ccc=+0_0_0_0_0_0_0_036}', "");
Expect(0, 1810, '\P{^Ccc=+0_0_0_0_0_0_0_036}', "");
Error('\p{Is_Canonical_Combining_Class=/a/CCC36}');
Error('\P{Is_Canonical_Combining_Class=/a/CCC36}');
Expect(1, 1809, '\p{Is_Canonical_Combining_Class=ccc36}', "");
Expect(0, 1809, '\p{^Is_Canonical_Combining_Class=ccc36}', "");
Expect(0, 1809, '\P{Is_Canonical_Combining_Class=ccc36}', "");
Expect(1, 1809, '\P{^Is_Canonical_Combining_Class=ccc36}', "");
Expect(0, 1810, '\p{Is_Canonical_Combining_Class=ccc36}', "");
Expect(1, 1810, '\p{^Is_Canonical_Combining_Class=ccc36}', "");
Expect(1, 1810, '\P{Is_Canonical_Combining_Class=ccc36}', "");
Expect(0, 1810, '\P{^Is_Canonical_Combining_Class=ccc36}', "");
Expect(1, 1809, '\p{Is_Canonical_Combining_Class=_-ccc36}', "");
Expect(0, 1809, '\p{^Is_Canonical_Combining_Class=_-ccc36}', "");
Expect(0, 1809, '\P{Is_Canonical_Combining_Class=_-ccc36}', "");
Expect(1, 1809, '\P{^Is_Canonical_Combining_Class=_-ccc36}', "");
Expect(0, 1810, '\p{Is_Canonical_Combining_Class=_-ccc36}', "");
Expect(1, 1810, '\p{^Is_Canonical_Combining_Class=_-ccc36}', "");
Expect(1, 1810, '\P{Is_Canonical_Combining_Class=_-ccc36}', "");
Expect(0, 1810, '\P{^Is_Canonical_Combining_Class=_-ccc36}', "");
Error('\p{Is_Ccc=/a/-+0_0_0_0_0_0_0036}');
Error('\P{Is_Ccc=/a/-+0_0_0_0_0_0_0036}');
Expect(1, 1809, '\p{Is_Ccc=000036}', "");
Expect(0, 1809, '\p{^Is_Ccc=000036}', "");
Expect(0, 1809, '\P{Is_Ccc=000036}', "");
Expect(1, 1809, '\P{^Is_Ccc=000036}', "");
Expect(0, 1810, '\p{Is_Ccc=000036}', "");
Expect(1, 1810, '\p{^Is_Ccc=000036}', "");
Expect(1, 1810, '\P{Is_Ccc=000036}', "");
Expect(0, 1810, '\P{^Is_Ccc=000036}', "");
Error('\p{Canonical_Combining_Class=:=__CCC84}');
Error('\P{Canonical_Combining_Class=:=__CCC84}');
Expect(1, 3157, '\p{Canonical_Combining_Class=ccc84}', "");
Expect(0, 3157, '\p{^Canonical_Combining_Class=ccc84}', "");
Expect(0, 3157, '\P{Canonical_Combining_Class=ccc84}', "");
Expect(1, 3157, '\P{^Canonical_Combining_Class=ccc84}', "");
Expect(0, 3158, '\p{Canonical_Combining_Class=ccc84}', "");
Expect(1, 3158, '\p{^Canonical_Combining_Class=ccc84}', "");
Expect(1, 3158, '\P{Canonical_Combining_Class=ccc84}', "");
Expect(0, 3158, '\P{^Canonical_Combining_Class=ccc84}', "");
Expect(1, 3157, '\p{Canonical_Combining_Class=-CCC84}', "");
Expect(0, 3157, '\p{^Canonical_Combining_Class=-CCC84}', "");
Expect(0, 3157, '\P{Canonical_Combining_Class=-CCC84}', "");
Expect(1, 3157, '\P{^Canonical_Combining_Class=-CCC84}', "");
Expect(0, 3158, '\p{Canonical_Combining_Class=-CCC84}', "");
Expect(1, 3158, '\p{^Canonical_Combining_Class=-CCC84}', "");
Expect(1, 3158, '\P{Canonical_Combining_Class=-CCC84}', "");
Expect(0, 3158, '\P{^Canonical_Combining_Class=-CCC84}', "");
Error('\p{Ccc:  :=0_0_0_0_084}');
Error('\P{Ccc:  :=0_0_0_0_084}');
Expect(1, 3157, '\p{Ccc: 00_08_4}', "");
Expect(0, 3157, '\p{^Ccc: 00_08_4}', "");
Expect(0, 3157, '\P{Ccc: 00_08_4}', "");
Expect(1, 3157, '\P{^Ccc: 00_08_4}', "");
Expect(0, 3158, '\p{Ccc: 00_08_4}', "");
Expect(1, 3158, '\p{^Ccc: 00_08_4}', "");
Expect(1, 3158, '\P{Ccc: 00_08_4}', "");
Expect(0, 3158, '\P{^Ccc: 00_08_4}', "");
Error('\p{Is_Canonical_Combining_Class=/a/--CCC84}');
Error('\P{Is_Canonical_Combining_Class=/a/--CCC84}');
Expect(1, 3157, '\p{Is_Canonical_Combining_Class=ccc84}', "");
Expect(0, 3157, '\p{^Is_Canonical_Combining_Class=ccc84}', "");
Expect(0, 3157, '\P{Is_Canonical_Combining_Class=ccc84}', "");
Expect(1, 3157, '\P{^Is_Canonical_Combining_Class=ccc84}', "");
Expect(0, 3158, '\p{Is_Canonical_Combining_Class=ccc84}', "");
Expect(1, 3158, '\p{^Is_Canonical_Combining_Class=ccc84}', "");
Expect(1, 3158, '\P{Is_Canonical_Combining_Class=ccc84}', "");
Expect(0, 3158, '\P{^Is_Canonical_Combining_Class=ccc84}', "");
Expect(1, 3157, '\p{Is_Canonical_Combining_Class=CCC84}', "");
Expect(0, 3157, '\p{^Is_Canonical_Combining_Class=CCC84}', "");
Expect(0, 3157, '\P{Is_Canonical_Combining_Class=CCC84}', "");
Expect(1, 3157, '\P{^Is_Canonical_Combining_Class=CCC84}', "");
Expect(0, 3158, '\p{Is_Canonical_Combining_Class=CCC84}', "");
Expect(1, 3158, '\p{^Is_Canonical_Combining_Class=CCC84}', "");
Expect(1, 3158, '\P{Is_Canonical_Combining_Class=CCC84}', "");
Expect(0, 3158, '\P{^Is_Canonical_Combining_Class=CCC84}', "");
Error('\p{Is_Ccc= :=0008_4}');
Error('\P{Is_Ccc= :=0008_4}');
Expect(1, 3157, '\p{Is_Ccc=08_4}', "");
Expect(0, 3157, '\p{^Is_Ccc=08_4}', "");
Expect(0, 3157, '\P{Is_Ccc=08_4}', "");
Expect(1, 3157, '\P{^Is_Ccc=08_4}', "");
Expect(0, 3158, '\p{Is_Ccc=08_4}', "");
Expect(1, 3158, '\p{^Is_Ccc=08_4}', "");
Expect(1, 3158, '\P{Is_Ccc=08_4}', "");
Expect(0, 3158, '\P{^Is_Ccc=08_4}', "");
Error('\p{Canonical_Combining_Class=--CCC91/a/}');
Error('\P{Canonical_Combining_Class=--CCC91/a/}');
Expect(1, 3158, '\p{Canonical_Combining_Class=ccc91}', "");
Expect(0, 3158, '\p{^Canonical_Combining_Class=ccc91}', "");
Expect(0, 3158, '\P{Canonical_Combining_Class=ccc91}', "");
Expect(1, 3158, '\P{^Canonical_Combining_Class=ccc91}', "");
Expect(0, 3159, '\p{Canonical_Combining_Class=ccc91}', "");
Expect(1, 3159, '\p{^Canonical_Combining_Class=ccc91}', "");
Expect(1, 3159, '\P{Canonical_Combining_Class=ccc91}', "");
Expect(0, 3159, '\P{^Canonical_Combining_Class=ccc91}', "");
Expect(1, 3158, '\p{Canonical_Combining_Class= _CCC91}', "");
Expect(0, 3158, '\p{^Canonical_Combining_Class= _CCC91}', "");
Expect(0, 3158, '\P{Canonical_Combining_Class= _CCC91}', "");
Expect(1, 3158, '\P{^Canonical_Combining_Class= _CCC91}', "");
Expect(0, 3159, '\p{Canonical_Combining_Class= _CCC91}', "");
Expect(1, 3159, '\p{^Canonical_Combining_Class= _CCC91}', "");
Expect(1, 3159, '\P{Canonical_Combining_Class= _CCC91}', "");
Expect(0, 3159, '\P{^Canonical_Combining_Class= _CCC91}', "");
Error('\p{Ccc=- +0009_1:=}');
Error('\P{Ccc=- +0009_1:=}');
Expect(1, 3158, '\p{Ccc:	000_009_1}', "");
Expect(0, 3158, '\p{^Ccc:	000_009_1}', "");
Expect(0, 3158, '\P{Ccc:	000_009_1}', "");
Expect(1, 3158, '\P{^Ccc:	000_009_1}', "");
Expect(0, 3159, '\p{Ccc:	000_009_1}', "");
Expect(1, 3159, '\p{^Ccc:	000_009_1}', "");
Expect(1, 3159, '\P{Ccc:	000_009_1}', "");
Expect(0, 3159, '\P{^Ccc:	000_009_1}', "");
Error('\p{Is_Canonical_Combining_Class=		CCC91/a/}');
Error('\P{Is_Canonical_Combining_Class=		CCC91/a/}');
Expect(1, 3158, '\p{Is_Canonical_Combining_Class=ccc91}', "");
Expect(0, 3158, '\p{^Is_Canonical_Combining_Class=ccc91}', "");
Expect(0, 3158, '\P{Is_Canonical_Combining_Class=ccc91}', "");
Expect(1, 3158, '\P{^Is_Canonical_Combining_Class=ccc91}', "");
Expect(0, 3159, '\p{Is_Canonical_Combining_Class=ccc91}', "");
Expect(1, 3159, '\p{^Is_Canonical_Combining_Class=ccc91}', "");
Expect(1, 3159, '\P{Is_Canonical_Combining_Class=ccc91}', "");
Expect(0, 3159, '\P{^Is_Canonical_Combining_Class=ccc91}', "");
Expect(1, 3158, '\p{Is_Canonical_Combining_Class=	 CCC91}', "");
Expect(0, 3158, '\p{^Is_Canonical_Combining_Class=	 CCC91}', "");
Expect(0, 3158, '\P{Is_Canonical_Combining_Class=	 CCC91}', "");
Expect(1, 3158, '\P{^Is_Canonical_Combining_Class=	 CCC91}', "");
Expect(0, 3159, '\p{Is_Canonical_Combining_Class=	 CCC91}', "");
Expect(1, 3159, '\p{^Is_Canonical_Combining_Class=	 CCC91}', "");
Expect(1, 3159, '\P{Is_Canonical_Combining_Class=	 CCC91}', "");
Expect(0, 3159, '\P{^Is_Canonical_Combining_Class=	 CCC91}', "");
Error('\p{Is_Ccc=:= 00_09_1}');
Error('\P{Is_Ccc=:= 00_09_1}');
Expect(1, 3158, '\p{Is_Ccc=+0000000091}', "");
Expect(0, 3158, '\p{^Is_Ccc=+0000000091}', "");
Expect(0, 3158, '\P{Is_Ccc=+0000000091}', "");
Expect(1, 3158, '\P{^Is_Ccc=+0000000091}', "");
Expect(0, 3159, '\p{Is_Ccc=+0000000091}', "");
Expect(1, 3159, '\p{^Is_Ccc=+0000000091}', "");
Expect(1, 3159, '\P{Is_Ccc=+0000000091}', "");
Expect(0, 3159, '\P{^Is_Ccc=+0000000091}', "");
Error('\p{Canonical_Combining_Class=:= Double_ABOVE}');
Error('\P{Canonical_Combining_Class=:= Double_ABOVE}');
Expect(1, 7629, '\p{Canonical_Combining_Class: doubleabove}', "");
Expect(0, 7629, '\p{^Canonical_Combining_Class: doubleabove}', "");
Expect(0, 7629, '\P{Canonical_Combining_Class: doubleabove}', "");
Expect(1, 7629, '\P{^Canonical_Combining_Class: doubleabove}', "");
Expect(0, 7630, '\p{Canonical_Combining_Class: doubleabove}', "");
Expect(1, 7630, '\p{^Canonical_Combining_Class: doubleabove}', "");
Expect(1, 7630, '\P{Canonical_Combining_Class: doubleabove}', "");
Expect(0, 7630, '\P{^Canonical_Combining_Class: doubleabove}', "");
Expect(1, 7629, '\p{Canonical_Combining_Class=__double_ABOVE}', "");
Expect(0, 7629, '\p{^Canonical_Combining_Class=__double_ABOVE}', "");
Expect(0, 7629, '\P{Canonical_Combining_Class=__double_ABOVE}', "");
Expect(1, 7629, '\P{^Canonical_Combining_Class=__double_ABOVE}', "");
Expect(0, 7630, '\p{Canonical_Combining_Class=__double_ABOVE}', "");
Expect(1, 7630, '\p{^Canonical_Combining_Class=__double_ABOVE}', "");
Expect(1, 7630, '\P{Canonical_Combining_Class=__double_ABOVE}', "");
Expect(0, 7630, '\P{^Canonical_Combining_Class=__double_ABOVE}', "");
Error('\p{Ccc= 	da:=}');
Error('\P{Ccc= 	da:=}');
Expect(1, 7629, '\p{Ccc=da}', "");
Expect(0, 7629, '\p{^Ccc=da}', "");
Expect(0, 7629, '\P{Ccc=da}', "");
Expect(1, 7629, '\P{^Ccc=da}', "");
Expect(0, 7630, '\p{Ccc=da}', "");
Expect(1, 7630, '\p{^Ccc=da}', "");
Expect(1, 7630, '\P{Ccc=da}', "");
Expect(0, 7630, '\P{^Ccc=da}', "");
Expect(1, 7629, '\p{Ccc=	 DA}', "");
Expect(0, 7629, '\p{^Ccc=	 DA}', "");
Expect(0, 7629, '\P{Ccc=	 DA}', "");
Expect(1, 7629, '\P{^Ccc=	 DA}', "");
Expect(0, 7630, '\p{Ccc=	 DA}', "");
Expect(1, 7630, '\p{^Ccc=	 DA}', "");
Expect(1, 7630, '\P{Ccc=	 DA}', "");
Expect(0, 7630, '\P{^Ccc=	 DA}', "");
Error('\p{Is_Canonical_Combining_Class=:= 0_0_0_0_0_0_2_34}');
Error('\P{Is_Canonical_Combining_Class=:= 0_0_0_0_0_0_2_34}');
Expect(1, 7629, '\p{Is_Canonical_Combining_Class=00000000234}', "");
Expect(0, 7629, '\p{^Is_Canonical_Combining_Class=00000000234}', "");
Expect(0, 7629, '\P{Is_Canonical_Combining_Class=00000000234}', "");
Expect(1, 7629, '\P{^Is_Canonical_Combining_Class=00000000234}', "");
Expect(0, 7630, '\p{Is_Canonical_Combining_Class=00000000234}', "");
Expect(1, 7630, '\p{^Is_Canonical_Combining_Class=00000000234}', "");
Expect(1, 7630, '\P{Is_Canonical_Combining_Class=00000000234}', "");
Expect(0, 7630, '\P{^Is_Canonical_Combining_Class=00000000234}', "");
Error('\p{Is_Ccc=_double_ABOVE/a/}');
Error('\P{Is_Ccc=_double_ABOVE/a/}');
Expect(1, 7629, '\p{Is_Ccc=doubleabove}', "");
Expect(0, 7629, '\p{^Is_Ccc=doubleabove}', "");
Expect(0, 7629, '\P{Is_Ccc=doubleabove}', "");
Expect(1, 7629, '\P{^Is_Ccc=doubleabove}', "");
Expect(0, 7630, '\p{Is_Ccc=doubleabove}', "");
Expect(1, 7630, '\p{^Is_Ccc=doubleabove}', "");
Expect(1, 7630, '\P{Is_Ccc=doubleabove}', "");
Expect(0, 7630, '\P{^Is_Ccc=doubleabove}', "");
Expect(1, 7629, '\p{Is_Ccc: Double_Above}', "");
Expect(0, 7629, '\p{^Is_Ccc: Double_Above}', "");
Expect(0, 7629, '\P{Is_Ccc: Double_Above}', "");
Expect(1, 7629, '\P{^Is_Ccc: Double_Above}', "");
Expect(0, 7630, '\p{Is_Ccc: Double_Above}', "");
Expect(1, 7630, '\p{^Is_Ccc: Double_Above}', "");
Expect(1, 7630, '\P{Is_Ccc: Double_Above}', "");
Expect(0, 7630, '\P{^Is_Ccc: Double_Above}', "");
Error('\p{Canonical_Combining_Class=_Double_Below/a/}');
Error('\P{Canonical_Combining_Class=_Double_Below/a/}');
Expect(1, 7676, '\p{Canonical_Combining_Class: doublebelow}', "");
Expect(0, 7676, '\p{^Canonical_Combining_Class: doublebelow}', "");
Expect(0, 7676, '\P{Canonical_Combining_Class: doublebelow}', "");
Expect(1, 7676, '\P{^Canonical_Combining_Class: doublebelow}', "");
Expect(0, 7677, '\p{Canonical_Combining_Class: doublebelow}', "");
Expect(1, 7677, '\p{^Canonical_Combining_Class: doublebelow}', "");
Expect(1, 7677, '\P{Canonical_Combining_Class: doublebelow}', "");
Expect(0, 7677, '\P{^Canonical_Combining_Class: doublebelow}', "");
Expect(1, 7676, '\p{Canonical_Combining_Class=-Double_BELOW}', "");
Expect(0, 7676, '\p{^Canonical_Combining_Class=-Double_BELOW}', "");
Expect(0, 7676, '\P{Canonical_Combining_Class=-Double_BELOW}', "");
Expect(1, 7676, '\P{^Canonical_Combining_Class=-Double_BELOW}', "");
Expect(0, 7677, '\p{Canonical_Combining_Class=-Double_BELOW}', "");
Expect(1, 7677, '\p{^Canonical_Combining_Class=-Double_BELOW}', "");
Expect(1, 7677, '\P{Canonical_Combining_Class=-Double_BELOW}', "");
Expect(0, 7677, '\P{^Canonical_Combining_Class=-Double_BELOW}', "");
Error('\p{Ccc= -DB:=}');
Error('\P{Ccc= -DB:=}');
Expect(1, 7676, '\p{Ccc=db}', "");
Expect(0, 7676, '\p{^Ccc=db}', "");
Expect(0, 7676, '\P{Ccc=db}', "");
Expect(1, 7676, '\P{^Ccc=db}', "");
Expect(0, 7677, '\p{Ccc=db}', "");
Expect(1, 7677, '\p{^Ccc=db}', "");
Expect(1, 7677, '\P{Ccc=db}', "");
Expect(0, 7677, '\P{^Ccc=db}', "");
Expect(1, 7676, '\p{Ccc= DB}', "");
Expect(0, 7676, '\p{^Ccc= DB}', "");
Expect(0, 7676, '\P{Ccc= DB}', "");
Expect(1, 7676, '\P{^Ccc= DB}', "");
Expect(0, 7677, '\p{Ccc= DB}', "");
Expect(1, 7677, '\p{^Ccc= DB}', "");
Expect(1, 7677, '\P{Ccc= DB}', "");
Expect(0, 7677, '\P{^Ccc= DB}', "");
Error('\p{Is_Canonical_Combining_Class:   /a/_+00000000233}');
Error('\P{Is_Canonical_Combining_Class:   /a/_+00000000233}');
Expect(1, 7676, '\p{Is_Canonical_Combining_Class=0233}', "");
Expect(0, 7676, '\p{^Is_Canonical_Combining_Class=0233}', "");
Expect(0, 7676, '\P{Is_Canonical_Combining_Class=0233}', "");
Expect(1, 7676, '\P{^Is_Canonical_Combining_Class=0233}', "");
Expect(0, 7677, '\p{Is_Canonical_Combining_Class=0233}', "");
Expect(1, 7677, '\p{^Is_Canonical_Combining_Class=0233}', "");
Expect(1, 7677, '\P{Is_Canonical_Combining_Class=0233}', "");
Expect(0, 7677, '\P{^Is_Canonical_Combining_Class=0233}', "");
Error('\p{Is_Ccc:   -_Double_Below/a/}');
Error('\P{Is_Ccc:   -_Double_Below/a/}');
Expect(1, 7676, '\p{Is_Ccc:doublebelow}', "");
Expect(0, 7676, '\p{^Is_Ccc:doublebelow}', "");
Expect(0, 7676, '\P{Is_Ccc:doublebelow}', "");
Expect(1, 7676, '\P{^Is_Ccc:doublebelow}', "");
Expect(0, 7677, '\p{Is_Ccc:doublebelow}', "");
Expect(1, 7677, '\p{^Is_Ccc:doublebelow}', "");
Expect(1, 7677, '\P{Is_Ccc:doublebelow}', "");
Expect(0, 7677, '\P{^Is_Ccc:doublebelow}', "");
Expect(1, 7676, '\p{Is_Ccc=	double_Below}', "");
Expect(0, 7676, '\p{^Is_Ccc=	double_Below}', "");
Expect(0, 7676, '\P{Is_Ccc=	double_Below}', "");
Expect(1, 7676, '\P{^Is_Ccc=	double_Below}', "");
Expect(0, 7677, '\p{Is_Ccc=	double_Below}', "");
Expect(1, 7677, '\p{^Is_Ccc=	double_Below}', "");
Expect(1, 7677, '\P{Is_Ccc=	double_Below}', "");
Expect(0, 7677, '\P{^Is_Ccc=	double_Below}', "");
Error('\p{Canonical_Combining_Class=-/a/Iota_subscript}');
Error('\P{Canonical_Combining_Class=-/a/Iota_subscript}');
Expect(1, 837, '\p{Canonical_Combining_Class=iotasubscript}', "");
Expect(0, 837, '\p{^Canonical_Combining_Class=iotasubscript}', "");
Expect(0, 837, '\P{Canonical_Combining_Class=iotasubscript}', "");
Expect(1, 837, '\P{^Canonical_Combining_Class=iotasubscript}', "");
Expect(0, 838, '\p{Canonical_Combining_Class=iotasubscript}', "");
Expect(1, 838, '\p{^Canonical_Combining_Class=iotasubscript}', "");
Expect(1, 838, '\P{Canonical_Combining_Class=iotasubscript}', "");
Expect(0, 838, '\P{^Canonical_Combining_Class=iotasubscript}', "");
Expect(1, 837, '\p{Canonical_Combining_Class=-Iota_Subscript}', "");
Expect(0, 837, '\p{^Canonical_Combining_Class=-Iota_Subscript}', "");
Expect(0, 837, '\P{Canonical_Combining_Class=-Iota_Subscript}', "");
Expect(1, 837, '\P{^Canonical_Combining_Class=-Iota_Subscript}', "");
Expect(0, 838, '\p{Canonical_Combining_Class=-Iota_Subscript}', "");
Expect(1, 838, '\p{^Canonical_Combining_Class=-Iota_Subscript}', "");
Expect(1, 838, '\P{Canonical_Combining_Class=-Iota_Subscript}', "");
Expect(0, 838, '\P{^Canonical_Combining_Class=-Iota_Subscript}', "");
Error('\p{Ccc=	is/a/}');
Error('\P{Ccc=	is/a/}');
Expect(1, 837, '\p{Ccc=is}', "");
Expect(0, 837, '\p{^Ccc=is}', "");
Expect(0, 837, '\P{Ccc=is}', "");
Expect(1, 837, '\P{^Ccc=is}', "");
Expect(0, 838, '\p{Ccc=is}', "");
Expect(1, 838, '\p{^Ccc=is}', "");
Expect(1, 838, '\P{Ccc=is}', "");
Expect(0, 838, '\P{^Ccc=is}', "");
Expect(1, 837, '\p{Ccc:    	IS}', "");
Expect(0, 837, '\p{^Ccc:    	IS}', "");
Expect(0, 837, '\P{Ccc:    	IS}', "");
Expect(1, 837, '\P{^Ccc:    	IS}', "");
Expect(0, 838, '\p{Ccc:    	IS}', "");
Expect(1, 838, '\p{^Ccc:    	IS}', "");
Expect(1, 838, '\P{Ccc:    	IS}', "");
Expect(0, 838, '\P{^Ccc:    	IS}', "");
Error('\p{Is_Canonical_Combining_Class=-000024_0:=}');
Error('\P{Is_Canonical_Combining_Class=-000024_0:=}');
Expect(1, 837, '\p{Is_Canonical_Combining_Class=0000240}', "");
Expect(0, 837, '\p{^Is_Canonical_Combining_Class=0000240}', "");
Expect(0, 837, '\P{Is_Canonical_Combining_Class=0000240}', "");
Expect(1, 837, '\P{^Is_Canonical_Combining_Class=0000240}', "");
Expect(0, 838, '\p{Is_Canonical_Combining_Class=0000240}', "");
Expect(1, 838, '\p{^Is_Canonical_Combining_Class=0000240}', "");
Expect(1, 838, '\P{Is_Canonical_Combining_Class=0000240}', "");
Expect(0, 838, '\P{^Is_Canonical_Combining_Class=0000240}', "");
Error('\p{Is_Ccc=	:=iota_Subscript}');
Error('\P{Is_Ccc=	:=iota_Subscript}');
Expect(1, 837, '\p{Is_Ccc=iotasubscript}', "");
Expect(0, 837, '\p{^Is_Ccc=iotasubscript}', "");
Expect(0, 837, '\P{Is_Ccc=iotasubscript}', "");
Expect(1, 837, '\P{^Is_Ccc=iotasubscript}', "");
Expect(0, 838, '\p{Is_Ccc=iotasubscript}', "");
Expect(1, 838, '\p{^Is_Ccc=iotasubscript}', "");
Expect(1, 838, '\P{Is_Ccc=iotasubscript}', "");
Expect(0, 838, '\P{^Is_Ccc=iotasubscript}', "");
Expect(1, 837, '\p{Is_Ccc=	-Iota_Subscript}', "");
Expect(0, 837, '\p{^Is_Ccc=	-Iota_Subscript}', "");
Expect(0, 837, '\P{Is_Ccc=	-Iota_Subscript}', "");
Expect(1, 837, '\P{^Is_Ccc=	-Iota_Subscript}', "");
Expect(0, 838, '\p{Is_Ccc=	-Iota_Subscript}', "");
Expect(1, 838, '\p{^Is_Ccc=	-Iota_Subscript}', "");
Expect(1, 838, '\P{Is_Ccc=	-Iota_Subscript}', "");
Expect(0, 838, '\P{^Is_Ccc=	-Iota_Subscript}', "");
Error('\p{Canonical_Combining_Class=	_KANA_voicing:=}');
Error('\P{Canonical_Combining_Class=	_KANA_voicing:=}');
Expect(1, 12442, '\p{Canonical_Combining_Class=kanavoicing}', "");
Expect(0, 12442, '\p{^Canonical_Combining_Class=kanavoicing}', "");
Expect(0, 12442, '\P{Canonical_Combining_Class=kanavoicing}', "");
Expect(1, 12442, '\P{^Canonical_Combining_Class=kanavoicing}', "");
Expect(0, 12443, '\p{Canonical_Combining_Class=kanavoicing}', "");
Expect(1, 12443, '\p{^Canonical_Combining_Class=kanavoicing}', "");
Expect(1, 12443, '\P{Canonical_Combining_Class=kanavoicing}', "");
Expect(0, 12443, '\P{^Canonical_Combining_Class=kanavoicing}', "");
Expect(1, 12442, '\p{Canonical_Combining_Class=--Kana_Voicing}', "");
Expect(0, 12442, '\p{^Canonical_Combining_Class=--Kana_Voicing}', "");
Expect(0, 12442, '\P{Canonical_Combining_Class=--Kana_Voicing}', "");
Expect(1, 12442, '\P{^Canonical_Combining_Class=--Kana_Voicing}', "");
Expect(0, 12443, '\p{Canonical_Combining_Class=--Kana_Voicing}', "");
Expect(1, 12443, '\p{^Canonical_Combining_Class=--Kana_Voicing}', "");
Expect(1, 12443, '\P{Canonical_Combining_Class=--Kana_Voicing}', "");
Expect(0, 12443, '\P{^Canonical_Combining_Class=--Kana_Voicing}', "");
Error('\p{Ccc=-:=KV}');
Error('\P{Ccc=-:=KV}');
Expect(1, 12442, '\p{Ccc=kv}', "");
Expect(0, 12442, '\p{^Ccc=kv}', "");
Expect(0, 12442, '\P{Ccc=kv}', "");
Expect(1, 12442, '\P{^Ccc=kv}', "");
Expect(0, 12443, '\p{Ccc=kv}', "");
Expect(1, 12443, '\p{^Ccc=kv}', "");
Expect(1, 12443, '\P{Ccc=kv}', "");
Expect(0, 12443, '\P{^Ccc=kv}', "");
Expect(1, 12442, '\p{Ccc=_KV}', "");
Expect(0, 12442, '\p{^Ccc=_KV}', "");
Expect(0, 12442, '\P{Ccc=_KV}', "");
Expect(1, 12442, '\P{^Ccc=_KV}', "");
Expect(0, 12443, '\p{Ccc=_KV}', "");
Expect(1, 12443, '\p{^Ccc=_KV}', "");
Expect(1, 12443, '\P{Ccc=_KV}', "");
Expect(0, 12443, '\P{^Ccc=_KV}', "");
Error('\p{Is_Canonical_Combining_Class=	 0008:=}');
Error('\P{Is_Canonical_Combining_Class=	 0008:=}');
Expect(1, 12442, '\p{Is_Canonical_Combining_Class=008}', "");
Expect(0, 12442, '\p{^Is_Canonical_Combining_Class=008}', "");
Expect(0, 12442, '\P{Is_Canonical_Combining_Class=008}', "");
Expect(1, 12442, '\P{^Is_Canonical_Combining_Class=008}', "");
Expect(0, 12443, '\p{Is_Canonical_Combining_Class=008}', "");
Expect(1, 12443, '\p{^Is_Canonical_Combining_Class=008}', "");
Expect(1, 12443, '\P{Is_Canonical_Combining_Class=008}', "");
Expect(0, 12443, '\P{^Is_Canonical_Combining_Class=008}', "");
Error('\p{Is_Ccc=/a/kana_VOICING}');
Error('\P{Is_Ccc=/a/kana_VOICING}');
Expect(1, 12442, '\p{Is_Ccc=kanavoicing}', "");
Expect(0, 12442, '\p{^Is_Ccc=kanavoicing}', "");
Expect(0, 12442, '\P{Is_Ccc=kanavoicing}', "");
Expect(1, 12442, '\P{^Is_Ccc=kanavoicing}', "");
Expect(0, 12443, '\p{Is_Ccc=kanavoicing}', "");
Expect(1, 12443, '\p{^Is_Ccc=kanavoicing}', "");
Expect(1, 12443, '\P{Is_Ccc=kanavoicing}', "");
Expect(0, 12443, '\P{^Is_Ccc=kanavoicing}', "");
Expect(1, 12442, '\p{Is_Ccc= Kana_Voicing}', "");
Expect(0, 12442, '\p{^Is_Ccc= Kana_Voicing}', "");
Expect(0, 12442, '\P{Is_Ccc= Kana_Voicing}', "");
Expect(1, 12442, '\P{^Is_Ccc= Kana_Voicing}', "");
Expect(0, 12443, '\p{Is_Ccc= Kana_Voicing}', "");
Expect(1, 12443, '\p{^Is_Ccc= Kana_Voicing}', "");
Expect(1, 12443, '\P{Is_Ccc= Kana_Voicing}', "");
Expect(0, 12443, '\P{^Is_Ccc= Kana_Voicing}', "");
Error('\p{Canonical_Combining_Class=/a/	Left}');
Error('\P{Canonical_Combining_Class=/a/	Left}');
Expect(1, 12335, '\p{Canonical_Combining_Class=left}', "");
Expect(0, 12335, '\p{^Canonical_Combining_Class=left}', "");
Expect(0, 12335, '\P{Canonical_Combining_Class=left}', "");
Expect(1, 12335, '\P{^Canonical_Combining_Class=left}', "");
Expect(0, 12336, '\p{Canonical_Combining_Class=left}', "");
Expect(1, 12336, '\p{^Canonical_Combining_Class=left}', "");
Expect(1, 12336, '\P{Canonical_Combining_Class=left}', "");
Expect(0, 12336, '\P{^Canonical_Combining_Class=left}', "");
Expect(1, 12335, '\p{Canonical_Combining_Class=Left}', "");
Expect(0, 12335, '\p{^Canonical_Combining_Class=Left}', "");
Expect(0, 12335, '\P{Canonical_Combining_Class=Left}', "");
Expect(1, 12335, '\P{^Canonical_Combining_Class=Left}', "");
Expect(0, 12336, '\p{Canonical_Combining_Class=Left}', "");
Expect(1, 12336, '\p{^Canonical_Combining_Class=Left}', "");
Expect(1, 12336, '\P{Canonical_Combining_Class=Left}', "");
Expect(0, 12336, '\P{^Canonical_Combining_Class=Left}', "");
Error('\p{Ccc:	-/a/L}');
Error('\P{Ccc:	-/a/L}');
Expect(1, 12335, '\p{Ccc=l}', "");
Expect(0, 12335, '\p{^Ccc=l}', "");
Expect(0, 12335, '\P{Ccc=l}', "");
Expect(1, 12335, '\P{^Ccc=l}', "");
Expect(0, 12336, '\p{Ccc=l}', "");
Expect(1, 12336, '\p{^Ccc=l}', "");
Expect(1, 12336, '\P{Ccc=l}', "");
Expect(0, 12336, '\P{^Ccc=l}', "");
Expect(1, 12335, '\p{Ccc=_ L}', "");
Expect(0, 12335, '\p{^Ccc=_ L}', "");
Expect(0, 12335, '\P{Ccc=_ L}', "");
Expect(1, 12335, '\P{^Ccc=_ L}', "");
Expect(0, 12336, '\p{Ccc=_ L}', "");
Expect(1, 12336, '\p{^Ccc=_ L}', "");
Expect(1, 12336, '\P{Ccc=_ L}', "");
Expect(0, 12336, '\P{^Ccc=_ L}', "");
Error('\p{Is_Canonical_Combining_Class=:=0_2_24}');
Error('\P{Is_Canonical_Combining_Class=:=0_2_24}');
Expect(1, 12335, '\p{Is_Canonical_Combining_Class=000224}', "");
Expect(0, 12335, '\p{^Is_Canonical_Combining_Class=000224}', "");
Expect(0, 12335, '\P{Is_Canonical_Combining_Class=000224}', "");
Expect(1, 12335, '\P{^Is_Canonical_Combining_Class=000224}', "");
Expect(0, 12336, '\p{Is_Canonical_Combining_Class=000224}', "");
Expect(1, 12336, '\p{^Is_Canonical_Combining_Class=000224}', "");
Expect(1, 12336, '\P{Is_Canonical_Combining_Class=000224}', "");
Expect(0, 12336, '\P{^Is_Canonical_Combining_Class=000224}', "");
Error('\p{Is_Ccc=_/a/Left}');
Error('\P{Is_Ccc=_/a/Left}');
Expect(1, 12335, '\p{Is_Ccc=left}', "");
Expect(0, 12335, '\p{^Is_Ccc=left}', "");
Expect(0, 12335, '\P{Is_Ccc=left}', "");
Expect(1, 12335, '\P{^Is_Ccc=left}', "");
Expect(0, 12336, '\p{Is_Ccc=left}', "");
Expect(1, 12336, '\p{^Is_Ccc=left}', "");
Expect(1, 12336, '\P{Is_Ccc=left}', "");
Expect(0, 12336, '\P{^Is_Ccc=left}', "");
Expect(1, 12335, '\p{Is_Ccc=- Left}', "");
Expect(0, 12335, '\p{^Is_Ccc=- Left}', "");
Expect(0, 12335, '\P{Is_Ccc=- Left}', "");
Expect(1, 12335, '\P{^Is_Ccc=- Left}', "");
Expect(0, 12336, '\p{Is_Ccc=- Left}', "");
Expect(1, 12336, '\p{^Is_Ccc=- Left}', "");
Expect(1, 12336, '\P{Is_Ccc=- Left}', "");
Expect(0, 12336, '\P{^Is_Ccc=- Left}', "");
Error('\p{Canonical_Combining_Class=:=NUKTA}');
Error('\P{Canonical_Combining_Class=:=NUKTA}');
Expect(1, 71351, '\p{Canonical_Combining_Class=nukta}', "");
Expect(0, 71351, '\p{^Canonical_Combining_Class=nukta}', "");
Expect(0, 71351, '\P{Canonical_Combining_Class=nukta}', "");
Expect(1, 71351, '\P{^Canonical_Combining_Class=nukta}', "");
Expect(0, 71352, '\p{Canonical_Combining_Class=nukta}', "");
Expect(1, 71352, '\p{^Canonical_Combining_Class=nukta}', "");
Expect(1, 71352, '\P{Canonical_Combining_Class=nukta}', "");
Expect(0, 71352, '\P{^Canonical_Combining_Class=nukta}', "");
Expect(1, 71351, '\p{Canonical_Combining_Class=_Nukta}', "");
Expect(0, 71351, '\p{^Canonical_Combining_Class=_Nukta}', "");
Expect(0, 71351, '\P{Canonical_Combining_Class=_Nukta}', "");
Expect(1, 71351, '\P{^Canonical_Combining_Class=_Nukta}', "");
Expect(0, 71352, '\p{Canonical_Combining_Class=_Nukta}', "");
Expect(1, 71352, '\p{^Canonical_Combining_Class=_Nukta}', "");
Expect(1, 71352, '\P{Canonical_Combining_Class=_Nukta}', "");
Expect(0, 71352, '\P{^Canonical_Combining_Class=_Nukta}', "");
Error('\p{Ccc: :=	 NK}');
Error('\P{Ccc: :=	 NK}');
Expect(1, 71351, '\p{Ccc: nk}', "");
Expect(0, 71351, '\p{^Ccc: nk}', "");
Expect(0, 71351, '\P{Ccc: nk}', "");
Expect(1, 71351, '\P{^Ccc: nk}', "");
Expect(0, 71352, '\p{Ccc: nk}', "");
Expect(1, 71352, '\p{^Ccc: nk}', "");
Expect(1, 71352, '\P{Ccc: nk}', "");
Expect(0, 71352, '\P{^Ccc: nk}', "");
Expect(1, 71351, '\p{Ccc=--NK}', "");
Expect(0, 71351, '\p{^Ccc=--NK}', "");
Expect(0, 71351, '\P{Ccc=--NK}', "");
Expect(1, 71351, '\P{^Ccc=--NK}', "");
Expect(0, 71352, '\p{Ccc=--NK}', "");
Expect(1, 71352, '\p{^Ccc=--NK}', "");
Expect(1, 71352, '\P{Ccc=--NK}', "");
Expect(0, 71352, '\P{^Ccc=--NK}', "");
Error('\p{Is_Canonical_Combining_Class=:=--+000007}');
Error('\P{Is_Canonical_Combining_Class=:=--+000007}');
Expect(1, 71351, '\p{Is_Canonical_Combining_Class:0_0_0_07}', "");
Expect(0, 71351, '\p{^Is_Canonical_Combining_Class:0_0_0_07}', "");
Expect(0, 71351, '\P{Is_Canonical_Combining_Class:0_0_0_07}', "");
Expect(1, 71351, '\P{^Is_Canonical_Combining_Class:0_0_0_07}', "");
Expect(0, 71352, '\p{Is_Canonical_Combining_Class:0_0_0_07}', "");
Expect(1, 71352, '\p{^Is_Canonical_Combining_Class:0_0_0_07}', "");
Expect(1, 71352, '\P{Is_Canonical_Combining_Class:0_0_0_07}', "");
Expect(0, 71352, '\P{^Is_Canonical_Combining_Class:0_0_0_07}', "");
Error('\p{Is_Ccc=_:=Nukta}');
Error('\P{Is_Ccc=_:=Nukta}');
Expect(1, 71351, '\p{Is_Ccc=nukta}', "");
Expect(0, 71351, '\p{^Is_Ccc=nukta}', "");
Expect(0, 71351, '\P{Is_Ccc=nukta}', "");
Expect(1, 71351, '\P{^Is_Ccc=nukta}', "");
Expect(0, 71352, '\p{Is_Ccc=nukta}', "");
Expect(1, 71352, '\p{^Is_Ccc=nukta}', "");
Expect(1, 71352, '\P{Is_Ccc=nukta}', "");
Expect(0, 71352, '\P{^Is_Ccc=nukta}', "");
Expect(1, 71351, '\p{Is_Ccc:	--nukta}', "");
Expect(0, 71351, '\p{^Is_Ccc:	--nukta}', "");
Expect(0, 71351, '\P{Is_Ccc:	--nukta}', "");
Expect(1, 71351, '\P{^Is_Ccc:	--nukta}', "");
Expect(0, 71352, '\p{Is_Ccc:	--nukta}', "");
Expect(1, 71352, '\p{^Is_Ccc:	--nukta}', "");
Expect(1, 71352, '\P{Is_Ccc:	--nukta}', "");
Expect(0, 71352, '\P{^Is_Ccc:	--nukta}', "");
Error('\p{Canonical_Combining_Class=-/a/NOT_reordered}');
Error('\P{Canonical_Combining_Class=-/a/NOT_reordered}');
Expect(1, 119365, '\p{Canonical_Combining_Class=notreordered}', "");
Expect(0, 119365, '\p{^Canonical_Combining_Class=notreordered}', "");
Expect(0, 119365, '\P{Canonical_Combining_Class=notreordered}', "");
Expect(1, 119365, '\P{^Canonical_Combining_Class=notreordered}', "");
Expect(0, 119364, '\p{Canonical_Combining_Class=notreordered}', "");
Expect(1, 119364, '\p{^Canonical_Combining_Class=notreordered}', "");
Expect(1, 119364, '\P{Canonical_Combining_Class=notreordered}', "");
Expect(0, 119364, '\P{^Canonical_Combining_Class=notreordered}', "");
Expect(1, 119365, '\p{Canonical_Combining_Class=	-not_REORDERED}', "");
Expect(0, 119365, '\p{^Canonical_Combining_Class=	-not_REORDERED}', "");
Expect(0, 119365, '\P{Canonical_Combining_Class=	-not_REORDERED}', "");
Expect(1, 119365, '\P{^Canonical_Combining_Class=	-not_REORDERED}', "");
Expect(0, 119364, '\p{Canonical_Combining_Class=	-not_REORDERED}', "");
Expect(1, 119364, '\p{^Canonical_Combining_Class=	-not_REORDERED}', "");
Expect(1, 119364, '\P{Canonical_Combining_Class=	-not_REORDERED}', "");
Expect(0, 119364, '\P{^Canonical_Combining_Class=	-not_REORDERED}', "");
Error('\p{Ccc=/a/ _nr}');
Error('\P{Ccc=/a/ _nr}');
Expect(1, 119365, '\p{Ccc=nr}', "");
Expect(0, 119365, '\p{^Ccc=nr}', "");
Expect(0, 119365, '\P{Ccc=nr}', "");
Expect(1, 119365, '\P{^Ccc=nr}', "");
Expect(0, 119364, '\p{Ccc=nr}', "");
Expect(1, 119364, '\p{^Ccc=nr}', "");
Expect(1, 119364, '\P{Ccc=nr}', "");
Expect(0, 119364, '\P{^Ccc=nr}', "");
Expect(1, 119365, '\p{Ccc=__NR}', "");
Expect(0, 119365, '\p{^Ccc=__NR}', "");
Expect(0, 119365, '\P{Ccc=__NR}', "");
Expect(1, 119365, '\P{^Ccc=__NR}', "");
Expect(0, 119364, '\p{Ccc=__NR}', "");
Expect(1, 119364, '\p{^Ccc=__NR}', "");
Expect(1, 119364, '\P{Ccc=__NR}', "");
Expect(0, 119364, '\P{^Ccc=__NR}', "");
Error('\p{Is_Canonical_Combining_Class= :=000000}');
Error('\P{Is_Canonical_Combining_Class= :=000000}');
Expect(1, 119365, '\p{Is_Canonical_Combining_Class=+0_0_00}', "");
Expect(0, 119365, '\p{^Is_Canonical_Combining_Class=+0_0_00}', "");
Expect(0, 119365, '\P{Is_Canonical_Combining_Class=+0_0_00}', "");
Expect(1, 119365, '\P{^Is_Canonical_Combining_Class=+0_0_00}', "");
Expect(0, 119364, '\p{Is_Canonical_Combining_Class=+0_0_00}', "");
Expect(1, 119364, '\p{^Is_Canonical_Combining_Class=+0_0_00}', "");
Expect(1, 119364, '\P{Is_Canonical_Combining_Class=+0_0_00}', "");
Expect(0, 119364, '\P{^Is_Canonical_Combining_Class=+0_0_00}', "");
Error('\p{Is_Ccc=	:=Not_Reordered}');
Error('\P{Is_Ccc=	:=Not_Reordered}');
Expect(1, 119365, '\p{Is_Ccc=notreordered}', "");
Expect(0, 119365, '\p{^Is_Ccc=notreordered}', "");
Expect(0, 119365, '\P{Is_Ccc=notreordered}', "");
Expect(1, 119365, '\P{^Is_Ccc=notreordered}', "");
Expect(0, 119364, '\p{Is_Ccc=notreordered}', "");
Expect(1, 119364, '\p{^Is_Ccc=notreordered}', "");
Expect(1, 119364, '\P{Is_Ccc=notreordered}', "");
Expect(0, 119364, '\P{^Is_Ccc=notreordered}', "");
Expect(1, 119365, '\p{Is_Ccc=  NOT_Reordered}', "");
Expect(0, 119365, '\p{^Is_Ccc=  NOT_Reordered}', "");
Expect(0, 119365, '\P{Is_Ccc=  NOT_Reordered}', "");
Expect(1, 119365, '\P{^Is_Ccc=  NOT_Reordered}', "");
Expect(0, 119364, '\p{Is_Ccc=  NOT_Reordered}', "");
Expect(1, 119364, '\p{^Is_Ccc=  NOT_Reordered}', "");
Expect(1, 119364, '\P{Is_Ccc=  NOT_Reordered}', "");
Expect(0, 119364, '\P{^Is_Ccc=  NOT_Reordered}', "");
Error('\p{Canonical_Combining_Class=/a/__OVERLAY}');
Error('\P{Canonical_Combining_Class=/a/__OVERLAY}');
Expect(1, 119145, '\p{Canonical_Combining_Class=overlay}', "");
Expect(0, 119145, '\p{^Canonical_Combining_Class=overlay}', "");
Expect(0, 119145, '\P{Canonical_Combining_Class=overlay}', "");
Expect(1, 119145, '\P{^Canonical_Combining_Class=overlay}', "");
Expect(0, 119146, '\p{Canonical_Combining_Class=overlay}', "");
Expect(1, 119146, '\p{^Canonical_Combining_Class=overlay}', "");
Expect(1, 119146, '\P{Canonical_Combining_Class=overlay}', "");
Expect(0, 119146, '\P{^Canonical_Combining_Class=overlay}', "");
Expect(1, 119145, '\p{Canonical_Combining_Class:   Overlay}', "");
Expect(0, 119145, '\p{^Canonical_Combining_Class:   Overlay}', "");
Expect(0, 119145, '\P{Canonical_Combining_Class:   Overlay}', "");
Expect(1, 119145, '\P{^Canonical_Combining_Class:   Overlay}', "");
Expect(0, 119146, '\p{Canonical_Combining_Class:   Overlay}', "");
Expect(1, 119146, '\p{^Canonical_Combining_Class:   Overlay}', "");
Expect(1, 119146, '\P{Canonical_Combining_Class:   Overlay}', "");
Expect(0, 119146, '\P{^Canonical_Combining_Class:   Overlay}', "");
Error('\p{Ccc=:=	OV}');
Error('\P{Ccc=:=	OV}');
Expect(1, 119145, '\p{Ccc:	ov}', "");
Expect(0, 119145, '\p{^Ccc:	ov}', "");
Expect(0, 119145, '\P{Ccc:	ov}', "");
Expect(1, 119145, '\P{^Ccc:	ov}', "");
Expect(0, 119146, '\p{Ccc:	ov}', "");
Expect(1, 119146, '\p{^Ccc:	ov}', "");
Expect(1, 119146, '\P{Ccc:	ov}', "");
Expect(0, 119146, '\P{^Ccc:	ov}', "");
Expect(1, 119145, '\p{Ccc=- OV}', "");
Expect(0, 119145, '\p{^Ccc=- OV}', "");
Expect(0, 119145, '\P{Ccc=- OV}', "");
Expect(1, 119145, '\P{^Ccc=- OV}', "");
Expect(0, 119146, '\p{Ccc=- OV}', "");
Expect(1, 119146, '\p{^Ccc=- OV}', "");
Expect(1, 119146, '\P{Ccc=- OV}', "");
Expect(0, 119146, '\P{^Ccc=- OV}', "");
Error('\p{Is_Canonical_Combining_Class=:=		00000001}');
Error('\P{Is_Canonical_Combining_Class=:=		00000001}');
Expect(1, 119145, '\p{Is_Canonical_Combining_Class=0001}', "");
Expect(0, 119145, '\p{^Is_Canonical_Combining_Class=0001}', "");
Expect(0, 119145, '\P{Is_Canonical_Combining_Class=0001}', "");
Expect(1, 119145, '\P{^Is_Canonical_Combining_Class=0001}', "");
Expect(0, 119146, '\p{Is_Canonical_Combining_Class=0001}', "");
Expect(1, 119146, '\p{^Is_Canonical_Combining_Class=0001}', "");
Expect(1, 119146, '\P{Is_Canonical_Combining_Class=0001}', "");
Expect(0, 119146, '\P{^Is_Canonical_Combining_Class=0001}', "");
Error('\p{Is_Ccc=/a/ -Overlay}');
Error('\P{Is_Ccc=/a/ -Overlay}');
Expect(1, 119145, '\p{Is_Ccc=overlay}', "");
Expect(0, 119145, '\p{^Is_Ccc=overlay}', "");
Expect(0, 119145, '\P{Is_Ccc=overlay}', "");
Expect(1, 119145, '\P{^Is_Ccc=overlay}', "");
Expect(0, 119146, '\p{Is_Ccc=overlay}', "");
Expect(1, 119146, '\p{^Is_Ccc=overlay}', "");
Expect(1, 119146, '\P{Is_Ccc=overlay}', "");
Expect(0, 119146, '\P{^Is_Ccc=overlay}', "");
Expect(1, 119145, '\p{Is_Ccc=__Overlay}', "");
Expect(0, 119145, '\p{^Is_Ccc=__Overlay}', "");
Expect(0, 119145, '\P{Is_Ccc=__Overlay}', "");
Expect(1, 119145, '\P{^Is_Ccc=__Overlay}', "");
Expect(0, 119146, '\p{Is_Ccc=__Overlay}', "");
Expect(1, 119146, '\p{^Is_Ccc=__Overlay}', "");
Expect(1, 119146, '\P{Is_Ccc=__Overlay}', "");
Expect(0, 119146, '\P{^Is_Ccc=__Overlay}', "");
Error('\p{Canonical_Combining_Class=	/a/RIGHT}');
Error('\P{Canonical_Combining_Class=	/a/RIGHT}');
Expect(1, 119149, '\p{Canonical_Combining_Class=right}', "");
Expect(0, 119149, '\p{^Canonical_Combining_Class=right}', "");
Expect(0, 119149, '\P{Canonical_Combining_Class=right}', "");
Expect(1, 119149, '\P{^Canonical_Combining_Class=right}', "");
Expect(0, 119150, '\p{Canonical_Combining_Class=right}', "");
Expect(1, 119150, '\p{^Canonical_Combining_Class=right}', "");
Expect(1, 119150, '\P{Canonical_Combining_Class=right}', "");
Expect(0, 119150, '\P{^Canonical_Combining_Class=right}', "");
Expect(1, 119149, '\p{Canonical_Combining_Class=	-Right}', "");
Expect(0, 119149, '\p{^Canonical_Combining_Class=	-Right}', "");
Expect(0, 119149, '\P{Canonical_Combining_Class=	-Right}', "");
Expect(1, 119149, '\P{^Canonical_Combining_Class=	-Right}', "");
Expect(0, 119150, '\p{Canonical_Combining_Class=	-Right}', "");
Expect(1, 119150, '\p{^Canonical_Combining_Class=	-Right}', "");
Expect(1, 119150, '\P{Canonical_Combining_Class=	-Right}', "");
Expect(0, 119150, '\P{^Canonical_Combining_Class=	-Right}', "");
Error('\p{Ccc=:= -r}');
Error('\P{Ccc=:= -r}');
Expect(1, 119149, '\p{Ccc=r}', "");
Expect(0, 119149, '\p{^Ccc=r}', "");
Expect(0, 119149, '\P{Ccc=r}', "");
Expect(1, 119149, '\P{^Ccc=r}', "");
Expect(0, 119150, '\p{Ccc=r}', "");
Expect(1, 119150, '\p{^Ccc=r}', "");
Expect(1, 119150, '\P{Ccc=r}', "");
Expect(0, 119150, '\P{^Ccc=r}', "");
Expect(1, 119149, '\p{Ccc=-_R}', "");
Expect(0, 119149, '\p{^Ccc=-_R}', "");
Expect(0, 119149, '\P{Ccc=-_R}', "");
Expect(1, 119149, '\P{^Ccc=-_R}', "");
Expect(0, 119150, '\p{Ccc=-_R}', "");
Expect(1, 119150, '\p{^Ccc=-_R}', "");
Expect(1, 119150, '\P{Ccc=-_R}', "");
Expect(0, 119150, '\P{^Ccc=-_R}', "");
Error('\p{Is_Canonical_Combining_Class= _00000226/a/}');
Error('\P{Is_Canonical_Combining_Class= _00000226/a/}');
Expect(1, 119149, '\p{Is_Canonical_Combining_Class=+00_00_00_00_02_26}', "");
Expect(0, 119149, '\p{^Is_Canonical_Combining_Class=+00_00_00_00_02_26}', "");
Expect(0, 119149, '\P{Is_Canonical_Combining_Class=+00_00_00_00_02_26}', "");
Expect(1, 119149, '\P{^Is_Canonical_Combining_Class=+00_00_00_00_02_26}', "");
Expect(0, 119150, '\p{Is_Canonical_Combining_Class=+00_00_00_00_02_26}', "");
Expect(1, 119150, '\p{^Is_Canonical_Combining_Class=+00_00_00_00_02_26}', "");
Expect(1, 119150, '\P{Is_Canonical_Combining_Class=+00_00_00_00_02_26}', "");
Expect(0, 119150, '\P{^Is_Canonical_Combining_Class=+00_00_00_00_02_26}', "");
Error('\p{Is_Ccc=:= 	right}');
Error('\P{Is_Ccc=:= 	right}');
Expect(1, 119149, '\p{Is_Ccc=right}', "");
Expect(0, 119149, '\p{^Is_Ccc=right}', "");
Expect(0, 119149, '\P{Is_Ccc=right}', "");
Expect(1, 119149, '\P{^Is_Ccc=right}', "");
Expect(0, 119150, '\p{Is_Ccc=right}', "");
Expect(1, 119150, '\p{^Is_Ccc=right}', "");
Expect(1, 119150, '\P{Is_Ccc=right}', "");
Expect(0, 119150, '\P{^Is_Ccc=right}', "");
Expect(1, 119149, '\p{Is_Ccc=- Right}', "");
Expect(0, 119149, '\p{^Is_Ccc=- Right}', "");
Expect(0, 119149, '\P{Is_Ccc=- Right}', "");
Expect(1, 119149, '\P{^Is_Ccc=- Right}', "");
Expect(0, 119150, '\p{Is_Ccc=- Right}', "");
Expect(1, 119150, '\p{^Is_Ccc=- Right}', "");
Expect(1, 119150, '\P{Is_Ccc=- Right}', "");
Expect(0, 119150, '\P{^Is_Ccc=- Right}', "");
Error('\p{Canonical_Combining_Class=-/a/VIRAMA}');
Error('\P{Canonical_Combining_Class=-/a/VIRAMA}');
Expect(1, 71350, '\p{Canonical_Combining_Class=virama}', "");
Expect(0, 71350, '\p{^Canonical_Combining_Class=virama}', "");
Expect(0, 71350, '\P{Canonical_Combining_Class=virama}', "");
Expect(1, 71350, '\P{^Canonical_Combining_Class=virama}', "");
Expect(0, 71351, '\p{Canonical_Combining_Class=virama}', "");
Expect(1, 71351, '\p{^Canonical_Combining_Class=virama}', "");
Expect(1, 71351, '\P{Canonical_Combining_Class=virama}', "");
Expect(0, 71351, '\P{^Canonical_Combining_Class=virama}', "");
Expect(1, 71350, '\p{Canonical_Combining_Class=_virama}', "");
Expect(0, 71350, '\p{^Canonical_Combining_Class=_virama}', "");
Expect(0, 71350, '\P{Canonical_Combining_Class=_virama}', "");
Expect(1, 71350, '\P{^Canonical_Combining_Class=_virama}', "");
Expect(0, 71351, '\p{Canonical_Combining_Class=_virama}', "");
Expect(1, 71351, '\p{^Canonical_Combining_Class=_virama}', "");
Expect(1, 71351, '\P{Canonical_Combining_Class=_virama}', "");
Expect(0, 71351, '\P{^Canonical_Combining_Class=_virama}', "");
Error('\p{Ccc=  vr/a/}');
Error('\P{Ccc=  vr/a/}');
Expect(1, 71350, '\p{Ccc=vr}', "");
Expect(0, 71350, '\p{^Ccc=vr}', "");
Expect(0, 71350, '\P{Ccc=vr}', "");
Expect(1, 71350, '\P{^Ccc=vr}', "");
Expect(0, 71351, '\p{Ccc=vr}', "");
Expect(1, 71351, '\p{^Ccc=vr}', "");
Expect(1, 71351, '\P{Ccc=vr}', "");
Expect(0, 71351, '\P{^Ccc=vr}', "");
Expect(1, 71350, '\p{Ccc= -VR}', "");
Expect(0, 71350, '\p{^Ccc= -VR}', "");
Expect(0, 71350, '\P{Ccc= -VR}', "");
Expect(1, 71350, '\P{^Ccc= -VR}', "");
Expect(0, 71351, '\p{Ccc= -VR}', "");
Expect(1, 71351, '\p{^Ccc= -VR}', "");
Expect(1, 71351, '\P{Ccc= -VR}', "");
Expect(0, 71351, '\P{^Ccc= -VR}', "");
Error('\p{Is_Canonical_Combining_Class:/a/  0000000_9}');
Error('\P{Is_Canonical_Combining_Class:/a/  0000000_9}');
Expect(1, 71350, '\p{Is_Canonical_Combining_Class=+0_0_09}', "");
Expect(0, 71350, '\p{^Is_Canonical_Combining_Class=+0_0_09}', "");
Expect(0, 71350, '\P{Is_Canonical_Combining_Class=+0_0_09}', "");
Expect(1, 71350, '\P{^Is_Canonical_Combining_Class=+0_0_09}', "");
Expect(0, 71351, '\p{Is_Canonical_Combining_Class=+0_0_09}', "");
Expect(1, 71351, '\p{^Is_Canonical_Combining_Class=+0_0_09}', "");
Expect(1, 71351, '\P{Is_Canonical_Combining_Class=+0_0_09}', "");
Expect(0, 71351, '\P{^Is_Canonical_Combining_Class=+0_0_09}', "");
Error('\p{Is_Ccc=:=virama}');
Error('\P{Is_Ccc=:=virama}');
Expect(1, 71350, '\p{Is_Ccc=virama}', "");
Expect(0, 71350, '\p{^Is_Ccc=virama}', "");
Expect(0, 71350, '\P{Is_Ccc=virama}', "");
Expect(1, 71350, '\P{^Is_Ccc=virama}', "");
Expect(0, 71351, '\p{Is_Ccc=virama}', "");
Expect(1, 71351, '\p{^Is_Ccc=virama}', "");
Expect(1, 71351, '\P{Is_Ccc=virama}', "");
Expect(0, 71351, '\P{^Is_Ccc=virama}', "");
Expect(1, 71350, '\p{Is_Ccc= virama}', "");
Expect(0, 71350, '\p{^Is_Ccc= virama}', "");
Expect(0, 71350, '\P{Is_Ccc= virama}', "");
Expect(1, 71350, '\P{^Is_Ccc= virama}', "");
Expect(0, 71351, '\p{Is_Ccc= virama}', "");
Expect(1, 71351, '\p{^Is_Ccc= virama}', "");
Expect(1, 71351, '\P{Is_Ccc= virama}', "");
Expect(0, 71351, '\P{^Is_Ccc= virama}', "");
Error('\p{Decomposition_Type= /a/canonical}');
Error('\P{Decomposition_Type= /a/canonical}');
Expect(1, 195101, '\p{Decomposition_Type=canonical}', "");
Expect(0, 195101, '\p{^Decomposition_Type=canonical}', "");
Expect(0, 195101, '\P{Decomposition_Type=canonical}', "");
Expect(1, 195101, '\P{^Decomposition_Type=canonical}', "");
Expect(0, 195102, '\p{Decomposition_Type=canonical}', "");
Expect(1, 195102, '\p{^Decomposition_Type=canonical}', "");
Expect(1, 195102, '\P{Decomposition_Type=canonical}', "");
Expect(0, 195102, '\P{^Decomposition_Type=canonical}', "");
Expect(1, 195101, '\p{Decomposition_Type=_Canonical}', "");
Expect(0, 195101, '\p{^Decomposition_Type=_Canonical}', "");
Expect(0, 195101, '\P{Decomposition_Type=_Canonical}', "");
Expect(1, 195101, '\P{^Decomposition_Type=_Canonical}', "");
Expect(0, 195102, '\p{Decomposition_Type=_Canonical}', "");
Expect(1, 195102, '\p{^Decomposition_Type=_Canonical}', "");
Expect(1, 195102, '\P{Decomposition_Type=_Canonical}', "");
Expect(0, 195102, '\P{^Decomposition_Type=_Canonical}', "");
Error('\p{Dt=/a/_can}');
Error('\P{Dt=/a/_can}');
Expect(1, 195101, '\p{Dt=can}', "");
Expect(0, 195101, '\p{^Dt=can}', "");
Expect(0, 195101, '\P{Dt=can}', "");
Expect(1, 195101, '\P{^Dt=can}', "");
Expect(0, 195102, '\p{Dt=can}', "");
Expect(1, 195102, '\p{^Dt=can}', "");
Expect(1, 195102, '\P{Dt=can}', "");
Expect(0, 195102, '\P{^Dt=can}', "");
Expect(1, 195101, '\p{Dt= 	CAN}', "");
Expect(0, 195101, '\p{^Dt= 	CAN}', "");
Expect(0, 195101, '\P{Dt= 	CAN}', "");
Expect(1, 195101, '\P{^Dt= 	CAN}', "");
Expect(0, 195102, '\p{Dt= 	CAN}', "");
Expect(1, 195102, '\p{^Dt= 	CAN}', "");
Expect(1, 195102, '\P{Dt= 	CAN}', "");
Expect(0, 195102, '\P{^Dt= 	CAN}', "");
Error('\p{Is_Decomposition_Type= :=CANONICAL}');
Error('\P{Is_Decomposition_Type= :=CANONICAL}');
Expect(1, 195101, '\p{Is_Decomposition_Type: canonical}', "");
Expect(0, 195101, '\p{^Is_Decomposition_Type: canonical}', "");
Expect(0, 195101, '\P{Is_Decomposition_Type: canonical}', "");
Expect(1, 195101, '\P{^Is_Decomposition_Type: canonical}', "");
Expect(0, 195102, '\p{Is_Decomposition_Type: canonical}', "");
Expect(1, 195102, '\p{^Is_Decomposition_Type: canonical}', "");
Expect(1, 195102, '\P{Is_Decomposition_Type: canonical}', "");
Expect(0, 195102, '\P{^Is_Decomposition_Type: canonical}', "");
Expect(1, 195101, '\p{Is_Decomposition_Type=__Canonical}', "");
Expect(0, 195101, '\p{^Is_Decomposition_Type=__Canonical}', "");
Expect(0, 195101, '\P{Is_Decomposition_Type=__Canonical}', "");
Expect(1, 195101, '\P{^Is_Decomposition_Type=__Canonical}', "");
Expect(0, 195102, '\p{Is_Decomposition_Type=__Canonical}', "");
Expect(1, 195102, '\p{^Is_Decomposition_Type=__Canonical}', "");
Expect(1, 195102, '\P{Is_Decomposition_Type=__Canonical}', "");
Expect(0, 195102, '\P{^Is_Decomposition_Type=__Canonical}', "");
Error('\p{Is_Dt=:= 	can}');
Error('\P{Is_Dt=:= 	can}');
Expect(1, 195101, '\p{Is_Dt:   can}', "");
Expect(0, 195101, '\p{^Is_Dt:   can}', "");
Expect(0, 195101, '\P{Is_Dt:   can}', "");
Expect(1, 195101, '\P{^Is_Dt:   can}', "");
Expect(0, 195102, '\p{Is_Dt:   can}', "");
Expect(1, 195102, '\p{^Is_Dt:   can}', "");
Expect(1, 195102, '\P{Is_Dt:   can}', "");
Expect(0, 195102, '\P{^Is_Dt:   can}', "");
Expect(1, 195101, '\p{Is_Dt=-_Can}', "");
Expect(0, 195101, '\p{^Is_Dt=-_Can}', "");
Expect(0, 195101, '\P{Is_Dt=-_Can}', "");
Expect(1, 195101, '\P{^Is_Dt=-_Can}', "");
Expect(0, 195102, '\p{Is_Dt=-_Can}', "");
Expect(1, 195102, '\p{^Is_Dt=-_Can}', "");
Expect(1, 195102, '\P{Is_Dt=-_Can}', "");
Expect(0, 195102, '\P{^Is_Dt=-_Can}', "");
Error('\p{Decomposition_Type:	:= -COMPAT}');
Error('\P{Decomposition_Type:	:= -COMPAT}');
Expect(1, 127560, '\p{Decomposition_Type:	compat}', "");
Expect(0, 127560, '\p{^Decomposition_Type:	compat}', "");
Expect(0, 127560, '\P{Decomposition_Type:	compat}', "");
Expect(1, 127560, '\P{^Decomposition_Type:	compat}', "");
Expect(0, 127561, '\p{Decomposition_Type:	compat}', "");
Expect(1, 127561, '\p{^Decomposition_Type:	compat}', "");
Expect(1, 127561, '\P{Decomposition_Type:	compat}', "");
Expect(0, 127561, '\P{^Decomposition_Type:	compat}', "");
Expect(1, 127560, '\p{Decomposition_Type=-	compat}', "");
Expect(0, 127560, '\p{^Decomposition_Type=-	compat}', "");
Expect(0, 127560, '\P{Decomposition_Type=-	compat}', "");
Expect(1, 127560, '\P{^Decomposition_Type=-	compat}', "");
Expect(0, 127561, '\p{Decomposition_Type=-	compat}', "");
Expect(1, 127561, '\p{^Decomposition_Type=-	compat}', "");
Expect(1, 127561, '\P{Decomposition_Type=-	compat}', "");
Expect(0, 127561, '\P{^Decomposition_Type=-	compat}', "");
Error('\p{Dt=-Com/a/}');
Error('\P{Dt=-Com/a/}');
Expect(1, 127560, '\p{Dt=com}', "");
Expect(0, 127560, '\p{^Dt=com}', "");
Expect(0, 127560, '\P{Dt=com}', "");
Expect(1, 127560, '\P{^Dt=com}', "");
Expect(0, 127561, '\p{Dt=com}', "");
Expect(1, 127561, '\p{^Dt=com}', "");
Expect(1, 127561, '\P{Dt=com}', "");
Expect(0, 127561, '\P{^Dt=com}', "");
Expect(1, 127560, '\p{Dt=	COM}', "");
Expect(0, 127560, '\p{^Dt=	COM}', "");
Expect(0, 127560, '\P{Dt=	COM}', "");
Expect(1, 127560, '\P{^Dt=	COM}', "");
Expect(0, 127561, '\p{Dt=	COM}', "");
Expect(1, 127561, '\p{^Dt=	COM}', "");
Expect(1, 127561, '\P{Dt=	COM}', "");
Expect(0, 127561, '\P{^Dt=	COM}', "");
Error('\p{Is_Decomposition_Type=:= _Compat}');
Error('\P{Is_Decomposition_Type=:= _Compat}');
Expect(1, 127560, '\p{Is_Decomposition_Type=compat}', "");
Expect(0, 127560, '\p{^Is_Decomposition_Type=compat}', "");
Expect(0, 127560, '\P{Is_Decomposition_Type=compat}', "");
Expect(1, 127560, '\P{^Is_Decomposition_Type=compat}', "");
Expect(0, 127561, '\p{Is_Decomposition_Type=compat}', "");
Expect(1, 127561, '\p{^Is_Decomposition_Type=compat}', "");
Expect(1, 127561, '\P{Is_Decomposition_Type=compat}', "");
Expect(0, 127561, '\P{^Is_Decomposition_Type=compat}', "");
Expect(1, 127560, '\p{Is_Decomposition_Type= Compat}', "");
Expect(0, 127560, '\p{^Is_Decomposition_Type= Compat}', "");
Expect(0, 127560, '\P{Is_Decomposition_Type= Compat}', "");
Expect(1, 127560, '\P{^Is_Decomposition_Type= Compat}', "");
Expect(0, 127561, '\p{Is_Decomposition_Type= Compat}', "");
Expect(1, 127561, '\p{^Is_Decomposition_Type= Compat}', "");
Expect(1, 127561, '\P{Is_Decomposition_Type= Compat}', "");
Expect(0, 127561, '\P{^Is_Decomposition_Type= Compat}', "");
Error('\p{Is_Dt=	/a/Com}');
Error('\P{Is_Dt=	/a/Com}');
Expect(1, 127560, '\p{Is_Dt=com}', "");
Expect(0, 127560, '\p{^Is_Dt=com}', "");
Expect(0, 127560, '\P{Is_Dt=com}', "");
Expect(1, 127560, '\P{^Is_Dt=com}', "");
Expect(0, 127561, '\p{Is_Dt=com}', "");
Expect(1, 127561, '\p{^Is_Dt=com}', "");
Expect(1, 127561, '\P{Is_Dt=com}', "");
Expect(0, 127561, '\P{^Is_Dt=com}', "");
Expect(1, 127560, '\p{Is_Dt= -Com}', "");
Expect(0, 127560, '\p{^Is_Dt= -Com}', "");
Expect(0, 127560, '\P{Is_Dt= -Com}', "");
Expect(1, 127560, '\P{^Is_Dt= -Com}', "");
Expect(0, 127561, '\p{Is_Dt= -Com}', "");
Expect(1, 127561, '\p{^Is_Dt= -Com}', "");
Expect(1, 127561, '\P{Is_Dt= -Com}', "");
Expect(0, 127561, '\P{^Is_Dt= -Com}', "");
Error('\p{Decomposition_Type=:=circle}');
Error('\P{Decomposition_Type=:=circle}');
Expect(1, 127569, '\p{Decomposition_Type=circle}', "");
Expect(0, 127569, '\p{^Decomposition_Type=circle}', "");
Expect(0, 127569, '\P{Decomposition_Type=circle}', "");
Expect(1, 127569, '\P{^Decomposition_Type=circle}', "");
Expect(0, 127570, '\p{Decomposition_Type=circle}', "");
Expect(1, 127570, '\p{^Decomposition_Type=circle}', "");
Expect(1, 127570, '\P{Decomposition_Type=circle}', "");
Expect(0, 127570, '\P{^Decomposition_Type=circle}', "");
Expect(1, 127569, '\p{Decomposition_Type=	Circle}', "");
Expect(0, 127569, '\p{^Decomposition_Type=	Circle}', "");
Expect(0, 127569, '\P{Decomposition_Type=	Circle}', "");
Expect(1, 127569, '\P{^Decomposition_Type=	Circle}', "");
Expect(0, 127570, '\p{Decomposition_Type=	Circle}', "");
Expect(1, 127570, '\p{^Decomposition_Type=	Circle}', "");
Expect(1, 127570, '\P{Decomposition_Type=	Circle}', "");
Expect(0, 127570, '\P{^Decomposition_Type=	Circle}', "");
Error('\p{Dt=/a/	_Enc}');
Error('\P{Dt=/a/	_Enc}');
Expect(1, 127569, '\p{Dt=enc}', "");
Expect(0, 127569, '\p{^Dt=enc}', "");
Expect(0, 127569, '\P{Dt=enc}', "");
Expect(1, 127569, '\P{^Dt=enc}', "");
Expect(0, 127570, '\p{Dt=enc}', "");
Expect(1, 127570, '\p{^Dt=enc}', "");
Expect(1, 127570, '\P{Dt=enc}', "");
Expect(0, 127570, '\P{^Dt=enc}', "");
Expect(1, 127569, '\p{Dt= _Enc}', "");
Expect(0, 127569, '\p{^Dt= _Enc}', "");
Expect(0, 127569, '\P{Dt= _Enc}', "");
Expect(1, 127569, '\P{^Dt= _Enc}', "");
Expect(0, 127570, '\p{Dt= _Enc}', "");
Expect(1, 127570, '\p{^Dt= _Enc}', "");
Expect(1, 127570, '\P{Dt= _Enc}', "");
Expect(0, 127570, '\P{^Dt= _Enc}', "");
Error('\p{Is_Decomposition_Type=-/a/Circle}');
Error('\P{Is_Decomposition_Type=-/a/Circle}');
Expect(1, 127569, '\p{Is_Decomposition_Type:	circle}', "");
Expect(0, 127569, '\p{^Is_Decomposition_Type:	circle}', "");
Expect(0, 127569, '\P{Is_Decomposition_Type:	circle}', "");
Expect(1, 127569, '\P{^Is_Decomposition_Type:	circle}', "");
Expect(0, 127570, '\p{Is_Decomposition_Type:	circle}', "");
Expect(1, 127570, '\p{^Is_Decomposition_Type:	circle}', "");
Expect(1, 127570, '\P{Is_Decomposition_Type:	circle}', "");
Expect(0, 127570, '\P{^Is_Decomposition_Type:	circle}', "");
Expect(1, 127569, '\p{Is_Decomposition_Type=  Circle}', "");
Expect(0, 127569, '\p{^Is_Decomposition_Type=  Circle}', "");
Expect(0, 127569, '\P{Is_Decomposition_Type=  Circle}', "");
Expect(1, 127569, '\P{^Is_Decomposition_Type=  Circle}', "");
Expect(0, 127570, '\p{Is_Decomposition_Type=  Circle}', "");
Expect(1, 127570, '\p{^Is_Decomposition_Type=  Circle}', "");
Expect(1, 127570, '\P{Is_Decomposition_Type=  Circle}', "");
Expect(0, 127570, '\P{^Is_Decomposition_Type=  Circle}', "");
Error('\p{Is_Dt=/a/		ENC}');
Error('\P{Is_Dt=/a/		ENC}');
Expect(1, 127569, '\p{Is_Dt=enc}', "");
Expect(0, 127569, '\p{^Is_Dt=enc}', "");
Expect(0, 127569, '\P{Is_Dt=enc}', "");
Expect(1, 127569, '\P{^Is_Dt=enc}', "");
Expect(0, 127570, '\p{Is_Dt=enc}', "");
Expect(1, 127570, '\p{^Is_Dt=enc}', "");
Expect(1, 127570, '\P{Is_Dt=enc}', "");
Expect(0, 127570, '\P{^Is_Dt=enc}', "");
Expect(1, 127569, '\p{Is_Dt= ENC}', "");
Expect(0, 127569, '\p{^Is_Dt= ENC}', "");
Expect(0, 127569, '\P{Is_Dt= ENC}', "");
Expect(1, 127569, '\P{^Is_Dt= ENC}', "");
Expect(0, 127570, '\p{Is_Dt= ENC}', "");
Expect(1, 127570, '\p{^Is_Dt= ENC}', "");
Expect(1, 127570, '\P{Is_Dt= ENC}', "");
Expect(0, 127570, '\P{^Is_Dt= ENC}', "");
Error('\p{Decomposition_Type=-Final/a/}');
Error('\P{Decomposition_Type=-Final/a/}');
Expect(1, 65276, '\p{Decomposition_Type=final}', "");
Expect(0, 65276, '\p{^Decomposition_Type=final}', "");
Expect(0, 65276, '\P{Decomposition_Type=final}', "");
Expect(1, 65276, '\P{^Decomposition_Type=final}', "");
Expect(0, 65277, '\p{Decomposition_Type=final}', "");
Expect(1, 65277, '\p{^Decomposition_Type=final}', "");
Expect(1, 65277, '\P{Decomposition_Type=final}', "");
Expect(0, 65277, '\P{^Decomposition_Type=final}', "");
Expect(1, 65276, '\p{Decomposition_Type=	-FINAL}', "");
Expect(0, 65276, '\p{^Decomposition_Type=	-FINAL}', "");
Expect(0, 65276, '\P{Decomposition_Type=	-FINAL}', "");
Expect(1, 65276, '\P{^Decomposition_Type=	-FINAL}', "");
Expect(0, 65277, '\p{Decomposition_Type=	-FINAL}', "");
Expect(1, 65277, '\p{^Decomposition_Type=	-FINAL}', "");
Expect(1, 65277, '\P{Decomposition_Type=	-FINAL}', "");
Expect(0, 65277, '\P{^Decomposition_Type=	-FINAL}', "");
Error('\p{Dt:/a/_	Fin}');
Error('\P{Dt:/a/_	Fin}');
Expect(1, 65276, '\p{Dt=fin}', "");
Expect(0, 65276, '\p{^Dt=fin}', "");
Expect(0, 65276, '\P{Dt=fin}', "");
Expect(1, 65276, '\P{^Dt=fin}', "");
Expect(0, 65277, '\p{Dt=fin}', "");
Expect(1, 65277, '\p{^Dt=fin}', "");
Expect(1, 65277, '\P{Dt=fin}', "");
Expect(0, 65277, '\P{^Dt=fin}', "");
Expect(1, 65276, '\p{Dt=Fin}', "");
Expect(0, 65276, '\p{^Dt=Fin}', "");
Expect(0, 65276, '\P{Dt=Fin}', "");
Expect(1, 65276, '\P{^Dt=Fin}', "");
Expect(0, 65277, '\p{Dt=Fin}', "");
Expect(1, 65277, '\p{^Dt=Fin}', "");
Expect(1, 65277, '\P{Dt=Fin}', "");
Expect(0, 65277, '\P{^Dt=Fin}', "");
Error('\p{Is_Decomposition_Type=:=	 Final}');
Error('\P{Is_Decomposition_Type=:=	 Final}');
Expect(1, 65276, '\p{Is_Decomposition_Type=final}', "");
Expect(0, 65276, '\p{^Is_Decomposition_Type=final}', "");
Expect(0, 65276, '\P{Is_Decomposition_Type=final}', "");
Expect(1, 65276, '\P{^Is_Decomposition_Type=final}', "");
Expect(0, 65277, '\p{Is_Decomposition_Type=final}', "");
Expect(1, 65277, '\p{^Is_Decomposition_Type=final}', "");
Expect(1, 65277, '\P{Is_Decomposition_Type=final}', "");
Expect(0, 65277, '\P{^Is_Decomposition_Type=final}', "");
Expect(1, 65276, '\p{Is_Decomposition_Type=-	Final}', "");
Expect(0, 65276, '\p{^Is_Decomposition_Type=-	Final}', "");
Expect(0, 65276, '\P{Is_Decomposition_Type=-	Final}', "");
Expect(1, 65276, '\P{^Is_Decomposition_Type=-	Final}', "");
Expect(0, 65277, '\p{Is_Decomposition_Type=-	Final}', "");
Expect(1, 65277, '\p{^Is_Decomposition_Type=-	Final}', "");
Expect(1, 65277, '\P{Is_Decomposition_Type=-	Final}', "");
Expect(0, 65277, '\P{^Is_Decomposition_Type=-	Final}', "");
Error('\p{Is_Dt=/a/- Fin}');
Error('\P{Is_Dt=/a/- Fin}');
Expect(1, 65276, '\p{Is_Dt=fin}', "");
Expect(0, 65276, '\p{^Is_Dt=fin}', "");
Expect(0, 65276, '\P{Is_Dt=fin}', "");
Expect(1, 65276, '\P{^Is_Dt=fin}', "");
Expect(0, 65277, '\p{Is_Dt=fin}', "");
Expect(1, 65277, '\p{^Is_Dt=fin}', "");
Expect(1, 65277, '\P{Is_Dt=fin}', "");
Expect(0, 65277, '\P{^Is_Dt=fin}', "");
Expect(1, 65276, '\p{Is_Dt=_ Fin}', "");
Expect(0, 65276, '\p{^Is_Dt=_ Fin}', "");
Expect(0, 65276, '\P{Is_Dt=_ Fin}', "");
Expect(1, 65276, '\P{^Is_Dt=_ Fin}', "");
Expect(0, 65277, '\p{Is_Dt=_ Fin}', "");
Expect(1, 65277, '\p{^Is_Dt=_ Fin}', "");
Expect(1, 65277, '\P{Is_Dt=_ Fin}', "");
Expect(0, 65277, '\P{^Is_Dt=_ Fin}', "");
Error('\p{Decomposition_Type=:=font}');
Error('\P{Decomposition_Type=:=font}');
Expect(1, 126651, '\p{Decomposition_Type=font}', "");
Expect(0, 126651, '\p{^Decomposition_Type=font}', "");
Expect(0, 126651, '\P{Decomposition_Type=font}', "");
Expect(1, 126651, '\P{^Decomposition_Type=font}', "");
Expect(0, 126652, '\p{Decomposition_Type=font}', "");
Expect(1, 126652, '\p{^Decomposition_Type=font}', "");
Expect(1, 126652, '\P{Decomposition_Type=font}', "");
Expect(0, 126652, '\P{^Decomposition_Type=font}', "");
Expect(1, 126651, '\p{Decomposition_Type=	Font}', "");
Expect(0, 126651, '\p{^Decomposition_Type=	Font}', "");
Expect(0, 126651, '\P{Decomposition_Type=	Font}', "");
Expect(1, 126651, '\P{^Decomposition_Type=	Font}', "");
Expect(0, 126652, '\p{Decomposition_Type=	Font}', "");
Expect(1, 126652, '\p{^Decomposition_Type=	Font}', "");
Expect(1, 126652, '\P{Decomposition_Type=	Font}', "");
Expect(0, 126652, '\P{^Decomposition_Type=	Font}', "");
Error('\p{Dt: :=  font}');
Error('\P{Dt: :=  font}');
Expect(1, 126651, '\p{Dt=font}', "");
Expect(0, 126651, '\p{^Dt=font}', "");
Expect(0, 126651, '\P{Dt=font}', "");
Expect(1, 126651, '\P{^Dt=font}', "");
Expect(0, 126652, '\p{Dt=font}', "");
Expect(1, 126652, '\p{^Dt=font}', "");
Expect(1, 126652, '\P{Dt=font}', "");
Expect(0, 126652, '\P{^Dt=font}', "");
Expect(1, 126651, '\p{Dt=_Font}', "");
Expect(0, 126651, '\p{^Dt=_Font}', "");
Expect(0, 126651, '\P{Dt=_Font}', "");
Expect(1, 126651, '\P{^Dt=_Font}', "");
Expect(0, 126652, '\p{Dt=_Font}', "");
Expect(1, 126652, '\p{^Dt=_Font}', "");
Expect(1, 126652, '\P{Dt=_Font}', "");
Expect(0, 126652, '\P{^Dt=_Font}', "");
Error('\p{Is_Decomposition_Type:  	Font:=}');
Error('\P{Is_Decomposition_Type:  	Font:=}');
Expect(1, 126651, '\p{Is_Decomposition_Type=font}', "");
Expect(0, 126651, '\p{^Is_Decomposition_Type=font}', "");
Expect(0, 126651, '\P{Is_Decomposition_Type=font}', "");
Expect(1, 126651, '\P{^Is_Decomposition_Type=font}', "");
Expect(0, 126652, '\p{Is_Decomposition_Type=font}', "");
Expect(1, 126652, '\p{^Is_Decomposition_Type=font}', "");
Expect(1, 126652, '\P{Is_Decomposition_Type=font}', "");
Expect(0, 126652, '\P{^Is_Decomposition_Type=font}', "");
Expect(1, 126651, '\p{Is_Decomposition_Type=--font}', "");
Expect(0, 126651, '\p{^Is_Decomposition_Type=--font}', "");
Expect(0, 126651, '\P{Is_Decomposition_Type=--font}', "");
Expect(1, 126651, '\P{^Is_Decomposition_Type=--font}', "");
Expect(0, 126652, '\p{Is_Decomposition_Type=--font}', "");
Expect(1, 126652, '\p{^Is_Decomposition_Type=--font}', "");
Expect(1, 126652, '\P{Is_Decomposition_Type=--font}', "");
Expect(0, 126652, '\P{^Is_Decomposition_Type=--font}', "");
Error('\p{Is_Dt=/a/FONT}');
Error('\P{Is_Dt=/a/FONT}');
Expect(1, 126651, '\p{Is_Dt=font}', "");
Expect(0, 126651, '\p{^Is_Dt=font}', "");
Expect(0, 126651, '\P{Is_Dt=font}', "");
Expect(1, 126651, '\P{^Is_Dt=font}', "");
Expect(0, 126652, '\p{Is_Dt=font}', "");
Expect(1, 126652, '\p{^Is_Dt=font}', "");
Expect(1, 126652, '\P{Is_Dt=font}', "");
Expect(0, 126652, '\P{^Is_Dt=font}', "");
Expect(1, 126651, '\p{Is_Dt=_	Font}', "");
Expect(0, 126651, '\p{^Is_Dt=_	Font}', "");
Expect(0, 126651, '\P{Is_Dt=_	Font}', "");
Expect(1, 126651, '\P{^Is_Dt=_	Font}', "");
Expect(0, 126652, '\p{Is_Dt=_	Font}', "");
Expect(1, 126652, '\p{^Is_Dt=_	Font}', "");
Expect(1, 126652, '\P{Is_Dt=_	Font}', "");
Expect(0, 126652, '\P{^Is_Dt=_	Font}', "");
Error('\p{Decomposition_Type:	:=	_Fraction}');
Error('\P{Decomposition_Type:	:=	_Fraction}');
Expect(1, 8585, '\p{Decomposition_Type=fraction}', "");
Expect(0, 8585, '\p{^Decomposition_Type=fraction}', "");
Expect(0, 8585, '\P{Decomposition_Type=fraction}', "");
Expect(1, 8585, '\P{^Decomposition_Type=fraction}', "");
Expect(0, 8586, '\p{Decomposition_Type=fraction}', "");
Expect(1, 8586, '\p{^Decomposition_Type=fraction}', "");
Expect(1, 8586, '\P{Decomposition_Type=fraction}', "");
Expect(0, 8586, '\P{^Decomposition_Type=fraction}', "");
Expect(1, 8585, '\p{Decomposition_Type=- FRACTION}', "");
Expect(0, 8585, '\p{^Decomposition_Type=- FRACTION}', "");
Expect(0, 8585, '\P{Decomposition_Type=- FRACTION}', "");
Expect(1, 8585, '\P{^Decomposition_Type=- FRACTION}', "");
Expect(0, 8586, '\p{Decomposition_Type=- FRACTION}', "");
Expect(1, 8586, '\p{^Decomposition_Type=- FRACTION}', "");
Expect(1, 8586, '\P{Decomposition_Type=- FRACTION}', "");
Expect(0, 8586, '\P{^Decomposition_Type=- FRACTION}', "");
Error('\p{Dt:	:=fra}');
Error('\P{Dt:	:=fra}');
Expect(1, 8585, '\p{Dt=fra}', "");
Expect(0, 8585, '\p{^Dt=fra}', "");
Expect(0, 8585, '\P{Dt=fra}', "");
Expect(1, 8585, '\P{^Dt=fra}', "");
Expect(0, 8586, '\p{Dt=fra}', "");
Expect(1, 8586, '\p{^Dt=fra}', "");
Expect(1, 8586, '\P{Dt=fra}', "");
Expect(0, 8586, '\P{^Dt=fra}', "");
Expect(1, 8585, '\p{Dt=-_Fra}', "");
Expect(0, 8585, '\p{^Dt=-_Fra}', "");
Expect(0, 8585, '\P{Dt=-_Fra}', "");
Expect(1, 8585, '\P{^Dt=-_Fra}', "");
Expect(0, 8586, '\p{Dt=-_Fra}', "");
Expect(1, 8586, '\p{^Dt=-_Fra}', "");
Expect(1, 8586, '\P{Dt=-_Fra}', "");
Expect(0, 8586, '\P{^Dt=-_Fra}', "");
Error('\p{Is_Decomposition_Type=-:=Fraction}');
Error('\P{Is_Decomposition_Type=-:=Fraction}');
Expect(1, 8585, '\p{Is_Decomposition_Type=fraction}', "");
Expect(0, 8585, '\p{^Is_Decomposition_Type=fraction}', "");
Expect(0, 8585, '\P{Is_Decomposition_Type=fraction}', "");
Expect(1, 8585, '\P{^Is_Decomposition_Type=fraction}', "");
Expect(0, 8586, '\p{Is_Decomposition_Type=fraction}', "");
Expect(1, 8586, '\p{^Is_Decomposition_Type=fraction}', "");
Expect(1, 8586, '\P{Is_Decomposition_Type=fraction}', "");
Expect(0, 8586, '\P{^Is_Decomposition_Type=fraction}', "");
Expect(1, 8585, '\p{Is_Decomposition_Type=_fraction}', "");
Expect(0, 8585, '\p{^Is_Decomposition_Type=_fraction}', "");
Expect(0, 8585, '\P{Is_Decomposition_Type=_fraction}', "");
Expect(1, 8585, '\P{^Is_Decomposition_Type=_fraction}', "");
Expect(0, 8586, '\p{Is_Decomposition_Type=_fraction}', "");
Expect(1, 8586, '\p{^Is_Decomposition_Type=_fraction}', "");
Expect(1, 8586, '\P{Is_Decomposition_Type=_fraction}', "");
Expect(0, 8586, '\P{^Is_Decomposition_Type=_fraction}', "");
Error('\p{Is_Dt=__Fra:=}');
Error('\P{Is_Dt=__Fra:=}');
Expect(1, 8585, '\p{Is_Dt=fra}', "");
Expect(0, 8585, '\p{^Is_Dt=fra}', "");
Expect(0, 8585, '\P{Is_Dt=fra}', "");
Expect(1, 8585, '\P{^Is_Dt=fra}', "");
Expect(0, 8586, '\p{Is_Dt=fra}', "");
Expect(1, 8586, '\p{^Is_Dt=fra}', "");
Expect(1, 8586, '\P{Is_Dt=fra}', "");
Expect(0, 8586, '\P{^Is_Dt=fra}', "");
Expect(1, 8585, '\p{Is_Dt=	_Fra}', "");
Expect(0, 8585, '\p{^Is_Dt=	_Fra}', "");
Expect(0, 8585, '\P{Is_Dt=	_Fra}', "");
Expect(1, 8585, '\P{^Is_Dt=	_Fra}', "");
Expect(0, 8586, '\p{Is_Dt=	_Fra}', "");
Expect(1, 8586, '\p{^Is_Dt=	_Fra}', "");
Expect(1, 8586, '\P{Is_Dt=	_Fra}', "");
Expect(0, 8586, '\P{^Is_Dt=	_Fra}', "");
Error('\p{Decomposition_Type=:=-	Initial}');
Error('\P{Decomposition_Type=:=-	Initial}');
Expect(1, 65267, '\p{Decomposition_Type=initial}', "");
Expect(0, 65267, '\p{^Decomposition_Type=initial}', "");
Expect(0, 65267, '\P{Decomposition_Type=initial}', "");
Expect(1, 65267, '\P{^Decomposition_Type=initial}', "");
Expect(0, 65268, '\p{Decomposition_Type=initial}', "");
Expect(1, 65268, '\p{^Decomposition_Type=initial}', "");
Expect(1, 65268, '\P{Decomposition_Type=initial}', "");
Expect(0, 65268, '\P{^Decomposition_Type=initial}', "");
Expect(1, 65267, '\p{Decomposition_Type=-	Initial}', "");
Expect(0, 65267, '\p{^Decomposition_Type=-	Initial}', "");
Expect(0, 65267, '\P{Decomposition_Type=-	Initial}', "");
Expect(1, 65267, '\P{^Decomposition_Type=-	Initial}', "");
Expect(0, 65268, '\p{Decomposition_Type=-	Initial}', "");
Expect(1, 65268, '\p{^Decomposition_Type=-	Initial}', "");
Expect(1, 65268, '\P{Decomposition_Type=-	Initial}', "");
Expect(0, 65268, '\P{^Decomposition_Type=-	Initial}', "");
Error('\p{Dt=/a/ INIT}');
Error('\P{Dt=/a/ INIT}');
Expect(1, 65267, '\p{Dt=init}', "");
Expect(0, 65267, '\p{^Dt=init}', "");
Expect(0, 65267, '\P{Dt=init}', "");
Expect(1, 65267, '\P{^Dt=init}', "");
Expect(0, 65268, '\p{Dt=init}', "");
Expect(1, 65268, '\p{^Dt=init}', "");
Expect(1, 65268, '\P{Dt=init}', "");
Expect(0, 65268, '\P{^Dt=init}', "");
Expect(1, 65267, '\p{Dt= -Init}', "");
Expect(0, 65267, '\p{^Dt= -Init}', "");
Expect(0, 65267, '\P{Dt= -Init}', "");
Expect(1, 65267, '\P{^Dt= -Init}', "");
Expect(0, 65268, '\p{Dt= -Init}', "");
Expect(1, 65268, '\p{^Dt= -Init}', "");
Expect(1, 65268, '\P{Dt= -Init}', "");
Expect(0, 65268, '\P{^Dt= -Init}', "");
Error('\p{Is_Decomposition_Type=--INITIAL:=}');
Error('\P{Is_Decomposition_Type=--INITIAL:=}');
Expect(1, 65267, '\p{Is_Decomposition_Type=initial}', "");
Expect(0, 65267, '\p{^Is_Decomposition_Type=initial}', "");
Expect(0, 65267, '\P{Is_Decomposition_Type=initial}', "");
Expect(1, 65267, '\P{^Is_Decomposition_Type=initial}', "");
Expect(0, 65268, '\p{Is_Decomposition_Type=initial}', "");
Expect(1, 65268, '\p{^Is_Decomposition_Type=initial}', "");
Expect(1, 65268, '\P{Is_Decomposition_Type=initial}', "");
Expect(0, 65268, '\P{^Is_Decomposition_Type=initial}', "");
Expect(1, 65267, '\p{Is_Decomposition_Type=-_Initial}', "");
Expect(0, 65267, '\p{^Is_Decomposition_Type=-_Initial}', "");
Expect(0, 65267, '\P{Is_Decomposition_Type=-_Initial}', "");
Expect(1, 65267, '\P{^Is_Decomposition_Type=-_Initial}', "");
Expect(0, 65268, '\p{Is_Decomposition_Type=-_Initial}', "");
Expect(1, 65268, '\p{^Is_Decomposition_Type=-_Initial}', "");
Expect(1, 65268, '\P{Is_Decomposition_Type=-_Initial}', "");
Expect(0, 65268, '\P{^Is_Decomposition_Type=-_Initial}', "");
Error('\p{Is_Dt=	 Init/a/}');
Error('\P{Is_Dt=	 Init/a/}');
Expect(1, 65267, '\p{Is_Dt=init}', "");
Expect(0, 65267, '\p{^Is_Dt=init}', "");
Expect(0, 65267, '\P{Is_Dt=init}', "");
Expect(1, 65267, '\P{^Is_Dt=init}', "");
Expect(0, 65268, '\p{Is_Dt=init}', "");
Expect(1, 65268, '\p{^Is_Dt=init}', "");
Expect(1, 65268, '\P{Is_Dt=init}', "");
Expect(0, 65268, '\P{^Is_Dt=init}', "");
Expect(1, 65267, '\p{Is_Dt=-_init}', "");
Expect(0, 65267, '\p{^Is_Dt=-_init}', "");
Expect(0, 65267, '\P{Is_Dt=-_init}', "");
Expect(1, 65267, '\P{^Is_Dt=-_init}', "");
Expect(0, 65268, '\p{Is_Dt=-_init}', "");
Expect(1, 65268, '\p{^Is_Dt=-_init}', "");
Expect(1, 65268, '\P{Is_Dt=-_init}', "");
Expect(0, 65268, '\P{^Is_Dt=-_init}', "");
Error('\p{Decomposition_Type:   _Isolated/a/}');
Error('\P{Decomposition_Type:   _Isolated/a/}');
Expect(1, 65275, '\p{Decomposition_Type=isolated}', "");
Expect(0, 65275, '\p{^Decomposition_Type=isolated}', "");
Expect(0, 65275, '\P{Decomposition_Type=isolated}', "");
Expect(1, 65275, '\P{^Decomposition_Type=isolated}', "");
Expect(0, 65276, '\p{Decomposition_Type=isolated}', "");
Expect(1, 65276, '\p{^Decomposition_Type=isolated}', "");
Expect(1, 65276, '\P{Decomposition_Type=isolated}', "");
Expect(0, 65276, '\P{^Decomposition_Type=isolated}', "");
Expect(1, 65275, '\p{Decomposition_Type=	_Isolated}', "");
Expect(0, 65275, '\p{^Decomposition_Type=	_Isolated}', "");
Expect(0, 65275, '\P{Decomposition_Type=	_Isolated}', "");
Expect(1, 65275, '\P{^Decomposition_Type=	_Isolated}', "");
Expect(0, 65276, '\p{Decomposition_Type=	_Isolated}', "");
Expect(1, 65276, '\p{^Decomposition_Type=	_Isolated}', "");
Expect(1, 65276, '\P{Decomposition_Type=	_Isolated}', "");
Expect(0, 65276, '\P{^Decomposition_Type=	_Isolated}', "");
Error('\p{Dt=:=	 iso}');
Error('\P{Dt=:=	 iso}');
Expect(1, 65275, '\p{Dt=iso}', "");
Expect(0, 65275, '\p{^Dt=iso}', "");
Expect(0, 65275, '\P{Dt=iso}', "");
Expect(1, 65275, '\P{^Dt=iso}', "");
Expect(0, 65276, '\p{Dt=iso}', "");
Expect(1, 65276, '\p{^Dt=iso}', "");
Expect(1, 65276, '\P{Dt=iso}', "");
Expect(0, 65276, '\P{^Dt=iso}', "");
Expect(1, 65275, '\p{Dt=_ Iso}', "");
Expect(0, 65275, '\p{^Dt=_ Iso}', "");
Expect(0, 65275, '\P{Dt=_ Iso}', "");
Expect(1, 65275, '\P{^Dt=_ Iso}', "");
Expect(0, 65276, '\p{Dt=_ Iso}', "");
Expect(1, 65276, '\p{^Dt=_ Iso}', "");
Expect(1, 65276, '\P{Dt=_ Iso}', "");
Expect(0, 65276, '\P{^Dt=_ Iso}', "");
Error('\p{Is_Decomposition_Type:   :=ISOLATED}');
Error('\P{Is_Decomposition_Type:   :=ISOLATED}');
Expect(1, 65275, '\p{Is_Decomposition_Type=isolated}', "");
Expect(0, 65275, '\p{^Is_Decomposition_Type=isolated}', "");
Expect(0, 65275, '\P{Is_Decomposition_Type=isolated}', "");
Expect(1, 65275, '\P{^Is_Decomposition_Type=isolated}', "");
Expect(0, 65276, '\p{Is_Decomposition_Type=isolated}', "");
Expect(1, 65276, '\p{^Is_Decomposition_Type=isolated}', "");
Expect(1, 65276, '\P{Is_Decomposition_Type=isolated}', "");
Expect(0, 65276, '\P{^Is_Decomposition_Type=isolated}', "");
Expect(1, 65275, '\p{Is_Decomposition_Type=	-isolated}', "");
Expect(0, 65275, '\p{^Is_Decomposition_Type=	-isolated}', "");
Expect(0, 65275, '\P{Is_Decomposition_Type=	-isolated}', "");
Expect(1, 65275, '\P{^Is_Decomposition_Type=	-isolated}', "");
Expect(0, 65276, '\p{Is_Decomposition_Type=	-isolated}', "");
Expect(1, 65276, '\p{^Is_Decomposition_Type=	-isolated}', "");
Expect(1, 65276, '\P{Is_Decomposition_Type=	-isolated}', "");
Expect(0, 65276, '\P{^Is_Decomposition_Type=	-isolated}', "");
Error('\p{Is_Dt=/a/-iso}');
Error('\P{Is_Dt=/a/-iso}');
Expect(1, 65275, '\p{Is_Dt:   iso}', "");
Expect(0, 65275, '\p{^Is_Dt:   iso}', "");
Expect(0, 65275, '\P{Is_Dt:   iso}', "");
Expect(1, 65275, '\P{^Is_Dt:   iso}', "");
Expect(0, 65276, '\p{Is_Dt:   iso}', "");
Expect(1, 65276, '\p{^Is_Dt:   iso}', "");
Expect(1, 65276, '\P{Is_Dt:   iso}', "");
Expect(0, 65276, '\P{^Is_Dt:   iso}', "");
Expect(1, 65275, '\p{Is_Dt: _ Iso}', "");
Expect(0, 65275, '\p{^Is_Dt: _ Iso}', "");
Expect(0, 65275, '\P{Is_Dt: _ Iso}', "");
Expect(1, 65275, '\P{^Is_Dt: _ Iso}', "");
Expect(0, 65276, '\p{Is_Dt: _ Iso}', "");
Expect(1, 65276, '\p{^Is_Dt: _ Iso}', "");
Expect(1, 65276, '\P{Is_Dt: _ Iso}', "");
Expect(0, 65276, '\P{^Is_Dt: _ Iso}', "");
Error('\p{Decomposition_Type=-:=Medial}');
Error('\P{Decomposition_Type=-:=Medial}');
Expect(1, 65268, '\p{Decomposition_Type=medial}', "");
Expect(0, 65268, '\p{^Decomposition_Type=medial}', "");
Expect(0, 65268, '\P{Decomposition_Type=medial}', "");
Expect(1, 65268, '\P{^Decomposition_Type=medial}', "");
Expect(0, 65269, '\p{Decomposition_Type=medial}', "");
Expect(1, 65269, '\p{^Decomposition_Type=medial}', "");
Expect(1, 65269, '\P{Decomposition_Type=medial}', "");
Expect(0, 65269, '\P{^Decomposition_Type=medial}', "");
Expect(1, 65268, '\p{Decomposition_Type= Medial}', "");
Expect(0, 65268, '\p{^Decomposition_Type= Medial}', "");
Expect(0, 65268, '\P{Decomposition_Type= Medial}', "");
Expect(1, 65268, '\P{^Decomposition_Type= Medial}', "");
Expect(0, 65269, '\p{Decomposition_Type= Medial}', "");
Expect(1, 65269, '\p{^Decomposition_Type= Medial}', "");
Expect(1, 65269, '\P{Decomposition_Type= Medial}', "");
Expect(0, 65269, '\P{^Decomposition_Type= Medial}', "");
Error('\p{Dt: :=Med}');
Error('\P{Dt: :=Med}');
Expect(1, 65268, '\p{Dt=med}', "");
Expect(0, 65268, '\p{^Dt=med}', "");
Expect(0, 65268, '\P{Dt=med}', "");
Expect(1, 65268, '\P{^Dt=med}', "");
Expect(0, 65269, '\p{Dt=med}', "");
Expect(1, 65269, '\p{^Dt=med}', "");
Expect(1, 65269, '\P{Dt=med}', "");
Expect(0, 65269, '\P{^Dt=med}', "");
Expect(1, 65268, '\p{Dt= 	Med}', "");
Expect(0, 65268, '\p{^Dt= 	Med}', "");
Expect(0, 65268, '\P{Dt= 	Med}', "");
Expect(1, 65268, '\P{^Dt= 	Med}', "");
Expect(0, 65269, '\p{Dt= 	Med}', "");
Expect(1, 65269, '\p{^Dt= 	Med}', "");
Expect(1, 65269, '\P{Dt= 	Med}', "");
Expect(0, 65269, '\P{^Dt= 	Med}', "");
Error('\p{Is_Decomposition_Type=	/a/Medial}');
Error('\P{Is_Decomposition_Type=	/a/Medial}');
Expect(1, 65268, '\p{Is_Decomposition_Type=medial}', "");
Expect(0, 65268, '\p{^Is_Decomposition_Type=medial}', "");
Expect(0, 65268, '\P{Is_Decomposition_Type=medial}', "");
Expect(1, 65268, '\P{^Is_Decomposition_Type=medial}', "");
Expect(0, 65269, '\p{Is_Decomposition_Type=medial}', "");
Expect(1, 65269, '\p{^Is_Decomposition_Type=medial}', "");
Expect(1, 65269, '\P{Is_Decomposition_Type=medial}', "");
Expect(0, 65269, '\P{^Is_Decomposition_Type=medial}', "");
Expect(1, 65268, '\p{Is_Decomposition_Type=--MEDIAL}', "");
Expect(0, 65268, '\p{^Is_Decomposition_Type=--MEDIAL}', "");
Expect(0, 65268, '\P{Is_Decomposition_Type=--MEDIAL}', "");
Expect(1, 65268, '\P{^Is_Decomposition_Type=--MEDIAL}', "");
Expect(0, 65269, '\p{Is_Decomposition_Type=--MEDIAL}', "");
Expect(1, 65269, '\p{^Is_Decomposition_Type=--MEDIAL}', "");
Expect(1, 65269, '\P{Is_Decomposition_Type=--MEDIAL}', "");
Expect(0, 65269, '\P{^Is_Decomposition_Type=--MEDIAL}', "");
Error('\p{Is_Dt:		/a/Med}');
Error('\P{Is_Dt:		/a/Med}');
Expect(1, 65268, '\p{Is_Dt=med}', "");
Expect(0, 65268, '\p{^Is_Dt=med}', "");
Expect(0, 65268, '\P{Is_Dt=med}', "");
Expect(1, 65268, '\P{^Is_Dt=med}', "");
Expect(0, 65269, '\p{Is_Dt=med}', "");
Expect(1, 65269, '\p{^Is_Dt=med}', "");
Expect(1, 65269, '\P{Is_Dt=med}', "");
Expect(0, 65269, '\P{^Is_Dt=med}', "");
Expect(1, 65268, '\p{Is_Dt=--Med}', "");
Expect(0, 65268, '\p{^Is_Dt=--Med}', "");
Expect(0, 65268, '\P{Is_Dt=--Med}', "");
Expect(1, 65268, '\P{^Is_Dt=--Med}', "");
Expect(0, 65269, '\p{Is_Dt=--Med}', "");
Expect(1, 65269, '\p{^Is_Dt=--Med}', "");
Expect(1, 65269, '\P{Is_Dt=--Med}', "");
Expect(0, 65269, '\P{^Is_Dt=--Med}', "");
Error('\p{Decomposition_Type=:=	narrow}');
Error('\P{Decomposition_Type=:=	narrow}');
Expect(1, 65518, '\p{Decomposition_Type=narrow}', "");
Expect(0, 65518, '\p{^Decomposition_Type=narrow}', "");
Expect(0, 65518, '\P{Decomposition_Type=narrow}', "");
Expect(1, 65518, '\P{^Decomposition_Type=narrow}', "");
Expect(0, 65519, '\p{Decomposition_Type=narrow}', "");
Expect(1, 65519, '\p{^Decomposition_Type=narrow}', "");
Expect(1, 65519, '\P{Decomposition_Type=narrow}', "");
Expect(0, 65519, '\P{^Decomposition_Type=narrow}', "");
Expect(1, 65518, '\p{Decomposition_Type=__Narrow}', "");
Expect(0, 65518, '\p{^Decomposition_Type=__Narrow}', "");
Expect(0, 65518, '\P{Decomposition_Type=__Narrow}', "");
Expect(1, 65518, '\P{^Decomposition_Type=__Narrow}', "");
Expect(0, 65519, '\p{Decomposition_Type=__Narrow}', "");
Expect(1, 65519, '\p{^Decomposition_Type=__Narrow}', "");
Expect(1, 65519, '\P{Decomposition_Type=__Narrow}', "");
Expect(0, 65519, '\P{^Decomposition_Type=__Narrow}', "");
Error('\p{Dt=/a/ _Nar}');
Error('\P{Dt=/a/ _Nar}');
Expect(1, 65518, '\p{Dt=nar}', "");
Expect(0, 65518, '\p{^Dt=nar}', "");
Expect(0, 65518, '\P{Dt=nar}', "");
Expect(1, 65518, '\P{^Dt=nar}', "");
Expect(0, 65519, '\p{Dt=nar}', "");
Expect(1, 65519, '\p{^Dt=nar}', "");
Expect(1, 65519, '\P{Dt=nar}', "");
Expect(0, 65519, '\P{^Dt=nar}', "");
Expect(1, 65518, '\p{Dt=	_nar}', "");
Expect(0, 65518, '\p{^Dt=	_nar}', "");
Expect(0, 65518, '\P{Dt=	_nar}', "");
Expect(1, 65518, '\P{^Dt=	_nar}', "");
Expect(0, 65519, '\p{Dt=	_nar}', "");
Expect(1, 65519, '\p{^Dt=	_nar}', "");
Expect(1, 65519, '\P{Dt=	_nar}', "");
Expect(0, 65519, '\P{^Dt=	_nar}', "");
Error('\p{Is_Decomposition_Type=_/a/narrow}');
Error('\P{Is_Decomposition_Type=_/a/narrow}');
Expect(1, 65518, '\p{Is_Decomposition_Type=narrow}', "");
Expect(0, 65518, '\p{^Is_Decomposition_Type=narrow}', "");
Expect(0, 65518, '\P{Is_Decomposition_Type=narrow}', "");
Expect(1, 65518, '\P{^Is_Decomposition_Type=narrow}', "");
Expect(0, 65519, '\p{Is_Decomposition_Type=narrow}', "");
Expect(1, 65519, '\p{^Is_Decomposition_Type=narrow}', "");
Expect(1, 65519, '\P{Is_Decomposition_Type=narrow}', "");
Expect(0, 65519, '\P{^Is_Decomposition_Type=narrow}', "");
Expect(1, 65518, '\p{Is_Decomposition_Type=--Narrow}', "");
Expect(0, 65518, '\p{^Is_Decomposition_Type=--Narrow}', "");
Expect(0, 65518, '\P{Is_Decomposition_Type=--Narrow}', "");
Expect(1, 65518, '\P{^Is_Decomposition_Type=--Narrow}', "");
Expect(0, 65519, '\p{Is_Decomposition_Type=--Narrow}', "");
Expect(1, 65519, '\p{^Is_Decomposition_Type=--Narrow}', "");
Expect(1, 65519, '\P{Is_Decomposition_Type=--Narrow}', "");
Expect(0, 65519, '\P{^Is_Decomposition_Type=--Narrow}', "");
Error('\p{Is_Dt:	/a/_Nar}');
Error('\P{Is_Dt:	/a/_Nar}');
Expect(1, 65518, '\p{Is_Dt=nar}', "");
Expect(0, 65518, '\p{^Is_Dt=nar}', "");
Expect(0, 65518, '\P{Is_Dt=nar}', "");
Expect(1, 65518, '\P{^Is_Dt=nar}', "");
Expect(0, 65519, '\p{Is_Dt=nar}', "");
Expect(1, 65519, '\p{^Is_Dt=nar}', "");
Expect(1, 65519, '\P{Is_Dt=nar}', "");
Expect(0, 65519, '\P{^Is_Dt=nar}', "");
Expect(1, 65518, '\p{Is_Dt=_Nar}', "");
Expect(0, 65518, '\p{^Is_Dt=_Nar}', "");
Expect(0, 65518, '\P{Is_Dt=_Nar}', "");
Expect(1, 65518, '\P{^Is_Dt=_Nar}', "");
Expect(0, 65519, '\p{Is_Dt=_Nar}', "");
Expect(1, 65519, '\p{^Is_Dt=_Nar}', "");
Expect(1, 65519, '\P{Is_Dt=_Nar}', "");
Expect(0, 65519, '\P{^Is_Dt=_Nar}', "");
Error('\p{Decomposition_Type=		Nobreak/a/}');
Error('\P{Decomposition_Type=		Nobreak/a/}');
Expect(1, 8239, '\p{Decomposition_Type=nobreak}', "");
Expect(0, 8239, '\p{^Decomposition_Type=nobreak}', "");
Expect(0, 8239, '\P{Decomposition_Type=nobreak}', "");
Expect(1, 8239, '\P{^Decomposition_Type=nobreak}', "");
Expect(0, 8240, '\p{Decomposition_Type=nobreak}', "");
Expect(1, 8240, '\p{^Decomposition_Type=nobreak}', "");
Expect(1, 8240, '\P{Decomposition_Type=nobreak}', "");
Expect(0, 8240, '\P{^Decomposition_Type=nobreak}', "");
Expect(1, 8239, '\p{Decomposition_Type=__Nobreak}', "");
Expect(0, 8239, '\p{^Decomposition_Type=__Nobreak}', "");
Expect(0, 8239, '\P{Decomposition_Type=__Nobreak}', "");
Expect(1, 8239, '\P{^Decomposition_Type=__Nobreak}', "");
Expect(0, 8240, '\p{Decomposition_Type=__Nobreak}', "");
Expect(1, 8240, '\p{^Decomposition_Type=__Nobreak}', "");
Expect(1, 8240, '\P{Decomposition_Type=__Nobreak}', "");
Expect(0, 8240, '\P{^Decomposition_Type=__Nobreak}', "");
Error('\p{Dt:   _NB:=}');
Error('\P{Dt:   _NB:=}');
Expect(1, 8239, '\p{Dt=nb}', "");
Expect(0, 8239, '\p{^Dt=nb}', "");
Expect(0, 8239, '\P{Dt=nb}', "");
Expect(1, 8239, '\P{^Dt=nb}', "");
Expect(0, 8240, '\p{Dt=nb}', "");
Expect(1, 8240, '\p{^Dt=nb}', "");
Expect(1, 8240, '\P{Dt=nb}', "");
Expect(0, 8240, '\P{^Dt=nb}', "");
Expect(1, 8239, '\p{Dt= _NB}', "");
Expect(0, 8239, '\p{^Dt= _NB}', "");
Expect(0, 8239, '\P{Dt= _NB}', "");
Expect(1, 8239, '\P{^Dt= _NB}', "");
Expect(0, 8240, '\p{Dt= _NB}', "");
Expect(1, 8240, '\p{^Dt= _NB}', "");
Expect(1, 8240, '\P{Dt= _NB}', "");
Expect(0, 8240, '\P{^Dt= _NB}', "");
Error('\p{Is_Decomposition_Type=_ Nobreak/a/}');
Error('\P{Is_Decomposition_Type=_ Nobreak/a/}');
Expect(1, 8239, '\p{Is_Decomposition_Type=nobreak}', "");
Expect(0, 8239, '\p{^Is_Decomposition_Type=nobreak}', "");
Expect(0, 8239, '\P{Is_Decomposition_Type=nobreak}', "");
Expect(1, 8239, '\P{^Is_Decomposition_Type=nobreak}', "");
Expect(0, 8240, '\p{Is_Decomposition_Type=nobreak}', "");
Expect(1, 8240, '\p{^Is_Decomposition_Type=nobreak}', "");
Expect(1, 8240, '\P{Is_Decomposition_Type=nobreak}', "");
Expect(0, 8240, '\P{^Is_Decomposition_Type=nobreak}', "");
Expect(1, 8239, '\p{Is_Decomposition_Type=	-NOBREAK}', "");
Expect(0, 8239, '\p{^Is_Decomposition_Type=	-NOBREAK}', "");
Expect(0, 8239, '\P{Is_Decomposition_Type=	-NOBREAK}', "");
Expect(1, 8239, '\P{^Is_Decomposition_Type=	-NOBREAK}', "");
Expect(0, 8240, '\p{Is_Decomposition_Type=	-NOBREAK}', "");
Expect(1, 8240, '\p{^Is_Decomposition_Type=	-NOBREAK}', "");
Expect(1, 8240, '\P{Is_Decomposition_Type=	-NOBREAK}', "");
Expect(0, 8240, '\P{^Is_Decomposition_Type=	-NOBREAK}', "");
Error('\p{Is_Dt=-_nb/a/}');
Error('\P{Is_Dt=-_nb/a/}');
Expect(1, 8239, '\p{Is_Dt=nb}', "");
Expect(0, 8239, '\p{^Is_Dt=nb}', "");
Expect(0, 8239, '\P{Is_Dt=nb}', "");
Expect(1, 8239, '\P{^Is_Dt=nb}', "");
Expect(0, 8240, '\p{Is_Dt=nb}', "");
Expect(1, 8240, '\p{^Is_Dt=nb}', "");
Expect(1, 8240, '\P{Is_Dt=nb}', "");
Expect(0, 8240, '\P{^Is_Dt=nb}', "");
Expect(1, 8239, '\p{Is_Dt=	_NB}', "");
Expect(0, 8239, '\p{^Is_Dt=	_NB}', "");
Expect(0, 8239, '\P{Is_Dt=	_NB}', "");
Expect(1, 8239, '\P{^Is_Dt=	_NB}', "");
Expect(0, 8240, '\p{Is_Dt=	_NB}', "");
Expect(1, 8240, '\p{^Is_Dt=	_NB}', "");
Expect(1, 8240, '\P{Is_Dt=	_NB}', "");
Expect(0, 8240, '\P{^Is_Dt=	_NB}', "");
Error('\p{Decomposition_Type=		Non_Canonical:=}');
Error('\P{Decomposition_Type=		Non_Canonical:=}');
Expect(1, 127569, '\p{Decomposition_Type=noncanonical}', "");
Expect(0, 127569, '\p{^Decomposition_Type=noncanonical}', "");
Expect(0, 127569, '\P{Decomposition_Type=noncanonical}', "");
Expect(1, 127569, '\P{^Decomposition_Type=noncanonical}', "");
Expect(0, 127570, '\p{Decomposition_Type=noncanonical}', "");
Expect(1, 127570, '\p{^Decomposition_Type=noncanonical}', "");
Expect(1, 127570, '\P{Decomposition_Type=noncanonical}', "");
Expect(0, 127570, '\P{^Decomposition_Type=noncanonical}', "");
Expect(1, 127569, '\p{Decomposition_Type: -NON_Canonical}', "");
Expect(0, 127569, '\p{^Decomposition_Type: -NON_Canonical}', "");
Expect(0, 127569, '\P{Decomposition_Type: -NON_Canonical}', "");
Expect(1, 127569, '\P{^Decomposition_Type: -NON_Canonical}', "");
Expect(0, 127570, '\p{Decomposition_Type: -NON_Canonical}', "");
Expect(1, 127570, '\p{^Decomposition_Type: -NON_Canonical}', "");
Expect(1, 127570, '\P{Decomposition_Type: -NON_Canonical}', "");
Expect(0, 127570, '\P{^Decomposition_Type: -NON_Canonical}', "");
Error('\p{Dt= /a/NON_canon}');
Error('\P{Dt= /a/NON_canon}');
Expect(1, 127569, '\p{Dt=noncanon}', "");
Expect(0, 127569, '\p{^Dt=noncanon}', "");
Expect(0, 127569, '\P{Dt=noncanon}', "");
Expect(1, 127569, '\P{^Dt=noncanon}', "");
Expect(0, 127570, '\p{Dt=noncanon}', "");
Expect(1, 127570, '\p{^Dt=noncanon}', "");
Expect(1, 127570, '\P{Dt=noncanon}', "");
Expect(0, 127570, '\P{^Dt=noncanon}', "");
Expect(1, 127569, '\p{Dt= -NON_Canon}', "");
Expect(0, 127569, '\p{^Dt= -NON_Canon}', "");
Expect(0, 127569, '\P{Dt= -NON_Canon}', "");
Expect(1, 127569, '\P{^Dt= -NON_Canon}', "");
Expect(0, 127570, '\p{Dt= -NON_Canon}', "");
Expect(1, 127570, '\p{^Dt= -NON_Canon}', "");
Expect(1, 127570, '\P{Dt= -NON_Canon}', "");
Expect(0, 127570, '\P{^Dt= -NON_Canon}', "");
Error('\p{Is_Decomposition_Type=-	Non_CANONICAL/a/}');
Error('\P{Is_Decomposition_Type=-	Non_CANONICAL/a/}');
Expect(1, 127569, '\p{Is_Decomposition_Type=noncanonical}', "");
Expect(0, 127569, '\p{^Is_Decomposition_Type=noncanonical}', "");
Expect(0, 127569, '\P{Is_Decomposition_Type=noncanonical}', "");
Expect(1, 127569, '\P{^Is_Decomposition_Type=noncanonical}', "");
Expect(0, 127570, '\p{Is_Decomposition_Type=noncanonical}', "");
Expect(1, 127570, '\p{^Is_Decomposition_Type=noncanonical}', "");
Expect(1, 127570, '\P{Is_Decomposition_Type=noncanonical}', "");
Expect(0, 127570, '\P{^Is_Decomposition_Type=noncanonical}', "");
Expect(1, 127569, '\p{Is_Decomposition_Type: 		NON_Canonical}', "");
Expect(0, 127569, '\p{^Is_Decomposition_Type: 		NON_Canonical}', "");
Expect(0, 127569, '\P{Is_Decomposition_Type: 		NON_Canonical}', "");
Expect(1, 127569, '\P{^Is_Decomposition_Type: 		NON_Canonical}', "");
Expect(0, 127570, '\p{Is_Decomposition_Type: 		NON_Canonical}', "");
Expect(1, 127570, '\p{^Is_Decomposition_Type: 		NON_Canonical}', "");
Expect(1, 127570, '\P{Is_Decomposition_Type: 		NON_Canonical}', "");
Expect(0, 127570, '\P{^Is_Decomposition_Type: 		NON_Canonical}', "");
Error('\p{Is_Dt=/a/ non_Canon}');
Error('\P{Is_Dt=/a/ non_Canon}');
Expect(1, 127569, '\p{Is_Dt=noncanon}', "");
Expect(0, 127569, '\p{^Is_Dt=noncanon}', "");
Expect(0, 127569, '\P{Is_Dt=noncanon}', "");
Expect(1, 127569, '\P{^Is_Dt=noncanon}', "");
Expect(0, 127570, '\p{Is_Dt=noncanon}', "");
Expect(1, 127570, '\p{^Is_Dt=noncanon}', "");
Expect(1, 127570, '\P{Is_Dt=noncanon}', "");
Expect(0, 127570, '\P{^Is_Dt=noncanon}', "");
Expect(1, 127569, '\p{Is_Dt=-	Non_Canon}', "");
Expect(0, 127569, '\p{^Is_Dt=-	Non_Canon}', "");
Expect(0, 127569, '\P{Is_Dt=-	Non_Canon}', "");
Expect(1, 127569, '\P{^Is_Dt=-	Non_Canon}', "");
Expect(0, 127570, '\p{Is_Dt=-	Non_Canon}', "");
Expect(1, 127570, '\p{^Is_Dt=-	Non_Canon}', "");
Expect(1, 127570, '\P{Is_Dt=-	Non_Canon}', "");
Expect(0, 127570, '\P{^Is_Dt=-	Non_Canon}', "");
Error('\p{Decomposition_Type= :=none}');
Error('\P{Decomposition_Type= :=none}');
Expect(1, 195102, '\p{Decomposition_Type:	none}', "");
Expect(0, 195102, '\p{^Decomposition_Type:	none}', "");
Expect(0, 195102, '\P{Decomposition_Type:	none}', "");
Expect(1, 195102, '\P{^Decomposition_Type:	none}', "");
Expect(0, 195101, '\p{Decomposition_Type:	none}', "");
Expect(1, 195101, '\p{^Decomposition_Type:	none}', "");
Expect(1, 195101, '\P{Decomposition_Type:	none}', "");
Expect(0, 195101, '\P{^Decomposition_Type:	none}', "");
Expect(1, 195102, '\p{Decomposition_Type=	NONE}', "");
Expect(0, 195102, '\p{^Decomposition_Type=	NONE}', "");
Expect(0, 195102, '\P{Decomposition_Type=	NONE}', "");
Expect(1, 195102, '\P{^Decomposition_Type=	NONE}', "");
Expect(0, 195101, '\p{Decomposition_Type=	NONE}', "");
Expect(1, 195101, '\p{^Decomposition_Type=	NONE}', "");
Expect(1, 195101, '\P{Decomposition_Type=	NONE}', "");
Expect(0, 195101, '\P{^Decomposition_Type=	NONE}', "");
Error('\p{Dt=_	NONE/a/}');
Error('\P{Dt=_	NONE/a/}');
Expect(1, 195102, '\p{Dt=none}', "");
Expect(0, 195102, '\p{^Dt=none}', "");
Expect(0, 195102, '\P{Dt=none}', "");
Expect(1, 195102, '\P{^Dt=none}', "");
Expect(0, 195101, '\p{Dt=none}', "");
Expect(1, 195101, '\p{^Dt=none}', "");
Expect(1, 195101, '\P{Dt=none}', "");
Expect(0, 195101, '\P{^Dt=none}', "");
Expect(1, 195102, '\p{Dt=-None}', "");
Expect(0, 195102, '\p{^Dt=-None}', "");
Expect(0, 195102, '\P{Dt=-None}', "");
Expect(1, 195102, '\P{^Dt=-None}', "");
Expect(0, 195101, '\p{Dt=-None}', "");
Expect(1, 195101, '\p{^Dt=-None}', "");
Expect(1, 195101, '\P{Dt=-None}', "");
Expect(0, 195101, '\P{^Dt=-None}', "");
Error('\p{Is_Decomposition_Type=/a/		None}');
Error('\P{Is_Decomposition_Type=/a/		None}');
Expect(1, 195102, '\p{Is_Decomposition_Type=none}', "");
Expect(0, 195102, '\p{^Is_Decomposition_Type=none}', "");
Expect(0, 195102, '\P{Is_Decomposition_Type=none}', "");
Expect(1, 195102, '\P{^Is_Decomposition_Type=none}', "");
Expect(0, 195101, '\p{Is_Decomposition_Type=none}', "");
Expect(1, 195101, '\p{^Is_Decomposition_Type=none}', "");
Expect(1, 195101, '\P{Is_Decomposition_Type=none}', "");
Expect(0, 195101, '\P{^Is_Decomposition_Type=none}', "");
Expect(1, 195102, '\p{Is_Decomposition_Type=	None}', "");
Expect(0, 195102, '\p{^Is_Decomposition_Type=	None}', "");
Expect(0, 195102, '\P{Is_Decomposition_Type=	None}', "");
Expect(1, 195102, '\P{^Is_Decomposition_Type=	None}', "");
Expect(0, 195101, '\p{Is_Decomposition_Type=	None}', "");
Expect(1, 195101, '\p{^Is_Decomposition_Type=	None}', "");
Expect(1, 195101, '\P{Is_Decomposition_Type=	None}', "");
Expect(0, 195101, '\P{^Is_Decomposition_Type=	None}', "");
Error('\p{Is_Dt=:=	-NONE}');
Error('\P{Is_Dt=:=	-NONE}');
Expect(1, 195102, '\p{Is_Dt=none}', "");
Expect(0, 195102, '\p{^Is_Dt=none}', "");
Expect(0, 195102, '\P{Is_Dt=none}', "");
Expect(1, 195102, '\P{^Is_Dt=none}', "");
Expect(0, 195101, '\p{Is_Dt=none}', "");
Expect(1, 195101, '\p{^Is_Dt=none}', "");
Expect(1, 195101, '\P{Is_Dt=none}', "");
Expect(0, 195101, '\P{^Is_Dt=none}', "");
Expect(1, 195102, '\p{Is_Dt=__None}', "");
Expect(0, 195102, '\p{^Is_Dt=__None}', "");
Expect(0, 195102, '\P{Is_Dt=__None}', "");
Expect(1, 195102, '\P{^Is_Dt=__None}', "");
Expect(0, 195101, '\p{Is_Dt=__None}', "");
Expect(1, 195101, '\p{^Is_Dt=__None}', "");
Expect(1, 195101, '\P{Is_Dt=__None}', "");
Expect(0, 195101, '\P{^Is_Dt=__None}', "");
Error('\p{Decomposition_Type=-_SMALL:=}');
Error('\P{Decomposition_Type=-_SMALL:=}');
Expect(1, 65131, '\p{Decomposition_Type=small}', "");
Expect(0, 65131, '\p{^Decomposition_Type=small}', "");
Expect(0, 65131, '\P{Decomposition_Type=small}', "");
Expect(1, 65131, '\P{^Decomposition_Type=small}', "");
Expect(0, 65132, '\p{Decomposition_Type=small}', "");
Expect(1, 65132, '\p{^Decomposition_Type=small}', "");
Expect(1, 65132, '\P{Decomposition_Type=small}', "");
Expect(0, 65132, '\P{^Decomposition_Type=small}', "");
Expect(1, 65131, '\p{Decomposition_Type=-Small}', "");
Expect(0, 65131, '\p{^Decomposition_Type=-Small}', "");
Expect(0, 65131, '\P{Decomposition_Type=-Small}', "");
Expect(1, 65131, '\P{^Decomposition_Type=-Small}', "");
Expect(0, 65132, '\p{Decomposition_Type=-Small}', "");
Expect(1, 65132, '\p{^Decomposition_Type=-Small}', "");
Expect(1, 65132, '\P{Decomposition_Type=-Small}', "");
Expect(0, 65132, '\P{^Decomposition_Type=-Small}', "");
Error('\p{Dt=		SML/a/}');
Error('\P{Dt=		SML/a/}');
Expect(1, 65131, '\p{Dt=sml}', "");
Expect(0, 65131, '\p{^Dt=sml}', "");
Expect(0, 65131, '\P{Dt=sml}', "");
Expect(1, 65131, '\P{^Dt=sml}', "");
Expect(0, 65132, '\p{Dt=sml}', "");
Expect(1, 65132, '\p{^Dt=sml}', "");
Expect(1, 65132, '\P{Dt=sml}', "");
Expect(0, 65132, '\P{^Dt=sml}', "");
Expect(1, 65131, '\p{Dt=_Sml}', "");
Expect(0, 65131, '\p{^Dt=_Sml}', "");
Expect(0, 65131, '\P{Dt=_Sml}', "");
Expect(1, 65131, '\P{^Dt=_Sml}', "");
Expect(0, 65132, '\p{Dt=_Sml}', "");
Expect(1, 65132, '\p{^Dt=_Sml}', "");
Expect(1, 65132, '\P{Dt=_Sml}', "");
Expect(0, 65132, '\P{^Dt=_Sml}', "");
Error('\p{Is_Decomposition_Type=/a/	Small}');
Error('\P{Is_Decomposition_Type=/a/	Small}');
Expect(1, 65131, '\p{Is_Decomposition_Type=small}', "");
Expect(0, 65131, '\p{^Is_Decomposition_Type=small}', "");
Expect(0, 65131, '\P{Is_Decomposition_Type=small}', "");
Expect(1, 65131, '\P{^Is_Decomposition_Type=small}', "");
Expect(0, 65132, '\p{Is_Decomposition_Type=small}', "");
Expect(1, 65132, '\p{^Is_Decomposition_Type=small}', "");
Expect(1, 65132, '\P{Is_Decomposition_Type=small}', "");
Expect(0, 65132, '\P{^Is_Decomposition_Type=small}', "");
Expect(1, 65131, '\p{Is_Decomposition_Type=	_small}', "");
Expect(0, 65131, '\p{^Is_Decomposition_Type=	_small}', "");
Expect(0, 65131, '\P{Is_Decomposition_Type=	_small}', "");
Expect(1, 65131, '\P{^Is_Decomposition_Type=	_small}', "");
Expect(0, 65132, '\p{Is_Decomposition_Type=	_small}', "");
Expect(1, 65132, '\p{^Is_Decomposition_Type=	_small}', "");
Expect(1, 65132, '\P{Is_Decomposition_Type=	_small}', "");
Expect(0, 65132, '\P{^Is_Decomposition_Type=	_small}', "");
Error('\p{Is_Dt=_SML:=}');
Error('\P{Is_Dt=_SML:=}');
Expect(1, 65131, '\p{Is_Dt=sml}', "");
Expect(0, 65131, '\p{^Is_Dt=sml}', "");
Expect(0, 65131, '\P{Is_Dt=sml}', "");
Expect(1, 65131, '\P{^Is_Dt=sml}', "");
Expect(0, 65132, '\p{Is_Dt=sml}', "");
Expect(1, 65132, '\p{^Is_Dt=sml}', "");
Expect(1, 65132, '\P{Is_Dt=sml}', "");
Expect(0, 65132, '\P{^Is_Dt=sml}', "");
Expect(1, 65131, '\p{Is_Dt=		Sml}', "");
Expect(0, 65131, '\p{^Is_Dt=		Sml}', "");
Expect(0, 65131, '\P{Is_Dt=		Sml}', "");
Expect(1, 65131, '\P{^Is_Dt=		Sml}', "");
Expect(0, 65132, '\p{Is_Dt=		Sml}', "");
Expect(1, 65132, '\p{^Is_Dt=		Sml}', "");
Expect(1, 65132, '\P{Is_Dt=		Sml}', "");
Expect(0, 65132, '\P{^Is_Dt=		Sml}', "");
Error('\p{Decomposition_Type=:=-_Square}');
Error('\P{Decomposition_Type=:=-_Square}');
Expect(1, 127546, '\p{Decomposition_Type=square}', "");
Expect(0, 127546, '\p{^Decomposition_Type=square}', "");
Expect(0, 127546, '\P{Decomposition_Type=square}', "");
Expect(1, 127546, '\P{^Decomposition_Type=square}', "");
Expect(0, 127547, '\p{Decomposition_Type=square}', "");
Expect(1, 127547, '\p{^Decomposition_Type=square}', "");
Expect(1, 127547, '\P{Decomposition_Type=square}', "");
Expect(0, 127547, '\P{^Decomposition_Type=square}', "");
Expect(1, 127546, '\p{Decomposition_Type:  square}', "");
Expect(0, 127546, '\p{^Decomposition_Type:  square}', "");
Expect(0, 127546, '\P{Decomposition_Type:  square}', "");
Expect(1, 127546, '\P{^Decomposition_Type:  square}', "");
Expect(0, 127547, '\p{Decomposition_Type:  square}', "");
Expect(1, 127547, '\p{^Decomposition_Type:  square}', "");
Expect(1, 127547, '\P{Decomposition_Type:  square}', "");
Expect(0, 127547, '\P{^Decomposition_Type:  square}', "");
Error('\p{Dt=-_SQR:=}');
Error('\P{Dt=-_SQR:=}');
Expect(1, 127546, '\p{Dt=sqr}', "");
Expect(0, 127546, '\p{^Dt=sqr}', "");
Expect(0, 127546, '\P{Dt=sqr}', "");
Expect(1, 127546, '\P{^Dt=sqr}', "");
Expect(0, 127547, '\p{Dt=sqr}', "");
Expect(1, 127547, '\p{^Dt=sqr}', "");
Expect(1, 127547, '\P{Dt=sqr}', "");
Expect(0, 127547, '\P{^Dt=sqr}', "");
Expect(1, 127546, '\p{Dt:	_	SQR}', "");
Expect(0, 127546, '\p{^Dt:	_	SQR}', "");
Expect(0, 127546, '\P{Dt:	_	SQR}', "");
Expect(1, 127546, '\P{^Dt:	_	SQR}', "");
Expect(0, 127547, '\p{Dt:	_	SQR}', "");
Expect(1, 127547, '\p{^Dt:	_	SQR}', "");
Expect(1, 127547, '\P{Dt:	_	SQR}', "");
Expect(0, 127547, '\P{^Dt:	_	SQR}', "");
Error('\p{Is_Decomposition_Type=  Square/a/}');
Error('\P{Is_Decomposition_Type=  Square/a/}');
Expect(1, 127546, '\p{Is_Decomposition_Type=square}', "");
Expect(0, 127546, '\p{^Is_Decomposition_Type=square}', "");
Expect(0, 127546, '\P{Is_Decomposition_Type=square}', "");
Expect(1, 127546, '\P{^Is_Decomposition_Type=square}', "");
Expect(0, 127547, '\p{Is_Decomposition_Type=square}', "");
Expect(1, 127547, '\p{^Is_Decomposition_Type=square}', "");
Expect(1, 127547, '\P{Is_Decomposition_Type=square}', "");
Expect(0, 127547, '\P{^Is_Decomposition_Type=square}', "");
Expect(1, 127546, '\p{Is_Decomposition_Type=	_square}', "");
Expect(0, 127546, '\p{^Is_Decomposition_Type=	_square}', "");
Expect(0, 127546, '\P{Is_Decomposition_Type=	_square}', "");
Expect(1, 127546, '\P{^Is_Decomposition_Type=	_square}', "");
Expect(0, 127547, '\p{Is_Decomposition_Type=	_square}', "");
Expect(1, 127547, '\p{^Is_Decomposition_Type=	_square}', "");
Expect(1, 127547, '\P{Is_Decomposition_Type=	_square}', "");
Expect(0, 127547, '\P{^Is_Decomposition_Type=	_square}', "");
Error('\p{Is_Dt=_	Sqr:=}');
Error('\P{Is_Dt=_	Sqr:=}');
Expect(1, 127546, '\p{Is_Dt=sqr}', "");
Expect(0, 127546, '\p{^Is_Dt=sqr}', "");
Expect(0, 127546, '\P{Is_Dt=sqr}', "");
Expect(1, 127546, '\P{^Is_Dt=sqr}', "");
Expect(0, 127547, '\p{Is_Dt=sqr}', "");
Expect(1, 127547, '\p{^Is_Dt=sqr}', "");
Expect(1, 127547, '\P{Is_Dt=sqr}', "");
Expect(0, 127547, '\P{^Is_Dt=sqr}', "");
Expect(1, 127546, '\p{Is_Dt=_sqr}', "");
Expect(0, 127546, '\p{^Is_Dt=_sqr}', "");
Expect(0, 127546, '\P{Is_Dt=_sqr}', "");
Expect(1, 127546, '\P{^Is_Dt=_sqr}', "");
Expect(0, 127547, '\p{Is_Dt=_sqr}', "");
Expect(1, 127547, '\p{^Is_Dt=_sqr}', "");
Expect(1, 127547, '\P{Is_Dt=_sqr}', "");
Expect(0, 127547, '\P{^Is_Dt=_sqr}', "");
Error('\p{Decomposition_Type=:= 	SUB}');
Error('\P{Decomposition_Type=:= 	SUB}');
Expect(1, 11388, '\p{Decomposition_Type=sub}', "");
Expect(0, 11388, '\p{^Decomposition_Type=sub}', "");
Expect(0, 11388, '\P{Decomposition_Type=sub}', "");
Expect(1, 11388, '\P{^Decomposition_Type=sub}', "");
Expect(0, 11389, '\p{Decomposition_Type=sub}', "");
Expect(1, 11389, '\p{^Decomposition_Type=sub}', "");
Expect(1, 11389, '\P{Decomposition_Type=sub}', "");
Expect(0, 11389, '\P{^Decomposition_Type=sub}', "");
Expect(1, 11388, '\p{Decomposition_Type=	_sub}', "");
Expect(0, 11388, '\p{^Decomposition_Type=	_sub}', "");
Expect(0, 11388, '\P{Decomposition_Type=	_sub}', "");
Expect(1, 11388, '\P{^Decomposition_Type=	_sub}', "");
Expect(0, 11389, '\p{Decomposition_Type=	_sub}', "");
Expect(1, 11389, '\p{^Decomposition_Type=	_sub}', "");
Expect(1, 11389, '\P{Decomposition_Type=	_sub}', "");
Expect(0, 11389, '\P{^Decomposition_Type=	_sub}', "");
Error('\p{Dt=:=Sub}');
Error('\P{Dt=:=Sub}');
Expect(1, 11388, '\p{Dt:	sub}', "");
Expect(0, 11388, '\p{^Dt:	sub}', "");
Expect(0, 11388, '\P{Dt:	sub}', "");
Expect(1, 11388, '\P{^Dt:	sub}', "");
Expect(0, 11389, '\p{Dt:	sub}', "");
Expect(1, 11389, '\p{^Dt:	sub}', "");
Expect(1, 11389, '\P{Dt:	sub}', "");
Expect(0, 11389, '\P{^Dt:	sub}', "");
Expect(1, 11388, '\p{Dt=-_Sub}', "");
Expect(0, 11388, '\p{^Dt=-_Sub}', "");
Expect(0, 11388, '\P{Dt=-_Sub}', "");
Expect(1, 11388, '\P{^Dt=-_Sub}', "");
Expect(0, 11389, '\p{Dt=-_Sub}', "");
Expect(1, 11389, '\p{^Dt=-_Sub}', "");
Expect(1, 11389, '\P{Dt=-_Sub}', "");
Expect(0, 11389, '\P{^Dt=-_Sub}', "");
Error('\p{Is_Decomposition_Type=-	sub:=}');
Error('\P{Is_Decomposition_Type=-	sub:=}');
Expect(1, 11388, '\p{Is_Decomposition_Type=sub}', "");
Expect(0, 11388, '\p{^Is_Decomposition_Type=sub}', "");
Expect(0, 11388, '\P{Is_Decomposition_Type=sub}', "");
Expect(1, 11388, '\P{^Is_Decomposition_Type=sub}', "");
Expect(0, 11389, '\p{Is_Decomposition_Type=sub}', "");
Expect(1, 11389, '\p{^Is_Decomposition_Type=sub}', "");
Expect(1, 11389, '\P{Is_Decomposition_Type=sub}', "");
Expect(0, 11389, '\P{^Is_Decomposition_Type=sub}', "");
Expect(1, 11388, '\p{Is_Decomposition_Type=_-SUB}', "");
Expect(0, 11388, '\p{^Is_Decomposition_Type=_-SUB}', "");
Expect(0, 11388, '\P{Is_Decomposition_Type=_-SUB}', "");
Expect(1, 11388, '\P{^Is_Decomposition_Type=_-SUB}', "");
Expect(0, 11389, '\p{Is_Decomposition_Type=_-SUB}', "");
Expect(1, 11389, '\p{^Is_Decomposition_Type=_-SUB}', "");
Expect(1, 11389, '\P{Is_Decomposition_Type=_-SUB}', "");
Expect(0, 11389, '\P{^Is_Decomposition_Type=_-SUB}', "");
Error('\p{Is_Dt=:=		sub}');
Error('\P{Is_Dt=:=		sub}');
Expect(1, 11388, '\p{Is_Dt=sub}', "");
Expect(0, 11388, '\p{^Is_Dt=sub}', "");
Expect(0, 11388, '\P{Is_Dt=sub}', "");
Expect(1, 11388, '\P{^Is_Dt=sub}', "");
Expect(0, 11389, '\p{Is_Dt=sub}', "");
Expect(1, 11389, '\p{^Is_Dt=sub}', "");
Expect(1, 11389, '\P{Is_Dt=sub}', "");
Expect(0, 11389, '\P{^Is_Dt=sub}', "");
Expect(1, 11388, '\p{Is_Dt= -SUB}', "");
Expect(0, 11388, '\p{^Is_Dt= -SUB}', "");
Expect(0, 11388, '\P{Is_Dt= -SUB}', "");
Expect(1, 11388, '\P{^Is_Dt= -SUB}', "");
Expect(0, 11389, '\p{Is_Dt= -SUB}', "");
Expect(1, 11389, '\p{^Is_Dt= -SUB}', "");
Expect(1, 11389, '\P{Is_Dt= -SUB}', "");
Expect(0, 11389, '\P{^Is_Dt= -SUB}', "");
Error('\p{Decomposition_Type=	:=super}');
Error('\P{Decomposition_Type=	:=super}');
Expect(1, 127339, '\p{Decomposition_Type=super}', "");
Expect(0, 127339, '\p{^Decomposition_Type=super}', "");
Expect(0, 127339, '\P{Decomposition_Type=super}', "");
Expect(1, 127339, '\P{^Decomposition_Type=super}', "");
Expect(0, 127340, '\p{Decomposition_Type=super}', "");
Expect(1, 127340, '\p{^Decomposition_Type=super}', "");
Expect(1, 127340, '\P{Decomposition_Type=super}', "");
Expect(0, 127340, '\P{^Decomposition_Type=super}', "");
Expect(1, 127339, '\p{Decomposition_Type=	_super}', "");
Expect(0, 127339, '\p{^Decomposition_Type=	_super}', "");
Expect(0, 127339, '\P{Decomposition_Type=	_super}', "");
Expect(1, 127339, '\P{^Decomposition_Type=	_super}', "");
Expect(0, 127340, '\p{Decomposition_Type=	_super}', "");
Expect(1, 127340, '\p{^Decomposition_Type=	_super}', "");
Expect(1, 127340, '\P{Decomposition_Type=	_super}', "");
Expect(0, 127340, '\P{^Decomposition_Type=	_super}', "");
Error('\p{Dt=:=-sup}');
Error('\P{Dt=:=-sup}');
Expect(1, 127339, '\p{Dt=sup}', "");
Expect(0, 127339, '\p{^Dt=sup}', "");
Expect(0, 127339, '\P{Dt=sup}', "");
Expect(1, 127339, '\P{^Dt=sup}', "");
Expect(0, 127340, '\p{Dt=sup}', "");
Expect(1, 127340, '\p{^Dt=sup}', "");
Expect(1, 127340, '\P{Dt=sup}', "");
Expect(0, 127340, '\P{^Dt=sup}', "");
Expect(1, 127339, '\p{Dt= Sup}', "");
Expect(0, 127339, '\p{^Dt= Sup}', "");
Expect(0, 127339, '\P{Dt= Sup}', "");
Expect(1, 127339, '\P{^Dt= Sup}', "");
Expect(0, 127340, '\p{Dt= Sup}', "");
Expect(1, 127340, '\p{^Dt= Sup}', "");
Expect(1, 127340, '\P{Dt= Sup}', "");
Expect(0, 127340, '\P{^Dt= Sup}', "");
Error('\p{Is_Decomposition_Type=/a/	-SUPER}');
Error('\P{Is_Decomposition_Type=/a/	-SUPER}');
Expect(1, 127339, '\p{Is_Decomposition_Type=super}', "");
Expect(0, 127339, '\p{^Is_Decomposition_Type=super}', "");
Expect(0, 127339, '\P{Is_Decomposition_Type=super}', "");
Expect(1, 127339, '\P{^Is_Decomposition_Type=super}', "");
Expect(0, 127340, '\p{Is_Decomposition_Type=super}', "");
Expect(1, 127340, '\p{^Is_Decomposition_Type=super}', "");
Expect(1, 127340, '\P{Is_Decomposition_Type=super}', "");
Expect(0, 127340, '\P{^Is_Decomposition_Type=super}', "");
Expect(1, 127339, '\p{Is_Decomposition_Type=  super}', "");
Expect(0, 127339, '\p{^Is_Decomposition_Type=  super}', "");
Expect(0, 127339, '\P{Is_Decomposition_Type=  super}', "");
Expect(1, 127339, '\P{^Is_Decomposition_Type=  super}', "");
Expect(0, 127340, '\p{Is_Decomposition_Type=  super}', "");
Expect(1, 127340, '\p{^Is_Decomposition_Type=  super}', "");
Expect(1, 127340, '\P{Is_Decomposition_Type=  super}', "");
Expect(0, 127340, '\P{^Is_Decomposition_Type=  super}', "");
Error('\p{Is_Dt=_Sup/a/}');
Error('\P{Is_Dt=_Sup/a/}');
Expect(1, 127339, '\p{Is_Dt=sup}', "");
Expect(0, 127339, '\p{^Is_Dt=sup}', "");
Expect(0, 127339, '\P{Is_Dt=sup}', "");
Expect(1, 127339, '\P{^Is_Dt=sup}', "");
Expect(0, 127340, '\p{Is_Dt=sup}', "");
Expect(1, 127340, '\p{^Is_Dt=sup}', "");
Expect(1, 127340, '\P{Is_Dt=sup}', "");
Expect(0, 127340, '\P{^Is_Dt=sup}', "");
Expect(1, 127339, '\p{Is_Dt=_ Sup}', "");
Expect(0, 127339, '\p{^Is_Dt=_ Sup}', "");
Expect(0, 127339, '\P{Is_Dt=_ Sup}', "");
Expect(1, 127339, '\P{^Is_Dt=_ Sup}', "");
Expect(0, 127340, '\p{Is_Dt=_ Sup}', "");
Expect(1, 127340, '\p{^Is_Dt=_ Sup}', "");
Expect(1, 127340, '\P{Is_Dt=_ Sup}', "");
Expect(0, 127340, '\P{^Is_Dt=_ Sup}', "");
Error('\p{Decomposition_Type=	Vertical/a/}');
Error('\P{Decomposition_Type=	Vertical/a/}');
Expect(1, 65096, '\p{Decomposition_Type=vertical}', "");
Expect(0, 65096, '\p{^Decomposition_Type=vertical}', "");
Expect(0, 65096, '\P{Decomposition_Type=vertical}', "");
Expect(1, 65096, '\P{^Decomposition_Type=vertical}', "");
Expect(0, 65097, '\p{Decomposition_Type=vertical}', "");
Expect(1, 65097, '\p{^Decomposition_Type=vertical}', "");
Expect(1, 65097, '\P{Decomposition_Type=vertical}', "");
Expect(0, 65097, '\P{^Decomposition_Type=vertical}', "");
Expect(1, 65096, '\p{Decomposition_Type=_-VERTICAL}', "");
Expect(0, 65096, '\p{^Decomposition_Type=_-VERTICAL}', "");
Expect(0, 65096, '\P{Decomposition_Type=_-VERTICAL}', "");
Expect(1, 65096, '\P{^Decomposition_Type=_-VERTICAL}', "");
Expect(0, 65097, '\p{Decomposition_Type=_-VERTICAL}', "");
Expect(1, 65097, '\p{^Decomposition_Type=_-VERTICAL}', "");
Expect(1, 65097, '\P{Decomposition_Type=_-VERTICAL}', "");
Expect(0, 65097, '\P{^Decomposition_Type=_-VERTICAL}', "");
Error('\p{Dt=_/a/Vert}');
Error('\P{Dt=_/a/Vert}');
Expect(1, 65096, '\p{Dt=vert}', "");
Expect(0, 65096, '\p{^Dt=vert}', "");
Expect(0, 65096, '\P{Dt=vert}', "");
Expect(1, 65096, '\P{^Dt=vert}', "");
Expect(0, 65097, '\p{Dt=vert}', "");
Expect(1, 65097, '\p{^Dt=vert}', "");
Expect(1, 65097, '\P{Dt=vert}', "");
Expect(0, 65097, '\P{^Dt=vert}', "");
Expect(1, 65096, '\p{Dt=	Vert}', "");
Expect(0, 65096, '\p{^Dt=	Vert}', "");
Expect(0, 65096, '\P{Dt=	Vert}', "");
Expect(1, 65096, '\P{^Dt=	Vert}', "");
Expect(0, 65097, '\p{Dt=	Vert}', "");
Expect(1, 65097, '\p{^Dt=	Vert}', "");
Expect(1, 65097, '\P{Dt=	Vert}', "");
Expect(0, 65097, '\P{^Dt=	Vert}', "");
Error('\p{Is_Decomposition_Type=/a/ _VERTICAL}');
Error('\P{Is_Decomposition_Type=/a/ _VERTICAL}');
Expect(1, 65096, '\p{Is_Decomposition_Type=vertical}', "");
Expect(0, 65096, '\p{^Is_Decomposition_Type=vertical}', "");
Expect(0, 65096, '\P{Is_Decomposition_Type=vertical}', "");
Expect(1, 65096, '\P{^Is_Decomposition_Type=vertical}', "");
Expect(0, 65097, '\p{Is_Decomposition_Type=vertical}', "");
Expect(1, 65097, '\p{^Is_Decomposition_Type=vertical}', "");
Expect(1, 65097, '\P{Is_Decomposition_Type=vertical}', "");
Expect(0, 65097, '\P{^Is_Decomposition_Type=vertical}', "");
Expect(1, 65096, '\p{Is_Decomposition_Type=-VERTICAL}', "");
Expect(0, 65096, '\p{^Is_Decomposition_Type=-VERTICAL}', "");
Expect(0, 65096, '\P{Is_Decomposition_Type=-VERTICAL}', "");
Expect(1, 65096, '\P{^Is_Decomposition_Type=-VERTICAL}', "");
Expect(0, 65097, '\p{Is_Decomposition_Type=-VERTICAL}', "");
Expect(1, 65097, '\p{^Is_Decomposition_Type=-VERTICAL}', "");
Expect(1, 65097, '\P{Is_Decomposition_Type=-VERTICAL}', "");
Expect(0, 65097, '\P{^Is_Decomposition_Type=-VERTICAL}', "");
Error('\p{Is_Dt=vert/a/}');
Error('\P{Is_Dt=vert/a/}');
Expect(1, 65096, '\p{Is_Dt=vert}', "");
Expect(0, 65096, '\p{^Is_Dt=vert}', "");
Expect(0, 65096, '\P{Is_Dt=vert}', "");
Expect(1, 65096, '\P{^Is_Dt=vert}', "");
Expect(0, 65097, '\p{Is_Dt=vert}', "");
Expect(1, 65097, '\p{^Is_Dt=vert}', "");
Expect(1, 65097, '\P{Is_Dt=vert}', "");
Expect(0, 65097, '\P{^Is_Dt=vert}', "");
Expect(1, 65096, '\p{Is_Dt=-	VERT}', "");
Expect(0, 65096, '\p{^Is_Dt=-	VERT}', "");
Expect(0, 65096, '\P{Is_Dt=-	VERT}', "");
Expect(1, 65096, '\P{^Is_Dt=-	VERT}', "");
Expect(0, 65097, '\p{Is_Dt=-	VERT}', "");
Expect(1, 65097, '\p{^Is_Dt=-	VERT}', "");
Expect(1, 65097, '\P{Is_Dt=-	VERT}', "");
Expect(0, 65097, '\P{^Is_Dt=-	VERT}', "");
Error('\p{Decomposition_Type= /a/Wide}');
Error('\P{Decomposition_Type= /a/Wide}');
Expect(1, 65510, '\p{Decomposition_Type=wide}', "");
Expect(0, 65510, '\p{^Decomposition_Type=wide}', "");
Expect(0, 65510, '\P{Decomposition_Type=wide}', "");
Expect(1, 65510, '\P{^Decomposition_Type=wide}', "");
Expect(0, 65511, '\p{Decomposition_Type=wide}', "");
Expect(1, 65511, '\p{^Decomposition_Type=wide}', "");
Expect(1, 65511, '\P{Decomposition_Type=wide}', "");
Expect(0, 65511, '\P{^Decomposition_Type=wide}', "");
Expect(1, 65510, '\p{Decomposition_Type=	wide}', "");
Expect(0, 65510, '\p{^Decomposition_Type=	wide}', "");
Expect(0, 65510, '\P{Decomposition_Type=	wide}', "");
Expect(1, 65510, '\P{^Decomposition_Type=	wide}', "");
Expect(0, 65511, '\p{Decomposition_Type=	wide}', "");
Expect(1, 65511, '\p{^Decomposition_Type=	wide}', "");
Expect(1, 65511, '\P{Decomposition_Type=	wide}', "");
Expect(0, 65511, '\P{^Decomposition_Type=	wide}', "");
Error('\p{Dt=		wide:=}');
Error('\P{Dt=		wide:=}');
Expect(1, 65510, '\p{Dt=wide}', "");
Expect(0, 65510, '\p{^Dt=wide}', "");
Expect(0, 65510, '\P{Dt=wide}', "");
Expect(1, 65510, '\P{^Dt=wide}', "");
Expect(0, 65511, '\p{Dt=wide}', "");
Expect(1, 65511, '\p{^Dt=wide}', "");
Expect(1, 65511, '\P{Dt=wide}', "");
Expect(0, 65511, '\P{^Dt=wide}', "");
Expect(1, 65510, '\p{Dt=_-Wide}', "");
Expect(0, 65510, '\p{^Dt=_-Wide}', "");
Expect(0, 65510, '\P{Dt=_-Wide}', "");
Expect(1, 65510, '\P{^Dt=_-Wide}', "");
Expect(0, 65511, '\p{Dt=_-Wide}', "");
Expect(1, 65511, '\p{^Dt=_-Wide}', "");
Expect(1, 65511, '\P{Dt=_-Wide}', "");
Expect(0, 65511, '\P{^Dt=_-Wide}', "");
Error('\p{Is_Decomposition_Type=-/a/WIDE}');
Error('\P{Is_Decomposition_Type=-/a/WIDE}');
Expect(1, 65510, '\p{Is_Decomposition_Type:wide}', "");
Expect(0, 65510, '\p{^Is_Decomposition_Type:wide}', "");
Expect(0, 65510, '\P{Is_Decomposition_Type:wide}', "");
Expect(1, 65510, '\P{^Is_Decomposition_Type:wide}', "");
Expect(0, 65511, '\p{Is_Decomposition_Type:wide}', "");
Expect(1, 65511, '\p{^Is_Decomposition_Type:wide}', "");
Expect(1, 65511, '\P{Is_Decomposition_Type:wide}', "");
Expect(0, 65511, '\P{^Is_Decomposition_Type:wide}', "");
Expect(1, 65510, '\p{Is_Decomposition_Type=	-Wide}', "");
Expect(0, 65510, '\p{^Is_Decomposition_Type=	-Wide}', "");
Expect(0, 65510, '\P{Is_Decomposition_Type=	-Wide}', "");
Expect(1, 65510, '\P{^Is_Decomposition_Type=	-Wide}', "");
Expect(0, 65511, '\p{Is_Decomposition_Type=	-Wide}', "");
Expect(1, 65511, '\p{^Is_Decomposition_Type=	-Wide}', "");
Expect(1, 65511, '\P{Is_Decomposition_Type=	-Wide}', "");
Expect(0, 65511, '\P{^Is_Decomposition_Type=	-Wide}', "");
Error('\p{Is_Dt=/a/-_Wide}');
Error('\P{Is_Dt=/a/-_Wide}');
Expect(1, 65510, '\p{Is_Dt:   wide}', "");
Expect(0, 65510, '\p{^Is_Dt:   wide}', "");
Expect(0, 65510, '\P{Is_Dt:   wide}', "");
Expect(1, 65510, '\P{^Is_Dt:   wide}', "");
Expect(0, 65511, '\p{Is_Dt:   wide}', "");
Expect(1, 65511, '\p{^Is_Dt:   wide}', "");
Expect(1, 65511, '\P{Is_Dt:   wide}', "");
Expect(0, 65511, '\P{^Is_Dt:   wide}', "");
Expect(1, 65510, '\p{Is_Dt=		WIDE}', "");
Expect(0, 65510, '\p{^Is_Dt=		WIDE}', "");
Expect(0, 65510, '\P{Is_Dt=		WIDE}', "");
Expect(1, 65510, '\P{^Is_Dt=		WIDE}', "");
Expect(0, 65511, '\p{Is_Dt=		WIDE}', "");
Expect(1, 65511, '\p{^Is_Dt=		WIDE}', "");
Expect(1, 65511, '\P{Is_Dt=		WIDE}', "");
Expect(0, 65511, '\P{^Is_Dt=		WIDE}', "");
Error('\p{East_Asian_Width=/a/ambiguous}');
Error('\P{East_Asian_Width=/a/ambiguous}');
Expect(1, 1114109, '\p{East_Asian_Width=ambiguous}', "");
Expect(0, 1114109, '\p{^East_Asian_Width=ambiguous}', "");
Expect(0, 1114109, '\P{East_Asian_Width=ambiguous}', "");
Expect(1, 1114109, '\P{^East_Asian_Width=ambiguous}', "");
Expect(0, 918000, '\p{East_Asian_Width=ambiguous}', "");
Expect(1, 918000, '\p{^East_Asian_Width=ambiguous}', "");
Expect(1, 918000, '\P{East_Asian_Width=ambiguous}', "");
Expect(0, 918000, '\P{^East_Asian_Width=ambiguous}', "");
Expect(1, 1114109, '\p{East_Asian_Width=-AMBIGUOUS}', "");
Expect(0, 1114109, '\p{^East_Asian_Width=-AMBIGUOUS}', "");
Expect(0, 1114109, '\P{East_Asian_Width=-AMBIGUOUS}', "");
Expect(1, 1114109, '\P{^East_Asian_Width=-AMBIGUOUS}', "");
Expect(0, 918000, '\p{East_Asian_Width=-AMBIGUOUS}', "");
Expect(1, 918000, '\p{^East_Asian_Width=-AMBIGUOUS}', "");
Expect(1, 918000, '\P{East_Asian_Width=-AMBIGUOUS}', "");
Expect(0, 918000, '\P{^East_Asian_Width=-AMBIGUOUS}', "");
Error('\p{Ea=:=	a}');
Error('\P{Ea=:=	a}');
Expect(1, 1114109, '\p{Ea=a}', "");
Expect(0, 1114109, '\p{^Ea=a}', "");
Expect(0, 1114109, '\P{Ea=a}', "");
Expect(1, 1114109, '\P{^Ea=a}', "");
Expect(0, 918000, '\p{Ea=a}', "");
Expect(1, 918000, '\p{^Ea=a}', "");
Expect(1, 918000, '\P{Ea=a}', "");
Expect(0, 918000, '\P{^Ea=a}', "");
Expect(1, 1114109, '\p{Ea=		a}', "");
Expect(0, 1114109, '\p{^Ea=		a}', "");
Expect(0, 1114109, '\P{Ea=		a}', "");
Expect(1, 1114109, '\P{^Ea=		a}', "");
Expect(0, 918000, '\p{Ea=		a}', "");
Expect(1, 918000, '\p{^Ea=		a}', "");
Expect(1, 918000, '\P{Ea=		a}', "");
Expect(0, 918000, '\P{^Ea=		a}', "");
Error('\p{Is_East_Asian_Width=-:=Ambiguous}');
Error('\P{Is_East_Asian_Width=-:=Ambiguous}');
Expect(1, 1114109, '\p{Is_East_Asian_Width=ambiguous}', "");
Expect(0, 1114109, '\p{^Is_East_Asian_Width=ambiguous}', "");
Expect(0, 1114109, '\P{Is_East_Asian_Width=ambiguous}', "");
Expect(1, 1114109, '\P{^Is_East_Asian_Width=ambiguous}', "");
Expect(0, 918000, '\p{Is_East_Asian_Width=ambiguous}', "");
Expect(1, 918000, '\p{^Is_East_Asian_Width=ambiguous}', "");
Expect(1, 918000, '\P{Is_East_Asian_Width=ambiguous}', "");
Expect(0, 918000, '\P{^Is_East_Asian_Width=ambiguous}', "");
Expect(1, 1114109, '\p{Is_East_Asian_Width=_ Ambiguous}', "");
Expect(0, 1114109, '\p{^Is_East_Asian_Width=_ Ambiguous}', "");
Expect(0, 1114109, '\P{Is_East_Asian_Width=_ Ambiguous}', "");
Expect(1, 1114109, '\P{^Is_East_Asian_Width=_ Ambiguous}', "");
Expect(0, 918000, '\p{Is_East_Asian_Width=_ Ambiguous}', "");
Expect(1, 918000, '\p{^Is_East_Asian_Width=_ Ambiguous}', "");
Expect(1, 918000, '\P{Is_East_Asian_Width=_ Ambiguous}', "");
Expect(0, 918000, '\P{^Is_East_Asian_Width=_ Ambiguous}', "");
Error('\p{Is_Ea=-	A:=}');
Error('\P{Is_Ea=-	A:=}');
Expect(1, 1114109, '\p{Is_Ea:a}', "");
Expect(0, 1114109, '\p{^Is_Ea:a}', "");
Expect(0, 1114109, '\P{Is_Ea:a}', "");
Expect(1, 1114109, '\P{^Is_Ea:a}', "");
Expect(0, 918000, '\p{Is_Ea:a}', "");
Expect(1, 918000, '\p{^Is_Ea:a}', "");
Expect(1, 918000, '\P{Is_Ea:a}', "");
Expect(0, 918000, '\P{^Is_Ea:a}', "");
Expect(1, 1114109, '\p{Is_Ea=	 A}', "");
Expect(0, 1114109, '\p{^Is_Ea=	 A}', "");
Expect(0, 1114109, '\P{Is_Ea=	 A}', "");
Expect(1, 1114109, '\P{^Is_Ea=	 A}', "");
Expect(0, 918000, '\p{Is_Ea=	 A}', "");
Expect(1, 918000, '\p{^Is_Ea=	 A}', "");
Expect(1, 918000, '\P{Is_Ea=	 A}', "");
Expect(0, 918000, '\P{^Is_Ea=	 A}', "");
Error('\p{East_Asian_Width=_ Fullwidth/a/}');
Error('\P{East_Asian_Width=_ Fullwidth/a/}');
Expect(1, 65510, '\p{East_Asian_Width=fullwidth}', "");
Expect(0, 65510, '\p{^East_Asian_Width=fullwidth}', "");
Expect(0, 65510, '\P{East_Asian_Width=fullwidth}', "");
Expect(1, 65510, '\P{^East_Asian_Width=fullwidth}', "");
Expect(0, 65511, '\p{East_Asian_Width=fullwidth}', "");
Expect(1, 65511, '\p{^East_Asian_Width=fullwidth}', "");
Expect(1, 65511, '\P{East_Asian_Width=fullwidth}', "");
Expect(0, 65511, '\P{^East_Asian_Width=fullwidth}', "");
Expect(1, 65510, '\p{East_Asian_Width=  Fullwidth}', "");
Expect(0, 65510, '\p{^East_Asian_Width=  Fullwidth}', "");
Expect(0, 65510, '\P{East_Asian_Width=  Fullwidth}', "");
Expect(1, 65510, '\P{^East_Asian_Width=  Fullwidth}', "");
Expect(0, 65511, '\p{East_Asian_Width=  Fullwidth}', "");
Expect(1, 65511, '\p{^East_Asian_Width=  Fullwidth}', "");
Expect(1, 65511, '\P{East_Asian_Width=  Fullwidth}', "");
Expect(0, 65511, '\P{^East_Asian_Width=  Fullwidth}', "");
Error('\p{Ea=/a/__F}');
Error('\P{Ea=/a/__F}');
Expect(1, 65510, '\p{Ea=f}', "");
Expect(0, 65510, '\p{^Ea=f}', "");
Expect(0, 65510, '\P{Ea=f}', "");
Expect(1, 65510, '\P{^Ea=f}', "");
Expect(0, 65511, '\p{Ea=f}', "");
Expect(1, 65511, '\p{^Ea=f}', "");
Expect(1, 65511, '\P{Ea=f}', "");
Expect(0, 65511, '\P{^Ea=f}', "");
Expect(1, 65510, '\p{Ea:_ F}', "");
Expect(0, 65510, '\p{^Ea:_ F}', "");
Expect(0, 65510, '\P{Ea:_ F}', "");
Expect(1, 65510, '\P{^Ea:_ F}', "");
Expect(0, 65511, '\p{Ea:_ F}', "");
Expect(1, 65511, '\p{^Ea:_ F}', "");
Expect(1, 65511, '\P{Ea:_ F}', "");
Expect(0, 65511, '\P{^Ea:_ F}', "");
Error('\p{Is_East_Asian_Width= -Fullwidth/a/}');
Error('\P{Is_East_Asian_Width= -Fullwidth/a/}');
Expect(1, 65510, '\p{Is_East_Asian_Width=fullwidth}', "");
Expect(0, 65510, '\p{^Is_East_Asian_Width=fullwidth}', "");
Expect(0, 65510, '\P{Is_East_Asian_Width=fullwidth}', "");
Expect(1, 65510, '\P{^Is_East_Asian_Width=fullwidth}', "");
Expect(0, 65511, '\p{Is_East_Asian_Width=fullwidth}', "");
Expect(1, 65511, '\p{^Is_East_Asian_Width=fullwidth}', "");
Expect(1, 65511, '\P{Is_East_Asian_Width=fullwidth}', "");
Expect(0, 65511, '\P{^Is_East_Asian_Width=fullwidth}', "");
Expect(1, 65510, '\p{Is_East_Asian_Width=	FULLWIDTH}', "");
Expect(0, 65510, '\p{^Is_East_Asian_Width=	FULLWIDTH}', "");
Expect(0, 65510, '\P{Is_East_Asian_Width=	FULLWIDTH}', "");
Expect(1, 65510, '\P{^Is_East_Asian_Width=	FULLWIDTH}', "");
Expect(0, 65511, '\p{Is_East_Asian_Width=	FULLWIDTH}', "");
Expect(1, 65511, '\p{^Is_East_Asian_Width=	FULLWIDTH}', "");
Expect(1, 65511, '\P{Is_East_Asian_Width=	FULLWIDTH}', "");
Expect(0, 65511, '\P{^Is_East_Asian_Width=	FULLWIDTH}', "");
Error('\p{Is_Ea= 	F/a/}');
Error('\P{Is_Ea= 	F/a/}');
Expect(1, 65510, '\p{Is_Ea=f}', "");
Expect(0, 65510, '\p{^Is_Ea=f}', "");
Expect(0, 65510, '\P{Is_Ea=f}', "");
Expect(1, 65510, '\P{^Is_Ea=f}', "");
Expect(0, 65511, '\p{Is_Ea=f}', "");
Expect(1, 65511, '\p{^Is_Ea=f}', "");
Expect(1, 65511, '\P{Is_Ea=f}', "");
Expect(0, 65511, '\P{^Is_Ea=f}', "");
Expect(1, 65510, '\p{Is_Ea=- F}', "");
Expect(0, 65510, '\p{^Is_Ea=- F}', "");
Expect(0, 65510, '\P{Is_Ea=- F}', "");
Expect(1, 65510, '\P{^Is_Ea=- F}', "");
Expect(0, 65511, '\p{Is_Ea=- F}', "");
Expect(1, 65511, '\p{^Is_Ea=- F}', "");
Expect(1, 65511, '\P{Is_Ea=- F}', "");
Expect(0, 65511, '\P{^Is_Ea=- F}', "");
Error('\p{East_Asian_Width=--Halfwidth/a/}');
Error('\P{East_Asian_Width=--Halfwidth/a/}');
Expect(1, 65518, '\p{East_Asian_Width=halfwidth}', "");
Expect(0, 65518, '\p{^East_Asian_Width=halfwidth}', "");
Expect(0, 65518, '\P{East_Asian_Width=halfwidth}', "");
Expect(1, 65518, '\P{^East_Asian_Width=halfwidth}', "");
Expect(0, 65519, '\p{East_Asian_Width=halfwidth}', "");
Expect(1, 65519, '\p{^East_Asian_Width=halfwidth}', "");
Expect(1, 65519, '\P{East_Asian_Width=halfwidth}', "");
Expect(0, 65519, '\P{^East_Asian_Width=halfwidth}', "");
Expect(1, 65518, '\p{East_Asian_Width= Halfwidth}', "");
Expect(0, 65518, '\p{^East_Asian_Width= Halfwidth}', "");
Expect(0, 65518, '\P{East_Asian_Width= Halfwidth}', "");
Expect(1, 65518, '\P{^East_Asian_Width= Halfwidth}', "");
Expect(0, 65519, '\p{East_Asian_Width= Halfwidth}', "");
Expect(1, 65519, '\p{^East_Asian_Width= Halfwidth}', "");
Expect(1, 65519, '\P{East_Asian_Width= Halfwidth}', "");
Expect(0, 65519, '\P{^East_Asian_Width= Halfwidth}', "");
Error('\p{Ea=-_h/a/}');
Error('\P{Ea=-_h/a/}');
Expect(1, 65518, '\p{Ea=h}', "");
Expect(0, 65518, '\p{^Ea=h}', "");
Expect(0, 65518, '\P{Ea=h}', "");
Expect(1, 65518, '\P{^Ea=h}', "");
Expect(0, 65519, '\p{Ea=h}', "");
Expect(1, 65519, '\p{^Ea=h}', "");
Expect(1, 65519, '\P{Ea=h}', "");
Expect(0, 65519, '\P{^Ea=h}', "");
Expect(1, 65518, '\p{Ea=	h}', "");
Expect(0, 65518, '\p{^Ea=	h}', "");
Expect(0, 65518, '\P{Ea=	h}', "");
Expect(1, 65518, '\P{^Ea=	h}', "");
Expect(0, 65519, '\p{Ea=	h}', "");
Expect(1, 65519, '\p{^Ea=	h}', "");
Expect(1, 65519, '\P{Ea=	h}', "");
Expect(0, 65519, '\P{^Ea=	h}', "");
Error('\p{Is_East_Asian_Width=	/a/HALFWIDTH}');
Error('\P{Is_East_Asian_Width=	/a/HALFWIDTH}');
Expect(1, 65518, '\p{Is_East_Asian_Width=halfwidth}', "");
Expect(0, 65518, '\p{^Is_East_Asian_Width=halfwidth}', "");
Expect(0, 65518, '\P{Is_East_Asian_Width=halfwidth}', "");
Expect(1, 65518, '\P{^Is_East_Asian_Width=halfwidth}', "");
Expect(0, 65519, '\p{Is_East_Asian_Width=halfwidth}', "");
Expect(1, 65519, '\p{^Is_East_Asian_Width=halfwidth}', "");
Expect(1, 65519, '\P{Is_East_Asian_Width=halfwidth}', "");
Expect(0, 65519, '\P{^Is_East_Asian_Width=halfwidth}', "");
Expect(1, 65518, '\p{Is_East_Asian_Width=_Halfwidth}', "");
Expect(0, 65518, '\p{^Is_East_Asian_Width=_Halfwidth}', "");
Expect(0, 65518, '\P{Is_East_Asian_Width=_Halfwidth}', "");
Expect(1, 65518, '\P{^Is_East_Asian_Width=_Halfwidth}', "");
Expect(0, 65519, '\p{Is_East_Asian_Width=_Halfwidth}', "");
Expect(1, 65519, '\p{^Is_East_Asian_Width=_Halfwidth}', "");
Expect(1, 65519, '\P{Is_East_Asian_Width=_Halfwidth}', "");
Expect(0, 65519, '\P{^Is_East_Asian_Width=_Halfwidth}', "");
Error('\p{Is_Ea=-:=H}');
Error('\P{Is_Ea=-:=H}');
Expect(1, 65518, '\p{Is_Ea=h}', "");
Expect(0, 65518, '\p{^Is_Ea=h}', "");
Expect(0, 65518, '\P{Is_Ea=h}', "");
Expect(1, 65518, '\P{^Is_Ea=h}', "");
Expect(0, 65519, '\p{Is_Ea=h}', "");
Expect(1, 65519, '\p{^Is_Ea=h}', "");
Expect(1, 65519, '\P{Is_Ea=h}', "");
Expect(0, 65519, '\P{^Is_Ea=h}', "");
Expect(1, 65518, '\p{Is_Ea=  H}', "");
Expect(0, 65518, '\p{^Is_Ea=  H}', "");
Expect(0, 65518, '\P{Is_Ea=  H}', "");
Expect(1, 65518, '\P{^Is_Ea=  H}', "");
Expect(0, 65519, '\p{Is_Ea=  H}', "");
Expect(1, 65519, '\p{^Is_Ea=  H}', "");
Expect(1, 65519, '\P{Is_Ea=  H}', "");
Expect(0, 65519, '\P{^Is_Ea=  H}', "");
Error('\p{East_Asian_Width=/a/Neutral}');
Error('\P{East_Asian_Width=/a/Neutral}');
Expect(1, 918000, '\p{East_Asian_Width=neutral}', "");
Expect(0, 918000, '\p{^East_Asian_Width=neutral}', "");
Expect(0, 918000, '\P{East_Asian_Width=neutral}', "");
Expect(1, 918000, '\P{^East_Asian_Width=neutral}', "");
Expect(0, 1114109, '\p{East_Asian_Width=neutral}', "");
Expect(1, 1114109, '\p{^East_Asian_Width=neutral}', "");
Expect(1, 1114109, '\P{East_Asian_Width=neutral}', "");
Expect(0, 1114109, '\P{^East_Asian_Width=neutral}', "");
Expect(1, 918000, '\p{East_Asian_Width=		neutral}', "");
Expect(0, 918000, '\p{^East_Asian_Width=		neutral}', "");
Expect(0, 918000, '\P{East_Asian_Width=		neutral}', "");
Expect(1, 918000, '\P{^East_Asian_Width=		neutral}', "");
Expect(0, 1114109, '\p{East_Asian_Width=		neutral}', "");
Expect(1, 1114109, '\p{^East_Asian_Width=		neutral}', "");
Expect(1, 1114109, '\P{East_Asian_Width=		neutral}', "");
Expect(0, 1114109, '\P{^East_Asian_Width=		neutral}', "");
Error('\p{Ea:	/a/ N}');
Error('\P{Ea:	/a/ N}');
Expect(1, 918000, '\p{Ea=n}', "");
Expect(0, 918000, '\p{^Ea=n}', "");
Expect(0, 918000, '\P{Ea=n}', "");
Expect(1, 918000, '\P{^Ea=n}', "");
Expect(0, 1114109, '\p{Ea=n}', "");
Expect(1, 1114109, '\p{^Ea=n}', "");
Expect(1, 1114109, '\P{Ea=n}', "");
Expect(0, 1114109, '\P{^Ea=n}', "");
Expect(1, 918000, '\p{Ea=-n}', "");
Expect(0, 918000, '\p{^Ea=-n}', "");
Expect(0, 918000, '\P{Ea=-n}', "");
Expect(1, 918000, '\P{^Ea=-n}', "");
Expect(0, 1114109, '\p{Ea=-n}', "");
Expect(1, 1114109, '\p{^Ea=-n}', "");
Expect(1, 1114109, '\P{Ea=-n}', "");
Expect(0, 1114109, '\P{^Ea=-n}', "");
Error('\p{Is_East_Asian_Width=_neutral:=}');
Error('\P{Is_East_Asian_Width=_neutral:=}');
Expect(1, 918000, '\p{Is_East_Asian_Width=neutral}', "");
Expect(0, 918000, '\p{^Is_East_Asian_Width=neutral}', "");
Expect(0, 918000, '\P{Is_East_Asian_Width=neutral}', "");
Expect(1, 918000, '\P{^Is_East_Asian_Width=neutral}', "");
Expect(0, 1114109, '\p{Is_East_Asian_Width=neutral}', "");
Expect(1, 1114109, '\p{^Is_East_Asian_Width=neutral}', "");
Expect(1, 1114109, '\P{Is_East_Asian_Width=neutral}', "");
Expect(0, 1114109, '\P{^Is_East_Asian_Width=neutral}', "");
Expect(1, 918000, '\p{Is_East_Asian_Width=_Neutral}', "");
Expect(0, 918000, '\p{^Is_East_Asian_Width=_Neutral}', "");
Expect(0, 918000, '\P{Is_East_Asian_Width=_Neutral}', "");
Expect(1, 918000, '\P{^Is_East_Asian_Width=_Neutral}', "");
Expect(0, 1114109, '\p{Is_East_Asian_Width=_Neutral}', "");
Expect(1, 1114109, '\p{^Is_East_Asian_Width=_Neutral}', "");
Expect(1, 1114109, '\P{Is_East_Asian_Width=_Neutral}', "");
Expect(0, 1114109, '\P{^Is_East_Asian_Width=_Neutral}', "");
Error('\p{Is_Ea=/a/ N}');
Error('\P{Is_Ea=/a/ N}');
Expect(1, 918000, '\p{Is_Ea=n}', "");
Expect(0, 918000, '\p{^Is_Ea=n}', "");
Expect(0, 918000, '\P{Is_Ea=n}', "");
Expect(1, 918000, '\P{^Is_Ea=n}', "");
Expect(0, 1114109, '\p{Is_Ea=n}', "");
Expect(1, 1114109, '\p{^Is_Ea=n}', "");
Expect(1, 1114109, '\P{Is_Ea=n}', "");
Expect(0, 1114109, '\P{^Is_Ea=n}', "");
Expect(1, 918000, '\p{Is_Ea: N}', "");
Expect(0, 918000, '\p{^Is_Ea: N}', "");
Expect(0, 918000, '\P{Is_Ea: N}', "");
Expect(1, 918000, '\P{^Is_Ea: N}', "");
Expect(0, 1114109, '\p{Is_Ea: N}', "");
Expect(1, 1114109, '\p{^Is_Ea: N}', "");
Expect(1, 1114109, '\P{Is_Ea: N}', "");
Expect(0, 1114109, '\P{^Is_Ea: N}', "");
Error('\p{East_Asian_Width=:=_narrow}');
Error('\P{East_Asian_Width=:=_narrow}');
Expect(1, 10630, '\p{East_Asian_Width=narrow}', "");
Expect(0, 10630, '\p{^East_Asian_Width=narrow}', "");
Expect(0, 10630, '\P{East_Asian_Width=narrow}', "");
Expect(1, 10630, '\P{^East_Asian_Width=narrow}', "");
Expect(0, 10631, '\p{East_Asian_Width=narrow}', "");
Expect(1, 10631, '\p{^East_Asian_Width=narrow}', "");
Expect(1, 10631, '\P{East_Asian_Width=narrow}', "");
Expect(0, 10631, '\P{^East_Asian_Width=narrow}', "");
Expect(1, 10630, '\p{East_Asian_Width=- Narrow}', "");
Expect(0, 10630, '\p{^East_Asian_Width=- Narrow}', "");
Expect(0, 10630, '\P{East_Asian_Width=- Narrow}', "");
Expect(1, 10630, '\P{^East_Asian_Width=- Narrow}', "");
Expect(0, 10631, '\p{East_Asian_Width=- Narrow}', "");
Expect(1, 10631, '\p{^East_Asian_Width=- Narrow}', "");
Expect(1, 10631, '\P{East_Asian_Width=- Narrow}', "");
Expect(0, 10631, '\P{^East_Asian_Width=- Narrow}', "");
Error('\p{Ea=:=  NA}');
Error('\P{Ea=:=  NA}');
Expect(1, 10630, '\p{Ea=na}', "");
Expect(0, 10630, '\p{^Ea=na}', "");
Expect(0, 10630, '\P{Ea=na}', "");
Expect(1, 10630, '\P{^Ea=na}', "");
Expect(0, 10631, '\p{Ea=na}', "");
Expect(1, 10631, '\p{^Ea=na}', "");
Expect(1, 10631, '\P{Ea=na}', "");
Expect(0, 10631, '\P{^Ea=na}', "");
Expect(1, 10630, '\p{Ea=_ Na}', "");
Expect(0, 10630, '\p{^Ea=_ Na}', "");
Expect(0, 10630, '\P{Ea=_ Na}', "");
Expect(1, 10630, '\P{^Ea=_ Na}', "");
Expect(0, 10631, '\p{Ea=_ Na}', "");
Expect(1, 10631, '\p{^Ea=_ Na}', "");
Expect(1, 10631, '\P{Ea=_ Na}', "");
Expect(0, 10631, '\P{^Ea=_ Na}', "");
Error('\p{Is_East_Asian_Width=_:=Narrow}');
Error('\P{Is_East_Asian_Width=_:=Narrow}');
Expect(1, 10630, '\p{Is_East_Asian_Width=narrow}', "");
Expect(0, 10630, '\p{^Is_East_Asian_Width=narrow}', "");
Expect(0, 10630, '\P{Is_East_Asian_Width=narrow}', "");
Expect(1, 10630, '\P{^Is_East_Asian_Width=narrow}', "");
Expect(0, 10631, '\p{Is_East_Asian_Width=narrow}', "");
Expect(1, 10631, '\p{^Is_East_Asian_Width=narrow}', "");
Expect(1, 10631, '\P{Is_East_Asian_Width=narrow}', "");
Expect(0, 10631, '\P{^Is_East_Asian_Width=narrow}', "");
Expect(1, 10630, '\p{Is_East_Asian_Width= narrow}', "");
Expect(0, 10630, '\p{^Is_East_Asian_Width= narrow}', "");
Expect(0, 10630, '\P{Is_East_Asian_Width= narrow}', "");
Expect(1, 10630, '\P{^Is_East_Asian_Width= narrow}', "");
Expect(0, 10631, '\p{Is_East_Asian_Width= narrow}', "");
Expect(1, 10631, '\p{^Is_East_Asian_Width= narrow}', "");
Expect(1, 10631, '\P{Is_East_Asian_Width= narrow}', "");
Expect(0, 10631, '\P{^Is_East_Asian_Width= narrow}', "");
Error('\p{Is_Ea=/a/	na}');
Error('\P{Is_Ea=/a/	na}');
Expect(1, 10630, '\p{Is_Ea=na}', "");
Expect(0, 10630, '\p{^Is_Ea=na}', "");
Expect(0, 10630, '\P{Is_Ea=na}', "");
Expect(1, 10630, '\P{^Is_Ea=na}', "");
Expect(0, 10631, '\p{Is_Ea=na}', "");
Expect(1, 10631, '\p{^Is_Ea=na}', "");
Expect(1, 10631, '\P{Is_Ea=na}', "");
Expect(0, 10631, '\P{^Is_Ea=na}', "");
Expect(1, 10630, '\p{Is_Ea=_Na}', "");
Expect(0, 10630, '\p{^Is_Ea=_Na}', "");
Expect(0, 10630, '\P{Is_Ea=_Na}', "");
Expect(1, 10630, '\P{^Is_Ea=_Na}', "");
Expect(0, 10631, '\p{Is_Ea=_Na}', "");
Expect(1, 10631, '\p{^Is_Ea=_Na}', "");
Expect(1, 10631, '\P{Is_Ea=_Na}', "");
Expect(0, 10631, '\P{^Is_Ea=_Na}', "");
Error('\p{East_Asian_Width=- wide/a/}');
Error('\P{East_Asian_Width=- wide/a/}');
Expect(1, 262141, '\p{East_Asian_Width=wide}', "");
Expect(0, 262141, '\p{^East_Asian_Width=wide}', "");
Expect(0, 262141, '\P{East_Asian_Width=wide}', "");
Expect(1, 262141, '\P{^East_Asian_Width=wide}', "");
Expect(0, 262144, '\p{East_Asian_Width=wide}', "");
Expect(1, 262144, '\p{^East_Asian_Width=wide}', "");
Expect(1, 262144, '\P{East_Asian_Width=wide}', "");
Expect(0, 262144, '\P{^East_Asian_Width=wide}', "");
Expect(1, 262141, '\p{East_Asian_Width= wide}', "");
Expect(0, 262141, '\p{^East_Asian_Width= wide}', "");
Expect(0, 262141, '\P{East_Asian_Width= wide}', "");
Expect(1, 262141, '\P{^East_Asian_Width= wide}', "");
Expect(0, 262144, '\p{East_Asian_Width= wide}', "");
Expect(1, 262144, '\p{^East_Asian_Width= wide}', "");
Expect(1, 262144, '\P{East_Asian_Width= wide}', "");
Expect(0, 262144, '\P{^East_Asian_Width= wide}', "");
Error('\p{Ea=		w/a/}');
Error('\P{Ea=		w/a/}');
Expect(1, 262141, '\p{Ea=w}', "");
Expect(0, 262141, '\p{^Ea=w}', "");
Expect(0, 262141, '\P{Ea=w}', "");
Expect(1, 262141, '\P{^Ea=w}', "");
Expect(0, 262144, '\p{Ea=w}', "");
Expect(1, 262144, '\p{^Ea=w}', "");
Expect(1, 262144, '\P{Ea=w}', "");
Expect(0, 262144, '\P{^Ea=w}', "");
Expect(1, 262141, '\p{Ea=	W}', "");
Expect(0, 262141, '\p{^Ea=	W}', "");
Expect(0, 262141, '\P{Ea=	W}', "");
Expect(1, 262141, '\P{^Ea=	W}', "");
Expect(0, 262144, '\p{Ea=	W}', "");
Expect(1, 262144, '\p{^Ea=	W}', "");
Expect(1, 262144, '\P{Ea=	W}', "");
Expect(0, 262144, '\P{^Ea=	W}', "");
Error('\p{Is_East_Asian_Width=-Wide:=}');
Error('\P{Is_East_Asian_Width=-Wide:=}');
Expect(1, 262141, '\p{Is_East_Asian_Width=wide}', "");
Expect(0, 262141, '\p{^Is_East_Asian_Width=wide}', "");
Expect(0, 262141, '\P{Is_East_Asian_Width=wide}', "");
Expect(1, 262141, '\P{^Is_East_Asian_Width=wide}', "");
Expect(0, 262144, '\p{Is_East_Asian_Width=wide}', "");
Expect(1, 262144, '\p{^Is_East_Asian_Width=wide}', "");
Expect(1, 262144, '\P{Is_East_Asian_Width=wide}', "");
Expect(0, 262144, '\P{^Is_East_Asian_Width=wide}', "");
Expect(1, 262141, '\p{Is_East_Asian_Width=WIDE}', "");
Expect(0, 262141, '\p{^Is_East_Asian_Width=WIDE}', "");
Expect(0, 262141, '\P{Is_East_Asian_Width=WIDE}', "");
Expect(1, 262141, '\P{^Is_East_Asian_Width=WIDE}', "");
Expect(0, 262144, '\p{Is_East_Asian_Width=WIDE}', "");
Expect(1, 262144, '\p{^Is_East_Asian_Width=WIDE}', "");
Expect(1, 262144, '\P{Is_East_Asian_Width=WIDE}', "");
Expect(0, 262144, '\P{^Is_East_Asian_Width=WIDE}', "");
Error('\p{Is_Ea=	 w:=}');
Error('\P{Is_Ea=	 w:=}');
Expect(1, 262141, '\p{Is_Ea=w}', "");
Expect(0, 262141, '\p{^Is_Ea=w}', "");
Expect(0, 262141, '\P{Is_Ea=w}', "");
Expect(1, 262141, '\P{^Is_Ea=w}', "");
Expect(0, 262144, '\p{Is_Ea=w}', "");
Expect(1, 262144, '\p{^Is_Ea=w}', "");
Expect(1, 262144, '\P{Is_Ea=w}', "");
Expect(0, 262144, '\P{^Is_Ea=w}', "");
Expect(1, 262141, '\p{Is_Ea=  W}', "");
Expect(0, 262141, '\p{^Is_Ea=  W}', "");
Expect(0, 262141, '\P{Is_Ea=  W}', "");
Expect(1, 262141, '\P{^Is_Ea=  W}', "");
Expect(0, 262144, '\p{Is_Ea=  W}', "");
Expect(1, 262144, '\p{^Is_Ea=  W}', "");
Expect(1, 262144, '\P{Is_Ea=  W}', "");
Expect(0, 262144, '\P{^Is_Ea=  W}', "");
Error('\p{General_Category:   /a/_	Other}');
Error('\P{General_Category:   /a/_	Other}');
Expect(1, 918000, '\p{General_Category: other}', "");
Expect(0, 918000, '\p{^General_Category: other}', "");
Expect(0, 918000, '\P{General_Category: other}', "");
Expect(1, 918000, '\P{^General_Category: other}', "");
Expect(0, 917999, '\p{General_Category: other}', "");
Expect(1, 917999, '\p{^General_Category: other}', "");
Expect(1, 917999, '\P{General_Category: other}', "");
Expect(0, 917999, '\P{^General_Category: other}', "");
Expect(1, 918000, '\p{General_Category:   - Other}', "");
Expect(0, 918000, '\p{^General_Category:   - Other}', "");
Expect(0, 918000, '\P{General_Category:   - Other}', "");
Expect(1, 918000, '\P{^General_Category:   - Other}', "");
Expect(0, 917999, '\p{General_Category:   - Other}', "");
Expect(1, 917999, '\p{^General_Category:   - Other}', "");
Expect(1, 917999, '\P{General_Category:   - Other}', "");
Expect(0, 917999, '\P{^General_Category:   - Other}', "");
Error('\p{Gc=-:=C}');
Error('\P{Gc=-:=C}');
Expect(1, 918000, '\p{Gc=c}', "");
Expect(0, 918000, '\p{^Gc=c}', "");
Expect(0, 918000, '\P{Gc=c}', "");
Expect(1, 918000, '\P{^Gc=c}', "");
Expect(0, 917999, '\p{Gc=c}', "");
Expect(1, 917999, '\p{^Gc=c}', "");
Expect(1, 917999, '\P{Gc=c}', "");
Expect(0, 917999, '\P{^Gc=c}', "");
Expect(1, 918000, '\p{Gc= 	C}', "");
Expect(0, 918000, '\p{^Gc= 	C}', "");
Expect(0, 918000, '\P{Gc= 	C}', "");
Expect(1, 918000, '\P{^Gc= 	C}', "");
Expect(0, 917999, '\p{Gc= 	C}', "");
Expect(1, 917999, '\p{^Gc= 	C}', "");
Expect(1, 917999, '\P{Gc= 	C}', "");
Expect(0, 917999, '\P{^Gc= 	C}', "");
Error('\p{Category=/a/_-Other}');
Error('\P{Category=/a/_-Other}');
Expect(1, 918000, '\p{Category=other}', "");
Expect(0, 918000, '\p{^Category=other}', "");
Expect(0, 918000, '\P{Category=other}', "");
Expect(1, 918000, '\P{^Category=other}', "");
Expect(0, 917999, '\p{Category=other}', "");
Expect(1, 917999, '\p{^Category=other}', "");
Expect(1, 917999, '\P{Category=other}', "");
Expect(0, 917999, '\P{^Category=other}', "");
Expect(1, 918000, '\p{Category= _other}', "");
Expect(0, 918000, '\p{^Category= _other}', "");
Expect(0, 918000, '\P{Category= _other}', "");
Expect(1, 918000, '\P{^Category= _other}', "");
Expect(0, 917999, '\p{Category= _other}', "");
Expect(1, 917999, '\p{^Category= _other}', "");
Expect(1, 917999, '\P{Category= _other}', "");
Expect(0, 917999, '\P{^Category= _other}', "");
Error('\p{Is_General_Category=	C/a/}');
Error('\P{Is_General_Category=	C/a/}');
Expect(1, 918000, '\p{Is_General_Category=c}', "");
Expect(0, 918000, '\p{^Is_General_Category=c}', "");
Expect(0, 918000, '\P{Is_General_Category=c}', "");
Expect(1, 918000, '\P{^Is_General_Category=c}', "");
Expect(0, 917999, '\p{Is_General_Category=c}', "");
Expect(1, 917999, '\p{^Is_General_Category=c}', "");
Expect(1, 917999, '\P{Is_General_Category=c}', "");
Expect(0, 917999, '\P{^Is_General_Category=c}', "");
Expect(1, 918000, '\p{Is_General_Category= C}', "");
Expect(0, 918000, '\p{^Is_General_Category= C}', "");
Expect(0, 918000, '\P{Is_General_Category= C}', "");
Expect(1, 918000, '\P{^Is_General_Category= C}', "");
Expect(0, 917999, '\p{Is_General_Category= C}', "");
Expect(1, 917999, '\p{^Is_General_Category= C}', "");
Expect(1, 917999, '\P{Is_General_Category= C}', "");
Expect(0, 917999, '\P{^Is_General_Category= C}', "");
Error('\p{Is_Gc=		Other:=}');
Error('\P{Is_Gc=		Other:=}');
Expect(1, 918000, '\p{Is_Gc=other}', "");
Expect(0, 918000, '\p{^Is_Gc=other}', "");
Expect(0, 918000, '\P{Is_Gc=other}', "");
Expect(1, 918000, '\P{^Is_Gc=other}', "");
Expect(0, 917999, '\p{Is_Gc=other}', "");
Expect(1, 917999, '\p{^Is_Gc=other}', "");
Expect(1, 917999, '\P{Is_Gc=other}', "");
Expect(0, 917999, '\P{^Is_Gc=other}', "");
Expect(1, 918000, '\p{Is_Gc:	Other}', "");
Expect(0, 918000, '\p{^Is_Gc:	Other}', "");
Expect(0, 918000, '\P{Is_Gc:	Other}', "");
Expect(1, 918000, '\P{^Is_Gc:	Other}', "");
Expect(0, 917999, '\p{Is_Gc:	Other}', "");
Expect(1, 917999, '\p{^Is_Gc:	Other}', "");
Expect(1, 917999, '\P{Is_Gc:	Other}', "");
Expect(0, 917999, '\P{^Is_Gc:	Other}', "");
Error('\p{Is_Category=-_C/a/}');
Error('\P{Is_Category=-_C/a/}');
Expect(1, 918000, '\p{Is_Category:   c}', "");
Expect(0, 918000, '\p{^Is_Category:   c}', "");
Expect(0, 918000, '\P{Is_Category:   c}', "");
Expect(1, 918000, '\P{^Is_Category:   c}', "");
Expect(0, 917999, '\p{Is_Category:   c}', "");
Expect(1, 917999, '\p{^Is_Category:   c}', "");
Expect(1, 917999, '\P{Is_Category:   c}', "");
Expect(0, 917999, '\P{^Is_Category:   c}', "");
Expect(1, 918000, '\p{Is_Category=_c}', "");
Expect(0, 918000, '\p{^Is_Category=_c}', "");
Expect(0, 918000, '\P{Is_Category=_c}', "");
Expect(1, 918000, '\P{^Is_Category=_c}', "");
Expect(0, 917999, '\p{Is_Category=_c}', "");
Expect(1, 917999, '\p{^Is_Category=_c}', "");
Expect(1, 917999, '\P{Is_Category=_c}', "");
Expect(0, 917999, '\P{^Is_Category=_c}', "");
Error('\p{General_Category= /a/CONTROL}');
Error('\P{General_Category= /a/CONTROL}');
Expect(1, 159, '\p{General_Category:   control}', "");
Expect(0, 159, '\p{^General_Category:   control}', "");
Expect(0, 159, '\P{General_Category:   control}', "");
Expect(1, 159, '\P{^General_Category:   control}', "");
Expect(0, 160, '\p{General_Category:   control}', "");
Expect(1, 160, '\p{^General_Category:   control}', "");
Expect(1, 160, '\P{General_Category:   control}', "");
Expect(0, 160, '\P{^General_Category:   control}', "");
Expect(1, 159, '\p{General_Category=	Control}', "");
Expect(0, 159, '\p{^General_Category=	Control}', "");
Expect(0, 159, '\P{General_Category=	Control}', "");
Expect(1, 159, '\P{^General_Category=	Control}', "");
Expect(0, 160, '\p{General_Category=	Control}', "");
Expect(1, 160, '\p{^General_Category=	Control}', "");
Expect(1, 160, '\P{General_Category=	Control}', "");
Expect(0, 160, '\P{^General_Category=	Control}', "");
Error('\p{Gc:   :=cc}');
Error('\P{Gc:   :=cc}');
Expect(1, 159, '\p{Gc=cc}', "");
Expect(0, 159, '\p{^Gc=cc}', "");
Expect(0, 159, '\P{Gc=cc}', "");
Expect(1, 159, '\P{^Gc=cc}', "");
Expect(0, 160, '\p{Gc=cc}', "");
Expect(1, 160, '\p{^Gc=cc}', "");
Expect(1, 160, '\P{Gc=cc}', "");
Expect(0, 160, '\P{^Gc=cc}', "");
Expect(1, 159, '\p{Gc= cc}', "");
Expect(0, 159, '\p{^Gc= cc}', "");
Expect(0, 159, '\P{Gc= cc}', "");
Expect(1, 159, '\P{^Gc= cc}', "");
Expect(0, 160, '\p{Gc= cc}', "");
Expect(1, 160, '\p{^Gc= cc}', "");
Expect(1, 160, '\P{Gc= cc}', "");
Expect(0, 160, '\P{^Gc= cc}', "");
Error('\p{Category=:=cntrl}');
Error('\P{Category=:=cntrl}');
Expect(1, 159, '\p{Category=cntrl}', "");
Expect(0, 159, '\p{^Category=cntrl}', "");
Expect(0, 159, '\P{Category=cntrl}', "");
Expect(1, 159, '\P{^Category=cntrl}', "");
Expect(0, 160, '\p{Category=cntrl}', "");
Expect(1, 160, '\p{^Category=cntrl}', "");
Expect(1, 160, '\P{Category=cntrl}', "");
Expect(0, 160, '\P{^Category=cntrl}', "");
Expect(1, 159, '\p{Category=	 cntrl}', "");
Expect(0, 159, '\p{^Category=	 cntrl}', "");
Expect(0, 159, '\P{Category=	 cntrl}', "");
Expect(1, 159, '\P{^Category=	 cntrl}', "");
Expect(0, 160, '\p{Category=	 cntrl}', "");
Expect(1, 160, '\p{^Category=	 cntrl}', "");
Expect(1, 160, '\P{Category=	 cntrl}', "");
Expect(0, 160, '\P{^Category=	 cntrl}', "");
Error('\p{Is_General_Category=/a/-Control}');
Error('\P{Is_General_Category=/a/-Control}');
Expect(1, 159, '\p{Is_General_Category=control}', "");
Expect(0, 159, '\p{^Is_General_Category=control}', "");
Expect(0, 159, '\P{Is_General_Category=control}', "");
Expect(1, 159, '\P{^Is_General_Category=control}', "");
Expect(0, 160, '\p{Is_General_Category=control}', "");
Expect(1, 160, '\p{^Is_General_Category=control}', "");
Expect(1, 160, '\P{Is_General_Category=control}', "");
Expect(0, 160, '\P{^Is_General_Category=control}', "");
Expect(1, 159, '\p{Is_General_Category=		control}', "");
Expect(0, 159, '\p{^Is_General_Category=		control}', "");
Expect(0, 159, '\P{Is_General_Category=		control}', "");
Expect(1, 159, '\P{^Is_General_Category=		control}', "");
Expect(0, 160, '\p{Is_General_Category=		control}', "");
Expect(1, 160, '\p{^Is_General_Category=		control}', "");
Expect(1, 160, '\P{Is_General_Category=		control}', "");
Expect(0, 160, '\P{^Is_General_Category=		control}', "");
Error('\p{Is_Gc=-Cc/a/}');
Error('\P{Is_Gc=-Cc/a/}');
Expect(1, 159, '\p{Is_Gc=cc}', "");
Expect(0, 159, '\p{^Is_Gc=cc}', "");
Expect(0, 159, '\P{Is_Gc=cc}', "");
Expect(1, 159, '\P{^Is_Gc=cc}', "");
Expect(0, 160, '\p{Is_Gc=cc}', "");
Expect(1, 160, '\p{^Is_Gc=cc}', "");
Expect(1, 160, '\P{Is_Gc=cc}', "");
Expect(0, 160, '\P{^Is_Gc=cc}', "");
Expect(1, 159, '\p{Is_Gc= _Cc}', "");
Expect(0, 159, '\p{^Is_Gc= _Cc}', "");
Expect(0, 159, '\P{Is_Gc= _Cc}', "");
Expect(1, 159, '\P{^Is_Gc= _Cc}', "");
Expect(0, 160, '\p{Is_Gc= _Cc}', "");
Expect(1, 160, '\p{^Is_Gc= _Cc}', "");
Expect(1, 160, '\P{Is_Gc= _Cc}', "");
Expect(0, 160, '\P{^Is_Gc= _Cc}', "");
Error('\p{Is_Category=-:=cntrl}');
Error('\P{Is_Category=-:=cntrl}');
Expect(1, 159, '\p{Is_Category=cntrl}', "");
Expect(0, 159, '\p{^Is_Category=cntrl}', "");
Expect(0, 159, '\P{Is_Category=cntrl}', "");
Expect(1, 159, '\P{^Is_Category=cntrl}', "");
Expect(0, 160, '\p{Is_Category=cntrl}', "");
Expect(1, 160, '\p{^Is_Category=cntrl}', "");
Expect(1, 160, '\P{Is_Category=cntrl}', "");
Expect(0, 160, '\P{^Is_Category=cntrl}', "");
Expect(1, 159, '\p{Is_Category=	CNTRL}', "");
Expect(0, 159, '\p{^Is_Category=	CNTRL}', "");
Expect(0, 159, '\P{Is_Category=	CNTRL}', "");
Expect(1, 159, '\P{^Is_Category=	CNTRL}', "");
Expect(0, 160, '\p{Is_Category=	CNTRL}', "");
Expect(1, 160, '\p{^Is_Category=	CNTRL}', "");
Expect(1, 160, '\P{Is_Category=	CNTRL}', "");
Expect(0, 160, '\P{^Is_Category=	CNTRL}', "");
Error('\p{General_Category=	/a/Format}');
Error('\P{General_Category=	/a/Format}');
Expect(1, 917631, '\p{General_Category=format}', "");
Expect(0, 917631, '\p{^General_Category=format}', "");
Expect(0, 917631, '\P{General_Category=format}', "");
Expect(1, 917631, '\P{^General_Category=format}', "");
Expect(0, 917632, '\p{General_Category=format}', "");
Expect(1, 917632, '\p{^General_Category=format}', "");
Expect(1, 917632, '\P{General_Category=format}', "");
Expect(0, 917632, '\P{^General_Category=format}', "");
Expect(1, 917631, '\p{General_Category=	Format}', "");
Expect(0, 917631, '\p{^General_Category=	Format}', "");
Expect(0, 917631, '\P{General_Category=	Format}', "");
Expect(1, 917631, '\P{^General_Category=	Format}', "");
Expect(0, 917632, '\p{General_Category=	Format}', "");
Expect(1, 917632, '\p{^General_Category=	Format}', "");
Expect(1, 917632, '\P{General_Category=	Format}', "");
Expect(0, 917632, '\P{^General_Category=	Format}', "");
Error('\p{Gc=	:=CF}');
Error('\P{Gc=	:=CF}');
Expect(1, 917631, '\p{Gc=cf}', "");
Expect(0, 917631, '\p{^Gc=cf}', "");
Expect(0, 917631, '\P{Gc=cf}', "");
Expect(1, 917631, '\P{^Gc=cf}', "");
Expect(0, 917632, '\p{Gc=cf}', "");
Expect(1, 917632, '\p{^Gc=cf}', "");
Expect(1, 917632, '\P{Gc=cf}', "");
Expect(0, 917632, '\P{^Gc=cf}', "");
Expect(1, 917631, '\p{Gc=-	Cf}', "");
Expect(0, 917631, '\p{^Gc=-	Cf}', "");
Expect(0, 917631, '\P{Gc=-	Cf}', "");
Expect(1, 917631, '\P{^Gc=-	Cf}', "");
Expect(0, 917632, '\p{Gc=-	Cf}', "");
Expect(1, 917632, '\p{^Gc=-	Cf}', "");
Expect(1, 917632, '\P{Gc=-	Cf}', "");
Expect(0, 917632, '\P{^Gc=-	Cf}', "");
Error('\p{Category=_:=Format}');
Error('\P{Category=_:=Format}');
Expect(1, 917631, '\p{Category=format}', "");
Expect(0, 917631, '\p{^Category=format}', "");
Expect(0, 917631, '\P{Category=format}', "");
Expect(1, 917631, '\P{^Category=format}', "");
Expect(0, 917632, '\p{Category=format}', "");
Expect(1, 917632, '\p{^Category=format}', "");
Expect(1, 917632, '\P{Category=format}', "");
Expect(0, 917632, '\P{^Category=format}', "");
Expect(1, 917631, '\p{Category=_-Format}', "");
Expect(0, 917631, '\p{^Category=_-Format}', "");
Expect(0, 917631, '\P{Category=_-Format}', "");
Expect(1, 917631, '\P{^Category=_-Format}', "");
Expect(0, 917632, '\p{Category=_-Format}', "");
Expect(1, 917632, '\p{^Category=_-Format}', "");
Expect(1, 917632, '\P{Category=_-Format}', "");
Expect(0, 917632, '\P{^Category=_-Format}', "");
Error('\p{Is_General_Category=/a/	Cf}');
Error('\P{Is_General_Category=/a/	Cf}');
Expect(1, 917631, '\p{Is_General_Category=cf}', "");
Expect(0, 917631, '\p{^Is_General_Category=cf}', "");
Expect(0, 917631, '\P{Is_General_Category=cf}', "");
Expect(1, 917631, '\P{^Is_General_Category=cf}', "");
Expect(0, 917632, '\p{Is_General_Category=cf}', "");
Expect(1, 917632, '\p{^Is_General_Category=cf}', "");
Expect(1, 917632, '\P{Is_General_Category=cf}', "");
Expect(0, 917632, '\P{^Is_General_Category=cf}', "");
Expect(1, 917631, '\p{Is_General_Category: -cf}', "");
Expect(0, 917631, '\p{^Is_General_Category: -cf}', "");
Expect(0, 917631, '\P{Is_General_Category: -cf}', "");
Expect(1, 917631, '\P{^Is_General_Category: -cf}', "");
Expect(0, 917632, '\p{Is_General_Category: -cf}', "");
Expect(1, 917632, '\p{^Is_General_Category: -cf}', "");
Expect(1, 917632, '\P{Is_General_Category: -cf}', "");
Expect(0, 917632, '\P{^Is_General_Category: -cf}', "");
Error('\p{Is_Gc=format/a/}');
Error('\P{Is_Gc=format/a/}');
Expect(1, 917631, '\p{Is_Gc=format}', "");
Expect(0, 917631, '\p{^Is_Gc=format}', "");
Expect(0, 917631, '\P{Is_Gc=format}', "");
Expect(1, 917631, '\P{^Is_Gc=format}', "");
Expect(0, 917632, '\p{Is_Gc=format}', "");
Expect(1, 917632, '\p{^Is_Gc=format}', "");
Expect(1, 917632, '\P{Is_Gc=format}', "");
Expect(0, 917632, '\P{^Is_Gc=format}', "");
Expect(1, 917631, '\p{Is_Gc=  FORMAT}', "");
Expect(0, 917631, '\p{^Is_Gc=  FORMAT}', "");
Expect(0, 917631, '\P{Is_Gc=  FORMAT}', "");
Expect(1, 917631, '\P{^Is_Gc=  FORMAT}', "");
Expect(0, 917632, '\p{Is_Gc=  FORMAT}', "");
Expect(1, 917632, '\p{^Is_Gc=  FORMAT}', "");
Expect(1, 917632, '\P{Is_Gc=  FORMAT}', "");
Expect(0, 917632, '\P{^Is_Gc=  FORMAT}', "");
Error('\p{Is_Category=--cf/a/}');
Error('\P{Is_Category=--cf/a/}');
Expect(1, 917631, '\p{Is_Category=cf}', "");
Expect(0, 917631, '\p{^Is_Category=cf}', "");
Expect(0, 917631, '\P{Is_Category=cf}', "");
Expect(1, 917631, '\P{^Is_Category=cf}', "");
Expect(0, 917632, '\p{Is_Category=cf}', "");
Expect(1, 917632, '\p{^Is_Category=cf}', "");
Expect(1, 917632, '\P{Is_Category=cf}', "");
Expect(0, 917632, '\P{^Is_Category=cf}', "");
Expect(1, 917631, '\p{Is_Category= -Cf}', "");
Expect(0, 917631, '\p{^Is_Category= -Cf}', "");
Expect(0, 917631, '\P{Is_Category= -Cf}', "");
Expect(1, 917631, '\P{^Is_Category= -Cf}', "");
Expect(0, 917632, '\p{Is_Category= -Cf}', "");
Expect(1, 917632, '\p{^Is_Category= -Cf}', "");
Expect(1, 917632, '\P{Is_Category= -Cf}', "");
Expect(0, 917632, '\P{^Is_Category= -Cf}', "");
Error('\p{General_Category=-Unassigned/a/}');
Error('\P{General_Category=-Unassigned/a/}');
Expect(1, 918000, '\p{General_Category=unassigned}', "");
Expect(0, 918000, '\p{^General_Category=unassigned}', "");
Expect(0, 918000, '\P{General_Category=unassigned}', "");
Expect(1, 918000, '\P{^General_Category=unassigned}', "");
Expect(0, 1114109, '\p{General_Category=unassigned}', "");
Expect(1, 1114109, '\p{^General_Category=unassigned}', "");
Expect(1, 1114109, '\P{General_Category=unassigned}', "");
Expect(0, 1114109, '\P{^General_Category=unassigned}', "");
Expect(1, 918000, '\p{General_Category= _Unassigned}', "");
Expect(0, 918000, '\p{^General_Category= _Unassigned}', "");
Expect(0, 918000, '\P{General_Category= _Unassigned}', "");
Expect(1, 918000, '\P{^General_Category= _Unassigned}', "");
Expect(0, 1114109, '\p{General_Category= _Unassigned}', "");
Expect(1, 1114109, '\p{^General_Category= _Unassigned}', "");
Expect(1, 1114109, '\P{General_Category= _Unassigned}', "");
Expect(0, 1114109, '\P{^General_Category= _Unassigned}', "");
Error('\p{Gc=	 cn:=}');
Error('\P{Gc=	 cn:=}');
Expect(1, 918000, '\p{Gc=cn}', "");
Expect(0, 918000, '\p{^Gc=cn}', "");
Expect(0, 918000, '\P{Gc=cn}', "");
Expect(1, 918000, '\P{^Gc=cn}', "");
Expect(0, 1114109, '\p{Gc=cn}', "");
Expect(1, 1114109, '\p{^Gc=cn}', "");
Expect(1, 1114109, '\P{Gc=cn}', "");
Expect(0, 1114109, '\P{^Gc=cn}', "");
Expect(1, 918000, '\p{Gc=_-Cn}', "");
Expect(0, 918000, '\p{^Gc=_-Cn}', "");
Expect(0, 918000, '\P{Gc=_-Cn}', "");
Expect(1, 918000, '\P{^Gc=_-Cn}', "");
Expect(0, 1114109, '\p{Gc=_-Cn}', "");
Expect(1, 1114109, '\p{^Gc=_-Cn}', "");
Expect(1, 1114109, '\P{Gc=_-Cn}', "");
Expect(0, 1114109, '\P{^Gc=_-Cn}', "");
Error('\p{Category=:=  Unassigned}');
Error('\P{Category=:=  Unassigned}');
Expect(1, 918000, '\p{Category=unassigned}', "");
Expect(0, 918000, '\p{^Category=unassigned}', "");
Expect(0, 918000, '\P{Category=unassigned}', "");
Expect(1, 918000, '\P{^Category=unassigned}', "");
Expect(0, 1114109, '\p{Category=unassigned}', "");
Expect(1, 1114109, '\p{^Category=unassigned}', "");
Expect(1, 1114109, '\P{Category=unassigned}', "");
Expect(0, 1114109, '\P{^Category=unassigned}', "");
Expect(1, 918000, '\p{Category=	-Unassigned}', "");
Expect(0, 918000, '\p{^Category=	-Unassigned}', "");
Expect(0, 918000, '\P{Category=	-Unassigned}', "");
Expect(1, 918000, '\P{^Category=	-Unassigned}', "");
Expect(0, 1114109, '\p{Category=	-Unassigned}', "");
Expect(1, 1114109, '\p{^Category=	-Unassigned}', "");
Expect(1, 1114109, '\P{Category=	-Unassigned}', "");
Expect(0, 1114109, '\P{^Category=	-Unassigned}', "");
Error('\p{Is_General_Category:-:=CN}');
Error('\P{Is_General_Category:-:=CN}');
Expect(1, 918000, '\p{Is_General_Category=cn}', "");
Expect(0, 918000, '\p{^Is_General_Category=cn}', "");
Expect(0, 918000, '\P{Is_General_Category=cn}', "");
Expect(1, 918000, '\P{^Is_General_Category=cn}', "");
Expect(0, 1114109, '\p{Is_General_Category=cn}', "");
Expect(1, 1114109, '\p{^Is_General_Category=cn}', "");
Expect(1, 1114109, '\P{Is_General_Category=cn}', "");
Expect(0, 1114109, '\P{^Is_General_Category=cn}', "");
Expect(1, 918000, '\p{Is_General_Category=__CN}', "");
Expect(0, 918000, '\p{^Is_General_Category=__CN}', "");
Expect(0, 918000, '\P{Is_General_Category=__CN}', "");
Expect(1, 918000, '\P{^Is_General_Category=__CN}', "");
Expect(0, 1114109, '\p{Is_General_Category=__CN}', "");
Expect(1, 1114109, '\p{^Is_General_Category=__CN}', "");
Expect(1, 1114109, '\P{Is_General_Category=__CN}', "");
Expect(0, 1114109, '\P{^Is_General_Category=__CN}', "");
Error('\p{Is_Gc=-UNASSIGNED:=}');
Error('\P{Is_Gc=-UNASSIGNED:=}');
Expect(1, 918000, '\p{Is_Gc=unassigned}', "");
Expect(0, 918000, '\p{^Is_Gc=unassigned}', "");
Expect(0, 918000, '\P{Is_Gc=unassigned}', "");
Expect(1, 918000, '\P{^Is_Gc=unassigned}', "");
Expect(0, 1114109, '\p{Is_Gc=unassigned}', "");
Expect(1, 1114109, '\p{^Is_Gc=unassigned}', "");
Expect(1, 1114109, '\P{Is_Gc=unassigned}', "");
Expect(0, 1114109, '\P{^Is_Gc=unassigned}', "");
Expect(1, 918000, '\p{Is_Gc=_Unassigned}', "");
Expect(0, 918000, '\p{^Is_Gc=_Unassigned}', "");
Expect(0, 918000, '\P{Is_Gc=_Unassigned}', "");
Expect(1, 918000, '\P{^Is_Gc=_Unassigned}', "");
Expect(0, 1114109, '\p{Is_Gc=_Unassigned}', "");
Expect(1, 1114109, '\p{^Is_Gc=_Unassigned}', "");
Expect(1, 1114109, '\P{Is_Gc=_Unassigned}', "");
Expect(0, 1114109, '\P{^Is_Gc=_Unassigned}', "");
Error('\p{Is_Category=-cn:=}');
Error('\P{Is_Category=-cn:=}');
Expect(1, 918000, '\p{Is_Category=cn}', "");
Expect(0, 918000, '\p{^Is_Category=cn}', "");
Expect(0, 918000, '\P{Is_Category=cn}', "");
Expect(1, 918000, '\P{^Is_Category=cn}', "");
Expect(0, 1114109, '\p{Is_Category=cn}', "");
Expect(1, 1114109, '\p{^Is_Category=cn}', "");
Expect(1, 1114109, '\P{Is_Category=cn}', "");
Expect(0, 1114109, '\P{^Is_Category=cn}', "");
Expect(1, 918000, '\p{Is_Category=		Cn}', "");
Expect(0, 918000, '\p{^Is_Category=		Cn}', "");
Expect(0, 918000, '\P{Is_Category=		Cn}', "");
Expect(1, 918000, '\P{^Is_Category=		Cn}', "");
Expect(0, 1114109, '\p{Is_Category=		Cn}', "");
Expect(1, 1114109, '\p{^Is_Category=		Cn}', "");
Expect(1, 1114109, '\P{Is_Category=		Cn}', "");
Expect(0, 1114109, '\P{^Is_Category=		Cn}', "");
Error('\p{General_Category:	PRIVATE_Use:=}');
Error('\P{General_Category:	PRIVATE_Use:=}');
Expect(1, 1114109, '\p{General_Category=privateuse}', "");
Expect(0, 1114109, '\p{^General_Category=privateuse}', "");
Expect(0, 1114109, '\P{General_Category=privateuse}', "");
Expect(1, 1114109, '\P{^General_Category=privateuse}', "");
Expect(0, 63744, '\p{General_Category=privateuse}', "");
Expect(1, 63744, '\p{^General_Category=privateuse}', "");
Expect(1, 63744, '\P{General_Category=privateuse}', "");
Expect(0, 63744, '\P{^General_Category=privateuse}', "");
Expect(1, 1114109, '\p{General_Category=	_Private_use}', "");
Expect(0, 1114109, '\p{^General_Category=	_Private_use}', "");
Expect(0, 1114109, '\P{General_Category=	_Private_use}', "");
Expect(1, 1114109, '\P{^General_Category=	_Private_use}', "");
Expect(0, 63744, '\p{General_Category=	_Private_use}', "");
Expect(1, 63744, '\p{^General_Category=	_Private_use}', "");
Expect(1, 63744, '\P{General_Category=	_Private_use}', "");
Expect(0, 63744, '\P{^General_Category=	_Private_use}', "");
Error('\p{Gc=	 CO/a/}');
Error('\P{Gc=	 CO/a/}');
Expect(1, 1114109, '\p{Gc:	co}', "");
Expect(0, 1114109, '\p{^Gc:	co}', "");
Expect(0, 1114109, '\P{Gc:	co}', "");
Expect(1, 1114109, '\P{^Gc:	co}', "");
Expect(0, 63744, '\p{Gc:	co}', "");
Expect(1, 63744, '\p{^Gc:	co}', "");
Expect(1, 63744, '\P{Gc:	co}', "");
Expect(0, 63744, '\P{^Gc:	co}', "");
Expect(1, 1114109, '\p{Gc=__Co}', "");
Expect(0, 1114109, '\p{^Gc=__Co}', "");
Expect(0, 1114109, '\P{Gc=__Co}', "");
Expect(1, 1114109, '\P{^Gc=__Co}', "");
Expect(0, 63744, '\p{Gc=__Co}', "");
Expect(1, 63744, '\p{^Gc=__Co}', "");
Expect(1, 63744, '\P{Gc=__Co}', "");
Expect(0, 63744, '\P{^Gc=__Co}', "");
Error('\p{Category=:=__PRIVATE_Use}');
Error('\P{Category=:=__PRIVATE_Use}');
Expect(1, 1114109, '\p{Category=privateuse}', "");
Expect(0, 1114109, '\p{^Category=privateuse}', "");
Expect(0, 1114109, '\P{Category=privateuse}', "");
Expect(1, 1114109, '\P{^Category=privateuse}', "");
Expect(0, 63744, '\p{Category=privateuse}', "");
Expect(1, 63744, '\p{^Category=privateuse}', "");
Expect(1, 63744, '\P{Category=privateuse}', "");
Expect(0, 63744, '\P{^Category=privateuse}', "");
Expect(1, 1114109, '\p{Category=	_private_USE}', "");
Expect(0, 1114109, '\p{^Category=	_private_USE}', "");
Expect(0, 1114109, '\P{Category=	_private_USE}', "");
Expect(1, 1114109, '\P{^Category=	_private_USE}', "");
Expect(0, 63744, '\p{Category=	_private_USE}', "");
Expect(1, 63744, '\p{^Category=	_private_USE}', "");
Expect(1, 63744, '\P{Category=	_private_USE}', "");
Expect(0, 63744, '\P{^Category=	_private_USE}', "");
Error('\p{Is_General_Category=/a/ -co}');
Error('\P{Is_General_Category=/a/ -co}');
Expect(1, 1114109, '\p{Is_General_Category=co}', "");
Expect(0, 1114109, '\p{^Is_General_Category=co}', "");
Expect(0, 1114109, '\P{Is_General_Category=co}', "");
Expect(1, 1114109, '\P{^Is_General_Category=co}', "");
Expect(0, 63744, '\p{Is_General_Category=co}', "");
Expect(1, 63744, '\p{^Is_General_Category=co}', "");
Expect(1, 63744, '\P{Is_General_Category=co}', "");
Expect(0, 63744, '\P{^Is_General_Category=co}', "");
Expect(1, 1114109, '\p{Is_General_Category=		Co}', "");
Expect(0, 1114109, '\p{^Is_General_Category=		Co}', "");
Expect(0, 1114109, '\P{Is_General_Category=		Co}', "");
Expect(1, 1114109, '\P{^Is_General_Category=		Co}', "");
Expect(0, 63744, '\p{Is_General_Category=		Co}', "");
Expect(1, 63744, '\p{^Is_General_Category=		Co}', "");
Expect(1, 63744, '\P{Is_General_Category=		Co}', "");
Expect(0, 63744, '\P{^Is_General_Category=		Co}', "");
Error('\p{Is_Gc: /a/private_USE}');
Error('\P{Is_Gc: /a/private_USE}');
Expect(1, 1114109, '\p{Is_Gc=privateuse}', "");
Expect(0, 1114109, '\p{^Is_Gc=privateuse}', "");
Expect(0, 1114109, '\P{Is_Gc=privateuse}', "");
Expect(1, 1114109, '\P{^Is_Gc=privateuse}', "");
Expect(0, 63744, '\p{Is_Gc=privateuse}', "");
Expect(1, 63744, '\p{^Is_Gc=privateuse}', "");
Expect(1, 63744, '\P{Is_Gc=privateuse}', "");
Expect(0, 63744, '\P{^Is_Gc=privateuse}', "");
Expect(1, 1114109, '\p{Is_Gc=_private_Use}', "");
Expect(0, 1114109, '\p{^Is_Gc=_private_Use}', "");
Expect(0, 1114109, '\P{Is_Gc=_private_Use}', "");
Expect(1, 1114109, '\P{^Is_Gc=_private_Use}', "");
Expect(0, 63744, '\p{Is_Gc=_private_Use}', "");
Expect(1, 63744, '\p{^Is_Gc=_private_Use}', "");
Expect(1, 63744, '\P{Is_Gc=_private_Use}', "");
Expect(0, 63744, '\P{^Is_Gc=_private_Use}', "");
Error('\p{Is_Category: :=_ CO}');
Error('\P{Is_Category: :=_ CO}');
Expect(1, 1114109, '\p{Is_Category:co}', "");
Expect(0, 1114109, '\p{^Is_Category:co}', "");
Expect(0, 1114109, '\P{Is_Category:co}', "");
Expect(1, 1114109, '\P{^Is_Category:co}', "");
Expect(0, 63744, '\p{Is_Category:co}', "");
Expect(1, 63744, '\p{^Is_Category:co}', "");
Expect(1, 63744, '\P{Is_Category:co}', "");
Expect(0, 63744, '\P{^Is_Category:co}', "");
Expect(1, 1114109, '\p{Is_Category=--Co}', "");
Expect(0, 1114109, '\p{^Is_Category=--Co}', "");
Expect(0, 1114109, '\P{Is_Category=--Co}', "");
Expect(1, 1114109, '\P{^Is_Category=--Co}', "");
Expect(0, 63744, '\p{Is_Category=--Co}', "");
Expect(1, 63744, '\p{^Is_Category=--Co}', "");
Expect(1, 63744, '\P{Is_Category=--Co}', "");
Expect(0, 63744, '\P{^Is_Category=--Co}', "");
Error('\p{General_Category= surrogate/a/}');
Error('\P{General_Category= surrogate/a/}');
Expect(1, 57343, '\p{General_Category=surrogate}', "");
Expect(0, 57343, '\p{^General_Category=surrogate}', "");
Expect(0, 57343, '\P{General_Category=surrogate}', "");
Expect(1, 57343, '\P{^General_Category=surrogate}', "");
Expect(0, 57344, '\p{General_Category=surrogate}', "");
Expect(1, 57344, '\p{^General_Category=surrogate}', "");
Expect(1, 57344, '\P{General_Category=surrogate}', "");
Expect(0, 57344, '\P{^General_Category=surrogate}', "");
Expect(1, 57343, '\p{General_Category= Surrogate}', "");
Expect(0, 57343, '\p{^General_Category= Surrogate}', "");
Expect(0, 57343, '\P{General_Category= Surrogate}', "");
Expect(1, 57343, '\P{^General_Category= Surrogate}', "");
Expect(0, 57344, '\p{General_Category= Surrogate}', "");
Expect(1, 57344, '\p{^General_Category= Surrogate}', "");
Expect(1, 57344, '\P{General_Category= Surrogate}', "");
Expect(0, 57344, '\P{^General_Category= Surrogate}', "");
Error('\p{Gc=-/a/CS}');
Error('\P{Gc=-/a/CS}');
Expect(1, 57343, '\p{Gc: cs}', "");
Expect(0, 57343, '\p{^Gc: cs}', "");
Expect(0, 57343, '\P{Gc: cs}', "");
Expect(1, 57343, '\P{^Gc: cs}', "");
Expect(0, 57344, '\p{Gc: cs}', "");
Expect(1, 57344, '\p{^Gc: cs}', "");
Expect(1, 57344, '\P{Gc: cs}', "");
Expect(0, 57344, '\P{^Gc: cs}', "");
Expect(1, 57343, '\p{Gc=-cs}', "");
Expect(0, 57343, '\p{^Gc=-cs}', "");
Expect(0, 57343, '\P{Gc=-cs}', "");
Expect(1, 57343, '\P{^Gc=-cs}', "");
Expect(0, 57344, '\p{Gc=-cs}', "");
Expect(1, 57344, '\p{^Gc=-cs}', "");
Expect(1, 57344, '\P{Gc=-cs}', "");
Expect(0, 57344, '\P{^Gc=-cs}', "");
Error('\p{Category=Surrogate:=}');
Error('\P{Category=Surrogate:=}');
Expect(1, 57343, '\p{Category=surrogate}', "");
Expect(0, 57343, '\p{^Category=surrogate}', "");
Expect(0, 57343, '\P{Category=surrogate}', "");
Expect(1, 57343, '\P{^Category=surrogate}', "");
Expect(0, 57344, '\p{Category=surrogate}', "");
Expect(1, 57344, '\p{^Category=surrogate}', "");
Expect(1, 57344, '\P{Category=surrogate}', "");
Expect(0, 57344, '\P{^Category=surrogate}', "");
Expect(1, 57343, '\p{Category=-Surrogate}', "");
Expect(0, 57343, '\p{^Category=-Surrogate}', "");
Expect(0, 57343, '\P{Category=-Surrogate}', "");
Expect(1, 57343, '\P{^Category=-Surrogate}', "");
Expect(0, 57344, '\p{Category=-Surrogate}', "");
Expect(1, 57344, '\p{^Category=-Surrogate}', "");
Expect(1, 57344, '\P{Category=-Surrogate}', "");
Expect(0, 57344, '\P{^Category=-Surrogate}', "");
Error('\p{Is_General_Category=/a/_ CS}');
Error('\P{Is_General_Category=/a/_ CS}');
Expect(1, 57343, '\p{Is_General_Category: cs}', "");
Expect(0, 57343, '\p{^Is_General_Category: cs}', "");
Expect(0, 57343, '\P{Is_General_Category: cs}', "");
Expect(1, 57343, '\P{^Is_General_Category: cs}', "");
Expect(0, 57344, '\p{Is_General_Category: cs}', "");
Expect(1, 57344, '\p{^Is_General_Category: cs}', "");
Expect(1, 57344, '\P{Is_General_Category: cs}', "");
Expect(0, 57344, '\P{^Is_General_Category: cs}', "");
Expect(1, 57343, '\p{Is_General_Category= _cs}', "");
Expect(0, 57343, '\p{^Is_General_Category= _cs}', "");
Expect(0, 57343, '\P{Is_General_Category= _cs}', "");
Expect(1, 57343, '\P{^Is_General_Category= _cs}', "");
Expect(0, 57344, '\p{Is_General_Category= _cs}', "");
Expect(1, 57344, '\p{^Is_General_Category= _cs}', "");
Expect(1, 57344, '\P{Is_General_Category= _cs}', "");
Expect(0, 57344, '\P{^Is_General_Category= _cs}', "");
Error('\p{Is_Gc=:=_	surrogate}');
Error('\P{Is_Gc=:=_	surrogate}');
Expect(1, 57343, '\p{Is_Gc=surrogate}', "");
Expect(0, 57343, '\p{^Is_Gc=surrogate}', "");
Expect(0, 57343, '\P{Is_Gc=surrogate}', "");
Expect(1, 57343, '\P{^Is_Gc=surrogate}', "");
Expect(0, 57344, '\p{Is_Gc=surrogate}', "");
Expect(1, 57344, '\p{^Is_Gc=surrogate}', "");
Expect(1, 57344, '\P{Is_Gc=surrogate}', "");
Expect(0, 57344, '\P{^Is_Gc=surrogate}', "");
Expect(1, 57343, '\p{Is_Gc=_-SURROGATE}', "");
Expect(0, 57343, '\p{^Is_Gc=_-SURROGATE}', "");
Expect(0, 57343, '\P{Is_Gc=_-SURROGATE}', "");
Expect(1, 57343, '\P{^Is_Gc=_-SURROGATE}', "");
Expect(0, 57344, '\p{Is_Gc=_-SURROGATE}', "");
Expect(1, 57344, '\p{^Is_Gc=_-SURROGATE}', "");
Expect(1, 57344, '\P{Is_Gc=_-SURROGATE}', "");
Expect(0, 57344, '\P{^Is_Gc=_-SURROGATE}', "");
Error('\p{Is_Category=_-cs:=}');
Error('\P{Is_Category=_-cs:=}');
Expect(1, 57343, '\p{Is_Category=cs}', "");
Expect(0, 57343, '\p{^Is_Category=cs}', "");
Expect(0, 57343, '\P{Is_Category=cs}', "");
Expect(1, 57343, '\P{^Is_Category=cs}', "");
Expect(0, 57344, '\p{Is_Category=cs}', "");
Expect(1, 57344, '\p{^Is_Category=cs}', "");
Expect(1, 57344, '\P{Is_Category=cs}', "");
Expect(0, 57344, '\P{^Is_Category=cs}', "");
Expect(1, 57343, '\p{Is_Category=  Cs}', "");
Expect(0, 57343, '\p{^Is_Category=  Cs}', "");
Expect(0, 57343, '\P{Is_Category=  Cs}', "");
Expect(1, 57343, '\P{^Is_Category=  Cs}', "");
Expect(0, 57344, '\p{Is_Category=  Cs}', "");
Expect(1, 57344, '\p{^Is_Category=  Cs}', "");
Expect(1, 57344, '\P{Is_Category=  Cs}', "");
Expect(0, 57344, '\P{^Is_Category=  Cs}', "");
Error('\p{General_Category=_:=Letter}');
Error('\P{General_Category=_:=Letter}');
Expect(1, 195101, '\p{General_Category=letter}', "");
Expect(0, 195101, '\p{^General_Category=letter}', "");
Expect(0, 195101, '\P{General_Category=letter}', "");
Expect(1, 195101, '\P{^General_Category=letter}', "");
Expect(0, 195102, '\p{General_Category=letter}', "");
Expect(1, 195102, '\p{^General_Category=letter}', "");
Expect(1, 195102, '\P{General_Category=letter}', "");
Expect(0, 195102, '\P{^General_Category=letter}', "");
Expect(1, 195101, '\p{General_Category= 	letter}', "");
Expect(0, 195101, '\p{^General_Category= 	letter}', "");
Expect(0, 195101, '\P{General_Category= 	letter}', "");
Expect(1, 195101, '\P{^General_Category= 	letter}', "");
Expect(0, 195102, '\p{General_Category= 	letter}', "");
Expect(1, 195102, '\p{^General_Category= 	letter}', "");
Expect(1, 195102, '\P{General_Category= 	letter}', "");
Expect(0, 195102, '\P{^General_Category= 	letter}', "");
Error('\p{Gc= L:=}');
Error('\P{Gc= L:=}');
Expect(1, 195101, '\p{Gc=l}', "");
Expect(0, 195101, '\p{^Gc=l}', "");
Expect(0, 195101, '\P{Gc=l}', "");
Expect(1, 195101, '\P{^Gc=l}', "");
Expect(0, 195102, '\p{Gc=l}', "");
Expect(1, 195102, '\p{^Gc=l}', "");
Expect(1, 195102, '\P{Gc=l}', "");
Expect(0, 195102, '\P{^Gc=l}', "");
Expect(1, 195101, '\p{Gc=--l}', "");
Expect(0, 195101, '\p{^Gc=--l}', "");
Expect(0, 195101, '\P{Gc=--l}', "");
Expect(1, 195101, '\P{^Gc=--l}', "");
Expect(0, 195102, '\p{Gc=--l}', "");
Expect(1, 195102, '\p{^Gc=--l}', "");
Expect(1, 195102, '\P{Gc=--l}', "");
Expect(0, 195102, '\P{^Gc=--l}', "");
Error('\p{Category=:= 	Letter}');
Error('\P{Category=:= 	Letter}');
Expect(1, 195101, '\p{Category:   letter}', "");
Expect(0, 195101, '\p{^Category:   letter}', "");
Expect(0, 195101, '\P{Category:   letter}', "");
Expect(1, 195101, '\P{^Category:   letter}', "");
Expect(0, 195102, '\p{Category:   letter}', "");
Expect(1, 195102, '\p{^Category:   letter}', "");
Expect(1, 195102, '\P{Category:   letter}', "");
Expect(0, 195102, '\P{^Category:   letter}', "");
Expect(1, 195101, '\p{Category: _-letter}', "");
Expect(0, 195101, '\p{^Category: _-letter}', "");
Expect(0, 195101, '\P{Category: _-letter}', "");
Expect(1, 195101, '\P{^Category: _-letter}', "");
Expect(0, 195102, '\p{Category: _-letter}', "");
Expect(1, 195102, '\p{^Category: _-letter}', "");
Expect(1, 195102, '\P{Category: _-letter}', "");
Expect(0, 195102, '\P{^Category: _-letter}', "");
Error('\p{Is_General_Category=	_L:=}');
Error('\P{Is_General_Category=	_L:=}');
Expect(1, 195101, '\p{Is_General_Category=l}', "");
Expect(0, 195101, '\p{^Is_General_Category=l}', "");
Expect(0, 195101, '\P{Is_General_Category=l}', "");
Expect(1, 195101, '\P{^Is_General_Category=l}', "");
Expect(0, 195102, '\p{Is_General_Category=l}', "");
Expect(1, 195102, '\p{^Is_General_Category=l}', "");
Expect(1, 195102, '\P{Is_General_Category=l}', "");
Expect(0, 195102, '\P{^Is_General_Category=l}', "");
Expect(1, 195101, '\p{Is_General_Category=_ L}', "");
Expect(0, 195101, '\p{^Is_General_Category=_ L}', "");
Expect(0, 195101, '\P{Is_General_Category=_ L}', "");
Expect(1, 195101, '\P{^Is_General_Category=_ L}', "");
Expect(0, 195102, '\p{Is_General_Category=_ L}', "");
Expect(1, 195102, '\p{^Is_General_Category=_ L}', "");
Expect(1, 195102, '\P{Is_General_Category=_ L}', "");
Expect(0, 195102, '\P{^Is_General_Category=_ L}', "");
Error('\p{Is_Gc=/a/ _LETTER}');
Error('\P{Is_Gc=/a/ _LETTER}');
Expect(1, 195101, '\p{Is_Gc=letter}', "");
Expect(0, 195101, '\p{^Is_Gc=letter}', "");
Expect(0, 195101, '\P{Is_Gc=letter}', "");
Expect(1, 195101, '\P{^Is_Gc=letter}', "");
Expect(0, 195102, '\p{Is_Gc=letter}', "");
Expect(1, 195102, '\p{^Is_Gc=letter}', "");
Expect(1, 195102, '\P{Is_Gc=letter}', "");
Expect(0, 195102, '\P{^Is_Gc=letter}', "");
Expect(1, 195101, '\p{Is_Gc:_Letter}', "");
Expect(0, 195101, '\p{^Is_Gc:_Letter}', "");
Expect(0, 195101, '\P{Is_Gc:_Letter}', "");
Expect(1, 195101, '\P{^Is_Gc:_Letter}', "");
Expect(0, 195102, '\p{Is_Gc:_Letter}', "");
Expect(1, 195102, '\p{^Is_Gc:_Letter}', "");
Expect(1, 195102, '\P{Is_Gc:_Letter}', "");
Expect(0, 195102, '\P{^Is_Gc:_Letter}', "");
Error('\p{Is_Category=-:=L}');
Error('\P{Is_Category=-:=L}');
Expect(1, 195101, '\p{Is_Category=l}', "");
Expect(0, 195101, '\p{^Is_Category=l}', "");
Expect(0, 195101, '\P{Is_Category=l}', "");
Expect(1, 195101, '\P{^Is_Category=l}', "");
Expect(0, 195102, '\p{Is_Category=l}', "");
Expect(1, 195102, '\p{^Is_Category=l}', "");
Expect(1, 195102, '\P{Is_Category=l}', "");
Expect(0, 195102, '\P{^Is_Category=l}', "");
Expect(1, 195101, '\p{Is_Category= L}', "");
Expect(0, 195101, '\p{^Is_Category= L}', "");
Expect(0, 195101, '\P{Is_Category= L}', "");
Expect(1, 195101, '\P{^Is_Category= L}', "");
Expect(0, 195102, '\p{Is_Category= L}', "");
Expect(1, 195102, '\p{^Is_Category= L}', "");
Expect(1, 195102, '\P{Is_Category= L}', "");
Expect(0, 195102, '\P{^Is_Category= L}', "");
Error('\p{General_Category=_Cased_letter:=}');
Error('\P{General_Category=_Cased_letter:=}');
Expect(1, 120779, '\p{General_Category=casedletter}', "");
Expect(0, 120779, '\p{^General_Category=casedletter}', "");
Expect(0, 120779, '\P{General_Category=casedletter}', "");
Expect(1, 120779, '\P{^General_Category=casedletter}', "");
Expect(0, 120780, '\p{General_Category=casedletter}', "");
Expect(1, 120780, '\p{^General_Category=casedletter}', "");
Expect(1, 120780, '\P{General_Category=casedletter}', "");
Expect(0, 120780, '\P{^General_Category=casedletter}', "");
Expect(1, 120779, '\p{General_Category=_CASED_Letter}', "");
Expect(0, 120779, '\p{^General_Category=_CASED_Letter}', "");
Expect(0, 120779, '\P{General_Category=_CASED_Letter}', "");
Expect(1, 120779, '\P{^General_Category=_CASED_Letter}', "");
Expect(0, 120780, '\p{General_Category=_CASED_Letter}', "");
Expect(1, 120780, '\p{^General_Category=_CASED_Letter}', "");
Expect(1, 120780, '\P{General_Category=_CASED_Letter}', "");
Expect(0, 120780, '\P{^General_Category=_CASED_Letter}', "");
Error('\p{Gc= LC/a/}');
Error('\P{Gc= LC/a/}');
Expect(1, 120779, '\p{Gc=lc}', "");
Expect(0, 120779, '\p{^Gc=lc}', "");
Expect(0, 120779, '\P{Gc=lc}', "");
Expect(1, 120779, '\P{^Gc=lc}', "");
Expect(0, 120780, '\p{Gc=lc}', "");
Expect(1, 120780, '\p{^Gc=lc}', "");
Expect(1, 120780, '\P{Gc=lc}', "");
Expect(0, 120780, '\P{^Gc=lc}', "");
Expect(1, 120779, '\p{Gc=_lc}', "");
Expect(0, 120779, '\p{^Gc=_lc}', "");
Expect(0, 120779, '\P{Gc=_lc}', "");
Expect(1, 120779, '\P{^Gc=_lc}', "");
Expect(0, 120780, '\p{Gc=_lc}', "");
Expect(1, 120780, '\p{^Gc=_lc}', "");
Expect(1, 120780, '\P{Gc=_lc}', "");
Expect(0, 120780, '\P{^Gc=_lc}', "");
Error('\p{Category=/a/	 L_}');
Error('\P{Category=/a/	 L_}');
Expect(1, 120779, '\p{Category=l_}', "");
Expect(0, 120779, '\p{^Category=l_}', "");
Expect(0, 120779, '\P{Category=l_}', "");
Expect(1, 120779, '\P{^Category=l_}', "");
Expect(0, 120780, '\p{Category=l_}', "");
Expect(1, 120780, '\p{^Category=l_}', "");
Expect(1, 120780, '\P{Category=l_}', "");
Expect(0, 120780, '\P{^Category=l_}', "");
Expect(1, 120779, '\p{Category=--l_}', "");
Expect(0, 120779, '\p{^Category=--l_}', "");
Expect(0, 120779, '\P{Category=--l_}', "");
Expect(1, 120779, '\P{^Category=--l_}', "");
Expect(0, 120780, '\p{Category=--l_}', "");
Expect(1, 120780, '\p{^Category=--l_}', "");
Expect(1, 120780, '\P{Category=--l_}', "");
Expect(0, 120780, '\P{^Category=--l_}', "");
Error('\p{Is_General_Category=_ l&:=}');
Error('\P{Is_General_Category=_ l&:=}');
Expect(1, 120779, '\p{Is_General_Category=l&}', "");
Expect(0, 120779, '\p{^Is_General_Category=l&}', "");
Expect(0, 120779, '\P{Is_General_Category=l&}', "");
Expect(1, 120779, '\P{^Is_General_Category=l&}', "");
Expect(0, 120780, '\p{Is_General_Category=l&}', "");
Expect(1, 120780, '\p{^Is_General_Category=l&}', "");
Expect(1, 120780, '\P{Is_General_Category=l&}', "");
Expect(0, 120780, '\P{^Is_General_Category=l&}', "");
Expect(1, 120779, '\p{Is_General_Category=_L&}', "");
Expect(0, 120779, '\p{^Is_General_Category=_L&}', "");
Expect(0, 120779, '\P{Is_General_Category=_L&}', "");
Expect(1, 120779, '\P{^Is_General_Category=_L&}', "");
Expect(0, 120780, '\p{Is_General_Category=_L&}', "");
Expect(1, 120780, '\p{^Is_General_Category=_L&}', "");
Expect(1, 120780, '\P{Is_General_Category=_L&}', "");
Expect(0, 120780, '\P{^Is_General_Category=_L&}', "");
Error('\p{Is_Gc=:=_	Cased_Letter}');
Error('\P{Is_Gc=:=_	Cased_Letter}');
Expect(1, 120779, '\p{Is_Gc=casedletter}', "");
Expect(0, 120779, '\p{^Is_Gc=casedletter}', "");
Expect(0, 120779, '\P{Is_Gc=casedletter}', "");
Expect(1, 120779, '\P{^Is_Gc=casedletter}', "");
Expect(0, 120780, '\p{Is_Gc=casedletter}', "");
Expect(1, 120780, '\p{^Is_Gc=casedletter}', "");
Expect(1, 120780, '\P{Is_Gc=casedletter}', "");
Expect(0, 120780, '\P{^Is_Gc=casedletter}', "");
Expect(1, 120779, '\p{Is_Gc= Cased_letter}', "");
Expect(0, 120779, '\p{^Is_Gc= Cased_letter}', "");
Expect(0, 120779, '\P{Is_Gc= Cased_letter}', "");
Expect(1, 120779, '\P{^Is_Gc= Cased_letter}', "");
Expect(0, 120780, '\p{Is_Gc= Cased_letter}', "");
Expect(1, 120780, '\p{^Is_Gc= Cased_letter}', "");
Expect(1, 120780, '\P{Is_Gc= Cased_letter}', "");
Expect(0, 120780, '\P{^Is_Gc= Cased_letter}', "");
Error('\p{Is_Category=_/a/LC}');
Error('\P{Is_Category=_/a/LC}');
Expect(1, 120779, '\p{Is_Category=lc}', "");
Expect(0, 120779, '\p{^Is_Category=lc}', "");
Expect(0, 120779, '\P{Is_Category=lc}', "");
Expect(1, 120779, '\P{^Is_Category=lc}', "");
Expect(0, 120780, '\p{Is_Category=lc}', "");
Expect(1, 120780, '\p{^Is_Category=lc}', "");
Expect(1, 120780, '\P{Is_Category=lc}', "");
Expect(0, 120780, '\P{^Is_Category=lc}', "");
Expect(1, 120779, '\p{Is_Category=-_LC}', "");
Expect(0, 120779, '\p{^Is_Category=-_LC}', "");
Expect(0, 120779, '\P{Is_Category=-_LC}', "");
Expect(1, 120779, '\P{^Is_Category=-_LC}', "");
Expect(0, 120780, '\p{Is_Category=-_LC}', "");
Expect(1, 120780, '\p{^Is_Category=-_LC}', "");
Expect(1, 120780, '\P{Is_Category=-_LC}', "");
Expect(0, 120780, '\P{^Is_Category=-_LC}', "");
Error('\p{General_Category=:= 	Lowercase_letter}');
Error('\P{General_Category=:= 	Lowercase_letter}');
Expect(1, 120779, '\p{General_Category=lowercaseletter}', "");
Expect(0, 120779, '\p{^General_Category=lowercaseletter}', "");
Expect(0, 120779, '\P{General_Category=lowercaseletter}', "");
Expect(1, 120779, '\P{^General_Category=lowercaseletter}', "");
Expect(0, 120780, '\p{General_Category=lowercaseletter}', "");
Expect(1, 120780, '\p{^General_Category=lowercaseletter}', "");
Expect(1, 120780, '\P{General_Category=lowercaseletter}', "");
Expect(0, 120780, '\P{^General_Category=lowercaseletter}', "");
Expect(1, 120779, '\p{General_Category=	LOWERCASE_Letter}', "");
Expect(0, 120779, '\p{^General_Category=	LOWERCASE_Letter}', "");
Expect(0, 120779, '\P{General_Category=	LOWERCASE_Letter}', "");
Expect(1, 120779, '\P{^General_Category=	LOWERCASE_Letter}', "");
Expect(0, 120780, '\p{General_Category=	LOWERCASE_Letter}', "");
Expect(1, 120780, '\p{^General_Category=	LOWERCASE_Letter}', "");
Expect(1, 120780, '\P{General_Category=	LOWERCASE_Letter}', "");
Expect(0, 120780, '\P{^General_Category=	LOWERCASE_Letter}', "");
Error('\p{Gc=-	ll/a/}');
Error('\P{Gc=-	ll/a/}');
Expect(1, 120779, '\p{Gc=ll}', "");
Expect(0, 120779, '\p{^Gc=ll}', "");
Expect(0, 120779, '\P{Gc=ll}', "");
Expect(1, 120779, '\P{^Gc=ll}', "");
Expect(0, 120780, '\p{Gc=ll}', "");
Expect(1, 120780, '\p{^Gc=ll}', "");
Expect(1, 120780, '\P{Gc=ll}', "");
Expect(0, 120780, '\P{^Gc=ll}', "");
Expect(1, 120779, '\p{Gc= Ll}', "");
Expect(0, 120779, '\p{^Gc= Ll}', "");
Expect(0, 120779, '\P{Gc= Ll}', "");
Expect(1, 120779, '\P{^Gc= Ll}', "");
Expect(0, 120780, '\p{Gc= Ll}', "");
Expect(1, 120780, '\p{^Gc= Ll}', "");
Expect(1, 120780, '\P{Gc= Ll}', "");
Expect(0, 120780, '\P{^Gc= Ll}', "");
Error('\p{Category=/a/	-Lowercase_letter}');
Error('\P{Category=/a/	-Lowercase_letter}');
Expect(1, 120779, '\p{Category=lowercaseletter}', "");
Expect(0, 120779, '\p{^Category=lowercaseletter}', "");
Expect(0, 120779, '\P{Category=lowercaseletter}', "");
Expect(1, 120779, '\P{^Category=lowercaseletter}', "");
Expect(0, 120780, '\p{Category=lowercaseletter}', "");
Expect(1, 120780, '\p{^Category=lowercaseletter}', "");
Expect(1, 120780, '\P{Category=lowercaseletter}', "");
Expect(0, 120780, '\P{^Category=lowercaseletter}', "");
Expect(1, 120779, '\p{Category= Lowercase_LETTER}', "");
Expect(0, 120779, '\p{^Category= Lowercase_LETTER}', "");
Expect(0, 120779, '\P{Category= Lowercase_LETTER}', "");
Expect(1, 120779, '\P{^Category= Lowercase_LETTER}', "");
Expect(0, 120780, '\p{Category= Lowercase_LETTER}', "");
Expect(1, 120780, '\p{^Category= Lowercase_LETTER}', "");
Expect(1, 120780, '\P{Category= Lowercase_LETTER}', "");
Expect(0, 120780, '\P{^Category= Lowercase_LETTER}', "");
Error('\p{Is_General_Category=/a/--Ll}');
Error('\P{Is_General_Category=/a/--Ll}');
Expect(1, 120779, '\p{Is_General_Category=ll}', "");
Expect(0, 120779, '\p{^Is_General_Category=ll}', "");
Expect(0, 120779, '\P{Is_General_Category=ll}', "");
Expect(1, 120779, '\P{^Is_General_Category=ll}', "");
Expect(0, 120780, '\p{Is_General_Category=ll}', "");
Expect(1, 120780, '\p{^Is_General_Category=ll}', "");
Expect(1, 120780, '\P{Is_General_Category=ll}', "");
Expect(0, 120780, '\P{^Is_General_Category=ll}', "");
Expect(1, 120779, '\p{Is_General_Category= _ll}', "");
Expect(0, 120779, '\p{^Is_General_Category= _ll}', "");
Expect(0, 120779, '\P{Is_General_Category= _ll}', "");
Expect(1, 120779, '\P{^Is_General_Category= _ll}', "");
Expect(0, 120780, '\p{Is_General_Category= _ll}', "");
Expect(1, 120780, '\p{^Is_General_Category= _ll}', "");
Expect(1, 120780, '\P{Is_General_Category= _ll}', "");
Expect(0, 120780, '\P{^Is_General_Category= _ll}', "");
Error('\p{Is_Gc: -:=lowercase_letter}');
Error('\P{Is_Gc: -:=lowercase_letter}');
Expect(1, 120779, '\p{Is_Gc=lowercaseletter}', "");
Expect(0, 120779, '\p{^Is_Gc=lowercaseletter}', "");
Expect(0, 120779, '\P{Is_Gc=lowercaseletter}', "");
Expect(1, 120779, '\P{^Is_Gc=lowercaseletter}', "");
Expect(0, 120780, '\p{Is_Gc=lowercaseletter}', "");
Expect(1, 120780, '\p{^Is_Gc=lowercaseletter}', "");
Expect(1, 120780, '\P{Is_Gc=lowercaseletter}', "");
Expect(0, 120780, '\P{^Is_Gc=lowercaseletter}', "");
Expect(1, 120779, '\p{Is_Gc=	 LOWERCASE_Letter}', "");
Expect(0, 120779, '\p{^Is_Gc=	 LOWERCASE_Letter}', "");
Expect(0, 120779, '\P{Is_Gc=	 LOWERCASE_Letter}', "");
Expect(1, 120779, '\P{^Is_Gc=	 LOWERCASE_Letter}', "");
Expect(0, 120780, '\p{Is_Gc=	 LOWERCASE_Letter}', "");
Expect(1, 120780, '\p{^Is_Gc=	 LOWERCASE_Letter}', "");
Expect(1, 120780, '\P{Is_Gc=	 LOWERCASE_Letter}', "");
Expect(0, 120780, '\P{^Is_Gc=	 LOWERCASE_Letter}', "");
Error('\p{Is_Category:   _/a/Ll}');
Error('\P{Is_Category:   _/a/Ll}');
Expect(1, 120779, '\p{Is_Category=ll}', "");
Expect(0, 120779, '\p{^Is_Category=ll}', "");
Expect(0, 120779, '\P{Is_Category=ll}', "");
Expect(1, 120779, '\P{^Is_Category=ll}', "");
Expect(0, 120780, '\p{Is_Category=ll}', "");
Expect(1, 120780, '\p{^Is_Category=ll}', "");
Expect(1, 120780, '\P{Is_Category=ll}', "");
Expect(0, 120780, '\P{^Is_Category=ll}', "");
Expect(1, 120779, '\p{Is_Category=_	ll}', "");
Expect(0, 120779, '\p{^Is_Category=_	ll}', "");
Expect(0, 120779, '\P{Is_Category=_	ll}', "");
Expect(1, 120779, '\P{^Is_Category=_	ll}', "");
Expect(0, 120780, '\p{Is_Category=_	ll}', "");
Expect(1, 120780, '\p{^Is_Category=_	ll}', "");
Expect(1, 120780, '\P{Is_Category=_	ll}', "");
Expect(0, 120780, '\P{^Is_Category=_	ll}', "");
Error('\p{General_Category=		modifier_Letter/a/}');
Error('\P{General_Category=		modifier_Letter/a/}');
Expect(1, 94111, '\p{General_Category=modifierletter}', "");
Expect(0, 94111, '\p{^General_Category=modifierletter}', "");
Expect(0, 94111, '\P{General_Category=modifierletter}', "");
Expect(1, 94111, '\P{^General_Category=modifierletter}', "");
Expect(0, 94112, '\p{General_Category=modifierletter}', "");
Expect(1, 94112, '\p{^General_Category=modifierletter}', "");
Expect(1, 94112, '\P{General_Category=modifierletter}', "");
Expect(0, 94112, '\P{^General_Category=modifierletter}', "");
Expect(1, 94111, '\p{General_Category=_MODIFIER_letter}', "");
Expect(0, 94111, '\p{^General_Category=_MODIFIER_letter}', "");
Expect(0, 94111, '\P{General_Category=_MODIFIER_letter}', "");
Expect(1, 94111, '\P{^General_Category=_MODIFIER_letter}', "");
Expect(0, 94112, '\p{General_Category=_MODIFIER_letter}', "");
Expect(1, 94112, '\p{^General_Category=_MODIFIER_letter}', "");
Expect(1, 94112, '\P{General_Category=_MODIFIER_letter}', "");
Expect(0, 94112, '\P{^General_Category=_MODIFIER_letter}', "");
Error('\p{Gc=/a/Lm}');
Error('\P{Gc=/a/Lm}');
Expect(1, 94111, '\p{Gc=lm}', "");
Expect(0, 94111, '\p{^Gc=lm}', "");
Expect(0, 94111, '\P{Gc=lm}', "");
Expect(1, 94111, '\P{^Gc=lm}', "");
Expect(0, 94112, '\p{Gc=lm}', "");
Expect(1, 94112, '\p{^Gc=lm}', "");
Expect(1, 94112, '\P{Gc=lm}', "");
Expect(0, 94112, '\P{^Gc=lm}', "");
Expect(1, 94111, '\p{Gc=_LM}', "");
Expect(0, 94111, '\p{^Gc=_LM}', "");
Expect(0, 94111, '\P{Gc=_LM}', "");
Expect(1, 94111, '\P{^Gc=_LM}', "");
Expect(0, 94112, '\p{Gc=_LM}', "");
Expect(1, 94112, '\p{^Gc=_LM}', "");
Expect(1, 94112, '\P{Gc=_LM}', "");
Expect(0, 94112, '\P{^Gc=_LM}', "");
Error('\p{Category= /a/modifier_letter}');
Error('\P{Category= /a/modifier_letter}');
Expect(1, 94111, '\p{Category=modifierletter}', "");
Expect(0, 94111, '\p{^Category=modifierletter}', "");
Expect(0, 94111, '\P{Category=modifierletter}', "");
Expect(1, 94111, '\P{^Category=modifierletter}', "");
Expect(0, 94112, '\p{Category=modifierletter}', "");
Expect(1, 94112, '\p{^Category=modifierletter}', "");
Expect(1, 94112, '\P{Category=modifierletter}', "");
Expect(0, 94112, '\P{^Category=modifierletter}', "");
Expect(1, 94111, '\p{Category:   -Modifier_Letter}', "");
Expect(0, 94111, '\p{^Category:   -Modifier_Letter}', "");
Expect(0, 94111, '\P{Category:   -Modifier_Letter}', "");
Expect(1, 94111, '\P{^Category:   -Modifier_Letter}', "");
Expect(0, 94112, '\p{Category:   -Modifier_Letter}', "");
Expect(1, 94112, '\p{^Category:   -Modifier_Letter}', "");
Expect(1, 94112, '\P{Category:   -Modifier_Letter}', "");
Expect(0, 94112, '\P{^Category:   -Modifier_Letter}', "");
Error('\p{Is_General_Category=/a/		LM}');
Error('\P{Is_General_Category=/a/		LM}');
Expect(1, 94111, '\p{Is_General_Category=lm}', "");
Expect(0, 94111, '\p{^Is_General_Category=lm}', "");
Expect(0, 94111, '\P{Is_General_Category=lm}', "");
Expect(1, 94111, '\P{^Is_General_Category=lm}', "");
Expect(0, 94112, '\p{Is_General_Category=lm}', "");
Expect(1, 94112, '\p{^Is_General_Category=lm}', "");
Expect(1, 94112, '\P{Is_General_Category=lm}', "");
Expect(0, 94112, '\P{^Is_General_Category=lm}', "");
Expect(1, 94111, '\p{Is_General_Category: _lm}', "");
Expect(0, 94111, '\p{^Is_General_Category: _lm}', "");
Expect(0, 94111, '\P{Is_General_Category: _lm}', "");
Expect(1, 94111, '\P{^Is_General_Category: _lm}', "");
Expect(0, 94112, '\p{Is_General_Category: _lm}', "");
Expect(1, 94112, '\p{^Is_General_Category: _lm}', "");
Expect(1, 94112, '\P{Is_General_Category: _lm}', "");
Expect(0, 94112, '\P{^Is_General_Category: _lm}', "");
Error('\p{Is_Gc=	Modifier_Letter/a/}');
Error('\P{Is_Gc=	Modifier_Letter/a/}');
Expect(1, 94111, '\p{Is_Gc=modifierletter}', "");
Expect(0, 94111, '\p{^Is_Gc=modifierletter}', "");
Expect(0, 94111, '\P{Is_Gc=modifierletter}', "");
Expect(1, 94111, '\P{^Is_Gc=modifierletter}', "");
Expect(0, 94112, '\p{Is_Gc=modifierletter}', "");
Expect(1, 94112, '\p{^Is_Gc=modifierletter}', "");
Expect(1, 94112, '\P{Is_Gc=modifierletter}', "");
Expect(0, 94112, '\P{^Is_Gc=modifierletter}', "");
Expect(1, 94111, '\p{Is_Gc= modifier_Letter}', "");
Expect(0, 94111, '\p{^Is_Gc= modifier_Letter}', "");
Expect(0, 94111, '\P{Is_Gc= modifier_Letter}', "");
Expect(1, 94111, '\P{^Is_Gc= modifier_Letter}', "");
Expect(0, 94112, '\p{Is_Gc= modifier_Letter}', "");
Expect(1, 94112, '\p{^Is_Gc= modifier_Letter}', "");
Expect(1, 94112, '\P{Is_Gc= modifier_Letter}', "");
Expect(0, 94112, '\P{^Is_Gc= modifier_Letter}', "");
Error('\p{Is_Category=_ Lm/a/}');
Error('\P{Is_Category=_ Lm/a/}');
Expect(1, 94111, '\p{Is_Category=lm}', "");
Expect(0, 94111, '\p{^Is_Category=lm}', "");
Expect(0, 94111, '\P{Is_Category=lm}', "");
Expect(1, 94111, '\P{^Is_Category=lm}', "");
Expect(0, 94112, '\p{Is_Category=lm}', "");
Expect(1, 94112, '\p{^Is_Category=lm}', "");
Expect(1, 94112, '\P{Is_Category=lm}', "");
Expect(0, 94112, '\P{^Is_Category=lm}', "");
Expect(1, 94111, '\p{Is_Category=-Lm}', "");
Expect(0, 94111, '\p{^Is_Category=-Lm}', "");
Expect(0, 94111, '\P{Is_Category=-Lm}', "");
Expect(1, 94111, '\P{^Is_Category=-Lm}', "");
Expect(0, 94112, '\p{Is_Category=-Lm}', "");
Expect(1, 94112, '\p{^Is_Category=-Lm}', "");
Expect(1, 94112, '\P{Is_Category=-Lm}', "");
Expect(0, 94112, '\P{^Is_Category=-Lm}', "");
Error('\p{General_Category=_Other_Letter/a/}');
Error('\P{General_Category=_Other_Letter/a/}');
Expect(1, 195101, '\p{General_Category=otherletter}', "");
Expect(0, 195101, '\p{^General_Category=otherletter}', "");
Expect(0, 195101, '\P{General_Category=otherletter}', "");
Expect(1, 195101, '\P{^General_Category=otherletter}', "");
Expect(0, 195102, '\p{General_Category=otherletter}', "");
Expect(1, 195102, '\p{^General_Category=otherletter}', "");
Expect(1, 195102, '\P{General_Category=otherletter}', "");
Expect(0, 195102, '\P{^General_Category=otherletter}', "");
Expect(1, 195101, '\p{General_Category= -Other_letter}', "");
Expect(0, 195101, '\p{^General_Category= -Other_letter}', "");
Expect(0, 195101, '\P{General_Category= -Other_letter}', "");
Expect(1, 195101, '\P{^General_Category= -Other_letter}', "");
Expect(0, 195102, '\p{General_Category= -Other_letter}', "");
Expect(1, 195102, '\p{^General_Category= -Other_letter}', "");
Expect(1, 195102, '\P{General_Category= -Other_letter}', "");
Expect(0, 195102, '\P{^General_Category= -Other_letter}', "");
Error('\p{Gc= lo:=}');
Error('\P{Gc= lo:=}');
Expect(1, 195101, '\p{Gc=lo}', "");
Expect(0, 195101, '\p{^Gc=lo}', "");
Expect(0, 195101, '\P{Gc=lo}', "");
Expect(1, 195101, '\P{^Gc=lo}', "");
Expect(0, 195102, '\p{Gc=lo}', "");
Expect(1, 195102, '\p{^Gc=lo}', "");
Expect(1, 195102, '\P{Gc=lo}', "");
Expect(0, 195102, '\P{^Gc=lo}', "");
Expect(1, 195101, '\p{Gc=	lo}', "");
Expect(0, 195101, '\p{^Gc=	lo}', "");
Expect(0, 195101, '\P{Gc=	lo}', "");
Expect(1, 195101, '\P{^Gc=	lo}', "");
Expect(0, 195102, '\p{Gc=	lo}', "");
Expect(1, 195102, '\p{^Gc=	lo}', "");
Expect(1, 195102, '\P{Gc=	lo}', "");
Expect(0, 195102, '\P{^Gc=	lo}', "");
Error('\p{Category=	 Other_Letter/a/}');
Error('\P{Category=	 Other_Letter/a/}');
Expect(1, 195101, '\p{Category=otherletter}', "");
Expect(0, 195101, '\p{^Category=otherletter}', "");
Expect(0, 195101, '\P{Category=otherletter}', "");
Expect(1, 195101, '\P{^Category=otherletter}', "");
Expect(0, 195102, '\p{Category=otherletter}', "");
Expect(1, 195102, '\p{^Category=otherletter}', "");
Expect(1, 195102, '\P{Category=otherletter}', "");
Expect(0, 195102, '\P{^Category=otherletter}', "");
Expect(1, 195101, '\p{Category:    OTHER_LETTER}', "");
Expect(0, 195101, '\p{^Category:    OTHER_LETTER}', "");
Expect(0, 195101, '\P{Category:    OTHER_LETTER}', "");
Expect(1, 195101, '\P{^Category:    OTHER_LETTER}', "");
Expect(0, 195102, '\p{Category:    OTHER_LETTER}', "");
Expect(1, 195102, '\p{^Category:    OTHER_LETTER}', "");
Expect(1, 195102, '\P{Category:    OTHER_LETTER}', "");
Expect(0, 195102, '\P{^Category:    OTHER_LETTER}', "");
Error('\p{Is_General_Category= -Lo/a/}');
Error('\P{Is_General_Category= -Lo/a/}');
Expect(1, 195101, '\p{Is_General_Category=lo}', "");
Expect(0, 195101, '\p{^Is_General_Category=lo}', "");
Expect(0, 195101, '\P{Is_General_Category=lo}', "");
Expect(1, 195101, '\P{^Is_General_Category=lo}', "");
Expect(0, 195102, '\p{Is_General_Category=lo}', "");
Expect(1, 195102, '\p{^Is_General_Category=lo}', "");
Expect(1, 195102, '\P{Is_General_Category=lo}', "");
Expect(0, 195102, '\P{^Is_General_Category=lo}', "");
Expect(1, 195101, '\p{Is_General_Category=_ lo}', "");
Expect(0, 195101, '\p{^Is_General_Category=_ lo}', "");
Expect(0, 195101, '\P{Is_General_Category=_ lo}', "");
Expect(1, 195101, '\P{^Is_General_Category=_ lo}', "");
Expect(0, 195102, '\p{Is_General_Category=_ lo}', "");
Expect(1, 195102, '\p{^Is_General_Category=_ lo}', "");
Expect(1, 195102, '\P{Is_General_Category=_ lo}', "");
Expect(0, 195102, '\P{^Is_General_Category=_ lo}', "");
Error('\p{Is_Gc:/a/_Other_LETTER}');
Error('\P{Is_Gc:/a/_Other_LETTER}');
Expect(1, 195101, '\p{Is_Gc=otherletter}', "");
Expect(0, 195101, '\p{^Is_Gc=otherletter}', "");
Expect(0, 195101, '\P{Is_Gc=otherletter}', "");
Expect(1, 195101, '\P{^Is_Gc=otherletter}', "");
Expect(0, 195102, '\p{Is_Gc=otherletter}', "");
Expect(1, 195102, '\p{^Is_Gc=otherletter}', "");
Expect(1, 195102, '\P{Is_Gc=otherletter}', "");
Expect(0, 195102, '\P{^Is_Gc=otherletter}', "");
Expect(1, 195101, '\p{Is_Gc=-	other_Letter}', "");
Expect(0, 195101, '\p{^Is_Gc=-	other_Letter}', "");
Expect(0, 195101, '\P{Is_Gc=-	other_Letter}', "");
Expect(1, 195101, '\P{^Is_Gc=-	other_Letter}', "");
Expect(0, 195102, '\p{Is_Gc=-	other_Letter}', "");
Expect(1, 195102, '\p{^Is_Gc=-	other_Letter}', "");
Expect(1, 195102, '\P{Is_Gc=-	other_Letter}', "");
Expect(0, 195102, '\P{^Is_Gc=-	other_Letter}', "");
Error('\p{Is_Category=__LO/a/}');
Error('\P{Is_Category=__LO/a/}');
Expect(1, 195101, '\p{Is_Category=lo}', "");
Expect(0, 195101, '\p{^Is_Category=lo}', "");
Expect(0, 195101, '\P{Is_Category=lo}', "");
Expect(1, 195101, '\P{^Is_Category=lo}', "");
Expect(0, 195102, '\p{Is_Category=lo}', "");
Expect(1, 195102, '\p{^Is_Category=lo}', "");
Expect(1, 195102, '\P{Is_Category=lo}', "");
Expect(0, 195102, '\P{^Is_Category=lo}', "");
Expect(1, 195101, '\p{Is_Category=		lo}', "");
Expect(0, 195101, '\p{^Is_Category=		lo}', "");
Expect(0, 195101, '\P{Is_Category=		lo}', "");
Expect(1, 195101, '\P{^Is_Category=		lo}', "");
Expect(0, 195102, '\p{Is_Category=		lo}', "");
Expect(1, 195102, '\p{^Is_Category=		lo}', "");
Expect(1, 195102, '\P{Is_Category=		lo}', "");
Expect(0, 195102, '\P{^Is_Category=		lo}', "");
Error('\p{General_Category:_/a/TITLECASE_letter}');
Error('\P{General_Category:_/a/TITLECASE_letter}');
Expect(1, 8188, '\p{General_Category=titlecaseletter}', "");
Expect(0, 8188, '\p{^General_Category=titlecaseletter}', "");
Expect(0, 8188, '\P{General_Category=titlecaseletter}', "");
Expect(1, 8188, '\P{^General_Category=titlecaseletter}', "");
Expect(0, 8189, '\p{General_Category=titlecaseletter}', "");
Expect(1, 8189, '\p{^General_Category=titlecaseletter}', "");
Expect(1, 8189, '\P{General_Category=titlecaseletter}', "");
Expect(0, 8189, '\P{^General_Category=titlecaseletter}', "");
Expect(1, 8188, '\p{General_Category: 	-titlecase_Letter}', "");
Expect(0, 8188, '\p{^General_Category: 	-titlecase_Letter}', "");
Expect(0, 8188, '\P{General_Category: 	-titlecase_Letter}', "");
Expect(1, 8188, '\P{^General_Category: 	-titlecase_Letter}', "");
Expect(0, 8189, '\p{General_Category: 	-titlecase_Letter}', "");
Expect(1, 8189, '\p{^General_Category: 	-titlecase_Letter}', "");
Expect(1, 8189, '\P{General_Category: 	-titlecase_Letter}', "");
Expect(0, 8189, '\P{^General_Category: 	-titlecase_Letter}', "");
Error('\p{Gc= -Lt/a/}');
Error('\P{Gc= -Lt/a/}');
Expect(1, 8188, '\p{Gc=lt}', "");
Expect(0, 8188, '\p{^Gc=lt}', "");
Expect(0, 8188, '\P{Gc=lt}', "");
Expect(1, 8188, '\P{^Gc=lt}', "");
Expect(0, 8189, '\p{Gc=lt}', "");
Expect(1, 8189, '\p{^Gc=lt}', "");
Expect(1, 8189, '\P{Gc=lt}', "");
Expect(0, 8189, '\P{^Gc=lt}', "");
Expect(1, 8188, '\p{Gc= 	LT}', "");
Expect(0, 8188, '\p{^Gc= 	LT}', "");
Expect(0, 8188, '\P{Gc= 	LT}', "");
Expect(1, 8188, '\P{^Gc= 	LT}', "");
Expect(0, 8189, '\p{Gc= 	LT}', "");
Expect(1, 8189, '\p{^Gc= 	LT}', "");
Expect(1, 8189, '\P{Gc= 	LT}', "");
Expect(0, 8189, '\P{^Gc= 	LT}', "");
Error('\p{Category=-Titlecase_Letter/a/}');
Error('\P{Category=-Titlecase_Letter/a/}');
Expect(1, 8188, '\p{Category=titlecaseletter}', "");
Expect(0, 8188, '\p{^Category=titlecaseletter}', "");
Expect(0, 8188, '\P{Category=titlecaseletter}', "");
Expect(1, 8188, '\P{^Category=titlecaseletter}', "");
Expect(0, 8189, '\p{Category=titlecaseletter}', "");
Expect(1, 8189, '\p{^Category=titlecaseletter}', "");
Expect(1, 8189, '\P{Category=titlecaseletter}', "");
Expect(0, 8189, '\P{^Category=titlecaseletter}', "");
Expect(1, 8188, '\p{Category=_TITLECASE_Letter}', "");
Expect(0, 8188, '\p{^Category=_TITLECASE_Letter}', "");
Expect(0, 8188, '\P{Category=_TITLECASE_Letter}', "");
Expect(1, 8188, '\P{^Category=_TITLECASE_Letter}', "");
Expect(0, 8189, '\p{Category=_TITLECASE_Letter}', "");
Expect(1, 8189, '\p{^Category=_TITLECASE_Letter}', "");
Expect(1, 8189, '\P{Category=_TITLECASE_Letter}', "");
Expect(0, 8189, '\P{^Category=_TITLECASE_Letter}', "");
Error('\p{Is_General_Category=_:=LT}');
Error('\P{Is_General_Category=_:=LT}');
Expect(1, 8188, '\p{Is_General_Category=lt}', "");
Expect(0, 8188, '\p{^Is_General_Category=lt}', "");
Expect(0, 8188, '\P{Is_General_Category=lt}', "");
Expect(1, 8188, '\P{^Is_General_Category=lt}', "");
Expect(0, 8189, '\p{Is_General_Category=lt}', "");
Expect(1, 8189, '\p{^Is_General_Category=lt}', "");
Expect(1, 8189, '\P{Is_General_Category=lt}', "");
Expect(0, 8189, '\P{^Is_General_Category=lt}', "");
Expect(1, 8188, '\p{Is_General_Category=-	Lt}', "");
Expect(0, 8188, '\p{^Is_General_Category=-	Lt}', "");
Expect(0, 8188, '\P{Is_General_Category=-	Lt}', "");
Expect(1, 8188, '\P{^Is_General_Category=-	Lt}', "");
Expect(0, 8189, '\p{Is_General_Category=-	Lt}', "");
Expect(1, 8189, '\p{^Is_General_Category=-	Lt}', "");
Expect(1, 8189, '\P{Is_General_Category=-	Lt}', "");
Expect(0, 8189, '\P{^Is_General_Category=-	Lt}', "");
Error('\p{Is_Gc=	 Titlecase_Letter/a/}');
Error('\P{Is_Gc=	 Titlecase_Letter/a/}');
Expect(1, 8188, '\p{Is_Gc:	titlecaseletter}', "");
Expect(0, 8188, '\p{^Is_Gc:	titlecaseletter}', "");
Expect(0, 8188, '\P{Is_Gc:	titlecaseletter}', "");
Expect(1, 8188, '\P{^Is_Gc:	titlecaseletter}', "");
Expect(0, 8189, '\p{Is_Gc:	titlecaseletter}', "");
Expect(1, 8189, '\p{^Is_Gc:	titlecaseletter}', "");
Expect(1, 8189, '\P{Is_Gc:	titlecaseletter}', "");
Expect(0, 8189, '\P{^Is_Gc:	titlecaseletter}', "");
Expect(1, 8188, '\p{Is_Gc=-titlecase_LETTER}', "");
Expect(0, 8188, '\p{^Is_Gc=-titlecase_LETTER}', "");
Expect(0, 8188, '\P{Is_Gc=-titlecase_LETTER}', "");
Expect(1, 8188, '\P{^Is_Gc=-titlecase_LETTER}', "");
Expect(0, 8189, '\p{Is_Gc=-titlecase_LETTER}', "");
Expect(1, 8189, '\p{^Is_Gc=-titlecase_LETTER}', "");
Expect(1, 8189, '\P{Is_Gc=-titlecase_LETTER}', "");
Expect(0, 8189, '\P{^Is_Gc=-titlecase_LETTER}', "");
Error('\p{Is_Category=-LT/a/}');
Error('\P{Is_Category=-LT/a/}');
Expect(1, 8188, '\p{Is_Category=lt}', "");
Expect(0, 8188, '\p{^Is_Category=lt}', "");
Expect(0, 8188, '\P{Is_Category=lt}', "");
Expect(1, 8188, '\P{^Is_Category=lt}', "");
Expect(0, 8189, '\p{Is_Category=lt}', "");
Expect(1, 8189, '\p{^Is_Category=lt}', "");
Expect(1, 8189, '\P{Is_Category=lt}', "");
Expect(0, 8189, '\P{^Is_Category=lt}', "");
Expect(1, 8188, '\p{Is_Category=_-Lt}', "");
Expect(0, 8188, '\p{^Is_Category=_-Lt}', "");
Expect(0, 8188, '\P{Is_Category=_-Lt}', "");
Expect(1, 8188, '\P{^Is_Category=_-Lt}', "");
Expect(0, 8189, '\p{Is_Category=_-Lt}', "");
Expect(1, 8189, '\p{^Is_Category=_-Lt}', "");
Expect(1, 8189, '\P{Is_Category=_-Lt}', "");
Expect(0, 8189, '\P{^Is_Category=_-Lt}', "");
Error('\p{General_Category=	_uppercase_Letter/a/}');
Error('\P{General_Category=	_uppercase_Letter/a/}');
Expect(1, 120778, '\p{General_Category:uppercaseletter}', "");
Expect(0, 120778, '\p{^General_Category:uppercaseletter}', "");
Expect(0, 120778, '\P{General_Category:uppercaseletter}', "");
Expect(1, 120778, '\P{^General_Category:uppercaseletter}', "");
Expect(0, 120779, '\p{General_Category:uppercaseletter}', "");
Expect(1, 120779, '\p{^General_Category:uppercaseletter}', "");
Expect(1, 120779, '\P{General_Category:uppercaseletter}', "");
Expect(0, 120779, '\P{^General_Category:uppercaseletter}', "");
Expect(1, 120778, '\p{General_Category= UPPERCASE_LETTER}', "");
Expect(0, 120778, '\p{^General_Category= UPPERCASE_LETTER}', "");
Expect(0, 120778, '\P{General_Category= UPPERCASE_LETTER}', "");
Expect(1, 120778, '\P{^General_Category= UPPERCASE_LETTER}', "");
Expect(0, 120779, '\p{General_Category= UPPERCASE_LETTER}', "");
Expect(1, 120779, '\p{^General_Category= UPPERCASE_LETTER}', "");
Expect(1, 120779, '\P{General_Category= UPPERCASE_LETTER}', "");
Expect(0, 120779, '\P{^General_Category= UPPERCASE_LETTER}', "");
Error('\p{Gc=-_Lu/a/}');
Error('\P{Gc=-_Lu/a/}');
Expect(1, 120778, '\p{Gc=lu}', "");
Expect(0, 120778, '\p{^Gc=lu}', "");
Expect(0, 120778, '\P{Gc=lu}', "");
Expect(1, 120778, '\P{^Gc=lu}', "");
Expect(0, 120779, '\p{Gc=lu}', "");
Expect(1, 120779, '\p{^Gc=lu}', "");
Expect(1, 120779, '\P{Gc=lu}', "");
Expect(0, 120779, '\P{^Gc=lu}', "");
Expect(1, 120778, '\p{Gc=	 Lu}', "");
Expect(0, 120778, '\p{^Gc=	 Lu}', "");
Expect(0, 120778, '\P{Gc=	 Lu}', "");
Expect(1, 120778, '\P{^Gc=	 Lu}', "");
Expect(0, 120779, '\p{Gc=	 Lu}', "");
Expect(1, 120779, '\p{^Gc=	 Lu}', "");
Expect(1, 120779, '\P{Gc=	 Lu}', "");
Expect(0, 120779, '\P{^Gc=	 Lu}', "");
Error('\p{Category=__UPPERCASE_letter:=}');
Error('\P{Category=__UPPERCASE_letter:=}');
Expect(1, 120778, '\p{Category=uppercaseletter}', "");
Expect(0, 120778, '\p{^Category=uppercaseletter}', "");
Expect(0, 120778, '\P{Category=uppercaseletter}', "");
Expect(1, 120778, '\P{^Category=uppercaseletter}', "");
Expect(0, 120779, '\p{Category=uppercaseletter}', "");
Expect(1, 120779, '\p{^Category=uppercaseletter}', "");
Expect(1, 120779, '\P{Category=uppercaseletter}', "");
Expect(0, 120779, '\P{^Category=uppercaseletter}', "");
Expect(1, 120778, '\p{Category= _uppercase_letter}', "");
Expect(0, 120778, '\p{^Category= _uppercase_letter}', "");
Expect(0, 120778, '\P{Category= _uppercase_letter}', "");
Expect(1, 120778, '\P{^Category= _uppercase_letter}', "");
Expect(0, 120779, '\p{Category= _uppercase_letter}', "");
Expect(1, 120779, '\p{^Category= _uppercase_letter}', "");
Expect(1, 120779, '\P{Category= _uppercase_letter}', "");
Expect(0, 120779, '\P{^Category= _uppercase_letter}', "");
Error('\p{Is_General_Category=/a/_	LU}');
Error('\P{Is_General_Category=/a/_	LU}');
Expect(1, 120778, '\p{Is_General_Category=lu}', "");
Expect(0, 120778, '\p{^Is_General_Category=lu}', "");
Expect(0, 120778, '\P{Is_General_Category=lu}', "");
Expect(1, 120778, '\P{^Is_General_Category=lu}', "");
Expect(0, 120779, '\p{Is_General_Category=lu}', "");
Expect(1, 120779, '\p{^Is_General_Category=lu}', "");
Expect(1, 120779, '\P{Is_General_Category=lu}', "");
Expect(0, 120779, '\P{^Is_General_Category=lu}', "");
Expect(1, 120778, '\p{Is_General_Category:-LU}', "");
Expect(0, 120778, '\p{^Is_General_Category:-LU}', "");
Expect(0, 120778, '\P{Is_General_Category:-LU}', "");
Expect(1, 120778, '\P{^Is_General_Category:-LU}', "");
Expect(0, 120779, '\p{Is_General_Category:-LU}', "");
Expect(1, 120779, '\p{^Is_General_Category:-LU}', "");
Expect(1, 120779, '\P{Is_General_Category:-LU}', "");
Expect(0, 120779, '\P{^Is_General_Category:-LU}', "");
Error('\p{Is_Gc=/a/Uppercase_Letter}');
Error('\P{Is_Gc=/a/Uppercase_Letter}');
Expect(1, 120778, '\p{Is_Gc=uppercaseletter}', "");
Expect(0, 120778, '\p{^Is_Gc=uppercaseletter}', "");
Expect(0, 120778, '\P{Is_Gc=uppercaseletter}', "");
Expect(1, 120778, '\P{^Is_Gc=uppercaseletter}', "");
Expect(0, 120779, '\p{Is_Gc=uppercaseletter}', "");
Expect(1, 120779, '\p{^Is_Gc=uppercaseletter}', "");
Expect(1, 120779, '\P{Is_Gc=uppercaseletter}', "");
Expect(0, 120779, '\P{^Is_Gc=uppercaseletter}', "");
Expect(1, 120778, '\p{Is_Gc=_UPPERCASE_Letter}', "");
Expect(0, 120778, '\p{^Is_Gc=_UPPERCASE_Letter}', "");
Expect(0, 120778, '\P{Is_Gc=_UPPERCASE_Letter}', "");
Expect(1, 120778, '\P{^Is_Gc=_UPPERCASE_Letter}', "");
Expect(0, 120779, '\p{Is_Gc=_UPPERCASE_Letter}', "");
Expect(1, 120779, '\p{^Is_Gc=_UPPERCASE_Letter}', "");
Expect(1, 120779, '\P{Is_Gc=_UPPERCASE_Letter}', "");
Expect(0, 120779, '\P{^Is_Gc=_UPPERCASE_Letter}', "");
Error('\p{Is_Category= LU/a/}');
Error('\P{Is_Category= LU/a/}');
Expect(1, 120778, '\p{Is_Category=lu}', "");
Expect(0, 120778, '\p{^Is_Category=lu}', "");
Expect(0, 120778, '\P{Is_Category=lu}', "");
Expect(1, 120778, '\P{^Is_Category=lu}', "");
Expect(0, 120779, '\p{Is_Category=lu}', "");
Expect(1, 120779, '\p{^Is_Category=lu}', "");
Expect(1, 120779, '\P{Is_Category=lu}', "");
Expect(0, 120779, '\P{^Is_Category=lu}', "");
Expect(1, 120778, '\p{Is_Category=_	Lu}', "");
Expect(0, 120778, '\p{^Is_Category=_	Lu}', "");
Expect(0, 120778, '\P{Is_Category=_	Lu}', "");
Expect(1, 120778, '\P{^Is_Category=_	Lu}', "");
Expect(0, 120779, '\p{Is_Category=_	Lu}', "");
Expect(1, 120779, '\p{^Is_Category=_	Lu}', "");
Expect(1, 120779, '\P{Is_Category=_	Lu}', "");
Expect(0, 120779, '\P{^Is_Category=_	Lu}', "");
Error('\p{General_Category=-/a/MARK}');
Error('\P{General_Category=-/a/MARK}');
Expect(1, 917999, '\p{General_Category=mark}', "");
Expect(0, 917999, '\p{^General_Category=mark}', "");
Expect(0, 917999, '\P{General_Category=mark}', "");
Expect(1, 917999, '\P{^General_Category=mark}', "");
Expect(0, 918000, '\p{General_Category=mark}', "");
Expect(1, 918000, '\p{^General_Category=mark}', "");
Expect(1, 918000, '\P{General_Category=mark}', "");
Expect(0, 918000, '\P{^General_Category=mark}', "");
Expect(1, 917999, '\p{General_Category=_-Mark}', "");
Expect(0, 917999, '\p{^General_Category=_-Mark}', "");
Expect(0, 917999, '\P{General_Category=_-Mark}', "");
Expect(1, 917999, '\P{^General_Category=_-Mark}', "");
Expect(0, 918000, '\p{General_Category=_-Mark}', "");
Expect(1, 918000, '\p{^General_Category=_-Mark}', "");
Expect(1, 918000, '\P{General_Category=_-Mark}', "");
Expect(0, 918000, '\P{^General_Category=_-Mark}', "");
Error('\p{Gc=	:=M}');
Error('\P{Gc=	:=M}');
Expect(1, 917999, '\p{Gc=m}', "");
Expect(0, 917999, '\p{^Gc=m}', "");
Expect(0, 917999, '\P{Gc=m}', "");
Expect(1, 917999, '\P{^Gc=m}', "");
Expect(0, 918000, '\p{Gc=m}', "");
Expect(1, 918000, '\p{^Gc=m}', "");
Expect(1, 918000, '\P{Gc=m}', "");
Expect(0, 918000, '\P{^Gc=m}', "");
Expect(1, 917999, '\p{Gc= -M}', "");
Expect(0, 917999, '\p{^Gc= -M}', "");
Expect(0, 917999, '\P{Gc= -M}', "");
Expect(1, 917999, '\P{^Gc= -M}', "");
Expect(0, 918000, '\p{Gc= -M}', "");
Expect(1, 918000, '\p{^Gc= -M}', "");
Expect(1, 918000, '\P{Gc= -M}', "");
Expect(0, 918000, '\P{^Gc= -M}', "");
Error('\p{Category=_ combining_mark:=}');
Error('\P{Category=_ combining_mark:=}');
Expect(1, 917999, '\p{Category=combiningmark}', "");
Expect(0, 917999, '\p{^Category=combiningmark}', "");
Expect(0, 917999, '\P{Category=combiningmark}', "");
Expect(1, 917999, '\P{^Category=combiningmark}', "");
Expect(0, 918000, '\p{Category=combiningmark}', "");
Expect(1, 918000, '\p{^Category=combiningmark}', "");
Expect(1, 918000, '\P{Category=combiningmark}', "");
Expect(0, 918000, '\P{^Category=combiningmark}', "");
Expect(1, 917999, '\p{Category=	 COMBINING_Mark}', "");
Expect(0, 917999, '\p{^Category=	 COMBINING_Mark}', "");
Expect(0, 917999, '\P{Category=	 COMBINING_Mark}', "");
Expect(1, 917999, '\P{^Category=	 COMBINING_Mark}', "");
Expect(0, 918000, '\p{Category=	 COMBINING_Mark}', "");
Expect(1, 918000, '\p{^Category=	 COMBINING_Mark}', "");
Expect(1, 918000, '\P{Category=	 COMBINING_Mark}', "");
Expect(0, 918000, '\P{^Category=	 COMBINING_Mark}', "");
Error('\p{Is_General_Category=:=	_MARK}');
Error('\P{Is_General_Category=:=	_MARK}');
Expect(1, 917999, '\p{Is_General_Category=mark}', "");
Expect(0, 917999, '\p{^Is_General_Category=mark}', "");
Expect(0, 917999, '\P{Is_General_Category=mark}', "");
Expect(1, 917999, '\P{^Is_General_Category=mark}', "");
Expect(0, 918000, '\p{Is_General_Category=mark}', "");
Expect(1, 918000, '\p{^Is_General_Category=mark}', "");
Expect(1, 918000, '\P{Is_General_Category=mark}', "");
Expect(0, 918000, '\P{^Is_General_Category=mark}', "");
Expect(1, 917999, '\p{Is_General_Category=_Mark}', "");
Expect(0, 917999, '\p{^Is_General_Category=_Mark}', "");
Expect(0, 917999, '\P{Is_General_Category=_Mark}', "");
Expect(1, 917999, '\P{^Is_General_Category=_Mark}', "");
Expect(0, 918000, '\p{Is_General_Category=_Mark}', "");
Expect(1, 918000, '\p{^Is_General_Category=_Mark}', "");
Expect(1, 918000, '\P{Is_General_Category=_Mark}', "");
Expect(0, 918000, '\P{^Is_General_Category=_Mark}', "");
Error('\p{Is_Gc=/a/ -M}');
Error('\P{Is_Gc=/a/ -M}');
Expect(1, 917999, '\p{Is_Gc:   m}', "");
Expect(0, 917999, '\p{^Is_Gc:   m}', "");
Expect(0, 917999, '\P{Is_Gc:   m}', "");
Expect(1, 917999, '\P{^Is_Gc:   m}', "");
Expect(0, 918000, '\p{Is_Gc:   m}', "");
Expect(1, 918000, '\p{^Is_Gc:   m}', "");
Expect(1, 918000, '\P{Is_Gc:   m}', "");
Expect(0, 918000, '\P{^Is_Gc:   m}', "");
Expect(1, 917999, '\p{Is_Gc=	_m}', "");
Expect(0, 917999, '\p{^Is_Gc=	_m}', "");
Expect(0, 917999, '\P{Is_Gc=	_m}', "");
Expect(1, 917999, '\P{^Is_Gc=	_m}', "");
Expect(0, 918000, '\p{Is_Gc=	_m}', "");
Expect(1, 918000, '\p{^Is_Gc=	_m}', "");
Expect(1, 918000, '\P{Is_Gc=	_m}', "");
Expect(0, 918000, '\P{^Is_Gc=	_m}', "");
Error('\p{Is_Category=:=COMBINING_mark}');
Error('\P{Is_Category=:=COMBINING_mark}');
Expect(1, 917999, '\p{Is_Category:combiningmark}', "");
Expect(0, 917999, '\p{^Is_Category:combiningmark}', "");
Expect(0, 917999, '\P{Is_Category:combiningmark}', "");
Expect(1, 917999, '\P{^Is_Category:combiningmark}', "");
Expect(0, 918000, '\p{Is_Category:combiningmark}', "");
Expect(1, 918000, '\p{^Is_Category:combiningmark}', "");
Expect(1, 918000, '\P{Is_Category:combiningmark}', "");
Expect(0, 918000, '\P{^Is_Category:combiningmark}', "");
Expect(1, 917999, '\p{Is_Category=-COMBINING_Mark}', "");
Expect(0, 917999, '\p{^Is_Category=-COMBINING_Mark}', "");
Expect(0, 917999, '\P{Is_Category=-COMBINING_Mark}', "");
Expect(1, 917999, '\P{^Is_Category=-COMBINING_Mark}', "");
Expect(0, 918000, '\p{Is_Category=-COMBINING_Mark}', "");
Expect(1, 918000, '\p{^Is_Category=-COMBINING_Mark}', "");
Expect(1, 918000, '\P{Is_Category=-COMBINING_Mark}', "");
Expect(0, 918000, '\P{^Is_Category=-COMBINING_Mark}', "");
Error('\p{General_Category=:=	Spacing_Mark}');
Error('\P{General_Category=:=	Spacing_Mark}');
Expect(1, 119154, '\p{General_Category=spacingmark}', "");
Expect(0, 119154, '\p{^General_Category=spacingmark}', "");
Expect(0, 119154, '\P{General_Category=spacingmark}', "");
Expect(1, 119154, '\P{^General_Category=spacingmark}', "");
Expect(0, 119155, '\p{General_Category=spacingmark}', "");
Expect(1, 119155, '\p{^General_Category=spacingmark}', "");
Expect(1, 119155, '\P{General_Category=spacingmark}', "");
Expect(0, 119155, '\P{^General_Category=spacingmark}', "");
Expect(1, 119154, '\p{General_Category=--SPACING_mark}', "");
Expect(0, 119154, '\p{^General_Category=--SPACING_mark}', "");
Expect(0, 119154, '\P{General_Category=--SPACING_mark}', "");
Expect(1, 119154, '\P{^General_Category=--SPACING_mark}', "");
Expect(0, 119155, '\p{General_Category=--SPACING_mark}', "");
Expect(1, 119155, '\p{^General_Category=--SPACING_mark}', "");
Expect(1, 119155, '\P{General_Category=--SPACING_mark}', "");
Expect(0, 119155, '\P{^General_Category=--SPACING_mark}', "");
Error('\p{Gc:    	MC/a/}');
Error('\P{Gc:    	MC/a/}');
Expect(1, 119154, '\p{Gc=mc}', "");
Expect(0, 119154, '\p{^Gc=mc}', "");
Expect(0, 119154, '\P{Gc=mc}', "");
Expect(1, 119154, '\P{^Gc=mc}', "");
Expect(0, 119155, '\p{Gc=mc}', "");
Expect(1, 119155, '\p{^Gc=mc}', "");
Expect(1, 119155, '\P{Gc=mc}', "");
Expect(0, 119155, '\P{^Gc=mc}', "");
Expect(1, 119154, '\p{Gc=	_Mc}', "");
Expect(0, 119154, '\p{^Gc=	_Mc}', "");
Expect(0, 119154, '\P{Gc=	_Mc}', "");
Expect(1, 119154, '\P{^Gc=	_Mc}', "");
Expect(0, 119155, '\p{Gc=	_Mc}', "");
Expect(1, 119155, '\p{^Gc=	_Mc}', "");
Expect(1, 119155, '\P{Gc=	_Mc}', "");
Expect(0, 119155, '\P{^Gc=	_Mc}', "");
Error('\p{Category=:=	 Spacing_Mark}');
Error('\P{Category=:=	 Spacing_Mark}');
Expect(1, 119154, '\p{Category=spacingmark}', "");
Expect(0, 119154, '\p{^Category=spacingmark}', "");
Expect(0, 119154, '\P{Category=spacingmark}', "");
Expect(1, 119154, '\P{^Category=spacingmark}', "");
Expect(0, 119155, '\p{Category=spacingmark}', "");
Expect(1, 119155, '\p{^Category=spacingmark}', "");
Expect(1, 119155, '\P{Category=spacingmark}', "");
Expect(0, 119155, '\P{^Category=spacingmark}', "");
Expect(1, 119154, '\p{Category= spacing_Mark}', "");
Expect(0, 119154, '\p{^Category= spacing_Mark}', "");
Expect(0, 119154, '\P{Category= spacing_Mark}', "");
Expect(1, 119154, '\P{^Category= spacing_Mark}', "");
Expect(0, 119155, '\p{Category= spacing_Mark}', "");
Expect(1, 119155, '\p{^Category= spacing_Mark}', "");
Expect(1, 119155, '\P{Category= spacing_Mark}', "");
Expect(0, 119155, '\P{^Category= spacing_Mark}', "");
Error('\p{Is_General_Category=_Mc/a/}');
Error('\P{Is_General_Category=_Mc/a/}');
Expect(1, 119154, '\p{Is_General_Category=mc}', "");
Expect(0, 119154, '\p{^Is_General_Category=mc}', "");
Expect(0, 119154, '\P{Is_General_Category=mc}', "");
Expect(1, 119154, '\P{^Is_General_Category=mc}', "");
Expect(0, 119155, '\p{Is_General_Category=mc}', "");
Expect(1, 119155, '\p{^Is_General_Category=mc}', "");
Expect(1, 119155, '\P{Is_General_Category=mc}', "");
Expect(0, 119155, '\P{^Is_General_Category=mc}', "");
Expect(1, 119154, '\p{Is_General_Category=--MC}', "");
Expect(0, 119154, '\p{^Is_General_Category=--MC}', "");
Expect(0, 119154, '\P{Is_General_Category=--MC}', "");
Expect(1, 119154, '\P{^Is_General_Category=--MC}', "");
Expect(0, 119155, '\p{Is_General_Category=--MC}', "");
Expect(1, 119155, '\p{^Is_General_Category=--MC}', "");
Expect(1, 119155, '\P{Is_General_Category=--MC}', "");
Expect(0, 119155, '\P{^Is_General_Category=--MC}', "");
Error('\p{Is_Gc=_Spacing_MARK/a/}');
Error('\P{Is_Gc=_Spacing_MARK/a/}');
Expect(1, 119154, '\p{Is_Gc=spacingmark}', "");
Expect(0, 119154, '\p{^Is_Gc=spacingmark}', "");
Expect(0, 119154, '\P{Is_Gc=spacingmark}', "");
Expect(1, 119154, '\P{^Is_Gc=spacingmark}', "");
Expect(0, 119155, '\p{Is_Gc=spacingmark}', "");
Expect(1, 119155, '\p{^Is_Gc=spacingmark}', "");
Expect(1, 119155, '\P{Is_Gc=spacingmark}', "");
Expect(0, 119155, '\P{^Is_Gc=spacingmark}', "");
Expect(1, 119154, '\p{Is_Gc= Spacing_mark}', "");
Expect(0, 119154, '\p{^Is_Gc= Spacing_mark}', "");
Expect(0, 119154, '\P{Is_Gc= Spacing_mark}', "");
Expect(1, 119154, '\P{^Is_Gc= Spacing_mark}', "");
Expect(0, 119155, '\p{Is_Gc= Spacing_mark}', "");
Expect(1, 119155, '\p{^Is_Gc= Spacing_mark}', "");
Expect(1, 119155, '\P{Is_Gc= Spacing_mark}', "");
Expect(0, 119155, '\P{^Is_Gc= Spacing_mark}', "");
Error('\p{Is_Category=_ mc/a/}');
Error('\P{Is_Category=_ mc/a/}');
Expect(1, 119154, '\p{Is_Category=mc}', "");
Expect(0, 119154, '\p{^Is_Category=mc}', "");
Expect(0, 119154, '\P{Is_Category=mc}', "");
Expect(1, 119154, '\P{^Is_Category=mc}', "");
Expect(0, 119155, '\p{Is_Category=mc}', "");
Expect(1, 119155, '\p{^Is_Category=mc}', "");
Expect(1, 119155, '\P{Is_Category=mc}', "");
Expect(0, 119155, '\P{^Is_Category=mc}', "");
Expect(1, 119154, '\p{Is_Category=-	mc}', "");
Expect(0, 119154, '\p{^Is_Category=-	mc}', "");
Expect(0, 119154, '\P{Is_Category=-	mc}', "");
Expect(1, 119154, '\P{^Is_Category=-	mc}', "");
Expect(0, 119155, '\p{Is_Category=-	mc}', "");
Expect(1, 119155, '\p{^Is_Category=-	mc}', "");
Expect(1, 119155, '\P{Is_Category=-	mc}', "");
Expect(0, 119155, '\P{^Is_Category=-	mc}', "");
Error('\p{General_Category=	-Enclosing_Mark:=}');
Error('\P{General_Category=	-Enclosing_Mark:=}');
Expect(1, 42610, '\p{General_Category=enclosingmark}', "");
Expect(0, 42610, '\p{^General_Category=enclosingmark}', "");
Expect(0, 42610, '\P{General_Category=enclosingmark}', "");
Expect(1, 42610, '\P{^General_Category=enclosingmark}', "");
Expect(0, 42611, '\p{General_Category=enclosingmark}', "");
Expect(1, 42611, '\p{^General_Category=enclosingmark}', "");
Expect(1, 42611, '\P{General_Category=enclosingmark}', "");
Expect(0, 42611, '\P{^General_Category=enclosingmark}', "");
Expect(1, 42610, '\p{General_Category=_	ENCLOSING_MARK}', "");
Expect(0, 42610, '\p{^General_Category=_	ENCLOSING_MARK}', "");
Expect(0, 42610, '\P{General_Category=_	ENCLOSING_MARK}', "");
Expect(1, 42610, '\P{^General_Category=_	ENCLOSING_MARK}', "");
Expect(0, 42611, '\p{General_Category=_	ENCLOSING_MARK}', "");
Expect(1, 42611, '\p{^General_Category=_	ENCLOSING_MARK}', "");
Expect(1, 42611, '\P{General_Category=_	ENCLOSING_MARK}', "");
Expect(0, 42611, '\P{^General_Category=_	ENCLOSING_MARK}', "");
Error('\p{Gc=_-Me:=}');
Error('\P{Gc=_-Me:=}');
Expect(1, 42610, '\p{Gc=me}', "");
Expect(0, 42610, '\p{^Gc=me}', "");
Expect(0, 42610, '\P{Gc=me}', "");
Expect(1, 42610, '\P{^Gc=me}', "");
Expect(0, 42611, '\p{Gc=me}', "");
Expect(1, 42611, '\p{^Gc=me}', "");
Expect(1, 42611, '\P{Gc=me}', "");
Expect(0, 42611, '\P{^Gc=me}', "");
Expect(1, 42610, '\p{Gc= ME}', "");
Expect(0, 42610, '\p{^Gc= ME}', "");
Expect(0, 42610, '\P{Gc= ME}', "");
Expect(1, 42610, '\P{^Gc= ME}', "");
Expect(0, 42611, '\p{Gc= ME}', "");
Expect(1, 42611, '\p{^Gc= ME}', "");
Expect(1, 42611, '\P{Gc= ME}', "");
Expect(0, 42611, '\P{^Gc= ME}', "");
Error('\p{Category=/a/ENCLOSING_mark}');
Error('\P{Category=/a/ENCLOSING_mark}');
Expect(1, 42610, '\p{Category=enclosingmark}', "");
Expect(0, 42610, '\p{^Category=enclosingmark}', "");
Expect(0, 42610, '\P{Category=enclosingmark}', "");
Expect(1, 42610, '\P{^Category=enclosingmark}', "");
Expect(0, 42611, '\p{Category=enclosingmark}', "");
Expect(1, 42611, '\p{^Category=enclosingmark}', "");
Expect(1, 42611, '\P{Category=enclosingmark}', "");
Expect(0, 42611, '\P{^Category=enclosingmark}', "");
Expect(1, 42610, '\p{Category= -Enclosing_Mark}', "");
Expect(0, 42610, '\p{^Category= -Enclosing_Mark}', "");
Expect(0, 42610, '\P{Category= -Enclosing_Mark}', "");
Expect(1, 42610, '\P{^Category= -Enclosing_Mark}', "");
Expect(0, 42611, '\p{Category= -Enclosing_Mark}', "");
Expect(1, 42611, '\p{^Category= -Enclosing_Mark}', "");
Expect(1, 42611, '\P{Category= -Enclosing_Mark}', "");
Expect(0, 42611, '\P{^Category= -Enclosing_Mark}', "");
Error('\p{Is_General_Category:	_Me:=}');
Error('\P{Is_General_Category:	_Me:=}');
Expect(1, 42610, '\p{Is_General_Category=me}', "");
Expect(0, 42610, '\p{^Is_General_Category=me}', "");
Expect(0, 42610, '\P{Is_General_Category=me}', "");
Expect(1, 42610, '\P{^Is_General_Category=me}', "");
Expect(0, 42611, '\p{Is_General_Category=me}', "");
Expect(1, 42611, '\p{^Is_General_Category=me}', "");
Expect(1, 42611, '\P{Is_General_Category=me}', "");
Expect(0, 42611, '\P{^Is_General_Category=me}', "");
Expect(1, 42610, '\p{Is_General_Category=  ME}', "");
Expect(0, 42610, '\p{^Is_General_Category=  ME}', "");
Expect(0, 42610, '\P{Is_General_Category=  ME}', "");
Expect(1, 42610, '\P{^Is_General_Category=  ME}', "");
Expect(0, 42611, '\p{Is_General_Category=  ME}', "");
Expect(1, 42611, '\p{^Is_General_Category=  ME}', "");
Expect(1, 42611, '\P{Is_General_Category=  ME}', "");
Expect(0, 42611, '\P{^Is_General_Category=  ME}', "");
Error('\p{Is_Gc=/a/ _ENCLOSING_MARK}');
Error('\P{Is_Gc=/a/ _ENCLOSING_MARK}');
Expect(1, 42610, '\p{Is_Gc=enclosingmark}', "");
Expect(0, 42610, '\p{^Is_Gc=enclosingmark}', "");
Expect(0, 42610, '\P{Is_Gc=enclosingmark}', "");
Expect(1, 42610, '\P{^Is_Gc=enclosingmark}', "");
Expect(0, 42611, '\p{Is_Gc=enclosingmark}', "");
Expect(1, 42611, '\p{^Is_Gc=enclosingmark}', "");
Expect(1, 42611, '\P{Is_Gc=enclosingmark}', "");
Expect(0, 42611, '\P{^Is_Gc=enclosingmark}', "");
Expect(1, 42610, '\p{Is_Gc=--ENCLOSING_MARK}', "");
Expect(0, 42610, '\p{^Is_Gc=--ENCLOSING_MARK}', "");
Expect(0, 42610, '\P{Is_Gc=--ENCLOSING_MARK}', "");
Expect(1, 42610, '\P{^Is_Gc=--ENCLOSING_MARK}', "");
Expect(0, 42611, '\p{Is_Gc=--ENCLOSING_MARK}', "");
Expect(1, 42611, '\p{^Is_Gc=--ENCLOSING_MARK}', "");
Expect(1, 42611, '\P{Is_Gc=--ENCLOSING_MARK}', "");
Expect(0, 42611, '\P{^Is_Gc=--ENCLOSING_MARK}', "");
Error('\p{Is_Category=	:=ME}');
Error('\P{Is_Category=	:=ME}');
Expect(1, 42610, '\p{Is_Category:	me}', "");
Expect(0, 42610, '\p{^Is_Category:	me}', "");
Expect(0, 42610, '\P{Is_Category:	me}', "");
Expect(1, 42610, '\P{^Is_Category:	me}', "");
Expect(0, 42611, '\p{Is_Category:	me}', "");
Expect(1, 42611, '\p{^Is_Category:	me}', "");
Expect(1, 42611, '\P{Is_Category:	me}', "");
Expect(0, 42611, '\P{^Is_Category:	me}', "");
Expect(1, 42610, '\p{Is_Category= -me}', "");
Expect(0, 42610, '\p{^Is_Category= -me}', "");
Expect(0, 42610, '\P{Is_Category= -me}', "");
Expect(1, 42610, '\P{^Is_Category= -me}', "");
Expect(0, 42611, '\p{Is_Category= -me}', "");
Expect(1, 42611, '\p{^Is_Category= -me}', "");
Expect(1, 42611, '\P{Is_Category= -me}', "");
Expect(0, 42611, '\P{^Is_Category= -me}', "");
Error('\p{General_Category=-/a/Nonspacing_Mark}');
Error('\P{General_Category=-/a/Nonspacing_Mark}');
Expect(1, 917999, '\p{General_Category=nonspacingmark}', "");
Expect(0, 917999, '\p{^General_Category=nonspacingmark}', "");
Expect(0, 917999, '\P{General_Category=nonspacingmark}', "");
Expect(1, 917999, '\P{^General_Category=nonspacingmark}', "");
Expect(0, 918000, '\p{General_Category=nonspacingmark}', "");
Expect(1, 918000, '\p{^General_Category=nonspacingmark}', "");
Expect(1, 918000, '\P{General_Category=nonspacingmark}', "");
Expect(0, 918000, '\P{^General_Category=nonspacingmark}', "");
Expect(1, 917999, '\p{General_Category: 	_NONSPACING_mark}', "");
Expect(0, 917999, '\p{^General_Category: 	_NONSPACING_mark}', "");
Expect(0, 917999, '\P{General_Category: 	_NONSPACING_mark}', "");
Expect(1, 917999, '\P{^General_Category: 	_NONSPACING_mark}', "");
Expect(0, 918000, '\p{General_Category: 	_NONSPACING_mark}', "");
Expect(1, 918000, '\p{^General_Category: 	_NONSPACING_mark}', "");
Expect(1, 918000, '\P{General_Category: 	_NONSPACING_mark}', "");
Expect(0, 918000, '\P{^General_Category: 	_NONSPACING_mark}', "");
Error('\p{Gc=:=Mn}');
Error('\P{Gc=:=Mn}');
Expect(1, 917999, '\p{Gc=mn}', "");
Expect(0, 917999, '\p{^Gc=mn}', "");
Expect(0, 917999, '\P{Gc=mn}', "");
Expect(1, 917999, '\P{^Gc=mn}', "");
Expect(0, 918000, '\p{Gc=mn}', "");
Expect(1, 918000, '\p{^Gc=mn}', "");
Expect(1, 918000, '\P{Gc=mn}', "");
Expect(0, 918000, '\P{^Gc=mn}', "");
Expect(1, 917999, '\p{Gc=	 Mn}', "");
Expect(0, 917999, '\p{^Gc=	 Mn}', "");
Expect(0, 917999, '\P{Gc=	 Mn}', "");
Expect(1, 917999, '\P{^Gc=	 Mn}', "");
Expect(0, 918000, '\p{Gc=	 Mn}', "");
Expect(1, 918000, '\p{^Gc=	 Mn}', "");
Expect(1, 918000, '\P{Gc=	 Mn}', "");
Expect(0, 918000, '\P{^Gc=	 Mn}', "");
Error('\p{Category=:=	_Nonspacing_Mark}');
Error('\P{Category=:=	_Nonspacing_Mark}');
Expect(1, 917999, '\p{Category=nonspacingmark}', "");
Expect(0, 917999, '\p{^Category=nonspacingmark}', "");
Expect(0, 917999, '\P{Category=nonspacingmark}', "");
Expect(1, 917999, '\P{^Category=nonspacingmark}', "");
Expect(0, 918000, '\p{Category=nonspacingmark}', "");
Expect(1, 918000, '\p{^Category=nonspacingmark}', "");
Expect(1, 918000, '\P{Category=nonspacingmark}', "");
Expect(0, 918000, '\P{^Category=nonspacingmark}', "");
Expect(1, 917999, '\p{Category=-	Nonspacing_Mark}', "");
Expect(0, 917999, '\p{^Category=-	Nonspacing_Mark}', "");
Expect(0, 917999, '\P{Category=-	Nonspacing_Mark}', "");
Expect(1, 917999, '\P{^Category=-	Nonspacing_Mark}', "");
Expect(0, 918000, '\p{Category=-	Nonspacing_Mark}', "");
Expect(1, 918000, '\p{^Category=-	Nonspacing_Mark}', "");
Expect(1, 918000, '\P{Category=-	Nonspacing_Mark}', "");
Expect(0, 918000, '\P{^Category=-	Nonspacing_Mark}', "");
Error('\p{Is_General_Category=__MN/a/}');
Error('\P{Is_General_Category=__MN/a/}');
Expect(1, 917999, '\p{Is_General_Category=mn}', "");
Expect(0, 917999, '\p{^Is_General_Category=mn}', "");
Expect(0, 917999, '\P{Is_General_Category=mn}', "");
Expect(1, 917999, '\P{^Is_General_Category=mn}', "");
Expect(0, 918000, '\p{Is_General_Category=mn}', "");
Expect(1, 918000, '\p{^Is_General_Category=mn}', "");
Expect(1, 918000, '\P{Is_General_Category=mn}', "");
Expect(0, 918000, '\P{^Is_General_Category=mn}', "");
Expect(1, 917999, '\p{Is_General_Category=		MN}', "");
Expect(0, 917999, '\p{^Is_General_Category=		MN}', "");
Expect(0, 917999, '\P{Is_General_Category=		MN}', "");
Expect(1, 917999, '\P{^Is_General_Category=		MN}', "");
Expect(0, 918000, '\p{Is_General_Category=		MN}', "");
Expect(1, 918000, '\p{^Is_General_Category=		MN}', "");
Expect(1, 918000, '\P{Is_General_Category=		MN}', "");
Expect(0, 918000, '\P{^Is_General_Category=		MN}', "");
Error('\p{Is_Gc=--nonspacing_mark:=}');
Error('\P{Is_Gc=--nonspacing_mark:=}');
Expect(1, 917999, '\p{Is_Gc=nonspacingmark}', "");
Expect(0, 917999, '\p{^Is_Gc=nonspacingmark}', "");
Expect(0, 917999, '\P{Is_Gc=nonspacingmark}', "");
Expect(1, 917999, '\P{^Is_Gc=nonspacingmark}', "");
Expect(0, 918000, '\p{Is_Gc=nonspacingmark}', "");
Expect(1, 918000, '\p{^Is_Gc=nonspacingmark}', "");
Expect(1, 918000, '\P{Is_Gc=nonspacingmark}', "");
Expect(0, 918000, '\P{^Is_Gc=nonspacingmark}', "");
Expect(1, 917999, '\p{Is_Gc=	NONSPACING_MARK}', "");
Expect(0, 917999, '\p{^Is_Gc=	NONSPACING_MARK}', "");
Expect(0, 917999, '\P{Is_Gc=	NONSPACING_MARK}', "");
Expect(1, 917999, '\P{^Is_Gc=	NONSPACING_MARK}', "");
Expect(0, 918000, '\p{Is_Gc=	NONSPACING_MARK}', "");
Expect(1, 918000, '\p{^Is_Gc=	NONSPACING_MARK}', "");
Expect(1, 918000, '\P{Is_Gc=	NONSPACING_MARK}', "");
Expect(0, 918000, '\P{^Is_Gc=	NONSPACING_MARK}', "");
Error('\p{Is_Category=/a/Mn}');
Error('\P{Is_Category=/a/Mn}');
Expect(1, 917999, '\p{Is_Category=mn}', "");
Expect(0, 917999, '\p{^Is_Category=mn}', "");
Expect(0, 917999, '\P{Is_Category=mn}', "");
Expect(1, 917999, '\P{^Is_Category=mn}', "");
Expect(0, 918000, '\p{Is_Category=mn}', "");
Expect(1, 918000, '\p{^Is_Category=mn}', "");
Expect(1, 918000, '\P{Is_Category=mn}', "");
Expect(0, 918000, '\P{^Is_Category=mn}', "");
Expect(1, 917999, '\p{Is_Category= -mn}', "");
Expect(0, 917999, '\p{^Is_Category= -mn}', "");
Expect(0, 917999, '\P{Is_Category= -mn}', "");
Expect(1, 917999, '\P{^Is_Category= -mn}', "");
Expect(0, 918000, '\p{Is_Category= -mn}', "");
Expect(1, 918000, '\p{^Is_Category= -mn}', "");
Expect(1, 918000, '\P{Is_Category= -mn}', "");
Expect(0, 918000, '\P{^Is_Category= -mn}', "");
Error('\p{General_Category= -number/a/}');
Error('\P{General_Category= -number/a/}');
Expect(1, 127242, '\p{General_Category:	number}', "");
Expect(0, 127242, '\p{^General_Category:	number}', "");
Expect(0, 127242, '\P{General_Category:	number}', "");
Expect(1, 127242, '\P{^General_Category:	number}', "");
Expect(0, 127243, '\p{General_Category:	number}', "");
Expect(1, 127243, '\p{^General_Category:	number}', "");
Expect(1, 127243, '\P{General_Category:	number}', "");
Expect(0, 127243, '\P{^General_Category:	number}', "");
Expect(1, 127242, '\p{General_Category= _number}', "");
Expect(0, 127242, '\p{^General_Category= _number}', "");
Expect(0, 127242, '\P{General_Category= _number}', "");
Expect(1, 127242, '\P{^General_Category= _number}', "");
Expect(0, 127243, '\p{General_Category= _number}', "");
Expect(1, 127243, '\p{^General_Category= _number}', "");
Expect(1, 127243, '\P{General_Category= _number}', "");
Expect(0, 127243, '\P{^General_Category= _number}', "");
Error('\p{Gc=:=	 N}');
Error('\P{Gc=:=	 N}');
Expect(1, 127242, '\p{Gc:n}', "");
Expect(0, 127242, '\p{^Gc:n}', "");
Expect(0, 127242, '\P{Gc:n}', "");
Expect(1, 127242, '\P{^Gc:n}', "");
Expect(0, 127243, '\p{Gc:n}', "");
Expect(1, 127243, '\p{^Gc:n}', "");
Expect(1, 127243, '\P{Gc:n}', "");
Expect(0, 127243, '\P{^Gc:n}', "");
Expect(1, 127242, '\p{Gc=  N}', "");
Expect(0, 127242, '\p{^Gc=  N}', "");
Expect(0, 127242, '\P{Gc=  N}', "");
Expect(1, 127242, '\P{^Gc=  N}', "");
Expect(0, 127243, '\p{Gc=  N}', "");
Expect(1, 127243, '\p{^Gc=  N}', "");
Expect(1, 127243, '\P{Gc=  N}', "");
Expect(0, 127243, '\P{^Gc=  N}', "");
Error('\p{Category=-/a/Number}');
Error('\P{Category=-/a/Number}');
Expect(1, 127242, '\p{Category=number}', "");
Expect(0, 127242, '\p{^Category=number}', "");
Expect(0, 127242, '\P{Category=number}', "");
Expect(1, 127242, '\P{^Category=number}', "");
Expect(0, 127243, '\p{Category=number}', "");
Expect(1, 127243, '\p{^Category=number}', "");
Expect(1, 127243, '\P{Category=number}', "");
Expect(0, 127243, '\P{^Category=number}', "");
Expect(1, 127242, '\p{Category:   _-NUMBER}', "");
Expect(0, 127242, '\p{^Category:   _-NUMBER}', "");
Expect(0, 127242, '\P{Category:   _-NUMBER}', "");
Expect(1, 127242, '\P{^Category:   _-NUMBER}', "");
Expect(0, 127243, '\p{Category:   _-NUMBER}', "");
Expect(1, 127243, '\p{^Category:   _-NUMBER}', "");
Expect(1, 127243, '\P{Category:   _-NUMBER}', "");
Expect(0, 127243, '\P{^Category:   _-NUMBER}', "");
Error('\p{Is_General_Category=	_N:=}');
Error('\P{Is_General_Category=	_N:=}');
Expect(1, 127242, '\p{Is_General_Category=n}', "");
Expect(0, 127242, '\p{^Is_General_Category=n}', "");
Expect(0, 127242, '\P{Is_General_Category=n}', "");
Expect(1, 127242, '\P{^Is_General_Category=n}', "");
Expect(0, 127243, '\p{Is_General_Category=n}', "");
Expect(1, 127243, '\p{^Is_General_Category=n}', "");
Expect(1, 127243, '\P{Is_General_Category=n}', "");
Expect(0, 127243, '\P{^Is_General_Category=n}', "");
Expect(1, 127242, '\p{Is_General_Category:			N}', "");
Expect(0, 127242, '\p{^Is_General_Category:			N}', "");
Expect(0, 127242, '\P{Is_General_Category:			N}', "");
Expect(1, 127242, '\P{^Is_General_Category:			N}', "");
Expect(0, 127243, '\p{Is_General_Category:			N}', "");
Expect(1, 127243, '\p{^Is_General_Category:			N}', "");
Expect(1, 127243, '\P{Is_General_Category:			N}', "");
Expect(0, 127243, '\P{^Is_General_Category:			N}', "");
Error('\p{Is_Gc=:=_-NUMBER}');
Error('\P{Is_Gc=:=_-NUMBER}');
Expect(1, 127242, '\p{Is_Gc=number}', "");
Expect(0, 127242, '\p{^Is_Gc=number}', "");
Expect(0, 127242, '\P{Is_Gc=number}', "");
Expect(1, 127242, '\P{^Is_Gc=number}', "");
Expect(0, 127243, '\p{Is_Gc=number}', "");
Expect(1, 127243, '\p{^Is_Gc=number}', "");
Expect(1, 127243, '\P{Is_Gc=number}', "");
Expect(0, 127243, '\P{^Is_Gc=number}', "");
Expect(1, 127242, '\p{Is_Gc=-number}', "");
Expect(0, 127242, '\p{^Is_Gc=-number}', "");
Expect(0, 127242, '\P{Is_Gc=-number}', "");
Expect(1, 127242, '\P{^Is_Gc=-number}', "");
Expect(0, 127243, '\p{Is_Gc=-number}', "");
Expect(1, 127243, '\p{^Is_Gc=-number}', "");
Expect(1, 127243, '\P{Is_Gc=-number}', "");
Expect(0, 127243, '\P{^Is_Gc=-number}', "");
Error('\p{Is_Category=:=_ N}');
Error('\P{Is_Category=:=_ N}');
Expect(1, 127242, '\p{Is_Category=n}', "");
Expect(0, 127242, '\p{^Is_Category=n}', "");
Expect(0, 127242, '\P{Is_Category=n}', "");
Expect(1, 127242, '\P{^Is_Category=n}', "");
Expect(0, 127243, '\p{Is_Category=n}', "");
Expect(1, 127243, '\p{^Is_Category=n}', "");
Expect(1, 127243, '\P{Is_Category=n}', "");
Expect(0, 127243, '\P{^Is_Category=n}', "");
Expect(1, 127242, '\p{Is_Category=	N}', "");
Expect(0, 127242, '\p{^Is_Category=	N}', "");
Expect(0, 127242, '\P{Is_Category=	N}', "");
Expect(1, 127242, '\P{^Is_Category=	N}', "");
Expect(0, 127243, '\p{Is_Category=	N}', "");
Expect(1, 127243, '\p{^Is_Category=	N}', "");
Expect(1, 127243, '\P{Is_Category=	N}', "");
Expect(0, 127243, '\P{^Is_Category=	N}', "");
Error('\p{General_Category=	_decimal_NUMBER:=}');
Error('\P{General_Category=	_decimal_NUMBER:=}');
Expect(1, 120831, '\p{General_Category=decimalnumber}', "");
Expect(0, 120831, '\p{^General_Category=decimalnumber}', "");
Expect(0, 120831, '\P{General_Category=decimalnumber}', "");
Expect(1, 120831, '\P{^General_Category=decimalnumber}', "");
Expect(0, 120832, '\p{General_Category=decimalnumber}', "");
Expect(1, 120832, '\p{^General_Category=decimalnumber}', "");
Expect(1, 120832, '\P{General_Category=decimalnumber}', "");
Expect(0, 120832, '\P{^General_Category=decimalnumber}', "");
Expect(1, 120831, '\p{General_Category=	-Decimal_NUMBER}', "");
Expect(0, 120831, '\p{^General_Category=	-Decimal_NUMBER}', "");
Expect(0, 120831, '\P{General_Category=	-Decimal_NUMBER}', "");
Expect(1, 120831, '\P{^General_Category=	-Decimal_NUMBER}', "");
Expect(0, 120832, '\p{General_Category=	-Decimal_NUMBER}', "");
Expect(1, 120832, '\p{^General_Category=	-Decimal_NUMBER}', "");
Expect(1, 120832, '\P{General_Category=	-Decimal_NUMBER}', "");
Expect(0, 120832, '\P{^General_Category=	-Decimal_NUMBER}', "");
Error('\p{Gc:		-nd/a/}');
Error('\P{Gc:		-nd/a/}');
Expect(1, 120831, '\p{Gc=nd}', "");
Expect(0, 120831, '\p{^Gc=nd}', "");
Expect(0, 120831, '\P{Gc=nd}', "");
Expect(1, 120831, '\P{^Gc=nd}', "");
Expect(0, 120832, '\p{Gc=nd}', "");
Expect(1, 120832, '\p{^Gc=nd}', "");
Expect(1, 120832, '\P{Gc=nd}', "");
Expect(0, 120832, '\P{^Gc=nd}', "");
Expect(1, 120831, '\p{Gc=Nd}', "");
Expect(0, 120831, '\p{^Gc=Nd}', "");
Expect(0, 120831, '\P{Gc=Nd}', "");
Expect(1, 120831, '\P{^Gc=Nd}', "");
Expect(0, 120832, '\p{Gc=Nd}', "");
Expect(1, 120832, '\p{^Gc=Nd}', "");
Expect(1, 120832, '\P{Gc=Nd}', "");
Expect(0, 120832, '\P{^Gc=Nd}', "");
Error('\p{Category= 	Digit/a/}');
Error('\P{Category= 	Digit/a/}');
Expect(1, 120831, '\p{Category=digit}', "");
Expect(0, 120831, '\p{^Category=digit}', "");
Expect(0, 120831, '\P{Category=digit}', "");
Expect(1, 120831, '\P{^Category=digit}', "");
Expect(0, 120832, '\p{Category=digit}', "");
Expect(1, 120832, '\p{^Category=digit}', "");
Expect(1, 120832, '\P{Category=digit}', "");
Expect(0, 120832, '\P{^Category=digit}', "");
Expect(1, 120831, '\p{Category=_-Digit}', "");
Expect(0, 120831, '\p{^Category=_-Digit}', "");
Expect(0, 120831, '\P{Category=_-Digit}', "");
Expect(1, 120831, '\P{^Category=_-Digit}', "");
Expect(0, 120832, '\p{Category=_-Digit}', "");
Expect(1, 120832, '\p{^Category=_-Digit}', "");
Expect(1, 120832, '\P{Category=_-Digit}', "");
Expect(0, 120832, '\P{^Category=_-Digit}', "");
Error('\p{Is_General_Category=	DECIMAL_Number/a/}');
Error('\P{Is_General_Category=	DECIMAL_Number/a/}');
Expect(1, 120831, '\p{Is_General_Category=decimalnumber}', "");
Expect(0, 120831, '\p{^Is_General_Category=decimalnumber}', "");
Expect(0, 120831, '\P{Is_General_Category=decimalnumber}', "");
Expect(1, 120831, '\P{^Is_General_Category=decimalnumber}', "");
Expect(0, 120832, '\p{Is_General_Category=decimalnumber}', "");
Expect(1, 120832, '\p{^Is_General_Category=decimalnumber}', "");
Expect(1, 120832, '\P{Is_General_Category=decimalnumber}', "");
Expect(0, 120832, '\P{^Is_General_Category=decimalnumber}', "");
Expect(1, 120831, '\p{Is_General_Category:		Decimal_Number}', "");
Expect(0, 120831, '\p{^Is_General_Category:		Decimal_Number}', "");
Expect(0, 120831, '\P{Is_General_Category:		Decimal_Number}', "");
Expect(1, 120831, '\P{^Is_General_Category:		Decimal_Number}', "");
Expect(0, 120832, '\p{Is_General_Category:		Decimal_Number}', "");
Expect(1, 120832, '\p{^Is_General_Category:		Decimal_Number}', "");
Expect(1, 120832, '\P{Is_General_Category:		Decimal_Number}', "");
Expect(0, 120832, '\P{^Is_General_Category:		Decimal_Number}', "");
Error('\p{Is_Gc:   -/a/nd}');
Error('\P{Is_Gc:   -/a/nd}');
Expect(1, 120831, '\p{Is_Gc=nd}', "");
Expect(0, 120831, '\p{^Is_Gc=nd}', "");
Expect(0, 120831, '\P{Is_Gc=nd}', "");
Expect(1, 120831, '\P{^Is_Gc=nd}', "");
Expect(0, 120832, '\p{Is_Gc=nd}', "");
Expect(1, 120832, '\p{^Is_Gc=nd}', "");
Expect(1, 120832, '\P{Is_Gc=nd}', "");
Expect(0, 120832, '\P{^Is_Gc=nd}', "");
Expect(1, 120831, '\p{Is_Gc= -Nd}', "");
Expect(0, 120831, '\p{^Is_Gc= -Nd}', "");
Expect(0, 120831, '\P{Is_Gc= -Nd}', "");
Expect(1, 120831, '\P{^Is_Gc= -Nd}', "");
Expect(0, 120832, '\p{Is_Gc= -Nd}', "");
Expect(1, 120832, '\p{^Is_Gc= -Nd}', "");
Expect(1, 120832, '\P{Is_Gc= -Nd}', "");
Expect(0, 120832, '\P{^Is_Gc= -Nd}', "");
Error('\p{Is_Category=-	digit:=}');
Error('\P{Is_Category=-	digit:=}');
Expect(1, 120831, '\p{Is_Category:digit}', "");
Expect(0, 120831, '\p{^Is_Category:digit}', "");
Expect(0, 120831, '\P{Is_Category:digit}', "");
Expect(1, 120831, '\P{^Is_Category:digit}', "");
Expect(0, 120832, '\p{Is_Category:digit}', "");
Expect(1, 120832, '\p{^Is_Category:digit}', "");
Expect(1, 120832, '\P{Is_Category:digit}', "");
Expect(0, 120832, '\P{^Is_Category:digit}', "");
Expect(1, 120831, '\p{Is_Category=-DIGIT}', "");
Expect(0, 120831, '\p{^Is_Category=-DIGIT}', "");
Expect(0, 120831, '\P{Is_Category=-DIGIT}', "");
Expect(1, 120831, '\P{^Is_Category=-DIGIT}', "");
Expect(0, 120832, '\p{Is_Category=-DIGIT}', "");
Expect(1, 120832, '\p{^Is_Category=-DIGIT}', "");
Expect(1, 120832, '\P{Is_Category=-DIGIT}', "");
Expect(0, 120832, '\P{^Is_Category=-DIGIT}', "");
Error('\p{General_Category=_Letter_NUMBER:=}');
Error('\P{General_Category=_Letter_NUMBER:=}');
Expect(1, 74850, '\p{General_Category=letternumber}', "");
Expect(0, 74850, '\p{^General_Category=letternumber}', "");
Expect(0, 74850, '\P{General_Category=letternumber}', "");
Expect(1, 74850, '\P{^General_Category=letternumber}', "");
Expect(0, 74851, '\p{General_Category=letternumber}', "");
Expect(1, 74851, '\p{^General_Category=letternumber}', "");
Expect(1, 74851, '\P{General_Category=letternumber}', "");
Expect(0, 74851, '\P{^General_Category=letternumber}', "");
Expect(1, 74850, '\p{General_Category= _Letter_number}', "");
Expect(0, 74850, '\p{^General_Category= _Letter_number}', "");
Expect(0, 74850, '\P{General_Category= _Letter_number}', "");
Expect(1, 74850, '\P{^General_Category= _Letter_number}', "");
Expect(0, 74851, '\p{General_Category= _Letter_number}', "");
Expect(1, 74851, '\p{^General_Category= _Letter_number}', "");
Expect(1, 74851, '\P{General_Category= _Letter_number}', "");
Expect(0, 74851, '\P{^General_Category= _Letter_number}', "");
Error('\p{Gc=-:=NL}');
Error('\P{Gc=-:=NL}');
Expect(1, 74850, '\p{Gc=nl}', "");
Expect(0, 74850, '\p{^Gc=nl}', "");
Expect(0, 74850, '\P{Gc=nl}', "");
Expect(1, 74850, '\P{^Gc=nl}', "");
Expect(0, 74851, '\p{Gc=nl}', "");
Expect(1, 74851, '\p{^Gc=nl}', "");
Expect(1, 74851, '\P{Gc=nl}', "");
Expect(0, 74851, '\P{^Gc=nl}', "");
Expect(1, 74850, '\p{Gc=- nl}', "");
Expect(0, 74850, '\p{^Gc=- nl}', "");
Expect(0, 74850, '\P{Gc=- nl}', "");
Expect(1, 74850, '\P{^Gc=- nl}', "");
Expect(0, 74851, '\p{Gc=- nl}', "");
Expect(1, 74851, '\p{^Gc=- nl}', "");
Expect(1, 74851, '\P{Gc=- nl}', "");
Expect(0, 74851, '\P{^Gc=- nl}', "");
Error('\p{Category=:= _LETTER_number}');
Error('\P{Category=:= _LETTER_number}');
Expect(1, 74850, '\p{Category=letternumber}', "");
Expect(0, 74850, '\p{^Category=letternumber}', "");
Expect(0, 74850, '\P{Category=letternumber}', "");
Expect(1, 74850, '\P{^Category=letternumber}', "");
Expect(0, 74851, '\p{Category=letternumber}', "");
Expect(1, 74851, '\p{^Category=letternumber}', "");
Expect(1, 74851, '\P{Category=letternumber}', "");
Expect(0, 74851, '\P{^Category=letternumber}', "");
Expect(1, 74850, '\p{Category=LETTER_Number}', "");
Expect(0, 74850, '\p{^Category=LETTER_Number}', "");
Expect(0, 74850, '\P{Category=LETTER_Number}', "");
Expect(1, 74850, '\P{^Category=LETTER_Number}', "");
Expect(0, 74851, '\p{Category=LETTER_Number}', "");
Expect(1, 74851, '\p{^Category=LETTER_Number}', "");
Expect(1, 74851, '\P{Category=LETTER_Number}', "");
Expect(0, 74851, '\P{^Category=LETTER_Number}', "");
Error('\p{Is_General_Category= :=Nl}');
Error('\P{Is_General_Category= :=Nl}');
Expect(1, 74850, '\p{Is_General_Category=nl}', "");
Expect(0, 74850, '\p{^Is_General_Category=nl}', "");
Expect(0, 74850, '\P{Is_General_Category=nl}', "");
Expect(1, 74850, '\P{^Is_General_Category=nl}', "");
Expect(0, 74851, '\p{Is_General_Category=nl}', "");
Expect(1, 74851, '\p{^Is_General_Category=nl}', "");
Expect(1, 74851, '\P{Is_General_Category=nl}', "");
Expect(0, 74851, '\P{^Is_General_Category=nl}', "");
Expect(1, 74850, '\p{Is_General_Category=_	NL}', "");
Expect(0, 74850, '\p{^Is_General_Category=_	NL}', "");
Expect(0, 74850, '\P{Is_General_Category=_	NL}', "");
Expect(1, 74850, '\P{^Is_General_Category=_	NL}', "");
Expect(0, 74851, '\p{Is_General_Category=_	NL}', "");
Expect(1, 74851, '\p{^Is_General_Category=_	NL}', "");
Expect(1, 74851, '\P{Is_General_Category=_	NL}', "");
Expect(0, 74851, '\P{^Is_General_Category=_	NL}', "");
Error('\p{Is_Gc::=_-letter_Number}');
Error('\P{Is_Gc::=_-letter_Number}');
Expect(1, 74850, '\p{Is_Gc=letternumber}', "");
Expect(0, 74850, '\p{^Is_Gc=letternumber}', "");
Expect(0, 74850, '\P{Is_Gc=letternumber}', "");
Expect(1, 74850, '\P{^Is_Gc=letternumber}', "");
Expect(0, 74851, '\p{Is_Gc=letternumber}', "");
Expect(1, 74851, '\p{^Is_Gc=letternumber}', "");
Expect(1, 74851, '\P{Is_Gc=letternumber}', "");
Expect(0, 74851, '\P{^Is_Gc=letternumber}', "");
Expect(1, 74850, '\p{Is_Gc= 	letter_number}', "");
Expect(0, 74850, '\p{^Is_Gc= 	letter_number}', "");
Expect(0, 74850, '\P{Is_Gc= 	letter_number}', "");
Expect(1, 74850, '\P{^Is_Gc= 	letter_number}', "");
Expect(0, 74851, '\p{Is_Gc= 	letter_number}', "");
Expect(1, 74851, '\p{^Is_Gc= 	letter_number}', "");
Expect(1, 74851, '\P{Is_Gc= 	letter_number}', "");
Expect(0, 74851, '\P{^Is_Gc= 	letter_number}', "");
Error('\p{Is_Category=_:=nl}');
Error('\P{Is_Category=_:=nl}');
Expect(1, 74850, '\p{Is_Category:	nl}', "");
Expect(0, 74850, '\p{^Is_Category:	nl}', "");
Expect(0, 74850, '\P{Is_Category:	nl}', "");
Expect(1, 74850, '\P{^Is_Category:	nl}', "");
Expect(0, 74851, '\p{Is_Category:	nl}', "");
Expect(1, 74851, '\p{^Is_Category:	nl}', "");
Expect(1, 74851, '\P{Is_Category:	nl}', "");
Expect(0, 74851, '\P{^Is_Category:	nl}', "");
Expect(1, 74850, '\p{Is_Category=Nl}', "");
Expect(0, 74850, '\p{^Is_Category=Nl}', "");
Expect(0, 74850, '\P{Is_Category=Nl}', "");
Expect(1, 74850, '\P{^Is_Category=Nl}', "");
Expect(0, 74851, '\p{Is_Category=Nl}', "");
Expect(1, 74851, '\p{^Is_Category=Nl}', "");
Expect(1, 74851, '\P{Is_Category=Nl}', "");
Expect(0, 74851, '\P{^Is_Category=Nl}', "");
Error('\p{General_Category=	:=OTHER_NUMBER}');
Error('\P{General_Category=	:=OTHER_NUMBER}');
Expect(1, 127242, '\p{General_Category=othernumber}', "");
Expect(0, 127242, '\p{^General_Category=othernumber}', "");
Expect(0, 127242, '\P{General_Category=othernumber}', "");
Expect(1, 127242, '\P{^General_Category=othernumber}', "");
Expect(0, 127243, '\p{General_Category=othernumber}', "");
Expect(1, 127243, '\p{^General_Category=othernumber}', "");
Expect(1, 127243, '\P{General_Category=othernumber}', "");
Expect(0, 127243, '\P{^General_Category=othernumber}', "");
Expect(1, 127242, '\p{General_Category=	Other_Number}', "");
Expect(0, 127242, '\p{^General_Category=	Other_Number}', "");
Expect(0, 127242, '\P{General_Category=	Other_Number}', "");
Expect(1, 127242, '\P{^General_Category=	Other_Number}', "");
Expect(0, 127243, '\p{General_Category=	Other_Number}', "");
Expect(1, 127243, '\p{^General_Category=	Other_Number}', "");
Expect(1, 127243, '\P{General_Category=	Other_Number}', "");
Expect(0, 127243, '\P{^General_Category=	Other_Number}', "");
Error('\p{Gc=		No/a/}');
Error('\P{Gc=		No/a/}');
Expect(1, 127242, '\p{Gc=no}', "");
Expect(0, 127242, '\p{^Gc=no}', "");
Expect(0, 127242, '\P{Gc=no}', "");
Expect(1, 127242, '\P{^Gc=no}', "");
Expect(0, 127243, '\p{Gc=no}', "");
Expect(1, 127243, '\p{^Gc=no}', "");
Expect(1, 127243, '\P{Gc=no}', "");
Expect(0, 127243, '\P{^Gc=no}', "");
Expect(1, 127242, '\p{Gc: _No}', "");
Expect(0, 127242, '\p{^Gc: _No}', "");
Expect(0, 127242, '\P{Gc: _No}', "");
Expect(1, 127242, '\P{^Gc: _No}', "");
Expect(0, 127243, '\p{Gc: _No}', "");
Expect(1, 127243, '\p{^Gc: _No}', "");
Expect(1, 127243, '\P{Gc: _No}', "");
Expect(0, 127243, '\P{^Gc: _No}', "");
Error('\p{Category= _OTHER_Number:=}');
Error('\P{Category= _OTHER_Number:=}');
Expect(1, 127242, '\p{Category=othernumber}', "");
Expect(0, 127242, '\p{^Category=othernumber}', "");
Expect(0, 127242, '\P{Category=othernumber}', "");
Expect(1, 127242, '\P{^Category=othernumber}', "");
Expect(0, 127243, '\p{Category=othernumber}', "");
Expect(1, 127243, '\p{^Category=othernumber}', "");
Expect(1, 127243, '\P{Category=othernumber}', "");
Expect(0, 127243, '\P{^Category=othernumber}', "");
Expect(1, 127242, '\p{Category= Other_Number}', "");
Expect(0, 127242, '\p{^Category= Other_Number}', "");
Expect(0, 127242, '\P{Category= Other_Number}', "");
Expect(1, 127242, '\P{^Category= Other_Number}', "");
Expect(0, 127243, '\p{Category= Other_Number}', "");
Expect(1, 127243, '\p{^Category= Other_Number}', "");
Expect(1, 127243, '\P{Category= Other_Number}', "");
Expect(0, 127243, '\P{^Category= Other_Number}', "");
Error('\p{Is_General_Category=/a/_	NO}');
Error('\P{Is_General_Category=/a/_	NO}');
Expect(1, 127242, '\p{Is_General_Category=no}', "");
Expect(0, 127242, '\p{^Is_General_Category=no}', "");
Expect(0, 127242, '\P{Is_General_Category=no}', "");
Expect(1, 127242, '\P{^Is_General_Category=no}', "");
Expect(0, 127243, '\p{Is_General_Category=no}', "");
Expect(1, 127243, '\p{^Is_General_Category=no}', "");
Expect(1, 127243, '\P{Is_General_Category=no}', "");
Expect(0, 127243, '\P{^Is_General_Category=no}', "");
Expect(1, 127242, '\p{Is_General_Category=_No}', "");
Expect(0, 127242, '\p{^Is_General_Category=_No}', "");
Expect(0, 127242, '\P{Is_General_Category=_No}', "");
Expect(1, 127242, '\P{^Is_General_Category=_No}', "");
Expect(0, 127243, '\p{Is_General_Category=_No}', "");
Expect(1, 127243, '\p{^Is_General_Category=_No}', "");
Expect(1, 127243, '\P{Is_General_Category=_No}', "");
Expect(0, 127243, '\P{^Is_General_Category=_No}', "");
Error('\p{Is_Gc:  	other_number/a/}');
Error('\P{Is_Gc:  	other_number/a/}');
Expect(1, 127242, '\p{Is_Gc=othernumber}', "");
Expect(0, 127242, '\p{^Is_Gc=othernumber}', "");
Expect(0, 127242, '\P{Is_Gc=othernumber}', "");
Expect(1, 127242, '\P{^Is_Gc=othernumber}', "");
Expect(0, 127243, '\p{Is_Gc=othernumber}', "");
Expect(1, 127243, '\p{^Is_Gc=othernumber}', "");
Expect(1, 127243, '\P{Is_Gc=othernumber}', "");
Expect(0, 127243, '\P{^Is_Gc=othernumber}', "");
Expect(1, 127242, '\p{Is_Gc=	-OTHER_Number}', "");
Expect(0, 127242, '\p{^Is_Gc=	-OTHER_Number}', "");
Expect(0, 127242, '\P{Is_Gc=	-OTHER_Number}', "");
Expect(1, 127242, '\P{^Is_Gc=	-OTHER_Number}', "");
Expect(0, 127243, '\p{Is_Gc=	-OTHER_Number}', "");
Expect(1, 127243, '\p{^Is_Gc=	-OTHER_Number}', "");
Expect(1, 127243, '\P{Is_Gc=	-OTHER_Number}', "");
Expect(0, 127243, '\P{^Is_Gc=	-OTHER_Number}', "");
Error('\p{Is_Category= _No:=}');
Error('\P{Is_Category= _No:=}');
Expect(1, 127242, '\p{Is_Category=no}', "");
Expect(0, 127242, '\p{^Is_Category=no}', "");
Expect(0, 127242, '\P{Is_Category=no}', "");
Expect(1, 127242, '\P{^Is_Category=no}', "");
Expect(0, 127243, '\p{Is_Category=no}', "");
Expect(1, 127243, '\p{^Is_Category=no}', "");
Expect(1, 127243, '\P{Is_Category=no}', "");
Expect(0, 127243, '\P{^Is_Category=no}', "");
Expect(1, 127242, '\p{Is_Category:--No}', "");
Expect(0, 127242, '\p{^Is_Category:--No}', "");
Expect(0, 127242, '\P{Is_Category:--No}', "");
Expect(1, 127242, '\P{^Is_Category:--No}', "");
Expect(0, 127243, '\p{Is_Category:--No}', "");
Expect(1, 127243, '\p{^Is_Category:--No}', "");
Expect(1, 127243, '\P{Is_Category:--No}', "");
Expect(0, 127243, '\P{^Is_Category:--No}', "");
Error('\p{General_Category=/a/ PUNCTUATION}');
Error('\P{General_Category=/a/ PUNCTUATION}');
Expect(1, 74867, '\p{General_Category:punctuation}', "");
Expect(0, 74867, '\p{^General_Category:punctuation}', "");
Expect(0, 74867, '\P{General_Category:punctuation}', "");
Expect(1, 74867, '\P{^General_Category:punctuation}', "");
Expect(0, 74868, '\p{General_Category:punctuation}', "");
Expect(1, 74868, '\p{^General_Category:punctuation}', "");
Expect(1, 74868, '\P{General_Category:punctuation}', "");
Expect(0, 74868, '\P{^General_Category:punctuation}', "");
Expect(1, 74867, '\p{General_Category= 	PUNCTUATION}', "");
Expect(0, 74867, '\p{^General_Category= 	PUNCTUATION}', "");
Expect(0, 74867, '\P{General_Category= 	PUNCTUATION}', "");
Expect(1, 74867, '\P{^General_Category= 	PUNCTUATION}', "");
Expect(0, 74868, '\p{General_Category= 	PUNCTUATION}', "");
Expect(1, 74868, '\p{^General_Category= 	PUNCTUATION}', "");
Expect(1, 74868, '\P{General_Category= 	PUNCTUATION}', "");
Expect(0, 74868, '\P{^General_Category= 	PUNCTUATION}', "");
Error('\p{Gc=/a/ p}');
Error('\P{Gc=/a/ p}');
Expect(1, 74867, '\p{Gc=p}', "");
Expect(0, 74867, '\p{^Gc=p}', "");
Expect(0, 74867, '\P{Gc=p}', "");
Expect(1, 74867, '\P{^Gc=p}', "");
Expect(0, 74868, '\p{Gc=p}', "");
Expect(1, 74868, '\p{^Gc=p}', "");
Expect(1, 74868, '\P{Gc=p}', "");
Expect(0, 74868, '\P{^Gc=p}', "");
Expect(1, 74867, '\p{Gc=_P}', "");
Expect(0, 74867, '\p{^Gc=_P}', "");
Expect(0, 74867, '\P{Gc=_P}', "");
Expect(1, 74867, '\P{^Gc=_P}', "");
Expect(0, 74868, '\p{Gc=_P}', "");
Expect(1, 74868, '\p{^Gc=_P}', "");
Expect(1, 74868, '\P{Gc=_P}', "");
Expect(0, 74868, '\P{^Gc=_P}', "");
Error('\p{Category=_punct/a/}');
Error('\P{Category=_punct/a/}');
Expect(1, 74867, '\p{Category=punct}', "");
Expect(0, 74867, '\p{^Category=punct}', "");
Expect(0, 74867, '\P{Category=punct}', "");
Expect(1, 74867, '\P{^Category=punct}', "");
Expect(0, 74868, '\p{Category=punct}', "");
Expect(1, 74868, '\p{^Category=punct}', "");
Expect(1, 74868, '\P{Category=punct}', "");
Expect(0, 74868, '\P{^Category=punct}', "");
Expect(1, 74867, '\p{Category=_ PUNCT}', "");
Expect(0, 74867, '\p{^Category=_ PUNCT}', "");
Expect(0, 74867, '\P{Category=_ PUNCT}', "");
Expect(1, 74867, '\P{^Category=_ PUNCT}', "");
Expect(0, 74868, '\p{Category=_ PUNCT}', "");
Expect(1, 74868, '\p{^Category=_ PUNCT}', "");
Expect(1, 74868, '\P{Category=_ PUNCT}', "");
Expect(0, 74868, '\P{^Category=_ PUNCT}', "");
Error('\p{Is_General_Category=  punctuation:=}');
Error('\P{Is_General_Category=  punctuation:=}');
Expect(1, 74867, '\p{Is_General_Category=punctuation}', "");
Expect(0, 74867, '\p{^Is_General_Category=punctuation}', "");
Expect(0, 74867, '\P{Is_General_Category=punctuation}', "");
Expect(1, 74867, '\P{^Is_General_Category=punctuation}', "");
Expect(0, 74868, '\p{Is_General_Category=punctuation}', "");
Expect(1, 74868, '\p{^Is_General_Category=punctuation}', "");
Expect(1, 74868, '\P{Is_General_Category=punctuation}', "");
Expect(0, 74868, '\P{^Is_General_Category=punctuation}', "");
Expect(1, 74867, '\p{Is_General_Category=-_PUNCTUATION}', "");
Expect(0, 74867, '\p{^Is_General_Category=-_PUNCTUATION}', "");
Expect(0, 74867, '\P{Is_General_Category=-_PUNCTUATION}', "");
Expect(1, 74867, '\P{^Is_General_Category=-_PUNCTUATION}', "");
Expect(0, 74868, '\p{Is_General_Category=-_PUNCTUATION}', "");
Expect(1, 74868, '\p{^Is_General_Category=-_PUNCTUATION}', "");
Expect(1, 74868, '\P{Is_General_Category=-_PUNCTUATION}', "");
Expect(0, 74868, '\P{^Is_General_Category=-_PUNCTUATION}', "");
Error('\p{Is_Gc=/a/-	p}');
Error('\P{Is_Gc=/a/-	p}');
Expect(1, 74867, '\p{Is_Gc=p}', "");
Expect(0, 74867, '\p{^Is_Gc=p}', "");
Expect(0, 74867, '\P{Is_Gc=p}', "");
Expect(1, 74867, '\P{^Is_Gc=p}', "");
Expect(0, 74868, '\p{Is_Gc=p}', "");
Expect(1, 74868, '\p{^Is_Gc=p}', "");
Expect(1, 74868, '\P{Is_Gc=p}', "");
Expect(0, 74868, '\P{^Is_Gc=p}', "");
Expect(1, 74867, '\p{Is_Gc=_	P}', "");
Expect(0, 74867, '\p{^Is_Gc=_	P}', "");
Expect(0, 74867, '\P{Is_Gc=_	P}', "");
Expect(1, 74867, '\P{^Is_Gc=_	P}', "");
Expect(0, 74868, '\p{Is_Gc=_	P}', "");
Expect(1, 74868, '\p{^Is_Gc=_	P}', "");
Expect(1, 74868, '\P{Is_Gc=_	P}', "");
Expect(0, 74868, '\P{^Is_Gc=_	P}', "");
Error('\p{Is_Category=/a/ Punct}');
Error('\P{Is_Category=/a/ Punct}');
Expect(1, 74867, '\p{Is_Category=punct}', "");
Expect(0, 74867, '\p{^Is_Category=punct}', "");
Expect(0, 74867, '\P{Is_Category=punct}', "");
Expect(1, 74867, '\P{^Is_Category=punct}', "");
Expect(0, 74868, '\p{Is_Category=punct}', "");
Expect(1, 74868, '\p{^Is_Category=punct}', "");
Expect(1, 74868, '\P{Is_Category=punct}', "");
Expect(0, 74868, '\P{^Is_Category=punct}', "");
Error('\p{General_Category= _Connector_PUNCTUATION:=}');
Error('\P{General_Category= _Connector_PUNCTUATION:=}');
Expect(1, 65343, '\p{General_Category=connectorpunctuation}', "");
Expect(0, 65343, '\p{^General_Category=connectorpunctuation}', "");
Expect(0, 65343, '\P{General_Category=connectorpunctuation}', "");
Expect(1, 65343, '\P{^General_Category=connectorpunctuation}', "");
Expect(0, 65344, '\p{General_Category=connectorpunctuation}', "");
Expect(1, 65344, '\p{^General_Category=connectorpunctuation}', "");
Expect(1, 65344, '\P{General_Category=connectorpunctuation}', "");
Expect(0, 65344, '\P{^General_Category=connectorpunctuation}', "");
Expect(1, 65343, '\p{General_Category=	Connector_PUNCTUATION}', "");
Expect(0, 65343, '\p{^General_Category=	Connector_PUNCTUATION}', "");
Expect(0, 65343, '\P{General_Category=	Connector_PUNCTUATION}', "");
Expect(1, 65343, '\P{^General_Category=	Connector_PUNCTUATION}', "");
Expect(0, 65344, '\p{General_Category=	Connector_PUNCTUATION}', "");
Expect(1, 65344, '\p{^General_Category=	Connector_PUNCTUATION}', "");
Expect(1, 65344, '\P{General_Category=	Connector_PUNCTUATION}', "");
Expect(0, 65344, '\P{^General_Category=	Connector_PUNCTUATION}', "");
Error('\p{Gc= /a/Pc}');
Error('\P{Gc= /a/Pc}');
Expect(1, 65343, '\p{Gc=pc}', "");
Expect(0, 65343, '\p{^Gc=pc}', "");
Expect(0, 65343, '\P{Gc=pc}', "");
Expect(1, 65343, '\P{^Gc=pc}', "");
Expect(0, 65344, '\p{Gc=pc}', "");
Expect(1, 65344, '\p{^Gc=pc}', "");
Expect(1, 65344, '\P{Gc=pc}', "");
Expect(0, 65344, '\P{^Gc=pc}', "");
Expect(1, 65343, '\p{Gc=	Pc}', "");
Expect(0, 65343, '\p{^Gc=	Pc}', "");
Expect(0, 65343, '\P{Gc=	Pc}', "");
Expect(1, 65343, '\P{^Gc=	Pc}', "");
Expect(0, 65344, '\p{Gc=	Pc}', "");
Expect(1, 65344, '\p{^Gc=	Pc}', "");
Expect(1, 65344, '\P{Gc=	Pc}', "");
Expect(0, 65344, '\P{^Gc=	Pc}', "");
Error('\p{Category=/a/ Connector_PUNCTUATION}');
Error('\P{Category=/a/ Connector_PUNCTUATION}');
Expect(1, 65343, '\p{Category:connectorpunctuation}', "");
Expect(0, 65343, '\p{^Category:connectorpunctuation}', "");
Expect(0, 65343, '\P{Category:connectorpunctuation}', "");
Expect(1, 65343, '\P{^Category:connectorpunctuation}', "");
Expect(0, 65344, '\p{Category:connectorpunctuation}', "");
Expect(1, 65344, '\p{^Category:connectorpunctuation}', "");
Expect(1, 65344, '\P{Category:connectorpunctuation}', "");
Expect(0, 65344, '\P{^Category:connectorpunctuation}', "");
Expect(1, 65343, '\p{Category=-_connector_PUNCTUATION}', "");
Expect(0, 65343, '\p{^Category=-_connector_PUNCTUATION}', "");
Expect(0, 65343, '\P{Category=-_connector_PUNCTUATION}', "");
Expect(1, 65343, '\P{^Category=-_connector_PUNCTUATION}', "");
Expect(0, 65344, '\p{Category=-_connector_PUNCTUATION}', "");
Expect(1, 65344, '\p{^Category=-_connector_PUNCTUATION}', "");
Expect(1, 65344, '\P{Category=-_connector_PUNCTUATION}', "");
Expect(0, 65344, '\P{^Category=-_connector_PUNCTUATION}', "");
Error('\p{Is_General_Category=_:=PC}');
Error('\P{Is_General_Category=_:=PC}');
Expect(1, 65343, '\p{Is_General_Category=pc}', "");
Expect(0, 65343, '\p{^Is_General_Category=pc}', "");
Expect(0, 65343, '\P{Is_General_Category=pc}', "");
Expect(1, 65343, '\P{^Is_General_Category=pc}', "");
Expect(0, 65344, '\p{Is_General_Category=pc}', "");
Expect(1, 65344, '\p{^Is_General_Category=pc}', "");
Expect(1, 65344, '\P{Is_General_Category=pc}', "");
Expect(0, 65344, '\P{^Is_General_Category=pc}', "");
Expect(1, 65343, '\p{Is_General_Category=Pc}', "");
Expect(0, 65343, '\p{^Is_General_Category=Pc}', "");
Expect(0, 65343, '\P{Is_General_Category=Pc}', "");
Expect(1, 65343, '\P{^Is_General_Category=Pc}', "");
Expect(0, 65344, '\p{Is_General_Category=Pc}', "");
Expect(1, 65344, '\p{^Is_General_Category=Pc}', "");
Expect(1, 65344, '\P{Is_General_Category=Pc}', "");
Expect(0, 65344, '\P{^Is_General_Category=Pc}', "");
Error('\p{Is_Gc=_:=Connector_Punctuation}');
Error('\P{Is_Gc=_:=Connector_Punctuation}');
Expect(1, 65343, '\p{Is_Gc=connectorpunctuation}', "");
Expect(0, 65343, '\p{^Is_Gc=connectorpunctuation}', "");
Expect(0, 65343, '\P{Is_Gc=connectorpunctuation}', "");
Expect(1, 65343, '\P{^Is_Gc=connectorpunctuation}', "");
Expect(0, 65344, '\p{Is_Gc=connectorpunctuation}', "");
Expect(1, 65344, '\p{^Is_Gc=connectorpunctuation}', "");
Expect(1, 65344, '\P{Is_Gc=connectorpunctuation}', "");
Expect(0, 65344, '\P{^Is_Gc=connectorpunctuation}', "");
Expect(1, 65343, '\p{Is_Gc=_	CONNECTOR_Punctuation}', "");
Expect(0, 65343, '\p{^Is_Gc=_	CONNECTOR_Punctuation}', "");
Expect(0, 65343, '\P{Is_Gc=_	CONNECTOR_Punctuation}', "");
Expect(1, 65343, '\P{^Is_Gc=_	CONNECTOR_Punctuation}', "");
Expect(0, 65344, '\p{Is_Gc=_	CONNECTOR_Punctuation}', "");
Expect(1, 65344, '\p{^Is_Gc=_	CONNECTOR_Punctuation}', "");
Expect(1, 65344, '\P{Is_Gc=_	CONNECTOR_Punctuation}', "");
Expect(0, 65344, '\P{^Is_Gc=_	CONNECTOR_Punctuation}', "");
Error('\p{Is_Category: :=- Pc}');
Error('\P{Is_Category: :=- Pc}');
Expect(1, 65343, '\p{Is_Category=pc}', "");
Expect(0, 65343, '\p{^Is_Category=pc}', "");
Expect(0, 65343, '\P{Is_Category=pc}', "");
Expect(1, 65343, '\P{^Is_Category=pc}', "");
Expect(0, 65344, '\p{Is_Category=pc}', "");
Expect(1, 65344, '\p{^Is_Category=pc}', "");
Expect(1, 65344, '\P{Is_Category=pc}', "");
Expect(0, 65344, '\P{^Is_Category=pc}', "");
Expect(1, 65343, '\p{Is_Category= Pc}', "");
Expect(0, 65343, '\p{^Is_Category= Pc}', "");
Expect(0, 65343, '\P{Is_Category= Pc}', "");
Expect(1, 65343, '\P{^Is_Category= Pc}', "");
Expect(0, 65344, '\p{Is_Category= Pc}', "");
Expect(1, 65344, '\p{^Is_Category= Pc}', "");
Expect(1, 65344, '\P{Is_Category= Pc}', "");
Expect(0, 65344, '\P{^Is_Category= Pc}', "");
Error('\p{General_Category=	-dash_Punctuation:=}');
Error('\P{General_Category=	-dash_Punctuation:=}');
Expect(1, 65293, '\p{General_Category=dashpunctuation}', "");
Expect(0, 65293, '\p{^General_Category=dashpunctuation}', "");
Expect(0, 65293, '\P{General_Category=dashpunctuation}', "");
Expect(1, 65293, '\P{^General_Category=dashpunctuation}', "");
Expect(0, 65294, '\p{General_Category=dashpunctuation}', "");
Expect(1, 65294, '\p{^General_Category=dashpunctuation}', "");
Expect(1, 65294, '\P{General_Category=dashpunctuation}', "");
Expect(0, 65294, '\P{^General_Category=dashpunctuation}', "");
Expect(1, 65293, '\p{General_Category=_	Dash_PUNCTUATION}', "");
Expect(0, 65293, '\p{^General_Category=_	Dash_PUNCTUATION}', "");
Expect(0, 65293, '\P{General_Category=_	Dash_PUNCTUATION}', "");
Expect(1, 65293, '\P{^General_Category=_	Dash_PUNCTUATION}', "");
Expect(0, 65294, '\p{General_Category=_	Dash_PUNCTUATION}', "");
Expect(1, 65294, '\p{^General_Category=_	Dash_PUNCTUATION}', "");
Expect(1, 65294, '\P{General_Category=_	Dash_PUNCTUATION}', "");
Expect(0, 65294, '\P{^General_Category=_	Dash_PUNCTUATION}', "");
Error('\p{Gc=/a/--PD}');
Error('\P{Gc=/a/--PD}');
Expect(1, 65293, '\p{Gc=pd}', "");
Expect(0, 65293, '\p{^Gc=pd}', "");
Expect(0, 65293, '\P{Gc=pd}', "");
Expect(1, 65293, '\P{^Gc=pd}', "");
Expect(0, 65294, '\p{Gc=pd}', "");
Expect(1, 65294, '\p{^Gc=pd}', "");
Expect(1, 65294, '\P{Gc=pd}', "");
Expect(0, 65294, '\P{^Gc=pd}', "");
Expect(1, 65293, '\p{Gc= -Pd}', "");
Expect(0, 65293, '\p{^Gc= -Pd}', "");
Expect(0, 65293, '\P{Gc= -Pd}', "");
Expect(1, 65293, '\P{^Gc= -Pd}', "");
Expect(0, 65294, '\p{Gc= -Pd}', "");
Expect(1, 65294, '\p{^Gc= -Pd}', "");
Expect(1, 65294, '\P{Gc= -Pd}', "");
Expect(0, 65294, '\P{^Gc= -Pd}', "");
Error('\p{Category=	-DASH_punctuation/a/}');
Error('\P{Category=	-DASH_punctuation/a/}');
Expect(1, 65293, '\p{Category=dashpunctuation}', "");
Expect(0, 65293, '\p{^Category=dashpunctuation}', "");
Expect(0, 65293, '\P{Category=dashpunctuation}', "");
Expect(1, 65293, '\P{^Category=dashpunctuation}', "");
Expect(0, 65294, '\p{Category=dashpunctuation}', "");
Expect(1, 65294, '\p{^Category=dashpunctuation}', "");
Expect(1, 65294, '\P{Category=dashpunctuation}', "");
Expect(0, 65294, '\P{^Category=dashpunctuation}', "");
Expect(1, 65293, '\p{Category=  Dash_Punctuation}', "");
Expect(0, 65293, '\p{^Category=  Dash_Punctuation}', "");
Expect(0, 65293, '\P{Category=  Dash_Punctuation}', "");
Expect(1, 65293, '\P{^Category=  Dash_Punctuation}', "");
Expect(0, 65294, '\p{Category=  Dash_Punctuation}', "");
Expect(1, 65294, '\p{^Category=  Dash_Punctuation}', "");
Expect(1, 65294, '\P{Category=  Dash_Punctuation}', "");
Expect(0, 65294, '\P{^Category=  Dash_Punctuation}', "");
Error('\p{Is_General_Category=:=__Pd}');
Error('\P{Is_General_Category=:=__Pd}');
Expect(1, 65293, '\p{Is_General_Category=pd}', "");
Expect(0, 65293, '\p{^Is_General_Category=pd}', "");
Expect(0, 65293, '\P{Is_General_Category=pd}', "");
Expect(1, 65293, '\P{^Is_General_Category=pd}', "");
Expect(0, 65294, '\p{Is_General_Category=pd}', "");
Expect(1, 65294, '\p{^Is_General_Category=pd}', "");
Expect(1, 65294, '\P{Is_General_Category=pd}', "");
Expect(0, 65294, '\P{^Is_General_Category=pd}', "");
Expect(1, 65293, '\p{Is_General_Category= -pd}', "");
Expect(0, 65293, '\p{^Is_General_Category= -pd}', "");
Expect(0, 65293, '\P{Is_General_Category= -pd}', "");
Expect(1, 65293, '\P{^Is_General_Category= -pd}', "");
Expect(0, 65294, '\p{Is_General_Category= -pd}', "");
Expect(1, 65294, '\p{^Is_General_Category= -pd}', "");
Expect(1, 65294, '\P{Is_General_Category= -pd}', "");
Expect(0, 65294, '\P{^Is_General_Category= -pd}', "");
Error('\p{Is_Gc=_ Dash_PUNCTUATION:=}');
Error('\P{Is_Gc=_ Dash_PUNCTUATION:=}');
Expect(1, 65293, '\p{Is_Gc=dashpunctuation}', "");
Expect(0, 65293, '\p{^Is_Gc=dashpunctuation}', "");
Expect(0, 65293, '\P{Is_Gc=dashpunctuation}', "");
Expect(1, 65293, '\P{^Is_Gc=dashpunctuation}', "");
Expect(0, 65294, '\p{Is_Gc=dashpunctuation}', "");
Expect(1, 65294, '\p{^Is_Gc=dashpunctuation}', "");
Expect(1, 65294, '\P{Is_Gc=dashpunctuation}', "");
Expect(0, 65294, '\P{^Is_Gc=dashpunctuation}', "");
Expect(1, 65293, '\p{Is_Gc: -	Dash_punctuation}', "");
Expect(0, 65293, '\p{^Is_Gc: -	Dash_punctuation}', "");
Expect(0, 65293, '\P{Is_Gc: -	Dash_punctuation}', "");
Expect(1, 65293, '\P{^Is_Gc: -	Dash_punctuation}', "");
Expect(0, 65294, '\p{Is_Gc: -	Dash_punctuation}', "");
Expect(1, 65294, '\p{^Is_Gc: -	Dash_punctuation}', "");
Expect(1, 65294, '\P{Is_Gc: -	Dash_punctuation}', "");
Expect(0, 65294, '\P{^Is_Gc: -	Dash_punctuation}', "");
Error('\p{Is_Category=/a/--Pd}');
Error('\P{Is_Category=/a/--Pd}');
Expect(1, 65293, '\p{Is_Category=pd}', "");
Expect(0, 65293, '\p{^Is_Category=pd}', "");
Expect(0, 65293, '\P{Is_Category=pd}', "");
Expect(1, 65293, '\P{^Is_Category=pd}', "");
Expect(0, 65294, '\p{Is_Category=pd}', "");
Expect(1, 65294, '\p{^Is_Category=pd}', "");
Expect(1, 65294, '\P{Is_Category=pd}', "");
Expect(0, 65294, '\P{^Is_Category=pd}', "");
Expect(1, 65293, '\p{Is_Category: -pd}', "");
Expect(0, 65293, '\p{^Is_Category: -pd}', "");
Expect(0, 65293, '\P{Is_Category: -pd}', "");
Expect(1, 65293, '\P{^Is_Category: -pd}', "");
Expect(0, 65294, '\p{Is_Category: -pd}', "");
Expect(1, 65294, '\p{^Is_Category: -pd}', "");
Expect(1, 65294, '\P{Is_Category: -pd}', "");
Expect(0, 65294, '\P{^Is_Category: -pd}', "");
Error('\p{General_Category=:=_Close_Punctuation}');
Error('\P{General_Category=:=_Close_Punctuation}');
Expect(1, 65379, '\p{General_Category:	closepunctuation}', "");
Expect(0, 65379, '\p{^General_Category:	closepunctuation}', "");
Expect(0, 65379, '\P{General_Category:	closepunctuation}', "");
Expect(1, 65379, '\P{^General_Category:	closepunctuation}', "");
Expect(0, 65380, '\p{General_Category:	closepunctuation}', "");
Expect(1, 65380, '\p{^General_Category:	closepunctuation}', "");
Expect(1, 65380, '\P{General_Category:	closepunctuation}', "");
Expect(0, 65380, '\P{^General_Category:	closepunctuation}', "");
Expect(1, 65379, '\p{General_Category=_-Close_PUNCTUATION}', "");
Expect(0, 65379, '\p{^General_Category=_-Close_PUNCTUATION}', "");
Expect(0, 65379, '\P{General_Category=_-Close_PUNCTUATION}', "");
Expect(1, 65379, '\P{^General_Category=_-Close_PUNCTUATION}', "");
Expect(0, 65380, '\p{General_Category=_-Close_PUNCTUATION}', "");
Expect(1, 65380, '\p{^General_Category=_-Close_PUNCTUATION}', "");
Expect(1, 65380, '\P{General_Category=_-Close_PUNCTUATION}', "");
Expect(0, 65380, '\P{^General_Category=_-Close_PUNCTUATION}', "");
Error('\p{Gc=-:=PE}');
Error('\P{Gc=-:=PE}');
Expect(1, 65379, '\p{Gc=pe}', "");
Expect(0, 65379, '\p{^Gc=pe}', "");
Expect(0, 65379, '\P{Gc=pe}', "");
Expect(1, 65379, '\P{^Gc=pe}', "");
Expect(0, 65380, '\p{Gc=pe}', "");
Expect(1, 65380, '\p{^Gc=pe}', "");
Expect(1, 65380, '\P{Gc=pe}', "");
Expect(0, 65380, '\P{^Gc=pe}', "");
Expect(1, 65379, '\p{Gc=Pe}', "");
Expect(0, 65379, '\p{^Gc=Pe}', "");
Expect(0, 65379, '\P{Gc=Pe}', "");
Expect(1, 65379, '\P{^Gc=Pe}', "");
Expect(0, 65380, '\p{Gc=Pe}', "");
Expect(1, 65380, '\p{^Gc=Pe}', "");
Expect(1, 65380, '\P{Gc=Pe}', "");
Expect(0, 65380, '\P{^Gc=Pe}', "");
Error('\p{Category=/a/_close_Punctuation}');
Error('\P{Category=/a/_close_Punctuation}');
Expect(1, 65379, '\p{Category=closepunctuation}', "");
Expect(0, 65379, '\p{^Category=closepunctuation}', "");
Expect(0, 65379, '\P{Category=closepunctuation}', "");
Expect(1, 65379, '\P{^Category=closepunctuation}', "");
Expect(0, 65380, '\p{Category=closepunctuation}', "");
Expect(1, 65380, '\p{^Category=closepunctuation}', "");
Expect(1, 65380, '\P{Category=closepunctuation}', "");
Expect(0, 65380, '\P{^Category=closepunctuation}', "");
Expect(1, 65379, '\p{Category=close_punctuation}', "");
Expect(0, 65379, '\p{^Category=close_punctuation}', "");
Expect(0, 65379, '\P{Category=close_punctuation}', "");
Expect(1, 65379, '\P{^Category=close_punctuation}', "");
Expect(0, 65380, '\p{Category=close_punctuation}', "");
Expect(1, 65380, '\p{^Category=close_punctuation}', "");
Expect(1, 65380, '\P{Category=close_punctuation}', "");
Expect(0, 65380, '\P{^Category=close_punctuation}', "");
Error('\p{Is_General_Category=/a/-PE}');
Error('\P{Is_General_Category=/a/-PE}');
Expect(1, 65379, '\p{Is_General_Category=pe}', "");
Expect(0, 65379, '\p{^Is_General_Category=pe}', "");
Expect(0, 65379, '\P{Is_General_Category=pe}', "");
Expect(1, 65379, '\P{^Is_General_Category=pe}', "");
Expect(0, 65380, '\p{Is_General_Category=pe}', "");
Expect(1, 65380, '\p{^Is_General_Category=pe}', "");
Expect(1, 65380, '\P{Is_General_Category=pe}', "");
Expect(0, 65380, '\P{^Is_General_Category=pe}', "");
Expect(1, 65379, '\p{Is_General_Category= Pe}', "");
Expect(0, 65379, '\p{^Is_General_Category= Pe}', "");
Expect(0, 65379, '\P{Is_General_Category= Pe}', "");
Expect(1, 65379, '\P{^Is_General_Category= Pe}', "");
Expect(0, 65380, '\p{Is_General_Category= Pe}', "");
Expect(1, 65380, '\p{^Is_General_Category= Pe}', "");
Expect(1, 65380, '\P{Is_General_Category= Pe}', "");
Expect(0, 65380, '\P{^Is_General_Category= Pe}', "");
Error('\p{Is_Gc= CLOSE_Punctuation:=}');
Error('\P{Is_Gc= CLOSE_Punctuation:=}');
Expect(1, 65379, '\p{Is_Gc:   closepunctuation}', "");
Expect(0, 65379, '\p{^Is_Gc:   closepunctuation}', "");
Expect(0, 65379, '\P{Is_Gc:   closepunctuation}', "");
Expect(1, 65379, '\P{^Is_Gc:   closepunctuation}', "");
Expect(0, 65380, '\p{Is_Gc:   closepunctuation}', "");
Expect(1, 65380, '\p{^Is_Gc:   closepunctuation}', "");
Expect(1, 65380, '\P{Is_Gc:   closepunctuation}', "");
Expect(0, 65380, '\P{^Is_Gc:   closepunctuation}', "");
Expect(1, 65379, '\p{Is_Gc=	 Close_Punctuation}', "");
Expect(0, 65379, '\p{^Is_Gc=	 Close_Punctuation}', "");
Expect(0, 65379, '\P{Is_Gc=	 Close_Punctuation}', "");
Expect(1, 65379, '\P{^Is_Gc=	 Close_Punctuation}', "");
Expect(0, 65380, '\p{Is_Gc=	 Close_Punctuation}', "");
Expect(1, 65380, '\p{^Is_Gc=	 Close_Punctuation}', "");
Expect(1, 65380, '\P{Is_Gc=	 Close_Punctuation}', "");
Expect(0, 65380, '\P{^Is_Gc=	 Close_Punctuation}', "");
Error('\p{Is_Category=-pe/a/}');
Error('\P{Is_Category=-pe/a/}');
Expect(1, 65379, '\p{Is_Category=pe}', "");
Expect(0, 65379, '\p{^Is_Category=pe}', "");
Expect(0, 65379, '\P{Is_Category=pe}', "");
Expect(1, 65379, '\P{^Is_Category=pe}', "");
Expect(0, 65380, '\p{Is_Category=pe}', "");
Expect(1, 65380, '\p{^Is_Category=pe}', "");
Expect(1, 65380, '\P{Is_Category=pe}', "");
Expect(0, 65380, '\P{^Is_Category=pe}', "");
Expect(1, 65379, '\p{Is_Category=-	Pe}', "");
Expect(0, 65379, '\p{^Is_Category=-	Pe}', "");
Expect(0, 65379, '\P{Is_Category=-	Pe}', "");
Expect(1, 65379, '\P{^Is_Category=-	Pe}', "");
Expect(0, 65380, '\p{Is_Category=-	Pe}', "");
Expect(1, 65380, '\p{^Is_Category=-	Pe}', "");
Expect(1, 65380, '\P{Is_Category=-	Pe}', "");
Expect(0, 65380, '\P{^Is_Category=-	Pe}', "");
Error('\p{General_Category=-:=Final_PUNCTUATION}');
Error('\P{General_Category=-:=Final_PUNCTUATION}');
Expect(1, 11809, '\p{General_Category=finalpunctuation}', "");
Expect(0, 11809, '\p{^General_Category=finalpunctuation}', "");
Expect(0, 11809, '\P{General_Category=finalpunctuation}', "");
Expect(1, 11809, '\P{^General_Category=finalpunctuation}', "");
Expect(0, 11810, '\p{General_Category=finalpunctuation}', "");
Expect(1, 11810, '\p{^General_Category=finalpunctuation}', "");
Expect(1, 11810, '\P{General_Category=finalpunctuation}', "");
Expect(0, 11810, '\P{^General_Category=finalpunctuation}', "");
Expect(1, 11809, '\p{General_Category:		-FINAL_punctuation}', "");
Expect(0, 11809, '\p{^General_Category:		-FINAL_punctuation}', "");
Expect(0, 11809, '\P{General_Category:		-FINAL_punctuation}', "");
Expect(1, 11809, '\P{^General_Category:		-FINAL_punctuation}', "");
Expect(0, 11810, '\p{General_Category:		-FINAL_punctuation}', "");
Expect(1, 11810, '\p{^General_Category:		-FINAL_punctuation}', "");
Expect(1, 11810, '\P{General_Category:		-FINAL_punctuation}', "");
Expect(0, 11810, '\P{^General_Category:		-FINAL_punctuation}', "");
Error('\p{Gc=_/a/pf}');
Error('\P{Gc=_/a/pf}');
Expect(1, 11809, '\p{Gc=pf}', "");
Expect(0, 11809, '\p{^Gc=pf}', "");
Expect(0, 11809, '\P{Gc=pf}', "");
Expect(1, 11809, '\P{^Gc=pf}', "");
Expect(0, 11810, '\p{Gc=pf}', "");
Expect(1, 11810, '\p{^Gc=pf}', "");
Expect(1, 11810, '\P{Gc=pf}', "");
Expect(0, 11810, '\P{^Gc=pf}', "");
Expect(1, 11809, '\p{Gc=__Pf}', "");
Expect(0, 11809, '\p{^Gc=__Pf}', "");
Expect(0, 11809, '\P{Gc=__Pf}', "");
Expect(1, 11809, '\P{^Gc=__Pf}', "");
Expect(0, 11810, '\p{Gc=__Pf}', "");
Expect(1, 11810, '\p{^Gc=__Pf}', "");
Expect(1, 11810, '\P{Gc=__Pf}', "");
Expect(0, 11810, '\P{^Gc=__Pf}', "");
Error('\p{Category=_ final_Punctuation:=}');
Error('\P{Category=_ final_Punctuation:=}');
Expect(1, 11809, '\p{Category=finalpunctuation}', "");
Expect(0, 11809, '\p{^Category=finalpunctuation}', "");
Expect(0, 11809, '\P{Category=finalpunctuation}', "");
Expect(1, 11809, '\P{^Category=finalpunctuation}', "");
Expect(0, 11810, '\p{Category=finalpunctuation}', "");
Expect(1, 11810, '\p{^Category=finalpunctuation}', "");
Expect(1, 11810, '\P{Category=finalpunctuation}', "");
Expect(0, 11810, '\P{^Category=finalpunctuation}', "");
Expect(1, 11809, '\p{Category:-_Final_punctuation}', "");
Expect(0, 11809, '\p{^Category:-_Final_punctuation}', "");
Expect(0, 11809, '\P{Category:-_Final_punctuation}', "");
Expect(1, 11809, '\P{^Category:-_Final_punctuation}', "");
Expect(0, 11810, '\p{Category:-_Final_punctuation}', "");
Expect(1, 11810, '\p{^Category:-_Final_punctuation}', "");
Expect(1, 11810, '\P{Category:-_Final_punctuation}', "");
Expect(0, 11810, '\P{^Category:-_Final_punctuation}', "");
Error('\p{Is_General_Category=--PF/a/}');
Error('\P{Is_General_Category=--PF/a/}');
Expect(1, 11809, '\p{Is_General_Category=pf}', "");
Expect(0, 11809, '\p{^Is_General_Category=pf}', "");
Expect(0, 11809, '\P{Is_General_Category=pf}', "");
Expect(1, 11809, '\P{^Is_General_Category=pf}', "");
Expect(0, 11810, '\p{Is_General_Category=pf}', "");
Expect(1, 11810, '\p{^Is_General_Category=pf}', "");
Expect(1, 11810, '\P{Is_General_Category=pf}', "");
Expect(0, 11810, '\P{^Is_General_Category=pf}', "");
Expect(1, 11809, '\p{Is_General_Category=_Pf}', "");
Expect(0, 11809, '\p{^Is_General_Category=_Pf}', "");
Expect(0, 11809, '\P{Is_General_Category=_Pf}', "");
Expect(1, 11809, '\P{^Is_General_Category=_Pf}', "");
Expect(0, 11810, '\p{Is_General_Category=_Pf}', "");
Expect(1, 11810, '\p{^Is_General_Category=_Pf}', "");
Expect(1, 11810, '\P{Is_General_Category=_Pf}', "");
Expect(0, 11810, '\P{^Is_General_Category=_Pf}', "");
Error('\p{Is_Gc:  Final_punctuation:=}');
Error('\P{Is_Gc:  Final_punctuation:=}');
Expect(1, 11809, '\p{Is_Gc=finalpunctuation}', "");
Expect(0, 11809, '\p{^Is_Gc=finalpunctuation}', "");
Expect(0, 11809, '\P{Is_Gc=finalpunctuation}', "");
Expect(1, 11809, '\P{^Is_Gc=finalpunctuation}', "");
Expect(0, 11810, '\p{Is_Gc=finalpunctuation}', "");
Expect(1, 11810, '\p{^Is_Gc=finalpunctuation}', "");
Expect(1, 11810, '\P{Is_Gc=finalpunctuation}', "");
Expect(0, 11810, '\P{^Is_Gc=finalpunctuation}', "");
Expect(1, 11809, '\p{Is_Gc= Final_Punctuation}', "");
Expect(0, 11809, '\p{^Is_Gc= Final_Punctuation}', "");
Expect(0, 11809, '\P{Is_Gc= Final_Punctuation}', "");
Expect(1, 11809, '\P{^Is_Gc= Final_Punctuation}', "");
Expect(0, 11810, '\p{Is_Gc= Final_Punctuation}', "");
Expect(1, 11810, '\p{^Is_Gc= Final_Punctuation}', "");
Expect(1, 11810, '\P{Is_Gc= Final_Punctuation}', "");
Expect(0, 11810, '\P{^Is_Gc= Final_Punctuation}', "");
Error('\p{Is_Category:  Pf/a/}');
Error('\P{Is_Category:  Pf/a/}');
Expect(1, 11809, '\p{Is_Category=pf}', "");
Expect(0, 11809, '\p{^Is_Category=pf}', "");
Expect(0, 11809, '\P{Is_Category=pf}', "");
Expect(1, 11809, '\P{^Is_Category=pf}', "");
Expect(0, 11810, '\p{Is_Category=pf}', "");
Expect(1, 11810, '\p{^Is_Category=pf}', "");
Expect(1, 11810, '\P{Is_Category=pf}', "");
Expect(0, 11810, '\P{^Is_Category=pf}', "");
Expect(1, 11809, '\p{Is_Category=	Pf}', "");
Expect(0, 11809, '\p{^Is_Category=	Pf}', "");
Expect(0, 11809, '\P{Is_Category=	Pf}', "");
Expect(1, 11809, '\P{^Is_Category=	Pf}', "");
Expect(0, 11810, '\p{Is_Category=	Pf}', "");
Expect(1, 11810, '\p{^Is_Category=	Pf}', "");
Expect(1, 11810, '\P{Is_Category=	Pf}', "");
Expect(0, 11810, '\P{^Is_Category=	Pf}', "");
Error('\p{General_Category=:=-_initial_punctuation}');
Error('\P{General_Category=:=-_initial_punctuation}');
Expect(1, 11808, '\p{General_Category=initialpunctuation}', "");
Expect(0, 11808, '\p{^General_Category=initialpunctuation}', "");
Expect(0, 11808, '\P{General_Category=initialpunctuation}', "");
Expect(1, 11808, '\P{^General_Category=initialpunctuation}', "");
Expect(0, 11809, '\p{General_Category=initialpunctuation}', "");
Expect(1, 11809, '\p{^General_Category=initialpunctuation}', "");
Expect(1, 11809, '\P{General_Category=initialpunctuation}', "");
Expect(0, 11809, '\P{^General_Category=initialpunctuation}', "");
Expect(1, 11808, '\p{General_Category= INITIAL_Punctuation}', "");
Expect(0, 11808, '\p{^General_Category= INITIAL_Punctuation}', "");
Expect(0, 11808, '\P{General_Category= INITIAL_Punctuation}', "");
Expect(1, 11808, '\P{^General_Category= INITIAL_Punctuation}', "");
Expect(0, 11809, '\p{General_Category= INITIAL_Punctuation}', "");
Expect(1, 11809, '\p{^General_Category= INITIAL_Punctuation}', "");
Expect(1, 11809, '\P{General_Category= INITIAL_Punctuation}', "");
Expect(0, 11809, '\P{^General_Category= INITIAL_Punctuation}', "");
Error('\p{Gc= -PI/a/}');
Error('\P{Gc= -PI/a/}');
Expect(1, 11808, '\p{Gc=pi}', "");
Expect(0, 11808, '\p{^Gc=pi}', "");
Expect(0, 11808, '\P{Gc=pi}', "");
Expect(1, 11808, '\P{^Gc=pi}', "");
Expect(0, 11809, '\p{Gc=pi}', "");
Expect(1, 11809, '\p{^Gc=pi}', "");
Expect(1, 11809, '\P{Gc=pi}', "");
Expect(0, 11809, '\P{^Gc=pi}', "");
Expect(1, 11808, '\p{Gc=__Pi}', "");
Expect(0, 11808, '\p{^Gc=__Pi}', "");
Expect(0, 11808, '\P{Gc=__Pi}', "");
Expect(1, 11808, '\P{^Gc=__Pi}', "");
Expect(0, 11809, '\p{Gc=__Pi}', "");
Expect(1, 11809, '\p{^Gc=__Pi}', "");
Expect(1, 11809, '\P{Gc=__Pi}', "");
Expect(0, 11809, '\P{^Gc=__Pi}', "");
Error('\p{Category=:=- Initial_Punctuation}');
Error('\P{Category=:=- Initial_Punctuation}');
Expect(1, 11808, '\p{Category=initialpunctuation}', "");
Expect(0, 11808, '\p{^Category=initialpunctuation}', "");
Expect(0, 11808, '\P{Category=initialpunctuation}', "");
Expect(1, 11808, '\P{^Category=initialpunctuation}', "");
Expect(0, 11809, '\p{Category=initialpunctuation}', "");
Expect(1, 11809, '\p{^Category=initialpunctuation}', "");
Expect(1, 11809, '\P{Category=initialpunctuation}', "");
Expect(0, 11809, '\P{^Category=initialpunctuation}', "");
Expect(1, 11808, '\p{Category=_Initial_Punctuation}', "");
Expect(0, 11808, '\p{^Category=_Initial_Punctuation}', "");
Expect(0, 11808, '\P{Category=_Initial_Punctuation}', "");
Expect(1, 11808, '\P{^Category=_Initial_Punctuation}', "");
Expect(0, 11809, '\p{Category=_Initial_Punctuation}', "");
Expect(1, 11809, '\p{^Category=_Initial_Punctuation}', "");
Expect(1, 11809, '\P{Category=_Initial_Punctuation}', "");
Expect(0, 11809, '\P{^Category=_Initial_Punctuation}', "");
Error('\p{Is_General_Category= /a/PI}');
Error('\P{Is_General_Category= /a/PI}');
Expect(1, 11808, '\p{Is_General_Category=pi}', "");
Expect(0, 11808, '\p{^Is_General_Category=pi}', "");
Expect(0, 11808, '\P{Is_General_Category=pi}', "");
Expect(1, 11808, '\P{^Is_General_Category=pi}', "");
Expect(0, 11809, '\p{Is_General_Category=pi}', "");
Expect(1, 11809, '\p{^Is_General_Category=pi}', "");
Expect(1, 11809, '\P{Is_General_Category=pi}', "");
Expect(0, 11809, '\P{^Is_General_Category=pi}', "");
Expect(1, 11808, '\p{Is_General_Category=Pi}', "");
Expect(0, 11808, '\p{^Is_General_Category=Pi}', "");
Expect(0, 11808, '\P{Is_General_Category=Pi}', "");
Expect(1, 11808, '\P{^Is_General_Category=Pi}', "");
Expect(0, 11809, '\p{Is_General_Category=Pi}', "");
Expect(1, 11809, '\p{^Is_General_Category=Pi}', "");
Expect(1, 11809, '\P{Is_General_Category=Pi}', "");
Expect(0, 11809, '\P{^Is_General_Category=Pi}', "");
Error('\p{Is_Gc=	/a/Initial_Punctuation}');
Error('\P{Is_Gc=	/a/Initial_Punctuation}');
Expect(1, 11808, '\p{Is_Gc=initialpunctuation}', "");
Expect(0, 11808, '\p{^Is_Gc=initialpunctuation}', "");
Expect(0, 11808, '\P{Is_Gc=initialpunctuation}', "");
Expect(1, 11808, '\P{^Is_Gc=initialpunctuation}', "");
Expect(0, 11809, '\p{Is_Gc=initialpunctuation}', "");
Expect(1, 11809, '\p{^Is_Gc=initialpunctuation}', "");
Expect(1, 11809, '\P{Is_Gc=initialpunctuation}', "");
Expect(0, 11809, '\P{^Is_Gc=initialpunctuation}', "");
Expect(1, 11808, '\p{Is_Gc=_	initial_punctuation}', "");
Expect(0, 11808, '\p{^Is_Gc=_	initial_punctuation}', "");
Expect(0, 11808, '\P{Is_Gc=_	initial_punctuation}', "");
Expect(1, 11808, '\P{^Is_Gc=_	initial_punctuation}', "");
Expect(0, 11809, '\p{Is_Gc=_	initial_punctuation}', "");
Expect(1, 11809, '\p{^Is_Gc=_	initial_punctuation}', "");
Expect(1, 11809, '\P{Is_Gc=_	initial_punctuation}', "");
Expect(0, 11809, '\P{^Is_Gc=_	initial_punctuation}', "");
Error('\p{Is_Category=_:=Pi}');
Error('\P{Is_Category=_:=Pi}');
Expect(1, 11808, '\p{Is_Category:   pi}', "");
Expect(0, 11808, '\p{^Is_Category:   pi}', "");
Expect(0, 11808, '\P{Is_Category:   pi}', "");
Expect(1, 11808, '\P{^Is_Category:   pi}', "");
Expect(0, 11809, '\p{Is_Category:   pi}', "");
Expect(1, 11809, '\p{^Is_Category:   pi}', "");
Expect(1, 11809, '\P{Is_Category:   pi}', "");
Expect(0, 11809, '\P{^Is_Category:   pi}', "");
Expect(1, 11808, '\p{Is_Category=	Pi}', "");
Expect(0, 11808, '\p{^Is_Category=	Pi}', "");
Expect(0, 11808, '\P{Is_Category=	Pi}', "");
Expect(1, 11808, '\P{^Is_Category=	Pi}', "");
Expect(0, 11809, '\p{Is_Category=	Pi}', "");
Expect(1, 11809, '\p{^Is_Category=	Pi}', "");
Expect(1, 11809, '\P{Is_Category=	Pi}', "");
Expect(0, 11809, '\P{^Is_Category=	Pi}', "");
Error('\p{General_Category=/a/Other_PUNCTUATION}');
Error('\P{General_Category=/a/Other_PUNCTUATION}');
Expect(1, 74867, '\p{General_Category=otherpunctuation}', "");
Expect(0, 74867, '\p{^General_Category=otherpunctuation}', "");
Expect(0, 74867, '\P{General_Category=otherpunctuation}', "");
Expect(1, 74867, '\P{^General_Category=otherpunctuation}', "");
Expect(0, 74868, '\p{General_Category=otherpunctuation}', "");
Expect(1, 74868, '\p{^General_Category=otherpunctuation}', "");
Expect(1, 74868, '\P{General_Category=otherpunctuation}', "");
Expect(0, 74868, '\P{^General_Category=otherpunctuation}', "");
Expect(1, 74867, '\p{General_Category=_	Other_Punctuation}', "");
Expect(0, 74867, '\p{^General_Category=_	Other_Punctuation}', "");
Expect(0, 74867, '\P{General_Category=_	Other_Punctuation}', "");
Expect(1, 74867, '\P{^General_Category=_	Other_Punctuation}', "");
Expect(0, 74868, '\p{General_Category=_	Other_Punctuation}', "");
Expect(1, 74868, '\p{^General_Category=_	Other_Punctuation}', "");
Expect(1, 74868, '\P{General_Category=_	Other_Punctuation}', "");
Expect(0, 74868, '\P{^General_Category=_	Other_Punctuation}', "");
Error('\p{Gc=:=-Po}');
Error('\P{Gc=:=-Po}');
Expect(1, 74867, '\p{Gc=po}', "");
Expect(0, 74867, '\p{^Gc=po}', "");
Expect(0, 74867, '\P{Gc=po}', "");
Expect(1, 74867, '\P{^Gc=po}', "");
Expect(0, 74868, '\p{Gc=po}', "");
Expect(1, 74868, '\p{^Gc=po}', "");
Expect(1, 74868, '\P{Gc=po}', "");
Expect(0, 74868, '\P{^Gc=po}', "");
Expect(1, 74867, '\p{Gc=-_PO}', "");
Expect(0, 74867, '\p{^Gc=-_PO}', "");
Expect(0, 74867, '\P{Gc=-_PO}', "");
Expect(1, 74867, '\P{^Gc=-_PO}', "");
Expect(0, 74868, '\p{Gc=-_PO}', "");
Expect(1, 74868, '\p{^Gc=-_PO}', "");
Expect(1, 74868, '\P{Gc=-_PO}', "");
Expect(0, 74868, '\P{^Gc=-_PO}', "");
Error('\p{Category=	 Other_Punctuation:=}');
Error('\P{Category=	 Other_Punctuation:=}');
Expect(1, 74867, '\p{Category:	otherpunctuation}', "");
Expect(0, 74867, '\p{^Category:	otherpunctuation}', "");
Expect(0, 74867, '\P{Category:	otherpunctuation}', "");
Expect(1, 74867, '\P{^Category:	otherpunctuation}', "");
Expect(0, 74868, '\p{Category:	otherpunctuation}', "");
Expect(1, 74868, '\p{^Category:	otherpunctuation}', "");
Expect(1, 74868, '\P{Category:	otherpunctuation}', "");
Expect(0, 74868, '\P{^Category:	otherpunctuation}', "");
Expect(1, 74867, '\p{Category=_-OTHER_PUNCTUATION}', "");
Expect(0, 74867, '\p{^Category=_-OTHER_PUNCTUATION}', "");
Expect(0, 74867, '\P{Category=_-OTHER_PUNCTUATION}', "");
Expect(1, 74867, '\P{^Category=_-OTHER_PUNCTUATION}', "");
Expect(0, 74868, '\p{Category=_-OTHER_PUNCTUATION}', "");
Expect(1, 74868, '\p{^Category=_-OTHER_PUNCTUATION}', "");
Expect(1, 74868, '\P{Category=_-OTHER_PUNCTUATION}', "");
Expect(0, 74868, '\P{^Category=_-OTHER_PUNCTUATION}', "");
Error('\p{Is_General_Category=-po:=}');
Error('\P{Is_General_Category=-po:=}');
Expect(1, 74867, '\p{Is_General_Category=po}', "");
Expect(0, 74867, '\p{^Is_General_Category=po}', "");
Expect(0, 74867, '\P{Is_General_Category=po}', "");
Expect(1, 74867, '\P{^Is_General_Category=po}', "");
Expect(0, 74868, '\p{Is_General_Category=po}', "");
Expect(1, 74868, '\p{^Is_General_Category=po}', "");
Expect(1, 74868, '\P{Is_General_Category=po}', "");
Expect(0, 74868, '\P{^Is_General_Category=po}', "");
Expect(1, 74867, '\p{Is_General_Category:   _ Po}', "");
Expect(0, 74867, '\p{^Is_General_Category:   _ Po}', "");
Expect(0, 74867, '\P{Is_General_Category:   _ Po}', "");
Expect(1, 74867, '\P{^Is_General_Category:   _ Po}', "");
Expect(0, 74868, '\p{Is_General_Category:   _ Po}', "");
Expect(1, 74868, '\p{^Is_General_Category:   _ Po}', "");
Expect(1, 74868, '\P{Is_General_Category:   _ Po}', "");
Expect(0, 74868, '\P{^Is_General_Category:   _ Po}', "");
Error('\p{Is_Gc=-_Other_Punctuation/a/}');
Error('\P{Is_Gc=-_Other_Punctuation/a/}');
Expect(1, 74867, '\p{Is_Gc=otherpunctuation}', "");
Expect(0, 74867, '\p{^Is_Gc=otherpunctuation}', "");
Expect(0, 74867, '\P{Is_Gc=otherpunctuation}', "");
Expect(1, 74867, '\P{^Is_Gc=otherpunctuation}', "");
Expect(0, 74868, '\p{Is_Gc=otherpunctuation}', "");
Expect(1, 74868, '\p{^Is_Gc=otherpunctuation}', "");
Expect(1, 74868, '\P{Is_Gc=otherpunctuation}', "");
Expect(0, 74868, '\P{^Is_Gc=otherpunctuation}', "");
Expect(1, 74867, '\p{Is_Gc=	Other_punctuation}', "");
Expect(0, 74867, '\p{^Is_Gc=	Other_punctuation}', "");
Expect(0, 74867, '\P{Is_Gc=	Other_punctuation}', "");
Expect(1, 74867, '\P{^Is_Gc=	Other_punctuation}', "");
Expect(0, 74868, '\p{Is_Gc=	Other_punctuation}', "");
Expect(1, 74868, '\p{^Is_Gc=	Other_punctuation}', "");
Expect(1, 74868, '\P{Is_Gc=	Other_punctuation}', "");
Expect(0, 74868, '\P{^Is_Gc=	Other_punctuation}', "");
Error('\p{Is_Category=		po:=}');
Error('\P{Is_Category=		po:=}');
Expect(1, 74867, '\p{Is_Category=po}', "");
Expect(0, 74867, '\p{^Is_Category=po}', "");
Expect(0, 74867, '\P{Is_Category=po}', "");
Expect(1, 74867, '\P{^Is_Category=po}', "");
Expect(0, 74868, '\p{Is_Category=po}', "");
Expect(1, 74868, '\p{^Is_Category=po}', "");
Expect(1, 74868, '\P{Is_Category=po}', "");
Expect(0, 74868, '\P{^Is_Category=po}', "");
Expect(1, 74867, '\p{Is_Category=-Po}', "");
Expect(0, 74867, '\p{^Is_Category=-Po}', "");
Expect(0, 74867, '\P{Is_Category=-Po}', "");
Expect(1, 74867, '\P{^Is_Category=-Po}', "");
Expect(0, 74868, '\p{Is_Category=-Po}', "");
Expect(1, 74868, '\p{^Is_Category=-Po}', "");
Expect(1, 74868, '\P{Is_Category=-Po}', "");
Expect(0, 74868, '\P{^Is_Category=-Po}', "");
Error('\p{General_Category=-	open_PUNCTUATION/a/}');
Error('\P{General_Category=-	open_PUNCTUATION/a/}');
Expect(1, 65378, '\p{General_Category=openpunctuation}', "");
Expect(0, 65378, '\p{^General_Category=openpunctuation}', "");
Expect(0, 65378, '\P{General_Category=openpunctuation}', "");
Expect(1, 65378, '\P{^General_Category=openpunctuation}', "");
Expect(0, 65379, '\p{General_Category=openpunctuation}', "");
Expect(1, 65379, '\p{^General_Category=openpunctuation}', "");
Expect(1, 65379, '\P{General_Category=openpunctuation}', "");
Expect(0, 65379, '\P{^General_Category=openpunctuation}', "");
Expect(1, 65378, '\p{General_Category= 	OPEN_punctuation}', "");
Expect(0, 65378, '\p{^General_Category= 	OPEN_punctuation}', "");
Expect(0, 65378, '\P{General_Category= 	OPEN_punctuation}', "");
Expect(1, 65378, '\P{^General_Category= 	OPEN_punctuation}', "");
Expect(0, 65379, '\p{General_Category= 	OPEN_punctuation}', "");
Expect(1, 65379, '\p{^General_Category= 	OPEN_punctuation}', "");
Expect(1, 65379, '\P{General_Category= 	OPEN_punctuation}', "");
Expect(0, 65379, '\P{^General_Category= 	OPEN_punctuation}', "");
Error('\p{Gc=	 Ps:=}');
Error('\P{Gc=	 Ps:=}');
Expect(1, 65378, '\p{Gc=ps}', "");
Expect(0, 65378, '\p{^Gc=ps}', "");
Expect(0, 65378, '\P{Gc=ps}', "");
Expect(1, 65378, '\P{^Gc=ps}', "");
Expect(0, 65379, '\p{Gc=ps}', "");
Expect(1, 65379, '\p{^Gc=ps}', "");
Expect(1, 65379, '\P{Gc=ps}', "");
Expect(0, 65379, '\P{^Gc=ps}', "");
Expect(1, 65378, '\p{Gc=-	Ps}', "");
Expect(0, 65378, '\p{^Gc=-	Ps}', "");
Expect(0, 65378, '\P{Gc=-	Ps}', "");
Expect(1, 65378, '\P{^Gc=-	Ps}', "");
Expect(0, 65379, '\p{Gc=-	Ps}', "");
Expect(1, 65379, '\p{^Gc=-	Ps}', "");
Expect(1, 65379, '\P{Gc=-	Ps}', "");
Expect(0, 65379, '\P{^Gc=-	Ps}', "");
Error('\p{Category=:=	_Open_Punctuation}');
Error('\P{Category=:=	_Open_Punctuation}');
Expect(1, 65378, '\p{Category=openpunctuation}', "");
Expect(0, 65378, '\p{^Category=openpunctuation}', "");
Expect(0, 65378, '\P{Category=openpunctuation}', "");
Expect(1, 65378, '\P{^Category=openpunctuation}', "");
Expect(0, 65379, '\p{Category=openpunctuation}', "");
Expect(1, 65379, '\p{^Category=openpunctuation}', "");
Expect(1, 65379, '\P{Category=openpunctuation}', "");
Expect(0, 65379, '\P{^Category=openpunctuation}', "");
Expect(1, 65378, '\p{Category= -Open_punctuation}', "");
Expect(0, 65378, '\p{^Category= -Open_punctuation}', "");
Expect(0, 65378, '\P{Category= -Open_punctuation}', "");
Expect(1, 65378, '\P{^Category= -Open_punctuation}', "");
Expect(0, 65379, '\p{Category= -Open_punctuation}', "");
Expect(1, 65379, '\p{^Category= -Open_punctuation}', "");
Expect(1, 65379, '\P{Category= -Open_punctuation}', "");
Expect(0, 65379, '\P{^Category= -Open_punctuation}', "");
Error('\p{Is_General_Category=_	Ps:=}');
Error('\P{Is_General_Category=_	Ps:=}');
Expect(1, 65378, '\p{Is_General_Category=ps}', "");
Expect(0, 65378, '\p{^Is_General_Category=ps}', "");
Expect(0, 65378, '\P{Is_General_Category=ps}', "");
Expect(1, 65378, '\P{^Is_General_Category=ps}', "");
Expect(0, 65379, '\p{Is_General_Category=ps}', "");
Expect(1, 65379, '\p{^Is_General_Category=ps}', "");
Expect(1, 65379, '\P{Is_General_Category=ps}', "");
Expect(0, 65379, '\P{^Is_General_Category=ps}', "");
Expect(1, 65378, '\p{Is_General_Category=	ps}', "");
Expect(0, 65378, '\p{^Is_General_Category=	ps}', "");
Expect(0, 65378, '\P{Is_General_Category=	ps}', "");
Expect(1, 65378, '\P{^Is_General_Category=	ps}', "");
Expect(0, 65379, '\p{Is_General_Category=	ps}', "");
Expect(1, 65379, '\p{^Is_General_Category=	ps}', "");
Expect(1, 65379, '\P{Is_General_Category=	ps}', "");
Expect(0, 65379, '\P{^Is_General_Category=	ps}', "");
Error('\p{Is_Gc=:= Open_punctuation}');
Error('\P{Is_Gc=:= Open_punctuation}');
Expect(1, 65378, '\p{Is_Gc=openpunctuation}', "");
Expect(0, 65378, '\p{^Is_Gc=openpunctuation}', "");
Expect(0, 65378, '\P{Is_Gc=openpunctuation}', "");
Expect(1, 65378, '\P{^Is_Gc=openpunctuation}', "");
Expect(0, 65379, '\p{Is_Gc=openpunctuation}', "");
Expect(1, 65379, '\p{^Is_Gc=openpunctuation}', "");
Expect(1, 65379, '\P{Is_Gc=openpunctuation}', "");
Expect(0, 65379, '\P{^Is_Gc=openpunctuation}', "");
Expect(1, 65378, '\p{Is_Gc=__Open_PUNCTUATION}', "");
Expect(0, 65378, '\p{^Is_Gc=__Open_PUNCTUATION}', "");
Expect(0, 65378, '\P{Is_Gc=__Open_PUNCTUATION}', "");
Expect(1, 65378, '\P{^Is_Gc=__Open_PUNCTUATION}', "");
Expect(0, 65379, '\p{Is_Gc=__Open_PUNCTUATION}', "");
Expect(1, 65379, '\p{^Is_Gc=__Open_PUNCTUATION}', "");
Expect(1, 65379, '\P{Is_Gc=__Open_PUNCTUATION}', "");
Expect(0, 65379, '\P{^Is_Gc=__Open_PUNCTUATION}', "");
Error('\p{Is_Category=/a/-Ps}');
Error('\P{Is_Category=/a/-Ps}');
Expect(1, 65378, '\p{Is_Category=ps}', "");
Expect(0, 65378, '\p{^Is_Category=ps}', "");
Expect(0, 65378, '\P{Is_Category=ps}', "");
Expect(1, 65378, '\P{^Is_Category=ps}', "");
Expect(0, 65379, '\p{Is_Category=ps}', "");
Expect(1, 65379, '\p{^Is_Category=ps}', "");
Expect(1, 65379, '\P{Is_Category=ps}', "");
Expect(0, 65379, '\P{^Is_Category=ps}', "");
Expect(1, 65378, '\p{Is_Category=		ps}', "");
Expect(0, 65378, '\p{^Is_Category=		ps}', "");
Expect(0, 65378, '\P{Is_Category=		ps}', "");
Expect(1, 65378, '\P{^Is_Category=		ps}', "");
Expect(0, 65379, '\p{Is_Category=		ps}', "");
Expect(1, 65379, '\p{^Is_Category=		ps}', "");
Expect(1, 65379, '\P{Is_Category=		ps}', "");
Expect(0, 65379, '\P{^Is_Category=		ps}', "");
Error('\p{General_Category=/a/ -SYMBOL}');
Error('\P{General_Category=/a/ -SYMBOL}');
Expect(1, 128883, '\p{General_Category=symbol}', "");
Expect(0, 128883, '\p{^General_Category=symbol}', "");
Expect(0, 128883, '\P{General_Category=symbol}', "");
Expect(1, 128883, '\P{^General_Category=symbol}', "");
Expect(0, 128884, '\p{General_Category=symbol}', "");
Expect(1, 128884, '\p{^General_Category=symbol}', "");
Expect(1, 128884, '\P{General_Category=symbol}', "");
Expect(0, 128884, '\P{^General_Category=symbol}', "");
Expect(1, 128883, '\p{General_Category=  symbol}', "");
Expect(0, 128883, '\p{^General_Category=  symbol}', "");
Expect(0, 128883, '\P{General_Category=  symbol}', "");
Expect(1, 128883, '\P{^General_Category=  symbol}', "");
Expect(0, 128884, '\p{General_Category=  symbol}', "");
Expect(1, 128884, '\p{^General_Category=  symbol}', "");
Expect(1, 128884, '\P{General_Category=  symbol}', "");
Expect(0, 128884, '\P{^General_Category=  symbol}', "");
Error('\p{Gc=- s/a/}');
Error('\P{Gc=- s/a/}');
Expect(1, 128883, '\p{Gc=s}', "");
Expect(0, 128883, '\p{^Gc=s}', "");
Expect(0, 128883, '\P{Gc=s}', "");
Expect(1, 128883, '\P{^Gc=s}', "");
Expect(0, 128884, '\p{Gc=s}', "");
Expect(1, 128884, '\p{^Gc=s}', "");
Expect(1, 128884, '\P{Gc=s}', "");
Expect(0, 128884, '\P{^Gc=s}', "");
Expect(1, 128883, '\p{Gc=-	s}', "");
Expect(0, 128883, '\p{^Gc=-	s}', "");
Expect(0, 128883, '\P{Gc=-	s}', "");
Expect(1, 128883, '\P{^Gc=-	s}', "");
Expect(0, 128884, '\p{Gc=-	s}', "");
Expect(1, 128884, '\p{^Gc=-	s}', "");
Expect(1, 128884, '\P{Gc=-	s}', "");
Expect(0, 128884, '\P{^Gc=-	s}', "");
Error('\p{Category=:= symbol}');
Error('\P{Category=:= symbol}');
Expect(1, 128883, '\p{Category=symbol}', "");
Expect(0, 128883, '\p{^Category=symbol}', "");
Expect(0, 128883, '\P{Category=symbol}', "");
Expect(1, 128883, '\P{^Category=symbol}', "");
Expect(0, 128884, '\p{Category=symbol}', "");
Expect(1, 128884, '\p{^Category=symbol}', "");
Expect(1, 128884, '\P{Category=symbol}', "");
Expect(0, 128884, '\P{^Category=symbol}', "");
Expect(1, 128883, '\p{Category=_Symbol}', "");
Expect(0, 128883, '\p{^Category=_Symbol}', "");
Expect(0, 128883, '\P{Category=_Symbol}', "");
Expect(1, 128883, '\P{^Category=_Symbol}', "");
Expect(0, 128884, '\p{Category=_Symbol}', "");
Expect(1, 128884, '\p{^Category=_Symbol}', "");
Expect(1, 128884, '\P{Category=_Symbol}', "");
Expect(0, 128884, '\P{^Category=_Symbol}', "");
Error('\p{Is_General_Category= :=s}');
Error('\P{Is_General_Category= :=s}');
Expect(1, 128883, '\p{Is_General_Category:   s}', "");
Expect(0, 128883, '\p{^Is_General_Category:   s}', "");
Expect(0, 128883, '\P{Is_General_Category:   s}', "");
Expect(1, 128883, '\P{^Is_General_Category:   s}', "");
Expect(0, 128884, '\p{Is_General_Category:   s}', "");
Expect(1, 128884, '\p{^Is_General_Category:   s}', "");
Expect(1, 128884, '\P{Is_General_Category:   s}', "");
Expect(0, 128884, '\P{^Is_General_Category:   s}', "");
Expect(1, 128883, '\p{Is_General_Category= -s}', "");
Expect(0, 128883, '\p{^Is_General_Category= -s}', "");
Expect(0, 128883, '\P{Is_General_Category= -s}', "");
Expect(1, 128883, '\P{^Is_General_Category= -s}', "");
Expect(0, 128884, '\p{Is_General_Category= -s}', "");
Expect(1, 128884, '\p{^Is_General_Category= -s}', "");
Expect(1, 128884, '\P{Is_General_Category= -s}', "");
Expect(0, 128884, '\P{^Is_General_Category= -s}', "");
Error('\p{Is_Gc=-	Symbol:=}');
Error('\P{Is_Gc=-	Symbol:=}');
Expect(1, 128883, '\p{Is_Gc=symbol}', "");
Expect(0, 128883, '\p{^Is_Gc=symbol}', "");
Expect(0, 128883, '\P{Is_Gc=symbol}', "");
Expect(1, 128883, '\P{^Is_Gc=symbol}', "");
Expect(0, 128884, '\p{Is_Gc=symbol}', "");
Expect(1, 128884, '\p{^Is_Gc=symbol}', "");
Expect(1, 128884, '\P{Is_Gc=symbol}', "");
Expect(0, 128884, '\P{^Is_Gc=symbol}', "");
Expect(1, 128883, '\p{Is_Gc=	 symbol}', "");
Expect(0, 128883, '\p{^Is_Gc=	 symbol}', "");
Expect(0, 128883, '\P{Is_Gc=	 symbol}', "");
Expect(1, 128883, '\P{^Is_Gc=	 symbol}', "");
Expect(0, 128884, '\p{Is_Gc=	 symbol}', "");
Expect(1, 128884, '\p{^Is_Gc=	 symbol}', "");
Expect(1, 128884, '\P{Is_Gc=	 symbol}', "");
Expect(0, 128884, '\P{^Is_Gc=	 symbol}', "");
Error('\p{Is_Category=	/a/S}');
Error('\P{Is_Category=	/a/S}');
Expect(1, 128883, '\p{Is_Category=s}', "");
Expect(0, 128883, '\p{^Is_Category=s}', "");
Expect(0, 128883, '\P{Is_Category=s}', "");
Expect(1, 128883, '\P{^Is_Category=s}', "");
Expect(0, 128884, '\p{Is_Category=s}', "");
Expect(1, 128884, '\p{^Is_Category=s}', "");
Expect(1, 128884, '\P{Is_Category=s}', "");
Expect(0, 128884, '\P{^Is_Category=s}', "");
Expect(1, 128883, '\p{Is_Category=S}', "");
Expect(0, 128883, '\p{^Is_Category=S}', "");
Expect(0, 128883, '\P{Is_Category=S}', "");
Expect(1, 128883, '\P{^Is_Category=S}', "");
Expect(0, 128884, '\p{Is_Category=S}', "");
Expect(1, 128884, '\p{^Is_Category=S}', "");
Expect(1, 128884, '\P{Is_Category=S}', "");
Expect(0, 128884, '\P{^Is_Category=S}', "");
Error('\p{General_Category=:=-currency_Symbol}');
Error('\P{General_Category=:=-currency_Symbol}');
Expect(1, 65510, '\p{General_Category=currencysymbol}', "");
Expect(0, 65510, '\p{^General_Category=currencysymbol}', "");
Expect(0, 65510, '\P{General_Category=currencysymbol}', "");
Expect(1, 65510, '\P{^General_Category=currencysymbol}', "");
Expect(0, 65511, '\p{General_Category=currencysymbol}', "");
Expect(1, 65511, '\p{^General_Category=currencysymbol}', "");
Expect(1, 65511, '\P{General_Category=currencysymbol}', "");
Expect(0, 65511, '\P{^General_Category=currencysymbol}', "");
Expect(1, 65510, '\p{General_Category=_currency_SYMBOL}', "");
Expect(0, 65510, '\p{^General_Category=_currency_SYMBOL}', "");
Expect(0, 65510, '\P{General_Category=_currency_SYMBOL}', "");
Expect(1, 65510, '\P{^General_Category=_currency_SYMBOL}', "");
Expect(0, 65511, '\p{General_Category=_currency_SYMBOL}', "");
Expect(1, 65511, '\p{^General_Category=_currency_SYMBOL}', "");
Expect(1, 65511, '\P{General_Category=_currency_SYMBOL}', "");
Expect(0, 65511, '\P{^General_Category=_currency_SYMBOL}', "");
Error('\p{Gc: /a/_Sc}');
Error('\P{Gc: /a/_Sc}');
Expect(1, 65510, '\p{Gc=sc}', "");
Expect(0, 65510, '\p{^Gc=sc}', "");
Expect(0, 65510, '\P{Gc=sc}', "");
Expect(1, 65510, '\P{^Gc=sc}', "");
Expect(0, 65511, '\p{Gc=sc}', "");
Expect(1, 65511, '\p{^Gc=sc}', "");
Expect(1, 65511, '\P{Gc=sc}', "");
Expect(0, 65511, '\P{^Gc=sc}', "");
Expect(1, 65510, '\p{Gc=-_Sc}', "");
Expect(0, 65510, '\p{^Gc=-_Sc}', "");
Expect(0, 65510, '\P{Gc=-_Sc}', "");
Expect(1, 65510, '\P{^Gc=-_Sc}', "");
Expect(0, 65511, '\p{Gc=-_Sc}', "");
Expect(1, 65511, '\p{^Gc=-_Sc}', "");
Expect(1, 65511, '\P{Gc=-_Sc}', "");
Expect(0, 65511, '\P{^Gc=-_Sc}', "");
Error('\p{Category= :=currency_SYMBOL}');
Error('\P{Category= :=currency_SYMBOL}');
Expect(1, 65510, '\p{Category=currencysymbol}', "");
Expect(0, 65510, '\p{^Category=currencysymbol}', "");
Expect(0, 65510, '\P{Category=currencysymbol}', "");
Expect(1, 65510, '\P{^Category=currencysymbol}', "");
Expect(0, 65511, '\p{Category=currencysymbol}', "");
Expect(1, 65511, '\p{^Category=currencysymbol}', "");
Expect(1, 65511, '\P{Category=currencysymbol}', "");
Expect(0, 65511, '\P{^Category=currencysymbol}', "");
Expect(1, 65510, '\p{Category=-	Currency_symbol}', "");
Expect(0, 65510, '\p{^Category=-	Currency_symbol}', "");
Expect(0, 65510, '\P{Category=-	Currency_symbol}', "");
Expect(1, 65510, '\P{^Category=-	Currency_symbol}', "");
Expect(0, 65511, '\p{Category=-	Currency_symbol}', "");
Expect(1, 65511, '\p{^Category=-	Currency_symbol}', "");
Expect(1, 65511, '\P{Category=-	Currency_symbol}', "");
Expect(0, 65511, '\P{^Category=-	Currency_symbol}', "");
Error('\p{Is_General_Category=-_Sc/a/}');
Error('\P{Is_General_Category=-_Sc/a/}');
Expect(1, 65510, '\p{Is_General_Category=sc}', "");
Expect(0, 65510, '\p{^Is_General_Category=sc}', "");
Expect(0, 65510, '\P{Is_General_Category=sc}', "");
Expect(1, 65510, '\P{^Is_General_Category=sc}', "");
Expect(0, 65511, '\p{Is_General_Category=sc}', "");
Expect(1, 65511, '\p{^Is_General_Category=sc}', "");
Expect(1, 65511, '\P{Is_General_Category=sc}', "");
Expect(0, 65511, '\P{^Is_General_Category=sc}', "");
Expect(1, 65510, '\p{Is_General_Category=		Sc}', "");
Expect(0, 65510, '\p{^Is_General_Category=		Sc}', "");
Expect(0, 65510, '\P{Is_General_Category=		Sc}', "");
Expect(1, 65510, '\P{^Is_General_Category=		Sc}', "");
Expect(0, 65511, '\p{Is_General_Category=		Sc}', "");
Expect(1, 65511, '\p{^Is_General_Category=		Sc}', "");
Expect(1, 65511, '\P{Is_General_Category=		Sc}', "");
Expect(0, 65511, '\P{^Is_General_Category=		Sc}', "");
Error('\p{Is_Gc=	 Currency_Symbol/a/}');
Error('\P{Is_Gc=	 Currency_Symbol/a/}');
Expect(1, 65510, '\p{Is_Gc=currencysymbol}', "");
Expect(0, 65510, '\p{^Is_Gc=currencysymbol}', "");
Expect(0, 65510, '\P{Is_Gc=currencysymbol}', "");
Expect(1, 65510, '\P{^Is_Gc=currencysymbol}', "");
Expect(0, 65511, '\p{Is_Gc=currencysymbol}', "");
Expect(1, 65511, '\p{^Is_Gc=currencysymbol}', "");
Expect(1, 65511, '\P{Is_Gc=currencysymbol}', "");
Expect(0, 65511, '\P{^Is_Gc=currencysymbol}', "");
Expect(1, 65510, '\p{Is_Gc=	-CURRENCY_Symbol}', "");
Expect(0, 65510, '\p{^Is_Gc=	-CURRENCY_Symbol}', "");
Expect(0, 65510, '\P{Is_Gc=	-CURRENCY_Symbol}', "");
Expect(1, 65510, '\P{^Is_Gc=	-CURRENCY_Symbol}', "");
Expect(0, 65511, '\p{Is_Gc=	-CURRENCY_Symbol}', "");
Expect(1, 65511, '\p{^Is_Gc=	-CURRENCY_Symbol}', "");
Expect(1, 65511, '\P{Is_Gc=	-CURRENCY_Symbol}', "");
Expect(0, 65511, '\P{^Is_Gc=	-CURRENCY_Symbol}', "");
Error('\p{Is_Category=		sc/a/}');
Error('\P{Is_Category=		sc/a/}');
Expect(1, 65510, '\p{Is_Category=sc}', "");
Expect(0, 65510, '\p{^Is_Category=sc}', "");
Expect(0, 65510, '\P{Is_Category=sc}', "");
Expect(1, 65510, '\P{^Is_Category=sc}', "");
Expect(0, 65511, '\p{Is_Category=sc}', "");
Expect(1, 65511, '\p{^Is_Category=sc}', "");
Expect(1, 65511, '\P{Is_Category=sc}', "");
Expect(0, 65511, '\P{^Is_Category=sc}', "");
Expect(1, 65510, '\p{Is_Category=	 SC}', "");
Expect(0, 65510, '\p{^Is_Category=	 SC}', "");
Expect(0, 65510, '\P{Is_Category=	 SC}', "");
Expect(1, 65510, '\P{^Is_Category=	 SC}', "");
Expect(0, 65511, '\p{Is_Category=	 SC}', "");
Expect(1, 65511, '\p{^Is_Category=	 SC}', "");
Expect(1, 65511, '\P{Is_Category=	 SC}', "");
Expect(0, 65511, '\P{^Is_Category=	 SC}', "");
Error('\p{General_Category= 	modifier_Symbol/a/}');
Error('\P{General_Category= 	modifier_Symbol/a/}');
Expect(1, 65507, '\p{General_Category: modifiersymbol}', "");
Expect(0, 65507, '\p{^General_Category: modifiersymbol}', "");
Expect(0, 65507, '\P{General_Category: modifiersymbol}', "");
Expect(1, 65507, '\P{^General_Category: modifiersymbol}', "");
Expect(0, 65508, '\p{General_Category: modifiersymbol}', "");
Expect(1, 65508, '\p{^General_Category: modifiersymbol}', "");
Expect(1, 65508, '\P{General_Category: modifiersymbol}', "");
Expect(0, 65508, '\P{^General_Category: modifiersymbol}', "");
Expect(1, 65507, '\p{General_Category=	-Modifier_Symbol}', "");
Expect(0, 65507, '\p{^General_Category=	-Modifier_Symbol}', "");
Expect(0, 65507, '\P{General_Category=	-Modifier_Symbol}', "");
Expect(1, 65507, '\P{^General_Category=	-Modifier_Symbol}', "");
Expect(0, 65508, '\p{General_Category=	-Modifier_Symbol}', "");
Expect(1, 65508, '\p{^General_Category=	-Modifier_Symbol}', "");
Expect(1, 65508, '\P{General_Category=	-Modifier_Symbol}', "");
Expect(0, 65508, '\P{^General_Category=	-Modifier_Symbol}', "");
Error('\p{Gc= /a/Sk}');
Error('\P{Gc= /a/Sk}');
Expect(1, 65507, '\p{Gc=sk}', "");
Expect(0, 65507, '\p{^Gc=sk}', "");
Expect(0, 65507, '\P{Gc=sk}', "");
Expect(1, 65507, '\P{^Gc=sk}', "");
Expect(0, 65508, '\p{Gc=sk}', "");
Expect(1, 65508, '\p{^Gc=sk}', "");
Expect(1, 65508, '\P{Gc=sk}', "");
Expect(0, 65508, '\P{^Gc=sk}', "");
Expect(1, 65507, '\p{Gc=-sk}', "");
Expect(0, 65507, '\p{^Gc=-sk}', "");
Expect(0, 65507, '\P{Gc=-sk}', "");
Expect(1, 65507, '\P{^Gc=-sk}', "");
Expect(0, 65508, '\p{Gc=-sk}', "");
Expect(1, 65508, '\p{^Gc=-sk}', "");
Expect(1, 65508, '\P{Gc=-sk}', "");
Expect(0, 65508, '\P{^Gc=-sk}', "");
Error('\p{Category=/a/	_Modifier_Symbol}');
Error('\P{Category=/a/	_Modifier_Symbol}');
Expect(1, 65507, '\p{Category=modifiersymbol}', "");
Expect(0, 65507, '\p{^Category=modifiersymbol}', "");
Expect(0, 65507, '\P{Category=modifiersymbol}', "");
Expect(1, 65507, '\P{^Category=modifiersymbol}', "");
Expect(0, 65508, '\p{Category=modifiersymbol}', "");
Expect(1, 65508, '\p{^Category=modifiersymbol}', "");
Expect(1, 65508, '\P{Category=modifiersymbol}', "");
Expect(0, 65508, '\P{^Category=modifiersymbol}', "");
Expect(1, 65507, '\p{Category:		-Modifier_symbol}', "");
Expect(0, 65507, '\p{^Category:		-Modifier_symbol}', "");
Expect(0, 65507, '\P{Category:		-Modifier_symbol}', "");
Expect(1, 65507, '\P{^Category:		-Modifier_symbol}', "");
Expect(0, 65508, '\p{Category:		-Modifier_symbol}', "");
Expect(1, 65508, '\p{^Category:		-Modifier_symbol}', "");
Expect(1, 65508, '\P{Category:		-Modifier_symbol}', "");
Expect(0, 65508, '\P{^Category:		-Modifier_symbol}', "");
Error('\p{Is_General_Category=:=  Sk}');
Error('\P{Is_General_Category=:=  Sk}');
Expect(1, 65507, '\p{Is_General_Category=sk}', "");
Expect(0, 65507, '\p{^Is_General_Category=sk}', "");
Expect(0, 65507, '\P{Is_General_Category=sk}', "");
Expect(1, 65507, '\P{^Is_General_Category=sk}', "");
Expect(0, 65508, '\p{Is_General_Category=sk}', "");
Expect(1, 65508, '\p{^Is_General_Category=sk}', "");
Expect(1, 65508, '\P{Is_General_Category=sk}', "");
Expect(0, 65508, '\P{^Is_General_Category=sk}', "");
Expect(1, 65507, '\p{Is_General_Category=	 Sk}', "");
Expect(0, 65507, '\p{^Is_General_Category=	 Sk}', "");
Expect(0, 65507, '\P{Is_General_Category=	 Sk}', "");
Expect(1, 65507, '\P{^Is_General_Category=	 Sk}', "");
Expect(0, 65508, '\p{Is_General_Category=	 Sk}', "");
Expect(1, 65508, '\p{^Is_General_Category=	 Sk}', "");
Expect(1, 65508, '\P{Is_General_Category=	 Sk}', "");
Expect(0, 65508, '\P{^Is_General_Category=	 Sk}', "");
Error('\p{Is_Gc=/a/MODIFIER_Symbol}');
Error('\P{Is_Gc=/a/MODIFIER_Symbol}');
Expect(1, 65507, '\p{Is_Gc=modifiersymbol}', "");
Expect(0, 65507, '\p{^Is_Gc=modifiersymbol}', "");
Expect(0, 65507, '\P{Is_Gc=modifiersymbol}', "");
Expect(1, 65507, '\P{^Is_Gc=modifiersymbol}', "");
Expect(0, 65508, '\p{Is_Gc=modifiersymbol}', "");
Expect(1, 65508, '\p{^Is_Gc=modifiersymbol}', "");
Expect(1, 65508, '\P{Is_Gc=modifiersymbol}', "");
Expect(0, 65508, '\P{^Is_Gc=modifiersymbol}', "");
Expect(1, 65507, '\p{Is_Gc=-modifier_SYMBOL}', "");
Expect(0, 65507, '\p{^Is_Gc=-modifier_SYMBOL}', "");
Expect(0, 65507, '\P{Is_Gc=-modifier_SYMBOL}', "");
Expect(1, 65507, '\P{^Is_Gc=-modifier_SYMBOL}', "");
Expect(0, 65508, '\p{Is_Gc=-modifier_SYMBOL}', "");
Expect(1, 65508, '\p{^Is_Gc=-modifier_SYMBOL}', "");
Expect(1, 65508, '\P{Is_Gc=-modifier_SYMBOL}', "");
Expect(0, 65508, '\P{^Is_Gc=-modifier_SYMBOL}', "");
Error('\p{Is_Category=  sk/a/}');
Error('\P{Is_Category=  sk/a/}');
Expect(1, 65507, '\p{Is_Category=sk}', "");
Expect(0, 65507, '\p{^Is_Category=sk}', "");
Expect(0, 65507, '\P{Is_Category=sk}', "");
Expect(1, 65507, '\P{^Is_Category=sk}', "");
Expect(0, 65508, '\p{Is_Category=sk}', "");
Expect(1, 65508, '\p{^Is_Category=sk}', "");
Expect(1, 65508, '\P{Is_Category=sk}', "");
Expect(0, 65508, '\P{^Is_Category=sk}', "");
Expect(1, 65507, '\p{Is_Category=_ sk}', "");
Expect(0, 65507, '\p{^Is_Category=_ sk}', "");
Expect(0, 65507, '\P{Is_Category=_ sk}', "");
Expect(1, 65507, '\P{^Is_Category=_ sk}', "");
Expect(0, 65508, '\p{Is_Category=_ sk}', "");
Expect(1, 65508, '\p{^Is_Category=_ sk}', "");
Expect(1, 65508, '\P{Is_Category=_ sk}', "");
Expect(0, 65508, '\P{^Is_Category=_ sk}', "");
Error('\p{General_Category=/a/	math_Symbol}');
Error('\P{General_Category=/a/	math_Symbol}');
Expect(1, 126705, '\p{General_Category=mathsymbol}', "");
Expect(0, 126705, '\p{^General_Category=mathsymbol}', "");
Expect(0, 126705, '\P{General_Category=mathsymbol}', "");
Expect(1, 126705, '\P{^General_Category=mathsymbol}', "");
Expect(0, 126706, '\p{General_Category=mathsymbol}', "");
Expect(1, 126706, '\p{^General_Category=mathsymbol}', "");
Expect(1, 126706, '\P{General_Category=mathsymbol}', "");
Expect(0, 126706, '\P{^General_Category=mathsymbol}', "");
Expect(1, 126705, '\p{General_Category= 	Math_SYMBOL}', "");
Expect(0, 126705, '\p{^General_Category= 	Math_SYMBOL}', "");
Expect(0, 126705, '\P{General_Category= 	Math_SYMBOL}', "");
Expect(1, 126705, '\P{^General_Category= 	Math_SYMBOL}', "");
Expect(0, 126706, '\p{General_Category= 	Math_SYMBOL}', "");
Expect(1, 126706, '\p{^General_Category= 	Math_SYMBOL}', "");
Expect(1, 126706, '\P{General_Category= 	Math_SYMBOL}', "");
Expect(0, 126706, '\P{^General_Category= 	Math_SYMBOL}', "");
Error('\p{Gc=	Sm:=}');
Error('\P{Gc=	Sm:=}');
Expect(1, 126705, '\p{Gc=sm}', "");
Expect(0, 126705, '\p{^Gc=sm}', "");
Expect(0, 126705, '\P{Gc=sm}', "");
Expect(1, 126705, '\P{^Gc=sm}', "");
Expect(0, 126706, '\p{Gc=sm}', "");
Expect(1, 126706, '\p{^Gc=sm}', "");
Expect(1, 126706, '\P{Gc=sm}', "");
Expect(0, 126706, '\P{^Gc=sm}', "");
Expect(1, 126705, '\p{Gc=	Sm}', "");
Expect(0, 126705, '\p{^Gc=	Sm}', "");
Expect(0, 126705, '\P{Gc=	Sm}', "");
Expect(1, 126705, '\P{^Gc=	Sm}', "");
Expect(0, 126706, '\p{Gc=	Sm}', "");
Expect(1, 126706, '\p{^Gc=	Sm}', "");
Expect(1, 126706, '\P{Gc=	Sm}', "");
Expect(0, 126706, '\P{^Gc=	Sm}', "");
Error('\p{Category=:=__math_Symbol}');
Error('\P{Category=:=__math_Symbol}');
Expect(1, 126705, '\p{Category: mathsymbol}', "");
Expect(0, 126705, '\p{^Category: mathsymbol}', "");
Expect(0, 126705, '\P{Category: mathsymbol}', "");
Expect(1, 126705, '\P{^Category: mathsymbol}', "");
Expect(0, 126706, '\p{Category: mathsymbol}', "");
Expect(1, 126706, '\p{^Category: mathsymbol}', "");
Expect(1, 126706, '\P{Category: mathsymbol}', "");
Expect(0, 126706, '\P{^Category: mathsymbol}', "");
Expect(1, 126705, '\p{Category= math_SYMBOL}', "");
Expect(0, 126705, '\p{^Category= math_SYMBOL}', "");
Expect(0, 126705, '\P{Category= math_SYMBOL}', "");
Expect(1, 126705, '\P{^Category= math_SYMBOL}', "");
Expect(0, 126706, '\p{Category= math_SYMBOL}', "");
Expect(1, 126706, '\p{^Category= math_SYMBOL}', "");
Expect(1, 126706, '\P{Category= math_SYMBOL}', "");
Expect(0, 126706, '\P{^Category= math_SYMBOL}', "");
Error('\p{Is_General_Category= :=SM}');
Error('\P{Is_General_Category= :=SM}');
Expect(1, 126705, '\p{Is_General_Category=sm}', "");
Expect(0, 126705, '\p{^Is_General_Category=sm}', "");
Expect(0, 126705, '\P{Is_General_Category=sm}', "");
Expect(1, 126705, '\P{^Is_General_Category=sm}', "");
Expect(0, 126706, '\p{Is_General_Category=sm}', "");
Expect(1, 126706, '\p{^Is_General_Category=sm}', "");
Expect(1, 126706, '\P{Is_General_Category=sm}', "");
Expect(0, 126706, '\P{^Is_General_Category=sm}', "");
Expect(1, 126705, '\p{Is_General_Category=_ Sm}', "");
Expect(0, 126705, '\p{^Is_General_Category=_ Sm}', "");
Expect(0, 126705, '\P{Is_General_Category=_ Sm}', "");
Expect(1, 126705, '\P{^Is_General_Category=_ Sm}', "");
Expect(0, 126706, '\p{Is_General_Category=_ Sm}', "");
Expect(1, 126706, '\p{^Is_General_Category=_ Sm}', "");
Expect(1, 126706, '\P{Is_General_Category=_ Sm}', "");
Expect(0, 126706, '\P{^Is_General_Category=_ Sm}', "");
Error('\p{Is_Gc=/a/	 Math_symbol}');
Error('\P{Is_Gc=/a/	 Math_symbol}');
Expect(1, 126705, '\p{Is_Gc=mathsymbol}', "");
Expect(0, 126705, '\p{^Is_Gc=mathsymbol}', "");
Expect(0, 126705, '\P{Is_Gc=mathsymbol}', "");
Expect(1, 126705, '\P{^Is_Gc=mathsymbol}', "");
Expect(0, 126706, '\p{Is_Gc=mathsymbol}', "");
Expect(1, 126706, '\p{^Is_Gc=mathsymbol}', "");
Expect(1, 126706, '\P{Is_Gc=mathsymbol}', "");
Expect(0, 126706, '\P{^Is_Gc=mathsymbol}', "");
Expect(1, 126705, '\p{Is_Gc=-Math_SYMBOL}', "");
Expect(0, 126705, '\p{^Is_Gc=-Math_SYMBOL}', "");
Expect(0, 126705, '\P{Is_Gc=-Math_SYMBOL}', "");
Expect(1, 126705, '\P{^Is_Gc=-Math_SYMBOL}', "");
Expect(0, 126706, '\p{Is_Gc=-Math_SYMBOL}', "");
Expect(1, 126706, '\p{^Is_Gc=-Math_SYMBOL}', "");
Expect(1, 126706, '\P{Is_Gc=-Math_SYMBOL}', "");
Expect(0, 126706, '\P{^Is_Gc=-Math_SYMBOL}', "");
Error('\p{Is_Category=_/a/SM}');
Error('\P{Is_Category=_/a/SM}');
Expect(1, 126705, '\p{Is_Category=sm}', "");
Expect(0, 126705, '\p{^Is_Category=sm}', "");
Expect(0, 126705, '\P{Is_Category=sm}', "");
Expect(1, 126705, '\P{^Is_Category=sm}', "");
Expect(0, 126706, '\p{Is_Category=sm}', "");
Expect(1, 126706, '\p{^Is_Category=sm}', "");
Expect(1, 126706, '\P{Is_Category=sm}', "");
Expect(0, 126706, '\P{^Is_Category=sm}', "");
Expect(1, 126705, '\p{Is_Category=_Sm}', "");
Expect(0, 126705, '\p{^Is_Category=_Sm}', "");
Expect(0, 126705, '\P{Is_Category=_Sm}', "");
Expect(1, 126705, '\P{^Is_Category=_Sm}', "");
Expect(0, 126706, '\p{Is_Category=_Sm}', "");
Expect(1, 126706, '\p{^Is_Category=_Sm}', "");
Expect(1, 126706, '\P{Is_Category=_Sm}', "");
Expect(0, 126706, '\P{^Is_Category=_Sm}', "");
Error('\p{General_Category=-	Other_Symbol/a/}');
Error('\P{General_Category=-	Other_Symbol/a/}');
Expect(1, 128883, '\p{General_Category=othersymbol}', "");
Expect(0, 128883, '\p{^General_Category=othersymbol}', "");
Expect(0, 128883, '\P{General_Category=othersymbol}', "");
Expect(1, 128883, '\P{^General_Category=othersymbol}', "");
Expect(0, 128884, '\p{General_Category=othersymbol}', "");
Expect(1, 128884, '\p{^General_Category=othersymbol}', "");
Expect(1, 128884, '\P{General_Category=othersymbol}', "");
Expect(0, 128884, '\P{^General_Category=othersymbol}', "");
Expect(1, 128883, '\p{General_Category: 	other_symbol}', "");
Expect(0, 128883, '\p{^General_Category: 	other_symbol}', "");
Expect(0, 128883, '\P{General_Category: 	other_symbol}', "");
Expect(1, 128883, '\P{^General_Category: 	other_symbol}', "");
Expect(0, 128884, '\p{General_Category: 	other_symbol}', "");
Expect(1, 128884, '\p{^General_Category: 	other_symbol}', "");
Expect(1, 128884, '\P{General_Category: 	other_symbol}', "");
Expect(0, 128884, '\P{^General_Category: 	other_symbol}', "");
Error('\p{Gc:	_/a/So}');
Error('\P{Gc:	_/a/So}');
Expect(1, 128883, '\p{Gc=so}', "");
Expect(0, 128883, '\p{^Gc=so}', "");
Expect(0, 128883, '\P{Gc=so}', "");
Expect(1, 128883, '\P{^Gc=so}', "");
Expect(0, 128884, '\p{Gc=so}', "");
Expect(1, 128884, '\p{^Gc=so}', "");
Expect(1, 128884, '\P{Gc=so}', "");
Expect(0, 128884, '\P{^Gc=so}', "");
Expect(1, 128883, '\p{Gc=-So}', "");
Expect(0, 128883, '\p{^Gc=-So}', "");
Expect(0, 128883, '\P{Gc=-So}', "");
Expect(1, 128883, '\P{^Gc=-So}', "");
Expect(0, 128884, '\p{Gc=-So}', "");
Expect(1, 128884, '\p{^Gc=-So}', "");
Expect(1, 128884, '\P{Gc=-So}', "");
Expect(0, 128884, '\P{^Gc=-So}', "");
Error('\p{Category=  other_symbol/a/}');
Error('\P{Category=  other_symbol/a/}');
Expect(1, 128883, '\p{Category=othersymbol}', "");
Expect(0, 128883, '\p{^Category=othersymbol}', "");
Expect(0, 128883, '\P{Category=othersymbol}', "");
Expect(1, 128883, '\P{^Category=othersymbol}', "");
Expect(0, 128884, '\p{Category=othersymbol}', "");
Expect(1, 128884, '\p{^Category=othersymbol}', "");
Expect(1, 128884, '\P{Category=othersymbol}', "");
Expect(0, 128884, '\P{^Category=othersymbol}', "");
Expect(1, 128883, '\p{Category=__other_SYMBOL}', "");
Expect(0, 128883, '\p{^Category=__other_SYMBOL}', "");
Expect(0, 128883, '\P{Category=__other_SYMBOL}', "");
Expect(1, 128883, '\P{^Category=__other_SYMBOL}', "");
Expect(0, 128884, '\p{Category=__other_SYMBOL}', "");
Expect(1, 128884, '\p{^Category=__other_SYMBOL}', "");
Expect(1, 128884, '\P{Category=__other_SYMBOL}', "");
Expect(0, 128884, '\P{^Category=__other_SYMBOL}', "");
Error('\p{Is_General_Category=-SO/a/}');
Error('\P{Is_General_Category=-SO/a/}');
Expect(1, 128883, '\p{Is_General_Category=so}', "");
Expect(0, 128883, '\p{^Is_General_Category=so}', "");
Expect(0, 128883, '\P{Is_General_Category=so}', "");
Expect(1, 128883, '\P{^Is_General_Category=so}', "");
Expect(0, 128884, '\p{Is_General_Category=so}', "");
Expect(1, 128884, '\p{^Is_General_Category=so}', "");
Expect(1, 128884, '\P{Is_General_Category=so}', "");
Expect(0, 128884, '\P{^Is_General_Category=so}', "");
Expect(1, 128883, '\p{Is_General_Category= so}', "");
Expect(0, 128883, '\p{^Is_General_Category= so}', "");
Expect(0, 128883, '\P{Is_General_Category= so}', "");
Expect(1, 128883, '\P{^Is_General_Category= so}', "");
Expect(0, 128884, '\p{Is_General_Category= so}', "");
Expect(1, 128884, '\p{^Is_General_Category= so}', "");
Expect(1, 128884, '\P{Is_General_Category= so}', "");
Expect(0, 128884, '\P{^Is_General_Category= so}', "");
Error('\p{Is_Gc:-/a/OTHER_Symbol}');
Error('\P{Is_Gc:-/a/OTHER_Symbol}');
Expect(1, 128883, '\p{Is_Gc:othersymbol}', "");
Expect(0, 128883, '\p{^Is_Gc:othersymbol}', "");
Expect(0, 128883, '\P{Is_Gc:othersymbol}', "");
Expect(1, 128883, '\P{^Is_Gc:othersymbol}', "");
Expect(0, 128884, '\p{Is_Gc:othersymbol}', "");
Expect(1, 128884, '\p{^Is_Gc:othersymbol}', "");
Expect(1, 128884, '\P{Is_Gc:othersymbol}', "");
Expect(0, 128884, '\P{^Is_Gc:othersymbol}', "");
Expect(1, 128883, '\p{Is_Gc=--Other_Symbol}', "");
Expect(0, 128883, '\p{^Is_Gc=--Other_Symbol}', "");
Expect(0, 128883, '\P{Is_Gc=--Other_Symbol}', "");
Expect(1, 128883, '\P{^Is_Gc=--Other_Symbol}', "");
Expect(0, 128884, '\p{Is_Gc=--Other_Symbol}', "");
Expect(1, 128884, '\p{^Is_Gc=--Other_Symbol}', "");
Expect(1, 128884, '\P{Is_Gc=--Other_Symbol}', "");
Expect(0, 128884, '\P{^Is_Gc=--Other_Symbol}', "");
Error('\p{Is_Category=:=so}');
Error('\P{Is_Category=:=so}');
Expect(1, 128883, '\p{Is_Category:so}', "");
Expect(0, 128883, '\p{^Is_Category:so}', "");
Expect(0, 128883, '\P{Is_Category:so}', "");
Expect(1, 128883, '\P{^Is_Category:so}', "");
Expect(0, 128884, '\p{Is_Category:so}', "");
Expect(1, 128884, '\p{^Is_Category:so}', "");
Expect(1, 128884, '\P{Is_Category:so}', "");
Expect(0, 128884, '\P{^Is_Category:so}', "");
Expect(1, 128883, '\p{Is_Category= -SO}', "");
Expect(0, 128883, '\p{^Is_Category= -SO}', "");
Expect(0, 128883, '\P{Is_Category= -SO}', "");
Expect(1, 128883, '\P{^Is_Category= -SO}', "");
Expect(0, 128884, '\p{Is_Category= -SO}', "");
Expect(1, 128884, '\p{^Is_Category= -SO}', "");
Expect(1, 128884, '\P{Is_Category= -SO}', "");
Expect(0, 128884, '\P{^Is_Category= -SO}', "");
Error('\p{General_Category=:=Separator}');
Error('\P{General_Category=:=Separator}');
Expect(1, 12288, '\p{General_Category=separator}', "");
Expect(0, 12288, '\p{^General_Category=separator}', "");
Expect(0, 12288, '\P{General_Category=separator}', "");
Expect(1, 12288, '\P{^General_Category=separator}', "");
Expect(0, 12289, '\p{General_Category=separator}', "");
Expect(1, 12289, '\p{^General_Category=separator}', "");
Expect(1, 12289, '\P{General_Category=separator}', "");
Expect(0, 12289, '\P{^General_Category=separator}', "");
Expect(1, 12288, '\p{General_Category=	-Separator}', "");
Expect(0, 12288, '\p{^General_Category=	-Separator}', "");
Expect(0, 12288, '\P{General_Category=	-Separator}', "");
Expect(1, 12288, '\P{^General_Category=	-Separator}', "");
Expect(0, 12289, '\p{General_Category=	-Separator}', "");
Expect(1, 12289, '\p{^General_Category=	-Separator}', "");
Expect(1, 12289, '\P{General_Category=	-Separator}', "");
Expect(0, 12289, '\P{^General_Category=	-Separator}', "");
Error('\p{Gc:	 -z:=}');
Error('\P{Gc:	 -z:=}');
Expect(1, 12288, '\p{Gc=z}', "");
Expect(0, 12288, '\p{^Gc=z}', "");
Expect(0, 12288, '\P{Gc=z}', "");
Expect(1, 12288, '\P{^Gc=z}', "");
Expect(0, 12289, '\p{Gc=z}', "");
Expect(1, 12289, '\p{^Gc=z}', "");
Expect(1, 12289, '\P{Gc=z}', "");
Expect(0, 12289, '\P{^Gc=z}', "");
Expect(1, 12288, '\p{Gc= _Z}', "");
Expect(0, 12288, '\p{^Gc= _Z}', "");
Expect(0, 12288, '\P{Gc= _Z}', "");
Expect(1, 12288, '\P{^Gc= _Z}', "");
Expect(0, 12289, '\p{Gc= _Z}', "");
Expect(1, 12289, '\p{^Gc= _Z}', "");
Expect(1, 12289, '\P{Gc= _Z}', "");
Expect(0, 12289, '\P{^Gc= _Z}', "");
Error('\p{Category=:=Separator}');
Error('\P{Category=:=Separator}');
Expect(1, 12288, '\p{Category=separator}', "");
Expect(0, 12288, '\p{^Category=separator}', "");
Expect(0, 12288, '\P{Category=separator}', "");
Expect(1, 12288, '\P{^Category=separator}', "");
Expect(0, 12289, '\p{Category=separator}', "");
Expect(1, 12289, '\p{^Category=separator}', "");
Expect(1, 12289, '\P{Category=separator}', "");
Expect(0, 12289, '\P{^Category=separator}', "");
Expect(1, 12288, '\p{Category=_Separator}', "");
Expect(0, 12288, '\p{^Category=_Separator}', "");
Expect(0, 12288, '\P{Category=_Separator}', "");
Expect(1, 12288, '\P{^Category=_Separator}', "");
Expect(0, 12289, '\p{Category=_Separator}', "");
Expect(1, 12289, '\p{^Category=_Separator}', "");
Expect(1, 12289, '\P{Category=_Separator}', "");
Expect(0, 12289, '\P{^Category=_Separator}', "");
Error('\p{Is_General_Category=/a/--Z}');
Error('\P{Is_General_Category=/a/--Z}');
Expect(1, 12288, '\p{Is_General_Category:z}', "");
Expect(0, 12288, '\p{^Is_General_Category:z}', "");
Expect(0, 12288, '\P{Is_General_Category:z}', "");
Expect(1, 12288, '\P{^Is_General_Category:z}', "");
Expect(0, 12289, '\p{Is_General_Category:z}', "");
Expect(1, 12289, '\p{^Is_General_Category:z}', "");
Expect(1, 12289, '\P{Is_General_Category:z}', "");
Expect(0, 12289, '\P{^Is_General_Category:z}', "");
Expect(1, 12288, '\p{Is_General_Category=-Z}', "");
Expect(0, 12288, '\p{^Is_General_Category=-Z}', "");
Expect(0, 12288, '\P{Is_General_Category=-Z}', "");
Expect(1, 12288, '\P{^Is_General_Category=-Z}', "");
Expect(0, 12289, '\p{Is_General_Category=-Z}', "");
Expect(1, 12289, '\p{^Is_General_Category=-Z}', "");
Expect(1, 12289, '\P{Is_General_Category=-Z}', "");
Expect(0, 12289, '\P{^Is_General_Category=-Z}', "");
Error('\p{Is_Gc=_:=Separator}');
Error('\P{Is_Gc=_:=Separator}');
Expect(1, 12288, '\p{Is_Gc=separator}', "");
Expect(0, 12288, '\p{^Is_Gc=separator}', "");
Expect(0, 12288, '\P{Is_Gc=separator}', "");
Expect(1, 12288, '\P{^Is_Gc=separator}', "");
Expect(0, 12289, '\p{Is_Gc=separator}', "");
Expect(1, 12289, '\p{^Is_Gc=separator}', "");
Expect(1, 12289, '\P{Is_Gc=separator}', "");
Expect(0, 12289, '\P{^Is_Gc=separator}', "");
Expect(1, 12288, '\p{Is_Gc= 	separator}', "");
Expect(0, 12288, '\p{^Is_Gc= 	separator}', "");
Expect(0, 12288, '\P{Is_Gc= 	separator}', "");
Expect(1, 12288, '\P{^Is_Gc= 	separator}', "");
Expect(0, 12289, '\p{Is_Gc= 	separator}', "");
Expect(1, 12289, '\p{^Is_Gc= 	separator}', "");
Expect(1, 12289, '\P{Is_Gc= 	separator}', "");
Expect(0, 12289, '\P{^Is_Gc= 	separator}', "");
Error('\p{Is_Category:    :=z}');
Error('\P{Is_Category:    :=z}');
Expect(1, 12288, '\p{Is_Category: z}', "");
Expect(0, 12288, '\p{^Is_Category: z}', "");
Expect(0, 12288, '\P{Is_Category: z}', "");
Expect(1, 12288, '\P{^Is_Category: z}', "");
Expect(0, 12289, '\p{Is_Category: z}', "");
Expect(1, 12289, '\p{^Is_Category: z}', "");
Expect(1, 12289, '\P{Is_Category: z}', "");
Expect(0, 12289, '\P{^Is_Category: z}', "");
Expect(1, 12288, '\p{Is_Category=-Z}', "");
Expect(0, 12288, '\p{^Is_Category=-Z}', "");
Expect(0, 12288, '\P{Is_Category=-Z}', "");
Expect(1, 12288, '\P{^Is_Category=-Z}', "");
Expect(0, 12289, '\p{Is_Category=-Z}', "");
Expect(1, 12289, '\p{^Is_Category=-Z}', "");
Expect(1, 12289, '\P{Is_Category=-Z}', "");
Expect(0, 12289, '\P{^Is_Category=-Z}', "");
Error('\p{General_Category=_:=line_Separator}');
Error('\P{General_Category=_:=line_Separator}');
Expect(1, 8232, '\p{General_Category=lineseparator}', "");
Expect(0, 8232, '\p{^General_Category=lineseparator}', "");
Expect(0, 8232, '\P{General_Category=lineseparator}', "");
Expect(1, 8232, '\P{^General_Category=lineseparator}', "");
Expect(0, 8233, '\p{General_Category=lineseparator}', "");
Expect(1, 8233, '\p{^General_Category=lineseparator}', "");
Expect(1, 8233, '\P{General_Category=lineseparator}', "");
Expect(0, 8233, '\P{^General_Category=lineseparator}', "");
Expect(1, 8232, '\p{General_Category=	_Line_Separator}', "");
Expect(0, 8232, '\p{^General_Category=	_Line_Separator}', "");
Expect(0, 8232, '\P{General_Category=	_Line_Separator}', "");
Expect(1, 8232, '\P{^General_Category=	_Line_Separator}', "");
Expect(0, 8233, '\p{General_Category=	_Line_Separator}', "");
Expect(1, 8233, '\p{^General_Category=	_Line_Separator}', "");
Expect(1, 8233, '\P{General_Category=	_Line_Separator}', "");
Expect(0, 8233, '\P{^General_Category=	_Line_Separator}', "");
Error('\p{Gc=:= zl}');
Error('\P{Gc=:= zl}');
Expect(1, 8232, '\p{Gc=zl}', "");
Expect(0, 8232, '\p{^Gc=zl}', "");
Expect(0, 8232, '\P{Gc=zl}', "");
Expect(1, 8232, '\P{^Gc=zl}', "");
Expect(0, 8233, '\p{Gc=zl}', "");
Expect(1, 8233, '\p{^Gc=zl}', "");
Expect(1, 8233, '\P{Gc=zl}', "");
Expect(0, 8233, '\P{^Gc=zl}', "");
Expect(1, 8232, '\p{Gc= -zl}', "");
Expect(0, 8232, '\p{^Gc= -zl}', "");
Expect(0, 8232, '\P{Gc= -zl}', "");
Expect(1, 8232, '\P{^Gc= -zl}', "");
Expect(0, 8233, '\p{Gc= -zl}', "");
Expect(1, 8233, '\p{^Gc= -zl}', "");
Expect(1, 8233, '\P{Gc= -zl}', "");
Expect(0, 8233, '\P{^Gc= -zl}', "");
Error('\p{Category=-	LINE_separator/a/}');
Error('\P{Category=-	LINE_separator/a/}');
Expect(1, 8232, '\p{Category=lineseparator}', "");
Expect(0, 8232, '\p{^Category=lineseparator}', "");
Expect(0, 8232, '\P{Category=lineseparator}', "");
Expect(1, 8232, '\P{^Category=lineseparator}', "");
Expect(0, 8233, '\p{Category=lineseparator}', "");
Expect(1, 8233, '\p{^Category=lineseparator}', "");
Expect(1, 8233, '\P{Category=lineseparator}', "");
Expect(0, 8233, '\P{^Category=lineseparator}', "");
Expect(1, 8232, '\p{Category=--Line_SEPARATOR}', "");
Expect(0, 8232, '\p{^Category=--Line_SEPARATOR}', "");
Expect(0, 8232, '\P{Category=--Line_SEPARATOR}', "");
Expect(1, 8232, '\P{^Category=--Line_SEPARATOR}', "");
Expect(0, 8233, '\p{Category=--Line_SEPARATOR}', "");
Expect(1, 8233, '\p{^Category=--Line_SEPARATOR}', "");
Expect(1, 8233, '\P{Category=--Line_SEPARATOR}', "");
Expect(0, 8233, '\P{^Category=--Line_SEPARATOR}', "");
Error('\p{Is_General_Category=/a/zl}');
Error('\P{Is_General_Category=/a/zl}');
Expect(1, 8232, '\p{Is_General_Category: zl}', "");
Expect(0, 8232, '\p{^Is_General_Category: zl}', "");
Expect(0, 8232, '\P{Is_General_Category: zl}', "");
Expect(1, 8232, '\P{^Is_General_Category: zl}', "");
Expect(0, 8233, '\p{Is_General_Category: zl}', "");
Expect(1, 8233, '\p{^Is_General_Category: zl}', "");
Expect(1, 8233, '\P{Is_General_Category: zl}', "");
Expect(0, 8233, '\P{^Is_General_Category: zl}', "");
Expect(1, 8232, '\p{Is_General_Category=_Zl}', "");
Expect(0, 8232, '\p{^Is_General_Category=_Zl}', "");
Expect(0, 8232, '\P{Is_General_Category=_Zl}', "");
Expect(1, 8232, '\P{^Is_General_Category=_Zl}', "");
Expect(0, 8233, '\p{Is_General_Category=_Zl}', "");
Expect(1, 8233, '\p{^Is_General_Category=_Zl}', "");
Expect(1, 8233, '\P{Is_General_Category=_Zl}', "");
Expect(0, 8233, '\P{^Is_General_Category=_Zl}', "");
Error('\p{Is_Gc=:=- Line_Separator}');
Error('\P{Is_Gc=:=- Line_Separator}');
Expect(1, 8232, '\p{Is_Gc=lineseparator}', "");
Expect(0, 8232, '\p{^Is_Gc=lineseparator}', "");
Expect(0, 8232, '\P{Is_Gc=lineseparator}', "");
Expect(1, 8232, '\P{^Is_Gc=lineseparator}', "");
Expect(0, 8233, '\p{Is_Gc=lineseparator}', "");
Expect(1, 8233, '\p{^Is_Gc=lineseparator}', "");
Expect(1, 8233, '\P{Is_Gc=lineseparator}', "");
Expect(0, 8233, '\P{^Is_Gc=lineseparator}', "");
Expect(1, 8232, '\p{Is_Gc=-	Line_Separator}', "");
Expect(0, 8232, '\p{^Is_Gc=-	Line_Separator}', "");
Expect(0, 8232, '\P{Is_Gc=-	Line_Separator}', "");
Expect(1, 8232, '\P{^Is_Gc=-	Line_Separator}', "");
Expect(0, 8233, '\p{Is_Gc=-	Line_Separator}', "");
Expect(1, 8233, '\p{^Is_Gc=-	Line_Separator}', "");
Expect(1, 8233, '\P{Is_Gc=-	Line_Separator}', "");
Expect(0, 8233, '\P{^Is_Gc=-	Line_Separator}', "");
Error('\p{Is_Category=-:=ZL}');
Error('\P{Is_Category=-:=ZL}');
Expect(1, 8232, '\p{Is_Category=zl}', "");
Expect(0, 8232, '\p{^Is_Category=zl}', "");
Expect(0, 8232, '\P{Is_Category=zl}', "");
Expect(1, 8232, '\P{^Is_Category=zl}', "");
Expect(0, 8233, '\p{Is_Category=zl}', "");
Expect(1, 8233, '\p{^Is_Category=zl}', "");
Expect(1, 8233, '\P{Is_Category=zl}', "");
Expect(0, 8233, '\P{^Is_Category=zl}', "");
Expect(1, 8232, '\p{Is_Category=	Zl}', "");
Expect(0, 8232, '\p{^Is_Category=	Zl}', "");
Expect(0, 8232, '\P{Is_Category=	Zl}', "");
Expect(1, 8232, '\P{^Is_Category=	Zl}', "");
Expect(0, 8233, '\p{Is_Category=	Zl}', "");
Expect(1, 8233, '\p{^Is_Category=	Zl}', "");
Expect(1, 8233, '\P{Is_Category=	Zl}', "");
Expect(0, 8233, '\P{^Is_Category=	Zl}', "");
Error('\p{General_Category=_/a/Paragraph_Separator}');
Error('\P{General_Category=_/a/Paragraph_Separator}');
Expect(1, 8233, '\p{General_Category=paragraphseparator}', "");
Expect(0, 8233, '\p{^General_Category=paragraphseparator}', "");
Expect(0, 8233, '\P{General_Category=paragraphseparator}', "");
Expect(1, 8233, '\P{^General_Category=paragraphseparator}', "");
Expect(0, 8234, '\p{General_Category=paragraphseparator}', "");
Expect(1, 8234, '\p{^General_Category=paragraphseparator}', "");
Expect(1, 8234, '\P{General_Category=paragraphseparator}', "");
Expect(0, 8234, '\P{^General_Category=paragraphseparator}', "");
Expect(1, 8233, '\p{General_Category=_Paragraph_Separator}', "");
Expect(0, 8233, '\p{^General_Category=_Paragraph_Separator}', "");
Expect(0, 8233, '\P{General_Category=_Paragraph_Separator}', "");
Expect(1, 8233, '\P{^General_Category=_Paragraph_Separator}', "");
Expect(0, 8234, '\p{General_Category=_Paragraph_Separator}', "");
Expect(1, 8234, '\p{^General_Category=_Paragraph_Separator}', "");
Expect(1, 8234, '\P{General_Category=_Paragraph_Separator}', "");
Expect(0, 8234, '\P{^General_Category=_Paragraph_Separator}', "");
Error('\p{Gc=/a/ Zp}');
Error('\P{Gc=/a/ Zp}');
Expect(1, 8233, '\p{Gc=zp}', "");
Expect(0, 8233, '\p{^Gc=zp}', "");
Expect(0, 8233, '\P{Gc=zp}', "");
Expect(1, 8233, '\P{^Gc=zp}', "");
Expect(0, 8234, '\p{Gc=zp}', "");
Expect(1, 8234, '\p{^Gc=zp}', "");
Expect(1, 8234, '\P{Gc=zp}', "");
Expect(0, 8234, '\P{^Gc=zp}', "");
Expect(1, 8233, '\p{Gc=-	Zp}', "");
Expect(0, 8233, '\p{^Gc=-	Zp}', "");
Expect(0, 8233, '\P{Gc=-	Zp}', "");
Expect(1, 8233, '\P{^Gc=-	Zp}', "");
Expect(0, 8234, '\p{Gc=-	Zp}', "");
Expect(1, 8234, '\p{^Gc=-	Zp}', "");
Expect(1, 8234, '\P{Gc=-	Zp}', "");
Expect(0, 8234, '\P{^Gc=-	Zp}', "");
Error('\p{Category=__PARAGRAPH_SEPARATOR/a/}');
Error('\P{Category=__PARAGRAPH_SEPARATOR/a/}');
Expect(1, 8233, '\p{Category=paragraphseparator}', "");
Expect(0, 8233, '\p{^Category=paragraphseparator}', "");
Expect(0, 8233, '\P{Category=paragraphseparator}', "");
Expect(1, 8233, '\P{^Category=paragraphseparator}', "");
Expect(0, 8234, '\p{Category=paragraphseparator}', "");
Expect(1, 8234, '\p{^Category=paragraphseparator}', "");
Expect(1, 8234, '\P{Category=paragraphseparator}', "");
Expect(0, 8234, '\P{^Category=paragraphseparator}', "");
Expect(1, 8233, '\p{Category=__PARAGRAPH_separator}', "");
Expect(0, 8233, '\p{^Category=__PARAGRAPH_separator}', "");
Expect(0, 8233, '\P{Category=__PARAGRAPH_separator}', "");
Expect(1, 8233, '\P{^Category=__PARAGRAPH_separator}', "");
Expect(0, 8234, '\p{Category=__PARAGRAPH_separator}', "");
Expect(1, 8234, '\p{^Category=__PARAGRAPH_separator}', "");
Expect(1, 8234, '\P{Category=__PARAGRAPH_separator}', "");
Expect(0, 8234, '\P{^Category=__PARAGRAPH_separator}', "");
Error('\p{Is_General_Category:	 /a/Zp}');
Error('\P{Is_General_Category:	 /a/Zp}');
Expect(1, 8233, '\p{Is_General_Category=zp}', "");
Expect(0, 8233, '\p{^Is_General_Category=zp}', "");
Expect(0, 8233, '\P{Is_General_Category=zp}', "");
Expect(1, 8233, '\P{^Is_General_Category=zp}', "");
Expect(0, 8234, '\p{Is_General_Category=zp}', "");
Expect(1, 8234, '\p{^Is_General_Category=zp}', "");
Expect(1, 8234, '\P{Is_General_Category=zp}', "");
Expect(0, 8234, '\P{^Is_General_Category=zp}', "");
Expect(1, 8233, '\p{Is_General_Category=	zp}', "");
Expect(0, 8233, '\p{^Is_General_Category=	zp}', "");
Expect(0, 8233, '\P{Is_General_Category=	zp}', "");
Expect(1, 8233, '\P{^Is_General_Category=	zp}', "");
Expect(0, 8234, '\p{Is_General_Category=	zp}', "");
Expect(1, 8234, '\p{^Is_General_Category=	zp}', "");
Expect(1, 8234, '\P{Is_General_Category=	zp}', "");
Expect(0, 8234, '\P{^Is_General_Category=	zp}', "");
Error('\p{Is_Gc= -paragraph_separator/a/}');
Error('\P{Is_Gc= -paragraph_separator/a/}');
Expect(1, 8233, '\p{Is_Gc=paragraphseparator}', "");
Expect(0, 8233, '\p{^Is_Gc=paragraphseparator}', "");
Expect(0, 8233, '\P{Is_Gc=paragraphseparator}', "");
Expect(1, 8233, '\P{^Is_Gc=paragraphseparator}', "");
Expect(0, 8234, '\p{Is_Gc=paragraphseparator}', "");
Expect(1, 8234, '\p{^Is_Gc=paragraphseparator}', "");
Expect(1, 8234, '\P{Is_Gc=paragraphseparator}', "");
Expect(0, 8234, '\P{^Is_Gc=paragraphseparator}', "");
Expect(1, 8233, '\p{Is_Gc=-Paragraph_separator}', "");
Expect(0, 8233, '\p{^Is_Gc=-Paragraph_separator}', "");
Expect(0, 8233, '\P{Is_Gc=-Paragraph_separator}', "");
Expect(1, 8233, '\P{^Is_Gc=-Paragraph_separator}', "");
Expect(0, 8234, '\p{Is_Gc=-Paragraph_separator}', "");
Expect(1, 8234, '\p{^Is_Gc=-Paragraph_separator}', "");
Expect(1, 8234, '\P{Is_Gc=-Paragraph_separator}', "");
Expect(0, 8234, '\P{^Is_Gc=-Paragraph_separator}', "");
Error('\p{Is_Category:	ZP:=}');
Error('\P{Is_Category:	ZP:=}');
Expect(1, 8233, '\p{Is_Category:   zp}', "");
Expect(0, 8233, '\p{^Is_Category:   zp}', "");
Expect(0, 8233, '\P{Is_Category:   zp}', "");
Expect(1, 8233, '\P{^Is_Category:   zp}', "");
Expect(0, 8234, '\p{Is_Category:   zp}', "");
Expect(1, 8234, '\p{^Is_Category:   zp}', "");
Expect(1, 8234, '\P{Is_Category:   zp}', "");
Expect(0, 8234, '\P{^Is_Category:   zp}', "");
Expect(1, 8233, '\p{Is_Category=_	Zp}', "");
Expect(0, 8233, '\p{^Is_Category=_	Zp}', "");
Expect(0, 8233, '\P{Is_Category=_	Zp}', "");
Expect(1, 8233, '\P{^Is_Category=_	Zp}', "");
Expect(0, 8234, '\p{Is_Category=_	Zp}', "");
Expect(1, 8234, '\p{^Is_Category=_	Zp}', "");
Expect(1, 8234, '\P{Is_Category=_	Zp}', "");
Expect(0, 8234, '\P{^Is_Category=_	Zp}', "");
Error('\p{General_Category=:=  Space_Separator}');
Error('\P{General_Category=:=  Space_Separator}');
Expect(1, 12288, '\p{General_Category=spaceseparator}', "");
Expect(0, 12288, '\p{^General_Category=spaceseparator}', "");
Expect(0, 12288, '\P{General_Category=spaceseparator}', "");
Expect(1, 12288, '\P{^General_Category=spaceseparator}', "");
Expect(0, 12289, '\p{General_Category=spaceseparator}', "");
Expect(1, 12289, '\p{^General_Category=spaceseparator}', "");
Expect(1, 12289, '\P{General_Category=spaceseparator}', "");
Expect(0, 12289, '\P{^General_Category=spaceseparator}', "");
Expect(1, 12288, '\p{General_Category=_ Space_SEPARATOR}', "");
Expect(0, 12288, '\p{^General_Category=_ Space_SEPARATOR}', "");
Expect(0, 12288, '\P{General_Category=_ Space_SEPARATOR}', "");
Expect(1, 12288, '\P{^General_Category=_ Space_SEPARATOR}', "");
Expect(0, 12289, '\p{General_Category=_ Space_SEPARATOR}', "");
Expect(1, 12289, '\p{^General_Category=_ Space_SEPARATOR}', "");
Expect(1, 12289, '\P{General_Category=_ Space_SEPARATOR}', "");
Expect(0, 12289, '\P{^General_Category=_ Space_SEPARATOR}', "");
Error('\p{Gc=/a/ Zs}');
Error('\P{Gc=/a/ Zs}');
Expect(1, 12288, '\p{Gc=zs}', "");
Expect(0, 12288, '\p{^Gc=zs}', "");
Expect(0, 12288, '\P{Gc=zs}', "");
Expect(1, 12288, '\P{^Gc=zs}', "");
Expect(0, 12289, '\p{Gc=zs}', "");
Expect(1, 12289, '\p{^Gc=zs}', "");
Expect(1, 12289, '\P{Gc=zs}', "");
Expect(0, 12289, '\P{^Gc=zs}', "");
Expect(1, 12288, '\p{Gc=-	Zs}', "");
Expect(0, 12288, '\p{^Gc=-	Zs}', "");
Expect(0, 12288, '\P{Gc=-	Zs}', "");
Expect(1, 12288, '\P{^Gc=-	Zs}', "");
Expect(0, 12289, '\p{Gc=-	Zs}', "");
Expect(1, 12289, '\p{^Gc=-	Zs}', "");
Expect(1, 12289, '\P{Gc=-	Zs}', "");
Expect(0, 12289, '\P{^Gc=-	Zs}', "");
Error('\p{Category=	-SPACE_Separator:=}');
Error('\P{Category=	-SPACE_Separator:=}');
Expect(1, 12288, '\p{Category=spaceseparator}', "");
Expect(0, 12288, '\p{^Category=spaceseparator}', "");
Expect(0, 12288, '\P{Category=spaceseparator}', "");
Expect(1, 12288, '\P{^Category=spaceseparator}', "");
Expect(0, 12289, '\p{Category=spaceseparator}', "");
Expect(1, 12289, '\p{^Category=spaceseparator}', "");
Expect(1, 12289, '\P{Category=spaceseparator}', "");
Expect(0, 12289, '\P{^Category=spaceseparator}', "");
Expect(1, 12288, '\p{Category=_ Space_Separator}', "");
Expect(0, 12288, '\p{^Category=_ Space_Separator}', "");
Expect(0, 12288, '\P{Category=_ Space_Separator}', "");
Expect(1, 12288, '\P{^Category=_ Space_Separator}', "");
Expect(0, 12289, '\p{Category=_ Space_Separator}', "");
Expect(1, 12289, '\p{^Category=_ Space_Separator}', "");
Expect(1, 12289, '\P{Category=_ Space_Separator}', "");
Expect(0, 12289, '\P{^Category=_ Space_Separator}', "");
Error('\p{Is_General_Category=/a/	 Zs}');
Error('\P{Is_General_Category=/a/	 Zs}');
Expect(1, 12288, '\p{Is_General_Category=zs}', "");
Expect(0, 12288, '\p{^Is_General_Category=zs}', "");
Expect(0, 12288, '\P{Is_General_Category=zs}', "");
Expect(1, 12288, '\P{^Is_General_Category=zs}', "");
Expect(0, 12289, '\p{Is_General_Category=zs}', "");
Expect(1, 12289, '\p{^Is_General_Category=zs}', "");
Expect(1, 12289, '\P{Is_General_Category=zs}', "");
Expect(0, 12289, '\P{^Is_General_Category=zs}', "");
Expect(1, 12288, '\p{Is_General_Category=__Zs}', "");
Expect(0, 12288, '\p{^Is_General_Category=__Zs}', "");
Expect(0, 12288, '\P{Is_General_Category=__Zs}', "");
Expect(1, 12288, '\P{^Is_General_Category=__Zs}', "");
Expect(0, 12289, '\p{Is_General_Category=__Zs}', "");
Expect(1, 12289, '\p{^Is_General_Category=__Zs}', "");
Expect(1, 12289, '\P{Is_General_Category=__Zs}', "");
Expect(0, 12289, '\P{^Is_General_Category=__Zs}', "");
Error('\p{Is_Gc=	:=SPACE_Separator}');
Error('\P{Is_Gc=	:=SPACE_Separator}');
Expect(1, 12288, '\p{Is_Gc=spaceseparator}', "");
Expect(0, 12288, '\p{^Is_Gc=spaceseparator}', "");
Expect(0, 12288, '\P{Is_Gc=spaceseparator}', "");
Expect(1, 12288, '\P{^Is_Gc=spaceseparator}', "");
Expect(0, 12289, '\p{Is_Gc=spaceseparator}', "");
Expect(1, 12289, '\p{^Is_Gc=spaceseparator}', "");
Expect(1, 12289, '\P{Is_Gc=spaceseparator}', "");
Expect(0, 12289, '\P{^Is_Gc=spaceseparator}', "");
Expect(1, 12288, '\p{Is_Gc=	space_Separator}', "");
Expect(0, 12288, '\p{^Is_Gc=	space_Separator}', "");
Expect(0, 12288, '\P{Is_Gc=	space_Separator}', "");
Expect(1, 12288, '\P{^Is_Gc=	space_Separator}', "");
Expect(0, 12289, '\p{Is_Gc=	space_Separator}', "");
Expect(1, 12289, '\p{^Is_Gc=	space_Separator}', "");
Expect(1, 12289, '\P{Is_Gc=	space_Separator}', "");
Expect(0, 12289, '\P{^Is_Gc=	space_Separator}', "");
Error('\p{Is_Category=-_zs/a/}');
Error('\P{Is_Category=-_zs/a/}');
Expect(1, 12288, '\p{Is_Category=zs}', "");
Expect(0, 12288, '\p{^Is_Category=zs}', "");
Expect(0, 12288, '\P{Is_Category=zs}', "");
Expect(1, 12288, '\P{^Is_Category=zs}', "");
Expect(0, 12289, '\p{Is_Category=zs}', "");
Expect(1, 12289, '\p{^Is_Category=zs}', "");
Expect(1, 12289, '\P{Is_Category=zs}', "");
Expect(0, 12289, '\P{^Is_Category=zs}', "");
Expect(1, 12288, '\p{Is_Category=_ ZS}', "");
Expect(0, 12288, '\p{^Is_Category=_ ZS}', "");
Expect(0, 12288, '\P{Is_Category=_ ZS}', "");
Expect(1, 12288, '\P{^Is_Category=_ ZS}', "");
Expect(0, 12289, '\p{Is_Category=_ ZS}', "");
Expect(1, 12289, '\p{^Is_Category=_ ZS}', "");
Expect(1, 12289, '\P{Is_Category=_ ZS}', "");
Expect(0, 12289, '\P{^Is_Category=_ ZS}', "");
Error('\p{Hangul_Syllable_Type=/a/ Leading_Jamo}');
Error('\P{Hangul_Syllable_Type=/a/ Leading_Jamo}');
Expect(1, 43388, '\p{Hangul_Syllable_Type=leadingjamo}', "");
Expect(0, 43388, '\p{^Hangul_Syllable_Type=leadingjamo}', "");
Expect(0, 43388, '\P{Hangul_Syllable_Type=leadingjamo}', "");
Expect(1, 43388, '\P{^Hangul_Syllable_Type=leadingjamo}', "");
Expect(0, 43389, '\p{Hangul_Syllable_Type=leadingjamo}', "");
Expect(1, 43389, '\p{^Hangul_Syllable_Type=leadingjamo}', "");
Expect(1, 43389, '\P{Hangul_Syllable_Type=leadingjamo}', "");
Expect(0, 43389, '\P{^Hangul_Syllable_Type=leadingjamo}', "");
Expect(1, 43388, '\p{Hangul_Syllable_Type=-	Leading_Jamo}', "");
Expect(0, 43388, '\p{^Hangul_Syllable_Type=-	Leading_Jamo}', "");
Expect(0, 43388, '\P{Hangul_Syllable_Type=-	Leading_Jamo}', "");
Expect(1, 43388, '\P{^Hangul_Syllable_Type=-	Leading_Jamo}', "");
Expect(0, 43389, '\p{Hangul_Syllable_Type=-	Leading_Jamo}', "");
Expect(1, 43389, '\p{^Hangul_Syllable_Type=-	Leading_Jamo}', "");
Expect(1, 43389, '\P{Hangul_Syllable_Type=-	Leading_Jamo}', "");
Expect(0, 43389, '\P{^Hangul_Syllable_Type=-	Leading_Jamo}', "");
Error('\p{Hst:	-L/a/}');
Error('\P{Hst:	-L/a/}');
Expect(1, 43388, '\p{Hst:	l}', "");
Expect(0, 43388, '\p{^Hst:	l}', "");
Expect(0, 43388, '\P{Hst:	l}', "");
Expect(1, 43388, '\P{^Hst:	l}', "");
Expect(0, 43389, '\p{Hst:	l}', "");
Expect(1, 43389, '\p{^Hst:	l}', "");
Expect(1, 43389, '\P{Hst:	l}', "");
Expect(0, 43389, '\P{^Hst:	l}', "");
Expect(1, 43388, '\p{Hst=--L}', "");
Expect(0, 43388, '\p{^Hst=--L}', "");
Expect(0, 43388, '\P{Hst=--L}', "");
Expect(1, 43388, '\P{^Hst=--L}', "");
Expect(0, 43389, '\p{Hst=--L}', "");
Expect(1, 43389, '\p{^Hst=--L}', "");
Expect(1, 43389, '\P{Hst=--L}', "");
Expect(0, 43389, '\P{^Hst=--L}', "");
Error('\p{Is_Hangul_Syllable_Type=-Leading_JAMO:=}');
Error('\P{Is_Hangul_Syllable_Type=-Leading_JAMO:=}');
Expect(1, 43388, '\p{Is_Hangul_Syllable_Type=leadingjamo}', "");
Expect(0, 43388, '\p{^Is_Hangul_Syllable_Type=leadingjamo}', "");
Expect(0, 43388, '\P{Is_Hangul_Syllable_Type=leadingjamo}', "");
Expect(1, 43388, '\P{^Is_Hangul_Syllable_Type=leadingjamo}', "");
Expect(0, 43389, '\p{Is_Hangul_Syllable_Type=leadingjamo}', "");
Expect(1, 43389, '\p{^Is_Hangul_Syllable_Type=leadingjamo}', "");
Expect(1, 43389, '\P{Is_Hangul_Syllable_Type=leadingjamo}', "");
Expect(0, 43389, '\P{^Is_Hangul_Syllable_Type=leadingjamo}', "");
Expect(1, 43388, '\p{Is_Hangul_Syllable_Type=_Leading_Jamo}', "");
Expect(0, 43388, '\p{^Is_Hangul_Syllable_Type=_Leading_Jamo}', "");
Expect(0, 43388, '\P{Is_Hangul_Syllable_Type=_Leading_Jamo}', "");
Expect(1, 43388, '\P{^Is_Hangul_Syllable_Type=_Leading_Jamo}', "");
Expect(0, 43389, '\p{Is_Hangul_Syllable_Type=_Leading_Jamo}', "");
Expect(1, 43389, '\p{^Is_Hangul_Syllable_Type=_Leading_Jamo}', "");
Expect(1, 43389, '\P{Is_Hangul_Syllable_Type=_Leading_Jamo}', "");
Expect(0, 43389, '\P{^Is_Hangul_Syllable_Type=_Leading_Jamo}', "");
Error('\p{Is_Hst=_:=L}');
Error('\P{Is_Hst=_:=L}');
Expect(1, 43388, '\p{Is_Hst:   l}', "");
Expect(0, 43388, '\p{^Is_Hst:   l}', "");
Expect(0, 43388, '\P{Is_Hst:   l}', "");
Expect(1, 43388, '\P{^Is_Hst:   l}', "");
Expect(0, 43389, '\p{Is_Hst:   l}', "");
Expect(1, 43389, '\p{^Is_Hst:   l}', "");
Expect(1, 43389, '\P{Is_Hst:   l}', "");
Expect(0, 43389, '\P{^Is_Hst:   l}', "");
Expect(1, 43388, '\p{Is_Hst=-	L}', "");
Expect(0, 43388, '\p{^Is_Hst=-	L}', "");
Expect(0, 43388, '\P{Is_Hst=-	L}', "");
Expect(1, 43388, '\P{^Is_Hst=-	L}', "");
Expect(0, 43389, '\p{Is_Hst=-	L}', "");
Expect(1, 43389, '\p{^Is_Hst=-	L}', "");
Expect(1, 43389, '\P{Is_Hst=-	L}', "");
Expect(0, 43389, '\P{^Is_Hst=-	L}', "");
Error('\p{Hangul_Syllable_Type:	:= LV_SYLLABLE}');
Error('\P{Hangul_Syllable_Type:	:= LV_SYLLABLE}');
Expect(1, 55176, '\p{Hangul_Syllable_Type=lvsyllable}', "");
Expect(0, 55176, '\p{^Hangul_Syllable_Type=lvsyllable}', "");
Expect(0, 55176, '\P{Hangul_Syllable_Type=lvsyllable}', "");
Expect(1, 55176, '\P{^Hangul_Syllable_Type=lvsyllable}', "");
Expect(0, 55177, '\p{Hangul_Syllable_Type=lvsyllable}', "");
Expect(1, 55177, '\p{^Hangul_Syllable_Type=lvsyllable}', "");
Expect(1, 55177, '\P{Hangul_Syllable_Type=lvsyllable}', "");
Expect(0, 55177, '\P{^Hangul_Syllable_Type=lvsyllable}', "");
Expect(1, 55176, '\p{Hangul_Syllable_Type=- LV_syllable}', "");
Expect(0, 55176, '\p{^Hangul_Syllable_Type=- LV_syllable}', "");
Expect(0, 55176, '\P{Hangul_Syllable_Type=- LV_syllable}', "");
Expect(1, 55176, '\P{^Hangul_Syllable_Type=- LV_syllable}', "");
Expect(0, 55177, '\p{Hangul_Syllable_Type=- LV_syllable}', "");
Expect(1, 55177, '\p{^Hangul_Syllable_Type=- LV_syllable}', "");
Expect(1, 55177, '\P{Hangul_Syllable_Type=- LV_syllable}', "");
Expect(0, 55177, '\P{^Hangul_Syllable_Type=- LV_syllable}', "");
Error('\p{Hst=_:=LV}');
Error('\P{Hst=_:=LV}');
Expect(1, 55176, '\p{Hst:lv}', "");
Expect(0, 55176, '\p{^Hst:lv}', "");
Expect(0, 55176, '\P{Hst:lv}', "");
Expect(1, 55176, '\P{^Hst:lv}', "");
Expect(0, 55177, '\p{Hst:lv}', "");
Expect(1, 55177, '\p{^Hst:lv}', "");
Expect(1, 55177, '\P{Hst:lv}', "");
Expect(0, 55177, '\P{^Hst:lv}', "");
Expect(1, 55176, '\p{Hst:_LV}', "");
Expect(0, 55176, '\p{^Hst:_LV}', "");
Expect(0, 55176, '\P{Hst:_LV}', "");
Expect(1, 55176, '\P{^Hst:_LV}', "");
Expect(0, 55177, '\p{Hst:_LV}', "");
Expect(1, 55177, '\p{^Hst:_LV}', "");
Expect(1, 55177, '\P{Hst:_LV}', "");
Expect(0, 55177, '\P{^Hst:_LV}', "");
Error('\p{Is_Hangul_Syllable_Type= LV_Syllable:=}');
Error('\P{Is_Hangul_Syllable_Type= LV_Syllable:=}');
Expect(1, 55176, '\p{Is_Hangul_Syllable_Type: lvsyllable}', "");
Expect(0, 55176, '\p{^Is_Hangul_Syllable_Type: lvsyllable}', "");
Expect(0, 55176, '\P{Is_Hangul_Syllable_Type: lvsyllable}', "");
Expect(1, 55176, '\P{^Is_Hangul_Syllable_Type: lvsyllable}', "");
Expect(0, 55177, '\p{Is_Hangul_Syllable_Type: lvsyllable}', "");
Expect(1, 55177, '\p{^Is_Hangul_Syllable_Type: lvsyllable}', "");
Expect(1, 55177, '\P{Is_Hangul_Syllable_Type: lvsyllable}', "");
Expect(0, 55177, '\P{^Is_Hangul_Syllable_Type: lvsyllable}', "");
Expect(1, 55176, '\p{Is_Hangul_Syllable_Type=-_lv_SYLLABLE}', "");
Expect(0, 55176, '\p{^Is_Hangul_Syllable_Type=-_lv_SYLLABLE}', "");
Expect(0, 55176, '\P{Is_Hangul_Syllable_Type=-_lv_SYLLABLE}', "");
Expect(1, 55176, '\P{^Is_Hangul_Syllable_Type=-_lv_SYLLABLE}', "");
Expect(0, 55177, '\p{Is_Hangul_Syllable_Type=-_lv_SYLLABLE}', "");
Expect(1, 55177, '\p{^Is_Hangul_Syllable_Type=-_lv_SYLLABLE}', "");
Expect(1, 55177, '\P{Is_Hangul_Syllable_Type=-_lv_SYLLABLE}', "");
Expect(0, 55177, '\P{^Is_Hangul_Syllable_Type=-_lv_SYLLABLE}', "");
Error('\p{Is_Hst=:=	LV}');
Error('\P{Is_Hst=:=	LV}');
Expect(1, 55176, '\p{Is_Hst=lv}', "");
Expect(0, 55176, '\p{^Is_Hst=lv}', "");
Expect(0, 55176, '\P{Is_Hst=lv}', "");
Expect(1, 55176, '\P{^Is_Hst=lv}', "");
Expect(0, 55177, '\p{Is_Hst=lv}', "");
Expect(1, 55177, '\p{^Is_Hst=lv}', "");
Expect(1, 55177, '\P{Is_Hst=lv}', "");
Expect(0, 55177, '\P{^Is_Hst=lv}', "");
Expect(1, 55176, '\p{Is_Hst=	LV}', "");
Expect(0, 55176, '\p{^Is_Hst=	LV}', "");
Expect(0, 55176, '\P{Is_Hst=	LV}', "");
Expect(1, 55176, '\P{^Is_Hst=	LV}', "");
Expect(0, 55177, '\p{Is_Hst=	LV}', "");
Expect(1, 55177, '\p{^Is_Hst=	LV}', "");
Expect(1, 55177, '\P{Is_Hst=	LV}', "");
Expect(0, 55177, '\P{^Is_Hst=	LV}', "");
Error('\p{Hangul_Syllable_Type=/a/	LVT_SYLLABLE}');
Error('\P{Hangul_Syllable_Type=/a/	LVT_SYLLABLE}');
Expect(1, 55203, '\p{Hangul_Syllable_Type=lvtsyllable}', "");
Expect(0, 55203, '\p{^Hangul_Syllable_Type=lvtsyllable}', "");
Expect(0, 55203, '\P{Hangul_Syllable_Type=lvtsyllable}', "");
Expect(1, 55203, '\P{^Hangul_Syllable_Type=lvtsyllable}', "");
Expect(0, 55204, '\p{Hangul_Syllable_Type=lvtsyllable}', "");
Expect(1, 55204, '\p{^Hangul_Syllable_Type=lvtsyllable}', "");
Expect(1, 55204, '\P{Hangul_Syllable_Type=lvtsyllable}', "");
Expect(0, 55204, '\P{^Hangul_Syllable_Type=lvtsyllable}', "");
Expect(1, 55203, '\p{Hangul_Syllable_Type= -LVT_syllable}', "");
Expect(0, 55203, '\p{^Hangul_Syllable_Type= -LVT_syllable}', "");
Expect(0, 55203, '\P{Hangul_Syllable_Type= -LVT_syllable}', "");
Expect(1, 55203, '\P{^Hangul_Syllable_Type= -LVT_syllable}', "");
Expect(0, 55204, '\p{Hangul_Syllable_Type= -LVT_syllable}', "");
Expect(1, 55204, '\p{^Hangul_Syllable_Type= -LVT_syllable}', "");
Expect(1, 55204, '\P{Hangul_Syllable_Type= -LVT_syllable}', "");
Expect(0, 55204, '\P{^Hangul_Syllable_Type= -LVT_syllable}', "");
Error('\p{Hst=_/a/LVT}');
Error('\P{Hst=_/a/LVT}');
Expect(1, 55203, '\p{Hst=lvt}', "");
Expect(0, 55203, '\p{^Hst=lvt}', "");
Expect(0, 55203, '\P{Hst=lvt}', "");
Expect(1, 55203, '\P{^Hst=lvt}', "");
Expect(0, 55204, '\p{Hst=lvt}', "");
Expect(1, 55204, '\p{^Hst=lvt}', "");
Expect(1, 55204, '\P{Hst=lvt}', "");
Expect(0, 55204, '\P{^Hst=lvt}', "");
Expect(1, 55203, '\p{Hst=LVT}', "");
Expect(0, 55203, '\p{^Hst=LVT}', "");
Expect(0, 55203, '\P{Hst=LVT}', "");
Expect(1, 55203, '\P{^Hst=LVT}', "");
Expect(0, 55204, '\p{Hst=LVT}', "");
Expect(1, 55204, '\p{^Hst=LVT}', "");
Expect(1, 55204, '\P{Hst=LVT}', "");
Expect(0, 55204, '\P{^Hst=LVT}', "");
Error('\p{Is_Hangul_Syllable_Type=/a/_-LVT_syllable}');
Error('\P{Is_Hangul_Syllable_Type=/a/_-LVT_syllable}');
Expect(1, 55203, '\p{Is_Hangul_Syllable_Type=lvtsyllable}', "");
Expect(0, 55203, '\p{^Is_Hangul_Syllable_Type=lvtsyllable}', "");
Expect(0, 55203, '\P{Is_Hangul_Syllable_Type=lvtsyllable}', "");
Expect(1, 55203, '\P{^Is_Hangul_Syllable_Type=lvtsyllable}', "");
Expect(0, 55204, '\p{Is_Hangul_Syllable_Type=lvtsyllable}', "");
Expect(1, 55204, '\p{^Is_Hangul_Syllable_Type=lvtsyllable}', "");
Expect(1, 55204, '\P{Is_Hangul_Syllable_Type=lvtsyllable}', "");
Expect(0, 55204, '\P{^Is_Hangul_Syllable_Type=lvtsyllable}', "");
Expect(1, 55203, '\p{Is_Hangul_Syllable_Type:	 	LVT_Syllable}', "");
Expect(0, 55203, '\p{^Is_Hangul_Syllable_Type:	 	LVT_Syllable}', "");
Expect(0, 55203, '\P{Is_Hangul_Syllable_Type:	 	LVT_Syllable}', "");
Expect(1, 55203, '\P{^Is_Hangul_Syllable_Type:	 	LVT_Syllable}', "");
Expect(0, 55204, '\p{Is_Hangul_Syllable_Type:	 	LVT_Syllable}', "");
Expect(1, 55204, '\p{^Is_Hangul_Syllable_Type:	 	LVT_Syllable}', "");
Expect(1, 55204, '\P{Is_Hangul_Syllable_Type:	 	LVT_Syllable}', "");
Expect(0, 55204, '\P{^Is_Hangul_Syllable_Type:	 	LVT_Syllable}', "");
Error('\p{Is_Hst=-/a/LVT}');
Error('\P{Is_Hst=-/a/LVT}');
Expect(1, 55203, '\p{Is_Hst=lvt}', "");
Expect(0, 55203, '\p{^Is_Hst=lvt}', "");
Expect(0, 55203, '\P{Is_Hst=lvt}', "");
Expect(1, 55203, '\P{^Is_Hst=lvt}', "");
Expect(0, 55204, '\p{Is_Hst=lvt}', "");
Expect(1, 55204, '\p{^Is_Hst=lvt}', "");
Expect(1, 55204, '\P{Is_Hst=lvt}', "");
Expect(0, 55204, '\P{^Is_Hst=lvt}', "");
Expect(1, 55203, '\p{Is_Hst=	-LVT}', "");
Expect(0, 55203, '\p{^Is_Hst=	-LVT}', "");
Expect(0, 55203, '\P{Is_Hst=	-LVT}', "");
Expect(1, 55203, '\P{^Is_Hst=	-LVT}', "");
Expect(0, 55204, '\p{Is_Hst=	-LVT}', "");
Expect(1, 55204, '\p{^Is_Hst=	-LVT}', "");
Expect(1, 55204, '\P{Is_Hst=	-LVT}', "");
Expect(0, 55204, '\P{^Is_Hst=	-LVT}', "");
Error('\p{Hangul_Syllable_Type: _ Not_Applicable:=}');
Error('\P{Hangul_Syllable_Type: _ Not_Applicable:=}');
Expect(1, 55292, '\p{Hangul_Syllable_Type=notapplicable}', "");
Expect(0, 55292, '\p{^Hangul_Syllable_Type=notapplicable}', "");
Expect(0, 55292, '\P{Hangul_Syllable_Type=notapplicable}', "");
Expect(1, 55292, '\P{^Hangul_Syllable_Type=notapplicable}', "");
Expect(0, 55291, '\p{Hangul_Syllable_Type=notapplicable}', "");
Expect(1, 55291, '\p{^Hangul_Syllable_Type=notapplicable}', "");
Expect(1, 55291, '\P{Hangul_Syllable_Type=notapplicable}', "");
Expect(0, 55291, '\P{^Hangul_Syllable_Type=notapplicable}', "");
Expect(1, 55292, '\p{Hangul_Syllable_Type=	_not_Applicable}', "");
Expect(0, 55292, '\p{^Hangul_Syllable_Type=	_not_Applicable}', "");
Expect(0, 55292, '\P{Hangul_Syllable_Type=	_not_Applicable}', "");
Expect(1, 55292, '\P{^Hangul_Syllable_Type=	_not_Applicable}', "");
Expect(0, 55291, '\p{Hangul_Syllable_Type=	_not_Applicable}', "");
Expect(1, 55291, '\p{^Hangul_Syllable_Type=	_not_Applicable}', "");
Expect(1, 55291, '\P{Hangul_Syllable_Type=	_not_Applicable}', "");
Expect(0, 55291, '\P{^Hangul_Syllable_Type=	_not_Applicable}', "");
Error('\p{Hst=:= NA}');
Error('\P{Hst=:= NA}');
Expect(1, 55292, '\p{Hst=na}', "");
Expect(0, 55292, '\p{^Hst=na}', "");
Expect(0, 55292, '\P{Hst=na}', "");
Expect(1, 55292, '\P{^Hst=na}', "");
Expect(0, 55291, '\p{Hst=na}', "");
Expect(1, 55291, '\p{^Hst=na}', "");
Expect(1, 55291, '\P{Hst=na}', "");
Expect(0, 55291, '\P{^Hst=na}', "");
Expect(1, 55292, '\p{Hst= 	NA}', "");
Expect(0, 55292, '\p{^Hst= 	NA}', "");
Expect(0, 55292, '\P{Hst= 	NA}', "");
Expect(1, 55292, '\P{^Hst= 	NA}', "");
Expect(0, 55291, '\p{Hst= 	NA}', "");
Expect(1, 55291, '\p{^Hst= 	NA}', "");
Expect(1, 55291, '\P{Hst= 	NA}', "");
Expect(0, 55291, '\P{^Hst= 	NA}', "");
Error('\p{Is_Hangul_Syllable_Type= :=Not_applicable}');
Error('\P{Is_Hangul_Syllable_Type= :=Not_applicable}');
Expect(1, 55292, '\p{Is_Hangul_Syllable_Type=notapplicable}', "");
Expect(0, 55292, '\p{^Is_Hangul_Syllable_Type=notapplicable}', "");
Expect(0, 55292, '\P{Is_Hangul_Syllable_Type=notapplicable}', "");
Expect(1, 55292, '\P{^Is_Hangul_Syllable_Type=notapplicable}', "");
Expect(0, 55291, '\p{Is_Hangul_Syllable_Type=notapplicable}', "");
Expect(1, 55291, '\p{^Is_Hangul_Syllable_Type=notapplicable}', "");
Expect(1, 55291, '\P{Is_Hangul_Syllable_Type=notapplicable}', "");
Expect(0, 55291, '\P{^Is_Hangul_Syllable_Type=notapplicable}', "");
Expect(1, 55292, '\p{Is_Hangul_Syllable_Type=--not_APPLICABLE}', "");
Expect(0, 55292, '\p{^Is_Hangul_Syllable_Type=--not_APPLICABLE}', "");
Expect(0, 55292, '\P{Is_Hangul_Syllable_Type=--not_APPLICABLE}', "");
Expect(1, 55292, '\P{^Is_Hangul_Syllable_Type=--not_APPLICABLE}', "");
Expect(0, 55291, '\p{Is_Hangul_Syllable_Type=--not_APPLICABLE}', "");
Expect(1, 55291, '\p{^Is_Hangul_Syllable_Type=--not_APPLICABLE}', "");
Expect(1, 55291, '\P{Is_Hangul_Syllable_Type=--not_APPLICABLE}', "");
Expect(0, 55291, '\P{^Is_Hangul_Syllable_Type=--not_APPLICABLE}', "");
Error('\p{Is_Hst=__na/a/}');
Error('\P{Is_Hst=__na/a/}');
Expect(1, 55292, '\p{Is_Hst=na}', "");
Expect(0, 55292, '\p{^Is_Hst=na}', "");
Expect(0, 55292, '\P{Is_Hst=na}', "");
Expect(1, 55292, '\P{^Is_Hst=na}', "");
Expect(0, 55291, '\p{Is_Hst=na}', "");
Expect(1, 55291, '\p{^Is_Hst=na}', "");
Expect(1, 55291, '\P{Is_Hst=na}', "");
Expect(0, 55291, '\P{^Is_Hst=na}', "");
Expect(1, 55292, '\p{Is_Hst:	 NA}', "");
Expect(0, 55292, '\p{^Is_Hst:	 NA}', "");
Expect(0, 55292, '\P{Is_Hst:	 NA}', "");
Expect(1, 55292, '\P{^Is_Hst:	 NA}', "");
Expect(0, 55291, '\p{Is_Hst:	 NA}', "");
Expect(1, 55291, '\p{^Is_Hst:	 NA}', "");
Expect(1, 55291, '\P{Is_Hst:	 NA}', "");
Expect(0, 55291, '\P{^Is_Hst:	 NA}', "");
Error('\p{Hangul_Syllable_Type=-/a/trailing_Jamo}');
Error('\P{Hangul_Syllable_Type=-/a/trailing_Jamo}');
Expect(1, 55291, '\p{Hangul_Syllable_Type:trailingjamo}', "");
Expect(0, 55291, '\p{^Hangul_Syllable_Type:trailingjamo}', "");
Expect(0, 55291, '\P{Hangul_Syllable_Type:trailingjamo}', "");
Expect(1, 55291, '\P{^Hangul_Syllable_Type:trailingjamo}', "");
Expect(0, 55292, '\p{Hangul_Syllable_Type:trailingjamo}', "");
Expect(1, 55292, '\p{^Hangul_Syllable_Type:trailingjamo}', "");
Expect(1, 55292, '\P{Hangul_Syllable_Type:trailingjamo}', "");
Expect(0, 55292, '\P{^Hangul_Syllable_Type:trailingjamo}', "");
Expect(1, 55291, '\p{Hangul_Syllable_Type=	trailing_jamo}', "");
Expect(0, 55291, '\p{^Hangul_Syllable_Type=	trailing_jamo}', "");
Expect(0, 55291, '\P{Hangul_Syllable_Type=	trailing_jamo}', "");
Expect(1, 55291, '\P{^Hangul_Syllable_Type=	trailing_jamo}', "");
Expect(0, 55292, '\p{Hangul_Syllable_Type=	trailing_jamo}', "");
Expect(1, 55292, '\p{^Hangul_Syllable_Type=	trailing_jamo}', "");
Expect(1, 55292, '\P{Hangul_Syllable_Type=	trailing_jamo}', "");
Expect(0, 55292, '\P{^Hangul_Syllable_Type=	trailing_jamo}', "");
Error('\p{Hst=:=T}');
Error('\P{Hst=:=T}');
Expect(1, 55291, '\p{Hst:	t}', "");
Expect(0, 55291, '\p{^Hst:	t}', "");
Expect(0, 55291, '\P{Hst:	t}', "");
Expect(1, 55291, '\P{^Hst:	t}', "");
Expect(0, 55292, '\p{Hst:	t}', "");
Expect(1, 55292, '\p{^Hst:	t}', "");
Expect(1, 55292, '\P{Hst:	t}', "");
Expect(0, 55292, '\P{^Hst:	t}', "");
Expect(1, 55291, '\p{Hst=- T}', "");
Expect(0, 55291, '\p{^Hst=- T}', "");
Expect(0, 55291, '\P{Hst=- T}', "");
Expect(1, 55291, '\P{^Hst=- T}', "");
Expect(0, 55292, '\p{Hst=- T}', "");
Expect(1, 55292, '\p{^Hst=- T}', "");
Expect(1, 55292, '\P{Hst=- T}', "");
Expect(0, 55292, '\P{^Hst=- T}', "");
Error('\p{Is_Hangul_Syllable_Type=	/a/trailing_Jamo}');
Error('\P{Is_Hangul_Syllable_Type=	/a/trailing_Jamo}');
Expect(1, 55291, '\p{Is_Hangul_Syllable_Type=trailingjamo}', "");
Expect(0, 55291, '\p{^Is_Hangul_Syllable_Type=trailingjamo}', "");
Expect(0, 55291, '\P{Is_Hangul_Syllable_Type=trailingjamo}', "");
Expect(1, 55291, '\P{^Is_Hangul_Syllable_Type=trailingjamo}', "");
Expect(0, 55292, '\p{Is_Hangul_Syllable_Type=trailingjamo}', "");
Expect(1, 55292, '\p{^Is_Hangul_Syllable_Type=trailingjamo}', "");
Expect(1, 55292, '\P{Is_Hangul_Syllable_Type=trailingjamo}', "");
Expect(0, 55292, '\P{^Is_Hangul_Syllable_Type=trailingjamo}', "");
Expect(1, 55291, '\p{Is_Hangul_Syllable_Type=-_TRAILING_JAMO}', "");
Expect(0, 55291, '\p{^Is_Hangul_Syllable_Type=-_TRAILING_JAMO}', "");
Expect(0, 55291, '\P{Is_Hangul_Syllable_Type=-_TRAILING_JAMO}', "");
Expect(1, 55291, '\P{^Is_Hangul_Syllable_Type=-_TRAILING_JAMO}', "");
Expect(0, 55292, '\p{Is_Hangul_Syllable_Type=-_TRAILING_JAMO}', "");
Expect(1, 55292, '\p{^Is_Hangul_Syllable_Type=-_TRAILING_JAMO}', "");
Expect(1, 55292, '\P{Is_Hangul_Syllable_Type=-_TRAILING_JAMO}', "");
Expect(0, 55292, '\P{^Is_Hangul_Syllable_Type=-_TRAILING_JAMO}', "");
Error('\p{Is_Hst=_/a/T}');
Error('\P{Is_Hst=_/a/T}');
Expect(1, 55291, '\p{Is_Hst=t}', "");
Expect(0, 55291, '\p{^Is_Hst=t}', "");
Expect(0, 55291, '\P{Is_Hst=t}', "");
Expect(1, 55291, '\P{^Is_Hst=t}', "");
Expect(0, 55292, '\p{Is_Hst=t}', "");
Expect(1, 55292, '\p{^Is_Hst=t}', "");
Expect(1, 55292, '\P{Is_Hst=t}', "");
Expect(0, 55292, '\P{^Is_Hst=t}', "");
Expect(1, 55291, '\p{Is_Hst=  T}', "");
Expect(0, 55291, '\p{^Is_Hst=  T}', "");
Expect(0, 55291, '\P{Is_Hst=  T}', "");
Expect(1, 55291, '\P{^Is_Hst=  T}', "");
Expect(0, 55292, '\p{Is_Hst=  T}', "");
Expect(1, 55292, '\p{^Is_Hst=  T}', "");
Expect(1, 55292, '\P{Is_Hst=  T}', "");
Expect(0, 55292, '\P{^Is_Hst=  T}', "");
Error('\p{Hangul_Syllable_Type=_Vowel_Jamo:=}');
Error('\P{Hangul_Syllable_Type=_Vowel_Jamo:=}');
Expect(1, 55238, '\p{Hangul_Syllable_Type=voweljamo}', "");
Expect(0, 55238, '\p{^Hangul_Syllable_Type=voweljamo}', "");
Expect(0, 55238, '\P{Hangul_Syllable_Type=voweljamo}', "");
Expect(1, 55238, '\P{^Hangul_Syllable_Type=voweljamo}', "");
Expect(0, 55239, '\p{Hangul_Syllable_Type=voweljamo}', "");
Expect(1, 55239, '\p{^Hangul_Syllable_Type=voweljamo}', "");
Expect(1, 55239, '\P{Hangul_Syllable_Type=voweljamo}', "");
Expect(0, 55239, '\P{^Hangul_Syllable_Type=voweljamo}', "");
Expect(1, 55238, '\p{Hangul_Syllable_Type= VOWEL_Jamo}', "");
Expect(0, 55238, '\p{^Hangul_Syllable_Type= VOWEL_Jamo}', "");
Expect(0, 55238, '\P{Hangul_Syllable_Type= VOWEL_Jamo}', "");
Expect(1, 55238, '\P{^Hangul_Syllable_Type= VOWEL_Jamo}', "");
Expect(0, 55239, '\p{Hangul_Syllable_Type= VOWEL_Jamo}', "");
Expect(1, 55239, '\p{^Hangul_Syllable_Type= VOWEL_Jamo}', "");
Expect(1, 55239, '\P{Hangul_Syllable_Type= VOWEL_Jamo}', "");
Expect(0, 55239, '\P{^Hangul_Syllable_Type= VOWEL_Jamo}', "");
Error('\p{Hst=/a/  v}');
Error('\P{Hst=/a/  v}');
Expect(1, 55238, '\p{Hst=v}', "");
Expect(0, 55238, '\p{^Hst=v}', "");
Expect(0, 55238, '\P{Hst=v}', "");
Expect(1, 55238, '\P{^Hst=v}', "");
Expect(0, 55239, '\p{Hst=v}', "");
Expect(1, 55239, '\p{^Hst=v}', "");
Expect(1, 55239, '\P{Hst=v}', "");
Expect(0, 55239, '\P{^Hst=v}', "");
Expect(1, 55238, '\p{Hst=	_v}', "");
Expect(0, 55238, '\p{^Hst=	_v}', "");
Expect(0, 55238, '\P{Hst=	_v}', "");
Expect(1, 55238, '\P{^Hst=	_v}', "");
Expect(0, 55239, '\p{Hst=	_v}', "");
Expect(1, 55239, '\p{^Hst=	_v}', "");
Expect(1, 55239, '\P{Hst=	_v}', "");
Expect(0, 55239, '\P{^Hst=	_v}', "");
Error('\p{Is_Hangul_Syllable_Type=/a/Vowel_jamo}');
Error('\P{Is_Hangul_Syllable_Type=/a/Vowel_jamo}');
Expect(1, 55238, '\p{Is_Hangul_Syllable_Type=voweljamo}', "");
Expect(0, 55238, '\p{^Is_Hangul_Syllable_Type=voweljamo}', "");
Expect(0, 55238, '\P{Is_Hangul_Syllable_Type=voweljamo}', "");
Expect(1, 55238, '\P{^Is_Hangul_Syllable_Type=voweljamo}', "");
Expect(0, 55239, '\p{Is_Hangul_Syllable_Type=voweljamo}', "");
Expect(1, 55239, '\p{^Is_Hangul_Syllable_Type=voweljamo}', "");
Expect(1, 55239, '\P{Is_Hangul_Syllable_Type=voweljamo}', "");
Expect(0, 55239, '\P{^Is_Hangul_Syllable_Type=voweljamo}', "");
Expect(1, 55238, '\p{Is_Hangul_Syllable_Type=__Vowel_Jamo}', "");
Expect(0, 55238, '\p{^Is_Hangul_Syllable_Type=__Vowel_Jamo}', "");
Expect(0, 55238, '\P{Is_Hangul_Syllable_Type=__Vowel_Jamo}', "");
Expect(1, 55238, '\P{^Is_Hangul_Syllable_Type=__Vowel_Jamo}', "");
Expect(0, 55239, '\p{Is_Hangul_Syllable_Type=__Vowel_Jamo}', "");
Expect(1, 55239, '\p{^Is_Hangul_Syllable_Type=__Vowel_Jamo}', "");
Expect(1, 55239, '\P{Is_Hangul_Syllable_Type=__Vowel_Jamo}', "");
Expect(0, 55239, '\P{^Is_Hangul_Syllable_Type=__Vowel_Jamo}', "");
Error('\p{Is_Hst=/a/_	V}');
Error('\P{Is_Hst=/a/_	V}');
Expect(1, 55238, '\p{Is_Hst=v}', "");
Expect(0, 55238, '\p{^Is_Hst=v}', "");
Expect(0, 55238, '\P{Is_Hst=v}', "");
Expect(1, 55238, '\P{^Is_Hst=v}', "");
Expect(0, 55239, '\p{Is_Hst=v}', "");
Expect(1, 55239, '\p{^Is_Hst=v}', "");
Expect(1, 55239, '\P{Is_Hst=v}', "");
Expect(0, 55239, '\P{^Is_Hst=v}', "");
Expect(1, 55238, '\p{Is_Hst= _v}', "");
Expect(0, 55238, '\p{^Is_Hst= _v}', "");
Expect(0, 55238, '\P{Is_Hst= _v}', "");
Expect(1, 55238, '\P{^Is_Hst= _v}', "");
Expect(0, 55239, '\p{Is_Hst= _v}', "");
Expect(1, 55239, '\p{^Is_Hst= _v}', "");
Expect(1, 55239, '\P{Is_Hst= _v}', "");
Expect(0, 55239, '\P{^Is_Hst= _v}', "");
Error('\p{Joining_Group=:=ain}');
Error('\P{Joining_Group=:=ain}');
Expect(1, 1887, '\p{Joining_Group=ain}', "");
Expect(0, 1887, '\p{^Joining_Group=ain}', "");
Expect(0, 1887, '\P{Joining_Group=ain}', "");
Expect(1, 1887, '\P{^Joining_Group=ain}', "");
Expect(0, 1888, '\p{Joining_Group=ain}', "");
Expect(1, 1888, '\p{^Joining_Group=ain}', "");
Expect(1, 1888, '\P{Joining_Group=ain}', "");
Expect(0, 1888, '\P{^Joining_Group=ain}', "");
Expect(1, 1887, '\p{Joining_Group=		AIN}', "");
Expect(0, 1887, '\p{^Joining_Group=		AIN}', "");
Expect(0, 1887, '\P{Joining_Group=		AIN}', "");
Expect(1, 1887, '\P{^Joining_Group=		AIN}', "");
Expect(0, 1888, '\p{Joining_Group=		AIN}', "");
Expect(1, 1888, '\p{^Joining_Group=		AIN}', "");
Expect(1, 1888, '\P{Joining_Group=		AIN}', "");
Expect(0, 1888, '\P{^Joining_Group=		AIN}', "");
Error('\p{Jg=:=- ain}');
Error('\P{Jg=:=- ain}');
Expect(1, 1887, '\p{Jg=ain}', "");
Expect(0, 1887, '\p{^Jg=ain}', "");
Expect(0, 1887, '\P{Jg=ain}', "");
Expect(1, 1887, '\P{^Jg=ain}', "");
Expect(0, 1888, '\p{Jg=ain}', "");
Expect(1, 1888, '\p{^Jg=ain}', "");
Expect(1, 1888, '\P{Jg=ain}', "");
Expect(0, 1888, '\P{^Jg=ain}', "");
Expect(1, 1887, '\p{Jg=	 ain}', "");
Expect(0, 1887, '\p{^Jg=	 ain}', "");
Expect(0, 1887, '\P{Jg=	 ain}', "");
Expect(1, 1887, '\P{^Jg=	 ain}', "");
Expect(0, 1888, '\p{Jg=	 ain}', "");
Expect(1, 1888, '\p{^Jg=	 ain}', "");
Expect(1, 1888, '\P{Jg=	 ain}', "");
Expect(0, 1888, '\P{^Jg=	 ain}', "");
Error('\p{Is_Joining_Group=/a/__ain}');
Error('\P{Is_Joining_Group=/a/__ain}');
Expect(1, 1887, '\p{Is_Joining_Group=ain}', "");
Expect(0, 1887, '\p{^Is_Joining_Group=ain}', "");
Expect(0, 1887, '\P{Is_Joining_Group=ain}', "");
Expect(1, 1887, '\P{^Is_Joining_Group=ain}', "");
Expect(0, 1888, '\p{Is_Joining_Group=ain}', "");
Expect(1, 1888, '\p{^Is_Joining_Group=ain}', "");
Expect(1, 1888, '\P{Is_Joining_Group=ain}', "");
Expect(0, 1888, '\P{^Is_Joining_Group=ain}', "");
Expect(1, 1887, '\p{Is_Joining_Group=- Ain}', "");
Expect(0, 1887, '\p{^Is_Joining_Group=- Ain}', "");
Expect(0, 1887, '\P{Is_Joining_Group=- Ain}', "");
Expect(1, 1887, '\P{^Is_Joining_Group=- Ain}', "");
Expect(0, 1888, '\p{Is_Joining_Group=- Ain}', "");
Expect(1, 1888, '\p{^Is_Joining_Group=- Ain}', "");
Expect(1, 1888, '\P{Is_Joining_Group=- Ain}', "");
Expect(0, 1888, '\P{^Is_Joining_Group=- Ain}', "");
Error('\p{Is_Jg= :=ain}');
Error('\P{Is_Jg= :=ain}');
Expect(1, 1887, '\p{Is_Jg=ain}', "");
Expect(0, 1887, '\p{^Is_Jg=ain}', "");
Expect(0, 1887, '\P{Is_Jg=ain}', "");
Expect(1, 1887, '\P{^Is_Jg=ain}', "");
Expect(0, 1888, '\p{Is_Jg=ain}', "");
Expect(1, 1888, '\p{^Is_Jg=ain}', "");
Expect(1, 1888, '\P{Is_Jg=ain}', "");
Expect(0, 1888, '\P{^Is_Jg=ain}', "");
Expect(1, 1887, '\p{Is_Jg=	-Ain}', "");
Expect(0, 1887, '\p{^Is_Jg=	-Ain}', "");
Expect(0, 1887, '\P{Is_Jg=	-Ain}', "");
Expect(1, 1887, '\P{^Is_Jg=	-Ain}', "");
Expect(0, 1888, '\p{Is_Jg=	-Ain}', "");
Expect(1, 1888, '\p{^Is_Jg=	-Ain}', "");
Expect(1, 1888, '\P{Is_Jg=	-Ain}', "");
Expect(0, 1888, '\P{^Is_Jg=	-Ain}', "");
Error('\p{Joining_Group=-ALAPH:=}');
Error('\P{Joining_Group=-ALAPH:=}');
Expect(1, 1808, '\p{Joining_Group=alaph}', "");
Expect(0, 1808, '\p{^Joining_Group=alaph}', "");
Expect(0, 1808, '\P{Joining_Group=alaph}', "");
Expect(1, 1808, '\P{^Joining_Group=alaph}', "");
Expect(0, 1809, '\p{Joining_Group=alaph}', "");
Expect(1, 1809, '\p{^Joining_Group=alaph}', "");
Expect(1, 1809, '\P{Joining_Group=alaph}', "");
Expect(0, 1809, '\P{^Joining_Group=alaph}', "");
Expect(1, 1808, '\p{Joining_Group=_alaph}', "");
Expect(0, 1808, '\p{^Joining_Group=_alaph}', "");
Expect(0, 1808, '\P{Joining_Group=_alaph}', "");
Expect(1, 1808, '\P{^Joining_Group=_alaph}', "");
Expect(0, 1809, '\p{Joining_Group=_alaph}', "");
Expect(1, 1809, '\p{^Joining_Group=_alaph}', "");
Expect(1, 1809, '\P{Joining_Group=_alaph}', "");
Expect(0, 1809, '\P{^Joining_Group=_alaph}', "");
Error('\p{Jg=	-Alaph/a/}');
Error('\P{Jg=	-Alaph/a/}');
Expect(1, 1808, '\p{Jg:   alaph}', "");
Expect(0, 1808, '\p{^Jg:   alaph}', "");
Expect(0, 1808, '\P{Jg:   alaph}', "");
Expect(1, 1808, '\P{^Jg:   alaph}', "");
Expect(0, 1809, '\p{Jg:   alaph}', "");
Expect(1, 1809, '\p{^Jg:   alaph}', "");
Expect(1, 1809, '\P{Jg:   alaph}', "");
Expect(0, 1809, '\P{^Jg:   alaph}', "");
Expect(1, 1808, '\p{Jg= 	Alaph}', "");
Expect(0, 1808, '\p{^Jg= 	Alaph}', "");
Expect(0, 1808, '\P{Jg= 	Alaph}', "");
Expect(1, 1808, '\P{^Jg= 	Alaph}', "");
Expect(0, 1809, '\p{Jg= 	Alaph}', "");
Expect(1, 1809, '\p{^Jg= 	Alaph}', "");
Expect(1, 1809, '\P{Jg= 	Alaph}', "");
Expect(0, 1809, '\P{^Jg= 	Alaph}', "");
Error('\p{Is_Joining_Group= alaph/a/}');
Error('\P{Is_Joining_Group= alaph/a/}');
Expect(1, 1808, '\p{Is_Joining_Group=alaph}', "");
Expect(0, 1808, '\p{^Is_Joining_Group=alaph}', "");
Expect(0, 1808, '\P{Is_Joining_Group=alaph}', "");
Expect(1, 1808, '\P{^Is_Joining_Group=alaph}', "");
Expect(0, 1809, '\p{Is_Joining_Group=alaph}', "");
Expect(1, 1809, '\p{^Is_Joining_Group=alaph}', "");
Expect(1, 1809, '\P{Is_Joining_Group=alaph}', "");
Expect(0, 1809, '\P{^Is_Joining_Group=alaph}', "");
Expect(1, 1808, '\p{Is_Joining_Group:   	 alaph}', "");
Expect(0, 1808, '\p{^Is_Joining_Group:   	 alaph}', "");
Expect(0, 1808, '\P{Is_Joining_Group:   	 alaph}', "");
Expect(1, 1808, '\P{^Is_Joining_Group:   	 alaph}', "");
Expect(0, 1809, '\p{Is_Joining_Group:   	 alaph}', "");
Expect(1, 1809, '\p{^Is_Joining_Group:   	 alaph}', "");
Expect(1, 1809, '\P{Is_Joining_Group:   	 alaph}', "");
Expect(0, 1809, '\P{^Is_Joining_Group:   	 alaph}', "");
Error('\p{Is_Jg=_:=ALAPH}');
Error('\P{Is_Jg=_:=ALAPH}');
Expect(1, 1808, '\p{Is_Jg=alaph}', "");
Expect(0, 1808, '\p{^Is_Jg=alaph}', "");
Expect(0, 1808, '\P{Is_Jg=alaph}', "");
Expect(1, 1808, '\P{^Is_Jg=alaph}', "");
Expect(0, 1809, '\p{Is_Jg=alaph}', "");
Expect(1, 1809, '\p{^Is_Jg=alaph}', "");
Expect(1, 1809, '\P{Is_Jg=alaph}', "");
Expect(0, 1809, '\P{^Is_Jg=alaph}', "");
Expect(1, 1808, '\p{Is_Jg=ALAPH}', "");
Expect(0, 1808, '\p{^Is_Jg=ALAPH}', "");
Expect(0, 1808, '\P{Is_Jg=ALAPH}', "");
Expect(1, 1808, '\P{^Is_Jg=ALAPH}', "");
Expect(0, 1809, '\p{Is_Jg=ALAPH}', "");
Expect(1, 1809, '\p{^Is_Jg=ALAPH}', "");
Expect(1, 1809, '\P{Is_Jg=ALAPH}', "");
Expect(0, 1809, '\P{^Is_Jg=ALAPH}', "");
Error('\p{Joining_Group=:=-Alef}');
Error('\P{Joining_Group=:=-Alef}');
Expect(1, 1908, '\p{Joining_Group=alef}', "");
Expect(0, 1908, '\p{^Joining_Group=alef}', "");
Expect(0, 1908, '\P{Joining_Group=alef}', "");
Expect(1, 1908, '\P{^Joining_Group=alef}', "");
Expect(0, 1909, '\p{Joining_Group=alef}', "");
Expect(1, 1909, '\p{^Joining_Group=alef}', "");
Expect(1, 1909, '\P{Joining_Group=alef}', "");
Expect(0, 1909, '\P{^Joining_Group=alef}', "");
Expect(1, 1908, '\p{Joining_Group= 	Alef}', "");
Expect(0, 1908, '\p{^Joining_Group= 	Alef}', "");
Expect(0, 1908, '\P{Joining_Group= 	Alef}', "");
Expect(1, 1908, '\P{^Joining_Group= 	Alef}', "");
Expect(0, 1909, '\p{Joining_Group= 	Alef}', "");
Expect(1, 1909, '\p{^Joining_Group= 	Alef}', "");
Expect(1, 1909, '\P{Joining_Group= 	Alef}', "");
Expect(0, 1909, '\P{^Joining_Group= 	Alef}', "");
Error('\p{Jg:/a/- alef}');
Error('\P{Jg:/a/- alef}');
Expect(1, 1908, '\p{Jg=alef}', "");
Expect(0, 1908, '\p{^Jg=alef}', "");
Expect(0, 1908, '\P{Jg=alef}', "");
Expect(1, 1908, '\P{^Jg=alef}', "");
Expect(0, 1909, '\p{Jg=alef}', "");
Expect(1, 1909, '\p{^Jg=alef}', "");
Expect(1, 1909, '\P{Jg=alef}', "");
Expect(0, 1909, '\P{^Jg=alef}', "");
Expect(1, 1908, '\p{Jg=  alef}', "");
Expect(0, 1908, '\p{^Jg=  alef}', "");
Expect(0, 1908, '\P{Jg=  alef}', "");
Expect(1, 1908, '\P{^Jg=  alef}', "");
Expect(0, 1909, '\p{Jg=  alef}', "");
Expect(1, 1909, '\p{^Jg=  alef}', "");
Expect(1, 1909, '\P{Jg=  alef}', "");
Expect(0, 1909, '\P{^Jg=  alef}', "");
Error('\p{Is_Joining_Group=_	Alef/a/}');
Error('\P{Is_Joining_Group=_	Alef/a/}');
Expect(1, 1908, '\p{Is_Joining_Group=alef}', "");
Expect(0, 1908, '\p{^Is_Joining_Group=alef}', "");
Expect(0, 1908, '\P{Is_Joining_Group=alef}', "");
Expect(1, 1908, '\P{^Is_Joining_Group=alef}', "");
Expect(0, 1909, '\p{Is_Joining_Group=alef}', "");
Expect(1, 1909, '\p{^Is_Joining_Group=alef}', "");
Expect(1, 1909, '\P{Is_Joining_Group=alef}', "");
Expect(0, 1909, '\P{^Is_Joining_Group=alef}', "");
Expect(1, 1908, '\p{Is_Joining_Group=	-alef}', "");
Expect(0, 1908, '\p{^Is_Joining_Group=	-alef}', "");
Expect(0, 1908, '\P{Is_Joining_Group=	-alef}', "");
Expect(1, 1908, '\P{^Is_Joining_Group=	-alef}', "");
Expect(0, 1909, '\p{Is_Joining_Group=	-alef}', "");
Expect(1, 1909, '\p{^Is_Joining_Group=	-alef}', "");
Expect(1, 1909, '\P{Is_Joining_Group=	-alef}', "");
Expect(0, 1909, '\P{^Is_Joining_Group=	-alef}', "");
Error('\p{Is_Jg=:=	alef}');
Error('\P{Is_Jg=:=	alef}');
Expect(1, 1908, '\p{Is_Jg=alef}', "");
Expect(0, 1908, '\p{^Is_Jg=alef}', "");
Expect(0, 1908, '\P{Is_Jg=alef}', "");
Expect(1, 1908, '\P{^Is_Jg=alef}', "");
Expect(0, 1909, '\p{Is_Jg=alef}', "");
Expect(1, 1909, '\p{^Is_Jg=alef}', "");
Expect(1, 1909, '\P{Is_Jg=alef}', "");
Expect(0, 1909, '\P{^Is_Jg=alef}', "");
Expect(1, 1908, '\p{Is_Jg=_Alef}', "");
Expect(0, 1908, '\p{^Is_Jg=_Alef}', "");
Expect(0, 1908, '\P{Is_Jg=_Alef}', "");
Expect(1, 1908, '\P{^Is_Jg=_Alef}', "");
Expect(0, 1909, '\p{Is_Jg=_Alef}', "");
Expect(1, 1909, '\p{^Is_Jg=_Alef}', "");
Expect(1, 1909, '\P{Is_Jg=_Alef}', "");
Expect(0, 1909, '\P{^Is_Jg=_Alef}', "");
Error('\p{Joining_Group=__Beh:=}');
Error('\P{Joining_Group=__Beh:=}');
Expect(1, 2208, '\p{Joining_Group=beh}', "");
Expect(0, 2208, '\p{^Joining_Group=beh}', "");
Expect(0, 2208, '\P{Joining_Group=beh}', "");
Expect(1, 2208, '\P{^Joining_Group=beh}', "");
Expect(0, 2209, '\p{Joining_Group=beh}', "");
Expect(1, 2209, '\p{^Joining_Group=beh}', "");
Expect(1, 2209, '\P{Joining_Group=beh}', "");
Expect(0, 2209, '\P{^Joining_Group=beh}', "");
Expect(1, 2208, '\p{Joining_Group=_Beh}', "");
Expect(0, 2208, '\p{^Joining_Group=_Beh}', "");
Expect(0, 2208, '\P{Joining_Group=_Beh}', "");
Expect(1, 2208, '\P{^Joining_Group=_Beh}', "");
Expect(0, 2209, '\p{Joining_Group=_Beh}', "");
Expect(1, 2209, '\p{^Joining_Group=_Beh}', "");
Expect(1, 2209, '\P{Joining_Group=_Beh}', "");
Expect(0, 2209, '\P{^Joining_Group=_Beh}', "");
Error('\p{Jg= :=beh}');
Error('\P{Jg= :=beh}');
Expect(1, 2208, '\p{Jg=beh}', "");
Expect(0, 2208, '\p{^Jg=beh}', "");
Expect(0, 2208, '\P{Jg=beh}', "");
Expect(1, 2208, '\P{^Jg=beh}', "");
Expect(0, 2209, '\p{Jg=beh}', "");
Expect(1, 2209, '\p{^Jg=beh}', "");
Expect(1, 2209, '\P{Jg=beh}', "");
Expect(0, 2209, '\P{^Jg=beh}', "");
Expect(1, 2208, '\p{Jg=_BEH}', "");
Expect(0, 2208, '\p{^Jg=_BEH}', "");
Expect(0, 2208, '\P{Jg=_BEH}', "");
Expect(1, 2208, '\P{^Jg=_BEH}', "");
Expect(0, 2209, '\p{Jg=_BEH}', "");
Expect(1, 2209, '\p{^Jg=_BEH}', "");
Expect(1, 2209, '\P{Jg=_BEH}', "");
Expect(0, 2209, '\P{^Jg=_BEH}', "");
Error('\p{Is_Joining_Group=_/a/BEH}');
Error('\P{Is_Joining_Group=_/a/BEH}');
Expect(1, 2208, '\p{Is_Joining_Group=beh}', "");
Expect(0, 2208, '\p{^Is_Joining_Group=beh}', "");
Expect(0, 2208, '\P{Is_Joining_Group=beh}', "");
Expect(1, 2208, '\P{^Is_Joining_Group=beh}', "");
Expect(0, 2209, '\p{Is_Joining_Group=beh}', "");
Expect(1, 2209, '\p{^Is_Joining_Group=beh}', "");
Expect(1, 2209, '\P{Is_Joining_Group=beh}', "");
Expect(0, 2209, '\P{^Is_Joining_Group=beh}', "");
Expect(1, 2208, '\p{Is_Joining_Group=	 beh}', "");
Expect(0, 2208, '\p{^Is_Joining_Group=	 beh}', "");
Expect(0, 2208, '\P{Is_Joining_Group=	 beh}', "");
Expect(1, 2208, '\P{^Is_Joining_Group=	 beh}', "");
Expect(0, 2209, '\p{Is_Joining_Group=	 beh}', "");
Expect(1, 2209, '\p{^Is_Joining_Group=	 beh}', "");
Expect(1, 2209, '\P{Is_Joining_Group=	 beh}', "");
Expect(0, 2209, '\P{^Is_Joining_Group=	 beh}', "");
Error('\p{Is_Jg=	:=BEH}');
Error('\P{Is_Jg=	:=BEH}');
Expect(1, 2208, '\p{Is_Jg=beh}', "");
Expect(0, 2208, '\p{^Is_Jg=beh}', "");
Expect(0, 2208, '\P{Is_Jg=beh}', "");
Expect(1, 2208, '\P{^Is_Jg=beh}', "");
Expect(0, 2209, '\p{Is_Jg=beh}', "");
Expect(1, 2209, '\p{^Is_Jg=beh}', "");
Expect(1, 2209, '\P{Is_Jg=beh}', "");
Expect(0, 2209, '\P{^Is_Jg=beh}', "");
Expect(1, 2208, '\p{Is_Jg=_	beh}', "");
Expect(0, 2208, '\p{^Is_Jg=_	beh}', "");
Expect(0, 2208, '\P{Is_Jg=_	beh}', "");
Expect(1, 2208, '\P{^Is_Jg=_	beh}', "");
Expect(0, 2209, '\p{Is_Jg=_	beh}', "");
Expect(1, 2209, '\p{^Is_Jg=_	beh}', "");
Expect(1, 2209, '\P{Is_Jg=_	beh}', "");
Expect(0, 2209, '\P{^Is_Jg=_	beh}', "");
Error('\p{Joining_Group= /a/Beth}');
Error('\P{Joining_Group= /a/Beth}');
Expect(1, 1837, '\p{Joining_Group=beth}', "");
Expect(0, 1837, '\p{^Joining_Group=beth}', "");
Expect(0, 1837, '\P{Joining_Group=beth}', "");
Expect(1, 1837, '\P{^Joining_Group=beth}', "");
Expect(0, 1838, '\p{Joining_Group=beth}', "");
Expect(1, 1838, '\p{^Joining_Group=beth}', "");
Expect(1, 1838, '\P{Joining_Group=beth}', "");
Expect(0, 1838, '\P{^Joining_Group=beth}', "");
Expect(1, 1837, '\p{Joining_Group= -Beth}', "");
Expect(0, 1837, '\p{^Joining_Group= -Beth}', "");
Expect(0, 1837, '\P{Joining_Group= -Beth}', "");
Expect(1, 1837, '\P{^Joining_Group= -Beth}', "");
Expect(0, 1838, '\p{Joining_Group= -Beth}', "");
Expect(1, 1838, '\p{^Joining_Group= -Beth}', "");
Expect(1, 1838, '\P{Joining_Group= -Beth}', "");
Expect(0, 1838, '\P{^Joining_Group= -Beth}', "");
Error('\p{Jg::=	Beth}');
Error('\P{Jg::=	Beth}');
Expect(1, 1837, '\p{Jg=beth}', "");
Expect(0, 1837, '\p{^Jg=beth}', "");
Expect(0, 1837, '\P{Jg=beth}', "");
Expect(1, 1837, '\P{^Jg=beth}', "");
Expect(0, 1838, '\p{Jg=beth}', "");
Expect(1, 1838, '\p{^Jg=beth}', "");
Expect(1, 1838, '\P{Jg=beth}', "");
Expect(0, 1838, '\P{^Jg=beth}', "");
Expect(1, 1837, '\p{Jg:_	BETH}', "");
Expect(0, 1837, '\p{^Jg:_	BETH}', "");
Expect(0, 1837, '\P{Jg:_	BETH}', "");
Expect(1, 1837, '\P{^Jg:_	BETH}', "");
Expect(0, 1838, '\p{Jg:_	BETH}', "");
Expect(1, 1838, '\p{^Jg:_	BETH}', "");
Expect(1, 1838, '\P{Jg:_	BETH}', "");
Expect(0, 1838, '\P{^Jg:_	BETH}', "");
Error('\p{Is_Joining_Group=/a/Beth}');
Error('\P{Is_Joining_Group=/a/Beth}');
Expect(1, 1837, '\p{Is_Joining_Group=beth}', "");
Expect(0, 1837, '\p{^Is_Joining_Group=beth}', "");
Expect(0, 1837, '\P{Is_Joining_Group=beth}', "");
Expect(1, 1837, '\P{^Is_Joining_Group=beth}', "");
Expect(0, 1838, '\p{Is_Joining_Group=beth}', "");
Expect(1, 1838, '\p{^Is_Joining_Group=beth}', "");
Expect(1, 1838, '\P{Is_Joining_Group=beth}', "");
Expect(0, 1838, '\P{^Is_Joining_Group=beth}', "");
Expect(1, 1837, '\p{Is_Joining_Group=-Beth}', "");
Expect(0, 1837, '\p{^Is_Joining_Group=-Beth}', "");
Expect(0, 1837, '\P{Is_Joining_Group=-Beth}', "");
Expect(1, 1837, '\P{^Is_Joining_Group=-Beth}', "");
Expect(0, 1838, '\p{Is_Joining_Group=-Beth}', "");
Expect(1, 1838, '\p{^Is_Joining_Group=-Beth}', "");
Expect(1, 1838, '\P{Is_Joining_Group=-Beth}', "");
Expect(0, 1838, '\P{^Is_Joining_Group=-Beth}', "");
Error('\p{Is_Jg=/a/-Beth}');
Error('\P{Is_Jg=/a/-Beth}');
Expect(1, 1837, '\p{Is_Jg=beth}', "");
Expect(0, 1837, '\p{^Is_Jg=beth}', "");
Expect(0, 1837, '\P{Is_Jg=beth}', "");
Expect(1, 1837, '\P{^Is_Jg=beth}', "");
Expect(0, 1838, '\p{Is_Jg=beth}', "");
Expect(1, 1838, '\p{^Is_Jg=beth}', "");
Expect(1, 1838, '\P{Is_Jg=beth}', "");
Expect(0, 1838, '\P{^Is_Jg=beth}', "");
Expect(1, 1837, '\p{Is_Jg= -BETH}', "");
Expect(0, 1837, '\p{^Is_Jg= -BETH}', "");
Expect(0, 1837, '\P{Is_Jg= -BETH}', "");
Expect(1, 1837, '\P{^Is_Jg= -BETH}', "");
Expect(0, 1838, '\p{Is_Jg= -BETH}', "");
Expect(1, 1838, '\p{^Is_Jg= -BETH}', "");
Expect(1, 1838, '\P{Is_Jg= -BETH}', "");
Expect(0, 1838, '\P{^Is_Jg= -BETH}', "");
Error('\p{Joining_Group=-Burushaski_Yeh_Barree/a/}');
Error('\P{Joining_Group=-Burushaski_Yeh_Barree/a/}');
Expect(1, 1915, '\p{Joining_Group=burushaskiyehbarree}', "");
Expect(0, 1915, '\p{^Joining_Group=burushaskiyehbarree}', "");
Expect(0, 1915, '\P{Joining_Group=burushaskiyehbarree}', "");
Expect(1, 1915, '\P{^Joining_Group=burushaskiyehbarree}', "");
Expect(0, 1916, '\p{Joining_Group=burushaskiyehbarree}', "");
Expect(1, 1916, '\p{^Joining_Group=burushaskiyehbarree}', "");
Expect(1, 1916, '\P{Joining_Group=burushaskiyehbarree}', "");
Expect(0, 1916, '\P{^Joining_Group=burushaskiyehbarree}', "");
Expect(1, 1915, '\p{Joining_Group=- burushaski_yeh_Barree}', "");
Expect(0, 1915, '\p{^Joining_Group=- burushaski_yeh_Barree}', "");
Expect(0, 1915, '\P{Joining_Group=- burushaski_yeh_Barree}', "");
Expect(1, 1915, '\P{^Joining_Group=- burushaski_yeh_Barree}', "");
Expect(0, 1916, '\p{Joining_Group=- burushaski_yeh_Barree}', "");
Expect(1, 1916, '\p{^Joining_Group=- burushaski_yeh_Barree}', "");
Expect(1, 1916, '\P{Joining_Group=- burushaski_yeh_Barree}', "");
Expect(0, 1916, '\P{^Joining_Group=- burushaski_yeh_Barree}', "");
Error('\p{Jg=/a/burushaski_Yeh_BARREE}');
Error('\P{Jg=/a/burushaski_Yeh_BARREE}');
Expect(1, 1915, '\p{Jg: burushaskiyehbarree}', "");
Expect(0, 1915, '\p{^Jg: burushaskiyehbarree}', "");
Expect(0, 1915, '\P{Jg: burushaskiyehbarree}', "");
Expect(1, 1915, '\P{^Jg: burushaskiyehbarree}', "");
Expect(0, 1916, '\p{Jg: burushaskiyehbarree}', "");
Expect(1, 1916, '\p{^Jg: burushaskiyehbarree}', "");
Expect(1, 1916, '\P{Jg: burushaskiyehbarree}', "");
Expect(0, 1916, '\P{^Jg: burushaskiyehbarree}', "");
Expect(1, 1915, '\p{Jg=	BURUSHASKI_Yeh_Barree}', "");
Expect(0, 1915, '\p{^Jg=	BURUSHASKI_Yeh_Barree}', "");
Expect(0, 1915, '\P{Jg=	BURUSHASKI_Yeh_Barree}', "");
Expect(1, 1915, '\P{^Jg=	BURUSHASKI_Yeh_Barree}', "");
Expect(0, 1916, '\p{Jg=	BURUSHASKI_Yeh_Barree}', "");
Expect(1, 1916, '\p{^Jg=	BURUSHASKI_Yeh_Barree}', "");
Expect(1, 1916, '\P{Jg=	BURUSHASKI_Yeh_Barree}', "");
Expect(0, 1916, '\P{^Jg=	BURUSHASKI_Yeh_Barree}', "");
Error('\p{Is_Joining_Group:   _/a/Burushaski_yeh_Barree}');
Error('\P{Is_Joining_Group:   _/a/Burushaski_yeh_Barree}');
Expect(1, 1915, '\p{Is_Joining_Group=burushaskiyehbarree}', "");
Expect(0, 1915, '\p{^Is_Joining_Group=burushaskiyehbarree}', "");
Expect(0, 1915, '\P{Is_Joining_Group=burushaskiyehbarree}', "");
Expect(1, 1915, '\P{^Is_Joining_Group=burushaskiyehbarree}', "");
Expect(0, 1916, '\p{Is_Joining_Group=burushaskiyehbarree}', "");
Expect(1, 1916, '\p{^Is_Joining_Group=burushaskiyehbarree}', "");
Expect(1, 1916, '\P{Is_Joining_Group=burushaskiyehbarree}', "");
Expect(0, 1916, '\P{^Is_Joining_Group=burushaskiyehbarree}', "");
Expect(1, 1915, '\p{Is_Joining_Group=-burushaski_Yeh_BARREE}', "");
Expect(0, 1915, '\p{^Is_Joining_Group=-burushaski_Yeh_BARREE}', "");
Expect(0, 1915, '\P{Is_Joining_Group=-burushaski_Yeh_BARREE}', "");
Expect(1, 1915, '\P{^Is_Joining_Group=-burushaski_Yeh_BARREE}', "");
Expect(0, 1916, '\p{Is_Joining_Group=-burushaski_Yeh_BARREE}', "");
Expect(1, 1916, '\p{^Is_Joining_Group=-burushaski_Yeh_BARREE}', "");
Expect(1, 1916, '\P{Is_Joining_Group=-burushaski_Yeh_BARREE}', "");
Expect(0, 1916, '\P{^Is_Joining_Group=-burushaski_Yeh_BARREE}', "");
Error('\p{Is_Jg=/a/_BURUSHASKI_Yeh_barree}');
Error('\P{Is_Jg=/a/_BURUSHASKI_Yeh_barree}');
Expect(1, 1915, '\p{Is_Jg: burushaskiyehbarree}', "");
Expect(0, 1915, '\p{^Is_Jg: burushaskiyehbarree}', "");
Expect(0, 1915, '\P{Is_Jg: burushaskiyehbarree}', "");
Expect(1, 1915, '\P{^Is_Jg: burushaskiyehbarree}', "");
Expect(0, 1916, '\p{Is_Jg: burushaskiyehbarree}', "");
Expect(1, 1916, '\p{^Is_Jg: burushaskiyehbarree}', "");
Expect(1, 1916, '\P{Is_Jg: burushaskiyehbarree}', "");
Expect(0, 1916, '\P{^Is_Jg: burushaskiyehbarree}', "");
Expect(1, 1915, '\p{Is_Jg= Burushaski_Yeh_Barree}', "");
Expect(0, 1915, '\p{^Is_Jg= Burushaski_Yeh_Barree}', "");
Expect(0, 1915, '\P{Is_Jg= Burushaski_Yeh_Barree}', "");
Expect(1, 1915, '\P{^Is_Jg= Burushaski_Yeh_Barree}', "");
Expect(0, 1916, '\p{Is_Jg= Burushaski_Yeh_Barree}', "");
Expect(1, 1916, '\p{^Is_Jg= Burushaski_Yeh_Barree}', "");
Expect(1, 1916, '\P{Is_Jg= Burushaski_Yeh_Barree}', "");
Expect(0, 1916, '\P{^Is_Jg= Burushaski_Yeh_Barree}', "");
Error('\p{Joining_Group= _Dal/a/}');
Error('\P{Joining_Group= _Dal/a/}');
Expect(1, 1882, '\p{Joining_Group=dal}', "");
Expect(0, 1882, '\p{^Joining_Group=dal}', "");
Expect(0, 1882, '\P{Joining_Group=dal}', "");
Expect(1, 1882, '\P{^Joining_Group=dal}', "");
Expect(0, 1883, '\p{Joining_Group=dal}', "");
Expect(1, 1883, '\p{^Joining_Group=dal}', "");
Expect(1, 1883, '\P{Joining_Group=dal}', "");
Expect(0, 1883, '\P{^Joining_Group=dal}', "");
Expect(1, 1882, '\p{Joining_Group=_ dal}', "");
Expect(0, 1882, '\p{^Joining_Group=_ dal}', "");
Expect(0, 1882, '\P{Joining_Group=_ dal}', "");
Expect(1, 1882, '\P{^Joining_Group=_ dal}', "");
Expect(0, 1883, '\p{Joining_Group=_ dal}', "");
Expect(1, 1883, '\p{^Joining_Group=_ dal}', "");
Expect(1, 1883, '\P{Joining_Group=_ dal}', "");
Expect(0, 1883, '\P{^Joining_Group=_ dal}', "");
Error('\p{Jg=-/a/Dal}');
Error('\P{Jg=-/a/Dal}');
Expect(1, 1882, '\p{Jg: dal}', "");
Expect(0, 1882, '\p{^Jg: dal}', "");
Expect(0, 1882, '\P{Jg: dal}', "");
Expect(1, 1882, '\P{^Jg: dal}', "");
Expect(0, 1883, '\p{Jg: dal}', "");
Expect(1, 1883, '\p{^Jg: dal}', "");
Expect(1, 1883, '\P{Jg: dal}', "");
Expect(0, 1883, '\P{^Jg: dal}', "");
Expect(1, 1882, '\p{Jg= _DAL}', "");
Expect(0, 1882, '\p{^Jg= _DAL}', "");
Expect(0, 1882, '\P{Jg= _DAL}', "");
Expect(1, 1882, '\P{^Jg= _DAL}', "");
Expect(0, 1883, '\p{Jg= _DAL}', "");
Expect(1, 1883, '\p{^Jg= _DAL}', "");
Expect(1, 1883, '\P{Jg= _DAL}', "");
Expect(0, 1883, '\P{^Jg= _DAL}', "");
Error('\p{Is_Joining_Group=_ Dal/a/}');
Error('\P{Is_Joining_Group=_ Dal/a/}');
Expect(1, 1882, '\p{Is_Joining_Group=dal}', "");
Expect(0, 1882, '\p{^Is_Joining_Group=dal}', "");
Expect(0, 1882, '\P{Is_Joining_Group=dal}', "");
Expect(1, 1882, '\P{^Is_Joining_Group=dal}', "");
Expect(0, 1883, '\p{Is_Joining_Group=dal}', "");
Expect(1, 1883, '\p{^Is_Joining_Group=dal}', "");
Expect(1, 1883, '\P{Is_Joining_Group=dal}', "");
Expect(0, 1883, '\P{^Is_Joining_Group=dal}', "");
Expect(1, 1882, '\p{Is_Joining_Group=		Dal}', "");
Expect(0, 1882, '\p{^Is_Joining_Group=		Dal}', "");
Expect(0, 1882, '\P{Is_Joining_Group=		Dal}', "");
Expect(1, 1882, '\P{^Is_Joining_Group=		Dal}', "");
Expect(0, 1883, '\p{Is_Joining_Group=		Dal}', "");
Expect(1, 1883, '\p{^Is_Joining_Group=		Dal}', "");
Expect(1, 1883, '\P{Is_Joining_Group=		Dal}', "");
Expect(0, 1883, '\P{^Is_Joining_Group=		Dal}', "");
Error('\p{Is_Jg=:=_ Dal}');
Error('\P{Is_Jg=:=_ Dal}');
Expect(1, 1882, '\p{Is_Jg=dal}', "");
Expect(0, 1882, '\p{^Is_Jg=dal}', "");
Expect(0, 1882, '\P{Is_Jg=dal}', "");
Expect(1, 1882, '\P{^Is_Jg=dal}', "");
Expect(0, 1883, '\p{Is_Jg=dal}', "");
Expect(1, 1883, '\p{^Is_Jg=dal}', "");
Expect(1, 1883, '\P{Is_Jg=dal}', "");
Expect(0, 1883, '\P{^Is_Jg=dal}', "");
Expect(1, 1882, '\p{Is_Jg:   Dal}', "");
Expect(0, 1882, '\p{^Is_Jg:   Dal}', "");
Expect(0, 1882, '\P{Is_Jg:   Dal}', "");
Expect(1, 1882, '\P{^Is_Jg:   Dal}', "");
Expect(0, 1883, '\p{Is_Jg:   Dal}', "");
Expect(1, 1883, '\p{^Is_Jg:   Dal}', "");
Expect(1, 1883, '\P{Is_Jg:   Dal}', "");
Expect(0, 1883, '\P{^Is_Jg:   Dal}', "");
Error('\p{Joining_Group=:=Dalath_RISH}');
Error('\P{Joining_Group=:=Dalath_RISH}');
Expect(1, 1839, '\p{Joining_Group=dalathrish}', "");
Expect(0, 1839, '\p{^Joining_Group=dalathrish}', "");
Expect(0, 1839, '\P{Joining_Group=dalathrish}', "");
Expect(1, 1839, '\P{^Joining_Group=dalathrish}', "");
Expect(0, 1840, '\p{Joining_Group=dalathrish}', "");
Expect(1, 1840, '\p{^Joining_Group=dalathrish}', "");
Expect(1, 1840, '\P{Joining_Group=dalathrish}', "");
Expect(0, 1840, '\P{^Joining_Group=dalathrish}', "");
Expect(1, 1839, '\p{Joining_Group:   -	dalath_Rish}', "");
Expect(0, 1839, '\p{^Joining_Group:   -	dalath_Rish}', "");
Expect(0, 1839, '\P{Joining_Group:   -	dalath_Rish}', "");
Expect(1, 1839, '\P{^Joining_Group:   -	dalath_Rish}', "");
Expect(0, 1840, '\p{Joining_Group:   -	dalath_Rish}', "");
Expect(1, 1840, '\p{^Joining_Group:   -	dalath_Rish}', "");
Expect(1, 1840, '\P{Joining_Group:   -	dalath_Rish}', "");
Expect(0, 1840, '\P{^Joining_Group:   -	dalath_Rish}', "");
Error('\p{Jg=_:=Dalath_RISH}');
Error('\P{Jg=_:=Dalath_RISH}');
Expect(1, 1839, '\p{Jg=dalathrish}', "");
Expect(0, 1839, '\p{^Jg=dalathrish}', "");
Expect(0, 1839, '\P{Jg=dalathrish}', "");
Expect(1, 1839, '\P{^Jg=dalathrish}', "");
Expect(0, 1840, '\p{Jg=dalathrish}', "");
Expect(1, 1840, '\p{^Jg=dalathrish}', "");
Expect(1, 1840, '\P{Jg=dalathrish}', "");
Expect(0, 1840, '\P{^Jg=dalathrish}', "");
Expect(1, 1839, '\p{Jg=-	DALATH_Rish}', "");
Expect(0, 1839, '\p{^Jg=-	DALATH_Rish}', "");
Expect(0, 1839, '\P{Jg=-	DALATH_Rish}', "");
Expect(1, 1839, '\P{^Jg=-	DALATH_Rish}', "");
Expect(0, 1840, '\p{Jg=-	DALATH_Rish}', "");
Expect(1, 1840, '\p{^Jg=-	DALATH_Rish}', "");
Expect(1, 1840, '\P{Jg=-	DALATH_Rish}', "");
Expect(0, 1840, '\P{^Jg=-	DALATH_Rish}', "");
Error('\p{Is_Joining_Group=:=_Dalath_Rish}');
Error('\P{Is_Joining_Group=:=_Dalath_Rish}');
Expect(1, 1839, '\p{Is_Joining_Group: dalathrish}', "");
Expect(0, 1839, '\p{^Is_Joining_Group: dalathrish}', "");
Expect(0, 1839, '\P{Is_Joining_Group: dalathrish}', "");
Expect(1, 1839, '\P{^Is_Joining_Group: dalathrish}', "");
Expect(0, 1840, '\p{Is_Joining_Group: dalathrish}', "");
Expect(1, 1840, '\p{^Is_Joining_Group: dalathrish}', "");
Expect(1, 1840, '\P{Is_Joining_Group: dalathrish}', "");
Expect(0, 1840, '\P{^Is_Joining_Group: dalathrish}', "");
Expect(1, 1839, '\p{Is_Joining_Group= 	DALATH_Rish}', "");
Expect(0, 1839, '\p{^Is_Joining_Group= 	DALATH_Rish}', "");
Expect(0, 1839, '\P{Is_Joining_Group= 	DALATH_Rish}', "");
Expect(1, 1839, '\P{^Is_Joining_Group= 	DALATH_Rish}', "");
Expect(0, 1840, '\p{Is_Joining_Group= 	DALATH_Rish}', "");
Expect(1, 1840, '\p{^Is_Joining_Group= 	DALATH_Rish}', "");
Expect(1, 1840, '\P{Is_Joining_Group= 	DALATH_Rish}', "");
Expect(0, 1840, '\P{^Is_Joining_Group= 	DALATH_Rish}', "");
Error('\p{Is_Jg=	_dalath_RISH:=}');
Error('\P{Is_Jg=	_dalath_RISH:=}');
Expect(1, 1839, '\p{Is_Jg=dalathrish}', "");
Expect(0, 1839, '\p{^Is_Jg=dalathrish}', "");
Expect(0, 1839, '\P{Is_Jg=dalathrish}', "");
Expect(1, 1839, '\P{^Is_Jg=dalathrish}', "");
Expect(0, 1840, '\p{Is_Jg=dalathrish}', "");
Expect(1, 1840, '\p{^Is_Jg=dalathrish}', "");
Expect(1, 1840, '\P{Is_Jg=dalathrish}', "");
Expect(0, 1840, '\P{^Is_Jg=dalathrish}', "");
Expect(1, 1839, '\p{Is_Jg=	_DALATH_rish}', "");
Expect(0, 1839, '\p{^Is_Jg=	_DALATH_rish}', "");
Expect(0, 1839, '\P{Is_Jg=	_DALATH_rish}', "");
Expect(1, 1839, '\P{^Is_Jg=	_DALATH_rish}', "");
Expect(0, 1840, '\p{Is_Jg=	_DALATH_rish}', "");
Expect(1, 1840, '\p{^Is_Jg=	_DALATH_rish}', "");
Expect(1, 1840, '\P{Is_Jg=	_DALATH_rish}', "");
Expect(0, 1840, '\P{^Is_Jg=	_DALATH_rish}', "");
Error('\p{Joining_Group=/a/__e}');
Error('\P{Joining_Group=/a/__e}');
Expect(1, 1829, '\p{Joining_Group=e}', "");
Expect(0, 1829, '\p{^Joining_Group=e}', "");
Expect(0, 1829, '\P{Joining_Group=e}', "");
Expect(1, 1829, '\P{^Joining_Group=e}', "");
Expect(0, 1830, '\p{Joining_Group=e}', "");
Expect(1, 1830, '\p{^Joining_Group=e}', "");
Expect(1, 1830, '\P{Joining_Group=e}', "");
Expect(0, 1830, '\P{^Joining_Group=e}', "");
Expect(1, 1829, '\p{Joining_Group=-E}', "");
Expect(0, 1829, '\p{^Joining_Group=-E}', "");
Expect(0, 1829, '\P{Joining_Group=-E}', "");
Expect(1, 1829, '\P{^Joining_Group=-E}', "");
Expect(0, 1830, '\p{Joining_Group=-E}', "");
Expect(1, 1830, '\p{^Joining_Group=-E}', "");
Expect(1, 1830, '\P{Joining_Group=-E}', "");
Expect(0, 1830, '\P{^Joining_Group=-E}', "");
Error('\p{Jg=E/a/}');
Error('\P{Jg=E/a/}');
Expect(1, 1829, '\p{Jg=e}', "");
Expect(0, 1829, '\p{^Jg=e}', "");
Expect(0, 1829, '\P{Jg=e}', "");
Expect(1, 1829, '\P{^Jg=e}', "");
Expect(0, 1830, '\p{Jg=e}', "");
Expect(1, 1830, '\p{^Jg=e}', "");
Expect(1, 1830, '\P{Jg=e}', "");
Expect(0, 1830, '\P{^Jg=e}', "");
Expect(1, 1829, '\p{Jg= 	E}', "");
Expect(0, 1829, '\p{^Jg= 	E}', "");
Expect(0, 1829, '\P{Jg= 	E}', "");
Expect(1, 1829, '\P{^Jg= 	E}', "");
Expect(0, 1830, '\p{Jg= 	E}', "");
Expect(1, 1830, '\p{^Jg= 	E}', "");
Expect(1, 1830, '\P{Jg= 	E}', "");
Expect(0, 1830, '\P{^Jg= 	E}', "");
Error('\p{Is_Joining_Group=_:=E}');
Error('\P{Is_Joining_Group=_:=E}');
Expect(1, 1829, '\p{Is_Joining_Group=e}', "");
Expect(0, 1829, '\p{^Is_Joining_Group=e}', "");
Expect(0, 1829, '\P{Is_Joining_Group=e}', "");
Expect(1, 1829, '\P{^Is_Joining_Group=e}', "");
Expect(0, 1830, '\p{Is_Joining_Group=e}', "");
Expect(1, 1830, '\p{^Is_Joining_Group=e}', "");
Expect(1, 1830, '\P{Is_Joining_Group=e}', "");
Expect(0, 1830, '\P{^Is_Joining_Group=e}', "");
Expect(1, 1829, '\p{Is_Joining_Group=	_E}', "");
Expect(0, 1829, '\p{^Is_Joining_Group=	_E}', "");
Expect(0, 1829, '\P{Is_Joining_Group=	_E}', "");
Expect(1, 1829, '\P{^Is_Joining_Group=	_E}', "");
Expect(0, 1830, '\p{Is_Joining_Group=	_E}', "");
Expect(1, 1830, '\p{^Is_Joining_Group=	_E}', "");
Expect(1, 1830, '\P{Is_Joining_Group=	_E}', "");
Expect(0, 1830, '\P{^Is_Joining_Group=	_E}', "");
Error('\p{Is_Jg=/a/--E}');
Error('\P{Is_Jg=/a/--E}');
Expect(1, 1829, '\p{Is_Jg=e}', "");
Expect(0, 1829, '\p{^Is_Jg=e}', "");
Expect(0, 1829, '\P{Is_Jg=e}', "");
Expect(1, 1829, '\P{^Is_Jg=e}', "");
Expect(0, 1830, '\p{Is_Jg=e}', "");
Expect(1, 1830, '\p{^Is_Jg=e}', "");
Expect(1, 1830, '\P{Is_Jg=e}', "");
Expect(0, 1830, '\P{^Is_Jg=e}', "");
Expect(1, 1829, '\p{Is_Jg= E}', "");
Expect(0, 1829, '\p{^Is_Jg= E}', "");
Expect(0, 1829, '\P{Is_Jg= E}', "");
Expect(1, 1829, '\P{^Is_Jg= E}', "");
Expect(0, 1830, '\p{Is_Jg= E}', "");
Expect(1, 1830, '\p{^Is_Jg= E}', "");
Expect(1, 1830, '\P{Is_Jg= E}', "");
Expect(0, 1830, '\P{^Is_Jg= E}', "");
Error('\p{Joining_Group=-FARSI_yeh/a/}');
Error('\P{Joining_Group=-FARSI_yeh/a/}');
Expect(1, 1910, '\p{Joining_Group=farsiyeh}', "");
Expect(0, 1910, '\p{^Joining_Group=farsiyeh}', "");
Expect(0, 1910, '\P{Joining_Group=farsiyeh}', "");
Expect(1, 1910, '\P{^Joining_Group=farsiyeh}', "");
Expect(0, 1911, '\p{Joining_Group=farsiyeh}', "");
Expect(1, 1911, '\p{^Joining_Group=farsiyeh}', "");
Expect(1, 1911, '\P{Joining_Group=farsiyeh}', "");
Expect(0, 1911, '\P{^Joining_Group=farsiyeh}', "");
Expect(1, 1910, '\p{Joining_Group=-	Farsi_YEH}', "");
Expect(0, 1910, '\p{^Joining_Group=-	Farsi_YEH}', "");
Expect(0, 1910, '\P{Joining_Group=-	Farsi_YEH}', "");
Expect(1, 1910, '\P{^Joining_Group=-	Farsi_YEH}', "");
Expect(0, 1911, '\p{Joining_Group=-	Farsi_YEH}', "");
Expect(1, 1911, '\p{^Joining_Group=-	Farsi_YEH}', "");
Expect(1, 1911, '\P{Joining_Group=-	Farsi_YEH}', "");
Expect(0, 1911, '\P{^Joining_Group=-	Farsi_YEH}', "");
Error('\p{Jg=-Farsi_Yeh:=}');
Error('\P{Jg=-Farsi_Yeh:=}');
Expect(1, 1910, '\p{Jg=farsiyeh}', "");
Expect(0, 1910, '\p{^Jg=farsiyeh}', "");
Expect(0, 1910, '\P{Jg=farsiyeh}', "");
Expect(1, 1910, '\P{^Jg=farsiyeh}', "");
Expect(0, 1911, '\p{Jg=farsiyeh}', "");
Expect(1, 1911, '\p{^Jg=farsiyeh}', "");
Expect(1, 1911, '\P{Jg=farsiyeh}', "");
Expect(0, 1911, '\P{^Jg=farsiyeh}', "");
Expect(1, 1910, '\p{Jg: __farsi_yeh}', "");
Expect(0, 1910, '\p{^Jg: __farsi_yeh}', "");
Expect(0, 1910, '\P{Jg: __farsi_yeh}', "");
Expect(1, 1910, '\P{^Jg: __farsi_yeh}', "");
Expect(0, 1911, '\p{Jg: __farsi_yeh}', "");
Expect(1, 1911, '\p{^Jg: __farsi_yeh}', "");
Expect(1, 1911, '\P{Jg: __farsi_yeh}', "");
Expect(0, 1911, '\P{^Jg: __farsi_yeh}', "");
Error('\p{Is_Joining_Group= -Farsi_yeh/a/}');
Error('\P{Is_Joining_Group= -Farsi_yeh/a/}');
Expect(1, 1910, '\p{Is_Joining_Group=farsiyeh}', "");
Expect(0, 1910, '\p{^Is_Joining_Group=farsiyeh}', "");
Expect(0, 1910, '\P{Is_Joining_Group=farsiyeh}', "");
Expect(1, 1910, '\P{^Is_Joining_Group=farsiyeh}', "");
Expect(0, 1911, '\p{Is_Joining_Group=farsiyeh}', "");
Expect(1, 1911, '\p{^Is_Joining_Group=farsiyeh}', "");
Expect(1, 1911, '\P{Is_Joining_Group=farsiyeh}', "");
Expect(0, 1911, '\P{^Is_Joining_Group=farsiyeh}', "");
Expect(1, 1910, '\p{Is_Joining_Group=_farsi_YEH}', "");
Expect(0, 1910, '\p{^Is_Joining_Group=_farsi_YEH}', "");
Expect(0, 1910, '\P{Is_Joining_Group=_farsi_YEH}', "");
Expect(1, 1910, '\P{^Is_Joining_Group=_farsi_YEH}', "");
Expect(0, 1911, '\p{Is_Joining_Group=_farsi_YEH}', "");
Expect(1, 1911, '\p{^Is_Joining_Group=_farsi_YEH}', "");
Expect(1, 1911, '\P{Is_Joining_Group=_farsi_YEH}', "");
Expect(0, 1911, '\P{^Is_Joining_Group=_farsi_YEH}', "");
Error('\p{Is_Jg=_ Farsi_Yeh:=}');
Error('\P{Is_Jg=_ Farsi_Yeh:=}');
Expect(1, 1910, '\p{Is_Jg: farsiyeh}', "");
Expect(0, 1910, '\p{^Is_Jg: farsiyeh}', "");
Expect(0, 1910, '\P{Is_Jg: farsiyeh}', "");
Expect(1, 1910, '\P{^Is_Jg: farsiyeh}', "");
Expect(0, 1911, '\p{Is_Jg: farsiyeh}', "");
Expect(1, 1911, '\p{^Is_Jg: farsiyeh}', "");
Expect(1, 1911, '\P{Is_Jg: farsiyeh}', "");
Expect(0, 1911, '\P{^Is_Jg: farsiyeh}', "");
Expect(1, 1910, '\p{Is_Jg=-_FARSI_yeh}', "");
Expect(0, 1910, '\p{^Is_Jg=-_FARSI_yeh}', "");
Expect(0, 1910, '\P{Is_Jg=-_FARSI_yeh}', "");
Expect(1, 1910, '\P{^Is_Jg=-_FARSI_yeh}', "");
Expect(0, 1911, '\p{Is_Jg=-_FARSI_yeh}', "");
Expect(1, 1911, '\p{^Is_Jg=-_FARSI_yeh}', "");
Expect(1, 1911, '\P{Is_Jg=-_FARSI_yeh}', "");
Expect(0, 1911, '\P{^Is_Jg=-_FARSI_yeh}', "");
Error('\p{Joining_Group:/a/		FE}');
Error('\P{Joining_Group:/a/		FE}');
Expect(1, 1871, '\p{Joining_Group=fe}', "");
Expect(0, 1871, '\p{^Joining_Group=fe}', "");
Expect(0, 1871, '\P{Joining_Group=fe}', "");
Expect(1, 1871, '\P{^Joining_Group=fe}', "");
Expect(0, 1872, '\p{Joining_Group=fe}', "");
Expect(1, 1872, '\p{^Joining_Group=fe}', "");
Expect(1, 1872, '\P{Joining_Group=fe}', "");
Expect(0, 1872, '\P{^Joining_Group=fe}', "");
Expect(1, 1871, '\p{Joining_Group:   -fe}', "");
Expect(0, 1871, '\p{^Joining_Group:   -fe}', "");
Expect(0, 1871, '\P{Joining_Group:   -fe}', "");
Expect(1, 1871, '\P{^Joining_Group:   -fe}', "");
Expect(0, 1872, '\p{Joining_Group:   -fe}', "");
Expect(1, 1872, '\p{^Joining_Group:   -fe}', "");
Expect(1, 1872, '\P{Joining_Group:   -fe}', "");
Expect(0, 1872, '\P{^Joining_Group:   -fe}', "");
Error('\p{Jg=-/a/Fe}');
Error('\P{Jg=-/a/Fe}');
Expect(1, 1871, '\p{Jg=fe}', "");
Expect(0, 1871, '\p{^Jg=fe}', "");
Expect(0, 1871, '\P{Jg=fe}', "");
Expect(1, 1871, '\P{^Jg=fe}', "");
Expect(0, 1872, '\p{Jg=fe}', "");
Expect(1, 1872, '\p{^Jg=fe}', "");
Expect(1, 1872, '\P{Jg=fe}', "");
Expect(0, 1872, '\P{^Jg=fe}', "");
Expect(1, 1871, '\p{Jg=		Fe}', "");
Expect(0, 1871, '\p{^Jg=		Fe}', "");
Expect(0, 1871, '\P{Jg=		Fe}', "");
Expect(1, 1871, '\P{^Jg=		Fe}', "");
Expect(0, 1872, '\p{Jg=		Fe}', "");
Expect(1, 1872, '\p{^Jg=		Fe}', "");
Expect(1, 1872, '\P{Jg=		Fe}', "");
Expect(0, 1872, '\P{^Jg=		Fe}', "");
Error('\p{Is_Joining_Group=:=-	Fe}');
Error('\P{Is_Joining_Group=:=-	Fe}');
Expect(1, 1871, '\p{Is_Joining_Group=fe}', "");
Expect(0, 1871, '\p{^Is_Joining_Group=fe}', "");
Expect(0, 1871, '\P{Is_Joining_Group=fe}', "");
Expect(1, 1871, '\P{^Is_Joining_Group=fe}', "");
Expect(0, 1872, '\p{Is_Joining_Group=fe}', "");
Expect(1, 1872, '\p{^Is_Joining_Group=fe}', "");
Expect(1, 1872, '\P{Is_Joining_Group=fe}', "");
Expect(0, 1872, '\P{^Is_Joining_Group=fe}', "");
Expect(1, 1871, '\p{Is_Joining_Group=	 FE}', "");
Expect(0, 1871, '\p{^Is_Joining_Group=	 FE}', "");
Expect(0, 1871, '\P{Is_Joining_Group=	 FE}', "");
Expect(1, 1871, '\P{^Is_Joining_Group=	 FE}', "");
Expect(0, 1872, '\p{Is_Joining_Group=	 FE}', "");
Expect(1, 1872, '\p{^Is_Joining_Group=	 FE}', "");
Expect(1, 1872, '\P{Is_Joining_Group=	 FE}', "");
Expect(0, 1872, '\P{^Is_Joining_Group=	 FE}', "");
Error('\p{Is_Jg=/a/Fe}');
Error('\P{Is_Jg=/a/Fe}');
Expect(1, 1871, '\p{Is_Jg=fe}', "");
Expect(0, 1871, '\p{^Is_Jg=fe}', "");
Expect(0, 1871, '\P{Is_Jg=fe}', "");
Expect(1, 1871, '\P{^Is_Jg=fe}', "");
Expect(0, 1872, '\p{Is_Jg=fe}', "");
Expect(1, 1872, '\p{^Is_Jg=fe}', "");
Expect(1, 1872, '\P{Is_Jg=fe}', "");
Expect(0, 1872, '\P{^Is_Jg=fe}', "");
Expect(1, 1871, '\p{Is_Jg=	 fe}', "");
Expect(0, 1871, '\p{^Is_Jg=	 fe}', "");
Expect(0, 1871, '\P{Is_Jg=	 fe}', "");
Expect(1, 1871, '\P{^Is_Jg=	 fe}', "");
Expect(0, 1872, '\p{Is_Jg=	 fe}', "");
Expect(1, 1872, '\p{^Is_Jg=	 fe}', "");
Expect(1, 1872, '\P{Is_Jg=	 fe}', "");
Expect(0, 1872, '\P{^Is_Jg=	 fe}', "");
Error('\p{Joining_Group:   -:=feh}');
Error('\P{Joining_Group:   -:=feh}');
Expect(1, 2212, '\p{Joining_Group=feh}', "");
Expect(0, 2212, '\p{^Joining_Group=feh}', "");
Expect(0, 2212, '\P{Joining_Group=feh}', "");
Expect(1, 2212, '\P{^Joining_Group=feh}', "");
Expect(0, 2213, '\p{Joining_Group=feh}', "");
Expect(1, 2213, '\p{^Joining_Group=feh}', "");
Expect(1, 2213, '\P{Joining_Group=feh}', "");
Expect(0, 2213, '\P{^Joining_Group=feh}', "");
Expect(1, 2212, '\p{Joining_Group=-_FEH}', "");
Expect(0, 2212, '\p{^Joining_Group=-_FEH}', "");
Expect(0, 2212, '\P{Joining_Group=-_FEH}', "");
Expect(1, 2212, '\P{^Joining_Group=-_FEH}', "");
Expect(0, 2213, '\p{Joining_Group=-_FEH}', "");
Expect(1, 2213, '\p{^Joining_Group=-_FEH}', "");
Expect(1, 2213, '\P{Joining_Group=-_FEH}', "");
Expect(0, 2213, '\P{^Joining_Group=-_FEH}', "");
Error('\p{Jg=:=	 Feh}');
Error('\P{Jg=:=	 Feh}');
Expect(1, 2212, '\p{Jg=feh}', "");
Expect(0, 2212, '\p{^Jg=feh}', "");
Expect(0, 2212, '\P{Jg=feh}', "");
Expect(1, 2212, '\P{^Jg=feh}', "");
Expect(0, 2213, '\p{Jg=feh}', "");
Expect(1, 2213, '\p{^Jg=feh}', "");
Expect(1, 2213, '\P{Jg=feh}', "");
Expect(0, 2213, '\P{^Jg=feh}', "");
Expect(1, 2212, '\p{Jg:  FEH}', "");
Expect(0, 2212, '\p{^Jg:  FEH}', "");
Expect(0, 2212, '\P{Jg:  FEH}', "");
Expect(1, 2212, '\P{^Jg:  FEH}', "");
Expect(0, 2213, '\p{Jg:  FEH}', "");
Expect(1, 2213, '\p{^Jg:  FEH}', "");
Expect(1, 2213, '\P{Jg:  FEH}', "");
Expect(0, 2213, '\P{^Jg:  FEH}', "");
Error('\p{Is_Joining_Group=:=__Feh}');
Error('\P{Is_Joining_Group=:=__Feh}');
Expect(1, 2212, '\p{Is_Joining_Group=feh}', "");
Expect(0, 2212, '\p{^Is_Joining_Group=feh}', "");
Expect(0, 2212, '\P{Is_Joining_Group=feh}', "");
Expect(1, 2212, '\P{^Is_Joining_Group=feh}', "");
Expect(0, 2213, '\p{Is_Joining_Group=feh}', "");
Expect(1, 2213, '\p{^Is_Joining_Group=feh}', "");
Expect(1, 2213, '\P{Is_Joining_Group=feh}', "");
Expect(0, 2213, '\P{^Is_Joining_Group=feh}', "");
Expect(1, 2212, '\p{Is_Joining_Group=- Feh}', "");
Expect(0, 2212, '\p{^Is_Joining_Group=- Feh}', "");
Expect(0, 2212, '\P{Is_Joining_Group=- Feh}', "");
Expect(1, 2212, '\P{^Is_Joining_Group=- Feh}', "");
Expect(0, 2213, '\p{Is_Joining_Group=- Feh}', "");
Expect(1, 2213, '\p{^Is_Joining_Group=- Feh}', "");
Expect(1, 2213, '\P{Is_Joining_Group=- Feh}', "");
Expect(0, 2213, '\P{^Is_Joining_Group=- Feh}', "");
Error('\p{Is_Jg=/a/Feh}');
Error('\P{Is_Jg=/a/Feh}');
Expect(1, 2212, '\p{Is_Jg=feh}', "");
Expect(0, 2212, '\p{^Is_Jg=feh}', "");
Expect(0, 2212, '\P{Is_Jg=feh}', "");
Expect(1, 2212, '\P{^Is_Jg=feh}', "");
Expect(0, 2213, '\p{Is_Jg=feh}', "");
Expect(1, 2213, '\p{^Is_Jg=feh}', "");
Expect(1, 2213, '\P{Is_Jg=feh}', "");
Expect(0, 2213, '\P{^Is_Jg=feh}', "");
Expect(1, 2212, '\p{Is_Jg=--feh}', "");
Expect(0, 2212, '\p{^Is_Jg=--feh}', "");
Expect(0, 2212, '\P{Is_Jg=--feh}', "");
Expect(1, 2212, '\P{^Is_Jg=--feh}', "");
Expect(0, 2213, '\p{Is_Jg=--feh}', "");
Expect(1, 2213, '\p{^Is_Jg=--feh}', "");
Expect(1, 2213, '\P{Is_Jg=--feh}', "");
Expect(0, 2213, '\P{^Is_Jg=--feh}', "");
Error('\p{Joining_Group=_:=Final_semkath}');
Error('\P{Joining_Group=_:=Final_semkath}');
Expect(1, 1828, '\p{Joining_Group=finalsemkath}', "");
Expect(0, 1828, '\p{^Joining_Group=finalsemkath}', "");
Expect(0, 1828, '\P{Joining_Group=finalsemkath}', "");
Expect(1, 1828, '\P{^Joining_Group=finalsemkath}', "");
Expect(0, 1829, '\p{Joining_Group=finalsemkath}', "");
Expect(1, 1829, '\p{^Joining_Group=finalsemkath}', "");
Expect(1, 1829, '\P{Joining_Group=finalsemkath}', "");
Expect(0, 1829, '\P{^Joining_Group=finalsemkath}', "");
Expect(1, 1828, '\p{Joining_Group=-final_Semkath}', "");
Expect(0, 1828, '\p{^Joining_Group=-final_Semkath}', "");
Expect(0, 1828, '\P{Joining_Group=-final_Semkath}', "");
Expect(1, 1828, '\P{^Joining_Group=-final_Semkath}', "");
Expect(0, 1829, '\p{Joining_Group=-final_Semkath}', "");
Expect(1, 1829, '\p{^Joining_Group=-final_Semkath}', "");
Expect(1, 1829, '\P{Joining_Group=-final_Semkath}', "");
Expect(0, 1829, '\P{^Joining_Group=-final_Semkath}', "");
Error('\p{Jg=	:=Final_semkath}');
Error('\P{Jg=	:=Final_semkath}');
Expect(1, 1828, '\p{Jg:finalsemkath}', "");
Expect(0, 1828, '\p{^Jg:finalsemkath}', "");
Expect(0, 1828, '\P{Jg:finalsemkath}', "");
Expect(1, 1828, '\P{^Jg:finalsemkath}', "");
Expect(0, 1829, '\p{Jg:finalsemkath}', "");
Expect(1, 1829, '\p{^Jg:finalsemkath}', "");
Expect(1, 1829, '\P{Jg:finalsemkath}', "");
Expect(0, 1829, '\P{^Jg:finalsemkath}', "");
Expect(1, 1828, '\p{Jg= _final_Semkath}', "");
Expect(0, 1828, '\p{^Jg= _final_Semkath}', "");
Expect(0, 1828, '\P{Jg= _final_Semkath}', "");
Expect(1, 1828, '\P{^Jg= _final_Semkath}', "");
Expect(0, 1829, '\p{Jg= _final_Semkath}', "");
Expect(1, 1829, '\p{^Jg= _final_Semkath}', "");
Expect(1, 1829, '\P{Jg= _final_Semkath}', "");
Expect(0, 1829, '\P{^Jg= _final_Semkath}', "");
Error('\p{Is_Joining_Group= Final_Semkath:=}');
Error('\P{Is_Joining_Group= Final_Semkath:=}');
Expect(1, 1828, '\p{Is_Joining_Group=finalsemkath}', "");
Expect(0, 1828, '\p{^Is_Joining_Group=finalsemkath}', "");
Expect(0, 1828, '\P{Is_Joining_Group=finalsemkath}', "");
Expect(1, 1828, '\P{^Is_Joining_Group=finalsemkath}', "");
Expect(0, 1829, '\p{Is_Joining_Group=finalsemkath}', "");
Expect(1, 1829, '\p{^Is_Joining_Group=finalsemkath}', "");
Expect(1, 1829, '\P{Is_Joining_Group=finalsemkath}', "");
Expect(0, 1829, '\P{^Is_Joining_Group=finalsemkath}', "");
Expect(1, 1828, '\p{Is_Joining_Group=  Final_semkath}', "");
Expect(0, 1828, '\p{^Is_Joining_Group=  Final_semkath}', "");
Expect(0, 1828, '\P{Is_Joining_Group=  Final_semkath}', "");
Expect(1, 1828, '\P{^Is_Joining_Group=  Final_semkath}', "");
Expect(0, 1829, '\p{Is_Joining_Group=  Final_semkath}', "");
Expect(1, 1829, '\p{^Is_Joining_Group=  Final_semkath}', "");
Expect(1, 1829, '\P{Is_Joining_Group=  Final_semkath}', "");
Expect(0, 1829, '\P{^Is_Joining_Group=  Final_semkath}', "");
Error('\p{Is_Jg= FINAL_SEMKATH/a/}');
Error('\P{Is_Jg= FINAL_SEMKATH/a/}');
Expect(1, 1828, '\p{Is_Jg=finalsemkath}', "");
Expect(0, 1828, '\p{^Is_Jg=finalsemkath}', "");
Expect(0, 1828, '\P{Is_Jg=finalsemkath}', "");
Expect(1, 1828, '\P{^Is_Jg=finalsemkath}', "");
Expect(0, 1829, '\p{Is_Jg=finalsemkath}', "");
Expect(1, 1829, '\p{^Is_Jg=finalsemkath}', "");
Expect(1, 1829, '\P{Is_Jg=finalsemkath}', "");
Expect(0, 1829, '\P{^Is_Jg=finalsemkath}', "");
Expect(1, 1828, '\p{Is_Jg= final_Semkath}', "");
Expect(0, 1828, '\p{^Is_Jg= final_Semkath}', "");
Expect(0, 1828, '\P{Is_Jg= final_Semkath}', "");
Expect(1, 1828, '\P{^Is_Jg= final_Semkath}', "");
Expect(0, 1829, '\p{Is_Jg= final_Semkath}', "");
Expect(1, 1829, '\p{^Is_Jg= final_Semkath}', "");
Expect(1, 1829, '\P{Is_Jg= final_Semkath}', "");
Expect(0, 1829, '\P{^Is_Jg= final_Semkath}', "");
Error('\p{Joining_Group=:=gaf}');
Error('\P{Joining_Group=:=gaf}');
Expect(1, 1892, '\p{Joining_Group=gaf}', "");
Expect(0, 1892, '\p{^Joining_Group=gaf}', "");
Expect(0, 1892, '\P{Joining_Group=gaf}', "");
Expect(1, 1892, '\P{^Joining_Group=gaf}', "");
Expect(0, 1893, '\p{Joining_Group=gaf}', "");
Expect(1, 1893, '\p{^Joining_Group=gaf}', "");
Expect(1, 1893, '\P{Joining_Group=gaf}', "");
Expect(0, 1893, '\P{^Joining_Group=gaf}', "");
Expect(1, 1892, '\p{Joining_Group=_gaf}', "");
Expect(0, 1892, '\p{^Joining_Group=_gaf}', "");
Expect(0, 1892, '\P{Joining_Group=_gaf}', "");
Expect(1, 1892, '\P{^Joining_Group=_gaf}', "");
Expect(0, 1893, '\p{Joining_Group=_gaf}', "");
Expect(1, 1893, '\p{^Joining_Group=_gaf}', "");
Expect(1, 1893, '\P{Joining_Group=_gaf}', "");
Expect(0, 1893, '\P{^Joining_Group=_gaf}', "");
Error('\p{Jg=/a/-_gaf}');
Error('\P{Jg=/a/-_gaf}');
Expect(1, 1892, '\p{Jg=gaf}', "");
Expect(0, 1892, '\p{^Jg=gaf}', "");
Expect(0, 1892, '\P{Jg=gaf}', "");
Expect(1, 1892, '\P{^Jg=gaf}', "");
Expect(0, 1893, '\p{Jg=gaf}', "");
Expect(1, 1893, '\p{^Jg=gaf}', "");
Expect(1, 1893, '\P{Jg=gaf}', "");
Expect(0, 1893, '\P{^Jg=gaf}', "");
Expect(1, 1892, '\p{Jg:- Gaf}', "");
Expect(0, 1892, '\p{^Jg:- Gaf}', "");
Expect(0, 1892, '\P{Jg:- Gaf}', "");
Expect(1, 1892, '\P{^Jg:- Gaf}', "");
Expect(0, 1893, '\p{Jg:- Gaf}', "");
Expect(1, 1893, '\p{^Jg:- Gaf}', "");
Expect(1, 1893, '\P{Jg:- Gaf}', "");
Expect(0, 1893, '\P{^Jg:- Gaf}', "");
Error('\p{Is_Joining_Group=-/a/Gaf}');
Error('\P{Is_Joining_Group=-/a/Gaf}');
Expect(1, 1892, '\p{Is_Joining_Group=gaf}', "");
Expect(0, 1892, '\p{^Is_Joining_Group=gaf}', "");
Expect(0, 1892, '\P{Is_Joining_Group=gaf}', "");
Expect(1, 1892, '\P{^Is_Joining_Group=gaf}', "");
Expect(0, 1893, '\p{Is_Joining_Group=gaf}', "");
Expect(1, 1893, '\p{^Is_Joining_Group=gaf}', "");
Expect(1, 1893, '\P{Is_Joining_Group=gaf}', "");
Expect(0, 1893, '\P{^Is_Joining_Group=gaf}', "");
Expect(1, 1892, '\p{Is_Joining_Group=__gaf}', "");
Expect(0, 1892, '\p{^Is_Joining_Group=__gaf}', "");
Expect(0, 1892, '\P{Is_Joining_Group=__gaf}', "");
Expect(1, 1892, '\P{^Is_Joining_Group=__gaf}', "");
Expect(0, 1893, '\p{Is_Joining_Group=__gaf}', "");
Expect(1, 1893, '\p{^Is_Joining_Group=__gaf}', "");
Expect(1, 1893, '\P{Is_Joining_Group=__gaf}', "");
Expect(0, 1893, '\P{^Is_Joining_Group=__gaf}', "");
Error('\p{Is_Jg=_/a/Gaf}');
Error('\P{Is_Jg=_/a/Gaf}');
Expect(1, 1892, '\p{Is_Jg=gaf}', "");
Expect(0, 1892, '\p{^Is_Jg=gaf}', "");
Expect(0, 1892, '\P{Is_Jg=gaf}', "");
Expect(1, 1892, '\P{^Is_Jg=gaf}', "");
Expect(0, 1893, '\p{Is_Jg=gaf}', "");
Expect(1, 1893, '\p{^Is_Jg=gaf}', "");
Expect(1, 1893, '\P{Is_Jg=gaf}', "");
Expect(0, 1893, '\P{^Is_Jg=gaf}', "");
Error('\p{Joining_Group=	Gamal:=}');
Error('\P{Joining_Group=	Gamal:=}');
Expect(1, 1838, '\p{Joining_Group:gamal}', "");
Expect(0, 1838, '\p{^Joining_Group:gamal}', "");
Expect(0, 1838, '\P{Joining_Group:gamal}', "");
Expect(1, 1838, '\P{^Joining_Group:gamal}', "");
Expect(0, 1839, '\p{Joining_Group:gamal}', "");
Expect(1, 1839, '\p{^Joining_Group:gamal}', "");
Expect(1, 1839, '\P{Joining_Group:gamal}', "");
Expect(0, 1839, '\P{^Joining_Group:gamal}', "");
Expect(1, 1838, '\p{Joining_Group= _gamal}', "");
Expect(0, 1838, '\p{^Joining_Group= _gamal}', "");
Expect(0, 1838, '\P{Joining_Group= _gamal}', "");
Expect(1, 1838, '\P{^Joining_Group= _gamal}', "");
Expect(0, 1839, '\p{Joining_Group= _gamal}', "");
Expect(1, 1839, '\p{^Joining_Group= _gamal}', "");
Expect(1, 1839, '\P{Joining_Group= _gamal}', "");
Expect(0, 1839, '\P{^Joining_Group= _gamal}', "");
Error('\p{Jg=_-Gamal:=}');
Error('\P{Jg=_-Gamal:=}');
Expect(1, 1838, '\p{Jg=gamal}', "");
Expect(0, 1838, '\p{^Jg=gamal}', "");
Expect(0, 1838, '\P{Jg=gamal}', "");
Expect(1, 1838, '\P{^Jg=gamal}', "");
Expect(0, 1839, '\p{Jg=gamal}', "");
Expect(1, 1839, '\p{^Jg=gamal}', "");
Expect(1, 1839, '\P{Jg=gamal}', "");
Expect(0, 1839, '\P{^Jg=gamal}', "");
Expect(1, 1838, '\p{Jg=	-Gamal}', "");
Expect(0, 1838, '\p{^Jg=	-Gamal}', "");
Expect(0, 1838, '\P{Jg=	-Gamal}', "");
Expect(1, 1838, '\P{^Jg=	-Gamal}', "");
Expect(0, 1839, '\p{Jg=	-Gamal}', "");
Expect(1, 1839, '\p{^Jg=	-Gamal}', "");
Expect(1, 1839, '\P{Jg=	-Gamal}', "");
Expect(0, 1839, '\P{^Jg=	-Gamal}', "");
Error('\p{Is_Joining_Group:/a/-	gamal}');
Error('\P{Is_Joining_Group:/a/-	gamal}');
Expect(1, 1838, '\p{Is_Joining_Group:	gamal}', "");
Expect(0, 1838, '\p{^Is_Joining_Group:	gamal}', "");
Expect(0, 1838, '\P{Is_Joining_Group:	gamal}', "");
Expect(1, 1838, '\P{^Is_Joining_Group:	gamal}', "");
Expect(0, 1839, '\p{Is_Joining_Group:	gamal}', "");
Expect(1, 1839, '\p{^Is_Joining_Group:	gamal}', "");
Expect(1, 1839, '\P{Is_Joining_Group:	gamal}', "");
Expect(0, 1839, '\P{^Is_Joining_Group:	gamal}', "");
Expect(1, 1838, '\p{Is_Joining_Group=	_Gamal}', "");
Expect(0, 1838, '\p{^Is_Joining_Group=	_Gamal}', "");
Expect(0, 1838, '\P{Is_Joining_Group=	_Gamal}', "");
Expect(1, 1838, '\P{^Is_Joining_Group=	_Gamal}', "");
Expect(0, 1839, '\p{Is_Joining_Group=	_Gamal}', "");
Expect(1, 1839, '\p{^Is_Joining_Group=	_Gamal}', "");
Expect(1, 1839, '\P{Is_Joining_Group=	_Gamal}', "");
Expect(0, 1839, '\P{^Is_Joining_Group=	_Gamal}', "");
Error('\p{Is_Jg=:=GAMAL}');
Error('\P{Is_Jg=:=GAMAL}');
Expect(1, 1838, '\p{Is_Jg=gamal}', "");
Expect(0, 1838, '\p{^Is_Jg=gamal}', "");
Expect(0, 1838, '\P{Is_Jg=gamal}', "");
Expect(1, 1838, '\P{^Is_Jg=gamal}', "");
Expect(0, 1839, '\p{Is_Jg=gamal}', "");
Expect(1, 1839, '\p{^Is_Jg=gamal}', "");
Expect(1, 1839, '\P{Is_Jg=gamal}', "");
Expect(0, 1839, '\P{^Is_Jg=gamal}', "");
Expect(1, 1838, '\p{Is_Jg= gamal}', "");
Expect(0, 1838, '\p{^Is_Jg= gamal}', "");
Expect(0, 1838, '\P{Is_Jg= gamal}', "");
Expect(1, 1838, '\P{^Is_Jg= gamal}', "");
Expect(0, 1839, '\p{Is_Jg= gamal}', "");
Expect(1, 1839, '\p{^Is_Jg= gamal}', "");
Expect(1, 1839, '\P{Is_Jg= gamal}', "");
Expect(0, 1839, '\P{^Is_Jg= gamal}', "");
Error('\p{Joining_Group=--hah/a/}');
Error('\P{Joining_Group=--hah/a/}');
Expect(1, 2210, '\p{Joining_Group=hah}', "");
Expect(0, 2210, '\p{^Joining_Group=hah}', "");
Expect(0, 2210, '\P{Joining_Group=hah}', "");
Expect(1, 2210, '\P{^Joining_Group=hah}', "");
Expect(0, 2211, '\p{Joining_Group=hah}', "");
Expect(1, 2211, '\p{^Joining_Group=hah}', "");
Expect(1, 2211, '\P{Joining_Group=hah}', "");
Expect(0, 2211, '\P{^Joining_Group=hah}', "");
Expect(1, 2210, '\p{Joining_Group=__hah}', "");
Expect(0, 2210, '\p{^Joining_Group=__hah}', "");
Expect(0, 2210, '\P{Joining_Group=__hah}', "");
Expect(1, 2210, '\P{^Joining_Group=__hah}', "");
Expect(0, 2211, '\p{Joining_Group=__hah}', "");
Expect(1, 2211, '\p{^Joining_Group=__hah}', "");
Expect(1, 2211, '\P{Joining_Group=__hah}', "");
Expect(0, 2211, '\P{^Joining_Group=__hah}', "");
Error('\p{Jg=:=	_Hah}');
Error('\P{Jg=:=	_Hah}');
Expect(1, 2210, '\p{Jg=hah}', "");
Expect(0, 2210, '\p{^Jg=hah}', "");
Expect(0, 2210, '\P{Jg=hah}', "");
Expect(1, 2210, '\P{^Jg=hah}', "");
Expect(0, 2211, '\p{Jg=hah}', "");
Expect(1, 2211, '\p{^Jg=hah}', "");
Expect(1, 2211, '\P{Jg=hah}', "");
Expect(0, 2211, '\P{^Jg=hah}', "");
Expect(1, 2210, '\p{Jg=-hah}', "");
Expect(0, 2210, '\p{^Jg=-hah}', "");
Expect(0, 2210, '\P{Jg=-hah}', "");
Expect(1, 2210, '\P{^Jg=-hah}', "");
Expect(0, 2211, '\p{Jg=-hah}', "");
Expect(1, 2211, '\p{^Jg=-hah}', "");
Expect(1, 2211, '\P{Jg=-hah}', "");
Expect(0, 2211, '\P{^Jg=-hah}', "");
Error('\p{Is_Joining_Group=/a/ _Hah}');
Error('\P{Is_Joining_Group=/a/ _Hah}');
Expect(1, 2210, '\p{Is_Joining_Group=hah}', "");
Expect(0, 2210, '\p{^Is_Joining_Group=hah}', "");
Expect(0, 2210, '\P{Is_Joining_Group=hah}', "");
Expect(1, 2210, '\P{^Is_Joining_Group=hah}', "");
Expect(0, 2211, '\p{Is_Joining_Group=hah}', "");
Expect(1, 2211, '\p{^Is_Joining_Group=hah}', "");
Expect(1, 2211, '\P{Is_Joining_Group=hah}', "");
Expect(0, 2211, '\P{^Is_Joining_Group=hah}', "");
Expect(1, 2210, '\p{Is_Joining_Group=	 HAH}', "");
Expect(0, 2210, '\p{^Is_Joining_Group=	 HAH}', "");
Expect(0, 2210, '\P{Is_Joining_Group=	 HAH}', "");
Expect(1, 2210, '\P{^Is_Joining_Group=	 HAH}', "");
Expect(0, 2211, '\p{Is_Joining_Group=	 HAH}', "");
Expect(1, 2211, '\p{^Is_Joining_Group=	 HAH}', "");
Expect(1, 2211, '\P{Is_Joining_Group=	 HAH}', "");
Expect(0, 2211, '\P{^Is_Joining_Group=	 HAH}', "");
Error('\p{Is_Jg:    /a/HAH}');
Error('\P{Is_Jg:    /a/HAH}');
Expect(1, 2210, '\p{Is_Jg=hah}', "");
Expect(0, 2210, '\p{^Is_Jg=hah}', "");
Expect(0, 2210, '\P{Is_Jg=hah}', "");
Expect(1, 2210, '\P{^Is_Jg=hah}', "");
Expect(0, 2211, '\p{Is_Jg=hah}', "");
Expect(1, 2211, '\p{^Is_Jg=hah}', "");
Expect(1, 2211, '\P{Is_Jg=hah}', "");
Expect(0, 2211, '\P{^Is_Jg=hah}', "");
Expect(1, 2210, '\p{Is_Jg=		HAH}', "");
Expect(0, 2210, '\p{^Is_Jg=		HAH}', "");
Expect(0, 2210, '\P{Is_Jg=		HAH}', "");
Expect(1, 2210, '\P{^Is_Jg=		HAH}', "");
Expect(0, 2211, '\p{Is_Jg=		HAH}', "");
Expect(1, 2211, '\p{^Is_Jg=		HAH}', "");
Expect(1, 2211, '\P{Is_Jg=		HAH}', "");
Expect(0, 2211, '\P{^Is_Jg=		HAH}', "");
Error('\p{Joining_Group=/a/He}');
Error('\P{Joining_Group=/a/He}');
Expect(1, 1815, '\p{Joining_Group:	he}', "");
Expect(0, 1815, '\p{^Joining_Group:	he}', "");
Expect(0, 1815, '\P{Joining_Group:	he}', "");
Expect(1, 1815, '\P{^Joining_Group:	he}', "");
Expect(0, 1816, '\p{Joining_Group:	he}', "");
Expect(1, 1816, '\p{^Joining_Group:	he}', "");
Expect(1, 1816, '\P{Joining_Group:	he}', "");
Expect(0, 1816, '\P{^Joining_Group:	he}', "");
Expect(1, 1815, '\p{Joining_Group=	_HE}', "");
Expect(0, 1815, '\p{^Joining_Group=	_HE}', "");
Expect(0, 1815, '\P{Joining_Group=	_HE}', "");
Expect(1, 1815, '\P{^Joining_Group=	_HE}', "");
Expect(0, 1816, '\p{Joining_Group=	_HE}', "");
Expect(1, 1816, '\p{^Joining_Group=	_HE}', "");
Expect(1, 1816, '\P{Joining_Group=	_HE}', "");
Expect(0, 1816, '\P{^Joining_Group=	_HE}', "");
Error('\p{Jg:	 HE:=}');
Error('\P{Jg:	 HE:=}');
Expect(1, 1815, '\p{Jg=he}', "");
Expect(0, 1815, '\p{^Jg=he}', "");
Expect(0, 1815, '\P{Jg=he}', "");
Expect(1, 1815, '\P{^Jg=he}', "");
Expect(0, 1816, '\p{Jg=he}', "");
Expect(1, 1816, '\p{^Jg=he}', "");
Expect(1, 1816, '\P{Jg=he}', "");
Expect(0, 1816, '\P{^Jg=he}', "");
Expect(1, 1815, '\p{Jg:    _he}', "");
Expect(0, 1815, '\p{^Jg:    _he}', "");
Expect(0, 1815, '\P{Jg:    _he}', "");
Expect(1, 1815, '\P{^Jg:    _he}', "");
Expect(0, 1816, '\p{Jg:    _he}', "");
Expect(1, 1816, '\p{^Jg:    _he}', "");
Expect(1, 1816, '\P{Jg:    _he}', "");
Expect(0, 1816, '\P{^Jg:    _he}', "");
Error('\p{Is_Joining_Group:	:= _He}');
Error('\P{Is_Joining_Group:	:= _He}');
Expect(1, 1815, '\p{Is_Joining_Group=he}', "");
Expect(0, 1815, '\p{^Is_Joining_Group=he}', "");
Expect(0, 1815, '\P{Is_Joining_Group=he}', "");
Expect(1, 1815, '\P{^Is_Joining_Group=he}', "");
Expect(0, 1816, '\p{Is_Joining_Group=he}', "");
Expect(1, 1816, '\p{^Is_Joining_Group=he}', "");
Expect(1, 1816, '\P{Is_Joining_Group=he}', "");
Expect(0, 1816, '\P{^Is_Joining_Group=he}', "");
Expect(1, 1815, '\p{Is_Joining_Group=-He}', "");
Expect(0, 1815, '\p{^Is_Joining_Group=-He}', "");
Expect(0, 1815, '\P{Is_Joining_Group=-He}', "");
Expect(1, 1815, '\P{^Is_Joining_Group=-He}', "");
Expect(0, 1816, '\p{Is_Joining_Group=-He}', "");
Expect(1, 1816, '\p{^Is_Joining_Group=-He}', "");
Expect(1, 1816, '\P{Is_Joining_Group=-He}', "");
Expect(0, 1816, '\P{^Is_Joining_Group=-He}', "");
Error('\p{Is_Jg=__he:=}');
Error('\P{Is_Jg=__he:=}');
Expect(1, 1815, '\p{Is_Jg:	he}', "");
Expect(0, 1815, '\p{^Is_Jg:	he}', "");
Expect(0, 1815, '\P{Is_Jg:	he}', "");
Expect(1, 1815, '\P{^Is_Jg:	he}', "");
Expect(0, 1816, '\p{Is_Jg:	he}', "");
Expect(1, 1816, '\p{^Is_Jg:	he}', "");
Expect(1, 1816, '\P{Is_Jg:	he}', "");
Expect(0, 1816, '\P{^Is_Jg:	he}', "");
Expect(1, 1815, '\p{Is_Jg=	 He}', "");
Expect(0, 1815, '\p{^Is_Jg=	 He}', "");
Expect(0, 1815, '\P{Is_Jg=	 He}', "");
Expect(1, 1815, '\P{^Is_Jg=	 He}', "");
Expect(0, 1816, '\p{Is_Jg=	 He}', "");
Expect(1, 1816, '\p{^Is_Jg=	 He}', "");
Expect(1, 1816, '\P{Is_Jg=	 He}', "");
Expect(0, 1816, '\P{^Is_Jg=	 He}', "");
Error('\p{Joining_Group: :=	_Heh}');
Error('\P{Joining_Group: :=	_Heh}');
Expect(1, 1607, '\p{Joining_Group=heh}', "");
Expect(0, 1607, '\p{^Joining_Group=heh}', "");
Expect(0, 1607, '\P{Joining_Group=heh}', "");
Expect(1, 1607, '\P{^Joining_Group=heh}', "");
Expect(0, 1608, '\p{Joining_Group=heh}', "");
Expect(1, 1608, '\p{^Joining_Group=heh}', "");
Expect(1, 1608, '\P{Joining_Group=heh}', "");
Expect(0, 1608, '\P{^Joining_Group=heh}', "");
Expect(1, 1607, '\p{Joining_Group: 	HEH}', "");
Expect(0, 1607, '\p{^Joining_Group: 	HEH}', "");
Expect(0, 1607, '\P{Joining_Group: 	HEH}', "");
Expect(1, 1607, '\P{^Joining_Group: 	HEH}', "");
Expect(0, 1608, '\p{Joining_Group: 	HEH}', "");
Expect(1, 1608, '\p{^Joining_Group: 	HEH}', "");
Expect(1, 1608, '\P{Joining_Group: 	HEH}', "");
Expect(0, 1608, '\P{^Joining_Group: 	HEH}', "");
Error('\p{Jg= Heh:=}');
Error('\P{Jg= Heh:=}');
Expect(1, 1607, '\p{Jg:heh}', "");
Expect(0, 1607, '\p{^Jg:heh}', "");
Expect(0, 1607, '\P{Jg:heh}', "");
Expect(1, 1607, '\P{^Jg:heh}', "");
Expect(0, 1608, '\p{Jg:heh}', "");
Expect(1, 1608, '\p{^Jg:heh}', "");
Expect(1, 1608, '\P{Jg:heh}', "");
Expect(0, 1608, '\P{^Jg:heh}', "");
Expect(1, 1607, '\p{Jg= -heh}', "");
Expect(0, 1607, '\p{^Jg= -heh}', "");
Expect(0, 1607, '\P{Jg= -heh}', "");
Expect(1, 1607, '\P{^Jg= -heh}', "");
Expect(0, 1608, '\p{Jg= -heh}', "");
Expect(1, 1608, '\p{^Jg= -heh}', "");
Expect(1, 1608, '\P{Jg= -heh}', "");
Expect(0, 1608, '\P{^Jg= -heh}', "");
Error('\p{Is_Joining_Group=/a/HEH}');
Error('\P{Is_Joining_Group=/a/HEH}');
Expect(1, 1607, '\p{Is_Joining_Group=heh}', "");
Expect(0, 1607, '\p{^Is_Joining_Group=heh}', "");
Expect(0, 1607, '\P{Is_Joining_Group=heh}', "");
Expect(1, 1607, '\P{^Is_Joining_Group=heh}', "");
Expect(0, 1608, '\p{Is_Joining_Group=heh}', "");
Expect(1, 1608, '\p{^Is_Joining_Group=heh}', "");
Expect(1, 1608, '\P{Is_Joining_Group=heh}', "");
Expect(0, 1608, '\P{^Is_Joining_Group=heh}', "");
Expect(1, 1607, '\p{Is_Joining_Group=	 heh}', "");
Expect(0, 1607, '\p{^Is_Joining_Group=	 heh}', "");
Expect(0, 1607, '\P{Is_Joining_Group=	 heh}', "");
Expect(1, 1607, '\P{^Is_Joining_Group=	 heh}', "");
Expect(0, 1608, '\p{Is_Joining_Group=	 heh}', "");
Expect(1, 1608, '\p{^Is_Joining_Group=	 heh}', "");
Expect(1, 1608, '\P{Is_Joining_Group=	 heh}', "");
Expect(0, 1608, '\P{^Is_Joining_Group=	 heh}', "");
Error('\p{Is_Jg=/a/		heh}');
Error('\P{Is_Jg=/a/		heh}');
Expect(1, 1607, '\p{Is_Jg=heh}', "");
Expect(0, 1607, '\p{^Is_Jg=heh}', "");
Expect(0, 1607, '\P{Is_Jg=heh}', "");
Expect(1, 1607, '\P{^Is_Jg=heh}', "");
Expect(0, 1608, '\p{Is_Jg=heh}', "");
Expect(1, 1608, '\p{^Is_Jg=heh}', "");
Expect(1, 1608, '\P{Is_Jg=heh}', "");
Expect(0, 1608, '\P{^Is_Jg=heh}', "");
Expect(1, 1607, '\p{Is_Jg: 	_HEH}', "");
Expect(0, 1607, '\p{^Is_Jg: 	_HEH}', "");
Expect(0, 1607, '\P{Is_Jg: 	_HEH}', "");
Expect(1, 1607, '\P{^Is_Jg: 	_HEH}', "");
Expect(0, 1608, '\p{Is_Jg: 	_HEH}', "");
Expect(1, 1608, '\p{^Is_Jg: 	_HEH}', "");
Expect(1, 1608, '\P{Is_Jg: 	_HEH}', "");
Expect(0, 1608, '\P{^Is_Jg: 	_HEH}', "");
Error('\p{Joining_Group=-:=heh_Goal}');
Error('\P{Joining_Group=-:=heh_Goal}');
Expect(1, 1730, '\p{Joining_Group=hehgoal}', "");
Expect(0, 1730, '\p{^Joining_Group=hehgoal}', "");
Expect(0, 1730, '\P{Joining_Group=hehgoal}', "");
Expect(1, 1730, '\P{^Joining_Group=hehgoal}', "");
Expect(0, 1731, '\p{Joining_Group=hehgoal}', "");
Expect(1, 1731, '\p{^Joining_Group=hehgoal}', "");
Expect(1, 1731, '\P{Joining_Group=hehgoal}', "");
Expect(0, 1731, '\P{^Joining_Group=hehgoal}', "");
Expect(1, 1730, '\p{Joining_Group=-HEH_GOAL}', "");
Expect(0, 1730, '\p{^Joining_Group=-HEH_GOAL}', "");
Expect(0, 1730, '\P{Joining_Group=-HEH_GOAL}', "");
Expect(1, 1730, '\P{^Joining_Group=-HEH_GOAL}', "");
Expect(0, 1731, '\p{Joining_Group=-HEH_GOAL}', "");
Expect(1, 1731, '\p{^Joining_Group=-HEH_GOAL}', "");
Expect(1, 1731, '\P{Joining_Group=-HEH_GOAL}', "");
Expect(0, 1731, '\P{^Joining_Group=-HEH_GOAL}', "");
Error('\p{Jg=/a/ _Heh_GOAL}');
Error('\P{Jg=/a/ _Heh_GOAL}');
Expect(1, 1730, '\p{Jg: hehgoal}', "");
Expect(0, 1730, '\p{^Jg: hehgoal}', "");
Expect(0, 1730, '\P{Jg: hehgoal}', "");
Expect(1, 1730, '\P{^Jg: hehgoal}', "");
Expect(0, 1731, '\p{Jg: hehgoal}', "");
Expect(1, 1731, '\p{^Jg: hehgoal}', "");
Expect(1, 1731, '\P{Jg: hehgoal}', "");
Expect(0, 1731, '\P{^Jg: hehgoal}', "");
Expect(1, 1730, '\p{Jg= -heh_Goal}', "");
Expect(0, 1730, '\p{^Jg= -heh_Goal}', "");
Expect(0, 1730, '\P{Jg= -heh_Goal}', "");
Expect(1, 1730, '\P{^Jg= -heh_Goal}', "");
Expect(0, 1731, '\p{Jg= -heh_Goal}', "");
Expect(1, 1731, '\p{^Jg= -heh_Goal}', "");
Expect(1, 1731, '\P{Jg= -heh_Goal}', "");
Expect(0, 1731, '\P{^Jg= -heh_Goal}', "");
Error('\p{Is_Joining_Group=_:=HEH_goal}');
Error('\P{Is_Joining_Group=_:=HEH_goal}');
Expect(1, 1730, '\p{Is_Joining_Group=hehgoal}', "");
Expect(0, 1730, '\p{^Is_Joining_Group=hehgoal}', "");
Expect(0, 1730, '\P{Is_Joining_Group=hehgoal}', "");
Expect(1, 1730, '\P{^Is_Joining_Group=hehgoal}', "");
Expect(0, 1731, '\p{Is_Joining_Group=hehgoal}', "");
Expect(1, 1731, '\p{^Is_Joining_Group=hehgoal}', "");
Expect(1, 1731, '\P{Is_Joining_Group=hehgoal}', "");
Expect(0, 1731, '\P{^Is_Joining_Group=hehgoal}', "");
Expect(1, 1730, '\p{Is_Joining_Group:__heh_Goal}', "");
Expect(0, 1730, '\p{^Is_Joining_Group:__heh_Goal}', "");
Expect(0, 1730, '\P{Is_Joining_Group:__heh_Goal}', "");
Expect(1, 1730, '\P{^Is_Joining_Group:__heh_Goal}', "");
Expect(0, 1731, '\p{Is_Joining_Group:__heh_Goal}', "");
Expect(1, 1731, '\p{^Is_Joining_Group:__heh_Goal}', "");
Expect(1, 1731, '\P{Is_Joining_Group:__heh_Goal}', "");
Expect(0, 1731, '\P{^Is_Joining_Group:__heh_Goal}', "");
Error('\p{Is_Jg=_-HEH_goal:=}');
Error('\P{Is_Jg=_-HEH_goal:=}');
Expect(1, 1730, '\p{Is_Jg=hehgoal}', "");
Expect(0, 1730, '\p{^Is_Jg=hehgoal}', "");
Expect(0, 1730, '\P{Is_Jg=hehgoal}', "");
Expect(1, 1730, '\P{^Is_Jg=hehgoal}', "");
Expect(0, 1731, '\p{Is_Jg=hehgoal}', "");
Expect(1, 1731, '\p{^Is_Jg=hehgoal}', "");
Expect(1, 1731, '\P{Is_Jg=hehgoal}', "");
Expect(0, 1731, '\P{^Is_Jg=hehgoal}', "");
Expect(1, 1730, '\p{Is_Jg=		HEH_Goal}', "");
Expect(0, 1730, '\p{^Is_Jg=		HEH_Goal}', "");
Expect(0, 1730, '\P{Is_Jg=		HEH_Goal}', "");
Expect(1, 1730, '\P{^Is_Jg=		HEH_Goal}', "");
Expect(0, 1731, '\p{Is_Jg=		HEH_Goal}', "");
Expect(1, 1731, '\p{^Is_Jg=		HEH_Goal}', "");
Expect(1, 1731, '\P{Is_Jg=		HEH_Goal}', "");
Expect(0, 1731, '\P{^Is_Jg=		HEH_Goal}', "");
Error('\p{Joining_Group=-:=Heth}');
Error('\P{Joining_Group=-:=Heth}');
Expect(1, 1818, '\p{Joining_Group=heth}', "");
Expect(0, 1818, '\p{^Joining_Group=heth}', "");
Expect(0, 1818, '\P{Joining_Group=heth}', "");
Expect(1, 1818, '\P{^Joining_Group=heth}', "");
Expect(0, 1819, '\p{Joining_Group=heth}', "");
Expect(1, 1819, '\p{^Joining_Group=heth}', "");
Expect(1, 1819, '\P{Joining_Group=heth}', "");
Expect(0, 1819, '\P{^Joining_Group=heth}', "");
Expect(1, 1818, '\p{Joining_Group=- heth}', "");
Expect(0, 1818, '\p{^Joining_Group=- heth}', "");
Expect(0, 1818, '\P{Joining_Group=- heth}', "");
Expect(1, 1818, '\P{^Joining_Group=- heth}', "");
Expect(0, 1819, '\p{Joining_Group=- heth}', "");
Expect(1, 1819, '\p{^Joining_Group=- heth}', "");
Expect(1, 1819, '\P{Joining_Group=- heth}', "");
Expect(0, 1819, '\P{^Joining_Group=- heth}', "");
Error('\p{Jg=_:=HETH}');
Error('\P{Jg=_:=HETH}');
Expect(1, 1818, '\p{Jg: heth}', "");
Expect(0, 1818, '\p{^Jg: heth}', "");
Expect(0, 1818, '\P{Jg: heth}', "");
Expect(1, 1818, '\P{^Jg: heth}', "");
Expect(0, 1819, '\p{Jg: heth}', "");
Expect(1, 1819, '\p{^Jg: heth}', "");
Expect(1, 1819, '\P{Jg: heth}', "");
Expect(0, 1819, '\P{^Jg: heth}', "");
Expect(1, 1818, '\p{Jg=_	HETH}', "");
Expect(0, 1818, '\p{^Jg=_	HETH}', "");
Expect(0, 1818, '\P{Jg=_	HETH}', "");
Expect(1, 1818, '\P{^Jg=_	HETH}', "");
Expect(0, 1819, '\p{Jg=_	HETH}', "");
Expect(1, 1819, '\p{^Jg=_	HETH}', "");
Expect(1, 1819, '\P{Jg=_	HETH}', "");
Expect(0, 1819, '\P{^Jg=_	HETH}', "");
Error('\p{Is_Joining_Group=	-Heth/a/}');
Error('\P{Is_Joining_Group=	-Heth/a/}');
Expect(1, 1818, '\p{Is_Joining_Group=heth}', "");
Expect(0, 1818, '\p{^Is_Joining_Group=heth}', "");
Expect(0, 1818, '\P{Is_Joining_Group=heth}', "");
Expect(1, 1818, '\P{^Is_Joining_Group=heth}', "");
Expect(0, 1819, '\p{Is_Joining_Group=heth}', "");
Expect(1, 1819, '\p{^Is_Joining_Group=heth}', "");
Expect(1, 1819, '\P{Is_Joining_Group=heth}', "");
Expect(0, 1819, '\P{^Is_Joining_Group=heth}', "");
Error('\p{Is_Jg:	 _Heth/a/}');
Error('\P{Is_Jg:	 _Heth/a/}');
Expect(1, 1818, '\p{Is_Jg:	heth}', "");
Expect(0, 1818, '\p{^Is_Jg:	heth}', "");
Expect(0, 1818, '\P{Is_Jg:	heth}', "");
Expect(1, 1818, '\P{^Is_Jg:	heth}', "");
Expect(0, 1819, '\p{Is_Jg:	heth}', "");
Expect(1, 1819, '\p{^Is_Jg:	heth}', "");
Expect(1, 1819, '\P{Is_Jg:	heth}', "");
Expect(0, 1819, '\P{^Is_Jg:	heth}', "");
Expect(1, 1818, '\p{Is_Jg=		HETH}', "");
Expect(0, 1818, '\p{^Is_Jg=		HETH}', "");
Expect(0, 1818, '\P{Is_Jg=		HETH}', "");
Expect(1, 1818, '\P{^Is_Jg=		HETH}', "");
Expect(0, 1819, '\p{Is_Jg=		HETH}', "");
Expect(1, 1819, '\p{^Is_Jg=		HETH}', "");
Expect(1, 1819, '\P{Is_Jg=		HETH}', "");
Expect(0, 1819, '\P{^Is_Jg=		HETH}', "");
Error('\p{Joining_Group=_/a/KAF}');
Error('\P{Joining_Group=_/a/KAF}');
Expect(1, 1919, '\p{Joining_Group=kaf}', "");
Expect(0, 1919, '\p{^Joining_Group=kaf}', "");
Expect(0, 1919, '\P{Joining_Group=kaf}', "");
Expect(1, 1919, '\P{^Joining_Group=kaf}', "");
Expect(0, 1920, '\p{Joining_Group=kaf}', "");
Expect(1, 1920, '\p{^Joining_Group=kaf}', "");
Expect(1, 1920, '\P{Joining_Group=kaf}', "");
Expect(0, 1920, '\P{^Joining_Group=kaf}', "");
Expect(1, 1919, '\p{Joining_Group=	KAF}', "");
Expect(0, 1919, '\p{^Joining_Group=	KAF}', "");
Expect(0, 1919, '\P{Joining_Group=	KAF}', "");
Expect(1, 1919, '\P{^Joining_Group=	KAF}', "");
Expect(0, 1920, '\p{Joining_Group=	KAF}', "");
Expect(1, 1920, '\p{^Joining_Group=	KAF}', "");
Expect(1, 1920, '\P{Joining_Group=	KAF}', "");
Expect(0, 1920, '\P{^Joining_Group=	KAF}', "");
Error('\p{Jg=/a/Kaf}');
Error('\P{Jg=/a/Kaf}');
Expect(1, 1919, '\p{Jg=kaf}', "");
Expect(0, 1919, '\p{^Jg=kaf}', "");
Expect(0, 1919, '\P{Jg=kaf}', "");
Expect(1, 1919, '\P{^Jg=kaf}', "");
Expect(0, 1920, '\p{Jg=kaf}', "");
Expect(1, 1920, '\p{^Jg=kaf}', "");
Expect(1, 1920, '\P{Jg=kaf}', "");
Expect(0, 1920, '\P{^Jg=kaf}', "");
Expect(1, 1919, '\p{Jg=_KAF}', "");
Expect(0, 1919, '\p{^Jg=_KAF}', "");
Expect(0, 1919, '\P{Jg=_KAF}', "");
Expect(1, 1919, '\P{^Jg=_KAF}', "");
Expect(0, 1920, '\p{Jg=_KAF}', "");
Expect(1, 1920, '\p{^Jg=_KAF}', "");
Expect(1, 1920, '\P{Jg=_KAF}', "");
Expect(0, 1920, '\P{^Jg=_KAF}', "");
Error('\p{Is_Joining_Group:	:=KAF}');
Error('\P{Is_Joining_Group:	:=KAF}');
Expect(1, 1919, '\p{Is_Joining_Group=kaf}', "");
Expect(0, 1919, '\p{^Is_Joining_Group=kaf}', "");
Expect(0, 1919, '\P{Is_Joining_Group=kaf}', "");
Expect(1, 1919, '\P{^Is_Joining_Group=kaf}', "");
Expect(0, 1920, '\p{Is_Joining_Group=kaf}', "");
Expect(1, 1920, '\p{^Is_Joining_Group=kaf}', "");
Expect(1, 1920, '\P{Is_Joining_Group=kaf}', "");
Expect(0, 1920, '\P{^Is_Joining_Group=kaf}', "");
Expect(1, 1919, '\p{Is_Joining_Group=-Kaf}', "");
Expect(0, 1919, '\p{^Is_Joining_Group=-Kaf}', "");
Expect(0, 1919, '\P{Is_Joining_Group=-Kaf}', "");
Expect(1, 1919, '\P{^Is_Joining_Group=-Kaf}', "");
Expect(0, 1920, '\p{Is_Joining_Group=-Kaf}', "");
Expect(1, 1920, '\p{^Is_Joining_Group=-Kaf}', "");
Expect(1, 1920, '\P{Is_Joining_Group=-Kaf}', "");
Expect(0, 1920, '\P{^Is_Joining_Group=-Kaf}', "");
Error('\p{Is_Jg=:=_kaf}');
Error('\P{Is_Jg=:=_kaf}');
Expect(1, 1919, '\p{Is_Jg=kaf}', "");
Expect(0, 1919, '\p{^Is_Jg=kaf}', "");
Expect(0, 1919, '\P{Is_Jg=kaf}', "");
Expect(1, 1919, '\P{^Is_Jg=kaf}', "");
Expect(0, 1920, '\p{Is_Jg=kaf}', "");
Expect(1, 1920, '\p{^Is_Jg=kaf}', "");
Expect(1, 1920, '\P{Is_Jg=kaf}', "");
Expect(0, 1920, '\P{^Is_Jg=kaf}', "");
Expect(1, 1919, '\p{Is_Jg= 	Kaf}', "");
Expect(0, 1919, '\p{^Is_Jg= 	Kaf}', "");
Expect(0, 1919, '\P{Is_Jg= 	Kaf}', "");
Expect(1, 1919, '\P{^Is_Jg= 	Kaf}', "");
Expect(0, 1920, '\p{Is_Jg= 	Kaf}', "");
Expect(1, 1920, '\p{^Is_Jg= 	Kaf}', "");
Expect(1, 1920, '\P{Is_Jg= 	Kaf}', "");
Expect(0, 1920, '\P{^Is_Jg= 	Kaf}', "");
Error('\p{Joining_Group=/a/KAPH}');
Error('\P{Joining_Group=/a/KAPH}');
Expect(1, 1823, '\p{Joining_Group=kaph}', "");
Expect(0, 1823, '\p{^Joining_Group=kaph}', "");
Expect(0, 1823, '\P{Joining_Group=kaph}', "");
Expect(1, 1823, '\P{^Joining_Group=kaph}', "");
Expect(0, 1824, '\p{Joining_Group=kaph}', "");
Expect(1, 1824, '\p{^Joining_Group=kaph}', "");
Expect(1, 1824, '\P{Joining_Group=kaph}', "");
Expect(0, 1824, '\P{^Joining_Group=kaph}', "");
Expect(1, 1823, '\p{Joining_Group=	KAPH}', "");
Expect(0, 1823, '\p{^Joining_Group=	KAPH}', "");
Expect(0, 1823, '\P{Joining_Group=	KAPH}', "");
Expect(1, 1823, '\P{^Joining_Group=	KAPH}', "");
Expect(0, 1824, '\p{Joining_Group=	KAPH}', "");
Expect(1, 1824, '\p{^Joining_Group=	KAPH}', "");
Expect(1, 1824, '\P{Joining_Group=	KAPH}', "");
Expect(0, 1824, '\P{^Joining_Group=	KAPH}', "");
Error('\p{Jg=__kaph:=}');
Error('\P{Jg=__kaph:=}');
Expect(1, 1823, '\p{Jg=kaph}', "");
Expect(0, 1823, '\p{^Jg=kaph}', "");
Expect(0, 1823, '\P{Jg=kaph}', "");
Expect(1, 1823, '\P{^Jg=kaph}', "");
Expect(0, 1824, '\p{Jg=kaph}', "");
Expect(1, 1824, '\p{^Jg=kaph}', "");
Expect(1, 1824, '\P{Jg=kaph}', "");
Expect(0, 1824, '\P{^Jg=kaph}', "");
Expect(1, 1823, '\p{Jg=_KAPH}', "");
Expect(0, 1823, '\p{^Jg=_KAPH}', "");
Expect(0, 1823, '\P{Jg=_KAPH}', "");
Expect(1, 1823, '\P{^Jg=_KAPH}', "");
Expect(0, 1824, '\p{Jg=_KAPH}', "");
Expect(1, 1824, '\p{^Jg=_KAPH}', "");
Expect(1, 1824, '\P{Jg=_KAPH}', "");
Expect(0, 1824, '\P{^Jg=_KAPH}', "");
Error('\p{Is_Joining_Group=_KAPH:=}');
Error('\P{Is_Joining_Group=_KAPH:=}');
Expect(1, 1823, '\p{Is_Joining_Group=kaph}', "");
Expect(0, 1823, '\p{^Is_Joining_Group=kaph}', "");
Expect(0, 1823, '\P{Is_Joining_Group=kaph}', "");
Expect(1, 1823, '\P{^Is_Joining_Group=kaph}', "");
Expect(0, 1824, '\p{Is_Joining_Group=kaph}', "");
Expect(1, 1824, '\p{^Is_Joining_Group=kaph}', "");
Expect(1, 1824, '\P{Is_Joining_Group=kaph}', "");
Expect(0, 1824, '\P{^Is_Joining_Group=kaph}', "");
Expect(1, 1823, '\p{Is_Joining_Group=	KAPH}', "");
Expect(0, 1823, '\p{^Is_Joining_Group=	KAPH}', "");
Expect(0, 1823, '\P{Is_Joining_Group=	KAPH}', "");
Expect(1, 1823, '\P{^Is_Joining_Group=	KAPH}', "");
Expect(0, 1824, '\p{Is_Joining_Group=	KAPH}', "");
Expect(1, 1824, '\p{^Is_Joining_Group=	KAPH}', "");
Expect(1, 1824, '\P{Is_Joining_Group=	KAPH}', "");
Expect(0, 1824, '\P{^Is_Joining_Group=	KAPH}', "");
Error('\p{Is_Jg=-KAPH/a/}');
Error('\P{Is_Jg=-KAPH/a/}');
Expect(1, 1823, '\p{Is_Jg=kaph}', "");
Expect(0, 1823, '\p{^Is_Jg=kaph}', "");
Expect(0, 1823, '\P{Is_Jg=kaph}', "");
Expect(1, 1823, '\P{^Is_Jg=kaph}', "");
Expect(0, 1824, '\p{Is_Jg=kaph}', "");
Expect(1, 1824, '\p{^Is_Jg=kaph}', "");
Expect(1, 1824, '\P{Is_Jg=kaph}', "");
Expect(0, 1824, '\P{^Is_Jg=kaph}', "");
Expect(1, 1823, '\p{Is_Jg=  kaph}', "");
Expect(0, 1823, '\p{^Is_Jg=  kaph}', "");
Expect(0, 1823, '\P{Is_Jg=  kaph}', "");
Expect(1, 1823, '\P{^Is_Jg=  kaph}', "");
Expect(0, 1824, '\p{Is_Jg=  kaph}', "");
Expect(1, 1824, '\p{^Is_Jg=  kaph}', "");
Expect(1, 1824, '\P{Is_Jg=  kaph}', "");
Expect(0, 1824, '\P{^Is_Jg=  kaph}', "");
Error('\p{Joining_Group=_Khaph:=}');
Error('\P{Joining_Group=_Khaph:=}');
Expect(1, 1870, '\p{Joining_Group=khaph}', "");
Expect(0, 1870, '\p{^Joining_Group=khaph}', "");
Expect(0, 1870, '\P{Joining_Group=khaph}', "");
Expect(1, 1870, '\P{^Joining_Group=khaph}', "");
Expect(0, 1871, '\p{Joining_Group=khaph}', "");
Expect(1, 1871, '\p{^Joining_Group=khaph}', "");
Expect(1, 1871, '\P{Joining_Group=khaph}', "");
Expect(0, 1871, '\P{^Joining_Group=khaph}', "");
Expect(1, 1870, '\p{Joining_Group=	_khaph}', "");
Expect(0, 1870, '\p{^Joining_Group=	_khaph}', "");
Expect(0, 1870, '\P{Joining_Group=	_khaph}', "");
Expect(1, 1870, '\P{^Joining_Group=	_khaph}', "");
Expect(0, 1871, '\p{Joining_Group=	_khaph}', "");
Expect(1, 1871, '\p{^Joining_Group=	_khaph}', "");
Expect(1, 1871, '\P{Joining_Group=	_khaph}', "");
Expect(0, 1871, '\P{^Joining_Group=	_khaph}', "");
Error('\p{Jg=_khaph:=}');
Error('\P{Jg=_khaph:=}');
Expect(1, 1870, '\p{Jg=khaph}', "");
Expect(0, 1870, '\p{^Jg=khaph}', "");
Expect(0, 1870, '\P{Jg=khaph}', "");
Expect(1, 1870, '\P{^Jg=khaph}', "");
Expect(0, 1871, '\p{Jg=khaph}', "");
Expect(1, 1871, '\p{^Jg=khaph}', "");
Expect(1, 1871, '\P{Jg=khaph}', "");
Expect(0, 1871, '\P{^Jg=khaph}', "");
Expect(1, 1870, '\p{Jg= Khaph}', "");
Expect(0, 1870, '\p{^Jg= Khaph}', "");
Expect(0, 1870, '\P{Jg= Khaph}', "");
Expect(1, 1870, '\P{^Jg= Khaph}', "");
Expect(0, 1871, '\p{Jg= Khaph}', "");
Expect(1, 1871, '\p{^Jg= Khaph}', "");
Expect(1, 1871, '\P{Jg= Khaph}', "");
Expect(0, 1871, '\P{^Jg= Khaph}', "");
Error('\p{Is_Joining_Group=:=_khaph}');
Error('\P{Is_Joining_Group=:=_khaph}');
Expect(1, 1870, '\p{Is_Joining_Group=khaph}', "");
Expect(0, 1870, '\p{^Is_Joining_Group=khaph}', "");
Expect(0, 1870, '\P{Is_Joining_Group=khaph}', "");
Expect(1, 1870, '\P{^Is_Joining_Group=khaph}', "");
Expect(0, 1871, '\p{Is_Joining_Group=khaph}', "");
Expect(1, 1871, '\p{^Is_Joining_Group=khaph}', "");
Expect(1, 1871, '\P{Is_Joining_Group=khaph}', "");
Expect(0, 1871, '\P{^Is_Joining_Group=khaph}', "");
Expect(1, 1870, '\p{Is_Joining_Group=-_Khaph}', "");
Expect(0, 1870, '\p{^Is_Joining_Group=-_Khaph}', "");
Expect(0, 1870, '\P{Is_Joining_Group=-_Khaph}', "");
Expect(1, 1870, '\P{^Is_Joining_Group=-_Khaph}', "");
Expect(0, 1871, '\p{Is_Joining_Group=-_Khaph}', "");
Expect(1, 1871, '\p{^Is_Joining_Group=-_Khaph}', "");
Expect(1, 1871, '\P{Is_Joining_Group=-_Khaph}', "");
Expect(0, 1871, '\P{^Is_Joining_Group=-_Khaph}', "");
Error('\p{Is_Jg= Khaph/a/}');
Error('\P{Is_Jg= Khaph/a/}');
Expect(1, 1870, '\p{Is_Jg=khaph}', "");
Expect(0, 1870, '\p{^Is_Jg=khaph}', "");
Expect(0, 1870, '\P{Is_Jg=khaph}', "");
Expect(1, 1870, '\P{^Is_Jg=khaph}', "");
Expect(0, 1871, '\p{Is_Jg=khaph}', "");
Expect(1, 1871, '\p{^Is_Jg=khaph}', "");
Expect(1, 1871, '\P{Is_Jg=khaph}', "");
Expect(0, 1871, '\P{^Is_Jg=khaph}', "");
Expect(1, 1870, '\p{Is_Jg=-	khaph}', "");
Expect(0, 1870, '\p{^Is_Jg=-	khaph}', "");
Expect(0, 1870, '\P{Is_Jg=-	khaph}', "");
Expect(1, 1870, '\P{^Is_Jg=-	khaph}', "");
Expect(0, 1871, '\p{Is_Jg=-	khaph}', "");
Expect(1, 1871, '\p{^Is_Jg=-	khaph}', "");
Expect(1, 1871, '\P{Is_Jg=-	khaph}', "");
Expect(0, 1871, '\P{^Is_Jg=-	khaph}', "");
Error('\p{Joining_Group=-:=Knotted_Heh}');
Error('\P{Joining_Group=-:=Knotted_Heh}');
Expect(1, 1791, '\p{Joining_Group=knottedheh}', "");
Expect(0, 1791, '\p{^Joining_Group=knottedheh}', "");
Expect(0, 1791, '\P{Joining_Group=knottedheh}', "");
Expect(1, 1791, '\P{^Joining_Group=knottedheh}', "");
Expect(0, 1792, '\p{Joining_Group=knottedheh}', "");
Expect(1, 1792, '\p{^Joining_Group=knottedheh}', "");
Expect(1, 1792, '\P{Joining_Group=knottedheh}', "");
Expect(0, 1792, '\P{^Joining_Group=knottedheh}', "");
Expect(1, 1791, '\p{Joining_Group=	 knotted_heh}', "");
Expect(0, 1791, '\p{^Joining_Group=	 knotted_heh}', "");
Expect(0, 1791, '\P{Joining_Group=	 knotted_heh}', "");
Expect(1, 1791, '\P{^Joining_Group=	 knotted_heh}', "");
Expect(0, 1792, '\p{Joining_Group=	 knotted_heh}', "");
Expect(1, 1792, '\p{^Joining_Group=	 knotted_heh}', "");
Expect(1, 1792, '\P{Joining_Group=	 knotted_heh}', "");
Expect(0, 1792, '\P{^Joining_Group=	 knotted_heh}', "");
Error('\p{Jg=	-Knotted_Heh:=}');
Error('\P{Jg=	-Knotted_Heh:=}');
Expect(1, 1791, '\p{Jg:   knottedheh}', "");
Expect(0, 1791, '\p{^Jg:   knottedheh}', "");
Expect(0, 1791, '\P{Jg:   knottedheh}', "");
Expect(1, 1791, '\P{^Jg:   knottedheh}', "");
Expect(0, 1792, '\p{Jg:   knottedheh}', "");
Expect(1, 1792, '\p{^Jg:   knottedheh}', "");
Expect(1, 1792, '\P{Jg:   knottedheh}', "");
Expect(0, 1792, '\P{^Jg:   knottedheh}', "");
Expect(1, 1791, '\p{Jg=		Knotted_Heh}', "");
Expect(0, 1791, '\p{^Jg=		Knotted_Heh}', "");
Expect(0, 1791, '\P{Jg=		Knotted_Heh}', "");
Expect(1, 1791, '\P{^Jg=		Knotted_Heh}', "");
Expect(0, 1792, '\p{Jg=		Knotted_Heh}', "");
Expect(1, 1792, '\p{^Jg=		Knotted_Heh}', "");
Expect(1, 1792, '\P{Jg=		Knotted_Heh}', "");
Expect(0, 1792, '\P{^Jg=		Knotted_Heh}', "");
Error('\p{Is_Joining_Group:  knotted_HEH:=}');
Error('\P{Is_Joining_Group:  knotted_HEH:=}');
Expect(1, 1791, '\p{Is_Joining_Group=knottedheh}', "");
Expect(0, 1791, '\p{^Is_Joining_Group=knottedheh}', "");
Expect(0, 1791, '\P{Is_Joining_Group=knottedheh}', "");
Expect(1, 1791, '\P{^Is_Joining_Group=knottedheh}', "");
Expect(0, 1792, '\p{Is_Joining_Group=knottedheh}', "");
Expect(1, 1792, '\p{^Is_Joining_Group=knottedheh}', "");
Expect(1, 1792, '\P{Is_Joining_Group=knottedheh}', "");
Expect(0, 1792, '\P{^Is_Joining_Group=knottedheh}', "");
Expect(1, 1791, '\p{Is_Joining_Group=_knotted_heh}', "");
Expect(0, 1791, '\p{^Is_Joining_Group=_knotted_heh}', "");
Expect(0, 1791, '\P{Is_Joining_Group=_knotted_heh}', "");
Expect(1, 1791, '\P{^Is_Joining_Group=_knotted_heh}', "");
Expect(0, 1792, '\p{Is_Joining_Group=_knotted_heh}', "");
Expect(1, 1792, '\p{^Is_Joining_Group=_knotted_heh}', "");
Expect(1, 1792, '\P{Is_Joining_Group=_knotted_heh}', "");
Expect(0, 1792, '\P{^Is_Joining_Group=_knotted_heh}', "");
Error('\p{Is_Jg: /a/KNOTTED_HEH}');
Error('\P{Is_Jg: /a/KNOTTED_HEH}');
Expect(1, 1791, '\p{Is_Jg=knottedheh}', "");
Expect(0, 1791, '\p{^Is_Jg=knottedheh}', "");
Expect(0, 1791, '\P{Is_Jg=knottedheh}', "");
Expect(1, 1791, '\P{^Is_Jg=knottedheh}', "");
Expect(0, 1792, '\p{Is_Jg=knottedheh}', "");
Expect(1, 1792, '\p{^Is_Jg=knottedheh}', "");
Expect(1, 1792, '\P{Is_Jg=knottedheh}', "");
Expect(0, 1792, '\P{^Is_Jg=knottedheh}', "");
Expect(1, 1791, '\p{Is_Jg=	-knotted_HEH}', "");
Expect(0, 1791, '\p{^Is_Jg=	-knotted_HEH}', "");
Expect(0, 1791, '\P{Is_Jg=	-knotted_HEH}', "");
Expect(1, 1791, '\P{^Is_Jg=	-knotted_HEH}', "");
Expect(0, 1792, '\p{Is_Jg=	-knotted_HEH}', "");
Expect(1, 1792, '\p{^Is_Jg=	-knotted_HEH}', "");
Expect(1, 1792, '\P{Is_Jg=	-knotted_HEH}', "");
Expect(0, 1792, '\P{^Is_Jg=	-knotted_HEH}', "");
Error('\p{Joining_Group: :=--lam}');
Error('\P{Joining_Group: :=--lam}');
Expect(1, 2214, '\p{Joining_Group:   lam}', "");
Expect(0, 2214, '\p{^Joining_Group:   lam}', "");
Expect(0, 2214, '\P{Joining_Group:   lam}', "");
Expect(1, 2214, '\P{^Joining_Group:   lam}', "");
Expect(0, 2215, '\p{Joining_Group:   lam}', "");
Expect(1, 2215, '\p{^Joining_Group:   lam}', "");
Expect(1, 2215, '\P{Joining_Group:   lam}', "");
Expect(0, 2215, '\P{^Joining_Group:   lam}', "");
Expect(1, 2214, '\p{Joining_Group=-lam}', "");
Expect(0, 2214, '\p{^Joining_Group=-lam}', "");
Expect(0, 2214, '\P{Joining_Group=-lam}', "");
Expect(1, 2214, '\P{^Joining_Group=-lam}', "");
Expect(0, 2215, '\p{Joining_Group=-lam}', "");
Expect(1, 2215, '\p{^Joining_Group=-lam}', "");
Expect(1, 2215, '\P{Joining_Group=-lam}', "");
Expect(0, 2215, '\P{^Joining_Group=-lam}', "");
Error('\p{Jg=	:=Lam}');
Error('\P{Jg=	:=Lam}');
Expect(1, 2214, '\p{Jg=lam}', "");
Expect(0, 2214, '\p{^Jg=lam}', "");
Expect(0, 2214, '\P{Jg=lam}', "");
Expect(1, 2214, '\P{^Jg=lam}', "");
Expect(0, 2215, '\p{Jg=lam}', "");
Expect(1, 2215, '\p{^Jg=lam}', "");
Expect(1, 2215, '\P{Jg=lam}', "");
Expect(0, 2215, '\P{^Jg=lam}', "");
Error('\p{Is_Joining_Group=/a/-Lam}');
Error('\P{Is_Joining_Group=/a/-Lam}');
Expect(1, 2214, '\p{Is_Joining_Group=lam}', "");
Expect(0, 2214, '\p{^Is_Joining_Group=lam}', "");
Expect(0, 2214, '\P{Is_Joining_Group=lam}', "");
Expect(1, 2214, '\P{^Is_Joining_Group=lam}', "");
Expect(0, 2215, '\p{Is_Joining_Group=lam}', "");
Expect(1, 2215, '\p{^Is_Joining_Group=lam}', "");
Expect(1, 2215, '\P{Is_Joining_Group=lam}', "");
Expect(0, 2215, '\P{^Is_Joining_Group=lam}', "");
Error('\p{Is_Jg=/a/	LAM}');
Error('\P{Is_Jg=/a/	LAM}');
Expect(1, 2214, '\p{Is_Jg=lam}', "");
Expect(0, 2214, '\p{^Is_Jg=lam}', "");
Expect(0, 2214, '\P{Is_Jg=lam}', "");
Expect(1, 2214, '\P{^Is_Jg=lam}', "");
Expect(0, 2215, '\p{Is_Jg=lam}', "");
Expect(1, 2215, '\p{^Is_Jg=lam}', "");
Expect(1, 2215, '\P{Is_Jg=lam}', "");
Expect(0, 2215, '\P{^Is_Jg=lam}', "");
Expect(1, 2214, '\p{Is_Jg=__LAM}', "");
Expect(0, 2214, '\p{^Is_Jg=__LAM}', "");
Expect(0, 2214, '\P{Is_Jg=__LAM}', "");
Expect(1, 2214, '\P{^Is_Jg=__LAM}', "");
Expect(0, 2215, '\p{Is_Jg=__LAM}', "");
Expect(1, 2215, '\p{^Is_Jg=__LAM}', "");
Expect(1, 2215, '\P{Is_Jg=__LAM}', "");
Expect(0, 2215, '\P{^Is_Jg=__LAM}', "");
Error('\p{Joining_Group=_Lamadh/a/}');
Error('\P{Joining_Group=_Lamadh/a/}');
Expect(1, 1824, '\p{Joining_Group=lamadh}', "");
Expect(0, 1824, '\p{^Joining_Group=lamadh}', "");
Expect(0, 1824, '\P{Joining_Group=lamadh}', "");
Expect(1, 1824, '\P{^Joining_Group=lamadh}', "");
Expect(0, 1825, '\p{Joining_Group=lamadh}', "");
Expect(1, 1825, '\p{^Joining_Group=lamadh}', "");
Expect(1, 1825, '\P{Joining_Group=lamadh}', "");
Expect(0, 1825, '\P{^Joining_Group=lamadh}', "");
Expect(1, 1824, '\p{Joining_Group= -Lamadh}', "");
Expect(0, 1824, '\p{^Joining_Group= -Lamadh}', "");
Expect(0, 1824, '\P{Joining_Group= -Lamadh}', "");
Expect(1, 1824, '\P{^Joining_Group= -Lamadh}', "");
Expect(0, 1825, '\p{Joining_Group= -Lamadh}', "");
Expect(1, 1825, '\p{^Joining_Group= -Lamadh}', "");
Expect(1, 1825, '\P{Joining_Group= -Lamadh}', "");
Expect(0, 1825, '\P{^Joining_Group= -Lamadh}', "");
Error('\p{Jg=_-Lamadh/a/}');
Error('\P{Jg=_-Lamadh/a/}');
Expect(1, 1824, '\p{Jg=lamadh}', "");
Expect(0, 1824, '\p{^Jg=lamadh}', "");
Expect(0, 1824, '\P{Jg=lamadh}', "");
Expect(1, 1824, '\P{^Jg=lamadh}', "");
Expect(0, 1825, '\p{Jg=lamadh}', "");
Expect(1, 1825, '\p{^Jg=lamadh}', "");
Expect(1, 1825, '\P{Jg=lamadh}', "");
Expect(0, 1825, '\P{^Jg=lamadh}', "");
Expect(1, 1824, '\p{Jg=_-lamadh}', "");
Expect(0, 1824, '\p{^Jg=_-lamadh}', "");
Expect(0, 1824, '\P{Jg=_-lamadh}', "");
Expect(1, 1824, '\P{^Jg=_-lamadh}', "");
Expect(0, 1825, '\p{Jg=_-lamadh}', "");
Expect(1, 1825, '\p{^Jg=_-lamadh}', "");
Expect(1, 1825, '\P{Jg=_-lamadh}', "");
Expect(0, 1825, '\P{^Jg=_-lamadh}', "");
Error('\p{Is_Joining_Group=_:=lamadh}');
Error('\P{Is_Joining_Group=_:=lamadh}');
Expect(1, 1824, '\p{Is_Joining_Group:lamadh}', "");
Expect(0, 1824, '\p{^Is_Joining_Group:lamadh}', "");
Expect(0, 1824, '\P{Is_Joining_Group:lamadh}', "");
Expect(1, 1824, '\P{^Is_Joining_Group:lamadh}', "");
Expect(0, 1825, '\p{Is_Joining_Group:lamadh}', "");
Expect(1, 1825, '\p{^Is_Joining_Group:lamadh}', "");
Expect(1, 1825, '\P{Is_Joining_Group:lamadh}', "");
Expect(0, 1825, '\P{^Is_Joining_Group:lamadh}', "");
Expect(1, 1824, '\p{Is_Joining_Group=-	LAMADH}', "");
Expect(0, 1824, '\p{^Is_Joining_Group=-	LAMADH}', "");
Expect(0, 1824, '\P{Is_Joining_Group=-	LAMADH}', "");
Expect(1, 1824, '\P{^Is_Joining_Group=-	LAMADH}', "");
Expect(0, 1825, '\p{Is_Joining_Group=-	LAMADH}', "");
Expect(1, 1825, '\p{^Is_Joining_Group=-	LAMADH}', "");
Expect(1, 1825, '\P{Is_Joining_Group=-	LAMADH}', "");
Expect(0, 1825, '\P{^Is_Joining_Group=-	LAMADH}', "");
Error('\p{Is_Jg= /a/Lamadh}');
Error('\P{Is_Jg= /a/Lamadh}');
Expect(1, 1824, '\p{Is_Jg=lamadh}', "");
Expect(0, 1824, '\p{^Is_Jg=lamadh}', "");
Expect(0, 1824, '\P{Is_Jg=lamadh}', "");
Expect(1, 1824, '\P{^Is_Jg=lamadh}', "");
Expect(0, 1825, '\p{Is_Jg=lamadh}', "");
Expect(1, 1825, '\p{^Is_Jg=lamadh}', "");
Expect(1, 1825, '\P{Is_Jg=lamadh}', "");
Expect(0, 1825, '\P{^Is_Jg=lamadh}', "");
Expect(1, 1824, '\p{Is_Jg=	-LAMADH}', "");
Expect(0, 1824, '\p{^Is_Jg=	-LAMADH}', "");
Expect(0, 1824, '\P{Is_Jg=	-LAMADH}', "");
Expect(1, 1824, '\P{^Is_Jg=	-LAMADH}', "");
Expect(0, 1825, '\p{Is_Jg=	-LAMADH}', "");
Expect(1, 1825, '\p{^Is_Jg=	-LAMADH}', "");
Expect(1, 1825, '\P{Is_Jg=	-LAMADH}', "");
Expect(0, 1825, '\P{^Is_Jg=	-LAMADH}', "");
Error('\p{Joining_Group=-/a/Meem}');
Error('\P{Joining_Group=-/a/Meem}');
Expect(1, 2215, '\p{Joining_Group=meem}', "");
Expect(0, 2215, '\p{^Joining_Group=meem}', "");
Expect(0, 2215, '\P{Joining_Group=meem}', "");
Expect(1, 2215, '\P{^Joining_Group=meem}', "");
Expect(0, 2216, '\p{Joining_Group=meem}', "");
Expect(1, 2216, '\p{^Joining_Group=meem}', "");
Expect(1, 2216, '\P{Joining_Group=meem}', "");
Expect(0, 2216, '\P{^Joining_Group=meem}', "");
Expect(1, 2215, '\p{Joining_Group:		meem}', "");
Expect(0, 2215, '\p{^Joining_Group:		meem}', "");
Expect(0, 2215, '\P{Joining_Group:		meem}', "");
Expect(1, 2215, '\P{^Joining_Group:		meem}', "");
Expect(0, 2216, '\p{Joining_Group:		meem}', "");
Expect(1, 2216, '\p{^Joining_Group:		meem}', "");
Expect(1, 2216, '\P{Joining_Group:		meem}', "");
Expect(0, 2216, '\P{^Joining_Group:		meem}', "");
Error('\p{Jg= _Meem:=}');
Error('\P{Jg= _Meem:=}');
Expect(1, 2215, '\p{Jg=meem}', "");
Expect(0, 2215, '\p{^Jg=meem}', "");
Expect(0, 2215, '\P{Jg=meem}', "");
Expect(1, 2215, '\P{^Jg=meem}', "");
Expect(0, 2216, '\p{Jg=meem}', "");
Expect(1, 2216, '\p{^Jg=meem}', "");
Expect(1, 2216, '\P{Jg=meem}', "");
Expect(0, 2216, '\P{^Jg=meem}', "");
Expect(1, 2215, '\p{Jg=- Meem}', "");
Expect(0, 2215, '\p{^Jg=- Meem}', "");
Expect(0, 2215, '\P{Jg=- Meem}', "");
Expect(1, 2215, '\P{^Jg=- Meem}', "");
Expect(0, 2216, '\p{Jg=- Meem}', "");
Expect(1, 2216, '\p{^Jg=- Meem}', "");
Expect(1, 2216, '\P{Jg=- Meem}', "");
Expect(0, 2216, '\P{^Jg=- Meem}', "");
Error('\p{Is_Joining_Group= :=Meem}');
Error('\P{Is_Joining_Group= :=Meem}');
Expect(1, 2215, '\p{Is_Joining_Group=meem}', "");
Expect(0, 2215, '\p{^Is_Joining_Group=meem}', "");
Expect(0, 2215, '\P{Is_Joining_Group=meem}', "");
Expect(1, 2215, '\P{^Is_Joining_Group=meem}', "");
Expect(0, 2216, '\p{Is_Joining_Group=meem}', "");
Expect(1, 2216, '\p{^Is_Joining_Group=meem}', "");
Expect(1, 2216, '\P{Is_Joining_Group=meem}', "");
Expect(0, 2216, '\P{^Is_Joining_Group=meem}', "");
Expect(1, 2215, '\p{Is_Joining_Group: 	MEEM}', "");
Expect(0, 2215, '\p{^Is_Joining_Group: 	MEEM}', "");
Expect(0, 2215, '\P{Is_Joining_Group: 	MEEM}', "");
Expect(1, 2215, '\P{^Is_Joining_Group: 	MEEM}', "");
Expect(0, 2216, '\p{Is_Joining_Group: 	MEEM}', "");
Expect(1, 2216, '\p{^Is_Joining_Group: 	MEEM}', "");
Expect(1, 2216, '\P{Is_Joining_Group: 	MEEM}', "");
Expect(0, 2216, '\P{^Is_Joining_Group: 	MEEM}', "");
Error('\p{Is_Jg=-/a/meem}');
Error('\P{Is_Jg=-/a/meem}');
Expect(1, 2215, '\p{Is_Jg=meem}', "");
Expect(0, 2215, '\p{^Is_Jg=meem}', "");
Expect(0, 2215, '\P{Is_Jg=meem}', "");
Expect(1, 2215, '\P{^Is_Jg=meem}', "");
Expect(0, 2216, '\p{Is_Jg=meem}', "");
Expect(1, 2216, '\p{^Is_Jg=meem}', "");
Expect(1, 2216, '\P{Is_Jg=meem}', "");
Expect(0, 2216, '\P{^Is_Jg=meem}', "");
Expect(1, 2215, '\p{Is_Jg=-	meem}', "");
Expect(0, 2215, '\p{^Is_Jg=-	meem}', "");
Expect(0, 2215, '\P{Is_Jg=-	meem}', "");
Expect(1, 2215, '\P{^Is_Jg=-	meem}', "");
Expect(0, 2216, '\p{Is_Jg=-	meem}', "");
Expect(1, 2216, '\p{^Is_Jg=-	meem}', "");
Expect(1, 2216, '\P{Is_Jg=-	meem}', "");
Expect(0, 2216, '\P{^Is_Jg=-	meem}', "");
Error('\p{Joining_Group=:=--mim}');
Error('\P{Joining_Group=:=--mim}');
Expect(1, 1825, '\p{Joining_Group=mim}', "");
Expect(0, 1825, '\p{^Joining_Group=mim}', "");
Expect(0, 1825, '\P{Joining_Group=mim}', "");
Expect(1, 1825, '\P{^Joining_Group=mim}', "");
Expect(0, 1826, '\p{Joining_Group=mim}', "");
Expect(1, 1826, '\p{^Joining_Group=mim}', "");
Expect(1, 1826, '\P{Joining_Group=mim}', "");
Expect(0, 1826, '\P{^Joining_Group=mim}', "");
Expect(1, 1825, '\p{Joining_Group=Mim}', "");
Expect(0, 1825, '\p{^Joining_Group=Mim}', "");
Expect(0, 1825, '\P{Joining_Group=Mim}', "");
Expect(1, 1825, '\P{^Joining_Group=Mim}', "");
Expect(0, 1826, '\p{Joining_Group=Mim}', "");
Expect(1, 1826, '\p{^Joining_Group=Mim}', "");
Expect(1, 1826, '\P{Joining_Group=Mim}', "");
Expect(0, 1826, '\P{^Joining_Group=Mim}', "");
Error('\p{Jg=__Mim:=}');
Error('\P{Jg=__Mim:=}');
Expect(1, 1825, '\p{Jg=mim}', "");
Expect(0, 1825, '\p{^Jg=mim}', "");
Expect(0, 1825, '\P{Jg=mim}', "");
Expect(1, 1825, '\P{^Jg=mim}', "");
Expect(0, 1826, '\p{Jg=mim}', "");
Expect(1, 1826, '\p{^Jg=mim}', "");
Expect(1, 1826, '\P{Jg=mim}', "");
Expect(0, 1826, '\P{^Jg=mim}', "");
Expect(1, 1825, '\p{Jg=-Mim}', "");
Expect(0, 1825, '\p{^Jg=-Mim}', "");
Expect(0, 1825, '\P{Jg=-Mim}', "");
Expect(1, 1825, '\P{^Jg=-Mim}', "");
Expect(0, 1826, '\p{Jg=-Mim}', "");
Expect(1, 1826, '\p{^Jg=-Mim}', "");
Expect(1, 1826, '\P{Jg=-Mim}', "");
Expect(0, 1826, '\P{^Jg=-Mim}', "");
Error('\p{Is_Joining_Group:	/a/mim}');
Error('\P{Is_Joining_Group:	/a/mim}');
Expect(1, 1825, '\p{Is_Joining_Group=mim}', "");
Expect(0, 1825, '\p{^Is_Joining_Group=mim}', "");
Expect(0, 1825, '\P{Is_Joining_Group=mim}', "");
Expect(1, 1825, '\P{^Is_Joining_Group=mim}', "");
Expect(0, 1826, '\p{Is_Joining_Group=mim}', "");
Expect(1, 1826, '\p{^Is_Joining_Group=mim}', "");
Expect(1, 1826, '\P{Is_Joining_Group=mim}', "");
Expect(0, 1826, '\P{^Is_Joining_Group=mim}', "");
Expect(1, 1825, '\p{Is_Joining_Group=_Mim}', "");
Expect(0, 1825, '\p{^Is_Joining_Group=_Mim}', "");
Expect(0, 1825, '\P{Is_Joining_Group=_Mim}', "");
Expect(1, 1825, '\P{^Is_Joining_Group=_Mim}', "");
Expect(0, 1826, '\p{Is_Joining_Group=_Mim}', "");
Expect(1, 1826, '\p{^Is_Joining_Group=_Mim}', "");
Expect(1, 1826, '\P{Is_Joining_Group=_Mim}', "");
Expect(0, 1826, '\P{^Is_Joining_Group=_Mim}', "");
Error('\p{Is_Jg=:=  Mim}');
Error('\P{Is_Jg=:=  Mim}');
Expect(1, 1825, '\p{Is_Jg=mim}', "");
Expect(0, 1825, '\p{^Is_Jg=mim}', "");
Expect(0, 1825, '\P{Is_Jg=mim}', "");
Expect(1, 1825, '\P{^Is_Jg=mim}', "");
Expect(0, 1826, '\p{Is_Jg=mim}', "");
Expect(1, 1826, '\p{^Is_Jg=mim}', "");
Expect(1, 1826, '\P{Is_Jg=mim}', "");
Expect(0, 1826, '\P{^Is_Jg=mim}', "");
Expect(1, 1825, '\p{Is_Jg=-mim}', "");
Expect(0, 1825, '\p{^Is_Jg=-mim}', "");
Expect(0, 1825, '\P{Is_Jg=-mim}', "");
Expect(1, 1825, '\P{^Is_Jg=-mim}', "");
Expect(0, 1826, '\p{Is_Jg=-mim}', "");
Expect(1, 1826, '\p{^Is_Jg=-mim}', "");
Expect(1, 1826, '\P{Is_Jg=-mim}', "");
Expect(0, 1826, '\P{^Is_Jg=-mim}', "");
Error('\p{Joining_Group=-	No_JOINING_group:=}');
Error('\P{Joining_Group=-	No_JOINING_group:=}');
Expect(1, 2221, '\p{Joining_Group=nojoininggroup}', "");
Expect(0, 2221, '\p{^Joining_Group=nojoininggroup}', "");
Expect(0, 2221, '\P{Joining_Group=nojoininggroup}', "");
Expect(1, 2221, '\P{^Joining_Group=nojoininggroup}', "");
Expect(0, 2220, '\p{Joining_Group=nojoininggroup}', "");
Expect(1, 2220, '\p{^Joining_Group=nojoininggroup}', "");
Expect(1, 2220, '\P{Joining_Group=nojoininggroup}', "");
Expect(0, 2220, '\P{^Joining_Group=nojoininggroup}', "");
Expect(1, 2221, '\p{Joining_Group:_no_JOINING_Group}', "");
Expect(0, 2221, '\p{^Joining_Group:_no_JOINING_Group}', "");
Expect(0, 2221, '\P{Joining_Group:_no_JOINING_Group}', "");
Expect(1, 2221, '\P{^Joining_Group:_no_JOINING_Group}', "");
Expect(0, 2220, '\p{Joining_Group:_no_JOINING_Group}', "");
Expect(1, 2220, '\p{^Joining_Group:_no_JOINING_Group}', "");
Expect(1, 2220, '\P{Joining_Group:_no_JOINING_Group}', "");
Expect(0, 2220, '\P{^Joining_Group:_no_JOINING_Group}', "");
Error('\p{Jg=_	NO_Joining_group/a/}');
Error('\P{Jg=_	NO_Joining_group/a/}');
Expect(1, 2221, '\p{Jg=nojoininggroup}', "");
Expect(0, 2221, '\p{^Jg=nojoininggroup}', "");
Expect(0, 2221, '\P{Jg=nojoininggroup}', "");
Expect(1, 2221, '\P{^Jg=nojoininggroup}', "");
Expect(0, 2220, '\p{Jg=nojoininggroup}', "");
Expect(1, 2220, '\p{^Jg=nojoininggroup}', "");
Expect(1, 2220, '\P{Jg=nojoininggroup}', "");
Expect(0, 2220, '\P{^Jg=nojoininggroup}', "");
Expect(1, 2221, '\p{Jg=_ No_Joining_GROUP}', "");
Expect(0, 2221, '\p{^Jg=_ No_Joining_GROUP}', "");
Expect(0, 2221, '\P{Jg=_ No_Joining_GROUP}', "");
Expect(1, 2221, '\P{^Jg=_ No_Joining_GROUP}', "");
Expect(0, 2220, '\p{Jg=_ No_Joining_GROUP}', "");
Expect(1, 2220, '\p{^Jg=_ No_Joining_GROUP}', "");
Expect(1, 2220, '\P{Jg=_ No_Joining_GROUP}', "");
Expect(0, 2220, '\P{^Jg=_ No_Joining_GROUP}', "");
Error('\p{Is_Joining_Group=:=-_No_Joining_Group}');
Error('\P{Is_Joining_Group=:=-_No_Joining_Group}');
Expect(1, 2221, '\p{Is_Joining_Group=nojoininggroup}', "");
Expect(0, 2221, '\p{^Is_Joining_Group=nojoininggroup}', "");
Expect(0, 2221, '\P{Is_Joining_Group=nojoininggroup}', "");
Expect(1, 2221, '\P{^Is_Joining_Group=nojoininggroup}', "");
Expect(0, 2220, '\p{Is_Joining_Group=nojoininggroup}', "");
Expect(1, 2220, '\p{^Is_Joining_Group=nojoininggroup}', "");
Expect(1, 2220, '\P{Is_Joining_Group=nojoininggroup}', "");
Expect(0, 2220, '\P{^Is_Joining_Group=nojoininggroup}', "");
Expect(1, 2221, '\p{Is_Joining_Group=No_Joining_group}', "");
Expect(0, 2221, '\p{^Is_Joining_Group=No_Joining_group}', "");
Expect(0, 2221, '\P{Is_Joining_Group=No_Joining_group}', "");
Expect(1, 2221, '\P{^Is_Joining_Group=No_Joining_group}', "");
Expect(0, 2220, '\p{Is_Joining_Group=No_Joining_group}', "");
Expect(1, 2220, '\p{^Is_Joining_Group=No_Joining_group}', "");
Expect(1, 2220, '\P{Is_Joining_Group=No_Joining_group}', "");
Expect(0, 2220, '\P{^Is_Joining_Group=No_Joining_group}', "");
Error('\p{Is_Jg=/a/	no_joining_Group}');
Error('\P{Is_Jg=/a/	no_joining_Group}');
Expect(1, 2221, '\p{Is_Jg: nojoininggroup}', "");
Expect(0, 2221, '\p{^Is_Jg: nojoininggroup}', "");
Expect(0, 2221, '\P{Is_Jg: nojoininggroup}', "");
Expect(1, 2221, '\P{^Is_Jg: nojoininggroup}', "");
Expect(0, 2220, '\p{Is_Jg: nojoininggroup}', "");
Expect(1, 2220, '\p{^Is_Jg: nojoininggroup}', "");
Expect(1, 2220, '\P{Is_Jg: nojoininggroup}', "");
Expect(0, 2220, '\P{^Is_Jg: nojoininggroup}', "");
Expect(1, 2221, '\p{Is_Jg=	_No_joining_Group}', "");
Expect(0, 2221, '\p{^Is_Jg=	_No_joining_Group}', "");
Expect(0, 2221, '\P{Is_Jg=	_No_joining_Group}', "");
Expect(1, 2221, '\P{^Is_Jg=	_No_joining_Group}', "");
Expect(0, 2220, '\p{Is_Jg=	_No_joining_Group}', "");
Expect(1, 2220, '\p{^Is_Jg=	_No_joining_Group}', "");
Expect(1, 2220, '\P{Is_Jg=	_No_joining_Group}', "");
Expect(0, 2220, '\P{^Is_Jg=	_No_joining_Group}', "");
Error('\p{Joining_Group= :=NOON}');
Error('\P{Joining_Group= :=NOON}');
Expect(1, 1897, '\p{Joining_Group=noon}', "");
Expect(0, 1897, '\p{^Joining_Group=noon}', "");
Expect(0, 1897, '\P{Joining_Group=noon}', "");
Expect(1, 1897, '\P{^Joining_Group=noon}', "");
Expect(0, 1898, '\p{Joining_Group=noon}', "");
Expect(1, 1898, '\p{^Joining_Group=noon}', "");
Expect(1, 1898, '\P{Joining_Group=noon}', "");
Expect(0, 1898, '\P{^Joining_Group=noon}', "");
Expect(1, 1897, '\p{Joining_Group=__Noon}', "");
Expect(0, 1897, '\p{^Joining_Group=__Noon}', "");
Expect(0, 1897, '\P{Joining_Group=__Noon}', "");
Expect(1, 1897, '\P{^Joining_Group=__Noon}', "");
Expect(0, 1898, '\p{Joining_Group=__Noon}', "");
Expect(1, 1898, '\p{^Joining_Group=__Noon}', "");
Expect(1, 1898, '\P{Joining_Group=__Noon}', "");
Expect(0, 1898, '\P{^Joining_Group=__Noon}', "");
Error('\p{Jg:   -/a/Noon}');
Error('\P{Jg:   -/a/Noon}');
Expect(1, 1897, '\p{Jg=noon}', "");
Expect(0, 1897, '\p{^Jg=noon}', "");
Expect(0, 1897, '\P{Jg=noon}', "");
Expect(1, 1897, '\P{^Jg=noon}', "");
Expect(0, 1898, '\p{Jg=noon}', "");
Expect(1, 1898, '\p{^Jg=noon}', "");
Expect(1, 1898, '\P{Jg=noon}', "");
Expect(0, 1898, '\P{^Jg=noon}', "");
Expect(1, 1897, '\p{Jg=-	NOON}', "");
Expect(0, 1897, '\p{^Jg=-	NOON}', "");
Expect(0, 1897, '\P{Jg=-	NOON}', "");
Expect(1, 1897, '\P{^Jg=-	NOON}', "");
Expect(0, 1898, '\p{Jg=-	NOON}', "");
Expect(1, 1898, '\p{^Jg=-	NOON}', "");
Expect(1, 1898, '\P{Jg=-	NOON}', "");
Expect(0, 1898, '\P{^Jg=-	NOON}', "");
Error('\p{Is_Joining_Group=/a/- Noon}');
Error('\P{Is_Joining_Group=/a/- Noon}');
Expect(1, 1897, '\p{Is_Joining_Group=noon}', "");
Expect(0, 1897, '\p{^Is_Joining_Group=noon}', "");
Expect(0, 1897, '\P{Is_Joining_Group=noon}', "");
Expect(1, 1897, '\P{^Is_Joining_Group=noon}', "");
Expect(0, 1898, '\p{Is_Joining_Group=noon}', "");
Expect(1, 1898, '\p{^Is_Joining_Group=noon}', "");
Expect(1, 1898, '\P{Is_Joining_Group=noon}', "");
Expect(0, 1898, '\P{^Is_Joining_Group=noon}', "");
Expect(1, 1897, '\p{Is_Joining_Group=	_Noon}', "");
Expect(0, 1897, '\p{^Is_Joining_Group=	_Noon}', "");
Expect(0, 1897, '\P{Is_Joining_Group=	_Noon}', "");
Expect(1, 1897, '\P{^Is_Joining_Group=	_Noon}', "");
Expect(0, 1898, '\p{Is_Joining_Group=	_Noon}', "");
Expect(1, 1898, '\p{^Is_Joining_Group=	_Noon}', "");
Expect(1, 1898, '\P{Is_Joining_Group=	_Noon}', "");
Expect(0, 1898, '\P{^Is_Joining_Group=	_Noon}', "");
Error('\p{Is_Jg= -Noon:=}');
Error('\P{Is_Jg= -Noon:=}');
Expect(1, 1897, '\p{Is_Jg=noon}', "");
Expect(0, 1897, '\p{^Is_Jg=noon}', "");
Expect(0, 1897, '\P{Is_Jg=noon}', "");
Expect(1, 1897, '\P{^Is_Jg=noon}', "");
Expect(0, 1898, '\p{Is_Jg=noon}', "");
Expect(1, 1898, '\p{^Is_Jg=noon}', "");
Expect(1, 1898, '\P{Is_Jg=noon}', "");
Expect(0, 1898, '\P{^Is_Jg=noon}', "");
Expect(1, 1897, '\p{Is_Jg=- noon}', "");
Expect(0, 1897, '\p{^Is_Jg=- noon}', "");
Expect(0, 1897, '\P{Is_Jg=- noon}', "");
Expect(1, 1897, '\P{^Is_Jg=- noon}', "");
Expect(0, 1898, '\p{Is_Jg=- noon}', "");
Expect(1, 1898, '\p{^Is_Jg=- noon}', "");
Expect(1, 1898, '\P{Is_Jg=- noon}', "");
Expect(0, 1898, '\P{^Is_Jg=- noon}', "");
Error('\p{Joining_Group=__NUN:=}');
Error('\P{Joining_Group=__NUN:=}');
Expect(1, 1826, '\p{Joining_Group=nun}', "");
Expect(0, 1826, '\p{^Joining_Group=nun}', "");
Expect(0, 1826, '\P{Joining_Group=nun}', "");
Expect(1, 1826, '\P{^Joining_Group=nun}', "");
Expect(0, 1827, '\p{Joining_Group=nun}', "");
Expect(1, 1827, '\p{^Joining_Group=nun}', "");
Expect(1, 1827, '\P{Joining_Group=nun}', "");
Expect(0, 1827, '\P{^Joining_Group=nun}', "");
Expect(1, 1826, '\p{Joining_Group=_ NUN}', "");
Expect(0, 1826, '\p{^Joining_Group=_ NUN}', "");
Expect(0, 1826, '\P{Joining_Group=_ NUN}', "");
Expect(1, 1826, '\P{^Joining_Group=_ NUN}', "");
Expect(0, 1827, '\p{Joining_Group=_ NUN}', "");
Expect(1, 1827, '\p{^Joining_Group=_ NUN}', "");
Expect(1, 1827, '\P{Joining_Group=_ NUN}', "");
Expect(0, 1827, '\P{^Joining_Group=_ NUN}', "");
Error('\p{Jg= /a/nun}');
Error('\P{Jg= /a/nun}');
Expect(1, 1826, '\p{Jg=nun}', "");
Expect(0, 1826, '\p{^Jg=nun}', "");
Expect(0, 1826, '\P{Jg=nun}', "");
Expect(1, 1826, '\P{^Jg=nun}', "");
Expect(0, 1827, '\p{Jg=nun}', "");
Expect(1, 1827, '\p{^Jg=nun}', "");
Expect(1, 1827, '\P{Jg=nun}', "");
Expect(0, 1827, '\P{^Jg=nun}', "");
Expect(1, 1826, '\p{Jg=	nun}', "");
Expect(0, 1826, '\p{^Jg=	nun}', "");
Expect(0, 1826, '\P{Jg=	nun}', "");
Expect(1, 1826, '\P{^Jg=	nun}', "");
Expect(0, 1827, '\p{Jg=	nun}', "");
Expect(1, 1827, '\p{^Jg=	nun}', "");
Expect(1, 1827, '\P{Jg=	nun}', "");
Expect(0, 1827, '\P{^Jg=	nun}', "");
Error('\p{Is_Joining_Group=Nun/a/}');
Error('\P{Is_Joining_Group=Nun/a/}');
Expect(1, 1826, '\p{Is_Joining_Group=nun}', "");
Expect(0, 1826, '\p{^Is_Joining_Group=nun}', "");
Expect(0, 1826, '\P{Is_Joining_Group=nun}', "");
Expect(1, 1826, '\P{^Is_Joining_Group=nun}', "");
Expect(0, 1827, '\p{Is_Joining_Group=nun}', "");
Expect(1, 1827, '\p{^Is_Joining_Group=nun}', "");
Expect(1, 1827, '\P{Is_Joining_Group=nun}', "");
Expect(0, 1827, '\P{^Is_Joining_Group=nun}', "");
Expect(1, 1826, '\p{Is_Joining_Group=_	Nun}', "");
Expect(0, 1826, '\p{^Is_Joining_Group=_	Nun}', "");
Expect(0, 1826, '\P{Is_Joining_Group=_	Nun}', "");
Expect(1, 1826, '\P{^Is_Joining_Group=_	Nun}', "");
Expect(0, 1827, '\p{Is_Joining_Group=_	Nun}', "");
Expect(1, 1827, '\p{^Is_Joining_Group=_	Nun}', "");
Expect(1, 1827, '\P{Is_Joining_Group=_	Nun}', "");
Expect(0, 1827, '\P{^Is_Joining_Group=_	Nun}', "");
Error('\p{Is_Jg=/a/Nun}');
Error('\P{Is_Jg=/a/Nun}');
Expect(1, 1826, '\p{Is_Jg=nun}', "");
Expect(0, 1826, '\p{^Is_Jg=nun}', "");
Expect(0, 1826, '\P{Is_Jg=nun}', "");
Expect(1, 1826, '\P{^Is_Jg=nun}', "");
Expect(0, 1827, '\p{Is_Jg=nun}', "");
Expect(1, 1827, '\p{^Is_Jg=nun}', "");
Expect(1, 1827, '\P{Is_Jg=nun}', "");
Expect(0, 1827, '\P{^Is_Jg=nun}', "");
Expect(1, 1826, '\p{Is_Jg: 	Nun}', "");
Expect(0, 1826, '\p{^Is_Jg: 	Nun}', "");
Expect(0, 1826, '\P{Is_Jg: 	Nun}', "");
Expect(1, 1826, '\P{^Is_Jg: 	Nun}', "");
Expect(0, 1827, '\p{Is_Jg: 	Nun}', "");
Expect(1, 1827, '\p{^Is_Jg: 	Nun}', "");
Expect(1, 1827, '\P{Is_Jg: 	Nun}', "");
Expect(0, 1827, '\P{^Is_Jg: 	Nun}', "");
Error('\p{Joining_Group=:=_Nya}');
Error('\P{Joining_Group=:=_Nya}');
Expect(1, 1725, '\p{Joining_Group=nya}', "");
Expect(0, 1725, '\p{^Joining_Group=nya}', "");
Expect(0, 1725, '\P{Joining_Group=nya}', "");
Expect(1, 1725, '\P{^Joining_Group=nya}', "");
Expect(0, 1726, '\p{Joining_Group=nya}', "");
Expect(1, 1726, '\p{^Joining_Group=nya}', "");
Expect(1, 1726, '\P{Joining_Group=nya}', "");
Expect(0, 1726, '\P{^Joining_Group=nya}', "");
Expect(1, 1725, '\p{Joining_Group=-NYA}', "");
Expect(0, 1725, '\p{^Joining_Group=-NYA}', "");
Expect(0, 1725, '\P{Joining_Group=-NYA}', "");
Expect(1, 1725, '\P{^Joining_Group=-NYA}', "");
Expect(0, 1726, '\p{Joining_Group=-NYA}', "");
Expect(1, 1726, '\p{^Joining_Group=-NYA}', "");
Expect(1, 1726, '\P{Joining_Group=-NYA}', "");
Expect(0, 1726, '\P{^Joining_Group=-NYA}', "");
Error('\p{Jg:   -nya/a/}');
Error('\P{Jg:   -nya/a/}');
Expect(1, 1725, '\p{Jg=nya}', "");
Expect(0, 1725, '\p{^Jg=nya}', "");
Expect(0, 1725, '\P{Jg=nya}', "");
Expect(1, 1725, '\P{^Jg=nya}', "");
Expect(0, 1726, '\p{Jg=nya}', "");
Expect(1, 1726, '\p{^Jg=nya}', "");
Expect(1, 1726, '\P{Jg=nya}', "");
Expect(0, 1726, '\P{^Jg=nya}', "");
Expect(1, 1725, '\p{Jg=	 Nya}', "");
Expect(0, 1725, '\p{^Jg=	 Nya}', "");
Expect(0, 1725, '\P{Jg=	 Nya}', "");
Expect(1, 1725, '\P{^Jg=	 Nya}', "");
Expect(0, 1726, '\p{Jg=	 Nya}', "");
Expect(1, 1726, '\p{^Jg=	 Nya}', "");
Expect(1, 1726, '\P{Jg=	 Nya}', "");
Expect(0, 1726, '\P{^Jg=	 Nya}', "");
Error('\p{Is_Joining_Group=:= Nya}');
Error('\P{Is_Joining_Group=:= Nya}');
Expect(1, 1725, '\p{Is_Joining_Group=nya}', "");
Expect(0, 1725, '\p{^Is_Joining_Group=nya}', "");
Expect(0, 1725, '\P{Is_Joining_Group=nya}', "");
Expect(1, 1725, '\P{^Is_Joining_Group=nya}', "");
Expect(0, 1726, '\p{Is_Joining_Group=nya}', "");
Expect(1, 1726, '\p{^Is_Joining_Group=nya}', "");
Expect(1, 1726, '\P{Is_Joining_Group=nya}', "");
Expect(0, 1726, '\P{^Is_Joining_Group=nya}', "");
Expect(1, 1725, '\p{Is_Joining_Group=	Nya}', "");
Expect(0, 1725, '\p{^Is_Joining_Group=	Nya}', "");
Expect(0, 1725, '\P{Is_Joining_Group=	Nya}', "");
Expect(1, 1725, '\P{^Is_Joining_Group=	Nya}', "");
Expect(0, 1726, '\p{Is_Joining_Group=	Nya}', "");
Expect(1, 1726, '\p{^Is_Joining_Group=	Nya}', "");
Expect(1, 1726, '\P{Is_Joining_Group=	Nya}', "");
Expect(0, 1726, '\P{^Is_Joining_Group=	Nya}', "");
Error('\p{Is_Jg:	:=_Nya}');
Error('\P{Is_Jg:	:=_Nya}');
Expect(1, 1725, '\p{Is_Jg=nya}', "");
Expect(0, 1725, '\p{^Is_Jg=nya}', "");
Expect(0, 1725, '\P{Is_Jg=nya}', "");
Expect(1, 1725, '\P{^Is_Jg=nya}', "");
Expect(0, 1726, '\p{Is_Jg=nya}', "");
Expect(1, 1726, '\p{^Is_Jg=nya}', "");
Expect(1, 1726, '\P{Is_Jg=nya}', "");
Expect(0, 1726, '\P{^Is_Jg=nya}', "");
Expect(1, 1725, '\p{Is_Jg=-nya}', "");
Expect(0, 1725, '\p{^Is_Jg=-nya}', "");
Expect(0, 1725, '\P{Is_Jg=-nya}', "");
Expect(1, 1725, '\P{^Is_Jg=-nya}', "");
Expect(0, 1726, '\p{Is_Jg=-nya}', "");
Expect(1, 1726, '\p{^Is_Jg=-nya}', "");
Expect(1, 1726, '\P{Is_Jg=-nya}', "");
Expect(0, 1726, '\P{^Is_Jg=-nya}', "");
Error('\p{Joining_Group=_:=PE}');
Error('\P{Joining_Group=_:=PE}');
Expect(1, 1830, '\p{Joining_Group=pe}', "");
Expect(0, 1830, '\p{^Joining_Group=pe}', "");
Expect(0, 1830, '\P{Joining_Group=pe}', "");
Expect(1, 1830, '\P{^Joining_Group=pe}', "");
Expect(0, 1831, '\p{Joining_Group=pe}', "");
Expect(1, 1831, '\p{^Joining_Group=pe}', "");
Expect(1, 1831, '\P{Joining_Group=pe}', "");
Expect(0, 1831, '\P{^Joining_Group=pe}', "");
Expect(1, 1830, '\p{Joining_Group=-	PE}', "");
Expect(0, 1830, '\p{^Joining_Group=-	PE}', "");
Expect(0, 1830, '\P{Joining_Group=-	PE}', "");
Expect(1, 1830, '\P{^Joining_Group=-	PE}', "");
Expect(0, 1831, '\p{Joining_Group=-	PE}', "");
Expect(1, 1831, '\p{^Joining_Group=-	PE}', "");
Expect(1, 1831, '\P{Joining_Group=-	PE}', "");
Expect(0, 1831, '\P{^Joining_Group=-	PE}', "");
Error('\p{Jg=_	Pe:=}');
Error('\P{Jg=_	Pe:=}');
Expect(1, 1830, '\p{Jg=pe}', "");
Expect(0, 1830, '\p{^Jg=pe}', "");
Expect(0, 1830, '\P{Jg=pe}', "");
Expect(1, 1830, '\P{^Jg=pe}', "");
Expect(0, 1831, '\p{Jg=pe}', "");
Expect(1, 1831, '\p{^Jg=pe}', "");
Expect(1, 1831, '\P{Jg=pe}', "");
Expect(0, 1831, '\P{^Jg=pe}', "");
Expect(1, 1830, '\p{Jg=	 Pe}', "");
Expect(0, 1830, '\p{^Jg=	 Pe}', "");
Expect(0, 1830, '\P{Jg=	 Pe}', "");
Expect(1, 1830, '\P{^Jg=	 Pe}', "");
Expect(0, 1831, '\p{Jg=	 Pe}', "");
Expect(1, 1831, '\p{^Jg=	 Pe}', "");
Expect(1, 1831, '\P{Jg=	 Pe}', "");
Expect(0, 1831, '\P{^Jg=	 Pe}', "");
Error('\p{Is_Joining_Group=/a/-Pe}');
Error('\P{Is_Joining_Group=/a/-Pe}');
Expect(1, 1830, '\p{Is_Joining_Group=pe}', "");
Expect(0, 1830, '\p{^Is_Joining_Group=pe}', "");
Expect(0, 1830, '\P{Is_Joining_Group=pe}', "");
Expect(1, 1830, '\P{^Is_Joining_Group=pe}', "");
Expect(0, 1831, '\p{Is_Joining_Group=pe}', "");
Expect(1, 1831, '\p{^Is_Joining_Group=pe}', "");
Expect(1, 1831, '\P{Is_Joining_Group=pe}', "");
Expect(0, 1831, '\P{^Is_Joining_Group=pe}', "");
Expect(1, 1830, '\p{Is_Joining_Group=-_Pe}', "");
Expect(0, 1830, '\p{^Is_Joining_Group=-_Pe}', "");
Expect(0, 1830, '\P{Is_Joining_Group=-_Pe}', "");
Expect(1, 1830, '\P{^Is_Joining_Group=-_Pe}', "");
Expect(0, 1831, '\p{Is_Joining_Group=-_Pe}', "");
Expect(1, 1831, '\p{^Is_Joining_Group=-_Pe}', "");
Expect(1, 1831, '\P{Is_Joining_Group=-_Pe}', "");
Expect(0, 1831, '\P{^Is_Joining_Group=-_Pe}', "");
Error('\p{Is_Jg=:=- PE}');
Error('\P{Is_Jg=:=- PE}');
Expect(1, 1830, '\p{Is_Jg:pe}', "");
Expect(0, 1830, '\p{^Is_Jg:pe}', "");
Expect(0, 1830, '\P{Is_Jg:pe}', "");
Expect(1, 1830, '\P{^Is_Jg:pe}', "");
Expect(0, 1831, '\p{Is_Jg:pe}', "");
Expect(1, 1831, '\p{^Is_Jg:pe}', "");
Expect(1, 1831, '\P{Is_Jg:pe}', "");
Expect(0, 1831, '\P{^Is_Jg:pe}', "");
Expect(1, 1830, '\p{Is_Jg:_pe}', "");
Expect(0, 1830, '\p{^Is_Jg:_pe}', "");
Expect(0, 1830, '\P{Is_Jg:_pe}', "");
Expect(1, 1830, '\P{^Is_Jg:_pe}', "");
Expect(0, 1831, '\p{Is_Jg:_pe}', "");
Expect(1, 1831, '\p{^Is_Jg:_pe}', "");
Expect(1, 1831, '\P{Is_Jg:_pe}', "");
Expect(0, 1831, '\P{^Is_Jg:_pe}', "");
Error('\p{Joining_Group=:= _Qaf}');
Error('\P{Joining_Group=:= _Qaf}');
Expect(1, 2213, '\p{Joining_Group=qaf}', "");
Expect(0, 2213, '\p{^Joining_Group=qaf}', "");
Expect(0, 2213, '\P{Joining_Group=qaf}', "");
Expect(1, 2213, '\P{^Joining_Group=qaf}', "");
Expect(0, 2214, '\p{Joining_Group=qaf}', "");
Expect(1, 2214, '\p{^Joining_Group=qaf}', "");
Expect(1, 2214, '\P{Joining_Group=qaf}', "");
Expect(0, 2214, '\P{^Joining_Group=qaf}', "");
Expect(1, 2213, '\p{Joining_Group=- QAF}', "");
Expect(0, 2213, '\p{^Joining_Group=- QAF}', "");
Expect(0, 2213, '\P{Joining_Group=- QAF}', "");
Expect(1, 2213, '\P{^Joining_Group=- QAF}', "");
Expect(0, 2214, '\p{Joining_Group=- QAF}', "");
Expect(1, 2214, '\p{^Joining_Group=- QAF}', "");
Expect(1, 2214, '\P{Joining_Group=- QAF}', "");
Expect(0, 2214, '\P{^Joining_Group=- QAF}', "");
Error('\p{Jg=-_qaf:=}');
Error('\P{Jg=-_qaf:=}');
Expect(1, 2213, '\p{Jg=qaf}', "");
Expect(0, 2213, '\p{^Jg=qaf}', "");
Expect(0, 2213, '\P{Jg=qaf}', "");
Expect(1, 2213, '\P{^Jg=qaf}', "");
Expect(0, 2214, '\p{Jg=qaf}', "");
Expect(1, 2214, '\p{^Jg=qaf}', "");
Expect(1, 2214, '\P{Jg=qaf}', "");
Expect(0, 2214, '\P{^Jg=qaf}', "");
Expect(1, 2213, '\p{Jg=- Qaf}', "");
Expect(0, 2213, '\p{^Jg=- Qaf}', "");
Expect(0, 2213, '\P{Jg=- Qaf}', "");
Expect(1, 2213, '\P{^Jg=- Qaf}', "");
Expect(0, 2214, '\p{Jg=- Qaf}', "");
Expect(1, 2214, '\p{^Jg=- Qaf}', "");
Expect(1, 2214, '\P{Jg=- Qaf}', "");
Expect(0, 2214, '\P{^Jg=- Qaf}', "");
Error('\p{Is_Joining_Group= /a/QAF}');
Error('\P{Is_Joining_Group= /a/QAF}');
Expect(1, 2213, '\p{Is_Joining_Group=qaf}', "");
Expect(0, 2213, '\p{^Is_Joining_Group=qaf}', "");
Expect(0, 2213, '\P{Is_Joining_Group=qaf}', "");
Expect(1, 2213, '\P{^Is_Joining_Group=qaf}', "");
Expect(0, 2214, '\p{Is_Joining_Group=qaf}', "");
Expect(1, 2214, '\p{^Is_Joining_Group=qaf}', "");
Expect(1, 2214, '\P{Is_Joining_Group=qaf}', "");
Expect(0, 2214, '\P{^Is_Joining_Group=qaf}', "");
Expect(1, 2213, '\p{Is_Joining_Group=--QAF}', "");
Expect(0, 2213, '\p{^Is_Joining_Group=--QAF}', "");
Expect(0, 2213, '\P{Is_Joining_Group=--QAF}', "");
Expect(1, 2213, '\P{^Is_Joining_Group=--QAF}', "");
Expect(0, 2214, '\p{Is_Joining_Group=--QAF}', "");
Expect(1, 2214, '\p{^Is_Joining_Group=--QAF}', "");
Expect(1, 2214, '\P{Is_Joining_Group=--QAF}', "");
Expect(0, 2214, '\P{^Is_Joining_Group=--QAF}', "");
Error('\p{Is_Jg=_-Qaf:=}');
Error('\P{Is_Jg=_-Qaf:=}');
Expect(1, 2213, '\p{Is_Jg=qaf}', "");
Expect(0, 2213, '\p{^Is_Jg=qaf}', "");
Expect(0, 2213, '\P{Is_Jg=qaf}', "");
Expect(1, 2213, '\P{^Is_Jg=qaf}', "");
Expect(0, 2214, '\p{Is_Jg=qaf}', "");
Expect(1, 2214, '\p{^Is_Jg=qaf}', "");
Expect(1, 2214, '\P{Is_Jg=qaf}', "");
Expect(0, 2214, '\P{^Is_Jg=qaf}', "");
Expect(1, 2213, '\p{Is_Jg=--Qaf}', "");
Expect(0, 2213, '\p{^Is_Jg=--Qaf}', "");
Expect(0, 2213, '\P{Is_Jg=--Qaf}', "");
Expect(1, 2213, '\P{^Is_Jg=--Qaf}', "");
Expect(0, 2214, '\p{Is_Jg=--Qaf}', "");
Expect(1, 2214, '\p{^Is_Jg=--Qaf}', "");
Expect(1, 2214, '\P{Is_Jg=--Qaf}', "");
Expect(0, 2214, '\P{^Is_Jg=--Qaf}', "");
Error('\p{Joining_Group:   /a/--qaph}');
Error('\P{Joining_Group:   /a/--qaph}');
Expect(1, 1833, '\p{Joining_Group=qaph}', "");
Expect(0, 1833, '\p{^Joining_Group=qaph}', "");
Expect(0, 1833, '\P{Joining_Group=qaph}', "");
Expect(1, 1833, '\P{^Joining_Group=qaph}', "");
Expect(0, 1834, '\p{Joining_Group=qaph}', "");
Expect(1, 1834, '\p{^Joining_Group=qaph}', "");
Expect(1, 1834, '\P{Joining_Group=qaph}', "");
Expect(0, 1834, '\P{^Joining_Group=qaph}', "");
Expect(1, 1833, '\p{Joining_Group= QAPH}', "");
Expect(0, 1833, '\p{^Joining_Group= QAPH}', "");
Expect(0, 1833, '\P{Joining_Group= QAPH}', "");
Expect(1, 1833, '\P{^Joining_Group= QAPH}', "");
Expect(0, 1834, '\p{Joining_Group= QAPH}', "");
Expect(1, 1834, '\p{^Joining_Group= QAPH}', "");
Expect(1, 1834, '\P{Joining_Group= QAPH}', "");
Expect(0, 1834, '\P{^Joining_Group= QAPH}', "");
Error('\p{Jg=/a/QAPH}');
Error('\P{Jg=/a/QAPH}');
Expect(1, 1833, '\p{Jg=qaph}', "");
Expect(0, 1833, '\p{^Jg=qaph}', "");
Expect(0, 1833, '\P{Jg=qaph}', "");
Expect(1, 1833, '\P{^Jg=qaph}', "");
Expect(0, 1834, '\p{Jg=qaph}', "");
Expect(1, 1834, '\p{^Jg=qaph}', "");
Expect(1, 1834, '\P{Jg=qaph}', "");
Expect(0, 1834, '\P{^Jg=qaph}', "");
Expect(1, 1833, '\p{Jg= -Qaph}', "");
Expect(0, 1833, '\p{^Jg= -Qaph}', "");
Expect(0, 1833, '\P{Jg= -Qaph}', "");
Expect(1, 1833, '\P{^Jg= -Qaph}', "");
Expect(0, 1834, '\p{Jg= -Qaph}', "");
Expect(1, 1834, '\p{^Jg= -Qaph}', "");
Expect(1, 1834, '\P{Jg= -Qaph}', "");
Expect(0, 1834, '\P{^Jg= -Qaph}', "");
Error('\p{Is_Joining_Group=/a/_-QAPH}');
Error('\P{Is_Joining_Group=/a/_-QAPH}');
Expect(1, 1833, '\p{Is_Joining_Group=qaph}', "");
Expect(0, 1833, '\p{^Is_Joining_Group=qaph}', "");
Expect(0, 1833, '\P{Is_Joining_Group=qaph}', "");
Expect(1, 1833, '\P{^Is_Joining_Group=qaph}', "");
Expect(0, 1834, '\p{Is_Joining_Group=qaph}', "");
Expect(1, 1834, '\p{^Is_Joining_Group=qaph}', "");
Expect(1, 1834, '\P{Is_Joining_Group=qaph}', "");
Expect(0, 1834, '\P{^Is_Joining_Group=qaph}', "");
Expect(1, 1833, '\p{Is_Joining_Group=-_QAPH}', "");
Expect(0, 1833, '\p{^Is_Joining_Group=-_QAPH}', "");
Expect(0, 1833, '\P{Is_Joining_Group=-_QAPH}', "");
Expect(1, 1833, '\P{^Is_Joining_Group=-_QAPH}', "");
Expect(0, 1834, '\p{Is_Joining_Group=-_QAPH}', "");
Expect(1, 1834, '\p{^Is_Joining_Group=-_QAPH}', "");
Expect(1, 1834, '\P{Is_Joining_Group=-_QAPH}', "");
Expect(0, 1834, '\P{^Is_Joining_Group=-_QAPH}', "");
Error('\p{Is_Jg=  Qaph/a/}');
Error('\P{Is_Jg=  Qaph/a/}');
Expect(1, 1833, '\p{Is_Jg=qaph}', "");
Expect(0, 1833, '\p{^Is_Jg=qaph}', "");
Expect(0, 1833, '\P{Is_Jg=qaph}', "");
Expect(1, 1833, '\P{^Is_Jg=qaph}', "");
Expect(0, 1834, '\p{Is_Jg=qaph}', "");
Expect(1, 1834, '\p{^Is_Jg=qaph}', "");
Expect(1, 1834, '\P{Is_Jg=qaph}', "");
Expect(0, 1834, '\P{^Is_Jg=qaph}', "");
Expect(1, 1833, '\p{Is_Jg= -Qaph}', "");
Expect(0, 1833, '\p{^Is_Jg= -Qaph}', "");
Expect(0, 1833, '\P{Is_Jg= -Qaph}', "");
Expect(1, 1833, '\P{^Is_Jg= -Qaph}', "");
Expect(0, 1834, '\p{Is_Jg= -Qaph}', "");
Expect(1, 1834, '\p{^Is_Jg= -Qaph}', "");
Expect(1, 1834, '\P{Is_Jg= -Qaph}', "");
Expect(0, 1834, '\P{^Is_Jg= -Qaph}', "");
Error('\p{Joining_Group=-Reh/a/}');
Error('\P{Joining_Group=-Reh/a/}');
Expect(1, 2218, '\p{Joining_Group=reh}', "");
Expect(0, 2218, '\p{^Joining_Group=reh}', "");
Expect(0, 2218, '\P{Joining_Group=reh}', "");
Expect(1, 2218, '\P{^Joining_Group=reh}', "");
Expect(0, 2219, '\p{Joining_Group=reh}', "");
Expect(1, 2219, '\p{^Joining_Group=reh}', "");
Expect(1, 2219, '\P{Joining_Group=reh}', "");
Expect(0, 2219, '\P{^Joining_Group=reh}', "");
Expect(1, 2218, '\p{Joining_Group=-	Reh}', "");
Expect(0, 2218, '\p{^Joining_Group=-	Reh}', "");
Expect(0, 2218, '\P{Joining_Group=-	Reh}', "");
Expect(1, 2218, '\P{^Joining_Group=-	Reh}', "");
Expect(0, 2219, '\p{Joining_Group=-	Reh}', "");
Expect(1, 2219, '\p{^Joining_Group=-	Reh}', "");
Expect(1, 2219, '\P{Joining_Group=-	Reh}', "");
Expect(0, 2219, '\P{^Joining_Group=-	Reh}', "");
Error('\p{Jg:	:= reh}');
Error('\P{Jg:	:= reh}');
Expect(1, 2218, '\p{Jg=reh}', "");
Expect(0, 2218, '\p{^Jg=reh}', "");
Expect(0, 2218, '\P{Jg=reh}', "");
Expect(1, 2218, '\P{^Jg=reh}', "");
Expect(0, 2219, '\p{Jg=reh}', "");
Expect(1, 2219, '\p{^Jg=reh}', "");
Expect(1, 2219, '\P{Jg=reh}', "");
Expect(0, 2219, '\P{^Jg=reh}', "");
Expect(1, 2218, '\p{Jg: 	_reh}', "");
Expect(0, 2218, '\p{^Jg: 	_reh}', "");
Expect(0, 2218, '\P{Jg: 	_reh}', "");
Expect(1, 2218, '\P{^Jg: 	_reh}', "");
Expect(0, 2219, '\p{Jg: 	_reh}', "");
Expect(1, 2219, '\p{^Jg: 	_reh}', "");
Expect(1, 2219, '\P{Jg: 	_reh}', "");
Expect(0, 2219, '\P{^Jg: 	_reh}', "");
Error('\p{Is_Joining_Group=:=--REH}');
Error('\P{Is_Joining_Group=:=--REH}');
Expect(1, 2218, '\p{Is_Joining_Group: reh}', "");
Expect(0, 2218, '\p{^Is_Joining_Group: reh}', "");
Expect(0, 2218, '\P{Is_Joining_Group: reh}', "");
Expect(1, 2218, '\P{^Is_Joining_Group: reh}', "");
Expect(0, 2219, '\p{Is_Joining_Group: reh}', "");
Expect(1, 2219, '\p{^Is_Joining_Group: reh}', "");
Expect(1, 2219, '\P{Is_Joining_Group: reh}', "");
Expect(0, 2219, '\P{^Is_Joining_Group: reh}', "");
Expect(1, 2218, '\p{Is_Joining_Group=__reh}', "");
Expect(0, 2218, '\p{^Is_Joining_Group=__reh}', "");
Expect(0, 2218, '\P{Is_Joining_Group=__reh}', "");
Expect(1, 2218, '\P{^Is_Joining_Group=__reh}', "");
Expect(0, 2219, '\p{Is_Joining_Group=__reh}', "");
Expect(1, 2219, '\p{^Is_Joining_Group=__reh}', "");
Expect(1, 2219, '\P{Is_Joining_Group=__reh}', "");
Expect(0, 2219, '\P{^Is_Joining_Group=__reh}', "");
Error('\p{Is_Jg=	_Reh:=}');
Error('\P{Is_Jg=	_Reh:=}');
Expect(1, 2218, '\p{Is_Jg=reh}', "");
Expect(0, 2218, '\p{^Is_Jg=reh}', "");
Expect(0, 2218, '\P{Is_Jg=reh}', "");
Expect(1, 2218, '\P{^Is_Jg=reh}', "");
Expect(0, 2219, '\p{Is_Jg=reh}', "");
Expect(1, 2219, '\p{^Is_Jg=reh}', "");
Expect(1, 2219, '\P{Is_Jg=reh}', "");
Expect(0, 2219, '\P{^Is_Jg=reh}', "");
Expect(1, 2218, '\p{Is_Jg= -REH}', "");
Expect(0, 2218, '\p{^Is_Jg= -REH}', "");
Expect(0, 2218, '\P{Is_Jg= -REH}', "");
Expect(1, 2218, '\P{^Is_Jg= -REH}', "");
Expect(0, 2219, '\p{Is_Jg= -REH}', "");
Expect(1, 2219, '\p{^Is_Jg= -REH}', "");
Expect(1, 2219, '\P{Is_Jg= -REH}', "");
Expect(0, 2219, '\P{^Is_Jg= -REH}', "");
Error('\p{Joining_Group=_:=REVERSED_PE}');
Error('\P{Joining_Group=_:=REVERSED_PE}');
Expect(1, 1831, '\p{Joining_Group:   reversedpe}', "");
Expect(0, 1831, '\p{^Joining_Group:   reversedpe}', "");
Expect(0, 1831, '\P{Joining_Group:   reversedpe}', "");
Expect(1, 1831, '\P{^Joining_Group:   reversedpe}', "");
Expect(0, 1832, '\p{Joining_Group:   reversedpe}', "");
Expect(1, 1832, '\p{^Joining_Group:   reversedpe}', "");
Expect(1, 1832, '\P{Joining_Group:   reversedpe}', "");
Expect(0, 1832, '\P{^Joining_Group:   reversedpe}', "");
Expect(1, 1831, '\p{Joining_Group=_-Reversed_Pe}', "");
Expect(0, 1831, '\p{^Joining_Group=_-Reversed_Pe}', "");
Expect(0, 1831, '\P{Joining_Group=_-Reversed_Pe}', "");
Expect(1, 1831, '\P{^Joining_Group=_-Reversed_Pe}', "");
Expect(0, 1832, '\p{Joining_Group=_-Reversed_Pe}', "");
Expect(1, 1832, '\p{^Joining_Group=_-Reversed_Pe}', "");
Expect(1, 1832, '\P{Joining_Group=_-Reversed_Pe}', "");
Expect(0, 1832, '\P{^Joining_Group=_-Reversed_Pe}', "");
Error('\p{Jg=:=REVERSED_PE}');
Error('\P{Jg=:=REVERSED_PE}');
Expect(1, 1831, '\p{Jg=reversedpe}', "");
Expect(0, 1831, '\p{^Jg=reversedpe}', "");
Expect(0, 1831, '\P{Jg=reversedpe}', "");
Expect(1, 1831, '\P{^Jg=reversedpe}', "");
Expect(0, 1832, '\p{Jg=reversedpe}', "");
Expect(1, 1832, '\p{^Jg=reversedpe}', "");
Expect(1, 1832, '\P{Jg=reversedpe}', "");
Expect(0, 1832, '\P{^Jg=reversedpe}', "");
Expect(1, 1831, '\p{Jg=REVERSED_Pe}', "");
Expect(0, 1831, '\p{^Jg=REVERSED_Pe}', "");
Expect(0, 1831, '\P{Jg=REVERSED_Pe}', "");
Expect(1, 1831, '\P{^Jg=REVERSED_Pe}', "");
Expect(0, 1832, '\p{Jg=REVERSED_Pe}', "");
Expect(1, 1832, '\p{^Jg=REVERSED_Pe}', "");
Expect(1, 1832, '\P{Jg=REVERSED_Pe}', "");
Expect(0, 1832, '\P{^Jg=REVERSED_Pe}', "");
Error('\p{Is_Joining_Group= -Reversed_PE/a/}');
Error('\P{Is_Joining_Group= -Reversed_PE/a/}');
Expect(1, 1831, '\p{Is_Joining_Group=reversedpe}', "");
Expect(0, 1831, '\p{^Is_Joining_Group=reversedpe}', "");
Expect(0, 1831, '\P{Is_Joining_Group=reversedpe}', "");
Expect(1, 1831, '\P{^Is_Joining_Group=reversedpe}', "");
Expect(0, 1832, '\p{Is_Joining_Group=reversedpe}', "");
Expect(1, 1832, '\p{^Is_Joining_Group=reversedpe}', "");
Expect(1, 1832, '\P{Is_Joining_Group=reversedpe}', "");
Expect(0, 1832, '\P{^Is_Joining_Group=reversedpe}', "");
Expect(1, 1831, '\p{Is_Joining_Group= -Reversed_pe}', "");
Expect(0, 1831, '\p{^Is_Joining_Group= -Reversed_pe}', "");
Expect(0, 1831, '\P{Is_Joining_Group= -Reversed_pe}', "");
Expect(1, 1831, '\P{^Is_Joining_Group= -Reversed_pe}', "");
Expect(0, 1832, '\p{Is_Joining_Group= -Reversed_pe}', "");
Expect(1, 1832, '\p{^Is_Joining_Group= -Reversed_pe}', "");
Expect(1, 1832, '\P{Is_Joining_Group= -Reversed_pe}', "");
Expect(0, 1832, '\P{^Is_Joining_Group= -Reversed_pe}', "");
Error('\p{Is_Jg=:=_REVERSED_Pe}');
Error('\P{Is_Jg=:=_REVERSED_Pe}');
Expect(1, 1831, '\p{Is_Jg=reversedpe}', "");
Expect(0, 1831, '\p{^Is_Jg=reversedpe}', "");
Expect(0, 1831, '\P{Is_Jg=reversedpe}', "");
Expect(1, 1831, '\P{^Is_Jg=reversedpe}', "");
Expect(0, 1832, '\p{Is_Jg=reversedpe}', "");
Expect(1, 1832, '\p{^Is_Jg=reversedpe}', "");
Expect(1, 1832, '\P{Is_Jg=reversedpe}', "");
Expect(0, 1832, '\P{^Is_Jg=reversedpe}', "");
Expect(1, 1831, '\p{Is_Jg=_Reversed_Pe}', "");
Expect(0, 1831, '\p{^Is_Jg=_Reversed_Pe}', "");
Expect(0, 1831, '\P{Is_Jg=_Reversed_Pe}', "");
Expect(1, 1831, '\P{^Is_Jg=_Reversed_Pe}', "");
Expect(0, 1832, '\p{Is_Jg=_Reversed_Pe}', "");
Expect(1, 1832, '\p{^Is_Jg=_Reversed_Pe}', "");
Expect(1, 1832, '\P{Is_Jg=_Reversed_Pe}', "");
Expect(0, 1832, '\P{^Is_Jg=_Reversed_Pe}', "");
Error('\p{Joining_Group= _Rohingya_yeh:=}');
Error('\P{Joining_Group= _Rohingya_yeh:=}');
Expect(1, 2220, '\p{Joining_Group=rohingyayeh}', "");
Expect(0, 2220, '\p{^Joining_Group=rohingyayeh}', "");
Expect(0, 2220, '\P{Joining_Group=rohingyayeh}', "");
Expect(1, 2220, '\P{^Joining_Group=rohingyayeh}', "");
Expect(0, 2221, '\p{Joining_Group=rohingyayeh}', "");
Expect(1, 2221, '\p{^Joining_Group=rohingyayeh}', "");
Expect(1, 2221, '\P{Joining_Group=rohingyayeh}', "");
Expect(0, 2221, '\P{^Joining_Group=rohingyayeh}', "");
Expect(1, 2220, '\p{Joining_Group=-Rohingya_Yeh}', "");
Expect(0, 2220, '\p{^Joining_Group=-Rohingya_Yeh}', "");
Expect(0, 2220, '\P{Joining_Group=-Rohingya_Yeh}', "");
Expect(1, 2220, '\P{^Joining_Group=-Rohingya_Yeh}', "");
Expect(0, 2221, '\p{Joining_Group=-Rohingya_Yeh}', "");
Expect(1, 2221, '\p{^Joining_Group=-Rohingya_Yeh}', "");
Expect(1, 2221, '\P{Joining_Group=-Rohingya_Yeh}', "");
Expect(0, 2221, '\P{^Joining_Group=-Rohingya_Yeh}', "");
Error('\p{Jg:		Rohingya_Yeh:=}');
Error('\P{Jg:		Rohingya_Yeh:=}');
Expect(1, 2220, '\p{Jg=rohingyayeh}', "");
Expect(0, 2220, '\p{^Jg=rohingyayeh}', "");
Expect(0, 2220, '\P{Jg=rohingyayeh}', "");
Expect(1, 2220, '\P{^Jg=rohingyayeh}', "");
Expect(0, 2221, '\p{Jg=rohingyayeh}', "");
Expect(1, 2221, '\p{^Jg=rohingyayeh}', "");
Expect(1, 2221, '\P{Jg=rohingyayeh}', "");
Expect(0, 2221, '\P{^Jg=rohingyayeh}', "");
Expect(1, 2220, '\p{Jg= Rohingya_yeh}', "");
Expect(0, 2220, '\p{^Jg= Rohingya_yeh}', "");
Expect(0, 2220, '\P{Jg= Rohingya_yeh}', "");
Expect(1, 2220, '\P{^Jg= Rohingya_yeh}', "");
Expect(0, 2221, '\p{Jg= Rohingya_yeh}', "");
Expect(1, 2221, '\p{^Jg= Rohingya_yeh}', "");
Expect(1, 2221, '\P{Jg= Rohingya_yeh}', "");
Expect(0, 2221, '\P{^Jg= Rohingya_yeh}', "");
Error('\p{Is_Joining_Group= -Rohingya_Yeh:=}');
Error('\P{Is_Joining_Group= -Rohingya_Yeh:=}');
Expect(1, 2220, '\p{Is_Joining_Group:	rohingyayeh}', "");
Expect(0, 2220, '\p{^Is_Joining_Group:	rohingyayeh}', "");
Expect(0, 2220, '\P{Is_Joining_Group:	rohingyayeh}', "");
Expect(1, 2220, '\P{^Is_Joining_Group:	rohingyayeh}', "");
Expect(0, 2221, '\p{Is_Joining_Group:	rohingyayeh}', "");
Expect(1, 2221, '\p{^Is_Joining_Group:	rohingyayeh}', "");
Expect(1, 2221, '\P{Is_Joining_Group:	rohingyayeh}', "");
Expect(0, 2221, '\P{^Is_Joining_Group:	rohingyayeh}', "");
Expect(1, 2220, '\p{Is_Joining_Group=rohingya_YEH}', "");
Expect(0, 2220, '\p{^Is_Joining_Group=rohingya_YEH}', "");
Expect(0, 2220, '\P{Is_Joining_Group=rohingya_YEH}', "");
Expect(1, 2220, '\P{^Is_Joining_Group=rohingya_YEH}', "");
Expect(0, 2221, '\p{Is_Joining_Group=rohingya_YEH}', "");
Expect(1, 2221, '\p{^Is_Joining_Group=rohingya_YEH}', "");
Expect(1, 2221, '\P{Is_Joining_Group=rohingya_YEH}', "");
Expect(0, 2221, '\P{^Is_Joining_Group=rohingya_YEH}', "");
Error('\p{Is_Jg=- rohingya_Yeh:=}');
Error('\P{Is_Jg=- rohingya_Yeh:=}');
Expect(1, 2220, '\p{Is_Jg=rohingyayeh}', "");
Expect(0, 2220, '\p{^Is_Jg=rohingyayeh}', "");
Expect(0, 2220, '\P{Is_Jg=rohingyayeh}', "");
Expect(1, 2220, '\P{^Is_Jg=rohingyayeh}', "");
Expect(0, 2221, '\p{Is_Jg=rohingyayeh}', "");
Expect(1, 2221, '\p{^Is_Jg=rohingyayeh}', "");
Expect(1, 2221, '\P{Is_Jg=rohingyayeh}', "");
Expect(0, 2221, '\P{^Is_Jg=rohingyayeh}', "");
Expect(1, 2220, '\p{Is_Jg=_-ROHINGYA_Yeh}', "");
Expect(0, 2220, '\p{^Is_Jg=_-ROHINGYA_Yeh}', "");
Expect(0, 2220, '\P{Is_Jg=_-ROHINGYA_Yeh}', "");
Expect(1, 2220, '\P{^Is_Jg=_-ROHINGYA_Yeh}', "");
Expect(0, 2221, '\p{Is_Jg=_-ROHINGYA_Yeh}', "");
Expect(1, 2221, '\p{^Is_Jg=_-ROHINGYA_Yeh}', "");
Expect(1, 2221, '\P{Is_Jg=_-ROHINGYA_Yeh}', "");
Expect(0, 2221, '\P{^Is_Jg=_-ROHINGYA_Yeh}', "");
Error('\p{Joining_Group=/a/Sad}');
Error('\P{Joining_Group=/a/Sad}');
Expect(1, 1787, '\p{Joining_Group=sad}', "");
Expect(0, 1787, '\p{^Joining_Group=sad}', "");
Expect(0, 1787, '\P{Joining_Group=sad}', "");
Expect(1, 1787, '\P{^Joining_Group=sad}', "");
Expect(0, 1788, '\p{Joining_Group=sad}', "");
Expect(1, 1788, '\p{^Joining_Group=sad}', "");
Expect(1, 1788, '\P{Joining_Group=sad}', "");
Expect(0, 1788, '\P{^Joining_Group=sad}', "");
Expect(1, 1787, '\p{Joining_Group:	_-Sad}', "");
Expect(0, 1787, '\p{^Joining_Group:	_-Sad}', "");
Expect(0, 1787, '\P{Joining_Group:	_-Sad}', "");
Expect(1, 1787, '\P{^Joining_Group:	_-Sad}', "");
Expect(0, 1788, '\p{Joining_Group:	_-Sad}', "");
Expect(1, 1788, '\p{^Joining_Group:	_-Sad}', "");
Expect(1, 1788, '\P{Joining_Group:	_-Sad}', "");
Expect(0, 1788, '\P{^Joining_Group:	_-Sad}', "");
Error('\p{Jg=/a/	-Sad}');
Error('\P{Jg=/a/	-Sad}');
Expect(1, 1787, '\p{Jg=sad}', "");
Expect(0, 1787, '\p{^Jg=sad}', "");
Expect(0, 1787, '\P{Jg=sad}', "");
Expect(1, 1787, '\P{^Jg=sad}', "");
Expect(0, 1788, '\p{Jg=sad}', "");
Expect(1, 1788, '\p{^Jg=sad}', "");
Expect(1, 1788, '\P{Jg=sad}', "");
Expect(0, 1788, '\P{^Jg=sad}', "");
Expect(1, 1787, '\p{Jg:    _SAD}', "");
Expect(0, 1787, '\p{^Jg:    _SAD}', "");
Expect(0, 1787, '\P{Jg:    _SAD}', "");
Expect(1, 1787, '\P{^Jg:    _SAD}', "");
Expect(0, 1788, '\p{Jg:    _SAD}', "");
Expect(1, 1788, '\p{^Jg:    _SAD}', "");
Expect(1, 1788, '\P{Jg:    _SAD}', "");
Expect(0, 1788, '\P{^Jg:    _SAD}', "");
Error('\p{Is_Joining_Group=_/a/sad}');
Error('\P{Is_Joining_Group=_/a/sad}');
Expect(1, 1787, '\p{Is_Joining_Group=sad}', "");
Expect(0, 1787, '\p{^Is_Joining_Group=sad}', "");
Expect(0, 1787, '\P{Is_Joining_Group=sad}', "");
Expect(1, 1787, '\P{^Is_Joining_Group=sad}', "");
Expect(0, 1788, '\p{Is_Joining_Group=sad}', "");
Expect(1, 1788, '\p{^Is_Joining_Group=sad}', "");
Expect(1, 1788, '\P{Is_Joining_Group=sad}', "");
Expect(0, 1788, '\P{^Is_Joining_Group=sad}', "");
Expect(1, 1787, '\p{Is_Joining_Group=_Sad}', "");
Expect(0, 1787, '\p{^Is_Joining_Group=_Sad}', "");
Expect(0, 1787, '\P{Is_Joining_Group=_Sad}', "");
Expect(1, 1787, '\P{^Is_Joining_Group=_Sad}', "");
Expect(0, 1788, '\p{Is_Joining_Group=_Sad}', "");
Expect(1, 1788, '\p{^Is_Joining_Group=_Sad}', "");
Expect(1, 1788, '\P{Is_Joining_Group=_Sad}', "");
Expect(0, 1788, '\P{^Is_Joining_Group=_Sad}', "");
Error('\p{Is_Jg:/a/_SAD}');
Error('\P{Is_Jg:/a/_SAD}');
Expect(1, 1787, '\p{Is_Jg=sad}', "");
Expect(0, 1787, '\p{^Is_Jg=sad}', "");
Expect(0, 1787, '\P{Is_Jg=sad}', "");
Expect(1, 1787, '\P{^Is_Jg=sad}', "");
Expect(0, 1788, '\p{Is_Jg=sad}', "");
Expect(1, 1788, '\p{^Is_Jg=sad}', "");
Expect(1, 1788, '\P{Is_Jg=sad}', "");
Expect(0, 1788, '\P{^Is_Jg=sad}', "");
Expect(1, 1787, '\p{Is_Jg=	 sad}', "");
Expect(0, 1787, '\p{^Is_Jg=	 sad}', "");
Expect(0, 1787, '\P{Is_Jg=	 sad}', "");
Expect(1, 1787, '\P{^Is_Jg=	 sad}', "");
Expect(0, 1788, '\p{Is_Jg=	 sad}', "");
Expect(1, 1788, '\p{^Is_Jg=	 sad}', "");
Expect(1, 1788, '\P{Is_Jg=	 sad}', "");
Expect(0, 1788, '\P{^Is_Jg=	 sad}', "");
Error('\p{Joining_Group=_/a/Sadhe}');
Error('\P{Joining_Group=_/a/Sadhe}');
Expect(1, 1832, '\p{Joining_Group=sadhe}', "");
Expect(0, 1832, '\p{^Joining_Group=sadhe}', "");
Expect(0, 1832, '\P{Joining_Group=sadhe}', "");
Expect(1, 1832, '\P{^Joining_Group=sadhe}', "");
Expect(0, 1833, '\p{Joining_Group=sadhe}', "");
Expect(1, 1833, '\p{^Joining_Group=sadhe}', "");
Expect(1, 1833, '\P{Joining_Group=sadhe}', "");
Expect(0, 1833, '\P{^Joining_Group=sadhe}', "");
Expect(1, 1832, '\p{Joining_Group=	_sadhe}', "");
Expect(0, 1832, '\p{^Joining_Group=	_sadhe}', "");
Expect(0, 1832, '\P{Joining_Group=	_sadhe}', "");
Expect(1, 1832, '\P{^Joining_Group=	_sadhe}', "");
Expect(0, 1833, '\p{Joining_Group=	_sadhe}', "");
Expect(1, 1833, '\p{^Joining_Group=	_sadhe}', "");
Expect(1, 1833, '\P{Joining_Group=	_sadhe}', "");
Expect(0, 1833, '\P{^Joining_Group=	_sadhe}', "");
Error('\p{Jg=--Sadhe/a/}');
Error('\P{Jg=--Sadhe/a/}');
Expect(1, 1832, '\p{Jg=sadhe}', "");
Expect(0, 1832, '\p{^Jg=sadhe}', "");
Expect(0, 1832, '\P{Jg=sadhe}', "");
Expect(1, 1832, '\P{^Jg=sadhe}', "");
Expect(0, 1833, '\p{Jg=sadhe}', "");
Expect(1, 1833, '\p{^Jg=sadhe}', "");
Expect(1, 1833, '\P{Jg=sadhe}', "");
Expect(0, 1833, '\P{^Jg=sadhe}', "");
Expect(1, 1832, '\p{Jg= 	Sadhe}', "");
Expect(0, 1832, '\p{^Jg= 	Sadhe}', "");
Expect(0, 1832, '\P{Jg= 	Sadhe}', "");
Expect(1, 1832, '\P{^Jg= 	Sadhe}', "");
Expect(0, 1833, '\p{Jg= 	Sadhe}', "");
Expect(1, 1833, '\p{^Jg= 	Sadhe}', "");
Expect(1, 1833, '\P{Jg= 	Sadhe}', "");
Expect(0, 1833, '\P{^Jg= 	Sadhe}', "");
Error('\p{Is_Joining_Group=/a/-	Sadhe}');
Error('\P{Is_Joining_Group=/a/-	Sadhe}');
Expect(1, 1832, '\p{Is_Joining_Group:sadhe}', "");
Expect(0, 1832, '\p{^Is_Joining_Group:sadhe}', "");
Expect(0, 1832, '\P{Is_Joining_Group:sadhe}', "");
Expect(1, 1832, '\P{^Is_Joining_Group:sadhe}', "");
Expect(0, 1833, '\p{Is_Joining_Group:sadhe}', "");
Expect(1, 1833, '\p{^Is_Joining_Group:sadhe}', "");
Expect(1, 1833, '\P{Is_Joining_Group:sadhe}', "");
Expect(0, 1833, '\P{^Is_Joining_Group:sadhe}', "");
Expect(1, 1832, '\p{Is_Joining_Group=Sadhe}', "");
Expect(0, 1832, '\p{^Is_Joining_Group=Sadhe}', "");
Expect(0, 1832, '\P{Is_Joining_Group=Sadhe}', "");
Expect(1, 1832, '\P{^Is_Joining_Group=Sadhe}', "");
Expect(0, 1833, '\p{Is_Joining_Group=Sadhe}', "");
Expect(1, 1833, '\p{^Is_Joining_Group=Sadhe}', "");
Expect(1, 1833, '\P{Is_Joining_Group=Sadhe}', "");
Expect(0, 1833, '\P{^Is_Joining_Group=Sadhe}', "");
Error('\p{Is_Jg=/a/_ SADHE}');
Error('\P{Is_Jg=/a/_ SADHE}');
Expect(1, 1832, '\p{Is_Jg=sadhe}', "");
Expect(0, 1832, '\p{^Is_Jg=sadhe}', "");
Expect(0, 1832, '\P{Is_Jg=sadhe}', "");
Expect(1, 1832, '\P{^Is_Jg=sadhe}', "");
Expect(0, 1833, '\p{Is_Jg=sadhe}', "");
Expect(1, 1833, '\p{^Is_Jg=sadhe}', "");
Expect(1, 1833, '\P{Is_Jg=sadhe}', "");
Expect(0, 1833, '\P{^Is_Jg=sadhe}', "");
Expect(1, 1832, '\p{Is_Jg= _Sadhe}', "");
Expect(0, 1832, '\p{^Is_Jg= _Sadhe}', "");
Expect(0, 1832, '\P{Is_Jg= _Sadhe}', "");
Expect(1, 1832, '\P{^Is_Jg= _Sadhe}', "");
Expect(0, 1833, '\p{Is_Jg= _Sadhe}', "");
Expect(1, 1833, '\p{^Is_Jg= _Sadhe}', "");
Expect(1, 1833, '\P{Is_Jg= _Sadhe}', "");
Expect(0, 1833, '\P{^Is_Jg= _Sadhe}', "");
Error('\p{Joining_Group=:= _seen}');
Error('\P{Joining_Group=:= _seen}');
Expect(1, 1918, '\p{Joining_Group=seen}', "");
Expect(0, 1918, '\p{^Joining_Group=seen}', "");
Expect(0, 1918, '\P{Joining_Group=seen}', "");
Expect(1, 1918, '\P{^Joining_Group=seen}', "");
Expect(0, 1919, '\p{Joining_Group=seen}', "");
Expect(1, 1919, '\p{^Joining_Group=seen}', "");
Expect(1, 1919, '\P{Joining_Group=seen}', "");
Expect(0, 1919, '\P{^Joining_Group=seen}', "");
Expect(1, 1918, '\p{Joining_Group=__Seen}', "");
Expect(0, 1918, '\p{^Joining_Group=__Seen}', "");
Expect(0, 1918, '\P{Joining_Group=__Seen}', "");
Expect(1, 1918, '\P{^Joining_Group=__Seen}', "");
Expect(0, 1919, '\p{Joining_Group=__Seen}', "");
Expect(1, 1919, '\p{^Joining_Group=__Seen}', "");
Expect(1, 1919, '\P{Joining_Group=__Seen}', "");
Expect(0, 1919, '\P{^Joining_Group=__Seen}', "");
Error('\p{Jg=	_seen/a/}');
Error('\P{Jg=	_seen/a/}');
Expect(1, 1918, '\p{Jg=seen}', "");
Expect(0, 1918, '\p{^Jg=seen}', "");
Expect(0, 1918, '\P{Jg=seen}', "");
Expect(1, 1918, '\P{^Jg=seen}', "");
Expect(0, 1919, '\p{Jg=seen}', "");
Expect(1, 1919, '\p{^Jg=seen}', "");
Expect(1, 1919, '\P{Jg=seen}', "");
Expect(0, 1919, '\P{^Jg=seen}', "");
Expect(1, 1918, '\p{Jg:   Seen}', "");
Expect(0, 1918, '\p{^Jg:   Seen}', "");
Expect(0, 1918, '\P{Jg:   Seen}', "");
Expect(1, 1918, '\P{^Jg:   Seen}', "");
Expect(0, 1919, '\p{Jg:   Seen}', "");
Expect(1, 1919, '\p{^Jg:   Seen}', "");
Expect(1, 1919, '\P{Jg:   Seen}', "");
Expect(0, 1919, '\P{^Jg:   Seen}', "");
Error('\p{Is_Joining_Group=:=Seen}');
Error('\P{Is_Joining_Group=:=Seen}');
Expect(1, 1918, '\p{Is_Joining_Group=seen}', "");
Expect(0, 1918, '\p{^Is_Joining_Group=seen}', "");
Expect(0, 1918, '\P{Is_Joining_Group=seen}', "");
Expect(1, 1918, '\P{^Is_Joining_Group=seen}', "");
Expect(0, 1919, '\p{Is_Joining_Group=seen}', "");
Expect(1, 1919, '\p{^Is_Joining_Group=seen}', "");
Expect(1, 1919, '\P{Is_Joining_Group=seen}', "");
Expect(0, 1919, '\P{^Is_Joining_Group=seen}', "");
Expect(1, 1918, '\p{Is_Joining_Group=_Seen}', "");
Expect(0, 1918, '\p{^Is_Joining_Group=_Seen}', "");
Expect(0, 1918, '\P{Is_Joining_Group=_Seen}', "");
Expect(1, 1918, '\P{^Is_Joining_Group=_Seen}', "");
Expect(0, 1919, '\p{Is_Joining_Group=_Seen}', "");
Expect(1, 1919, '\p{^Is_Joining_Group=_Seen}', "");
Expect(1, 1919, '\P{Is_Joining_Group=_Seen}', "");
Expect(0, 1919, '\P{^Is_Joining_Group=_Seen}', "");
Error('\p{Is_Jg=:=-Seen}');
Error('\P{Is_Jg=:=-Seen}');
Expect(1, 1918, '\p{Is_Jg:	seen}', "");
Expect(0, 1918, '\p{^Is_Jg:	seen}', "");
Expect(0, 1918, '\P{Is_Jg:	seen}', "");
Expect(1, 1918, '\P{^Is_Jg:	seen}', "");
Expect(0, 1919, '\p{Is_Jg:	seen}', "");
Expect(1, 1919, '\p{^Is_Jg:	seen}', "");
Expect(1, 1919, '\P{Is_Jg:	seen}', "");
Expect(0, 1919, '\P{^Is_Jg:	seen}', "");
Expect(1, 1918, '\p{Is_Jg=_Seen}', "");
Expect(0, 1918, '\p{^Is_Jg=_Seen}', "");
Expect(0, 1918, '\P{Is_Jg=_Seen}', "");
Expect(1, 1918, '\P{^Is_Jg=_Seen}', "");
Expect(0, 1919, '\p{Is_Jg=_Seen}', "");
Expect(1, 1919, '\p{^Is_Jg=_Seen}', "");
Expect(1, 1919, '\P{Is_Jg=_Seen}', "");
Expect(0, 1919, '\P{^Is_Jg=_Seen}', "");
Error('\p{Joining_Group=-:=Semkath}');
Error('\P{Joining_Group=-:=Semkath}');
Expect(1, 1827, '\p{Joining_Group=semkath}', "");
Expect(0, 1827, '\p{^Joining_Group=semkath}', "");
Expect(0, 1827, '\P{Joining_Group=semkath}', "");
Expect(1, 1827, '\P{^Joining_Group=semkath}', "");
Expect(0, 1828, '\p{Joining_Group=semkath}', "");
Expect(1, 1828, '\p{^Joining_Group=semkath}', "");
Expect(1, 1828, '\P{Joining_Group=semkath}', "");
Expect(0, 1828, '\P{^Joining_Group=semkath}', "");
Expect(1, 1827, '\p{Joining_Group=	-Semkath}', "");
Expect(0, 1827, '\p{^Joining_Group=	-Semkath}', "");
Expect(0, 1827, '\P{Joining_Group=	-Semkath}', "");
Expect(1, 1827, '\P{^Joining_Group=	-Semkath}', "");
Expect(0, 1828, '\p{Joining_Group=	-Semkath}', "");
Expect(1, 1828, '\p{^Joining_Group=	-Semkath}', "");
Expect(1, 1828, '\P{Joining_Group=	-Semkath}', "");
Expect(0, 1828, '\P{^Joining_Group=	-Semkath}', "");
Error('\p{Jg=		SEMKATH:=}');
Error('\P{Jg=		SEMKATH:=}');
Expect(1, 1827, '\p{Jg=semkath}', "");
Expect(0, 1827, '\p{^Jg=semkath}', "");
Expect(0, 1827, '\P{Jg=semkath}', "");
Expect(1, 1827, '\P{^Jg=semkath}', "");
Expect(0, 1828, '\p{Jg=semkath}', "");
Expect(1, 1828, '\p{^Jg=semkath}', "");
Expect(1, 1828, '\P{Jg=semkath}', "");
Expect(0, 1828, '\P{^Jg=semkath}', "");
Expect(1, 1827, '\p{Jg=-Semkath}', "");
Expect(0, 1827, '\p{^Jg=-Semkath}', "");
Expect(0, 1827, '\P{Jg=-Semkath}', "");
Expect(1, 1827, '\P{^Jg=-Semkath}', "");
Expect(0, 1828, '\p{Jg=-Semkath}', "");
Expect(1, 1828, '\p{^Jg=-Semkath}', "");
Expect(1, 1828, '\P{Jg=-Semkath}', "");
Expect(0, 1828, '\P{^Jg=-Semkath}', "");
Error('\p{Is_Joining_Group=		semkath/a/}');
Error('\P{Is_Joining_Group=		semkath/a/}');
Expect(1, 1827, '\p{Is_Joining_Group: semkath}', "");
Expect(0, 1827, '\p{^Is_Joining_Group: semkath}', "");
Expect(0, 1827, '\P{Is_Joining_Group: semkath}', "");
Expect(1, 1827, '\P{^Is_Joining_Group: semkath}', "");
Expect(0, 1828, '\p{Is_Joining_Group: semkath}', "");
Expect(1, 1828, '\p{^Is_Joining_Group: semkath}', "");
Expect(1, 1828, '\P{Is_Joining_Group: semkath}', "");
Expect(0, 1828, '\P{^Is_Joining_Group: semkath}', "");
Expect(1, 1827, '\p{Is_Joining_Group=_-Semkath}', "");
Expect(0, 1827, '\p{^Is_Joining_Group=_-Semkath}', "");
Expect(0, 1827, '\P{Is_Joining_Group=_-Semkath}', "");
Expect(1, 1827, '\P{^Is_Joining_Group=_-Semkath}', "");
Expect(0, 1828, '\p{Is_Joining_Group=_-Semkath}', "");
Expect(1, 1828, '\p{^Is_Joining_Group=_-Semkath}', "");
Expect(1, 1828, '\P{Is_Joining_Group=_-Semkath}', "");
Expect(0, 1828, '\P{^Is_Joining_Group=_-Semkath}', "");
Error('\p{Is_Jg=:=_	Semkath}');
Error('\P{Is_Jg=:=_	Semkath}');
Expect(1, 1827, '\p{Is_Jg=semkath}', "");
Expect(0, 1827, '\p{^Is_Jg=semkath}', "");
Expect(0, 1827, '\P{Is_Jg=semkath}', "");
Expect(1, 1827, '\P{^Is_Jg=semkath}', "");
Expect(0, 1828, '\p{Is_Jg=semkath}', "");
Expect(1, 1828, '\p{^Is_Jg=semkath}', "");
Expect(1, 1828, '\P{Is_Jg=semkath}', "");
Expect(0, 1828, '\P{^Is_Jg=semkath}', "");
Expect(1, 1827, '\p{Is_Jg:   __Semkath}', "");
Expect(0, 1827, '\p{^Is_Jg:   __Semkath}', "");
Expect(0, 1827, '\P{Is_Jg:   __Semkath}', "");
Expect(1, 1827, '\P{^Is_Jg:   __Semkath}', "");
Expect(0, 1828, '\p{Is_Jg:   __Semkath}', "");
Expect(1, 1828, '\p{^Is_Jg:   __Semkath}', "");
Expect(1, 1828, '\P{Is_Jg:   __Semkath}', "");
Expect(0, 1828, '\P{^Is_Jg:   __Semkath}', "");
Error('\p{Joining_Group=	/a/SHIN}');
Error('\P{Joining_Group=	/a/SHIN}');
Expect(1, 1835, '\p{Joining_Group=shin}', "");
Expect(0, 1835, '\p{^Joining_Group=shin}', "");
Expect(0, 1835, '\P{Joining_Group=shin}', "");
Expect(1, 1835, '\P{^Joining_Group=shin}', "");
Expect(0, 1836, '\p{Joining_Group=shin}', "");
Expect(1, 1836, '\p{^Joining_Group=shin}', "");
Expect(1, 1836, '\P{Joining_Group=shin}', "");
Expect(0, 1836, '\P{^Joining_Group=shin}', "");
Expect(1, 1835, '\p{Joining_Group: __Shin}', "");
Expect(0, 1835, '\p{^Joining_Group: __Shin}', "");
Expect(0, 1835, '\P{Joining_Group: __Shin}', "");
Expect(1, 1835, '\P{^Joining_Group: __Shin}', "");
Expect(0, 1836, '\p{Joining_Group: __Shin}', "");
Expect(1, 1836, '\p{^Joining_Group: __Shin}', "");
Expect(1, 1836, '\P{Joining_Group: __Shin}', "");
Expect(0, 1836, '\P{^Joining_Group: __Shin}', "");
Error('\p{Jg=-:=SHIN}');
Error('\P{Jg=-:=SHIN}');
Expect(1, 1835, '\p{Jg=shin}', "");
Expect(0, 1835, '\p{^Jg=shin}', "");
Expect(0, 1835, '\P{Jg=shin}', "");
Expect(1, 1835, '\P{^Jg=shin}', "");
Expect(0, 1836, '\p{Jg=shin}', "");
Expect(1, 1836, '\p{^Jg=shin}', "");
Expect(1, 1836, '\P{Jg=shin}', "");
Expect(0, 1836, '\P{^Jg=shin}', "");
Expect(1, 1835, '\p{Jg=	Shin}', "");
Expect(0, 1835, '\p{^Jg=	Shin}', "");
Expect(0, 1835, '\P{Jg=	Shin}', "");
Expect(1, 1835, '\P{^Jg=	Shin}', "");
Expect(0, 1836, '\p{Jg=	Shin}', "");
Expect(1, 1836, '\p{^Jg=	Shin}', "");
Expect(1, 1836, '\P{Jg=	Shin}', "");
Expect(0, 1836, '\P{^Jg=	Shin}', "");
Error('\p{Is_Joining_Group= SHIN/a/}');
Error('\P{Is_Joining_Group= SHIN/a/}');
Expect(1, 1835, '\p{Is_Joining_Group=shin}', "");
Expect(0, 1835, '\p{^Is_Joining_Group=shin}', "");
Expect(0, 1835, '\P{Is_Joining_Group=shin}', "");
Expect(1, 1835, '\P{^Is_Joining_Group=shin}', "");
Expect(0, 1836, '\p{Is_Joining_Group=shin}', "");
Expect(1, 1836, '\p{^Is_Joining_Group=shin}', "");
Expect(1, 1836, '\P{Is_Joining_Group=shin}', "");
Expect(0, 1836, '\P{^Is_Joining_Group=shin}', "");
Expect(1, 1835, '\p{Is_Joining_Group= _Shin}', "");
Expect(0, 1835, '\p{^Is_Joining_Group= _Shin}', "");
Expect(0, 1835, '\P{Is_Joining_Group= _Shin}', "");
Expect(1, 1835, '\P{^Is_Joining_Group= _Shin}', "");
Expect(0, 1836, '\p{Is_Joining_Group= _Shin}', "");
Expect(1, 1836, '\p{^Is_Joining_Group= _Shin}', "");
Expect(1, 1836, '\P{Is_Joining_Group= _Shin}', "");
Expect(0, 1836, '\P{^Is_Joining_Group= _Shin}', "");
Error('\p{Is_Jg=/a/_SHIN}');
Error('\P{Is_Jg=/a/_SHIN}');
Expect(1, 1835, '\p{Is_Jg=shin}', "");
Expect(0, 1835, '\p{^Is_Jg=shin}', "");
Expect(0, 1835, '\P{Is_Jg=shin}', "");
Expect(1, 1835, '\P{^Is_Jg=shin}', "");
Expect(0, 1836, '\p{Is_Jg=shin}', "");
Expect(1, 1836, '\p{^Is_Jg=shin}', "");
Expect(1, 1836, '\P{Is_Jg=shin}', "");
Expect(0, 1836, '\P{^Is_Jg=shin}', "");
Expect(1, 1835, '\p{Is_Jg=	shin}', "");
Expect(0, 1835, '\p{^Is_Jg=	shin}', "");
Expect(0, 1835, '\P{Is_Jg=	shin}', "");
Expect(1, 1835, '\P{^Is_Jg=	shin}', "");
Expect(0, 1836, '\p{Is_Jg=	shin}', "");
Expect(1, 1836, '\p{^Is_Jg=	shin}', "");
Expect(1, 1836, '\P{Is_Jg=	shin}', "");
Expect(0, 1836, '\P{^Is_Jg=	shin}', "");
Error('\p{Joining_Group:	SWASH_Kaf/a/}');
Error('\P{Joining_Group:	SWASH_Kaf/a/}');
Expect(1, 1706, '\p{Joining_Group=swashkaf}', "");
Expect(0, 1706, '\p{^Joining_Group=swashkaf}', "");
Expect(0, 1706, '\P{Joining_Group=swashkaf}', "");
Expect(1, 1706, '\P{^Joining_Group=swashkaf}', "");
Expect(0, 1707, '\p{Joining_Group=swashkaf}', "");
Expect(1, 1707, '\p{^Joining_Group=swashkaf}', "");
Expect(1, 1707, '\P{Joining_Group=swashkaf}', "");
Expect(0, 1707, '\P{^Joining_Group=swashkaf}', "");
Expect(1, 1706, '\p{Joining_Group=_Swash_Kaf}', "");
Expect(0, 1706, '\p{^Joining_Group=_Swash_Kaf}', "");
Expect(0, 1706, '\P{Joining_Group=_Swash_Kaf}', "");
Expect(1, 1706, '\P{^Joining_Group=_Swash_Kaf}', "");
Expect(0, 1707, '\p{Joining_Group=_Swash_Kaf}', "");
Expect(1, 1707, '\p{^Joining_Group=_Swash_Kaf}', "");
Expect(1, 1707, '\P{Joining_Group=_Swash_Kaf}', "");
Expect(0, 1707, '\P{^Joining_Group=_Swash_Kaf}', "");
Error('\p{Jg=-:=Swash_Kaf}');
Error('\P{Jg=-:=Swash_Kaf}');
Expect(1, 1706, '\p{Jg=swashkaf}', "");
Expect(0, 1706, '\p{^Jg=swashkaf}', "");
Expect(0, 1706, '\P{Jg=swashkaf}', "");
Expect(1, 1706, '\P{^Jg=swashkaf}', "");
Expect(0, 1707, '\p{Jg=swashkaf}', "");
Expect(1, 1707, '\p{^Jg=swashkaf}', "");
Expect(1, 1707, '\P{Jg=swashkaf}', "");
Expect(0, 1707, '\P{^Jg=swashkaf}', "");
Expect(1, 1706, '\p{Jg=_	Swash_kaf}', "");
Expect(0, 1706, '\p{^Jg=_	Swash_kaf}', "");
Expect(0, 1706, '\P{Jg=_	Swash_kaf}', "");
Expect(1, 1706, '\P{^Jg=_	Swash_kaf}', "");
Expect(0, 1707, '\p{Jg=_	Swash_kaf}', "");
Expect(1, 1707, '\p{^Jg=_	Swash_kaf}', "");
Expect(1, 1707, '\P{Jg=_	Swash_kaf}', "");
Expect(0, 1707, '\P{^Jg=_	Swash_kaf}', "");
Error('\p{Is_Joining_Group=:=SWASH_KAF}');
Error('\P{Is_Joining_Group=:=SWASH_KAF}');
Expect(1, 1706, '\p{Is_Joining_Group=swashkaf}', "");
Expect(0, 1706, '\p{^Is_Joining_Group=swashkaf}', "");
Expect(0, 1706, '\P{Is_Joining_Group=swashkaf}', "");
Expect(1, 1706, '\P{^Is_Joining_Group=swashkaf}', "");
Expect(0, 1707, '\p{Is_Joining_Group=swashkaf}', "");
Expect(1, 1707, '\p{^Is_Joining_Group=swashkaf}', "");
Expect(1, 1707, '\P{Is_Joining_Group=swashkaf}', "");
Expect(0, 1707, '\P{^Is_Joining_Group=swashkaf}', "");
Expect(1, 1706, '\p{Is_Joining_Group=-Swash_kaf}', "");
Expect(0, 1706, '\p{^Is_Joining_Group=-Swash_kaf}', "");
Expect(0, 1706, '\P{Is_Joining_Group=-Swash_kaf}', "");
Expect(1, 1706, '\P{^Is_Joining_Group=-Swash_kaf}', "");
Expect(0, 1707, '\p{Is_Joining_Group=-Swash_kaf}', "");
Expect(1, 1707, '\p{^Is_Joining_Group=-Swash_kaf}', "");
Expect(1, 1707, '\P{Is_Joining_Group=-Swash_kaf}', "");
Expect(0, 1707, '\P{^Is_Joining_Group=-Swash_kaf}', "");
Error('\p{Is_Jg=		SWASH_KAF:=}');
Error('\P{Is_Jg=		SWASH_KAF:=}');
Expect(1, 1706, '\p{Is_Jg=swashkaf}', "");
Expect(0, 1706, '\p{^Is_Jg=swashkaf}', "");
Expect(0, 1706, '\P{Is_Jg=swashkaf}', "");
Expect(1, 1706, '\P{^Is_Jg=swashkaf}', "");
Expect(0, 1707, '\p{Is_Jg=swashkaf}', "");
Expect(1, 1707, '\p{^Is_Jg=swashkaf}', "");
Expect(1, 1707, '\P{Is_Jg=swashkaf}', "");
Expect(0, 1707, '\P{^Is_Jg=swashkaf}', "");
Expect(1, 1706, '\p{Is_Jg=-_SWASH_Kaf}', "");
Expect(0, 1706, '\p{^Is_Jg=-_SWASH_Kaf}', "");
Expect(0, 1706, '\P{Is_Jg=-_SWASH_Kaf}', "");
Expect(1, 1706, '\P{^Is_Jg=-_SWASH_Kaf}', "");
Expect(0, 1707, '\p{Is_Jg=-_SWASH_Kaf}', "");
Expect(1, 1707, '\p{^Is_Jg=-_SWASH_Kaf}', "");
Expect(1, 1707, '\P{Is_Jg=-_SWASH_Kaf}', "");
Expect(0, 1707, '\P{^Is_Jg=-_SWASH_Kaf}', "");
Error('\p{Joining_Group=:=Syriac_Waw}');
Error('\P{Joining_Group=:=Syriac_Waw}');
Expect(1, 1816, '\p{Joining_Group=syriacwaw}', "");
Expect(0, 1816, '\p{^Joining_Group=syriacwaw}', "");
Expect(0, 1816, '\P{Joining_Group=syriacwaw}', "");
Expect(1, 1816, '\P{^Joining_Group=syriacwaw}', "");
Expect(0, 1817, '\p{Joining_Group=syriacwaw}', "");
Expect(1, 1817, '\p{^Joining_Group=syriacwaw}', "");
Expect(1, 1817, '\P{Joining_Group=syriacwaw}', "");
Expect(0, 1817, '\P{^Joining_Group=syriacwaw}', "");
Expect(1, 1816, '\p{Joining_Group=-Syriac_waw}', "");
Expect(0, 1816, '\p{^Joining_Group=-Syriac_waw}', "");
Expect(0, 1816, '\P{Joining_Group=-Syriac_waw}', "");
Expect(1, 1816, '\P{^Joining_Group=-Syriac_waw}', "");
Expect(0, 1817, '\p{Joining_Group=-Syriac_waw}', "");
Expect(1, 1817, '\p{^Joining_Group=-Syriac_waw}', "");
Expect(1, 1817, '\P{Joining_Group=-Syriac_waw}', "");
Expect(0, 1817, '\P{^Joining_Group=-Syriac_waw}', "");
Error('\p{Jg=:=	syriac_Waw}');
Error('\P{Jg=:=	syriac_Waw}');
Expect(1, 1816, '\p{Jg=syriacwaw}', "");
Expect(0, 1816, '\p{^Jg=syriacwaw}', "");
Expect(0, 1816, '\P{Jg=syriacwaw}', "");
Expect(1, 1816, '\P{^Jg=syriacwaw}', "");
Expect(0, 1817, '\p{Jg=syriacwaw}', "");
Expect(1, 1817, '\p{^Jg=syriacwaw}', "");
Expect(1, 1817, '\P{Jg=syriacwaw}', "");
Expect(0, 1817, '\P{^Jg=syriacwaw}', "");
Expect(1, 1816, '\p{Jg=-Syriac_Waw}', "");
Expect(0, 1816, '\p{^Jg=-Syriac_Waw}', "");
Expect(0, 1816, '\P{Jg=-Syriac_Waw}', "");
Expect(1, 1816, '\P{^Jg=-Syriac_Waw}', "");
Expect(0, 1817, '\p{Jg=-Syriac_Waw}', "");
Expect(1, 1817, '\p{^Jg=-Syriac_Waw}', "");
Expect(1, 1817, '\P{Jg=-Syriac_Waw}', "");
Expect(0, 1817, '\P{^Jg=-Syriac_Waw}', "");
Error('\p{Is_Joining_Group=--SYRIAC_Waw/a/}');
Error('\P{Is_Joining_Group=--SYRIAC_Waw/a/}');
Expect(1, 1816, '\p{Is_Joining_Group=syriacwaw}', "");
Expect(0, 1816, '\p{^Is_Joining_Group=syriacwaw}', "");
Expect(0, 1816, '\P{Is_Joining_Group=syriacwaw}', "");
Expect(1, 1816, '\P{^Is_Joining_Group=syriacwaw}', "");
Expect(0, 1817, '\p{Is_Joining_Group=syriacwaw}', "");
Expect(1, 1817, '\p{^Is_Joining_Group=syriacwaw}', "");
Expect(1, 1817, '\P{Is_Joining_Group=syriacwaw}', "");
Expect(0, 1817, '\P{^Is_Joining_Group=syriacwaw}', "");
Expect(1, 1816, '\p{Is_Joining_Group:SYRIAC_waw}', "");
Expect(0, 1816, '\p{^Is_Joining_Group:SYRIAC_waw}', "");
Expect(0, 1816, '\P{Is_Joining_Group:SYRIAC_waw}', "");
Expect(1, 1816, '\P{^Is_Joining_Group:SYRIAC_waw}', "");
Expect(0, 1817, '\p{Is_Joining_Group:SYRIAC_waw}', "");
Expect(1, 1817, '\p{^Is_Joining_Group:SYRIAC_waw}', "");
Expect(1, 1817, '\P{Is_Joining_Group:SYRIAC_waw}', "");
Expect(0, 1817, '\P{^Is_Joining_Group:SYRIAC_waw}', "");
Error('\p{Is_Jg=:= Syriac_waw}');
Error('\P{Is_Jg=:= Syriac_waw}');
Expect(1, 1816, '\p{Is_Jg=syriacwaw}', "");
Expect(0, 1816, '\p{^Is_Jg=syriacwaw}', "");
Expect(0, 1816, '\P{Is_Jg=syriacwaw}', "");
Expect(1, 1816, '\P{^Is_Jg=syriacwaw}', "");
Expect(0, 1817, '\p{Is_Jg=syriacwaw}', "");
Expect(1, 1817, '\p{^Is_Jg=syriacwaw}', "");
Expect(1, 1817, '\P{Is_Jg=syriacwaw}', "");
Expect(0, 1817, '\P{^Is_Jg=syriacwaw}', "");
Expect(1, 1816, '\p{Is_Jg= 	Syriac_Waw}', "");
Expect(0, 1816, '\p{^Is_Jg= 	Syriac_Waw}', "");
Expect(0, 1816, '\P{Is_Jg= 	Syriac_Waw}', "");
Expect(1, 1816, '\P{^Is_Jg= 	Syriac_Waw}', "");
Expect(0, 1817, '\p{Is_Jg= 	Syriac_Waw}', "");
Expect(1, 1817, '\p{^Is_Jg= 	Syriac_Waw}', "");
Expect(1, 1817, '\P{Is_Jg= 	Syriac_Waw}', "");
Expect(0, 1817, '\P{^Is_Jg= 	Syriac_Waw}', "");
Error('\p{Joining_Group=-Tah:=}');
Error('\P{Joining_Group=-Tah:=}');
Expect(1, 2211, '\p{Joining_Group=tah}', "");
Expect(0, 2211, '\p{^Joining_Group=tah}', "");
Expect(0, 2211, '\P{Joining_Group=tah}', "");
Expect(1, 2211, '\P{^Joining_Group=tah}', "");
Expect(0, 2212, '\p{Joining_Group=tah}', "");
Expect(1, 2212, '\p{^Joining_Group=tah}', "");
Expect(1, 2212, '\P{Joining_Group=tah}', "");
Expect(0, 2212, '\P{^Joining_Group=tah}', "");
Expect(1, 2211, '\p{Joining_Group:	TAH}', "");
Expect(0, 2211, '\p{^Joining_Group:	TAH}', "");
Expect(0, 2211, '\P{Joining_Group:	TAH}', "");
Expect(1, 2211, '\P{^Joining_Group:	TAH}', "");
Expect(0, 2212, '\p{Joining_Group:	TAH}', "");
Expect(1, 2212, '\p{^Joining_Group:	TAH}', "");
Expect(1, 2212, '\P{Joining_Group:	TAH}', "");
Expect(0, 2212, '\P{^Joining_Group:	TAH}', "");
Error('\p{Jg=/a/ _Tah}');
Error('\P{Jg=/a/ _Tah}');
Expect(1, 2211, '\p{Jg=tah}', "");
Expect(0, 2211, '\p{^Jg=tah}', "");
Expect(0, 2211, '\P{Jg=tah}', "");
Expect(1, 2211, '\P{^Jg=tah}', "");
Expect(0, 2212, '\p{Jg=tah}', "");
Expect(1, 2212, '\p{^Jg=tah}', "");
Expect(1, 2212, '\P{Jg=tah}', "");
Expect(0, 2212, '\P{^Jg=tah}', "");
Expect(1, 2211, '\p{Jg:			Tah}', "");
Expect(0, 2211, '\p{^Jg:			Tah}', "");
Expect(0, 2211, '\P{Jg:			Tah}', "");
Expect(1, 2211, '\P{^Jg:			Tah}', "");
Expect(0, 2212, '\p{Jg:			Tah}', "");
Expect(1, 2212, '\p{^Jg:			Tah}', "");
Expect(1, 2212, '\P{Jg:			Tah}', "");
Expect(0, 2212, '\P{^Jg:			Tah}', "");
Error('\p{Is_Joining_Group=--TAH:=}');
Error('\P{Is_Joining_Group=--TAH:=}');
Expect(1, 2211, '\p{Is_Joining_Group:   tah}', "");
Expect(0, 2211, '\p{^Is_Joining_Group:   tah}', "");
Expect(0, 2211, '\P{Is_Joining_Group:   tah}', "");
Expect(1, 2211, '\P{^Is_Joining_Group:   tah}', "");
Expect(0, 2212, '\p{Is_Joining_Group:   tah}', "");
Expect(1, 2212, '\p{^Is_Joining_Group:   tah}', "");
Expect(1, 2212, '\P{Is_Joining_Group:   tah}', "");
Expect(0, 2212, '\P{^Is_Joining_Group:   tah}', "");
Expect(1, 2211, '\p{Is_Joining_Group=- TAH}', "");
Expect(0, 2211, '\p{^Is_Joining_Group=- TAH}', "");
Expect(0, 2211, '\P{Is_Joining_Group=- TAH}', "");
Expect(1, 2211, '\P{^Is_Joining_Group=- TAH}', "");
Expect(0, 2212, '\p{Is_Joining_Group=- TAH}', "");
Expect(1, 2212, '\p{^Is_Joining_Group=- TAH}', "");
Expect(1, 2212, '\P{Is_Joining_Group=- TAH}', "");
Expect(0, 2212, '\P{^Is_Joining_Group=- TAH}', "");
Error('\p{Is_Jg= -TAH/a/}');
Error('\P{Is_Jg= -TAH/a/}');
Expect(1, 2211, '\p{Is_Jg=tah}', "");
Expect(0, 2211, '\p{^Is_Jg=tah}', "");
Expect(0, 2211, '\P{Is_Jg=tah}', "");
Expect(1, 2211, '\P{^Is_Jg=tah}', "");
Expect(0, 2212, '\p{Is_Jg=tah}', "");
Expect(1, 2212, '\p{^Is_Jg=tah}', "");
Expect(1, 2212, '\P{Is_Jg=tah}', "");
Expect(0, 2212, '\P{^Is_Jg=tah}', "");
Expect(1, 2211, '\p{Is_Jg=  Tah}', "");
Expect(0, 2211, '\p{^Is_Jg=  Tah}', "");
Expect(0, 2211, '\P{Is_Jg=  Tah}', "");
Expect(1, 2211, '\P{^Is_Jg=  Tah}', "");
Expect(0, 2212, '\p{Is_Jg=  Tah}', "");
Expect(1, 2212, '\p{^Is_Jg=  Tah}', "");
Expect(1, 2212, '\P{Is_Jg=  Tah}', "");
Expect(0, 2212, '\P{^Is_Jg=  Tah}', "");
Error('\p{Joining_Group=:= _Taw}');
Error('\P{Joining_Group=:= _Taw}');
Expect(1, 1836, '\p{Joining_Group=taw}', "");
Expect(0, 1836, '\p{^Joining_Group=taw}', "");
Expect(0, 1836, '\P{Joining_Group=taw}', "");
Expect(1, 1836, '\P{^Joining_Group=taw}', "");
Expect(0, 1837, '\p{Joining_Group=taw}', "");
Expect(1, 1837, '\p{^Joining_Group=taw}', "");
Expect(1, 1837, '\P{Joining_Group=taw}', "");
Expect(0, 1837, '\P{^Joining_Group=taw}', "");
Expect(1, 1836, '\p{Joining_Group= Taw}', "");
Expect(0, 1836, '\p{^Joining_Group= Taw}', "");
Expect(0, 1836, '\P{Joining_Group= Taw}', "");
Expect(1, 1836, '\P{^Joining_Group= Taw}', "");
Expect(0, 1837, '\p{Joining_Group= Taw}', "");
Expect(1, 1837, '\p{^Joining_Group= Taw}', "");
Expect(1, 1837, '\P{Joining_Group= Taw}', "");
Expect(0, 1837, '\P{^Joining_Group= Taw}', "");
Error('\p{Jg: -:=TAW}');
Error('\P{Jg: -:=TAW}');
Expect(1, 1836, '\p{Jg=taw}', "");
Expect(0, 1836, '\p{^Jg=taw}', "");
Expect(0, 1836, '\P{Jg=taw}', "");
Expect(1, 1836, '\P{^Jg=taw}', "");
Expect(0, 1837, '\p{Jg=taw}', "");
Expect(1, 1837, '\p{^Jg=taw}', "");
Expect(1, 1837, '\P{Jg=taw}', "");
Expect(0, 1837, '\P{^Jg=taw}', "");
Expect(1, 1836, '\p{Jg=_taw}', "");
Expect(0, 1836, '\p{^Jg=_taw}', "");
Expect(0, 1836, '\P{Jg=_taw}', "");
Expect(1, 1836, '\P{^Jg=_taw}', "");
Expect(0, 1837, '\p{Jg=_taw}', "");
Expect(1, 1837, '\p{^Jg=_taw}', "");
Expect(1, 1837, '\P{Jg=_taw}', "");
Expect(0, 1837, '\P{^Jg=_taw}', "");
Error('\p{Is_Joining_Group=-_Taw/a/}');
Error('\P{Is_Joining_Group=-_Taw/a/}');
Expect(1, 1836, '\p{Is_Joining_Group=taw}', "");
Expect(0, 1836, '\p{^Is_Joining_Group=taw}', "");
Expect(0, 1836, '\P{Is_Joining_Group=taw}', "");
Expect(1, 1836, '\P{^Is_Joining_Group=taw}', "");
Expect(0, 1837, '\p{Is_Joining_Group=taw}', "");
Expect(1, 1837, '\p{^Is_Joining_Group=taw}', "");
Expect(1, 1837, '\P{Is_Joining_Group=taw}', "");
Expect(0, 1837, '\P{^Is_Joining_Group=taw}', "");
Expect(1, 1836, '\p{Is_Joining_Group=- taw}', "");
Expect(0, 1836, '\p{^Is_Joining_Group=- taw}', "");
Expect(0, 1836, '\P{Is_Joining_Group=- taw}', "");
Expect(1, 1836, '\P{^Is_Joining_Group=- taw}', "");
Expect(0, 1837, '\p{Is_Joining_Group=- taw}', "");
Expect(1, 1837, '\p{^Is_Joining_Group=- taw}', "");
Expect(1, 1837, '\P{Is_Joining_Group=- taw}', "");
Expect(0, 1837, '\P{^Is_Joining_Group=- taw}', "");
Error('\p{Is_Jg= /a/Taw}');
Error('\P{Is_Jg= /a/Taw}');
Expect(1, 1836, '\p{Is_Jg=taw}', "");
Expect(0, 1836, '\p{^Is_Jg=taw}', "");
Expect(0, 1836, '\P{Is_Jg=taw}', "");
Expect(1, 1836, '\P{^Is_Jg=taw}', "");
Expect(0, 1837, '\p{Is_Jg=taw}', "");
Expect(1, 1837, '\p{^Is_Jg=taw}', "");
Expect(1, 1837, '\P{Is_Jg=taw}', "");
Expect(0, 1837, '\P{^Is_Jg=taw}', "");
Expect(1, 1836, '\p{Is_Jg= taw}', "");
Expect(0, 1836, '\p{^Is_Jg= taw}', "");
Expect(0, 1836, '\P{Is_Jg= taw}', "");
Expect(1, 1836, '\P{^Is_Jg= taw}', "");
Expect(0, 1837, '\p{Is_Jg= taw}', "");
Expect(1, 1837, '\p{^Is_Jg= taw}', "");
Expect(1, 1837, '\P{Is_Jg= taw}', "");
Expect(0, 1837, '\P{^Is_Jg= taw}', "");
Error('\p{Joining_Group=_:=Teh_MARBUTA}');
Error('\P{Joining_Group=_:=Teh_MARBUTA}');
Expect(1, 1749, '\p{Joining_Group=tehmarbuta}', "");
Expect(0, 1749, '\p{^Joining_Group=tehmarbuta}', "");
Expect(0, 1749, '\P{Joining_Group=tehmarbuta}', "");
Expect(1, 1749, '\P{^Joining_Group=tehmarbuta}', "");
Expect(0, 1750, '\p{Joining_Group=tehmarbuta}', "");
Expect(1, 1750, '\p{^Joining_Group=tehmarbuta}', "");
Expect(1, 1750, '\P{Joining_Group=tehmarbuta}', "");
Expect(0, 1750, '\P{^Joining_Group=tehmarbuta}', "");
Expect(1, 1749, '\p{Joining_Group=_teh_Marbuta}', "");
Expect(0, 1749, '\p{^Joining_Group=_teh_Marbuta}', "");
Expect(0, 1749, '\P{Joining_Group=_teh_Marbuta}', "");
Expect(1, 1749, '\P{^Joining_Group=_teh_Marbuta}', "");
Expect(0, 1750, '\p{Joining_Group=_teh_Marbuta}', "");
Expect(1, 1750, '\p{^Joining_Group=_teh_Marbuta}', "");
Expect(1, 1750, '\P{Joining_Group=_teh_Marbuta}', "");
Expect(0, 1750, '\P{^Joining_Group=_teh_Marbuta}', "");
Error('\p{Jg=_ teh_Marbuta:=}');
Error('\P{Jg=_ teh_Marbuta:=}');
Expect(1, 1749, '\p{Jg=tehmarbuta}', "");
Expect(0, 1749, '\p{^Jg=tehmarbuta}', "");
Expect(0, 1749, '\P{Jg=tehmarbuta}', "");
Expect(1, 1749, '\P{^Jg=tehmarbuta}', "");
Expect(0, 1750, '\p{Jg=tehmarbuta}', "");
Expect(1, 1750, '\p{^Jg=tehmarbuta}', "");
Expect(1, 1750, '\P{Jg=tehmarbuta}', "");
Expect(0, 1750, '\P{^Jg=tehmarbuta}', "");
Expect(1, 1749, '\p{Jg=__Teh_MARBUTA}', "");
Expect(0, 1749, '\p{^Jg=__Teh_MARBUTA}', "");
Expect(0, 1749, '\P{Jg=__Teh_MARBUTA}', "");
Expect(1, 1749, '\P{^Jg=__Teh_MARBUTA}', "");
Expect(0, 1750, '\p{Jg=__Teh_MARBUTA}', "");
Expect(1, 1750, '\p{^Jg=__Teh_MARBUTA}', "");
Expect(1, 1750, '\P{Jg=__Teh_MARBUTA}', "");
Expect(0, 1750, '\P{^Jg=__Teh_MARBUTA}', "");
Error('\p{Is_Joining_Group:   -Teh_Marbuta:=}');
Error('\P{Is_Joining_Group:   -Teh_Marbuta:=}');
Expect(1, 1749, '\p{Is_Joining_Group=tehmarbuta}', "");
Expect(0, 1749, '\p{^Is_Joining_Group=tehmarbuta}', "");
Expect(0, 1749, '\P{Is_Joining_Group=tehmarbuta}', "");
Expect(1, 1749, '\P{^Is_Joining_Group=tehmarbuta}', "");
Expect(0, 1750, '\p{Is_Joining_Group=tehmarbuta}', "");
Expect(1, 1750, '\p{^Is_Joining_Group=tehmarbuta}', "");
Expect(1, 1750, '\P{Is_Joining_Group=tehmarbuta}', "");
Expect(0, 1750, '\P{^Is_Joining_Group=tehmarbuta}', "");
Expect(1, 1749, '\p{Is_Joining_Group=- Teh_marbuta}', "");
Expect(0, 1749, '\p{^Is_Joining_Group=- Teh_marbuta}', "");
Expect(0, 1749, '\P{Is_Joining_Group=- Teh_marbuta}', "");
Expect(1, 1749, '\P{^Is_Joining_Group=- Teh_marbuta}', "");
Expect(0, 1750, '\p{Is_Joining_Group=- Teh_marbuta}', "");
Expect(1, 1750, '\p{^Is_Joining_Group=- Teh_marbuta}', "");
Expect(1, 1750, '\P{Is_Joining_Group=- Teh_marbuta}', "");
Expect(0, 1750, '\P{^Is_Joining_Group=- Teh_marbuta}', "");
Error('\p{Is_Jg=-	Teh_Marbuta:=}');
Error('\P{Is_Jg=-	Teh_Marbuta:=}');
Expect(1, 1749, '\p{Is_Jg=tehmarbuta}', "");
Expect(0, 1749, '\p{^Is_Jg=tehmarbuta}', "");
Expect(0, 1749, '\P{Is_Jg=tehmarbuta}', "");
Expect(1, 1749, '\P{^Is_Jg=tehmarbuta}', "");
Expect(0, 1750, '\p{Is_Jg=tehmarbuta}', "");
Expect(1, 1750, '\p{^Is_Jg=tehmarbuta}', "");
Expect(1, 1750, '\P{Is_Jg=tehmarbuta}', "");
Expect(0, 1750, '\P{^Is_Jg=tehmarbuta}', "");
Expect(1, 1749, '\p{Is_Jg=- TEH_marbuta}', "");
Expect(0, 1749, '\p{^Is_Jg=- TEH_marbuta}', "");
Expect(0, 1749, '\P{Is_Jg=- TEH_marbuta}', "");
Expect(1, 1749, '\P{^Is_Jg=- TEH_marbuta}', "");
Expect(0, 1750, '\p{Is_Jg=- TEH_marbuta}', "");
Expect(1, 1750, '\p{^Is_Jg=- TEH_marbuta}', "");
Expect(1, 1750, '\P{Is_Jg=- TEH_marbuta}', "");
Expect(0, 1750, '\P{^Is_Jg=- TEH_marbuta}', "");
Error('\p{Joining_Group=/a/Hamza_On_HEH_goal}');
Error('\P{Joining_Group=/a/Hamza_On_HEH_goal}');
Expect(1, 1731, '\p{Joining_Group:	hamzaonhehgoal}', "");
Expect(0, 1731, '\p{^Joining_Group:	hamzaonhehgoal}', "");
Expect(0, 1731, '\P{Joining_Group:	hamzaonhehgoal}', "");
Expect(1, 1731, '\P{^Joining_Group:	hamzaonhehgoal}', "");
Expect(0, 1732, '\p{Joining_Group:	hamzaonhehgoal}', "");
Expect(1, 1732, '\p{^Joining_Group:	hamzaonhehgoal}', "");
Expect(1, 1732, '\P{Joining_Group:	hamzaonhehgoal}', "");
Expect(0, 1732, '\P{^Joining_Group:	hamzaonhehgoal}', "");
Expect(1, 1731, '\p{Joining_Group=	Hamza_on_Heh_goal}', "");
Expect(0, 1731, '\p{^Joining_Group=	Hamza_on_Heh_goal}', "");
Expect(0, 1731, '\P{Joining_Group=	Hamza_on_Heh_goal}', "");
Expect(1, 1731, '\P{^Joining_Group=	Hamza_on_Heh_goal}', "");
Expect(0, 1732, '\p{Joining_Group=	Hamza_on_Heh_goal}', "");
Expect(1, 1732, '\p{^Joining_Group=	Hamza_on_Heh_goal}', "");
Expect(1, 1732, '\P{Joining_Group=	Hamza_on_Heh_goal}', "");
Expect(0, 1732, '\P{^Joining_Group=	Hamza_on_Heh_goal}', "");
Error('\p{Jg=:=-	Teh_Marbuta_Goal}');
Error('\P{Jg=:=-	Teh_Marbuta_Goal}');
Expect(1, 1731, '\p{Jg=tehmarbutagoal}', "");
Expect(0, 1731, '\p{^Jg=tehmarbutagoal}', "");
Expect(0, 1731, '\P{Jg=tehmarbutagoal}', "");
Expect(1, 1731, '\P{^Jg=tehmarbutagoal}', "");
Expect(0, 1732, '\p{Jg=tehmarbutagoal}', "");
Expect(1, 1732, '\p{^Jg=tehmarbutagoal}', "");
Expect(1, 1732, '\P{Jg=tehmarbutagoal}', "");
Expect(0, 1732, '\P{^Jg=tehmarbutagoal}', "");
Expect(1, 1731, '\p{Jg=__Teh_marbuta_goal}', "");
Expect(0, 1731, '\p{^Jg=__Teh_marbuta_goal}', "");
Expect(0, 1731, '\P{Jg=__Teh_marbuta_goal}', "");
Expect(1, 1731, '\P{^Jg=__Teh_marbuta_goal}', "");
Expect(0, 1732, '\p{Jg=__Teh_marbuta_goal}', "");
Expect(1, 1732, '\p{^Jg=__Teh_marbuta_goal}', "");
Expect(1, 1732, '\P{Jg=__Teh_marbuta_goal}', "");
Expect(0, 1732, '\P{^Jg=__Teh_marbuta_goal}', "");
Error('\p{Is_Joining_Group=	 Hamza_On_Heh_goal/a/}');
Error('\P{Is_Joining_Group=	 Hamza_On_Heh_goal/a/}');
Expect(1, 1731, '\p{Is_Joining_Group=hamzaonhehgoal}', "");
Expect(0, 1731, '\p{^Is_Joining_Group=hamzaonhehgoal}', "");
Expect(0, 1731, '\P{Is_Joining_Group=hamzaonhehgoal}', "");
Expect(1, 1731, '\P{^Is_Joining_Group=hamzaonhehgoal}', "");
Expect(0, 1732, '\p{Is_Joining_Group=hamzaonhehgoal}', "");
Expect(1, 1732, '\p{^Is_Joining_Group=hamzaonhehgoal}', "");
Expect(1, 1732, '\P{Is_Joining_Group=hamzaonhehgoal}', "");
Expect(0, 1732, '\P{^Is_Joining_Group=hamzaonhehgoal}', "");
Expect(1, 1731, '\p{Is_Joining_Group=-	hamza_ON_heh_goal}', "");
Expect(0, 1731, '\p{^Is_Joining_Group=-	hamza_ON_heh_goal}', "");
Expect(0, 1731, '\P{Is_Joining_Group=-	hamza_ON_heh_goal}', "");
Expect(1, 1731, '\P{^Is_Joining_Group=-	hamza_ON_heh_goal}', "");
Expect(0, 1732, '\p{Is_Joining_Group=-	hamza_ON_heh_goal}', "");
Expect(1, 1732, '\p{^Is_Joining_Group=-	hamza_ON_heh_goal}', "");
Expect(1, 1732, '\P{Is_Joining_Group=-	hamza_ON_heh_goal}', "");
Expect(0, 1732, '\P{^Is_Joining_Group=-	hamza_ON_heh_goal}', "");
Error('\p{Is_Jg=teh_MARBUTA_goal:=}');
Error('\P{Is_Jg=teh_MARBUTA_goal:=}');
Expect(1, 1731, '\p{Is_Jg=tehmarbutagoal}', "");
Expect(0, 1731, '\p{^Is_Jg=tehmarbutagoal}', "");
Expect(0, 1731, '\P{Is_Jg=tehmarbutagoal}', "");
Expect(1, 1731, '\P{^Is_Jg=tehmarbutagoal}', "");
Expect(0, 1732, '\p{Is_Jg=tehmarbutagoal}', "");
Expect(1, 1732, '\p{^Is_Jg=tehmarbutagoal}', "");
Expect(1, 1732, '\P{Is_Jg=tehmarbutagoal}', "");
Expect(0, 1732, '\P{^Is_Jg=tehmarbutagoal}', "");
Expect(1, 1731, '\p{Is_Jg=_Teh_MARBUTA_GOAL}', "");
Expect(0, 1731, '\p{^Is_Jg=_Teh_MARBUTA_GOAL}', "");
Expect(0, 1731, '\P{Is_Jg=_Teh_MARBUTA_GOAL}', "");
Expect(1, 1731, '\P{^Is_Jg=_Teh_MARBUTA_GOAL}', "");
Expect(0, 1732, '\p{Is_Jg=_Teh_MARBUTA_GOAL}', "");
Expect(1, 1732, '\p{^Is_Jg=_Teh_MARBUTA_GOAL}', "");
Expect(1, 1732, '\P{Is_Jg=_Teh_MARBUTA_GOAL}', "");
Expect(0, 1732, '\P{^Is_Jg=_Teh_MARBUTA_GOAL}', "");
Error('\p{Joining_Group=:=-TETH}');
Error('\P{Joining_Group=:=-TETH}');
Expect(1, 1820, '\p{Joining_Group=teth}', "");
Expect(0, 1820, '\p{^Joining_Group=teth}', "");
Expect(0, 1820, '\P{Joining_Group=teth}', "");
Expect(1, 1820, '\P{^Joining_Group=teth}', "");
Expect(0, 1821, '\p{Joining_Group=teth}', "");
Expect(1, 1821, '\p{^Joining_Group=teth}', "");
Expect(1, 1821, '\P{Joining_Group=teth}', "");
Expect(0, 1821, '\P{^Joining_Group=teth}', "");
Expect(1, 1820, '\p{Joining_Group: _-Teth}', "");
Expect(0, 1820, '\p{^Joining_Group: _-Teth}', "");
Expect(0, 1820, '\P{Joining_Group: _-Teth}', "");
Expect(1, 1820, '\P{^Joining_Group: _-Teth}', "");
Expect(0, 1821, '\p{Joining_Group: _-Teth}', "");
Expect(1, 1821, '\p{^Joining_Group: _-Teth}', "");
Expect(1, 1821, '\P{Joining_Group: _-Teth}', "");
Expect(0, 1821, '\P{^Joining_Group: _-Teth}', "");
Error('\p{Jg= -teth/a/}');
Error('\P{Jg= -teth/a/}');
Expect(1, 1820, '\p{Jg=teth}', "");
Expect(0, 1820, '\p{^Jg=teth}', "");
Expect(0, 1820, '\P{Jg=teth}', "");
Expect(1, 1820, '\P{^Jg=teth}', "");
Expect(0, 1821, '\p{Jg=teth}', "");
Expect(1, 1821, '\p{^Jg=teth}', "");
Expect(1, 1821, '\P{Jg=teth}', "");
Expect(0, 1821, '\P{^Jg=teth}', "");
Expect(1, 1820, '\p{Jg=-Teth}', "");
Expect(0, 1820, '\p{^Jg=-Teth}', "");
Expect(0, 1820, '\P{Jg=-Teth}', "");
Expect(1, 1820, '\P{^Jg=-Teth}', "");
Expect(0, 1821, '\p{Jg=-Teth}', "");
Expect(1, 1821, '\p{^Jg=-Teth}', "");
Expect(1, 1821, '\P{Jg=-Teth}', "");
Expect(0, 1821, '\P{^Jg=-Teth}', "");
Error('\p{Is_Joining_Group=:=- TETH}');
Error('\P{Is_Joining_Group=:=- TETH}');
Expect(1, 1820, '\p{Is_Joining_Group=teth}', "");
Expect(0, 1820, '\p{^Is_Joining_Group=teth}', "");
Expect(0, 1820, '\P{Is_Joining_Group=teth}', "");
Expect(1, 1820, '\P{^Is_Joining_Group=teth}', "");
Expect(0, 1821, '\p{Is_Joining_Group=teth}', "");
Expect(1, 1821, '\p{^Is_Joining_Group=teth}', "");
Expect(1, 1821, '\P{Is_Joining_Group=teth}', "");
Expect(0, 1821, '\P{^Is_Joining_Group=teth}', "");
Expect(1, 1820, '\p{Is_Joining_Group=	-teth}', "");
Expect(0, 1820, '\p{^Is_Joining_Group=	-teth}', "");
Expect(0, 1820, '\P{Is_Joining_Group=	-teth}', "");
Expect(1, 1820, '\P{^Is_Joining_Group=	-teth}', "");
Expect(0, 1821, '\p{Is_Joining_Group=	-teth}', "");
Expect(1, 1821, '\p{^Is_Joining_Group=	-teth}', "");
Expect(1, 1821, '\P{Is_Joining_Group=	-teth}', "");
Expect(0, 1821, '\P{^Is_Joining_Group=	-teth}', "");
Error('\p{Is_Jg=	/a/Teth}');
Error('\P{Is_Jg=	/a/Teth}');
Expect(1, 1820, '\p{Is_Jg=teth}', "");
Expect(0, 1820, '\p{^Is_Jg=teth}', "");
Expect(0, 1820, '\P{Is_Jg=teth}', "");
Expect(1, 1820, '\P{^Is_Jg=teth}', "");
Expect(0, 1821, '\p{Is_Jg=teth}', "");
Expect(1, 1821, '\p{^Is_Jg=teth}', "");
Expect(1, 1821, '\P{Is_Jg=teth}', "");
Expect(0, 1821, '\P{^Is_Jg=teth}', "");
Expect(1, 1820, '\p{Is_Jg=-	Teth}', "");
Expect(0, 1820, '\p{^Is_Jg=-	Teth}', "");
Expect(0, 1820, '\P{Is_Jg=-	Teth}', "");
Expect(1, 1820, '\P{^Is_Jg=-	Teth}', "");
Expect(0, 1821, '\p{Is_Jg=-	Teth}', "");
Expect(1, 1821, '\p{^Is_Jg=-	Teth}', "");
Expect(1, 1821, '\P{Is_Jg=-	Teth}', "");
Expect(0, 1821, '\P{^Is_Jg=-	Teth}', "");
Error('\p{Joining_Group= 	Waw/a/}');
Error('\P{Joining_Group= 	Waw/a/}');
Expect(1, 2219, '\p{Joining_Group=waw}', "");
Expect(0, 2219, '\p{^Joining_Group=waw}', "");
Expect(0, 2219, '\P{Joining_Group=waw}', "");
Expect(1, 2219, '\P{^Joining_Group=waw}', "");
Expect(0, 2220, '\p{Joining_Group=waw}', "");
Expect(1, 2220, '\p{^Joining_Group=waw}', "");
Expect(1, 2220, '\P{Joining_Group=waw}', "");
Expect(0, 2220, '\P{^Joining_Group=waw}', "");
Expect(1, 2219, '\p{Joining_Group=_	waw}', "");
Expect(0, 2219, '\p{^Joining_Group=_	waw}', "");
Expect(0, 2219, '\P{Joining_Group=_	waw}', "");
Expect(1, 2219, '\P{^Joining_Group=_	waw}', "");
Expect(0, 2220, '\p{Joining_Group=_	waw}', "");
Expect(1, 2220, '\p{^Joining_Group=_	waw}', "");
Expect(1, 2220, '\P{Joining_Group=_	waw}', "");
Expect(0, 2220, '\P{^Joining_Group=_	waw}', "");
Error('\p{Jg=	 WAW/a/}');
Error('\P{Jg=	 WAW/a/}');
Expect(1, 2219, '\p{Jg: waw}', "");
Expect(0, 2219, '\p{^Jg: waw}', "");
Expect(0, 2219, '\P{Jg: waw}', "");
Expect(1, 2219, '\P{^Jg: waw}', "");
Expect(0, 2220, '\p{Jg: waw}', "");
Expect(1, 2220, '\p{^Jg: waw}', "");
Expect(1, 2220, '\P{Jg: waw}', "");
Expect(0, 2220, '\P{^Jg: waw}', "");
Expect(1, 2219, '\p{Jg=-_Waw}', "");
Expect(0, 2219, '\p{^Jg=-_Waw}', "");
Expect(0, 2219, '\P{Jg=-_Waw}', "");
Expect(1, 2219, '\P{^Jg=-_Waw}', "");
Expect(0, 2220, '\p{Jg=-_Waw}', "");
Expect(1, 2220, '\p{^Jg=-_Waw}', "");
Expect(1, 2220, '\P{Jg=-_Waw}', "");
Expect(0, 2220, '\P{^Jg=-_Waw}', "");
Error('\p{Is_Joining_Group=:=_-Waw}');
Error('\P{Is_Joining_Group=:=_-Waw}');
Expect(1, 2219, '\p{Is_Joining_Group=waw}', "");
Expect(0, 2219, '\p{^Is_Joining_Group=waw}', "");
Expect(0, 2219, '\P{Is_Joining_Group=waw}', "");
Expect(1, 2219, '\P{^Is_Joining_Group=waw}', "");
Expect(0, 2220, '\p{Is_Joining_Group=waw}', "");
Expect(1, 2220, '\p{^Is_Joining_Group=waw}', "");
Expect(1, 2220, '\P{Is_Joining_Group=waw}', "");
Expect(0, 2220, '\P{^Is_Joining_Group=waw}', "");
Expect(1, 2219, '\p{Is_Joining_Group=  Waw}', "");
Expect(0, 2219, '\p{^Is_Joining_Group=  Waw}', "");
Expect(0, 2219, '\P{Is_Joining_Group=  Waw}', "");
Expect(1, 2219, '\P{^Is_Joining_Group=  Waw}', "");
Expect(0, 2220, '\p{Is_Joining_Group=  Waw}', "");
Expect(1, 2220, '\p{^Is_Joining_Group=  Waw}', "");
Expect(1, 2220, '\P{Is_Joining_Group=  Waw}', "");
Expect(0, 2220, '\P{^Is_Joining_Group=  Waw}', "");
Error('\p{Is_Jg= -Waw/a/}');
Error('\P{Is_Jg= -Waw/a/}');
Expect(1, 2219, '\p{Is_Jg=waw}', "");
Expect(0, 2219, '\p{^Is_Jg=waw}', "");
Expect(0, 2219, '\P{Is_Jg=waw}', "");
Expect(1, 2219, '\P{^Is_Jg=waw}', "");
Expect(0, 2220, '\p{Is_Jg=waw}', "");
Expect(1, 2220, '\p{^Is_Jg=waw}', "");
Expect(1, 2220, '\P{Is_Jg=waw}', "");
Expect(0, 2220, '\P{^Is_Jg=waw}', "");
Expect(1, 2219, '\p{Is_Jg=	 Waw}', "");
Expect(0, 2219, '\p{^Is_Jg=	 Waw}', "");
Expect(0, 2219, '\P{Is_Jg=	 Waw}', "");
Expect(1, 2219, '\P{^Is_Jg=	 Waw}', "");
Expect(0, 2220, '\p{Is_Jg=	 Waw}', "");
Expect(1, 2220, '\p{^Is_Jg=	 Waw}', "");
Expect(1, 2220, '\P{Is_Jg=	 Waw}', "");
Expect(0, 2220, '\P{^Is_Jg=	 Waw}', "");
Error('\p{Joining_Group=_-Yeh/a/}');
Error('\P{Joining_Group=_-Yeh/a/}');
Expect(1, 2217, '\p{Joining_Group=yeh}', "");
Expect(0, 2217, '\p{^Joining_Group=yeh}', "");
Expect(0, 2217, '\P{Joining_Group=yeh}', "");
Expect(1, 2217, '\P{^Joining_Group=yeh}', "");
Expect(0, 2218, '\p{Joining_Group=yeh}', "");
Expect(1, 2218, '\p{^Joining_Group=yeh}', "");
Expect(1, 2218, '\P{Joining_Group=yeh}', "");
Expect(0, 2218, '\P{^Joining_Group=yeh}', "");
Expect(1, 2217, '\p{Joining_Group=-_yeh}', "");
Expect(0, 2217, '\p{^Joining_Group=-_yeh}', "");
Expect(0, 2217, '\P{Joining_Group=-_yeh}', "");
Expect(1, 2217, '\P{^Joining_Group=-_yeh}', "");
Expect(0, 2218, '\p{Joining_Group=-_yeh}', "");
Expect(1, 2218, '\p{^Joining_Group=-_yeh}', "");
Expect(1, 2218, '\P{Joining_Group=-_yeh}', "");
Expect(0, 2218, '\P{^Joining_Group=-_yeh}', "");
Error('\p{Jg:	_	Yeh:=}');
Error('\P{Jg:	_	Yeh:=}');
Expect(1, 2217, '\p{Jg=yeh}', "");
Expect(0, 2217, '\p{^Jg=yeh}', "");
Expect(0, 2217, '\P{Jg=yeh}', "");
Expect(1, 2217, '\P{^Jg=yeh}', "");
Expect(0, 2218, '\p{Jg=yeh}', "");
Expect(1, 2218, '\p{^Jg=yeh}', "");
Expect(1, 2218, '\P{Jg=yeh}', "");
Expect(0, 2218, '\P{^Jg=yeh}', "");
Expect(1, 2217, '\p{Jg=-Yeh}', "");
Expect(0, 2217, '\p{^Jg=-Yeh}', "");
Expect(0, 2217, '\P{Jg=-Yeh}', "");
Expect(1, 2217, '\P{^Jg=-Yeh}', "");
Expect(0, 2218, '\p{Jg=-Yeh}', "");
Expect(1, 2218, '\p{^Jg=-Yeh}', "");
Expect(1, 2218, '\P{Jg=-Yeh}', "");
Expect(0, 2218, '\P{^Jg=-Yeh}', "");
Error('\p{Is_Joining_Group=:= -Yeh}');
Error('\P{Is_Joining_Group=:= -Yeh}');
Expect(1, 2217, '\p{Is_Joining_Group=yeh}', "");
Expect(0, 2217, '\p{^Is_Joining_Group=yeh}', "");
Expect(0, 2217, '\P{Is_Joining_Group=yeh}', "");
Expect(1, 2217, '\P{^Is_Joining_Group=yeh}', "");
Expect(0, 2218, '\p{Is_Joining_Group=yeh}', "");
Expect(1, 2218, '\p{^Is_Joining_Group=yeh}', "");
Expect(1, 2218, '\P{Is_Joining_Group=yeh}', "");
Expect(0, 2218, '\P{^Is_Joining_Group=yeh}', "");
Expect(1, 2217, '\p{Is_Joining_Group= 	Yeh}', "");
Expect(0, 2217, '\p{^Is_Joining_Group= 	Yeh}', "");
Expect(0, 2217, '\P{Is_Joining_Group= 	Yeh}', "");
Expect(1, 2217, '\P{^Is_Joining_Group= 	Yeh}', "");
Expect(0, 2218, '\p{Is_Joining_Group= 	Yeh}', "");
Expect(1, 2218, '\p{^Is_Joining_Group= 	Yeh}', "");
Expect(1, 2218, '\P{Is_Joining_Group= 	Yeh}', "");
Expect(0, 2218, '\P{^Is_Joining_Group= 	Yeh}', "");
Error('\p{Is_Jg= _YEH:=}');
Error('\P{Is_Jg= _YEH:=}');
Expect(1, 2217, '\p{Is_Jg=yeh}', "");
Expect(0, 2217, '\p{^Is_Jg=yeh}', "");
Expect(0, 2217, '\P{Is_Jg=yeh}', "");
Expect(1, 2217, '\P{^Is_Jg=yeh}', "");
Expect(0, 2218, '\p{Is_Jg=yeh}', "");
Expect(1, 2218, '\p{^Is_Jg=yeh}', "");
Expect(1, 2218, '\P{Is_Jg=yeh}', "");
Expect(0, 2218, '\P{^Is_Jg=yeh}', "");
Expect(1, 2217, '\p{Is_Jg=	YEH}', "");
Expect(0, 2217, '\p{^Is_Jg=	YEH}', "");
Expect(0, 2217, '\P{Is_Jg=	YEH}', "");
Expect(1, 2217, '\P{^Is_Jg=	YEH}', "");
Expect(0, 2218, '\p{Is_Jg=	YEH}', "");
Expect(1, 2218, '\p{^Is_Jg=	YEH}', "");
Expect(1, 2218, '\P{Is_Jg=	YEH}', "");
Expect(0, 2218, '\P{^Is_Jg=	YEH}', "");
Error('\p{Joining_Group=_Yeh_barree:=}');
Error('\P{Joining_Group=_Yeh_barree:=}');
Expect(1, 1747, '\p{Joining_Group=yehbarree}', "");
Expect(0, 1747, '\p{^Joining_Group=yehbarree}', "");
Expect(0, 1747, '\P{Joining_Group=yehbarree}', "");
Expect(1, 1747, '\P{^Joining_Group=yehbarree}', "");
Expect(0, 1748, '\p{Joining_Group=yehbarree}', "");
Expect(1, 1748, '\p{^Joining_Group=yehbarree}', "");
Expect(1, 1748, '\P{Joining_Group=yehbarree}', "");
Expect(0, 1748, '\P{^Joining_Group=yehbarree}', "");
Expect(1, 1747, '\p{Joining_Group=		YEH_BARREE}', "");
Expect(0, 1747, '\p{^Joining_Group=		YEH_BARREE}', "");
Expect(0, 1747, '\P{Joining_Group=		YEH_BARREE}', "");
Expect(1, 1747, '\P{^Joining_Group=		YEH_BARREE}', "");
Expect(0, 1748, '\p{Joining_Group=		YEH_BARREE}', "");
Expect(1, 1748, '\p{^Joining_Group=		YEH_BARREE}', "");
Expect(1, 1748, '\P{Joining_Group=		YEH_BARREE}', "");
Expect(0, 1748, '\P{^Joining_Group=		YEH_BARREE}', "");
Error('\p{Jg:   _-Yeh_Barree/a/}');
Error('\P{Jg:   _-Yeh_Barree/a/}');
Expect(1, 1747, '\p{Jg=yehbarree}', "");
Expect(0, 1747, '\p{^Jg=yehbarree}', "");
Expect(0, 1747, '\P{Jg=yehbarree}', "");
Expect(1, 1747, '\P{^Jg=yehbarree}', "");
Expect(0, 1748, '\p{Jg=yehbarree}', "");
Expect(1, 1748, '\p{^Jg=yehbarree}', "");
Expect(1, 1748, '\P{Jg=yehbarree}', "");
Expect(0, 1748, '\P{^Jg=yehbarree}', "");
Expect(1, 1747, '\p{Jg= -Yeh_Barree}', "");
Expect(0, 1747, '\p{^Jg= -Yeh_Barree}', "");
Expect(0, 1747, '\P{Jg= -Yeh_Barree}', "");
Expect(1, 1747, '\P{^Jg= -Yeh_Barree}', "");
Expect(0, 1748, '\p{Jg= -Yeh_Barree}', "");
Expect(1, 1748, '\p{^Jg= -Yeh_Barree}', "");
Expect(1, 1748, '\P{Jg= -Yeh_Barree}', "");
Expect(0, 1748, '\P{^Jg= -Yeh_Barree}', "");
Error('\p{Is_Joining_Group:   	/a/Yeh_Barree}');
Error('\P{Is_Joining_Group:   	/a/Yeh_Barree}');
Expect(1, 1747, '\p{Is_Joining_Group=yehbarree}', "");
Expect(0, 1747, '\p{^Is_Joining_Group=yehbarree}', "");
Expect(0, 1747, '\P{Is_Joining_Group=yehbarree}', "");
Expect(1, 1747, '\P{^Is_Joining_Group=yehbarree}', "");
Expect(0, 1748, '\p{Is_Joining_Group=yehbarree}', "");
Expect(1, 1748, '\p{^Is_Joining_Group=yehbarree}', "");
Expect(1, 1748, '\P{Is_Joining_Group=yehbarree}', "");
Expect(0, 1748, '\P{^Is_Joining_Group=yehbarree}', "");
Expect(1, 1747, '\p{Is_Joining_Group=		yeh_BARREE}', "");
Expect(0, 1747, '\p{^Is_Joining_Group=		yeh_BARREE}', "");
Expect(0, 1747, '\P{Is_Joining_Group=		yeh_BARREE}', "");
Expect(1, 1747, '\P{^Is_Joining_Group=		yeh_BARREE}', "");
Expect(0, 1748, '\p{Is_Joining_Group=		yeh_BARREE}', "");
Expect(1, 1748, '\p{^Is_Joining_Group=		yeh_BARREE}', "");
Expect(1, 1748, '\P{Is_Joining_Group=		yeh_BARREE}', "");
Expect(0, 1748, '\P{^Is_Joining_Group=		yeh_BARREE}', "");
Error('\p{Is_Jg=_YEH_Barree:=}');
Error('\P{Is_Jg=_YEH_Barree:=}');
Expect(1, 1747, '\p{Is_Jg=yehbarree}', "");
Expect(0, 1747, '\p{^Is_Jg=yehbarree}', "");
Expect(0, 1747, '\P{Is_Jg=yehbarree}', "");
Expect(1, 1747, '\P{^Is_Jg=yehbarree}', "");
Expect(0, 1748, '\p{Is_Jg=yehbarree}', "");
Expect(1, 1748, '\p{^Is_Jg=yehbarree}', "");
Expect(1, 1748, '\P{Is_Jg=yehbarree}', "");
Expect(0, 1748, '\P{^Is_Jg=yehbarree}', "");
Expect(1, 1747, '\p{Is_Jg=_Yeh_barree}', "");
Expect(0, 1747, '\p{^Is_Jg=_Yeh_barree}', "");
Expect(0, 1747, '\P{Is_Jg=_Yeh_barree}', "");
Expect(1, 1747, '\P{^Is_Jg=_Yeh_barree}', "");
Expect(0, 1748, '\p{Is_Jg=_Yeh_barree}', "");
Expect(1, 1748, '\p{^Is_Jg=_Yeh_barree}', "");
Expect(1, 1748, '\P{Is_Jg=_Yeh_barree}', "");
Expect(0, 1748, '\P{^Is_Jg=_Yeh_barree}', "");
Error('\p{Joining_Group::=_Yeh_With_Tail}');
Error('\P{Joining_Group::=_Yeh_With_Tail}');
Expect(1, 1741, '\p{Joining_Group:yehwithtail}', "");
Expect(0, 1741, '\p{^Joining_Group:yehwithtail}', "");
Expect(0, 1741, '\P{Joining_Group:yehwithtail}', "");
Expect(1, 1741, '\P{^Joining_Group:yehwithtail}', "");
Expect(0, 1742, '\p{Joining_Group:yehwithtail}', "");
Expect(1, 1742, '\p{^Joining_Group:yehwithtail}', "");
Expect(1, 1742, '\P{Joining_Group:yehwithtail}', "");
Expect(0, 1742, '\P{^Joining_Group:yehwithtail}', "");
Expect(1, 1741, '\p{Joining_Group= Yeh_With_tail}', "");
Expect(0, 1741, '\p{^Joining_Group= Yeh_With_tail}', "");
Expect(0, 1741, '\P{Joining_Group= Yeh_With_tail}', "");
Expect(1, 1741, '\P{^Joining_Group= Yeh_With_tail}', "");
Expect(0, 1742, '\p{Joining_Group= Yeh_With_tail}', "");
Expect(1, 1742, '\p{^Joining_Group= Yeh_With_tail}', "");
Expect(1, 1742, '\P{Joining_Group= Yeh_With_tail}', "");
Expect(0, 1742, '\P{^Joining_Group= Yeh_With_tail}', "");
Error('\p{Jg=_:=YEH_With_Tail}');
Error('\P{Jg=_:=YEH_With_Tail}');
Expect(1, 1741, '\p{Jg=yehwithtail}', "");
Expect(0, 1741, '\p{^Jg=yehwithtail}', "");
Expect(0, 1741, '\P{Jg=yehwithtail}', "");
Expect(1, 1741, '\P{^Jg=yehwithtail}', "");
Expect(0, 1742, '\p{Jg=yehwithtail}', "");
Expect(1, 1742, '\p{^Jg=yehwithtail}', "");
Expect(1, 1742, '\P{Jg=yehwithtail}', "");
Expect(0, 1742, '\P{^Jg=yehwithtail}', "");
Expect(1, 1741, '\p{Jg=--Yeh_WITH_TAIL}', "");
Expect(0, 1741, '\p{^Jg=--Yeh_WITH_TAIL}', "");
Expect(0, 1741, '\P{Jg=--Yeh_WITH_TAIL}', "");
Expect(1, 1741, '\P{^Jg=--Yeh_WITH_TAIL}', "");
Expect(0, 1742, '\p{Jg=--Yeh_WITH_TAIL}', "");
Expect(1, 1742, '\p{^Jg=--Yeh_WITH_TAIL}', "");
Expect(1, 1742, '\P{Jg=--Yeh_WITH_TAIL}', "");
Expect(0, 1742, '\P{^Jg=--Yeh_WITH_TAIL}', "");
Error('\p{Is_Joining_Group=-YEH_With_TAIL/a/}');
Error('\P{Is_Joining_Group=-YEH_With_TAIL/a/}');
Expect(1, 1741, '\p{Is_Joining_Group=yehwithtail}', "");
Expect(0, 1741, '\p{^Is_Joining_Group=yehwithtail}', "");
Expect(0, 1741, '\P{Is_Joining_Group=yehwithtail}', "");
Expect(1, 1741, '\P{^Is_Joining_Group=yehwithtail}', "");
Expect(0, 1742, '\p{Is_Joining_Group=yehwithtail}', "");
Expect(1, 1742, '\p{^Is_Joining_Group=yehwithtail}', "");
Expect(1, 1742, '\P{Is_Joining_Group=yehwithtail}', "");
Expect(0, 1742, '\P{^Is_Joining_Group=yehwithtail}', "");
Expect(1, 1741, '\p{Is_Joining_Group=	-yeh_With_Tail}', "");
Expect(0, 1741, '\p{^Is_Joining_Group=	-yeh_With_Tail}', "");
Expect(0, 1741, '\P{Is_Joining_Group=	-yeh_With_Tail}', "");
Expect(1, 1741, '\P{^Is_Joining_Group=	-yeh_With_Tail}', "");
Expect(0, 1742, '\p{Is_Joining_Group=	-yeh_With_Tail}', "");
Expect(1, 1742, '\p{^Is_Joining_Group=	-yeh_With_Tail}', "");
Expect(1, 1742, '\P{Is_Joining_Group=	-yeh_With_Tail}', "");
Expect(0, 1742, '\P{^Is_Joining_Group=	-yeh_With_Tail}', "");
Error('\p{Is_Jg=/a/ -YEH_With_TAIL}');
Error('\P{Is_Jg=/a/ -YEH_With_TAIL}');
Expect(1, 1741, '\p{Is_Jg=yehwithtail}', "");
Expect(0, 1741, '\p{^Is_Jg=yehwithtail}', "");
Expect(0, 1741, '\P{Is_Jg=yehwithtail}', "");
Expect(1, 1741, '\P{^Is_Jg=yehwithtail}', "");
Expect(0, 1742, '\p{Is_Jg=yehwithtail}', "");
Expect(1, 1742, '\p{^Is_Jg=yehwithtail}', "");
Expect(1, 1742, '\P{Is_Jg=yehwithtail}', "");
Expect(0, 1742, '\P{^Is_Jg=yehwithtail}', "");
Expect(1, 1741, '\p{Is_Jg=	yeh_WITH_TAIL}', "");
Expect(0, 1741, '\p{^Is_Jg=	yeh_WITH_TAIL}', "");
Expect(0, 1741, '\P{Is_Jg=	yeh_WITH_TAIL}', "");
Expect(1, 1741, '\P{^Is_Jg=	yeh_WITH_TAIL}', "");
Expect(0, 1742, '\p{Is_Jg=	yeh_WITH_TAIL}', "");
Expect(1, 1742, '\p{^Is_Jg=	yeh_WITH_TAIL}', "");
Expect(1, 1742, '\P{Is_Jg=	yeh_WITH_TAIL}', "");
Expect(0, 1742, '\P{^Is_Jg=	yeh_WITH_TAIL}', "");
Error('\p{Joining_Group:	 YUDH/a/}');
Error('\P{Joining_Group:	 YUDH/a/}');
Expect(1, 1821, '\p{Joining_Group=yudh}', "");
Expect(0, 1821, '\p{^Joining_Group=yudh}', "");
Expect(0, 1821, '\P{Joining_Group=yudh}', "");
Expect(1, 1821, '\P{^Joining_Group=yudh}', "");
Expect(0, 1822, '\p{Joining_Group=yudh}', "");
Expect(1, 1822, '\p{^Joining_Group=yudh}', "");
Expect(1, 1822, '\P{Joining_Group=yudh}', "");
Expect(0, 1822, '\P{^Joining_Group=yudh}', "");
Expect(1, 1821, '\p{Joining_Group= YUDH}', "");
Expect(0, 1821, '\p{^Joining_Group= YUDH}', "");
Expect(0, 1821, '\P{Joining_Group= YUDH}', "");
Expect(1, 1821, '\P{^Joining_Group= YUDH}', "");
Expect(0, 1822, '\p{Joining_Group= YUDH}', "");
Expect(1, 1822, '\p{^Joining_Group= YUDH}', "");
Expect(1, 1822, '\P{Joining_Group= YUDH}', "");
Expect(0, 1822, '\P{^Joining_Group= YUDH}', "");
Error('\p{Jg=	:=Yudh}');
Error('\P{Jg=	:=Yudh}');
Expect(1, 1821, '\p{Jg=yudh}', "");
Expect(0, 1821, '\p{^Jg=yudh}', "");
Expect(0, 1821, '\P{Jg=yudh}', "");
Expect(1, 1821, '\P{^Jg=yudh}', "");
Expect(0, 1822, '\p{Jg=yudh}', "");
Expect(1, 1822, '\p{^Jg=yudh}', "");
Expect(1, 1822, '\P{Jg=yudh}', "");
Expect(0, 1822, '\P{^Jg=yudh}', "");
Expect(1, 1821, '\p{Jg= YUDH}', "");
Expect(0, 1821, '\p{^Jg= YUDH}', "");
Expect(0, 1821, '\P{Jg= YUDH}', "");
Expect(1, 1821, '\P{^Jg= YUDH}', "");
Expect(0, 1822, '\p{Jg= YUDH}', "");
Expect(1, 1822, '\p{^Jg= YUDH}', "");
Expect(1, 1822, '\P{Jg= YUDH}', "");
Expect(0, 1822, '\P{^Jg= YUDH}', "");
Error('\p{Is_Joining_Group=/a/	-Yudh}');
Error('\P{Is_Joining_Group=/a/	-Yudh}');
Expect(1, 1821, '\p{Is_Joining_Group=yudh}', "");
Expect(0, 1821, '\p{^Is_Joining_Group=yudh}', "");
Expect(0, 1821, '\P{Is_Joining_Group=yudh}', "");
Expect(1, 1821, '\P{^Is_Joining_Group=yudh}', "");
Expect(0, 1822, '\p{Is_Joining_Group=yudh}', "");
Expect(1, 1822, '\p{^Is_Joining_Group=yudh}', "");
Expect(1, 1822, '\P{Is_Joining_Group=yudh}', "");
Expect(0, 1822, '\P{^Is_Joining_Group=yudh}', "");
Expect(1, 1821, '\p{Is_Joining_Group=_YUDH}', "");
Expect(0, 1821, '\p{^Is_Joining_Group=_YUDH}', "");
Expect(0, 1821, '\P{Is_Joining_Group=_YUDH}', "");
Expect(1, 1821, '\P{^Is_Joining_Group=_YUDH}', "");
Expect(0, 1822, '\p{Is_Joining_Group=_YUDH}', "");
Expect(1, 1822, '\p{^Is_Joining_Group=_YUDH}', "");
Expect(1, 1822, '\P{Is_Joining_Group=_YUDH}', "");
Expect(0, 1822, '\P{^Is_Joining_Group=_YUDH}', "");
Error('\p{Is_Jg= /a/Yudh}');
Error('\P{Is_Jg= /a/Yudh}');
Expect(1, 1821, '\p{Is_Jg=yudh}', "");
Expect(0, 1821, '\p{^Is_Jg=yudh}', "");
Expect(0, 1821, '\P{Is_Jg=yudh}', "");
Expect(1, 1821, '\P{^Is_Jg=yudh}', "");
Expect(0, 1822, '\p{Is_Jg=yudh}', "");
Expect(1, 1822, '\p{^Is_Jg=yudh}', "");
Expect(1, 1822, '\P{Is_Jg=yudh}', "");
Expect(0, 1822, '\P{^Is_Jg=yudh}', "");
Expect(1, 1821, '\p{Is_Jg= 	Yudh}', "");
Expect(0, 1821, '\p{^Is_Jg= 	Yudh}', "");
Expect(0, 1821, '\P{Is_Jg= 	Yudh}', "");
Expect(1, 1821, '\P{^Is_Jg= 	Yudh}', "");
Expect(0, 1822, '\p{Is_Jg= 	Yudh}', "");
Expect(1, 1822, '\p{^Is_Jg= 	Yudh}', "");
Expect(1, 1822, '\P{Is_Jg= 	Yudh}', "");
Expect(0, 1822, '\P{^Is_Jg= 	Yudh}', "");
Error('\p{Joining_Group=	:=Yudh_he}');
Error('\P{Joining_Group=	:=Yudh_he}');
Expect(1, 1822, '\p{Joining_Group=yudhhe}', "");
Expect(0, 1822, '\p{^Joining_Group=yudhhe}', "");
Expect(0, 1822, '\P{Joining_Group=yudhhe}', "");
Expect(1, 1822, '\P{^Joining_Group=yudhhe}', "");
Expect(0, 1823, '\p{Joining_Group=yudhhe}', "");
Expect(1, 1823, '\p{^Joining_Group=yudhhe}', "");
Expect(1, 1823, '\P{Joining_Group=yudhhe}', "");
Expect(0, 1823, '\P{^Joining_Group=yudhhe}', "");
Expect(1, 1822, '\p{Joining_Group=	_YUDH_HE}', "");
Expect(0, 1822, '\p{^Joining_Group=	_YUDH_HE}', "");
Expect(0, 1822, '\P{Joining_Group=	_YUDH_HE}', "");
Expect(1, 1822, '\P{^Joining_Group=	_YUDH_HE}', "");
Expect(0, 1823, '\p{Joining_Group=	_YUDH_HE}', "");
Expect(1, 1823, '\p{^Joining_Group=	_YUDH_HE}', "");
Expect(1, 1823, '\P{Joining_Group=	_YUDH_HE}', "");
Expect(0, 1823, '\P{^Joining_Group=	_YUDH_HE}', "");
Error('\p{Jg::=--Yudh_he}');
Error('\P{Jg::=--Yudh_he}');
Expect(1, 1822, '\p{Jg=yudhhe}', "");
Expect(0, 1822, '\p{^Jg=yudhhe}', "");
Expect(0, 1822, '\P{Jg=yudhhe}', "");
Expect(1, 1822, '\P{^Jg=yudhhe}', "");
Expect(0, 1823, '\p{Jg=yudhhe}', "");
Expect(1, 1823, '\p{^Jg=yudhhe}', "");
Expect(1, 1823, '\P{Jg=yudhhe}', "");
Expect(0, 1823, '\P{^Jg=yudhhe}', "");
Expect(1, 1822, '\p{Jg=-Yudh_HE}', "");
Expect(0, 1822, '\p{^Jg=-Yudh_HE}', "");
Expect(0, 1822, '\P{Jg=-Yudh_HE}', "");
Expect(1, 1822, '\P{^Jg=-Yudh_HE}', "");
Expect(0, 1823, '\p{Jg=-Yudh_HE}', "");
Expect(1, 1823, '\p{^Jg=-Yudh_HE}', "");
Expect(1, 1823, '\P{Jg=-Yudh_HE}', "");
Expect(0, 1823, '\P{^Jg=-Yudh_HE}', "");
Error('\p{Is_Joining_Group=  Yudh_He/a/}');
Error('\P{Is_Joining_Group=  Yudh_He/a/}');
Expect(1, 1822, '\p{Is_Joining_Group=yudhhe}', "");
Expect(0, 1822, '\p{^Is_Joining_Group=yudhhe}', "");
Expect(0, 1822, '\P{Is_Joining_Group=yudhhe}', "");
Expect(1, 1822, '\P{^Is_Joining_Group=yudhhe}', "");
Expect(0, 1823, '\p{Is_Joining_Group=yudhhe}', "");
Expect(1, 1823, '\p{^Is_Joining_Group=yudhhe}', "");
Expect(1, 1823, '\P{Is_Joining_Group=yudhhe}', "");
Expect(0, 1823, '\P{^Is_Joining_Group=yudhhe}', "");
Expect(1, 1822, '\p{Is_Joining_Group=		Yudh_HE}', "");
Expect(0, 1822, '\p{^Is_Joining_Group=		Yudh_HE}', "");
Expect(0, 1822, '\P{Is_Joining_Group=		Yudh_HE}', "");
Expect(1, 1822, '\P{^Is_Joining_Group=		Yudh_HE}', "");
Expect(0, 1823, '\p{Is_Joining_Group=		Yudh_HE}', "");
Expect(1, 1823, '\p{^Is_Joining_Group=		Yudh_HE}', "");
Expect(1, 1823, '\P{Is_Joining_Group=		Yudh_HE}', "");
Expect(0, 1823, '\P{^Is_Joining_Group=		Yudh_HE}', "");
Error('\p{Is_Jg=:=Yudh_He}');
Error('\P{Is_Jg=:=Yudh_He}');
Expect(1, 1822, '\p{Is_Jg=yudhhe}', "");
Expect(0, 1822, '\p{^Is_Jg=yudhhe}', "");
Expect(0, 1822, '\P{Is_Jg=yudhhe}', "");
Expect(1, 1822, '\P{^Is_Jg=yudhhe}', "");
Expect(0, 1823, '\p{Is_Jg=yudhhe}', "");
Expect(1, 1823, '\p{^Is_Jg=yudhhe}', "");
Expect(1, 1823, '\P{Is_Jg=yudhhe}', "");
Expect(0, 1823, '\P{^Is_Jg=yudhhe}', "");
Expect(1, 1822, '\p{Is_Jg=_ Yudh_he}', "");
Expect(0, 1822, '\p{^Is_Jg=_ Yudh_he}', "");
Expect(0, 1822, '\P{Is_Jg=_ Yudh_he}', "");
Expect(1, 1822, '\P{^Is_Jg=_ Yudh_he}', "");
Expect(0, 1823, '\p{Is_Jg=_ Yudh_he}', "");
Expect(1, 1823, '\p{^Is_Jg=_ Yudh_he}', "");
Expect(1, 1823, '\P{Is_Jg=_ Yudh_he}', "");
Expect(0, 1823, '\P{^Is_Jg=_ Yudh_he}', "");
Error('\p{Joining_Group=-_Zain/a/}');
Error('\P{Joining_Group=-_Zain/a/}');
Expect(1, 1817, '\p{Joining_Group=zain}', "");
Expect(0, 1817, '\p{^Joining_Group=zain}', "");
Expect(0, 1817, '\P{Joining_Group=zain}', "");
Expect(1, 1817, '\P{^Joining_Group=zain}', "");
Expect(0, 1818, '\p{Joining_Group=zain}', "");
Expect(1, 1818, '\p{^Joining_Group=zain}', "");
Expect(1, 1818, '\P{Joining_Group=zain}', "");
Expect(0, 1818, '\P{^Joining_Group=zain}', "");
Expect(1, 1817, '\p{Joining_Group= _Zain}', "");
Expect(0, 1817, '\p{^Joining_Group= _Zain}', "");
Expect(0, 1817, '\P{Joining_Group= _Zain}', "");
Expect(1, 1817, '\P{^Joining_Group= _Zain}', "");
Expect(0, 1818, '\p{Joining_Group= _Zain}', "");
Expect(1, 1818, '\p{^Joining_Group= _Zain}', "");
Expect(1, 1818, '\P{Joining_Group= _Zain}', "");
Expect(0, 1818, '\P{^Joining_Group= _Zain}', "");
Error('\p{Jg=/a/	_ZAIN}');
Error('\P{Jg=/a/	_ZAIN}');
Expect(1, 1817, '\p{Jg=zain}', "");
Expect(0, 1817, '\p{^Jg=zain}', "");
Expect(0, 1817, '\P{Jg=zain}', "");
Expect(1, 1817, '\P{^Jg=zain}', "");
Expect(0, 1818, '\p{Jg=zain}', "");
Expect(1, 1818, '\p{^Jg=zain}', "");
Expect(1, 1818, '\P{Jg=zain}', "");
Expect(0, 1818, '\P{^Jg=zain}', "");
Expect(1, 1817, '\p{Jg=  ZAIN}', "");
Expect(0, 1817, '\p{^Jg=  ZAIN}', "");
Expect(0, 1817, '\P{Jg=  ZAIN}', "");
Expect(1, 1817, '\P{^Jg=  ZAIN}', "");
Expect(0, 1818, '\p{Jg=  ZAIN}', "");
Expect(1, 1818, '\p{^Jg=  ZAIN}', "");
Expect(1, 1818, '\P{Jg=  ZAIN}', "");
Expect(0, 1818, '\P{^Jg=  ZAIN}', "");
Error('\p{Is_Joining_Group=	-Zain/a/}');
Error('\P{Is_Joining_Group=	-Zain/a/}');
Expect(1, 1817, '\p{Is_Joining_Group=zain}', "");
Expect(0, 1817, '\p{^Is_Joining_Group=zain}', "");
Expect(0, 1817, '\P{Is_Joining_Group=zain}', "");
Expect(1, 1817, '\P{^Is_Joining_Group=zain}', "");
Expect(0, 1818, '\p{Is_Joining_Group=zain}', "");
Expect(1, 1818, '\p{^Is_Joining_Group=zain}', "");
Expect(1, 1818, '\P{Is_Joining_Group=zain}', "");
Expect(0, 1818, '\P{^Is_Joining_Group=zain}', "");
Expect(1, 1817, '\p{Is_Joining_Group=-	ZAIN}', "");
Expect(0, 1817, '\p{^Is_Joining_Group=-	ZAIN}', "");
Expect(0, 1817, '\P{Is_Joining_Group=-	ZAIN}', "");
Expect(1, 1817, '\P{^Is_Joining_Group=-	ZAIN}', "");
Expect(0, 1818, '\p{Is_Joining_Group=-	ZAIN}', "");
Expect(1, 1818, '\p{^Is_Joining_Group=-	ZAIN}', "");
Expect(1, 1818, '\P{Is_Joining_Group=-	ZAIN}', "");
Expect(0, 1818, '\P{^Is_Joining_Group=-	ZAIN}', "");
Error('\p{Is_Jg=:=Zain}');
Error('\P{Is_Jg=:=Zain}');
Expect(1, 1817, '\p{Is_Jg=zain}', "");
Expect(0, 1817, '\p{^Is_Jg=zain}', "");
Expect(0, 1817, '\P{Is_Jg=zain}', "");
Expect(1, 1817, '\P{^Is_Jg=zain}', "");
Expect(0, 1818, '\p{Is_Jg=zain}', "");
Expect(1, 1818, '\p{^Is_Jg=zain}', "");
Expect(1, 1818, '\P{Is_Jg=zain}', "");
Expect(0, 1818, '\P{^Is_Jg=zain}', "");
Expect(1, 1817, '\p{Is_Jg= Zain}', "");
Expect(0, 1817, '\p{^Is_Jg= Zain}', "");
Expect(0, 1817, '\P{Is_Jg= Zain}', "");
Expect(1, 1817, '\P{^Is_Jg= Zain}', "");
Expect(0, 1818, '\p{Is_Jg= Zain}', "");
Expect(1, 1818, '\p{^Is_Jg= Zain}', "");
Expect(1, 1818, '\P{Is_Jg= Zain}', "");
Expect(0, 1818, '\P{^Is_Jg= Zain}', "");
Error('\p{Joining_Group:   /a/	Zhain}');
Error('\P{Joining_Group:   /a/	Zhain}');
Expect(1, 1869, '\p{Joining_Group:   zhain}', "");
Expect(0, 1869, '\p{^Joining_Group:   zhain}', "");
Expect(0, 1869, '\P{Joining_Group:   zhain}', "");
Expect(1, 1869, '\P{^Joining_Group:   zhain}', "");
Expect(0, 1870, '\p{Joining_Group:   zhain}', "");
Expect(1, 1870, '\p{^Joining_Group:   zhain}', "");
Expect(1, 1870, '\P{Joining_Group:   zhain}', "");
Expect(0, 1870, '\P{^Joining_Group:   zhain}', "");
Expect(1, 1869, '\p{Joining_Group=__zhain}', "");
Expect(0, 1869, '\p{^Joining_Group=__zhain}', "");
Expect(0, 1869, '\P{Joining_Group=__zhain}', "");
Expect(1, 1869, '\P{^Joining_Group=__zhain}', "");
Expect(0, 1870, '\p{Joining_Group=__zhain}', "");
Expect(1, 1870, '\p{^Joining_Group=__zhain}', "");
Expect(1, 1870, '\P{Joining_Group=__zhain}', "");
Expect(0, 1870, '\P{^Joining_Group=__zhain}', "");
Error('\p{Jg=/a/	-zhain}');
Error('\P{Jg=/a/	-zhain}');
Expect(1, 1869, '\p{Jg=zhain}', "");
Expect(0, 1869, '\p{^Jg=zhain}', "");
Expect(0, 1869, '\P{Jg=zhain}', "");
Expect(1, 1869, '\P{^Jg=zhain}', "");
Expect(0, 1870, '\p{Jg=zhain}', "");
Expect(1, 1870, '\p{^Jg=zhain}', "");
Expect(1, 1870, '\P{Jg=zhain}', "");
Expect(0, 1870, '\P{^Jg=zhain}', "");
Expect(1, 1869, '\p{Jg=	ZHAIN}', "");
Expect(0, 1869, '\p{^Jg=	ZHAIN}', "");
Expect(0, 1869, '\P{Jg=	ZHAIN}', "");
Expect(1, 1869, '\P{^Jg=	ZHAIN}', "");
Expect(0, 1870, '\p{Jg=	ZHAIN}', "");
Expect(1, 1870, '\p{^Jg=	ZHAIN}', "");
Expect(1, 1870, '\P{Jg=	ZHAIN}', "");
Expect(0, 1870, '\P{^Jg=	ZHAIN}', "");
Error('\p{Is_Joining_Group=	Zhain/a/}');
Error('\P{Is_Joining_Group=	Zhain/a/}');
Expect(1, 1869, '\p{Is_Joining_Group=zhain}', "");
Expect(0, 1869, '\p{^Is_Joining_Group=zhain}', "");
Expect(0, 1869, '\P{Is_Joining_Group=zhain}', "");
Expect(1, 1869, '\P{^Is_Joining_Group=zhain}', "");
Expect(0, 1870, '\p{Is_Joining_Group=zhain}', "");
Expect(1, 1870, '\p{^Is_Joining_Group=zhain}', "");
Expect(1, 1870, '\P{Is_Joining_Group=zhain}', "");
Expect(0, 1870, '\P{^Is_Joining_Group=zhain}', "");
Expect(1, 1869, '\p{Is_Joining_Group=	_Zhain}', "");
Expect(0, 1869, '\p{^Is_Joining_Group=	_Zhain}', "");
Expect(0, 1869, '\P{Is_Joining_Group=	_Zhain}', "");
Expect(1, 1869, '\P{^Is_Joining_Group=	_Zhain}', "");
Expect(0, 1870, '\p{Is_Joining_Group=	_Zhain}', "");
Expect(1, 1870, '\p{^Is_Joining_Group=	_Zhain}', "");
Expect(1, 1870, '\P{Is_Joining_Group=	_Zhain}', "");
Expect(0, 1870, '\P{^Is_Joining_Group=	_Zhain}', "");
Error('\p{Is_Jg=__ZHAIN:=}');
Error('\P{Is_Jg=__ZHAIN:=}');
Expect(1, 1869, '\p{Is_Jg=zhain}', "");
Expect(0, 1869, '\p{^Is_Jg=zhain}', "");
Expect(0, 1869, '\P{Is_Jg=zhain}', "");
Expect(1, 1869, '\P{^Is_Jg=zhain}', "");
Expect(0, 1870, '\p{Is_Jg=zhain}', "");
Expect(1, 1870, '\p{^Is_Jg=zhain}', "");
Expect(1, 1870, '\P{Is_Jg=zhain}', "");
Expect(0, 1870, '\P{^Is_Jg=zhain}', "");
Expect(1, 1869, '\p{Is_Jg=__zhain}', "");
Expect(0, 1869, '\p{^Is_Jg=__zhain}', "");
Expect(0, 1869, '\P{Is_Jg=__zhain}', "");
Expect(1, 1869, '\P{^Is_Jg=__zhain}', "");
Expect(0, 1870, '\p{Is_Jg=__zhain}', "");
Expect(1, 1870, '\p{^Is_Jg=__zhain}', "");
Expect(1, 1870, '\P{Is_Jg=__zhain}', "");
Expect(0, 1870, '\P{^Is_Jg=__zhain}', "");
Error('\p{Joining_Type= join_CAUSING/a/}');
Error('\P{Joining_Type= join_CAUSING/a/}');
Expect(1, 8205, '\p{Joining_Type=joincausing}', "");
Expect(0, 8205, '\p{^Joining_Type=joincausing}', "");
Expect(0, 8205, '\P{Joining_Type=joincausing}', "");
Expect(1, 8205, '\P{^Joining_Type=joincausing}', "");
Expect(0, 8206, '\p{Joining_Type=joincausing}', "");
Expect(1, 8206, '\p{^Joining_Type=joincausing}', "");
Expect(1, 8206, '\P{Joining_Type=joincausing}', "");
Expect(0, 8206, '\P{^Joining_Type=joincausing}', "");
Expect(1, 8205, '\p{Joining_Type=-Join_CAUSING}', "");
Expect(0, 8205, '\p{^Joining_Type=-Join_CAUSING}', "");
Expect(0, 8205, '\P{Joining_Type=-Join_CAUSING}', "");
Expect(1, 8205, '\P{^Joining_Type=-Join_CAUSING}', "");
Expect(0, 8206, '\p{Joining_Type=-Join_CAUSING}', "");
Expect(1, 8206, '\p{^Joining_Type=-Join_CAUSING}', "");
Expect(1, 8206, '\P{Joining_Type=-Join_CAUSING}', "");
Expect(0, 8206, '\P{^Joining_Type=-Join_CAUSING}', "");
Error('\p{Jt= /a/C}');
Error('\P{Jt= /a/C}');
Expect(1, 8205, '\p{Jt=c}', "");
Expect(0, 8205, '\p{^Jt=c}', "");
Expect(0, 8205, '\P{Jt=c}', "");
Expect(1, 8205, '\P{^Jt=c}', "");
Expect(0, 8206, '\p{Jt=c}', "");
Expect(1, 8206, '\p{^Jt=c}', "");
Expect(1, 8206, '\P{Jt=c}', "");
Expect(0, 8206, '\P{^Jt=c}', "");
Expect(1, 8205, '\p{Jt: 	C}', "");
Expect(0, 8205, '\p{^Jt: 	C}', "");
Expect(0, 8205, '\P{Jt: 	C}', "");
Expect(1, 8205, '\P{^Jt: 	C}', "");
Expect(0, 8206, '\p{Jt: 	C}', "");
Expect(1, 8206, '\p{^Jt: 	C}', "");
Expect(1, 8206, '\P{Jt: 	C}', "");
Expect(0, 8206, '\P{^Jt: 	C}', "");
Error('\p{Is_Joining_Type=:=JOIN_Causing}');
Error('\P{Is_Joining_Type=:=JOIN_Causing}');
Expect(1, 8205, '\p{Is_Joining_Type=joincausing}', "");
Expect(0, 8205, '\p{^Is_Joining_Type=joincausing}', "");
Expect(0, 8205, '\P{Is_Joining_Type=joincausing}', "");
Expect(1, 8205, '\P{^Is_Joining_Type=joincausing}', "");
Expect(0, 8206, '\p{Is_Joining_Type=joincausing}', "");
Expect(1, 8206, '\p{^Is_Joining_Type=joincausing}', "");
Expect(1, 8206, '\P{Is_Joining_Type=joincausing}', "");
Expect(0, 8206, '\P{^Is_Joining_Type=joincausing}', "");
Expect(1, 8205, '\p{Is_Joining_Type= JOIN_Causing}', "");
Expect(0, 8205, '\p{^Is_Joining_Type= JOIN_Causing}', "");
Expect(0, 8205, '\P{Is_Joining_Type= JOIN_Causing}', "");
Expect(1, 8205, '\P{^Is_Joining_Type= JOIN_Causing}', "");
Expect(0, 8206, '\p{Is_Joining_Type= JOIN_Causing}', "");
Expect(1, 8206, '\p{^Is_Joining_Type= JOIN_Causing}', "");
Expect(1, 8206, '\P{Is_Joining_Type= JOIN_Causing}', "");
Expect(0, 8206, '\P{^Is_Joining_Type= JOIN_Causing}', "");
Error('\p{Is_Jt=:=	-C}');
Error('\P{Is_Jt=:=	-C}');
Expect(1, 8205, '\p{Is_Jt=c}', "");
Expect(0, 8205, '\p{^Is_Jt=c}', "");
Expect(0, 8205, '\P{Is_Jt=c}', "");
Expect(1, 8205, '\P{^Is_Jt=c}', "");
Expect(0, 8206, '\p{Is_Jt=c}', "");
Expect(1, 8206, '\p{^Is_Jt=c}', "");
Expect(1, 8206, '\P{Is_Jt=c}', "");
Expect(0, 8206, '\P{^Is_Jt=c}', "");
Expect(1, 8205, '\p{Is_Jt=_ C}', "");
Expect(0, 8205, '\p{^Is_Jt=_ C}', "");
Expect(0, 8205, '\P{Is_Jt=_ C}', "");
Expect(1, 8205, '\P{^Is_Jt=_ C}', "");
Expect(0, 8206, '\p{Is_Jt=_ C}', "");
Expect(1, 8206, '\p{^Is_Jt=_ C}', "");
Expect(1, 8206, '\P{Is_Jt=_ C}', "");
Expect(0, 8206, '\P{^Is_Jt=_ C}', "");
Error('\p{Joining_Type=/a/ Dual_JOINING}');
Error('\P{Joining_Type=/a/ Dual_JOINING}');
Expect(1, 43121, '\p{Joining_Type=dualjoining}', "");
Expect(0, 43121, '\p{^Joining_Type=dualjoining}', "");
Expect(0, 43121, '\P{Joining_Type=dualjoining}', "");
Expect(1, 43121, '\P{^Joining_Type=dualjoining}', "");
Expect(0, 43122, '\p{Joining_Type=dualjoining}', "");
Expect(1, 43122, '\p{^Joining_Type=dualjoining}', "");
Expect(1, 43122, '\P{Joining_Type=dualjoining}', "");
Expect(0, 43122, '\P{^Joining_Type=dualjoining}', "");
Expect(1, 43121, '\p{Joining_Type=	Dual_JOINING}', "");
Expect(0, 43121, '\p{^Joining_Type=	Dual_JOINING}', "");
Expect(0, 43121, '\P{Joining_Type=	Dual_JOINING}', "");
Expect(1, 43121, '\P{^Joining_Type=	Dual_JOINING}', "");
Expect(0, 43122, '\p{Joining_Type=	Dual_JOINING}', "");
Expect(1, 43122, '\p{^Joining_Type=	Dual_JOINING}', "");
Expect(1, 43122, '\P{Joining_Type=	Dual_JOINING}', "");
Expect(0, 43122, '\P{^Joining_Type=	Dual_JOINING}', "");
Error('\p{Jt= /a/d}');
Error('\P{Jt= /a/d}');
Expect(1, 43121, '\p{Jt=d}', "");
Expect(0, 43121, '\p{^Jt=d}', "");
Expect(0, 43121, '\P{Jt=d}', "");
Expect(1, 43121, '\P{^Jt=d}', "");
Expect(0, 43122, '\p{Jt=d}', "");
Expect(1, 43122, '\p{^Jt=d}', "");
Expect(1, 43122, '\P{Jt=d}', "");
Expect(0, 43122, '\P{^Jt=d}', "");
Expect(1, 43121, '\p{Jt=_-D}', "");
Expect(0, 43121, '\p{^Jt=_-D}', "");
Expect(0, 43121, '\P{Jt=_-D}', "");
Expect(1, 43121, '\P{^Jt=_-D}', "");
Expect(0, 43122, '\p{Jt=_-D}', "");
Expect(1, 43122, '\p{^Jt=_-D}', "");
Expect(1, 43122, '\P{Jt=_-D}', "");
Expect(0, 43122, '\P{^Jt=_-D}', "");
Error('\p{Is_Joining_Type=-	Dual_JOINING/a/}');
Error('\P{Is_Joining_Type=-	Dual_JOINING/a/}');
Expect(1, 43121, '\p{Is_Joining_Type=dualjoining}', "");
Expect(0, 43121, '\p{^Is_Joining_Type=dualjoining}', "");
Expect(0, 43121, '\P{Is_Joining_Type=dualjoining}', "");
Expect(1, 43121, '\P{^Is_Joining_Type=dualjoining}', "");
Expect(0, 43122, '\p{Is_Joining_Type=dualjoining}', "");
Expect(1, 43122, '\p{^Is_Joining_Type=dualjoining}', "");
Expect(1, 43122, '\P{Is_Joining_Type=dualjoining}', "");
Expect(0, 43122, '\P{^Is_Joining_Type=dualjoining}', "");
Expect(1, 43121, '\p{Is_Joining_Type=	Dual_JOINING}', "");
Expect(0, 43121, '\p{^Is_Joining_Type=	Dual_JOINING}', "");
Expect(0, 43121, '\P{Is_Joining_Type=	Dual_JOINING}', "");
Expect(1, 43121, '\P{^Is_Joining_Type=	Dual_JOINING}', "");
Expect(0, 43122, '\p{Is_Joining_Type=	Dual_JOINING}', "");
Expect(1, 43122, '\p{^Is_Joining_Type=	Dual_JOINING}', "");
Expect(1, 43122, '\P{Is_Joining_Type=	Dual_JOINING}', "");
Expect(0, 43122, '\P{^Is_Joining_Type=	Dual_JOINING}', "");
Error('\p{Is_Jt=__D:=}');
Error('\P{Is_Jt=__D:=}');
Expect(1, 43121, '\p{Is_Jt=d}', "");
Expect(0, 43121, '\p{^Is_Jt=d}', "");
Expect(0, 43121, '\P{Is_Jt=d}', "");
Expect(1, 43121, '\P{^Is_Jt=d}', "");
Expect(0, 43122, '\p{Is_Jt=d}', "");
Expect(1, 43122, '\p{^Is_Jt=d}', "");
Expect(1, 43122, '\P{Is_Jt=d}', "");
Expect(0, 43122, '\P{^Is_Jt=d}', "");
Expect(1, 43121, '\p{Is_Jt= d}', "");
Expect(0, 43121, '\p{^Is_Jt= d}', "");
Expect(0, 43121, '\P{Is_Jt= d}', "");
Expect(1, 43121, '\P{^Is_Jt= d}', "");
Expect(0, 43122, '\p{Is_Jt= d}', "");
Expect(1, 43122, '\p{^Is_Jt= d}', "");
Expect(1, 43122, '\P{Is_Jt= d}', "");
Expect(0, 43122, '\P{^Is_Jt= d}', "");
Error('\p{Joining_Type:   __left_JOINING:=}');
Error('\P{Joining_Type:   __left_JOINING:=}');
Expect(1, 43122, '\p{Joining_Type=leftjoining}', "");
Expect(0, 43122, '\p{^Joining_Type=leftjoining}', "");
Expect(0, 43122, '\P{Joining_Type=leftjoining}', "");
Expect(1, 43122, '\P{^Joining_Type=leftjoining}', "");
Expect(0, 43123, '\p{Joining_Type=leftjoining}', "");
Expect(1, 43123, '\p{^Joining_Type=leftjoining}', "");
Expect(1, 43123, '\P{Joining_Type=leftjoining}', "");
Expect(0, 43123, '\P{^Joining_Type=leftjoining}', "");
Expect(1, 43122, '\p{Joining_Type= Left_Joining}', "");
Expect(0, 43122, '\p{^Joining_Type= Left_Joining}', "");
Expect(0, 43122, '\P{Joining_Type= Left_Joining}', "");
Expect(1, 43122, '\P{^Joining_Type= Left_Joining}', "");
Expect(0, 43123, '\p{Joining_Type= Left_Joining}', "");
Expect(1, 43123, '\p{^Joining_Type= Left_Joining}', "");
Expect(1, 43123, '\P{Joining_Type= Left_Joining}', "");
Expect(0, 43123, '\P{^Joining_Type= Left_Joining}', "");
Error('\p{Jt=/a/l}');
Error('\P{Jt=/a/l}');
Expect(1, 43122, '\p{Jt=l}', "");
Expect(0, 43122, '\p{^Jt=l}', "");
Expect(0, 43122, '\P{Jt=l}', "");
Expect(1, 43122, '\P{^Jt=l}', "");
Expect(0, 43123, '\p{Jt=l}', "");
Expect(1, 43123, '\p{^Jt=l}', "");
Expect(1, 43123, '\P{Jt=l}', "");
Expect(0, 43123, '\P{^Jt=l}', "");
Error('\p{Is_Joining_Type=left_JOINING:=}');
Error('\P{Is_Joining_Type=left_JOINING:=}');
Expect(1, 43122, '\p{Is_Joining_Type=leftjoining}', "");
Expect(0, 43122, '\p{^Is_Joining_Type=leftjoining}', "");
Expect(0, 43122, '\P{Is_Joining_Type=leftjoining}', "");
Expect(1, 43122, '\P{^Is_Joining_Type=leftjoining}', "");
Expect(0, 43123, '\p{Is_Joining_Type=leftjoining}', "");
Expect(1, 43123, '\p{^Is_Joining_Type=leftjoining}', "");
Expect(1, 43123, '\P{Is_Joining_Type=leftjoining}', "");
Expect(0, 43123, '\P{^Is_Joining_Type=leftjoining}', "");
Expect(1, 43122, '\p{Is_Joining_Type=__Left_joining}', "");
Expect(0, 43122, '\p{^Is_Joining_Type=__Left_joining}', "");
Expect(0, 43122, '\P{Is_Joining_Type=__Left_joining}', "");
Expect(1, 43122, '\P{^Is_Joining_Type=__Left_joining}', "");
Expect(0, 43123, '\p{Is_Joining_Type=__Left_joining}', "");
Expect(1, 43123, '\p{^Is_Joining_Type=__Left_joining}', "");
Expect(1, 43123, '\P{Is_Joining_Type=__Left_joining}', "");
Expect(0, 43123, '\P{^Is_Joining_Type=__Left_joining}', "");
Error('\p{Is_Jt=/a/	L}');
Error('\P{Is_Jt=/a/	L}');
Expect(1, 43122, '\p{Is_Jt:	l}', "");
Expect(0, 43122, '\p{^Is_Jt:	l}', "");
Expect(0, 43122, '\P{Is_Jt:	l}', "");
Expect(1, 43122, '\P{^Is_Jt:	l}', "");
Expect(0, 43123, '\p{Is_Jt:	l}', "");
Expect(1, 43123, '\p{^Is_Jt:	l}', "");
Expect(1, 43123, '\P{Is_Jt:	l}', "");
Expect(0, 43123, '\P{^Is_Jt:	l}', "");
Expect(1, 43122, '\p{Is_Jt=-L}', "");
Expect(0, 43122, '\p{^Is_Jt=-L}', "");
Expect(0, 43122, '\P{Is_Jt=-L}', "");
Expect(1, 43122, '\P{^Is_Jt=-L}', "");
Expect(0, 43123, '\p{Is_Jt=-L}', "");
Expect(1, 43123, '\p{^Is_Jt=-L}', "");
Expect(1, 43123, '\P{Is_Jt=-L}', "");
Expect(0, 43123, '\P{^Is_Jt=-L}', "");
Error('\p{Joining_Type=_	RIGHT_Joining/a/}');
Error('\P{Joining_Type=_	RIGHT_Joining/a/}');
Expect(1, 2220, '\p{Joining_Type=rightjoining}', "");
Expect(0, 2220, '\p{^Joining_Type=rightjoining}', "");
Expect(0, 2220, '\P{Joining_Type=rightjoining}', "");
Expect(1, 2220, '\P{^Joining_Type=rightjoining}', "");
Expect(0, 2221, '\p{Joining_Type=rightjoining}', "");
Expect(1, 2221, '\p{^Joining_Type=rightjoining}', "");
Expect(1, 2221, '\P{Joining_Type=rightjoining}', "");
Expect(0, 2221, '\P{^Joining_Type=rightjoining}', "");
Expect(1, 2220, '\p{Joining_Type=	RIGHT_joining}', "");
Expect(0, 2220, '\p{^Joining_Type=	RIGHT_joining}', "");
Expect(0, 2220, '\P{Joining_Type=	RIGHT_joining}', "");
Expect(1, 2220, '\P{^Joining_Type=	RIGHT_joining}', "");
Expect(0, 2221, '\p{Joining_Type=	RIGHT_joining}', "");
Expect(1, 2221, '\p{^Joining_Type=	RIGHT_joining}', "");
Expect(1, 2221, '\P{Joining_Type=	RIGHT_joining}', "");
Expect(0, 2221, '\P{^Joining_Type=	RIGHT_joining}', "");
Error('\p{Jt=:=r}');
Error('\P{Jt=:=r}');
Expect(1, 2220, '\p{Jt=r}', "");
Expect(0, 2220, '\p{^Jt=r}', "");
Expect(0, 2220, '\P{Jt=r}', "");
Expect(1, 2220, '\P{^Jt=r}', "");
Expect(0, 2221, '\p{Jt=r}', "");
Expect(1, 2221, '\p{^Jt=r}', "");
Expect(1, 2221, '\P{Jt=r}', "");
Expect(0, 2221, '\P{^Jt=r}', "");
Expect(1, 2220, '\p{Jt:		R}', "");
Expect(0, 2220, '\p{^Jt:		R}', "");
Expect(0, 2220, '\P{Jt:		R}', "");
Expect(1, 2220, '\P{^Jt:		R}', "");
Expect(0, 2221, '\p{Jt:		R}', "");
Expect(1, 2221, '\p{^Jt:		R}', "");
Expect(1, 2221, '\P{Jt:		R}', "");
Expect(0, 2221, '\P{^Jt:		R}', "");
Error('\p{Is_Joining_Type=:=	Right_JOINING}');
Error('\P{Is_Joining_Type=:=	Right_JOINING}');
Expect(1, 2220, '\p{Is_Joining_Type=rightjoining}', "");
Expect(0, 2220, '\p{^Is_Joining_Type=rightjoining}', "");
Expect(0, 2220, '\P{Is_Joining_Type=rightjoining}', "");
Expect(1, 2220, '\P{^Is_Joining_Type=rightjoining}', "");
Expect(0, 2221, '\p{Is_Joining_Type=rightjoining}', "");
Expect(1, 2221, '\p{^Is_Joining_Type=rightjoining}', "");
Expect(1, 2221, '\P{Is_Joining_Type=rightjoining}', "");
Expect(0, 2221, '\P{^Is_Joining_Type=rightjoining}', "");
Expect(1, 2220, '\p{Is_Joining_Type= 	Right_JOINING}', "");
Expect(0, 2220, '\p{^Is_Joining_Type= 	Right_JOINING}', "");
Expect(0, 2220, '\P{Is_Joining_Type= 	Right_JOINING}', "");
Expect(1, 2220, '\P{^Is_Joining_Type= 	Right_JOINING}', "");
Expect(0, 2221, '\p{Is_Joining_Type= 	Right_JOINING}', "");
Expect(1, 2221, '\p{^Is_Joining_Type= 	Right_JOINING}', "");
Expect(1, 2221, '\P{Is_Joining_Type= 	Right_JOINING}', "");
Expect(0, 2221, '\P{^Is_Joining_Type= 	Right_JOINING}', "");
Error('\p{Is_Jt= :=R}');
Error('\P{Is_Jt= :=R}');
Expect(1, 2220, '\p{Is_Jt=r}', "");
Expect(0, 2220, '\p{^Is_Jt=r}', "");
Expect(0, 2220, '\P{Is_Jt=r}', "");
Expect(1, 2220, '\P{^Is_Jt=r}', "");
Expect(0, 2221, '\p{Is_Jt=r}', "");
Expect(1, 2221, '\p{^Is_Jt=r}', "");
Expect(1, 2221, '\P{Is_Jt=r}', "");
Expect(0, 2221, '\P{^Is_Jt=r}', "");
Expect(1, 2220, '\p{Is_Jt=-	R}', "");
Expect(0, 2220, '\p{^Is_Jt=-	R}', "");
Expect(0, 2220, '\P{Is_Jt=-	R}', "");
Expect(1, 2220, '\P{^Is_Jt=-	R}', "");
Expect(0, 2221, '\p{Is_Jt=-	R}', "");
Expect(1, 2221, '\p{^Is_Jt=-	R}', "");
Expect(1, 2221, '\P{Is_Jt=-	R}', "");
Expect(0, 2221, '\P{^Is_Jt=-	R}', "");
Error('\p{Joining_Type=/a/	Transparent}');
Error('\P{Joining_Type=/a/	Transparent}');
Expect(1, 917999, '\p{Joining_Type: transparent}', "");
Expect(0, 917999, '\p{^Joining_Type: transparent}', "");
Expect(0, 917999, '\P{Joining_Type: transparent}', "");
Expect(1, 917999, '\P{^Joining_Type: transparent}', "");
Expect(0, 918000, '\p{Joining_Type: transparent}', "");
Expect(1, 918000, '\p{^Joining_Type: transparent}', "");
Expect(1, 918000, '\P{Joining_Type: transparent}', "");
Expect(0, 918000, '\P{^Joining_Type: transparent}', "");
Expect(1, 917999, '\p{Joining_Type=-TRANSPARENT}', "");
Expect(0, 917999, '\p{^Joining_Type=-TRANSPARENT}', "");
Expect(0, 917999, '\P{Joining_Type=-TRANSPARENT}', "");
Expect(1, 917999, '\P{^Joining_Type=-TRANSPARENT}', "");
Expect(0, 918000, '\p{Joining_Type=-TRANSPARENT}', "");
Expect(1, 918000, '\p{^Joining_Type=-TRANSPARENT}', "");
Expect(1, 918000, '\P{Joining_Type=-TRANSPARENT}', "");
Expect(0, 918000, '\P{^Joining_Type=-TRANSPARENT}', "");
Error('\p{Jt=-/a/T}');
Error('\P{Jt=-/a/T}');
Expect(1, 917999, '\p{Jt:t}', "");
Expect(0, 917999, '\p{^Jt:t}', "");
Expect(0, 917999, '\P{Jt:t}', "");
Expect(1, 917999, '\P{^Jt:t}', "");
Expect(0, 918000, '\p{Jt:t}', "");
Expect(1, 918000, '\p{^Jt:t}', "");
Expect(1, 918000, '\P{Jt:t}', "");
Expect(0, 918000, '\P{^Jt:t}', "");
Expect(1, 917999, '\p{Jt=T}', "");
Expect(0, 917999, '\p{^Jt=T}', "");
Expect(0, 917999, '\P{Jt=T}', "");
Expect(1, 917999, '\P{^Jt=T}', "");
Expect(0, 918000, '\p{Jt=T}', "");
Expect(1, 918000, '\p{^Jt=T}', "");
Expect(1, 918000, '\P{Jt=T}', "");
Expect(0, 918000, '\P{^Jt=T}', "");
Error('\p{Is_Joining_Type=-:=Transparent}');
Error('\P{Is_Joining_Type=-:=Transparent}');
Expect(1, 917999, '\p{Is_Joining_Type=transparent}', "");
Expect(0, 917999, '\p{^Is_Joining_Type=transparent}', "");
Expect(0, 917999, '\P{Is_Joining_Type=transparent}', "");
Expect(1, 917999, '\P{^Is_Joining_Type=transparent}', "");
Expect(0, 918000, '\p{Is_Joining_Type=transparent}', "");
Expect(1, 918000, '\p{^Is_Joining_Type=transparent}', "");
Expect(1, 918000, '\P{Is_Joining_Type=transparent}', "");
Expect(0, 918000, '\P{^Is_Joining_Type=transparent}', "");
Expect(1, 917999, '\p{Is_Joining_Type: _transparent}', "");
Expect(0, 917999, '\p{^Is_Joining_Type: _transparent}', "");
Expect(0, 917999, '\P{Is_Joining_Type: _transparent}', "");
Expect(1, 917999, '\P{^Is_Joining_Type: _transparent}', "");
Expect(0, 918000, '\p{Is_Joining_Type: _transparent}', "");
Expect(1, 918000, '\p{^Is_Joining_Type: _transparent}', "");
Expect(1, 918000, '\P{Is_Joining_Type: _transparent}', "");
Expect(0, 918000, '\P{^Is_Joining_Type: _transparent}', "");
Error('\p{Is_Jt=-_T:=}');
Error('\P{Is_Jt=-_T:=}');
Expect(1, 917999, '\p{Is_Jt=t}', "");
Expect(0, 917999, '\p{^Is_Jt=t}', "");
Expect(0, 917999, '\P{Is_Jt=t}', "");
Expect(1, 917999, '\P{^Is_Jt=t}', "");
Expect(0, 918000, '\p{Is_Jt=t}', "");
Expect(1, 918000, '\p{^Is_Jt=t}', "");
Expect(1, 918000, '\P{Is_Jt=t}', "");
Expect(0, 918000, '\P{^Is_Jt=t}', "");
Expect(1, 917999, '\p{Is_Jt=__T}', "");
Expect(0, 917999, '\p{^Is_Jt=__T}', "");
Expect(0, 917999, '\P{Is_Jt=__T}', "");
Expect(1, 917999, '\P{^Is_Jt=__T}', "");
Expect(0, 918000, '\p{Is_Jt=__T}', "");
Expect(1, 918000, '\p{^Is_Jt=__T}', "");
Expect(1, 918000, '\P{Is_Jt=__T}', "");
Expect(0, 918000, '\P{^Is_Jt=__T}', "");
Error('\p{Joining_Type=/a/-Non_JOINING}');
Error('\P{Joining_Type=/a/-Non_JOINING}');
Expect(1, 918000, '\p{Joining_Type=nonjoining}', "");
Expect(0, 918000, '\p{^Joining_Type=nonjoining}', "");
Expect(0, 918000, '\P{Joining_Type=nonjoining}', "");
Expect(1, 918000, '\P{^Joining_Type=nonjoining}', "");
Expect(0, 917999, '\p{Joining_Type=nonjoining}', "");
Expect(1, 917999, '\p{^Joining_Type=nonjoining}', "");
Expect(1, 917999, '\P{Joining_Type=nonjoining}', "");
Expect(0, 917999, '\P{^Joining_Type=nonjoining}', "");
Expect(1, 918000, '\p{Joining_Type=_	non_joining}', "");
Expect(0, 918000, '\p{^Joining_Type=_	non_joining}', "");
Expect(0, 918000, '\P{Joining_Type=_	non_joining}', "");
Expect(1, 918000, '\P{^Joining_Type=_	non_joining}', "");
Expect(0, 917999, '\p{Joining_Type=_	non_joining}', "");
Expect(1, 917999, '\p{^Joining_Type=_	non_joining}', "");
Expect(1, 917999, '\P{Joining_Type=_	non_joining}', "");
Expect(0, 917999, '\P{^Joining_Type=_	non_joining}', "");
Error('\p{Jt=	U:=}');
Error('\P{Jt=	U:=}');
Expect(1, 918000, '\p{Jt=u}', "");
Expect(0, 918000, '\p{^Jt=u}', "");
Expect(0, 918000, '\P{Jt=u}', "");
Expect(1, 918000, '\P{^Jt=u}', "");
Expect(0, 917999, '\p{Jt=u}', "");
Expect(1, 917999, '\p{^Jt=u}', "");
Expect(1, 917999, '\P{Jt=u}', "");
Expect(0, 917999, '\P{^Jt=u}', "");
Expect(1, 918000, '\p{Jt=	u}', "");
Expect(0, 918000, '\p{^Jt=	u}', "");
Expect(0, 918000, '\P{Jt=	u}', "");
Expect(1, 918000, '\P{^Jt=	u}', "");
Expect(0, 917999, '\p{Jt=	u}', "");
Expect(1, 917999, '\p{^Jt=	u}', "");
Expect(1, 917999, '\P{Jt=	u}', "");
Expect(0, 917999, '\P{^Jt=	u}', "");
Error('\p{Is_Joining_Type=/a/ 	NON_JOINING}');
Error('\P{Is_Joining_Type=/a/ 	NON_JOINING}');
Expect(1, 918000, '\p{Is_Joining_Type=nonjoining}', "");
Expect(0, 918000, '\p{^Is_Joining_Type=nonjoining}', "");
Expect(0, 918000, '\P{Is_Joining_Type=nonjoining}', "");
Expect(1, 918000, '\P{^Is_Joining_Type=nonjoining}', "");
Expect(0, 917999, '\p{Is_Joining_Type=nonjoining}', "");
Expect(1, 917999, '\p{^Is_Joining_Type=nonjoining}', "");
Expect(1, 917999, '\P{Is_Joining_Type=nonjoining}', "");
Expect(0, 917999, '\P{^Is_Joining_Type=nonjoining}', "");
Expect(1, 918000, '\p{Is_Joining_Type=		Non_joining}', "");
Expect(0, 918000, '\p{^Is_Joining_Type=		Non_joining}', "");
Expect(0, 918000, '\P{Is_Joining_Type=		Non_joining}', "");
Expect(1, 918000, '\P{^Is_Joining_Type=		Non_joining}', "");
Expect(0, 917999, '\p{Is_Joining_Type=		Non_joining}', "");
Expect(1, 917999, '\p{^Is_Joining_Type=		Non_joining}', "");
Expect(1, 917999, '\P{Is_Joining_Type=		Non_joining}', "");
Expect(0, 917999, '\P{^Is_Joining_Type=		Non_joining}', "");
Error('\p{Is_Jt=_/a/U}');
Error('\P{Is_Jt=_/a/U}');
Expect(1, 918000, '\p{Is_Jt=u}', "");
Expect(0, 918000, '\p{^Is_Jt=u}', "");
Expect(0, 918000, '\P{Is_Jt=u}', "");
Expect(1, 918000, '\P{^Is_Jt=u}', "");
Expect(0, 917999, '\p{Is_Jt=u}', "");
Expect(1, 917999, '\p{^Is_Jt=u}', "");
Expect(1, 917999, '\P{Is_Jt=u}', "");
Expect(0, 917999, '\P{^Is_Jt=u}', "");
Expect(1, 918000, '\p{Is_Jt=		U}', "");
Expect(0, 918000, '\p{^Is_Jt=		U}', "");
Expect(0, 918000, '\P{Is_Jt=		U}', "");
Expect(1, 918000, '\P{^Is_Jt=		U}', "");
Expect(0, 917999, '\p{Is_Jt=		U}', "");
Expect(1, 917999, '\p{^Is_Jt=		U}', "");
Expect(1, 917999, '\P{Is_Jt=		U}', "");
Expect(0, 917999, '\P{^Is_Jt=		U}', "");
Error('\p{Line_Break=_AMBIGUOUS/a/}');
Error('\P{Line_Break=_AMBIGUOUS/a/}');
Expect(1, 127386, '\p{Line_Break=ambiguous}', "");
Expect(0, 127386, '\p{^Line_Break=ambiguous}', "");
Expect(0, 127386, '\P{Line_Break=ambiguous}', "");
Expect(1, 127386, '\P{^Line_Break=ambiguous}', "");
Expect(0, 127387, '\p{Line_Break=ambiguous}', "");
Expect(1, 127387, '\p{^Line_Break=ambiguous}', "");
Expect(1, 127387, '\P{Line_Break=ambiguous}', "");
Expect(0, 127387, '\P{^Line_Break=ambiguous}', "");
Expect(1, 127386, '\p{Line_Break=__Ambiguous}', "");
Expect(0, 127386, '\p{^Line_Break=__Ambiguous}', "");
Expect(0, 127386, '\P{Line_Break=__Ambiguous}', "");
Expect(1, 127386, '\P{^Line_Break=__Ambiguous}', "");
Expect(0, 127387, '\p{Line_Break=__Ambiguous}', "");
Expect(1, 127387, '\p{^Line_Break=__Ambiguous}', "");
Expect(1, 127387, '\P{Line_Break=__Ambiguous}', "");
Expect(0, 127387, '\P{^Line_Break=__Ambiguous}', "");
Error('\p{Lb=/a/ _AI}');
Error('\P{Lb=/a/ _AI}');
Expect(1, 127386, '\p{Lb:ai}', "");
Expect(0, 127386, '\p{^Lb:ai}', "");
Expect(0, 127386, '\P{Lb:ai}', "");
Expect(1, 127386, '\P{^Lb:ai}', "");
Expect(0, 127387, '\p{Lb:ai}', "");
Expect(1, 127387, '\p{^Lb:ai}', "");
Expect(1, 127387, '\P{Lb:ai}', "");
Expect(0, 127387, '\P{^Lb:ai}', "");
Expect(1, 127386, '\p{Lb=AI}', "");
Expect(0, 127386, '\p{^Lb=AI}', "");
Expect(0, 127386, '\P{Lb=AI}', "");
Expect(1, 127386, '\P{^Lb=AI}', "");
Expect(0, 127387, '\p{Lb=AI}', "");
Expect(1, 127387, '\p{^Lb=AI}', "");
Expect(1, 127387, '\P{Lb=AI}', "");
Expect(0, 127387, '\P{^Lb=AI}', "");
Error('\p{Is_Line_Break=-/a/Ambiguous}');
Error('\P{Is_Line_Break=-/a/Ambiguous}');
Expect(1, 127386, '\p{Is_Line_Break=ambiguous}', "");
Expect(0, 127386, '\p{^Is_Line_Break=ambiguous}', "");
Expect(0, 127386, '\P{Is_Line_Break=ambiguous}', "");
Expect(1, 127386, '\P{^Is_Line_Break=ambiguous}', "");
Expect(0, 127387, '\p{Is_Line_Break=ambiguous}', "");
Expect(1, 127387, '\p{^Is_Line_Break=ambiguous}', "");
Expect(1, 127387, '\P{Is_Line_Break=ambiguous}', "");
Expect(0, 127387, '\P{^Is_Line_Break=ambiguous}', "");
Expect(1, 127386, '\p{Is_Line_Break=_-ambiguous}', "");
Expect(0, 127386, '\p{^Is_Line_Break=_-ambiguous}', "");
Expect(0, 127386, '\P{Is_Line_Break=_-ambiguous}', "");
Expect(1, 127386, '\P{^Is_Line_Break=_-ambiguous}', "");
Expect(0, 127387, '\p{Is_Line_Break=_-ambiguous}', "");
Expect(1, 127387, '\p{^Is_Line_Break=_-ambiguous}', "");
Expect(1, 127387, '\P{Is_Line_Break=_-ambiguous}', "");
Expect(0, 127387, '\P{^Is_Line_Break=_-ambiguous}', "");
Error('\p{Is_Lb=-AI:=}');
Error('\P{Is_Lb=-AI:=}');
Expect(1, 127386, '\p{Is_Lb=ai}', "");
Expect(0, 127386, '\p{^Is_Lb=ai}', "");
Expect(0, 127386, '\P{Is_Lb=ai}', "");
Expect(1, 127386, '\P{^Is_Lb=ai}', "");
Expect(0, 127387, '\p{Is_Lb=ai}', "");
Expect(1, 127387, '\p{^Is_Lb=ai}', "");
Expect(1, 127387, '\P{Is_Lb=ai}', "");
Expect(0, 127387, '\P{^Is_Lb=ai}', "");
Expect(1, 127386, '\p{Is_Lb=	-AI}', "");
Expect(0, 127386, '\p{^Is_Lb=	-AI}', "");
Expect(0, 127386, '\P{Is_Lb=	-AI}', "");
Expect(1, 127386, '\P{^Is_Lb=	-AI}', "");
Expect(0, 127387, '\p{Is_Lb=	-AI}', "");
Expect(1, 127387, '\p{^Is_Lb=	-AI}', "");
Expect(1, 127387, '\P{Is_Lb=	-AI}', "");
Expect(0, 127387, '\P{^Is_Lb=	-AI}', "");
Error('\p{Line_Break:    	alphabetic:=}');
Error('\P{Line_Break:    	alphabetic:=}');
Expect(1, 128883, '\p{Line_Break: alphabetic}', "");
Expect(0, 128883, '\p{^Line_Break: alphabetic}', "");
Expect(0, 128883, '\P{Line_Break: alphabetic}', "");
Expect(1, 128883, '\P{^Line_Break: alphabetic}', "");
Expect(0, 128884, '\p{Line_Break: alphabetic}', "");
Expect(1, 128884, '\p{^Line_Break: alphabetic}', "");
Expect(1, 128884, '\P{Line_Break: alphabetic}', "");
Expect(0, 128884, '\P{^Line_Break: alphabetic}', "");
Expect(1, 128883, '\p{Line_Break=		alphabetic}', "");
Expect(0, 128883, '\p{^Line_Break=		alphabetic}', "");
Expect(0, 128883, '\P{Line_Break=		alphabetic}', "");
Expect(1, 128883, '\P{^Line_Break=		alphabetic}', "");
Expect(0, 128884, '\p{Line_Break=		alphabetic}', "");
Expect(1, 128884, '\p{^Line_Break=		alphabetic}', "");
Expect(1, 128884, '\P{Line_Break=		alphabetic}', "");
Expect(0, 128884, '\P{^Line_Break=		alphabetic}', "");
Error('\p{Lb=_:=AL}');
Error('\P{Lb=_:=AL}');
Expect(1, 128883, '\p{Lb=al}', "");
Expect(0, 128883, '\p{^Lb=al}', "");
Expect(0, 128883, '\P{Lb=al}', "");
Expect(1, 128883, '\P{^Lb=al}', "");
Expect(0, 128884, '\p{Lb=al}', "");
Expect(1, 128884, '\p{^Lb=al}', "");
Expect(1, 128884, '\P{Lb=al}', "");
Expect(0, 128884, '\P{^Lb=al}', "");
Expect(1, 128883, '\p{Lb=-_AL}', "");
Expect(0, 128883, '\p{^Lb=-_AL}', "");
Expect(0, 128883, '\P{Lb=-_AL}', "");
Expect(1, 128883, '\P{^Lb=-_AL}', "");
Expect(0, 128884, '\p{Lb=-_AL}', "");
Expect(1, 128884, '\p{^Lb=-_AL}', "");
Expect(1, 128884, '\P{Lb=-_AL}', "");
Expect(0, 128884, '\P{^Lb=-_AL}', "");
Error('\p{Is_Line_Break=/a/alphabetic}');
Error('\P{Is_Line_Break=/a/alphabetic}');
Expect(1, 128883, '\p{Is_Line_Break=alphabetic}', "");
Expect(0, 128883, '\p{^Is_Line_Break=alphabetic}', "");
Expect(0, 128883, '\P{Is_Line_Break=alphabetic}', "");
Expect(1, 128883, '\P{^Is_Line_Break=alphabetic}', "");
Expect(0, 128884, '\p{Is_Line_Break=alphabetic}', "");
Expect(1, 128884, '\p{^Is_Line_Break=alphabetic}', "");
Expect(1, 128884, '\P{Is_Line_Break=alphabetic}', "");
Expect(0, 128884, '\P{^Is_Line_Break=alphabetic}', "");
Expect(1, 128883, '\p{Is_Line_Break=	_ALPHABETIC}', "");
Expect(0, 128883, '\p{^Is_Line_Break=	_ALPHABETIC}', "");
Expect(0, 128883, '\P{Is_Line_Break=	_ALPHABETIC}', "");
Expect(1, 128883, '\P{^Is_Line_Break=	_ALPHABETIC}', "");
Expect(0, 128884, '\p{Is_Line_Break=	_ALPHABETIC}', "");
Expect(1, 128884, '\p{^Is_Line_Break=	_ALPHABETIC}', "");
Expect(1, 128884, '\P{Is_Line_Break=	_ALPHABETIC}', "");
Expect(0, 128884, '\P{^Is_Line_Break=	_ALPHABETIC}', "");
Error('\p{Is_Lb=		AL:=}');
Error('\P{Is_Lb=		AL:=}');
Expect(1, 128883, '\p{Is_Lb=al}', "");
Expect(0, 128883, '\p{^Is_Lb=al}', "");
Expect(0, 128883, '\P{Is_Lb=al}', "");
Expect(1, 128883, '\P{^Is_Lb=al}', "");
Expect(0, 128884, '\p{Is_Lb=al}', "");
Expect(1, 128884, '\p{^Is_Lb=al}', "");
Expect(1, 128884, '\P{Is_Lb=al}', "");
Expect(0, 128884, '\P{^Is_Lb=al}', "");
Expect(1, 128883, '\p{Is_Lb= AL}', "");
Expect(0, 128883, '\p{^Is_Lb= AL}', "");
Expect(0, 128883, '\P{Is_Lb= AL}', "");
Expect(1, 128883, '\P{^Is_Lb= AL}', "");
Expect(0, 128884, '\p{Is_Lb= AL}', "");
Expect(1, 128884, '\p{^Is_Lb= AL}', "");
Expect(1, 128884, '\P{Is_Lb= AL}', "");
Expect(0, 128884, '\P{^Is_Lb= AL}', "");
Error('\p{Line_Break::=-_break_both}');
Error('\P{Line_Break::=-_break_both}');
Expect(1, 11835, '\p{Line_Break=breakboth}', "");
Expect(0, 11835, '\p{^Line_Break=breakboth}', "");
Expect(0, 11835, '\P{Line_Break=breakboth}', "");
Expect(1, 11835, '\P{^Line_Break=breakboth}', "");
Expect(0, 11836, '\p{Line_Break=breakboth}', "");
Expect(1, 11836, '\p{^Line_Break=breakboth}', "");
Expect(1, 11836, '\P{Line_Break=breakboth}', "");
Expect(0, 11836, '\P{^Line_Break=breakboth}', "");
Expect(1, 11835, '\p{Line_Break=	 Break_Both}', "");
Expect(0, 11835, '\p{^Line_Break=	 Break_Both}', "");
Expect(0, 11835, '\P{Line_Break=	 Break_Both}', "");
Expect(1, 11835, '\P{^Line_Break=	 Break_Both}', "");
Expect(0, 11836, '\p{Line_Break=	 Break_Both}', "");
Expect(1, 11836, '\p{^Line_Break=	 Break_Both}', "");
Expect(1, 11836, '\P{Line_Break=	 Break_Both}', "");
Expect(0, 11836, '\P{^Line_Break=	 Break_Both}', "");
Error('\p{Lb=	:=B2}');
Error('\P{Lb=	:=B2}');
Expect(1, 11835, '\p{Lb=b2}', "");
Expect(0, 11835, '\p{^Lb=b2}', "");
Expect(0, 11835, '\P{Lb=b2}', "");
Expect(1, 11835, '\P{^Lb=b2}', "");
Expect(0, 11836, '\p{Lb=b2}', "");
Expect(1, 11836, '\p{^Lb=b2}', "");
Expect(1, 11836, '\P{Lb=b2}', "");
Expect(0, 11836, '\P{^Lb=b2}', "");
Expect(1, 11835, '\p{Lb=	 B2}', "");
Expect(0, 11835, '\p{^Lb=	 B2}', "");
Expect(0, 11835, '\P{Lb=	 B2}', "");
Expect(1, 11835, '\P{^Lb=	 B2}', "");
Expect(0, 11836, '\p{Lb=	 B2}', "");
Expect(1, 11836, '\p{^Lb=	 B2}', "");
Expect(1, 11836, '\P{Lb=	 B2}', "");
Expect(0, 11836, '\P{^Lb=	 B2}', "");
Error('\p{Is_Line_Break=/a/	BREAK_Both}');
Error('\P{Is_Line_Break=/a/	BREAK_Both}');
Expect(1, 11835, '\p{Is_Line_Break=breakboth}', "");
Expect(0, 11835, '\p{^Is_Line_Break=breakboth}', "");
Expect(0, 11835, '\P{Is_Line_Break=breakboth}', "");
Expect(1, 11835, '\P{^Is_Line_Break=breakboth}', "");
Expect(0, 11836, '\p{Is_Line_Break=breakboth}', "");
Expect(1, 11836, '\p{^Is_Line_Break=breakboth}', "");
Expect(1, 11836, '\P{Is_Line_Break=breakboth}', "");
Expect(0, 11836, '\P{^Is_Line_Break=breakboth}', "");
Expect(1, 11835, '\p{Is_Line_Break:  Break_BOTH}', "");
Expect(0, 11835, '\p{^Is_Line_Break:  Break_BOTH}', "");
Expect(0, 11835, '\P{Is_Line_Break:  Break_BOTH}', "");
Expect(1, 11835, '\P{^Is_Line_Break:  Break_BOTH}', "");
Expect(0, 11836, '\p{Is_Line_Break:  Break_BOTH}', "");
Expect(1, 11836, '\p{^Is_Line_Break:  Break_BOTH}', "");
Expect(1, 11836, '\P{Is_Line_Break:  Break_BOTH}', "");
Expect(0, 11836, '\P{^Is_Line_Break:  Break_BOTH}', "");
Error('\p{Is_Lb=-/a/B2}');
Error('\P{Is_Lb=-/a/B2}');
Expect(1, 11835, '\p{Is_Lb=b2}', "");
Expect(0, 11835, '\p{^Is_Lb=b2}', "");
Expect(0, 11835, '\P{Is_Lb=b2}', "");
Expect(1, 11835, '\P{^Is_Lb=b2}', "");
Expect(0, 11836, '\p{Is_Lb=b2}', "");
Expect(1, 11836, '\p{^Is_Lb=b2}', "");
Expect(1, 11836, '\P{Is_Lb=b2}', "");
Expect(0, 11836, '\P{^Is_Lb=b2}', "");
Expect(1, 11835, '\p{Is_Lb: b2}', "");
Expect(0, 11835, '\p{^Is_Lb: b2}', "");
Expect(0, 11835, '\P{Is_Lb: b2}', "");
Expect(1, 11835, '\P{^Is_Lb: b2}', "");
Expect(0, 11836, '\p{Is_Lb: b2}', "");
Expect(1, 11836, '\p{^Is_Lb: b2}', "");
Expect(1, 11836, '\P{Is_Lb: b2}', "");
Expect(0, 11836, '\P{^Is_Lb: b2}', "");
Error('\p{Line_Break=_:=Break_After}');
Error('\P{Line_Break=_:=Break_After}');
Expect(1, 74867, '\p{Line_Break=breakafter}', "");
Expect(0, 74867, '\p{^Line_Break=breakafter}', "");
Expect(0, 74867, '\P{Line_Break=breakafter}', "");
Expect(1, 74867, '\P{^Line_Break=breakafter}', "");
Expect(0, 74868, '\p{Line_Break=breakafter}', "");
Expect(1, 74868, '\p{^Line_Break=breakafter}', "");
Expect(1, 74868, '\P{Line_Break=breakafter}', "");
Expect(0, 74868, '\P{^Line_Break=breakafter}', "");
Expect(1, 74867, '\p{Line_Break=_	break_AFTER}', "");
Expect(0, 74867, '\p{^Line_Break=_	break_AFTER}', "");
Expect(0, 74867, '\P{Line_Break=_	break_AFTER}', "");
Expect(1, 74867, '\P{^Line_Break=_	break_AFTER}', "");
Expect(0, 74868, '\p{Line_Break=_	break_AFTER}', "");
Expect(1, 74868, '\p{^Line_Break=_	break_AFTER}', "");
Expect(1, 74868, '\P{Line_Break=_	break_AFTER}', "");
Expect(0, 74868, '\P{^Line_Break=_	break_AFTER}', "");
Error('\p{Lb=/a/BA}');
Error('\P{Lb=/a/BA}');
Expect(1, 74867, '\p{Lb=ba}', "");
Expect(0, 74867, '\p{^Lb=ba}', "");
Expect(0, 74867, '\P{Lb=ba}', "");
Expect(1, 74867, '\P{^Lb=ba}', "");
Expect(0, 74868, '\p{Lb=ba}', "");
Expect(1, 74868, '\p{^Lb=ba}', "");
Expect(1, 74868, '\P{Lb=ba}', "");
Expect(0, 74868, '\P{^Lb=ba}', "");
Expect(1, 74867, '\p{Lb=	 ba}', "");
Expect(0, 74867, '\p{^Lb=	 ba}', "");
Expect(0, 74867, '\P{Lb=	 ba}', "");
Expect(1, 74867, '\P{^Lb=	 ba}', "");
Expect(0, 74868, '\p{Lb=	 ba}', "");
Expect(1, 74868, '\p{^Lb=	 ba}', "");
Expect(1, 74868, '\P{Lb=	 ba}', "");
Expect(0, 74868, '\P{^Lb=	 ba}', "");
Error('\p{Is_Line_Break=- Break_After:=}');
Error('\P{Is_Line_Break=- Break_After:=}');
Expect(1, 74867, '\p{Is_Line_Break=breakafter}', "");
Expect(0, 74867, '\p{^Is_Line_Break=breakafter}', "");
Expect(0, 74867, '\P{Is_Line_Break=breakafter}', "");
Expect(1, 74867, '\P{^Is_Line_Break=breakafter}', "");
Expect(0, 74868, '\p{Is_Line_Break=breakafter}', "");
Expect(1, 74868, '\p{^Is_Line_Break=breakafter}', "");
Expect(1, 74868, '\P{Is_Line_Break=breakafter}', "");
Expect(0, 74868, '\P{^Is_Line_Break=breakafter}', "");
Expect(1, 74867, '\p{Is_Line_Break=  Break_AFTER}', "");
Expect(0, 74867, '\p{^Is_Line_Break=  Break_AFTER}', "");
Expect(0, 74867, '\P{Is_Line_Break=  Break_AFTER}', "");
Expect(1, 74867, '\P{^Is_Line_Break=  Break_AFTER}', "");
Expect(0, 74868, '\p{Is_Line_Break=  Break_AFTER}', "");
Expect(1, 74868, '\p{^Is_Line_Break=  Break_AFTER}', "");
Expect(1, 74868, '\P{Is_Line_Break=  Break_AFTER}', "");
Expect(0, 74868, '\P{^Is_Line_Break=  Break_AFTER}', "");
Error('\p{Is_Lb=/a/_BA}');
Error('\P{Is_Lb=/a/_BA}');
Expect(1, 74867, '\p{Is_Lb=ba}', "");
Expect(0, 74867, '\p{^Is_Lb=ba}', "");
Expect(0, 74867, '\P{Is_Lb=ba}', "");
Expect(1, 74867, '\P{^Is_Lb=ba}', "");
Expect(0, 74868, '\p{Is_Lb=ba}', "");
Expect(1, 74868, '\p{^Is_Lb=ba}', "");
Expect(1, 74868, '\P{Is_Lb=ba}', "");
Expect(0, 74868, '\P{^Is_Lb=ba}', "");
Expect(1, 74867, '\p{Is_Lb=		BA}', "");
Expect(0, 74867, '\p{^Is_Lb=		BA}', "");
Expect(0, 74867, '\P{Is_Lb=		BA}', "");
Expect(1, 74867, '\P{^Is_Lb=		BA}', "");
Expect(0, 74868, '\p{Is_Lb=		BA}', "");
Expect(1, 74868, '\p{^Is_Lb=		BA}', "");
Expect(1, 74868, '\P{Is_Lb=		BA}', "");
Expect(0, 74868, '\P{^Is_Lb=		BA}', "");
Error('\p{Line_Break= Break_Before/a/}');
Error('\P{Line_Break= Break_Before/a/}');
Expect(1, 43125, '\p{Line_Break=breakbefore}', "");
Expect(0, 43125, '\p{^Line_Break=breakbefore}', "");
Expect(0, 43125, '\P{Line_Break=breakbefore}', "");
Expect(1, 43125, '\P{^Line_Break=breakbefore}', "");
Expect(0, 43126, '\p{Line_Break=breakbefore}', "");
Expect(1, 43126, '\p{^Line_Break=breakbefore}', "");
Expect(1, 43126, '\P{Line_Break=breakbefore}', "");
Expect(0, 43126, '\P{^Line_Break=breakbefore}', "");
Expect(1, 43125, '\p{Line_Break=_break_Before}', "");
Expect(0, 43125, '\p{^Line_Break=_break_Before}', "");
Expect(0, 43125, '\P{Line_Break=_break_Before}', "");
Expect(1, 43125, '\P{^Line_Break=_break_Before}', "");
Expect(0, 43126, '\p{Line_Break=_break_Before}', "");
Expect(1, 43126, '\p{^Line_Break=_break_Before}', "");
Expect(1, 43126, '\P{Line_Break=_break_Before}', "");
Expect(0, 43126, '\P{^Line_Break=_break_Before}', "");
Error('\p{Lb=:= -bb}');
Error('\P{Lb=:= -bb}');
Expect(1, 43125, '\p{Lb=bb}', "");
Expect(0, 43125, '\p{^Lb=bb}', "");
Expect(0, 43125, '\P{Lb=bb}', "");
Expect(1, 43125, '\P{^Lb=bb}', "");
Expect(0, 43126, '\p{Lb=bb}', "");
Expect(1, 43126, '\p{^Lb=bb}', "");
Expect(1, 43126, '\P{Lb=bb}', "");
Expect(0, 43126, '\P{^Lb=bb}', "");
Expect(1, 43125, '\p{Lb=-	BB}', "");
Expect(0, 43125, '\p{^Lb=-	BB}', "");
Expect(0, 43125, '\P{Lb=-	BB}', "");
Expect(1, 43125, '\P{^Lb=-	BB}', "");
Expect(0, 43126, '\p{Lb=-	BB}', "");
Expect(1, 43126, '\p{^Lb=-	BB}', "");
Expect(1, 43126, '\P{Lb=-	BB}', "");
Expect(0, 43126, '\P{^Lb=-	BB}', "");
Error('\p{Is_Line_Break=:=_	BREAK_Before}');
Error('\P{Is_Line_Break=:=_	BREAK_Before}');
Expect(1, 43125, '\p{Is_Line_Break=breakbefore}', "");
Expect(0, 43125, '\p{^Is_Line_Break=breakbefore}', "");
Expect(0, 43125, '\P{Is_Line_Break=breakbefore}', "");
Expect(1, 43125, '\P{^Is_Line_Break=breakbefore}', "");
Expect(0, 43126, '\p{Is_Line_Break=breakbefore}', "");
Expect(1, 43126, '\p{^Is_Line_Break=breakbefore}', "");
Expect(1, 43126, '\P{Is_Line_Break=breakbefore}', "");
Expect(0, 43126, '\P{^Is_Line_Break=breakbefore}', "");
Expect(1, 43125, '\p{Is_Line_Break:		 Break_BEFORE}', "");
Expect(0, 43125, '\p{^Is_Line_Break:		 Break_BEFORE}', "");
Expect(0, 43125, '\P{Is_Line_Break:		 Break_BEFORE}', "");
Expect(1, 43125, '\P{^Is_Line_Break:		 Break_BEFORE}', "");
Expect(0, 43126, '\p{Is_Line_Break:		 Break_BEFORE}', "");
Expect(1, 43126, '\p{^Is_Line_Break:		 Break_BEFORE}', "");
Expect(1, 43126, '\P{Is_Line_Break:		 Break_BEFORE}', "");
Expect(0, 43126, '\P{^Is_Line_Break:		 Break_BEFORE}', "");
Error('\p{Is_Lb=__BB/a/}');
Error('\P{Is_Lb=__BB/a/}');
Expect(1, 43125, '\p{Is_Lb=bb}', "");
Expect(0, 43125, '\p{^Is_Lb=bb}', "");
Expect(0, 43125, '\P{Is_Lb=bb}', "");
Expect(1, 43125, '\P{^Is_Lb=bb}', "");
Expect(0, 43126, '\p{Is_Lb=bb}', "");
Expect(1, 43126, '\p{^Is_Lb=bb}', "");
Expect(1, 43126, '\P{Is_Lb=bb}', "");
Expect(0, 43126, '\P{^Is_Lb=bb}', "");
Expect(1, 43125, '\p{Is_Lb= bb}', "");
Expect(0, 43125, '\p{^Is_Lb= bb}', "");
Expect(0, 43125, '\P{Is_Lb= bb}', "");
Expect(1, 43125, '\P{^Is_Lb= bb}', "");
Expect(0, 43126, '\p{Is_Lb= bb}', "");
Expect(1, 43126, '\p{^Is_Lb= bb}', "");
Expect(1, 43126, '\P{Is_Lb= bb}', "");
Expect(0, 43126, '\P{^Is_Lb= bb}', "");
Error('\p{Line_Break=-:=MANDATORY_break}');
Error('\P{Line_Break=-:=MANDATORY_break}');
Expect(1, 8233, '\p{Line_Break=mandatorybreak}', "");
Expect(0, 8233, '\p{^Line_Break=mandatorybreak}', "");
Expect(0, 8233, '\P{Line_Break=mandatorybreak}', "");
Expect(1, 8233, '\P{^Line_Break=mandatorybreak}', "");
Expect(0, 8234, '\p{Line_Break=mandatorybreak}', "");
Expect(1, 8234, '\p{^Line_Break=mandatorybreak}', "");
Expect(1, 8234, '\P{Line_Break=mandatorybreak}', "");
Expect(0, 8234, '\P{^Line_Break=mandatorybreak}', "");
Expect(1, 8233, '\p{Line_Break=		Mandatory_BREAK}', "");
Expect(0, 8233, '\p{^Line_Break=		Mandatory_BREAK}', "");
Expect(0, 8233, '\P{Line_Break=		Mandatory_BREAK}', "");
Expect(1, 8233, '\P{^Line_Break=		Mandatory_BREAK}', "");
Expect(0, 8234, '\p{Line_Break=		Mandatory_BREAK}', "");
Expect(1, 8234, '\p{^Line_Break=		Mandatory_BREAK}', "");
Expect(1, 8234, '\P{Line_Break=		Mandatory_BREAK}', "");
Expect(0, 8234, '\P{^Line_Break=		Mandatory_BREAK}', "");
Error('\p{Lb= BK:=}');
Error('\P{Lb= BK:=}');
Expect(1, 8233, '\p{Lb=bk}', "");
Expect(0, 8233, '\p{^Lb=bk}', "");
Expect(0, 8233, '\P{Lb=bk}', "");
Expect(1, 8233, '\P{^Lb=bk}', "");
Expect(0, 8234, '\p{Lb=bk}', "");
Expect(1, 8234, '\p{^Lb=bk}', "");
Expect(1, 8234, '\P{Lb=bk}', "");
Expect(0, 8234, '\P{^Lb=bk}', "");
Expect(1, 8233, '\p{Lb= bk}', "");
Expect(0, 8233, '\p{^Lb= bk}', "");
Expect(0, 8233, '\P{Lb= bk}', "");
Expect(1, 8233, '\P{^Lb= bk}', "");
Expect(0, 8234, '\p{Lb= bk}', "");
Expect(1, 8234, '\p{^Lb= bk}', "");
Expect(1, 8234, '\P{Lb= bk}', "");
Expect(0, 8234, '\P{^Lb= bk}', "");
Error('\p{Is_Line_Break=/a/_Mandatory_Break}');
Error('\P{Is_Line_Break=/a/_Mandatory_Break}');
Expect(1, 8233, '\p{Is_Line_Break:   mandatorybreak}', "");
Expect(0, 8233, '\p{^Is_Line_Break:   mandatorybreak}', "");
Expect(0, 8233, '\P{Is_Line_Break:   mandatorybreak}', "");
Expect(1, 8233, '\P{^Is_Line_Break:   mandatorybreak}', "");
Expect(0, 8234, '\p{Is_Line_Break:   mandatorybreak}', "");
Expect(1, 8234, '\p{^Is_Line_Break:   mandatorybreak}', "");
Expect(1, 8234, '\P{Is_Line_Break:   mandatorybreak}', "");
Expect(0, 8234, '\P{^Is_Line_Break:   mandatorybreak}', "");
Expect(1, 8233, '\p{Is_Line_Break=_ Mandatory_break}', "");
Expect(0, 8233, '\p{^Is_Line_Break=_ Mandatory_break}', "");
Expect(0, 8233, '\P{Is_Line_Break=_ Mandatory_break}', "");
Expect(1, 8233, '\P{^Is_Line_Break=_ Mandatory_break}', "");
Expect(0, 8234, '\p{Is_Line_Break=_ Mandatory_break}', "");
Expect(1, 8234, '\p{^Is_Line_Break=_ Mandatory_break}', "");
Expect(1, 8234, '\P{Is_Line_Break=_ Mandatory_break}', "");
Expect(0, 8234, '\P{^Is_Line_Break=_ Mandatory_break}', "");
Error('\p{Is_Lb=__BK:=}');
Error('\P{Is_Lb=__BK:=}');
Expect(1, 8233, '\p{Is_Lb:   bk}', "");
Expect(0, 8233, '\p{^Is_Lb:   bk}', "");
Expect(0, 8233, '\P{Is_Lb:   bk}', "");
Expect(1, 8233, '\P{^Is_Lb:   bk}', "");
Expect(0, 8234, '\p{Is_Lb:   bk}', "");
Expect(1, 8234, '\p{^Is_Lb:   bk}', "");
Expect(1, 8234, '\P{Is_Lb:   bk}', "");
Expect(0, 8234, '\P{^Is_Lb:   bk}', "");
Expect(1, 8233, '\p{Is_Lb:    BK}', "");
Expect(0, 8233, '\p{^Is_Lb:    BK}', "");
Expect(0, 8233, '\P{Is_Lb:    BK}', "");
Expect(1, 8233, '\P{^Is_Lb:    BK}', "");
Expect(0, 8234, '\p{Is_Lb:    BK}', "");
Expect(1, 8234, '\p{^Is_Lb:    BK}', "");
Expect(1, 8234, '\P{Is_Lb:    BK}', "");
Expect(0, 8234, '\P{^Is_Lb:    BK}', "");
Error('\p{Line_Break=_Contingent_Break/a/}');
Error('\P{Line_Break=_Contingent_Break/a/}');
Expect(1, 65532, '\p{Line_Break=contingentbreak}', "");
Expect(0, 65532, '\p{^Line_Break=contingentbreak}', "");
Expect(0, 65532, '\P{Line_Break=contingentbreak}', "");
Expect(1, 65532, '\P{^Line_Break=contingentbreak}', "");
Expect(0, 65533, '\p{Line_Break=contingentbreak}', "");
Expect(1, 65533, '\p{^Line_Break=contingentbreak}', "");
Expect(1, 65533, '\P{Line_Break=contingentbreak}', "");
Expect(0, 65533, '\P{^Line_Break=contingentbreak}', "");
Expect(1, 65532, '\p{Line_Break=	 Contingent_break}', "");
Expect(0, 65532, '\p{^Line_Break=	 Contingent_break}', "");
Expect(0, 65532, '\P{Line_Break=	 Contingent_break}', "");
Expect(1, 65532, '\P{^Line_Break=	 Contingent_break}', "");
Expect(0, 65533, '\p{Line_Break=	 Contingent_break}', "");
Expect(1, 65533, '\p{^Line_Break=	 Contingent_break}', "");
Expect(1, 65533, '\P{Line_Break=	 Contingent_break}', "");
Expect(0, 65533, '\P{^Line_Break=	 Contingent_break}', "");
Error('\p{Lb=/a/_CB}');
Error('\P{Lb=/a/_CB}');
Expect(1, 65532, '\p{Lb=cb}', "");
Expect(0, 65532, '\p{^Lb=cb}', "");
Expect(0, 65532, '\P{Lb=cb}', "");
Expect(1, 65532, '\P{^Lb=cb}', "");
Expect(0, 65533, '\p{Lb=cb}', "");
Expect(1, 65533, '\p{^Lb=cb}', "");
Expect(1, 65533, '\P{Lb=cb}', "");
Expect(0, 65533, '\P{^Lb=cb}', "");
Expect(1, 65532, '\p{Lb=--CB}', "");
Expect(0, 65532, '\p{^Lb=--CB}', "");
Expect(0, 65532, '\P{Lb=--CB}', "");
Expect(1, 65532, '\P{^Lb=--CB}', "");
Expect(0, 65533, '\p{Lb=--CB}', "");
Expect(1, 65533, '\p{^Lb=--CB}', "");
Expect(1, 65533, '\P{Lb=--CB}', "");
Expect(0, 65533, '\P{^Lb=--CB}', "");
Error('\p{Is_Line_Break=_/a/Contingent_BREAK}');
Error('\P{Is_Line_Break=_/a/Contingent_BREAK}');
Expect(1, 65532, '\p{Is_Line_Break=contingentbreak}', "");
Expect(0, 65532, '\p{^Is_Line_Break=contingentbreak}', "");
Expect(0, 65532, '\P{Is_Line_Break=contingentbreak}', "");
Expect(1, 65532, '\P{^Is_Line_Break=contingentbreak}', "");
Expect(0, 65533, '\p{Is_Line_Break=contingentbreak}', "");
Expect(1, 65533, '\p{^Is_Line_Break=contingentbreak}', "");
Expect(1, 65533, '\P{Is_Line_Break=contingentbreak}', "");
Expect(0, 65533, '\P{^Is_Line_Break=contingentbreak}', "");
Expect(1, 65532, '\p{Is_Line_Break=-_Contingent_BREAK}', "");
Expect(0, 65532, '\p{^Is_Line_Break=-_Contingent_BREAK}', "");
Expect(0, 65532, '\P{Is_Line_Break=-_Contingent_BREAK}', "");
Expect(1, 65532, '\P{^Is_Line_Break=-_Contingent_BREAK}', "");
Expect(0, 65533, '\p{Is_Line_Break=-_Contingent_BREAK}', "");
Expect(1, 65533, '\p{^Is_Line_Break=-_Contingent_BREAK}', "");
Expect(1, 65533, '\P{Is_Line_Break=-_Contingent_BREAK}', "");
Expect(0, 65533, '\P{^Is_Line_Break=-_Contingent_BREAK}', "");
Error('\p{Is_Lb=:=_CB}');
Error('\P{Is_Lb=:=_CB}');
Expect(1, 65532, '\p{Is_Lb=cb}', "");
Expect(0, 65532, '\p{^Is_Lb=cb}', "");
Expect(0, 65532, '\P{Is_Lb=cb}', "");
Expect(1, 65532, '\P{^Is_Lb=cb}', "");
Expect(0, 65533, '\p{Is_Lb=cb}', "");
Expect(1, 65533, '\p{^Is_Lb=cb}', "");
Expect(1, 65533, '\P{Is_Lb=cb}', "");
Expect(0, 65533, '\P{^Is_Lb=cb}', "");
Expect(1, 65532, '\p{Is_Lb=- CB}', "");
Expect(0, 65532, '\p{^Is_Lb=- CB}', "");
Expect(0, 65532, '\P{Is_Lb=- CB}', "");
Expect(1, 65532, '\P{^Is_Lb=- CB}', "");
Expect(0, 65533, '\p{Is_Lb=- CB}', "");
Expect(1, 65533, '\p{^Is_Lb=- CB}', "");
Expect(1, 65533, '\P{Is_Lb=- CB}', "");
Expect(0, 65533, '\P{^Is_Lb=- CB}', "");
Error('\p{Line_Break=__CONDITIONAL_japanese_starter:=}');
Error('\P{Line_Break=__CONDITIONAL_japanese_starter:=}');
Expect(1, 65392, '\p{Line_Break=conditionaljapanesestarter}', "");
Expect(0, 65392, '\p{^Line_Break=conditionaljapanesestarter}', "");
Expect(0, 65392, '\P{Line_Break=conditionaljapanesestarter}', "");
Expect(1, 65392, '\P{^Line_Break=conditionaljapanesestarter}', "");
Expect(0, 65393, '\p{Line_Break=conditionaljapanesestarter}', "");
Expect(1, 65393, '\p{^Line_Break=conditionaljapanesestarter}', "");
Expect(1, 65393, '\P{Line_Break=conditionaljapanesestarter}', "");
Expect(0, 65393, '\P{^Line_Break=conditionaljapanesestarter}', "");
Expect(1, 65392, '\p{Line_Break=_ Conditional_Japanese_Starter}', "");
Expect(0, 65392, '\p{^Line_Break=_ Conditional_Japanese_Starter}', "");
Expect(0, 65392, '\P{Line_Break=_ Conditional_Japanese_Starter}', "");
Expect(1, 65392, '\P{^Line_Break=_ Conditional_Japanese_Starter}', "");
Expect(0, 65393, '\p{Line_Break=_ Conditional_Japanese_Starter}', "");
Expect(1, 65393, '\p{^Line_Break=_ Conditional_Japanese_Starter}', "");
Expect(1, 65393, '\P{Line_Break=_ Conditional_Japanese_Starter}', "");
Expect(0, 65393, '\P{^Line_Break=_ Conditional_Japanese_Starter}', "");
Error('\p{Lb=/a/	 cj}');
Error('\P{Lb=/a/	 cj}');
Expect(1, 65392, '\p{Lb=cj}', "");
Expect(0, 65392, '\p{^Lb=cj}', "");
Expect(0, 65392, '\P{Lb=cj}', "");
Expect(1, 65392, '\P{^Lb=cj}', "");
Expect(0, 65393, '\p{Lb=cj}', "");
Expect(1, 65393, '\p{^Lb=cj}', "");
Expect(1, 65393, '\P{Lb=cj}', "");
Expect(0, 65393, '\P{^Lb=cj}', "");
Expect(1, 65392, '\p{Lb=--CJ}', "");
Expect(0, 65392, '\p{^Lb=--CJ}', "");
Expect(0, 65392, '\P{Lb=--CJ}', "");
Expect(1, 65392, '\P{^Lb=--CJ}', "");
Expect(0, 65393, '\p{Lb=--CJ}', "");
Expect(1, 65393, '\p{^Lb=--CJ}', "");
Expect(1, 65393, '\P{Lb=--CJ}', "");
Expect(0, 65393, '\P{^Lb=--CJ}', "");
Error('\p{Is_Line_Break=:=-	Conditional_Japanese_Starter}');
Error('\P{Is_Line_Break=:=-	Conditional_Japanese_Starter}');
Expect(1, 65392, '\p{Is_Line_Break=conditionaljapanesestarter}', "");
Expect(0, 65392, '\p{^Is_Line_Break=conditionaljapanesestarter}', "");
Expect(0, 65392, '\P{Is_Line_Break=conditionaljapanesestarter}', "");
Expect(1, 65392, '\P{^Is_Line_Break=conditionaljapanesestarter}', "");
Expect(0, 65393, '\p{Is_Line_Break=conditionaljapanesestarter}', "");
Expect(1, 65393, '\p{^Is_Line_Break=conditionaljapanesestarter}', "");
Expect(1, 65393, '\P{Is_Line_Break=conditionaljapanesestarter}', "");
Expect(0, 65393, '\P{^Is_Line_Break=conditionaljapanesestarter}', "");
Expect(1, 65392, '\p{Is_Line_Break:	-Conditional_japanese_Starter}', "");
Expect(0, 65392, '\p{^Is_Line_Break:	-Conditional_japanese_Starter}', "");
Expect(0, 65392, '\P{Is_Line_Break:	-Conditional_japanese_Starter}', "");
Expect(1, 65392, '\P{^Is_Line_Break:	-Conditional_japanese_Starter}', "");
Expect(0, 65393, '\p{Is_Line_Break:	-Conditional_japanese_Starter}', "");
Expect(1, 65393, '\p{^Is_Line_Break:	-Conditional_japanese_Starter}', "");
Expect(1, 65393, '\P{Is_Line_Break:	-Conditional_japanese_Starter}', "");
Expect(0, 65393, '\P{^Is_Line_Break:	-Conditional_japanese_Starter}', "");
Error('\p{Is_Lb=	_CJ/a/}');
Error('\P{Is_Lb=	_CJ/a/}');
Expect(1, 65392, '\p{Is_Lb=cj}', "");
Expect(0, 65392, '\p{^Is_Lb=cj}', "");
Expect(0, 65392, '\P{Is_Lb=cj}', "");
Expect(1, 65392, '\P{^Is_Lb=cj}', "");
Expect(0, 65393, '\p{Is_Lb=cj}', "");
Expect(1, 65393, '\p{^Is_Lb=cj}', "");
Expect(1, 65393, '\P{Is_Lb=cj}', "");
Expect(0, 65393, '\P{^Is_Lb=cj}', "");
Expect(1, 65392, '\p{Is_Lb=		cj}', "");
Expect(0, 65392, '\p{^Is_Lb=		cj}', "");
Expect(0, 65392, '\P{Is_Lb=		cj}', "");
Expect(1, 65392, '\P{^Is_Lb=		cj}', "");
Expect(0, 65393, '\p{Is_Lb=		cj}', "");
Expect(1, 65393, '\p{^Is_Lb=		cj}', "");
Expect(1, 65393, '\P{Is_Lb=		cj}', "");
Expect(0, 65393, '\P{^Is_Lb=		cj}', "");
Error('\p{Line_Break=/a/-	Close_Punctuation}');
Error('\P{Line_Break=/a/-	Close_Punctuation}');
Expect(1, 78715, '\p{Line_Break=closepunctuation}', "");
Expect(0, 78715, '\p{^Line_Break=closepunctuation}', "");
Expect(0, 78715, '\P{Line_Break=closepunctuation}', "");
Expect(1, 78715, '\P{^Line_Break=closepunctuation}', "");
Expect(0, 78716, '\p{Line_Break=closepunctuation}', "");
Expect(1, 78716, '\p{^Line_Break=closepunctuation}', "");
Expect(1, 78716, '\P{Line_Break=closepunctuation}', "");
Expect(0, 78716, '\P{^Line_Break=closepunctuation}', "");
Expect(1, 78715, '\p{Line_Break= 	Close_PUNCTUATION}', "");
Expect(0, 78715, '\p{^Line_Break= 	Close_PUNCTUATION}', "");
Expect(0, 78715, '\P{Line_Break= 	Close_PUNCTUATION}', "");
Expect(1, 78715, '\P{^Line_Break= 	Close_PUNCTUATION}', "");
Expect(0, 78716, '\p{Line_Break= 	Close_PUNCTUATION}', "");
Expect(1, 78716, '\p{^Line_Break= 	Close_PUNCTUATION}', "");
Expect(1, 78716, '\P{Line_Break= 	Close_PUNCTUATION}', "");
Expect(0, 78716, '\P{^Line_Break= 	Close_PUNCTUATION}', "");
Error('\p{Lb=/a/ _cl}');
Error('\P{Lb=/a/ _cl}');
Expect(1, 78715, '\p{Lb=cl}', "");
Expect(0, 78715, '\p{^Lb=cl}', "");
Expect(0, 78715, '\P{Lb=cl}', "");
Expect(1, 78715, '\P{^Lb=cl}', "");
Expect(0, 78716, '\p{Lb=cl}', "");
Expect(1, 78716, '\p{^Lb=cl}', "");
Expect(1, 78716, '\P{Lb=cl}', "");
Expect(0, 78716, '\P{^Lb=cl}', "");
Expect(1, 78715, '\p{Lb=-CL}', "");
Expect(0, 78715, '\p{^Lb=-CL}', "");
Expect(0, 78715, '\P{Lb=-CL}', "");
Expect(1, 78715, '\P{^Lb=-CL}', "");
Expect(0, 78716, '\p{Lb=-CL}', "");
Expect(1, 78716, '\p{^Lb=-CL}', "");
Expect(1, 78716, '\P{Lb=-CL}', "");
Expect(0, 78716, '\P{^Lb=-CL}', "");
Error('\p{Is_Line_Break=	CLOSE_PUNCTUATION:=}');
Error('\P{Is_Line_Break=	CLOSE_PUNCTUATION:=}');
Expect(1, 78715, '\p{Is_Line_Break=closepunctuation}', "");
Expect(0, 78715, '\p{^Is_Line_Break=closepunctuation}', "");
Expect(0, 78715, '\P{Is_Line_Break=closepunctuation}', "");
Expect(1, 78715, '\P{^Is_Line_Break=closepunctuation}', "");
Expect(0, 78716, '\p{Is_Line_Break=closepunctuation}', "");
Expect(1, 78716, '\p{^Is_Line_Break=closepunctuation}', "");
Expect(1, 78716, '\P{Is_Line_Break=closepunctuation}', "");
Expect(0, 78716, '\P{^Is_Line_Break=closepunctuation}', "");
Expect(1, 78715, '\p{Is_Line_Break=  CLOSE_Punctuation}', "");
Expect(0, 78715, '\p{^Is_Line_Break=  CLOSE_Punctuation}', "");
Expect(0, 78715, '\P{Is_Line_Break=  CLOSE_Punctuation}', "");
Expect(1, 78715, '\P{^Is_Line_Break=  CLOSE_Punctuation}', "");
Expect(0, 78716, '\p{Is_Line_Break=  CLOSE_Punctuation}', "");
Expect(1, 78716, '\p{^Is_Line_Break=  CLOSE_Punctuation}', "");
Expect(1, 78716, '\P{Is_Line_Break=  CLOSE_Punctuation}', "");
Expect(0, 78716, '\P{^Is_Line_Break=  CLOSE_Punctuation}', "");
Error('\p{Is_Lb=	CL/a/}');
Error('\P{Is_Lb=	CL/a/}');
Expect(1, 78715, '\p{Is_Lb=cl}', "");
Expect(0, 78715, '\p{^Is_Lb=cl}', "");
Expect(0, 78715, '\P{Is_Lb=cl}', "");
Expect(1, 78715, '\P{^Is_Lb=cl}', "");
Expect(0, 78716, '\p{Is_Lb=cl}', "");
Expect(1, 78716, '\p{^Is_Lb=cl}', "");
Expect(1, 78716, '\P{Is_Lb=cl}', "");
Expect(0, 78716, '\P{^Is_Lb=cl}', "");
Expect(1, 78715, '\p{Is_Lb=__CL}', "");
Expect(0, 78715, '\p{^Is_Lb=__CL}', "");
Expect(0, 78715, '\P{Is_Lb=__CL}', "");
Expect(1, 78715, '\P{^Is_Lb=__CL}', "");
Expect(0, 78716, '\p{Is_Lb=__CL}', "");
Expect(1, 78716, '\p{^Is_Lb=__CL}', "");
Expect(1, 78716, '\P{Is_Lb=__CL}', "");
Expect(0, 78716, '\P{^Is_Lb=__CL}', "");
Error('\p{Line_Break= /a/COMBINING_Mark}');
Error('\P{Line_Break= /a/COMBINING_Mark}');
Expect(1, 917999, '\p{Line_Break:	combiningmark}', "");
Expect(0, 917999, '\p{^Line_Break:	combiningmark}', "");
Expect(0, 917999, '\P{Line_Break:	combiningmark}', "");
Expect(1, 917999, '\P{^Line_Break:	combiningmark}', "");
Expect(0, 918000, '\p{Line_Break:	combiningmark}', "");
Expect(1, 918000, '\p{^Line_Break:	combiningmark}', "");
Expect(1, 918000, '\P{Line_Break:	combiningmark}', "");
Expect(0, 918000, '\P{^Line_Break:	combiningmark}', "");
Expect(1, 917999, '\p{Line_Break=Combining_mark}', "");
Expect(0, 917999, '\p{^Line_Break=Combining_mark}', "");
Expect(0, 917999, '\P{Line_Break=Combining_mark}', "");
Expect(1, 917999, '\P{^Line_Break=Combining_mark}', "");
Expect(0, 918000, '\p{Line_Break=Combining_mark}', "");
Expect(1, 918000, '\p{^Line_Break=Combining_mark}', "");
Expect(1, 918000, '\P{Line_Break=Combining_mark}', "");
Expect(0, 918000, '\P{^Line_Break=Combining_mark}', "");
Error('\p{Lb: CM:=}');
Error('\P{Lb: CM:=}');
Expect(1, 917999, '\p{Lb=cm}', "");
Expect(0, 917999, '\p{^Lb=cm}', "");
Expect(0, 917999, '\P{Lb=cm}', "");
Expect(1, 917999, '\P{^Lb=cm}', "");
Expect(0, 918000, '\p{Lb=cm}', "");
Expect(1, 918000, '\p{^Lb=cm}', "");
Expect(1, 918000, '\P{Lb=cm}', "");
Expect(0, 918000, '\P{^Lb=cm}', "");
Expect(1, 917999, '\p{Lb= 	CM}', "");
Expect(0, 917999, '\p{^Lb= 	CM}', "");
Expect(0, 917999, '\P{Lb= 	CM}', "");
Expect(1, 917999, '\P{^Lb= 	CM}', "");
Expect(0, 918000, '\p{Lb= 	CM}', "");
Expect(1, 918000, '\p{^Lb= 	CM}', "");
Expect(1, 918000, '\P{Lb= 	CM}', "");
Expect(0, 918000, '\P{^Lb= 	CM}', "");
Error('\p{Is_Line_Break:-:=Combining_Mark}');
Error('\P{Is_Line_Break:-:=Combining_Mark}');
Expect(1, 917999, '\p{Is_Line_Break=combiningmark}', "");
Expect(0, 917999, '\p{^Is_Line_Break=combiningmark}', "");
Expect(0, 917999, '\P{Is_Line_Break=combiningmark}', "");
Expect(1, 917999, '\P{^Is_Line_Break=combiningmark}', "");
Expect(0, 918000, '\p{Is_Line_Break=combiningmark}', "");
Expect(1, 918000, '\p{^Is_Line_Break=combiningmark}', "");
Expect(1, 918000, '\P{Is_Line_Break=combiningmark}', "");
Expect(0, 918000, '\P{^Is_Line_Break=combiningmark}', "");
Expect(1, 917999, '\p{Is_Line_Break:	_	Combining_mark}', "");
Expect(0, 917999, '\p{^Is_Line_Break:	_	Combining_mark}', "");
Expect(0, 917999, '\P{Is_Line_Break:	_	Combining_mark}', "");
Expect(1, 917999, '\P{^Is_Line_Break:	_	Combining_mark}', "");
Expect(0, 918000, '\p{Is_Line_Break:	_	Combining_mark}', "");
Expect(1, 918000, '\p{^Is_Line_Break:	_	Combining_mark}', "");
Expect(1, 918000, '\P{Is_Line_Break:	_	Combining_mark}', "");
Expect(0, 918000, '\P{^Is_Line_Break:	_	Combining_mark}', "");
Error('\p{Is_Lb=	-cm:=}');
Error('\P{Is_Lb=	-cm:=}');
Expect(1, 917999, '\p{Is_Lb=cm}', "");
Expect(0, 917999, '\p{^Is_Lb=cm}', "");
Expect(0, 917999, '\P{Is_Lb=cm}', "");
Expect(1, 917999, '\P{^Is_Lb=cm}', "");
Expect(0, 918000, '\p{Is_Lb=cm}', "");
Expect(1, 918000, '\p{^Is_Lb=cm}', "");
Expect(1, 918000, '\P{Is_Lb=cm}', "");
Expect(0, 918000, '\P{^Is_Lb=cm}', "");
Expect(1, 917999, '\p{Is_Lb=_CM}', "");
Expect(0, 917999, '\p{^Is_Lb=_CM}', "");
Expect(0, 917999, '\P{Is_Lb=_CM}', "");
Expect(1, 917999, '\P{^Is_Lb=_CM}', "");
Expect(0, 918000, '\p{Is_Lb=_CM}', "");
Expect(1, 918000, '\p{^Is_Lb=_CM}', "");
Expect(1, 918000, '\P{Is_Lb=_CM}', "");
Expect(0, 918000, '\P{^Is_Lb=_CM}', "");
Error('\p{Line_Break=_/a/Close_Parenthesis}');
Error('\P{Line_Break=_/a/Close_Parenthesis}');
Expect(1, 93, '\p{Line_Break=closeparenthesis}', "");
Expect(0, 93, '\p{^Line_Break=closeparenthesis}', "");
Expect(0, 93, '\P{Line_Break=closeparenthesis}', "");
Expect(1, 93, '\P{^Line_Break=closeparenthesis}', "");
Expect(0, 94, '\p{Line_Break=closeparenthesis}', "");
Expect(1, 94, '\p{^Line_Break=closeparenthesis}', "");
Expect(1, 94, '\P{Line_Break=closeparenthesis}', "");
Expect(0, 94, '\P{^Line_Break=closeparenthesis}', "");
Expect(1, 93, '\p{Line_Break: _	close_Parenthesis}', "");
Expect(0, 93, '\p{^Line_Break: _	close_Parenthesis}', "");
Expect(0, 93, '\P{Line_Break: _	close_Parenthesis}', "");
Expect(1, 93, '\P{^Line_Break: _	close_Parenthesis}', "");
Expect(0, 94, '\p{Line_Break: _	close_Parenthesis}', "");
Expect(1, 94, '\p{^Line_Break: _	close_Parenthesis}', "");
Expect(1, 94, '\P{Line_Break: _	close_Parenthesis}', "");
Expect(0, 94, '\P{^Line_Break: _	close_Parenthesis}', "");
Error('\p{Lb=	/a/CP}');
Error('\P{Lb=	/a/CP}');
Expect(1, 93, '\p{Lb=cp}', "");
Expect(0, 93, '\p{^Lb=cp}', "");
Expect(0, 93, '\P{Lb=cp}', "");
Expect(1, 93, '\P{^Lb=cp}', "");
Expect(0, 94, '\p{Lb=cp}', "");
Expect(1, 94, '\p{^Lb=cp}', "");
Expect(1, 94, '\P{Lb=cp}', "");
Expect(0, 94, '\P{^Lb=cp}', "");
Expect(1, 93, '\p{Lb=-CP}', "");
Expect(0, 93, '\p{^Lb=-CP}', "");
Expect(0, 93, '\P{Lb=-CP}', "");
Expect(1, 93, '\P{^Lb=-CP}', "");
Expect(0, 94, '\p{Lb=-CP}', "");
Expect(1, 94, '\p{^Lb=-CP}', "");
Expect(1, 94, '\P{Lb=-CP}', "");
Expect(0, 94, '\P{^Lb=-CP}', "");
Error('\p{Is_Line_Break=_Close_Parenthesis:=}');
Error('\P{Is_Line_Break=_Close_Parenthesis:=}');
Expect(1, 93, '\p{Is_Line_Break=closeparenthesis}', "");
Expect(0, 93, '\p{^Is_Line_Break=closeparenthesis}', "");
Expect(0, 93, '\P{Is_Line_Break=closeparenthesis}', "");
Expect(1, 93, '\P{^Is_Line_Break=closeparenthesis}', "");
Expect(0, 94, '\p{Is_Line_Break=closeparenthesis}', "");
Expect(1, 94, '\p{^Is_Line_Break=closeparenthesis}', "");
Expect(1, 94, '\P{Is_Line_Break=closeparenthesis}', "");
Expect(0, 94, '\P{^Is_Line_Break=closeparenthesis}', "");
Expect(1, 93, '\p{Is_Line_Break=_	close_PARENTHESIS}', "");
Expect(0, 93, '\p{^Is_Line_Break=_	close_PARENTHESIS}', "");
Expect(0, 93, '\P{Is_Line_Break=_	close_PARENTHESIS}', "");
Expect(1, 93, '\P{^Is_Line_Break=_	close_PARENTHESIS}', "");
Expect(0, 94, '\p{Is_Line_Break=_	close_PARENTHESIS}', "");
Expect(1, 94, '\p{^Is_Line_Break=_	close_PARENTHESIS}', "");
Expect(1, 94, '\P{Is_Line_Break=_	close_PARENTHESIS}', "");
Expect(0, 94, '\P{^Is_Line_Break=_	close_PARENTHESIS}', "");
Error('\p{Is_Lb=:=_-CP}');
Error('\P{Is_Lb=:=_-CP}');
Expect(1, 93, '\p{Is_Lb:   cp}', "");
Expect(0, 93, '\p{^Is_Lb:   cp}', "");
Expect(0, 93, '\P{Is_Lb:   cp}', "");
Expect(1, 93, '\P{^Is_Lb:   cp}', "");
Expect(0, 94, '\p{Is_Lb:   cp}', "");
Expect(1, 94, '\p{^Is_Lb:   cp}', "");
Expect(1, 94, '\P{Is_Lb:   cp}', "");
Expect(0, 94, '\P{^Is_Lb:   cp}', "");
Expect(1, 93, '\p{Is_Lb= CP}', "");
Expect(0, 93, '\p{^Is_Lb= CP}', "");
Expect(0, 93, '\P{Is_Lb= CP}', "");
Expect(1, 93, '\P{^Is_Lb= CP}', "");
Expect(0, 94, '\p{Is_Lb= CP}', "");
Expect(1, 94, '\p{^Is_Lb= CP}', "");
Expect(1, 94, '\P{Is_Lb= CP}', "");
Expect(0, 94, '\P{^Is_Lb= CP}', "");
Error('\p{Line_Break=_	Carriage_return/a/}');
Error('\P{Line_Break=_	Carriage_return/a/}');
Expect(1, 13, '\p{Line_Break=carriagereturn}', "");
Expect(0, 13, '\p{^Line_Break=carriagereturn}', "");
Expect(0, 13, '\P{Line_Break=carriagereturn}', "");
Expect(1, 13, '\P{^Line_Break=carriagereturn}', "");
Expect(0, 14, '\p{Line_Break=carriagereturn}', "");
Expect(1, 14, '\p{^Line_Break=carriagereturn}', "");
Expect(1, 14, '\P{Line_Break=carriagereturn}', "");
Expect(0, 14, '\P{^Line_Break=carriagereturn}', "");
Expect(1, 13, '\p{Line_Break= -Carriage_RETURN}', "");
Expect(0, 13, '\p{^Line_Break= -Carriage_RETURN}', "");
Expect(0, 13, '\P{Line_Break= -Carriage_RETURN}', "");
Expect(1, 13, '\P{^Line_Break= -Carriage_RETURN}', "");
Expect(0, 14, '\p{Line_Break= -Carriage_RETURN}', "");
Expect(1, 14, '\p{^Line_Break= -Carriage_RETURN}', "");
Expect(1, 14, '\P{Line_Break= -Carriage_RETURN}', "");
Expect(0, 14, '\P{^Line_Break= -Carriage_RETURN}', "");
Error('\p{Lb=:=--CR}');
Error('\P{Lb=:=--CR}');
Expect(1, 13, '\p{Lb=cr}', "");
Expect(0, 13, '\p{^Lb=cr}', "");
Expect(0, 13, '\P{Lb=cr}', "");
Expect(1, 13, '\P{^Lb=cr}', "");
Expect(0, 14, '\p{Lb=cr}', "");
Expect(1, 14, '\p{^Lb=cr}', "");
Expect(1, 14, '\P{Lb=cr}', "");
Expect(0, 14, '\P{^Lb=cr}', "");
Expect(1, 13, '\p{Lb=- cr}', "");
Expect(0, 13, '\p{^Lb=- cr}', "");
Expect(0, 13, '\P{Lb=- cr}', "");
Expect(1, 13, '\P{^Lb=- cr}', "");
Expect(0, 14, '\p{Lb=- cr}', "");
Expect(1, 14, '\p{^Lb=- cr}', "");
Expect(1, 14, '\P{Lb=- cr}', "");
Expect(0, 14, '\P{^Lb=- cr}', "");
Error('\p{Is_Line_Break=	:=Carriage_Return}');
Error('\P{Is_Line_Break=	:=Carriage_Return}');
Expect(1, 13, '\p{Is_Line_Break:	carriagereturn}', "");
Expect(0, 13, '\p{^Is_Line_Break:	carriagereturn}', "");
Expect(0, 13, '\P{Is_Line_Break:	carriagereturn}', "");
Expect(1, 13, '\P{^Is_Line_Break:	carriagereturn}', "");
Expect(0, 14, '\p{Is_Line_Break:	carriagereturn}', "");
Expect(1, 14, '\p{^Is_Line_Break:	carriagereturn}', "");
Expect(1, 14, '\P{Is_Line_Break:	carriagereturn}', "");
Expect(0, 14, '\P{^Is_Line_Break:	carriagereturn}', "");
Expect(1, 13, '\p{Is_Line_Break=	Carriage_Return}', "");
Expect(0, 13, '\p{^Is_Line_Break=	Carriage_Return}', "");
Expect(0, 13, '\P{Is_Line_Break=	Carriage_Return}', "");
Expect(1, 13, '\P{^Is_Line_Break=	Carriage_Return}', "");
Expect(0, 14, '\p{Is_Line_Break=	Carriage_Return}', "");
Expect(1, 14, '\p{^Is_Line_Break=	Carriage_Return}', "");
Expect(1, 14, '\P{Is_Line_Break=	Carriage_Return}', "");
Expect(0, 14, '\P{^Is_Line_Break=	Carriage_Return}', "");
Error('\p{Is_Lb:	:=CR}');
Error('\P{Is_Lb:	:=CR}');
Expect(1, 13, '\p{Is_Lb=cr}', "");
Expect(0, 13, '\p{^Is_Lb=cr}', "");
Expect(0, 13, '\P{Is_Lb=cr}', "");
Expect(1, 13, '\P{^Is_Lb=cr}', "");
Expect(0, 14, '\p{Is_Lb=cr}', "");
Expect(1, 14, '\p{^Is_Lb=cr}', "");
Expect(1, 14, '\P{Is_Lb=cr}', "");
Expect(0, 14, '\P{^Is_Lb=cr}', "");
Expect(1, 13, '\p{Is_Lb= -CR}', "");
Expect(0, 13, '\p{^Is_Lb= -CR}', "");
Expect(0, 13, '\P{Is_Lb= -CR}', "");
Expect(1, 13, '\P{^Is_Lb= -CR}', "");
Expect(0, 14, '\p{Is_Lb= -CR}', "");
Expect(1, 14, '\p{^Is_Lb= -CR}', "");
Expect(1, 14, '\P{Is_Lb= -CR}', "");
Expect(0, 14, '\P{^Is_Lb= -CR}', "");
Error('\p{Line_Break=:= -exclamation}');
Error('\P{Line_Break=:= -exclamation}');
Expect(1, 65311, '\p{Line_Break=exclamation}', "");
Expect(0, 65311, '\p{^Line_Break=exclamation}', "");
Expect(0, 65311, '\P{Line_Break=exclamation}', "");
Expect(1, 65311, '\P{^Line_Break=exclamation}', "");
Expect(0, 65312, '\p{Line_Break=exclamation}', "");
Expect(1, 65312, '\p{^Line_Break=exclamation}', "");
Expect(1, 65312, '\P{Line_Break=exclamation}', "");
Expect(0, 65312, '\P{^Line_Break=exclamation}', "");
Expect(1, 65311, '\p{Line_Break=_	exclamation}', "");
Expect(0, 65311, '\p{^Line_Break=_	exclamation}', "");
Expect(0, 65311, '\P{Line_Break=_	exclamation}', "");
Expect(1, 65311, '\P{^Line_Break=_	exclamation}', "");
Expect(0, 65312, '\p{Line_Break=_	exclamation}', "");
Expect(1, 65312, '\p{^Line_Break=_	exclamation}', "");
Expect(1, 65312, '\P{Line_Break=_	exclamation}', "");
Expect(0, 65312, '\P{^Line_Break=_	exclamation}', "");
Error('\p{Lb=-/a/EX}');
Error('\P{Lb=-/a/EX}');
Expect(1, 65311, '\p{Lb=ex}', "");
Expect(0, 65311, '\p{^Lb=ex}', "");
Expect(0, 65311, '\P{Lb=ex}', "");
Expect(1, 65311, '\P{^Lb=ex}', "");
Expect(0, 65312, '\p{Lb=ex}', "");
Expect(1, 65312, '\p{^Lb=ex}', "");
Expect(1, 65312, '\P{Lb=ex}', "");
Expect(0, 65312, '\P{^Lb=ex}', "");
Expect(1, 65311, '\p{Lb=		EX}', "");
Expect(0, 65311, '\p{^Lb=		EX}', "");
Expect(0, 65311, '\P{Lb=		EX}', "");
Expect(1, 65311, '\P{^Lb=		EX}', "");
Expect(0, 65312, '\p{Lb=		EX}', "");
Expect(1, 65312, '\p{^Lb=		EX}', "");
Expect(1, 65312, '\P{Lb=		EX}', "");
Expect(0, 65312, '\P{^Lb=		EX}', "");
Error('\p{Is_Line_Break=/a/- EXCLAMATION}');
Error('\P{Is_Line_Break=/a/- EXCLAMATION}');
Expect(1, 65311, '\p{Is_Line_Break=exclamation}', "");
Expect(0, 65311, '\p{^Is_Line_Break=exclamation}', "");
Expect(0, 65311, '\P{Is_Line_Break=exclamation}', "");
Expect(1, 65311, '\P{^Is_Line_Break=exclamation}', "");
Expect(0, 65312, '\p{Is_Line_Break=exclamation}', "");
Expect(1, 65312, '\p{^Is_Line_Break=exclamation}', "");
Expect(1, 65312, '\P{Is_Line_Break=exclamation}', "");
Expect(0, 65312, '\P{^Is_Line_Break=exclamation}', "");
Expect(1, 65311, '\p{Is_Line_Break=_	exclamation}', "");
Expect(0, 65311, '\p{^Is_Line_Break=_	exclamation}', "");
Expect(0, 65311, '\P{Is_Line_Break=_	exclamation}', "");
Expect(1, 65311, '\P{^Is_Line_Break=_	exclamation}', "");
Expect(0, 65312, '\p{Is_Line_Break=_	exclamation}', "");
Expect(1, 65312, '\p{^Is_Line_Break=_	exclamation}', "");
Expect(1, 65312, '\P{Is_Line_Break=_	exclamation}', "");
Expect(0, 65312, '\P{^Is_Line_Break=_	exclamation}', "");
Error('\p{Is_Lb=/a/EX}');
Error('\P{Is_Lb=/a/EX}');
Expect(1, 65311, '\p{Is_Lb=ex}', "");
Expect(0, 65311, '\p{^Is_Lb=ex}', "");
Expect(0, 65311, '\P{Is_Lb=ex}', "");
Expect(1, 65311, '\P{^Is_Lb=ex}', "");
Expect(0, 65312, '\p{Is_Lb=ex}', "");
Expect(1, 65312, '\p{^Is_Lb=ex}', "");
Expect(1, 65312, '\P{Is_Lb=ex}', "");
Expect(0, 65312, '\P{^Is_Lb=ex}', "");
Expect(1, 65311, '\p{Is_Lb=	EX}', "");
Expect(0, 65311, '\p{^Is_Lb=	EX}', "");
Expect(0, 65311, '\P{Is_Lb=	EX}', "");
Expect(1, 65311, '\P{^Is_Lb=	EX}', "");
Expect(0, 65312, '\p{Is_Lb=	EX}', "");
Expect(1, 65312, '\p{^Is_Lb=	EX}', "");
Expect(1, 65312, '\P{Is_Lb=	EX}', "");
Expect(0, 65312, '\P{^Is_Lb=	EX}', "");
Error('\p{Line_Break=-_Glue:=}');
Error('\P{Line_Break=-_Glue:=}');
Expect(1, 8239, '\p{Line_Break=glue}', "");
Expect(0, 8239, '\p{^Line_Break=glue}', "");
Expect(0, 8239, '\P{Line_Break=glue}', "");
Expect(1, 8239, '\P{^Line_Break=glue}', "");
Expect(0, 8240, '\p{Line_Break=glue}', "");
Expect(1, 8240, '\p{^Line_Break=glue}', "");
Expect(1, 8240, '\P{Line_Break=glue}', "");
Expect(0, 8240, '\P{^Line_Break=glue}', "");
Expect(1, 8239, '\p{Line_Break= glue}', "");
Expect(0, 8239, '\p{^Line_Break= glue}', "");
Expect(0, 8239, '\P{Line_Break= glue}', "");
Expect(1, 8239, '\P{^Line_Break= glue}', "");
Expect(0, 8240, '\p{Line_Break= glue}', "");
Expect(1, 8240, '\p{^Line_Break= glue}', "");
Expect(1, 8240, '\P{Line_Break= glue}', "");
Expect(0, 8240, '\P{^Line_Break= glue}', "");
Error('\p{Lb:  :=GL}');
Error('\P{Lb:  :=GL}');
Expect(1, 8239, '\p{Lb=gl}', "");
Expect(0, 8239, '\p{^Lb=gl}', "");
Expect(0, 8239, '\P{Lb=gl}', "");
Expect(1, 8239, '\P{^Lb=gl}', "");
Expect(0, 8240, '\p{Lb=gl}', "");
Expect(1, 8240, '\p{^Lb=gl}', "");
Expect(1, 8240, '\P{Lb=gl}', "");
Expect(0, 8240, '\P{^Lb=gl}', "");
Expect(1, 8239, '\p{Lb=	_GL}', "");
Expect(0, 8239, '\p{^Lb=	_GL}', "");
Expect(0, 8239, '\P{Lb=	_GL}', "");
Expect(1, 8239, '\P{^Lb=	_GL}', "");
Expect(0, 8240, '\p{Lb=	_GL}', "");
Expect(1, 8240, '\p{^Lb=	_GL}', "");
Expect(1, 8240, '\P{Lb=	_GL}', "");
Expect(0, 8240, '\P{^Lb=	_GL}', "");
Error('\p{Is_Line_Break=-_Glue:=}');
Error('\P{Is_Line_Break=-_Glue:=}');
Expect(1, 8239, '\p{Is_Line_Break=glue}', "");
Expect(0, 8239, '\p{^Is_Line_Break=glue}', "");
Expect(0, 8239, '\P{Is_Line_Break=glue}', "");
Expect(1, 8239, '\P{^Is_Line_Break=glue}', "");
Expect(0, 8240, '\p{Is_Line_Break=glue}', "");
Expect(1, 8240, '\p{^Is_Line_Break=glue}', "");
Expect(1, 8240, '\P{Is_Line_Break=glue}', "");
Expect(0, 8240, '\P{^Is_Line_Break=glue}', "");
Expect(1, 8239, '\p{Is_Line_Break=_ Glue}', "");
Expect(0, 8239, '\p{^Is_Line_Break=_ Glue}', "");
Expect(0, 8239, '\P{Is_Line_Break=_ Glue}', "");
Expect(1, 8239, '\P{^Is_Line_Break=_ Glue}', "");
Expect(0, 8240, '\p{Is_Line_Break=_ Glue}', "");
Expect(1, 8240, '\p{^Is_Line_Break=_ Glue}', "");
Expect(1, 8240, '\P{Is_Line_Break=_ Glue}', "");
Expect(0, 8240, '\P{^Is_Line_Break=_ Glue}', "");
Error('\p{Is_Lb=_ GL:=}');
Error('\P{Is_Lb=_ GL:=}');
Expect(1, 8239, '\p{Is_Lb=gl}', "");
Expect(0, 8239, '\p{^Is_Lb=gl}', "");
Expect(0, 8239, '\P{Is_Lb=gl}', "");
Expect(1, 8239, '\P{^Is_Lb=gl}', "");
Expect(0, 8240, '\p{Is_Lb=gl}', "");
Expect(1, 8240, '\p{^Is_Lb=gl}', "");
Expect(1, 8240, '\P{Is_Lb=gl}', "");
Expect(0, 8240, '\P{^Is_Lb=gl}', "");
Expect(1, 8239, '\p{Is_Lb=	-GL}', "");
Expect(0, 8239, '\p{^Is_Lb=	-GL}', "");
Expect(0, 8239, '\P{Is_Lb=	-GL}', "");
Expect(1, 8239, '\P{^Is_Lb=	-GL}', "");
Expect(0, 8240, '\p{Is_Lb=	-GL}', "");
Expect(1, 8240, '\p{^Is_Lb=	-GL}', "");
Expect(1, 8240, '\P{Is_Lb=	-GL}', "");
Expect(0, 8240, '\P{^Is_Lb=	-GL}', "");
Error('\p{Line_Break=/a/--H2}');
Error('\P{Line_Break=/a/--H2}');
Expect(1, 55176, '\p{Line_Break=h2}', "");
Expect(0, 55176, '\p{^Line_Break=h2}', "");
Expect(0, 55176, '\P{Line_Break=h2}', "");
Expect(1, 55176, '\P{^Line_Break=h2}', "");
Expect(0, 55177, '\p{Line_Break=h2}', "");
Expect(1, 55177, '\p{^Line_Break=h2}', "");
Expect(1, 55177, '\P{Line_Break=h2}', "");
Expect(0, 55177, '\P{^Line_Break=h2}', "");
Expect(1, 55176, '\p{Line_Break=-_H2}', "");
Expect(0, 55176, '\p{^Line_Break=-_H2}', "");
Expect(0, 55176, '\P{Line_Break=-_H2}', "");
Expect(1, 55176, '\P{^Line_Break=-_H2}', "");
Expect(0, 55177, '\p{Line_Break=-_H2}', "");
Expect(1, 55177, '\p{^Line_Break=-_H2}', "");
Expect(1, 55177, '\P{Line_Break=-_H2}', "");
Expect(0, 55177, '\P{^Line_Break=-_H2}', "");
Error('\p{Lb:   -:=H2}');
Error('\P{Lb:   -:=H2}');
Expect(1, 55176, '\p{Lb=h2}', "");
Expect(0, 55176, '\p{^Lb=h2}', "");
Expect(0, 55176, '\P{Lb=h2}', "");
Expect(1, 55176, '\P{^Lb=h2}', "");
Expect(0, 55177, '\p{Lb=h2}', "");
Expect(1, 55177, '\p{^Lb=h2}', "");
Expect(1, 55177, '\P{Lb=h2}', "");
Expect(0, 55177, '\P{^Lb=h2}', "");
Expect(1, 55176, '\p{Lb=- H2}', "");
Expect(0, 55176, '\p{^Lb=- H2}', "");
Expect(0, 55176, '\P{Lb=- H2}', "");
Expect(1, 55176, '\P{^Lb=- H2}', "");
Expect(0, 55177, '\p{Lb=- H2}', "");
Expect(1, 55177, '\p{^Lb=- H2}', "");
Expect(1, 55177, '\P{Lb=- H2}', "");
Expect(0, 55177, '\P{^Lb=- H2}', "");
Error('\p{Is_Line_Break:	_	h2:=}');
Error('\P{Is_Line_Break:	_	h2:=}');
Expect(1, 55176, '\p{Is_Line_Break=h2}', "");
Expect(0, 55176, '\p{^Is_Line_Break=h2}', "");
Expect(0, 55176, '\P{Is_Line_Break=h2}', "");
Expect(1, 55176, '\P{^Is_Line_Break=h2}', "");
Expect(0, 55177, '\p{Is_Line_Break=h2}', "");
Expect(1, 55177, '\p{^Is_Line_Break=h2}', "");
Expect(1, 55177, '\P{Is_Line_Break=h2}', "");
Expect(0, 55177, '\P{^Is_Line_Break=h2}', "");
Expect(1, 55176, '\p{Is_Line_Break=	H2}', "");
Expect(0, 55176, '\p{^Is_Line_Break=	H2}', "");
Expect(0, 55176, '\P{Is_Line_Break=	H2}', "");
Expect(1, 55176, '\P{^Is_Line_Break=	H2}', "");
Expect(0, 55177, '\p{Is_Line_Break=	H2}', "");
Expect(1, 55177, '\p{^Is_Line_Break=	H2}', "");
Expect(1, 55177, '\P{Is_Line_Break=	H2}', "");
Expect(0, 55177, '\P{^Is_Line_Break=	H2}', "");
Error('\p{Is_Lb=		H2/a/}');
Error('\P{Is_Lb=		H2/a/}');
Expect(1, 55176, '\p{Is_Lb=h2}', "");
Expect(0, 55176, '\p{^Is_Lb=h2}', "");
Expect(0, 55176, '\P{Is_Lb=h2}', "");
Expect(1, 55176, '\P{^Is_Lb=h2}', "");
Expect(0, 55177, '\p{Is_Lb=h2}', "");
Expect(1, 55177, '\p{^Is_Lb=h2}', "");
Expect(1, 55177, '\P{Is_Lb=h2}', "");
Expect(0, 55177, '\P{^Is_Lb=h2}', "");
Expect(1, 55176, '\p{Is_Lb=_ H2}', "");
Expect(0, 55176, '\p{^Is_Lb=_ H2}', "");
Expect(0, 55176, '\P{Is_Lb=_ H2}', "");
Expect(1, 55176, '\P{^Is_Lb=_ H2}', "");
Expect(0, 55177, '\p{Is_Lb=_ H2}', "");
Expect(1, 55177, '\p{^Is_Lb=_ H2}', "");
Expect(1, 55177, '\P{Is_Lb=_ H2}', "");
Expect(0, 55177, '\P{^Is_Lb=_ H2}', "");
Error('\p{Line_Break=/a/-	H3}');
Error('\P{Line_Break=/a/-	H3}');
Expect(1, 55203, '\p{Line_Break: h3}', "");
Expect(0, 55203, '\p{^Line_Break: h3}', "");
Expect(0, 55203, '\P{Line_Break: h3}', "");
Expect(1, 55203, '\P{^Line_Break: h3}', "");
Expect(0, 55204, '\p{Line_Break: h3}', "");
Expect(1, 55204, '\p{^Line_Break: h3}', "");
Expect(1, 55204, '\P{Line_Break: h3}', "");
Expect(0, 55204, '\P{^Line_Break: h3}', "");
Expect(1, 55203, '\p{Line_Break=_	H3}', "");
Expect(0, 55203, '\p{^Line_Break=_	H3}', "");
Expect(0, 55203, '\P{Line_Break=_	H3}', "");
Expect(1, 55203, '\P{^Line_Break=_	H3}', "");
Expect(0, 55204, '\p{Line_Break=_	H3}', "");
Expect(1, 55204, '\p{^Line_Break=_	H3}', "");
Expect(1, 55204, '\P{Line_Break=_	H3}', "");
Expect(0, 55204, '\P{^Line_Break=_	H3}', "");
Error('\p{Lb=_h3:=}');
Error('\P{Lb=_h3:=}');
Expect(1, 55203, '\p{Lb=h3}', "");
Expect(0, 55203, '\p{^Lb=h3}', "");
Expect(0, 55203, '\P{Lb=h3}', "");
Expect(1, 55203, '\P{^Lb=h3}', "");
Expect(0, 55204, '\p{Lb=h3}', "");
Expect(1, 55204, '\p{^Lb=h3}', "");
Expect(1, 55204, '\P{Lb=h3}', "");
Expect(0, 55204, '\P{^Lb=h3}', "");
Expect(1, 55203, '\p{Lb:    H3}', "");
Expect(0, 55203, '\p{^Lb:    H3}', "");
Expect(0, 55203, '\P{Lb:    H3}', "");
Expect(1, 55203, '\P{^Lb:    H3}', "");
Expect(0, 55204, '\p{Lb:    H3}', "");
Expect(1, 55204, '\p{^Lb:    H3}', "");
Expect(1, 55204, '\P{Lb:    H3}', "");
Expect(0, 55204, '\P{^Lb:    H3}', "");
Error('\p{Is_Line_Break=/a/_H3}');
Error('\P{Is_Line_Break=/a/_H3}');
Expect(1, 55203, '\p{Is_Line_Break:h3}', "");
Expect(0, 55203, '\p{^Is_Line_Break:h3}', "");
Expect(0, 55203, '\P{Is_Line_Break:h3}', "");
Expect(1, 55203, '\P{^Is_Line_Break:h3}', "");
Expect(0, 55204, '\p{Is_Line_Break:h3}', "");
Expect(1, 55204, '\p{^Is_Line_Break:h3}', "");
Expect(1, 55204, '\P{Is_Line_Break:h3}', "");
Expect(0, 55204, '\P{^Is_Line_Break:h3}', "");
Expect(1, 55203, '\p{Is_Line_Break=_ H3}', "");
Expect(0, 55203, '\p{^Is_Line_Break=_ H3}', "");
Expect(0, 55203, '\P{Is_Line_Break=_ H3}', "");
Expect(1, 55203, '\P{^Is_Line_Break=_ H3}', "");
Expect(0, 55204, '\p{Is_Line_Break=_ H3}', "");
Expect(1, 55204, '\p{^Is_Line_Break=_ H3}', "");
Expect(1, 55204, '\P{Is_Line_Break=_ H3}', "");
Expect(0, 55204, '\P{^Is_Line_Break=_ H3}', "");
Error('\p{Is_Lb=-:=H3}');
Error('\P{Is_Lb=-:=H3}');
Expect(1, 55203, '\p{Is_Lb=h3}', "");
Expect(0, 55203, '\p{^Is_Lb=h3}', "");
Expect(0, 55203, '\P{Is_Lb=h3}', "");
Expect(1, 55203, '\P{^Is_Lb=h3}', "");
Expect(0, 55204, '\p{Is_Lb=h3}', "");
Expect(1, 55204, '\p{^Is_Lb=h3}', "");
Expect(1, 55204, '\P{Is_Lb=h3}', "");
Expect(0, 55204, '\P{^Is_Lb=h3}', "");
Expect(1, 55203, '\p{Is_Lb=_-H3}', "");
Expect(0, 55203, '\p{^Is_Lb=_-H3}', "");
Expect(0, 55203, '\P{Is_Lb=_-H3}', "");
Expect(1, 55203, '\P{^Is_Lb=_-H3}', "");
Expect(0, 55204, '\p{Is_Lb=_-H3}', "");
Expect(1, 55204, '\p{^Is_Lb=_-H3}', "");
Expect(1, 55204, '\P{Is_Lb=_-H3}', "");
Expect(0, 55204, '\P{^Is_Lb=_-H3}', "");
Error('\p{Line_Break=/a/_ HEBREW_letter}');
Error('\P{Line_Break=/a/_ HEBREW_letter}');
Expect(1, 64335, '\p{Line_Break=hebrewletter}', "");
Expect(0, 64335, '\p{^Line_Break=hebrewletter}', "");
Expect(0, 64335, '\P{Line_Break=hebrewletter}', "");
Expect(1, 64335, '\P{^Line_Break=hebrewletter}', "");
Expect(0, 64336, '\p{Line_Break=hebrewletter}', "");
Expect(1, 64336, '\p{^Line_Break=hebrewletter}', "");
Expect(1, 64336, '\P{Line_Break=hebrewletter}', "");
Expect(0, 64336, '\P{^Line_Break=hebrewletter}', "");
Expect(1, 64335, '\p{Line_Break:   _Hebrew_Letter}', "");
Expect(0, 64335, '\p{^Line_Break:   _Hebrew_Letter}', "");
Expect(0, 64335, '\P{Line_Break:   _Hebrew_Letter}', "");
Expect(1, 64335, '\P{^Line_Break:   _Hebrew_Letter}', "");
Expect(0, 64336, '\p{Line_Break:   _Hebrew_Letter}', "");
Expect(1, 64336, '\p{^Line_Break:   _Hebrew_Letter}', "");
Expect(1, 64336, '\P{Line_Break:   _Hebrew_Letter}', "");
Expect(0, 64336, '\P{^Line_Break:   _Hebrew_Letter}', "");
Error('\p{Lb=--hl:=}');
Error('\P{Lb=--hl:=}');
Expect(1, 64335, '\p{Lb:   hl}', "");
Expect(0, 64335, '\p{^Lb:   hl}', "");
Expect(0, 64335, '\P{Lb:   hl}', "");
Expect(1, 64335, '\P{^Lb:   hl}', "");
Expect(0, 64336, '\p{Lb:   hl}', "");
Expect(1, 64336, '\p{^Lb:   hl}', "");
Expect(1, 64336, '\P{Lb:   hl}', "");
Expect(0, 64336, '\P{^Lb:   hl}', "");
Expect(1, 64335, '\p{Lb=--HL}', "");
Expect(0, 64335, '\p{^Lb=--HL}', "");
Expect(0, 64335, '\P{Lb=--HL}', "");
Expect(1, 64335, '\P{^Lb=--HL}', "");
Expect(0, 64336, '\p{Lb=--HL}', "");
Expect(1, 64336, '\p{^Lb=--HL}', "");
Expect(1, 64336, '\P{Lb=--HL}', "");
Expect(0, 64336, '\P{^Lb=--HL}', "");
Error('\p{Is_Line_Break=--hebrew_Letter/a/}');
Error('\P{Is_Line_Break=--hebrew_Letter/a/}');
Expect(1, 64335, '\p{Is_Line_Break=hebrewletter}', "");
Expect(0, 64335, '\p{^Is_Line_Break=hebrewletter}', "");
Expect(0, 64335, '\P{Is_Line_Break=hebrewletter}', "");
Expect(1, 64335, '\P{^Is_Line_Break=hebrewletter}', "");
Expect(0, 64336, '\p{Is_Line_Break=hebrewletter}', "");
Expect(1, 64336, '\p{^Is_Line_Break=hebrewletter}', "");
Expect(1, 64336, '\P{Is_Line_Break=hebrewletter}', "");
Expect(0, 64336, '\P{^Is_Line_Break=hebrewletter}', "");
Expect(1, 64335, '\p{Is_Line_Break=-HEBREW_letter}', "");
Expect(0, 64335, '\p{^Is_Line_Break=-HEBREW_letter}', "");
Expect(0, 64335, '\P{Is_Line_Break=-HEBREW_letter}', "");
Expect(1, 64335, '\P{^Is_Line_Break=-HEBREW_letter}', "");
Expect(0, 64336, '\p{Is_Line_Break=-HEBREW_letter}', "");
Expect(1, 64336, '\p{^Is_Line_Break=-HEBREW_letter}', "");
Expect(1, 64336, '\P{Is_Line_Break=-HEBREW_letter}', "");
Expect(0, 64336, '\P{^Is_Line_Break=-HEBREW_letter}', "");
Error('\p{Is_Lb=/a/ hl}');
Error('\P{Is_Lb=/a/ hl}');
Expect(1, 64335, '\p{Is_Lb=hl}', "");
Expect(0, 64335, '\p{^Is_Lb=hl}', "");
Expect(0, 64335, '\P{Is_Lb=hl}', "");
Expect(1, 64335, '\P{^Is_Lb=hl}', "");
Expect(0, 64336, '\p{Is_Lb=hl}', "");
Expect(1, 64336, '\p{^Is_Lb=hl}', "");
Expect(1, 64336, '\P{Is_Lb=hl}', "");
Expect(0, 64336, '\P{^Is_Lb=hl}', "");
Expect(1, 64335, '\p{Is_Lb=_	HL}', "");
Expect(0, 64335, '\p{^Is_Lb=_	HL}', "");
Expect(0, 64335, '\P{Is_Lb=_	HL}', "");
Expect(1, 64335, '\P{^Is_Lb=_	HL}', "");
Expect(0, 64336, '\p{Is_Lb=_	HL}', "");
Expect(1, 64336, '\p{^Is_Lb=_	HL}', "");
Expect(1, 64336, '\P{Is_Lb=_	HL}', "");
Expect(0, 64336, '\P{^Is_Lb=_	HL}', "");
Error('\p{Line_Break= :=Hyphen}');
Error('\P{Line_Break= :=Hyphen}');
Expect(1, 45, '\p{Line_Break=hyphen}', "");
Expect(0, 45, '\p{^Line_Break=hyphen}', "");
Expect(0, 45, '\P{Line_Break=hyphen}', "");
Expect(1, 45, '\P{^Line_Break=hyphen}', "");
Expect(0, 46, '\p{Line_Break=hyphen}', "");
Expect(1, 46, '\p{^Line_Break=hyphen}', "");
Expect(1, 46, '\P{Line_Break=hyphen}', "");
Expect(0, 46, '\P{^Line_Break=hyphen}', "");
Expect(1, 45, '\p{Line_Break=	Hyphen}', "");
Expect(0, 45, '\p{^Line_Break=	Hyphen}', "");
Expect(0, 45, '\P{Line_Break=	Hyphen}', "");
Expect(1, 45, '\P{^Line_Break=	Hyphen}', "");
Expect(0, 46, '\p{Line_Break=	Hyphen}', "");
Expect(1, 46, '\p{^Line_Break=	Hyphen}', "");
Expect(1, 46, '\P{Line_Break=	Hyphen}', "");
Expect(0, 46, '\P{^Line_Break=	Hyphen}', "");
Error('\p{Lb=__hy/a/}');
Error('\P{Lb=__hy/a/}');
Expect(1, 45, '\p{Lb=hy}', "");
Expect(0, 45, '\p{^Lb=hy}', "");
Expect(0, 45, '\P{Lb=hy}', "");
Expect(1, 45, '\P{^Lb=hy}', "");
Expect(0, 46, '\p{Lb=hy}', "");
Expect(1, 46, '\p{^Lb=hy}', "");
Expect(1, 46, '\P{Lb=hy}', "");
Expect(0, 46, '\P{^Lb=hy}', "");
Expect(1, 45, '\p{Lb:	--HY}', "");
Expect(0, 45, '\p{^Lb:	--HY}', "");
Expect(0, 45, '\P{Lb:	--HY}', "");
Expect(1, 45, '\P{^Lb:	--HY}', "");
Expect(0, 46, '\p{Lb:	--HY}', "");
Expect(1, 46, '\p{^Lb:	--HY}', "");
Expect(1, 46, '\P{Lb:	--HY}', "");
Expect(0, 46, '\P{^Lb:	--HY}', "");
Error('\p{Is_Line_Break=/a/	 Hyphen}');
Error('\P{Is_Line_Break=/a/	 Hyphen}');
Expect(1, 45, '\p{Is_Line_Break=hyphen}', "");
Expect(0, 45, '\p{^Is_Line_Break=hyphen}', "");
Expect(0, 45, '\P{Is_Line_Break=hyphen}', "");
Expect(1, 45, '\P{^Is_Line_Break=hyphen}', "");
Expect(0, 46, '\p{Is_Line_Break=hyphen}', "");
Expect(1, 46, '\p{^Is_Line_Break=hyphen}', "");
Expect(1, 46, '\P{Is_Line_Break=hyphen}', "");
Expect(0, 46, '\P{^Is_Line_Break=hyphen}', "");
Expect(1, 45, '\p{Is_Line_Break=	Hyphen}', "");
Expect(0, 45, '\p{^Is_Line_Break=	Hyphen}', "");
Expect(0, 45, '\P{Is_Line_Break=	Hyphen}', "");
Expect(1, 45, '\P{^Is_Line_Break=	Hyphen}', "");
Expect(0, 46, '\p{Is_Line_Break=	Hyphen}', "");
Expect(1, 46, '\p{^Is_Line_Break=	Hyphen}', "");
Expect(1, 46, '\P{Is_Line_Break=	Hyphen}', "");
Expect(0, 46, '\P{^Is_Line_Break=	Hyphen}', "");
Error('\p{Is_Lb=	-hy/a/}');
Error('\P{Is_Lb=	-hy/a/}');
Expect(1, 45, '\p{Is_Lb=hy}', "");
Expect(0, 45, '\p{^Is_Lb=hy}', "");
Expect(0, 45, '\P{Is_Lb=hy}', "");
Expect(1, 45, '\P{^Is_Lb=hy}', "");
Expect(0, 46, '\p{Is_Lb=hy}', "");
Expect(1, 46, '\p{^Is_Lb=hy}', "");
Expect(1, 46, '\P{Is_Lb=hy}', "");
Expect(0, 46, '\P{^Is_Lb=hy}', "");
Expect(1, 45, '\p{Is_Lb=  hy}', "");
Expect(0, 45, '\p{^Is_Lb=  hy}', "");
Expect(0, 45, '\P{Is_Lb=  hy}', "");
Expect(1, 45, '\P{^Is_Lb=  hy}', "");
Expect(0, 46, '\p{Is_Lb=  hy}', "");
Expect(1, 46, '\p{^Is_Lb=  hy}', "");
Expect(1, 46, '\P{Is_Lb=  hy}', "");
Expect(0, 46, '\P{^Is_Lb=  hy}', "");
Error('\p{Line_Break=:= IDEOGRAPHIC}');
Error('\P{Line_Break=:= IDEOGRAPHIC}');
Expect(1, 262141, '\p{Line_Break=ideographic}', "");
Expect(0, 262141, '\p{^Line_Break=ideographic}', "");
Expect(0, 262141, '\P{Line_Break=ideographic}', "");
Expect(1, 262141, '\P{^Line_Break=ideographic}', "");
Expect(0, 262144, '\p{Line_Break=ideographic}', "");
Expect(1, 262144, '\p{^Line_Break=ideographic}', "");
Expect(1, 262144, '\P{Line_Break=ideographic}', "");
Expect(0, 262144, '\P{^Line_Break=ideographic}', "");
Expect(1, 262141, '\p{Line_Break=	ideographic}', "");
Expect(0, 262141, '\p{^Line_Break=	ideographic}', "");
Expect(0, 262141, '\P{Line_Break=	ideographic}', "");
Expect(1, 262141, '\P{^Line_Break=	ideographic}', "");
Expect(0, 262144, '\p{Line_Break=	ideographic}', "");
Expect(1, 262144, '\p{^Line_Break=	ideographic}', "");
Expect(1, 262144, '\P{Line_Break=	ideographic}', "");
Expect(0, 262144, '\P{^Line_Break=	ideographic}', "");
Error('\p{Lb= :=ID}');
Error('\P{Lb= :=ID}');
Expect(1, 262141, '\p{Lb=id}', "");
Expect(0, 262141, '\p{^Lb=id}', "");
Expect(0, 262141, '\P{Lb=id}', "");
Expect(1, 262141, '\P{^Lb=id}', "");
Expect(0, 262144, '\p{Lb=id}', "");
Expect(1, 262144, '\p{^Lb=id}', "");
Expect(1, 262144, '\P{Lb=id}', "");
Expect(0, 262144, '\P{^Lb=id}', "");
Expect(1, 262141, '\p{Lb=  ID}', "");
Expect(0, 262141, '\p{^Lb=  ID}', "");
Expect(0, 262141, '\P{Lb=  ID}', "");
Expect(1, 262141, '\P{^Lb=  ID}', "");
Expect(0, 262144, '\p{Lb=  ID}', "");
Expect(1, 262144, '\p{^Lb=  ID}', "");
Expect(1, 262144, '\P{Lb=  ID}', "");
Expect(0, 262144, '\P{^Lb=  ID}', "");
Error('\p{Is_Line_Break=/a/		Ideographic}');
Error('\P{Is_Line_Break=/a/		Ideographic}');
Expect(1, 262141, '\p{Is_Line_Break=ideographic}', "");
Expect(0, 262141, '\p{^Is_Line_Break=ideographic}', "");
Expect(0, 262141, '\P{Is_Line_Break=ideographic}', "");
Expect(1, 262141, '\P{^Is_Line_Break=ideographic}', "");
Expect(0, 262144, '\p{Is_Line_Break=ideographic}', "");
Expect(1, 262144, '\p{^Is_Line_Break=ideographic}', "");
Expect(1, 262144, '\P{Is_Line_Break=ideographic}', "");
Expect(0, 262144, '\P{^Is_Line_Break=ideographic}', "");
Expect(1, 262141, '\p{Is_Line_Break= Ideographic}', "");
Expect(0, 262141, '\p{^Is_Line_Break= Ideographic}', "");
Expect(0, 262141, '\P{Is_Line_Break= Ideographic}', "");
Expect(1, 262141, '\P{^Is_Line_Break= Ideographic}', "");
Expect(0, 262144, '\p{Is_Line_Break= Ideographic}', "");
Expect(1, 262144, '\p{^Is_Line_Break= Ideographic}', "");
Expect(1, 262144, '\P{Is_Line_Break= Ideographic}', "");
Expect(0, 262144, '\P{^Is_Line_Break= Ideographic}', "");
Error('\p{Is_Lb=	:=ID}');
Error('\P{Is_Lb=	:=ID}');
Expect(1, 262141, '\p{Is_Lb=id}', "");
Expect(0, 262141, '\p{^Is_Lb=id}', "");
Expect(0, 262141, '\P{Is_Lb=id}', "");
Expect(1, 262141, '\P{^Is_Lb=id}', "");
Expect(0, 262144, '\p{Is_Lb=id}', "");
Expect(1, 262144, '\p{^Is_Lb=id}', "");
Expect(1, 262144, '\P{Is_Lb=id}', "");
Expect(0, 262144, '\P{^Is_Lb=id}', "");
Expect(1, 262141, '\p{Is_Lb= _ID}', "");
Expect(0, 262141, '\p{^Is_Lb= _ID}', "");
Expect(0, 262141, '\P{Is_Lb= _ID}', "");
Expect(1, 262141, '\P{^Is_Lb= _ID}', "");
Expect(0, 262144, '\p{Is_Lb= _ID}', "");
Expect(1, 262144, '\p{^Is_Lb= _ID}', "");
Expect(1, 262144, '\P{Is_Lb= _ID}', "");
Expect(0, 262144, '\P{^Is_Lb= _ID}', "");
Error('\p{Line_Break=:=Inseparable}');
Error('\P{Line_Break=:=Inseparable}');
Expect(1, 65049, '\p{Line_Break=inseparable}', "");
Expect(0, 65049, '\p{^Line_Break=inseparable}', "");
Expect(0, 65049, '\P{Line_Break=inseparable}', "");
Expect(1, 65049, '\P{^Line_Break=inseparable}', "");
Expect(0, 65050, '\p{Line_Break=inseparable}', "");
Expect(1, 65050, '\p{^Line_Break=inseparable}', "");
Expect(1, 65050, '\P{Line_Break=inseparable}', "");
Expect(0, 65050, '\P{^Line_Break=inseparable}', "");
Expect(1, 65049, '\p{Line_Break=_-Inseparable}', "");
Expect(0, 65049, '\p{^Line_Break=_-Inseparable}', "");
Expect(0, 65049, '\P{Line_Break=_-Inseparable}', "");
Expect(1, 65049, '\P{^Line_Break=_-Inseparable}', "");
Expect(0, 65050, '\p{Line_Break=_-Inseparable}', "");
Expect(1, 65050, '\p{^Line_Break=_-Inseparable}', "");
Expect(1, 65050, '\P{Line_Break=_-Inseparable}', "");
Expect(0, 65050, '\P{^Line_Break=_-Inseparable}', "");
Error('\p{Lb=_IN:=}');
Error('\P{Lb=_IN:=}');
Expect(1, 65049, '\p{Lb=in}', "");
Expect(0, 65049, '\p{^Lb=in}', "");
Expect(0, 65049, '\P{Lb=in}', "");
Expect(1, 65049, '\P{^Lb=in}', "");
Expect(0, 65050, '\p{Lb=in}', "");
Expect(1, 65050, '\p{^Lb=in}', "");
Expect(1, 65050, '\P{Lb=in}', "");
Expect(0, 65050, '\P{^Lb=in}', "");
Expect(1, 65049, '\p{Lb=-IN}', "");
Expect(0, 65049, '\p{^Lb=-IN}', "");
Expect(0, 65049, '\P{Lb=-IN}', "");
Expect(1, 65049, '\P{^Lb=-IN}', "");
Expect(0, 65050, '\p{Lb=-IN}', "");
Expect(1, 65050, '\p{^Lb=-IN}', "");
Expect(1, 65050, '\P{Lb=-IN}', "");
Expect(0, 65050, '\P{^Lb=-IN}', "");
Error('\p{Is_Line_Break=/a/- INSEPERABLE}');
Error('\P{Is_Line_Break=/a/- INSEPERABLE}');
Expect(1, 65049, '\p{Is_Line_Break=inseperable}', "");
Expect(0, 65049, '\p{^Is_Line_Break=inseperable}', "");
Expect(0, 65049, '\P{Is_Line_Break=inseperable}', "");
Expect(1, 65049, '\P{^Is_Line_Break=inseperable}', "");
Expect(0, 65050, '\p{Is_Line_Break=inseperable}', "");
Expect(1, 65050, '\p{^Is_Line_Break=inseperable}', "");
Expect(1, 65050, '\P{Is_Line_Break=inseperable}', "");
Expect(0, 65050, '\P{^Is_Line_Break=inseperable}', "");
Expect(1, 65049, '\p{Is_Line_Break= Inseperable}', "");
Expect(0, 65049, '\p{^Is_Line_Break= Inseperable}', "");
Expect(0, 65049, '\P{Is_Line_Break= Inseperable}', "");
Expect(1, 65049, '\P{^Is_Line_Break= Inseperable}', "");
Expect(0, 65050, '\p{Is_Line_Break= Inseperable}', "");
Expect(1, 65050, '\p{^Is_Line_Break= Inseperable}', "");
Expect(1, 65050, '\P{Is_Line_Break= Inseperable}', "");
Expect(0, 65050, '\P{^Is_Line_Break= Inseperable}', "");
Error('\p{Is_Lb=:=_Inseparable}');
Error('\P{Is_Lb=:=_Inseparable}');
Expect(1, 65049, '\p{Is_Lb=inseparable}', "");
Expect(0, 65049, '\p{^Is_Lb=inseparable}', "");
Expect(0, 65049, '\P{Is_Lb=inseparable}', "");
Expect(1, 65049, '\P{^Is_Lb=inseparable}', "");
Expect(0, 65050, '\p{Is_Lb=inseparable}', "");
Expect(1, 65050, '\p{^Is_Lb=inseparable}', "");
Expect(1, 65050, '\P{Is_Lb=inseparable}', "");
Expect(0, 65050, '\P{^Is_Lb=inseparable}', "");
Expect(1, 65049, '\p{Is_Lb= Inseparable}', "");
Expect(0, 65049, '\p{^Is_Lb= Inseparable}', "");
Expect(0, 65049, '\P{Is_Lb= Inseparable}', "");
Expect(1, 65049, '\P{^Is_Lb= Inseparable}', "");
Expect(0, 65050, '\p{Is_Lb= Inseparable}', "");
Expect(1, 65050, '\p{^Is_Lb= Inseparable}', "");
Expect(1, 65050, '\P{Is_Lb= Inseparable}', "");
Expect(0, 65050, '\P{^Is_Lb= Inseparable}', "");
Error('\p{Line_Break=/a/		INFIX_numeric}');
Error('\P{Line_Break=/a/		INFIX_numeric}');
Expect(1, 65044, '\p{Line_Break=infixnumeric}', "");
Expect(0, 65044, '\p{^Line_Break=infixnumeric}', "");
Expect(0, 65044, '\P{Line_Break=infixnumeric}', "");
Expect(1, 65044, '\P{^Line_Break=infixnumeric}', "");
Expect(0, 65045, '\p{Line_Break=infixnumeric}', "");
Expect(1, 65045, '\p{^Line_Break=infixnumeric}', "");
Expect(1, 65045, '\P{Line_Break=infixnumeric}', "");
Expect(0, 65045, '\P{^Line_Break=infixnumeric}', "");
Expect(1, 65044, '\p{Line_Break:	 Infix_NUMERIC}', "");
Expect(0, 65044, '\p{^Line_Break:	 Infix_NUMERIC}', "");
Expect(0, 65044, '\P{Line_Break:	 Infix_NUMERIC}', "");
Expect(1, 65044, '\P{^Line_Break:	 Infix_NUMERIC}', "");
Expect(0, 65045, '\p{Line_Break:	 Infix_NUMERIC}', "");
Expect(1, 65045, '\p{^Line_Break:	 Infix_NUMERIC}', "");
Expect(1, 65045, '\P{Line_Break:	 Infix_NUMERIC}', "");
Expect(0, 65045, '\P{^Line_Break:	 Infix_NUMERIC}', "");
Error('\p{Lb= -IS/a/}');
Error('\P{Lb= -IS/a/}');
Expect(1, 65044, '\p{Lb=is}', "");
Expect(0, 65044, '\p{^Lb=is}', "");
Expect(0, 65044, '\P{Lb=is}', "");
Expect(1, 65044, '\P{^Lb=is}', "");
Expect(0, 65045, '\p{Lb=is}', "");
Expect(1, 65045, '\p{^Lb=is}', "");
Expect(1, 65045, '\P{Lb=is}', "");
Expect(0, 65045, '\P{^Lb=is}', "");
Expect(1, 65044, '\p{Lb=_-IS}', "");
Expect(0, 65044, '\p{^Lb=_-IS}', "");
Expect(0, 65044, '\P{Lb=_-IS}', "");
Expect(1, 65044, '\P{^Lb=_-IS}', "");
Expect(0, 65045, '\p{Lb=_-IS}', "");
Expect(1, 65045, '\p{^Lb=_-IS}', "");
Expect(1, 65045, '\P{Lb=_-IS}', "");
Expect(0, 65045, '\P{^Lb=_-IS}', "");
Error('\p{Is_Line_Break= :=Infix_Numeric}');
Error('\P{Is_Line_Break= :=Infix_Numeric}');
Expect(1, 65044, '\p{Is_Line_Break=infixnumeric}', "");
Expect(0, 65044, '\p{^Is_Line_Break=infixnumeric}', "");
Expect(0, 65044, '\P{Is_Line_Break=infixnumeric}', "");
Expect(1, 65044, '\P{^Is_Line_Break=infixnumeric}', "");
Expect(0, 65045, '\p{Is_Line_Break=infixnumeric}', "");
Expect(1, 65045, '\p{^Is_Line_Break=infixnumeric}', "");
Expect(1, 65045, '\P{Is_Line_Break=infixnumeric}', "");
Expect(0, 65045, '\P{^Is_Line_Break=infixnumeric}', "");
Expect(1, 65044, '\p{Is_Line_Break=	-Infix_Numeric}', "");
Expect(0, 65044, '\p{^Is_Line_Break=	-Infix_Numeric}', "");
Expect(0, 65044, '\P{Is_Line_Break=	-Infix_Numeric}', "");
Expect(1, 65044, '\P{^Is_Line_Break=	-Infix_Numeric}', "");
Expect(0, 65045, '\p{Is_Line_Break=	-Infix_Numeric}', "");
Expect(1, 65045, '\p{^Is_Line_Break=	-Infix_Numeric}', "");
Expect(1, 65045, '\P{Is_Line_Break=	-Infix_Numeric}', "");
Expect(0, 65045, '\P{^Is_Line_Break=	-Infix_Numeric}', "");
Error('\p{Is_Lb=-:=IS}');
Error('\P{Is_Lb=-:=IS}');
Expect(1, 65044, '\p{Is_Lb=is}', "");
Expect(0, 65044, '\p{^Is_Lb=is}', "");
Expect(0, 65044, '\P{Is_Lb=is}', "");
Expect(1, 65044, '\P{^Is_Lb=is}', "");
Expect(0, 65045, '\p{Is_Lb=is}', "");
Expect(1, 65045, '\p{^Is_Lb=is}', "");
Expect(1, 65045, '\P{Is_Lb=is}', "");
Expect(0, 65045, '\P{^Is_Lb=is}', "");
Expect(1, 65044, '\p{Is_Lb:   -_IS}', "");
Expect(0, 65044, '\p{^Is_Lb:   -_IS}', "");
Expect(0, 65044, '\P{Is_Lb:   -_IS}', "");
Expect(1, 65044, '\P{^Is_Lb:   -_IS}', "");
Expect(0, 65045, '\p{Is_Lb:   -_IS}', "");
Expect(1, 65045, '\p{^Is_Lb:   -_IS}', "");
Expect(1, 65045, '\P{Is_Lb:   -_IS}', "");
Expect(0, 65045, '\P{^Is_Lb:   -_IS}', "");
Error('\p{Line_Break=	 JL:=}');
Error('\P{Line_Break=	 JL:=}');
Expect(1, 43388, '\p{Line_Break=jl}', "");
Expect(0, 43388, '\p{^Line_Break=jl}', "");
Expect(0, 43388, '\P{Line_Break=jl}', "");
Expect(1, 43388, '\P{^Line_Break=jl}', "");
Expect(0, 43389, '\p{Line_Break=jl}', "");
Expect(1, 43389, '\p{^Line_Break=jl}', "");
Expect(1, 43389, '\P{Line_Break=jl}', "");
Expect(0, 43389, '\P{^Line_Break=jl}', "");
Expect(1, 43388, '\p{Line_Break=_JL}', "");
Expect(0, 43388, '\p{^Line_Break=_JL}', "");
Expect(0, 43388, '\P{Line_Break=_JL}', "");
Expect(1, 43388, '\P{^Line_Break=_JL}', "");
Expect(0, 43389, '\p{Line_Break=_JL}', "");
Expect(1, 43389, '\p{^Line_Break=_JL}', "");
Expect(1, 43389, '\P{Line_Break=_JL}', "");
Expect(0, 43389, '\P{^Line_Break=_JL}', "");
Error('\p{Lb=/a/	 JL}');
Error('\P{Lb=/a/	 JL}');
Expect(1, 43388, '\p{Lb=jl}', "");
Expect(0, 43388, '\p{^Lb=jl}', "");
Expect(0, 43388, '\P{Lb=jl}', "");
Expect(1, 43388, '\P{^Lb=jl}', "");
Expect(0, 43389, '\p{Lb=jl}', "");
Expect(1, 43389, '\p{^Lb=jl}', "");
Expect(1, 43389, '\P{Lb=jl}', "");
Expect(0, 43389, '\P{^Lb=jl}', "");
Expect(1, 43388, '\p{Lb=	 JL}', "");
Expect(0, 43388, '\p{^Lb=	 JL}', "");
Expect(0, 43388, '\P{Lb=	 JL}', "");
Expect(1, 43388, '\P{^Lb=	 JL}', "");
Expect(0, 43389, '\p{Lb=	 JL}', "");
Expect(1, 43389, '\p{^Lb=	 JL}', "");
Expect(1, 43389, '\P{Lb=	 JL}', "");
Expect(0, 43389, '\P{^Lb=	 JL}', "");
Error('\p{Is_Line_Break=:=		jl}');
Error('\P{Is_Line_Break=:=		jl}');
Expect(1, 43388, '\p{Is_Line_Break=jl}', "");
Expect(0, 43388, '\p{^Is_Line_Break=jl}', "");
Expect(0, 43388, '\P{Is_Line_Break=jl}', "");
Expect(1, 43388, '\P{^Is_Line_Break=jl}', "");
Expect(0, 43389, '\p{Is_Line_Break=jl}', "");
Expect(1, 43389, '\p{^Is_Line_Break=jl}', "");
Expect(1, 43389, '\P{Is_Line_Break=jl}', "");
Expect(0, 43389, '\P{^Is_Line_Break=jl}', "");
Expect(1, 43388, '\p{Is_Line_Break=_ jl}', "");
Expect(0, 43388, '\p{^Is_Line_Break=_ jl}', "");
Expect(0, 43388, '\P{Is_Line_Break=_ jl}', "");
Expect(1, 43388, '\P{^Is_Line_Break=_ jl}', "");
Expect(0, 43389, '\p{Is_Line_Break=_ jl}', "");
Expect(1, 43389, '\p{^Is_Line_Break=_ jl}', "");
Expect(1, 43389, '\P{Is_Line_Break=_ jl}', "");
Expect(0, 43389, '\P{^Is_Line_Break=_ jl}', "");
Error('\p{Is_Lb= jl/a/}');
Error('\P{Is_Lb= jl/a/}');
Expect(1, 43388, '\p{Is_Lb=jl}', "");
Expect(0, 43388, '\p{^Is_Lb=jl}', "");
Expect(0, 43388, '\P{Is_Lb=jl}', "");
Expect(1, 43388, '\P{^Is_Lb=jl}', "");
Expect(0, 43389, '\p{Is_Lb=jl}', "");
Expect(1, 43389, '\p{^Is_Lb=jl}', "");
Expect(1, 43389, '\P{Is_Lb=jl}', "");
Expect(0, 43389, '\P{^Is_Lb=jl}', "");
Expect(1, 43388, '\p{Is_Lb:-JL}', "");
Expect(0, 43388, '\p{^Is_Lb:-JL}', "");
Expect(0, 43388, '\P{Is_Lb:-JL}', "");
Expect(1, 43388, '\P{^Is_Lb:-JL}', "");
Expect(0, 43389, '\p{Is_Lb:-JL}', "");
Expect(1, 43389, '\p{^Is_Lb:-JL}', "");
Expect(1, 43389, '\P{Is_Lb:-JL}', "");
Expect(0, 43389, '\P{^Is_Lb:-JL}', "");
Error('\p{Line_Break=:= -JT}');
Error('\P{Line_Break=:= -JT}');
Expect(1, 55291, '\p{Line_Break:	jt}', "");
Expect(0, 55291, '\p{^Line_Break:	jt}', "");
Expect(0, 55291, '\P{Line_Break:	jt}', "");
Expect(1, 55291, '\P{^Line_Break:	jt}', "");
Expect(0, 55292, '\p{Line_Break:	jt}', "");
Expect(1, 55292, '\p{^Line_Break:	jt}', "");
Expect(1, 55292, '\P{Line_Break:	jt}', "");
Expect(0, 55292, '\P{^Line_Break:	jt}', "");
Expect(1, 55291, '\p{Line_Break=__JT}', "");
Expect(0, 55291, '\p{^Line_Break=__JT}', "");
Expect(0, 55291, '\P{Line_Break=__JT}', "");
Expect(1, 55291, '\P{^Line_Break=__JT}', "");
Expect(0, 55292, '\p{Line_Break=__JT}', "");
Expect(1, 55292, '\p{^Line_Break=__JT}', "");
Expect(1, 55292, '\P{Line_Break=__JT}', "");
Expect(0, 55292, '\P{^Line_Break=__JT}', "");
Error('\p{Lb=:=_-JT}');
Error('\P{Lb=:=_-JT}');
Expect(1, 55291, '\p{Lb=jt}', "");
Expect(0, 55291, '\p{^Lb=jt}', "");
Expect(0, 55291, '\P{Lb=jt}', "");
Expect(1, 55291, '\P{^Lb=jt}', "");
Expect(0, 55292, '\p{Lb=jt}', "");
Expect(1, 55292, '\p{^Lb=jt}', "");
Expect(1, 55292, '\P{Lb=jt}', "");
Expect(0, 55292, '\P{^Lb=jt}', "");
Expect(1, 55291, '\p{Lb=	_jt}', "");
Expect(0, 55291, '\p{^Lb=	_jt}', "");
Expect(0, 55291, '\P{Lb=	_jt}', "");
Expect(1, 55291, '\P{^Lb=	_jt}', "");
Expect(0, 55292, '\p{Lb=	_jt}', "");
Expect(1, 55292, '\p{^Lb=	_jt}', "");
Expect(1, 55292, '\P{Lb=	_jt}', "");
Expect(0, 55292, '\P{^Lb=	_jt}', "");
Error('\p{Is_Line_Break=	-jt/a/}');
Error('\P{Is_Line_Break=	-jt/a/}');
Expect(1, 55291, '\p{Is_Line_Break=jt}', "");
Expect(0, 55291, '\p{^Is_Line_Break=jt}', "");
Expect(0, 55291, '\P{Is_Line_Break=jt}', "");
Expect(1, 55291, '\P{^Is_Line_Break=jt}', "");
Expect(0, 55292, '\p{Is_Line_Break=jt}', "");
Expect(1, 55292, '\p{^Is_Line_Break=jt}', "");
Expect(1, 55292, '\P{Is_Line_Break=jt}', "");
Expect(0, 55292, '\P{^Is_Line_Break=jt}', "");
Expect(1, 55291, '\p{Is_Line_Break: JT}', "");
Expect(0, 55291, '\p{^Is_Line_Break: JT}', "");
Expect(0, 55291, '\P{Is_Line_Break: JT}', "");
Expect(1, 55291, '\P{^Is_Line_Break: JT}', "");
Expect(0, 55292, '\p{Is_Line_Break: JT}', "");
Expect(1, 55292, '\p{^Is_Line_Break: JT}', "");
Expect(1, 55292, '\P{Is_Line_Break: JT}', "");
Expect(0, 55292, '\P{^Is_Line_Break: JT}', "");
Error('\p{Is_Lb=_-JT/a/}');
Error('\P{Is_Lb=_-JT/a/}');
Expect(1, 55291, '\p{Is_Lb=jt}', "");
Expect(0, 55291, '\p{^Is_Lb=jt}', "");
Expect(0, 55291, '\P{Is_Lb=jt}', "");
Expect(1, 55291, '\P{^Is_Lb=jt}', "");
Expect(0, 55292, '\p{Is_Lb=jt}', "");
Expect(1, 55292, '\p{^Is_Lb=jt}', "");
Expect(1, 55292, '\P{Is_Lb=jt}', "");
Expect(0, 55292, '\P{^Is_Lb=jt}', "");
Expect(1, 55291, '\p{Is_Lb= 	JT}', "");
Expect(0, 55291, '\p{^Is_Lb= 	JT}', "");
Expect(0, 55291, '\P{Is_Lb= 	JT}', "");
Expect(1, 55291, '\P{^Is_Lb= 	JT}', "");
Expect(0, 55292, '\p{Is_Lb= 	JT}', "");
Expect(1, 55292, '\p{^Is_Lb= 	JT}', "");
Expect(1, 55292, '\P{Is_Lb= 	JT}', "");
Expect(0, 55292, '\P{^Is_Lb= 	JT}', "");
Error('\p{Line_Break=/a/- JV}');
Error('\P{Line_Break=/a/- JV}');
Expect(1, 55238, '\p{Line_Break=jv}', "");
Expect(0, 55238, '\p{^Line_Break=jv}', "");
Expect(0, 55238, '\P{Line_Break=jv}', "");
Expect(1, 55238, '\P{^Line_Break=jv}', "");
Expect(0, 55239, '\p{Line_Break=jv}', "");
Expect(1, 55239, '\p{^Line_Break=jv}', "");
Expect(1, 55239, '\P{Line_Break=jv}', "");
Expect(0, 55239, '\P{^Line_Break=jv}', "");
Expect(1, 55238, '\p{Line_Break= JV}', "");
Expect(0, 55238, '\p{^Line_Break= JV}', "");
Expect(0, 55238, '\P{Line_Break= JV}', "");
Expect(1, 55238, '\P{^Line_Break= JV}', "");
Expect(0, 55239, '\p{Line_Break= JV}', "");
Expect(1, 55239, '\p{^Line_Break= JV}', "");
Expect(1, 55239, '\P{Line_Break= JV}', "");
Expect(0, 55239, '\P{^Line_Break= JV}', "");
Error('\p{Lb=__jv/a/}');
Error('\P{Lb=__jv/a/}');
Expect(1, 55238, '\p{Lb=jv}', "");
Expect(0, 55238, '\p{^Lb=jv}', "");
Expect(0, 55238, '\P{Lb=jv}', "");
Expect(1, 55238, '\P{^Lb=jv}', "");
Expect(0, 55239, '\p{Lb=jv}', "");
Expect(1, 55239, '\p{^Lb=jv}', "");
Expect(1, 55239, '\P{Lb=jv}', "");
Expect(0, 55239, '\P{^Lb=jv}', "");
Expect(1, 55238, '\p{Lb= jv}', "");
Expect(0, 55238, '\p{^Lb= jv}', "");
Expect(0, 55238, '\P{Lb= jv}', "");
Expect(1, 55238, '\P{^Lb= jv}', "");
Expect(0, 55239, '\p{Lb= jv}', "");
Expect(1, 55239, '\p{^Lb= jv}', "");
Expect(1, 55239, '\P{Lb= jv}', "");
Expect(0, 55239, '\P{^Lb= jv}', "");
Error('\p{Is_Line_Break=  JV:=}');
Error('\P{Is_Line_Break=  JV:=}');
Expect(1, 55238, '\p{Is_Line_Break:	jv}', "");
Expect(0, 55238, '\p{^Is_Line_Break:	jv}', "");
Expect(0, 55238, '\P{Is_Line_Break:	jv}', "");
Expect(1, 55238, '\P{^Is_Line_Break:	jv}', "");
Expect(0, 55239, '\p{Is_Line_Break:	jv}', "");
Expect(1, 55239, '\p{^Is_Line_Break:	jv}', "");
Expect(1, 55239, '\P{Is_Line_Break:	jv}', "");
Expect(0, 55239, '\P{^Is_Line_Break:	jv}', "");
Expect(1, 55238, '\p{Is_Line_Break=	-jv}', "");
Expect(0, 55238, '\p{^Is_Line_Break=	-jv}', "");
Expect(0, 55238, '\P{Is_Line_Break=	-jv}', "");
Expect(1, 55238, '\P{^Is_Line_Break=	-jv}', "");
Expect(0, 55239, '\p{Is_Line_Break=	-jv}', "");
Expect(1, 55239, '\p{^Is_Line_Break=	-jv}', "");
Expect(1, 55239, '\P{Is_Line_Break=	-jv}', "");
Expect(0, 55239, '\P{^Is_Line_Break=	-jv}', "");
Error('\p{Is_Lb=_/a/JV}');
Error('\P{Is_Lb=_/a/JV}');
Expect(1, 55238, '\p{Is_Lb: jv}', "");
Expect(0, 55238, '\p{^Is_Lb: jv}', "");
Expect(0, 55238, '\P{Is_Lb: jv}', "");
Expect(1, 55238, '\P{^Is_Lb: jv}', "");
Expect(0, 55239, '\p{Is_Lb: jv}', "");
Expect(1, 55239, '\p{^Is_Lb: jv}', "");
Expect(1, 55239, '\P{Is_Lb: jv}', "");
Expect(0, 55239, '\P{^Is_Lb: jv}', "");
Expect(1, 55238, '\p{Is_Lb= 	JV}', "");
Expect(0, 55238, '\p{^Is_Lb= 	JV}', "");
Expect(0, 55238, '\P{Is_Lb= 	JV}', "");
Expect(1, 55238, '\P{^Is_Lb= 	JV}', "");
Expect(0, 55239, '\p{Is_Lb= 	JV}', "");
Expect(1, 55239, '\p{^Is_Lb= 	JV}', "");
Expect(1, 55239, '\P{Is_Lb= 	JV}', "");
Expect(0, 55239, '\P{^Is_Lb= 	JV}', "");
Error('\p{Line_Break=	line_feed/a/}');
Error('\P{Line_Break=	line_feed/a/}');
Expect(1, 10, '\p{Line_Break=linefeed}', "");
Expect(0, 10, '\p{^Line_Break=linefeed}', "");
Expect(0, 10, '\P{Line_Break=linefeed}', "");
Expect(1, 10, '\P{^Line_Break=linefeed}', "");
Expect(0, 11, '\p{Line_Break=linefeed}', "");
Expect(1, 11, '\p{^Line_Break=linefeed}', "");
Expect(1, 11, '\P{Line_Break=linefeed}', "");
Expect(0, 11, '\P{^Line_Break=linefeed}', "");
Expect(1, 10, '\p{Line_Break:	_ line_Feed}', "");
Expect(0, 10, '\p{^Line_Break:	_ line_Feed}', "");
Expect(0, 10, '\P{Line_Break:	_ line_Feed}', "");
Expect(1, 10, '\P{^Line_Break:	_ line_Feed}', "");
Expect(0, 11, '\p{Line_Break:	_ line_Feed}', "");
Expect(1, 11, '\p{^Line_Break:	_ line_Feed}', "");
Expect(1, 11, '\P{Line_Break:	_ line_Feed}', "");
Expect(0, 11, '\P{^Line_Break:	_ line_Feed}', "");
Error('\p{Lb=	_lf/a/}');
Error('\P{Lb=	_lf/a/}');
Expect(1, 10, '\p{Lb:lf}', "");
Expect(0, 10, '\p{^Lb:lf}', "");
Expect(0, 10, '\P{Lb:lf}', "");
Expect(1, 10, '\P{^Lb:lf}', "");
Expect(0, 11, '\p{Lb:lf}', "");
Expect(1, 11, '\p{^Lb:lf}', "");
Expect(1, 11, '\P{Lb:lf}', "");
Expect(0, 11, '\P{^Lb:lf}', "");
Expect(1, 10, '\p{Lb: 	 LF}', "");
Expect(0, 10, '\p{^Lb: 	 LF}', "");
Expect(0, 10, '\P{Lb: 	 LF}', "");
Expect(1, 10, '\P{^Lb: 	 LF}', "");
Expect(0, 11, '\p{Lb: 	 LF}', "");
Expect(1, 11, '\p{^Lb: 	 LF}', "");
Expect(1, 11, '\P{Lb: 	 LF}', "");
Expect(0, 11, '\P{^Lb: 	 LF}', "");
Error('\p{Is_Line_Break=:= Line_FEED}');
Error('\P{Is_Line_Break=:= Line_FEED}');
Expect(1, 10, '\p{Is_Line_Break=linefeed}', "");
Expect(0, 10, '\p{^Is_Line_Break=linefeed}', "");
Expect(0, 10, '\P{Is_Line_Break=linefeed}', "");
Expect(1, 10, '\P{^Is_Line_Break=linefeed}', "");
Expect(0, 11, '\p{Is_Line_Break=linefeed}', "");
Expect(1, 11, '\p{^Is_Line_Break=linefeed}', "");
Expect(1, 11, '\P{Is_Line_Break=linefeed}', "");
Expect(0, 11, '\P{^Is_Line_Break=linefeed}', "");
Expect(1, 10, '\p{Is_Line_Break: _ Line_FEED}', "");
Expect(0, 10, '\p{^Is_Line_Break: _ Line_FEED}', "");
Expect(0, 10, '\P{Is_Line_Break: _ Line_FEED}', "");
Expect(1, 10, '\P{^Is_Line_Break: _ Line_FEED}', "");
Expect(0, 11, '\p{Is_Line_Break: _ Line_FEED}', "");
Expect(1, 11, '\p{^Is_Line_Break: _ Line_FEED}', "");
Expect(1, 11, '\P{Is_Line_Break: _ Line_FEED}', "");
Expect(0, 11, '\P{^Is_Line_Break: _ Line_FEED}', "");
Error('\p{Is_Lb=/a/	lf}');
Error('\P{Is_Lb=/a/	lf}');
Expect(1, 10, '\p{Is_Lb=lf}', "");
Expect(0, 10, '\p{^Is_Lb=lf}', "");
Expect(0, 10, '\P{Is_Lb=lf}', "");
Expect(1, 10, '\P{^Is_Lb=lf}', "");
Expect(0, 11, '\p{Is_Lb=lf}', "");
Expect(1, 11, '\p{^Is_Lb=lf}', "");
Expect(1, 11, '\P{Is_Lb=lf}', "");
Expect(0, 11, '\P{^Is_Lb=lf}', "");
Expect(1, 10, '\p{Is_Lb=-lf}', "");
Expect(0, 10, '\p{^Is_Lb=-lf}', "");
Expect(0, 10, '\P{Is_Lb=-lf}', "");
Expect(1, 10, '\P{^Is_Lb=-lf}', "");
Expect(0, 11, '\p{Is_Lb=-lf}', "");
Expect(1, 11, '\p{^Is_Lb=-lf}', "");
Expect(1, 11, '\P{Is_Lb=-lf}', "");
Expect(0, 11, '\P{^Is_Lb=-lf}', "");
Error('\p{Line_Break=:=Next_Line}');
Error('\P{Line_Break=:=Next_Line}');
Expect(1, 133, '\p{Line_Break:   nextline}', "");
Expect(0, 133, '\p{^Line_Break:   nextline}', "");
Expect(0, 133, '\P{Line_Break:   nextline}', "");
Expect(1, 133, '\P{^Line_Break:   nextline}', "");
Expect(0, 134, '\p{Line_Break:   nextline}', "");
Expect(1, 134, '\p{^Line_Break:   nextline}', "");
Expect(1, 134, '\P{Line_Break:   nextline}', "");
Expect(0, 134, '\P{^Line_Break:   nextline}', "");
Expect(1, 133, '\p{Line_Break=	_Next_LINE}', "");
Expect(0, 133, '\p{^Line_Break=	_Next_LINE}', "");
Expect(0, 133, '\P{Line_Break=	_Next_LINE}', "");
Expect(1, 133, '\P{^Line_Break=	_Next_LINE}', "");
Expect(0, 134, '\p{Line_Break=	_Next_LINE}', "");
Expect(1, 134, '\p{^Line_Break=	_Next_LINE}', "");
Expect(1, 134, '\P{Line_Break=	_Next_LINE}', "");
Expect(0, 134, '\P{^Line_Break=	_Next_LINE}', "");
Error('\p{Lb=:=	 NL}');
Error('\P{Lb=:=	 NL}');
Expect(1, 133, '\p{Lb=nl}', "");
Expect(0, 133, '\p{^Lb=nl}', "");
Expect(0, 133, '\P{Lb=nl}', "");
Expect(1, 133, '\P{^Lb=nl}', "");
Expect(0, 134, '\p{Lb=nl}', "");
Expect(1, 134, '\p{^Lb=nl}', "");
Expect(1, 134, '\P{Lb=nl}', "");
Expect(0, 134, '\P{^Lb=nl}', "");
Expect(1, 133, '\p{Lb=-_NL}', "");
Expect(0, 133, '\p{^Lb=-_NL}', "");
Expect(0, 133, '\P{Lb=-_NL}', "");
Expect(1, 133, '\P{^Lb=-_NL}', "");
Expect(0, 134, '\p{Lb=-_NL}', "");
Expect(1, 134, '\p{^Lb=-_NL}', "");
Expect(1, 134, '\P{Lb=-_NL}', "");
Expect(0, 134, '\P{^Lb=-_NL}', "");
Error('\p{Is_Line_Break= next_line:=}');
Error('\P{Is_Line_Break= next_line:=}');
Expect(1, 133, '\p{Is_Line_Break=nextline}', "");
Expect(0, 133, '\p{^Is_Line_Break=nextline}', "");
Expect(0, 133, '\P{Is_Line_Break=nextline}', "");
Expect(1, 133, '\P{^Is_Line_Break=nextline}', "");
Expect(0, 134, '\p{Is_Line_Break=nextline}', "");
Expect(1, 134, '\p{^Is_Line_Break=nextline}', "");
Expect(1, 134, '\P{Is_Line_Break=nextline}', "");
Expect(0, 134, '\P{^Is_Line_Break=nextline}', "");
Expect(1, 133, '\p{Is_Line_Break=_ next_LINE}', "");
Expect(0, 133, '\p{^Is_Line_Break=_ next_LINE}', "");
Expect(0, 133, '\P{Is_Line_Break=_ next_LINE}', "");
Expect(1, 133, '\P{^Is_Line_Break=_ next_LINE}', "");
Expect(0, 134, '\p{Is_Line_Break=_ next_LINE}', "");
Expect(1, 134, '\p{^Is_Line_Break=_ next_LINE}', "");
Expect(1, 134, '\P{Is_Line_Break=_ next_LINE}', "");
Expect(0, 134, '\P{^Is_Line_Break=_ next_LINE}', "");
Error('\p{Is_Lb:	_NL/a/}');
Error('\P{Is_Lb:	_NL/a/}');
Expect(1, 133, '\p{Is_Lb=nl}', "");
Expect(0, 133, '\p{^Is_Lb=nl}', "");
Expect(0, 133, '\P{Is_Lb=nl}', "");
Expect(1, 133, '\P{^Is_Lb=nl}', "");
Expect(0, 134, '\p{Is_Lb=nl}', "");
Expect(1, 134, '\p{^Is_Lb=nl}', "");
Expect(1, 134, '\P{Is_Lb=nl}', "");
Expect(0, 134, '\P{^Is_Lb=nl}', "");
Expect(1, 133, '\p{Is_Lb=	-NL}', "");
Expect(0, 133, '\p{^Is_Lb=	-NL}', "");
Expect(0, 133, '\P{Is_Lb=	-NL}', "");
Expect(1, 133, '\P{^Is_Lb=	-NL}', "");
Expect(0, 134, '\p{Is_Lb=	-NL}', "");
Expect(1, 134, '\p{^Is_Lb=	-NL}', "");
Expect(1, 134, '\P{Is_Lb=	-NL}', "");
Expect(0, 134, '\P{^Is_Lb=	-NL}', "");
Error('\p{Line_Break= _NONSTARTER:=}');
Error('\P{Line_Break= _NONSTARTER:=}');
Expect(1, 65439, '\p{Line_Break=nonstarter}', "");
Expect(0, 65439, '\p{^Line_Break=nonstarter}', "");
Expect(0, 65439, '\P{Line_Break=nonstarter}', "");
Expect(1, 65439, '\P{^Line_Break=nonstarter}', "");
Expect(0, 65440, '\p{Line_Break=nonstarter}', "");
Expect(1, 65440, '\p{^Line_Break=nonstarter}', "");
Expect(1, 65440, '\P{Line_Break=nonstarter}', "");
Expect(0, 65440, '\P{^Line_Break=nonstarter}', "");
Expect(1, 65439, '\p{Line_Break=_NONSTARTER}', "");
Expect(0, 65439, '\p{^Line_Break=_NONSTARTER}', "");
Expect(0, 65439, '\P{Line_Break=_NONSTARTER}', "");
Expect(1, 65439, '\P{^Line_Break=_NONSTARTER}', "");
Expect(0, 65440, '\p{Line_Break=_NONSTARTER}', "");
Expect(1, 65440, '\p{^Line_Break=_NONSTARTER}', "");
Expect(1, 65440, '\P{Line_Break=_NONSTARTER}', "");
Expect(0, 65440, '\P{^Line_Break=_NONSTARTER}', "");
Error('\p{Lb= /a/NS}');
Error('\P{Lb= /a/NS}');
Expect(1, 65439, '\p{Lb=ns}', "");
Expect(0, 65439, '\p{^Lb=ns}', "");
Expect(0, 65439, '\P{Lb=ns}', "");
Expect(1, 65439, '\P{^Lb=ns}', "");
Expect(0, 65440, '\p{Lb=ns}', "");
Expect(1, 65440, '\p{^Lb=ns}', "");
Expect(1, 65440, '\P{Lb=ns}', "");
Expect(0, 65440, '\P{^Lb=ns}', "");
Expect(1, 65439, '\p{Lb=	_ns}', "");
Expect(0, 65439, '\p{^Lb=	_ns}', "");
Expect(0, 65439, '\P{Lb=	_ns}', "");
Expect(1, 65439, '\P{^Lb=	_ns}', "");
Expect(0, 65440, '\p{Lb=	_ns}', "");
Expect(1, 65440, '\p{^Lb=	_ns}', "");
Expect(1, 65440, '\P{Lb=	_ns}', "");
Expect(0, 65440, '\P{^Lb=	_ns}', "");
Error('\p{Is_Line_Break=	:=Nonstarter}');
Error('\P{Is_Line_Break=	:=Nonstarter}');
Expect(1, 65439, '\p{Is_Line_Break=nonstarter}', "");
Expect(0, 65439, '\p{^Is_Line_Break=nonstarter}', "");
Expect(0, 65439, '\P{Is_Line_Break=nonstarter}', "");
Expect(1, 65439, '\P{^Is_Line_Break=nonstarter}', "");
Expect(0, 65440, '\p{Is_Line_Break=nonstarter}', "");
Expect(1, 65440, '\p{^Is_Line_Break=nonstarter}', "");
Expect(1, 65440, '\P{Is_Line_Break=nonstarter}', "");
Expect(0, 65440, '\P{^Is_Line_Break=nonstarter}', "");
Expect(1, 65439, '\p{Is_Line_Break=-	Nonstarter}', "");
Expect(0, 65439, '\p{^Is_Line_Break=-	Nonstarter}', "");
Expect(0, 65439, '\P{Is_Line_Break=-	Nonstarter}', "");
Expect(1, 65439, '\P{^Is_Line_Break=-	Nonstarter}', "");
Expect(0, 65440, '\p{Is_Line_Break=-	Nonstarter}', "");
Expect(1, 65440, '\p{^Is_Line_Break=-	Nonstarter}', "");
Expect(1, 65440, '\P{Is_Line_Break=-	Nonstarter}', "");
Expect(0, 65440, '\P{^Is_Line_Break=-	Nonstarter}', "");
Error('\p{Is_Lb=-/a/NS}');
Error('\P{Is_Lb=-/a/NS}');
Expect(1, 65439, '\p{Is_Lb=ns}', "");
Expect(0, 65439, '\p{^Is_Lb=ns}', "");
Expect(0, 65439, '\P{Is_Lb=ns}', "");
Expect(1, 65439, '\P{^Is_Lb=ns}', "");
Expect(0, 65440, '\p{Is_Lb=ns}', "");
Expect(1, 65440, '\p{^Is_Lb=ns}', "");
Expect(1, 65440, '\P{Is_Lb=ns}', "");
Expect(0, 65440, '\P{^Is_Lb=ns}', "");
Expect(1, 65439, '\p{Is_Lb: 	NS}', "");
Expect(0, 65439, '\p{^Is_Lb: 	NS}', "");
Expect(0, 65439, '\P{Is_Lb: 	NS}', "");
Expect(1, 65439, '\P{^Is_Lb: 	NS}', "");
Expect(0, 65440, '\p{Is_Lb: 	NS}', "");
Expect(1, 65440, '\p{^Is_Lb: 	NS}', "");
Expect(1, 65440, '\P{Is_Lb: 	NS}', "");
Expect(0, 65440, '\P{^Is_Lb: 	NS}', "");
Error('\p{Line_Break=--numeric:=}');
Error('\P{Line_Break=--numeric:=}');
Expect(1, 120831, '\p{Line_Break:   numeric}', "");
Expect(0, 120831, '\p{^Line_Break:   numeric}', "");
Expect(0, 120831, '\P{Line_Break:   numeric}', "");
Expect(1, 120831, '\P{^Line_Break:   numeric}', "");
Expect(0, 120832, '\p{Line_Break:   numeric}', "");
Expect(1, 120832, '\p{^Line_Break:   numeric}', "");
Expect(1, 120832, '\P{Line_Break:   numeric}', "");
Expect(0, 120832, '\P{^Line_Break:   numeric}', "");
Expect(1, 120831, '\p{Line_Break=-numeric}', "");
Expect(0, 120831, '\p{^Line_Break=-numeric}', "");
Expect(0, 120831, '\P{Line_Break=-numeric}', "");
Expect(1, 120831, '\P{^Line_Break=-numeric}', "");
Expect(0, 120832, '\p{Line_Break=-numeric}', "");
Expect(1, 120832, '\p{^Line_Break=-numeric}', "");
Expect(1, 120832, '\P{Line_Break=-numeric}', "");
Expect(0, 120832, '\P{^Line_Break=-numeric}', "");
Error('\p{Lb=:=_	nu}');
Error('\P{Lb=:=_	nu}');
Expect(1, 120831, '\p{Lb=nu}', "");
Expect(0, 120831, '\p{^Lb=nu}', "");
Expect(0, 120831, '\P{Lb=nu}', "");
Expect(1, 120831, '\P{^Lb=nu}', "");
Expect(0, 120832, '\p{Lb=nu}', "");
Expect(1, 120832, '\p{^Lb=nu}', "");
Expect(1, 120832, '\P{Lb=nu}', "");
Expect(0, 120832, '\P{^Lb=nu}', "");
Expect(1, 120831, '\p{Lb= NU}', "");
Expect(0, 120831, '\p{^Lb= NU}', "");
Expect(0, 120831, '\P{Lb= NU}', "");
Expect(1, 120831, '\P{^Lb= NU}', "");
Expect(0, 120832, '\p{Lb= NU}', "");
Expect(1, 120832, '\p{^Lb= NU}', "");
Expect(1, 120832, '\P{Lb= NU}', "");
Expect(0, 120832, '\P{^Lb= NU}', "");
Error('\p{Is_Line_Break=:= 	NUMERIC}');
Error('\P{Is_Line_Break=:= 	NUMERIC}');
Expect(1, 120831, '\p{Is_Line_Break=numeric}', "");
Expect(0, 120831, '\p{^Is_Line_Break=numeric}', "");
Expect(0, 120831, '\P{Is_Line_Break=numeric}', "");
Expect(1, 120831, '\P{^Is_Line_Break=numeric}', "");
Expect(0, 120832, '\p{Is_Line_Break=numeric}', "");
Expect(1, 120832, '\p{^Is_Line_Break=numeric}', "");
Expect(1, 120832, '\P{Is_Line_Break=numeric}', "");
Expect(0, 120832, '\P{^Is_Line_Break=numeric}', "");
Expect(1, 120831, '\p{Is_Line_Break=_Numeric}', "");
Expect(0, 120831, '\p{^Is_Line_Break=_Numeric}', "");
Expect(0, 120831, '\P{Is_Line_Break=_Numeric}', "");
Expect(1, 120831, '\P{^Is_Line_Break=_Numeric}', "");
Expect(0, 120832, '\p{Is_Line_Break=_Numeric}', "");
Expect(1, 120832, '\p{^Is_Line_Break=_Numeric}', "");
Expect(1, 120832, '\P{Is_Line_Break=_Numeric}', "");
Expect(0, 120832, '\P{^Is_Line_Break=_Numeric}', "");
Error('\p{Is_Lb=-	NU:=}');
Error('\P{Is_Lb=-	NU:=}');
Expect(1, 120831, '\p{Is_Lb=nu}', "");
Expect(0, 120831, '\p{^Is_Lb=nu}', "");
Expect(0, 120831, '\P{Is_Lb=nu}', "");
Expect(1, 120831, '\P{^Is_Lb=nu}', "");
Expect(0, 120832, '\p{Is_Lb=nu}', "");
Expect(1, 120832, '\p{^Is_Lb=nu}', "");
Expect(1, 120832, '\P{Is_Lb=nu}', "");
Expect(0, 120832, '\P{^Is_Lb=nu}', "");
Expect(1, 120831, '\p{Is_Lb=	-NU}', "");
Expect(0, 120831, '\p{^Is_Lb=	-NU}', "");
Expect(0, 120831, '\P{Is_Lb=	-NU}', "");
Expect(1, 120831, '\P{^Is_Lb=	-NU}', "");
Expect(0, 120832, '\p{Is_Lb=	-NU}', "");
Expect(1, 120832, '\p{^Is_Lb=	-NU}', "");
Expect(1, 120832, '\P{Is_Lb=	-NU}', "");
Expect(0, 120832, '\P{^Is_Lb=	-NU}', "");
Error('\p{Line_Break=/a/		open_Punctuation}');
Error('\P{Line_Break=/a/		open_Punctuation}');
Expect(1, 78713, '\p{Line_Break=openpunctuation}', "");
Expect(0, 78713, '\p{^Line_Break=openpunctuation}', "");
Expect(0, 78713, '\P{Line_Break=openpunctuation}', "");
Expect(1, 78713, '\P{^Line_Break=openpunctuation}', "");
Expect(0, 78714, '\p{Line_Break=openpunctuation}', "");
Expect(1, 78714, '\p{^Line_Break=openpunctuation}', "");
Expect(1, 78714, '\P{Line_Break=openpunctuation}', "");
Expect(0, 78714, '\P{^Line_Break=openpunctuation}', "");
Expect(1, 78713, '\p{Line_Break= OPEN_punctuation}', "");
Expect(0, 78713, '\p{^Line_Break= OPEN_punctuation}', "");
Expect(0, 78713, '\P{Line_Break= OPEN_punctuation}', "");
Expect(1, 78713, '\P{^Line_Break= OPEN_punctuation}', "");
Expect(0, 78714, '\p{Line_Break= OPEN_punctuation}', "");
Expect(1, 78714, '\p{^Line_Break= OPEN_punctuation}', "");
Expect(1, 78714, '\P{Line_Break= OPEN_punctuation}', "");
Expect(0, 78714, '\P{^Line_Break= OPEN_punctuation}', "");
Error('\p{Lb:/a/ op}');
Error('\P{Lb:/a/ op}');
Expect(1, 78713, '\p{Lb=op}', "");
Expect(0, 78713, '\p{^Lb=op}', "");
Expect(0, 78713, '\P{Lb=op}', "");
Expect(1, 78713, '\P{^Lb=op}', "");
Expect(0, 78714, '\p{Lb=op}', "");
Expect(1, 78714, '\p{^Lb=op}', "");
Expect(1, 78714, '\P{Lb=op}', "");
Expect(0, 78714, '\P{^Lb=op}', "");
Expect(1, 78713, '\p{Lb=-OP}', "");
Expect(0, 78713, '\p{^Lb=-OP}', "");
Expect(0, 78713, '\P{Lb=-OP}', "");
Expect(1, 78713, '\P{^Lb=-OP}', "");
Expect(0, 78714, '\p{Lb=-OP}', "");
Expect(1, 78714, '\p{^Lb=-OP}', "");
Expect(1, 78714, '\P{Lb=-OP}', "");
Expect(0, 78714, '\P{^Lb=-OP}', "");
Error('\p{Is_Line_Break= OPEN_punctuation/a/}');
Error('\P{Is_Line_Break= OPEN_punctuation/a/}');
Expect(1, 78713, '\p{Is_Line_Break=openpunctuation}', "");
Expect(0, 78713, '\p{^Is_Line_Break=openpunctuation}', "");
Expect(0, 78713, '\P{Is_Line_Break=openpunctuation}', "");
Expect(1, 78713, '\P{^Is_Line_Break=openpunctuation}', "");
Expect(0, 78714, '\p{Is_Line_Break=openpunctuation}', "");
Expect(1, 78714, '\p{^Is_Line_Break=openpunctuation}', "");
Expect(1, 78714, '\P{Is_Line_Break=openpunctuation}', "");
Expect(0, 78714, '\P{^Is_Line_Break=openpunctuation}', "");
Expect(1, 78713, '\p{Is_Line_Break=OPEN_punctuation}', "");
Expect(0, 78713, '\p{^Is_Line_Break=OPEN_punctuation}', "");
Expect(0, 78713, '\P{Is_Line_Break=OPEN_punctuation}', "");
Expect(1, 78713, '\P{^Is_Line_Break=OPEN_punctuation}', "");
Expect(0, 78714, '\p{Is_Line_Break=OPEN_punctuation}', "");
Expect(1, 78714, '\p{^Is_Line_Break=OPEN_punctuation}', "");
Expect(1, 78714, '\P{Is_Line_Break=OPEN_punctuation}', "");
Expect(0, 78714, '\P{^Is_Line_Break=OPEN_punctuation}', "");
Error('\p{Is_Lb=	:=OP}');
Error('\P{Is_Lb=	:=OP}');
Expect(1, 78713, '\p{Is_Lb=op}', "");
Expect(0, 78713, '\p{^Is_Lb=op}', "");
Expect(0, 78713, '\P{Is_Lb=op}', "");
Expect(1, 78713, '\P{^Is_Lb=op}', "");
Expect(0, 78714, '\p{Is_Lb=op}', "");
Expect(1, 78714, '\p{^Is_Lb=op}', "");
Expect(1, 78714, '\P{Is_Lb=op}', "");
Expect(0, 78714, '\P{^Is_Lb=op}', "");
Expect(1, 78713, '\p{Is_Lb=--OP}', "");
Expect(0, 78713, '\p{^Is_Lb=--OP}', "");
Expect(0, 78713, '\P{Is_Lb=--OP}', "");
Expect(1, 78713, '\P{^Is_Lb=--OP}', "");
Expect(0, 78714, '\p{Is_Lb=--OP}', "");
Expect(1, 78714, '\p{^Is_Lb=--OP}', "");
Expect(1, 78714, '\P{Is_Lb=--OP}', "");
Expect(0, 78714, '\P{^Is_Lb=--OP}', "");
Error('\p{Line_Break: /a/	 Postfix_NUMERIC}');
Error('\P{Line_Break: /a/	 Postfix_NUMERIC}');
Expect(1, 65504, '\p{Line_Break=postfixnumeric}', "");
Expect(0, 65504, '\p{^Line_Break=postfixnumeric}', "");
Expect(0, 65504, '\P{Line_Break=postfixnumeric}', "");
Expect(1, 65504, '\P{^Line_Break=postfixnumeric}', "");
Expect(0, 65505, '\p{Line_Break=postfixnumeric}', "");
Expect(1, 65505, '\p{^Line_Break=postfixnumeric}', "");
Expect(1, 65505, '\P{Line_Break=postfixnumeric}', "");
Expect(0, 65505, '\P{^Line_Break=postfixnumeric}', "");
Expect(1, 65504, '\p{Line_Break=--postfix_Numeric}', "");
Expect(0, 65504, '\p{^Line_Break=--postfix_Numeric}', "");
Expect(0, 65504, '\P{Line_Break=--postfix_Numeric}', "");
Expect(1, 65504, '\P{^Line_Break=--postfix_Numeric}', "");
Expect(0, 65505, '\p{Line_Break=--postfix_Numeric}', "");
Expect(1, 65505, '\p{^Line_Break=--postfix_Numeric}', "");
Expect(1, 65505, '\P{Line_Break=--postfix_Numeric}', "");
Expect(0, 65505, '\P{^Line_Break=--postfix_Numeric}', "");
Error('\p{Lb=	 PO:=}');
Error('\P{Lb=	 PO:=}');
Expect(1, 65504, '\p{Lb=po}', "");
Expect(0, 65504, '\p{^Lb=po}', "");
Expect(0, 65504, '\P{Lb=po}', "");
Expect(1, 65504, '\P{^Lb=po}', "");
Expect(0, 65505, '\p{Lb=po}', "");
Expect(1, 65505, '\p{^Lb=po}', "");
Expect(1, 65505, '\P{Lb=po}', "");
Expect(0, 65505, '\P{^Lb=po}', "");
Expect(1, 65504, '\p{Lb=-_PO}', "");
Expect(0, 65504, '\p{^Lb=-_PO}', "");
Expect(0, 65504, '\P{Lb=-_PO}', "");
Expect(1, 65504, '\P{^Lb=-_PO}', "");
Expect(0, 65505, '\p{Lb=-_PO}', "");
Expect(1, 65505, '\p{^Lb=-_PO}', "");
Expect(1, 65505, '\P{Lb=-_PO}', "");
Expect(0, 65505, '\P{^Lb=-_PO}', "");
Error('\p{Is_Line_Break=	:=POSTFIX_NUMERIC}');
Error('\P{Is_Line_Break=	:=POSTFIX_NUMERIC}');
Expect(1, 65504, '\p{Is_Line_Break=postfixnumeric}', "");
Expect(0, 65504, '\p{^Is_Line_Break=postfixnumeric}', "");
Expect(0, 65504, '\P{Is_Line_Break=postfixnumeric}', "");
Expect(1, 65504, '\P{^Is_Line_Break=postfixnumeric}', "");
Expect(0, 65505, '\p{Is_Line_Break=postfixnumeric}', "");
Expect(1, 65505, '\p{^Is_Line_Break=postfixnumeric}', "");
Expect(1, 65505, '\P{Is_Line_Break=postfixnumeric}', "");
Expect(0, 65505, '\P{^Is_Line_Break=postfixnumeric}', "");
Expect(1, 65504, '\p{Is_Line_Break= Postfix_NUMERIC}', "");
Expect(0, 65504, '\p{^Is_Line_Break= Postfix_NUMERIC}', "");
Expect(0, 65504, '\P{Is_Line_Break= Postfix_NUMERIC}', "");
Expect(1, 65504, '\P{^Is_Line_Break= Postfix_NUMERIC}', "");
Expect(0, 65505, '\p{Is_Line_Break= Postfix_NUMERIC}', "");
Expect(1, 65505, '\p{^Is_Line_Break= Postfix_NUMERIC}', "");
Expect(1, 65505, '\P{Is_Line_Break= Postfix_NUMERIC}', "");
Expect(0, 65505, '\P{^Is_Line_Break= Postfix_NUMERIC}', "");
Error('\p{Is_Lb=_/a/PO}');
Error('\P{Is_Lb=_/a/PO}');
Expect(1, 65504, '\p{Is_Lb:   po}', "");
Expect(0, 65504, '\p{^Is_Lb:   po}', "");
Expect(0, 65504, '\P{Is_Lb:   po}', "");
Expect(1, 65504, '\P{^Is_Lb:   po}', "");
Expect(0, 65505, '\p{Is_Lb:   po}', "");
Expect(1, 65505, '\p{^Is_Lb:   po}', "");
Expect(1, 65505, '\P{Is_Lb:   po}', "");
Expect(0, 65505, '\P{^Is_Lb:   po}', "");
Expect(1, 65504, '\p{Is_Lb=_	PO}', "");
Expect(0, 65504, '\p{^Is_Lb=_	PO}', "");
Expect(0, 65504, '\P{Is_Lb=_	PO}', "");
Expect(1, 65504, '\P{^Is_Lb=_	PO}', "");
Expect(0, 65505, '\p{Is_Lb=_	PO}', "");
Expect(1, 65505, '\p{^Is_Lb=_	PO}', "");
Expect(1, 65505, '\P{Is_Lb=_	PO}', "");
Expect(0, 65505, '\P{^Is_Lb=_	PO}', "");
Error('\p{Line_Break=/a/		PREFIX_Numeric}');
Error('\P{Line_Break=/a/		PREFIX_Numeric}');
Expect(1, 65510, '\p{Line_Break:	prefixnumeric}', "");
Expect(0, 65510, '\p{^Line_Break:	prefixnumeric}', "");
Expect(0, 65510, '\P{Line_Break:	prefixnumeric}', "");
Expect(1, 65510, '\P{^Line_Break:	prefixnumeric}', "");
Expect(0, 65511, '\p{Line_Break:	prefixnumeric}', "");
Expect(1, 65511, '\p{^Line_Break:	prefixnumeric}', "");
Expect(1, 65511, '\P{Line_Break:	prefixnumeric}', "");
Expect(0, 65511, '\P{^Line_Break:	prefixnumeric}', "");
Expect(1, 65510, '\p{Line_Break= PREFIX_Numeric}', "");
Expect(0, 65510, '\p{^Line_Break= PREFIX_Numeric}', "");
Expect(0, 65510, '\P{Line_Break= PREFIX_Numeric}', "");
Expect(1, 65510, '\P{^Line_Break= PREFIX_Numeric}', "");
Expect(0, 65511, '\p{Line_Break= PREFIX_Numeric}', "");
Expect(1, 65511, '\p{^Line_Break= PREFIX_Numeric}', "");
Expect(1, 65511, '\P{Line_Break= PREFIX_Numeric}', "");
Expect(0, 65511, '\P{^Line_Break= PREFIX_Numeric}', "");
Error('\p{Lb=_	PR/a/}');
Error('\P{Lb=_	PR/a/}');
Expect(1, 65510, '\p{Lb: pr}', "");
Expect(0, 65510, '\p{^Lb: pr}', "");
Expect(0, 65510, '\P{Lb: pr}', "");
Expect(1, 65510, '\P{^Lb: pr}', "");
Expect(0, 65511, '\p{Lb: pr}', "");
Expect(1, 65511, '\p{^Lb: pr}', "");
Expect(1, 65511, '\P{Lb: pr}', "");
Expect(0, 65511, '\P{^Lb: pr}', "");
Expect(1, 65510, '\p{Lb=_PR}', "");
Expect(0, 65510, '\p{^Lb=_PR}', "");
Expect(0, 65510, '\P{Lb=_PR}', "");
Expect(1, 65510, '\P{^Lb=_PR}', "");
Expect(0, 65511, '\p{Lb=_PR}', "");
Expect(1, 65511, '\p{^Lb=_PR}', "");
Expect(1, 65511, '\P{Lb=_PR}', "");
Expect(0, 65511, '\P{^Lb=_PR}', "");
Error('\p{Is_Line_Break: 	PREFIX_Numeric:=}');
Error('\P{Is_Line_Break: 	PREFIX_Numeric:=}');
Expect(1, 65510, '\p{Is_Line_Break=prefixnumeric}', "");
Expect(0, 65510, '\p{^Is_Line_Break=prefixnumeric}', "");
Expect(0, 65510, '\P{Is_Line_Break=prefixnumeric}', "");
Expect(1, 65510, '\P{^Is_Line_Break=prefixnumeric}', "");
Expect(0, 65511, '\p{Is_Line_Break=prefixnumeric}', "");
Expect(1, 65511, '\p{^Is_Line_Break=prefixnumeric}', "");
Expect(1, 65511, '\P{Is_Line_Break=prefixnumeric}', "");
Expect(0, 65511, '\P{^Is_Line_Break=prefixnumeric}', "");
Expect(1, 65510, '\p{Is_Line_Break=	Prefix_numeric}', "");
Expect(0, 65510, '\p{^Is_Line_Break=	Prefix_numeric}', "");
Expect(0, 65510, '\P{Is_Line_Break=	Prefix_numeric}', "");
Expect(1, 65510, '\P{^Is_Line_Break=	Prefix_numeric}', "");
Expect(0, 65511, '\p{Is_Line_Break=	Prefix_numeric}', "");
Expect(1, 65511, '\p{^Is_Line_Break=	Prefix_numeric}', "");
Expect(1, 65511, '\P{Is_Line_Break=	Prefix_numeric}', "");
Expect(0, 65511, '\P{^Is_Line_Break=	Prefix_numeric}', "");
Error('\p{Is_Lb=	/a/PR}');
Error('\P{Is_Lb=	/a/PR}');
Expect(1, 65510, '\p{Is_Lb=pr}', "");
Expect(0, 65510, '\p{^Is_Lb=pr}', "");
Expect(0, 65510, '\P{Is_Lb=pr}', "");
Expect(1, 65510, '\P{^Is_Lb=pr}', "");
Expect(0, 65511, '\p{Is_Lb=pr}', "");
Expect(1, 65511, '\p{^Is_Lb=pr}', "");
Expect(1, 65511, '\P{Is_Lb=pr}', "");
Expect(0, 65511, '\P{^Is_Lb=pr}', "");
Expect(1, 65510, '\p{Is_Lb=	 pr}', "");
Expect(0, 65510, '\p{^Is_Lb=	 pr}', "");
Expect(0, 65510, '\P{Is_Lb=	 pr}', "");
Expect(1, 65510, '\P{^Is_Lb=	 pr}', "");
Expect(0, 65511, '\p{Is_Lb=	 pr}', "");
Expect(1, 65511, '\p{^Is_Lb=	 pr}', "");
Expect(1, 65511, '\P{Is_Lb=	 pr}', "");
Expect(0, 65511, '\P{^Is_Lb=	 pr}', "");
Error('\p{Line_Break=/a/	Quotation}');
Error('\P{Line_Break=/a/	Quotation}');
Expect(1, 11809, '\p{Line_Break=quotation}', "");
Expect(0, 11809, '\p{^Line_Break=quotation}', "");
Expect(0, 11809, '\P{Line_Break=quotation}', "");
Expect(1, 11809, '\P{^Line_Break=quotation}', "");
Expect(0, 11810, '\p{Line_Break=quotation}', "");
Expect(1, 11810, '\p{^Line_Break=quotation}', "");
Expect(1, 11810, '\P{Line_Break=quotation}', "");
Expect(0, 11810, '\P{^Line_Break=quotation}', "");
Expect(1, 11809, '\p{Line_Break=	-Quotation}', "");
Expect(0, 11809, '\p{^Line_Break=	-Quotation}', "");
Expect(0, 11809, '\P{Line_Break=	-Quotation}', "");
Expect(1, 11809, '\P{^Line_Break=	-Quotation}', "");
Expect(0, 11810, '\p{Line_Break=	-Quotation}', "");
Expect(1, 11810, '\p{^Line_Break=	-Quotation}', "");
Expect(1, 11810, '\P{Line_Break=	-Quotation}', "");
Expect(0, 11810, '\P{^Line_Break=	-Quotation}', "");
Error('\p{Lb=/a/ 	qu}');
Error('\P{Lb=/a/ 	qu}');
Expect(1, 11809, '\p{Lb=qu}', "");
Expect(0, 11809, '\p{^Lb=qu}', "");
Expect(0, 11809, '\P{Lb=qu}', "");
Expect(1, 11809, '\P{^Lb=qu}', "");
Expect(0, 11810, '\p{Lb=qu}', "");
Expect(1, 11810, '\p{^Lb=qu}', "");
Expect(1, 11810, '\P{Lb=qu}', "");
Expect(0, 11810, '\P{^Lb=qu}', "");
Expect(1, 11809, '\p{Lb:	_ QU}', "");
Expect(0, 11809, '\p{^Lb:	_ QU}', "");
Expect(0, 11809, '\P{Lb:	_ QU}', "");
Expect(1, 11809, '\P{^Lb:	_ QU}', "");
Expect(0, 11810, '\p{Lb:	_ QU}', "");
Expect(1, 11810, '\p{^Lb:	_ QU}', "");
Expect(1, 11810, '\P{Lb:	_ QU}', "");
Expect(0, 11810, '\P{^Lb:	_ QU}', "");
Error('\p{Is_Line_Break= :=Quotation}');
Error('\P{Is_Line_Break= :=Quotation}');
Expect(1, 11809, '\p{Is_Line_Break=quotation}', "");
Expect(0, 11809, '\p{^Is_Line_Break=quotation}', "");
Expect(0, 11809, '\P{Is_Line_Break=quotation}', "");
Expect(1, 11809, '\P{^Is_Line_Break=quotation}', "");
Expect(0, 11810, '\p{Is_Line_Break=quotation}', "");
Expect(1, 11810, '\p{^Is_Line_Break=quotation}', "");
Expect(1, 11810, '\P{Is_Line_Break=quotation}', "");
Expect(0, 11810, '\P{^Is_Line_Break=quotation}', "");
Expect(1, 11809, '\p{Is_Line_Break=- QUOTATION}', "");
Expect(0, 11809, '\p{^Is_Line_Break=- QUOTATION}', "");
Expect(0, 11809, '\P{Is_Line_Break=- QUOTATION}', "");
Expect(1, 11809, '\P{^Is_Line_Break=- QUOTATION}', "");
Expect(0, 11810, '\p{Is_Line_Break=- QUOTATION}', "");
Expect(1, 11810, '\p{^Is_Line_Break=- QUOTATION}', "");
Expect(1, 11810, '\P{Is_Line_Break=- QUOTATION}', "");
Expect(0, 11810, '\P{^Is_Line_Break=- QUOTATION}', "");
Error('\p{Is_Lb=-	qu/a/}');
Error('\P{Is_Lb=-	qu/a/}');
Expect(1, 11809, '\p{Is_Lb=qu}', "");
Expect(0, 11809, '\p{^Is_Lb=qu}', "");
Expect(0, 11809, '\P{Is_Lb=qu}', "");
Expect(1, 11809, '\P{^Is_Lb=qu}', "");
Expect(0, 11810, '\p{Is_Lb=qu}', "");
Expect(1, 11810, '\p{^Is_Lb=qu}', "");
Expect(1, 11810, '\P{Is_Lb=qu}', "");
Expect(0, 11810, '\P{^Is_Lb=qu}', "");
Expect(1, 11809, '\p{Is_Lb=-QU}', "");
Expect(0, 11809, '\p{^Is_Lb=-QU}', "");
Expect(0, 11809, '\P{Is_Lb=-QU}', "");
Expect(1, 11809, '\P{^Is_Lb=-QU}', "");
Expect(0, 11810, '\p{Is_Lb=-QU}', "");
Expect(1, 11810, '\p{^Is_Lb=-QU}', "");
Expect(1, 11810, '\P{Is_Lb=-QU}', "");
Expect(0, 11810, '\P{^Is_Lb=-QU}', "");
Error('\p{Line_Break= _Regional_INDICATOR:=}');
Error('\P{Line_Break= _Regional_INDICATOR:=}');
Expect(1, 127487, '\p{Line_Break=regionalindicator}', "");
Expect(0, 127487, '\p{^Line_Break=regionalindicator}', "");
Expect(0, 127487, '\P{Line_Break=regionalindicator}', "");
Expect(1, 127487, '\P{^Line_Break=regionalindicator}', "");
Expect(0, 127488, '\p{Line_Break=regionalindicator}', "");
Expect(1, 127488, '\p{^Line_Break=regionalindicator}', "");
Expect(1, 127488, '\P{Line_Break=regionalindicator}', "");
Expect(0, 127488, '\P{^Line_Break=regionalindicator}', "");
Expect(1, 127487, '\p{Line_Break=- regional_Indicator}', "");
Expect(0, 127487, '\p{^Line_Break=- regional_Indicator}', "");
Expect(0, 127487, '\P{Line_Break=- regional_Indicator}', "");
Expect(1, 127487, '\P{^Line_Break=- regional_Indicator}', "");
Expect(0, 127488, '\p{Line_Break=- regional_Indicator}', "");
Expect(1, 127488, '\p{^Line_Break=- regional_Indicator}', "");
Expect(1, 127488, '\P{Line_Break=- regional_Indicator}', "");
Expect(0, 127488, '\P{^Line_Break=- regional_Indicator}', "");
Error('\p{Lb=-/a/RI}');
Error('\P{Lb=-/a/RI}');
Expect(1, 127487, '\p{Lb=ri}', "");
Expect(0, 127487, '\p{^Lb=ri}', "");
Expect(0, 127487, '\P{Lb=ri}', "");
Expect(1, 127487, '\P{^Lb=ri}', "");
Expect(0, 127488, '\p{Lb=ri}', "");
Expect(1, 127488, '\p{^Lb=ri}', "");
Expect(1, 127488, '\P{Lb=ri}', "");
Expect(0, 127488, '\P{^Lb=ri}', "");
Expect(1, 127487, '\p{Lb=	_RI}', "");
Expect(0, 127487, '\p{^Lb=	_RI}', "");
Expect(0, 127487, '\P{Lb=	_RI}', "");
Expect(1, 127487, '\P{^Lb=	_RI}', "");
Expect(0, 127488, '\p{Lb=	_RI}', "");
Expect(1, 127488, '\p{^Lb=	_RI}', "");
Expect(1, 127488, '\P{Lb=	_RI}', "");
Expect(0, 127488, '\P{^Lb=	_RI}', "");
Error('\p{Is_Line_Break:	_-REGIONAL_INDICATOR/a/}');
Error('\P{Is_Line_Break:	_-REGIONAL_INDICATOR/a/}');
Expect(1, 127487, '\p{Is_Line_Break:	regionalindicator}', "");
Expect(0, 127487, '\p{^Is_Line_Break:	regionalindicator}', "");
Expect(0, 127487, '\P{Is_Line_Break:	regionalindicator}', "");
Expect(1, 127487, '\P{^Is_Line_Break:	regionalindicator}', "");
Expect(0, 127488, '\p{Is_Line_Break:	regionalindicator}', "");
Expect(1, 127488, '\p{^Is_Line_Break:	regionalindicator}', "");
Expect(1, 127488, '\P{Is_Line_Break:	regionalindicator}', "");
Expect(0, 127488, '\P{^Is_Line_Break:	regionalindicator}', "");
Expect(1, 127487, '\p{Is_Line_Break=_Regional_indicator}', "");
Expect(0, 127487, '\p{^Is_Line_Break=_Regional_indicator}', "");
Expect(0, 127487, '\P{Is_Line_Break=_Regional_indicator}', "");
Expect(1, 127487, '\P{^Is_Line_Break=_Regional_indicator}', "");
Expect(0, 127488, '\p{Is_Line_Break=_Regional_indicator}', "");
Expect(1, 127488, '\p{^Is_Line_Break=_Regional_indicator}', "");
Expect(1, 127488, '\P{Is_Line_Break=_Regional_indicator}', "");
Expect(0, 127488, '\P{^Is_Line_Break=_Regional_indicator}', "");
Error('\p{Is_Lb=:=RI}');
Error('\P{Is_Lb=:=RI}');
Expect(1, 127487, '\p{Is_Lb=ri}', "");
Expect(0, 127487, '\p{^Is_Lb=ri}', "");
Expect(0, 127487, '\P{Is_Lb=ri}', "");
Expect(1, 127487, '\P{^Is_Lb=ri}', "");
Expect(0, 127488, '\p{Is_Lb=ri}', "");
Expect(1, 127488, '\p{^Is_Lb=ri}', "");
Expect(1, 127488, '\P{Is_Lb=ri}', "");
Expect(0, 127488, '\P{^Is_Lb=ri}', "");
Expect(1, 127487, '\p{Is_Lb: 		RI}', "");
Expect(0, 127487, '\p{^Is_Lb: 		RI}', "");
Expect(0, 127487, '\P{Is_Lb: 		RI}', "");
Expect(1, 127487, '\P{^Is_Lb: 		RI}', "");
Expect(0, 127488, '\p{Is_Lb: 		RI}', "");
Expect(1, 127488, '\p{^Is_Lb: 		RI}', "");
Expect(1, 127488, '\P{Is_Lb: 		RI}', "");
Expect(0, 127488, '\P{^Is_Lb: 		RI}', "");
Error('\p{Line_Break=_/a/Complex_context}');
Error('\P{Line_Break=_/a/Complex_context}');
Expect(1, 43743, '\p{Line_Break=complexcontext}', "");
Expect(0, 43743, '\p{^Line_Break=complexcontext}', "");
Expect(0, 43743, '\P{Line_Break=complexcontext}', "");
Expect(1, 43743, '\P{^Line_Break=complexcontext}', "");
Expect(0, 43744, '\p{Line_Break=complexcontext}', "");
Expect(1, 43744, '\p{^Line_Break=complexcontext}', "");
Expect(1, 43744, '\P{Line_Break=complexcontext}', "");
Expect(0, 43744, '\P{^Line_Break=complexcontext}', "");
Expect(1, 43743, '\p{Line_Break=-	complex_context}', "");
Expect(0, 43743, '\p{^Line_Break=-	complex_context}', "");
Expect(0, 43743, '\P{Line_Break=-	complex_context}', "");
Expect(1, 43743, '\P{^Line_Break=-	complex_context}', "");
Expect(0, 43744, '\p{Line_Break=-	complex_context}', "");
Expect(1, 43744, '\p{^Line_Break=-	complex_context}', "");
Expect(1, 43744, '\P{Line_Break=-	complex_context}', "");
Expect(0, 43744, '\P{^Line_Break=-	complex_context}', "");
Error('\p{Lb=/a/SA}');
Error('\P{Lb=/a/SA}');
Expect(1, 43743, '\p{Lb=sa}', "");
Expect(0, 43743, '\p{^Lb=sa}', "");
Expect(0, 43743, '\P{Lb=sa}', "");
Expect(1, 43743, '\P{^Lb=sa}', "");
Expect(0, 43744, '\p{Lb=sa}', "");
Expect(1, 43744, '\p{^Lb=sa}', "");
Expect(1, 43744, '\P{Lb=sa}', "");
Expect(0, 43744, '\P{^Lb=sa}', "");
Expect(1, 43743, '\p{Lb=__SA}', "");
Expect(0, 43743, '\p{^Lb=__SA}', "");
Expect(0, 43743, '\P{Lb=__SA}', "");
Expect(1, 43743, '\P{^Lb=__SA}', "");
Expect(0, 43744, '\p{Lb=__SA}', "");
Expect(1, 43744, '\p{^Lb=__SA}', "");
Expect(1, 43744, '\P{Lb=__SA}', "");
Expect(0, 43744, '\P{^Lb=__SA}', "");
Error('\p{Is_Line_Break=/a/_	complex_context}');
Error('\P{Is_Line_Break=/a/_	complex_context}');
Expect(1, 43743, '\p{Is_Line_Break=complexcontext}', "");
Expect(0, 43743, '\p{^Is_Line_Break=complexcontext}', "");
Expect(0, 43743, '\P{Is_Line_Break=complexcontext}', "");
Expect(1, 43743, '\P{^Is_Line_Break=complexcontext}', "");
Expect(0, 43744, '\p{Is_Line_Break=complexcontext}', "");
Expect(1, 43744, '\p{^Is_Line_Break=complexcontext}', "");
Expect(1, 43744, '\P{Is_Line_Break=complexcontext}', "");
Expect(0, 43744, '\P{^Is_Line_Break=complexcontext}', "");
Expect(1, 43743, '\p{Is_Line_Break: 	Complex_Context}', "");
Expect(0, 43743, '\p{^Is_Line_Break: 	Complex_Context}', "");
Expect(0, 43743, '\P{Is_Line_Break: 	Complex_Context}', "");
Expect(1, 43743, '\P{^Is_Line_Break: 	Complex_Context}', "");
Expect(0, 43744, '\p{Is_Line_Break: 	Complex_Context}', "");
Expect(1, 43744, '\p{^Is_Line_Break: 	Complex_Context}', "");
Expect(1, 43744, '\P{Is_Line_Break: 	Complex_Context}', "");
Expect(0, 43744, '\P{^Is_Line_Break: 	Complex_Context}', "");
Error('\p{Is_Lb=_	sa/a/}');
Error('\P{Is_Lb=_	sa/a/}');
Expect(1, 43743, '\p{Is_Lb=sa}', "");
Expect(0, 43743, '\p{^Is_Lb=sa}', "");
Expect(0, 43743, '\P{Is_Lb=sa}', "");
Expect(1, 43743, '\P{^Is_Lb=sa}', "");
Expect(0, 43744, '\p{Is_Lb=sa}', "");
Expect(1, 43744, '\p{^Is_Lb=sa}', "");
Expect(1, 43744, '\P{Is_Lb=sa}', "");
Expect(0, 43744, '\P{^Is_Lb=sa}', "");
Expect(1, 43743, '\p{Is_Lb= SA}', "");
Expect(0, 43743, '\p{^Is_Lb= SA}', "");
Expect(0, 43743, '\P{Is_Lb= SA}', "");
Expect(1, 43743, '\P{^Is_Lb= SA}', "");
Expect(0, 43744, '\p{Is_Lb= SA}', "");
Expect(1, 43744, '\p{^Is_Lb= SA}', "");
Expect(1, 43744, '\P{Is_Lb= SA}', "");
Expect(0, 43744, '\P{^Is_Lb= SA}', "");
Error('\p{Line_Break=:=_ surrogate}');
Error('\P{Line_Break=:=_ surrogate}');
Expect(1, 57343, '\p{Line_Break=surrogate}', 'deprecated');
Expect(0, 57343, '\p{^Line_Break=surrogate}', 'deprecated');
Expect(0, 57343, '\P{Line_Break=surrogate}', 'deprecated');
Expect(1, 57343, '\P{^Line_Break=surrogate}', 'deprecated');
Expect(0, 57344, '\p{Line_Break=surrogate}', 'deprecated');
Expect(1, 57344, '\p{^Line_Break=surrogate}', 'deprecated');
Expect(1, 57344, '\P{Line_Break=surrogate}', 'deprecated');
Expect(0, 57344, '\P{^Line_Break=surrogate}', 'deprecated');
Expect(1, 57343, '\p{Line_Break=-	surrogate}', 'deprecated');
Expect(0, 57343, '\p{^Line_Break=-	surrogate}', 'deprecated');
Expect(0, 57343, '\P{Line_Break=-	surrogate}', 'deprecated');
Expect(1, 57343, '\P{^Line_Break=-	surrogate}', 'deprecated');
Expect(0, 57344, '\p{Line_Break=-	surrogate}', 'deprecated');
Expect(1, 57344, '\p{^Line_Break=-	surrogate}', 'deprecated');
Expect(1, 57344, '\P{Line_Break=-	surrogate}', 'deprecated');
Expect(0, 57344, '\P{^Line_Break=-	surrogate}', 'deprecated');
Error('\p{Lb=/a/_sg}');
Error('\P{Lb=/a/_sg}');
Expect(1, 57343, '\p{Lb=sg}', 'deprecated');
Expect(0, 57343, '\p{^Lb=sg}', 'deprecated');
Expect(0, 57343, '\P{Lb=sg}', 'deprecated');
Expect(1, 57343, '\P{^Lb=sg}', 'deprecated');
Expect(0, 57344, '\p{Lb=sg}', 'deprecated');
Expect(1, 57344, '\p{^Lb=sg}', 'deprecated');
Expect(1, 57344, '\P{Lb=sg}', 'deprecated');
Expect(0, 57344, '\P{^Lb=sg}', 'deprecated');
Expect(1, 57343, '\p{Lb=	_SG}', 'deprecated');
Expect(0, 57343, '\p{^Lb=	_SG}', 'deprecated');
Expect(0, 57343, '\P{Lb=	_SG}', 'deprecated');
Expect(1, 57343, '\P{^Lb=	_SG}', 'deprecated');
Expect(0, 57344, '\p{Lb=	_SG}', 'deprecated');
Expect(1, 57344, '\p{^Lb=	_SG}', 'deprecated');
Expect(1, 57344, '\P{Lb=	_SG}', 'deprecated');
Expect(0, 57344, '\P{^Lb=	_SG}', 'deprecated');
Error('\p{Is_Line_Break=-Surrogate/a/}');
Error('\P{Is_Line_Break=-Surrogate/a/}');
Expect(1, 57343, '\p{Is_Line_Break=surrogate}', 'deprecated');
Expect(0, 57343, '\p{^Is_Line_Break=surrogate}', 'deprecated');
Expect(0, 57343, '\P{Is_Line_Break=surrogate}', 'deprecated');
Expect(1, 57343, '\P{^Is_Line_Break=surrogate}', 'deprecated');
Expect(0, 57344, '\p{Is_Line_Break=surrogate}', 'deprecated');
Expect(1, 57344, '\p{^Is_Line_Break=surrogate}', 'deprecated');
Expect(1, 57344, '\P{Is_Line_Break=surrogate}', 'deprecated');
Expect(0, 57344, '\P{^Is_Line_Break=surrogate}', 'deprecated');
Expect(1, 57343, '\p{Is_Line_Break:   -Surrogate}', 'deprecated');
Expect(0, 57343, '\p{^Is_Line_Break:   -Surrogate}', 'deprecated');
Expect(0, 57343, '\P{Is_Line_Break:   -Surrogate}', 'deprecated');
Expect(1, 57343, '\P{^Is_Line_Break:   -Surrogate}', 'deprecated');
Expect(0, 57344, '\p{Is_Line_Break:   -Surrogate}', 'deprecated');
Expect(1, 57344, '\p{^Is_Line_Break:   -Surrogate}', 'deprecated');
Expect(1, 57344, '\P{Is_Line_Break:   -Surrogate}', 'deprecated');
Expect(0, 57344, '\P{^Is_Line_Break:   -Surrogate}', 'deprecated');
Error('\p{Is_Lb= /a/SG}');
Error('\P{Is_Lb= /a/SG}');
Expect(1, 57343, '\p{Is_Lb=sg}', 'deprecated');
Expect(0, 57343, '\p{^Is_Lb=sg}', 'deprecated');
Expect(0, 57343, '\P{Is_Lb=sg}', 'deprecated');
Expect(1, 57343, '\P{^Is_Lb=sg}', 'deprecated');
Expect(0, 57344, '\p{Is_Lb=sg}', 'deprecated');
Expect(1, 57344, '\p{^Is_Lb=sg}', 'deprecated');
Expect(1, 57344, '\P{Is_Lb=sg}', 'deprecated');
Expect(0, 57344, '\P{^Is_Lb=sg}', 'deprecated');
Expect(1, 57343, '\p{Is_Lb=- sg}', 'deprecated');
Expect(0, 57343, '\p{^Is_Lb=- sg}', 'deprecated');
Expect(0, 57343, '\P{Is_Lb=- sg}', 'deprecated');
Expect(1, 57343, '\P{^Is_Lb=- sg}', 'deprecated');
Expect(0, 57344, '\p{Is_Lb=- sg}', 'deprecated');
Expect(1, 57344, '\p{^Is_Lb=- sg}', 'deprecated');
Expect(1, 57344, '\P{Is_Lb=- sg}', 'deprecated');
Expect(0, 57344, '\P{^Is_Lb=- sg}', 'deprecated');
Error('\p{Line_Break: _:=Space}');
Error('\P{Line_Break: _:=Space}');
Expect(1, 32, '\p{Line_Break=space}', "");
Expect(0, 32, '\p{^Line_Break=space}', "");
Expect(0, 32, '\P{Line_Break=space}', "");
Expect(1, 32, '\P{^Line_Break=space}', "");
Expect(0, 33, '\p{Line_Break=space}', "");
Expect(1, 33, '\p{^Line_Break=space}', "");
Expect(1, 33, '\P{Line_Break=space}', "");
Expect(0, 33, '\P{^Line_Break=space}', "");
Expect(1, 32, '\p{Line_Break=	SPACE}', "");
Expect(0, 32, '\p{^Line_Break=	SPACE}', "");
Expect(0, 32, '\P{Line_Break=	SPACE}', "");
Expect(1, 32, '\P{^Line_Break=	SPACE}', "");
Expect(0, 33, '\p{Line_Break=	SPACE}', "");
Expect(1, 33, '\p{^Line_Break=	SPACE}', "");
Expect(1, 33, '\P{Line_Break=	SPACE}', "");
Expect(0, 33, '\P{^Line_Break=	SPACE}', "");
Error('\p{Lb=__SP/a/}');
Error('\P{Lb=__SP/a/}');
Expect(1, 32, '\p{Lb=sp}', "");
Expect(0, 32, '\p{^Lb=sp}', "");
Expect(0, 32, '\P{Lb=sp}', "");
Expect(1, 32, '\P{^Lb=sp}', "");
Expect(0, 33, '\p{Lb=sp}', "");
Expect(1, 33, '\p{^Lb=sp}', "");
Expect(1, 33, '\P{Lb=sp}', "");
Expect(0, 33, '\P{^Lb=sp}', "");
Expect(1, 32, '\p{Lb=SP}', "");
Expect(0, 32, '\p{^Lb=SP}', "");
Expect(0, 32, '\P{Lb=SP}', "");
Expect(1, 32, '\P{^Lb=SP}', "");
Expect(0, 33, '\p{Lb=SP}', "");
Expect(1, 33, '\p{^Lb=SP}', "");
Expect(1, 33, '\P{Lb=SP}', "");
Expect(0, 33, '\P{^Lb=SP}', "");
Error('\p{Is_Line_Break=:=-space}');
Error('\P{Is_Line_Break=:=-space}');
Expect(1, 32, '\p{Is_Line_Break=space}', "");
Expect(0, 32, '\p{^Is_Line_Break=space}', "");
Expect(0, 32, '\P{Is_Line_Break=space}', "");
Expect(1, 32, '\P{^Is_Line_Break=space}', "");
Expect(0, 33, '\p{Is_Line_Break=space}', "");
Expect(1, 33, '\p{^Is_Line_Break=space}', "");
Expect(1, 33, '\P{Is_Line_Break=space}', "");
Expect(0, 33, '\P{^Is_Line_Break=space}', "");
Expect(1, 32, '\p{Is_Line_Break=--Space}', "");
Expect(0, 32, '\p{^Is_Line_Break=--Space}', "");
Expect(0, 32, '\P{Is_Line_Break=--Space}', "");
Expect(1, 32, '\P{^Is_Line_Break=--Space}', "");
Expect(0, 33, '\p{Is_Line_Break=--Space}', "");
Expect(1, 33, '\p{^Is_Line_Break=--Space}', "");
Expect(1, 33, '\P{Is_Line_Break=--Space}', "");
Expect(0, 33, '\P{^Is_Line_Break=--Space}', "");
Error('\p{Is_Lb=	/a/SP}');
Error('\P{Is_Lb=	/a/SP}');
Expect(1, 32, '\p{Is_Lb=sp}', "");
Expect(0, 32, '\p{^Is_Lb=sp}', "");
Expect(0, 32, '\P{Is_Lb=sp}', "");
Expect(1, 32, '\P{^Is_Lb=sp}', "");
Expect(0, 33, '\p{Is_Lb=sp}', "");
Expect(1, 33, '\p{^Is_Lb=sp}', "");
Expect(1, 33, '\P{Is_Lb=sp}', "");
Expect(0, 33, '\P{^Is_Lb=sp}', "");
Expect(1, 32, '\p{Is_Lb= 	SP}', "");
Expect(0, 32, '\p{^Is_Lb= 	SP}', "");
Expect(0, 32, '\P{Is_Lb= 	SP}', "");
Expect(1, 32, '\P{^Is_Lb= 	SP}', "");
Expect(0, 33, '\p{Is_Lb= 	SP}', "");
Expect(1, 33, '\p{^Is_Lb= 	SP}', "");
Expect(1, 33, '\P{Is_Lb= 	SP}', "");
Expect(0, 33, '\P{^Is_Lb= 	SP}', "");
Error('\p{Line_Break=	-break_SYMBOLS:=}');
Error('\P{Line_Break=	-break_SYMBOLS:=}');
Expect(1, 47, '\p{Line_Break=breaksymbols}', "");
Expect(0, 47, '\p{^Line_Break=breaksymbols}', "");
Expect(0, 47, '\P{Line_Break=breaksymbols}', "");
Expect(1, 47, '\P{^Line_Break=breaksymbols}', "");
Expect(0, 48, '\p{Line_Break=breaksymbols}', "");
Expect(1, 48, '\p{^Line_Break=breaksymbols}', "");
Expect(1, 48, '\P{Line_Break=breaksymbols}', "");
Expect(0, 48, '\P{^Line_Break=breaksymbols}', "");
Expect(1, 47, '\p{Line_Break=	_Break_Symbols}', "");
Expect(0, 47, '\p{^Line_Break=	_Break_Symbols}', "");
Expect(0, 47, '\P{Line_Break=	_Break_Symbols}', "");
Expect(1, 47, '\P{^Line_Break=	_Break_Symbols}', "");
Expect(0, 48, '\p{Line_Break=	_Break_Symbols}', "");
Expect(1, 48, '\p{^Line_Break=	_Break_Symbols}', "");
Expect(1, 48, '\P{Line_Break=	_Break_Symbols}', "");
Expect(0, 48, '\P{^Line_Break=	_Break_Symbols}', "");
Error('\p{Lb=_	SY/a/}');
Error('\P{Lb=_	SY/a/}');
Expect(1, 47, '\p{Lb:sy}', "");
Expect(0, 47, '\p{^Lb:sy}', "");
Expect(0, 47, '\P{Lb:sy}', "");
Expect(1, 47, '\P{^Lb:sy}', "");
Expect(0, 48, '\p{Lb:sy}', "");
Expect(1, 48, '\p{^Lb:sy}', "");
Expect(1, 48, '\P{Lb:sy}', "");
Expect(0, 48, '\P{^Lb:sy}', "");
Expect(1, 47, '\p{Lb=_SY}', "");
Expect(0, 47, '\p{^Lb=_SY}', "");
Expect(0, 47, '\P{Lb=_SY}', "");
Expect(1, 47, '\P{^Lb=_SY}', "");
Expect(0, 48, '\p{Lb=_SY}', "");
Expect(1, 48, '\p{^Lb=_SY}', "");
Expect(1, 48, '\P{Lb=_SY}', "");
Expect(0, 48, '\P{^Lb=_SY}', "");
Error('\p{Is_Line_Break=_Break_Symbols:=}');
Error('\P{Is_Line_Break=_Break_Symbols:=}');
Expect(1, 47, '\p{Is_Line_Break=breaksymbols}', "");
Expect(0, 47, '\p{^Is_Line_Break=breaksymbols}', "");
Expect(0, 47, '\P{Is_Line_Break=breaksymbols}', "");
Expect(1, 47, '\P{^Is_Line_Break=breaksymbols}', "");
Expect(0, 48, '\p{Is_Line_Break=breaksymbols}', "");
Expect(1, 48, '\p{^Is_Line_Break=breaksymbols}', "");
Expect(1, 48, '\P{Is_Line_Break=breaksymbols}', "");
Expect(0, 48, '\P{^Is_Line_Break=breaksymbols}', "");
Expect(1, 47, '\p{Is_Line_Break=--BREAK_SYMBOLS}', "");
Expect(0, 47, '\p{^Is_Line_Break=--BREAK_SYMBOLS}', "");
Expect(0, 47, '\P{Is_Line_Break=--BREAK_SYMBOLS}', "");
Expect(1, 47, '\P{^Is_Line_Break=--BREAK_SYMBOLS}', "");
Expect(0, 48, '\p{Is_Line_Break=--BREAK_SYMBOLS}', "");
Expect(1, 48, '\p{^Is_Line_Break=--BREAK_SYMBOLS}', "");
Expect(1, 48, '\P{Is_Line_Break=--BREAK_SYMBOLS}', "");
Expect(0, 48, '\P{^Is_Line_Break=--BREAK_SYMBOLS}', "");
Error('\p{Is_Lb=:= 	SY}');
Error('\P{Is_Lb=:= 	SY}');
Expect(1, 47, '\p{Is_Lb=sy}', "");
Expect(0, 47, '\p{^Is_Lb=sy}', "");
Expect(0, 47, '\P{Is_Lb=sy}', "");
Expect(1, 47, '\P{^Is_Lb=sy}', "");
Expect(0, 48, '\p{Is_Lb=sy}', "");
Expect(1, 48, '\p{^Is_Lb=sy}', "");
Expect(1, 48, '\P{Is_Lb=sy}', "");
Expect(0, 48, '\P{^Is_Lb=sy}', "");
Expect(1, 47, '\p{Is_Lb=-	SY}', "");
Expect(0, 47, '\p{^Is_Lb=-	SY}', "");
Expect(0, 47, '\P{Is_Lb=-	SY}', "");
Expect(1, 47, '\P{^Is_Lb=-	SY}', "");
Expect(0, 48, '\p{Is_Lb=-	SY}', "");
Expect(1, 48, '\p{^Is_Lb=-	SY}', "");
Expect(1, 48, '\P{Is_Lb=-	SY}', "");
Expect(0, 48, '\P{^Is_Lb=-	SY}', "");
Error('\p{Line_Break=	Word_joiner:=}');
Error('\P{Line_Break=	Word_joiner:=}');
Expect(1, 65279, '\p{Line_Break=wordjoiner}', "");
Expect(0, 65279, '\p{^Line_Break=wordjoiner}', "");
Expect(0, 65279, '\P{Line_Break=wordjoiner}', "");
Expect(1, 65279, '\P{^Line_Break=wordjoiner}', "");
Expect(0, 65280, '\p{Line_Break=wordjoiner}', "");
Expect(1, 65280, '\p{^Line_Break=wordjoiner}', "");
Expect(1, 65280, '\P{Line_Break=wordjoiner}', "");
Expect(0, 65280, '\P{^Line_Break=wordjoiner}', "");
Expect(1, 65279, '\p{Line_Break:   -	word_joiner}', "");
Expect(0, 65279, '\p{^Line_Break:   -	word_joiner}', "");
Expect(0, 65279, '\P{Line_Break:   -	word_joiner}', "");
Expect(1, 65279, '\P{^Line_Break:   -	word_joiner}', "");
Expect(0, 65280, '\p{Line_Break:   -	word_joiner}', "");
Expect(1, 65280, '\p{^Line_Break:   -	word_joiner}', "");
Expect(1, 65280, '\P{Line_Break:   -	word_joiner}', "");
Expect(0, 65280, '\P{^Line_Break:   -	word_joiner}', "");
Error('\p{Lb=	:=wj}');
Error('\P{Lb=	:=wj}');
Expect(1, 65279, '\p{Lb=wj}', "");
Expect(0, 65279, '\p{^Lb=wj}', "");
Expect(0, 65279, '\P{Lb=wj}', "");
Expect(1, 65279, '\P{^Lb=wj}', "");
Expect(0, 65280, '\p{Lb=wj}', "");
Expect(1, 65280, '\p{^Lb=wj}', "");
Expect(1, 65280, '\P{Lb=wj}', "");
Expect(0, 65280, '\P{^Lb=wj}', "");
Expect(1, 65279, '\p{Lb: -_WJ}', "");
Expect(0, 65279, '\p{^Lb: -_WJ}', "");
Expect(0, 65279, '\P{Lb: -_WJ}', "");
Expect(1, 65279, '\P{^Lb: -_WJ}', "");
Expect(0, 65280, '\p{Lb: -_WJ}', "");
Expect(1, 65280, '\p{^Lb: -_WJ}', "");
Expect(1, 65280, '\P{Lb: -_WJ}', "");
Expect(0, 65280, '\P{^Lb: -_WJ}', "");
Error('\p{Is_Line_Break=	Word_Joiner:=}');
Error('\P{Is_Line_Break=	Word_Joiner:=}');
Expect(1, 65279, '\p{Is_Line_Break=wordjoiner}', "");
Expect(0, 65279, '\p{^Is_Line_Break=wordjoiner}', "");
Expect(0, 65279, '\P{Is_Line_Break=wordjoiner}', "");
Expect(1, 65279, '\P{^Is_Line_Break=wordjoiner}', "");
Expect(0, 65280, '\p{Is_Line_Break=wordjoiner}', "");
Expect(1, 65280, '\p{^Is_Line_Break=wordjoiner}', "");
Expect(1, 65280, '\P{Is_Line_Break=wordjoiner}', "");
Expect(0, 65280, '\P{^Is_Line_Break=wordjoiner}', "");
Expect(1, 65279, '\p{Is_Line_Break= 	word_joiner}', "");
Expect(0, 65279, '\p{^Is_Line_Break= 	word_joiner}', "");
Expect(0, 65279, '\P{Is_Line_Break= 	word_joiner}', "");
Expect(1, 65279, '\P{^Is_Line_Break= 	word_joiner}', "");
Expect(0, 65280, '\p{Is_Line_Break= 	word_joiner}', "");
Expect(1, 65280, '\p{^Is_Line_Break= 	word_joiner}', "");
Expect(1, 65280, '\P{Is_Line_Break= 	word_joiner}', "");
Expect(0, 65280, '\P{^Is_Line_Break= 	word_joiner}', "");
Error('\p{Is_Lb=	_WJ/a/}');
Error('\P{Is_Lb=	_WJ/a/}');
Expect(1, 65279, '\p{Is_Lb=wj}', "");
Expect(0, 65279, '\p{^Is_Lb=wj}', "");
Expect(0, 65279, '\P{Is_Lb=wj}', "");
Expect(1, 65279, '\P{^Is_Lb=wj}', "");
Expect(0, 65280, '\p{Is_Lb=wj}', "");
Expect(1, 65280, '\p{^Is_Lb=wj}', "");
Expect(1, 65280, '\P{Is_Lb=wj}', "");
Expect(0, 65280, '\P{^Is_Lb=wj}', "");
Expect(1, 65279, '\p{Is_Lb=-WJ}', "");
Expect(0, 65279, '\p{^Is_Lb=-WJ}', "");
Expect(0, 65279, '\P{Is_Lb=-WJ}', "");
Expect(1, 65279, '\P{^Is_Lb=-WJ}', "");
Expect(0, 65280, '\p{Is_Lb=-WJ}', "");
Expect(1, 65280, '\p{^Is_Lb=-WJ}', "");
Expect(1, 65280, '\P{Is_Lb=-WJ}', "");
Expect(0, 65280, '\P{^Is_Lb=-WJ}', "");
Error('\p{Line_Break=	:=unknown}');
Error('\P{Line_Break=	:=unknown}');
Expect(1, 918000, '\p{Line_Break=unknown}', "");
Expect(0, 918000, '\p{^Line_Break=unknown}', "");
Expect(0, 918000, '\P{Line_Break=unknown}', "");
Expect(1, 918000, '\P{^Line_Break=unknown}', "");
Expect(0, 917999, '\p{Line_Break=unknown}', "");
Expect(1, 917999, '\p{^Line_Break=unknown}', "");
Expect(1, 917999, '\P{Line_Break=unknown}', "");
Expect(0, 917999, '\P{^Line_Break=unknown}', "");
Expect(1, 918000, '\p{Line_Break=-Unknown}', "");
Expect(0, 918000, '\p{^Line_Break=-Unknown}', "");
Expect(0, 918000, '\P{Line_Break=-Unknown}', "");
Expect(1, 918000, '\P{^Line_Break=-Unknown}', "");
Expect(0, 917999, '\p{Line_Break=-Unknown}', "");
Expect(1, 917999, '\p{^Line_Break=-Unknown}', "");
Expect(1, 917999, '\P{Line_Break=-Unknown}', "");
Expect(0, 917999, '\P{^Line_Break=-Unknown}', "");
Error('\p{Lb=:=XX}');
Error('\P{Lb=:=XX}');
Expect(1, 918000, '\p{Lb=xx}', "");
Expect(0, 918000, '\p{^Lb=xx}', "");
Expect(0, 918000, '\P{Lb=xx}', "");
Expect(1, 918000, '\P{^Lb=xx}', "");
Expect(0, 917999, '\p{Lb=xx}', "");
Expect(1, 917999, '\p{^Lb=xx}', "");
Expect(1, 917999, '\P{Lb=xx}', "");
Expect(0, 917999, '\P{^Lb=xx}', "");
Expect(1, 918000, '\p{Lb=XX}', "");
Expect(0, 918000, '\p{^Lb=XX}', "");
Expect(0, 918000, '\P{Lb=XX}', "");
Expect(1, 918000, '\P{^Lb=XX}', "");
Expect(0, 917999, '\p{Lb=XX}', "");
Expect(1, 917999, '\p{^Lb=XX}', "");
Expect(1, 917999, '\P{Lb=XX}', "");
Expect(0, 917999, '\P{^Lb=XX}', "");
Error('\p{Is_Line_Break=/a/unknown}');
Error('\P{Is_Line_Break=/a/unknown}');
Expect(1, 918000, '\p{Is_Line_Break=unknown}', "");
Expect(0, 918000, '\p{^Is_Line_Break=unknown}', "");
Expect(0, 918000, '\P{Is_Line_Break=unknown}', "");
Expect(1, 918000, '\P{^Is_Line_Break=unknown}', "");
Expect(0, 917999, '\p{Is_Line_Break=unknown}', "");
Expect(1, 917999, '\p{^Is_Line_Break=unknown}', "");
Expect(1, 917999, '\P{Is_Line_Break=unknown}', "");
Expect(0, 917999, '\P{^Is_Line_Break=unknown}', "");
Expect(1, 918000, '\p{Is_Line_Break= unknown}', "");
Expect(0, 918000, '\p{^Is_Line_Break= unknown}', "");
Expect(0, 918000, '\P{Is_Line_Break= unknown}', "");
Expect(1, 918000, '\P{^Is_Line_Break= unknown}', "");
Expect(0, 917999, '\p{Is_Line_Break= unknown}', "");
Expect(1, 917999, '\p{^Is_Line_Break= unknown}', "");
Expect(1, 917999, '\P{Is_Line_Break= unknown}', "");
Expect(0, 917999, '\P{^Is_Line_Break= unknown}', "");
Error('\p{Is_Lb=	 XX/a/}');
Error('\P{Is_Lb=	 XX/a/}');
Expect(1, 918000, '\p{Is_Lb:   xx}', "");
Expect(0, 918000, '\p{^Is_Lb:   xx}', "");
Expect(0, 918000, '\P{Is_Lb:   xx}', "");
Expect(1, 918000, '\P{^Is_Lb:   xx}', "");
Expect(0, 917999, '\p{Is_Lb:   xx}', "");
Expect(1, 917999, '\p{^Is_Lb:   xx}', "");
Expect(1, 917999, '\P{Is_Lb:   xx}', "");
Expect(0, 917999, '\P{^Is_Lb:   xx}', "");
Expect(1, 918000, '\p{Is_Lb= XX}', "");
Expect(0, 918000, '\p{^Is_Lb= XX}', "");
Expect(0, 918000, '\P{Is_Lb= XX}', "");
Expect(1, 918000, '\P{^Is_Lb= XX}', "");
Expect(0, 917999, '\p{Is_Lb= XX}', "");
Expect(1, 917999, '\p{^Is_Lb= XX}', "");
Expect(1, 917999, '\P{Is_Lb= XX}', "");
Expect(0, 917999, '\P{^Is_Lb= XX}', "");
Error('\p{Line_Break=_ZWSpace:=}');
Error('\P{Line_Break=_ZWSpace:=}');
Expect(1, 8203, '\p{Line_Break=zwspace}', "");
Expect(0, 8203, '\p{^Line_Break=zwspace}', "");
Expect(0, 8203, '\P{Line_Break=zwspace}', "");
Expect(1, 8203, '\P{^Line_Break=zwspace}', "");
Expect(0, 8204, '\p{Line_Break=zwspace}', "");
Expect(1, 8204, '\p{^Line_Break=zwspace}', "");
Expect(1, 8204, '\P{Line_Break=zwspace}', "");
Expect(0, 8204, '\P{^Line_Break=zwspace}', "");
Expect(1, 8203, '\p{Line_Break=-_ZWSPACE}', "");
Expect(0, 8203, '\p{^Line_Break=-_ZWSPACE}', "");
Expect(0, 8203, '\P{Line_Break=-_ZWSPACE}', "");
Expect(1, 8203, '\P{^Line_Break=-_ZWSPACE}', "");
Expect(0, 8204, '\p{Line_Break=-_ZWSPACE}', "");
Expect(1, 8204, '\p{^Line_Break=-_ZWSPACE}', "");
Expect(1, 8204, '\P{Line_Break=-_ZWSPACE}', "");
Expect(0, 8204, '\P{^Line_Break=-_ZWSPACE}', "");
Error('\p{Lb=/a/-	ZW}');
Error('\P{Lb=/a/-	ZW}');
Expect(1, 8203, '\p{Lb=zw}', "");
Expect(0, 8203, '\p{^Lb=zw}', "");
Expect(0, 8203, '\P{Lb=zw}', "");
Expect(1, 8203, '\P{^Lb=zw}', "");
Expect(0, 8204, '\p{Lb=zw}', "");
Expect(1, 8204, '\p{^Lb=zw}', "");
Expect(1, 8204, '\P{Lb=zw}', "");
Expect(0, 8204, '\P{^Lb=zw}', "");
Expect(1, 8203, '\p{Lb=	_ZW}', "");
Expect(0, 8203, '\p{^Lb=	_ZW}', "");
Expect(0, 8203, '\P{Lb=	_ZW}', "");
Expect(1, 8203, '\P{^Lb=	_ZW}', "");
Expect(0, 8204, '\p{Lb=	_ZW}', "");
Expect(1, 8204, '\p{^Lb=	_ZW}', "");
Expect(1, 8204, '\P{Lb=	_ZW}', "");
Expect(0, 8204, '\P{^Lb=	_ZW}', "");
Error('\p{Is_Line_Break=/a/ZWSpace}');
Error('\P{Is_Line_Break=/a/ZWSpace}');
Expect(1, 8203, '\p{Is_Line_Break=zwspace}', "");
Expect(0, 8203, '\p{^Is_Line_Break=zwspace}', "");
Expect(0, 8203, '\P{Is_Line_Break=zwspace}', "");
Expect(1, 8203, '\P{^Is_Line_Break=zwspace}', "");
Expect(0, 8204, '\p{Is_Line_Break=zwspace}', "");
Expect(1, 8204, '\p{^Is_Line_Break=zwspace}', "");
Expect(1, 8204, '\P{Is_Line_Break=zwspace}', "");
Expect(0, 8204, '\P{^Is_Line_Break=zwspace}', "");
Expect(1, 8203, '\p{Is_Line_Break=_ZWSpace}', "");
Expect(0, 8203, '\p{^Is_Line_Break=_ZWSpace}', "");
Expect(0, 8203, '\P{Is_Line_Break=_ZWSpace}', "");
Expect(1, 8203, '\P{^Is_Line_Break=_ZWSpace}', "");
Expect(0, 8204, '\p{Is_Line_Break=_ZWSpace}', "");
Expect(1, 8204, '\p{^Is_Line_Break=_ZWSpace}', "");
Expect(1, 8204, '\P{Is_Line_Break=_ZWSpace}', "");
Expect(0, 8204, '\P{^Is_Line_Break=_ZWSpace}', "");
Error('\p{Is_Lb=/a/		ZW}');
Error('\P{Is_Lb=/a/		ZW}');
Expect(1, 8203, '\p{Is_Lb=zw}', "");
Expect(0, 8203, '\p{^Is_Lb=zw}', "");
Expect(0, 8203, '\P{Is_Lb=zw}', "");
Expect(1, 8203, '\P{^Is_Lb=zw}', "");
Expect(0, 8204, '\p{Is_Lb=zw}', "");
Expect(1, 8204, '\p{^Is_Lb=zw}', "");
Expect(1, 8204, '\P{Is_Lb=zw}', "");
Expect(0, 8204, '\P{^Is_Lb=zw}', "");
Expect(1, 8203, '\p{Is_Lb=-ZW}', "");
Expect(0, 8203, '\p{^Is_Lb=-ZW}', "");
Expect(0, 8203, '\P{Is_Lb=-ZW}', "");
Expect(1, 8203, '\P{^Is_Lb=-ZW}', "");
Expect(0, 8204, '\p{Is_Lb=-ZW}', "");
Expect(1, 8204, '\p{^Is_Lb=-ZW}', "");
Expect(1, 8204, '\P{Is_Lb=-ZW}', "");
Expect(0, 8204, '\P{^Is_Lb=-ZW}', "");
Error('\p{Numeric_Type=:=__Decimal}');
Error('\P{Numeric_Type=:=__Decimal}');
Expect(1, 120831, '\p{Numeric_Type=decimal}', "");
Expect(0, 120831, '\p{^Numeric_Type=decimal}', "");
Expect(0, 120831, '\P{Numeric_Type=decimal}', "");
Expect(1, 120831, '\P{^Numeric_Type=decimal}', "");
Expect(0, 120832, '\p{Numeric_Type=decimal}', "");
Expect(1, 120832, '\p{^Numeric_Type=decimal}', "");
Expect(1, 120832, '\P{Numeric_Type=decimal}', "");
Expect(0, 120832, '\P{^Numeric_Type=decimal}', "");
Expect(1, 120831, '\p{Numeric_Type= DECIMAL}', "");
Expect(0, 120831, '\p{^Numeric_Type= DECIMAL}', "");
Expect(0, 120831, '\P{Numeric_Type= DECIMAL}', "");
Expect(1, 120831, '\P{^Numeric_Type= DECIMAL}', "");
Expect(0, 120832, '\p{Numeric_Type= DECIMAL}', "");
Expect(1, 120832, '\p{^Numeric_Type= DECIMAL}', "");
Expect(1, 120832, '\P{Numeric_Type= DECIMAL}', "");
Expect(0, 120832, '\P{^Numeric_Type= DECIMAL}', "");
Error('\p{Nt=-De:=}');
Error('\P{Nt=-De:=}');
Expect(1, 120831, '\p{Nt=de}', "");
Expect(0, 120831, '\p{^Nt=de}', "");
Expect(0, 120831, '\P{Nt=de}', "");
Expect(1, 120831, '\P{^Nt=de}', "");
Expect(0, 120832, '\p{Nt=de}', "");
Expect(1, 120832, '\p{^Nt=de}', "");
Expect(1, 120832, '\P{Nt=de}', "");
Expect(0, 120832, '\P{^Nt=de}', "");
Expect(1, 120831, '\p{Nt=	DE}', "");
Expect(0, 120831, '\p{^Nt=	DE}', "");
Expect(0, 120831, '\P{Nt=	DE}', "");
Expect(1, 120831, '\P{^Nt=	DE}', "");
Expect(0, 120832, '\p{Nt=	DE}', "");
Expect(1, 120832, '\p{^Nt=	DE}', "");
Expect(1, 120832, '\P{Nt=	DE}', "");
Expect(0, 120832, '\P{^Nt=	DE}', "");
Error('\p{Is_Numeric_Type:-_decimal:=}');
Error('\P{Is_Numeric_Type:-_decimal:=}');
Expect(1, 120831, '\p{Is_Numeric_Type=decimal}', "");
Expect(0, 120831, '\p{^Is_Numeric_Type=decimal}', "");
Expect(0, 120831, '\P{Is_Numeric_Type=decimal}', "");
Expect(1, 120831, '\P{^Is_Numeric_Type=decimal}', "");
Expect(0, 120832, '\p{Is_Numeric_Type=decimal}', "");
Expect(1, 120832, '\p{^Is_Numeric_Type=decimal}', "");
Expect(1, 120832, '\P{Is_Numeric_Type=decimal}', "");
Expect(0, 120832, '\P{^Is_Numeric_Type=decimal}', "");
Expect(1, 120831, '\p{Is_Numeric_Type= -decimal}', "");
Expect(0, 120831, '\p{^Is_Numeric_Type= -decimal}', "");
Expect(0, 120831, '\P{Is_Numeric_Type= -decimal}', "");
Expect(1, 120831, '\P{^Is_Numeric_Type= -decimal}', "");
Expect(0, 120832, '\p{Is_Numeric_Type= -decimal}', "");
Expect(1, 120832, '\p{^Is_Numeric_Type= -decimal}', "");
Expect(1, 120832, '\P{Is_Numeric_Type= -decimal}', "");
Expect(0, 120832, '\P{^Is_Numeric_Type= -decimal}', "");
Error('\p{Is_Nt=	/a/De}');
Error('\P{Is_Nt=	/a/De}');
Expect(1, 120831, '\p{Is_Nt=de}', "");
Expect(0, 120831, '\p{^Is_Nt=de}', "");
Expect(0, 120831, '\P{Is_Nt=de}', "");
Expect(1, 120831, '\P{^Is_Nt=de}', "");
Expect(0, 120832, '\p{Is_Nt=de}', "");
Expect(1, 120832, '\p{^Is_Nt=de}', "");
Expect(1, 120832, '\P{Is_Nt=de}', "");
Expect(0, 120832, '\P{^Is_Nt=de}', "");
Expect(1, 120831, '\p{Is_Nt=	De}', "");
Expect(0, 120831, '\p{^Is_Nt=	De}', "");
Expect(0, 120831, '\P{Is_Nt=	De}', "");
Expect(1, 120831, '\P{^Is_Nt=	De}', "");
Expect(0, 120832, '\p{Is_Nt=	De}', "");
Expect(1, 120832, '\p{^Is_Nt=	De}', "");
Expect(1, 120832, '\P{Is_Nt=	De}', "");
Expect(0, 120832, '\P{^Is_Nt=	De}', "");
Error('\p{Numeric_Type=:=__digit}');
Error('\P{Numeric_Type=:=__digit}');
Expect(1, 127242, '\p{Numeric_Type=digit}', "");
Expect(0, 127242, '\p{^Numeric_Type=digit}', "");
Expect(0, 127242, '\P{Numeric_Type=digit}', "");
Expect(1, 127242, '\P{^Numeric_Type=digit}', "");
Expect(0, 127243, '\p{Numeric_Type=digit}', "");
Expect(1, 127243, '\p{^Numeric_Type=digit}', "");
Expect(1, 127243, '\P{Numeric_Type=digit}', "");
Expect(0, 127243, '\P{^Numeric_Type=digit}', "");
Expect(1, 127242, '\p{Numeric_Type=--DIGIT}', "");
Expect(0, 127242, '\p{^Numeric_Type=--DIGIT}', "");
Expect(0, 127242, '\P{Numeric_Type=--DIGIT}', "");
Expect(1, 127242, '\P{^Numeric_Type=--DIGIT}', "");
Expect(0, 127243, '\p{Numeric_Type=--DIGIT}', "");
Expect(1, 127243, '\p{^Numeric_Type=--DIGIT}', "");
Expect(1, 127243, '\P{Numeric_Type=--DIGIT}', "");
Expect(0, 127243, '\P{^Numeric_Type=--DIGIT}', "");
Error('\p{Nt=:=_-di}');
Error('\P{Nt=:=_-di}');
Expect(1, 127242, '\p{Nt=di}', "");
Expect(0, 127242, '\p{^Nt=di}', "");
Expect(0, 127242, '\P{Nt=di}', "");
Expect(1, 127242, '\P{^Nt=di}', "");
Expect(0, 127243, '\p{Nt=di}', "");
Expect(1, 127243, '\p{^Nt=di}', "");
Expect(1, 127243, '\P{Nt=di}', "");
Expect(0, 127243, '\P{^Nt=di}', "");
Expect(1, 127242, '\p{Nt: 	DI}', "");
Expect(0, 127242, '\p{^Nt: 	DI}', "");
Expect(0, 127242, '\P{Nt: 	DI}', "");
Expect(1, 127242, '\P{^Nt: 	DI}', "");
Expect(0, 127243, '\p{Nt: 	DI}', "");
Expect(1, 127243, '\p{^Nt: 	DI}', "");
Expect(1, 127243, '\P{Nt: 	DI}', "");
Expect(0, 127243, '\P{^Nt: 	DI}', "");
Error('\p{Is_Numeric_Type=:=		Digit}');
Error('\P{Is_Numeric_Type=:=		Digit}');
Expect(1, 127242, '\p{Is_Numeric_Type=digit}', "");
Expect(0, 127242, '\p{^Is_Numeric_Type=digit}', "");
Expect(0, 127242, '\P{Is_Numeric_Type=digit}', "");
Expect(1, 127242, '\P{^Is_Numeric_Type=digit}', "");
Expect(0, 127243, '\p{Is_Numeric_Type=digit}', "");
Expect(1, 127243, '\p{^Is_Numeric_Type=digit}', "");
Expect(1, 127243, '\P{Is_Numeric_Type=digit}', "");
Expect(0, 127243, '\P{^Is_Numeric_Type=digit}', "");
Expect(1, 127242, '\p{Is_Numeric_Type:  DIGIT}', "");
Expect(0, 127242, '\p{^Is_Numeric_Type:  DIGIT}', "");
Expect(0, 127242, '\P{Is_Numeric_Type:  DIGIT}', "");
Expect(1, 127242, '\P{^Is_Numeric_Type:  DIGIT}', "");
Expect(0, 127243, '\p{Is_Numeric_Type:  DIGIT}', "");
Expect(1, 127243, '\p{^Is_Numeric_Type:  DIGIT}', "");
Expect(1, 127243, '\P{Is_Numeric_Type:  DIGIT}', "");
Expect(0, 127243, '\P{^Is_Numeric_Type:  DIGIT}', "");
Error('\p{Is_Nt=	 Di/a/}');
Error('\P{Is_Nt=	 Di/a/}');
Expect(1, 127242, '\p{Is_Nt:di}', "");
Expect(0, 127242, '\p{^Is_Nt:di}', "");
Expect(0, 127242, '\P{Is_Nt:di}', "");
Expect(1, 127242, '\P{^Is_Nt:di}', "");
Expect(0, 127243, '\p{Is_Nt:di}', "");
Expect(1, 127243, '\p{^Is_Nt:di}', "");
Expect(1, 127243, '\P{Is_Nt:di}', "");
Expect(0, 127243, '\P{^Is_Nt:di}', "");
Error('\p{Numeric_Type=/a/	None}');
Error('\P{Numeric_Type=/a/	None}');
Expect(1, 194705, '\p{Numeric_Type=none}', "");
Expect(0, 194705, '\p{^Numeric_Type=none}', "");
Expect(0, 194705, '\P{Numeric_Type=none}', "");
Expect(1, 194705, '\P{^Numeric_Type=none}', "");
Expect(0, 194704, '\p{Numeric_Type=none}', "");
Expect(1, 194704, '\p{^Numeric_Type=none}', "");
Expect(1, 194704, '\P{Numeric_Type=none}', "");
Expect(0, 194704, '\P{^Numeric_Type=none}', "");
Expect(1, 194705, '\p{Numeric_Type:   --none}', "");
Expect(0, 194705, '\p{^Numeric_Type:   --none}', "");
Expect(0, 194705, '\P{Numeric_Type:   --none}', "");
Expect(1, 194705, '\P{^Numeric_Type:   --none}', "");
Expect(0, 194704, '\p{Numeric_Type:   --none}', "");
Expect(1, 194704, '\p{^Numeric_Type:   --none}', "");
Expect(1, 194704, '\P{Numeric_Type:   --none}', "");
Expect(0, 194704, '\P{^Numeric_Type:   --none}', "");
Error('\p{Nt=-_NONE/a/}');
Error('\P{Nt=-_NONE/a/}');
Expect(1, 194705, '\p{Nt=none}', "");
Expect(0, 194705, '\p{^Nt=none}', "");
Expect(0, 194705, '\P{Nt=none}', "");
Expect(1, 194705, '\P{^Nt=none}', "");
Expect(0, 194704, '\p{Nt=none}', "");
Expect(1, 194704, '\p{^Nt=none}', "");
Expect(1, 194704, '\P{Nt=none}', "");
Expect(0, 194704, '\P{^Nt=none}', "");
Expect(1, 194705, '\p{Nt:	-NONE}', "");
Expect(0, 194705, '\p{^Nt:	-NONE}', "");
Expect(0, 194705, '\P{Nt:	-NONE}', "");
Expect(1, 194705, '\P{^Nt:	-NONE}', "");
Expect(0, 194704, '\p{Nt:	-NONE}', "");
Expect(1, 194704, '\p{^Nt:	-NONE}', "");
Expect(1, 194704, '\P{Nt:	-NONE}', "");
Expect(0, 194704, '\P{^Nt:	-NONE}', "");
Error('\p{Is_Numeric_Type=:=None}');
Error('\P{Is_Numeric_Type=:=None}');
Expect(1, 194705, '\p{Is_Numeric_Type=none}', "");
Expect(0, 194705, '\p{^Is_Numeric_Type=none}', "");
Expect(0, 194705, '\P{Is_Numeric_Type=none}', "");
Expect(1, 194705, '\P{^Is_Numeric_Type=none}', "");
Expect(0, 194704, '\p{Is_Numeric_Type=none}', "");
Expect(1, 194704, '\p{^Is_Numeric_Type=none}', "");
Expect(1, 194704, '\P{Is_Numeric_Type=none}', "");
Expect(0, 194704, '\P{^Is_Numeric_Type=none}', "");
Expect(1, 194705, '\p{Is_Numeric_Type=_None}', "");
Expect(0, 194705, '\p{^Is_Numeric_Type=_None}', "");
Expect(0, 194705, '\P{Is_Numeric_Type=_None}', "");
Expect(1, 194705, '\P{^Is_Numeric_Type=_None}', "");
Expect(0, 194704, '\p{Is_Numeric_Type=_None}', "");
Expect(1, 194704, '\p{^Is_Numeric_Type=_None}', "");
Expect(1, 194704, '\P{Is_Numeric_Type=_None}', "");
Expect(0, 194704, '\P{^Is_Numeric_Type=_None}', "");
Error('\p{Is_Nt::=None}');
Error('\P{Is_Nt::=None}');
Expect(1, 194705, '\p{Is_Nt=none}', "");
Expect(0, 194705, '\p{^Is_Nt=none}', "");
Expect(0, 194705, '\P{Is_Nt=none}', "");
Expect(1, 194705, '\P{^Is_Nt=none}', "");
Expect(0, 194704, '\p{Is_Nt=none}', "");
Expect(1, 194704, '\p{^Is_Nt=none}', "");
Expect(1, 194704, '\P{Is_Nt=none}', "");
Expect(0, 194704, '\P{^Is_Nt=none}', "");
Expect(1, 194705, '\p{Is_Nt=_-None}', "");
Expect(0, 194705, '\p{^Is_Nt=_-None}', "");
Expect(0, 194705, '\P{Is_Nt=_-None}', "");
Expect(1, 194705, '\P{^Is_Nt=_-None}', "");
Expect(0, 194704, '\p{Is_Nt=_-None}', "");
Expect(1, 194704, '\p{^Is_Nt=_-None}', "");
Expect(1, 194704, '\P{Is_Nt=_-None}', "");
Expect(0, 194704, '\P{^Is_Nt=_-None}', "");
Error('\p{Numeric_Type=/a/Numeric}');
Error('\P{Numeric_Type=/a/Numeric}');
Expect(1, 194704, '\p{Numeric_Type:   numeric}', "");
Expect(0, 194704, '\p{^Numeric_Type:   numeric}', "");
Expect(0, 194704, '\P{Numeric_Type:   numeric}', "");
Expect(1, 194704, '\P{^Numeric_Type:   numeric}', "");
Expect(0, 194705, '\p{Numeric_Type:   numeric}', "");
Expect(1, 194705, '\p{^Numeric_Type:   numeric}', "");
Expect(1, 194705, '\P{Numeric_Type:   numeric}', "");
Expect(0, 194705, '\P{^Numeric_Type:   numeric}', "");
Expect(1, 194704, '\p{Numeric_Type=_ NUMERIC}', "");
Expect(0, 194704, '\p{^Numeric_Type=_ NUMERIC}', "");
Expect(0, 194704, '\P{Numeric_Type=_ NUMERIC}', "");
Expect(1, 194704, '\P{^Numeric_Type=_ NUMERIC}', "");
Expect(0, 194705, '\p{Numeric_Type=_ NUMERIC}', "");
Expect(1, 194705, '\p{^Numeric_Type=_ NUMERIC}', "");
Expect(1, 194705, '\P{Numeric_Type=_ NUMERIC}', "");
Expect(0, 194705, '\P{^Numeric_Type=_ NUMERIC}', "");
Error('\p{Nt= _Nu:=}');
Error('\P{Nt= _Nu:=}');
Expect(1, 194704, '\p{Nt=nu}', "");
Expect(0, 194704, '\p{^Nt=nu}', "");
Expect(0, 194704, '\P{Nt=nu}', "");
Expect(1, 194704, '\P{^Nt=nu}', "");
Expect(0, 194705, '\p{Nt=nu}', "");
Expect(1, 194705, '\p{^Nt=nu}', "");
Expect(1, 194705, '\P{Nt=nu}', "");
Expect(0, 194705, '\P{^Nt=nu}', "");
Expect(1, 194704, '\p{Nt= -Nu}', "");
Expect(0, 194704, '\p{^Nt= -Nu}', "");
Expect(0, 194704, '\P{Nt= -Nu}', "");
Expect(1, 194704, '\P{^Nt= -Nu}', "");
Expect(0, 194705, '\p{Nt= -Nu}', "");
Expect(1, 194705, '\p{^Nt= -Nu}', "");
Expect(1, 194705, '\P{Nt= -Nu}', "");
Expect(0, 194705, '\P{^Nt= -Nu}', "");
Error('\p{Is_Numeric_Type=-_numeric:=}');
Error('\P{Is_Numeric_Type=-_numeric:=}');
Expect(1, 194704, '\p{Is_Numeric_Type=numeric}', "");
Expect(0, 194704, '\p{^Is_Numeric_Type=numeric}', "");
Expect(0, 194704, '\P{Is_Numeric_Type=numeric}', "");
Expect(1, 194704, '\P{^Is_Numeric_Type=numeric}', "");
Expect(0, 194705, '\p{Is_Numeric_Type=numeric}', "");
Expect(1, 194705, '\p{^Is_Numeric_Type=numeric}', "");
Expect(1, 194705, '\P{Is_Numeric_Type=numeric}', "");
Expect(0, 194705, '\P{^Is_Numeric_Type=numeric}', "");
Expect(1, 194704, '\p{Is_Numeric_Type:   -	NUMERIC}', "");
Expect(0, 194704, '\p{^Is_Numeric_Type:   -	NUMERIC}', "");
Expect(0, 194704, '\P{Is_Numeric_Type:   -	NUMERIC}', "");
Expect(1, 194704, '\P{^Is_Numeric_Type:   -	NUMERIC}', "");
Expect(0, 194705, '\p{Is_Numeric_Type:   -	NUMERIC}', "");
Expect(1, 194705, '\p{^Is_Numeric_Type:   -	NUMERIC}', "");
Expect(1, 194705, '\P{Is_Numeric_Type:   -	NUMERIC}', "");
Expect(0, 194705, '\P{^Is_Numeric_Type:   -	NUMERIC}', "");
Error('\p{Is_Nt=/a/_ Nu}');
Error('\P{Is_Nt=/a/_ Nu}');
Expect(1, 194704, '\p{Is_Nt:   nu}', "");
Expect(0, 194704, '\p{^Is_Nt:   nu}', "");
Expect(0, 194704, '\P{Is_Nt:   nu}', "");
Expect(1, 194704, '\P{^Is_Nt:   nu}', "");
Expect(0, 194705, '\p{Is_Nt:   nu}', "");
Expect(1, 194705, '\p{^Is_Nt:   nu}', "");
Expect(1, 194705, '\P{Is_Nt:   nu}', "");
Expect(0, 194705, '\P{^Is_Nt:   nu}', "");
Expect(1, 194704, '\p{Is_Nt=- Nu}', "");
Expect(0, 194704, '\p{^Is_Nt=- Nu}', "");
Expect(0, 194704, '\P{Is_Nt=- Nu}', "");
Expect(1, 194704, '\P{^Is_Nt=- Nu}', "");
Expect(0, 194705, '\p{Is_Nt=- Nu}', "");
Expect(1, 194705, '\p{^Is_Nt=- Nu}', "");
Expect(1, 194705, '\P{Is_Nt=- Nu}', "");
Expect(0, 194705, '\P{^Is_Nt=- Nu}', "");
Error('\p{Numeric_Value=	/a/-000000001/002}');
Error('\P{Numeric_Value=	/a/-000000001/002}');
Expect(1, 3891, '\p{Numeric_Value=-00000001/000000002}', "");
Expect(0, 3891, '\p{^Numeric_Value=-00000001/000000002}', "");
Expect(0, 3891, '\P{Numeric_Value=-00000001/000000002}', "");
Expect(1, 3891, '\P{^Numeric_Value=-00000001/000000002}', "");
Expect(0, 3892, '\p{Numeric_Value=-00000001/000000002}', "");
Expect(1, 3892, '\p{^Numeric_Value=-00000001/000000002}', "");
Expect(1, 3892, '\P{Numeric_Value=-00000001/000000002}', "");
Expect(0, 3892, '\P{^Numeric_Value=-00000001/000000002}', "");
Expect(1, 3891, '\p{Numeric_Value=-0.500}', "");
Expect(0, 3891, '\p{^Numeric_Value=-0.500}', "");
Expect(0, 3891, '\P{Numeric_Value=-0.500}', "");
Expect(1, 3891, '\P{^Numeric_Value=-0.500}', "");
Expect(0, 3892, '\p{Numeric_Value=-0.500}', "");
Expect(1, 3892, '\p{^Numeric_Value=-0.500}', "");
Expect(1, 3892, '\P{Numeric_Value=-0.500}', "");
Expect(0, 3892, '\P{^Numeric_Value=-0.500}', "");
Error('\p{Nv=:=-	-000000001/0000000002}');
Error('\P{Nv=:=-	-000000001/0000000002}');
Expect(1, 3891, '\p{Nv=-00000001/02}', "");
Expect(0, 3891, '\p{^Nv=-00000001/02}', "");
Expect(0, 3891, '\P{Nv=-00000001/02}', "");
Expect(1, 3891, '\P{^Nv=-00000001/02}', "");
Expect(0, 3892, '\p{Nv=-00000001/02}', "");
Expect(1, 3892, '\p{^Nv=-00000001/02}', "");
Expect(1, 3892, '\P{Nv=-00000001/02}', "");
Expect(0, 3892, '\P{^Nv=-00000001/02}', "");
Expect(1, 3891, '\p{Nv:   -0.500}', "");
Expect(0, 3891, '\p{^Nv:   -0.500}', "");
Expect(0, 3891, '\P{Nv:   -0.500}', "");
Expect(1, 3891, '\P{^Nv:   -0.500}', "");
Expect(0, 3892, '\p{Nv:   -0.500}', "");
Expect(1, 3892, '\p{^Nv:   -0.500}', "");
Expect(1, 3892, '\P{Nv:   -0.500}', "");
Expect(0, 3892, '\P{^Nv:   -0.500}', "");
Error('\p{Is_Numeric_Value=_ -00000001/00002/a/}');
Error('\P{Is_Numeric_Value=_ -00000001/00002/a/}');
Expect(1, 3891, '\p{Is_Numeric_Value=-00001/00002}', "");
Expect(0, 3891, '\p{^Is_Numeric_Value=-00001/00002}', "");
Expect(0, 3891, '\P{Is_Numeric_Value=-00001/00002}', "");
Expect(1, 3891, '\P{^Is_Numeric_Value=-00001/00002}', "");
Expect(0, 3892, '\p{Is_Numeric_Value=-00001/00002}', "");
Expect(1, 3892, '\p{^Is_Numeric_Value=-00001/00002}', "");
Expect(1, 3892, '\P{Is_Numeric_Value=-00001/00002}', "");
Expect(0, 3892, '\P{^Is_Numeric_Value=-00001/00002}', "");
Expect(1, 3891, '\p{Is_Numeric_Value=-0.500}', "");
Expect(0, 3891, '\p{^Is_Numeric_Value=-0.500}', "");
Expect(0, 3891, '\P{Is_Numeric_Value=-0.500}', "");
Expect(1, 3891, '\P{^Is_Numeric_Value=-0.500}', "");
Expect(0, 3892, '\p{Is_Numeric_Value=-0.500}', "");
Expect(1, 3892, '\p{^Is_Numeric_Value=-0.500}', "");
Expect(1, 3892, '\P{Is_Numeric_Value=-0.500}', "");
Expect(0, 3892, '\P{^Is_Numeric_Value=-0.500}', "");
Error('\p{Is_Nv=:=-	-00001/000000002}');
Error('\P{Is_Nv=:=-	-00001/000000002}');
Expect(1, 3891, '\p{Is_Nv=-00000001/000000002}', "");
Expect(0, 3891, '\p{^Is_Nv=-00000001/000000002}', "");
Expect(0, 3891, '\P{Is_Nv=-00000001/000000002}', "");
Expect(1, 3891, '\P{^Is_Nv=-00000001/000000002}', "");
Expect(0, 3892, '\p{Is_Nv=-00000001/000000002}', "");
Expect(1, 3892, '\p{^Is_Nv=-00000001/000000002}', "");
Expect(1, 3892, '\P{Is_Nv=-00000001/000000002}', "");
Expect(0, 3892, '\P{^Is_Nv=-00000001/000000002}', "");
Expect(1, 3891, '\p{Is_Nv: -0.500}', "");
Expect(0, 3891, '\p{^Is_Nv: -0.500}', "");
Expect(0, 3891, '\P{Is_Nv: -0.500}', "");
Expect(1, 3891, '\P{^Is_Nv: -0.500}', "");
Expect(0, 3892, '\p{Is_Nv: -0.500}', "");
Expect(1, 3892, '\p{^Is_Nv: -0.500}', "");
Expect(1, 3892, '\P{Is_Nv: -0.500}', "");
Expect(0, 3892, '\P{^Is_Nv: -0.500}', "");
Error('\p{Numeric_Value=/a/--+0_0}');
Error('\P{Numeric_Value=/a/--+0_0}');
Expect(1, 127233, '\p{Numeric_Value=00000}', "");
Expect(0, 127233, '\p{^Numeric_Value=00000}', "");
Expect(0, 127233, '\P{Numeric_Value=00000}', "");
Expect(1, 127233, '\P{^Numeric_Value=00000}', "");
Expect(0, 127234, '\p{Numeric_Value=00000}', "");
Expect(1, 127234, '\p{^Numeric_Value=00000}', "");
Expect(1, 127234, '\P{Numeric_Value=00000}', "");
Expect(0, 127234, '\P{^Numeric_Value=00000}', "");
Error('\p{Nv= /a/00}');
Error('\P{Nv= /a/00}');
Expect(1, 127233, '\p{Nv=+00000_0}', "");
Expect(0, 127233, '\p{^Nv=+00000_0}', "");
Expect(0, 127233, '\P{Nv=+00000_0}', "");
Expect(1, 127233, '\P{^Nv=+00000_0}', "");
Expect(0, 127234, '\p{Nv=+00000_0}', "");
Expect(1, 127234, '\p{^Nv=+00000_0}', "");
Expect(1, 127234, '\P{Nv=+00000_0}', "");
Expect(0, 127234, '\P{^Nv=+00000_0}', "");
Error('\p{Is_Numeric_Value=	:=00_00_00_0}');
Error('\P{Is_Numeric_Value=	:=00_00_00_0}');
Expect(1, 127233, '\p{Is_Numeric_Value=0000000}', "");
Expect(0, 127233, '\p{^Is_Numeric_Value=0000000}', "");
Expect(0, 127233, '\P{Is_Numeric_Value=0000000}', "");
Expect(1, 127233, '\P{^Is_Numeric_Value=0000000}', "");
Expect(0, 127234, '\p{Is_Numeric_Value=0000000}', "");
Expect(1, 127234, '\p{^Is_Numeric_Value=0000000}', "");
Expect(1, 127234, '\P{Is_Numeric_Value=0000000}', "");
Expect(0, 127234, '\P{^Is_Numeric_Value=0000000}', "");
Error('\p{Is_Nv:   	/a/0000000000}');
Error('\P{Is_Nv:   	/a/0000000000}');
Expect(1, 127233, '\p{Is_Nv=000_0}', "");
Expect(0, 127233, '\p{^Is_Nv=000_0}', "");
Expect(0, 127233, '\P{Is_Nv=000_0}', "");
Expect(1, 127233, '\P{^Is_Nv=000_0}', "");
Expect(0, 127234, '\p{Is_Nv=000_0}', "");
Expect(1, 127234, '\p{^Is_Nv=000_0}', "");
Expect(1, 127234, '\P{Is_Nv=000_0}', "");
Expect(0, 127234, '\P{^Is_Nv=000_0}', "");
Error('\p{Numeric_Value= +0_0_0_0_0_0_001/a/}');
Error('\P{Numeric_Value= +0_0_0_0_0_0_001/a/}');
Expect(1, 133418, '\p{Numeric_Value=+001}', "");
Expect(0, 133418, '\p{^Numeric_Value=+001}', "");
Expect(0, 133418, '\P{Numeric_Value=+001}', "");
Expect(1, 133418, '\P{^Numeric_Value=+001}', "");
Expect(0, 133419, '\p{Numeric_Value=+001}', "");
Expect(1, 133419, '\p{^Numeric_Value=+001}', "");
Expect(1, 133419, '\P{Numeric_Value=+001}', "");
Expect(0, 133419, '\P{^Numeric_Value=+001}', "");
Error('\p{Nv= _000000001/a/}');
Error('\P{Nv= _000000001/a/}');
Expect(1, 133418, '\p{Nv=1}', "");
Expect(0, 133418, '\p{^Nv=1}', "");
Expect(0, 133418, '\P{Nv=1}', "");
Expect(1, 133418, '\P{^Nv=1}', "");
Expect(0, 133419, '\p{Nv=1}', "");
Expect(1, 133419, '\p{^Nv=1}', "");
Expect(1, 133419, '\P{Nv=1}', "");
Expect(0, 133419, '\P{^Nv=1}', "");
Error('\p{Is_Numeric_Value=_/a/+001}');
Error('\P{Is_Numeric_Value=_/a/+001}');
Expect(1, 133418, '\p{Is_Numeric_Value: 00001}', "");
Expect(0, 133418, '\p{^Is_Numeric_Value: 00001}', "");
Expect(0, 133418, '\P{Is_Numeric_Value: 00001}', "");
Expect(1, 133418, '\P{^Is_Numeric_Value: 00001}', "");
Expect(0, 133419, '\p{Is_Numeric_Value: 00001}', "");
Expect(1, 133419, '\p{^Is_Numeric_Value: 00001}', "");
Expect(1, 133419, '\P{Is_Numeric_Value: 00001}', "");
Expect(0, 133419, '\P{^Is_Numeric_Value: 00001}', "");
Error('\p{Is_Nv= +000_000_1:=}');
Error('\P{Is_Nv= +000_000_1:=}');
Expect(1, 133418, '\p{Is_Nv=0_1}', "");
Expect(0, 133418, '\p{^Is_Nv=0_1}', "");
Expect(0, 133418, '\P{Is_Nv=0_1}', "");
Expect(1, 133418, '\P{^Is_Nv=0_1}', "");
Expect(0, 133419, '\p{Is_Nv=0_1}', "");
Expect(1, 133419, '\p{^Is_Nv=0_1}', "");
Expect(1, 133419, '\P{Is_Nv=0_1}', "");
Expect(0, 133419, '\P{^Is_Nv=0_1}', "");
Error('\p{Numeric_Value=/a/-_00000001/0000010}');
Error('\P{Numeric_Value=/a/-_00000001/0000010}');
Expect(1, 8530, '\p{Numeric_Value=+1/00000000010}', "");
Expect(0, 8530, '\p{^Numeric_Value=+1/00000000010}', "");
Expect(0, 8530, '\P{Numeric_Value=+1/00000000010}', "");
Expect(1, 8530, '\P{^Numeric_Value=+1/00000000010}', "");
Expect(0, 8531, '\p{Numeric_Value=+1/00000000010}', "");
Expect(1, 8531, '\p{^Numeric_Value=+1/00000000010}', "");
Expect(1, 8531, '\P{Numeric_Value=+1/00000000010}', "");
Expect(0, 8531, '\P{^Numeric_Value=+1/00000000010}', "");
Expect(1, 8530, '\p{Numeric_Value=0.100}', "");
Expect(0, 8530, '\p{^Numeric_Value=0.100}', "");
Expect(0, 8530, '\P{Numeric_Value=0.100}', "");
Expect(1, 8530, '\P{^Numeric_Value=0.100}', "");
Expect(0, 8531, '\p{Numeric_Value=0.100}', "");
Expect(1, 8531, '\p{^Numeric_Value=0.100}', "");
Expect(1, 8531, '\P{Numeric_Value=0.100}', "");
Expect(0, 8531, '\P{^Numeric_Value=0.100}', "");
Error('\p{Nv=_	0001/00010/a/}');
Error('\P{Nv=_	0001/00010/a/}');
Expect(1, 8530, '\p{Nv=01/010}', "");
Expect(0, 8530, '\p{^Nv=01/010}', "");
Expect(0, 8530, '\P{Nv=01/010}', "");
Expect(1, 8530, '\P{^Nv=01/010}', "");
Expect(0, 8531, '\p{Nv=01/010}', "");
Expect(1, 8531, '\p{^Nv=01/010}', "");
Expect(1, 8531, '\P{Nv=01/010}', "");
Expect(0, 8531, '\P{^Nv=01/010}', "");
Expect(1, 8530, '\p{Nv=0.100}', "");
Expect(0, 8530, '\p{^Nv=0.100}', "");
Expect(0, 8530, '\P{Nv=0.100}', "");
Expect(1, 8530, '\P{^Nv=0.100}', "");
Expect(0, 8531, '\p{Nv=0.100}', "");
Expect(1, 8531, '\p{^Nv=0.100}', "");
Expect(1, 8531, '\P{Nv=0.100}', "");
Expect(0, 8531, '\P{^Nv=0.100}', "");
Error('\p{Is_Numeric_Value=  00000001/000010/a/}');
Error('\P{Is_Numeric_Value=  00000001/000010/a/}');
Expect(1, 8530, '\p{Is_Numeric_Value=0000000001/0010}', "");
Expect(0, 8530, '\p{^Is_Numeric_Value=0000000001/0010}', "");
Expect(0, 8530, '\P{Is_Numeric_Value=0000000001/0010}', "");
Expect(1, 8530, '\P{^Is_Numeric_Value=0000000001/0010}', "");
Expect(0, 8531, '\p{Is_Numeric_Value=0000000001/0010}', "");
Expect(1, 8531, '\p{^Is_Numeric_Value=0000000001/0010}', "");
Expect(1, 8531, '\P{Is_Numeric_Value=0000000001/0010}', "");
Expect(0, 8531, '\P{^Is_Numeric_Value=0000000001/0010}', "");
Expect(1, 8530, '\p{Is_Numeric_Value=0.100}', "");
Expect(0, 8530, '\p{^Is_Numeric_Value=0.100}', "");
Expect(0, 8530, '\P{Is_Numeric_Value=0.100}', "");
Expect(1, 8530, '\P{^Is_Numeric_Value=0.100}', "");
Expect(0, 8531, '\p{Is_Numeric_Value=0.100}', "");
Expect(1, 8531, '\p{^Is_Numeric_Value=0.100}', "");
Expect(1, 8531, '\P{Is_Numeric_Value=0.100}', "");
Expect(0, 8531, '\P{^Is_Numeric_Value=0.100}', "");
Error('\p{Is_Nv= 	01/00000000010:=}');
Error('\P{Is_Nv= 	01/00000000010:=}');
Expect(1, 8530, '\p{Is_Nv=+001/0000000010}', "");
Expect(0, 8530, '\p{^Is_Nv=+001/0000000010}', "");
Expect(0, 8530, '\P{Is_Nv=+001/0000000010}', "");
Expect(1, 8530, '\P{^Is_Nv=+001/0000000010}', "");
Expect(0, 8531, '\p{Is_Nv=+001/0000000010}', "");
Expect(1, 8531, '\p{^Is_Nv=+001/0000000010}', "");
Expect(1, 8531, '\P{Is_Nv=+001/0000000010}', "");
Expect(0, 8531, '\P{^Is_Nv=+001/0000000010}', "");
Expect(1, 8530, '\p{Is_Nv: 0.100}', "");
Expect(0, 8530, '\p{^Is_Nv: 0.100}', "");
Expect(0, 8530, '\P{Is_Nv: 0.100}', "");
Expect(1, 8530, '\P{^Is_Nv: 0.100}', "");
Expect(0, 8531, '\p{Is_Nv: 0.100}', "");
Expect(1, 8531, '\p{^Is_Nv: 0.100}', "");
Expect(1, 8531, '\P{Is_Nv: 0.100}', "");
Expect(0, 8531, '\P{^Is_Nv: 0.100}', "");
Error('\p{Numeric_Value=:=- 0000001/00000016}');
Error('\P{Numeric_Value=:=- 0000001/00000016}');
Expect(1, 43059, '\p{Numeric_Value=+0001/00000016}', "");
Expect(0, 43059, '\p{^Numeric_Value=+0001/00000016}', "");
Expect(0, 43059, '\P{Numeric_Value=+0001/00000016}', "");
Expect(1, 43059, '\P{^Numeric_Value=+0001/00000016}', "");
Expect(0, 43060, '\p{Numeric_Value=+0001/00000016}', "");
Expect(1, 43060, '\p{^Numeric_Value=+0001/00000016}', "");
Expect(1, 43060, '\P{Numeric_Value=+0001/00000016}', "");
Expect(0, 43060, '\P{^Numeric_Value=+0001/00000016}', "");
Error('\p{Numeric_Value=0.06}');
Error('\P{Numeric_Value=0.06}');
Expect(1, 43059, '\p{Numeric_Value=0.062}', "");
Expect(0, 43059, '\p{^Numeric_Value=0.062}', "");
Expect(0, 43059, '\P{Numeric_Value=0.062}', "");
Expect(1, 43059, '\P{^Numeric_Value=0.062}', "");
Expect(0, 43060, '\p{Numeric_Value=0.062}', "");
Expect(1, 43060, '\p{^Numeric_Value=0.062}', "");
Expect(1, 43060, '\P{Numeric_Value=0.062}', "");
Expect(0, 43060, '\P{^Numeric_Value=0.062}', "");
Error('\p{Nv=-0001/00000000016/a/}');
Error('\P{Nv=-0001/00000000016/a/}');
Expect(1, 43059, '\p{Nv=0000000001/000000016}', "");
Expect(0, 43059, '\p{^Nv=0000000001/000000016}', "");
Expect(0, 43059, '\P{Nv=0000000001/000000016}', "");
Expect(1, 43059, '\P{^Nv=0000000001/000000016}', "");
Expect(0, 43060, '\p{Nv=0000000001/000000016}', "");
Expect(1, 43060, '\p{^Nv=0000000001/000000016}', "");
Expect(1, 43060, '\P{Nv=0000000001/000000016}', "");
Expect(0, 43060, '\P{^Nv=0000000001/000000016}', "");
Error('\p{Nv=0.06}');
Error('\P{Nv=0.06}');
Expect(1, 43059, '\p{Nv=0.062}', "");
Expect(0, 43059, '\p{^Nv=0.062}', "");
Expect(0, 43059, '\P{Nv=0.062}', "");
Expect(1, 43059, '\P{^Nv=0.062}', "");
Expect(0, 43060, '\p{Nv=0.062}', "");
Expect(1, 43060, '\p{^Nv=0.062}', "");
Expect(1, 43060, '\P{Nv=0.062}', "");
Expect(0, 43060, '\P{^Nv=0.062}', "");
Error('\p{Is_Numeric_Value=_/a/+00001/000016}');
Error('\P{Is_Numeric_Value=_/a/+00001/000016}');
Expect(1, 43059, '\p{Is_Numeric_Value=01/16}', "");
Expect(0, 43059, '\p{^Is_Numeric_Value=01/16}', "");
Expect(0, 43059, '\P{Is_Numeric_Value=01/16}', "");
Expect(1, 43059, '\P{^Is_Numeric_Value=01/16}', "");
Expect(0, 43060, '\p{Is_Numeric_Value=01/16}', "");
Expect(1, 43060, '\p{^Is_Numeric_Value=01/16}', "");
Expect(1, 43060, '\P{Is_Numeric_Value=01/16}', "");
Expect(0, 43060, '\P{^Is_Numeric_Value=01/16}', "");
Error('\p{Is_Numeric_Value=0.06}');
Error('\P{Is_Numeric_Value=0.06}');
Expect(1, 43059, '\p{Is_Numeric_Value=0.062}', "");
Expect(0, 43059, '\p{^Is_Numeric_Value=0.062}', "");
Expect(0, 43059, '\P{Is_Numeric_Value=0.062}', "");
Expect(1, 43059, '\P{^Is_Numeric_Value=0.062}', "");
Expect(0, 43060, '\p{Is_Numeric_Value=0.062}', "");
Expect(1, 43060, '\p{^Is_Numeric_Value=0.062}', "");
Expect(1, 43060, '\P{Is_Numeric_Value=0.062}', "");
Expect(0, 43060, '\P{^Is_Numeric_Value=0.062}', "");
Error('\p{Is_Nv:	/a/+1/16}');
Error('\P{Is_Nv:	/a/+1/16}');
Expect(1, 43059, '\p{Is_Nv=000000001/0000016}', "");
Expect(0, 43059, '\p{^Is_Nv=000000001/0000016}', "");
Expect(0, 43059, '\P{Is_Nv=000000001/0000016}', "");
Expect(1, 43059, '\P{^Is_Nv=000000001/0000016}', "");
Expect(0, 43060, '\p{Is_Nv=000000001/0000016}', "");
Expect(1, 43060, '\p{^Is_Nv=000000001/0000016}', "");
Expect(1, 43060, '\P{Is_Nv=000000001/0000016}', "");
Expect(0, 43060, '\P{^Is_Nv=000000001/0000016}', "");
Error('\p{Is_Nv=0.06}');
Error('\P{Is_Nv=0.06}');
Expect(1, 43059, '\p{Is_Nv=0.062}', "");
Expect(0, 43059, '\p{^Is_Nv=0.062}', "");
Expect(0, 43059, '\P{Is_Nv=0.062}', "");
Expect(1, 43059, '\P{^Is_Nv=0.062}', "");
Expect(0, 43060, '\p{Is_Nv=0.062}', "");
Expect(1, 43060, '\p{^Is_Nv=0.062}', "");
Expect(1, 43060, '\P{Is_Nv=0.062}', "");
Expect(0, 43060, '\P{^Is_Nv=0.062}', "");
Error('\p{Numeric_Value=/a/_-0001/000000002}');
Error('\P{Numeric_Value=/a/_-0001/000000002}');
Expect(1, 69243, '\p{Numeric_Value: 00000001/000000002}', "");
Expect(0, 69243, '\p{^Numeric_Value: 00000001/000000002}', "");
Expect(0, 69243, '\P{Numeric_Value: 00000001/000000002}', "");
Expect(1, 69243, '\P{^Numeric_Value: 00000001/000000002}', "");
Expect(0, 69244, '\p{Numeric_Value: 00000001/000000002}', "");
Expect(1, 69244, '\p{^Numeric_Value: 00000001/000000002}', "");
Expect(1, 69244, '\P{Numeric_Value: 00000001/000000002}', "");
Expect(0, 69244, '\P{^Numeric_Value: 00000001/000000002}', "");
Expect(1, 69243, '\p{Numeric_Value=0.500}', "");
Expect(0, 69243, '\p{^Numeric_Value=0.500}', "");
Expect(0, 69243, '\P{Numeric_Value=0.500}', "");
Expect(1, 69243, '\P{^Numeric_Value=0.500}', "");
Expect(0, 69244, '\p{Numeric_Value=0.500}', "");
Expect(1, 69244, '\p{^Numeric_Value=0.500}', "");
Expect(1, 69244, '\P{Numeric_Value=0.500}', "");
Expect(0, 69244, '\P{^Numeric_Value=0.500}', "");
Error('\p{Nv:	0000000001/002/a/}');
Error('\P{Nv:	0000000001/002/a/}');
Expect(1, 69243, '\p{Nv=01/000002}', "");
Expect(0, 69243, '\p{^Nv=01/000002}', "");
Expect(0, 69243, '\P{Nv=01/000002}', "");
Expect(1, 69243, '\P{^Nv=01/000002}', "");
Expect(0, 69244, '\p{Nv=01/000002}', "");
Expect(1, 69244, '\p{^Nv=01/000002}', "");
Expect(1, 69244, '\P{Nv=01/000002}', "");
Expect(0, 69244, '\P{^Nv=01/000002}', "");
Expect(1, 69243, '\p{Nv:   0.500}', "");
Expect(0, 69243, '\p{^Nv:   0.500}', "");
Expect(0, 69243, '\P{Nv:   0.500}', "");
Expect(1, 69243, '\P{^Nv:   0.500}', "");
Expect(0, 69244, '\p{Nv:   0.500}', "");
Expect(1, 69244, '\p{^Nv:   0.500}', "");
Expect(1, 69244, '\P{Nv:   0.500}', "");
Expect(0, 69244, '\P{^Nv:   0.500}', "");
Error('\p{Is_Numeric_Value=		+00001/0002:=}');
Error('\P{Is_Numeric_Value=		+00001/0002:=}');
Expect(1, 69243, '\p{Is_Numeric_Value=+001/002}', "");
Expect(0, 69243, '\p{^Is_Numeric_Value=+001/002}', "");
Expect(0, 69243, '\P{Is_Numeric_Value=+001/002}', "");
Expect(1, 69243, '\P{^Is_Numeric_Value=+001/002}', "");
Expect(0, 69244, '\p{Is_Numeric_Value=+001/002}', "");
Expect(1, 69244, '\p{^Is_Numeric_Value=+001/002}', "");
Expect(1, 69244, '\P{Is_Numeric_Value=+001/002}', "");
Expect(0, 69244, '\P{^Is_Numeric_Value=+001/002}', "");
Expect(1, 69243, '\p{Is_Numeric_Value=0.500}', "");
Expect(0, 69243, '\p{^Is_Numeric_Value=0.500}', "");
Expect(0, 69243, '\P{Is_Numeric_Value=0.500}', "");
Expect(1, 69243, '\P{^Is_Numeric_Value=0.500}', "");
Expect(0, 69244, '\p{Is_Numeric_Value=0.500}', "");
Expect(1, 69244, '\p{^Is_Numeric_Value=0.500}', "");
Expect(1, 69244, '\P{Is_Numeric_Value=0.500}', "");
Expect(0, 69244, '\P{^Is_Numeric_Value=0.500}', "");
Error('\p{Is_Nv:  :=01/0000000002}');
Error('\P{Is_Nv:  :=01/0000000002}');
Expect(1, 69243, '\p{Is_Nv=00001/000000002}', "");
Expect(0, 69243, '\p{^Is_Nv=00001/000000002}', "");
Expect(0, 69243, '\P{Is_Nv=00001/000000002}', "");
Expect(1, 69243, '\P{^Is_Nv=00001/000000002}', "");
Expect(0, 69244, '\p{Is_Nv=00001/000000002}', "");
Expect(1, 69244, '\p{^Is_Nv=00001/000000002}', "");
Expect(1, 69244, '\P{Is_Nv=00001/000000002}', "");
Expect(0, 69244, '\P{^Is_Nv=00001/000000002}', "");
Expect(1, 69243, '\p{Is_Nv=0.500}', "");
Expect(0, 69243, '\p{^Is_Nv=0.500}', "");
Expect(0, 69243, '\P{Is_Nv=0.500}', "");
Expect(1, 69243, '\P{^Is_Nv=0.500}', "");
Expect(0, 69244, '\p{Is_Nv=0.500}', "");
Expect(1, 69244, '\p{^Is_Nv=0.500}', "");
Expect(1, 69244, '\P{Is_Nv=0.500}', "");
Expect(0, 69244, '\P{^Is_Nv=0.500}', "");
Error('\p{Numeric_Value=:=--0000000001/000003}');
Error('\P{Numeric_Value=:=--0000000001/000003}');
Expect(1, 74845, '\p{Numeric_Value=000000001/000000003}', "");
Expect(0, 74845, '\p{^Numeric_Value=000000001/000000003}', "");
Expect(0, 74845, '\P{Numeric_Value=000000001/000000003}', "");
Expect(1, 74845, '\P{^Numeric_Value=000000001/000000003}', "");
Expect(0, 74846, '\p{Numeric_Value=000000001/000000003}', "");
Expect(1, 74846, '\p{^Numeric_Value=000000001/000000003}', "");
Expect(1, 74846, '\P{Numeric_Value=000000001/000000003}', "");
Expect(0, 74846, '\P{^Numeric_Value=000000001/000000003}', "");
Error('\p{Numeric_Value=0.3}');
Error('\P{Numeric_Value=0.3}');
Error('\p{Numeric_Value=0.33}');
Error('\P{Numeric_Value=0.33}');
Expect(1, 74845, '\p{Numeric_Value=0.333}', "");
Expect(0, 74845, '\p{^Numeric_Value=0.333}', "");
Expect(0, 74845, '\P{Numeric_Value=0.333}', "");
Expect(1, 74845, '\P{^Numeric_Value=0.333}', "");
Expect(0, 74846, '\p{Numeric_Value=0.333}', "");
Expect(1, 74846, '\p{^Numeric_Value=0.333}', "");
Expect(1, 74846, '\P{Numeric_Value=0.333}', "");
Expect(0, 74846, '\P{^Numeric_Value=0.333}', "");
Error('\p{Nv=-/a/00001/00000003}');
Error('\P{Nv=-/a/00001/00000003}');
Expect(1, 74845, '\p{Nv=00001/3}', "");
Expect(0, 74845, '\p{^Nv=00001/3}', "");
Expect(0, 74845, '\P{Nv=00001/3}', "");
Expect(1, 74845, '\P{^Nv=00001/3}', "");
Expect(0, 74846, '\p{Nv=00001/3}', "");
Expect(1, 74846, '\p{^Nv=00001/3}', "");
Expect(1, 74846, '\P{Nv=00001/3}', "");
Expect(0, 74846, '\P{^Nv=00001/3}', "");
Error('\p{Nv=0.3}');
Error('\P{Nv=0.3}');
Error('\p{Nv: 0.33}');
Error('\P{Nv: 0.33}');
Expect(1, 74845, '\p{Nv=0.333}', "");
Expect(0, 74845, '\p{^Nv=0.333}', "");
Expect(0, 74845, '\P{Nv=0.333}', "");
Expect(1, 74845, '\P{^Nv=0.333}', "");
Expect(0, 74846, '\p{Nv=0.333}', "");
Expect(1, 74846, '\p{^Nv=0.333}', "");
Expect(1, 74846, '\P{Nv=0.333}', "");
Expect(0, 74846, '\P{^Nv=0.333}', "");
Error('\p{Is_Numeric_Value= :=01/003}');
Error('\P{Is_Numeric_Value= :=01/003}');
Expect(1, 74845, '\p{Is_Numeric_Value=01/3}', "");
Expect(0, 74845, '\p{^Is_Numeric_Value=01/3}', "");
Expect(0, 74845, '\P{Is_Numeric_Value=01/3}', "");
Expect(1, 74845, '\P{^Is_Numeric_Value=01/3}', "");
Expect(0, 74846, '\p{Is_Numeric_Value=01/3}', "");
Expect(1, 74846, '\p{^Is_Numeric_Value=01/3}', "");
Expect(1, 74846, '\P{Is_Numeric_Value=01/3}', "");
Expect(0, 74846, '\P{^Is_Numeric_Value=01/3}', "");
Error('\p{Is_Numeric_Value=0.3}');
Error('\P{Is_Numeric_Value=0.3}');
Error('\p{Is_Numeric_Value=0.33}');
Error('\P{Is_Numeric_Value=0.33}');
Expect(1, 74845, '\p{Is_Numeric_Value=0.333}', "");
Expect(0, 74845, '\p{^Is_Numeric_Value=0.333}', "");
Expect(0, 74845, '\P{Is_Numeric_Value=0.333}', "");
Expect(1, 74845, '\P{^Is_Numeric_Value=0.333}', "");
Expect(0, 74846, '\p{Is_Numeric_Value=0.333}', "");
Expect(1, 74846, '\p{^Is_Numeric_Value=0.333}', "");
Expect(1, 74846, '\P{Is_Numeric_Value=0.333}', "");
Expect(0, 74846, '\P{^Is_Numeric_Value=0.333}', "");
Error('\p{Is_Nv=/a/-_001/000000003}');
Error('\P{Is_Nv=/a/-_001/000000003}');
Expect(1, 74845, '\p{Is_Nv=000001/000003}', "");
Expect(0, 74845, '\p{^Is_Nv=000001/000003}', "");
Expect(0, 74845, '\P{Is_Nv=000001/000003}', "");
Expect(1, 74845, '\P{^Is_Nv=000001/000003}', "");
Expect(0, 74846, '\p{Is_Nv=000001/000003}', "");
Expect(1, 74846, '\p{^Is_Nv=000001/000003}', "");
Expect(1, 74846, '\P{Is_Nv=000001/000003}', "");
Expect(0, 74846, '\P{^Is_Nv=000001/000003}', "");
Error('\p{Is_Nv=0.3}');
Error('\P{Is_Nv=0.3}');
Error('\p{Is_Nv=0.33}');
Error('\P{Is_Nv=0.33}');
Expect(1, 74845, '\p{Is_Nv=0.333}', "");
Expect(0, 74845, '\p{^Is_Nv=0.333}', "");
Expect(0, 74845, '\P{Is_Nv=0.333}', "");
Expect(1, 74845, '\P{^Is_Nv=0.333}', "");
Expect(0, 74846, '\p{Is_Nv=0.333}', "");
Expect(1, 74846, '\p{^Is_Nv=0.333}', "");
Expect(1, 74846, '\P{Is_Nv=0.333}', "");
Expect(0, 74846, '\P{^Is_Nv=0.333}', "");
Error('\p{Numeric_Value=:=	000000001/000000004}');
Error('\P{Numeric_Value=:=	000000001/000000004}');
Expect(1, 74850, '\p{Numeric_Value=01/0000000004}', "");
Expect(0, 74850, '\p{^Numeric_Value=01/0000000004}', "");
Expect(0, 74850, '\P{Numeric_Value=01/0000000004}', "");
Expect(1, 74850, '\P{^Numeric_Value=01/0000000004}', "");
Expect(0, 74851, '\p{Numeric_Value=01/0000000004}', "");
Expect(1, 74851, '\p{^Numeric_Value=01/0000000004}', "");
Expect(1, 74851, '\P{Numeric_Value=01/0000000004}', "");
Expect(0, 74851, '\P{^Numeric_Value=01/0000000004}', "");
Expect(1, 74850, '\p{Numeric_Value:	0.250}', "");
Expect(0, 74850, '\p{^Numeric_Value:	0.250}', "");
Expect(0, 74850, '\P{Numeric_Value:	0.250}', "");
Expect(1, 74850, '\P{^Numeric_Value:	0.250}', "");
Expect(0, 74851, '\p{Numeric_Value:	0.250}', "");
Expect(1, 74851, '\p{^Numeric_Value:	0.250}', "");
Expect(1, 74851, '\P{Numeric_Value:	0.250}', "");
Expect(0, 74851, '\P{^Numeric_Value:	0.250}', "");
Error('\p{Nv: :=+1/04}');
Error('\P{Nv: :=+1/04}');
Expect(1, 74850, '\p{Nv=00001/000004}', "");
Expect(0, 74850, '\p{^Nv=00001/000004}', "");
Expect(0, 74850, '\P{Nv=00001/000004}', "");
Expect(1, 74850, '\P{^Nv=00001/000004}', "");
Expect(0, 74851, '\p{Nv=00001/000004}', "");
Expect(1, 74851, '\p{^Nv=00001/000004}', "");
Expect(1, 74851, '\P{Nv=00001/000004}', "");
Expect(0, 74851, '\P{^Nv=00001/000004}', "");
Expect(1, 74850, '\p{Nv=0.250}', "");
Expect(0, 74850, '\p{^Nv=0.250}', "");
Expect(0, 74850, '\P{Nv=0.250}', "");
Expect(1, 74850, '\P{^Nv=0.250}', "");
Expect(0, 74851, '\p{Nv=0.250}', "");
Expect(1, 74851, '\p{^Nv=0.250}', "");
Expect(1, 74851, '\P{Nv=0.250}', "");
Expect(0, 74851, '\P{^Nv=0.250}', "");
Error('\p{Is_Numeric_Value= -+000001/4:=}');
Error('\P{Is_Numeric_Value= -+000001/4:=}');
Expect(1, 74850, '\p{Is_Numeric_Value=+000001/000004}', "");
Expect(0, 74850, '\p{^Is_Numeric_Value=+000001/000004}', "");
Expect(0, 74850, '\P{Is_Numeric_Value=+000001/000004}', "");
Expect(1, 74850, '\P{^Is_Numeric_Value=+000001/000004}', "");
Expect(0, 74851, '\p{Is_Numeric_Value=+000001/000004}', "");
Expect(1, 74851, '\p{^Is_Numeric_Value=+000001/000004}', "");
Expect(1, 74851, '\P{Is_Numeric_Value=+000001/000004}', "");
Expect(0, 74851, '\P{^Is_Numeric_Value=+000001/000004}', "");
Expect(1, 74850, '\p{Is_Numeric_Value: 0.250}', "");
Expect(0, 74850, '\p{^Is_Numeric_Value: 0.250}', "");
Expect(0, 74850, '\P{Is_Numeric_Value: 0.250}', "");
Expect(1, 74850, '\P{^Is_Numeric_Value: 0.250}', "");
Expect(0, 74851, '\p{Is_Numeric_Value: 0.250}', "");
Expect(1, 74851, '\p{^Is_Numeric_Value: 0.250}', "");
Expect(1, 74851, '\P{Is_Numeric_Value: 0.250}', "");
Expect(0, 74851, '\P{^Is_Numeric_Value: 0.250}', "");
Error('\p{Is_Nv: /a/	 0001/04}');
Error('\P{Is_Nv: /a/	 0001/04}');
Expect(1, 74850, '\p{Is_Nv=1/000000004}', "");
Expect(0, 74850, '\p{^Is_Nv=1/000000004}', "");
Expect(0, 74850, '\P{Is_Nv=1/000000004}', "");
Expect(1, 74850, '\P{^Is_Nv=1/000000004}', "");
Expect(0, 74851, '\p{Is_Nv=1/000000004}', "");
Expect(1, 74851, '\p{^Is_Nv=1/000000004}', "");
Expect(1, 74851, '\P{Is_Nv=1/000000004}', "");
Expect(0, 74851, '\P{^Is_Nv=1/000000004}', "");
Expect(1, 74850, '\p{Is_Nv=0.250}', "");
Expect(0, 74850, '\p{^Is_Nv=0.250}', "");
Expect(0, 74850, '\P{Is_Nv=0.250}', "");
Expect(1, 74850, '\P{^Is_Nv=0.250}', "");
Expect(0, 74851, '\p{Is_Nv=0.250}', "");
Expect(1, 74851, '\p{^Is_Nv=0.250}', "");
Expect(1, 74851, '\P{Is_Nv=0.250}', "");
Expect(0, 74851, '\P{^Is_Nv=0.250}', "");
Error('\p{Numeric_Value=	_+0000001/0000000005:=}');
Error('\P{Numeric_Value=	_+0000001/0000000005:=}');
Expect(1, 8533, '\p{Numeric_Value=000001/00000005}', "");
Expect(0, 8533, '\p{^Numeric_Value=000001/00000005}', "");
Expect(0, 8533, '\P{Numeric_Value=000001/00000005}', "");
Expect(1, 8533, '\P{^Numeric_Value=000001/00000005}', "");
Expect(0, 8534, '\p{Numeric_Value=000001/00000005}', "");
Expect(1, 8534, '\p{^Numeric_Value=000001/00000005}', "");
Expect(1, 8534, '\P{Numeric_Value=000001/00000005}', "");
Expect(0, 8534, '\P{^Numeric_Value=000001/00000005}', "");
Expect(1, 8533, '\p{Numeric_Value=0.200}', "");
Expect(0, 8533, '\p{^Numeric_Value=0.200}', "");
Expect(0, 8533, '\P{Numeric_Value=0.200}', "");
Expect(1, 8533, '\P{^Numeric_Value=0.200}', "");
Expect(0, 8534, '\p{Numeric_Value=0.200}', "");
Expect(1, 8534, '\p{^Numeric_Value=0.200}', "");
Expect(1, 8534, '\P{Numeric_Value=0.200}', "");
Expect(0, 8534, '\P{^Numeric_Value=0.200}', "");
Error('\p{Nv:    :=000000001/000000005}');
Error('\P{Nv:    :=000000001/000000005}');
Expect(1, 8533, '\p{Nv=+00000001/0000005}', "");
Expect(0, 8533, '\p{^Nv=+00000001/0000005}', "");
Expect(0, 8533, '\P{Nv=+00000001/0000005}', "");
Expect(1, 8533, '\P{^Nv=+00000001/0000005}', "");
Expect(0, 8534, '\p{Nv=+00000001/0000005}', "");
Expect(1, 8534, '\p{^Nv=+00000001/0000005}', "");
Expect(1, 8534, '\P{Nv=+00000001/0000005}', "");
Expect(0, 8534, '\P{^Nv=+00000001/0000005}', "");
Expect(1, 8533, '\p{Nv=0.200}', "");
Expect(0, 8533, '\p{^Nv=0.200}', "");
Expect(0, 8533, '\P{Nv=0.200}', "");
Expect(1, 8533, '\P{^Nv=0.200}', "");
Expect(0, 8534, '\p{Nv=0.200}', "");
Expect(1, 8534, '\p{^Nv=0.200}', "");
Expect(1, 8534, '\P{Nv=0.200}', "");
Expect(0, 8534, '\P{^Nv=0.200}', "");
Error('\p{Is_Numeric_Value=-_000000001/0000005/a/}');
Error('\P{Is_Numeric_Value=-_000000001/0000005/a/}');
Expect(1, 8533, '\p{Is_Numeric_Value=00001/005}', "");
Expect(0, 8533, '\p{^Is_Numeric_Value=00001/005}', "");
Expect(0, 8533, '\P{Is_Numeric_Value=00001/005}', "");
Expect(1, 8533, '\P{^Is_Numeric_Value=00001/005}', "");
Expect(0, 8534, '\p{Is_Numeric_Value=00001/005}', "");
Expect(1, 8534, '\p{^Is_Numeric_Value=00001/005}', "");
Expect(1, 8534, '\P{Is_Numeric_Value=00001/005}', "");
Expect(0, 8534, '\P{^Is_Numeric_Value=00001/005}', "");
Expect(1, 8533, '\p{Is_Numeric_Value=0.200}', "");
Expect(0, 8533, '\p{^Is_Numeric_Value=0.200}', "");
Expect(0, 8533, '\P{Is_Numeric_Value=0.200}', "");
Expect(1, 8533, '\P{^Is_Numeric_Value=0.200}', "");
Expect(0, 8534, '\p{Is_Numeric_Value=0.200}', "");
Expect(1, 8534, '\p{^Is_Numeric_Value=0.200}', "");
Expect(1, 8534, '\P{Is_Numeric_Value=0.200}', "");
Expect(0, 8534, '\P{^Is_Numeric_Value=0.200}', "");
Error('\p{Is_Nv=:=_-000001/0005}');
Error('\P{Is_Nv=:=_-000001/0005}');
Expect(1, 8533, '\p{Is_Nv=0000000001/00005}', "");
Expect(0, 8533, '\p{^Is_Nv=0000000001/00005}', "");
Expect(0, 8533, '\P{Is_Nv=0000000001/00005}', "");
Expect(1, 8533, '\P{^Is_Nv=0000000001/00005}', "");
Expect(0, 8534, '\p{Is_Nv=0000000001/00005}', "");
Expect(1, 8534, '\p{^Is_Nv=0000000001/00005}', "");
Expect(1, 8534, '\P{Is_Nv=0000000001/00005}', "");
Expect(0, 8534, '\P{^Is_Nv=0000000001/00005}', "");
Expect(1, 8533, '\p{Is_Nv=0.200}', "");
Expect(0, 8533, '\p{^Is_Nv=0.200}', "");
Expect(0, 8533, '\P{Is_Nv=0.200}', "");
Expect(1, 8533, '\P{^Is_Nv=0.200}', "");
Expect(0, 8534, '\p{Is_Nv=0.200}', "");
Expect(1, 8534, '\p{^Is_Nv=0.200}', "");
Expect(1, 8534, '\P{Is_Nv=0.200}', "");
Expect(0, 8534, '\P{^Is_Nv=0.200}', "");
Error('\p{Numeric_Value=	-01/000000006/a/}');
Error('\P{Numeric_Value=	-01/000000006/a/}');
Expect(1, 74849, '\p{Numeric_Value=1/06}', "");
Expect(0, 74849, '\p{^Numeric_Value=1/06}', "");
Expect(0, 74849, '\P{Numeric_Value=1/06}', "");
Expect(1, 74849, '\P{^Numeric_Value=1/06}', "");
Expect(0, 74850, '\p{Numeric_Value=1/06}', "");
Expect(1, 74850, '\p{^Numeric_Value=1/06}', "");
Expect(1, 74850, '\P{Numeric_Value=1/06}', "");
Expect(0, 74850, '\P{^Numeric_Value=1/06}', "");
Error('\p{Numeric_Value:   0.17}');
Error('\P{Numeric_Value:   0.17}');
Expect(1, 74849, '\p{Numeric_Value=0.167}', "");
Expect(0, 74849, '\p{^Numeric_Value=0.167}', "");
Expect(0, 74849, '\P{Numeric_Value=0.167}', "");
Expect(1, 74849, '\P{^Numeric_Value=0.167}', "");
Expect(0, 74850, '\p{Numeric_Value=0.167}', "");
Expect(1, 74850, '\p{^Numeric_Value=0.167}', "");
Expect(1, 74850, '\P{Numeric_Value=0.167}', "");
Expect(0, 74850, '\P{^Numeric_Value=0.167}', "");
Error('\p{Nv=	:=001/0006}');
Error('\P{Nv=	:=001/0006}');
Expect(1, 74849, '\p{Nv=00001/006}', "");
Expect(0, 74849, '\p{^Nv=00001/006}', "");
Expect(0, 74849, '\P{Nv=00001/006}', "");
Expect(1, 74849, '\P{^Nv=00001/006}', "");
Expect(0, 74850, '\p{Nv=00001/006}', "");
Expect(1, 74850, '\p{^Nv=00001/006}', "");
Expect(1, 74850, '\P{Nv=00001/006}', "");
Expect(0, 74850, '\P{^Nv=00001/006}', "");
Error('\p{Nv=0.17}');
Error('\P{Nv=0.17}');
Expect(1, 74849, '\p{Nv=0.167}', "");
Expect(0, 74849, '\p{^Nv=0.167}', "");
Expect(0, 74849, '\P{Nv=0.167}', "");
Expect(1, 74849, '\P{^Nv=0.167}', "");
Expect(0, 74850, '\p{Nv=0.167}', "");
Expect(1, 74850, '\p{^Nv=0.167}', "");
Expect(1, 74850, '\P{Nv=0.167}', "");
Expect(0, 74850, '\P{^Nv=0.167}', "");
Error('\p{Is_Numeric_Value=	 00000001/0000000006/a/}');
Error('\P{Is_Numeric_Value=	 00000001/0000000006/a/}');
Expect(1, 74849, '\p{Is_Numeric_Value=+00000001/0000006}', "");
Expect(0, 74849, '\p{^Is_Numeric_Value=+00000001/0000006}', "");
Expect(0, 74849, '\P{Is_Numeric_Value=+00000001/0000006}', "");
Expect(1, 74849, '\P{^Is_Numeric_Value=+00000001/0000006}', "");
Expect(0, 74850, '\p{Is_Numeric_Value=+00000001/0000006}', "");
Expect(1, 74850, '\p{^Is_Numeric_Value=+00000001/0000006}', "");
Expect(1, 74850, '\P{Is_Numeric_Value=+00000001/0000006}', "");
Expect(0, 74850, '\P{^Is_Numeric_Value=+00000001/0000006}', "");
Error('\p{Is_Numeric_Value=0.17}');
Error('\P{Is_Numeric_Value=0.17}');
Expect(1, 74849, '\p{Is_Numeric_Value=0.167}', "");
Expect(0, 74849, '\p{^Is_Numeric_Value=0.167}', "");
Expect(0, 74849, '\P{Is_Numeric_Value=0.167}', "");
Expect(1, 74849, '\P{^Is_Numeric_Value=0.167}', "");
Expect(0, 74850, '\p{Is_Numeric_Value=0.167}', "");
Expect(1, 74850, '\p{^Is_Numeric_Value=0.167}', "");
Expect(1, 74850, '\P{Is_Numeric_Value=0.167}', "");
Expect(0, 74850, '\P{^Is_Numeric_Value=0.167}', "");
Error('\p{Is_Nv= _000001/006:=}');
Error('\P{Is_Nv= _000001/006:=}');
Expect(1, 74849, '\p{Is_Nv=00000001/0000006}', "");
Expect(0, 74849, '\p{^Is_Nv=00000001/0000006}', "");
Expect(0, 74849, '\P{Is_Nv=00000001/0000006}', "");
Expect(1, 74849, '\P{^Is_Nv=00000001/0000006}', "");
Expect(0, 74850, '\p{Is_Nv=00000001/0000006}', "");
Expect(1, 74850, '\p{^Is_Nv=00000001/0000006}', "");
Expect(1, 74850, '\P{Is_Nv=00000001/0000006}', "");
Expect(0, 74850, '\P{^Is_Nv=00000001/0000006}', "");
Error('\p{Is_Nv=0.17}');
Error('\P{Is_Nv=0.17}');
Expect(1, 74849, '\p{Is_Nv=0.167}', "");
Expect(0, 74849, '\p{^Is_Nv=0.167}', "");
Expect(0, 74849, '\P{Is_Nv=0.167}', "");
Expect(1, 74849, '\P{^Is_Nv=0.167}', "");
Expect(0, 74850, '\p{Is_Nv=0.167}', "");
Expect(1, 74850, '\p{^Is_Nv=0.167}', "");
Expect(1, 74850, '\P{Is_Nv=0.167}', "");
Expect(0, 74850, '\P{^Is_Nv=0.167}', "");
Error('\p{Numeric_Value=:=_-1/0000000007}');
Error('\P{Numeric_Value=:=_-1/0000000007}');
Expect(1, 8528, '\p{Numeric_Value=000001/00007}', "");
Expect(0, 8528, '\p{^Numeric_Value=000001/00007}', "");
Expect(0, 8528, '\P{Numeric_Value=000001/00007}', "");
Expect(1, 8528, '\P{^Numeric_Value=000001/00007}', "");
Expect(0, 8529, '\p{Numeric_Value=000001/00007}', "");
Expect(1, 8529, '\p{^Numeric_Value=000001/00007}', "");
Expect(1, 8529, '\P{Numeric_Value=000001/00007}', "");
Expect(0, 8529, '\P{^Numeric_Value=000001/00007}', "");
Error('\p{Numeric_Value=0.14}');
Error('\P{Numeric_Value=0.14}');
Expect(1, 8528, '\p{Numeric_Value=0.143}', "");
Expect(0, 8528, '\p{^Numeric_Value=0.143}', "");
Expect(0, 8528, '\P{Numeric_Value=0.143}', "");
Expect(1, 8528, '\P{^Numeric_Value=0.143}', "");
Expect(0, 8529, '\p{Numeric_Value=0.143}', "");
Expect(1, 8529, '\p{^Numeric_Value=0.143}', "");
Expect(1, 8529, '\P{Numeric_Value=0.143}', "");
Expect(0, 8529, '\P{^Numeric_Value=0.143}', "");
Error('\p{Nv=/a/-00001/000007}');
Error('\P{Nv=/a/-00001/000007}');
Expect(1, 8528, '\p{Nv=000001/00000007}', "");
Expect(0, 8528, '\p{^Nv=000001/00000007}', "");
Expect(0, 8528, '\P{Nv=000001/00000007}', "");
Expect(1, 8528, '\P{^Nv=000001/00000007}', "");
Expect(0, 8529, '\p{Nv=000001/00000007}', "");
Expect(1, 8529, '\p{^Nv=000001/00000007}', "");
Expect(1, 8529, '\P{Nv=000001/00000007}', "");
Expect(0, 8529, '\P{^Nv=000001/00000007}', "");
Error('\p{Nv:   0.14}');
Error('\P{Nv:   0.14}');
Expect(1, 8528, '\p{Nv=0.143}', "");
Expect(0, 8528, '\p{^Nv=0.143}', "");
Expect(0, 8528, '\P{Nv=0.143}', "");
Expect(1, 8528, '\P{^Nv=0.143}', "");
Expect(0, 8529, '\p{Nv=0.143}', "");
Expect(1, 8529, '\p{^Nv=0.143}', "");
Expect(1, 8529, '\P{Nv=0.143}', "");
Expect(0, 8529, '\P{^Nv=0.143}', "");
Error('\p{Is_Numeric_Value=-/a/000001/000007}');
Error('\P{Is_Numeric_Value=-/a/000001/000007}');
Expect(1, 8528, '\p{Is_Numeric_Value=0000000001/07}', "");
Expect(0, 8528, '\p{^Is_Numeric_Value=0000000001/07}', "");
Expect(0, 8528, '\P{Is_Numeric_Value=0000000001/07}', "");
Expect(1, 8528, '\P{^Is_Numeric_Value=0000000001/07}', "");
Expect(0, 8529, '\p{Is_Numeric_Value=0000000001/07}', "");
Expect(1, 8529, '\p{^Is_Numeric_Value=0000000001/07}', "");
Expect(1, 8529, '\P{Is_Numeric_Value=0000000001/07}', "");
Expect(0, 8529, '\P{^Is_Numeric_Value=0000000001/07}', "");
Error('\p{Is_Numeric_Value=0.14}');
Error('\P{Is_Numeric_Value=0.14}');
Expect(1, 8528, '\p{Is_Numeric_Value=0.143}', "");
Expect(0, 8528, '\p{^Is_Numeric_Value=0.143}', "");
Expect(0, 8528, '\P{Is_Numeric_Value=0.143}', "");
Expect(1, 8528, '\P{^Is_Numeric_Value=0.143}', "");
Expect(0, 8529, '\p{Is_Numeric_Value=0.143}', "");
Expect(1, 8529, '\p{^Is_Numeric_Value=0.143}', "");
Expect(1, 8529, '\P{Is_Numeric_Value=0.143}', "");
Expect(0, 8529, '\P{^Is_Numeric_Value=0.143}', "");
Error('\p{Is_Nv=	:=0001/000000007}');
Error('\P{Is_Nv=	:=0001/000000007}');
Expect(1, 8528, '\p{Is_Nv=01/000000007}', "");
Expect(0, 8528, '\p{^Is_Nv=01/000000007}', "");
Expect(0, 8528, '\P{Is_Nv=01/000000007}', "");
Expect(1, 8528, '\P{^Is_Nv=01/000000007}', "");
Expect(0, 8529, '\p{Is_Nv=01/000000007}', "");
Expect(1, 8529, '\p{^Is_Nv=01/000000007}', "");
Expect(1, 8529, '\P{Is_Nv=01/000000007}', "");
Expect(0, 8529, '\P{^Is_Nv=01/000000007}', "");
Error('\p{Is_Nv=0.14}');
Error('\P{Is_Nv=0.14}');
Expect(1, 8528, '\p{Is_Nv=0.143}', "");
Expect(0, 8528, '\p{^Is_Nv=0.143}', "");
Expect(0, 8528, '\P{Is_Nv=0.143}', "");
Expect(1, 8528, '\P{^Is_Nv=0.143}', "");
Expect(0, 8529, '\p{Is_Nv=0.143}', "");
Expect(1, 8529, '\p{^Is_Nv=0.143}', "");
Expect(1, 8529, '\P{Is_Nv=0.143}', "");
Expect(0, 8529, '\P{^Is_Nv=0.143}', "");
Error('\p{Numeric_Value= :=1/08}');
Error('\P{Numeric_Value= :=1/08}');
Expect(1, 74847, '\p{Numeric_Value=001/0000008}', "");
Expect(0, 74847, '\p{^Numeric_Value=001/0000008}', "");
Expect(0, 74847, '\P{Numeric_Value=001/0000008}', "");
Expect(1, 74847, '\P{^Numeric_Value=001/0000008}', "");
Expect(0, 74848, '\p{Numeric_Value=001/0000008}', "");
Expect(1, 74848, '\p{^Numeric_Value=001/0000008}', "");
Expect(1, 74848, '\P{Numeric_Value=001/0000008}', "");
Expect(0, 74848, '\P{^Numeric_Value=001/0000008}', "");
Error('\p{Numeric_Value=0.12}');
Error('\P{Numeric_Value=0.12}');
Expect(1, 74847, '\p{Numeric_Value=0.125}', "");
Expect(0, 74847, '\p{^Numeric_Value=0.125}', "");
Expect(0, 74847, '\P{Numeric_Value=0.125}', "");
Expect(1, 74847, '\P{^Numeric_Value=0.125}', "");
Expect(0, 74848, '\p{Numeric_Value=0.125}', "");
Expect(1, 74848, '\p{^Numeric_Value=0.125}', "");
Expect(1, 74848, '\P{Numeric_Value=0.125}', "");
Expect(0, 74848, '\P{^Numeric_Value=0.125}', "");
Error('\p{Nv=-:=000000001/000000008}');
Error('\P{Nv=-:=000000001/000000008}');
Expect(1, 74847, '\p{Nv=00000001/08}', "");
Expect(0, 74847, '\p{^Nv=00000001/08}', "");
Expect(0, 74847, '\P{Nv=00000001/08}', "");
Expect(1, 74847, '\P{^Nv=00000001/08}', "");
Expect(0, 74848, '\p{Nv=00000001/08}', "");
Expect(1, 74848, '\p{^Nv=00000001/08}', "");
Expect(1, 74848, '\P{Nv=00000001/08}', "");
Expect(0, 74848, '\P{^Nv=00000001/08}', "");
Error('\p{Nv=0.12}');
Error('\P{Nv=0.12}');
Expect(1, 74847, '\p{Nv=0.125}', "");
Expect(0, 74847, '\p{^Nv=0.125}', "");
Expect(0, 74847, '\P{Nv=0.125}', "");
Expect(1, 74847, '\P{^Nv=0.125}', "");
Expect(0, 74848, '\p{Nv=0.125}', "");
Expect(1, 74848, '\p{^Nv=0.125}', "");
Expect(1, 74848, '\P{Nv=0.125}', "");
Expect(0, 74848, '\P{^Nv=0.125}', "");
Error('\p{Is_Numeric_Value=_	0000000001/8:=}');
Error('\P{Is_Numeric_Value=_	0000000001/8:=}');
Expect(1, 74847, '\p{Is_Numeric_Value=0001/8}', "");
Expect(0, 74847, '\p{^Is_Numeric_Value=0001/8}', "");
Expect(0, 74847, '\P{Is_Numeric_Value=0001/8}', "");
Expect(1, 74847, '\P{^Is_Numeric_Value=0001/8}', "");
Expect(0, 74848, '\p{Is_Numeric_Value=0001/8}', "");
Expect(1, 74848, '\p{^Is_Numeric_Value=0001/8}', "");
Expect(1, 74848, '\P{Is_Numeric_Value=0001/8}', "");
Expect(0, 74848, '\P{^Is_Numeric_Value=0001/8}', "");
Error('\p{Is_Numeric_Value=0.12}');
Error('\P{Is_Numeric_Value=0.12}');
Expect(1, 74847, '\p{Is_Numeric_Value=0.125}', "");
Expect(0, 74847, '\p{^Is_Numeric_Value=0.125}', "");
Expect(0, 74847, '\P{Is_Numeric_Value=0.125}', "");
Expect(1, 74847, '\P{^Is_Numeric_Value=0.125}', "");
Expect(0, 74848, '\p{Is_Numeric_Value=0.125}', "");
Expect(1, 74848, '\p{^Is_Numeric_Value=0.125}', "");
Expect(1, 74848, '\P{Is_Numeric_Value=0.125}', "");
Expect(0, 74848, '\P{^Is_Numeric_Value=0.125}', "");
Error('\p{Is_Nv=  0000000001/8:=}');
Error('\P{Is_Nv=  0000000001/8:=}');
Expect(1, 74847, '\p{Is_Nv=+00000001/00008}', "");
Expect(0, 74847, '\p{^Is_Nv=+00000001/00008}', "");
Expect(0, 74847, '\P{Is_Nv=+00000001/00008}', "");
Expect(1, 74847, '\P{^Is_Nv=+00000001/00008}', "");
Expect(0, 74848, '\p{Is_Nv=+00000001/00008}', "");
Expect(1, 74848, '\p{^Is_Nv=+00000001/00008}', "");
Expect(1, 74848, '\P{Is_Nv=+00000001/00008}', "");
Expect(0, 74848, '\P{^Is_Nv=+00000001/00008}', "");
Error('\p{Is_Nv=0.12}');
Error('\P{Is_Nv=0.12}');
Expect(1, 74847, '\p{Is_Nv=0.125}', "");
Expect(0, 74847, '\p{^Is_Nv=0.125}', "");
Expect(0, 74847, '\P{Is_Nv=0.125}', "");
Expect(1, 74847, '\P{^Is_Nv=0.125}', "");
Expect(0, 74848, '\p{Is_Nv=0.125}', "");
Expect(1, 74848, '\p{^Is_Nv=0.125}', "");
Expect(1, 74848, '\P{Is_Nv=0.125}', "");
Expect(0, 74848, '\P{^Is_Nv=0.125}', "");
Error('\p{Numeric_Value:   _-00001/00000009:=}');
Error('\P{Numeric_Value:   _-00001/00000009:=}');
Expect(1, 8529, '\p{Numeric_Value=000001/0009}', "");
Expect(0, 8529, '\p{^Numeric_Value=000001/0009}', "");
Expect(0, 8529, '\P{Numeric_Value=000001/0009}', "");
Expect(1, 8529, '\P{^Numeric_Value=000001/0009}', "");
Expect(0, 8530, '\p{Numeric_Value=000001/0009}', "");
Expect(1, 8530, '\p{^Numeric_Value=000001/0009}', "");
Expect(1, 8530, '\P{Numeric_Value=000001/0009}', "");
Expect(0, 8530, '\P{^Numeric_Value=000001/0009}', "");
Error('\p{Numeric_Value=0.11}');
Error('\P{Numeric_Value=0.11}');
Expect(1, 8529, '\p{Numeric_Value:   0.111}', "");
Expect(0, 8529, '\p{^Numeric_Value:   0.111}', "");
Expect(0, 8529, '\P{Numeric_Value:   0.111}', "");
Expect(1, 8529, '\P{^Numeric_Value:   0.111}', "");
Expect(0, 8530, '\p{Numeric_Value:   0.111}', "");
Expect(1, 8530, '\p{^Numeric_Value:   0.111}', "");
Expect(1, 8530, '\P{Numeric_Value:   0.111}', "");
Expect(0, 8530, '\P{^Numeric_Value:   0.111}', "");
Error('\p{Nv=_+00000001/00000009/a/}');
Error('\P{Nv=_+00000001/00000009/a/}');
Expect(1, 8529, '\p{Nv=0000001/0000000009}', "");
Expect(0, 8529, '\p{^Nv=0000001/0000000009}', "");
Expect(0, 8529, '\P{Nv=0000001/0000000009}', "");
Expect(1, 8529, '\P{^Nv=0000001/0000000009}', "");
Expect(0, 8530, '\p{Nv=0000001/0000000009}', "");
Expect(1, 8530, '\p{^Nv=0000001/0000000009}', "");
Expect(1, 8530, '\P{Nv=0000001/0000000009}', "");
Expect(0, 8530, '\P{^Nv=0000001/0000000009}', "");
Error('\p{Nv=0.11}');
Error('\P{Nv=0.11}');
Expect(1, 8529, '\p{Nv=0.111}', "");
Expect(0, 8529, '\p{^Nv=0.111}', "");
Expect(0, 8529, '\P{Nv=0.111}', "");
Expect(1, 8529, '\P{^Nv=0.111}', "");
Expect(0, 8530, '\p{Nv=0.111}', "");
Expect(1, 8530, '\p{^Nv=0.111}', "");
Expect(1, 8530, '\P{Nv=0.111}', "");
Expect(0, 8530, '\P{^Nv=0.111}', "");
Error('\p{Is_Numeric_Value=/a/_01/00009}');
Error('\P{Is_Numeric_Value=/a/_01/00009}');
Expect(1, 8529, '\p{Is_Numeric_Value=+1/0000009}', "");
Expect(0, 8529, '\p{^Is_Numeric_Value=+1/0000009}', "");
Expect(0, 8529, '\P{Is_Numeric_Value=+1/0000009}', "");
Expect(1, 8529, '\P{^Is_Numeric_Value=+1/0000009}', "");
Expect(0, 8530, '\p{Is_Numeric_Value=+1/0000009}', "");
Expect(1, 8530, '\p{^Is_Numeric_Value=+1/0000009}', "");
Expect(1, 8530, '\P{Is_Numeric_Value=+1/0000009}', "");
Expect(0, 8530, '\P{^Is_Numeric_Value=+1/0000009}', "");
Error('\p{Is_Numeric_Value: 0.11}');
Error('\P{Is_Numeric_Value: 0.11}');
Expect(1, 8529, '\p{Is_Numeric_Value=0.111}', "");
Expect(0, 8529, '\p{^Is_Numeric_Value=0.111}', "");
Expect(0, 8529, '\P{Is_Numeric_Value=0.111}', "");
Expect(1, 8529, '\P{^Is_Numeric_Value=0.111}', "");
Expect(0, 8530, '\p{Is_Numeric_Value=0.111}', "");
Expect(1, 8530, '\p{^Is_Numeric_Value=0.111}', "");
Expect(1, 8530, '\P{Is_Numeric_Value=0.111}', "");
Expect(0, 8530, '\P{^Is_Numeric_Value=0.111}', "");
Error('\p{Is_Nv=		000000001/000000009/a/}');
Error('\P{Is_Nv=		000000001/000000009/a/}');
Expect(1, 8529, '\p{Is_Nv=01/09}', "");
Expect(0, 8529, '\p{^Is_Nv=01/09}', "");
Expect(0, 8529, '\P{Is_Nv=01/09}', "");
Expect(1, 8529, '\P{^Is_Nv=01/09}', "");
Expect(0, 8530, '\p{Is_Nv=01/09}', "");
Expect(1, 8530, '\p{^Is_Nv=01/09}', "");
Expect(1, 8530, '\P{Is_Nv=01/09}', "");
Expect(0, 8530, '\P{^Is_Nv=01/09}', "");
Error('\p{Is_Nv=0.11}');
Error('\P{Is_Nv=0.11}');
Expect(1, 8529, '\p{Is_Nv=0.111}', "");
Expect(0, 8529, '\p{^Is_Nv=0.111}', "");
Expect(0, 8529, '\P{Is_Nv=0.111}', "");
Expect(1, 8529, '\P{^Is_Nv=0.111}', "");
Expect(0, 8530, '\p{Is_Nv=0.111}', "");
Expect(1, 8530, '\p{^Is_Nv=0.111}', "");
Expect(1, 8530, '\P{Is_Nv=0.111}', "");
Expect(0, 8530, '\P{^Is_Nv=0.111}', "");
Error('\p{Numeric_Value=:=		0_0_0_10}');
Error('\P{Numeric_Value=:=		0_0_0_10}');
Expect(1, 119657, '\p{Numeric_Value:	000010}', "");
Expect(0, 119657, '\p{^Numeric_Value:	000010}', "");
Expect(0, 119657, '\P{Numeric_Value:	000010}', "");
Expect(1, 119657, '\P{^Numeric_Value:	000010}', "");
Expect(0, 119658, '\p{Numeric_Value:	000010}', "");
Expect(1, 119658, '\p{^Numeric_Value:	000010}', "");
Expect(1, 119658, '\P{Numeric_Value:	000010}', "");
Expect(0, 119658, '\P{^Numeric_Value:	000010}', "");
Error('\p{Nv=:=	01_0}');
Error('\P{Nv=:=	01_0}');
Expect(1, 119657, '\p{Nv=0_0_0_0_0_0_10}', "");
Expect(0, 119657, '\p{^Nv=0_0_0_0_0_0_10}', "");
Expect(0, 119657, '\P{Nv=0_0_0_0_0_0_10}', "");
Expect(1, 119657, '\P{^Nv=0_0_0_0_0_0_10}', "");
Expect(0, 119658, '\p{Nv=0_0_0_0_0_0_10}', "");
Expect(1, 119658, '\p{^Nv=0_0_0_0_0_0_10}', "");
Expect(1, 119658, '\P{Nv=0_0_0_0_0_0_10}', "");
Expect(0, 119658, '\P{^Nv=0_0_0_0_0_0_10}', "");
Error('\p{Is_Numeric_Value=:=-+0001_0}');
Error('\P{Is_Numeric_Value=:=-+0001_0}');
Expect(1, 119657, '\p{Is_Numeric_Value=0_0_10}', "");
Expect(0, 119657, '\p{^Is_Numeric_Value=0_0_10}', "");
Expect(0, 119657, '\P{Is_Numeric_Value=0_0_10}', "");
Expect(1, 119657, '\P{^Is_Numeric_Value=0_0_10}', "");
Expect(0, 119658, '\p{Is_Numeric_Value=0_0_10}', "");
Expect(1, 119658, '\p{^Is_Numeric_Value=0_0_10}', "");
Expect(1, 119658, '\P{Is_Numeric_Value=0_0_10}', "");
Expect(0, 119658, '\P{^Is_Numeric_Value=0_0_10}', "");
Error('\p{Is_Nv=/a/_00000010}');
Error('\P{Is_Nv=/a/_00000010}');
Expect(1, 119657, '\p{Is_Nv: +00000000010}', "");
Expect(0, 119657, '\p{^Is_Nv: +00000000010}', "");
Expect(0, 119657, '\P{Is_Nv: +00000000010}', "");
Expect(1, 119657, '\P{^Is_Nv: +00000000010}', "");
Expect(0, 119658, '\p{Is_Nv: +00000000010}', "");
Expect(1, 119658, '\p{^Is_Nv: +00000000010}', "");
Expect(1, 119658, '\P{Is_Nv: +00000000010}', "");
Expect(0, 119658, '\P{^Is_Nv: +00000000010}', "");
Error('\p{Numeric_Value=	-000000100:=}');
Error('\P{Numeric_Value=	-000000100:=}');
Expect(1, 69732, '\p{Numeric_Value=+0010_0}', "");
Expect(0, 69732, '\p{^Numeric_Value=+0010_0}', "");
Expect(0, 69732, '\P{Numeric_Value=+0010_0}', "");
Expect(1, 69732, '\P{^Numeric_Value=+0010_0}', "");
Expect(0, 69733, '\p{Numeric_Value=+0010_0}', "");
Expect(1, 69733, '\p{^Numeric_Value=+0010_0}', "");
Expect(1, 69733, '\P{Numeric_Value=+0010_0}', "");
Expect(0, 69733, '\P{^Numeric_Value=+0010_0}', "");
Error('\p{Nv=/a/-_00100}');
Error('\P{Nv=/a/-_00100}');
Expect(1, 69732, '\p{Nv=+0_0_0_100}', "");
Expect(0, 69732, '\p{^Nv=+0_0_0_100}', "");
Expect(0, 69732, '\P{Nv=+0_0_0_100}', "");
Expect(1, 69732, '\P{^Nv=+0_0_0_100}', "");
Expect(0, 69733, '\p{Nv=+0_0_0_100}', "");
Expect(1, 69733, '\p{^Nv=+0_0_0_100}', "");
Expect(1, 69733, '\P{Nv=+0_0_0_100}', "");
Expect(0, 69733, '\P{^Nv=+0_0_0_100}', "");
Error('\p{Is_Numeric_Value=	:=000_001_00}');
Error('\P{Is_Numeric_Value=	:=000_001_00}');
Expect(1, 69732, '\p{Is_Numeric_Value=+00000000100}', "");
Expect(0, 69732, '\p{^Is_Numeric_Value=+00000000100}', "");
Expect(0, 69732, '\P{Is_Numeric_Value=+00000000100}', "");
Expect(1, 69732, '\P{^Is_Numeric_Value=+00000000100}', "");
Expect(0, 69733, '\p{Is_Numeric_Value=+00000000100}', "");
Expect(1, 69733, '\p{^Is_Numeric_Value=+00000000100}', "");
Expect(1, 69733, '\P{Is_Numeric_Value=+00000000100}', "");
Expect(0, 69733, '\P{^Is_Numeric_Value=+00000000100}', "");
Error('\p{Is_Nv=/a/ 100}');
Error('\P{Is_Nv=/a/ 100}');
Expect(1, 69732, '\p{Is_Nv=000_000_000_100}', "");
Expect(0, 69732, '\p{^Is_Nv=000_000_000_100}', "");
Expect(0, 69732, '\P{Is_Nv=000_000_000_100}', "");
Expect(1, 69732, '\P{^Is_Nv=000_000_000_100}', "");
Expect(0, 69733, '\p{Is_Nv=000_000_000_100}', "");
Expect(1, 69733, '\p{^Is_Nv=000_000_000_100}', "");
Expect(1, 69733, '\P{Is_Nv=000_000_000_100}', "");
Expect(0, 69733, '\P{^Is_Nv=000_000_000_100}', "");
Error('\p{Numeric_Value=:=		0_0_0_0_1000}');
Error('\P{Numeric_Value=:=		0_0_0_0_1000}');
Expect(1, 69733, '\p{Numeric_Value: 01000}', "");
Expect(0, 69733, '\p{^Numeric_Value: 01000}', "");
Expect(0, 69733, '\P{Numeric_Value: 01000}', "");
Expect(1, 69733, '\P{^Numeric_Value: 01000}', "");
Expect(0, 69734, '\p{Numeric_Value: 01000}', "");
Expect(1, 69734, '\p{^Numeric_Value: 01000}', "");
Expect(1, 69734, '\P{Numeric_Value: 01000}', "");
Expect(0, 69734, '\P{^Numeric_Value: 01000}', "");
Error('\p{Nv:	_ 01000/a/}');
Error('\P{Nv:	_ 01000/a/}');
Expect(1, 69733, '\p{Nv=001000}', "");
Expect(0, 69733, '\p{^Nv=001000}', "");
Expect(0, 69733, '\P{Nv=001000}', "");
Expect(1, 69733, '\P{^Nv=001000}', "");
Expect(0, 69734, '\p{Nv=001000}', "");
Expect(1, 69734, '\p{^Nv=001000}', "");
Expect(1, 69734, '\P{Nv=001000}', "");
Expect(0, 69734, '\P{^Nv=001000}', "");
Error('\p{Is_Numeric_Value=  0000_0010_00/a/}');
Error('\P{Is_Numeric_Value=  0000_0010_00/a/}');
Expect(1, 69733, '\p{Is_Numeric_Value=1000}', "");
Expect(0, 69733, '\p{^Is_Numeric_Value=1000}', "");
Expect(0, 69733, '\P{Is_Numeric_Value=1000}', "");
Expect(1, 69733, '\P{^Is_Numeric_Value=1000}', "");
Expect(0, 69734, '\p{Is_Numeric_Value=1000}', "");
Expect(1, 69734, '\p{^Is_Numeric_Value=1000}', "");
Expect(1, 69734, '\P{Is_Numeric_Value=1000}', "");
Expect(0, 69734, '\P{^Is_Numeric_Value=1000}', "");
Error('\p{Is_Nv=:=0_0_0_0_1000}');
Error('\P{Is_Nv=:=0_0_0_0_1000}');
Expect(1, 69733, '\p{Is_Nv:	100_0}', "");
Expect(0, 69733, '\p{^Is_Nv:	100_0}', "");
Expect(0, 69733, '\P{Is_Nv:	100_0}', "");
Expect(1, 69733, '\P{^Is_Nv:	100_0}', "");
Expect(0, 69734, '\p{Is_Nv:	100_0}', "");
Expect(1, 69734, '\p{^Is_Nv:	100_0}', "");
Expect(1, 69734, '\P{Is_Nv:	100_0}', "");
Expect(0, 69734, '\P{^Is_Nv:	100_0}', "");
Error('\p{Numeric_Value=/a/_-01_00_00}');
Error('\P{Numeric_Value=/a/_-01_00_00}');
Expect(1, 67679, '\p{Numeric_Value=000000010000}', "");
Expect(0, 67679, '\p{^Numeric_Value=000000010000}', "");
Expect(0, 67679, '\P{Numeric_Value=000000010000}', "");
Expect(1, 67679, '\P{^Numeric_Value=000000010000}', "");
Expect(0, 67680, '\p{Numeric_Value=000000010000}', "");
Expect(1, 67680, '\p{^Numeric_Value=000000010000}', "");
Expect(1, 67680, '\P{Numeric_Value=000000010000}', "");
Expect(0, 67680, '\P{^Numeric_Value=000000010000}', "");
Error('\p{Nv=- 00010000/a/}');
Error('\P{Nv=- 00010000/a/}');
Expect(1, 67679, '\p{Nv=+00000010000}', "");
Expect(0, 67679, '\p{^Nv=+00000010000}', "");
Expect(0, 67679, '\P{Nv=+00000010000}', "");
Expect(1, 67679, '\P{^Nv=+00000010000}', "");
Expect(0, 67680, '\p{Nv=+00000010000}', "");
Expect(1, 67680, '\p{^Nv=+00000010000}', "");
Expect(1, 67680, '\P{Nv=+00000010000}', "");
Expect(0, 67680, '\P{^Nv=+00000010000}', "");
Error('\p{Is_Numeric_Value=--00000010000/a/}');
Error('\P{Is_Numeric_Value=--00000010000/a/}');
Expect(1, 67679, '\p{Is_Numeric_Value=0_0_0_0_0_0_0_1_0_000}', "");
Expect(0, 67679, '\p{^Is_Numeric_Value=0_0_0_0_0_0_0_1_0_000}', "");
Expect(0, 67679, '\P{Is_Numeric_Value=0_0_0_0_0_0_0_1_0_000}', "");
Expect(1, 67679, '\P{^Is_Numeric_Value=0_0_0_0_0_0_0_1_0_000}', "");
Expect(0, 67680, '\p{Is_Numeric_Value=0_0_0_0_0_0_0_1_0_000}', "");
Expect(1, 67680, '\p{^Is_Numeric_Value=0_0_0_0_0_0_0_1_0_000}', "");
Expect(1, 67680, '\P{Is_Numeric_Value=0_0_0_0_0_0_0_1_0_000}', "");
Expect(0, 67680, '\P{^Is_Numeric_Value=0_0_0_0_0_0_0_1_0_000}', "");
Error('\p{Is_Nv=_:=+00000010000}');
Error('\P{Is_Nv=_:=+00000010000}');
Expect(1, 67679, '\p{Is_Nv=00000010000}', "");
Expect(0, 67679, '\p{^Is_Nv=00000010000}', "");
Expect(0, 67679, '\P{Is_Nv=00000010000}', "");
Expect(1, 67679, '\P{^Is_Nv=00000010000}', "");
Expect(0, 67680, '\p{Is_Nv=00000010000}', "");
Expect(1, 67680, '\p{^Is_Nv=00000010000}', "");
Expect(1, 67680, '\P{Is_Nv=00000010000}', "");
Expect(0, 67680, '\P{^Is_Nv=00000010000}', "");
Error('\p{Numeric_Value:   _/a/00100000}');
Error('\P{Numeric_Value:   _/a/00100000}');
Expect(1, 8584, '\p{Numeric_Value=0000100000}', "");
Expect(0, 8584, '\p{^Numeric_Value=0000100000}', "");
Expect(0, 8584, '\P{Numeric_Value=0000100000}', "");
Expect(1, 8584, '\P{^Numeric_Value=0000100000}', "");
Expect(0, 8585, '\p{Numeric_Value=0000100000}', "");
Expect(1, 8585, '\p{^Numeric_Value=0000100000}', "");
Expect(1, 8585, '\P{Numeric_Value=0000100000}', "");
Expect(0, 8585, '\P{^Numeric_Value=0000100000}', "");
Error('\p{Nv=:= 000100000}');
Error('\P{Nv=:= 000100000}');
Expect(1, 8584, '\p{Nv=00000100000}', "");
Expect(0, 8584, '\p{^Nv=00000100000}', "");
Expect(0, 8584, '\P{Nv=00000100000}', "");
Expect(1, 8584, '\P{^Nv=00000100000}', "");
Expect(0, 8585, '\p{Nv=00000100000}', "");
Expect(1, 8585, '\p{^Nv=00000100000}', "");
Expect(1, 8585, '\P{Nv=00000100000}', "");
Expect(0, 8585, '\P{^Nv=00000100000}', "");
Error('\p{Is_Numeric_Value=/a/ -10000_0}');
Error('\P{Is_Numeric_Value=/a/ -10000_0}');
Expect(1, 8584, '\p{Is_Numeric_Value=00000100000}', "");
Expect(0, 8584, '\p{^Is_Numeric_Value=00000100000}', "");
Expect(0, 8584, '\P{Is_Numeric_Value=00000100000}', "");
Expect(1, 8584, '\P{^Is_Numeric_Value=00000100000}', "");
Expect(0, 8585, '\p{Is_Numeric_Value=00000100000}', "");
Expect(1, 8585, '\p{^Is_Numeric_Value=00000100000}', "");
Expect(1, 8585, '\P{Is_Numeric_Value=00000100000}', "");
Expect(0, 8585, '\P{^Is_Numeric_Value=00000100000}', "");
Error('\p{Is_Nv= 00100000/a/}');
Error('\P{Is_Nv= 00100000/a/}');
Expect(1, 8584, '\p{Is_Nv=+00100000}', "");
Expect(0, 8584, '\p{^Is_Nv=+00100000}', "");
Expect(0, 8584, '\P{Is_Nv=+00100000}', "");
Expect(1, 8584, '\P{^Is_Nv=+00100000}', "");
Expect(0, 8585, '\p{Is_Nv=+00100000}', "");
Expect(1, 8585, '\p{^Is_Nv=+00100000}', "");
Expect(1, 8585, '\P{Is_Nv=+00100000}', "");
Expect(0, 8585, '\P{^Is_Nv=+00100000}', "");
Error('\p{Numeric_Value=/a/__0_0_0_1_0_0_000000}');
Error('\P{Numeric_Value=/a/__0_0_0_1_0_0_000000}');
Expect(1, 20740, '\p{Numeric_Value: 0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_00}', "");
Expect(0, 20740, '\p{^Numeric_Value: 0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_00}', "");
Expect(0, 20740, '\P{Numeric_Value: 0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_00}', "");
Expect(1, 20740, '\P{^Numeric_Value: 0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_00}', "");
Expect(0, 20741, '\p{Numeric_Value: 0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_00}', "");
Expect(1, 20741, '\p{^Numeric_Value: 0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_00}', "");
Expect(1, 20741, '\P{Numeric_Value: 0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_00}', "");
Expect(0, 20741, '\P{^Numeric_Value: 0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_00}', "");
Error('\p{Nv= /a/000000100000000}');
Error('\P{Nv= /a/000000100000000}');
Expect(1, 20740, '\p{Nv:   00010000000_0}', "");
Expect(0, 20740, '\p{^Nv:   00010000000_0}', "");
Expect(0, 20740, '\P{Nv:   00010000000_0}', "");
Expect(1, 20740, '\P{^Nv:   00010000000_0}', "");
Expect(0, 20741, '\p{Nv:   00010000000_0}', "");
Expect(1, 20741, '\p{^Nv:   00010000000_0}', "");
Expect(1, 20741, '\P{Nv:   00010000000_0}', "");
Expect(0, 20741, '\P{^Nv:   00010000000_0}', "");
Error('\p{Is_Numeric_Value:	+00000_00100_00000_0:=}');
Error('\P{Is_Numeric_Value:	+00000_00100_00000_0:=}');
Expect(1, 20740, '\p{Is_Numeric_Value:100000000}', "");
Expect(0, 20740, '\p{^Is_Numeric_Value:100000000}', "");
Expect(0, 20740, '\P{Is_Numeric_Value:100000000}', "");
Expect(1, 20740, '\P{^Is_Numeric_Value:100000000}', "");
Expect(0, 20741, '\p{Is_Numeric_Value:100000000}', "");
Expect(1, 20741, '\p{^Is_Numeric_Value:100000000}', "");
Expect(1, 20741, '\P{Is_Numeric_Value:100000000}', "");
Expect(0, 20741, '\P{^Is_Numeric_Value:100000000}', "");
Error('\p{Is_Nv=000100000000:=}');
Error('\P{Is_Nv=000100000000:=}');
Expect(1, 20740, '\p{Is_Nv=0000100000000}', "");
Expect(0, 20740, '\p{^Is_Nv=0000100000000}', "");
Expect(0, 20740, '\P{Is_Nv=0000100000000}', "");
Expect(1, 20740, '\P{^Is_Nv=0000100000000}', "");
Expect(0, 20741, '\p{Is_Nv=0000100000000}', "");
Expect(1, 20741, '\p{^Is_Nv=0000100000000}', "");
Expect(1, 20741, '\P{Is_Nv=0000100000000}', "");
Expect(0, 20741, '\P{^Is_Nv=0000100000000}', "");
Error('\p{Numeric_Value=:=_-01000000000000}');
Error('\P{Numeric_Value=:=_-01000000000000}');
Expect(1, 20806, '\p{Numeric_Value=1000000000000}', "");
Expect(0, 20806, '\p{^Numeric_Value=1000000000000}', "");
Expect(0, 20806, '\P{Numeric_Value=1000000000000}', "");
Expect(1, 20806, '\P{^Numeric_Value=1000000000000}', "");
Expect(0, 20807, '\p{Numeric_Value=1000000000000}', "");
Expect(1, 20807, '\p{^Numeric_Value=1000000000000}', "");
Expect(1, 20807, '\P{Numeric_Value=1000000000000}', "");
Expect(0, 20807, '\P{^Numeric_Value=1000000000000}', "");
Error('\p{Nv=/a/	_+00000001000000000000}');
Error('\P{Nv=/a/	_+00000001000000000000}');
Expect(1, 20806, '\p{Nv=0000001000000000000}', "");
Expect(0, 20806, '\p{^Nv=0000001000000000000}', "");
Expect(0, 20806, '\P{Nv=0000001000000000000}', "");
Expect(1, 20806, '\P{^Nv=0000001000000000000}', "");
Expect(0, 20807, '\p{Nv=0000001000000000000}', "");
Expect(1, 20807, '\p{^Nv=0000001000000000000}', "");
Expect(1, 20807, '\P{Nv=0000001000000000000}', "");
Expect(0, 20807, '\P{^Nv=0000001000000000000}', "");
Error('\p{Is_Numeric_Value=/a/01000000000000}');
Error('\P{Is_Numeric_Value=/a/01000000000000}');
Expect(1, 20806, '\p{Is_Numeric_Value=+0000000001000000000000}', "");
Expect(0, 20806, '\p{^Is_Numeric_Value=+0000000001000000000000}', "");
Expect(0, 20806, '\P{Is_Numeric_Value=+0000000001000000000000}', "");
Expect(1, 20806, '\P{^Is_Numeric_Value=+0000000001000000000000}', "");
Expect(0, 20807, '\p{Is_Numeric_Value=+0000000001000000000000}', "");
Expect(1, 20807, '\p{^Is_Numeric_Value=+0000000001000000000000}', "");
Expect(1, 20807, '\P{Is_Numeric_Value=+0000000001000000000000}', "");
Expect(0, 20807, '\P{^Is_Numeric_Value=+0000000001000000000000}', "");
Error('\p{Is_Nv=/a/  00_00_00_00_01_00_00_00_000000}');
Error('\P{Is_Nv=/a/  00_00_00_00_01_00_00_00_000000}');
Expect(1, 20806, '\p{Is_Nv=+0_0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_0_0_0000}', "");
Expect(0, 20806, '\p{^Is_Nv=+0_0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_0_0_0000}', "");
Expect(0, 20806, '\P{Is_Nv=+0_0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_0_0_0000}', "");
Expect(1, 20806, '\P{^Is_Nv=+0_0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_0_0_0000}', "");
Expect(0, 20807, '\p{Is_Nv=+0_0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_0_0_0000}', "");
Expect(1, 20807, '\p{^Is_Nv=+0_0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_0_0_0000}', "");
Expect(1, 20807, '\P{Is_Nv=+0_0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_0_0_0000}', "");
Expect(0, 20807, '\P{^Is_Nv=+0_0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_0_0_0000}', "");
Error('\p{Numeric_Value=		000011:=}');
Error('\P{Numeric_Value=		000011:=}');
Expect(1, 9451, '\p{Numeric_Value=0_0_11}', "");
Expect(0, 9451, '\p{^Numeric_Value=0_0_11}', "");
Expect(0, 9451, '\P{Numeric_Value=0_0_11}', "");
Expect(1, 9451, '\P{^Numeric_Value=0_0_11}', "");
Expect(0, 9452, '\p{Numeric_Value=0_0_11}', "");
Expect(1, 9452, '\p{^Numeric_Value=0_0_11}', "");
Expect(1, 9452, '\P{Numeric_Value=0_0_11}', "");
Expect(0, 9452, '\P{^Numeric_Value=0_0_11}', "");
Error('\p{Nv=-	00000011/a/}');
Error('\P{Nv=-	00000011/a/}');
Expect(1, 9451, '\p{Nv:   +11}', "");
Expect(0, 9451, '\p{^Nv:   +11}', "");
Expect(0, 9451, '\P{Nv:   +11}', "");
Expect(1, 9451, '\P{^Nv:   +11}', "");
Expect(0, 9452, '\p{Nv:   +11}', "");
Expect(1, 9452, '\p{^Nv:   +11}', "");
Expect(1, 9452, '\P{Nv:   +11}', "");
Expect(0, 9452, '\P{^Nv:   +11}', "");
Error('\p{Is_Numeric_Value=/a/	01_1}');
Error('\P{Is_Numeric_Value=/a/	01_1}');
Expect(1, 9451, '\p{Is_Numeric_Value=0_0_0_0_0_011}', "");
Expect(0, 9451, '\p{^Is_Numeric_Value=0_0_0_0_0_011}', "");
Expect(0, 9451, '\P{Is_Numeric_Value=0_0_0_0_0_011}', "");
Expect(1, 9451, '\P{^Is_Numeric_Value=0_0_0_0_0_011}', "");
Expect(0, 9452, '\p{Is_Numeric_Value=0_0_0_0_0_011}', "");
Expect(1, 9452, '\p{^Is_Numeric_Value=0_0_0_0_0_011}', "");
Expect(1, 9452, '\P{Is_Numeric_Value=0_0_0_0_0_011}', "");
Expect(0, 9452, '\P{^Is_Numeric_Value=0_0_0_0_0_011}', "");
Error('\p{Is_Nv=	:=0000_0001_1}');
Error('\P{Is_Nv=	:=0000_0001_1}');
Expect(1, 9451, '\p{Is_Nv=0000000011}', "");
Expect(0, 9451, '\p{^Is_Nv=0000000011}', "");
Expect(0, 9451, '\P{Is_Nv=0000000011}', "");
Expect(1, 9451, '\P{^Is_Nv=0000000011}', "");
Expect(0, 9452, '\p{Is_Nv=0000000011}', "");
Expect(1, 9452, '\p{^Is_Nv=0000000011}', "");
Expect(1, 9452, '\P{Is_Nv=0000000011}', "");
Expect(0, 9452, '\P{^Is_Nv=0000000011}', "");
Error('\p{Numeric_Value=:= 00000000011/000002}');
Error('\P{Numeric_Value=:= 00000000011/000002}');
Expect(1, 3887, '\p{Numeric_Value=000000011/0002}', "");
Expect(0, 3887, '\p{^Numeric_Value=000000011/0002}', "");
Expect(0, 3887, '\P{Numeric_Value=000000011/0002}', "");
Expect(1, 3887, '\P{^Numeric_Value=000000011/0002}', "");
Expect(0, 3888, '\p{Numeric_Value=000000011/0002}', "");
Expect(1, 3888, '\p{^Numeric_Value=000000011/0002}', "");
Expect(1, 3888, '\P{Numeric_Value=000000011/0002}', "");
Expect(0, 3888, '\P{^Numeric_Value=000000011/0002}', "");
Expect(1, 3887, '\p{Numeric_Value=5.500}', "");
Expect(0, 3887, '\p{^Numeric_Value=5.500}', "");
Expect(0, 3887, '\P{Numeric_Value=5.500}', "");
Expect(1, 3887, '\P{^Numeric_Value=5.500}', "");
Expect(0, 3888, '\p{Numeric_Value=5.500}', "");
Expect(1, 3888, '\p{^Numeric_Value=5.500}', "");
Expect(1, 3888, '\P{Numeric_Value=5.500}', "");
Expect(0, 3888, '\P{^Numeric_Value=5.500}', "");
Error('\p{Nv=/a/_-+11/02}');
Error('\P{Nv=/a/_-+11/02}');
Expect(1, 3887, '\p{Nv=0011/0000002}', "");
Expect(0, 3887, '\p{^Nv=0011/0000002}', "");
Expect(0, 3887, '\P{Nv=0011/0000002}', "");
Expect(1, 3887, '\P{^Nv=0011/0000002}', "");
Expect(0, 3888, '\p{Nv=0011/0000002}', "");
Expect(1, 3888, '\p{^Nv=0011/0000002}', "");
Expect(1, 3888, '\P{Nv=0011/0000002}', "");
Expect(0, 3888, '\P{^Nv=0011/0000002}', "");
Expect(1, 3887, '\p{Nv=5.500}', "");
Expect(0, 3887, '\p{^Nv=5.500}', "");
Expect(0, 3887, '\P{Nv=5.500}', "");
Expect(1, 3887, '\P{^Nv=5.500}', "");
Expect(0, 3888, '\p{Nv=5.500}', "");
Expect(1, 3888, '\p{^Nv=5.500}', "");
Expect(1, 3888, '\P{Nv=5.500}', "");
Expect(0, 3888, '\P{^Nv=5.500}', "");
Error('\p{Is_Numeric_Value=/a/ _11/00002}');
Error('\P{Is_Numeric_Value=/a/ _11/00002}');
Expect(1, 3887, '\p{Is_Numeric_Value=000000011/0000002}', "");
Expect(0, 3887, '\p{^Is_Numeric_Value=000000011/0000002}', "");
Expect(0, 3887, '\P{Is_Numeric_Value=000000011/0000002}', "");
Expect(1, 3887, '\P{^Is_Numeric_Value=000000011/0000002}', "");
Expect(0, 3888, '\p{Is_Numeric_Value=000000011/0000002}', "");
Expect(1, 3888, '\p{^Is_Numeric_Value=000000011/0000002}', "");
Expect(1, 3888, '\P{Is_Numeric_Value=000000011/0000002}', "");
Expect(0, 3888, '\P{^Is_Numeric_Value=000000011/0000002}', "");
Expect(1, 3887, '\p{Is_Numeric_Value=5.500}', "");
Expect(0, 3887, '\p{^Is_Numeric_Value=5.500}', "");
Expect(0, 3887, '\P{Is_Numeric_Value=5.500}', "");
Expect(1, 3887, '\P{^Is_Numeric_Value=5.500}', "");
Expect(0, 3888, '\p{Is_Numeric_Value=5.500}', "");
Expect(1, 3888, '\p{^Is_Numeric_Value=5.500}', "");
Expect(1, 3888, '\P{Is_Numeric_Value=5.500}', "");
Expect(0, 3888, '\P{^Is_Numeric_Value=5.500}', "");
Error('\p{Is_Nv=-:=+011/0002}');
Error('\P{Is_Nv=-:=+011/0002}');
Expect(1, 3887, '\p{Is_Nv=00000011/0002}', "");
Expect(0, 3887, '\p{^Is_Nv=00000011/0002}', "");
Expect(0, 3887, '\P{Is_Nv=00000011/0002}', "");
Expect(1, 3887, '\P{^Is_Nv=00000011/0002}', "");
Expect(0, 3888, '\p{Is_Nv=00000011/0002}', "");
Expect(1, 3888, '\p{^Is_Nv=00000011/0002}', "");
Expect(1, 3888, '\P{Is_Nv=00000011/0002}', "");
Expect(0, 3888, '\P{^Is_Nv=00000011/0002}', "");
Expect(1, 3887, '\p{Is_Nv=5.500}', "");
Expect(0, 3887, '\p{^Is_Nv=5.500}', "");
Expect(0, 3887, '\P{Is_Nv=5.500}', "");
Expect(1, 3887, '\P{^Is_Nv=5.500}', "");
Expect(0, 3888, '\p{Is_Nv=5.500}', "");
Expect(1, 3888, '\p{^Is_Nv=5.500}', "");
Expect(1, 3888, '\P{Is_Nv=5.500}', "");
Expect(0, 3888, '\P{^Is_Nv=5.500}', "");
Error('\p{Numeric_Value=	-+0_0_0_0_0_12:=}');
Error('\P{Numeric_Value=	-+0_0_0_0_0_12:=}');
Expect(1, 9452, '\p{Numeric_Value=000000012}', "");
Expect(0, 9452, '\p{^Numeric_Value=000000012}', "");
Expect(0, 9452, '\P{Numeric_Value=000000012}', "");
Expect(1, 9452, '\P{^Numeric_Value=000000012}', "");
Expect(0, 9453, '\p{Numeric_Value=000000012}', "");
Expect(1, 9453, '\p{^Numeric_Value=000000012}', "");
Expect(1, 9453, '\P{Numeric_Value=000000012}', "");
Expect(0, 9453, '\P{^Numeric_Value=000000012}', "");
Error('\p{Nv=_+001_2/a/}');
Error('\P{Nv=_+001_2/a/}');
Expect(1, 9452, '\p{Nv=+0000012}', "");
Expect(0, 9452, '\p{^Nv=+0000012}', "");
Expect(0, 9452, '\P{Nv=+0000012}', "");
Expect(1, 9452, '\P{^Nv=+0000012}', "");
Expect(0, 9453, '\p{Nv=+0000012}', "");
Expect(1, 9453, '\p{^Nv=+0000012}', "");
Expect(1, 9453, '\P{Nv=+0000012}', "");
Expect(0, 9453, '\P{^Nv=+0000012}', "");
Error('\p{Is_Numeric_Value=:=	 0_0_0_0_0_0_0_12}');
Error('\P{Is_Numeric_Value=:=	 0_0_0_0_0_0_0_12}');
Expect(1, 9452, '\p{Is_Numeric_Value=+00_00_00_00_01_2}', "");
Expect(0, 9452, '\p{^Is_Numeric_Value=+00_00_00_00_01_2}', "");
Expect(0, 9452, '\P{Is_Numeric_Value=+00_00_00_00_01_2}', "");
Expect(1, 9452, '\P{^Is_Numeric_Value=+00_00_00_00_01_2}', "");
Expect(0, 9453, '\p{Is_Numeric_Value=+00_00_00_00_01_2}', "");
Expect(1, 9453, '\p{^Is_Numeric_Value=+00_00_00_00_01_2}', "");
Expect(1, 9453, '\P{Is_Numeric_Value=+00_00_00_00_01_2}', "");
Expect(0, 9453, '\P{^Is_Numeric_Value=+00_00_00_00_01_2}', "");
Error('\p{Is_Nv=- 00_00_00_12/a/}');
Error('\P{Is_Nv=- 00_00_00_12/a/}');
Expect(1, 9452, '\p{Is_Nv=+000000012}', "");
Expect(0, 9452, '\p{^Is_Nv=+000000012}', "");
Expect(0, 9452, '\P{Is_Nv=+000000012}', "");
Expect(1, 9452, '\P{^Is_Nv=+000000012}', "");
Expect(0, 9453, '\p{Is_Nv=+000000012}', "");
Expect(1, 9453, '\p{^Is_Nv=+000000012}', "");
Expect(1, 9453, '\P{Is_Nv=+000000012}', "");
Expect(0, 9453, '\P{^Is_Nv=+000000012}', "");
Error('\p{Numeric_Value=_:=013}');
Error('\P{Numeric_Value=_:=013}');
Expect(1, 9453, '\p{Numeric_Value=00000001_3}', "");
Expect(0, 9453, '\p{^Numeric_Value=00000001_3}', "");
Expect(0, 9453, '\P{Numeric_Value=00000001_3}', "");
Expect(1, 9453, '\P{^Numeric_Value=00000001_3}', "");
Expect(0, 9454, '\p{Numeric_Value=00000001_3}', "");
Expect(1, 9454, '\p{^Numeric_Value=00000001_3}', "");
Expect(1, 9454, '\P{Numeric_Value=00000001_3}', "");
Expect(0, 9454, '\P{^Numeric_Value=00000001_3}', "");
Error('\p{Nv=:=	01_3}');
Error('\P{Nv=:=	01_3}');
Expect(1, 9453, '\p{Nv=+01_3}', "");
Expect(0, 9453, '\p{^Nv=+01_3}', "");
Expect(0, 9453, '\P{Nv=+01_3}', "");
Expect(1, 9453, '\P{^Nv=+01_3}', "");
Expect(0, 9454, '\p{Nv=+01_3}', "");
Expect(1, 9454, '\p{^Nv=+01_3}', "");
Expect(1, 9454, '\P{Nv=+01_3}', "");
Expect(0, 9454, '\P{^Nv=+01_3}', "");
Error('\p{Is_Numeric_Value=_/a/000013}');
Error('\P{Is_Numeric_Value=_/a/000013}');
Expect(1, 9453, '\p{Is_Numeric_Value=000000001_3}', "");
Expect(0, 9453, '\p{^Is_Numeric_Value=000000001_3}', "");
Expect(0, 9453, '\P{Is_Numeric_Value=000000001_3}', "");
Expect(1, 9453, '\P{^Is_Numeric_Value=000000001_3}', "");
Expect(0, 9454, '\p{Is_Numeric_Value=000000001_3}', "");
Expect(1, 9454, '\p{^Is_Numeric_Value=000000001_3}', "");
Expect(1, 9454, '\P{Is_Numeric_Value=000000001_3}', "");
Expect(0, 9454, '\P{^Is_Numeric_Value=000000001_3}', "");
Error('\p{Is_Nv=_-+0001_3:=}');
Error('\P{Is_Nv=_-+0001_3:=}');
Expect(1, 9453, '\p{Is_Nv=+0000_0001_3}', "");
Expect(0, 9453, '\p{^Is_Nv=+0000_0001_3}', "");
Expect(0, 9453, '\P{Is_Nv=+0000_0001_3}', "");
Expect(1, 9453, '\P{^Is_Nv=+0000_0001_3}', "");
Expect(0, 9454, '\p{Is_Nv=+0000_0001_3}', "");
Expect(1, 9454, '\p{^Is_Nv=+0000_0001_3}', "");
Expect(1, 9454, '\P{Is_Nv=+0000_0001_3}', "");
Expect(0, 9454, '\P{^Is_Nv=+0000_0001_3}', "");
Error('\p{Numeric_Value=-/a/000000013/00002}');
Error('\P{Numeric_Value=-/a/000000013/00002}');
Expect(1, 3888, '\p{Numeric_Value:	0000000013/002}', "");
Expect(0, 3888, '\p{^Numeric_Value:	0000000013/002}', "");
Expect(0, 3888, '\P{Numeric_Value:	0000000013/002}', "");
Expect(1, 3888, '\P{^Numeric_Value:	0000000013/002}', "");
Expect(0, 3889, '\p{Numeric_Value:	0000000013/002}', "");
Expect(1, 3889, '\p{^Numeric_Value:	0000000013/002}', "");
Expect(1, 3889, '\P{Numeric_Value:	0000000013/002}', "");
Expect(0, 3889, '\P{^Numeric_Value:	0000000013/002}', "");
Expect(1, 3888, '\p{Numeric_Value=6.500}', "");
Expect(0, 3888, '\p{^Numeric_Value=6.500}', "");
Expect(0, 3888, '\P{Numeric_Value=6.500}', "");
Expect(1, 3888, '\P{^Numeric_Value=6.500}', "");
Expect(0, 3889, '\p{Numeric_Value=6.500}', "");
Expect(1, 3889, '\p{^Numeric_Value=6.500}', "");
Expect(1, 3889, '\P{Numeric_Value=6.500}', "");
Expect(0, 3889, '\P{^Numeric_Value=6.500}', "");
Error('\p{Nv=	000013/002/a/}');
Error('\P{Nv=	000013/002/a/}');
Expect(1, 3888, '\p{Nv=013/00000002}', "");
Expect(0, 3888, '\p{^Nv=013/00000002}', "");
Expect(0, 3888, '\P{Nv=013/00000002}', "");
Expect(1, 3888, '\P{^Nv=013/00000002}', "");
Expect(0, 3889, '\p{Nv=013/00000002}', "");
Expect(1, 3889, '\p{^Nv=013/00000002}', "");
Expect(1, 3889, '\P{Nv=013/00000002}', "");
Expect(0, 3889, '\P{^Nv=013/00000002}', "");
Expect(1, 3888, '\p{Nv=6.500}', "");
Expect(0, 3888, '\p{^Nv=6.500}', "");
Expect(0, 3888, '\P{Nv=6.500}', "");
Expect(1, 3888, '\P{^Nv=6.500}', "");
Expect(0, 3889, '\p{Nv=6.500}', "");
Expect(1, 3889, '\p{^Nv=6.500}', "");
Expect(1, 3889, '\P{Nv=6.500}', "");
Expect(0, 3889, '\P{^Nv=6.500}', "");
Error('\p{Is_Numeric_Value=:=  +000013/00002}');
Error('\P{Is_Numeric_Value=:=  +000013/00002}');
Expect(1, 3888, '\p{Is_Numeric_Value=0013/2}', "");
Expect(0, 3888, '\p{^Is_Numeric_Value=0013/2}', "");
Expect(0, 3888, '\P{Is_Numeric_Value=0013/2}', "");
Expect(1, 3888, '\P{^Is_Numeric_Value=0013/2}', "");
Expect(0, 3889, '\p{Is_Numeric_Value=0013/2}', "");
Expect(1, 3889, '\p{^Is_Numeric_Value=0013/2}', "");
Expect(1, 3889, '\P{Is_Numeric_Value=0013/2}', "");
Expect(0, 3889, '\P{^Is_Numeric_Value=0013/2}', "");
Expect(1, 3888, '\p{Is_Numeric_Value=6.500}', "");
Expect(0, 3888, '\p{^Is_Numeric_Value=6.500}', "");
Expect(0, 3888, '\P{Is_Numeric_Value=6.500}', "");
Expect(1, 3888, '\P{^Is_Numeric_Value=6.500}', "");
Expect(0, 3889, '\p{Is_Numeric_Value=6.500}', "");
Expect(1, 3889, '\p{^Is_Numeric_Value=6.500}', "");
Expect(1, 3889, '\P{Is_Numeric_Value=6.500}', "");
Expect(0, 3889, '\P{^Is_Numeric_Value=6.500}', "");
Error('\p{Is_Nv:   :=_ 13/0002}');
Error('\P{Is_Nv:   :=_ 13/0002}');
Expect(1, 3888, '\p{Is_Nv=0013/02}', "");
Expect(0, 3888, '\p{^Is_Nv=0013/02}', "");
Expect(0, 3888, '\P{Is_Nv=0013/02}', "");
Expect(1, 3888, '\P{^Is_Nv=0013/02}', "");
Expect(0, 3889, '\p{Is_Nv=0013/02}', "");
Expect(1, 3889, '\p{^Is_Nv=0013/02}', "");
Expect(1, 3889, '\P{Is_Nv=0013/02}', "");
Expect(0, 3889, '\P{^Is_Nv=0013/02}', "");
Expect(1, 3888, '\p{Is_Nv=6.500}', "");
Expect(0, 3888, '\p{^Is_Nv=6.500}', "");
Expect(0, 3888, '\P{Is_Nv=6.500}', "");
Expect(1, 3888, '\P{^Is_Nv=6.500}', "");
Expect(0, 3889, '\p{Is_Nv=6.500}', "");
Expect(1, 3889, '\p{^Is_Nv=6.500}', "");
Expect(1, 3889, '\P{Is_Nv=6.500}', "");
Expect(0, 3889, '\P{^Is_Nv=6.500}', "");
Error('\p{Numeric_Value:   :=--0_0_0_0_0_0_00014}');
Error('\P{Numeric_Value:   :=--0_0_0_0_0_0_00014}');
Expect(1, 9454, '\p{Numeric_Value=000000001_4}', "");
Expect(0, 9454, '\p{^Numeric_Value=000000001_4}', "");
Expect(0, 9454, '\P{Numeric_Value=000000001_4}', "");
Expect(1, 9454, '\P{^Numeric_Value=000000001_4}', "");
Expect(0, 9455, '\p{Numeric_Value=000000001_4}', "");
Expect(1, 9455, '\p{^Numeric_Value=000000001_4}', "");
Expect(1, 9455, '\P{Numeric_Value=000000001_4}', "");
Expect(0, 9455, '\P{^Numeric_Value=000000001_4}', "");
Error('\p{Nv:   /a/-	00000014}');
Error('\P{Nv:   /a/-	00000014}');
Expect(1, 9454, '\p{Nv=+14}', "");
Expect(0, 9454, '\p{^Nv=+14}', "");
Expect(0, 9454, '\P{Nv=+14}', "");
Expect(1, 9454, '\P{^Nv=+14}', "");
Expect(0, 9455, '\p{Nv=+14}', "");
Expect(1, 9455, '\p{^Nv=+14}', "");
Expect(1, 9455, '\P{Nv=+14}', "");
Expect(0, 9455, '\P{^Nv=+14}', "");
Error('\p{Is_Numeric_Value=/a/-	+0000_0000_14}');
Error('\P{Is_Numeric_Value=/a/-	+0000_0000_14}');
Expect(1, 9454, '\p{Is_Numeric_Value=+01_4}', "");
Expect(0, 9454, '\p{^Is_Numeric_Value=+01_4}', "");
Expect(0, 9454, '\P{Is_Numeric_Value=+01_4}', "");
Expect(1, 9454, '\P{^Is_Numeric_Value=+01_4}', "");
Expect(0, 9455, '\p{Is_Numeric_Value=+01_4}', "");
Expect(1, 9455, '\p{^Is_Numeric_Value=+01_4}', "");
Expect(1, 9455, '\P{Is_Numeric_Value=+01_4}', "");
Expect(0, 9455, '\P{^Is_Numeric_Value=+01_4}', "");
Error('\p{Is_Nv=:=	014}');
Error('\P{Is_Nv=:=	014}');
Expect(1, 9454, '\p{Is_Nv=014}', "");
Expect(0, 9454, '\p{^Is_Nv=014}', "");
Expect(0, 9454, '\P{Is_Nv=014}', "");
Expect(1, 9454, '\P{^Is_Nv=014}', "");
Expect(0, 9455, '\p{Is_Nv=014}', "");
Expect(1, 9455, '\p{^Is_Nv=014}', "");
Expect(1, 9455, '\P{Is_Nv=014}', "");
Expect(0, 9455, '\P{^Is_Nv=014}', "");
Error('\p{Numeric_Value=:=--0000015}');
Error('\P{Numeric_Value=:=--0000015}');
Expect(1, 9455, '\p{Numeric_Value=00015}', "");
Expect(0, 9455, '\p{^Numeric_Value=00015}', "");
Expect(0, 9455, '\P{Numeric_Value=00015}', "");
Expect(1, 9455, '\P{^Numeric_Value=00015}', "");
Expect(0, 9456, '\p{Numeric_Value=00015}', "");
Expect(1, 9456, '\p{^Numeric_Value=00015}', "");
Expect(1, 9456, '\P{Numeric_Value=00015}', "");
Expect(0, 9456, '\P{^Numeric_Value=00015}', "");
Error('\p{Nv=:=_-+15}');
Error('\P{Nv=:=_-+15}');
Expect(1, 9455, '\p{Nv=+0000015}', "");
Expect(0, 9455, '\p{^Nv=+0000015}', "");
Expect(0, 9455, '\P{Nv=+0000015}', "");
Expect(1, 9455, '\P{^Nv=+0000015}', "");
Expect(0, 9456, '\p{Nv=+0000015}', "");
Expect(1, 9456, '\p{^Nv=+0000015}', "");
Expect(1, 9456, '\P{Nv=+0000015}', "");
Expect(0, 9456, '\P{^Nv=+0000015}', "");
Error('\p{Is_Numeric_Value:   	/a/+015}');
Error('\P{Is_Numeric_Value:   	/a/+015}');
Expect(1, 9455, '\p{Is_Numeric_Value=000000015}', "");
Expect(0, 9455, '\p{^Is_Numeric_Value=000000015}', "");
Expect(0, 9455, '\P{Is_Numeric_Value=000000015}', "");
Expect(1, 9455, '\P{^Is_Numeric_Value=000000015}', "");
Expect(0, 9456, '\p{Is_Numeric_Value=000000015}', "");
Expect(1, 9456, '\p{^Is_Numeric_Value=000000015}', "");
Expect(1, 9456, '\P{Is_Numeric_Value=000000015}', "");
Expect(0, 9456, '\P{^Is_Numeric_Value=000000015}', "");
Error('\p{Is_Nv=- 00000000015/a/}');
Error('\P{Is_Nv=- 00000000015/a/}');
Expect(1, 9455, '\p{Is_Nv=000000015}', "");
Expect(0, 9455, '\p{^Is_Nv=000000015}', "");
Expect(0, 9455, '\P{Is_Nv=000000015}', "");
Expect(1, 9455, '\P{^Is_Nv=000000015}', "");
Expect(0, 9456, '\p{Is_Nv=000000015}', "");
Expect(1, 9456, '\p{^Is_Nv=000000015}', "");
Expect(1, 9456, '\P{Is_Nv=000000015}', "");
Expect(0, 9456, '\P{^Is_Nv=000000015}', "");
Error('\p{Numeric_Value=:=-	000015/0000002}');
Error('\P{Numeric_Value=:=-	000015/0000002}');
Expect(1, 3889, '\p{Numeric_Value=15/2}', "");
Expect(0, 3889, '\p{^Numeric_Value=15/2}', "");
Expect(0, 3889, '\P{Numeric_Value=15/2}', "");
Expect(1, 3889, '\P{^Numeric_Value=15/2}', "");
Expect(0, 3890, '\p{Numeric_Value=15/2}', "");
Expect(1, 3890, '\p{^Numeric_Value=15/2}', "");
Expect(1, 3890, '\P{Numeric_Value=15/2}', "");
Expect(0, 3890, '\P{^Numeric_Value=15/2}', "");
Expect(1, 3889, '\p{Numeric_Value=7.500}', "");
Expect(0, 3889, '\p{^Numeric_Value=7.500}', "");
Expect(0, 3889, '\P{Numeric_Value=7.500}', "");
Expect(1, 3889, '\P{^Numeric_Value=7.500}', "");
Expect(0, 3890, '\p{Numeric_Value=7.500}', "");
Expect(1, 3890, '\p{^Numeric_Value=7.500}', "");
Expect(1, 3890, '\P{Numeric_Value=7.500}', "");
Expect(0, 3890, '\P{^Numeric_Value=7.500}', "");
Error('\p{Nv=-15/002/a/}');
Error('\P{Nv=-15/002/a/}');
Expect(1, 3889, '\p{Nv=00015/02}', "");
Expect(0, 3889, '\p{^Nv=00015/02}', "");
Expect(0, 3889, '\P{Nv=00015/02}', "");
Expect(1, 3889, '\P{^Nv=00015/02}', "");
Expect(0, 3890, '\p{Nv=00015/02}', "");
Expect(1, 3890, '\p{^Nv=00015/02}', "");
Expect(1, 3890, '\P{Nv=00015/02}', "");
Expect(0, 3890, '\P{^Nv=00015/02}', "");
Expect(1, 3889, '\p{Nv=7.500}', "");
Expect(0, 3889, '\p{^Nv=7.500}', "");
Expect(0, 3889, '\P{Nv=7.500}', "");
Expect(1, 3889, '\P{^Nv=7.500}', "");
Expect(0, 3890, '\p{Nv=7.500}', "");
Expect(1, 3890, '\p{^Nv=7.500}', "");
Expect(1, 3890, '\P{Nv=7.500}', "");
Expect(0, 3890, '\P{^Nv=7.500}', "");
Error('\p{Is_Numeric_Value=:= +00015/00002}');
Error('\P{Is_Numeric_Value=:= +00015/00002}');
Expect(1, 3889, '\p{Is_Numeric_Value=0015/00002}', "");
Expect(0, 3889, '\p{^Is_Numeric_Value=0015/00002}', "");
Expect(0, 3889, '\P{Is_Numeric_Value=0015/00002}', "");
Expect(1, 3889, '\P{^Is_Numeric_Value=0015/00002}', "");
Expect(0, 3890, '\p{Is_Numeric_Value=0015/00002}', "");
Expect(1, 3890, '\p{^Is_Numeric_Value=0015/00002}', "");
Expect(1, 3890, '\P{Is_Numeric_Value=0015/00002}', "");
Expect(0, 3890, '\P{^Is_Numeric_Value=0015/00002}', "");
Expect(1, 3889, '\p{Is_Numeric_Value=7.500}', "");
Expect(0, 3889, '\p{^Is_Numeric_Value=7.500}', "");
Expect(0, 3889, '\P{Is_Numeric_Value=7.500}', "");
Expect(1, 3889, '\P{^Is_Numeric_Value=7.500}', "");
Expect(0, 3890, '\p{Is_Numeric_Value=7.500}', "");
Expect(1, 3890, '\p{^Is_Numeric_Value=7.500}', "");
Expect(1, 3890, '\P{Is_Numeric_Value=7.500}', "");
Expect(0, 3890, '\P{^Is_Numeric_Value=7.500}', "");
Error('\p{Is_Nv:   __015/0000000002/a/}');
Error('\P{Is_Nv:   __015/0000000002/a/}');
Expect(1, 3889, '\p{Is_Nv=+15/0000000002}', "");
Expect(0, 3889, '\p{^Is_Nv=+15/0000000002}', "");
Expect(0, 3889, '\P{Is_Nv=+15/0000000002}', "");
Expect(1, 3889, '\P{^Is_Nv=+15/0000000002}', "");
Expect(0, 3890, '\p{Is_Nv=+15/0000000002}', "");
Expect(1, 3890, '\p{^Is_Nv=+15/0000000002}', "");
Expect(1, 3890, '\P{Is_Nv=+15/0000000002}', "");
Expect(0, 3890, '\P{^Is_Nv=+15/0000000002}', "");
Expect(1, 3889, '\p{Is_Nv=7.500}', "");
Expect(0, 3889, '\p{^Is_Nv=7.500}', "");
Expect(0, 3889, '\P{Is_Nv=7.500}', "");
Expect(1, 3889, '\P{^Is_Nv=7.500}', "");
Expect(0, 3890, '\p{Is_Nv=7.500}', "");
Expect(1, 3890, '\p{^Is_Nv=7.500}', "");
Expect(1, 3890, '\P{Is_Nv=7.500}', "");
Expect(0, 3890, '\P{^Is_Nv=7.500}', "");
Error('\p{Numeric_Value=	/a/+00_00_00_16}');
Error('\P{Numeric_Value=	/a/+00_00_00_16}');
Expect(1, 9456, '\p{Numeric_Value=1_6}', "");
Expect(0, 9456, '\p{^Numeric_Value=1_6}', "");
Expect(0, 9456, '\P{Numeric_Value=1_6}', "");
Expect(1, 9456, '\P{^Numeric_Value=1_6}', "");
Expect(0, 9457, '\p{Numeric_Value=1_6}', "");
Expect(1, 9457, '\p{^Numeric_Value=1_6}', "");
Expect(1, 9457, '\P{Numeric_Value=1_6}', "");
Expect(0, 9457, '\P{^Numeric_Value=1_6}', "");
Error('\p{Nv=	:=0_0_0_0_0_0_00016}');
Error('\P{Nv=	:=0_0_0_0_0_0_00016}');
Expect(1, 9456, '\p{Nv=+001_6}', "");
Expect(0, 9456, '\p{^Nv=+001_6}', "");
Expect(0, 9456, '\P{Nv=+001_6}', "");
Expect(1, 9456, '\P{^Nv=+001_6}', "");
Expect(0, 9457, '\p{Nv=+001_6}', "");
Expect(1, 9457, '\p{^Nv=+001_6}', "");
Expect(1, 9457, '\P{Nv=+001_6}', "");
Expect(0, 9457, '\P{^Nv=+001_6}', "");
Error('\p{Is_Numeric_Value=:=-_+0_0_0_0_0_0016}');
Error('\P{Is_Numeric_Value=:=-_+0_0_0_0_0_0016}');
Expect(1, 9456, '\p{Is_Numeric_Value=0_0_0_0_0_0_0_016}', "");
Expect(0, 9456, '\p{^Is_Numeric_Value=0_0_0_0_0_0_0_016}', "");
Expect(0, 9456, '\P{Is_Numeric_Value=0_0_0_0_0_0_0_016}', "");
Expect(1, 9456, '\P{^Is_Numeric_Value=0_0_0_0_0_0_0_016}', "");
Expect(0, 9457, '\p{Is_Numeric_Value=0_0_0_0_0_0_0_016}', "");
Expect(1, 9457, '\p{^Is_Numeric_Value=0_0_0_0_0_0_0_016}', "");
Expect(1, 9457, '\P{Is_Numeric_Value=0_0_0_0_0_0_0_016}', "");
Expect(0, 9457, '\P{^Is_Numeric_Value=0_0_0_0_0_0_0_016}', "");
Error('\p{Is_Nv=-_+00_00_00_01_6:=}');
Error('\P{Is_Nv=-_+00_00_00_01_6:=}');
Expect(1, 9456, '\p{Is_Nv=00000000016}', "");
Expect(0, 9456, '\p{^Is_Nv=00000000016}', "");
Expect(0, 9456, '\P{Is_Nv=00000000016}', "");
Expect(1, 9456, '\P{^Is_Nv=00000000016}', "");
Expect(0, 9457, '\p{Is_Nv=00000000016}', "");
Expect(1, 9457, '\p{^Is_Nv=00000000016}', "");
Expect(1, 9457, '\P{Is_Nv=00000000016}', "");
Expect(0, 9457, '\P{^Is_Nv=00000000016}', "");
Error('\p{Numeric_Value=:= -1_7}');
Error('\P{Numeric_Value=:= -1_7}');
Expect(1, 9457, '\p{Numeric_Value=+0017}', "");
Expect(0, 9457, '\p{^Numeric_Value=+0017}', "");
Expect(0, 9457, '\P{Numeric_Value=+0017}', "");
Expect(1, 9457, '\P{^Numeric_Value=+0017}', "");
Expect(0, 9458, '\p{Numeric_Value=+0017}', "");
Expect(1, 9458, '\p{^Numeric_Value=+0017}', "");
Expect(1, 9458, '\P{Numeric_Value=+0017}', "");
Expect(0, 9458, '\P{^Numeric_Value=+0017}', "");
Error('\p{Nv= _+0_0_0_17:=}');
Error('\P{Nv= _+0_0_0_17:=}');
Expect(1, 9457, '\p{Nv=+00_00_17}', "");
Expect(0, 9457, '\p{^Nv=+00_00_17}', "");
Expect(0, 9457, '\P{Nv=+00_00_17}', "");
Expect(1, 9457, '\P{^Nv=+00_00_17}', "");
Expect(0, 9458, '\p{Nv=+00_00_17}', "");
Expect(1, 9458, '\p{^Nv=+00_00_17}', "");
Expect(1, 9458, '\P{Nv=+00_00_17}', "");
Expect(0, 9458, '\P{^Nv=+00_00_17}', "");
Error('\p{Is_Numeric_Value=	_0001_7:=}');
Error('\P{Is_Numeric_Value=	_0001_7:=}');
Expect(1, 9457, '\p{Is_Numeric_Value=0017}', "");
Expect(0, 9457, '\p{^Is_Numeric_Value=0017}', "");
Expect(0, 9457, '\P{Is_Numeric_Value=0017}', "");
Expect(1, 9457, '\P{^Is_Numeric_Value=0017}', "");
Expect(0, 9458, '\p{Is_Numeric_Value=0017}', "");
Expect(1, 9458, '\p{^Is_Numeric_Value=0017}', "");
Expect(1, 9458, '\P{Is_Numeric_Value=0017}', "");
Expect(0, 9458, '\P{^Is_Numeric_Value=0017}', "");
Error('\p{Is_Nv=  001_7:=}');
Error('\P{Is_Nv=  001_7:=}');
Expect(1, 9457, '\p{Is_Nv=+1_7}', "");
Expect(0, 9457, '\p{^Is_Nv=+1_7}', "");
Expect(0, 9457, '\P{Is_Nv=+1_7}', "");
Expect(1, 9457, '\P{^Is_Nv=+1_7}', "");
Expect(0, 9458, '\p{Is_Nv=+1_7}', "");
Expect(1, 9458, '\p{^Is_Nv=+1_7}', "");
Expect(1, 9458, '\P{Is_Nv=+1_7}', "");
Expect(0, 9458, '\P{^Is_Nv=+1_7}', "");
Error('\p{Numeric_Value=	-0000000017/02/a/}');
Error('\P{Numeric_Value=	-0000000017/02/a/}');
Expect(1, 3890, '\p{Numeric_Value=+000000017/002}', "");
Expect(0, 3890, '\p{^Numeric_Value=+000000017/002}', "");
Expect(0, 3890, '\P{Numeric_Value=+000000017/002}', "");
Expect(1, 3890, '\P{^Numeric_Value=+000000017/002}', "");
Expect(0, 3891, '\p{Numeric_Value=+000000017/002}', "");
Expect(1, 3891, '\p{^Numeric_Value=+000000017/002}', "");
Expect(1, 3891, '\P{Numeric_Value=+000000017/002}', "");
Expect(0, 3891, '\P{^Numeric_Value=+000000017/002}', "");
Expect(1, 3890, '\p{Numeric_Value=8.500}', "");
Expect(0, 3890, '\p{^Numeric_Value=8.500}', "");
Expect(0, 3890, '\P{Numeric_Value=8.500}', "");
Expect(1, 3890, '\P{^Numeric_Value=8.500}', "");
Expect(0, 3891, '\p{Numeric_Value=8.500}', "");
Expect(1, 3891, '\p{^Numeric_Value=8.500}', "");
Expect(1, 3891, '\P{Numeric_Value=8.500}', "");
Expect(0, 3891, '\P{^Numeric_Value=8.500}', "");
Error('\p{Nv=	 000017/00000002/a/}');
Error('\P{Nv=	 000017/00000002/a/}');
Expect(1, 3890, '\p{Nv: 0000000017/0000002}', "");
Expect(0, 3890, '\p{^Nv: 0000000017/0000002}', "");
Expect(0, 3890, '\P{Nv: 0000000017/0000002}', "");
Expect(1, 3890, '\P{^Nv: 0000000017/0000002}', "");
Expect(0, 3891, '\p{Nv: 0000000017/0000002}', "");
Expect(1, 3891, '\p{^Nv: 0000000017/0000002}', "");
Expect(1, 3891, '\P{Nv: 0000000017/0000002}', "");
Expect(0, 3891, '\P{^Nv: 0000000017/0000002}', "");
Expect(1, 3890, '\p{Nv=8.500}', "");
Expect(0, 3890, '\p{^Nv=8.500}', "");
Expect(0, 3890, '\P{Nv=8.500}', "");
Expect(1, 3890, '\P{^Nv=8.500}', "");
Expect(0, 3891, '\p{Nv=8.500}', "");
Expect(1, 3891, '\p{^Nv=8.500}', "");
Expect(1, 3891, '\P{Nv=8.500}', "");
Expect(0, 3891, '\P{^Nv=8.500}', "");
Error('\p{Is_Numeric_Value= 00017/00000002:=}');
Error('\P{Is_Numeric_Value= 00017/00000002:=}');
Expect(1, 3890, '\p{Is_Numeric_Value=017/00000002}', "");
Expect(0, 3890, '\p{^Is_Numeric_Value=017/00000002}', "");
Expect(0, 3890, '\P{Is_Numeric_Value=017/00000002}', "");
Expect(1, 3890, '\P{^Is_Numeric_Value=017/00000002}', "");
Expect(0, 3891, '\p{Is_Numeric_Value=017/00000002}', "");
Expect(1, 3891, '\p{^Is_Numeric_Value=017/00000002}', "");
Expect(1, 3891, '\P{Is_Numeric_Value=017/00000002}', "");
Expect(0, 3891, '\P{^Is_Numeric_Value=017/00000002}', "");
Expect(1, 3890, '\p{Is_Numeric_Value=8.500}', "");
Expect(0, 3890, '\p{^Is_Numeric_Value=8.500}', "");
Expect(0, 3890, '\P{Is_Numeric_Value=8.500}', "");
Expect(1, 3890, '\P{^Is_Numeric_Value=8.500}', "");
Expect(0, 3891, '\p{Is_Numeric_Value=8.500}', "");
Expect(1, 3891, '\p{^Is_Numeric_Value=8.500}', "");
Expect(1, 3891, '\P{Is_Numeric_Value=8.500}', "");
Expect(0, 3891, '\P{^Is_Numeric_Value=8.500}', "");
Error('\p{Is_Nv:   /a/-_+0000017/2}');
Error('\P{Is_Nv:   /a/-_+0000017/2}');
Expect(1, 3890, '\p{Is_Nv:   000017/0000000002}', "");
Expect(0, 3890, '\p{^Is_Nv:   000017/0000000002}', "");
Expect(0, 3890, '\P{Is_Nv:   000017/0000000002}', "");
Expect(1, 3890, '\P{^Is_Nv:   000017/0000000002}', "");
Expect(0, 3891, '\p{Is_Nv:   000017/0000000002}', "");
Expect(1, 3891, '\p{^Is_Nv:   000017/0000000002}', "");
Expect(1, 3891, '\P{Is_Nv:   000017/0000000002}', "");
Expect(0, 3891, '\P{^Is_Nv:   000017/0000000002}', "");
Expect(1, 3890, '\p{Is_Nv: 8.500}', "");
Expect(0, 3890, '\p{^Is_Nv: 8.500}', "");
Expect(0, 3890, '\P{Is_Nv: 8.500}', "");
Expect(1, 3890, '\P{^Is_Nv: 8.500}', "");
Expect(0, 3891, '\p{Is_Nv: 8.500}', "");
Expect(1, 3891, '\p{^Is_Nv: 8.500}', "");
Expect(1, 3891, '\P{Is_Nv: 8.500}', "");
Expect(0, 3891, '\P{^Is_Nv: 8.500}', "");
Error('\p{Numeric_Value=		0_0_0_018:=}');
Error('\P{Numeric_Value=		0_0_0_018:=}');
Expect(1, 9458, '\p{Numeric_Value=001_8}', "");
Expect(0, 9458, '\p{^Numeric_Value=001_8}', "");
Expect(0, 9458, '\P{Numeric_Value=001_8}', "");
Expect(1, 9458, '\P{^Numeric_Value=001_8}', "");
Expect(0, 9459, '\p{Numeric_Value=001_8}', "");
Expect(1, 9459, '\p{^Numeric_Value=001_8}', "");
Expect(1, 9459, '\P{Numeric_Value=001_8}', "");
Expect(0, 9459, '\P{^Numeric_Value=001_8}', "");
Error('\p{Nv=:=-0_0_0_0_0_0_0_0018}');
Error('\P{Nv=:=-0_0_0_0_0_0_0_0018}');
Expect(1, 9458, '\p{Nv=+00_00_01_8}', "");
Expect(0, 9458, '\p{^Nv=+00_00_01_8}', "");
Expect(0, 9458, '\P{Nv=+00_00_01_8}', "");
Expect(1, 9458, '\P{^Nv=+00_00_01_8}', "");
Expect(0, 9459, '\p{Nv=+00_00_01_8}', "");
Expect(1, 9459, '\p{^Nv=+00_00_01_8}', "");
Expect(1, 9459, '\P{Nv=+00_00_01_8}', "");
Expect(0, 9459, '\P{^Nv=+00_00_01_8}', "");
Error('\p{Is_Numeric_Value=/a/_18}');
Error('\P{Is_Numeric_Value=/a/_18}');
Expect(1, 9458, '\p{Is_Numeric_Value=+00000018}', "");
Expect(0, 9458, '\p{^Is_Numeric_Value=+00000018}', "");
Expect(0, 9458, '\P{Is_Numeric_Value=+00000018}', "");
Expect(1, 9458, '\P{^Is_Numeric_Value=+00000018}', "");
Expect(0, 9459, '\p{Is_Numeric_Value=+00000018}', "");
Expect(1, 9459, '\p{^Is_Numeric_Value=+00000018}', "");
Expect(1, 9459, '\P{Is_Numeric_Value=+00000018}', "");
Expect(0, 9459, '\P{^Is_Numeric_Value=+00000018}', "");
Error('\p{Is_Nv=/a/ _00_00_00_18}');
Error('\P{Is_Nv=/a/ _00_00_00_18}');
Expect(1, 9458, '\p{Is_Nv:   01_8}', "");
Expect(0, 9458, '\p{^Is_Nv:   01_8}', "");
Expect(0, 9458, '\P{Is_Nv:   01_8}', "");
Expect(1, 9458, '\P{^Is_Nv:   01_8}', "");
Expect(0, 9459, '\p{Is_Nv:   01_8}', "");
Expect(1, 9459, '\p{^Is_Nv:   01_8}', "");
Expect(1, 9459, '\P{Is_Nv:   01_8}', "");
Expect(0, 9459, '\P{^Is_Nv:   01_8}', "");
Error('\p{Numeric_Value=_0_0_0_0_0_0019:=}');
Error('\P{Numeric_Value=_0_0_0_0_0_0019:=}');
Expect(1, 9459, '\p{Numeric_Value=+019}', "");
Expect(0, 9459, '\p{^Numeric_Value=+019}', "");
Expect(0, 9459, '\P{Numeric_Value=+019}', "");
Expect(1, 9459, '\P{^Numeric_Value=+019}', "");
Expect(0, 9460, '\p{Numeric_Value=+019}', "");
Expect(1, 9460, '\p{^Numeric_Value=+019}', "");
Expect(1, 9460, '\P{Numeric_Value=+019}', "");
Expect(0, 9460, '\P{^Numeric_Value=+019}', "");
Error('\p{Nv= :=000_000_19}');
Error('\P{Nv= :=000_000_19}');
Expect(1, 9459, '\p{Nv=0019}', "");
Expect(0, 9459, '\p{^Nv=0019}', "");
Expect(0, 9459, '\P{Nv=0019}', "");
Expect(1, 9459, '\P{^Nv=0019}', "");
Expect(0, 9460, '\p{Nv=0019}', "");
Expect(1, 9460, '\p{^Nv=0019}', "");
Expect(1, 9460, '\P{Nv=0019}', "");
Expect(0, 9460, '\P{^Nv=0019}', "");
Error('\p{Is_Numeric_Value= _00001_9/a/}');
Error('\P{Is_Numeric_Value= _00001_9/a/}');
Expect(1, 9459, '\p{Is_Numeric_Value=+000000019}', "");
Expect(0, 9459, '\p{^Is_Numeric_Value=+000000019}', "");
Expect(0, 9459, '\P{Is_Numeric_Value=+000000019}', "");
Expect(1, 9459, '\P{^Is_Numeric_Value=+000000019}', "");
Expect(0, 9460, '\p{Is_Numeric_Value=+000000019}', "");
Expect(1, 9460, '\p{^Is_Numeric_Value=+000000019}', "");
Expect(1, 9460, '\P{Is_Numeric_Value=+000000019}', "");
Expect(0, 9460, '\P{^Is_Numeric_Value=+000000019}', "");
Error('\p{Is_Nv:	_:=0000001_9}');
Error('\P{Is_Nv:	_:=0000001_9}');
Expect(1, 9459, '\p{Is_Nv=019}', "");
Expect(0, 9459, '\p{^Is_Nv=019}', "");
Expect(0, 9459, '\P{Is_Nv=019}', "");
Expect(1, 9459, '\P{^Is_Nv=019}', "");
Expect(0, 9460, '\p{Is_Nv=019}', "");
Expect(1, 9460, '\p{^Is_Nv=019}', "");
Expect(1, 9460, '\P{Is_Nv=019}', "");
Expect(0, 9460, '\P{^Is_Nv=019}', "");
Error('\p{Numeric_Value=_:=02}');
Error('\P{Numeric_Value=_:=02}');
Expect(1, 140176, '\p{Numeric_Value=0_0_02}', "");
Expect(0, 140176, '\p{^Numeric_Value=0_0_02}', "");
Expect(0, 140176, '\P{Numeric_Value=0_0_02}', "");
Expect(1, 140176, '\P{^Numeric_Value=0_0_02}', "");
Expect(0, 140177, '\p{Numeric_Value=0_0_02}', "");
Expect(1, 140177, '\p{^Numeric_Value=0_0_02}', "");
Expect(1, 140177, '\P{Numeric_Value=0_0_02}', "");
Expect(0, 140177, '\P{^Numeric_Value=0_0_02}', "");
Error('\p{Nv=  +0002:=}');
Error('\P{Nv=  +0002:=}');
Expect(1, 140176, '\p{Nv=0_0_0_0_0_0_0002}', "");
Expect(0, 140176, '\p{^Nv=0_0_0_0_0_0_0002}', "");
Expect(0, 140176, '\P{Nv=0_0_0_0_0_0_0002}', "");
Expect(1, 140176, '\P{^Nv=0_0_0_0_0_0_0002}', "");
Expect(0, 140177, '\p{Nv=0_0_0_0_0_0_0002}', "");
Expect(1, 140177, '\p{^Nv=0_0_0_0_0_0_0002}', "");
Expect(1, 140177, '\P{Nv=0_0_0_0_0_0_0002}', "");
Expect(0, 140177, '\P{^Nv=0_0_0_0_0_0_0002}', "");
Error('\p{Is_Numeric_Value=	-000002/a/}');
Error('\P{Is_Numeric_Value=	-000002/a/}');
Expect(1, 140176, '\p{Is_Numeric_Value=+2}', "");
Expect(0, 140176, '\p{^Is_Numeric_Value=+2}', "");
Expect(0, 140176, '\P{Is_Numeric_Value=+2}', "");
Expect(1, 140176, '\P{^Is_Numeric_Value=+2}', "");
Expect(0, 140177, '\p{Is_Numeric_Value=+2}', "");
Expect(1, 140177, '\p{^Is_Numeric_Value=+2}', "");
Expect(1, 140177, '\P{Is_Numeric_Value=+2}', "");
Expect(0, 140177, '\P{^Is_Numeric_Value=+2}', "");
Error('\p{Is_Nv=-/a/0000000002}');
Error('\P{Is_Nv=-/a/0000000002}');
Expect(1, 140176, '\p{Is_Nv=0_0_0_0_0_00002}', "");
Expect(0, 140176, '\p{^Is_Nv=0_0_0_0_0_00002}', "");
Expect(0, 140176, '\P{Is_Nv=0_0_0_0_0_00002}', "");
Expect(1, 140176, '\P{^Is_Nv=0_0_0_0_0_00002}', "");
Expect(0, 140177, '\p{Is_Nv=0_0_0_0_0_00002}', "");
Expect(1, 140177, '\p{^Is_Nv=0_0_0_0_0_00002}', "");
Expect(1, 140177, '\P{Is_Nv=0_0_0_0_0_00002}', "");
Expect(0, 140177, '\P{^Is_Nv=0_0_0_0_0_00002}', "");
Error('\p{Numeric_Value= 0000002/00003:=}');
Error('\P{Numeric_Value= 0000002/00003:=}');
Expect(1, 74846, '\p{Numeric_Value:	000000002/00003}', "");
Expect(0, 74846, '\p{^Numeric_Value:	000000002/00003}', "");
Expect(0, 74846, '\P{Numeric_Value:	000000002/00003}', "");
Expect(1, 74846, '\P{^Numeric_Value:	000000002/00003}', "");
Expect(0, 74847, '\p{Numeric_Value:	000000002/00003}', "");
Expect(1, 74847, '\p{^Numeric_Value:	000000002/00003}', "");
Expect(1, 74847, '\P{Numeric_Value:	000000002/00003}', "");
Expect(0, 74847, '\P{^Numeric_Value:	000000002/00003}', "");
Error('\p{Numeric_Value=0.7}');
Error('\P{Numeric_Value=0.7}');
Error('\p{Numeric_Value=0.67}');
Error('\P{Numeric_Value=0.67}');
Expect(1, 74846, '\p{Numeric_Value=0.667}', "");
Expect(0, 74846, '\p{^Numeric_Value=0.667}', "");
Expect(0, 74846, '\P{Numeric_Value=0.667}', "");
Expect(1, 74846, '\P{^Numeric_Value=0.667}', "");
Expect(0, 74847, '\p{Numeric_Value=0.667}', "");
Expect(1, 74847, '\p{^Numeric_Value=0.667}', "");
Expect(1, 74847, '\P{Numeric_Value=0.667}', "");
Expect(0, 74847, '\P{^Numeric_Value=0.667}', "");
Error('\p{Nv:	+00000002/3/a/}');
Error('\P{Nv:	+00000002/3/a/}');
Expect(1, 74846, '\p{Nv=00000002/000000003}', "");
Expect(0, 74846, '\p{^Nv=00000002/000000003}', "");
Expect(0, 74846, '\P{Nv=00000002/000000003}', "");
Expect(1, 74846, '\P{^Nv=00000002/000000003}', "");
Expect(0, 74847, '\p{Nv=00000002/000000003}', "");
Expect(1, 74847, '\p{^Nv=00000002/000000003}', "");
Expect(1, 74847, '\P{Nv=00000002/000000003}', "");
Expect(0, 74847, '\P{^Nv=00000002/000000003}', "");
Error('\p{Nv: 0.7}');
Error('\P{Nv: 0.7}');
Error('\p{Nv=0.67}');
Error('\P{Nv=0.67}');
Expect(1, 74846, '\p{Nv:	0.667}', "");
Expect(0, 74846, '\p{^Nv:	0.667}', "");
Expect(0, 74846, '\P{Nv:	0.667}', "");
Expect(1, 74846, '\P{^Nv:	0.667}', "");
Expect(0, 74847, '\p{Nv:	0.667}', "");
Expect(1, 74847, '\p{^Nv:	0.667}', "");
Expect(1, 74847, '\P{Nv:	0.667}', "");
Expect(0, 74847, '\P{^Nv:	0.667}', "");
Error('\p{Is_Numeric_Value=/a/  0000000002/3}');
Error('\P{Is_Numeric_Value=/a/  0000000002/3}');
Expect(1, 74846, '\p{Is_Numeric_Value=+00000002/0000000003}', "");
Expect(0, 74846, '\p{^Is_Numeric_Value=+00000002/0000000003}', "");
Expect(0, 74846, '\P{Is_Numeric_Value=+00000002/0000000003}', "");
Expect(1, 74846, '\P{^Is_Numeric_Value=+00000002/0000000003}', "");
Expect(0, 74847, '\p{Is_Numeric_Value=+00000002/0000000003}', "");
Expect(1, 74847, '\p{^Is_Numeric_Value=+00000002/0000000003}', "");
Expect(1, 74847, '\P{Is_Numeric_Value=+00000002/0000000003}', "");
Expect(0, 74847, '\P{^Is_Numeric_Value=+00000002/0000000003}', "");
Error('\p{Is_Numeric_Value=0.7}');
Error('\P{Is_Numeric_Value=0.7}');
Error('\p{Is_Numeric_Value=0.67}');
Error('\P{Is_Numeric_Value=0.67}');
Expect(1, 74846, '\p{Is_Numeric_Value=0.667}', "");
Expect(0, 74846, '\p{^Is_Numeric_Value=0.667}', "");
Expect(0, 74846, '\P{Is_Numeric_Value=0.667}', "");
Expect(1, 74846, '\P{^Is_Numeric_Value=0.667}', "");
Expect(0, 74847, '\p{Is_Numeric_Value=0.667}', "");
Expect(1, 74847, '\p{^Is_Numeric_Value=0.667}', "");
Expect(1, 74847, '\P{Is_Numeric_Value=0.667}', "");
Expect(0, 74847, '\P{^Is_Numeric_Value=0.667}', "");
Error('\p{Is_Nv= /a/+000002/003}');
Error('\P{Is_Nv= /a/+000002/003}');
Expect(1, 74846, '\p{Is_Nv=2/3}', "");
Expect(0, 74846, '\p{^Is_Nv=2/3}', "");
Expect(0, 74846, '\P{Is_Nv=2/3}', "");
Expect(1, 74846, '\P{^Is_Nv=2/3}', "");
Expect(0, 74847, '\p{Is_Nv=2/3}', "");
Expect(1, 74847, '\p{^Is_Nv=2/3}', "");
Expect(1, 74847, '\P{Is_Nv=2/3}', "");
Expect(0, 74847, '\P{^Is_Nv=2/3}', "");
Error('\p{Is_Nv=0.7}');
Error('\P{Is_Nv=0.7}');
Error('\p{Is_Nv=0.67}');
Error('\P{Is_Nv=0.67}');
Expect(1, 74846, '\p{Is_Nv=0.667}', "");
Expect(0, 74846, '\p{^Is_Nv=0.667}', "");
Expect(0, 74846, '\P{Is_Nv=0.667}', "");
Expect(1, 74846, '\P{^Is_Nv=0.667}', "");
Expect(0, 74847, '\p{Is_Nv=0.667}', "");
Expect(1, 74847, '\p{^Is_Nv=0.667}', "");
Expect(1, 74847, '\P{Is_Nv=0.667}', "");
Expect(0, 74847, '\P{^Is_Nv=0.667}', "");
Error('\p{Numeric_Value=:= 000002/5}');
Error('\P{Numeric_Value=:= 000002/5}');
Expect(1, 8534, '\p{Numeric_Value:	+000000002/00000005}', "");
Expect(0, 8534, '\p{^Numeric_Value:	+000000002/00000005}', "");
Expect(0, 8534, '\P{Numeric_Value:	+000000002/00000005}', "");
Expect(1, 8534, '\P{^Numeric_Value:	+000000002/00000005}', "");
Expect(0, 8535, '\p{Numeric_Value:	+000000002/00000005}', "");
Expect(1, 8535, '\p{^Numeric_Value:	+000000002/00000005}', "");
Expect(1, 8535, '\P{Numeric_Value:	+000000002/00000005}', "");
Expect(0, 8535, '\P{^Numeric_Value:	+000000002/00000005}', "");
Expect(1, 8534, '\p{Numeric_Value=0.400}', "");
Expect(0, 8534, '\p{^Numeric_Value=0.400}', "");
Expect(0, 8534, '\P{Numeric_Value=0.400}', "");
Expect(1, 8534, '\P{^Numeric_Value=0.400}', "");
Expect(0, 8535, '\p{Numeric_Value=0.400}', "");
Expect(1, 8535, '\p{^Numeric_Value=0.400}', "");
Expect(1, 8535, '\P{Numeric_Value=0.400}', "");
Expect(0, 8535, '\P{^Numeric_Value=0.400}', "");
Error('\p{Nv=	:=000002/0005}');
Error('\P{Nv=	:=000002/0005}');
Expect(1, 8534, '\p{Nv: 000002/0000000005}', "");
Expect(0, 8534, '\p{^Nv: 000002/0000000005}', "");
Expect(0, 8534, '\P{Nv: 000002/0000000005}', "");
Expect(1, 8534, '\P{^Nv: 000002/0000000005}', "");
Expect(0, 8535, '\p{Nv: 000002/0000000005}', "");
Expect(1, 8535, '\p{^Nv: 000002/0000000005}', "");
Expect(1, 8535, '\P{Nv: 000002/0000000005}', "");
Expect(0, 8535, '\P{^Nv: 000002/0000000005}', "");
Expect(1, 8534, '\p{Nv=0.400}', "");
Expect(0, 8534, '\p{^Nv=0.400}', "");
Expect(0, 8534, '\P{Nv=0.400}', "");
Expect(1, 8534, '\P{^Nv=0.400}', "");
Expect(0, 8535, '\p{Nv=0.400}', "");
Expect(1, 8535, '\p{^Nv=0.400}', "");
Expect(1, 8535, '\P{Nv=0.400}', "");
Expect(0, 8535, '\P{^Nv=0.400}', "");
Error('\p{Is_Numeric_Value: _00002/0005/a/}');
Error('\P{Is_Numeric_Value: _00002/0005/a/}');
Expect(1, 8534, '\p{Is_Numeric_Value=000002/0000000005}', "");
Expect(0, 8534, '\p{^Is_Numeric_Value=000002/0000000005}', "");
Expect(0, 8534, '\P{Is_Numeric_Value=000002/0000000005}', "");
Expect(1, 8534, '\P{^Is_Numeric_Value=000002/0000000005}', "");
Expect(0, 8535, '\p{Is_Numeric_Value=000002/0000000005}', "");
Expect(1, 8535, '\p{^Is_Numeric_Value=000002/0000000005}', "");
Expect(1, 8535, '\P{Is_Numeric_Value=000002/0000000005}', "");
Expect(0, 8535, '\P{^Is_Numeric_Value=000002/0000000005}', "");
Expect(1, 8534, '\p{Is_Numeric_Value=0.400}', "");
Expect(0, 8534, '\p{^Is_Numeric_Value=0.400}', "");
Expect(0, 8534, '\P{Is_Numeric_Value=0.400}', "");
Expect(1, 8534, '\P{^Is_Numeric_Value=0.400}', "");
Expect(0, 8535, '\p{Is_Numeric_Value=0.400}', "");
Expect(1, 8535, '\p{^Is_Numeric_Value=0.400}', "");
Expect(1, 8535, '\P{Is_Numeric_Value=0.400}', "");
Expect(0, 8535, '\P{^Is_Numeric_Value=0.400}', "");
Error('\p{Is_Nv=/a/ _+2/00005}');
Error('\P{Is_Nv=/a/ _+2/00005}');
Expect(1, 8534, '\p{Is_Nv=+0000002/05}', "");
Expect(0, 8534, '\p{^Is_Nv=+0000002/05}', "");
Expect(0, 8534, '\P{Is_Nv=+0000002/05}', "");
Expect(1, 8534, '\P{^Is_Nv=+0000002/05}', "");
Expect(0, 8535, '\p{Is_Nv=+0000002/05}', "");
Expect(1, 8535, '\p{^Is_Nv=+0000002/05}', "");
Expect(1, 8535, '\P{Is_Nv=+0000002/05}', "");
Expect(0, 8535, '\P{^Is_Nv=+0000002/05}', "");
Expect(1, 8534, '\p{Is_Nv=0.400}', "");
Expect(0, 8534, '\p{^Is_Nv=0.400}', "");
Expect(0, 8534, '\P{Is_Nv=0.400}', "");
Expect(1, 8534, '\P{^Is_Nv=0.400}', "");
Expect(0, 8535, '\p{Is_Nv=0.400}', "");
Expect(1, 8535, '\p{^Is_Nv=0.400}', "");
Expect(1, 8535, '\P{Is_Nv=0.400}', "");
Expect(0, 8535, '\P{^Is_Nv=0.400}', "");
Error('\p{Numeric_Value=-/a/0000000020}');
Error('\P{Numeric_Value=-/a/0000000020}');
Expect(1, 119658, '\p{Numeric_Value=002_0}', "");
Expect(0, 119658, '\p{^Numeric_Value=002_0}', "");
Expect(0, 119658, '\P{Numeric_Value=002_0}', "");
Expect(1, 119658, '\P{^Numeric_Value=002_0}', "");
Expect(0, 119659, '\p{Numeric_Value=002_0}', "");
Expect(1, 119659, '\p{^Numeric_Value=002_0}', "");
Expect(1, 119659, '\P{Numeric_Value=002_0}', "");
Expect(0, 119659, '\P{^Numeric_Value=002_0}', "");
Error('\p{Nv=/a/	000020}');
Error('\P{Nv=/a/	000020}');
Expect(1, 119658, '\p{Nv=0_0_0_0_0_00020}', "");
Expect(0, 119658, '\p{^Nv=0_0_0_0_0_00020}', "");
Expect(0, 119658, '\P{Nv=0_0_0_0_0_00020}', "");
Expect(1, 119658, '\P{^Nv=0_0_0_0_0_00020}', "");
Expect(0, 119659, '\p{Nv=0_0_0_0_0_00020}', "");
Expect(1, 119659, '\p{^Nv=0_0_0_0_0_00020}', "");
Expect(1, 119659, '\P{Nv=0_0_0_0_0_00020}', "");
Expect(0, 119659, '\P{^Nv=0_0_0_0_0_00020}', "");
Error('\p{Is_Numeric_Value=-:=+0020}');
Error('\P{Is_Numeric_Value=-:=+0020}');
Expect(1, 119658, '\p{Is_Numeric_Value=0000020}', "");
Expect(0, 119658, '\p{^Is_Numeric_Value=0000020}', "");
Expect(0, 119658, '\P{Is_Numeric_Value=0000020}', "");
Expect(1, 119658, '\P{^Is_Numeric_Value=0000020}', "");
Expect(0, 119659, '\p{Is_Numeric_Value=0000020}', "");
Expect(1, 119659, '\p{^Is_Numeric_Value=0000020}', "");
Expect(1, 119659, '\P{Is_Numeric_Value=0000020}', "");
Expect(0, 119659, '\P{^Is_Numeric_Value=0000020}', "");
Error('\p{Is_Nv=:=_	00_00_00_00_20}');
Error('\P{Is_Nv=:=_	00_00_00_00_20}');
Expect(1, 119658, '\p{Is_Nv: 00_00_00_00_20}', "");
Expect(0, 119658, '\p{^Is_Nv: 00_00_00_00_20}', "");
Expect(0, 119658, '\P{Is_Nv: 00_00_00_00_20}', "");
Expect(1, 119658, '\P{^Is_Nv: 00_00_00_00_20}', "");
Expect(0, 119659, '\p{Is_Nv: 00_00_00_00_20}', "");
Expect(1, 119659, '\p{^Is_Nv: 00_00_00_00_20}', "");
Expect(1, 119659, '\P{Is_Nv: 00_00_00_00_20}', "");
Expect(0, 119659, '\P{^Is_Nv: 00_00_00_00_20}', "");
Error('\p{Numeric_Value=_:=0_0_0_0_0_200}');
Error('\P{Numeric_Value=_:=0_0_0_0_0_200}');
Expect(1, 69235, '\p{Numeric_Value=0_0_0_0_0_0_2_00}', "");
Expect(0, 69235, '\p{^Numeric_Value=0_0_0_0_0_0_2_00}', "");
Expect(0, 69235, '\P{Numeric_Value=0_0_0_0_0_0_2_00}', "");
Expect(1, 69235, '\P{^Numeric_Value=0_0_0_0_0_0_2_00}', "");
Expect(0, 69236, '\p{Numeric_Value=0_0_0_0_0_0_2_00}', "");
Expect(1, 69236, '\p{^Numeric_Value=0_0_0_0_0_0_2_00}', "");
Expect(1, 69236, '\P{Numeric_Value=0_0_0_0_0_0_2_00}', "");
Expect(0, 69236, '\P{^Numeric_Value=0_0_0_0_0_0_2_00}', "");
Error('\p{Nv=	0_0_0_0_0_0_000200/a/}');
Error('\P{Nv=	0_0_0_0_0_0_000200/a/}');
Expect(1, 69235, '\p{Nv: 20_0}', "");
Expect(0, 69235, '\p{^Nv: 20_0}', "");
Expect(0, 69235, '\P{Nv: 20_0}', "");
Expect(1, 69235, '\P{^Nv: 20_0}', "");
Expect(0, 69236, '\p{Nv: 20_0}', "");
Expect(1, 69236, '\p{^Nv: 20_0}', "");
Expect(1, 69236, '\P{Nv: 20_0}', "");
Expect(0, 69236, '\P{^Nv: 20_0}', "");
Error('\p{Is_Numeric_Value=:=  000000200}');
Error('\P{Is_Numeric_Value=:=  000000200}');
Expect(1, 69235, '\p{Is_Numeric_Value=0200}', "");
Expect(0, 69235, '\p{^Is_Numeric_Value=0200}', "");
Expect(0, 69235, '\P{Is_Numeric_Value=0200}', "");
Expect(1, 69235, '\P{^Is_Numeric_Value=0200}', "");
Expect(0, 69236, '\p{Is_Numeric_Value=0200}', "");
Expect(1, 69236, '\p{^Is_Numeric_Value=0200}', "");
Expect(1, 69236, '\P{Is_Numeric_Value=0200}', "");
Expect(0, 69236, '\P{^Is_Numeric_Value=0200}', "");
Error('\p{Is_Nv=	+00_00_00_200/a/}');
Error('\P{Is_Nv=	+00_00_00_200/a/}');
Expect(1, 69235, '\p{Is_Nv=0_0_0_0_2_00}', "");
Expect(0, 69235, '\p{^Is_Nv=0_0_0_0_2_00}', "");
Expect(0, 69235, '\P{Is_Nv=0_0_0_0_2_00}', "");
Expect(1, 69235, '\P{^Is_Nv=0_0_0_0_2_00}', "");
Expect(0, 69236, '\p{Is_Nv=0_0_0_0_2_00}', "");
Expect(1, 69236, '\p{^Is_Nv=0_0_0_0_2_00}', "");
Expect(1, 69236, '\P{Is_Nv=0_0_0_0_2_00}', "");
Expect(0, 69236, '\P{^Is_Nv=0_0_0_0_2_00}', "");
Error('\p{Numeric_Value=:=	 +0_0_0_0_2_0_00}');
Error('\P{Numeric_Value=:=	 +0_0_0_0_2_0_00}');
Expect(1, 65827, '\p{Numeric_Value=0_0_0_0_2000}', "");
Expect(0, 65827, '\p{^Numeric_Value=0_0_0_0_2000}', "");
Expect(0, 65827, '\P{Numeric_Value=0_0_0_0_2000}', "");
Expect(1, 65827, '\P{^Numeric_Value=0_0_0_0_2000}', "");
Expect(0, 65828, '\p{Numeric_Value=0_0_0_0_2000}', "");
Expect(1, 65828, '\p{^Numeric_Value=0_0_0_0_2000}', "");
Expect(1, 65828, '\P{Numeric_Value=0_0_0_0_2000}', "");
Expect(0, 65828, '\P{^Numeric_Value=0_0_0_0_2000}', "");
Error('\p{Nv:	_+00000002000:=}');
Error('\P{Nv:	_+00000002000:=}');
Expect(1, 65827, '\p{Nv=0_0_0_0_2000}', "");
Expect(0, 65827, '\p{^Nv=0_0_0_0_2000}', "");
Expect(0, 65827, '\P{Nv=0_0_0_0_2000}', "");
Expect(1, 65827, '\P{^Nv=0_0_0_0_2000}', "");
Expect(0, 65828, '\p{Nv=0_0_0_0_2000}', "");
Expect(1, 65828, '\p{^Nv=0_0_0_0_2000}', "");
Expect(1, 65828, '\P{Nv=0_0_0_0_2000}', "");
Expect(0, 65828, '\P{^Nv=0_0_0_0_2000}', "");
Error('\p{Is_Numeric_Value=:=	_+2_0_00}');
Error('\P{Is_Numeric_Value=:=	_+2_0_00}');
Expect(1, 65827, '\p{Is_Numeric_Value=+000002000}', "");
Expect(0, 65827, '\p{^Is_Numeric_Value=+000002000}', "");
Expect(0, 65827, '\P{Is_Numeric_Value=+000002000}', "");
Expect(1, 65827, '\P{^Is_Numeric_Value=+000002000}', "");
Expect(0, 65828, '\p{Is_Numeric_Value=+000002000}', "");
Expect(1, 65828, '\p{^Is_Numeric_Value=+000002000}', "");
Expect(1, 65828, '\P{Is_Numeric_Value=+000002000}', "");
Expect(0, 65828, '\P{^Is_Numeric_Value=+000002000}', "");
Error('\p{Is_Nv=__0000000002000:=}');
Error('\P{Is_Nv=__0000000002000:=}');
Expect(1, 65827, '\p{Is_Nv=2000}', "");
Expect(0, 65827, '\p{^Is_Nv=2000}', "");
Expect(0, 65827, '\P{Is_Nv=2000}', "");
Expect(1, 65827, '\P{^Is_Nv=2000}', "");
Expect(0, 65828, '\p{Is_Nv=2000}', "");
Expect(1, 65828, '\p{^Is_Nv=2000}', "");
Expect(1, 65828, '\P{Is_Nv=2000}', "");
Expect(0, 65828, '\P{^Is_Nv=2000}', "");
Error('\p{Numeric_Value=/a/	-+000000020000}');
Error('\P{Numeric_Value=/a/	-+000000020000}');
Expect(1, 65836, '\p{Numeric_Value=+0_0_0_0_0_0_0_0_20000}', "");
Expect(0, 65836, '\p{^Numeric_Value=+0_0_0_0_0_0_0_0_20000}', "");
Expect(0, 65836, '\P{Numeric_Value=+0_0_0_0_0_0_0_0_20000}', "");
Expect(1, 65836, '\P{^Numeric_Value=+0_0_0_0_0_0_0_0_20000}', "");
Expect(0, 65837, '\p{Numeric_Value=+0_0_0_0_0_0_0_0_20000}', "");
Expect(1, 65837, '\p{^Numeric_Value=+0_0_0_0_0_0_0_0_20000}', "");
Expect(1, 65837, '\P{Numeric_Value=+0_0_0_0_0_0_0_0_20000}', "");
Expect(0, 65837, '\P{^Numeric_Value=+0_0_0_0_0_0_0_0_20000}', "");
Error('\p{Nv=:=	_2000_0}');
Error('\P{Nv=:=	_2000_0}');
Expect(1, 65836, '\p{Nv=+0000000020000}', "");
Expect(0, 65836, '\p{^Nv=+0000000020000}', "");
Expect(0, 65836, '\P{Nv=+0000000020000}', "");
Expect(1, 65836, '\P{^Nv=+0000000020000}', "");
Expect(0, 65837, '\p{Nv=+0000000020000}', "");
Expect(1, 65837, '\p{^Nv=+0000000020000}', "");
Expect(1, 65837, '\P{Nv=+0000000020000}', "");
Expect(0, 65837, '\P{^Nv=+0000000020000}', "");
Error('\p{Is_Numeric_Value=-_0000020000:=}');
Error('\P{Is_Numeric_Value=-_0000020000:=}');
Expect(1, 65836, '\p{Is_Numeric_Value=020000}', "");
Expect(0, 65836, '\p{^Is_Numeric_Value=020000}', "");
Expect(0, 65836, '\P{Is_Numeric_Value=020000}', "");
Expect(1, 65836, '\P{^Is_Numeric_Value=020000}', "");
Expect(0, 65837, '\p{Is_Numeric_Value=020000}', "");
Expect(1, 65837, '\p{^Is_Numeric_Value=020000}', "");
Expect(1, 65837, '\P{Is_Numeric_Value=020000}', "");
Expect(0, 65837, '\P{^Is_Numeric_Value=020000}', "");
Error('\p{Is_Nv=:=  2000_0}');
Error('\P{Is_Nv=:=  2000_0}');
Expect(1, 65836, '\p{Is_Nv=+000000_000200_00}', "");
Expect(0, 65836, '\p{^Is_Nv=+000000_000200_00}', "");
Expect(0, 65836, '\P{Is_Nv=+000000_000200_00}', "");
Expect(1, 65836, '\P{^Is_Nv=+000000_000200_00}', "");
Expect(0, 65837, '\p{Is_Nv=+000000_000200_00}', "");
Expect(1, 65837, '\p{^Is_Nv=+000000_000200_00}', "");
Expect(1, 65837, '\P{Is_Nv=+000000_000200_00}', "");
Expect(0, 65837, '\P{^Is_Nv=+000000_000200_00}', "");
Error('\p{Numeric_Value=	0_0_0_0_0_0_0_0_0_21/a/}');
Error('\P{Numeric_Value=	0_0_0_0_0_0_0_0_0_21/a/}');
Expect(1, 12881, '\p{Numeric_Value=0000021}', "");
Expect(0, 12881, '\p{^Numeric_Value=0000021}', "");
Expect(0, 12881, '\P{Numeric_Value=0000021}', "");
Expect(1, 12881, '\P{^Numeric_Value=0000021}', "");
Expect(0, 12882, '\p{Numeric_Value=0000021}', "");
Expect(1, 12882, '\p{^Numeric_Value=0000021}', "");
Expect(1, 12882, '\P{Numeric_Value=0000021}', "");
Expect(0, 12882, '\P{^Numeric_Value=0000021}', "");
Error('\p{Nv= 	0_0_0_21:=}');
Error('\P{Nv= 	0_0_0_21:=}');
Expect(1, 12881, '\p{Nv:0021}', "");
Expect(0, 12881, '\p{^Nv:0021}', "");
Expect(0, 12881, '\P{Nv:0021}', "");
Expect(1, 12881, '\P{^Nv:0021}', "");
Expect(0, 12882, '\p{Nv:0021}', "");
Expect(1, 12882, '\p{^Nv:0021}', "");
Expect(1, 12882, '\P{Nv:0021}', "");
Expect(0, 12882, '\P{^Nv:0021}', "");
Error('\p{Is_Numeric_Value=  +00000000021/a/}');
Error('\P{Is_Numeric_Value=  +00000000021/a/}');
Expect(1, 12881, '\p{Is_Numeric_Value=+00_00_00_021}', "");
Expect(0, 12881, '\p{^Is_Numeric_Value=+00_00_00_021}', "");
Expect(0, 12881, '\P{Is_Numeric_Value=+00_00_00_021}', "");
Expect(1, 12881, '\P{^Is_Numeric_Value=+00_00_00_021}', "");
Expect(0, 12882, '\p{Is_Numeric_Value=+00_00_00_021}', "");
Expect(1, 12882, '\p{^Is_Numeric_Value=+00_00_00_021}', "");
Expect(1, 12882, '\P{Is_Numeric_Value=+00_00_00_021}', "");
Expect(0, 12882, '\P{^Is_Numeric_Value=+00_00_00_021}', "");
Error('\p{Is_Nv=:=		0_0_21}');
Error('\P{Is_Nv=:=		0_0_21}');
Expect(1, 12881, '\p{Is_Nv=000021}', "");
Expect(0, 12881, '\p{^Is_Nv=000021}', "");
Expect(0, 12881, '\P{Is_Nv=000021}', "");
Expect(1, 12881, '\P{^Is_Nv=000021}', "");
Expect(0, 12882, '\p{Is_Nv=000021}', "");
Expect(1, 12882, '\p{^Is_Nv=000021}', "");
Expect(1, 12882, '\P{Is_Nv=000021}', "");
Expect(0, 12882, '\P{^Is_Nv=000021}', "");
Error('\p{Numeric_Value:	/a/ +00000000216000}');
Error('\P{Numeric_Value:	/a/ +00000000216000}');
Expect(1, 74802, '\p{Numeric_Value=+000000216000}', "");
Expect(0, 74802, '\p{^Numeric_Value=+000000216000}', "");
Expect(0, 74802, '\P{Numeric_Value=+000000216000}', "");
Expect(1, 74802, '\P{^Numeric_Value=+000000216000}', "");
Expect(0, 74803, '\p{Numeric_Value=+000000216000}', "");
Expect(1, 74803, '\p{^Numeric_Value=+000000216000}', "");
Expect(1, 74803, '\P{Numeric_Value=+000000216000}', "");
Expect(0, 74803, '\P{^Numeric_Value=+000000216000}', "");
Error('\p{Nv: 	/a/0_0_0_0_0_0_0_0_2_1_6_0_00}');
Error('\P{Nv: 	/a/0_0_0_0_0_0_0_0_2_1_6_0_00}');
Expect(1, 74802, '\p{Nv=000000216000}', "");
Expect(0, 74802, '\p{^Nv=000000216000}', "");
Expect(0, 74802, '\P{Nv=000000216000}', "");
Expect(1, 74802, '\P{^Nv=000000216000}', "");
Expect(0, 74803, '\p{Nv=000000216000}', "");
Expect(1, 74803, '\p{^Nv=000000216000}', "");
Expect(1, 74803, '\P{Nv=000000216000}', "");
Expect(0, 74803, '\P{^Nv=000000216000}', "");
Error('\p{Is_Numeric_Value= 00000216000/a/}');
Error('\P{Is_Numeric_Value= 00000216000/a/}');
Expect(1, 74802, '\p{Is_Numeric_Value=216000}', "");
Expect(0, 74802, '\p{^Is_Numeric_Value=216000}', "");
Expect(0, 74802, '\P{Is_Numeric_Value=216000}', "");
Expect(1, 74802, '\P{^Is_Numeric_Value=216000}', "");
Expect(0, 74803, '\p{Is_Numeric_Value=216000}', "");
Expect(1, 74803, '\p{^Is_Numeric_Value=216000}', "");
Expect(1, 74803, '\P{Is_Numeric_Value=216000}', "");
Expect(0, 74803, '\P{^Is_Numeric_Value=216000}', "");
Error('\p{Is_Nv=-/a/000000000216000}');
Error('\P{Is_Nv=-/a/000000000216000}');
Expect(1, 74802, '\p{Is_Nv=0_0_0_0_2_1_6_0_00}', "");
Expect(0, 74802, '\p{^Is_Nv=0_0_0_0_2_1_6_0_00}', "");
Expect(0, 74802, '\P{Is_Nv=0_0_0_0_2_1_6_0_00}', "");
Expect(1, 74802, '\P{^Is_Nv=0_0_0_0_2_1_6_0_00}', "");
Expect(0, 74803, '\p{Is_Nv=0_0_0_0_2_1_6_0_00}', "");
Expect(1, 74803, '\p{^Is_Nv=0_0_0_0_2_1_6_0_00}', "");
Expect(1, 74803, '\P{Is_Nv=0_0_0_0_2_1_6_0_00}', "");
Expect(0, 74803, '\P{^Is_Nv=0_0_0_0_2_1_6_0_00}', "");
Error('\p{Numeric_Value=+00000000022/a/}');
Error('\P{Numeric_Value=+00000000022/a/}');
Expect(1, 12882, '\p{Numeric_Value:   00_00_22}', "");
Expect(0, 12882, '\p{^Numeric_Value:   00_00_22}', "");
Expect(0, 12882, '\P{Numeric_Value:   00_00_22}', "");
Expect(1, 12882, '\P{^Numeric_Value:   00_00_22}', "");
Expect(0, 12883, '\p{Numeric_Value:   00_00_22}', "");
Expect(1, 12883, '\p{^Numeric_Value:   00_00_22}', "");
Expect(1, 12883, '\P{Numeric_Value:   00_00_22}', "");
Expect(0, 12883, '\P{^Numeric_Value:   00_00_22}', "");
Error('\p{Nv=:=  +00000022}');
Error('\P{Nv=:=  +00000022}');
Expect(1, 12882, '\p{Nv=+0000_0002_2}', "");
Expect(0, 12882, '\p{^Nv=+0000_0002_2}', "");
Expect(0, 12882, '\P{Nv=+0000_0002_2}', "");
Expect(1, 12882, '\P{^Nv=+0000_0002_2}', "");
Expect(0, 12883, '\p{Nv=+0000_0002_2}', "");
Expect(1, 12883, '\p{^Nv=+0000_0002_2}', "");
Expect(1, 12883, '\P{Nv=+0000_0002_2}', "");
Expect(0, 12883, '\P{^Nv=+0000_0002_2}', "");
Error('\p{Is_Numeric_Value=	 +02_2/a/}');
Error('\P{Is_Numeric_Value=	 +02_2/a/}');
Expect(1, 12882, '\p{Is_Numeric_Value=+000_000_002_2}', "");
Expect(0, 12882, '\p{^Is_Numeric_Value=+000_000_002_2}', "");
Expect(0, 12882, '\P{Is_Numeric_Value=+000_000_002_2}', "");
Expect(1, 12882, '\P{^Is_Numeric_Value=+000_000_002_2}', "");
Expect(0, 12883, '\p{Is_Numeric_Value=+000_000_002_2}', "");
Expect(1, 12883, '\p{^Is_Numeric_Value=+000_000_002_2}', "");
Expect(1, 12883, '\P{Is_Numeric_Value=+000_000_002_2}', "");
Expect(0, 12883, '\P{^Is_Numeric_Value=+000_000_002_2}', "");
Error('\p{Is_Nv=-_0000000022/a/}');
Error('\P{Is_Nv=-_0000000022/a/}');
Expect(1, 12882, '\p{Is_Nv=2_2}', "");
Expect(0, 12882, '\p{^Is_Nv=2_2}', "");
Expect(0, 12882, '\P{Is_Nv=2_2}', "");
Expect(1, 12882, '\P{^Is_Nv=2_2}', "");
Expect(0, 12883, '\p{Is_Nv=2_2}', "");
Expect(1, 12883, '\p{^Is_Nv=2_2}', "");
Expect(1, 12883, '\P{Is_Nv=2_2}', "");
Expect(0, 12883, '\P{^Is_Nv=2_2}', "");
Error('\p{Numeric_Value=:=_	+0000000023}');
Error('\P{Numeric_Value=:=_	+0000000023}');
Expect(1, 12883, '\p{Numeric_Value=23}', "");
Expect(0, 12883, '\p{^Numeric_Value=23}', "");
Expect(0, 12883, '\P{Numeric_Value=23}', "");
Expect(1, 12883, '\P{^Numeric_Value=23}', "");
Expect(0, 12884, '\p{Numeric_Value=23}', "");
Expect(1, 12884, '\p{^Numeric_Value=23}', "");
Expect(1, 12884, '\P{Numeric_Value=23}', "");
Expect(0, 12884, '\P{^Numeric_Value=23}', "");
Error('\p{Nv::= 0002_3}');
Error('\P{Nv::= 0002_3}');
Expect(1, 12883, '\p{Nv=000_000_23}', "");
Expect(0, 12883, '\p{^Nv=000_000_23}', "");
Expect(0, 12883, '\P{Nv=000_000_23}', "");
Expect(1, 12883, '\P{^Nv=000_000_23}', "");
Expect(0, 12884, '\p{Nv=000_000_23}', "");
Expect(1, 12884, '\p{^Nv=000_000_23}', "");
Expect(1, 12884, '\P{Nv=000_000_23}', "");
Expect(0, 12884, '\P{^Nv=000_000_23}', "");
Error('\p{Is_Numeric_Value=/a/		0000000023}');
Error('\P{Is_Numeric_Value=/a/		0000000023}');
Expect(1, 12883, '\p{Is_Numeric_Value=0000000023}', "");
Expect(0, 12883, '\p{^Is_Numeric_Value=0000000023}', "");
Expect(0, 12883, '\P{Is_Numeric_Value=0000000023}', "");
Expect(1, 12883, '\P{^Is_Numeric_Value=0000000023}', "");
Expect(0, 12884, '\p{Is_Numeric_Value=0000000023}', "");
Expect(1, 12884, '\p{^Is_Numeric_Value=0000000023}', "");
Expect(1, 12884, '\P{Is_Numeric_Value=0000000023}', "");
Expect(0, 12884, '\P{^Is_Numeric_Value=0000000023}', "");
Error('\p{Is_Nv=:=- +00002_3}');
Error('\P{Is_Nv=:=- +00002_3}');
Expect(1, 12883, '\p{Is_Nv=+00000023}', "");
Expect(0, 12883, '\p{^Is_Nv=+00000023}', "");
Expect(0, 12883, '\P{Is_Nv=+00000023}', "");
Expect(1, 12883, '\P{^Is_Nv=+00000023}', "");
Expect(0, 12884, '\p{Is_Nv=+00000023}', "");
Expect(1, 12884, '\p{^Is_Nv=+00000023}', "");
Expect(1, 12884, '\P{Is_Nv=+00000023}', "");
Expect(0, 12884, '\P{^Is_Nv=+00000023}', "");
Error('\p{Numeric_Value=:= _024}');
Error('\P{Numeric_Value=:= _024}');
Expect(1, 12884, '\p{Numeric_Value=+0_0_0_24}', "");
Expect(0, 12884, '\p{^Numeric_Value=+0_0_0_24}', "");
Expect(0, 12884, '\P{Numeric_Value=+0_0_0_24}', "");
Expect(1, 12884, '\P{^Numeric_Value=+0_0_0_24}', "");
Expect(0, 12885, '\p{Numeric_Value=+0_0_0_24}', "");
Expect(1, 12885, '\p{^Numeric_Value=+0_0_0_24}', "");
Expect(1, 12885, '\P{Numeric_Value=+0_0_0_24}', "");
Expect(0, 12885, '\P{^Numeric_Value=+0_0_0_24}', "");
Error('\p{Nv=/a/	-+00000000024}');
Error('\P{Nv=/a/	-+00000000024}');
Expect(1, 12884, '\p{Nv=+00000024}', "");
Expect(0, 12884, '\p{^Nv=+00000024}', "");
Expect(0, 12884, '\P{Nv=+00000024}', "");
Expect(1, 12884, '\P{^Nv=+00000024}', "");
Expect(0, 12885, '\p{Nv=+00000024}', "");
Expect(1, 12885, '\p{^Nv=+00000024}', "");
Expect(1, 12885, '\P{Nv=+00000024}', "");
Expect(0, 12885, '\P{^Nv=+00000024}', "");
Error('\p{Is_Numeric_Value=:=__00_00_00_02_4}');
Error('\P{Is_Numeric_Value=:=__00_00_00_02_4}');
Expect(1, 12884, '\p{Is_Numeric_Value=024}', "");
Expect(0, 12884, '\p{^Is_Numeric_Value=024}', "");
Expect(0, 12884, '\P{Is_Numeric_Value=024}', "");
Expect(1, 12884, '\P{^Is_Numeric_Value=024}', "");
Expect(0, 12885, '\p{Is_Numeric_Value=024}', "");
Expect(1, 12885, '\p{^Is_Numeric_Value=024}', "");
Expect(1, 12885, '\P{Is_Numeric_Value=024}', "");
Expect(0, 12885, '\P{^Is_Numeric_Value=024}', "");
Error('\p{Is_Nv=:=+0000002_4}');
Error('\P{Is_Nv=:=+0000002_4}');
Expect(1, 12884, '\p{Is_Nv=0_0_24}', "");
Expect(0, 12884, '\p{^Is_Nv=0_0_24}', "");
Expect(0, 12884, '\P{Is_Nv=0_0_24}', "");
Expect(1, 12884, '\P{^Is_Nv=0_0_24}', "");
Expect(0, 12885, '\p{Is_Nv=0_0_24}', "");
Expect(1, 12885, '\p{^Is_Nv=0_0_24}', "");
Expect(1, 12885, '\P{Is_Nv=0_0_24}', "");
Expect(0, 12885, '\P{^Is_Nv=0_0_24}', "");
Error('\p{Numeric_Value=:=_-2_5}');
Error('\P{Numeric_Value=:=_-2_5}');
Expect(1, 12885, '\p{Numeric_Value=0_0_0_0_025}', "");
Expect(0, 12885, '\p{^Numeric_Value=0_0_0_0_025}', "");
Expect(0, 12885, '\P{Numeric_Value=0_0_0_0_025}', "");
Expect(1, 12885, '\P{^Numeric_Value=0_0_0_0_025}', "");
Expect(0, 12886, '\p{Numeric_Value=0_0_0_0_025}', "");
Expect(1, 12886, '\p{^Numeric_Value=0_0_0_0_025}', "");
Expect(1, 12886, '\P{Numeric_Value=0_0_0_0_025}', "");
Expect(0, 12886, '\P{^Numeric_Value=0_0_0_0_025}', "");
Error('\p{Nv=/a/ -00_02_5}');
Error('\P{Nv=/a/ -00_02_5}');
Expect(1, 12885, '\p{Nv=0_0_0_0_0_00025}', "");
Expect(0, 12885, '\p{^Nv=0_0_0_0_0_00025}', "");
Expect(0, 12885, '\P{Nv=0_0_0_0_0_00025}', "");
Expect(1, 12885, '\P{^Nv=0_0_0_0_0_00025}', "");
Expect(0, 12886, '\p{Nv=0_0_0_0_0_00025}', "");
Expect(1, 12886, '\p{^Nv=0_0_0_0_0_00025}', "");
Expect(1, 12886, '\P{Nv=0_0_0_0_0_00025}', "");
Expect(0, 12886, '\P{^Nv=0_0_0_0_0_00025}', "");
Error('\p{Is_Numeric_Value=:=__+0_0_0_0_0_0_25}');
Error('\P{Is_Numeric_Value=:=__+0_0_0_0_0_0_25}');
Expect(1, 12885, '\p{Is_Numeric_Value=0_0_25}', "");
Expect(0, 12885, '\p{^Is_Numeric_Value=0_0_25}', "");
Expect(0, 12885, '\P{Is_Numeric_Value=0_0_25}', "");
Expect(1, 12885, '\P{^Is_Numeric_Value=0_0_25}', "");
Expect(0, 12886, '\p{Is_Numeric_Value=0_0_25}', "");
Expect(1, 12886, '\p{^Is_Numeric_Value=0_0_25}', "");
Expect(1, 12886, '\P{Is_Numeric_Value=0_0_25}', "");
Expect(0, 12886, '\P{^Is_Numeric_Value=0_0_25}', "");
Error('\p{Is_Nv= :=000000025}');
Error('\P{Is_Nv= :=000000025}');
Expect(1, 12885, '\p{Is_Nv:   2_5}', "");
Expect(0, 12885, '\p{^Is_Nv:   2_5}', "");
Expect(0, 12885, '\P{Is_Nv:   2_5}', "");
Expect(1, 12885, '\P{^Is_Nv:   2_5}', "");
Expect(0, 12886, '\p{Is_Nv:   2_5}', "");
Expect(1, 12886, '\p{^Is_Nv:   2_5}', "");
Expect(1, 12886, '\P{Is_Nv:   2_5}', "");
Expect(0, 12886, '\P{^Is_Nv:   2_5}', "");
Error('\p{Numeric_Value= :=0000000026}');
Error('\P{Numeric_Value= :=0000000026}');
Expect(1, 12886, '\p{Numeric_Value=0000026}', "");
Expect(0, 12886, '\p{^Numeric_Value=0000026}', "");
Expect(0, 12886, '\P{Numeric_Value=0000026}', "");
Expect(1, 12886, '\P{^Numeric_Value=0000026}', "");
Expect(0, 12887, '\p{Numeric_Value=0000026}', "");
Expect(1, 12887, '\p{^Numeric_Value=0000026}', "");
Expect(1, 12887, '\P{Numeric_Value=0000026}', "");
Expect(0, 12887, '\P{^Numeric_Value=0000026}', "");
Error('\p{Nv: /a/-00002_6}');
Error('\P{Nv: /a/-00002_6}');
Expect(1, 12886, '\p{Nv=+026}', "");
Expect(0, 12886, '\p{^Nv=+026}', "");
Expect(0, 12886, '\P{Nv=+026}', "");
Expect(1, 12886, '\P{^Nv=+026}', "");
Expect(0, 12887, '\p{Nv=+026}', "");
Expect(1, 12887, '\p{^Nv=+026}', "");
Expect(1, 12887, '\P{Nv=+026}', "");
Expect(0, 12887, '\P{^Nv=+026}', "");
Error('\p{Is_Numeric_Value=:=_2_6}');
Error('\P{Is_Numeric_Value=:=_2_6}');
Expect(1, 12886, '\p{Is_Numeric_Value: 0_0_26}', "");
Expect(0, 12886, '\p{^Is_Numeric_Value: 0_0_26}', "");
Expect(0, 12886, '\P{Is_Numeric_Value: 0_0_26}', "");
Expect(1, 12886, '\P{^Is_Numeric_Value: 0_0_26}', "");
Expect(0, 12887, '\p{Is_Numeric_Value: 0_0_26}', "");
Expect(1, 12887, '\p{^Is_Numeric_Value: 0_0_26}', "");
Expect(1, 12887, '\P{Is_Numeric_Value: 0_0_26}', "");
Expect(0, 12887, '\P{^Is_Numeric_Value: 0_0_26}', "");
Error('\p{Is_Nv=__+00_02_6/a/}');
Error('\P{Is_Nv=__+00_02_6/a/}');
Expect(1, 12886, '\p{Is_Nv=0_0_0_26}', "");
Expect(0, 12886, '\p{^Is_Nv=0_0_0_26}', "");
Expect(0, 12886, '\P{Is_Nv=0_0_0_26}', "");
Expect(1, 12886, '\P{^Is_Nv=0_0_0_26}', "");
Expect(0, 12887, '\p{Is_Nv=0_0_0_26}', "");
Expect(1, 12887, '\p{^Is_Nv=0_0_0_26}', "");
Expect(1, 12887, '\P{Is_Nv=0_0_0_26}', "");
Expect(0, 12887, '\P{^Is_Nv=0_0_0_26}', "");
Error('\p{Numeric_Value=:=+0_0_0_0_0_0027}');
Error('\P{Numeric_Value=:=+0_0_0_0_0_0027}');
Expect(1, 12887, '\p{Numeric_Value=000027}', "");
Expect(0, 12887, '\p{^Numeric_Value=000027}', "");
Expect(0, 12887, '\P{Numeric_Value=000027}', "");
Expect(1, 12887, '\P{^Numeric_Value=000027}', "");
Expect(0, 12888, '\p{Numeric_Value=000027}', "");
Expect(1, 12888, '\p{^Numeric_Value=000027}', "");
Expect(1, 12888, '\P{Numeric_Value=000027}', "");
Expect(0, 12888, '\P{^Numeric_Value=000027}', "");
Error('\p{Nv=:=  0027}');
Error('\P{Nv=:=  0027}');
Expect(1, 12887, '\p{Nv=+2_7}', "");
Expect(0, 12887, '\p{^Nv=+2_7}', "");
Expect(0, 12887, '\P{Nv=+2_7}', "");
Expect(1, 12887, '\P{^Nv=+2_7}', "");
Expect(0, 12888, '\p{Nv=+2_7}', "");
Expect(1, 12888, '\p{^Nv=+2_7}', "");
Expect(1, 12888, '\P{Nv=+2_7}', "");
Expect(0, 12888, '\P{^Nv=+2_7}', "");
Error('\p{Is_Numeric_Value:	-+0027:=}');
Error('\P{Is_Numeric_Value:	-+0027:=}');
Expect(1, 12887, '\p{Is_Numeric_Value=2_7}', "");
Expect(0, 12887, '\p{^Is_Numeric_Value=2_7}', "");
Expect(0, 12887, '\P{Is_Numeric_Value=2_7}', "");
Expect(1, 12887, '\P{^Is_Numeric_Value=2_7}', "");
Expect(0, 12888, '\p{Is_Numeric_Value=2_7}', "");
Expect(1, 12888, '\p{^Is_Numeric_Value=2_7}', "");
Expect(1, 12888, '\P{Is_Numeric_Value=2_7}', "");
Expect(0, 12888, '\P{^Is_Numeric_Value=2_7}', "");
Error('\p{Is_Nv=_/a/027}');
Error('\P{Is_Nv=_/a/027}');
Expect(1, 12887, '\p{Is_Nv:+00_00_00_02_7}', "");
Expect(0, 12887, '\p{^Is_Nv:+00_00_00_02_7}', "");
Expect(0, 12887, '\P{Is_Nv:+00_00_00_02_7}', "");
Expect(1, 12887, '\P{^Is_Nv:+00_00_00_02_7}', "");
Expect(0, 12888, '\p{Is_Nv:+00_00_00_02_7}', "");
Expect(1, 12888, '\p{^Is_Nv:+00_00_00_02_7}', "");
Expect(1, 12888, '\P{Is_Nv:+00_00_00_02_7}', "");
Expect(0, 12888, '\P{^Is_Nv:+00_00_00_02_7}', "");
Error('\p{Numeric_Value= /a/+0_0_0_0_0_0_0_0_28}');
Error('\P{Numeric_Value= /a/+0_0_0_0_0_0_0_0_28}');
Expect(1, 12888, '\p{Numeric_Value=+0028}', "");
Expect(0, 12888, '\p{^Numeric_Value=+0028}', "");
Expect(0, 12888, '\P{Numeric_Value=+0028}', "");
Expect(1, 12888, '\P{^Numeric_Value=+0028}', "");
Expect(0, 12889, '\p{Numeric_Value=+0028}', "");
Expect(1, 12889, '\p{^Numeric_Value=+0028}', "");
Expect(1, 12889, '\P{Numeric_Value=+0028}', "");
Expect(0, 12889, '\P{^Numeric_Value=+0028}', "");
Error('\p{Nv=:= _002_8}');
Error('\P{Nv=:= _002_8}');
Expect(1, 12888, '\p{Nv=0000002_8}', "");
Expect(0, 12888, '\p{^Nv=0000002_8}', "");
Expect(0, 12888, '\P{Nv=0000002_8}', "");
Expect(1, 12888, '\P{^Nv=0000002_8}', "");
Expect(0, 12889, '\p{Nv=0000002_8}', "");
Expect(1, 12889, '\p{^Nv=0000002_8}', "");
Expect(1, 12889, '\P{Nv=0000002_8}', "");
Expect(0, 12889, '\P{^Nv=0000002_8}', "");
Error('\p{Is_Numeric_Value=/a/00000_00002_8}');
Error('\P{Is_Numeric_Value=/a/00000_00002_8}');
Expect(1, 12888, '\p{Is_Numeric_Value=+0000028}', "");
Expect(0, 12888, '\p{^Is_Numeric_Value=+0000028}', "");
Expect(0, 12888, '\P{Is_Numeric_Value=+0000028}', "");
Expect(1, 12888, '\P{^Is_Numeric_Value=+0000028}', "");
Expect(0, 12889, '\p{Is_Numeric_Value=+0000028}', "");
Expect(1, 12889, '\p{^Is_Numeric_Value=+0000028}', "");
Expect(1, 12889, '\P{Is_Numeric_Value=+0000028}', "");
Expect(0, 12889, '\P{^Is_Numeric_Value=+0000028}', "");
Error('\p{Is_Nv=/a/ 	02_8}');
Error('\P{Is_Nv=/a/ 	02_8}');
Expect(1, 12888, '\p{Is_Nv=0_0_0_0_0_0_028}', "");
Expect(0, 12888, '\p{^Is_Nv=0_0_0_0_0_0_028}', "");
Expect(0, 12888, '\P{Is_Nv=0_0_0_0_0_0_028}', "");
Expect(1, 12888, '\P{^Is_Nv=0_0_0_0_0_0_028}', "");
Expect(0, 12889, '\p{Is_Nv=0_0_0_0_0_0_028}', "");
Expect(1, 12889, '\p{^Is_Nv=0_0_0_0_0_0_028}', "");
Expect(1, 12889, '\P{Is_Nv=0_0_0_0_0_0_028}', "");
Expect(0, 12889, '\P{^Is_Nv=0_0_0_0_0_0_028}', "");
Error('\p{Numeric_Value=	/a/0000000029}');
Error('\P{Numeric_Value=	/a/0000000029}');
Expect(1, 12889, '\p{Numeric_Value:2_9}', "");
Expect(0, 12889, '\p{^Numeric_Value:2_9}', "");
Expect(0, 12889, '\P{Numeric_Value:2_9}', "");
Expect(1, 12889, '\P{^Numeric_Value:2_9}', "");
Expect(0, 12890, '\p{Numeric_Value:2_9}', "");
Expect(1, 12890, '\p{^Numeric_Value:2_9}', "");
Expect(1, 12890, '\P{Numeric_Value:2_9}', "");
Expect(0, 12890, '\P{^Numeric_Value:2_9}', "");
Error('\p{Nv=/a/	_00_00_29}');
Error('\P{Nv=/a/	_00_00_29}');
Expect(1, 12889, '\p{Nv=0_0_0_0_0_0_29}', "");
Expect(0, 12889, '\p{^Nv=0_0_0_0_0_0_29}', "");
Expect(0, 12889, '\P{Nv=0_0_0_0_0_0_29}', "");
Expect(1, 12889, '\P{^Nv=0_0_0_0_0_0_29}', "");
Expect(0, 12890, '\p{Nv=0_0_0_0_0_0_29}', "");
Expect(1, 12890, '\p{^Nv=0_0_0_0_0_0_29}', "");
Expect(1, 12890, '\P{Nv=0_0_0_0_0_0_29}', "");
Expect(0, 12890, '\P{^Nv=0_0_0_0_0_0_29}', "");
Error('\p{Is_Numeric_Value= /a/+000000029}');
Error('\P{Is_Numeric_Value= /a/+000000029}');
Expect(1, 12889, '\p{Is_Numeric_Value=+029}', "");
Expect(0, 12889, '\p{^Is_Numeric_Value=+029}', "");
Expect(0, 12889, '\P{Is_Numeric_Value=+029}', "");
Expect(1, 12889, '\P{^Is_Numeric_Value=+029}', "");
Expect(0, 12890, '\p{Is_Numeric_Value=+029}', "");
Expect(1, 12890, '\p{^Is_Numeric_Value=+029}', "");
Expect(1, 12890, '\P{Is_Numeric_Value=+029}', "");
Expect(0, 12890, '\P{^Is_Numeric_Value=+029}', "");
Error('\p{Is_Nv=	/a/+0_0_29}');
Error('\P{Is_Nv=	/a/+0_0_29}');
Expect(1, 12889, '\p{Is_Nv=00002_9}', "");
Expect(0, 12889, '\p{^Is_Nv=00002_9}', "");
Expect(0, 12889, '\P{Is_Nv=00002_9}', "");
Expect(1, 12889, '\P{^Is_Nv=00002_9}', "");
Expect(0, 12890, '\p{Is_Nv=00002_9}', "");
Expect(1, 12890, '\p{^Is_Nv=00002_9}', "");
Expect(1, 12890, '\P{Is_Nv=00002_9}', "");
Expect(0, 12890, '\P{^Is_Nv=00002_9}', "");
Error('\p{Numeric_Value= -+00003/a/}');
Error('\P{Numeric_Value= -+00003/a/}');
Expect(1, 146203, '\p{Numeric_Value=+0_0_0_0_0_0_0003}', "");
Expect(0, 146203, '\p{^Numeric_Value=+0_0_0_0_0_0_0003}', "");
Expect(0, 146203, '\P{Numeric_Value=+0_0_0_0_0_0_0003}', "");
Expect(1, 146203, '\P{^Numeric_Value=+0_0_0_0_0_0_0003}', "");
Expect(0, 146204, '\p{Numeric_Value=+0_0_0_0_0_0_0003}', "");
Expect(1, 146204, '\p{^Numeric_Value=+0_0_0_0_0_0_0003}', "");
Expect(1, 146204, '\P{Numeric_Value=+0_0_0_0_0_0_0003}', "");
Expect(0, 146204, '\P{^Numeric_Value=+0_0_0_0_0_0_0003}', "");
Error('\p{Nv=-:=00_3}');
Error('\P{Nv=-:=00_3}');
Expect(1, 146203, '\p{Nv=00_3}', "");
Expect(0, 146203, '\p{^Nv=00_3}', "");
Expect(0, 146203, '\P{Nv=00_3}', "");
Expect(1, 146203, '\P{^Nv=00_3}', "");
Expect(0, 146204, '\p{Nv=00_3}', "");
Expect(1, 146204, '\p{^Nv=00_3}', "");
Expect(1, 146204, '\P{Nv=00_3}', "");
Expect(0, 146204, '\P{^Nv=00_3}', "");
Error('\p{Is_Numeric_Value=_/a/003}');
Error('\P{Is_Numeric_Value=_/a/003}');
Expect(1, 146203, '\p{Is_Numeric_Value=+00000003}', "");
Expect(0, 146203, '\p{^Is_Numeric_Value=+00000003}', "");
Expect(0, 146203, '\P{Is_Numeric_Value=+00000003}', "");
Expect(1, 146203, '\P{^Is_Numeric_Value=+00000003}', "");
Expect(0, 146204, '\p{Is_Numeric_Value=+00000003}', "");
Expect(1, 146204, '\p{^Is_Numeric_Value=+00000003}', "");
Expect(1, 146204, '\P{Is_Numeric_Value=+00000003}', "");
Expect(0, 146204, '\P{^Is_Numeric_Value=+00000003}', "");
Error('\p{Is_Nv=	 0_0_0_0_0_0_0_003:=}');
Error('\P{Is_Nv=	 0_0_0_0_0_0_0_003:=}');
Expect(1, 146203, '\p{Is_Nv=0003}', "");
Expect(0, 146203, '\p{^Is_Nv=0003}', "");
Expect(0, 146203, '\P{Is_Nv=0003}', "");
Expect(1, 146203, '\P{^Is_Nv=0003}', "");
Expect(0, 146204, '\p{Is_Nv=0003}', "");
Expect(1, 146204, '\p{^Is_Nv=0003}', "");
Expect(1, 146204, '\P{Is_Nv=0003}', "");
Expect(0, 146204, '\P{^Is_Nv=0003}', "");
Error('\p{Numeric_Value=/a/_	003/0000000016}');
Error('\P{Numeric_Value=/a/_	003/0000000016}');
Expect(1, 43061, '\p{Numeric_Value=0000003/0016}', "");
Expect(0, 43061, '\p{^Numeric_Value=0000003/0016}', "");
Expect(0, 43061, '\P{Numeric_Value=0000003/0016}', "");
Expect(1, 43061, '\P{^Numeric_Value=0000003/0016}', "");
Expect(0, 43062, '\p{Numeric_Value=0000003/0016}', "");
Expect(1, 43062, '\p{^Numeric_Value=0000003/0016}', "");
Expect(1, 43062, '\P{Numeric_Value=0000003/0016}', "");
Expect(0, 43062, '\P{^Numeric_Value=0000003/0016}', "");
Error('\p{Numeric_Value=0.19}');
Error('\P{Numeric_Value=0.19}');
Expect(1, 43061, '\p{Numeric_Value: 0.188}', "");
Expect(0, 43061, '\p{^Numeric_Value: 0.188}', "");
Expect(0, 43061, '\P{Numeric_Value: 0.188}', "");
Expect(1, 43061, '\P{^Numeric_Value: 0.188}', "");
Expect(0, 43062, '\p{Numeric_Value: 0.188}', "");
Expect(1, 43062, '\p{^Numeric_Value: 0.188}', "");
Expect(1, 43062, '\P{Numeric_Value: 0.188}', "");
Expect(0, 43062, '\P{^Numeric_Value: 0.188}', "");
Error('\p{Nv=_3/016/a/}');
Error('\P{Nv=_3/016/a/}');
Expect(1, 43061, '\p{Nv=+0000000003/0000000016}', "");
Expect(0, 43061, '\p{^Nv=+0000000003/0000000016}', "");
Expect(0, 43061, '\P{Nv=+0000000003/0000000016}', "");
Expect(1, 43061, '\P{^Nv=+0000000003/0000000016}', "");
Expect(0, 43062, '\p{Nv=+0000000003/0000000016}', "");
Expect(1, 43062, '\p{^Nv=+0000000003/0000000016}', "");
Expect(1, 43062, '\P{Nv=+0000000003/0000000016}', "");
Expect(0, 43062, '\P{^Nv=+0000000003/0000000016}', "");
Error('\p{Nv=0.19}');
Error('\P{Nv=0.19}');
Expect(1, 43061, '\p{Nv=0.188}', "");
Expect(0, 43061, '\p{^Nv=0.188}', "");
Expect(0, 43061, '\P{Nv=0.188}', "");
Expect(1, 43061, '\P{^Nv=0.188}', "");
Expect(0, 43062, '\p{Nv=0.188}', "");
Expect(1, 43062, '\p{^Nv=0.188}', "");
Expect(1, 43062, '\P{Nv=0.188}', "");
Expect(0, 43062, '\P{^Nv=0.188}', "");
Error('\p{Is_Numeric_Value:   /a/_3/0016}');
Error('\P{Is_Numeric_Value:   /a/_3/0016}');
Expect(1, 43061, '\p{Is_Numeric_Value:   +00003/000000016}', "");
Expect(0, 43061, '\p{^Is_Numeric_Value:   +00003/000000016}', "");
Expect(0, 43061, '\P{Is_Numeric_Value:   +00003/000000016}', "");
Expect(1, 43061, '\P{^Is_Numeric_Value:   +00003/000000016}', "");
Expect(0, 43062, '\p{Is_Numeric_Value:   +00003/000000016}', "");
Expect(1, 43062, '\p{^Is_Numeric_Value:   +00003/000000016}', "");
Expect(1, 43062, '\P{Is_Numeric_Value:   +00003/000000016}', "");
Expect(0, 43062, '\P{^Is_Numeric_Value:   +00003/000000016}', "");
Error('\p{Is_Numeric_Value=0.19}');
Error('\P{Is_Numeric_Value=0.19}');
Expect(1, 43061, '\p{Is_Numeric_Value=0.188}', "");
Expect(0, 43061, '\p{^Is_Numeric_Value=0.188}', "");
Expect(0, 43061, '\P{Is_Numeric_Value=0.188}', "");
Expect(1, 43061, '\P{^Is_Numeric_Value=0.188}', "");
Expect(0, 43062, '\p{Is_Numeric_Value=0.188}', "");
Expect(1, 43062, '\p{^Is_Numeric_Value=0.188}', "");
Expect(1, 43062, '\P{Is_Numeric_Value=0.188}', "");
Expect(0, 43062, '\P{^Is_Numeric_Value=0.188}', "");
Error('\p{Is_Nv= 0000000003/016/a/}');
Error('\P{Is_Nv= 0000000003/016/a/}');
Expect(1, 43061, '\p{Is_Nv=+000000003/0016}', "");
Expect(0, 43061, '\p{^Is_Nv=+000000003/0016}', "");
Expect(0, 43061, '\P{Is_Nv=+000000003/0016}', "");
Expect(1, 43061, '\P{^Is_Nv=+000000003/0016}', "");
Expect(0, 43062, '\p{Is_Nv=+000000003/0016}', "");
Expect(1, 43062, '\p{^Is_Nv=+000000003/0016}', "");
Expect(1, 43062, '\P{Is_Nv=+000000003/0016}', "");
Expect(0, 43062, '\P{^Is_Nv=+000000003/0016}', "");
Error('\p{Is_Nv=0.19}');
Error('\P{Is_Nv=0.19}');
Expect(1, 43061, '\p{Is_Nv=0.188}', "");
Expect(0, 43061, '\p{^Is_Nv=0.188}', "");
Expect(0, 43061, '\P{Is_Nv=0.188}', "");
Expect(1, 43061, '\P{^Is_Nv=0.188}', "");
Expect(0, 43062, '\p{Is_Nv=0.188}', "");
Expect(1, 43062, '\p{^Is_Nv=0.188}', "");
Expect(1, 43062, '\P{Is_Nv=0.188}', "");
Expect(0, 43062, '\P{^Is_Nv=0.188}', "");
Error('\p{Numeric_Value:	-/a/003/002}');
Error('\P{Numeric_Value:	-/a/003/002}');
Expect(1, 3883, '\p{Numeric_Value=+00000003/000000002}', "");
Expect(0, 3883, '\p{^Numeric_Value=+00000003/000000002}', "");
Expect(0, 3883, '\P{Numeric_Value=+00000003/000000002}', "");
Expect(1, 3883, '\P{^Numeric_Value=+00000003/000000002}', "");
Expect(0, 3884, '\p{Numeric_Value=+00000003/000000002}', "");
Expect(1, 3884, '\p{^Numeric_Value=+00000003/000000002}', "");
Expect(1, 3884, '\P{Numeric_Value=+00000003/000000002}', "");
Expect(0, 3884, '\P{^Numeric_Value=+00000003/000000002}', "");
Expect(1, 3883, '\p{Numeric_Value=1.500}', "");
Expect(0, 3883, '\p{^Numeric_Value=1.500}', "");
Expect(0, 3883, '\P{Numeric_Value=1.500}', "");
Expect(1, 3883, '\P{^Numeric_Value=1.500}', "");
Expect(0, 3884, '\p{Numeric_Value=1.500}', "");
Expect(1, 3884, '\p{^Numeric_Value=1.500}', "");
Expect(1, 3884, '\P{Numeric_Value=1.500}', "");
Expect(0, 3884, '\P{^Numeric_Value=1.500}', "");
Error('\p{Nv= +3/002:=}');
Error('\P{Nv= +3/002:=}');
Expect(1, 3883, '\p{Nv=+0000000003/0000002}', "");
Expect(0, 3883, '\p{^Nv=+0000000003/0000002}', "");
Expect(0, 3883, '\P{Nv=+0000000003/0000002}', "");
Expect(1, 3883, '\P{^Nv=+0000000003/0000002}', "");
Expect(0, 3884, '\p{Nv=+0000000003/0000002}', "");
Expect(1, 3884, '\p{^Nv=+0000000003/0000002}', "");
Expect(1, 3884, '\P{Nv=+0000000003/0000002}', "");
Expect(0, 3884, '\P{^Nv=+0000000003/0000002}', "");
Expect(1, 3883, '\p{Nv=1.500}', "");
Expect(0, 3883, '\p{^Nv=1.500}', "");
Expect(0, 3883, '\P{Nv=1.500}', "");
Expect(1, 3883, '\P{^Nv=1.500}', "");
Expect(0, 3884, '\p{Nv=1.500}', "");
Expect(1, 3884, '\p{^Nv=1.500}', "");
Expect(1, 3884, '\P{Nv=1.500}', "");
Expect(0, 3884, '\P{^Nv=1.500}', "");
Error('\p{Is_Numeric_Value=/a/	_0000000003/0000002}');
Error('\P{Is_Numeric_Value=/a/	_0000000003/0000002}');
Expect(1, 3883, '\p{Is_Numeric_Value=+03/000002}', "");
Expect(0, 3883, '\p{^Is_Numeric_Value=+03/000002}', "");
Expect(0, 3883, '\P{Is_Numeric_Value=+03/000002}', "");
Expect(1, 3883, '\P{^Is_Numeric_Value=+03/000002}', "");
Expect(0, 3884, '\p{Is_Numeric_Value=+03/000002}', "");
Expect(1, 3884, '\p{^Is_Numeric_Value=+03/000002}', "");
Expect(1, 3884, '\P{Is_Numeric_Value=+03/000002}', "");
Expect(0, 3884, '\P{^Is_Numeric_Value=+03/000002}', "");
Expect(1, 3883, '\p{Is_Numeric_Value=1.500}', "");
Expect(0, 3883, '\p{^Is_Numeric_Value=1.500}', "");
Expect(0, 3883, '\P{Is_Numeric_Value=1.500}', "");
Expect(1, 3883, '\P{^Is_Numeric_Value=1.500}', "");
Expect(0, 3884, '\p{Is_Numeric_Value=1.500}', "");
Expect(1, 3884, '\p{^Is_Numeric_Value=1.500}', "");
Expect(1, 3884, '\P{Is_Numeric_Value=1.500}', "");
Expect(0, 3884, '\P{^Is_Numeric_Value=1.500}', "");
Error('\p{Is_Nv:		000000003/002/a/}');
Error('\P{Is_Nv:		000000003/002/a/}');
Expect(1, 3883, '\p{Is_Nv=03/000002}', "");
Expect(0, 3883, '\p{^Is_Nv=03/000002}', "");
Expect(0, 3883, '\P{Is_Nv=03/000002}', "");
Expect(1, 3883, '\P{^Is_Nv=03/000002}', "");
Expect(0, 3884, '\p{Is_Nv=03/000002}', "");
Expect(1, 3884, '\p{^Is_Nv=03/000002}', "");
Expect(1, 3884, '\P{Is_Nv=03/000002}', "");
Expect(0, 3884, '\P{^Is_Nv=03/000002}', "");
Expect(1, 3883, '\p{Is_Nv=1.500}', "");
Expect(0, 3883, '\p{^Is_Nv=1.500}', "");
Expect(0, 3883, '\P{Is_Nv=1.500}', "");
Expect(1, 3883, '\P{^Is_Nv=1.500}', "");
Expect(0, 3884, '\p{Is_Nv=1.500}', "");
Expect(1, 3884, '\p{^Is_Nv=1.500}', "");
Expect(1, 3884, '\P{Is_Nv=1.500}', "");
Expect(0, 3884, '\P{^Is_Nv=1.500}', "");
Error('\p{Numeric_Value=_	+000000003/00004/a/}');
Error('\P{Numeric_Value=_	+000000003/00004/a/}');
Expect(1, 65912, '\p{Numeric_Value:	00003/00000004}', "");
Expect(0, 65912, '\p{^Numeric_Value:	00003/00000004}', "");
Expect(0, 65912, '\P{Numeric_Value:	00003/00000004}', "");
Expect(1, 65912, '\P{^Numeric_Value:	00003/00000004}', "");
Expect(0, 65913, '\p{Numeric_Value:	00003/00000004}', "");
Expect(1, 65913, '\p{^Numeric_Value:	00003/00000004}', "");
Expect(1, 65913, '\P{Numeric_Value:	00003/00000004}', "");
Expect(0, 65913, '\P{^Numeric_Value:	00003/00000004}', "");
Expect(1, 65912, '\p{Numeric_Value=0.750}', "");
Expect(0, 65912, '\p{^Numeric_Value=0.750}', "");
Expect(0, 65912, '\P{Numeric_Value=0.750}', "");
Expect(1, 65912, '\P{^Numeric_Value=0.750}', "");
Expect(0, 65913, '\p{Numeric_Value=0.750}', "");
Expect(1, 65913, '\p{^Numeric_Value=0.750}', "");
Expect(1, 65913, '\P{Numeric_Value=0.750}', "");
Expect(0, 65913, '\P{^Numeric_Value=0.750}', "");
Error('\p{Nv=--000000003/4:=}');
Error('\P{Nv=--000000003/4:=}');
Expect(1, 65912, '\p{Nv=+00000003/0000004}', "");
Expect(0, 65912, '\p{^Nv=+00000003/0000004}', "");
Expect(0, 65912, '\P{Nv=+00000003/0000004}', "");
Expect(1, 65912, '\P{^Nv=+00000003/0000004}', "");
Expect(0, 65913, '\p{Nv=+00000003/0000004}', "");
Expect(1, 65913, '\p{^Nv=+00000003/0000004}', "");
Expect(1, 65913, '\P{Nv=+00000003/0000004}', "");
Expect(0, 65913, '\P{^Nv=+00000003/0000004}', "");
Expect(1, 65912, '\p{Nv:0.750}', "");
Expect(0, 65912, '\p{^Nv:0.750}', "");
Expect(0, 65912, '\P{Nv:0.750}', "");
Expect(1, 65912, '\P{^Nv:0.750}', "");
Expect(0, 65913, '\p{Nv:0.750}', "");
Expect(1, 65913, '\p{^Nv:0.750}', "");
Expect(1, 65913, '\P{Nv:0.750}', "");
Expect(0, 65913, '\P{^Nv:0.750}', "");
Error('\p{Is_Numeric_Value=-/a/+3/00000004}');
Error('\P{Is_Numeric_Value=-/a/+3/00000004}');
Expect(1, 65912, '\p{Is_Numeric_Value=3/4}', "");
Expect(0, 65912, '\p{^Is_Numeric_Value=3/4}', "");
Expect(0, 65912, '\P{Is_Numeric_Value=3/4}', "");
Expect(1, 65912, '\P{^Is_Numeric_Value=3/4}', "");
Expect(0, 65913, '\p{Is_Numeric_Value=3/4}', "");
Expect(1, 65913, '\p{^Is_Numeric_Value=3/4}', "");
Expect(1, 65913, '\P{Is_Numeric_Value=3/4}', "");
Expect(0, 65913, '\P{^Is_Numeric_Value=3/4}', "");
Expect(1, 65912, '\p{Is_Numeric_Value:0.750}', "");
Expect(0, 65912, '\p{^Is_Numeric_Value:0.750}', "");
Expect(0, 65912, '\P{Is_Numeric_Value:0.750}', "");
Expect(1, 65912, '\P{^Is_Numeric_Value:0.750}', "");
Expect(0, 65913, '\p{Is_Numeric_Value:0.750}', "");
Expect(1, 65913, '\p{^Is_Numeric_Value:0.750}', "");
Expect(1, 65913, '\P{Is_Numeric_Value:0.750}', "");
Expect(0, 65913, '\P{^Is_Numeric_Value:0.750}', "");
Error('\p{Is_Nv=-/a/00003/000000004}');
Error('\P{Is_Nv=-/a/00003/000000004}');
Expect(1, 65912, '\p{Is_Nv:	0000000003/04}', "");
Expect(0, 65912, '\p{^Is_Nv:	0000000003/04}', "");
Expect(0, 65912, '\P{Is_Nv:	0000000003/04}', "");
Expect(1, 65912, '\P{^Is_Nv:	0000000003/04}', "");
Expect(0, 65913, '\p{Is_Nv:	0000000003/04}', "");
Expect(1, 65913, '\p{^Is_Nv:	0000000003/04}', "");
Expect(1, 65913, '\P{Is_Nv:	0000000003/04}', "");
Expect(0, 65913, '\P{^Is_Nv:	0000000003/04}', "");
Expect(1, 65912, '\p{Is_Nv:0.750}', "");
Expect(0, 65912, '\p{^Is_Nv:0.750}', "");
Expect(0, 65912, '\P{Is_Nv:0.750}', "");
Expect(1, 65912, '\P{^Is_Nv:0.750}', "");
Expect(0, 65913, '\p{Is_Nv:0.750}', "");
Expect(1, 65913, '\p{^Is_Nv:0.750}', "");
Expect(1, 65913, '\P{Is_Nv:0.750}', "");
Expect(0, 65913, '\P{^Is_Nv:0.750}', "");
Error('\p{Numeric_Value=	/a/+000000003/00005}');
Error('\P{Numeric_Value=	/a/+000000003/00005}');
Expect(1, 8535, '\p{Numeric_Value=00000003/0000005}', "");
Expect(0, 8535, '\p{^Numeric_Value=00000003/0000005}', "");
Expect(0, 8535, '\P{Numeric_Value=00000003/0000005}', "");
Expect(1, 8535, '\P{^Numeric_Value=00000003/0000005}', "");
Expect(0, 8536, '\p{Numeric_Value=00000003/0000005}', "");
Expect(1, 8536, '\p{^Numeric_Value=00000003/0000005}', "");
Expect(1, 8536, '\P{Numeric_Value=00000003/0000005}', "");
Expect(0, 8536, '\P{^Numeric_Value=00000003/0000005}', "");
Expect(1, 8535, '\p{Numeric_Value=0.600}', "");
Expect(0, 8535, '\p{^Numeric_Value=0.600}', "");
Expect(0, 8535, '\P{Numeric_Value=0.600}', "");
Expect(1, 8535, '\P{^Numeric_Value=0.600}', "");
Expect(0, 8536, '\p{Numeric_Value=0.600}', "");
Expect(1, 8536, '\p{^Numeric_Value=0.600}', "");
Expect(1, 8536, '\P{Numeric_Value=0.600}', "");
Expect(0, 8536, '\P{^Numeric_Value=0.600}', "");
Error('\p{Nv=003/05:=}');
Error('\P{Nv=003/05:=}');
Expect(1, 8535, '\p{Nv=0000003/00005}', "");
Expect(0, 8535, '\p{^Nv=0000003/00005}', "");
Expect(0, 8535, '\P{Nv=0000003/00005}', "");
Expect(1, 8535, '\P{^Nv=0000003/00005}', "");
Expect(0, 8536, '\p{Nv=0000003/00005}', "");
Expect(1, 8536, '\p{^Nv=0000003/00005}', "");
Expect(1, 8536, '\P{Nv=0000003/00005}', "");
Expect(0, 8536, '\P{^Nv=0000003/00005}', "");
Expect(1, 8535, '\p{Nv=0.600}', "");
Expect(0, 8535, '\p{^Nv=0.600}', "");
Expect(0, 8535, '\P{Nv=0.600}', "");
Expect(1, 8535, '\P{^Nv=0.600}', "");
Expect(0, 8536, '\p{Nv=0.600}', "");
Expect(1, 8536, '\p{^Nv=0.600}', "");
Expect(1, 8536, '\P{Nv=0.600}', "");
Expect(0, 8536, '\P{^Nv=0.600}', "");
Error('\p{Is_Numeric_Value=_/a/00000003/000000005}');
Error('\P{Is_Numeric_Value=_/a/00000003/000000005}');
Expect(1, 8535, '\p{Is_Numeric_Value=00000003/00000005}', "");
Expect(0, 8535, '\p{^Is_Numeric_Value=00000003/00000005}', "");
Expect(0, 8535, '\P{Is_Numeric_Value=00000003/00000005}', "");
Expect(1, 8535, '\P{^Is_Numeric_Value=00000003/00000005}', "");
Expect(0, 8536, '\p{Is_Numeric_Value=00000003/00000005}', "");
Expect(1, 8536, '\p{^Is_Numeric_Value=00000003/00000005}', "");
Expect(1, 8536, '\P{Is_Numeric_Value=00000003/00000005}', "");
Expect(0, 8536, '\P{^Is_Numeric_Value=00000003/00000005}', "");
Expect(1, 8535, '\p{Is_Numeric_Value=0.600}', "");
Expect(0, 8535, '\p{^Is_Numeric_Value=0.600}', "");
Expect(0, 8535, '\P{Is_Numeric_Value=0.600}', "");
Expect(1, 8535, '\P{^Is_Numeric_Value=0.600}', "");
Expect(0, 8536, '\p{Is_Numeric_Value=0.600}', "");
Expect(1, 8536, '\p{^Is_Numeric_Value=0.600}', "");
Expect(1, 8536, '\P{Is_Numeric_Value=0.600}', "");
Expect(0, 8536, '\P{^Is_Numeric_Value=0.600}', "");
Error('\p{Is_Nv=:= 0003/000000005}');
Error('\P{Is_Nv=:= 0003/000000005}');
Expect(1, 8535, '\p{Is_Nv=3/0000005}', "");
Expect(0, 8535, '\p{^Is_Nv=3/0000005}', "");
Expect(0, 8535, '\P{Is_Nv=3/0000005}', "");
Expect(1, 8535, '\P{^Is_Nv=3/0000005}', "");
Expect(0, 8536, '\p{Is_Nv=3/0000005}', "");
Expect(1, 8536, '\p{^Is_Nv=3/0000005}', "");
Expect(1, 8536, '\P{Is_Nv=3/0000005}', "");
Expect(0, 8536, '\P{^Is_Nv=3/0000005}', "");
Expect(1, 8535, '\p{Is_Nv=0.600}', "");
Expect(0, 8535, '\p{^Is_Nv=0.600}', "");
Expect(0, 8535, '\P{Is_Nv=0.600}', "");
Expect(1, 8535, '\P{^Is_Nv=0.600}', "");
Expect(0, 8536, '\p{Is_Nv=0.600}', "");
Expect(1, 8536, '\p{^Is_Nv=0.600}', "");
Expect(1, 8536, '\P{Is_Nv=0.600}', "");
Expect(0, 8536, '\P{^Is_Nv=0.600}', "");
Error('\p{Numeric_Value=	/a/+003/8}');
Error('\P{Numeric_Value=	/a/+003/8}');
Expect(1, 8540, '\p{Numeric_Value=0000000003/08}', "");
Expect(0, 8540, '\p{^Numeric_Value=0000000003/08}', "");
Expect(0, 8540, '\P{Numeric_Value=0000000003/08}', "");
Expect(1, 8540, '\P{^Numeric_Value=0000000003/08}', "");
Expect(0, 8541, '\p{Numeric_Value=0000000003/08}', "");
Expect(1, 8541, '\p{^Numeric_Value=0000000003/08}', "");
Expect(1, 8541, '\P{Numeric_Value=0000000003/08}', "");
Expect(0, 8541, '\P{^Numeric_Value=0000000003/08}', "");
Error('\p{Numeric_Value=0.38}');
Error('\P{Numeric_Value=0.38}');
Expect(1, 8540, '\p{Numeric_Value=0.375}', "");
Expect(0, 8540, '\p{^Numeric_Value=0.375}', "");
Expect(0, 8540, '\P{Numeric_Value=0.375}', "");
Expect(1, 8540, '\P{^Numeric_Value=0.375}', "");
Expect(0, 8541, '\p{Numeric_Value=0.375}', "");
Expect(1, 8541, '\p{^Numeric_Value=0.375}', "");
Expect(1, 8541, '\P{Numeric_Value=0.375}', "");
Expect(0, 8541, '\P{^Numeric_Value=0.375}', "");
Error('\p{Nv=_:=0000003/0000008}');
Error('\P{Nv=_:=0000003/0000008}');
Expect(1, 8540, '\p{Nv=03/008}', "");
Expect(0, 8540, '\p{^Nv=03/008}', "");
Expect(0, 8540, '\P{Nv=03/008}', "");
Expect(1, 8540, '\P{^Nv=03/008}', "");
Expect(0, 8541, '\p{Nv=03/008}', "");
Expect(1, 8541, '\p{^Nv=03/008}', "");
Expect(1, 8541, '\P{Nv=03/008}', "");
Expect(0, 8541, '\P{^Nv=03/008}', "");
Error('\p{Nv:0.38}');
Error('\P{Nv:0.38}');
Expect(1, 8540, '\p{Nv=0.375}', "");
Expect(0, 8540, '\p{^Nv=0.375}', "");
Expect(0, 8540, '\P{Nv=0.375}', "");
Expect(1, 8540, '\P{^Nv=0.375}', "");
Expect(0, 8541, '\p{Nv=0.375}', "");
Expect(1, 8541, '\p{^Nv=0.375}', "");
Expect(1, 8541, '\P{Nv=0.375}', "");
Expect(0, 8541, '\P{^Nv=0.375}', "");
Error('\p{Is_Numeric_Value=_:=+003/8}');
Error('\P{Is_Numeric_Value=_:=+003/8}');
Expect(1, 8540, '\p{Is_Numeric_Value=3/008}', "");
Expect(0, 8540, '\p{^Is_Numeric_Value=3/008}', "");
Expect(0, 8540, '\P{Is_Numeric_Value=3/008}', "");
Expect(1, 8540, '\P{^Is_Numeric_Value=3/008}', "");
Expect(0, 8541, '\p{Is_Numeric_Value=3/008}', "");
Expect(1, 8541, '\p{^Is_Numeric_Value=3/008}', "");
Expect(1, 8541, '\P{Is_Numeric_Value=3/008}', "");
Expect(0, 8541, '\P{^Is_Numeric_Value=3/008}', "");
Error('\p{Is_Numeric_Value: 0.38}');
Error('\P{Is_Numeric_Value: 0.38}');
Expect(1, 8540, '\p{Is_Numeric_Value=0.375}', "");
Expect(0, 8540, '\p{^Is_Numeric_Value=0.375}', "");
Expect(0, 8540, '\P{Is_Numeric_Value=0.375}', "");
Expect(1, 8540, '\P{^Is_Numeric_Value=0.375}', "");
Expect(0, 8541, '\p{Is_Numeric_Value=0.375}', "");
Expect(1, 8541, '\p{^Is_Numeric_Value=0.375}', "");
Expect(1, 8541, '\P{Is_Numeric_Value=0.375}', "");
Expect(0, 8541, '\P{^Is_Numeric_Value=0.375}', "");
Error('\p{Is_Nv=	:=0000003/0000008}');
Error('\P{Is_Nv=	:=0000003/0000008}');
Expect(1, 8540, '\p{Is_Nv=00000003/0000008}', "");
Expect(0, 8540, '\p{^Is_Nv=00000003/0000008}', "");
Expect(0, 8540, '\P{Is_Nv=00000003/0000008}', "");
Expect(1, 8540, '\P{^Is_Nv=00000003/0000008}', "");
Expect(0, 8541, '\p{Is_Nv=00000003/0000008}', "");
Expect(1, 8541, '\p{^Is_Nv=00000003/0000008}', "");
Expect(1, 8541, '\P{Is_Nv=00000003/0000008}', "");
Expect(0, 8541, '\P{^Is_Nv=00000003/0000008}', "");
Error('\p{Is_Nv=0.38}');
Error('\P{Is_Nv=0.38}');
Expect(1, 8540, '\p{Is_Nv=0.375}', "");
Expect(0, 8540, '\p{^Is_Nv=0.375}', "");
Expect(0, 8540, '\P{Is_Nv=0.375}', "");
Expect(1, 8540, '\P{^Is_Nv=0.375}', "");
Expect(0, 8541, '\p{Is_Nv=0.375}', "");
Expect(1, 8541, '\p{^Is_Nv=0.375}', "");
Expect(1, 8541, '\P{Is_Nv=0.375}', "");
Expect(0, 8541, '\P{^Is_Nv=0.375}', "");
Error('\p{Numeric_Value=/a/- 00000000030}');
Error('\P{Numeric_Value=/a/- 00000000030}');
Expect(1, 133507, '\p{Numeric_Value=+0_0_0_0_0_00030}', "");
Expect(0, 133507, '\p{^Numeric_Value=+0_0_0_0_0_00030}', "");
Expect(0, 133507, '\P{Numeric_Value=+0_0_0_0_0_00030}', "");
Expect(1, 133507, '\P{^Numeric_Value=+0_0_0_0_0_00030}', "");
Expect(0, 133508, '\p{Numeric_Value=+0_0_0_0_0_00030}', "");
Expect(1, 133508, '\p{^Numeric_Value=+0_0_0_0_0_00030}', "");
Expect(1, 133508, '\P{Numeric_Value=+0_0_0_0_0_00030}', "");
Expect(0, 133508, '\P{^Numeric_Value=+0_0_0_0_0_00030}', "");
Error('\p{Nv=:=	_0_0_0_030}');
Error('\P{Nv=:=	_0_0_0_030}');
Expect(1, 133507, '\p{Nv=0_0_0_0_0_0_30}', "");
Expect(0, 133507, '\p{^Nv=0_0_0_0_0_0_30}', "");
Expect(0, 133507, '\P{Nv=0_0_0_0_0_0_30}', "");
Expect(1, 133507, '\P{^Nv=0_0_0_0_0_0_30}', "");
Expect(0, 133508, '\p{Nv=0_0_0_0_0_0_30}', "");
Expect(1, 133508, '\p{^Nv=0_0_0_0_0_0_30}', "");
Expect(1, 133508, '\P{Nv=0_0_0_0_0_0_30}', "");
Expect(0, 133508, '\P{^Nv=0_0_0_0_0_0_30}', "");
Error('\p{Is_Numeric_Value=-:=0_0_0_0_0030}');
Error('\P{Is_Numeric_Value=-:=0_0_0_0_0030}');
Expect(1, 133507, '\p{Is_Numeric_Value:   0_0_0_0_0_0_30}', "");
Expect(0, 133507, '\p{^Is_Numeric_Value:   0_0_0_0_0_0_30}', "");
Expect(0, 133507, '\P{Is_Numeric_Value:   0_0_0_0_0_0_30}', "");
Expect(1, 133507, '\P{^Is_Numeric_Value:   0_0_0_0_0_0_30}', "");
Expect(0, 133508, '\p{Is_Numeric_Value:   0_0_0_0_0_0_30}', "");
Expect(1, 133508, '\p{^Is_Numeric_Value:   0_0_0_0_0_0_30}', "");
Expect(1, 133508, '\P{Is_Numeric_Value:   0_0_0_0_0_0_30}', "");
Expect(0, 133508, '\P{^Is_Numeric_Value:   0_0_0_0_0_0_30}', "");
Error('\p{Is_Nv=	+00000030:=}');
Error('\P{Is_Nv=	+00000030:=}');
Expect(1, 133507, '\p{Is_Nv=00000030}', "");
Expect(0, 133507, '\p{^Is_Nv=00000030}', "");
Expect(0, 133507, '\P{Is_Nv=00000030}', "");
Expect(1, 133507, '\P{^Is_Nv=00000030}', "");
Expect(0, 133508, '\p{Is_Nv=00000030}', "");
Expect(1, 133508, '\p{^Is_Nv=00000030}', "");
Expect(1, 133508, '\P{Is_Nv=00000030}', "");
Expect(0, 133508, '\P{^Is_Nv=00000030}', "");
Error('\p{Numeric_Value=:=	00300}');
Error('\P{Numeric_Value=:=	00300}');
Expect(1, 69236, '\p{Numeric_Value:   +00000000300}', "");
Expect(0, 69236, '\p{^Numeric_Value:   +00000000300}', "");
Expect(0, 69236, '\P{Numeric_Value:   +00000000300}', "");
Expect(1, 69236, '\P{^Numeric_Value:   +00000000300}', "");
Expect(0, 69237, '\p{Numeric_Value:   +00000000300}', "");
Expect(1, 69237, '\p{^Numeric_Value:   +00000000300}', "");
Expect(1, 69237, '\P{Numeric_Value:   +00000000300}', "");
Expect(0, 69237, '\P{^Numeric_Value:   +00000000300}', "");
Error('\p{Nv= -00_00_30_0/a/}');
Error('\P{Nv= -00_00_30_0/a/}');
Expect(1, 69236, '\p{Nv=+00_00_03_00}', "");
Expect(0, 69236, '\p{^Nv=+00_00_03_00}', "");
Expect(0, 69236, '\P{Nv=+00_00_03_00}', "");
Expect(1, 69236, '\P{^Nv=+00_00_03_00}', "");
Expect(0, 69237, '\p{Nv=+00_00_03_00}', "");
Expect(1, 69237, '\p{^Nv=+00_00_03_00}', "");
Expect(1, 69237, '\P{Nv=+00_00_03_00}', "");
Expect(0, 69237, '\P{^Nv=+00_00_03_00}', "");
Error('\p{Is_Numeric_Value= /a/0300}');
Error('\P{Is_Numeric_Value= /a/0300}');
Expect(1, 69236, '\p{Is_Numeric_Value=+0_0_0_0_0_00300}', "");
Expect(0, 69236, '\p{^Is_Numeric_Value=+0_0_0_0_0_00300}', "");
Expect(0, 69236, '\P{Is_Numeric_Value=+0_0_0_0_0_00300}', "");
Expect(1, 69236, '\P{^Is_Numeric_Value=+0_0_0_0_0_00300}', "");
Expect(0, 69237, '\p{Is_Numeric_Value=+0_0_0_0_0_00300}', "");
Expect(1, 69237, '\p{^Is_Numeric_Value=+0_0_0_0_0_00300}', "");
Expect(1, 69237, '\P{Is_Numeric_Value=+0_0_0_0_0_00300}', "");
Expect(0, 69237, '\P{^Is_Numeric_Value=+0_0_0_0_0_00300}', "");
Error('\p{Is_Nv= _000000300:=}');
Error('\P{Is_Nv= _000000300:=}');
Expect(1, 69236, '\p{Is_Nv:	00000030_0}', "");
Expect(0, 69236, '\p{^Is_Nv:	00000030_0}', "");
Expect(0, 69236, '\P{Is_Nv:	00000030_0}', "");
Expect(1, 69236, '\P{^Is_Nv:	00000030_0}', "");
Expect(0, 69237, '\p{Is_Nv:	00000030_0}', "");
Expect(1, 69237, '\p{^Is_Nv:	00000030_0}', "");
Expect(1, 69237, '\P{Is_Nv:	00000030_0}', "");
Expect(0, 69237, '\P{^Is_Nv:	00000030_0}', "");
Error('\p{Numeric_Value=:=- 00000003000}');
Error('\P{Numeric_Value=:=- 00000003000}');
Expect(1, 65828, '\p{Numeric_Value=000000003000}', "");
Expect(0, 65828, '\p{^Numeric_Value=000000003000}', "");
Expect(0, 65828, '\P{Numeric_Value=000000003000}', "");
Expect(1, 65828, '\P{^Numeric_Value=000000003000}', "");
Expect(0, 65829, '\p{Numeric_Value=000000003000}', "");
Expect(1, 65829, '\p{^Numeric_Value=000000003000}', "");
Expect(1, 65829, '\P{Numeric_Value=000000003000}', "");
Expect(0, 65829, '\P{^Numeric_Value=000000003000}', "");
Error('\p{Nv=:=_00300_0}');
Error('\P{Nv=:=_00300_0}');
Expect(1, 65828, '\p{Nv=+0000000003000}', "");
Expect(0, 65828, '\p{^Nv=+0000000003000}', "");
Expect(0, 65828, '\P{Nv=+0000000003000}', "");
Expect(1, 65828, '\P{^Nv=+0000000003000}', "");
Expect(0, 65829, '\p{Nv=+0000000003000}', "");
Expect(1, 65829, '\p{^Nv=+0000000003000}', "");
Expect(1, 65829, '\P{Nv=+0000000003000}', "");
Expect(0, 65829, '\P{^Nv=+0000000003000}', "");
Error('\p{Is_Numeric_Value=/a/-003000}');
Error('\P{Is_Numeric_Value=/a/-003000}');
Expect(1, 65828, '\p{Is_Numeric_Value=3000}', "");
Expect(0, 65828, '\p{^Is_Numeric_Value=3000}', "");
Expect(0, 65828, '\P{Is_Numeric_Value=3000}', "");
Expect(1, 65828, '\P{^Is_Numeric_Value=3000}', "");
Expect(0, 65829, '\p{Is_Numeric_Value=3000}', "");
Expect(1, 65829, '\p{^Is_Numeric_Value=3000}', "");
Expect(1, 65829, '\P{Is_Numeric_Value=3000}', "");
Expect(0, 65829, '\P{^Is_Numeric_Value=3000}', "");
Error('\p{Is_Nv=	_00_00_00_00_30_00/a/}');
Error('\P{Is_Nv=	_00_00_00_00_30_00/a/}');
Expect(1, 65828, '\p{Is_Nv=000_030_00}', "");
Expect(0, 65828, '\p{^Is_Nv=000_030_00}', "");
Expect(0, 65828, '\P{Is_Nv=000_030_00}', "");
Expect(1, 65828, '\P{^Is_Nv=000_030_00}', "");
Expect(0, 65829, '\p{Is_Nv=000_030_00}', "");
Expect(1, 65829, '\p{^Is_Nv=000_030_00}', "");
Expect(1, 65829, '\P{Is_Nv=000_030_00}', "");
Expect(0, 65829, '\P{^Is_Nv=000_030_00}', "");
Error('\p{Numeric_Value=-:=+000000030000}');
Error('\P{Numeric_Value=-:=+000000030000}');
Expect(1, 65837, '\p{Numeric_Value=0000030000}', "");
Expect(0, 65837, '\p{^Numeric_Value=0000030000}', "");
Expect(0, 65837, '\P{Numeric_Value=0000030000}', "");
Expect(1, 65837, '\P{^Numeric_Value=0000030000}', "");
Expect(0, 65838, '\p{Numeric_Value=0000030000}', "");
Expect(1, 65838, '\p{^Numeric_Value=0000030000}', "");
Expect(1, 65838, '\P{Numeric_Value=0000030000}', "");
Expect(0, 65838, '\P{^Numeric_Value=0000030000}', "");
Error('\p{Nv=	-030000/a/}');
Error('\P{Nv=	-030000/a/}');
Expect(1, 65837, '\p{Nv=0000000030000}', "");
Expect(0, 65837, '\p{^Nv=0000000030000}', "");
Expect(0, 65837, '\P{Nv=0000000030000}', "");
Expect(1, 65837, '\P{^Nv=0000000030000}', "");
Expect(0, 65838, '\p{Nv=0000000030000}', "");
Expect(1, 65838, '\p{^Nv=0000000030000}', "");
Expect(1, 65838, '\P{Nv=0000000030000}', "");
Expect(0, 65838, '\P{^Nv=0000000030000}', "");
Error('\p{Is_Numeric_Value=-:=0_0_0_0_0_0_0_3_0_0_00}');
Error('\P{Is_Numeric_Value=-:=0_0_0_0_0_0_0_3_0_0_00}');
Expect(1, 65837, '\p{Is_Numeric_Value:+030000}', "");
Expect(0, 65837, '\p{^Is_Numeric_Value:+030000}', "");
Expect(0, 65837, '\P{Is_Numeric_Value:+030000}', "");
Expect(1, 65837, '\P{^Is_Numeric_Value:+030000}', "");
Expect(0, 65838, '\p{Is_Numeric_Value:+030000}', "");
Expect(1, 65838, '\p{^Is_Numeric_Value:+030000}', "");
Expect(1, 65838, '\P{Is_Numeric_Value:+030000}', "");
Expect(0, 65838, '\P{^Is_Numeric_Value:+030000}', "");
Error('\p{Is_Nv=_/a/0_0_0_0_0_0_0_0_0_3_0_0_00}');
Error('\P{Is_Nv=_/a/0_0_0_0_0_0_0_0_0_3_0_0_00}');
Expect(1, 65837, '\p{Is_Nv=000000030000}', "");
Expect(0, 65837, '\p{^Is_Nv=000000030000}', "");
Expect(0, 65837, '\P{Is_Nv=000000030000}', "");
Expect(1, 65837, '\P{^Is_Nv=000000030000}', "");
Expect(0, 65838, '\p{Is_Nv=000000030000}', "");
Expect(1, 65838, '\p{^Is_Nv=000000030000}', "");
Expect(1, 65838, '\P{Is_Nv=000000030000}', "");
Expect(0, 65838, '\P{^Is_Nv=000000030000}', "");
Error('\p{Numeric_Value:   /a/-0031}');
Error('\P{Numeric_Value:   /a/-0031}');
Expect(1, 12891, '\p{Numeric_Value=31}', "");
Expect(0, 12891, '\p{^Numeric_Value=31}', "");
Expect(0, 12891, '\P{Numeric_Value=31}', "");
Expect(1, 12891, '\P{^Numeric_Value=31}', "");
Expect(0, 12892, '\p{Numeric_Value=31}', "");
Expect(1, 12892, '\p{^Numeric_Value=31}', "");
Expect(1, 12892, '\P{Numeric_Value=31}', "");
Expect(0, 12892, '\P{^Numeric_Value=31}', "");
Error('\p{Nv=-+00000031/a/}');
Error('\P{Nv=-+00000031/a/}');
Expect(1, 12891, '\p{Nv=0_0_0_0_031}', "");
Expect(0, 12891, '\p{^Nv=0_0_0_0_031}', "");
Expect(0, 12891, '\P{Nv=0_0_0_0_031}', "");
Expect(1, 12891, '\P{^Nv=0_0_0_0_031}', "");
Expect(0, 12892, '\p{Nv=0_0_0_0_031}', "");
Expect(1, 12892, '\p{^Nv=0_0_0_0_031}', "");
Expect(1, 12892, '\P{Nv=0_0_0_0_031}', "");
Expect(0, 12892, '\P{^Nv=0_0_0_0_031}', "");
Error('\p{Is_Numeric_Value=:=  00031}');
Error('\P{Is_Numeric_Value=:=  00031}');
Expect(1, 12891, '\p{Is_Numeric_Value=0_0_0_31}', "");
Expect(0, 12891, '\p{^Is_Numeric_Value=0_0_0_31}', "");
Expect(0, 12891, '\P{Is_Numeric_Value=0_0_0_31}', "");
Expect(1, 12891, '\P{^Is_Numeric_Value=0_0_0_31}', "");
Expect(0, 12892, '\p{Is_Numeric_Value=0_0_0_31}', "");
Expect(1, 12892, '\p{^Is_Numeric_Value=0_0_0_31}', "");
Expect(1, 12892, '\P{Is_Numeric_Value=0_0_0_31}', "");
Expect(0, 12892, '\P{^Is_Numeric_Value=0_0_0_31}', "");
Error('\p{Is_Nv= /a/+00000000031}');
Error('\P{Is_Nv= /a/+00000000031}');
Expect(1, 12891, '\p{Is_Nv=+00000031}', "");
Expect(0, 12891, '\p{^Is_Nv=+00000031}', "");
Expect(0, 12891, '\P{Is_Nv=+00000031}', "");
Expect(1, 12891, '\P{^Is_Nv=+00000031}', "");
Expect(0, 12892, '\p{Is_Nv=+00000031}', "");
Expect(1, 12892, '\p{^Is_Nv=+00000031}', "");
Expect(1, 12892, '\P{Is_Nv=+00000031}', "");
Expect(0, 12892, '\P{^Is_Nv=+00000031}', "");
Error('\p{Numeric_Value=/a/	+3_2}');
Error('\P{Numeric_Value=/a/	+3_2}');
Expect(1, 12892, '\p{Numeric_Value=+0000000032}', "");
Expect(0, 12892, '\p{^Numeric_Value=+0000000032}', "");
Expect(0, 12892, '\P{Numeric_Value=+0000000032}', "");
Expect(1, 12892, '\P{^Numeric_Value=+0000000032}', "");
Expect(0, 12893, '\p{Numeric_Value=+0000000032}', "");
Expect(1, 12893, '\p{^Numeric_Value=+0000000032}', "");
Expect(1, 12893, '\P{Numeric_Value=+0000000032}', "");
Expect(0, 12893, '\P{^Numeric_Value=+0000000032}', "");
Error('\p{Nv=:= 	03_2}');
Error('\P{Nv=:= 	03_2}');
Expect(1, 12892, '\p{Nv=+3_2}', "");
Expect(0, 12892, '\p{^Nv=+3_2}', "");
Expect(0, 12892, '\P{Nv=+3_2}', "");
Expect(1, 12892, '\P{^Nv=+3_2}', "");
Expect(0, 12893, '\p{Nv=+3_2}', "");
Expect(1, 12893, '\p{^Nv=+3_2}', "");
Expect(1, 12893, '\P{Nv=+3_2}', "");
Expect(0, 12893, '\P{^Nv=+3_2}', "");
Error('\p{Is_Numeric_Value:	 /a/+0_0_0_32}');
Error('\P{Is_Numeric_Value:	 /a/+0_0_0_32}');
Expect(1, 12892, '\p{Is_Numeric_Value=0000_0003_2}', "");
Expect(0, 12892, '\p{^Is_Numeric_Value=0000_0003_2}', "");
Expect(0, 12892, '\P{Is_Numeric_Value=0000_0003_2}', "");
Expect(1, 12892, '\P{^Is_Numeric_Value=0000_0003_2}', "");
Expect(0, 12893, '\p{Is_Numeric_Value=0000_0003_2}', "");
Expect(1, 12893, '\p{^Is_Numeric_Value=0000_0003_2}', "");
Expect(1, 12893, '\P{Is_Numeric_Value=0000_0003_2}', "");
Expect(0, 12893, '\P{^Is_Numeric_Value=0000_0003_2}', "");
Error('\p{Is_Nv=-/a/03_2}');
Error('\P{Is_Nv=-/a/03_2}');
Expect(1, 12892, '\p{Is_Nv=0_0_0_0_032}', "");
Expect(0, 12892, '\p{^Is_Nv=0_0_0_0_032}', "");
Expect(0, 12892, '\P{Is_Nv=0_0_0_0_032}', "");
Expect(1, 12892, '\P{^Is_Nv=0_0_0_0_032}', "");
Expect(0, 12893, '\p{Is_Nv=0_0_0_0_032}', "");
Expect(1, 12893, '\p{^Is_Nv=0_0_0_0_032}', "");
Expect(1, 12893, '\P{Is_Nv=0_0_0_0_032}', "");
Expect(0, 12893, '\P{^Is_Nv=0_0_0_0_032}', "");
Error('\p{Numeric_Value=/a/0000000033}');
Error('\P{Numeric_Value=/a/0000000033}');
Expect(1, 12893, '\p{Numeric_Value:033}', "");
Expect(0, 12893, '\p{^Numeric_Value:033}', "");
Expect(0, 12893, '\P{Numeric_Value:033}', "");
Expect(1, 12893, '\P{^Numeric_Value:033}', "");
Expect(0, 12894, '\p{Numeric_Value:033}', "");
Expect(1, 12894, '\p{^Numeric_Value:033}', "");
Expect(1, 12894, '\P{Numeric_Value:033}', "");
Expect(0, 12894, '\P{^Numeric_Value:033}', "");
Error('\p{Nv=/a/	 00033}');
Error('\P{Nv=/a/	 00033}');
Expect(1, 12893, '\p{Nv=33}', "");
Expect(0, 12893, '\p{^Nv=33}', "");
Expect(0, 12893, '\P{Nv=33}', "");
Expect(1, 12893, '\P{^Nv=33}', "");
Expect(0, 12894, '\p{Nv=33}', "");
Expect(1, 12894, '\p{^Nv=33}', "");
Expect(1, 12894, '\P{Nv=33}', "");
Expect(0, 12894, '\P{^Nv=33}', "");
Error('\p{Is_Numeric_Value:   /a/-_003_3}');
Error('\P{Is_Numeric_Value:   /a/-_003_3}');
Expect(1, 12893, '\p{Is_Numeric_Value=0_0_0_0_033}', "");
Expect(0, 12893, '\p{^Is_Numeric_Value=0_0_0_0_033}', "");
Expect(0, 12893, '\P{Is_Numeric_Value=0_0_0_0_033}', "");
Expect(1, 12893, '\P{^Is_Numeric_Value=0_0_0_0_033}', "");
Expect(0, 12894, '\p{Is_Numeric_Value=0_0_0_0_033}', "");
Expect(1, 12894, '\p{^Is_Numeric_Value=0_0_0_0_033}', "");
Expect(1, 12894, '\P{Is_Numeric_Value=0_0_0_0_033}', "");
Expect(0, 12894, '\P{^Is_Numeric_Value=0_0_0_0_033}', "");
Error('\p{Is_Nv=_:=033}');
Error('\P{Is_Nv=_:=033}');
Expect(1, 12893, '\p{Is_Nv=0000000033}', "");
Expect(0, 12893, '\p{^Is_Nv=0000000033}', "");
Expect(0, 12893, '\P{Is_Nv=0000000033}', "");
Expect(1, 12893, '\P{^Is_Nv=0000000033}', "");
Expect(0, 12894, '\p{Is_Nv=0000000033}', "");
Expect(1, 12894, '\p{^Is_Nv=0000000033}', "");
Expect(1, 12894, '\P{Is_Nv=0000000033}', "");
Expect(0, 12894, '\P{^Is_Nv=0000000033}', "");
Error('\p{Numeric_Value=	_0000034:=}');
Error('\P{Numeric_Value=	_0000034:=}');
Expect(1, 12894, '\p{Numeric_Value=00000034}', "");
Expect(0, 12894, '\p{^Numeric_Value=00000034}', "");
Expect(0, 12894, '\P{Numeric_Value=00000034}', "");
Expect(1, 12894, '\P{^Numeric_Value=00000034}', "");
Expect(0, 12895, '\p{Numeric_Value=00000034}', "");
Expect(1, 12895, '\p{^Numeric_Value=00000034}', "");
Expect(1, 12895, '\P{Numeric_Value=00000034}', "");
Expect(0, 12895, '\P{^Numeric_Value=00000034}', "");
Error('\p{Nv=:=-0003_4}');
Error('\P{Nv=:=-0003_4}');
Expect(1, 12894, '\p{Nv=+0000034}', "");
Expect(0, 12894, '\p{^Nv=+0000034}', "");
Expect(0, 12894, '\P{Nv=+0000034}', "");
Expect(1, 12894, '\P{^Nv=+0000034}', "");
Expect(0, 12895, '\p{Nv=+0000034}', "");
Expect(1, 12895, '\p{^Nv=+0000034}', "");
Expect(1, 12895, '\P{Nv=+0000034}', "");
Expect(0, 12895, '\P{^Nv=+0000034}', "");
Error('\p{Is_Numeric_Value=_:=+0000003_4}');
Error('\P{Is_Numeric_Value=_:=+0000003_4}');
Expect(1, 12894, '\p{Is_Numeric_Value=03_4}', "");
Expect(0, 12894, '\p{^Is_Numeric_Value=03_4}', "");
Expect(0, 12894, '\P{Is_Numeric_Value=03_4}', "");
Expect(1, 12894, '\P{^Is_Numeric_Value=03_4}', "");
Expect(0, 12895, '\p{Is_Numeric_Value=03_4}', "");
Expect(1, 12895, '\p{^Is_Numeric_Value=03_4}', "");
Expect(1, 12895, '\P{Is_Numeric_Value=03_4}', "");
Expect(0, 12895, '\P{^Is_Numeric_Value=03_4}', "");
Error('\p{Is_Nv:	- 0034:=}');
Error('\P{Is_Nv:	- 0034:=}');
Expect(1, 12894, '\p{Is_Nv=034}', "");
Expect(0, 12894, '\p{^Is_Nv=034}', "");
Expect(0, 12894, '\P{Is_Nv=034}', "");
Expect(1, 12894, '\P{^Is_Nv=034}', "");
Expect(0, 12895, '\p{Is_Nv=034}', "");
Expect(1, 12895, '\p{^Is_Nv=034}', "");
Expect(1, 12895, '\P{Is_Nv=034}', "");
Expect(0, 12895, '\P{^Is_Nv=034}', "");
Error('\p{Numeric_Value=/a/-00_00_00_35}');
Error('\P{Numeric_Value=/a/-00_00_00_35}');
Expect(1, 12895, '\p{Numeric_Value=000000035}', "");
Expect(0, 12895, '\p{^Numeric_Value=000000035}', "");
Expect(0, 12895, '\P{Numeric_Value=000000035}', "");
Expect(1, 12895, '\P{^Numeric_Value=000000035}', "");
Expect(0, 12896, '\p{Numeric_Value=000000035}', "");
Expect(1, 12896, '\p{^Numeric_Value=000000035}', "");
Expect(1, 12896, '\P{Numeric_Value=000000035}', "");
Expect(0, 12896, '\P{^Numeric_Value=000000035}', "");
Error('\p{Nv=  0000035:=}');
Error('\P{Nv=  0000035:=}');
Expect(1, 12895, '\p{Nv=00_00_00_035}', "");
Expect(0, 12895, '\p{^Nv=00_00_00_035}', "");
Expect(0, 12895, '\P{Nv=00_00_00_035}', "");
Expect(1, 12895, '\P{^Nv=00_00_00_035}', "");
Expect(0, 12896, '\p{Nv=00_00_00_035}', "");
Expect(1, 12896, '\p{^Nv=00_00_00_035}', "");
Expect(1, 12896, '\P{Nv=00_00_00_035}', "");
Expect(0, 12896, '\P{^Nv=00_00_00_035}', "");
Error('\p{Is_Numeric_Value=/a/	0035}');
Error('\P{Is_Numeric_Value=/a/	0035}');
Expect(1, 12895, '\p{Is_Numeric_Value=+000000035}', "");
Expect(0, 12895, '\p{^Is_Numeric_Value=+000000035}', "");
Expect(0, 12895, '\P{Is_Numeric_Value=+000000035}', "");
Expect(1, 12895, '\P{^Is_Numeric_Value=+000000035}', "");
Expect(0, 12896, '\p{Is_Numeric_Value=+000000035}', "");
Expect(1, 12896, '\p{^Is_Numeric_Value=+000000035}', "");
Expect(1, 12896, '\P{Is_Numeric_Value=+000000035}', "");
Expect(0, 12896, '\P{^Is_Numeric_Value=+000000035}', "");
Error('\p{Is_Nv=-/a/+0003_5}');
Error('\P{Is_Nv=-/a/+0003_5}');
Expect(1, 12895, '\p{Is_Nv=0003_5}', "");
Expect(0, 12895, '\p{^Is_Nv=0003_5}', "");
Expect(0, 12895, '\P{Is_Nv=0003_5}', "");
Expect(1, 12895, '\P{^Is_Nv=0003_5}', "");
Expect(0, 12896, '\p{Is_Nv=0003_5}', "");
Expect(1, 12896, '\p{^Is_Nv=0003_5}', "");
Expect(1, 12896, '\P{Is_Nv=0003_5}', "");
Expect(0, 12896, '\P{^Is_Nv=0003_5}', "");
Error('\p{Numeric_Value=_ 036/a/}');
Error('\P{Numeric_Value=_ 036/a/}');
Expect(1, 12977, '\p{Numeric_Value=00000003_6}', "");
Expect(0, 12977, '\p{^Numeric_Value=00000003_6}', "");
Expect(0, 12977, '\P{Numeric_Value=00000003_6}', "");
Expect(1, 12977, '\P{^Numeric_Value=00000003_6}', "");
Expect(0, 12978, '\p{Numeric_Value=00000003_6}', "");
Expect(1, 12978, '\p{^Numeric_Value=00000003_6}', "");
Expect(1, 12978, '\P{Numeric_Value=00000003_6}', "");
Expect(0, 12978, '\P{^Numeric_Value=00000003_6}', "");
Error('\p{Nv=/a/- 03_6}');
Error('\P{Nv=/a/- 03_6}');
Expect(1, 12977, '\p{Nv=+00036}', "");
Expect(0, 12977, '\p{^Nv=+00036}', "");
Expect(0, 12977, '\P{Nv=+00036}', "");
Expect(1, 12977, '\P{^Nv=+00036}', "");
Expect(0, 12978, '\p{Nv=+00036}', "");
Expect(1, 12978, '\p{^Nv=+00036}', "");
Expect(1, 12978, '\P{Nv=+00036}', "");
Expect(0, 12978, '\P{^Nv=+00036}', "");
Error('\p{Is_Numeric_Value=/a/-_00000036}');
Error('\P{Is_Numeric_Value=/a/-_00000036}');
Expect(1, 12977, '\p{Is_Numeric_Value=00000000036}', "");
Expect(0, 12977, '\p{^Is_Numeric_Value=00000000036}', "");
Expect(0, 12977, '\P{Is_Numeric_Value=00000000036}', "");
Expect(1, 12977, '\P{^Is_Numeric_Value=00000000036}', "");
Expect(0, 12978, '\p{Is_Numeric_Value=00000000036}', "");
Expect(1, 12978, '\p{^Is_Numeric_Value=00000000036}', "");
Expect(1, 12978, '\P{Is_Numeric_Value=00000000036}', "");
Expect(0, 12978, '\P{^Is_Numeric_Value=00000000036}', "");
Error('\p{Is_Nv= 000_000_36:=}');
Error('\P{Is_Nv= 000_000_36:=}');
Expect(1, 12977, '\p{Is_Nv=0036}', "");
Expect(0, 12977, '\p{^Is_Nv=0036}', "");
Expect(0, 12977, '\P{Is_Nv=0036}', "");
Expect(1, 12977, '\P{^Is_Nv=0036}', "");
Expect(0, 12978, '\p{Is_Nv=0036}', "");
Expect(1, 12978, '\p{^Is_Nv=0036}', "");
Expect(1, 12978, '\P{Is_Nv=0036}', "");
Expect(0, 12978, '\P{^Is_Nv=0036}', "");
Error('\p{Numeric_Value: 	00037:=}');
Error('\P{Numeric_Value: 	00037:=}');
Expect(1, 12978, '\p{Numeric_Value=0_0_37}', "");
Expect(0, 12978, '\p{^Numeric_Value=0_0_37}', "");
Expect(0, 12978, '\P{Numeric_Value=0_0_37}', "");
Expect(1, 12978, '\P{^Numeric_Value=0_0_37}', "");
Expect(0, 12979, '\p{Numeric_Value=0_0_37}', "");
Expect(1, 12979, '\p{^Numeric_Value=0_0_37}', "");
Expect(1, 12979, '\P{Numeric_Value=0_0_37}', "");
Expect(0, 12979, '\P{^Numeric_Value=0_0_37}', "");
Error('\p{Nv=-37/a/}');
Error('\P{Nv=-37/a/}');
Expect(1, 12978, '\p{Nv=+0000003_7}', "");
Expect(0, 12978, '\p{^Nv=+0000003_7}', "");
Expect(0, 12978, '\P{Nv=+0000003_7}', "");
Expect(1, 12978, '\P{^Nv=+0000003_7}', "");
Expect(0, 12979, '\p{Nv=+0000003_7}', "");
Expect(1, 12979, '\p{^Nv=+0000003_7}', "");
Expect(1, 12979, '\P{Nv=+0000003_7}', "");
Expect(0, 12979, '\P{^Nv=+0000003_7}', "");
Error('\p{Is_Numeric_Value=_/a/+000000037}');
Error('\P{Is_Numeric_Value=_/a/+000000037}');
Expect(1, 12978, '\p{Is_Numeric_Value=00037}', "");
Expect(0, 12978, '\p{^Is_Numeric_Value=00037}', "");
Expect(0, 12978, '\P{Is_Numeric_Value=00037}', "");
Expect(1, 12978, '\P{^Is_Numeric_Value=00037}', "");
Expect(0, 12979, '\p{Is_Numeric_Value=00037}', "");
Expect(1, 12979, '\p{^Is_Numeric_Value=00037}', "");
Expect(1, 12979, '\P{Is_Numeric_Value=00037}', "");
Expect(0, 12979, '\P{^Is_Numeric_Value=00037}', "");
Error('\p{Is_Nv= _000037:=}');
Error('\P{Is_Nv= _000037:=}');
Expect(1, 12978, '\p{Is_Nv=03_7}', "");
Expect(0, 12978, '\p{^Is_Nv=03_7}', "");
Expect(0, 12978, '\P{Is_Nv=03_7}', "");
Expect(1, 12978, '\P{^Is_Nv=03_7}', "");
Expect(0, 12979, '\p{Is_Nv=03_7}', "");
Expect(1, 12979, '\p{^Is_Nv=03_7}', "");
Expect(1, 12979, '\P{Is_Nv=03_7}', "");
Expect(0, 12979, '\P{^Is_Nv=03_7}', "");
Error('\p{Numeric_Value= -0003_8:=}');
Error('\P{Numeric_Value= -0003_8:=}');
Expect(1, 12979, '\p{Numeric_Value=+0_0_0_0_0_038}', "");
Expect(0, 12979, '\p{^Numeric_Value=+0_0_0_0_0_038}', "");
Expect(0, 12979, '\P{Numeric_Value=+0_0_0_0_0_038}', "");
Expect(1, 12979, '\P{^Numeric_Value=+0_0_0_0_0_038}', "");
Expect(0, 12980, '\p{Numeric_Value=+0_0_0_0_0_038}', "");
Expect(1, 12980, '\p{^Numeric_Value=+0_0_0_0_0_038}', "");
Expect(1, 12980, '\P{Numeric_Value=+0_0_0_0_0_038}', "");
Expect(0, 12980, '\P{^Numeric_Value=+0_0_0_0_0_038}', "");
Error('\p{Nv=-	+0_0_0_0_0_0_0_0_38:=}');
Error('\P{Nv=-	+0_0_0_0_0_0_0_0_38:=}');
Expect(1, 12979, '\p{Nv: 38}', "");
Expect(0, 12979, '\p{^Nv: 38}', "");
Expect(0, 12979, '\P{Nv: 38}', "");
Expect(1, 12979, '\P{^Nv: 38}', "");
Expect(0, 12980, '\p{Nv: 38}', "");
Expect(1, 12980, '\p{^Nv: 38}', "");
Expect(1, 12980, '\P{Nv: 38}', "");
Expect(0, 12980, '\P{^Nv: 38}', "");
Error('\p{Is_Numeric_Value=-	00000000038:=}');
Error('\P{Is_Numeric_Value=-	00000000038:=}');
Expect(1, 12979, '\p{Is_Numeric_Value=0_0_0_0_38}', "");
Expect(0, 12979, '\p{^Is_Numeric_Value=0_0_0_0_38}', "");
Expect(0, 12979, '\P{Is_Numeric_Value=0_0_0_0_38}', "");
Expect(1, 12979, '\P{^Is_Numeric_Value=0_0_0_0_38}', "");
Expect(0, 12980, '\p{Is_Numeric_Value=0_0_0_0_38}', "");
Expect(1, 12980, '\p{^Is_Numeric_Value=0_0_0_0_38}', "");
Expect(1, 12980, '\P{Is_Numeric_Value=0_0_0_0_38}', "");
Expect(0, 12980, '\P{^Is_Numeric_Value=0_0_0_0_38}', "");
Error('\p{Is_Nv: /a/ -00_00_00_00_038}');
Error('\P{Is_Nv: /a/ -00_00_00_00_038}');
Expect(1, 12979, '\p{Is_Nv=0_0_0_0_0_0_0_038}', "");
Expect(0, 12979, '\p{^Is_Nv=0_0_0_0_0_0_0_038}', "");
Expect(0, 12979, '\P{Is_Nv=0_0_0_0_0_0_0_038}', "");
Expect(1, 12979, '\P{^Is_Nv=0_0_0_0_0_0_0_038}', "");
Expect(0, 12980, '\p{Is_Nv=0_0_0_0_0_0_0_038}', "");
Expect(1, 12980, '\p{^Is_Nv=0_0_0_0_0_0_0_038}', "");
Expect(1, 12980, '\P{Is_Nv=0_0_0_0_0_0_0_038}', "");
Expect(0, 12980, '\P{^Is_Nv=0_0_0_0_0_0_0_038}', "");
Error('\p{Numeric_Value=	/a/000003_9}');
Error('\P{Numeric_Value=	/a/000003_9}');
Expect(1, 12980, '\p{Numeric_Value=0_0_0_0_0_0_00039}', "");
Expect(0, 12980, '\p{^Numeric_Value=0_0_0_0_0_0_00039}', "");
Expect(0, 12980, '\P{Numeric_Value=0_0_0_0_0_0_00039}', "");
Expect(1, 12980, '\P{^Numeric_Value=0_0_0_0_0_0_00039}', "");
Expect(0, 12981, '\p{Numeric_Value=0_0_0_0_0_0_00039}', "");
Expect(1, 12981, '\p{^Numeric_Value=0_0_0_0_0_0_00039}', "");
Expect(1, 12981, '\P{Numeric_Value=0_0_0_0_0_0_00039}', "");
Expect(0, 12981, '\P{^Numeric_Value=0_0_0_0_0_0_00039}', "");
Error('\p{Nv=/a/	0_0_0_0_0_039}');
Error('\P{Nv=/a/	0_0_0_0_0_039}');
Expect(1, 12980, '\p{Nv=39}', "");
Expect(0, 12980, '\p{^Nv=39}', "");
Expect(0, 12980, '\P{Nv=39}', "");
Expect(1, 12980, '\P{^Nv=39}', "");
Expect(0, 12981, '\p{Nv=39}', "");
Expect(1, 12981, '\p{^Nv=39}', "");
Expect(1, 12981, '\P{Nv=39}', "");
Expect(0, 12981, '\P{^Nv=39}', "");
Error('\p{Is_Numeric_Value=:= -+039}');
Error('\P{Is_Numeric_Value=:= -+039}');
Expect(1, 12980, '\p{Is_Numeric_Value=0000003_9}', "");
Expect(0, 12980, '\p{^Is_Numeric_Value=0000003_9}', "");
Expect(0, 12980, '\P{Is_Numeric_Value=0000003_9}', "");
Expect(1, 12980, '\P{^Is_Numeric_Value=0000003_9}', "");
Expect(0, 12981, '\p{Is_Numeric_Value=0000003_9}', "");
Expect(1, 12981, '\p{^Is_Numeric_Value=0000003_9}', "");
Expect(1, 12981, '\P{Is_Numeric_Value=0000003_9}', "");
Expect(0, 12981, '\P{^Is_Numeric_Value=0000003_9}', "");
Error('\p{Is_Nv=_/a/0039}');
Error('\P{Is_Nv=_/a/0039}');
Expect(1, 12980, '\p{Is_Nv=0_0_0_0_0_00039}', "");
Expect(0, 12980, '\p{^Is_Nv=0_0_0_0_0_00039}', "");
Expect(0, 12980, '\P{Is_Nv=0_0_0_0_0_00039}', "");
Expect(1, 12980, '\P{^Is_Nv=0_0_0_0_0_00039}', "");
Expect(0, 12981, '\p{Is_Nv=0_0_0_0_0_00039}', "");
Expect(1, 12981, '\p{^Is_Nv=0_0_0_0_0_00039}', "");
Expect(1, 12981, '\P{Is_Nv=0_0_0_0_0_00039}', "");
Expect(0, 12981, '\P{^Is_Nv=0_0_0_0_0_00039}', "");
Error('\p{Numeric_Value=	/a/0000_4}');
Error('\P{Numeric_Value=	/a/0000_4}');
Expect(1, 156269, '\p{Numeric_Value=000000004}', "");
Expect(0, 156269, '\p{^Numeric_Value=000000004}', "");
Expect(0, 156269, '\P{Numeric_Value=000000004}', "");
Expect(1, 156269, '\P{^Numeric_Value=000000004}', "");
Expect(0, 156270, '\p{Numeric_Value=000000004}', "");
Expect(1, 156270, '\p{^Numeric_Value=000000004}', "");
Expect(1, 156270, '\P{Numeric_Value=000000004}', "");
Expect(0, 156270, '\P{^Numeric_Value=000000004}', "");
Error('\p{Nv=:= 0_0_04}');
Error('\P{Nv=:= 0_0_04}');
Expect(1, 156269, '\p{Nv=04}', "");
Expect(0, 156269, '\p{^Nv=04}', "");
Expect(0, 156269, '\P{Nv=04}', "");
Expect(1, 156269, '\P{^Nv=04}', "");
Expect(0, 156270, '\p{Nv=04}', "");
Expect(1, 156270, '\p{^Nv=04}', "");
Expect(1, 156270, '\P{Nv=04}', "");
Expect(0, 156270, '\P{^Nv=04}', "");
Error('\p{Is_Numeric_Value=/a/0_0_0_0_04}');
Error('\P{Is_Numeric_Value=/a/0_0_0_0_04}');
Expect(1, 156269, '\p{Is_Numeric_Value=+00_00_00_004}', "");
Expect(0, 156269, '\p{^Is_Numeric_Value=+00_00_00_004}', "");
Expect(0, 156269, '\P{Is_Numeric_Value=+00_00_00_004}', "");
Expect(1, 156269, '\P{^Is_Numeric_Value=+00_00_00_004}', "");
Expect(0, 156270, '\p{Is_Numeric_Value=+00_00_00_004}', "");
Expect(1, 156270, '\p{^Is_Numeric_Value=+00_00_00_004}', "");
Expect(1, 156270, '\P{Is_Numeric_Value=+00_00_00_004}', "");
Expect(0, 156270, '\P{^Is_Numeric_Value=+00_00_00_004}', "");
Error('\p{Is_Nv=_/a/+000004}');
Error('\P{Is_Nv=_/a/+000004}');
Expect(1, 156269, '\p{Is_Nv:+00_00_4}', "");
Expect(0, 156269, '\p{^Is_Nv:+00_00_4}', "");
Expect(0, 156269, '\P{Is_Nv:+00_00_4}', "");
Expect(1, 156269, '\P{^Is_Nv:+00_00_4}', "");
Expect(0, 156270, '\p{Is_Nv:+00_00_4}', "");
Expect(1, 156270, '\p{^Is_Nv:+00_00_4}', "");
Expect(1, 156270, '\P{Is_Nv:+00_00_4}', "");
Expect(0, 156270, '\P{^Is_Nv:+00_00_4}', "");
Error('\p{Numeric_Value:   /a/-000000004/000005}');
Error('\P{Numeric_Value:   /a/-000000004/000005}');
Expect(1, 8536, '\p{Numeric_Value=4/00000005}', "");
Expect(0, 8536, '\p{^Numeric_Value=4/00000005}', "");
Expect(0, 8536, '\P{Numeric_Value=4/00000005}', "");
Expect(1, 8536, '\P{^Numeric_Value=4/00000005}', "");
Expect(0, 8537, '\p{Numeric_Value=4/00000005}', "");
Expect(1, 8537, '\p{^Numeric_Value=4/00000005}', "");
Expect(1, 8537, '\P{Numeric_Value=4/00000005}', "");
Expect(0, 8537, '\P{^Numeric_Value=4/00000005}', "");
Expect(1, 8536, '\p{Numeric_Value=0.800}', "");
Expect(0, 8536, '\p{^Numeric_Value=0.800}', "");
Expect(0, 8536, '\P{Numeric_Value=0.800}', "");
Expect(1, 8536, '\P{^Numeric_Value=0.800}', "");
Expect(0, 8537, '\p{Numeric_Value=0.800}', "");
Expect(1, 8537, '\p{^Numeric_Value=0.800}', "");
Expect(1, 8537, '\P{Numeric_Value=0.800}', "");
Expect(0, 8537, '\P{^Numeric_Value=0.800}', "");
Error('\p{Nv=_/a/0000000004/00005}');
Error('\P{Nv=_/a/0000000004/00005}');
Expect(1, 8536, '\p{Nv=+000000004/05}', "");
Expect(0, 8536, '\p{^Nv=+000000004/05}', "");
Expect(0, 8536, '\P{Nv=+000000004/05}', "");
Expect(1, 8536, '\P{^Nv=+000000004/05}', "");
Expect(0, 8537, '\p{Nv=+000000004/05}', "");
Expect(1, 8537, '\p{^Nv=+000000004/05}', "");
Expect(1, 8537, '\P{Nv=+000000004/05}', "");
Expect(0, 8537, '\P{^Nv=+000000004/05}', "");
Expect(1, 8536, '\p{Nv:	0.800}', "");
Expect(0, 8536, '\p{^Nv:	0.800}', "");
Expect(0, 8536, '\P{Nv:	0.800}', "");
Expect(1, 8536, '\P{^Nv:	0.800}', "");
Expect(0, 8537, '\p{Nv:	0.800}', "");
Expect(1, 8537, '\p{^Nv:	0.800}', "");
Expect(1, 8537, '\P{Nv:	0.800}', "");
Expect(0, 8537, '\P{^Nv:	0.800}', "");
Error('\p{Is_Numeric_Value=	-00004/005/a/}');
Error('\P{Is_Numeric_Value=	-00004/005/a/}');
Expect(1, 8536, '\p{Is_Numeric_Value=000000004/000005}', "");
Expect(0, 8536, '\p{^Is_Numeric_Value=000000004/000005}', "");
Expect(0, 8536, '\P{Is_Numeric_Value=000000004/000005}', "");
Expect(1, 8536, '\P{^Is_Numeric_Value=000000004/000005}', "");
Expect(0, 8537, '\p{Is_Numeric_Value=000000004/000005}', "");
Expect(1, 8537, '\p{^Is_Numeric_Value=000000004/000005}', "");
Expect(1, 8537, '\P{Is_Numeric_Value=000000004/000005}', "");
Expect(0, 8537, '\P{^Is_Numeric_Value=000000004/000005}', "");
Expect(1, 8536, '\p{Is_Numeric_Value: 0.800}', "");
Expect(0, 8536, '\p{^Is_Numeric_Value: 0.800}', "");
Expect(0, 8536, '\P{Is_Numeric_Value: 0.800}', "");
Expect(1, 8536, '\P{^Is_Numeric_Value: 0.800}', "");
Expect(0, 8537, '\p{Is_Numeric_Value: 0.800}', "");
Expect(1, 8537, '\p{^Is_Numeric_Value: 0.800}', "");
Expect(1, 8537, '\P{Is_Numeric_Value: 0.800}', "");
Expect(0, 8537, '\P{^Is_Numeric_Value: 0.800}', "");
Error('\p{Is_Nv=__0004/0005/a/}');
Error('\P{Is_Nv=__0004/0005/a/}');
Expect(1, 8536, '\p{Is_Nv=00000004/0005}', "");
Expect(0, 8536, '\p{^Is_Nv=00000004/0005}', "");
Expect(0, 8536, '\P{Is_Nv=00000004/0005}', "");
Expect(1, 8536, '\P{^Is_Nv=00000004/0005}', "");
Expect(0, 8537, '\p{Is_Nv=00000004/0005}', "");
Expect(1, 8537, '\p{^Is_Nv=00000004/0005}', "");
Expect(1, 8537, '\P{Is_Nv=00000004/0005}', "");
Expect(0, 8537, '\P{^Is_Nv=00000004/0005}', "");
Expect(1, 8536, '\p{Is_Nv:   0.800}', "");
Expect(0, 8536, '\p{^Is_Nv:   0.800}', "");
Expect(0, 8536, '\P{Is_Nv:   0.800}', "");
Expect(1, 8536, '\P{^Is_Nv:   0.800}', "");
Expect(0, 8537, '\p{Is_Nv:   0.800}', "");
Expect(1, 8537, '\p{^Is_Nv:   0.800}', "");
Expect(1, 8537, '\P{Is_Nv:   0.800}', "");
Expect(0, 8537, '\P{^Is_Nv:   0.800}', "");
Error('\p{Numeric_Value=- 000040/a/}');
Error('\P{Numeric_Value=- 000040/a/}');
Expect(1, 133532, '\p{Numeric_Value=04_0}', "");
Expect(0, 133532, '\p{^Numeric_Value=04_0}', "");
Expect(0, 133532, '\P{Numeric_Value=04_0}', "");
Expect(1, 133532, '\P{^Numeric_Value=04_0}', "");
Expect(0, 133533, '\p{Numeric_Value=04_0}', "");
Expect(1, 133533, '\p{^Numeric_Value=04_0}', "");
Expect(1, 133533, '\P{Numeric_Value=04_0}', "");
Expect(0, 133533, '\P{^Numeric_Value=04_0}', "");
Error('\p{Nv=:=- 0000040}');
Error('\P{Nv=:=- 0000040}');
Expect(1, 133532, '\p{Nv:04_0}', "");
Expect(0, 133532, '\p{^Nv:04_0}', "");
Expect(0, 133532, '\P{Nv:04_0}', "");
Expect(1, 133532, '\P{^Nv:04_0}', "");
Expect(0, 133533, '\p{Nv:04_0}', "");
Expect(1, 133533, '\p{^Nv:04_0}', "");
Expect(1, 133533, '\P{Nv:04_0}', "");
Expect(0, 133533, '\P{^Nv:04_0}', "");
Error('\p{Is_Numeric_Value=-:=00004_0}');
Error('\P{Is_Numeric_Value=-:=00004_0}');
Expect(1, 133532, '\p{Is_Numeric_Value=000000040}', "");
Expect(0, 133532, '\p{^Is_Numeric_Value=000000040}', "");
Expect(0, 133532, '\P{Is_Numeric_Value=000000040}', "");
Expect(1, 133532, '\P{^Is_Numeric_Value=000000040}', "");
Expect(0, 133533, '\p{Is_Numeric_Value=000000040}', "");
Expect(1, 133533, '\p{^Is_Numeric_Value=000000040}', "");
Expect(1, 133533, '\P{Is_Numeric_Value=000000040}', "");
Expect(0, 133533, '\P{^Is_Numeric_Value=000000040}', "");
Error('\p{Is_Nv=	/a/+0000000040}');
Error('\P{Is_Nv=	/a/+0000000040}');
Expect(1, 133532, '\p{Is_Nv:   00000040}', "");
Expect(0, 133532, '\p{^Is_Nv:   00000040}', "");
Expect(0, 133532, '\P{Is_Nv:   00000040}', "");
Expect(1, 133532, '\P{^Is_Nv:   00000040}', "");
Expect(0, 133533, '\p{Is_Nv:   00000040}', "");
Expect(1, 133533, '\p{^Is_Nv:   00000040}', "");
Expect(1, 133533, '\P{Is_Nv:   00000040}', "");
Expect(0, 133533, '\P{^Is_Nv:   00000040}', "");
Error('\p{Numeric_Value: 	 +00_00_00_00_400:=}');
Error('\P{Numeric_Value: 	 +00_00_00_00_400:=}');
Expect(1, 69237, '\p{Numeric_Value=+400}', "");
Expect(0, 69237, '\p{^Numeric_Value=+400}', "");
Expect(0, 69237, '\P{Numeric_Value=+400}', "");
Expect(1, 69237, '\P{^Numeric_Value=+400}', "");
Expect(0, 69238, '\p{Numeric_Value=+400}', "");
Expect(1, 69238, '\p{^Numeric_Value=+400}', "");
Expect(1, 69238, '\P{Numeric_Value=+400}', "");
Expect(0, 69238, '\P{^Numeric_Value=+400}', "");
Error('\p{Nv=/a/_ 040_0}');
Error('\P{Nv=/a/_ 040_0}');
Expect(1, 69237, '\p{Nv=00_00_04_00}', "");
Expect(0, 69237, '\p{^Nv=00_00_04_00}', "");
Expect(0, 69237, '\P{Nv=00_00_04_00}', "");
Expect(1, 69237, '\P{^Nv=00_00_04_00}', "");
Expect(0, 69238, '\p{Nv=00_00_04_00}', "");
Expect(1, 69238, '\p{^Nv=00_00_04_00}', "");
Expect(1, 69238, '\P{Nv=00_00_04_00}', "");
Expect(0, 69238, '\P{^Nv=00_00_04_00}', "");
Error('\p{Is_Numeric_Value=/a/0_0_0_400}');
Error('\P{Is_Numeric_Value=/a/0_0_0_400}');
Expect(1, 69237, '\p{Is_Numeric_Value:	00_04_00}', "");
Expect(0, 69237, '\p{^Is_Numeric_Value:	00_04_00}', "");
Expect(0, 69237, '\P{Is_Numeric_Value:	00_04_00}', "");
Expect(1, 69237, '\P{^Is_Numeric_Value:	00_04_00}', "");
Expect(0, 69238, '\p{Is_Numeric_Value:	00_04_00}', "");
Expect(1, 69238, '\p{^Is_Numeric_Value:	00_04_00}', "");
Expect(1, 69238, '\P{Is_Numeric_Value:	00_04_00}', "");
Expect(0, 69238, '\P{^Is_Numeric_Value:	00_04_00}', "");
Error('\p{Is_Nv=/a/-_0_0_0_0_0_0_0_4_00}');
Error('\P{Is_Nv=/a/-_0_0_0_0_0_0_0_4_00}');
Expect(1, 69237, '\p{Is_Nv=0_0_0_0_0_0_400}', "");
Expect(0, 69237, '\p{^Is_Nv=0_0_0_0_0_0_400}', "");
Expect(0, 69237, '\P{Is_Nv=0_0_0_0_0_0_400}', "");
Expect(1, 69237, '\P{^Is_Nv=0_0_0_0_0_0_400}', "");
Expect(0, 69238, '\p{Is_Nv=0_0_0_0_0_0_400}', "");
Expect(1, 69238, '\p{^Is_Nv=0_0_0_0_0_0_400}', "");
Expect(1, 69238, '\P{Is_Nv=0_0_0_0_0_0_400}', "");
Expect(0, 69238, '\P{^Is_Nv=0_0_0_0_0_0_400}', "");
Error('\p{Numeric_Value=/a/-_0_0_0_0_0_4000}');
Error('\P{Numeric_Value=/a/-_0_0_0_0_0_4000}');
Expect(1, 65829, '\p{Numeric_Value=+0_0_4_000}', "");
Expect(0, 65829, '\p{^Numeric_Value=+0_0_4_000}', "");
Expect(0, 65829, '\P{Numeric_Value=+0_0_4_000}', "");
Expect(1, 65829, '\P{^Numeric_Value=+0_0_4_000}', "");
Expect(0, 65830, '\p{Numeric_Value=+0_0_4_000}', "");
Expect(1, 65830, '\p{^Numeric_Value=+0_0_4_000}', "");
Expect(1, 65830, '\P{Numeric_Value=+0_0_4_000}', "");
Expect(0, 65830, '\P{^Numeric_Value=+0_0_4_000}', "");
Error('\p{Nv=:=--04000}');
Error('\P{Nv=:=--04000}');
Expect(1, 65829, '\p{Nv:	0000_0040_00}', "");
Expect(0, 65829, '\p{^Nv:	0000_0040_00}', "");
Expect(0, 65829, '\P{Nv:	0000_0040_00}', "");
Expect(1, 65829, '\P{^Nv:	0000_0040_00}', "");
Expect(0, 65830, '\p{Nv:	0000_0040_00}', "");
Expect(1, 65830, '\p{^Nv:	0000_0040_00}', "");
Expect(1, 65830, '\P{Nv:	0000_0040_00}', "");
Expect(0, 65830, '\P{^Nv:	0000_0040_00}', "");
Error('\p{Is_Numeric_Value:   _:=0_0_0_0_0_0_0_004000}');
Error('\P{Is_Numeric_Value:   _:=0_0_0_0_0_0_0_004000}');
Expect(1, 65829, '\p{Is_Numeric_Value=00000004000}', "");
Expect(0, 65829, '\p{^Is_Numeric_Value=00000004000}', "");
Expect(0, 65829, '\P{Is_Numeric_Value=00000004000}', "");
Expect(1, 65829, '\P{^Is_Numeric_Value=00000004000}', "");
Expect(0, 65830, '\p{Is_Numeric_Value=00000004000}', "");
Expect(1, 65830, '\p{^Is_Numeric_Value=00000004000}', "");
Expect(1, 65830, '\P{Is_Numeric_Value=00000004000}', "");
Expect(0, 65830, '\P{^Is_Numeric_Value=00000004000}', "");
Error('\p{Is_Nv=-+000000004000/a/}');
Error('\P{Is_Nv=-+000000004000/a/}');
Expect(1, 65829, '\p{Is_Nv=0_0_0_0_0_0_0_4_000}', "");
Expect(0, 65829, '\p{^Is_Nv=0_0_0_0_0_0_0_4_000}', "");
Expect(0, 65829, '\P{Is_Nv=0_0_0_0_0_0_0_4_000}', "");
Expect(1, 65829, '\P{^Is_Nv=0_0_0_0_0_0_0_4_000}', "");
Expect(0, 65830, '\p{Is_Nv=0_0_0_0_0_0_0_4_000}', "");
Expect(1, 65830, '\p{^Is_Nv=0_0_0_0_0_0_0_4_000}', "");
Expect(1, 65830, '\P{Is_Nv=0_0_0_0_0_0_0_4_000}', "");
Expect(0, 65830, '\P{^Is_Nv=0_0_0_0_0_0_0_4_000}', "");
Error('\p{Numeric_Value=/a/+00000040000}');
Error('\P{Numeric_Value=/a/+00000040000}');
Expect(1, 65838, '\p{Numeric_Value=0000000040000}', "");
Expect(0, 65838, '\p{^Numeric_Value=0000000040000}', "");
Expect(0, 65838, '\P{Numeric_Value=0000000040000}', "");
Expect(1, 65838, '\P{^Numeric_Value=0000000040000}', "");
Expect(0, 65839, '\p{Numeric_Value=0000000040000}', "");
Expect(1, 65839, '\p{^Numeric_Value=0000000040000}', "");
Expect(1, 65839, '\P{Numeric_Value=0000000040000}', "");
Expect(0, 65839, '\P{^Numeric_Value=0000000040000}', "");
Error('\p{Nv=_/a/4000_0}');
Error('\P{Nv=_/a/4000_0}');
Expect(1, 65838, '\p{Nv=00000000040000}', "");
Expect(0, 65838, '\p{^Nv=00000000040000}', "");
Expect(0, 65838, '\P{Nv=00000000040000}', "");
Expect(1, 65838, '\P{^Nv=00000000040000}', "");
Expect(0, 65839, '\p{Nv=00000000040000}', "");
Expect(1, 65839, '\p{^Nv=00000000040000}', "");
Expect(1, 65839, '\P{Nv=00000000040000}', "");
Expect(0, 65839, '\P{^Nv=00000000040000}', "");
Error('\p{Is_Numeric_Value=-:=00000040000}');
Error('\P{Is_Numeric_Value=-:=00000040000}');
Expect(1, 65838, '\p{Is_Numeric_Value=40_00_0}', "");
Expect(0, 65838, '\p{^Is_Numeric_Value=40_00_0}', "");
Expect(0, 65838, '\P{Is_Numeric_Value=40_00_0}', "");
Expect(1, 65838, '\P{^Is_Numeric_Value=40_00_0}', "");
Expect(0, 65839, '\p{Is_Numeric_Value=40_00_0}', "");
Expect(1, 65839, '\p{^Is_Numeric_Value=40_00_0}', "");
Expect(1, 65839, '\P{Is_Numeric_Value=40_00_0}', "");
Expect(0, 65839, '\P{^Is_Numeric_Value=40_00_0}', "");
Error('\p{Is_Nv= :=00_00_00_00_04_0000}');
Error('\P{Is_Nv= :=00_00_00_00_04_0000}');
Expect(1, 65838, '\p{Is_Nv=+00000040000}', "");
Expect(0, 65838, '\p{^Is_Nv=+00000040000}', "");
Expect(0, 65838, '\P{Is_Nv=+00000040000}', "");
Expect(1, 65838, '\P{^Is_Nv=+00000040000}', "");
Expect(0, 65839, '\p{Is_Nv=+00000040000}', "");
Expect(1, 65839, '\p{^Is_Nv=+00000040000}', "");
Expect(1, 65839, '\P{Is_Nv=+00000040000}', "");
Expect(0, 65839, '\P{^Is_Nv=+00000040000}', "");
Error('\p{Numeric_Value: /a/__0_0_0_0_0_0041}');
Error('\P{Numeric_Value: /a/__0_0_0_0_0_0041}');
Expect(1, 12982, '\p{Numeric_Value=0_0_0_0_0_0_0041}', "");
Expect(0, 12982, '\p{^Numeric_Value=0_0_0_0_0_0_0041}', "");
Expect(0, 12982, '\P{Numeric_Value=0_0_0_0_0_0_0041}', "");
Expect(1, 12982, '\P{^Numeric_Value=0_0_0_0_0_0_0041}', "");
Expect(0, 12983, '\p{Numeric_Value=0_0_0_0_0_0_0041}', "");
Expect(1, 12983, '\p{^Numeric_Value=0_0_0_0_0_0_0041}', "");
Expect(1, 12983, '\P{Numeric_Value=0_0_0_0_0_0_0041}', "");
Expect(0, 12983, '\P{^Numeric_Value=0_0_0_0_0_0_0041}', "");
Error('\p{Nv=-	+4_1:=}');
Error('\P{Nv=-	+4_1:=}');
Expect(1, 12982, '\p{Nv=0000000041}', "");
Expect(0, 12982, '\p{^Nv=0000000041}', "");
Expect(0, 12982, '\P{Nv=0000000041}', "");
Expect(1, 12982, '\P{^Nv=0000000041}', "");
Expect(0, 12983, '\p{Nv=0000000041}', "");
Expect(1, 12983, '\p{^Nv=0000000041}', "");
Expect(1, 12983, '\P{Nv=0000000041}', "");
Expect(0, 12983, '\P{^Nv=0000000041}', "");
Error('\p{Is_Numeric_Value=__000004_1:=}');
Error('\P{Is_Numeric_Value=__000004_1:=}');
Expect(1, 12982, '\p{Is_Numeric_Value=00000000041}', "");
Expect(0, 12982, '\p{^Is_Numeric_Value=00000000041}', "");
Expect(0, 12982, '\P{Is_Numeric_Value=00000000041}', "");
Expect(1, 12982, '\P{^Is_Numeric_Value=00000000041}', "");
Expect(0, 12983, '\p{Is_Numeric_Value=00000000041}', "");
Expect(1, 12983, '\p{^Is_Numeric_Value=00000000041}', "");
Expect(1, 12983, '\P{Is_Numeric_Value=00000000041}', "");
Expect(0, 12983, '\P{^Is_Numeric_Value=00000000041}', "");
Error('\p{Is_Nv=:= 	00000041}');
Error('\P{Is_Nv=:= 	00000041}');
Expect(1, 12982, '\p{Is_Nv=0_0_0_41}', "");
Expect(0, 12982, '\p{^Is_Nv=0_0_0_41}', "");
Expect(0, 12982, '\P{Is_Nv=0_0_0_41}', "");
Expect(1, 12982, '\P{^Is_Nv=0_0_0_41}', "");
Expect(0, 12983, '\p{Is_Nv=0_0_0_41}', "");
Expect(1, 12983, '\p{^Is_Nv=0_0_0_41}', "");
Expect(1, 12983, '\P{Is_Nv=0_0_0_41}', "");
Expect(0, 12983, '\P{^Is_Nv=0_0_0_41}', "");
Error('\p{Numeric_Value= +00000042/a/}');
Error('\P{Numeric_Value= +00000042/a/}');
Expect(1, 12983, '\p{Numeric_Value=+04_2}', "");
Expect(0, 12983, '\p{^Numeric_Value=+04_2}', "");
Expect(0, 12983, '\P{Numeric_Value=+04_2}', "");
Expect(1, 12983, '\P{^Numeric_Value=+04_2}', "");
Expect(0, 12984, '\p{Numeric_Value=+04_2}', "");
Expect(1, 12984, '\p{^Numeric_Value=+04_2}', "");
Expect(1, 12984, '\P{Numeric_Value=+04_2}', "");
Expect(0, 12984, '\P{^Numeric_Value=+04_2}', "");
Error('\p{Nv=_:=+000000042}');
Error('\P{Nv=_:=+000000042}');
Expect(1, 12983, '\p{Nv:	4_2}', "");
Expect(0, 12983, '\p{^Nv:	4_2}', "");
Expect(0, 12983, '\P{Nv:	4_2}', "");
Expect(1, 12983, '\P{^Nv:	4_2}', "");
Expect(0, 12984, '\p{Nv:	4_2}', "");
Expect(1, 12984, '\p{^Nv:	4_2}', "");
Expect(1, 12984, '\P{Nv:	4_2}', "");
Expect(0, 12984, '\P{^Nv:	4_2}', "");
Error('\p{Is_Numeric_Value= :=+0_0_0_042}');
Error('\P{Is_Numeric_Value= :=+0_0_0_042}');
Expect(1, 12983, '\p{Is_Numeric_Value=0_0_42}', "");
Expect(0, 12983, '\p{^Is_Numeric_Value=0_0_42}', "");
Expect(0, 12983, '\P{Is_Numeric_Value=0_0_42}', "");
Expect(1, 12983, '\P{^Is_Numeric_Value=0_0_42}', "");
Expect(0, 12984, '\p{Is_Numeric_Value=0_0_42}', "");
Expect(1, 12984, '\p{^Is_Numeric_Value=0_0_42}', "");
Expect(1, 12984, '\P{Is_Numeric_Value=0_0_42}', "");
Expect(0, 12984, '\P{^Is_Numeric_Value=0_0_42}', "");
Error('\p{Is_Nv=:=-_+000004_2}');
Error('\P{Is_Nv=:=-_+000004_2}');
Expect(1, 12983, '\p{Is_Nv=042}', "");
Expect(0, 12983, '\p{^Is_Nv=042}', "");
Expect(0, 12983, '\P{Is_Nv=042}', "");
Expect(1, 12983, '\P{^Is_Nv=042}', "");
Expect(0, 12984, '\p{Is_Nv=042}', "");
Expect(1, 12984, '\p{^Is_Nv=042}', "");
Expect(1, 12984, '\P{Is_Nv=042}', "");
Expect(0, 12984, '\P{^Is_Nv=042}', "");
Error('\p{Numeric_Value=:=-00000000043}');
Error('\P{Numeric_Value=:=-00000000043}');
Expect(1, 12984, '\p{Numeric_Value=+43}', "");
Expect(0, 12984, '\p{^Numeric_Value=+43}', "");
Expect(0, 12984, '\P{Numeric_Value=+43}', "");
Expect(1, 12984, '\P{^Numeric_Value=+43}', "");
Expect(0, 12985, '\p{Numeric_Value=+43}', "");
Expect(1, 12985, '\p{^Numeric_Value=+43}', "");
Expect(1, 12985, '\P{Numeric_Value=+43}', "");
Expect(0, 12985, '\P{^Numeric_Value=+43}', "");
Error('\p{Nv=/a/		+00_04_3}');
Error('\P{Nv=/a/		+00_04_3}');
Expect(1, 12984, '\p{Nv=00_00_43}', "");
Expect(0, 12984, '\p{^Nv=00_00_43}', "");
Expect(0, 12984, '\P{Nv=00_00_43}', "");
Expect(1, 12984, '\P{^Nv=00_00_43}', "");
Expect(0, 12985, '\p{Nv=00_00_43}', "");
Expect(1, 12985, '\p{^Nv=00_00_43}', "");
Expect(1, 12985, '\P{Nv=00_00_43}', "");
Expect(0, 12985, '\P{^Nv=00_00_43}', "");
Error('\p{Is_Numeric_Value=_:=0004_3}');
Error('\P{Is_Numeric_Value=_:=0004_3}');
Expect(1, 12984, '\p{Is_Numeric_Value=4_3}', "");
Expect(0, 12984, '\p{^Is_Numeric_Value=4_3}', "");
Expect(0, 12984, '\P{Is_Numeric_Value=4_3}', "");
Expect(1, 12984, '\P{^Is_Numeric_Value=4_3}', "");
Expect(0, 12985, '\p{Is_Numeric_Value=4_3}', "");
Expect(1, 12985, '\p{^Is_Numeric_Value=4_3}', "");
Expect(1, 12985, '\P{Is_Numeric_Value=4_3}', "");
Expect(0, 12985, '\P{^Is_Numeric_Value=4_3}', "");
Error('\p{Is_Nv=:=	-0000043}');
Error('\P{Is_Nv=:=	-0000043}');
Expect(1, 12984, '\p{Is_Nv=+000000043}', "");
Expect(0, 12984, '\p{^Is_Nv=+000000043}', "");
Expect(0, 12984, '\P{Is_Nv=+000000043}', "");
Expect(1, 12984, '\P{^Is_Nv=+000000043}', "");
Expect(0, 12985, '\p{Is_Nv=+000000043}', "");
Expect(1, 12985, '\p{^Is_Nv=+000000043}', "");
Expect(1, 12985, '\P{Is_Nv=+000000043}', "");
Expect(0, 12985, '\P{^Is_Nv=+000000043}', "");
Error('\p{Numeric_Value=_+0000000432000:=}');
Error('\P{Numeric_Value=_+0000000432000:=}');
Expect(1, 74803, '\p{Numeric_Value: 04_32_00_0}', "");
Expect(0, 74803, '\p{^Numeric_Value: 04_32_00_0}', "");
Expect(0, 74803, '\P{Numeric_Value: 04_32_00_0}', "");
Expect(1, 74803, '\P{^Numeric_Value: 04_32_00_0}', "");
Expect(0, 74804, '\p{Numeric_Value: 04_32_00_0}', "");
Expect(1, 74804, '\p{^Numeric_Value: 04_32_00_0}', "");
Expect(1, 74804, '\P{Numeric_Value: 04_32_00_0}', "");
Expect(0, 74804, '\P{^Numeric_Value: 04_32_00_0}', "");
Error('\p{Nv=_+00043200_0:=}');
Error('\P{Nv=_+00043200_0:=}');
Expect(1, 74803, '\p{Nv=+0_4_3_2_000}', "");
Expect(0, 74803, '\p{^Nv=+0_4_3_2_000}', "");
Expect(0, 74803, '\P{Nv=+0_4_3_2_000}', "");
Expect(1, 74803, '\P{^Nv=+0_4_3_2_000}', "");
Expect(0, 74804, '\p{Nv=+0_4_3_2_000}', "");
Expect(1, 74804, '\p{^Nv=+0_4_3_2_000}', "");
Expect(1, 74804, '\P{Nv=+0_4_3_2_000}', "");
Expect(0, 74804, '\P{^Nv=+0_4_3_2_000}', "");
Error('\p{Is_Numeric_Value= /a/0000_0000_0432_000}');
Error('\P{Is_Numeric_Value= /a/0000_0000_0432_000}');
Expect(1, 74803, '\p{Is_Numeric_Value=0_0_0_0_0_0_0_0_432000}', "");
Expect(0, 74803, '\p{^Is_Numeric_Value=0_0_0_0_0_0_0_0_432000}', "");
Expect(0, 74803, '\P{Is_Numeric_Value=0_0_0_0_0_0_0_0_432000}', "");
Expect(1, 74803, '\P{^Is_Numeric_Value=0_0_0_0_0_0_0_0_432000}', "");
Expect(0, 74804, '\p{Is_Numeric_Value=0_0_0_0_0_0_0_0_432000}', "");
Expect(1, 74804, '\p{^Is_Numeric_Value=0_0_0_0_0_0_0_0_432000}', "");
Expect(1, 74804, '\P{Is_Numeric_Value=0_0_0_0_0_0_0_0_432000}', "");
Expect(0, 74804, '\P{^Is_Numeric_Value=0_0_0_0_0_0_0_0_432000}', "");
Error('\p{Is_Nv=:=	-000000000432000}');
Error('\P{Is_Nv=:=	-000000000432000}');
Expect(1, 74803, '\p{Is_Nv:	00_00_00_00_04_32_00_0}', "");
Expect(0, 74803, '\p{^Is_Nv:	00_00_00_00_04_32_00_0}', "");
Expect(0, 74803, '\P{Is_Nv:	00_00_00_00_04_32_00_0}', "");
Expect(1, 74803, '\P{^Is_Nv:	00_00_00_00_04_32_00_0}', "");
Expect(0, 74804, '\p{Is_Nv:	00_00_00_00_04_32_00_0}', "");
Expect(1, 74804, '\p{^Is_Nv:	00_00_00_00_04_32_00_0}', "");
Expect(1, 74804, '\P{Is_Nv:	00_00_00_00_04_32_00_0}', "");
Expect(0, 74804, '\P{^Is_Nv:	00_00_00_00_04_32_00_0}', "");
Error('\p{Numeric_Value= -0_0_0_0_044/a/}');
Error('\P{Numeric_Value= -0_0_0_0_044/a/}');
Expect(1, 12985, '\p{Numeric_Value=000000044}', "");
Expect(0, 12985, '\p{^Numeric_Value=000000044}', "");
Expect(0, 12985, '\P{Numeric_Value=000000044}', "");
Expect(1, 12985, '\P{^Numeric_Value=000000044}', "");
Expect(0, 12986, '\p{Numeric_Value=000000044}', "");
Expect(1, 12986, '\p{^Numeric_Value=000000044}', "");
Expect(1, 12986, '\P{Numeric_Value=000000044}', "");
Expect(0, 12986, '\P{^Numeric_Value=000000044}', "");
Error('\p{Nv:   := +00_04_4}');
Error('\P{Nv:   := +00_04_4}');
Expect(1, 12985, '\p{Nv=+0_0_0_044}', "");
Expect(0, 12985, '\p{^Nv=+0_0_0_044}', "");
Expect(0, 12985, '\P{Nv=+0_0_0_044}', "");
Expect(1, 12985, '\P{^Nv=+0_0_0_044}', "");
Expect(0, 12986, '\p{Nv=+0_0_0_044}', "");
Expect(1, 12986, '\p{^Nv=+0_0_0_044}', "");
Expect(1, 12986, '\P{Nv=+0_0_0_044}', "");
Expect(0, 12986, '\P{^Nv=+0_0_0_044}', "");
Error('\p{Is_Numeric_Value=-/a/0004_4}');
Error('\P{Is_Numeric_Value=-/a/0004_4}');
Expect(1, 12985, '\p{Is_Numeric_Value=0044}', "");
Expect(0, 12985, '\p{^Is_Numeric_Value=0044}', "");
Expect(0, 12985, '\P{Is_Numeric_Value=0044}', "");
Expect(1, 12985, '\P{^Is_Numeric_Value=0044}', "");
Expect(0, 12986, '\p{Is_Numeric_Value=0044}', "");
Expect(1, 12986, '\p{^Is_Numeric_Value=0044}', "");
Expect(1, 12986, '\P{Is_Numeric_Value=0044}', "");
Expect(0, 12986, '\P{^Is_Numeric_Value=0044}', "");
Error('\p{Is_Nv=-_0004_4:=}');
Error('\P{Is_Nv=-_0004_4:=}');
Expect(1, 12985, '\p{Is_Nv=00_00_00_04_4}', "");
Expect(0, 12985, '\p{^Is_Nv=00_00_00_04_4}', "");
Expect(0, 12985, '\P{Is_Nv=00_00_00_04_4}', "");
Expect(1, 12985, '\P{^Is_Nv=00_00_00_04_4}', "");
Expect(0, 12986, '\p{Is_Nv=00_00_00_04_4}', "");
Expect(1, 12986, '\p{^Is_Nv=00_00_00_04_4}', "");
Expect(1, 12986, '\P{Is_Nv=00_00_00_04_4}', "");
Expect(0, 12986, '\P{^Is_Nv=00_00_00_04_4}', "");
Error('\p{Numeric_Value=:=-	00045}');
Error('\P{Numeric_Value=:=-	00045}');
Expect(1, 12986, '\p{Numeric_Value=00045}', "");
Expect(0, 12986, '\p{^Numeric_Value=00045}', "");
Expect(0, 12986, '\P{Numeric_Value=00045}', "");
Expect(1, 12986, '\P{^Numeric_Value=00045}', "");
Expect(0, 12987, '\p{Numeric_Value=00045}', "");
Expect(1, 12987, '\p{^Numeric_Value=00045}', "");
Expect(1, 12987, '\P{Numeric_Value=00045}', "");
Expect(0, 12987, '\P{^Numeric_Value=00045}', "");
Error('\p{Nv=	-+0000000045:=}');
Error('\P{Nv=	-+0000000045:=}');
Expect(1, 12986, '\p{Nv=045}', "");
Expect(0, 12986, '\p{^Nv=045}', "");
Expect(0, 12986, '\P{Nv=045}', "");
Expect(1, 12986, '\P{^Nv=045}', "");
Expect(0, 12987, '\p{Nv=045}', "");
Expect(1, 12987, '\p{^Nv=045}', "");
Expect(1, 12987, '\P{Nv=045}', "");
Expect(0, 12987, '\P{^Nv=045}', "");
Error('\p{Is_Numeric_Value=/a/	-000000045}');
Error('\P{Is_Numeric_Value=/a/	-000000045}');
Expect(1, 12986, '\p{Is_Numeric_Value=000004_5}', "");
Expect(0, 12986, '\p{^Is_Numeric_Value=000004_5}', "");
Expect(0, 12986, '\P{Is_Numeric_Value=000004_5}', "");
Expect(1, 12986, '\P{^Is_Numeric_Value=000004_5}', "");
Expect(0, 12987, '\p{Is_Numeric_Value=000004_5}', "");
Expect(1, 12987, '\p{^Is_Numeric_Value=000004_5}', "");
Expect(1, 12987, '\P{Is_Numeric_Value=000004_5}', "");
Expect(0, 12987, '\P{^Is_Numeric_Value=000004_5}', "");
Error('\p{Is_Nv=_/a/00045}');
Error('\P{Is_Nv=_/a/00045}');
Expect(1, 12986, '\p{Is_Nv: 0_0_0_0_0_0_0_0_0_45}', "");
Expect(0, 12986, '\p{^Is_Nv: 0_0_0_0_0_0_0_0_0_45}', "");
Expect(0, 12986, '\P{Is_Nv: 0_0_0_0_0_0_0_0_0_45}', "");
Expect(1, 12986, '\P{^Is_Nv: 0_0_0_0_0_0_0_0_0_45}', "");
Expect(0, 12987, '\p{Is_Nv: 0_0_0_0_0_0_0_0_0_45}', "");
Expect(1, 12987, '\p{^Is_Nv: 0_0_0_0_0_0_0_0_0_45}', "");
Expect(1, 12987, '\P{Is_Nv: 0_0_0_0_0_0_0_0_0_45}', "");
Expect(0, 12987, '\P{^Is_Nv: 0_0_0_0_0_0_0_0_0_45}', "");
Error('\p{Numeric_Value=_/a/00046}');
Error('\P{Numeric_Value=_/a/00046}');
Expect(1, 12987, '\p{Numeric_Value=46}', "");
Expect(0, 12987, '\p{^Numeric_Value=46}', "");
Expect(0, 12987, '\P{Numeric_Value=46}', "");
Expect(1, 12987, '\P{^Numeric_Value=46}', "");
Expect(0, 12988, '\p{Numeric_Value=46}', "");
Expect(1, 12988, '\p{^Numeric_Value=46}', "");
Expect(1, 12988, '\P{Numeric_Value=46}', "");
Expect(0, 12988, '\P{^Numeric_Value=46}', "");
Error('\p{Nv:   :=	0_0_0_0_0046}');
Error('\P{Nv:   :=	0_0_0_0_0046}');
Expect(1, 12987, '\p{Nv=+000000046}', "");
Expect(0, 12987, '\p{^Nv=+000000046}', "");
Expect(0, 12987, '\P{Nv=+000000046}', "");
Expect(1, 12987, '\P{^Nv=+000000046}', "");
Expect(0, 12988, '\p{Nv=+000000046}', "");
Expect(1, 12988, '\p{^Nv=+000000046}', "");
Expect(1, 12988, '\P{Nv=+000000046}', "");
Expect(0, 12988, '\P{^Nv=+000000046}', "");
Error('\p{Is_Numeric_Value=	/a/+00046}');
Error('\P{Is_Numeric_Value=	/a/+00046}');
Expect(1, 12987, '\p{Is_Numeric_Value=00_00_00_00_046}', "");
Expect(0, 12987, '\p{^Is_Numeric_Value=00_00_00_00_046}', "");
Expect(0, 12987, '\P{Is_Numeric_Value=00_00_00_00_046}', "");
Expect(1, 12987, '\P{^Is_Numeric_Value=00_00_00_00_046}', "");
Expect(0, 12988, '\p{Is_Numeric_Value=00_00_00_00_046}', "");
Expect(1, 12988, '\p{^Is_Numeric_Value=00_00_00_00_046}', "");
Expect(1, 12988, '\P{Is_Numeric_Value=00_00_00_00_046}', "");
Expect(0, 12988, '\P{^Is_Numeric_Value=00_00_00_00_046}', "");
Error('\p{Is_Nv=:=00000000046}');
Error('\P{Is_Nv=:=00000000046}');
Expect(1, 12987, '\p{Is_Nv=0004_6}', "");
Expect(0, 12987, '\p{^Is_Nv=0004_6}', "");
Expect(0, 12987, '\P{Is_Nv=0004_6}', "");
Expect(1, 12987, '\P{^Is_Nv=0004_6}', "");
Expect(0, 12988, '\p{Is_Nv=0004_6}', "");
Expect(1, 12988, '\p{^Is_Nv=0004_6}', "");
Expect(1, 12988, '\P{Is_Nv=0004_6}', "");
Expect(0, 12988, '\P{^Is_Nv=0004_6}', "");
Error('\p{Numeric_Value:/a/ 	000000047}');
Error('\P{Numeric_Value:/a/ 	000000047}');
Expect(1, 12988, '\p{Numeric_Value=+0000004_7}', "");
Expect(0, 12988, '\p{^Numeric_Value=+0000004_7}', "");
Expect(0, 12988, '\P{Numeric_Value=+0000004_7}', "");
Expect(1, 12988, '\P{^Numeric_Value=+0000004_7}', "");
Expect(0, 12989, '\p{Numeric_Value=+0000004_7}', "");
Expect(1, 12989, '\p{^Numeric_Value=+0000004_7}', "");
Expect(1, 12989, '\P{Numeric_Value=+0000004_7}', "");
Expect(0, 12989, '\P{^Numeric_Value=+0000004_7}', "");
Error('\p{Nv=	_0_0_0_0_0_0047/a/}');
Error('\P{Nv=	_0_0_0_0_0_0047/a/}');
Expect(1, 12988, '\p{Nv=0047}', "");
Expect(0, 12988, '\p{^Nv=0047}', "");
Expect(0, 12988, '\P{Nv=0047}', "");
Expect(1, 12988, '\P{^Nv=0047}', "");
Expect(0, 12989, '\p{Nv=0047}', "");
Expect(1, 12989, '\p{^Nv=0047}', "");
Expect(1, 12989, '\P{Nv=0047}', "");
Expect(0, 12989, '\P{^Nv=0047}', "");
Error('\p{Is_Numeric_Value=/a/	0047}');
Error('\P{Is_Numeric_Value=/a/	0047}');
Expect(1, 12988, '\p{Is_Numeric_Value=00000047}', "");
Expect(0, 12988, '\p{^Is_Numeric_Value=00000047}', "");
Expect(0, 12988, '\P{Is_Numeric_Value=00000047}', "");
Expect(1, 12988, '\P{^Is_Numeric_Value=00000047}', "");
Expect(0, 12989, '\p{Is_Numeric_Value=00000047}', "");
Expect(1, 12989, '\p{^Is_Numeric_Value=00000047}', "");
Expect(1, 12989, '\P{Is_Numeric_Value=00000047}', "");
Expect(0, 12989, '\P{^Is_Numeric_Value=00000047}', "");
Error('\p{Is_Nv: :=- 00000000047}');
Error('\P{Is_Nv: :=- 00000000047}');
Expect(1, 12988, '\p{Is_Nv=+0047}', "");
Expect(0, 12988, '\p{^Is_Nv=+0047}', "");
Expect(0, 12988, '\P{Is_Nv=+0047}', "");
Expect(1, 12988, '\P{^Is_Nv=+0047}', "");
Expect(0, 12989, '\p{Is_Nv=+0047}', "");
Expect(1, 12989, '\p{^Is_Nv=+0047}', "");
Expect(1, 12989, '\P{Is_Nv=+0047}', "");
Expect(0, 12989, '\P{^Is_Nv=+0047}', "");
Error('\p{Numeric_Value=_+00000004_8/a/}');
Error('\P{Numeric_Value=_+00000004_8/a/}');
Expect(1, 12989, '\p{Numeric_Value=+0_0_0_0_048}', "");
Expect(0, 12989, '\p{^Numeric_Value=+0_0_0_0_048}', "");
Expect(0, 12989, '\P{Numeric_Value=+0_0_0_0_048}', "");
Expect(1, 12989, '\P{^Numeric_Value=+0_0_0_0_048}', "");
Expect(0, 12990, '\p{Numeric_Value=+0_0_0_0_048}', "");
Expect(1, 12990, '\p{^Numeric_Value=+0_0_0_0_048}', "");
Expect(1, 12990, '\P{Numeric_Value=+0_0_0_0_048}', "");
Expect(0, 12990, '\P{^Numeric_Value=+0_0_0_0_048}', "");
Error('\p{Nv=4_8:=}');
Error('\P{Nv=4_8:=}');
Expect(1, 12989, '\p{Nv=+000048}', "");
Expect(0, 12989, '\p{^Nv=+000048}', "");
Expect(0, 12989, '\P{Nv=+000048}', "");
Expect(1, 12989, '\P{^Nv=+000048}', "");
Expect(0, 12990, '\p{Nv=+000048}', "");
Expect(1, 12990, '\p{^Nv=+000048}', "");
Expect(1, 12990, '\P{Nv=+000048}', "");
Expect(0, 12990, '\P{^Nv=+000048}', "");
Error('\p{Is_Numeric_Value=-:=+48}');
Error('\P{Is_Numeric_Value=-:=+48}');
Expect(1, 12989, '\p{Is_Numeric_Value=0_0_0_048}', "");
Expect(0, 12989, '\p{^Is_Numeric_Value=0_0_0_048}', "");
Expect(0, 12989, '\P{Is_Numeric_Value=0_0_0_048}', "");
Expect(1, 12989, '\P{^Is_Numeric_Value=0_0_0_048}', "");
Expect(0, 12990, '\p{Is_Numeric_Value=0_0_0_048}', "");
Expect(1, 12990, '\p{^Is_Numeric_Value=0_0_0_048}', "");
Expect(1, 12990, '\P{Is_Numeric_Value=0_0_0_048}', "");
Expect(0, 12990, '\P{^Is_Numeric_Value=0_0_0_048}', "");
Error('\p{Is_Nv=+004_8/a/}');
Error('\P{Is_Nv=+004_8/a/}');
Expect(1, 12989, '\p{Is_Nv=00_00_00_04_8}', "");
Expect(0, 12989, '\p{^Is_Nv=00_00_00_04_8}', "");
Expect(0, 12989, '\P{Is_Nv=00_00_00_04_8}', "");
Expect(1, 12989, '\P{^Is_Nv=00_00_00_04_8}', "");
Expect(0, 12990, '\p{Is_Nv=00_00_00_04_8}', "");
Expect(1, 12990, '\p{^Is_Nv=00_00_00_04_8}', "");
Expect(1, 12990, '\P{Is_Nv=00_00_00_04_8}', "");
Expect(0, 12990, '\P{^Is_Nv=00_00_00_04_8}', "");
Error('\p{Numeric_Value=:=_	00049}');
Error('\P{Numeric_Value=:=_	00049}');
Expect(1, 12990, '\p{Numeric_Value=000049}', "");
Expect(0, 12990, '\p{^Numeric_Value=000049}', "");
Expect(0, 12990, '\P{Numeric_Value=000049}', "");
Expect(1, 12990, '\P{^Numeric_Value=000049}', "");
Expect(0, 12991, '\p{Numeric_Value=000049}', "");
Expect(1, 12991, '\p{^Numeric_Value=000049}', "");
Expect(1, 12991, '\P{Numeric_Value=000049}', "");
Expect(0, 12991, '\P{^Numeric_Value=000049}', "");
Error('\p{Nv=-+4_9:=}');
Error('\P{Nv=-+4_9:=}');
Expect(1, 12990, '\p{Nv=000000049}', "");
Expect(0, 12990, '\p{^Nv=000000049}', "");
Expect(0, 12990, '\P{Nv=000000049}', "");
Expect(1, 12990, '\P{^Nv=000000049}', "");
Expect(0, 12991, '\p{Nv=000000049}', "");
Expect(1, 12991, '\p{^Nv=000000049}', "");
Expect(1, 12991, '\P{Nv=000000049}', "");
Expect(0, 12991, '\P{^Nv=000000049}', "");
Error('\p{Is_Numeric_Value:    :=+00_00_00_049}');
Error('\P{Is_Numeric_Value:    :=+00_00_00_049}');
Expect(1, 12990, '\p{Is_Numeric_Value=0049}', "");
Expect(0, 12990, '\p{^Is_Numeric_Value=0049}', "");
Expect(0, 12990, '\P{Is_Numeric_Value=0049}', "");
Expect(1, 12990, '\P{^Is_Numeric_Value=0049}', "");
Expect(0, 12991, '\p{Is_Numeric_Value=0049}', "");
Expect(1, 12991, '\p{^Is_Numeric_Value=0049}', "");
Expect(1, 12991, '\P{Is_Numeric_Value=0049}', "");
Expect(0, 12991, '\P{^Is_Numeric_Value=0049}', "");
Error('\p{Is_Nv=-0000000049:=}');
Error('\P{Is_Nv=-0000000049:=}');
Expect(1, 12990, '\p{Is_Nv:00000049}', "");
Expect(0, 12990, '\p{^Is_Nv:00000049}', "");
Expect(0, 12990, '\P{Is_Nv:00000049}', "");
Expect(1, 12990, '\P{^Is_Nv:00000049}', "");
Expect(0, 12991, '\p{Is_Nv:00000049}', "");
Expect(1, 12991, '\p{^Is_Nv:00000049}', "");
Expect(1, 12991, '\P{Is_Nv:00000049}', "");
Expect(0, 12991, '\P{^Is_Nv:00000049}', "");
Error('\p{Numeric_Value:	 	00_5/a/}');
Error('\P{Numeric_Value:	 	00_5/a/}');
Expect(1, 131361, '\p{Numeric_Value=0_0_0_0_0_00005}', "");
Expect(0, 131361, '\p{^Numeric_Value=0_0_0_0_0_00005}', "");
Expect(0, 131361, '\P{Numeric_Value=0_0_0_0_0_00005}', "");
Expect(1, 131361, '\P{^Numeric_Value=0_0_0_0_0_00005}', "");
Expect(0, 131362, '\p{Numeric_Value=0_0_0_0_0_00005}', "");
Expect(1, 131362, '\p{^Numeric_Value=0_0_0_0_0_00005}', "");
Expect(1, 131362, '\P{Numeric_Value=0_0_0_0_0_00005}', "");
Expect(0, 131362, '\P{^Numeric_Value=0_0_0_0_0_00005}', "");
Error('\p{Nv=-/a/0000005}');
Error('\P{Nv=-/a/0000005}');
Expect(1, 131361, '\p{Nv=00005}', "");
Expect(0, 131361, '\p{^Nv=00005}', "");
Expect(0, 131361, '\P{Nv=00005}', "");
Expect(1, 131361, '\P{^Nv=00005}', "");
Expect(0, 131362, '\p{Nv=00005}', "");
Expect(1, 131362, '\p{^Nv=00005}', "");
Expect(1, 131362, '\P{Nv=00005}', "");
Expect(0, 131362, '\P{^Nv=00005}', "");
Error('\p{Is_Numeric_Value=/a/_ 5}');
Error('\P{Is_Numeric_Value=/a/_ 5}');
Expect(1, 131361, '\p{Is_Numeric_Value=+0000000_5}', "");
Expect(0, 131361, '\p{^Is_Numeric_Value=+0000000_5}', "");
Expect(0, 131361, '\P{Is_Numeric_Value=+0000000_5}', "");
Expect(1, 131361, '\P{^Is_Numeric_Value=+0000000_5}', "");
Expect(0, 131362, '\p{Is_Numeric_Value=+0000000_5}', "");
Expect(1, 131362, '\p{^Is_Numeric_Value=+0000000_5}', "");
Expect(1, 131362, '\P{Is_Numeric_Value=+0000000_5}', "");
Expect(0, 131362, '\P{^Is_Numeric_Value=+0000000_5}', "");
Error('\p{Is_Nv=_-+00_00_00_05/a/}');
Error('\P{Is_Nv=_-+00_00_00_05/a/}');
Expect(1, 131361, '\p{Is_Nv=+05}', "");
Expect(0, 131361, '\p{^Is_Nv=+05}', "");
Expect(0, 131361, '\P{Is_Nv=+05}', "");
Expect(1, 131361, '\P{^Is_Nv=+05}', "");
Expect(0, 131362, '\p{Is_Nv=+05}', "");
Expect(1, 131362, '\p{^Is_Nv=+05}', "");
Expect(1, 131362, '\P{Is_Nv=+05}', "");
Expect(0, 131362, '\P{^Is_Nv=+05}', "");
Error('\p{Numeric_Value=/a/+0005/002}');
Error('\P{Numeric_Value=/a/+0005/002}');
Expect(1, 3884, '\p{Numeric_Value=0000000005/00002}', "");
Expect(0, 3884, '\p{^Numeric_Value=0000000005/00002}', "");
Expect(0, 3884, '\P{Numeric_Value=0000000005/00002}', "");
Expect(1, 3884, '\P{^Numeric_Value=0000000005/00002}', "");
Expect(0, 3885, '\p{Numeric_Value=0000000005/00002}', "");
Expect(1, 3885, '\p{^Numeric_Value=0000000005/00002}', "");
Expect(1, 3885, '\P{Numeric_Value=0000000005/00002}', "");
Expect(0, 3885, '\P{^Numeric_Value=0000000005/00002}', "");
Expect(1, 3884, '\p{Numeric_Value=2.500}', "");
Expect(0, 3884, '\p{^Numeric_Value=2.500}', "");
Expect(0, 3884, '\P{Numeric_Value=2.500}', "");
Expect(1, 3884, '\P{^Numeric_Value=2.500}', "");
Expect(0, 3885, '\p{Numeric_Value=2.500}', "");
Expect(1, 3885, '\p{^Numeric_Value=2.500}', "");
Expect(1, 3885, '\P{Numeric_Value=2.500}', "");
Expect(0, 3885, '\P{^Numeric_Value=2.500}', "");
Error('\p{Nv=-0005/00002/a/}');
Error('\P{Nv=-0005/00002/a/}');
Expect(1, 3884, '\p{Nv=0000005/000002}', "");
Expect(0, 3884, '\p{^Nv=0000005/000002}', "");
Expect(0, 3884, '\P{Nv=0000005/000002}', "");
Expect(1, 3884, '\P{^Nv=0000005/000002}', "");
Expect(0, 3885, '\p{Nv=0000005/000002}', "");
Expect(1, 3885, '\p{^Nv=0000005/000002}', "");
Expect(1, 3885, '\P{Nv=0000005/000002}', "");
Expect(0, 3885, '\P{^Nv=0000005/000002}', "");
Expect(1, 3884, '\p{Nv=2.500}', "");
Expect(0, 3884, '\p{^Nv=2.500}', "");
Expect(0, 3884, '\P{Nv=2.500}', "");
Expect(1, 3884, '\P{^Nv=2.500}', "");
Expect(0, 3885, '\p{Nv=2.500}', "");
Expect(1, 3885, '\p{^Nv=2.500}', "");
Expect(1, 3885, '\P{Nv=2.500}', "");
Expect(0, 3885, '\P{^Nv=2.500}', "");
Error('\p{Is_Numeric_Value=_/a/+000005/0000002}');
Error('\P{Is_Numeric_Value=_/a/+000005/0000002}');
Expect(1, 3884, '\p{Is_Numeric_Value=00000005/000000002}', "");
Expect(0, 3884, '\p{^Is_Numeric_Value=00000005/000000002}', "");
Expect(0, 3884, '\P{Is_Numeric_Value=00000005/000000002}', "");
Expect(1, 3884, '\P{^Is_Numeric_Value=00000005/000000002}', "");
Expect(0, 3885, '\p{Is_Numeric_Value=00000005/000000002}', "");
Expect(1, 3885, '\p{^Is_Numeric_Value=00000005/000000002}', "");
Expect(1, 3885, '\P{Is_Numeric_Value=00000005/000000002}', "");
Expect(0, 3885, '\P{^Is_Numeric_Value=00000005/000000002}', "");
Expect(1, 3884, '\p{Is_Numeric_Value=2.500}', "");
Expect(0, 3884, '\p{^Is_Numeric_Value=2.500}', "");
Expect(0, 3884, '\P{Is_Numeric_Value=2.500}', "");
Expect(1, 3884, '\P{^Is_Numeric_Value=2.500}', "");
Expect(0, 3885, '\p{Is_Numeric_Value=2.500}', "");
Expect(1, 3885, '\p{^Is_Numeric_Value=2.500}', "");
Expect(1, 3885, '\P{Is_Numeric_Value=2.500}', "");
Expect(0, 3885, '\P{^Is_Numeric_Value=2.500}', "");
Error('\p{Is_Nv:	:=5/000002}');
Error('\P{Is_Nv:	:=5/000002}');
Expect(1, 3884, '\p{Is_Nv=00000005/00000002}', "");
Expect(0, 3884, '\p{^Is_Nv=00000005/00000002}', "");
Expect(0, 3884, '\P{Is_Nv=00000005/00000002}', "");
Expect(1, 3884, '\P{^Is_Nv=00000005/00000002}', "");
Expect(0, 3885, '\p{Is_Nv=00000005/00000002}', "");
Expect(1, 3885, '\p{^Is_Nv=00000005/00000002}', "");
Expect(1, 3885, '\P{Is_Nv=00000005/00000002}', "");
Expect(0, 3885, '\P{^Is_Nv=00000005/00000002}', "");
Expect(1, 3884, '\p{Is_Nv=2.500}', "");
Expect(0, 3884, '\p{^Is_Nv=2.500}', "");
Expect(0, 3884, '\P{Is_Nv=2.500}', "");
Expect(1, 3884, '\P{^Is_Nv=2.500}', "");
Expect(0, 3885, '\p{Is_Nv=2.500}', "");
Expect(1, 3885, '\p{^Is_Nv=2.500}', "");
Expect(1, 3885, '\P{Is_Nv=2.500}', "");
Expect(0, 3885, '\P{^Is_Nv=2.500}', "");
Error('\p{Numeric_Value=/a/0000005/0006}');
Error('\P{Numeric_Value=/a/0000005/0006}');
Expect(1, 74844, '\p{Numeric_Value=+005/0000000006}', "");
Expect(0, 74844, '\p{^Numeric_Value=+005/0000000006}', "");
Expect(0, 74844, '\P{Numeric_Value=+005/0000000006}', "");
Expect(1, 74844, '\P{^Numeric_Value=+005/0000000006}', "");
Expect(0, 74845, '\p{Numeric_Value=+005/0000000006}', "");
Expect(1, 74845, '\p{^Numeric_Value=+005/0000000006}', "");
Expect(1, 74845, '\P{Numeric_Value=+005/0000000006}', "");
Expect(0, 74845, '\P{^Numeric_Value=+005/0000000006}', "");
Error('\p{Numeric_Value=0.83}');
Error('\P{Numeric_Value=0.83}');
Expect(1, 74844, '\p{Numeric_Value=0.833}', "");
Expect(0, 74844, '\p{^Numeric_Value=0.833}', "");
Expect(0, 74844, '\P{Numeric_Value=0.833}', "");
Expect(1, 74844, '\P{^Numeric_Value=0.833}', "");
Expect(0, 74845, '\p{Numeric_Value=0.833}', "");
Expect(1, 74845, '\p{^Numeric_Value=0.833}', "");
Expect(1, 74845, '\P{Numeric_Value=0.833}', "");
Expect(0, 74845, '\P{^Numeric_Value=0.833}', "");
Error('\p{Nv= _+0000000005/0006/a/}');
Error('\P{Nv= _+0000000005/0006/a/}');
Expect(1, 74844, '\p{Nv=+005/0006}', "");
Expect(0, 74844, '\p{^Nv=+005/0006}', "");
Expect(0, 74844, '\P{Nv=+005/0006}', "");
Expect(1, 74844, '\P{^Nv=+005/0006}', "");
Expect(0, 74845, '\p{Nv=+005/0006}', "");
Expect(1, 74845, '\p{^Nv=+005/0006}', "");
Expect(1, 74845, '\P{Nv=+005/0006}', "");
Expect(0, 74845, '\P{^Nv=+005/0006}', "");
Error('\p{Nv=0.83}');
Error('\P{Nv=0.83}');
Expect(1, 74844, '\p{Nv=0.833}', "");
Expect(0, 74844, '\p{^Nv=0.833}', "");
Expect(0, 74844, '\P{Nv=0.833}', "");
Expect(1, 74844, '\P{^Nv=0.833}', "");
Expect(0, 74845, '\p{Nv=0.833}', "");
Expect(1, 74845, '\p{^Nv=0.833}', "");
Expect(1, 74845, '\P{Nv=0.833}', "");
Expect(0, 74845, '\P{^Nv=0.833}', "");
Error('\p{Is_Numeric_Value:		/a/+000005/0000006}');
Error('\P{Is_Numeric_Value:		/a/+000005/0000006}');
Expect(1, 74844, '\p{Is_Numeric_Value=0000000005/00000006}', "");
Expect(0, 74844, '\p{^Is_Numeric_Value=0000000005/00000006}', "");
Expect(0, 74844, '\P{Is_Numeric_Value=0000000005/00000006}', "");
Expect(1, 74844, '\P{^Is_Numeric_Value=0000000005/00000006}', "");
Expect(0, 74845, '\p{Is_Numeric_Value=0000000005/00000006}', "");
Expect(1, 74845, '\p{^Is_Numeric_Value=0000000005/00000006}', "");
Expect(1, 74845, '\P{Is_Numeric_Value=0000000005/00000006}', "");
Expect(0, 74845, '\P{^Is_Numeric_Value=0000000005/00000006}', "");
Error('\p{Is_Numeric_Value=0.83}');
Error('\P{Is_Numeric_Value=0.83}');
Expect(1, 74844, '\p{Is_Numeric_Value=0.833}', "");
Expect(0, 74844, '\p{^Is_Numeric_Value=0.833}', "");
Expect(0, 74844, '\P{Is_Numeric_Value=0.833}', "");
Expect(1, 74844, '\P{^Is_Numeric_Value=0.833}', "");
Expect(0, 74845, '\p{Is_Numeric_Value=0.833}', "");
Expect(1, 74845, '\p{^Is_Numeric_Value=0.833}', "");
Expect(1, 74845, '\P{Is_Numeric_Value=0.833}', "");
Expect(0, 74845, '\P{^Is_Numeric_Value=0.833}', "");
Error('\p{Is_Nv=	-+00000005/000006:=}');
Error('\P{Is_Nv=	-+00000005/000006:=}');
Expect(1, 74844, '\p{Is_Nv=+5/00000006}', "");
Expect(0, 74844, '\p{^Is_Nv=+5/00000006}', "");
Expect(0, 74844, '\P{Is_Nv=+5/00000006}', "");
Expect(1, 74844, '\P{^Is_Nv=+5/00000006}', "");
Expect(0, 74845, '\p{Is_Nv=+5/00000006}', "");
Expect(1, 74845, '\p{^Is_Nv=+5/00000006}', "");
Expect(1, 74845, '\P{Is_Nv=+5/00000006}', "");
Expect(0, 74845, '\P{^Is_Nv=+5/00000006}', "");
Error('\p{Is_Nv=0.83}');
Error('\P{Is_Nv=0.83}');
Expect(1, 74844, '\p{Is_Nv:   0.833}', "");
Expect(0, 74844, '\p{^Is_Nv:   0.833}', "");
Expect(0, 74844, '\P{Is_Nv:   0.833}', "");
Expect(1, 74844, '\P{^Is_Nv:   0.833}', "");
Expect(0, 74845, '\p{Is_Nv:   0.833}', "");
Expect(1, 74845, '\p{^Is_Nv:   0.833}', "");
Expect(1, 74845, '\P{Is_Nv:   0.833}', "");
Expect(0, 74845, '\P{^Is_Nv:   0.833}', "");
Error('\p{Numeric_Value=-/a/0000005/00000008}');
Error('\P{Numeric_Value=-/a/0000005/00000008}');
Expect(1, 8541, '\p{Numeric_Value=00000005/0000008}', "");
Expect(0, 8541, '\p{^Numeric_Value=00000005/0000008}', "");
Expect(0, 8541, '\P{Numeric_Value=00000005/0000008}', "");
Expect(1, 8541, '\P{^Numeric_Value=00000005/0000008}', "");
Expect(0, 8542, '\p{Numeric_Value=00000005/0000008}', "");
Expect(1, 8542, '\p{^Numeric_Value=00000005/0000008}', "");
Expect(1, 8542, '\P{Numeric_Value=00000005/0000008}', "");
Expect(0, 8542, '\P{^Numeric_Value=00000005/0000008}', "");
Error('\p{Numeric_Value=0.62}');
Error('\P{Numeric_Value=0.62}');
Expect(1, 8541, '\p{Numeric_Value=0.625}', "");
Expect(0, 8541, '\p{^Numeric_Value=0.625}', "");
Expect(0, 8541, '\P{Numeric_Value=0.625}', "");
Expect(1, 8541, '\P{^Numeric_Value=0.625}', "");
Expect(0, 8542, '\p{Numeric_Value=0.625}', "");
Expect(1, 8542, '\p{^Numeric_Value=0.625}', "");
Expect(1, 8542, '\P{Numeric_Value=0.625}', "");
Expect(0, 8542, '\P{^Numeric_Value=0.625}', "");
Error('\p{Nv: _005/8/a/}');
Error('\P{Nv: _005/8/a/}');
Expect(1, 8541, '\p{Nv=+0000005/00000008}', "");
Expect(0, 8541, '\p{^Nv=+0000005/00000008}', "");
Expect(0, 8541, '\P{Nv=+0000005/00000008}', "");
Expect(1, 8541, '\P{^Nv=+0000005/00000008}', "");
Expect(0, 8542, '\p{Nv=+0000005/00000008}', "");
Expect(1, 8542, '\p{^Nv=+0000005/00000008}', "");
Expect(1, 8542, '\P{Nv=+0000005/00000008}', "");
Expect(0, 8542, '\P{^Nv=+0000005/00000008}', "");
Error('\p{Nv=0.62}');
Error('\P{Nv=0.62}');
Expect(1, 8541, '\p{Nv=0.625}', "");
Expect(0, 8541, '\p{^Nv=0.625}', "");
Expect(0, 8541, '\P{Nv=0.625}', "");
Expect(1, 8541, '\P{^Nv=0.625}', "");
Expect(0, 8542, '\p{Nv=0.625}', "");
Expect(1, 8542, '\p{^Nv=0.625}', "");
Expect(1, 8542, '\P{Nv=0.625}', "");
Expect(0, 8542, '\P{^Nv=0.625}', "");
Error('\p{Is_Numeric_Value=/a/--005/0000008}');
Error('\P{Is_Numeric_Value=/a/--005/0000008}');
Expect(1, 8541, '\p{Is_Numeric_Value=005/008}', "");
Expect(0, 8541, '\p{^Is_Numeric_Value=005/008}', "");
Expect(0, 8541, '\P{Is_Numeric_Value=005/008}', "");
Expect(1, 8541, '\P{^Is_Numeric_Value=005/008}', "");
Expect(0, 8542, '\p{Is_Numeric_Value=005/008}', "");
Expect(1, 8542, '\p{^Is_Numeric_Value=005/008}', "");
Expect(1, 8542, '\P{Is_Numeric_Value=005/008}', "");
Expect(0, 8542, '\P{^Is_Numeric_Value=005/008}', "");
Error('\p{Is_Numeric_Value=0.62}');
Error('\P{Is_Numeric_Value=0.62}');
Expect(1, 8541, '\p{Is_Numeric_Value=0.625}', "");
Expect(0, 8541, '\p{^Is_Numeric_Value=0.625}', "");
Expect(0, 8541, '\P{Is_Numeric_Value=0.625}', "");
Expect(1, 8541, '\P{^Is_Numeric_Value=0.625}', "");
Expect(0, 8542, '\p{Is_Numeric_Value=0.625}', "");
Expect(1, 8542, '\p{^Is_Numeric_Value=0.625}', "");
Expect(1, 8542, '\P{Is_Numeric_Value=0.625}', "");
Expect(0, 8542, '\P{^Is_Numeric_Value=0.625}', "");
Error('\p{Is_Nv=/a/_	5/08}');
Error('\P{Is_Nv=/a/_	5/08}');
Expect(1, 8541, '\p{Is_Nv=+00000005/0000008}', "");
Expect(0, 8541, '\p{^Is_Nv=+00000005/0000008}', "");
Expect(0, 8541, '\P{Is_Nv=+00000005/0000008}', "");
Expect(1, 8541, '\P{^Is_Nv=+00000005/0000008}', "");
Expect(0, 8542, '\p{Is_Nv=+00000005/0000008}', "");
Expect(1, 8542, '\p{^Is_Nv=+00000005/0000008}', "");
Expect(1, 8542, '\P{Is_Nv=+00000005/0000008}', "");
Expect(0, 8542, '\P{^Is_Nv=+00000005/0000008}', "");
Error('\p{Is_Nv=0.62}');
Error('\P{Is_Nv=0.62}');
Expect(1, 8541, '\p{Is_Nv=0.625}', "");
Expect(0, 8541, '\p{^Is_Nv=0.625}', "");
Expect(0, 8541, '\P{Is_Nv=0.625}', "");
Expect(1, 8541, '\P{^Is_Nv=0.625}', "");
Expect(0, 8542, '\p{Is_Nv=0.625}', "");
Expect(1, 8542, '\p{^Is_Nv=0.625}', "");
Expect(1, 8542, '\P{Is_Nv=0.625}', "");
Expect(0, 8542, '\P{^Is_Nv=0.625}', "");
Error('\p{Numeric_Value=:=		+00000050}');
Error('\P{Numeric_Value=:=		+00000050}');
Expect(1, 119661, '\p{Numeric_Value=0_0_0_0_0_0_50}', "");
Expect(0, 119661, '\p{^Numeric_Value=0_0_0_0_0_0_50}', "");
Expect(0, 119661, '\P{Numeric_Value=0_0_0_0_0_0_50}', "");
Expect(1, 119661, '\P{^Numeric_Value=0_0_0_0_0_0_50}', "");
Expect(0, 119662, '\p{Numeric_Value=0_0_0_0_0_0_50}', "");
Expect(1, 119662, '\p{^Numeric_Value=0_0_0_0_0_0_50}', "");
Expect(1, 119662, '\P{Numeric_Value=0_0_0_0_0_0_50}', "");
Expect(0, 119662, '\P{^Numeric_Value=0_0_0_0_0_0_50}', "");
Error('\p{Nv=/a/	000050}');
Error('\P{Nv=/a/	000050}');
Expect(1, 119661, '\p{Nv=+050}', "");
Expect(0, 119661, '\p{^Nv=+050}', "");
Expect(0, 119661, '\P{Nv=+050}', "");
Expect(1, 119661, '\P{^Nv=+050}', "");
Expect(0, 119662, '\p{Nv=+050}', "");
Expect(1, 119662, '\p{^Nv=+050}', "");
Expect(1, 119662, '\P{Nv=+050}', "");
Expect(0, 119662, '\P{^Nv=+050}', "");
Error('\p{Is_Numeric_Value= _+05_0/a/}');
Error('\P{Is_Numeric_Value= _+05_0/a/}');
Expect(1, 119661, '\p{Is_Numeric_Value=00050}', "");
Expect(0, 119661, '\p{^Is_Numeric_Value=00050}', "");
Expect(0, 119661, '\P{Is_Numeric_Value=00050}', "");
Expect(1, 119661, '\P{^Is_Numeric_Value=00050}', "");
Expect(0, 119662, '\p{Is_Numeric_Value=00050}', "");
Expect(1, 119662, '\p{^Is_Numeric_Value=00050}', "");
Expect(1, 119662, '\P{Is_Numeric_Value=00050}', "");
Expect(0, 119662, '\P{^Is_Numeric_Value=00050}', "");
Error('\p{Is_Nv=_:=+0000050}');
Error('\P{Is_Nv=_:=+0000050}');
Expect(1, 119661, '\p{Is_Nv=0000000050}', "");
Expect(0, 119661, '\p{^Is_Nv=0000000050}', "");
Expect(0, 119661, '\P{Is_Nv=0000000050}', "");
Expect(1, 119661, '\P{^Is_Nv=0000000050}', "");
Expect(0, 119662, '\p{Is_Nv=0000000050}', "");
Expect(1, 119662, '\p{^Is_Nv=0000000050}', "");
Expect(1, 119662, '\P{Is_Nv=0000000050}', "");
Expect(0, 119662, '\P{^Is_Nv=0000000050}', "");
Error('\p{Numeric_Value::=	 +000500}');
Error('\P{Numeric_Value::=	 +000500}');
Expect(1, 69238, '\p{Numeric_Value:	0500}', "");
Expect(0, 69238, '\p{^Numeric_Value:	0500}', "");
Expect(0, 69238, '\P{Numeric_Value:	0500}', "");
Expect(1, 69238, '\P{^Numeric_Value:	0500}', "");
Expect(0, 69239, '\p{Numeric_Value:	0500}', "");
Expect(1, 69239, '\p{^Numeric_Value:	0500}', "");
Expect(1, 69239, '\P{Numeric_Value:	0500}', "");
Expect(0, 69239, '\P{^Numeric_Value:	0500}', "");
Error('\p{Nv=:=0500}');
Error('\P{Nv=:=0500}');
Expect(1, 69238, '\p{Nv=0_0_0_0_0_0_0_500}', "");
Expect(0, 69238, '\p{^Nv=0_0_0_0_0_0_0_500}', "");
Expect(0, 69238, '\P{Nv=0_0_0_0_0_0_0_500}', "");
Expect(1, 69238, '\P{^Nv=0_0_0_0_0_0_0_500}', "");
Expect(0, 69239, '\p{Nv=0_0_0_0_0_0_0_500}', "");
Expect(1, 69239, '\p{^Nv=0_0_0_0_0_0_0_500}', "");
Expect(1, 69239, '\P{Nv=0_0_0_0_0_0_0_500}', "");
Expect(0, 69239, '\P{^Nv=0_0_0_0_0_0_0_500}', "");
Error('\p{Is_Numeric_Value=:=_	000_050_0}');
Error('\P{Is_Numeric_Value=:=_	000_050_0}');
Expect(1, 69238, '\p{Is_Numeric_Value=000000000500}', "");
Expect(0, 69238, '\p{^Is_Numeric_Value=000000000500}', "");
Expect(0, 69238, '\P{Is_Numeric_Value=000000000500}', "");
Expect(1, 69238, '\P{^Is_Numeric_Value=000000000500}', "");
Expect(0, 69239, '\p{Is_Numeric_Value=000000000500}', "");
Expect(1, 69239, '\p{^Is_Numeric_Value=000000000500}', "");
Expect(1, 69239, '\P{Is_Numeric_Value=000000000500}', "");
Expect(0, 69239, '\P{^Is_Numeric_Value=000000000500}', "");
Error('\p{Is_Nv=	 +0000500:=}');
Error('\P{Is_Nv=	 +0000500:=}');
Expect(1, 69238, '\p{Is_Nv:   000000000500}', "");
Expect(0, 69238, '\p{^Is_Nv:   000000000500}', "");
Expect(0, 69238, '\P{Is_Nv:   000000000500}', "");
Expect(1, 69238, '\P{^Is_Nv:   000000000500}', "");
Expect(0, 69239, '\p{Is_Nv:   000000000500}', "");
Expect(1, 69239, '\p{^Is_Nv:   000000000500}', "");
Expect(1, 69239, '\P{Is_Nv:   000000000500}', "");
Expect(0, 69239, '\P{^Is_Nv:   000000000500}', "");
Error('\p{Numeric_Value=/a/ 05000}');
Error('\P{Numeric_Value=/a/ 05000}');
Expect(1, 65906, '\p{Numeric_Value=+000005000}', "");
Expect(0, 65906, '\p{^Numeric_Value=+000005000}', "");
Expect(0, 65906, '\P{Numeric_Value=+000005000}', "");
Expect(1, 65906, '\P{^Numeric_Value=+000005000}', "");
Expect(0, 65907, '\p{Numeric_Value=+000005000}', "");
Expect(1, 65907, '\p{^Numeric_Value=+000005000}', "");
Expect(1, 65907, '\P{Numeric_Value=+000005000}', "");
Expect(0, 65907, '\P{^Numeric_Value=+000005000}', "");
Error('\p{Nv=_-0000000005000/a/}');
Error('\P{Nv=_-0000000005000/a/}');
Expect(1, 65906, '\p{Nv=00_00_00_00_05_00_0}', "");
Expect(0, 65906, '\p{^Nv=00_00_00_00_05_00_0}', "");
Expect(0, 65906, '\P{Nv=00_00_00_00_05_00_0}', "");
Expect(1, 65906, '\P{^Nv=00_00_00_00_05_00_0}', "");
Expect(0, 65907, '\p{Nv=00_00_00_00_05_00_0}', "");
Expect(1, 65907, '\p{^Nv=00_00_00_00_05_00_0}', "");
Expect(1, 65907, '\P{Nv=00_00_00_00_05_00_0}', "");
Expect(0, 65907, '\P{^Nv=00_00_00_00_05_00_0}', "");
Error('\p{Is_Numeric_Value:   /a/-	05_00_0}');
Error('\P{Is_Numeric_Value:   /a/-	05_00_0}');
Expect(1, 65906, '\p{Is_Numeric_Value=005000}', "");
Expect(0, 65906, '\p{^Is_Numeric_Value=005000}', "");
Expect(0, 65906, '\P{Is_Numeric_Value=005000}', "");
Expect(1, 65906, '\P{^Is_Numeric_Value=005000}', "");
Expect(0, 65907, '\p{Is_Numeric_Value=005000}', "");
Expect(1, 65907, '\p{^Is_Numeric_Value=005000}', "");
Expect(1, 65907, '\P{Is_Numeric_Value=005000}', "");
Expect(0, 65907, '\P{^Is_Numeric_Value=005000}', "");
Error('\p{Is_Nv=+0_0_0_0_0_5000:=}');
Error('\P{Is_Nv=+0_0_0_0_0_5000:=}');
Expect(1, 65906, '\p{Is_Nv=05000}', "");
Expect(0, 65906, '\p{^Is_Nv=05000}', "");
Expect(0, 65906, '\P{Is_Nv=05000}', "");
Expect(1, 65906, '\P{^Is_Nv=05000}', "");
Expect(0, 65907, '\p{Is_Nv=05000}', "");
Expect(1, 65907, '\p{^Is_Nv=05000}', "");
Expect(1, 65907, '\P{Is_Nv=05000}', "");
Expect(0, 65907, '\P{^Is_Nv=05000}', "");
Error('\p{Numeric_Value=_-+0_5_0_0_00/a/}');
Error('\P{Numeric_Value=_-+0_5_0_0_00/a/}');
Expect(1, 65878, '\p{Numeric_Value=+005_000_0}', "");
Expect(0, 65878, '\p{^Numeric_Value=+005_000_0}', "");
Expect(0, 65878, '\P{Numeric_Value=+005_000_0}', "");
Expect(1, 65878, '\P{^Numeric_Value=+005_000_0}', "");
Expect(0, 65879, '\p{Numeric_Value=+005_000_0}', "");
Expect(1, 65879, '\p{^Numeric_Value=+005_000_0}', "");
Expect(1, 65879, '\P{Numeric_Value=+005_000_0}', "");
Expect(0, 65879, '\P{^Numeric_Value=+005_000_0}', "");
Error('\p{Nv=_ 0_0_5_0_0_00/a/}');
Error('\P{Nv=_ 0_0_5_0_0_00/a/}');
Expect(1, 65878, '\p{Nv=+50000}', "");
Expect(0, 65878, '\p{^Nv=+50000}', "");
Expect(0, 65878, '\P{Nv=+50000}', "");
Expect(1, 65878, '\P{^Nv=+50000}', "");
Expect(0, 65879, '\p{Nv=+50000}', "");
Expect(1, 65879, '\p{^Nv=+50000}', "");
Expect(1, 65879, '\P{Nv=+50000}', "");
Expect(0, 65879, '\P{^Nv=+50000}', "");
Error('\p{Is_Numeric_Value=-_000050000:=}');
Error('\P{Is_Numeric_Value=-_000050000:=}');
Expect(1, 65878, '\p{Is_Numeric_Value=+0_5_0_0_00}', "");
Expect(0, 65878, '\p{^Is_Numeric_Value=+0_5_0_0_00}', "");
Expect(0, 65878, '\P{Is_Numeric_Value=+0_5_0_0_00}', "");
Expect(1, 65878, '\P{^Is_Numeric_Value=+0_5_0_0_00}', "");
Expect(0, 65879, '\p{Is_Numeric_Value=+0_5_0_0_00}', "");
Expect(1, 65879, '\p{^Is_Numeric_Value=+0_5_0_0_00}', "");
Expect(1, 65879, '\P{Is_Numeric_Value=+0_5_0_0_00}', "");
Expect(0, 65879, '\P{^Is_Numeric_Value=+0_5_0_0_00}', "");
Error('\p{Is_Nv=:=		00_00_50_00_0}');
Error('\P{Is_Nv=:=		00_00_50_00_0}');
Expect(1, 65878, '\p{Is_Nv=00_00_00_00_50_000}', "");
Expect(0, 65878, '\p{^Is_Nv=00_00_00_00_50_000}', "");
Expect(0, 65878, '\P{Is_Nv=00_00_00_00_50_000}', "");
Expect(1, 65878, '\P{^Is_Nv=00_00_00_00_50_000}', "");
Expect(0, 65879, '\p{Is_Nv=00_00_00_00_50_000}', "");
Expect(1, 65879, '\p{^Is_Nv=00_00_00_00_50_000}', "");
Expect(1, 65879, '\P{Is_Nv=00_00_00_00_50_000}', "");
Expect(0, 65879, '\P{^Is_Nv=00_00_00_00_50_000}', "");
Error('\p{Numeric_Value=-/a/00_6}');
Error('\P{Numeric_Value=-/a/00_6}');
Expect(1, 133866, '\p{Numeric_Value=0000000006}', "");
Expect(0, 133866, '\p{^Numeric_Value=0000000006}', "");
Expect(0, 133866, '\P{Numeric_Value=0000000006}', "");
Expect(1, 133866, '\P{^Numeric_Value=0000000006}', "");
Expect(0, 133867, '\p{Numeric_Value=0000000006}', "");
Expect(1, 133867, '\p{^Numeric_Value=0000000006}', "");
Expect(1, 133867, '\P{Numeric_Value=0000000006}', "");
Expect(0, 133867, '\P{^Numeric_Value=0000000006}', "");
Error('\p{Nv= /a/0_6}');
Error('\P{Nv= /a/0_6}');
Expect(1, 133866, '\p{Nv=+6}', "");
Expect(0, 133866, '\p{^Nv=+6}', "");
Expect(0, 133866, '\P{Nv=+6}', "");
Expect(1, 133866, '\P{^Nv=+6}', "");
Expect(0, 133867, '\p{Nv=+6}', "");
Expect(1, 133867, '\p{^Nv=+6}', "");
Expect(1, 133867, '\P{Nv=+6}', "");
Expect(0, 133867, '\P{^Nv=+6}', "");
Error('\p{Is_Numeric_Value= /a/0_0_06}');
Error('\P{Is_Numeric_Value= /a/0_0_06}');
Expect(1, 133866, '\p{Is_Numeric_Value=0000000006}', "");
Expect(0, 133866, '\p{^Is_Numeric_Value=0000000006}', "");
Expect(0, 133866, '\P{Is_Numeric_Value=0000000006}', "");
Expect(1, 133866, '\P{^Is_Numeric_Value=0000000006}', "");
Expect(0, 133867, '\p{Is_Numeric_Value=0000000006}', "");
Expect(1, 133867, '\p{^Is_Numeric_Value=0000000006}', "");
Expect(1, 133867, '\P{Is_Numeric_Value=0000000006}', "");
Expect(0, 133867, '\P{^Is_Numeric_Value=0000000006}', "");
Error('\p{Is_Nv=/a/-_+00000006}');
Error('\P{Is_Nv=/a/-_+00000006}');
Expect(1, 133866, '\p{Is_Nv=+0_0_0_0_0_06}', "");
Expect(0, 133866, '\p{^Is_Nv=+0_0_0_0_0_06}', "");
Expect(0, 133866, '\P{Is_Nv=+0_0_0_0_0_06}', "");
Expect(1, 133866, '\P{^Is_Nv=+0_0_0_0_0_06}', "");
Expect(0, 133867, '\p{Is_Nv=+0_0_0_0_0_06}', "");
Expect(1, 133867, '\p{^Is_Nv=+0_0_0_0_0_06}', "");
Expect(1, 133867, '\P{Is_Nv=+0_0_0_0_0_06}', "");
Expect(0, 133867, '\P{^Is_Nv=+0_0_0_0_0_06}', "");
Error('\p{Numeric_Value=:=- 0000006_0}');
Error('\P{Numeric_Value=:=- 0000006_0}');
Expect(1, 119662, '\p{Numeric_Value=00_00_00_060}', "");
Expect(0, 119662, '\p{^Numeric_Value=00_00_00_060}', "");
Expect(0, 119662, '\P{Numeric_Value=00_00_00_060}', "");
Expect(1, 119662, '\P{^Numeric_Value=00_00_00_060}', "");
Expect(0, 119663, '\p{Numeric_Value=00_00_00_060}', "");
Expect(1, 119663, '\p{^Numeric_Value=00_00_00_060}', "");
Expect(1, 119663, '\P{Numeric_Value=00_00_00_060}', "");
Expect(0, 119663, '\P{^Numeric_Value=00_00_00_060}', "");
Error('\p{Nv=_/a/00000060}');
Error('\P{Nv=_/a/00000060}');
Expect(1, 119662, '\p{Nv=000060}', "");
Expect(0, 119662, '\p{^Nv=000060}', "");
Expect(0, 119662, '\P{Nv=000060}', "");
Expect(1, 119662, '\P{^Nv=000060}', "");
Expect(0, 119663, '\p{Nv=000060}', "");
Expect(1, 119663, '\p{^Nv=000060}', "");
Expect(1, 119663, '\P{Nv=000060}', "");
Expect(0, 119663, '\P{^Nv=000060}', "");
Error('\p{Is_Numeric_Value=  +0_0_0_0_0_0_0060/a/}');
Error('\P{Is_Numeric_Value=  +0_0_0_0_0_0_0060/a/}');
Expect(1, 119662, '\p{Is_Numeric_Value=0000000060}', "");
Expect(0, 119662, '\p{^Is_Numeric_Value=0000000060}', "");
Expect(0, 119662, '\P{Is_Numeric_Value=0000000060}', "");
Expect(1, 119662, '\P{^Is_Numeric_Value=0000000060}', "");
Expect(0, 119663, '\p{Is_Numeric_Value=0000000060}', "");
Expect(1, 119663, '\p{^Is_Numeric_Value=0000000060}', "");
Expect(1, 119663, '\P{Is_Numeric_Value=0000000060}', "");
Expect(0, 119663, '\P{^Is_Numeric_Value=0000000060}', "");
Error('\p{Is_Nv=_0_0_0_0_0_60:=}');
Error('\P{Is_Nv=_0_0_0_0_0_60:=}');
Expect(1, 119662, '\p{Is_Nv=0000006_0}', "");
Expect(0, 119662, '\p{^Is_Nv=0000006_0}', "");
Expect(0, 119662, '\P{Is_Nv=0000006_0}', "");
Expect(1, 119662, '\P{^Is_Nv=0000006_0}', "");
Expect(0, 119663, '\p{Is_Nv=0000006_0}', "");
Expect(1, 119663, '\p{^Is_Nv=0000006_0}', "");
Expect(1, 119663, '\P{Is_Nv=0000006_0}', "");
Expect(0, 119663, '\P{^Is_Nv=0000006_0}', "");
Error('\p{Numeric_Value=-/a/0000600}');
Error('\P{Numeric_Value=-/a/0000600}');
Expect(1, 69239, '\p{Numeric_Value=60_0}', "");
Expect(0, 69239, '\p{^Numeric_Value=60_0}', "");
Expect(0, 69239, '\P{Numeric_Value=60_0}', "");
Expect(1, 69239, '\P{^Numeric_Value=60_0}', "");
Expect(0, 69240, '\p{Numeric_Value=60_0}', "");
Expect(1, 69240, '\p{^Numeric_Value=60_0}', "");
Expect(1, 69240, '\P{Numeric_Value=60_0}', "");
Expect(0, 69240, '\P{^Numeric_Value=60_0}', "");
Error('\p{Nv=/a/ 000060_0}');
Error('\P{Nv=/a/ 000060_0}');
Expect(1, 69239, '\p{Nv=+0_0_0_0_0_0_0_0_0600}', "");
Expect(0, 69239, '\p{^Nv=+0_0_0_0_0_0_0_0_0600}', "");
Expect(0, 69239, '\P{Nv=+0_0_0_0_0_0_0_0_0600}', "");
Expect(1, 69239, '\P{^Nv=+0_0_0_0_0_0_0_0_0600}', "");
Expect(0, 69240, '\p{Nv=+0_0_0_0_0_0_0_0_0600}', "");
Expect(1, 69240, '\p{^Nv=+0_0_0_0_0_0_0_0_0600}', "");
Expect(1, 69240, '\P{Nv=+0_0_0_0_0_0_0_0_0600}', "");
Expect(0, 69240, '\P{^Nv=+0_0_0_0_0_0_0_0_0600}', "");
Error('\p{Is_Numeric_Value=:=000600}');
Error('\P{Is_Numeric_Value=:=000600}');
Expect(1, 69239, '\p{Is_Numeric_Value=00000600}', "");
Expect(0, 69239, '\p{^Is_Numeric_Value=00000600}', "");
Expect(0, 69239, '\P{Is_Numeric_Value=00000600}', "");
Expect(1, 69239, '\P{^Is_Numeric_Value=00000600}', "");
Expect(0, 69240, '\p{Is_Numeric_Value=00000600}', "");
Expect(1, 69240, '\p{^Is_Numeric_Value=00000600}', "");
Expect(1, 69240, '\P{Is_Numeric_Value=00000600}', "");
Expect(0, 69240, '\P{^Is_Numeric_Value=00000600}', "");
Error('\p{Is_Nv=-	+00_00_00_00_60_0:=}');
Error('\P{Is_Nv=-	+00_00_00_00_60_0:=}');
Expect(1, 69239, '\p{Is_Nv=+000000600}', "");
Expect(0, 69239, '\p{^Is_Nv=+000000600}', "");
Expect(0, 69239, '\P{Is_Nv=+000000600}', "");
Expect(1, 69239, '\P{^Is_Nv=+000000600}', "");
Expect(0, 69240, '\p{Is_Nv=+000000600}', "");
Expect(1, 69240, '\p{^Is_Nv=+000000600}', "");
Expect(1, 69240, '\P{Is_Nv=+000000600}', "");
Expect(0, 69240, '\P{^Is_Nv=+000000600}', "");
Error('\p{Numeric_Value:_+0_0_0_0_0_0_0_0_06000:=}');
Error('\P{Numeric_Value:_+0_0_0_0_0_0_0_0_06000:=}');
Expect(1, 65831, '\p{Numeric_Value=00_00_00_06_00_0}', "");
Expect(0, 65831, '\p{^Numeric_Value=00_00_00_06_00_0}', "");
Expect(0, 65831, '\P{Numeric_Value=00_00_00_06_00_0}', "");
Expect(1, 65831, '\P{^Numeric_Value=00_00_00_06_00_0}', "");
Expect(0, 65832, '\p{Numeric_Value=00_00_00_06_00_0}', "");
Expect(1, 65832, '\p{^Numeric_Value=00_00_00_06_00_0}', "");
Expect(1, 65832, '\P{Numeric_Value=00_00_00_06_00_0}', "");
Expect(0, 65832, '\P{^Numeric_Value=00_00_00_06_00_0}', "");
Error('\p{Nv=:= 0000_0060_00}');
Error('\P{Nv=:= 0000_0060_00}');
Expect(1, 65831, '\p{Nv=06000}', "");
Expect(0, 65831, '\p{^Nv=06000}', "");
Expect(0, 65831, '\P{Nv=06000}', "");
Expect(1, 65831, '\P{^Nv=06000}', "");
Expect(0, 65832, '\p{Nv=06000}', "");
Expect(1, 65832, '\p{^Nv=06000}', "");
Expect(1, 65832, '\P{Nv=06000}', "");
Expect(0, 65832, '\P{^Nv=06000}', "");
Error('\p{Is_Numeric_Value=/a/ -0_0_0_6_000}');
Error('\P{Is_Numeric_Value=/a/ -0_0_0_6_000}');
Expect(1, 65831, '\p{Is_Numeric_Value=00_00_60_00}', "");
Expect(0, 65831, '\p{^Is_Numeric_Value=00_00_60_00}', "");
Expect(0, 65831, '\P{Is_Numeric_Value=00_00_60_00}', "");
Expect(1, 65831, '\P{^Is_Numeric_Value=00_00_60_00}', "");
Expect(0, 65832, '\p{Is_Numeric_Value=00_00_60_00}', "");
Expect(1, 65832, '\p{^Is_Numeric_Value=00_00_60_00}', "");
Expect(1, 65832, '\P{Is_Numeric_Value=00_00_60_00}', "");
Expect(0, 65832, '\P{^Is_Numeric_Value=00_00_60_00}', "");
Error('\p{Is_Nv= /a/+0000006000}');
Error('\P{Is_Nv= /a/+0000006000}');
Expect(1, 65831, '\p{Is_Nv=+0_0_0_0_0_0_6_000}', "");
Expect(0, 65831, '\p{^Is_Nv=+0_0_0_0_0_0_6_000}', "");
Expect(0, 65831, '\P{Is_Nv=+0_0_0_0_0_0_6_000}', "");
Expect(1, 65831, '\P{^Is_Nv=+0_0_0_0_0_0_6_000}', "");
Expect(0, 65832, '\p{Is_Nv=+0_0_0_0_0_0_6_000}', "");
Expect(1, 65832, '\p{^Is_Nv=+0_0_0_0_0_0_6_000}', "");
Expect(1, 65832, '\P{Is_Nv=+0_0_0_0_0_0_6_000}', "");
Expect(0, 65832, '\P{^Is_Nv=+0_0_0_0_0_0_6_000}', "");
Error('\p{Numeric_Value=:= _+060000}');
Error('\P{Numeric_Value=:= _+060000}');
Expect(1, 65840, '\p{Numeric_Value:   0060000}', "");
Expect(0, 65840, '\p{^Numeric_Value:   0060000}', "");
Expect(0, 65840, '\P{Numeric_Value:   0060000}', "");
Expect(1, 65840, '\P{^Numeric_Value:   0060000}', "");
Expect(0, 65841, '\p{Numeric_Value:   0060000}', "");
Expect(1, 65841, '\p{^Numeric_Value:   0060000}', "");
Expect(1, 65841, '\P{Numeric_Value:   0060000}', "");
Expect(0, 65841, '\P{^Numeric_Value:   0060000}', "");
Error('\p{Nv=-000000060000:=}');
Error('\P{Nv=-000000060000:=}');
Expect(1, 65840, '\p{Nv=00000000060000}', "");
Expect(0, 65840, '\p{^Nv=00000000060000}', "");
Expect(0, 65840, '\P{Nv=00000000060000}', "");
Expect(1, 65840, '\P{^Nv=00000000060000}', "");
Expect(0, 65841, '\p{Nv=00000000060000}', "");
Expect(1, 65841, '\p{^Nv=00000000060000}', "");
Expect(1, 65841, '\P{Nv=00000000060000}', "");
Expect(0, 65841, '\P{^Nv=00000000060000}', "");
Error('\p{Is_Numeric_Value= :=+0_0_0_0_0_0_6_0_000}');
Error('\P{Is_Numeric_Value= :=+0_0_0_0_0_0_6_0_000}');
Expect(1, 65840, '\p{Is_Numeric_Value=00_06_00_00}', "");
Expect(0, 65840, '\p{^Is_Numeric_Value=00_06_00_00}', "");
Expect(0, 65840, '\P{Is_Numeric_Value=00_06_00_00}', "");
Expect(1, 65840, '\P{^Is_Numeric_Value=00_06_00_00}', "");
Expect(0, 65841, '\p{Is_Numeric_Value=00_06_00_00}', "");
Expect(1, 65841, '\p{^Is_Numeric_Value=00_06_00_00}', "");
Expect(1, 65841, '\P{Is_Numeric_Value=00_06_00_00}', "");
Expect(0, 65841, '\P{^Is_Numeric_Value=00_06_00_00}', "");
Error('\p{Is_Nv=-000060000:=}');
Error('\P{Is_Nv=-000060000:=}');
Expect(1, 65840, '\p{Is_Nv=000060000}', "");
Expect(0, 65840, '\p{^Is_Nv=000060000}', "");
Expect(0, 65840, '\P{Is_Nv=000060000}', "");
Expect(1, 65840, '\P{^Is_Nv=000060000}', "");
Expect(0, 65841, '\p{Is_Nv=000060000}', "");
Expect(1, 65841, '\p{^Is_Nv=000060000}', "");
Expect(1, 65841, '\P{Is_Nv=000060000}', "");
Expect(0, 65841, '\P{^Is_Nv=000060000}', "");
Error('\p{Numeric_Value= +00000007:=}');
Error('\P{Numeric_Value= +00000007:=}');
Expect(1, 131073, '\p{Numeric_Value=00007}', "");
Expect(0, 131073, '\p{^Numeric_Value=00007}', "");
Expect(0, 131073, '\P{Numeric_Value=00007}', "");
Expect(1, 131073, '\P{^Numeric_Value=00007}', "");
Expect(0, 131074, '\p{Numeric_Value=00007}', "");
Expect(1, 131074, '\p{^Numeric_Value=00007}', "");
Expect(1, 131074, '\P{Numeric_Value=00007}', "");
Expect(0, 131074, '\P{^Numeric_Value=00007}', "");
Error('\p{Nv=	 +00_00_7/a/}');
Error('\P{Nv=	 +00_00_7/a/}');
Expect(1, 131073, '\p{Nv=0_0_0_0_0_0_0_0_07}', "");
Expect(0, 131073, '\p{^Nv=0_0_0_0_0_0_0_0_07}', "");
Expect(0, 131073, '\P{Nv=0_0_0_0_0_0_0_0_07}', "");
Expect(1, 131073, '\P{^Nv=0_0_0_0_0_0_0_0_07}', "");
Expect(0, 131074, '\p{Nv=0_0_0_0_0_0_0_0_07}', "");
Expect(1, 131074, '\p{^Nv=0_0_0_0_0_0_0_0_07}', "");
Expect(1, 131074, '\P{Nv=0_0_0_0_0_0_0_0_07}', "");
Expect(0, 131074, '\P{^Nv=0_0_0_0_0_0_0_0_07}', "");
Error('\p{Is_Numeric_Value=	 +0_0_0_0_0_0_007/a/}');
Error('\P{Is_Numeric_Value=	 +0_0_0_0_0_0_007/a/}');
Expect(1, 131073, '\p{Is_Numeric_Value=0_0_0_0_0_00007}', "");
Expect(0, 131073, '\p{^Is_Numeric_Value=0_0_0_0_0_00007}', "");
Expect(0, 131073, '\P{Is_Numeric_Value=0_0_0_0_0_00007}', "");
Expect(1, 131073, '\P{^Is_Numeric_Value=0_0_0_0_0_00007}', "");
Expect(0, 131074, '\p{Is_Numeric_Value=0_0_0_0_0_00007}', "");
Expect(1, 131074, '\p{^Is_Numeric_Value=0_0_0_0_0_00007}', "");
Expect(1, 131074, '\P{Is_Numeric_Value=0_0_0_0_0_00007}', "");
Expect(0, 131074, '\P{^Is_Numeric_Value=0_0_0_0_0_00007}', "");
Error('\p{Is_Nv=	:=000007}');
Error('\P{Is_Nv=	:=000007}');
Expect(1, 131073, '\p{Is_Nv=0_0_0_0_0_0_07}', "");
Expect(0, 131073, '\p{^Is_Nv=0_0_0_0_0_0_07}', "");
Expect(0, 131073, '\P{Is_Nv=0_0_0_0_0_0_07}', "");
Expect(1, 131073, '\P{^Is_Nv=0_0_0_0_0_0_07}', "");
Expect(0, 131074, '\p{Is_Nv=0_0_0_0_0_0_07}', "");
Expect(1, 131074, '\p{^Is_Nv=0_0_0_0_0_0_07}', "");
Expect(1, 131074, '\P{Is_Nv=0_0_0_0_0_0_07}', "");
Expect(0, 131074, '\P{^Is_Nv=0_0_0_0_0_0_07}', "");
Error('\p{Numeric_Value=/a/+0007/002}');
Error('\P{Numeric_Value=/a/+0007/002}');
Expect(1, 3885, '\p{Numeric_Value=0000007/000002}', "");
Expect(0, 3885, '\p{^Numeric_Value=0000007/000002}', "");
Expect(0, 3885, '\P{Numeric_Value=0000007/000002}', "");
Expect(1, 3885, '\P{^Numeric_Value=0000007/000002}', "");
Expect(0, 3886, '\p{Numeric_Value=0000007/000002}', "");
Expect(1, 3886, '\p{^Numeric_Value=0000007/000002}', "");
Expect(1, 3886, '\P{Numeric_Value=0000007/000002}', "");
Expect(0, 3886, '\P{^Numeric_Value=0000007/000002}', "");
Expect(1, 3885, '\p{Numeric_Value:3.500}', "");
Expect(0, 3885, '\p{^Numeric_Value:3.500}', "");
Expect(0, 3885, '\P{Numeric_Value:3.500}', "");
Expect(1, 3885, '\P{^Numeric_Value:3.500}', "");
Expect(0, 3886, '\p{Numeric_Value:3.500}', "");
Expect(1, 3886, '\p{^Numeric_Value:3.500}', "");
Expect(1, 3886, '\P{Numeric_Value:3.500}', "");
Expect(0, 3886, '\P{^Numeric_Value:3.500}', "");
Error('\p{Nv=/a/- 000007/0002}');
Error('\P{Nv=/a/- 000007/0002}');
Expect(1, 3885, '\p{Nv=0007/00002}', "");
Expect(0, 3885, '\p{^Nv=0007/00002}', "");
Expect(0, 3885, '\P{Nv=0007/00002}', "");
Expect(1, 3885, '\P{^Nv=0007/00002}', "");
Expect(0, 3886, '\p{Nv=0007/00002}', "");
Expect(1, 3886, '\p{^Nv=0007/00002}', "");
Expect(1, 3886, '\P{Nv=0007/00002}', "");
Expect(0, 3886, '\P{^Nv=0007/00002}', "");
Expect(1, 3885, '\p{Nv=3.500}', "");
Expect(0, 3885, '\p{^Nv=3.500}', "");
Expect(0, 3885, '\P{Nv=3.500}', "");
Expect(1, 3885, '\P{^Nv=3.500}', "");
Expect(0, 3886, '\p{Nv=3.500}', "");
Expect(1, 3886, '\p{^Nv=3.500}', "");
Expect(1, 3886, '\P{Nv=3.500}', "");
Expect(0, 3886, '\P{^Nv=3.500}', "");
Error('\p{Is_Numeric_Value=/a/ 7/002}');
Error('\P{Is_Numeric_Value=/a/ 7/002}');
Expect(1, 3885, '\p{Is_Numeric_Value=+0000007/0000002}', "");
Expect(0, 3885, '\p{^Is_Numeric_Value=+0000007/0000002}', "");
Expect(0, 3885, '\P{Is_Numeric_Value=+0000007/0000002}', "");
Expect(1, 3885, '\P{^Is_Numeric_Value=+0000007/0000002}', "");
Expect(0, 3886, '\p{Is_Numeric_Value=+0000007/0000002}', "");
Expect(1, 3886, '\p{^Is_Numeric_Value=+0000007/0000002}', "");
Expect(1, 3886, '\P{Is_Numeric_Value=+0000007/0000002}', "");
Expect(0, 3886, '\P{^Is_Numeric_Value=+0000007/0000002}', "");
Expect(1, 3885, '\p{Is_Numeric_Value=3.500}', "");
Expect(0, 3885, '\p{^Is_Numeric_Value=3.500}', "");
Expect(0, 3885, '\P{Is_Numeric_Value=3.500}', "");
Expect(1, 3885, '\P{^Is_Numeric_Value=3.500}', "");
Expect(0, 3886, '\p{Is_Numeric_Value=3.500}', "");
Expect(1, 3886, '\p{^Is_Numeric_Value=3.500}', "");
Expect(1, 3886, '\P{Is_Numeric_Value=3.500}', "");
Expect(0, 3886, '\P{^Is_Numeric_Value=3.500}', "");
Error('\p{Is_Nv=	_+0007/02/a/}');
Error('\P{Is_Nv=	_+0007/02/a/}');
Expect(1, 3885, '\p{Is_Nv=+007/0000000002}', "");
Expect(0, 3885, '\p{^Is_Nv=+007/0000000002}', "");
Expect(0, 3885, '\P{Is_Nv=+007/0000000002}', "");
Expect(1, 3885, '\P{^Is_Nv=+007/0000000002}', "");
Expect(0, 3886, '\p{Is_Nv=+007/0000000002}', "");
Expect(1, 3886, '\p{^Is_Nv=+007/0000000002}', "");
Expect(1, 3886, '\P{Is_Nv=+007/0000000002}', "");
Expect(0, 3886, '\P{^Is_Nv=+007/0000000002}', "");
Expect(1, 3885, '\p{Is_Nv=3.500}', "");
Expect(0, 3885, '\p{^Is_Nv=3.500}', "");
Expect(0, 3885, '\P{Is_Nv=3.500}', "");
Expect(1, 3885, '\P{^Is_Nv=3.500}', "");
Expect(0, 3886, '\p{Is_Nv=3.500}', "");
Expect(1, 3886, '\p{^Is_Nv=3.500}', "");
Expect(1, 3886, '\P{Is_Nv=3.500}', "");
Expect(0, 3886, '\P{^Is_Nv=3.500}', "");
Error('\p{Numeric_Value=_0000007/0000008:=}');
Error('\P{Numeric_Value=_0000007/0000008:=}');
Expect(1, 8542, '\p{Numeric_Value=+0000007/000008}', "");
Expect(0, 8542, '\p{^Numeric_Value=+0000007/000008}', "");
Expect(0, 8542, '\P{Numeric_Value=+0000007/000008}', "");
Expect(1, 8542, '\P{^Numeric_Value=+0000007/000008}', "");
Expect(0, 8543, '\p{Numeric_Value=+0000007/000008}', "");
Expect(1, 8543, '\p{^Numeric_Value=+0000007/000008}', "");
Expect(1, 8543, '\P{Numeric_Value=+0000007/000008}', "");
Expect(0, 8543, '\P{^Numeric_Value=+0000007/000008}', "");
Error('\p{Numeric_Value=0.9}');
Error('\P{Numeric_Value=0.9}');
Error('\p{Numeric_Value=0.88}');
Error('\P{Numeric_Value=0.88}');
Expect(1, 8542, '\p{Numeric_Value=0.875}', "");
Expect(0, 8542, '\p{^Numeric_Value=0.875}', "");
Expect(0, 8542, '\P{Numeric_Value=0.875}', "");
Expect(1, 8542, '\P{^Numeric_Value=0.875}', "");
Expect(0, 8543, '\p{Numeric_Value=0.875}', "");
Expect(1, 8543, '\p{^Numeric_Value=0.875}', "");
Expect(1, 8543, '\P{Numeric_Value=0.875}', "");
Expect(0, 8543, '\P{^Numeric_Value=0.875}', "");
Error('\p{Nv=/a/-_07/0000008}');
Error('\P{Nv=/a/-_07/0000008}');
Expect(1, 8542, '\p{Nv=007/0000008}', "");
Expect(0, 8542, '\p{^Nv=007/0000008}', "");
Expect(0, 8542, '\P{Nv=007/0000008}', "");
Expect(1, 8542, '\P{^Nv=007/0000008}', "");
Expect(0, 8543, '\p{Nv=007/0000008}', "");
Expect(1, 8543, '\p{^Nv=007/0000008}', "");
Expect(1, 8543, '\P{Nv=007/0000008}', "");
Expect(0, 8543, '\P{^Nv=007/0000008}', "");
Error('\p{Nv=0.9}');
Error('\P{Nv=0.9}');
Error('\p{Nv=0.88}');
Error('\P{Nv=0.88}');
Expect(1, 8542, '\p{Nv:	0.875}', "");
Expect(0, 8542, '\p{^Nv:	0.875}', "");
Expect(0, 8542, '\P{Nv:	0.875}', "");
Expect(1, 8542, '\P{^Nv:	0.875}', "");
Expect(0, 8543, '\p{Nv:	0.875}', "");
Expect(1, 8543, '\p{^Nv:	0.875}', "");
Expect(1, 8543, '\P{Nv:	0.875}', "");
Expect(0, 8543, '\P{^Nv:	0.875}', "");
Error('\p{Is_Numeric_Value=-00007/00008/a/}');
Error('\P{Is_Numeric_Value=-00007/00008/a/}');
Expect(1, 8542, '\p{Is_Numeric_Value=0000007/0008}', "");
Expect(0, 8542, '\p{^Is_Numeric_Value=0000007/0008}', "");
Expect(0, 8542, '\P{Is_Numeric_Value=0000007/0008}', "");
Expect(1, 8542, '\P{^Is_Numeric_Value=0000007/0008}', "");
Expect(0, 8543, '\p{Is_Numeric_Value=0000007/0008}', "");
Expect(1, 8543, '\p{^Is_Numeric_Value=0000007/0008}', "");
Expect(1, 8543, '\P{Is_Numeric_Value=0000007/0008}', "");
Expect(0, 8543, '\P{^Is_Numeric_Value=0000007/0008}', "");
Error('\p{Is_Numeric_Value=0.9}');
Error('\P{Is_Numeric_Value=0.9}');
Error('\p{Is_Numeric_Value=0.88}');
Error('\P{Is_Numeric_Value=0.88}');
Expect(1, 8542, '\p{Is_Numeric_Value=0.875}', "");
Expect(0, 8542, '\p{^Is_Numeric_Value=0.875}', "");
Expect(0, 8542, '\P{Is_Numeric_Value=0.875}', "");
Expect(1, 8542, '\P{^Is_Numeric_Value=0.875}', "");
Expect(0, 8543, '\p{Is_Numeric_Value=0.875}', "");
Expect(1, 8543, '\p{^Is_Numeric_Value=0.875}', "");
Expect(1, 8543, '\P{Is_Numeric_Value=0.875}', "");
Expect(0, 8543, '\P{^Is_Numeric_Value=0.875}', "");
Error('\p{Is_Nv=:=000007/0000000008}');
Error('\P{Is_Nv=:=000007/0000000008}');
Expect(1, 8542, '\p{Is_Nv=+007/008}', "");
Expect(0, 8542, '\p{^Is_Nv=+007/008}', "");
Expect(0, 8542, '\P{Is_Nv=+007/008}', "");
Expect(1, 8542, '\P{^Is_Nv=+007/008}', "");
Expect(0, 8543, '\p{Is_Nv=+007/008}', "");
Expect(1, 8543, '\p{^Is_Nv=+007/008}', "");
Expect(1, 8543, '\P{Is_Nv=+007/008}', "");
Expect(0, 8543, '\P{^Is_Nv=+007/008}', "");
Error('\p{Is_Nv=0.9}');
Error('\P{Is_Nv=0.9}');
Error('\p{Is_Nv=0.88}');
Error('\P{Is_Nv=0.88}');
Expect(1, 8542, '\p{Is_Nv:   0.875}', "");
Expect(0, 8542, '\p{^Is_Nv:   0.875}', "");
Expect(0, 8542, '\P{Is_Nv:   0.875}', "");
Expect(1, 8542, '\P{^Is_Nv:   0.875}', "");
Expect(0, 8543, '\p{Is_Nv:   0.875}', "");
Expect(1, 8543, '\p{^Is_Nv:   0.875}', "");
Expect(1, 8543, '\P{Is_Nv:   0.875}', "");
Expect(0, 8543, '\P{^Is_Nv:   0.875}', "");
Error('\p{Numeric_Value=_+7_0:=}');
Error('\P{Numeric_Value=_+7_0:=}');
Expect(1, 119663, '\p{Numeric_Value:   0_0_0_0_0_70}', "");
Expect(0, 119663, '\p{^Numeric_Value:   0_0_0_0_0_70}', "");
Expect(0, 119663, '\P{Numeric_Value:   0_0_0_0_0_70}', "");
Expect(1, 119663, '\P{^Numeric_Value:   0_0_0_0_0_70}', "");
Expect(0, 119664, '\p{Numeric_Value:   0_0_0_0_0_70}', "");
Expect(1, 119664, '\p{^Numeric_Value:   0_0_0_0_0_70}', "");
Expect(1, 119664, '\P{Numeric_Value:   0_0_0_0_0_70}', "");
Expect(0, 119664, '\P{^Numeric_Value:   0_0_0_0_0_70}', "");
Error('\p{Nv=:=  +00000070}');
Error('\P{Nv=:=  +00000070}');
Expect(1, 119663, '\p{Nv=+000000070}', "");
Expect(0, 119663, '\p{^Nv=+000000070}', "");
Expect(0, 119663, '\P{Nv=+000000070}', "");
Expect(1, 119663, '\P{^Nv=+000000070}', "");
Expect(0, 119664, '\p{Nv=+000000070}', "");
Expect(1, 119664, '\p{^Nv=+000000070}', "");
Expect(1, 119664, '\P{Nv=+000000070}', "");
Expect(0, 119664, '\P{^Nv=+000000070}', "");
Error('\p{Is_Numeric_Value=-:=+070}');
Error('\P{Is_Numeric_Value=-:=+070}');
Expect(1, 119663, '\p{Is_Numeric_Value=007_0}', "");
Expect(0, 119663, '\p{^Is_Numeric_Value=007_0}', "");
Expect(0, 119663, '\P{Is_Numeric_Value=007_0}', "");
Expect(1, 119663, '\P{^Is_Numeric_Value=007_0}', "");
Expect(0, 119664, '\p{Is_Numeric_Value=007_0}', "");
Expect(1, 119664, '\p{^Is_Numeric_Value=007_0}', "");
Expect(1, 119664, '\P{Is_Numeric_Value=007_0}', "");
Expect(0, 119664, '\P{^Is_Numeric_Value=007_0}', "");
Error('\p{Is_Nv=- +0000070/a/}');
Error('\P{Is_Nv=- +0000070/a/}');
Expect(1, 119663, '\p{Is_Nv=+00070}', "");
Expect(0, 119663, '\p{^Is_Nv=+00070}', "");
Expect(0, 119663, '\P{Is_Nv=+00070}', "");
Expect(1, 119663, '\P{^Is_Nv=+00070}', "");
Expect(0, 119664, '\p{Is_Nv=+00070}', "");
Expect(1, 119664, '\p{^Is_Nv=+00070}', "");
Expect(1, 119664, '\P{Is_Nv=+00070}', "");
Expect(0, 119664, '\P{^Is_Nv=+00070}', "");
Error('\p{Numeric_Value=:=_0_0_0_0_0_0_0_0_700}');
Error('\P{Numeric_Value=:=_0_0_0_0_0_0_0_0_700}');
Expect(1, 69240, '\p{Numeric_Value=+00000000700}', "");
Expect(0, 69240, '\p{^Numeric_Value=+00000000700}', "");
Expect(0, 69240, '\P{Numeric_Value=+00000000700}', "");
Expect(1, 69240, '\P{^Numeric_Value=+00000000700}', "");
Expect(0, 69241, '\p{Numeric_Value=+00000000700}', "");
Expect(1, 69241, '\p{^Numeric_Value=+00000000700}', "");
Expect(1, 69241, '\P{Numeric_Value=+00000000700}', "");
Expect(0, 69241, '\P{^Numeric_Value=+00000000700}', "");
Error('\p{Nv=:= 	+70_0}');
Error('\P{Nv=:= 	+70_0}');
Expect(1, 69240, '\p{Nv=+000070_0}', "");
Expect(0, 69240, '\p{^Nv=+000070_0}', "");
Expect(0, 69240, '\P{Nv=+000070_0}', "");
Expect(1, 69240, '\P{^Nv=+000070_0}', "");
Expect(0, 69241, '\p{Nv=+000070_0}', "");
Expect(1, 69241, '\p{^Nv=+000070_0}', "");
Expect(1, 69241, '\P{Nv=+000070_0}', "");
Expect(0, 69241, '\P{^Nv=+000070_0}', "");
Error('\p{Is_Numeric_Value=_/a/+00000700}');
Error('\P{Is_Numeric_Value=_/a/+00000700}');
Expect(1, 69240, '\p{Is_Numeric_Value=000700}', "");
Expect(0, 69240, '\p{^Is_Numeric_Value=000700}', "");
Expect(0, 69240, '\P{Is_Numeric_Value=000700}', "");
Expect(1, 69240, '\P{^Is_Numeric_Value=000700}', "");
Expect(0, 69241, '\p{Is_Numeric_Value=000700}', "");
Expect(1, 69241, '\p{^Is_Numeric_Value=000700}', "");
Expect(1, 69241, '\P{Is_Numeric_Value=000700}', "");
Expect(0, 69241, '\P{^Is_Numeric_Value=000700}', "");
Error('\p{Is_Nv=:=_000000000700}');
Error('\P{Is_Nv=:=_000000000700}');
Expect(1, 69240, '\p{Is_Nv=+0_0_0_0_0_700}', "");
Expect(0, 69240, '\p{^Is_Nv=+0_0_0_0_0_700}', "");
Expect(0, 69240, '\P{Is_Nv=+0_0_0_0_0_700}', "");
Expect(1, 69240, '\P{^Is_Nv=+0_0_0_0_0_700}', "");
Expect(0, 69241, '\p{Is_Nv=+0_0_0_0_0_700}', "");
Expect(1, 69241, '\p{^Is_Nv=+0_0_0_0_0_700}', "");
Expect(1, 69241, '\P{Is_Nv=+0_0_0_0_0_700}', "");
Expect(0, 69241, '\P{^Is_Nv=+0_0_0_0_0_700}', "");
Error('\p{Numeric_Value=/a/  0007000}');
Error('\P{Numeric_Value=/a/  0007000}');
Expect(1, 65832, '\p{Numeric_Value:   +000_070_00}', "");
Expect(0, 65832, '\p{^Numeric_Value:   +000_070_00}', "");
Expect(0, 65832, '\P{Numeric_Value:   +000_070_00}', "");
Expect(1, 65832, '\P{^Numeric_Value:   +000_070_00}', "");
Expect(0, 65833, '\p{Numeric_Value:   +000_070_00}', "");
Expect(1, 65833, '\p{^Numeric_Value:   +000_070_00}', "");
Expect(1, 65833, '\P{Numeric_Value:   +000_070_00}', "");
Expect(0, 65833, '\P{^Numeric_Value:   +000_070_00}', "");
Error('\p{Nv=0007000/a/}');
Error('\P{Nv=0007000/a/}');
Expect(1, 65832, '\p{Nv=7000}', "");
Expect(0, 65832, '\p{^Nv=7000}', "");
Expect(0, 65832, '\P{Nv=7000}', "");
Expect(1, 65832, '\P{^Nv=7000}', "");
Expect(0, 65833, '\p{Nv=7000}', "");
Expect(1, 65833, '\p{^Nv=7000}', "");
Expect(1, 65833, '\P{Nv=7000}', "");
Expect(0, 65833, '\P{^Nv=7000}', "");
Error('\p{Is_Numeric_Value=/a/ 00_00_07_00_0}');
Error('\P{Is_Numeric_Value=/a/ 00_00_07_00_0}');
Expect(1, 65832, '\p{Is_Numeric_Value=+0000_0700_0}', "");
Expect(0, 65832, '\p{^Is_Numeric_Value=+0000_0700_0}', "");
Expect(0, 65832, '\P{Is_Numeric_Value=+0000_0700_0}', "");
Expect(1, 65832, '\P{^Is_Numeric_Value=+0000_0700_0}', "");
Expect(0, 65833, '\p{Is_Numeric_Value=+0000_0700_0}', "");
Expect(1, 65833, '\p{^Is_Numeric_Value=+0000_0700_0}', "");
Expect(1, 65833, '\P{Is_Numeric_Value=+0000_0700_0}', "");
Expect(0, 65833, '\P{^Is_Numeric_Value=+0000_0700_0}', "");
Error('\p{Is_Nv= _0_0_0_0_0_0_0_0_0_7_000:=}');
Error('\P{Is_Nv= _0_0_0_0_0_0_0_0_0_7_000:=}');
Expect(1, 65832, '\p{Is_Nv=0_0_0_0_0_7_0_00}', "");
Expect(0, 65832, '\p{^Is_Nv=0_0_0_0_0_7_0_00}', "");
Expect(0, 65832, '\P{Is_Nv=0_0_0_0_0_7_0_00}', "");
Expect(1, 65832, '\P{^Is_Nv=0_0_0_0_0_7_0_00}', "");
Expect(0, 65833, '\p{Is_Nv=0_0_0_0_0_7_0_00}', "");
Expect(1, 65833, '\p{^Is_Nv=0_0_0_0_0_7_0_00}', "");
Expect(1, 65833, '\P{Is_Nv=0_0_0_0_0_7_0_00}', "");
Expect(0, 65833, '\P{^Is_Nv=0_0_0_0_0_7_0_00}', "");
Error('\p{Numeric_Value=/a/--0_0_0_7_0_000}');
Error('\P{Numeric_Value=/a/--0_0_0_7_0_000}');
Expect(1, 65841, '\p{Numeric_Value=+0_0_0_0_7_0000}', "");
Expect(0, 65841, '\p{^Numeric_Value=+0_0_0_0_7_0000}', "");
Expect(0, 65841, '\P{Numeric_Value=+0_0_0_0_7_0000}', "");
Expect(1, 65841, '\P{^Numeric_Value=+0_0_0_0_7_0000}', "");
Expect(0, 65842, '\p{Numeric_Value=+0_0_0_0_7_0000}', "");
Expect(1, 65842, '\p{^Numeric_Value=+0_0_0_0_7_0000}', "");
Expect(1, 65842, '\P{Numeric_Value=+0_0_0_0_7_0000}', "");
Expect(0, 65842, '\P{^Numeric_Value=+0_0_0_0_7_0000}', "");
Error('\p{Nv:	 :=000070000}');
Error('\P{Nv:	 :=000070000}');
Expect(1, 65841, '\p{Nv=0000000070000}', "");
Expect(0, 65841, '\p{^Nv=0000000070000}', "");
Expect(0, 65841, '\P{Nv=0000000070000}', "");
Expect(1, 65841, '\P{^Nv=0000000070000}', "");
Expect(0, 65842, '\p{Nv=0000000070000}', "");
Expect(1, 65842, '\p{^Nv=0000000070000}', "");
Expect(1, 65842, '\P{Nv=0000000070000}', "");
Expect(0, 65842, '\P{^Nv=0000000070000}', "");
Error('\p{Is_Numeric_Value= 	+00_00_00_00_07_00_00:=}');
Error('\P{Is_Numeric_Value= 	+00_00_00_00_07_00_00:=}');
Expect(1, 65841, '\p{Is_Numeric_Value: +0_0_0_0_0_0_0_7_0_000}', "");
Expect(0, 65841, '\p{^Is_Numeric_Value: +0_0_0_0_0_0_0_7_0_000}', "");
Expect(0, 65841, '\P{Is_Numeric_Value: +0_0_0_0_0_0_0_7_0_000}', "");
Expect(1, 65841, '\P{^Is_Numeric_Value: +0_0_0_0_0_0_0_7_0_000}', "");
Expect(0, 65842, '\p{Is_Numeric_Value: +0_0_0_0_0_0_0_7_0_000}', "");
Expect(1, 65842, '\p{^Is_Numeric_Value: +0_0_0_0_0_0_0_7_0_000}', "");
Expect(1, 65842, '\P{Is_Numeric_Value: +0_0_0_0_0_0_0_7_0_000}', "");
Expect(0, 65842, '\P{^Is_Numeric_Value: +0_0_0_0_0_0_0_7_0_000}', "");
Error('\p{Is_Nv=__0000_0700_00:=}');
Error('\P{Is_Nv=__0000_0700_00:=}');
Expect(1, 65841, '\p{Is_Nv:	0000000007000_0}', "");
Expect(0, 65841, '\p{^Is_Nv:	0000000007000_0}', "");
Expect(0, 65841, '\P{Is_Nv:	0000000007000_0}', "");
Expect(1, 65841, '\P{^Is_Nv:	0000000007000_0}', "");
Expect(0, 65842, '\p{Is_Nv:	0000000007000_0}', "");
Expect(1, 65842, '\p{^Is_Nv:	0000000007000_0}', "");
Expect(1, 65842, '\P{Is_Nv:	0000000007000_0}', "");
Expect(0, 65842, '\P{^Is_Nv:	0000000007000_0}', "");
Error('\p{Numeric_Value=		0_0_0_0_0_0_008/a/}');
Error('\P{Numeric_Value=		0_0_0_0_0_0_008/a/}');
Expect(1, 127241, '\p{Numeric_Value=+0_0_0_0_0_0_08}', "");
Expect(0, 127241, '\p{^Numeric_Value=+0_0_0_0_0_0_08}', "");
Expect(0, 127241, '\P{Numeric_Value=+0_0_0_0_0_0_08}', "");
Expect(1, 127241, '\P{^Numeric_Value=+0_0_0_0_0_0_08}', "");
Expect(0, 127242, '\p{Numeric_Value=+0_0_0_0_0_0_08}', "");
Expect(1, 127242, '\p{^Numeric_Value=+0_0_0_0_0_0_08}', "");
Expect(1, 127242, '\P{Numeric_Value=+0_0_0_0_0_0_08}', "");
Expect(0, 127242, '\P{^Numeric_Value=+0_0_0_0_0_0_08}', "");
Error('\p{Nv=/a/ _+0000008}');
Error('\P{Nv=/a/ _+0000008}');
Expect(1, 127241, '\p{Nv=0000000008}', "");
Expect(0, 127241, '\p{^Nv=0000000008}', "");
Expect(0, 127241, '\P{Nv=0000000008}', "");
Expect(1, 127241, '\P{^Nv=0000000008}', "");
Expect(0, 127242, '\p{Nv=0000000008}', "");
Expect(1, 127242, '\p{^Nv=0000000008}', "");
Expect(1, 127242, '\P{Nv=0000000008}', "");
Expect(0, 127242, '\P{^Nv=0000000008}', "");
Error('\p{Is_Numeric_Value=:=_0_8}');
Error('\P{Is_Numeric_Value=:=_0_8}');
Expect(1, 127241, '\p{Is_Numeric_Value=0_0_0_0_0_0_0_08}', "");
Expect(0, 127241, '\p{^Is_Numeric_Value=0_0_0_0_0_0_0_08}', "");
Expect(0, 127241, '\P{Is_Numeric_Value=0_0_0_0_0_0_0_08}', "");
Expect(1, 127241, '\P{^Is_Numeric_Value=0_0_0_0_0_0_0_08}', "");
Expect(0, 127242, '\p{Is_Numeric_Value=0_0_0_0_0_0_0_08}', "");
Expect(1, 127242, '\p{^Is_Numeric_Value=0_0_0_0_0_0_0_08}', "");
Expect(1, 127242, '\P{Is_Numeric_Value=0_0_0_0_0_0_0_08}', "");
Expect(0, 127242, '\P{^Is_Numeric_Value=0_0_0_0_0_0_0_08}', "");
Error('\p{Is_Nv=_:=+000000008}');
Error('\P{Is_Nv=_:=+000000008}');
Expect(1, 127241, '\p{Is_Nv:0_0_0_08}', "");
Expect(0, 127241, '\p{^Is_Nv:0_0_0_08}', "");
Expect(0, 127241, '\P{Is_Nv:0_0_0_08}', "");
Expect(1, 127241, '\P{^Is_Nv:0_0_0_08}', "");
Expect(0, 127242, '\p{Is_Nv:0_0_0_08}', "");
Expect(1, 127242, '\p{^Is_Nv:0_0_0_08}', "");
Expect(1, 127242, '\P{Is_Nv:0_0_0_08}', "");
Expect(0, 127242, '\P{^Is_Nv:0_0_0_08}', "");
Error('\p{Numeric_Value=-/a/+080}');
Error('\P{Numeric_Value=-/a/+080}');
Expect(1, 119664, '\p{Numeric_Value=000_008_0}', "");
Expect(0, 119664, '\p{^Numeric_Value=000_008_0}', "");
Expect(0, 119664, '\P{Numeric_Value=000_008_0}', "");
Expect(1, 119664, '\P{^Numeric_Value=000_008_0}', "");
Expect(0, 119665, '\p{Numeric_Value=000_008_0}', "");
Expect(1, 119665, '\p{^Numeric_Value=000_008_0}', "");
Expect(1, 119665, '\P{Numeric_Value=000_008_0}', "");
Expect(0, 119665, '\P{^Numeric_Value=000_008_0}', "");
Error('\p{Nv=/a/-_000000080}');
Error('\P{Nv=/a/-_000000080}');
Expect(1, 119664, '\p{Nv=+0080}', "");
Expect(0, 119664, '\p{^Nv=+0080}', "");
Expect(0, 119664, '\P{Nv=+0080}', "");
Expect(1, 119664, '\P{^Nv=+0080}', "");
Expect(0, 119665, '\p{Nv=+0080}', "");
Expect(1, 119665, '\p{^Nv=+0080}', "");
Expect(1, 119665, '\P{Nv=+0080}', "");
Expect(0, 119665, '\P{^Nv=+0080}', "");
Error('\p{Is_Numeric_Value=_/a/+0_0_0_0_0_0_0080}');
Error('\P{Is_Numeric_Value=_/a/+0_0_0_0_0_0_0080}');
Expect(1, 119664, '\p{Is_Numeric_Value=000080}', "");
Expect(0, 119664, '\p{^Is_Numeric_Value=000080}', "");
Expect(0, 119664, '\P{Is_Numeric_Value=000080}', "");
Expect(1, 119664, '\P{^Is_Numeric_Value=000080}', "");
Expect(0, 119665, '\p{Is_Numeric_Value=000080}', "");
Expect(1, 119665, '\p{^Is_Numeric_Value=000080}', "");
Expect(1, 119665, '\P{Is_Numeric_Value=000080}', "");
Expect(0, 119665, '\P{^Is_Numeric_Value=000080}', "");
Error('\p{Is_Nv=/a/_+08_0}');
Error('\P{Is_Nv=/a/_+08_0}');
Expect(1, 119664, '\p{Is_Nv=0_0_0_0_0_0_0_80}', "");
Expect(0, 119664, '\p{^Is_Nv=0_0_0_0_0_0_0_80}', "");
Expect(0, 119664, '\P{Is_Nv=0_0_0_0_0_0_0_80}', "");
Expect(1, 119664, '\P{^Is_Nv=0_0_0_0_0_0_0_80}', "");
Expect(0, 119665, '\p{Is_Nv=0_0_0_0_0_0_0_80}', "");
Expect(1, 119665, '\p{^Is_Nv=0_0_0_0_0_0_0_80}', "");
Expect(1, 119665, '\P{Is_Nv=0_0_0_0_0_0_0_80}', "");
Expect(0, 119665, '\P{^Is_Nv=0_0_0_0_0_0_0_80}', "");
Error('\p{Numeric_Value:	 +00000800/a/}');
Error('\P{Numeric_Value:	 +00000800/a/}');
Expect(1, 69241, '\p{Numeric_Value=000000000800}', "");
Expect(0, 69241, '\p{^Numeric_Value=000000000800}', "");
Expect(0, 69241, '\P{Numeric_Value=000000000800}', "");
Expect(1, 69241, '\P{^Numeric_Value=000000000800}', "");
Expect(0, 69242, '\p{Numeric_Value=000000000800}', "");
Expect(1, 69242, '\p{^Numeric_Value=000000000800}', "");
Expect(1, 69242, '\P{Numeric_Value=000000000800}', "");
Expect(0, 69242, '\P{^Numeric_Value=000000000800}', "");
Error('\p{Nv=	:=0000000800}');
Error('\P{Nv=	:=0000000800}');
Expect(1, 69241, '\p{Nv=+0000000800}', "");
Expect(0, 69241, '\p{^Nv=+0000000800}', "");
Expect(0, 69241, '\P{Nv=+0000000800}', "");
Expect(1, 69241, '\P{^Nv=+0000000800}', "");
Expect(0, 69242, '\p{Nv=+0000000800}', "");
Expect(1, 69242, '\p{^Nv=+0000000800}', "");
Expect(1, 69242, '\P{Nv=+0000000800}', "");
Expect(0, 69242, '\P{^Nv=+0000000800}', "");
Error('\p{Is_Numeric_Value=-:=00_00_80_0}');
Error('\P{Is_Numeric_Value=-:=00_00_80_0}');
Expect(1, 69241, '\p{Is_Numeric_Value=00000080_0}', "");
Expect(0, 69241, '\p{^Is_Numeric_Value=00000080_0}', "");
Expect(0, 69241, '\P{Is_Numeric_Value=00000080_0}', "");
Expect(1, 69241, '\P{^Is_Numeric_Value=00000080_0}', "");
Expect(0, 69242, '\p{Is_Numeric_Value=00000080_0}', "");
Expect(1, 69242, '\p{^Is_Numeric_Value=00000080_0}', "");
Expect(1, 69242, '\P{Is_Numeric_Value=00000080_0}', "");
Expect(0, 69242, '\P{^Is_Numeric_Value=00000080_0}', "");
Error('\p{Is_Nv= _+800:=}');
Error('\P{Is_Nv= _+800:=}');
Expect(1, 69241, '\p{Is_Nv=0000800}', "");
Expect(0, 69241, '\p{^Is_Nv=0000800}', "");
Expect(0, 69241, '\P{Is_Nv=0000800}', "");
Expect(1, 69241, '\P{^Is_Nv=0000800}', "");
Expect(0, 69242, '\p{Is_Nv=0000800}', "");
Expect(1, 69242, '\p{^Is_Nv=0000800}', "");
Expect(1, 69242, '\P{Is_Nv=0000800}', "");
Expect(0, 69242, '\P{^Is_Nv=0000800}', "");
Error('\p{Numeric_Value=/a/__000_800_0}');
Error('\P{Numeric_Value=/a/__000_800_0}');
Expect(1, 65833, '\p{Numeric_Value=0000008000}', "");
Expect(0, 65833, '\p{^Numeric_Value=0000008000}', "");
Expect(0, 65833, '\P{Numeric_Value=0000008000}', "");
Expect(1, 65833, '\P{^Numeric_Value=0000008000}', "");
Expect(0, 65834, '\p{Numeric_Value=0000008000}', "");
Expect(1, 65834, '\p{^Numeric_Value=0000008000}', "");
Expect(1, 65834, '\P{Numeric_Value=0000008000}', "");
Expect(0, 65834, '\P{^Numeric_Value=0000008000}', "");
Error('\p{Nv:	 000000008000:=}');
Error('\P{Nv:	 000000008000:=}');
Expect(1, 65833, '\p{Nv=+00800_0}', "");
Expect(0, 65833, '\p{^Nv=+00800_0}', "");
Expect(0, 65833, '\P{Nv=+00800_0}', "");
Expect(1, 65833, '\P{^Nv=+00800_0}', "");
Expect(0, 65834, '\p{Nv=+00800_0}', "");
Expect(1, 65834, '\p{^Nv=+00800_0}', "");
Expect(1, 65834, '\P{Nv=+00800_0}', "");
Expect(0, 65834, '\P{^Nv=+00800_0}', "");
Error('\p{Is_Numeric_Value=/a/_000000008000}');
Error('\P{Is_Numeric_Value=/a/_000000008000}');
Expect(1, 65833, '\p{Is_Numeric_Value=00008000}', "");
Expect(0, 65833, '\p{^Is_Numeric_Value=00008000}', "");
Expect(0, 65833, '\P{Is_Numeric_Value=00008000}', "");
Expect(1, 65833, '\P{^Is_Numeric_Value=00008000}', "");
Expect(0, 65834, '\p{Is_Numeric_Value=00008000}', "");
Expect(1, 65834, '\p{^Is_Numeric_Value=00008000}', "");
Expect(1, 65834, '\P{Is_Numeric_Value=00008000}', "");
Expect(0, 65834, '\P{^Is_Numeric_Value=00008000}', "");
Error('\p{Is_Nv:    :=800_0}');
Error('\P{Is_Nv:    :=800_0}');
Expect(1, 65833, '\p{Is_Nv=000008000}', "");
Expect(0, 65833, '\p{^Is_Nv=000008000}', "");
Expect(0, 65833, '\P{Is_Nv=000008000}', "");
Expect(1, 65833, '\P{^Is_Nv=000008000}', "");
Expect(0, 65834, '\p{Is_Nv=000008000}', "");
Expect(1, 65834, '\p{^Is_Nv=000008000}', "");
Expect(1, 65834, '\P{Is_Nv=000008000}', "");
Expect(0, 65834, '\P{^Is_Nv=000008000}', "");
Error('\p{Numeric_Value=:=	_0000000080000}');
Error('\P{Numeric_Value=:=	_0000000080000}');
Expect(1, 65842, '\p{Numeric_Value=00_00_80_000}', "");
Expect(0, 65842, '\p{^Numeric_Value=00_00_80_000}', "");
Expect(0, 65842, '\P{Numeric_Value=00_00_80_000}', "");
Expect(1, 65842, '\P{^Numeric_Value=00_00_80_000}', "");
Expect(0, 65843, '\p{Numeric_Value=00_00_80_000}', "");
Expect(1, 65843, '\p{^Numeric_Value=00_00_80_000}', "");
Expect(1, 65843, '\P{Numeric_Value=00_00_80_000}', "");
Expect(0, 65843, '\P{^Numeric_Value=00_00_80_000}', "");
Error('\p{Nv=:= 0_0_0_0_0_0_0_0_8_0_000}');
Error('\P{Nv=:= 0_0_0_0_0_0_0_0_8_0_000}');
Expect(1, 65842, '\p{Nv=0_0_0_0_0_0_0_0_080000}', "");
Expect(0, 65842, '\p{^Nv=0_0_0_0_0_0_0_0_080000}', "");
Expect(0, 65842, '\P{Nv=0_0_0_0_0_0_0_0_080000}', "");
Expect(1, 65842, '\P{^Nv=0_0_0_0_0_0_0_0_080000}', "");
Expect(0, 65843, '\p{Nv=0_0_0_0_0_0_0_0_080000}', "");
Expect(1, 65843, '\p{^Nv=0_0_0_0_0_0_0_0_080000}', "");
Expect(1, 65843, '\P{Nv=0_0_0_0_0_0_0_0_080000}', "");
Expect(0, 65843, '\P{^Nv=0_0_0_0_0_0_0_0_080000}', "");
Error('\p{Is_Numeric_Value: :=- +0_0_0_8_0000}');
Error('\P{Is_Numeric_Value: :=- +0_0_0_8_0000}');
Expect(1, 65842, '\p{Is_Numeric_Value: 0_0_0_0_0_0_0_0_0_8_0_0_00}', "");
Expect(0, 65842, '\p{^Is_Numeric_Value: 0_0_0_0_0_0_0_0_0_8_0_0_00}', "");
Expect(0, 65842, '\P{Is_Numeric_Value: 0_0_0_0_0_0_0_0_0_8_0_0_00}', "");
Expect(1, 65842, '\P{^Is_Numeric_Value: 0_0_0_0_0_0_0_0_0_8_0_0_00}', "");
Expect(0, 65843, '\p{Is_Numeric_Value: 0_0_0_0_0_0_0_0_0_8_0_0_00}', "");
Expect(1, 65843, '\p{^Is_Numeric_Value: 0_0_0_0_0_0_0_0_0_8_0_0_00}', "");
Expect(1, 65843, '\P{Is_Numeric_Value: 0_0_0_0_0_0_0_0_0_8_0_0_00}', "");
Expect(0, 65843, '\P{^Is_Numeric_Value: 0_0_0_0_0_0_0_0_0_8_0_0_00}', "");
Error('\p{Is_Nv= -+0000080000/a/}');
Error('\P{Is_Nv= -+0000080000/a/}');
Expect(1, 65842, '\p{Is_Nv=000_800_00}', "");
Expect(0, 65842, '\p{^Is_Nv=000_800_00}', "");
Expect(0, 65842, '\P{Is_Nv=000_800_00}', "");
Expect(1, 65842, '\P{^Is_Nv=000_800_00}', "");
Expect(0, 65843, '\p{Is_Nv=000_800_00}', "");
Expect(1, 65843, '\p{^Is_Nv=000_800_00}', "");
Expect(1, 65843, '\P{Is_Nv=000_800_00}', "");
Expect(0, 65843, '\P{^Is_Nv=000_800_00}', "");
Error('\p{Numeric_Value= :=00009}');
Error('\P{Numeric_Value= :=00009}');
Expect(1, 194704, '\p{Numeric_Value=+0_0_0_0_0_00009}', "");
Expect(0, 194704, '\p{^Numeric_Value=+0_0_0_0_0_00009}', "");
Expect(0, 194704, '\P{Numeric_Value=+0_0_0_0_0_00009}', "");
Expect(1, 194704, '\P{^Numeric_Value=+0_0_0_0_0_00009}', "");
Expect(0, 194705, '\p{Numeric_Value=+0_0_0_0_0_00009}', "");
Expect(1, 194705, '\p{^Numeric_Value=+0_0_0_0_0_00009}', "");
Expect(1, 194705, '\P{Numeric_Value=+0_0_0_0_0_00009}', "");
Expect(0, 194705, '\P{^Numeric_Value=+0_0_0_0_0_00009}', "");
Error('\p{Nv=/a/-+0000000009}');
Error('\P{Nv=/a/-+0000000009}');
Expect(1, 194704, '\p{Nv=0000009}', "");
Expect(0, 194704, '\p{^Nv=0000009}', "");
Expect(0, 194704, '\P{Nv=0000009}', "");
Expect(1, 194704, '\P{^Nv=0000009}', "");
Expect(0, 194705, '\p{Nv=0000009}', "");
Expect(1, 194705, '\p{^Nv=0000009}', "");
Expect(1, 194705, '\P{Nv=0000009}', "");
Expect(0, 194705, '\P{^Nv=0000009}', "");
Error('\p{Is_Numeric_Value=-	0_0_0_0_0_009:=}');
Error('\P{Is_Numeric_Value=-	0_0_0_0_0_009:=}');
Expect(1, 194704, '\p{Is_Numeric_Value=0000000009}', "");
Expect(0, 194704, '\p{^Is_Numeric_Value=0000000009}', "");
Expect(0, 194704, '\P{Is_Numeric_Value=0000000009}', "");
Expect(1, 194704, '\P{^Is_Numeric_Value=0000000009}', "");
Expect(0, 194705, '\p{Is_Numeric_Value=0000000009}', "");
Expect(1, 194705, '\p{^Is_Numeric_Value=0000000009}', "");
Expect(1, 194705, '\P{Is_Numeric_Value=0000000009}', "");
Expect(0, 194705, '\P{^Is_Numeric_Value=0000000009}', "");
Error('\p{Is_Nv=:=_-000_000_09}');
Error('\P{Is_Nv=:=_-000_000_09}');
Expect(1, 194704, '\p{Is_Nv=9}', "");
Expect(0, 194704, '\p{^Is_Nv=9}', "");
Expect(0, 194704, '\P{Is_Nv=9}', "");
Expect(1, 194704, '\P{^Is_Nv=9}', "");
Expect(0, 194705, '\p{Is_Nv=9}', "");
Expect(1, 194705, '\p{^Is_Nv=9}', "");
Expect(1, 194705, '\P{Is_Nv=9}', "");
Expect(0, 194705, '\P{^Is_Nv=9}', "");
Error('\p{Numeric_Value=:=__9/000000002}');
Error('\P{Numeric_Value=:=__9/000000002}');
Expect(1, 3886, '\p{Numeric_Value=0000009/2}', "");
Expect(0, 3886, '\p{^Numeric_Value=0000009/2}', "");
Expect(0, 3886, '\P{Numeric_Value=0000009/2}', "");
Expect(1, 3886, '\P{^Numeric_Value=0000009/2}', "");
Expect(0, 3887, '\p{Numeric_Value=0000009/2}', "");
Expect(1, 3887, '\p{^Numeric_Value=0000009/2}', "");
Expect(1, 3887, '\P{Numeric_Value=0000009/2}', "");
Expect(0, 3887, '\P{^Numeric_Value=0000009/2}', "");
Expect(1, 3886, '\p{Numeric_Value=4.500}', "");
Expect(0, 3886, '\p{^Numeric_Value=4.500}', "");
Expect(0, 3886, '\P{Numeric_Value=4.500}', "");
Expect(1, 3886, '\P{^Numeric_Value=4.500}', "");
Expect(0, 3887, '\p{Numeric_Value=4.500}', "");
Expect(1, 3887, '\p{^Numeric_Value=4.500}', "");
Expect(1, 3887, '\P{Numeric_Value=4.500}', "");
Expect(0, 3887, '\P{^Numeric_Value=4.500}', "");
Error('\p{Nv=_:=+9/00000002}');
Error('\P{Nv=_:=+9/00000002}');
Expect(1, 3886, '\p{Nv=00000009/2}', "");
Expect(0, 3886, '\p{^Nv=00000009/2}', "");
Expect(0, 3886, '\P{Nv=00000009/2}', "");
Expect(1, 3886, '\P{^Nv=00000009/2}', "");
Expect(0, 3887, '\p{Nv=00000009/2}', "");
Expect(1, 3887, '\p{^Nv=00000009/2}', "");
Expect(1, 3887, '\P{Nv=00000009/2}', "");
Expect(0, 3887, '\P{^Nv=00000009/2}', "");
Expect(1, 3886, '\p{Nv=4.500}', "");
Expect(0, 3886, '\p{^Nv=4.500}', "");
Expect(0, 3886, '\P{Nv=4.500}', "");
Expect(1, 3886, '\P{^Nv=4.500}', "");
Expect(0, 3887, '\p{Nv=4.500}', "");
Expect(1, 3887, '\p{^Nv=4.500}', "");
Expect(1, 3887, '\P{Nv=4.500}', "");
Expect(0, 3887, '\P{^Nv=4.500}', "");
Error('\p{Is_Numeric_Value=/a/__+9/0002}');
Error('\P{Is_Numeric_Value=/a/__+9/0002}');
Expect(1, 3886, '\p{Is_Numeric_Value=9/0000000002}', "");
Expect(0, 3886, '\p{^Is_Numeric_Value=9/0000000002}', "");
Expect(0, 3886, '\P{Is_Numeric_Value=9/0000000002}', "");
Expect(1, 3886, '\P{^Is_Numeric_Value=9/0000000002}', "");
Expect(0, 3887, '\p{Is_Numeric_Value=9/0000000002}', "");
Expect(1, 3887, '\p{^Is_Numeric_Value=9/0000000002}', "");
Expect(1, 3887, '\P{Is_Numeric_Value=9/0000000002}', "");
Expect(0, 3887, '\P{^Is_Numeric_Value=9/0000000002}', "");
Expect(1, 3886, '\p{Is_Numeric_Value=4.500}', "");
Expect(0, 3886, '\p{^Is_Numeric_Value=4.500}', "");
Expect(0, 3886, '\P{Is_Numeric_Value=4.500}', "");
Expect(1, 3886, '\P{^Is_Numeric_Value=4.500}', "");
Expect(0, 3887, '\p{Is_Numeric_Value=4.500}', "");
Expect(1, 3887, '\p{^Is_Numeric_Value=4.500}', "");
Expect(1, 3887, '\P{Is_Numeric_Value=4.500}', "");
Expect(0, 3887, '\P{^Is_Numeric_Value=4.500}', "");
Error('\p{Is_Nv=/a/ 	00000009/0000002}');
Error('\P{Is_Nv=/a/ 	00000009/0000002}');
Expect(1, 3886, '\p{Is_Nv=00009/000000002}', "");
Expect(0, 3886, '\p{^Is_Nv=00009/000000002}', "");
Expect(0, 3886, '\P{Is_Nv=00009/000000002}', "");
Expect(1, 3886, '\P{^Is_Nv=00009/000000002}', "");
Expect(0, 3887, '\p{Is_Nv=00009/000000002}', "");
Expect(1, 3887, '\p{^Is_Nv=00009/000000002}', "");
Expect(1, 3887, '\P{Is_Nv=00009/000000002}', "");
Expect(0, 3887, '\P{^Is_Nv=00009/000000002}', "");
Expect(1, 3886, '\p{Is_Nv=4.500}', "");
Expect(0, 3886, '\p{^Is_Nv=4.500}', "");
Expect(0, 3886, '\P{Is_Nv=4.500}', "");
Expect(1, 3886, '\P{^Is_Nv=4.500}', "");
Expect(0, 3887, '\p{Is_Nv=4.500}', "");
Expect(1, 3887, '\p{^Is_Nv=4.500}', "");
Expect(1, 3887, '\P{Is_Nv=4.500}', "");
Expect(0, 3887, '\P{^Is_Nv=4.500}', "");
Error('\p{Numeric_Value=_	000090/a/}');
Error('\P{Numeric_Value=_	000090/a/}');
Expect(1, 119665, '\p{Numeric_Value=00090}', "");
Expect(0, 119665, '\p{^Numeric_Value=00090}', "");
Expect(0, 119665, '\P{Numeric_Value=00090}', "");
Expect(1, 119665, '\P{^Numeric_Value=00090}', "");
Expect(0, 119666, '\p{Numeric_Value=00090}', "");
Expect(1, 119666, '\p{^Numeric_Value=00090}', "");
Expect(1, 119666, '\P{Numeric_Value=00090}', "");
Expect(0, 119666, '\P{^Numeric_Value=00090}', "");
Error('\p{Nv:/a/90}');
Error('\P{Nv:/a/90}');
Expect(1, 119665, '\p{Nv=0_0_0_0_0_0_0_0090}', "");
Expect(0, 119665, '\p{^Nv=0_0_0_0_0_0_0_0090}', "");
Expect(0, 119665, '\P{Nv=0_0_0_0_0_0_0_0090}', "");
Expect(1, 119665, '\P{^Nv=0_0_0_0_0_0_0_0090}', "");
Expect(0, 119666, '\p{Nv=0_0_0_0_0_0_0_0090}', "");
Expect(1, 119666, '\p{^Nv=0_0_0_0_0_0_0_0090}', "");
Expect(1, 119666, '\P{Nv=0_0_0_0_0_0_0_0090}', "");
Expect(0, 119666, '\P{^Nv=0_0_0_0_0_0_0_0090}', "");
Error('\p{Is_Numeric_Value=_-000000090/a/}');
Error('\P{Is_Numeric_Value=_-000000090/a/}');
Expect(1, 119665, '\p{Is_Numeric_Value=00090}', "");
Expect(0, 119665, '\p{^Is_Numeric_Value=00090}', "");
Expect(0, 119665, '\P{Is_Numeric_Value=00090}', "");
Expect(1, 119665, '\P{^Is_Numeric_Value=00090}', "");
Expect(0, 119666, '\p{Is_Numeric_Value=00090}', "");
Expect(1, 119666, '\p{^Is_Numeric_Value=00090}', "");
Expect(1, 119666, '\P{Is_Numeric_Value=00090}', "");
Expect(0, 119666, '\P{^Is_Numeric_Value=00090}', "");
Error('\p{Is_Nv=:=_	+00000000090}');
Error('\P{Is_Nv=:=_	+00000000090}');
Expect(1, 119665, '\p{Is_Nv=+00_00_90}', "");
Expect(0, 119665, '\p{^Is_Nv=+00_00_90}', "");
Expect(0, 119665, '\P{Is_Nv=+00_00_90}', "");
Expect(1, 119665, '\P{^Is_Nv=+00_00_90}', "");
Expect(0, 119666, '\p{Is_Nv=+00_00_90}', "");
Expect(1, 119666, '\p{^Is_Nv=+00_00_90}', "");
Expect(1, 119666, '\P{Is_Nv=+00_00_90}', "");
Expect(0, 119666, '\P{^Is_Nv=+00_00_90}', "");
Error('\p{Numeric_Value=	/a/+00000000090_0}');
Error('\P{Numeric_Value=	/a/+00000000090_0}');
Expect(1, 69242, '\p{Numeric_Value=0_0_0_0_0_900}', "");
Expect(0, 69242, '\p{^Numeric_Value=0_0_0_0_0_900}', "");
Expect(0, 69242, '\P{Numeric_Value=0_0_0_0_0_900}', "");
Expect(1, 69242, '\P{^Numeric_Value=0_0_0_0_0_900}', "");
Expect(0, 69243, '\p{Numeric_Value=0_0_0_0_0_900}', "");
Expect(1, 69243, '\p{^Numeric_Value=0_0_0_0_0_900}', "");
Expect(1, 69243, '\P{Numeric_Value=0_0_0_0_0_900}', "");
Expect(0, 69243, '\P{^Numeric_Value=0_0_0_0_0_900}', "");
Error('\p{Nv=:=_0000000090_0}');
Error('\P{Nv=:=_0000000090_0}');
Expect(1, 69242, '\p{Nv=+00000000900}', "");
Expect(0, 69242, '\p{^Nv=+00000000900}', "");
Expect(0, 69242, '\P{Nv=+00000000900}', "");
Expect(1, 69242, '\P{^Nv=+00000000900}', "");
Expect(0, 69243, '\p{Nv=+00000000900}', "");
Expect(1, 69243, '\p{^Nv=+00000000900}', "");
Expect(1, 69243, '\P{Nv=+00000000900}', "");
Expect(0, 69243, '\P{^Nv=+00000000900}', "");
Error('\p{Is_Numeric_Value=:=	 00_90_0}');
Error('\P{Is_Numeric_Value=:=	 00_90_0}');
Expect(1, 69242, '\p{Is_Numeric_Value=0900}', "");
Expect(0, 69242, '\p{^Is_Numeric_Value=0900}', "");
Expect(0, 69242, '\P{Is_Numeric_Value=0900}', "");
Expect(1, 69242, '\P{^Is_Numeric_Value=0900}', "");
Expect(0, 69243, '\p{Is_Numeric_Value=0900}', "");
Expect(1, 69243, '\p{^Is_Numeric_Value=0900}', "");
Expect(1, 69243, '\P{Is_Numeric_Value=0900}', "");
Expect(0, 69243, '\P{^Is_Numeric_Value=0900}', "");
Error('\p{Is_Nv=  900:=}');
Error('\P{Is_Nv=  900:=}');
Expect(1, 69242, '\p{Is_Nv=0000000090_0}', "");
Expect(0, 69242, '\p{^Is_Nv=0000000090_0}', "");
Expect(0, 69242, '\P{Is_Nv=0000000090_0}', "");
Expect(1, 69242, '\P{^Is_Nv=0000000090_0}', "");
Expect(0, 69243, '\p{Is_Nv=0000000090_0}', "");
Expect(1, 69243, '\p{^Is_Nv=0000000090_0}', "");
Expect(1, 69243, '\P{Is_Nv=0000000090_0}', "");
Expect(0, 69243, '\P{^Is_Nv=0000000090_0}', "");
Error('\p{Numeric_Value=/a/ 00000_00090_00}');
Error('\P{Numeric_Value=/a/ 00000_00090_00}');
Expect(1, 65834, '\p{Numeric_Value=+9000}', "");
Expect(0, 65834, '\p{^Numeric_Value=+9000}', "");
Expect(0, 65834, '\P{Numeric_Value=+9000}', "");
Expect(1, 65834, '\P{^Numeric_Value=+9000}', "");
Expect(0, 65835, '\p{Numeric_Value=+9000}', "");
Expect(1, 65835, '\p{^Numeric_Value=+9000}', "");
Expect(1, 65835, '\P{Numeric_Value=+9000}', "");
Expect(0, 65835, '\P{^Numeric_Value=+9000}', "");
Error('\p{Nv= 0000_0900_0:=}');
Error('\P{Nv= 0000_0900_0:=}');
Expect(1, 65834, '\p{Nv=000000900_0}', "");
Expect(0, 65834, '\p{^Nv=000000900_0}', "");
Expect(0, 65834, '\P{Nv=000000900_0}', "");
Expect(1, 65834, '\P{^Nv=000000900_0}', "");
Expect(0, 65835, '\p{Nv=000000900_0}', "");
Expect(1, 65835, '\p{^Nv=000000900_0}', "");
Expect(1, 65835, '\P{Nv=000000900_0}', "");
Expect(0, 65835, '\P{^Nv=000000900_0}', "");
Error('\p{Is_Numeric_Value:-/a/+0000000009000}');
Error('\P{Is_Numeric_Value:-/a/+0000000009000}');
Expect(1, 65834, '\p{Is_Numeric_Value=00_00_00_09_000}', "");
Expect(0, 65834, '\p{^Is_Numeric_Value=00_00_00_09_000}', "");
Expect(0, 65834, '\P{Is_Numeric_Value=00_00_00_09_000}', "");
Expect(1, 65834, '\P{^Is_Numeric_Value=00_00_00_09_000}', "");
Expect(0, 65835, '\p{Is_Numeric_Value=00_00_00_09_000}', "");
Expect(1, 65835, '\p{^Is_Numeric_Value=00_00_00_09_000}', "");
Expect(1, 65835, '\P{Is_Numeric_Value=00_00_00_09_000}', "");
Expect(0, 65835, '\P{^Is_Numeric_Value=00_00_00_09_000}', "");
Error('\p{Is_Nv=	/a/0_0_9_000}');
Error('\P{Is_Nv=	/a/0_0_9_000}');
Expect(1, 65834, '\p{Is_Nv=+0000009000}', "");
Expect(0, 65834, '\p{^Is_Nv=+0000009000}', "");
Expect(0, 65834, '\P{Is_Nv=+0000009000}', "");
Expect(1, 65834, '\P{^Is_Nv=+0000009000}', "");
Expect(0, 65835, '\p{Is_Nv=+0000009000}', "");
Expect(1, 65835, '\p{^Is_Nv=+0000009000}', "");
Expect(1, 65835, '\P{Is_Nv=+0000009000}', "");
Expect(0, 65835, '\P{^Is_Nv=+0000009000}', "");
Error('\p{Numeric_Value=	_+09_00_00/a/}');
Error('\P{Numeric_Value=	_+09_00_00/a/}');
Expect(1, 65843, '\p{Numeric_Value=+00000000090000}', "");
Expect(0, 65843, '\p{^Numeric_Value=+00000000090000}', "");
Expect(0, 65843, '\P{Numeric_Value=+00000000090000}', "");
Expect(1, 65843, '\P{^Numeric_Value=+00000000090000}', "");
Expect(0, 65844, '\p{Numeric_Value=+00000000090000}', "");
Expect(1, 65844, '\p{^Numeric_Value=+00000000090000}', "");
Expect(1, 65844, '\P{Numeric_Value=+00000000090000}', "");
Expect(0, 65844, '\P{^Numeric_Value=+00000000090000}', "");
Error('\p{Nv=-	90000:=}');
Error('\P{Nv=-	90000:=}');
Expect(1, 65843, '\p{Nv=00090000}', "");
Expect(0, 65843, '\p{^Nv=00090000}', "");
Expect(0, 65843, '\P{Nv=00090000}', "");
Expect(1, 65843, '\P{^Nv=00090000}', "");
Expect(0, 65844, '\p{Nv=00090000}', "");
Expect(1, 65844, '\p{^Nv=00090000}', "");
Expect(1, 65844, '\P{Nv=00090000}', "");
Expect(0, 65844, '\P{^Nv=00090000}', "");
Error('\p{Is_Numeric_Value=	/a/0_0_9_0_000}');
Error('\P{Is_Numeric_Value=	/a/0_0_9_0_000}');
Expect(1, 65843, '\p{Is_Numeric_Value=0_0_0_0_0_0_9_0000}', "");
Expect(0, 65843, '\p{^Is_Numeric_Value=0_0_0_0_0_0_9_0000}', "");
Expect(0, 65843, '\P{Is_Numeric_Value=0_0_0_0_0_0_9_0000}', "");
Expect(1, 65843, '\P{^Is_Numeric_Value=0_0_0_0_0_0_9_0000}', "");
Expect(0, 65844, '\p{Is_Numeric_Value=0_0_0_0_0_0_9_0000}', "");
Expect(1, 65844, '\p{^Is_Numeric_Value=0_0_0_0_0_0_9_0000}', "");
Expect(1, 65844, '\P{Is_Numeric_Value=0_0_0_0_0_0_9_0000}', "");
Expect(0, 65844, '\P{^Is_Numeric_Value=0_0_0_0_0_0_9_0000}', "");
Error('\p{Is_Nv=_	0_0_0_0_0_0_0_9_0_0_00:=}');
Error('\P{Is_Nv=_	0_0_0_0_0_0_0_9_0_0_00:=}');
Expect(1, 65843, '\p{Is_Nv:	00000090000}', "");
Expect(0, 65843, '\p{^Is_Nv:	00000090000}', "");
Expect(0, 65843, '\P{Is_Nv:	00000090000}', "");
Expect(1, 65843, '\P{^Is_Nv:	00000090000}', "");
Expect(0, 65844, '\p{Is_Nv:	00000090000}', "");
Expect(1, 65844, '\p{^Is_Nv:	00000090000}', "");
Expect(1, 65844, '\P{Is_Nv:	00000090000}', "");
Expect(0, 65844, '\P{^Is_Nv:	00000090000}', "");
Error('\p{Numeric_Value=-:=NaN}');
Error('\P{Numeric_Value=-:=NaN}');
Expect(1, 194705, '\p{Numeric_Value=nan}', "");
Expect(0, 194705, '\p{^Numeric_Value=nan}', "");
Expect(0, 194705, '\P{Numeric_Value=nan}', "");
Expect(1, 194705, '\P{^Numeric_Value=nan}', "");
Expect(0, 194704, '\p{Numeric_Value=nan}', "");
Expect(1, 194704, '\p{^Numeric_Value=nan}', "");
Expect(1, 194704, '\P{Numeric_Value=nan}', "");
Expect(0, 194704, '\P{^Numeric_Value=nan}', "");
Expect(1, 194705, '\p{Numeric_Value=_nan}', "");
Expect(0, 194705, '\p{^Numeric_Value=_nan}', "");
Expect(0, 194705, '\P{Numeric_Value=_nan}', "");
Expect(1, 194705, '\P{^Numeric_Value=_nan}', "");
Expect(0, 194704, '\p{Numeric_Value=_nan}', "");
Expect(1, 194704, '\p{^Numeric_Value=_nan}', "");
Expect(1, 194704, '\P{Numeric_Value=_nan}', "");
Expect(0, 194704, '\P{^Numeric_Value=_nan}', "");
Error('\p{Nv= 	NaN:=}');
Error('\P{Nv= 	NaN:=}');
Expect(1, 194705, '\p{Nv:	nan}', "");
Expect(0, 194705, '\p{^Nv:	nan}', "");
Expect(0, 194705, '\P{Nv:	nan}', "");
Expect(1, 194705, '\P{^Nv:	nan}', "");
Expect(0, 194704, '\p{Nv:	nan}', "");
Expect(1, 194704, '\p{^Nv:	nan}', "");
Expect(1, 194704, '\P{Nv:	nan}', "");
Expect(0, 194704, '\P{^Nv:	nan}', "");
Expect(1, 194705, '\p{Nv=_	NaN}', "");
Expect(0, 194705, '\p{^Nv=_	NaN}', "");
Expect(0, 194705, '\P{Nv=_	NaN}', "");
Expect(1, 194705, '\P{^Nv=_	NaN}', "");
Expect(0, 194704, '\p{Nv=_	NaN}', "");
Expect(1, 194704, '\p{^Nv=_	NaN}', "");
Expect(1, 194704, '\P{Nv=_	NaN}', "");
Expect(0, 194704, '\P{^Nv=_	NaN}', "");
Error('\p{Is_Numeric_Value=/a/-	nan}');
Error('\P{Is_Numeric_Value=/a/-	nan}');
Expect(1, 194705, '\p{Is_Numeric_Value=nan}', "");
Expect(0, 194705, '\p{^Is_Numeric_Value=nan}', "");
Expect(0, 194705, '\P{Is_Numeric_Value=nan}', "");
Expect(1, 194705, '\P{^Is_Numeric_Value=nan}', "");
Expect(0, 194704, '\p{Is_Numeric_Value=nan}', "");
Expect(1, 194704, '\p{^Is_Numeric_Value=nan}', "");
Expect(1, 194704, '\P{Is_Numeric_Value=nan}', "");
Expect(0, 194704, '\P{^Is_Numeric_Value=nan}', "");
Expect(1, 194705, '\p{Is_Numeric_Value=- NAN}', "");
Expect(0, 194705, '\p{^Is_Numeric_Value=- NAN}', "");
Expect(0, 194705, '\P{Is_Numeric_Value=- NAN}', "");
Expect(1, 194705, '\P{^Is_Numeric_Value=- NAN}', "");
Expect(0, 194704, '\p{Is_Numeric_Value=- NAN}', "");
Expect(1, 194704, '\p{^Is_Numeric_Value=- NAN}', "");
Expect(1, 194704, '\P{Is_Numeric_Value=- NAN}', "");
Expect(0, 194704, '\P{^Is_Numeric_Value=- NAN}', "");
Error('\p{Is_Nv=:= NAN}');
Error('\P{Is_Nv=:= NAN}');
Expect(1, 194705, '\p{Is_Nv=nan}', "");
Expect(0, 194705, '\p{^Is_Nv=nan}', "");
Expect(0, 194705, '\P{Is_Nv=nan}', "");
Expect(1, 194705, '\P{^Is_Nv=nan}', "");
Expect(0, 194704, '\p{Is_Nv=nan}', "");
Expect(1, 194704, '\p{^Is_Nv=nan}', "");
Expect(1, 194704, '\P{Is_Nv=nan}', "");
Expect(0, 194704, '\P{^Is_Nv=nan}', "");
Expect(1, 194705, '\p{Is_Nv=_ NaN}', "");
Expect(0, 194705, '\p{^Is_Nv=_ NaN}', "");
Expect(0, 194705, '\P{Is_Nv=_ NaN}', "");
Expect(1, 194705, '\P{^Is_Nv=_ NaN}', "");
Expect(0, 194704, '\p{Is_Nv=_ NaN}', "");
Expect(1, 194704, '\p{^Is_Nv=_ NaN}', "");
Expect(1, 194704, '\P{Is_Nv=_ NaN}', "");
Expect(0, 194704, '\P{^Is_Nv=_ NaN}', "");
Error('\p{:=_-ASCII}');
Error('\P{:=_-ASCII}');
Expect(1, 127, '\p{ascii}', "");
Expect(0, 127, '\p{^ascii}', "");
Expect(0, 127, '\P{ascii}', "");
Expect(1, 127, '\P{^ascii}', "");
Expect(0, 128, '\p{ascii}', "");
Expect(1, 128, '\p{^ascii}', "");
Expect(1, 128, '\P{ascii}', "");
Expect(0, 128, '\P{^ascii}', "");
Expect(1, 127, '\p{ ascii}', "");
Expect(0, 127, '\p{^ ascii}', "");
Expect(0, 127, '\P{ ascii}', "");
Expect(1, 127, '\P{^ ascii}', "");
Expect(0, 128, '\p{ ascii}', "");
Expect(1, 128, '\p{^ ascii}', "");
Expect(1, 128, '\P{ ascii}', "");
Expect(0, 128, '\P{^ ascii}', "");
Error('\p{	Is_ASCII:=}');
Error('\P{	Is_ASCII:=}');
Expect(1, 127, '\p{isascii}', "");
Expect(0, 127, '\p{^isascii}', "");
Expect(0, 127, '\P{isascii}', "");
Expect(1, 127, '\P{^isascii}', "");
Expect(0, 128, '\p{isascii}', "");
Expect(1, 128, '\p{^isascii}', "");
Expect(1, 128, '\P{isascii}', "");
Expect(0, 128, '\P{^isascii}', "");
Expect(1, 127, '\p{	IS_ASCII}', "");
Expect(0, 127, '\p{^	IS_ASCII}', "");
Expect(0, 127, '\P{	IS_ASCII}', "");
Expect(1, 127, '\P{^	IS_ASCII}', "");
Expect(0, 128, '\p{	IS_ASCII}', "");
Expect(1, 128, '\p{^	IS_ASCII}', "");
Expect(1, 128, '\P{	IS_ASCII}', "");
Expect(0, 128, '\P{^	IS_ASCII}', "");
Error('\p{__basic_Latin:=}');
Error('\P{__basic_Latin:=}');
Expect(1, 127, '\p{basiclatin}', "");
Expect(0, 127, '\p{^basiclatin}', "");
Expect(0, 127, '\P{basiclatin}', "");
Expect(1, 127, '\P{^basiclatin}', "");
Expect(0, 128, '\p{basiclatin}', "");
Expect(1, 128, '\p{^basiclatin}', "");
Expect(1, 128, '\P{basiclatin}', "");
Expect(0, 128, '\P{^basiclatin}', "");
Expect(1, 127, '\p{--Basic_LATIN}', "");
Expect(0, 127, '\p{^--Basic_LATIN}', "");
Expect(0, 127, '\P{--Basic_LATIN}', "");
Expect(1, 127, '\P{^--Basic_LATIN}', "");
Expect(0, 128, '\p{--Basic_LATIN}', "");
Expect(1, 128, '\p{^--Basic_LATIN}', "");
Expect(1, 128, '\P{--Basic_LATIN}', "");
Expect(0, 128, '\P{^--Basic_LATIN}', "");
Error('\p{ Is_Basic_Latin/a/}');
Error('\P{ Is_Basic_Latin/a/}');
Expect(1, 127, '\p{isbasiclatin}', "");
Expect(0, 127, '\p{^isbasiclatin}', "");
Expect(0, 127, '\P{isbasiclatin}', "");
Expect(1, 127, '\P{^isbasiclatin}', "");
Expect(0, 128, '\p{isbasiclatin}', "");
Expect(1, 128, '\p{^isbasiclatin}', "");
Expect(1, 128, '\P{isbasiclatin}', "");
Expect(0, 128, '\P{^isbasiclatin}', "");
Expect(1, 127, '\p{  IS_basic_LATIN}', "");
Expect(0, 127, '\p{^  IS_basic_LATIN}', "");
Expect(0, 127, '\P{  IS_basic_LATIN}', "");
Expect(1, 127, '\P{^  IS_basic_LATIN}', "");
Expect(0, 128, '\p{  IS_basic_LATIN}', "");
Expect(1, 128, '\p{^  IS_basic_LATIN}', "");
Expect(1, 128, '\P{  IS_basic_LATIN}', "");
Expect(0, 128, '\P{^  IS_basic_LATIN}', "");
Error('\p{-/a/In_Basic_Latin}');
Error('\P{-/a/In_Basic_Latin}');
Expect(1, 127, '\p{inbasiclatin}', "");
Expect(0, 127, '\p{^inbasiclatin}', "");
Expect(0, 127, '\P{inbasiclatin}', "");
Expect(1, 127, '\P{^inbasiclatin}', "");
Expect(0, 128, '\p{inbasiclatin}', "");
Expect(1, 128, '\p{^inbasiclatin}', "");
Expect(1, 128, '\P{inbasiclatin}', "");
Expect(0, 128, '\P{^inbasiclatin}', "");
Expect(1, 127, '\p{-In_Basic_Latin}', "");
Expect(0, 127, '\p{^-In_Basic_Latin}', "");
Expect(0, 127, '\P{-In_Basic_Latin}', "");
Expect(1, 127, '\P{^-In_Basic_Latin}', "");
Expect(0, 128, '\p{-In_Basic_Latin}', "");
Expect(1, 128, '\p{^-In_Basic_Latin}', "");
Expect(1, 128, '\P{-In_Basic_Latin}', "");
Expect(0, 128, '\P{^-In_Basic_Latin}', "");
Error('\p{/a/	IN_ASCII}');
Error('\P{/a/	IN_ASCII}');
Expect(1, 127, '\p{inascii}', "");
Expect(0, 127, '\p{^inascii}', "");
Expect(0, 127, '\P{inascii}', "");
Expect(1, 127, '\P{^inascii}', "");
Expect(0, 128, '\p{inascii}', "");
Expect(1, 128, '\p{^inascii}', "");
Expect(1, 128, '\P{inascii}', "");
Expect(0, 128, '\P{^inascii}', "");
Expect(1, 127, '\p{_ IN_ASCII}', "");
Expect(0, 127, '\p{^_ IN_ASCII}', "");
Expect(0, 127, '\P{_ IN_ASCII}', "");
Expect(1, 127, '\P{^_ IN_ASCII}', "");
Expect(0, 128, '\p{_ IN_ASCII}', "");
Expect(1, 128, '\p{^_ IN_ASCII}', "");
Expect(1, 128, '\P{_ IN_ASCII}', "");
Expect(0, 128, '\P{^_ IN_ASCII}', "");
Error('\p{ :=aegean_Numbers}');
Error('\P{ :=aegean_Numbers}');
Expect(1, 65855, '\p{aegeannumbers}', "");
Expect(0, 65855, '\p{^aegeannumbers}', "");
Expect(0, 65855, '\P{aegeannumbers}', "");
Expect(1, 65855, '\P{^aegeannumbers}', "");
Expect(0, 65856, '\p{aegeannumbers}', "");
Expect(1, 65856, '\p{^aegeannumbers}', "");
Expect(1, 65856, '\P{aegeannumbers}', "");
Expect(0, 65856, '\P{^aegeannumbers}', "");
Expect(1, 65855, '\p{-	Aegean_Numbers}', "");
Expect(0, 65855, '\p{^-	Aegean_Numbers}', "");
Expect(0, 65855, '\P{-	Aegean_Numbers}', "");
Expect(1, 65855, '\P{^-	Aegean_Numbers}', "");
Expect(0, 65856, '\p{-	Aegean_Numbers}', "");
Expect(1, 65856, '\p{^-	Aegean_Numbers}', "");
Expect(1, 65856, '\P{-	Aegean_Numbers}', "");
Expect(0, 65856, '\P{^-	Aegean_Numbers}', "");
Error('\p{:= Is_Aegean_Numbers}');
Error('\P{:= Is_Aegean_Numbers}');
Expect(1, 65855, '\p{isaegeannumbers}', "");
Expect(0, 65855, '\p{^isaegeannumbers}', "");
Expect(0, 65855, '\P{isaegeannumbers}', "");
Expect(1, 65855, '\P{^isaegeannumbers}', "");
Expect(0, 65856, '\p{isaegeannumbers}', "");
Expect(1, 65856, '\p{^isaegeannumbers}', "");
Expect(1, 65856, '\P{isaegeannumbers}', "");
Expect(0, 65856, '\P{^isaegeannumbers}', "");
Expect(1, 65855, '\p{ _is_Aegean_numbers}', "");
Expect(0, 65855, '\p{^ _is_Aegean_numbers}', "");
Expect(0, 65855, '\P{ _is_Aegean_numbers}', "");
Expect(1, 65855, '\P{^ _is_Aegean_numbers}', "");
Expect(0, 65856, '\p{ _is_Aegean_numbers}', "");
Expect(1, 65856, '\p{^ _is_Aegean_numbers}', "");
Expect(1, 65856, '\P{ _is_Aegean_numbers}', "");
Expect(0, 65856, '\P{^ _is_Aegean_numbers}', "");
Error('\p{/a/	_In_aegean_numbers}');
Error('\P{/a/	_In_aegean_numbers}');
Expect(1, 65855, '\p{inaegeannumbers}', "");
Expect(0, 65855, '\p{^inaegeannumbers}', "");
Expect(0, 65855, '\P{inaegeannumbers}', "");
Expect(1, 65855, '\P{^inaegeannumbers}', "");
Expect(0, 65856, '\p{inaegeannumbers}', "");
Expect(1, 65856, '\p{^inaegeannumbers}', "");
Expect(1, 65856, '\P{inaegeannumbers}', "");
Expect(0, 65856, '\P{^inaegeannumbers}', "");
Expect(1, 65855, '\p{ _In_AEGEAN_Numbers}', "");
Expect(0, 65855, '\p{^ _In_AEGEAN_Numbers}', "");
Expect(0, 65855, '\P{ _In_AEGEAN_Numbers}', "");
Expect(1, 65855, '\P{^ _In_AEGEAN_Numbers}', "");
Expect(0, 65856, '\p{ _In_AEGEAN_Numbers}', "");
Expect(1, 65856, '\p{^ _In_AEGEAN_Numbers}', "");
Expect(1, 65856, '\P{ _In_AEGEAN_Numbers}', "");
Expect(0, 65856, '\P{^ _In_AEGEAN_Numbers}', "");
Error('\p{ alchemical_Symbols:=}');
Error('\P{ alchemical_Symbols:=}');
Expect(1, 128895, '\p{alchemicalsymbols}', "");
Expect(0, 128895, '\p{^alchemicalsymbols}', "");
Expect(0, 128895, '\P{alchemicalsymbols}', "");
Expect(1, 128895, '\P{^alchemicalsymbols}', "");
Expect(0, 128896, '\p{alchemicalsymbols}', "");
Expect(1, 128896, '\p{^alchemicalsymbols}', "");
Expect(1, 128896, '\P{alchemicalsymbols}', "");
Expect(0, 128896, '\P{^alchemicalsymbols}', "");
Expect(1, 128895, '\p{ALCHEMICAL_Symbols}', "");
Expect(0, 128895, '\p{^ALCHEMICAL_Symbols}', "");
Expect(0, 128895, '\P{ALCHEMICAL_Symbols}', "");
Expect(1, 128895, '\P{^ALCHEMICAL_Symbols}', "");
Expect(0, 128896, '\p{ALCHEMICAL_Symbols}', "");
Expect(1, 128896, '\p{^ALCHEMICAL_Symbols}', "");
Expect(1, 128896, '\P{ALCHEMICAL_Symbols}', "");
Expect(0, 128896, '\P{^ALCHEMICAL_Symbols}', "");
Error('\p{	:=Is_alchemical_SYMBOLS}');
Error('\P{	:=Is_alchemical_SYMBOLS}');
Expect(1, 128895, '\p{isalchemicalsymbols}', "");
Expect(0, 128895, '\p{^isalchemicalsymbols}', "");
Expect(0, 128895, '\P{isalchemicalsymbols}', "");
Expect(1, 128895, '\P{^isalchemicalsymbols}', "");
Expect(0, 128896, '\p{isalchemicalsymbols}', "");
Expect(1, 128896, '\p{^isalchemicalsymbols}', "");
Expect(1, 128896, '\P{isalchemicalsymbols}', "");
Expect(0, 128896, '\P{^isalchemicalsymbols}', "");
Expect(1, 128895, '\p{_ IS_Alchemical_SYMBOLS}', "");
Expect(0, 128895, '\p{^_ IS_Alchemical_SYMBOLS}', "");
Expect(0, 128895, '\P{_ IS_Alchemical_SYMBOLS}', "");
Expect(1, 128895, '\P{^_ IS_Alchemical_SYMBOLS}', "");
Expect(0, 128896, '\p{_ IS_Alchemical_SYMBOLS}', "");
Expect(1, 128896, '\p{^_ IS_Alchemical_SYMBOLS}', "");
Expect(1, 128896, '\P{_ IS_Alchemical_SYMBOLS}', "");
Expect(0, 128896, '\P{^_ IS_Alchemical_SYMBOLS}', "");
Error('\p{:= In_ALCHEMICAL_Symbols}');
Error('\P{:= In_ALCHEMICAL_Symbols}');
Expect(1, 128895, '\p{inalchemicalsymbols}', "");
Expect(0, 128895, '\p{^inalchemicalsymbols}', "");
Expect(0, 128895, '\P{inalchemicalsymbols}', "");
Expect(1, 128895, '\P{^inalchemicalsymbols}', "");
Expect(0, 128896, '\p{inalchemicalsymbols}', "");
Expect(1, 128896, '\p{^inalchemicalsymbols}', "");
Expect(1, 128896, '\P{inalchemicalsymbols}', "");
Expect(0, 128896, '\P{^inalchemicalsymbols}', "");
Expect(1, 128895, '\p{_-in_alchemical_Symbols}', "");
Expect(0, 128895, '\p{^_-in_alchemical_Symbols}', "");
Expect(0, 128895, '\P{_-in_alchemical_Symbols}', "");
Expect(1, 128895, '\P{^_-in_alchemical_Symbols}', "");
Expect(0, 128896, '\p{_-in_alchemical_Symbols}', "");
Expect(1, 128896, '\p{^_-in_alchemical_Symbols}', "");
Expect(1, 128896, '\P{_-in_alchemical_Symbols}', "");
Expect(0, 128896, '\P{^_-in_alchemical_Symbols}', "");
Error('\p{:=  Alchemical}');
Error('\P{:=  Alchemical}');
Expect(1, 128895, '\p{alchemical}', "");
Expect(0, 128895, '\p{^alchemical}', "");
Expect(0, 128895, '\P{alchemical}', "");
Expect(1, 128895, '\P{^alchemical}', "");
Expect(0, 128896, '\p{alchemical}', "");
Expect(1, 128896, '\p{^alchemical}', "");
Expect(1, 128896, '\P{alchemical}', "");
Expect(0, 128896, '\P{^alchemical}', "");
Expect(1, 128895, '\p{_-Alchemical}', "");
Expect(0, 128895, '\p{^_-Alchemical}', "");
Expect(0, 128895, '\P{_-Alchemical}', "");
Expect(1, 128895, '\P{^_-Alchemical}', "");
Expect(0, 128896, '\p{_-Alchemical}', "");
Expect(1, 128896, '\p{^_-Alchemical}', "");
Expect(1, 128896, '\P{_-Alchemical}', "");
Expect(0, 128896, '\P{^_-Alchemical}', "");
Error('\p{	_Is_Alchemical:=}');
Error('\P{	_Is_Alchemical:=}');
Expect(1, 128895, '\p{isalchemical}', "");
Expect(0, 128895, '\p{^isalchemical}', "");
Expect(0, 128895, '\P{isalchemical}', "");
Expect(1, 128895, '\P{^isalchemical}', "");
Expect(0, 128896, '\p{isalchemical}', "");
Expect(1, 128896, '\p{^isalchemical}', "");
Expect(1, 128896, '\P{isalchemical}', "");
Expect(0, 128896, '\P{^isalchemical}', "");
Expect(1, 128895, '\p{ is_Alchemical}', "");
Expect(0, 128895, '\p{^ is_Alchemical}', "");
Expect(0, 128895, '\P{ is_Alchemical}', "");
Expect(1, 128895, '\P{^ is_Alchemical}', "");
Expect(0, 128896, '\p{ is_Alchemical}', "");
Expect(1, 128896, '\p{^ is_Alchemical}', "");
Expect(1, 128896, '\P{ is_Alchemical}', "");
Expect(0, 128896, '\P{^ is_Alchemical}', "");
Error('\p{_IN_ALCHEMICAL/a/}');
Error('\P{_IN_ALCHEMICAL/a/}');
Expect(1, 128895, '\p{inalchemical}', "");
Expect(0, 128895, '\p{^inalchemical}', "");
Expect(0, 128895, '\P{inalchemical}', "");
Expect(1, 128895, '\P{^inalchemical}', "");
Expect(0, 128896, '\p{inalchemical}', "");
Expect(1, 128896, '\p{^inalchemical}', "");
Expect(1, 128896, '\P{inalchemical}', "");
Expect(0, 128896, '\P{^inalchemical}', "");
Expect(1, 128895, '\p{_In_ALCHEMICAL}', "");
Expect(0, 128895, '\p{^_In_ALCHEMICAL}', "");
Expect(0, 128895, '\P{_In_ALCHEMICAL}', "");
Expect(1, 128895, '\P{^_In_ALCHEMICAL}', "");
Expect(0, 128896, '\p{_In_ALCHEMICAL}', "");
Expect(1, 128896, '\p{^_In_ALCHEMICAL}', "");
Expect(1, 128896, '\P{_In_ALCHEMICAL}', "");
Expect(0, 128896, '\P{^_In_ALCHEMICAL}', "");
Error('\p{	:=All}');
Error('\P{	:=All}');
Expect(1, 1, '\p{all}', "");
Expect(0, 1, '\p{^all}', "");
Expect(0, 1, '\P{all}', "");
Expect(1, 1, '\P{^all}', "");
Expect(1, 1, '\p{_	ALL}', "");
Expect(0, 1, '\p{^_	ALL}', "");
Expect(0, 1, '\P{_	ALL}', "");
Expect(1, 1, '\P{^_	ALL}', "");
Error('\p{- Is_All/a/}');
Error('\P{- Is_All/a/}');
Expect(1, 1, '\p{isall}', "");
Expect(0, 1, '\p{^isall}', "");
Expect(0, 1, '\P{isall}', "");
Expect(1, 1, '\P{^isall}', "");
Expect(1, 1, '\p{	 Is_all}', "");
Expect(0, 1, '\p{^	 Is_all}', "");
Expect(0, 1, '\P{	 Is_all}', "");
Expect(1, 1, '\P{^	 Is_all}', "");
Error('\p{_/a/Alnum}');
Error('\P{_/a/Alnum}');
Expect(1, 195101, '\p{alnum}', "");
Expect(0, 195101, '\p{^alnum}', "");
Expect(0, 195101, '\P{alnum}', "");
Expect(1, 195101, '\P{^alnum}', "");
Expect(0, 195102, '\p{alnum}', "");
Expect(1, 195102, '\p{^alnum}', "");
Expect(1, 195102, '\P{alnum}', "");
Expect(0, 195102, '\P{^alnum}', "");
Expect(1, 195101, '\p{--Alnum}', "");
Expect(0, 195101, '\p{^--Alnum}', "");
Expect(0, 195101, '\P{--Alnum}', "");
Expect(1, 195101, '\P{^--Alnum}', "");
Expect(0, 195102, '\p{--Alnum}', "");
Expect(1, 195102, '\p{^--Alnum}', "");
Expect(1, 195102, '\P{--Alnum}', "");
Expect(0, 195102, '\P{^--Alnum}', "");
Error('\p{/a/-_XPosixAlnum}');
Error('\P{/a/-_XPosixAlnum}');
Expect(1, 195101, '\p{xposixalnum}', "");
Expect(0, 195101, '\p{^xposixalnum}', "");
Expect(0, 195101, '\P{xposixalnum}', "");
Expect(1, 195101, '\P{^xposixalnum}', "");
Expect(0, 195102, '\p{xposixalnum}', "");
Expect(1, 195102, '\p{^xposixalnum}', "");
Expect(1, 195102, '\P{xposixalnum}', "");
Expect(0, 195102, '\P{^xposixalnum}', "");
Expect(1, 195101, '\p{-XPosixAlnum}', "");
Expect(0, 195101, '\p{^-XPosixAlnum}', "");
Expect(0, 195101, '\P{-XPosixAlnum}', "");
Expect(1, 195101, '\P{^-XPosixAlnum}', "");
Expect(0, 195102, '\p{-XPosixAlnum}', "");
Expect(1, 195102, '\p{^-XPosixAlnum}', "");
Expect(1, 195102, '\P{-XPosixAlnum}', "");
Expect(0, 195102, '\P{^-XPosixAlnum}', "");
Error('\p{__is_Alnum/a/}');
Error('\P{__is_Alnum/a/}');
Expect(1, 195101, '\p{isalnum}', "");
Expect(0, 195101, '\p{^isalnum}', "");
Expect(0, 195101, '\P{isalnum}', "");
Expect(1, 195101, '\P{^isalnum}', "");
Expect(0, 195102, '\p{isalnum}', "");
Expect(1, 195102, '\p{^isalnum}', "");
Expect(1, 195102, '\P{isalnum}', "");
Expect(0, 195102, '\P{^isalnum}', "");
Expect(1, 195101, '\p{-	Is_ALNUM}', "");
Expect(0, 195101, '\p{^-	Is_ALNUM}', "");
Expect(0, 195101, '\P{-	Is_ALNUM}', "");
Expect(1, 195101, '\P{^-	Is_ALNUM}', "");
Expect(0, 195102, '\p{-	Is_ALNUM}', "");
Expect(1, 195102, '\p{^-	Is_ALNUM}', "");
Expect(1, 195102, '\P{-	Is_ALNUM}', "");
Expect(0, 195102, '\P{^-	Is_ALNUM}', "");
Error('\p{ Is_XPOSIXALNUM/a/}');
Error('\P{ Is_XPOSIXALNUM/a/}');
Expect(1, 195101, '\p{isxposixalnum}', "");
Expect(0, 195101, '\p{^isxposixalnum}', "");
Expect(0, 195101, '\P{isxposixalnum}', "");
Expect(1, 195101, '\P{^isxposixalnum}', "");
Expect(0, 195102, '\p{isxposixalnum}', "");
Expect(1, 195102, '\p{^isxposixalnum}', "");
Expect(1, 195102, '\P{isxposixalnum}', "");
Expect(0, 195102, '\P{^isxposixalnum}', "");
Expect(1, 195101, '\p{ Is_XPosixAlnum}', "");
Expect(0, 195101, '\p{^ Is_XPosixAlnum}', "");
Expect(0, 195101, '\P{ Is_XPosixAlnum}', "");
Expect(1, 195101, '\P{^ Is_XPosixAlnum}', "");
Expect(0, 195102, '\p{ Is_XPosixAlnum}', "");
Expect(1, 195102, '\p{^ Is_XPosixAlnum}', "");
Expect(1, 195102, '\P{ Is_XPosixAlnum}', "");
Expect(0, 195102, '\P{^ Is_XPosixAlnum}', "");
Error('\p{_	ALPHA:=}');
Error('\P{_	ALPHA:=}');
Expect(1, 195101, '\p{alpha}', "");
Expect(0, 195101, '\p{^alpha}', "");
Expect(0, 195101, '\P{alpha}', "");
Expect(1, 195101, '\P{^alpha}', "");
Expect(0, 195102, '\p{alpha}', "");
Expect(1, 195102, '\p{^alpha}', "");
Expect(1, 195102, '\P{alpha}', "");
Expect(0, 195102, '\P{^alpha}', "");
Expect(1, 195101, '\p{_ ALPHA}', "");
Expect(0, 195101, '\p{^_ ALPHA}', "");
Expect(0, 195101, '\P{_ ALPHA}', "");
Expect(1, 195101, '\P{^_ ALPHA}', "");
Expect(0, 195102, '\p{_ ALPHA}', "");
Expect(1, 195102, '\p{^_ ALPHA}', "");
Expect(1, 195102, '\P{_ ALPHA}', "");
Expect(0, 195102, '\P{^_ ALPHA}', "");
Error('\p{_ XPosixAlpha:=}');
Error('\P{_ XPosixAlpha:=}');
Expect(1, 195101, '\p{xposixalpha}', "");
Expect(0, 195101, '\p{^xposixalpha}', "");
Expect(0, 195101, '\P{xposixalpha}', "");
Expect(1, 195101, '\P{^xposixalpha}', "");
Expect(0, 195102, '\p{xposixalpha}', "");
Expect(1, 195102, '\p{^xposixalpha}', "");
Expect(1, 195102, '\P{xposixalpha}', "");
Expect(0, 195102, '\P{^xposixalpha}', "");
Expect(1, 195101, '\p{ xposixalpha}', "");
Expect(0, 195101, '\p{^ xposixalpha}', "");
Expect(0, 195101, '\P{ xposixalpha}', "");
Expect(1, 195101, '\P{^ xposixalpha}', "");
Expect(0, 195102, '\p{ xposixalpha}', "");
Expect(1, 195102, '\p{^ xposixalpha}', "");
Expect(1, 195102, '\P{ xposixalpha}', "");
Expect(0, 195102, '\P{^ xposixalpha}', "");
Error('\p{:= _is_ALPHA}');
Error('\P{:= _is_ALPHA}');
Expect(1, 195101, '\p{isalpha}', "");
Expect(0, 195101, '\p{^isalpha}', "");
Expect(0, 195101, '\P{isalpha}', "");
Expect(1, 195101, '\P{^isalpha}', "");
Expect(0, 195102, '\p{isalpha}', "");
Expect(1, 195102, '\p{^isalpha}', "");
Expect(1, 195102, '\P{isalpha}', "");
Expect(0, 195102, '\P{^isalpha}', "");
Expect(1, 195101, '\p{-is_alpha}', "");
Expect(0, 195101, '\p{^-is_alpha}', "");
Expect(0, 195101, '\P{-is_alpha}', "");
Expect(1, 195101, '\P{^-is_alpha}', "");
Expect(0, 195102, '\p{-is_alpha}', "");
Expect(1, 195102, '\p{^-is_alpha}', "");
Expect(1, 195102, '\P{-is_alpha}', "");
Expect(0, 195102, '\P{^-is_alpha}', "");
Error('\p{ :=IS_XPosixAlpha}');
Error('\P{ :=IS_XPosixAlpha}');
Expect(1, 195101, '\p{isxposixalpha}', "");
Expect(0, 195101, '\p{^isxposixalpha}', "");
Expect(0, 195101, '\P{isxposixalpha}', "");
Expect(1, 195101, '\P{^isxposixalpha}', "");
Expect(0, 195102, '\p{isxposixalpha}', "");
Expect(1, 195102, '\p{^isxposixalpha}', "");
Expect(1, 195102, '\P{isxposixalpha}', "");
Expect(0, 195102, '\P{^isxposixalpha}', "");
Expect(1, 195101, '\p{_ is_XPOSIXALPHA}', "");
Expect(0, 195101, '\p{^_ is_XPOSIXALPHA}', "");
Expect(0, 195101, '\P{_ is_XPOSIXALPHA}', "");
Expect(1, 195101, '\P{^_ is_XPOSIXALPHA}', "");
Expect(0, 195102, '\p{_ is_XPOSIXALPHA}', "");
Expect(1, 195102, '\p{^_ is_XPOSIXALPHA}', "");
Expect(1, 195102, '\P{_ is_XPOSIXALPHA}', "");
Expect(0, 195102, '\P{^_ is_XPOSIXALPHA}', "");
Error('\p{ /a/Alphabetic}');
Error('\P{ /a/Alphabetic}');
Expect(1, 195101, '\p{alphabetic}', "");
Expect(0, 195101, '\p{^alphabetic}', "");
Expect(0, 195101, '\P{alphabetic}', "");
Expect(1, 195101, '\P{^alphabetic}', "");
Expect(0, 195102, '\p{alphabetic}', "");
Expect(1, 195102, '\p{^alphabetic}', "");
Expect(1, 195102, '\P{alphabetic}', "");
Expect(0, 195102, '\P{^alphabetic}', "");
Expect(1, 195101, '\p{-Alphabetic}', "");
Expect(0, 195101, '\p{^-Alphabetic}', "");
Expect(0, 195101, '\P{-Alphabetic}', "");
Expect(1, 195101, '\P{^-Alphabetic}', "");
Expect(0, 195102, '\p{-Alphabetic}', "");
Expect(1, 195102, '\p{^-Alphabetic}', "");
Expect(1, 195102, '\P{-Alphabetic}', "");
Expect(0, 195102, '\P{^-Alphabetic}', "");
Error('\p{_-Is_alphabetic:=}');
Error('\P{_-Is_alphabetic:=}');
Expect(1, 195101, '\p{isalphabetic}', "");
Expect(0, 195101, '\p{^isalphabetic}', "");
Expect(0, 195101, '\P{isalphabetic}', "");
Expect(1, 195101, '\P{^isalphabetic}', "");
Expect(0, 195102, '\p{isalphabetic}', "");
Expect(1, 195102, '\p{^isalphabetic}', "");
Expect(1, 195102, '\P{isalphabetic}', "");
Expect(0, 195102, '\P{^isalphabetic}', "");
Expect(1, 195101, '\p{_Is_ALPHABETIC}', "");
Expect(0, 195101, '\p{^_Is_ALPHABETIC}', "");
Expect(0, 195101, '\P{_Is_ALPHABETIC}', "");
Expect(1, 195101, '\P{^_Is_ALPHABETIC}', "");
Expect(0, 195102, '\p{_Is_ALPHABETIC}', "");
Expect(1, 195102, '\p{^_Is_ALPHABETIC}', "");
Expect(1, 195102, '\P{_Is_ALPHABETIC}', "");
Expect(0, 195102, '\P{^_Is_ALPHABETIC}', "");
Error('\p{	:=Alphabetic_Presentation_Forms}');
Error('\P{	:=Alphabetic_Presentation_Forms}');
Expect(1, 64335, '\p{alphabeticpresentationforms}', "");
Expect(0, 64335, '\p{^alphabeticpresentationforms}', "");
Expect(0, 64335, '\P{alphabeticpresentationforms}', "");
Expect(1, 64335, '\P{^alphabeticpresentationforms}', "");
Expect(0, 64336, '\p{alphabeticpresentationforms}', "");
Expect(1, 64336, '\p{^alphabeticpresentationforms}', "");
Expect(1, 64336, '\P{alphabeticpresentationforms}', "");
Expect(0, 64336, '\P{^alphabeticpresentationforms}', "");
Expect(1, 64335, '\p{_ Alphabetic_Presentation_FORMS}', "");
Expect(0, 64335, '\p{^_ Alphabetic_Presentation_FORMS}', "");
Expect(0, 64335, '\P{_ Alphabetic_Presentation_FORMS}', "");
Expect(1, 64335, '\P{^_ Alphabetic_Presentation_FORMS}', "");
Expect(0, 64336, '\p{_ Alphabetic_Presentation_FORMS}', "");
Expect(1, 64336, '\p{^_ Alphabetic_Presentation_FORMS}', "");
Expect(1, 64336, '\P{_ Alphabetic_Presentation_FORMS}', "");
Expect(0, 64336, '\P{^_ Alphabetic_Presentation_FORMS}', "");
Error('\p{__Is_alphabetic_PRESENTATION_forms/a/}');
Error('\P{__Is_alphabetic_PRESENTATION_forms/a/}');
Expect(1, 64335, '\p{isalphabeticpresentationforms}', "");
Expect(0, 64335, '\p{^isalphabeticpresentationforms}', "");
Expect(0, 64335, '\P{isalphabeticpresentationforms}', "");
Expect(1, 64335, '\P{^isalphabeticpresentationforms}', "");
Expect(0, 64336, '\p{isalphabeticpresentationforms}', "");
Expect(1, 64336, '\p{^isalphabeticpresentationforms}', "");
Expect(1, 64336, '\P{isalphabeticpresentationforms}', "");
Expect(0, 64336, '\P{^isalphabeticpresentationforms}', "");
Expect(1, 64335, '\p{-_Is_Alphabetic_PRESENTATION_forms}', "");
Expect(0, 64335, '\p{^-_Is_Alphabetic_PRESENTATION_forms}', "");
Expect(0, 64335, '\P{-_Is_Alphabetic_PRESENTATION_forms}', "");
Expect(1, 64335, '\P{^-_Is_Alphabetic_PRESENTATION_forms}', "");
Expect(0, 64336, '\p{-_Is_Alphabetic_PRESENTATION_forms}', "");
Expect(1, 64336, '\p{^-_Is_Alphabetic_PRESENTATION_forms}', "");
Expect(1, 64336, '\P{-_Is_Alphabetic_PRESENTATION_forms}', "");
Expect(0, 64336, '\P{^-_Is_Alphabetic_PRESENTATION_forms}', "");
Error('\p{ In_alphabetic_Presentation_forms:=}');
Error('\P{ In_alphabetic_Presentation_forms:=}');
Expect(1, 64335, '\p{inalphabeticpresentationforms}', "");
Expect(0, 64335, '\p{^inalphabeticpresentationforms}', "");
Expect(0, 64335, '\P{inalphabeticpresentationforms}', "");
Expect(1, 64335, '\P{^inalphabeticpresentationforms}', "");
Expect(0, 64336, '\p{inalphabeticpresentationforms}', "");
Expect(1, 64336, '\p{^inalphabeticpresentationforms}', "");
Expect(1, 64336, '\P{inalphabeticpresentationforms}', "");
Expect(0, 64336, '\P{^inalphabeticpresentationforms}', "");
Expect(1, 64335, '\p{_In_alphabetic_Presentation_Forms}', "");
Expect(0, 64335, '\p{^_In_alphabetic_Presentation_Forms}', "");
Expect(0, 64335, '\P{_In_alphabetic_Presentation_Forms}', "");
Expect(1, 64335, '\P{^_In_alphabetic_Presentation_Forms}', "");
Expect(0, 64336, '\p{_In_alphabetic_Presentation_Forms}', "");
Expect(1, 64336, '\p{^_In_alphabetic_Presentation_Forms}', "");
Expect(1, 64336, '\P{_In_alphabetic_Presentation_Forms}', "");
Expect(0, 64336, '\P{^_In_alphabetic_Presentation_Forms}', "");
Error('\p{Alphabetic_pf/a/}');
Error('\P{Alphabetic_pf/a/}');
Expect(1, 64335, '\p{alphabeticpf}', "");
Expect(0, 64335, '\p{^alphabeticpf}', "");
Expect(0, 64335, '\P{alphabeticpf}', "");
Expect(1, 64335, '\P{^alphabeticpf}', "");
Expect(0, 64336, '\p{alphabeticpf}', "");
Expect(1, 64336, '\p{^alphabeticpf}', "");
Expect(1, 64336, '\P{alphabeticpf}', "");
Expect(0, 64336, '\P{^alphabeticpf}', "");
Expect(1, 64335, '\p{Alphabetic_PF}', "");
Expect(0, 64335, '\p{^Alphabetic_PF}', "");
Expect(0, 64335, '\P{Alphabetic_PF}', "");
Expect(1, 64335, '\P{^Alphabetic_PF}', "");
Expect(0, 64336, '\p{Alphabetic_PF}', "");
Expect(1, 64336, '\p{^Alphabetic_PF}', "");
Expect(1, 64336, '\P{Alphabetic_PF}', "");
Expect(0, 64336, '\P{^Alphabetic_PF}', "");
Error('\p{ _Is_ALPHABETIC_PF/a/}');
Error('\P{ _Is_ALPHABETIC_PF/a/}');
Expect(1, 64335, '\p{isalphabeticpf}', "");
Expect(0, 64335, '\p{^isalphabeticpf}', "");
Expect(0, 64335, '\P{isalphabeticpf}', "");
Expect(1, 64335, '\P{^isalphabeticpf}', "");
Expect(0, 64336, '\p{isalphabeticpf}', "");
Expect(1, 64336, '\p{^isalphabeticpf}', "");
Expect(1, 64336, '\P{isalphabeticpf}', "");
Expect(0, 64336, '\P{^isalphabeticpf}', "");
Expect(1, 64335, '\p{--IS_Alphabetic_pf}', "");
Expect(0, 64335, '\p{^--IS_Alphabetic_pf}', "");
Expect(0, 64335, '\P{--IS_Alphabetic_pf}', "");
Expect(1, 64335, '\P{^--IS_Alphabetic_pf}', "");
Expect(0, 64336, '\p{--IS_Alphabetic_pf}', "");
Expect(1, 64336, '\p{^--IS_Alphabetic_pf}', "");
Expect(1, 64336, '\P{--IS_Alphabetic_pf}', "");
Expect(0, 64336, '\P{^--IS_Alphabetic_pf}', "");
Error('\p{:= -IN_alphabetic_pf}');
Error('\P{:= -IN_alphabetic_pf}');
Expect(1, 64335, '\p{inalphabeticpf}', "");
Expect(0, 64335, '\p{^inalphabeticpf}', "");
Expect(0, 64335, '\P{inalphabeticpf}', "");
Expect(1, 64335, '\P{^inalphabeticpf}', "");
Expect(0, 64336, '\p{inalphabeticpf}', "");
Expect(1, 64336, '\p{^inalphabeticpf}', "");
Expect(1, 64336, '\P{inalphabeticpf}', "");
Expect(0, 64336, '\P{^inalphabeticpf}', "");
Expect(1, 64335, '\p{	-In_Alphabetic_PF}', "");
Expect(0, 64335, '\p{^	-In_Alphabetic_PF}', "");
Expect(0, 64335, '\P{	-In_Alphabetic_PF}', "");
Expect(1, 64335, '\P{^	-In_Alphabetic_PF}', "");
Expect(0, 64336, '\p{	-In_Alphabetic_PF}', "");
Expect(1, 64336, '\p{^	-In_Alphabetic_PF}', "");
Expect(1, 64336, '\P{	-In_Alphabetic_PF}', "");
Expect(0, 64336, '\P{^	-In_Alphabetic_PF}', "");
Error('\p{_:=ancient_greek_MUSICAL_NOTATION}');
Error('\P{_:=ancient_greek_MUSICAL_NOTATION}');
Expect(1, 119375, '\p{ancientgreekmusicalnotation}', "");
Expect(0, 119375, '\p{^ancientgreekmusicalnotation}', "");
Expect(0, 119375, '\P{ancientgreekmusicalnotation}', "");
Expect(1, 119375, '\P{^ancientgreekmusicalnotation}', "");
Expect(0, 119376, '\p{ancientgreekmusicalnotation}', "");
Expect(1, 119376, '\p{^ancientgreekmusicalnotation}', "");
Expect(1, 119376, '\P{ancientgreekmusicalnotation}', "");
Expect(0, 119376, '\P{^ancientgreekmusicalnotation}', "");
Expect(1, 119375, '\p{	-ANCIENT_Greek_musical_Notation}', "");
Expect(0, 119375, '\p{^	-ANCIENT_Greek_musical_Notation}', "");
Expect(0, 119375, '\P{	-ANCIENT_Greek_musical_Notation}', "");
Expect(1, 119375, '\P{^	-ANCIENT_Greek_musical_Notation}', "");
Expect(0, 119376, '\p{	-ANCIENT_Greek_musical_Notation}', "");
Expect(1, 119376, '\p{^	-ANCIENT_Greek_musical_Notation}', "");
Expect(1, 119376, '\P{	-ANCIENT_Greek_musical_Notation}', "");
Expect(0, 119376, '\P{^	-ANCIENT_Greek_musical_Notation}', "");
Error('\p{ -Is_Ancient_GREEK_Musical_NOTATION/a/}');
Error('\P{ -Is_Ancient_GREEK_Musical_NOTATION/a/}');
Expect(1, 119375, '\p{isancientgreekmusicalnotation}', "");
Expect(0, 119375, '\p{^isancientgreekmusicalnotation}', "");
Expect(0, 119375, '\P{isancientgreekmusicalnotation}', "");
Expect(1, 119375, '\P{^isancientgreekmusicalnotation}', "");
Expect(0, 119376, '\p{isancientgreekmusicalnotation}', "");
Expect(1, 119376, '\p{^isancientgreekmusicalnotation}', "");
Expect(1, 119376, '\P{isancientgreekmusicalnotation}', "");
Expect(0, 119376, '\P{^isancientgreekmusicalnotation}', "");
Expect(1, 119375, '\p{_is_ancient_Greek_Musical_Notation}', "");
Expect(0, 119375, '\p{^_is_ancient_Greek_Musical_Notation}', "");
Expect(0, 119375, '\P{_is_ancient_Greek_Musical_Notation}', "");
Expect(1, 119375, '\P{^_is_ancient_Greek_Musical_Notation}', "");
Expect(0, 119376, '\p{_is_ancient_Greek_Musical_Notation}', "");
Expect(1, 119376, '\p{^_is_ancient_Greek_Musical_Notation}', "");
Expect(1, 119376, '\P{_is_ancient_Greek_Musical_Notation}', "");
Expect(0, 119376, '\P{^_is_ancient_Greek_Musical_Notation}', "");
Error('\p{--in_Ancient_greek_Musical_Notation/a/}');
Error('\P{--in_Ancient_greek_Musical_Notation/a/}');
Expect(1, 119375, '\p{inancientgreekmusicalnotation}', "");
Expect(0, 119375, '\p{^inancientgreekmusicalnotation}', "");
Expect(0, 119375, '\P{inancientgreekmusicalnotation}', "");
Expect(1, 119375, '\P{^inancientgreekmusicalnotation}', "");
Expect(0, 119376, '\p{inancientgreekmusicalnotation}', "");
Expect(1, 119376, '\p{^inancientgreekmusicalnotation}', "");
Expect(1, 119376, '\P{inancientgreekmusicalnotation}', "");
Expect(0, 119376, '\P{^inancientgreekmusicalnotation}', "");
Expect(1, 119375, '\p{ 	In_Ancient_GREEK_Musical_notation}', "");
Expect(0, 119375, '\p{^ 	In_Ancient_GREEK_Musical_notation}', "");
Expect(0, 119375, '\P{ 	In_Ancient_GREEK_Musical_notation}', "");
Expect(1, 119375, '\P{^ 	In_Ancient_GREEK_Musical_notation}', "");
Expect(0, 119376, '\p{ 	In_Ancient_GREEK_Musical_notation}', "");
Expect(1, 119376, '\p{^ 	In_Ancient_GREEK_Musical_notation}', "");
Expect(1, 119376, '\P{ 	In_Ancient_GREEK_Musical_notation}', "");
Expect(0, 119376, '\P{^ 	In_Ancient_GREEK_Musical_notation}', "");
Error('\p{:=- ancient_Greek_music}');
Error('\P{:=- ancient_Greek_music}');
Expect(1, 119375, '\p{ancientgreekmusic}', "");
Expect(0, 119375, '\p{^ancientgreekmusic}', "");
Expect(0, 119375, '\P{ancientgreekmusic}', "");
Expect(1, 119375, '\P{^ancientgreekmusic}', "");
Expect(0, 119376, '\p{ancientgreekmusic}', "");
Expect(1, 119376, '\p{^ancientgreekmusic}', "");
Expect(1, 119376, '\P{ancientgreekmusic}', "");
Expect(0, 119376, '\P{^ancientgreekmusic}', "");
Expect(1, 119375, '\p{ -Ancient_Greek_Music}', "");
Expect(0, 119375, '\p{^ -Ancient_Greek_Music}', "");
Expect(0, 119375, '\P{ -Ancient_Greek_Music}', "");
Expect(1, 119375, '\P{^ -Ancient_Greek_Music}', "");
Expect(0, 119376, '\p{ -Ancient_Greek_Music}', "");
Expect(1, 119376, '\p{^ -Ancient_Greek_Music}', "");
Expect(1, 119376, '\P{ -Ancient_Greek_Music}', "");
Expect(0, 119376, '\P{^ -Ancient_Greek_Music}', "");
Error('\p{_ Is_ancient_Greek_music/a/}');
Error('\P{_ Is_ancient_Greek_music/a/}');
Expect(1, 119375, '\p{isancientgreekmusic}', "");
Expect(0, 119375, '\p{^isancientgreekmusic}', "");
Expect(0, 119375, '\P{isancientgreekmusic}', "");
Expect(1, 119375, '\P{^isancientgreekmusic}', "");
Expect(0, 119376, '\p{isancientgreekmusic}', "");
Expect(1, 119376, '\p{^isancientgreekmusic}', "");
Expect(1, 119376, '\P{isancientgreekmusic}', "");
Expect(0, 119376, '\P{^isancientgreekmusic}', "");
Expect(1, 119375, '\p{-	IS_Ancient_Greek_music}', "");
Expect(0, 119375, '\p{^-	IS_Ancient_Greek_music}', "");
Expect(0, 119375, '\P{-	IS_Ancient_Greek_music}', "");
Expect(1, 119375, '\P{^-	IS_Ancient_Greek_music}', "");
Expect(0, 119376, '\p{-	IS_Ancient_Greek_music}', "");
Expect(1, 119376, '\p{^-	IS_Ancient_Greek_music}', "");
Expect(1, 119376, '\P{-	IS_Ancient_Greek_music}', "");
Expect(0, 119376, '\P{^-	IS_Ancient_Greek_music}', "");
Error('\p{/a/	IN_Ancient_greek_Music}');
Error('\P{/a/	IN_Ancient_greek_Music}');
Expect(1, 119375, '\p{inancientgreekmusic}', "");
Expect(0, 119375, '\p{^inancientgreekmusic}', "");
Expect(0, 119375, '\P{inancientgreekmusic}', "");
Expect(1, 119375, '\P{^inancientgreekmusic}', "");
Expect(0, 119376, '\p{inancientgreekmusic}', "");
Expect(1, 119376, '\p{^inancientgreekmusic}', "");
Expect(1, 119376, '\P{inancientgreekmusic}', "");
Expect(0, 119376, '\P{^inancientgreekmusic}', "");
Expect(1, 119375, '\p{_IN_ancient_Greek_music}', "");
Expect(0, 119375, '\p{^_IN_ancient_Greek_music}', "");
Expect(0, 119375, '\P{_IN_ancient_Greek_music}', "");
Expect(1, 119375, '\P{^_IN_ancient_Greek_music}', "");
Expect(0, 119376, '\p{_IN_ancient_Greek_music}', "");
Expect(1, 119376, '\p{^_IN_ancient_Greek_music}', "");
Expect(1, 119376, '\P{_IN_ancient_Greek_music}', "");
Expect(0, 119376, '\P{^_IN_ancient_Greek_music}', "");
Error('\p{ :=Ancient_Greek_numbers}');
Error('\P{ :=Ancient_Greek_numbers}');
Expect(1, 65935, '\p{ancientgreeknumbers}', "");
Expect(0, 65935, '\p{^ancientgreeknumbers}', "");
Expect(0, 65935, '\P{ancientgreeknumbers}', "");
Expect(1, 65935, '\P{^ancientgreeknumbers}', "");
Expect(0, 65936, '\p{ancientgreeknumbers}', "");
Expect(1, 65936, '\p{^ancientgreeknumbers}', "");
Expect(1, 65936, '\P{ancientgreeknumbers}', "");
Expect(0, 65936, '\P{^ancientgreeknumbers}', "");
Expect(1, 65935, '\p{ Ancient_Greek_Numbers}', "");
Expect(0, 65935, '\p{^ Ancient_Greek_Numbers}', "");
Expect(0, 65935, '\P{ Ancient_Greek_Numbers}', "");
Expect(1, 65935, '\P{^ Ancient_Greek_Numbers}', "");
Expect(0, 65936, '\p{ Ancient_Greek_Numbers}', "");
Expect(1, 65936, '\p{^ Ancient_Greek_Numbers}', "");
Expect(1, 65936, '\P{ Ancient_Greek_Numbers}', "");
Expect(0, 65936, '\P{^ Ancient_Greek_Numbers}', "");
Error('\p{/a/		IS_ANCIENT_Greek_NUMBERS}');
Error('\P{/a/		IS_ANCIENT_Greek_NUMBERS}');
Expect(1, 65935, '\p{isancientgreeknumbers}', "");
Expect(0, 65935, '\p{^isancientgreeknumbers}', "");
Expect(0, 65935, '\P{isancientgreeknumbers}', "");
Expect(1, 65935, '\P{^isancientgreeknumbers}', "");
Expect(0, 65936, '\p{isancientgreeknumbers}', "");
Expect(1, 65936, '\p{^isancientgreeknumbers}', "");
Expect(1, 65936, '\P{isancientgreeknumbers}', "");
Expect(0, 65936, '\P{^isancientgreeknumbers}', "");
Expect(1, 65935, '\p{--Is_ANCIENT_GREEK_NUMBERS}', "");
Expect(0, 65935, '\p{^--Is_ANCIENT_GREEK_NUMBERS}', "");
Expect(0, 65935, '\P{--Is_ANCIENT_GREEK_NUMBERS}', "");
Expect(1, 65935, '\P{^--Is_ANCIENT_GREEK_NUMBERS}', "");
Expect(0, 65936, '\p{--Is_ANCIENT_GREEK_NUMBERS}', "");
Expect(1, 65936, '\p{^--Is_ANCIENT_GREEK_NUMBERS}', "");
Expect(1, 65936, '\P{--Is_ANCIENT_GREEK_NUMBERS}', "");
Expect(0, 65936, '\P{^--Is_ANCIENT_GREEK_NUMBERS}', "");
Error('\p{/a/In_ANCIENT_Greek_numbers}');
Error('\P{/a/In_ANCIENT_Greek_numbers}');
Expect(1, 65935, '\p{inancientgreeknumbers}', "");
Expect(0, 65935, '\p{^inancientgreeknumbers}', "");
Expect(0, 65935, '\P{inancientgreeknumbers}', "");
Expect(1, 65935, '\P{^inancientgreeknumbers}', "");
Expect(0, 65936, '\p{inancientgreeknumbers}', "");
Expect(1, 65936, '\p{^inancientgreeknumbers}', "");
Expect(1, 65936, '\P{inancientgreeknumbers}', "");
Expect(0, 65936, '\P{^inancientgreeknumbers}', "");
Expect(1, 65935, '\p{--In_Ancient_greek_numbers}', "");
Expect(0, 65935, '\p{^--In_Ancient_greek_numbers}', "");
Expect(0, 65935, '\P{--In_Ancient_greek_numbers}', "");
Expect(1, 65935, '\P{^--In_Ancient_greek_numbers}', "");
Expect(0, 65936, '\p{--In_Ancient_greek_numbers}', "");
Expect(1, 65936, '\p{^--In_Ancient_greek_numbers}', "");
Expect(1, 65936, '\P{--In_Ancient_greek_numbers}', "");
Expect(0, 65936, '\P{^--In_Ancient_greek_numbers}', "");
Error('\p{Ancient_Symbols/a/}');
Error('\P{Ancient_Symbols/a/}');
Expect(1, 65999, '\p{ancientsymbols}', "");
Expect(0, 65999, '\p{^ancientsymbols}', "");
Expect(0, 65999, '\P{ancientsymbols}', "");
Expect(1, 65999, '\P{^ancientsymbols}', "");
Expect(0, 66000, '\p{ancientsymbols}', "");
Expect(1, 66000, '\p{^ancientsymbols}', "");
Expect(1, 66000, '\P{ancientsymbols}', "");
Expect(0, 66000, '\P{^ancientsymbols}', "");
Expect(1, 65999, '\p{	-Ancient_Symbols}', "");
Expect(0, 65999, '\p{^	-Ancient_Symbols}', "");
Expect(0, 65999, '\P{	-Ancient_Symbols}', "");
Expect(1, 65999, '\P{^	-Ancient_Symbols}', "");
Expect(0, 66000, '\p{	-Ancient_Symbols}', "");
Expect(1, 66000, '\p{^	-Ancient_Symbols}', "");
Expect(1, 66000, '\P{	-Ancient_Symbols}', "");
Expect(0, 66000, '\P{^	-Ancient_Symbols}', "");
Error('\p{_:=is_ANCIENT_Symbols}');
Error('\P{_:=is_ANCIENT_Symbols}');
Expect(1, 65999, '\p{isancientsymbols}', "");
Expect(0, 65999, '\p{^isancientsymbols}', "");
Expect(0, 65999, '\P{isancientsymbols}', "");
Expect(1, 65999, '\P{^isancientsymbols}', "");
Expect(0, 66000, '\p{isancientsymbols}', "");
Expect(1, 66000, '\p{^isancientsymbols}', "");
Expect(1, 66000, '\P{isancientsymbols}', "");
Expect(0, 66000, '\P{^isancientsymbols}', "");
Expect(1, 65999, '\p{ is_ancient_Symbols}', "");
Expect(0, 65999, '\p{^ is_ancient_Symbols}', "");
Expect(0, 65999, '\P{ is_ancient_Symbols}', "");
Expect(1, 65999, '\P{^ is_ancient_Symbols}', "");
Expect(0, 66000, '\p{ is_ancient_Symbols}', "");
Expect(1, 66000, '\p{^ is_ancient_Symbols}', "");
Expect(1, 66000, '\P{ is_ancient_Symbols}', "");
Expect(0, 66000, '\P{^ is_ancient_Symbols}', "");
Error('\p{	/a/In_Ancient_symbols}');
Error('\P{	/a/In_Ancient_symbols}');
Expect(1, 65999, '\p{inancientsymbols}', "");
Expect(0, 65999, '\p{^inancientsymbols}', "");
Expect(0, 65999, '\P{inancientsymbols}', "");
Expect(1, 65999, '\P{^inancientsymbols}', "");
Expect(0, 66000, '\p{inancientsymbols}', "");
Expect(1, 66000, '\p{^inancientsymbols}', "");
Expect(1, 66000, '\P{inancientsymbols}', "");
Expect(0, 66000, '\P{^inancientsymbols}', "");
Expect(1, 65999, '\p{ 	In_Ancient_symbols}', "");
Expect(0, 65999, '\p{^ 	In_Ancient_symbols}', "");
Expect(0, 65999, '\P{ 	In_Ancient_symbols}', "");
Expect(1, 65999, '\P{^ 	In_Ancient_symbols}', "");
Expect(0, 66000, '\p{ 	In_Ancient_symbols}', "");
Expect(1, 66000, '\p{^ 	In_Ancient_symbols}', "");
Expect(1, 66000, '\P{ 	In_Ancient_symbols}', "");
Expect(0, 66000, '\P{^ 	In_Ancient_symbols}', "");
Error('\p{/a/ _Any}');
Error('\P{/a/ _Any}');
Expect(1, 1, '\p{any}', "");
Expect(0, 1, '\p{^any}', "");
Expect(0, 1, '\P{any}', "");
Expect(1, 1, '\P{^any}', "");
Expect(0, 8912887, '\p{any}', "");
Expect(1, 8912887, '\p{^any}', "");
Expect(1, 8912887, '\P{any}', "");
Expect(0, 8912887, '\P{^any}', "");
Expect(1, 1, '\p{_ Any}', "");
Expect(0, 1, '\p{^_ Any}', "");
Expect(0, 1, '\P{_ Any}', "");
Expect(1, 1, '\P{^_ Any}', "");
Expect(0, 8912887, '\p{_ Any}', "");
Expect(1, 8912887, '\p{^_ Any}', "");
Expect(1, 8912887, '\P{_ Any}', "");
Expect(0, 8912887, '\P{^_ Any}', "");
Error('\p{-:=unicode}');
Error('\P{-:=unicode}');
Expect(1, 1, '\p{unicode}', "");
Expect(0, 1, '\p{^unicode}', "");
Expect(0, 1, '\P{unicode}', "");
Expect(1, 1, '\P{^unicode}', "");
Expect(0, 8912887, '\p{unicode}', "");
Expect(1, 8912887, '\p{^unicode}', "");
Expect(1, 8912887, '\P{unicode}', "");
Expect(0, 8912887, '\P{^unicode}', "");
Expect(1, 1, '\p{-	Unicode}', "");
Expect(0, 1, '\p{^-	Unicode}', "");
Expect(0, 1, '\P{-	Unicode}', "");
Expect(1, 1, '\P{^-	Unicode}', "");
Expect(0, 8912887, '\p{-	Unicode}', "");
Expect(1, 8912887, '\p{^-	Unicode}', "");
Expect(1, 8912887, '\P{-	Unicode}', "");
Expect(0, 8912887, '\P{^-	Unicode}', "");
Error('\p{:= is_ANY}');
Error('\P{:= is_ANY}');
Expect(1, 1, '\p{isany}', "");
Expect(0, 1, '\p{^isany}', "");
Expect(0, 1, '\P{isany}', "");
Expect(1, 1, '\P{^isany}', "");
Expect(0, 8912887, '\p{isany}', "");
Expect(1, 8912887, '\p{^isany}', "");
Expect(1, 8912887, '\P{isany}', "");
Expect(0, 8912887, '\P{^isany}', "");
Expect(1, 1, '\p{-_is_any}', "");
Expect(0, 1, '\p{^-_is_any}', "");
Expect(0, 1, '\P{-_is_any}', "");
Expect(1, 1, '\P{^-_is_any}', "");
Expect(0, 8912887, '\p{-_is_any}', "");
Expect(1, 8912887, '\p{^-_is_any}', "");
Expect(1, 8912887, '\P{-_is_any}', "");
Expect(0, 8912887, '\P{^-_is_any}', "");
Error('\p{-:=IS_Unicode}');
Error('\P{-:=IS_Unicode}');
Expect(1, 1, '\p{isunicode}', "");
Expect(0, 1, '\p{^isunicode}', "");
Expect(0, 1, '\P{isunicode}', "");
Expect(1, 1, '\P{^isunicode}', "");
Expect(0, 8912887, '\p{isunicode}', "");
Expect(1, 8912887, '\p{^isunicode}', "");
Expect(1, 8912887, '\P{isunicode}', "");
Expect(0, 8912887, '\P{^isunicode}', "");
Expect(1, 1, '\p{  Is_UNICODE}', "");
Expect(0, 1, '\p{^  Is_UNICODE}', "");
Expect(0, 1, '\P{  Is_UNICODE}', "");
Expect(1, 1, '\P{^  Is_UNICODE}', "");
Expect(0, 8912887, '\p{  Is_UNICODE}', "");
Expect(1, 8912887, '\p{^  Is_UNICODE}', "");
Expect(1, 8912887, '\P{  Is_UNICODE}', "");
Expect(0, 8912887, '\P{^  Is_UNICODE}', "");
Error('\p{:=-	ARABIC}');
Error('\P{:=-	ARABIC}');
Expect(1, 126705, '\p{arabic}', "");
Expect(0, 126705, '\p{^arabic}', "");
Expect(0, 126705, '\P{arabic}', "");
Expect(1, 126705, '\P{^arabic}', "");
Expect(0, 126706, '\p{arabic}', "");
Expect(1, 126706, '\p{^arabic}', "");
Expect(1, 126706, '\P{arabic}', "");
Expect(0, 126706, '\P{^arabic}', "");
Expect(1, 126705, '\p{	 arabic}', "");
Expect(0, 126705, '\p{^	 arabic}', "");
Expect(0, 126705, '\P{	 arabic}', "");
Expect(1, 126705, '\P{^	 arabic}', "");
Expect(0, 126706, '\p{	 arabic}', "");
Expect(1, 126706, '\p{^	 arabic}', "");
Expect(1, 126706, '\P{	 arabic}', "");
Expect(0, 126706, '\P{^	 arabic}', "");
Error('\p{-:=Is_ARABIC}');
Error('\P{-:=Is_ARABIC}');
Expect(1, 126705, '\p{isarabic}', "");
Expect(0, 126705, '\p{^isarabic}', "");
Expect(0, 126705, '\P{isarabic}', "");
Expect(1, 126705, '\P{^isarabic}', "");
Expect(0, 126706, '\p{isarabic}', "");
Expect(1, 126706, '\p{^isarabic}', "");
Expect(1, 126706, '\P{isarabic}', "");
Expect(0, 126706, '\P{^isarabic}', "");
Expect(1, 126705, '\p{_is_Arabic}', "");
Expect(0, 126705, '\p{^_is_Arabic}', "");
Expect(0, 126705, '\P{_is_Arabic}', "");
Expect(1, 126705, '\P{^_is_Arabic}', "");
Expect(0, 126706, '\p{_is_Arabic}', "");
Expect(1, 126706, '\p{^_is_Arabic}', "");
Expect(1, 126706, '\P{_is_Arabic}', "");
Expect(0, 126706, '\P{^_is_Arabic}', "");
Error('\p{:=ARAB}');
Error('\P{:=ARAB}');
Expect(1, 126705, '\p{arab}', "");
Expect(0, 126705, '\p{^arab}', "");
Expect(0, 126705, '\P{arab}', "");
Expect(1, 126705, '\P{^arab}', "");
Expect(0, 126706, '\p{arab}', "");
Expect(1, 126706, '\p{^arab}', "");
Expect(1, 126706, '\P{arab}', "");
Expect(0, 126706, '\P{^arab}', "");
Expect(1, 126705, '\p{-Arab}', "");
Expect(0, 126705, '\p{^-Arab}', "");
Expect(0, 126705, '\P{-Arab}', "");
Expect(1, 126705, '\P{^-Arab}', "");
Expect(0, 126706, '\p{-Arab}', "");
Expect(1, 126706, '\p{^-Arab}', "");
Expect(1, 126706, '\P{-Arab}', "");
Expect(0, 126706, '\P{^-Arab}', "");
Error('\p{:=IS_Arab}');
Error('\P{:=IS_Arab}');
Expect(1, 126705, '\p{isarab}', "");
Expect(0, 126705, '\p{^isarab}', "");
Expect(0, 126705, '\P{isarab}', "");
Expect(1, 126705, '\P{^isarab}', "");
Expect(0, 126706, '\p{isarab}', "");
Expect(1, 126706, '\p{^isarab}', "");
Expect(1, 126706, '\P{isarab}', "");
Expect(0, 126706, '\P{^isarab}', "");
Expect(1, 126705, '\p{ 	IS_Arab}', "");
Expect(0, 126705, '\p{^ 	IS_Arab}', "");
Expect(0, 126705, '\P{ 	IS_Arab}', "");
Expect(1, 126705, '\P{^ 	IS_Arab}', "");
Expect(0, 126706, '\p{ 	IS_Arab}', "");
Expect(1, 126706, '\p{^ 	IS_Arab}', "");
Expect(1, 126706, '\P{ 	IS_Arab}', "");
Expect(0, 126706, '\P{^ 	IS_Arab}', "");
Error('\p{/a/ -Arabic_Extended_A}');
Error('\P{/a/ -Arabic_Extended_A}');
Expect(1, 2303, '\p{arabicextendeda}', "");
Expect(0, 2303, '\p{^arabicextendeda}', "");
Expect(0, 2303, '\P{arabicextendeda}', "");
Expect(1, 2303, '\P{^arabicextendeda}', "");
Expect(0, 2304, '\p{arabicextendeda}', "");
Expect(1, 2304, '\p{^arabicextendeda}', "");
Expect(1, 2304, '\P{arabicextendeda}', "");
Expect(0, 2304, '\P{^arabicextendeda}', "");
Expect(1, 2303, '\p{_ arabic_extended_A}', "");
Expect(0, 2303, '\p{^_ arabic_extended_A}', "");
Expect(0, 2303, '\P{_ arabic_extended_A}', "");
Expect(1, 2303, '\P{^_ arabic_extended_A}', "");
Expect(0, 2304, '\p{_ arabic_extended_A}', "");
Expect(1, 2304, '\p{^_ arabic_extended_A}', "");
Expect(1, 2304, '\P{_ arabic_extended_A}', "");
Expect(0, 2304, '\P{^_ arabic_extended_A}', "");
Error('\p{--Is_Arabic_EXTENDED_a/a/}');
Error('\P{--Is_Arabic_EXTENDED_a/a/}');
Expect(1, 2303, '\p{isarabicextendeda}', "");
Expect(0, 2303, '\p{^isarabicextendeda}', "");
Expect(0, 2303, '\P{isarabicextendeda}', "");
Expect(1, 2303, '\P{^isarabicextendeda}', "");
Expect(0, 2304, '\p{isarabicextendeda}', "");
Expect(1, 2304, '\p{^isarabicextendeda}', "");
Expect(1, 2304, '\P{isarabicextendeda}', "");
Expect(0, 2304, '\P{^isarabicextendeda}', "");
Expect(1, 2303, '\p{		Is_Arabic_extended_A}', "");
Expect(0, 2303, '\p{^		Is_Arabic_extended_A}', "");
Expect(0, 2303, '\P{		Is_Arabic_extended_A}', "");
Expect(1, 2303, '\P{^		Is_Arabic_extended_A}', "");
Expect(0, 2304, '\p{		Is_Arabic_extended_A}', "");
Expect(1, 2304, '\p{^		Is_Arabic_extended_A}', "");
Expect(1, 2304, '\P{		Is_Arabic_extended_A}', "");
Expect(0, 2304, '\P{^		Is_Arabic_extended_A}', "");
Error('\p{:=		In_Arabic_Extended_A}');
Error('\P{:=		In_Arabic_Extended_A}');
Expect(1, 2303, '\p{inarabicextendeda}', "");
Expect(0, 2303, '\p{^inarabicextendeda}', "");
Expect(0, 2303, '\P{inarabicextendeda}', "");
Expect(1, 2303, '\P{^inarabicextendeda}', "");
Expect(0, 2304, '\p{inarabicextendeda}', "");
Expect(1, 2304, '\p{^inarabicextendeda}', "");
Expect(1, 2304, '\P{inarabicextendeda}', "");
Expect(0, 2304, '\P{^inarabicextendeda}', "");
Expect(1, 2303, '\p{	 In_arabic_extended_A}', "");
Expect(0, 2303, '\p{^	 In_arabic_extended_A}', "");
Expect(0, 2303, '\P{	 In_arabic_extended_A}', "");
Expect(1, 2303, '\P{^	 In_arabic_extended_A}', "");
Expect(0, 2304, '\p{	 In_arabic_extended_A}', "");
Expect(1, 2304, '\p{^	 In_arabic_extended_A}', "");
Expect(1, 2304, '\P{	 In_arabic_extended_A}', "");
Expect(0, 2304, '\P{^	 In_arabic_extended_A}', "");
Error('\p{/a/	 Arabic_ext_A}');
Error('\P{/a/	 Arabic_ext_A}');
Expect(1, 2303, '\p{arabicexta}', "");
Expect(0, 2303, '\p{^arabicexta}', "");
Expect(0, 2303, '\P{arabicexta}', "");
Expect(1, 2303, '\P{^arabicexta}', "");
Expect(0, 2304, '\p{arabicexta}', "");
Expect(1, 2304, '\p{^arabicexta}', "");
Expect(1, 2304, '\P{arabicexta}', "");
Expect(0, 2304, '\P{^arabicexta}', "");
Expect(1, 2303, '\p{-	Arabic_EXT_A}', "");
Expect(0, 2303, '\p{^-	Arabic_EXT_A}', "");
Expect(0, 2303, '\P{-	Arabic_EXT_A}', "");
Expect(1, 2303, '\P{^-	Arabic_EXT_A}', "");
Expect(0, 2304, '\p{-	Arabic_EXT_A}', "");
Expect(1, 2304, '\p{^-	Arabic_EXT_A}', "");
Expect(1, 2304, '\P{-	Arabic_EXT_A}', "");
Expect(0, 2304, '\P{^-	Arabic_EXT_A}', "");
Error('\p{/a/ _Is_arabic_ext_A}');
Error('\P{/a/ _Is_arabic_ext_A}');
Expect(1, 2303, '\p{isarabicexta}', "");
Expect(0, 2303, '\p{^isarabicexta}', "");
Expect(0, 2303, '\P{isarabicexta}', "");
Expect(1, 2303, '\P{^isarabicexta}', "");
Expect(0, 2304, '\p{isarabicexta}', "");
Expect(1, 2304, '\p{^isarabicexta}', "");
Expect(1, 2304, '\P{isarabicexta}', "");
Expect(0, 2304, '\P{^isarabicexta}', "");
Expect(1, 2303, '\p{ _IS_Arabic_EXT_a}', "");
Expect(0, 2303, '\p{^ _IS_Arabic_EXT_a}', "");
Expect(0, 2303, '\P{ _IS_Arabic_EXT_a}', "");
Expect(1, 2303, '\P{^ _IS_Arabic_EXT_a}', "");
Expect(0, 2304, '\p{ _IS_Arabic_EXT_a}', "");
Expect(1, 2304, '\p{^ _IS_Arabic_EXT_a}', "");
Expect(1, 2304, '\P{ _IS_Arabic_EXT_a}', "");
Expect(0, 2304, '\P{^ _IS_Arabic_EXT_a}', "");
Error('\p{		in_arabic_EXT_A/a/}');
Error('\P{		in_arabic_EXT_A/a/}');
Expect(1, 2303, '\p{inarabicexta}', "");
Expect(0, 2303, '\p{^inarabicexta}', "");
Expect(0, 2303, '\P{inarabicexta}', "");
Expect(1, 2303, '\P{^inarabicexta}', "");
Expect(0, 2304, '\p{inarabicexta}', "");
Expect(1, 2304, '\p{^inarabicexta}', "");
Expect(1, 2304, '\P{inarabicexta}', "");
Expect(0, 2304, '\P{^inarabicexta}', "");
Expect(1, 2303, '\p{_	in_Arabic_Ext_A}', "");
Expect(0, 2303, '\p{^_	in_Arabic_Ext_A}', "");
Expect(0, 2303, '\P{_	in_Arabic_Ext_A}', "");
Expect(1, 2303, '\P{^_	in_Arabic_Ext_A}', "");
Expect(0, 2304, '\p{_	in_Arabic_Ext_A}', "");
Expect(1, 2304, '\p{^_	in_Arabic_Ext_A}', "");
Expect(1, 2304, '\P{_	in_Arabic_Ext_A}', "");
Expect(0, 2304, '\P{^_	in_Arabic_Ext_A}', "");
Error('\p{-	arabic_MATHEMATICAL_ALPHABETIC_Symbols/a/}');
Error('\P{-	arabic_MATHEMATICAL_ALPHABETIC_Symbols/a/}');
Expect(1, 126719, '\p{arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\p{^arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\P{arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\P{^arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\p{arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\p{^arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\P{arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\P{^arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\p{_	Arabic_MATHEMATICAL_Alphabetic_symbols}', "");
Expect(0, 126719, '\p{^_	Arabic_MATHEMATICAL_Alphabetic_symbols}', "");
Expect(0, 126719, '\P{_	Arabic_MATHEMATICAL_Alphabetic_symbols}', "");
Expect(1, 126719, '\P{^_	Arabic_MATHEMATICAL_Alphabetic_symbols}', "");
Expect(0, 126720, '\p{_	Arabic_MATHEMATICAL_Alphabetic_symbols}', "");
Expect(1, 126720, '\p{^_	Arabic_MATHEMATICAL_Alphabetic_symbols}', "");
Expect(1, 126720, '\P{_	Arabic_MATHEMATICAL_Alphabetic_symbols}', "");
Expect(0, 126720, '\P{^_	Arabic_MATHEMATICAL_Alphabetic_symbols}', "");
Error('\p{ 	IS_Arabic_mathematical_ALPHABETIC_SYMBOLS:=}');
Error('\P{ 	IS_Arabic_mathematical_ALPHABETIC_SYMBOLS:=}');
Expect(1, 126719, '\p{isarabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\p{^isarabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\P{isarabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\P{^isarabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\p{isarabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\p{^isarabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\P{isarabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\P{^isarabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\p{ -Is_ARABIC_MATHEMATICAL_Alphabetic_SYMBOLS}', "");
Expect(0, 126719, '\p{^ -Is_ARABIC_MATHEMATICAL_Alphabetic_SYMBOLS}', "");
Expect(0, 126719, '\P{ -Is_ARABIC_MATHEMATICAL_Alphabetic_SYMBOLS}', "");
Expect(1, 126719, '\P{^ -Is_ARABIC_MATHEMATICAL_Alphabetic_SYMBOLS}', "");
Expect(0, 126720, '\p{ -Is_ARABIC_MATHEMATICAL_Alphabetic_SYMBOLS}', "");
Expect(1, 126720, '\p{^ -Is_ARABIC_MATHEMATICAL_Alphabetic_SYMBOLS}', "");
Expect(1, 126720, '\P{ -Is_ARABIC_MATHEMATICAL_Alphabetic_SYMBOLS}', "");
Expect(0, 126720, '\P{^ -Is_ARABIC_MATHEMATICAL_Alphabetic_SYMBOLS}', "");
Error('\p{-:=In_Arabic_MATHEMATICAL_Alphabetic_symbols}');
Error('\P{-:=In_Arabic_MATHEMATICAL_Alphabetic_symbols}');
Expect(1, 126719, '\p{inarabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\p{^inarabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\P{inarabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\P{^inarabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\p{inarabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\p{^inarabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\P{inarabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\P{^inarabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\p{	_In_Arabic_MATHEMATICAL_alphabetic_Symbols}', "");
Expect(0, 126719, '\p{^	_In_Arabic_MATHEMATICAL_alphabetic_Symbols}', "");
Expect(0, 126719, '\P{	_In_Arabic_MATHEMATICAL_alphabetic_Symbols}', "");
Expect(1, 126719, '\P{^	_In_Arabic_MATHEMATICAL_alphabetic_Symbols}', "");
Expect(0, 126720, '\p{	_In_Arabic_MATHEMATICAL_alphabetic_Symbols}', "");
Expect(1, 126720, '\p{^	_In_Arabic_MATHEMATICAL_alphabetic_Symbols}', "");
Expect(1, 126720, '\P{	_In_Arabic_MATHEMATICAL_alphabetic_Symbols}', "");
Expect(0, 126720, '\P{^	_In_Arabic_MATHEMATICAL_alphabetic_Symbols}', "");
Error('\p{:= 	Arabic_Math}');
Error('\P{:= 	Arabic_Math}');
Expect(1, 126719, '\p{arabicmath}', "");
Expect(0, 126719, '\p{^arabicmath}', "");
Expect(0, 126719, '\P{arabicmath}', "");
Expect(1, 126719, '\P{^arabicmath}', "");
Expect(0, 126720, '\p{arabicmath}', "");
Expect(1, 126720, '\p{^arabicmath}', "");
Expect(1, 126720, '\P{arabicmath}', "");
Expect(0, 126720, '\P{^arabicmath}', "");
Expect(1, 126719, '\p{Arabic_Math}', "");
Expect(0, 126719, '\p{^Arabic_Math}', "");
Expect(0, 126719, '\P{Arabic_Math}', "");
Expect(1, 126719, '\P{^Arabic_Math}', "");
Expect(0, 126720, '\p{Arabic_Math}', "");
Expect(1, 126720, '\p{^Arabic_Math}', "");
Expect(1, 126720, '\P{Arabic_Math}', "");
Expect(0, 126720, '\P{^Arabic_Math}', "");
Error('\p{-/a/Is_arabic_MATH}');
Error('\P{-/a/Is_arabic_MATH}');
Expect(1, 126719, '\p{isarabicmath}', "");
Expect(0, 126719, '\p{^isarabicmath}', "");
Expect(0, 126719, '\P{isarabicmath}', "");
Expect(1, 126719, '\P{^isarabicmath}', "");
Expect(0, 126720, '\p{isarabicmath}', "");
Expect(1, 126720, '\p{^isarabicmath}', "");
Expect(1, 126720, '\P{isarabicmath}', "");
Expect(0, 126720, '\P{^isarabicmath}', "");
Expect(1, 126719, '\p{_-IS_Arabic_math}', "");
Expect(0, 126719, '\p{^_-IS_Arabic_math}', "");
Expect(0, 126719, '\P{_-IS_Arabic_math}', "");
Expect(1, 126719, '\P{^_-IS_Arabic_math}', "");
Expect(0, 126720, '\p{_-IS_Arabic_math}', "");
Expect(1, 126720, '\p{^_-IS_Arabic_math}', "");
Expect(1, 126720, '\P{_-IS_Arabic_math}', "");
Expect(0, 126720, '\P{^_-IS_Arabic_math}', "");
Error('\p{:=In_ARABIC_Math}');
Error('\P{:=In_ARABIC_Math}');
Expect(1, 126719, '\p{inarabicmath}', "");
Expect(0, 126719, '\p{^inarabicmath}', "");
Expect(0, 126719, '\P{inarabicmath}', "");
Expect(1, 126719, '\P{^inarabicmath}', "");
Expect(0, 126720, '\p{inarabicmath}', "");
Expect(1, 126720, '\p{^inarabicmath}', "");
Expect(1, 126720, '\P{inarabicmath}', "");
Expect(0, 126720, '\P{^inarabicmath}', "");
Expect(1, 126719, '\p{_In_Arabic_Math}', "");
Expect(0, 126719, '\p{^_In_Arabic_Math}', "");
Expect(0, 126719, '\P{_In_Arabic_Math}', "");
Expect(1, 126719, '\P{^_In_Arabic_Math}', "");
Expect(0, 126720, '\p{_In_Arabic_Math}', "");
Expect(1, 126720, '\p{^_In_Arabic_Math}', "");
Expect(1, 126720, '\P{_In_Arabic_Math}', "");
Expect(0, 126720, '\P{^_In_Arabic_Math}', "");
Error('\p{_:=ARABIC_presentation_forms_A}');
Error('\P{_:=ARABIC_presentation_forms_A}');
Expect(1, 65023, '\p{arabicpresentationformsa}', "");
Expect(0, 65023, '\p{^arabicpresentationformsa}', "");
Expect(0, 65023, '\P{arabicpresentationformsa}', "");
Expect(1, 65023, '\P{^arabicpresentationformsa}', "");
Expect(0, 65024, '\p{arabicpresentationformsa}', "");
Expect(1, 65024, '\p{^arabicpresentationformsa}', "");
Expect(1, 65024, '\P{arabicpresentationformsa}', "");
Expect(0, 65024, '\P{^arabicpresentationformsa}', "");
Expect(1, 65023, '\p{ _Arabic_presentation_Forms_A}', "");
Expect(0, 65023, '\p{^ _Arabic_presentation_Forms_A}', "");
Expect(0, 65023, '\P{ _Arabic_presentation_Forms_A}', "");
Expect(1, 65023, '\P{^ _Arabic_presentation_Forms_A}', "");
Expect(0, 65024, '\p{ _Arabic_presentation_Forms_A}', "");
Expect(1, 65024, '\p{^ _Arabic_presentation_Forms_A}', "");
Expect(1, 65024, '\P{ _Arabic_presentation_Forms_A}', "");
Expect(0, 65024, '\P{^ _Arabic_presentation_Forms_A}', "");
Error('\p{/a/Is_Arabic_PRESENTATION_Forms_A}');
Error('\P{/a/Is_Arabic_PRESENTATION_Forms_A}');
Expect(1, 65023, '\p{isarabicpresentationformsa}', "");
Expect(0, 65023, '\p{^isarabicpresentationformsa}', "");
Expect(0, 65023, '\P{isarabicpresentationformsa}', "");
Expect(1, 65023, '\P{^isarabicpresentationformsa}', "");
Expect(0, 65024, '\p{isarabicpresentationformsa}', "");
Expect(1, 65024, '\p{^isarabicpresentationformsa}', "");
Expect(1, 65024, '\P{isarabicpresentationformsa}', "");
Expect(0, 65024, '\P{^isarabicpresentationformsa}', "");
Expect(1, 65023, '\p{__Is_Arabic_Presentation_Forms_A}', "");
Expect(0, 65023, '\p{^__Is_Arabic_Presentation_Forms_A}', "");
Expect(0, 65023, '\P{__Is_Arabic_Presentation_Forms_A}', "");
Expect(1, 65023, '\P{^__Is_Arabic_Presentation_Forms_A}', "");
Expect(0, 65024, '\p{__Is_Arabic_Presentation_Forms_A}', "");
Expect(1, 65024, '\p{^__Is_Arabic_Presentation_Forms_A}', "");
Expect(1, 65024, '\P{__Is_Arabic_Presentation_Forms_A}', "");
Expect(0, 65024, '\P{^__Is_Arabic_Presentation_Forms_A}', "");
Error('\p{	_in_Arabic_presentation_Forms_A/a/}');
Error('\P{	_in_Arabic_presentation_Forms_A/a/}');
Expect(1, 65023, '\p{inarabicpresentationformsa}', "");
Expect(0, 65023, '\p{^inarabicpresentationformsa}', "");
Expect(0, 65023, '\P{inarabicpresentationformsa}', "");
Expect(1, 65023, '\P{^inarabicpresentationformsa}', "");
Expect(0, 65024, '\p{inarabicpresentationformsa}', "");
Expect(1, 65024, '\p{^inarabicpresentationformsa}', "");
Expect(1, 65024, '\P{inarabicpresentationformsa}', "");
Expect(0, 65024, '\P{^inarabicpresentationformsa}', "");
Expect(1, 65023, '\p{ IN_Arabic_Presentation_Forms_A}', "");
Expect(0, 65023, '\p{^ IN_Arabic_Presentation_Forms_A}', "");
Expect(0, 65023, '\P{ IN_Arabic_Presentation_Forms_A}', "");
Expect(1, 65023, '\P{^ IN_Arabic_Presentation_Forms_A}', "");
Expect(0, 65024, '\p{ IN_Arabic_Presentation_Forms_A}', "");
Expect(1, 65024, '\p{^ IN_Arabic_Presentation_Forms_A}', "");
Expect(1, 65024, '\P{ IN_Arabic_Presentation_Forms_A}', "");
Expect(0, 65024, '\P{^ IN_Arabic_Presentation_Forms_A}', "");
Error('\p{/a/_	Arabic_PF_A}');
Error('\P{/a/_	Arabic_PF_A}');
Expect(1, 65023, '\p{arabicpfa}', "");
Expect(0, 65023, '\p{^arabicpfa}', "");
Expect(0, 65023, '\P{arabicpfa}', "");
Expect(1, 65023, '\P{^arabicpfa}', "");
Expect(0, 65024, '\p{arabicpfa}', "");
Expect(1, 65024, '\p{^arabicpfa}', "");
Expect(1, 65024, '\P{arabicpfa}', "");
Expect(0, 65024, '\P{^arabicpfa}', "");
Expect(1, 65023, '\p{__Arabic_PF_A}', "");
Expect(0, 65023, '\p{^__Arabic_PF_A}', "");
Expect(0, 65023, '\P{__Arabic_PF_A}', "");
Expect(1, 65023, '\P{^__Arabic_PF_A}', "");
Expect(0, 65024, '\p{__Arabic_PF_A}', "");
Expect(1, 65024, '\p{^__Arabic_PF_A}', "");
Expect(1, 65024, '\P{__Arabic_PF_A}', "");
Expect(0, 65024, '\P{^__Arabic_PF_A}', "");
Error('\p{:=- Is_Arabic_PF_A}');
Error('\P{:=- Is_Arabic_PF_A}');
Expect(1, 65023, '\p{isarabicpfa}', "");
Expect(0, 65023, '\p{^isarabicpfa}', "");
Expect(0, 65023, '\P{isarabicpfa}', "");
Expect(1, 65023, '\P{^isarabicpfa}', "");
Expect(0, 65024, '\p{isarabicpfa}', "");
Expect(1, 65024, '\p{^isarabicpfa}', "");
Expect(1, 65024, '\P{isarabicpfa}', "");
Expect(0, 65024, '\P{^isarabicpfa}', "");
Expect(1, 65023, '\p{ is_arabic_pf_a}', "");
Expect(0, 65023, '\p{^ is_arabic_pf_a}', "");
Expect(0, 65023, '\P{ is_arabic_pf_a}', "");
Expect(1, 65023, '\P{^ is_arabic_pf_a}', "");
Expect(0, 65024, '\p{ is_arabic_pf_a}', "");
Expect(1, 65024, '\p{^ is_arabic_pf_a}', "");
Expect(1, 65024, '\P{ is_arabic_pf_a}', "");
Expect(0, 65024, '\P{^ is_arabic_pf_a}', "");
Error('\p{/a/	 in_arabic_pf_a}');
Error('\P{/a/	 in_arabic_pf_a}');
Expect(1, 65023, '\p{inarabicpfa}', "");
Expect(0, 65023, '\p{^inarabicpfa}', "");
Expect(0, 65023, '\P{inarabicpfa}', "");
Expect(1, 65023, '\P{^inarabicpfa}', "");
Expect(0, 65024, '\p{inarabicpfa}', "");
Expect(1, 65024, '\p{^inarabicpfa}', "");
Expect(1, 65024, '\P{inarabicpfa}', "");
Expect(0, 65024, '\P{^inarabicpfa}', "");
Expect(1, 65023, '\p{__in_Arabic_PF_a}', "");
Expect(0, 65023, '\p{^__in_Arabic_PF_a}', "");
Expect(0, 65023, '\P{__in_Arabic_PF_a}', "");
Expect(1, 65023, '\P{^__in_Arabic_PF_a}', "");
Expect(0, 65024, '\p{__in_Arabic_PF_a}', "");
Expect(1, 65024, '\p{^__in_Arabic_PF_a}', "");
Expect(1, 65024, '\P{__in_Arabic_PF_a}', "");
Expect(0, 65024, '\P{^__in_Arabic_PF_a}', "");
Error('\p{/a/ 	Arabic_Presentation_Forms_B}');
Error('\P{/a/ 	Arabic_Presentation_Forms_B}');
Expect(1, 65279, '\p{arabicpresentationformsb}', "");
Expect(0, 65279, '\p{^arabicpresentationformsb}', "");
Expect(0, 65279, '\P{arabicpresentationformsb}', "");
Expect(1, 65279, '\P{^arabicpresentationformsb}', "");
Expect(0, 65280, '\p{arabicpresentationformsb}', "");
Expect(1, 65280, '\p{^arabicpresentationformsb}', "");
Expect(1, 65280, '\P{arabicpresentationformsb}', "");
Expect(0, 65280, '\P{^arabicpresentationformsb}', "");
Expect(1, 65279, '\p{-arabic_PRESENTATION_Forms_b}', "");
Expect(0, 65279, '\p{^-arabic_PRESENTATION_Forms_b}', "");
Expect(0, 65279, '\P{-arabic_PRESENTATION_Forms_b}', "");
Expect(1, 65279, '\P{^-arabic_PRESENTATION_Forms_b}', "");
Expect(0, 65280, '\p{-arabic_PRESENTATION_Forms_b}', "");
Expect(1, 65280, '\p{^-arabic_PRESENTATION_Forms_b}', "");
Expect(1, 65280, '\P{-arabic_PRESENTATION_Forms_b}', "");
Expect(0, 65280, '\P{^-arabic_PRESENTATION_Forms_b}', "");
Error('\p{_Is_Arabic_Presentation_Forms_b:=}');
Error('\P{_Is_Arabic_Presentation_Forms_b:=}');
Expect(1, 65279, '\p{isarabicpresentationformsb}', "");
Expect(0, 65279, '\p{^isarabicpresentationformsb}', "");
Expect(0, 65279, '\P{isarabicpresentationformsb}', "");
Expect(1, 65279, '\P{^isarabicpresentationformsb}', "");
Expect(0, 65280, '\p{isarabicpresentationformsb}', "");
Expect(1, 65280, '\p{^isarabicpresentationformsb}', "");
Expect(1, 65280, '\P{isarabicpresentationformsb}', "");
Expect(0, 65280, '\P{^isarabicpresentationformsb}', "");
Expect(1, 65279, '\p{-is_Arabic_Presentation_Forms_B}', "");
Expect(0, 65279, '\p{^-is_Arabic_Presentation_Forms_B}', "");
Expect(0, 65279, '\P{-is_Arabic_Presentation_Forms_B}', "");
Expect(1, 65279, '\P{^-is_Arabic_Presentation_Forms_B}', "");
Expect(0, 65280, '\p{-is_Arabic_Presentation_Forms_B}', "");
Expect(1, 65280, '\p{^-is_Arabic_Presentation_Forms_B}', "");
Expect(1, 65280, '\P{-is_Arabic_Presentation_Forms_B}', "");
Expect(0, 65280, '\P{^-is_Arabic_Presentation_Forms_B}', "");
Error('\p{ _IN_Arabic_PRESENTATION_Forms_b/a/}');
Error('\P{ _IN_Arabic_PRESENTATION_Forms_b/a/}');
Expect(1, 65279, '\p{inarabicpresentationformsb}', "");
Expect(0, 65279, '\p{^inarabicpresentationformsb}', "");
Expect(0, 65279, '\P{inarabicpresentationformsb}', "");
Expect(1, 65279, '\P{^inarabicpresentationformsb}', "");
Expect(0, 65280, '\p{inarabicpresentationformsb}', "");
Expect(1, 65280, '\p{^inarabicpresentationformsb}', "");
Expect(1, 65280, '\P{inarabicpresentationformsb}', "");
Expect(0, 65280, '\P{^inarabicpresentationformsb}', "");
Expect(1, 65279, '\p{	_in_arabic_presentation_FORMS_B}', "");
Expect(0, 65279, '\p{^	_in_arabic_presentation_FORMS_B}', "");
Expect(0, 65279, '\P{	_in_arabic_presentation_FORMS_B}', "");
Expect(1, 65279, '\P{^	_in_arabic_presentation_FORMS_B}', "");
Expect(0, 65280, '\p{	_in_arabic_presentation_FORMS_B}', "");
Expect(1, 65280, '\p{^	_in_arabic_presentation_FORMS_B}', "");
Expect(1, 65280, '\P{	_in_arabic_presentation_FORMS_B}', "");
Expect(0, 65280, '\P{^	_in_arabic_presentation_FORMS_B}', "");
Error('\p{__Arabic_PF_B/a/}');
Error('\P{__Arabic_PF_B/a/}');
Expect(1, 65279, '\p{arabicpfb}', "");
Expect(0, 65279, '\p{^arabicpfb}', "");
Expect(0, 65279, '\P{arabicpfb}', "");
Expect(1, 65279, '\P{^arabicpfb}', "");
Expect(0, 65280, '\p{arabicpfb}', "");
Expect(1, 65280, '\p{^arabicpfb}', "");
Expect(1, 65280, '\P{arabicpfb}', "");
Expect(0, 65280, '\P{^arabicpfb}', "");
Expect(1, 65279, '\p{-_Arabic_pf_B}', "");
Expect(0, 65279, '\p{^-_Arabic_pf_B}', "");
Expect(0, 65279, '\P{-_Arabic_pf_B}', "");
Expect(1, 65279, '\P{^-_Arabic_pf_B}', "");
Expect(0, 65280, '\p{-_Arabic_pf_B}', "");
Expect(1, 65280, '\p{^-_Arabic_pf_B}', "");
Expect(1, 65280, '\P{-_Arabic_pf_B}', "");
Expect(0, 65280, '\P{^-_Arabic_pf_B}', "");
Error('\p{_/a/Is_Arabic_PF_B}');
Error('\P{_/a/Is_Arabic_PF_B}');
Expect(1, 65279, '\p{isarabicpfb}', "");
Expect(0, 65279, '\p{^isarabicpfb}', "");
Expect(0, 65279, '\P{isarabicpfb}', "");
Expect(1, 65279, '\P{^isarabicpfb}', "");
Expect(0, 65280, '\p{isarabicpfb}', "");
Expect(1, 65280, '\p{^isarabicpfb}', "");
Expect(1, 65280, '\P{isarabicpfb}', "");
Expect(0, 65280, '\P{^isarabicpfb}', "");
Expect(1, 65279, '\p{	Is_ARABIC_pf_b}', "");
Expect(0, 65279, '\p{^	Is_ARABIC_pf_b}', "");
Expect(0, 65279, '\P{	Is_ARABIC_pf_b}', "");
Expect(1, 65279, '\P{^	Is_ARABIC_pf_b}', "");
Expect(0, 65280, '\p{	Is_ARABIC_pf_b}', "");
Expect(1, 65280, '\p{^	Is_ARABIC_pf_b}', "");
Expect(1, 65280, '\P{	Is_ARABIC_pf_b}', "");
Expect(0, 65280, '\P{^	Is_ARABIC_pf_b}', "");
Error('\p{-/a/In_Arabic_PF_B}');
Error('\P{-/a/In_Arabic_PF_B}');
Expect(1, 65279, '\p{inarabicpfb}', "");
Expect(0, 65279, '\p{^inarabicpfb}', "");
Expect(0, 65279, '\P{inarabicpfb}', "");
Expect(1, 65279, '\P{^inarabicpfb}', "");
Expect(0, 65280, '\p{inarabicpfb}', "");
Expect(1, 65280, '\p{^inarabicpfb}', "");
Expect(1, 65280, '\P{inarabicpfb}', "");
Expect(0, 65280, '\P{^inarabicpfb}', "");
Expect(1, 65279, '\p{__In_arabic_PF_B}', "");
Expect(0, 65279, '\p{^__In_arabic_PF_B}', "");
Expect(0, 65279, '\P{__In_arabic_PF_B}', "");
Expect(1, 65279, '\P{^__In_arabic_PF_B}', "");
Expect(0, 65280, '\p{__In_arabic_PF_B}', "");
Expect(1, 65280, '\p{^__In_arabic_PF_B}', "");
Expect(1, 65280, '\P{__In_arabic_PF_B}', "");
Expect(0, 65280, '\P{^__In_arabic_PF_B}', "");
Error('\p{	/a/Arabic_SUPPLEMENT}');
Error('\P{	/a/Arabic_SUPPLEMENT}');
Expect(1, 1919, '\p{arabicsupplement}', "");
Expect(0, 1919, '\p{^arabicsupplement}', "");
Expect(0, 1919, '\P{arabicsupplement}', "");
Expect(1, 1919, '\P{^arabicsupplement}', "");
Expect(0, 1920, '\p{arabicsupplement}', "");
Expect(1, 1920, '\p{^arabicsupplement}', "");
Expect(1, 1920, '\P{arabicsupplement}', "");
Expect(0, 1920, '\P{^arabicsupplement}', "");
Expect(1, 1919, '\p{_ arabic_Supplement}', "");
Expect(0, 1919, '\p{^_ arabic_Supplement}', "");
Expect(0, 1919, '\P{_ arabic_Supplement}', "");
Expect(1, 1919, '\P{^_ arabic_Supplement}', "");
Expect(0, 1920, '\p{_ arabic_Supplement}', "");
Expect(1, 1920, '\p{^_ arabic_Supplement}', "");
Expect(1, 1920, '\P{_ arabic_Supplement}', "");
Expect(0, 1920, '\P{^_ arabic_Supplement}', "");
Error('\p{:=Is_ARABIC_Supplement}');
Error('\P{:=Is_ARABIC_Supplement}');
Expect(1, 1919, '\p{isarabicsupplement}', "");
Expect(0, 1919, '\p{^isarabicsupplement}', "");
Expect(0, 1919, '\P{isarabicsupplement}', "");
Expect(1, 1919, '\P{^isarabicsupplement}', "");
Expect(0, 1920, '\p{isarabicsupplement}', "");
Expect(1, 1920, '\p{^isarabicsupplement}', "");
Expect(1, 1920, '\P{isarabicsupplement}', "");
Expect(0, 1920, '\P{^isarabicsupplement}', "");
Expect(1, 1919, '\p{-is_Arabic_Supplement}', "");
Expect(0, 1919, '\p{^-is_Arabic_Supplement}', "");
Expect(0, 1919, '\P{-is_Arabic_Supplement}', "");
Expect(1, 1919, '\P{^-is_Arabic_Supplement}', "");
Expect(0, 1920, '\p{-is_Arabic_Supplement}', "");
Expect(1, 1920, '\p{^-is_Arabic_Supplement}', "");
Expect(1, 1920, '\P{-is_Arabic_Supplement}', "");
Expect(0, 1920, '\P{^-is_Arabic_Supplement}', "");
Error('\p{-/a/in_Arabic_Supplement}');
Error('\P{-/a/in_Arabic_Supplement}');
Expect(1, 1919, '\p{inarabicsupplement}', "");
Expect(0, 1919, '\p{^inarabicsupplement}', "");
Expect(0, 1919, '\P{inarabicsupplement}', "");
Expect(1, 1919, '\P{^inarabicsupplement}', "");
Expect(0, 1920, '\p{inarabicsupplement}', "");
Expect(1, 1920, '\p{^inarabicsupplement}', "");
Expect(1, 1920, '\P{inarabicsupplement}', "");
Expect(0, 1920, '\P{^inarabicsupplement}', "");
Expect(1, 1919, '\p{ 	In_arabic_Supplement}', "");
Expect(0, 1919, '\p{^ 	In_arabic_Supplement}', "");
Expect(0, 1919, '\P{ 	In_arabic_Supplement}', "");
Expect(1, 1919, '\P{^ 	In_arabic_Supplement}', "");
Expect(0, 1920, '\p{ 	In_arabic_Supplement}', "");
Expect(1, 1920, '\p{^ 	In_arabic_Supplement}', "");
Expect(1, 1920, '\P{ 	In_arabic_Supplement}', "");
Expect(0, 1920, '\P{^ 	In_arabic_Supplement}', "");
Error('\p{		Arabic_Sup/a/}');
Error('\P{		Arabic_Sup/a/}');
Expect(1, 1919, '\p{arabicsup}', "");
Expect(0, 1919, '\p{^arabicsup}', "");
Expect(0, 1919, '\P{arabicsup}', "");
Expect(1, 1919, '\P{^arabicsup}', "");
Expect(0, 1920, '\p{arabicsup}', "");
Expect(1, 1920, '\p{^arabicsup}', "");
Expect(1, 1920, '\P{arabicsup}', "");
Expect(0, 1920, '\P{^arabicsup}', "");
Expect(1, 1919, '\p{	_ARABIC_Sup}', "");
Expect(0, 1919, '\p{^	_ARABIC_Sup}', "");
Expect(0, 1919, '\P{	_ARABIC_Sup}', "");
Expect(1, 1919, '\P{^	_ARABIC_Sup}', "");
Expect(0, 1920, '\p{	_ARABIC_Sup}', "");
Expect(1, 1920, '\p{^	_ARABIC_Sup}', "");
Expect(1, 1920, '\P{	_ARABIC_Sup}', "");
Expect(0, 1920, '\P{^	_ARABIC_Sup}', "");
Error('\p{IS_Arabic_Sup:=}');
Error('\P{IS_Arabic_Sup:=}');
Expect(1, 1919, '\p{isarabicsup}', "");
Expect(0, 1919, '\p{^isarabicsup}', "");
Expect(0, 1919, '\P{isarabicsup}', "");
Expect(1, 1919, '\P{^isarabicsup}', "");
Expect(0, 1920, '\p{isarabicsup}', "");
Expect(1, 1920, '\p{^isarabicsup}', "");
Expect(1, 1920, '\P{isarabicsup}', "");
Expect(0, 1920, '\P{^isarabicsup}', "");
Expect(1, 1919, '\p{-	IS_ARABIC_Sup}', "");
Expect(0, 1919, '\p{^-	IS_ARABIC_Sup}', "");
Expect(0, 1919, '\P{-	IS_ARABIC_Sup}', "");
Expect(1, 1919, '\P{^-	IS_ARABIC_Sup}', "");
Expect(0, 1920, '\p{-	IS_ARABIC_Sup}', "");
Expect(1, 1920, '\p{^-	IS_ARABIC_Sup}', "");
Expect(1, 1920, '\P{-	IS_ARABIC_Sup}', "");
Expect(0, 1920, '\P{^-	IS_ARABIC_Sup}', "");
Error('\p{_IN_Arabic_Sup/a/}');
Error('\P{_IN_Arabic_Sup/a/}');
Expect(1, 1919, '\p{inarabicsup}', "");
Expect(0, 1919, '\p{^inarabicsup}', "");
Expect(0, 1919, '\P{inarabicsup}', "");
Expect(1, 1919, '\P{^inarabicsup}', "");
Expect(0, 1920, '\p{inarabicsup}', "");
Expect(1, 1920, '\p{^inarabicsup}', "");
Expect(1, 1920, '\P{inarabicsup}', "");
Expect(0, 1920, '\P{^inarabicsup}', "");
Expect(1, 1919, '\p{-	In_Arabic_Sup}', "");
Expect(0, 1919, '\p{^-	In_Arabic_Sup}', "");
Expect(0, 1919, '\P{-	In_Arabic_Sup}', "");
Expect(1, 1919, '\P{^-	In_Arabic_Sup}', "");
Expect(0, 1920, '\p{-	In_Arabic_Sup}', "");
Expect(1, 1920, '\p{^-	In_Arabic_Sup}', "");
Expect(1, 1920, '\P{-	In_Arabic_Sup}', "");
Expect(0, 1920, '\P{^-	In_Arabic_Sup}', "");
Error('\p{-ARMENIAN:=}');
Error('\P{-ARMENIAN:=}');
Expect(1, 64279, '\p{armenian}', "");
Expect(0, 64279, '\p{^armenian}', "");
Expect(0, 64279, '\P{armenian}', "");
Expect(1, 64279, '\P{^armenian}', "");
Expect(0, 64280, '\p{armenian}', "");
Expect(1, 64280, '\p{^armenian}', "");
Expect(1, 64280, '\P{armenian}', "");
Expect(0, 64280, '\P{^armenian}', "");
Expect(1, 64279, '\p{		ARMENIAN}', "");
Expect(0, 64279, '\p{^		ARMENIAN}', "");
Expect(0, 64279, '\P{		ARMENIAN}', "");
Expect(1, 64279, '\P{^		ARMENIAN}', "");
Expect(0, 64280, '\p{		ARMENIAN}', "");
Expect(1, 64280, '\p{^		ARMENIAN}', "");
Expect(1, 64280, '\P{		ARMENIAN}', "");
Expect(0, 64280, '\P{^		ARMENIAN}', "");
Error('\p{:=	Is_Armenian}');
Error('\P{:=	Is_Armenian}');
Expect(1, 64279, '\p{isarmenian}', "");
Expect(0, 64279, '\p{^isarmenian}', "");
Expect(0, 64279, '\P{isarmenian}', "");
Expect(1, 64279, '\P{^isarmenian}', "");
Expect(0, 64280, '\p{isarmenian}', "");
Expect(1, 64280, '\p{^isarmenian}', "");
Expect(1, 64280, '\P{isarmenian}', "");
Expect(0, 64280, '\P{^isarmenian}', "");
Expect(1, 64279, '\p{	-is_Armenian}', "");
Expect(0, 64279, '\p{^	-is_Armenian}', "");
Expect(0, 64279, '\P{	-is_Armenian}', "");
Expect(1, 64279, '\P{^	-is_Armenian}', "");
Expect(0, 64280, '\p{	-is_Armenian}', "");
Expect(1, 64280, '\p{^	-is_Armenian}', "");
Expect(1, 64280, '\P{	-is_Armenian}', "");
Expect(0, 64280, '\P{^	-is_Armenian}', "");
Error('\p{:=_ARMN}');
Error('\P{:=_ARMN}');
Expect(1, 64279, '\p{armn}', "");
Expect(0, 64279, '\p{^armn}', "");
Expect(0, 64279, '\P{armn}', "");
Expect(1, 64279, '\P{^armn}', "");
Expect(0, 64280, '\p{armn}', "");
Expect(1, 64280, '\p{^armn}', "");
Expect(1, 64280, '\P{armn}', "");
Expect(0, 64280, '\P{^armn}', "");
Expect(1, 64279, '\p{	Armn}', "");
Expect(0, 64279, '\p{^	Armn}', "");
Expect(0, 64279, '\P{	Armn}', "");
Expect(1, 64279, '\P{^	Armn}', "");
Expect(0, 64280, '\p{	Armn}', "");
Expect(1, 64280, '\p{^	Armn}', "");
Expect(1, 64280, '\P{	Armn}', "");
Expect(0, 64280, '\P{^	Armn}', "");
Error('\p{ :=Is_Armn}');
Error('\P{ :=Is_Armn}');
Expect(1, 64279, '\p{isarmn}', "");
Expect(0, 64279, '\p{^isarmn}', "");
Expect(0, 64279, '\P{isarmn}', "");
Expect(1, 64279, '\P{^isarmn}', "");
Expect(0, 64280, '\p{isarmn}', "");
Expect(1, 64280, '\p{^isarmn}', "");
Expect(1, 64280, '\P{isarmn}', "");
Expect(0, 64280, '\P{^isarmn}', "");
Expect(1, 64279, '\p{ 	Is_armn}', "");
Expect(0, 64279, '\p{^ 	Is_armn}', "");
Expect(0, 64279, '\P{ 	Is_armn}', "");
Expect(1, 64279, '\P{^ 	Is_armn}', "");
Expect(0, 64280, '\p{ 	Is_armn}', "");
Expect(1, 64280, '\p{^ 	Is_armn}', "");
Expect(1, 64280, '\P{ 	Is_armn}', "");
Expect(0, 64280, '\P{^ 	Is_armn}', "");
Error('\p{__Arrows/a/}');
Error('\P{__Arrows/a/}');
Expect(1, 8703, '\p{arrows}', "");
Expect(0, 8703, '\p{^arrows}', "");
Expect(0, 8703, '\P{arrows}', "");
Expect(1, 8703, '\P{^arrows}', "");
Expect(0, 8704, '\p{arrows}', "");
Expect(1, 8704, '\p{^arrows}', "");
Expect(1, 8704, '\P{arrows}', "");
Expect(0, 8704, '\P{^arrows}', "");
Expect(1, 8703, '\p{-arrows}', "");
Expect(0, 8703, '\p{^-arrows}', "");
Expect(0, 8703, '\P{-arrows}', "");
Expect(1, 8703, '\P{^-arrows}', "");
Expect(0, 8704, '\p{-arrows}', "");
Expect(1, 8704, '\p{^-arrows}', "");
Expect(1, 8704, '\P{-arrows}', "");
Expect(0, 8704, '\P{^-arrows}', "");
Error('\p{/a/ Is_Arrows}');
Error('\P{/a/ Is_Arrows}');
Expect(1, 8703, '\p{isarrows}', "");
Expect(0, 8703, '\p{^isarrows}', "");
Expect(0, 8703, '\P{isarrows}', "");
Expect(1, 8703, '\P{^isarrows}', "");
Expect(0, 8704, '\p{isarrows}', "");
Expect(1, 8704, '\p{^isarrows}', "");
Expect(1, 8704, '\P{isarrows}', "");
Expect(0, 8704, '\P{^isarrows}', "");
Expect(1, 8703, '\p{-_Is_ARROWS}', "");
Expect(0, 8703, '\p{^-_Is_ARROWS}', "");
Expect(0, 8703, '\P{-_Is_ARROWS}', "");
Expect(1, 8703, '\P{^-_Is_ARROWS}', "");
Expect(0, 8704, '\p{-_Is_ARROWS}', "");
Expect(1, 8704, '\p{^-_Is_ARROWS}', "");
Expect(1, 8704, '\P{-_Is_ARROWS}', "");
Expect(0, 8704, '\P{^-_Is_ARROWS}', "");
Error('\p{-In_ARROWS/a/}');
Error('\P{-In_ARROWS/a/}');
Expect(1, 8703, '\p{inarrows}', "");
Expect(0, 8703, '\p{^inarrows}', "");
Expect(0, 8703, '\P{inarrows}', "");
Expect(1, 8703, '\P{^inarrows}', "");
Expect(0, 8704, '\p{inarrows}', "");
Expect(1, 8704, '\p{^inarrows}', "");
Expect(1, 8704, '\P{inarrows}', "");
Expect(0, 8704, '\P{^inarrows}', "");
Expect(1, 8703, '\p{ -In_arrows}', "");
Expect(0, 8703, '\p{^ -In_arrows}', "");
Expect(0, 8703, '\P{ -In_arrows}', "");
Expect(1, 8703, '\P{^ -In_arrows}', "");
Expect(0, 8704, '\p{ -In_arrows}', "");
Expect(1, 8704, '\p{^ -In_arrows}', "");
Expect(1, 8704, '\P{ -In_arrows}', "");
Expect(0, 8704, '\P{^ -In_arrows}', "");
Error('\p{	/a/Assigned}');
Error('\P{	/a/Assigned}');
Expect(1, 1114109, '\p{assigned}', "");
Expect(0, 1114109, '\p{^assigned}', "");
Expect(0, 1114109, '\P{assigned}', "");
Expect(1, 1114109, '\P{^assigned}', "");
Expect(0, 918000, '\p{assigned}', "");
Expect(1, 918000, '\p{^assigned}', "");
Expect(1, 918000, '\P{assigned}', "");
Expect(0, 918000, '\P{^assigned}', "");
Expect(1, 1114109, '\p{	Assigned}', "");
Expect(0, 1114109, '\p{^	Assigned}', "");
Expect(0, 1114109, '\P{	Assigned}', "");
Expect(1, 1114109, '\P{^	Assigned}', "");
Expect(0, 918000, '\p{	Assigned}', "");
Expect(1, 918000, '\p{^	Assigned}', "");
Expect(1, 918000, '\P{	Assigned}', "");
Expect(0, 918000, '\P{^	Assigned}', "");
Error('\p{:=is_ASSIGNED}');
Error('\P{:=is_ASSIGNED}');
Expect(1, 1114109, '\p{isassigned}', "");
Expect(0, 1114109, '\p{^isassigned}', "");
Expect(0, 1114109, '\P{isassigned}', "");
Expect(1, 1114109, '\P{^isassigned}', "");
Expect(0, 918000, '\p{isassigned}', "");
Expect(1, 918000, '\p{^isassigned}', "");
Expect(1, 918000, '\P{isassigned}', "");
Expect(0, 918000, '\P{^isassigned}', "");
Expect(1, 1114109, '\p{	Is_ASSIGNED}', "");
Expect(0, 1114109, '\p{^	Is_ASSIGNED}', "");
Expect(0, 1114109, '\P{	Is_ASSIGNED}', "");
Expect(1, 1114109, '\P{^	Is_ASSIGNED}', "");
Expect(0, 918000, '\p{	Is_ASSIGNED}', "");
Expect(1, 918000, '\p{^	Is_ASSIGNED}', "");
Expect(1, 918000, '\P{	Is_ASSIGNED}', "");
Expect(0, 918000, '\P{^	Is_ASSIGNED}', "");
Error('\p{:= -AVESTAN}');
Error('\P{:= -AVESTAN}');
Expect(1, 68415, '\p{avestan}', "");
Expect(0, 68415, '\p{^avestan}', "");
Expect(0, 68415, '\P{avestan}', "");
Expect(1, 68415, '\P{^avestan}', "");
Expect(0, 68416, '\p{avestan}', "");
Expect(1, 68416, '\p{^avestan}', "");
Expect(1, 68416, '\P{avestan}', "");
Expect(0, 68416, '\P{^avestan}', "");
Expect(1, 68415, '\p{_	Avestan}', "");
Expect(0, 68415, '\p{^_	Avestan}', "");
Expect(0, 68415, '\P{_	Avestan}', "");
Expect(1, 68415, '\P{^_	Avestan}', "");
Expect(0, 68416, '\p{_	Avestan}', "");
Expect(1, 68416, '\p{^_	Avestan}', "");
Expect(1, 68416, '\P{_	Avestan}', "");
Expect(0, 68416, '\P{^_	Avestan}', "");
Error('\p{/a/		is_avestan}');
Error('\P{/a/		is_avestan}');
Expect(1, 68415, '\p{isavestan}', "");
Expect(0, 68415, '\p{^isavestan}', "");
Expect(0, 68415, '\P{isavestan}', "");
Expect(1, 68415, '\P{^isavestan}', "");
Expect(0, 68416, '\p{isavestan}', "");
Expect(1, 68416, '\p{^isavestan}', "");
Expect(1, 68416, '\P{isavestan}', "");
Expect(0, 68416, '\P{^isavestan}', "");
Expect(1, 68415, '\p{  Is_Avestan}', "");
Expect(0, 68415, '\p{^  Is_Avestan}', "");
Expect(0, 68415, '\P{  Is_Avestan}', "");
Expect(1, 68415, '\P{^  Is_Avestan}', "");
Expect(0, 68416, '\p{  Is_Avestan}', "");
Expect(1, 68416, '\p{^  Is_Avestan}', "");
Expect(1, 68416, '\P{  Is_Avestan}', "");
Expect(0, 68416, '\P{^  Is_Avestan}', "");
Error('\p{/a/-_AVST}');
Error('\P{/a/-_AVST}');
Expect(1, 68415, '\p{avst}', "");
Expect(0, 68415, '\p{^avst}', "");
Expect(0, 68415, '\P{avst}', "");
Expect(1, 68415, '\P{^avst}', "");
Expect(0, 68416, '\p{avst}', "");
Expect(1, 68416, '\p{^avst}', "");
Expect(1, 68416, '\P{avst}', "");
Expect(0, 68416, '\P{^avst}', "");
Expect(1, 68415, '\p{ 	Avst}', "");
Expect(0, 68415, '\p{^ 	Avst}', "");
Expect(0, 68415, '\P{ 	Avst}', "");
Expect(1, 68415, '\P{^ 	Avst}', "");
Expect(0, 68416, '\p{ 	Avst}', "");
Expect(1, 68416, '\p{^ 	Avst}', "");
Expect(1, 68416, '\P{ 	Avst}', "");
Expect(0, 68416, '\P{^ 	Avst}', "");
Error('\p{:=_-IS_Avst}');
Error('\P{:=_-IS_Avst}');
Expect(1, 68415, '\p{isavst}', "");
Expect(0, 68415, '\p{^isavst}', "");
Expect(0, 68415, '\P{isavst}', "");
Expect(1, 68415, '\P{^isavst}', "");
Expect(0, 68416, '\p{isavst}', "");
Expect(1, 68416, '\p{^isavst}', "");
Expect(1, 68416, '\P{isavst}', "");
Expect(0, 68416, '\P{^isavst}', "");
Expect(1, 68415, '\p{		IS_Avst}', "");
Expect(0, 68415, '\p{^		IS_Avst}', "");
Expect(0, 68415, '\P{		IS_Avst}', "");
Expect(1, 68415, '\P{^		IS_Avst}', "");
Expect(0, 68416, '\p{		IS_Avst}', "");
Expect(1, 68416, '\p{^		IS_Avst}', "");
Expect(1, 68416, '\P{		IS_Avst}', "");
Expect(0, 68416, '\P{^		IS_Avst}', "");
Error('\p{:=BALINESE}');
Error('\P{:=BALINESE}');
Expect(1, 7036, '\p{balinese}', "");
Expect(0, 7036, '\p{^balinese}', "");
Expect(0, 7036, '\P{balinese}', "");
Expect(1, 7036, '\P{^balinese}', "");
Expect(0, 7037, '\p{balinese}', "");
Expect(1, 7037, '\p{^balinese}', "");
Expect(1, 7037, '\P{balinese}', "");
Expect(0, 7037, '\P{^balinese}', "");
Expect(1, 7036, '\p{-Balinese}', "");
Expect(0, 7036, '\p{^-Balinese}', "");
Expect(0, 7036, '\P{-Balinese}', "");
Expect(1, 7036, '\P{^-Balinese}', "");
Expect(0, 7037, '\p{-Balinese}', "");
Expect(1, 7037, '\p{^-Balinese}', "");
Expect(1, 7037, '\P{-Balinese}', "");
Expect(0, 7037, '\P{^-Balinese}', "");
Error('\p{	-Is_BALINESE:=}');
Error('\P{	-Is_BALINESE:=}');
Expect(1, 7036, '\p{isbalinese}', "");
Expect(0, 7036, '\p{^isbalinese}', "");
Expect(0, 7036, '\P{isbalinese}', "");
Expect(1, 7036, '\P{^isbalinese}', "");
Expect(0, 7037, '\p{isbalinese}', "");
Expect(1, 7037, '\p{^isbalinese}', "");
Expect(1, 7037, '\P{isbalinese}', "");
Expect(0, 7037, '\P{^isbalinese}', "");
Expect(1, 7036, '\p{  Is_Balinese}', "");
Expect(0, 7036, '\p{^  Is_Balinese}', "");
Expect(0, 7036, '\P{  Is_Balinese}', "");
Expect(1, 7036, '\P{^  Is_Balinese}', "");
Expect(0, 7037, '\p{  Is_Balinese}', "");
Expect(1, 7037, '\p{^  Is_Balinese}', "");
Expect(1, 7037, '\P{  Is_Balinese}', "");
Expect(0, 7037, '\P{^  Is_Balinese}', "");
Error('\p{/a/-bali}');
Error('\P{/a/-bali}');
Expect(1, 7036, '\p{bali}', "");
Expect(0, 7036, '\p{^bali}', "");
Expect(0, 7036, '\P{bali}', "");
Expect(1, 7036, '\P{^bali}', "");
Expect(0, 7037, '\p{bali}', "");
Expect(1, 7037, '\p{^bali}', "");
Expect(1, 7037, '\P{bali}', "");
Expect(0, 7037, '\P{^bali}', "");
Expect(1, 7036, '\p{	 Bali}', "");
Expect(0, 7036, '\p{^	 Bali}', "");
Expect(0, 7036, '\P{	 Bali}', "");
Expect(1, 7036, '\P{^	 Bali}', "");
Expect(0, 7037, '\p{	 Bali}', "");
Expect(1, 7037, '\p{^	 Bali}', "");
Expect(1, 7037, '\P{	 Bali}', "");
Expect(0, 7037, '\P{^	 Bali}', "");
Error('\p{	/a/is_Bali}');
Error('\P{	/a/is_Bali}');
Expect(1, 7036, '\p{isbali}', "");
Expect(0, 7036, '\p{^isbali}', "");
Expect(0, 7036, '\P{isbali}', "");
Expect(1, 7036, '\P{^isbali}', "");
Expect(0, 7037, '\p{isbali}', "");
Expect(1, 7037, '\p{^isbali}', "");
Expect(1, 7037, '\P{isbali}', "");
Expect(0, 7037, '\P{^isbali}', "");
Expect(1, 7036, '\p{ -IS_Bali}', "");
Expect(0, 7036, '\p{^ -IS_Bali}', "");
Expect(0, 7036, '\P{ -IS_Bali}', "");
Expect(1, 7036, '\P{^ -IS_Bali}', "");
Expect(0, 7037, '\p{ -IS_Bali}', "");
Expect(1, 7037, '\p{^ -IS_Bali}', "");
Expect(1, 7037, '\P{ -IS_Bali}', "");
Expect(0, 7037, '\P{^ -IS_Bali}', "");
Error('\p{:=- Bamum}');
Error('\P{:=- Bamum}');
Expect(1, 92728, '\p{bamum}', "");
Expect(0, 92728, '\p{^bamum}', "");
Expect(0, 92728, '\P{bamum}', "");
Expect(1, 92728, '\P{^bamum}', "");
Expect(0, 92729, '\p{bamum}', "");
Expect(1, 92729, '\p{^bamum}', "");
Expect(1, 92729, '\P{bamum}', "");
Expect(0, 92729, '\P{^bamum}', "");
Expect(1, 92728, '\p{	_Bamum}', "");
Expect(0, 92728, '\p{^	_Bamum}', "");
Expect(0, 92728, '\P{	_Bamum}', "");
Expect(1, 92728, '\P{^	_Bamum}', "");
Expect(0, 92729, '\p{	_Bamum}', "");
Expect(1, 92729, '\p{^	_Bamum}', "");
Expect(1, 92729, '\P{	_Bamum}', "");
Expect(0, 92729, '\P{^	_Bamum}', "");
Error('\p{		IS_Bamum:=}');
Error('\P{		IS_Bamum:=}');
Expect(1, 92728, '\p{isbamum}', "");
Expect(0, 92728, '\p{^isbamum}', "");
Expect(0, 92728, '\P{isbamum}', "");
Expect(1, 92728, '\P{^isbamum}', "");
Expect(0, 92729, '\p{isbamum}', "");
Expect(1, 92729, '\p{^isbamum}', "");
Expect(1, 92729, '\P{isbamum}', "");
Expect(0, 92729, '\P{^isbamum}', "");
Expect(1, 92728, '\p{ 	is_Bamum}', "");
Expect(0, 92728, '\p{^ 	is_Bamum}', "");
Expect(0, 92728, '\P{ 	is_Bamum}', "");
Expect(1, 92728, '\P{^ 	is_Bamum}', "");
Expect(0, 92729, '\p{ 	is_Bamum}', "");
Expect(1, 92729, '\p{^ 	is_Bamum}', "");
Expect(1, 92729, '\P{ 	is_Bamum}', "");
Expect(0, 92729, '\P{^ 	is_Bamum}', "");
Error('\p{ /a/Bamu}');
Error('\P{ /a/Bamu}');
Expect(1, 92728, '\p{bamu}', "");
Expect(0, 92728, '\p{^bamu}', "");
Expect(0, 92728, '\P{bamu}', "");
Expect(1, 92728, '\P{^bamu}', "");
Expect(0, 92729, '\p{bamu}', "");
Expect(1, 92729, '\p{^bamu}', "");
Expect(1, 92729, '\P{bamu}', "");
Expect(0, 92729, '\P{^bamu}', "");
Expect(1, 92728, '\p{  Bamu}', "");
Expect(0, 92728, '\p{^  Bamu}', "");
Expect(0, 92728, '\P{  Bamu}', "");
Expect(1, 92728, '\P{^  Bamu}', "");
Expect(0, 92729, '\p{  Bamu}', "");
Expect(1, 92729, '\p{^  Bamu}', "");
Expect(1, 92729, '\P{  Bamu}', "");
Expect(0, 92729, '\P{^  Bamu}', "");
Error('\p{_-IS_Bamu/a/}');
Error('\P{_-IS_Bamu/a/}');
Expect(1, 92728, '\p{isbamu}', "");
Expect(0, 92728, '\p{^isbamu}', "");
Expect(0, 92728, '\P{isbamu}', "");
Expect(1, 92728, '\P{^isbamu}', "");
Expect(0, 92729, '\p{isbamu}', "");
Expect(1, 92729, '\p{^isbamu}', "");
Expect(1, 92729, '\P{isbamu}', "");
Expect(0, 92729, '\P{^isbamu}', "");
Expect(1, 92728, '\p{-	Is_Bamu}', "");
Expect(0, 92728, '\p{^-	Is_Bamu}', "");
Expect(0, 92728, '\P{-	Is_Bamu}', "");
Expect(1, 92728, '\P{^-	Is_Bamu}', "");
Expect(0, 92729, '\p{-	Is_Bamu}', "");
Expect(1, 92729, '\p{^-	Is_Bamu}', "");
Expect(1, 92729, '\P{-	Is_Bamu}', "");
Expect(0, 92729, '\P{^-	Is_Bamu}', "");
Error('\p{- Bamum_Supplement/a/}');
Error('\P{- Bamum_Supplement/a/}');
Expect(1, 92735, '\p{bamumsupplement}', "");
Expect(0, 92735, '\p{^bamumsupplement}', "");
Expect(0, 92735, '\P{bamumsupplement}', "");
Expect(1, 92735, '\P{^bamumsupplement}', "");
Expect(0, 92736, '\p{bamumsupplement}', "");
Expect(1, 92736, '\p{^bamumsupplement}', "");
Expect(1, 92736, '\P{bamumsupplement}', "");
Expect(0, 92736, '\P{^bamumsupplement}', "");
Expect(1, 92735, '\p{-BAMUM_Supplement}', "");
Expect(0, 92735, '\p{^-BAMUM_Supplement}', "");
Expect(0, 92735, '\P{-BAMUM_Supplement}', "");
Expect(1, 92735, '\P{^-BAMUM_Supplement}', "");
Expect(0, 92736, '\p{-BAMUM_Supplement}', "");
Expect(1, 92736, '\p{^-BAMUM_Supplement}', "");
Expect(1, 92736, '\P{-BAMUM_Supplement}', "");
Expect(0, 92736, '\P{^-BAMUM_Supplement}', "");
Error('\p{_IS_bamum_supplement:=}');
Error('\P{_IS_bamum_supplement:=}');
Expect(1, 92735, '\p{isbamumsupplement}', "");
Expect(0, 92735, '\p{^isbamumsupplement}', "");
Expect(0, 92735, '\P{isbamumsupplement}', "");
Expect(1, 92735, '\P{^isbamumsupplement}', "");
Expect(0, 92736, '\p{isbamumsupplement}', "");
Expect(1, 92736, '\p{^isbamumsupplement}', "");
Expect(1, 92736, '\P{isbamumsupplement}', "");
Expect(0, 92736, '\P{^isbamumsupplement}', "");
Expect(1, 92735, '\p{	is_BAMUM_Supplement}', "");
Expect(0, 92735, '\p{^	is_BAMUM_Supplement}', "");
Expect(0, 92735, '\P{	is_BAMUM_Supplement}', "");
Expect(1, 92735, '\P{^	is_BAMUM_Supplement}', "");
Expect(0, 92736, '\p{	is_BAMUM_Supplement}', "");
Expect(1, 92736, '\p{^	is_BAMUM_Supplement}', "");
Expect(1, 92736, '\P{	is_BAMUM_Supplement}', "");
Expect(0, 92736, '\P{^	is_BAMUM_Supplement}', "");
Error('\p{in_Bamum_Supplement:=}');
Error('\P{in_Bamum_Supplement:=}');
Expect(1, 92735, '\p{inbamumsupplement}', "");
Expect(0, 92735, '\p{^inbamumsupplement}', "");
Expect(0, 92735, '\P{inbamumsupplement}', "");
Expect(1, 92735, '\P{^inbamumsupplement}', "");
Expect(0, 92736, '\p{inbamumsupplement}', "");
Expect(1, 92736, '\p{^inbamumsupplement}', "");
Expect(1, 92736, '\P{inbamumsupplement}', "");
Expect(0, 92736, '\P{^inbamumsupplement}', "");
Expect(1, 92735, '\p{_IN_Bamum_Supplement}', "");
Expect(0, 92735, '\p{^_IN_Bamum_Supplement}', "");
Expect(0, 92735, '\P{_IN_Bamum_Supplement}', "");
Expect(1, 92735, '\P{^_IN_Bamum_Supplement}', "");
Expect(0, 92736, '\p{_IN_Bamum_Supplement}', "");
Expect(1, 92736, '\p{^_IN_Bamum_Supplement}', "");
Expect(1, 92736, '\P{_IN_Bamum_Supplement}', "");
Expect(0, 92736, '\P{^_IN_Bamum_Supplement}', "");
Error('\p{-BAMUM_Sup:=}');
Error('\P{-BAMUM_Sup:=}');
Expect(1, 92735, '\p{bamumsup}', "");
Expect(0, 92735, '\p{^bamumsup}', "");
Expect(0, 92735, '\P{bamumsup}', "");
Expect(1, 92735, '\P{^bamumsup}', "");
Expect(0, 92736, '\p{bamumsup}', "");
Expect(1, 92736, '\p{^bamumsup}', "");
Expect(1, 92736, '\P{bamumsup}', "");
Expect(0, 92736, '\P{^bamumsup}', "");
Expect(1, 92735, '\p{	_bamum_SUP}', "");
Expect(0, 92735, '\p{^	_bamum_SUP}', "");
Expect(0, 92735, '\P{	_bamum_SUP}', "");
Expect(1, 92735, '\P{^	_bamum_SUP}', "");
Expect(0, 92736, '\p{	_bamum_SUP}', "");
Expect(1, 92736, '\p{^	_bamum_SUP}', "");
Expect(1, 92736, '\P{	_bamum_SUP}', "");
Expect(0, 92736, '\P{^	_bamum_SUP}', "");
Error('\p{	/a/IS_BAMUM_Sup}');
Error('\P{	/a/IS_BAMUM_Sup}');
Expect(1, 92735, '\p{isbamumsup}', "");
Expect(0, 92735, '\p{^isbamumsup}', "");
Expect(0, 92735, '\P{isbamumsup}', "");
Expect(1, 92735, '\P{^isbamumsup}', "");
Expect(0, 92736, '\p{isbamumsup}', "");
Expect(1, 92736, '\p{^isbamumsup}', "");
Expect(1, 92736, '\P{isbamumsup}', "");
Expect(0, 92736, '\P{^isbamumsup}', "");
Expect(1, 92735, '\p{  is_Bamum_SUP}', "");
Expect(0, 92735, '\p{^  is_Bamum_SUP}', "");
Expect(0, 92735, '\P{  is_Bamum_SUP}', "");
Expect(1, 92735, '\P{^  is_Bamum_SUP}', "");
Expect(0, 92736, '\p{  is_Bamum_SUP}', "");
Expect(1, 92736, '\p{^  is_Bamum_SUP}', "");
Expect(1, 92736, '\P{  is_Bamum_SUP}', "");
Expect(0, 92736, '\P{^  is_Bamum_SUP}', "");
Error('\p{ :=in_bamum_Sup}');
Error('\P{ :=in_bamum_Sup}');
Expect(1, 92735, '\p{inbamumsup}', "");
Expect(0, 92735, '\p{^inbamumsup}', "");
Expect(0, 92735, '\P{inbamumsup}', "");
Expect(1, 92735, '\P{^inbamumsup}', "");
Expect(0, 92736, '\p{inbamumsup}', "");
Expect(1, 92736, '\p{^inbamumsup}', "");
Expect(1, 92736, '\P{inbamumsup}', "");
Expect(0, 92736, '\P{^inbamumsup}', "");
Expect(1, 92735, '\p{-In_Bamum_SUP}', "");
Expect(0, 92735, '\p{^-In_Bamum_SUP}', "");
Expect(0, 92735, '\P{-In_Bamum_SUP}', "");
Expect(1, 92735, '\P{^-In_Bamum_SUP}', "");
Expect(0, 92736, '\p{-In_Bamum_SUP}', "");
Expect(1, 92736, '\p{^-In_Bamum_SUP}', "");
Expect(1, 92736, '\P{-In_Bamum_SUP}', "");
Expect(0, 92736, '\P{^-In_Bamum_SUP}', "");
Error('\p{	:=BATAK}');
Error('\P{	:=BATAK}');
Expect(1, 7167, '\p{batak}', "");
Expect(0, 7167, '\p{^batak}', "");
Expect(0, 7167, '\P{batak}', "");
Expect(1, 7167, '\P{^batak}', "");
Expect(0, 7168, '\p{batak}', "");
Expect(1, 7168, '\p{^batak}', "");
Expect(1, 7168, '\P{batak}', "");
Expect(0, 7168, '\P{^batak}', "");
Expect(1, 7167, '\p{-BATAK}', "");
Expect(0, 7167, '\p{^-BATAK}', "");
Expect(0, 7167, '\P{-BATAK}', "");
Expect(1, 7167, '\P{^-BATAK}', "");
Expect(0, 7168, '\p{-BATAK}', "");
Expect(1, 7168, '\p{^-BATAK}', "");
Expect(1, 7168, '\P{-BATAK}', "");
Expect(0, 7168, '\P{^-BATAK}', "");
Error('\p{:=- is_BATAK}');
Error('\P{:=- is_BATAK}');
Expect(1, 7167, '\p{isbatak}', "");
Expect(0, 7167, '\p{^isbatak}', "");
Expect(0, 7167, '\P{isbatak}', "");
Expect(1, 7167, '\P{^isbatak}', "");
Expect(0, 7168, '\p{isbatak}', "");
Expect(1, 7168, '\p{^isbatak}', "");
Expect(1, 7168, '\P{isbatak}', "");
Expect(0, 7168, '\P{^isbatak}', "");
Expect(1, 7167, '\p{_Is_BATAK}', "");
Expect(0, 7167, '\p{^_Is_BATAK}', "");
Expect(0, 7167, '\P{_Is_BATAK}', "");
Expect(1, 7167, '\P{^_Is_BATAK}', "");
Expect(0, 7168, '\p{_Is_BATAK}', "");
Expect(1, 7168, '\p{^_Is_BATAK}', "");
Expect(1, 7168, '\P{_Is_BATAK}', "");
Expect(0, 7168, '\P{^_Is_BATAK}', "");
Error('\p{:=-	Batk}');
Error('\P{:=-	Batk}');
Expect(1, 7167, '\p{batk}', "");
Expect(0, 7167, '\p{^batk}', "");
Expect(0, 7167, '\P{batk}', "");
Expect(1, 7167, '\P{^batk}', "");
Expect(0, 7168, '\p{batk}', "");
Expect(1, 7168, '\p{^batk}', "");
Expect(1, 7168, '\P{batk}', "");
Expect(0, 7168, '\P{^batk}', "");
Expect(1, 7167, '\p{- Batk}', "");
Expect(0, 7167, '\p{^- Batk}', "");
Expect(0, 7167, '\P{- Batk}', "");
Expect(1, 7167, '\P{^- Batk}', "");
Expect(0, 7168, '\p{- Batk}', "");
Expect(1, 7168, '\p{^- Batk}', "");
Expect(1, 7168, '\P{- Batk}', "");
Expect(0, 7168, '\P{^- Batk}', "");
Error('\p{:=-Is_batk}');
Error('\P{:=-Is_batk}');
Expect(1, 7167, '\p{isbatk}', "");
Expect(0, 7167, '\p{^isbatk}', "");
Expect(0, 7167, '\P{isbatk}', "");
Expect(1, 7167, '\P{^isbatk}', "");
Expect(0, 7168, '\p{isbatk}', "");
Expect(1, 7168, '\p{^isbatk}', "");
Expect(1, 7168, '\P{isbatk}', "");
Expect(0, 7168, '\P{^isbatk}', "");
Expect(1, 7167, '\p{ 	Is_Batk}', "");
Expect(0, 7167, '\p{^ 	Is_Batk}', "");
Expect(0, 7167, '\P{ 	Is_Batk}', "");
Expect(1, 7167, '\P{^ 	Is_Batk}', "");
Expect(0, 7168, '\p{ 	Is_Batk}', "");
Expect(1, 7168, '\p{^ 	Is_Batk}', "");
Expect(1, 7168, '\P{ 	Is_Batk}', "");
Expect(0, 7168, '\P{^ 	Is_Batk}', "");
Error('\p{/a/--Bengali}');
Error('\P{/a/--Bengali}');
Expect(1, 2555, '\p{bengali}', "");
Expect(0, 2555, '\p{^bengali}', "");
Expect(0, 2555, '\P{bengali}', "");
Expect(1, 2555, '\P{^bengali}', "");
Expect(0, 2556, '\p{bengali}', "");
Expect(1, 2556, '\p{^bengali}', "");
Expect(1, 2556, '\P{bengali}', "");
Expect(0, 2556, '\P{^bengali}', "");
Expect(1, 2555, '\p{ _bengali}', "");
Expect(0, 2555, '\p{^ _bengali}', "");
Expect(0, 2555, '\P{ _bengali}', "");
Expect(1, 2555, '\P{^ _bengali}', "");
Expect(0, 2556, '\p{ _bengali}', "");
Expect(1, 2556, '\p{^ _bengali}', "");
Expect(1, 2556, '\P{ _bengali}', "");
Expect(0, 2556, '\P{^ _bengali}', "");
Error('\p{- Is_BENGALI/a/}');
Error('\P{- Is_BENGALI/a/}');
Expect(1, 2555, '\p{isbengali}', "");
Expect(0, 2555, '\p{^isbengali}', "");
Expect(0, 2555, '\P{isbengali}', "");
Expect(1, 2555, '\P{^isbengali}', "");
Expect(0, 2556, '\p{isbengali}', "");
Expect(1, 2556, '\p{^isbengali}', "");
Expect(1, 2556, '\P{isbengali}', "");
Expect(0, 2556, '\P{^isbengali}', "");
Expect(1, 2555, '\p{	-IS_Bengali}', "");
Expect(0, 2555, '\p{^	-IS_Bengali}', "");
Expect(0, 2555, '\P{	-IS_Bengali}', "");
Expect(1, 2555, '\P{^	-IS_Bengali}', "");
Expect(0, 2556, '\p{	-IS_Bengali}', "");
Expect(1, 2556, '\p{^	-IS_Bengali}', "");
Expect(1, 2556, '\P{	-IS_Bengali}', "");
Expect(0, 2556, '\P{^	-IS_Bengali}', "");
Error('\p{:=	BENG}');
Error('\P{:=	BENG}');
Expect(1, 2555, '\p{beng}', "");
Expect(0, 2555, '\p{^beng}', "");
Expect(0, 2555, '\P{beng}', "");
Expect(1, 2555, '\P{^beng}', "");
Expect(0, 2556, '\p{beng}', "");
Expect(1, 2556, '\p{^beng}', "");
Expect(1, 2556, '\P{beng}', "");
Expect(0, 2556, '\P{^beng}', "");
Expect(1, 2555, '\p{-_BENG}', "");
Expect(0, 2555, '\p{^-_BENG}', "");
Expect(0, 2555, '\P{-_BENG}', "");
Expect(1, 2555, '\P{^-_BENG}', "");
Expect(0, 2556, '\p{-_BENG}', "");
Expect(1, 2556, '\p{^-_BENG}', "");
Expect(1, 2556, '\P{-_BENG}', "");
Expect(0, 2556, '\P{^-_BENG}', "");
Error('\p{	:=IS_beng}');
Error('\P{	:=IS_beng}');
Expect(1, 2555, '\p{isbeng}', "");
Expect(0, 2555, '\p{^isbeng}', "");
Expect(0, 2555, '\P{isbeng}', "");
Expect(1, 2555, '\P{^isbeng}', "");
Expect(0, 2556, '\p{isbeng}', "");
Expect(1, 2556, '\p{^isbeng}', "");
Expect(1, 2556, '\P{isbeng}', "");
Expect(0, 2556, '\P{^isbeng}', "");
Expect(1, 2555, '\p{	is_Beng}', "");
Expect(0, 2555, '\p{^	is_Beng}', "");
Expect(0, 2555, '\P{	is_Beng}', "");
Expect(1, 2555, '\P{^	is_Beng}', "");
Expect(0, 2556, '\p{	is_Beng}', "");
Expect(1, 2556, '\p{^	is_Beng}', "");
Expect(1, 2556, '\P{	is_Beng}', "");
Expect(0, 2556, '\P{^	is_Beng}', "");
Error('\p{ /a/Bidi_Control}');
Error('\P{ /a/Bidi_Control}');
Expect(1, 8297, '\p{bidicontrol}', "");
Expect(0, 8297, '\p{^bidicontrol}', "");
Expect(0, 8297, '\P{bidicontrol}', "");
Expect(1, 8297, '\P{^bidicontrol}', "");
Expect(0, 8298, '\p{bidicontrol}', "");
Expect(1, 8298, '\p{^bidicontrol}', "");
Expect(1, 8298, '\P{bidicontrol}', "");
Expect(0, 8298, '\P{^bidicontrol}', "");
Expect(1, 8297, '\p{--Bidi_CONTROL}', "");
Expect(0, 8297, '\p{^--Bidi_CONTROL}', "");
Expect(0, 8297, '\P{--Bidi_CONTROL}', "");
Expect(1, 8297, '\P{^--Bidi_CONTROL}', "");
Expect(0, 8298, '\p{--Bidi_CONTROL}', "");
Expect(1, 8298, '\p{^--Bidi_CONTROL}', "");
Expect(1, 8298, '\P{--Bidi_CONTROL}', "");
Expect(0, 8298, '\P{^--Bidi_CONTROL}', "");
Error('\p{/a/-	IS_Bidi_Control}');
Error('\P{/a/-	IS_Bidi_Control}');
Expect(1, 8297, '\p{isbidicontrol}', "");
Expect(0, 8297, '\p{^isbidicontrol}', "");
Expect(0, 8297, '\P{isbidicontrol}', "");
Expect(1, 8297, '\P{^isbidicontrol}', "");
Expect(0, 8298, '\p{isbidicontrol}', "");
Expect(1, 8298, '\p{^isbidicontrol}', "");
Expect(1, 8298, '\P{isbidicontrol}', "");
Expect(0, 8298, '\P{^isbidicontrol}', "");
Expect(1, 8297, '\p{	-IS_Bidi_Control}', "");
Expect(0, 8297, '\p{^	-IS_Bidi_Control}', "");
Expect(0, 8297, '\P{	-IS_Bidi_Control}', "");
Expect(1, 8297, '\P{^	-IS_Bidi_Control}', "");
Expect(0, 8298, '\p{	-IS_Bidi_Control}', "");
Expect(1, 8298, '\p{^	-IS_Bidi_Control}', "");
Expect(1, 8298, '\P{	-IS_Bidi_Control}', "");
Expect(0, 8298, '\P{^	-IS_Bidi_Control}', "");
Error('\p{ /a/Bidi_c}');
Error('\P{ /a/Bidi_c}');
Expect(1, 8297, '\p{bidic}', "");
Expect(0, 8297, '\p{^bidic}', "");
Expect(0, 8297, '\P{bidic}', "");
Expect(1, 8297, '\P{^bidic}', "");
Expect(0, 8298, '\p{bidic}', "");
Expect(1, 8298, '\p{^bidic}', "");
Expect(1, 8298, '\P{bidic}', "");
Expect(0, 8298, '\P{^bidic}', "");
Expect(1, 8297, '\p{	_bidi_c}', "");
Expect(0, 8297, '\p{^	_bidi_c}', "");
Expect(0, 8297, '\P{	_bidi_c}', "");
Expect(1, 8297, '\P{^	_bidi_c}', "");
Expect(0, 8298, '\p{	_bidi_c}', "");
Expect(1, 8298, '\p{^	_bidi_c}', "");
Expect(1, 8298, '\P{	_bidi_c}', "");
Expect(0, 8298, '\P{^	_bidi_c}', "");
Error('\p{/a/ 	is_BIDI_c}');
Error('\P{/a/ 	is_BIDI_c}');
Expect(1, 8297, '\p{isbidic}', "");
Expect(0, 8297, '\p{^isbidic}', "");
Expect(0, 8297, '\P{isbidic}', "");
Expect(1, 8297, '\P{^isbidic}', "");
Expect(0, 8298, '\p{isbidic}', "");
Expect(1, 8298, '\p{^isbidic}', "");
Expect(1, 8298, '\P{isbidic}', "");
Expect(0, 8298, '\P{^isbidic}', "");
Expect(1, 8297, '\p{-IS_Bidi_c}', "");
Expect(0, 8297, '\p{^-IS_Bidi_c}', "");
Expect(0, 8297, '\P{-IS_Bidi_c}', "");
Expect(1, 8297, '\P{^-IS_Bidi_c}', "");
Expect(0, 8298, '\p{-IS_Bidi_c}', "");
Expect(1, 8298, '\p{^-IS_Bidi_c}', "");
Expect(1, 8298, '\P{-IS_Bidi_c}', "");
Expect(0, 8298, '\P{^-IS_Bidi_c}', "");
Error('\p{_:=Bidi_Mirrored}');
Error('\P{_:=Bidi_Mirrored}');
Expect(1, 120771, '\p{bidimirrored}', "");
Expect(0, 120771, '\p{^bidimirrored}', "");
Expect(0, 120771, '\P{bidimirrored}', "");
Expect(1, 120771, '\P{^bidimirrored}', "");
Expect(0, 120772, '\p{bidimirrored}', "");
Expect(1, 120772, '\p{^bidimirrored}', "");
Expect(1, 120772, '\P{bidimirrored}', "");
Expect(0, 120772, '\P{^bidimirrored}', "");
Expect(1, 120771, '\p{	 BIDI_mirrored}', "");
Expect(0, 120771, '\p{^	 BIDI_mirrored}', "");
Expect(0, 120771, '\P{	 BIDI_mirrored}', "");
Expect(1, 120771, '\P{^	 BIDI_mirrored}', "");
Expect(0, 120772, '\p{	 BIDI_mirrored}', "");
Expect(1, 120772, '\p{^	 BIDI_mirrored}', "");
Expect(1, 120772, '\P{	 BIDI_mirrored}', "");
Expect(0, 120772, '\P{^	 BIDI_mirrored}', "");
Error('\p{:=Is_BIDI_Mirrored}');
Error('\P{:=Is_BIDI_Mirrored}');
Expect(1, 120771, '\p{isbidimirrored}', "");
Expect(0, 120771, '\p{^isbidimirrored}', "");
Expect(0, 120771, '\P{isbidimirrored}', "");
Expect(1, 120771, '\P{^isbidimirrored}', "");
Expect(0, 120772, '\p{isbidimirrored}', "");
Expect(1, 120772, '\p{^isbidimirrored}', "");
Expect(1, 120772, '\P{isbidimirrored}', "");
Expect(0, 120772, '\P{^isbidimirrored}', "");
Expect(1, 120771, '\p{	is_bidi_MIRRORED}', "");
Expect(0, 120771, '\p{^	is_bidi_MIRRORED}', "");
Expect(0, 120771, '\P{	is_bidi_MIRRORED}', "");
Expect(1, 120771, '\P{^	is_bidi_MIRRORED}', "");
Expect(0, 120772, '\p{	is_bidi_MIRRORED}', "");
Expect(1, 120772, '\p{^	is_bidi_MIRRORED}', "");
Expect(1, 120772, '\P{	is_bidi_MIRRORED}', "");
Expect(0, 120772, '\P{^	is_bidi_MIRRORED}', "");
Error('\p{ :=BIDI_M}');
Error('\P{ :=BIDI_M}');
Expect(1, 120771, '\p{bidim}', "");
Expect(0, 120771, '\p{^bidim}', "");
Expect(0, 120771, '\P{bidim}', "");
Expect(1, 120771, '\P{^bidim}', "");
Expect(0, 120772, '\p{bidim}', "");
Expect(1, 120772, '\p{^bidim}', "");
Expect(1, 120772, '\P{bidim}', "");
Expect(0, 120772, '\P{^bidim}', "");
Expect(1, 120771, '\p{-Bidi_M}', "");
Expect(0, 120771, '\p{^-Bidi_M}', "");
Expect(0, 120771, '\P{-Bidi_M}', "");
Expect(1, 120771, '\P{^-Bidi_M}', "");
Expect(0, 120772, '\p{-Bidi_M}', "");
Expect(1, 120772, '\p{^-Bidi_M}', "");
Expect(1, 120772, '\P{-Bidi_M}', "");
Expect(0, 120772, '\P{^-Bidi_M}', "");
Error('\p{ IS_Bidi_m/a/}');
Error('\P{ IS_Bidi_m/a/}');
Expect(1, 120771, '\p{isbidim}', "");
Expect(0, 120771, '\p{^isbidim}', "");
Expect(0, 120771, '\P{isbidim}', "");
Expect(1, 120771, '\P{^isbidim}', "");
Expect(0, 120772, '\p{isbidim}', "");
Expect(1, 120772, '\p{^isbidim}', "");
Expect(1, 120772, '\P{isbidim}', "");
Expect(0, 120772, '\P{^isbidim}', "");
Expect(1, 120771, '\p{_Is_BIDI_m}', "");
Expect(0, 120771, '\p{^_Is_BIDI_m}', "");
Expect(0, 120771, '\P{_Is_BIDI_m}', "");
Expect(1, 120771, '\P{^_Is_BIDI_m}', "");
Expect(0, 120772, '\p{_Is_BIDI_m}', "");
Expect(1, 120772, '\p{^_Is_BIDI_m}', "");
Expect(1, 120772, '\P{_Is_BIDI_m}', "");
Expect(0, 120772, '\P{^_Is_BIDI_m}', "");
Error('\p{ 	Blank:=}');
Error('\P{ 	Blank:=}');
Expect(1, 12288, '\p{blank}', "");
Expect(0, 12288, '\p{^blank}', "");
Expect(0, 12288, '\P{blank}', "");
Expect(1, 12288, '\P{^blank}', "");
Expect(0, 12289, '\p{blank}', "");
Expect(1, 12289, '\p{^blank}', "");
Expect(1, 12289, '\P{blank}', "");
Expect(0, 12289, '\P{^blank}', "");
Expect(1, 12288, '\p{	Blank}', "");
Expect(0, 12288, '\p{^	Blank}', "");
Expect(0, 12288, '\P{	Blank}', "");
Expect(1, 12288, '\P{^	Blank}', "");
Expect(0, 12289, '\p{	Blank}', "");
Expect(1, 12289, '\p{^	Blank}', "");
Expect(1, 12289, '\P{	Blank}', "");
Expect(0, 12289, '\P{^	Blank}', "");
Error('\p{/a/ 	HORIZSPACE}');
Error('\P{/a/ 	HORIZSPACE}');
Expect(1, 12288, '\p{horizspace}', "");
Expect(0, 12288, '\p{^horizspace}', "");
Expect(0, 12288, '\P{horizspace}', "");
Expect(1, 12288, '\P{^horizspace}', "");
Expect(0, 12289, '\p{horizspace}', "");
Expect(1, 12289, '\p{^horizspace}', "");
Expect(1, 12289, '\P{horizspace}', "");
Expect(0, 12289, '\P{^horizspace}', "");
Expect(1, 12288, '\p{	_HORIZSPACE}', "");
Expect(0, 12288, '\p{^	_HORIZSPACE}', "");
Expect(0, 12288, '\P{	_HORIZSPACE}', "");
Expect(1, 12288, '\P{^	_HORIZSPACE}', "");
Expect(0, 12289, '\p{	_HORIZSPACE}', "");
Expect(1, 12289, '\p{^	_HORIZSPACE}', "");
Expect(1, 12289, '\P{	_HORIZSPACE}', "");
Expect(0, 12289, '\P{^	_HORIZSPACE}', "");
Error('\p{:=	XPOSIXBLANK}');
Error('\P{:=	XPOSIXBLANK}');
Expect(1, 12288, '\p{xposixblank}', "");
Expect(0, 12288, '\p{^xposixblank}', "");
Expect(0, 12288, '\P{xposixblank}', "");
Expect(1, 12288, '\P{^xposixblank}', "");
Expect(0, 12289, '\p{xposixblank}', "");
Expect(1, 12289, '\p{^xposixblank}', "");
Expect(1, 12289, '\P{xposixblank}', "");
Expect(0, 12289, '\P{^xposixblank}', "");
Expect(1, 12288, '\p{	XPOSIXBLANK}', "");
Expect(0, 12288, '\p{^	XPOSIXBLANK}', "");
Expect(0, 12288, '\P{	XPOSIXBLANK}', "");
Expect(1, 12288, '\P{^	XPOSIXBLANK}', "");
Expect(0, 12289, '\p{	XPOSIXBLANK}', "");
Expect(1, 12289, '\p{^	XPOSIXBLANK}', "");
Expect(1, 12289, '\P{	XPOSIXBLANK}', "");
Expect(0, 12289, '\P{^	XPOSIXBLANK}', "");
Error('\p{	_Is_blank/a/}');
Error('\P{	_Is_blank/a/}');
Expect(1, 12288, '\p{isblank}', "");
Expect(0, 12288, '\p{^isblank}', "");
Expect(0, 12288, '\P{isblank}', "");
Expect(1, 12288, '\P{^isblank}', "");
Expect(0, 12289, '\p{isblank}', "");
Expect(1, 12289, '\p{^isblank}', "");
Expect(1, 12289, '\P{isblank}', "");
Expect(0, 12289, '\P{^isblank}', "");
Expect(1, 12288, '\p{_	Is_blank}', "");
Expect(0, 12288, '\p{^_	Is_blank}', "");
Expect(0, 12288, '\P{_	Is_blank}', "");
Expect(1, 12288, '\P{^_	Is_blank}', "");
Expect(0, 12289, '\p{_	Is_blank}', "");
Expect(1, 12289, '\p{^_	Is_blank}', "");
Expect(1, 12289, '\P{_	Is_blank}', "");
Expect(0, 12289, '\P{^_	Is_blank}', "");
Error('\p{- Is_HorizSpace/a/}');
Error('\P{- Is_HorizSpace/a/}');
Expect(1, 12288, '\p{ishorizspace}', "");
Expect(0, 12288, '\p{^ishorizspace}', "");
Expect(0, 12288, '\P{ishorizspace}', "");
Expect(1, 12288, '\P{^ishorizspace}', "");
Expect(0, 12289, '\p{ishorizspace}', "");
Expect(1, 12289, '\p{^ishorizspace}', "");
Expect(1, 12289, '\P{ishorizspace}', "");
Expect(0, 12289, '\P{^ishorizspace}', "");
Expect(1, 12288, '\p{_ is_horizspace}', "");
Expect(0, 12288, '\p{^_ is_horizspace}', "");
Expect(0, 12288, '\P{_ is_horizspace}', "");
Expect(1, 12288, '\P{^_ is_horizspace}', "");
Expect(0, 12289, '\p{_ is_horizspace}', "");
Expect(1, 12289, '\p{^_ is_horizspace}', "");
Expect(1, 12289, '\P{_ is_horizspace}', "");
Expect(0, 12289, '\P{^_ is_horizspace}', "");
Error('\p{/a/ Is_XPosixBlank}');
Error('\P{/a/ Is_XPosixBlank}');
Expect(1, 12288, '\p{isxposixblank}', "");
Expect(0, 12288, '\p{^isxposixblank}', "");
Expect(0, 12288, '\P{isxposixblank}', "");
Expect(1, 12288, '\P{^isxposixblank}', "");
Expect(0, 12289, '\p{isxposixblank}', "");
Expect(1, 12289, '\p{^isxposixblank}', "");
Expect(1, 12289, '\P{isxposixblank}', "");
Expect(0, 12289, '\P{^isxposixblank}', "");
Expect(1, 12288, '\p{-Is_xposixblank}', "");
Expect(0, 12288, '\p{^-Is_xposixblank}', "");
Expect(0, 12288, '\P{-Is_xposixblank}', "");
Expect(1, 12288, '\P{^-Is_xposixblank}', "");
Expect(0, 12289, '\p{-Is_xposixblank}', "");
Expect(1, 12289, '\p{^-Is_xposixblank}', "");
Expect(1, 12289, '\P{-Is_xposixblank}', "");
Expect(0, 12289, '\P{^-Is_xposixblank}', "");
Error('\p{:=BLOCK_ELEMENTS}');
Error('\P{:=BLOCK_ELEMENTS}');
Expect(1, 9631, '\p{blockelements}', "");
Expect(0, 9631, '\p{^blockelements}', "");
Expect(0, 9631, '\P{blockelements}', "");
Expect(1, 9631, '\P{^blockelements}', "");
Expect(0, 9632, '\p{blockelements}', "");
Expect(1, 9632, '\p{^blockelements}', "");
Expect(1, 9632, '\P{blockelements}', "");
Expect(0, 9632, '\P{^blockelements}', "");
Expect(1, 9631, '\p{--block_Elements}', "");
Expect(0, 9631, '\p{^--block_Elements}', "");
Expect(0, 9631, '\P{--block_Elements}', "");
Expect(1, 9631, '\P{^--block_Elements}', "");
Expect(0, 9632, '\p{--block_Elements}', "");
Expect(1, 9632, '\p{^--block_Elements}', "");
Expect(1, 9632, '\P{--block_Elements}', "");
Expect(0, 9632, '\P{^--block_Elements}', "");
Error('\p{	/a/is_Block_elements}');
Error('\P{	/a/is_Block_elements}');
Expect(1, 9631, '\p{isblockelements}', "");
Expect(0, 9631, '\p{^isblockelements}', "");
Expect(0, 9631, '\P{isblockelements}', "");
Expect(1, 9631, '\P{^isblockelements}', "");
Expect(0, 9632, '\p{isblockelements}', "");
Expect(1, 9632, '\p{^isblockelements}', "");
Expect(1, 9632, '\P{isblockelements}', "");
Expect(0, 9632, '\P{^isblockelements}', "");
Expect(1, 9631, '\p{  Is_BLOCK_Elements}', "");
Expect(0, 9631, '\p{^  Is_BLOCK_Elements}', "");
Expect(0, 9631, '\P{  Is_BLOCK_Elements}', "");
Expect(1, 9631, '\P{^  Is_BLOCK_Elements}', "");
Expect(0, 9632, '\p{  Is_BLOCK_Elements}', "");
Expect(1, 9632, '\p{^  Is_BLOCK_Elements}', "");
Expect(1, 9632, '\P{  Is_BLOCK_Elements}', "");
Expect(0, 9632, '\P{^  Is_BLOCK_Elements}', "");
Error('\p{_In_BLOCK_ELEMENTS:=}');
Error('\P{_In_BLOCK_ELEMENTS:=}');
Expect(1, 9631, '\p{inblockelements}', "");
Expect(0, 9631, '\p{^inblockelements}', "");
Expect(0, 9631, '\P{inblockelements}', "");
Expect(1, 9631, '\P{^inblockelements}', "");
Expect(0, 9632, '\p{inblockelements}', "");
Expect(1, 9632, '\p{^inblockelements}', "");
Expect(1, 9632, '\P{inblockelements}', "");
Expect(0, 9632, '\P{^inblockelements}', "");
Expect(1, 9631, '\p{  IN_BLOCK_Elements}', "");
Expect(0, 9631, '\p{^  IN_BLOCK_Elements}', "");
Expect(0, 9631, '\P{  IN_BLOCK_Elements}', "");
Expect(1, 9631, '\P{^  IN_BLOCK_Elements}', "");
Expect(0, 9632, '\p{  IN_BLOCK_Elements}', "");
Expect(1, 9632, '\p{^  IN_BLOCK_Elements}', "");
Expect(1, 9632, '\P{  IN_BLOCK_Elements}', "");
Expect(0, 9632, '\P{^  IN_BLOCK_Elements}', "");
Error('\p{/a/__BOPOMOFO}');
Error('\P{/a/__BOPOMOFO}');
Expect(1, 12730, '\p{bopomofo}', "");
Expect(0, 12730, '\p{^bopomofo}', "");
Expect(0, 12730, '\P{bopomofo}', "");
Expect(1, 12730, '\P{^bopomofo}', "");
Expect(0, 12731, '\p{bopomofo}', "");
Expect(1, 12731, '\p{^bopomofo}', "");
Expect(1, 12731, '\P{bopomofo}', "");
Expect(0, 12731, '\P{^bopomofo}', "");
Expect(1, 12730, '\p{-Bopomofo}', "");
Expect(0, 12730, '\p{^-Bopomofo}', "");
Expect(0, 12730, '\P{-Bopomofo}', "");
Expect(1, 12730, '\P{^-Bopomofo}', "");
Expect(0, 12731, '\p{-Bopomofo}', "");
Expect(1, 12731, '\p{^-Bopomofo}', "");
Expect(1, 12731, '\P{-Bopomofo}', "");
Expect(0, 12731, '\P{^-Bopomofo}', "");
Error('\p{-is_bopomofo:=}');
Error('\P{-is_bopomofo:=}');
Expect(1, 12730, '\p{isbopomofo}', "");
Expect(0, 12730, '\p{^isbopomofo}', "");
Expect(0, 12730, '\P{isbopomofo}', "");
Expect(1, 12730, '\P{^isbopomofo}', "");
Expect(0, 12731, '\p{isbopomofo}', "");
Expect(1, 12731, '\p{^isbopomofo}', "");
Expect(1, 12731, '\P{isbopomofo}', "");
Expect(0, 12731, '\P{^isbopomofo}', "");
Expect(1, 12730, '\p{_Is_Bopomofo}', "");
Expect(0, 12730, '\p{^_Is_Bopomofo}', "");
Expect(0, 12730, '\P{_Is_Bopomofo}', "");
Expect(1, 12730, '\P{^_Is_Bopomofo}', "");
Expect(0, 12731, '\p{_Is_Bopomofo}', "");
Expect(1, 12731, '\p{^_Is_Bopomofo}', "");
Expect(1, 12731, '\P{_Is_Bopomofo}', "");
Expect(0, 12731, '\P{^_Is_Bopomofo}', "");
Error('\p{:=_ Bopo}');
Error('\P{:=_ Bopo}');
Expect(1, 12730, '\p{bopo}', "");
Expect(0, 12730, '\p{^bopo}', "");
Expect(0, 12730, '\P{bopo}', "");
Expect(1, 12730, '\P{^bopo}', "");
Expect(0, 12731, '\p{bopo}', "");
Expect(1, 12731, '\p{^bopo}', "");
Expect(1, 12731, '\P{bopo}', "");
Expect(0, 12731, '\P{^bopo}', "");
Expect(1, 12730, '\p{ -bopo}', "");
Expect(0, 12730, '\p{^ -bopo}', "");
Expect(0, 12730, '\P{ -bopo}', "");
Expect(1, 12730, '\P{^ -bopo}', "");
Expect(0, 12731, '\p{ -bopo}', "");
Expect(1, 12731, '\p{^ -bopo}', "");
Expect(1, 12731, '\P{ -bopo}', "");
Expect(0, 12731, '\P{^ -bopo}', "");
Error('\p{:=IS_BOPO}');
Error('\P{:=IS_BOPO}');
Expect(1, 12730, '\p{isbopo}', "");
Expect(0, 12730, '\p{^isbopo}', "");
Expect(0, 12730, '\P{isbopo}', "");
Expect(1, 12730, '\P{^isbopo}', "");
Expect(0, 12731, '\p{isbopo}', "");
Expect(1, 12731, '\p{^isbopo}', "");
Expect(1, 12731, '\P{isbopo}', "");
Expect(0, 12731, '\P{^isbopo}', "");
Expect(1, 12730, '\p{-	is_BOPO}', "");
Expect(0, 12730, '\p{^-	is_BOPO}', "");
Expect(0, 12730, '\P{-	is_BOPO}', "");
Expect(1, 12730, '\P{^-	is_BOPO}', "");
Expect(0, 12731, '\p{-	is_BOPO}', "");
Expect(1, 12731, '\p{^-	is_BOPO}', "");
Expect(1, 12731, '\P{-	is_BOPO}', "");
Expect(0, 12731, '\P{^-	is_BOPO}', "");
Error('\p{/a/		bopomofo_extended}');
Error('\P{/a/		bopomofo_extended}');
Expect(1, 12735, '\p{bopomofoextended}', "");
Expect(0, 12735, '\p{^bopomofoextended}', "");
Expect(0, 12735, '\P{bopomofoextended}', "");
Expect(1, 12735, '\P{^bopomofoextended}', "");
Expect(0, 12736, '\p{bopomofoextended}', "");
Expect(1, 12736, '\p{^bopomofoextended}', "");
Expect(1, 12736, '\P{bopomofoextended}', "");
Expect(0, 12736, '\P{^bopomofoextended}', "");
Expect(1, 12735, '\p{ 	BOPOMOFO_Extended}', "");
Expect(0, 12735, '\p{^ 	BOPOMOFO_Extended}', "");
Expect(0, 12735, '\P{ 	BOPOMOFO_Extended}', "");
Expect(1, 12735, '\P{^ 	BOPOMOFO_Extended}', "");
Expect(0, 12736, '\p{ 	BOPOMOFO_Extended}', "");
Expect(1, 12736, '\p{^ 	BOPOMOFO_Extended}', "");
Expect(1, 12736, '\P{ 	BOPOMOFO_Extended}', "");
Expect(0, 12736, '\P{^ 	BOPOMOFO_Extended}', "");
Error('\p{:=	IS_Bopomofo_EXTENDED}');
Error('\P{:=	IS_Bopomofo_EXTENDED}');
Expect(1, 12735, '\p{isbopomofoextended}', "");
Expect(0, 12735, '\p{^isbopomofoextended}', "");
Expect(0, 12735, '\P{isbopomofoextended}', "");
Expect(1, 12735, '\P{^isbopomofoextended}', "");
Expect(0, 12736, '\p{isbopomofoextended}', "");
Expect(1, 12736, '\p{^isbopomofoextended}', "");
Expect(1, 12736, '\P{isbopomofoextended}', "");
Expect(0, 12736, '\P{^isbopomofoextended}', "");
Expect(1, 12735, '\p{ _Is_Bopomofo_Extended}', "");
Expect(0, 12735, '\p{^ _Is_Bopomofo_Extended}', "");
Expect(0, 12735, '\P{ _Is_Bopomofo_Extended}', "");
Expect(1, 12735, '\P{^ _Is_Bopomofo_Extended}', "");
Expect(0, 12736, '\p{ _Is_Bopomofo_Extended}', "");
Expect(1, 12736, '\p{^ _Is_Bopomofo_Extended}', "");
Expect(1, 12736, '\P{ _Is_Bopomofo_Extended}', "");
Expect(0, 12736, '\P{^ _Is_Bopomofo_Extended}', "");
Error('\p{:= -IN_Bopomofo_Extended}');
Error('\P{:= -IN_Bopomofo_Extended}');
Expect(1, 12735, '\p{inbopomofoextended}', "");
Expect(0, 12735, '\p{^inbopomofoextended}', "");
Expect(0, 12735, '\P{inbopomofoextended}', "");
Expect(1, 12735, '\P{^inbopomofoextended}', "");
Expect(0, 12736, '\p{inbopomofoextended}', "");
Expect(1, 12736, '\p{^inbopomofoextended}', "");
Expect(1, 12736, '\P{inbopomofoextended}', "");
Expect(0, 12736, '\P{^inbopomofoextended}', "");
Expect(1, 12735, '\p{_ In_bopomofo_extended}', "");
Expect(0, 12735, '\p{^_ In_bopomofo_extended}', "");
Expect(0, 12735, '\P{_ In_bopomofo_extended}', "");
Expect(1, 12735, '\P{^_ In_bopomofo_extended}', "");
Expect(0, 12736, '\p{_ In_bopomofo_extended}', "");
Expect(1, 12736, '\p{^_ In_bopomofo_extended}', "");
Expect(1, 12736, '\P{_ In_bopomofo_extended}', "");
Expect(0, 12736, '\P{^_ In_bopomofo_extended}', "");
Error('\p{:=-_bopomofo_EXT}');
Error('\P{:=-_bopomofo_EXT}');
Expect(1, 12735, '\p{bopomofoext}', "");
Expect(0, 12735, '\p{^bopomofoext}', "");
Expect(0, 12735, '\P{bopomofoext}', "");
Expect(1, 12735, '\P{^bopomofoext}', "");
Expect(0, 12736, '\p{bopomofoext}', "");
Expect(1, 12736, '\p{^bopomofoext}', "");
Expect(1, 12736, '\P{bopomofoext}', "");
Expect(0, 12736, '\P{^bopomofoext}', "");
Expect(1, 12735, '\p{	-bopomofo_EXT}', "");
Expect(0, 12735, '\p{^	-bopomofo_EXT}', "");
Expect(0, 12735, '\P{	-bopomofo_EXT}', "");
Expect(1, 12735, '\P{^	-bopomofo_EXT}', "");
Expect(0, 12736, '\p{	-bopomofo_EXT}', "");
Expect(1, 12736, '\p{^	-bopomofo_EXT}', "");
Expect(1, 12736, '\P{	-bopomofo_EXT}', "");
Expect(0, 12736, '\P{^	-bopomofo_EXT}', "");
Error('\p{	-IS_BOPOMOFO_Ext/a/}');
Error('\P{	-IS_BOPOMOFO_Ext/a/}');
Expect(1, 12735, '\p{isbopomofoext}', "");
Expect(0, 12735, '\p{^isbopomofoext}', "");
Expect(0, 12735, '\P{isbopomofoext}', "");
Expect(1, 12735, '\P{^isbopomofoext}', "");
Expect(0, 12736, '\p{isbopomofoext}', "");
Expect(1, 12736, '\p{^isbopomofoext}', "");
Expect(1, 12736, '\P{isbopomofoext}', "");
Expect(0, 12736, '\P{^isbopomofoext}', "");
Expect(1, 12735, '\p{	 is_Bopomofo_Ext}', "");
Expect(0, 12735, '\p{^	 is_Bopomofo_Ext}', "");
Expect(0, 12735, '\P{	 is_Bopomofo_Ext}', "");
Expect(1, 12735, '\P{^	 is_Bopomofo_Ext}', "");
Expect(0, 12736, '\p{	 is_Bopomofo_Ext}', "");
Expect(1, 12736, '\p{^	 is_Bopomofo_Ext}', "");
Expect(1, 12736, '\P{	 is_Bopomofo_Ext}', "");
Expect(0, 12736, '\P{^	 is_Bopomofo_Ext}', "");
Error('\p{	:=In_bopomofo_ext}');
Error('\P{	:=In_bopomofo_ext}');
Expect(1, 12735, '\p{inbopomofoext}', "");
Expect(0, 12735, '\p{^inbopomofoext}', "");
Expect(0, 12735, '\P{inbopomofoext}', "");
Expect(1, 12735, '\P{^inbopomofoext}', "");
Expect(0, 12736, '\p{inbopomofoext}', "");
Expect(1, 12736, '\p{^inbopomofoext}', "");
Expect(1, 12736, '\P{inbopomofoext}', "");
Expect(0, 12736, '\P{^inbopomofoext}', "");
Expect(1, 12735, '\p{_In_Bopomofo_Ext}', "");
Expect(0, 12735, '\p{^_In_Bopomofo_Ext}', "");
Expect(0, 12735, '\P{_In_Bopomofo_Ext}', "");
Expect(1, 12735, '\P{^_In_Bopomofo_Ext}', "");
Expect(0, 12736, '\p{_In_Bopomofo_Ext}', "");
Expect(1, 12736, '\p{^_In_Bopomofo_Ext}', "");
Expect(1, 12736, '\P{_In_Bopomofo_Ext}', "");
Expect(0, 12736, '\P{^_In_Bopomofo_Ext}', "");
Error('\p{-/a/Box_Drawing}');
Error('\P{-/a/Box_Drawing}');
Expect(1, 9599, '\p{boxdrawing}', "");
Expect(0, 9599, '\p{^boxdrawing}', "");
Expect(0, 9599, '\P{boxdrawing}', "");
Expect(1, 9599, '\P{^boxdrawing}', "");
Expect(0, 9600, '\p{boxdrawing}', "");
Expect(1, 9600, '\p{^boxdrawing}', "");
Expect(1, 9600, '\P{boxdrawing}', "");
Expect(0, 9600, '\P{^boxdrawing}', "");
Expect(1, 9599, '\p{_-Box_drawing}', "");
Expect(0, 9599, '\p{^_-Box_drawing}', "");
Expect(0, 9599, '\P{_-Box_drawing}', "");
Expect(1, 9599, '\P{^_-Box_drawing}', "");
Expect(0, 9600, '\p{_-Box_drawing}', "");
Expect(1, 9600, '\p{^_-Box_drawing}', "");
Expect(1, 9600, '\P{_-Box_drawing}', "");
Expect(0, 9600, '\P{^_-Box_drawing}', "");
Error('\p{	/a/is_Box_Drawing}');
Error('\P{	/a/is_Box_Drawing}');
Expect(1, 9599, '\p{isboxdrawing}', "");
Expect(0, 9599, '\p{^isboxdrawing}', "");
Expect(0, 9599, '\P{isboxdrawing}', "");
Expect(1, 9599, '\P{^isboxdrawing}', "");
Expect(0, 9600, '\p{isboxdrawing}', "");
Expect(1, 9600, '\p{^isboxdrawing}', "");
Expect(1, 9600, '\P{isboxdrawing}', "");
Expect(0, 9600, '\P{^isboxdrawing}', "");
Expect(1, 9599, '\p{-IS_Box_Drawing}', "");
Expect(0, 9599, '\p{^-IS_Box_Drawing}', "");
Expect(0, 9599, '\P{-IS_Box_Drawing}', "");
Expect(1, 9599, '\P{^-IS_Box_Drawing}', "");
Expect(0, 9600, '\p{-IS_Box_Drawing}', "");
Expect(1, 9600, '\p{^-IS_Box_Drawing}', "");
Expect(1, 9600, '\P{-IS_Box_Drawing}', "");
Expect(0, 9600, '\P{^-IS_Box_Drawing}', "");
Error('\p{:=	 in_Box_drawing}');
Error('\P{:=	 in_Box_drawing}');
Expect(1, 9599, '\p{inboxdrawing}', "");
Expect(0, 9599, '\p{^inboxdrawing}', "");
Expect(0, 9599, '\P{inboxdrawing}', "");
Expect(1, 9599, '\P{^inboxdrawing}', "");
Expect(0, 9600, '\p{inboxdrawing}', "");
Expect(1, 9600, '\p{^inboxdrawing}', "");
Expect(1, 9600, '\P{inboxdrawing}', "");
Expect(0, 9600, '\P{^inboxdrawing}', "");
Expect(1, 9599, '\p{ -In_Box_drawing}', "");
Expect(0, 9599, '\p{^ -In_Box_drawing}', "");
Expect(0, 9599, '\P{ -In_Box_drawing}', "");
Expect(1, 9599, '\P{^ -In_Box_drawing}', "");
Expect(0, 9600, '\p{ -In_Box_drawing}', "");
Expect(1, 9600, '\p{^ -In_Box_drawing}', "");
Expect(1, 9600, '\P{ -In_Box_drawing}', "");
Expect(0, 9600, '\P{^ -In_Box_drawing}', "");
Error('\p{:=Brahmi}');
Error('\P{:=Brahmi}');
Expect(1, 69743, '\p{brahmi}', "");
Expect(0, 69743, '\p{^brahmi}', "");
Expect(0, 69743, '\P{brahmi}', "");
Expect(1, 69743, '\P{^brahmi}', "");
Expect(0, 69744, '\p{brahmi}', "");
Expect(1, 69744, '\p{^brahmi}', "");
Expect(1, 69744, '\P{brahmi}', "");
Expect(0, 69744, '\P{^brahmi}', "");
Expect(1, 69743, '\p{ _Brahmi}', "");
Expect(0, 69743, '\p{^ _Brahmi}', "");
Expect(0, 69743, '\P{ _Brahmi}', "");
Expect(1, 69743, '\P{^ _Brahmi}', "");
Expect(0, 69744, '\p{ _Brahmi}', "");
Expect(1, 69744, '\p{^ _Brahmi}', "");
Expect(1, 69744, '\P{ _Brahmi}', "");
Expect(0, 69744, '\P{^ _Brahmi}', "");
Error('\p{		IS_Brahmi/a/}');
Error('\P{		IS_Brahmi/a/}');
Expect(1, 69743, '\p{isbrahmi}', "");
Expect(0, 69743, '\p{^isbrahmi}', "");
Expect(0, 69743, '\P{isbrahmi}', "");
Expect(1, 69743, '\P{^isbrahmi}', "");
Expect(0, 69744, '\p{isbrahmi}', "");
Expect(1, 69744, '\p{^isbrahmi}', "");
Expect(1, 69744, '\P{isbrahmi}', "");
Expect(0, 69744, '\P{^isbrahmi}', "");
Expect(1, 69743, '\p{ _Is_BRAHMI}', "");
Expect(0, 69743, '\p{^ _Is_BRAHMI}', "");
Expect(0, 69743, '\P{ _Is_BRAHMI}', "");
Expect(1, 69743, '\P{^ _Is_BRAHMI}', "");
Expect(0, 69744, '\p{ _Is_BRAHMI}', "");
Expect(1, 69744, '\p{^ _Is_BRAHMI}', "");
Expect(1, 69744, '\P{ _Is_BRAHMI}', "");
Expect(0, 69744, '\P{^ _Is_BRAHMI}', "");
Error('\p{-/a/Brah}');
Error('\P{-/a/Brah}');
Expect(1, 69743, '\p{brah}', "");
Expect(0, 69743, '\p{^brah}', "");
Expect(0, 69743, '\P{brah}', "");
Expect(1, 69743, '\P{^brah}', "");
Expect(0, 69744, '\p{brah}', "");
Expect(1, 69744, '\p{^brah}', "");
Expect(1, 69744, '\P{brah}', "");
Expect(0, 69744, '\P{^brah}', "");
Expect(1, 69743, '\p{	_Brah}', "");
Expect(0, 69743, '\p{^	_Brah}', "");
Expect(0, 69743, '\P{	_Brah}', "");
Expect(1, 69743, '\P{^	_Brah}', "");
Expect(0, 69744, '\p{	_Brah}', "");
Expect(1, 69744, '\p{^	_Brah}', "");
Expect(1, 69744, '\P{	_Brah}', "");
Expect(0, 69744, '\P{^	_Brah}', "");
Error('\p{/a/_ IS_brah}');
Error('\P{/a/_ IS_brah}');
Expect(1, 69743, '\p{isbrah}', "");
Expect(0, 69743, '\p{^isbrah}', "");
Expect(0, 69743, '\P{isbrah}', "");
Expect(1, 69743, '\P{^isbrah}', "");
Expect(0, 69744, '\p{isbrah}', "");
Expect(1, 69744, '\p{^isbrah}', "");
Expect(1, 69744, '\P{isbrah}', "");
Expect(0, 69744, '\P{^isbrah}', "");
Expect(1, 69743, '\p{_	IS_Brah}', "");
Expect(0, 69743, '\p{^_	IS_Brah}', "");
Expect(0, 69743, '\P{_	IS_Brah}', "");
Expect(1, 69743, '\P{^_	IS_Brah}', "");
Expect(0, 69744, '\p{_	IS_Brah}', "");
Expect(1, 69744, '\p{^_	IS_Brah}', "");
Expect(1, 69744, '\P{_	IS_Brah}', "");
Expect(0, 69744, '\P{^_	IS_Brah}', "");
Error('\p{:=_braille}');
Error('\P{:=_braille}');
Expect(1, 10495, '\p{braille}', "");
Expect(0, 10495, '\p{^braille}', "");
Expect(0, 10495, '\P{braille}', "");
Expect(1, 10495, '\P{^braille}', "");
Expect(0, 10496, '\p{braille}', "");
Expect(1, 10496, '\p{^braille}', "");
Expect(1, 10496, '\P{braille}', "");
Expect(0, 10496, '\P{^braille}', "");
Expect(1, 10495, '\p{_Braille}', "");
Expect(0, 10495, '\p{^_Braille}', "");
Expect(0, 10495, '\P{_Braille}', "");
Expect(1, 10495, '\P{^_Braille}', "");
Expect(0, 10496, '\p{_Braille}', "");
Expect(1, 10496, '\p{^_Braille}', "");
Expect(1, 10496, '\P{_Braille}', "");
Expect(0, 10496, '\P{^_Braille}', "");
Error('\p{ -Is_braille/a/}');
Error('\P{ -Is_braille/a/}');
Expect(1, 10495, '\p{isbraille}', "");
Expect(0, 10495, '\p{^isbraille}', "");
Expect(0, 10495, '\P{isbraille}', "");
Expect(1, 10495, '\P{^isbraille}', "");
Expect(0, 10496, '\p{isbraille}', "");
Expect(1, 10496, '\p{^isbraille}', "");
Expect(1, 10496, '\P{isbraille}', "");
Expect(0, 10496, '\P{^isbraille}', "");
Expect(1, 10495, '\p{ Is_Braille}', "");
Expect(0, 10495, '\p{^ Is_Braille}', "");
Expect(0, 10495, '\P{ Is_Braille}', "");
Expect(1, 10495, '\P{^ Is_Braille}', "");
Expect(0, 10496, '\p{ Is_Braille}', "");
Expect(1, 10496, '\p{^ Is_Braille}', "");
Expect(1, 10496, '\P{ Is_Braille}', "");
Expect(0, 10496, '\P{^ Is_Braille}', "");
Error('\p{/a/_-brai}');
Error('\P{/a/_-brai}');
Expect(1, 10495, '\p{brai}', "");
Expect(0, 10495, '\p{^brai}', "");
Expect(0, 10495, '\P{brai}', "");
Expect(1, 10495, '\P{^brai}', "");
Expect(0, 10496, '\p{brai}', "");
Expect(1, 10496, '\p{^brai}', "");
Expect(1, 10496, '\P{brai}', "");
Expect(0, 10496, '\P{^brai}', "");
Expect(1, 10495, '\p{--Brai}', "");
Expect(0, 10495, '\p{^--Brai}', "");
Expect(0, 10495, '\P{--Brai}', "");
Expect(1, 10495, '\P{^--Brai}', "");
Expect(0, 10496, '\p{--Brai}', "");
Expect(1, 10496, '\p{^--Brai}', "");
Expect(1, 10496, '\P{--Brai}', "");
Expect(0, 10496, '\P{^--Brai}', "");
Error('\p{:= IS_Brai}');
Error('\P{:= IS_Brai}');
Expect(1, 10495, '\p{isbrai}', "");
Expect(0, 10495, '\p{^isbrai}', "");
Expect(0, 10495, '\P{isbrai}', "");
Expect(1, 10495, '\P{^isbrai}', "");
Expect(0, 10496, '\p{isbrai}', "");
Expect(1, 10496, '\p{^isbrai}', "");
Expect(1, 10496, '\P{isbrai}', "");
Expect(0, 10496, '\P{^isbrai}', "");
Expect(1, 10495, '\p{-_Is_Brai}', "");
Expect(0, 10495, '\p{^-_Is_Brai}', "");
Expect(0, 10495, '\P{-_Is_Brai}', "");
Expect(1, 10495, '\P{^-_Is_Brai}', "");
Expect(0, 10496, '\p{-_Is_Brai}', "");
Expect(1, 10496, '\p{^-_Is_Brai}', "");
Expect(1, 10496, '\P{-_Is_Brai}', "");
Expect(0, 10496, '\P{^-_Is_Brai}', "");
Error('\p{_-braille_Patterns:=}');
Error('\P{_-braille_Patterns:=}');
Expect(1, 10495, '\p{braillepatterns}', "");
Expect(0, 10495, '\p{^braillepatterns}', "");
Expect(0, 10495, '\P{braillepatterns}', "");
Expect(1, 10495, '\P{^braillepatterns}', "");
Expect(0, 10496, '\p{braillepatterns}', "");
Expect(1, 10496, '\p{^braillepatterns}', "");
Expect(1, 10496, '\P{braillepatterns}', "");
Expect(0, 10496, '\P{^braillepatterns}', "");
Expect(1, 10495, '\p{-Braille_Patterns}', "");
Expect(0, 10495, '\p{^-Braille_Patterns}', "");
Expect(0, 10495, '\P{-Braille_Patterns}', "");
Expect(1, 10495, '\P{^-Braille_Patterns}', "");
Expect(0, 10496, '\p{-Braille_Patterns}', "");
Expect(1, 10496, '\p{^-Braille_Patterns}', "");
Expect(1, 10496, '\P{-Braille_Patterns}', "");
Expect(0, 10496, '\P{^-Braille_Patterns}', "");
Error('\p{ :=IS_Braille_patterns}');
Error('\P{ :=IS_Braille_patterns}');
Expect(1, 10495, '\p{isbraillepatterns}', "");
Expect(0, 10495, '\p{^isbraillepatterns}', "");
Expect(0, 10495, '\P{isbraillepatterns}', "");
Expect(1, 10495, '\P{^isbraillepatterns}', "");
Expect(0, 10496, '\p{isbraillepatterns}', "");
Expect(1, 10496, '\p{^isbraillepatterns}', "");
Expect(1, 10496, '\P{isbraillepatterns}', "");
Expect(0, 10496, '\P{^isbraillepatterns}', "");
Expect(1, 10495, '\p{_	Is_braille_PATTERNS}', "");
Expect(0, 10495, '\p{^_	Is_braille_PATTERNS}', "");
Expect(0, 10495, '\P{_	Is_braille_PATTERNS}', "");
Expect(1, 10495, '\P{^_	Is_braille_PATTERNS}', "");
Expect(0, 10496, '\p{_	Is_braille_PATTERNS}', "");
Expect(1, 10496, '\p{^_	Is_braille_PATTERNS}', "");
Expect(1, 10496, '\P{_	Is_braille_PATTERNS}', "");
Expect(0, 10496, '\P{^_	Is_braille_PATTERNS}', "");
Error('\p{:=--in_Braille_Patterns}');
Error('\P{:=--in_Braille_Patterns}');
Expect(1, 10495, '\p{inbraillepatterns}', "");
Expect(0, 10495, '\p{^inbraillepatterns}', "");
Expect(0, 10495, '\P{inbraillepatterns}', "");
Expect(1, 10495, '\P{^inbraillepatterns}', "");
Expect(0, 10496, '\p{inbraillepatterns}', "");
Expect(1, 10496, '\p{^inbraillepatterns}', "");
Expect(1, 10496, '\P{inbraillepatterns}', "");
Expect(0, 10496, '\P{^inbraillepatterns}', "");
Expect(1, 10495, '\p{ In_Braille_Patterns}', "");
Expect(0, 10495, '\p{^ In_Braille_Patterns}', "");
Expect(0, 10495, '\P{ In_Braille_Patterns}', "");
Expect(1, 10495, '\P{^ In_Braille_Patterns}', "");
Expect(0, 10496, '\p{ In_Braille_Patterns}', "");
Expect(1, 10496, '\p{^ In_Braille_Patterns}', "");
Expect(1, 10496, '\P{ In_Braille_Patterns}', "");
Expect(0, 10496, '\P{^ In_Braille_Patterns}', "");
Error('\p{/a/ _In_Braille}');
Error('\P{/a/ _In_Braille}');
Expect(1, 10495, '\p{inbraille}', "");
Expect(0, 10495, '\p{^inbraille}', "");
Expect(0, 10495, '\P{inbraille}', "");
Expect(1, 10495, '\P{^inbraille}', "");
Expect(0, 10496, '\p{inbraille}', "");
Expect(1, 10496, '\p{^inbraille}', "");
Expect(1, 10496, '\P{inbraille}', "");
Expect(0, 10496, '\P{^inbraille}', "");
Expect(1, 10495, '\p{_ in_Braille}', "");
Expect(0, 10495, '\p{^_ in_Braille}', "");
Expect(0, 10495, '\P{_ in_Braille}', "");
Expect(1, 10495, '\P{^_ in_Braille}', "");
Expect(0, 10496, '\p{_ in_Braille}', "");
Expect(1, 10496, '\p{^_ in_Braille}', "");
Expect(1, 10496, '\P{_ in_Braille}', "");
Expect(0, 10496, '\P{^_ in_Braille}', "");
Error('\p{	-BUGINESE/a/}');
Error('\P{	-BUGINESE/a/}');
Expect(1, 6687, '\p{buginese}', "");
Expect(0, 6687, '\p{^buginese}', "");
Expect(0, 6687, '\P{buginese}', "");
Expect(1, 6687, '\P{^buginese}', "");
Expect(0, 6688, '\p{buginese}', "");
Expect(1, 6688, '\p{^buginese}', "");
Expect(1, 6688, '\P{buginese}', "");
Expect(0, 6688, '\P{^buginese}', "");
Expect(1, 6687, '\p{-Buginese}', "");
Expect(0, 6687, '\p{^-Buginese}', "");
Expect(0, 6687, '\P{-Buginese}', "");
Expect(1, 6687, '\P{^-Buginese}', "");
Expect(0, 6688, '\p{-Buginese}', "");
Expect(1, 6688, '\p{^-Buginese}', "");
Expect(1, 6688, '\P{-Buginese}', "");
Expect(0, 6688, '\P{^-Buginese}', "");
Error('\p{-:=IS_Buginese}');
Error('\P{-:=IS_Buginese}');
Expect(1, 6687, '\p{isbuginese}', "");
Expect(0, 6687, '\p{^isbuginese}', "");
Expect(0, 6687, '\P{isbuginese}', "");
Expect(1, 6687, '\P{^isbuginese}', "");
Expect(0, 6688, '\p{isbuginese}', "");
Expect(1, 6688, '\p{^isbuginese}', "");
Expect(1, 6688, '\P{isbuginese}', "");
Expect(0, 6688, '\P{^isbuginese}', "");
Expect(1, 6687, '\p{_ IS_BUGINESE}', "");
Expect(0, 6687, '\p{^_ IS_BUGINESE}', "");
Expect(0, 6687, '\P{_ IS_BUGINESE}', "");
Expect(1, 6687, '\P{^_ IS_BUGINESE}', "");
Expect(0, 6688, '\p{_ IS_BUGINESE}', "");
Expect(1, 6688, '\p{^_ IS_BUGINESE}', "");
Expect(1, 6688, '\P{_ IS_BUGINESE}', "");
Expect(0, 6688, '\P{^_ IS_BUGINESE}', "");
Error('\p{-	bugi:=}');
Error('\P{-	bugi:=}');
Expect(1, 6687, '\p{bugi}', "");
Expect(0, 6687, '\p{^bugi}', "");
Expect(0, 6687, '\P{bugi}', "");
Expect(1, 6687, '\P{^bugi}', "");
Expect(0, 6688, '\p{bugi}', "");
Expect(1, 6688, '\p{^bugi}', "");
Expect(1, 6688, '\P{bugi}', "");
Expect(0, 6688, '\P{^bugi}', "");
Expect(1, 6687, '\p{- Bugi}', "");
Expect(0, 6687, '\p{^- Bugi}', "");
Expect(0, 6687, '\P{- Bugi}', "");
Expect(1, 6687, '\P{^- Bugi}', "");
Expect(0, 6688, '\p{- Bugi}', "");
Expect(1, 6688, '\p{^- Bugi}', "");
Expect(1, 6688, '\P{- Bugi}', "");
Expect(0, 6688, '\P{^- Bugi}', "");
Error('\p{	:=Is_Bugi}');
Error('\P{	:=Is_Bugi}');
Expect(1, 6687, '\p{isbugi}', "");
Expect(0, 6687, '\p{^isbugi}', "");
Expect(0, 6687, '\P{isbugi}', "");
Expect(1, 6687, '\P{^isbugi}', "");
Expect(0, 6688, '\p{isbugi}', "");
Expect(1, 6688, '\p{^isbugi}', "");
Expect(1, 6688, '\P{isbugi}', "");
Expect(0, 6688, '\P{^isbugi}', "");
Expect(1, 6687, '\p{_IS_Bugi}', "");
Expect(0, 6687, '\p{^_IS_Bugi}', "");
Expect(0, 6687, '\P{_IS_Bugi}', "");
Expect(1, 6687, '\P{^_IS_Bugi}', "");
Expect(0, 6688, '\p{_IS_Bugi}', "");
Expect(1, 6688, '\p{^_IS_Bugi}', "");
Expect(1, 6688, '\P{_IS_Bugi}', "");
Expect(0, 6688, '\P{^_IS_Bugi}', "");
Error('\p{  Buhid:=}');
Error('\P{  Buhid:=}');
Expect(1, 5971, '\p{buhid}', "");
Expect(0, 5971, '\p{^buhid}', "");
Expect(0, 5971, '\P{buhid}', "");
Expect(1, 5971, '\P{^buhid}', "");
Expect(0, 5972, '\p{buhid}', "");
Expect(1, 5972, '\p{^buhid}', "");
Expect(1, 5972, '\P{buhid}', "");
Expect(0, 5972, '\P{^buhid}', "");
Expect(1, 5971, '\p{ 	BUHID}', "");
Expect(0, 5971, '\p{^ 	BUHID}', "");
Expect(0, 5971, '\P{ 	BUHID}', "");
Expect(1, 5971, '\P{^ 	BUHID}', "");
Expect(0, 5972, '\p{ 	BUHID}', "");
Expect(1, 5972, '\p{^ 	BUHID}', "");
Expect(1, 5972, '\P{ 	BUHID}', "");
Expect(0, 5972, '\P{^ 	BUHID}', "");
Error('\p{		Is_BUHID/a/}');
Error('\P{		Is_BUHID/a/}');
Expect(1, 5971, '\p{isbuhid}', "");
Expect(0, 5971, '\p{^isbuhid}', "");
Expect(0, 5971, '\P{isbuhid}', "");
Expect(1, 5971, '\P{^isbuhid}', "");
Expect(0, 5972, '\p{isbuhid}', "");
Expect(1, 5972, '\p{^isbuhid}', "");
Expect(1, 5972, '\P{isbuhid}', "");
Expect(0, 5972, '\P{^isbuhid}', "");
Expect(1, 5971, '\p{ -IS_Buhid}', "");
Expect(0, 5971, '\p{^ -IS_Buhid}', "");
Expect(0, 5971, '\P{ -IS_Buhid}', "");
Expect(1, 5971, '\P{^ -IS_Buhid}', "");
Expect(0, 5972, '\p{ -IS_Buhid}', "");
Expect(1, 5972, '\p{^ -IS_Buhid}', "");
Expect(1, 5972, '\P{ -IS_Buhid}', "");
Expect(0, 5972, '\P{^ -IS_Buhid}', "");
Error('\p{/a/Buhd}');
Error('\P{/a/Buhd}');
Expect(1, 5971, '\p{buhd}', "");
Expect(0, 5971, '\p{^buhd}', "");
Expect(0, 5971, '\P{buhd}', "");
Expect(1, 5971, '\P{^buhd}', "");
Expect(0, 5972, '\p{buhd}', "");
Expect(1, 5972, '\p{^buhd}', "");
Expect(1, 5972, '\P{buhd}', "");
Expect(0, 5972, '\P{^buhd}', "");
Expect(1, 5971, '\p{_Buhd}', "");
Expect(0, 5971, '\p{^_Buhd}', "");
Expect(0, 5971, '\P{_Buhd}', "");
Expect(1, 5971, '\P{^_Buhd}', "");
Expect(0, 5972, '\p{_Buhd}', "");
Expect(1, 5972, '\p{^_Buhd}', "");
Expect(1, 5972, '\P{_Buhd}', "");
Expect(0, 5972, '\P{^_Buhd}', "");
Error('\p{		IS_Buhd:=}');
Error('\P{		IS_Buhd:=}');
Expect(1, 5971, '\p{isbuhd}', "");
Expect(0, 5971, '\p{^isbuhd}', "");
Expect(0, 5971, '\P{isbuhd}', "");
Expect(1, 5971, '\P{^isbuhd}', "");
Expect(0, 5972, '\p{isbuhd}', "");
Expect(1, 5972, '\p{^isbuhd}', "");
Expect(1, 5972, '\P{isbuhd}', "");
Expect(0, 5972, '\P{^isbuhd}', "");
Expect(1, 5971, '\p{ _Is_Buhd}', "");
Expect(0, 5971, '\p{^ _Is_Buhd}', "");
Expect(0, 5971, '\P{ _Is_Buhd}', "");
Expect(1, 5971, '\P{^ _Is_Buhd}', "");
Expect(0, 5972, '\p{ _Is_Buhd}', "");
Expect(1, 5972, '\p{^ _Is_Buhd}', "");
Expect(1, 5972, '\P{ _Is_Buhd}', "");
Expect(0, 5972, '\P{^ _Is_Buhd}', "");
Error('\p{_-byzantine_musical_Symbols:=}');
Error('\P{_-byzantine_musical_Symbols:=}');
Expect(1, 119039, '\p{byzantinemusicalsymbols}', "");
Expect(0, 119039, '\p{^byzantinemusicalsymbols}', "");
Expect(0, 119039, '\P{byzantinemusicalsymbols}', "");
Expect(1, 119039, '\P{^byzantinemusicalsymbols}', "");
Expect(0, 119040, '\p{byzantinemusicalsymbols}', "");
Expect(1, 119040, '\p{^byzantinemusicalsymbols}', "");
Expect(1, 119040, '\P{byzantinemusicalsymbols}', "");
Expect(0, 119040, '\P{^byzantinemusicalsymbols}', "");
Expect(1, 119039, '\p{-Byzantine_MUSICAL_Symbols}', "");
Expect(0, 119039, '\p{^-Byzantine_MUSICAL_Symbols}', "");
Expect(0, 119039, '\P{-Byzantine_MUSICAL_Symbols}', "");
Expect(1, 119039, '\P{^-Byzantine_MUSICAL_Symbols}', "");
Expect(0, 119040, '\p{-Byzantine_MUSICAL_Symbols}', "");
Expect(1, 119040, '\p{^-Byzantine_MUSICAL_Symbols}', "");
Expect(1, 119040, '\P{-Byzantine_MUSICAL_Symbols}', "");
Expect(0, 119040, '\P{^-Byzantine_MUSICAL_Symbols}', "");
Error('\p{-is_Byzantine_musical_SYMBOLS/a/}');
Error('\P{-is_Byzantine_musical_SYMBOLS/a/}');
Expect(1, 119039, '\p{isbyzantinemusicalsymbols}', "");
Expect(0, 119039, '\p{^isbyzantinemusicalsymbols}', "");
Expect(0, 119039, '\P{isbyzantinemusicalsymbols}', "");
Expect(1, 119039, '\P{^isbyzantinemusicalsymbols}', "");
Expect(0, 119040, '\p{isbyzantinemusicalsymbols}', "");
Expect(1, 119040, '\p{^isbyzantinemusicalsymbols}', "");
Expect(1, 119040, '\P{isbyzantinemusicalsymbols}', "");
Expect(0, 119040, '\P{^isbyzantinemusicalsymbols}', "");
Expect(1, 119039, '\p{ Is_Byzantine_Musical_Symbols}', "");
Expect(0, 119039, '\p{^ Is_Byzantine_Musical_Symbols}', "");
Expect(0, 119039, '\P{ Is_Byzantine_Musical_Symbols}', "");
Expect(1, 119039, '\P{^ Is_Byzantine_Musical_Symbols}', "");
Expect(0, 119040, '\p{ Is_Byzantine_Musical_Symbols}', "");
Expect(1, 119040, '\p{^ Is_Byzantine_Musical_Symbols}', "");
Expect(1, 119040, '\P{ Is_Byzantine_Musical_Symbols}', "");
Expect(0, 119040, '\P{^ Is_Byzantine_Musical_Symbols}', "");
Error('\p{	:=In_Byzantine_Musical_Symbols}');
Error('\P{	:=In_Byzantine_Musical_Symbols}');
Expect(1, 119039, '\p{inbyzantinemusicalsymbols}', "");
Expect(0, 119039, '\p{^inbyzantinemusicalsymbols}', "");
Expect(0, 119039, '\P{inbyzantinemusicalsymbols}', "");
Expect(1, 119039, '\P{^inbyzantinemusicalsymbols}', "");
Expect(0, 119040, '\p{inbyzantinemusicalsymbols}', "");
Expect(1, 119040, '\p{^inbyzantinemusicalsymbols}', "");
Expect(1, 119040, '\P{inbyzantinemusicalsymbols}', "");
Expect(0, 119040, '\P{^inbyzantinemusicalsymbols}', "");
Expect(1, 119039, '\p{--IN_BYZANTINE_Musical_Symbols}', "");
Expect(0, 119039, '\p{^--IN_BYZANTINE_Musical_Symbols}', "");
Expect(0, 119039, '\P{--IN_BYZANTINE_Musical_Symbols}', "");
Expect(1, 119039, '\P{^--IN_BYZANTINE_Musical_Symbols}', "");
Expect(0, 119040, '\p{--IN_BYZANTINE_Musical_Symbols}', "");
Expect(1, 119040, '\p{^--IN_BYZANTINE_Musical_Symbols}', "");
Expect(1, 119040, '\P{--IN_BYZANTINE_Musical_Symbols}', "");
Expect(0, 119040, '\P{^--IN_BYZANTINE_Musical_Symbols}', "");
Error('\p{:=  Byzantine_music}');
Error('\P{:=  Byzantine_music}');
Expect(1, 119039, '\p{byzantinemusic}', "");
Expect(0, 119039, '\p{^byzantinemusic}', "");
Expect(0, 119039, '\P{byzantinemusic}', "");
Expect(1, 119039, '\P{^byzantinemusic}', "");
Expect(0, 119040, '\p{byzantinemusic}', "");
Expect(1, 119040, '\p{^byzantinemusic}', "");
Expect(1, 119040, '\P{byzantinemusic}', "");
Expect(0, 119040, '\P{^byzantinemusic}', "");
Expect(1, 119039, '\p{__Byzantine_Music}', "");
Expect(0, 119039, '\p{^__Byzantine_Music}', "");
Expect(0, 119039, '\P{__Byzantine_Music}', "");
Expect(1, 119039, '\P{^__Byzantine_Music}', "");
Expect(0, 119040, '\p{__Byzantine_Music}', "");
Expect(1, 119040, '\p{^__Byzantine_Music}', "");
Expect(1, 119040, '\P{__Byzantine_Music}', "");
Expect(0, 119040, '\P{^__Byzantine_Music}', "");
Error('\p{:=IS_BYZANTINE_music}');
Error('\P{:=IS_BYZANTINE_music}');
Expect(1, 119039, '\p{isbyzantinemusic}', "");
Expect(0, 119039, '\p{^isbyzantinemusic}', "");
Expect(0, 119039, '\P{isbyzantinemusic}', "");
Expect(1, 119039, '\P{^isbyzantinemusic}', "");
Expect(0, 119040, '\p{isbyzantinemusic}', "");
Expect(1, 119040, '\p{^isbyzantinemusic}', "");
Expect(1, 119040, '\P{isbyzantinemusic}', "");
Expect(0, 119040, '\P{^isbyzantinemusic}', "");
Expect(1, 119039, '\p{- is_BYZANTINE_Music}', "");
Expect(0, 119039, '\p{^- is_BYZANTINE_Music}', "");
Expect(0, 119039, '\P{- is_BYZANTINE_Music}', "");
Expect(1, 119039, '\P{^- is_BYZANTINE_Music}', "");
Expect(0, 119040, '\p{- is_BYZANTINE_Music}', "");
Expect(1, 119040, '\p{^- is_BYZANTINE_Music}', "");
Expect(1, 119040, '\P{- is_BYZANTINE_Music}', "");
Expect(0, 119040, '\P{^- is_BYZANTINE_Music}', "");
Error('\p{/a/_-In_BYZANTINE_MUSIC}');
Error('\P{/a/_-In_BYZANTINE_MUSIC}');
Expect(1, 119039, '\p{inbyzantinemusic}', "");
Expect(0, 119039, '\p{^inbyzantinemusic}', "");
Expect(0, 119039, '\P{inbyzantinemusic}', "");
Expect(1, 119039, '\P{^inbyzantinemusic}', "");
Expect(0, 119040, '\p{inbyzantinemusic}', "");
Expect(1, 119040, '\p{^inbyzantinemusic}', "");
Expect(1, 119040, '\P{inbyzantinemusic}', "");
Expect(0, 119040, '\P{^inbyzantinemusic}', "");
Expect(1, 119039, '\p{  in_byzantine_Music}', "");
Expect(0, 119039, '\p{^  in_byzantine_Music}', "");
Expect(0, 119039, '\P{  in_byzantine_Music}', "");
Expect(1, 119039, '\P{^  in_byzantine_Music}', "");
Expect(0, 119040, '\p{  in_byzantine_Music}', "");
Expect(1, 119040, '\p{^  in_byzantine_Music}', "");
Expect(1, 119040, '\P{  in_byzantine_Music}', "");
Expect(0, 119040, '\P{^  in_byzantine_Music}', "");
Error('\p{-CJK_COMPATIBILITY/a/}');
Error('\P{-CJK_COMPATIBILITY/a/}');
Expect(1, 13311, '\p{cjkcompatibility}', "");
Expect(0, 13311, '\p{^cjkcompatibility}', "");
Expect(0, 13311, '\P{cjkcompatibility}', "");
Expect(1, 13311, '\P{^cjkcompatibility}', "");
Expect(0, 13312, '\p{cjkcompatibility}', "");
Expect(1, 13312, '\p{^cjkcompatibility}', "");
Expect(1, 13312, '\P{cjkcompatibility}', "");
Expect(0, 13312, '\P{^cjkcompatibility}', "");
Expect(1, 13311, '\p{ 	CJK_COMPATIBILITY}', "");
Expect(0, 13311, '\p{^ 	CJK_COMPATIBILITY}', "");
Expect(0, 13311, '\P{ 	CJK_COMPATIBILITY}', "");
Expect(1, 13311, '\P{^ 	CJK_COMPATIBILITY}', "");
Expect(0, 13312, '\p{ 	CJK_COMPATIBILITY}', "");
Expect(1, 13312, '\p{^ 	CJK_COMPATIBILITY}', "");
Expect(1, 13312, '\P{ 	CJK_COMPATIBILITY}', "");
Expect(0, 13312, '\P{^ 	CJK_COMPATIBILITY}', "");
Error('\p{-/a/IS_cjk_compatibility}');
Error('\P{-/a/IS_cjk_compatibility}');
Expect(1, 13311, '\p{iscjkcompatibility}', "");
Expect(0, 13311, '\p{^iscjkcompatibility}', "");
Expect(0, 13311, '\P{iscjkcompatibility}', "");
Expect(1, 13311, '\P{^iscjkcompatibility}', "");
Expect(0, 13312, '\p{iscjkcompatibility}', "");
Expect(1, 13312, '\p{^iscjkcompatibility}', "");
Expect(1, 13312, '\P{iscjkcompatibility}', "");
Expect(0, 13312, '\P{^iscjkcompatibility}', "");
Expect(1, 13311, '\p{	Is_CJK_COMPATIBILITY}', "");
Expect(0, 13311, '\p{^	Is_CJK_COMPATIBILITY}', "");
Expect(0, 13311, '\P{	Is_CJK_COMPATIBILITY}', "");
Expect(1, 13311, '\P{^	Is_CJK_COMPATIBILITY}', "");
Expect(0, 13312, '\p{	Is_CJK_COMPATIBILITY}', "");
Expect(1, 13312, '\p{^	Is_CJK_COMPATIBILITY}', "");
Expect(1, 13312, '\P{	Is_CJK_COMPATIBILITY}', "");
Expect(0, 13312, '\P{^	Is_CJK_COMPATIBILITY}', "");
Error('\p{_	In_CJK_Compatibility/a/}');
Error('\P{_	In_CJK_Compatibility/a/}');
Expect(1, 13311, '\p{incjkcompatibility}', "");
Expect(0, 13311, '\p{^incjkcompatibility}', "");
Expect(0, 13311, '\P{incjkcompatibility}', "");
Expect(1, 13311, '\P{^incjkcompatibility}', "");
Expect(0, 13312, '\p{incjkcompatibility}', "");
Expect(1, 13312, '\p{^incjkcompatibility}', "");
Expect(1, 13312, '\P{incjkcompatibility}', "");
Expect(0, 13312, '\P{^incjkcompatibility}', "");
Expect(1, 13311, '\p{-_In_CJK_Compatibility}', "");
Expect(0, 13311, '\p{^-_In_CJK_Compatibility}', "");
Expect(0, 13311, '\P{-_In_CJK_Compatibility}', "");
Expect(1, 13311, '\P{^-_In_CJK_Compatibility}', "");
Expect(0, 13312, '\p{-_In_CJK_Compatibility}', "");
Expect(1, 13312, '\p{^-_In_CJK_Compatibility}', "");
Expect(1, 13312, '\P{-_In_CJK_Compatibility}', "");
Expect(0, 13312, '\P{^-_In_CJK_Compatibility}', "");
Error('\p{:=_CJK_Compat}');
Error('\P{:=_CJK_Compat}');
Expect(1, 13311, '\p{cjkcompat}', "");
Expect(0, 13311, '\p{^cjkcompat}', "");
Expect(0, 13311, '\P{cjkcompat}', "");
Expect(1, 13311, '\P{^cjkcompat}', "");
Expect(0, 13312, '\p{cjkcompat}', "");
Expect(1, 13312, '\p{^cjkcompat}', "");
Expect(1, 13312, '\P{cjkcompat}', "");
Expect(0, 13312, '\P{^cjkcompat}', "");
Expect(1, 13311, '\p{ 	cjk_compat}', "");
Expect(0, 13311, '\p{^ 	cjk_compat}', "");
Expect(0, 13311, '\P{ 	cjk_compat}', "");
Expect(1, 13311, '\P{^ 	cjk_compat}', "");
Expect(0, 13312, '\p{ 	cjk_compat}', "");
Expect(1, 13312, '\p{^ 	cjk_compat}', "");
Expect(1, 13312, '\P{ 	cjk_compat}', "");
Expect(0, 13312, '\P{^ 	cjk_compat}', "");
Error('\p{-/a/IS_CJK_COMPAT}');
Error('\P{-/a/IS_CJK_COMPAT}');
Expect(1, 13311, '\p{iscjkcompat}', "");
Expect(0, 13311, '\p{^iscjkcompat}', "");
Expect(0, 13311, '\P{iscjkcompat}', "");
Expect(1, 13311, '\P{^iscjkcompat}', "");
Expect(0, 13312, '\p{iscjkcompat}', "");
Expect(1, 13312, '\p{^iscjkcompat}', "");
Expect(1, 13312, '\P{iscjkcompat}', "");
Expect(0, 13312, '\P{^iscjkcompat}', "");
Expect(1, 13311, '\p{_ Is_CJK_COMPAT}', "");
Expect(0, 13311, '\p{^_ Is_CJK_COMPAT}', "");
Expect(0, 13311, '\P{_ Is_CJK_COMPAT}', "");
Expect(1, 13311, '\P{^_ Is_CJK_COMPAT}', "");
Expect(0, 13312, '\p{_ Is_CJK_COMPAT}', "");
Expect(1, 13312, '\p{^_ Is_CJK_COMPAT}', "");
Expect(1, 13312, '\P{_ Is_CJK_COMPAT}', "");
Expect(0, 13312, '\P{^_ Is_CJK_COMPAT}', "");
Error('\p{		IN_cjk_Compat/a/}');
Error('\P{		IN_cjk_Compat/a/}');
Expect(1, 13311, '\p{incjkcompat}', "");
Expect(0, 13311, '\p{^incjkcompat}', "");
Expect(0, 13311, '\P{incjkcompat}', "");
Expect(1, 13311, '\P{^incjkcompat}', "");
Expect(0, 13312, '\p{incjkcompat}', "");
Expect(1, 13312, '\p{^incjkcompat}', "");
Expect(1, 13312, '\P{incjkcompat}', "");
Expect(0, 13312, '\P{^incjkcompat}', "");
Expect(1, 13311, '\p{-In_CJK_COMPAT}', "");
Expect(0, 13311, '\p{^-In_CJK_COMPAT}', "");
Expect(0, 13311, '\P{-In_CJK_COMPAT}', "");
Expect(1, 13311, '\P{^-In_CJK_COMPAT}', "");
Expect(0, 13312, '\p{-In_CJK_COMPAT}', "");
Expect(1, 13312, '\p{^-In_CJK_COMPAT}', "");
Expect(1, 13312, '\P{-In_CJK_COMPAT}', "");
Expect(0, 13312, '\P{^-In_CJK_COMPAT}', "");
Error('\p{-:=cjk_compatibility_forms}');
Error('\P{-:=cjk_compatibility_forms}');
Expect(1, 65103, '\p{cjkcompatibilityforms}', "");
Expect(0, 65103, '\p{^cjkcompatibilityforms}', "");
Expect(0, 65103, '\P{cjkcompatibilityforms}', "");
Expect(1, 65103, '\P{^cjkcompatibilityforms}', "");
Expect(0, 65104, '\p{cjkcompatibilityforms}', "");
Expect(1, 65104, '\p{^cjkcompatibilityforms}', "");
Expect(1, 65104, '\P{cjkcompatibilityforms}', "");
Expect(0, 65104, '\P{^cjkcompatibilityforms}', "");
Expect(1, 65103, '\p{-cjk_compatibility_forms}', "");
Expect(0, 65103, '\p{^-cjk_compatibility_forms}', "");
Expect(0, 65103, '\P{-cjk_compatibility_forms}', "");
Expect(1, 65103, '\P{^-cjk_compatibility_forms}', "");
Expect(0, 65104, '\p{-cjk_compatibility_forms}', "");
Expect(1, 65104, '\p{^-cjk_compatibility_forms}', "");
Expect(1, 65104, '\P{-cjk_compatibility_forms}', "");
Expect(0, 65104, '\P{^-cjk_compatibility_forms}', "");
Error('\p{_/a/is_CJK_COMPATIBILITY_Forms}');
Error('\P{_/a/is_CJK_COMPATIBILITY_Forms}');
Expect(1, 65103, '\p{iscjkcompatibilityforms}', "");
Expect(0, 65103, '\p{^iscjkcompatibilityforms}', "");
Expect(0, 65103, '\P{iscjkcompatibilityforms}', "");
Expect(1, 65103, '\P{^iscjkcompatibilityforms}', "");
Expect(0, 65104, '\p{iscjkcompatibilityforms}', "");
Expect(1, 65104, '\p{^iscjkcompatibilityforms}', "");
Expect(1, 65104, '\P{iscjkcompatibilityforms}', "");
Expect(0, 65104, '\P{^iscjkcompatibilityforms}', "");
Expect(1, 65103, '\p{ -IS_CJK_Compatibility_FORMS}', "");
Expect(0, 65103, '\p{^ -IS_CJK_Compatibility_FORMS}', "");
Expect(0, 65103, '\P{ -IS_CJK_Compatibility_FORMS}', "");
Expect(1, 65103, '\P{^ -IS_CJK_Compatibility_FORMS}', "");
Expect(0, 65104, '\p{ -IS_CJK_Compatibility_FORMS}', "");
Expect(1, 65104, '\p{^ -IS_CJK_Compatibility_FORMS}', "");
Expect(1, 65104, '\P{ -IS_CJK_Compatibility_FORMS}', "");
Expect(0, 65104, '\P{^ -IS_CJK_Compatibility_FORMS}', "");
Error('\p{-:=in_CJK_Compatibility_forms}');
Error('\P{-:=in_CJK_Compatibility_forms}');
Expect(1, 65103, '\p{incjkcompatibilityforms}', "");
Expect(0, 65103, '\p{^incjkcompatibilityforms}', "");
Expect(0, 65103, '\P{incjkcompatibilityforms}', "");
Expect(1, 65103, '\P{^incjkcompatibilityforms}', "");
Expect(0, 65104, '\p{incjkcompatibilityforms}', "");
Expect(1, 65104, '\p{^incjkcompatibilityforms}', "");
Expect(1, 65104, '\P{incjkcompatibilityforms}', "");
Expect(0, 65104, '\P{^incjkcompatibilityforms}', "");
Expect(1, 65103, '\p{	_in_cjk_compatibility_forms}', "");
Expect(0, 65103, '\p{^	_in_cjk_compatibility_forms}', "");
Expect(0, 65103, '\P{	_in_cjk_compatibility_forms}', "");
Expect(1, 65103, '\P{^	_in_cjk_compatibility_forms}', "");
Expect(0, 65104, '\p{	_in_cjk_compatibility_forms}', "");
Expect(1, 65104, '\p{^	_in_cjk_compatibility_forms}', "");
Expect(1, 65104, '\P{	_in_cjk_compatibility_forms}', "");
Expect(0, 65104, '\P{^	_in_cjk_compatibility_forms}', "");
Error('\p{--CJK_COMPAT_Forms/a/}');
Error('\P{--CJK_COMPAT_Forms/a/}');
Expect(1, 65103, '\p{cjkcompatforms}', "");
Expect(0, 65103, '\p{^cjkcompatforms}', "");
Expect(0, 65103, '\P{cjkcompatforms}', "");
Expect(1, 65103, '\P{^cjkcompatforms}', "");
Expect(0, 65104, '\p{cjkcompatforms}', "");
Expect(1, 65104, '\p{^cjkcompatforms}', "");
Expect(1, 65104, '\P{cjkcompatforms}', "");
Expect(0, 65104, '\P{^cjkcompatforms}', "");
Expect(1, 65103, '\p{	 CJK_Compat_Forms}', "");
Expect(0, 65103, '\p{^	 CJK_Compat_Forms}', "");
Expect(0, 65103, '\P{	 CJK_Compat_Forms}', "");
Expect(1, 65103, '\P{^	 CJK_Compat_Forms}', "");
Expect(0, 65104, '\p{	 CJK_Compat_Forms}', "");
Expect(1, 65104, '\p{^	 CJK_Compat_Forms}', "");
Expect(1, 65104, '\P{	 CJK_Compat_Forms}', "");
Expect(0, 65104, '\P{^	 CJK_Compat_Forms}', "");
Error('\p{/a/_	IS_cjk_Compat_FORMS}');
Error('\P{/a/_	IS_cjk_Compat_FORMS}');
Expect(1, 65103, '\p{iscjkcompatforms}', "");
Expect(0, 65103, '\p{^iscjkcompatforms}', "");
Expect(0, 65103, '\P{iscjkcompatforms}', "");
Expect(1, 65103, '\P{^iscjkcompatforms}', "");
Expect(0, 65104, '\p{iscjkcompatforms}', "");
Expect(1, 65104, '\p{^iscjkcompatforms}', "");
Expect(1, 65104, '\P{iscjkcompatforms}', "");
Expect(0, 65104, '\P{^iscjkcompatforms}', "");
Expect(1, 65103, '\p{--Is_CJK_Compat_forms}', "");
Expect(0, 65103, '\p{^--Is_CJK_Compat_forms}', "");
Expect(0, 65103, '\P{--Is_CJK_Compat_forms}', "");
Expect(1, 65103, '\P{^--Is_CJK_Compat_forms}', "");
Expect(0, 65104, '\p{--Is_CJK_Compat_forms}', "");
Expect(1, 65104, '\p{^--Is_CJK_Compat_forms}', "");
Expect(1, 65104, '\P{--Is_CJK_Compat_forms}', "");
Expect(0, 65104, '\P{^--Is_CJK_Compat_forms}', "");
Error('\p{:=__In_CJK_Compat_Forms}');
Error('\P{:=__In_CJK_Compat_Forms}');
Expect(1, 65103, '\p{incjkcompatforms}', "");
Expect(0, 65103, '\p{^incjkcompatforms}', "");
Expect(0, 65103, '\P{incjkcompatforms}', "");
Expect(1, 65103, '\P{^incjkcompatforms}', "");
Expect(0, 65104, '\p{incjkcompatforms}', "");
Expect(1, 65104, '\p{^incjkcompatforms}', "");
Expect(1, 65104, '\P{incjkcompatforms}', "");
Expect(0, 65104, '\P{^incjkcompatforms}', "");
Expect(1, 65103, '\p{_ In_CJK_COMPAT_Forms}', "");
Expect(0, 65103, '\p{^_ In_CJK_COMPAT_Forms}', "");
Expect(0, 65103, '\P{_ In_CJK_COMPAT_Forms}', "");
Expect(1, 65103, '\P{^_ In_CJK_COMPAT_Forms}', "");
Expect(0, 65104, '\p{_ In_CJK_COMPAT_Forms}', "");
Expect(1, 65104, '\p{^_ In_CJK_COMPAT_Forms}', "");
Expect(1, 65104, '\P{_ In_CJK_COMPAT_Forms}', "");
Expect(0, 65104, '\P{^_ In_CJK_COMPAT_Forms}', "");
Error('\p{	:=CJK_Compatibility_ideographs}');
Error('\P{	:=CJK_Compatibility_ideographs}');
Expect(1, 64255, '\p{cjkcompatibilityideographs}', "");
Expect(0, 64255, '\p{^cjkcompatibilityideographs}', "");
Expect(0, 64255, '\P{cjkcompatibilityideographs}', "");
Expect(1, 64255, '\P{^cjkcompatibilityideographs}', "");
Expect(0, 64256, '\p{cjkcompatibilityideographs}', "");
Expect(1, 64256, '\p{^cjkcompatibilityideographs}', "");
Expect(1, 64256, '\P{cjkcompatibilityideographs}', "");
Expect(0, 64256, '\P{^cjkcompatibilityideographs}', "");
Expect(1, 64255, '\p{	 CJK_compatibility_ideographs}', "");
Expect(0, 64255, '\p{^	 CJK_compatibility_ideographs}', "");
Expect(0, 64255, '\P{	 CJK_compatibility_ideographs}', "");
Expect(1, 64255, '\P{^	 CJK_compatibility_ideographs}', "");
Expect(0, 64256, '\p{	 CJK_compatibility_ideographs}', "");
Expect(1, 64256, '\p{^	 CJK_compatibility_ideographs}', "");
Expect(1, 64256, '\P{	 CJK_compatibility_ideographs}', "");
Expect(0, 64256, '\P{^	 CJK_compatibility_ideographs}', "");
Error('\p{:=	_IS_CJK_Compatibility_IDEOGRAPHS}');
Error('\P{:=	_IS_CJK_Compatibility_IDEOGRAPHS}');
Expect(1, 64255, '\p{iscjkcompatibilityideographs}', "");
Expect(0, 64255, '\p{^iscjkcompatibilityideographs}', "");
Expect(0, 64255, '\P{iscjkcompatibilityideographs}', "");
Expect(1, 64255, '\P{^iscjkcompatibilityideographs}', "");
Expect(0, 64256, '\p{iscjkcompatibilityideographs}', "");
Expect(1, 64256, '\p{^iscjkcompatibilityideographs}', "");
Expect(1, 64256, '\P{iscjkcompatibilityideographs}', "");
Expect(0, 64256, '\P{^iscjkcompatibilityideographs}', "");
Expect(1, 64255, '\p{	IS_cjk_COMPATIBILITY_ideographs}', "");
Expect(0, 64255, '\p{^	IS_cjk_COMPATIBILITY_ideographs}', "");
Expect(0, 64255, '\P{	IS_cjk_COMPATIBILITY_ideographs}', "");
Expect(1, 64255, '\P{^	IS_cjk_COMPATIBILITY_ideographs}', "");
Expect(0, 64256, '\p{	IS_cjk_COMPATIBILITY_ideographs}', "");
Expect(1, 64256, '\p{^	IS_cjk_COMPATIBILITY_ideographs}', "");
Expect(1, 64256, '\P{	IS_cjk_COMPATIBILITY_ideographs}', "");
Expect(0, 64256, '\P{^	IS_cjk_COMPATIBILITY_ideographs}', "");
Error('\p{		In_cjk_COMPATIBILITY_Ideographs:=}');
Error('\P{		In_cjk_COMPATIBILITY_Ideographs:=}');
Expect(1, 64255, '\p{incjkcompatibilityideographs}', "");
Expect(0, 64255, '\p{^incjkcompatibilityideographs}', "");
Expect(0, 64255, '\P{incjkcompatibilityideographs}', "");
Expect(1, 64255, '\P{^incjkcompatibilityideographs}', "");
Expect(0, 64256, '\p{incjkcompatibilityideographs}', "");
Expect(1, 64256, '\p{^incjkcompatibilityideographs}', "");
Expect(1, 64256, '\P{incjkcompatibilityideographs}', "");
Expect(0, 64256, '\P{^incjkcompatibilityideographs}', "");
Expect(1, 64255, '\p{ in_CJK_compatibility_ideographs}', "");
Expect(0, 64255, '\p{^ in_CJK_compatibility_ideographs}', "");
Expect(0, 64255, '\P{ in_CJK_compatibility_ideographs}', "");
Expect(1, 64255, '\P{^ in_CJK_compatibility_ideographs}', "");
Expect(0, 64256, '\p{ in_CJK_compatibility_ideographs}', "");
Expect(1, 64256, '\p{^ in_CJK_compatibility_ideographs}', "");
Expect(1, 64256, '\P{ in_CJK_compatibility_ideographs}', "");
Expect(0, 64256, '\P{^ in_CJK_compatibility_ideographs}', "");
Error('\p{:=_ CJK_compat_Ideographs}');
Error('\P{:=_ CJK_compat_Ideographs}');
Expect(1, 64255, '\p{cjkcompatideographs}', "");
Expect(0, 64255, '\p{^cjkcompatideographs}', "");
Expect(0, 64255, '\P{cjkcompatideographs}', "");
Expect(1, 64255, '\P{^cjkcompatideographs}', "");
Expect(0, 64256, '\p{cjkcompatideographs}', "");
Expect(1, 64256, '\p{^cjkcompatideographs}', "");
Expect(1, 64256, '\P{cjkcompatideographs}', "");
Expect(0, 64256, '\P{^cjkcompatideographs}', "");
Expect(1, 64255, '\p{	-CJK_Compat_Ideographs}', "");
Expect(0, 64255, '\p{^	-CJK_Compat_Ideographs}', "");
Expect(0, 64255, '\P{	-CJK_Compat_Ideographs}', "");
Expect(1, 64255, '\P{^	-CJK_Compat_Ideographs}', "");
Expect(0, 64256, '\p{	-CJK_Compat_Ideographs}', "");
Expect(1, 64256, '\p{^	-CJK_Compat_Ideographs}', "");
Expect(1, 64256, '\P{	-CJK_Compat_Ideographs}', "");
Expect(0, 64256, '\P{^	-CJK_Compat_Ideographs}', "");
Error('\p{ Is_CJK_compat_IDEOGRAPHS:=}');
Error('\P{ Is_CJK_compat_IDEOGRAPHS:=}');
Expect(1, 64255, '\p{iscjkcompatideographs}', "");
Expect(0, 64255, '\p{^iscjkcompatideographs}', "");
Expect(0, 64255, '\P{iscjkcompatideographs}', "");
Expect(1, 64255, '\P{^iscjkcompatideographs}', "");
Expect(0, 64256, '\p{iscjkcompatideographs}', "");
Expect(1, 64256, '\p{^iscjkcompatideographs}', "");
Expect(1, 64256, '\P{iscjkcompatideographs}', "");
Expect(0, 64256, '\P{^iscjkcompatideographs}', "");
Expect(1, 64255, '\p{_Is_CJK_Compat_ideographs}', "");
Expect(0, 64255, '\p{^_Is_CJK_Compat_ideographs}', "");
Expect(0, 64255, '\P{_Is_CJK_Compat_ideographs}', "");
Expect(1, 64255, '\P{^_Is_CJK_Compat_ideographs}', "");
Expect(0, 64256, '\p{_Is_CJK_Compat_ideographs}', "");
Expect(1, 64256, '\p{^_Is_CJK_Compat_ideographs}', "");
Expect(1, 64256, '\P{_Is_CJK_Compat_ideographs}', "");
Expect(0, 64256, '\P{^_Is_CJK_Compat_ideographs}', "");
Error('\p{/a/-In_cjk_Compat_ideographs}');
Error('\P{/a/-In_cjk_Compat_ideographs}');
Expect(1, 64255, '\p{incjkcompatideographs}', "");
Expect(0, 64255, '\p{^incjkcompatideographs}', "");
Expect(0, 64255, '\P{incjkcompatideographs}', "");
Expect(1, 64255, '\P{^incjkcompatideographs}', "");
Expect(0, 64256, '\p{incjkcompatideographs}', "");
Expect(1, 64256, '\p{^incjkcompatideographs}', "");
Expect(1, 64256, '\P{incjkcompatideographs}', "");
Expect(0, 64256, '\P{^incjkcompatideographs}', "");
Expect(1, 64255, '\p{IN_CJK_Compat_Ideographs}', "");
Expect(0, 64255, '\p{^IN_CJK_Compat_Ideographs}', "");
Expect(0, 64255, '\P{IN_CJK_Compat_Ideographs}', "");
Expect(1, 64255, '\P{^IN_CJK_Compat_Ideographs}', "");
Expect(0, 64256, '\p{IN_CJK_Compat_Ideographs}', "");
Expect(1, 64256, '\p{^IN_CJK_Compat_Ideographs}', "");
Expect(1, 64256, '\P{IN_CJK_Compat_Ideographs}', "");
Expect(0, 64256, '\P{^IN_CJK_Compat_Ideographs}', "");
Error('\p{_	CJK_COMPATIBILITY_Ideographs_Supplement/a/}');
Error('\P{_	CJK_COMPATIBILITY_Ideographs_Supplement/a/}');
Expect(1, 195103, '\p{cjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\p{^cjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\P{cjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\P{^cjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\p{cjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\p{^cjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\P{cjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\P{^cjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\p{_-CJK_Compatibility_IDEOGRAPHS_supplement}', "");
Expect(0, 195103, '\p{^_-CJK_Compatibility_IDEOGRAPHS_supplement}', "");
Expect(0, 195103, '\P{_-CJK_Compatibility_IDEOGRAPHS_supplement}', "");
Expect(1, 195103, '\P{^_-CJK_Compatibility_IDEOGRAPHS_supplement}', "");
Expect(0, 195104, '\p{_-CJK_Compatibility_IDEOGRAPHS_supplement}', "");
Expect(1, 195104, '\p{^_-CJK_Compatibility_IDEOGRAPHS_supplement}', "");
Expect(1, 195104, '\P{_-CJK_Compatibility_IDEOGRAPHS_supplement}', "");
Expect(0, 195104, '\P{^_-CJK_Compatibility_IDEOGRAPHS_supplement}', "");
Error('\p{_-IS_CJK_compatibility_Ideographs_supplement/a/}');
Error('\P{_-IS_CJK_compatibility_Ideographs_supplement/a/}');
Expect(1, 195103, '\p{iscjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\p{^iscjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\P{iscjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\P{^iscjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\p{iscjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\p{^iscjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\P{iscjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\P{^iscjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\p{-is_CJK_Compatibility_Ideographs_Supplement}', "");
Expect(0, 195103, '\p{^-is_CJK_Compatibility_Ideographs_Supplement}', "");
Expect(0, 195103, '\P{-is_CJK_Compatibility_Ideographs_Supplement}', "");
Expect(1, 195103, '\P{^-is_CJK_Compatibility_Ideographs_Supplement}', "");
Expect(0, 195104, '\p{-is_CJK_Compatibility_Ideographs_Supplement}', "");
Expect(1, 195104, '\p{^-is_CJK_Compatibility_Ideographs_Supplement}', "");
Expect(1, 195104, '\P{-is_CJK_Compatibility_Ideographs_Supplement}', "");
Expect(0, 195104, '\P{^-is_CJK_Compatibility_Ideographs_Supplement}', "");
Error('\p{:=in_cjk_compatibility_Ideographs_supplement}');
Error('\P{:=in_cjk_compatibility_Ideographs_supplement}');
Expect(1, 195103, '\p{incjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\p{^incjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\P{incjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\P{^incjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\p{incjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\p{^incjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\P{incjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\P{^incjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\p{_In_CJK_Compatibility_Ideographs_SUPPLEMENT}', "");
Expect(0, 195103, '\p{^_In_CJK_Compatibility_Ideographs_SUPPLEMENT}', "");
Expect(0, 195103, '\P{_In_CJK_Compatibility_Ideographs_SUPPLEMENT}', "");
Expect(1, 195103, '\P{^_In_CJK_Compatibility_Ideographs_SUPPLEMENT}', "");
Expect(0, 195104, '\p{_In_CJK_Compatibility_Ideographs_SUPPLEMENT}', "");
Expect(1, 195104, '\p{^_In_CJK_Compatibility_Ideographs_SUPPLEMENT}', "");
Expect(1, 195104, '\P{_In_CJK_Compatibility_Ideographs_SUPPLEMENT}', "");
Expect(0, 195104, '\P{^_In_CJK_Compatibility_Ideographs_SUPPLEMENT}', "");
Error('\p{ /a/CJK_compat_Ideographs_SUP}');
Error('\P{ /a/CJK_compat_Ideographs_SUP}');
Expect(1, 195103, '\p{cjkcompatideographssup}', "");
Expect(0, 195103, '\p{^cjkcompatideographssup}', "");
Expect(0, 195103, '\P{cjkcompatideographssup}', "");
Expect(1, 195103, '\P{^cjkcompatideographssup}', "");
Expect(0, 195104, '\p{cjkcompatideographssup}', "");
Expect(1, 195104, '\p{^cjkcompatideographssup}', "");
Expect(1, 195104, '\P{cjkcompatideographssup}', "");
Expect(0, 195104, '\P{^cjkcompatideographssup}', "");
Expect(1, 195103, '\p{_	CJK_COMPAT_Ideographs_Sup}', "");
Expect(0, 195103, '\p{^_	CJK_COMPAT_Ideographs_Sup}', "");
Expect(0, 195103, '\P{_	CJK_COMPAT_Ideographs_Sup}', "");
Expect(1, 195103, '\P{^_	CJK_COMPAT_Ideographs_Sup}', "");
Expect(0, 195104, '\p{_	CJK_COMPAT_Ideographs_Sup}', "");
Expect(1, 195104, '\p{^_	CJK_COMPAT_Ideographs_Sup}', "");
Expect(1, 195104, '\P{_	CJK_COMPAT_Ideographs_Sup}', "");
Expect(0, 195104, '\P{^_	CJK_COMPAT_Ideographs_Sup}', "");
Error('\p{	:=IS_CJK_Compat_ideographs_SUP}');
Error('\P{	:=IS_CJK_Compat_ideographs_SUP}');
Expect(1, 195103, '\p{iscjkcompatideographssup}', "");
Expect(0, 195103, '\p{^iscjkcompatideographssup}', "");
Expect(0, 195103, '\P{iscjkcompatideographssup}', "");
Expect(1, 195103, '\P{^iscjkcompatideographssup}', "");
Expect(0, 195104, '\p{iscjkcompatideographssup}', "");
Expect(1, 195104, '\p{^iscjkcompatideographssup}', "");
Expect(1, 195104, '\P{iscjkcompatideographssup}', "");
Expect(0, 195104, '\P{^iscjkcompatideographssup}', "");
Expect(1, 195103, '\p{_ Is_CJK_Compat_IDEOGRAPHS_Sup}', "");
Expect(0, 195103, '\p{^_ Is_CJK_Compat_IDEOGRAPHS_Sup}', "");
Expect(0, 195103, '\P{_ Is_CJK_Compat_IDEOGRAPHS_Sup}', "");
Expect(1, 195103, '\P{^_ Is_CJK_Compat_IDEOGRAPHS_Sup}', "");
Expect(0, 195104, '\p{_ Is_CJK_Compat_IDEOGRAPHS_Sup}', "");
Expect(1, 195104, '\p{^_ Is_CJK_Compat_IDEOGRAPHS_Sup}', "");
Expect(1, 195104, '\P{_ Is_CJK_Compat_IDEOGRAPHS_Sup}', "");
Expect(0, 195104, '\P{^_ Is_CJK_Compat_IDEOGRAPHS_Sup}', "");
Error('\p{-In_CJK_compat_Ideographs_SUP:=}');
Error('\P{-In_CJK_compat_Ideographs_SUP:=}');
Expect(1, 195103, '\p{incjkcompatideographssup}', "");
Expect(0, 195103, '\p{^incjkcompatideographssup}', "");
Expect(0, 195103, '\P{incjkcompatideographssup}', "");
Expect(1, 195103, '\P{^incjkcompatideographssup}', "");
Expect(0, 195104, '\p{incjkcompatideographssup}', "");
Expect(1, 195104, '\p{^incjkcompatideographssup}', "");
Expect(1, 195104, '\P{incjkcompatideographssup}', "");
Expect(0, 195104, '\P{^incjkcompatideographssup}', "");
Expect(1, 195103, '\p{ -in_cjk_Compat_Ideographs_Sup}', "");
Expect(0, 195103, '\p{^ -in_cjk_Compat_Ideographs_Sup}', "");
Expect(0, 195103, '\P{ -in_cjk_Compat_Ideographs_Sup}', "");
Expect(1, 195103, '\P{^ -in_cjk_Compat_Ideographs_Sup}', "");
Expect(0, 195104, '\p{ -in_cjk_Compat_Ideographs_Sup}', "");
Expect(1, 195104, '\p{^ -in_cjk_Compat_Ideographs_Sup}', "");
Expect(1, 195104, '\P{ -in_cjk_Compat_Ideographs_Sup}', "");
Expect(0, 195104, '\P{^ -in_cjk_Compat_Ideographs_Sup}', "");
Error('\p{:=_CJK_Radicals_SUPPLEMENT}');
Error('\P{:=_CJK_Radicals_SUPPLEMENT}');
Expect(1, 12031, '\p{cjkradicalssupplement}', "");
Expect(0, 12031, '\p{^cjkradicalssupplement}', "");
Expect(0, 12031, '\P{cjkradicalssupplement}', "");
Expect(1, 12031, '\P{^cjkradicalssupplement}', "");
Expect(0, 12032, '\p{cjkradicalssupplement}', "");
Expect(1, 12032, '\p{^cjkradicalssupplement}', "");
Expect(1, 12032, '\P{cjkradicalssupplement}', "");
Expect(0, 12032, '\P{^cjkradicalssupplement}', "");
Expect(1, 12031, '\p{_cjk_RADICALS_SUPPLEMENT}', "");
Expect(0, 12031, '\p{^_cjk_RADICALS_SUPPLEMENT}', "");
Expect(0, 12031, '\P{_cjk_RADICALS_SUPPLEMENT}', "");
Expect(1, 12031, '\P{^_cjk_RADICALS_SUPPLEMENT}', "");
Expect(0, 12032, '\p{_cjk_RADICALS_SUPPLEMENT}', "");
Expect(1, 12032, '\p{^_cjk_RADICALS_SUPPLEMENT}', "");
Expect(1, 12032, '\P{_cjk_RADICALS_SUPPLEMENT}', "");
Expect(0, 12032, '\P{^_cjk_RADICALS_SUPPLEMENT}', "");
Error('\p{:=	_Is_CJK_Radicals_Supplement}');
Error('\P{:=	_Is_CJK_Radicals_Supplement}');
Expect(1, 12031, '\p{iscjkradicalssupplement}', "");
Expect(0, 12031, '\p{^iscjkradicalssupplement}', "");
Expect(0, 12031, '\P{iscjkradicalssupplement}', "");
Expect(1, 12031, '\P{^iscjkradicalssupplement}', "");
Expect(0, 12032, '\p{iscjkradicalssupplement}', "");
Expect(1, 12032, '\p{^iscjkradicalssupplement}', "");
Expect(1, 12032, '\P{iscjkradicalssupplement}', "");
Expect(0, 12032, '\P{^iscjkradicalssupplement}', "");
Expect(1, 12031, '\p{  IS_CJK_Radicals_supplement}', "");
Expect(0, 12031, '\p{^  IS_CJK_Radicals_supplement}', "");
Expect(0, 12031, '\P{  IS_CJK_Radicals_supplement}', "");
Expect(1, 12031, '\P{^  IS_CJK_Radicals_supplement}', "");
Expect(0, 12032, '\p{  IS_CJK_Radicals_supplement}', "");
Expect(1, 12032, '\p{^  IS_CJK_Radicals_supplement}', "");
Expect(1, 12032, '\P{  IS_CJK_Radicals_supplement}', "");
Expect(0, 12032, '\P{^  IS_CJK_Radicals_supplement}', "");
Error('\p{ /a/In_cjk_radicals_supplement}');
Error('\P{ /a/In_cjk_radicals_supplement}');
Expect(1, 12031, '\p{incjkradicalssupplement}', "");
Expect(0, 12031, '\p{^incjkradicalssupplement}', "");
Expect(0, 12031, '\P{incjkradicalssupplement}', "");
Expect(1, 12031, '\P{^incjkradicalssupplement}', "");
Expect(0, 12032, '\p{incjkradicalssupplement}', "");
Expect(1, 12032, '\p{^incjkradicalssupplement}', "");
Expect(1, 12032, '\P{incjkradicalssupplement}', "");
Expect(0, 12032, '\P{^incjkradicalssupplement}', "");
Expect(1, 12031, '\p{-_IN_cjk_Radicals_Supplement}', "");
Expect(0, 12031, '\p{^-_IN_cjk_Radicals_Supplement}', "");
Expect(0, 12031, '\P{-_IN_cjk_Radicals_Supplement}', "");
Expect(1, 12031, '\P{^-_IN_cjk_Radicals_Supplement}', "");
Expect(0, 12032, '\p{-_IN_cjk_Radicals_Supplement}', "");
Expect(1, 12032, '\p{^-_IN_cjk_Radicals_Supplement}', "");
Expect(1, 12032, '\P{-_IN_cjk_Radicals_Supplement}', "");
Expect(0, 12032, '\P{^-_IN_cjk_Radicals_Supplement}', "");
Error('\p{  CJK_RADICALS_Sup:=}');
Error('\P{  CJK_RADICALS_Sup:=}');
Expect(1, 12031, '\p{cjkradicalssup}', "");
Expect(0, 12031, '\p{^cjkradicalssup}', "");
Expect(0, 12031, '\P{cjkradicalssup}', "");
Expect(1, 12031, '\P{^cjkradicalssup}', "");
Expect(0, 12032, '\p{cjkradicalssup}', "");
Expect(1, 12032, '\p{^cjkradicalssup}', "");
Expect(1, 12032, '\P{cjkradicalssup}', "");
Expect(0, 12032, '\P{^cjkradicalssup}', "");
Expect(1, 12031, '\p{ 	CJK_RADICALS_sup}', "");
Expect(0, 12031, '\p{^ 	CJK_RADICALS_sup}', "");
Expect(0, 12031, '\P{ 	CJK_RADICALS_sup}', "");
Expect(1, 12031, '\P{^ 	CJK_RADICALS_sup}', "");
Expect(0, 12032, '\p{ 	CJK_RADICALS_sup}', "");
Expect(1, 12032, '\p{^ 	CJK_RADICALS_sup}', "");
Expect(1, 12032, '\P{ 	CJK_RADICALS_sup}', "");
Expect(0, 12032, '\P{^ 	CJK_RADICALS_sup}', "");
Error('\p{_ is_CJK_radicals_Sup/a/}');
Error('\P{_ is_CJK_radicals_Sup/a/}');
Expect(1, 12031, '\p{iscjkradicalssup}', "");
Expect(0, 12031, '\p{^iscjkradicalssup}', "");
Expect(0, 12031, '\P{iscjkradicalssup}', "");
Expect(1, 12031, '\P{^iscjkradicalssup}', "");
Expect(0, 12032, '\p{iscjkradicalssup}', "");
Expect(1, 12032, '\p{^iscjkradicalssup}', "");
Expect(1, 12032, '\P{iscjkradicalssup}', "");
Expect(0, 12032, '\P{^iscjkradicalssup}', "");
Expect(1, 12031, '\p{		IS_CJK_Radicals_SUP}', "");
Expect(0, 12031, '\p{^		IS_CJK_Radicals_SUP}', "");
Expect(0, 12031, '\P{		IS_CJK_Radicals_SUP}', "");
Expect(1, 12031, '\P{^		IS_CJK_Radicals_SUP}', "");
Expect(0, 12032, '\p{		IS_CJK_Radicals_SUP}', "");
Expect(1, 12032, '\p{^		IS_CJK_Radicals_SUP}', "");
Expect(1, 12032, '\P{		IS_CJK_Radicals_SUP}', "");
Expect(0, 12032, '\P{^		IS_CJK_Radicals_SUP}', "");
Error('\p{--in_CJK_Radicals_Sup/a/}');
Error('\P{--in_CJK_Radicals_Sup/a/}');
Expect(1, 12031, '\p{incjkradicalssup}', "");
Expect(0, 12031, '\p{^incjkradicalssup}', "");
Expect(0, 12031, '\P{incjkradicalssup}', "");
Expect(1, 12031, '\P{^incjkradicalssup}', "");
Expect(0, 12032, '\p{incjkradicalssup}', "");
Expect(1, 12032, '\p{^incjkradicalssup}', "");
Expect(1, 12032, '\P{incjkradicalssup}', "");
Expect(0, 12032, '\P{^incjkradicalssup}', "");
Expect(1, 12031, '\p{ IN_CJK_Radicals_Sup}', "");
Expect(0, 12031, '\p{^ IN_CJK_Radicals_Sup}', "");
Expect(0, 12031, '\P{ IN_CJK_Radicals_Sup}', "");
Expect(1, 12031, '\P{^ IN_CJK_Radicals_Sup}', "");
Expect(0, 12032, '\p{ IN_CJK_Radicals_Sup}', "");
Expect(1, 12032, '\p{^ IN_CJK_Radicals_Sup}', "");
Expect(1, 12032, '\P{ IN_CJK_Radicals_Sup}', "");
Expect(0, 12032, '\P{^ IN_CJK_Radicals_Sup}', "");
Error('\p{:=  CJK_Strokes}');
Error('\P{:=  CJK_Strokes}');
Expect(1, 12783, '\p{cjkstrokes}', "");
Expect(0, 12783, '\p{^cjkstrokes}', "");
Expect(0, 12783, '\P{cjkstrokes}', "");
Expect(1, 12783, '\P{^cjkstrokes}', "");
Expect(0, 12784, '\p{cjkstrokes}', "");
Expect(1, 12784, '\p{^cjkstrokes}', "");
Expect(1, 12784, '\P{cjkstrokes}', "");
Expect(0, 12784, '\P{^cjkstrokes}', "");
Expect(1, 12783, '\p{	 CJK_STROKES}', "");
Expect(0, 12783, '\p{^	 CJK_STROKES}', "");
Expect(0, 12783, '\P{	 CJK_STROKES}', "");
Expect(1, 12783, '\P{^	 CJK_STROKES}', "");
Expect(0, 12784, '\p{	 CJK_STROKES}', "");
Expect(1, 12784, '\p{^	 CJK_STROKES}', "");
Expect(1, 12784, '\P{	 CJK_STROKES}', "");
Expect(0, 12784, '\P{^	 CJK_STROKES}', "");
Error('\p{_/a/Is_cjk_STROKES}');
Error('\P{_/a/Is_cjk_STROKES}');
Expect(1, 12783, '\p{iscjkstrokes}', "");
Expect(0, 12783, '\p{^iscjkstrokes}', "");
Expect(0, 12783, '\P{iscjkstrokes}', "");
Expect(1, 12783, '\P{^iscjkstrokes}', "");
Expect(0, 12784, '\p{iscjkstrokes}', "");
Expect(1, 12784, '\p{^iscjkstrokes}', "");
Expect(1, 12784, '\P{iscjkstrokes}', "");
Expect(0, 12784, '\P{^iscjkstrokes}', "");
Expect(1, 12783, '\p{_IS_cjk_strokes}', "");
Expect(0, 12783, '\p{^_IS_cjk_strokes}', "");
Expect(0, 12783, '\P{_IS_cjk_strokes}', "");
Expect(1, 12783, '\P{^_IS_cjk_strokes}', "");
Expect(0, 12784, '\p{_IS_cjk_strokes}', "");
Expect(1, 12784, '\p{^_IS_cjk_strokes}', "");
Expect(1, 12784, '\P{_IS_cjk_strokes}', "");
Expect(0, 12784, '\P{^_IS_cjk_strokes}', "");
Error('\p{_IN_CJK_Strokes:=}');
Error('\P{_IN_CJK_Strokes:=}');
Expect(1, 12783, '\p{incjkstrokes}', "");
Expect(0, 12783, '\p{^incjkstrokes}', "");
Expect(0, 12783, '\P{incjkstrokes}', "");
Expect(1, 12783, '\P{^incjkstrokes}', "");
Expect(0, 12784, '\p{incjkstrokes}', "");
Expect(1, 12784, '\p{^incjkstrokes}', "");
Expect(1, 12784, '\P{incjkstrokes}', "");
Expect(0, 12784, '\P{^incjkstrokes}', "");
Expect(1, 12783, '\p{ in_cjk_Strokes}', "");
Expect(0, 12783, '\p{^ in_cjk_Strokes}', "");
Expect(0, 12783, '\P{ in_cjk_Strokes}', "");
Expect(1, 12783, '\P{^ in_cjk_Strokes}', "");
Expect(0, 12784, '\p{ in_cjk_Strokes}', "");
Expect(1, 12784, '\p{^ in_cjk_Strokes}', "");
Expect(1, 12784, '\P{ in_cjk_Strokes}', "");
Expect(0, 12784, '\P{^ in_cjk_Strokes}', "");
Error('\p{:=	cjk_SYMBOLS_AND_PUNCTUATION}');
Error('\P{:=	cjk_SYMBOLS_AND_PUNCTUATION}');
Expect(1, 12351, '\p{cjksymbolsandpunctuation}', "");
Expect(0, 12351, '\p{^cjksymbolsandpunctuation}', "");
Expect(0, 12351, '\P{cjksymbolsandpunctuation}', "");
Expect(1, 12351, '\P{^cjksymbolsandpunctuation}', "");
Expect(0, 12352, '\p{cjksymbolsandpunctuation}', "");
Expect(1, 12352, '\p{^cjksymbolsandpunctuation}', "");
Expect(1, 12352, '\P{cjksymbolsandpunctuation}', "");
Expect(0, 12352, '\P{^cjksymbolsandpunctuation}', "");
Expect(1, 12351, '\p{-_CJK_Symbols_and_Punctuation}', "");
Expect(0, 12351, '\p{^-_CJK_Symbols_and_Punctuation}', "");
Expect(0, 12351, '\P{-_CJK_Symbols_and_Punctuation}', "");
Expect(1, 12351, '\P{^-_CJK_Symbols_and_Punctuation}', "");
Expect(0, 12352, '\p{-_CJK_Symbols_and_Punctuation}', "");
Expect(1, 12352, '\p{^-_CJK_Symbols_and_Punctuation}', "");
Expect(1, 12352, '\P{-_CJK_Symbols_and_Punctuation}', "");
Expect(0, 12352, '\P{^-_CJK_Symbols_and_Punctuation}', "");
Error('\p{_/a/Is_CJK_symbols_AND_PUNCTUATION}');
Error('\P{_/a/Is_CJK_symbols_AND_PUNCTUATION}');
Expect(1, 12351, '\p{iscjksymbolsandpunctuation}', "");
Expect(0, 12351, '\p{^iscjksymbolsandpunctuation}', "");
Expect(0, 12351, '\P{iscjksymbolsandpunctuation}', "");
Expect(1, 12351, '\P{^iscjksymbolsandpunctuation}', "");
Expect(0, 12352, '\p{iscjksymbolsandpunctuation}', "");
Expect(1, 12352, '\p{^iscjksymbolsandpunctuation}', "");
Expect(1, 12352, '\P{iscjksymbolsandpunctuation}', "");
Expect(0, 12352, '\P{^iscjksymbolsandpunctuation}', "");
Expect(1, 12351, '\p{--Is_cjk_SYMBOLS_And_Punctuation}', "");
Expect(0, 12351, '\p{^--Is_cjk_SYMBOLS_And_Punctuation}', "");
Expect(0, 12351, '\P{--Is_cjk_SYMBOLS_And_Punctuation}', "");
Expect(1, 12351, '\P{^--Is_cjk_SYMBOLS_And_Punctuation}', "");
Expect(0, 12352, '\p{--Is_cjk_SYMBOLS_And_Punctuation}', "");
Expect(1, 12352, '\p{^--Is_cjk_SYMBOLS_And_Punctuation}', "");
Expect(1, 12352, '\P{--Is_cjk_SYMBOLS_And_Punctuation}', "");
Expect(0, 12352, '\P{^--Is_cjk_SYMBOLS_And_Punctuation}', "");
Error('\p{--in_CJK_symbols_And_Punctuation:=}');
Error('\P{--in_CJK_symbols_And_Punctuation:=}');
Expect(1, 12351, '\p{incjksymbolsandpunctuation}', "");
Expect(0, 12351, '\p{^incjksymbolsandpunctuation}', "");
Expect(0, 12351, '\P{incjksymbolsandpunctuation}', "");
Expect(1, 12351, '\P{^incjksymbolsandpunctuation}', "");
Expect(0, 12352, '\p{incjksymbolsandpunctuation}', "");
Expect(1, 12352, '\p{^incjksymbolsandpunctuation}', "");
Expect(1, 12352, '\P{incjksymbolsandpunctuation}', "");
Expect(0, 12352, '\P{^incjksymbolsandpunctuation}', "");
Expect(1, 12351, '\p{	_in_CJK_symbols_AND_PUNCTUATION}', "");
Expect(0, 12351, '\p{^	_in_CJK_symbols_AND_PUNCTUATION}', "");
Expect(0, 12351, '\P{	_in_CJK_symbols_AND_PUNCTUATION}', "");
Expect(1, 12351, '\P{^	_in_CJK_symbols_AND_PUNCTUATION}', "");
Expect(0, 12352, '\p{	_in_CJK_symbols_AND_PUNCTUATION}', "");
Expect(1, 12352, '\p{^	_in_CJK_symbols_AND_PUNCTUATION}', "");
Expect(1, 12352, '\P{	_in_CJK_symbols_AND_PUNCTUATION}', "");
Expect(0, 12352, '\P{^	_in_CJK_symbols_AND_PUNCTUATION}', "");
Error('\p{-:=CJK_SYMBOLS}');
Error('\P{-:=CJK_SYMBOLS}');
Expect(1, 12351, '\p{cjksymbols}', "");
Expect(0, 12351, '\p{^cjksymbols}', "");
Expect(0, 12351, '\P{cjksymbols}', "");
Expect(1, 12351, '\P{^cjksymbols}', "");
Expect(0, 12352, '\p{cjksymbols}', "");
Expect(1, 12352, '\p{^cjksymbols}', "");
Expect(1, 12352, '\P{cjksymbols}', "");
Expect(0, 12352, '\P{^cjksymbols}', "");
Expect(1, 12351, '\p{ cjk_symbols}', "");
Expect(0, 12351, '\p{^ cjk_symbols}', "");
Expect(0, 12351, '\P{ cjk_symbols}', "");
Expect(1, 12351, '\P{^ cjk_symbols}', "");
Expect(0, 12352, '\p{ cjk_symbols}', "");
Expect(1, 12352, '\p{^ cjk_symbols}', "");
Expect(1, 12352, '\P{ cjk_symbols}', "");
Expect(0, 12352, '\P{^ cjk_symbols}', "");
Error('\p{:= IS_CJK_Symbols}');
Error('\P{:= IS_CJK_Symbols}');
Expect(1, 12351, '\p{iscjksymbols}', "");
Expect(0, 12351, '\p{^iscjksymbols}', "");
Expect(0, 12351, '\P{iscjksymbols}', "");
Expect(1, 12351, '\P{^iscjksymbols}', "");
Expect(0, 12352, '\p{iscjksymbols}', "");
Expect(1, 12352, '\p{^iscjksymbols}', "");
Expect(1, 12352, '\P{iscjksymbols}', "");
Expect(0, 12352, '\P{^iscjksymbols}', "");
Expect(1, 12351, '\p{ _Is_CJK_symbols}', "");
Expect(0, 12351, '\p{^ _Is_CJK_symbols}', "");
Expect(0, 12351, '\P{ _Is_CJK_symbols}', "");
Expect(1, 12351, '\P{^ _Is_CJK_symbols}', "");
Expect(0, 12352, '\p{ _Is_CJK_symbols}', "");
Expect(1, 12352, '\p{^ _Is_CJK_symbols}', "");
Expect(1, 12352, '\P{ _Is_CJK_symbols}', "");
Expect(0, 12352, '\P{^ _Is_CJK_symbols}', "");
Error('\p{:=-In_CJK_Symbols}');
Error('\P{:=-In_CJK_Symbols}');
Expect(1, 12351, '\p{incjksymbols}', "");
Expect(0, 12351, '\p{^incjksymbols}', "");
Expect(0, 12351, '\P{incjksymbols}', "");
Expect(1, 12351, '\P{^incjksymbols}', "");
Expect(0, 12352, '\p{incjksymbols}', "");
Expect(1, 12352, '\p{^incjksymbols}', "");
Expect(1, 12352, '\P{incjksymbols}', "");
Expect(0, 12352, '\P{^incjksymbols}', "");
Expect(1, 12351, '\p{	-In_cjk_SYMBOLS}', "");
Expect(0, 12351, '\p{^	-In_cjk_SYMBOLS}', "");
Expect(0, 12351, '\P{	-In_cjk_SYMBOLS}', "");
Expect(1, 12351, '\P{^	-In_cjk_SYMBOLS}', "");
Expect(0, 12352, '\p{	-In_cjk_SYMBOLS}', "");
Expect(1, 12352, '\p{^	-In_cjk_SYMBOLS}', "");
Expect(1, 12352, '\P{	-In_cjk_SYMBOLS}', "");
Expect(0, 12352, '\P{^	-In_cjk_SYMBOLS}', "");
Error('\p{	_cjk_Unified_ideographs/a/}');
Error('\P{	_cjk_Unified_ideographs/a/}');
Expect(1, 40959, '\p{cjkunifiedideographs}', "");
Expect(0, 40959, '\p{^cjkunifiedideographs}', "");
Expect(0, 40959, '\P{cjkunifiedideographs}', "");
Expect(1, 40959, '\P{^cjkunifiedideographs}', "");
Expect(0, 40960, '\p{cjkunifiedideographs}', "");
Expect(1, 40960, '\p{^cjkunifiedideographs}', "");
Expect(1, 40960, '\P{cjkunifiedideographs}', "");
Expect(0, 40960, '\P{^cjkunifiedideographs}', "");
Expect(1, 40959, '\p{	CJK_Unified_ideographs}', "");
Expect(0, 40959, '\p{^	CJK_Unified_ideographs}', "");
Expect(0, 40959, '\P{	CJK_Unified_ideographs}', "");
Expect(1, 40959, '\P{^	CJK_Unified_ideographs}', "");
Expect(0, 40960, '\p{	CJK_Unified_ideographs}', "");
Expect(1, 40960, '\p{^	CJK_Unified_ideographs}', "");
Expect(1, 40960, '\P{	CJK_Unified_ideographs}', "");
Expect(0, 40960, '\P{^	CJK_Unified_ideographs}', "");
Error('\p{_ IS_CJK_unified_Ideographs/a/}');
Error('\P{_ IS_CJK_unified_Ideographs/a/}');
Expect(1, 40959, '\p{iscjkunifiedideographs}', "");
Expect(0, 40959, '\p{^iscjkunifiedideographs}', "");
Expect(0, 40959, '\P{iscjkunifiedideographs}', "");
Expect(1, 40959, '\P{^iscjkunifiedideographs}', "");
Expect(0, 40960, '\p{iscjkunifiedideographs}', "");
Expect(1, 40960, '\p{^iscjkunifiedideographs}', "");
Expect(1, 40960, '\P{iscjkunifiedideographs}', "");
Expect(0, 40960, '\P{^iscjkunifiedideographs}', "");
Expect(1, 40959, '\p{	Is_CJK_Unified_ideographs}', "");
Expect(0, 40959, '\p{^	Is_CJK_Unified_ideographs}', "");
Expect(0, 40959, '\P{	Is_CJK_Unified_ideographs}', "");
Expect(1, 40959, '\P{^	Is_CJK_Unified_ideographs}', "");
Expect(0, 40960, '\p{	Is_CJK_Unified_ideographs}', "");
Expect(1, 40960, '\p{^	Is_CJK_Unified_ideographs}', "");
Expect(1, 40960, '\P{	Is_CJK_Unified_ideographs}', "");
Expect(0, 40960, '\P{^	Is_CJK_Unified_ideographs}', "");
Error('\p{_/a/IN_CJK_unified_Ideographs}');
Error('\P{_/a/IN_CJK_unified_Ideographs}');
Expect(1, 40959, '\p{incjkunifiedideographs}', "");
Expect(0, 40959, '\p{^incjkunifiedideographs}', "");
Expect(0, 40959, '\P{incjkunifiedideographs}', "");
Expect(1, 40959, '\P{^incjkunifiedideographs}', "");
Expect(0, 40960, '\p{incjkunifiedideographs}', "");
Expect(1, 40960, '\p{^incjkunifiedideographs}', "");
Expect(1, 40960, '\P{incjkunifiedideographs}', "");
Expect(0, 40960, '\P{^incjkunifiedideographs}', "");
Expect(1, 40959, '\p{_In_CJK_UNIFIED_Ideographs}', "");
Expect(0, 40959, '\p{^_In_CJK_UNIFIED_Ideographs}', "");
Expect(0, 40959, '\P{_In_CJK_UNIFIED_Ideographs}', "");
Expect(1, 40959, '\P{^_In_CJK_UNIFIED_Ideographs}', "");
Expect(0, 40960, '\p{_In_CJK_UNIFIED_Ideographs}', "");
Expect(1, 40960, '\p{^_In_CJK_UNIFIED_Ideographs}', "");
Expect(1, 40960, '\P{_In_CJK_UNIFIED_Ideographs}', "");
Expect(0, 40960, '\P{^_In_CJK_UNIFIED_Ideographs}', "");
Error('\p{ :=CJK}');
Error('\P{ :=CJK}');
Expect(1, 40959, '\p{cjk}', "");
Expect(0, 40959, '\p{^cjk}', "");
Expect(0, 40959, '\P{cjk}', "");
Expect(1, 40959, '\P{^cjk}', "");
Expect(0, 40960, '\p{cjk}', "");
Expect(1, 40960, '\p{^cjk}', "");
Expect(1, 40960, '\P{cjk}', "");
Expect(0, 40960, '\P{^cjk}', "");
Expect(1, 40959, '\p{ cjk}', "");
Expect(0, 40959, '\p{^ cjk}', "");
Expect(0, 40959, '\P{ cjk}', "");
Expect(1, 40959, '\P{^ cjk}', "");
Expect(0, 40960, '\p{ cjk}', "");
Expect(1, 40960, '\p{^ cjk}', "");
Expect(1, 40960, '\P{ cjk}', "");
Expect(0, 40960, '\P{^ cjk}', "");
Error('\p{_ Is_cjk/a/}');
Error('\P{_ Is_cjk/a/}');
Expect(1, 40959, '\p{iscjk}', "");
Expect(0, 40959, '\p{^iscjk}', "");
Expect(0, 40959, '\P{iscjk}', "");
Expect(1, 40959, '\P{^iscjk}', "");
Expect(0, 40960, '\p{iscjk}', "");
Expect(1, 40960, '\p{^iscjk}', "");
Expect(1, 40960, '\P{iscjk}', "");
Expect(0, 40960, '\P{^iscjk}', "");
Expect(1, 40959, '\p{	Is_CJK}', "");
Expect(0, 40959, '\p{^	Is_CJK}', "");
Expect(0, 40959, '\P{	Is_CJK}', "");
Expect(1, 40959, '\P{^	Is_CJK}', "");
Expect(0, 40960, '\p{	Is_CJK}', "");
Expect(1, 40960, '\p{^	Is_CJK}', "");
Expect(1, 40960, '\P{	Is_CJK}', "");
Expect(0, 40960, '\P{^	Is_CJK}', "");
Error('\p{/a/In_CJK}');
Error('\P{/a/In_CJK}');
Expect(1, 40959, '\p{incjk}', "");
Expect(0, 40959, '\p{^incjk}', "");
Expect(0, 40959, '\P{incjk}', "");
Expect(1, 40959, '\P{^incjk}', "");
Expect(0, 40960, '\p{incjk}', "");
Expect(1, 40960, '\p{^incjk}', "");
Expect(1, 40960, '\P{incjk}', "");
Expect(0, 40960, '\P{^incjk}', "");
Expect(1, 40959, '\p{_	In_CJK}', "");
Expect(0, 40959, '\p{^_	In_CJK}', "");
Expect(0, 40959, '\P{_	In_CJK}', "");
Expect(1, 40959, '\P{^_	In_CJK}', "");
Expect(0, 40960, '\p{_	In_CJK}', "");
Expect(1, 40960, '\p{^_	In_CJK}', "");
Expect(1, 40960, '\P{_	In_CJK}', "");
Expect(0, 40960, '\P{^_	In_CJK}', "");
Error('\p{/a/_CJK_Unified_Ideographs_Extension_a}');
Error('\P{/a/_CJK_Unified_Ideographs_Extension_a}');
Expect(1, 19903, '\p{cjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\p{^cjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\P{cjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\P{^cjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\p{cjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\p{^cjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\P{cjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\P{^cjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\p{CJK_Unified_ideographs_extension_A}', "");
Expect(0, 19903, '\p{^CJK_Unified_ideographs_extension_A}', "");
Expect(0, 19903, '\P{CJK_Unified_ideographs_extension_A}', "");
Expect(1, 19903, '\P{^CJK_Unified_ideographs_extension_A}', "");
Expect(0, 19904, '\p{CJK_Unified_ideographs_extension_A}', "");
Expect(1, 19904, '\p{^CJK_Unified_ideographs_extension_A}', "");
Expect(1, 19904, '\P{CJK_Unified_ideographs_extension_A}', "");
Expect(0, 19904, '\P{^CJK_Unified_ideographs_extension_A}', "");
Error('\p{_ is_cjk_Unified_ideographs_extension_A:=}');
Error('\P{_ is_cjk_Unified_ideographs_extension_A:=}');
Expect(1, 19903, '\p{iscjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\p{^iscjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\P{iscjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\P{^iscjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\p{iscjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\p{^iscjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\P{iscjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\P{^iscjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\p{- IS_CJK_Unified_Ideographs_extension_A}', "");
Expect(0, 19903, '\p{^- IS_CJK_Unified_Ideographs_extension_A}', "");
Expect(0, 19903, '\P{- IS_CJK_Unified_Ideographs_extension_A}', "");
Expect(1, 19903, '\P{^- IS_CJK_Unified_Ideographs_extension_A}', "");
Expect(0, 19904, '\p{- IS_CJK_Unified_Ideographs_extension_A}', "");
Expect(1, 19904, '\p{^- IS_CJK_Unified_Ideographs_extension_A}', "");
Expect(1, 19904, '\P{- IS_CJK_Unified_Ideographs_extension_A}', "");
Expect(0, 19904, '\P{^- IS_CJK_Unified_Ideographs_extension_A}', "");
Error('\p{-IN_CJK_unified_Ideographs_Extension_A:=}');
Error('\P{-IN_CJK_unified_Ideographs_Extension_A:=}');
Expect(1, 19903, '\p{incjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\p{^incjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\P{incjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\P{^incjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\p{incjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\p{^incjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\P{incjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\P{^incjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\p{	 In_CJK_unified_IDEOGRAPHS_Extension_A}', "");
Expect(0, 19903, '\p{^	 In_CJK_unified_IDEOGRAPHS_Extension_A}', "");
Expect(0, 19903, '\P{	 In_CJK_unified_IDEOGRAPHS_Extension_A}', "");
Expect(1, 19903, '\P{^	 In_CJK_unified_IDEOGRAPHS_Extension_A}', "");
Expect(0, 19904, '\p{	 In_CJK_unified_IDEOGRAPHS_Extension_A}', "");
Expect(1, 19904, '\p{^	 In_CJK_unified_IDEOGRAPHS_Extension_A}', "");
Expect(1, 19904, '\P{	 In_CJK_unified_IDEOGRAPHS_Extension_A}', "");
Expect(0, 19904, '\P{^	 In_CJK_unified_IDEOGRAPHS_Extension_A}', "");
Error('\p{/a/		CJK_EXT_A}');
Error('\P{/a/		CJK_EXT_A}');
Expect(1, 19903, '\p{cjkexta}', "");
Expect(0, 19903, '\p{^cjkexta}', "");
Expect(0, 19903, '\P{cjkexta}', "");
Expect(1, 19903, '\P{^cjkexta}', "");
Expect(0, 19904, '\p{cjkexta}', "");
Expect(1, 19904, '\p{^cjkexta}', "");
Expect(1, 19904, '\P{cjkexta}', "");
Expect(0, 19904, '\P{^cjkexta}', "");
Expect(1, 19903, '\p{- CJK_ext_a}', "");
Expect(0, 19903, '\p{^- CJK_ext_a}', "");
Expect(0, 19903, '\P{- CJK_ext_a}', "");
Expect(1, 19903, '\P{^- CJK_ext_a}', "");
Expect(0, 19904, '\p{- CJK_ext_a}', "");
Expect(1, 19904, '\p{^- CJK_ext_a}', "");
Expect(1, 19904, '\P{- CJK_ext_a}', "");
Expect(0, 19904, '\P{^- CJK_ext_a}', "");
Error('\p{/a/IS_CJK_EXT_a}');
Error('\P{/a/IS_CJK_EXT_a}');
Expect(1, 19903, '\p{iscjkexta}', "");
Expect(0, 19903, '\p{^iscjkexta}', "");
Expect(0, 19903, '\P{iscjkexta}', "");
Expect(1, 19903, '\P{^iscjkexta}', "");
Expect(0, 19904, '\p{iscjkexta}', "");
Expect(1, 19904, '\p{^iscjkexta}', "");
Expect(1, 19904, '\P{iscjkexta}', "");
Expect(0, 19904, '\P{^iscjkexta}', "");
Expect(1, 19903, '\p{ -Is_cjk_Ext_A}', "");
Expect(0, 19903, '\p{^ -Is_cjk_Ext_A}', "");
Expect(0, 19903, '\P{ -Is_cjk_Ext_A}', "");
Expect(1, 19903, '\P{^ -Is_cjk_Ext_A}', "");
Expect(0, 19904, '\p{ -Is_cjk_Ext_A}', "");
Expect(1, 19904, '\p{^ -Is_cjk_Ext_A}', "");
Expect(1, 19904, '\P{ -Is_cjk_Ext_A}', "");
Expect(0, 19904, '\P{^ -Is_cjk_Ext_A}', "");
Error('\p{_:=In_CJK_Ext_A}');
Error('\P{_:=In_CJK_Ext_A}');
Expect(1, 19903, '\p{incjkexta}', "");
Expect(0, 19903, '\p{^incjkexta}', "");
Expect(0, 19903, '\P{incjkexta}', "");
Expect(1, 19903, '\P{^incjkexta}', "");
Expect(0, 19904, '\p{incjkexta}', "");
Expect(1, 19904, '\p{^incjkexta}', "");
Expect(1, 19904, '\P{incjkexta}', "");
Expect(0, 19904, '\P{^incjkexta}', "");
Expect(1, 19903, '\p{  IN_CJK_EXT_A}', "");
Expect(0, 19903, '\p{^  IN_CJK_EXT_A}', "");
Expect(0, 19903, '\P{  IN_CJK_EXT_A}', "");
Expect(1, 19903, '\P{^  IN_CJK_EXT_A}', "");
Expect(0, 19904, '\p{  IN_CJK_EXT_A}', "");
Expect(1, 19904, '\p{^  IN_CJK_EXT_A}', "");
Expect(1, 19904, '\P{  IN_CJK_EXT_A}', "");
Expect(0, 19904, '\P{^  IN_CJK_EXT_A}', "");
Error('\p{:=	-CJK_UNIFIED_ideographs_Extension_B}');
Error('\P{:=	-CJK_UNIFIED_ideographs_Extension_B}');
Expect(1, 173791, '\p{cjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\p{^cjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\P{cjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\P{^cjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\p{cjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\p{^cjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\P{cjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\P{^cjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\p{_CJK_Unified_ideographs_EXTENSION_B}', "");
Expect(0, 173791, '\p{^_CJK_Unified_ideographs_EXTENSION_B}', "");
Expect(0, 173791, '\P{_CJK_Unified_ideographs_EXTENSION_B}', "");
Expect(1, 173791, '\P{^_CJK_Unified_ideographs_EXTENSION_B}', "");
Expect(0, 173792, '\p{_CJK_Unified_ideographs_EXTENSION_B}', "");
Expect(1, 173792, '\p{^_CJK_Unified_ideographs_EXTENSION_B}', "");
Expect(1, 173792, '\P{_CJK_Unified_ideographs_EXTENSION_B}', "");
Expect(0, 173792, '\P{^_CJK_Unified_ideographs_EXTENSION_B}', "");
Error('\p{:=		Is_cjk_unified_IDEOGRAPHS_EXTENSION_B}');
Error('\P{:=		Is_cjk_unified_IDEOGRAPHS_EXTENSION_B}');
Expect(1, 173791, '\p{iscjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\p{^iscjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\P{iscjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\P{^iscjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\p{iscjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\p{^iscjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\P{iscjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\P{^iscjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\p{	_is_CJK_Unified_Ideographs_Extension_B}', "");
Expect(0, 173791, '\p{^	_is_CJK_Unified_Ideographs_Extension_B}', "");
Expect(0, 173791, '\P{	_is_CJK_Unified_Ideographs_Extension_B}', "");
Expect(1, 173791, '\P{^	_is_CJK_Unified_Ideographs_Extension_B}', "");
Expect(0, 173792, '\p{	_is_CJK_Unified_Ideographs_Extension_B}', "");
Expect(1, 173792, '\p{^	_is_CJK_Unified_Ideographs_Extension_B}', "");
Expect(1, 173792, '\P{	_is_CJK_Unified_Ideographs_Extension_B}', "");
Expect(0, 173792, '\P{^	_is_CJK_Unified_Ideographs_Extension_B}', "");
Error('\p{_	In_CJK_unified_IDEOGRAPHS_Extension_B/a/}');
Error('\P{_	In_CJK_unified_IDEOGRAPHS_Extension_B/a/}');
Expect(1, 173791, '\p{incjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\p{^incjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\P{incjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\P{^incjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\p{incjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\p{^incjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\P{incjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\P{^incjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\p{-in_CJK_Unified_IDEOGRAPHS_Extension_B}', "");
Expect(0, 173791, '\p{^-in_CJK_Unified_IDEOGRAPHS_Extension_B}', "");
Expect(0, 173791, '\P{-in_CJK_Unified_IDEOGRAPHS_Extension_B}', "");
Expect(1, 173791, '\P{^-in_CJK_Unified_IDEOGRAPHS_Extension_B}', "");
Expect(0, 173792, '\p{-in_CJK_Unified_IDEOGRAPHS_Extension_B}', "");
Expect(1, 173792, '\p{^-in_CJK_Unified_IDEOGRAPHS_Extension_B}', "");
Expect(1, 173792, '\P{-in_CJK_Unified_IDEOGRAPHS_Extension_B}', "");
Expect(0, 173792, '\P{^-in_CJK_Unified_IDEOGRAPHS_Extension_B}', "");
Error('\p{/a/- CJK_Ext_B}');
Error('\P{/a/- CJK_Ext_B}');
Expect(1, 173791, '\p{cjkextb}', "");
Expect(0, 173791, '\p{^cjkextb}', "");
Expect(0, 173791, '\P{cjkextb}', "");
Expect(1, 173791, '\P{^cjkextb}', "");
Expect(0, 173792, '\p{cjkextb}', "");
Expect(1, 173792, '\p{^cjkextb}', "");
Expect(1, 173792, '\P{cjkextb}', "");
Expect(0, 173792, '\P{^cjkextb}', "");
Expect(1, 173791, '\p{ CJK_ext_B}', "");
Expect(0, 173791, '\p{^ CJK_ext_B}', "");
Expect(0, 173791, '\P{ CJK_ext_B}', "");
Expect(1, 173791, '\P{^ CJK_ext_B}', "");
Expect(0, 173792, '\p{ CJK_ext_B}', "");
Expect(1, 173792, '\p{^ CJK_ext_B}', "");
Expect(1, 173792, '\P{ CJK_ext_B}', "");
Expect(0, 173792, '\P{^ CJK_ext_B}', "");
Error('\p{_/a/Is_cjk_EXT_b}');
Error('\P{_/a/Is_cjk_EXT_b}');
Expect(1, 173791, '\p{iscjkextb}', "");
Expect(0, 173791, '\p{^iscjkextb}', "");
Expect(0, 173791, '\P{iscjkextb}', "");
Expect(1, 173791, '\P{^iscjkextb}', "");
Expect(0, 173792, '\p{iscjkextb}', "");
Expect(1, 173792, '\p{^iscjkextb}', "");
Expect(1, 173792, '\P{iscjkextb}', "");
Expect(0, 173792, '\P{^iscjkextb}', "");
Expect(1, 173791, '\p{--IS_cjk_ext_B}', "");
Expect(0, 173791, '\p{^--IS_cjk_ext_B}', "");
Expect(0, 173791, '\P{--IS_cjk_ext_B}', "");
Expect(1, 173791, '\P{^--IS_cjk_ext_B}', "");
Expect(0, 173792, '\p{--IS_cjk_ext_B}', "");
Expect(1, 173792, '\p{^--IS_cjk_ext_B}', "");
Expect(1, 173792, '\P{--IS_cjk_ext_B}', "");
Expect(0, 173792, '\P{^--IS_cjk_ext_B}', "");
Error('\p{	/a/IN_CJK_ext_B}');
Error('\P{	/a/IN_CJK_ext_B}');
Expect(1, 173791, '\p{incjkextb}', "");
Expect(0, 173791, '\p{^incjkextb}', "");
Expect(0, 173791, '\P{incjkextb}', "");
Expect(1, 173791, '\P{^incjkextb}', "");
Expect(0, 173792, '\p{incjkextb}', "");
Expect(1, 173792, '\p{^incjkextb}', "");
Expect(1, 173792, '\P{incjkextb}', "");
Expect(0, 173792, '\P{^incjkextb}', "");
Expect(1, 173791, '\p{	-in_CJK_Ext_b}', "");
Expect(0, 173791, '\p{^	-in_CJK_Ext_b}', "");
Expect(0, 173791, '\P{	-in_CJK_Ext_b}', "");
Expect(1, 173791, '\P{^	-in_CJK_Ext_b}', "");
Expect(0, 173792, '\p{	-in_CJK_Ext_b}', "");
Expect(1, 173792, '\p{^	-in_CJK_Ext_b}', "");
Expect(1, 173792, '\P{	-in_CJK_Ext_b}', "");
Expect(0, 173792, '\P{^	-in_CJK_Ext_b}', "");
Error('\p{		CJK_unified_ideographs_Extension_C:=}');
Error('\P{		CJK_unified_ideographs_Extension_C:=}');
Expect(1, 177983, '\p{cjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\p{^cjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\P{cjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\P{^cjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\p{cjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\p{^cjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\P{cjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\P{^cjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\p{	-CJK_Unified_IDEOGRAPHS_EXTENSION_C}', "");
Expect(0, 177983, '\p{^	-CJK_Unified_IDEOGRAPHS_EXTENSION_C}', "");
Expect(0, 177983, '\P{	-CJK_Unified_IDEOGRAPHS_EXTENSION_C}', "");
Expect(1, 177983, '\P{^	-CJK_Unified_IDEOGRAPHS_EXTENSION_C}', "");
Expect(0, 177984, '\p{	-CJK_Unified_IDEOGRAPHS_EXTENSION_C}', "");
Expect(1, 177984, '\p{^	-CJK_Unified_IDEOGRAPHS_EXTENSION_C}', "");
Expect(1, 177984, '\P{	-CJK_Unified_IDEOGRAPHS_EXTENSION_C}', "");
Expect(0, 177984, '\P{^	-CJK_Unified_IDEOGRAPHS_EXTENSION_C}', "");
Error('\p{ _Is_CJK_Unified_Ideographs_extension_C/a/}');
Error('\P{ _Is_CJK_Unified_Ideographs_extension_C/a/}');
Expect(1, 177983, '\p{iscjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\p{^iscjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\P{iscjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\P{^iscjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\p{iscjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\p{^iscjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\P{iscjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\P{^iscjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\p{	 Is_cjk_UNIFIED_Ideographs_Extension_C}', "");
Expect(0, 177983, '\p{^	 Is_cjk_UNIFIED_Ideographs_Extension_C}', "");
Expect(0, 177983, '\P{	 Is_cjk_UNIFIED_Ideographs_Extension_C}', "");
Expect(1, 177983, '\P{^	 Is_cjk_UNIFIED_Ideographs_Extension_C}', "");
Expect(0, 177984, '\p{	 Is_cjk_UNIFIED_Ideographs_Extension_C}', "");
Expect(1, 177984, '\p{^	 Is_cjk_UNIFIED_Ideographs_Extension_C}', "");
Expect(1, 177984, '\P{	 Is_cjk_UNIFIED_Ideographs_Extension_C}', "");
Expect(0, 177984, '\P{^	 Is_cjk_UNIFIED_Ideographs_Extension_C}', "");
Error('\p{_:=In_CJK_Unified_Ideographs_extension_C}');
Error('\P{_:=In_CJK_Unified_Ideographs_extension_C}');
Expect(1, 177983, '\p{incjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\p{^incjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\P{incjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\P{^incjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\p{incjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\p{^incjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\P{incjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\P{^incjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\p{	in_CJK_Unified_IDEOGRAPHS_Extension_C}', "");
Expect(0, 177983, '\p{^	in_CJK_Unified_IDEOGRAPHS_Extension_C}', "");
Expect(0, 177983, '\P{	in_CJK_Unified_IDEOGRAPHS_Extension_C}', "");
Expect(1, 177983, '\P{^	in_CJK_Unified_IDEOGRAPHS_Extension_C}', "");
Expect(0, 177984, '\p{	in_CJK_Unified_IDEOGRAPHS_Extension_C}', "");
Expect(1, 177984, '\p{^	in_CJK_Unified_IDEOGRAPHS_Extension_C}', "");
Expect(1, 177984, '\P{	in_CJK_Unified_IDEOGRAPHS_Extension_C}', "");
Expect(0, 177984, '\P{^	in_CJK_Unified_IDEOGRAPHS_Extension_C}', "");
Error('\p{ /a/CJK_Ext_C}');
Error('\P{ /a/CJK_Ext_C}');
Expect(1, 177983, '\p{cjkextc}', "");
Expect(0, 177983, '\p{^cjkextc}', "");
Expect(0, 177983, '\P{cjkextc}', "");
Expect(1, 177983, '\P{^cjkextc}', "");
Expect(0, 177984, '\p{cjkextc}', "");
Expect(1, 177984, '\p{^cjkextc}', "");
Expect(1, 177984, '\P{cjkextc}', "");
Expect(0, 177984, '\P{^cjkextc}', "");
Expect(1, 177983, '\p{_-CJK_ext_C}', "");
Expect(0, 177983, '\p{^_-CJK_ext_C}', "");
Expect(0, 177983, '\P{_-CJK_ext_C}', "");
Expect(1, 177983, '\P{^_-CJK_ext_C}', "");
Expect(0, 177984, '\p{_-CJK_ext_C}', "");
Expect(1, 177984, '\p{^_-CJK_ext_C}', "");
Expect(1, 177984, '\P{_-CJK_ext_C}', "");
Expect(0, 177984, '\P{^_-CJK_ext_C}', "");
Error('\p{:=-IS_CJK_ext_c}');
Error('\P{:=-IS_CJK_ext_c}');
Expect(1, 177983, '\p{iscjkextc}', "");
Expect(0, 177983, '\p{^iscjkextc}', "");
Expect(0, 177983, '\P{iscjkextc}', "");
Expect(1, 177983, '\P{^iscjkextc}', "");
Expect(0, 177984, '\p{iscjkextc}', "");
Expect(1, 177984, '\p{^iscjkextc}', "");
Expect(1, 177984, '\P{iscjkextc}', "");
Expect(0, 177984, '\P{^iscjkextc}', "");
Expect(1, 177983, '\p{ IS_cjk_Ext_c}', "");
Expect(0, 177983, '\p{^ IS_cjk_Ext_c}', "");
Expect(0, 177983, '\P{ IS_cjk_Ext_c}', "");
Expect(1, 177983, '\P{^ IS_cjk_Ext_c}', "");
Expect(0, 177984, '\p{ IS_cjk_Ext_c}', "");
Expect(1, 177984, '\p{^ IS_cjk_Ext_c}', "");
Expect(1, 177984, '\P{ IS_cjk_Ext_c}', "");
Expect(0, 177984, '\P{^ IS_cjk_Ext_c}', "");
Error('\p{	/a/IN_CJK_Ext_C}');
Error('\P{	/a/IN_CJK_Ext_C}');
Expect(1, 177983, '\p{incjkextc}', "");
Expect(0, 177983, '\p{^incjkextc}', "");
Expect(0, 177983, '\P{incjkextc}', "");
Expect(1, 177983, '\P{^incjkextc}', "");
Expect(0, 177984, '\p{incjkextc}', "");
Expect(1, 177984, '\p{^incjkextc}', "");
Expect(1, 177984, '\P{incjkextc}', "");
Expect(0, 177984, '\P{^incjkextc}', "");
Expect(1, 177983, '\p{ In_CJK_ext_C}', "");
Expect(0, 177983, '\p{^ In_CJK_ext_C}', "");
Expect(0, 177983, '\P{ In_CJK_ext_C}', "");
Expect(1, 177983, '\P{^ In_CJK_ext_C}', "");
Expect(0, 177984, '\p{ In_CJK_ext_C}', "");
Expect(1, 177984, '\p{^ In_CJK_ext_C}', "");
Expect(1, 177984, '\P{ In_CJK_ext_C}', "");
Expect(0, 177984, '\P{^ In_CJK_ext_C}', "");
Error('\p{_:=CJK_Unified_ideographs_Extension_D}');
Error('\P{_:=CJK_Unified_ideographs_Extension_D}');
Expect(1, 178207, '\p{cjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\p{^cjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\P{cjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\P{^cjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\p{cjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\p{^cjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\P{cjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\P{^cjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\p{_CJK_UNIFIED_Ideographs_Extension_d}', "");
Expect(0, 178207, '\p{^_CJK_UNIFIED_Ideographs_Extension_d}', "");
Expect(0, 178207, '\P{_CJK_UNIFIED_Ideographs_Extension_d}', "");
Expect(1, 178207, '\P{^_CJK_UNIFIED_Ideographs_Extension_d}', "");
Expect(0, 178208, '\p{_CJK_UNIFIED_Ideographs_Extension_d}', "");
Expect(1, 178208, '\p{^_CJK_UNIFIED_Ideographs_Extension_d}', "");
Expect(1, 178208, '\P{_CJK_UNIFIED_Ideographs_Extension_d}', "");
Expect(0, 178208, '\P{^_CJK_UNIFIED_Ideographs_Extension_d}', "");
Error('\p{ -is_CJK_UNIFIED_IDEOGRAPHS_Extension_D/a/}');
Error('\P{ -is_CJK_UNIFIED_IDEOGRAPHS_Extension_D/a/}');
Expect(1, 178207, '\p{iscjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\p{^iscjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\P{iscjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\P{^iscjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\p{iscjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\p{^iscjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\P{iscjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\P{^iscjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\p{_	IS_CJK_Unified_IDEOGRAPHS_EXTENSION_d}', "");
Expect(0, 178207, '\p{^_	IS_CJK_Unified_IDEOGRAPHS_EXTENSION_d}', "");
Expect(0, 178207, '\P{_	IS_CJK_Unified_IDEOGRAPHS_EXTENSION_d}', "");
Expect(1, 178207, '\P{^_	IS_CJK_Unified_IDEOGRAPHS_EXTENSION_d}', "");
Expect(0, 178208, '\p{_	IS_CJK_Unified_IDEOGRAPHS_EXTENSION_d}', "");
Expect(1, 178208, '\p{^_	IS_CJK_Unified_IDEOGRAPHS_EXTENSION_d}', "");
Expect(1, 178208, '\P{_	IS_CJK_Unified_IDEOGRAPHS_EXTENSION_d}', "");
Expect(0, 178208, '\P{^_	IS_CJK_Unified_IDEOGRAPHS_EXTENSION_d}', "");
Error('\p{:=	-In_CJK_unified_ideographs_Extension_d}');
Error('\P{:=	-In_CJK_unified_ideographs_Extension_d}');
Expect(1, 178207, '\p{incjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\p{^incjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\P{incjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\P{^incjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\p{incjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\p{^incjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\P{incjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\P{^incjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\p{	 in_cjk_UNIFIED_IDEOGRAPHS_Extension_d}', "");
Expect(0, 178207, '\p{^	 in_cjk_UNIFIED_IDEOGRAPHS_Extension_d}', "");
Expect(0, 178207, '\P{	 in_cjk_UNIFIED_IDEOGRAPHS_Extension_d}', "");
Expect(1, 178207, '\P{^	 in_cjk_UNIFIED_IDEOGRAPHS_Extension_d}', "");
Expect(0, 178208, '\p{	 in_cjk_UNIFIED_IDEOGRAPHS_Extension_d}', "");
Expect(1, 178208, '\p{^	 in_cjk_UNIFIED_IDEOGRAPHS_Extension_d}', "");
Expect(1, 178208, '\P{	 in_cjk_UNIFIED_IDEOGRAPHS_Extension_d}', "");
Expect(0, 178208, '\P{^	 in_cjk_UNIFIED_IDEOGRAPHS_Extension_d}', "");
Error('\p{/a/-	CJK_ext_D}');
Error('\P{/a/-	CJK_ext_D}');
Expect(1, 178207, '\p{cjkextd}', "");
Expect(0, 178207, '\p{^cjkextd}', "");
Expect(0, 178207, '\P{cjkextd}', "");
Expect(1, 178207, '\P{^cjkextd}', "");
Expect(0, 178208, '\p{cjkextd}', "");
Expect(1, 178208, '\p{^cjkextd}', "");
Expect(1, 178208, '\P{cjkextd}', "");
Expect(0, 178208, '\P{^cjkextd}', "");
Expect(1, 178207, '\p{_CJK_Ext_d}', "");
Expect(0, 178207, '\p{^_CJK_Ext_d}', "");
Expect(0, 178207, '\P{_CJK_Ext_d}', "");
Expect(1, 178207, '\P{^_CJK_Ext_d}', "");
Expect(0, 178208, '\p{_CJK_Ext_d}', "");
Expect(1, 178208, '\p{^_CJK_Ext_d}', "");
Expect(1, 178208, '\P{_CJK_Ext_d}', "");
Expect(0, 178208, '\P{^_CJK_Ext_d}', "");
Error('\p{/a/__IS_CJK_ext_D}');
Error('\P{/a/__IS_CJK_ext_D}');
Expect(1, 178207, '\p{iscjkextd}', "");
Expect(0, 178207, '\p{^iscjkextd}', "");
Expect(0, 178207, '\P{iscjkextd}', "");
Expect(1, 178207, '\P{^iscjkextd}', "");
Expect(0, 178208, '\p{iscjkextd}', "");
Expect(1, 178208, '\p{^iscjkextd}', "");
Expect(1, 178208, '\P{iscjkextd}', "");
Expect(0, 178208, '\P{^iscjkextd}', "");
Expect(1, 178207, '\p{_	Is_CJK_ext_d}', "");
Expect(0, 178207, '\p{^_	Is_CJK_ext_d}', "");
Expect(0, 178207, '\P{_	Is_CJK_ext_d}', "");
Expect(1, 178207, '\P{^_	Is_CJK_ext_d}', "");
Expect(0, 178208, '\p{_	Is_CJK_ext_d}', "");
Expect(1, 178208, '\p{^_	Is_CJK_ext_d}', "");
Expect(1, 178208, '\P{_	Is_CJK_ext_d}', "");
Expect(0, 178208, '\P{^_	Is_CJK_ext_d}', "");
Error('\p{:=_	In_CJK_Ext_D}');
Error('\P{:=_	In_CJK_Ext_D}');
Expect(1, 178207, '\p{incjkextd}', "");
Expect(0, 178207, '\p{^incjkextd}', "");
Expect(0, 178207, '\P{incjkextd}', "");
Expect(1, 178207, '\P{^incjkextd}', "");
Expect(0, 178208, '\p{incjkextd}', "");
Expect(1, 178208, '\p{^incjkextd}', "");
Expect(1, 178208, '\P{incjkextd}', "");
Expect(0, 178208, '\P{^incjkextd}', "");
Expect(1, 178207, '\p{  in_CJK_Ext_D}', "");
Expect(0, 178207, '\p{^  in_CJK_Ext_D}', "");
Expect(0, 178207, '\P{  in_CJK_Ext_D}', "");
Expect(1, 178207, '\P{^  in_CJK_Ext_D}', "");
Expect(0, 178208, '\p{  in_CJK_Ext_D}', "");
Expect(1, 178208, '\p{^  in_CJK_Ext_D}', "");
Expect(1, 178208, '\P{  in_CJK_Ext_D}', "");
Expect(0, 178208, '\P{^  in_CJK_Ext_D}', "");
Error('\p{/a/	CANADIAN_ABORIGINAL}');
Error('\P{/a/	CANADIAN_ABORIGINAL}');
Expect(1, 6389, '\p{canadianaboriginal}', "");
Expect(0, 6389, '\p{^canadianaboriginal}', "");
Expect(0, 6389, '\P{canadianaboriginal}', "");
Expect(1, 6389, '\P{^canadianaboriginal}', "");
Expect(0, 6390, '\p{canadianaboriginal}', "");
Expect(1, 6390, '\p{^canadianaboriginal}', "");
Expect(1, 6390, '\P{canadianaboriginal}', "");
Expect(0, 6390, '\P{^canadianaboriginal}', "");
Expect(1, 6389, '\p{_CANADIAN_ABORIGINAL}', "");
Expect(0, 6389, '\p{^_CANADIAN_ABORIGINAL}', "");
Expect(0, 6389, '\P{_CANADIAN_ABORIGINAL}', "");
Expect(1, 6389, '\P{^_CANADIAN_ABORIGINAL}', "");
Expect(0, 6390, '\p{_CANADIAN_ABORIGINAL}', "");
Expect(1, 6390, '\p{^_CANADIAN_ABORIGINAL}', "");
Expect(1, 6390, '\P{_CANADIAN_ABORIGINAL}', "");
Expect(0, 6390, '\P{^_CANADIAN_ABORIGINAL}', "");
Error('\p{/a/is_CANADIAN_Aboriginal}');
Error('\P{/a/is_CANADIAN_Aboriginal}');
Expect(1, 6389, '\p{iscanadianaboriginal}', "");
Expect(0, 6389, '\p{^iscanadianaboriginal}', "");
Expect(0, 6389, '\P{iscanadianaboriginal}', "");
Expect(1, 6389, '\P{^iscanadianaboriginal}', "");
Expect(0, 6390, '\p{iscanadianaboriginal}', "");
Expect(1, 6390, '\p{^iscanadianaboriginal}', "");
Expect(1, 6390, '\P{iscanadianaboriginal}', "");
Expect(0, 6390, '\P{^iscanadianaboriginal}', "");
Expect(1, 6389, '\p{-_is_Canadian_aboriginal}', "");
Expect(0, 6389, '\p{^-_is_Canadian_aboriginal}', "");
Expect(0, 6389, '\P{-_is_Canadian_aboriginal}', "");
Expect(1, 6389, '\P{^-_is_Canadian_aboriginal}', "");
Expect(0, 6390, '\p{-_is_Canadian_aboriginal}', "");
Expect(1, 6390, '\p{^-_is_Canadian_aboriginal}', "");
Expect(1, 6390, '\P{-_is_Canadian_aboriginal}', "");
Expect(0, 6390, '\P{^-_is_Canadian_aboriginal}', "");
Error('\p{:=CANS}');
Error('\P{:=CANS}');
Expect(1, 6389, '\p{cans}', "");
Expect(0, 6389, '\p{^cans}', "");
Expect(0, 6389, '\P{cans}', "");
Expect(1, 6389, '\P{^cans}', "");
Expect(0, 6390, '\p{cans}', "");
Expect(1, 6390, '\p{^cans}', "");
Expect(1, 6390, '\P{cans}', "");
Expect(0, 6390, '\P{^cans}', "");
Expect(1, 6389, '\p{_	CANS}', "");
Expect(0, 6389, '\p{^_	CANS}', "");
Expect(0, 6389, '\P{_	CANS}', "");
Expect(1, 6389, '\P{^_	CANS}', "");
Expect(0, 6390, '\p{_	CANS}', "");
Expect(1, 6390, '\p{^_	CANS}', "");
Expect(1, 6390, '\P{_	CANS}', "");
Expect(0, 6390, '\P{^_	CANS}', "");
Error('\p{:= 	Is_cans}');
Error('\P{:= 	Is_cans}');
Expect(1, 6389, '\p{iscans}', "");
Expect(0, 6389, '\p{^iscans}', "");
Expect(0, 6389, '\P{iscans}', "");
Expect(1, 6389, '\P{^iscans}', "");
Expect(0, 6390, '\p{iscans}', "");
Expect(1, 6390, '\p{^iscans}', "");
Expect(1, 6390, '\P{iscans}', "");
Expect(0, 6390, '\P{^iscans}', "");
Expect(1, 6389, '\p{ _is_cans}', "");
Expect(0, 6389, '\p{^ _is_cans}', "");
Expect(0, 6389, '\P{ _is_cans}', "");
Expect(1, 6389, '\P{^ _is_cans}', "");
Expect(0, 6390, '\p{ _is_cans}', "");
Expect(1, 6390, '\p{^ _is_cans}', "");
Expect(1, 6390, '\P{ _is_cans}', "");
Expect(0, 6390, '\P{^ _is_cans}', "");
Error('\p{ /a/carian}');
Error('\P{ /a/carian}');
Expect(1, 66256, '\p{carian}', "");
Expect(0, 66256, '\p{^carian}', "");
Expect(0, 66256, '\P{carian}', "");
Expect(1, 66256, '\P{^carian}', "");
Expect(0, 66257, '\p{carian}', "");
Expect(1, 66257, '\p{^carian}', "");
Expect(1, 66257, '\P{carian}', "");
Expect(0, 66257, '\P{^carian}', "");
Expect(1, 66256, '\p{--Carian}', "");
Expect(0, 66256, '\p{^--Carian}', "");
Expect(0, 66256, '\P{--Carian}', "");
Expect(1, 66256, '\P{^--Carian}', "");
Expect(0, 66257, '\p{--Carian}', "");
Expect(1, 66257, '\p{^--Carian}', "");
Expect(1, 66257, '\P{--Carian}', "");
Expect(0, 66257, '\P{^--Carian}', "");
Error('\p{		Is_Carian:=}');
Error('\P{		Is_Carian:=}');
Expect(1, 66256, '\p{iscarian}', "");
Expect(0, 66256, '\p{^iscarian}', "");
Expect(0, 66256, '\P{iscarian}', "");
Expect(1, 66256, '\P{^iscarian}', "");
Expect(0, 66257, '\p{iscarian}', "");
Expect(1, 66257, '\p{^iscarian}', "");
Expect(1, 66257, '\P{iscarian}', "");
Expect(0, 66257, '\P{^iscarian}', "");
Expect(1, 66256, '\p{-_IS_Carian}', "");
Expect(0, 66256, '\p{^-_IS_Carian}', "");
Expect(0, 66256, '\P{-_IS_Carian}', "");
Expect(1, 66256, '\P{^-_IS_Carian}', "");
Expect(0, 66257, '\p{-_IS_Carian}', "");
Expect(1, 66257, '\p{^-_IS_Carian}', "");
Expect(1, 66257, '\P{-_IS_Carian}', "");
Expect(0, 66257, '\P{^-_IS_Carian}', "");
Error('\p{-:=Cari}');
Error('\P{-:=Cari}');
Expect(1, 66256, '\p{cari}', "");
Expect(0, 66256, '\p{^cari}', "");
Expect(0, 66256, '\P{cari}', "");
Expect(1, 66256, '\P{^cari}', "");
Expect(0, 66257, '\p{cari}', "");
Expect(1, 66257, '\p{^cari}', "");
Expect(1, 66257, '\P{cari}', "");
Expect(0, 66257, '\P{^cari}', "");
Expect(1, 66256, '\p{ CARI}', "");
Expect(0, 66256, '\p{^ CARI}', "");
Expect(0, 66256, '\P{ CARI}', "");
Expect(1, 66256, '\P{^ CARI}', "");
Expect(0, 66257, '\p{ CARI}', "");
Expect(1, 66257, '\p{^ CARI}', "");
Expect(1, 66257, '\P{ CARI}', "");
Expect(0, 66257, '\P{^ CARI}', "");
Error('\p{:=	_is_Cari}');
Error('\P{:=	_is_Cari}');
Expect(1, 66256, '\p{iscari}', "");
Expect(0, 66256, '\p{^iscari}', "");
Expect(0, 66256, '\P{iscari}', "");
Expect(1, 66256, '\P{^iscari}', "");
Expect(0, 66257, '\p{iscari}', "");
Expect(1, 66257, '\p{^iscari}', "");
Expect(1, 66257, '\P{iscari}', "");
Expect(0, 66257, '\P{^iscari}', "");
Expect(1, 66256, '\p{_-Is_Cari}', "");
Expect(0, 66256, '\p{^_-Is_Cari}', "");
Expect(0, 66256, '\P{_-Is_Cari}', "");
Expect(1, 66256, '\P{^_-Is_Cari}', "");
Expect(0, 66257, '\p{_-Is_Cari}', "");
Expect(1, 66257, '\p{^_-Is_Cari}', "");
Expect(1, 66257, '\P{_-Is_Cari}', "");
Expect(0, 66257, '\P{^_-Is_Cari}', "");
Error('\p{	:=Cased}');
Error('\P{	:=Cased}');
Expect(1, 120779, '\p{cased}', "");
Expect(0, 120779, '\p{^cased}', "");
Expect(0, 120779, '\P{cased}', "");
Expect(1, 120779, '\P{^cased}', "");
Expect(0, 120780, '\p{cased}', "");
Expect(1, 120780, '\p{^cased}', "");
Expect(1, 120780, '\P{cased}', "");
Expect(0, 120780, '\P{^cased}', "");
Expect(1, 120779, '\p{-CASED}', "");
Expect(0, 120779, '\p{^-CASED}', "");
Expect(0, 120779, '\P{-CASED}', "");
Expect(1, 120779, '\P{^-CASED}', "");
Expect(0, 120780, '\p{-CASED}', "");
Expect(1, 120780, '\p{^-CASED}', "");
Expect(1, 120780, '\P{-CASED}', "");
Expect(0, 120780, '\P{^-CASED}', "");
Error('\p{:=-Is_Cased}');
Error('\P{:=-Is_Cased}');
Expect(1, 120779, '\p{iscased}', "");
Expect(0, 120779, '\p{^iscased}', "");
Expect(0, 120779, '\P{iscased}', "");
Expect(1, 120779, '\P{^iscased}', "");
Expect(0, 120780, '\p{iscased}', "");
Expect(1, 120780, '\p{^iscased}', "");
Expect(1, 120780, '\P{iscased}', "");
Expect(0, 120780, '\P{^iscased}', "");
Expect(1, 120779, '\p{-Is_CASED}', "");
Expect(0, 120779, '\p{^-Is_CASED}', "");
Expect(0, 120779, '\P{-Is_CASED}', "");
Expect(1, 120779, '\P{^-Is_CASED}', "");
Expect(0, 120780, '\p{-Is_CASED}', "");
Expect(1, 120780, '\p{^-Is_CASED}', "");
Expect(1, 120780, '\P{-Is_CASED}', "");
Expect(0, 120780, '\P{^-Is_CASED}', "");
Error('\p{	CASED_Letter/a/}');
Error('\P{	CASED_Letter/a/}');
Expect(1, 120779, '\p{casedletter}', "");
Expect(0, 120779, '\p{^casedletter}', "");
Expect(0, 120779, '\P{casedletter}', "");
Expect(1, 120779, '\P{^casedletter}', "");
Expect(0, 120780, '\p{casedletter}', "");
Expect(1, 120780, '\p{^casedletter}', "");
Expect(1, 120780, '\P{casedletter}', "");
Expect(0, 120780, '\P{^casedletter}', "");
Expect(1, 120779, '\p{ _Cased_letter}', "");
Expect(0, 120779, '\p{^ _Cased_letter}', "");
Expect(0, 120779, '\P{ _Cased_letter}', "");
Expect(1, 120779, '\P{^ _Cased_letter}', "");
Expect(0, 120780, '\p{ _Cased_letter}', "");
Expect(1, 120780, '\p{^ _Cased_letter}', "");
Expect(1, 120780, '\P{ _Cased_letter}', "");
Expect(0, 120780, '\P{^ _Cased_letter}', "");
Error('\p{_ IS_Cased_Letter:=}');
Error('\P{_ IS_Cased_Letter:=}');
Expect(1, 120779, '\p{iscasedletter}', "");
Expect(0, 120779, '\p{^iscasedletter}', "");
Expect(0, 120779, '\P{iscasedletter}', "");
Expect(1, 120779, '\P{^iscasedletter}', "");
Expect(0, 120780, '\p{iscasedletter}', "");
Expect(1, 120780, '\p{^iscasedletter}', "");
Expect(1, 120780, '\P{iscasedletter}', "");
Expect(0, 120780, '\P{^iscasedletter}', "");
Expect(1, 120779, '\p{IS_cased_Letter}', "");
Expect(0, 120779, '\p{^IS_cased_Letter}', "");
Expect(0, 120779, '\P{IS_cased_Letter}', "");
Expect(1, 120779, '\P{^IS_cased_Letter}', "");
Expect(0, 120780, '\p{IS_cased_Letter}', "");
Expect(1, 120780, '\p{^IS_cased_Letter}', "");
Expect(1, 120780, '\P{IS_cased_Letter}', "");
Expect(0, 120780, '\P{^IS_cased_Letter}', "");
Error('\p{/a/ LC}');
Error('\P{/a/ LC}');
Expect(1, 120779, '\p{lc}', "");
Expect(0, 120779, '\p{^lc}', "");
Expect(0, 120779, '\P{lc}', "");
Expect(1, 120779, '\P{^lc}', "");
Expect(0, 120780, '\p{lc}', "");
Expect(1, 120780, '\p{^lc}', "");
Expect(1, 120780, '\P{lc}', "");
Expect(0, 120780, '\P{^lc}', "");
Expect(1, 120779, '\p{-LC}', "");
Expect(0, 120779, '\p{^-LC}', "");
Expect(0, 120779, '\P{-LC}', "");
Expect(1, 120779, '\P{^-LC}', "");
Expect(0, 120780, '\p{-LC}', "");
Expect(1, 120780, '\p{^-LC}', "");
Expect(1, 120780, '\P{-LC}', "");
Expect(0, 120780, '\P{^-LC}', "");
Error('\p{_	is_LC/a/}');
Error('\P{_	is_LC/a/}');
Expect(1, 120779, '\p{islc}', "");
Expect(0, 120779, '\p{^islc}', "");
Expect(0, 120779, '\P{islc}', "");
Expect(1, 120779, '\P{^islc}', "");
Expect(0, 120780, '\p{islc}', "");
Expect(1, 120780, '\p{^islc}', "");
Expect(1, 120780, '\P{islc}', "");
Expect(0, 120780, '\P{^islc}', "");
Expect(1, 120779, '\p{ _IS_lc}', "");
Expect(0, 120779, '\p{^ _IS_lc}', "");
Expect(0, 120779, '\P{ _IS_lc}', "");
Expect(1, 120779, '\P{^ _IS_lc}', "");
Expect(0, 120780, '\p{ _IS_lc}', "");
Expect(1, 120780, '\p{^ _IS_lc}', "");
Expect(1, 120780, '\P{ _IS_lc}', "");
Expect(0, 120780, '\P{^ _IS_lc}', "");
Error('\p{:=-_L_}');
Error('\P{:=-_L_}');
Expect(1, 120779, '\p{l_}', "");
Expect(0, 120779, '\p{^l_}', "");
Expect(0, 120779, '\P{l_}', "");
Expect(1, 120779, '\P{^l_}', "");
Expect(0, 120780, '\p{l_}', "");
Expect(1, 120780, '\p{^l_}', "");
Expect(1, 120780, '\P{l_}', "");
Expect(0, 120780, '\P{^l_}', "");
Expect(1, 120779, '\p{_L_}', "");
Expect(0, 120779, '\p{^_L_}', "");
Expect(0, 120779, '\P{_L_}', "");
Expect(1, 120779, '\P{^_L_}', "");
Expect(0, 120780, '\p{_L_}', "");
Expect(1, 120780, '\p{^_L_}', "");
Expect(1, 120780, '\P{_L_}', "");
Expect(0, 120780, '\P{^_L_}', "");
Error('\p{ _Is_L_:=}');
Error('\P{ _Is_L_:=}');
Expect(1, 120779, '\p{isl_}', "");
Expect(0, 120779, '\p{^isl_}', "");
Expect(0, 120779, '\P{isl_}', "");
Expect(1, 120779, '\P{^isl_}', "");
Expect(0, 120780, '\p{isl_}', "");
Expect(1, 120780, '\p{^isl_}', "");
Expect(1, 120780, '\P{isl_}', "");
Expect(0, 120780, '\P{^isl_}', "");
Expect(1, 120779, '\p{- Is_L_}', "");
Expect(0, 120779, '\p{^- Is_L_}', "");
Expect(0, 120779, '\P{- Is_L_}', "");
Expect(1, 120779, '\P{^- Is_L_}', "");
Expect(0, 120780, '\p{- Is_L_}', "");
Expect(1, 120780, '\p{^- Is_L_}', "");
Expect(1, 120780, '\P{- Is_L_}', "");
Expect(0, 120780, '\P{^- Is_L_}', "");
Error('\p{:=	 l&}');
Error('\P{:=	 l&}');
Expect(1, 120779, '\p{l&}', "");
Expect(0, 120779, '\p{^l&}', "");
Expect(0, 120779, '\P{l&}', "");
Expect(1, 120779, '\P{^l&}', "");
Expect(0, 120780, '\p{l&}', "");
Expect(1, 120780, '\p{^l&}', "");
Expect(1, 120780, '\P{l&}', "");
Expect(0, 120780, '\P{^l&}', "");
Expect(1, 120779, '\p{ L&}', "");
Expect(0, 120779, '\p{^ L&}', "");
Expect(0, 120779, '\P{ L&}', "");
Expect(1, 120779, '\P{^ L&}', "");
Expect(0, 120780, '\p{ L&}', "");
Expect(1, 120780, '\p{^ L&}', "");
Expect(1, 120780, '\P{ L&}', "");
Expect(0, 120780, '\P{^ L&}', "");
Error('\p{-_Is_L&/a/}');
Error('\P{-_Is_L&/a/}');
Expect(1, 120779, '\p{isl&}', "");
Expect(0, 120779, '\p{^isl&}', "");
Expect(0, 120779, '\P{isl&}', "");
Expect(1, 120779, '\P{^isl&}', "");
Expect(0, 120780, '\p{isl&}', "");
Expect(1, 120780, '\p{^isl&}', "");
Expect(1, 120780, '\P{isl&}', "");
Expect(0, 120780, '\P{^isl&}', "");
Expect(1, 120779, '\p{-IS_L&}', "");
Expect(0, 120779, '\p{^-IS_L&}', "");
Expect(0, 120779, '\P{-IS_L&}', "");
Expect(1, 120779, '\P{^-IS_L&}', "");
Expect(0, 120780, '\p{-IS_L&}', "");
Expect(1, 120780, '\p{^-IS_L&}', "");
Expect(1, 120780, '\P{-IS_L&}', "");
Expect(0, 120780, '\P{^-IS_L&}', "");
Error('\p{ 	chakma:=}');
Error('\P{ 	chakma:=}');
Expect(1, 69955, '\p{chakma}', "");
Expect(0, 69955, '\p{^chakma}', "");
Expect(0, 69955, '\P{chakma}', "");
Expect(1, 69955, '\P{^chakma}', "");
Expect(0, 69956, '\p{chakma}', "");
Expect(1, 69956, '\p{^chakma}', "");
Expect(1, 69956, '\P{chakma}', "");
Expect(0, 69956, '\P{^chakma}', "");
Expect(1, 69955, '\p{	chakma}', "");
Expect(0, 69955, '\p{^	chakma}', "");
Expect(0, 69955, '\P{	chakma}', "");
Expect(1, 69955, '\P{^	chakma}', "");
Expect(0, 69956, '\p{	chakma}', "");
Expect(1, 69956, '\p{^	chakma}', "");
Expect(1, 69956, '\P{	chakma}', "");
Expect(0, 69956, '\P{^	chakma}', "");
Error('\p{:=-	Is_Chakma}');
Error('\P{:=-	Is_Chakma}');
Expect(1, 69955, '\p{ischakma}', "");
Expect(0, 69955, '\p{^ischakma}', "");
Expect(0, 69955, '\P{ischakma}', "");
Expect(1, 69955, '\P{^ischakma}', "");
Expect(0, 69956, '\p{ischakma}', "");
Expect(1, 69956, '\p{^ischakma}', "");
Expect(1, 69956, '\P{ischakma}', "");
Expect(0, 69956, '\P{^ischakma}', "");
Expect(1, 69955, '\p{-	Is_chakma}', "");
Expect(0, 69955, '\p{^-	Is_chakma}', "");
Expect(0, 69955, '\P{-	Is_chakma}', "");
Expect(1, 69955, '\P{^-	Is_chakma}', "");
Expect(0, 69956, '\p{-	Is_chakma}', "");
Expect(1, 69956, '\p{^-	Is_chakma}', "");
Expect(1, 69956, '\P{-	Is_chakma}', "");
Expect(0, 69956, '\P{^-	Is_chakma}', "");
Error('\p{Cakm/a/}');
Error('\P{Cakm/a/}');
Expect(1, 69955, '\p{cakm}', "");
Expect(0, 69955, '\p{^cakm}', "");
Expect(0, 69955, '\P{cakm}', "");
Expect(1, 69955, '\P{^cakm}', "");
Expect(0, 69956, '\p{cakm}', "");
Expect(1, 69956, '\p{^cakm}', "");
Expect(1, 69956, '\P{cakm}', "");
Expect(0, 69956, '\P{^cakm}', "");
Expect(1, 69955, '\p{	 Cakm}', "");
Expect(0, 69955, '\p{^	 Cakm}', "");
Expect(0, 69955, '\P{	 Cakm}', "");
Expect(1, 69955, '\P{^	 Cakm}', "");
Expect(0, 69956, '\p{	 Cakm}', "");
Expect(1, 69956, '\p{^	 Cakm}', "");
Expect(1, 69956, '\P{	 Cakm}', "");
Expect(0, 69956, '\P{^	 Cakm}', "");
Error('\p{:=-	Is_CAKM}');
Error('\P{:=-	Is_CAKM}');
Expect(1, 69955, '\p{iscakm}', "");
Expect(0, 69955, '\p{^iscakm}', "");
Expect(0, 69955, '\P{iscakm}', "");
Expect(1, 69955, '\P{^iscakm}', "");
Expect(0, 69956, '\p{iscakm}', "");
Expect(1, 69956, '\p{^iscakm}', "");
Expect(1, 69956, '\P{iscakm}', "");
Expect(0, 69956, '\P{^iscakm}', "");
Expect(1, 69955, '\p{  is_CAKM}', "");
Expect(0, 69955, '\p{^  is_CAKM}', "");
Expect(0, 69955, '\P{  is_CAKM}', "");
Expect(1, 69955, '\P{^  is_CAKM}', "");
Expect(0, 69956, '\p{  is_CAKM}', "");
Expect(1, 69956, '\p{^  is_CAKM}', "");
Expect(1, 69956, '\P{  is_CAKM}', "");
Expect(0, 69956, '\P{^  is_CAKM}', "");
Error('\p{/a/	CHAM}');
Error('\P{/a/	CHAM}');
Expect(1, 43615, '\p{cham}', "");
Expect(0, 43615, '\p{^cham}', "");
Expect(0, 43615, '\P{cham}', "");
Expect(1, 43615, '\P{^cham}', "");
Expect(0, 43616, '\p{cham}', "");
Expect(1, 43616, '\p{^cham}', "");
Expect(1, 43616, '\P{cham}', "");
Expect(0, 43616, '\P{^cham}', "");
Expect(1, 43615, '\p{-_Cham}', "");
Expect(0, 43615, '\p{^-_Cham}', "");
Expect(0, 43615, '\P{-_Cham}', "");
Expect(1, 43615, '\P{^-_Cham}', "");
Expect(0, 43616, '\p{-_Cham}', "");
Expect(1, 43616, '\p{^-_Cham}', "");
Expect(1, 43616, '\P{-_Cham}', "");
Expect(0, 43616, '\P{^-_Cham}', "");
Error('\p{-	Is_cham/a/}');
Error('\P{-	Is_cham/a/}');
Expect(1, 43615, '\p{ischam}', "");
Expect(0, 43615, '\p{^ischam}', "");
Expect(0, 43615, '\P{ischam}', "");
Expect(1, 43615, '\P{^ischam}', "");
Expect(0, 43616, '\p{ischam}', "");
Expect(1, 43616, '\p{^ischam}', "");
Expect(1, 43616, '\P{ischam}', "");
Expect(0, 43616, '\P{^ischam}', "");
Expect(1, 43615, '\p{__IS_cham}', "");
Expect(0, 43615, '\p{^__IS_cham}', "");
Expect(0, 43615, '\P{__IS_cham}', "");
Expect(1, 43615, '\P{^__IS_cham}', "");
Expect(0, 43616, '\p{__IS_cham}', "");
Expect(1, 43616, '\p{^__IS_cham}', "");
Expect(1, 43616, '\P{__IS_cham}', "");
Expect(0, 43616, '\P{^__IS_cham}', "");
Error('\p{:=__changes_When_Casefolded}');
Error('\P{:=__changes_When_Casefolded}');
Expect(1, 66599, '\p{changeswhencasefolded}', "");
Expect(0, 66599, '\p{^changeswhencasefolded}', "");
Expect(0, 66599, '\P{changeswhencasefolded}', "");
Expect(1, 66599, '\P{^changeswhencasefolded}', "");
Expect(0, 66600, '\p{changeswhencasefolded}', "");
Expect(1, 66600, '\p{^changeswhencasefolded}', "");
Expect(1, 66600, '\P{changeswhencasefolded}', "");
Expect(0, 66600, '\P{^changeswhencasefolded}', "");
Expect(1, 66599, '\p{ _Changes_When_CASEFOLDED}', "");
Expect(0, 66599, '\p{^ _Changes_When_CASEFOLDED}', "");
Expect(0, 66599, '\P{ _Changes_When_CASEFOLDED}', "");
Expect(1, 66599, '\P{^ _Changes_When_CASEFOLDED}', "");
Expect(0, 66600, '\p{ _Changes_When_CASEFOLDED}', "");
Expect(1, 66600, '\p{^ _Changes_When_CASEFOLDED}', "");
Expect(1, 66600, '\P{ _Changes_When_CASEFOLDED}', "");
Expect(0, 66600, '\P{^ _Changes_When_CASEFOLDED}', "");
Error('\p{/a/ Is_Changes_WHEN_Casefolded}');
Error('\P{/a/ Is_Changes_WHEN_Casefolded}');
Expect(1, 66599, '\p{ischangeswhencasefolded}', "");
Expect(0, 66599, '\p{^ischangeswhencasefolded}', "");
Expect(0, 66599, '\P{ischangeswhencasefolded}', "");
Expect(1, 66599, '\P{^ischangeswhencasefolded}', "");
Expect(0, 66600, '\p{ischangeswhencasefolded}', "");
Expect(1, 66600, '\p{^ischangeswhencasefolded}', "");
Expect(1, 66600, '\P{ischangeswhencasefolded}', "");
Expect(0, 66600, '\P{^ischangeswhencasefolded}', "");
Expect(1, 66599, '\p{  Is_CHANGES_When_CASEFOLDED}', "");
Expect(0, 66599, '\p{^  Is_CHANGES_When_CASEFOLDED}', "");
Expect(0, 66599, '\P{  Is_CHANGES_When_CASEFOLDED}', "");
Expect(1, 66599, '\P{^  Is_CHANGES_When_CASEFOLDED}', "");
Expect(0, 66600, '\p{  Is_CHANGES_When_CASEFOLDED}', "");
Expect(1, 66600, '\p{^  Is_CHANGES_When_CASEFOLDED}', "");
Expect(1, 66600, '\P{  Is_CHANGES_When_CASEFOLDED}', "");
Expect(0, 66600, '\P{^  Is_CHANGES_When_CASEFOLDED}', "");
Error('\p{:=_CWCF}');
Error('\P{:=_CWCF}');
Expect(1, 66599, '\p{cwcf}', "");
Expect(0, 66599, '\p{^cwcf}', "");
Expect(0, 66599, '\P{cwcf}', "");
Expect(1, 66599, '\P{^cwcf}', "");
Expect(0, 66600, '\p{cwcf}', "");
Expect(1, 66600, '\p{^cwcf}', "");
Expect(1, 66600, '\P{cwcf}', "");
Expect(0, 66600, '\P{^cwcf}', "");
Expect(1, 66599, '\p{ CWCF}', "");
Expect(0, 66599, '\p{^ CWCF}', "");
Expect(0, 66599, '\P{ CWCF}', "");
Expect(1, 66599, '\P{^ CWCF}', "");
Expect(0, 66600, '\p{ CWCF}', "");
Expect(1, 66600, '\p{^ CWCF}', "");
Expect(1, 66600, '\P{ CWCF}', "");
Expect(0, 66600, '\P{^ CWCF}', "");
Error('\p{-	is_cwcf:=}');
Error('\P{-	is_cwcf:=}');
Expect(1, 66599, '\p{iscwcf}', "");
Expect(0, 66599, '\p{^iscwcf}', "");
Expect(0, 66599, '\P{iscwcf}', "");
Expect(1, 66599, '\P{^iscwcf}', "");
Expect(0, 66600, '\p{iscwcf}', "");
Expect(1, 66600, '\p{^iscwcf}', "");
Expect(1, 66600, '\P{iscwcf}', "");
Expect(0, 66600, '\P{^iscwcf}', "");
Expect(1, 66599, '\p{ Is_CWCF}', "");
Expect(0, 66599, '\p{^ Is_CWCF}', "");
Expect(0, 66599, '\P{ Is_CWCF}', "");
Expect(1, 66599, '\P{^ Is_CWCF}', "");
Expect(0, 66600, '\p{ Is_CWCF}', "");
Expect(1, 66600, '\p{^ Is_CWCF}', "");
Expect(1, 66600, '\P{ Is_CWCF}', "");
Expect(0, 66600, '\P{^ Is_CWCF}', "");
Error('\p{/a/_ Changes_When_casemapped}');
Error('\P{/a/_ Changes_When_casemapped}');
Expect(1, 66639, '\p{changeswhencasemapped}', "");
Expect(0, 66639, '\p{^changeswhencasemapped}', "");
Expect(0, 66639, '\P{changeswhencasemapped}', "");
Expect(1, 66639, '\P{^changeswhencasemapped}', "");
Expect(0, 66640, '\p{changeswhencasemapped}', "");
Expect(1, 66640, '\p{^changeswhencasemapped}', "");
Expect(1, 66640, '\P{changeswhencasemapped}', "");
Expect(0, 66640, '\P{^changeswhencasemapped}', "");
Expect(1, 66639, '\p{	_CHANGES_When_Casemapped}', "");
Expect(0, 66639, '\p{^	_CHANGES_When_Casemapped}', "");
Expect(0, 66639, '\P{	_CHANGES_When_Casemapped}', "");
Expect(1, 66639, '\P{^	_CHANGES_When_Casemapped}', "");
Expect(0, 66640, '\p{	_CHANGES_When_Casemapped}', "");
Expect(1, 66640, '\p{^	_CHANGES_When_Casemapped}', "");
Expect(1, 66640, '\P{	_CHANGES_When_Casemapped}', "");
Expect(0, 66640, '\P{^	_CHANGES_When_Casemapped}', "");
Error('\p{:=	_Is_CHANGES_When_Casemapped}');
Error('\P{:=	_Is_CHANGES_When_Casemapped}');
Expect(1, 66639, '\p{ischangeswhencasemapped}', "");
Expect(0, 66639, '\p{^ischangeswhencasemapped}', "");
Expect(0, 66639, '\P{ischangeswhencasemapped}', "");
Expect(1, 66639, '\P{^ischangeswhencasemapped}', "");
Expect(0, 66640, '\p{ischangeswhencasemapped}', "");
Expect(1, 66640, '\p{^ischangeswhencasemapped}', "");
Expect(1, 66640, '\P{ischangeswhencasemapped}', "");
Expect(0, 66640, '\P{^ischangeswhencasemapped}', "");
Expect(1, 66639, '\p{_ Is_changes_When_CASEMAPPED}', "");
Expect(0, 66639, '\p{^_ Is_changes_When_CASEMAPPED}', "");
Expect(0, 66639, '\P{_ Is_changes_When_CASEMAPPED}', "");
Expect(1, 66639, '\P{^_ Is_changes_When_CASEMAPPED}', "");
Expect(0, 66640, '\p{_ Is_changes_When_CASEMAPPED}', "");
Expect(1, 66640, '\p{^_ Is_changes_When_CASEMAPPED}', "");
Expect(1, 66640, '\P{_ Is_changes_When_CASEMAPPED}', "");
Expect(0, 66640, '\P{^_ Is_changes_When_CASEMAPPED}', "");
Error('\p{CWCM:=}');
Error('\P{CWCM:=}');
Expect(1, 66639, '\p{cwcm}', "");
Expect(0, 66639, '\p{^cwcm}', "");
Expect(0, 66639, '\P{cwcm}', "");
Expect(1, 66639, '\P{^cwcm}', "");
Expect(0, 66640, '\p{cwcm}', "");
Expect(1, 66640, '\p{^cwcm}', "");
Expect(1, 66640, '\P{cwcm}', "");
Expect(0, 66640, '\P{^cwcm}', "");
Expect(1, 66639, '\p{	cwcm}', "");
Expect(0, 66639, '\p{^	cwcm}', "");
Expect(0, 66639, '\P{	cwcm}', "");
Expect(1, 66639, '\P{^	cwcm}', "");
Expect(0, 66640, '\p{	cwcm}', "");
Expect(1, 66640, '\p{^	cwcm}', "");
Expect(1, 66640, '\P{	cwcm}', "");
Expect(0, 66640, '\P{^	cwcm}', "");
Error('\p{_:=Is_CWCM}');
Error('\P{_:=Is_CWCM}');
Expect(1, 66639, '\p{iscwcm}', "");
Expect(0, 66639, '\p{^iscwcm}', "");
Expect(0, 66639, '\P{iscwcm}', "");
Expect(1, 66639, '\P{^iscwcm}', "");
Expect(0, 66640, '\p{iscwcm}', "");
Expect(1, 66640, '\p{^iscwcm}', "");
Expect(1, 66640, '\P{iscwcm}', "");
Expect(0, 66640, '\P{^iscwcm}', "");
Expect(1, 66639, '\p{  Is_CWCM}', "");
Expect(0, 66639, '\p{^  Is_CWCM}', "");
Expect(0, 66639, '\P{  Is_CWCM}', "");
Expect(1, 66639, '\P{^  Is_CWCM}', "");
Expect(0, 66640, '\p{  Is_CWCM}', "");
Expect(1, 66640, '\p{^  Is_CWCM}', "");
Expect(1, 66640, '\P{  Is_CWCM}', "");
Expect(0, 66640, '\P{^  Is_CWCM}', "");
Error('\p{/a/ Changes_When_LOWERCASED}');
Error('\P{/a/ Changes_When_LOWERCASED}');
Expect(1, 66599, '\p{changeswhenlowercased}', "");
Expect(0, 66599, '\p{^changeswhenlowercased}', "");
Expect(0, 66599, '\P{changeswhenlowercased}', "");
Expect(1, 66599, '\P{^changeswhenlowercased}', "");
Expect(0, 66600, '\p{changeswhenlowercased}', "");
Expect(1, 66600, '\p{^changeswhenlowercased}', "");
Expect(1, 66600, '\P{changeswhenlowercased}', "");
Expect(0, 66600, '\P{^changeswhenlowercased}', "");
Expect(1, 66599, '\p{	changes_When_lowercased}', "");
Expect(0, 66599, '\p{^	changes_When_lowercased}', "");
Expect(0, 66599, '\P{	changes_When_lowercased}', "");
Expect(1, 66599, '\P{^	changes_When_lowercased}', "");
Expect(0, 66600, '\p{	changes_When_lowercased}', "");
Expect(1, 66600, '\p{^	changes_When_lowercased}', "");
Expect(1, 66600, '\P{	changes_When_lowercased}', "");
Expect(0, 66600, '\P{^	changes_When_lowercased}', "");
Error('\p{_is_Changes_When_LOWERCASED:=}');
Error('\P{_is_Changes_When_LOWERCASED:=}');
Expect(1, 66599, '\p{ischangeswhenlowercased}', "");
Expect(0, 66599, '\p{^ischangeswhenlowercased}', "");
Expect(0, 66599, '\P{ischangeswhenlowercased}', "");
Expect(1, 66599, '\P{^ischangeswhenlowercased}', "");
Expect(0, 66600, '\p{ischangeswhenlowercased}', "");
Expect(1, 66600, '\p{^ischangeswhenlowercased}', "");
Expect(1, 66600, '\P{ischangeswhenlowercased}', "");
Expect(0, 66600, '\P{^ischangeswhenlowercased}', "");
Expect(1, 66599, '\p{ IS_Changes_When_lowercased}', "");
Expect(0, 66599, '\p{^ IS_Changes_When_lowercased}', "");
Expect(0, 66599, '\P{ IS_Changes_When_lowercased}', "");
Expect(1, 66599, '\P{^ IS_Changes_When_lowercased}', "");
Expect(0, 66600, '\p{ IS_Changes_When_lowercased}', "");
Expect(1, 66600, '\p{^ IS_Changes_When_lowercased}', "");
Expect(1, 66600, '\P{ IS_Changes_When_lowercased}', "");
Expect(0, 66600, '\P{^ IS_Changes_When_lowercased}', "");
Error('\p{ :=CWL}');
Error('\P{ :=CWL}');
Expect(1, 66599, '\p{cwl}', "");
Expect(0, 66599, '\p{^cwl}', "");
Expect(0, 66599, '\P{cwl}', "");
Expect(1, 66599, '\P{^cwl}', "");
Expect(0, 66600, '\p{cwl}', "");
Expect(1, 66600, '\p{^cwl}', "");
Expect(1, 66600, '\P{cwl}', "");
Expect(0, 66600, '\P{^cwl}', "");
Expect(1, 66599, '\p{ CWL}', "");
Expect(0, 66599, '\p{^ CWL}', "");
Expect(0, 66599, '\P{ CWL}', "");
Expect(1, 66599, '\P{^ CWL}', "");
Expect(0, 66600, '\p{ CWL}', "");
Expect(1, 66600, '\p{^ CWL}', "");
Expect(1, 66600, '\P{ CWL}', "");
Expect(0, 66600, '\P{^ CWL}', "");
Error('\p{_is_CWL:=}');
Error('\P{_is_CWL:=}');
Expect(1, 66599, '\p{iscwl}', "");
Expect(0, 66599, '\p{^iscwl}', "");
Expect(0, 66599, '\P{iscwl}', "");
Expect(1, 66599, '\P{^iscwl}', "");
Expect(0, 66600, '\p{iscwl}', "");
Expect(1, 66600, '\p{^iscwl}', "");
Expect(1, 66600, '\P{iscwl}', "");
Expect(0, 66600, '\P{^iscwl}', "");
Expect(1, 66599, '\p{-IS_CWL}', "");
Expect(0, 66599, '\p{^-IS_CWL}', "");
Expect(0, 66599, '\P{-IS_CWL}', "");
Expect(1, 66599, '\P{^-IS_CWL}', "");
Expect(0, 66600, '\p{-IS_CWL}', "");
Expect(1, 66600, '\p{^-IS_CWL}', "");
Expect(1, 66600, '\P{-IS_CWL}', "");
Expect(0, 66600, '\P{^-IS_CWL}', "");
Error('\p{-CHANGES_When_NFKC_CASEFOLDED/a/}');
Error('\P{-CHANGES_When_NFKC_CASEFOLDED/a/}');
Expect(1, 921599, '\p{changeswhennfkccasefolded}', "");
Expect(0, 921599, '\p{^changeswhennfkccasefolded}', "");
Expect(0, 921599, '\P{changeswhennfkccasefolded}', "");
Expect(1, 921599, '\P{^changeswhennfkccasefolded}', "");
Expect(0, 921600, '\p{changeswhennfkccasefolded}', "");
Expect(1, 921600, '\p{^changeswhennfkccasefolded}', "");
Expect(1, 921600, '\P{changeswhennfkccasefolded}', "");
Expect(0, 921600, '\P{^changeswhennfkccasefolded}', "");
Expect(1, 921599, '\p{- changes_When_nfkc_CASEFOLDED}', "");
Expect(0, 921599, '\p{^- changes_When_nfkc_CASEFOLDED}', "");
Expect(0, 921599, '\P{- changes_When_nfkc_CASEFOLDED}', "");
Expect(1, 921599, '\P{^- changes_When_nfkc_CASEFOLDED}', "");
Expect(0, 921600, '\p{- changes_When_nfkc_CASEFOLDED}', "");
Expect(1, 921600, '\p{^- changes_When_nfkc_CASEFOLDED}', "");
Expect(1, 921600, '\P{- changes_When_nfkc_CASEFOLDED}', "");
Expect(0, 921600, '\P{^- changes_When_nfkc_CASEFOLDED}', "");
Error('\p{/a/- Is_CHANGES_WHEN_NFKC_casefolded}');
Error('\P{/a/- Is_CHANGES_WHEN_NFKC_casefolded}');
Expect(1, 921599, '\p{ischangeswhennfkccasefolded}', "");
Expect(0, 921599, '\p{^ischangeswhennfkccasefolded}', "");
Expect(0, 921599, '\P{ischangeswhennfkccasefolded}', "");
Expect(1, 921599, '\P{^ischangeswhennfkccasefolded}', "");
Expect(0, 921600, '\p{ischangeswhennfkccasefolded}', "");
Expect(1, 921600, '\p{^ischangeswhennfkccasefolded}', "");
Expect(1, 921600, '\P{ischangeswhennfkccasefolded}', "");
Expect(0, 921600, '\P{^ischangeswhennfkccasefolded}', "");
Expect(1, 921599, '\p{	 Is_changes_WHEN_NFKC_Casefolded}', "");
Expect(0, 921599, '\p{^	 Is_changes_WHEN_NFKC_Casefolded}', "");
Expect(0, 921599, '\P{	 Is_changes_WHEN_NFKC_Casefolded}', "");
Expect(1, 921599, '\P{^	 Is_changes_WHEN_NFKC_Casefolded}', "");
Expect(0, 921600, '\p{	 Is_changes_WHEN_NFKC_Casefolded}', "");
Expect(1, 921600, '\p{^	 Is_changes_WHEN_NFKC_Casefolded}', "");
Expect(1, 921600, '\P{	 Is_changes_WHEN_NFKC_Casefolded}', "");
Expect(0, 921600, '\P{^	 Is_changes_WHEN_NFKC_Casefolded}', "");
Error('\p{/a/ 	CWKCF}');
Error('\P{/a/ 	CWKCF}');
Expect(1, 921599, '\p{cwkcf}', "");
Expect(0, 921599, '\p{^cwkcf}', "");
Expect(0, 921599, '\P{cwkcf}', "");
Expect(1, 921599, '\P{^cwkcf}', "");
Expect(0, 921600, '\p{cwkcf}', "");
Expect(1, 921600, '\p{^cwkcf}', "");
Expect(1, 921600, '\P{cwkcf}', "");
Expect(0, 921600, '\P{^cwkcf}', "");
Expect(1, 921599, '\p{	CWKCF}', "");
Expect(0, 921599, '\p{^	CWKCF}', "");
Expect(0, 921599, '\P{	CWKCF}', "");
Expect(1, 921599, '\P{^	CWKCF}', "");
Expect(0, 921600, '\p{	CWKCF}', "");
Expect(1, 921600, '\p{^	CWKCF}', "");
Expect(1, 921600, '\P{	CWKCF}', "");
Expect(0, 921600, '\P{^	CWKCF}', "");
Error('\p{_:=Is_CWKCF}');
Error('\P{_:=Is_CWKCF}');
Expect(1, 921599, '\p{iscwkcf}', "");
Expect(0, 921599, '\p{^iscwkcf}', "");
Expect(0, 921599, '\P{iscwkcf}', "");
Expect(1, 921599, '\P{^iscwkcf}', "");
Expect(0, 921600, '\p{iscwkcf}', "");
Expect(1, 921600, '\p{^iscwkcf}', "");
Expect(1, 921600, '\P{iscwkcf}', "");
Expect(0, 921600, '\P{^iscwkcf}', "");
Expect(1, 921599, '\p{-_is_CWKCF}', "");
Expect(0, 921599, '\p{^-_is_CWKCF}', "");
Expect(0, 921599, '\P{-_is_CWKCF}', "");
Expect(1, 921599, '\P{^-_is_CWKCF}', "");
Expect(0, 921600, '\p{-_is_CWKCF}', "");
Expect(1, 921600, '\p{^-_is_CWKCF}', "");
Expect(1, 921600, '\P{-_is_CWKCF}', "");
Expect(0, 921600, '\P{^-_is_CWKCF}', "");
Error('\p{ -Changes_When_Titlecased/a/}');
Error('\P{ -Changes_When_Titlecased/a/}');
Expect(1, 66639, '\p{changeswhentitlecased}', "");
Expect(0, 66639, '\p{^changeswhentitlecased}', "");
Expect(0, 66639, '\P{changeswhentitlecased}', "");
Expect(1, 66639, '\P{^changeswhentitlecased}', "");
Expect(0, 66640, '\p{changeswhentitlecased}', "");
Expect(1, 66640, '\p{^changeswhentitlecased}', "");
Expect(1, 66640, '\P{changeswhentitlecased}', "");
Expect(0, 66640, '\P{^changeswhentitlecased}', "");
Expect(1, 66639, '\p{__Changes_When_Titlecased}', "");
Expect(0, 66639, '\p{^__Changes_When_Titlecased}', "");
Expect(0, 66639, '\P{__Changes_When_Titlecased}', "");
Expect(1, 66639, '\P{^__Changes_When_Titlecased}', "");
Expect(0, 66640, '\p{__Changes_When_Titlecased}', "");
Expect(1, 66640, '\p{^__Changes_When_Titlecased}', "");
Expect(1, 66640, '\P{__Changes_When_Titlecased}', "");
Expect(0, 66640, '\P{^__Changes_When_Titlecased}', "");
Error('\p{	/a/IS_changes_When_Titlecased}');
Error('\P{	/a/IS_changes_When_Titlecased}');
Expect(1, 66639, '\p{ischangeswhentitlecased}', "");
Expect(0, 66639, '\p{^ischangeswhentitlecased}', "");
Expect(0, 66639, '\P{ischangeswhentitlecased}', "");
Expect(1, 66639, '\P{^ischangeswhentitlecased}', "");
Expect(0, 66640, '\p{ischangeswhentitlecased}', "");
Expect(1, 66640, '\p{^ischangeswhentitlecased}', "");
Expect(1, 66640, '\P{ischangeswhentitlecased}', "");
Expect(0, 66640, '\P{^ischangeswhentitlecased}', "");
Expect(1, 66639, '\p{_Is_Changes_when_Titlecased}', "");
Expect(0, 66639, '\p{^_Is_Changes_when_Titlecased}', "");
Expect(0, 66639, '\P{_Is_Changes_when_Titlecased}', "");
Expect(1, 66639, '\P{^_Is_Changes_when_Titlecased}', "");
Expect(0, 66640, '\p{_Is_Changes_when_Titlecased}', "");
Expect(1, 66640, '\p{^_Is_Changes_when_Titlecased}', "");
Expect(1, 66640, '\P{_Is_Changes_when_Titlecased}', "");
Expect(0, 66640, '\P{^_Is_Changes_when_Titlecased}', "");
Error('\p{-:=CWT}');
Error('\P{-:=CWT}');
Expect(1, 66639, '\p{cwt}', "");
Expect(0, 66639, '\p{^cwt}', "");
Expect(0, 66639, '\P{cwt}', "");
Expect(1, 66639, '\P{^cwt}', "");
Expect(0, 66640, '\p{cwt}', "");
Expect(1, 66640, '\p{^cwt}', "");
Expect(1, 66640, '\P{cwt}', "");
Expect(0, 66640, '\P{^cwt}', "");
Expect(1, 66639, '\p{--CWT}', "");
Expect(0, 66639, '\p{^--CWT}', "");
Expect(0, 66639, '\P{--CWT}', "");
Expect(1, 66639, '\P{^--CWT}', "");
Expect(0, 66640, '\p{--CWT}', "");
Expect(1, 66640, '\p{^--CWT}', "");
Expect(1, 66640, '\P{--CWT}', "");
Expect(0, 66640, '\P{^--CWT}', "");
Error('\p{:=	-Is_cwt}');
Error('\P{:=	-Is_cwt}');
Expect(1, 66639, '\p{iscwt}', "");
Expect(0, 66639, '\p{^iscwt}', "");
Expect(0, 66639, '\P{iscwt}', "");
Expect(1, 66639, '\P{^iscwt}', "");
Expect(0, 66640, '\p{iscwt}', "");
Expect(1, 66640, '\p{^iscwt}', "");
Expect(1, 66640, '\P{iscwt}', "");
Expect(0, 66640, '\P{^iscwt}', "");
Expect(1, 66639, '\p{		Is_CWT}', "");
Expect(0, 66639, '\p{^		Is_CWT}', "");
Expect(0, 66639, '\P{		Is_CWT}', "");
Expect(1, 66639, '\P{^		Is_CWT}', "");
Expect(0, 66640, '\p{		Is_CWT}', "");
Expect(1, 66640, '\p{^		Is_CWT}', "");
Expect(1, 66640, '\P{		Is_CWT}', "");
Expect(0, 66640, '\P{^		Is_CWT}', "");
Error('\p{_:=Changes_WHEN_uppercased}');
Error('\P{_:=Changes_WHEN_uppercased}');
Expect(1, 66639, '\p{changeswhenuppercased}', "");
Expect(0, 66639, '\p{^changeswhenuppercased}', "");
Expect(0, 66639, '\P{changeswhenuppercased}', "");
Expect(1, 66639, '\P{^changeswhenuppercased}', "");
Expect(0, 66640, '\p{changeswhenuppercased}', "");
Expect(1, 66640, '\p{^changeswhenuppercased}', "");
Expect(1, 66640, '\P{changeswhenuppercased}', "");
Expect(0, 66640, '\P{^changeswhenuppercased}', "");
Expect(1, 66639, '\p{_ changes_when_uppercased}', "");
Expect(0, 66639, '\p{^_ changes_when_uppercased}', "");
Expect(0, 66639, '\P{_ changes_when_uppercased}', "");
Expect(1, 66639, '\P{^_ changes_when_uppercased}', "");
Expect(0, 66640, '\p{_ changes_when_uppercased}', "");
Expect(1, 66640, '\p{^_ changes_when_uppercased}', "");
Expect(1, 66640, '\P{_ changes_when_uppercased}', "");
Expect(0, 66640, '\P{^_ changes_when_uppercased}', "");
Error('\p{ :=Is_changes_when_Uppercased}');
Error('\P{ :=Is_changes_when_Uppercased}');
Expect(1, 66639, '\p{ischangeswhenuppercased}', "");
Expect(0, 66639, '\p{^ischangeswhenuppercased}', "");
Expect(0, 66639, '\P{ischangeswhenuppercased}', "");
Expect(1, 66639, '\P{^ischangeswhenuppercased}', "");
Expect(0, 66640, '\p{ischangeswhenuppercased}', "");
Expect(1, 66640, '\p{^ischangeswhenuppercased}', "");
Expect(1, 66640, '\P{ischangeswhenuppercased}', "");
Expect(0, 66640, '\P{^ischangeswhenuppercased}', "");
Expect(1, 66639, '\p{ 	Is_Changes_When_Uppercased}', "");
Expect(0, 66639, '\p{^ 	Is_Changes_When_Uppercased}', "");
Expect(0, 66639, '\P{ 	Is_Changes_When_Uppercased}', "");
Expect(1, 66639, '\P{^ 	Is_Changes_When_Uppercased}', "");
Expect(0, 66640, '\p{ 	Is_Changes_When_Uppercased}', "");
Expect(1, 66640, '\p{^ 	Is_Changes_When_Uppercased}', "");
Expect(1, 66640, '\P{ 	Is_Changes_When_Uppercased}', "");
Expect(0, 66640, '\P{^ 	Is_Changes_When_Uppercased}', "");
Error('\p{ CWU:=}');
Error('\P{ CWU:=}');
Expect(1, 66639, '\p{cwu}', "");
Expect(0, 66639, '\p{^cwu}', "");
Expect(0, 66639, '\P{cwu}', "");
Expect(1, 66639, '\P{^cwu}', "");
Expect(0, 66640, '\p{cwu}', "");
Expect(1, 66640, '\p{^cwu}', "");
Expect(1, 66640, '\P{cwu}', "");
Expect(0, 66640, '\P{^cwu}', "");
Expect(1, 66639, '\p{- CWU}', "");
Expect(0, 66639, '\p{^- CWU}', "");
Expect(0, 66639, '\P{- CWU}', "");
Expect(1, 66639, '\P{^- CWU}', "");
Expect(0, 66640, '\p{- CWU}', "");
Expect(1, 66640, '\p{^- CWU}', "");
Expect(1, 66640, '\P{- CWU}', "");
Expect(0, 66640, '\P{^- CWU}', "");
Error('\p{_Is_cwu:=}');
Error('\P{_Is_cwu:=}');
Expect(1, 66639, '\p{iscwu}', "");
Expect(0, 66639, '\p{^iscwu}', "");
Expect(0, 66639, '\P{iscwu}', "");
Expect(1, 66639, '\P{^iscwu}', "");
Expect(0, 66640, '\p{iscwu}', "");
Expect(1, 66640, '\p{^iscwu}', "");
Expect(1, 66640, '\P{iscwu}', "");
Expect(0, 66640, '\P{^iscwu}', "");
Expect(1, 66639, '\p{ -Is_CWU}', "");
Expect(0, 66639, '\p{^ -Is_CWU}', "");
Expect(0, 66639, '\P{ -Is_CWU}', "");
Expect(1, 66639, '\P{^ -Is_CWU}', "");
Expect(0, 66640, '\p{ -Is_CWU}', "");
Expect(1, 66640, '\p{^ -Is_CWU}', "");
Expect(1, 66640, '\P{ -Is_CWU}', "");
Expect(0, 66640, '\P{^ -Is_CWU}', "");
Error('\p{	-cherokee/a/}');
Error('\P{	-cherokee/a/}');
Expect(1, 5108, '\p{cherokee}', "");
Expect(0, 5108, '\p{^cherokee}', "");
Expect(0, 5108, '\P{cherokee}', "");
Expect(1, 5108, '\P{^cherokee}', "");
Expect(0, 5109, '\p{cherokee}', "");
Expect(1, 5109, '\p{^cherokee}', "");
Expect(1, 5109, '\P{cherokee}', "");
Expect(0, 5109, '\P{^cherokee}', "");
Expect(1, 5108, '\p{_cherokee}', "");
Expect(0, 5108, '\p{^_cherokee}', "");
Expect(0, 5108, '\P{_cherokee}', "");
Expect(1, 5108, '\P{^_cherokee}', "");
Expect(0, 5109, '\p{_cherokee}', "");
Expect(1, 5109, '\p{^_cherokee}', "");
Expect(1, 5109, '\P{_cherokee}', "");
Expect(0, 5109, '\P{^_cherokee}', "");
Error('\p{/a/ 	IS_CHEROKEE}');
Error('\P{/a/ 	IS_CHEROKEE}');
Expect(1, 5108, '\p{ischerokee}', "");
Expect(0, 5108, '\p{^ischerokee}', "");
Expect(0, 5108, '\P{ischerokee}', "");
Expect(1, 5108, '\P{^ischerokee}', "");
Expect(0, 5109, '\p{ischerokee}', "");
Expect(1, 5109, '\p{^ischerokee}', "");
Expect(1, 5109, '\P{ischerokee}', "");
Expect(0, 5109, '\P{^ischerokee}', "");
Expect(1, 5108, '\p{__Is_Cherokee}', "");
Expect(0, 5108, '\p{^__Is_Cherokee}', "");
Expect(0, 5108, '\P{__Is_Cherokee}', "");
Expect(1, 5108, '\P{^__Is_Cherokee}', "");
Expect(0, 5109, '\p{__Is_Cherokee}', "");
Expect(1, 5109, '\p{^__Is_Cherokee}', "");
Expect(1, 5109, '\P{__Is_Cherokee}', "");
Expect(0, 5109, '\P{^__Is_Cherokee}', "");
Error('\p{ Cher:=}');
Error('\P{ Cher:=}');
Expect(1, 5108, '\p{cher}', "");
Expect(0, 5108, '\p{^cher}', "");
Expect(0, 5108, '\P{cher}', "");
Expect(1, 5108, '\P{^cher}', "");
Expect(0, 5109, '\p{cher}', "");
Expect(1, 5109, '\p{^cher}', "");
Expect(1, 5109, '\P{cher}', "");
Expect(0, 5109, '\P{^cher}', "");
Expect(1, 5108, '\p{ _Cher}', "");
Expect(0, 5108, '\p{^ _Cher}', "");
Expect(0, 5108, '\P{ _Cher}', "");
Expect(1, 5108, '\P{^ _Cher}', "");
Expect(0, 5109, '\p{ _Cher}', "");
Expect(1, 5109, '\p{^ _Cher}', "");
Expect(1, 5109, '\P{ _Cher}', "");
Expect(0, 5109, '\P{^ _Cher}', "");
Error('\p{:=- Is_cher}');
Error('\P{:=- Is_cher}');
Expect(1, 5108, '\p{ischer}', "");
Expect(0, 5108, '\p{^ischer}', "");
Expect(0, 5108, '\P{ischer}', "");
Expect(1, 5108, '\P{^ischer}', "");
Expect(0, 5109, '\p{ischer}', "");
Expect(1, 5109, '\p{^ischer}', "");
Expect(1, 5109, '\P{ischer}', "");
Expect(0, 5109, '\P{^ischer}', "");
Expect(1, 5108, '\p{  IS_Cher}', "");
Expect(0, 5108, '\p{^  IS_Cher}', "");
Expect(0, 5108, '\P{  IS_Cher}', "");
Expect(1, 5108, '\P{^  IS_Cher}', "");
Expect(0, 5109, '\p{  IS_Cher}', "");
Expect(1, 5109, '\p{^  IS_Cher}', "");
Expect(1, 5109, '\P{  IS_Cher}', "");
Expect(0, 5109, '\P{^  IS_Cher}', "");
Error('\p{_	Close_PUNCTUATION:=}');
Error('\P{_	Close_PUNCTUATION:=}');
Expect(1, 65379, '\p{closepunctuation}', "");
Expect(0, 65379, '\p{^closepunctuation}', "");
Expect(0, 65379, '\P{closepunctuation}', "");
Expect(1, 65379, '\P{^closepunctuation}', "");
Expect(0, 65380, '\p{closepunctuation}', "");
Expect(1, 65380, '\p{^closepunctuation}', "");
Expect(1, 65380, '\P{closepunctuation}', "");
Expect(0, 65380, '\P{^closepunctuation}', "");
Expect(1, 65379, '\p{ 	Close_punctuation}', "");
Expect(0, 65379, '\p{^ 	Close_punctuation}', "");
Expect(0, 65379, '\P{ 	Close_punctuation}', "");
Expect(1, 65379, '\P{^ 	Close_punctuation}', "");
Expect(0, 65380, '\p{ 	Close_punctuation}', "");
Expect(1, 65380, '\p{^ 	Close_punctuation}', "");
Expect(1, 65380, '\P{ 	Close_punctuation}', "");
Expect(0, 65380, '\P{^ 	Close_punctuation}', "");
Error('\p{/a/_ is_Close_PUNCTUATION}');
Error('\P{/a/_ is_Close_PUNCTUATION}');
Expect(1, 65379, '\p{isclosepunctuation}', "");
Expect(0, 65379, '\p{^isclosepunctuation}', "");
Expect(0, 65379, '\P{isclosepunctuation}', "");
Expect(1, 65379, '\P{^isclosepunctuation}', "");
Expect(0, 65380, '\p{isclosepunctuation}', "");
Expect(1, 65380, '\p{^isclosepunctuation}', "");
Expect(1, 65380, '\P{isclosepunctuation}', "");
Expect(0, 65380, '\P{^isclosepunctuation}', "");
Expect(1, 65379, '\p{--is_CLOSE_punctuation}', "");
Expect(0, 65379, '\p{^--is_CLOSE_punctuation}', "");
Expect(0, 65379, '\P{--is_CLOSE_punctuation}', "");
Expect(1, 65379, '\P{^--is_CLOSE_punctuation}', "");
Expect(0, 65380, '\p{--is_CLOSE_punctuation}', "");
Expect(1, 65380, '\p{^--is_CLOSE_punctuation}', "");
Expect(1, 65380, '\P{--is_CLOSE_punctuation}', "");
Expect(0, 65380, '\P{^--is_CLOSE_punctuation}', "");
Error('\p{-:=PE}');
Error('\P{-:=PE}');
Expect(1, 65379, '\p{pe}', "");
Expect(0, 65379, '\p{^pe}', "");
Expect(0, 65379, '\P{pe}', "");
Expect(1, 65379, '\P{^pe}', "");
Expect(0, 65380, '\p{pe}', "");
Expect(1, 65380, '\p{^pe}', "");
Expect(1, 65380, '\P{pe}', "");
Expect(0, 65380, '\P{^pe}', "");
Expect(1, 65379, '\p{ Pe}', "");
Expect(0, 65379, '\p{^ Pe}', "");
Expect(0, 65379, '\P{ Pe}', "");
Expect(1, 65379, '\P{^ Pe}', "");
Expect(0, 65380, '\p{ Pe}', "");
Expect(1, 65380, '\p{^ Pe}', "");
Expect(1, 65380, '\P{ Pe}', "");
Expect(0, 65380, '\P{^ Pe}', "");
Error('\p{:=is_pe}');
Error('\P{:=is_pe}');
Expect(1, 65379, '\p{ispe}', "");
Expect(0, 65379, '\p{^ispe}', "");
Expect(0, 65379, '\P{ispe}', "");
Expect(1, 65379, '\P{^ispe}', "");
Expect(0, 65380, '\p{ispe}', "");
Expect(1, 65380, '\p{^ispe}', "");
Expect(1, 65380, '\P{ispe}', "");
Expect(0, 65380, '\P{^ispe}', "");
Expect(1, 65379, '\p{		Is_Pe}', "");
Expect(0, 65379, '\p{^		Is_Pe}', "");
Expect(0, 65379, '\P{		Is_Pe}', "");
Expect(1, 65379, '\P{^		Is_Pe}', "");
Expect(0, 65380, '\p{		Is_Pe}', "");
Expect(1, 65380, '\p{^		Is_Pe}', "");
Expect(1, 65380, '\P{		Is_Pe}', "");
Expect(0, 65380, '\P{^		Is_Pe}', "");
Error('\p{:=cntrl}');
Error('\P{:=cntrl}');
Expect(1, 159, '\p{cntrl}', "");
Expect(0, 159, '\p{^cntrl}', "");
Expect(0, 159, '\P{cntrl}', "");
Expect(1, 159, '\P{^cntrl}', "");
Expect(0, 160, '\p{cntrl}', "");
Expect(1, 160, '\p{^cntrl}', "");
Expect(1, 160, '\P{cntrl}', "");
Expect(0, 160, '\P{^cntrl}', "");
Expect(1, 159, '\p{ 	CNTRL}', "");
Expect(0, 159, '\p{^ 	CNTRL}', "");
Expect(0, 159, '\P{ 	CNTRL}', "");
Expect(1, 159, '\P{^ 	CNTRL}', "");
Expect(0, 160, '\p{ 	CNTRL}', "");
Expect(1, 160, '\p{^ 	CNTRL}', "");
Expect(1, 160, '\P{ 	CNTRL}', "");
Expect(0, 160, '\P{^ 	CNTRL}', "");
Error('\p{-_XPosixCntrl:=}');
Error('\P{-_XPosixCntrl:=}');
Expect(1, 159, '\p{xposixcntrl}', "");
Expect(0, 159, '\p{^xposixcntrl}', "");
Expect(0, 159, '\P{xposixcntrl}', "");
Expect(1, 159, '\P{^xposixcntrl}', "");
Expect(0, 160, '\p{xposixcntrl}', "");
Expect(1, 160, '\p{^xposixcntrl}', "");
Expect(1, 160, '\P{xposixcntrl}', "");
Expect(0, 160, '\P{^xposixcntrl}', "");
Expect(1, 159, '\p{--XPOSIXCNTRL}', "");
Expect(0, 159, '\p{^--XPOSIXCNTRL}', "");
Expect(0, 159, '\P{--XPOSIXCNTRL}', "");
Expect(1, 159, '\P{^--XPOSIXCNTRL}', "");
Expect(0, 160, '\p{--XPOSIXCNTRL}', "");
Expect(1, 160, '\p{^--XPOSIXCNTRL}', "");
Expect(1, 160, '\P{--XPOSIXCNTRL}', "");
Expect(0, 160, '\P{^--XPOSIXCNTRL}', "");
Error('\p{:=Is_Cntrl}');
Error('\P{:=Is_Cntrl}');
Expect(1, 159, '\p{iscntrl}', "");
Expect(0, 159, '\p{^iscntrl}', "");
Expect(0, 159, '\P{iscntrl}', "");
Expect(1, 159, '\P{^iscntrl}', "");
Expect(0, 160, '\p{iscntrl}', "");
Expect(1, 160, '\p{^iscntrl}', "");
Expect(1, 160, '\P{iscntrl}', "");
Expect(0, 160, '\P{^iscntrl}', "");
Expect(1, 159, '\p{-_Is_Cntrl}', "");
Expect(0, 159, '\p{^-_Is_Cntrl}', "");
Expect(0, 159, '\P{-_Is_Cntrl}', "");
Expect(1, 159, '\P{^-_Is_Cntrl}', "");
Expect(0, 160, '\p{-_Is_Cntrl}', "");
Expect(1, 160, '\p{^-_Is_Cntrl}', "");
Expect(1, 160, '\P{-_Is_Cntrl}', "");
Expect(0, 160, '\P{^-_Is_Cntrl}', "");
Error('\p{:=_Is_XPosixCntrl}');
Error('\P{:=_Is_XPosixCntrl}');
Expect(1, 159, '\p{isxposixcntrl}', "");
Expect(0, 159, '\p{^isxposixcntrl}', "");
Expect(0, 159, '\P{isxposixcntrl}', "");
Expect(1, 159, '\P{^isxposixcntrl}', "");
Expect(0, 160, '\p{isxposixcntrl}', "");
Expect(1, 160, '\p{^isxposixcntrl}', "");
Expect(1, 160, '\P{isxposixcntrl}', "");
Expect(0, 160, '\P{^isxposixcntrl}', "");
Expect(1, 159, '\p{-Is_XPOSIXCNTRL}', "");
Expect(0, 159, '\p{^-Is_XPOSIXCNTRL}', "");
Expect(0, 159, '\P{-Is_XPOSIXCNTRL}', "");
Expect(1, 159, '\P{^-Is_XPOSIXCNTRL}', "");
Expect(0, 160, '\p{-Is_XPOSIXCNTRL}', "");
Expect(1, 160, '\p{^-Is_XPOSIXCNTRL}', "");
Expect(1, 160, '\P{-Is_XPOSIXCNTRL}', "");
Expect(0, 160, '\P{^-Is_XPOSIXCNTRL}', "");
Error('\p{ /a/Control}');
Error('\P{ /a/Control}');
Expect(1, 159, '\p{control}', "");
Expect(0, 159, '\p{^control}', "");
Expect(0, 159, '\P{control}', "");
Expect(1, 159, '\P{^control}', "");
Expect(0, 160, '\p{control}', "");
Expect(1, 160, '\p{^control}', "");
Expect(1, 160, '\P{control}', "");
Expect(0, 160, '\P{^control}', "");
Expect(1, 159, '\p{-control}', "");
Expect(0, 159, '\p{^-control}', "");
Expect(0, 159, '\P{-control}', "");
Expect(1, 159, '\P{^-control}', "");
Expect(0, 160, '\p{-control}', "");
Expect(1, 160, '\p{^-control}', "");
Expect(1, 160, '\P{-control}', "");
Expect(0, 160, '\P{^-control}', "");
Error('\p{/a/Is_CONTROL}');
Error('\P{/a/Is_CONTROL}');
Expect(1, 159, '\p{iscontrol}', "");
Expect(0, 159, '\p{^iscontrol}', "");
Expect(0, 159, '\P{iscontrol}', "");
Expect(1, 159, '\P{^iscontrol}', "");
Expect(0, 160, '\p{iscontrol}', "");
Expect(1, 160, '\p{^iscontrol}', "");
Expect(1, 160, '\P{iscontrol}', "");
Expect(0, 160, '\P{^iscontrol}', "");
Expect(1, 159, '\p{	-Is_control}', "");
Expect(0, 159, '\p{^	-Is_control}', "");
Expect(0, 159, '\P{	-Is_control}', "");
Expect(1, 159, '\P{^	-Is_control}', "");
Expect(0, 160, '\p{	-Is_control}', "");
Expect(1, 160, '\p{^	-Is_control}', "");
Expect(1, 160, '\P{	-Is_control}', "");
Expect(0, 160, '\P{^	-Is_control}', "");
Error('\p{_Cc:=}');
Error('\P{_Cc:=}');
Expect(1, 159, '\p{cc}', "");
Expect(0, 159, '\p{^cc}', "");
Expect(0, 159, '\P{cc}', "");
Expect(1, 159, '\P{^cc}', "");
Expect(0, 160, '\p{cc}', "");
Expect(1, 160, '\p{^cc}', "");
Expect(1, 160, '\P{cc}', "");
Expect(0, 160, '\P{^cc}', "");
Expect(1, 159, '\p{		Cc}', "");
Expect(0, 159, '\p{^		Cc}', "");
Expect(0, 159, '\P{		Cc}', "");
Expect(1, 159, '\P{^		Cc}', "");
Expect(0, 160, '\p{		Cc}', "");
Expect(1, 160, '\p{^		Cc}', "");
Expect(1, 160, '\P{		Cc}', "");
Expect(0, 160, '\P{^		Cc}', "");
Error('\p{:=IS_CC}');
Error('\P{:=IS_CC}');
Expect(1, 159, '\p{iscc}', "");
Expect(0, 159, '\p{^iscc}', "");
Expect(0, 159, '\P{iscc}', "");
Expect(1, 159, '\P{^iscc}', "");
Expect(0, 160, '\p{iscc}', "");
Expect(1, 160, '\p{^iscc}', "");
Expect(1, 160, '\P{iscc}', "");
Expect(0, 160, '\P{^iscc}', "");
Expect(1, 159, '\p{__Is_cc}', "");
Expect(0, 159, '\p{^__Is_cc}', "");
Expect(0, 159, '\P{__Is_cc}', "");
Expect(1, 159, '\P{^__Is_cc}', "");
Expect(0, 160, '\p{__Is_cc}', "");
Expect(1, 160, '\p{^__Is_cc}', "");
Expect(1, 160, '\P{__Is_cc}', "");
Expect(0, 160, '\P{^__Is_cc}', "");
Error('\p{-Combining_Diacritical_MARKS:=}');
Error('\P{-Combining_Diacritical_MARKS:=}');
Expect(1, 879, '\p{combiningdiacriticalmarks}', "");
Expect(0, 879, '\p{^combiningdiacriticalmarks}', "");
Expect(0, 879, '\P{combiningdiacriticalmarks}', "");
Expect(1, 879, '\P{^combiningdiacriticalmarks}', "");
Expect(0, 880, '\p{combiningdiacriticalmarks}', "");
Expect(1, 880, '\p{^combiningdiacriticalmarks}', "");
Expect(1, 880, '\P{combiningdiacriticalmarks}', "");
Expect(0, 880, '\P{^combiningdiacriticalmarks}', "");
Expect(1, 879, '\p{__Combining_Diacritical_Marks}', "");
Expect(0, 879, '\p{^__Combining_Diacritical_Marks}', "");
Expect(0, 879, '\P{__Combining_Diacritical_Marks}', "");
Expect(1, 879, '\P{^__Combining_Diacritical_Marks}', "");
Expect(0, 880, '\p{__Combining_Diacritical_Marks}', "");
Expect(1, 880, '\p{^__Combining_Diacritical_Marks}', "");
Expect(1, 880, '\P{__Combining_Diacritical_Marks}', "");
Expect(0, 880, '\P{^__Combining_Diacritical_Marks}', "");
Error('\p{ _Is_combining_Diacritical_Marks:=}');
Error('\P{ _Is_combining_Diacritical_Marks:=}');
Expect(1, 879, '\p{iscombiningdiacriticalmarks}', "");
Expect(0, 879, '\p{^iscombiningdiacriticalmarks}', "");
Expect(0, 879, '\P{iscombiningdiacriticalmarks}', "");
Expect(1, 879, '\P{^iscombiningdiacriticalmarks}', "");
Expect(0, 880, '\p{iscombiningdiacriticalmarks}', "");
Expect(1, 880, '\p{^iscombiningdiacriticalmarks}', "");
Expect(1, 880, '\P{iscombiningdiacriticalmarks}', "");
Expect(0, 880, '\P{^iscombiningdiacriticalmarks}', "");
Expect(1, 879, '\p{  IS_Combining_Diacritical_Marks}', "");
Expect(0, 879, '\p{^  IS_Combining_Diacritical_Marks}', "");
Expect(0, 879, '\P{  IS_Combining_Diacritical_Marks}', "");
Expect(1, 879, '\P{^  IS_Combining_Diacritical_Marks}', "");
Expect(0, 880, '\p{  IS_Combining_Diacritical_Marks}', "");
Expect(1, 880, '\p{^  IS_Combining_Diacritical_Marks}', "");
Expect(1, 880, '\P{  IS_Combining_Diacritical_Marks}', "");
Expect(0, 880, '\P{^  IS_Combining_Diacritical_Marks}', "");
Error('\p{ 	IN_Combining_Diacritical_MARKS:=}');
Error('\P{ 	IN_Combining_Diacritical_MARKS:=}');
Expect(1, 879, '\p{incombiningdiacriticalmarks}', "");
Expect(0, 879, '\p{^incombiningdiacriticalmarks}', "");
Expect(0, 879, '\P{incombiningdiacriticalmarks}', "");
Expect(1, 879, '\P{^incombiningdiacriticalmarks}', "");
Expect(0, 880, '\p{incombiningdiacriticalmarks}', "");
Expect(1, 880, '\p{^incombiningdiacriticalmarks}', "");
Expect(1, 880, '\P{incombiningdiacriticalmarks}', "");
Expect(0, 880, '\P{^incombiningdiacriticalmarks}', "");
Expect(1, 879, '\p{-IN_combining_diacritical_Marks}', "");
Expect(0, 879, '\p{^-IN_combining_diacritical_Marks}', "");
Expect(0, 879, '\P{-IN_combining_diacritical_Marks}', "");
Expect(1, 879, '\P{^-IN_combining_diacritical_Marks}', "");
Expect(0, 880, '\p{-IN_combining_diacritical_Marks}', "");
Expect(1, 880, '\p{^-IN_combining_diacritical_Marks}', "");
Expect(1, 880, '\P{-IN_combining_diacritical_Marks}', "");
Expect(0, 880, '\P{^-IN_combining_diacritical_Marks}', "");
Error('\p{_:=diacriticals}');
Error('\P{_:=diacriticals}');
Expect(1, 879, '\p{diacriticals}', "");
Expect(0, 879, '\p{^diacriticals}', "");
Expect(0, 879, '\P{diacriticals}', "");
Expect(1, 879, '\P{^diacriticals}', "");
Expect(0, 880, '\p{diacriticals}', "");
Expect(1, 880, '\p{^diacriticals}', "");
Expect(1, 880, '\P{diacriticals}', "");
Expect(0, 880, '\P{^diacriticals}', "");
Expect(1, 879, '\p{_ Diacriticals}', "");
Expect(0, 879, '\p{^_ Diacriticals}', "");
Expect(0, 879, '\P{_ Diacriticals}', "");
Expect(1, 879, '\P{^_ Diacriticals}', "");
Expect(0, 880, '\p{_ Diacriticals}', "");
Expect(1, 880, '\p{^_ Diacriticals}', "");
Expect(1, 880, '\P{_ Diacriticals}', "");
Expect(0, 880, '\P{^_ Diacriticals}', "");
Error('\p{/a/-_IS_diacriticals}');
Error('\P{/a/-_IS_diacriticals}');
Expect(1, 879, '\p{isdiacriticals}', "");
Expect(0, 879, '\p{^isdiacriticals}', "");
Expect(0, 879, '\P{isdiacriticals}', "");
Expect(1, 879, '\P{^isdiacriticals}', "");
Expect(0, 880, '\p{isdiacriticals}', "");
Expect(1, 880, '\p{^isdiacriticals}', "");
Expect(1, 880, '\P{isdiacriticals}', "");
Expect(0, 880, '\P{^isdiacriticals}', "");
Expect(1, 879, '\p{ Is_Diacriticals}', "");
Expect(0, 879, '\p{^ Is_Diacriticals}', "");
Expect(0, 879, '\P{ Is_Diacriticals}', "");
Expect(1, 879, '\P{^ Is_Diacriticals}', "");
Expect(0, 880, '\p{ Is_Diacriticals}', "");
Expect(1, 880, '\p{^ Is_Diacriticals}', "");
Expect(1, 880, '\P{ Is_Diacriticals}', "");
Expect(0, 880, '\P{^ Is_Diacriticals}', "");
Error('\p{_/a/in_Diacriticals}');
Error('\P{_/a/in_Diacriticals}');
Expect(1, 879, '\p{indiacriticals}', "");
Expect(0, 879, '\p{^indiacriticals}', "");
Expect(0, 879, '\P{indiacriticals}', "");
Expect(1, 879, '\P{^indiacriticals}', "");
Expect(0, 880, '\p{indiacriticals}', "");
Expect(1, 880, '\p{^indiacriticals}', "");
Expect(1, 880, '\P{indiacriticals}', "");
Expect(0, 880, '\P{^indiacriticals}', "");
Expect(1, 879, '\p{ -In_DIACRITICALS}', "");
Expect(0, 879, '\p{^ -In_DIACRITICALS}', "");
Expect(0, 879, '\P{ -In_DIACRITICALS}', "");
Expect(1, 879, '\P{^ -In_DIACRITICALS}', "");
Expect(0, 880, '\p{ -In_DIACRITICALS}', "");
Expect(1, 880, '\p{^ -In_DIACRITICALS}', "");
Expect(1, 880, '\P{ -In_DIACRITICALS}', "");
Expect(0, 880, '\P{^ -In_DIACRITICALS}', "");
Error('\p{- COMBINING_Diacritical_MARKS_For_symbols:=}');
Error('\P{- COMBINING_Diacritical_MARKS_For_symbols:=}');
Expect(1, 8447, '\p{combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\p{^combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\P{combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\P{^combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\p{combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\p{^combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\P{combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\P{^combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\p{ -Combining_Diacritical_Marks_FOR_symbols}', "");
Expect(0, 8447, '\p{^ -Combining_Diacritical_Marks_FOR_symbols}', "");
Expect(0, 8447, '\P{ -Combining_Diacritical_Marks_FOR_symbols}', "");
Expect(1, 8447, '\P{^ -Combining_Diacritical_Marks_FOR_symbols}', "");
Expect(0, 8448, '\p{ -Combining_Diacritical_Marks_FOR_symbols}', "");
Expect(1, 8448, '\p{^ -Combining_Diacritical_Marks_FOR_symbols}', "");
Expect(1, 8448, '\P{ -Combining_Diacritical_Marks_FOR_symbols}', "");
Expect(0, 8448, '\P{^ -Combining_Diacritical_Marks_FOR_symbols}', "");
Error('\p{	:=Is_Combining_DIACRITICAL_Marks_For_symbols}');
Error('\P{	:=Is_Combining_DIACRITICAL_Marks_For_symbols}');
Expect(1, 8447, '\p{iscombiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\p{^iscombiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\P{iscombiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\P{^iscombiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\p{iscombiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\p{^iscombiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\P{iscombiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\P{^iscombiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\p{	Is_Combining_Diacritical_Marks_For_symbols}', "");
Expect(0, 8447, '\p{^	Is_Combining_Diacritical_Marks_For_symbols}', "");
Expect(0, 8447, '\P{	Is_Combining_Diacritical_Marks_For_symbols}', "");
Expect(1, 8447, '\P{^	Is_Combining_Diacritical_Marks_For_symbols}', "");
Expect(0, 8448, '\p{	Is_Combining_Diacritical_Marks_For_symbols}', "");
Expect(1, 8448, '\p{^	Is_Combining_Diacritical_Marks_For_symbols}', "");
Expect(1, 8448, '\P{	Is_Combining_Diacritical_Marks_For_symbols}', "");
Expect(0, 8448, '\P{^	Is_Combining_Diacritical_Marks_For_symbols}', "");
Error('\p{/a/	-In_combining_DIACRITICAL_MARKS_For_SYMBOLS}');
Error('\P{/a/	-In_combining_DIACRITICAL_MARKS_For_SYMBOLS}');
Expect(1, 8447, '\p{incombiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\p{^incombiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\P{incombiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\P{^incombiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\p{incombiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\p{^incombiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\P{incombiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\P{^incombiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\p{	In_combining_Diacritical_MARKS_for_Symbols}', "");
Expect(0, 8447, '\p{^	In_combining_Diacritical_MARKS_for_Symbols}', "");
Expect(0, 8447, '\P{	In_combining_Diacritical_MARKS_for_Symbols}', "");
Expect(1, 8447, '\P{^	In_combining_Diacritical_MARKS_for_Symbols}', "");
Expect(0, 8448, '\p{	In_combining_Diacritical_MARKS_for_Symbols}', "");
Expect(1, 8448, '\p{^	In_combining_Diacritical_MARKS_for_Symbols}', "");
Expect(1, 8448, '\P{	In_combining_Diacritical_MARKS_for_Symbols}', "");
Expect(0, 8448, '\P{^	In_combining_Diacritical_MARKS_for_Symbols}', "");
Error('\p{/a/DIACRITICALS_For_symbols}');
Error('\P{/a/DIACRITICALS_For_symbols}');
Expect(1, 8447, '\p{diacriticalsforsymbols}', "");
Expect(0, 8447, '\p{^diacriticalsforsymbols}', "");
Expect(0, 8447, '\P{diacriticalsforsymbols}', "");
Expect(1, 8447, '\P{^diacriticalsforsymbols}', "");
Expect(0, 8448, '\p{diacriticalsforsymbols}', "");
Expect(1, 8448, '\p{^diacriticalsforsymbols}', "");
Expect(1, 8448, '\P{diacriticalsforsymbols}', "");
Expect(0, 8448, '\P{^diacriticalsforsymbols}', "");
Expect(1, 8447, '\p{		Diacriticals_for_Symbols}', "");
Expect(0, 8447, '\p{^		Diacriticals_for_Symbols}', "");
Expect(0, 8447, '\P{		Diacriticals_for_Symbols}', "");
Expect(1, 8447, '\P{^		Diacriticals_for_Symbols}', "");
Expect(0, 8448, '\p{		Diacriticals_for_Symbols}', "");
Expect(1, 8448, '\p{^		Diacriticals_for_Symbols}', "");
Expect(1, 8448, '\P{		Diacriticals_for_Symbols}', "");
Expect(0, 8448, '\P{^		Diacriticals_for_Symbols}', "");
Error('\p{- IS_DIACRITICALS_For_symbols:=}');
Error('\P{- IS_DIACRITICALS_For_symbols:=}');
Expect(1, 8447, '\p{isdiacriticalsforsymbols}', "");
Expect(0, 8447, '\p{^isdiacriticalsforsymbols}', "");
Expect(0, 8447, '\P{isdiacriticalsforsymbols}', "");
Expect(1, 8447, '\P{^isdiacriticalsforsymbols}', "");
Expect(0, 8448, '\p{isdiacriticalsforsymbols}', "");
Expect(1, 8448, '\p{^isdiacriticalsforsymbols}', "");
Expect(1, 8448, '\P{isdiacriticalsforsymbols}', "");
Expect(0, 8448, '\P{^isdiacriticalsforsymbols}', "");
Expect(1, 8447, '\p{	Is_Diacriticals_FOR_symbols}', "");
Expect(0, 8447, '\p{^	Is_Diacriticals_FOR_symbols}', "");
Expect(0, 8447, '\P{	Is_Diacriticals_FOR_symbols}', "");
Expect(1, 8447, '\P{^	Is_Diacriticals_FOR_symbols}', "");
Expect(0, 8448, '\p{	Is_Diacriticals_FOR_symbols}', "");
Expect(1, 8448, '\p{^	Is_Diacriticals_FOR_symbols}', "");
Expect(1, 8448, '\P{	Is_Diacriticals_FOR_symbols}', "");
Expect(0, 8448, '\P{^	Is_Diacriticals_FOR_symbols}', "");
Error('\p{:=		IN_DIACRITICALS_For_Symbols}');
Error('\P{:=		IN_DIACRITICALS_For_Symbols}');
Expect(1, 8447, '\p{indiacriticalsforsymbols}', "");
Expect(0, 8447, '\p{^indiacriticalsforsymbols}', "");
Expect(0, 8447, '\P{indiacriticalsforsymbols}', "");
Expect(1, 8447, '\P{^indiacriticalsforsymbols}', "");
Expect(0, 8448, '\p{indiacriticalsforsymbols}', "");
Expect(1, 8448, '\p{^indiacriticalsforsymbols}', "");
Expect(1, 8448, '\P{indiacriticalsforsymbols}', "");
Expect(0, 8448, '\P{^indiacriticalsforsymbols}', "");
Expect(1, 8447, '\p{--in_Diacriticals_for_symbols}', "");
Expect(0, 8447, '\p{^--in_Diacriticals_for_symbols}', "");
Expect(0, 8447, '\P{--in_Diacriticals_for_symbols}', "");
Expect(1, 8447, '\P{^--in_Diacriticals_for_symbols}', "");
Expect(0, 8448, '\p{--in_Diacriticals_for_symbols}', "");
Expect(1, 8448, '\p{^--in_Diacriticals_for_symbols}', "");
Expect(1, 8448, '\P{--in_Diacriticals_for_symbols}', "");
Expect(0, 8448, '\P{^--in_Diacriticals_for_symbols}', "");
Error('\p{	COMBINING_Marks_for_Symbols/a/}');
Error('\P{	COMBINING_Marks_for_Symbols/a/}');
Expect(1, 8447, '\p{combiningmarksforsymbols}', "");
Expect(0, 8447, '\p{^combiningmarksforsymbols}', "");
Expect(0, 8447, '\P{combiningmarksforsymbols}', "");
Expect(1, 8447, '\P{^combiningmarksforsymbols}', "");
Expect(0, 8448, '\p{combiningmarksforsymbols}', "");
Expect(1, 8448, '\p{^combiningmarksforsymbols}', "");
Expect(1, 8448, '\P{combiningmarksforsymbols}', "");
Expect(0, 8448, '\P{^combiningmarksforsymbols}', "");
Expect(1, 8447, '\p{_combining_Marks_for_Symbols}', "");
Expect(0, 8447, '\p{^_combining_Marks_for_Symbols}', "");
Expect(0, 8447, '\P{_combining_Marks_for_Symbols}', "");
Expect(1, 8447, '\P{^_combining_Marks_for_Symbols}', "");
Expect(0, 8448, '\p{_combining_Marks_for_Symbols}', "");
Expect(1, 8448, '\p{^_combining_Marks_for_Symbols}', "");
Expect(1, 8448, '\P{_combining_Marks_for_Symbols}', "");
Expect(0, 8448, '\P{^_combining_Marks_for_Symbols}', "");
Error('\p{_/a/is_combining_Marks_For_symbols}');
Error('\P{_/a/is_combining_Marks_For_symbols}');
Expect(1, 8447, '\p{iscombiningmarksforsymbols}', "");
Expect(0, 8447, '\p{^iscombiningmarksforsymbols}', "");
Expect(0, 8447, '\P{iscombiningmarksforsymbols}', "");
Expect(1, 8447, '\P{^iscombiningmarksforsymbols}', "");
Expect(0, 8448, '\p{iscombiningmarksforsymbols}', "");
Expect(1, 8448, '\p{^iscombiningmarksforsymbols}', "");
Expect(1, 8448, '\P{iscombiningmarksforsymbols}', "");
Expect(0, 8448, '\P{^iscombiningmarksforsymbols}', "");
Expect(1, 8447, '\p{ Is_Combining_marks_FOR_symbols}', "");
Expect(0, 8447, '\p{^ Is_Combining_marks_FOR_symbols}', "");
Expect(0, 8447, '\P{ Is_Combining_marks_FOR_symbols}', "");
Expect(1, 8447, '\P{^ Is_Combining_marks_FOR_symbols}', "");
Expect(0, 8448, '\p{ Is_Combining_marks_FOR_symbols}', "");
Expect(1, 8448, '\p{^ Is_Combining_marks_FOR_symbols}', "");
Expect(1, 8448, '\P{ Is_Combining_marks_FOR_symbols}', "");
Expect(0, 8448, '\P{^ Is_Combining_marks_FOR_symbols}', "");
Error('\p{ /a/IN_Combining_marks_For_Symbols}');
Error('\P{ /a/IN_Combining_marks_For_Symbols}');
Expect(1, 8447, '\p{incombiningmarksforsymbols}', "");
Expect(0, 8447, '\p{^incombiningmarksforsymbols}', "");
Expect(0, 8447, '\P{incombiningmarksforsymbols}', "");
Expect(1, 8447, '\P{^incombiningmarksforsymbols}', "");
Expect(0, 8448, '\p{incombiningmarksforsymbols}', "");
Expect(1, 8448, '\p{^incombiningmarksforsymbols}', "");
Expect(1, 8448, '\P{incombiningmarksforsymbols}', "");
Expect(0, 8448, '\P{^incombiningmarksforsymbols}', "");
Expect(1, 8447, '\p{ 	IN_Combining_marks_For_Symbols}', "");
Expect(0, 8447, '\p{^ 	IN_Combining_marks_For_Symbols}', "");
Expect(0, 8447, '\P{ 	IN_Combining_marks_For_Symbols}', "");
Expect(1, 8447, '\P{^ 	IN_Combining_marks_For_Symbols}', "");
Expect(0, 8448, '\p{ 	IN_Combining_marks_For_Symbols}', "");
Expect(1, 8448, '\p{^ 	IN_Combining_marks_For_Symbols}', "");
Expect(1, 8448, '\P{ 	IN_Combining_marks_For_Symbols}', "");
Expect(0, 8448, '\P{^ 	IN_Combining_marks_For_Symbols}', "");
Error('\p{-Combining_diacritical_Marks_supplement/a/}');
Error('\P{-Combining_diacritical_Marks_supplement/a/}');
Expect(1, 7679, '\p{combiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\p{^combiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\P{combiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\P{^combiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\p{combiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\p{^combiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\P{combiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\P{^combiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\p{-_COMBINING_DIACRITICAL_MARKS_supplement}', "");
Expect(0, 7679, '\p{^-_COMBINING_DIACRITICAL_MARKS_supplement}', "");
Expect(0, 7679, '\P{-_COMBINING_DIACRITICAL_MARKS_supplement}', "");
Expect(1, 7679, '\P{^-_COMBINING_DIACRITICAL_MARKS_supplement}', "");
Expect(0, 7680, '\p{-_COMBINING_DIACRITICAL_MARKS_supplement}', "");
Expect(1, 7680, '\p{^-_COMBINING_DIACRITICAL_MARKS_supplement}', "");
Expect(1, 7680, '\P{-_COMBINING_DIACRITICAL_MARKS_supplement}', "");
Expect(0, 7680, '\P{^-_COMBINING_DIACRITICAL_MARKS_supplement}', "");
Error('\p{ 	is_COMBINING_diacritical_MARKS_Supplement:=}');
Error('\P{ 	is_COMBINING_diacritical_MARKS_Supplement:=}');
Expect(1, 7679, '\p{iscombiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\p{^iscombiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\P{iscombiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\P{^iscombiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\p{iscombiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\p{^iscombiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\P{iscombiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\P{^iscombiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\p{	 Is_Combining_DIACRITICAL_marks_Supplement}', "");
Expect(0, 7679, '\p{^	 Is_Combining_DIACRITICAL_marks_Supplement}', "");
Expect(0, 7679, '\P{	 Is_Combining_DIACRITICAL_marks_Supplement}', "");
Expect(1, 7679, '\P{^	 Is_Combining_DIACRITICAL_marks_Supplement}', "");
Expect(0, 7680, '\p{	 Is_Combining_DIACRITICAL_marks_Supplement}', "");
Expect(1, 7680, '\p{^	 Is_Combining_DIACRITICAL_marks_Supplement}', "");
Expect(1, 7680, '\P{	 Is_Combining_DIACRITICAL_marks_Supplement}', "");
Expect(0, 7680, '\P{^	 Is_Combining_DIACRITICAL_marks_Supplement}', "");
Error('\p{/a/-	IN_Combining_diacritical_marks_supplement}');
Error('\P{/a/-	IN_Combining_diacritical_marks_supplement}');
Expect(1, 7679, '\p{incombiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\p{^incombiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\P{incombiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\P{^incombiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\p{incombiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\p{^incombiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\P{incombiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\P{^incombiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\p{	 in_Combining_diacritical_Marks_SUPPLEMENT}', "");
Expect(0, 7679, '\p{^	 in_Combining_diacritical_Marks_SUPPLEMENT}', "");
Expect(0, 7679, '\P{	 in_Combining_diacritical_Marks_SUPPLEMENT}', "");
Expect(1, 7679, '\P{^	 in_Combining_diacritical_Marks_SUPPLEMENT}', "");
Expect(0, 7680, '\p{	 in_Combining_diacritical_Marks_SUPPLEMENT}', "");
Expect(1, 7680, '\p{^	 in_Combining_diacritical_Marks_SUPPLEMENT}', "");
Expect(1, 7680, '\P{	 in_Combining_diacritical_Marks_SUPPLEMENT}', "");
Expect(0, 7680, '\P{^	 in_Combining_diacritical_Marks_SUPPLEMENT}', "");
Error('\p{-_DIACRITICALS_sup/a/}');
Error('\P{-_DIACRITICALS_sup/a/}');
Expect(1, 7679, '\p{diacriticalssup}', "");
Expect(0, 7679, '\p{^diacriticalssup}', "");
Expect(0, 7679, '\P{diacriticalssup}', "");
Expect(1, 7679, '\P{^diacriticalssup}', "");
Expect(0, 7680, '\p{diacriticalssup}', "");
Expect(1, 7680, '\p{^diacriticalssup}', "");
Expect(1, 7680, '\P{diacriticalssup}', "");
Expect(0, 7680, '\P{^diacriticalssup}', "");
Expect(1, 7679, '\p{diacriticals_sup}', "");
Expect(0, 7679, '\p{^diacriticals_sup}', "");
Expect(0, 7679, '\P{diacriticals_sup}', "");
Expect(1, 7679, '\P{^diacriticals_sup}', "");
Expect(0, 7680, '\p{diacriticals_sup}', "");
Expect(1, 7680, '\p{^diacriticals_sup}', "");
Expect(1, 7680, '\P{diacriticals_sup}', "");
Expect(0, 7680, '\P{^diacriticals_sup}', "");
Error('\p{:=-_Is_DIACRITICALS_SUP}');
Error('\P{:=-_Is_DIACRITICALS_SUP}');
Expect(1, 7679, '\p{isdiacriticalssup}', "");
Expect(0, 7679, '\p{^isdiacriticalssup}', "");
Expect(0, 7679, '\P{isdiacriticalssup}', "");
Expect(1, 7679, '\P{^isdiacriticalssup}', "");
Expect(0, 7680, '\p{isdiacriticalssup}', "");
Expect(1, 7680, '\p{^isdiacriticalssup}', "");
Expect(1, 7680, '\P{isdiacriticalssup}', "");
Expect(0, 7680, '\P{^isdiacriticalssup}', "");
Expect(1, 7679, '\p{_ IS_DIACRITICALS_SUP}', "");
Expect(0, 7679, '\p{^_ IS_DIACRITICALS_SUP}', "");
Expect(0, 7679, '\P{_ IS_DIACRITICALS_SUP}', "");
Expect(1, 7679, '\P{^_ IS_DIACRITICALS_SUP}', "");
Expect(0, 7680, '\p{_ IS_DIACRITICALS_SUP}', "");
Expect(1, 7680, '\p{^_ IS_DIACRITICALS_SUP}', "");
Expect(1, 7680, '\P{_ IS_DIACRITICALS_SUP}', "");
Expect(0, 7680, '\P{^_ IS_DIACRITICALS_SUP}', "");
Error('\p{_-In_DIACRITICALS_sup:=}');
Error('\P{_-In_DIACRITICALS_sup:=}');
Expect(1, 7679, '\p{indiacriticalssup}', "");
Expect(0, 7679, '\p{^indiacriticalssup}', "");
Expect(0, 7679, '\P{indiacriticalssup}', "");
Expect(1, 7679, '\P{^indiacriticalssup}', "");
Expect(0, 7680, '\p{indiacriticalssup}', "");
Expect(1, 7680, '\p{^indiacriticalssup}', "");
Expect(1, 7680, '\P{indiacriticalssup}', "");
Expect(0, 7680, '\P{^indiacriticalssup}', "");
Expect(1, 7679, '\p{-_IN_diacriticals_Sup}', "");
Expect(0, 7679, '\p{^-_IN_diacriticals_Sup}', "");
Expect(0, 7679, '\P{-_IN_diacriticals_Sup}', "");
Expect(1, 7679, '\P{^-_IN_diacriticals_Sup}', "");
Expect(0, 7680, '\p{-_IN_diacriticals_Sup}', "");
Expect(1, 7680, '\p{^-_IN_diacriticals_Sup}', "");
Expect(1, 7680, '\P{-_IN_diacriticals_Sup}', "");
Expect(0, 7680, '\P{^-_IN_diacriticals_Sup}', "");
Error('\p{ Combining_HALF_MARKS/a/}');
Error('\P{ Combining_HALF_MARKS/a/}');
Expect(1, 65071, '\p{combininghalfmarks}', "");
Expect(0, 65071, '\p{^combininghalfmarks}', "");
Expect(0, 65071, '\P{combininghalfmarks}', "");
Expect(1, 65071, '\P{^combininghalfmarks}', "");
Expect(0, 65072, '\p{combininghalfmarks}', "");
Expect(1, 65072, '\p{^combininghalfmarks}', "");
Expect(1, 65072, '\P{combininghalfmarks}', "");
Expect(0, 65072, '\P{^combininghalfmarks}', "");
Expect(1, 65071, '\p{-COMBINING_Half_Marks}', "");
Expect(0, 65071, '\p{^-COMBINING_Half_Marks}', "");
Expect(0, 65071, '\P{-COMBINING_Half_Marks}', "");
Expect(1, 65071, '\P{^-COMBINING_Half_Marks}', "");
Expect(0, 65072, '\p{-COMBINING_Half_Marks}', "");
Expect(1, 65072, '\p{^-COMBINING_Half_Marks}', "");
Expect(1, 65072, '\P{-COMBINING_Half_Marks}', "");
Expect(0, 65072, '\P{^-COMBINING_Half_Marks}', "");
Error('\p{ 	is_COMBINING_HALF_Marks:=}');
Error('\P{ 	is_COMBINING_HALF_Marks:=}');
Expect(1, 65071, '\p{iscombininghalfmarks}', "");
Expect(0, 65071, '\p{^iscombininghalfmarks}', "");
Expect(0, 65071, '\P{iscombininghalfmarks}', "");
Expect(1, 65071, '\P{^iscombininghalfmarks}', "");
Expect(0, 65072, '\p{iscombininghalfmarks}', "");
Expect(1, 65072, '\p{^iscombininghalfmarks}', "");
Expect(1, 65072, '\P{iscombininghalfmarks}', "");
Expect(0, 65072, '\P{^iscombininghalfmarks}', "");
Expect(1, 65071, '\p{	IS_COMBINING_HALF_marks}', "");
Expect(0, 65071, '\p{^	IS_COMBINING_HALF_marks}', "");
Expect(0, 65071, '\P{	IS_COMBINING_HALF_marks}', "");
Expect(1, 65071, '\P{^	IS_COMBINING_HALF_marks}', "");
Expect(0, 65072, '\p{	IS_COMBINING_HALF_marks}', "");
Expect(1, 65072, '\p{^	IS_COMBINING_HALF_marks}', "");
Expect(1, 65072, '\P{	IS_COMBINING_HALF_marks}', "");
Expect(0, 65072, '\P{^	IS_COMBINING_HALF_marks}', "");
Error('\p{_:=IN_Combining_half_Marks}');
Error('\P{_:=IN_Combining_half_Marks}');
Expect(1, 65071, '\p{incombininghalfmarks}', "");
Expect(0, 65071, '\p{^incombininghalfmarks}', "");
Expect(0, 65071, '\P{incombininghalfmarks}', "");
Expect(1, 65071, '\P{^incombininghalfmarks}', "");
Expect(0, 65072, '\p{incombininghalfmarks}', "");
Expect(1, 65072, '\p{^incombininghalfmarks}', "");
Expect(1, 65072, '\P{incombininghalfmarks}', "");
Expect(0, 65072, '\P{^incombininghalfmarks}', "");
Expect(1, 65071, '\p{  IN_COMBINING_Half_Marks}', "");
Expect(0, 65071, '\p{^  IN_COMBINING_Half_Marks}', "");
Expect(0, 65071, '\P{  IN_COMBINING_Half_Marks}', "");
Expect(1, 65071, '\P{^  IN_COMBINING_Half_Marks}', "");
Expect(0, 65072, '\p{  IN_COMBINING_Half_Marks}', "");
Expect(1, 65072, '\p{^  IN_COMBINING_Half_Marks}', "");
Expect(1, 65072, '\P{  IN_COMBINING_Half_Marks}', "");
Expect(0, 65072, '\P{^  IN_COMBINING_Half_Marks}', "");
Error('\p{	:=half_MARKS}');
Error('\P{	:=half_MARKS}');
Expect(1, 65071, '\p{halfmarks}', "");
Expect(0, 65071, '\p{^halfmarks}', "");
Expect(0, 65071, '\P{halfmarks}', "");
Expect(1, 65071, '\P{^halfmarks}', "");
Expect(0, 65072, '\p{halfmarks}', "");
Expect(1, 65072, '\p{^halfmarks}', "");
Expect(1, 65072, '\P{halfmarks}', "");
Expect(0, 65072, '\P{^halfmarks}', "");
Expect(1, 65071, '\p{ Half_Marks}', "");
Expect(0, 65071, '\p{^ Half_Marks}', "");
Expect(0, 65071, '\P{ Half_Marks}', "");
Expect(1, 65071, '\P{^ Half_Marks}', "");
Expect(0, 65072, '\p{ Half_Marks}', "");
Expect(1, 65072, '\p{^ Half_Marks}', "");
Expect(1, 65072, '\P{ Half_Marks}', "");
Expect(0, 65072, '\P{^ Half_Marks}', "");
Error('\p{_/a/Is_Half_Marks}');
Error('\P{_/a/Is_Half_Marks}');
Expect(1, 65071, '\p{ishalfmarks}', "");
Expect(0, 65071, '\p{^ishalfmarks}', "");
Expect(0, 65071, '\P{ishalfmarks}', "");
Expect(1, 65071, '\P{^ishalfmarks}', "");
Expect(0, 65072, '\p{ishalfmarks}', "");
Expect(1, 65072, '\p{^ishalfmarks}', "");
Expect(1, 65072, '\P{ishalfmarks}', "");
Expect(0, 65072, '\P{^ishalfmarks}', "");
Expect(1, 65071, '\p{- IS_HALF_marks}', "");
Expect(0, 65071, '\p{^- IS_HALF_marks}', "");
Expect(0, 65071, '\P{- IS_HALF_marks}', "");
Expect(1, 65071, '\P{^- IS_HALF_marks}', "");
Expect(0, 65072, '\p{- IS_HALF_marks}', "");
Expect(1, 65072, '\p{^- IS_HALF_marks}', "");
Expect(1, 65072, '\P{- IS_HALF_marks}', "");
Expect(0, 65072, '\P{^- IS_HALF_marks}', "");
Error('\p{/a/	-In_HALF_MARKS}');
Error('\P{/a/	-In_HALF_MARKS}');
Expect(1, 65071, '\p{inhalfmarks}', "");
Expect(0, 65071, '\p{^inhalfmarks}', "");
Expect(0, 65071, '\P{inhalfmarks}', "");
Expect(1, 65071, '\P{^inhalfmarks}', "");
Expect(0, 65072, '\p{inhalfmarks}', "");
Expect(1, 65072, '\p{^inhalfmarks}', "");
Expect(1, 65072, '\P{inhalfmarks}', "");
Expect(0, 65072, '\P{^inhalfmarks}', "");
Expect(1, 65071, '\p{ 	in_Half_MARKS}', "");
Expect(0, 65071, '\p{^ 	in_Half_MARKS}', "");
Expect(0, 65071, '\P{ 	in_Half_MARKS}', "");
Expect(1, 65071, '\P{^ 	in_Half_MARKS}', "");
Expect(0, 65072, '\p{ 	in_Half_MARKS}', "");
Expect(1, 65072, '\p{^ 	in_Half_MARKS}', "");
Expect(1, 65072, '\P{ 	in_Half_MARKS}', "");
Expect(0, 65072, '\P{^ 	in_Half_MARKS}', "");
Error('\p{-/a/Common}');
Error('\P{-/a/Common}');
Expect(1, 917631, '\p{common}', "");
Expect(0, 917631, '\p{^common}', "");
Expect(0, 917631, '\P{common}', "");
Expect(1, 917631, '\P{^common}', "");
Expect(0, 917632, '\p{common}', "");
Expect(1, 917632, '\p{^common}', "");
Expect(1, 917632, '\P{common}', "");
Expect(0, 917632, '\P{^common}', "");
Expect(1, 917631, '\p{_common}', "");
Expect(0, 917631, '\p{^_common}', "");
Expect(0, 917631, '\P{_common}', "");
Expect(1, 917631, '\P{^_common}', "");
Expect(0, 917632, '\p{_common}', "");
Expect(1, 917632, '\p{^_common}', "");
Expect(1, 917632, '\P{_common}', "");
Expect(0, 917632, '\P{^_common}', "");
Error('\p{/a/Is_common}');
Error('\P{/a/Is_common}');
Expect(1, 917631, '\p{iscommon}', "");
Expect(0, 917631, '\p{^iscommon}', "");
Expect(0, 917631, '\P{iscommon}', "");
Expect(1, 917631, '\P{^iscommon}', "");
Expect(0, 917632, '\p{iscommon}', "");
Expect(1, 917632, '\p{^iscommon}', "");
Expect(1, 917632, '\P{iscommon}', "");
Expect(0, 917632, '\P{^iscommon}', "");
Expect(1, 917631, '\p{- Is_common}', "");
Expect(0, 917631, '\p{^- Is_common}', "");
Expect(0, 917631, '\P{- Is_common}', "");
Expect(1, 917631, '\P{^- Is_common}', "");
Expect(0, 917632, '\p{- Is_common}', "");
Expect(1, 917632, '\p{^- Is_common}', "");
Expect(1, 917632, '\P{- Is_common}', "");
Expect(0, 917632, '\P{^- Is_common}', "");
Error('\p{:= -Zyyy}');
Error('\P{:= -Zyyy}');
Expect(1, 917631, '\p{zyyy}', "");
Expect(0, 917631, '\p{^zyyy}', "");
Expect(0, 917631, '\P{zyyy}', "");
Expect(1, 917631, '\P{^zyyy}', "");
Expect(0, 917632, '\p{zyyy}', "");
Expect(1, 917632, '\p{^zyyy}', "");
Expect(1, 917632, '\P{zyyy}', "");
Expect(0, 917632, '\P{^zyyy}', "");
Expect(1, 917631, '\p{-zyyy}', "");
Expect(0, 917631, '\p{^-zyyy}', "");
Expect(0, 917631, '\P{-zyyy}', "");
Expect(1, 917631, '\P{^-zyyy}', "");
Expect(0, 917632, '\p{-zyyy}', "");
Expect(1, 917632, '\p{^-zyyy}', "");
Expect(1, 917632, '\P{-zyyy}', "");
Expect(0, 917632, '\P{^-zyyy}', "");
Error('\p{	Is_ZYYY:=}');
Error('\P{	Is_ZYYY:=}');
Expect(1, 917631, '\p{iszyyy}', "");
Expect(0, 917631, '\p{^iszyyy}', "");
Expect(0, 917631, '\P{iszyyy}', "");
Expect(1, 917631, '\P{^iszyyy}', "");
Expect(0, 917632, '\p{iszyyy}', "");
Expect(1, 917632, '\p{^iszyyy}', "");
Expect(1, 917632, '\P{iszyyy}', "");
Expect(0, 917632, '\P{^iszyyy}', "");
Expect(1, 917631, '\p{_	IS_Zyyy}', "");
Expect(0, 917631, '\p{^_	IS_Zyyy}', "");
Expect(0, 917631, '\P{_	IS_Zyyy}', "");
Expect(1, 917631, '\P{^_	IS_Zyyy}', "");
Expect(0, 917632, '\p{_	IS_Zyyy}', "");
Expect(1, 917632, '\p{^_	IS_Zyyy}', "");
Expect(1, 917632, '\P{_	IS_Zyyy}', "");
Expect(0, 917632, '\P{^_	IS_Zyyy}', "");
Error('\p{/a/ _Common_INDIC_Number_forms}');
Error('\P{/a/ _Common_INDIC_Number_forms}');
Expect(1, 43071, '\p{commonindicnumberforms}', "");
Expect(0, 43071, '\p{^commonindicnumberforms}', "");
Expect(0, 43071, '\P{commonindicnumberforms}', "");
Expect(1, 43071, '\P{^commonindicnumberforms}', "");
Expect(0, 43072, '\p{commonindicnumberforms}', "");
Expect(1, 43072, '\p{^commonindicnumberforms}', "");
Expect(1, 43072, '\P{commonindicnumberforms}', "");
Expect(0, 43072, '\P{^commonindicnumberforms}', "");
Expect(1, 43071, '\p{ 	Common_Indic_Number_FORMS}', "");
Expect(0, 43071, '\p{^ 	Common_Indic_Number_FORMS}', "");
Expect(0, 43071, '\P{ 	Common_Indic_Number_FORMS}', "");
Expect(1, 43071, '\P{^ 	Common_Indic_Number_FORMS}', "");
Expect(0, 43072, '\p{ 	Common_Indic_Number_FORMS}', "");
Expect(1, 43072, '\p{^ 	Common_Indic_Number_FORMS}', "");
Expect(1, 43072, '\P{ 	Common_Indic_Number_FORMS}', "");
Expect(0, 43072, '\P{^ 	Common_Indic_Number_FORMS}', "");
Error('\p{:=		IS_COMMON_Indic_Number_forms}');
Error('\P{:=		IS_COMMON_Indic_Number_forms}');
Expect(1, 43071, '\p{iscommonindicnumberforms}', "");
Expect(0, 43071, '\p{^iscommonindicnumberforms}', "");
Expect(0, 43071, '\P{iscommonindicnumberforms}', "");
Expect(1, 43071, '\P{^iscommonindicnumberforms}', "");
Expect(0, 43072, '\p{iscommonindicnumberforms}', "");
Expect(1, 43072, '\p{^iscommonindicnumberforms}', "");
Expect(1, 43072, '\P{iscommonindicnumberforms}', "");
Expect(0, 43072, '\P{^iscommonindicnumberforms}', "");
Expect(1, 43071, '\p{	-Is_Common_Indic_Number_FORMS}', "");
Expect(0, 43071, '\p{^	-Is_Common_Indic_Number_FORMS}', "");
Expect(0, 43071, '\P{	-Is_Common_Indic_Number_FORMS}', "");
Expect(1, 43071, '\P{^	-Is_Common_Indic_Number_FORMS}', "");
Expect(0, 43072, '\p{	-Is_Common_Indic_Number_FORMS}', "");
Expect(1, 43072, '\p{^	-Is_Common_Indic_Number_FORMS}', "");
Expect(1, 43072, '\P{	-Is_Common_Indic_Number_FORMS}', "");
Expect(0, 43072, '\P{^	-Is_Common_Indic_Number_FORMS}', "");
Error('\p{/a/- In_Common_Indic_Number_Forms}');
Error('\P{/a/- In_Common_Indic_Number_Forms}');
Expect(1, 43071, '\p{incommonindicnumberforms}', "");
Expect(0, 43071, '\p{^incommonindicnumberforms}', "");
Expect(0, 43071, '\P{incommonindicnumberforms}', "");
Expect(1, 43071, '\P{^incommonindicnumberforms}', "");
Expect(0, 43072, '\p{incommonindicnumberforms}', "");
Expect(1, 43072, '\p{^incommonindicnumberforms}', "");
Expect(1, 43072, '\P{incommonindicnumberforms}', "");
Expect(0, 43072, '\P{^incommonindicnumberforms}', "");
Expect(1, 43071, '\p{ 	In_Common_Indic_Number_Forms}', "");
Expect(0, 43071, '\p{^ 	In_Common_Indic_Number_Forms}', "");
Expect(0, 43071, '\P{ 	In_Common_Indic_Number_Forms}', "");
Expect(1, 43071, '\P{^ 	In_Common_Indic_Number_Forms}', "");
Expect(0, 43072, '\p{ 	In_Common_Indic_Number_Forms}', "");
Expect(1, 43072, '\p{^ 	In_Common_Indic_Number_Forms}', "");
Expect(1, 43072, '\P{ 	In_Common_Indic_Number_Forms}', "");
Expect(0, 43072, '\P{^ 	In_Common_Indic_Number_Forms}', "");
Error('\p{-_Indic_Number_forms:=}');
Error('\P{-_Indic_Number_forms:=}');
Expect(1, 43071, '\p{indicnumberforms}', "");
Expect(0, 43071, '\p{^indicnumberforms}', "");
Expect(0, 43071, '\P{indicnumberforms}', "");
Expect(1, 43071, '\P{^indicnumberforms}', "");
Expect(0, 43072, '\p{indicnumberforms}', "");
Expect(1, 43072, '\p{^indicnumberforms}', "");
Expect(1, 43072, '\P{indicnumberforms}', "");
Expect(0, 43072, '\P{^indicnumberforms}', "");
Expect(1, 43071, '\p{	indic_Number_FORMS}', "");
Expect(0, 43071, '\p{^	indic_Number_FORMS}', "");
Expect(0, 43071, '\P{	indic_Number_FORMS}', "");
Expect(1, 43071, '\P{^	indic_Number_FORMS}', "");
Expect(0, 43072, '\p{	indic_Number_FORMS}', "");
Expect(1, 43072, '\p{^	indic_Number_FORMS}', "");
Expect(1, 43072, '\P{	indic_Number_FORMS}', "");
Expect(0, 43072, '\P{^	indic_Number_FORMS}', "");
Error('\p{ is_INDIC_NUMBER_FORMS:=}');
Error('\P{ is_INDIC_NUMBER_FORMS:=}');
Expect(1, 43071, '\p{isindicnumberforms}', "");
Expect(0, 43071, '\p{^isindicnumberforms}', "");
Expect(0, 43071, '\P{isindicnumberforms}', "");
Expect(1, 43071, '\P{^isindicnumberforms}', "");
Expect(0, 43072, '\p{isindicnumberforms}', "");
Expect(1, 43072, '\p{^isindicnumberforms}', "");
Expect(1, 43072, '\P{isindicnumberforms}', "");
Expect(0, 43072, '\P{^isindicnumberforms}', "");
Expect(1, 43071, '\p{-	IS_Indic_number_FORMS}', "");
Expect(0, 43071, '\p{^-	IS_Indic_number_FORMS}', "");
Expect(0, 43071, '\P{-	IS_Indic_number_FORMS}', "");
Expect(1, 43071, '\P{^-	IS_Indic_number_FORMS}', "");
Expect(0, 43072, '\p{-	IS_Indic_number_FORMS}', "");
Expect(1, 43072, '\p{^-	IS_Indic_number_FORMS}', "");
Expect(1, 43072, '\P{-	IS_Indic_number_FORMS}', "");
Expect(0, 43072, '\P{^-	IS_Indic_number_FORMS}', "");
Error('\p{-_In_INDIC_number_Forms/a/}');
Error('\P{-_In_INDIC_number_Forms/a/}');
Expect(1, 43071, '\p{inindicnumberforms}', "");
Expect(0, 43071, '\p{^inindicnumberforms}', "");
Expect(0, 43071, '\P{inindicnumberforms}', "");
Expect(1, 43071, '\P{^inindicnumberforms}', "");
Expect(0, 43072, '\p{inindicnumberforms}', "");
Expect(1, 43072, '\p{^inindicnumberforms}', "");
Expect(1, 43072, '\P{inindicnumberforms}', "");
Expect(0, 43072, '\P{^inindicnumberforms}', "");
Expect(1, 43071, '\p{-In_Indic_Number_Forms}', "");
Expect(0, 43071, '\p{^-In_Indic_Number_Forms}', "");
Expect(0, 43071, '\P{-In_Indic_Number_Forms}', "");
Expect(1, 43071, '\P{^-In_Indic_Number_Forms}', "");
Expect(0, 43072, '\p{-In_Indic_Number_Forms}', "");
Expect(1, 43072, '\p{^-In_Indic_Number_Forms}', "");
Expect(1, 43072, '\P{-In_Indic_Number_Forms}', "");
Expect(0, 43072, '\P{^-In_Indic_Number_Forms}', "");
Error('\p{		Composition_Exclusion/a/}');
Error('\P{		Composition_Exclusion/a/}');
Expect(1, 119232, '\p{compositionexclusion}', "");
Expect(0, 119232, '\p{^compositionexclusion}', "");
Expect(0, 119232, '\P{compositionexclusion}', "");
Expect(1, 119232, '\P{^compositionexclusion}', "");
Expect(0, 119233, '\p{compositionexclusion}', "");
Expect(1, 119233, '\p{^compositionexclusion}', "");
Expect(1, 119233, '\P{compositionexclusion}', "");
Expect(0, 119233, '\P{^compositionexclusion}', "");
Expect(1, 119232, '\p{__COMPOSITION_Exclusion}', "");
Expect(0, 119232, '\p{^__COMPOSITION_Exclusion}', "");
Expect(0, 119232, '\P{__COMPOSITION_Exclusion}', "");
Expect(1, 119232, '\P{^__COMPOSITION_Exclusion}', "");
Expect(0, 119233, '\p{__COMPOSITION_Exclusion}', "");
Expect(1, 119233, '\p{^__COMPOSITION_Exclusion}', "");
Expect(1, 119233, '\P{__COMPOSITION_Exclusion}', "");
Expect(0, 119233, '\P{^__COMPOSITION_Exclusion}', "");
Error('\p{/a/  is_composition_EXCLUSION}');
Error('\P{/a/  is_composition_EXCLUSION}');
Expect(1, 119232, '\p{iscompositionexclusion}', "");
Expect(0, 119232, '\p{^iscompositionexclusion}', "");
Expect(0, 119232, '\P{iscompositionexclusion}', "");
Expect(1, 119232, '\P{^iscompositionexclusion}', "");
Expect(0, 119233, '\p{iscompositionexclusion}', "");
Expect(1, 119233, '\p{^iscompositionexclusion}', "");
Expect(1, 119233, '\P{iscompositionexclusion}', "");
Expect(0, 119233, '\P{^iscompositionexclusion}', "");
Expect(1, 119232, '\p{	Is_composition_exclusion}', "");
Expect(0, 119232, '\p{^	Is_composition_exclusion}', "");
Expect(0, 119232, '\P{	Is_composition_exclusion}', "");
Expect(1, 119232, '\P{^	Is_composition_exclusion}', "");
Expect(0, 119233, '\p{	Is_composition_exclusion}', "");
Expect(1, 119233, '\p{^	Is_composition_exclusion}', "");
Expect(1, 119233, '\P{	Is_composition_exclusion}', "");
Expect(0, 119233, '\P{^	Is_composition_exclusion}', "");
Error('\p{/a/	CE}');
Error('\P{/a/	CE}');
Expect(1, 119232, '\p{ce}', "");
Expect(0, 119232, '\p{^ce}', "");
Expect(0, 119232, '\P{ce}', "");
Expect(1, 119232, '\P{^ce}', "");
Expect(0, 119233, '\p{ce}', "");
Expect(1, 119233, '\p{^ce}', "");
Expect(1, 119233, '\P{ce}', "");
Expect(0, 119233, '\P{^ce}', "");
Expect(1, 119232, '\p{--CE}', "");
Expect(0, 119232, '\p{^--CE}', "");
Expect(0, 119232, '\P{--CE}', "");
Expect(1, 119232, '\P{^--CE}', "");
Expect(0, 119233, '\p{--CE}', "");
Expect(1, 119233, '\p{^--CE}', "");
Expect(1, 119233, '\P{--CE}', "");
Expect(0, 119233, '\P{^--CE}', "");
Error('\p{	:=is_CE}');
Error('\P{	:=is_CE}');
Expect(1, 119232, '\p{isce}', "");
Expect(0, 119232, '\p{^isce}', "");
Expect(0, 119232, '\P{isce}', "");
Expect(1, 119232, '\P{^isce}', "");
Expect(0, 119233, '\p{isce}', "");
Expect(1, 119233, '\p{^isce}', "");
Expect(1, 119233, '\P{isce}', "");
Expect(0, 119233, '\P{^isce}', "");
Expect(1, 119232, '\p{ is_CE}', "");
Expect(0, 119232, '\p{^ is_CE}', "");
Expect(0, 119232, '\P{ is_CE}', "");
Expect(1, 119232, '\P{^ is_CE}', "");
Expect(0, 119233, '\p{ is_CE}', "");
Expect(1, 119233, '\p{^ is_CE}', "");
Expect(1, 119233, '\P{ is_CE}', "");
Expect(0, 119233, '\P{^ is_CE}', "");
Error('\p{	/a/Connector_PUNCTUATION}');
Error('\P{	/a/Connector_PUNCTUATION}');
Expect(1, 65343, '\p{connectorpunctuation}', "");
Expect(0, 65343, '\p{^connectorpunctuation}', "");
Expect(0, 65343, '\P{connectorpunctuation}', "");
Expect(1, 65343, '\P{^connectorpunctuation}', "");
Expect(0, 65344, '\p{connectorpunctuation}', "");
Expect(1, 65344, '\p{^connectorpunctuation}', "");
Expect(1, 65344, '\P{connectorpunctuation}', "");
Expect(0, 65344, '\P{^connectorpunctuation}', "");
Expect(1, 65343, '\p{  CONNECTOR_Punctuation}', "");
Expect(0, 65343, '\p{^  CONNECTOR_Punctuation}', "");
Expect(0, 65343, '\P{  CONNECTOR_Punctuation}', "");
Expect(1, 65343, '\P{^  CONNECTOR_Punctuation}', "");
Expect(0, 65344, '\p{  CONNECTOR_Punctuation}', "");
Expect(1, 65344, '\p{^  CONNECTOR_Punctuation}', "");
Expect(1, 65344, '\P{  CONNECTOR_Punctuation}', "");
Expect(0, 65344, '\P{^  CONNECTOR_Punctuation}', "");
Error('\p{/a/-_IS_Connector_Punctuation}');
Error('\P{/a/-_IS_Connector_Punctuation}');
Expect(1, 65343, '\p{isconnectorpunctuation}', "");
Expect(0, 65343, '\p{^isconnectorpunctuation}', "");
Expect(0, 65343, '\P{isconnectorpunctuation}', "");
Expect(1, 65343, '\P{^isconnectorpunctuation}', "");
Expect(0, 65344, '\p{isconnectorpunctuation}', "");
Expect(1, 65344, '\p{^isconnectorpunctuation}', "");
Expect(1, 65344, '\P{isconnectorpunctuation}', "");
Expect(0, 65344, '\P{^isconnectorpunctuation}', "");
Expect(1, 65343, '\p{--IS_CONNECTOR_Punctuation}', "");
Expect(0, 65343, '\p{^--IS_CONNECTOR_Punctuation}', "");
Expect(0, 65343, '\P{--IS_CONNECTOR_Punctuation}', "");
Expect(1, 65343, '\P{^--IS_CONNECTOR_Punctuation}', "");
Expect(0, 65344, '\p{--IS_CONNECTOR_Punctuation}', "");
Expect(1, 65344, '\p{^--IS_CONNECTOR_Punctuation}', "");
Expect(1, 65344, '\P{--IS_CONNECTOR_Punctuation}', "");
Expect(0, 65344, '\P{^--IS_CONNECTOR_Punctuation}', "");
Error('\p{  Pc:=}');
Error('\P{  Pc:=}');
Expect(1, 65343, '\p{pc}', "");
Expect(0, 65343, '\p{^pc}', "");
Expect(0, 65343, '\P{pc}', "");
Expect(1, 65343, '\P{^pc}', "");
Expect(0, 65344, '\p{pc}', "");
Expect(1, 65344, '\p{^pc}', "");
Expect(1, 65344, '\P{pc}', "");
Expect(0, 65344, '\P{^pc}', "");
Expect(1, 65343, '\p{	-Pc}', "");
Expect(0, 65343, '\p{^	-Pc}', "");
Expect(0, 65343, '\P{	-Pc}', "");
Expect(1, 65343, '\P{^	-Pc}', "");
Expect(0, 65344, '\p{	-Pc}', "");
Expect(1, 65344, '\p{^	-Pc}', "");
Expect(1, 65344, '\P{	-Pc}', "");
Expect(0, 65344, '\P{^	-Pc}', "");
Error('\p{_/a/IS_pc}');
Error('\P{_/a/IS_pc}');
Expect(1, 65343, '\p{ispc}', "");
Expect(0, 65343, '\p{^ispc}', "");
Expect(0, 65343, '\P{ispc}', "");
Expect(1, 65343, '\P{^ispc}', "");
Expect(0, 65344, '\p{ispc}', "");
Expect(1, 65344, '\p{^ispc}', "");
Expect(1, 65344, '\P{ispc}', "");
Expect(0, 65344, '\P{^ispc}', "");
Expect(1, 65343, '\p{ Is_pc}', "");
Expect(0, 65343, '\p{^ Is_pc}', "");
Expect(0, 65343, '\P{ Is_pc}', "");
Expect(1, 65343, '\P{^ Is_pc}', "");
Expect(0, 65344, '\p{ Is_pc}', "");
Expect(1, 65344, '\p{^ Is_pc}', "");
Expect(1, 65344, '\P{ Is_pc}', "");
Expect(0, 65344, '\P{^ Is_pc}', "");
Error('\p{/a/-Control_pictures}');
Error('\P{/a/-Control_pictures}');
Expect(1, 9279, '\p{controlpictures}', "");
Expect(0, 9279, '\p{^controlpictures}', "");
Expect(0, 9279, '\P{controlpictures}', "");
Expect(1, 9279, '\P{^controlpictures}', "");
Expect(0, 9280, '\p{controlpictures}', "");
Expect(1, 9280, '\p{^controlpictures}', "");
Expect(1, 9280, '\P{controlpictures}', "");
Expect(0, 9280, '\P{^controlpictures}', "");
Expect(1, 9279, '\p{	_Control_Pictures}', "");
Expect(0, 9279, '\p{^	_Control_Pictures}', "");
Expect(0, 9279, '\P{	_Control_Pictures}', "");
Expect(1, 9279, '\P{^	_Control_Pictures}', "");
Expect(0, 9280, '\p{	_Control_Pictures}', "");
Expect(1, 9280, '\p{^	_Control_Pictures}', "");
Expect(1, 9280, '\P{	_Control_Pictures}', "");
Expect(0, 9280, '\P{^	_Control_Pictures}', "");
Error('\p{ /a/is_Control_Pictures}');
Error('\P{ /a/is_Control_Pictures}');
Expect(1, 9279, '\p{iscontrolpictures}', "");
Expect(0, 9279, '\p{^iscontrolpictures}', "");
Expect(0, 9279, '\P{iscontrolpictures}', "");
Expect(1, 9279, '\P{^iscontrolpictures}', "");
Expect(0, 9280, '\p{iscontrolpictures}', "");
Expect(1, 9280, '\p{^iscontrolpictures}', "");
Expect(1, 9280, '\P{iscontrolpictures}', "");
Expect(0, 9280, '\P{^iscontrolpictures}', "");
Expect(1, 9279, '\p{		Is_Control_pictures}', "");
Expect(0, 9279, '\p{^		Is_Control_pictures}', "");
Expect(0, 9279, '\P{		Is_Control_pictures}', "");
Expect(1, 9279, '\P{^		Is_Control_pictures}', "");
Expect(0, 9280, '\p{		Is_Control_pictures}', "");
Expect(1, 9280, '\p{^		Is_Control_pictures}', "");
Expect(1, 9280, '\P{		Is_Control_pictures}', "");
Expect(0, 9280, '\P{^		Is_Control_pictures}', "");
Error('\p{ _In_Control_PICTURES:=}');
Error('\P{ _In_Control_PICTURES:=}');
Expect(1, 9279, '\p{incontrolpictures}', "");
Expect(0, 9279, '\p{^incontrolpictures}', "");
Expect(0, 9279, '\P{incontrolpictures}', "");
Expect(1, 9279, '\P{^incontrolpictures}', "");
Expect(0, 9280, '\p{incontrolpictures}', "");
Expect(1, 9280, '\p{^incontrolpictures}', "");
Expect(1, 9280, '\P{incontrolpictures}', "");
Expect(0, 9280, '\P{^incontrolpictures}', "");
Expect(1, 9279, '\p{ _IN_control_PICTURES}', "");
Expect(0, 9279, '\p{^ _IN_control_PICTURES}', "");
Expect(0, 9279, '\P{ _IN_control_PICTURES}', "");
Expect(1, 9279, '\P{^ _IN_control_PICTURES}', "");
Expect(0, 9280, '\p{ _IN_control_PICTURES}', "");
Expect(1, 9280, '\p{^ _IN_control_PICTURES}', "");
Expect(1, 9280, '\P{ _IN_control_PICTURES}', "");
Expect(0, 9280, '\P{^ _IN_control_PICTURES}', "");
Error('\p{/a/ _Coptic}');
Error('\P{/a/ _Coptic}');
Expect(1, 11519, '\p{coptic}', "");
Expect(0, 11519, '\p{^coptic}', "");
Expect(0, 11519, '\P{coptic}', "");
Expect(1, 11519, '\P{^coptic}', "");
Expect(0, 11520, '\p{coptic}', "");
Expect(1, 11520, '\p{^coptic}', "");
Expect(1, 11520, '\P{coptic}', "");
Expect(0, 11520, '\P{^coptic}', "");
Expect(1, 11519, '\p{ -Coptic}', "");
Expect(0, 11519, '\p{^ -Coptic}', "");
Expect(0, 11519, '\P{ -Coptic}', "");
Expect(1, 11519, '\P{^ -Coptic}', "");
Expect(0, 11520, '\p{ -Coptic}', "");
Expect(1, 11520, '\p{^ -Coptic}', "");
Expect(1, 11520, '\P{ -Coptic}', "");
Expect(0, 11520, '\P{^ -Coptic}', "");
Error('\p{:=_Is_Coptic}');
Error('\P{:=_Is_Coptic}');
Expect(1, 11519, '\p{iscoptic}', "");
Expect(0, 11519, '\p{^iscoptic}', "");
Expect(0, 11519, '\P{iscoptic}', "");
Expect(1, 11519, '\P{^iscoptic}', "");
Expect(0, 11520, '\p{iscoptic}', "");
Expect(1, 11520, '\p{^iscoptic}', "");
Expect(1, 11520, '\P{iscoptic}', "");
Expect(0, 11520, '\P{^iscoptic}', "");
Expect(1, 11519, '\p{	Is_COPTIC}', "");
Expect(0, 11519, '\p{^	Is_COPTIC}', "");
Expect(0, 11519, '\P{	Is_COPTIC}', "");
Expect(1, 11519, '\P{^	Is_COPTIC}', "");
Expect(0, 11520, '\p{	Is_COPTIC}', "");
Expect(1, 11520, '\p{^	Is_COPTIC}', "");
Expect(1, 11520, '\P{	Is_COPTIC}', "");
Expect(0, 11520, '\P{^	Is_COPTIC}', "");
Error('\p{/a/ 	Copt}');
Error('\P{/a/ 	Copt}');
Expect(1, 11519, '\p{copt}', "");
Expect(0, 11519, '\p{^copt}', "");
Expect(0, 11519, '\P{copt}', "");
Expect(1, 11519, '\P{^copt}', "");
Expect(0, 11520, '\p{copt}', "");
Expect(1, 11520, '\p{^copt}', "");
Expect(1, 11520, '\P{copt}', "");
Expect(0, 11520, '\P{^copt}', "");
Expect(1, 11519, '\p{	COPT}', "");
Expect(0, 11519, '\p{^	COPT}', "");
Expect(0, 11519, '\P{	COPT}', "");
Expect(1, 11519, '\P{^	COPT}', "");
Expect(0, 11520, '\p{	COPT}', "");
Expect(1, 11520, '\p{^	COPT}', "");
Expect(1, 11520, '\P{	COPT}', "");
Expect(0, 11520, '\P{^	COPT}', "");
Error('\p{:=- Is_Copt}');
Error('\P{:=- Is_Copt}');
Expect(1, 11519, '\p{iscopt}', "");
Expect(0, 11519, '\p{^iscopt}', "");
Expect(0, 11519, '\P{iscopt}', "");
Expect(1, 11519, '\P{^iscopt}', "");
Expect(0, 11520, '\p{iscopt}', "");
Expect(1, 11520, '\p{^iscopt}', "");
Expect(1, 11520, '\P{iscopt}', "");
Expect(0, 11520, '\P{^iscopt}', "");
Expect(1, 11519, '\p{_-Is_Copt}', "");
Expect(0, 11519, '\p{^_-Is_Copt}', "");
Expect(0, 11519, '\P{_-Is_Copt}', "");
Expect(1, 11519, '\P{^_-Is_Copt}', "");
Expect(0, 11520, '\p{_-Is_Copt}', "");
Expect(1, 11520, '\p{^_-Is_Copt}', "");
Expect(1, 11520, '\P{_-Is_Copt}', "");
Expect(0, 11520, '\P{^_-Is_Copt}', "");
Error('\p{	Qaac/a/}');
Error('\P{	Qaac/a/}');
Expect(1, 11519, '\p{qaac}', "");
Expect(0, 11519, '\p{^qaac}', "");
Expect(0, 11519, '\P{qaac}', "");
Expect(1, 11519, '\P{^qaac}', "");
Expect(0, 11520, '\p{qaac}', "");
Expect(1, 11520, '\p{^qaac}', "");
Expect(1, 11520, '\P{qaac}', "");
Expect(0, 11520, '\P{^qaac}', "");
Expect(1, 11519, '\p{- Qaac}', "");
Expect(0, 11519, '\p{^- Qaac}', "");
Expect(0, 11519, '\P{- Qaac}', "");
Expect(1, 11519, '\P{^- Qaac}', "");
Expect(0, 11520, '\p{- Qaac}', "");
Expect(1, 11520, '\p{^- Qaac}', "");
Expect(1, 11520, '\P{- Qaac}', "");
Expect(0, 11520, '\P{^- Qaac}', "");
Error('\p{-/a/Is_Qaac}');
Error('\P{-/a/Is_Qaac}');
Expect(1, 11519, '\p{isqaac}', "");
Expect(0, 11519, '\p{^isqaac}', "");
Expect(0, 11519, '\P{isqaac}', "");
Expect(1, 11519, '\P{^isqaac}', "");
Expect(0, 11520, '\p{isqaac}', "");
Expect(1, 11520, '\p{^isqaac}', "");
Expect(1, 11520, '\P{isqaac}', "");
Expect(0, 11520, '\P{^isqaac}', "");
Expect(1, 11519, '\p{  IS_qaac}', "");
Expect(0, 11519, '\p{^  IS_qaac}', "");
Expect(0, 11519, '\P{  IS_qaac}', "");
Expect(1, 11519, '\P{^  IS_qaac}', "");
Expect(0, 11520, '\p{  IS_qaac}', "");
Expect(1, 11520, '\p{^  IS_qaac}', "");
Expect(1, 11520, '\P{  IS_qaac}', "");
Expect(0, 11520, '\P{^  IS_qaac}', "");
Error('\p{:=Counting_ROD_NUMERALS}');
Error('\P{:=Counting_ROD_NUMERALS}');
Expect(1, 119679, '\p{countingrodnumerals}', "");
Expect(0, 119679, '\p{^countingrodnumerals}', "");
Expect(0, 119679, '\P{countingrodnumerals}', "");
Expect(1, 119679, '\P{^countingrodnumerals}', "");
Expect(0, 119680, '\p{countingrodnumerals}', "");
Expect(1, 119680, '\p{^countingrodnumerals}', "");
Expect(1, 119680, '\P{countingrodnumerals}', "");
Expect(0, 119680, '\P{^countingrodnumerals}', "");
Expect(1, 119679, '\p{ -Counting_rod_Numerals}', "");
Expect(0, 119679, '\p{^ -Counting_rod_Numerals}', "");
Expect(0, 119679, '\P{ -Counting_rod_Numerals}', "");
Expect(1, 119679, '\P{^ -Counting_rod_Numerals}', "");
Expect(0, 119680, '\p{ -Counting_rod_Numerals}', "");
Expect(1, 119680, '\p{^ -Counting_rod_Numerals}', "");
Expect(1, 119680, '\P{ -Counting_rod_Numerals}', "");
Expect(0, 119680, '\P{^ -Counting_rod_Numerals}', "");
Error('\p{-/a/Is_Counting_Rod_Numerals}');
Error('\P{-/a/Is_Counting_Rod_Numerals}');
Expect(1, 119679, '\p{iscountingrodnumerals}', "");
Expect(0, 119679, '\p{^iscountingrodnumerals}', "");
Expect(0, 119679, '\P{iscountingrodnumerals}', "");
Expect(1, 119679, '\P{^iscountingrodnumerals}', "");
Expect(0, 119680, '\p{iscountingrodnumerals}', "");
Expect(1, 119680, '\p{^iscountingrodnumerals}', "");
Expect(1, 119680, '\P{iscountingrodnumerals}', "");
Expect(0, 119680, '\P{^iscountingrodnumerals}', "");
Expect(1, 119679, '\p{ Is_Counting_ROD_Numerals}', "");
Expect(0, 119679, '\p{^ Is_Counting_ROD_Numerals}', "");
Expect(0, 119679, '\P{ Is_Counting_ROD_Numerals}', "");
Expect(1, 119679, '\P{^ Is_Counting_ROD_Numerals}', "");
Expect(0, 119680, '\p{ Is_Counting_ROD_Numerals}', "");
Expect(1, 119680, '\p{^ Is_Counting_ROD_Numerals}', "");
Expect(1, 119680, '\P{ Is_Counting_ROD_Numerals}', "");
Expect(0, 119680, '\P{^ Is_Counting_ROD_Numerals}', "");
Error('\p{_ In_Counting_rod_numerals:=}');
Error('\P{_ In_Counting_rod_numerals:=}');
Expect(1, 119679, '\p{incountingrodnumerals}', "");
Expect(0, 119679, '\p{^incountingrodnumerals}', "");
Expect(0, 119679, '\P{incountingrodnumerals}', "");
Expect(1, 119679, '\P{^incountingrodnumerals}', "");
Expect(0, 119680, '\p{incountingrodnumerals}', "");
Expect(1, 119680, '\p{^incountingrodnumerals}', "");
Expect(1, 119680, '\P{incountingrodnumerals}', "");
Expect(0, 119680, '\P{^incountingrodnumerals}', "");
Expect(1, 119679, '\p{	IN_COUNTING_Rod_numerals}', "");
Expect(0, 119679, '\p{^	IN_COUNTING_Rod_numerals}', "");
Expect(0, 119679, '\P{	IN_COUNTING_Rod_numerals}', "");
Expect(1, 119679, '\P{^	IN_COUNTING_Rod_numerals}', "");
Expect(0, 119680, '\p{	IN_COUNTING_Rod_numerals}', "");
Expect(1, 119680, '\p{^	IN_COUNTING_Rod_numerals}', "");
Expect(1, 119680, '\P{	IN_COUNTING_Rod_numerals}', "");
Expect(0, 119680, '\P{^	IN_COUNTING_Rod_numerals}', "");
Error('\p{/a/ _COUNTING_Rod}');
Error('\P{/a/ _COUNTING_Rod}');
Expect(1, 119679, '\p{countingrod}', "");
Expect(0, 119679, '\p{^countingrod}', "");
Expect(0, 119679, '\P{countingrod}', "");
Expect(1, 119679, '\P{^countingrod}', "");
Expect(0, 119680, '\p{countingrod}', "");
Expect(1, 119680, '\p{^countingrod}', "");
Expect(1, 119680, '\P{countingrod}', "");
Expect(0, 119680, '\P{^countingrod}', "");
Expect(1, 119679, '\p{- Counting_Rod}', "");
Expect(0, 119679, '\p{^- Counting_Rod}', "");
Expect(0, 119679, '\P{- Counting_Rod}', "");
Expect(1, 119679, '\P{^- Counting_Rod}', "");
Expect(0, 119680, '\p{- Counting_Rod}', "");
Expect(1, 119680, '\p{^- Counting_Rod}', "");
Expect(1, 119680, '\P{- Counting_Rod}', "");
Expect(0, 119680, '\P{^- Counting_Rod}', "");
Error('\p{-IS_Counting_Rod:=}');
Error('\P{-IS_Counting_Rod:=}');
Expect(1, 119679, '\p{iscountingrod}', "");
Expect(0, 119679, '\p{^iscountingrod}', "");
Expect(0, 119679, '\P{iscountingrod}', "");
Expect(1, 119679, '\P{^iscountingrod}', "");
Expect(0, 119680, '\p{iscountingrod}', "");
Expect(1, 119680, '\p{^iscountingrod}', "");
Expect(1, 119680, '\P{iscountingrod}', "");
Expect(0, 119680, '\P{^iscountingrod}', "");
Expect(1, 119679, '\p{ Is_Counting_rod}', "");
Expect(0, 119679, '\p{^ Is_Counting_rod}', "");
Expect(0, 119679, '\P{ Is_Counting_rod}', "");
Expect(1, 119679, '\P{^ Is_Counting_rod}', "");
Expect(0, 119680, '\p{ Is_Counting_rod}', "");
Expect(1, 119680, '\p{^ Is_Counting_rod}', "");
Expect(1, 119680, '\P{ Is_Counting_rod}', "");
Expect(0, 119680, '\P{^ Is_Counting_rod}', "");
Error('\p{/a/ in_COUNTING_Rod}');
Error('\P{/a/ in_COUNTING_Rod}');
Expect(1, 119679, '\p{incountingrod}', "");
Expect(0, 119679, '\p{^incountingrod}', "");
Expect(0, 119679, '\P{incountingrod}', "");
Expect(1, 119679, '\P{^incountingrod}', "");
Expect(0, 119680, '\p{incountingrod}', "");
Expect(1, 119680, '\p{^incountingrod}', "");
Expect(1, 119680, '\P{incountingrod}', "");
Expect(0, 119680, '\P{^incountingrod}', "");
Expect(1, 119679, '\p{ 	In_counting_Rod}', "");
Expect(0, 119679, '\p{^ 	In_counting_Rod}', "");
Expect(0, 119679, '\P{ 	In_counting_Rod}', "");
Expect(1, 119679, '\P{^ 	In_counting_Rod}', "");
Expect(0, 119680, '\p{ 	In_counting_Rod}', "");
Expect(1, 119680, '\p{^ 	In_counting_Rod}', "");
Expect(1, 119680, '\P{ 	In_counting_Rod}', "");
Expect(0, 119680, '\P{^ 	In_counting_Rod}', "");
Error('\p{-cuneiform/a/}');
Error('\P{-cuneiform/a/}');
Expect(1, 74867, '\p{cuneiform}', "");
Expect(0, 74867, '\p{^cuneiform}', "");
Expect(0, 74867, '\P{cuneiform}', "");
Expect(1, 74867, '\P{^cuneiform}', "");
Expect(0, 74868, '\p{cuneiform}', "");
Expect(1, 74868, '\p{^cuneiform}', "");
Expect(1, 74868, '\P{cuneiform}', "");
Expect(0, 74868, '\P{^cuneiform}', "");
Expect(1, 74867, '\p{  cuneiform}', "");
Expect(0, 74867, '\p{^  cuneiform}', "");
Expect(0, 74867, '\P{  cuneiform}', "");
Expect(1, 74867, '\P{^  cuneiform}', "");
Expect(0, 74868, '\p{  cuneiform}', "");
Expect(1, 74868, '\p{^  cuneiform}', "");
Expect(1, 74868, '\P{  cuneiform}', "");
Expect(0, 74868, '\P{^  cuneiform}', "");
Error('\p{_IS_cuneiform/a/}');
Error('\P{_IS_cuneiform/a/}');
Expect(1, 74867, '\p{iscuneiform}', "");
Expect(0, 74867, '\p{^iscuneiform}', "");
Expect(0, 74867, '\P{iscuneiform}', "");
Expect(1, 74867, '\P{^iscuneiform}', "");
Expect(0, 74868, '\p{iscuneiform}', "");
Expect(1, 74868, '\p{^iscuneiform}', "");
Expect(1, 74868, '\P{iscuneiform}', "");
Expect(0, 74868, '\P{^iscuneiform}', "");
Expect(1, 74867, '\p{	IS_Cuneiform}', "");
Expect(0, 74867, '\p{^	IS_Cuneiform}', "");
Expect(0, 74867, '\P{	IS_Cuneiform}', "");
Expect(1, 74867, '\P{^	IS_Cuneiform}', "");
Expect(0, 74868, '\p{	IS_Cuneiform}', "");
Expect(1, 74868, '\p{^	IS_Cuneiform}', "");
Expect(1, 74868, '\P{	IS_Cuneiform}', "");
Expect(0, 74868, '\P{^	IS_Cuneiform}', "");
Error('\p{:=_xsux}');
Error('\P{:=_xsux}');
Expect(1, 74867, '\p{xsux}', "");
Expect(0, 74867, '\p{^xsux}', "");
Expect(0, 74867, '\P{xsux}', "");
Expect(1, 74867, '\P{^xsux}', "");
Expect(0, 74868, '\p{xsux}', "");
Expect(1, 74868, '\p{^xsux}', "");
Expect(1, 74868, '\P{xsux}', "");
Expect(0, 74868, '\P{^xsux}', "");
Expect(1, 74867, '\p{ xsux}', "");
Expect(0, 74867, '\p{^ xsux}', "");
Expect(0, 74867, '\P{ xsux}', "");
Expect(1, 74867, '\P{^ xsux}', "");
Expect(0, 74868, '\p{ xsux}', "");
Expect(1, 74868, '\p{^ xsux}', "");
Expect(1, 74868, '\P{ xsux}', "");
Expect(0, 74868, '\P{^ xsux}', "");
Error('\p{		is_Xsux/a/}');
Error('\P{		is_Xsux/a/}');
Expect(1, 74867, '\p{isxsux}', "");
Expect(0, 74867, '\p{^isxsux}', "");
Expect(0, 74867, '\P{isxsux}', "");
Expect(1, 74867, '\P{^isxsux}', "");
Expect(0, 74868, '\p{isxsux}', "");
Expect(1, 74868, '\p{^isxsux}', "");
Expect(1, 74868, '\P{isxsux}', "");
Expect(0, 74868, '\P{^isxsux}', "");
Expect(1, 74867, '\p{IS_xsux}', "");
Expect(0, 74867, '\p{^IS_xsux}', "");
Expect(0, 74867, '\P{IS_xsux}', "");
Expect(1, 74867, '\P{^IS_xsux}', "");
Expect(0, 74868, '\p{IS_xsux}', "");
Expect(1, 74868, '\p{^IS_xsux}', "");
Expect(1, 74868, '\P{IS_xsux}', "");
Expect(0, 74868, '\P{^IS_xsux}', "");
Error('\p{_/a/CUNEIFORM_numbers_And_punctuation}');
Error('\P{_/a/CUNEIFORM_numbers_And_punctuation}');
Expect(1, 74879, '\p{cuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\p{^cuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\P{cuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\P{^cuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\p{cuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\p{^cuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\P{cuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\P{^cuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\p{_ cuneiform_numbers_And_punctuation}', "");
Expect(0, 74879, '\p{^_ cuneiform_numbers_And_punctuation}', "");
Expect(0, 74879, '\P{_ cuneiform_numbers_And_punctuation}', "");
Expect(1, 74879, '\P{^_ cuneiform_numbers_And_punctuation}', "");
Expect(0, 74880, '\p{_ cuneiform_numbers_And_punctuation}', "");
Expect(1, 74880, '\p{^_ cuneiform_numbers_And_punctuation}', "");
Expect(1, 74880, '\P{_ cuneiform_numbers_And_punctuation}', "");
Expect(0, 74880, '\P{^_ cuneiform_numbers_And_punctuation}', "");
Error('\p{/a/Is_cuneiform_NUMBERS_And_punctuation}');
Error('\P{/a/Is_cuneiform_NUMBERS_And_punctuation}');
Expect(1, 74879, '\p{iscuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\p{^iscuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\P{iscuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\P{^iscuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\p{iscuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\p{^iscuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\P{iscuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\P{^iscuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\p{--IS_Cuneiform_numbers_and_Punctuation}', "");
Expect(0, 74879, '\p{^--IS_Cuneiform_numbers_and_Punctuation}', "");
Expect(0, 74879, '\P{--IS_Cuneiform_numbers_and_Punctuation}', "");
Expect(1, 74879, '\P{^--IS_Cuneiform_numbers_and_Punctuation}', "");
Expect(0, 74880, '\p{--IS_Cuneiform_numbers_and_Punctuation}', "");
Expect(1, 74880, '\p{^--IS_Cuneiform_numbers_and_Punctuation}', "");
Expect(1, 74880, '\P{--IS_Cuneiform_numbers_and_Punctuation}', "");
Expect(0, 74880, '\P{^--IS_Cuneiform_numbers_and_Punctuation}', "");
Error('\p{ 	in_CUNEIFORM_Numbers_And_Punctuation:=}');
Error('\P{ 	in_CUNEIFORM_Numbers_And_Punctuation:=}');
Expect(1, 74879, '\p{incuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\p{^incuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\P{incuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\P{^incuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\p{incuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\p{^incuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\P{incuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\P{^incuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\p{_	In_CUNEIFORM_NUMBERS_And_punctuation}', "");
Expect(0, 74879, '\p{^_	In_CUNEIFORM_NUMBERS_And_punctuation}', "");
Expect(0, 74879, '\P{_	In_CUNEIFORM_NUMBERS_And_punctuation}', "");
Expect(1, 74879, '\P{^_	In_CUNEIFORM_NUMBERS_And_punctuation}', "");
Expect(0, 74880, '\p{_	In_CUNEIFORM_NUMBERS_And_punctuation}', "");
Expect(1, 74880, '\p{^_	In_CUNEIFORM_NUMBERS_And_punctuation}', "");
Expect(1, 74880, '\P{_	In_CUNEIFORM_NUMBERS_And_punctuation}', "");
Expect(0, 74880, '\P{^_	In_CUNEIFORM_NUMBERS_And_punctuation}', "");
Error('\p{ _cuneiform_Numbers/a/}');
Error('\P{ _cuneiform_Numbers/a/}');
Expect(1, 74879, '\p{cuneiformnumbers}', "");
Expect(0, 74879, '\p{^cuneiformnumbers}', "");
Expect(0, 74879, '\P{cuneiformnumbers}', "");
Expect(1, 74879, '\P{^cuneiformnumbers}', "");
Expect(0, 74880, '\p{cuneiformnumbers}', "");
Expect(1, 74880, '\p{^cuneiformnumbers}', "");
Expect(1, 74880, '\P{cuneiformnumbers}', "");
Expect(0, 74880, '\P{^cuneiformnumbers}', "");
Expect(1, 74879, '\p{ Cuneiform_numbers}', "");
Expect(0, 74879, '\p{^ Cuneiform_numbers}', "");
Expect(0, 74879, '\P{ Cuneiform_numbers}', "");
Expect(1, 74879, '\P{^ Cuneiform_numbers}', "");
Expect(0, 74880, '\p{ Cuneiform_numbers}', "");
Expect(1, 74880, '\p{^ Cuneiform_numbers}', "");
Expect(1, 74880, '\P{ Cuneiform_numbers}', "");
Expect(0, 74880, '\P{^ Cuneiform_numbers}', "");
Error('\p{:=--IS_Cuneiform_Numbers}');
Error('\P{:=--IS_Cuneiform_Numbers}');
Expect(1, 74879, '\p{iscuneiformnumbers}', "");
Expect(0, 74879, '\p{^iscuneiformnumbers}', "");
Expect(0, 74879, '\P{iscuneiformnumbers}', "");
Expect(1, 74879, '\P{^iscuneiformnumbers}', "");
Expect(0, 74880, '\p{iscuneiformnumbers}', "");
Expect(1, 74880, '\p{^iscuneiformnumbers}', "");
Expect(1, 74880, '\P{iscuneiformnumbers}', "");
Expect(0, 74880, '\P{^iscuneiformnumbers}', "");
Expect(1, 74879, '\p{_Is_Cuneiform_Numbers}', "");
Expect(0, 74879, '\p{^_Is_Cuneiform_Numbers}', "");
Expect(0, 74879, '\P{_Is_Cuneiform_Numbers}', "");
Expect(1, 74879, '\P{^_Is_Cuneiform_Numbers}', "");
Expect(0, 74880, '\p{_Is_Cuneiform_Numbers}', "");
Expect(1, 74880, '\p{^_Is_Cuneiform_Numbers}', "");
Expect(1, 74880, '\P{_Is_Cuneiform_Numbers}', "");
Expect(0, 74880, '\P{^_Is_Cuneiform_Numbers}', "");
Error('\p{	/a/In_cuneiform_Numbers}');
Error('\P{	/a/In_cuneiform_Numbers}');
Expect(1, 74879, '\p{incuneiformnumbers}', "");
Expect(0, 74879, '\p{^incuneiformnumbers}', "");
Expect(0, 74879, '\P{incuneiformnumbers}', "");
Expect(1, 74879, '\P{^incuneiformnumbers}', "");
Expect(0, 74880, '\p{incuneiformnumbers}', "");
Expect(1, 74880, '\p{^incuneiformnumbers}', "");
Expect(1, 74880, '\P{incuneiformnumbers}', "");
Expect(0, 74880, '\P{^incuneiformnumbers}', "");
Expect(1, 74879, '\p{	In_cuneiform_numbers}', "");
Expect(0, 74879, '\p{^	In_cuneiform_numbers}', "");
Expect(0, 74879, '\P{	In_cuneiform_numbers}', "");
Expect(1, 74879, '\P{^	In_cuneiform_numbers}', "");
Expect(0, 74880, '\p{	In_cuneiform_numbers}', "");
Expect(1, 74880, '\p{^	In_cuneiform_numbers}', "");
Expect(1, 74880, '\P{	In_cuneiform_numbers}', "");
Expect(0, 74880, '\P{^	In_cuneiform_numbers}', "");
Error('\p{ :=Currency_Symbol}');
Error('\P{ :=Currency_Symbol}');
Expect(1, 65510, '\p{currencysymbol}', "");
Expect(0, 65510, '\p{^currencysymbol}', "");
Expect(0, 65510, '\P{currencysymbol}', "");
Expect(1, 65510, '\P{^currencysymbol}', "");
Expect(0, 65511, '\p{currencysymbol}', "");
Expect(1, 65511, '\p{^currencysymbol}', "");
Expect(1, 65511, '\P{currencysymbol}', "");
Expect(0, 65511, '\P{^currencysymbol}', "");
Expect(1, 65510, '\p{		CURRENCY_Symbol}', "");
Expect(0, 65510, '\p{^		CURRENCY_Symbol}', "");
Expect(0, 65510, '\P{		CURRENCY_Symbol}', "");
Expect(1, 65510, '\P{^		CURRENCY_Symbol}', "");
Expect(0, 65511, '\p{		CURRENCY_Symbol}', "");
Expect(1, 65511, '\p{^		CURRENCY_Symbol}', "");
Expect(1, 65511, '\P{		CURRENCY_Symbol}', "");
Expect(0, 65511, '\P{^		CURRENCY_Symbol}', "");
Error('\p{- IS_CURRENCY_SYMBOL/a/}');
Error('\P{- IS_CURRENCY_SYMBOL/a/}');
Expect(1, 65510, '\p{iscurrencysymbol}', "");
Expect(0, 65510, '\p{^iscurrencysymbol}', "");
Expect(0, 65510, '\P{iscurrencysymbol}', "");
Expect(1, 65510, '\P{^iscurrencysymbol}', "");
Expect(0, 65511, '\p{iscurrencysymbol}', "");
Expect(1, 65511, '\p{^iscurrencysymbol}', "");
Expect(1, 65511, '\P{iscurrencysymbol}', "");
Expect(0, 65511, '\P{^iscurrencysymbol}', "");
Expect(1, 65510, '\p{-IS_Currency_SYMBOL}', "");
Expect(0, 65510, '\p{^-IS_Currency_SYMBOL}', "");
Expect(0, 65510, '\P{-IS_Currency_SYMBOL}', "");
Expect(1, 65510, '\P{^-IS_Currency_SYMBOL}', "");
Expect(0, 65511, '\p{-IS_Currency_SYMBOL}', "");
Expect(1, 65511, '\p{^-IS_Currency_SYMBOL}', "");
Expect(1, 65511, '\P{-IS_Currency_SYMBOL}', "");
Expect(0, 65511, '\P{^-IS_Currency_SYMBOL}', "");
Error('\p{	/a/sc}');
Error('\P{	/a/sc}');
Expect(1, 65510, '\p{sc}', "");
Expect(0, 65510, '\p{^sc}', "");
Expect(0, 65510, '\P{sc}', "");
Expect(1, 65510, '\P{^sc}', "");
Expect(0, 65511, '\p{sc}', "");
Expect(1, 65511, '\p{^sc}', "");
Expect(1, 65511, '\P{sc}', "");
Expect(0, 65511, '\P{^sc}', "");
Expect(1, 65510, '\p{_SC}', "");
Expect(0, 65510, '\p{^_SC}', "");
Expect(0, 65510, '\P{_SC}', "");
Expect(1, 65510, '\P{^_SC}', "");
Expect(0, 65511, '\p{_SC}', "");
Expect(1, 65511, '\p{^_SC}', "");
Expect(1, 65511, '\P{_SC}', "");
Expect(0, 65511, '\P{^_SC}', "");
Error('\p{/a/-	Is_Sc}');
Error('\P{/a/-	Is_Sc}');
Expect(1, 65510, '\p{issc}', "");
Expect(0, 65510, '\p{^issc}', "");
Expect(0, 65510, '\P{issc}', "");
Expect(1, 65510, '\P{^issc}', "");
Expect(0, 65511, '\p{issc}', "");
Expect(1, 65511, '\p{^issc}', "");
Expect(1, 65511, '\P{issc}', "");
Expect(0, 65511, '\P{^issc}', "");
Expect(1, 65510, '\p{-IS_SC}', "");
Expect(0, 65510, '\p{^-IS_SC}', "");
Expect(0, 65510, '\P{-IS_SC}', "");
Expect(1, 65510, '\P{^-IS_SC}', "");
Expect(0, 65511, '\p{-IS_SC}', "");
Expect(1, 65511, '\p{^-IS_SC}', "");
Expect(1, 65511, '\P{-IS_SC}', "");
Expect(0, 65511, '\P{^-IS_SC}', "");
Error('\p{	-currency_Symbols/a/}');
Error('\P{	-currency_Symbols/a/}');
Expect(1, 8399, '\p{currencysymbols}', "");
Expect(0, 8399, '\p{^currencysymbols}', "");
Expect(0, 8399, '\P{currencysymbols}', "");
Expect(1, 8399, '\P{^currencysymbols}', "");
Expect(0, 8400, '\p{currencysymbols}', "");
Expect(1, 8400, '\p{^currencysymbols}', "");
Expect(1, 8400, '\P{currencysymbols}', "");
Expect(0, 8400, '\P{^currencysymbols}', "");
Expect(1, 8399, '\p{_	currency_Symbols}', "");
Expect(0, 8399, '\p{^_	currency_Symbols}', "");
Expect(0, 8399, '\P{_	currency_Symbols}', "");
Expect(1, 8399, '\P{^_	currency_Symbols}', "");
Expect(0, 8400, '\p{_	currency_Symbols}', "");
Expect(1, 8400, '\p{^_	currency_Symbols}', "");
Expect(1, 8400, '\P{_	currency_Symbols}', "");
Expect(0, 8400, '\P{^_	currency_Symbols}', "");
Error('\p{/a/	is_Currency_Symbols}');
Error('\P{/a/	is_Currency_Symbols}');
Expect(1, 8399, '\p{iscurrencysymbols}', "");
Expect(0, 8399, '\p{^iscurrencysymbols}', "");
Expect(0, 8399, '\P{iscurrencysymbols}', "");
Expect(1, 8399, '\P{^iscurrencysymbols}', "");
Expect(0, 8400, '\p{iscurrencysymbols}', "");
Expect(1, 8400, '\p{^iscurrencysymbols}', "");
Expect(1, 8400, '\P{iscurrencysymbols}', "");
Expect(0, 8400, '\P{^iscurrencysymbols}', "");
Expect(1, 8399, '\p{ -Is_currency_SYMBOLS}', "");
Expect(0, 8399, '\p{^ -Is_currency_SYMBOLS}', "");
Expect(0, 8399, '\P{ -Is_currency_SYMBOLS}', "");
Expect(1, 8399, '\P{^ -Is_currency_SYMBOLS}', "");
Expect(0, 8400, '\p{ -Is_currency_SYMBOLS}', "");
Expect(1, 8400, '\p{^ -Is_currency_SYMBOLS}', "");
Expect(1, 8400, '\P{ -Is_currency_SYMBOLS}', "");
Expect(0, 8400, '\P{^ -Is_currency_SYMBOLS}', "");
Error('\p{ _In_Currency_Symbols:=}');
Error('\P{ _In_Currency_Symbols:=}');
Expect(1, 8399, '\p{incurrencysymbols}', "");
Expect(0, 8399, '\p{^incurrencysymbols}', "");
Expect(0, 8399, '\P{incurrencysymbols}', "");
Expect(1, 8399, '\P{^incurrencysymbols}', "");
Expect(0, 8400, '\p{incurrencysymbols}', "");
Expect(1, 8400, '\p{^incurrencysymbols}', "");
Expect(1, 8400, '\P{incurrencysymbols}', "");
Expect(0, 8400, '\P{^incurrencysymbols}', "");
Expect(1, 8399, '\p{--in_currency_Symbols}', "");
Expect(0, 8399, '\p{^--in_currency_Symbols}', "");
Expect(0, 8399, '\P{--in_currency_Symbols}', "");
Expect(1, 8399, '\P{^--in_currency_Symbols}', "");
Expect(0, 8400, '\p{--in_currency_Symbols}', "");
Expect(1, 8400, '\p{^--in_currency_Symbols}', "");
Expect(1, 8400, '\P{--in_currency_Symbols}', "");
Expect(0, 8400, '\P{^--in_currency_Symbols}', "");
Error('\p{/a/ _Cypriot}');
Error('\P{/a/ _Cypriot}');
Expect(1, 67647, '\p{cypriot}', "");
Expect(0, 67647, '\p{^cypriot}', "");
Expect(0, 67647, '\P{cypriot}', "");
Expect(1, 67647, '\P{^cypriot}', "");
Expect(0, 67648, '\p{cypriot}', "");
Expect(1, 67648, '\p{^cypriot}', "");
Expect(1, 67648, '\P{cypriot}', "");
Expect(0, 67648, '\P{^cypriot}', "");
Expect(1, 67647, '\p{-Cypriot}', "");
Expect(0, 67647, '\p{^-Cypriot}', "");
Expect(0, 67647, '\P{-Cypriot}', "");
Expect(1, 67647, '\P{^-Cypriot}', "");
Expect(0, 67648, '\p{-Cypriot}', "");
Expect(1, 67648, '\p{^-Cypriot}', "");
Expect(1, 67648, '\P{-Cypriot}', "");
Expect(0, 67648, '\P{^-Cypriot}', "");
Error('\p{__Is_Cypriot/a/}');
Error('\P{__Is_Cypriot/a/}');
Expect(1, 67647, '\p{iscypriot}', "");
Expect(0, 67647, '\p{^iscypriot}', "");
Expect(0, 67647, '\P{iscypriot}', "");
Expect(1, 67647, '\P{^iscypriot}', "");
Expect(0, 67648, '\p{iscypriot}', "");
Expect(1, 67648, '\p{^iscypriot}', "");
Expect(1, 67648, '\P{iscypriot}', "");
Expect(0, 67648, '\P{^iscypriot}', "");
Expect(1, 67647, '\p{	 IS_Cypriot}', "");
Expect(0, 67647, '\p{^	 IS_Cypriot}', "");
Expect(0, 67647, '\P{	 IS_Cypriot}', "");
Expect(1, 67647, '\P{^	 IS_Cypriot}', "");
Expect(0, 67648, '\p{	 IS_Cypriot}', "");
Expect(1, 67648, '\p{^	 IS_Cypriot}', "");
Expect(1, 67648, '\P{	 IS_Cypriot}', "");
Expect(0, 67648, '\P{^	 IS_Cypriot}', "");
Error('\p{-/a/Cprt}');
Error('\P{-/a/Cprt}');
Expect(1, 67647, '\p{cprt}', "");
Expect(0, 67647, '\p{^cprt}', "");
Expect(0, 67647, '\P{cprt}', "");
Expect(1, 67647, '\P{^cprt}', "");
Expect(0, 67648, '\p{cprt}', "");
Expect(1, 67648, '\p{^cprt}', "");
Expect(1, 67648, '\P{cprt}', "");
Expect(0, 67648, '\P{^cprt}', "");
Expect(1, 67647, '\p{-CPRT}', "");
Expect(0, 67647, '\p{^-CPRT}', "");
Expect(0, 67647, '\P{-CPRT}', "");
Expect(1, 67647, '\P{^-CPRT}', "");
Expect(0, 67648, '\p{-CPRT}', "");
Expect(1, 67648, '\p{^-CPRT}', "");
Expect(1, 67648, '\P{-CPRT}', "");
Expect(0, 67648, '\P{^-CPRT}', "");
Error('\p{_ is_cprt/a/}');
Error('\P{_ is_cprt/a/}');
Expect(1, 67647, '\p{iscprt}', "");
Expect(0, 67647, '\p{^iscprt}', "");
Expect(0, 67647, '\P{iscprt}', "");
Expect(1, 67647, '\P{^iscprt}', "");
Expect(0, 67648, '\p{iscprt}', "");
Expect(1, 67648, '\p{^iscprt}', "");
Expect(1, 67648, '\P{iscprt}', "");
Expect(0, 67648, '\P{^iscprt}', "");
Expect(1, 67647, '\p{  Is_Cprt}', "");
Expect(0, 67647, '\p{^  Is_Cprt}', "");
Expect(0, 67647, '\P{  Is_Cprt}', "");
Expect(1, 67647, '\P{^  Is_Cprt}', "");
Expect(0, 67648, '\p{  Is_Cprt}', "");
Expect(1, 67648, '\p{^  Is_Cprt}', "");
Expect(1, 67648, '\P{  Is_Cprt}', "");
Expect(0, 67648, '\P{^  Is_Cprt}', "");
Error('\p{_ Cypriot_Syllabary:=}');
Error('\P{_ Cypriot_Syllabary:=}');
Expect(1, 67647, '\p{cypriotsyllabary}', "");
Expect(0, 67647, '\p{^cypriotsyllabary}', "");
Expect(0, 67647, '\P{cypriotsyllabary}', "");
Expect(1, 67647, '\P{^cypriotsyllabary}', "");
Expect(0, 67648, '\p{cypriotsyllabary}', "");
Expect(1, 67648, '\p{^cypriotsyllabary}', "");
Expect(1, 67648, '\P{cypriotsyllabary}', "");
Expect(0, 67648, '\P{^cypriotsyllabary}', "");
Expect(1, 67647, '\p{-_Cypriot_Syllabary}', "");
Expect(0, 67647, '\p{^-_Cypriot_Syllabary}', "");
Expect(0, 67647, '\P{-_Cypriot_Syllabary}', "");
Expect(1, 67647, '\P{^-_Cypriot_Syllabary}', "");
Expect(0, 67648, '\p{-_Cypriot_Syllabary}', "");
Expect(1, 67648, '\p{^-_Cypriot_Syllabary}', "");
Expect(1, 67648, '\P{-_Cypriot_Syllabary}', "");
Expect(0, 67648, '\P{^-_Cypriot_Syllabary}', "");
Error('\p{	-Is_Cypriot_SYLLABARY/a/}');
Error('\P{	-Is_Cypriot_SYLLABARY/a/}');
Expect(1, 67647, '\p{iscypriotsyllabary}', "");
Expect(0, 67647, '\p{^iscypriotsyllabary}', "");
Expect(0, 67647, '\P{iscypriotsyllabary}', "");
Expect(1, 67647, '\P{^iscypriotsyllabary}', "");
Expect(0, 67648, '\p{iscypriotsyllabary}', "");
Expect(1, 67648, '\p{^iscypriotsyllabary}', "");
Expect(1, 67648, '\P{iscypriotsyllabary}', "");
Expect(0, 67648, '\P{^iscypriotsyllabary}', "");
Expect(1, 67647, '\p{_	Is_Cypriot_Syllabary}', "");
Expect(0, 67647, '\p{^_	Is_Cypriot_Syllabary}', "");
Expect(0, 67647, '\P{_	Is_Cypriot_Syllabary}', "");
Expect(1, 67647, '\P{^_	Is_Cypriot_Syllabary}', "");
Expect(0, 67648, '\p{_	Is_Cypriot_Syllabary}', "");
Expect(1, 67648, '\p{^_	Is_Cypriot_Syllabary}', "");
Expect(1, 67648, '\P{_	Is_Cypriot_Syllabary}', "");
Expect(0, 67648, '\P{^_	Is_Cypriot_Syllabary}', "");
Error('\p{/a/		In_CYPRIOT_SYLLABARY}');
Error('\P{/a/		In_CYPRIOT_SYLLABARY}');
Expect(1, 67647, '\p{incypriotsyllabary}', "");
Expect(0, 67647, '\p{^incypriotsyllabary}', "");
Expect(0, 67647, '\P{incypriotsyllabary}', "");
Expect(1, 67647, '\P{^incypriotsyllabary}', "");
Expect(0, 67648, '\p{incypriotsyllabary}', "");
Expect(1, 67648, '\p{^incypriotsyllabary}', "");
Expect(1, 67648, '\P{incypriotsyllabary}', "");
Expect(0, 67648, '\P{^incypriotsyllabary}', "");
Expect(1, 67647, '\p{IN_cypriot_Syllabary}', "");
Expect(0, 67647, '\p{^IN_cypriot_Syllabary}', "");
Expect(0, 67647, '\P{IN_cypriot_Syllabary}', "");
Expect(1, 67647, '\P{^IN_cypriot_Syllabary}', "");
Expect(0, 67648, '\p{IN_cypriot_Syllabary}', "");
Expect(1, 67648, '\p{^IN_cypriot_Syllabary}', "");
Expect(1, 67648, '\P{IN_cypriot_Syllabary}', "");
Expect(0, 67648, '\P{^IN_cypriot_Syllabary}', "");
Error('\p{/a/CYRILLIC}');
Error('\P{/a/CYRILLIC}');
Expect(1, 42655, '\p{cyrillic}', "");
Expect(0, 42655, '\p{^cyrillic}', "");
Expect(0, 42655, '\P{cyrillic}', "");
Expect(1, 42655, '\P{^cyrillic}', "");
Expect(0, 42656, '\p{cyrillic}', "");
Expect(1, 42656, '\p{^cyrillic}', "");
Expect(1, 42656, '\P{cyrillic}', "");
Expect(0, 42656, '\P{^cyrillic}', "");
Expect(1, 42655, '\p{-	Cyrillic}', "");
Expect(0, 42655, '\p{^-	Cyrillic}', "");
Expect(0, 42655, '\P{-	Cyrillic}', "");
Expect(1, 42655, '\P{^-	Cyrillic}', "");
Expect(0, 42656, '\p{-	Cyrillic}', "");
Expect(1, 42656, '\p{^-	Cyrillic}', "");
Expect(1, 42656, '\P{-	Cyrillic}', "");
Expect(0, 42656, '\P{^-	Cyrillic}', "");
Error('\p{_/a/Is_cyrillic}');
Error('\P{_/a/Is_cyrillic}');
Expect(1, 42655, '\p{iscyrillic}', "");
Expect(0, 42655, '\p{^iscyrillic}', "");
Expect(0, 42655, '\P{iscyrillic}', "");
Expect(1, 42655, '\P{^iscyrillic}', "");
Expect(0, 42656, '\p{iscyrillic}', "");
Expect(1, 42656, '\p{^iscyrillic}', "");
Expect(1, 42656, '\P{iscyrillic}', "");
Expect(0, 42656, '\P{^iscyrillic}', "");
Expect(1, 42655, '\p{	 IS_cyrillic}', "");
Expect(0, 42655, '\p{^	 IS_cyrillic}', "");
Expect(0, 42655, '\P{	 IS_cyrillic}', "");
Expect(1, 42655, '\P{^	 IS_cyrillic}', "");
Expect(0, 42656, '\p{	 IS_cyrillic}', "");
Expect(1, 42656, '\p{^	 IS_cyrillic}', "");
Expect(1, 42656, '\P{	 IS_cyrillic}', "");
Expect(0, 42656, '\P{^	 IS_cyrillic}', "");
Error('\p{__cyrl:=}');
Error('\P{__cyrl:=}');
Expect(1, 42655, '\p{cyrl}', "");
Expect(0, 42655, '\p{^cyrl}', "");
Expect(0, 42655, '\P{cyrl}', "");
Expect(1, 42655, '\P{^cyrl}', "");
Expect(0, 42656, '\p{cyrl}', "");
Expect(1, 42656, '\p{^cyrl}', "");
Expect(1, 42656, '\P{cyrl}', "");
Expect(0, 42656, '\P{^cyrl}', "");
Expect(1, 42655, '\p{	Cyrl}', "");
Expect(0, 42655, '\p{^	Cyrl}', "");
Expect(0, 42655, '\P{	Cyrl}', "");
Expect(1, 42655, '\P{^	Cyrl}', "");
Expect(0, 42656, '\p{	Cyrl}', "");
Expect(1, 42656, '\p{^	Cyrl}', "");
Expect(1, 42656, '\P{	Cyrl}', "");
Expect(0, 42656, '\P{^	Cyrl}', "");
Error('\p{	IS_Cyrl:=}');
Error('\P{	IS_Cyrl:=}');
Expect(1, 42655, '\p{iscyrl}', "");
Expect(0, 42655, '\p{^iscyrl}', "");
Expect(0, 42655, '\P{iscyrl}', "");
Expect(1, 42655, '\P{^iscyrl}', "");
Expect(0, 42656, '\p{iscyrl}', "");
Expect(1, 42656, '\p{^iscyrl}', "");
Expect(1, 42656, '\P{iscyrl}', "");
Expect(0, 42656, '\P{^iscyrl}', "");
Expect(1, 42655, '\p{	-Is_cyrl}', "");
Expect(0, 42655, '\p{^	-Is_cyrl}', "");
Expect(0, 42655, '\P{	-Is_cyrl}', "");
Expect(1, 42655, '\P{^	-Is_cyrl}', "");
Expect(0, 42656, '\p{	-Is_cyrl}', "");
Expect(1, 42656, '\p{^	-Is_cyrl}', "");
Expect(1, 42656, '\P{	-Is_cyrl}', "");
Expect(0, 42656, '\P{^	-Is_cyrl}', "");
Error('\p{:=  CYRILLIC_Extended_A}');
Error('\P{:=  CYRILLIC_Extended_A}');
Expect(1, 11775, '\p{cyrillicextendeda}', "");
Expect(0, 11775, '\p{^cyrillicextendeda}', "");
Expect(0, 11775, '\P{cyrillicextendeda}', "");
Expect(1, 11775, '\P{^cyrillicextendeda}', "");
Expect(0, 11776, '\p{cyrillicextendeda}', "");
Expect(1, 11776, '\p{^cyrillicextendeda}', "");
Expect(1, 11776, '\P{cyrillicextendeda}', "");
Expect(0, 11776, '\P{^cyrillicextendeda}', "");
Expect(1, 11775, '\p{_CYRILLIC_extended_A}', "");
Expect(0, 11775, '\p{^_CYRILLIC_extended_A}', "");
Expect(0, 11775, '\P{_CYRILLIC_extended_A}', "");
Expect(1, 11775, '\P{^_CYRILLIC_extended_A}', "");
Expect(0, 11776, '\p{_CYRILLIC_extended_A}', "");
Expect(1, 11776, '\p{^_CYRILLIC_extended_A}', "");
Expect(1, 11776, '\P{_CYRILLIC_extended_A}', "");
Expect(0, 11776, '\P{^_CYRILLIC_extended_A}', "");
Error('\p{	:=IS_CYRILLIC_Extended_A}');
Error('\P{	:=IS_CYRILLIC_Extended_A}');
Expect(1, 11775, '\p{iscyrillicextendeda}', "");
Expect(0, 11775, '\p{^iscyrillicextendeda}', "");
Expect(0, 11775, '\P{iscyrillicextendeda}', "");
Expect(1, 11775, '\P{^iscyrillicextendeda}', "");
Expect(0, 11776, '\p{iscyrillicextendeda}', "");
Expect(1, 11776, '\p{^iscyrillicextendeda}', "");
Expect(1, 11776, '\P{iscyrillicextendeda}', "");
Expect(0, 11776, '\P{^iscyrillicextendeda}', "");
Expect(1, 11775, '\p{_is_Cyrillic_Extended_a}', "");
Expect(0, 11775, '\p{^_is_Cyrillic_Extended_a}', "");
Expect(0, 11775, '\P{_is_Cyrillic_Extended_a}', "");
Expect(1, 11775, '\P{^_is_Cyrillic_Extended_a}', "");
Expect(0, 11776, '\p{_is_Cyrillic_Extended_a}', "");
Expect(1, 11776, '\p{^_is_Cyrillic_Extended_a}', "");
Expect(1, 11776, '\P{_is_Cyrillic_Extended_a}', "");
Expect(0, 11776, '\P{^_is_Cyrillic_Extended_a}', "");
Error('\p{--In_Cyrillic_extended_A/a/}');
Error('\P{--In_Cyrillic_extended_A/a/}');
Expect(1, 11775, '\p{incyrillicextendeda}', "");
Expect(0, 11775, '\p{^incyrillicextendeda}', "");
Expect(0, 11775, '\P{incyrillicextendeda}', "");
Expect(1, 11775, '\P{^incyrillicextendeda}', "");
Expect(0, 11776, '\p{incyrillicextendeda}', "");
Expect(1, 11776, '\p{^incyrillicextendeda}', "");
Expect(1, 11776, '\P{incyrillicextendeda}', "");
Expect(0, 11776, '\P{^incyrillicextendeda}', "");
Expect(1, 11775, '\p{		In_Cyrillic_EXTENDED_A}', "");
Expect(0, 11775, '\p{^		In_Cyrillic_EXTENDED_A}', "");
Expect(0, 11775, '\P{		In_Cyrillic_EXTENDED_A}', "");
Expect(1, 11775, '\P{^		In_Cyrillic_EXTENDED_A}', "");
Expect(0, 11776, '\p{		In_Cyrillic_EXTENDED_A}', "");
Expect(1, 11776, '\p{^		In_Cyrillic_EXTENDED_A}', "");
Expect(1, 11776, '\P{		In_Cyrillic_EXTENDED_A}', "");
Expect(0, 11776, '\P{^		In_Cyrillic_EXTENDED_A}', "");
Error('\p{:=  CYRILLIC_Ext_A}');
Error('\P{:=  CYRILLIC_Ext_A}');
Expect(1, 11775, '\p{cyrillicexta}', "");
Expect(0, 11775, '\p{^cyrillicexta}', "");
Expect(0, 11775, '\P{cyrillicexta}', "");
Expect(1, 11775, '\P{^cyrillicexta}', "");
Expect(0, 11776, '\p{cyrillicexta}', "");
Expect(1, 11776, '\p{^cyrillicexta}', "");
Expect(1, 11776, '\P{cyrillicexta}', "");
Expect(0, 11776, '\P{^cyrillicexta}', "");
Expect(1, 11775, '\p{ -Cyrillic_ext_a}', "");
Expect(0, 11775, '\p{^ -Cyrillic_ext_a}', "");
Expect(0, 11775, '\P{ -Cyrillic_ext_a}', "");
Expect(1, 11775, '\P{^ -Cyrillic_ext_a}', "");
Expect(0, 11776, '\p{ -Cyrillic_ext_a}', "");
Expect(1, 11776, '\p{^ -Cyrillic_ext_a}', "");
Expect(1, 11776, '\P{ -Cyrillic_ext_a}', "");
Expect(0, 11776, '\P{^ -Cyrillic_ext_a}', "");
Error('\p{	_Is_Cyrillic_Ext_A/a/}');
Error('\P{	_Is_Cyrillic_Ext_A/a/}');
Expect(1, 11775, '\p{iscyrillicexta}', "");
Expect(0, 11775, '\p{^iscyrillicexta}', "");
Expect(0, 11775, '\P{iscyrillicexta}', "");
Expect(1, 11775, '\P{^iscyrillicexta}', "");
Expect(0, 11776, '\p{iscyrillicexta}', "");
Expect(1, 11776, '\p{^iscyrillicexta}', "");
Expect(1, 11776, '\P{iscyrillicexta}', "");
Expect(0, 11776, '\P{^iscyrillicexta}', "");
Expect(1, 11775, '\p{_ IS_cyrillic_Ext_A}', "");
Expect(0, 11775, '\p{^_ IS_cyrillic_Ext_A}', "");
Expect(0, 11775, '\P{_ IS_cyrillic_Ext_A}', "");
Expect(1, 11775, '\P{^_ IS_cyrillic_Ext_A}', "");
Expect(0, 11776, '\p{_ IS_cyrillic_Ext_A}', "");
Expect(1, 11776, '\p{^_ IS_cyrillic_Ext_A}', "");
Expect(1, 11776, '\P{_ IS_cyrillic_Ext_A}', "");
Expect(0, 11776, '\P{^_ IS_cyrillic_Ext_A}', "");
Error('\p{:=_-IN_Cyrillic_EXT_A}');
Error('\P{:=_-IN_Cyrillic_EXT_A}');
Expect(1, 11775, '\p{incyrillicexta}', "");
Expect(0, 11775, '\p{^incyrillicexta}', "");
Expect(0, 11775, '\P{incyrillicexta}', "");
Expect(1, 11775, '\P{^incyrillicexta}', "");
Expect(0, 11776, '\p{incyrillicexta}', "");
Expect(1, 11776, '\p{^incyrillicexta}', "");
Expect(1, 11776, '\P{incyrillicexta}', "");
Expect(0, 11776, '\P{^incyrillicexta}', "");
Expect(1, 11775, '\p{	In_Cyrillic_ext_A}', "");
Expect(0, 11775, '\p{^	In_Cyrillic_ext_A}', "");
Expect(0, 11775, '\P{	In_Cyrillic_ext_A}', "");
Expect(1, 11775, '\P{^	In_Cyrillic_ext_A}', "");
Expect(0, 11776, '\p{	In_Cyrillic_ext_A}', "");
Expect(1, 11776, '\p{^	In_Cyrillic_ext_A}', "");
Expect(1, 11776, '\P{	In_Cyrillic_ext_A}', "");
Expect(0, 11776, '\P{^	In_Cyrillic_ext_A}', "");
Error('\p{ :=Cyrillic_extended_B}');
Error('\P{ :=Cyrillic_extended_B}');
Expect(1, 42655, '\p{cyrillicextendedb}', "");
Expect(0, 42655, '\p{^cyrillicextendedb}', "");
Expect(0, 42655, '\P{cyrillicextendedb}', "");
Expect(1, 42655, '\P{^cyrillicextendedb}', "");
Expect(0, 42656, '\p{cyrillicextendedb}', "");
Expect(1, 42656, '\p{^cyrillicextendedb}', "");
Expect(1, 42656, '\P{cyrillicextendedb}', "");
Expect(0, 42656, '\P{^cyrillicextendedb}', "");
Expect(1, 42655, '\p{	 Cyrillic_extended_b}', "");
Expect(0, 42655, '\p{^	 Cyrillic_extended_b}', "");
Expect(0, 42655, '\P{	 Cyrillic_extended_b}', "");
Expect(1, 42655, '\P{^	 Cyrillic_extended_b}', "");
Expect(0, 42656, '\p{	 Cyrillic_extended_b}', "");
Expect(1, 42656, '\p{^	 Cyrillic_extended_b}', "");
Expect(1, 42656, '\P{	 Cyrillic_extended_b}', "");
Expect(0, 42656, '\P{^	 Cyrillic_extended_b}', "");
Error('\p{:=_IS_Cyrillic_Extended_B}');
Error('\P{:=_IS_Cyrillic_Extended_B}');
Expect(1, 42655, '\p{iscyrillicextendedb}', "");
Expect(0, 42655, '\p{^iscyrillicextendedb}', "");
Expect(0, 42655, '\P{iscyrillicextendedb}', "");
Expect(1, 42655, '\P{^iscyrillicextendedb}', "");
Expect(0, 42656, '\p{iscyrillicextendedb}', "");
Expect(1, 42656, '\p{^iscyrillicextendedb}', "");
Expect(1, 42656, '\P{iscyrillicextendedb}', "");
Expect(0, 42656, '\P{^iscyrillicextendedb}', "");
Expect(1, 42655, '\p{  Is_Cyrillic_extended_B}', "");
Expect(0, 42655, '\p{^  Is_Cyrillic_extended_B}', "");
Expect(0, 42655, '\P{  Is_Cyrillic_extended_B}', "");
Expect(1, 42655, '\P{^  Is_Cyrillic_extended_B}', "");
Expect(0, 42656, '\p{  Is_Cyrillic_extended_B}', "");
Expect(1, 42656, '\p{^  Is_Cyrillic_extended_B}', "");
Expect(1, 42656, '\P{  Is_Cyrillic_extended_B}', "");
Expect(0, 42656, '\P{^  Is_Cyrillic_extended_B}', "");
Error('\p{/a/	_In_Cyrillic_EXTENDED_B}');
Error('\P{/a/	_In_Cyrillic_EXTENDED_B}');
Expect(1, 42655, '\p{incyrillicextendedb}', "");
Expect(0, 42655, '\p{^incyrillicextendedb}', "");
Expect(0, 42655, '\P{incyrillicextendedb}', "");
Expect(1, 42655, '\P{^incyrillicextendedb}', "");
Expect(0, 42656, '\p{incyrillicextendedb}', "");
Expect(1, 42656, '\p{^incyrillicextendedb}', "");
Expect(1, 42656, '\P{incyrillicextendedb}', "");
Expect(0, 42656, '\P{^incyrillicextendedb}', "");
Expect(1, 42655, '\p{ In_Cyrillic_Extended_B}', "");
Expect(0, 42655, '\p{^ In_Cyrillic_Extended_B}', "");
Expect(0, 42655, '\P{ In_Cyrillic_Extended_B}', "");
Expect(1, 42655, '\P{^ In_Cyrillic_Extended_B}', "");
Expect(0, 42656, '\p{ In_Cyrillic_Extended_B}', "");
Expect(1, 42656, '\p{^ In_Cyrillic_Extended_B}', "");
Expect(1, 42656, '\P{ In_Cyrillic_Extended_B}', "");
Expect(0, 42656, '\P{^ In_Cyrillic_Extended_B}', "");
Error('\p{	_Cyrillic_ext_b/a/}');
Error('\P{	_Cyrillic_ext_b/a/}');
Expect(1, 42655, '\p{cyrillicextb}', "");
Expect(0, 42655, '\p{^cyrillicextb}', "");
Expect(0, 42655, '\P{cyrillicextb}', "");
Expect(1, 42655, '\P{^cyrillicextb}', "");
Expect(0, 42656, '\p{cyrillicextb}', "");
Expect(1, 42656, '\p{^cyrillicextb}', "");
Expect(1, 42656, '\P{cyrillicextb}', "");
Expect(0, 42656, '\P{^cyrillicextb}', "");
Expect(1, 42655, '\p{_Cyrillic_Ext_B}', "");
Expect(0, 42655, '\p{^_Cyrillic_Ext_B}', "");
Expect(0, 42655, '\P{_Cyrillic_Ext_B}', "");
Expect(1, 42655, '\P{^_Cyrillic_Ext_B}', "");
Expect(0, 42656, '\p{_Cyrillic_Ext_B}', "");
Expect(1, 42656, '\p{^_Cyrillic_Ext_B}', "");
Expect(1, 42656, '\P{_Cyrillic_Ext_B}', "");
Expect(0, 42656, '\P{^_Cyrillic_Ext_B}', "");
Error('\p{/a/-_IS_cyrillic_Ext_b}');
Error('\P{/a/-_IS_cyrillic_Ext_b}');
Expect(1, 42655, '\p{iscyrillicextb}', "");
Expect(0, 42655, '\p{^iscyrillicextb}', "");
Expect(0, 42655, '\P{iscyrillicextb}', "");
Expect(1, 42655, '\P{^iscyrillicextb}', "");
Expect(0, 42656, '\p{iscyrillicextb}', "");
Expect(1, 42656, '\p{^iscyrillicextb}', "");
Expect(1, 42656, '\P{iscyrillicextb}', "");
Expect(0, 42656, '\P{^iscyrillicextb}', "");
Expect(1, 42655, '\p{--is_Cyrillic_ext_B}', "");
Expect(0, 42655, '\p{^--is_Cyrillic_ext_B}', "");
Expect(0, 42655, '\P{--is_Cyrillic_ext_B}', "");
Expect(1, 42655, '\P{^--is_Cyrillic_ext_B}', "");
Expect(0, 42656, '\p{--is_Cyrillic_ext_B}', "");
Expect(1, 42656, '\p{^--is_Cyrillic_ext_B}', "");
Expect(1, 42656, '\P{--is_Cyrillic_ext_B}', "");
Expect(0, 42656, '\P{^--is_Cyrillic_ext_B}', "");
Error('\p{/a/_	In_cyrillic_Ext_B}');
Error('\P{/a/_	In_cyrillic_Ext_B}');
Expect(1, 42655, '\p{incyrillicextb}', "");
Expect(0, 42655, '\p{^incyrillicextb}', "");
Expect(0, 42655, '\P{incyrillicextb}', "");
Expect(1, 42655, '\P{^incyrillicextb}', "");
Expect(0, 42656, '\p{incyrillicextb}', "");
Expect(1, 42656, '\p{^incyrillicextb}', "");
Expect(1, 42656, '\P{incyrillicextb}', "");
Expect(0, 42656, '\P{^incyrillicextb}', "");
Expect(1, 42655, '\p{-	IN_CYRILLIC_EXT_B}', "");
Expect(0, 42655, '\p{^-	IN_CYRILLIC_EXT_B}', "");
Expect(0, 42655, '\P{-	IN_CYRILLIC_EXT_B}', "");
Expect(1, 42655, '\P{^-	IN_CYRILLIC_EXT_B}', "");
Expect(0, 42656, '\p{-	IN_CYRILLIC_EXT_B}', "");
Expect(1, 42656, '\p{^-	IN_CYRILLIC_EXT_B}', "");
Expect(1, 42656, '\P{-	IN_CYRILLIC_EXT_B}', "");
Expect(0, 42656, '\P{^-	IN_CYRILLIC_EXT_B}', "");
Error('\p{/a/cyrillic_supplement}');
Error('\P{/a/cyrillic_supplement}');
Expect(1, 1327, '\p{cyrillicsupplement}', "");
Expect(0, 1327, '\p{^cyrillicsupplement}', "");
Expect(0, 1327, '\P{cyrillicsupplement}', "");
Expect(1, 1327, '\P{^cyrillicsupplement}', "");
Expect(0, 1328, '\p{cyrillicsupplement}', "");
Expect(1, 1328, '\p{^cyrillicsupplement}', "");
Expect(1, 1328, '\P{cyrillicsupplement}', "");
Expect(0, 1328, '\P{^cyrillicsupplement}', "");
Expect(1, 1327, '\p{__cyrillic_Supplement}', "");
Expect(0, 1327, '\p{^__cyrillic_Supplement}', "");
Expect(0, 1327, '\P{__cyrillic_Supplement}', "");
Expect(1, 1327, '\P{^__cyrillic_Supplement}', "");
Expect(0, 1328, '\p{__cyrillic_Supplement}', "");
Expect(1, 1328, '\p{^__cyrillic_Supplement}', "");
Expect(1, 1328, '\P{__cyrillic_Supplement}', "");
Expect(0, 1328, '\P{^__cyrillic_Supplement}', "");
Error('\p{:=IS_Cyrillic_Supplement}');
Error('\P{:=IS_Cyrillic_Supplement}');
Expect(1, 1327, '\p{iscyrillicsupplement}', "");
Expect(0, 1327, '\p{^iscyrillicsupplement}', "");
Expect(0, 1327, '\P{iscyrillicsupplement}', "");
Expect(1, 1327, '\P{^iscyrillicsupplement}', "");
Expect(0, 1328, '\p{iscyrillicsupplement}', "");
Expect(1, 1328, '\p{^iscyrillicsupplement}', "");
Expect(1, 1328, '\P{iscyrillicsupplement}', "");
Expect(0, 1328, '\P{^iscyrillicsupplement}', "");
Expect(1, 1327, '\p{_Is_cyrillic_SUPPLEMENT}', "");
Expect(0, 1327, '\p{^_Is_cyrillic_SUPPLEMENT}', "");
Expect(0, 1327, '\P{_Is_cyrillic_SUPPLEMENT}', "");
Expect(1, 1327, '\P{^_Is_cyrillic_SUPPLEMENT}', "");
Expect(0, 1328, '\p{_Is_cyrillic_SUPPLEMENT}', "");
Expect(1, 1328, '\p{^_Is_cyrillic_SUPPLEMENT}', "");
Expect(1, 1328, '\P{_Is_cyrillic_SUPPLEMENT}', "");
Expect(0, 1328, '\P{^_Is_cyrillic_SUPPLEMENT}', "");
Error('\p{:=_In_Cyrillic_Supplement}');
Error('\P{:=_In_Cyrillic_Supplement}');
Expect(1, 1327, '\p{incyrillicsupplement}', "");
Expect(0, 1327, '\p{^incyrillicsupplement}', "");
Expect(0, 1327, '\P{incyrillicsupplement}', "");
Expect(1, 1327, '\P{^incyrillicsupplement}', "");
Expect(0, 1328, '\p{incyrillicsupplement}', "");
Expect(1, 1328, '\p{^incyrillicsupplement}', "");
Expect(1, 1328, '\P{incyrillicsupplement}', "");
Expect(0, 1328, '\P{^incyrillicsupplement}', "");
Expect(1, 1327, '\p{--IN_Cyrillic_SUPPLEMENT}', "");
Expect(0, 1327, '\p{^--IN_Cyrillic_SUPPLEMENT}', "");
Expect(0, 1327, '\P{--IN_Cyrillic_SUPPLEMENT}', "");
Expect(1, 1327, '\P{^--IN_Cyrillic_SUPPLEMENT}', "");
Expect(0, 1328, '\p{--IN_Cyrillic_SUPPLEMENT}', "");
Expect(1, 1328, '\p{^--IN_Cyrillic_SUPPLEMENT}', "");
Expect(1, 1328, '\P{--IN_Cyrillic_SUPPLEMENT}', "");
Expect(0, 1328, '\P{^--IN_Cyrillic_SUPPLEMENT}', "");
Error('\p{_-CYRILLIC_sup/a/}');
Error('\P{_-CYRILLIC_sup/a/}');
Expect(1, 1327, '\p{cyrillicsup}', "");
Expect(0, 1327, '\p{^cyrillicsup}', "");
Expect(0, 1327, '\P{cyrillicsup}', "");
Expect(1, 1327, '\P{^cyrillicsup}', "");
Expect(0, 1328, '\p{cyrillicsup}', "");
Expect(1, 1328, '\p{^cyrillicsup}', "");
Expect(1, 1328, '\P{cyrillicsup}', "");
Expect(0, 1328, '\P{^cyrillicsup}', "");
Expect(1, 1327, '\p{_-Cyrillic_sup}', "");
Expect(0, 1327, '\p{^_-Cyrillic_sup}', "");
Expect(0, 1327, '\P{_-Cyrillic_sup}', "");
Expect(1, 1327, '\P{^_-Cyrillic_sup}', "");
Expect(0, 1328, '\p{_-Cyrillic_sup}', "");
Expect(1, 1328, '\p{^_-Cyrillic_sup}', "");
Expect(1, 1328, '\P{_-Cyrillic_sup}', "");
Expect(0, 1328, '\P{^_-Cyrillic_sup}', "");
Error('\p{:=_	Is_CYRILLIC_SUP}');
Error('\P{:=_	Is_CYRILLIC_SUP}');
Expect(1, 1327, '\p{iscyrillicsup}', "");
Expect(0, 1327, '\p{^iscyrillicsup}', "");
Expect(0, 1327, '\P{iscyrillicsup}', "");
Expect(1, 1327, '\P{^iscyrillicsup}', "");
Expect(0, 1328, '\p{iscyrillicsup}', "");
Expect(1, 1328, '\p{^iscyrillicsup}', "");
Expect(1, 1328, '\P{iscyrillicsup}', "");
Expect(0, 1328, '\P{^iscyrillicsup}', "");
Expect(1, 1327, '\p{_Is_Cyrillic_Sup}', "");
Expect(0, 1327, '\p{^_Is_Cyrillic_Sup}', "");
Expect(0, 1327, '\P{_Is_Cyrillic_Sup}', "");
Expect(1, 1327, '\P{^_Is_Cyrillic_Sup}', "");
Expect(0, 1328, '\p{_Is_Cyrillic_Sup}', "");
Expect(1, 1328, '\p{^_Is_Cyrillic_Sup}', "");
Expect(1, 1328, '\P{_Is_Cyrillic_Sup}', "");
Expect(0, 1328, '\P{^_Is_Cyrillic_Sup}', "");
Error('\p{_ in_Cyrillic_sup:=}');
Error('\P{_ in_Cyrillic_sup:=}');
Expect(1, 1327, '\p{incyrillicsup}', "");
Expect(0, 1327, '\p{^incyrillicsup}', "");
Expect(0, 1327, '\P{incyrillicsup}', "");
Expect(1, 1327, '\P{^incyrillicsup}', "");
Expect(0, 1328, '\p{incyrillicsup}', "");
Expect(1, 1328, '\p{^incyrillicsup}', "");
Expect(1, 1328, '\P{incyrillicsup}', "");
Expect(0, 1328, '\P{^incyrillicsup}', "");
Expect(1, 1327, '\p{_-In_Cyrillic_Sup}', "");
Expect(0, 1327, '\p{^_-In_Cyrillic_Sup}', "");
Expect(0, 1327, '\P{_-In_Cyrillic_Sup}', "");
Expect(1, 1327, '\P{^_-In_Cyrillic_Sup}', "");
Expect(0, 1328, '\p{_-In_Cyrillic_Sup}', "");
Expect(1, 1328, '\p{^_-In_Cyrillic_Sup}', "");
Expect(1, 1328, '\P{_-In_Cyrillic_Sup}', "");
Expect(0, 1328, '\P{^_-In_Cyrillic_Sup}', "");
Error('\p{:=Cyrillic_SUPPLEMENTARY}');
Error('\P{:=Cyrillic_SUPPLEMENTARY}');
Expect(1, 1327, '\p{cyrillicsupplementary}', "");
Expect(0, 1327, '\p{^cyrillicsupplementary}', "");
Expect(0, 1327, '\P{cyrillicsupplementary}', "");
Expect(1, 1327, '\P{^cyrillicsupplementary}', "");
Expect(0, 1328, '\p{cyrillicsupplementary}', "");
Expect(1, 1328, '\p{^cyrillicsupplementary}', "");
Expect(1, 1328, '\P{cyrillicsupplementary}', "");
Expect(0, 1328, '\P{^cyrillicsupplementary}', "");
Expect(1, 1327, '\p{ cyrillic_Supplementary}', "");
Expect(0, 1327, '\p{^ cyrillic_Supplementary}', "");
Expect(0, 1327, '\P{ cyrillic_Supplementary}', "");
Expect(1, 1327, '\P{^ cyrillic_Supplementary}', "");
Expect(0, 1328, '\p{ cyrillic_Supplementary}', "");
Expect(1, 1328, '\p{^ cyrillic_Supplementary}', "");
Expect(1, 1328, '\P{ cyrillic_Supplementary}', "");
Expect(0, 1328, '\P{^ cyrillic_Supplementary}', "");
Error('\p{_is_Cyrillic_Supplementary/a/}');
Error('\P{_is_Cyrillic_Supplementary/a/}');
Expect(1, 1327, '\p{iscyrillicsupplementary}', "");
Expect(0, 1327, '\p{^iscyrillicsupplementary}', "");
Expect(0, 1327, '\P{iscyrillicsupplementary}', "");
Expect(1, 1327, '\P{^iscyrillicsupplementary}', "");
Expect(0, 1328, '\p{iscyrillicsupplementary}', "");
Expect(1, 1328, '\p{^iscyrillicsupplementary}', "");
Expect(1, 1328, '\P{iscyrillicsupplementary}', "");
Expect(0, 1328, '\P{^iscyrillicsupplementary}', "");
Expect(1, 1327, '\p{ -is_cyrillic_SUPPLEMENTARY}', "");
Expect(0, 1327, '\p{^ -is_cyrillic_SUPPLEMENTARY}', "");
Expect(0, 1327, '\P{ -is_cyrillic_SUPPLEMENTARY}', "");
Expect(1, 1327, '\P{^ -is_cyrillic_SUPPLEMENTARY}', "");
Expect(0, 1328, '\p{ -is_cyrillic_SUPPLEMENTARY}', "");
Expect(1, 1328, '\p{^ -is_cyrillic_SUPPLEMENTARY}', "");
Expect(1, 1328, '\P{ -is_cyrillic_SUPPLEMENTARY}', "");
Expect(0, 1328, '\P{^ -is_cyrillic_SUPPLEMENTARY}', "");
Error('\p{In_Cyrillic_Supplementary:=}');
Error('\P{In_Cyrillic_Supplementary:=}');
Expect(1, 1327, '\p{incyrillicsupplementary}', "");
Expect(0, 1327, '\p{^incyrillicsupplementary}', "");
Expect(0, 1327, '\P{incyrillicsupplementary}', "");
Expect(1, 1327, '\P{^incyrillicsupplementary}', "");
Expect(0, 1328, '\p{incyrillicsupplementary}', "");
Expect(1, 1328, '\p{^incyrillicsupplementary}', "");
Expect(1, 1328, '\P{incyrillicsupplementary}', "");
Expect(0, 1328, '\P{^incyrillicsupplementary}', "");
Expect(1, 1327, '\p{- IN_cyrillic_Supplementary}', "");
Expect(0, 1327, '\p{^- IN_cyrillic_Supplementary}', "");
Expect(0, 1327, '\P{- IN_cyrillic_Supplementary}', "");
Expect(1, 1327, '\P{^- IN_cyrillic_Supplementary}', "");
Expect(0, 1328, '\p{- IN_cyrillic_Supplementary}', "");
Expect(1, 1328, '\p{^- IN_cyrillic_Supplementary}', "");
Expect(1, 1328, '\P{- IN_cyrillic_Supplementary}', "");
Expect(0, 1328, '\P{^- IN_cyrillic_Supplementary}', "");
Error('\p{:=Dash}');
Error('\P{:=Dash}');
Expect(1, 65293, '\p{dash}', "");
Expect(0, 65293, '\p{^dash}', "");
Expect(0, 65293, '\P{dash}', "");
Expect(1, 65293, '\P{^dash}', "");
Expect(0, 65294, '\p{dash}', "");
Expect(1, 65294, '\p{^dash}', "");
Expect(1, 65294, '\P{dash}', "");
Expect(0, 65294, '\P{^dash}', "");
Expect(1, 65293, '\p{__DASH}', "");
Expect(0, 65293, '\p{^__DASH}', "");
Expect(0, 65293, '\P{__DASH}', "");
Expect(1, 65293, '\P{^__DASH}', "");
Expect(0, 65294, '\p{__DASH}', "");
Expect(1, 65294, '\p{^__DASH}', "");
Expect(1, 65294, '\P{__DASH}', "");
Expect(0, 65294, '\P{^__DASH}', "");
Error('\p{:=  Is_DASH}');
Error('\P{:=  Is_DASH}');
Expect(1, 65293, '\p{isdash}', "");
Expect(0, 65293, '\p{^isdash}', "");
Expect(0, 65293, '\P{isdash}', "");
Expect(1, 65293, '\P{^isdash}', "");
Expect(0, 65294, '\p{isdash}', "");
Expect(1, 65294, '\p{^isdash}', "");
Expect(1, 65294, '\P{isdash}', "");
Expect(0, 65294, '\P{^isdash}', "");
Expect(1, 65293, '\p{	_is_DASH}', "");
Expect(0, 65293, '\p{^	_is_DASH}', "");
Expect(0, 65293, '\P{	_is_DASH}', "");
Expect(1, 65293, '\P{^	_is_DASH}', "");
Expect(0, 65294, '\p{	_is_DASH}', "");
Expect(1, 65294, '\p{^	_is_DASH}', "");
Expect(1, 65294, '\P{	_is_DASH}', "");
Expect(0, 65294, '\P{^	_is_DASH}', "");
Error('\p{	Dash_Punctuation:=}');
Error('\P{	Dash_Punctuation:=}');
Expect(1, 65293, '\p{dashpunctuation}', "");
Expect(0, 65293, '\p{^dashpunctuation}', "");
Expect(0, 65293, '\P{dashpunctuation}', "");
Expect(1, 65293, '\P{^dashpunctuation}', "");
Expect(0, 65294, '\p{dashpunctuation}', "");
Expect(1, 65294, '\p{^dashpunctuation}', "");
Expect(1, 65294, '\P{dashpunctuation}', "");
Expect(0, 65294, '\P{^dashpunctuation}', "");
Expect(1, 65293, '\p{	dash_Punctuation}', "");
Expect(0, 65293, '\p{^	dash_Punctuation}', "");
Expect(0, 65293, '\P{	dash_Punctuation}', "");
Expect(1, 65293, '\P{^	dash_Punctuation}', "");
Expect(0, 65294, '\p{	dash_Punctuation}', "");
Expect(1, 65294, '\p{^	dash_Punctuation}', "");
Expect(1, 65294, '\P{	dash_Punctuation}', "");
Expect(0, 65294, '\P{^	dash_Punctuation}', "");
Error('\p{:=	_Is_dash_PUNCTUATION}');
Error('\P{:=	_Is_dash_PUNCTUATION}');
Expect(1, 65293, '\p{isdashpunctuation}', "");
Expect(0, 65293, '\p{^isdashpunctuation}', "");
Expect(0, 65293, '\P{isdashpunctuation}', "");
Expect(1, 65293, '\P{^isdashpunctuation}', "");
Expect(0, 65294, '\p{isdashpunctuation}', "");
Expect(1, 65294, '\p{^isdashpunctuation}', "");
Expect(1, 65294, '\P{isdashpunctuation}', "");
Expect(0, 65294, '\P{^isdashpunctuation}', "");
Expect(1, 65293, '\p{__IS_Dash_Punctuation}', "");
Expect(0, 65293, '\p{^__IS_Dash_Punctuation}', "");
Expect(0, 65293, '\P{__IS_Dash_Punctuation}', "");
Expect(1, 65293, '\P{^__IS_Dash_Punctuation}', "");
Expect(0, 65294, '\p{__IS_Dash_Punctuation}', "");
Expect(1, 65294, '\p{^__IS_Dash_Punctuation}', "");
Expect(1, 65294, '\P{__IS_Dash_Punctuation}', "");
Expect(0, 65294, '\P{^__IS_Dash_Punctuation}', "");
Error('\p{:=_-Pd}');
Error('\P{:=_-Pd}');
Expect(1, 65293, '\p{pd}', "");
Expect(0, 65293, '\p{^pd}', "");
Expect(0, 65293, '\P{pd}', "");
Expect(1, 65293, '\P{^pd}', "");
Expect(0, 65294, '\p{pd}', "");
Expect(1, 65294, '\p{^pd}', "");
Expect(1, 65294, '\P{pd}', "");
Expect(0, 65294, '\P{^pd}', "");
Expect(1, 65293, '\p{	-pd}', "");
Expect(0, 65293, '\p{^	-pd}', "");
Expect(0, 65293, '\P{	-pd}', "");
Expect(1, 65293, '\P{^	-pd}', "");
Expect(0, 65294, '\p{	-pd}', "");
Expect(1, 65294, '\p{^	-pd}', "");
Expect(1, 65294, '\P{	-pd}', "");
Expect(0, 65294, '\P{^	-pd}', "");
Error('\p{--Is_Pd/a/}');
Error('\P{--Is_Pd/a/}');
Expect(1, 65293, '\p{ispd}', "");
Expect(0, 65293, '\p{^ispd}', "");
Expect(0, 65293, '\P{ispd}', "");
Expect(1, 65293, '\P{^ispd}', "");
Expect(0, 65294, '\p{ispd}', "");
Expect(1, 65294, '\p{^ispd}', "");
Expect(1, 65294, '\P{ispd}', "");
Expect(0, 65294, '\P{^ispd}', "");
Expect(1, 65293, '\p{--IS_Pd}', "");
Expect(0, 65293, '\p{^--IS_Pd}', "");
Expect(0, 65293, '\P{--IS_Pd}', "");
Expect(1, 65293, '\P{^--IS_Pd}', "");
Expect(0, 65294, '\p{--IS_Pd}', "");
Expect(1, 65294, '\p{^--IS_Pd}', "");
Expect(1, 65294, '\P{--IS_Pd}', "");
Expect(0, 65294, '\P{^--IS_Pd}', "");
Error('\p{:=	-DEFAULT_ignorable_code_POINT}');
Error('\P{:=	-DEFAULT_ignorable_code_POINT}');
Expect(1, 921599, '\p{defaultignorablecodepoint}', "");
Expect(0, 921599, '\p{^defaultignorablecodepoint}', "");
Expect(0, 921599, '\P{defaultignorablecodepoint}', "");
Expect(1, 921599, '\P{^defaultignorablecodepoint}', "");
Expect(0, 921600, '\p{defaultignorablecodepoint}', "");
Expect(1, 921600, '\p{^defaultignorablecodepoint}', "");
Expect(1, 921600, '\P{defaultignorablecodepoint}', "");
Expect(0, 921600, '\P{^defaultignorablecodepoint}', "");
Expect(1, 921599, '\p{	 DEFAULT_ignorable_CODE_Point}', "");
Expect(0, 921599, '\p{^	 DEFAULT_ignorable_CODE_Point}', "");
Expect(0, 921599, '\P{	 DEFAULT_ignorable_CODE_Point}', "");
Expect(1, 921599, '\P{^	 DEFAULT_ignorable_CODE_Point}', "");
Expect(0, 921600, '\p{	 DEFAULT_ignorable_CODE_Point}', "");
Expect(1, 921600, '\p{^	 DEFAULT_ignorable_CODE_Point}', "");
Expect(1, 921600, '\P{	 DEFAULT_ignorable_CODE_Point}', "");
Expect(0, 921600, '\P{^	 DEFAULT_ignorable_CODE_Point}', "");
Error('\p{/a/is_Default_IGNORABLE_Code_point}');
Error('\P{/a/is_Default_IGNORABLE_Code_point}');
Expect(1, 921599, '\p{isdefaultignorablecodepoint}', "");
Expect(0, 921599, '\p{^isdefaultignorablecodepoint}', "");
Expect(0, 921599, '\P{isdefaultignorablecodepoint}', "");
Expect(1, 921599, '\P{^isdefaultignorablecodepoint}', "");
Expect(0, 921600, '\p{isdefaultignorablecodepoint}', "");
Expect(1, 921600, '\p{^isdefaultignorablecodepoint}', "");
Expect(1, 921600, '\P{isdefaultignorablecodepoint}', "");
Expect(0, 921600, '\P{^isdefaultignorablecodepoint}', "");
Expect(1, 921599, '\p{-_is_DEFAULT_IGNORABLE_CODE_Point}', "");
Expect(0, 921599, '\p{^-_is_DEFAULT_IGNORABLE_CODE_Point}', "");
Expect(0, 921599, '\P{-_is_DEFAULT_IGNORABLE_CODE_Point}', "");
Expect(1, 921599, '\P{^-_is_DEFAULT_IGNORABLE_CODE_Point}', "");
Expect(0, 921600, '\p{-_is_DEFAULT_IGNORABLE_CODE_Point}', "");
Expect(1, 921600, '\p{^-_is_DEFAULT_IGNORABLE_CODE_Point}', "");
Expect(1, 921600, '\P{-_is_DEFAULT_IGNORABLE_CODE_Point}', "");
Expect(0, 921600, '\P{^-_is_DEFAULT_IGNORABLE_CODE_Point}', "");
Error('\p{:=_	DI}');
Error('\P{:=_	DI}');
Expect(1, 921599, '\p{di}', "");
Expect(0, 921599, '\p{^di}', "");
Expect(0, 921599, '\P{di}', "");
Expect(1, 921599, '\P{^di}', "");
Expect(0, 921600, '\p{di}', "");
Expect(1, 921600, '\p{^di}', "");
Expect(1, 921600, '\P{di}', "");
Expect(0, 921600, '\P{^di}', "");
Expect(1, 921599, '\p{ DI}', "");
Expect(0, 921599, '\p{^ DI}', "");
Expect(0, 921599, '\P{ DI}', "");
Expect(1, 921599, '\P{^ DI}', "");
Expect(0, 921600, '\p{ DI}', "");
Expect(1, 921600, '\p{^ DI}', "");
Expect(1, 921600, '\P{ DI}', "");
Expect(0, 921600, '\P{^ DI}', "");
Error('\p{:=--is_DI}');
Error('\P{:=--is_DI}');
Expect(1, 921599, '\p{isdi}', "");
Expect(0, 921599, '\p{^isdi}', "");
Expect(0, 921599, '\P{isdi}', "");
Expect(1, 921599, '\P{^isdi}', "");
Expect(0, 921600, '\p{isdi}', "");
Expect(1, 921600, '\p{^isdi}', "");
Expect(1, 921600, '\P{isdi}', "");
Expect(0, 921600, '\P{^isdi}', "");
Expect(1, 921599, '\p{_ IS_DI}', "");
Expect(0, 921599, '\p{^_ IS_DI}', "");
Expect(0, 921599, '\P{_ IS_DI}', "");
Expect(1, 921599, '\P{^_ IS_DI}', "");
Expect(0, 921600, '\p{_ IS_DI}', "");
Expect(1, 921600, '\p{^_ IS_DI}', "");
Expect(1, 921600, '\P{_ IS_DI}', "");
Expect(0, 921600, '\P{^_ IS_DI}', "");
Error('\p{ /a/Deprecated}');
Error('\P{ /a/Deprecated}');
Expect(1, 917631, '\p{deprecated}', "");
Expect(0, 917631, '\p{^deprecated}', "");
Expect(0, 917631, '\P{deprecated}', "");
Expect(1, 917631, '\P{^deprecated}', "");
Expect(0, 917632, '\p{deprecated}', "");
Expect(1, 917632, '\p{^deprecated}', "");
Expect(1, 917632, '\P{deprecated}', "");
Expect(0, 917632, '\P{^deprecated}', "");
Expect(1, 917631, '\p{-_deprecated}', "");
Expect(0, 917631, '\p{^-_deprecated}', "");
Expect(0, 917631, '\P{-_deprecated}', "");
Expect(1, 917631, '\P{^-_deprecated}', "");
Expect(0, 917632, '\p{-_deprecated}', "");
Expect(1, 917632, '\p{^-_deprecated}', "");
Expect(1, 917632, '\P{-_deprecated}', "");
Expect(0, 917632, '\P{^-_deprecated}', "");
Error('\p{_IS_deprecated/a/}');
Error('\P{_IS_deprecated/a/}');
Expect(1, 917631, '\p{isdeprecated}', "");
Expect(0, 917631, '\p{^isdeprecated}', "");
Expect(0, 917631, '\P{isdeprecated}', "");
Expect(1, 917631, '\P{^isdeprecated}', "");
Expect(0, 917632, '\p{isdeprecated}', "");
Expect(1, 917632, '\p{^isdeprecated}', "");
Expect(1, 917632, '\P{isdeprecated}', "");
Expect(0, 917632, '\P{^isdeprecated}', "");
Expect(1, 917631, '\p{-	Is_Deprecated}', "");
Expect(0, 917631, '\p{^-	Is_Deprecated}', "");
Expect(0, 917631, '\P{-	Is_Deprecated}', "");
Expect(1, 917631, '\P{^-	Is_Deprecated}', "");
Expect(0, 917632, '\p{-	Is_Deprecated}', "");
Expect(1, 917632, '\p{^-	Is_Deprecated}', "");
Expect(1, 917632, '\P{-	Is_Deprecated}', "");
Expect(0, 917632, '\P{^-	Is_Deprecated}', "");
Error('\p{ /a/Dep}');
Error('\P{ /a/Dep}');
Expect(1, 917631, '\p{dep}', "");
Expect(0, 917631, '\p{^dep}', "");
Expect(0, 917631, '\P{dep}', "");
Expect(1, 917631, '\P{^dep}', "");
Expect(0, 917632, '\p{dep}', "");
Expect(1, 917632, '\p{^dep}', "");
Expect(1, 917632, '\P{dep}', "");
Expect(0, 917632, '\P{^dep}', "");
Expect(1, 917631, '\p{_Dep}', "");
Expect(0, 917631, '\p{^_Dep}', "");
Expect(0, 917631, '\P{_Dep}', "");
Expect(1, 917631, '\P{^_Dep}', "");
Expect(0, 917632, '\p{_Dep}', "");
Expect(1, 917632, '\p{^_Dep}', "");
Expect(1, 917632, '\P{_Dep}', "");
Expect(0, 917632, '\P{^_Dep}', "");
Error('\p{ _Is_DEP:=}');
Error('\P{ _Is_DEP:=}');
Expect(1, 917631, '\p{isdep}', "");
Expect(0, 917631, '\p{^isdep}', "");
Expect(0, 917631, '\P{isdep}', "");
Expect(1, 917631, '\P{^isdep}', "");
Expect(0, 917632, '\p{isdep}', "");
Expect(1, 917632, '\p{^isdep}', "");
Expect(1, 917632, '\P{isdep}', "");
Expect(0, 917632, '\P{^isdep}', "");
Expect(1, 917631, '\p{	is_DEP}', "");
Expect(0, 917631, '\p{^	is_DEP}', "");
Expect(0, 917631, '\P{	is_DEP}', "");
Expect(1, 917631, '\P{^	is_DEP}', "");
Expect(0, 917632, '\p{	is_DEP}', "");
Expect(1, 917632, '\p{^	is_DEP}', "");
Expect(1, 917632, '\P{	is_DEP}', "");
Expect(0, 917632, '\P{^	is_DEP}', "");
Error('\p{_:=Deseret}');
Error('\P{_:=Deseret}');
Expect(1, 66639, '\p{deseret}', "");
Expect(0, 66639, '\p{^deseret}', "");
Expect(0, 66639, '\P{deseret}', "");
Expect(1, 66639, '\P{^deseret}', "");
Expect(0, 66640, '\p{deseret}', "");
Expect(1, 66640, '\p{^deseret}', "");
Expect(1, 66640, '\P{deseret}', "");
Expect(0, 66640, '\P{^deseret}', "");
Expect(1, 66639, '\p{- deseret}', "");
Expect(0, 66639, '\p{^- deseret}', "");
Expect(0, 66639, '\P{- deseret}', "");
Expect(1, 66639, '\P{^- deseret}', "");
Expect(0, 66640, '\p{- deseret}', "");
Expect(1, 66640, '\p{^- deseret}', "");
Expect(1, 66640, '\P{- deseret}', "");
Expect(0, 66640, '\P{^- deseret}', "");
Error('\p{-	Is_Deseret/a/}');
Error('\P{-	Is_Deseret/a/}');
Expect(1, 66639, '\p{isdeseret}', "");
Expect(0, 66639, '\p{^isdeseret}', "");
Expect(0, 66639, '\P{isdeseret}', "");
Expect(1, 66639, '\P{^isdeseret}', "");
Expect(0, 66640, '\p{isdeseret}', "");
Expect(1, 66640, '\p{^isdeseret}', "");
Expect(1, 66640, '\P{isdeseret}', "");
Expect(0, 66640, '\P{^isdeseret}', "");
Expect(1, 66639, '\p{	-IS_Deseret}', "");
Expect(0, 66639, '\p{^	-IS_Deseret}', "");
Expect(0, 66639, '\P{	-IS_Deseret}', "");
Expect(1, 66639, '\P{^	-IS_Deseret}', "");
Expect(0, 66640, '\p{	-IS_Deseret}', "");
Expect(1, 66640, '\p{^	-IS_Deseret}', "");
Expect(1, 66640, '\P{	-IS_Deseret}', "");
Expect(0, 66640, '\P{^	-IS_Deseret}', "");
Error('\p{	:=DSRT}');
Error('\P{	:=DSRT}');
Expect(1, 66639, '\p{dsrt}', "");
Expect(0, 66639, '\p{^dsrt}', "");
Expect(0, 66639, '\P{dsrt}', "");
Expect(1, 66639, '\P{^dsrt}', "");
Expect(0, 66640, '\p{dsrt}', "");
Expect(1, 66640, '\p{^dsrt}', "");
Expect(1, 66640, '\P{dsrt}', "");
Expect(0, 66640, '\P{^dsrt}', "");
Expect(1, 66639, '\p{ -Dsrt}', "");
Expect(0, 66639, '\p{^ -Dsrt}', "");
Expect(0, 66639, '\P{ -Dsrt}', "");
Expect(1, 66639, '\P{^ -Dsrt}', "");
Expect(0, 66640, '\p{ -Dsrt}', "");
Expect(1, 66640, '\p{^ -Dsrt}', "");
Expect(1, 66640, '\P{ -Dsrt}', "");
Expect(0, 66640, '\P{^ -Dsrt}', "");
Error('\p{_/a/is_dsrt}');
Error('\P{_/a/is_dsrt}');
Expect(1, 66639, '\p{isdsrt}', "");
Expect(0, 66639, '\p{^isdsrt}', "");
Expect(0, 66639, '\P{isdsrt}', "");
Expect(1, 66639, '\P{^isdsrt}', "");
Expect(0, 66640, '\p{isdsrt}', "");
Expect(1, 66640, '\p{^isdsrt}', "");
Expect(1, 66640, '\P{isdsrt}', "");
Expect(0, 66640, '\P{^isdsrt}', "");
Expect(1, 66639, '\p{_	IS_Dsrt}', "");
Expect(0, 66639, '\p{^_	IS_Dsrt}', "");
Expect(0, 66639, '\P{_	IS_Dsrt}', "");
Expect(1, 66639, '\P{^_	IS_Dsrt}', "");
Expect(0, 66640, '\p{_	IS_Dsrt}', "");
Expect(1, 66640, '\p{^_	IS_Dsrt}', "");
Expect(1, 66640, '\P{_	IS_Dsrt}', "");
Expect(0, 66640, '\P{^_	IS_Dsrt}', "");
Error('\p{/a/ Devanagari}');
Error('\P{/a/ Devanagari}');
Expect(1, 43259, '\p{devanagari}', "");
Expect(0, 43259, '\p{^devanagari}', "");
Expect(0, 43259, '\P{devanagari}', "");
Expect(1, 43259, '\P{^devanagari}', "");
Expect(0, 43260, '\p{devanagari}', "");
Expect(1, 43260, '\p{^devanagari}', "");
Expect(1, 43260, '\P{devanagari}', "");
Expect(0, 43260, '\P{^devanagari}', "");
Expect(1, 43259, '\p{	 DEVANAGARI}', "");
Expect(0, 43259, '\p{^	 DEVANAGARI}', "");
Expect(0, 43259, '\P{	 DEVANAGARI}', "");
Expect(1, 43259, '\P{^	 DEVANAGARI}', "");
Expect(0, 43260, '\p{	 DEVANAGARI}', "");
Expect(1, 43260, '\p{^	 DEVANAGARI}', "");
Expect(1, 43260, '\P{	 DEVANAGARI}', "");
Expect(0, 43260, '\P{^	 DEVANAGARI}', "");
Error('\p{-:=Is_Devanagari}');
Error('\P{-:=Is_Devanagari}');
Expect(1, 43259, '\p{isdevanagari}', "");
Expect(0, 43259, '\p{^isdevanagari}', "");
Expect(0, 43259, '\P{isdevanagari}', "");
Expect(1, 43259, '\P{^isdevanagari}', "");
Expect(0, 43260, '\p{isdevanagari}', "");
Expect(1, 43260, '\p{^isdevanagari}', "");
Expect(1, 43260, '\P{isdevanagari}', "");
Expect(0, 43260, '\P{^isdevanagari}', "");
Expect(1, 43259, '\p{	_is_devanagari}', "");
Expect(0, 43259, '\p{^	_is_devanagari}', "");
Expect(0, 43259, '\P{	_is_devanagari}', "");
Expect(1, 43259, '\P{^	_is_devanagari}', "");
Expect(0, 43260, '\p{	_is_devanagari}', "");
Expect(1, 43260, '\p{^	_is_devanagari}', "");
Expect(1, 43260, '\P{	_is_devanagari}', "");
Expect(0, 43260, '\P{^	_is_devanagari}', "");
Error('\p{:= _deva}');
Error('\P{:= _deva}');
Expect(1, 43259, '\p{deva}', "");
Expect(0, 43259, '\p{^deva}', "");
Expect(0, 43259, '\P{deva}', "");
Expect(1, 43259, '\P{^deva}', "");
Expect(0, 43260, '\p{deva}', "");
Expect(1, 43260, '\p{^deva}', "");
Expect(1, 43260, '\P{deva}', "");
Expect(0, 43260, '\P{^deva}', "");
Expect(1, 43259, '\p{_-Deva}', "");
Expect(0, 43259, '\p{^_-Deva}', "");
Expect(0, 43259, '\P{_-Deva}', "");
Expect(1, 43259, '\P{^_-Deva}', "");
Expect(0, 43260, '\p{_-Deva}', "");
Expect(1, 43260, '\p{^_-Deva}', "");
Expect(1, 43260, '\P{_-Deva}', "");
Expect(0, 43260, '\P{^_-Deva}', "");
Error('\p{-/a/Is_deva}');
Error('\P{-/a/Is_deva}');
Expect(1, 43259, '\p{isdeva}', "");
Expect(0, 43259, '\p{^isdeva}', "");
Expect(0, 43259, '\P{isdeva}', "");
Expect(1, 43259, '\P{^isdeva}', "");
Expect(0, 43260, '\p{isdeva}', "");
Expect(1, 43260, '\p{^isdeva}', "");
Expect(1, 43260, '\P{isdeva}', "");
Expect(0, 43260, '\P{^isdeva}', "");
Expect(1, 43259, '\p{ IS_Deva}', "");
Expect(0, 43259, '\p{^ IS_Deva}', "");
Expect(0, 43259, '\P{ IS_Deva}', "");
Expect(1, 43259, '\P{^ IS_Deva}', "");
Expect(0, 43260, '\p{ IS_Deva}', "");
Expect(1, 43260, '\p{^ IS_Deva}', "");
Expect(1, 43260, '\P{ IS_Deva}', "");
Expect(0, 43260, '\P{^ IS_Deva}', "");
Error('\p{:=_-Devanagari_Extended}');
Error('\P{:=_-Devanagari_Extended}');
Expect(1, 43263, '\p{devanagariextended}', "");
Expect(0, 43263, '\p{^devanagariextended}', "");
Expect(0, 43263, '\P{devanagariextended}', "");
Expect(1, 43263, '\P{^devanagariextended}', "");
Expect(0, 43264, '\p{devanagariextended}', "");
Expect(1, 43264, '\p{^devanagariextended}', "");
Expect(1, 43264, '\P{devanagariextended}', "");
Expect(0, 43264, '\P{^devanagariextended}', "");
Expect(1, 43263, '\p{ DEVANAGARI_Extended}', "");
Expect(0, 43263, '\p{^ DEVANAGARI_Extended}', "");
Expect(0, 43263, '\P{ DEVANAGARI_Extended}', "");
Expect(1, 43263, '\P{^ DEVANAGARI_Extended}', "");
Expect(0, 43264, '\p{ DEVANAGARI_Extended}', "");
Expect(1, 43264, '\p{^ DEVANAGARI_Extended}', "");
Expect(1, 43264, '\P{ DEVANAGARI_Extended}', "");
Expect(0, 43264, '\P{^ DEVANAGARI_Extended}', "");
Error('\p{/a/_Is_devanagari_extended}');
Error('\P{/a/_Is_devanagari_extended}');
Expect(1, 43263, '\p{isdevanagariextended}', "");
Expect(0, 43263, '\p{^isdevanagariextended}', "");
Expect(0, 43263, '\P{isdevanagariextended}', "");
Expect(1, 43263, '\P{^isdevanagariextended}', "");
Expect(0, 43264, '\p{isdevanagariextended}', "");
Expect(1, 43264, '\p{^isdevanagariextended}', "");
Expect(1, 43264, '\P{isdevanagariextended}', "");
Expect(0, 43264, '\P{^isdevanagariextended}', "");
Expect(1, 43263, '\p{ Is_devanagari_Extended}', "");
Expect(0, 43263, '\p{^ Is_devanagari_Extended}', "");
Expect(0, 43263, '\P{ Is_devanagari_Extended}', "");
Expect(1, 43263, '\P{^ Is_devanagari_Extended}', "");
Expect(0, 43264, '\p{ Is_devanagari_Extended}', "");
Expect(1, 43264, '\p{^ Is_devanagari_Extended}', "");
Expect(1, 43264, '\P{ Is_devanagari_Extended}', "");
Expect(0, 43264, '\P{^ Is_devanagari_Extended}', "");
Error('\p{_in_Devanagari_extended:=}');
Error('\P{_in_Devanagari_extended:=}');
Expect(1, 43263, '\p{indevanagariextended}', "");
Expect(0, 43263, '\p{^indevanagariextended}', "");
Expect(0, 43263, '\P{indevanagariextended}', "");
Expect(1, 43263, '\P{^indevanagariextended}', "");
Expect(0, 43264, '\p{indevanagariextended}', "");
Expect(1, 43264, '\p{^indevanagariextended}', "");
Expect(1, 43264, '\P{indevanagariextended}', "");
Expect(0, 43264, '\P{^indevanagariextended}', "");
Expect(1, 43263, '\p{- IN_Devanagari_extended}', "");
Expect(0, 43263, '\p{^- IN_Devanagari_extended}', "");
Expect(0, 43263, '\P{- IN_Devanagari_extended}', "");
Expect(1, 43263, '\P{^- IN_Devanagari_extended}', "");
Expect(0, 43264, '\p{- IN_Devanagari_extended}', "");
Expect(1, 43264, '\p{^- IN_Devanagari_extended}', "");
Expect(1, 43264, '\P{- IN_Devanagari_extended}', "");
Expect(0, 43264, '\P{^- IN_Devanagari_extended}', "");
Error('\p{/a/ _Devanagari_EXT}');
Error('\P{/a/ _Devanagari_EXT}');
Expect(1, 43263, '\p{devanagariext}', "");
Expect(0, 43263, '\p{^devanagariext}', "");
Expect(0, 43263, '\P{devanagariext}', "");
Expect(1, 43263, '\P{^devanagariext}', "");
Expect(0, 43264, '\p{devanagariext}', "");
Expect(1, 43264, '\p{^devanagariext}', "");
Expect(1, 43264, '\P{devanagariext}', "");
Expect(0, 43264, '\P{^devanagariext}', "");
Expect(1, 43263, '\p{-Devanagari_ext}', "");
Expect(0, 43263, '\p{^-Devanagari_ext}', "");
Expect(0, 43263, '\P{-Devanagari_ext}', "");
Expect(1, 43263, '\P{^-Devanagari_ext}', "");
Expect(0, 43264, '\p{-Devanagari_ext}', "");
Expect(1, 43264, '\p{^-Devanagari_ext}', "");
Expect(1, 43264, '\P{-Devanagari_ext}', "");
Expect(0, 43264, '\P{^-Devanagari_ext}', "");
Error('\p{	IS_devanagari_ext/a/}');
Error('\P{	IS_devanagari_ext/a/}');
Expect(1, 43263, '\p{isdevanagariext}', "");
Expect(0, 43263, '\p{^isdevanagariext}', "");
Expect(0, 43263, '\P{isdevanagariext}', "");
Expect(1, 43263, '\P{^isdevanagariext}', "");
Expect(0, 43264, '\p{isdevanagariext}', "");
Expect(1, 43264, '\p{^isdevanagariext}', "");
Expect(1, 43264, '\P{isdevanagariext}', "");
Expect(0, 43264, '\P{^isdevanagariext}', "");
Expect(1, 43263, '\p{	-Is_Devanagari_EXT}', "");
Expect(0, 43263, '\p{^	-Is_Devanagari_EXT}', "");
Expect(0, 43263, '\P{	-Is_Devanagari_EXT}', "");
Expect(1, 43263, '\P{^	-Is_Devanagari_EXT}', "");
Expect(0, 43264, '\p{	-Is_Devanagari_EXT}', "");
Expect(1, 43264, '\p{^	-Is_Devanagari_EXT}', "");
Expect(1, 43264, '\P{	-Is_Devanagari_EXT}', "");
Expect(0, 43264, '\P{^	-Is_Devanagari_EXT}', "");
Error('\p{/a/_In_devanagari_EXT}');
Error('\P{/a/_In_devanagari_EXT}');
Expect(1, 43263, '\p{indevanagariext}', "");
Expect(0, 43263, '\p{^indevanagariext}', "");
Expect(0, 43263, '\P{indevanagariext}', "");
Expect(1, 43263, '\P{^indevanagariext}', "");
Expect(0, 43264, '\p{indevanagariext}', "");
Expect(1, 43264, '\p{^indevanagariext}', "");
Expect(1, 43264, '\P{indevanagariext}', "");
Expect(0, 43264, '\P{^indevanagariext}', "");
Expect(1, 43263, '\p{	In_devanagari_EXT}', "");
Expect(0, 43263, '\p{^	In_devanagari_EXT}', "");
Expect(0, 43263, '\P{	In_devanagari_EXT}', "");
Expect(1, 43263, '\P{^	In_devanagari_EXT}', "");
Expect(0, 43264, '\p{	In_devanagari_EXT}', "");
Expect(1, 43264, '\p{^	In_devanagari_EXT}', "");
Expect(1, 43264, '\P{	In_devanagari_EXT}', "");
Expect(0, 43264, '\P{^	In_devanagari_EXT}', "");
Error('\p{-:=Diacritic}');
Error('\P{-:=Diacritic}');
Expect(1, 119213, '\p{diacritic}', "");
Expect(0, 119213, '\p{^diacritic}', "");
Expect(0, 119213, '\P{diacritic}', "");
Expect(1, 119213, '\P{^diacritic}', "");
Expect(0, 119214, '\p{diacritic}', "");
Expect(1, 119214, '\p{^diacritic}', "");
Expect(1, 119214, '\P{diacritic}', "");
Expect(0, 119214, '\P{^diacritic}', "");
Expect(1, 119213, '\p{  DIACRITIC}', "");
Expect(0, 119213, '\p{^  DIACRITIC}', "");
Expect(0, 119213, '\P{  DIACRITIC}', "");
Expect(1, 119213, '\P{^  DIACRITIC}', "");
Expect(0, 119214, '\p{  DIACRITIC}', "");
Expect(1, 119214, '\p{^  DIACRITIC}', "");
Expect(1, 119214, '\P{  DIACRITIC}', "");
Expect(0, 119214, '\P{^  DIACRITIC}', "");
Error('\p{:=is_Diacritic}');
Error('\P{:=is_Diacritic}');
Expect(1, 119213, '\p{isdiacritic}', "");
Expect(0, 119213, '\p{^isdiacritic}', "");
Expect(0, 119213, '\P{isdiacritic}', "");
Expect(1, 119213, '\P{^isdiacritic}', "");
Expect(0, 119214, '\p{isdiacritic}', "");
Expect(1, 119214, '\p{^isdiacritic}', "");
Expect(1, 119214, '\P{isdiacritic}', "");
Expect(0, 119214, '\P{^isdiacritic}', "");
Expect(1, 119213, '\p{-is_diacritic}', "");
Expect(0, 119213, '\p{^-is_diacritic}', "");
Expect(0, 119213, '\P{-is_diacritic}', "");
Expect(1, 119213, '\P{^-is_diacritic}', "");
Expect(0, 119214, '\p{-is_diacritic}', "");
Expect(1, 119214, '\p{^-is_diacritic}', "");
Expect(1, 119214, '\P{-is_diacritic}', "");
Expect(0, 119214, '\P{^-is_diacritic}', "");
Error('\p{:=Dia}');
Error('\P{:=Dia}');
Expect(1, 119213, '\p{dia}', "");
Expect(0, 119213, '\p{^dia}', "");
Expect(0, 119213, '\P{dia}', "");
Expect(1, 119213, '\P{^dia}', "");
Expect(0, 119214, '\p{dia}', "");
Expect(1, 119214, '\p{^dia}', "");
Expect(1, 119214, '\P{dia}', "");
Expect(0, 119214, '\P{^dia}', "");
Expect(1, 119213, '\p{	Dia}', "");
Expect(0, 119213, '\p{^	Dia}', "");
Expect(0, 119213, '\P{	Dia}', "");
Expect(1, 119213, '\P{^	Dia}', "");
Expect(0, 119214, '\p{	Dia}', "");
Expect(1, 119214, '\p{^	Dia}', "");
Expect(1, 119214, '\P{	Dia}', "");
Expect(0, 119214, '\P{^	Dia}', "");
Error('\p{ _Is_Dia:=}');
Error('\P{ _Is_Dia:=}');
Expect(1, 119213, '\p{isdia}', "");
Expect(0, 119213, '\p{^isdia}', "");
Expect(0, 119213, '\P{isdia}', "");
Expect(1, 119213, '\P{^isdia}', "");
Expect(0, 119214, '\p{isdia}', "");
Expect(1, 119214, '\p{^isdia}', "");
Expect(1, 119214, '\P{isdia}', "");
Expect(0, 119214, '\P{^isdia}', "");
Expect(1, 119213, '\p{ Is_DIA}', "");
Expect(0, 119213, '\p{^ Is_DIA}', "");
Expect(0, 119213, '\P{ Is_DIA}', "");
Expect(1, 119213, '\P{^ Is_DIA}', "");
Expect(0, 119214, '\p{ Is_DIA}', "");
Expect(1, 119214, '\p{^ Is_DIA}', "");
Expect(1, 119214, '\P{ Is_DIA}', "");
Expect(0, 119214, '\P{^ Is_DIA}', "");
Error('\p{:=- DIGIT}');
Error('\P{:=- DIGIT}');
Expect(1, 120831, '\p{digit}', "");
Expect(0, 120831, '\p{^digit}', "");
Expect(0, 120831, '\P{digit}', "");
Expect(1, 120831, '\P{^digit}', "");
Expect(0, 120832, '\p{digit}', "");
Expect(1, 120832, '\p{^digit}', "");
Expect(1, 120832, '\P{digit}', "");
Expect(0, 120832, '\P{^digit}', "");
Expect(1, 120831, '\p{Digit}', "");
Expect(0, 120831, '\p{^Digit}', "");
Expect(0, 120831, '\P{Digit}', "");
Expect(1, 120831, '\P{^Digit}', "");
Expect(0, 120832, '\p{Digit}', "");
Expect(1, 120832, '\p{^Digit}', "");
Expect(1, 120832, '\P{Digit}', "");
Expect(0, 120832, '\P{^Digit}', "");
Error('\p{	:=XPosixDigit}');
Error('\P{	:=XPosixDigit}');
Expect(1, 120831, '\p{xposixdigit}', "");
Expect(0, 120831, '\p{^xposixdigit}', "");
Expect(0, 120831, '\P{xposixdigit}', "");
Expect(1, 120831, '\P{^xposixdigit}', "");
Expect(0, 120832, '\p{xposixdigit}', "");
Expect(1, 120832, '\p{^xposixdigit}', "");
Expect(1, 120832, '\P{xposixdigit}', "");
Expect(0, 120832, '\P{^xposixdigit}', "");
Expect(1, 120831, '\p{_-XPOSIXDIGIT}', "");
Expect(0, 120831, '\p{^_-XPOSIXDIGIT}', "");
Expect(0, 120831, '\P{_-XPOSIXDIGIT}', "");
Expect(1, 120831, '\P{^_-XPOSIXDIGIT}', "");
Expect(0, 120832, '\p{_-XPOSIXDIGIT}', "");
Expect(1, 120832, '\p{^_-XPOSIXDIGIT}', "");
Expect(1, 120832, '\P{_-XPOSIXDIGIT}', "");
Expect(0, 120832, '\P{^_-XPOSIXDIGIT}', "");
Error('\p{:=	 Is_digit}');
Error('\P{:=	 Is_digit}');
Expect(1, 120831, '\p{isdigit}', "");
Expect(0, 120831, '\p{^isdigit}', "");
Expect(0, 120831, '\P{isdigit}', "");
Expect(1, 120831, '\P{^isdigit}', "");
Expect(0, 120832, '\p{isdigit}', "");
Expect(1, 120832, '\p{^isdigit}', "");
Expect(1, 120832, '\P{isdigit}', "");
Expect(0, 120832, '\P{^isdigit}', "");
Expect(1, 120831, '\p{-Is_Digit}', "");
Expect(0, 120831, '\p{^-Is_Digit}', "");
Expect(0, 120831, '\P{-Is_Digit}', "");
Expect(1, 120831, '\P{^-Is_Digit}', "");
Expect(0, 120832, '\p{-Is_Digit}', "");
Expect(1, 120832, '\p{^-Is_Digit}', "");
Expect(1, 120832, '\P{-Is_Digit}', "");
Expect(0, 120832, '\P{^-Is_Digit}', "");
Error('\p{:=_-IS_xposixdigit}');
Error('\P{:=_-IS_xposixdigit}');
Expect(1, 120831, '\p{isxposixdigit}', "");
Expect(0, 120831, '\p{^isxposixdigit}', "");
Expect(0, 120831, '\P{isxposixdigit}', "");
Expect(1, 120831, '\P{^isxposixdigit}', "");
Expect(0, 120832, '\p{isxposixdigit}', "");
Expect(1, 120832, '\p{^isxposixdigit}', "");
Expect(1, 120832, '\P{isxposixdigit}', "");
Expect(0, 120832, '\P{^isxposixdigit}', "");
Expect(1, 120831, '\p{--is_xposixdigit}', "");
Expect(0, 120831, '\p{^--is_xposixdigit}', "");
Expect(0, 120831, '\P{--is_xposixdigit}', "");
Expect(1, 120831, '\P{^--is_xposixdigit}', "");
Expect(0, 120832, '\p{--is_xposixdigit}', "");
Expect(1, 120832, '\p{^--is_xposixdigit}', "");
Expect(1, 120832, '\P{--is_xposixdigit}', "");
Expect(0, 120832, '\P{^--is_xposixdigit}', "");
Error('\p{/a/_-Decimal_Number}');
Error('\P{/a/_-Decimal_Number}');
Expect(1, 120831, '\p{decimalnumber}', "");
Expect(0, 120831, '\p{^decimalnumber}', "");
Expect(0, 120831, '\P{decimalnumber}', "");
Expect(1, 120831, '\P{^decimalnumber}', "");
Expect(0, 120832, '\p{decimalnumber}', "");
Expect(1, 120832, '\p{^decimalnumber}', "");
Expect(1, 120832, '\P{decimalnumber}', "");
Expect(0, 120832, '\P{^decimalnumber}', "");
Expect(1, 120831, '\p{-	Decimal_number}', "");
Expect(0, 120831, '\p{^-	Decimal_number}', "");
Expect(0, 120831, '\P{-	Decimal_number}', "");
Expect(1, 120831, '\P{^-	Decimal_number}', "");
Expect(0, 120832, '\p{-	Decimal_number}', "");
Expect(1, 120832, '\p{^-	Decimal_number}', "");
Expect(1, 120832, '\P{-	Decimal_number}', "");
Expect(0, 120832, '\P{^-	Decimal_number}', "");
Error('\p{/a/-IS_Decimal_Number}');
Error('\P{/a/-IS_Decimal_Number}');
Expect(1, 120831, '\p{isdecimalnumber}', "");
Expect(0, 120831, '\p{^isdecimalnumber}', "");
Expect(0, 120831, '\P{isdecimalnumber}', "");
Expect(1, 120831, '\P{^isdecimalnumber}', "");
Expect(0, 120832, '\p{isdecimalnumber}', "");
Expect(1, 120832, '\p{^isdecimalnumber}', "");
Expect(1, 120832, '\P{isdecimalnumber}', "");
Expect(0, 120832, '\P{^isdecimalnumber}', "");
Expect(1, 120831, '\p{_Is_Decimal_NUMBER}', "");
Expect(0, 120831, '\p{^_Is_Decimal_NUMBER}', "");
Expect(0, 120831, '\P{_Is_Decimal_NUMBER}', "");
Expect(1, 120831, '\P{^_Is_Decimal_NUMBER}', "");
Expect(0, 120832, '\p{_Is_Decimal_NUMBER}', "");
Expect(1, 120832, '\p{^_Is_Decimal_NUMBER}', "");
Expect(1, 120832, '\P{_Is_Decimal_NUMBER}', "");
Expect(0, 120832, '\P{^_Is_Decimal_NUMBER}', "");
Error('\p{:=_ ND}');
Error('\P{:=_ ND}');
Expect(1, 120831, '\p{nd}', "");
Expect(0, 120831, '\p{^nd}', "");
Expect(0, 120831, '\P{nd}', "");
Expect(1, 120831, '\P{^nd}', "");
Expect(0, 120832, '\p{nd}', "");
Expect(1, 120832, '\p{^nd}', "");
Expect(1, 120832, '\P{nd}', "");
Expect(0, 120832, '\P{^nd}', "");
Expect(1, 120831, '\p{-nd}', "");
Expect(0, 120831, '\p{^-nd}', "");
Expect(0, 120831, '\P{-nd}', "");
Expect(1, 120831, '\P{^-nd}', "");
Expect(0, 120832, '\p{-nd}', "");
Expect(1, 120832, '\p{^-nd}', "");
Expect(1, 120832, '\P{-nd}', "");
Expect(0, 120832, '\P{^-nd}', "");
Error('\p{:=__IS_nd}');
Error('\P{:=__IS_nd}');
Expect(1, 120831, '\p{isnd}', "");
Expect(0, 120831, '\p{^isnd}', "");
Expect(0, 120831, '\P{isnd}', "");
Expect(1, 120831, '\P{^isnd}', "");
Expect(0, 120832, '\p{isnd}', "");
Expect(1, 120832, '\p{^isnd}', "");
Expect(1, 120832, '\P{isnd}', "");
Expect(0, 120832, '\P{^isnd}', "");
Expect(1, 120831, '\p{ IS_Nd}', "");
Expect(0, 120831, '\p{^ IS_Nd}', "");
Expect(0, 120831, '\P{ IS_Nd}', "");
Expect(1, 120831, '\P{^ IS_Nd}', "");
Expect(0, 120832, '\p{ IS_Nd}', "");
Expect(1, 120832, '\p{^ IS_Nd}', "");
Expect(1, 120832, '\P{ IS_Nd}', "");
Expect(0, 120832, '\P{^ IS_Nd}', "");
Error('\p{:=_ Dingbats}');
Error('\P{:=_ Dingbats}');
Expect(1, 10175, '\p{dingbats}', "");
Expect(0, 10175, '\p{^dingbats}', "");
Expect(0, 10175, '\P{dingbats}', "");
Expect(1, 10175, '\P{^dingbats}', "");
Expect(0, 10176, '\p{dingbats}', "");
Expect(1, 10176, '\p{^dingbats}', "");
Expect(1, 10176, '\P{dingbats}', "");
Expect(0, 10176, '\P{^dingbats}', "");
Expect(1, 10175, '\p{ -Dingbats}', "");
Expect(0, 10175, '\p{^ -Dingbats}', "");
Expect(0, 10175, '\P{ -Dingbats}', "");
Expect(1, 10175, '\P{^ -Dingbats}', "");
Expect(0, 10176, '\p{ -Dingbats}', "");
Expect(1, 10176, '\p{^ -Dingbats}', "");
Expect(1, 10176, '\P{ -Dingbats}', "");
Expect(0, 10176, '\P{^ -Dingbats}', "");
Error('\p{  Is_DINGBATS/a/}');
Error('\P{  Is_DINGBATS/a/}');
Expect(1, 10175, '\p{isdingbats}', "");
Expect(0, 10175, '\p{^isdingbats}', "");
Expect(0, 10175, '\P{isdingbats}', "");
Expect(1, 10175, '\P{^isdingbats}', "");
Expect(0, 10176, '\p{isdingbats}', "");
Expect(1, 10176, '\p{^isdingbats}', "");
Expect(1, 10176, '\P{isdingbats}', "");
Expect(0, 10176, '\P{^isdingbats}', "");
Expect(1, 10175, '\p{	-is_Dingbats}', "");
Expect(0, 10175, '\p{^	-is_Dingbats}', "");
Expect(0, 10175, '\P{	-is_Dingbats}', "");
Expect(1, 10175, '\P{^	-is_Dingbats}', "");
Expect(0, 10176, '\p{	-is_Dingbats}', "");
Expect(1, 10176, '\p{^	-is_Dingbats}', "");
Expect(1, 10176, '\P{	-is_Dingbats}', "");
Expect(0, 10176, '\P{^	-is_Dingbats}', "");
Error('\p{/a/in_dingbats}');
Error('\P{/a/in_dingbats}');
Expect(1, 10175, '\p{indingbats}', "");
Expect(0, 10175, '\p{^indingbats}', "");
Expect(0, 10175, '\P{indingbats}', "");
Expect(1, 10175, '\P{^indingbats}', "");
Expect(0, 10176, '\p{indingbats}', "");
Expect(1, 10176, '\p{^indingbats}', "");
Expect(1, 10176, '\P{indingbats}', "");
Expect(0, 10176, '\P{^indingbats}', "");
Expect(1, 10175, '\p{_In_dingbats}', "");
Expect(0, 10175, '\p{^_In_dingbats}', "");
Expect(0, 10175, '\P{_In_dingbats}', "");
Expect(1, 10175, '\P{^_In_dingbats}', "");
Expect(0, 10176, '\p{_In_dingbats}', "");
Expect(1, 10176, '\p{^_In_dingbats}', "");
Expect(1, 10176, '\P{_In_dingbats}', "");
Expect(0, 10176, '\P{^_In_dingbats}', "");
Error('\p{/a/ domino_Tiles}');
Error('\P{/a/ domino_Tiles}');
Expect(1, 127135, '\p{dominotiles}', "");
Expect(0, 127135, '\p{^dominotiles}', "");
Expect(0, 127135, '\P{dominotiles}', "");
Expect(1, 127135, '\P{^dominotiles}', "");
Expect(0, 127136, '\p{dominotiles}', "");
Expect(1, 127136, '\p{^dominotiles}', "");
Expect(1, 127136, '\P{dominotiles}', "");
Expect(0, 127136, '\P{^dominotiles}', "");
Expect(1, 127135, '\p{_ Domino_TILES}', "");
Expect(0, 127135, '\p{^_ Domino_TILES}', "");
Expect(0, 127135, '\P{_ Domino_TILES}', "");
Expect(1, 127135, '\P{^_ Domino_TILES}', "");
Expect(0, 127136, '\p{_ Domino_TILES}', "");
Expect(1, 127136, '\p{^_ Domino_TILES}', "");
Expect(1, 127136, '\P{_ Domino_TILES}', "");
Expect(0, 127136, '\P{^_ Domino_TILES}', "");
Error('\p{:=_is_Domino_Tiles}');
Error('\P{:=_is_Domino_Tiles}');
Expect(1, 127135, '\p{isdominotiles}', "");
Expect(0, 127135, '\p{^isdominotiles}', "");
Expect(0, 127135, '\P{isdominotiles}', "");
Expect(1, 127135, '\P{^isdominotiles}', "");
Expect(0, 127136, '\p{isdominotiles}', "");
Expect(1, 127136, '\p{^isdominotiles}', "");
Expect(1, 127136, '\P{isdominotiles}', "");
Expect(0, 127136, '\P{^isdominotiles}', "");
Expect(1, 127135, '\p{		Is_Domino_tiles}', "");
Expect(0, 127135, '\p{^		Is_Domino_tiles}', "");
Expect(0, 127135, '\P{		Is_Domino_tiles}', "");
Expect(1, 127135, '\P{^		Is_Domino_tiles}', "");
Expect(0, 127136, '\p{		Is_Domino_tiles}', "");
Expect(1, 127136, '\p{^		Is_Domino_tiles}', "");
Expect(1, 127136, '\P{		Is_Domino_tiles}', "");
Expect(0, 127136, '\P{^		Is_Domino_tiles}', "");
Error('\p{/a/_IN_Domino_tiles}');
Error('\P{/a/_IN_Domino_tiles}');
Expect(1, 127135, '\p{indominotiles}', "");
Expect(0, 127135, '\p{^indominotiles}', "");
Expect(0, 127135, '\P{indominotiles}', "");
Expect(1, 127135, '\P{^indominotiles}', "");
Expect(0, 127136, '\p{indominotiles}', "");
Expect(1, 127136, '\p{^indominotiles}', "");
Expect(1, 127136, '\P{indominotiles}', "");
Expect(0, 127136, '\P{^indominotiles}', "");
Expect(1, 127135, '\p{	_In_Domino_tiles}', "");
Expect(0, 127135, '\p{^	_In_Domino_tiles}', "");
Expect(0, 127135, '\P{	_In_Domino_tiles}', "");
Expect(1, 127135, '\P{^	_In_Domino_tiles}', "");
Expect(0, 127136, '\p{	_In_Domino_tiles}', "");
Expect(1, 127136, '\p{^	_In_Domino_tiles}', "");
Expect(1, 127136, '\P{	_In_Domino_tiles}', "");
Expect(0, 127136, '\P{^	_In_Domino_tiles}', "");
Error('\p{	:=Domino}');
Error('\P{	:=Domino}');
Expect(1, 127135, '\p{domino}', "");
Expect(0, 127135, '\p{^domino}', "");
Expect(0, 127135, '\P{domino}', "");
Expect(1, 127135, '\P{^domino}', "");
Expect(0, 127136, '\p{domino}', "");
Expect(1, 127136, '\p{^domino}', "");
Expect(1, 127136, '\P{domino}', "");
Expect(0, 127136, '\P{^domino}', "");
Expect(1, 127135, '\p{-domino}', "");
Expect(0, 127135, '\p{^-domino}', "");
Expect(0, 127135, '\P{-domino}', "");
Expect(1, 127135, '\P{^-domino}', "");
Expect(0, 127136, '\p{-domino}', "");
Expect(1, 127136, '\p{^-domino}', "");
Expect(1, 127136, '\P{-domino}', "");
Expect(0, 127136, '\P{^-domino}', "");
Error('\p{:=is_Domino}');
Error('\P{:=is_Domino}');
Expect(1, 127135, '\p{isdomino}', "");
Expect(0, 127135, '\p{^isdomino}', "");
Expect(0, 127135, '\P{isdomino}', "");
Expect(1, 127135, '\P{^isdomino}', "");
Expect(0, 127136, '\p{isdomino}', "");
Expect(1, 127136, '\p{^isdomino}', "");
Expect(1, 127136, '\P{isdomino}', "");
Expect(0, 127136, '\P{^isdomino}', "");
Expect(1, 127135, '\p{--Is_Domino}', "");
Expect(0, 127135, '\p{^--Is_Domino}', "");
Expect(0, 127135, '\P{--Is_Domino}', "");
Expect(1, 127135, '\P{^--Is_Domino}', "");
Expect(0, 127136, '\p{--Is_Domino}', "");
Expect(1, 127136, '\p{^--Is_Domino}', "");
Expect(1, 127136, '\P{--Is_Domino}', "");
Expect(0, 127136, '\P{^--Is_Domino}', "");
Error('\p{:=_	in_Domino}');
Error('\P{:=_	in_Domino}');
Expect(1, 127135, '\p{indomino}', "");
Expect(0, 127135, '\p{^indomino}', "");
Expect(0, 127135, '\P{indomino}', "");
Expect(1, 127135, '\P{^indomino}', "");
Expect(0, 127136, '\p{indomino}', "");
Expect(1, 127136, '\p{^indomino}', "");
Expect(1, 127136, '\P{indomino}', "");
Expect(0, 127136, '\P{^indomino}', "");
Expect(1, 127135, '\p{In_Domino}', "");
Expect(0, 127135, '\p{^In_Domino}', "");
Expect(0, 127135, '\P{In_Domino}', "");
Expect(1, 127135, '\P{^In_Domino}', "");
Expect(0, 127136, '\p{In_Domino}', "");
Expect(1, 127136, '\p{^In_Domino}', "");
Expect(1, 127136, '\P{In_Domino}', "");
Expect(0, 127136, '\P{^In_Domino}', "");
Error('\p{:=	-Egyptian_Hieroglyphs}');
Error('\P{:=	-Egyptian_Hieroglyphs}');
Expect(1, 78894, '\p{egyptianhieroglyphs}', "");
Expect(0, 78894, '\p{^egyptianhieroglyphs}', "");
Expect(0, 78894, '\P{egyptianhieroglyphs}', "");
Expect(1, 78894, '\P{^egyptianhieroglyphs}', "");
Expect(0, 78895, '\p{egyptianhieroglyphs}', "");
Expect(1, 78895, '\p{^egyptianhieroglyphs}', "");
Expect(1, 78895, '\P{egyptianhieroglyphs}', "");
Expect(0, 78895, '\P{^egyptianhieroglyphs}', "");
Expect(1, 78894, '\p{-_EGYPTIAN_Hieroglyphs}', "");
Expect(0, 78894, '\p{^-_EGYPTIAN_Hieroglyphs}', "");
Expect(0, 78894, '\P{-_EGYPTIAN_Hieroglyphs}', "");
Expect(1, 78894, '\P{^-_EGYPTIAN_Hieroglyphs}', "");
Expect(0, 78895, '\p{-_EGYPTIAN_Hieroglyphs}', "");
Expect(1, 78895, '\p{^-_EGYPTIAN_Hieroglyphs}', "");
Expect(1, 78895, '\P{-_EGYPTIAN_Hieroglyphs}', "");
Expect(0, 78895, '\P{^-_EGYPTIAN_Hieroglyphs}', "");
Error('\p{	:=Is_Egyptian_Hieroglyphs}');
Error('\P{	:=Is_Egyptian_Hieroglyphs}');
Expect(1, 78894, '\p{isegyptianhieroglyphs}', "");
Expect(0, 78894, '\p{^isegyptianhieroglyphs}', "");
Expect(0, 78894, '\P{isegyptianhieroglyphs}', "");
Expect(1, 78894, '\P{^isegyptianhieroglyphs}', "");
Expect(0, 78895, '\p{isegyptianhieroglyphs}', "");
Expect(1, 78895, '\p{^isegyptianhieroglyphs}', "");
Expect(1, 78895, '\P{isegyptianhieroglyphs}', "");
Expect(0, 78895, '\P{^isegyptianhieroglyphs}', "");
Expect(1, 78894, '\p{ is_Egyptian_hieroglyphs}', "");
Expect(0, 78894, '\p{^ is_Egyptian_hieroglyphs}', "");
Expect(0, 78894, '\P{ is_Egyptian_hieroglyphs}', "");
Expect(1, 78894, '\P{^ is_Egyptian_hieroglyphs}', "");
Expect(0, 78895, '\p{ is_Egyptian_hieroglyphs}', "");
Expect(1, 78895, '\p{^ is_Egyptian_hieroglyphs}', "");
Expect(1, 78895, '\P{ is_Egyptian_hieroglyphs}', "");
Expect(0, 78895, '\P{^ is_Egyptian_hieroglyphs}', "");
Error('\p{--Egyp:=}');
Error('\P{--Egyp:=}');
Expect(1, 78894, '\p{egyp}', "");
Expect(0, 78894, '\p{^egyp}', "");
Expect(0, 78894, '\P{egyp}', "");
Expect(1, 78894, '\P{^egyp}', "");
Expect(0, 78895, '\p{egyp}', "");
Expect(1, 78895, '\p{^egyp}', "");
Expect(1, 78895, '\P{egyp}', "");
Expect(0, 78895, '\P{^egyp}', "");
Expect(1, 78894, '\p{--EGYP}', "");
Expect(0, 78894, '\p{^--EGYP}', "");
Expect(0, 78894, '\P{--EGYP}', "");
Expect(1, 78894, '\P{^--EGYP}', "");
Expect(0, 78895, '\p{--EGYP}', "");
Expect(1, 78895, '\p{^--EGYP}', "");
Expect(1, 78895, '\P{--EGYP}', "");
Expect(0, 78895, '\P{^--EGYP}', "");
Error('\p{	/a/Is_Egyp}');
Error('\P{	/a/Is_Egyp}');
Expect(1, 78894, '\p{isegyp}', "");
Expect(0, 78894, '\p{^isegyp}', "");
Expect(0, 78894, '\P{isegyp}', "");
Expect(1, 78894, '\P{^isegyp}', "");
Expect(0, 78895, '\p{isegyp}', "");
Expect(1, 78895, '\p{^isegyp}', "");
Expect(1, 78895, '\P{isegyp}', "");
Expect(0, 78895, '\P{^isegyp}', "");
Expect(1, 78894, '\p{		IS_EGYP}', "");
Expect(0, 78894, '\p{^		IS_EGYP}', "");
Expect(0, 78894, '\P{		IS_EGYP}', "");
Expect(1, 78894, '\P{^		IS_EGYP}', "");
Expect(0, 78895, '\p{		IS_EGYP}', "");
Expect(1, 78895, '\p{^		IS_EGYP}', "");
Expect(1, 78895, '\P{		IS_EGYP}', "");
Expect(0, 78895, '\P{^		IS_EGYP}', "");
Error('\p{/a/_ Emoticons}');
Error('\P{/a/_ Emoticons}');
Expect(1, 128591, '\p{emoticons}', "");
Expect(0, 128591, '\p{^emoticons}', "");
Expect(0, 128591, '\P{emoticons}', "");
Expect(1, 128591, '\P{^emoticons}', "");
Expect(0, 128592, '\p{emoticons}', "");
Expect(1, 128592, '\p{^emoticons}', "");
Expect(1, 128592, '\P{emoticons}', "");
Expect(0, 128592, '\P{^emoticons}', "");
Expect(1, 128591, '\p{_Emoticons}', "");
Expect(0, 128591, '\p{^_Emoticons}', "");
Expect(0, 128591, '\P{_Emoticons}', "");
Expect(1, 128591, '\P{^_Emoticons}', "");
Expect(0, 128592, '\p{_Emoticons}', "");
Expect(1, 128592, '\p{^_Emoticons}', "");
Expect(1, 128592, '\P{_Emoticons}', "");
Expect(0, 128592, '\P{^_Emoticons}', "");
Error('\p{-	IS_Emoticons/a/}');
Error('\P{-	IS_Emoticons/a/}');
Expect(1, 128591, '\p{isemoticons}', "");
Expect(0, 128591, '\p{^isemoticons}', "");
Expect(0, 128591, '\P{isemoticons}', "");
Expect(1, 128591, '\P{^isemoticons}', "");
Expect(0, 128592, '\p{isemoticons}', "");
Expect(1, 128592, '\p{^isemoticons}', "");
Expect(1, 128592, '\P{isemoticons}', "");
Expect(0, 128592, '\P{^isemoticons}', "");
Expect(1, 128591, '\p{- Is_Emoticons}', "");
Expect(0, 128591, '\p{^- Is_Emoticons}', "");
Expect(0, 128591, '\P{- Is_Emoticons}', "");
Expect(1, 128591, '\P{^- Is_Emoticons}', "");
Expect(0, 128592, '\p{- Is_Emoticons}', "");
Expect(1, 128592, '\p{^- Is_Emoticons}', "");
Expect(1, 128592, '\P{- Is_Emoticons}', "");
Expect(0, 128592, '\P{^- Is_Emoticons}', "");
Error('\p{-_In_EMOTICONS:=}');
Error('\P{-_In_EMOTICONS:=}');
Expect(1, 128591, '\p{inemoticons}', "");
Expect(0, 128591, '\p{^inemoticons}', "");
Expect(0, 128591, '\P{inemoticons}', "");
Expect(1, 128591, '\P{^inemoticons}', "");
Expect(0, 128592, '\p{inemoticons}', "");
Expect(1, 128592, '\p{^inemoticons}', "");
Expect(1, 128592, '\P{inemoticons}', "");
Expect(0, 128592, '\P{^inemoticons}', "");
Expect(1, 128591, '\p{		IN_Emoticons}', "");
Expect(0, 128591, '\p{^		IN_Emoticons}', "");
Expect(0, 128591, '\P{		IN_Emoticons}', "");
Expect(1, 128591, '\P{^		IN_Emoticons}', "");
Expect(0, 128592, '\p{		IN_Emoticons}', "");
Expect(1, 128592, '\p{^		IN_Emoticons}', "");
Expect(1, 128592, '\P{		IN_Emoticons}', "");
Expect(0, 128592, '\P{^		IN_Emoticons}', "");
Error('\p{-/a/enclosed_Alphanumeric_Supplement}');
Error('\P{-/a/enclosed_Alphanumeric_Supplement}');
Expect(1, 127487, '\p{enclosedalphanumericsupplement}', "");
Expect(0, 127487, '\p{^enclosedalphanumericsupplement}', "");
Expect(0, 127487, '\P{enclosedalphanumericsupplement}', "");
Expect(1, 127487, '\P{^enclosedalphanumericsupplement}', "");
Expect(0, 127488, '\p{enclosedalphanumericsupplement}', "");
Expect(1, 127488, '\p{^enclosedalphanumericsupplement}', "");
Expect(1, 127488, '\P{enclosedalphanumericsupplement}', "");
Expect(0, 127488, '\P{^enclosedalphanumericsupplement}', "");
Expect(1, 127487, '\p{-	enclosed_alphanumeric_Supplement}', "");
Expect(0, 127487, '\p{^-	enclosed_alphanumeric_Supplement}', "");
Expect(0, 127487, '\P{-	enclosed_alphanumeric_Supplement}', "");
Expect(1, 127487, '\P{^-	enclosed_alphanumeric_Supplement}', "");
Expect(0, 127488, '\p{-	enclosed_alphanumeric_Supplement}', "");
Expect(1, 127488, '\p{^-	enclosed_alphanumeric_Supplement}', "");
Expect(1, 127488, '\P{-	enclosed_alphanumeric_Supplement}', "");
Expect(0, 127488, '\P{^-	enclosed_alphanumeric_Supplement}', "");
Error('\p{	:=IS_enclosed_ALPHANUMERIC_Supplement}');
Error('\P{	:=IS_enclosed_ALPHANUMERIC_Supplement}');
Expect(1, 127487, '\p{isenclosedalphanumericsupplement}', "");
Expect(0, 127487, '\p{^isenclosedalphanumericsupplement}', "");
Expect(0, 127487, '\P{isenclosedalphanumericsupplement}', "");
Expect(1, 127487, '\P{^isenclosedalphanumericsupplement}', "");
Expect(0, 127488, '\p{isenclosedalphanumericsupplement}', "");
Expect(1, 127488, '\p{^isenclosedalphanumericsupplement}', "");
Expect(1, 127488, '\P{isenclosedalphanumericsupplement}', "");
Expect(0, 127488, '\P{^isenclosedalphanumericsupplement}', "");
Expect(1, 127487, '\p{__is_enclosed_ALPHANUMERIC_supplement}', "");
Expect(0, 127487, '\p{^__is_enclosed_ALPHANUMERIC_supplement}', "");
Expect(0, 127487, '\P{__is_enclosed_ALPHANUMERIC_supplement}', "");
Expect(1, 127487, '\P{^__is_enclosed_ALPHANUMERIC_supplement}', "");
Expect(0, 127488, '\p{__is_enclosed_ALPHANUMERIC_supplement}', "");
Expect(1, 127488, '\p{^__is_enclosed_ALPHANUMERIC_supplement}', "");
Expect(1, 127488, '\P{__is_enclosed_ALPHANUMERIC_supplement}', "");
Expect(0, 127488, '\P{^__is_enclosed_ALPHANUMERIC_supplement}', "");
Error('\p{In_enclosed_ALPHANUMERIC_SUPPLEMENT:=}');
Error('\P{In_enclosed_ALPHANUMERIC_SUPPLEMENT:=}');
Expect(1, 127487, '\p{inenclosedalphanumericsupplement}', "");
Expect(0, 127487, '\p{^inenclosedalphanumericsupplement}', "");
Expect(0, 127487, '\P{inenclosedalphanumericsupplement}', "");
Expect(1, 127487, '\P{^inenclosedalphanumericsupplement}', "");
Expect(0, 127488, '\p{inenclosedalphanumericsupplement}', "");
Expect(1, 127488, '\p{^inenclosedalphanumericsupplement}', "");
Expect(1, 127488, '\P{inenclosedalphanumericsupplement}', "");
Expect(0, 127488, '\P{^inenclosedalphanumericsupplement}', "");
Expect(1, 127487, '\p{In_enclosed_alphanumeric_SUPPLEMENT}', "");
Expect(0, 127487, '\p{^In_enclosed_alphanumeric_SUPPLEMENT}', "");
Expect(0, 127487, '\P{In_enclosed_alphanumeric_SUPPLEMENT}', "");
Expect(1, 127487, '\P{^In_enclosed_alphanumeric_SUPPLEMENT}', "");
Expect(0, 127488, '\p{In_enclosed_alphanumeric_SUPPLEMENT}', "");
Expect(1, 127488, '\p{^In_enclosed_alphanumeric_SUPPLEMENT}', "");
Expect(1, 127488, '\P{In_enclosed_alphanumeric_SUPPLEMENT}', "");
Expect(0, 127488, '\P{^In_enclosed_alphanumeric_SUPPLEMENT}', "");
Error('\p{-ENCLOSED_Alphanum_sup:=}');
Error('\P{-ENCLOSED_Alphanum_sup:=}');
Expect(1, 127487, '\p{enclosedalphanumsup}', "");
Expect(0, 127487, '\p{^enclosedalphanumsup}', "");
Expect(0, 127487, '\P{enclosedalphanumsup}', "");
Expect(1, 127487, '\P{^enclosedalphanumsup}', "");
Expect(0, 127488, '\p{enclosedalphanumsup}', "");
Expect(1, 127488, '\p{^enclosedalphanumsup}', "");
Expect(1, 127488, '\P{enclosedalphanumsup}', "");
Expect(0, 127488, '\P{^enclosedalphanumsup}', "");
Expect(1, 127487, '\p{ Enclosed_ALPHANUM_Sup}', "");
Expect(0, 127487, '\p{^ Enclosed_ALPHANUM_Sup}', "");
Expect(0, 127487, '\P{ Enclosed_ALPHANUM_Sup}', "");
Expect(1, 127487, '\P{^ Enclosed_ALPHANUM_Sup}', "");
Expect(0, 127488, '\p{ Enclosed_ALPHANUM_Sup}', "");
Expect(1, 127488, '\p{^ Enclosed_ALPHANUM_Sup}', "");
Expect(1, 127488, '\P{ Enclosed_ALPHANUM_Sup}', "");
Expect(0, 127488, '\P{^ Enclosed_ALPHANUM_Sup}', "");
Error('\p{	-IS_ENCLOSED_alphanum_Sup/a/}');
Error('\P{	-IS_ENCLOSED_alphanum_Sup/a/}');
Expect(1, 127487, '\p{isenclosedalphanumsup}', "");
Expect(0, 127487, '\p{^isenclosedalphanumsup}', "");
Expect(0, 127487, '\P{isenclosedalphanumsup}', "");
Expect(1, 127487, '\P{^isenclosedalphanumsup}', "");
Expect(0, 127488, '\p{isenclosedalphanumsup}', "");
Expect(1, 127488, '\p{^isenclosedalphanumsup}', "");
Expect(1, 127488, '\P{isenclosedalphanumsup}', "");
Expect(0, 127488, '\P{^isenclosedalphanumsup}', "");
Expect(1, 127487, '\p{-Is_Enclosed_Alphanum_Sup}', "");
Expect(0, 127487, '\p{^-Is_Enclosed_Alphanum_Sup}', "");
Expect(0, 127487, '\P{-Is_Enclosed_Alphanum_Sup}', "");
Expect(1, 127487, '\P{^-Is_Enclosed_Alphanum_Sup}', "");
Expect(0, 127488, '\p{-Is_Enclosed_Alphanum_Sup}', "");
Expect(1, 127488, '\p{^-Is_Enclosed_Alphanum_Sup}', "");
Expect(1, 127488, '\P{-Is_Enclosed_Alphanum_Sup}', "");
Expect(0, 127488, '\P{^-Is_Enclosed_Alphanum_Sup}', "");
Error('\p{:=_ IN_enclosed_Alphanum_sup}');
Error('\P{:=_ IN_enclosed_Alphanum_sup}');
Expect(1, 127487, '\p{inenclosedalphanumsup}', "");
Expect(0, 127487, '\p{^inenclosedalphanumsup}', "");
Expect(0, 127487, '\P{inenclosedalphanumsup}', "");
Expect(1, 127487, '\P{^inenclosedalphanumsup}', "");
Expect(0, 127488, '\p{inenclosedalphanumsup}', "");
Expect(1, 127488, '\p{^inenclosedalphanumsup}', "");
Expect(1, 127488, '\P{inenclosedalphanumsup}', "");
Expect(0, 127488, '\P{^inenclosedalphanumsup}', "");
Expect(1, 127487, '\p{	_in_ENCLOSED_alphanum_SUP}', "");
Expect(0, 127487, '\p{^	_in_ENCLOSED_alphanum_SUP}', "");
Expect(0, 127487, '\P{	_in_ENCLOSED_alphanum_SUP}', "");
Expect(1, 127487, '\P{^	_in_ENCLOSED_alphanum_SUP}', "");
Expect(0, 127488, '\p{	_in_ENCLOSED_alphanum_SUP}', "");
Expect(1, 127488, '\p{^	_in_ENCLOSED_alphanum_SUP}', "");
Expect(1, 127488, '\P{	_in_ENCLOSED_alphanum_SUP}', "");
Expect(0, 127488, '\P{^	_in_ENCLOSED_alphanum_SUP}', "");
Error('\p{	:=ENCLOSED_Alphanumerics}');
Error('\P{	:=ENCLOSED_Alphanumerics}');
Expect(1, 9471, '\p{enclosedalphanumerics}', "");
Expect(0, 9471, '\p{^enclosedalphanumerics}', "");
Expect(0, 9471, '\P{enclosedalphanumerics}', "");
Expect(1, 9471, '\P{^enclosedalphanumerics}', "");
Expect(0, 9472, '\p{enclosedalphanumerics}', "");
Expect(1, 9472, '\p{^enclosedalphanumerics}', "");
Expect(1, 9472, '\P{enclosedalphanumerics}', "");
Expect(0, 9472, '\P{^enclosedalphanumerics}', "");
Expect(1, 9471, '\p{	_enclosed_Alphanumerics}', "");
Expect(0, 9471, '\p{^	_enclosed_Alphanumerics}', "");
Expect(0, 9471, '\P{	_enclosed_Alphanumerics}', "");
Expect(1, 9471, '\P{^	_enclosed_Alphanumerics}', "");
Expect(0, 9472, '\p{	_enclosed_Alphanumerics}', "");
Expect(1, 9472, '\p{^	_enclosed_Alphanumerics}', "");
Expect(1, 9472, '\P{	_enclosed_Alphanumerics}', "");
Expect(0, 9472, '\P{^	_enclosed_Alphanumerics}', "");
Error('\p{/a/	Is_Enclosed_ALPHANUMERICS}');
Error('\P{/a/	Is_Enclosed_ALPHANUMERICS}');
Expect(1, 9471, '\p{isenclosedalphanumerics}', "");
Expect(0, 9471, '\p{^isenclosedalphanumerics}', "");
Expect(0, 9471, '\P{isenclosedalphanumerics}', "");
Expect(1, 9471, '\P{^isenclosedalphanumerics}', "");
Expect(0, 9472, '\p{isenclosedalphanumerics}', "");
Expect(1, 9472, '\p{^isenclosedalphanumerics}', "");
Expect(1, 9472, '\P{isenclosedalphanumerics}', "");
Expect(0, 9472, '\P{^isenclosedalphanumerics}', "");
Expect(1, 9471, '\p{	-Is_ENCLOSED_Alphanumerics}', "");
Expect(0, 9471, '\p{^	-Is_ENCLOSED_Alphanumerics}', "");
Expect(0, 9471, '\P{	-Is_ENCLOSED_Alphanumerics}', "");
Expect(1, 9471, '\P{^	-Is_ENCLOSED_Alphanumerics}', "");
Expect(0, 9472, '\p{	-Is_ENCLOSED_Alphanumerics}', "");
Expect(1, 9472, '\p{^	-Is_ENCLOSED_Alphanumerics}', "");
Expect(1, 9472, '\P{	-Is_ENCLOSED_Alphanumerics}', "");
Expect(0, 9472, '\P{^	-Is_ENCLOSED_Alphanumerics}', "");
Error('\p{/a/_ IN_enclosed_ALPHANUMERICS}');
Error('\P{/a/_ IN_enclosed_ALPHANUMERICS}');
Expect(1, 9471, '\p{inenclosedalphanumerics}', "");
Expect(0, 9471, '\p{^inenclosedalphanumerics}', "");
Expect(0, 9471, '\P{inenclosedalphanumerics}', "");
Expect(1, 9471, '\P{^inenclosedalphanumerics}', "");
Expect(0, 9472, '\p{inenclosedalphanumerics}', "");
Expect(1, 9472, '\p{^inenclosedalphanumerics}', "");
Expect(1, 9472, '\P{inenclosedalphanumerics}', "");
Expect(0, 9472, '\P{^inenclosedalphanumerics}', "");
Expect(1, 9471, '\p{ In_Enclosed_ALPHANUMERICS}', "");
Expect(0, 9471, '\p{^ In_Enclosed_ALPHANUMERICS}', "");
Expect(0, 9471, '\P{ In_Enclosed_ALPHANUMERICS}', "");
Expect(1, 9471, '\P{^ In_Enclosed_ALPHANUMERICS}', "");
Expect(0, 9472, '\p{ In_Enclosed_ALPHANUMERICS}', "");
Expect(1, 9472, '\p{^ In_Enclosed_ALPHANUMERICS}', "");
Expect(1, 9472, '\P{ In_Enclosed_ALPHANUMERICS}', "");
Expect(0, 9472, '\P{^ In_Enclosed_ALPHANUMERICS}', "");
Error('\p{:=  Enclosed_Alphanum}');
Error('\P{:=  Enclosed_Alphanum}');
Expect(1, 9471, '\p{enclosedalphanum}', "");
Expect(0, 9471, '\p{^enclosedalphanum}', "");
Expect(0, 9471, '\P{enclosedalphanum}', "");
Expect(1, 9471, '\P{^enclosedalphanum}', "");
Expect(0, 9472, '\p{enclosedalphanum}', "");
Expect(1, 9472, '\p{^enclosedalphanum}', "");
Expect(1, 9472, '\P{enclosedalphanum}', "");
Expect(0, 9472, '\P{^enclosedalphanum}', "");
Expect(1, 9471, '\p{ _ENCLOSED_alphanum}', "");
Expect(0, 9471, '\p{^ _ENCLOSED_alphanum}', "");
Expect(0, 9471, '\P{ _ENCLOSED_alphanum}', "");
Expect(1, 9471, '\P{^ _ENCLOSED_alphanum}', "");
Expect(0, 9472, '\p{ _ENCLOSED_alphanum}', "");
Expect(1, 9472, '\p{^ _ENCLOSED_alphanum}', "");
Expect(1, 9472, '\P{ _ENCLOSED_alphanum}', "");
Expect(0, 9472, '\P{^ _ENCLOSED_alphanum}', "");
Error('\p{/a/-	Is_ENCLOSED_Alphanum}');
Error('\P{/a/-	Is_ENCLOSED_Alphanum}');
Expect(1, 9471, '\p{isenclosedalphanum}', "");
Expect(0, 9471, '\p{^isenclosedalphanum}', "");
Expect(0, 9471, '\P{isenclosedalphanum}', "");
Expect(1, 9471, '\P{^isenclosedalphanum}', "");
Expect(0, 9472, '\p{isenclosedalphanum}', "");
Expect(1, 9472, '\p{^isenclosedalphanum}', "");
Expect(1, 9472, '\P{isenclosedalphanum}', "");
Expect(0, 9472, '\P{^isenclosedalphanum}', "");
Expect(1, 9471, '\p{IS_enclosed_alphanum}', "");
Expect(0, 9471, '\p{^IS_enclosed_alphanum}', "");
Expect(0, 9471, '\P{IS_enclosed_alphanum}', "");
Expect(1, 9471, '\P{^IS_enclosed_alphanum}', "");
Expect(0, 9472, '\p{IS_enclosed_alphanum}', "");
Expect(1, 9472, '\p{^IS_enclosed_alphanum}', "");
Expect(1, 9472, '\P{IS_enclosed_alphanum}', "");
Expect(0, 9472, '\P{^IS_enclosed_alphanum}', "");
Error('\p{/a/In_ENCLOSED_alphanum}');
Error('\P{/a/In_ENCLOSED_alphanum}');
Expect(1, 9471, '\p{inenclosedalphanum}', "");
Expect(0, 9471, '\p{^inenclosedalphanum}', "");
Expect(0, 9471, '\P{inenclosedalphanum}', "");
Expect(1, 9471, '\P{^inenclosedalphanum}', "");
Expect(0, 9472, '\p{inenclosedalphanum}', "");
Expect(1, 9472, '\p{^inenclosedalphanum}', "");
Expect(1, 9472, '\P{inenclosedalphanum}', "");
Expect(0, 9472, '\P{^inenclosedalphanum}', "");
Expect(1, 9471, '\p{_ in_enclosed_ALPHANUM}', "");
Expect(0, 9471, '\p{^_ in_enclosed_ALPHANUM}', "");
Expect(0, 9471, '\P{_ in_enclosed_ALPHANUM}', "");
Expect(1, 9471, '\P{^_ in_enclosed_ALPHANUM}', "");
Expect(0, 9472, '\p{_ in_enclosed_ALPHANUM}', "");
Expect(1, 9472, '\p{^_ in_enclosed_ALPHANUM}', "");
Expect(1, 9472, '\P{_ in_enclosed_ALPHANUM}', "");
Expect(0, 9472, '\P{^_ in_enclosed_ALPHANUM}', "");
Error('\p{ :=Enclosed_cjk_Letters_and_Months}');
Error('\P{ :=Enclosed_cjk_Letters_and_Months}');
Expect(1, 13055, '\p{enclosedcjklettersandmonths}', "");
Expect(0, 13055, '\p{^enclosedcjklettersandmonths}', "");
Expect(0, 13055, '\P{enclosedcjklettersandmonths}', "");
Expect(1, 13055, '\P{^enclosedcjklettersandmonths}', "");
Expect(0, 13056, '\p{enclosedcjklettersandmonths}', "");
Expect(1, 13056, '\p{^enclosedcjklettersandmonths}', "");
Expect(1, 13056, '\P{enclosedcjklettersandmonths}', "");
Expect(0, 13056, '\P{^enclosedcjklettersandmonths}', "");
Expect(1, 13055, '\p{_-Enclosed_CJK_LETTERS_AND_months}', "");
Expect(0, 13055, '\p{^_-Enclosed_CJK_LETTERS_AND_months}', "");
Expect(0, 13055, '\P{_-Enclosed_CJK_LETTERS_AND_months}', "");
Expect(1, 13055, '\P{^_-Enclosed_CJK_LETTERS_AND_months}', "");
Expect(0, 13056, '\p{_-Enclosed_CJK_LETTERS_AND_months}', "");
Expect(1, 13056, '\p{^_-Enclosed_CJK_LETTERS_AND_months}', "");
Expect(1, 13056, '\P{_-Enclosed_CJK_LETTERS_AND_months}', "");
Expect(0, 13056, '\P{^_-Enclosed_CJK_LETTERS_AND_months}', "");
Error('\p{ :=Is_ENCLOSED_cjk_Letters_AND_Months}');
Error('\P{ :=Is_ENCLOSED_cjk_Letters_AND_Months}');
Expect(1, 13055, '\p{isenclosedcjklettersandmonths}', "");
Expect(0, 13055, '\p{^isenclosedcjklettersandmonths}', "");
Expect(0, 13055, '\P{isenclosedcjklettersandmonths}', "");
Expect(1, 13055, '\P{^isenclosedcjklettersandmonths}', "");
Expect(0, 13056, '\p{isenclosedcjklettersandmonths}', "");
Expect(1, 13056, '\p{^isenclosedcjklettersandmonths}', "");
Expect(1, 13056, '\P{isenclosedcjklettersandmonths}', "");
Expect(0, 13056, '\P{^isenclosedcjklettersandmonths}', "");
Expect(1, 13055, '\p{	_Is_Enclosed_CJK_letters_And_months}', "");
Expect(0, 13055, '\p{^	_Is_Enclosed_CJK_letters_And_months}', "");
Expect(0, 13055, '\P{	_Is_Enclosed_CJK_letters_And_months}', "");
Expect(1, 13055, '\P{^	_Is_Enclosed_CJK_letters_And_months}', "");
Expect(0, 13056, '\p{	_Is_Enclosed_CJK_letters_And_months}', "");
Expect(1, 13056, '\p{^	_Is_Enclosed_CJK_letters_And_months}', "");
Expect(1, 13056, '\P{	_Is_Enclosed_CJK_letters_And_months}', "");
Expect(0, 13056, '\P{^	_Is_Enclosed_CJK_letters_And_months}', "");
Error('\p{__In_Enclosed_CJK_letters_And_MONTHS:=}');
Error('\P{__In_Enclosed_CJK_letters_And_MONTHS:=}');
Expect(1, 13055, '\p{inenclosedcjklettersandmonths}', "");
Expect(0, 13055, '\p{^inenclosedcjklettersandmonths}', "");
Expect(0, 13055, '\P{inenclosedcjklettersandmonths}', "");
Expect(1, 13055, '\P{^inenclosedcjklettersandmonths}', "");
Expect(0, 13056, '\p{inenclosedcjklettersandmonths}', "");
Expect(1, 13056, '\p{^inenclosedcjklettersandmonths}', "");
Expect(1, 13056, '\P{inenclosedcjklettersandmonths}', "");
Expect(0, 13056, '\P{^inenclosedcjklettersandmonths}', "");
Expect(1, 13055, '\p{-_in_ENCLOSED_CJK_LETTERS_And_months}', "");
Expect(0, 13055, '\p{^-_in_ENCLOSED_CJK_LETTERS_And_months}', "");
Expect(0, 13055, '\P{-_in_ENCLOSED_CJK_LETTERS_And_months}', "");
Expect(1, 13055, '\P{^-_in_ENCLOSED_CJK_LETTERS_And_months}', "");
Expect(0, 13056, '\p{-_in_ENCLOSED_CJK_LETTERS_And_months}', "");
Expect(1, 13056, '\p{^-_in_ENCLOSED_CJK_LETTERS_And_months}', "");
Expect(1, 13056, '\P{-_in_ENCLOSED_CJK_LETTERS_And_months}', "");
Expect(0, 13056, '\P{^-_in_ENCLOSED_CJK_LETTERS_And_months}', "");
Error('\p{	 ENCLOSED_CJK/a/}');
Error('\P{	 ENCLOSED_CJK/a/}');
Expect(1, 13055, '\p{enclosedcjk}', "");
Expect(0, 13055, '\p{^enclosedcjk}', "");
Expect(0, 13055, '\P{enclosedcjk}', "");
Expect(1, 13055, '\P{^enclosedcjk}', "");
Expect(0, 13056, '\p{enclosedcjk}', "");
Expect(1, 13056, '\p{^enclosedcjk}', "");
Expect(1, 13056, '\P{enclosedcjk}', "");
Expect(0, 13056, '\P{^enclosedcjk}', "");
Expect(1, 13055, '\p{-Enclosed_CJK}', "");
Expect(0, 13055, '\p{^-Enclosed_CJK}', "");
Expect(0, 13055, '\P{-Enclosed_CJK}', "");
Expect(1, 13055, '\P{^-Enclosed_CJK}', "");
Expect(0, 13056, '\p{-Enclosed_CJK}', "");
Expect(1, 13056, '\p{^-Enclosed_CJK}', "");
Expect(1, 13056, '\P{-Enclosed_CJK}', "");
Expect(0, 13056, '\P{^-Enclosed_CJK}', "");
Error('\p{_	is_ENCLOSED_CJK/a/}');
Error('\P{_	is_ENCLOSED_CJK/a/}');
Expect(1, 13055, '\p{isenclosedcjk}', "");
Expect(0, 13055, '\p{^isenclosedcjk}', "");
Expect(0, 13055, '\P{isenclosedcjk}', "");
Expect(1, 13055, '\P{^isenclosedcjk}', "");
Expect(0, 13056, '\p{isenclosedcjk}', "");
Expect(1, 13056, '\p{^isenclosedcjk}', "");
Expect(1, 13056, '\P{isenclosedcjk}', "");
Expect(0, 13056, '\P{^isenclosedcjk}', "");
Expect(1, 13055, '\p{_-Is_Enclosed_CJK}', "");
Expect(0, 13055, '\p{^_-Is_Enclosed_CJK}', "");
Expect(0, 13055, '\P{_-Is_Enclosed_CJK}', "");
Expect(1, 13055, '\P{^_-Is_Enclosed_CJK}', "");
Expect(0, 13056, '\p{_-Is_Enclosed_CJK}', "");
Expect(1, 13056, '\p{^_-Is_Enclosed_CJK}', "");
Expect(1, 13056, '\P{_-Is_Enclosed_CJK}', "");
Expect(0, 13056, '\P{^_-Is_Enclosed_CJK}', "");
Error('\p{ /a/IN_Enclosed_CJK}');
Error('\P{ /a/IN_Enclosed_CJK}');
Expect(1, 13055, '\p{inenclosedcjk}', "");
Expect(0, 13055, '\p{^inenclosedcjk}', "");
Expect(0, 13055, '\P{inenclosedcjk}', "");
Expect(1, 13055, '\P{^inenclosedcjk}', "");
Expect(0, 13056, '\p{inenclosedcjk}', "");
Expect(1, 13056, '\p{^inenclosedcjk}', "");
Expect(1, 13056, '\P{inenclosedcjk}', "");
Expect(0, 13056, '\P{^inenclosedcjk}', "");
Expect(1, 13055, '\p{ -In_Enclosed_cjk}', "");
Expect(0, 13055, '\p{^ -In_Enclosed_cjk}', "");
Expect(0, 13055, '\P{ -In_Enclosed_cjk}', "");
Expect(1, 13055, '\P{^ -In_Enclosed_cjk}', "");
Expect(0, 13056, '\p{ -In_Enclosed_cjk}', "");
Expect(1, 13056, '\p{^ -In_Enclosed_cjk}', "");
Expect(1, 13056, '\P{ -In_Enclosed_cjk}', "");
Expect(0, 13056, '\P{^ -In_Enclosed_cjk}', "");
Error('\p{-	Enclosed_Ideographic_supplement/a/}');
Error('\P{-	Enclosed_Ideographic_supplement/a/}');
Expect(1, 127743, '\p{enclosedideographicsupplement}', "");
Expect(0, 127743, '\p{^enclosedideographicsupplement}', "");
Expect(0, 127743, '\P{enclosedideographicsupplement}', "");
Expect(1, 127743, '\P{^enclosedideographicsupplement}', "");
Expect(0, 127744, '\p{enclosedideographicsupplement}', "");
Expect(1, 127744, '\p{^enclosedideographicsupplement}', "");
Expect(1, 127744, '\P{enclosedideographicsupplement}', "");
Expect(0, 127744, '\P{^enclosedideographicsupplement}', "");
Expect(1, 127743, '\p{	-Enclosed_IDEOGRAPHIC_Supplement}', "");
Expect(0, 127743, '\p{^	-Enclosed_IDEOGRAPHIC_Supplement}', "");
Expect(0, 127743, '\P{	-Enclosed_IDEOGRAPHIC_Supplement}', "");
Expect(1, 127743, '\P{^	-Enclosed_IDEOGRAPHIC_Supplement}', "");
Expect(0, 127744, '\p{	-Enclosed_IDEOGRAPHIC_Supplement}', "");
Expect(1, 127744, '\p{^	-Enclosed_IDEOGRAPHIC_Supplement}', "");
Expect(1, 127744, '\P{	-Enclosed_IDEOGRAPHIC_Supplement}', "");
Expect(0, 127744, '\P{^	-Enclosed_IDEOGRAPHIC_Supplement}', "");
Error('\p{	/a/Is_ENCLOSED_Ideographic_Supplement}');
Error('\P{	/a/Is_ENCLOSED_Ideographic_Supplement}');
Expect(1, 127743, '\p{isenclosedideographicsupplement}', "");
Expect(0, 127743, '\p{^isenclosedideographicsupplement}', "");
Expect(0, 127743, '\P{isenclosedideographicsupplement}', "");
Expect(1, 127743, '\P{^isenclosedideographicsupplement}', "");
Expect(0, 127744, '\p{isenclosedideographicsupplement}', "");
Expect(1, 127744, '\p{^isenclosedideographicsupplement}', "");
Expect(1, 127744, '\P{isenclosedideographicsupplement}', "");
Expect(0, 127744, '\P{^isenclosedideographicsupplement}', "");
Expect(1, 127743, '\p{--is_enclosed_IDEOGRAPHIC_SUPPLEMENT}', "");
Expect(0, 127743, '\p{^--is_enclosed_IDEOGRAPHIC_SUPPLEMENT}', "");
Expect(0, 127743, '\P{--is_enclosed_IDEOGRAPHIC_SUPPLEMENT}', "");
Expect(1, 127743, '\P{^--is_enclosed_IDEOGRAPHIC_SUPPLEMENT}', "");
Expect(0, 127744, '\p{--is_enclosed_IDEOGRAPHIC_SUPPLEMENT}', "");
Expect(1, 127744, '\p{^--is_enclosed_IDEOGRAPHIC_SUPPLEMENT}', "");
Expect(1, 127744, '\P{--is_enclosed_IDEOGRAPHIC_SUPPLEMENT}', "");
Expect(0, 127744, '\P{^--is_enclosed_IDEOGRAPHIC_SUPPLEMENT}', "");
Error('\p{/a/	 In_ENCLOSED_Ideographic_Supplement}');
Error('\P{/a/	 In_ENCLOSED_Ideographic_Supplement}');
Expect(1, 127743, '\p{inenclosedideographicsupplement}', "");
Expect(0, 127743, '\p{^inenclosedideographicsupplement}', "");
Expect(0, 127743, '\P{inenclosedideographicsupplement}', "");
Expect(1, 127743, '\P{^inenclosedideographicsupplement}', "");
Expect(0, 127744, '\p{inenclosedideographicsupplement}', "");
Expect(1, 127744, '\p{^inenclosedideographicsupplement}', "");
Expect(1, 127744, '\P{inenclosedideographicsupplement}', "");
Expect(0, 127744, '\P{^inenclosedideographicsupplement}', "");
Expect(1, 127743, '\p{-_In_Enclosed_ideographic_supplement}', "");
Expect(0, 127743, '\p{^-_In_Enclosed_ideographic_supplement}', "");
Expect(0, 127743, '\P{-_In_Enclosed_ideographic_supplement}', "");
Expect(1, 127743, '\P{^-_In_Enclosed_ideographic_supplement}', "");
Expect(0, 127744, '\p{-_In_Enclosed_ideographic_supplement}', "");
Expect(1, 127744, '\p{^-_In_Enclosed_ideographic_supplement}', "");
Expect(1, 127744, '\P{-_In_Enclosed_ideographic_supplement}', "");
Expect(0, 127744, '\P{^-_In_Enclosed_ideographic_supplement}', "");
Error('\p{:=- ENCLOSED_Ideographic_Sup}');
Error('\P{:=- ENCLOSED_Ideographic_Sup}');
Expect(1, 127743, '\p{enclosedideographicsup}', "");
Expect(0, 127743, '\p{^enclosedideographicsup}', "");
Expect(0, 127743, '\P{enclosedideographicsup}', "");
Expect(1, 127743, '\P{^enclosedideographicsup}', "");
Expect(0, 127744, '\p{enclosedideographicsup}', "");
Expect(1, 127744, '\p{^enclosedideographicsup}', "");
Expect(1, 127744, '\P{enclosedideographicsup}', "");
Expect(0, 127744, '\P{^enclosedideographicsup}', "");
Expect(1, 127743, '\p{ -Enclosed_Ideographic_SUP}', "");
Expect(0, 127743, '\p{^ -Enclosed_Ideographic_SUP}', "");
Expect(0, 127743, '\P{ -Enclosed_Ideographic_SUP}', "");
Expect(1, 127743, '\P{^ -Enclosed_Ideographic_SUP}', "");
Expect(0, 127744, '\p{ -Enclosed_Ideographic_SUP}', "");
Expect(1, 127744, '\p{^ -Enclosed_Ideographic_SUP}', "");
Expect(1, 127744, '\P{ -Enclosed_Ideographic_SUP}', "");
Expect(0, 127744, '\P{^ -Enclosed_Ideographic_SUP}', "");
Error('\p{--Is_Enclosed_IDEOGRAPHIC_sup:=}');
Error('\P{--Is_Enclosed_IDEOGRAPHIC_sup:=}');
Expect(1, 127743, '\p{isenclosedideographicsup}', "");
Expect(0, 127743, '\p{^isenclosedideographicsup}', "");
Expect(0, 127743, '\P{isenclosedideographicsup}', "");
Expect(1, 127743, '\P{^isenclosedideographicsup}', "");
Expect(0, 127744, '\p{isenclosedideographicsup}', "");
Expect(1, 127744, '\p{^isenclosedideographicsup}', "");
Expect(1, 127744, '\P{isenclosedideographicsup}', "");
Expect(0, 127744, '\P{^isenclosedideographicsup}', "");
Expect(1, 127743, '\p{Is_ENCLOSED_ideographic_Sup}', "");
Expect(0, 127743, '\p{^Is_ENCLOSED_ideographic_Sup}', "");
Expect(0, 127743, '\P{Is_ENCLOSED_ideographic_Sup}', "");
Expect(1, 127743, '\P{^Is_ENCLOSED_ideographic_Sup}', "");
Expect(0, 127744, '\p{Is_ENCLOSED_ideographic_Sup}', "");
Expect(1, 127744, '\p{^Is_ENCLOSED_ideographic_Sup}', "");
Expect(1, 127744, '\P{Is_ENCLOSED_ideographic_Sup}', "");
Expect(0, 127744, '\P{^Is_ENCLOSED_ideographic_Sup}', "");
Error('\p{/a/	_In_Enclosed_Ideographic_SUP}');
Error('\P{/a/	_In_Enclosed_Ideographic_SUP}');
Expect(1, 127743, '\p{inenclosedideographicsup}', "");
Expect(0, 127743, '\p{^inenclosedideographicsup}', "");
Expect(0, 127743, '\P{inenclosedideographicsup}', "");
Expect(1, 127743, '\P{^inenclosedideographicsup}', "");
Expect(0, 127744, '\p{inenclosedideographicsup}', "");
Expect(1, 127744, '\p{^inenclosedideographicsup}', "");
Expect(1, 127744, '\P{inenclosedideographicsup}', "");
Expect(0, 127744, '\P{^inenclosedideographicsup}', "");
Expect(1, 127743, '\p{-_In_Enclosed_IDEOGRAPHIC_Sup}', "");
Expect(0, 127743, '\p{^-_In_Enclosed_IDEOGRAPHIC_Sup}', "");
Expect(0, 127743, '\P{-_In_Enclosed_IDEOGRAPHIC_Sup}', "");
Expect(1, 127743, '\P{^-_In_Enclosed_IDEOGRAPHIC_Sup}', "");
Expect(0, 127744, '\p{-_In_Enclosed_IDEOGRAPHIC_Sup}', "");
Expect(1, 127744, '\p{^-_In_Enclosed_IDEOGRAPHIC_Sup}', "");
Expect(1, 127744, '\P{-_In_Enclosed_IDEOGRAPHIC_Sup}', "");
Expect(0, 127744, '\P{^-_In_Enclosed_IDEOGRAPHIC_Sup}', "");
Error('\p{__ENCLOSING_Mark/a/}');
Error('\P{__ENCLOSING_Mark/a/}');
Expect(1, 42610, '\p{enclosingmark}', "");
Expect(0, 42610, '\p{^enclosingmark}', "");
Expect(0, 42610, '\P{enclosingmark}', "");
Expect(1, 42610, '\P{^enclosingmark}', "");
Expect(0, 42611, '\p{enclosingmark}', "");
Expect(1, 42611, '\p{^enclosingmark}', "");
Expect(1, 42611, '\P{enclosingmark}', "");
Expect(0, 42611, '\P{^enclosingmark}', "");
Expect(1, 42610, '\p{- Enclosing_MARK}', "");
Expect(0, 42610, '\p{^- Enclosing_MARK}', "");
Expect(0, 42610, '\P{- Enclosing_MARK}', "");
Expect(1, 42610, '\P{^- Enclosing_MARK}', "");
Expect(0, 42611, '\p{- Enclosing_MARK}', "");
Expect(1, 42611, '\p{^- Enclosing_MARK}', "");
Expect(1, 42611, '\P{- Enclosing_MARK}', "");
Expect(0, 42611, '\P{^- Enclosing_MARK}', "");
Error('\p{_Is_ENCLOSING_Mark/a/}');
Error('\P{_Is_ENCLOSING_Mark/a/}');
Expect(1, 42610, '\p{isenclosingmark}', "");
Expect(0, 42610, '\p{^isenclosingmark}', "");
Expect(0, 42610, '\P{isenclosingmark}', "");
Expect(1, 42610, '\P{^isenclosingmark}', "");
Expect(0, 42611, '\p{isenclosingmark}', "");
Expect(1, 42611, '\p{^isenclosingmark}', "");
Expect(1, 42611, '\P{isenclosingmark}', "");
Expect(0, 42611, '\P{^isenclosingmark}', "");
Expect(1, 42610, '\p{ _Is_ENCLOSING_mark}', "");
Expect(0, 42610, '\p{^ _Is_ENCLOSING_mark}', "");
Expect(0, 42610, '\P{ _Is_ENCLOSING_mark}', "");
Expect(1, 42610, '\P{^ _Is_ENCLOSING_mark}', "");
Expect(0, 42611, '\p{ _Is_ENCLOSING_mark}', "");
Expect(1, 42611, '\p{^ _Is_ENCLOSING_mark}', "");
Expect(1, 42611, '\P{ _Is_ENCLOSING_mark}', "");
Expect(0, 42611, '\P{^ _Is_ENCLOSING_mark}', "");
Error('\p{:=- Me}');
Error('\P{:=- Me}');
Expect(1, 42610, '\p{me}', "");
Expect(0, 42610, '\p{^me}', "");
Expect(0, 42610, '\P{me}', "");
Expect(1, 42610, '\P{^me}', "");
Expect(0, 42611, '\p{me}', "");
Expect(1, 42611, '\p{^me}', "");
Expect(1, 42611, '\P{me}', "");
Expect(0, 42611, '\P{^me}', "");
Expect(1, 42610, '\p{-me}', "");
Expect(0, 42610, '\p{^-me}', "");
Expect(0, 42610, '\P{-me}', "");
Expect(1, 42610, '\P{^-me}', "");
Expect(0, 42611, '\p{-me}', "");
Expect(1, 42611, '\p{^-me}', "");
Expect(1, 42611, '\P{-me}', "");
Expect(0, 42611, '\P{^-me}', "");
Error('\p{__IS_me:=}');
Error('\P{__IS_me:=}');
Expect(1, 42610, '\p{isme}', "");
Expect(0, 42610, '\p{^isme}', "");
Expect(0, 42610, '\P{isme}', "");
Expect(1, 42610, '\P{^isme}', "");
Expect(0, 42611, '\p{isme}', "");
Expect(1, 42611, '\p{^isme}', "");
Expect(1, 42611, '\P{isme}', "");
Expect(0, 42611, '\P{^isme}', "");
Expect(1, 42610, '\p{-	Is_me}', "");
Expect(0, 42610, '\p{^-	Is_me}', "");
Expect(0, 42610, '\P{-	Is_me}', "");
Expect(1, 42610, '\P{^-	Is_me}', "");
Expect(0, 42611, '\p{-	Is_me}', "");
Expect(1, 42611, '\p{^-	Is_me}', "");
Expect(1, 42611, '\P{-	Is_me}', "");
Expect(0, 42611, '\P{^-	Is_me}', "");
Error('\p{-ETHIOPIC/a/}');
Error('\P{-ETHIOPIC/a/}');
Expect(1, 43822, '\p{ethiopic}', "");
Expect(0, 43822, '\p{^ethiopic}', "");
Expect(0, 43822, '\P{ethiopic}', "");
Expect(1, 43822, '\P{^ethiopic}', "");
Expect(0, 43823, '\p{ethiopic}', "");
Expect(1, 43823, '\p{^ethiopic}', "");
Expect(1, 43823, '\P{ethiopic}', "");
Expect(0, 43823, '\P{^ethiopic}', "");
Expect(1, 43822, '\p{--ethiopic}', "");
Expect(0, 43822, '\p{^--ethiopic}', "");
Expect(0, 43822, '\P{--ethiopic}', "");
Expect(1, 43822, '\P{^--ethiopic}', "");
Expect(0, 43823, '\p{--ethiopic}', "");
Expect(1, 43823, '\p{^--ethiopic}', "");
Expect(1, 43823, '\P{--ethiopic}', "");
Expect(0, 43823, '\P{^--ethiopic}', "");
Error('\p{	:=Is_ETHIOPIC}');
Error('\P{	:=Is_ETHIOPIC}');
Expect(1, 43822, '\p{isethiopic}', "");
Expect(0, 43822, '\p{^isethiopic}', "");
Expect(0, 43822, '\P{isethiopic}', "");
Expect(1, 43822, '\P{^isethiopic}', "");
Expect(0, 43823, '\p{isethiopic}', "");
Expect(1, 43823, '\p{^isethiopic}', "");
Expect(1, 43823, '\P{isethiopic}', "");
Expect(0, 43823, '\P{^isethiopic}', "");
Expect(1, 43822, '\p{_ Is_ETHIOPIC}', "");
Expect(0, 43822, '\p{^_ Is_ETHIOPIC}', "");
Expect(0, 43822, '\P{_ Is_ETHIOPIC}', "");
Expect(1, 43822, '\P{^_ Is_ETHIOPIC}', "");
Expect(0, 43823, '\p{_ Is_ETHIOPIC}', "");
Expect(1, 43823, '\p{^_ Is_ETHIOPIC}', "");
Expect(1, 43823, '\P{_ Is_ETHIOPIC}', "");
Expect(0, 43823, '\P{^_ Is_ETHIOPIC}', "");
Error('\p{/a/Ethi}');
Error('\P{/a/Ethi}');
Expect(1, 43822, '\p{ethi}', "");
Expect(0, 43822, '\p{^ethi}', "");
Expect(0, 43822, '\P{ethi}', "");
Expect(1, 43822, '\P{^ethi}', "");
Expect(0, 43823, '\p{ethi}', "");
Expect(1, 43823, '\p{^ethi}', "");
Expect(1, 43823, '\P{ethi}', "");
Expect(0, 43823, '\P{^ethi}', "");
Expect(1, 43822, '\p{ 	Ethi}', "");
Expect(0, 43822, '\p{^ 	Ethi}', "");
Expect(0, 43822, '\P{ 	Ethi}', "");
Expect(1, 43822, '\P{^ 	Ethi}', "");
Expect(0, 43823, '\p{ 	Ethi}', "");
Expect(1, 43823, '\p{^ 	Ethi}', "");
Expect(1, 43823, '\P{ 	Ethi}', "");
Expect(0, 43823, '\P{^ 	Ethi}', "");
Error('\p{  Is_ETHI:=}');
Error('\P{  Is_ETHI:=}');
Expect(1, 43822, '\p{isethi}', "");
Expect(0, 43822, '\p{^isethi}', "");
Expect(0, 43822, '\P{isethi}', "");
Expect(1, 43822, '\P{^isethi}', "");
Expect(0, 43823, '\p{isethi}', "");
Expect(1, 43823, '\p{^isethi}', "");
Expect(1, 43823, '\P{isethi}', "");
Expect(0, 43823, '\P{^isethi}', "");
Expect(1, 43822, '\p{Is_ethi}', "");
Expect(0, 43822, '\p{^Is_ethi}', "");
Expect(0, 43822, '\P{Is_ethi}', "");
Expect(1, 43822, '\P{^Is_ethi}', "");
Expect(0, 43823, '\p{Is_ethi}', "");
Expect(1, 43823, '\p{^Is_ethi}', "");
Expect(1, 43823, '\P{Is_ethi}', "");
Expect(0, 43823, '\P{^Is_ethi}', "");
Error('\p{/a/ Ethiopic_extended}');
Error('\P{/a/ Ethiopic_extended}');
Expect(1, 11743, '\p{ethiopicextended}', "");
Expect(0, 11743, '\p{^ethiopicextended}', "");
Expect(0, 11743, '\P{ethiopicextended}', "");
Expect(1, 11743, '\P{^ethiopicextended}', "");
Expect(0, 11744, '\p{ethiopicextended}', "");
Expect(1, 11744, '\p{^ethiopicextended}', "");
Expect(1, 11744, '\P{ethiopicextended}', "");
Expect(0, 11744, '\P{^ethiopicextended}', "");
Expect(1, 11743, '\p{		Ethiopic_Extended}', "");
Expect(0, 11743, '\p{^		Ethiopic_Extended}', "");
Expect(0, 11743, '\P{		Ethiopic_Extended}', "");
Expect(1, 11743, '\P{^		Ethiopic_Extended}', "");
Expect(0, 11744, '\p{		Ethiopic_Extended}', "");
Expect(1, 11744, '\p{^		Ethiopic_Extended}', "");
Expect(1, 11744, '\P{		Ethiopic_Extended}', "");
Expect(0, 11744, '\P{^		Ethiopic_Extended}', "");
Error('\p{  Is_Ethiopic_EXTENDED:=}');
Error('\P{  Is_Ethiopic_EXTENDED:=}');
Expect(1, 11743, '\p{isethiopicextended}', "");
Expect(0, 11743, '\p{^isethiopicextended}', "");
Expect(0, 11743, '\P{isethiopicextended}', "");
Expect(1, 11743, '\P{^isethiopicextended}', "");
Expect(0, 11744, '\p{isethiopicextended}', "");
Expect(1, 11744, '\p{^isethiopicextended}', "");
Expect(1, 11744, '\P{isethiopicextended}', "");
Expect(0, 11744, '\P{^isethiopicextended}', "");
Expect(1, 11743, '\p{	 Is_ethiopic_EXTENDED}', "");
Expect(0, 11743, '\p{^	 Is_ethiopic_EXTENDED}', "");
Expect(0, 11743, '\P{	 Is_ethiopic_EXTENDED}', "");
Expect(1, 11743, '\P{^	 Is_ethiopic_EXTENDED}', "");
Expect(0, 11744, '\p{	 Is_ethiopic_EXTENDED}', "");
Expect(1, 11744, '\p{^	 Is_ethiopic_EXTENDED}', "");
Expect(1, 11744, '\P{	 Is_ethiopic_EXTENDED}', "");
Expect(0, 11744, '\P{^	 Is_ethiopic_EXTENDED}', "");
Error('\p{ :=In_ethiopic_extended}');
Error('\P{ :=In_ethiopic_extended}');
Expect(1, 11743, '\p{inethiopicextended}', "");
Expect(0, 11743, '\p{^inethiopicextended}', "");
Expect(0, 11743, '\P{inethiopicextended}', "");
Expect(1, 11743, '\P{^inethiopicextended}', "");
Expect(0, 11744, '\p{inethiopicextended}', "");
Expect(1, 11744, '\p{^inethiopicextended}', "");
Expect(1, 11744, '\P{inethiopicextended}', "");
Expect(0, 11744, '\P{^inethiopicextended}', "");
Expect(1, 11743, '\p{  In_ethiopic_Extended}', "");
Expect(0, 11743, '\p{^  In_ethiopic_Extended}', "");
Expect(0, 11743, '\P{  In_ethiopic_Extended}', "");
Expect(1, 11743, '\P{^  In_ethiopic_Extended}', "");
Expect(0, 11744, '\p{  In_ethiopic_Extended}', "");
Expect(1, 11744, '\p{^  In_ethiopic_Extended}', "");
Expect(1, 11744, '\P{  In_ethiopic_Extended}', "");
Expect(0, 11744, '\P{^  In_ethiopic_Extended}', "");
Error('\p{ /a/Ethiopic_ext}');
Error('\P{ /a/Ethiopic_ext}');
Expect(1, 11743, '\p{ethiopicext}', "");
Expect(0, 11743, '\p{^ethiopicext}', "");
Expect(0, 11743, '\P{ethiopicext}', "");
Expect(1, 11743, '\P{^ethiopicext}', "");
Expect(0, 11744, '\p{ethiopicext}', "");
Expect(1, 11744, '\p{^ethiopicext}', "");
Expect(1, 11744, '\P{ethiopicext}', "");
Expect(0, 11744, '\P{^ethiopicext}', "");
Expect(1, 11743, '\p{--Ethiopic_Ext}', "");
Expect(0, 11743, '\p{^--Ethiopic_Ext}', "");
Expect(0, 11743, '\P{--Ethiopic_Ext}', "");
Expect(1, 11743, '\P{^--Ethiopic_Ext}', "");
Expect(0, 11744, '\p{--Ethiopic_Ext}', "");
Expect(1, 11744, '\p{^--Ethiopic_Ext}', "");
Expect(1, 11744, '\P{--Ethiopic_Ext}', "");
Expect(0, 11744, '\P{^--Ethiopic_Ext}', "");
Error('\p{/a/_is_Ethiopic_Ext}');
Error('\P{/a/_is_Ethiopic_Ext}');
Expect(1, 11743, '\p{isethiopicext}', "");
Expect(0, 11743, '\p{^isethiopicext}', "");
Expect(0, 11743, '\P{isethiopicext}', "");
Expect(1, 11743, '\P{^isethiopicext}', "");
Expect(0, 11744, '\p{isethiopicext}', "");
Expect(1, 11744, '\p{^isethiopicext}', "");
Expect(1, 11744, '\P{isethiopicext}', "");
Expect(0, 11744, '\P{^isethiopicext}', "");
Expect(1, 11743, '\p{ 	Is_Ethiopic_Ext}', "");
Expect(0, 11743, '\p{^ 	Is_Ethiopic_Ext}', "");
Expect(0, 11743, '\P{ 	Is_Ethiopic_Ext}', "");
Expect(1, 11743, '\P{^ 	Is_Ethiopic_Ext}', "");
Expect(0, 11744, '\p{ 	Is_Ethiopic_Ext}', "");
Expect(1, 11744, '\p{^ 	Is_Ethiopic_Ext}', "");
Expect(1, 11744, '\P{ 	Is_Ethiopic_Ext}', "");
Expect(0, 11744, '\P{^ 	Is_Ethiopic_Ext}', "");
Error('\p{ :=In_ethiopic_ext}');
Error('\P{ :=In_ethiopic_ext}');
Expect(1, 11743, '\p{inethiopicext}', "");
Expect(0, 11743, '\p{^inethiopicext}', "");
Expect(0, 11743, '\P{inethiopicext}', "");
Expect(1, 11743, '\P{^inethiopicext}', "");
Expect(0, 11744, '\p{inethiopicext}', "");
Expect(1, 11744, '\p{^inethiopicext}', "");
Expect(1, 11744, '\P{inethiopicext}', "");
Expect(0, 11744, '\P{^inethiopicext}', "");
Expect(1, 11743, '\p{ 	in_Ethiopic_ext}', "");
Expect(0, 11743, '\p{^ 	in_Ethiopic_ext}', "");
Expect(0, 11743, '\P{ 	in_Ethiopic_ext}', "");
Expect(1, 11743, '\P{^ 	in_Ethiopic_ext}', "");
Expect(0, 11744, '\p{ 	in_Ethiopic_ext}', "");
Expect(1, 11744, '\p{^ 	in_Ethiopic_ext}', "");
Expect(1, 11744, '\P{ 	in_Ethiopic_ext}', "");
Expect(0, 11744, '\P{^ 	in_Ethiopic_ext}', "");
Error('\p{	:=ETHIOPIC_EXTENDED_A}');
Error('\P{	:=ETHIOPIC_EXTENDED_A}');
Expect(1, 43823, '\p{ethiopicextendeda}', "");
Expect(0, 43823, '\p{^ethiopicextendeda}', "");
Expect(0, 43823, '\P{ethiopicextendeda}', "");
Expect(1, 43823, '\P{^ethiopicextendeda}', "");
Expect(0, 43824, '\p{ethiopicextendeda}', "");
Expect(1, 43824, '\p{^ethiopicextendeda}', "");
Expect(1, 43824, '\P{ethiopicextendeda}', "");
Expect(0, 43824, '\P{^ethiopicextendeda}', "");
Expect(1, 43823, '\p{_Ethiopic_Extended_a}', "");
Expect(0, 43823, '\p{^_Ethiopic_Extended_a}', "");
Expect(0, 43823, '\P{_Ethiopic_Extended_a}', "");
Expect(1, 43823, '\P{^_Ethiopic_Extended_a}', "");
Expect(0, 43824, '\p{_Ethiopic_Extended_a}', "");
Expect(1, 43824, '\p{^_Ethiopic_Extended_a}', "");
Expect(1, 43824, '\P{_Ethiopic_Extended_a}', "");
Expect(0, 43824, '\P{^_Ethiopic_Extended_a}', "");
Error('\p{-IS_Ethiopic_Extended_A:=}');
Error('\P{-IS_Ethiopic_Extended_A:=}');
Expect(1, 43823, '\p{isethiopicextendeda}', "");
Expect(0, 43823, '\p{^isethiopicextendeda}', "");
Expect(0, 43823, '\P{isethiopicextendeda}', "");
Expect(1, 43823, '\P{^isethiopicextendeda}', "");
Expect(0, 43824, '\p{isethiopicextendeda}', "");
Expect(1, 43824, '\p{^isethiopicextendeda}', "");
Expect(1, 43824, '\P{isethiopicextendeda}', "");
Expect(0, 43824, '\P{^isethiopicextendeda}', "");
Expect(1, 43823, '\p{ 	Is_ethiopic_Extended_A}', "");
Expect(0, 43823, '\p{^ 	Is_ethiopic_Extended_A}', "");
Expect(0, 43823, '\P{ 	Is_ethiopic_Extended_A}', "");
Expect(1, 43823, '\P{^ 	Is_ethiopic_Extended_A}', "");
Expect(0, 43824, '\p{ 	Is_ethiopic_Extended_A}', "");
Expect(1, 43824, '\p{^ 	Is_ethiopic_Extended_A}', "");
Expect(1, 43824, '\P{ 	Is_ethiopic_Extended_A}', "");
Expect(0, 43824, '\P{^ 	Is_ethiopic_Extended_A}', "");
Error('\p{ /a/in_Ethiopic_extended_A}');
Error('\P{ /a/in_Ethiopic_extended_A}');
Expect(1, 43823, '\p{inethiopicextendeda}', "");
Expect(0, 43823, '\p{^inethiopicextendeda}', "");
Expect(0, 43823, '\P{inethiopicextendeda}', "");
Expect(1, 43823, '\P{^inethiopicextendeda}', "");
Expect(0, 43824, '\p{inethiopicextendeda}', "");
Expect(1, 43824, '\p{^inethiopicextendeda}', "");
Expect(1, 43824, '\P{inethiopicextendeda}', "");
Expect(0, 43824, '\P{^inethiopicextendeda}', "");
Expect(1, 43823, '\p{ -In_Ethiopic_extended_A}', "");
Expect(0, 43823, '\p{^ -In_Ethiopic_extended_A}', "");
Expect(0, 43823, '\P{ -In_Ethiopic_extended_A}', "");
Expect(1, 43823, '\P{^ -In_Ethiopic_extended_A}', "");
Expect(0, 43824, '\p{ -In_Ethiopic_extended_A}', "");
Expect(1, 43824, '\p{^ -In_Ethiopic_extended_A}', "");
Expect(1, 43824, '\P{ -In_Ethiopic_extended_A}', "");
Expect(0, 43824, '\P{^ -In_Ethiopic_extended_A}', "");
Error('\p{/a/		Ethiopic_ext_A}');
Error('\P{/a/		Ethiopic_ext_A}');
Expect(1, 43823, '\p{ethiopicexta}', "");
Expect(0, 43823, '\p{^ethiopicexta}', "");
Expect(0, 43823, '\P{ethiopicexta}', "");
Expect(1, 43823, '\P{^ethiopicexta}', "");
Expect(0, 43824, '\p{ethiopicexta}', "");
Expect(1, 43824, '\p{^ethiopicexta}', "");
Expect(1, 43824, '\P{ethiopicexta}', "");
Expect(0, 43824, '\P{^ethiopicexta}', "");
Expect(1, 43823, '\p{-_ETHIOPIC_EXT_A}', "");
Expect(0, 43823, '\p{^-_ETHIOPIC_EXT_A}', "");
Expect(0, 43823, '\P{-_ETHIOPIC_EXT_A}', "");
Expect(1, 43823, '\P{^-_ETHIOPIC_EXT_A}', "");
Expect(0, 43824, '\p{-_ETHIOPIC_EXT_A}', "");
Expect(1, 43824, '\p{^-_ETHIOPIC_EXT_A}', "");
Expect(1, 43824, '\P{-_ETHIOPIC_EXT_A}', "");
Expect(0, 43824, '\P{^-_ETHIOPIC_EXT_A}', "");
Error('\p{_/a/Is_ethiopic_EXT_A}');
Error('\P{_/a/Is_ethiopic_EXT_A}');
Expect(1, 43823, '\p{isethiopicexta}', "");
Expect(0, 43823, '\p{^isethiopicexta}', "");
Expect(0, 43823, '\P{isethiopicexta}', "");
Expect(1, 43823, '\P{^isethiopicexta}', "");
Expect(0, 43824, '\p{isethiopicexta}', "");
Expect(1, 43824, '\p{^isethiopicexta}', "");
Expect(1, 43824, '\P{isethiopicexta}', "");
Expect(0, 43824, '\P{^isethiopicexta}', "");
Expect(1, 43823, '\p{	-Is_Ethiopic_Ext_A}', "");
Expect(0, 43823, '\p{^	-Is_Ethiopic_Ext_A}', "");
Expect(0, 43823, '\P{	-Is_Ethiopic_Ext_A}', "");
Expect(1, 43823, '\P{^	-Is_Ethiopic_Ext_A}', "");
Expect(0, 43824, '\p{	-Is_Ethiopic_Ext_A}', "");
Expect(1, 43824, '\p{^	-Is_Ethiopic_Ext_A}', "");
Expect(1, 43824, '\P{	-Is_Ethiopic_Ext_A}', "");
Expect(0, 43824, '\P{^	-Is_Ethiopic_Ext_A}', "");
Error('\p{ -In_Ethiopic_Ext_a:=}');
Error('\P{ -In_Ethiopic_Ext_a:=}');
Expect(1, 43823, '\p{inethiopicexta}', "");
Expect(0, 43823, '\p{^inethiopicexta}', "");
Expect(0, 43823, '\P{inethiopicexta}', "");
Expect(1, 43823, '\P{^inethiopicexta}', "");
Expect(0, 43824, '\p{inethiopicexta}', "");
Expect(1, 43824, '\p{^inethiopicexta}', "");
Expect(1, 43824, '\P{inethiopicexta}', "");
Expect(0, 43824, '\P{^inethiopicexta}', "");
Expect(1, 43823, '\p{	 In_Ethiopic_Ext_A}', "");
Expect(0, 43823, '\p{^	 In_Ethiopic_Ext_A}', "");
Expect(0, 43823, '\P{	 In_Ethiopic_Ext_A}', "");
Expect(1, 43823, '\P{^	 In_Ethiopic_Ext_A}', "");
Expect(0, 43824, '\p{	 In_Ethiopic_Ext_A}', "");
Expect(1, 43824, '\p{^	 In_Ethiopic_Ext_A}', "");
Expect(1, 43824, '\P{	 In_Ethiopic_Ext_A}', "");
Expect(0, 43824, '\P{^	 In_Ethiopic_Ext_A}', "");
Error('\p{	ethiopic_Supplement/a/}');
Error('\P{	ethiopic_Supplement/a/}');
Expect(1, 5023, '\p{ethiopicsupplement}', "");
Expect(0, 5023, '\p{^ethiopicsupplement}', "");
Expect(0, 5023, '\P{ethiopicsupplement}', "");
Expect(1, 5023, '\P{^ethiopicsupplement}', "");
Expect(0, 5024, '\p{ethiopicsupplement}', "");
Expect(1, 5024, '\p{^ethiopicsupplement}', "");
Expect(1, 5024, '\P{ethiopicsupplement}', "");
Expect(0, 5024, '\P{^ethiopicsupplement}', "");
Expect(1, 5023, '\p{__Ethiopic_SUPPLEMENT}', "");
Expect(0, 5023, '\p{^__Ethiopic_SUPPLEMENT}', "");
Expect(0, 5023, '\P{__Ethiopic_SUPPLEMENT}', "");
Expect(1, 5023, '\P{^__Ethiopic_SUPPLEMENT}', "");
Expect(0, 5024, '\p{__Ethiopic_SUPPLEMENT}', "");
Expect(1, 5024, '\p{^__Ethiopic_SUPPLEMENT}', "");
Expect(1, 5024, '\P{__Ethiopic_SUPPLEMENT}', "");
Expect(0, 5024, '\P{^__Ethiopic_SUPPLEMENT}', "");
Error('\p{:=- is_Ethiopic_SUPPLEMENT}');
Error('\P{:=- is_Ethiopic_SUPPLEMENT}');
Expect(1, 5023, '\p{isethiopicsupplement}', "");
Expect(0, 5023, '\p{^isethiopicsupplement}', "");
Expect(0, 5023, '\P{isethiopicsupplement}', "");
Expect(1, 5023, '\P{^isethiopicsupplement}', "");
Expect(0, 5024, '\p{isethiopicsupplement}', "");
Expect(1, 5024, '\p{^isethiopicsupplement}', "");
Expect(1, 5024, '\P{isethiopicsupplement}', "");
Expect(0, 5024, '\P{^isethiopicsupplement}', "");
Expect(1, 5023, '\p{ -Is_ethiopic_Supplement}', "");
Expect(0, 5023, '\p{^ -Is_ethiopic_Supplement}', "");
Expect(0, 5023, '\P{ -Is_ethiopic_Supplement}', "");
Expect(1, 5023, '\P{^ -Is_ethiopic_Supplement}', "");
Expect(0, 5024, '\p{ -Is_ethiopic_Supplement}', "");
Expect(1, 5024, '\p{^ -Is_ethiopic_Supplement}', "");
Expect(1, 5024, '\P{ -Is_ethiopic_Supplement}', "");
Expect(0, 5024, '\P{^ -Is_ethiopic_Supplement}', "");
Error('\p{__IN_Ethiopic_SUPPLEMENT:=}');
Error('\P{__IN_Ethiopic_SUPPLEMENT:=}');
Expect(1, 5023, '\p{inethiopicsupplement}', "");
Expect(0, 5023, '\p{^inethiopicsupplement}', "");
Expect(0, 5023, '\P{inethiopicsupplement}', "");
Expect(1, 5023, '\P{^inethiopicsupplement}', "");
Expect(0, 5024, '\p{inethiopicsupplement}', "");
Expect(1, 5024, '\p{^inethiopicsupplement}', "");
Expect(1, 5024, '\P{inethiopicsupplement}', "");
Expect(0, 5024, '\P{^inethiopicsupplement}', "");
Expect(1, 5023, '\p{- IN_ethiopic_Supplement}', "");
Expect(0, 5023, '\p{^- IN_ethiopic_Supplement}', "");
Expect(0, 5023, '\P{- IN_ethiopic_Supplement}', "");
Expect(1, 5023, '\P{^- IN_ethiopic_Supplement}', "");
Expect(0, 5024, '\p{- IN_ethiopic_Supplement}', "");
Expect(1, 5024, '\p{^- IN_ethiopic_Supplement}', "");
Expect(1, 5024, '\P{- IN_ethiopic_Supplement}', "");
Expect(0, 5024, '\P{^- IN_ethiopic_Supplement}', "");
Error('\p{:=-	ethiopic_Sup}');
Error('\P{:=-	ethiopic_Sup}');
Expect(1, 5023, '\p{ethiopicsup}', "");
Expect(0, 5023, '\p{^ethiopicsup}', "");
Expect(0, 5023, '\P{ethiopicsup}', "");
Expect(1, 5023, '\P{^ethiopicsup}', "");
Expect(0, 5024, '\p{ethiopicsup}', "");
Expect(1, 5024, '\p{^ethiopicsup}', "");
Expect(1, 5024, '\P{ethiopicsup}', "");
Expect(0, 5024, '\P{^ethiopicsup}', "");
Expect(1, 5023, '\p{	 Ethiopic_Sup}', "");
Expect(0, 5023, '\p{^	 Ethiopic_Sup}', "");
Expect(0, 5023, '\P{	 Ethiopic_Sup}', "");
Expect(1, 5023, '\P{^	 Ethiopic_Sup}', "");
Expect(0, 5024, '\p{	 Ethiopic_Sup}', "");
Expect(1, 5024, '\p{^	 Ethiopic_Sup}', "");
Expect(1, 5024, '\P{	 Ethiopic_Sup}', "");
Expect(0, 5024, '\P{^	 Ethiopic_Sup}', "");
Error('\p{__is_ETHIOPIC_Sup/a/}');
Error('\P{__is_ETHIOPIC_Sup/a/}');
Expect(1, 5023, '\p{isethiopicsup}', "");
Expect(0, 5023, '\p{^isethiopicsup}', "");
Expect(0, 5023, '\P{isethiopicsup}', "");
Expect(1, 5023, '\P{^isethiopicsup}', "");
Expect(0, 5024, '\p{isethiopicsup}', "");
Expect(1, 5024, '\p{^isethiopicsup}', "");
Expect(1, 5024, '\P{isethiopicsup}', "");
Expect(0, 5024, '\P{^isethiopicsup}', "");
Expect(1, 5023, '\p{_Is_ETHIOPIC_SUP}', "");
Expect(0, 5023, '\p{^_Is_ETHIOPIC_SUP}', "");
Expect(0, 5023, '\P{_Is_ETHIOPIC_SUP}', "");
Expect(1, 5023, '\P{^_Is_ETHIOPIC_SUP}', "");
Expect(0, 5024, '\p{_Is_ETHIOPIC_SUP}', "");
Expect(1, 5024, '\p{^_Is_ETHIOPIC_SUP}', "");
Expect(1, 5024, '\P{_Is_ETHIOPIC_SUP}', "");
Expect(0, 5024, '\P{^_Is_ETHIOPIC_SUP}', "");
Error('\p{-/a/In_Ethiopic_SUP}');
Error('\P{-/a/In_Ethiopic_SUP}');
Expect(1, 5023, '\p{inethiopicsup}', "");
Expect(0, 5023, '\p{^inethiopicsup}', "");
Expect(0, 5023, '\P{inethiopicsup}', "");
Expect(1, 5023, '\P{^inethiopicsup}', "");
Expect(0, 5024, '\p{inethiopicsup}', "");
Expect(1, 5024, '\p{^inethiopicsup}', "");
Expect(1, 5024, '\P{inethiopicsup}', "");
Expect(0, 5024, '\P{^inethiopicsup}', "");
Expect(1, 5023, '\p{	 In_Ethiopic_sup}', "");
Expect(0, 5023, '\p{^	 In_Ethiopic_sup}', "");
Expect(0, 5023, '\P{	 In_Ethiopic_sup}', "");
Expect(1, 5023, '\P{^	 In_Ethiopic_sup}', "");
Expect(0, 5024, '\p{	 In_Ethiopic_sup}', "");
Expect(1, 5024, '\p{^	 In_Ethiopic_sup}', "");
Expect(1, 5024, '\P{	 In_Ethiopic_sup}', "");
Expect(0, 5024, '\P{^	 In_Ethiopic_sup}', "");
Error('\p{Expands_On_NFC}');
Error('\P{Expands_On_NFC}');
Error('\p{Is_Expands_On_NFC}');
Error('\P{Is_Expands_On_NFC}');
Error('\p{XO_NFC}');
Error('\P{XO_NFC}');
Error('\p{Is_XO_NFC}');
Error('\P{Is_XO_NFC}');
Error('\p{Expands_On_NFD}');
Error('\P{Expands_On_NFD}');
Error('\p{Is_Expands_On_NFD}');
Error('\P{Is_Expands_On_NFD}');
Error('\p{XO_NFD}');
Error('\P{XO_NFD}');
Error('\p{Is_XO_NFD}');
Error('\P{Is_XO_NFD}');
Error('\p{Expands_On_NFKC}');
Error('\P{Expands_On_NFKC}');
Error('\p{Is_Expands_On_NFKC}');
Error('\P{Is_Expands_On_NFKC}');
Error('\p{XO_NFKC}');
Error('\P{XO_NFKC}');
Error('\p{Is_XO_NFKC}');
Error('\P{Is_XO_NFKC}');
Error('\p{Expands_On_NFKD}');
Error('\P{Expands_On_NFKD}');
Error('\p{Is_Expands_On_NFKD}');
Error('\P{Is_Expands_On_NFKD}');
Error('\p{XO_NFKD}');
Error('\P{XO_NFKD}');
Error('\p{Is_XO_NFKD}');
Error('\P{Is_XO_NFKD}');
Error('\p{- EXTENDER/a/}');
Error('\P{- EXTENDER/a/}');
Expect(1, 65392, '\p{extender}', "");
Expect(0, 65392, '\p{^extender}', "");
Expect(0, 65392, '\P{extender}', "");
Expect(1, 65392, '\P{^extender}', "");
Expect(0, 65393, '\p{extender}', "");
Expect(1, 65393, '\p{^extender}', "");
Expect(1, 65393, '\P{extender}', "");
Expect(0, 65393, '\P{^extender}', "");
Expect(1, 65392, '\p{ EXTENDER}', "");
Expect(0, 65392, '\p{^ EXTENDER}', "");
Expect(0, 65392, '\P{ EXTENDER}', "");
Expect(1, 65392, '\P{^ EXTENDER}', "");
Expect(0, 65393, '\p{ EXTENDER}', "");
Expect(1, 65393, '\p{^ EXTENDER}', "");
Expect(1, 65393, '\P{ EXTENDER}', "");
Expect(0, 65393, '\P{^ EXTENDER}', "");
Error('\p{/a/Is_EXTENDER}');
Error('\P{/a/Is_EXTENDER}');
Expect(1, 65392, '\p{isextender}', "");
Expect(0, 65392, '\p{^isextender}', "");
Expect(0, 65392, '\P{isextender}', "");
Expect(1, 65392, '\P{^isextender}', "");
Expect(0, 65393, '\p{isextender}', "");
Expect(1, 65393, '\p{^isextender}', "");
Expect(1, 65393, '\P{isextender}', "");
Expect(0, 65393, '\P{^isextender}', "");
Expect(1, 65392, '\p{ -Is_extender}', "");
Expect(0, 65392, '\p{^ -Is_extender}', "");
Expect(0, 65392, '\P{ -Is_extender}', "");
Expect(1, 65392, '\P{^ -Is_extender}', "");
Expect(0, 65393, '\p{ -Is_extender}', "");
Expect(1, 65393, '\p{^ -Is_extender}', "");
Expect(1, 65393, '\P{ -Is_extender}', "");
Expect(0, 65393, '\P{^ -Is_extender}', "");
Error('\p{ 	Ext:=}');
Error('\P{ 	Ext:=}');
Expect(1, 65392, '\p{ext}', "");
Expect(0, 65392, '\p{^ext}', "");
Expect(0, 65392, '\P{ext}', "");
Expect(1, 65392, '\P{^ext}', "");
Expect(0, 65393, '\p{ext}', "");
Expect(1, 65393, '\p{^ext}', "");
Expect(1, 65393, '\P{ext}', "");
Expect(0, 65393, '\P{^ext}', "");
Expect(1, 65392, '\p{		Ext}', "");
Expect(0, 65392, '\p{^		Ext}', "");
Expect(0, 65392, '\P{		Ext}', "");
Expect(1, 65392, '\P{^		Ext}', "");
Expect(0, 65393, '\p{		Ext}', "");
Expect(1, 65393, '\p{^		Ext}', "");
Expect(1, 65393, '\P{		Ext}', "");
Expect(0, 65393, '\P{^		Ext}', "");
Error('\p{ Is_Ext/a/}');
Error('\P{ Is_Ext/a/}');
Expect(1, 65392, '\p{isext}', "");
Expect(0, 65392, '\p{^isext}', "");
Expect(0, 65392, '\P{isext}', "");
Expect(1, 65392, '\P{^isext}', "");
Expect(0, 65393, '\p{isext}', "");
Expect(1, 65393, '\p{^isext}', "");
Expect(1, 65393, '\P{isext}', "");
Expect(0, 65393, '\P{^isext}', "");
Expect(1, 65392, '\p{  Is_Ext}', "");
Expect(0, 65392, '\p{^  Is_Ext}', "");
Expect(0, 65392, '\P{  Is_Ext}', "");
Expect(1, 65392, '\P{^  Is_Ext}', "");
Expect(0, 65393, '\p{  Is_Ext}', "");
Expect(1, 65393, '\p{^  Is_Ext}', "");
Expect(1, 65393, '\P{  Is_Ext}', "");
Expect(0, 65393, '\P{^  Is_Ext}', "");
Error('\p{:=_	final_PUNCTUATION}');
Error('\P{:=_	final_PUNCTUATION}');
Expect(1, 11809, '\p{finalpunctuation}', "");
Expect(0, 11809, '\p{^finalpunctuation}', "");
Expect(0, 11809, '\P{finalpunctuation}', "");
Expect(1, 11809, '\P{^finalpunctuation}', "");
Expect(0, 11810, '\p{finalpunctuation}', "");
Expect(1, 11810, '\p{^finalpunctuation}', "");
Expect(1, 11810, '\P{finalpunctuation}', "");
Expect(0, 11810, '\P{^finalpunctuation}', "");
Expect(1, 11809, '\p{	-final_punctuation}', "");
Expect(0, 11809, '\p{^	-final_punctuation}', "");
Expect(0, 11809, '\P{	-final_punctuation}', "");
Expect(1, 11809, '\P{^	-final_punctuation}', "");
Expect(0, 11810, '\p{	-final_punctuation}', "");
Expect(1, 11810, '\p{^	-final_punctuation}', "");
Expect(1, 11810, '\P{	-final_punctuation}', "");
Expect(0, 11810, '\P{^	-final_punctuation}', "");
Error('\p{	 Is_Final_Punctuation/a/}');
Error('\P{	 Is_Final_Punctuation/a/}');
Expect(1, 11809, '\p{isfinalpunctuation}', "");
Expect(0, 11809, '\p{^isfinalpunctuation}', "");
Expect(0, 11809, '\P{isfinalpunctuation}', "");
Expect(1, 11809, '\P{^isfinalpunctuation}', "");
Expect(0, 11810, '\p{isfinalpunctuation}', "");
Expect(1, 11810, '\p{^isfinalpunctuation}', "");
Expect(1, 11810, '\P{isfinalpunctuation}', "");
Expect(0, 11810, '\P{^isfinalpunctuation}', "");
Expect(1, 11809, '\p{	_is_Final_PUNCTUATION}', "");
Expect(0, 11809, '\p{^	_is_Final_PUNCTUATION}', "");
Expect(0, 11809, '\P{	_is_Final_PUNCTUATION}', "");
Expect(1, 11809, '\P{^	_is_Final_PUNCTUATION}', "");
Expect(0, 11810, '\p{	_is_Final_PUNCTUATION}', "");
Expect(1, 11810, '\p{^	_is_Final_PUNCTUATION}', "");
Expect(1, 11810, '\P{	_is_Final_PUNCTUATION}', "");
Expect(0, 11810, '\P{^	_is_Final_PUNCTUATION}', "");
Error('\p{:=_	PF}');
Error('\P{:=_	PF}');
Expect(1, 11809, '\p{pf}', "");
Expect(0, 11809, '\p{^pf}', "");
Expect(0, 11809, '\P{pf}', "");
Expect(1, 11809, '\P{^pf}', "");
Expect(0, 11810, '\p{pf}', "");
Expect(1, 11810, '\p{^pf}', "");
Expect(1, 11810, '\P{pf}', "");
Expect(0, 11810, '\P{^pf}', "");
Expect(1, 11809, '\p{ -PF}', "");
Expect(0, 11809, '\p{^ -PF}', "");
Expect(0, 11809, '\P{ -PF}', "");
Expect(1, 11809, '\P{^ -PF}', "");
Expect(0, 11810, '\p{ -PF}', "");
Expect(1, 11810, '\p{^ -PF}', "");
Expect(1, 11810, '\P{ -PF}', "");
Expect(0, 11810, '\P{^ -PF}', "");
Error('\p{:=Is_Pf}');
Error('\P{:=Is_Pf}');
Expect(1, 11809, '\p{ispf}', "");
Expect(0, 11809, '\p{^ispf}', "");
Expect(0, 11809, '\P{ispf}', "");
Expect(1, 11809, '\P{^ispf}', "");
Expect(0, 11810, '\p{ispf}', "");
Expect(1, 11810, '\p{^ispf}', "");
Expect(1, 11810, '\P{ispf}', "");
Expect(0, 11810, '\P{^ispf}', "");
Expect(1, 11809, '\p{_IS_Pf}', "");
Expect(0, 11809, '\p{^_IS_Pf}', "");
Expect(0, 11809, '\P{_IS_Pf}', "");
Expect(1, 11809, '\P{^_IS_Pf}', "");
Expect(0, 11810, '\p{_IS_Pf}', "");
Expect(1, 11810, '\p{^_IS_Pf}', "");
Expect(1, 11810, '\P{_IS_Pf}', "");
Expect(0, 11810, '\P{^_IS_Pf}', "");
Error('\p{:=		Format}');
Error('\P{:=		Format}');
Expect(1, 917631, '\p{format}', "");
Expect(0, 917631, '\p{^format}', "");
Expect(0, 917631, '\P{format}', "");
Expect(1, 917631, '\P{^format}', "");
Expect(0, 917632, '\p{format}', "");
Expect(1, 917632, '\p{^format}', "");
Expect(1, 917632, '\P{format}', "");
Expect(0, 917632, '\P{^format}', "");
Expect(1, 917631, '\p{-Format}', "");
Expect(0, 917631, '\p{^-Format}', "");
Expect(0, 917631, '\P{-Format}', "");
Expect(1, 917631, '\P{^-Format}', "");
Expect(0, 917632, '\p{-Format}', "");
Expect(1, 917632, '\p{^-Format}', "");
Expect(1, 917632, '\P{-Format}', "");
Expect(0, 917632, '\P{^-Format}', "");
Error('\p{-is_FORMAT:=}');
Error('\P{-is_FORMAT:=}');
Expect(1, 917631, '\p{isformat}', "");
Expect(0, 917631, '\p{^isformat}', "");
Expect(0, 917631, '\P{isformat}', "");
Expect(1, 917631, '\P{^isformat}', "");
Expect(0, 917632, '\p{isformat}', "");
Expect(1, 917632, '\p{^isformat}', "");
Expect(1, 917632, '\P{isformat}', "");
Expect(0, 917632, '\P{^isformat}', "");
Expect(1, 917631, '\p{		Is_Format}', "");
Expect(0, 917631, '\p{^		Is_Format}', "");
Expect(0, 917631, '\P{		Is_Format}', "");
Expect(1, 917631, '\P{^		Is_Format}', "");
Expect(0, 917632, '\p{		Is_Format}', "");
Expect(1, 917632, '\p{^		Is_Format}', "");
Expect(1, 917632, '\P{		Is_Format}', "");
Expect(0, 917632, '\P{^		Is_Format}', "");
Error('\p{	:=CF}');
Error('\P{	:=CF}');
Expect(1, 917631, '\p{cf}', "");
Expect(0, 917631, '\p{^cf}', "");
Expect(0, 917631, '\P{cf}', "");
Expect(1, 917631, '\P{^cf}', "");
Expect(0, 917632, '\p{cf}', "");
Expect(1, 917632, '\p{^cf}', "");
Expect(1, 917632, '\P{cf}', "");
Expect(0, 917632, '\P{^cf}', "");
Expect(1, 917631, '\p{ Cf}', "");
Expect(0, 917631, '\p{^ Cf}', "");
Expect(0, 917631, '\P{ Cf}', "");
Expect(1, 917631, '\P{^ Cf}', "");
Expect(0, 917632, '\p{ Cf}', "");
Expect(1, 917632, '\p{^ Cf}', "");
Expect(1, 917632, '\P{ Cf}', "");
Expect(0, 917632, '\P{^ Cf}', "");
Error('\p{_:=IS_Cf}');
Error('\P{_:=IS_Cf}');
Expect(1, 917631, '\p{iscf}', "");
Expect(0, 917631, '\p{^iscf}', "");
Expect(0, 917631, '\P{iscf}', "");
Expect(1, 917631, '\P{^iscf}', "");
Expect(0, 917632, '\p{iscf}', "");
Expect(1, 917632, '\p{^iscf}', "");
Expect(1, 917632, '\P{iscf}', "");
Expect(0, 917632, '\P{^iscf}', "");
Expect(1, 917631, '\p{		is_CF}', "");
Expect(0, 917631, '\p{^		is_CF}', "");
Expect(0, 917631, '\P{		is_CF}', "");
Expect(1, 917631, '\P{^		is_CF}', "");
Expect(0, 917632, '\p{		is_CF}', "");
Expect(1, 917632, '\p{^		is_CF}', "");
Expect(1, 917632, '\P{		is_CF}', "");
Expect(0, 917632, '\P{^		is_CF}', "");
Error('\p{/a/ Full_COMPOSITION_exclusion}');
Error('\P{/a/ Full_COMPOSITION_exclusion}');
Expect(1, 195101, '\p{fullcompositionexclusion}', "");
Expect(0, 195101, '\p{^fullcompositionexclusion}', "");
Expect(0, 195101, '\P{fullcompositionexclusion}', "");
Expect(1, 195101, '\P{^fullcompositionexclusion}', "");
Expect(0, 195102, '\p{fullcompositionexclusion}', "");
Expect(1, 195102, '\p{^fullcompositionexclusion}', "");
Expect(1, 195102, '\P{fullcompositionexclusion}', "");
Expect(0, 195102, '\P{^fullcompositionexclusion}', "");
Expect(1, 195101, '\p{--FULL_COMPOSITION_Exclusion}', "");
Expect(0, 195101, '\p{^--FULL_COMPOSITION_Exclusion}', "");
Expect(0, 195101, '\P{--FULL_COMPOSITION_Exclusion}', "");
Expect(1, 195101, '\P{^--FULL_COMPOSITION_Exclusion}', "");
Expect(0, 195102, '\p{--FULL_COMPOSITION_Exclusion}', "");
Expect(1, 195102, '\p{^--FULL_COMPOSITION_Exclusion}', "");
Expect(1, 195102, '\P{--FULL_COMPOSITION_Exclusion}', "");
Expect(0, 195102, '\P{^--FULL_COMPOSITION_Exclusion}', "");
Error('\p{:= Is_FULL_Composition_exclusion}');
Error('\P{:= Is_FULL_Composition_exclusion}');
Expect(1, 195101, '\p{isfullcompositionexclusion}', "");
Expect(0, 195101, '\p{^isfullcompositionexclusion}', "");
Expect(0, 195101, '\P{isfullcompositionexclusion}', "");
Expect(1, 195101, '\P{^isfullcompositionexclusion}', "");
Expect(0, 195102, '\p{isfullcompositionexclusion}', "");
Expect(1, 195102, '\p{^isfullcompositionexclusion}', "");
Expect(1, 195102, '\P{isfullcompositionexclusion}', "");
Expect(0, 195102, '\P{^isfullcompositionexclusion}', "");
Expect(1, 195101, '\p{		Is_Full_COMPOSITION_Exclusion}', "");
Expect(0, 195101, '\p{^		Is_Full_COMPOSITION_Exclusion}', "");
Expect(0, 195101, '\P{		Is_Full_COMPOSITION_Exclusion}', "");
Expect(1, 195101, '\P{^		Is_Full_COMPOSITION_Exclusion}', "");
Expect(0, 195102, '\p{		Is_Full_COMPOSITION_Exclusion}', "");
Expect(1, 195102, '\p{^		Is_Full_COMPOSITION_Exclusion}', "");
Expect(1, 195102, '\P{		Is_Full_COMPOSITION_Exclusion}', "");
Expect(0, 195102, '\P{^		Is_Full_COMPOSITION_Exclusion}', "");
Error('\p{/a/--comp_Ex}');
Error('\P{/a/--comp_Ex}');
Expect(1, 195101, '\p{compex}', "");
Expect(0, 195101, '\p{^compex}', "");
Expect(0, 195101, '\P{compex}', "");
Expect(1, 195101, '\P{^compex}', "");
Expect(0, 195102, '\p{compex}', "");
Expect(1, 195102, '\p{^compex}', "");
Expect(1, 195102, '\P{compex}', "");
Expect(0, 195102, '\P{^compex}', "");
Expect(1, 195101, '\p{		Comp_Ex}', "");
Expect(0, 195101, '\p{^		Comp_Ex}', "");
Expect(0, 195101, '\P{		Comp_Ex}', "");
Expect(1, 195101, '\P{^		Comp_Ex}', "");
Expect(0, 195102, '\p{		Comp_Ex}', "");
Expect(1, 195102, '\p{^		Comp_Ex}', "");
Expect(1, 195102, '\P{		Comp_Ex}', "");
Expect(0, 195102, '\P{^		Comp_Ex}', "");
Error('\p{_:=Is_comp_Ex}');
Error('\P{_:=Is_comp_Ex}');
Expect(1, 195101, '\p{iscompex}', "");
Expect(0, 195101, '\p{^iscompex}', "");
Expect(0, 195101, '\P{iscompex}', "");
Expect(1, 195101, '\P{^iscompex}', "");
Expect(0, 195102, '\p{iscompex}', "");
Expect(1, 195102, '\p{^iscompex}', "");
Expect(1, 195102, '\P{iscompex}', "");
Expect(0, 195102, '\P{^iscompex}', "");
Expect(1, 195101, '\p{ 	is_Comp_Ex}', "");
Expect(0, 195101, '\p{^ 	is_Comp_Ex}', "");
Expect(0, 195101, '\P{ 	is_Comp_Ex}', "");
Expect(1, 195101, '\P{^ 	is_Comp_Ex}', "");
Expect(0, 195102, '\p{ 	is_Comp_Ex}', "");
Expect(1, 195102, '\p{^ 	is_Comp_Ex}', "");
Expect(1, 195102, '\P{ 	is_Comp_Ex}', "");
Expect(0, 195102, '\P{^ 	is_Comp_Ex}', "");
Error('\p{/a/ General_punctuation}');
Error('\P{/a/ General_punctuation}');
Expect(1, 8303, '\p{generalpunctuation}', "");
Expect(0, 8303, '\p{^generalpunctuation}', "");
Expect(0, 8303, '\P{generalpunctuation}', "");
Expect(1, 8303, '\P{^generalpunctuation}', "");
Expect(0, 8304, '\p{generalpunctuation}', "");
Expect(1, 8304, '\p{^generalpunctuation}', "");
Expect(1, 8304, '\P{generalpunctuation}', "");
Expect(0, 8304, '\P{^generalpunctuation}', "");
Expect(1, 8303, '\p{__General_punctuation}', "");
Expect(0, 8303, '\p{^__General_punctuation}', "");
Expect(0, 8303, '\P{__General_punctuation}', "");
Expect(1, 8303, '\P{^__General_punctuation}', "");
Expect(0, 8304, '\p{__General_punctuation}', "");
Expect(1, 8304, '\p{^__General_punctuation}', "");
Expect(1, 8304, '\P{__General_punctuation}', "");
Expect(0, 8304, '\P{^__General_punctuation}', "");
Error('\p{_ Is_General_punctuation:=}');
Error('\P{_ Is_General_punctuation:=}');
Expect(1, 8303, '\p{isgeneralpunctuation}', "");
Expect(0, 8303, '\p{^isgeneralpunctuation}', "");
Expect(0, 8303, '\P{isgeneralpunctuation}', "");
Expect(1, 8303, '\P{^isgeneralpunctuation}', "");
Expect(0, 8304, '\p{isgeneralpunctuation}', "");
Expect(1, 8304, '\p{^isgeneralpunctuation}', "");
Expect(1, 8304, '\P{isgeneralpunctuation}', "");
Expect(0, 8304, '\P{^isgeneralpunctuation}', "");
Expect(1, 8303, '\p{	 Is_General_PUNCTUATION}', "");
Expect(0, 8303, '\p{^	 Is_General_PUNCTUATION}', "");
Expect(0, 8303, '\P{	 Is_General_PUNCTUATION}', "");
Expect(1, 8303, '\P{^	 Is_General_PUNCTUATION}', "");
Expect(0, 8304, '\p{	 Is_General_PUNCTUATION}', "");
Expect(1, 8304, '\p{^	 Is_General_PUNCTUATION}', "");
Expect(1, 8304, '\P{	 Is_General_PUNCTUATION}', "");
Expect(0, 8304, '\P{^	 Is_General_PUNCTUATION}', "");
Error('\p{ 	in_GENERAL_PUNCTUATION/a/}');
Error('\P{ 	in_GENERAL_PUNCTUATION/a/}');
Expect(1, 8303, '\p{ingeneralpunctuation}', "");
Expect(0, 8303, '\p{^ingeneralpunctuation}', "");
Expect(0, 8303, '\P{ingeneralpunctuation}', "");
Expect(1, 8303, '\P{^ingeneralpunctuation}', "");
Expect(0, 8304, '\p{ingeneralpunctuation}', "");
Expect(1, 8304, '\p{^ingeneralpunctuation}', "");
Expect(1, 8304, '\P{ingeneralpunctuation}', "");
Expect(0, 8304, '\P{^ingeneralpunctuation}', "");
Expect(1, 8303, '\p{	-IN_GENERAL_PUNCTUATION}', "");
Expect(0, 8303, '\p{^	-IN_GENERAL_PUNCTUATION}', "");
Expect(0, 8303, '\P{	-IN_GENERAL_PUNCTUATION}', "");
Expect(1, 8303, '\P{^	-IN_GENERAL_PUNCTUATION}', "");
Expect(0, 8304, '\p{	-IN_GENERAL_PUNCTUATION}', "");
Expect(1, 8304, '\p{^	-IN_GENERAL_PUNCTUATION}', "");
Expect(1, 8304, '\P{	-IN_GENERAL_PUNCTUATION}', "");
Expect(0, 8304, '\P{^	-IN_GENERAL_PUNCTUATION}', "");
Error('\p{-/a/IN_Punctuation}');
Error('\P{-/a/IN_Punctuation}');
Expect(1, 8303, '\p{inpunctuation}', "");
Expect(0, 8303, '\p{^inpunctuation}', "");
Expect(0, 8303, '\P{inpunctuation}', "");
Expect(1, 8303, '\P{^inpunctuation}', "");
Expect(0, 8304, '\p{inpunctuation}', "");
Expect(1, 8304, '\p{^inpunctuation}', "");
Expect(1, 8304, '\P{inpunctuation}', "");
Expect(0, 8304, '\P{^inpunctuation}', "");
Expect(1, 8303, '\p{		in_PUNCTUATION}', "");
Expect(0, 8303, '\p{^		in_PUNCTUATION}', "");
Expect(0, 8303, '\P{		in_PUNCTUATION}', "");
Expect(1, 8303, '\P{^		in_PUNCTUATION}', "");
Expect(0, 8304, '\p{		in_PUNCTUATION}', "");
Expect(1, 8304, '\p{^		in_PUNCTUATION}', "");
Expect(1, 8304, '\P{		in_PUNCTUATION}', "");
Expect(0, 8304, '\P{^		in_PUNCTUATION}', "");
Error('\p{:=- geometric_Shapes}');
Error('\P{:=- geometric_Shapes}');
Expect(1, 9727, '\p{geometricshapes}', "");
Expect(0, 9727, '\p{^geometricshapes}', "");
Expect(0, 9727, '\P{geometricshapes}', "");
Expect(1, 9727, '\P{^geometricshapes}', "");
Expect(0, 9728, '\p{geometricshapes}', "");
Expect(1, 9728, '\p{^geometricshapes}', "");
Expect(1, 9728, '\P{geometricshapes}', "");
Expect(0, 9728, '\P{^geometricshapes}', "");
Expect(1, 9727, '\p{ 	Geometric_Shapes}', "");
Expect(0, 9727, '\p{^ 	Geometric_Shapes}', "");
Expect(0, 9727, '\P{ 	Geometric_Shapes}', "");
Expect(1, 9727, '\P{^ 	Geometric_Shapes}', "");
Expect(0, 9728, '\p{ 	Geometric_Shapes}', "");
Expect(1, 9728, '\p{^ 	Geometric_Shapes}', "");
Expect(1, 9728, '\P{ 	Geometric_Shapes}', "");
Expect(0, 9728, '\P{^ 	Geometric_Shapes}', "");
Error('\p{/a/	Is_GEOMETRIC_shapes}');
Error('\P{/a/	Is_GEOMETRIC_shapes}');
Expect(1, 9727, '\p{isgeometricshapes}', "");
Expect(0, 9727, '\p{^isgeometricshapes}', "");
Expect(0, 9727, '\P{isgeometricshapes}', "");
Expect(1, 9727, '\P{^isgeometricshapes}', "");
Expect(0, 9728, '\p{isgeometricshapes}', "");
Expect(1, 9728, '\p{^isgeometricshapes}', "");
Expect(1, 9728, '\P{isgeometricshapes}', "");
Expect(0, 9728, '\P{^isgeometricshapes}', "");
Expect(1, 9727, '\p{ Is_Geometric_Shapes}', "");
Expect(0, 9727, '\p{^ Is_Geometric_Shapes}', "");
Expect(0, 9727, '\P{ Is_Geometric_Shapes}', "");
Expect(1, 9727, '\P{^ Is_Geometric_Shapes}', "");
Expect(0, 9728, '\p{ Is_Geometric_Shapes}', "");
Expect(1, 9728, '\p{^ Is_Geometric_Shapes}', "");
Expect(1, 9728, '\P{ Is_Geometric_Shapes}', "");
Expect(0, 9728, '\P{^ Is_Geometric_Shapes}', "");
Error('\p{_:=In_geometric_Shapes}');
Error('\P{_:=In_geometric_Shapes}');
Expect(1, 9727, '\p{ingeometricshapes}', "");
Expect(0, 9727, '\p{^ingeometricshapes}', "");
Expect(0, 9727, '\P{ingeometricshapes}', "");
Expect(1, 9727, '\P{^ingeometricshapes}', "");
Expect(0, 9728, '\p{ingeometricshapes}', "");
Expect(1, 9728, '\p{^ingeometricshapes}', "");
Expect(1, 9728, '\P{ingeometricshapes}', "");
Expect(0, 9728, '\P{^ingeometricshapes}', "");
Expect(1, 9727, '\p{ In_Geometric_SHAPES}', "");
Expect(0, 9727, '\p{^ In_Geometric_SHAPES}', "");
Expect(0, 9727, '\P{ In_Geometric_SHAPES}', "");
Expect(1, 9727, '\P{^ In_Geometric_SHAPES}', "");
Expect(0, 9728, '\p{ In_Geometric_SHAPES}', "");
Expect(1, 9728, '\p{^ In_Geometric_SHAPES}', "");
Expect(1, 9728, '\P{ In_Geometric_SHAPES}', "");
Expect(0, 9728, '\P{^ In_Geometric_SHAPES}', "");
Error('\p{ Georgian/a/}');
Error('\P{ Georgian/a/}');
Expect(1, 11565, '\p{georgian}', "");
Expect(0, 11565, '\p{^georgian}', "");
Expect(0, 11565, '\P{georgian}', "");
Expect(1, 11565, '\P{^georgian}', "");
Expect(0, 11566, '\p{georgian}', "");
Expect(1, 11566, '\p{^georgian}', "");
Expect(1, 11566, '\P{georgian}', "");
Expect(0, 11566, '\P{^georgian}', "");
Expect(1, 11565, '\p{	_GEORGIAN}', "");
Expect(0, 11565, '\p{^	_GEORGIAN}', "");
Expect(0, 11565, '\P{	_GEORGIAN}', "");
Expect(1, 11565, '\P{^	_GEORGIAN}', "");
Expect(0, 11566, '\p{	_GEORGIAN}', "");
Expect(1, 11566, '\p{^	_GEORGIAN}', "");
Expect(1, 11566, '\P{	_GEORGIAN}', "");
Expect(0, 11566, '\P{^	_GEORGIAN}', "");
Error('\p{		IS_georgian:=}');
Error('\P{		IS_georgian:=}');
Expect(1, 11565, '\p{isgeorgian}', "");
Expect(0, 11565, '\p{^isgeorgian}', "");
Expect(0, 11565, '\P{isgeorgian}', "");
Expect(1, 11565, '\P{^isgeorgian}', "");
Expect(0, 11566, '\p{isgeorgian}', "");
Expect(1, 11566, '\p{^isgeorgian}', "");
Expect(1, 11566, '\P{isgeorgian}', "");
Expect(0, 11566, '\P{^isgeorgian}', "");
Expect(1, 11565, '\p{	is_Georgian}', "");
Expect(0, 11565, '\p{^	is_Georgian}', "");
Expect(0, 11565, '\P{	is_Georgian}', "");
Expect(1, 11565, '\P{^	is_Georgian}', "");
Expect(0, 11566, '\p{	is_Georgian}', "");
Expect(1, 11566, '\p{^	is_Georgian}', "");
Expect(1, 11566, '\P{	is_Georgian}', "");
Expect(0, 11566, '\P{^	is_Georgian}', "");
Error('\p{	geor:=}');
Error('\P{	geor:=}');
Expect(1, 11565, '\p{geor}', "");
Expect(0, 11565, '\p{^geor}', "");
Expect(0, 11565, '\P{geor}', "");
Expect(1, 11565, '\P{^geor}', "");
Expect(0, 11566, '\p{geor}', "");
Expect(1, 11566, '\p{^geor}', "");
Expect(1, 11566, '\P{geor}', "");
Expect(0, 11566, '\P{^geor}', "");
Expect(1, 11565, '\p{ 	Geor}', "");
Expect(0, 11565, '\p{^ 	Geor}', "");
Expect(0, 11565, '\P{ 	Geor}', "");
Expect(1, 11565, '\P{^ 	Geor}', "");
Expect(0, 11566, '\p{ 	Geor}', "");
Expect(1, 11566, '\p{^ 	Geor}', "");
Expect(1, 11566, '\P{ 	Geor}', "");
Expect(0, 11566, '\P{^ 	Geor}', "");
Error('\p{:=-_IS_Geor}');
Error('\P{:=-_IS_Geor}');
Expect(1, 11565, '\p{isgeor}', "");
Expect(0, 11565, '\p{^isgeor}', "");
Expect(0, 11565, '\P{isgeor}', "");
Expect(1, 11565, '\P{^isgeor}', "");
Expect(0, 11566, '\p{isgeor}', "");
Expect(1, 11566, '\p{^isgeor}', "");
Expect(1, 11566, '\P{isgeor}', "");
Expect(0, 11566, '\P{^isgeor}', "");
Expect(1, 11565, '\p{ 	Is_Geor}', "");
Expect(0, 11565, '\p{^ 	Is_Geor}', "");
Expect(0, 11565, '\P{ 	Is_Geor}', "");
Expect(1, 11565, '\P{^ 	Is_Geor}', "");
Expect(0, 11566, '\p{ 	Is_Geor}', "");
Expect(1, 11566, '\p{^ 	Is_Geor}', "");
Expect(1, 11566, '\P{ 	Is_Geor}', "");
Expect(0, 11566, '\P{^ 	Is_Geor}', "");
Error('\p{/a/--Georgian_supplement}');
Error('\P{/a/--Georgian_supplement}');
Expect(1, 11567, '\p{georgiansupplement}', "");
Expect(0, 11567, '\p{^georgiansupplement}', "");
Expect(0, 11567, '\P{georgiansupplement}', "");
Expect(1, 11567, '\P{^georgiansupplement}', "");
Expect(0, 11568, '\p{georgiansupplement}', "");
Expect(1, 11568, '\p{^georgiansupplement}', "");
Expect(1, 11568, '\P{georgiansupplement}', "");
Expect(0, 11568, '\P{^georgiansupplement}', "");
Expect(1, 11567, '\p{_ GEORGIAN_supplement}', "");
Expect(0, 11567, '\p{^_ GEORGIAN_supplement}', "");
Expect(0, 11567, '\P{_ GEORGIAN_supplement}', "");
Expect(1, 11567, '\P{^_ GEORGIAN_supplement}', "");
Expect(0, 11568, '\p{_ GEORGIAN_supplement}', "");
Expect(1, 11568, '\p{^_ GEORGIAN_supplement}', "");
Expect(1, 11568, '\P{_ GEORGIAN_supplement}', "");
Expect(0, 11568, '\P{^_ GEORGIAN_supplement}', "");
Error('\p{-:=IS_Georgian_Supplement}');
Error('\P{-:=IS_Georgian_Supplement}');
Expect(1, 11567, '\p{isgeorgiansupplement}', "");
Expect(0, 11567, '\p{^isgeorgiansupplement}', "");
Expect(0, 11567, '\P{isgeorgiansupplement}', "");
Expect(1, 11567, '\P{^isgeorgiansupplement}', "");
Expect(0, 11568, '\p{isgeorgiansupplement}', "");
Expect(1, 11568, '\p{^isgeorgiansupplement}', "");
Expect(1, 11568, '\P{isgeorgiansupplement}', "");
Expect(0, 11568, '\P{^isgeorgiansupplement}', "");
Expect(1, 11567, '\p{__IS_GEORGIAN_supplement}', "");
Expect(0, 11567, '\p{^__IS_GEORGIAN_supplement}', "");
Expect(0, 11567, '\P{__IS_GEORGIAN_supplement}', "");
Expect(1, 11567, '\P{^__IS_GEORGIAN_supplement}', "");
Expect(0, 11568, '\p{__IS_GEORGIAN_supplement}', "");
Expect(1, 11568, '\p{^__IS_GEORGIAN_supplement}', "");
Expect(1, 11568, '\P{__IS_GEORGIAN_supplement}', "");
Expect(0, 11568, '\P{^__IS_GEORGIAN_supplement}', "");
Error('\p{/a/ _in_Georgian_SUPPLEMENT}');
Error('\P{/a/ _in_Georgian_SUPPLEMENT}');
Expect(1, 11567, '\p{ingeorgiansupplement}', "");
Expect(0, 11567, '\p{^ingeorgiansupplement}', "");
Expect(0, 11567, '\P{ingeorgiansupplement}', "");
Expect(1, 11567, '\P{^ingeorgiansupplement}', "");
Expect(0, 11568, '\p{ingeorgiansupplement}', "");
Expect(1, 11568, '\p{^ingeorgiansupplement}', "");
Expect(1, 11568, '\P{ingeorgiansupplement}', "");
Expect(0, 11568, '\P{^ingeorgiansupplement}', "");
Expect(1, 11567, '\p{	 In_georgian_SUPPLEMENT}', "");
Expect(0, 11567, '\p{^	 In_georgian_SUPPLEMENT}', "");
Expect(0, 11567, '\P{	 In_georgian_SUPPLEMENT}', "");
Expect(1, 11567, '\P{^	 In_georgian_SUPPLEMENT}', "");
Expect(0, 11568, '\p{	 In_georgian_SUPPLEMENT}', "");
Expect(1, 11568, '\p{^	 In_georgian_SUPPLEMENT}', "");
Expect(1, 11568, '\P{	 In_georgian_SUPPLEMENT}', "");
Expect(0, 11568, '\P{^	 In_georgian_SUPPLEMENT}', "");
Error('\p{georgian_Sup/a/}');
Error('\P{georgian_Sup/a/}');
Expect(1, 11567, '\p{georgiansup}', "");
Expect(0, 11567, '\p{^georgiansup}', "");
Expect(0, 11567, '\P{georgiansup}', "");
Expect(1, 11567, '\P{^georgiansup}', "");
Expect(0, 11568, '\p{georgiansup}', "");
Expect(1, 11568, '\p{^georgiansup}', "");
Expect(1, 11568, '\P{georgiansup}', "");
Expect(0, 11568, '\P{^georgiansup}', "");
Expect(1, 11567, '\p{  georgian_sup}', "");
Expect(0, 11567, '\p{^  georgian_sup}', "");
Expect(0, 11567, '\P{  georgian_sup}', "");
Expect(1, 11567, '\P{^  georgian_sup}', "");
Expect(0, 11568, '\p{  georgian_sup}', "");
Expect(1, 11568, '\p{^  georgian_sup}', "");
Expect(1, 11568, '\P{  georgian_sup}', "");
Expect(0, 11568, '\P{^  georgian_sup}', "");
Error('\p{_Is_Georgian_sup:=}');
Error('\P{_Is_Georgian_sup:=}');
Expect(1, 11567, '\p{isgeorgiansup}', "");
Expect(0, 11567, '\p{^isgeorgiansup}', "");
Expect(0, 11567, '\P{isgeorgiansup}', "");
Expect(1, 11567, '\P{^isgeorgiansup}', "");
Expect(0, 11568, '\p{isgeorgiansup}', "");
Expect(1, 11568, '\p{^isgeorgiansup}', "");
Expect(1, 11568, '\P{isgeorgiansup}', "");
Expect(0, 11568, '\P{^isgeorgiansup}', "");
Expect(1, 11567, '\p{ -IS_Georgian_Sup}', "");
Expect(0, 11567, '\p{^ -IS_Georgian_Sup}', "");
Expect(0, 11567, '\P{ -IS_Georgian_Sup}', "");
Expect(1, 11567, '\P{^ -IS_Georgian_Sup}', "");
Expect(0, 11568, '\p{ -IS_Georgian_Sup}', "");
Expect(1, 11568, '\p{^ -IS_Georgian_Sup}', "");
Expect(1, 11568, '\P{ -IS_Georgian_Sup}', "");
Expect(0, 11568, '\P{^ -IS_Georgian_Sup}', "");
Error('\p{ /a/In_GEORGIAN_Sup}');
Error('\P{ /a/In_GEORGIAN_Sup}');
Expect(1, 11567, '\p{ingeorgiansup}', "");
Expect(0, 11567, '\p{^ingeorgiansup}', "");
Expect(0, 11567, '\P{ingeorgiansup}', "");
Expect(1, 11567, '\P{^ingeorgiansup}', "");
Expect(0, 11568, '\p{ingeorgiansup}', "");
Expect(1, 11568, '\p{^ingeorgiansup}', "");
Expect(1, 11568, '\P{ingeorgiansup}', "");
Expect(0, 11568, '\P{^ingeorgiansup}', "");
Expect(1, 11567, '\p{		in_Georgian_SUP}', "");
Expect(0, 11567, '\p{^		in_Georgian_SUP}', "");
Expect(0, 11567, '\P{		in_Georgian_SUP}', "");
Expect(1, 11567, '\P{^		in_Georgian_SUP}', "");
Expect(0, 11568, '\p{		in_Georgian_SUP}', "");
Expect(1, 11568, '\p{^		in_Georgian_SUP}', "");
Expect(1, 11568, '\P{		in_Georgian_SUP}', "");
Expect(0, 11568, '\P{^		in_Georgian_SUP}', "");
Error('\p{/a/__Glagolitic}');
Error('\P{/a/__Glagolitic}');
Expect(1, 11358, '\p{glagolitic}', "");
Expect(0, 11358, '\p{^glagolitic}', "");
Expect(0, 11358, '\P{glagolitic}', "");
Expect(1, 11358, '\P{^glagolitic}', "");
Expect(0, 11359, '\p{glagolitic}', "");
Expect(1, 11359, '\p{^glagolitic}', "");
Expect(1, 11359, '\P{glagolitic}', "");
Expect(0, 11359, '\P{^glagolitic}', "");
Expect(1, 11358, '\p{	glagolitic}', "");
Expect(0, 11358, '\p{^	glagolitic}', "");
Expect(0, 11358, '\P{	glagolitic}', "");
Expect(1, 11358, '\P{^	glagolitic}', "");
Expect(0, 11359, '\p{	glagolitic}', "");
Expect(1, 11359, '\p{^	glagolitic}', "");
Expect(1, 11359, '\P{	glagolitic}', "");
Expect(0, 11359, '\P{^	glagolitic}', "");
Error('\p{/a/--IS_glagolitic}');
Error('\P{/a/--IS_glagolitic}');
Expect(1, 11358, '\p{isglagolitic}', "");
Expect(0, 11358, '\p{^isglagolitic}', "");
Expect(0, 11358, '\P{isglagolitic}', "");
Expect(1, 11358, '\P{^isglagolitic}', "");
Expect(0, 11359, '\p{isglagolitic}', "");
Expect(1, 11359, '\p{^isglagolitic}', "");
Expect(1, 11359, '\P{isglagolitic}', "");
Expect(0, 11359, '\P{^isglagolitic}', "");
Expect(1, 11358, '\p{_-Is_Glagolitic}', "");
Expect(0, 11358, '\p{^_-Is_Glagolitic}', "");
Expect(0, 11358, '\P{_-Is_Glagolitic}', "");
Expect(1, 11358, '\P{^_-Is_Glagolitic}', "");
Expect(0, 11359, '\p{_-Is_Glagolitic}', "");
Expect(1, 11359, '\p{^_-Is_Glagolitic}', "");
Expect(1, 11359, '\P{_-Is_Glagolitic}', "");
Expect(0, 11359, '\P{^_-Is_Glagolitic}', "");
Error('\p{	Glag/a/}');
Error('\P{	Glag/a/}');
Expect(1, 11358, '\p{glag}', "");
Expect(0, 11358, '\p{^glag}', "");
Expect(0, 11358, '\P{glag}', "");
Expect(1, 11358, '\P{^glag}', "");
Expect(0, 11359, '\p{glag}', "");
Expect(1, 11359, '\p{^glag}', "");
Expect(1, 11359, '\P{glag}', "");
Expect(0, 11359, '\P{^glag}', "");
Expect(1, 11358, '\p{	GLAG}', "");
Expect(0, 11358, '\p{^	GLAG}', "");
Expect(0, 11358, '\P{	GLAG}', "");
Expect(1, 11358, '\P{^	GLAG}', "");
Expect(0, 11359, '\p{	GLAG}', "");
Expect(1, 11359, '\p{^	GLAG}', "");
Expect(1, 11359, '\P{	GLAG}', "");
Expect(0, 11359, '\P{^	GLAG}', "");
Error('\p{/a/	_Is_glag}');
Error('\P{/a/	_Is_glag}');
Expect(1, 11358, '\p{isglag}', "");
Expect(0, 11358, '\p{^isglag}', "");
Expect(0, 11358, '\P{isglag}', "");
Expect(1, 11358, '\P{^isglag}', "");
Expect(0, 11359, '\p{isglag}', "");
Expect(1, 11359, '\p{^isglag}', "");
Expect(1, 11359, '\P{isglag}', "");
Expect(0, 11359, '\P{^isglag}', "");
Expect(1, 11358, '\p{		is_Glag}', "");
Expect(0, 11358, '\p{^		is_Glag}', "");
Expect(0, 11358, '\P{		is_Glag}', "");
Expect(1, 11358, '\P{^		is_Glag}', "");
Expect(0, 11359, '\p{		is_Glag}', "");
Expect(1, 11359, '\p{^		is_Glag}', "");
Expect(1, 11359, '\P{		is_Glag}', "");
Expect(0, 11359, '\P{^		is_Glag}', "");
Error('\p{ /a/Gothic}');
Error('\P{ /a/Gothic}');
Expect(1, 66378, '\p{gothic}', "");
Expect(0, 66378, '\p{^gothic}', "");
Expect(0, 66378, '\P{gothic}', "");
Expect(1, 66378, '\P{^gothic}', "");
Expect(0, 66379, '\p{gothic}', "");
Expect(1, 66379, '\p{^gothic}', "");
Expect(1, 66379, '\P{gothic}', "");
Expect(0, 66379, '\P{^gothic}', "");
Expect(1, 66378, '\p{ 	GOTHIC}', "");
Expect(0, 66378, '\p{^ 	GOTHIC}', "");
Expect(0, 66378, '\P{ 	GOTHIC}', "");
Expect(1, 66378, '\P{^ 	GOTHIC}', "");
Expect(0, 66379, '\p{ 	GOTHIC}', "");
Expect(1, 66379, '\p{^ 	GOTHIC}', "");
Expect(1, 66379, '\P{ 	GOTHIC}', "");
Expect(0, 66379, '\P{^ 	GOTHIC}', "");
Error('\p{ _is_Gothic:=}');
Error('\P{ _is_Gothic:=}');
Expect(1, 66378, '\p{isgothic}', "");
Expect(0, 66378, '\p{^isgothic}', "");
Expect(0, 66378, '\P{isgothic}', "");
Expect(1, 66378, '\P{^isgothic}', "");
Expect(0, 66379, '\p{isgothic}', "");
Expect(1, 66379, '\p{^isgothic}', "");
Expect(1, 66379, '\P{isgothic}', "");
Expect(0, 66379, '\P{^isgothic}', "");
Expect(1, 66378, '\p{_ IS_gothic}', "");
Expect(0, 66378, '\p{^_ IS_gothic}', "");
Expect(0, 66378, '\P{_ IS_gothic}', "");
Expect(1, 66378, '\P{^_ IS_gothic}', "");
Expect(0, 66379, '\p{_ IS_gothic}', "");
Expect(1, 66379, '\p{^_ IS_gothic}', "");
Expect(1, 66379, '\P{_ IS_gothic}', "");
Expect(0, 66379, '\P{^_ IS_gothic}', "");
Error('\p{--GOTH:=}');
Error('\P{--GOTH:=}');
Expect(1, 66378, '\p{goth}', "");
Expect(0, 66378, '\p{^goth}', "");
Expect(0, 66378, '\P{goth}', "");
Expect(1, 66378, '\P{^goth}', "");
Expect(0, 66379, '\p{goth}', "");
Expect(1, 66379, '\p{^goth}', "");
Expect(1, 66379, '\P{goth}', "");
Expect(0, 66379, '\P{^goth}', "");
Expect(1, 66378, '\p{	-goth}', "");
Expect(0, 66378, '\p{^	-goth}', "");
Expect(0, 66378, '\P{	-goth}', "");
Expect(1, 66378, '\P{^	-goth}', "");
Expect(0, 66379, '\p{	-goth}', "");
Expect(1, 66379, '\p{^	-goth}', "");
Expect(1, 66379, '\P{	-goth}', "");
Expect(0, 66379, '\P{^	-goth}', "");
Error('\p{:=  Is_GOTH}');
Error('\P{:=  Is_GOTH}');
Expect(1, 66378, '\p{isgoth}', "");
Expect(0, 66378, '\p{^isgoth}', "");
Expect(0, 66378, '\P{isgoth}', "");
Expect(1, 66378, '\P{^isgoth}', "");
Expect(0, 66379, '\p{isgoth}', "");
Expect(1, 66379, '\p{^isgoth}', "");
Expect(1, 66379, '\P{isgoth}', "");
Expect(0, 66379, '\P{^isgoth}', "");
Expect(1, 66378, '\p{__IS_Goth}', "");
Expect(0, 66378, '\p{^__IS_Goth}', "");
Expect(0, 66378, '\P{__IS_Goth}', "");
Expect(1, 66378, '\P{^__IS_Goth}', "");
Expect(0, 66379, '\p{__IS_Goth}', "");
Expect(1, 66379, '\p{^__IS_Goth}', "");
Expect(1, 66379, '\P{__IS_Goth}', "");
Expect(0, 66379, '\P{^__IS_Goth}', "");
Error('\p{	graph:=}');
Error('\P{	graph:=}');
Expect(1, 1114109, '\p{graph}', "");
Expect(0, 1114109, '\p{^graph}', "");
Expect(0, 1114109, '\P{graph}', "");
Expect(1, 1114109, '\P{^graph}', "");
Expect(0, 918000, '\p{graph}', "");
Expect(1, 918000, '\p{^graph}', "");
Expect(1, 918000, '\P{graph}', "");
Expect(0, 918000, '\P{^graph}', "");
Expect(1, 1114109, '\p{-Graph}', "");
Expect(0, 1114109, '\p{^-Graph}', "");
Expect(0, 1114109, '\P{-Graph}', "");
Expect(1, 1114109, '\P{^-Graph}', "");
Expect(0, 918000, '\p{-Graph}', "");
Expect(1, 918000, '\p{^-Graph}', "");
Expect(1, 918000, '\P{-Graph}', "");
Expect(0, 918000, '\P{^-Graph}', "");
Error('\p{__XPosixGraph/a/}');
Error('\P{__XPosixGraph/a/}');
Expect(1, 1114109, '\p{xposixgraph}', "");
Expect(0, 1114109, '\p{^xposixgraph}', "");
Expect(0, 1114109, '\P{xposixgraph}', "");
Expect(1, 1114109, '\P{^xposixgraph}', "");
Expect(0, 918000, '\p{xposixgraph}', "");
Expect(1, 918000, '\p{^xposixgraph}', "");
Expect(1, 918000, '\P{xposixgraph}', "");
Expect(0, 918000, '\P{^xposixgraph}', "");
Expect(1, 1114109, '\p{ _XPOSIXGRAPH}', "");
Expect(0, 1114109, '\p{^ _XPOSIXGRAPH}', "");
Expect(0, 1114109, '\P{ _XPOSIXGRAPH}', "");
Expect(1, 1114109, '\P{^ _XPOSIXGRAPH}', "");
Expect(0, 918000, '\p{ _XPOSIXGRAPH}', "");
Expect(1, 918000, '\p{^ _XPOSIXGRAPH}', "");
Expect(1, 918000, '\P{ _XPOSIXGRAPH}', "");
Expect(0, 918000, '\P{^ _XPOSIXGRAPH}', "");
Error('\p{ :=Is_GRAPH}');
Error('\P{ :=Is_GRAPH}');
Expect(1, 1114109, '\p{isgraph}', "");
Expect(0, 1114109, '\p{^isgraph}', "");
Expect(0, 1114109, '\P{isgraph}', "");
Expect(1, 1114109, '\P{^isgraph}', "");
Expect(0, 918000, '\p{isgraph}', "");
Expect(1, 918000, '\p{^isgraph}', "");
Expect(1, 918000, '\P{isgraph}', "");
Expect(0, 918000, '\P{^isgraph}', "");
Expect(1, 1114109, '\p{_	is_Graph}', "");
Expect(0, 1114109, '\p{^_	is_Graph}', "");
Expect(0, 1114109, '\P{_	is_Graph}', "");
Expect(1, 1114109, '\P{^_	is_Graph}', "");
Expect(0, 918000, '\p{_	is_Graph}', "");
Expect(1, 918000, '\p{^_	is_Graph}', "");
Expect(1, 918000, '\P{_	is_Graph}', "");
Expect(0, 918000, '\P{^_	is_Graph}', "");
Error('\p{:=		Is_XPosixGraph}');
Error('\P{:=		Is_XPosixGraph}');
Expect(1, 1114109, '\p{isxposixgraph}', "");
Expect(0, 1114109, '\p{^isxposixgraph}', "");
Expect(0, 1114109, '\P{isxposixgraph}', "");
Expect(1, 1114109, '\P{^isxposixgraph}', "");
Expect(0, 918000, '\p{isxposixgraph}', "");
Expect(1, 918000, '\p{^isxposixgraph}', "");
Expect(1, 918000, '\P{isxposixgraph}', "");
Expect(0, 918000, '\P{^isxposixgraph}', "");
Expect(1, 1114109, '\p{ 	is_XPOSIXGRAPH}', "");
Expect(0, 1114109, '\p{^ 	is_XPOSIXGRAPH}', "");
Expect(0, 1114109, '\P{ 	is_XPOSIXGRAPH}', "");
Expect(1, 1114109, '\P{^ 	is_XPOSIXGRAPH}', "");
Expect(0, 918000, '\p{ 	is_XPOSIXGRAPH}', "");
Expect(1, 918000, '\p{^ 	is_XPOSIXGRAPH}', "");
Expect(1, 918000, '\P{ 	is_XPOSIXGRAPH}', "");
Expect(0, 918000, '\P{^ 	is_XPOSIXGRAPH}', "");
Error('\p{:=_Grapheme_Base}');
Error('\P{:=_Grapheme_Base}');
Expect(1, 195101, '\p{graphemebase}', "");
Expect(0, 195101, '\p{^graphemebase}', "");
Expect(0, 195101, '\P{graphemebase}', "");
Expect(1, 195101, '\P{^graphemebase}', "");
Expect(0, 195102, '\p{graphemebase}', "");
Expect(1, 195102, '\p{^graphemebase}', "");
Expect(1, 195102, '\P{graphemebase}', "");
Expect(0, 195102, '\P{^graphemebase}', "");
Expect(1, 195101, '\p{-grapheme_Base}', "");
Expect(0, 195101, '\p{^-grapheme_Base}', "");
Expect(0, 195101, '\P{-grapheme_Base}', "");
Expect(1, 195101, '\P{^-grapheme_Base}', "");
Expect(0, 195102, '\p{-grapheme_Base}', "");
Expect(1, 195102, '\p{^-grapheme_Base}', "");
Expect(1, 195102, '\P{-grapheme_Base}', "");
Expect(0, 195102, '\P{^-grapheme_Base}', "");
Error('\p{_/a/Is_grapheme_base}');
Error('\P{_/a/Is_grapheme_base}');
Expect(1, 195101, '\p{isgraphemebase}', "");
Expect(0, 195101, '\p{^isgraphemebase}', "");
Expect(0, 195101, '\P{isgraphemebase}', "");
Expect(1, 195101, '\P{^isgraphemebase}', "");
Expect(0, 195102, '\p{isgraphemebase}', "");
Expect(1, 195102, '\p{^isgraphemebase}', "");
Expect(1, 195102, '\P{isgraphemebase}', "");
Expect(0, 195102, '\P{^isgraphemebase}', "");
Expect(1, 195101, '\p{--IS_Grapheme_Base}', "");
Expect(0, 195101, '\p{^--IS_Grapheme_Base}', "");
Expect(0, 195101, '\P{--IS_Grapheme_Base}', "");
Expect(1, 195101, '\P{^--IS_Grapheme_Base}', "");
Expect(0, 195102, '\p{--IS_Grapheme_Base}', "");
Expect(1, 195102, '\p{^--IS_Grapheme_Base}', "");
Expect(1, 195102, '\P{--IS_Grapheme_Base}', "");
Expect(0, 195102, '\P{^--IS_Grapheme_Base}', "");
Error('\p{/a/ 	Gr_base}');
Error('\P{/a/ 	Gr_base}');
Expect(1, 195101, '\p{grbase}', "");
Expect(0, 195101, '\p{^grbase}', "");
Expect(0, 195101, '\P{grbase}', "");
Expect(1, 195101, '\P{^grbase}', "");
Expect(0, 195102, '\p{grbase}', "");
Expect(1, 195102, '\p{^grbase}', "");
Expect(1, 195102, '\P{grbase}', "");
Expect(0, 195102, '\P{^grbase}', "");
Expect(1, 195101, '\p{_ GR_Base}', "");
Expect(0, 195101, '\p{^_ GR_Base}', "");
Expect(0, 195101, '\P{_ GR_Base}', "");
Expect(1, 195101, '\P{^_ GR_Base}', "");
Expect(0, 195102, '\p{_ GR_Base}', "");
Expect(1, 195102, '\p{^_ GR_Base}', "");
Expect(1, 195102, '\P{_ GR_Base}', "");
Expect(0, 195102, '\P{^_ GR_Base}', "");
Error('\p{/a/--is_gr_base}');
Error('\P{/a/--is_gr_base}');
Expect(1, 195101, '\p{isgrbase}', "");
Expect(0, 195101, '\p{^isgrbase}', "");
Expect(0, 195101, '\P{isgrbase}', "");
Expect(1, 195101, '\P{^isgrbase}', "");
Expect(0, 195102, '\p{isgrbase}', "");
Expect(1, 195102, '\p{^isgrbase}', "");
Expect(1, 195102, '\P{isgrbase}', "");
Expect(0, 195102, '\P{^isgrbase}', "");
Expect(1, 195101, '\p{Is_Gr_Base}', "");
Expect(0, 195101, '\p{^Is_Gr_Base}', "");
Expect(0, 195101, '\P{Is_Gr_Base}', "");
Expect(1, 195101, '\P{^Is_Gr_Base}', "");
Expect(0, 195102, '\p{Is_Gr_Base}', "");
Expect(1, 195102, '\p{^Is_Gr_Base}', "");
Expect(1, 195102, '\P{Is_Gr_Base}', "");
Expect(0, 195102, '\P{^Is_Gr_Base}', "");
Error('\p{:=	Grapheme_Extend}');
Error('\P{:=	Grapheme_Extend}');
Expect(1, 917999, '\p{graphemeextend}', "");
Expect(0, 917999, '\p{^graphemeextend}', "");
Expect(0, 917999, '\P{graphemeextend}', "");
Expect(1, 917999, '\P{^graphemeextend}', "");
Expect(0, 918000, '\p{graphemeextend}', "");
Expect(1, 918000, '\p{^graphemeextend}', "");
Expect(1, 918000, '\P{graphemeextend}', "");
Expect(0, 918000, '\P{^graphemeextend}', "");
Expect(1, 917999, '\p{ -Grapheme_EXTEND}', "");
Expect(0, 917999, '\p{^ -Grapheme_EXTEND}', "");
Expect(0, 917999, '\P{ -Grapheme_EXTEND}', "");
Expect(1, 917999, '\P{^ -Grapheme_EXTEND}', "");
Expect(0, 918000, '\p{ -Grapheme_EXTEND}', "");
Expect(1, 918000, '\p{^ -Grapheme_EXTEND}', "");
Expect(1, 918000, '\P{ -Grapheme_EXTEND}', "");
Expect(0, 918000, '\P{^ -Grapheme_EXTEND}', "");
Error('\p{	:=is_GRAPHEME_Extend}');
Error('\P{	:=is_GRAPHEME_Extend}');
Expect(1, 917999, '\p{isgraphemeextend}', "");
Expect(0, 917999, '\p{^isgraphemeextend}', "");
Expect(0, 917999, '\P{isgraphemeextend}', "");
Expect(1, 917999, '\P{^isgraphemeextend}', "");
Expect(0, 918000, '\p{isgraphemeextend}', "");
Expect(1, 918000, '\p{^isgraphemeextend}', "");
Expect(1, 918000, '\P{isgraphemeextend}', "");
Expect(0, 918000, '\P{^isgraphemeextend}', "");
Expect(1, 917999, '\p{-	is_Grapheme_Extend}', "");
Expect(0, 917999, '\p{^-	is_Grapheme_Extend}', "");
Expect(0, 917999, '\P{-	is_Grapheme_Extend}', "");
Expect(1, 917999, '\P{^-	is_Grapheme_Extend}', "");
Expect(0, 918000, '\p{-	is_Grapheme_Extend}', "");
Expect(1, 918000, '\p{^-	is_Grapheme_Extend}', "");
Expect(1, 918000, '\P{-	is_Grapheme_Extend}', "");
Expect(0, 918000, '\P{^-	is_Grapheme_Extend}', "");
Error('\p{:=_GR_ext}');
Error('\P{:=_GR_ext}');
Expect(1, 917999, '\p{grext}', "");
Expect(0, 917999, '\p{^grext}', "");
Expect(0, 917999, '\P{grext}', "");
Expect(1, 917999, '\P{^grext}', "");
Expect(0, 918000, '\p{grext}', "");
Expect(1, 918000, '\p{^grext}', "");
Expect(1, 918000, '\P{grext}', "");
Expect(0, 918000, '\P{^grext}', "");
Expect(1, 917999, '\p{		Gr_Ext}', "");
Expect(0, 917999, '\p{^		Gr_Ext}', "");
Expect(0, 917999, '\P{		Gr_Ext}', "");
Expect(1, 917999, '\P{^		Gr_Ext}', "");
Expect(0, 918000, '\p{		Gr_Ext}', "");
Expect(1, 918000, '\p{^		Gr_Ext}', "");
Expect(1, 918000, '\P{		Gr_Ext}', "");
Expect(0, 918000, '\P{^		Gr_Ext}', "");
Error('\p{/a/-IS_GR_EXT}');
Error('\P{/a/-IS_GR_EXT}');
Expect(1, 917999, '\p{isgrext}', "");
Expect(0, 917999, '\p{^isgrext}', "");
Expect(0, 917999, '\P{isgrext}', "");
Expect(1, 917999, '\P{^isgrext}', "");
Expect(0, 918000, '\p{isgrext}', "");
Expect(1, 918000, '\p{^isgrext}', "");
Expect(1, 918000, '\P{isgrext}', "");
Expect(0, 918000, '\P{^isgrext}', "");
Expect(1, 917999, '\p{		Is_Gr_Ext}', "");
Expect(0, 917999, '\p{^		Is_Gr_Ext}', "");
Expect(0, 917999, '\P{		Is_Gr_Ext}', "");
Expect(1, 917999, '\P{^		Is_Gr_Ext}', "");
Expect(0, 918000, '\p{		Is_Gr_Ext}', "");
Expect(1, 918000, '\p{^		Is_Gr_Ext}', "");
Expect(1, 918000, '\P{		Is_Gr_Ext}', "");
Expect(0, 918000, '\P{^		Is_Gr_Ext}', "");
Error('\p{Grapheme_Link}');
Error('\P{Grapheme_Link}');
Error('\p{Is_Grapheme_Link}');
Error('\P{Is_Grapheme_Link}');
Error('\p{Gr_Link}');
Error('\P{Gr_Link}');
Error('\p{Is_Gr_Link}');
Error('\P{Is_Gr_Link}');
Error('\p{:= -greek}');
Error('\P{:= -greek}');
Expect(1, 119365, '\p{greek}', "");
Expect(0, 119365, '\p{^greek}', "");
Expect(0, 119365, '\P{greek}', "");
Expect(1, 119365, '\P{^greek}', "");
Expect(0, 119366, '\p{greek}', "");
Expect(1, 119366, '\p{^greek}', "");
Expect(1, 119366, '\P{greek}', "");
Expect(0, 119366, '\P{^greek}', "");
Expect(1, 119365, '\p{_ Greek}', "");
Expect(0, 119365, '\p{^_ Greek}', "");
Expect(0, 119365, '\P{_ Greek}', "");
Expect(1, 119365, '\P{^_ Greek}', "");
Expect(0, 119366, '\p{_ Greek}', "");
Expect(1, 119366, '\p{^_ Greek}', "");
Expect(1, 119366, '\P{_ Greek}', "");
Expect(0, 119366, '\P{^_ Greek}', "");
Error('\p{/a/_	IS_greek}');
Error('\P{/a/_	IS_greek}');
Expect(1, 119365, '\p{isgreek}', "");
Expect(0, 119365, '\p{^isgreek}', "");
Expect(0, 119365, '\P{isgreek}', "");
Expect(1, 119365, '\P{^isgreek}', "");
Expect(0, 119366, '\p{isgreek}', "");
Expect(1, 119366, '\p{^isgreek}', "");
Expect(1, 119366, '\P{isgreek}', "");
Expect(0, 119366, '\P{^isgreek}', "");
Expect(1, 119365, '\p{ IS_greek}', "");
Expect(0, 119365, '\p{^ IS_greek}', "");
Expect(0, 119365, '\P{ IS_greek}', "");
Expect(1, 119365, '\P{^ IS_greek}', "");
Expect(0, 119366, '\p{ IS_greek}', "");
Expect(1, 119366, '\p{^ IS_greek}', "");
Expect(1, 119366, '\P{ IS_greek}', "");
Expect(0, 119366, '\P{^ IS_greek}', "");
Error('\p{/a/ GREK}');
Error('\P{/a/ GREK}');
Expect(1, 119365, '\p{grek}', "");
Expect(0, 119365, '\p{^grek}', "");
Expect(0, 119365, '\P{grek}', "");
Expect(1, 119365, '\P{^grek}', "");
Expect(0, 119366, '\p{grek}', "");
Expect(1, 119366, '\p{^grek}', "");
Expect(1, 119366, '\P{grek}', "");
Expect(0, 119366, '\P{^grek}', "");
Expect(1, 119365, '\p{	 Grek}', "");
Expect(0, 119365, '\p{^	 Grek}', "");
Expect(0, 119365, '\P{	 Grek}', "");
Expect(1, 119365, '\P{^	 Grek}', "");
Expect(0, 119366, '\p{	 Grek}', "");
Expect(1, 119366, '\p{^	 Grek}', "");
Expect(1, 119366, '\P{	 Grek}', "");
Expect(0, 119366, '\P{^	 Grek}', "");
Error('\p{	is_Grek:=}');
Error('\P{	is_Grek:=}');
Expect(1, 119365, '\p{isgrek}', "");
Expect(0, 119365, '\p{^isgrek}', "");
Expect(0, 119365, '\P{isgrek}', "");
Expect(1, 119365, '\P{^isgrek}', "");
Expect(0, 119366, '\p{isgrek}', "");
Expect(1, 119366, '\p{^isgrek}', "");
Expect(1, 119366, '\P{isgrek}', "");
Expect(0, 119366, '\P{^isgrek}', "");
Expect(1, 119365, '\p{	is_Grek}', "");
Expect(0, 119365, '\p{^	is_Grek}', "");
Expect(0, 119365, '\P{	is_Grek}', "");
Expect(1, 119365, '\P{^	is_Grek}', "");
Expect(0, 119366, '\p{	is_Grek}', "");
Expect(1, 119366, '\p{^	is_Grek}', "");
Expect(1, 119366, '\P{	is_Grek}', "");
Expect(0, 119366, '\P{^	is_Grek}', "");
Error('\p{ :=Greek_AND_Coptic}');
Error('\P{ :=Greek_AND_Coptic}');
Expect(1, 1023, '\p{greekandcoptic}', "");
Expect(0, 1023, '\p{^greekandcoptic}', "");
Expect(0, 1023, '\P{greekandcoptic}', "");
Expect(1, 1023, '\P{^greekandcoptic}', "");
Expect(0, 1024, '\p{greekandcoptic}', "");
Expect(1, 1024, '\p{^greekandcoptic}', "");
Expect(1, 1024, '\P{greekandcoptic}', "");
Expect(0, 1024, '\P{^greekandcoptic}', "");
Expect(1, 1023, '\p{	-GREEK_And_coptic}', "");
Expect(0, 1023, '\p{^	-GREEK_And_coptic}', "");
Expect(0, 1023, '\P{	-GREEK_And_coptic}', "");
Expect(1, 1023, '\P{^	-GREEK_And_coptic}', "");
Expect(0, 1024, '\p{	-GREEK_And_coptic}', "");
Expect(1, 1024, '\p{^	-GREEK_And_coptic}', "");
Expect(1, 1024, '\P{	-GREEK_And_coptic}', "");
Expect(0, 1024, '\P{^	-GREEK_And_coptic}', "");
Error('\p{:=is_Greek_And_Coptic}');
Error('\P{:=is_Greek_And_Coptic}');
Expect(1, 1023, '\p{isgreekandcoptic}', "");
Expect(0, 1023, '\p{^isgreekandcoptic}', "");
Expect(0, 1023, '\P{isgreekandcoptic}', "");
Expect(1, 1023, '\P{^isgreekandcoptic}', "");
Expect(0, 1024, '\p{isgreekandcoptic}', "");
Expect(1, 1024, '\p{^isgreekandcoptic}', "");
Expect(1, 1024, '\P{isgreekandcoptic}', "");
Expect(0, 1024, '\P{^isgreekandcoptic}', "");
Expect(1, 1023, '\p{_IS_Greek_AND_COPTIC}', "");
Expect(0, 1023, '\p{^_IS_Greek_AND_COPTIC}', "");
Expect(0, 1023, '\P{_IS_Greek_AND_COPTIC}', "");
Expect(1, 1023, '\P{^_IS_Greek_AND_COPTIC}', "");
Expect(0, 1024, '\p{_IS_Greek_AND_COPTIC}', "");
Expect(1, 1024, '\p{^_IS_Greek_AND_COPTIC}', "");
Expect(1, 1024, '\P{_IS_Greek_AND_COPTIC}', "");
Expect(0, 1024, '\P{^_IS_Greek_AND_COPTIC}', "");
Error('\p{-	In_greek_And_Coptic:=}');
Error('\P{-	In_greek_And_Coptic:=}');
Expect(1, 1023, '\p{ingreekandcoptic}', "");
Expect(0, 1023, '\p{^ingreekandcoptic}', "");
Expect(0, 1023, '\P{ingreekandcoptic}', "");
Expect(1, 1023, '\P{^ingreekandcoptic}', "");
Expect(0, 1024, '\p{ingreekandcoptic}', "");
Expect(1, 1024, '\p{^ingreekandcoptic}', "");
Expect(1, 1024, '\P{ingreekandcoptic}', "");
Expect(0, 1024, '\P{^ingreekandcoptic}', "");
Expect(1, 1023, '\p{	 IN_Greek_And_COPTIC}', "");
Expect(0, 1023, '\p{^	 IN_Greek_And_COPTIC}', "");
Expect(0, 1023, '\P{	 IN_Greek_And_COPTIC}', "");
Expect(1, 1023, '\P{^	 IN_Greek_And_COPTIC}', "");
Expect(0, 1024, '\p{	 IN_Greek_And_COPTIC}', "");
Expect(1, 1024, '\p{^	 IN_Greek_And_COPTIC}', "");
Expect(1, 1024, '\P{	 IN_Greek_And_COPTIC}', "");
Expect(0, 1024, '\P{^	 IN_Greek_And_COPTIC}', "");
Error('\p{_	In_greek/a/}');
Error('\P{_	In_greek/a/}');
Expect(1, 1023, '\p{ingreek}', "");
Expect(0, 1023, '\p{^ingreek}', "");
Expect(0, 1023, '\P{ingreek}', "");
Expect(1, 1023, '\P{^ingreek}', "");
Expect(0, 1024, '\p{ingreek}', "");
Expect(1, 1024, '\p{^ingreek}', "");
Expect(1, 1024, '\P{ingreek}', "");
Expect(0, 1024, '\P{^ingreek}', "");
Expect(1, 1023, '\p{ 	in_Greek}', "");
Expect(0, 1023, '\p{^ 	in_Greek}', "");
Expect(0, 1023, '\P{ 	in_Greek}', "");
Expect(1, 1023, '\P{^ 	in_Greek}', "");
Expect(0, 1024, '\p{ 	in_Greek}', "");
Expect(1, 1024, '\p{^ 	in_Greek}', "");
Expect(1, 1024, '\P{ 	in_Greek}', "");
Expect(0, 1024, '\P{^ 	in_Greek}', "");
Error('\p{/a/GREEK_Extended}');
Error('\P{/a/GREEK_Extended}');
Expect(1, 8191, '\p{greekextended}', "");
Expect(0, 8191, '\p{^greekextended}', "");
Expect(0, 8191, '\P{greekextended}', "");
Expect(1, 8191, '\P{^greekextended}', "");
Expect(0, 8192, '\p{greekextended}', "");
Expect(1, 8192, '\p{^greekextended}', "");
Expect(1, 8192, '\P{greekextended}', "");
Expect(0, 8192, '\P{^greekextended}', "");
Expect(1, 8191, '\p{ _Greek_Extended}', "");
Expect(0, 8191, '\p{^ _Greek_Extended}', "");
Expect(0, 8191, '\P{ _Greek_Extended}', "");
Expect(1, 8191, '\P{^ _Greek_Extended}', "");
Expect(0, 8192, '\p{ _Greek_Extended}', "");
Expect(1, 8192, '\p{^ _Greek_Extended}', "");
Expect(1, 8192, '\P{ _Greek_Extended}', "");
Expect(0, 8192, '\P{^ _Greek_Extended}', "");
Error('\p{-IS_greek_extended:=}');
Error('\P{-IS_greek_extended:=}');
Expect(1, 8191, '\p{isgreekextended}', "");
Expect(0, 8191, '\p{^isgreekextended}', "");
Expect(0, 8191, '\P{isgreekextended}', "");
Expect(1, 8191, '\P{^isgreekextended}', "");
Expect(0, 8192, '\p{isgreekextended}', "");
Expect(1, 8192, '\p{^isgreekextended}', "");
Expect(1, 8192, '\P{isgreekextended}', "");
Expect(0, 8192, '\P{^isgreekextended}', "");
Expect(1, 8191, '\p{--Is_GREEK_EXTENDED}', "");
Expect(0, 8191, '\p{^--Is_GREEK_EXTENDED}', "");
Expect(0, 8191, '\P{--Is_GREEK_EXTENDED}', "");
Expect(1, 8191, '\P{^--Is_GREEK_EXTENDED}', "");
Expect(0, 8192, '\p{--Is_GREEK_EXTENDED}', "");
Expect(1, 8192, '\p{^--Is_GREEK_EXTENDED}', "");
Expect(1, 8192, '\P{--Is_GREEK_EXTENDED}', "");
Expect(0, 8192, '\P{^--Is_GREEK_EXTENDED}', "");
Error('\p{-:=IN_Greek_extended}');
Error('\P{-:=IN_Greek_extended}');
Expect(1, 8191, '\p{ingreekextended}', "");
Expect(0, 8191, '\p{^ingreekextended}', "");
Expect(0, 8191, '\P{ingreekextended}', "");
Expect(1, 8191, '\P{^ingreekextended}', "");
Expect(0, 8192, '\p{ingreekextended}', "");
Expect(1, 8192, '\p{^ingreekextended}', "");
Expect(1, 8192, '\P{ingreekextended}', "");
Expect(0, 8192, '\P{^ingreekextended}', "");
Expect(1, 8191, '\p{	_In_GREEK_extended}', "");
Expect(0, 8191, '\p{^	_In_GREEK_extended}', "");
Expect(0, 8191, '\P{	_In_GREEK_extended}', "");
Expect(1, 8191, '\P{^	_In_GREEK_extended}', "");
Expect(0, 8192, '\p{	_In_GREEK_extended}', "");
Expect(1, 8192, '\p{^	_In_GREEK_extended}', "");
Expect(1, 8192, '\P{	_In_GREEK_extended}', "");
Expect(0, 8192, '\P{^	_In_GREEK_extended}', "");
Error('\p{/a/	 Greek_Ext}');
Error('\P{/a/	 Greek_Ext}');
Expect(1, 8191, '\p{greekext}', "");
Expect(0, 8191, '\p{^greekext}', "");
Expect(0, 8191, '\P{greekext}', "");
Expect(1, 8191, '\P{^greekext}', "");
Expect(0, 8192, '\p{greekext}', "");
Expect(1, 8192, '\p{^greekext}', "");
Expect(1, 8192, '\P{greekext}', "");
Expect(0, 8192, '\P{^greekext}', "");
Expect(1, 8191, '\p{-	GREEK_Ext}', "");
Expect(0, 8191, '\p{^-	GREEK_Ext}', "");
Expect(0, 8191, '\P{-	GREEK_Ext}', "");
Expect(1, 8191, '\P{^-	GREEK_Ext}', "");
Expect(0, 8192, '\p{-	GREEK_Ext}', "");
Expect(1, 8192, '\p{^-	GREEK_Ext}', "");
Expect(1, 8192, '\P{-	GREEK_Ext}', "");
Expect(0, 8192, '\P{^-	GREEK_Ext}', "");
Error('\p{_	is_GREEK_Ext/a/}');
Error('\P{_	is_GREEK_Ext/a/}');
Expect(1, 8191, '\p{isgreekext}', "");
Expect(0, 8191, '\p{^isgreekext}', "");
Expect(0, 8191, '\P{isgreekext}', "");
Expect(1, 8191, '\P{^isgreekext}', "");
Expect(0, 8192, '\p{isgreekext}', "");
Expect(1, 8192, '\p{^isgreekext}', "");
Expect(1, 8192, '\P{isgreekext}', "");
Expect(0, 8192, '\P{^isgreekext}', "");
Expect(1, 8191, '\p{	Is_GREEK_Ext}', "");
Expect(0, 8191, '\p{^	Is_GREEK_Ext}', "");
Expect(0, 8191, '\P{	Is_GREEK_Ext}', "");
Expect(1, 8191, '\P{^	Is_GREEK_Ext}', "");
Expect(0, 8192, '\p{	Is_GREEK_Ext}', "");
Expect(1, 8192, '\p{^	Is_GREEK_Ext}', "");
Expect(1, 8192, '\P{	Is_GREEK_Ext}', "");
Expect(0, 8192, '\P{^	Is_GREEK_Ext}', "");
Error('\p{_:=IN_Greek_ext}');
Error('\P{_:=IN_Greek_ext}');
Expect(1, 8191, '\p{ingreekext}', "");
Expect(0, 8191, '\p{^ingreekext}', "");
Expect(0, 8191, '\P{ingreekext}', "");
Expect(1, 8191, '\P{^ingreekext}', "");
Expect(0, 8192, '\p{ingreekext}', "");
Expect(1, 8192, '\p{^ingreekext}', "");
Expect(1, 8192, '\P{ingreekext}', "");
Expect(0, 8192, '\P{^ingreekext}', "");
Expect(1, 8191, '\p{		in_Greek_EXT}', "");
Expect(0, 8191, '\p{^		in_Greek_EXT}', "");
Expect(0, 8191, '\P{		in_Greek_EXT}', "");
Expect(1, 8191, '\P{^		in_Greek_EXT}', "");
Expect(0, 8192, '\p{		in_Greek_EXT}', "");
Expect(1, 8192, '\p{^		in_Greek_EXT}', "");
Expect(1, 8192, '\P{		in_Greek_EXT}', "");
Expect(0, 8192, '\P{^		in_Greek_EXT}', "");
Error('\p{/a/-Gujarati}');
Error('\P{/a/-Gujarati}');
Expect(1, 2801, '\p{gujarati}', "");
Expect(0, 2801, '\p{^gujarati}', "");
Expect(0, 2801, '\P{gujarati}', "");
Expect(1, 2801, '\P{^gujarati}', "");
Expect(0, 2802, '\p{gujarati}', "");
Expect(1, 2802, '\p{^gujarati}', "");
Expect(1, 2802, '\P{gujarati}', "");
Expect(0, 2802, '\P{^gujarati}', "");
Expect(1, 2801, '\p{_ GUJARATI}', "");
Expect(0, 2801, '\p{^_ GUJARATI}', "");
Expect(0, 2801, '\P{_ GUJARATI}', "");
Expect(1, 2801, '\P{^_ GUJARATI}', "");
Expect(0, 2802, '\p{_ GUJARATI}', "");
Expect(1, 2802, '\p{^_ GUJARATI}', "");
Expect(1, 2802, '\P{_ GUJARATI}', "");
Expect(0, 2802, '\P{^_ GUJARATI}', "");
Error('\p{/a/_Is_GUJARATI}');
Error('\P{/a/_Is_GUJARATI}');
Expect(1, 2801, '\p{isgujarati}', "");
Expect(0, 2801, '\p{^isgujarati}', "");
Expect(0, 2801, '\P{isgujarati}', "");
Expect(1, 2801, '\P{^isgujarati}', "");
Expect(0, 2802, '\p{isgujarati}', "");
Expect(1, 2802, '\p{^isgujarati}', "");
Expect(1, 2802, '\P{isgujarati}', "");
Expect(0, 2802, '\P{^isgujarati}', "");
Expect(1, 2801, '\p{__Is_Gujarati}', "");
Expect(0, 2801, '\p{^__Is_Gujarati}', "");
Expect(0, 2801, '\P{__Is_Gujarati}', "");
Expect(1, 2801, '\P{^__Is_Gujarati}', "");
Expect(0, 2802, '\p{__Is_Gujarati}', "");
Expect(1, 2802, '\p{^__Is_Gujarati}', "");
Expect(1, 2802, '\P{__Is_Gujarati}', "");
Expect(0, 2802, '\P{^__Is_Gujarati}', "");
Error('\p{:=__GUJR}');
Error('\P{:=__GUJR}');
Expect(1, 2801, '\p{gujr}', "");
Expect(0, 2801, '\p{^gujr}', "");
Expect(0, 2801, '\P{gujr}', "");
Expect(1, 2801, '\P{^gujr}', "");
Expect(0, 2802, '\p{gujr}', "");
Expect(1, 2802, '\p{^gujr}', "");
Expect(1, 2802, '\P{gujr}', "");
Expect(0, 2802, '\P{^gujr}', "");
Expect(1, 2801, '\p{	_Gujr}', "");
Expect(0, 2801, '\p{^	_Gujr}', "");
Expect(0, 2801, '\P{	_Gujr}', "");
Expect(1, 2801, '\P{^	_Gujr}', "");
Expect(0, 2802, '\p{	_Gujr}', "");
Expect(1, 2802, '\p{^	_Gujr}', "");
Expect(1, 2802, '\P{	_Gujr}', "");
Expect(0, 2802, '\P{^	_Gujr}', "");
Error('\p{/a/-_Is_Gujr}');
Error('\P{/a/-_Is_Gujr}');
Expect(1, 2801, '\p{isgujr}', "");
Expect(0, 2801, '\p{^isgujr}', "");
Expect(0, 2801, '\P{isgujr}', "");
Expect(1, 2801, '\P{^isgujr}', "");
Expect(0, 2802, '\p{isgujr}', "");
Expect(1, 2802, '\p{^isgujr}', "");
Expect(1, 2802, '\P{isgujr}', "");
Expect(0, 2802, '\P{^isgujr}', "");
Expect(1, 2801, '\p{_	Is_GUJR}', "");
Expect(0, 2801, '\p{^_	Is_GUJR}', "");
Expect(0, 2801, '\P{_	Is_GUJR}', "");
Expect(1, 2801, '\P{^_	Is_GUJR}', "");
Expect(0, 2802, '\p{_	Is_GUJR}', "");
Expect(1, 2802, '\p{^_	Is_GUJR}', "");
Expect(1, 2802, '\P{_	Is_GUJR}', "");
Expect(0, 2802, '\P{^_	Is_GUJR}', "");
Error('\p{:=_Gurmukhi}');
Error('\P{:=_Gurmukhi}');
Expect(1, 2677, '\p{gurmukhi}', "");
Expect(0, 2677, '\p{^gurmukhi}', "");
Expect(0, 2677, '\P{gurmukhi}', "");
Expect(1, 2677, '\P{^gurmukhi}', "");
Expect(0, 2678, '\p{gurmukhi}', "");
Expect(1, 2678, '\p{^gurmukhi}', "");
Expect(1, 2678, '\P{gurmukhi}', "");
Expect(0, 2678, '\P{^gurmukhi}', "");
Expect(1, 2677, '\p{ Gurmukhi}', "");
Expect(0, 2677, '\p{^ Gurmukhi}', "");
Expect(0, 2677, '\P{ Gurmukhi}', "");
Expect(1, 2677, '\P{^ Gurmukhi}', "");
Expect(0, 2678, '\p{ Gurmukhi}', "");
Expect(1, 2678, '\p{^ Gurmukhi}', "");
Expect(1, 2678, '\P{ Gurmukhi}', "");
Expect(0, 2678, '\P{^ Gurmukhi}', "");
Error('\p{ is_gurmukhi/a/}');
Error('\P{ is_gurmukhi/a/}');
Expect(1, 2677, '\p{isgurmukhi}', "");
Expect(0, 2677, '\p{^isgurmukhi}', "");
Expect(0, 2677, '\P{isgurmukhi}', "");
Expect(1, 2677, '\P{^isgurmukhi}', "");
Expect(0, 2678, '\p{isgurmukhi}', "");
Expect(1, 2678, '\p{^isgurmukhi}', "");
Expect(1, 2678, '\P{isgurmukhi}', "");
Expect(0, 2678, '\P{^isgurmukhi}', "");
Expect(1, 2677, '\p{-Is_Gurmukhi}', "");
Expect(0, 2677, '\p{^-Is_Gurmukhi}', "");
Expect(0, 2677, '\P{-Is_Gurmukhi}', "");
Expect(1, 2677, '\P{^-Is_Gurmukhi}', "");
Expect(0, 2678, '\p{-Is_Gurmukhi}', "");
Expect(1, 2678, '\p{^-Is_Gurmukhi}', "");
Expect(1, 2678, '\P{-Is_Gurmukhi}', "");
Expect(0, 2678, '\P{^-Is_Gurmukhi}', "");
Error('\p{:=_	guru}');
Error('\P{:=_	guru}');
Expect(1, 2677, '\p{guru}', "");
Expect(0, 2677, '\p{^guru}', "");
Expect(0, 2677, '\P{guru}', "");
Expect(1, 2677, '\P{^guru}', "");
Expect(0, 2678, '\p{guru}', "");
Expect(1, 2678, '\p{^guru}', "");
Expect(1, 2678, '\P{guru}', "");
Expect(0, 2678, '\P{^guru}', "");
Expect(1, 2677, '\p{-_guru}', "");
Expect(0, 2677, '\p{^-_guru}', "");
Expect(0, 2677, '\P{-_guru}', "");
Expect(1, 2677, '\P{^-_guru}', "");
Expect(0, 2678, '\p{-_guru}', "");
Expect(1, 2678, '\p{^-_guru}', "");
Expect(1, 2678, '\P{-_guru}', "");
Expect(0, 2678, '\P{^-_guru}', "");
Error('\p{/a/IS_Guru}');
Error('\P{/a/IS_Guru}');
Expect(1, 2677, '\p{isguru}', "");
Expect(0, 2677, '\p{^isguru}', "");
Expect(0, 2677, '\P{isguru}', "");
Expect(1, 2677, '\P{^isguru}', "");
Expect(0, 2678, '\p{isguru}', "");
Expect(1, 2678, '\p{^isguru}', "");
Expect(1, 2678, '\P{isguru}', "");
Expect(0, 2678, '\P{^isguru}', "");
Expect(1, 2677, '\p{_-Is_Guru}', "");
Expect(0, 2677, '\p{^_-Is_Guru}', "");
Expect(0, 2677, '\P{_-Is_Guru}', "");
Expect(1, 2677, '\P{^_-Is_Guru}', "");
Expect(0, 2678, '\p{_-Is_Guru}', "");
Expect(1, 2678, '\p{^_-Is_Guru}', "");
Expect(1, 2678, '\P{_-Is_Guru}', "");
Expect(0, 2678, '\P{^_-Is_Guru}', "");
Error('\p{	 halfwidth_And_fullwidth_forms:=}');
Error('\P{	 halfwidth_And_fullwidth_forms:=}');
Expect(1, 65519, '\p{halfwidthandfullwidthforms}', "");
Expect(0, 65519, '\p{^halfwidthandfullwidthforms}', "");
Expect(0, 65519, '\P{halfwidthandfullwidthforms}', "");
Expect(1, 65519, '\P{^halfwidthandfullwidthforms}', "");
Expect(0, 65520, '\p{halfwidthandfullwidthforms}', "");
Expect(1, 65520, '\p{^halfwidthandfullwidthforms}', "");
Expect(1, 65520, '\P{halfwidthandfullwidthforms}', "");
Expect(0, 65520, '\P{^halfwidthandfullwidthforms}', "");
Expect(1, 65519, '\p{ HALFWIDTH_And_Fullwidth_forms}', "");
Expect(0, 65519, '\p{^ HALFWIDTH_And_Fullwidth_forms}', "");
Expect(0, 65519, '\P{ HALFWIDTH_And_Fullwidth_forms}', "");
Expect(1, 65519, '\P{^ HALFWIDTH_And_Fullwidth_forms}', "");
Expect(0, 65520, '\p{ HALFWIDTH_And_Fullwidth_forms}', "");
Expect(1, 65520, '\p{^ HALFWIDTH_And_Fullwidth_forms}', "");
Expect(1, 65520, '\P{ HALFWIDTH_And_Fullwidth_forms}', "");
Expect(0, 65520, '\P{^ HALFWIDTH_And_Fullwidth_forms}', "");
Error('\p{__Is_halfwidth_and_Fullwidth_Forms:=}');
Error('\P{__Is_halfwidth_and_Fullwidth_Forms:=}');
Expect(1, 65519, '\p{ishalfwidthandfullwidthforms}', "");
Expect(0, 65519, '\p{^ishalfwidthandfullwidthforms}', "");
Expect(0, 65519, '\P{ishalfwidthandfullwidthforms}', "");
Expect(1, 65519, '\P{^ishalfwidthandfullwidthforms}', "");
Expect(0, 65520, '\p{ishalfwidthandfullwidthforms}', "");
Expect(1, 65520, '\p{^ishalfwidthandfullwidthforms}', "");
Expect(1, 65520, '\P{ishalfwidthandfullwidthforms}', "");
Expect(0, 65520, '\P{^ishalfwidthandfullwidthforms}', "");
Expect(1, 65519, '\p{ -IS_Halfwidth_and_Fullwidth_Forms}', "");
Expect(0, 65519, '\p{^ -IS_Halfwidth_and_Fullwidth_Forms}', "");
Expect(0, 65519, '\P{ -IS_Halfwidth_and_Fullwidth_Forms}', "");
Expect(1, 65519, '\P{^ -IS_Halfwidth_and_Fullwidth_Forms}', "");
Expect(0, 65520, '\p{ -IS_Halfwidth_and_Fullwidth_Forms}', "");
Expect(1, 65520, '\p{^ -IS_Halfwidth_and_Fullwidth_Forms}', "");
Expect(1, 65520, '\P{ -IS_Halfwidth_and_Fullwidth_Forms}', "");
Expect(0, 65520, '\P{^ -IS_Halfwidth_and_Fullwidth_Forms}', "");
Error('\p{	:=In_Halfwidth_And_Fullwidth_Forms}');
Error('\P{	:=In_Halfwidth_And_Fullwidth_Forms}');
Expect(1, 65519, '\p{inhalfwidthandfullwidthforms}', "");
Expect(0, 65519, '\p{^inhalfwidthandfullwidthforms}', "");
Expect(0, 65519, '\P{inhalfwidthandfullwidthforms}', "");
Expect(1, 65519, '\P{^inhalfwidthandfullwidthforms}', "");
Expect(0, 65520, '\p{inhalfwidthandfullwidthforms}', "");
Expect(1, 65520, '\p{^inhalfwidthandfullwidthforms}', "");
Expect(1, 65520, '\P{inhalfwidthandfullwidthforms}', "");
Expect(0, 65520, '\P{^inhalfwidthandfullwidthforms}', "");
Expect(1, 65519, '\p{-	In_Halfwidth_And_Fullwidth_FORMS}', "");
Expect(0, 65519, '\p{^-	In_Halfwidth_And_Fullwidth_FORMS}', "");
Expect(0, 65519, '\P{-	In_Halfwidth_And_Fullwidth_FORMS}', "");
Expect(1, 65519, '\P{^-	In_Halfwidth_And_Fullwidth_FORMS}', "");
Expect(0, 65520, '\p{-	In_Halfwidth_And_Fullwidth_FORMS}', "");
Expect(1, 65520, '\p{^-	In_Halfwidth_And_Fullwidth_FORMS}', "");
Expect(1, 65520, '\P{-	In_Halfwidth_And_Fullwidth_FORMS}', "");
Expect(0, 65520, '\P{^-	In_Halfwidth_And_Fullwidth_FORMS}', "");
Error('\p{	/a/Half_and_full_Forms}');
Error('\P{	/a/Half_and_full_Forms}');
Expect(1, 65519, '\p{halfandfullforms}', "");
Expect(0, 65519, '\p{^halfandfullforms}', "");
Expect(0, 65519, '\P{halfandfullforms}', "");
Expect(1, 65519, '\P{^halfandfullforms}', "");
Expect(0, 65520, '\p{halfandfullforms}', "");
Expect(1, 65520, '\p{^halfandfullforms}', "");
Expect(1, 65520, '\P{halfandfullforms}', "");
Expect(0, 65520, '\P{^halfandfullforms}', "");
Expect(1, 65519, '\p{_	HALF_AND_Full_FORMS}', "");
Expect(0, 65519, '\p{^_	HALF_AND_Full_FORMS}', "");
Expect(0, 65519, '\P{_	HALF_AND_Full_FORMS}', "");
Expect(1, 65519, '\P{^_	HALF_AND_Full_FORMS}', "");
Expect(0, 65520, '\p{_	HALF_AND_Full_FORMS}', "");
Expect(1, 65520, '\p{^_	HALF_AND_Full_FORMS}', "");
Expect(1, 65520, '\P{_	HALF_AND_Full_FORMS}', "");
Expect(0, 65520, '\P{^_	HALF_AND_Full_FORMS}', "");
Error('\p{	/a/Is_half_And_Full_forms}');
Error('\P{	/a/Is_half_And_Full_forms}');
Expect(1, 65519, '\p{ishalfandfullforms}', "");
Expect(0, 65519, '\p{^ishalfandfullforms}', "");
Expect(0, 65519, '\P{ishalfandfullforms}', "");
Expect(1, 65519, '\P{^ishalfandfullforms}', "");
Expect(0, 65520, '\p{ishalfandfullforms}', "");
Expect(1, 65520, '\p{^ishalfandfullforms}', "");
Expect(1, 65520, '\P{ishalfandfullforms}', "");
Expect(0, 65520, '\P{^ishalfandfullforms}', "");
Expect(1, 65519, '\p{ -IS_HALF_And_Full_Forms}', "");
Expect(0, 65519, '\p{^ -IS_HALF_And_Full_Forms}', "");
Expect(0, 65519, '\P{ -IS_HALF_And_Full_Forms}', "");
Expect(1, 65519, '\P{^ -IS_HALF_And_Full_Forms}', "");
Expect(0, 65520, '\p{ -IS_HALF_And_Full_Forms}', "");
Expect(1, 65520, '\p{^ -IS_HALF_And_Full_Forms}', "");
Expect(1, 65520, '\P{ -IS_HALF_And_Full_Forms}', "");
Expect(0, 65520, '\P{^ -IS_HALF_And_Full_Forms}', "");
Error('\p{	_In_HALF_And_Full_forms/a/}');
Error('\P{	_In_HALF_And_Full_forms/a/}');
Expect(1, 65519, '\p{inhalfandfullforms}', "");
Expect(0, 65519, '\p{^inhalfandfullforms}', "");
Expect(0, 65519, '\P{inhalfandfullforms}', "");
Expect(1, 65519, '\P{^inhalfandfullforms}', "");
Expect(0, 65520, '\p{inhalfandfullforms}', "");
Expect(1, 65520, '\p{^inhalfandfullforms}', "");
Expect(1, 65520, '\P{inhalfandfullforms}', "");
Expect(0, 65520, '\P{^inhalfandfullforms}', "");
Expect(1, 65519, '\p{in_Half_And_Full_Forms}', "");
Expect(0, 65519, '\p{^in_Half_And_Full_Forms}', "");
Expect(0, 65519, '\P{in_Half_And_Full_Forms}', "");
Expect(1, 65519, '\P{^in_Half_And_Full_Forms}', "");
Expect(0, 65520, '\p{in_Half_And_Full_Forms}', "");
Expect(1, 65520, '\p{^in_Half_And_Full_Forms}', "");
Expect(1, 65520, '\P{in_Half_And_Full_Forms}', "");
Expect(0, 65520, '\P{^in_Half_And_Full_Forms}', "");
Error('\p{:=	HAN}');
Error('\P{:=	HAN}');
Expect(1, 195101, '\p{han}', "");
Expect(0, 195101, '\p{^han}', "");
Expect(0, 195101, '\P{han}', "");
Expect(1, 195101, '\P{^han}', "");
Expect(0, 195102, '\p{han}', "");
Expect(1, 195102, '\p{^han}', "");
Expect(1, 195102, '\P{han}', "");
Expect(0, 195102, '\P{^han}', "");
Expect(1, 195101, '\p{_-Han}', "");
Expect(0, 195101, '\p{^_-Han}', "");
Expect(0, 195101, '\P{_-Han}', "");
Expect(1, 195101, '\P{^_-Han}', "");
Expect(0, 195102, '\p{_-Han}', "");
Expect(1, 195102, '\p{^_-Han}', "");
Expect(1, 195102, '\P{_-Han}', "");
Expect(0, 195102, '\P{^_-Han}', "");
Error('\p{-/a/IS_Han}');
Error('\P{-/a/IS_Han}');
Expect(1, 195101, '\p{ishan}', "");
Expect(0, 195101, '\p{^ishan}', "");
Expect(0, 195101, '\P{ishan}', "");
Expect(1, 195101, '\P{^ishan}', "");
Expect(0, 195102, '\p{ishan}', "");
Expect(1, 195102, '\p{^ishan}', "");
Expect(1, 195102, '\P{ishan}', "");
Expect(0, 195102, '\P{^ishan}', "");
Expect(1, 195101, '\p{-	Is_han}', "");
Expect(0, 195101, '\p{^-	Is_han}', "");
Expect(0, 195101, '\P{-	Is_han}', "");
Expect(1, 195101, '\P{^-	Is_han}', "");
Expect(0, 195102, '\p{-	Is_han}', "");
Expect(1, 195102, '\p{^-	Is_han}', "");
Expect(1, 195102, '\P{-	Is_han}', "");
Expect(0, 195102, '\P{^-	Is_han}', "");
Error('\p{_-HANI:=}');
Error('\P{_-HANI:=}');
Expect(1, 195101, '\p{hani}', "");
Expect(0, 195101, '\p{^hani}', "");
Expect(0, 195101, '\P{hani}', "");
Expect(1, 195101, '\P{^hani}', "");
Expect(0, 195102, '\p{hani}', "");
Expect(1, 195102, '\p{^hani}', "");
Expect(1, 195102, '\P{hani}', "");
Expect(0, 195102, '\P{^hani}', "");
Expect(1, 195101, '\p{	 Hani}', "");
Expect(0, 195101, '\p{^	 Hani}', "");
Expect(0, 195101, '\P{	 Hani}', "");
Expect(1, 195101, '\P{^	 Hani}', "");
Expect(0, 195102, '\p{	 Hani}', "");
Expect(1, 195102, '\p{^	 Hani}', "");
Expect(1, 195102, '\P{	 Hani}', "");
Expect(0, 195102, '\P{^	 Hani}', "");
Error('\p{  Is_Hani/a/}');
Error('\P{  Is_Hani/a/}');
Expect(1, 195101, '\p{ishani}', "");
Expect(0, 195101, '\p{^ishani}', "");
Expect(0, 195101, '\P{ishani}', "");
Expect(1, 195101, '\P{^ishani}', "");
Expect(0, 195102, '\p{ishani}', "");
Expect(1, 195102, '\p{^ishani}', "");
Expect(1, 195102, '\P{ishani}', "");
Expect(0, 195102, '\P{^ishani}', "");
Expect(1, 195101, '\p{_-Is_Hani}', "");
Expect(0, 195101, '\p{^_-Is_Hani}', "");
Expect(0, 195101, '\P{_-Is_Hani}', "");
Expect(1, 195101, '\P{^_-Is_Hani}', "");
Expect(0, 195102, '\p{_-Is_Hani}', "");
Expect(1, 195102, '\p{^_-Is_Hani}', "");
Expect(1, 195102, '\P{_-Is_Hani}', "");
Expect(0, 195102, '\P{^_-Is_Hani}', "");
Error('\p{:=--hangul}');
Error('\P{:=--hangul}');
Expect(1, 65500, '\p{hangul}', "");
Expect(0, 65500, '\p{^hangul}', "");
Expect(0, 65500, '\P{hangul}', "");
Expect(1, 65500, '\P{^hangul}', "");
Expect(0, 65501, '\p{hangul}', "");
Expect(1, 65501, '\p{^hangul}', "");
Expect(1, 65501, '\P{hangul}', "");
Expect(0, 65501, '\P{^hangul}', "");
Expect(1, 65500, '\p{_-Hangul}', "");
Expect(0, 65500, '\p{^_-Hangul}', "");
Expect(0, 65500, '\P{_-Hangul}', "");
Expect(1, 65500, '\P{^_-Hangul}', "");
Expect(0, 65501, '\p{_-Hangul}', "");
Expect(1, 65501, '\p{^_-Hangul}', "");
Expect(1, 65501, '\P{_-Hangul}', "");
Expect(0, 65501, '\P{^_-Hangul}', "");
Error('\p{/a/ -Is_Hangul}');
Error('\P{/a/ -Is_Hangul}');
Expect(1, 65500, '\p{ishangul}', "");
Expect(0, 65500, '\p{^ishangul}', "");
Expect(0, 65500, '\P{ishangul}', "");
Expect(1, 65500, '\P{^ishangul}', "");
Expect(0, 65501, '\p{ishangul}', "");
Expect(1, 65501, '\p{^ishangul}', "");
Expect(1, 65501, '\P{ishangul}', "");
Expect(0, 65501, '\P{^ishangul}', "");
Expect(1, 65500, '\p{_	Is_HANGUL}', "");
Expect(0, 65500, '\p{^_	Is_HANGUL}', "");
Expect(0, 65500, '\P{_	Is_HANGUL}', "");
Expect(1, 65500, '\P{^_	Is_HANGUL}', "");
Expect(0, 65501, '\p{_	Is_HANGUL}', "");
Expect(1, 65501, '\p{^_	Is_HANGUL}', "");
Expect(1, 65501, '\P{_	Is_HANGUL}', "");
Expect(0, 65501, '\P{^_	Is_HANGUL}', "");
Error('\p{--hang:=}');
Error('\P{--hang:=}');
Expect(1, 65500, '\p{hang}', "");
Expect(0, 65500, '\p{^hang}', "");
Expect(0, 65500, '\P{hang}', "");
Expect(1, 65500, '\P{^hang}', "");
Expect(0, 65501, '\p{hang}', "");
Expect(1, 65501, '\p{^hang}', "");
Expect(1, 65501, '\P{hang}', "");
Expect(0, 65501, '\P{^hang}', "");
Expect(1, 65500, '\p{Hang}', "");
Expect(0, 65500, '\p{^Hang}', "");
Expect(0, 65500, '\P{Hang}', "");
Expect(1, 65500, '\P{^Hang}', "");
Expect(0, 65501, '\p{Hang}', "");
Expect(1, 65501, '\p{^Hang}', "");
Expect(1, 65501, '\P{Hang}', "");
Expect(0, 65501, '\P{^Hang}', "");
Error('\p{/a/--IS_Hang}');
Error('\P{/a/--IS_Hang}');
Expect(1, 65500, '\p{ishang}', "");
Expect(0, 65500, '\p{^ishang}', "");
Expect(0, 65500, '\P{ishang}', "");
Expect(1, 65500, '\P{^ishang}', "");
Expect(0, 65501, '\p{ishang}', "");
Expect(1, 65501, '\p{^ishang}', "");
Expect(1, 65501, '\P{ishang}', "");
Expect(0, 65501, '\P{^ishang}', "");
Expect(1, 65500, '\p{_-Is_HANG}', "");
Expect(0, 65500, '\p{^_-Is_HANG}', "");
Expect(0, 65500, '\P{_-Is_HANG}', "");
Expect(1, 65500, '\P{^_-Is_HANG}', "");
Expect(0, 65501, '\p{_-Is_HANG}', "");
Expect(1, 65501, '\p{^_-Is_HANG}', "");
Expect(1, 65501, '\P{_-Is_HANG}', "");
Expect(0, 65501, '\P{^_-Is_HANG}', "");
Error('\p{__hangul_COMPATIBILITY_Jamo:=}');
Error('\P{__hangul_COMPATIBILITY_Jamo:=}');
Expect(1, 12687, '\p{hangulcompatibilityjamo}', "");
Expect(0, 12687, '\p{^hangulcompatibilityjamo}', "");
Expect(0, 12687, '\P{hangulcompatibilityjamo}', "");
Expect(1, 12687, '\P{^hangulcompatibilityjamo}', "");
Expect(0, 12688, '\p{hangulcompatibilityjamo}', "");
Expect(1, 12688, '\p{^hangulcompatibilityjamo}', "");
Expect(1, 12688, '\P{hangulcompatibilityjamo}', "");
Expect(0, 12688, '\P{^hangulcompatibilityjamo}', "");
Expect(1, 12687, '\p{_hangul_Compatibility_JAMO}', "");
Expect(0, 12687, '\p{^_hangul_Compatibility_JAMO}', "");
Expect(0, 12687, '\P{_hangul_Compatibility_JAMO}', "");
Expect(1, 12687, '\P{^_hangul_Compatibility_JAMO}', "");
Expect(0, 12688, '\p{_hangul_Compatibility_JAMO}', "");
Expect(1, 12688, '\p{^_hangul_Compatibility_JAMO}', "");
Expect(1, 12688, '\P{_hangul_Compatibility_JAMO}', "");
Expect(0, 12688, '\P{^_hangul_Compatibility_JAMO}', "");
Error('\p{-:=Is_HANGUL_Compatibility_JAMO}');
Error('\P{-:=Is_HANGUL_Compatibility_JAMO}');
Expect(1, 12687, '\p{ishangulcompatibilityjamo}', "");
Expect(0, 12687, '\p{^ishangulcompatibilityjamo}', "");
Expect(0, 12687, '\P{ishangulcompatibilityjamo}', "");
Expect(1, 12687, '\P{^ishangulcompatibilityjamo}', "");
Expect(0, 12688, '\p{ishangulcompatibilityjamo}', "");
Expect(1, 12688, '\p{^ishangulcompatibilityjamo}', "");
Expect(1, 12688, '\P{ishangulcompatibilityjamo}', "");
Expect(0, 12688, '\P{^ishangulcompatibilityjamo}', "");
Expect(1, 12687, '\p{-Is_Hangul_Compatibility_Jamo}', "");
Expect(0, 12687, '\p{^-Is_Hangul_Compatibility_Jamo}', "");
Expect(0, 12687, '\P{-Is_Hangul_Compatibility_Jamo}', "");
Expect(1, 12687, '\P{^-Is_Hangul_Compatibility_Jamo}', "");
Expect(0, 12688, '\p{-Is_Hangul_Compatibility_Jamo}', "");
Expect(1, 12688, '\p{^-Is_Hangul_Compatibility_Jamo}', "");
Expect(1, 12688, '\P{-Is_Hangul_Compatibility_Jamo}', "");
Expect(0, 12688, '\P{^-Is_Hangul_Compatibility_Jamo}', "");
Error('\p{	:=In_Hangul_Compatibility_JAMO}');
Error('\P{	:=In_Hangul_Compatibility_JAMO}');
Expect(1, 12687, '\p{inhangulcompatibilityjamo}', "");
Expect(0, 12687, '\p{^inhangulcompatibilityjamo}', "");
Expect(0, 12687, '\P{inhangulcompatibilityjamo}', "");
Expect(1, 12687, '\P{^inhangulcompatibilityjamo}', "");
Expect(0, 12688, '\p{inhangulcompatibilityjamo}', "");
Expect(1, 12688, '\p{^inhangulcompatibilityjamo}', "");
Expect(1, 12688, '\P{inhangulcompatibilityjamo}', "");
Expect(0, 12688, '\P{^inhangulcompatibilityjamo}', "");
Expect(1, 12687, '\p{ -In_hangul_COMPATIBILITY_JAMO}', "");
Expect(0, 12687, '\p{^ -In_hangul_COMPATIBILITY_JAMO}', "");
Expect(0, 12687, '\P{ -In_hangul_COMPATIBILITY_JAMO}', "");
Expect(1, 12687, '\P{^ -In_hangul_COMPATIBILITY_JAMO}', "");
Expect(0, 12688, '\p{ -In_hangul_COMPATIBILITY_JAMO}', "");
Expect(1, 12688, '\p{^ -In_hangul_COMPATIBILITY_JAMO}', "");
Expect(1, 12688, '\P{ -In_hangul_COMPATIBILITY_JAMO}', "");
Expect(0, 12688, '\P{^ -In_hangul_COMPATIBILITY_JAMO}', "");
Error('\p{:=--Compat_jamo}');
Error('\P{:=--Compat_jamo}');
Expect(1, 12687, '\p{compatjamo}', "");
Expect(0, 12687, '\p{^compatjamo}', "");
Expect(0, 12687, '\P{compatjamo}', "");
Expect(1, 12687, '\P{^compatjamo}', "");
Expect(0, 12688, '\p{compatjamo}', "");
Expect(1, 12688, '\p{^compatjamo}', "");
Expect(1, 12688, '\P{compatjamo}', "");
Expect(0, 12688, '\P{^compatjamo}', "");
Expect(1, 12687, '\p{	-compat_Jamo}', "");
Expect(0, 12687, '\p{^	-compat_Jamo}', "");
Expect(0, 12687, '\P{	-compat_Jamo}', "");
Expect(1, 12687, '\P{^	-compat_Jamo}', "");
Expect(0, 12688, '\p{	-compat_Jamo}', "");
Expect(1, 12688, '\p{^	-compat_Jamo}', "");
Expect(1, 12688, '\P{	-compat_Jamo}', "");
Expect(0, 12688, '\P{^	-compat_Jamo}', "");
Error('\p{--Is_Compat_jamo:=}');
Error('\P{--Is_Compat_jamo:=}');
Expect(1, 12687, '\p{iscompatjamo}', "");
Expect(0, 12687, '\p{^iscompatjamo}', "");
Expect(0, 12687, '\P{iscompatjamo}', "");
Expect(1, 12687, '\P{^iscompatjamo}', "");
Expect(0, 12688, '\p{iscompatjamo}', "");
Expect(1, 12688, '\p{^iscompatjamo}', "");
Expect(1, 12688, '\P{iscompatjamo}', "");
Expect(0, 12688, '\P{^iscompatjamo}', "");
Expect(1, 12687, '\p{_IS_compat_Jamo}', "");
Expect(0, 12687, '\p{^_IS_compat_Jamo}', "");
Expect(0, 12687, '\P{_IS_compat_Jamo}', "");
Expect(1, 12687, '\P{^_IS_compat_Jamo}', "");
Expect(0, 12688, '\p{_IS_compat_Jamo}', "");
Expect(1, 12688, '\p{^_IS_compat_Jamo}', "");
Expect(1, 12688, '\P{_IS_compat_Jamo}', "");
Expect(0, 12688, '\P{^_IS_compat_Jamo}', "");
Error('\p{--In_compat_JAMO:=}');
Error('\P{--In_compat_JAMO:=}');
Expect(1, 12687, '\p{incompatjamo}', "");
Expect(0, 12687, '\p{^incompatjamo}', "");
Expect(0, 12687, '\P{incompatjamo}', "");
Expect(1, 12687, '\P{^incompatjamo}', "");
Expect(0, 12688, '\p{incompatjamo}', "");
Expect(1, 12688, '\p{^incompatjamo}', "");
Expect(1, 12688, '\P{incompatjamo}', "");
Expect(0, 12688, '\P{^incompatjamo}', "");
Expect(1, 12687, '\p{	-in_Compat_Jamo}', "");
Expect(0, 12687, '\p{^	-in_Compat_Jamo}', "");
Expect(0, 12687, '\P{	-in_Compat_Jamo}', "");
Expect(1, 12687, '\P{^	-in_Compat_Jamo}', "");
Expect(0, 12688, '\p{	-in_Compat_Jamo}', "");
Expect(1, 12688, '\p{^	-in_Compat_Jamo}', "");
Expect(1, 12688, '\P{	-in_Compat_Jamo}', "");
Expect(0, 12688, '\P{^	-in_Compat_Jamo}', "");
Error('\p{/a/-	Hangul_jamo}');
Error('\P{/a/-	Hangul_jamo}');
Expect(1, 4607, '\p{hanguljamo}', "");
Expect(0, 4607, '\p{^hanguljamo}', "");
Expect(0, 4607, '\P{hanguljamo}', "");
Expect(1, 4607, '\P{^hanguljamo}', "");
Expect(0, 4608, '\p{hanguljamo}', "");
Expect(1, 4608, '\p{^hanguljamo}', "");
Expect(1, 4608, '\P{hanguljamo}', "");
Expect(0, 4608, '\P{^hanguljamo}', "");
Expect(1, 4607, '\p{_ Hangul_jamo}', "");
Expect(0, 4607, '\p{^_ Hangul_jamo}', "");
Expect(0, 4607, '\P{_ Hangul_jamo}', "");
Expect(1, 4607, '\P{^_ Hangul_jamo}', "");
Expect(0, 4608, '\p{_ Hangul_jamo}', "");
Expect(1, 4608, '\p{^_ Hangul_jamo}', "");
Expect(1, 4608, '\P{_ Hangul_jamo}', "");
Expect(0, 4608, '\P{^_ Hangul_jamo}', "");
Error('\p{Is_hangul_Jamo/a/}');
Error('\P{Is_hangul_Jamo/a/}');
Expect(1, 4607, '\p{ishanguljamo}', "");
Expect(0, 4607, '\p{^ishanguljamo}', "");
Expect(0, 4607, '\P{ishanguljamo}', "");
Expect(1, 4607, '\P{^ishanguljamo}', "");
Expect(0, 4608, '\p{ishanguljamo}', "");
Expect(1, 4608, '\p{^ishanguljamo}', "");
Expect(1, 4608, '\P{ishanguljamo}', "");
Expect(0, 4608, '\P{^ishanguljamo}', "");
Expect(1, 4607, '\p{ 	Is_Hangul_Jamo}', "");
Expect(0, 4607, '\p{^ 	Is_Hangul_Jamo}', "");
Expect(0, 4607, '\P{ 	Is_Hangul_Jamo}', "");
Expect(1, 4607, '\P{^ 	Is_Hangul_Jamo}', "");
Expect(0, 4608, '\p{ 	Is_Hangul_Jamo}', "");
Expect(1, 4608, '\p{^ 	Is_Hangul_Jamo}', "");
Expect(1, 4608, '\P{ 	Is_Hangul_Jamo}', "");
Expect(0, 4608, '\P{^ 	Is_Hangul_Jamo}', "");
Error('\p{_:=In_Hangul_JAMO}');
Error('\P{_:=In_Hangul_JAMO}');
Expect(1, 4607, '\p{inhanguljamo}', "");
Expect(0, 4607, '\p{^inhanguljamo}', "");
Expect(0, 4607, '\P{inhanguljamo}', "");
Expect(1, 4607, '\P{^inhanguljamo}', "");
Expect(0, 4608, '\p{inhanguljamo}', "");
Expect(1, 4608, '\p{^inhanguljamo}', "");
Expect(1, 4608, '\P{inhanguljamo}', "");
Expect(0, 4608, '\P{^inhanguljamo}', "");
Expect(1, 4607, '\p{- IN_HANGUL_jamo}', "");
Expect(0, 4607, '\p{^- IN_HANGUL_jamo}', "");
Expect(0, 4607, '\P{- IN_HANGUL_jamo}', "");
Expect(1, 4607, '\P{^- IN_HANGUL_jamo}', "");
Expect(0, 4608, '\p{- IN_HANGUL_jamo}', "");
Expect(1, 4608, '\p{^- IN_HANGUL_jamo}', "");
Expect(1, 4608, '\P{- IN_HANGUL_jamo}', "");
Expect(0, 4608, '\P{^- IN_HANGUL_jamo}', "");
Error('\p{	:=Jamo}');
Error('\P{	:=Jamo}');
Expect(1, 4607, '\p{jamo}', "");
Expect(0, 4607, '\p{^jamo}', "");
Expect(0, 4607, '\P{jamo}', "");
Expect(1, 4607, '\P{^jamo}', "");
Expect(0, 4608, '\p{jamo}', "");
Expect(1, 4608, '\p{^jamo}', "");
Expect(1, 4608, '\P{jamo}', "");
Expect(0, 4608, '\P{^jamo}', "");
Expect(1, 4607, '\p{	_Jamo}', "");
Expect(0, 4607, '\p{^	_Jamo}', "");
Expect(0, 4607, '\P{	_Jamo}', "");
Expect(1, 4607, '\P{^	_Jamo}', "");
Expect(0, 4608, '\p{	_Jamo}', "");
Expect(1, 4608, '\p{^	_Jamo}', "");
Expect(1, 4608, '\P{	_Jamo}', "");
Expect(0, 4608, '\P{^	_Jamo}', "");
Error('\p{	:=Is_Jamo}');
Error('\P{	:=Is_Jamo}');
Expect(1, 4607, '\p{isjamo}', "");
Expect(0, 4607, '\p{^isjamo}', "");
Expect(0, 4607, '\P{isjamo}', "");
Expect(1, 4607, '\P{^isjamo}', "");
Expect(0, 4608, '\p{isjamo}', "");
Expect(1, 4608, '\p{^isjamo}', "");
Expect(1, 4608, '\P{isjamo}', "");
Expect(0, 4608, '\P{^isjamo}', "");
Expect(1, 4607, '\p{--Is_Jamo}', "");
Expect(0, 4607, '\p{^--Is_Jamo}', "");
Expect(0, 4607, '\P{--Is_Jamo}', "");
Expect(1, 4607, '\P{^--Is_Jamo}', "");
Expect(0, 4608, '\p{--Is_Jamo}', "");
Expect(1, 4608, '\p{^--Is_Jamo}', "");
Expect(1, 4608, '\P{--Is_Jamo}', "");
Expect(0, 4608, '\P{^--Is_Jamo}', "");
Error('\p{_:=in_Jamo}');
Error('\P{_:=in_Jamo}');
Expect(1, 4607, '\p{injamo}', "");
Expect(0, 4607, '\p{^injamo}', "");
Expect(0, 4607, '\P{injamo}', "");
Expect(1, 4607, '\P{^injamo}', "");
Expect(0, 4608, '\p{injamo}', "");
Expect(1, 4608, '\p{^injamo}', "");
Expect(1, 4608, '\P{injamo}', "");
Expect(0, 4608, '\P{^injamo}', "");
Expect(1, 4607, '\p{_In_Jamo}', "");
Expect(0, 4607, '\p{^_In_Jamo}', "");
Expect(0, 4607, '\P{_In_Jamo}', "");
Expect(1, 4607, '\P{^_In_Jamo}', "");
Expect(0, 4608, '\p{_In_Jamo}', "");
Expect(1, 4608, '\p{^_In_Jamo}', "");
Expect(1, 4608, '\P{_In_Jamo}', "");
Expect(0, 4608, '\P{^_In_Jamo}', "");
Error('\p{  Hangul_JAMO_Extended_A/a/}');
Error('\P{  Hangul_JAMO_Extended_A/a/}');
Expect(1, 43391, '\p{hanguljamoextendeda}', "");
Expect(0, 43391, '\p{^hanguljamoextendeda}', "");
Expect(0, 43391, '\P{hanguljamoextendeda}', "");
Expect(1, 43391, '\P{^hanguljamoextendeda}', "");
Expect(0, 43392, '\p{hanguljamoextendeda}', "");
Expect(1, 43392, '\p{^hanguljamoextendeda}', "");
Expect(1, 43392, '\P{hanguljamoextendeda}', "");
Expect(0, 43392, '\P{^hanguljamoextendeda}', "");
Expect(1, 43391, '\p{	Hangul_jamo_Extended_A}', "");
Expect(0, 43391, '\p{^	Hangul_jamo_Extended_A}', "");
Expect(0, 43391, '\P{	Hangul_jamo_Extended_A}', "");
Expect(1, 43391, '\P{^	Hangul_jamo_Extended_A}', "");
Expect(0, 43392, '\p{	Hangul_jamo_Extended_A}', "");
Expect(1, 43392, '\p{^	Hangul_jamo_Extended_A}', "");
Expect(1, 43392, '\P{	Hangul_jamo_Extended_A}', "");
Expect(0, 43392, '\P{^	Hangul_jamo_Extended_A}', "");
Error('\p{	 Is_Hangul_Jamo_extended_a/a/}');
Error('\P{	 Is_Hangul_Jamo_extended_a/a/}');
Expect(1, 43391, '\p{ishanguljamoextendeda}', "");
Expect(0, 43391, '\p{^ishanguljamoextendeda}', "");
Expect(0, 43391, '\P{ishanguljamoextendeda}', "");
Expect(1, 43391, '\P{^ishanguljamoextendeda}', "");
Expect(0, 43392, '\p{ishanguljamoextendeda}', "");
Expect(1, 43392, '\p{^ishanguljamoextendeda}', "");
Expect(1, 43392, '\P{ishanguljamoextendeda}', "");
Expect(0, 43392, '\P{^ishanguljamoextendeda}', "");
Expect(1, 43391, '\p{	 Is_Hangul_Jamo_Extended_A}', "");
Expect(0, 43391, '\p{^	 Is_Hangul_Jamo_Extended_A}', "");
Expect(0, 43391, '\P{	 Is_Hangul_Jamo_Extended_A}', "");
Expect(1, 43391, '\P{^	 Is_Hangul_Jamo_Extended_A}', "");
Expect(0, 43392, '\p{	 Is_Hangul_Jamo_Extended_A}', "");
Expect(1, 43392, '\p{^	 Is_Hangul_Jamo_Extended_A}', "");
Expect(1, 43392, '\P{	 Is_Hangul_Jamo_Extended_A}', "");
Expect(0, 43392, '\P{^	 Is_Hangul_Jamo_Extended_A}', "");
Error('\p{_/a/In_Hangul_Jamo_EXTENDED_A}');
Error('\P{_/a/In_Hangul_Jamo_EXTENDED_A}');
Expect(1, 43391, '\p{inhanguljamoextendeda}', "");
Expect(0, 43391, '\p{^inhanguljamoextendeda}', "");
Expect(0, 43391, '\P{inhanguljamoextendeda}', "");
Expect(1, 43391, '\P{^inhanguljamoextendeda}', "");
Expect(0, 43392, '\p{inhanguljamoextendeda}', "");
Expect(1, 43392, '\p{^inhanguljamoextendeda}', "");
Expect(1, 43392, '\P{inhanguljamoextendeda}', "");
Expect(0, 43392, '\P{^inhanguljamoextendeda}', "");
Expect(1, 43391, '\p{_IN_Hangul_Jamo_extended_A}', "");
Expect(0, 43391, '\p{^_IN_Hangul_Jamo_extended_A}', "");
Expect(0, 43391, '\P{_IN_Hangul_Jamo_extended_A}', "");
Expect(1, 43391, '\P{^_IN_Hangul_Jamo_extended_A}', "");
Expect(0, 43392, '\p{_IN_Hangul_Jamo_extended_A}', "");
Expect(1, 43392, '\p{^_IN_Hangul_Jamo_extended_A}', "");
Expect(1, 43392, '\P{_IN_Hangul_Jamo_extended_A}', "");
Expect(0, 43392, '\P{^_IN_Hangul_Jamo_extended_A}', "");
Error('\p{-:=Jamo_Ext_A}');
Error('\P{-:=Jamo_Ext_A}');
Expect(1, 43391, '\p{jamoexta}', "");
Expect(0, 43391, '\p{^jamoexta}', "");
Expect(0, 43391, '\P{jamoexta}', "");
Expect(1, 43391, '\P{^jamoexta}', "");
Expect(0, 43392, '\p{jamoexta}', "");
Expect(1, 43392, '\p{^jamoexta}', "");
Expect(1, 43392, '\P{jamoexta}', "");
Expect(0, 43392, '\P{^jamoexta}', "");
Expect(1, 43391, '\p{-_jamo_Ext_a}', "");
Expect(0, 43391, '\p{^-_jamo_Ext_a}', "");
Expect(0, 43391, '\P{-_jamo_Ext_a}', "");
Expect(1, 43391, '\P{^-_jamo_Ext_a}', "");
Expect(0, 43392, '\p{-_jamo_Ext_a}', "");
Expect(1, 43392, '\p{^-_jamo_Ext_a}', "");
Expect(1, 43392, '\P{-_jamo_Ext_a}', "");
Expect(0, 43392, '\P{^-_jamo_Ext_a}', "");
Error('\p{ :=Is_JAMO_Ext_a}');
Error('\P{ :=Is_JAMO_Ext_a}');
Expect(1, 43391, '\p{isjamoexta}', "");
Expect(0, 43391, '\p{^isjamoexta}', "");
Expect(0, 43391, '\P{isjamoexta}', "");
Expect(1, 43391, '\P{^isjamoexta}', "");
Expect(0, 43392, '\p{isjamoexta}', "");
Expect(1, 43392, '\p{^isjamoexta}', "");
Expect(1, 43392, '\P{isjamoexta}', "");
Expect(0, 43392, '\P{^isjamoexta}', "");
Expect(1, 43391, '\p{	 IS_JAMO_ext_A}', "");
Expect(0, 43391, '\p{^	 IS_JAMO_ext_A}', "");
Expect(0, 43391, '\P{	 IS_JAMO_ext_A}', "");
Expect(1, 43391, '\P{^	 IS_JAMO_ext_A}', "");
Expect(0, 43392, '\p{	 IS_JAMO_ext_A}', "");
Expect(1, 43392, '\p{^	 IS_JAMO_ext_A}', "");
Expect(1, 43392, '\P{	 IS_JAMO_ext_A}', "");
Expect(0, 43392, '\P{^	 IS_JAMO_ext_A}', "");
Error('\p{_In_Jamo_EXT_A/a/}');
Error('\P{_In_Jamo_EXT_A/a/}');
Expect(1, 43391, '\p{injamoexta}', "");
Expect(0, 43391, '\p{^injamoexta}', "");
Expect(0, 43391, '\P{injamoexta}', "");
Expect(1, 43391, '\P{^injamoexta}', "");
Expect(0, 43392, '\p{injamoexta}', "");
Expect(1, 43392, '\p{^injamoexta}', "");
Expect(1, 43392, '\P{injamoexta}', "");
Expect(0, 43392, '\P{^injamoexta}', "");
Expect(1, 43391, '\p{-	In_jamo_Ext_A}', "");
Expect(0, 43391, '\p{^-	In_jamo_Ext_A}', "");
Expect(0, 43391, '\P{-	In_jamo_Ext_A}', "");
Expect(1, 43391, '\P{^-	In_jamo_Ext_A}', "");
Expect(0, 43392, '\p{-	In_jamo_Ext_A}', "");
Expect(1, 43392, '\p{^-	In_jamo_Ext_A}', "");
Expect(1, 43392, '\P{-	In_jamo_Ext_A}', "");
Expect(0, 43392, '\P{^-	In_jamo_Ext_A}', "");
Error('\p{/a/ _HANGUL_JAMO_extended_b}');
Error('\P{/a/ _HANGUL_JAMO_extended_b}');
Expect(1, 55295, '\p{hanguljamoextendedb}', "");
Expect(0, 55295, '\p{^hanguljamoextendedb}', "");
Expect(0, 55295, '\P{hanguljamoextendedb}', "");
Expect(1, 55295, '\P{^hanguljamoextendedb}', "");
Expect(0, 57344, '\p{hanguljamoextendedb}', "");
Expect(1, 57344, '\p{^hanguljamoextendedb}', "");
Expect(1, 57344, '\P{hanguljamoextendedb}', "");
Expect(0, 57344, '\P{^hanguljamoextendedb}', "");
Expect(1, 55295, '\p{_hangul_JAMO_extended_B}', "");
Expect(0, 55295, '\p{^_hangul_JAMO_extended_B}', "");
Expect(0, 55295, '\P{_hangul_JAMO_extended_B}', "");
Expect(1, 55295, '\P{^_hangul_JAMO_extended_B}', "");
Expect(0, 57344, '\p{_hangul_JAMO_extended_B}', "");
Expect(1, 57344, '\p{^_hangul_JAMO_extended_B}', "");
Expect(1, 57344, '\P{_hangul_JAMO_extended_B}', "");
Expect(0, 57344, '\P{^_hangul_JAMO_extended_B}', "");
Error('\p{ _is_Hangul_Jamo_Extended_b/a/}');
Error('\P{ _is_Hangul_Jamo_Extended_b/a/}');
Expect(1, 55295, '\p{ishanguljamoextendedb}', "");
Expect(0, 55295, '\p{^ishanguljamoextendedb}', "");
Expect(0, 55295, '\P{ishanguljamoextendedb}', "");
Expect(1, 55295, '\P{^ishanguljamoextendedb}', "");
Expect(0, 57344, '\p{ishanguljamoextendedb}', "");
Expect(1, 57344, '\p{^ishanguljamoextendedb}', "");
Expect(1, 57344, '\P{ishanguljamoextendedb}', "");
Expect(0, 57344, '\P{^ishanguljamoextendedb}', "");
Expect(1, 55295, '\p{-	is_HANGUL_Jamo_Extended_B}', "");
Expect(0, 55295, '\p{^-	is_HANGUL_Jamo_Extended_B}', "");
Expect(0, 55295, '\P{-	is_HANGUL_Jamo_Extended_B}', "");
Expect(1, 55295, '\P{^-	is_HANGUL_Jamo_Extended_B}', "");
Expect(0, 57344, '\p{-	is_HANGUL_Jamo_Extended_B}', "");
Expect(1, 57344, '\p{^-	is_HANGUL_Jamo_Extended_B}', "");
Expect(1, 57344, '\P{-	is_HANGUL_Jamo_Extended_B}', "");
Expect(0, 57344, '\P{^-	is_HANGUL_Jamo_Extended_B}', "");
Error('\p{/a/IN_Hangul_JAMO_Extended_B}');
Error('\P{/a/IN_Hangul_JAMO_Extended_B}');
Expect(1, 55295, '\p{inhanguljamoextendedb}', "");
Expect(0, 55295, '\p{^inhanguljamoextendedb}', "");
Expect(0, 55295, '\P{inhanguljamoextendedb}', "");
Expect(1, 55295, '\P{^inhanguljamoextendedb}', "");
Expect(0, 57344, '\p{inhanguljamoextendedb}', "");
Expect(1, 57344, '\p{^inhanguljamoextendedb}', "");
Expect(1, 57344, '\P{inhanguljamoextendedb}', "");
Expect(0, 57344, '\P{^inhanguljamoextendedb}', "");
Expect(1, 55295, '\p{ -In_Hangul_JAMO_Extended_b}', "");
Expect(0, 55295, '\p{^ -In_Hangul_JAMO_Extended_b}', "");
Expect(0, 55295, '\P{ -In_Hangul_JAMO_Extended_b}', "");
Expect(1, 55295, '\P{^ -In_Hangul_JAMO_Extended_b}', "");
Expect(0, 57344, '\p{ -In_Hangul_JAMO_Extended_b}', "");
Expect(1, 57344, '\p{^ -In_Hangul_JAMO_Extended_b}', "");
Expect(1, 57344, '\P{ -In_Hangul_JAMO_Extended_b}', "");
Expect(0, 57344, '\P{^ -In_Hangul_JAMO_Extended_b}', "");
Error('\p{ JAMO_Ext_B:=}');
Error('\P{ JAMO_Ext_B:=}');
Expect(1, 55295, '\p{jamoextb}', "");
Expect(0, 55295, '\p{^jamoextb}', "");
Expect(0, 55295, '\P{jamoextb}', "");
Expect(1, 55295, '\P{^jamoextb}', "");
Expect(0, 57344, '\p{jamoextb}', "");
Expect(1, 57344, '\p{^jamoextb}', "");
Expect(1, 57344, '\P{jamoextb}', "");
Expect(0, 57344, '\P{^jamoextb}', "");
Expect(1, 55295, '\p{_Jamo_Ext_B}', "");
Expect(0, 55295, '\p{^_Jamo_Ext_B}', "");
Expect(0, 55295, '\P{_Jamo_Ext_B}', "");
Expect(1, 55295, '\P{^_Jamo_Ext_B}', "");
Expect(0, 57344, '\p{_Jamo_Ext_B}', "");
Expect(1, 57344, '\p{^_Jamo_Ext_B}', "");
Expect(1, 57344, '\P{_Jamo_Ext_B}', "");
Expect(0, 57344, '\P{^_Jamo_Ext_B}', "");
Error('\p{	-Is_JAMO_EXT_B:=}');
Error('\P{	-Is_JAMO_EXT_B:=}');
Expect(1, 55295, '\p{isjamoextb}', "");
Expect(0, 55295, '\p{^isjamoextb}', "");
Expect(0, 55295, '\P{isjamoextb}', "");
Expect(1, 55295, '\P{^isjamoextb}', "");
Expect(0, 57344, '\p{isjamoextb}', "");
Expect(1, 57344, '\p{^isjamoextb}', "");
Expect(1, 57344, '\P{isjamoextb}', "");
Expect(0, 57344, '\P{^isjamoextb}', "");
Expect(1, 55295, '\p{_is_Jamo_ext_B}', "");
Expect(0, 55295, '\p{^_is_Jamo_ext_B}', "");
Expect(0, 55295, '\P{_is_Jamo_ext_B}', "");
Expect(1, 55295, '\P{^_is_Jamo_ext_B}', "");
Expect(0, 57344, '\p{_is_Jamo_ext_B}', "");
Expect(1, 57344, '\p{^_is_Jamo_ext_B}', "");
Expect(1, 57344, '\P{_is_Jamo_ext_B}', "");
Expect(0, 57344, '\P{^_is_Jamo_ext_B}', "");
Error('\p{	_In_jamo_Ext_B:=}');
Error('\P{	_In_jamo_Ext_B:=}');
Expect(1, 55295, '\p{injamoextb}', "");
Expect(0, 55295, '\p{^injamoextb}', "");
Expect(0, 55295, '\P{injamoextb}', "");
Expect(1, 55295, '\P{^injamoextb}', "");
Expect(0, 57344, '\p{injamoextb}', "");
Expect(1, 57344, '\p{^injamoextb}', "");
Expect(1, 57344, '\P{injamoextb}', "");
Expect(0, 57344, '\P{^injamoextb}', "");
Expect(1, 55295, '\p{		IN_JAMO_ext_b}', "");
Expect(0, 55295, '\p{^		IN_JAMO_ext_b}', "");
Expect(0, 55295, '\P{		IN_JAMO_ext_b}', "");
Expect(1, 55295, '\P{^		IN_JAMO_ext_b}', "");
Expect(0, 57344, '\p{		IN_JAMO_ext_b}', "");
Expect(1, 57344, '\p{^		IN_JAMO_ext_b}', "");
Expect(1, 57344, '\P{		IN_JAMO_ext_b}', "");
Expect(0, 57344, '\P{^		IN_JAMO_ext_b}', "");
Error('\p{ _HANGUL_SYLLABLES:=}');
Error('\P{ _HANGUL_SYLLABLES:=}');
Expect(1, 55215, '\p{hangulsyllables}', "");
Expect(0, 55215, '\p{^hangulsyllables}', "");
Expect(0, 55215, '\P{hangulsyllables}', "");
Expect(1, 55215, '\P{^hangulsyllables}', "");
Expect(0, 55216, '\p{hangulsyllables}', "");
Expect(1, 55216, '\p{^hangulsyllables}', "");
Expect(1, 55216, '\P{hangulsyllables}', "");
Expect(0, 55216, '\P{^hangulsyllables}', "");
Expect(1, 55215, '\p{		Hangul_SYLLABLES}', "");
Expect(0, 55215, '\p{^		Hangul_SYLLABLES}', "");
Expect(0, 55215, '\P{		Hangul_SYLLABLES}', "");
Expect(1, 55215, '\P{^		Hangul_SYLLABLES}', "");
Expect(0, 55216, '\p{		Hangul_SYLLABLES}', "");
Expect(1, 55216, '\p{^		Hangul_SYLLABLES}', "");
Expect(1, 55216, '\P{		Hangul_SYLLABLES}', "");
Expect(0, 55216, '\P{^		Hangul_SYLLABLES}', "");
Error('\p{:=is_Hangul_Syllables}');
Error('\P{:=is_Hangul_Syllables}');
Expect(1, 55215, '\p{ishangulsyllables}', "");
Expect(0, 55215, '\p{^ishangulsyllables}', "");
Expect(0, 55215, '\P{ishangulsyllables}', "");
Expect(1, 55215, '\P{^ishangulsyllables}', "");
Expect(0, 55216, '\p{ishangulsyllables}', "");
Expect(1, 55216, '\p{^ishangulsyllables}', "");
Expect(1, 55216, '\P{ishangulsyllables}', "");
Expect(0, 55216, '\P{^ishangulsyllables}', "");
Expect(1, 55215, '\p{-is_Hangul_Syllables}', "");
Expect(0, 55215, '\p{^-is_Hangul_Syllables}', "");
Expect(0, 55215, '\P{-is_Hangul_Syllables}', "");
Expect(1, 55215, '\P{^-is_Hangul_Syllables}', "");
Expect(0, 55216, '\p{-is_Hangul_Syllables}', "");
Expect(1, 55216, '\p{^-is_Hangul_Syllables}', "");
Expect(1, 55216, '\P{-is_Hangul_Syllables}', "");
Expect(0, 55216, '\P{^-is_Hangul_Syllables}', "");
Error('\p{/a/ -in_Hangul_syllables}');
Error('\P{/a/ -in_Hangul_syllables}');
Expect(1, 55215, '\p{inhangulsyllables}', "");
Expect(0, 55215, '\p{^inhangulsyllables}', "");
Expect(0, 55215, '\P{inhangulsyllables}', "");
Expect(1, 55215, '\P{^inhangulsyllables}', "");
Expect(0, 55216, '\p{inhangulsyllables}', "");
Expect(1, 55216, '\p{^inhangulsyllables}', "");
Expect(1, 55216, '\P{inhangulsyllables}', "");
Expect(0, 55216, '\P{^inhangulsyllables}', "");
Expect(1, 55215, '\p{IN_HANGUL_Syllables}', "");
Expect(0, 55215, '\p{^IN_HANGUL_Syllables}', "");
Expect(0, 55215, '\P{IN_HANGUL_Syllables}', "");
Expect(1, 55215, '\P{^IN_HANGUL_Syllables}', "");
Expect(0, 55216, '\p{IN_HANGUL_Syllables}', "");
Expect(1, 55216, '\p{^IN_HANGUL_Syllables}', "");
Expect(1, 55216, '\P{IN_HANGUL_Syllables}', "");
Expect(0, 55216, '\P{^IN_HANGUL_Syllables}', "");
Error('\p{:=	In_Hangul}');
Error('\P{:=	In_Hangul}');
Expect(1, 55215, '\p{inhangul}', "");
Expect(0, 55215, '\p{^inhangul}', "");
Expect(0, 55215, '\P{inhangul}', "");
Expect(1, 55215, '\P{^inhangul}', "");
Expect(0, 55216, '\p{inhangul}', "");
Expect(1, 55216, '\p{^inhangul}', "");
Expect(1, 55216, '\P{inhangul}', "");
Expect(0, 55216, '\P{^inhangul}', "");
Expect(1, 55215, '\p{ 	In_Hangul}', "");
Expect(0, 55215, '\p{^ 	In_Hangul}', "");
Expect(0, 55215, '\P{ 	In_Hangul}', "");
Expect(1, 55215, '\P{^ 	In_Hangul}', "");
Expect(0, 55216, '\p{ 	In_Hangul}', "");
Expect(1, 55216, '\p{^ 	In_Hangul}', "");
Expect(1, 55216, '\P{ 	In_Hangul}', "");
Expect(0, 55216, '\P{^ 	In_Hangul}', "");
Error('\p{:=	 HANUNOO}');
Error('\P{:=	 HANUNOO}');
Expect(1, 5940, '\p{hanunoo}', "");
Expect(0, 5940, '\p{^hanunoo}', "");
Expect(0, 5940, '\P{hanunoo}', "");
Expect(1, 5940, '\P{^hanunoo}', "");
Expect(0, 5941, '\p{hanunoo}', "");
Expect(1, 5941, '\p{^hanunoo}', "");
Expect(1, 5941, '\P{hanunoo}', "");
Expect(0, 5941, '\P{^hanunoo}', "");
Expect(1, 5940, '\p{__Hanunoo}', "");
Expect(0, 5940, '\p{^__Hanunoo}', "");
Expect(0, 5940, '\P{__Hanunoo}', "");
Expect(1, 5940, '\P{^__Hanunoo}', "");
Expect(0, 5941, '\p{__Hanunoo}', "");
Expect(1, 5941, '\p{^__Hanunoo}', "");
Expect(1, 5941, '\P{__Hanunoo}', "");
Expect(0, 5941, '\P{^__Hanunoo}', "");
Error('\p{:= Is_HANUNOO}');
Error('\P{:= Is_HANUNOO}');
Expect(1, 5940, '\p{ishanunoo}', "");
Expect(0, 5940, '\p{^ishanunoo}', "");
Expect(0, 5940, '\P{ishanunoo}', "");
Expect(1, 5940, '\P{^ishanunoo}', "");
Expect(0, 5941, '\p{ishanunoo}', "");
Expect(1, 5941, '\p{^ishanunoo}', "");
Expect(1, 5941, '\P{ishanunoo}', "");
Expect(0, 5941, '\P{^ishanunoo}', "");
Expect(1, 5940, '\p{-_Is_hanunoo}', "");
Expect(0, 5940, '\p{^-_Is_hanunoo}', "");
Expect(0, 5940, '\P{-_Is_hanunoo}', "");
Expect(1, 5940, '\P{^-_Is_hanunoo}', "");
Expect(0, 5941, '\p{-_Is_hanunoo}', "");
Expect(1, 5941, '\p{^-_Is_hanunoo}', "");
Expect(1, 5941, '\P{-_Is_hanunoo}', "");
Expect(0, 5941, '\P{^-_Is_hanunoo}', "");
Error('\p{ /a/HANO}');
Error('\P{ /a/HANO}');
Expect(1, 5940, '\p{hano}', "");
Expect(0, 5940, '\p{^hano}', "");
Expect(0, 5940, '\P{hano}', "");
Expect(1, 5940, '\P{^hano}', "");
Expect(0, 5941, '\p{hano}', "");
Expect(1, 5941, '\p{^hano}', "");
Expect(1, 5941, '\P{hano}', "");
Expect(0, 5941, '\P{^hano}', "");
Expect(1, 5940, '\p{--HANO}', "");
Expect(0, 5940, '\p{^--HANO}', "");
Expect(0, 5940, '\P{--HANO}', "");
Expect(1, 5940, '\P{^--HANO}', "");
Expect(0, 5941, '\p{--HANO}', "");
Expect(1, 5941, '\p{^--HANO}', "");
Expect(1, 5941, '\P{--HANO}', "");
Expect(0, 5941, '\P{^--HANO}', "");
Error('\p{/a/-is_HANO}');
Error('\P{/a/-is_HANO}');
Expect(1, 5940, '\p{ishano}', "");
Expect(0, 5940, '\p{^ishano}', "");
Expect(0, 5940, '\P{ishano}', "");
Expect(1, 5940, '\P{^ishano}', "");
Expect(0, 5941, '\p{ishano}', "");
Expect(1, 5941, '\p{^ishano}', "");
Expect(1, 5941, '\P{ishano}', "");
Expect(0, 5941, '\P{^ishano}', "");
Expect(1, 5940, '\p{Is_Hano}', "");
Expect(0, 5940, '\p{^Is_Hano}', "");
Expect(0, 5940, '\P{Is_Hano}', "");
Expect(1, 5940, '\P{^Is_Hano}', "");
Expect(0, 5941, '\p{Is_Hano}', "");
Expect(1, 5941, '\p{^Is_Hano}', "");
Expect(1, 5941, '\P{Is_Hano}', "");
Expect(0, 5941, '\P{^Is_Hano}', "");
Error('\p{	HEBREW:=}');
Error('\P{	HEBREW:=}');
Expect(1, 64335, '\p{hebrew}', "");
Expect(0, 64335, '\p{^hebrew}', "");
Expect(0, 64335, '\P{hebrew}', "");
Expect(1, 64335, '\P{^hebrew}', "");
Expect(0, 64336, '\p{hebrew}', "");
Expect(1, 64336, '\p{^hebrew}', "");
Expect(1, 64336, '\P{hebrew}', "");
Expect(0, 64336, '\P{^hebrew}', "");
Expect(1, 64335, '\p{ 	Hebrew}', "");
Expect(0, 64335, '\p{^ 	Hebrew}', "");
Expect(0, 64335, '\P{ 	Hebrew}', "");
Expect(1, 64335, '\P{^ 	Hebrew}', "");
Expect(0, 64336, '\p{ 	Hebrew}', "");
Expect(1, 64336, '\p{^ 	Hebrew}', "");
Expect(1, 64336, '\P{ 	Hebrew}', "");
Expect(0, 64336, '\P{^ 	Hebrew}', "");
Error('\p{:=Is_Hebrew}');
Error('\P{:=Is_Hebrew}');
Expect(1, 64335, '\p{ishebrew}', "");
Expect(0, 64335, '\p{^ishebrew}', "");
Expect(0, 64335, '\P{ishebrew}', "");
Expect(1, 64335, '\P{^ishebrew}', "");
Expect(0, 64336, '\p{ishebrew}', "");
Expect(1, 64336, '\p{^ishebrew}', "");
Expect(1, 64336, '\P{ishebrew}', "");
Expect(0, 64336, '\P{^ishebrew}', "");
Expect(1, 64335, '\p{-	Is_hebrew}', "");
Expect(0, 64335, '\p{^-	Is_hebrew}', "");
Expect(0, 64335, '\P{-	Is_hebrew}', "");
Expect(1, 64335, '\P{^-	Is_hebrew}', "");
Expect(0, 64336, '\p{-	Is_hebrew}', "");
Expect(1, 64336, '\p{^-	Is_hebrew}', "");
Expect(1, 64336, '\P{-	Is_hebrew}', "");
Expect(0, 64336, '\P{^-	Is_hebrew}', "");
Error('\p{/a/ _Hebr}');
Error('\P{/a/ _Hebr}');
Expect(1, 64335, '\p{hebr}', "");
Expect(0, 64335, '\p{^hebr}', "");
Expect(0, 64335, '\P{hebr}', "");
Expect(1, 64335, '\P{^hebr}', "");
Expect(0, 64336, '\p{hebr}', "");
Expect(1, 64336, '\p{^hebr}', "");
Expect(1, 64336, '\P{hebr}', "");
Expect(0, 64336, '\P{^hebr}', "");
Expect(1, 64335, '\p{	-Hebr}', "");
Expect(0, 64335, '\p{^	-Hebr}', "");
Expect(0, 64335, '\P{	-Hebr}', "");
Expect(1, 64335, '\P{^	-Hebr}', "");
Expect(0, 64336, '\p{	-Hebr}', "");
Expect(1, 64336, '\p{^	-Hebr}', "");
Expect(1, 64336, '\P{	-Hebr}', "");
Expect(0, 64336, '\P{^	-Hebr}', "");
Error('\p{_-IS_Hebr:=}');
Error('\P{_-IS_Hebr:=}');
Expect(1, 64335, '\p{ishebr}', "");
Expect(0, 64335, '\p{^ishebr}', "");
Expect(0, 64335, '\P{ishebr}', "");
Expect(1, 64335, '\P{^ishebr}', "");
Expect(0, 64336, '\p{ishebr}', "");
Expect(1, 64336, '\p{^ishebr}', "");
Expect(1, 64336, '\P{ishebr}', "");
Expect(0, 64336, '\P{^ishebr}', "");
Expect(1, 64335, '\p{	Is_HEBR}', "");
Expect(0, 64335, '\p{^	Is_HEBR}', "");
Expect(0, 64335, '\P{	Is_HEBR}', "");
Expect(1, 64335, '\P{^	Is_HEBR}', "");
Expect(0, 64336, '\p{	Is_HEBR}', "");
Expect(1, 64336, '\p{^	Is_HEBR}', "");
Expect(1, 64336, '\P{	Is_HEBR}', "");
Expect(0, 64336, '\P{^	Is_HEBR}', "");
Error('\p{_high_Private_use_SURROGATES/a/}');
Error('\P{_high_Private_use_SURROGATES/a/}');
Expect(1, 56319, '\p{highprivateusesurrogates}', "");
Expect(0, 56319, '\p{^highprivateusesurrogates}', "");
Expect(0, 56319, '\P{highprivateusesurrogates}', "");
Expect(1, 56319, '\P{^highprivateusesurrogates}', "");
Expect(0, 57344, '\p{highprivateusesurrogates}', "");
Expect(1, 57344, '\p{^highprivateusesurrogates}', "");
Expect(1, 57344, '\P{highprivateusesurrogates}', "");
Expect(0, 57344, '\P{^highprivateusesurrogates}', "");
Expect(1, 56319, '\p{-High_Private_use_Surrogates}', "");
Expect(0, 56319, '\p{^-High_Private_use_Surrogates}', "");
Expect(0, 56319, '\P{-High_Private_use_Surrogates}', "");
Expect(1, 56319, '\P{^-High_Private_use_Surrogates}', "");
Expect(0, 57344, '\p{-High_Private_use_Surrogates}', "");
Expect(1, 57344, '\p{^-High_Private_use_Surrogates}', "");
Expect(1, 57344, '\P{-High_Private_use_Surrogates}', "");
Expect(0, 57344, '\P{^-High_Private_use_Surrogates}', "");
Error('\p{_IS_High_Private_Use_SURROGATES/a/}');
Error('\P{_IS_High_Private_Use_SURROGATES/a/}');
Expect(1, 56319, '\p{ishighprivateusesurrogates}', "");
Expect(0, 56319, '\p{^ishighprivateusesurrogates}', "");
Expect(0, 56319, '\P{ishighprivateusesurrogates}', "");
Expect(1, 56319, '\P{^ishighprivateusesurrogates}', "");
Expect(0, 57344, '\p{ishighprivateusesurrogates}', "");
Expect(1, 57344, '\p{^ishighprivateusesurrogates}', "");
Expect(1, 57344, '\P{ishighprivateusesurrogates}', "");
Expect(0, 57344, '\P{^ishighprivateusesurrogates}', "");
Expect(1, 56319, '\p{_ Is_HIGH_Private_Use_Surrogates}', "");
Expect(0, 56319, '\p{^_ Is_HIGH_Private_Use_Surrogates}', "");
Expect(0, 56319, '\P{_ Is_HIGH_Private_Use_Surrogates}', "");
Expect(1, 56319, '\P{^_ Is_HIGH_Private_Use_Surrogates}', "");
Expect(0, 57344, '\p{_ Is_HIGH_Private_Use_Surrogates}', "");
Expect(1, 57344, '\p{^_ Is_HIGH_Private_Use_Surrogates}', "");
Expect(1, 57344, '\P{_ Is_HIGH_Private_Use_Surrogates}', "");
Expect(0, 57344, '\P{^_ Is_HIGH_Private_Use_Surrogates}', "");
Error('\p{	in_High_PRIVATE_Use_Surrogates:=}');
Error('\P{	in_High_PRIVATE_Use_Surrogates:=}');
Expect(1, 56319, '\p{inhighprivateusesurrogates}', "");
Expect(0, 56319, '\p{^inhighprivateusesurrogates}', "");
Expect(0, 56319, '\P{inhighprivateusesurrogates}', "");
Expect(1, 56319, '\P{^inhighprivateusesurrogates}', "");
Expect(0, 57344, '\p{inhighprivateusesurrogates}', "");
Expect(1, 57344, '\p{^inhighprivateusesurrogates}', "");
Expect(1, 57344, '\P{inhighprivateusesurrogates}', "");
Expect(0, 57344, '\P{^inhighprivateusesurrogates}', "");
Expect(1, 56319, '\p{ _In_high_private_Use_surrogates}', "");
Expect(0, 56319, '\p{^ _In_high_private_Use_surrogates}', "");
Expect(0, 56319, '\P{ _In_high_private_Use_surrogates}', "");
Expect(1, 56319, '\P{^ _In_high_private_Use_surrogates}', "");
Expect(0, 57344, '\p{ _In_high_private_Use_surrogates}', "");
Expect(1, 57344, '\p{^ _In_high_private_Use_surrogates}', "");
Expect(1, 57344, '\P{ _In_high_private_Use_surrogates}', "");
Expect(0, 57344, '\P{^ _In_high_private_Use_surrogates}', "");
Error('\p{_high_pu_Surrogates:=}');
Error('\P{_high_pu_Surrogates:=}');
Expect(1, 56319, '\p{highpusurrogates}', "");
Expect(0, 56319, '\p{^highpusurrogates}', "");
Expect(0, 56319, '\P{highpusurrogates}', "");
Expect(1, 56319, '\P{^highpusurrogates}', "");
Expect(0, 57344, '\p{highpusurrogates}', "");
Expect(1, 57344, '\p{^highpusurrogates}', "");
Expect(1, 57344, '\P{highpusurrogates}', "");
Expect(0, 57344, '\P{^highpusurrogates}', "");
Expect(1, 56319, '\p{_High_pu_surrogates}', "");
Expect(0, 56319, '\p{^_High_pu_surrogates}', "");
Expect(0, 56319, '\P{_High_pu_surrogates}', "");
Expect(1, 56319, '\P{^_High_pu_surrogates}', "");
Expect(0, 57344, '\p{_High_pu_surrogates}', "");
Expect(1, 57344, '\p{^_High_pu_surrogates}', "");
Expect(1, 57344, '\P{_High_pu_surrogates}', "");
Expect(0, 57344, '\P{^_High_pu_surrogates}', "");
Error('\p{-Is_high_PU_surrogates/a/}');
Error('\P{-Is_high_PU_surrogates/a/}');
Expect(1, 56319, '\p{ishighpusurrogates}', "");
Expect(0, 56319, '\p{^ishighpusurrogates}', "");
Expect(0, 56319, '\P{ishighpusurrogates}', "");
Expect(1, 56319, '\P{^ishighpusurrogates}', "");
Expect(0, 57344, '\p{ishighpusurrogates}', "");
Expect(1, 57344, '\p{^ishighpusurrogates}', "");
Expect(1, 57344, '\P{ishighpusurrogates}', "");
Expect(0, 57344, '\P{^ishighpusurrogates}', "");
Expect(1, 56319, '\p{  Is_high_pu_Surrogates}', "");
Expect(0, 56319, '\p{^  Is_high_pu_Surrogates}', "");
Expect(0, 56319, '\P{  Is_high_pu_Surrogates}', "");
Expect(1, 56319, '\P{^  Is_high_pu_Surrogates}', "");
Expect(0, 57344, '\p{  Is_high_pu_Surrogates}', "");
Expect(1, 57344, '\p{^  Is_high_pu_Surrogates}', "");
Expect(1, 57344, '\P{  Is_high_pu_Surrogates}', "");
Expect(0, 57344, '\P{^  Is_high_pu_Surrogates}', "");
Error('\p{	/a/In_HIGH_pu_SURROGATES}');
Error('\P{	/a/In_HIGH_pu_SURROGATES}');
Expect(1, 56319, '\p{inhighpusurrogates}', "");
Expect(0, 56319, '\p{^inhighpusurrogates}', "");
Expect(0, 56319, '\P{inhighpusurrogates}', "");
Expect(1, 56319, '\P{^inhighpusurrogates}', "");
Expect(0, 57344, '\p{inhighpusurrogates}', "");
Expect(1, 57344, '\p{^inhighpusurrogates}', "");
Expect(1, 57344, '\P{inhighpusurrogates}', "");
Expect(0, 57344, '\P{^inhighpusurrogates}', "");
Expect(1, 56319, '\p{-In_High_PU_surrogates}', "");
Expect(0, 56319, '\p{^-In_High_PU_surrogates}', "");
Expect(0, 56319, '\P{-In_High_PU_surrogates}', "");
Expect(1, 56319, '\P{^-In_High_PU_surrogates}', "");
Expect(0, 57344, '\p{-In_High_PU_surrogates}', "");
Expect(1, 57344, '\p{^-In_High_PU_surrogates}', "");
Expect(1, 57344, '\P{-In_High_PU_surrogates}', "");
Expect(0, 57344, '\P{^-In_High_PU_surrogates}', "");
Error('\p{- HIGH_surrogates/a/}');
Error('\P{- HIGH_surrogates/a/}');
Expect(1, 56191, '\p{highsurrogates}', "");
Expect(0, 56191, '\p{^highsurrogates}', "");
Expect(0, 56191, '\P{highsurrogates}', "");
Expect(1, 56191, '\P{^highsurrogates}', "");
Expect(0, 57344, '\p{highsurrogates}', "");
Expect(1, 57344, '\p{^highsurrogates}', "");
Expect(1, 57344, '\P{highsurrogates}', "");
Expect(0, 57344, '\P{^highsurrogates}', "");
Expect(1, 56191, '\p{-_High_Surrogates}', "");
Expect(0, 56191, '\p{^-_High_Surrogates}', "");
Expect(0, 56191, '\P{-_High_Surrogates}', "");
Expect(1, 56191, '\P{^-_High_Surrogates}', "");
Expect(0, 57344, '\p{-_High_Surrogates}', "");
Expect(1, 57344, '\p{^-_High_Surrogates}', "");
Expect(1, 57344, '\P{-_High_Surrogates}', "");
Expect(0, 57344, '\P{^-_High_Surrogates}', "");
Error('\p{	/a/IS_High_Surrogates}');
Error('\P{	/a/IS_High_Surrogates}');
Expect(1, 56191, '\p{ishighsurrogates}', "");
Expect(0, 56191, '\p{^ishighsurrogates}', "");
Expect(0, 56191, '\P{ishighsurrogates}', "");
Expect(1, 56191, '\P{^ishighsurrogates}', "");
Expect(0, 57344, '\p{ishighsurrogates}', "");
Expect(1, 57344, '\p{^ishighsurrogates}', "");
Expect(1, 57344, '\P{ishighsurrogates}', "");
Expect(0, 57344, '\P{^ishighsurrogates}', "");
Expect(1, 56191, '\p{	IS_HIGH_SURROGATES}', "");
Expect(0, 56191, '\p{^	IS_HIGH_SURROGATES}', "");
Expect(0, 56191, '\P{	IS_HIGH_SURROGATES}', "");
Expect(1, 56191, '\P{^	IS_HIGH_SURROGATES}', "");
Expect(0, 57344, '\p{	IS_HIGH_SURROGATES}', "");
Expect(1, 57344, '\p{^	IS_HIGH_SURROGATES}', "");
Expect(1, 57344, '\P{	IS_HIGH_SURROGATES}', "");
Expect(0, 57344, '\P{^	IS_HIGH_SURROGATES}', "");
Error('\p{ 	In_high_Surrogates/a/}');
Error('\P{ 	In_high_Surrogates/a/}');
Expect(1, 56191, '\p{inhighsurrogates}', "");
Expect(0, 56191, '\p{^inhighsurrogates}', "");
Expect(0, 56191, '\P{inhighsurrogates}', "");
Expect(1, 56191, '\P{^inhighsurrogates}', "");
Expect(0, 57344, '\p{inhighsurrogates}', "");
Expect(1, 57344, '\p{^inhighsurrogates}', "");
Expect(1, 57344, '\P{inhighsurrogates}', "");
Expect(0, 57344, '\P{^inhighsurrogates}', "");
Expect(1, 56191, '\p{--In_HIGH_SURROGATES}', "");
Expect(0, 56191, '\p{^--In_HIGH_SURROGATES}', "");
Expect(0, 56191, '\P{--In_HIGH_SURROGATES}', "");
Expect(1, 56191, '\P{^--In_HIGH_SURROGATES}', "");
Expect(0, 57344, '\p{--In_HIGH_SURROGATES}', "");
Expect(1, 57344, '\p{^--In_HIGH_SURROGATES}', "");
Expect(1, 57344, '\P{--In_HIGH_SURROGATES}', "");
Expect(0, 57344, '\P{^--In_HIGH_SURROGATES}', "");
Error('\p{:=_ Hiragana}');
Error('\P{:=_ Hiragana}');
Expect(1, 127488, '\p{hiragana}', "");
Expect(0, 127488, '\p{^hiragana}', "");
Expect(0, 127488, '\P{hiragana}', "");
Expect(1, 127488, '\P{^hiragana}', "");
Expect(0, 127489, '\p{hiragana}', "");
Expect(1, 127489, '\p{^hiragana}', "");
Expect(1, 127489, '\P{hiragana}', "");
Expect(0, 127489, '\P{^hiragana}', "");
Expect(1, 127488, '\p{	Hiragana}', "");
Expect(0, 127488, '\p{^	Hiragana}', "");
Expect(0, 127488, '\P{	Hiragana}', "");
Expect(1, 127488, '\P{^	Hiragana}', "");
Expect(0, 127489, '\p{	Hiragana}', "");
Expect(1, 127489, '\p{^	Hiragana}', "");
Expect(1, 127489, '\P{	Hiragana}', "");
Expect(0, 127489, '\P{^	Hiragana}', "");
Error('\p{:=-_Is_Hiragana}');
Error('\P{:=-_Is_Hiragana}');
Expect(1, 127488, '\p{ishiragana}', "");
Expect(0, 127488, '\p{^ishiragana}', "");
Expect(0, 127488, '\P{ishiragana}', "");
Expect(1, 127488, '\P{^ishiragana}', "");
Expect(0, 127489, '\p{ishiragana}', "");
Expect(1, 127489, '\p{^ishiragana}', "");
Expect(1, 127489, '\P{ishiragana}', "");
Expect(0, 127489, '\P{^ishiragana}', "");
Expect(1, 127488, '\p{_Is_Hiragana}', "");
Expect(0, 127488, '\p{^_Is_Hiragana}', "");
Expect(0, 127488, '\P{_Is_Hiragana}', "");
Expect(1, 127488, '\P{^_Is_Hiragana}', "");
Expect(0, 127489, '\p{_Is_Hiragana}', "");
Expect(1, 127489, '\p{^_Is_Hiragana}', "");
Expect(1, 127489, '\P{_Is_Hiragana}', "");
Expect(0, 127489, '\P{^_Is_Hiragana}', "");
Error('\p{ _HIRA:=}');
Error('\P{ _HIRA:=}');
Expect(1, 127488, '\p{hira}', "");
Expect(0, 127488, '\p{^hira}', "");
Expect(0, 127488, '\P{hira}', "");
Expect(1, 127488, '\P{^hira}', "");
Expect(0, 127489, '\p{hira}', "");
Expect(1, 127489, '\p{^hira}', "");
Expect(1, 127489, '\P{hira}', "");
Expect(0, 127489, '\P{^hira}', "");
Expect(1, 127488, '\p{-hira}', "");
Expect(0, 127488, '\p{^-hira}', "");
Expect(0, 127488, '\P{-hira}', "");
Expect(1, 127488, '\P{^-hira}', "");
Expect(0, 127489, '\p{-hira}', "");
Expect(1, 127489, '\p{^-hira}', "");
Expect(1, 127489, '\P{-hira}', "");
Expect(0, 127489, '\P{^-hira}', "");
Error('\p{	/a/IS_Hira}');
Error('\P{	/a/IS_Hira}');
Expect(1, 127488, '\p{ishira}', "");
Expect(0, 127488, '\p{^ishira}', "");
Expect(0, 127488, '\P{ishira}', "");
Expect(1, 127488, '\P{^ishira}', "");
Expect(0, 127489, '\p{ishira}', "");
Expect(1, 127489, '\p{^ishira}', "");
Expect(1, 127489, '\P{ishira}', "");
Expect(0, 127489, '\P{^ishira}', "");
Expect(1, 127488, '\p{-Is_Hira}', "");
Expect(0, 127488, '\p{^-Is_Hira}', "");
Expect(0, 127488, '\P{-Is_Hira}', "");
Expect(1, 127488, '\P{^-Is_Hira}', "");
Expect(0, 127489, '\p{-Is_Hira}', "");
Expect(1, 127489, '\p{^-Is_Hira}', "");
Expect(1, 127489, '\P{-Is_Hira}', "");
Expect(0, 127489, '\P{^-Is_Hira}', "");
Error('\p{:= _HYPHEN}');
Error('\P{:= _HYPHEN}');
Expect(1, 65381, '\p{hyphen}', 'deprecated');
Expect(0, 65381, '\p{^hyphen}', 'deprecated');
Expect(0, 65381, '\P{hyphen}', 'deprecated');
Expect(1, 65381, '\P{^hyphen}', 'deprecated');
Expect(0, 65382, '\p{hyphen}', 'deprecated');
Expect(1, 65382, '\p{^hyphen}', 'deprecated');
Expect(1, 65382, '\P{hyphen}', 'deprecated');
Expect(0, 65382, '\P{^hyphen}', 'deprecated');
Expect(1, 65381, '\p{ 	Hyphen}', 'deprecated');
Expect(0, 65381, '\p{^ 	Hyphen}', 'deprecated');
Expect(0, 65381, '\P{ 	Hyphen}', 'deprecated');
Expect(1, 65381, '\P{^ 	Hyphen}', 'deprecated');
Expect(0, 65382, '\p{ 	Hyphen}', 'deprecated');
Expect(1, 65382, '\p{^ 	Hyphen}', 'deprecated');
Expect(1, 65382, '\P{ 	Hyphen}', 'deprecated');
Expect(0, 65382, '\P{^ 	Hyphen}', 'deprecated');
Error('\p{-/a/Is_Hyphen}');
Error('\P{-/a/Is_Hyphen}');
Expect(1, 65381, '\p{ishyphen}', 'deprecated');
Expect(0, 65381, '\p{^ishyphen}', 'deprecated');
Expect(0, 65381, '\P{ishyphen}', 'deprecated');
Expect(1, 65381, '\P{^ishyphen}', 'deprecated');
Expect(0, 65382, '\p{ishyphen}', 'deprecated');
Expect(1, 65382, '\p{^ishyphen}', 'deprecated');
Expect(1, 65382, '\P{ishyphen}', 'deprecated');
Expect(0, 65382, '\P{^ishyphen}', 'deprecated');
Expect(1, 65381, '\p{ -is_Hyphen}', 'deprecated');
Expect(0, 65381, '\p{^ -is_Hyphen}', 'deprecated');
Expect(0, 65381, '\P{ -is_Hyphen}', 'deprecated');
Expect(1, 65381, '\P{^ -is_Hyphen}', 'deprecated');
Expect(0, 65382, '\p{ -is_Hyphen}', 'deprecated');
Expect(1, 65382, '\p{^ -is_Hyphen}', 'deprecated');
Expect(1, 65382, '\P{ -is_Hyphen}', 'deprecated');
Expect(0, 65382, '\P{^ -is_Hyphen}', 'deprecated');
Error('\p{-/a/IDS_Binary_Operator}');
Error('\P{-/a/IDS_Binary_Operator}');
Expect(1, 12283, '\p{idsbinaryoperator}', "");
Expect(0, 12283, '\p{^idsbinaryoperator}', "");
Expect(0, 12283, '\P{idsbinaryoperator}', "");
Expect(1, 12283, '\P{^idsbinaryoperator}', "");
Expect(0, 12284, '\p{idsbinaryoperator}', "");
Expect(1, 12284, '\p{^idsbinaryoperator}', "");
Expect(1, 12284, '\P{idsbinaryoperator}', "");
Expect(0, 12284, '\P{^idsbinaryoperator}', "");
Expect(1, 12283, '\p{_ IDS_BINARY_operator}', "");
Expect(0, 12283, '\p{^_ IDS_BINARY_operator}', "");
Expect(0, 12283, '\P{_ IDS_BINARY_operator}', "");
Expect(1, 12283, '\P{^_ IDS_BINARY_operator}', "");
Expect(0, 12284, '\p{_ IDS_BINARY_operator}', "");
Expect(1, 12284, '\p{^_ IDS_BINARY_operator}', "");
Expect(1, 12284, '\P{_ IDS_BINARY_operator}', "");
Expect(0, 12284, '\P{^_ IDS_BINARY_operator}', "");
Error('\p{ /a/is_IDS_Binary_Operator}');
Error('\P{ /a/is_IDS_Binary_Operator}');
Expect(1, 12283, '\p{isidsbinaryoperator}', "");
Expect(0, 12283, '\p{^isidsbinaryoperator}', "");
Expect(0, 12283, '\P{isidsbinaryoperator}', "");
Expect(1, 12283, '\P{^isidsbinaryoperator}', "");
Expect(0, 12284, '\p{isidsbinaryoperator}', "");
Expect(1, 12284, '\p{^isidsbinaryoperator}', "");
Expect(1, 12284, '\P{isidsbinaryoperator}', "");
Expect(0, 12284, '\P{^isidsbinaryoperator}', "");
Expect(1, 12283, '\p{_	IS_IDS_BINARY_Operator}', "");
Expect(0, 12283, '\p{^_	IS_IDS_BINARY_Operator}', "");
Expect(0, 12283, '\P{_	IS_IDS_BINARY_Operator}', "");
Expect(1, 12283, '\P{^_	IS_IDS_BINARY_Operator}', "");
Expect(0, 12284, '\p{_	IS_IDS_BINARY_Operator}', "");
Expect(1, 12284, '\p{^_	IS_IDS_BINARY_Operator}', "");
Expect(1, 12284, '\P{_	IS_IDS_BINARY_Operator}', "");
Expect(0, 12284, '\P{^_	IS_IDS_BINARY_Operator}', "");
Error('\p{:=	 IDSB}');
Error('\P{:=	 IDSB}');
Expect(1, 12283, '\p{idsb}', "");
Expect(0, 12283, '\p{^idsb}', "");
Expect(0, 12283, '\P{idsb}', "");
Expect(1, 12283, '\P{^idsb}', "");
Expect(0, 12284, '\p{idsb}', "");
Expect(1, 12284, '\p{^idsb}', "");
Expect(1, 12284, '\P{idsb}', "");
Expect(0, 12284, '\P{^idsb}', "");
Expect(1, 12283, '\p{	 idsb}', "");
Expect(0, 12283, '\p{^	 idsb}', "");
Expect(0, 12283, '\P{	 idsb}', "");
Expect(1, 12283, '\P{^	 idsb}', "");
Expect(0, 12284, '\p{	 idsb}', "");
Expect(1, 12284, '\p{^	 idsb}', "");
Expect(1, 12284, '\P{	 idsb}', "");
Expect(0, 12284, '\P{^	 idsb}', "");
Error('\p{:=_IS_idsb}');
Error('\P{:=_IS_idsb}');
Expect(1, 12283, '\p{isidsb}', "");
Expect(0, 12283, '\p{^isidsb}', "");
Expect(0, 12283, '\P{isidsb}', "");
Expect(1, 12283, '\P{^isidsb}', "");
Expect(0, 12284, '\p{isidsb}', "");
Expect(1, 12284, '\p{^isidsb}', "");
Expect(1, 12284, '\P{isidsb}', "");
Expect(0, 12284, '\P{^isidsb}', "");
Expect(1, 12283, '\p{	-IS_IDSB}', "");
Expect(0, 12283, '\p{^	-IS_IDSB}', "");
Expect(0, 12283, '\P{	-IS_IDSB}', "");
Expect(1, 12283, '\P{^	-IS_IDSB}', "");
Expect(0, 12284, '\p{	-IS_IDSB}', "");
Expect(1, 12284, '\p{^	-IS_IDSB}', "");
Expect(1, 12284, '\P{	-IS_IDSB}', "");
Expect(0, 12284, '\P{^	-IS_IDSB}', "");
Error('\p{_IDS_Trinary_Operator/a/}');
Error('\P{_IDS_Trinary_Operator/a/}');
Expect(1, 12275, '\p{idstrinaryoperator}', "");
Expect(0, 12275, '\p{^idstrinaryoperator}', "");
Expect(0, 12275, '\P{idstrinaryoperator}', "");
Expect(1, 12275, '\P{^idstrinaryoperator}', "");
Expect(0, 12276, '\p{idstrinaryoperator}', "");
Expect(1, 12276, '\p{^idstrinaryoperator}', "");
Expect(1, 12276, '\P{idstrinaryoperator}', "");
Expect(0, 12276, '\P{^idstrinaryoperator}', "");
Expect(1, 12275, '\p{-IDS_TRINARY_operator}', "");
Expect(0, 12275, '\p{^-IDS_TRINARY_operator}', "");
Expect(0, 12275, '\P{-IDS_TRINARY_operator}', "");
Expect(1, 12275, '\P{^-IDS_TRINARY_operator}', "");
Expect(0, 12276, '\p{-IDS_TRINARY_operator}', "");
Expect(1, 12276, '\p{^-IDS_TRINARY_operator}', "");
Expect(1, 12276, '\P{-IDS_TRINARY_operator}', "");
Expect(0, 12276, '\P{^-IDS_TRINARY_operator}', "");
Error('\p{-_IS_ids_Trinary_Operator:=}');
Error('\P{-_IS_ids_Trinary_Operator:=}');
Expect(1, 12275, '\p{isidstrinaryoperator}', "");
Expect(0, 12275, '\p{^isidstrinaryoperator}', "");
Expect(0, 12275, '\P{isidstrinaryoperator}', "");
Expect(1, 12275, '\P{^isidstrinaryoperator}', "");
Expect(0, 12276, '\p{isidstrinaryoperator}', "");
Expect(1, 12276, '\p{^isidstrinaryoperator}', "");
Expect(1, 12276, '\P{isidstrinaryoperator}', "");
Expect(0, 12276, '\P{^isidstrinaryoperator}', "");
Expect(1, 12275, '\p{_-IS_IDS_Trinary_Operator}', "");
Expect(0, 12275, '\p{^_-IS_IDS_Trinary_Operator}', "");
Expect(0, 12275, '\P{_-IS_IDS_Trinary_Operator}', "");
Expect(1, 12275, '\P{^_-IS_IDS_Trinary_Operator}', "");
Expect(0, 12276, '\p{_-IS_IDS_Trinary_Operator}', "");
Expect(1, 12276, '\p{^_-IS_IDS_Trinary_Operator}', "");
Expect(1, 12276, '\P{_-IS_IDS_Trinary_Operator}', "");
Expect(0, 12276, '\P{^_-IS_IDS_Trinary_Operator}', "");
Error('\p{:=_IDST}');
Error('\P{:=_IDST}');
Expect(1, 12275, '\p{idst}', "");
Expect(0, 12275, '\p{^idst}', "");
Expect(0, 12275, '\P{idst}', "");
Expect(1, 12275, '\P{^idst}', "");
Expect(0, 12276, '\p{idst}', "");
Expect(1, 12276, '\p{^idst}', "");
Expect(1, 12276, '\P{idst}', "");
Expect(0, 12276, '\P{^idst}', "");
Expect(1, 12275, '\p{		IDST}', "");
Expect(0, 12275, '\p{^		IDST}', "");
Expect(0, 12275, '\P{		IDST}', "");
Expect(1, 12275, '\P{^		IDST}', "");
Expect(0, 12276, '\p{		IDST}', "");
Expect(1, 12276, '\p{^		IDST}', "");
Expect(1, 12276, '\P{		IDST}', "");
Expect(0, 12276, '\P{^		IDST}', "");
Error('\p{_/a/is_IDST}');
Error('\P{_/a/is_IDST}');
Expect(1, 12275, '\p{isidst}', "");
Expect(0, 12275, '\p{^isidst}', "");
Expect(0, 12275, '\P{isidst}', "");
Expect(1, 12275, '\P{^isidst}', "");
Expect(0, 12276, '\p{isidst}', "");
Expect(1, 12276, '\p{^isidst}', "");
Expect(1, 12276, '\P{isidst}', "");
Expect(0, 12276, '\P{^isidst}', "");
Expect(1, 12275, '\p{_	IS_idst}', "");
Expect(0, 12275, '\p{^_	IS_idst}', "");
Expect(0, 12275, '\P{_	IS_idst}', "");
Expect(1, 12275, '\P{^_	IS_idst}', "");
Expect(0, 12276, '\p{_	IS_idst}', "");
Expect(1, 12276, '\p{^_	IS_idst}', "");
Expect(1, 12276, '\P{_	IS_idst}', "");
Expect(0, 12276, '\P{^_	IS_idst}', "");
Error('\p{:=-ID_Continue}');
Error('\P{:=-ID_Continue}');
Expect(1, 917999, '\p{idcontinue}', "");
Expect(0, 917999, '\p{^idcontinue}', "");
Expect(0, 917999, '\P{idcontinue}', "");
Expect(1, 917999, '\P{^idcontinue}', "");
Expect(0, 918000, '\p{idcontinue}', "");
Expect(1, 918000, '\p{^idcontinue}', "");
Expect(1, 918000, '\P{idcontinue}', "");
Expect(0, 918000, '\P{^idcontinue}', "");
Expect(1, 917999, '\p{	_ID_CONTINUE}', "");
Expect(0, 917999, '\p{^	_ID_CONTINUE}', "");
Expect(0, 917999, '\P{	_ID_CONTINUE}', "");
Expect(1, 917999, '\P{^	_ID_CONTINUE}', "");
Expect(0, 918000, '\p{	_ID_CONTINUE}', "");
Expect(1, 918000, '\p{^	_ID_CONTINUE}', "");
Expect(1, 918000, '\P{	_ID_CONTINUE}', "");
Expect(0, 918000, '\P{^	_ID_CONTINUE}', "");
Error('\p{/a/ -Is_ID_continue}');
Error('\P{/a/ -Is_ID_continue}');
Expect(1, 917999, '\p{isidcontinue}', "");
Expect(0, 917999, '\p{^isidcontinue}', "");
Expect(0, 917999, '\P{isidcontinue}', "");
Expect(1, 917999, '\P{^isidcontinue}', "");
Expect(0, 918000, '\p{isidcontinue}', "");
Expect(1, 918000, '\p{^isidcontinue}', "");
Expect(1, 918000, '\P{isidcontinue}', "");
Expect(0, 918000, '\P{^isidcontinue}', "");
Expect(1, 917999, '\p{	-Is_ID_Continue}', "");
Expect(0, 917999, '\p{^	-Is_ID_Continue}', "");
Expect(0, 917999, '\P{	-Is_ID_Continue}', "");
Expect(1, 917999, '\P{^	-Is_ID_Continue}', "");
Expect(0, 918000, '\p{	-Is_ID_Continue}', "");
Expect(1, 918000, '\p{^	-Is_ID_Continue}', "");
Expect(1, 918000, '\P{	-Is_ID_Continue}', "");
Expect(0, 918000, '\P{^	-Is_ID_Continue}', "");
Error('\p{-	IDC:=}');
Error('\P{-	IDC:=}');
Expect(1, 917999, '\p{idc}', "");
Expect(0, 917999, '\p{^idc}', "");
Expect(0, 917999, '\P{idc}', "");
Expect(1, 917999, '\P{^idc}', "");
Expect(0, 918000, '\p{idc}', "");
Expect(1, 918000, '\p{^idc}', "");
Expect(1, 918000, '\P{idc}', "");
Expect(0, 918000, '\P{^idc}', "");
Error('\p{:= -IS_IDC}');
Error('\P{:= -IS_IDC}');
Expect(1, 917999, '\p{isidc}', "");
Expect(0, 917999, '\p{^isidc}', "");
Expect(0, 917999, '\P{isidc}', "");
Expect(1, 917999, '\P{^isidc}', "");
Expect(0, 918000, '\p{isidc}', "");
Expect(1, 918000, '\p{^isidc}', "");
Expect(1, 918000, '\P{isidc}', "");
Expect(0, 918000, '\P{^isidc}', "");
Expect(1, 917999, '\p{ -is_idc}', "");
Expect(0, 917999, '\p{^ -is_idc}', "");
Expect(0, 917999, '\P{ -is_idc}', "");
Expect(1, 917999, '\P{^ -is_idc}', "");
Expect(0, 918000, '\p{ -is_idc}', "");
Expect(1, 918000, '\p{^ -is_idc}', "");
Expect(1, 918000, '\P{ -is_idc}', "");
Expect(0, 918000, '\P{^ -is_idc}', "");
Error('\p{ 	ID_START/a/}');
Error('\P{ 	ID_START/a/}');
Expect(1, 195101, '\p{idstart}', "");
Expect(0, 195101, '\p{^idstart}', "");
Expect(0, 195101, '\P{idstart}', "");
Expect(1, 195101, '\P{^idstart}', "");
Expect(0, 195102, '\p{idstart}', "");
Expect(1, 195102, '\p{^idstart}', "");
Expect(1, 195102, '\P{idstart}', "");
Expect(0, 195102, '\P{^idstart}', "");
Expect(1, 195101, '\p{-ID_Start}', "");
Expect(0, 195101, '\p{^-ID_Start}', "");
Expect(0, 195101, '\P{-ID_Start}', "");
Expect(1, 195101, '\P{^-ID_Start}', "");
Expect(0, 195102, '\p{-ID_Start}', "");
Expect(1, 195102, '\p{^-ID_Start}', "");
Expect(1, 195102, '\P{-ID_Start}', "");
Expect(0, 195102, '\P{^-ID_Start}', "");
Error('\p{/a/is_ID_START}');
Error('\P{/a/is_ID_START}');
Expect(1, 195101, '\p{isidstart}', "");
Expect(0, 195101, '\p{^isidstart}', "");
Expect(0, 195101, '\P{isidstart}', "");
Expect(1, 195101, '\P{^isidstart}', "");
Expect(0, 195102, '\p{isidstart}', "");
Expect(1, 195102, '\p{^isidstart}', "");
Expect(1, 195102, '\P{isidstart}', "");
Expect(0, 195102, '\P{^isidstart}', "");
Expect(1, 195101, '\p{ Is_ID_Start}', "");
Expect(0, 195101, '\p{^ Is_ID_Start}', "");
Expect(0, 195101, '\P{ Is_ID_Start}', "");
Expect(1, 195101, '\P{^ Is_ID_Start}', "");
Expect(0, 195102, '\p{ Is_ID_Start}', "");
Expect(1, 195102, '\p{^ Is_ID_Start}', "");
Expect(1, 195102, '\P{ Is_ID_Start}', "");
Expect(0, 195102, '\P{^ Is_ID_Start}', "");
Error('\p{_ IDS:=}');
Error('\P{_ IDS:=}');
Expect(1, 195101, '\p{ids}', "");
Expect(0, 195101, '\p{^ids}', "");
Expect(0, 195101, '\P{ids}', "");
Expect(1, 195101, '\P{^ids}', "");
Expect(0, 195102, '\p{ids}', "");
Expect(1, 195102, '\p{^ids}', "");
Expect(1, 195102, '\P{ids}', "");
Expect(0, 195102, '\P{^ids}', "");
Expect(1, 195101, '\p{	IDS}', "");
Expect(0, 195101, '\p{^	IDS}', "");
Expect(0, 195101, '\P{	IDS}', "");
Expect(1, 195101, '\P{^	IDS}', "");
Expect(0, 195102, '\p{	IDS}', "");
Expect(1, 195102, '\p{^	IDS}', "");
Expect(1, 195102, '\P{	IDS}', "");
Expect(0, 195102, '\P{^	IDS}', "");
Error('\p{-is_IDS/a/}');
Error('\P{-is_IDS/a/}');
Expect(1, 195101, '\p{isids}', "");
Expect(0, 195101, '\p{^isids}', "");
Expect(0, 195101, '\P{isids}', "");
Expect(1, 195101, '\P{^isids}', "");
Expect(0, 195102, '\p{isids}', "");
Expect(1, 195102, '\p{^isids}', "");
Expect(1, 195102, '\P{isids}', "");
Expect(0, 195102, '\P{^isids}', "");
Expect(1, 195101, '\p{Is_IDS}', "");
Expect(0, 195101, '\p{^Is_IDS}', "");
Expect(0, 195101, '\P{Is_IDS}', "");
Expect(1, 195101, '\P{^Is_IDS}', "");
Expect(0, 195102, '\p{Is_IDS}', "");
Expect(1, 195102, '\p{^Is_IDS}', "");
Expect(1, 195102, '\P{Is_IDS}', "");
Expect(0, 195102, '\P{^Is_IDS}', "");
Error('\p{	_IPA_Extensions:=}');
Error('\P{	_IPA_Extensions:=}');
Expect(1, 687, '\p{ipaextensions}', "");
Expect(0, 687, '\p{^ipaextensions}', "");
Expect(0, 687, '\P{ipaextensions}', "");
Expect(1, 687, '\P{^ipaextensions}', "");
Expect(0, 688, '\p{ipaextensions}', "");
Expect(1, 688, '\p{^ipaextensions}', "");
Expect(1, 688, '\P{ipaextensions}', "");
Expect(0, 688, '\P{^ipaextensions}', "");
Expect(1, 687, '\p{	_IPA_Extensions}', "");
Expect(0, 687, '\p{^	_IPA_Extensions}', "");
Expect(0, 687, '\P{	_IPA_Extensions}', "");
Expect(1, 687, '\P{^	_IPA_Extensions}', "");
Expect(0, 688, '\p{	_IPA_Extensions}', "");
Expect(1, 688, '\p{^	_IPA_Extensions}', "");
Expect(1, 688, '\P{	_IPA_Extensions}', "");
Expect(0, 688, '\P{^	_IPA_Extensions}', "");
Error('\p{ _is_IPA_extensions:=}');
Error('\P{ _is_IPA_extensions:=}');
Expect(1, 687, '\p{isipaextensions}', "");
Expect(0, 687, '\p{^isipaextensions}', "");
Expect(0, 687, '\P{isipaextensions}', "");
Expect(1, 687, '\P{^isipaextensions}', "");
Expect(0, 688, '\p{isipaextensions}', "");
Expect(1, 688, '\p{^isipaextensions}', "");
Expect(1, 688, '\P{isipaextensions}', "");
Expect(0, 688, '\P{^isipaextensions}', "");
Expect(1, 687, '\p{	is_IPA_EXTENSIONS}', "");
Expect(0, 687, '\p{^	is_IPA_EXTENSIONS}', "");
Expect(0, 687, '\P{	is_IPA_EXTENSIONS}', "");
Expect(1, 687, '\P{^	is_IPA_EXTENSIONS}', "");
Expect(0, 688, '\p{	is_IPA_EXTENSIONS}', "");
Expect(1, 688, '\p{^	is_IPA_EXTENSIONS}', "");
Expect(1, 688, '\P{	is_IPA_EXTENSIONS}', "");
Expect(0, 688, '\P{^	is_IPA_EXTENSIONS}', "");
Error('\p{_In_IPA_Extensions:=}');
Error('\P{_In_IPA_Extensions:=}');
Expect(1, 687, '\p{inipaextensions}', "");
Expect(0, 687, '\p{^inipaextensions}', "");
Expect(0, 687, '\P{inipaextensions}', "");
Expect(1, 687, '\P{^inipaextensions}', "");
Expect(0, 688, '\p{inipaextensions}', "");
Expect(1, 688, '\p{^inipaextensions}', "");
Expect(1, 688, '\P{inipaextensions}', "");
Expect(0, 688, '\P{^inipaextensions}', "");
Expect(1, 687, '\p{	 In_IPA_extensions}', "");
Expect(0, 687, '\p{^	 In_IPA_extensions}', "");
Expect(0, 687, '\P{	 In_IPA_extensions}', "");
Expect(1, 687, '\P{^	 In_IPA_extensions}', "");
Expect(0, 688, '\p{	 In_IPA_extensions}', "");
Expect(1, 688, '\p{^	 In_IPA_extensions}', "");
Expect(1, 688, '\P{	 In_IPA_extensions}', "");
Expect(0, 688, '\P{^	 In_IPA_extensions}', "");
Error('\p{:=IPA_EXT}');
Error('\P{:=IPA_EXT}');
Expect(1, 687, '\p{ipaext}', "");
Expect(0, 687, '\p{^ipaext}', "");
Expect(0, 687, '\P{ipaext}', "");
Expect(1, 687, '\P{^ipaext}', "");
Expect(0, 688, '\p{ipaext}', "");
Expect(1, 688, '\p{^ipaext}', "");
Expect(1, 688, '\P{ipaext}', "");
Expect(0, 688, '\P{^ipaext}', "");
Expect(1, 687, '\p{	_ipa_ext}', "");
Expect(0, 687, '\p{^	_ipa_ext}', "");
Expect(0, 687, '\P{	_ipa_ext}', "");
Expect(1, 687, '\P{^	_ipa_ext}', "");
Expect(0, 688, '\p{	_ipa_ext}', "");
Expect(1, 688, '\p{^	_ipa_ext}', "");
Expect(1, 688, '\P{	_ipa_ext}', "");
Expect(0, 688, '\P{^	_ipa_ext}', "");
Error('\p{:= -is_IPA_Ext}');
Error('\P{:= -is_IPA_Ext}');
Expect(1, 687, '\p{isipaext}', "");
Expect(0, 687, '\p{^isipaext}', "");
Expect(0, 687, '\P{isipaext}', "");
Expect(1, 687, '\P{^isipaext}', "");
Expect(0, 688, '\p{isipaext}', "");
Expect(1, 688, '\p{^isipaext}', "");
Expect(1, 688, '\P{isipaext}', "");
Expect(0, 688, '\P{^isipaext}', "");
Expect(1, 687, '\p{		IS_ipa_Ext}', "");
Expect(0, 687, '\p{^		IS_ipa_Ext}', "");
Expect(0, 687, '\P{		IS_ipa_Ext}', "");
Expect(1, 687, '\P{^		IS_ipa_Ext}', "");
Expect(0, 688, '\p{		IS_ipa_Ext}', "");
Expect(1, 688, '\p{^		IS_ipa_Ext}', "");
Expect(1, 688, '\P{		IS_ipa_Ext}', "");
Expect(0, 688, '\P{^		IS_ipa_Ext}', "");
Error('\p{-	In_IPA_Ext/a/}');
Error('\P{-	In_IPA_Ext/a/}');
Expect(1, 687, '\p{inipaext}', "");
Expect(0, 687, '\p{^inipaext}', "");
Expect(0, 687, '\P{inipaext}', "");
Expect(1, 687, '\P{^inipaext}', "");
Expect(0, 688, '\p{inipaext}', "");
Expect(1, 688, '\p{^inipaext}', "");
Expect(1, 688, '\P{inipaext}', "");
Expect(0, 688, '\P{^inipaext}', "");
Expect(1, 687, '\p{	_in_IPA_EXT}', "");
Expect(0, 687, '\p{^	_in_IPA_EXT}', "");
Expect(0, 687, '\P{	_in_IPA_EXT}', "");
Expect(1, 687, '\P{^	_in_IPA_EXT}', "");
Expect(0, 688, '\p{	_in_IPA_EXT}', "");
Expect(1, 688, '\p{^	_in_IPA_EXT}', "");
Expect(1, 688, '\P{	_in_IPA_EXT}', "");
Expect(0, 688, '\P{^	_in_IPA_EXT}', "");
Error('\p{	Ideographic/a/}');
Error('\P{	Ideographic/a/}');
Expect(1, 195101, '\p{ideographic}', "");
Expect(0, 195101, '\p{^ideographic}', "");
Expect(0, 195101, '\P{ideographic}', "");
Expect(1, 195101, '\P{^ideographic}', "");
Expect(0, 195102, '\p{ideographic}', "");
Expect(1, 195102, '\p{^ideographic}', "");
Expect(1, 195102, '\P{ideographic}', "");
Expect(0, 195102, '\P{^ideographic}', "");
Expect(1, 195101, '\p{_-Ideographic}', "");
Expect(0, 195101, '\p{^_-Ideographic}', "");
Expect(0, 195101, '\P{_-Ideographic}', "");
Expect(1, 195101, '\P{^_-Ideographic}', "");
Expect(0, 195102, '\p{_-Ideographic}', "");
Expect(1, 195102, '\p{^_-Ideographic}', "");
Expect(1, 195102, '\P{_-Ideographic}', "");
Expect(0, 195102, '\P{^_-Ideographic}', "");
Error('\p{/a/ _Is_Ideographic}');
Error('\P{/a/ _Is_Ideographic}');
Expect(1, 195101, '\p{isideographic}', "");
Expect(0, 195101, '\p{^isideographic}', "");
Expect(0, 195101, '\P{isideographic}', "");
Expect(1, 195101, '\P{^isideographic}', "");
Expect(0, 195102, '\p{isideographic}', "");
Expect(1, 195102, '\p{^isideographic}', "");
Expect(1, 195102, '\P{isideographic}', "");
Expect(0, 195102, '\P{^isideographic}', "");
Expect(1, 195101, '\p{	_IS_Ideographic}', "");
Expect(0, 195101, '\p{^	_IS_Ideographic}', "");
Expect(0, 195101, '\P{	_IS_Ideographic}', "");
Expect(1, 195101, '\P{^	_IS_Ideographic}', "");
Expect(0, 195102, '\p{	_IS_Ideographic}', "");
Expect(1, 195102, '\p{^	_IS_Ideographic}', "");
Expect(1, 195102, '\P{	_IS_Ideographic}', "");
Expect(0, 195102, '\P{^	_IS_Ideographic}', "");
Error('\p{--IDEO/a/}');
Error('\P{--IDEO/a/}');
Expect(1, 195101, '\p{ideo}', "");
Expect(0, 195101, '\p{^ideo}', "");
Expect(0, 195101, '\P{ideo}', "");
Expect(1, 195101, '\P{^ideo}', "");
Expect(0, 195102, '\p{ideo}', "");
Expect(1, 195102, '\p{^ideo}', "");
Expect(1, 195102, '\P{ideo}', "");
Expect(0, 195102, '\P{^ideo}', "");
Expect(1, 195101, '\p{_ Ideo}', "");
Expect(0, 195101, '\p{^_ Ideo}', "");
Expect(0, 195101, '\P{_ Ideo}', "");
Expect(1, 195101, '\P{^_ Ideo}', "");
Expect(0, 195102, '\p{_ Ideo}', "");
Expect(1, 195102, '\p{^_ Ideo}', "");
Expect(1, 195102, '\P{_ Ideo}', "");
Expect(0, 195102, '\P{^_ Ideo}', "");
Error('\p{-IS_ideo:=}');
Error('\P{-IS_ideo:=}');
Expect(1, 195101, '\p{isideo}', "");
Expect(0, 195101, '\p{^isideo}', "");
Expect(0, 195101, '\P{isideo}', "");
Expect(1, 195101, '\P{^isideo}', "");
Expect(0, 195102, '\p{isideo}', "");
Expect(1, 195102, '\p{^isideo}', "");
Expect(1, 195102, '\P{isideo}', "");
Expect(0, 195102, '\P{^isideo}', "");
Expect(1, 195101, '\p{-	IS_Ideo}', "");
Expect(0, 195101, '\p{^-	IS_Ideo}', "");
Expect(0, 195101, '\P{-	IS_Ideo}', "");
Expect(1, 195101, '\P{^-	IS_Ideo}', "");
Expect(0, 195102, '\p{-	IS_Ideo}', "");
Expect(1, 195102, '\p{^-	IS_Ideo}', "");
Expect(1, 195102, '\P{-	IS_Ideo}', "");
Expect(0, 195102, '\P{^-	IS_Ideo}', "");
Error('\p{-Ideographic_description_CHARACTERS/a/}');
Error('\P{-Ideographic_description_CHARACTERS/a/}');
Expect(1, 12287, '\p{ideographicdescriptioncharacters}', "");
Expect(0, 12287, '\p{^ideographicdescriptioncharacters}', "");
Expect(0, 12287, '\P{ideographicdescriptioncharacters}', "");
Expect(1, 12287, '\P{^ideographicdescriptioncharacters}', "");
Expect(0, 12288, '\p{ideographicdescriptioncharacters}', "");
Expect(1, 12288, '\p{^ideographicdescriptioncharacters}', "");
Expect(1, 12288, '\P{ideographicdescriptioncharacters}', "");
Expect(0, 12288, '\P{^ideographicdescriptioncharacters}', "");
Expect(1, 12287, '\p{ Ideographic_description_Characters}', "");
Expect(0, 12287, '\p{^ Ideographic_description_Characters}', "");
Expect(0, 12287, '\P{ Ideographic_description_Characters}', "");
Expect(1, 12287, '\P{^ Ideographic_description_Characters}', "");
Expect(0, 12288, '\p{ Ideographic_description_Characters}', "");
Expect(1, 12288, '\p{^ Ideographic_description_Characters}', "");
Expect(1, 12288, '\P{ Ideographic_description_Characters}', "");
Expect(0, 12288, '\P{^ Ideographic_description_Characters}', "");
Error('\p{:= _IS_IDEOGRAPHIC_description_CHARACTERS}');
Error('\P{:= _IS_IDEOGRAPHIC_description_CHARACTERS}');
Expect(1, 12287, '\p{isideographicdescriptioncharacters}', "");
Expect(0, 12287, '\p{^isideographicdescriptioncharacters}', "");
Expect(0, 12287, '\P{isideographicdescriptioncharacters}', "");
Expect(1, 12287, '\P{^isideographicdescriptioncharacters}', "");
Expect(0, 12288, '\p{isideographicdescriptioncharacters}', "");
Expect(1, 12288, '\p{^isideographicdescriptioncharacters}', "");
Expect(1, 12288, '\P{isideographicdescriptioncharacters}', "");
Expect(0, 12288, '\P{^isideographicdescriptioncharacters}', "");
Expect(1, 12287, '\p{	_IS_IDEOGRAPHIC_description_characters}', "");
Expect(0, 12287, '\p{^	_IS_IDEOGRAPHIC_description_characters}', "");
Expect(0, 12287, '\P{	_IS_IDEOGRAPHIC_description_characters}', "");
Expect(1, 12287, '\P{^	_IS_IDEOGRAPHIC_description_characters}', "");
Expect(0, 12288, '\p{	_IS_IDEOGRAPHIC_description_characters}', "");
Expect(1, 12288, '\p{^	_IS_IDEOGRAPHIC_description_characters}', "");
Expect(1, 12288, '\P{	_IS_IDEOGRAPHIC_description_characters}', "");
Expect(0, 12288, '\P{^	_IS_IDEOGRAPHIC_description_characters}', "");
Error('\p{_:=in_ideographic_description_CHARACTERS}');
Error('\P{_:=in_ideographic_description_CHARACTERS}');
Expect(1, 12287, '\p{inideographicdescriptioncharacters}', "");
Expect(0, 12287, '\p{^inideographicdescriptioncharacters}', "");
Expect(0, 12287, '\P{inideographicdescriptioncharacters}', "");
Expect(1, 12287, '\P{^inideographicdescriptioncharacters}', "");
Expect(0, 12288, '\p{inideographicdescriptioncharacters}', "");
Expect(1, 12288, '\p{^inideographicdescriptioncharacters}', "");
Expect(1, 12288, '\P{inideographicdescriptioncharacters}', "");
Expect(0, 12288, '\P{^inideographicdescriptioncharacters}', "");
Expect(1, 12287, '\p{ 	in_Ideographic_Description_Characters}', "");
Expect(0, 12287, '\p{^ 	in_Ideographic_Description_Characters}', "");
Expect(0, 12287, '\P{ 	in_Ideographic_Description_Characters}', "");
Expect(1, 12287, '\P{^ 	in_Ideographic_Description_Characters}', "");
Expect(0, 12288, '\p{ 	in_Ideographic_Description_Characters}', "");
Expect(1, 12288, '\p{^ 	in_Ideographic_Description_Characters}', "");
Expect(1, 12288, '\P{ 	in_Ideographic_Description_Characters}', "");
Expect(0, 12288, '\P{^ 	in_Ideographic_Description_Characters}', "");
Error('\p{/a/_	In_IDC}');
Error('\P{/a/_	In_IDC}');
Expect(1, 12287, '\p{inidc}', "");
Expect(0, 12287, '\p{^inidc}', "");
Expect(0, 12287, '\P{inidc}', "");
Expect(1, 12287, '\P{^inidc}', "");
Expect(0, 12288, '\p{inidc}', "");
Expect(1, 12288, '\p{^inidc}', "");
Expect(1, 12288, '\P{inidc}', "");
Expect(0, 12288, '\P{^inidc}', "");
Expect(1, 12287, '\p{In_idc}', "");
Expect(0, 12287, '\p{^In_idc}', "");
Expect(0, 12287, '\P{In_idc}', "");
Expect(1, 12287, '\P{^In_idc}', "");
Expect(0, 12288, '\p{In_idc}', "");
Expect(1, 12288, '\p{^In_idc}', "");
Expect(1, 12288, '\P{In_idc}', "");
Expect(0, 12288, '\P{^In_idc}', "");
Error('\p{:=__imperial_ARAMAIC}');
Error('\P{:=__imperial_ARAMAIC}');
Expect(1, 67679, '\p{imperialaramaic}', "");
Expect(0, 67679, '\p{^imperialaramaic}', "");
Expect(0, 67679, '\P{imperialaramaic}', "");
Expect(1, 67679, '\P{^imperialaramaic}', "");
Expect(0, 67680, '\p{imperialaramaic}', "");
Expect(1, 67680, '\p{^imperialaramaic}', "");
Expect(1, 67680, '\P{imperialaramaic}', "");
Expect(0, 67680, '\P{^imperialaramaic}', "");
Expect(1, 67679, '\p{-IMPERIAL_aramaic}', "");
Expect(0, 67679, '\p{^-IMPERIAL_aramaic}', "");
Expect(0, 67679, '\P{-IMPERIAL_aramaic}', "");
Expect(1, 67679, '\P{^-IMPERIAL_aramaic}', "");
Expect(0, 67680, '\p{-IMPERIAL_aramaic}', "");
Expect(1, 67680, '\p{^-IMPERIAL_aramaic}', "");
Expect(1, 67680, '\P{-IMPERIAL_aramaic}', "");
Expect(0, 67680, '\P{^-IMPERIAL_aramaic}', "");
Error('\p{-:=Is_imperial_Aramaic}');
Error('\P{-:=Is_imperial_Aramaic}');
Expect(1, 67679, '\p{isimperialaramaic}', "");
Expect(0, 67679, '\p{^isimperialaramaic}', "");
Expect(0, 67679, '\P{isimperialaramaic}', "");
Expect(1, 67679, '\P{^isimperialaramaic}', "");
Expect(0, 67680, '\p{isimperialaramaic}', "");
Expect(1, 67680, '\p{^isimperialaramaic}', "");
Expect(1, 67680, '\P{isimperialaramaic}', "");
Expect(0, 67680, '\P{^isimperialaramaic}', "");
Expect(1, 67679, '\p{ -is_Imperial_Aramaic}', "");
Expect(0, 67679, '\p{^ -is_Imperial_Aramaic}', "");
Expect(0, 67679, '\P{ -is_Imperial_Aramaic}', "");
Expect(1, 67679, '\P{^ -is_Imperial_Aramaic}', "");
Expect(0, 67680, '\p{ -is_Imperial_Aramaic}', "");
Expect(1, 67680, '\p{^ -is_Imperial_Aramaic}', "");
Expect(1, 67680, '\P{ -is_Imperial_Aramaic}', "");
Expect(0, 67680, '\P{^ -is_Imperial_Aramaic}', "");
Error('\p{:=_Armi}');
Error('\P{:=_Armi}');
Expect(1, 67679, '\p{armi}', "");
Expect(0, 67679, '\p{^armi}', "");
Expect(0, 67679, '\P{armi}', "");
Expect(1, 67679, '\P{^armi}', "");
Expect(0, 67680, '\p{armi}', "");
Expect(1, 67680, '\p{^armi}', "");
Expect(1, 67680, '\P{armi}', "");
Expect(0, 67680, '\P{^armi}', "");
Expect(1, 67679, '\p{ ARMI}', "");
Expect(0, 67679, '\p{^ ARMI}', "");
Expect(0, 67679, '\P{ ARMI}', "");
Expect(1, 67679, '\P{^ ARMI}', "");
Expect(0, 67680, '\p{ ARMI}', "");
Expect(1, 67680, '\p{^ ARMI}', "");
Expect(1, 67680, '\P{ ARMI}', "");
Expect(0, 67680, '\P{^ ARMI}', "");
Error('\p{-/a/Is_ARMI}');
Error('\P{-/a/Is_ARMI}');
Expect(1, 67679, '\p{isarmi}', "");
Expect(0, 67679, '\p{^isarmi}', "");
Expect(0, 67679, '\P{isarmi}', "");
Expect(1, 67679, '\P{^isarmi}', "");
Expect(0, 67680, '\p{isarmi}', "");
Expect(1, 67680, '\p{^isarmi}', "");
Expect(1, 67680, '\P{isarmi}', "");
Expect(0, 67680, '\P{^isarmi}', "");
Expect(1, 67679, '\p{	Is_Armi}', "");
Expect(0, 67679, '\p{^	Is_Armi}', "");
Expect(0, 67679, '\P{	Is_Armi}', "");
Expect(1, 67679, '\P{^	Is_Armi}', "");
Expect(0, 67680, '\p{	Is_Armi}', "");
Expect(1, 67680, '\p{^	Is_Armi}', "");
Expect(1, 67680, '\P{	Is_Armi}', "");
Expect(0, 67680, '\P{^	Is_Armi}', "");
Error('\p{/a/- In_Arabic}');
Error('\P{/a/- In_Arabic}');
Expect(1, 1791, '\p{inarabic}', "");
Expect(0, 1791, '\p{^inarabic}', "");
Expect(0, 1791, '\P{inarabic}', "");
Expect(1, 1791, '\P{^inarabic}', "");
Expect(0, 1792, '\p{inarabic}', "");
Expect(1, 1792, '\p{^inarabic}', "");
Expect(1, 1792, '\P{inarabic}', "");
Expect(0, 1792, '\P{^inarabic}', "");
Expect(1, 1791, '\p{--In_arabic}', "");
Expect(0, 1791, '\p{^--In_arabic}', "");
Expect(0, 1791, '\P{--In_arabic}', "");
Expect(1, 1791, '\P{^--In_arabic}', "");
Expect(0, 1792, '\p{--In_arabic}', "");
Expect(1, 1792, '\p{^--In_arabic}', "");
Expect(1, 1792, '\P{--In_arabic}', "");
Expect(0, 1792, '\P{^--In_arabic}', "");
Error('\p{:=__in_Arabic}');
Error('\P{:=__in_Arabic}');
Expect(1, 1791, '\p{-_IN_Arabic}', "");
Expect(0, 1791, '\p{^-_IN_Arabic}', "");
Expect(0, 1791, '\P{-_IN_Arabic}', "");
Expect(1, 1791, '\P{^-_IN_Arabic}', "");
Expect(0, 1792, '\p{-_IN_Arabic}', "");
Expect(1, 1792, '\p{^-_IN_Arabic}', "");
Expect(1, 1792, '\P{-_IN_Arabic}', "");
Expect(0, 1792, '\P{^-_IN_Arabic}', "");
Error('\p{:=-IN_ARMENIAN}');
Error('\P{:=-IN_ARMENIAN}');
Expect(1, 1423, '\p{inarmenian}', "");
Expect(0, 1423, '\p{^inarmenian}', "");
Expect(0, 1423, '\P{inarmenian}', "");
Expect(1, 1423, '\P{^inarmenian}', "");
Expect(0, 1424, '\p{inarmenian}', "");
Expect(1, 1424, '\p{^inarmenian}', "");
Expect(1, 1424, '\P{inarmenian}', "");
Expect(0, 1424, '\P{^inarmenian}', "");
Expect(1, 1423, '\p{__IN_armenian}', "");
Expect(0, 1423, '\p{^__IN_armenian}', "");
Expect(0, 1423, '\P{__IN_armenian}', "");
Expect(1, 1423, '\P{^__IN_armenian}', "");
Expect(0, 1424, '\p{__IN_armenian}', "");
Expect(1, 1424, '\p{^__IN_armenian}', "");
Expect(1, 1424, '\P{__IN_armenian}', "");
Expect(0, 1424, '\P{^__IN_armenian}', "");
Error('\p{-/a/IN_Armenian}');
Error('\P{-/a/IN_Armenian}');
Expect(1, 1423, '\p{_	In_Armenian}', "");
Expect(0, 1423, '\p{^_	In_Armenian}', "");
Expect(0, 1423, '\P{_	In_Armenian}', "");
Expect(1, 1423, '\P{^_	In_Armenian}', "");
Expect(0, 1424, '\p{_	In_Armenian}', "");
Expect(1, 1424, '\p{^_	In_Armenian}', "");
Expect(1, 1424, '\P{_	In_Armenian}', "");
Expect(0, 1424, '\P{^_	In_Armenian}', "");
Error('\p{-IN_Avestan:=}');
Error('\P{-IN_Avestan:=}');
Expect(1, 68415, '\p{inavestan}', "");
Expect(0, 68415, '\p{^inavestan}', "");
Expect(0, 68415, '\P{inavestan}', "");
Expect(1, 68415, '\P{^inavestan}', "");
Expect(0, 68416, '\p{inavestan}', "");
Expect(1, 68416, '\p{^inavestan}', "");
Expect(1, 68416, '\P{inavestan}', "");
Expect(0, 68416, '\P{^inavestan}', "");
Expect(1, 68415, '\p{ 	IN_Avestan}', "");
Expect(0, 68415, '\p{^ 	IN_Avestan}', "");
Expect(0, 68415, '\P{ 	IN_Avestan}', "");
Expect(1, 68415, '\P{^ 	IN_Avestan}', "");
Expect(0, 68416, '\p{ 	IN_Avestan}', "");
Expect(1, 68416, '\p{^ 	IN_Avestan}', "");
Expect(1, 68416, '\P{ 	IN_Avestan}', "");
Expect(0, 68416, '\P{^ 	IN_Avestan}', "");
Error('\p{:=-IN_avestan}');
Error('\P{:=-IN_avestan}');
Expect(1, 68415, '\p{- IN_avestan}', "");
Expect(0, 68415, '\p{^- IN_avestan}', "");
Expect(0, 68415, '\P{- IN_avestan}', "");
Expect(1, 68415, '\P{^- IN_avestan}', "");
Expect(0, 68416, '\p{- IN_avestan}', "");
Expect(1, 68416, '\p{^- IN_avestan}', "");
Expect(1, 68416, '\P{- IN_avestan}', "");
Expect(0, 68416, '\P{^- IN_avestan}', "");
Error('\p{/a/In_Balinese}');
Error('\P{/a/In_Balinese}');
Expect(1, 7039, '\p{inbalinese}', "");
Expect(0, 7039, '\p{^inbalinese}', "");
Expect(0, 7039, '\P{inbalinese}', "");
Expect(1, 7039, '\P{^inbalinese}', "");
Expect(0, 7040, '\p{inbalinese}', "");
Expect(1, 7040, '\p{^inbalinese}', "");
Expect(1, 7040, '\P{inbalinese}', "");
Expect(0, 7040, '\P{^inbalinese}', "");
Expect(1, 7039, '\p{-IN_Balinese}', "");
Expect(0, 7039, '\p{^-IN_Balinese}', "");
Expect(0, 7039, '\P{-IN_Balinese}', "");
Expect(1, 7039, '\P{^-IN_Balinese}', "");
Expect(0, 7040, '\p{-IN_Balinese}', "");
Expect(1, 7040, '\p{^-IN_Balinese}', "");
Expect(1, 7040, '\P{-IN_Balinese}', "");
Expect(0, 7040, '\P{^-IN_Balinese}', "");
Error('\p{/a/- In_Balinese}');
Error('\P{/a/- In_Balinese}');
Expect(1, 7039, '\p{_In_Balinese}', "");
Expect(0, 7039, '\p{^_In_Balinese}', "");
Expect(0, 7039, '\P{_In_Balinese}', "");
Expect(1, 7039, '\P{^_In_Balinese}', "");
Expect(0, 7040, '\p{_In_Balinese}', "");
Expect(1, 7040, '\p{^_In_Balinese}', "");
Expect(1, 7040, '\P{_In_Balinese}', "");
Expect(0, 7040, '\P{^_In_Balinese}', "");
Error('\p{	in_BAMUM:=}');
Error('\P{	in_BAMUM:=}');
Expect(1, 42751, '\p{inbamum}', "");
Expect(0, 42751, '\p{^inbamum}', "");
Expect(0, 42751, '\P{inbamum}', "");
Expect(1, 42751, '\P{^inbamum}', "");
Expect(0, 42752, '\p{inbamum}', "");
Expect(1, 42752, '\p{^inbamum}', "");
Expect(1, 42752, '\P{inbamum}', "");
Expect(0, 42752, '\P{^inbamum}', "");
Expect(1, 42751, '\p{		In_Bamum}', "");
Expect(0, 42751, '\p{^		In_Bamum}', "");
Expect(0, 42751, '\P{		In_Bamum}', "");
Expect(1, 42751, '\P{^		In_Bamum}', "");
Expect(0, 42752, '\p{		In_Bamum}', "");
Expect(1, 42752, '\p{^		In_Bamum}', "");
Expect(1, 42752, '\P{		In_Bamum}', "");
Expect(0, 42752, '\P{^		In_Bamum}', "");
Error('\p{ :=in_BAMUM}');
Error('\P{ :=in_BAMUM}');
Expect(1, 42751, '\p{	In_BAMUM}', "");
Expect(0, 42751, '\p{^	In_BAMUM}', "");
Expect(0, 42751, '\P{	In_BAMUM}', "");
Expect(1, 42751, '\P{^	In_BAMUM}', "");
Expect(0, 42752, '\p{	In_BAMUM}', "");
Expect(1, 42752, '\p{^	In_BAMUM}', "");
Expect(1, 42752, '\P{	In_BAMUM}', "");
Expect(0, 42752, '\P{^	In_BAMUM}', "");
Error('\p{_In_BATAK:=}');
Error('\P{_In_BATAK:=}');
Expect(1, 7167, '\p{inbatak}', "");
Expect(0, 7167, '\p{^inbatak}', "");
Expect(0, 7167, '\P{inbatak}', "");
Expect(1, 7167, '\P{^inbatak}', "");
Expect(0, 7168, '\p{inbatak}', "");
Expect(1, 7168, '\p{^inbatak}', "");
Expect(1, 7168, '\P{inbatak}', "");
Expect(0, 7168, '\P{^inbatak}', "");
Expect(1, 7167, '\p{ In_batak}', "");
Expect(0, 7167, '\p{^ In_batak}', "");
Expect(0, 7167, '\P{ In_batak}', "");
Expect(1, 7167, '\P{^ In_batak}', "");
Expect(0, 7168, '\p{ In_batak}', "");
Expect(1, 7168, '\p{^ In_batak}', "");
Expect(1, 7168, '\P{ In_batak}', "");
Expect(0, 7168, '\P{^ In_batak}', "");
Error('\p{/a/_-IN_BATAK}');
Error('\P{/a/_-IN_BATAK}');
Expect(1, 7167, '\p{_ In_Batak}', "");
Expect(0, 7167, '\p{^_ In_Batak}', "");
Expect(0, 7167, '\P{_ In_Batak}', "");
Expect(1, 7167, '\P{^_ In_Batak}', "");
Expect(0, 7168, '\p{_ In_Batak}', "");
Expect(1, 7168, '\p{^_ In_Batak}', "");
Expect(1, 7168, '\P{_ In_Batak}', "");
Expect(0, 7168, '\P{^_ In_Batak}', "");
Error('\p{_In_Bengali/a/}');
Error('\P{_In_Bengali/a/}');
Expect(1, 2559, '\p{inbengali}', "");
Expect(0, 2559, '\p{^inbengali}', "");
Expect(0, 2559, '\P{inbengali}', "");
Expect(1, 2559, '\P{^inbengali}', "");
Expect(0, 2560, '\p{inbengali}', "");
Expect(1, 2560, '\p{^inbengali}', "");
Expect(1, 2560, '\P{inbengali}', "");
Expect(0, 2560, '\P{^inbengali}', "");
Expect(1, 2559, '\p{--In_Bengali}', "");
Expect(0, 2559, '\p{^--In_Bengali}', "");
Expect(0, 2559, '\P{--In_Bengali}', "");
Expect(1, 2559, '\P{^--In_Bengali}', "");
Expect(0, 2560, '\p{--In_Bengali}', "");
Expect(1, 2560, '\p{^--In_Bengali}', "");
Expect(1, 2560, '\P{--In_Bengali}', "");
Expect(0, 2560, '\P{^--In_Bengali}', "");
Error('\p{/a/--In_bengali}');
Error('\P{/a/--In_bengali}');
Expect(1, 2559, '\p{__in_BENGALI}', "");
Expect(0, 2559, '\p{^__in_BENGALI}', "");
Expect(0, 2559, '\P{__in_BENGALI}', "");
Expect(1, 2559, '\P{^__in_BENGALI}', "");
Expect(0, 2560, '\p{__in_BENGALI}', "");
Expect(1, 2560, '\p{^__in_BENGALI}', "");
Expect(1, 2560, '\P{__in_BENGALI}', "");
Expect(0, 2560, '\P{^__in_BENGALI}', "");
Error('\p{/a/  IN_Bopomofo}');
Error('\P{/a/  IN_Bopomofo}');
Expect(1, 12591, '\p{inbopomofo}', "");
Expect(0, 12591, '\p{^inbopomofo}', "");
Expect(0, 12591, '\P{inbopomofo}', "");
Expect(1, 12591, '\P{^inbopomofo}', "");
Expect(0, 12592, '\p{inbopomofo}', "");
Expect(1, 12592, '\p{^inbopomofo}', "");
Expect(1, 12592, '\P{inbopomofo}', "");
Expect(0, 12592, '\P{^inbopomofo}', "");
Expect(1, 12591, '\p{	In_Bopomofo}', "");
Expect(0, 12591, '\p{^	In_Bopomofo}', "");
Expect(0, 12591, '\P{	In_Bopomofo}', "");
Expect(1, 12591, '\P{^	In_Bopomofo}', "");
Expect(0, 12592, '\p{	In_Bopomofo}', "");
Expect(1, 12592, '\p{^	In_Bopomofo}', "");
Expect(1, 12592, '\P{	In_Bopomofo}', "");
Expect(0, 12592, '\P{^	In_Bopomofo}', "");
Error('\p{:=	_in_Bopomofo}');
Error('\P{:=	_in_Bopomofo}');
Expect(1, 12591, '\p{--IN_Bopomofo}', "");
Expect(0, 12591, '\p{^--IN_Bopomofo}', "");
Expect(0, 12591, '\P{--IN_Bopomofo}', "");
Expect(1, 12591, '\P{^--IN_Bopomofo}', "");
Expect(0, 12592, '\p{--IN_Bopomofo}', "");
Expect(1, 12592, '\p{^--IN_Bopomofo}', "");
Expect(1, 12592, '\P{--IN_Bopomofo}', "");
Expect(0, 12592, '\P{^--IN_Bopomofo}', "");
Error('\p{:=--in_Brahmi}');
Error('\P{:=--in_Brahmi}');
Expect(1, 69759, '\p{inbrahmi}', "");
Expect(0, 69759, '\p{^inbrahmi}', "");
Expect(0, 69759, '\P{inbrahmi}', "");
Expect(1, 69759, '\P{^inbrahmi}', "");
Expect(0, 69760, '\p{inbrahmi}', "");
Expect(1, 69760, '\p{^inbrahmi}', "");
Expect(1, 69760, '\P{inbrahmi}', "");
Expect(0, 69760, '\P{^inbrahmi}', "");
Expect(1, 69759, '\p{		IN_brahmi}', "");
Expect(0, 69759, '\p{^		IN_brahmi}', "");
Expect(0, 69759, '\P{		IN_brahmi}', "");
Expect(1, 69759, '\P{^		IN_brahmi}', "");
Expect(0, 69760, '\p{		IN_brahmi}', "");
Expect(1, 69760, '\p{^		IN_brahmi}', "");
Expect(1, 69760, '\P{		IN_brahmi}', "");
Expect(0, 69760, '\P{^		IN_brahmi}', "");
Error('\p{_/a/IN_brahmi}');
Error('\P{_/a/IN_brahmi}');
Expect(1, 69759, '\p{ _In_BRAHMI}', "");
Expect(0, 69759, '\p{^ _In_BRAHMI}', "");
Expect(0, 69759, '\P{ _In_BRAHMI}', "");
Expect(1, 69759, '\P{^ _In_BRAHMI}', "");
Expect(0, 69760, '\p{ _In_BRAHMI}', "");
Expect(1, 69760, '\p{^ _In_BRAHMI}', "");
Expect(1, 69760, '\P{ _In_BRAHMI}', "");
Expect(0, 69760, '\P{^ _In_BRAHMI}', "");
Error('\p{	-in_Buginese:=}');
Error('\P{	-in_Buginese:=}');
Expect(1, 6687, '\p{inbuginese}', "");
Expect(0, 6687, '\p{^inbuginese}', "");
Expect(0, 6687, '\P{inbuginese}', "");
Expect(1, 6687, '\P{^inbuginese}', "");
Expect(0, 6688, '\p{inbuginese}', "");
Expect(1, 6688, '\p{^inbuginese}', "");
Expect(1, 6688, '\P{inbuginese}', "");
Expect(0, 6688, '\P{^inbuginese}', "");
Expect(1, 6687, '\p{_-In_BUGINESE}', "");
Expect(0, 6687, '\p{^_-In_BUGINESE}', "");
Expect(0, 6687, '\P{_-In_BUGINESE}', "");
Expect(1, 6687, '\P{^_-In_BUGINESE}', "");
Expect(0, 6688, '\p{_-In_BUGINESE}', "");
Expect(1, 6688, '\p{^_-In_BUGINESE}', "");
Expect(1, 6688, '\P{_-In_BUGINESE}', "");
Expect(0, 6688, '\P{^_-In_BUGINESE}', "");
Error('\p{:=_ In_buginese}');
Error('\P{:=_ In_buginese}');
Expect(1, 6687, '\p{_ In_Buginese}', "");
Expect(0, 6687, '\p{^_ In_Buginese}', "");
Expect(0, 6687, '\P{_ In_Buginese}', "");
Expect(1, 6687, '\P{^_ In_Buginese}', "");
Expect(0, 6688, '\p{_ In_Buginese}', "");
Expect(1, 6688, '\p{^_ In_Buginese}', "");
Expect(1, 6688, '\P{_ In_Buginese}', "");
Expect(0, 6688, '\P{^_ In_Buginese}', "");
Error('\p{	_IN_Buhid:=}');
Error('\P{	_IN_Buhid:=}');
Expect(1, 5983, '\p{inbuhid}', "");
Expect(0, 5983, '\p{^inbuhid}', "");
Expect(0, 5983, '\P{inbuhid}', "");
Expect(1, 5983, '\P{^inbuhid}', "");
Expect(0, 5984, '\p{inbuhid}', "");
Expect(1, 5984, '\p{^inbuhid}', "");
Expect(1, 5984, '\P{inbuhid}', "");
Expect(0, 5984, '\P{^inbuhid}', "");
Expect(1, 5983, '\p{ In_buhid}', "");
Expect(0, 5983, '\p{^ In_buhid}', "");
Expect(0, 5983, '\P{ In_buhid}', "");
Expect(1, 5983, '\P{^ In_buhid}', "");
Expect(0, 5984, '\p{ In_buhid}', "");
Expect(1, 5984, '\p{^ In_buhid}', "");
Expect(1, 5984, '\P{ In_buhid}', "");
Expect(0, 5984, '\P{^ In_buhid}', "");
Error('\p{:=	IN_BUHID}');
Error('\P{:=	IN_BUHID}');
Expect(1, 5983, '\p{		In_Buhid}', "");
Expect(0, 5983, '\p{^		In_Buhid}', "");
Expect(0, 5983, '\P{		In_Buhid}', "");
Expect(1, 5983, '\P{^		In_Buhid}', "");
Expect(0, 5984, '\p{		In_Buhid}', "");
Expect(1, 5984, '\p{^		In_Buhid}', "");
Expect(1, 5984, '\P{		In_Buhid}', "");
Expect(0, 5984, '\P{^		In_Buhid}', "");
Error('\p{-In_CARIAN/a/}');
Error('\P{-In_CARIAN/a/}');
Expect(1, 66271, '\p{incarian}', "");
Expect(0, 66271, '\p{^incarian}', "");
Expect(0, 66271, '\P{incarian}', "");
Expect(1, 66271, '\P{^incarian}', "");
Expect(0, 66272, '\p{incarian}', "");
Expect(1, 66272, '\p{^incarian}', "");
Expect(1, 66272, '\P{incarian}', "");
Expect(0, 66272, '\P{^incarian}', "");
Expect(1, 66271, '\p{_ In_CARIAN}', "");
Expect(0, 66271, '\p{^_ In_CARIAN}', "");
Expect(0, 66271, '\P{_ In_CARIAN}', "");
Expect(1, 66271, '\P{^_ In_CARIAN}', "");
Expect(0, 66272, '\p{_ In_CARIAN}', "");
Expect(1, 66272, '\p{^_ In_CARIAN}', "");
Expect(1, 66272, '\P{_ In_CARIAN}', "");
Expect(0, 66272, '\P{^_ In_CARIAN}', "");
Error('\p{	In_Carian/a/}');
Error('\P{	In_Carian/a/}');
Expect(1, 66271, '\p{-	In_carian}', "");
Expect(0, 66271, '\p{^-	In_carian}', "");
Expect(0, 66271, '\P{-	In_carian}', "");
Expect(1, 66271, '\P{^-	In_carian}', "");
Expect(0, 66272, '\p{-	In_carian}', "");
Expect(1, 66272, '\p{^-	In_carian}', "");
Expect(1, 66272, '\P{-	In_carian}', "");
Expect(0, 66272, '\P{^-	In_carian}', "");
Error('\p{-/a/In_Chakma}');
Error('\P{-/a/In_Chakma}');
Expect(1, 69967, '\p{inchakma}', "");
Expect(0, 69967, '\p{^inchakma}', "");
Expect(0, 69967, '\P{inchakma}', "");
Expect(1, 69967, '\P{^inchakma}', "");
Expect(0, 69968, '\p{inchakma}', "");
Expect(1, 69968, '\p{^inchakma}', "");
Expect(1, 69968, '\P{inchakma}', "");
Expect(0, 69968, '\P{^inchakma}', "");
Expect(1, 69967, '\p{ In_Chakma}', "");
Expect(0, 69967, '\p{^ In_Chakma}', "");
Expect(0, 69967, '\P{ In_Chakma}', "");
Expect(1, 69967, '\P{^ In_Chakma}', "");
Expect(0, 69968, '\p{ In_Chakma}', "");
Expect(1, 69968, '\p{^ In_Chakma}', "");
Expect(1, 69968, '\P{ In_Chakma}', "");
Expect(0, 69968, '\P{^ In_Chakma}', "");
Error('\p{/a/_IN_CHAKMA}');
Error('\P{/a/_IN_CHAKMA}');
Expect(1, 69967, '\p{__In_Chakma}', "");
Expect(0, 69967, '\p{^__In_Chakma}', "");
Expect(0, 69967, '\P{__In_Chakma}', "");
Expect(1, 69967, '\P{^__In_Chakma}', "");
Expect(0, 69968, '\p{__In_Chakma}', "");
Expect(1, 69968, '\p{^__In_Chakma}', "");
Expect(1, 69968, '\P{__In_Chakma}', "");
Expect(0, 69968, '\P{^__In_Chakma}', "");
Error('\p{/a/-In_Cham}');
Error('\P{/a/-In_Cham}');
Expect(1, 43615, '\p{incham}', "");
Expect(0, 43615, '\p{^incham}', "");
Expect(0, 43615, '\P{incham}', "");
Expect(1, 43615, '\P{^incham}', "");
Expect(0, 43616, '\p{incham}', "");
Expect(1, 43616, '\p{^incham}', "");
Expect(1, 43616, '\P{incham}', "");
Expect(0, 43616, '\P{^incham}', "");
Expect(1, 43615, '\p{	 In_CHAM}', "");
Expect(0, 43615, '\p{^	 In_CHAM}', "");
Expect(0, 43615, '\P{	 In_CHAM}', "");
Expect(1, 43615, '\P{^	 In_CHAM}', "");
Expect(0, 43616, '\p{	 In_CHAM}', "");
Expect(1, 43616, '\p{^	 In_CHAM}', "");
Expect(1, 43616, '\P{	 In_CHAM}', "");
Expect(0, 43616, '\P{^	 In_CHAM}', "");
Error('\p{:=- In_Cham}');
Error('\P{:=- In_Cham}');
Expect(1, 43615, '\p{		In_Cham}', "");
Expect(0, 43615, '\p{^		In_Cham}', "");
Expect(0, 43615, '\P{		In_Cham}', "");
Expect(1, 43615, '\P{^		In_Cham}', "");
Expect(0, 43616, '\p{		In_Cham}', "");
Expect(1, 43616, '\p{^		In_Cham}', "");
Expect(1, 43616, '\P{		In_Cham}', "");
Expect(0, 43616, '\P{^		In_Cham}', "");
Error('\p{ _In_Cherokee:=}');
Error('\P{ _In_Cherokee:=}');
Expect(1, 5119, '\p{incherokee}', "");
Expect(0, 5119, '\p{^incherokee}', "");
Expect(0, 5119, '\P{incherokee}', "");
Expect(1, 5119, '\P{^incherokee}', "");
Expect(0, 5120, '\p{incherokee}', "");
Expect(1, 5120, '\p{^incherokee}', "");
Expect(1, 5120, '\P{incherokee}', "");
Expect(0, 5120, '\P{^incherokee}', "");
Expect(1, 5119, '\p{	in_Cherokee}', "");
Expect(0, 5119, '\p{^	in_Cherokee}', "");
Expect(0, 5119, '\P{	in_Cherokee}', "");
Expect(1, 5119, '\P{^	in_Cherokee}', "");
Expect(0, 5120, '\p{	in_Cherokee}', "");
Expect(1, 5120, '\p{^	in_Cherokee}', "");
Expect(1, 5120, '\P{	in_Cherokee}', "");
Expect(0, 5120, '\P{^	in_Cherokee}', "");
Error('\p{ /a/IN_Cherokee}');
Error('\P{ /a/IN_Cherokee}');
Expect(1, 5119, '\p{	_In_cherokee}', "");
Expect(0, 5119, '\p{^	_In_cherokee}', "");
Expect(0, 5119, '\P{	_In_cherokee}', "");
Expect(1, 5119, '\P{^	_In_cherokee}', "");
Expect(0, 5120, '\p{	_In_cherokee}', "");
Expect(1, 5120, '\p{^	_In_cherokee}', "");
Expect(1, 5120, '\P{	_In_cherokee}', "");
Expect(0, 5120, '\P{^	_In_cherokee}', "");
Error('\p{-:=in_Coptic}');
Error('\P{-:=in_Coptic}');
Expect(1, 11519, '\p{incoptic}', "");
Expect(0, 11519, '\p{^incoptic}', "");
Expect(0, 11519, '\P{incoptic}', "");
Expect(1, 11519, '\P{^incoptic}', "");
Expect(0, 11520, '\p{incoptic}', "");
Expect(1, 11520, '\p{^incoptic}', "");
Expect(1, 11520, '\P{incoptic}', "");
Expect(0, 11520, '\P{^incoptic}', "");
Expect(1, 11519, '\p{	IN_COPTIC}', "");
Expect(0, 11519, '\p{^	IN_COPTIC}', "");
Expect(0, 11519, '\P{	IN_COPTIC}', "");
Expect(1, 11519, '\P{^	IN_COPTIC}', "");
Expect(0, 11520, '\p{	IN_COPTIC}', "");
Expect(1, 11520, '\p{^	IN_COPTIC}', "");
Expect(1, 11520, '\P{	IN_COPTIC}', "");
Expect(0, 11520, '\P{^	IN_COPTIC}', "");
Error('\p{/a/	 in_Coptic}');
Error('\P{/a/	 in_Coptic}');
Expect(1, 11519, '\p{-	in_Coptic}', "");
Expect(0, 11519, '\p{^-	in_Coptic}', "");
Expect(0, 11519, '\P{-	in_Coptic}', "");
Expect(1, 11519, '\P{^-	in_Coptic}', "");
Expect(0, 11520, '\p{-	in_Coptic}', "");
Expect(1, 11520, '\p{^-	in_Coptic}', "");
Expect(1, 11520, '\P{-	in_Coptic}', "");
Expect(0, 11520, '\P{^-	in_Coptic}', "");
Error('\p{/a/-	In_CUNEIFORM}');
Error('\P{/a/-	In_CUNEIFORM}');
Expect(1, 74751, '\p{incuneiform}', "");
Expect(0, 74751, '\p{^incuneiform}', "");
Expect(0, 74751, '\P{incuneiform}', "");
Expect(1, 74751, '\P{^incuneiform}', "");
Expect(0, 74752, '\p{incuneiform}', "");
Expect(1, 74752, '\p{^incuneiform}', "");
Expect(1, 74752, '\P{incuneiform}', "");
Expect(0, 74752, '\P{^incuneiform}', "");
Expect(1, 74751, '\p{_-In_cuneiform}', "");
Expect(0, 74751, '\p{^_-In_cuneiform}', "");
Expect(0, 74751, '\P{_-In_cuneiform}', "");
Expect(1, 74751, '\P{^_-In_cuneiform}', "");
Expect(0, 74752, '\p{_-In_cuneiform}', "");
Expect(1, 74752, '\p{^_-In_cuneiform}', "");
Expect(1, 74752, '\P{_-In_cuneiform}', "");
Expect(0, 74752, '\P{^_-In_cuneiform}', "");
Error('\p{ In_Cuneiform/a/}');
Error('\P{ In_Cuneiform/a/}');
Expect(1, 74751, '\p{  IN_CUNEIFORM}', "");
Expect(0, 74751, '\p{^  IN_CUNEIFORM}', "");
Expect(0, 74751, '\P{  IN_CUNEIFORM}', "");
Expect(1, 74751, '\P{^  IN_CUNEIFORM}', "");
Expect(0, 74752, '\p{  IN_CUNEIFORM}', "");
Expect(1, 74752, '\p{^  IN_CUNEIFORM}', "");
Expect(1, 74752, '\P{  IN_CUNEIFORM}', "");
Expect(0, 74752, '\P{^  IN_CUNEIFORM}', "");
Error('\p{-In_CYRILLIC/a/}');
Error('\P{-In_CYRILLIC/a/}');
Expect(1, 1279, '\p{incyrillic}', "");
Expect(0, 1279, '\p{^incyrillic}', "");
Expect(0, 1279, '\P{incyrillic}', "");
Expect(1, 1279, '\P{^incyrillic}', "");
Expect(0, 1280, '\p{incyrillic}', "");
Expect(1, 1280, '\p{^incyrillic}', "");
Expect(1, 1280, '\P{incyrillic}', "");
Expect(0, 1280, '\P{^incyrillic}', "");
Expect(1, 1279, '\p{		in_Cyrillic}', "");
Expect(0, 1279, '\p{^		in_Cyrillic}', "");
Expect(0, 1279, '\P{		in_Cyrillic}', "");
Expect(1, 1279, '\P{^		in_Cyrillic}', "");
Expect(0, 1280, '\p{		in_Cyrillic}', "");
Expect(1, 1280, '\p{^		in_Cyrillic}', "");
Expect(1, 1280, '\P{		in_Cyrillic}', "");
Expect(0, 1280, '\P{^		in_Cyrillic}', "");
Error('\p{ -in_CYRILLIC:=}');
Error('\P{ -in_CYRILLIC:=}');
Expect(1, 1279, '\p{_-In_CYRILLIC}', "");
Expect(0, 1279, '\p{^_-In_CYRILLIC}', "");
Expect(0, 1279, '\P{_-In_CYRILLIC}', "");
Expect(1, 1279, '\P{^_-In_CYRILLIC}', "");
Expect(0, 1280, '\p{_-In_CYRILLIC}', "");
Expect(1, 1280, '\p{^_-In_CYRILLIC}', "");
Expect(1, 1280, '\P{_-In_CYRILLIC}', "");
Expect(0, 1280, '\P{^_-In_CYRILLIC}', "");
Error('\p{/a/_	IN_deseret}');
Error('\P{/a/_	IN_deseret}');
Expect(1, 66639, '\p{indeseret}', "");
Expect(0, 66639, '\p{^indeseret}', "");
Expect(0, 66639, '\P{indeseret}', "");
Expect(1, 66639, '\P{^indeseret}', "");
Expect(0, 66640, '\p{indeseret}', "");
Expect(1, 66640, '\p{^indeseret}', "");
Expect(1, 66640, '\P{indeseret}', "");
Expect(0, 66640, '\P{^indeseret}', "");
Expect(1, 66639, '\p{_-in_deseret}', "");
Expect(0, 66639, '\p{^_-in_deseret}', "");
Expect(0, 66639, '\P{_-in_deseret}', "");
Expect(1, 66639, '\P{^_-in_deseret}', "");
Expect(0, 66640, '\p{_-in_deseret}', "");
Expect(1, 66640, '\p{^_-in_deseret}', "");
Expect(1, 66640, '\P{_-in_deseret}', "");
Expect(0, 66640, '\P{^_-in_deseret}', "");
Error('\p{_:=In_DESERET}');
Error('\P{_:=In_DESERET}');
Expect(1, 66639, '\p{ In_Deseret}', "");
Expect(0, 66639, '\p{^ In_Deseret}', "");
Expect(0, 66639, '\P{ In_Deseret}', "");
Expect(1, 66639, '\P{^ In_Deseret}', "");
Expect(0, 66640, '\p{ In_Deseret}', "");
Expect(1, 66640, '\p{^ In_Deseret}', "");
Expect(1, 66640, '\P{ In_Deseret}', "");
Expect(0, 66640, '\P{^ In_Deseret}', "");
Error('\p{:= 	in_DEVANAGARI}');
Error('\P{:= 	in_DEVANAGARI}');
Expect(1, 2431, '\p{indevanagari}', "");
Expect(0, 2431, '\p{^indevanagari}', "");
Expect(0, 2431, '\P{indevanagari}', "");
Expect(1, 2431, '\P{^indevanagari}', "");
Expect(0, 2432, '\p{indevanagari}', "");
Expect(1, 2432, '\p{^indevanagari}', "");
Expect(1, 2432, '\P{indevanagari}', "");
Expect(0, 2432, '\P{^indevanagari}', "");
Expect(1, 2431, '\p{- in_devanagari}', "");
Expect(0, 2431, '\p{^- in_devanagari}', "");
Expect(0, 2431, '\P{- in_devanagari}', "");
Expect(1, 2431, '\P{^- in_devanagari}', "");
Expect(0, 2432, '\p{- in_devanagari}', "");
Expect(1, 2432, '\p{^- in_devanagari}', "");
Expect(1, 2432, '\P{- in_devanagari}', "");
Expect(0, 2432, '\P{^- in_devanagari}', "");
Error('\p{__In_devanagari/a/}');
Error('\P{__In_devanagari/a/}');
Expect(1, 2431, '\p{_-in_Devanagari}', "");
Expect(0, 2431, '\p{^_-in_Devanagari}', "");
Expect(0, 2431, '\P{_-in_Devanagari}', "");
Expect(1, 2431, '\P{^_-in_Devanagari}', "");
Expect(0, 2432, '\p{_-in_Devanagari}', "");
Expect(1, 2432, '\p{^_-in_Devanagari}', "");
Expect(1, 2432, '\P{_-in_Devanagari}', "");
Expect(0, 2432, '\P{^_-in_Devanagari}', "");
Error('\p{	:=In_EGYPTIAN_hieroglyphs}');
Error('\P{	:=In_EGYPTIAN_hieroglyphs}');
Expect(1, 78895, '\p{inegyptianhieroglyphs}', "");
Expect(0, 78895, '\p{^inegyptianhieroglyphs}', "");
Expect(0, 78895, '\P{inegyptianhieroglyphs}', "");
Expect(1, 78895, '\P{^inegyptianhieroglyphs}', "");
Expect(0, 78896, '\p{inegyptianhieroglyphs}', "");
Expect(1, 78896, '\p{^inegyptianhieroglyphs}', "");
Expect(1, 78896, '\P{inegyptianhieroglyphs}', "");
Expect(0, 78896, '\P{^inegyptianhieroglyphs}', "");
Expect(1, 78895, '\p{ IN_EGYPTIAN_hieroglyphs}', "");
Expect(0, 78895, '\p{^ IN_EGYPTIAN_hieroglyphs}', "");
Expect(0, 78895, '\P{ IN_EGYPTIAN_hieroglyphs}', "");
Expect(1, 78895, '\P{^ IN_EGYPTIAN_hieroglyphs}', "");
Expect(0, 78896, '\p{ IN_EGYPTIAN_hieroglyphs}', "");
Expect(1, 78896, '\p{^ IN_EGYPTIAN_hieroglyphs}', "");
Expect(1, 78896, '\P{ IN_EGYPTIAN_hieroglyphs}', "");
Expect(0, 78896, '\P{^ IN_EGYPTIAN_hieroglyphs}', "");
Error('\p{	:=in_Egyptian_HIEROGLYPHS}');
Error('\P{	:=in_Egyptian_HIEROGLYPHS}');
Expect(1, 78895, '\p{_IN_Egyptian_HIEROGLYPHS}', "");
Expect(0, 78895, '\p{^_IN_Egyptian_HIEROGLYPHS}', "");
Expect(0, 78895, '\P{_IN_Egyptian_HIEROGLYPHS}', "");
Expect(1, 78895, '\P{^_IN_Egyptian_HIEROGLYPHS}', "");
Expect(0, 78896, '\p{_IN_Egyptian_HIEROGLYPHS}', "");
Expect(1, 78896, '\p{^_IN_Egyptian_HIEROGLYPHS}', "");
Expect(1, 78896, '\P{_IN_Egyptian_HIEROGLYPHS}', "");
Expect(0, 78896, '\P{^_IN_Egyptian_HIEROGLYPHS}', "");
Error('\p{	_In_ethiopic/a/}');
Error('\P{	_In_ethiopic/a/}');
Expect(1, 4991, '\p{inethiopic}', "");
Expect(0, 4991, '\p{^inethiopic}', "");
Expect(0, 4991, '\P{inethiopic}', "");
Expect(1, 4991, '\P{^inethiopic}', "");
Expect(0, 4992, '\p{inethiopic}', "");
Expect(1, 4992, '\p{^inethiopic}', "");
Expect(1, 4992, '\P{inethiopic}', "");
Expect(0, 4992, '\P{^inethiopic}', "");
Expect(1, 4991, '\p{	-In_ethiopic}', "");
Expect(0, 4991, '\p{^	-In_ethiopic}', "");
Expect(0, 4991, '\P{	-In_ethiopic}', "");
Expect(1, 4991, '\P{^	-In_ethiopic}', "");
Expect(0, 4992, '\p{	-In_ethiopic}', "");
Expect(1, 4992, '\p{^	-In_ethiopic}', "");
Expect(1, 4992, '\P{	-In_ethiopic}', "");
Expect(0, 4992, '\P{^	-In_ethiopic}', "");
Error('\p{:=	In_Ethiopic}');
Error('\P{:=	In_Ethiopic}');
Expect(1, 4991, '\p{ 	In_ETHIOPIC}', "");
Expect(0, 4991, '\p{^ 	In_ETHIOPIC}', "");
Expect(0, 4991, '\P{ 	In_ETHIOPIC}', "");
Expect(1, 4991, '\P{^ 	In_ETHIOPIC}', "");
Expect(0, 4992, '\p{ 	In_ETHIOPIC}', "");
Expect(1, 4992, '\p{^ 	In_ETHIOPIC}', "");
Expect(1, 4992, '\P{ 	In_ETHIOPIC}', "");
Expect(0, 4992, '\P{^ 	In_ETHIOPIC}', "");
Error('\p{	/a/in_Georgian}');
Error('\P{	/a/in_Georgian}');
Expect(1, 4351, '\p{ingeorgian}', "");
Expect(0, 4351, '\p{^ingeorgian}', "");
Expect(0, 4351, '\P{ingeorgian}', "");
Expect(1, 4351, '\P{^ingeorgian}', "");
Expect(0, 4352, '\p{ingeorgian}', "");
Expect(1, 4352, '\p{^ingeorgian}', "");
Expect(1, 4352, '\P{ingeorgian}', "");
Expect(0, 4352, '\P{^ingeorgian}', "");
Expect(1, 4351, '\p{	-In_Georgian}', "");
Expect(0, 4351, '\p{^	-In_Georgian}', "");
Expect(0, 4351, '\P{	-In_Georgian}', "");
Expect(1, 4351, '\P{^	-In_Georgian}', "");
Expect(0, 4352, '\p{	-In_Georgian}', "");
Expect(1, 4352, '\p{^	-In_Georgian}', "");
Expect(1, 4352, '\P{	-In_Georgian}', "");
Expect(0, 4352, '\P{^	-In_Georgian}', "");
Error('\p{_:=In_georgian}');
Error('\P{_:=In_georgian}');
Expect(1, 4351, '\p{_IN_GEORGIAN}', "");
Expect(0, 4351, '\p{^_IN_GEORGIAN}', "");
Expect(0, 4351, '\P{_IN_GEORGIAN}', "");
Expect(1, 4351, '\P{^_IN_GEORGIAN}', "");
Expect(0, 4352, '\p{_IN_GEORGIAN}', "");
Expect(1, 4352, '\p{^_IN_GEORGIAN}', "");
Expect(1, 4352, '\P{_IN_GEORGIAN}', "");
Expect(0, 4352, '\P{^_IN_GEORGIAN}', "");
Error('\p{-_In_Glagolitic/a/}');
Error('\P{-_In_Glagolitic/a/}');
Expect(1, 11359, '\p{inglagolitic}', "");
Expect(0, 11359, '\p{^inglagolitic}', "");
Expect(0, 11359, '\P{inglagolitic}', "");
Expect(1, 11359, '\P{^inglagolitic}', "");
Expect(0, 11360, '\p{inglagolitic}', "");
Expect(1, 11360, '\p{^inglagolitic}', "");
Expect(1, 11360, '\P{inglagolitic}', "");
Expect(0, 11360, '\P{^inglagolitic}', "");
Expect(1, 11359, '\p{	-in_glagolitic}', "");
Expect(0, 11359, '\p{^	-in_glagolitic}', "");
Expect(0, 11359, '\P{	-in_glagolitic}', "");
Expect(1, 11359, '\P{^	-in_glagolitic}', "");
Expect(0, 11360, '\p{	-in_glagolitic}', "");
Expect(1, 11360, '\p{^	-in_glagolitic}', "");
Expect(1, 11360, '\P{	-in_glagolitic}', "");
Expect(0, 11360, '\P{^	-in_glagolitic}', "");
Error('\p{-In_Glagolitic:=}');
Error('\P{-In_Glagolitic:=}');
Expect(1, 11359, '\p{_-in_Glagolitic}', "");
Expect(0, 11359, '\p{^_-in_Glagolitic}', "");
Expect(0, 11359, '\P{_-in_Glagolitic}', "");
Expect(1, 11359, '\P{^_-in_Glagolitic}', "");
Expect(0, 11360, '\p{_-in_Glagolitic}', "");
Expect(1, 11360, '\p{^_-in_Glagolitic}', "");
Expect(1, 11360, '\P{_-in_Glagolitic}', "");
Expect(0, 11360, '\P{^_-in_Glagolitic}', "");
Error('\p{ :=IN_gothic}');
Error('\P{ :=IN_gothic}');
Expect(1, 66383, '\p{ingothic}', "");
Expect(0, 66383, '\p{^ingothic}', "");
Expect(0, 66383, '\P{ingothic}', "");
Expect(1, 66383, '\P{^ingothic}', "");
Expect(0, 66384, '\p{ingothic}', "");
Expect(1, 66384, '\p{^ingothic}', "");
Expect(1, 66384, '\P{ingothic}', "");
Expect(0, 66384, '\P{^ingothic}', "");
Expect(1, 66383, '\p{	In_Gothic}', "");
Expect(0, 66383, '\p{^	In_Gothic}', "");
Expect(0, 66383, '\P{	In_Gothic}', "");
Expect(1, 66383, '\P{^	In_Gothic}', "");
Expect(0, 66384, '\p{	In_Gothic}', "");
Expect(1, 66384, '\p{^	In_Gothic}', "");
Expect(1, 66384, '\P{	In_Gothic}', "");
Expect(0, 66384, '\P{^	In_Gothic}', "");
Error('\p{_-In_GOTHIC:=}');
Error('\P{_-In_GOTHIC:=}');
Expect(1, 66383, '\p{ in_Gothic}', "");
Expect(0, 66383, '\p{^ in_Gothic}', "");
Expect(0, 66383, '\P{ in_Gothic}', "");
Expect(1, 66383, '\P{^ in_Gothic}', "");
Expect(0, 66384, '\p{ in_Gothic}', "");
Expect(1, 66384, '\p{^ in_Gothic}', "");
Expect(1, 66384, '\P{ in_Gothic}', "");
Expect(0, 66384, '\P{^ in_Gothic}', "");
Error('\p{_-In_GUJARATI/a/}');
Error('\P{_-In_GUJARATI/a/}');
Expect(1, 2815, '\p{ingujarati}', "");
Expect(0, 2815, '\p{^ingujarati}', "");
Expect(0, 2815, '\P{ingujarati}', "");
Expect(1, 2815, '\P{^ingujarati}', "");
Expect(0, 2816, '\p{ingujarati}', "");
Expect(1, 2816, '\p{^ingujarati}', "");
Expect(1, 2816, '\P{ingujarati}', "");
Expect(0, 2816, '\P{^ingujarati}', "");
Expect(1, 2815, '\p{_IN_GUJARATI}', "");
Expect(0, 2815, '\p{^_IN_GUJARATI}', "");
Expect(0, 2815, '\P{_IN_GUJARATI}', "");
Expect(1, 2815, '\P{^_IN_GUJARATI}', "");
Expect(0, 2816, '\p{_IN_GUJARATI}', "");
Expect(1, 2816, '\p{^_IN_GUJARATI}', "");
Expect(1, 2816, '\P{_IN_GUJARATI}', "");
Expect(0, 2816, '\P{^_IN_GUJARATI}', "");
Error('\p{ _In_Gujarati/a/}');
Error('\P{ _In_Gujarati/a/}');
Expect(1, 2815, '\p{	 In_GUJARATI}', "");
Expect(0, 2815, '\p{^	 In_GUJARATI}', "");
Expect(0, 2815, '\P{	 In_GUJARATI}', "");
Expect(1, 2815, '\P{^	 In_GUJARATI}', "");
Expect(0, 2816, '\p{	 In_GUJARATI}', "");
Expect(1, 2816, '\p{^	 In_GUJARATI}', "");
Expect(1, 2816, '\P{	 In_GUJARATI}', "");
Expect(0, 2816, '\P{^	 In_GUJARATI}', "");
Error('\p{	/a/In_gurmukhi}');
Error('\P{	/a/In_gurmukhi}');
Expect(1, 2687, '\p{ingurmukhi}', "");
Expect(0, 2687, '\p{^ingurmukhi}', "");
Expect(0, 2687, '\P{ingurmukhi}', "");
Expect(1, 2687, '\P{^ingurmukhi}', "");
Expect(0, 2688, '\p{ingurmukhi}', "");
Expect(1, 2688, '\p{^ingurmukhi}', "");
Expect(1, 2688, '\P{ingurmukhi}', "");
Expect(0, 2688, '\P{^ingurmukhi}', "");
Expect(1, 2687, '\p{- In_Gurmukhi}', "");
Expect(0, 2687, '\p{^- In_Gurmukhi}', "");
Expect(0, 2687, '\P{- In_Gurmukhi}', "");
Expect(1, 2687, '\P{^- In_Gurmukhi}', "");
Expect(0, 2688, '\p{- In_Gurmukhi}', "");
Expect(1, 2688, '\p{^- In_Gurmukhi}', "");
Expect(1, 2688, '\P{- In_Gurmukhi}', "");
Expect(0, 2688, '\P{^- In_Gurmukhi}', "");
Error('\p{/a/	 In_Gurmukhi}');
Error('\P{/a/	 In_Gurmukhi}');
Expect(1, 2687, '\p{ _In_gurmukhi}', "");
Expect(0, 2687, '\p{^ _In_gurmukhi}', "");
Expect(0, 2687, '\P{ _In_gurmukhi}', "");
Expect(1, 2687, '\P{^ _In_gurmukhi}', "");
Expect(0, 2688, '\p{ _In_gurmukhi}', "");
Expect(1, 2688, '\p{^ _In_gurmukhi}', "");
Expect(1, 2688, '\P{ _In_gurmukhi}', "");
Expect(0, 2688, '\P{^ _In_gurmukhi}', "");
Error('\p{:=--IN_Hanunoo}');
Error('\P{:=--IN_Hanunoo}');
Expect(1, 5951, '\p{inhanunoo}', "");
Expect(0, 5951, '\p{^inhanunoo}', "");
Expect(0, 5951, '\P{inhanunoo}', "");
Expect(1, 5951, '\P{^inhanunoo}', "");
Expect(0, 5952, '\p{inhanunoo}', "");
Expect(1, 5952, '\p{^inhanunoo}', "");
Expect(1, 5952, '\P{inhanunoo}', "");
Expect(0, 5952, '\P{^inhanunoo}', "");
Expect(1, 5951, '\p{  in_hanunoo}', "");
Expect(0, 5951, '\p{^  in_hanunoo}', "");
Expect(0, 5951, '\P{  in_hanunoo}', "");
Expect(1, 5951, '\P{^  in_hanunoo}', "");
Expect(0, 5952, '\p{  in_hanunoo}', "");
Expect(1, 5952, '\p{^  in_hanunoo}', "");
Expect(1, 5952, '\P{  in_hanunoo}', "");
Expect(0, 5952, '\P{^  in_hanunoo}', "");
Error('\p{	/a/in_Hanunoo}');
Error('\P{	/a/in_Hanunoo}');
Expect(1, 5951, '\p{-In_hanunoo}', "");
Expect(0, 5951, '\p{^-In_hanunoo}', "");
Expect(0, 5951, '\P{-In_hanunoo}', "");
Expect(1, 5951, '\P{^-In_hanunoo}', "");
Expect(0, 5952, '\p{-In_hanunoo}', "");
Expect(1, 5952, '\p{^-In_hanunoo}', "");
Expect(1, 5952, '\P{-In_hanunoo}', "");
Expect(0, 5952, '\P{^-In_hanunoo}', "");
Error('\p{-/a/In_HEBREW}');
Error('\P{-/a/In_HEBREW}');
Expect(1, 1535, '\p{inhebrew}', "");
Expect(0, 1535, '\p{^inhebrew}', "");
Expect(0, 1535, '\P{inhebrew}', "");
Expect(1, 1535, '\P{^inhebrew}', "");
Expect(0, 1536, '\p{inhebrew}', "");
Expect(1, 1536, '\p{^inhebrew}', "");
Expect(1, 1536, '\P{inhebrew}', "");
Expect(0, 1536, '\P{^inhebrew}', "");
Expect(1, 1535, '\p{- IN_hebrew}', "");
Expect(0, 1535, '\p{^- IN_hebrew}', "");
Expect(0, 1535, '\P{- IN_hebrew}', "");
Expect(1, 1535, '\P{^- IN_hebrew}', "");
Expect(0, 1536, '\p{- IN_hebrew}', "");
Expect(1, 1536, '\p{^- IN_hebrew}', "");
Expect(1, 1536, '\P{- IN_hebrew}', "");
Expect(0, 1536, '\P{^- IN_hebrew}', "");
Error('\p{/a/-_In_Hebrew}');
Error('\P{/a/-_In_Hebrew}');
Expect(1, 1535, '\p{-_IN_hebrew}', "");
Expect(0, 1535, '\p{^-_IN_hebrew}', "");
Expect(0, 1535, '\P{-_IN_hebrew}', "");
Expect(1, 1535, '\P{^-_IN_hebrew}', "");
Expect(0, 1536, '\p{-_IN_hebrew}', "");
Expect(1, 1536, '\p{^-_IN_hebrew}', "");
Expect(1, 1536, '\P{-_IN_hebrew}', "");
Expect(0, 1536, '\P{^-_IN_hebrew}', "");
Error('\p{:=in_Hiragana}');
Error('\P{:=in_Hiragana}');
Expect(1, 12447, '\p{inhiragana}', "");
Expect(0, 12447, '\p{^inhiragana}', "");
Expect(0, 12447, '\P{inhiragana}', "");
Expect(1, 12447, '\P{^inhiragana}', "");
Expect(0, 12448, '\p{inhiragana}', "");
Expect(1, 12448, '\p{^inhiragana}', "");
Expect(1, 12448, '\P{inhiragana}', "");
Expect(0, 12448, '\P{^inhiragana}', "");
Expect(1, 12447, '\p{ In_Hiragana}', "");
Expect(0, 12447, '\p{^ In_Hiragana}', "");
Expect(0, 12447, '\P{ In_Hiragana}', "");
Expect(1, 12447, '\P{^ In_Hiragana}', "");
Expect(0, 12448, '\p{ In_Hiragana}', "");
Expect(1, 12448, '\p{^ In_Hiragana}', "");
Expect(1, 12448, '\P{ In_Hiragana}', "");
Expect(0, 12448, '\P{^ In_Hiragana}', "");
Error('\p{ _In_Hiragana:=}');
Error('\P{ _In_Hiragana:=}');
Expect(1, 12447, '\p{	in_HIRAGANA}', "");
Expect(0, 12447, '\p{^	in_HIRAGANA}', "");
Expect(0, 12447, '\P{	in_HIRAGANA}', "");
Expect(1, 12447, '\P{^	in_HIRAGANA}', "");
Expect(0, 12448, '\p{	in_HIRAGANA}', "");
Expect(1, 12448, '\p{^	in_HIRAGANA}', "");
Expect(1, 12448, '\P{	in_HIRAGANA}', "");
Expect(0, 12448, '\P{^	in_HIRAGANA}', "");
Error('\p{	/a/In_Imperial_ARAMAIC}');
Error('\P{	/a/In_Imperial_ARAMAIC}');
Expect(1, 67679, '\p{inimperialaramaic}', "");
Expect(0, 67679, '\p{^inimperialaramaic}', "");
Expect(0, 67679, '\P{inimperialaramaic}', "");
Expect(1, 67679, '\P{^inimperialaramaic}', "");
Expect(0, 67680, '\p{inimperialaramaic}', "");
Expect(1, 67680, '\p{^inimperialaramaic}', "");
Expect(1, 67680, '\P{inimperialaramaic}', "");
Expect(0, 67680, '\P{^inimperialaramaic}', "");
Expect(1, 67679, '\p{  in_IMPERIAL_Aramaic}', "");
Expect(0, 67679, '\p{^  in_IMPERIAL_Aramaic}', "");
Expect(0, 67679, '\P{  in_IMPERIAL_Aramaic}', "");
Expect(1, 67679, '\P{^  in_IMPERIAL_Aramaic}', "");
Expect(0, 67680, '\p{  in_IMPERIAL_Aramaic}', "");
Expect(1, 67680, '\p{^  in_IMPERIAL_Aramaic}', "");
Expect(1, 67680, '\P{  in_IMPERIAL_Aramaic}', "");
Expect(0, 67680, '\P{^  in_IMPERIAL_Aramaic}', "");
Error('\p{/a/__In_imperial_Aramaic}');
Error('\P{/a/__In_imperial_Aramaic}');
Expect(1, 67679, '\p{	In_IMPERIAL_Aramaic}', "");
Expect(0, 67679, '\p{^	In_IMPERIAL_Aramaic}', "");
Expect(0, 67679, '\P{	In_IMPERIAL_Aramaic}', "");
Expect(1, 67679, '\P{^	In_IMPERIAL_Aramaic}', "");
Expect(0, 67680, '\p{	In_IMPERIAL_Aramaic}', "");
Expect(1, 67680, '\p{^	In_IMPERIAL_Aramaic}', "");
Expect(1, 67680, '\P{	In_IMPERIAL_Aramaic}', "");
Expect(0, 67680, '\P{^	In_IMPERIAL_Aramaic}', "");
Error('\p{	/a/IN_Inscriptional_pahlavi}');
Error('\P{	/a/IN_Inscriptional_pahlavi}');
Expect(1, 68479, '\p{ininscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^ininscriptionalpahlavi}', "");
Expect(0, 68479, '\P{ininscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^ininscriptionalpahlavi}', "");
Expect(0, 68480, '\p{ininscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^ininscriptionalpahlavi}', "");
Expect(1, 68480, '\P{ininscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^ininscriptionalpahlavi}', "");
Expect(1, 68479, '\p{	In_inscriptional_Pahlavi}', "");
Expect(0, 68479, '\p{^	In_inscriptional_Pahlavi}', "");
Expect(0, 68479, '\P{	In_inscriptional_Pahlavi}', "");
Expect(1, 68479, '\P{^	In_inscriptional_Pahlavi}', "");
Expect(0, 68480, '\p{	In_inscriptional_Pahlavi}', "");
Expect(1, 68480, '\p{^	In_inscriptional_Pahlavi}', "");
Expect(1, 68480, '\P{	In_inscriptional_Pahlavi}', "");
Expect(0, 68480, '\P{^	In_inscriptional_Pahlavi}', "");
Error('\p{/a/-	In_Inscriptional_pahlavi}');
Error('\P{/a/-	In_Inscriptional_pahlavi}');
Expect(1, 68479, '\p{ 	In_Inscriptional_PAHLAVI}', "");
Expect(0, 68479, '\p{^ 	In_Inscriptional_PAHLAVI}', "");
Expect(0, 68479, '\P{ 	In_Inscriptional_PAHLAVI}', "");
Expect(1, 68479, '\P{^ 	In_Inscriptional_PAHLAVI}', "");
Expect(0, 68480, '\p{ 	In_Inscriptional_PAHLAVI}', "");
Expect(1, 68480, '\p{^ 	In_Inscriptional_PAHLAVI}', "");
Expect(1, 68480, '\P{ 	In_Inscriptional_PAHLAVI}', "");
Expect(0, 68480, '\P{^ 	In_Inscriptional_PAHLAVI}', "");
Error('\p{__IN_INSCRIPTIONAL_parthian/a/}');
Error('\P{__IN_INSCRIPTIONAL_parthian/a/}');
Expect(1, 68447, '\p{ininscriptionalparthian}', "");
Expect(0, 68447, '\p{^ininscriptionalparthian}', "");
Expect(0, 68447, '\P{ininscriptionalparthian}', "");
Expect(1, 68447, '\P{^ininscriptionalparthian}', "");
Expect(0, 68448, '\p{ininscriptionalparthian}', "");
Expect(1, 68448, '\p{^ininscriptionalparthian}', "");
Expect(1, 68448, '\P{ininscriptionalparthian}', "");
Expect(0, 68448, '\P{^ininscriptionalparthian}', "");
Expect(1, 68447, '\p{_in_INSCRIPTIONAL_Parthian}', "");
Expect(0, 68447, '\p{^_in_INSCRIPTIONAL_Parthian}', "");
Expect(0, 68447, '\P{_in_INSCRIPTIONAL_Parthian}', "");
Expect(1, 68447, '\P{^_in_INSCRIPTIONAL_Parthian}', "");
Expect(0, 68448, '\p{_in_INSCRIPTIONAL_Parthian}', "");
Expect(1, 68448, '\p{^_in_INSCRIPTIONAL_Parthian}', "");
Expect(1, 68448, '\P{_in_INSCRIPTIONAL_Parthian}', "");
Expect(0, 68448, '\P{^_in_INSCRIPTIONAL_Parthian}', "");
Error('\p{/a/	In_inscriptional_Parthian}');
Error('\P{/a/	In_inscriptional_Parthian}');
Expect(1, 68447, '\p{  in_Inscriptional_Parthian}', "");
Expect(0, 68447, '\p{^  in_Inscriptional_Parthian}', "");
Expect(0, 68447, '\P{  in_Inscriptional_Parthian}', "");
Expect(1, 68447, '\P{^  in_Inscriptional_Parthian}', "");
Expect(0, 68448, '\p{  in_Inscriptional_Parthian}', "");
Expect(1, 68448, '\p{^  in_Inscriptional_Parthian}', "");
Expect(1, 68448, '\P{  in_Inscriptional_Parthian}', "");
Expect(0, 68448, '\P{^  in_Inscriptional_Parthian}', "");
Error('\p{/a/ In_javanese}');
Error('\P{/a/ In_javanese}');
Expect(1, 43487, '\p{injavanese}', "");
Expect(0, 43487, '\p{^injavanese}', "");
Expect(0, 43487, '\P{injavanese}', "");
Expect(1, 43487, '\P{^injavanese}', "");
Expect(0, 43488, '\p{injavanese}', "");
Expect(1, 43488, '\p{^injavanese}', "");
Expect(1, 43488, '\P{injavanese}', "");
Expect(0, 43488, '\P{^injavanese}', "");
Expect(1, 43487, '\p{	 in_Javanese}', "");
Expect(0, 43487, '\p{^	 in_Javanese}', "");
Expect(0, 43487, '\P{	 in_Javanese}', "");
Expect(1, 43487, '\P{^	 in_Javanese}', "");
Expect(0, 43488, '\p{	 in_Javanese}', "");
Expect(1, 43488, '\p{^	 in_Javanese}', "");
Expect(1, 43488, '\P{	 in_Javanese}', "");
Expect(0, 43488, '\P{^	 in_Javanese}', "");
Error('\p{  in_javanese/a/}');
Error('\P{  in_javanese/a/}');
Expect(1, 43487, '\p{in_Javanese}', "");
Expect(0, 43487, '\p{^in_Javanese}', "");
Expect(0, 43487, '\P{in_Javanese}', "");
Expect(1, 43487, '\P{^in_Javanese}', "");
Expect(0, 43488, '\p{in_Javanese}', "");
Expect(1, 43488, '\p{^in_Javanese}', "");
Expect(1, 43488, '\P{in_Javanese}', "");
Expect(0, 43488, '\P{^in_Javanese}', "");
Error('\p{ 	in_kaithi/a/}');
Error('\P{ 	in_kaithi/a/}');
Expect(1, 69839, '\p{inkaithi}', "");
Expect(0, 69839, '\p{^inkaithi}', "");
Expect(0, 69839, '\P{inkaithi}', "");
Expect(1, 69839, '\P{^inkaithi}', "");
Expect(0, 69840, '\p{inkaithi}', "");
Expect(1, 69840, '\p{^inkaithi}', "");
Expect(1, 69840, '\P{inkaithi}', "");
Expect(0, 69840, '\P{^inkaithi}', "");
Expect(1, 69839, '\p{_In_kaithi}', "");
Expect(0, 69839, '\p{^_In_kaithi}', "");
Expect(0, 69839, '\P{_In_kaithi}', "");
Expect(1, 69839, '\P{^_In_kaithi}', "");
Expect(0, 69840, '\p{_In_kaithi}', "");
Expect(1, 69840, '\p{^_In_kaithi}', "");
Expect(1, 69840, '\P{_In_kaithi}', "");
Expect(0, 69840, '\P{^_In_kaithi}', "");
Error('\p{/a/ -In_kaithi}');
Error('\P{/a/ -In_kaithi}');
Expect(1, 69839, '\p{-	in_kaithi}', "");
Expect(0, 69839, '\p{^-	in_kaithi}', "");
Expect(0, 69839, '\P{-	in_kaithi}', "");
Expect(1, 69839, '\P{^-	in_kaithi}', "");
Expect(0, 69840, '\p{-	in_kaithi}', "");
Expect(1, 69840, '\p{^-	in_kaithi}', "");
Expect(1, 69840, '\P{-	in_kaithi}', "");
Expect(0, 69840, '\P{^-	in_kaithi}', "");
Error('\p{:=--In_Kannada}');
Error('\P{:=--In_Kannada}');
Expect(1, 3327, '\p{inkannada}', "");
Expect(0, 3327, '\p{^inkannada}', "");
Expect(0, 3327, '\P{inkannada}', "");
Expect(1, 3327, '\P{^inkannada}', "");
Expect(0, 3328, '\p{inkannada}', "");
Expect(1, 3328, '\p{^inkannada}', "");
Expect(1, 3328, '\P{inkannada}', "");
Expect(0, 3328, '\P{^inkannada}', "");
Expect(1, 3327, '\p{_	In_Kannada}', "");
Expect(0, 3327, '\p{^_	In_Kannada}', "");
Expect(0, 3327, '\P{_	In_Kannada}', "");
Expect(1, 3327, '\P{^_	In_Kannada}', "");
Expect(0, 3328, '\p{_	In_Kannada}', "");
Expect(1, 3328, '\p{^_	In_Kannada}', "");
Expect(1, 3328, '\P{_	In_Kannada}', "");
Expect(0, 3328, '\P{^_	In_Kannada}', "");
Error('\p{-	In_Kannada/a/}');
Error('\P{-	In_Kannada/a/}');
Expect(1, 3327, '\p{ in_Kannada}', "");
Expect(0, 3327, '\p{^ in_Kannada}', "");
Expect(0, 3327, '\P{ in_Kannada}', "");
Expect(1, 3327, '\P{^ in_Kannada}', "");
Expect(0, 3328, '\p{ in_Kannada}', "");
Expect(1, 3328, '\p{^ in_Kannada}', "");
Expect(1, 3328, '\P{ in_Kannada}', "");
Expect(0, 3328, '\P{^ in_Kannada}', "");
Error('\p{_/a/IN_Katakana}');
Error('\P{_/a/IN_Katakana}');
Expect(1, 12543, '\p{inkatakana}', "");
Expect(0, 12543, '\p{^inkatakana}', "");
Expect(0, 12543, '\P{inkatakana}', "");
Expect(1, 12543, '\P{^inkatakana}', "");
Expect(0, 12544, '\p{inkatakana}', "");
Expect(1, 12544, '\p{^inkatakana}', "");
Expect(1, 12544, '\P{inkatakana}', "");
Expect(0, 12544, '\P{^inkatakana}', "");
Expect(1, 12543, '\p{	 IN_Katakana}', "");
Expect(0, 12543, '\p{^	 IN_Katakana}', "");
Expect(0, 12543, '\P{	 IN_Katakana}', "");
Expect(1, 12543, '\P{^	 IN_Katakana}', "");
Expect(0, 12544, '\p{	 IN_Katakana}', "");
Expect(1, 12544, '\p{^	 IN_Katakana}', "");
Expect(1, 12544, '\P{	 IN_Katakana}', "");
Expect(0, 12544, '\P{^	 IN_Katakana}', "");
Error('\p{  IN_Katakana:=}');
Error('\P{  IN_Katakana:=}');
Expect(1, 12543, '\p{_In_katakana}', "");
Expect(0, 12543, '\p{^_In_katakana}', "");
Expect(0, 12543, '\P{_In_katakana}', "");
Expect(1, 12543, '\P{^_In_katakana}', "");
Expect(0, 12544, '\p{_In_katakana}', "");
Expect(1, 12544, '\p{^_In_katakana}', "");
Expect(1, 12544, '\P{_In_katakana}', "");
Expect(0, 12544, '\P{^_In_katakana}', "");
Error('\p{	:=In_Kayah_Li}');
Error('\P{	:=In_Kayah_Li}');
Expect(1, 43311, '\p{inkayahli}', "");
Expect(0, 43311, '\p{^inkayahli}', "");
Expect(0, 43311, '\P{inkayahli}', "");
Expect(1, 43311, '\P{^inkayahli}', "");
Expect(0, 43312, '\p{inkayahli}', "");
Expect(1, 43312, '\p{^inkayahli}', "");
Expect(1, 43312, '\P{inkayahli}', "");
Expect(0, 43312, '\P{^inkayahli}', "");
Expect(1, 43311, '\p{	_in_kayah_Li}', "");
Expect(0, 43311, '\p{^	_in_kayah_Li}', "");
Expect(0, 43311, '\P{	_in_kayah_Li}', "");
Expect(1, 43311, '\P{^	_in_kayah_Li}', "");
Expect(0, 43312, '\p{	_in_kayah_Li}', "");
Expect(1, 43312, '\p{^	_in_kayah_Li}', "");
Expect(1, 43312, '\P{	_in_kayah_Li}', "");
Expect(0, 43312, '\P{^	_in_kayah_Li}', "");
Error('\p{	In_Kayah_Li/a/}');
Error('\P{	In_Kayah_Li/a/}');
Expect(1, 43311, '\p{-In_Kayah_Li}', "");
Expect(0, 43311, '\p{^-In_Kayah_Li}', "");
Expect(0, 43311, '\P{-In_Kayah_Li}', "");
Expect(1, 43311, '\P{^-In_Kayah_Li}', "");
Expect(0, 43312, '\p{-In_Kayah_Li}', "");
Expect(1, 43312, '\p{^-In_Kayah_Li}', "");
Expect(1, 43312, '\P{-In_Kayah_Li}', "");
Expect(0, 43312, '\P{^-In_Kayah_Li}', "");
Error('\p{-:=In_Kharoshthi}');
Error('\P{-:=In_Kharoshthi}');
Expect(1, 68191, '\p{inkharoshthi}', "");
Expect(0, 68191, '\p{^inkharoshthi}', "");
Expect(0, 68191, '\P{inkharoshthi}', "");
Expect(1, 68191, '\P{^inkharoshthi}', "");
Expect(0, 68192, '\p{inkharoshthi}', "");
Expect(1, 68192, '\p{^inkharoshthi}', "");
Expect(1, 68192, '\P{inkharoshthi}', "");
Expect(0, 68192, '\P{^inkharoshthi}', "");
Expect(1, 68191, '\p{ In_Kharoshthi}', "");
Expect(0, 68191, '\p{^ In_Kharoshthi}', "");
Expect(0, 68191, '\P{ In_Kharoshthi}', "");
Expect(1, 68191, '\P{^ In_Kharoshthi}', "");
Expect(0, 68192, '\p{ In_Kharoshthi}', "");
Expect(1, 68192, '\p{^ In_Kharoshthi}', "");
Expect(1, 68192, '\P{ In_Kharoshthi}', "");
Expect(0, 68192, '\P{^ In_Kharoshthi}', "");
Error('\p{	/a/In_Kharoshthi}');
Error('\P{	/a/In_Kharoshthi}');
Expect(1, 68191, '\p{_In_KHAROSHTHI}', "");
Expect(0, 68191, '\p{^_In_KHAROSHTHI}', "");
Expect(0, 68191, '\P{_In_KHAROSHTHI}', "");
Expect(1, 68191, '\P{^_In_KHAROSHTHI}', "");
Expect(0, 68192, '\p{_In_KHAROSHTHI}', "");
Expect(1, 68192, '\p{^_In_KHAROSHTHI}', "");
Expect(1, 68192, '\P{_In_KHAROSHTHI}', "");
Expect(0, 68192, '\P{^_In_KHAROSHTHI}', "");
Error('\p{:=- IN_Khmer}');
Error('\P{:=- IN_Khmer}');
Expect(1, 6143, '\p{inkhmer}', "");
Expect(0, 6143, '\p{^inkhmer}', "");
Expect(0, 6143, '\P{inkhmer}', "");
Expect(1, 6143, '\P{^inkhmer}', "");
Expect(0, 6144, '\p{inkhmer}', "");
Expect(1, 6144, '\p{^inkhmer}', "");
Expect(1, 6144, '\P{inkhmer}', "");
Expect(0, 6144, '\P{^inkhmer}', "");
Expect(1, 6143, '\p{-IN_KHMER}', "");
Expect(0, 6143, '\p{^-IN_KHMER}', "");
Expect(0, 6143, '\P{-IN_KHMER}', "");
Expect(1, 6143, '\P{^-IN_KHMER}', "");
Expect(0, 6144, '\p{-IN_KHMER}', "");
Expect(1, 6144, '\p{^-IN_KHMER}', "");
Expect(1, 6144, '\P{-IN_KHMER}', "");
Expect(0, 6144, '\P{^-IN_KHMER}', "");
Error('\p{ in_KHMER/a/}');
Error('\P{ in_KHMER/a/}');
Expect(1, 6143, '\p{	 in_Khmer}', "");
Expect(0, 6143, '\p{^	 in_Khmer}', "");
Expect(0, 6143, '\P{	 in_Khmer}', "");
Expect(1, 6143, '\P{^	 in_Khmer}', "");
Expect(0, 6144, '\p{	 in_Khmer}', "");
Expect(1, 6144, '\p{^	 in_Khmer}', "");
Expect(1, 6144, '\P{	 in_Khmer}', "");
Expect(0, 6144, '\P{^	 in_Khmer}', "");
Error('\p{_	IN_Lao:=}');
Error('\P{_	IN_Lao:=}');
Expect(1, 3839, '\p{inlao}', "");
Expect(0, 3839, '\p{^inlao}', "");
Expect(0, 3839, '\P{inlao}', "");
Expect(1, 3839, '\P{^inlao}', "");
Expect(0, 3840, '\p{inlao}', "");
Expect(1, 3840, '\p{^inlao}', "");
Expect(1, 3840, '\P{inlao}', "");
Expect(0, 3840, '\P{^inlao}', "");
Expect(1, 3839, '\p{	_In_LAO}', "");
Expect(0, 3839, '\p{^	_In_LAO}', "");
Expect(0, 3839, '\P{	_In_LAO}', "");
Expect(1, 3839, '\P{^	_In_LAO}', "");
Expect(0, 3840, '\p{	_In_LAO}', "");
Expect(1, 3840, '\p{^	_In_LAO}', "");
Expect(1, 3840, '\P{	_In_LAO}', "");
Expect(0, 3840, '\P{^	_In_LAO}', "");
Error('\p{	IN_Lao:=}');
Error('\P{	IN_Lao:=}');
Expect(1, 3839, '\p{ In_Lao}', "");
Expect(0, 3839, '\p{^ In_Lao}', "");
Expect(0, 3839, '\P{ In_Lao}', "");
Expect(1, 3839, '\P{^ In_Lao}', "");
Expect(0, 3840, '\p{ In_Lao}', "");
Expect(1, 3840, '\p{^ In_Lao}', "");
Expect(1, 3840, '\P{ In_Lao}', "");
Expect(0, 3840, '\P{^ In_Lao}', "");
Error('\p{/a/ In_LEPCHA}');
Error('\P{/a/ In_LEPCHA}');
Expect(1, 7247, '\p{inlepcha}', "");
Expect(0, 7247, '\p{^inlepcha}', "");
Expect(0, 7247, '\P{inlepcha}', "");
Expect(1, 7247, '\P{^inlepcha}', "");
Expect(0, 7248, '\p{inlepcha}', "");
Expect(1, 7248, '\p{^inlepcha}', "");
Expect(1, 7248, '\P{inlepcha}', "");
Expect(0, 7248, '\P{^inlepcha}', "");
Expect(1, 7247, '\p{ -IN_Lepcha}', "");
Expect(0, 7247, '\p{^ -IN_Lepcha}', "");
Expect(0, 7247, '\P{ -IN_Lepcha}', "");
Expect(1, 7247, '\P{^ -IN_Lepcha}', "");
Expect(0, 7248, '\p{ -IN_Lepcha}', "");
Expect(1, 7248, '\p{^ -IN_Lepcha}', "");
Expect(1, 7248, '\P{ -IN_Lepcha}', "");
Expect(0, 7248, '\P{^ -IN_Lepcha}', "");
Error('\p{	_In_lepcha:=}');
Error('\P{	_In_lepcha:=}');
Expect(1, 7247, '\p{In_Lepcha}', "");
Expect(0, 7247, '\p{^In_Lepcha}', "");
Expect(0, 7247, '\P{In_Lepcha}', "");
Expect(1, 7247, '\P{^In_Lepcha}', "");
Expect(0, 7248, '\p{In_Lepcha}', "");
Expect(1, 7248, '\p{^In_Lepcha}', "");
Expect(1, 7248, '\P{In_Lepcha}', "");
Expect(0, 7248, '\P{^In_Lepcha}', "");
Error('\p{:= _IN_Limbu}');
Error('\P{:= _IN_Limbu}');
Expect(1, 6479, '\p{inlimbu}', "");
Expect(0, 6479, '\p{^inlimbu}', "");
Expect(0, 6479, '\P{inlimbu}', "");
Expect(1, 6479, '\P{^inlimbu}', "");
Expect(0, 6480, '\p{inlimbu}', "");
Expect(1, 6480, '\p{^inlimbu}', "");
Expect(1, 6480, '\P{inlimbu}', "");
Expect(0, 6480, '\P{^inlimbu}', "");
Expect(1, 6479, '\p{	_in_Limbu}', "");
Expect(0, 6479, '\p{^	_in_Limbu}', "");
Expect(0, 6479, '\P{	_in_Limbu}', "");
Expect(1, 6479, '\P{^	_in_Limbu}', "");
Expect(0, 6480, '\p{	_in_Limbu}', "");
Expect(1, 6480, '\p{^	_in_Limbu}', "");
Expect(1, 6480, '\P{	_in_Limbu}', "");
Expect(0, 6480, '\P{^	_in_Limbu}', "");
Error('\p{:=-_in_Limbu}');
Error('\P{:=-_in_Limbu}');
Expect(1, 6479, '\p{	_In_limbu}', "");
Expect(0, 6479, '\p{^	_In_limbu}', "");
Expect(0, 6479, '\P{	_In_limbu}', "");
Expect(1, 6479, '\P{^	_In_limbu}', "");
Expect(0, 6480, '\p{	_In_limbu}', "");
Expect(1, 6480, '\p{^	_In_limbu}', "");
Expect(1, 6480, '\P{	_In_limbu}', "");
Expect(0, 6480, '\P{^	_In_limbu}', "");
Error('\p{-IN_Lisu:=}');
Error('\P{-IN_Lisu:=}');
Expect(1, 42239, '\p{inlisu}', "");
Expect(0, 42239, '\p{^inlisu}', "");
Expect(0, 42239, '\P{inlisu}', "");
Expect(1, 42239, '\P{^inlisu}', "");
Expect(0, 42240, '\p{inlisu}', "");
Expect(1, 42240, '\p{^inlisu}', "");
Expect(1, 42240, '\P{inlisu}', "");
Expect(0, 42240, '\P{^inlisu}', "");
Expect(1, 42239, '\p{__IN_Lisu}', "");
Expect(0, 42239, '\p{^__IN_Lisu}', "");
Expect(0, 42239, '\P{__IN_Lisu}', "");
Expect(1, 42239, '\P{^__IN_Lisu}', "");
Expect(0, 42240, '\p{__IN_Lisu}', "");
Expect(1, 42240, '\p{^__IN_Lisu}', "");
Expect(1, 42240, '\P{__IN_Lisu}', "");
Expect(0, 42240, '\P{^__IN_Lisu}', "");
Error('\p{-:=in_Lisu}');
Error('\P{-:=in_Lisu}');
Expect(1, 42239, '\p{-	In_Lisu}', "");
Expect(0, 42239, '\p{^-	In_Lisu}', "");
Expect(0, 42239, '\P{-	In_Lisu}', "");
Expect(1, 42239, '\P{^-	In_Lisu}', "");
Expect(0, 42240, '\p{-	In_Lisu}', "");
Expect(1, 42240, '\p{^-	In_Lisu}', "");
Expect(1, 42240, '\P{-	In_Lisu}', "");
Expect(0, 42240, '\P{^-	In_Lisu}', "");
Error('\p{:=	-in_lycian}');
Error('\P{:=	-in_lycian}');
Expect(1, 66207, '\p{inlycian}', "");
Expect(0, 66207, '\p{^inlycian}', "");
Expect(0, 66207, '\P{inlycian}', "");
Expect(1, 66207, '\P{^inlycian}', "");
Expect(0, 66208, '\p{inlycian}', "");
Expect(1, 66208, '\p{^inlycian}', "");
Expect(1, 66208, '\P{inlycian}', "");
Expect(0, 66208, '\P{^inlycian}', "");
Expect(1, 66207, '\p{ _IN_Lycian}', "");
Expect(0, 66207, '\p{^ _IN_Lycian}', "");
Expect(0, 66207, '\P{ _IN_Lycian}', "");
Expect(1, 66207, '\P{^ _IN_Lycian}', "");
Expect(0, 66208, '\p{ _IN_Lycian}', "");
Expect(1, 66208, '\p{^ _IN_Lycian}', "");
Expect(1, 66208, '\P{ _IN_Lycian}', "");
Expect(0, 66208, '\P{^ _IN_Lycian}', "");
Error('\p{-:=In_lycian}');
Error('\P{-:=In_lycian}');
Expect(1, 66207, '\p{_ in_Lycian}', "");
Expect(0, 66207, '\p{^_ in_Lycian}', "");
Expect(0, 66207, '\P{_ in_Lycian}', "");
Expect(1, 66207, '\P{^_ in_Lycian}', "");
Expect(0, 66208, '\p{_ in_Lycian}', "");
Expect(1, 66208, '\p{^_ in_Lycian}', "");
Expect(1, 66208, '\P{_ in_Lycian}', "");
Expect(0, 66208, '\P{^_ in_Lycian}', "");
Error('\p{:=In_Lydian}');
Error('\P{:=In_Lydian}');
Expect(1, 67903, '\p{inlydian}', "");
Expect(0, 67903, '\p{^inlydian}', "");
Expect(0, 67903, '\P{inlydian}', "");
Expect(1, 67903, '\P{^inlydian}', "");
Expect(0, 67904, '\p{inlydian}', "");
Expect(1, 67904, '\p{^inlydian}', "");
Expect(1, 67904, '\P{inlydian}', "");
Expect(0, 67904, '\P{^inlydian}', "");
Expect(1, 67903, '\p{_in_LYDIAN}', "");
Expect(0, 67903, '\p{^_in_LYDIAN}', "");
Expect(0, 67903, '\P{_in_LYDIAN}', "");
Expect(1, 67903, '\P{^_in_LYDIAN}', "");
Expect(0, 67904, '\p{_in_LYDIAN}', "");
Expect(1, 67904, '\p{^_in_LYDIAN}', "");
Expect(1, 67904, '\P{_in_LYDIAN}', "");
Expect(0, 67904, '\P{^_in_LYDIAN}', "");
Error('\p{/a/ 	IN_lydian}');
Error('\P{/a/ 	IN_lydian}');
Expect(1, 67903, '\p{	in_Lydian}', "");
Expect(0, 67903, '\p{^	in_Lydian}', "");
Expect(0, 67903, '\P{	in_Lydian}', "");
Expect(1, 67903, '\P{^	in_Lydian}', "");
Expect(0, 67904, '\p{	in_Lydian}', "");
Expect(1, 67904, '\p{^	in_Lydian}', "");
Expect(1, 67904, '\P{	in_Lydian}', "");
Expect(0, 67904, '\P{^	in_Lydian}', "");
Error('\p{	in_Malayalam/a/}');
Error('\P{	in_Malayalam/a/}');
Expect(1, 3455, '\p{inmalayalam}', "");
Expect(0, 3455, '\p{^inmalayalam}', "");
Expect(0, 3455, '\P{inmalayalam}', "");
Expect(1, 3455, '\P{^inmalayalam}', "");
Expect(0, 3456, '\p{inmalayalam}', "");
Expect(1, 3456, '\p{^inmalayalam}', "");
Expect(1, 3456, '\P{inmalayalam}', "");
Expect(0, 3456, '\P{^inmalayalam}', "");
Expect(1, 3455, '\p{	_In_malayalam}', "");
Expect(0, 3455, '\p{^	_In_malayalam}', "");
Expect(0, 3455, '\P{	_In_malayalam}', "");
Expect(1, 3455, '\P{^	_In_malayalam}', "");
Expect(0, 3456, '\p{	_In_malayalam}', "");
Expect(1, 3456, '\p{^	_In_malayalam}', "");
Expect(1, 3456, '\P{	_In_malayalam}', "");
Expect(0, 3456, '\P{^	_In_malayalam}', "");
Error('\p{_ in_Malayalam:=}');
Error('\P{_ in_Malayalam:=}');
Expect(1, 3455, '\p{ -In_Malayalam}', "");
Expect(0, 3455, '\p{^ -In_Malayalam}', "");
Expect(0, 3455, '\P{ -In_Malayalam}', "");
Expect(1, 3455, '\P{^ -In_Malayalam}', "");
Expect(0, 3456, '\p{ -In_Malayalam}', "");
Expect(1, 3456, '\p{^ -In_Malayalam}', "");
Expect(1, 3456, '\P{ -In_Malayalam}', "");
Expect(0, 3456, '\P{^ -In_Malayalam}', "");
Error('\p{	/a/In_Mandaic}');
Error('\P{	/a/In_Mandaic}');
Expect(1, 2143, '\p{inmandaic}', "");
Expect(0, 2143, '\p{^inmandaic}', "");
Expect(0, 2143, '\P{inmandaic}', "");
Expect(1, 2143, '\P{^inmandaic}', "");
Expect(0, 2144, '\p{inmandaic}', "");
Expect(1, 2144, '\p{^inmandaic}', "");
Expect(1, 2144, '\P{inmandaic}', "");
Expect(0, 2144, '\P{^inmandaic}', "");
Expect(1, 2143, '\p{-In_Mandaic}', "");
Expect(0, 2143, '\p{^-In_Mandaic}', "");
Expect(0, 2143, '\P{-In_Mandaic}', "");
Expect(1, 2143, '\P{^-In_Mandaic}', "");
Expect(0, 2144, '\p{-In_Mandaic}', "");
Expect(1, 2144, '\p{^-In_Mandaic}', "");
Expect(1, 2144, '\P{-In_Mandaic}', "");
Expect(0, 2144, '\P{^-In_Mandaic}', "");
Error('\p{-:=In_mandaic}');
Error('\P{-:=In_mandaic}');
Expect(1, 2143, '\p{	_IN_Mandaic}', "");
Expect(0, 2143, '\p{^	_IN_Mandaic}', "");
Expect(0, 2143, '\P{	_IN_Mandaic}', "");
Expect(1, 2143, '\P{^	_IN_Mandaic}', "");
Expect(0, 2144, '\p{	_IN_Mandaic}', "");
Expect(1, 2144, '\p{^	_IN_Mandaic}', "");
Expect(1, 2144, '\P{	_IN_Mandaic}', "");
Expect(0, 2144, '\P{^	_IN_Mandaic}', "");
Error('\p{/a/_ IN_Meetei_MAYEK}');
Error('\P{/a/_ IN_Meetei_MAYEK}');
Expect(1, 44031, '\p{inmeeteimayek}', "");
Expect(0, 44031, '\p{^inmeeteimayek}', "");
Expect(0, 44031, '\P{inmeeteimayek}', "");
Expect(1, 44031, '\P{^inmeeteimayek}', "");
Expect(0, 44032, '\p{inmeeteimayek}', "");
Expect(1, 44032, '\p{^inmeeteimayek}', "");
Expect(1, 44032, '\P{inmeeteimayek}', "");
Expect(0, 44032, '\P{^inmeeteimayek}', "");
Expect(1, 44031, '\p{-	In_Meetei_MAYEK}', "");
Expect(0, 44031, '\p{^-	In_Meetei_MAYEK}', "");
Expect(0, 44031, '\P{-	In_Meetei_MAYEK}', "");
Expect(1, 44031, '\P{^-	In_Meetei_MAYEK}', "");
Expect(0, 44032, '\p{-	In_Meetei_MAYEK}', "");
Expect(1, 44032, '\p{^-	In_Meetei_MAYEK}', "");
Expect(1, 44032, '\P{-	In_Meetei_MAYEK}', "");
Expect(0, 44032, '\P{^-	In_Meetei_MAYEK}', "");
Error('\p{ -IN_Meetei_mayek/a/}');
Error('\P{ -IN_Meetei_mayek/a/}');
Expect(1, 44031, '\p{		In_Meetei_MAYEK}', "");
Expect(0, 44031, '\p{^		In_Meetei_MAYEK}', "");
Expect(0, 44031, '\P{		In_Meetei_MAYEK}', "");
Expect(1, 44031, '\P{^		In_Meetei_MAYEK}', "");
Expect(0, 44032, '\p{		In_Meetei_MAYEK}', "");
Expect(1, 44032, '\p{^		In_Meetei_MAYEK}', "");
Expect(1, 44032, '\P{		In_Meetei_MAYEK}', "");
Expect(0, 44032, '\P{^		In_Meetei_MAYEK}', "");
Error('\p{/a/ 	In_MEROITIC_CURSIVE}');
Error('\P{/a/ 	In_MEROITIC_CURSIVE}');
Expect(1, 68095, '\p{inmeroiticcursive}', "");
Expect(0, 68095, '\p{^inmeroiticcursive}', "");
Expect(0, 68095, '\P{inmeroiticcursive}', "");
Expect(1, 68095, '\P{^inmeroiticcursive}', "");
Expect(0, 68096, '\p{inmeroiticcursive}', "");
Expect(1, 68096, '\p{^inmeroiticcursive}', "");
Expect(1, 68096, '\P{inmeroiticcursive}', "");
Expect(0, 68096, '\P{^inmeroiticcursive}', "");
Expect(1, 68095, '\p{ 	In_MEROITIC_cursive}', "");
Expect(0, 68095, '\p{^ 	In_MEROITIC_cursive}', "");
Expect(0, 68095, '\P{ 	In_MEROITIC_cursive}', "");
Expect(1, 68095, '\P{^ 	In_MEROITIC_cursive}', "");
Expect(0, 68096, '\p{ 	In_MEROITIC_cursive}', "");
Expect(1, 68096, '\p{^ 	In_MEROITIC_cursive}', "");
Expect(1, 68096, '\P{ 	In_MEROITIC_cursive}', "");
Expect(0, 68096, '\P{^ 	In_MEROITIC_cursive}', "");
Error('\p{-:=in_Meroitic_CURSIVE}');
Error('\P{-:=in_Meroitic_CURSIVE}');
Expect(1, 68095, '\p{ 	In_Meroitic_Cursive}', "");
Expect(0, 68095, '\p{^ 	In_Meroitic_Cursive}', "");
Expect(0, 68095, '\P{ 	In_Meroitic_Cursive}', "");
Expect(1, 68095, '\P{^ 	In_Meroitic_Cursive}', "");
Expect(0, 68096, '\p{ 	In_Meroitic_Cursive}', "");
Expect(1, 68096, '\p{^ 	In_Meroitic_Cursive}', "");
Expect(1, 68096, '\P{ 	In_Meroitic_Cursive}', "");
Expect(0, 68096, '\P{^ 	In_Meroitic_Cursive}', "");
Error('\p{-_In_meroitic_Hieroglyphs:=}');
Error('\P{-_In_meroitic_Hieroglyphs:=}');
Expect(1, 67999, '\p{inmeroitichieroglyphs}', "");
Expect(0, 67999, '\p{^inmeroitichieroglyphs}', "");
Expect(0, 67999, '\P{inmeroitichieroglyphs}', "");
Expect(1, 67999, '\P{^inmeroitichieroglyphs}', "");
Expect(0, 68000, '\p{inmeroitichieroglyphs}', "");
Expect(1, 68000, '\p{^inmeroitichieroglyphs}', "");
Expect(1, 68000, '\P{inmeroitichieroglyphs}', "");
Expect(0, 68000, '\P{^inmeroitichieroglyphs}', "");
Expect(1, 67999, '\p{		In_Meroitic_HIEROGLYPHS}', "");
Expect(0, 67999, '\p{^		In_Meroitic_HIEROGLYPHS}', "");
Expect(0, 67999, '\P{		In_Meroitic_HIEROGLYPHS}', "");
Expect(1, 67999, '\P{^		In_Meroitic_HIEROGLYPHS}', "");
Expect(0, 68000, '\p{		In_Meroitic_HIEROGLYPHS}', "");
Expect(1, 68000, '\p{^		In_Meroitic_HIEROGLYPHS}', "");
Expect(1, 68000, '\P{		In_Meroitic_HIEROGLYPHS}', "");
Expect(0, 68000, '\P{^		In_Meroitic_HIEROGLYPHS}', "");
Error('\p{	:=in_meroitic_Hieroglyphs}');
Error('\P{	:=in_meroitic_Hieroglyphs}');
Expect(1, 67999, '\p{ -in_meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\p{^ -in_meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\P{ -in_meroitic_Hieroglyphs}', "");
Expect(1, 67999, '\P{^ -in_meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\p{ -in_meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\p{^ -in_meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\P{ -in_meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\P{^ -in_meroitic_Hieroglyphs}', "");
Error('\p{:=	_in_Miao}');
Error('\P{:=	_in_Miao}');
Expect(1, 94111, '\p{inmiao}', "");
Expect(0, 94111, '\p{^inmiao}', "");
Expect(0, 94111, '\P{inmiao}', "");
Expect(1, 94111, '\P{^inmiao}', "");
Expect(0, 94112, '\p{inmiao}', "");
Expect(1, 94112, '\p{^inmiao}', "");
Expect(1, 94112, '\P{inmiao}', "");
Expect(0, 94112, '\P{^inmiao}', "");
Expect(1, 94111, '\p{-	In_Miao}', "");
Expect(0, 94111, '\p{^-	In_Miao}', "");
Expect(0, 94111, '\P{-	In_Miao}', "");
Expect(1, 94111, '\P{^-	In_Miao}', "");
Expect(0, 94112, '\p{-	In_Miao}', "");
Expect(1, 94112, '\p{^-	In_Miao}', "");
Expect(1, 94112, '\P{-	In_Miao}', "");
Expect(0, 94112, '\P{^-	In_Miao}', "");
Error('\p{-:=IN_MIAO}');
Error('\P{-:=IN_MIAO}');
Expect(1, 94111, '\p{	in_MIAO}', "");
Expect(0, 94111, '\p{^	in_MIAO}', "");
Expect(0, 94111, '\P{	in_MIAO}', "");
Expect(1, 94111, '\P{^	in_MIAO}', "");
Expect(0, 94112, '\p{	in_MIAO}', "");
Expect(1, 94112, '\p{^	in_MIAO}', "");
Expect(1, 94112, '\P{	in_MIAO}', "");
Expect(0, 94112, '\P{^	in_MIAO}', "");
Error('\p{/a/  IN_mongolian}');
Error('\P{/a/  IN_mongolian}');
Expect(1, 6319, '\p{inmongolian}', "");
Expect(0, 6319, '\p{^inmongolian}', "");
Expect(0, 6319, '\P{inmongolian}', "");
Expect(1, 6319, '\P{^inmongolian}', "");
Expect(0, 6320, '\p{inmongolian}', "");
Expect(1, 6320, '\p{^inmongolian}', "");
Expect(1, 6320, '\P{inmongolian}', "");
Expect(0, 6320, '\P{^inmongolian}', "");
Expect(1, 6319, '\p{	 IN_Mongolian}', "");
Expect(0, 6319, '\p{^	 IN_Mongolian}', "");
Expect(0, 6319, '\P{	 IN_Mongolian}', "");
Expect(1, 6319, '\P{^	 IN_Mongolian}', "");
Expect(0, 6320, '\p{	 IN_Mongolian}', "");
Expect(1, 6320, '\p{^	 IN_Mongolian}', "");
Expect(1, 6320, '\P{	 IN_Mongolian}', "");
Expect(0, 6320, '\P{^	 IN_Mongolian}', "");
Error('\p{-/a/In_MONGOLIAN}');
Error('\P{-/a/In_MONGOLIAN}');
Expect(1, 6319, '\p{ 	IN_Mongolian}', "");
Expect(0, 6319, '\p{^ 	IN_Mongolian}', "");
Expect(0, 6319, '\P{ 	IN_Mongolian}', "");
Expect(1, 6319, '\P{^ 	IN_Mongolian}', "");
Expect(0, 6320, '\p{ 	IN_Mongolian}', "");
Expect(1, 6320, '\p{^ 	IN_Mongolian}', "");
Expect(1, 6320, '\P{ 	IN_Mongolian}', "");
Expect(0, 6320, '\P{^ 	IN_Mongolian}', "");
Error('\p{:=		In_Myanmar}');
Error('\P{:=		In_Myanmar}');
Expect(1, 4255, '\p{inmyanmar}', "");
Expect(0, 4255, '\p{^inmyanmar}', "");
Expect(0, 4255, '\P{inmyanmar}', "");
Expect(1, 4255, '\P{^inmyanmar}', "");
Expect(0, 4256, '\p{inmyanmar}', "");
Expect(1, 4256, '\p{^inmyanmar}', "");
Expect(1, 4256, '\P{inmyanmar}', "");
Expect(0, 4256, '\P{^inmyanmar}', "");
Expect(1, 4255, '\p{	IN_Myanmar}', "");
Expect(0, 4255, '\p{^	IN_Myanmar}', "");
Expect(0, 4255, '\P{	IN_Myanmar}', "");
Expect(1, 4255, '\P{^	IN_Myanmar}', "");
Expect(0, 4256, '\p{	IN_Myanmar}', "");
Expect(1, 4256, '\p{^	IN_Myanmar}', "");
Expect(1, 4256, '\P{	IN_Myanmar}', "");
Expect(0, 4256, '\P{^	IN_Myanmar}', "");
Error('\p{_/a/In_Myanmar}');
Error('\P{_/a/In_Myanmar}');
Expect(1, 4255, '\p{_ in_myanmar}', "");
Expect(0, 4255, '\p{^_ in_myanmar}', "");
Expect(0, 4255, '\P{_ in_myanmar}', "");
Expect(1, 4255, '\P{^_ in_myanmar}', "");
Expect(0, 4256, '\p{_ in_myanmar}', "");
Expect(1, 4256, '\p{^_ in_myanmar}', "");
Expect(1, 4256, '\P{_ in_myanmar}', "");
Expect(0, 4256, '\P{^_ in_myanmar}', "");
Error('\p{ /a/IN_nko}');
Error('\P{ /a/IN_nko}');
Expect(1, 2047, '\p{innko}', "");
Expect(0, 2047, '\p{^innko}', "");
Expect(0, 2047, '\P{innko}', "");
Expect(1, 2047, '\P{^innko}', "");
Expect(0, 2048, '\p{innko}', "");
Expect(1, 2048, '\p{^innko}', "");
Expect(1, 2048, '\P{innko}', "");
Expect(0, 2048, '\P{^innko}', "");
Expect(1, 2047, '\p{-In_NKo}', "");
Expect(0, 2047, '\p{^-In_NKo}', "");
Expect(0, 2047, '\P{-In_NKo}', "");
Expect(1, 2047, '\P{^-In_NKo}', "");
Expect(0, 2048, '\p{-In_NKo}', "");
Expect(1, 2048, '\p{^-In_NKo}', "");
Expect(1, 2048, '\P{-In_NKo}', "");
Expect(0, 2048, '\P{^-In_NKo}', "");
Error('\p{	 IN_NKo:=}');
Error('\P{	 IN_NKo:=}');
Expect(1, 2047, '\p{_in_NKO}', "");
Expect(0, 2047, '\p{^_in_NKO}', "");
Expect(0, 2047, '\P{_in_NKO}', "");
Expect(1, 2047, '\P{^_in_NKO}', "");
Expect(0, 2048, '\p{_in_NKO}', "");
Expect(1, 2048, '\p{^_in_NKO}', "");
Expect(1, 2048, '\P{_in_NKO}', "");
Expect(0, 2048, '\P{^_in_NKO}', "");
Error('\p{ :=IN_NEW_TAI_Lue}');
Error('\P{ :=IN_NEW_TAI_Lue}');
Expect(1, 6623, '\p{innewtailue}', "");
Expect(0, 6623, '\p{^innewtailue}', "");
Expect(0, 6623, '\P{innewtailue}', "");
Expect(1, 6623, '\P{^innewtailue}', "");
Expect(0, 6624, '\p{innewtailue}', "");
Expect(1, 6624, '\p{^innewtailue}', "");
Expect(1, 6624, '\P{innewtailue}', "");
Expect(0, 6624, '\P{^innewtailue}', "");
Expect(1, 6623, '\p{	IN_New_Tai_lue}', "");
Expect(0, 6623, '\p{^	IN_New_Tai_lue}', "");
Expect(0, 6623, '\P{	IN_New_Tai_lue}', "");
Expect(1, 6623, '\P{^	IN_New_Tai_lue}', "");
Expect(0, 6624, '\p{	IN_New_Tai_lue}', "");
Expect(1, 6624, '\p{^	IN_New_Tai_lue}', "");
Expect(1, 6624, '\P{	IN_New_Tai_lue}', "");
Expect(0, 6624, '\P{^	IN_New_Tai_lue}', "");
Error('\p{-	in_new_TAI_lue/a/}');
Error('\P{-	in_new_TAI_lue/a/}');
Expect(1, 6623, '\p{ 	In_New_tai_lue}', "");
Expect(0, 6623, '\p{^ 	In_New_tai_lue}', "");
Expect(0, 6623, '\P{ 	In_New_tai_lue}', "");
Expect(1, 6623, '\P{^ 	In_New_tai_lue}', "");
Expect(0, 6624, '\p{ 	In_New_tai_lue}', "");
Expect(1, 6624, '\p{^ 	In_New_tai_lue}', "");
Expect(1, 6624, '\P{ 	In_New_tai_lue}', "");
Expect(0, 6624, '\P{^ 	In_New_tai_lue}', "");
Error('\p{/a/ 	In_Ogham}');
Error('\P{/a/ 	In_Ogham}');
Expect(1, 5791, '\p{inogham}', "");
Expect(0, 5791, '\p{^inogham}', "");
Expect(0, 5791, '\P{inogham}', "");
Expect(1, 5791, '\P{^inogham}', "");
Expect(0, 5792, '\p{inogham}', "");
Expect(1, 5792, '\p{^inogham}', "");
Expect(1, 5792, '\P{inogham}', "");
Expect(0, 5792, '\P{^inogham}', "");
Expect(1, 5791, '\p{_In_OGHAM}', "");
Expect(0, 5791, '\p{^_In_OGHAM}', "");
Expect(0, 5791, '\P{_In_OGHAM}', "");
Expect(1, 5791, '\P{^_In_OGHAM}', "");
Expect(0, 5792, '\p{_In_OGHAM}', "");
Expect(1, 5792, '\p{^_In_OGHAM}', "");
Expect(1, 5792, '\P{_In_OGHAM}', "");
Expect(0, 5792, '\P{^_In_OGHAM}', "");
Error('\p{--In_OGHAM:=}');
Error('\P{--In_OGHAM:=}');
Expect(1, 5791, '\p{	-In_OGHAM}', "");
Expect(0, 5791, '\p{^	-In_OGHAM}', "");
Expect(0, 5791, '\P{	-In_OGHAM}', "");
Expect(1, 5791, '\P{^	-In_OGHAM}', "");
Expect(0, 5792, '\p{	-In_OGHAM}', "");
Expect(1, 5792, '\p{^	-In_OGHAM}', "");
Expect(1, 5792, '\P{	-In_OGHAM}', "");
Expect(0, 5792, '\P{^	-In_OGHAM}', "");
Error('\p{_:=in_Ol_Chiki}');
Error('\P{_:=in_Ol_Chiki}');
Expect(1, 7295, '\p{inolchiki}', "");
Expect(0, 7295, '\p{^inolchiki}', "");
Expect(0, 7295, '\P{inolchiki}', "");
Expect(1, 7295, '\P{^inolchiki}', "");
Expect(0, 7296, '\p{inolchiki}', "");
Expect(1, 7296, '\p{^inolchiki}', "");
Expect(1, 7296, '\P{inolchiki}', "");
Expect(0, 7296, '\P{^inolchiki}', "");
Expect(1, 7295, '\p{	-In_ol_Chiki}', "");
Expect(0, 7295, '\p{^	-In_ol_Chiki}', "");
Expect(0, 7295, '\P{	-In_ol_Chiki}', "");
Expect(1, 7295, '\P{^	-In_ol_Chiki}', "");
Expect(0, 7296, '\p{	-In_ol_Chiki}', "");
Expect(1, 7296, '\p{^	-In_ol_Chiki}', "");
Expect(1, 7296, '\P{	-In_ol_Chiki}', "");
Expect(0, 7296, '\P{^	-In_ol_Chiki}', "");
Error('\p{/a/ 	In_OL_CHIKI}');
Error('\P{/a/ 	In_OL_CHIKI}');
Expect(1, 7295, '\p{_IN_OL_CHIKI}', "");
Expect(0, 7295, '\p{^_IN_OL_CHIKI}', "");
Expect(0, 7295, '\P{_IN_OL_CHIKI}', "");
Expect(1, 7295, '\P{^_IN_OL_CHIKI}', "");
Expect(0, 7296, '\p{_IN_OL_CHIKI}', "");
Expect(1, 7296, '\p{^_IN_OL_CHIKI}', "");
Expect(1, 7296, '\P{_IN_OL_CHIKI}', "");
Expect(0, 7296, '\P{^_IN_OL_CHIKI}', "");
Error('\p{-IN_Old_italic/a/}');
Error('\P{-IN_Old_italic/a/}');
Expect(1, 66351, '\p{inolditalic}', "");
Expect(0, 66351, '\p{^inolditalic}', "");
Expect(0, 66351, '\P{inolditalic}', "");
Expect(1, 66351, '\P{^inolditalic}', "");
Expect(0, 66352, '\p{inolditalic}', "");
Expect(1, 66352, '\p{^inolditalic}', "");
Expect(1, 66352, '\P{inolditalic}', "");
Expect(0, 66352, '\P{^inolditalic}', "");
Expect(1, 66351, '\p{	_in_Old_Italic}', "");
Expect(0, 66351, '\p{^	_in_Old_Italic}', "");
Expect(0, 66351, '\P{	_in_Old_Italic}', "");
Expect(1, 66351, '\P{^	_in_Old_Italic}', "");
Expect(0, 66352, '\p{	_in_Old_Italic}', "");
Expect(1, 66352, '\p{^	_in_Old_Italic}', "");
Expect(1, 66352, '\P{	_in_Old_Italic}', "");
Expect(0, 66352, '\P{^	_in_Old_Italic}', "");
Error('\p{__in_Old_Italic/a/}');
Error('\P{__in_Old_Italic/a/}');
Expect(1, 66351, '\p{	In_Old_Italic}', "");
Expect(0, 66351, '\p{^	In_Old_Italic}', "");
Expect(0, 66351, '\P{	In_Old_Italic}', "");
Expect(1, 66351, '\P{^	In_Old_Italic}', "");
Expect(0, 66352, '\p{	In_Old_Italic}', "");
Expect(1, 66352, '\p{^	In_Old_Italic}', "");
Expect(1, 66352, '\P{	In_Old_Italic}', "");
Expect(0, 66352, '\P{^	In_Old_Italic}', "");
Error('\p{_	IN_OLD_persian/a/}');
Error('\P{_	IN_OLD_persian/a/}');
Expect(1, 66527, '\p{inoldpersian}', "");
Expect(0, 66527, '\p{^inoldpersian}', "");
Expect(0, 66527, '\P{inoldpersian}', "");
Expect(1, 66527, '\P{^inoldpersian}', "");
Expect(0, 66528, '\p{inoldpersian}', "");
Expect(1, 66528, '\p{^inoldpersian}', "");
Expect(1, 66528, '\P{inoldpersian}', "");
Expect(0, 66528, '\P{^inoldpersian}', "");
Expect(1, 66527, '\p{	in_old_Persian}', "");
Expect(0, 66527, '\p{^	in_old_Persian}', "");
Expect(0, 66527, '\P{	in_old_Persian}', "");
Expect(1, 66527, '\P{^	in_old_Persian}', "");
Expect(0, 66528, '\p{	in_old_Persian}', "");
Expect(1, 66528, '\p{^	in_old_Persian}', "");
Expect(1, 66528, '\P{	in_old_Persian}', "");
Expect(0, 66528, '\P{^	in_old_Persian}', "");
Error('\p{--In_Old_PERSIAN:=}');
Error('\P{--In_Old_PERSIAN:=}');
Expect(1, 66527, '\p{_-IN_old_Persian}', "");
Expect(0, 66527, '\p{^_-IN_old_Persian}', "");
Expect(0, 66527, '\P{_-IN_old_Persian}', "");
Expect(1, 66527, '\P{^_-IN_old_Persian}', "");
Expect(0, 66528, '\p{_-IN_old_Persian}', "");
Expect(1, 66528, '\p{^_-IN_old_Persian}', "");
Expect(1, 66528, '\P{_-IN_old_Persian}', "");
Expect(0, 66528, '\P{^_-IN_old_Persian}', "");
Error('\p{ /a/In_Old_South_arabian}');
Error('\P{ /a/In_Old_South_arabian}');
Expect(1, 68223, '\p{inoldsoutharabian}', "");
Expect(0, 68223, '\p{^inoldsoutharabian}', "");
Expect(0, 68223, '\P{inoldsoutharabian}', "");
Expect(1, 68223, '\P{^inoldsoutharabian}', "");
Expect(0, 68224, '\p{inoldsoutharabian}', "");
Expect(1, 68224, '\p{^inoldsoutharabian}', "");
Expect(1, 68224, '\P{inoldsoutharabian}', "");
Expect(0, 68224, '\P{^inoldsoutharabian}', "");
Expect(1, 68223, '\p{__IN_old_south_arabian}', "");
Expect(0, 68223, '\p{^__IN_old_south_arabian}', "");
Expect(0, 68223, '\P{__IN_old_south_arabian}', "");
Expect(1, 68223, '\P{^__IN_old_south_arabian}', "");
Expect(0, 68224, '\p{__IN_old_south_arabian}', "");
Expect(1, 68224, '\p{^__IN_old_south_arabian}', "");
Expect(1, 68224, '\P{__IN_old_south_arabian}', "");
Expect(0, 68224, '\P{^__IN_old_south_arabian}', "");
Error('\p{_ In_OLD_South_Arabian/a/}');
Error('\P{_ In_OLD_South_Arabian/a/}');
Expect(1, 68223, '\p{-	IN_Old_SOUTH_arabian}', "");
Expect(0, 68223, '\p{^-	IN_Old_SOUTH_arabian}', "");
Expect(0, 68223, '\P{-	IN_Old_SOUTH_arabian}', "");
Expect(1, 68223, '\P{^-	IN_Old_SOUTH_arabian}', "");
Expect(0, 68224, '\p{-	IN_Old_SOUTH_arabian}', "");
Expect(1, 68224, '\p{^-	IN_Old_SOUTH_arabian}', "");
Expect(1, 68224, '\P{-	IN_Old_SOUTH_arabian}', "");
Expect(0, 68224, '\P{^-	IN_Old_SOUTH_arabian}', "");
Error('\p{ /a/In_Old_turkic}');
Error('\P{ /a/In_Old_turkic}');
Expect(1, 68687, '\p{inoldturkic}', "");
Expect(0, 68687, '\p{^inoldturkic}', "");
Expect(0, 68687, '\P{inoldturkic}', "");
Expect(1, 68687, '\P{^inoldturkic}', "");
Expect(0, 68688, '\p{inoldturkic}', "");
Expect(1, 68688, '\p{^inoldturkic}', "");
Expect(1, 68688, '\P{inoldturkic}', "");
Expect(0, 68688, '\P{^inoldturkic}', "");
Expect(1, 68687, '\p{-	IN_Old_Turkic}', "");
Expect(0, 68687, '\p{^-	IN_Old_Turkic}', "");
Expect(0, 68687, '\P{-	IN_Old_Turkic}', "");
Expect(1, 68687, '\P{^-	IN_Old_Turkic}', "");
Expect(0, 68688, '\p{-	IN_Old_Turkic}', "");
Expect(1, 68688, '\p{^-	IN_Old_Turkic}', "");
Expect(1, 68688, '\P{-	IN_Old_Turkic}', "");
Expect(0, 68688, '\P{^-	IN_Old_Turkic}', "");
Error('\p{-:=In_Old_Turkic}');
Error('\P{-:=In_Old_Turkic}');
Expect(1, 68687, '\p{	 in_Old_turkic}', "");
Expect(0, 68687, '\p{^	 in_Old_turkic}', "");
Expect(0, 68687, '\P{	 in_Old_turkic}', "");
Expect(1, 68687, '\P{^	 in_Old_turkic}', "");
Expect(0, 68688, '\p{	 in_Old_turkic}', "");
Expect(1, 68688, '\p{^	 in_Old_turkic}', "");
Expect(1, 68688, '\P{	 in_Old_turkic}', "");
Expect(0, 68688, '\P{^	 in_Old_turkic}', "");
Error('\p{-in_ORIYA:=}');
Error('\P{-in_ORIYA:=}');
Expect(1, 2943, '\p{inoriya}', "");
Expect(0, 2943, '\p{^inoriya}', "");
Expect(0, 2943, '\P{inoriya}', "");
Expect(1, 2943, '\P{^inoriya}', "");
Expect(0, 2944, '\p{inoriya}', "");
Expect(1, 2944, '\p{^inoriya}', "");
Expect(1, 2944, '\P{inoriya}', "");
Expect(0, 2944, '\P{^inoriya}', "");
Expect(1, 2943, '\p{ in_oriya}', "");
Expect(0, 2943, '\p{^ in_oriya}', "");
Expect(0, 2943, '\P{ in_oriya}', "");
Expect(1, 2943, '\P{^ in_oriya}', "");
Expect(0, 2944, '\p{ in_oriya}', "");
Expect(1, 2944, '\p{^ in_oriya}', "");
Expect(1, 2944, '\P{ in_oriya}', "");
Expect(0, 2944, '\P{^ in_oriya}', "");
Error('\p{_	in_ORIYA:=}');
Error('\P{_	in_ORIYA:=}');
Expect(1, 2943, '\p{--In_ORIYA}', "");
Expect(0, 2943, '\p{^--In_ORIYA}', "");
Expect(0, 2943, '\P{--In_ORIYA}', "");
Expect(1, 2943, '\P{^--In_ORIYA}', "");
Expect(0, 2944, '\p{--In_ORIYA}', "");
Expect(1, 2944, '\p{^--In_ORIYA}', "");
Expect(1, 2944, '\P{--In_ORIYA}', "");
Expect(0, 2944, '\P{^--In_ORIYA}', "");
Error('\p{	_in_Osmanya/a/}');
Error('\P{	_in_Osmanya/a/}');
Expect(1, 66735, '\p{inosmanya}', "");
Expect(0, 66735, '\p{^inosmanya}', "");
Expect(0, 66735, '\P{inosmanya}', "");
Expect(1, 66735, '\P{^inosmanya}', "");
Expect(0, 66736, '\p{inosmanya}', "");
Expect(1, 66736, '\p{^inosmanya}', "");
Expect(1, 66736, '\P{inosmanya}', "");
Expect(0, 66736, '\P{^inosmanya}', "");
Expect(1, 66735, '\p{ -In_osmanya}', "");
Expect(0, 66735, '\p{^ -In_osmanya}', "");
Expect(0, 66735, '\P{ -In_osmanya}', "");
Expect(1, 66735, '\P{^ -In_osmanya}', "");
Expect(0, 66736, '\p{ -In_osmanya}', "");
Expect(1, 66736, '\p{^ -In_osmanya}', "");
Expect(1, 66736, '\P{ -In_osmanya}', "");
Expect(0, 66736, '\P{^ -In_osmanya}', "");
Error('\p{_/a/In_osmanya}');
Error('\P{_/a/In_osmanya}');
Expect(1, 66735, '\p{_-in_OSMANYA}', "");
Expect(0, 66735, '\p{^_-in_OSMANYA}', "");
Expect(0, 66735, '\P{_-in_OSMANYA}', "");
Expect(1, 66735, '\P{^_-in_OSMANYA}', "");
Expect(0, 66736, '\p{_-in_OSMANYA}', "");
Expect(1, 66736, '\p{^_-in_OSMANYA}', "");
Expect(1, 66736, '\P{_-in_OSMANYA}', "");
Expect(0, 66736, '\P{^_-in_OSMANYA}', "");
Error('\p{:= IN_Phags_Pa}');
Error('\P{:= IN_Phags_Pa}');
Expect(1, 43135, '\p{inphagspa}', "");
Expect(0, 43135, '\p{^inphagspa}', "");
Expect(0, 43135, '\P{inphagspa}', "");
Expect(1, 43135, '\P{^inphagspa}', "");
Expect(0, 43136, '\p{inphagspa}', "");
Expect(1, 43136, '\p{^inphagspa}', "");
Expect(1, 43136, '\P{inphagspa}', "");
Expect(0, 43136, '\P{^inphagspa}', "");
Expect(1, 43135, '\p{_-In_PHAGS_pa}', "");
Expect(0, 43135, '\p{^_-In_PHAGS_pa}', "");
Expect(0, 43135, '\P{_-In_PHAGS_pa}', "");
Expect(1, 43135, '\P{^_-In_PHAGS_pa}', "");
Expect(0, 43136, '\p{_-In_PHAGS_pa}', "");
Expect(1, 43136, '\p{^_-In_PHAGS_pa}', "");
Expect(1, 43136, '\P{_-In_PHAGS_pa}', "");
Expect(0, 43136, '\P{^_-In_PHAGS_pa}', "");
Error('\p{_/a/IN_Phags_Pa}');
Error('\P{_/a/IN_Phags_Pa}');
Expect(1, 43135, '\p{		In_PHAGS_pa}', "");
Expect(0, 43135, '\p{^		In_PHAGS_pa}', "");
Expect(0, 43135, '\P{		In_PHAGS_pa}', "");
Expect(1, 43135, '\P{^		In_PHAGS_pa}', "");
Expect(0, 43136, '\p{		In_PHAGS_pa}', "");
Expect(1, 43136, '\p{^		In_PHAGS_pa}', "");
Expect(1, 43136, '\P{		In_PHAGS_pa}', "");
Expect(0, 43136, '\P{^		In_PHAGS_pa}', "");
Error('\p{-	in_phoenician/a/}');
Error('\P{-	in_phoenician/a/}');
Expect(1, 67871, '\p{inphoenician}', "");
Expect(0, 67871, '\p{^inphoenician}', "");
Expect(0, 67871, '\P{inphoenician}', "");
Expect(1, 67871, '\P{^inphoenician}', "");
Expect(0, 67872, '\p{inphoenician}', "");
Expect(1, 67872, '\p{^inphoenician}', "");
Expect(1, 67872, '\P{inphoenician}', "");
Expect(0, 67872, '\P{^inphoenician}', "");
Expect(1, 67871, '\p{ 	in_Phoenician}', "");
Expect(0, 67871, '\p{^ 	in_Phoenician}', "");
Expect(0, 67871, '\P{ 	in_Phoenician}', "");
Expect(1, 67871, '\P{^ 	in_Phoenician}', "");
Expect(0, 67872, '\p{ 	in_Phoenician}', "");
Expect(1, 67872, '\p{^ 	in_Phoenician}', "");
Expect(1, 67872, '\P{ 	in_Phoenician}', "");
Expect(0, 67872, '\P{^ 	in_Phoenician}', "");
Error('\p{_:=IN_Phoenician}');
Error('\P{_:=IN_Phoenician}');
Expect(1, 67871, '\p{_ in_Phoenician}', "");
Expect(0, 67871, '\p{^_ in_Phoenician}', "");
Expect(0, 67871, '\P{_ in_Phoenician}', "");
Expect(1, 67871, '\P{^_ in_Phoenician}', "");
Expect(0, 67872, '\p{_ in_Phoenician}', "");
Expect(1, 67872, '\p{^_ in_Phoenician}', "");
Expect(1, 67872, '\P{_ in_Phoenician}', "");
Expect(0, 67872, '\P{^_ in_Phoenician}', "");
Error('\p{/a/	 in_Rejang}');
Error('\P{/a/	 in_Rejang}');
Expect(1, 43359, '\p{inrejang}', "");
Expect(0, 43359, '\p{^inrejang}', "");
Expect(0, 43359, '\P{inrejang}', "");
Expect(1, 43359, '\P{^inrejang}', "");
Expect(0, 43360, '\p{inrejang}', "");
Expect(1, 43360, '\p{^inrejang}', "");
Expect(1, 43360, '\P{inrejang}', "");
Expect(0, 43360, '\P{^inrejang}', "");
Expect(1, 43359, '\p{-	in_REJANG}', "");
Expect(0, 43359, '\p{^-	in_REJANG}', "");
Expect(0, 43359, '\P{-	in_REJANG}', "");
Expect(1, 43359, '\P{^-	in_REJANG}', "");
Expect(0, 43360, '\p{-	in_REJANG}', "");
Expect(1, 43360, '\p{^-	in_REJANG}', "");
Expect(1, 43360, '\P{-	in_REJANG}', "");
Expect(0, 43360, '\P{^-	in_REJANG}', "");
Error('\p{:=-IN_Rejang}');
Error('\P{:=-IN_Rejang}');
Expect(1, 43359, '\p{		In_Rejang}', "");
Expect(0, 43359, '\p{^		In_Rejang}', "");
Expect(0, 43359, '\P{		In_Rejang}', "");
Expect(1, 43359, '\P{^		In_Rejang}', "");
Expect(0, 43360, '\p{		In_Rejang}', "");
Expect(1, 43360, '\p{^		In_Rejang}', "");
Expect(1, 43360, '\P{		In_Rejang}', "");
Expect(0, 43360, '\P{^		In_Rejang}', "");
Error('\p{-:=In_Runic}');
Error('\P{-:=In_Runic}');
Expect(1, 5887, '\p{inrunic}', "");
Expect(0, 5887, '\p{^inrunic}', "");
Expect(0, 5887, '\P{inrunic}', "");
Expect(1, 5887, '\P{^inrunic}', "");
Expect(0, 5888, '\p{inrunic}', "");
Expect(1, 5888, '\p{^inrunic}', "");
Expect(1, 5888, '\P{inrunic}', "");
Expect(0, 5888, '\P{^inrunic}', "");
Expect(1, 5887, '\p{--In_Runic}', "");
Expect(0, 5887, '\p{^--In_Runic}', "");
Expect(0, 5887, '\P{--In_Runic}', "");
Expect(1, 5887, '\P{^--In_Runic}', "");
Expect(0, 5888, '\p{--In_Runic}', "");
Expect(1, 5888, '\p{^--In_Runic}', "");
Expect(1, 5888, '\P{--In_Runic}', "");
Expect(0, 5888, '\P{^--In_Runic}', "");
Error('\p{-:=in_Runic}');
Error('\P{-:=in_Runic}');
Expect(1, 5887, '\p{-in_RUNIC}', "");
Expect(0, 5887, '\p{^-in_RUNIC}', "");
Expect(0, 5887, '\P{-in_RUNIC}', "");
Expect(1, 5887, '\P{^-in_RUNIC}', "");
Expect(0, 5888, '\p{-in_RUNIC}', "");
Expect(1, 5888, '\p{^-in_RUNIC}', "");
Expect(1, 5888, '\P{-in_RUNIC}', "");
Expect(0, 5888, '\P{^-in_RUNIC}', "");
Error('\p{_In_Samaritan/a/}');
Error('\P{_In_Samaritan/a/}');
Expect(1, 2111, '\p{insamaritan}', "");
Expect(0, 2111, '\p{^insamaritan}', "");
Expect(0, 2111, '\P{insamaritan}', "");
Expect(1, 2111, '\P{^insamaritan}', "");
Expect(0, 2112, '\p{insamaritan}', "");
Expect(1, 2112, '\p{^insamaritan}', "");
Expect(1, 2112, '\P{insamaritan}', "");
Expect(0, 2112, '\P{^insamaritan}', "");
Expect(1, 2111, '\p{in_Samaritan}', "");
Expect(0, 2111, '\p{^in_Samaritan}', "");
Expect(0, 2111, '\P{in_Samaritan}', "");
Expect(1, 2111, '\P{^in_Samaritan}', "");
Expect(0, 2112, '\p{in_Samaritan}', "");
Expect(1, 2112, '\p{^in_Samaritan}', "");
Expect(1, 2112, '\P{in_Samaritan}', "");
Expect(0, 2112, '\P{^in_Samaritan}', "");
Error('\p{:= 	in_Samaritan}');
Error('\P{:= 	in_Samaritan}');
Expect(1, 2111, '\p{ _IN_Samaritan}', "");
Expect(0, 2111, '\p{^ _IN_Samaritan}', "");
Expect(0, 2111, '\P{ _IN_Samaritan}', "");
Expect(1, 2111, '\P{^ _IN_Samaritan}', "");
Expect(0, 2112, '\p{ _IN_Samaritan}', "");
Expect(1, 2112, '\p{^ _IN_Samaritan}', "");
Expect(1, 2112, '\P{ _IN_Samaritan}', "");
Expect(0, 2112, '\P{^ _IN_Samaritan}', "");
Error('\p{--In_saurashtra:=}');
Error('\P{--In_saurashtra:=}');
Expect(1, 43231, '\p{insaurashtra}', "");
Expect(0, 43231, '\p{^insaurashtra}', "");
Expect(0, 43231, '\P{insaurashtra}', "");
Expect(1, 43231, '\P{^insaurashtra}', "");
Expect(0, 43232, '\p{insaurashtra}', "");
Expect(1, 43232, '\p{^insaurashtra}', "");
Expect(1, 43232, '\P{insaurashtra}', "");
Expect(0, 43232, '\P{^insaurashtra}', "");
Expect(1, 43231, '\p{_-In_Saurashtra}', "");
Expect(0, 43231, '\p{^_-In_Saurashtra}', "");
Expect(0, 43231, '\P{_-In_Saurashtra}', "");
Expect(1, 43231, '\P{^_-In_Saurashtra}', "");
Expect(0, 43232, '\p{_-In_Saurashtra}', "");
Expect(1, 43232, '\p{^_-In_Saurashtra}', "");
Expect(1, 43232, '\P{_-In_Saurashtra}', "");
Expect(0, 43232, '\P{^_-In_Saurashtra}', "");
Error('\p{/a/In_SAURASHTRA}');
Error('\P{/a/In_SAURASHTRA}');
Expect(1, 43231, '\p{_In_saurashtra}', "");
Expect(0, 43231, '\p{^_In_saurashtra}', "");
Expect(0, 43231, '\P{_In_saurashtra}', "");
Expect(1, 43231, '\P{^_In_saurashtra}', "");
Expect(0, 43232, '\p{_In_saurashtra}', "");
Expect(1, 43232, '\p{^_In_saurashtra}', "");
Expect(1, 43232, '\P{_In_saurashtra}', "");
Expect(0, 43232, '\P{^_In_saurashtra}', "");
Error('\p{/a/	-In_sharada}');
Error('\P{/a/	-In_sharada}');
Expect(1, 70111, '\p{insharada}', "");
Expect(0, 70111, '\p{^insharada}', "");
Expect(0, 70111, '\P{insharada}', "");
Expect(1, 70111, '\P{^insharada}', "");
Expect(0, 70112, '\p{insharada}', "");
Expect(1, 70112, '\p{^insharada}', "");
Expect(1, 70112, '\P{insharada}', "");
Expect(0, 70112, '\P{^insharada}', "");
Expect(1, 70111, '\p{	In_SHARADA}', "");
Expect(0, 70111, '\p{^	In_SHARADA}', "");
Expect(0, 70111, '\P{	In_SHARADA}', "");
Expect(1, 70111, '\P{^	In_SHARADA}', "");
Expect(0, 70112, '\p{	In_SHARADA}', "");
Expect(1, 70112, '\p{^	In_SHARADA}', "");
Expect(1, 70112, '\P{	In_SHARADA}', "");
Expect(0, 70112, '\P{^	In_SHARADA}', "");
Error('\p{	/a/IN_Sharada}');
Error('\P{	/a/IN_Sharada}');
Expect(1, 70111, '\p{_-in_Sharada}', "");
Expect(0, 70111, '\p{^_-in_Sharada}', "");
Expect(0, 70111, '\P{_-in_Sharada}', "");
Expect(1, 70111, '\P{^_-in_Sharada}', "");
Expect(0, 70112, '\p{_-in_Sharada}', "");
Expect(1, 70112, '\p{^_-in_Sharada}', "");
Expect(1, 70112, '\P{_-in_Sharada}', "");
Expect(0, 70112, '\P{^_-in_Sharada}', "");
Error('\p{/a/	In_shavian}');
Error('\P{/a/	In_shavian}');
Expect(1, 66687, '\p{inshavian}', "");
Expect(0, 66687, '\p{^inshavian}', "");
Expect(0, 66687, '\P{inshavian}', "");
Expect(1, 66687, '\P{^inshavian}', "");
Expect(0, 66688, '\p{inshavian}', "");
Expect(1, 66688, '\p{^inshavian}', "");
Expect(1, 66688, '\P{inshavian}', "");
Expect(0, 66688, '\P{^inshavian}', "");
Expect(1, 66687, '\p{ 	IN_SHAVIAN}', "");
Expect(0, 66687, '\p{^ 	IN_SHAVIAN}', "");
Expect(0, 66687, '\P{ 	IN_SHAVIAN}', "");
Expect(1, 66687, '\P{^ 	IN_SHAVIAN}', "");
Expect(0, 66688, '\p{ 	IN_SHAVIAN}', "");
Expect(1, 66688, '\p{^ 	IN_SHAVIAN}', "");
Expect(1, 66688, '\P{ 	IN_SHAVIAN}', "");
Expect(0, 66688, '\P{^ 	IN_SHAVIAN}', "");
Error('\p{_ IN_Shavian/a/}');
Error('\P{_ IN_Shavian/a/}');
Expect(1, 66687, '\p{ In_Shavian}', "");
Expect(0, 66687, '\p{^ In_Shavian}', "");
Expect(0, 66687, '\P{ In_Shavian}', "");
Expect(1, 66687, '\P{^ In_Shavian}', "");
Expect(0, 66688, '\p{ In_Shavian}', "");
Expect(1, 66688, '\p{^ In_Shavian}', "");
Expect(1, 66688, '\P{ In_Shavian}', "");
Expect(0, 66688, '\P{^ In_Shavian}', "");
Error('\p{/a/ In_SINHALA}');
Error('\P{/a/ In_SINHALA}');
Expect(1, 3583, '\p{insinhala}', "");
Expect(0, 3583, '\p{^insinhala}', "");
Expect(0, 3583, '\P{insinhala}', "");
Expect(1, 3583, '\P{^insinhala}', "");
Expect(0, 3584, '\p{insinhala}', "");
Expect(1, 3584, '\p{^insinhala}', "");
Expect(1, 3584, '\P{insinhala}', "");
Expect(0, 3584, '\P{^insinhala}', "");
Expect(1, 3583, '\p{- in_Sinhala}', "");
Expect(0, 3583, '\p{^- in_Sinhala}', "");
Expect(0, 3583, '\P{- in_Sinhala}', "");
Expect(1, 3583, '\P{^- in_Sinhala}', "");
Expect(0, 3584, '\p{- in_Sinhala}', "");
Expect(1, 3584, '\p{^- in_Sinhala}', "");
Expect(1, 3584, '\P{- in_Sinhala}', "");
Expect(0, 3584, '\P{^- in_Sinhala}', "");
Error('\p{/a/	_In_Sinhala}');
Error('\P{/a/	_In_Sinhala}');
Expect(1, 3583, '\p{ IN_Sinhala}', "");
Expect(0, 3583, '\p{^ IN_Sinhala}', "");
Expect(0, 3583, '\P{ IN_Sinhala}', "");
Expect(1, 3583, '\P{^ IN_Sinhala}', "");
Expect(0, 3584, '\p{ IN_Sinhala}', "");
Expect(1, 3584, '\p{^ IN_Sinhala}', "");
Expect(1, 3584, '\P{ IN_Sinhala}', "");
Expect(0, 3584, '\P{^ IN_Sinhala}', "");
Error('\p{ in_Sora_sompeng/a/}');
Error('\P{ in_Sora_sompeng/a/}');
Expect(1, 69887, '\p{insorasompeng}', "");
Expect(0, 69887, '\p{^insorasompeng}', "");
Expect(0, 69887, '\P{insorasompeng}', "");
Expect(1, 69887, '\P{^insorasompeng}', "");
Expect(0, 69888, '\p{insorasompeng}', "");
Expect(1, 69888, '\p{^insorasompeng}', "");
Expect(1, 69888, '\P{insorasompeng}', "");
Expect(0, 69888, '\P{^insorasompeng}', "");
Expect(1, 69887, '\p{ _in_Sora_sompeng}', "");
Expect(0, 69887, '\p{^ _in_Sora_sompeng}', "");
Expect(0, 69887, '\P{ _in_Sora_sompeng}', "");
Expect(1, 69887, '\P{^ _in_Sora_sompeng}', "");
Expect(0, 69888, '\p{ _in_Sora_sompeng}', "");
Expect(1, 69888, '\p{^ _in_Sora_sompeng}', "");
Expect(1, 69888, '\P{ _in_Sora_sompeng}', "");
Expect(0, 69888, '\P{^ _in_Sora_sompeng}', "");
Error('\p{-:=In_sora_Sompeng}');
Error('\P{-:=In_sora_Sompeng}');
Expect(1, 69887, '\p{_in_sora_Sompeng}', "");
Expect(0, 69887, '\p{^_in_sora_Sompeng}', "");
Expect(0, 69887, '\P{_in_sora_Sompeng}', "");
Expect(1, 69887, '\P{^_in_sora_Sompeng}', "");
Expect(0, 69888, '\p{_in_sora_Sompeng}', "");
Expect(1, 69888, '\p{^_in_sora_Sompeng}', "");
Expect(1, 69888, '\P{_in_sora_Sompeng}', "");
Expect(0, 69888, '\P{^_in_sora_Sompeng}', "");
Error('\p{:=- In_sundanese}');
Error('\P{:=- In_sundanese}');
Expect(1, 7103, '\p{insundanese}', "");
Expect(0, 7103, '\p{^insundanese}', "");
Expect(0, 7103, '\P{insundanese}', "");
Expect(1, 7103, '\P{^insundanese}', "");
Expect(0, 7104, '\p{insundanese}', "");
Expect(1, 7104, '\p{^insundanese}', "");
Expect(1, 7104, '\P{insundanese}', "");
Expect(0, 7104, '\P{^insundanese}', "");
Expect(1, 7103, '\p{__in_SUNDANESE}', "");
Expect(0, 7103, '\p{^__in_SUNDANESE}', "");
Expect(0, 7103, '\P{__in_SUNDANESE}', "");
Expect(1, 7103, '\P{^__in_SUNDANESE}', "");
Expect(0, 7104, '\p{__in_SUNDANESE}', "");
Expect(1, 7104, '\p{^__in_SUNDANESE}', "");
Expect(1, 7104, '\P{__in_SUNDANESE}', "");
Expect(0, 7104, '\P{^__in_SUNDANESE}', "");
Error('\p{/a/_-In_SUNDANESE}');
Error('\P{/a/_-In_SUNDANESE}');
Expect(1, 7103, '\p{		In_SUNDANESE}', "");
Expect(0, 7103, '\p{^		In_SUNDANESE}', "");
Expect(0, 7103, '\P{		In_SUNDANESE}', "");
Expect(1, 7103, '\P{^		In_SUNDANESE}', "");
Expect(0, 7104, '\p{		In_SUNDANESE}', "");
Expect(1, 7104, '\p{^		In_SUNDANESE}', "");
Expect(1, 7104, '\P{		In_SUNDANESE}', "");
Expect(0, 7104, '\P{^		In_SUNDANESE}', "");
Error('\p{ /a/In_syloti_Nagri}');
Error('\P{ /a/In_syloti_Nagri}');
Expect(1, 43055, '\p{insylotinagri}', "");
Expect(0, 43055, '\p{^insylotinagri}', "");
Expect(0, 43055, '\P{insylotinagri}', "");
Expect(1, 43055, '\P{^insylotinagri}', "");
Expect(0, 43056, '\p{insylotinagri}', "");
Expect(1, 43056, '\p{^insylotinagri}', "");
Expect(1, 43056, '\P{insylotinagri}', "");
Expect(0, 43056, '\P{^insylotinagri}', "");
Expect(1, 43055, '\p{-In_syloti_nagri}', "");
Expect(0, 43055, '\p{^-In_syloti_nagri}', "");
Expect(0, 43055, '\P{-In_syloti_nagri}', "");
Expect(1, 43055, '\P{^-In_syloti_nagri}', "");
Expect(0, 43056, '\p{-In_syloti_nagri}', "");
Expect(1, 43056, '\p{^-In_syloti_nagri}', "");
Expect(1, 43056, '\P{-In_syloti_nagri}', "");
Expect(0, 43056, '\P{^-In_syloti_nagri}', "");
Error('\p{- in_SYLOTI_NAGRI/a/}');
Error('\P{- in_SYLOTI_NAGRI/a/}');
Expect(1, 43055, '\p{--IN_Syloti_Nagri}', "");
Expect(0, 43055, '\p{^--IN_Syloti_Nagri}', "");
Expect(0, 43055, '\P{--IN_Syloti_Nagri}', "");
Expect(1, 43055, '\P{^--IN_Syloti_Nagri}', "");
Expect(0, 43056, '\p{--IN_Syloti_Nagri}', "");
Expect(1, 43056, '\p{^--IN_Syloti_Nagri}', "");
Expect(1, 43056, '\P{--IN_Syloti_Nagri}', "");
Expect(0, 43056, '\P{^--IN_Syloti_Nagri}', "");
Error('\p{/a/	-IN_Syriac}');
Error('\P{/a/	-IN_Syriac}');
Expect(1, 1871, '\p{insyriac}', "");
Expect(0, 1871, '\p{^insyriac}', "");
Expect(0, 1871, '\P{insyriac}', "");
Expect(1, 1871, '\P{^insyriac}', "");
Expect(0, 1872, '\p{insyriac}', "");
Expect(1, 1872, '\p{^insyriac}', "");
Expect(1, 1872, '\P{insyriac}', "");
Expect(0, 1872, '\P{^insyriac}', "");
Expect(1, 1871, '\p{	in_SYRIAC}', "");
Expect(0, 1871, '\p{^	in_SYRIAC}', "");
Expect(0, 1871, '\P{	in_SYRIAC}', "");
Expect(1, 1871, '\P{^	in_SYRIAC}', "");
Expect(0, 1872, '\p{	in_SYRIAC}', "");
Expect(1, 1872, '\p{^	in_SYRIAC}', "");
Expect(1, 1872, '\P{	in_SYRIAC}', "");
Expect(0, 1872, '\P{^	in_SYRIAC}', "");
Error('\p{ /a/in_SYRIAC}');
Error('\P{ /a/in_SYRIAC}');
Expect(1, 1871, '\p{	In_syriac}', "");
Expect(0, 1871, '\p{^	In_syriac}', "");
Expect(0, 1871, '\P{	In_syriac}', "");
Expect(1, 1871, '\P{^	In_syriac}', "");
Expect(0, 1872, '\p{	In_syriac}', "");
Expect(1, 1872, '\p{^	In_syriac}', "");
Expect(1, 1872, '\P{	In_syriac}', "");
Expect(0, 1872, '\P{^	In_syriac}', "");
Error('\p{-/a/In_Tagalog}');
Error('\P{-/a/In_Tagalog}');
Expect(1, 5919, '\p{intagalog}', "");
Expect(0, 5919, '\p{^intagalog}', "");
Expect(0, 5919, '\P{intagalog}', "");
Expect(1, 5919, '\P{^intagalog}', "");
Expect(0, 5920, '\p{intagalog}', "");
Expect(1, 5920, '\p{^intagalog}', "");
Expect(1, 5920, '\P{intagalog}', "");
Expect(0, 5920, '\P{^intagalog}', "");
Expect(1, 5919, '\p{	 In_Tagalog}', "");
Expect(0, 5919, '\p{^	 In_Tagalog}', "");
Expect(0, 5919, '\P{	 In_Tagalog}', "");
Expect(1, 5919, '\P{^	 In_Tagalog}', "");
Expect(0, 5920, '\p{	 In_Tagalog}', "");
Expect(1, 5920, '\p{^	 In_Tagalog}', "");
Expect(1, 5920, '\P{	 In_Tagalog}', "");
Expect(0, 5920, '\P{^	 In_Tagalog}', "");
Error('\p{/a/IN_TAGALOG}');
Error('\P{/a/IN_TAGALOG}');
Expect(1, 5919, '\p{_ In_TAGALOG}', "");
Expect(0, 5919, '\p{^_ In_TAGALOG}', "");
Expect(0, 5919, '\P{_ In_TAGALOG}', "");
Expect(1, 5919, '\P{^_ In_TAGALOG}', "");
Expect(0, 5920, '\p{_ In_TAGALOG}', "");
Expect(1, 5920, '\p{^_ In_TAGALOG}', "");
Expect(1, 5920, '\P{_ In_TAGALOG}', "");
Expect(0, 5920, '\P{^_ In_TAGALOG}', "");
Error('\p{/a/in_Tagbanwa}');
Error('\P{/a/in_Tagbanwa}');
Expect(1, 6015, '\p{intagbanwa}', "");
Expect(0, 6015, '\p{^intagbanwa}', "");
Expect(0, 6015, '\P{intagbanwa}', "");
Expect(1, 6015, '\P{^intagbanwa}', "");
Expect(0, 6016, '\p{intagbanwa}', "");
Expect(1, 6016, '\p{^intagbanwa}', "");
Expect(1, 6016, '\P{intagbanwa}', "");
Expect(0, 6016, '\P{^intagbanwa}', "");
Expect(1, 6015, '\p{-In_TAGBANWA}', "");
Expect(0, 6015, '\p{^-In_TAGBANWA}', "");
Expect(0, 6015, '\P{-In_TAGBANWA}', "");
Expect(1, 6015, '\P{^-In_TAGBANWA}', "");
Expect(0, 6016, '\p{-In_TAGBANWA}', "");
Expect(1, 6016, '\p{^-In_TAGBANWA}', "");
Expect(1, 6016, '\P{-In_TAGBANWA}', "");
Expect(0, 6016, '\P{^-In_TAGBANWA}', "");
Error('\p{/a/-In_Tagbanwa}');
Error('\P{/a/-In_Tagbanwa}');
Expect(1, 6015, '\p{	IN_Tagbanwa}', "");
Expect(0, 6015, '\p{^	IN_Tagbanwa}', "");
Expect(0, 6015, '\P{	IN_Tagbanwa}', "");
Expect(1, 6015, '\P{^	IN_Tagbanwa}', "");
Expect(0, 6016, '\p{	IN_Tagbanwa}', "");
Expect(1, 6016, '\p{^	IN_Tagbanwa}', "");
Expect(1, 6016, '\P{	IN_Tagbanwa}', "");
Expect(0, 6016, '\P{^	IN_Tagbanwa}', "");
Error('\p{/a/-_In_tai_Le}');
Error('\P{/a/-_In_tai_Le}');
Expect(1, 6527, '\p{intaile}', "");
Expect(0, 6527, '\p{^intaile}', "");
Expect(0, 6527, '\P{intaile}', "");
Expect(1, 6527, '\P{^intaile}', "");
Expect(0, 6528, '\p{intaile}', "");
Expect(1, 6528, '\p{^intaile}', "");
Expect(1, 6528, '\P{intaile}', "");
Expect(0, 6528, '\P{^intaile}', "");
Expect(1, 6527, '\p{	_in_TAI_le}', "");
Expect(0, 6527, '\p{^	_in_TAI_le}', "");
Expect(0, 6527, '\P{	_in_TAI_le}', "");
Expect(1, 6527, '\P{^	_in_TAI_le}', "");
Expect(0, 6528, '\p{	_in_TAI_le}', "");
Expect(1, 6528, '\p{^	_in_TAI_le}', "");
Expect(1, 6528, '\P{	_in_TAI_le}', "");
Expect(0, 6528, '\P{^	_in_TAI_le}', "");
Error('\p{ /a/in_Tai_Le}');
Error('\P{ /a/in_Tai_Le}');
Expect(1, 6527, '\p{ 	in_Tai_le}', "");
Expect(0, 6527, '\p{^ 	in_Tai_le}', "");
Expect(0, 6527, '\P{ 	in_Tai_le}', "");
Expect(1, 6527, '\P{^ 	in_Tai_le}', "");
Expect(0, 6528, '\p{ 	in_Tai_le}', "");
Expect(1, 6528, '\p{^ 	in_Tai_le}', "");
Expect(1, 6528, '\P{ 	in_Tai_le}', "");
Expect(0, 6528, '\P{^ 	in_Tai_le}', "");
Error('\p{-	In_tai_Tham/a/}');
Error('\P{-	In_tai_Tham/a/}');
Expect(1, 6831, '\p{intaitham}', "");
Expect(0, 6831, '\p{^intaitham}', "");
Expect(0, 6831, '\P{intaitham}', "");
Expect(1, 6831, '\P{^intaitham}', "");
Expect(0, 6832, '\p{intaitham}', "");
Expect(1, 6832, '\p{^intaitham}', "");
Expect(1, 6832, '\P{intaitham}', "");
Expect(0, 6832, '\P{^intaitham}', "");
Expect(1, 6831, '\p{	In_Tai_tham}', "");
Expect(0, 6831, '\p{^	In_Tai_tham}', "");
Expect(0, 6831, '\P{	In_Tai_tham}', "");
Expect(1, 6831, '\P{^	In_Tai_tham}', "");
Expect(0, 6832, '\p{	In_Tai_tham}', "");
Expect(1, 6832, '\p{^	In_Tai_tham}', "");
Expect(1, 6832, '\P{	In_Tai_tham}', "");
Expect(0, 6832, '\P{^	In_Tai_tham}', "");
Error('\p{-:=In_Tai_THAM}');
Error('\P{-:=In_Tai_THAM}');
Expect(1, 6831, '\p{- In_Tai_tham}', "");
Expect(0, 6831, '\p{^- In_Tai_tham}', "");
Expect(0, 6831, '\P{- In_Tai_tham}', "");
Expect(1, 6831, '\P{^- In_Tai_tham}', "");
Expect(0, 6832, '\p{- In_Tai_tham}', "");
Expect(1, 6832, '\p{^- In_Tai_tham}', "");
Expect(1, 6832, '\P{- In_Tai_tham}', "");
Expect(0, 6832, '\P{^- In_Tai_tham}', "");
Error('\p{/a/--In_TAI_viet}');
Error('\P{/a/--In_TAI_viet}');
Expect(1, 43743, '\p{intaiviet}', "");
Expect(0, 43743, '\p{^intaiviet}', "");
Expect(0, 43743, '\P{intaiviet}', "");
Expect(1, 43743, '\P{^intaiviet}', "");
Expect(0, 43744, '\p{intaiviet}', "");
Expect(1, 43744, '\p{^intaiviet}', "");
Expect(1, 43744, '\P{intaiviet}', "");
Expect(0, 43744, '\P{^intaiviet}', "");
Expect(1, 43743, '\p{-	In_Tai_Viet}', "");
Expect(0, 43743, '\p{^-	In_Tai_Viet}', "");
Expect(0, 43743, '\P{-	In_Tai_Viet}', "");
Expect(1, 43743, '\P{^-	In_Tai_Viet}', "");
Expect(0, 43744, '\p{-	In_Tai_Viet}', "");
Expect(1, 43744, '\p{^-	In_Tai_Viet}', "");
Expect(1, 43744, '\P{-	In_Tai_Viet}', "");
Expect(0, 43744, '\P{^-	In_Tai_Viet}', "");
Error('\p{	_In_tai_Viet/a/}');
Error('\P{	_In_tai_Viet/a/}');
Expect(1, 43743, '\p{ In_Tai_VIET}', "");
Expect(0, 43743, '\p{^ In_Tai_VIET}', "");
Expect(0, 43743, '\P{ In_Tai_VIET}', "");
Expect(1, 43743, '\P{^ In_Tai_VIET}', "");
Expect(0, 43744, '\p{ In_Tai_VIET}', "");
Expect(1, 43744, '\p{^ In_Tai_VIET}', "");
Expect(1, 43744, '\P{ In_Tai_VIET}', "");
Expect(0, 43744, '\P{^ In_Tai_VIET}', "");
Error('\p{	_in_Takri/a/}');
Error('\P{	_in_Takri/a/}');
Expect(1, 71375, '\p{intakri}', "");
Expect(0, 71375, '\p{^intakri}', "");
Expect(0, 71375, '\P{intakri}', "");
Expect(1, 71375, '\P{^intakri}', "");
Expect(0, 71376, '\p{intakri}', "");
Expect(1, 71376, '\p{^intakri}', "");
Expect(1, 71376, '\P{intakri}', "");
Expect(0, 71376, '\P{^intakri}', "");
Expect(1, 71375, '\p{	in_Takri}', "");
Expect(0, 71375, '\p{^	in_Takri}', "");
Expect(0, 71375, '\P{	in_Takri}', "");
Expect(1, 71375, '\P{^	in_Takri}', "");
Expect(0, 71376, '\p{	in_Takri}', "");
Expect(1, 71376, '\p{^	in_Takri}', "");
Expect(1, 71376, '\P{	in_Takri}', "");
Expect(0, 71376, '\P{^	in_Takri}', "");
Error('\p{	In_TAKRI:=}');
Error('\P{	In_TAKRI:=}');
Expect(1, 71375, '\p{_ IN_TAKRI}', "");
Expect(0, 71375, '\p{^_ IN_TAKRI}', "");
Expect(0, 71375, '\P{_ IN_TAKRI}', "");
Expect(1, 71375, '\P{^_ IN_TAKRI}', "");
Expect(0, 71376, '\p{_ IN_TAKRI}', "");
Expect(1, 71376, '\p{^_ IN_TAKRI}', "");
Expect(1, 71376, '\P{_ IN_TAKRI}', "");
Expect(0, 71376, '\P{^_ IN_TAKRI}', "");
Error('\p{	_In_Tamil:=}');
Error('\P{	_In_Tamil:=}');
Expect(1, 3071, '\p{intamil}', "");
Expect(0, 3071, '\p{^intamil}', "");
Expect(0, 3071, '\P{intamil}', "");
Expect(1, 3071, '\P{^intamil}', "");
Expect(0, 3072, '\p{intamil}', "");
Expect(1, 3072, '\p{^intamil}', "");
Expect(1, 3072, '\P{intamil}', "");
Expect(0, 3072, '\P{^intamil}', "");
Expect(1, 3071, '\p{ 	in_tamil}', "");
Expect(0, 3071, '\p{^ 	in_tamil}', "");
Expect(0, 3071, '\P{ 	in_tamil}', "");
Expect(1, 3071, '\P{^ 	in_tamil}', "");
Expect(0, 3072, '\p{ 	in_tamil}', "");
Expect(1, 3072, '\p{^ 	in_tamil}', "");
Expect(1, 3072, '\P{ 	in_tamil}', "");
Expect(0, 3072, '\P{^ 	in_tamil}', "");
Error('\p{_:=IN_Tamil}');
Error('\P{_:=IN_Tamil}');
Expect(1, 3071, '\p{	 In_Tamil}', "");
Expect(0, 3071, '\p{^	 In_Tamil}', "");
Expect(0, 3071, '\P{	 In_Tamil}', "");
Expect(1, 3071, '\P{^	 In_Tamil}', "");
Expect(0, 3072, '\p{	 In_Tamil}', "");
Expect(1, 3072, '\p{^	 In_Tamil}', "");
Expect(1, 3072, '\P{	 In_Tamil}', "");
Expect(0, 3072, '\P{^	 In_Tamil}', "");
Error('\p{/a/-IN_Telugu}');
Error('\P{/a/-IN_Telugu}');
Expect(1, 3199, '\p{intelugu}', "");
Expect(0, 3199, '\p{^intelugu}', "");
Expect(0, 3199, '\P{intelugu}', "");
Expect(1, 3199, '\P{^intelugu}', "");
Expect(0, 3200, '\p{intelugu}', "");
Expect(1, 3200, '\p{^intelugu}', "");
Expect(1, 3200, '\P{intelugu}', "");
Expect(0, 3200, '\P{^intelugu}', "");
Expect(1, 3199, '\p{--In_Telugu}', "");
Expect(0, 3199, '\p{^--In_Telugu}', "");
Expect(0, 3199, '\P{--In_Telugu}', "");
Expect(1, 3199, '\P{^--In_Telugu}', "");
Expect(0, 3200, '\p{--In_Telugu}', "");
Expect(1, 3200, '\p{^--In_Telugu}', "");
Expect(1, 3200, '\P{--In_Telugu}', "");
Expect(0, 3200, '\P{^--In_Telugu}', "");
Error('\p{_/a/in_Telugu}');
Error('\P{_/a/in_Telugu}');
Expect(1, 3199, '\p{-	IN_telugu}', "");
Expect(0, 3199, '\p{^-	IN_telugu}', "");
Expect(0, 3199, '\P{-	IN_telugu}', "");
Expect(1, 3199, '\P{^-	IN_telugu}', "");
Expect(0, 3200, '\p{-	IN_telugu}', "");
Expect(1, 3200, '\p{^-	IN_telugu}', "");
Expect(1, 3200, '\P{-	IN_telugu}', "");
Expect(0, 3200, '\P{^-	IN_telugu}', "");
Error('\p{:=-_In_thaana}');
Error('\P{:=-_In_thaana}');
Expect(1, 1983, '\p{inthaana}', "");
Expect(0, 1983, '\p{^inthaana}', "");
Expect(0, 1983, '\P{inthaana}', "");
Expect(1, 1983, '\P{^inthaana}', "");
Expect(0, 1984, '\p{inthaana}', "");
Expect(1, 1984, '\p{^inthaana}', "");
Expect(1, 1984, '\P{inthaana}', "");
Expect(0, 1984, '\P{^inthaana}', "");
Expect(1, 1983, '\p{--In_Thaana}', "");
Expect(0, 1983, '\p{^--In_Thaana}', "");
Expect(0, 1983, '\P{--In_Thaana}', "");
Expect(1, 1983, '\P{^--In_Thaana}', "");
Expect(0, 1984, '\p{--In_Thaana}', "");
Expect(1, 1984, '\p{^--In_Thaana}', "");
Expect(1, 1984, '\P{--In_Thaana}', "");
Expect(0, 1984, '\P{^--In_Thaana}', "");
Error('\p{ :=in_thaana}');
Error('\P{ :=in_thaana}');
Expect(1, 1983, '\p{- IN_thaana}', "");
Expect(0, 1983, '\p{^- IN_thaana}', "");
Expect(0, 1983, '\P{- IN_thaana}', "");
Expect(1, 1983, '\P{^- IN_thaana}', "");
Expect(0, 1984, '\p{- IN_thaana}', "");
Expect(1, 1984, '\p{^- IN_thaana}', "");
Expect(1, 1984, '\P{- IN_thaana}', "");
Expect(0, 1984, '\P{^- IN_thaana}', "");
Error('\p{:=	In_Thai}');
Error('\P{:=	In_Thai}');
Expect(1, 3711, '\p{inthai}', "");
Expect(0, 3711, '\p{^inthai}', "");
Expect(0, 3711, '\P{inthai}', "");
Expect(1, 3711, '\P{^inthai}', "");
Expect(0, 3712, '\p{inthai}', "");
Expect(1, 3712, '\p{^inthai}', "");
Expect(1, 3712, '\P{inthai}', "");
Expect(0, 3712, '\P{^inthai}', "");
Expect(1, 3711, '\p{	 IN_Thai}', "");
Expect(0, 3711, '\p{^	 IN_Thai}', "");
Expect(0, 3711, '\P{	 IN_Thai}', "");
Expect(1, 3711, '\P{^	 IN_Thai}', "");
Expect(0, 3712, '\p{	 IN_Thai}', "");
Expect(1, 3712, '\p{^	 IN_Thai}', "");
Expect(1, 3712, '\P{	 IN_Thai}', "");
Expect(0, 3712, '\P{^	 IN_Thai}', "");
Error('\p{-:=In_Thai}');
Error('\P{-:=In_Thai}');
Expect(1, 3711, '\p{ In_Thai}', "");
Expect(0, 3711, '\p{^ In_Thai}', "");
Expect(0, 3711, '\P{ In_Thai}', "");
Expect(1, 3711, '\P{^ In_Thai}', "");
Expect(0, 3712, '\p{ In_Thai}', "");
Expect(1, 3712, '\p{^ In_Thai}', "");
Expect(1, 3712, '\P{ In_Thai}', "");
Expect(0, 3712, '\P{^ In_Thai}', "");
Error('\p{:=-In_TIBETAN}');
Error('\P{:=-In_TIBETAN}');
Expect(1, 4095, '\p{intibetan}', "");
Expect(0, 4095, '\p{^intibetan}', "");
Expect(0, 4095, '\P{intibetan}', "");
Expect(1, 4095, '\P{^intibetan}', "");
Expect(0, 4096, '\p{intibetan}', "");
Expect(1, 4096, '\p{^intibetan}', "");
Expect(1, 4096, '\P{intibetan}', "");
Expect(0, 4096, '\P{^intibetan}', "");
Expect(1, 4095, '\p{- IN_Tibetan}', "");
Expect(0, 4095, '\p{^- IN_Tibetan}', "");
Expect(0, 4095, '\P{- IN_Tibetan}', "");
Expect(1, 4095, '\P{^- IN_Tibetan}', "");
Expect(0, 4096, '\p{- IN_Tibetan}', "");
Expect(1, 4096, '\p{^- IN_Tibetan}', "");
Expect(1, 4096, '\P{- IN_Tibetan}', "");
Expect(0, 4096, '\P{^- IN_Tibetan}', "");
Error('\p{:=- in_Tibetan}');
Error('\P{:=- in_Tibetan}');
Expect(1, 4095, '\p{-	in_Tibetan}', "");
Expect(0, 4095, '\p{^-	in_Tibetan}', "");
Expect(0, 4095, '\P{-	in_Tibetan}', "");
Expect(1, 4095, '\P{^-	in_Tibetan}', "");
Expect(0, 4096, '\p{-	in_Tibetan}', "");
Expect(1, 4096, '\p{^-	in_Tibetan}', "");
Expect(1, 4096, '\P{-	in_Tibetan}', "");
Expect(0, 4096, '\P{^-	in_Tibetan}', "");
Error('\p{:=- In_TIFINAGH}');
Error('\P{:=- In_TIFINAGH}');
Expect(1, 11647, '\p{intifinagh}', "");
Expect(0, 11647, '\p{^intifinagh}', "");
Expect(0, 11647, '\P{intifinagh}', "");
Expect(1, 11647, '\P{^intifinagh}', "");
Expect(0, 11648, '\p{intifinagh}', "");
Expect(1, 11648, '\p{^intifinagh}', "");
Expect(1, 11648, '\P{intifinagh}', "");
Expect(0, 11648, '\P{^intifinagh}', "");
Expect(1, 11647, '\p{	IN_TIFINAGH}', "");
Expect(0, 11647, '\p{^	IN_TIFINAGH}', "");
Expect(0, 11647, '\P{	IN_TIFINAGH}', "");
Expect(1, 11647, '\P{^	IN_TIFINAGH}', "");
Expect(0, 11648, '\p{	IN_TIFINAGH}', "");
Expect(1, 11648, '\p{^	IN_TIFINAGH}', "");
Expect(1, 11648, '\P{	IN_TIFINAGH}', "");
Expect(0, 11648, '\P{^	IN_TIFINAGH}', "");
Error('\p{ 	In_TIFINAGH:=}');
Error('\P{ 	In_TIFINAGH:=}');
Expect(1, 11647, '\p{_-In_TIFINAGH}', "");
Expect(0, 11647, '\p{^_-In_TIFINAGH}', "");
Expect(0, 11647, '\P{_-In_TIFINAGH}', "");
Expect(1, 11647, '\P{^_-In_TIFINAGH}', "");
Expect(0, 11648, '\p{_-In_TIFINAGH}', "");
Expect(1, 11648, '\p{^_-In_TIFINAGH}', "");
Expect(1, 11648, '\P{_-In_TIFINAGH}', "");
Expect(0, 11648, '\P{^_-In_TIFINAGH}', "");
Error('\p{ _In_UGARITIC:=}');
Error('\P{ _In_UGARITIC:=}');
Expect(1, 66463, '\p{inugaritic}', "");
Expect(0, 66463, '\p{^inugaritic}', "");
Expect(0, 66463, '\P{inugaritic}', "");
Expect(1, 66463, '\P{^inugaritic}', "");
Expect(0, 66464, '\p{inugaritic}', "");
Expect(1, 66464, '\p{^inugaritic}', "");
Expect(1, 66464, '\P{inugaritic}', "");
Expect(0, 66464, '\P{^inugaritic}', "");
Expect(1, 66463, '\p{	-IN_Ugaritic}', "");
Expect(0, 66463, '\p{^	-IN_Ugaritic}', "");
Expect(0, 66463, '\P{	-IN_Ugaritic}', "");
Expect(1, 66463, '\P{^	-IN_Ugaritic}', "");
Expect(0, 66464, '\p{	-IN_Ugaritic}', "");
Expect(1, 66464, '\p{^	-IN_Ugaritic}', "");
Expect(1, 66464, '\P{	-IN_Ugaritic}', "");
Expect(0, 66464, '\P{^	-IN_Ugaritic}', "");
Error('\p{/a/__IN_Ugaritic}');
Error('\P{/a/__IN_Ugaritic}');
Expect(1, 66463, '\p{ 	In_UGARITIC}', "");
Expect(0, 66463, '\p{^ 	In_UGARITIC}', "");
Expect(0, 66463, '\P{ 	In_UGARITIC}', "");
Expect(1, 66463, '\P{^ 	In_UGARITIC}', "");
Expect(0, 66464, '\p{ 	In_UGARITIC}', "");
Expect(1, 66464, '\p{^ 	In_UGARITIC}', "");
Expect(1, 66464, '\P{ 	In_UGARITIC}', "");
Expect(0, 66464, '\P{^ 	In_UGARITIC}', "");
Error('\p{:=_ In_Vai}');
Error('\P{:=_ In_Vai}');
Expect(1, 42559, '\p{invai}', "");
Expect(0, 42559, '\p{^invai}', "");
Expect(0, 42559, '\P{invai}', "");
Expect(1, 42559, '\P{^invai}', "");
Expect(0, 42560, '\p{invai}', "");
Expect(1, 42560, '\p{^invai}', "");
Expect(1, 42560, '\P{invai}', "");
Expect(0, 42560, '\P{^invai}', "");
Expect(1, 42559, '\p{-In_Vai}', "");
Expect(0, 42559, '\p{^-In_Vai}', "");
Expect(0, 42559, '\P{-In_Vai}', "");
Expect(1, 42559, '\P{^-In_Vai}', "");
Expect(0, 42560, '\p{-In_Vai}', "");
Expect(1, 42560, '\p{^-In_Vai}', "");
Expect(1, 42560, '\P{-In_Vai}', "");
Expect(0, 42560, '\P{^-In_Vai}', "");
Error('\p{/a/_	In_vai}');
Error('\P{/a/_	In_vai}');
Expect(1, 42559, '\p{-in_VAI}', "");
Expect(0, 42559, '\p{^-in_VAI}', "");
Expect(0, 42559, '\P{-in_VAI}', "");
Expect(1, 42559, '\P{^-in_VAI}', "");
Expect(0, 42560, '\p{-in_VAI}', "");
Expect(1, 42560, '\p{^-in_VAI}', "");
Expect(1, 42560, '\P{-in_VAI}', "");
Expect(0, 42560, '\P{^-in_VAI}', "");
Error('\p{_/a/INHERITED}');
Error('\P{_/a/INHERITED}');
Expect(1, 917999, '\p{inherited}', "");
Expect(0, 917999, '\p{^inherited}', "");
Expect(0, 917999, '\P{inherited}', "");
Expect(1, 917999, '\P{^inherited}', "");
Expect(0, 918000, '\p{inherited}', "");
Expect(1, 918000, '\p{^inherited}', "");
Expect(1, 918000, '\P{inherited}', "");
Expect(0, 918000, '\P{^inherited}', "");
Expect(1, 917999, '\p{	_Inherited}', "");
Expect(0, 917999, '\p{^	_Inherited}', "");
Expect(0, 917999, '\P{	_Inherited}', "");
Expect(1, 917999, '\P{^	_Inherited}', "");
Expect(0, 918000, '\p{	_Inherited}', "");
Expect(1, 918000, '\p{^	_Inherited}', "");
Expect(1, 918000, '\P{	_Inherited}', "");
Expect(0, 918000, '\P{^	_Inherited}', "");
Error('\p{:=--IS_INHERITED}');
Error('\P{:=--IS_INHERITED}');
Expect(1, 917999, '\p{isinherited}', "");
Expect(0, 917999, '\p{^isinherited}', "");
Expect(0, 917999, '\P{isinherited}', "");
Expect(1, 917999, '\P{^isinherited}', "");
Expect(0, 918000, '\p{isinherited}', "");
Expect(1, 918000, '\p{^isinherited}', "");
Expect(1, 918000, '\P{isinherited}', "");
Expect(0, 918000, '\P{^isinherited}', "");
Expect(1, 917999, '\p{	_Is_INHERITED}', "");
Expect(0, 917999, '\p{^	_Is_INHERITED}', "");
Expect(0, 917999, '\P{	_Is_INHERITED}', "");
Expect(1, 917999, '\P{^	_Is_INHERITED}', "");
Expect(0, 918000, '\p{	_Is_INHERITED}', "");
Expect(1, 918000, '\p{^	_Is_INHERITED}', "");
Expect(1, 918000, '\P{	_Is_INHERITED}', "");
Expect(0, 918000, '\P{^	_Is_INHERITED}', "");
Error('\p{  Zinh/a/}');
Error('\P{  Zinh/a/}');
Expect(1, 917999, '\p{zinh}', "");
Expect(0, 917999, '\p{^zinh}', "");
Expect(0, 917999, '\P{zinh}', "");
Expect(1, 917999, '\P{^zinh}', "");
Expect(0, 918000, '\p{zinh}', "");
Expect(1, 918000, '\p{^zinh}', "");
Expect(1, 918000, '\P{zinh}', "");
Expect(0, 918000, '\P{^zinh}', "");
Expect(1, 917999, '\p{-Zinh}', "");
Expect(0, 917999, '\p{^-Zinh}', "");
Expect(0, 917999, '\P{-Zinh}', "");
Expect(1, 917999, '\P{^-Zinh}', "");
Expect(0, 918000, '\p{-Zinh}', "");
Expect(1, 918000, '\p{^-Zinh}', "");
Expect(1, 918000, '\P{-Zinh}', "");
Expect(0, 918000, '\P{^-Zinh}', "");
Error('\p{:= Is_Zinh}');
Error('\P{:= Is_Zinh}');
Expect(1, 917999, '\p{iszinh}', "");
Expect(0, 917999, '\p{^iszinh}', "");
Expect(0, 917999, '\P{iszinh}', "");
Expect(1, 917999, '\P{^iszinh}', "");
Expect(0, 918000, '\p{iszinh}', "");
Expect(1, 918000, '\p{^iszinh}', "");
Expect(1, 918000, '\P{iszinh}', "");
Expect(0, 918000, '\P{^iszinh}', "");
Expect(1, 917999, '\p{Is_Zinh}', "");
Expect(0, 917999, '\p{^Is_Zinh}', "");
Expect(0, 917999, '\P{Is_Zinh}', "");
Expect(1, 917999, '\P{^Is_Zinh}', "");
Expect(0, 918000, '\p{Is_Zinh}', "");
Expect(1, 918000, '\p{^Is_Zinh}', "");
Expect(1, 918000, '\P{Is_Zinh}', "");
Expect(0, 918000, '\P{^Is_Zinh}', "");
Error('\p{/a/-	qaai}');
Error('\P{/a/-	qaai}');
Expect(1, 917999, '\p{qaai}', "");
Expect(0, 917999, '\p{^qaai}', "");
Expect(0, 917999, '\P{qaai}', "");
Expect(1, 917999, '\P{^qaai}', "");
Expect(0, 918000, '\p{qaai}', "");
Expect(1, 918000, '\p{^qaai}', "");
Expect(1, 918000, '\P{qaai}', "");
Expect(0, 918000, '\P{^qaai}', "");
Error('\p{/a/__is_Qaai}');
Error('\P{/a/__is_Qaai}');
Expect(1, 917999, '\p{isqaai}', "");
Expect(0, 917999, '\p{^isqaai}', "");
Expect(0, 917999, '\P{isqaai}', "");
Expect(1, 917999, '\P{^isqaai}', "");
Expect(0, 918000, '\p{isqaai}', "");
Expect(1, 918000, '\p{^isqaai}', "");
Expect(1, 918000, '\P{isqaai}', "");
Expect(0, 918000, '\P{^isqaai}', "");
Expect(1, 917999, '\p{_ IS_Qaai}', "");
Expect(0, 917999, '\p{^_ IS_Qaai}', "");
Expect(0, 917999, '\P{_ IS_Qaai}', "");
Expect(1, 917999, '\P{^_ IS_Qaai}', "");
Expect(0, 918000, '\p{_ IS_Qaai}', "");
Expect(1, 918000, '\p{^_ IS_Qaai}', "");
Expect(1, 918000, '\P{_ IS_Qaai}', "");
Expect(0, 918000, '\P{^_ IS_Qaai}', "");
Error('\p{-:=INITIAL_Punctuation}');
Error('\P{-:=INITIAL_Punctuation}');
Expect(1, 11808, '\p{initialpunctuation}', "");
Expect(0, 11808, '\p{^initialpunctuation}', "");
Expect(0, 11808, '\P{initialpunctuation}', "");
Expect(1, 11808, '\P{^initialpunctuation}', "");
Expect(0, 11809, '\p{initialpunctuation}', "");
Expect(1, 11809, '\p{^initialpunctuation}', "");
Expect(1, 11809, '\P{initialpunctuation}', "");
Expect(0, 11809, '\P{^initialpunctuation}', "");
Expect(1, 11808, '\p{	 Initial_PUNCTUATION}', "");
Expect(0, 11808, '\p{^	 Initial_PUNCTUATION}', "");
Expect(0, 11808, '\P{	 Initial_PUNCTUATION}', "");
Expect(1, 11808, '\P{^	 Initial_PUNCTUATION}', "");
Expect(0, 11809, '\p{	 Initial_PUNCTUATION}', "");
Expect(1, 11809, '\p{^	 Initial_PUNCTUATION}', "");
Expect(1, 11809, '\P{	 Initial_PUNCTUATION}', "");
Expect(0, 11809, '\P{^	 Initial_PUNCTUATION}', "");
Error('\p{:=Is_initial_punctuation}');
Error('\P{:=Is_initial_punctuation}');
Expect(1, 11808, '\p{isinitialpunctuation}', "");
Expect(0, 11808, '\p{^isinitialpunctuation}', "");
Expect(0, 11808, '\P{isinitialpunctuation}', "");
Expect(1, 11808, '\P{^isinitialpunctuation}', "");
Expect(0, 11809, '\p{isinitialpunctuation}', "");
Expect(1, 11809, '\p{^isinitialpunctuation}', "");
Expect(1, 11809, '\P{isinitialpunctuation}', "");
Expect(0, 11809, '\P{^isinitialpunctuation}', "");
Expect(1, 11808, '\p{	-Is_Initial_Punctuation}', "");
Expect(0, 11808, '\p{^	-Is_Initial_Punctuation}', "");
Expect(0, 11808, '\P{	-Is_Initial_Punctuation}', "");
Expect(1, 11808, '\P{^	-Is_Initial_Punctuation}', "");
Expect(0, 11809, '\p{	-Is_Initial_Punctuation}', "");
Expect(1, 11809, '\p{^	-Is_Initial_Punctuation}', "");
Expect(1, 11809, '\P{	-Is_Initial_Punctuation}', "");
Expect(0, 11809, '\P{^	-Is_Initial_Punctuation}', "");
Error('\p{_:=pi}');
Error('\P{_:=pi}');
Expect(1, 11808, '\p{pi}', "");
Expect(0, 11808, '\p{^pi}', "");
Expect(0, 11808, '\P{pi}', "");
Expect(1, 11808, '\P{^pi}', "");
Expect(0, 11809, '\p{pi}', "");
Expect(1, 11809, '\p{^pi}', "");
Expect(1, 11809, '\P{pi}', "");
Expect(0, 11809, '\P{^pi}', "");
Expect(1, 11808, '\p{Pi}', "");
Expect(0, 11808, '\p{^Pi}', "");
Expect(0, 11808, '\P{Pi}', "");
Expect(1, 11808, '\P{^Pi}', "");
Expect(0, 11809, '\p{Pi}', "");
Expect(1, 11809, '\p{^Pi}', "");
Expect(1, 11809, '\P{Pi}', "");
Expect(0, 11809, '\P{^Pi}', "");
Error('\p{	/a/is_PI}');
Error('\P{	/a/is_PI}');
Expect(1, 11808, '\p{ispi}', "");
Expect(0, 11808, '\p{^ispi}', "");
Expect(0, 11808, '\P{ispi}', "");
Expect(1, 11808, '\P{^ispi}', "");
Expect(0, 11809, '\p{ispi}', "");
Expect(1, 11809, '\p{^ispi}', "");
Expect(1, 11809, '\P{ispi}', "");
Expect(0, 11809, '\P{^ispi}', "");
Expect(1, 11808, '\p{_Is_PI}', "");
Expect(0, 11808, '\p{^_Is_PI}', "");
Expect(0, 11808, '\P{_Is_PI}', "");
Expect(1, 11808, '\P{^_Is_PI}', "");
Expect(0, 11809, '\p{_Is_PI}', "");
Expect(1, 11809, '\p{^_Is_PI}', "");
Expect(1, 11809, '\P{_Is_PI}', "");
Expect(0, 11809, '\P{^_Is_PI}', "");
Error('\p{	Inscriptional_Pahlavi/a/}');
Error('\P{	Inscriptional_Pahlavi/a/}');
Expect(1, 68479, '\p{inscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^inscriptionalpahlavi}', "");
Expect(0, 68479, '\P{inscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^inscriptionalpahlavi}', "");
Expect(0, 68480, '\p{inscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^inscriptionalpahlavi}', "");
Expect(1, 68480, '\P{inscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^inscriptionalpahlavi}', "");
Expect(1, 68479, '\p{	_Inscriptional_Pahlavi}', "");
Expect(0, 68479, '\p{^	_Inscriptional_Pahlavi}', "");
Expect(0, 68479, '\P{	_Inscriptional_Pahlavi}', "");
Expect(1, 68479, '\P{^	_Inscriptional_Pahlavi}', "");
Expect(0, 68480, '\p{	_Inscriptional_Pahlavi}', "");
Expect(1, 68480, '\p{^	_Inscriptional_Pahlavi}', "");
Expect(1, 68480, '\P{	_Inscriptional_Pahlavi}', "");
Expect(0, 68480, '\P{^	_Inscriptional_Pahlavi}', "");
Error('\p{  is_Inscriptional_Pahlavi:=}');
Error('\P{  is_Inscriptional_Pahlavi:=}');
Expect(1, 68479, '\p{isinscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^isinscriptionalpahlavi}', "");
Expect(0, 68479, '\P{isinscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^isinscriptionalpahlavi}', "");
Expect(0, 68480, '\p{isinscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^isinscriptionalpahlavi}', "");
Expect(1, 68480, '\P{isinscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^isinscriptionalpahlavi}', "");
Expect(1, 68479, '\p{__is_INSCRIPTIONAL_pahlavi}', "");
Expect(0, 68479, '\p{^__is_INSCRIPTIONAL_pahlavi}', "");
Expect(0, 68479, '\P{__is_INSCRIPTIONAL_pahlavi}', "");
Expect(1, 68479, '\P{^__is_INSCRIPTIONAL_pahlavi}', "");
Expect(0, 68480, '\p{__is_INSCRIPTIONAL_pahlavi}', "");
Expect(1, 68480, '\p{^__is_INSCRIPTIONAL_pahlavi}', "");
Expect(1, 68480, '\P{__is_INSCRIPTIONAL_pahlavi}', "");
Expect(0, 68480, '\P{^__is_INSCRIPTIONAL_pahlavi}', "");
Error('\p{_	PHLI/a/}');
Error('\P{_	PHLI/a/}');
Expect(1, 68479, '\p{phli}', "");
Expect(0, 68479, '\p{^phli}', "");
Expect(0, 68479, '\P{phli}', "");
Expect(1, 68479, '\P{^phli}', "");
Expect(0, 68480, '\p{phli}', "");
Expect(1, 68480, '\p{^phli}', "");
Expect(1, 68480, '\P{phli}', "");
Expect(0, 68480, '\P{^phli}', "");
Expect(1, 68479, '\p{_Phli}', "");
Expect(0, 68479, '\p{^_Phli}', "");
Expect(0, 68479, '\P{_Phli}', "");
Expect(1, 68479, '\P{^_Phli}', "");
Expect(0, 68480, '\p{_Phli}', "");
Expect(1, 68480, '\p{^_Phli}', "");
Expect(1, 68480, '\P{_Phli}', "");
Expect(0, 68480, '\P{^_Phli}', "");
Error('\p{:=Is_PHLI}');
Error('\P{:=Is_PHLI}');
Expect(1, 68479, '\p{isphli}', "");
Expect(0, 68479, '\p{^isphli}', "");
Expect(0, 68479, '\P{isphli}', "");
Expect(1, 68479, '\P{^isphli}', "");
Expect(0, 68480, '\p{isphli}', "");
Expect(1, 68480, '\p{^isphli}', "");
Expect(1, 68480, '\P{isphli}', "");
Expect(0, 68480, '\P{^isphli}', "");
Expect(1, 68479, '\p{_IS_Phli}', "");
Expect(0, 68479, '\p{^_IS_Phli}', "");
Expect(0, 68479, '\P{_IS_Phli}', "");
Expect(1, 68479, '\P{^_IS_Phli}', "");
Expect(0, 68480, '\p{_IS_Phli}', "");
Expect(1, 68480, '\p{^_IS_Phli}', "");
Expect(1, 68480, '\P{_IS_Phli}', "");
Expect(0, 68480, '\P{^_IS_Phli}', "");
Error('\p{  INSCRIPTIONAL_Parthian/a/}');
Error('\P{  INSCRIPTIONAL_Parthian/a/}');
Expect(1, 68447, '\p{inscriptionalparthian}', "");
Expect(0, 68447, '\p{^inscriptionalparthian}', "");
Expect(0, 68447, '\P{inscriptionalparthian}', "");
Expect(1, 68447, '\P{^inscriptionalparthian}', "");
Expect(0, 68448, '\p{inscriptionalparthian}', "");
Expect(1, 68448, '\p{^inscriptionalparthian}', "");
Expect(1, 68448, '\P{inscriptionalparthian}', "");
Expect(0, 68448, '\P{^inscriptionalparthian}', "");
Expect(1, 68447, '\p{ -Inscriptional_Parthian}', "");
Expect(0, 68447, '\p{^ -Inscriptional_Parthian}', "");
Expect(0, 68447, '\P{ -Inscriptional_Parthian}', "");
Expect(1, 68447, '\P{^ -Inscriptional_Parthian}', "");
Expect(0, 68448, '\p{ -Inscriptional_Parthian}', "");
Expect(1, 68448, '\p{^ -Inscriptional_Parthian}', "");
Expect(1, 68448, '\P{ -Inscriptional_Parthian}', "");
Expect(0, 68448, '\P{^ -Inscriptional_Parthian}', "");
Error('\p{_is_Inscriptional_PARTHIAN/a/}');
Error('\P{_is_Inscriptional_PARTHIAN/a/}');
Expect(1, 68447, '\p{isinscriptionalparthian}', "");
Expect(0, 68447, '\p{^isinscriptionalparthian}', "");
Expect(0, 68447, '\P{isinscriptionalparthian}', "");
Expect(1, 68447, '\P{^isinscriptionalparthian}', "");
Expect(0, 68448, '\p{isinscriptionalparthian}', "");
Expect(1, 68448, '\p{^isinscriptionalparthian}', "");
Expect(1, 68448, '\P{isinscriptionalparthian}', "");
Expect(0, 68448, '\P{^isinscriptionalparthian}', "");
Expect(1, 68447, '\p{-Is_INSCRIPTIONAL_parthian}', "");
Expect(0, 68447, '\p{^-Is_INSCRIPTIONAL_parthian}', "");
Expect(0, 68447, '\P{-Is_INSCRIPTIONAL_parthian}', "");
Expect(1, 68447, '\P{^-Is_INSCRIPTIONAL_parthian}', "");
Expect(0, 68448, '\p{-Is_INSCRIPTIONAL_parthian}', "");
Expect(1, 68448, '\p{^-Is_INSCRIPTIONAL_parthian}', "");
Expect(1, 68448, '\P{-Is_INSCRIPTIONAL_parthian}', "");
Expect(0, 68448, '\P{^-Is_INSCRIPTIONAL_parthian}', "");
Error('\p{/a/-	Prti}');
Error('\P{/a/-	Prti}');
Expect(1, 68447, '\p{prti}', "");
Expect(0, 68447, '\p{^prti}', "");
Expect(0, 68447, '\P{prti}', "");
Expect(1, 68447, '\P{^prti}', "");
Expect(0, 68448, '\p{prti}', "");
Expect(1, 68448, '\p{^prti}', "");
Expect(1, 68448, '\P{prti}', "");
Expect(0, 68448, '\P{^prti}', "");
Expect(1, 68447, '\p{ Prti}', "");
Expect(0, 68447, '\p{^ Prti}', "");
Expect(0, 68447, '\P{ Prti}', "");
Expect(1, 68447, '\P{^ Prti}', "");
Expect(0, 68448, '\p{ Prti}', "");
Expect(1, 68448, '\p{^ Prti}', "");
Expect(1, 68448, '\P{ Prti}', "");
Expect(0, 68448, '\P{^ Prti}', "");
Error('\p{	-IS_prti/a/}');
Error('\P{	-IS_prti/a/}');
Expect(1, 68447, '\p{isprti}', "");
Expect(0, 68447, '\p{^isprti}', "");
Expect(0, 68447, '\P{isprti}', "");
Expect(1, 68447, '\P{^isprti}', "");
Expect(0, 68448, '\p{isprti}', "");
Expect(1, 68448, '\p{^isprti}', "");
Expect(1, 68448, '\P{isprti}', "");
Expect(0, 68448, '\P{^isprti}', "");
Expect(1, 68447, '\p{_	IS_PRTI}', "");
Expect(0, 68447, '\p{^_	IS_PRTI}', "");
Expect(0, 68447, '\P{_	IS_PRTI}', "");
Expect(1, 68447, '\P{^_	IS_PRTI}', "");
Expect(0, 68448, '\p{_	IS_PRTI}', "");
Expect(1, 68448, '\p{^_	IS_PRTI}', "");
Expect(1, 68448, '\P{_	IS_PRTI}', "");
Expect(0, 68448, '\P{^_	IS_PRTI}', "");
Error('\p{	:=Javanese}');
Error('\P{	:=Javanese}');
Expect(1, 43487, '\p{javanese}', "");
Expect(0, 43487, '\p{^javanese}', "");
Expect(0, 43487, '\P{javanese}', "");
Expect(1, 43487, '\P{^javanese}', "");
Expect(0, 43488, '\p{javanese}', "");
Expect(1, 43488, '\p{^javanese}', "");
Expect(1, 43488, '\P{javanese}', "");
Expect(0, 43488, '\P{^javanese}', "");
Expect(1, 43487, '\p{		Javanese}', "");
Expect(0, 43487, '\p{^		Javanese}', "");
Expect(0, 43487, '\P{		Javanese}', "");
Expect(1, 43487, '\P{^		Javanese}', "");
Expect(0, 43488, '\p{		Javanese}', "");
Expect(1, 43488, '\p{^		Javanese}', "");
Expect(1, 43488, '\P{		Javanese}', "");
Expect(0, 43488, '\P{^		Javanese}', "");
Error('\p{:=_-is_Javanese}');
Error('\P{:=_-is_Javanese}');
Expect(1, 43487, '\p{isjavanese}', "");
Expect(0, 43487, '\p{^isjavanese}', "");
Expect(0, 43487, '\P{isjavanese}', "");
Expect(1, 43487, '\P{^isjavanese}', "");
Expect(0, 43488, '\p{isjavanese}', "");
Expect(1, 43488, '\p{^isjavanese}', "");
Expect(1, 43488, '\P{isjavanese}', "");
Expect(0, 43488, '\P{^isjavanese}', "");
Expect(1, 43487, '\p{	 Is_Javanese}', "");
Expect(0, 43487, '\p{^	 Is_Javanese}', "");
Expect(0, 43487, '\P{	 Is_Javanese}', "");
Expect(1, 43487, '\P{^	 Is_Javanese}', "");
Expect(0, 43488, '\p{	 Is_Javanese}', "");
Expect(1, 43488, '\p{^	 Is_Javanese}', "");
Expect(1, 43488, '\P{	 Is_Javanese}', "");
Expect(0, 43488, '\P{^	 Is_Javanese}', "");
Error('\p{_	JAVA/a/}');
Error('\P{_	JAVA/a/}');
Expect(1, 43487, '\p{java}', "");
Expect(0, 43487, '\p{^java}', "");
Expect(0, 43487, '\P{java}', "");
Expect(1, 43487, '\P{^java}', "");
Expect(0, 43488, '\p{java}', "");
Expect(1, 43488, '\p{^java}', "");
Expect(1, 43488, '\P{java}', "");
Expect(0, 43488, '\P{^java}', "");
Expect(1, 43487, '\p{__JAVA}', "");
Expect(0, 43487, '\p{^__JAVA}', "");
Expect(0, 43487, '\P{__JAVA}', "");
Expect(1, 43487, '\P{^__JAVA}', "");
Expect(0, 43488, '\p{__JAVA}', "");
Expect(1, 43488, '\p{^__JAVA}', "");
Expect(1, 43488, '\P{__JAVA}', "");
Expect(0, 43488, '\P{^__JAVA}', "");
Error('\p{:=IS_java}');
Error('\P{:=IS_java}');
Expect(1, 43487, '\p{isjava}', "");
Expect(0, 43487, '\p{^isjava}', "");
Expect(0, 43487, '\P{isjava}', "");
Expect(1, 43487, '\P{^isjava}', "");
Expect(0, 43488, '\p{isjava}', "");
Expect(1, 43488, '\p{^isjava}', "");
Expect(1, 43488, '\P{isjava}', "");
Expect(0, 43488, '\P{^isjava}', "");
Expect(1, 43487, '\p{	 Is_Java}', "");
Expect(0, 43487, '\p{^	 Is_Java}', "");
Expect(0, 43487, '\P{	 Is_Java}', "");
Expect(1, 43487, '\P{^	 Is_Java}', "");
Expect(0, 43488, '\p{	 Is_Java}', "");
Expect(1, 43488, '\p{^	 Is_Java}', "");
Expect(1, 43488, '\P{	 Is_Java}', "");
Expect(0, 43488, '\P{^	 Is_Java}', "");
Error('\p{-/a/Join_Control}');
Error('\P{-/a/Join_Control}');
Expect(1, 8205, '\p{joincontrol}', "");
Expect(0, 8205, '\p{^joincontrol}', "");
Expect(0, 8205, '\P{joincontrol}', "");
Expect(1, 8205, '\P{^joincontrol}', "");
Expect(0, 8206, '\p{joincontrol}', "");
Expect(1, 8206, '\p{^joincontrol}', "");
Expect(1, 8206, '\P{joincontrol}', "");
Expect(0, 8206, '\P{^joincontrol}', "");
Expect(1, 8205, '\p{ join_Control}', "");
Expect(0, 8205, '\p{^ join_Control}', "");
Expect(0, 8205, '\P{ join_Control}', "");
Expect(1, 8205, '\P{^ join_Control}', "");
Expect(0, 8206, '\p{ join_Control}', "");
Expect(1, 8206, '\p{^ join_Control}', "");
Expect(1, 8206, '\P{ join_Control}', "");
Expect(0, 8206, '\P{^ join_Control}', "");
Error('\p{ -is_Join_CONTROL/a/}');
Error('\P{ -is_Join_CONTROL/a/}');
Expect(1, 8205, '\p{isjoincontrol}', "");
Expect(0, 8205, '\p{^isjoincontrol}', "");
Expect(0, 8205, '\P{isjoincontrol}', "");
Expect(1, 8205, '\P{^isjoincontrol}', "");
Expect(0, 8206, '\p{isjoincontrol}', "");
Expect(1, 8206, '\p{^isjoincontrol}', "");
Expect(1, 8206, '\P{isjoincontrol}', "");
Expect(0, 8206, '\P{^isjoincontrol}', "");
Expect(1, 8205, '\p{	is_JOIN_control}', "");
Expect(0, 8205, '\p{^	is_JOIN_control}', "");
Expect(0, 8205, '\P{	is_JOIN_control}', "");
Expect(1, 8205, '\P{^	is_JOIN_control}', "");
Expect(0, 8206, '\p{	is_JOIN_control}', "");
Expect(1, 8206, '\p{^	is_JOIN_control}', "");
Expect(1, 8206, '\P{	is_JOIN_control}', "");
Expect(0, 8206, '\P{^	is_JOIN_control}', "");
Error('\p{_Join_C/a/}');
Error('\P{_Join_C/a/}');
Expect(1, 8205, '\p{joinc}', "");
Expect(0, 8205, '\p{^joinc}', "");
Expect(0, 8205, '\P{joinc}', "");
Expect(1, 8205, '\P{^joinc}', "");
Expect(0, 8206, '\p{joinc}', "");
Expect(1, 8206, '\p{^joinc}', "");
Expect(1, 8206, '\P{joinc}', "");
Expect(0, 8206, '\P{^joinc}', "");
Expect(1, 8205, '\p{-Join_C}', "");
Expect(0, 8205, '\p{^-Join_C}', "");
Expect(0, 8205, '\P{-Join_C}', "");
Expect(1, 8205, '\P{^-Join_C}', "");
Expect(0, 8206, '\p{-Join_C}', "");
Expect(1, 8206, '\p{^-Join_C}', "");
Expect(1, 8206, '\P{-Join_C}', "");
Expect(0, 8206, '\P{^-Join_C}', "");
Error('\p{_:=Is_Join_c}');
Error('\P{_:=Is_Join_c}');
Expect(1, 8205, '\p{isjoinc}', "");
Expect(0, 8205, '\p{^isjoinc}', "");
Expect(0, 8205, '\P{isjoinc}', "");
Expect(1, 8205, '\P{^isjoinc}', "");
Expect(0, 8206, '\p{isjoinc}', "");
Expect(1, 8206, '\p{^isjoinc}', "");
Expect(1, 8206, '\P{isjoinc}', "");
Expect(0, 8206, '\P{^isjoinc}', "");
Expect(1, 8205, '\p{_	Is_join_C}', "");
Expect(0, 8205, '\p{^_	Is_join_C}', "");
Expect(0, 8205, '\P{_	Is_join_C}', "");
Expect(1, 8205, '\P{^_	Is_join_C}', "");
Expect(0, 8206, '\p{_	Is_join_C}', "");
Expect(1, 8206, '\p{^_	Is_join_C}', "");
Expect(1, 8206, '\P{_	Is_join_C}', "");
Expect(0, 8206, '\P{^_	Is_join_C}', "");
Error('\p{-	kaithi/a/}');
Error('\P{-	kaithi/a/}');
Expect(1, 69825, '\p{kaithi}', "");
Expect(0, 69825, '\p{^kaithi}', "");
Expect(0, 69825, '\P{kaithi}', "");
Expect(1, 69825, '\P{^kaithi}', "");
Expect(0, 69826, '\p{kaithi}', "");
Expect(1, 69826, '\p{^kaithi}', "");
Expect(1, 69826, '\P{kaithi}', "");
Expect(0, 69826, '\P{^kaithi}', "");
Expect(1, 69825, '\p{-_kaithi}', "");
Expect(0, 69825, '\p{^-_kaithi}', "");
Expect(0, 69825, '\P{-_kaithi}', "");
Expect(1, 69825, '\P{^-_kaithi}', "");
Expect(0, 69826, '\p{-_kaithi}', "");
Expect(1, 69826, '\p{^-_kaithi}', "");
Expect(1, 69826, '\P{-_kaithi}', "");
Expect(0, 69826, '\P{^-_kaithi}', "");
Error('\p{	/a/Is_KAITHI}');
Error('\P{	/a/Is_KAITHI}');
Expect(1, 69825, '\p{iskaithi}', "");
Expect(0, 69825, '\p{^iskaithi}', "");
Expect(0, 69825, '\P{iskaithi}', "");
Expect(1, 69825, '\P{^iskaithi}', "");
Expect(0, 69826, '\p{iskaithi}', "");
Expect(1, 69826, '\p{^iskaithi}', "");
Expect(1, 69826, '\P{iskaithi}', "");
Expect(0, 69826, '\P{^iskaithi}', "");
Expect(1, 69825, '\p{	is_KAITHI}', "");
Expect(0, 69825, '\p{^	is_KAITHI}', "");
Expect(0, 69825, '\P{	is_KAITHI}', "");
Expect(1, 69825, '\P{^	is_KAITHI}', "");
Expect(0, 69826, '\p{	is_KAITHI}', "");
Expect(1, 69826, '\p{^	is_KAITHI}', "");
Expect(1, 69826, '\P{	is_KAITHI}', "");
Expect(0, 69826, '\P{^	is_KAITHI}', "");
Error('\p{	:=KTHI}');
Error('\P{	:=KTHI}');
Expect(1, 69825, '\p{kthi}', "");
Expect(0, 69825, '\p{^kthi}', "");
Expect(0, 69825, '\P{kthi}', "");
Expect(1, 69825, '\P{^kthi}', "");
Expect(0, 69826, '\p{kthi}', "");
Expect(1, 69826, '\p{^kthi}', "");
Expect(1, 69826, '\P{kthi}', "");
Expect(0, 69826, '\P{^kthi}', "");
Expect(1, 69825, '\p{_ KTHI}', "");
Expect(0, 69825, '\p{^_ KTHI}', "");
Expect(0, 69825, '\P{_ KTHI}', "");
Expect(1, 69825, '\P{^_ KTHI}', "");
Expect(0, 69826, '\p{_ KTHI}', "");
Expect(1, 69826, '\p{^_ KTHI}', "");
Expect(1, 69826, '\P{_ KTHI}', "");
Expect(0, 69826, '\P{^_ KTHI}', "");
Error('\p{-/a/is_Kthi}');
Error('\P{-/a/is_Kthi}');
Expect(1, 69825, '\p{iskthi}', "");
Expect(0, 69825, '\p{^iskthi}', "");
Expect(0, 69825, '\P{iskthi}', "");
Expect(1, 69825, '\P{^iskthi}', "");
Expect(0, 69826, '\p{iskthi}', "");
Expect(1, 69826, '\p{^iskthi}', "");
Expect(1, 69826, '\P{iskthi}', "");
Expect(0, 69826, '\P{^iskthi}', "");
Expect(1, 69825, '\p{  Is_kthi}', "");
Expect(0, 69825, '\p{^  Is_kthi}', "");
Expect(0, 69825, '\P{  Is_kthi}', "");
Expect(1, 69825, '\P{^  Is_kthi}', "");
Expect(0, 69826, '\p{  Is_kthi}', "");
Expect(1, 69826, '\p{^  Is_kthi}', "");
Expect(1, 69826, '\P{  Is_kthi}', "");
Expect(0, 69826, '\P{^  Is_kthi}', "");
Error('\p{_/a/KANA_Supplement}');
Error('\P{_/a/KANA_Supplement}');
Expect(1, 110847, '\p{kanasupplement}', "");
Expect(0, 110847, '\p{^kanasupplement}', "");
Expect(0, 110847, '\P{kanasupplement}', "");
Expect(1, 110847, '\P{^kanasupplement}', "");
Expect(0, 110848, '\p{kanasupplement}', "");
Expect(1, 110848, '\p{^kanasupplement}', "");
Expect(1, 110848, '\P{kanasupplement}', "");
Expect(0, 110848, '\P{^kanasupplement}', "");
Expect(1, 110847, '\p{	kana_supplement}', "");
Expect(0, 110847, '\p{^	kana_supplement}', "");
Expect(0, 110847, '\P{	kana_supplement}', "");
Expect(1, 110847, '\P{^	kana_supplement}', "");
Expect(0, 110848, '\p{	kana_supplement}', "");
Expect(1, 110848, '\p{^	kana_supplement}', "");
Expect(1, 110848, '\P{	kana_supplement}', "");
Expect(0, 110848, '\P{^	kana_supplement}', "");
Error('\p{:=-is_KANA_SUPPLEMENT}');
Error('\P{:=-is_KANA_SUPPLEMENT}');
Expect(1, 110847, '\p{iskanasupplement}', "");
Expect(0, 110847, '\p{^iskanasupplement}', "");
Expect(0, 110847, '\P{iskanasupplement}', "");
Expect(1, 110847, '\P{^iskanasupplement}', "");
Expect(0, 110848, '\p{iskanasupplement}', "");
Expect(1, 110848, '\p{^iskanasupplement}', "");
Expect(1, 110848, '\P{iskanasupplement}', "");
Expect(0, 110848, '\P{^iskanasupplement}', "");
Expect(1, 110847, '\p{	 Is_Kana_Supplement}', "");
Expect(0, 110847, '\p{^	 Is_Kana_Supplement}', "");
Expect(0, 110847, '\P{	 Is_Kana_Supplement}', "");
Expect(1, 110847, '\P{^	 Is_Kana_Supplement}', "");
Expect(0, 110848, '\p{	 Is_Kana_Supplement}', "");
Expect(1, 110848, '\p{^	 Is_Kana_Supplement}', "");
Expect(1, 110848, '\P{	 Is_Kana_Supplement}', "");
Expect(0, 110848, '\P{^	 Is_Kana_Supplement}', "");
Error('\p{:=_in_kana_Supplement}');
Error('\P{:=_in_kana_Supplement}');
Expect(1, 110847, '\p{inkanasupplement}', "");
Expect(0, 110847, '\p{^inkanasupplement}', "");
Expect(0, 110847, '\P{inkanasupplement}', "");
Expect(1, 110847, '\P{^inkanasupplement}', "");
Expect(0, 110848, '\p{inkanasupplement}', "");
Expect(1, 110848, '\p{^inkanasupplement}', "");
Expect(1, 110848, '\P{inkanasupplement}', "");
Expect(0, 110848, '\P{^inkanasupplement}', "");
Expect(1, 110847, '\p{	 In_kana_Supplement}', "");
Expect(0, 110847, '\p{^	 In_kana_Supplement}', "");
Expect(0, 110847, '\P{	 In_kana_Supplement}', "");
Expect(1, 110847, '\P{^	 In_kana_Supplement}', "");
Expect(0, 110848, '\p{	 In_kana_Supplement}', "");
Expect(1, 110848, '\p{^	 In_kana_Supplement}', "");
Expect(1, 110848, '\P{	 In_kana_Supplement}', "");
Expect(0, 110848, '\P{^	 In_kana_Supplement}', "");
Error('\p{:=- Kana_Sup}');
Error('\P{:=- Kana_Sup}');
Expect(1, 110847, '\p{kanasup}', "");
Expect(0, 110847, '\p{^kanasup}', "");
Expect(0, 110847, '\P{kanasup}', "");
Expect(1, 110847, '\P{^kanasup}', "");
Expect(0, 110848, '\p{kanasup}', "");
Expect(1, 110848, '\p{^kanasup}', "");
Expect(1, 110848, '\P{kanasup}', "");
Expect(0, 110848, '\P{^kanasup}', "");
Expect(1, 110847, '\p{	Kana_sup}', "");
Expect(0, 110847, '\p{^	Kana_sup}', "");
Expect(0, 110847, '\P{	Kana_sup}', "");
Expect(1, 110847, '\P{^	Kana_sup}', "");
Expect(0, 110848, '\p{	Kana_sup}', "");
Expect(1, 110848, '\p{^	Kana_sup}', "");
Expect(1, 110848, '\P{	Kana_sup}', "");
Expect(0, 110848, '\P{^	Kana_sup}', "");
Error('\p{/a/-Is_kana_Sup}');
Error('\P{/a/-Is_kana_Sup}');
Expect(1, 110847, '\p{iskanasup}', "");
Expect(0, 110847, '\p{^iskanasup}', "");
Expect(0, 110847, '\P{iskanasup}', "");
Expect(1, 110847, '\P{^iskanasup}', "");
Expect(0, 110848, '\p{iskanasup}', "");
Expect(1, 110848, '\p{^iskanasup}', "");
Expect(1, 110848, '\P{iskanasup}', "");
Expect(0, 110848, '\P{^iskanasup}', "");
Expect(1, 110847, '\p{-_Is_KANA_SUP}', "");
Expect(0, 110847, '\p{^-_Is_KANA_SUP}', "");
Expect(0, 110847, '\P{-_Is_KANA_SUP}', "");
Expect(1, 110847, '\P{^-_Is_KANA_SUP}', "");
Expect(0, 110848, '\p{-_Is_KANA_SUP}', "");
Expect(1, 110848, '\p{^-_Is_KANA_SUP}', "");
Expect(1, 110848, '\P{-_Is_KANA_SUP}', "");
Expect(0, 110848, '\P{^-_Is_KANA_SUP}', "");
Error('\p{	-In_Kana_Sup:=}');
Error('\P{	-In_Kana_Sup:=}');
Expect(1, 110847, '\p{inkanasup}', "");
Expect(0, 110847, '\p{^inkanasup}', "");
Expect(0, 110847, '\P{inkanasup}', "");
Expect(1, 110847, '\P{^inkanasup}', "");
Expect(0, 110848, '\p{inkanasup}', "");
Expect(1, 110848, '\p{^inkanasup}', "");
Expect(1, 110848, '\P{inkanasup}', "");
Expect(0, 110848, '\P{^inkanasup}', "");
Expect(1, 110847, '\p{	_In_KANA_Sup}', "");
Expect(0, 110847, '\p{^	_In_KANA_Sup}', "");
Expect(0, 110847, '\P{	_In_KANA_Sup}', "");
Expect(1, 110847, '\P{^	_In_KANA_Sup}', "");
Expect(0, 110848, '\p{	_In_KANA_Sup}', "");
Expect(1, 110848, '\p{^	_In_KANA_Sup}', "");
Expect(1, 110848, '\P{	_In_KANA_Sup}', "");
Expect(0, 110848, '\P{^	_In_KANA_Sup}', "");
Error('\p{-_Kanbun/a/}');
Error('\P{-_Kanbun/a/}');
Expect(1, 12703, '\p{kanbun}', "");
Expect(0, 12703, '\p{^kanbun}', "");
Expect(0, 12703, '\P{kanbun}', "");
Expect(1, 12703, '\P{^kanbun}', "");
Expect(0, 12704, '\p{kanbun}', "");
Expect(1, 12704, '\p{^kanbun}', "");
Expect(1, 12704, '\P{kanbun}', "");
Expect(0, 12704, '\P{^kanbun}', "");
Expect(1, 12703, '\p{- Kanbun}', "");
Expect(0, 12703, '\p{^- Kanbun}', "");
Expect(0, 12703, '\P{- Kanbun}', "");
Expect(1, 12703, '\P{^- Kanbun}', "");
Expect(0, 12704, '\p{- Kanbun}', "");
Expect(1, 12704, '\p{^- Kanbun}', "");
Expect(1, 12704, '\P{- Kanbun}', "");
Expect(0, 12704, '\P{^- Kanbun}', "");
Error('\p{-:=is_Kanbun}');
Error('\P{-:=is_Kanbun}');
Expect(1, 12703, '\p{iskanbun}', "");
Expect(0, 12703, '\p{^iskanbun}', "");
Expect(0, 12703, '\P{iskanbun}', "");
Expect(1, 12703, '\P{^iskanbun}', "");
Expect(0, 12704, '\p{iskanbun}', "");
Expect(1, 12704, '\p{^iskanbun}', "");
Expect(1, 12704, '\P{iskanbun}', "");
Expect(0, 12704, '\P{^iskanbun}', "");
Expect(1, 12703, '\p{_	Is_KANBUN}', "");
Expect(0, 12703, '\p{^_	Is_KANBUN}', "");
Expect(0, 12703, '\P{_	Is_KANBUN}', "");
Expect(1, 12703, '\P{^_	Is_KANBUN}', "");
Expect(0, 12704, '\p{_	Is_KANBUN}', "");
Expect(1, 12704, '\p{^_	Is_KANBUN}', "");
Expect(1, 12704, '\P{_	Is_KANBUN}', "");
Expect(0, 12704, '\P{^_	Is_KANBUN}', "");
Error('\p{:=IN_kanbun}');
Error('\P{:=IN_kanbun}');
Expect(1, 12703, '\p{inkanbun}', "");
Expect(0, 12703, '\p{^inkanbun}', "");
Expect(0, 12703, '\P{inkanbun}', "");
Expect(1, 12703, '\P{^inkanbun}', "");
Expect(0, 12704, '\p{inkanbun}', "");
Expect(1, 12704, '\p{^inkanbun}', "");
Expect(1, 12704, '\P{inkanbun}', "");
Expect(0, 12704, '\P{^inkanbun}', "");
Expect(1, 12703, '\p{ 	in_Kanbun}', "");
Expect(0, 12703, '\p{^ 	in_Kanbun}', "");
Expect(0, 12703, '\P{ 	in_Kanbun}', "");
Expect(1, 12703, '\P{^ 	in_Kanbun}', "");
Expect(0, 12704, '\p{ 	in_Kanbun}', "");
Expect(1, 12704, '\p{^ 	in_Kanbun}', "");
Expect(1, 12704, '\P{ 	in_Kanbun}', "");
Expect(0, 12704, '\P{^ 	in_Kanbun}', "");
Error('\p{:=-kangxi_RADICALS}');
Error('\P{:=-kangxi_RADICALS}');
Expect(1, 12255, '\p{kangxiradicals}', "");
Expect(0, 12255, '\p{^kangxiradicals}', "");
Expect(0, 12255, '\P{kangxiradicals}', "");
Expect(1, 12255, '\P{^kangxiradicals}', "");
Expect(0, 12256, '\p{kangxiradicals}', "");
Expect(1, 12256, '\p{^kangxiradicals}', "");
Expect(1, 12256, '\P{kangxiradicals}', "");
Expect(0, 12256, '\P{^kangxiradicals}', "");
Expect(1, 12255, '\p{_	KANGXI_radicals}', "");
Expect(0, 12255, '\p{^_	KANGXI_radicals}', "");
Expect(0, 12255, '\P{_	KANGXI_radicals}', "");
Expect(1, 12255, '\P{^_	KANGXI_radicals}', "");
Expect(0, 12256, '\p{_	KANGXI_radicals}', "");
Expect(1, 12256, '\p{^_	KANGXI_radicals}', "");
Expect(1, 12256, '\P{_	KANGXI_radicals}', "");
Expect(0, 12256, '\P{^_	KANGXI_radicals}', "");
Error('\p{/a/		Is_Kangxi_Radicals}');
Error('\P{/a/		Is_Kangxi_Radicals}');
Expect(1, 12255, '\p{iskangxiradicals}', "");
Expect(0, 12255, '\p{^iskangxiradicals}', "");
Expect(0, 12255, '\P{iskangxiradicals}', "");
Expect(1, 12255, '\P{^iskangxiradicals}', "");
Expect(0, 12256, '\p{iskangxiradicals}', "");
Expect(1, 12256, '\p{^iskangxiradicals}', "");
Expect(1, 12256, '\P{iskangxiradicals}', "");
Expect(0, 12256, '\P{^iskangxiradicals}', "");
Expect(1, 12255, '\p{- Is_KANGXI_RADICALS}', "");
Expect(0, 12255, '\p{^- Is_KANGXI_RADICALS}', "");
Expect(0, 12255, '\P{- Is_KANGXI_RADICALS}', "");
Expect(1, 12255, '\P{^- Is_KANGXI_RADICALS}', "");
Expect(0, 12256, '\p{- Is_KANGXI_RADICALS}', "");
Expect(1, 12256, '\p{^- Is_KANGXI_RADICALS}', "");
Expect(1, 12256, '\P{- Is_KANGXI_RADICALS}', "");
Expect(0, 12256, '\P{^- Is_KANGXI_RADICALS}', "");
Error('\p{/a/In_KANGXI_Radicals}');
Error('\P{/a/In_KANGXI_Radicals}');
Expect(1, 12255, '\p{inkangxiradicals}', "");
Expect(0, 12255, '\p{^inkangxiradicals}', "");
Expect(0, 12255, '\P{inkangxiradicals}', "");
Expect(1, 12255, '\P{^inkangxiradicals}', "");
Expect(0, 12256, '\p{inkangxiradicals}', "");
Expect(1, 12256, '\p{^inkangxiradicals}', "");
Expect(1, 12256, '\P{inkangxiradicals}', "");
Expect(0, 12256, '\P{^inkangxiradicals}', "");
Expect(1, 12255, '\p{	In_KANGXI_RADICALS}', "");
Expect(0, 12255, '\p{^	In_KANGXI_RADICALS}', "");
Expect(0, 12255, '\P{	In_KANGXI_RADICALS}', "");
Expect(1, 12255, '\P{^	In_KANGXI_RADICALS}', "");
Expect(0, 12256, '\p{	In_KANGXI_RADICALS}', "");
Expect(1, 12256, '\p{^	In_KANGXI_RADICALS}', "");
Expect(1, 12256, '\P{	In_KANGXI_RADICALS}', "");
Expect(0, 12256, '\P{^	In_KANGXI_RADICALS}', "");
Error('\p{-:=KANGXI}');
Error('\P{-:=KANGXI}');
Expect(1, 12255, '\p{kangxi}', "");
Expect(0, 12255, '\p{^kangxi}', "");
Expect(0, 12255, '\P{kangxi}', "");
Expect(1, 12255, '\P{^kangxi}', "");
Expect(0, 12256, '\p{kangxi}', "");
Expect(1, 12256, '\p{^kangxi}', "");
Expect(1, 12256, '\P{kangxi}', "");
Expect(0, 12256, '\P{^kangxi}', "");
Expect(1, 12255, '\p{_-KANGXI}', "");
Expect(0, 12255, '\p{^_-KANGXI}', "");
Expect(0, 12255, '\P{_-KANGXI}', "");
Expect(1, 12255, '\P{^_-KANGXI}', "");
Expect(0, 12256, '\p{_-KANGXI}', "");
Expect(1, 12256, '\p{^_-KANGXI}', "");
Expect(1, 12256, '\P{_-KANGXI}', "");
Expect(0, 12256, '\P{^_-KANGXI}', "");
Error('\p{-Is_Kangxi:=}');
Error('\P{-Is_Kangxi:=}');
Expect(1, 12255, '\p{iskangxi}', "");
Expect(0, 12255, '\p{^iskangxi}', "");
Expect(0, 12255, '\P{iskangxi}', "");
Expect(1, 12255, '\P{^iskangxi}', "");
Expect(0, 12256, '\p{iskangxi}', "");
Expect(1, 12256, '\p{^iskangxi}', "");
Expect(1, 12256, '\P{iskangxi}', "");
Expect(0, 12256, '\P{^iskangxi}', "");
Expect(1, 12255, '\p{	 is_Kangxi}', "");
Expect(0, 12255, '\p{^	 is_Kangxi}', "");
Expect(0, 12255, '\P{	 is_Kangxi}', "");
Expect(1, 12255, '\P{^	 is_Kangxi}', "");
Expect(0, 12256, '\p{	 is_Kangxi}', "");
Expect(1, 12256, '\p{^	 is_Kangxi}', "");
Expect(1, 12256, '\P{	 is_Kangxi}', "");
Expect(0, 12256, '\P{^	 is_Kangxi}', "");
Error('\p{_In_kangxi:=}');
Error('\P{_In_kangxi:=}');
Expect(1, 12255, '\p{inkangxi}', "");
Expect(0, 12255, '\p{^inkangxi}', "");
Expect(0, 12255, '\P{inkangxi}', "");
Expect(1, 12255, '\P{^inkangxi}', "");
Expect(0, 12256, '\p{inkangxi}', "");
Expect(1, 12256, '\p{^inkangxi}', "");
Expect(1, 12256, '\P{inkangxi}', "");
Expect(0, 12256, '\P{^inkangxi}', "");
Expect(1, 12255, '\p{-in_Kangxi}', "");
Expect(0, 12255, '\p{^-in_Kangxi}', "");
Expect(0, 12255, '\P{-in_Kangxi}', "");
Expect(1, 12255, '\P{^-in_Kangxi}', "");
Expect(0, 12256, '\p{-in_Kangxi}', "");
Expect(1, 12256, '\p{^-in_Kangxi}', "");
Expect(1, 12256, '\P{-in_Kangxi}', "");
Expect(0, 12256, '\P{^-in_Kangxi}', "");
Error('\p{/a/-_kannada}');
Error('\P{/a/-_kannada}');
Expect(1, 3314, '\p{kannada}', "");
Expect(0, 3314, '\p{^kannada}', "");
Expect(0, 3314, '\P{kannada}', "");
Expect(1, 3314, '\P{^kannada}', "");
Expect(0, 3315, '\p{kannada}', "");
Expect(1, 3315, '\p{^kannada}', "");
Expect(1, 3315, '\P{kannada}', "");
Expect(0, 3315, '\P{^kannada}', "");
Expect(1, 3314, '\p{	 Kannada}', "");
Expect(0, 3314, '\p{^	 Kannada}', "");
Expect(0, 3314, '\P{	 Kannada}', "");
Expect(1, 3314, '\P{^	 Kannada}', "");
Expect(0, 3315, '\p{	 Kannada}', "");
Expect(1, 3315, '\p{^	 Kannada}', "");
Expect(1, 3315, '\P{	 Kannada}', "");
Expect(0, 3315, '\P{^	 Kannada}', "");
Error('\p{-	Is_Kannada:=}');
Error('\P{-	Is_Kannada:=}');
Expect(1, 3314, '\p{iskannada}', "");
Expect(0, 3314, '\p{^iskannada}', "");
Expect(0, 3314, '\P{iskannada}', "");
Expect(1, 3314, '\P{^iskannada}', "");
Expect(0, 3315, '\p{iskannada}', "");
Expect(1, 3315, '\p{^iskannada}', "");
Expect(1, 3315, '\P{iskannada}', "");
Expect(0, 3315, '\P{^iskannada}', "");
Expect(1, 3314, '\p{ is_Kannada}', "");
Expect(0, 3314, '\p{^ is_Kannada}', "");
Expect(0, 3314, '\P{ is_Kannada}', "");
Expect(1, 3314, '\P{^ is_Kannada}', "");
Expect(0, 3315, '\p{ is_Kannada}', "");
Expect(1, 3315, '\p{^ is_Kannada}', "");
Expect(1, 3315, '\P{ is_Kannada}', "");
Expect(0, 3315, '\P{^ is_Kannada}', "");
Error('\p{	_knda/a/}');
Error('\P{	_knda/a/}');
Expect(1, 3314, '\p{knda}', "");
Expect(0, 3314, '\p{^knda}', "");
Expect(0, 3314, '\P{knda}', "");
Expect(1, 3314, '\P{^knda}', "");
Expect(0, 3315, '\p{knda}', "");
Expect(1, 3315, '\p{^knda}', "");
Expect(1, 3315, '\P{knda}', "");
Expect(0, 3315, '\P{^knda}', "");
Expect(1, 3314, '\p{ -Knda}', "");
Expect(0, 3314, '\p{^ -Knda}', "");
Expect(0, 3314, '\P{ -Knda}', "");
Expect(1, 3314, '\P{^ -Knda}', "");
Expect(0, 3315, '\p{ -Knda}', "");
Expect(1, 3315, '\p{^ -Knda}', "");
Expect(1, 3315, '\P{ -Knda}', "");
Expect(0, 3315, '\P{^ -Knda}', "");
Error('\p{_is_knda/a/}');
Error('\P{_is_knda/a/}');
Expect(1, 3314, '\p{isknda}', "");
Expect(0, 3314, '\p{^isknda}', "");
Expect(0, 3314, '\P{isknda}', "");
Expect(1, 3314, '\P{^isknda}', "");
Expect(0, 3315, '\p{isknda}', "");
Expect(1, 3315, '\p{^isknda}', "");
Expect(1, 3315, '\P{isknda}', "");
Expect(0, 3315, '\P{^isknda}', "");
Expect(1, 3314, '\p{-Is_knda}', "");
Expect(0, 3314, '\p{^-Is_knda}', "");
Expect(0, 3314, '\P{-Is_knda}', "");
Expect(1, 3314, '\P{^-Is_knda}', "");
Expect(0, 3315, '\p{-Is_knda}', "");
Expect(1, 3315, '\p{^-Is_knda}', "");
Expect(1, 3315, '\P{-Is_knda}', "");
Expect(0, 3315, '\P{^-Is_knda}', "");
Error('\p{/a/_katakana}');
Error('\P{/a/_katakana}');
Expect(1, 110592, '\p{katakana}', "");
Expect(0, 110592, '\p{^katakana}', "");
Expect(0, 110592, '\P{katakana}', "");
Expect(1, 110592, '\P{^katakana}', "");
Expect(0, 110593, '\p{katakana}', "");
Expect(1, 110593, '\p{^katakana}', "");
Expect(1, 110593, '\P{katakana}', "");
Expect(0, 110593, '\P{^katakana}', "");
Expect(1, 110592, '\p{ KATAKANA}', "");
Expect(0, 110592, '\p{^ KATAKANA}', "");
Expect(0, 110592, '\P{ KATAKANA}', "");
Expect(1, 110592, '\P{^ KATAKANA}', "");
Expect(0, 110593, '\p{ KATAKANA}', "");
Expect(1, 110593, '\p{^ KATAKANA}', "");
Expect(1, 110593, '\P{ KATAKANA}', "");
Expect(0, 110593, '\P{^ KATAKANA}', "");
Error('\p{		IS_Katakana/a/}');
Error('\P{		IS_Katakana/a/}');
Expect(1, 110592, '\p{iskatakana}', "");
Expect(0, 110592, '\p{^iskatakana}', "");
Expect(0, 110592, '\P{iskatakana}', "");
Expect(1, 110592, '\P{^iskatakana}', "");
Expect(0, 110593, '\p{iskatakana}', "");
Expect(1, 110593, '\p{^iskatakana}', "");
Expect(1, 110593, '\P{iskatakana}', "");
Expect(0, 110593, '\P{^iskatakana}', "");
Expect(1, 110592, '\p{	_Is_Katakana}', "");
Expect(0, 110592, '\p{^	_Is_Katakana}', "");
Expect(0, 110592, '\P{	_Is_Katakana}', "");
Expect(1, 110592, '\P{^	_Is_Katakana}', "");
Expect(0, 110593, '\p{	_Is_Katakana}', "");
Expect(1, 110593, '\p{^	_Is_Katakana}', "");
Expect(1, 110593, '\P{	_Is_Katakana}', "");
Expect(0, 110593, '\P{^	_Is_Katakana}', "");
Error('\p{:=-_KANA}');
Error('\P{:=-_KANA}');
Expect(1, 110592, '\p{kana}', "");
Expect(0, 110592, '\p{^kana}', "");
Expect(0, 110592, '\P{kana}', "");
Expect(1, 110592, '\P{^kana}', "");
Expect(0, 110593, '\p{kana}', "");
Expect(1, 110593, '\p{^kana}', "");
Expect(1, 110593, '\P{kana}', "");
Expect(0, 110593, '\P{^kana}', "");
Expect(1, 110592, '\p{		Kana}', "");
Expect(0, 110592, '\p{^		Kana}', "");
Expect(0, 110592, '\P{		Kana}', "");
Expect(1, 110592, '\P{^		Kana}', "");
Expect(0, 110593, '\p{		Kana}', "");
Expect(1, 110593, '\p{^		Kana}', "");
Expect(1, 110593, '\P{		Kana}', "");
Expect(0, 110593, '\P{^		Kana}', "");
Error('\p{/a/_Is_Kana}');
Error('\P{/a/_Is_Kana}');
Expect(1, 110592, '\p{iskana}', "");
Expect(0, 110592, '\p{^iskana}', "");
Expect(0, 110592, '\P{iskana}', "");
Expect(1, 110592, '\P{^iskana}', "");
Expect(0, 110593, '\p{iskana}', "");
Expect(1, 110593, '\p{^iskana}', "");
Expect(1, 110593, '\P{iskana}', "");
Expect(0, 110593, '\P{^iskana}', "");
Expect(1, 110592, '\p{Is_Kana}', "");
Expect(0, 110592, '\p{^Is_Kana}', "");
Expect(0, 110592, '\P{Is_Kana}', "");
Expect(1, 110592, '\P{^Is_Kana}', "");
Expect(0, 110593, '\p{Is_Kana}', "");
Expect(1, 110593, '\p{^Is_Kana}', "");
Expect(1, 110593, '\P{Is_Kana}', "");
Expect(0, 110593, '\P{^Is_Kana}', "");
Error('\p{Katakana_Or_Hiragana}');
Error('\P{Katakana_Or_Hiragana}');
Error('\p{Is_Katakana_Or_Hiragana}');
Error('\P{Is_Katakana_Or_Hiragana}');
Error('\p{Hrkt}');
Error('\P{Hrkt}');
Error('\p{Is_Hrkt}');
Error('\P{Is_Hrkt}');
Error('\p{_/a/Katakana_PHONETIC_EXTENSIONS}');
Error('\P{_/a/Katakana_PHONETIC_EXTENSIONS}');
Expect(1, 12799, '\p{katakanaphoneticextensions}', "");
Expect(0, 12799, '\p{^katakanaphoneticextensions}', "");
Expect(0, 12799, '\P{katakanaphoneticextensions}', "");
Expect(1, 12799, '\P{^katakanaphoneticextensions}', "");
Expect(0, 12800, '\p{katakanaphoneticextensions}', "");
Expect(1, 12800, '\p{^katakanaphoneticextensions}', "");
Expect(1, 12800, '\P{katakanaphoneticextensions}', "");
Expect(0, 12800, '\P{^katakanaphoneticextensions}', "");
Expect(1, 12799, '\p{_ KATAKANA_Phonetic_EXTENSIONS}', "");
Expect(0, 12799, '\p{^_ KATAKANA_Phonetic_EXTENSIONS}', "");
Expect(0, 12799, '\P{_ KATAKANA_Phonetic_EXTENSIONS}', "");
Expect(1, 12799, '\P{^_ KATAKANA_Phonetic_EXTENSIONS}', "");
Expect(0, 12800, '\p{_ KATAKANA_Phonetic_EXTENSIONS}', "");
Expect(1, 12800, '\p{^_ KATAKANA_Phonetic_EXTENSIONS}', "");
Expect(1, 12800, '\P{_ KATAKANA_Phonetic_EXTENSIONS}', "");
Expect(0, 12800, '\P{^_ KATAKANA_Phonetic_EXTENSIONS}', "");
Error('\p{is_Katakana_Phonetic_Extensions:=}');
Error('\P{is_Katakana_Phonetic_Extensions:=}');
Expect(1, 12799, '\p{iskatakanaphoneticextensions}', "");
Expect(0, 12799, '\p{^iskatakanaphoneticextensions}', "");
Expect(0, 12799, '\P{iskatakanaphoneticextensions}', "");
Expect(1, 12799, '\P{^iskatakanaphoneticextensions}', "");
Expect(0, 12800, '\p{iskatakanaphoneticextensions}', "");
Expect(1, 12800, '\p{^iskatakanaphoneticextensions}', "");
Expect(1, 12800, '\P{iskatakanaphoneticextensions}', "");
Expect(0, 12800, '\P{^iskatakanaphoneticextensions}', "");
Expect(1, 12799, '\p{_-is_Katakana_Phonetic_extensions}', "");
Expect(0, 12799, '\p{^_-is_Katakana_Phonetic_extensions}', "");
Expect(0, 12799, '\P{_-is_Katakana_Phonetic_extensions}', "");
Expect(1, 12799, '\P{^_-is_Katakana_Phonetic_extensions}', "");
Expect(0, 12800, '\p{_-is_Katakana_Phonetic_extensions}', "");
Expect(1, 12800, '\p{^_-is_Katakana_Phonetic_extensions}', "");
Expect(1, 12800, '\P{_-is_Katakana_Phonetic_extensions}', "");
Expect(0, 12800, '\P{^_-is_Katakana_Phonetic_extensions}', "");
Error('\p{/a/_-In_Katakana_phonetic_extensions}');
Error('\P{/a/_-In_Katakana_phonetic_extensions}');
Expect(1, 12799, '\p{inkatakanaphoneticextensions}', "");
Expect(0, 12799, '\p{^inkatakanaphoneticextensions}', "");
Expect(0, 12799, '\P{inkatakanaphoneticextensions}', "");
Expect(1, 12799, '\P{^inkatakanaphoneticextensions}', "");
Expect(0, 12800, '\p{inkatakanaphoneticextensions}', "");
Expect(1, 12800, '\p{^inkatakanaphoneticextensions}', "");
Expect(1, 12800, '\P{inkatakanaphoneticextensions}', "");
Expect(0, 12800, '\P{^inkatakanaphoneticextensions}', "");
Expect(1, 12799, '\p{	_in_KATAKANA_Phonetic_EXTENSIONS}', "");
Expect(0, 12799, '\p{^	_in_KATAKANA_Phonetic_EXTENSIONS}', "");
Expect(0, 12799, '\P{	_in_KATAKANA_Phonetic_EXTENSIONS}', "");
Expect(1, 12799, '\P{^	_in_KATAKANA_Phonetic_EXTENSIONS}', "");
Expect(0, 12800, '\p{	_in_KATAKANA_Phonetic_EXTENSIONS}', "");
Expect(1, 12800, '\p{^	_in_KATAKANA_Phonetic_EXTENSIONS}', "");
Expect(1, 12800, '\P{	_in_KATAKANA_Phonetic_EXTENSIONS}', "");
Expect(0, 12800, '\P{^	_in_KATAKANA_Phonetic_EXTENSIONS}', "");
Error('\p{  Katakana_Ext/a/}');
Error('\P{  Katakana_Ext/a/}');
Expect(1, 12799, '\p{katakanaext}', "");
Expect(0, 12799, '\p{^katakanaext}', "");
Expect(0, 12799, '\P{katakanaext}', "");
Expect(1, 12799, '\P{^katakanaext}', "");
Expect(0, 12800, '\p{katakanaext}', "");
Expect(1, 12800, '\p{^katakanaext}', "");
Expect(1, 12800, '\P{katakanaext}', "");
Expect(0, 12800, '\P{^katakanaext}', "");
Expect(1, 12799, '\p{ -Katakana_ext}', "");
Expect(0, 12799, '\p{^ -Katakana_ext}', "");
Expect(0, 12799, '\P{ -Katakana_ext}', "");
Expect(1, 12799, '\P{^ -Katakana_ext}', "");
Expect(0, 12800, '\p{ -Katakana_ext}', "");
Expect(1, 12800, '\p{^ -Katakana_ext}', "");
Expect(1, 12800, '\P{ -Katakana_ext}', "");
Expect(0, 12800, '\P{^ -Katakana_ext}', "");
Error('\p{-/a/Is_Katakana_ext}');
Error('\P{-/a/Is_Katakana_ext}');
Expect(1, 12799, '\p{iskatakanaext}', "");
Expect(0, 12799, '\p{^iskatakanaext}', "");
Expect(0, 12799, '\P{iskatakanaext}', "");
Expect(1, 12799, '\P{^iskatakanaext}', "");
Expect(0, 12800, '\p{iskatakanaext}', "");
Expect(1, 12800, '\p{^iskatakanaext}', "");
Expect(1, 12800, '\P{iskatakanaext}', "");
Expect(0, 12800, '\P{^iskatakanaext}', "");
Expect(1, 12799, '\p{	IS_Katakana_EXT}', "");
Expect(0, 12799, '\p{^	IS_Katakana_EXT}', "");
Expect(0, 12799, '\P{	IS_Katakana_EXT}', "");
Expect(1, 12799, '\P{^	IS_Katakana_EXT}', "");
Expect(0, 12800, '\p{	IS_Katakana_EXT}', "");
Expect(1, 12800, '\p{^	IS_Katakana_EXT}', "");
Expect(1, 12800, '\P{	IS_Katakana_EXT}', "");
Expect(0, 12800, '\P{^	IS_Katakana_EXT}', "");
Error('\p{	 In_Katakana_EXT:=}');
Error('\P{	 In_Katakana_EXT:=}');
Expect(1, 12799, '\p{inkatakanaext}', "");
Expect(0, 12799, '\p{^inkatakanaext}', "");
Expect(0, 12799, '\P{inkatakanaext}', "");
Expect(1, 12799, '\P{^inkatakanaext}', "");
Expect(0, 12800, '\p{inkatakanaext}', "");
Expect(1, 12800, '\p{^inkatakanaext}', "");
Expect(1, 12800, '\P{inkatakanaext}', "");
Expect(0, 12800, '\P{^inkatakanaext}', "");
Expect(1, 12799, '\p{ -in_katakana_EXT}', "");
Expect(0, 12799, '\p{^ -in_katakana_EXT}', "");
Expect(0, 12799, '\P{ -in_katakana_EXT}', "");
Expect(1, 12799, '\P{^ -in_katakana_EXT}', "");
Expect(0, 12800, '\p{ -in_katakana_EXT}', "");
Expect(1, 12800, '\p{^ -in_katakana_EXT}', "");
Expect(1, 12800, '\P{ -in_katakana_EXT}', "");
Expect(0, 12800, '\P{^ -in_katakana_EXT}', "");
Error('\p{_/a/kayah_LI}');
Error('\P{_/a/kayah_LI}');
Expect(1, 43311, '\p{kayahli}', "");
Expect(0, 43311, '\p{^kayahli}', "");
Expect(0, 43311, '\P{kayahli}', "");
Expect(1, 43311, '\P{^kayahli}', "");
Expect(0, 43312, '\p{kayahli}', "");
Expect(1, 43312, '\p{^kayahli}', "");
Expect(1, 43312, '\P{kayahli}', "");
Expect(0, 43312, '\P{^kayahli}', "");
Expect(1, 43311, '\p{- KAYAH_Li}', "");
Expect(0, 43311, '\p{^- KAYAH_Li}', "");
Expect(0, 43311, '\P{- KAYAH_Li}', "");
Expect(1, 43311, '\P{^- KAYAH_Li}', "");
Expect(0, 43312, '\p{- KAYAH_Li}', "");
Expect(1, 43312, '\p{^- KAYAH_Li}', "");
Expect(1, 43312, '\P{- KAYAH_Li}', "");
Expect(0, 43312, '\P{^- KAYAH_Li}', "");
Error('\p{/a/- IS_KAYAH_Li}');
Error('\P{/a/- IS_KAYAH_Li}');
Expect(1, 43311, '\p{iskayahli}', "");
Expect(0, 43311, '\p{^iskayahli}', "");
Expect(0, 43311, '\P{iskayahli}', "");
Expect(1, 43311, '\P{^iskayahli}', "");
Expect(0, 43312, '\p{iskayahli}', "");
Expect(1, 43312, '\p{^iskayahli}', "");
Expect(1, 43312, '\P{iskayahli}', "");
Expect(0, 43312, '\P{^iskayahli}', "");
Expect(1, 43311, '\p{		is_kayah_li}', "");
Expect(0, 43311, '\p{^		is_kayah_li}', "");
Expect(0, 43311, '\P{		is_kayah_li}', "");
Expect(1, 43311, '\P{^		is_kayah_li}', "");
Expect(0, 43312, '\p{		is_kayah_li}', "");
Expect(1, 43312, '\p{^		is_kayah_li}', "");
Expect(1, 43312, '\P{		is_kayah_li}', "");
Expect(0, 43312, '\P{^		is_kayah_li}', "");
Error('\p{-:=KALI}');
Error('\P{-:=KALI}');
Expect(1, 43311, '\p{kali}', "");
Expect(0, 43311, '\p{^kali}', "");
Expect(0, 43311, '\P{kali}', "");
Expect(1, 43311, '\P{^kali}', "");
Expect(0, 43312, '\p{kali}', "");
Expect(1, 43312, '\p{^kali}', "");
Expect(1, 43312, '\P{kali}', "");
Expect(0, 43312, '\P{^kali}', "");
Expect(1, 43311, '\p{-_kali}', "");
Expect(0, 43311, '\p{^-_kali}', "");
Expect(0, 43311, '\P{-_kali}', "");
Expect(1, 43311, '\P{^-_kali}', "");
Expect(0, 43312, '\p{-_kali}', "");
Expect(1, 43312, '\p{^-_kali}', "");
Expect(1, 43312, '\P{-_kali}', "");
Expect(0, 43312, '\P{^-_kali}', "");
Error('\p{is_Kali:=}');
Error('\P{is_Kali:=}');
Expect(1, 43311, '\p{iskali}', "");
Expect(0, 43311, '\p{^iskali}', "");
Expect(0, 43311, '\P{iskali}', "");
Expect(1, 43311, '\P{^iskali}', "");
Expect(0, 43312, '\p{iskali}', "");
Expect(1, 43312, '\p{^iskali}', "");
Expect(1, 43312, '\P{iskali}', "");
Expect(0, 43312, '\P{^iskali}', "");
Expect(1, 43311, '\p{ 	IS_kali}', "");
Expect(0, 43311, '\p{^ 	IS_kali}', "");
Expect(0, 43311, '\P{ 	IS_kali}', "");
Expect(1, 43311, '\P{^ 	IS_kali}', "");
Expect(0, 43312, '\p{ 	IS_kali}', "");
Expect(1, 43312, '\p{^ 	IS_kali}', "");
Expect(1, 43312, '\P{ 	IS_kali}', "");
Expect(0, 43312, '\P{^ 	IS_kali}', "");
Error('\p{_KHAROSHTHI:=}');
Error('\P{_KHAROSHTHI:=}');
Expect(1, 68184, '\p{kharoshthi}', "");
Expect(0, 68184, '\p{^kharoshthi}', "");
Expect(0, 68184, '\P{kharoshthi}', "");
Expect(1, 68184, '\P{^kharoshthi}', "");
Expect(0, 68185, '\p{kharoshthi}', "");
Expect(1, 68185, '\p{^kharoshthi}', "");
Expect(1, 68185, '\P{kharoshthi}', "");
Expect(0, 68185, '\P{^kharoshthi}', "");
Expect(1, 68184, '\p{	Kharoshthi}', "");
Expect(0, 68184, '\p{^	Kharoshthi}', "");
Expect(0, 68184, '\P{	Kharoshthi}', "");
Expect(1, 68184, '\P{^	Kharoshthi}', "");
Expect(0, 68185, '\p{	Kharoshthi}', "");
Expect(1, 68185, '\p{^	Kharoshthi}', "");
Expect(1, 68185, '\P{	Kharoshthi}', "");
Expect(0, 68185, '\P{^	Kharoshthi}', "");
Error('\p{/a/- IS_Kharoshthi}');
Error('\P{/a/- IS_Kharoshthi}');
Expect(1, 68184, '\p{iskharoshthi}', "");
Expect(0, 68184, '\p{^iskharoshthi}', "");
Expect(0, 68184, '\P{iskharoshthi}', "");
Expect(1, 68184, '\P{^iskharoshthi}', "");
Expect(0, 68185, '\p{iskharoshthi}', "");
Expect(1, 68185, '\p{^iskharoshthi}', "");
Expect(1, 68185, '\P{iskharoshthi}', "");
Expect(0, 68185, '\P{^iskharoshthi}', "");
Expect(1, 68184, '\p{	-is_Kharoshthi}', "");
Expect(0, 68184, '\p{^	-is_Kharoshthi}', "");
Expect(0, 68184, '\P{	-is_Kharoshthi}', "");
Expect(1, 68184, '\P{^	-is_Kharoshthi}', "");
Expect(0, 68185, '\p{	-is_Kharoshthi}', "");
Expect(1, 68185, '\p{^	-is_Kharoshthi}', "");
Expect(1, 68185, '\P{	-is_Kharoshthi}', "");
Expect(0, 68185, '\P{^	-is_Kharoshthi}', "");
Error('\p{ :=KHAR}');
Error('\P{ :=KHAR}');
Expect(1, 68184, '\p{khar}', "");
Expect(0, 68184, '\p{^khar}', "");
Expect(0, 68184, '\P{khar}', "");
Expect(1, 68184, '\P{^khar}', "");
Expect(0, 68185, '\p{khar}', "");
Expect(1, 68185, '\p{^khar}', "");
Expect(1, 68185, '\P{khar}', "");
Expect(0, 68185, '\P{^khar}', "");
Expect(1, 68184, '\p{	Khar}', "");
Expect(0, 68184, '\p{^	Khar}', "");
Expect(0, 68184, '\P{	Khar}', "");
Expect(1, 68184, '\P{^	Khar}', "");
Expect(0, 68185, '\p{	Khar}', "");
Expect(1, 68185, '\p{^	Khar}', "");
Expect(1, 68185, '\P{	Khar}', "");
Expect(0, 68185, '\P{^	Khar}', "");
Error('\p{/a/	Is_Khar}');
Error('\P{/a/	Is_Khar}');
Expect(1, 68184, '\p{iskhar}', "");
Expect(0, 68184, '\p{^iskhar}', "");
Expect(0, 68184, '\P{iskhar}', "");
Expect(1, 68184, '\P{^iskhar}', "");
Expect(0, 68185, '\p{iskhar}', "");
Expect(1, 68185, '\p{^iskhar}', "");
Expect(1, 68185, '\P{iskhar}', "");
Expect(0, 68185, '\P{^iskhar}', "");
Expect(1, 68184, '\p{	-is_KHAR}', "");
Expect(0, 68184, '\p{^	-is_KHAR}', "");
Expect(0, 68184, '\P{	-is_KHAR}', "");
Expect(1, 68184, '\P{^	-is_KHAR}', "");
Expect(0, 68185, '\p{	-is_KHAR}', "");
Expect(1, 68185, '\p{^	-is_KHAR}', "");
Expect(1, 68185, '\P{	-is_KHAR}', "");
Expect(0, 68185, '\P{^	-is_KHAR}', "");
Error('\p{	:=Khmer}');
Error('\P{	:=Khmer}');
Expect(1, 6655, '\p{khmer}', "");
Expect(0, 6655, '\p{^khmer}', "");
Expect(0, 6655, '\P{khmer}', "");
Expect(1, 6655, '\P{^khmer}', "");
Expect(0, 6656, '\p{khmer}', "");
Expect(1, 6656, '\p{^khmer}', "");
Expect(1, 6656, '\P{khmer}', "");
Expect(0, 6656, '\P{^khmer}', "");
Expect(1, 6655, '\p{	_Khmer}', "");
Expect(0, 6655, '\p{^	_Khmer}', "");
Expect(0, 6655, '\P{	_Khmer}', "");
Expect(1, 6655, '\P{^	_Khmer}', "");
Expect(0, 6656, '\p{	_Khmer}', "");
Expect(1, 6656, '\p{^	_Khmer}', "");
Expect(1, 6656, '\P{	_Khmer}', "");
Expect(0, 6656, '\P{^	_Khmer}', "");
Error('\p{-/a/is_Khmer}');
Error('\P{-/a/is_Khmer}');
Expect(1, 6655, '\p{iskhmer}', "");
Expect(0, 6655, '\p{^iskhmer}', "");
Expect(0, 6655, '\P{iskhmer}', "");
Expect(1, 6655, '\P{^iskhmer}', "");
Expect(0, 6656, '\p{iskhmer}', "");
Expect(1, 6656, '\p{^iskhmer}', "");
Expect(1, 6656, '\P{iskhmer}', "");
Expect(0, 6656, '\P{^iskhmer}', "");
Expect(1, 6655, '\p{ -Is_Khmer}', "");
Expect(0, 6655, '\p{^ -Is_Khmer}', "");
Expect(0, 6655, '\P{ -Is_Khmer}', "");
Expect(1, 6655, '\P{^ -Is_Khmer}', "");
Expect(0, 6656, '\p{ -Is_Khmer}', "");
Expect(1, 6656, '\p{^ -Is_Khmer}', "");
Expect(1, 6656, '\P{ -Is_Khmer}', "");
Expect(0, 6656, '\P{^ -Is_Khmer}', "");
Error('\p{_	KHMR:=}');
Error('\P{_	KHMR:=}');
Expect(1, 6655, '\p{khmr}', "");
Expect(0, 6655, '\p{^khmr}', "");
Expect(0, 6655, '\P{khmr}', "");
Expect(1, 6655, '\P{^khmr}', "");
Expect(0, 6656, '\p{khmr}', "");
Expect(1, 6656, '\p{^khmr}', "");
Expect(1, 6656, '\P{khmr}', "");
Expect(0, 6656, '\P{^khmr}', "");
Expect(1, 6655, '\p{-KHMR}', "");
Expect(0, 6655, '\p{^-KHMR}', "");
Expect(0, 6655, '\P{-KHMR}', "");
Expect(1, 6655, '\P{^-KHMR}', "");
Expect(0, 6656, '\p{-KHMR}', "");
Expect(1, 6656, '\p{^-KHMR}', "");
Expect(1, 6656, '\P{-KHMR}', "");
Expect(0, 6656, '\P{^-KHMR}', "");
Error('\p{_:=Is_khmr}');
Error('\P{_:=Is_khmr}');
Expect(1, 6655, '\p{iskhmr}', "");
Expect(0, 6655, '\p{^iskhmr}', "");
Expect(0, 6655, '\P{iskhmr}', "");
Expect(1, 6655, '\P{^iskhmr}', "");
Expect(0, 6656, '\p{iskhmr}', "");
Expect(1, 6656, '\p{^iskhmr}', "");
Expect(1, 6656, '\P{iskhmr}', "");
Expect(0, 6656, '\P{^iskhmr}', "");
Expect(1, 6655, '\p{  Is_Khmr}', "");
Expect(0, 6655, '\p{^  Is_Khmr}', "");
Expect(0, 6655, '\P{  Is_Khmr}', "");
Expect(1, 6655, '\P{^  Is_Khmr}', "");
Expect(0, 6656, '\p{  Is_Khmr}', "");
Expect(1, 6656, '\p{^  Is_Khmr}', "");
Expect(1, 6656, '\P{  Is_Khmr}', "");
Expect(0, 6656, '\P{^  Is_Khmr}', "");
Error('\p{-:=khmer_Symbols}');
Error('\P{-:=khmer_Symbols}');
Expect(1, 6655, '\p{khmersymbols}', "");
Expect(0, 6655, '\p{^khmersymbols}', "");
Expect(0, 6655, '\P{khmersymbols}', "");
Expect(1, 6655, '\P{^khmersymbols}', "");
Expect(0, 6656, '\p{khmersymbols}', "");
Expect(1, 6656, '\p{^khmersymbols}', "");
Expect(1, 6656, '\P{khmersymbols}', "");
Expect(0, 6656, '\P{^khmersymbols}', "");
Expect(1, 6655, '\p{__KHMER_Symbols}', "");
Expect(0, 6655, '\p{^__KHMER_Symbols}', "");
Expect(0, 6655, '\P{__KHMER_Symbols}', "");
Expect(1, 6655, '\P{^__KHMER_Symbols}', "");
Expect(0, 6656, '\p{__KHMER_Symbols}', "");
Expect(1, 6656, '\p{^__KHMER_Symbols}', "");
Expect(1, 6656, '\P{__KHMER_Symbols}', "");
Expect(0, 6656, '\P{^__KHMER_Symbols}', "");
Error('\p{:=is_Khmer_SYMBOLS}');
Error('\P{:=is_Khmer_SYMBOLS}');
Expect(1, 6655, '\p{iskhmersymbols}', "");
Expect(0, 6655, '\p{^iskhmersymbols}', "");
Expect(0, 6655, '\P{iskhmersymbols}', "");
Expect(1, 6655, '\P{^iskhmersymbols}', "");
Expect(0, 6656, '\p{iskhmersymbols}', "");
Expect(1, 6656, '\p{^iskhmersymbols}', "");
Expect(1, 6656, '\P{iskhmersymbols}', "");
Expect(0, 6656, '\P{^iskhmersymbols}', "");
Expect(1, 6655, '\p{	 Is_KHMER_Symbols}', "");
Expect(0, 6655, '\p{^	 Is_KHMER_Symbols}', "");
Expect(0, 6655, '\P{	 Is_KHMER_Symbols}', "");
Expect(1, 6655, '\P{^	 Is_KHMER_Symbols}', "");
Expect(0, 6656, '\p{	 Is_KHMER_Symbols}', "");
Expect(1, 6656, '\p{^	 Is_KHMER_Symbols}', "");
Expect(1, 6656, '\P{	 Is_KHMER_Symbols}', "");
Expect(0, 6656, '\P{^	 Is_KHMER_Symbols}', "");
Error('\p{	 In_KHMER_symbols/a/}');
Error('\P{	 In_KHMER_symbols/a/}');
Expect(1, 6655, '\p{inkhmersymbols}', "");
Expect(0, 6655, '\p{^inkhmersymbols}', "");
Expect(0, 6655, '\P{inkhmersymbols}', "");
Expect(1, 6655, '\P{^inkhmersymbols}', "");
Expect(0, 6656, '\p{inkhmersymbols}', "");
Expect(1, 6656, '\p{^inkhmersymbols}', "");
Expect(1, 6656, '\P{inkhmersymbols}', "");
Expect(0, 6656, '\P{^inkhmersymbols}', "");
Expect(1, 6655, '\p{	In_Khmer_SYMBOLS}', "");
Expect(0, 6655, '\p{^	In_Khmer_SYMBOLS}', "");
Expect(0, 6655, '\P{	In_Khmer_SYMBOLS}', "");
Expect(1, 6655, '\P{^	In_Khmer_SYMBOLS}', "");
Expect(0, 6656, '\p{	In_Khmer_SYMBOLS}', "");
Expect(1, 6656, '\p{^	In_Khmer_SYMBOLS}', "");
Expect(1, 6656, '\P{	In_Khmer_SYMBOLS}', "");
Expect(0, 6656, '\P{^	In_Khmer_SYMBOLS}', "");
Error('\p{:=Lao}');
Error('\P{:=Lao}');
Expect(1, 3807, '\p{lao}', "");
Expect(0, 3807, '\p{^lao}', "");
Expect(0, 3807, '\P{lao}', "");
Expect(1, 3807, '\P{^lao}', "");
Expect(0, 3808, '\p{lao}', "");
Expect(1, 3808, '\p{^lao}', "");
Expect(1, 3808, '\P{lao}', "");
Expect(0, 3808, '\P{^lao}', "");
Expect(1, 3807, '\p{- Lao}', "");
Expect(0, 3807, '\p{^- Lao}', "");
Expect(0, 3807, '\P{- Lao}', "");
Expect(1, 3807, '\P{^- Lao}', "");
Expect(0, 3808, '\p{- Lao}', "");
Expect(1, 3808, '\p{^- Lao}', "");
Expect(1, 3808, '\P{- Lao}', "");
Expect(0, 3808, '\P{^- Lao}', "");
Error('\p{/a/IS_Lao}');
Error('\P{/a/IS_Lao}');
Expect(1, 3807, '\p{islao}', "");
Expect(0, 3807, '\p{^islao}', "");
Expect(0, 3807, '\P{islao}', "");
Expect(1, 3807, '\P{^islao}', "");
Expect(0, 3808, '\p{islao}', "");
Expect(1, 3808, '\p{^islao}', "");
Expect(1, 3808, '\P{islao}', "");
Expect(0, 3808, '\P{^islao}', "");
Expect(1, 3807, '\p{-	IS_lao}', "");
Expect(0, 3807, '\p{^-	IS_lao}', "");
Expect(0, 3807, '\P{-	IS_lao}', "");
Expect(1, 3807, '\P{^-	IS_lao}', "");
Expect(0, 3808, '\p{-	IS_lao}', "");
Expect(1, 3808, '\p{^-	IS_lao}', "");
Expect(1, 3808, '\P{-	IS_lao}', "");
Expect(0, 3808, '\P{^-	IS_lao}', "");
Error('\p{	:=LAOO}');
Error('\P{	:=LAOO}');
Expect(1, 3807, '\p{laoo}', "");
Expect(0, 3807, '\p{^laoo}', "");
Expect(0, 3807, '\P{laoo}', "");
Expect(1, 3807, '\P{^laoo}', "");
Expect(0, 3808, '\p{laoo}', "");
Expect(1, 3808, '\p{^laoo}', "");
Expect(1, 3808, '\P{laoo}', "");
Expect(0, 3808, '\P{^laoo}', "");
Expect(1, 3807, '\p{- LAOO}', "");
Expect(0, 3807, '\p{^- LAOO}', "");
Expect(0, 3807, '\P{- LAOO}', "");
Expect(1, 3807, '\P{^- LAOO}', "");
Expect(0, 3808, '\p{- LAOO}', "");
Expect(1, 3808, '\p{^- LAOO}', "");
Expect(1, 3808, '\P{- LAOO}', "");
Expect(0, 3808, '\P{^- LAOO}', "");
Error('\p{	_Is_Laoo:=}');
Error('\P{	_Is_Laoo:=}');
Expect(1, 3807, '\p{islaoo}', "");
Expect(0, 3807, '\p{^islaoo}', "");
Expect(0, 3807, '\P{islaoo}', "");
Expect(1, 3807, '\P{^islaoo}', "");
Expect(0, 3808, '\p{islaoo}', "");
Expect(1, 3808, '\p{^islaoo}', "");
Expect(1, 3808, '\P{islaoo}', "");
Expect(0, 3808, '\P{^islaoo}', "");
Expect(1, 3807, '\p{	 IS_laoo}', "");
Expect(0, 3807, '\p{^	 IS_laoo}', "");
Expect(0, 3807, '\P{	 IS_laoo}', "");
Expect(1, 3807, '\P{^	 IS_laoo}', "");
Expect(0, 3808, '\p{	 IS_laoo}', "");
Expect(1, 3808, '\p{^	 IS_laoo}', "");
Expect(1, 3808, '\P{	 IS_laoo}', "");
Expect(0, 3808, '\P{^	 IS_laoo}', "");
Error('\p{ _Latin/a/}');
Error('\P{ _Latin/a/}');
Expect(1, 65370, '\p{latin}', "");
Expect(0, 65370, '\p{^latin}', "");
Expect(0, 65370, '\P{latin}', "");
Expect(1, 65370, '\P{^latin}', "");
Expect(0, 65371, '\p{latin}', "");
Expect(1, 65371, '\p{^latin}', "");
Expect(1, 65371, '\P{latin}', "");
Expect(0, 65371, '\P{^latin}', "");
Expect(1, 65370, '\p{_latin}', "");
Expect(0, 65370, '\p{^_latin}', "");
Expect(0, 65370, '\P{_latin}', "");
Expect(1, 65370, '\P{^_latin}', "");
Expect(0, 65371, '\p{_latin}', "");
Expect(1, 65371, '\p{^_latin}', "");
Expect(1, 65371, '\P{_latin}', "");
Expect(0, 65371, '\P{^_latin}', "");
Error('\p{/a/	IS_Latin}');
Error('\P{/a/	IS_Latin}');
Expect(1, 65370, '\p{islatin}', "");
Expect(0, 65370, '\p{^islatin}', "");
Expect(0, 65370, '\P{islatin}', "");
Expect(1, 65370, '\P{^islatin}', "");
Expect(0, 65371, '\p{islatin}', "");
Expect(1, 65371, '\p{^islatin}', "");
Expect(1, 65371, '\P{islatin}', "");
Expect(0, 65371, '\P{^islatin}', "");
Expect(1, 65370, '\p{IS_Latin}', "");
Expect(0, 65370, '\p{^IS_Latin}', "");
Expect(0, 65370, '\P{IS_Latin}', "");
Expect(1, 65370, '\P{^IS_Latin}', "");
Expect(0, 65371, '\p{IS_Latin}', "");
Expect(1, 65371, '\p{^IS_Latin}', "");
Expect(1, 65371, '\P{IS_Latin}', "");
Expect(0, 65371, '\P{^IS_Latin}', "");
Error('\p{/a/Latn}');
Error('\P{/a/Latn}');
Expect(1, 65370, '\p{latn}', "");
Expect(0, 65370, '\p{^latn}', "");
Expect(0, 65370, '\P{latn}', "");
Expect(1, 65370, '\P{^latn}', "");
Expect(0, 65371, '\p{latn}', "");
Expect(1, 65371, '\p{^latn}', "");
Expect(1, 65371, '\P{latn}', "");
Expect(0, 65371, '\P{^latn}', "");
Expect(1, 65370, '\p{	Latn}', "");
Expect(0, 65370, '\p{^	Latn}', "");
Expect(0, 65370, '\P{	Latn}', "");
Expect(1, 65370, '\P{^	Latn}', "");
Expect(0, 65371, '\p{	Latn}', "");
Expect(1, 65371, '\p{^	Latn}', "");
Expect(1, 65371, '\P{	Latn}', "");
Expect(0, 65371, '\P{^	Latn}', "");
Error('\p{_Is_LATN/a/}');
Error('\P{_Is_LATN/a/}');
Expect(1, 65370, '\p{islatn}', "");
Expect(0, 65370, '\p{^islatn}', "");
Expect(0, 65370, '\P{islatn}', "");
Expect(1, 65370, '\P{^islatn}', "");
Expect(0, 65371, '\p{islatn}', "");
Expect(1, 65371, '\p{^islatn}', "");
Expect(1, 65371, '\P{islatn}', "");
Expect(0, 65371, '\P{^islatn}', "");
Expect(1, 65370, '\p{--Is_Latn}', "");
Expect(0, 65370, '\p{^--Is_Latn}', "");
Expect(0, 65370, '\P{--Is_Latn}', "");
Expect(1, 65370, '\P{^--Is_Latn}', "");
Expect(0, 65371, '\p{--Is_Latn}', "");
Expect(1, 65371, '\p{^--Is_Latn}', "");
Expect(1, 65371, '\P{--Is_Latn}', "");
Expect(0, 65371, '\P{^--Is_Latn}', "");
Error('\p{- Latin_1_SUPPLEMENT:=}');
Error('\P{- Latin_1_SUPPLEMENT:=}');
Expect(1, 255, '\p{latin1supplement}', "");
Expect(0, 255, '\p{^latin1supplement}', "");
Expect(0, 255, '\P{latin1supplement}', "");
Expect(1, 255, '\P{^latin1supplement}', "");
Expect(0, 256, '\p{latin1supplement}', "");
Expect(1, 256, '\p{^latin1supplement}', "");
Expect(1, 256, '\P{latin1supplement}', "");
Expect(0, 256, '\P{^latin1supplement}', "");
Expect(1, 255, '\p{ -LATIN_1_supplement}', "");
Expect(0, 255, '\p{^ -LATIN_1_supplement}', "");
Expect(0, 255, '\P{ -LATIN_1_supplement}', "");
Expect(1, 255, '\P{^ -LATIN_1_supplement}', "");
Expect(0, 256, '\p{ -LATIN_1_supplement}', "");
Expect(1, 256, '\p{^ -LATIN_1_supplement}', "");
Expect(1, 256, '\P{ -LATIN_1_supplement}', "");
Expect(0, 256, '\P{^ -LATIN_1_supplement}', "");
Error('\p{	is_Latin_1_supplement:=}');
Error('\P{	is_Latin_1_supplement:=}');
Expect(1, 255, '\p{islatin1supplement}', "");
Expect(0, 255, '\p{^islatin1supplement}', "");
Expect(0, 255, '\P{islatin1supplement}', "");
Expect(1, 255, '\P{^islatin1supplement}', "");
Expect(0, 256, '\p{islatin1supplement}', "");
Expect(1, 256, '\p{^islatin1supplement}', "");
Expect(1, 256, '\P{islatin1supplement}', "");
Expect(0, 256, '\P{^islatin1supplement}', "");
Expect(1, 255, '\p{  Is_Latin_1_Supplement}', "");
Expect(0, 255, '\p{^  Is_Latin_1_Supplement}', "");
Expect(0, 255, '\P{  Is_Latin_1_Supplement}', "");
Expect(1, 255, '\P{^  Is_Latin_1_Supplement}', "");
Expect(0, 256, '\p{  Is_Latin_1_Supplement}', "");
Expect(1, 256, '\p{^  Is_Latin_1_Supplement}', "");
Expect(1, 256, '\P{  Is_Latin_1_Supplement}', "");
Expect(0, 256, '\P{^  Is_Latin_1_Supplement}', "");
Error('\p{-IN_Latin_1_SUPPLEMENT:=}');
Error('\P{-IN_Latin_1_SUPPLEMENT:=}');
Expect(1, 255, '\p{inlatin1supplement}', "");
Expect(0, 255, '\p{^inlatin1supplement}', "");
Expect(0, 255, '\P{inlatin1supplement}', "");
Expect(1, 255, '\P{^inlatin1supplement}', "");
Expect(0, 256, '\p{inlatin1supplement}', "");
Expect(1, 256, '\p{^inlatin1supplement}', "");
Expect(1, 256, '\P{inlatin1supplement}', "");
Expect(0, 256, '\P{^inlatin1supplement}', "");
Expect(1, 255, '\p{ -In_LATIN_1_Supplement}', "");
Expect(0, 255, '\p{^ -In_LATIN_1_Supplement}', "");
Expect(0, 255, '\P{ -In_LATIN_1_Supplement}', "");
Expect(1, 255, '\P{^ -In_LATIN_1_Supplement}', "");
Expect(0, 256, '\p{ -In_LATIN_1_Supplement}', "");
Expect(1, 256, '\p{^ -In_LATIN_1_Supplement}', "");
Expect(1, 256, '\P{ -In_LATIN_1_Supplement}', "");
Expect(0, 256, '\P{^ -In_LATIN_1_Supplement}', "");
Error('\p{-:=LATIN_1_Sup}');
Error('\P{-:=LATIN_1_Sup}');
Expect(1, 255, '\p{latin1sup}', "");
Expect(0, 255, '\p{^latin1sup}', "");
Expect(0, 255, '\P{latin1sup}', "");
Expect(1, 255, '\P{^latin1sup}', "");
Expect(0, 256, '\p{latin1sup}', "");
Expect(1, 256, '\p{^latin1sup}', "");
Expect(1, 256, '\P{latin1sup}', "");
Expect(0, 256, '\P{^latin1sup}', "");
Expect(1, 255, '\p{-_Latin_1_sup}', "");
Expect(0, 255, '\p{^-_Latin_1_sup}', "");
Expect(0, 255, '\P{-_Latin_1_sup}', "");
Expect(1, 255, '\P{^-_Latin_1_sup}', "");
Expect(0, 256, '\p{-_Latin_1_sup}', "");
Expect(1, 256, '\p{^-_Latin_1_sup}', "");
Expect(1, 256, '\P{-_Latin_1_sup}', "");
Expect(0, 256, '\P{^-_Latin_1_sup}', "");
Error('\p{/a/	_is_latin_1_Sup}');
Error('\P{/a/	_is_latin_1_Sup}');
Expect(1, 255, '\p{islatin1sup}', "");
Expect(0, 255, '\p{^islatin1sup}', "");
Expect(0, 255, '\P{islatin1sup}', "");
Expect(1, 255, '\P{^islatin1sup}', "");
Expect(0, 256, '\p{islatin1sup}', "");
Expect(1, 256, '\p{^islatin1sup}', "");
Expect(1, 256, '\P{islatin1sup}', "");
Expect(0, 256, '\P{^islatin1sup}', "");
Expect(1, 255, '\p{	-Is_Latin_1_Sup}', "");
Expect(0, 255, '\p{^	-Is_Latin_1_Sup}', "");
Expect(0, 255, '\P{	-Is_Latin_1_Sup}', "");
Expect(1, 255, '\P{^	-Is_Latin_1_Sup}', "");
Expect(0, 256, '\p{	-Is_Latin_1_Sup}', "");
Expect(1, 256, '\p{^	-Is_Latin_1_Sup}', "");
Expect(1, 256, '\P{	-Is_Latin_1_Sup}', "");
Expect(0, 256, '\P{^	-Is_Latin_1_Sup}', "");
Error('\p{_	IN_LATIN_1_Sup/a/}');
Error('\P{_	IN_LATIN_1_Sup/a/}');
Expect(1, 255, '\p{inlatin1sup}', "");
Expect(0, 255, '\p{^inlatin1sup}', "");
Expect(0, 255, '\P{inlatin1sup}', "");
Expect(1, 255, '\P{^inlatin1sup}', "");
Expect(0, 256, '\p{inlatin1sup}', "");
Expect(1, 256, '\p{^inlatin1sup}', "");
Expect(1, 256, '\P{inlatin1sup}', "");
Expect(0, 256, '\P{^inlatin1sup}', "");
Expect(1, 255, '\p{- In_LATIN_1_SUP}', "");
Expect(0, 255, '\p{^- In_LATIN_1_SUP}', "");
Expect(0, 255, '\P{- In_LATIN_1_SUP}', "");
Expect(1, 255, '\P{^- In_LATIN_1_SUP}', "");
Expect(0, 256, '\p{- In_LATIN_1_SUP}', "");
Expect(1, 256, '\p{^- In_LATIN_1_SUP}', "");
Expect(1, 256, '\P{- In_LATIN_1_SUP}', "");
Expect(0, 256, '\P{^- In_LATIN_1_SUP}', "");
Error('\p{:=Latin_1}');
Error('\P{:=Latin_1}');
Expect(1, 255, '\p{latin1}', "");
Expect(0, 255, '\p{^latin1}', "");
Expect(0, 255, '\P{latin1}', "");
Expect(1, 255, '\P{^latin1}', "");
Expect(0, 256, '\p{latin1}', "");
Expect(1, 256, '\p{^latin1}', "");
Expect(1, 256, '\P{latin1}', "");
Expect(0, 256, '\P{^latin1}', "");
Expect(1, 255, '\p{	-Latin_1}', "");
Expect(0, 255, '\p{^	-Latin_1}', "");
Expect(0, 255, '\P{	-Latin_1}', "");
Expect(1, 255, '\P{^	-Latin_1}', "");
Expect(0, 256, '\p{	-Latin_1}', "");
Expect(1, 256, '\p{^	-Latin_1}', "");
Expect(1, 256, '\P{	-Latin_1}', "");
Expect(0, 256, '\P{^	-Latin_1}', "");
Error('\p{	/a/IS_Latin_1}');
Error('\P{	/a/IS_Latin_1}');
Expect(1, 255, '\p{islatin1}', "");
Expect(0, 255, '\p{^islatin1}', "");
Expect(0, 255, '\P{islatin1}', "");
Expect(1, 255, '\P{^islatin1}', "");
Expect(0, 256, '\p{islatin1}', "");
Expect(1, 256, '\p{^islatin1}', "");
Expect(1, 256, '\P{islatin1}', "");
Expect(0, 256, '\P{^islatin1}', "");
Expect(1, 255, '\p{ _Is_Latin_1}', "");
Expect(0, 255, '\p{^ _Is_Latin_1}', "");
Expect(0, 255, '\P{ _Is_Latin_1}', "");
Expect(1, 255, '\P{^ _Is_Latin_1}', "");
Expect(0, 256, '\p{ _Is_Latin_1}', "");
Expect(1, 256, '\p{^ _Is_Latin_1}', "");
Expect(1, 256, '\P{ _Is_Latin_1}', "");
Expect(0, 256, '\P{^ _Is_Latin_1}', "");
Error('\p{:= -In_latin_1}');
Error('\P{:= -In_latin_1}');
Expect(1, 255, '\p{inlatin1}', "");
Expect(0, 255, '\p{^inlatin1}', "");
Expect(0, 255, '\P{inlatin1}', "");
Expect(1, 255, '\P{^inlatin1}', "");
Expect(0, 256, '\p{inlatin1}', "");
Expect(1, 256, '\p{^inlatin1}', "");
Expect(1, 256, '\P{inlatin1}', "");
Expect(0, 256, '\P{^inlatin1}', "");
Expect(1, 255, '\p{_In_Latin_1}', "");
Expect(0, 255, '\p{^_In_Latin_1}', "");
Expect(0, 255, '\P{_In_Latin_1}', "");
Expect(1, 255, '\P{^_In_Latin_1}', "");
Expect(0, 256, '\p{_In_Latin_1}', "");
Expect(1, 256, '\p{^_In_Latin_1}', "");
Expect(1, 256, '\P{_In_Latin_1}', "");
Expect(0, 256, '\P{^_In_Latin_1}', "");
Error('\p{/a/		Latin_Extended_A}');
Error('\P{/a/		Latin_Extended_A}');
Expect(1, 383, '\p{latinextendeda}', "");
Expect(0, 383, '\p{^latinextendeda}', "");
Expect(0, 383, '\P{latinextendeda}', "");
Expect(1, 383, '\P{^latinextendeda}', "");
Expect(0, 384, '\p{latinextendeda}', "");
Expect(1, 384, '\p{^latinextendeda}', "");
Expect(1, 384, '\P{latinextendeda}', "");
Expect(0, 384, '\P{^latinextendeda}', "");
Expect(1, 383, '\p{	Latin_EXTENDED_a}', "");
Expect(0, 383, '\p{^	Latin_EXTENDED_a}', "");
Expect(0, 383, '\P{	Latin_EXTENDED_a}', "");
Expect(1, 383, '\P{^	Latin_EXTENDED_a}', "");
Expect(0, 384, '\p{	Latin_EXTENDED_a}', "");
Expect(1, 384, '\p{^	Latin_EXTENDED_a}', "");
Expect(1, 384, '\P{	Latin_EXTENDED_a}', "");
Expect(0, 384, '\P{^	Latin_EXTENDED_a}', "");
Error('\p{/a/-	Is_Latin_extended_a}');
Error('\P{/a/-	Is_Latin_extended_a}');
Expect(1, 383, '\p{islatinextendeda}', "");
Expect(0, 383, '\p{^islatinextendeda}', "");
Expect(0, 383, '\P{islatinextendeda}', "");
Expect(1, 383, '\P{^islatinextendeda}', "");
Expect(0, 384, '\p{islatinextendeda}', "");
Expect(1, 384, '\p{^islatinextendeda}', "");
Expect(1, 384, '\P{islatinextendeda}', "");
Expect(0, 384, '\P{^islatinextendeda}', "");
Expect(1, 383, '\p{-Is_Latin_Extended_A}', "");
Expect(0, 383, '\p{^-Is_Latin_Extended_A}', "");
Expect(0, 383, '\P{-Is_Latin_Extended_A}', "");
Expect(1, 383, '\P{^-Is_Latin_Extended_A}', "");
Expect(0, 384, '\p{-Is_Latin_Extended_A}', "");
Expect(1, 384, '\p{^-Is_Latin_Extended_A}', "");
Expect(1, 384, '\P{-Is_Latin_Extended_A}', "");
Expect(0, 384, '\P{^-Is_Latin_Extended_A}', "");
Error('\p{/a/ IN_Latin_extended_a}');
Error('\P{/a/ IN_Latin_extended_a}');
Expect(1, 383, '\p{inlatinextendeda}', "");
Expect(0, 383, '\p{^inlatinextendeda}', "");
Expect(0, 383, '\P{inlatinextendeda}', "");
Expect(1, 383, '\P{^inlatinextendeda}', "");
Expect(0, 384, '\p{inlatinextendeda}', "");
Expect(1, 384, '\p{^inlatinextendeda}', "");
Expect(1, 384, '\P{inlatinextendeda}', "");
Expect(0, 384, '\P{^inlatinextendeda}', "");
Expect(1, 383, '\p{_-In_Latin_extended_a}', "");
Expect(0, 383, '\p{^_-In_Latin_extended_a}', "");
Expect(0, 383, '\P{_-In_Latin_extended_a}', "");
Expect(1, 383, '\P{^_-In_Latin_extended_a}', "");
Expect(0, 384, '\p{_-In_Latin_extended_a}', "");
Expect(1, 384, '\p{^_-In_Latin_extended_a}', "");
Expect(1, 384, '\P{_-In_Latin_extended_a}', "");
Expect(0, 384, '\P{^_-In_Latin_extended_a}', "");
Error('\p{:=LATIN_EXT_A}');
Error('\P{:=LATIN_EXT_A}');
Expect(1, 383, '\p{latinexta}', "");
Expect(0, 383, '\p{^latinexta}', "");
Expect(0, 383, '\P{latinexta}', "");
Expect(1, 383, '\P{^latinexta}', "");
Expect(0, 384, '\p{latinexta}', "");
Expect(1, 384, '\p{^latinexta}', "");
Expect(1, 384, '\P{latinexta}', "");
Expect(0, 384, '\P{^latinexta}', "");
Expect(1, 383, '\p{ Latin_Ext_A}', "");
Expect(0, 383, '\p{^ Latin_Ext_A}', "");
Expect(0, 383, '\P{ Latin_Ext_A}', "");
Expect(1, 383, '\P{^ Latin_Ext_A}', "");
Expect(0, 384, '\p{ Latin_Ext_A}', "");
Expect(1, 384, '\p{^ Latin_Ext_A}', "");
Expect(1, 384, '\P{ Latin_Ext_A}', "");
Expect(0, 384, '\P{^ Latin_Ext_A}', "");
Error('\p{		IS_Latin_Ext_A/a/}');
Error('\P{		IS_Latin_Ext_A/a/}');
Expect(1, 383, '\p{islatinexta}', "");
Expect(0, 383, '\p{^islatinexta}', "");
Expect(0, 383, '\P{islatinexta}', "");
Expect(1, 383, '\P{^islatinexta}', "");
Expect(0, 384, '\p{islatinexta}', "");
Expect(1, 384, '\p{^islatinexta}', "");
Expect(1, 384, '\P{islatinexta}', "");
Expect(0, 384, '\P{^islatinexta}', "");
Expect(1, 383, '\p{-Is_LATIN_EXT_A}', "");
Expect(0, 383, '\p{^-Is_LATIN_EXT_A}', "");
Expect(0, 383, '\P{-Is_LATIN_EXT_A}', "");
Expect(1, 383, '\P{^-Is_LATIN_EXT_A}', "");
Expect(0, 384, '\p{-Is_LATIN_EXT_A}', "");
Expect(1, 384, '\p{^-Is_LATIN_EXT_A}', "");
Expect(1, 384, '\P{-Is_LATIN_EXT_A}', "");
Expect(0, 384, '\P{^-Is_LATIN_EXT_A}', "");
Error('\p{:=  In_LATIN_Ext_a}');
Error('\P{:=  In_LATIN_Ext_a}');
Expect(1, 383, '\p{inlatinexta}', "");
Expect(0, 383, '\p{^inlatinexta}', "");
Expect(0, 383, '\P{inlatinexta}', "");
Expect(1, 383, '\P{^inlatinexta}', "");
Expect(0, 384, '\p{inlatinexta}', "");
Expect(1, 384, '\p{^inlatinexta}', "");
Expect(1, 384, '\P{inlatinexta}', "");
Expect(0, 384, '\P{^inlatinexta}', "");
Expect(1, 383, '\p{-_IN_Latin_ext_A}', "");
Expect(0, 383, '\p{^-_IN_Latin_ext_A}', "");
Expect(0, 383, '\P{-_IN_Latin_ext_A}', "");
Expect(1, 383, '\P{^-_IN_Latin_ext_A}', "");
Expect(0, 384, '\p{-_IN_Latin_ext_A}', "");
Expect(1, 384, '\p{^-_IN_Latin_ext_A}', "");
Expect(1, 384, '\P{-_IN_Latin_ext_A}', "");
Expect(0, 384, '\P{^-_IN_Latin_ext_A}', "");
Error('\p{:=_	Latin_EXTENDED_additional}');
Error('\P{:=_	Latin_EXTENDED_additional}');
Expect(1, 7935, '\p{latinextendedadditional}', "");
Expect(0, 7935, '\p{^latinextendedadditional}', "");
Expect(0, 7935, '\P{latinextendedadditional}', "");
Expect(1, 7935, '\P{^latinextendedadditional}', "");
Expect(0, 7936, '\p{latinextendedadditional}', "");
Expect(1, 7936, '\p{^latinextendedadditional}', "");
Expect(1, 7936, '\P{latinextendedadditional}', "");
Expect(0, 7936, '\P{^latinextendedadditional}', "");
Expect(1, 7935, '\p{-Latin_Extended_additional}', "");
Expect(0, 7935, '\p{^-Latin_Extended_additional}', "");
Expect(0, 7935, '\P{-Latin_Extended_additional}', "");
Expect(1, 7935, '\P{^-Latin_Extended_additional}', "");
Expect(0, 7936, '\p{-Latin_Extended_additional}', "");
Expect(1, 7936, '\p{^-Latin_Extended_additional}', "");
Expect(1, 7936, '\P{-Latin_Extended_additional}', "");
Expect(0, 7936, '\P{^-Latin_Extended_additional}', "");
Error('\p{ /a/Is_Latin_extended_additional}');
Error('\P{ /a/Is_Latin_extended_additional}');
Expect(1, 7935, '\p{islatinextendedadditional}', "");
Expect(0, 7935, '\p{^islatinextendedadditional}', "");
Expect(0, 7935, '\P{islatinextendedadditional}', "");
Expect(1, 7935, '\P{^islatinextendedadditional}', "");
Expect(0, 7936, '\p{islatinextendedadditional}', "");
Expect(1, 7936, '\p{^islatinextendedadditional}', "");
Expect(1, 7936, '\P{islatinextendedadditional}', "");
Expect(0, 7936, '\P{^islatinextendedadditional}', "");
Expect(1, 7935, '\p{	-IS_Latin_Extended_additional}', "");
Expect(0, 7935, '\p{^	-IS_Latin_Extended_additional}', "");
Expect(0, 7935, '\P{	-IS_Latin_Extended_additional}', "");
Expect(1, 7935, '\P{^	-IS_Latin_Extended_additional}', "");
Expect(0, 7936, '\p{	-IS_Latin_Extended_additional}', "");
Expect(1, 7936, '\p{^	-IS_Latin_Extended_additional}', "");
Expect(1, 7936, '\P{	-IS_Latin_Extended_additional}', "");
Expect(0, 7936, '\P{^	-IS_Latin_Extended_additional}', "");
Error('\p{-/a/IN_Latin_EXTENDED_ADDITIONAL}');
Error('\P{-/a/IN_Latin_EXTENDED_ADDITIONAL}');
Expect(1, 7935, '\p{inlatinextendedadditional}', "");
Expect(0, 7935, '\p{^inlatinextendedadditional}', "");
Expect(0, 7935, '\P{inlatinextendedadditional}', "");
Expect(1, 7935, '\P{^inlatinextendedadditional}', "");
Expect(0, 7936, '\p{inlatinextendedadditional}', "");
Expect(1, 7936, '\p{^inlatinextendedadditional}', "");
Expect(1, 7936, '\P{inlatinextendedadditional}', "");
Expect(0, 7936, '\P{^inlatinextendedadditional}', "");
Expect(1, 7935, '\p{__In_Latin_EXTENDED_additional}', "");
Expect(0, 7935, '\p{^__In_Latin_EXTENDED_additional}', "");
Expect(0, 7935, '\P{__In_Latin_EXTENDED_additional}', "");
Expect(1, 7935, '\P{^__In_Latin_EXTENDED_additional}', "");
Expect(0, 7936, '\p{__In_Latin_EXTENDED_additional}', "");
Expect(1, 7936, '\p{^__In_Latin_EXTENDED_additional}', "");
Expect(1, 7936, '\P{__In_Latin_EXTENDED_additional}', "");
Expect(0, 7936, '\P{^__In_Latin_EXTENDED_additional}', "");
Error('\p{	/a/LATIN_ext_Additional}');
Error('\P{	/a/LATIN_ext_Additional}');
Expect(1, 7935, '\p{latinextadditional}', "");
Expect(0, 7935, '\p{^latinextadditional}', "");
Expect(0, 7935, '\P{latinextadditional}', "");
Expect(1, 7935, '\P{^latinextadditional}', "");
Expect(0, 7936, '\p{latinextadditional}', "");
Expect(1, 7936, '\p{^latinextadditional}', "");
Expect(1, 7936, '\P{latinextadditional}', "");
Expect(0, 7936, '\P{^latinextadditional}', "");
Expect(1, 7935, '\p{_ Latin_EXT_ADDITIONAL}', "");
Expect(0, 7935, '\p{^_ Latin_EXT_ADDITIONAL}', "");
Expect(0, 7935, '\P{_ Latin_EXT_ADDITIONAL}', "");
Expect(1, 7935, '\P{^_ Latin_EXT_ADDITIONAL}', "");
Expect(0, 7936, '\p{_ Latin_EXT_ADDITIONAL}', "");
Expect(1, 7936, '\p{^_ Latin_EXT_ADDITIONAL}', "");
Expect(1, 7936, '\P{_ Latin_EXT_ADDITIONAL}', "");
Expect(0, 7936, '\P{^_ Latin_EXT_ADDITIONAL}', "");
Error('\p{:=--is_Latin_Ext_additional}');
Error('\P{:=--is_Latin_Ext_additional}');
Expect(1, 7935, '\p{islatinextadditional}', "");
Expect(0, 7935, '\p{^islatinextadditional}', "");
Expect(0, 7935, '\P{islatinextadditional}', "");
Expect(1, 7935, '\P{^islatinextadditional}', "");
Expect(0, 7936, '\p{islatinextadditional}', "");
Expect(1, 7936, '\p{^islatinextadditional}', "");
Expect(1, 7936, '\P{islatinextadditional}', "");
Expect(0, 7936, '\P{^islatinextadditional}', "");
Expect(1, 7935, '\p{_ is_latin_ext_additional}', "");
Expect(0, 7935, '\p{^_ is_latin_ext_additional}', "");
Expect(0, 7935, '\P{_ is_latin_ext_additional}', "");
Expect(1, 7935, '\P{^_ is_latin_ext_additional}', "");
Expect(0, 7936, '\p{_ is_latin_ext_additional}', "");
Expect(1, 7936, '\p{^_ is_latin_ext_additional}', "");
Expect(1, 7936, '\P{_ is_latin_ext_additional}', "");
Expect(0, 7936, '\P{^_ is_latin_ext_additional}', "");
Error('\p{	IN_Latin_ext_ADDITIONAL:=}');
Error('\P{	IN_Latin_ext_ADDITIONAL:=}');
Expect(1, 7935, '\p{inlatinextadditional}', "");
Expect(0, 7935, '\p{^inlatinextadditional}', "");
Expect(0, 7935, '\P{inlatinextadditional}', "");
Expect(1, 7935, '\P{^inlatinextadditional}', "");
Expect(0, 7936, '\p{inlatinextadditional}', "");
Expect(1, 7936, '\p{^inlatinextadditional}', "");
Expect(1, 7936, '\P{inlatinextadditional}', "");
Expect(0, 7936, '\P{^inlatinextadditional}', "");
Expect(1, 7935, '\p{-	In_Latin_EXT_Additional}', "");
Expect(0, 7935, '\p{^-	In_Latin_EXT_Additional}', "");
Expect(0, 7935, '\P{-	In_Latin_EXT_Additional}', "");
Expect(1, 7935, '\P{^-	In_Latin_EXT_Additional}', "");
Expect(0, 7936, '\p{-	In_Latin_EXT_Additional}', "");
Expect(1, 7936, '\p{^-	In_Latin_EXT_Additional}', "");
Expect(1, 7936, '\P{-	In_Latin_EXT_Additional}', "");
Expect(0, 7936, '\P{^-	In_Latin_EXT_Additional}', "");
Error('\p{-/a/LATIN_Extended_B}');
Error('\P{-/a/LATIN_Extended_B}');
Expect(1, 591, '\p{latinextendedb}', "");
Expect(0, 591, '\p{^latinextendedb}', "");
Expect(0, 591, '\P{latinextendedb}', "");
Expect(1, 591, '\P{^latinextendedb}', "");
Expect(0, 592, '\p{latinextendedb}', "");
Expect(1, 592, '\p{^latinextendedb}', "");
Expect(1, 592, '\P{latinextendedb}', "");
Expect(0, 592, '\P{^latinextendedb}', "");
Expect(1, 591, '\p{ LATIN_Extended_B}', "");
Expect(0, 591, '\p{^ LATIN_Extended_B}', "");
Expect(0, 591, '\P{ LATIN_Extended_B}', "");
Expect(1, 591, '\P{^ LATIN_Extended_B}', "");
Expect(0, 592, '\p{ LATIN_Extended_B}', "");
Expect(1, 592, '\p{^ LATIN_Extended_B}', "");
Expect(1, 592, '\P{ LATIN_Extended_B}', "");
Expect(0, 592, '\P{^ LATIN_Extended_B}', "");
Error('\p{_/a/IS_Latin_Extended_B}');
Error('\P{_/a/IS_Latin_Extended_B}');
Expect(1, 591, '\p{islatinextendedb}', "");
Expect(0, 591, '\p{^islatinextendedb}', "");
Expect(0, 591, '\P{islatinextendedb}', "");
Expect(1, 591, '\P{^islatinextendedb}', "");
Expect(0, 592, '\p{islatinextendedb}', "");
Expect(1, 592, '\p{^islatinextendedb}', "");
Expect(1, 592, '\P{islatinextendedb}', "");
Expect(0, 592, '\P{^islatinextendedb}', "");
Expect(1, 591, '\p{_	IS_Latin_Extended_b}', "");
Expect(0, 591, '\p{^_	IS_Latin_Extended_b}', "");
Expect(0, 591, '\P{_	IS_Latin_Extended_b}', "");
Expect(1, 591, '\P{^_	IS_Latin_Extended_b}', "");
Expect(0, 592, '\p{_	IS_Latin_Extended_b}', "");
Expect(1, 592, '\p{^_	IS_Latin_Extended_b}', "");
Expect(1, 592, '\P{_	IS_Latin_Extended_b}', "");
Expect(0, 592, '\P{^_	IS_Latin_Extended_b}', "");
Error('\p{/a/_in_Latin_Extended_B}');
Error('\P{/a/_in_Latin_Extended_B}');
Expect(1, 591, '\p{inlatinextendedb}', "");
Expect(0, 591, '\p{^inlatinextendedb}', "");
Expect(0, 591, '\P{inlatinextendedb}', "");
Expect(1, 591, '\P{^inlatinextendedb}', "");
Expect(0, 592, '\p{inlatinextendedb}', "");
Expect(1, 592, '\p{^inlatinextendedb}', "");
Expect(1, 592, '\P{inlatinextendedb}', "");
Expect(0, 592, '\P{^inlatinextendedb}', "");
Expect(1, 591, '\p{-in_latin_EXTENDED_B}', "");
Expect(0, 591, '\p{^-in_latin_EXTENDED_B}', "");
Expect(0, 591, '\P{-in_latin_EXTENDED_B}', "");
Expect(1, 591, '\P{^-in_latin_EXTENDED_B}', "");
Expect(0, 592, '\p{-in_latin_EXTENDED_B}', "");
Expect(1, 592, '\p{^-in_latin_EXTENDED_B}', "");
Expect(1, 592, '\P{-in_latin_EXTENDED_B}', "");
Expect(0, 592, '\P{^-in_latin_EXTENDED_B}', "");
Error('\p{_ latin_ext_B:=}');
Error('\P{_ latin_ext_B:=}');
Expect(1, 591, '\p{latinextb}', "");
Expect(0, 591, '\p{^latinextb}', "");
Expect(0, 591, '\P{latinextb}', "");
Expect(1, 591, '\P{^latinextb}', "");
Expect(0, 592, '\p{latinextb}', "");
Expect(1, 592, '\p{^latinextb}', "");
Expect(1, 592, '\P{latinextb}', "");
Expect(0, 592, '\P{^latinextb}', "");
Expect(1, 591, '\p{ Latin_Ext_B}', "");
Expect(0, 591, '\p{^ Latin_Ext_B}', "");
Expect(0, 591, '\P{ Latin_Ext_B}', "");
Expect(1, 591, '\P{^ Latin_Ext_B}', "");
Expect(0, 592, '\p{ Latin_Ext_B}', "");
Expect(1, 592, '\p{^ Latin_Ext_B}', "");
Expect(1, 592, '\P{ Latin_Ext_B}', "");
Expect(0, 592, '\P{^ Latin_Ext_B}', "");
Error('\p{/a/_Is_Latin_Ext_B}');
Error('\P{/a/_Is_Latin_Ext_B}');
Expect(1, 591, '\p{islatinextb}', "");
Expect(0, 591, '\p{^islatinextb}', "");
Expect(0, 591, '\P{islatinextb}', "");
Expect(1, 591, '\P{^islatinextb}', "");
Expect(0, 592, '\p{islatinextb}', "");
Expect(1, 592, '\p{^islatinextb}', "");
Expect(1, 592, '\P{islatinextb}', "");
Expect(0, 592, '\P{^islatinextb}', "");
Expect(1, 591, '\p{_ IS_Latin_ext_B}', "");
Expect(0, 591, '\p{^_ IS_Latin_ext_B}', "");
Expect(0, 591, '\P{_ IS_Latin_ext_B}', "");
Expect(1, 591, '\P{^_ IS_Latin_ext_B}', "");
Expect(0, 592, '\p{_ IS_Latin_ext_B}', "");
Expect(1, 592, '\p{^_ IS_Latin_ext_B}', "");
Expect(1, 592, '\P{_ IS_Latin_ext_B}', "");
Expect(0, 592, '\P{^_ IS_Latin_ext_B}', "");
Error('\p{ -IN_Latin_Ext_B:=}');
Error('\P{ -IN_Latin_Ext_B:=}');
Expect(1, 591, '\p{inlatinextb}', "");
Expect(0, 591, '\p{^inlatinextb}', "");
Expect(0, 591, '\P{inlatinextb}', "");
Expect(1, 591, '\P{^inlatinextb}', "");
Expect(0, 592, '\p{inlatinextb}', "");
Expect(1, 592, '\p{^inlatinextb}', "");
Expect(1, 592, '\P{inlatinextb}', "");
Expect(0, 592, '\P{^inlatinextb}', "");
Expect(1, 591, '\p{		In_LATIN_Ext_b}', "");
Expect(0, 591, '\p{^		In_LATIN_Ext_b}', "");
Expect(0, 591, '\P{		In_LATIN_Ext_b}', "");
Expect(1, 591, '\P{^		In_LATIN_Ext_b}', "");
Expect(0, 592, '\p{		In_LATIN_Ext_b}', "");
Expect(1, 592, '\p{^		In_LATIN_Ext_b}', "");
Expect(1, 592, '\P{		In_LATIN_Ext_b}', "");
Expect(0, 592, '\P{^		In_LATIN_Ext_b}', "");
Error('\p{:=	 Latin_Extended_C}');
Error('\P{:=	 Latin_Extended_C}');
Expect(1, 11391, '\p{latinextendedc}', "");
Expect(0, 11391, '\p{^latinextendedc}', "");
Expect(0, 11391, '\P{latinextendedc}', "");
Expect(1, 11391, '\P{^latinextendedc}', "");
Expect(0, 11392, '\p{latinextendedc}', "");
Expect(1, 11392, '\p{^latinextendedc}', "");
Expect(1, 11392, '\P{latinextendedc}', "");
Expect(0, 11392, '\P{^latinextendedc}', "");
Expect(1, 11391, '\p{_Latin_Extended_C}', "");
Expect(0, 11391, '\p{^_Latin_Extended_C}', "");
Expect(0, 11391, '\P{_Latin_Extended_C}', "");
Expect(1, 11391, '\P{^_Latin_Extended_C}', "");
Expect(0, 11392, '\p{_Latin_Extended_C}', "");
Expect(1, 11392, '\p{^_Latin_Extended_C}', "");
Expect(1, 11392, '\P{_Latin_Extended_C}', "");
Expect(0, 11392, '\P{^_Latin_Extended_C}', "");
Error('\p{-is_Latin_Extended_C/a/}');
Error('\P{-is_Latin_Extended_C/a/}');
Expect(1, 11391, '\p{islatinextendedc}', "");
Expect(0, 11391, '\p{^islatinextendedc}', "");
Expect(0, 11391, '\P{islatinextendedc}', "");
Expect(1, 11391, '\P{^islatinextendedc}', "");
Expect(0, 11392, '\p{islatinextendedc}', "");
Expect(1, 11392, '\p{^islatinextendedc}', "");
Expect(1, 11392, '\P{islatinextendedc}', "");
Expect(0, 11392, '\P{^islatinextendedc}', "");
Expect(1, 11391, '\p{- Is_Latin_Extended_C}', "");
Expect(0, 11391, '\p{^- Is_Latin_Extended_C}', "");
Expect(0, 11391, '\P{- Is_Latin_Extended_C}', "");
Expect(1, 11391, '\P{^- Is_Latin_Extended_C}', "");
Expect(0, 11392, '\p{- Is_Latin_Extended_C}', "");
Expect(1, 11392, '\p{^- Is_Latin_Extended_C}', "");
Expect(1, 11392, '\P{- Is_Latin_Extended_C}', "");
Expect(0, 11392, '\P{^- Is_Latin_Extended_C}', "");
Error('\p{ -In_Latin_extended_c:=}');
Error('\P{ -In_Latin_extended_c:=}');
Expect(1, 11391, '\p{inlatinextendedc}', "");
Expect(0, 11391, '\p{^inlatinextendedc}', "");
Expect(0, 11391, '\P{inlatinextendedc}', "");
Expect(1, 11391, '\P{^inlatinextendedc}', "");
Expect(0, 11392, '\p{inlatinextendedc}', "");
Expect(1, 11392, '\p{^inlatinextendedc}', "");
Expect(1, 11392, '\P{inlatinextendedc}', "");
Expect(0, 11392, '\P{^inlatinextendedc}', "");
Expect(1, 11391, '\p{  In_Latin_EXTENDED_C}', "");
Expect(0, 11391, '\p{^  In_Latin_EXTENDED_C}', "");
Expect(0, 11391, '\P{  In_Latin_EXTENDED_C}', "");
Expect(1, 11391, '\P{^  In_Latin_EXTENDED_C}', "");
Expect(0, 11392, '\p{  In_Latin_EXTENDED_C}', "");
Expect(1, 11392, '\p{^  In_Latin_EXTENDED_C}', "");
Expect(1, 11392, '\P{  In_Latin_EXTENDED_C}', "");
Expect(0, 11392, '\P{^  In_Latin_EXTENDED_C}', "");
Error('\p{/a/-latin_EXT_C}');
Error('\P{/a/-latin_EXT_C}');
Expect(1, 11391, '\p{latinextc}', "");
Expect(0, 11391, '\p{^latinextc}', "");
Expect(0, 11391, '\P{latinextc}', "");
Expect(1, 11391, '\P{^latinextc}', "");
Expect(0, 11392, '\p{latinextc}', "");
Expect(1, 11392, '\p{^latinextc}', "");
Expect(1, 11392, '\P{latinextc}', "");
Expect(0, 11392, '\P{^latinextc}', "");
Expect(1, 11391, '\p{	Latin_Ext_C}', "");
Expect(0, 11391, '\p{^	Latin_Ext_C}', "");
Expect(0, 11391, '\P{	Latin_Ext_C}', "");
Expect(1, 11391, '\P{^	Latin_Ext_C}', "");
Expect(0, 11392, '\p{	Latin_Ext_C}', "");
Expect(1, 11392, '\p{^	Latin_Ext_C}', "");
Expect(1, 11392, '\P{	Latin_Ext_C}', "");
Expect(0, 11392, '\P{^	Latin_Ext_C}', "");
Error('\p{-/a/IS_LATIN_EXT_c}');
Error('\P{-/a/IS_LATIN_EXT_c}');
Expect(1, 11391, '\p{islatinextc}', "");
Expect(0, 11391, '\p{^islatinextc}', "");
Expect(0, 11391, '\P{islatinextc}', "");
Expect(1, 11391, '\P{^islatinextc}', "");
Expect(0, 11392, '\p{islatinextc}', "");
Expect(1, 11392, '\p{^islatinextc}', "");
Expect(1, 11392, '\P{islatinextc}', "");
Expect(0, 11392, '\P{^islatinextc}', "");
Expect(1, 11391, '\p{	Is_Latin_Ext_c}', "");
Expect(0, 11391, '\p{^	Is_Latin_Ext_c}', "");
Expect(0, 11391, '\P{	Is_Latin_Ext_c}', "");
Expect(1, 11391, '\P{^	Is_Latin_Ext_c}', "");
Expect(0, 11392, '\p{	Is_Latin_Ext_c}', "");
Expect(1, 11392, '\p{^	Is_Latin_Ext_c}', "");
Expect(1, 11392, '\P{	Is_Latin_Ext_c}', "");
Expect(0, 11392, '\P{^	Is_Latin_Ext_c}', "");
Error('\p{	In_latin_Ext_C:=}');
Error('\P{	In_latin_Ext_C:=}');
Expect(1, 11391, '\p{inlatinextc}', "");
Expect(0, 11391, '\p{^inlatinextc}', "");
Expect(0, 11391, '\P{inlatinextc}', "");
Expect(1, 11391, '\P{^inlatinextc}', "");
Expect(0, 11392, '\p{inlatinextc}', "");
Expect(1, 11392, '\p{^inlatinextc}', "");
Expect(1, 11392, '\P{inlatinextc}', "");
Expect(0, 11392, '\P{^inlatinextc}', "");
Expect(1, 11391, '\p{	_IN_Latin_EXT_c}', "");
Expect(0, 11391, '\p{^	_IN_Latin_EXT_c}', "");
Expect(0, 11391, '\P{	_IN_Latin_EXT_c}', "");
Expect(1, 11391, '\P{^	_IN_Latin_EXT_c}', "");
Expect(0, 11392, '\p{	_IN_Latin_EXT_c}', "");
Expect(1, 11392, '\p{^	_IN_Latin_EXT_c}', "");
Expect(1, 11392, '\P{	_IN_Latin_EXT_c}', "");
Expect(0, 11392, '\P{^	_IN_Latin_EXT_c}', "");
Error('\p{-_latin_Extended_D:=}');
Error('\P{-_latin_Extended_D:=}');
Expect(1, 43007, '\p{latinextendedd}', "");
Expect(0, 43007, '\p{^latinextendedd}', "");
Expect(0, 43007, '\P{latinextendedd}', "");
Expect(1, 43007, '\P{^latinextendedd}', "");
Expect(0, 43008, '\p{latinextendedd}', "");
Expect(1, 43008, '\p{^latinextendedd}', "");
Expect(1, 43008, '\P{latinextendedd}', "");
Expect(0, 43008, '\P{^latinextendedd}', "");
Expect(1, 43007, '\p{-LATIN_Extended_D}', "");
Expect(0, 43007, '\p{^-LATIN_Extended_D}', "");
Expect(0, 43007, '\P{-LATIN_Extended_D}', "");
Expect(1, 43007, '\P{^-LATIN_Extended_D}', "");
Expect(0, 43008, '\p{-LATIN_Extended_D}', "");
Expect(1, 43008, '\p{^-LATIN_Extended_D}', "");
Expect(1, 43008, '\P{-LATIN_Extended_D}', "");
Expect(0, 43008, '\P{^-LATIN_Extended_D}', "");
Error('\p{-:=Is_Latin_Extended_D}');
Error('\P{-:=Is_Latin_Extended_D}');
Expect(1, 43007, '\p{islatinextendedd}', "");
Expect(0, 43007, '\p{^islatinextendedd}', "");
Expect(0, 43007, '\P{islatinextendedd}', "");
Expect(1, 43007, '\P{^islatinextendedd}', "");
Expect(0, 43008, '\p{islatinextendedd}', "");
Expect(1, 43008, '\p{^islatinextendedd}', "");
Expect(1, 43008, '\P{islatinextendedd}', "");
Expect(0, 43008, '\P{^islatinextendedd}', "");
Expect(1, 43007, '\p{	is_latin_Extended_D}', "");
Expect(0, 43007, '\p{^	is_latin_Extended_D}', "");
Expect(0, 43007, '\P{	is_latin_Extended_D}', "");
Expect(1, 43007, '\P{^	is_latin_Extended_D}', "");
Expect(0, 43008, '\p{	is_latin_Extended_D}', "");
Expect(1, 43008, '\p{^	is_latin_Extended_D}', "");
Expect(1, 43008, '\P{	is_latin_Extended_D}', "");
Expect(0, 43008, '\P{^	is_latin_Extended_D}', "");
Error('\p{/a/  In_Latin_Extended_D}');
Error('\P{/a/  In_Latin_Extended_D}');
Expect(1, 43007, '\p{inlatinextendedd}', "");
Expect(0, 43007, '\p{^inlatinextendedd}', "");
Expect(0, 43007, '\P{inlatinextendedd}', "");
Expect(1, 43007, '\P{^inlatinextendedd}', "");
Expect(0, 43008, '\p{inlatinextendedd}', "");
Expect(1, 43008, '\p{^inlatinextendedd}', "");
Expect(1, 43008, '\P{inlatinextendedd}', "");
Expect(0, 43008, '\P{^inlatinextendedd}', "");
Expect(1, 43007, '\p{_ In_Latin_Extended_D}', "");
Expect(0, 43007, '\p{^_ In_Latin_Extended_D}', "");
Expect(0, 43007, '\P{_ In_Latin_Extended_D}', "");
Expect(1, 43007, '\P{^_ In_Latin_Extended_D}', "");
Expect(0, 43008, '\p{_ In_Latin_Extended_D}', "");
Expect(1, 43008, '\p{^_ In_Latin_Extended_D}', "");
Expect(1, 43008, '\P{_ In_Latin_Extended_D}', "");
Expect(0, 43008, '\P{^_ In_Latin_Extended_D}', "");
Error('\p{-_Latin_EXT_D/a/}');
Error('\P{-_Latin_EXT_D/a/}');
Expect(1, 43007, '\p{latinextd}', "");
Expect(0, 43007, '\p{^latinextd}', "");
Expect(0, 43007, '\P{latinextd}', "");
Expect(1, 43007, '\P{^latinextd}', "");
Expect(0, 43008, '\p{latinextd}', "");
Expect(1, 43008, '\p{^latinextd}', "");
Expect(1, 43008, '\P{latinextd}', "");
Expect(0, 43008, '\P{^latinextd}', "");
Expect(1, 43007, '\p{  latin_ext_d}', "");
Expect(0, 43007, '\p{^  latin_ext_d}', "");
Expect(0, 43007, '\P{  latin_ext_d}', "");
Expect(1, 43007, '\P{^  latin_ext_d}', "");
Expect(0, 43008, '\p{  latin_ext_d}', "");
Expect(1, 43008, '\p{^  latin_ext_d}', "");
Expect(1, 43008, '\P{  latin_ext_d}', "");
Expect(0, 43008, '\P{^  latin_ext_d}', "");
Error('\p{/a/_is_latin_Ext_D}');
Error('\P{/a/_is_latin_Ext_D}');
Expect(1, 43007, '\p{islatinextd}', "");
Expect(0, 43007, '\p{^islatinextd}', "");
Expect(0, 43007, '\P{islatinextd}', "");
Expect(1, 43007, '\P{^islatinextd}', "");
Expect(0, 43008, '\p{islatinextd}', "");
Expect(1, 43008, '\p{^islatinextd}', "");
Expect(1, 43008, '\P{islatinextd}', "");
Expect(0, 43008, '\P{^islatinextd}', "");
Expect(1, 43007, '\p{	-is_Latin_EXT_D}', "");
Expect(0, 43007, '\p{^	-is_Latin_EXT_D}', "");
Expect(0, 43007, '\P{	-is_Latin_EXT_D}', "");
Expect(1, 43007, '\P{^	-is_Latin_EXT_D}', "");
Expect(0, 43008, '\p{	-is_Latin_EXT_D}', "");
Expect(1, 43008, '\p{^	-is_Latin_EXT_D}', "");
Expect(1, 43008, '\P{	-is_Latin_EXT_D}', "");
Expect(0, 43008, '\P{^	-is_Latin_EXT_D}', "");
Error('\p{-/a/In_Latin_ext_d}');
Error('\P{-/a/In_Latin_ext_d}');
Expect(1, 43007, '\p{inlatinextd}', "");
Expect(0, 43007, '\p{^inlatinextd}', "");
Expect(0, 43007, '\P{inlatinextd}', "");
Expect(1, 43007, '\P{^inlatinextd}', "");
Expect(0, 43008, '\p{inlatinextd}', "");
Expect(1, 43008, '\p{^inlatinextd}', "");
Expect(1, 43008, '\P{inlatinextd}', "");
Expect(0, 43008, '\P{^inlatinextd}', "");
Expect(1, 43007, '\p{_In_latin_Ext_D}', "");
Expect(0, 43007, '\p{^_In_latin_Ext_D}', "");
Expect(0, 43007, '\P{_In_latin_Ext_D}', "");
Expect(1, 43007, '\P{^_In_latin_Ext_D}', "");
Expect(0, 43008, '\p{_In_latin_Ext_D}', "");
Expect(1, 43008, '\p{^_In_latin_Ext_D}', "");
Expect(1, 43008, '\P{_In_latin_Ext_D}', "");
Expect(0, 43008, '\P{^_In_latin_Ext_D}', "");
Error('\p{-_LEPCHA:=}');
Error('\P{-_LEPCHA:=}');
Expect(1, 7247, '\p{lepcha}', "");
Expect(0, 7247, '\p{^lepcha}', "");
Expect(0, 7247, '\P{lepcha}', "");
Expect(1, 7247, '\P{^lepcha}', "");
Expect(0, 7248, '\p{lepcha}', "");
Expect(1, 7248, '\p{^lepcha}', "");
Expect(1, 7248, '\P{lepcha}', "");
Expect(0, 7248, '\P{^lepcha}', "");
Expect(1, 7247, '\p{	-LEPCHA}', "");
Expect(0, 7247, '\p{^	-LEPCHA}', "");
Expect(0, 7247, '\P{	-LEPCHA}', "");
Expect(1, 7247, '\P{^	-LEPCHA}', "");
Expect(0, 7248, '\p{	-LEPCHA}', "");
Expect(1, 7248, '\p{^	-LEPCHA}', "");
Expect(1, 7248, '\P{	-LEPCHA}', "");
Expect(0, 7248, '\P{^	-LEPCHA}', "");
Error('\p{ Is_LEPCHA/a/}');
Error('\P{ Is_LEPCHA/a/}');
Expect(1, 7247, '\p{islepcha}', "");
Expect(0, 7247, '\p{^islepcha}', "");
Expect(0, 7247, '\P{islepcha}', "");
Expect(1, 7247, '\P{^islepcha}', "");
Expect(0, 7248, '\p{islepcha}', "");
Expect(1, 7248, '\p{^islepcha}', "");
Expect(1, 7248, '\P{islepcha}', "");
Expect(0, 7248, '\P{^islepcha}', "");
Expect(1, 7247, '\p{_Is_Lepcha}', "");
Expect(0, 7247, '\p{^_Is_Lepcha}', "");
Expect(0, 7247, '\P{_Is_Lepcha}', "");
Expect(1, 7247, '\P{^_Is_Lepcha}', "");
Expect(0, 7248, '\p{_Is_Lepcha}', "");
Expect(1, 7248, '\p{^_Is_Lepcha}', "");
Expect(1, 7248, '\P{_Is_Lepcha}', "");
Expect(0, 7248, '\P{^_Is_Lepcha}', "");
Error('\p{	-Lepc/a/}');
Error('\P{	-Lepc/a/}');
Expect(1, 7247, '\p{lepc}', "");
Expect(0, 7247, '\p{^lepc}', "");
Expect(0, 7247, '\P{lepc}', "");
Expect(1, 7247, '\P{^lepc}', "");
Expect(0, 7248, '\p{lepc}', "");
Expect(1, 7248, '\p{^lepc}', "");
Expect(1, 7248, '\P{lepc}', "");
Expect(0, 7248, '\P{^lepc}', "");
Expect(1, 7247, '\p{ _lepc}', "");
Expect(0, 7247, '\p{^ _lepc}', "");
Expect(0, 7247, '\P{ _lepc}', "");
Expect(1, 7247, '\P{^ _lepc}', "");
Expect(0, 7248, '\p{ _lepc}', "");
Expect(1, 7248, '\p{^ _lepc}', "");
Expect(1, 7248, '\P{ _lepc}', "");
Expect(0, 7248, '\P{^ _lepc}', "");
Error('\p{/a/-IS_lepc}');
Error('\P{/a/-IS_lepc}');
Expect(1, 7247, '\p{islepc}', "");
Expect(0, 7247, '\p{^islepc}', "");
Expect(0, 7247, '\P{islepc}', "");
Expect(1, 7247, '\P{^islepc}', "");
Expect(0, 7248, '\p{islepc}', "");
Expect(1, 7248, '\p{^islepc}', "");
Expect(1, 7248, '\P{islepc}', "");
Expect(0, 7248, '\P{^islepc}', "");
Expect(1, 7247, '\p{ _IS_Lepc}', "");
Expect(0, 7247, '\p{^ _IS_Lepc}', "");
Expect(0, 7247, '\P{ _IS_Lepc}', "");
Expect(1, 7247, '\P{^ _IS_Lepc}', "");
Expect(0, 7248, '\p{ _IS_Lepc}', "");
Expect(1, 7248, '\p{^ _IS_Lepc}', "");
Expect(1, 7248, '\P{ _IS_Lepc}', "");
Expect(0, 7248, '\P{^ _IS_Lepc}', "");
Error('\p{/a/	_Letter}');
Error('\P{/a/	_Letter}');
Expect(1, 195101, '\p{letter}', "");
Expect(0, 195101, '\p{^letter}', "");
Expect(0, 195101, '\P{letter}', "");
Expect(1, 195101, '\P{^letter}', "");
Expect(0, 195102, '\p{letter}', "");
Expect(1, 195102, '\p{^letter}', "");
Expect(1, 195102, '\P{letter}', "");
Expect(0, 195102, '\P{^letter}', "");
Expect(1, 195101, '\p{ _LETTER}', "");
Expect(0, 195101, '\p{^ _LETTER}', "");
Expect(0, 195101, '\P{ _LETTER}', "");
Expect(1, 195101, '\P{^ _LETTER}', "");
Expect(0, 195102, '\p{ _LETTER}', "");
Expect(1, 195102, '\p{^ _LETTER}', "");
Expect(1, 195102, '\P{ _LETTER}', "");
Expect(0, 195102, '\P{^ _LETTER}', "");
Error('\p{_IS_Letter:=}');
Error('\P{_IS_Letter:=}');
Expect(1, 195101, '\p{isletter}', "");
Expect(0, 195101, '\p{^isletter}', "");
Expect(0, 195101, '\P{isletter}', "");
Expect(1, 195101, '\P{^isletter}', "");
Expect(0, 195102, '\p{isletter}', "");
Expect(1, 195102, '\p{^isletter}', "");
Expect(1, 195102, '\P{isletter}', "");
Expect(0, 195102, '\P{^isletter}', "");
Expect(1, 195101, '\p{ 	Is_Letter}', "");
Expect(0, 195101, '\p{^ 	Is_Letter}', "");
Expect(0, 195101, '\P{ 	Is_Letter}', "");
Expect(1, 195101, '\P{^ 	Is_Letter}', "");
Expect(0, 195102, '\p{ 	Is_Letter}', "");
Expect(1, 195102, '\p{^ 	Is_Letter}', "");
Expect(1, 195102, '\P{ 	Is_Letter}', "");
Expect(0, 195102, '\P{^ 	Is_Letter}', "");
Error('\p{:=  L}');
Error('\P{:=  L}');
Expect(1, 195101, '\p{l}', "");
Expect(0, 195101, '\p{^l}', "");
Expect(0, 195101, '\P{l}', "");
Expect(1, 195101, '\P{^l}', "");
Expect(0, 195102, '\p{l}', "");
Expect(1, 195102, '\p{^l}', "");
Expect(1, 195102, '\P{l}', "");
Expect(0, 195102, '\P{^l}', "");
Expect(1, 195101, '\p{ _L}', "");
Expect(0, 195101, '\p{^ _L}', "");
Expect(0, 195101, '\P{ _L}', "");
Expect(1, 195101, '\P{^ _L}', "");
Expect(0, 195102, '\p{ _L}', "");
Expect(1, 195102, '\p{^ _L}', "");
Expect(1, 195102, '\P{ _L}', "");
Expect(0, 195102, '\P{^ _L}', "");
Error('\p{_-IS_l/a/}');
Error('\P{_-IS_l/a/}');
Expect(1, 195101, '\p{isl}', "");
Expect(0, 195101, '\p{^isl}', "");
Expect(0, 195101, '\P{isl}', "");
Expect(1, 195101, '\P{^isl}', "");
Expect(0, 195102, '\p{isl}', "");
Expect(1, 195102, '\p{^isl}', "");
Expect(1, 195102, '\P{isl}', "");
Expect(0, 195102, '\P{^isl}', "");
Expect(1, 195101, '\p{-	IS_l}', "");
Expect(0, 195101, '\p{^-	IS_l}', "");
Expect(0, 195101, '\P{-	IS_l}', "");
Expect(1, 195101, '\P{^-	IS_l}', "");
Expect(0, 195102, '\p{-	IS_l}', "");
Expect(1, 195102, '\p{^-	IS_l}', "");
Expect(1, 195102, '\P{-	IS_l}', "");
Expect(0, 195102, '\P{^-	IS_l}', "");
Error('\p{-Letter_number:=}');
Error('\P{-Letter_number:=}');
Expect(1, 74850, '\p{letternumber}', "");
Expect(0, 74850, '\p{^letternumber}', "");
Expect(0, 74850, '\P{letternumber}', "");
Expect(1, 74850, '\P{^letternumber}', "");
Expect(0, 74851, '\p{letternumber}', "");
Expect(1, 74851, '\p{^letternumber}', "");
Expect(1, 74851, '\P{letternumber}', "");
Expect(0, 74851, '\P{^letternumber}', "");
Expect(1, 74850, '\p{	_Letter_number}', "");
Expect(0, 74850, '\p{^	_Letter_number}', "");
Expect(0, 74850, '\P{	_Letter_number}', "");
Expect(1, 74850, '\P{^	_Letter_number}', "");
Expect(0, 74851, '\p{	_Letter_number}', "");
Expect(1, 74851, '\p{^	_Letter_number}', "");
Expect(1, 74851, '\P{	_Letter_number}', "");
Expect(0, 74851, '\P{^	_Letter_number}', "");
Error('\p{/a/ -is_Letter_Number}');
Error('\P{/a/ -is_Letter_Number}');
Expect(1, 74850, '\p{isletternumber}', "");
Expect(0, 74850, '\p{^isletternumber}', "");
Expect(0, 74850, '\P{isletternumber}', "");
Expect(1, 74850, '\P{^isletternumber}', "");
Expect(0, 74851, '\p{isletternumber}', "");
Expect(1, 74851, '\p{^isletternumber}', "");
Expect(1, 74851, '\P{isletternumber}', "");
Expect(0, 74851, '\P{^isletternumber}', "");
Expect(1, 74850, '\p{_Is_LETTER_NUMBER}', "");
Expect(0, 74850, '\p{^_Is_LETTER_NUMBER}', "");
Expect(0, 74850, '\P{_Is_LETTER_NUMBER}', "");
Expect(1, 74850, '\P{^_Is_LETTER_NUMBER}', "");
Expect(0, 74851, '\p{_Is_LETTER_NUMBER}', "");
Expect(1, 74851, '\p{^_Is_LETTER_NUMBER}', "");
Expect(1, 74851, '\P{_Is_LETTER_NUMBER}', "");
Expect(0, 74851, '\P{^_Is_LETTER_NUMBER}', "");
Error('\p{ /a/Nl}');
Error('\P{ /a/Nl}');
Expect(1, 74850, '\p{nl}', "");
Expect(0, 74850, '\p{^nl}', "");
Expect(0, 74850, '\P{nl}', "");
Expect(1, 74850, '\P{^nl}', "");
Expect(0, 74851, '\p{nl}', "");
Expect(1, 74851, '\p{^nl}', "");
Expect(1, 74851, '\P{nl}', "");
Expect(0, 74851, '\P{^nl}', "");
Expect(1, 74850, '\p{	_nl}', "");
Expect(0, 74850, '\p{^	_nl}', "");
Expect(0, 74850, '\P{	_nl}', "");
Expect(1, 74850, '\P{^	_nl}', "");
Expect(0, 74851, '\p{	_nl}', "");
Expect(1, 74851, '\p{^	_nl}', "");
Expect(1, 74851, '\P{	_nl}', "");
Expect(0, 74851, '\P{^	_nl}', "");
Error('\p{/a/  Is_Nl}');
Error('\P{/a/  Is_Nl}');
Expect(1, 74850, '\p{isnl}', "");
Expect(0, 74850, '\p{^isnl}', "");
Expect(0, 74850, '\P{isnl}', "");
Expect(1, 74850, '\P{^isnl}', "");
Expect(0, 74851, '\p{isnl}', "");
Expect(1, 74851, '\p{^isnl}', "");
Expect(1, 74851, '\P{isnl}', "");
Expect(0, 74851, '\P{^isnl}', "");
Expect(1, 74850, '\p{_ is_NL}', "");
Expect(0, 74850, '\p{^_ is_NL}', "");
Expect(0, 74850, '\P{_ is_NL}', "");
Expect(1, 74850, '\P{^_ is_NL}', "");
Expect(0, 74851, '\p{_ is_NL}', "");
Expect(1, 74851, '\p{^_ is_NL}', "");
Expect(1, 74851, '\P{_ is_NL}', "");
Expect(0, 74851, '\P{^_ is_NL}', "");
Error('\p{ :=letterlike_symbols}');
Error('\P{ :=letterlike_symbols}');
Expect(1, 8527, '\p{letterlikesymbols}', "");
Expect(0, 8527, '\p{^letterlikesymbols}', "");
Expect(0, 8527, '\P{letterlikesymbols}', "");
Expect(1, 8527, '\P{^letterlikesymbols}', "");
Expect(0, 8528, '\p{letterlikesymbols}', "");
Expect(1, 8528, '\p{^letterlikesymbols}', "");
Expect(1, 8528, '\P{letterlikesymbols}', "");
Expect(0, 8528, '\P{^letterlikesymbols}', "");
Expect(1, 8527, '\p{ Letterlike_symbols}', "");
Expect(0, 8527, '\p{^ Letterlike_symbols}', "");
Expect(0, 8527, '\P{ Letterlike_symbols}', "");
Expect(1, 8527, '\P{^ Letterlike_symbols}', "");
Expect(0, 8528, '\p{ Letterlike_symbols}', "");
Expect(1, 8528, '\p{^ Letterlike_symbols}', "");
Expect(1, 8528, '\P{ Letterlike_symbols}', "");
Expect(0, 8528, '\P{^ Letterlike_symbols}', "");
Error('\p{:=-_IS_Letterlike_SYMBOLS}');
Error('\P{:=-_IS_Letterlike_SYMBOLS}');
Expect(1, 8527, '\p{isletterlikesymbols}', "");
Expect(0, 8527, '\p{^isletterlikesymbols}', "");
Expect(0, 8527, '\P{isletterlikesymbols}', "");
Expect(1, 8527, '\P{^isletterlikesymbols}', "");
Expect(0, 8528, '\p{isletterlikesymbols}', "");
Expect(1, 8528, '\p{^isletterlikesymbols}', "");
Expect(1, 8528, '\P{isletterlikesymbols}', "");
Expect(0, 8528, '\P{^isletterlikesymbols}', "");
Expect(1, 8527, '\p{__Is_Letterlike_SYMBOLS}', "");
Expect(0, 8527, '\p{^__Is_Letterlike_SYMBOLS}', "");
Expect(0, 8527, '\P{__Is_Letterlike_SYMBOLS}', "");
Expect(1, 8527, '\P{^__Is_Letterlike_SYMBOLS}', "");
Expect(0, 8528, '\p{__Is_Letterlike_SYMBOLS}', "");
Expect(1, 8528, '\p{^__Is_Letterlike_SYMBOLS}', "");
Expect(1, 8528, '\P{__Is_Letterlike_SYMBOLS}', "");
Expect(0, 8528, '\P{^__Is_Letterlike_SYMBOLS}', "");
Error('\p{:=_in_Letterlike_Symbols}');
Error('\P{:=_in_Letterlike_Symbols}');
Expect(1, 8527, '\p{inletterlikesymbols}', "");
Expect(0, 8527, '\p{^inletterlikesymbols}', "");
Expect(0, 8527, '\P{inletterlikesymbols}', "");
Expect(1, 8527, '\P{^inletterlikesymbols}', "");
Expect(0, 8528, '\p{inletterlikesymbols}', "");
Expect(1, 8528, '\p{^inletterlikesymbols}', "");
Expect(1, 8528, '\P{inletterlikesymbols}', "");
Expect(0, 8528, '\P{^inletterlikesymbols}', "");
Expect(1, 8527, '\p{In_Letterlike_symbols}', "");
Expect(0, 8527, '\p{^In_Letterlike_symbols}', "");
Expect(0, 8527, '\P{In_Letterlike_symbols}', "");
Expect(1, 8527, '\P{^In_Letterlike_symbols}', "");
Expect(0, 8528, '\p{In_Letterlike_symbols}', "");
Expect(1, 8528, '\p{^In_Letterlike_symbols}', "");
Expect(1, 8528, '\P{In_Letterlike_symbols}', "");
Expect(0, 8528, '\P{^In_Letterlike_symbols}', "");
Error('\p{:=	 Limbu}');
Error('\P{:=	 Limbu}');
Expect(1, 6479, '\p{limbu}', "");
Expect(0, 6479, '\p{^limbu}', "");
Expect(0, 6479, '\P{limbu}', "");
Expect(1, 6479, '\P{^limbu}', "");
Expect(0, 6480, '\p{limbu}', "");
Expect(1, 6480, '\p{^limbu}', "");
Expect(1, 6480, '\P{limbu}', "");
Expect(0, 6480, '\P{^limbu}', "");
Expect(1, 6479, '\p{ 	Limbu}', "");
Expect(0, 6479, '\p{^ 	Limbu}', "");
Expect(0, 6479, '\P{ 	Limbu}', "");
Expect(1, 6479, '\P{^ 	Limbu}', "");
Expect(0, 6480, '\p{ 	Limbu}', "");
Expect(1, 6480, '\p{^ 	Limbu}', "");
Expect(1, 6480, '\P{ 	Limbu}', "");
Expect(0, 6480, '\P{^ 	Limbu}', "");
Error('\p{:=IS_Limbu}');
Error('\P{:=IS_Limbu}');
Expect(1, 6479, '\p{islimbu}', "");
Expect(0, 6479, '\p{^islimbu}', "");
Expect(0, 6479, '\P{islimbu}', "");
Expect(1, 6479, '\P{^islimbu}', "");
Expect(0, 6480, '\p{islimbu}', "");
Expect(1, 6480, '\p{^islimbu}', "");
Expect(1, 6480, '\P{islimbu}', "");
Expect(0, 6480, '\P{^islimbu}', "");
Expect(1, 6479, '\p{_IS_LIMBU}', "");
Expect(0, 6479, '\p{^_IS_LIMBU}', "");
Expect(0, 6479, '\P{_IS_LIMBU}', "");
Expect(1, 6479, '\P{^_IS_LIMBU}', "");
Expect(0, 6480, '\p{_IS_LIMBU}', "");
Expect(1, 6480, '\p{^_IS_LIMBU}', "");
Expect(1, 6480, '\P{_IS_LIMBU}', "");
Expect(0, 6480, '\P{^_IS_LIMBU}', "");
Error('\p{_	Limb/a/}');
Error('\P{_	Limb/a/}');
Expect(1, 6479, '\p{limb}', "");
Expect(0, 6479, '\p{^limb}', "");
Expect(0, 6479, '\P{limb}', "");
Expect(1, 6479, '\P{^limb}', "");
Expect(0, 6480, '\p{limb}', "");
Expect(1, 6480, '\p{^limb}', "");
Expect(1, 6480, '\P{limb}', "");
Expect(0, 6480, '\P{^limb}', "");
Expect(1, 6479, '\p{-Limb}', "");
Expect(0, 6479, '\p{^-Limb}', "");
Expect(0, 6479, '\P{-Limb}', "");
Expect(1, 6479, '\P{^-Limb}', "");
Expect(0, 6480, '\p{-Limb}', "");
Expect(1, 6480, '\p{^-Limb}', "");
Expect(1, 6480, '\P{-Limb}', "");
Expect(0, 6480, '\P{^-Limb}', "");
Error('\p{_ Is_Limb/a/}');
Error('\P{_ Is_Limb/a/}');
Expect(1, 6479, '\p{islimb}', "");
Expect(0, 6479, '\p{^islimb}', "");
Expect(0, 6479, '\P{islimb}', "");
Expect(1, 6479, '\P{^islimb}', "");
Expect(0, 6480, '\p{islimb}', "");
Expect(1, 6480, '\p{^islimb}', "");
Expect(1, 6480, '\P{islimb}', "");
Expect(0, 6480, '\P{^islimb}', "");
Expect(1, 6479, '\p{-is_limb}', "");
Expect(0, 6479, '\p{^-is_limb}', "");
Expect(0, 6479, '\P{-is_limb}', "");
Expect(1, 6479, '\P{^-is_limb}', "");
Expect(0, 6480, '\p{-is_limb}', "");
Expect(1, 6480, '\p{^-is_limb}', "");
Expect(1, 6480, '\P{-is_limb}', "");
Expect(0, 6480, '\P{^-is_limb}', "");
Error('\p{:= -LINE_separator}');
Error('\P{:= -LINE_separator}');
Expect(1, 8232, '\p{lineseparator}', "");
Expect(0, 8232, '\p{^lineseparator}', "");
Expect(0, 8232, '\P{lineseparator}', "");
Expect(1, 8232, '\P{^lineseparator}', "");
Expect(0, 8233, '\p{lineseparator}', "");
Expect(1, 8233, '\p{^lineseparator}', "");
Expect(1, 8233, '\P{lineseparator}', "");
Expect(0, 8233, '\P{^lineseparator}', "");
Expect(1, 8232, '\p{ Line_SEPARATOR}', "");
Expect(0, 8232, '\p{^ Line_SEPARATOR}', "");
Expect(0, 8232, '\P{ Line_SEPARATOR}', "");
Expect(1, 8232, '\P{^ Line_SEPARATOR}', "");
Expect(0, 8233, '\p{ Line_SEPARATOR}', "");
Expect(1, 8233, '\p{^ Line_SEPARATOR}', "");
Expect(1, 8233, '\P{ Line_SEPARATOR}', "");
Expect(0, 8233, '\P{^ Line_SEPARATOR}', "");
Error('\p{:=Is_line_Separator}');
Error('\P{:=Is_line_Separator}');
Expect(1, 8232, '\p{islineseparator}', "");
Expect(0, 8232, '\p{^islineseparator}', "");
Expect(0, 8232, '\P{islineseparator}', "");
Expect(1, 8232, '\P{^islineseparator}', "");
Expect(0, 8233, '\p{islineseparator}', "");
Expect(1, 8233, '\p{^islineseparator}', "");
Expect(1, 8233, '\P{islineseparator}', "");
Expect(0, 8233, '\P{^islineseparator}', "");
Expect(1, 8232, '\p{ IS_Line_Separator}', "");
Expect(0, 8232, '\p{^ IS_Line_Separator}', "");
Expect(0, 8232, '\P{ IS_Line_Separator}', "");
Expect(1, 8232, '\P{^ IS_Line_Separator}', "");
Expect(0, 8233, '\p{ IS_Line_Separator}', "");
Expect(1, 8233, '\p{^ IS_Line_Separator}', "");
Expect(1, 8233, '\P{ IS_Line_Separator}', "");
Expect(0, 8233, '\P{^ IS_Line_Separator}', "");
Error('\p{_Zl/a/}');
Error('\P{_Zl/a/}');
Expect(1, 8232, '\p{zl}', "");
Expect(0, 8232, '\p{^zl}', "");
Expect(0, 8232, '\P{zl}', "");
Expect(1, 8232, '\P{^zl}', "");
Expect(0, 8233, '\p{zl}', "");
Expect(1, 8233, '\p{^zl}', "");
Expect(1, 8233, '\P{zl}', "");
Expect(0, 8233, '\P{^zl}', "");
Expect(1, 8232, '\p{_-zl}', "");
Expect(0, 8232, '\p{^_-zl}', "");
Expect(0, 8232, '\P{_-zl}', "");
Expect(1, 8232, '\P{^_-zl}', "");
Expect(0, 8233, '\p{_-zl}', "");
Expect(1, 8233, '\p{^_-zl}', "");
Expect(1, 8233, '\P{_-zl}', "");
Expect(0, 8233, '\P{^_-zl}', "");
Error('\p{-Is_Zl/a/}');
Error('\P{-Is_Zl/a/}');
Expect(1, 8232, '\p{iszl}', "");
Expect(0, 8232, '\p{^iszl}', "");
Expect(0, 8232, '\P{iszl}', "");
Expect(1, 8232, '\P{^iszl}', "");
Expect(0, 8233, '\p{iszl}', "");
Expect(1, 8233, '\p{^iszl}', "");
Expect(1, 8233, '\P{iszl}', "");
Expect(0, 8233, '\P{^iszl}', "");
Expect(1, 8232, '\p{		Is_zl}', "");
Expect(0, 8232, '\p{^		Is_zl}', "");
Expect(0, 8232, '\P{		Is_zl}', "");
Expect(1, 8232, '\P{^		Is_zl}', "");
Expect(0, 8233, '\p{		Is_zl}', "");
Expect(1, 8233, '\p{^		Is_zl}', "");
Expect(1, 8233, '\P{		Is_zl}', "");
Expect(0, 8233, '\P{^		Is_zl}', "");
Error('\p{  LINEAR_b/a/}');
Error('\P{  LINEAR_b/a/}');
Expect(1, 65786, '\p{linearb}', "");
Expect(0, 65786, '\p{^linearb}', "");
Expect(0, 65786, '\P{linearb}', "");
Expect(1, 65786, '\P{^linearb}', "");
Expect(0, 65787, '\p{linearb}', "");
Expect(1, 65787, '\p{^linearb}', "");
Expect(1, 65787, '\P{linearb}', "");
Expect(0, 65787, '\P{^linearb}', "");
Expect(1, 65786, '\p{-_LINEAR_B}', "");
Expect(0, 65786, '\p{^-_LINEAR_B}', "");
Expect(0, 65786, '\P{-_LINEAR_B}', "");
Expect(1, 65786, '\P{^-_LINEAR_B}', "");
Expect(0, 65787, '\p{-_LINEAR_B}', "");
Expect(1, 65787, '\p{^-_LINEAR_B}', "");
Expect(1, 65787, '\P{-_LINEAR_B}', "");
Expect(0, 65787, '\P{^-_LINEAR_B}', "");
Error('\p{-_Is_Linear_B/a/}');
Error('\P{-_Is_Linear_B/a/}');
Expect(1, 65786, '\p{islinearb}', "");
Expect(0, 65786, '\p{^islinearb}', "");
Expect(0, 65786, '\P{islinearb}', "");
Expect(1, 65786, '\P{^islinearb}', "");
Expect(0, 65787, '\p{islinearb}', "");
Expect(1, 65787, '\p{^islinearb}', "");
Expect(1, 65787, '\P{islinearb}', "");
Expect(0, 65787, '\P{^islinearb}', "");
Expect(1, 65786, '\p{	 is_LINEAR_b}', "");
Expect(0, 65786, '\p{^	 is_LINEAR_b}', "");
Expect(0, 65786, '\P{	 is_LINEAR_b}', "");
Expect(1, 65786, '\P{^	 is_LINEAR_b}', "");
Expect(0, 65787, '\p{	 is_LINEAR_b}', "");
Expect(1, 65787, '\p{^	 is_LINEAR_b}', "");
Expect(1, 65787, '\P{	 is_LINEAR_b}', "");
Expect(0, 65787, '\P{^	 is_LINEAR_b}', "");
Error('\p{:= 	Linb}');
Error('\P{:= 	Linb}');
Expect(1, 65786, '\p{linb}', "");
Expect(0, 65786, '\p{^linb}', "");
Expect(0, 65786, '\P{linb}', "");
Expect(1, 65786, '\P{^linb}', "");
Expect(0, 65787, '\p{linb}', "");
Expect(1, 65787, '\p{^linb}', "");
Expect(1, 65787, '\P{linb}', "");
Expect(0, 65787, '\P{^linb}', "");
Expect(1, 65786, '\p{ 	Linb}', "");
Expect(0, 65786, '\p{^ 	Linb}', "");
Expect(0, 65786, '\P{ 	Linb}', "");
Expect(1, 65786, '\P{^ 	Linb}', "");
Expect(0, 65787, '\p{ 	Linb}', "");
Expect(1, 65787, '\p{^ 	Linb}', "");
Expect(1, 65787, '\P{ 	Linb}', "");
Expect(0, 65787, '\P{^ 	Linb}', "");
Error('\p{:=	-Is_LINB}');
Error('\P{:=	-Is_LINB}');
Expect(1, 65786, '\p{islinb}', "");
Expect(0, 65786, '\p{^islinb}', "");
Expect(0, 65786, '\P{islinb}', "");
Expect(1, 65786, '\P{^islinb}', "");
Expect(0, 65787, '\p{islinb}', "");
Expect(1, 65787, '\p{^islinb}', "");
Expect(1, 65787, '\P{islinb}', "");
Expect(0, 65787, '\P{^islinb}', "");
Expect(1, 65786, '\p{ IS_LINB}', "");
Expect(0, 65786, '\p{^ IS_LINB}', "");
Expect(0, 65786, '\P{ IS_LINB}', "");
Expect(1, 65786, '\P{^ IS_LINB}', "");
Expect(0, 65787, '\p{ IS_LINB}', "");
Expect(1, 65787, '\p{^ IS_LINB}', "");
Expect(1, 65787, '\P{ IS_LINB}', "");
Expect(0, 65787, '\P{^ IS_LINB}', "");
Error('\p{-	Linear_b_IDEOGRAMS:=}');
Error('\P{-	Linear_b_IDEOGRAMS:=}');
Expect(1, 65791, '\p{linearbideograms}', "");
Expect(0, 65791, '\p{^linearbideograms}', "");
Expect(0, 65791, '\P{linearbideograms}', "");
Expect(1, 65791, '\P{^linearbideograms}', "");
Expect(0, 65792, '\p{linearbideograms}', "");
Expect(1, 65792, '\p{^linearbideograms}', "");
Expect(1, 65792, '\P{linearbideograms}', "");
Expect(0, 65792, '\P{^linearbideograms}', "");
Expect(1, 65791, '\p{-Linear_B_IDEOGRAMS}', "");
Expect(0, 65791, '\p{^-Linear_B_IDEOGRAMS}', "");
Expect(0, 65791, '\P{-Linear_B_IDEOGRAMS}', "");
Expect(1, 65791, '\P{^-Linear_B_IDEOGRAMS}', "");
Expect(0, 65792, '\p{-Linear_B_IDEOGRAMS}', "");
Expect(1, 65792, '\p{^-Linear_B_IDEOGRAMS}', "");
Expect(1, 65792, '\P{-Linear_B_IDEOGRAMS}', "");
Expect(0, 65792, '\P{^-Linear_B_IDEOGRAMS}', "");
Error('\p{:=	IS_LINEAR_b_Ideograms}');
Error('\P{:=	IS_LINEAR_b_Ideograms}');
Expect(1, 65791, '\p{islinearbideograms}', "");
Expect(0, 65791, '\p{^islinearbideograms}', "");
Expect(0, 65791, '\P{islinearbideograms}', "");
Expect(1, 65791, '\P{^islinearbideograms}', "");
Expect(0, 65792, '\p{islinearbideograms}', "");
Expect(1, 65792, '\p{^islinearbideograms}', "");
Expect(1, 65792, '\P{islinearbideograms}', "");
Expect(0, 65792, '\P{^islinearbideograms}', "");
Expect(1, 65791, '\p{  Is_LINEAR_B_Ideograms}', "");
Expect(0, 65791, '\p{^  Is_LINEAR_B_Ideograms}', "");
Expect(0, 65791, '\P{  Is_LINEAR_B_Ideograms}', "");
Expect(1, 65791, '\P{^  Is_LINEAR_B_Ideograms}', "");
Expect(0, 65792, '\p{  Is_LINEAR_B_Ideograms}', "");
Expect(1, 65792, '\p{^  Is_LINEAR_B_Ideograms}', "");
Expect(1, 65792, '\P{  Is_LINEAR_B_Ideograms}', "");
Expect(0, 65792, '\P{^  Is_LINEAR_B_Ideograms}', "");
Error('\p{ In_Linear_B_Ideograms/a/}');
Error('\P{ In_Linear_B_Ideograms/a/}');
Expect(1, 65791, '\p{inlinearbideograms}', "");
Expect(0, 65791, '\p{^inlinearbideograms}', "");
Expect(0, 65791, '\P{inlinearbideograms}', "");
Expect(1, 65791, '\P{^inlinearbideograms}', "");
Expect(0, 65792, '\p{inlinearbideograms}', "");
Expect(1, 65792, '\p{^inlinearbideograms}', "");
Expect(1, 65792, '\P{inlinearbideograms}', "");
Expect(0, 65792, '\P{^inlinearbideograms}', "");
Expect(1, 65791, '\p{		in_Linear_B_Ideograms}', "");
Expect(0, 65791, '\p{^		in_Linear_B_Ideograms}', "");
Expect(0, 65791, '\P{		in_Linear_B_Ideograms}', "");
Expect(1, 65791, '\P{^		in_Linear_B_Ideograms}', "");
Expect(0, 65792, '\p{		in_Linear_B_Ideograms}', "");
Expect(1, 65792, '\p{^		in_Linear_B_Ideograms}', "");
Expect(1, 65792, '\P{		in_Linear_B_Ideograms}', "");
Expect(0, 65792, '\P{^		in_Linear_B_Ideograms}', "");
Error('\p{_/a/Linear_B_SYLLABARY}');
Error('\P{_/a/Linear_B_SYLLABARY}');
Expect(1, 65663, '\p{linearbsyllabary}', "");
Expect(0, 65663, '\p{^linearbsyllabary}', "");
Expect(0, 65663, '\P{linearbsyllabary}', "");
Expect(1, 65663, '\P{^linearbsyllabary}', "");
Expect(0, 65664, '\p{linearbsyllabary}', "");
Expect(1, 65664, '\p{^linearbsyllabary}', "");
Expect(1, 65664, '\P{linearbsyllabary}', "");
Expect(0, 65664, '\P{^linearbsyllabary}', "");
Expect(1, 65663, '\p{	-LINEAR_B_SYLLABARY}', "");
Expect(0, 65663, '\p{^	-LINEAR_B_SYLLABARY}', "");
Expect(0, 65663, '\P{	-LINEAR_B_SYLLABARY}', "");
Expect(1, 65663, '\P{^	-LINEAR_B_SYLLABARY}', "");
Expect(0, 65664, '\p{	-LINEAR_B_SYLLABARY}', "");
Expect(1, 65664, '\p{^	-LINEAR_B_SYLLABARY}', "");
Expect(1, 65664, '\P{	-LINEAR_B_SYLLABARY}', "");
Expect(0, 65664, '\P{^	-LINEAR_B_SYLLABARY}', "");
Error('\p{/a/_IS_Linear_B_Syllabary}');
Error('\P{/a/_IS_Linear_B_Syllabary}');
Expect(1, 65663, '\p{islinearbsyllabary}', "");
Expect(0, 65663, '\p{^islinearbsyllabary}', "");
Expect(0, 65663, '\P{islinearbsyllabary}', "");
Expect(1, 65663, '\P{^islinearbsyllabary}', "");
Expect(0, 65664, '\p{islinearbsyllabary}', "");
Expect(1, 65664, '\p{^islinearbsyllabary}', "");
Expect(1, 65664, '\P{islinearbsyllabary}', "");
Expect(0, 65664, '\P{^islinearbsyllabary}', "");
Expect(1, 65663, '\p{ -IS_linear_B_SYLLABARY}', "");
Expect(0, 65663, '\p{^ -IS_linear_B_SYLLABARY}', "");
Expect(0, 65663, '\P{ -IS_linear_B_SYLLABARY}', "");
Expect(1, 65663, '\P{^ -IS_linear_B_SYLLABARY}', "");
Expect(0, 65664, '\p{ -IS_linear_B_SYLLABARY}', "");
Expect(1, 65664, '\p{^ -IS_linear_B_SYLLABARY}', "");
Expect(1, 65664, '\P{ -IS_linear_B_SYLLABARY}', "");
Expect(0, 65664, '\P{^ -IS_linear_B_SYLLABARY}', "");
Error('\p{/a/_ In_linear_b_Syllabary}');
Error('\P{/a/_ In_linear_b_Syllabary}');
Expect(1, 65663, '\p{inlinearbsyllabary}', "");
Expect(0, 65663, '\p{^inlinearbsyllabary}', "");
Expect(0, 65663, '\P{inlinearbsyllabary}', "");
Expect(1, 65663, '\P{^inlinearbsyllabary}', "");
Expect(0, 65664, '\p{inlinearbsyllabary}', "");
Expect(1, 65664, '\p{^inlinearbsyllabary}', "");
Expect(1, 65664, '\P{inlinearbsyllabary}', "");
Expect(0, 65664, '\P{^inlinearbsyllabary}', "");
Expect(1, 65663, '\p{ 	In_Linear_B_syllabary}', "");
Expect(0, 65663, '\p{^ 	In_Linear_B_syllabary}', "");
Expect(0, 65663, '\P{ 	In_Linear_B_syllabary}', "");
Expect(1, 65663, '\P{^ 	In_Linear_B_syllabary}', "");
Expect(0, 65664, '\p{ 	In_Linear_B_syllabary}', "");
Expect(1, 65664, '\p{^ 	In_Linear_B_syllabary}', "");
Expect(1, 65664, '\P{ 	In_Linear_B_syllabary}', "");
Expect(0, 65664, '\P{^ 	In_Linear_B_syllabary}', "");
Error('\p{	/a/LISU}');
Error('\P{	/a/LISU}');
Expect(1, 42239, '\p{lisu}', "");
Expect(0, 42239, '\p{^lisu}', "");
Expect(0, 42239, '\P{lisu}', "");
Expect(1, 42239, '\P{^lisu}', "");
Expect(0, 42240, '\p{lisu}', "");
Expect(1, 42240, '\p{^lisu}', "");
Expect(1, 42240, '\P{lisu}', "");
Expect(0, 42240, '\P{^lisu}', "");
Expect(1, 42239, '\p{	Lisu}', "");
Expect(0, 42239, '\p{^	Lisu}', "");
Expect(0, 42239, '\P{	Lisu}', "");
Expect(1, 42239, '\P{^	Lisu}', "");
Expect(0, 42240, '\p{	Lisu}', "");
Expect(1, 42240, '\p{^	Lisu}', "");
Expect(1, 42240, '\P{	Lisu}', "");
Expect(0, 42240, '\P{^	Lisu}', "");
Error('\p{	is_Lisu/a/}');
Error('\P{	is_Lisu/a/}');
Expect(1, 42239, '\p{islisu}', "");
Expect(0, 42239, '\p{^islisu}', "");
Expect(0, 42239, '\P{islisu}', "");
Expect(1, 42239, '\P{^islisu}', "");
Expect(0, 42240, '\p{islisu}', "");
Expect(1, 42240, '\p{^islisu}', "");
Expect(1, 42240, '\P{islisu}', "");
Expect(0, 42240, '\P{^islisu}', "");
Expect(1, 42239, '\p{__Is_LISU}', "");
Expect(0, 42239, '\p{^__Is_LISU}', "");
Expect(0, 42239, '\P{__Is_LISU}', "");
Expect(1, 42239, '\P{^__Is_LISU}', "");
Expect(0, 42240, '\p{__Is_LISU}', "");
Expect(1, 42240, '\p{^__Is_LISU}', "");
Expect(1, 42240, '\P{__Is_LISU}', "");
Expect(0, 42240, '\P{^__Is_LISU}', "");
Error('\p{	_Logical_Order_Exception/a/}');
Error('\P{	_Logical_Order_Exception/a/}');
Expect(1, 43708, '\p{logicalorderexception}', "");
Expect(0, 43708, '\p{^logicalorderexception}', "");
Expect(0, 43708, '\P{logicalorderexception}', "");
Expect(1, 43708, '\P{^logicalorderexception}', "");
Expect(0, 43709, '\p{logicalorderexception}', "");
Expect(1, 43709, '\p{^logicalorderexception}', "");
Expect(1, 43709, '\P{logicalorderexception}', "");
Expect(0, 43709, '\P{^logicalorderexception}', "");
Expect(1, 43708, '\p{  Logical_ORDER_EXCEPTION}', "");
Expect(0, 43708, '\p{^  Logical_ORDER_EXCEPTION}', "");
Expect(0, 43708, '\P{  Logical_ORDER_EXCEPTION}', "");
Expect(1, 43708, '\P{^  Logical_ORDER_EXCEPTION}', "");
Expect(0, 43709, '\p{  Logical_ORDER_EXCEPTION}', "");
Expect(1, 43709, '\p{^  Logical_ORDER_EXCEPTION}', "");
Expect(1, 43709, '\P{  Logical_ORDER_EXCEPTION}', "");
Expect(0, 43709, '\P{^  Logical_ORDER_EXCEPTION}', "");
Error('\p{/a/_Is_logical_order_Exception}');
Error('\P{/a/_Is_logical_order_Exception}');
Expect(1, 43708, '\p{islogicalorderexception}', "");
Expect(0, 43708, '\p{^islogicalorderexception}', "");
Expect(0, 43708, '\P{islogicalorderexception}', "");
Expect(1, 43708, '\P{^islogicalorderexception}', "");
Expect(0, 43709, '\p{islogicalorderexception}', "");
Expect(1, 43709, '\p{^islogicalorderexception}', "");
Expect(1, 43709, '\P{islogicalorderexception}', "");
Expect(0, 43709, '\P{^islogicalorderexception}', "");
Expect(1, 43708, '\p{ _Is_logical_ORDER_exception}', "");
Expect(0, 43708, '\p{^ _Is_logical_ORDER_exception}', "");
Expect(0, 43708, '\P{ _Is_logical_ORDER_exception}', "");
Expect(1, 43708, '\P{^ _Is_logical_ORDER_exception}', "");
Expect(0, 43709, '\p{ _Is_logical_ORDER_exception}', "");
Expect(1, 43709, '\p{^ _Is_logical_ORDER_exception}', "");
Expect(1, 43709, '\P{ _Is_logical_ORDER_exception}', "");
Expect(0, 43709, '\P{^ _Is_logical_ORDER_exception}', "");
Error('\p{	_LOE:=}');
Error('\P{	_LOE:=}');
Expect(1, 43708, '\p{loe}', "");
Expect(0, 43708, '\p{^loe}', "");
Expect(0, 43708, '\P{loe}', "");
Expect(1, 43708, '\P{^loe}', "");
Expect(0, 43709, '\p{loe}', "");
Expect(1, 43709, '\p{^loe}', "");
Expect(1, 43709, '\P{loe}', "");
Expect(0, 43709, '\P{^loe}', "");
Expect(1, 43708, '\p{	loe}', "");
Expect(0, 43708, '\p{^	loe}', "");
Expect(0, 43708, '\P{	loe}', "");
Expect(1, 43708, '\P{^	loe}', "");
Expect(0, 43709, '\p{	loe}', "");
Expect(1, 43709, '\p{^	loe}', "");
Expect(1, 43709, '\P{	loe}', "");
Expect(0, 43709, '\P{^	loe}', "");
Error('\p{/a/-is_LOE}');
Error('\P{/a/-is_LOE}');
Expect(1, 43708, '\p{isloe}', "");
Expect(0, 43708, '\p{^isloe}', "");
Expect(0, 43708, '\P{isloe}', "");
Expect(1, 43708, '\P{^isloe}', "");
Expect(0, 43709, '\p{isloe}', "");
Expect(1, 43709, '\p{^isloe}', "");
Expect(1, 43709, '\P{isloe}', "");
Expect(0, 43709, '\P{^isloe}', "");
Expect(1, 43708, '\p{	_Is_loe}', "");
Expect(0, 43708, '\p{^	_Is_loe}', "");
Expect(0, 43708, '\P{	_Is_loe}', "");
Expect(1, 43708, '\P{^	_Is_loe}', "");
Expect(0, 43709, '\p{	_Is_loe}', "");
Expect(1, 43709, '\p{^	_Is_loe}', "");
Expect(1, 43709, '\P{	_Is_loe}', "");
Expect(0, 43709, '\P{^	_Is_loe}', "");
Error('\p{:=	_Low_Surrogates}');
Error('\P{:=	_Low_Surrogates}');
Expect(1, 57343, '\p{lowsurrogates}', "");
Expect(0, 57343, '\p{^lowsurrogates}', "");
Expect(0, 57343, '\P{lowsurrogates}', "");
Expect(1, 57343, '\P{^lowsurrogates}', "");
Expect(0, 57344, '\p{lowsurrogates}', "");
Expect(1, 57344, '\p{^lowsurrogates}', "");
Expect(1, 57344, '\P{lowsurrogates}', "");
Expect(0, 57344, '\P{^lowsurrogates}', "");
Expect(1, 57343, '\p{	LOW_SURROGATES}', "");
Expect(0, 57343, '\p{^	LOW_SURROGATES}', "");
Expect(0, 57343, '\P{	LOW_SURROGATES}', "");
Expect(1, 57343, '\P{^	LOW_SURROGATES}', "");
Expect(0, 57344, '\p{	LOW_SURROGATES}', "");
Expect(1, 57344, '\p{^	LOW_SURROGATES}', "");
Expect(1, 57344, '\P{	LOW_SURROGATES}', "");
Expect(0, 57344, '\P{^	LOW_SURROGATES}', "");
Error('\p{:=_Is_Low_SURROGATES}');
Error('\P{:=_Is_Low_SURROGATES}');
Expect(1, 57343, '\p{islowsurrogates}', "");
Expect(0, 57343, '\p{^islowsurrogates}', "");
Expect(0, 57343, '\P{islowsurrogates}', "");
Expect(1, 57343, '\P{^islowsurrogates}', "");
Expect(0, 57344, '\p{islowsurrogates}', "");
Expect(1, 57344, '\p{^islowsurrogates}', "");
Expect(1, 57344, '\P{islowsurrogates}', "");
Expect(0, 57344, '\P{^islowsurrogates}', "");
Expect(1, 57343, '\p{_-Is_low_surrogates}', "");
Expect(0, 57343, '\p{^_-Is_low_surrogates}', "");
Expect(0, 57343, '\P{_-Is_low_surrogates}', "");
Expect(1, 57343, '\P{^_-Is_low_surrogates}', "");
Expect(0, 57344, '\p{_-Is_low_surrogates}', "");
Expect(1, 57344, '\p{^_-Is_low_surrogates}', "");
Expect(1, 57344, '\P{_-Is_low_surrogates}', "");
Expect(0, 57344, '\P{^_-Is_low_surrogates}', "");
Error('\p{-/a/In_LOW_Surrogates}');
Error('\P{-/a/In_LOW_Surrogates}');
Expect(1, 57343, '\p{inlowsurrogates}', "");
Expect(0, 57343, '\p{^inlowsurrogates}', "");
Expect(0, 57343, '\P{inlowsurrogates}', "");
Expect(1, 57343, '\P{^inlowsurrogates}', "");
Expect(0, 57344, '\p{inlowsurrogates}', "");
Expect(1, 57344, '\p{^inlowsurrogates}', "");
Expect(1, 57344, '\P{inlowsurrogates}', "");
Expect(0, 57344, '\P{^inlowsurrogates}', "");
Expect(1, 57343, '\p{_	In_LOW_SURROGATES}', "");
Expect(0, 57343, '\p{^_	In_LOW_SURROGATES}', "");
Expect(0, 57343, '\P{_	In_LOW_SURROGATES}', "");
Expect(1, 57343, '\P{^_	In_LOW_SURROGATES}', "");
Expect(0, 57344, '\p{_	In_LOW_SURROGATES}', "");
Expect(1, 57344, '\p{^_	In_LOW_SURROGATES}', "");
Expect(1, 57344, '\P{_	In_LOW_SURROGATES}', "");
Expect(0, 57344, '\P{^_	In_LOW_SURROGATES}', "");
Error('\p{-	lower/a/}');
Error('\P{-	lower/a/}');
Expect(1, 120779, '\p{lower}', "");
Expect(0, 120779, '\p{^lower}', "");
Expect(0, 120779, '\P{lower}', "");
Expect(1, 120779, '\P{^lower}', "");
Expect(0, 120780, '\p{lower}', "");
Expect(1, 120780, '\p{^lower}', "");
Expect(1, 120780, '\P{lower}', "");
Expect(0, 120780, '\P{^lower}', "");
Expect(1, 120779, '\p{-lower}', "");
Expect(0, 120779, '\p{^-lower}', "");
Expect(0, 120779, '\P{-lower}', "");
Expect(1, 120779, '\P{^-lower}', "");
Expect(0, 120780, '\p{-lower}', "");
Expect(1, 120780, '\p{^-lower}', "");
Expect(1, 120780, '\P{-lower}', "");
Expect(0, 120780, '\P{^-lower}', "");
Error('\p{/a/-XPosixLower}');
Error('\P{/a/-XPosixLower}');
Expect(1, 120779, '\p{xposixlower}', "");
Expect(0, 120779, '\p{^xposixlower}', "");
Expect(0, 120779, '\P{xposixlower}', "");
Expect(1, 120779, '\P{^xposixlower}', "");
Expect(0, 120780, '\p{xposixlower}', "");
Expect(1, 120780, '\p{^xposixlower}', "");
Expect(1, 120780, '\P{xposixlower}', "");
Expect(0, 120780, '\P{^xposixlower}', "");
Expect(1, 120779, '\p{	 XPosixLower}', "");
Expect(0, 120779, '\p{^	 XPosixLower}', "");
Expect(0, 120779, '\P{	 XPosixLower}', "");
Expect(1, 120779, '\P{^	 XPosixLower}', "");
Expect(0, 120780, '\p{	 XPosixLower}', "");
Expect(1, 120780, '\p{^	 XPosixLower}', "");
Expect(1, 120780, '\P{	 XPosixLower}', "");
Expect(0, 120780, '\P{^	 XPosixLower}', "");
Error('\p{_-Is_Lower/a/}');
Error('\P{_-Is_Lower/a/}');
Expect(1, 120779, '\p{islower}', "");
Expect(0, 120779, '\p{^islower}', "");
Expect(0, 120779, '\P{islower}', "");
Expect(1, 120779, '\P{^islower}', "");
Expect(0, 120780, '\p{islower}', "");
Expect(1, 120780, '\p{^islower}', "");
Expect(1, 120780, '\P{islower}', "");
Expect(0, 120780, '\P{^islower}', "");
Expect(1, 120779, '\p{__is_lower}', "");
Expect(0, 120779, '\p{^__is_lower}', "");
Expect(0, 120779, '\P{__is_lower}', "");
Expect(1, 120779, '\P{^__is_lower}', "");
Expect(0, 120780, '\p{__is_lower}', "");
Expect(1, 120780, '\p{^__is_lower}', "");
Expect(1, 120780, '\P{__is_lower}', "");
Expect(0, 120780, '\P{^__is_lower}', "");
Error('\p{/a/ 	Is_XPosixLower}');
Error('\P{/a/ 	Is_XPosixLower}');
Expect(1, 120779, '\p{isxposixlower}', "");
Expect(0, 120779, '\p{^isxposixlower}', "");
Expect(0, 120779, '\P{isxposixlower}', "");
Expect(1, 120779, '\P{^isxposixlower}', "");
Expect(0, 120780, '\p{isxposixlower}', "");
Expect(1, 120780, '\p{^isxposixlower}', "");
Expect(1, 120780, '\P{isxposixlower}', "");
Expect(0, 120780, '\P{^isxposixlower}', "");
Expect(1, 120779, '\p{	_is_XPOSIXLOWER}', "");
Expect(0, 120779, '\p{^	_is_XPOSIXLOWER}', "");
Expect(0, 120779, '\P{	_is_XPOSIXLOWER}', "");
Expect(1, 120779, '\P{^	_is_XPOSIXLOWER}', "");
Expect(0, 120780, '\p{	_is_XPOSIXLOWER}', "");
Expect(1, 120780, '\p{^	_is_XPOSIXLOWER}', "");
Expect(1, 120780, '\P{	_is_XPOSIXLOWER}', "");
Expect(0, 120780, '\P{^	_is_XPOSIXLOWER}', "");
Error('\p{/a/-Lowercase}');
Error('\P{/a/-Lowercase}');
Expect(1, 120779, '\p{lowercase}', "");
Expect(0, 120779, '\p{^lowercase}', "");
Expect(0, 120779, '\P{lowercase}', "");
Expect(1, 120779, '\P{^lowercase}', "");
Expect(0, 120780, '\p{lowercase}', "");
Expect(1, 120780, '\p{^lowercase}', "");
Expect(1, 120780, '\P{lowercase}', "");
Expect(0, 120780, '\P{^lowercase}', "");
Expect(1, 120779, '\p{_ Lowercase}', "");
Expect(0, 120779, '\p{^_ Lowercase}', "");
Expect(0, 120779, '\P{_ Lowercase}', "");
Expect(1, 120779, '\P{^_ Lowercase}', "");
Expect(0, 120780, '\p{_ Lowercase}', "");
Expect(1, 120780, '\p{^_ Lowercase}', "");
Expect(1, 120780, '\P{_ Lowercase}', "");
Expect(0, 120780, '\P{^_ Lowercase}', "");
Error('\p{	-Is_Lowercase:=}');
Error('\P{	-Is_Lowercase:=}');
Expect(1, 120779, '\p{islowercase}', "");
Expect(0, 120779, '\p{^islowercase}', "");
Expect(0, 120779, '\P{islowercase}', "");
Expect(1, 120779, '\P{^islowercase}', "");
Expect(0, 120780, '\p{islowercase}', "");
Expect(1, 120780, '\p{^islowercase}', "");
Expect(1, 120780, '\P{islowercase}', "");
Expect(0, 120780, '\P{^islowercase}', "");
Expect(1, 120779, '\p{-_IS_Lowercase}', "");
Expect(0, 120779, '\p{^-_IS_Lowercase}', "");
Expect(0, 120779, '\P{-_IS_Lowercase}', "");
Expect(1, 120779, '\P{^-_IS_Lowercase}', "");
Expect(0, 120780, '\p{-_IS_Lowercase}', "");
Expect(1, 120780, '\p{^-_IS_Lowercase}', "");
Expect(1, 120780, '\P{-_IS_Lowercase}', "");
Expect(0, 120780, '\P{^-_IS_Lowercase}', "");
Error('\p{ :=Lowercase_Letter}');
Error('\P{ :=Lowercase_Letter}');
Expect(1, 120779, '\p{lowercaseletter}', "");
Expect(0, 120779, '\p{^lowercaseletter}', "");
Expect(0, 120779, '\P{lowercaseletter}', "");
Expect(1, 120779, '\P{^lowercaseletter}', "");
Expect(0, 120780, '\p{lowercaseletter}', "");
Expect(1, 120780, '\p{^lowercaseletter}', "");
Expect(1, 120780, '\P{lowercaseletter}', "");
Expect(0, 120780, '\P{^lowercaseletter}', "");
Expect(1, 120779, '\p{_lowercase_letter}', "");
Expect(0, 120779, '\p{^_lowercase_letter}', "");
Expect(0, 120779, '\P{_lowercase_letter}', "");
Expect(1, 120779, '\P{^_lowercase_letter}', "");
Expect(0, 120780, '\p{_lowercase_letter}', "");
Expect(1, 120780, '\p{^_lowercase_letter}', "");
Expect(1, 120780, '\P{_lowercase_letter}', "");
Expect(0, 120780, '\P{^_lowercase_letter}', "");
Error('\p{	 is_lowercase_Letter:=}');
Error('\P{	 is_lowercase_Letter:=}');
Expect(1, 120779, '\p{islowercaseletter}', "");
Expect(0, 120779, '\p{^islowercaseletter}', "");
Expect(0, 120779, '\P{islowercaseletter}', "");
Expect(1, 120779, '\P{^islowercaseletter}', "");
Expect(0, 120780, '\p{islowercaseletter}', "");
Expect(1, 120780, '\p{^islowercaseletter}', "");
Expect(1, 120780, '\P{islowercaseletter}', "");
Expect(0, 120780, '\P{^islowercaseletter}', "");
Expect(1, 120779, '\p{_Is_lowercase_Letter}', "");
Expect(0, 120779, '\p{^_Is_lowercase_Letter}', "");
Expect(0, 120779, '\P{_Is_lowercase_Letter}', "");
Expect(1, 120779, '\P{^_Is_lowercase_Letter}', "");
Expect(0, 120780, '\p{_Is_lowercase_Letter}', "");
Expect(1, 120780, '\p{^_Is_lowercase_Letter}', "");
Expect(1, 120780, '\P{_Is_lowercase_Letter}', "");
Expect(0, 120780, '\P{^_Is_lowercase_Letter}', "");
Error('\p{-	LL/a/}');
Error('\P{-	LL/a/}');
Expect(1, 120779, '\p{ll}', "");
Expect(0, 120779, '\p{^ll}', "");
Expect(0, 120779, '\P{ll}', "");
Expect(1, 120779, '\P{^ll}', "");
Expect(0, 120780, '\p{ll}', "");
Expect(1, 120780, '\p{^ll}', "");
Expect(1, 120780, '\P{ll}', "");
Expect(0, 120780, '\P{^ll}', "");
Expect(1, 120779, '\p{-ll}', "");
Expect(0, 120779, '\p{^-ll}', "");
Expect(0, 120779, '\P{-ll}', "");
Expect(1, 120779, '\P{^-ll}', "");
Expect(0, 120780, '\p{-ll}', "");
Expect(1, 120780, '\p{^-ll}', "");
Expect(1, 120780, '\P{-ll}', "");
Expect(0, 120780, '\P{^-ll}', "");
Error('\p{	IS_Ll:=}');
Error('\P{	IS_Ll:=}');
Expect(1, 120779, '\p{isll}', "");
Expect(0, 120779, '\p{^isll}', "");
Expect(0, 120779, '\P{isll}', "");
Expect(1, 120779, '\P{^isll}', "");
Expect(0, 120780, '\p{isll}', "");
Expect(1, 120780, '\p{^isll}', "");
Expect(1, 120780, '\P{isll}', "");
Expect(0, 120780, '\P{^isll}', "");
Expect(1, 120779, '\p{-Is_Ll}', "");
Expect(0, 120779, '\p{^-Is_Ll}', "");
Expect(0, 120779, '\P{-Is_Ll}', "");
Expect(1, 120779, '\P{^-Is_Ll}', "");
Expect(0, 120780, '\p{-Is_Ll}', "");
Expect(1, 120780, '\p{^-Is_Ll}', "");
Expect(1, 120780, '\P{-Is_Ll}', "");
Expect(0, 120780, '\P{^-Is_Ll}', "");
Error('\p{:=	_Lycian}');
Error('\P{:=	_Lycian}');
Expect(1, 66204, '\p{lycian}', "");
Expect(0, 66204, '\p{^lycian}', "");
Expect(0, 66204, '\P{lycian}', "");
Expect(1, 66204, '\P{^lycian}', "");
Expect(0, 66205, '\p{lycian}', "");
Expect(1, 66205, '\p{^lycian}', "");
Expect(1, 66205, '\P{lycian}', "");
Expect(0, 66205, '\P{^lycian}', "");
Expect(1, 66204, '\p{__Lycian}', "");
Expect(0, 66204, '\p{^__Lycian}', "");
Expect(0, 66204, '\P{__Lycian}', "");
Expect(1, 66204, '\P{^__Lycian}', "");
Expect(0, 66205, '\p{__Lycian}', "");
Expect(1, 66205, '\p{^__Lycian}', "");
Expect(1, 66205, '\P{__Lycian}', "");
Expect(0, 66205, '\P{^__Lycian}', "");
Error('\p{_:=Is_lycian}');
Error('\P{_:=Is_lycian}');
Expect(1, 66204, '\p{islycian}', "");
Expect(0, 66204, '\p{^islycian}', "");
Expect(0, 66204, '\P{islycian}', "");
Expect(1, 66204, '\P{^islycian}', "");
Expect(0, 66205, '\p{islycian}', "");
Expect(1, 66205, '\p{^islycian}', "");
Expect(1, 66205, '\P{islycian}', "");
Expect(0, 66205, '\P{^islycian}', "");
Expect(1, 66204, '\p{  Is_Lycian}', "");
Expect(0, 66204, '\p{^  Is_Lycian}', "");
Expect(0, 66204, '\P{  Is_Lycian}', "");
Expect(1, 66204, '\P{^  Is_Lycian}', "");
Expect(0, 66205, '\p{  Is_Lycian}', "");
Expect(1, 66205, '\p{^  Is_Lycian}', "");
Expect(1, 66205, '\P{  Is_Lycian}', "");
Expect(0, 66205, '\P{^  Is_Lycian}', "");
Error('\p{/a/  Lyci}');
Error('\P{/a/  Lyci}');
Expect(1, 66204, '\p{lyci}', "");
Expect(0, 66204, '\p{^lyci}', "");
Expect(0, 66204, '\P{lyci}', "");
Expect(1, 66204, '\P{^lyci}', "");
Expect(0, 66205, '\p{lyci}', "");
Expect(1, 66205, '\p{^lyci}', "");
Expect(1, 66205, '\P{lyci}', "");
Expect(0, 66205, '\P{^lyci}', "");
Expect(1, 66204, '\p{-Lyci}', "");
Expect(0, 66204, '\p{^-Lyci}', "");
Expect(0, 66204, '\P{-Lyci}', "");
Expect(1, 66204, '\P{^-Lyci}', "");
Expect(0, 66205, '\p{-Lyci}', "");
Expect(1, 66205, '\p{^-Lyci}', "");
Expect(1, 66205, '\P{-Lyci}', "");
Expect(0, 66205, '\P{^-Lyci}', "");
Error('\p{-:=Is_Lyci}');
Error('\P{-:=Is_Lyci}');
Expect(1, 66204, '\p{islyci}', "");
Expect(0, 66204, '\p{^islyci}', "");
Expect(0, 66204, '\P{islyci}', "");
Expect(1, 66204, '\P{^islyci}', "");
Expect(0, 66205, '\p{islyci}', "");
Expect(1, 66205, '\p{^islyci}', "");
Expect(1, 66205, '\P{islyci}', "");
Expect(0, 66205, '\P{^islyci}', "");
Expect(1, 66204, '\p{		IS_Lyci}', "");
Expect(0, 66204, '\p{^		IS_Lyci}', "");
Expect(0, 66204, '\P{		IS_Lyci}', "");
Expect(1, 66204, '\P{^		IS_Lyci}', "");
Expect(0, 66205, '\p{		IS_Lyci}', "");
Expect(1, 66205, '\p{^		IS_Lyci}', "");
Expect(1, 66205, '\P{		IS_Lyci}', "");
Expect(0, 66205, '\P{^		IS_Lyci}', "");
Error('\p{_ Lydian:=}');
Error('\P{_ Lydian:=}');
Expect(1, 67903, '\p{lydian}', "");
Expect(0, 67903, '\p{^lydian}', "");
Expect(0, 67903, '\P{lydian}', "");
Expect(1, 67903, '\P{^lydian}', "");
Expect(0, 67904, '\p{lydian}', "");
Expect(1, 67904, '\p{^lydian}', "");
Expect(1, 67904, '\P{lydian}', "");
Expect(0, 67904, '\P{^lydian}', "");
Expect(1, 67903, '\p{	LYDIAN}', "");
Expect(0, 67903, '\p{^	LYDIAN}', "");
Expect(0, 67903, '\P{	LYDIAN}', "");
Expect(1, 67903, '\P{^	LYDIAN}', "");
Expect(0, 67904, '\p{	LYDIAN}', "");
Expect(1, 67904, '\p{^	LYDIAN}', "");
Expect(1, 67904, '\P{	LYDIAN}', "");
Expect(0, 67904, '\P{^	LYDIAN}', "");
Error('\p{--Is_Lydian/a/}');
Error('\P{--Is_Lydian/a/}');
Expect(1, 67903, '\p{islydian}', "");
Expect(0, 67903, '\p{^islydian}', "");
Expect(0, 67903, '\P{islydian}', "");
Expect(1, 67903, '\P{^islydian}', "");
Expect(0, 67904, '\p{islydian}', "");
Expect(1, 67904, '\p{^islydian}', "");
Expect(1, 67904, '\P{islydian}', "");
Expect(0, 67904, '\P{^islydian}', "");
Expect(1, 67903, '\p{ is_LYDIAN}', "");
Expect(0, 67903, '\p{^ is_LYDIAN}', "");
Expect(0, 67903, '\P{ is_LYDIAN}', "");
Expect(1, 67903, '\P{^ is_LYDIAN}', "");
Expect(0, 67904, '\p{ is_LYDIAN}', "");
Expect(1, 67904, '\p{^ is_LYDIAN}', "");
Expect(1, 67904, '\P{ is_LYDIAN}', "");
Expect(0, 67904, '\P{^ is_LYDIAN}', "");
Error('\p{ :=lydi}');
Error('\P{ :=lydi}');
Expect(1, 67903, '\p{lydi}', "");
Expect(0, 67903, '\p{^lydi}', "");
Expect(0, 67903, '\P{lydi}', "");
Expect(1, 67903, '\P{^lydi}', "");
Expect(0, 67904, '\p{lydi}', "");
Expect(1, 67904, '\p{^lydi}', "");
Expect(1, 67904, '\P{lydi}', "");
Expect(0, 67904, '\P{^lydi}', "");
Expect(1, 67903, '\p{	lydi}', "");
Expect(0, 67903, '\p{^	lydi}', "");
Expect(0, 67903, '\P{	lydi}', "");
Expect(1, 67903, '\P{^	lydi}', "");
Expect(0, 67904, '\p{	lydi}', "");
Expect(1, 67904, '\p{^	lydi}', "");
Expect(1, 67904, '\P{	lydi}', "");
Expect(0, 67904, '\P{^	lydi}', "");
Error('\p{ Is_Lydi:=}');
Error('\P{ Is_Lydi:=}');
Expect(1, 67903, '\p{islydi}', "");
Expect(0, 67903, '\p{^islydi}', "");
Expect(0, 67903, '\P{islydi}', "");
Expect(1, 67903, '\P{^islydi}', "");
Expect(0, 67904, '\p{islydi}', "");
Expect(1, 67904, '\p{^islydi}', "");
Expect(1, 67904, '\P{islydi}', "");
Expect(0, 67904, '\P{^islydi}', "");
Expect(1, 67903, '\p{		is_Lydi}', "");
Expect(0, 67903, '\p{^		is_Lydi}', "");
Expect(0, 67903, '\P{		is_Lydi}', "");
Expect(1, 67903, '\P{^		is_Lydi}', "");
Expect(0, 67904, '\p{		is_Lydi}', "");
Expect(1, 67904, '\p{^		is_Lydi}', "");
Expect(1, 67904, '\P{		is_Lydi}', "");
Expect(0, 67904, '\P{^		is_Lydi}', "");
Error('\p{Mahjong_Tiles:=}');
Error('\P{Mahjong_Tiles:=}');
Expect(1, 127023, '\p{mahjongtiles}', "");
Expect(0, 127023, '\p{^mahjongtiles}', "");
Expect(0, 127023, '\P{mahjongtiles}', "");
Expect(1, 127023, '\P{^mahjongtiles}', "");
Expect(0, 127024, '\p{mahjongtiles}', "");
Expect(1, 127024, '\p{^mahjongtiles}', "");
Expect(1, 127024, '\P{mahjongtiles}', "");
Expect(0, 127024, '\P{^mahjongtiles}', "");
Expect(1, 127023, '\p{--mahjong_tiles}', "");
Expect(0, 127023, '\p{^--mahjong_tiles}', "");
Expect(0, 127023, '\P{--mahjong_tiles}', "");
Expect(1, 127023, '\P{^--mahjong_tiles}', "");
Expect(0, 127024, '\p{--mahjong_tiles}', "");
Expect(1, 127024, '\p{^--mahjong_tiles}', "");
Expect(1, 127024, '\P{--mahjong_tiles}', "");
Expect(0, 127024, '\P{^--mahjong_tiles}', "");
Error('\p{	_IS_MAHJONG_TILES/a/}');
Error('\P{	_IS_MAHJONG_TILES/a/}');
Expect(1, 127023, '\p{ismahjongtiles}', "");
Expect(0, 127023, '\p{^ismahjongtiles}', "");
Expect(0, 127023, '\P{ismahjongtiles}', "");
Expect(1, 127023, '\P{^ismahjongtiles}', "");
Expect(0, 127024, '\p{ismahjongtiles}', "");
Expect(1, 127024, '\p{^ismahjongtiles}', "");
Expect(1, 127024, '\P{ismahjongtiles}', "");
Expect(0, 127024, '\P{^ismahjongtiles}', "");
Expect(1, 127023, '\p{-_IS_mahjong_tiles}', "");
Expect(0, 127023, '\p{^-_IS_mahjong_tiles}', "");
Expect(0, 127023, '\P{-_IS_mahjong_tiles}', "");
Expect(1, 127023, '\P{^-_IS_mahjong_tiles}', "");
Expect(0, 127024, '\p{-_IS_mahjong_tiles}', "");
Expect(1, 127024, '\p{^-_IS_mahjong_tiles}', "");
Expect(1, 127024, '\P{-_IS_mahjong_tiles}', "");
Expect(0, 127024, '\P{^-_IS_mahjong_tiles}', "");
Error('\p{	/a/IN_MAHJONG_Tiles}');
Error('\P{	/a/IN_MAHJONG_Tiles}');
Expect(1, 127023, '\p{inmahjongtiles}', "");
Expect(0, 127023, '\p{^inmahjongtiles}', "");
Expect(0, 127023, '\P{inmahjongtiles}', "");
Expect(1, 127023, '\P{^inmahjongtiles}', "");
Expect(0, 127024, '\p{inmahjongtiles}', "");
Expect(1, 127024, '\p{^inmahjongtiles}', "");
Expect(1, 127024, '\P{inmahjongtiles}', "");
Expect(0, 127024, '\P{^inmahjongtiles}', "");
Expect(1, 127023, '\p{_in_MAHJONG_Tiles}', "");
Expect(0, 127023, '\p{^_in_MAHJONG_Tiles}', "");
Expect(0, 127023, '\P{_in_MAHJONG_Tiles}', "");
Expect(1, 127023, '\P{^_in_MAHJONG_Tiles}', "");
Expect(0, 127024, '\p{_in_MAHJONG_Tiles}', "");
Expect(1, 127024, '\p{^_in_MAHJONG_Tiles}', "");
Expect(1, 127024, '\P{_in_MAHJONG_Tiles}', "");
Expect(0, 127024, '\P{^_in_MAHJONG_Tiles}', "");
Error('\p{__Mahjong/a/}');
Error('\P{__Mahjong/a/}');
Expect(1, 127023, '\p{mahjong}', "");
Expect(0, 127023, '\p{^mahjong}', "");
Expect(0, 127023, '\P{mahjong}', "");
Expect(1, 127023, '\P{^mahjong}', "");
Expect(0, 127024, '\p{mahjong}', "");
Expect(1, 127024, '\p{^mahjong}', "");
Expect(1, 127024, '\P{mahjong}', "");
Expect(0, 127024, '\P{^mahjong}', "");
Expect(1, 127023, '\p{ 	MAHJONG}', "");
Expect(0, 127023, '\p{^ 	MAHJONG}', "");
Expect(0, 127023, '\P{ 	MAHJONG}', "");
Expect(1, 127023, '\P{^ 	MAHJONG}', "");
Expect(0, 127024, '\p{ 	MAHJONG}', "");
Expect(1, 127024, '\p{^ 	MAHJONG}', "");
Expect(1, 127024, '\P{ 	MAHJONG}', "");
Expect(0, 127024, '\P{^ 	MAHJONG}', "");
Error('\p{/a/ 	IS_mahjong}');
Error('\P{/a/ 	IS_mahjong}');
Expect(1, 127023, '\p{ismahjong}', "");
Expect(0, 127023, '\p{^ismahjong}', "");
Expect(0, 127023, '\P{ismahjong}', "");
Expect(1, 127023, '\P{^ismahjong}', "");
Expect(0, 127024, '\p{ismahjong}', "");
Expect(1, 127024, '\p{^ismahjong}', "");
Expect(1, 127024, '\P{ismahjong}', "");
Expect(0, 127024, '\P{^ismahjong}', "");
Expect(1, 127023, '\p{_ Is_Mahjong}', "");
Expect(0, 127023, '\p{^_ Is_Mahjong}', "");
Expect(0, 127023, '\P{_ Is_Mahjong}', "");
Expect(1, 127023, '\P{^_ Is_Mahjong}', "");
Expect(0, 127024, '\p{_ Is_Mahjong}', "");
Expect(1, 127024, '\p{^_ Is_Mahjong}', "");
Expect(1, 127024, '\P{_ Is_Mahjong}', "");
Expect(0, 127024, '\P{^_ Is_Mahjong}', "");
Error('\p{	/a/In_MAHJONG}');
Error('\P{	/a/In_MAHJONG}');
Expect(1, 127023, '\p{inmahjong}', "");
Expect(0, 127023, '\p{^inmahjong}', "");
Expect(0, 127023, '\P{inmahjong}', "");
Expect(1, 127023, '\P{^inmahjong}', "");
Expect(0, 127024, '\p{inmahjong}', "");
Expect(1, 127024, '\p{^inmahjong}', "");
Expect(1, 127024, '\P{inmahjong}', "");
Expect(0, 127024, '\P{^inmahjong}', "");
Expect(1, 127023, '\p{--in_mahjong}', "");
Expect(0, 127023, '\p{^--in_mahjong}', "");
Expect(0, 127023, '\P{--in_mahjong}', "");
Expect(1, 127023, '\P{^--in_mahjong}', "");
Expect(0, 127024, '\p{--in_mahjong}', "");
Expect(1, 127024, '\p{^--in_mahjong}', "");
Expect(1, 127024, '\P{--in_mahjong}', "");
Expect(0, 127024, '\P{^--in_mahjong}', "");
Error('\p{/a/ _malayalam}');
Error('\P{/a/ _malayalam}');
Expect(1, 3455, '\p{malayalam}', "");
Expect(0, 3455, '\p{^malayalam}', "");
Expect(0, 3455, '\P{malayalam}', "");
Expect(1, 3455, '\P{^malayalam}', "");
Expect(0, 3456, '\p{malayalam}', "");
Expect(1, 3456, '\p{^malayalam}', "");
Expect(1, 3456, '\P{malayalam}', "");
Expect(0, 3456, '\P{^malayalam}', "");
Expect(1, 3455, '\p{-malayalam}', "");
Expect(0, 3455, '\p{^-malayalam}', "");
Expect(0, 3455, '\P{-malayalam}', "");
Expect(1, 3455, '\P{^-malayalam}', "");
Expect(0, 3456, '\p{-malayalam}', "");
Expect(1, 3456, '\p{^-malayalam}', "");
Expect(1, 3456, '\P{-malayalam}', "");
Expect(0, 3456, '\P{^-malayalam}', "");
Error('\p{/a/-Is_MALAYALAM}');
Error('\P{/a/-Is_MALAYALAM}');
Expect(1, 3455, '\p{ismalayalam}', "");
Expect(0, 3455, '\p{^ismalayalam}', "");
Expect(0, 3455, '\P{ismalayalam}', "");
Expect(1, 3455, '\P{^ismalayalam}', "");
Expect(0, 3456, '\p{ismalayalam}', "");
Expect(1, 3456, '\p{^ismalayalam}', "");
Expect(1, 3456, '\P{ismalayalam}', "");
Expect(0, 3456, '\P{^ismalayalam}', "");
Expect(1, 3455, '\p{ _Is_malayalam}', "");
Expect(0, 3455, '\p{^ _Is_malayalam}', "");
Expect(0, 3455, '\P{ _Is_malayalam}', "");
Expect(1, 3455, '\P{^ _Is_malayalam}', "");
Expect(0, 3456, '\p{ _Is_malayalam}', "");
Expect(1, 3456, '\p{^ _Is_malayalam}', "");
Expect(1, 3456, '\P{ _Is_malayalam}', "");
Expect(0, 3456, '\P{^ _Is_malayalam}', "");
Error('\p{/a/ mlym}');
Error('\P{/a/ mlym}');
Expect(1, 3455, '\p{mlym}', "");
Expect(0, 3455, '\p{^mlym}', "");
Expect(0, 3455, '\P{mlym}', "");
Expect(1, 3455, '\P{^mlym}', "");
Expect(0, 3456, '\p{mlym}', "");
Expect(1, 3456, '\p{^mlym}', "");
Expect(1, 3456, '\P{mlym}', "");
Expect(0, 3456, '\P{^mlym}', "");
Expect(1, 3455, '\p{_ Mlym}', "");
Expect(0, 3455, '\p{^_ Mlym}', "");
Expect(0, 3455, '\P{_ Mlym}', "");
Expect(1, 3455, '\P{^_ Mlym}', "");
Expect(0, 3456, '\p{_ Mlym}', "");
Expect(1, 3456, '\p{^_ Mlym}', "");
Expect(1, 3456, '\P{_ Mlym}', "");
Expect(0, 3456, '\P{^_ Mlym}', "");
Error('\p{/a/_is_MLYM}');
Error('\P{/a/_is_MLYM}');
Expect(1, 3455, '\p{ismlym}', "");
Expect(0, 3455, '\p{^ismlym}', "");
Expect(0, 3455, '\P{ismlym}', "");
Expect(1, 3455, '\P{^ismlym}', "");
Expect(0, 3456, '\p{ismlym}', "");
Expect(1, 3456, '\p{^ismlym}', "");
Expect(1, 3456, '\P{ismlym}', "");
Expect(0, 3456, '\P{^ismlym}', "");
Expect(1, 3455, '\p{-_is_Mlym}', "");
Expect(0, 3455, '\p{^-_is_Mlym}', "");
Expect(0, 3455, '\P{-_is_Mlym}', "");
Expect(1, 3455, '\P{^-_is_Mlym}', "");
Expect(0, 3456, '\p{-_is_Mlym}', "");
Expect(1, 3456, '\p{^-_is_Mlym}', "");
Expect(1, 3456, '\P{-_is_Mlym}', "");
Expect(0, 3456, '\P{^-_is_Mlym}', "");
Error('\p{:=		MANDAIC}');
Error('\P{:=		MANDAIC}');
Expect(1, 2142, '\p{mandaic}', "");
Expect(0, 2142, '\p{^mandaic}', "");
Expect(0, 2142, '\P{mandaic}', "");
Expect(1, 2142, '\P{^mandaic}', "");
Expect(0, 2143, '\p{mandaic}', "");
Expect(1, 2143, '\p{^mandaic}', "");
Expect(1, 2143, '\P{mandaic}', "");
Expect(0, 2143, '\P{^mandaic}', "");
Expect(1, 2142, '\p{ _MANDAIC}', "");
Expect(0, 2142, '\p{^ _MANDAIC}', "");
Expect(0, 2142, '\P{ _MANDAIC}', "");
Expect(1, 2142, '\P{^ _MANDAIC}', "");
Expect(0, 2143, '\p{ _MANDAIC}', "");
Expect(1, 2143, '\p{^ _MANDAIC}', "");
Expect(1, 2143, '\P{ _MANDAIC}', "");
Expect(0, 2143, '\P{^ _MANDAIC}', "");
Error('\p{_-Is_Mandaic/a/}');
Error('\P{_-Is_Mandaic/a/}');
Expect(1, 2142, '\p{ismandaic}', "");
Expect(0, 2142, '\p{^ismandaic}', "");
Expect(0, 2142, '\P{ismandaic}', "");
Expect(1, 2142, '\P{^ismandaic}', "");
Expect(0, 2143, '\p{ismandaic}', "");
Expect(1, 2143, '\p{^ismandaic}', "");
Expect(1, 2143, '\P{ismandaic}', "");
Expect(0, 2143, '\P{^ismandaic}', "");
Expect(1, 2142, '\p{_IS_Mandaic}', "");
Expect(0, 2142, '\p{^_IS_Mandaic}', "");
Expect(0, 2142, '\P{_IS_Mandaic}', "");
Expect(1, 2142, '\P{^_IS_Mandaic}', "");
Expect(0, 2143, '\p{_IS_Mandaic}', "");
Expect(1, 2143, '\p{^_IS_Mandaic}', "");
Expect(1, 2143, '\P{_IS_Mandaic}', "");
Expect(0, 2143, '\P{^_IS_Mandaic}', "");
Error('\p{-MAND:=}');
Error('\P{-MAND:=}');
Expect(1, 2142, '\p{mand}', "");
Expect(0, 2142, '\p{^mand}', "");
Expect(0, 2142, '\P{mand}', "");
Expect(1, 2142, '\P{^mand}', "");
Expect(0, 2143, '\p{mand}', "");
Expect(1, 2143, '\p{^mand}', "");
Expect(1, 2143, '\P{mand}', "");
Expect(0, 2143, '\P{^mand}', "");
Expect(1, 2142, '\p{	mand}', "");
Expect(0, 2142, '\p{^	mand}', "");
Expect(0, 2142, '\P{	mand}', "");
Expect(1, 2142, '\P{^	mand}', "");
Expect(0, 2143, '\p{	mand}', "");
Expect(1, 2143, '\p{^	mand}', "");
Expect(1, 2143, '\P{	mand}', "");
Expect(0, 2143, '\P{^	mand}', "");
Error('\p{	_Is_Mand:=}');
Error('\P{	_Is_Mand:=}');
Expect(1, 2142, '\p{ismand}', "");
Expect(0, 2142, '\p{^ismand}', "");
Expect(0, 2142, '\P{ismand}', "");
Expect(1, 2142, '\P{^ismand}', "");
Expect(0, 2143, '\p{ismand}', "");
Expect(1, 2143, '\p{^ismand}', "");
Expect(1, 2143, '\P{ismand}', "");
Expect(0, 2143, '\P{^ismand}', "");
Expect(1, 2142, '\p{		Is_Mand}', "");
Expect(0, 2142, '\p{^		Is_Mand}', "");
Expect(0, 2142, '\P{		Is_Mand}', "");
Expect(1, 2142, '\P{^		Is_Mand}', "");
Expect(0, 2143, '\p{		Is_Mand}', "");
Expect(1, 2143, '\p{^		Is_Mand}', "");
Expect(1, 2143, '\P{		Is_Mand}', "");
Expect(0, 2143, '\P{^		Is_Mand}', "");
Error('\p{:=_-mark}');
Error('\P{:=_-mark}');
Expect(1, 917999, '\p{mark}', "");
Expect(0, 917999, '\p{^mark}', "");
Expect(0, 917999, '\P{mark}', "");
Expect(1, 917999, '\P{^mark}', "");
Expect(0, 918000, '\p{mark}', "");
Expect(1, 918000, '\p{^mark}', "");
Expect(1, 918000, '\P{mark}', "");
Expect(0, 918000, '\P{^mark}', "");
Expect(1, 917999, '\p{Mark}', "");
Expect(0, 917999, '\p{^Mark}', "");
Expect(0, 917999, '\P{Mark}', "");
Expect(1, 917999, '\P{^Mark}', "");
Expect(0, 918000, '\p{Mark}', "");
Expect(1, 918000, '\p{^Mark}', "");
Expect(1, 918000, '\P{Mark}', "");
Expect(0, 918000, '\P{^Mark}', "");
Error('\p{- Is_MARK:=}');
Error('\P{- Is_MARK:=}');
Expect(1, 917999, '\p{ismark}', "");
Expect(0, 917999, '\p{^ismark}', "");
Expect(0, 917999, '\P{ismark}', "");
Expect(1, 917999, '\P{^ismark}', "");
Expect(0, 918000, '\p{ismark}', "");
Expect(1, 918000, '\p{^ismark}', "");
Expect(1, 918000, '\P{ismark}', "");
Expect(0, 918000, '\P{^ismark}', "");
Expect(1, 917999, '\p{--Is_Mark}', "");
Expect(0, 917999, '\p{^--Is_Mark}', "");
Expect(0, 917999, '\P{--Is_Mark}', "");
Expect(1, 917999, '\P{^--Is_Mark}', "");
Expect(0, 918000, '\p{--Is_Mark}', "");
Expect(1, 918000, '\p{^--Is_Mark}', "");
Expect(1, 918000, '\P{--Is_Mark}', "");
Expect(0, 918000, '\P{^--Is_Mark}', "");
Error('\p{:=		M}');
Error('\P{:=		M}');
Expect(1, 917999, '\p{m}', "");
Expect(0, 917999, '\p{^m}', "");
Expect(0, 917999, '\P{m}', "");
Expect(1, 917999, '\P{^m}', "");
Expect(0, 918000, '\p{m}', "");
Expect(1, 918000, '\p{^m}', "");
Expect(1, 918000, '\P{m}', "");
Expect(0, 918000, '\P{^m}', "");
Expect(1, 917999, '\p{_	m}', "");
Expect(0, 917999, '\p{^_	m}', "");
Expect(0, 917999, '\P{_	m}', "");
Expect(1, 917999, '\P{^_	m}', "");
Expect(0, 918000, '\p{_	m}', "");
Expect(1, 918000, '\p{^_	m}', "");
Expect(1, 918000, '\P{_	m}', "");
Expect(0, 918000, '\P{^_	m}', "");
Error('\p{:=__Is_m}');
Error('\P{:=__Is_m}');
Expect(1, 917999, '\p{ism}', "");
Expect(0, 917999, '\p{^ism}', "");
Expect(0, 917999, '\P{ism}', "");
Expect(1, 917999, '\P{^ism}', "");
Expect(0, 918000, '\p{ism}', "");
Expect(1, 918000, '\p{^ism}', "");
Expect(1, 918000, '\P{ism}', "");
Expect(0, 918000, '\P{^ism}', "");
Expect(1, 917999, '\p{ 	Is_M}', "");
Expect(0, 917999, '\p{^ 	Is_M}', "");
Expect(0, 917999, '\P{ 	Is_M}', "");
Expect(1, 917999, '\P{^ 	Is_M}', "");
Expect(0, 918000, '\p{ 	Is_M}', "");
Expect(1, 918000, '\p{^ 	Is_M}', "");
Expect(1, 918000, '\P{ 	Is_M}', "");
Expect(0, 918000, '\P{^ 	Is_M}', "");
Error('\p{:= 	Combining_mark}');
Error('\P{:= 	Combining_mark}');
Expect(1, 917999, '\p{combiningmark}', "");
Expect(0, 917999, '\p{^combiningmark}', "");
Expect(0, 917999, '\P{combiningmark}', "");
Expect(1, 917999, '\P{^combiningmark}', "");
Expect(0, 918000, '\p{combiningmark}', "");
Expect(1, 918000, '\p{^combiningmark}', "");
Expect(1, 918000, '\P{combiningmark}', "");
Expect(0, 918000, '\P{^combiningmark}', "");
Expect(1, 917999, '\p{  COMBINING_Mark}', "");
Expect(0, 917999, '\p{^  COMBINING_Mark}', "");
Expect(0, 917999, '\P{  COMBINING_Mark}', "");
Expect(1, 917999, '\P{^  COMBINING_Mark}', "");
Expect(0, 918000, '\p{  COMBINING_Mark}', "");
Expect(1, 918000, '\p{^  COMBINING_Mark}', "");
Expect(1, 918000, '\P{  COMBINING_Mark}', "");
Expect(0, 918000, '\P{^  COMBINING_Mark}', "");
Error('\p{ _IS_Combining_Mark/a/}');
Error('\P{ _IS_Combining_Mark/a/}');
Expect(1, 917999, '\p{iscombiningmark}', "");
Expect(0, 917999, '\p{^iscombiningmark}', "");
Expect(0, 917999, '\P{iscombiningmark}', "");
Expect(1, 917999, '\P{^iscombiningmark}', "");
Expect(0, 918000, '\p{iscombiningmark}', "");
Expect(1, 918000, '\p{^iscombiningmark}', "");
Expect(1, 918000, '\P{iscombiningmark}', "");
Expect(0, 918000, '\P{^iscombiningmark}', "");
Expect(1, 917999, '\p{_ Is_combining_mark}', "");
Expect(0, 917999, '\p{^_ Is_combining_mark}', "");
Expect(0, 917999, '\P{_ Is_combining_mark}', "");
Expect(1, 917999, '\P{^_ Is_combining_mark}', "");
Expect(0, 918000, '\p{_ Is_combining_mark}', "");
Expect(1, 918000, '\p{^_ Is_combining_mark}', "");
Expect(1, 918000, '\P{_ Is_combining_mark}', "");
Expect(0, 918000, '\P{^_ Is_combining_mark}', "");
Error('\p{:= -MATH}');
Error('\P{:= -MATH}');
Expect(1, 126705, '\p{math}', "");
Expect(0, 126705, '\p{^math}', "");
Expect(0, 126705, '\P{math}', "");
Expect(1, 126705, '\P{^math}', "");
Expect(0, 126706, '\p{math}', "");
Expect(1, 126706, '\p{^math}', "");
Expect(1, 126706, '\P{math}', "");
Expect(0, 126706, '\P{^math}', "");
Expect(1, 126705, '\p{	_math}', "");
Expect(0, 126705, '\p{^	_math}', "");
Expect(0, 126705, '\P{	_math}', "");
Expect(1, 126705, '\P{^	_math}', "");
Expect(0, 126706, '\p{	_math}', "");
Expect(1, 126706, '\p{^	_math}', "");
Expect(1, 126706, '\P{	_math}', "");
Expect(0, 126706, '\P{^	_math}', "");
Error('\p{-/a/Is_MATH}');
Error('\P{-/a/Is_MATH}');
Expect(1, 126705, '\p{ismath}', "");
Expect(0, 126705, '\p{^ismath}', "");
Expect(0, 126705, '\P{ismath}', "");
Expect(1, 126705, '\P{^ismath}', "");
Expect(0, 126706, '\p{ismath}', "");
Expect(1, 126706, '\p{^ismath}', "");
Expect(1, 126706, '\P{ismath}', "");
Expect(0, 126706, '\P{^ismath}', "");
Expect(1, 126705, '\p{_-Is_Math}', "");
Expect(0, 126705, '\p{^_-Is_Math}', "");
Expect(0, 126705, '\P{_-Is_Math}', "");
Expect(1, 126705, '\P{^_-Is_Math}', "");
Expect(0, 126706, '\p{_-Is_Math}', "");
Expect(1, 126706, '\p{^_-Is_Math}', "");
Expect(1, 126706, '\P{_-Is_Math}', "");
Expect(0, 126706, '\P{^_-Is_Math}', "");
Error('\p{/a/-MATH_symbol}');
Error('\P{/a/-MATH_symbol}');
Expect(1, 126705, '\p{mathsymbol}', "");
Expect(0, 126705, '\p{^mathsymbol}', "");
Expect(0, 126705, '\P{mathsymbol}', "");
Expect(1, 126705, '\P{^mathsymbol}', "");
Expect(0, 126706, '\p{mathsymbol}', "");
Expect(1, 126706, '\p{^mathsymbol}', "");
Expect(1, 126706, '\P{mathsymbol}', "");
Expect(0, 126706, '\P{^mathsymbol}', "");
Expect(1, 126705, '\p{_ Math_Symbol}', "");
Expect(0, 126705, '\p{^_ Math_Symbol}', "");
Expect(0, 126705, '\P{_ Math_Symbol}', "");
Expect(1, 126705, '\P{^_ Math_Symbol}', "");
Expect(0, 126706, '\p{_ Math_Symbol}', "");
Expect(1, 126706, '\p{^_ Math_Symbol}', "");
Expect(1, 126706, '\P{_ Math_Symbol}', "");
Expect(0, 126706, '\P{^_ Math_Symbol}', "");
Error('\p{ /a/is_math_SYMBOL}');
Error('\P{ /a/is_math_SYMBOL}');
Expect(1, 126705, '\p{ismathsymbol}', "");
Expect(0, 126705, '\p{^ismathsymbol}', "");
Expect(0, 126705, '\P{ismathsymbol}', "");
Expect(1, 126705, '\P{^ismathsymbol}', "");
Expect(0, 126706, '\p{ismathsymbol}', "");
Expect(1, 126706, '\p{^ismathsymbol}', "");
Expect(1, 126706, '\P{ismathsymbol}', "");
Expect(0, 126706, '\P{^ismathsymbol}', "");
Expect(1, 126705, '\p{_-Is_math_Symbol}', "");
Expect(0, 126705, '\p{^_-Is_math_Symbol}', "");
Expect(0, 126705, '\P{_-Is_math_Symbol}', "");
Expect(1, 126705, '\P{^_-Is_math_Symbol}', "");
Expect(0, 126706, '\p{_-Is_math_Symbol}', "");
Expect(1, 126706, '\p{^_-Is_math_Symbol}', "");
Expect(1, 126706, '\P{_-Is_math_Symbol}', "");
Expect(0, 126706, '\P{^_-Is_math_Symbol}', "");
Error('\p{- Sm:=}');
Error('\P{- Sm:=}');
Expect(1, 126705, '\p{sm}', "");
Expect(0, 126705, '\p{^sm}', "");
Expect(0, 126705, '\P{sm}', "");
Expect(1, 126705, '\P{^sm}', "");
Expect(0, 126706, '\p{sm}', "");
Expect(1, 126706, '\p{^sm}', "");
Expect(1, 126706, '\P{sm}', "");
Expect(0, 126706, '\P{^sm}', "");
Expect(1, 126705, '\p{_-Sm}', "");
Expect(0, 126705, '\p{^_-Sm}', "");
Expect(0, 126705, '\P{_-Sm}', "");
Expect(1, 126705, '\P{^_-Sm}', "");
Expect(0, 126706, '\p{_-Sm}', "");
Expect(1, 126706, '\p{^_-Sm}', "");
Expect(1, 126706, '\P{_-Sm}', "");
Expect(0, 126706, '\P{^_-Sm}', "");
Error('\p{--IS_SM:=}');
Error('\P{--IS_SM:=}');
Expect(1, 126705, '\p{issm}', "");
Expect(0, 126705, '\p{^issm}', "");
Expect(0, 126705, '\P{issm}', "");
Expect(1, 126705, '\P{^issm}', "");
Expect(0, 126706, '\p{issm}', "");
Expect(1, 126706, '\p{^issm}', "");
Expect(1, 126706, '\P{issm}', "");
Expect(0, 126706, '\P{^issm}', "");
Expect(1, 126705, '\p{	is_Sm}', "");
Expect(0, 126705, '\p{^	is_Sm}', "");
Expect(0, 126705, '\P{	is_Sm}', "");
Expect(1, 126705, '\P{^	is_Sm}', "");
Expect(0, 126706, '\p{	is_Sm}', "");
Expect(1, 126706, '\p{^	is_Sm}', "");
Expect(1, 126706, '\P{	is_Sm}', "");
Expect(0, 126706, '\P{^	is_Sm}', "");
Error('\p{:=		Mathematical_alphanumeric_SYMBOLS}');
Error('\P{:=		Mathematical_alphanumeric_SYMBOLS}');
Expect(1, 120831, '\p{mathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\p{^mathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\P{mathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\P{^mathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\p{mathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\p{^mathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\P{mathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\P{^mathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\p{- Mathematical_Alphanumeric_SYMBOLS}', "");
Expect(0, 120831, '\p{^- Mathematical_Alphanumeric_SYMBOLS}', "");
Expect(0, 120831, '\P{- Mathematical_Alphanumeric_SYMBOLS}', "");
Expect(1, 120831, '\P{^- Mathematical_Alphanumeric_SYMBOLS}', "");
Expect(0, 120832, '\p{- Mathematical_Alphanumeric_SYMBOLS}', "");
Expect(1, 120832, '\p{^- Mathematical_Alphanumeric_SYMBOLS}', "");
Expect(1, 120832, '\P{- Mathematical_Alphanumeric_SYMBOLS}', "");
Expect(0, 120832, '\P{^- Mathematical_Alphanumeric_SYMBOLS}', "");
Error('\p{__IS_mathematical_Alphanumeric_SYMBOLS:=}');
Error('\P{__IS_mathematical_Alphanumeric_SYMBOLS:=}');
Expect(1, 120831, '\p{ismathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\p{^ismathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\P{ismathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\P{^ismathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\p{ismathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\p{^ismathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\P{ismathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\P{^ismathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\p{	IS_Mathematical_Alphanumeric_Symbols}', "");
Expect(0, 120831, '\p{^	IS_Mathematical_Alphanumeric_Symbols}', "");
Expect(0, 120831, '\P{	IS_Mathematical_Alphanumeric_Symbols}', "");
Expect(1, 120831, '\P{^	IS_Mathematical_Alphanumeric_Symbols}', "");
Expect(0, 120832, '\p{	IS_Mathematical_Alphanumeric_Symbols}', "");
Expect(1, 120832, '\p{^	IS_Mathematical_Alphanumeric_Symbols}', "");
Expect(1, 120832, '\P{	IS_Mathematical_Alphanumeric_Symbols}', "");
Expect(0, 120832, '\P{^	IS_Mathematical_Alphanumeric_Symbols}', "");
Error('\p{--IN_Mathematical_Alphanumeric_symbols:=}');
Error('\P{--IN_Mathematical_Alphanumeric_symbols:=}');
Expect(1, 120831, '\p{inmathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\p{^inmathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\P{inmathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\P{^inmathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\p{inmathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\p{^inmathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\P{inmathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\P{^inmathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\p{	In_Mathematical_ALPHANUMERIC_Symbols}', "");
Expect(0, 120831, '\p{^	In_Mathematical_ALPHANUMERIC_Symbols}', "");
Expect(0, 120831, '\P{	In_Mathematical_ALPHANUMERIC_Symbols}', "");
Expect(1, 120831, '\P{^	In_Mathematical_ALPHANUMERIC_Symbols}', "");
Expect(0, 120832, '\p{	In_Mathematical_ALPHANUMERIC_Symbols}', "");
Expect(1, 120832, '\p{^	In_Mathematical_ALPHANUMERIC_Symbols}', "");
Expect(1, 120832, '\P{	In_Mathematical_ALPHANUMERIC_Symbols}', "");
Expect(0, 120832, '\P{^	In_Mathematical_ALPHANUMERIC_Symbols}', "");
Error('\p{ /a/math_Alphanum}');
Error('\P{ /a/math_Alphanum}');
Expect(1, 120831, '\p{mathalphanum}', "");
Expect(0, 120831, '\p{^mathalphanum}', "");
Expect(0, 120831, '\P{mathalphanum}', "");
Expect(1, 120831, '\P{^mathalphanum}', "");
Expect(0, 120832, '\p{mathalphanum}', "");
Expect(1, 120832, '\p{^mathalphanum}', "");
Expect(1, 120832, '\P{mathalphanum}', "");
Expect(0, 120832, '\P{^mathalphanum}', "");
Expect(1, 120831, '\p{		Math_Alphanum}', "");
Expect(0, 120831, '\p{^		Math_Alphanum}', "");
Expect(0, 120831, '\P{		Math_Alphanum}', "");
Expect(1, 120831, '\P{^		Math_Alphanum}', "");
Expect(0, 120832, '\p{		Math_Alphanum}', "");
Expect(1, 120832, '\p{^		Math_Alphanum}', "");
Expect(1, 120832, '\P{		Math_Alphanum}', "");
Expect(0, 120832, '\P{^		Math_Alphanum}', "");
Error('\p{:=_-is_math_ALPHANUM}');
Error('\P{:=_-is_math_ALPHANUM}');
Expect(1, 120831, '\p{ismathalphanum}', "");
Expect(0, 120831, '\p{^ismathalphanum}', "");
Expect(0, 120831, '\P{ismathalphanum}', "");
Expect(1, 120831, '\P{^ismathalphanum}', "");
Expect(0, 120832, '\p{ismathalphanum}', "");
Expect(1, 120832, '\p{^ismathalphanum}', "");
Expect(1, 120832, '\P{ismathalphanum}', "");
Expect(0, 120832, '\P{^ismathalphanum}', "");
Expect(1, 120831, '\p{ _IS_MATH_Alphanum}', "");
Expect(0, 120831, '\p{^ _IS_MATH_Alphanum}', "");
Expect(0, 120831, '\P{ _IS_MATH_Alphanum}', "");
Expect(1, 120831, '\P{^ _IS_MATH_Alphanum}', "");
Expect(0, 120832, '\p{ _IS_MATH_Alphanum}', "");
Expect(1, 120832, '\p{^ _IS_MATH_Alphanum}', "");
Expect(1, 120832, '\P{ _IS_MATH_Alphanum}', "");
Expect(0, 120832, '\P{^ _IS_MATH_Alphanum}', "");
Error('\p{:= In_math_Alphanum}');
Error('\P{:= In_math_Alphanum}');
Expect(1, 120831, '\p{inmathalphanum}', "");
Expect(0, 120831, '\p{^inmathalphanum}', "");
Expect(0, 120831, '\P{inmathalphanum}', "");
Expect(1, 120831, '\P{^inmathalphanum}', "");
Expect(0, 120832, '\p{inmathalphanum}', "");
Expect(1, 120832, '\p{^inmathalphanum}', "");
Expect(1, 120832, '\P{inmathalphanum}', "");
Expect(0, 120832, '\P{^inmathalphanum}', "");
Expect(1, 120831, '\p{ In_Math_alphanum}', "");
Expect(0, 120831, '\p{^ In_Math_alphanum}', "");
Expect(0, 120831, '\P{ In_Math_alphanum}', "");
Expect(1, 120831, '\P{^ In_Math_alphanum}', "");
Expect(0, 120832, '\p{ In_Math_alphanum}', "");
Expect(1, 120832, '\p{^ In_Math_alphanum}', "");
Expect(1, 120832, '\P{ In_Math_alphanum}', "");
Expect(0, 120832, '\P{^ In_Math_alphanum}', "");
Error('\p{:=		MATHEMATICAL_Operators}');
Error('\P{:=		MATHEMATICAL_Operators}');
Expect(1, 8959, '\p{mathematicaloperators}', "");
Expect(0, 8959, '\p{^mathematicaloperators}', "");
Expect(0, 8959, '\P{mathematicaloperators}', "");
Expect(1, 8959, '\P{^mathematicaloperators}', "");
Expect(0, 8960, '\p{mathematicaloperators}', "");
Expect(1, 8960, '\p{^mathematicaloperators}', "");
Expect(1, 8960, '\P{mathematicaloperators}', "");
Expect(0, 8960, '\P{^mathematicaloperators}', "");
Expect(1, 8959, '\p{_mathematical_Operators}', "");
Expect(0, 8959, '\p{^_mathematical_Operators}', "");
Expect(0, 8959, '\P{_mathematical_Operators}', "");
Expect(1, 8959, '\P{^_mathematical_Operators}', "");
Expect(0, 8960, '\p{_mathematical_Operators}', "");
Expect(1, 8960, '\p{^_mathematical_Operators}', "");
Expect(1, 8960, '\P{_mathematical_Operators}', "");
Expect(0, 8960, '\P{^_mathematical_Operators}', "");
Error('\p{		Is_mathematical_Operators:=}');
Error('\P{		Is_mathematical_Operators:=}');
Expect(1, 8959, '\p{ismathematicaloperators}', "");
Expect(0, 8959, '\p{^ismathematicaloperators}', "");
Expect(0, 8959, '\P{ismathematicaloperators}', "");
Expect(1, 8959, '\P{^ismathematicaloperators}', "");
Expect(0, 8960, '\p{ismathematicaloperators}', "");
Expect(1, 8960, '\p{^ismathematicaloperators}', "");
Expect(1, 8960, '\P{ismathematicaloperators}', "");
Expect(0, 8960, '\P{^ismathematicaloperators}', "");
Expect(1, 8959, '\p{_is_MATHEMATICAL_OPERATORS}', "");
Expect(0, 8959, '\p{^_is_MATHEMATICAL_OPERATORS}', "");
Expect(0, 8959, '\P{_is_MATHEMATICAL_OPERATORS}', "");
Expect(1, 8959, '\P{^_is_MATHEMATICAL_OPERATORS}', "");
Expect(0, 8960, '\p{_is_MATHEMATICAL_OPERATORS}', "");
Expect(1, 8960, '\p{^_is_MATHEMATICAL_OPERATORS}', "");
Expect(1, 8960, '\P{_is_MATHEMATICAL_OPERATORS}', "");
Expect(0, 8960, '\P{^_is_MATHEMATICAL_OPERATORS}', "");
Error('\p{_ In_MATHEMATICAL_Operators/a/}');
Error('\P{_ In_MATHEMATICAL_Operators/a/}');
Expect(1, 8959, '\p{inmathematicaloperators}', "");
Expect(0, 8959, '\p{^inmathematicaloperators}', "");
Expect(0, 8959, '\P{inmathematicaloperators}', "");
Expect(1, 8959, '\P{^inmathematicaloperators}', "");
Expect(0, 8960, '\p{inmathematicaloperators}', "");
Expect(1, 8960, '\p{^inmathematicaloperators}', "");
Expect(1, 8960, '\P{inmathematicaloperators}', "");
Expect(0, 8960, '\P{^inmathematicaloperators}', "");
Expect(1, 8959, '\p{ 	in_Mathematical_operators}', "");
Expect(0, 8959, '\p{^ 	in_Mathematical_operators}', "");
Expect(0, 8959, '\P{ 	in_Mathematical_operators}', "");
Expect(1, 8959, '\P{^ 	in_Mathematical_operators}', "");
Expect(0, 8960, '\p{ 	in_Mathematical_operators}', "");
Expect(1, 8960, '\p{^ 	in_Mathematical_operators}', "");
Expect(1, 8960, '\P{ 	in_Mathematical_operators}', "");
Expect(0, 8960, '\P{^ 	in_Mathematical_operators}', "");
Error('\p{:=_ MATH_OPERATORS}');
Error('\P{:=_ MATH_OPERATORS}');
Expect(1, 8959, '\p{mathoperators}', "");
Expect(0, 8959, '\p{^mathoperators}', "");
Expect(0, 8959, '\P{mathoperators}', "");
Expect(1, 8959, '\P{^mathoperators}', "");
Expect(0, 8960, '\p{mathoperators}', "");
Expect(1, 8960, '\p{^mathoperators}', "");
Expect(1, 8960, '\P{mathoperators}', "");
Expect(0, 8960, '\P{^mathoperators}', "");
Expect(1, 8959, '\p{--math_OPERATORS}', "");
Expect(0, 8959, '\p{^--math_OPERATORS}', "");
Expect(0, 8959, '\P{--math_OPERATORS}', "");
Expect(1, 8959, '\P{^--math_OPERATORS}', "");
Expect(0, 8960, '\p{--math_OPERATORS}', "");
Expect(1, 8960, '\p{^--math_OPERATORS}', "");
Expect(1, 8960, '\P{--math_OPERATORS}', "");
Expect(0, 8960, '\P{^--math_OPERATORS}', "");
Error('\p{/a/- is_Math_Operators}');
Error('\P{/a/- is_Math_Operators}');
Expect(1, 8959, '\p{ismathoperators}', "");
Expect(0, 8959, '\p{^ismathoperators}', "");
Expect(0, 8959, '\P{ismathoperators}', "");
Expect(1, 8959, '\P{^ismathoperators}', "");
Expect(0, 8960, '\p{ismathoperators}', "");
Expect(1, 8960, '\p{^ismathoperators}', "");
Expect(1, 8960, '\P{ismathoperators}', "");
Expect(0, 8960, '\P{^ismathoperators}', "");
Expect(1, 8959, '\p{ 	Is_math_Operators}', "");
Expect(0, 8959, '\p{^ 	Is_math_Operators}', "");
Expect(0, 8959, '\P{ 	Is_math_Operators}', "");
Expect(1, 8959, '\P{^ 	Is_math_Operators}', "");
Expect(0, 8960, '\p{ 	Is_math_Operators}', "");
Expect(1, 8960, '\p{^ 	Is_math_Operators}', "");
Expect(1, 8960, '\P{ 	Is_math_Operators}', "");
Expect(0, 8960, '\P{^ 	Is_math_Operators}', "");
Error('\p{ _In_MATH_Operators/a/}');
Error('\P{ _In_MATH_Operators/a/}');
Expect(1, 8959, '\p{inmathoperators}', "");
Expect(0, 8959, '\p{^inmathoperators}', "");
Expect(0, 8959, '\P{inmathoperators}', "");
Expect(1, 8959, '\P{^inmathoperators}', "");
Expect(0, 8960, '\p{inmathoperators}', "");
Expect(1, 8960, '\p{^inmathoperators}', "");
Expect(1, 8960, '\P{inmathoperators}', "");
Expect(0, 8960, '\P{^inmathoperators}', "");
Expect(1, 8959, '\p{-	In_math_Operators}', "");
Expect(0, 8959, '\p{^-	In_math_Operators}', "");
Expect(0, 8959, '\P{-	In_math_Operators}', "");
Expect(1, 8959, '\P{^-	In_math_Operators}', "");
Expect(0, 8960, '\p{-	In_math_Operators}', "");
Expect(1, 8960, '\p{^-	In_math_Operators}', "");
Expect(1, 8960, '\P{-	In_math_Operators}', "");
Expect(0, 8960, '\P{^-	In_math_Operators}', "");
Error('\p{  MEETEI_Mayek/a/}');
Error('\P{  MEETEI_Mayek/a/}');
Expect(1, 44025, '\p{meeteimayek}', "");
Expect(0, 44025, '\p{^meeteimayek}', "");
Expect(0, 44025, '\P{meeteimayek}', "");
Expect(1, 44025, '\P{^meeteimayek}', "");
Expect(0, 44026, '\p{meeteimayek}', "");
Expect(1, 44026, '\p{^meeteimayek}', "");
Expect(1, 44026, '\P{meeteimayek}', "");
Expect(0, 44026, '\P{^meeteimayek}', "");
Expect(1, 44025, '\p{__MEETEI_mayek}', "");
Expect(0, 44025, '\p{^__MEETEI_mayek}', "");
Expect(0, 44025, '\P{__MEETEI_mayek}', "");
Expect(1, 44025, '\P{^__MEETEI_mayek}', "");
Expect(0, 44026, '\p{__MEETEI_mayek}', "");
Expect(1, 44026, '\p{^__MEETEI_mayek}', "");
Expect(1, 44026, '\P{__MEETEI_mayek}', "");
Expect(0, 44026, '\P{^__MEETEI_mayek}', "");
Error('\p{_:=Is_meetei_Mayek}');
Error('\P{_:=Is_meetei_Mayek}');
Expect(1, 44025, '\p{ismeeteimayek}', "");
Expect(0, 44025, '\p{^ismeeteimayek}', "");
Expect(0, 44025, '\P{ismeeteimayek}', "");
Expect(1, 44025, '\P{^ismeeteimayek}', "");
Expect(0, 44026, '\p{ismeeteimayek}', "");
Expect(1, 44026, '\p{^ismeeteimayek}', "");
Expect(1, 44026, '\P{ismeeteimayek}', "");
Expect(0, 44026, '\P{^ismeeteimayek}', "");
Expect(1, 44025, '\p{_ is_Meetei_Mayek}', "");
Expect(0, 44025, '\p{^_ is_Meetei_Mayek}', "");
Expect(0, 44025, '\P{_ is_Meetei_Mayek}', "");
Expect(1, 44025, '\P{^_ is_Meetei_Mayek}', "");
Expect(0, 44026, '\p{_ is_Meetei_Mayek}', "");
Expect(1, 44026, '\p{^_ is_Meetei_Mayek}', "");
Expect(1, 44026, '\P{_ is_Meetei_Mayek}', "");
Expect(0, 44026, '\P{^_ is_Meetei_Mayek}', "");
Error('\p{	:=MTEI}');
Error('\P{	:=MTEI}');
Expect(1, 44025, '\p{mtei}', "");
Expect(0, 44025, '\p{^mtei}', "");
Expect(0, 44025, '\P{mtei}', "");
Expect(1, 44025, '\P{^mtei}', "");
Expect(0, 44026, '\p{mtei}', "");
Expect(1, 44026, '\p{^mtei}', "");
Expect(1, 44026, '\P{mtei}', "");
Expect(0, 44026, '\P{^mtei}', "");
Expect(1, 44025, '\p{  Mtei}', "");
Expect(0, 44025, '\p{^  Mtei}', "");
Expect(0, 44025, '\P{  Mtei}', "");
Expect(1, 44025, '\P{^  Mtei}', "");
Expect(0, 44026, '\p{  Mtei}', "");
Expect(1, 44026, '\p{^  Mtei}', "");
Expect(1, 44026, '\P{  Mtei}', "");
Expect(0, 44026, '\P{^  Mtei}', "");
Error('\p{-:=IS_Mtei}');
Error('\P{-:=IS_Mtei}');
Expect(1, 44025, '\p{ismtei}', "");
Expect(0, 44025, '\p{^ismtei}', "");
Expect(0, 44025, '\P{ismtei}', "");
Expect(1, 44025, '\P{^ismtei}', "");
Expect(0, 44026, '\p{ismtei}', "");
Expect(1, 44026, '\p{^ismtei}', "");
Expect(1, 44026, '\P{ismtei}', "");
Expect(0, 44026, '\P{^ismtei}', "");
Expect(1, 44025, '\p{_is_mtei}', "");
Expect(0, 44025, '\p{^_is_mtei}', "");
Expect(0, 44025, '\P{_is_mtei}', "");
Expect(1, 44025, '\P{^_is_mtei}', "");
Expect(0, 44026, '\p{_is_mtei}', "");
Expect(1, 44026, '\p{^_is_mtei}', "");
Expect(1, 44026, '\P{_is_mtei}', "");
Expect(0, 44026, '\P{^_is_mtei}', "");
Error('\p{_/a/Meetei_mayek_extensions}');
Error('\P{_/a/Meetei_mayek_extensions}');
Expect(1, 43775, '\p{meeteimayekextensions}', "");
Expect(0, 43775, '\p{^meeteimayekextensions}', "");
Expect(0, 43775, '\P{meeteimayekextensions}', "");
Expect(1, 43775, '\P{^meeteimayekextensions}', "");
Expect(0, 43776, '\p{meeteimayekextensions}', "");
Expect(1, 43776, '\p{^meeteimayekextensions}', "");
Expect(1, 43776, '\P{meeteimayekextensions}', "");
Expect(0, 43776, '\P{^meeteimayekextensions}', "");
Expect(1, 43775, '\p{  Meetei_Mayek_EXTENSIONS}', "");
Expect(0, 43775, '\p{^  Meetei_Mayek_EXTENSIONS}', "");
Expect(0, 43775, '\P{  Meetei_Mayek_EXTENSIONS}', "");
Expect(1, 43775, '\P{^  Meetei_Mayek_EXTENSIONS}', "");
Expect(0, 43776, '\p{  Meetei_Mayek_EXTENSIONS}', "");
Expect(1, 43776, '\p{^  Meetei_Mayek_EXTENSIONS}', "");
Expect(1, 43776, '\P{  Meetei_Mayek_EXTENSIONS}', "");
Expect(0, 43776, '\P{^  Meetei_Mayek_EXTENSIONS}', "");
Error('\p{:=-_is_Meetei_mayek_EXTENSIONS}');
Error('\P{:=-_is_Meetei_mayek_EXTENSIONS}');
Expect(1, 43775, '\p{ismeeteimayekextensions}', "");
Expect(0, 43775, '\p{^ismeeteimayekextensions}', "");
Expect(0, 43775, '\P{ismeeteimayekextensions}', "");
Expect(1, 43775, '\P{^ismeeteimayekextensions}', "");
Expect(0, 43776, '\p{ismeeteimayekextensions}', "");
Expect(1, 43776, '\p{^ismeeteimayekextensions}', "");
Expect(1, 43776, '\P{ismeeteimayekextensions}', "");
Expect(0, 43776, '\P{^ismeeteimayekextensions}', "");
Expect(1, 43775, '\p{  IS_MEETEI_Mayek_Extensions}', "");
Expect(0, 43775, '\p{^  IS_MEETEI_Mayek_Extensions}', "");
Expect(0, 43775, '\P{  IS_MEETEI_Mayek_Extensions}', "");
Expect(1, 43775, '\P{^  IS_MEETEI_Mayek_Extensions}', "");
Expect(0, 43776, '\p{  IS_MEETEI_Mayek_Extensions}', "");
Expect(1, 43776, '\p{^  IS_MEETEI_Mayek_Extensions}', "");
Expect(1, 43776, '\P{  IS_MEETEI_Mayek_Extensions}', "");
Expect(0, 43776, '\P{^  IS_MEETEI_Mayek_Extensions}', "");
Error('\p{/a/ IN_MEETEI_mayek_EXTENSIONS}');
Error('\P{/a/ IN_MEETEI_mayek_EXTENSIONS}');
Expect(1, 43775, '\p{inmeeteimayekextensions}', "");
Expect(0, 43775, '\p{^inmeeteimayekextensions}', "");
Expect(0, 43775, '\P{inmeeteimayekextensions}', "");
Expect(1, 43775, '\P{^inmeeteimayekextensions}', "");
Expect(0, 43776, '\p{inmeeteimayekextensions}', "");
Expect(1, 43776, '\p{^inmeeteimayekextensions}', "");
Expect(1, 43776, '\P{inmeeteimayekextensions}', "");
Expect(0, 43776, '\P{^inmeeteimayekextensions}', "");
Expect(1, 43775, '\p{_ IN_Meetei_Mayek_EXTENSIONS}', "");
Expect(0, 43775, '\p{^_ IN_Meetei_Mayek_EXTENSIONS}', "");
Expect(0, 43775, '\P{_ IN_Meetei_Mayek_EXTENSIONS}', "");
Expect(1, 43775, '\P{^_ IN_Meetei_Mayek_EXTENSIONS}', "");
Expect(0, 43776, '\p{_ IN_Meetei_Mayek_EXTENSIONS}', "");
Expect(1, 43776, '\p{^_ IN_Meetei_Mayek_EXTENSIONS}', "");
Expect(1, 43776, '\P{_ IN_Meetei_Mayek_EXTENSIONS}', "");
Expect(0, 43776, '\P{^_ IN_Meetei_Mayek_EXTENSIONS}', "");
Error('\p{- MEETEI_mayek_Ext:=}');
Error('\P{- MEETEI_mayek_Ext:=}');
Expect(1, 43775, '\p{meeteimayekext}', "");
Expect(0, 43775, '\p{^meeteimayekext}', "");
Expect(0, 43775, '\P{meeteimayekext}', "");
Expect(1, 43775, '\P{^meeteimayekext}', "");
Expect(0, 43776, '\p{meeteimayekext}', "");
Expect(1, 43776, '\p{^meeteimayekext}', "");
Expect(1, 43776, '\P{meeteimayekext}', "");
Expect(0, 43776, '\P{^meeteimayekext}', "");
Expect(1, 43775, '\p{ Meetei_MAYEK_Ext}', "");
Expect(0, 43775, '\p{^ Meetei_MAYEK_Ext}', "");
Expect(0, 43775, '\P{ Meetei_MAYEK_Ext}', "");
Expect(1, 43775, '\P{^ Meetei_MAYEK_Ext}', "");
Expect(0, 43776, '\p{ Meetei_MAYEK_Ext}', "");
Expect(1, 43776, '\p{^ Meetei_MAYEK_Ext}', "");
Expect(1, 43776, '\P{ Meetei_MAYEK_Ext}', "");
Expect(0, 43776, '\P{^ Meetei_MAYEK_Ext}', "");
Error('\p{:=_-is_MEETEI_Mayek_ext}');
Error('\P{:=_-is_MEETEI_Mayek_ext}');
Expect(1, 43775, '\p{ismeeteimayekext}', "");
Expect(0, 43775, '\p{^ismeeteimayekext}', "");
Expect(0, 43775, '\P{ismeeteimayekext}', "");
Expect(1, 43775, '\P{^ismeeteimayekext}', "");
Expect(0, 43776, '\p{ismeeteimayekext}', "");
Expect(1, 43776, '\p{^ismeeteimayekext}', "");
Expect(1, 43776, '\P{ismeeteimayekext}', "");
Expect(0, 43776, '\P{^ismeeteimayekext}', "");
Expect(1, 43775, '\p{ Is_MEETEI_mayek_Ext}', "");
Expect(0, 43775, '\p{^ Is_MEETEI_mayek_Ext}', "");
Expect(0, 43775, '\P{ Is_MEETEI_mayek_Ext}', "");
Expect(1, 43775, '\P{^ Is_MEETEI_mayek_Ext}', "");
Expect(0, 43776, '\p{ Is_MEETEI_mayek_Ext}', "");
Expect(1, 43776, '\p{^ Is_MEETEI_mayek_Ext}', "");
Expect(1, 43776, '\P{ Is_MEETEI_mayek_Ext}', "");
Expect(0, 43776, '\P{^ Is_MEETEI_mayek_Ext}', "");
Error('\p{ In_MEETEI_Mayek_ext:=}');
Error('\P{ In_MEETEI_Mayek_ext:=}');
Expect(1, 43775, '\p{inmeeteimayekext}', "");
Expect(0, 43775, '\p{^inmeeteimayekext}', "");
Expect(0, 43775, '\P{inmeeteimayekext}', "");
Expect(1, 43775, '\P{^inmeeteimayekext}', "");
Expect(0, 43776, '\p{inmeeteimayekext}', "");
Expect(1, 43776, '\p{^inmeeteimayekext}', "");
Expect(1, 43776, '\P{inmeeteimayekext}', "");
Expect(0, 43776, '\P{^inmeeteimayekext}', "");
Expect(1, 43775, '\p{_IN_meetei_mayek_Ext}', "");
Expect(0, 43775, '\p{^_IN_meetei_mayek_Ext}', "");
Expect(0, 43775, '\P{_IN_meetei_mayek_Ext}', "");
Expect(1, 43775, '\P{^_IN_meetei_mayek_Ext}', "");
Expect(0, 43776, '\p{_IN_meetei_mayek_Ext}', "");
Expect(1, 43776, '\p{^_IN_meetei_mayek_Ext}', "");
Expect(1, 43776, '\P{_IN_meetei_mayek_Ext}', "");
Expect(0, 43776, '\P{^_IN_meetei_mayek_Ext}', "");
Error('\p{/a/_	meroitic_Cursive}');
Error('\P{/a/_	meroitic_Cursive}');
Expect(1, 68031, '\p{meroiticcursive}', "");
Expect(0, 68031, '\p{^meroiticcursive}', "");
Expect(0, 68031, '\P{meroiticcursive}', "");
Expect(1, 68031, '\P{^meroiticcursive}', "");
Expect(0, 68032, '\p{meroiticcursive}', "");
Expect(1, 68032, '\p{^meroiticcursive}', "");
Expect(1, 68032, '\P{meroiticcursive}', "");
Expect(0, 68032, '\P{^meroiticcursive}', "");
Expect(1, 68031, '\p{_meroitic_CURSIVE}', "");
Expect(0, 68031, '\p{^_meroitic_CURSIVE}', "");
Expect(0, 68031, '\P{_meroitic_CURSIVE}', "");
Expect(1, 68031, '\P{^_meroitic_CURSIVE}', "");
Expect(0, 68032, '\p{_meroitic_CURSIVE}', "");
Expect(1, 68032, '\p{^_meroitic_CURSIVE}', "");
Expect(1, 68032, '\P{_meroitic_CURSIVE}', "");
Expect(0, 68032, '\P{^_meroitic_CURSIVE}', "");
Error('\p{/a/--Is_Meroitic_Cursive}');
Error('\P{/a/--Is_Meroitic_Cursive}');
Expect(1, 68031, '\p{ismeroiticcursive}', "");
Expect(0, 68031, '\p{^ismeroiticcursive}', "");
Expect(0, 68031, '\P{ismeroiticcursive}', "");
Expect(1, 68031, '\P{^ismeroiticcursive}', "");
Expect(0, 68032, '\p{ismeroiticcursive}', "");
Expect(1, 68032, '\p{^ismeroiticcursive}', "");
Expect(1, 68032, '\P{ismeroiticcursive}', "");
Expect(0, 68032, '\P{^ismeroiticcursive}', "");
Expect(1, 68031, '\p{--is_Meroitic_CURSIVE}', "");
Expect(0, 68031, '\p{^--is_Meroitic_CURSIVE}', "");
Expect(0, 68031, '\P{--is_Meroitic_CURSIVE}', "");
Expect(1, 68031, '\P{^--is_Meroitic_CURSIVE}', "");
Expect(0, 68032, '\p{--is_Meroitic_CURSIVE}', "");
Expect(1, 68032, '\p{^--is_Meroitic_CURSIVE}', "");
Expect(1, 68032, '\P{--is_Meroitic_CURSIVE}', "");
Expect(0, 68032, '\P{^--is_Meroitic_CURSIVE}', "");
Error('\p{	 MERC:=}');
Error('\P{	 MERC:=}');
Expect(1, 68031, '\p{merc}', "");
Expect(0, 68031, '\p{^merc}', "");
Expect(0, 68031, '\P{merc}', "");
Expect(1, 68031, '\P{^merc}', "");
Expect(0, 68032, '\p{merc}', "");
Expect(1, 68032, '\p{^merc}', "");
Expect(1, 68032, '\P{merc}', "");
Expect(0, 68032, '\P{^merc}', "");
Expect(1, 68031, '\p{ _Merc}', "");
Expect(0, 68031, '\p{^ _Merc}', "");
Expect(0, 68031, '\P{ _Merc}', "");
Expect(1, 68031, '\P{^ _Merc}', "");
Expect(0, 68032, '\p{ _Merc}', "");
Expect(1, 68032, '\p{^ _Merc}', "");
Expect(1, 68032, '\P{ _Merc}', "");
Expect(0, 68032, '\P{^ _Merc}', "");
Error('\p{:=- is_merc}');
Error('\P{:=- is_merc}');
Expect(1, 68031, '\p{ismerc}', "");
Expect(0, 68031, '\p{^ismerc}', "");
Expect(0, 68031, '\P{ismerc}', "");
Expect(1, 68031, '\P{^ismerc}', "");
Expect(0, 68032, '\p{ismerc}', "");
Expect(1, 68032, '\p{^ismerc}', "");
Expect(1, 68032, '\P{ismerc}', "");
Expect(0, 68032, '\P{^ismerc}', "");
Expect(1, 68031, '\p{ _IS_MERC}', "");
Expect(0, 68031, '\p{^ _IS_MERC}', "");
Expect(0, 68031, '\P{ _IS_MERC}', "");
Expect(1, 68031, '\P{^ _IS_MERC}', "");
Expect(0, 68032, '\p{ _IS_MERC}', "");
Expect(1, 68032, '\p{^ _IS_MERC}', "");
Expect(1, 68032, '\P{ _IS_MERC}', "");
Expect(0, 68032, '\P{^ _IS_MERC}', "");
Error('\p{_Meroitic_HIEROGLYPHS:=}');
Error('\P{_Meroitic_HIEROGLYPHS:=}');
Expect(1, 67999, '\p{meroitichieroglyphs}', "");
Expect(0, 67999, '\p{^meroitichieroglyphs}', "");
Expect(0, 67999, '\P{meroitichieroglyphs}', "");
Expect(1, 67999, '\P{^meroitichieroglyphs}', "");
Expect(0, 68000, '\p{meroitichieroglyphs}', "");
Expect(1, 68000, '\p{^meroitichieroglyphs}', "");
Expect(1, 68000, '\P{meroitichieroglyphs}', "");
Expect(0, 68000, '\P{^meroitichieroglyphs}', "");
Expect(1, 67999, '\p{Meroitic_HIEROGLYPHS}', "");
Expect(0, 67999, '\p{^Meroitic_HIEROGLYPHS}', "");
Expect(0, 67999, '\P{Meroitic_HIEROGLYPHS}', "");
Expect(1, 67999, '\P{^Meroitic_HIEROGLYPHS}', "");
Expect(0, 68000, '\p{Meroitic_HIEROGLYPHS}', "");
Expect(1, 68000, '\p{^Meroitic_HIEROGLYPHS}', "");
Expect(1, 68000, '\P{Meroitic_HIEROGLYPHS}', "");
Expect(0, 68000, '\P{^Meroitic_HIEROGLYPHS}', "");
Error('\p{	:=is_Meroitic_Hieroglyphs}');
Error('\P{	:=is_Meroitic_Hieroglyphs}');
Expect(1, 67999, '\p{ismeroitichieroglyphs}', "");
Expect(0, 67999, '\p{^ismeroitichieroglyphs}', "");
Expect(0, 67999, '\P{ismeroitichieroglyphs}', "");
Expect(1, 67999, '\P{^ismeroitichieroglyphs}', "");
Expect(0, 68000, '\p{ismeroitichieroglyphs}', "");
Expect(1, 68000, '\p{^ismeroitichieroglyphs}', "");
Expect(1, 68000, '\P{ismeroitichieroglyphs}', "");
Expect(0, 68000, '\P{^ismeroitichieroglyphs}', "");
Expect(1, 67999, '\p{	_is_MEROITIC_Hieroglyphs}', "");
Expect(0, 67999, '\p{^	_is_MEROITIC_Hieroglyphs}', "");
Expect(0, 67999, '\P{	_is_MEROITIC_Hieroglyphs}', "");
Expect(1, 67999, '\P{^	_is_MEROITIC_Hieroglyphs}', "");
Expect(0, 68000, '\p{	_is_MEROITIC_Hieroglyphs}', "");
Expect(1, 68000, '\p{^	_is_MEROITIC_Hieroglyphs}', "");
Expect(1, 68000, '\P{	_is_MEROITIC_Hieroglyphs}', "");
Expect(0, 68000, '\P{^	_is_MEROITIC_Hieroglyphs}', "");
Error('\p{_:=MERO}');
Error('\P{_:=MERO}');
Expect(1, 67999, '\p{mero}', "");
Expect(0, 67999, '\p{^mero}', "");
Expect(0, 67999, '\P{mero}', "");
Expect(1, 67999, '\P{^mero}', "");
Expect(0, 68000, '\p{mero}', "");
Expect(1, 68000, '\p{^mero}', "");
Expect(1, 68000, '\P{mero}', "");
Expect(0, 68000, '\P{^mero}', "");
Expect(1, 67999, '\p{_ Mero}', "");
Expect(0, 67999, '\p{^_ Mero}', "");
Expect(0, 67999, '\P{_ Mero}', "");
Expect(1, 67999, '\P{^_ Mero}', "");
Expect(0, 68000, '\p{_ Mero}', "");
Expect(1, 68000, '\p{^_ Mero}', "");
Expect(1, 68000, '\P{_ Mero}', "");
Expect(0, 68000, '\P{^_ Mero}', "");
Error('\p{__Is_Mero:=}');
Error('\P{__Is_Mero:=}');
Expect(1, 67999, '\p{ismero}', "");
Expect(0, 67999, '\p{^ismero}', "");
Expect(0, 67999, '\P{ismero}', "");
Expect(1, 67999, '\P{^ismero}', "");
Expect(0, 68000, '\p{ismero}', "");
Expect(1, 68000, '\p{^ismero}', "");
Expect(1, 68000, '\P{ismero}', "");
Expect(0, 68000, '\P{^ismero}', "");
Expect(1, 67999, '\p{-Is_Mero}', "");
Expect(0, 67999, '\p{^-Is_Mero}', "");
Expect(0, 67999, '\P{-Is_Mero}', "");
Expect(1, 67999, '\P{^-Is_Mero}', "");
Expect(0, 68000, '\p{-Is_Mero}', "");
Expect(1, 68000, '\p{^-Is_Mero}', "");
Expect(1, 68000, '\P{-Is_Mero}', "");
Expect(0, 68000, '\P{^-Is_Mero}', "");
Error('\p{:=	-miao}');
Error('\P{:=	-miao}');
Expect(1, 94111, '\p{miao}', "");
Expect(0, 94111, '\p{^miao}', "");
Expect(0, 94111, '\P{miao}', "");
Expect(1, 94111, '\P{^miao}', "");
Expect(0, 94112, '\p{miao}', "");
Expect(1, 94112, '\p{^miao}', "");
Expect(1, 94112, '\P{miao}', "");
Expect(0, 94112, '\P{^miao}', "");
Expect(1, 94111, '\p{-MIAO}', "");
Expect(0, 94111, '\p{^-MIAO}', "");
Expect(0, 94111, '\P{-MIAO}', "");
Expect(1, 94111, '\P{^-MIAO}', "");
Expect(0, 94112, '\p{-MIAO}', "");
Expect(1, 94112, '\p{^-MIAO}', "");
Expect(1, 94112, '\P{-MIAO}', "");
Expect(0, 94112, '\P{^-MIAO}', "");
Error('\p{_/a/Is_MIAO}');
Error('\P{_/a/Is_MIAO}');
Expect(1, 94111, '\p{ismiao}', "");
Expect(0, 94111, '\p{^ismiao}', "");
Expect(0, 94111, '\P{ismiao}', "");
Expect(1, 94111, '\P{^ismiao}', "");
Expect(0, 94112, '\p{ismiao}', "");
Expect(1, 94112, '\p{^ismiao}', "");
Expect(1, 94112, '\P{ismiao}', "");
Expect(0, 94112, '\P{^ismiao}', "");
Expect(1, 94111, '\p{_ is_Miao}', "");
Expect(0, 94111, '\p{^_ is_Miao}', "");
Expect(0, 94111, '\P{_ is_Miao}', "");
Expect(1, 94111, '\P{^_ is_Miao}', "");
Expect(0, 94112, '\p{_ is_Miao}', "");
Expect(1, 94112, '\p{^_ is_Miao}', "");
Expect(1, 94112, '\P{_ is_Miao}', "");
Expect(0, 94112, '\P{^_ is_Miao}', "");
Error('\p{-_Plrd:=}');
Error('\P{-_Plrd:=}');
Expect(1, 94111, '\p{plrd}', "");
Expect(0, 94111, '\p{^plrd}', "");
Expect(0, 94111, '\P{plrd}', "");
Expect(1, 94111, '\P{^plrd}', "");
Expect(0, 94112, '\p{plrd}', "");
Expect(1, 94112, '\p{^plrd}', "");
Expect(1, 94112, '\P{plrd}', "");
Expect(0, 94112, '\P{^plrd}', "");
Expect(1, 94111, '\p{ 	plrd}', "");
Expect(0, 94111, '\p{^ 	plrd}', "");
Expect(0, 94111, '\P{ 	plrd}', "");
Expect(1, 94111, '\P{^ 	plrd}', "");
Expect(0, 94112, '\p{ 	plrd}', "");
Expect(1, 94112, '\p{^ 	plrd}', "");
Expect(1, 94112, '\P{ 	plrd}', "");
Expect(0, 94112, '\P{^ 	plrd}', "");
Error('\p{:=is_Plrd}');
Error('\P{:=is_Plrd}');
Expect(1, 94111, '\p{isplrd}', "");
Expect(0, 94111, '\p{^isplrd}', "");
Expect(0, 94111, '\P{isplrd}', "");
Expect(1, 94111, '\P{^isplrd}', "");
Expect(0, 94112, '\p{isplrd}', "");
Expect(1, 94112, '\p{^isplrd}', "");
Expect(1, 94112, '\P{isplrd}', "");
Expect(0, 94112, '\P{^isplrd}', "");
Expect(1, 94111, '\p{	-is_PLRD}', "");
Expect(0, 94111, '\p{^	-is_PLRD}', "");
Expect(0, 94111, '\P{	-is_PLRD}', "");
Expect(1, 94111, '\P{^	-is_PLRD}', "");
Expect(0, 94112, '\p{	-is_PLRD}', "");
Expect(1, 94112, '\p{^	-is_PLRD}', "");
Expect(1, 94112, '\P{	-is_PLRD}', "");
Expect(0, 94112, '\P{^	-is_PLRD}', "");
Error('\p{/a/_	Miscellaneous_Mathematical_Symbols_a}');
Error('\P{/a/_	Miscellaneous_Mathematical_Symbols_a}');
Expect(1, 10223, '\p{miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\p{^miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\P{miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\P{^miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\p{miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\p{^miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\P{miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\P{^miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\p{	_Miscellaneous_mathematical_SYMBOLS_A}', "");
Expect(0, 10223, '\p{^	_Miscellaneous_mathematical_SYMBOLS_A}', "");
Expect(0, 10223, '\P{	_Miscellaneous_mathematical_SYMBOLS_A}', "");
Expect(1, 10223, '\P{^	_Miscellaneous_mathematical_SYMBOLS_A}', "");
Expect(0, 10224, '\p{	_Miscellaneous_mathematical_SYMBOLS_A}', "");
Expect(1, 10224, '\p{^	_Miscellaneous_mathematical_SYMBOLS_A}', "");
Expect(1, 10224, '\P{	_Miscellaneous_mathematical_SYMBOLS_A}', "");
Expect(0, 10224, '\P{^	_Miscellaneous_mathematical_SYMBOLS_A}', "");
Error('\p{/a/	IS_Miscellaneous_Mathematical_SYMBOLS_a}');
Error('\P{/a/	IS_Miscellaneous_Mathematical_SYMBOLS_a}');
Expect(1, 10223, '\p{ismiscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\p{^ismiscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\P{ismiscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\P{^ismiscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\p{ismiscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\p{^ismiscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\P{ismiscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\P{^ismiscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\p{_is_miscellaneous_Mathematical_Symbols_A}', "");
Expect(0, 10223, '\p{^_is_miscellaneous_Mathematical_Symbols_A}', "");
Expect(0, 10223, '\P{_is_miscellaneous_Mathematical_Symbols_A}', "");
Expect(1, 10223, '\P{^_is_miscellaneous_Mathematical_Symbols_A}', "");
Expect(0, 10224, '\p{_is_miscellaneous_Mathematical_Symbols_A}', "");
Expect(1, 10224, '\p{^_is_miscellaneous_Mathematical_Symbols_A}', "");
Expect(1, 10224, '\P{_is_miscellaneous_Mathematical_Symbols_A}', "");
Expect(0, 10224, '\P{^_is_miscellaneous_Mathematical_Symbols_A}', "");
Error('\p{	-In_Miscellaneous_Mathematical_SYMBOLS_A:=}');
Error('\P{	-In_Miscellaneous_Mathematical_SYMBOLS_A:=}');
Expect(1, 10223, '\p{inmiscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\p{^inmiscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\P{inmiscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\P{^inmiscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\p{inmiscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\p{^inmiscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\P{inmiscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\P{^inmiscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\p{in_MISCELLANEOUS_Mathematical_SYMBOLS_A}', "");
Expect(0, 10223, '\p{^in_MISCELLANEOUS_Mathematical_SYMBOLS_A}', "");
Expect(0, 10223, '\P{in_MISCELLANEOUS_Mathematical_SYMBOLS_A}', "");
Expect(1, 10223, '\P{^in_MISCELLANEOUS_Mathematical_SYMBOLS_A}', "");
Expect(0, 10224, '\p{in_MISCELLANEOUS_Mathematical_SYMBOLS_A}', "");
Expect(1, 10224, '\p{^in_MISCELLANEOUS_Mathematical_SYMBOLS_A}', "");
Expect(1, 10224, '\P{in_MISCELLANEOUS_Mathematical_SYMBOLS_A}', "");
Expect(0, 10224, '\P{^in_MISCELLANEOUS_Mathematical_SYMBOLS_A}', "");
Error('\p{ :=MISC_MATH_SYMBOLS_A}');
Error('\P{ :=MISC_MATH_SYMBOLS_A}');
Expect(1, 10223, '\p{miscmathsymbolsa}', "");
Expect(0, 10223, '\p{^miscmathsymbolsa}', "");
Expect(0, 10223, '\P{miscmathsymbolsa}', "");
Expect(1, 10223, '\P{^miscmathsymbolsa}', "");
Expect(0, 10224, '\p{miscmathsymbolsa}', "");
Expect(1, 10224, '\p{^miscmathsymbolsa}', "");
Expect(1, 10224, '\P{miscmathsymbolsa}', "");
Expect(0, 10224, '\P{^miscmathsymbolsa}', "");
Expect(1, 10223, '\p{ -Misc_MATH_Symbols_A}', "");
Expect(0, 10223, '\p{^ -Misc_MATH_Symbols_A}', "");
Expect(0, 10223, '\P{ -Misc_MATH_Symbols_A}', "");
Expect(1, 10223, '\P{^ -Misc_MATH_Symbols_A}', "");
Expect(0, 10224, '\p{ -Misc_MATH_Symbols_A}', "");
Expect(1, 10224, '\p{^ -Misc_MATH_Symbols_A}', "");
Expect(1, 10224, '\P{ -Misc_MATH_Symbols_A}', "");
Expect(0, 10224, '\P{^ -Misc_MATH_Symbols_A}', "");
Error('\p{ /a/Is_Misc_Math_Symbols_A}');
Error('\P{ /a/Is_Misc_Math_Symbols_A}');
Expect(1, 10223, '\p{ismiscmathsymbolsa}', "");
Expect(0, 10223, '\p{^ismiscmathsymbolsa}', "");
Expect(0, 10223, '\P{ismiscmathsymbolsa}', "");
Expect(1, 10223, '\P{^ismiscmathsymbolsa}', "");
Expect(0, 10224, '\p{ismiscmathsymbolsa}', "");
Expect(1, 10224, '\p{^ismiscmathsymbolsa}', "");
Expect(1, 10224, '\P{ismiscmathsymbolsa}', "");
Expect(0, 10224, '\P{^ismiscmathsymbolsa}', "");
Expect(1, 10223, '\p{  Is_MISC_Math_Symbols_A}', "");
Expect(0, 10223, '\p{^  Is_MISC_Math_Symbols_A}', "");
Expect(0, 10223, '\P{  Is_MISC_Math_Symbols_A}', "");
Expect(1, 10223, '\P{^  Is_MISC_Math_Symbols_A}', "");
Expect(0, 10224, '\p{  Is_MISC_Math_Symbols_A}', "");
Expect(1, 10224, '\p{^  Is_MISC_Math_Symbols_A}', "");
Expect(1, 10224, '\P{  Is_MISC_Math_Symbols_A}', "");
Expect(0, 10224, '\P{^  Is_MISC_Math_Symbols_A}', "");
Error('\p{-in_Misc_MATH_symbols_A/a/}');
Error('\P{-in_Misc_MATH_symbols_A/a/}');
Expect(1, 10223, '\p{inmiscmathsymbolsa}', "");
Expect(0, 10223, '\p{^inmiscmathsymbolsa}', "");
Expect(0, 10223, '\P{inmiscmathsymbolsa}', "");
Expect(1, 10223, '\P{^inmiscmathsymbolsa}', "");
Expect(0, 10224, '\p{inmiscmathsymbolsa}', "");
Expect(1, 10224, '\p{^inmiscmathsymbolsa}', "");
Expect(1, 10224, '\P{inmiscmathsymbolsa}', "");
Expect(0, 10224, '\P{^inmiscmathsymbolsa}', "");
Expect(1, 10223, '\p{	_in_MISC_MATH_Symbols_A}', "");
Expect(0, 10223, '\p{^	_in_MISC_MATH_Symbols_A}', "");
Expect(0, 10223, '\P{	_in_MISC_MATH_Symbols_A}', "");
Expect(1, 10223, '\P{^	_in_MISC_MATH_Symbols_A}', "");
Expect(0, 10224, '\p{	_in_MISC_MATH_Symbols_A}', "");
Expect(1, 10224, '\p{^	_in_MISC_MATH_Symbols_A}', "");
Expect(1, 10224, '\P{	_in_MISC_MATH_Symbols_A}', "");
Expect(0, 10224, '\P{^	_in_MISC_MATH_Symbols_A}', "");
Error('\p{ :=MISCELLANEOUS_Mathematical_symbols_B}');
Error('\P{ :=MISCELLANEOUS_Mathematical_symbols_B}');
Expect(1, 10751, '\p{miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\p{^miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\P{miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\P{^miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\p{miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\p{^miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\P{miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\P{^miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\p{ 	miscellaneous_MATHEMATICAL_Symbols_B}', "");
Expect(0, 10751, '\p{^ 	miscellaneous_MATHEMATICAL_Symbols_B}', "");
Expect(0, 10751, '\P{ 	miscellaneous_MATHEMATICAL_Symbols_B}', "");
Expect(1, 10751, '\P{^ 	miscellaneous_MATHEMATICAL_Symbols_B}', "");
Expect(0, 10752, '\p{ 	miscellaneous_MATHEMATICAL_Symbols_B}', "");
Expect(1, 10752, '\p{^ 	miscellaneous_MATHEMATICAL_Symbols_B}', "");
Expect(1, 10752, '\P{ 	miscellaneous_MATHEMATICAL_Symbols_B}', "");
Expect(0, 10752, '\P{^ 	miscellaneous_MATHEMATICAL_Symbols_B}', "");
Error('\p{-:=Is_Miscellaneous_Mathematical_Symbols_B}');
Error('\P{-:=Is_Miscellaneous_Mathematical_Symbols_B}');
Expect(1, 10751, '\p{ismiscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\p{^ismiscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\P{ismiscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\P{^ismiscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\p{ismiscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\p{^ismiscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\P{ismiscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\P{^ismiscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\p{	-Is_MISCELLANEOUS_mathematical_SYMBOLS_B}', "");
Expect(0, 10751, '\p{^	-Is_MISCELLANEOUS_mathematical_SYMBOLS_B}', "");
Expect(0, 10751, '\P{	-Is_MISCELLANEOUS_mathematical_SYMBOLS_B}', "");
Expect(1, 10751, '\P{^	-Is_MISCELLANEOUS_mathematical_SYMBOLS_B}', "");
Expect(0, 10752, '\p{	-Is_MISCELLANEOUS_mathematical_SYMBOLS_B}', "");
Expect(1, 10752, '\p{^	-Is_MISCELLANEOUS_mathematical_SYMBOLS_B}', "");
Expect(1, 10752, '\P{	-Is_MISCELLANEOUS_mathematical_SYMBOLS_B}', "");
Expect(0, 10752, '\P{^	-Is_MISCELLANEOUS_mathematical_SYMBOLS_B}', "");
Error('\p{ /a/IN_Miscellaneous_Mathematical_symbols_b}');
Error('\P{ /a/IN_Miscellaneous_Mathematical_symbols_b}');
Expect(1, 10751, '\p{inmiscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\p{^inmiscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\P{inmiscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\P{^inmiscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\p{inmiscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\p{^inmiscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\P{inmiscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\P{^inmiscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\p{_-In_Miscellaneous_MATHEMATICAL_Symbols_B}', "");
Expect(0, 10751, '\p{^_-In_Miscellaneous_MATHEMATICAL_Symbols_B}', "");
Expect(0, 10751, '\P{_-In_Miscellaneous_MATHEMATICAL_Symbols_B}', "");
Expect(1, 10751, '\P{^_-In_Miscellaneous_MATHEMATICAL_Symbols_B}', "");
Expect(0, 10752, '\p{_-In_Miscellaneous_MATHEMATICAL_Symbols_B}', "");
Expect(1, 10752, '\p{^_-In_Miscellaneous_MATHEMATICAL_Symbols_B}', "");
Expect(1, 10752, '\P{_-In_Miscellaneous_MATHEMATICAL_Symbols_B}', "");
Expect(0, 10752, '\P{^_-In_Miscellaneous_MATHEMATICAL_Symbols_B}', "");
Error('\p{ :=MISC_MATH_symbols_B}');
Error('\P{ :=MISC_MATH_symbols_B}');
Expect(1, 10751, '\p{miscmathsymbolsb}', "");
Expect(0, 10751, '\p{^miscmathsymbolsb}', "");
Expect(0, 10751, '\P{miscmathsymbolsb}', "");
Expect(1, 10751, '\P{^miscmathsymbolsb}', "");
Expect(0, 10752, '\p{miscmathsymbolsb}', "");
Expect(1, 10752, '\p{^miscmathsymbolsb}', "");
Expect(1, 10752, '\P{miscmathsymbolsb}', "");
Expect(0, 10752, '\P{^miscmathsymbolsb}', "");
Expect(1, 10751, '\p{ -Misc_Math_Symbols_B}', "");
Expect(0, 10751, '\p{^ -Misc_Math_Symbols_B}', "");
Expect(0, 10751, '\P{ -Misc_Math_Symbols_B}', "");
Expect(1, 10751, '\P{^ -Misc_Math_Symbols_B}', "");
Expect(0, 10752, '\p{ -Misc_Math_Symbols_B}', "");
Expect(1, 10752, '\p{^ -Misc_Math_Symbols_B}', "");
Expect(1, 10752, '\P{ -Misc_Math_Symbols_B}', "");
Expect(0, 10752, '\P{^ -Misc_Math_Symbols_B}', "");
Error('\p{-_is_Misc_Math_Symbols_b/a/}');
Error('\P{-_is_Misc_Math_Symbols_b/a/}');
Expect(1, 10751, '\p{ismiscmathsymbolsb}', "");
Expect(0, 10751, '\p{^ismiscmathsymbolsb}', "");
Expect(0, 10751, '\P{ismiscmathsymbolsb}', "");
Expect(1, 10751, '\P{^ismiscmathsymbolsb}', "");
Expect(0, 10752, '\p{ismiscmathsymbolsb}', "");
Expect(1, 10752, '\p{^ismiscmathsymbolsb}', "");
Expect(1, 10752, '\P{ismiscmathsymbolsb}', "");
Expect(0, 10752, '\P{^ismiscmathsymbolsb}', "");
Expect(1, 10751, '\p{	-is_misc_Math_symbols_B}', "");
Expect(0, 10751, '\p{^	-is_misc_Math_symbols_B}', "");
Expect(0, 10751, '\P{	-is_misc_Math_symbols_B}', "");
Expect(1, 10751, '\P{^	-is_misc_Math_symbols_B}', "");
Expect(0, 10752, '\p{	-is_misc_Math_symbols_B}', "");
Expect(1, 10752, '\p{^	-is_misc_Math_symbols_B}', "");
Expect(1, 10752, '\P{	-is_misc_Math_symbols_B}', "");
Expect(0, 10752, '\P{^	-is_misc_Math_symbols_B}', "");
Error('\p{	In_MISC_Math_SYMBOLS_b:=}');
Error('\P{	In_MISC_Math_SYMBOLS_b:=}');
Expect(1, 10751, '\p{inmiscmathsymbolsb}', "");
Expect(0, 10751, '\p{^inmiscmathsymbolsb}', "");
Expect(0, 10751, '\P{inmiscmathsymbolsb}', "");
Expect(1, 10751, '\P{^inmiscmathsymbolsb}', "");
Expect(0, 10752, '\p{inmiscmathsymbolsb}', "");
Expect(1, 10752, '\p{^inmiscmathsymbolsb}', "");
Expect(1, 10752, '\P{inmiscmathsymbolsb}', "");
Expect(0, 10752, '\P{^inmiscmathsymbolsb}', "");
Expect(1, 10751, '\p{ In_Misc_Math_symbols_B}', "");
Expect(0, 10751, '\p{^ In_Misc_Math_symbols_B}', "");
Expect(0, 10751, '\P{ In_Misc_Math_symbols_B}', "");
Expect(1, 10751, '\P{^ In_Misc_Math_symbols_B}', "");
Expect(0, 10752, '\p{ In_Misc_Math_symbols_B}', "");
Expect(1, 10752, '\p{^ In_Misc_Math_symbols_B}', "");
Expect(1, 10752, '\P{ In_Misc_Math_symbols_B}', "");
Expect(0, 10752, '\P{^ In_Misc_Math_symbols_B}', "");
Error('\p{	Miscellaneous_Symbols:=}');
Error('\P{	Miscellaneous_Symbols:=}');
Expect(1, 9983, '\p{miscellaneoussymbols}', "");
Expect(0, 9983, '\p{^miscellaneoussymbols}', "");
Expect(0, 9983, '\P{miscellaneoussymbols}', "");
Expect(1, 9983, '\P{^miscellaneoussymbols}', "");
Expect(0, 9984, '\p{miscellaneoussymbols}', "");
Expect(1, 9984, '\p{^miscellaneoussymbols}', "");
Expect(1, 9984, '\P{miscellaneoussymbols}', "");
Expect(0, 9984, '\P{^miscellaneoussymbols}', "");
Expect(1, 9983, '\p{ Miscellaneous_SYMBOLS}', "");
Expect(0, 9983, '\p{^ Miscellaneous_SYMBOLS}', "");
Expect(0, 9983, '\P{ Miscellaneous_SYMBOLS}', "");
Expect(1, 9983, '\P{^ Miscellaneous_SYMBOLS}', "");
Expect(0, 9984, '\p{ Miscellaneous_SYMBOLS}', "");
Expect(1, 9984, '\p{^ Miscellaneous_SYMBOLS}', "");
Expect(1, 9984, '\P{ Miscellaneous_SYMBOLS}', "");
Expect(0, 9984, '\P{^ Miscellaneous_SYMBOLS}', "");
Error('\p{:=-_is_miscellaneous_SYMBOLS}');
Error('\P{:=-_is_miscellaneous_SYMBOLS}');
Expect(1, 9983, '\p{ismiscellaneoussymbols}', "");
Expect(0, 9983, '\p{^ismiscellaneoussymbols}', "");
Expect(0, 9983, '\P{ismiscellaneoussymbols}', "");
Expect(1, 9983, '\P{^ismiscellaneoussymbols}', "");
Expect(0, 9984, '\p{ismiscellaneoussymbols}', "");
Expect(1, 9984, '\p{^ismiscellaneoussymbols}', "");
Expect(1, 9984, '\P{ismiscellaneoussymbols}', "");
Expect(0, 9984, '\P{^ismiscellaneoussymbols}', "");
Expect(1, 9983, '\p{_-Is_MISCELLANEOUS_symbols}', "");
Expect(0, 9983, '\p{^_-Is_MISCELLANEOUS_symbols}', "");
Expect(0, 9983, '\P{_-Is_MISCELLANEOUS_symbols}', "");
Expect(1, 9983, '\P{^_-Is_MISCELLANEOUS_symbols}', "");
Expect(0, 9984, '\p{_-Is_MISCELLANEOUS_symbols}', "");
Expect(1, 9984, '\p{^_-Is_MISCELLANEOUS_symbols}', "");
Expect(1, 9984, '\P{_-Is_MISCELLANEOUS_symbols}', "");
Expect(0, 9984, '\P{^_-Is_MISCELLANEOUS_symbols}', "");
Error('\p{/a/ In_Miscellaneous_Symbols}');
Error('\P{/a/ In_Miscellaneous_Symbols}');
Expect(1, 9983, '\p{inmiscellaneoussymbols}', "");
Expect(0, 9983, '\p{^inmiscellaneoussymbols}', "");
Expect(0, 9983, '\P{inmiscellaneoussymbols}', "");
Expect(1, 9983, '\P{^inmiscellaneoussymbols}', "");
Expect(0, 9984, '\p{inmiscellaneoussymbols}', "");
Expect(1, 9984, '\p{^inmiscellaneoussymbols}', "");
Expect(1, 9984, '\P{inmiscellaneoussymbols}', "");
Expect(0, 9984, '\P{^inmiscellaneoussymbols}', "");
Expect(1, 9983, '\p{	IN_miscellaneous_Symbols}', "");
Expect(0, 9983, '\p{^	IN_miscellaneous_Symbols}', "");
Expect(0, 9983, '\P{	IN_miscellaneous_Symbols}', "");
Expect(1, 9983, '\P{^	IN_miscellaneous_Symbols}', "");
Expect(0, 9984, '\p{	IN_miscellaneous_Symbols}', "");
Expect(1, 9984, '\p{^	IN_miscellaneous_Symbols}', "");
Expect(1, 9984, '\P{	IN_miscellaneous_Symbols}', "");
Expect(0, 9984, '\P{^	IN_miscellaneous_Symbols}', "");
Error('\p{ :=Misc_symbols}');
Error('\P{ :=Misc_symbols}');
Expect(1, 9983, '\p{miscsymbols}', "");
Expect(0, 9983, '\p{^miscsymbols}', "");
Expect(0, 9983, '\P{miscsymbols}', "");
Expect(1, 9983, '\P{^miscsymbols}', "");
Expect(0, 9984, '\p{miscsymbols}', "");
Expect(1, 9984, '\p{^miscsymbols}', "");
Expect(1, 9984, '\P{miscsymbols}', "");
Expect(0, 9984, '\P{^miscsymbols}', "");
Expect(1, 9983, '\p{__MISC_symbols}', "");
Expect(0, 9983, '\p{^__MISC_symbols}', "");
Expect(0, 9983, '\P{__MISC_symbols}', "");
Expect(1, 9983, '\P{^__MISC_symbols}', "");
Expect(0, 9984, '\p{__MISC_symbols}', "");
Expect(1, 9984, '\p{^__MISC_symbols}', "");
Expect(1, 9984, '\P{__MISC_symbols}', "");
Expect(0, 9984, '\P{^__MISC_symbols}', "");
Error('\p{_IS_misc_Symbols/a/}');
Error('\P{_IS_misc_Symbols/a/}');
Expect(1, 9983, '\p{ismiscsymbols}', "");
Expect(0, 9983, '\p{^ismiscsymbols}', "");
Expect(0, 9983, '\P{ismiscsymbols}', "");
Expect(1, 9983, '\P{^ismiscsymbols}', "");
Expect(0, 9984, '\p{ismiscsymbols}', "");
Expect(1, 9984, '\p{^ismiscsymbols}', "");
Expect(1, 9984, '\P{ismiscsymbols}', "");
Expect(0, 9984, '\P{^ismiscsymbols}', "");
Expect(1, 9983, '\p{-Is_Misc_Symbols}', "");
Expect(0, 9983, '\p{^-Is_Misc_Symbols}', "");
Expect(0, 9983, '\P{-Is_Misc_Symbols}', "");
Expect(1, 9983, '\P{^-Is_Misc_Symbols}', "");
Expect(0, 9984, '\p{-Is_Misc_Symbols}', "");
Expect(1, 9984, '\p{^-Is_Misc_Symbols}', "");
Expect(1, 9984, '\P{-Is_Misc_Symbols}', "");
Expect(0, 9984, '\P{^-Is_Misc_Symbols}', "");
Error('\p{_ in_MISC_Symbols/a/}');
Error('\P{_ in_MISC_Symbols/a/}');
Expect(1, 9983, '\p{inmiscsymbols}', "");
Expect(0, 9983, '\p{^inmiscsymbols}', "");
Expect(0, 9983, '\P{inmiscsymbols}', "");
Expect(1, 9983, '\P{^inmiscsymbols}', "");
Expect(0, 9984, '\p{inmiscsymbols}', "");
Expect(1, 9984, '\p{^inmiscsymbols}', "");
Expect(1, 9984, '\P{inmiscsymbols}', "");
Expect(0, 9984, '\P{^inmiscsymbols}', "");
Expect(1, 9983, '\p{_In_MISC_SYMBOLS}', "");
Expect(0, 9983, '\p{^_In_MISC_SYMBOLS}', "");
Expect(0, 9983, '\P{_In_MISC_SYMBOLS}', "");
Expect(1, 9983, '\P{^_In_MISC_SYMBOLS}', "");
Expect(0, 9984, '\p{_In_MISC_SYMBOLS}', "");
Expect(1, 9984, '\p{^_In_MISC_SYMBOLS}', "");
Expect(1, 9984, '\P{_In_MISC_SYMBOLS}', "");
Expect(0, 9984, '\P{^_In_MISC_SYMBOLS}', "");
Error('\p{		Miscellaneous_symbols_and_ARROWS:=}');
Error('\P{		Miscellaneous_symbols_and_ARROWS:=}');
Expect(1, 11263, '\p{miscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\p{^miscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\P{miscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\P{^miscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\p{miscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\p{^miscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\P{miscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\P{^miscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\p{-miscellaneous_SYMBOLS_And_Arrows}', "");
Expect(0, 11263, '\p{^-miscellaneous_SYMBOLS_And_Arrows}', "");
Expect(0, 11263, '\P{-miscellaneous_SYMBOLS_And_Arrows}', "");
Expect(1, 11263, '\P{^-miscellaneous_SYMBOLS_And_Arrows}', "");
Expect(0, 11264, '\p{-miscellaneous_SYMBOLS_And_Arrows}', "");
Expect(1, 11264, '\p{^-miscellaneous_SYMBOLS_And_Arrows}', "");
Expect(1, 11264, '\P{-miscellaneous_SYMBOLS_And_Arrows}', "");
Expect(0, 11264, '\P{^-miscellaneous_SYMBOLS_And_Arrows}', "");
Error('\p{-:=IS_miscellaneous_symbols_And_Arrows}');
Error('\P{-:=IS_miscellaneous_symbols_And_Arrows}');
Expect(1, 11263, '\p{ismiscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\p{^ismiscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\P{ismiscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\P{^ismiscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\p{ismiscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\p{^ismiscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\P{ismiscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\P{^ismiscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\p{-	IS_MISCELLANEOUS_Symbols_AND_Arrows}', "");
Expect(0, 11263, '\p{^-	IS_MISCELLANEOUS_Symbols_AND_Arrows}', "");
Expect(0, 11263, '\P{-	IS_MISCELLANEOUS_Symbols_AND_Arrows}', "");
Expect(1, 11263, '\P{^-	IS_MISCELLANEOUS_Symbols_AND_Arrows}', "");
Expect(0, 11264, '\p{-	IS_MISCELLANEOUS_Symbols_AND_Arrows}', "");
Expect(1, 11264, '\p{^-	IS_MISCELLANEOUS_Symbols_AND_Arrows}', "");
Expect(1, 11264, '\P{-	IS_MISCELLANEOUS_Symbols_AND_Arrows}', "");
Expect(0, 11264, '\P{^-	IS_MISCELLANEOUS_Symbols_AND_Arrows}', "");
Error('\p{/a/  In_Miscellaneous_symbols_And_Arrows}');
Error('\P{/a/  In_Miscellaneous_symbols_And_Arrows}');
Expect(1, 11263, '\p{inmiscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\p{^inmiscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\P{inmiscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\P{^inmiscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\p{inmiscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\p{^inmiscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\P{inmiscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\P{^inmiscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\p{-In_miscellaneous_symbols_And_ARROWS}', "");
Expect(0, 11263, '\p{^-In_miscellaneous_symbols_And_ARROWS}', "");
Expect(0, 11263, '\P{-In_miscellaneous_symbols_And_ARROWS}', "");
Expect(1, 11263, '\P{^-In_miscellaneous_symbols_And_ARROWS}', "");
Expect(0, 11264, '\p{-In_miscellaneous_symbols_And_ARROWS}', "");
Expect(1, 11264, '\p{^-In_miscellaneous_symbols_And_ARROWS}', "");
Expect(1, 11264, '\P{-In_miscellaneous_symbols_And_ARROWS}', "");
Expect(0, 11264, '\P{^-In_miscellaneous_symbols_And_ARROWS}', "");
Error('\p{/a/_ Misc_Arrows}');
Error('\P{/a/_ Misc_Arrows}');
Expect(1, 11263, '\p{miscarrows}', "");
Expect(0, 11263, '\p{^miscarrows}', "");
Expect(0, 11263, '\P{miscarrows}', "");
Expect(1, 11263, '\P{^miscarrows}', "");
Expect(0, 11264, '\p{miscarrows}', "");
Expect(1, 11264, '\p{^miscarrows}', "");
Expect(1, 11264, '\P{miscarrows}', "");
Expect(0, 11264, '\P{^miscarrows}', "");
Expect(1, 11263, '\p{-Misc_Arrows}', "");
Expect(0, 11263, '\p{^-Misc_Arrows}', "");
Expect(0, 11263, '\P{-Misc_Arrows}', "");
Expect(1, 11263, '\P{^-Misc_Arrows}', "");
Expect(0, 11264, '\p{-Misc_Arrows}', "");
Expect(1, 11264, '\p{^-Misc_Arrows}', "");
Expect(1, 11264, '\P{-Misc_Arrows}', "");
Expect(0, 11264, '\P{^-Misc_Arrows}', "");
Error('\p{ :=IS_misc_Arrows}');
Error('\P{ :=IS_misc_Arrows}');
Expect(1, 11263, '\p{ismiscarrows}', "");
Expect(0, 11263, '\p{^ismiscarrows}', "");
Expect(0, 11263, '\P{ismiscarrows}', "");
Expect(1, 11263, '\P{^ismiscarrows}', "");
Expect(0, 11264, '\p{ismiscarrows}', "");
Expect(1, 11264, '\p{^ismiscarrows}', "");
Expect(1, 11264, '\P{ismiscarrows}', "");
Expect(0, 11264, '\P{^ismiscarrows}', "");
Expect(1, 11263, '\p{-	Is_misc_Arrows}', "");
Expect(0, 11263, '\p{^-	Is_misc_Arrows}', "");
Expect(0, 11263, '\P{-	Is_misc_Arrows}', "");
Expect(1, 11263, '\P{^-	Is_misc_Arrows}', "");
Expect(0, 11264, '\p{-	Is_misc_Arrows}', "");
Expect(1, 11264, '\p{^-	Is_misc_Arrows}', "");
Expect(1, 11264, '\P{-	Is_misc_Arrows}', "");
Expect(0, 11264, '\P{^-	Is_misc_Arrows}', "");
Error('\p{-_In_Misc_ARROWS/a/}');
Error('\P{-_In_Misc_ARROWS/a/}');
Expect(1, 11263, '\p{inmiscarrows}', "");
Expect(0, 11263, '\p{^inmiscarrows}', "");
Expect(0, 11263, '\P{inmiscarrows}', "");
Expect(1, 11263, '\P{^inmiscarrows}', "");
Expect(0, 11264, '\p{inmiscarrows}', "");
Expect(1, 11264, '\p{^inmiscarrows}', "");
Expect(1, 11264, '\P{inmiscarrows}', "");
Expect(0, 11264, '\P{^inmiscarrows}', "");
Expect(1, 11263, '\p{	in_MISC_Arrows}', "");
Expect(0, 11263, '\p{^	in_MISC_Arrows}', "");
Expect(0, 11263, '\P{	in_MISC_Arrows}', "");
Expect(1, 11263, '\P{^	in_MISC_Arrows}', "");
Expect(0, 11264, '\p{	in_MISC_Arrows}', "");
Expect(1, 11264, '\p{^	in_MISC_Arrows}', "");
Expect(1, 11264, '\P{	in_MISC_Arrows}', "");
Expect(0, 11264, '\P{^	in_MISC_Arrows}', "");
Error('\p{_-Miscellaneous_Symbols_And_Pictographs:=}');
Error('\P{_-Miscellaneous_Symbols_And_Pictographs:=}');
Expect(1, 128511, '\p{miscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\p{^miscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\P{miscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\P{^miscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\p{miscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\p{^miscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\P{miscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\P{^miscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\p{_miscellaneous_SYMBOLS_And_pictographs}', "");
Expect(0, 128511, '\p{^_miscellaneous_SYMBOLS_And_pictographs}', "");
Expect(0, 128511, '\P{_miscellaneous_SYMBOLS_And_pictographs}', "");
Expect(1, 128511, '\P{^_miscellaneous_SYMBOLS_And_pictographs}', "");
Expect(0, 128512, '\p{_miscellaneous_SYMBOLS_And_pictographs}', "");
Expect(1, 128512, '\p{^_miscellaneous_SYMBOLS_And_pictographs}', "");
Expect(1, 128512, '\P{_miscellaneous_SYMBOLS_And_pictographs}', "");
Expect(0, 128512, '\P{^_miscellaneous_SYMBOLS_And_pictographs}', "");
Error('\p{/a/	_Is_Miscellaneous_SYMBOLS_and_PICTOGRAPHS}');
Error('\P{/a/	_Is_Miscellaneous_SYMBOLS_and_PICTOGRAPHS}');
Expect(1, 128511, '\p{ismiscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\p{^ismiscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\P{ismiscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\P{^ismiscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\p{ismiscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\p{^ismiscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\P{ismiscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\P{^ismiscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\p{ IS_MISCELLANEOUS_symbols_And_Pictographs}', "");
Expect(0, 128511, '\p{^ IS_MISCELLANEOUS_symbols_And_Pictographs}', "");
Expect(0, 128511, '\P{ IS_MISCELLANEOUS_symbols_And_Pictographs}', "");
Expect(1, 128511, '\P{^ IS_MISCELLANEOUS_symbols_And_Pictographs}', "");
Expect(0, 128512, '\p{ IS_MISCELLANEOUS_symbols_And_Pictographs}', "");
Expect(1, 128512, '\p{^ IS_MISCELLANEOUS_symbols_And_Pictographs}', "");
Expect(1, 128512, '\P{ IS_MISCELLANEOUS_symbols_And_Pictographs}', "");
Expect(0, 128512, '\P{^ IS_MISCELLANEOUS_symbols_And_Pictographs}', "");
Error('\p{:=_ In_Miscellaneous_symbols_AND_Pictographs}');
Error('\P{:=_ In_Miscellaneous_symbols_AND_Pictographs}');
Expect(1, 128511, '\p{inmiscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\p{^inmiscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\P{inmiscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\P{^inmiscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\p{inmiscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\p{^inmiscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\P{inmiscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\P{^inmiscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\p{		in_Miscellaneous_Symbols_and_PICTOGRAPHS}', "");
Expect(0, 128511, '\p{^		in_Miscellaneous_Symbols_and_PICTOGRAPHS}', "");
Expect(0, 128511, '\P{		in_Miscellaneous_Symbols_and_PICTOGRAPHS}', "");
Expect(1, 128511, '\P{^		in_Miscellaneous_Symbols_and_PICTOGRAPHS}', "");
Expect(0, 128512, '\p{		in_Miscellaneous_Symbols_and_PICTOGRAPHS}', "");
Expect(1, 128512, '\p{^		in_Miscellaneous_Symbols_and_PICTOGRAPHS}', "");
Expect(1, 128512, '\P{		in_Miscellaneous_Symbols_and_PICTOGRAPHS}', "");
Expect(0, 128512, '\P{^		in_Miscellaneous_Symbols_and_PICTOGRAPHS}', "");
Error('\p{ -Misc_pictographs/a/}');
Error('\P{ -Misc_pictographs/a/}');
Expect(1, 128511, '\p{miscpictographs}', "");
Expect(0, 128511, '\p{^miscpictographs}', "");
Expect(0, 128511, '\P{miscpictographs}', "");
Expect(1, 128511, '\P{^miscpictographs}', "");
Expect(0, 128512, '\p{miscpictographs}', "");
Expect(1, 128512, '\p{^miscpictographs}', "");
Expect(1, 128512, '\P{miscpictographs}', "");
Expect(0, 128512, '\P{^miscpictographs}', "");
Expect(1, 128511, '\p{ -Misc_PICTOGRAPHS}', "");
Expect(0, 128511, '\p{^ -Misc_PICTOGRAPHS}', "");
Expect(0, 128511, '\P{ -Misc_PICTOGRAPHS}', "");
Expect(1, 128511, '\P{^ -Misc_PICTOGRAPHS}', "");
Expect(0, 128512, '\p{ -Misc_PICTOGRAPHS}', "");
Expect(1, 128512, '\p{^ -Misc_PICTOGRAPHS}', "");
Expect(1, 128512, '\P{ -Misc_PICTOGRAPHS}', "");
Expect(0, 128512, '\P{^ -Misc_PICTOGRAPHS}', "");
Error('\p{-/a/IS_Misc_Pictographs}');
Error('\P{-/a/IS_Misc_Pictographs}');
Expect(1, 128511, '\p{ismiscpictographs}', "");
Expect(0, 128511, '\p{^ismiscpictographs}', "");
Expect(0, 128511, '\P{ismiscpictographs}', "");
Expect(1, 128511, '\P{^ismiscpictographs}', "");
Expect(0, 128512, '\p{ismiscpictographs}', "");
Expect(1, 128512, '\p{^ismiscpictographs}', "");
Expect(1, 128512, '\P{ismiscpictographs}', "");
Expect(0, 128512, '\P{^ismiscpictographs}', "");
Expect(1, 128511, '\p{	Is_Misc_pictographs}', "");
Expect(0, 128511, '\p{^	Is_Misc_pictographs}', "");
Expect(0, 128511, '\P{	Is_Misc_pictographs}', "");
Expect(1, 128511, '\P{^	Is_Misc_pictographs}', "");
Expect(0, 128512, '\p{	Is_Misc_pictographs}', "");
Expect(1, 128512, '\p{^	Is_Misc_pictographs}', "");
Expect(1, 128512, '\P{	Is_Misc_pictographs}', "");
Expect(0, 128512, '\P{^	Is_Misc_pictographs}', "");
Error('\p{-:=In_MISC_Pictographs}');
Error('\P{-:=In_MISC_Pictographs}');
Expect(1, 128511, '\p{inmiscpictographs}', "");
Expect(0, 128511, '\p{^inmiscpictographs}', "");
Expect(0, 128511, '\P{inmiscpictographs}', "");
Expect(1, 128511, '\P{^inmiscpictographs}', "");
Expect(0, 128512, '\p{inmiscpictographs}', "");
Expect(1, 128512, '\p{^inmiscpictographs}', "");
Expect(1, 128512, '\P{inmiscpictographs}', "");
Expect(0, 128512, '\P{^inmiscpictographs}', "");
Expect(1, 128511, '\p{	_in_Misc_pictographs}', "");
Expect(0, 128511, '\p{^	_in_Misc_pictographs}', "");
Expect(0, 128511, '\P{	_in_Misc_pictographs}', "");
Expect(1, 128511, '\P{^	_in_Misc_pictographs}', "");
Expect(0, 128512, '\p{	_in_Misc_pictographs}', "");
Expect(1, 128512, '\p{^	_in_Misc_pictographs}', "");
Expect(1, 128512, '\P{	_in_Misc_pictographs}', "");
Expect(0, 128512, '\P{^	_in_Misc_pictographs}', "");
Error('\p{/a/ _miscellaneous_Technical}');
Error('\P{/a/ _miscellaneous_Technical}');
Expect(1, 9215, '\p{miscellaneoustechnical}', "");
Expect(0, 9215, '\p{^miscellaneoustechnical}', "");
Expect(0, 9215, '\P{miscellaneoustechnical}', "");
Expect(1, 9215, '\P{^miscellaneoustechnical}', "");
Expect(0, 9216, '\p{miscellaneoustechnical}', "");
Expect(1, 9216, '\p{^miscellaneoustechnical}', "");
Expect(1, 9216, '\P{miscellaneoustechnical}', "");
Expect(0, 9216, '\P{^miscellaneoustechnical}', "");
Expect(1, 9215, '\p{- miscellaneous_Technical}', "");
Expect(0, 9215, '\p{^- miscellaneous_Technical}', "");
Expect(0, 9215, '\P{- miscellaneous_Technical}', "");
Expect(1, 9215, '\P{^- miscellaneous_Technical}', "");
Expect(0, 9216, '\p{- miscellaneous_Technical}', "");
Expect(1, 9216, '\p{^- miscellaneous_Technical}', "");
Expect(1, 9216, '\P{- miscellaneous_Technical}', "");
Expect(0, 9216, '\P{^- miscellaneous_Technical}', "");
Error('\p{/a/_-IS_Miscellaneous_Technical}');
Error('\P{/a/_-IS_Miscellaneous_Technical}');
Expect(1, 9215, '\p{ismiscellaneoustechnical}', "");
Expect(0, 9215, '\p{^ismiscellaneoustechnical}', "");
Expect(0, 9215, '\P{ismiscellaneoustechnical}', "");
Expect(1, 9215, '\P{^ismiscellaneoustechnical}', "");
Expect(0, 9216, '\p{ismiscellaneoustechnical}', "");
Expect(1, 9216, '\p{^ismiscellaneoustechnical}', "");
Expect(1, 9216, '\P{ismiscellaneoustechnical}', "");
Expect(0, 9216, '\P{^ismiscellaneoustechnical}', "");
Expect(1, 9215, '\p{	_Is_Miscellaneous_technical}', "");
Expect(0, 9215, '\p{^	_Is_Miscellaneous_technical}', "");
Expect(0, 9215, '\P{	_Is_Miscellaneous_technical}', "");
Expect(1, 9215, '\P{^	_Is_Miscellaneous_technical}', "");
Expect(0, 9216, '\p{	_Is_Miscellaneous_technical}', "");
Expect(1, 9216, '\p{^	_Is_Miscellaneous_technical}', "");
Expect(1, 9216, '\P{	_Is_Miscellaneous_technical}', "");
Expect(0, 9216, '\P{^	_Is_Miscellaneous_technical}', "");
Error('\p{  in_miscellaneous_technical:=}');
Error('\P{  in_miscellaneous_technical:=}');
Expect(1, 9215, '\p{inmiscellaneoustechnical}', "");
Expect(0, 9215, '\p{^inmiscellaneoustechnical}', "");
Expect(0, 9215, '\P{inmiscellaneoustechnical}', "");
Expect(1, 9215, '\P{^inmiscellaneoustechnical}', "");
Expect(0, 9216, '\p{inmiscellaneoustechnical}', "");
Expect(1, 9216, '\p{^inmiscellaneoustechnical}', "");
Expect(1, 9216, '\P{inmiscellaneoustechnical}', "");
Expect(0, 9216, '\P{^inmiscellaneoustechnical}', "");
Expect(1, 9215, '\p{ In_Miscellaneous_technical}', "");
Expect(0, 9215, '\p{^ In_Miscellaneous_technical}', "");
Expect(0, 9215, '\P{ In_Miscellaneous_technical}', "");
Expect(1, 9215, '\P{^ In_Miscellaneous_technical}', "");
Expect(0, 9216, '\p{ In_Miscellaneous_technical}', "");
Expect(1, 9216, '\p{^ In_Miscellaneous_technical}', "");
Expect(1, 9216, '\P{ In_Miscellaneous_technical}', "");
Expect(0, 9216, '\P{^ In_Miscellaneous_technical}', "");
Error('\p{/a/-misc_technical}');
Error('\P{/a/-misc_technical}');
Expect(1, 9215, '\p{misctechnical}', "");
Expect(0, 9215, '\p{^misctechnical}', "");
Expect(0, 9215, '\P{misctechnical}', "");
Expect(1, 9215, '\P{^misctechnical}', "");
Expect(0, 9216, '\p{misctechnical}', "");
Expect(1, 9216, '\p{^misctechnical}', "");
Expect(1, 9216, '\P{misctechnical}', "");
Expect(0, 9216, '\P{^misctechnical}', "");
Expect(1, 9215, '\p{_Misc_TECHNICAL}', "");
Expect(0, 9215, '\p{^_Misc_TECHNICAL}', "");
Expect(0, 9215, '\P{_Misc_TECHNICAL}', "");
Expect(1, 9215, '\P{^_Misc_TECHNICAL}', "");
Expect(0, 9216, '\p{_Misc_TECHNICAL}', "");
Expect(1, 9216, '\p{^_Misc_TECHNICAL}', "");
Expect(1, 9216, '\P{_Misc_TECHNICAL}', "");
Expect(0, 9216, '\P{^_Misc_TECHNICAL}', "");
Error('\p{/a/ is_MISC_Technical}');
Error('\P{/a/ is_MISC_Technical}');
Expect(1, 9215, '\p{ismisctechnical}', "");
Expect(0, 9215, '\p{^ismisctechnical}', "");
Expect(0, 9215, '\P{ismisctechnical}', "");
Expect(1, 9215, '\P{^ismisctechnical}', "");
Expect(0, 9216, '\p{ismisctechnical}', "");
Expect(1, 9216, '\p{^ismisctechnical}', "");
Expect(1, 9216, '\P{ismisctechnical}', "");
Expect(0, 9216, '\P{^ismisctechnical}', "");
Expect(1, 9215, '\p{_Is_misc_Technical}', "");
Expect(0, 9215, '\p{^_Is_misc_Technical}', "");
Expect(0, 9215, '\P{_Is_misc_Technical}', "");
Expect(1, 9215, '\P{^_Is_misc_Technical}', "");
Expect(0, 9216, '\p{_Is_misc_Technical}', "");
Expect(1, 9216, '\p{^_Is_misc_Technical}', "");
Expect(1, 9216, '\P{_Is_misc_Technical}', "");
Expect(0, 9216, '\P{^_Is_misc_Technical}', "");
Error('\p{ IN_misc_Technical:=}');
Error('\P{ IN_misc_Technical:=}');
Expect(1, 9215, '\p{inmisctechnical}', "");
Expect(0, 9215, '\p{^inmisctechnical}', "");
Expect(0, 9215, '\P{inmisctechnical}', "");
Expect(1, 9215, '\P{^inmisctechnical}', "");
Expect(0, 9216, '\p{inmisctechnical}', "");
Expect(1, 9216, '\p{^inmisctechnical}', "");
Expect(1, 9216, '\P{inmisctechnical}', "");
Expect(0, 9216, '\P{^inmisctechnical}', "");
Expect(1, 9215, '\p{ _In_misc_TECHNICAL}', "");
Expect(0, 9215, '\p{^ _In_misc_TECHNICAL}', "");
Expect(0, 9215, '\P{ _In_misc_TECHNICAL}', "");
Expect(1, 9215, '\P{^ _In_misc_TECHNICAL}', "");
Expect(0, 9216, '\p{ _In_misc_TECHNICAL}', "");
Expect(1, 9216, '\p{^ _In_misc_TECHNICAL}', "");
Expect(1, 9216, '\P{ _In_misc_TECHNICAL}', "");
Expect(0, 9216, '\P{^ _In_misc_TECHNICAL}', "");
Error('\p{-/a/modifier_Letter}');
Error('\P{-/a/modifier_Letter}');
Expect(1, 94111, '\p{modifierletter}', "");
Expect(0, 94111, '\p{^modifierletter}', "");
Expect(0, 94111, '\P{modifierletter}', "");
Expect(1, 94111, '\P{^modifierletter}', "");
Expect(0, 94112, '\p{modifierletter}', "");
Expect(1, 94112, '\p{^modifierletter}', "");
Expect(1, 94112, '\P{modifierletter}', "");
Expect(0, 94112, '\P{^modifierletter}', "");
Expect(1, 94111, '\p{_ MODIFIER_LETTER}', "");
Expect(0, 94111, '\p{^_ MODIFIER_LETTER}', "");
Expect(0, 94111, '\P{_ MODIFIER_LETTER}', "");
Expect(1, 94111, '\P{^_ MODIFIER_LETTER}', "");
Expect(0, 94112, '\p{_ MODIFIER_LETTER}', "");
Expect(1, 94112, '\p{^_ MODIFIER_LETTER}', "");
Expect(1, 94112, '\P{_ MODIFIER_LETTER}', "");
Expect(0, 94112, '\P{^_ MODIFIER_LETTER}', "");
Error('\p{	 IS_Modifier_LETTER:=}');
Error('\P{	 IS_Modifier_LETTER:=}');
Expect(1, 94111, '\p{ismodifierletter}', "");
Expect(0, 94111, '\p{^ismodifierletter}', "");
Expect(0, 94111, '\P{ismodifierletter}', "");
Expect(1, 94111, '\P{^ismodifierletter}', "");
Expect(0, 94112, '\p{ismodifierletter}', "");
Expect(1, 94112, '\p{^ismodifierletter}', "");
Expect(1, 94112, '\P{ismodifierletter}', "");
Expect(0, 94112, '\P{^ismodifierletter}', "");
Expect(1, 94111, '\p{-IS_modifier_letter}', "");
Expect(0, 94111, '\p{^-IS_modifier_letter}', "");
Expect(0, 94111, '\P{-IS_modifier_letter}', "");
Expect(1, 94111, '\P{^-IS_modifier_letter}', "");
Expect(0, 94112, '\p{-IS_modifier_letter}', "");
Expect(1, 94112, '\p{^-IS_modifier_letter}', "");
Expect(1, 94112, '\P{-IS_modifier_letter}', "");
Expect(0, 94112, '\P{^-IS_modifier_letter}', "");
Error('\p{ -Lm/a/}');
Error('\P{ -Lm/a/}');
Expect(1, 94111, '\p{lm}', "");
Expect(0, 94111, '\p{^lm}', "");
Expect(0, 94111, '\P{lm}', "");
Expect(1, 94111, '\P{^lm}', "");
Expect(0, 94112, '\p{lm}', "");
Expect(1, 94112, '\p{^lm}', "");
Expect(1, 94112, '\P{lm}', "");
Expect(0, 94112, '\P{^lm}', "");
Expect(1, 94111, '\p{ _lm}', "");
Expect(0, 94111, '\p{^ _lm}', "");
Expect(0, 94111, '\P{ _lm}', "");
Expect(1, 94111, '\P{^ _lm}', "");
Expect(0, 94112, '\p{ _lm}', "");
Expect(1, 94112, '\p{^ _lm}', "");
Expect(1, 94112, '\P{ _lm}', "");
Expect(0, 94112, '\P{^ _lm}', "");
Error('\p{-_Is_Lm/a/}');
Error('\P{-_Is_Lm/a/}');
Expect(1, 94111, '\p{islm}', "");
Expect(0, 94111, '\p{^islm}', "");
Expect(0, 94111, '\P{islm}', "");
Expect(1, 94111, '\P{^islm}', "");
Expect(0, 94112, '\p{islm}', "");
Expect(1, 94112, '\p{^islm}', "");
Expect(1, 94112, '\P{islm}', "");
Expect(0, 94112, '\P{^islm}', "");
Expect(1, 94111, '\p{-Is_lm}', "");
Expect(0, 94111, '\p{^-Is_lm}', "");
Expect(0, 94111, '\P{-Is_lm}', "");
Expect(1, 94111, '\P{^-Is_lm}', "");
Expect(0, 94112, '\p{-Is_lm}', "");
Expect(1, 94112, '\p{^-Is_lm}', "");
Expect(1, 94112, '\P{-Is_lm}', "");
Expect(0, 94112, '\P{^-Is_lm}', "");
Error('\p{:=		Modifier_Symbol}');
Error('\P{:=		Modifier_Symbol}');
Expect(1, 65507, '\p{modifiersymbol}', "");
Expect(0, 65507, '\p{^modifiersymbol}', "");
Expect(0, 65507, '\P{modifiersymbol}', "");
Expect(1, 65507, '\P{^modifiersymbol}', "");
Expect(0, 65508, '\p{modifiersymbol}', "");
Expect(1, 65508, '\p{^modifiersymbol}', "");
Expect(1, 65508, '\P{modifiersymbol}', "");
Expect(0, 65508, '\P{^modifiersymbol}', "");
Expect(1, 65507, '\p{_MODIFIER_Symbol}', "");
Expect(0, 65507, '\p{^_MODIFIER_Symbol}', "");
Expect(0, 65507, '\P{_MODIFIER_Symbol}', "");
Expect(1, 65507, '\P{^_MODIFIER_Symbol}', "");
Expect(0, 65508, '\p{_MODIFIER_Symbol}', "");
Expect(1, 65508, '\p{^_MODIFIER_Symbol}', "");
Expect(1, 65508, '\P{_MODIFIER_Symbol}', "");
Expect(0, 65508, '\P{^_MODIFIER_Symbol}', "");
Error('\p{/a/__Is_Modifier_Symbol}');
Error('\P{/a/__Is_Modifier_Symbol}');
Expect(1, 65507, '\p{ismodifiersymbol}', "");
Expect(0, 65507, '\p{^ismodifiersymbol}', "");
Expect(0, 65507, '\P{ismodifiersymbol}', "");
Expect(1, 65507, '\P{^ismodifiersymbol}', "");
Expect(0, 65508, '\p{ismodifiersymbol}', "");
Expect(1, 65508, '\p{^ismodifiersymbol}', "");
Expect(1, 65508, '\P{ismodifiersymbol}', "");
Expect(0, 65508, '\P{^ismodifiersymbol}', "");
Expect(1, 65507, '\p{-_IS_Modifier_SYMBOL}', "");
Expect(0, 65507, '\p{^-_IS_Modifier_SYMBOL}', "");
Expect(0, 65507, '\P{-_IS_Modifier_SYMBOL}', "");
Expect(1, 65507, '\P{^-_IS_Modifier_SYMBOL}', "");
Expect(0, 65508, '\p{-_IS_Modifier_SYMBOL}', "");
Expect(1, 65508, '\p{^-_IS_Modifier_SYMBOL}', "");
Expect(1, 65508, '\P{-_IS_Modifier_SYMBOL}', "");
Expect(0, 65508, '\P{^-_IS_Modifier_SYMBOL}', "");
Error('\p{:=	_sk}');
Error('\P{:=	_sk}');
Expect(1, 65507, '\p{sk}', "");
Expect(0, 65507, '\p{^sk}', "");
Expect(0, 65507, '\P{sk}', "");
Expect(1, 65507, '\P{^sk}', "");
Expect(0, 65508, '\p{sk}', "");
Expect(1, 65508, '\p{^sk}', "");
Expect(1, 65508, '\P{sk}', "");
Expect(0, 65508, '\P{^sk}', "");
Expect(1, 65507, '\p{-_SK}', "");
Expect(0, 65507, '\p{^-_SK}', "");
Expect(0, 65507, '\P{-_SK}', "");
Expect(1, 65507, '\P{^-_SK}', "");
Expect(0, 65508, '\p{-_SK}', "");
Expect(1, 65508, '\p{^-_SK}', "");
Expect(1, 65508, '\P{-_SK}', "");
Expect(0, 65508, '\P{^-_SK}', "");
Error('\p{	 Is_sk/a/}');
Error('\P{	 Is_sk/a/}');
Expect(1, 65507, '\p{issk}', "");
Expect(0, 65507, '\p{^issk}', "");
Expect(0, 65507, '\P{issk}', "");
Expect(1, 65507, '\P{^issk}', "");
Expect(0, 65508, '\p{issk}', "");
Expect(1, 65508, '\p{^issk}', "");
Expect(1, 65508, '\P{issk}', "");
Expect(0, 65508, '\P{^issk}', "");
Expect(1, 65507, '\p{_is_Sk}', "");
Expect(0, 65507, '\p{^_is_Sk}', "");
Expect(0, 65507, '\P{_is_Sk}', "");
Expect(1, 65507, '\P{^_is_Sk}', "");
Expect(0, 65508, '\p{_is_Sk}', "");
Expect(1, 65508, '\p{^_is_Sk}', "");
Expect(1, 65508, '\P{_is_Sk}', "");
Expect(0, 65508, '\P{^_is_Sk}', "");
Error('\p{	Modifier_tone_Letters:=}');
Error('\P{	Modifier_tone_Letters:=}');
Expect(1, 42783, '\p{modifiertoneletters}', "");
Expect(0, 42783, '\p{^modifiertoneletters}', "");
Expect(0, 42783, '\P{modifiertoneletters}', "");
Expect(1, 42783, '\P{^modifiertoneletters}', "");
Expect(0, 42784, '\p{modifiertoneletters}', "");
Expect(1, 42784, '\p{^modifiertoneletters}', "");
Expect(1, 42784, '\P{modifiertoneletters}', "");
Expect(0, 42784, '\P{^modifiertoneletters}', "");
Expect(1, 42783, '\p{ modifier_tone_LETTERS}', "");
Expect(0, 42783, '\p{^ modifier_tone_LETTERS}', "");
Expect(0, 42783, '\P{ modifier_tone_LETTERS}', "");
Expect(1, 42783, '\P{^ modifier_tone_LETTERS}', "");
Expect(0, 42784, '\p{ modifier_tone_LETTERS}', "");
Expect(1, 42784, '\p{^ modifier_tone_LETTERS}', "");
Expect(1, 42784, '\P{ modifier_tone_LETTERS}', "");
Expect(0, 42784, '\P{^ modifier_tone_LETTERS}', "");
Error('\p{		Is_Modifier_TONE_Letters:=}');
Error('\P{		Is_Modifier_TONE_Letters:=}');
Expect(1, 42783, '\p{ismodifiertoneletters}', "");
Expect(0, 42783, '\p{^ismodifiertoneletters}', "");
Expect(0, 42783, '\P{ismodifiertoneletters}', "");
Expect(1, 42783, '\P{^ismodifiertoneletters}', "");
Expect(0, 42784, '\p{ismodifiertoneletters}', "");
Expect(1, 42784, '\p{^ismodifiertoneletters}', "");
Expect(1, 42784, '\P{ismodifiertoneletters}', "");
Expect(0, 42784, '\P{^ismodifiertoneletters}', "");
Expect(1, 42783, '\p{_is_Modifier_TONE_Letters}', "");
Expect(0, 42783, '\p{^_is_Modifier_TONE_Letters}', "");
Expect(0, 42783, '\P{_is_Modifier_TONE_Letters}', "");
Expect(1, 42783, '\P{^_is_Modifier_TONE_Letters}', "");
Expect(0, 42784, '\p{_is_Modifier_TONE_Letters}', "");
Expect(1, 42784, '\p{^_is_Modifier_TONE_Letters}', "");
Expect(1, 42784, '\P{_is_Modifier_TONE_Letters}', "");
Expect(0, 42784, '\P{^_is_Modifier_TONE_Letters}', "");
Error('\p{/a/	-in_modifier_Tone_letters}');
Error('\P{/a/	-in_modifier_Tone_letters}');
Expect(1, 42783, '\p{inmodifiertoneletters}', "");
Expect(0, 42783, '\p{^inmodifiertoneletters}', "");
Expect(0, 42783, '\P{inmodifiertoneletters}', "");
Expect(1, 42783, '\P{^inmodifiertoneletters}', "");
Expect(0, 42784, '\p{inmodifiertoneletters}', "");
Expect(1, 42784, '\p{^inmodifiertoneletters}', "");
Expect(1, 42784, '\P{inmodifiertoneletters}', "");
Expect(0, 42784, '\P{^inmodifiertoneletters}', "");
Expect(1, 42783, '\p{	 In_MODIFIER_Tone_Letters}', "");
Expect(0, 42783, '\p{^	 In_MODIFIER_Tone_Letters}', "");
Expect(0, 42783, '\P{	 In_MODIFIER_Tone_Letters}', "");
Expect(1, 42783, '\P{^	 In_MODIFIER_Tone_Letters}', "");
Expect(0, 42784, '\p{	 In_MODIFIER_Tone_Letters}', "");
Expect(1, 42784, '\p{^	 In_MODIFIER_Tone_Letters}', "");
Expect(1, 42784, '\P{	 In_MODIFIER_Tone_Letters}', "");
Expect(0, 42784, '\P{^	 In_MODIFIER_Tone_Letters}', "");
Error('\p{-mongolian:=}');
Error('\P{-mongolian:=}');
Expect(1, 6314, '\p{mongolian}', "");
Expect(0, 6314, '\p{^mongolian}', "");
Expect(0, 6314, '\P{mongolian}', "");
Expect(1, 6314, '\P{^mongolian}', "");
Expect(0, 6315, '\p{mongolian}', "");
Expect(1, 6315, '\p{^mongolian}', "");
Expect(1, 6315, '\P{mongolian}', "");
Expect(0, 6315, '\P{^mongolian}', "");
Expect(1, 6314, '\p{	_MONGOLIAN}', "");
Expect(0, 6314, '\p{^	_MONGOLIAN}', "");
Expect(0, 6314, '\P{	_MONGOLIAN}', "");
Expect(1, 6314, '\P{^	_MONGOLIAN}', "");
Expect(0, 6315, '\p{	_MONGOLIAN}', "");
Expect(1, 6315, '\p{^	_MONGOLIAN}', "");
Expect(1, 6315, '\P{	_MONGOLIAN}', "");
Expect(0, 6315, '\P{^	_MONGOLIAN}', "");
Error('\p{ :=IS_Mongolian}');
Error('\P{ :=IS_Mongolian}');
Expect(1, 6314, '\p{ismongolian}', "");
Expect(0, 6314, '\p{^ismongolian}', "");
Expect(0, 6314, '\P{ismongolian}', "");
Expect(1, 6314, '\P{^ismongolian}', "");
Expect(0, 6315, '\p{ismongolian}', "");
Expect(1, 6315, '\p{^ismongolian}', "");
Expect(1, 6315, '\P{ismongolian}', "");
Expect(0, 6315, '\P{^ismongolian}', "");
Expect(1, 6314, '\p{_	is_mongolian}', "");
Expect(0, 6314, '\p{^_	is_mongolian}', "");
Expect(0, 6314, '\P{_	is_mongolian}', "");
Expect(1, 6314, '\P{^_	is_mongolian}', "");
Expect(0, 6315, '\p{_	is_mongolian}', "");
Expect(1, 6315, '\p{^_	is_mongolian}', "");
Expect(1, 6315, '\P{_	is_mongolian}', "");
Expect(0, 6315, '\P{^_	is_mongolian}', "");
Error('\p{:=		mong}');
Error('\P{:=		mong}');
Expect(1, 6314, '\p{mong}', "");
Expect(0, 6314, '\p{^mong}', "");
Expect(0, 6314, '\P{mong}', "");
Expect(1, 6314, '\P{^mong}', "");
Expect(0, 6315, '\p{mong}', "");
Expect(1, 6315, '\p{^mong}', "");
Expect(1, 6315, '\P{mong}', "");
Expect(0, 6315, '\P{^mong}', "");
Expect(1, 6314, '\p{	mong}', "");
Expect(0, 6314, '\p{^	mong}', "");
Expect(0, 6314, '\P{	mong}', "");
Expect(1, 6314, '\P{^	mong}', "");
Expect(0, 6315, '\p{	mong}', "");
Expect(1, 6315, '\p{^	mong}', "");
Expect(1, 6315, '\P{	mong}', "");
Expect(0, 6315, '\P{^	mong}', "");
Error('\p{/a/	 is_mong}');
Error('\P{/a/	 is_mong}');
Expect(1, 6314, '\p{ismong}', "");
Expect(0, 6314, '\p{^ismong}', "");
Expect(0, 6314, '\P{ismong}', "");
Expect(1, 6314, '\P{^ismong}', "");
Expect(0, 6315, '\p{ismong}', "");
Expect(1, 6315, '\p{^ismong}', "");
Expect(1, 6315, '\P{ismong}', "");
Expect(0, 6315, '\P{^ismong}', "");
Expect(1, 6314, '\p{ 	IS_MONG}', "");
Expect(0, 6314, '\p{^ 	IS_MONG}', "");
Expect(0, 6314, '\P{ 	IS_MONG}', "");
Expect(1, 6314, '\P{^ 	IS_MONG}', "");
Expect(0, 6315, '\p{ 	IS_MONG}', "");
Expect(1, 6315, '\p{^ 	IS_MONG}', "");
Expect(1, 6315, '\P{ 	IS_MONG}', "");
Expect(0, 6315, '\P{^ 	IS_MONG}', "");
Error('\p{/a/Musical_Symbols}');
Error('\P{/a/Musical_Symbols}');
Expect(1, 119295, '\p{musicalsymbols}', "");
Expect(0, 119295, '\p{^musicalsymbols}', "");
Expect(0, 119295, '\P{musicalsymbols}', "");
Expect(1, 119295, '\P{^musicalsymbols}', "");
Expect(0, 119296, '\p{musicalsymbols}', "");
Expect(1, 119296, '\p{^musicalsymbols}', "");
Expect(1, 119296, '\P{musicalsymbols}', "");
Expect(0, 119296, '\P{^musicalsymbols}', "");
Expect(1, 119295, '\p{	-MUSICAL_Symbols}', "");
Expect(0, 119295, '\p{^	-MUSICAL_Symbols}', "");
Expect(0, 119295, '\P{	-MUSICAL_Symbols}', "");
Expect(1, 119295, '\P{^	-MUSICAL_Symbols}', "");
Expect(0, 119296, '\p{	-MUSICAL_Symbols}', "");
Expect(1, 119296, '\p{^	-MUSICAL_Symbols}', "");
Expect(1, 119296, '\P{	-MUSICAL_Symbols}', "");
Expect(0, 119296, '\P{^	-MUSICAL_Symbols}', "");
Error('\p{	Is_Musical_SYMBOLS/a/}');
Error('\P{	Is_Musical_SYMBOLS/a/}');
Expect(1, 119295, '\p{ismusicalsymbols}', "");
Expect(0, 119295, '\p{^ismusicalsymbols}', "");
Expect(0, 119295, '\P{ismusicalsymbols}', "");
Expect(1, 119295, '\P{^ismusicalsymbols}', "");
Expect(0, 119296, '\p{ismusicalsymbols}', "");
Expect(1, 119296, '\p{^ismusicalsymbols}', "");
Expect(1, 119296, '\P{ismusicalsymbols}', "");
Expect(0, 119296, '\P{^ismusicalsymbols}', "");
Expect(1, 119295, '\p{- IS_Musical_Symbols}', "");
Expect(0, 119295, '\p{^- IS_Musical_Symbols}', "");
Expect(0, 119295, '\P{- IS_Musical_Symbols}', "");
Expect(1, 119295, '\P{^- IS_Musical_Symbols}', "");
Expect(0, 119296, '\p{- IS_Musical_Symbols}', "");
Expect(1, 119296, '\p{^- IS_Musical_Symbols}', "");
Expect(1, 119296, '\P{- IS_Musical_Symbols}', "");
Expect(0, 119296, '\P{^- IS_Musical_Symbols}', "");
Error('\p{ /a/In_Musical_symbols}');
Error('\P{ /a/In_Musical_symbols}');
Expect(1, 119295, '\p{inmusicalsymbols}', "");
Expect(0, 119295, '\p{^inmusicalsymbols}', "");
Expect(0, 119295, '\P{inmusicalsymbols}', "");
Expect(1, 119295, '\P{^inmusicalsymbols}', "");
Expect(0, 119296, '\p{inmusicalsymbols}', "");
Expect(1, 119296, '\p{^inmusicalsymbols}', "");
Expect(1, 119296, '\P{inmusicalsymbols}', "");
Expect(0, 119296, '\P{^inmusicalsymbols}', "");
Expect(1, 119295, '\p{ 	In_MUSICAL_SYMBOLS}', "");
Expect(0, 119295, '\p{^ 	In_MUSICAL_SYMBOLS}', "");
Expect(0, 119295, '\P{ 	In_MUSICAL_SYMBOLS}', "");
Expect(1, 119295, '\P{^ 	In_MUSICAL_SYMBOLS}', "");
Expect(0, 119296, '\p{ 	In_MUSICAL_SYMBOLS}', "");
Expect(1, 119296, '\p{^ 	In_MUSICAL_SYMBOLS}', "");
Expect(1, 119296, '\P{ 	In_MUSICAL_SYMBOLS}', "");
Expect(0, 119296, '\P{^ 	In_MUSICAL_SYMBOLS}', "");
Error('\p{/a/_-music}');
Error('\P{/a/_-music}');
Expect(1, 119295, '\p{music}', "");
Expect(0, 119295, '\p{^music}', "");
Expect(0, 119295, '\P{music}', "");
Expect(1, 119295, '\P{^music}', "");
Expect(0, 119296, '\p{music}', "");
Expect(1, 119296, '\p{^music}', "");
Expect(1, 119296, '\P{music}', "");
Expect(0, 119296, '\P{^music}', "");
Expect(1, 119295, '\p{-music}', "");
Expect(0, 119295, '\p{^-music}', "");
Expect(0, 119295, '\P{-music}', "");
Expect(1, 119295, '\P{^-music}', "");
Expect(0, 119296, '\p{-music}', "");
Expect(1, 119296, '\p{^-music}', "");
Expect(1, 119296, '\P{-music}', "");
Expect(0, 119296, '\P{^-music}', "");
Error('\p{/a/-Is_Music}');
Error('\P{/a/-Is_Music}');
Expect(1, 119295, '\p{ismusic}', "");
Expect(0, 119295, '\p{^ismusic}', "");
Expect(0, 119295, '\P{ismusic}', "");
Expect(1, 119295, '\P{^ismusic}', "");
Expect(0, 119296, '\p{ismusic}', "");
Expect(1, 119296, '\p{^ismusic}', "");
Expect(1, 119296, '\P{ismusic}', "");
Expect(0, 119296, '\P{^ismusic}', "");
Expect(1, 119295, '\p{	IS_MUSIC}', "");
Expect(0, 119295, '\p{^	IS_MUSIC}', "");
Expect(0, 119295, '\P{	IS_MUSIC}', "");
Expect(1, 119295, '\P{^	IS_MUSIC}', "");
Expect(0, 119296, '\p{	IS_MUSIC}', "");
Expect(1, 119296, '\p{^	IS_MUSIC}', "");
Expect(1, 119296, '\P{	IS_MUSIC}', "");
Expect(0, 119296, '\P{^	IS_MUSIC}', "");
Error('\p{ /a/in_music}');
Error('\P{ /a/in_music}');
Expect(1, 119295, '\p{inmusic}', "");
Expect(0, 119295, '\p{^inmusic}', "");
Expect(0, 119295, '\P{inmusic}', "");
Expect(1, 119295, '\P{^inmusic}', "");
Expect(0, 119296, '\p{inmusic}', "");
Expect(1, 119296, '\p{^inmusic}', "");
Expect(1, 119296, '\P{inmusic}', "");
Expect(0, 119296, '\P{^inmusic}', "");
Expect(1, 119295, '\p{ -IN_music}', "");
Expect(0, 119295, '\p{^ -IN_music}', "");
Expect(0, 119295, '\P{ -IN_music}', "");
Expect(1, 119295, '\P{^ -IN_music}', "");
Expect(0, 119296, '\p{ -IN_music}', "");
Expect(1, 119296, '\p{^ -IN_music}', "");
Expect(1, 119296, '\P{ -IN_music}', "");
Expect(0, 119296, '\P{^ -IN_music}', "");
Error('\p{_/a/Myanmar}');
Error('\P{_/a/Myanmar}');
Expect(1, 43643, '\p{myanmar}', "");
Expect(0, 43643, '\p{^myanmar}', "");
Expect(0, 43643, '\P{myanmar}', "");
Expect(1, 43643, '\P{^myanmar}', "");
Expect(0, 43644, '\p{myanmar}', "");
Expect(1, 43644, '\p{^myanmar}', "");
Expect(1, 43644, '\P{myanmar}', "");
Expect(0, 43644, '\P{^myanmar}', "");
Expect(1, 43643, '\p{ _Myanmar}', "");
Expect(0, 43643, '\p{^ _Myanmar}', "");
Expect(0, 43643, '\P{ _Myanmar}', "");
Expect(1, 43643, '\P{^ _Myanmar}', "");
Expect(0, 43644, '\p{ _Myanmar}', "");
Expect(1, 43644, '\p{^ _Myanmar}', "");
Expect(1, 43644, '\P{ _Myanmar}', "");
Expect(0, 43644, '\P{^ _Myanmar}', "");
Error('\p{:=- is_Myanmar}');
Error('\P{:=- is_Myanmar}');
Expect(1, 43643, '\p{ismyanmar}', "");
Expect(0, 43643, '\p{^ismyanmar}', "");
Expect(0, 43643, '\P{ismyanmar}', "");
Expect(1, 43643, '\P{^ismyanmar}', "");
Expect(0, 43644, '\p{ismyanmar}', "");
Expect(1, 43644, '\p{^ismyanmar}', "");
Expect(1, 43644, '\P{ismyanmar}', "");
Expect(0, 43644, '\P{^ismyanmar}', "");
Expect(1, 43643, '\p{ 	Is_Myanmar}', "");
Expect(0, 43643, '\p{^ 	Is_Myanmar}', "");
Expect(0, 43643, '\P{ 	Is_Myanmar}', "");
Expect(1, 43643, '\P{^ 	Is_Myanmar}', "");
Expect(0, 43644, '\p{ 	Is_Myanmar}', "");
Expect(1, 43644, '\p{^ 	Is_Myanmar}', "");
Expect(1, 43644, '\P{ 	Is_Myanmar}', "");
Expect(0, 43644, '\P{^ 	Is_Myanmar}', "");
Error('\p{_/a/MYMR}');
Error('\P{_/a/MYMR}');
Expect(1, 43643, '\p{mymr}', "");
Expect(0, 43643, '\p{^mymr}', "");
Expect(0, 43643, '\P{mymr}', "");
Expect(1, 43643, '\P{^mymr}', "");
Expect(0, 43644, '\p{mymr}', "");
Expect(1, 43644, '\p{^mymr}', "");
Expect(1, 43644, '\P{mymr}', "");
Expect(0, 43644, '\P{^mymr}', "");
Expect(1, 43643, '\p{_ mymr}', "");
Expect(0, 43643, '\p{^_ mymr}', "");
Expect(0, 43643, '\P{_ mymr}', "");
Expect(1, 43643, '\P{^_ mymr}', "");
Expect(0, 43644, '\p{_ mymr}', "");
Expect(1, 43644, '\p{^_ mymr}', "");
Expect(1, 43644, '\P{_ mymr}', "");
Expect(0, 43644, '\P{^_ mymr}', "");
Error('\p{_ Is_MYMR/a/}');
Error('\P{_ Is_MYMR/a/}');
Expect(1, 43643, '\p{ismymr}', "");
Expect(0, 43643, '\p{^ismymr}', "");
Expect(0, 43643, '\P{ismymr}', "");
Expect(1, 43643, '\P{^ismymr}', "");
Expect(0, 43644, '\p{ismymr}', "");
Expect(1, 43644, '\p{^ismymr}', "");
Expect(1, 43644, '\P{ismymr}', "");
Expect(0, 43644, '\P{^ismymr}', "");
Expect(1, 43643, '\p{	IS_Mymr}', "");
Expect(0, 43643, '\p{^	IS_Mymr}', "");
Expect(0, 43643, '\P{	IS_Mymr}', "");
Expect(1, 43643, '\P{^	IS_Mymr}', "");
Expect(0, 43644, '\p{	IS_Mymr}', "");
Expect(1, 43644, '\p{^	IS_Mymr}', "");
Expect(1, 43644, '\P{	IS_Mymr}', "");
Expect(0, 43644, '\P{^	IS_Mymr}', "");
Error('\p{:=myanmar_Extended_a}');
Error('\P{:=myanmar_Extended_a}');
Expect(1, 43647, '\p{myanmarextendeda}', "");
Expect(0, 43647, '\p{^myanmarextendeda}', "");
Expect(0, 43647, '\P{myanmarextendeda}', "");
Expect(1, 43647, '\P{^myanmarextendeda}', "");
Expect(0, 43648, '\p{myanmarextendeda}', "");
Expect(1, 43648, '\p{^myanmarextendeda}', "");
Expect(1, 43648, '\P{myanmarextendeda}', "");
Expect(0, 43648, '\P{^myanmarextendeda}', "");
Expect(1, 43647, '\p{		Myanmar_Extended_A}', "");
Expect(0, 43647, '\p{^		Myanmar_Extended_A}', "");
Expect(0, 43647, '\P{		Myanmar_Extended_A}', "");
Expect(1, 43647, '\P{^		Myanmar_Extended_A}', "");
Expect(0, 43648, '\p{		Myanmar_Extended_A}', "");
Expect(1, 43648, '\p{^		Myanmar_Extended_A}', "");
Expect(1, 43648, '\P{		Myanmar_Extended_A}', "");
Expect(0, 43648, '\P{^		Myanmar_Extended_A}', "");
Error('\p{	_Is_myanmar_Extended_A/a/}');
Error('\P{	_Is_myanmar_Extended_A/a/}');
Expect(1, 43647, '\p{ismyanmarextendeda}', "");
Expect(0, 43647, '\p{^ismyanmarextendeda}', "");
Expect(0, 43647, '\P{ismyanmarextendeda}', "");
Expect(1, 43647, '\P{^ismyanmarextendeda}', "");
Expect(0, 43648, '\p{ismyanmarextendeda}', "");
Expect(1, 43648, '\p{^ismyanmarextendeda}', "");
Expect(1, 43648, '\P{ismyanmarextendeda}', "");
Expect(0, 43648, '\P{^ismyanmarextendeda}', "");
Expect(1, 43647, '\p{_	Is_Myanmar_EXTENDED_A}', "");
Expect(0, 43647, '\p{^_	Is_Myanmar_EXTENDED_A}', "");
Expect(0, 43647, '\P{_	Is_Myanmar_EXTENDED_A}', "");
Expect(1, 43647, '\P{^_	Is_Myanmar_EXTENDED_A}', "");
Expect(0, 43648, '\p{_	Is_Myanmar_EXTENDED_A}', "");
Expect(1, 43648, '\p{^_	Is_Myanmar_EXTENDED_A}', "");
Expect(1, 43648, '\P{_	Is_Myanmar_EXTENDED_A}', "");
Expect(0, 43648, '\P{^_	Is_Myanmar_EXTENDED_A}', "");
Error('\p{:=_ IN_Myanmar_extended_A}');
Error('\P{:=_ IN_Myanmar_extended_A}');
Expect(1, 43647, '\p{inmyanmarextendeda}', "");
Expect(0, 43647, '\p{^inmyanmarextendeda}', "");
Expect(0, 43647, '\P{inmyanmarextendeda}', "");
Expect(1, 43647, '\P{^inmyanmarextendeda}', "");
Expect(0, 43648, '\p{inmyanmarextendeda}', "");
Expect(1, 43648, '\p{^inmyanmarextendeda}', "");
Expect(1, 43648, '\P{inmyanmarextendeda}', "");
Expect(0, 43648, '\P{^inmyanmarextendeda}', "");
Expect(1, 43647, '\p{- In_MYANMAR_extended_A}', "");
Expect(0, 43647, '\p{^- In_MYANMAR_extended_A}', "");
Expect(0, 43647, '\P{- In_MYANMAR_extended_A}', "");
Expect(1, 43647, '\P{^- In_MYANMAR_extended_A}', "");
Expect(0, 43648, '\p{- In_MYANMAR_extended_A}', "");
Expect(1, 43648, '\p{^- In_MYANMAR_extended_A}', "");
Expect(1, 43648, '\P{- In_MYANMAR_extended_A}', "");
Expect(0, 43648, '\P{^- In_MYANMAR_extended_A}', "");
Error('\p{ :=Myanmar_Ext_a}');
Error('\P{ :=Myanmar_Ext_a}');
Expect(1, 43647, '\p{myanmarexta}', "");
Expect(0, 43647, '\p{^myanmarexta}', "");
Expect(0, 43647, '\P{myanmarexta}', "");
Expect(1, 43647, '\P{^myanmarexta}', "");
Expect(0, 43648, '\p{myanmarexta}', "");
Expect(1, 43648, '\p{^myanmarexta}', "");
Expect(1, 43648, '\P{myanmarexta}', "");
Expect(0, 43648, '\P{^myanmarexta}', "");
Expect(1, 43647, '\p{ MYANMAR_ext_A}', "");
Expect(0, 43647, '\p{^ MYANMAR_ext_A}', "");
Expect(0, 43647, '\P{ MYANMAR_ext_A}', "");
Expect(1, 43647, '\P{^ MYANMAR_ext_A}', "");
Expect(0, 43648, '\p{ MYANMAR_ext_A}', "");
Expect(1, 43648, '\p{^ MYANMAR_ext_A}', "");
Expect(1, 43648, '\P{ MYANMAR_ext_A}', "");
Expect(0, 43648, '\P{^ MYANMAR_ext_A}', "");
Error('\p{_:=is_Myanmar_Ext_A}');
Error('\P{_:=is_Myanmar_Ext_A}');
Expect(1, 43647, '\p{ismyanmarexta}', "");
Expect(0, 43647, '\p{^ismyanmarexta}', "");
Expect(0, 43647, '\P{ismyanmarexta}', "");
Expect(1, 43647, '\P{^ismyanmarexta}', "");
Expect(0, 43648, '\p{ismyanmarexta}', "");
Expect(1, 43648, '\p{^ismyanmarexta}', "");
Expect(1, 43648, '\P{ismyanmarexta}', "");
Expect(0, 43648, '\P{^ismyanmarexta}', "");
Expect(1, 43647, '\p{	Is_Myanmar_ext_A}', "");
Expect(0, 43647, '\p{^	Is_Myanmar_ext_A}', "");
Expect(0, 43647, '\P{	Is_Myanmar_ext_A}', "");
Expect(1, 43647, '\P{^	Is_Myanmar_ext_A}', "");
Expect(0, 43648, '\p{	Is_Myanmar_ext_A}', "");
Expect(1, 43648, '\p{^	Is_Myanmar_ext_A}', "");
Expect(1, 43648, '\P{	Is_Myanmar_ext_A}', "");
Expect(0, 43648, '\P{^	Is_Myanmar_ext_A}', "");
Error('\p{_IN_myanmar_ext_a/a/}');
Error('\P{_IN_myanmar_ext_a/a/}');
Expect(1, 43647, '\p{inmyanmarexta}', "");
Expect(0, 43647, '\p{^inmyanmarexta}', "");
Expect(0, 43647, '\P{inmyanmarexta}', "");
Expect(1, 43647, '\P{^inmyanmarexta}', "");
Expect(0, 43648, '\p{inmyanmarexta}', "");
Expect(1, 43648, '\p{^inmyanmarexta}', "");
Expect(1, 43648, '\P{inmyanmarexta}', "");
Expect(0, 43648, '\P{^inmyanmarexta}', "");
Expect(1, 43647, '\p{_	IN_Myanmar_Ext_a}', "");
Expect(0, 43647, '\p{^_	IN_Myanmar_Ext_a}', "");
Expect(0, 43647, '\P{_	IN_Myanmar_Ext_a}', "");
Expect(1, 43647, '\P{^_	IN_Myanmar_Ext_a}', "");
Expect(0, 43648, '\p{_	IN_Myanmar_Ext_a}', "");
Expect(1, 43648, '\p{^_	IN_Myanmar_Ext_a}', "");
Expect(1, 43648, '\P{_	IN_Myanmar_Ext_a}', "");
Expect(0, 43648, '\P{^_	IN_Myanmar_Ext_a}', "");
Error('\p{:=New_Tai_Lue}');
Error('\P{:=New_Tai_Lue}');
Expect(1, 6623, '\p{newtailue}', "");
Expect(0, 6623, '\p{^newtailue}', "");
Expect(0, 6623, '\P{newtailue}', "");
Expect(1, 6623, '\P{^newtailue}', "");
Expect(0, 6624, '\p{newtailue}', "");
Expect(1, 6624, '\p{^newtailue}', "");
Expect(1, 6624, '\P{newtailue}', "");
Expect(0, 6624, '\P{^newtailue}', "");
Expect(1, 6623, '\p{- New_Tai_LUE}', "");
Expect(0, 6623, '\p{^- New_Tai_LUE}', "");
Expect(0, 6623, '\P{- New_Tai_LUE}', "");
Expect(1, 6623, '\P{^- New_Tai_LUE}', "");
Expect(0, 6624, '\p{- New_Tai_LUE}', "");
Expect(1, 6624, '\p{^- New_Tai_LUE}', "");
Expect(1, 6624, '\P{- New_Tai_LUE}', "");
Expect(0, 6624, '\P{^- New_Tai_LUE}', "");
Error('\p{/a/	Is_NEW_Tai_LUE}');
Error('\P{/a/	Is_NEW_Tai_LUE}');
Expect(1, 6623, '\p{isnewtailue}', "");
Expect(0, 6623, '\p{^isnewtailue}', "");
Expect(0, 6623, '\P{isnewtailue}', "");
Expect(1, 6623, '\P{^isnewtailue}', "");
Expect(0, 6624, '\p{isnewtailue}', "");
Expect(1, 6624, '\p{^isnewtailue}', "");
Expect(1, 6624, '\P{isnewtailue}', "");
Expect(0, 6624, '\P{^isnewtailue}', "");
Expect(1, 6623, '\p{	 Is_NEW_tai_Lue}', "");
Expect(0, 6623, '\p{^	 Is_NEW_tai_Lue}', "");
Expect(0, 6623, '\P{	 Is_NEW_tai_Lue}', "");
Expect(1, 6623, '\P{^	 Is_NEW_tai_Lue}', "");
Expect(0, 6624, '\p{	 Is_NEW_tai_Lue}', "");
Expect(1, 6624, '\p{^	 Is_NEW_tai_Lue}', "");
Expect(1, 6624, '\P{	 Is_NEW_tai_Lue}', "");
Expect(0, 6624, '\P{^	 Is_NEW_tai_Lue}', "");
Error('\p{/a/-talu}');
Error('\P{/a/-talu}');
Expect(1, 6623, '\p{talu}', "");
Expect(0, 6623, '\p{^talu}', "");
Expect(0, 6623, '\P{talu}', "");
Expect(1, 6623, '\P{^talu}', "");
Expect(0, 6624, '\p{talu}', "");
Expect(1, 6624, '\p{^talu}', "");
Expect(1, 6624, '\P{talu}', "");
Expect(0, 6624, '\P{^talu}', "");
Expect(1, 6623, '\p{- Talu}', "");
Expect(0, 6623, '\p{^- Talu}', "");
Expect(0, 6623, '\P{- Talu}', "");
Expect(1, 6623, '\P{^- Talu}', "");
Expect(0, 6624, '\p{- Talu}', "");
Expect(1, 6624, '\p{^- Talu}', "");
Expect(1, 6624, '\P{- Talu}', "");
Expect(0, 6624, '\P{^- Talu}', "");
Error('\p{/a/ 	Is_Talu}');
Error('\P{/a/ 	Is_Talu}');
Expect(1, 6623, '\p{istalu}', "");
Expect(0, 6623, '\p{^istalu}', "");
Expect(0, 6623, '\P{istalu}', "");
Expect(1, 6623, '\P{^istalu}', "");
Expect(0, 6624, '\p{istalu}', "");
Expect(1, 6624, '\p{^istalu}', "");
Expect(1, 6624, '\P{istalu}', "");
Expect(0, 6624, '\P{^istalu}', "");
Expect(1, 6623, '\p{_ IS_Talu}', "");
Expect(0, 6623, '\p{^_ IS_Talu}', "");
Expect(0, 6623, '\P{_ IS_Talu}', "");
Expect(1, 6623, '\P{^_ IS_Talu}', "");
Expect(0, 6624, '\p{_ IS_Talu}', "");
Expect(1, 6624, '\p{^_ IS_Talu}', "");
Expect(1, 6624, '\P{_ IS_Talu}', "");
Expect(0, 6624, '\P{^_ IS_Talu}', "");
Error('\p{/a/-	Nko}');
Error('\P{/a/-	Nko}');
Expect(1, 2042, '\p{nko}', "");
Expect(0, 2042, '\p{^nko}', "");
Expect(0, 2042, '\P{nko}', "");
Expect(1, 2042, '\P{^nko}', "");
Expect(0, 2043, '\p{nko}', "");
Expect(1, 2043, '\p{^nko}', "");
Expect(1, 2043, '\P{nko}', "");
Expect(0, 2043, '\P{^nko}', "");
Expect(1, 2042, '\p{_ nko}', "");
Expect(0, 2042, '\p{^_ nko}', "");
Expect(0, 2042, '\P{_ nko}', "");
Expect(1, 2042, '\P{^_ nko}', "");
Expect(0, 2043, '\p{_ nko}', "");
Expect(1, 2043, '\p{^_ nko}', "");
Expect(1, 2043, '\P{_ nko}', "");
Expect(0, 2043, '\P{^_ nko}', "");
Error('\p{ 	Is_Nko:=}');
Error('\P{ 	Is_Nko:=}');
Expect(1, 2042, '\p{isnko}', "");
Expect(0, 2042, '\p{^isnko}', "");
Expect(0, 2042, '\P{isnko}', "");
Expect(1, 2042, '\P{^isnko}', "");
Expect(0, 2043, '\p{isnko}', "");
Expect(1, 2043, '\p{^isnko}', "");
Expect(1, 2043, '\P{isnko}', "");
Expect(0, 2043, '\P{^isnko}', "");
Expect(1, 2042, '\p{		Is_Nko}', "");
Expect(0, 2042, '\p{^		Is_Nko}', "");
Expect(0, 2042, '\P{		Is_Nko}', "");
Expect(1, 2042, '\P{^		Is_Nko}', "");
Expect(0, 2043, '\p{		Is_Nko}', "");
Expect(1, 2043, '\p{^		Is_Nko}', "");
Expect(1, 2043, '\P{		Is_Nko}', "");
Expect(0, 2043, '\P{^		Is_Nko}', "");
Error('\p{/a/ nkoo}');
Error('\P{/a/ nkoo}');
Expect(1, 2042, '\p{nkoo}', "");
Expect(0, 2042, '\p{^nkoo}', "");
Expect(0, 2042, '\P{nkoo}', "");
Expect(1, 2042, '\P{^nkoo}', "");
Expect(0, 2043, '\p{nkoo}', "");
Expect(1, 2043, '\p{^nkoo}', "");
Expect(1, 2043, '\P{nkoo}', "");
Expect(0, 2043, '\P{^nkoo}', "");
Expect(1, 2042, '\p{_nkoo}', "");
Expect(0, 2042, '\p{^_nkoo}', "");
Expect(0, 2042, '\P{_nkoo}', "");
Expect(1, 2042, '\P{^_nkoo}', "");
Expect(0, 2043, '\p{_nkoo}', "");
Expect(1, 2043, '\p{^_nkoo}', "");
Expect(1, 2043, '\P{_nkoo}', "");
Expect(0, 2043, '\P{^_nkoo}', "");
Error('\p{/a/ 	Is_Nkoo}');
Error('\P{/a/ 	Is_Nkoo}');
Expect(1, 2042, '\p{isnkoo}', "");
Expect(0, 2042, '\p{^isnkoo}', "");
Expect(0, 2042, '\P{isnkoo}', "");
Expect(1, 2042, '\P{^isnkoo}', "");
Expect(0, 2043, '\p{isnkoo}', "");
Expect(1, 2043, '\p{^isnkoo}', "");
Expect(1, 2043, '\P{isnkoo}', "");
Expect(0, 2043, '\P{^isnkoo}', "");
Expect(1, 2042, '\p{	Is_NKOO}', "");
Expect(0, 2042, '\p{^	Is_NKOO}', "");
Expect(0, 2042, '\P{	Is_NKOO}', "");
Expect(1, 2042, '\P{^	Is_NKOO}', "");
Expect(0, 2043, '\p{	Is_NKOO}', "");
Expect(1, 2043, '\p{^	Is_NKOO}', "");
Expect(1, 2043, '\P{	Is_NKOO}', "");
Expect(0, 2043, '\P{^	Is_NKOO}', "");
Error('\p{-_NO_block:=}');
Error('\P{-_NO_block:=}');
Expect(1, 918000, '\p{noblock}', "");
Expect(0, 918000, '\p{^noblock}', "");
Expect(0, 918000, '\P{noblock}', "");
Expect(1, 918000, '\P{^noblock}', "");
Expect(0, 983040, '\p{noblock}', "");
Expect(1, 983040, '\p{^noblock}', "");
Expect(1, 983040, '\P{noblock}', "");
Expect(0, 983040, '\P{^noblock}', "");
Expect(1, 918000, '\p{_	No_Block}', "");
Expect(0, 918000, '\p{^_	No_Block}', "");
Expect(0, 918000, '\P{_	No_Block}', "");
Expect(1, 918000, '\P{^_	No_Block}', "");
Expect(0, 983040, '\p{_	No_Block}', "");
Expect(1, 983040, '\p{^_	No_Block}', "");
Expect(1, 983040, '\P{_	No_Block}', "");
Expect(0, 983040, '\P{^_	No_Block}', "");
Error('\p{- Is_NO_Block/a/}');
Error('\P{- Is_NO_Block/a/}');
Expect(1, 918000, '\p{isnoblock}', "");
Expect(0, 918000, '\p{^isnoblock}', "");
Expect(0, 918000, '\P{isnoblock}', "");
Expect(1, 918000, '\P{^isnoblock}', "");
Expect(0, 983040, '\p{isnoblock}', "");
Expect(1, 983040, '\p{^isnoblock}', "");
Expect(1, 983040, '\P{isnoblock}', "");
Expect(0, 983040, '\P{^isnoblock}', "");
Expect(1, 918000, '\p{  Is_NO_Block}', "");
Expect(0, 918000, '\p{^  Is_NO_Block}', "");
Expect(0, 918000, '\P{  Is_NO_Block}', "");
Expect(1, 918000, '\P{^  Is_NO_Block}', "");
Expect(0, 983040, '\p{  Is_NO_Block}', "");
Expect(1, 983040, '\p{^  Is_NO_Block}', "");
Expect(1, 983040, '\P{  Is_NO_Block}', "");
Expect(0, 983040, '\P{^  Is_NO_Block}', "");
Error('\p{_:=In_NO_Block}');
Error('\P{_:=In_NO_Block}');
Expect(1, 918000, '\p{innoblock}', "");
Expect(0, 918000, '\p{^innoblock}', "");
Expect(0, 918000, '\P{innoblock}', "");
Expect(1, 918000, '\P{^innoblock}', "");
Expect(0, 983040, '\p{innoblock}', "");
Expect(1, 983040, '\p{^innoblock}', "");
Expect(1, 983040, '\P{innoblock}', "");
Expect(0, 983040, '\P{^innoblock}', "");
Expect(1, 918000, '\p{	_in_No_block}', "");
Expect(0, 918000, '\p{^	_in_No_block}', "");
Expect(0, 918000, '\P{	_in_No_block}', "");
Expect(1, 918000, '\P{^	_in_No_block}', "");
Expect(0, 983040, '\p{	_in_No_block}', "");
Expect(1, 983040, '\p{^	_in_No_block}', "");
Expect(1, 983040, '\P{	_in_No_block}', "");
Expect(0, 983040, '\P{^	_in_No_block}', "");
Error('\p{/a/	NB}');
Error('\P{/a/	NB}');
Expect(1, 918000, '\p{nb}', "");
Expect(0, 918000, '\p{^nb}', "");
Expect(0, 918000, '\P{nb}', "");
Expect(1, 918000, '\P{^nb}', "");
Expect(0, 983040, '\p{nb}', "");
Expect(1, 983040, '\p{^nb}', "");
Expect(1, 983040, '\P{nb}', "");
Expect(0, 983040, '\P{^nb}', "");
Expect(1, 918000, '\p{ NB}', "");
Expect(0, 918000, '\p{^ NB}', "");
Expect(0, 918000, '\P{ NB}', "");
Expect(1, 918000, '\P{^ NB}', "");
Expect(0, 983040, '\p{ NB}', "");
Expect(1, 983040, '\p{^ NB}', "");
Expect(1, 983040, '\P{ NB}', "");
Expect(0, 983040, '\P{^ NB}', "");
Error('\p{/a/is_NB}');
Error('\P{/a/is_NB}');
Expect(1, 918000, '\p{isnb}', "");
Expect(0, 918000, '\p{^isnb}', "");
Expect(0, 918000, '\P{isnb}', "");
Expect(1, 918000, '\P{^isnb}', "");
Expect(0, 983040, '\p{isnb}', "");
Expect(1, 983040, '\p{^isnb}', "");
Expect(1, 983040, '\P{isnb}', "");
Expect(0, 983040, '\P{^isnb}', "");
Expect(1, 918000, '\p{	_Is_NB}', "");
Expect(0, 918000, '\p{^	_Is_NB}', "");
Expect(0, 918000, '\P{	_Is_NB}', "");
Expect(1, 918000, '\P{^	_Is_NB}', "");
Expect(0, 983040, '\p{	_Is_NB}', "");
Expect(1, 983040, '\p{^	_Is_NB}', "");
Expect(1, 983040, '\P{	_Is_NB}', "");
Expect(0, 983040, '\P{^	_Is_NB}', "");
Error('\p{-/a/In_NB}');
Error('\P{-/a/In_NB}');
Expect(1, 918000, '\p{innb}', "");
Expect(0, 918000, '\p{^innb}', "");
Expect(0, 918000, '\P{innb}', "");
Expect(1, 918000, '\P{^innb}', "");
Expect(0, 983040, '\p{innb}', "");
Expect(1, 983040, '\p{^innb}', "");
Expect(1, 983040, '\P{innb}', "");
Expect(0, 983040, '\P{^innb}', "");
Expect(1, 918000, '\p{ _in_NB}', "");
Expect(0, 918000, '\p{^ _in_NB}', "");
Expect(0, 918000, '\P{ _in_NB}', "");
Expect(1, 918000, '\P{^ _in_NB}', "");
Expect(0, 983040, '\p{ _in_NB}', "");
Expect(1, 983040, '\p{^ _in_NB}', "");
Expect(1, 983040, '\P{ _in_NB}', "");
Expect(0, 983040, '\P{^ _in_NB}', "");
Error('\p{-	noncharacter_CODE_Point:=}');
Error('\P{-	noncharacter_CODE_Point:=}');
Expect(1, 1114111, '\p{noncharactercodepoint}', "");
Expect(0, 1114111, '\p{^noncharactercodepoint}', "");
Expect(0, 1114111, '\P{noncharactercodepoint}', "");
Expect(1, 1114111, '\P{^noncharactercodepoint}', "");
Expect(0, 1114109, '\p{noncharactercodepoint}', "");
Expect(1, 1114109, '\p{^noncharactercodepoint}', "");
Expect(1, 1114109, '\P{noncharactercodepoint}', "");
Expect(0, 1114109, '\P{^noncharactercodepoint}', "");
Expect(1, 1114111, '\p{ _Noncharacter_Code_point}', "");
Expect(0, 1114111, '\p{^ _Noncharacter_Code_point}', "");
Expect(0, 1114111, '\P{ _Noncharacter_Code_point}', "");
Expect(1, 1114111, '\P{^ _Noncharacter_Code_point}', "");
Expect(0, 1114109, '\p{ _Noncharacter_Code_point}', "");
Expect(1, 1114109, '\p{^ _Noncharacter_Code_point}', "");
Expect(1, 1114109, '\P{ _Noncharacter_Code_point}', "");
Expect(0, 1114109, '\P{^ _Noncharacter_Code_point}', "");
Error('\p{	Is_Noncharacter_CODE_Point/a/}');
Error('\P{	Is_Noncharacter_CODE_Point/a/}');
Expect(1, 1114111, '\p{isnoncharactercodepoint}', "");
Expect(0, 1114111, '\p{^isnoncharactercodepoint}', "");
Expect(0, 1114111, '\P{isnoncharactercodepoint}', "");
Expect(1, 1114111, '\P{^isnoncharactercodepoint}', "");
Expect(0, 1114109, '\p{isnoncharactercodepoint}', "");
Expect(1, 1114109, '\p{^isnoncharactercodepoint}', "");
Expect(1, 1114109, '\P{isnoncharactercodepoint}', "");
Expect(0, 1114109, '\P{^isnoncharactercodepoint}', "");
Expect(1, 1114111, '\p{	Is_noncharacter_code_point}', "");
Expect(0, 1114111, '\p{^	Is_noncharacter_code_point}', "");
Expect(0, 1114111, '\P{	Is_noncharacter_code_point}', "");
Expect(1, 1114111, '\P{^	Is_noncharacter_code_point}', "");
Expect(0, 1114109, '\p{	Is_noncharacter_code_point}', "");
Expect(1, 1114109, '\p{^	Is_noncharacter_code_point}', "");
Expect(1, 1114109, '\P{	Is_noncharacter_code_point}', "");
Expect(0, 1114109, '\P{^	Is_noncharacter_code_point}', "");
Error('\p{_:=NChar}');
Error('\P{_:=NChar}');
Expect(1, 1114111, '\p{nchar}', "");
Expect(0, 1114111, '\p{^nchar}', "");
Expect(0, 1114111, '\P{nchar}', "");
Expect(1, 1114111, '\P{^nchar}', "");
Expect(0, 1114109, '\p{nchar}', "");
Expect(1, 1114109, '\p{^nchar}', "");
Expect(1, 1114109, '\P{nchar}', "");
Expect(0, 1114109, '\P{^nchar}', "");
Expect(1, 1114111, '\p{--NCHAR}', "");
Expect(0, 1114111, '\p{^--NCHAR}', "");
Expect(0, 1114111, '\P{--NCHAR}', "");
Expect(1, 1114111, '\P{^--NCHAR}', "");
Expect(0, 1114109, '\p{--NCHAR}', "");
Expect(1, 1114109, '\p{^--NCHAR}', "");
Expect(1, 1114109, '\P{--NCHAR}', "");
Expect(0, 1114109, '\P{^--NCHAR}', "");
Error('\p{-	is_NCHAR:=}');
Error('\P{-	is_NCHAR:=}');
Expect(1, 1114111, '\p{isnchar}', "");
Expect(0, 1114111, '\p{^isnchar}', "");
Expect(0, 1114111, '\P{isnchar}', "");
Expect(1, 1114111, '\P{^isnchar}', "");
Expect(0, 1114109, '\p{isnchar}', "");
Expect(1, 1114109, '\p{^isnchar}', "");
Expect(1, 1114109, '\P{isnchar}', "");
Expect(0, 1114109, '\P{^isnchar}', "");
Expect(1, 1114111, '\p{-	Is_NChar}', "");
Expect(0, 1114111, '\p{^-	Is_NChar}', "");
Expect(0, 1114111, '\P{-	Is_NChar}', "");
Expect(1, 1114111, '\P{^-	Is_NChar}', "");
Expect(0, 1114109, '\p{-	Is_NChar}', "");
Expect(1, 1114109, '\p{^-	Is_NChar}', "");
Expect(1, 1114109, '\P{-	Is_NChar}', "");
Expect(0, 1114109, '\P{^-	Is_NChar}', "");
Error('\p{-:=Nonspacing_mark}');
Error('\P{-:=Nonspacing_mark}');
Expect(1, 917999, '\p{nonspacingmark}', "");
Expect(0, 917999, '\p{^nonspacingmark}', "");
Expect(0, 917999, '\P{nonspacingmark}', "");
Expect(1, 917999, '\P{^nonspacingmark}', "");
Expect(0, 918000, '\p{nonspacingmark}', "");
Expect(1, 918000, '\p{^nonspacingmark}', "");
Expect(1, 918000, '\P{nonspacingmark}', "");
Expect(0, 918000, '\P{^nonspacingmark}', "");
Expect(1, 917999, '\p{	Nonspacing_mark}', "");
Expect(0, 917999, '\p{^	Nonspacing_mark}', "");
Expect(0, 917999, '\P{	Nonspacing_mark}', "");
Expect(1, 917999, '\P{^	Nonspacing_mark}', "");
Expect(0, 918000, '\p{	Nonspacing_mark}', "");
Expect(1, 918000, '\p{^	Nonspacing_mark}', "");
Expect(1, 918000, '\P{	Nonspacing_mark}', "");
Expect(0, 918000, '\P{^	Nonspacing_mark}', "");
Error('\p{:=	Is_Nonspacing_mark}');
Error('\P{:=	Is_Nonspacing_mark}');
Expect(1, 917999, '\p{isnonspacingmark}', "");
Expect(0, 917999, '\p{^isnonspacingmark}', "");
Expect(0, 917999, '\P{isnonspacingmark}', "");
Expect(1, 917999, '\P{^isnonspacingmark}', "");
Expect(0, 918000, '\p{isnonspacingmark}', "");
Expect(1, 918000, '\p{^isnonspacingmark}', "");
Expect(1, 918000, '\P{isnonspacingmark}', "");
Expect(0, 918000, '\P{^isnonspacingmark}', "");
Expect(1, 917999, '\p{ is_Nonspacing_Mark}', "");
Expect(0, 917999, '\p{^ is_Nonspacing_Mark}', "");
Expect(0, 917999, '\P{ is_Nonspacing_Mark}', "");
Expect(1, 917999, '\P{^ is_Nonspacing_Mark}', "");
Expect(0, 918000, '\p{ is_Nonspacing_Mark}', "");
Expect(1, 918000, '\p{^ is_Nonspacing_Mark}', "");
Expect(1, 918000, '\P{ is_Nonspacing_Mark}', "");
Expect(0, 918000, '\P{^ is_Nonspacing_Mark}', "");
Error('\p{ /a/mn}');
Error('\P{ /a/mn}');
Expect(1, 917999, '\p{mn}', "");
Expect(0, 917999, '\p{^mn}', "");
Expect(0, 917999, '\P{mn}', "");
Expect(1, 917999, '\P{^mn}', "");
Expect(0, 918000, '\p{mn}', "");
Expect(1, 918000, '\p{^mn}', "");
Expect(1, 918000, '\P{mn}', "");
Expect(0, 918000, '\P{^mn}', "");
Expect(1, 917999, '\p{ 	Mn}', "");
Expect(0, 917999, '\p{^ 	Mn}', "");
Expect(0, 917999, '\P{ 	Mn}', "");
Expect(1, 917999, '\P{^ 	Mn}', "");
Expect(0, 918000, '\p{ 	Mn}', "");
Expect(1, 918000, '\p{^ 	Mn}', "");
Expect(1, 918000, '\P{ 	Mn}', "");
Expect(0, 918000, '\P{^ 	Mn}', "");
Error('\p{	 Is_Mn/a/}');
Error('\P{	 Is_Mn/a/}');
Expect(1, 917999, '\p{ismn}', "");
Expect(0, 917999, '\p{^ismn}', "");
Expect(0, 917999, '\P{ismn}', "");
Expect(1, 917999, '\P{^ismn}', "");
Expect(0, 918000, '\p{ismn}', "");
Expect(1, 918000, '\p{^ismn}', "");
Expect(1, 918000, '\P{ismn}', "");
Expect(0, 918000, '\P{^ismn}', "");
Expect(1, 917999, '\p{ Is_Mn}', "");
Expect(0, 917999, '\p{^ Is_Mn}', "");
Expect(0, 917999, '\P{ Is_Mn}', "");
Expect(1, 917999, '\P{^ Is_Mn}', "");
Expect(0, 918000, '\p{ Is_Mn}', "");
Expect(1, 918000, '\p{^ Is_Mn}', "");
Expect(1, 918000, '\P{ Is_Mn}', "");
Expect(0, 918000, '\P{^ Is_Mn}', "");
Error('\p{_:=number}');
Error('\P{_:=number}');
Expect(1, 127242, '\p{number}', "");
Expect(0, 127242, '\p{^number}', "");
Expect(0, 127242, '\P{number}', "");
Expect(1, 127242, '\P{^number}', "");
Expect(0, 127243, '\p{number}', "");
Expect(1, 127243, '\p{^number}', "");
Expect(1, 127243, '\P{number}', "");
Expect(0, 127243, '\P{^number}', "");
Expect(1, 127242, '\p{	number}', "");
Expect(0, 127242, '\p{^	number}', "");
Expect(0, 127242, '\P{	number}', "");
Expect(1, 127242, '\P{^	number}', "");
Expect(0, 127243, '\p{	number}', "");
Expect(1, 127243, '\p{^	number}', "");
Expect(1, 127243, '\P{	number}', "");
Expect(0, 127243, '\P{^	number}', "");
Error('\p{:= -Is_NUMBER}');
Error('\P{:= -Is_NUMBER}');
Expect(1, 127242, '\p{isnumber}', "");
Expect(0, 127242, '\p{^isnumber}', "");
Expect(0, 127242, '\P{isnumber}', "");
Expect(1, 127242, '\P{^isnumber}', "");
Expect(0, 127243, '\p{isnumber}', "");
Expect(1, 127243, '\p{^isnumber}', "");
Expect(1, 127243, '\P{isnumber}', "");
Expect(0, 127243, '\P{^isnumber}', "");
Expect(1, 127242, '\p{__is_NUMBER}', "");
Expect(0, 127242, '\p{^__is_NUMBER}', "");
Expect(0, 127242, '\P{__is_NUMBER}', "");
Expect(1, 127242, '\P{^__is_NUMBER}', "");
Expect(0, 127243, '\p{__is_NUMBER}', "");
Expect(1, 127243, '\p{^__is_NUMBER}', "");
Expect(1, 127243, '\P{__is_NUMBER}', "");
Expect(0, 127243, '\P{^__is_NUMBER}', "");
Error('\p{/a/ n}');
Error('\P{/a/ n}');
Expect(1, 127242, '\p{n}', "");
Expect(0, 127242, '\p{^n}', "");
Expect(0, 127242, '\P{n}', "");
Expect(1, 127242, '\P{^n}', "");
Expect(0, 127243, '\p{n}', "");
Expect(1, 127243, '\p{^n}', "");
Expect(1, 127243, '\P{n}', "");
Expect(0, 127243, '\P{^n}', "");
Expect(1, 127242, '\p{	 n}', "");
Expect(0, 127242, '\p{^	 n}', "");
Expect(0, 127242, '\P{	 n}', "");
Expect(1, 127242, '\P{^	 n}', "");
Expect(0, 127243, '\p{	 n}', "");
Expect(1, 127243, '\p{^	 n}', "");
Expect(1, 127243, '\P{	 n}', "");
Expect(0, 127243, '\P{^	 n}', "");
Error('\p{	 IS_N/a/}');
Error('\P{	 IS_N/a/}');
Expect(1, 127242, '\p{isn}', "");
Expect(0, 127242, '\p{^isn}', "");
Expect(0, 127242, '\P{isn}', "");
Expect(1, 127242, '\P{^isn}', "");
Expect(0, 127243, '\p{isn}', "");
Expect(1, 127243, '\p{^isn}', "");
Expect(1, 127243, '\P{isn}', "");
Expect(0, 127243, '\P{^isn}', "");
Expect(1, 127242, '\p{	 Is_N}', "");
Expect(0, 127242, '\p{^	 Is_N}', "");
Expect(0, 127242, '\P{	 Is_N}', "");
Expect(1, 127242, '\P{^	 Is_N}', "");
Expect(0, 127243, '\p{	 Is_N}', "");
Expect(1, 127243, '\p{^	 Is_N}', "");
Expect(1, 127243, '\P{	 Is_N}', "");
Expect(0, 127243, '\P{^	 Is_N}', "");
Error('\p{-_number_Forms/a/}');
Error('\P{-_number_Forms/a/}');
Expect(1, 8591, '\p{numberforms}', "");
Expect(0, 8591, '\p{^numberforms}', "");
Expect(0, 8591, '\P{numberforms}', "");
Expect(1, 8591, '\P{^numberforms}', "");
Expect(0, 8592, '\p{numberforms}', "");
Expect(1, 8592, '\p{^numberforms}', "");
Expect(1, 8592, '\P{numberforms}', "");
Expect(0, 8592, '\P{^numberforms}', "");
Expect(1, 8591, '\p{-Number_Forms}', "");
Expect(0, 8591, '\p{^-Number_Forms}', "");
Expect(0, 8591, '\P{-Number_Forms}', "");
Expect(1, 8591, '\P{^-Number_Forms}', "");
Expect(0, 8592, '\p{-Number_Forms}', "");
Expect(1, 8592, '\p{^-Number_Forms}', "");
Expect(1, 8592, '\P{-Number_Forms}', "");
Expect(0, 8592, '\P{^-Number_Forms}', "");
Error('\p{-IS_NUMBER_FORMS/a/}');
Error('\P{-IS_NUMBER_FORMS/a/}');
Expect(1, 8591, '\p{isnumberforms}', "");
Expect(0, 8591, '\p{^isnumberforms}', "");
Expect(0, 8591, '\P{isnumberforms}', "");
Expect(1, 8591, '\P{^isnumberforms}', "");
Expect(0, 8592, '\p{isnumberforms}', "");
Expect(1, 8592, '\p{^isnumberforms}', "");
Expect(1, 8592, '\P{isnumberforms}', "");
Expect(0, 8592, '\P{^isnumberforms}', "");
Expect(1, 8591, '\p{Is_Number_FORMS}', "");
Expect(0, 8591, '\p{^Is_Number_FORMS}', "");
Expect(0, 8591, '\P{Is_Number_FORMS}', "");
Expect(1, 8591, '\P{^Is_Number_FORMS}', "");
Expect(0, 8592, '\p{Is_Number_FORMS}', "");
Expect(1, 8592, '\p{^Is_Number_FORMS}', "");
Expect(1, 8592, '\P{Is_Number_FORMS}', "");
Expect(0, 8592, '\P{^Is_Number_FORMS}', "");
Error('\p{:=__in_number_Forms}');
Error('\P{:=__in_number_Forms}');
Expect(1, 8591, '\p{innumberforms}', "");
Expect(0, 8591, '\p{^innumberforms}', "");
Expect(0, 8591, '\P{innumberforms}', "");
Expect(1, 8591, '\P{^innumberforms}', "");
Expect(0, 8592, '\p{innumberforms}', "");
Expect(1, 8592, '\p{^innumberforms}', "");
Expect(1, 8592, '\P{innumberforms}', "");
Expect(0, 8592, '\P{^innumberforms}', "");
Expect(1, 8591, '\p{-in_Number_forms}', "");
Expect(0, 8591, '\p{^-in_Number_forms}', "");
Expect(0, 8591, '\P{-in_Number_forms}', "");
Expect(1, 8591, '\P{^-in_Number_forms}', "");
Expect(0, 8592, '\p{-in_Number_forms}', "");
Expect(1, 8592, '\p{^-in_Number_forms}', "");
Expect(1, 8592, '\P{-in_Number_forms}', "");
Expect(0, 8592, '\P{^-in_Number_forms}', "");
Error('\p{:= _ogham}');
Error('\P{:= _ogham}');
Expect(1, 5788, '\p{ogham}', "");
Expect(0, 5788, '\p{^ogham}', "");
Expect(0, 5788, '\P{ogham}', "");
Expect(1, 5788, '\P{^ogham}', "");
Expect(0, 5789, '\p{ogham}', "");
Expect(1, 5789, '\p{^ogham}', "");
Expect(1, 5789, '\P{ogham}', "");
Expect(0, 5789, '\P{^ogham}', "");
Expect(1, 5788, '\p{_-OGHAM}', "");
Expect(0, 5788, '\p{^_-OGHAM}', "");
Expect(0, 5788, '\P{_-OGHAM}', "");
Expect(1, 5788, '\P{^_-OGHAM}', "");
Expect(0, 5789, '\p{_-OGHAM}', "");
Expect(1, 5789, '\p{^_-OGHAM}', "");
Expect(1, 5789, '\P{_-OGHAM}', "");
Expect(0, 5789, '\P{^_-OGHAM}', "");
Error('\p{-/a/IS_OGHAM}');
Error('\P{-/a/IS_OGHAM}');
Expect(1, 5788, '\p{isogham}', "");
Expect(0, 5788, '\p{^isogham}', "");
Expect(0, 5788, '\P{isogham}', "");
Expect(1, 5788, '\P{^isogham}', "");
Expect(0, 5789, '\p{isogham}', "");
Expect(1, 5789, '\p{^isogham}', "");
Expect(1, 5789, '\P{isogham}', "");
Expect(0, 5789, '\P{^isogham}', "");
Expect(1, 5788, '\p{_IS_Ogham}', "");
Expect(0, 5788, '\p{^_IS_Ogham}', "");
Expect(0, 5788, '\P{_IS_Ogham}', "");
Expect(1, 5788, '\P{^_IS_Ogham}', "");
Expect(0, 5789, '\p{_IS_Ogham}', "");
Expect(1, 5789, '\p{^_IS_Ogham}', "");
Expect(1, 5789, '\P{_IS_Ogham}', "");
Expect(0, 5789, '\P{^_IS_Ogham}', "");
Error('\p{/a/__Ogam}');
Error('\P{/a/__Ogam}');
Expect(1, 5788, '\p{ogam}', "");
Expect(0, 5788, '\p{^ogam}', "");
Expect(0, 5788, '\P{ogam}', "");
Expect(1, 5788, '\P{^ogam}', "");
Expect(0, 5789, '\p{ogam}', "");
Expect(1, 5789, '\p{^ogam}', "");
Expect(1, 5789, '\P{ogam}', "");
Expect(0, 5789, '\P{^ogam}', "");
Expect(1, 5788, '\p{_ Ogam}', "");
Expect(0, 5788, '\p{^_ Ogam}', "");
Expect(0, 5788, '\P{_ Ogam}', "");
Expect(1, 5788, '\P{^_ Ogam}', "");
Expect(0, 5789, '\p{_ Ogam}', "");
Expect(1, 5789, '\p{^_ Ogam}', "");
Expect(1, 5789, '\P{_ Ogam}', "");
Expect(0, 5789, '\P{^_ Ogam}', "");
Error('\p{/a/is_Ogam}');
Error('\P{/a/is_Ogam}');
Expect(1, 5788, '\p{isogam}', "");
Expect(0, 5788, '\p{^isogam}', "");
Expect(0, 5788, '\P{isogam}', "");
Expect(1, 5788, '\P{^isogam}', "");
Expect(0, 5789, '\p{isogam}', "");
Expect(1, 5789, '\p{^isogam}', "");
Expect(1, 5789, '\P{isogam}', "");
Expect(0, 5789, '\P{^isogam}', "");
Expect(1, 5788, '\p{_Is_Ogam}', "");
Expect(0, 5788, '\p{^_Is_Ogam}', "");
Expect(0, 5788, '\P{_Is_Ogam}', "");
Expect(1, 5788, '\P{^_Is_Ogam}', "");
Expect(0, 5789, '\p{_Is_Ogam}', "");
Expect(1, 5789, '\p{^_Is_Ogam}', "");
Expect(1, 5789, '\P{_Is_Ogam}', "");
Expect(0, 5789, '\P{^_Is_Ogam}', "");
Error('\p{--Ol_Chiki/a/}');
Error('\P{--Ol_Chiki/a/}');
Expect(1, 7295, '\p{olchiki}', "");
Expect(0, 7295, '\p{^olchiki}', "");
Expect(0, 7295, '\P{olchiki}', "");
Expect(1, 7295, '\P{^olchiki}', "");
Expect(0, 7296, '\p{olchiki}', "");
Expect(1, 7296, '\p{^olchiki}', "");
Expect(1, 7296, '\P{olchiki}', "");
Expect(0, 7296, '\P{^olchiki}', "");
Expect(1, 7295, '\p{_OL_Chiki}', "");
Expect(0, 7295, '\p{^_OL_Chiki}', "");
Expect(0, 7295, '\P{_OL_Chiki}', "");
Expect(1, 7295, '\P{^_OL_Chiki}', "");
Expect(0, 7296, '\p{_OL_Chiki}', "");
Expect(1, 7296, '\p{^_OL_Chiki}', "");
Expect(1, 7296, '\P{_OL_Chiki}', "");
Expect(0, 7296, '\P{^_OL_Chiki}', "");
Error('\p{  Is_Ol_Chiki/a/}');
Error('\P{  Is_Ol_Chiki/a/}');
Expect(1, 7295, '\p{isolchiki}', "");
Expect(0, 7295, '\p{^isolchiki}', "");
Expect(0, 7295, '\P{isolchiki}', "");
Expect(1, 7295, '\P{^isolchiki}', "");
Expect(0, 7296, '\p{isolchiki}', "");
Expect(1, 7296, '\p{^isolchiki}', "");
Expect(1, 7296, '\P{isolchiki}', "");
Expect(0, 7296, '\P{^isolchiki}', "");
Expect(1, 7295, '\p{-	is_ol_CHIKI}', "");
Expect(0, 7295, '\p{^-	is_ol_CHIKI}', "");
Expect(0, 7295, '\P{-	is_ol_CHIKI}', "");
Expect(1, 7295, '\P{^-	is_ol_CHIKI}', "");
Expect(0, 7296, '\p{-	is_ol_CHIKI}', "");
Expect(1, 7296, '\p{^-	is_ol_CHIKI}', "");
Expect(1, 7296, '\P{-	is_ol_CHIKI}', "");
Expect(0, 7296, '\P{^-	is_ol_CHIKI}', "");
Error('\p{ /a/olck}');
Error('\P{ /a/olck}');
Expect(1, 7295, '\p{olck}', "");
Expect(0, 7295, '\p{^olck}', "");
Expect(0, 7295, '\P{olck}', "");
Expect(1, 7295, '\P{^olck}', "");
Expect(0, 7296, '\p{olck}', "");
Expect(1, 7296, '\p{^olck}', "");
Expect(1, 7296, '\P{olck}', "");
Expect(0, 7296, '\P{^olck}', "");
Expect(1, 7295, '\p{_Olck}', "");
Expect(0, 7295, '\p{^_Olck}', "");
Expect(0, 7295, '\P{_Olck}', "");
Expect(1, 7295, '\P{^_Olck}', "");
Expect(0, 7296, '\p{_Olck}', "");
Expect(1, 7296, '\p{^_Olck}', "");
Expect(1, 7296, '\P{_Olck}', "");
Expect(0, 7296, '\P{^_Olck}', "");
Error('\p{/a/	-Is_Olck}');
Error('\P{/a/	-Is_Olck}');
Expect(1, 7295, '\p{isolck}', "");
Expect(0, 7295, '\p{^isolck}', "");
Expect(0, 7295, '\P{isolck}', "");
Expect(1, 7295, '\P{^isolck}', "");
Expect(0, 7296, '\p{isolck}', "");
Expect(1, 7296, '\p{^isolck}', "");
Expect(1, 7296, '\P{isolck}', "");
Expect(0, 7296, '\P{^isolck}', "");
Expect(1, 7295, '\p{- Is_OLCK}', "");
Expect(0, 7295, '\p{^- Is_OLCK}', "");
Expect(0, 7295, '\P{- Is_OLCK}', "");
Expect(1, 7295, '\P{^- Is_OLCK}', "");
Expect(0, 7296, '\p{- Is_OLCK}', "");
Expect(1, 7296, '\p{^- Is_OLCK}', "");
Expect(1, 7296, '\P{- Is_OLCK}', "");
Expect(0, 7296, '\P{^- Is_OLCK}', "");
Error('\p{/a/ _OLD_italic}');
Error('\P{/a/ _OLD_italic}');
Expect(1, 66339, '\p{olditalic}', "");
Expect(0, 66339, '\p{^olditalic}', "");
Expect(0, 66339, '\P{olditalic}', "");
Expect(1, 66339, '\P{^olditalic}', "");
Expect(0, 66340, '\p{olditalic}', "");
Expect(1, 66340, '\p{^olditalic}', "");
Expect(1, 66340, '\P{olditalic}', "");
Expect(0, 66340, '\P{^olditalic}', "");
Expect(1, 66339, '\p{_Old_Italic}', "");
Expect(0, 66339, '\p{^_Old_Italic}', "");
Expect(0, 66339, '\P{_Old_Italic}', "");
Expect(1, 66339, '\P{^_Old_Italic}', "");
Expect(0, 66340, '\p{_Old_Italic}', "");
Expect(1, 66340, '\p{^_Old_Italic}', "");
Expect(1, 66340, '\P{_Old_Italic}', "");
Expect(0, 66340, '\P{^_Old_Italic}', "");
Error('\p{ Is_Old_italic:=}');
Error('\P{ Is_Old_italic:=}');
Expect(1, 66339, '\p{isolditalic}', "");
Expect(0, 66339, '\p{^isolditalic}', "");
Expect(0, 66339, '\P{isolditalic}', "");
Expect(1, 66339, '\P{^isolditalic}', "");
Expect(0, 66340, '\p{isolditalic}', "");
Expect(1, 66340, '\p{^isolditalic}', "");
Expect(1, 66340, '\P{isolditalic}', "");
Expect(0, 66340, '\P{^isolditalic}', "");
Expect(1, 66339, '\p{_ Is_Old_Italic}', "");
Expect(0, 66339, '\p{^_ Is_Old_Italic}', "");
Expect(0, 66339, '\P{_ Is_Old_Italic}', "");
Expect(1, 66339, '\P{^_ Is_Old_Italic}', "");
Expect(0, 66340, '\p{_ Is_Old_Italic}', "");
Expect(1, 66340, '\p{^_ Is_Old_Italic}', "");
Expect(1, 66340, '\P{_ Is_Old_Italic}', "");
Expect(0, 66340, '\P{^_ Is_Old_Italic}', "");
Error('\p{	ital/a/}');
Error('\P{	ital/a/}');
Expect(1, 66339, '\p{ital}', "");
Expect(0, 66339, '\p{^ital}', "");
Expect(0, 66339, '\P{ital}', "");
Expect(1, 66339, '\P{^ital}', "");
Expect(0, 66340, '\p{ital}', "");
Expect(1, 66340, '\p{^ital}', "");
Expect(1, 66340, '\P{ital}', "");
Expect(0, 66340, '\P{^ital}', "");
Expect(1, 66339, '\p{_ ital}', "");
Expect(0, 66339, '\p{^_ ital}', "");
Expect(0, 66339, '\P{_ ital}', "");
Expect(1, 66339, '\P{^_ ital}', "");
Expect(0, 66340, '\p{_ ital}', "");
Expect(1, 66340, '\p{^_ ital}', "");
Expect(1, 66340, '\P{_ ital}', "");
Expect(0, 66340, '\P{^_ ital}', "");
Error('\p{/a/	IS_Ital}');
Error('\P{/a/	IS_Ital}');
Expect(1, 66339, '\p{isital}', "");
Expect(0, 66339, '\p{^isital}', "");
Expect(0, 66339, '\P{isital}', "");
Expect(1, 66339, '\P{^isital}', "");
Expect(0, 66340, '\p{isital}', "");
Expect(1, 66340, '\p{^isital}', "");
Expect(1, 66340, '\P{isital}', "");
Expect(0, 66340, '\P{^isital}', "");
Expect(1, 66339, '\p{-	IS_Ital}', "");
Expect(0, 66339, '\p{^-	IS_Ital}', "");
Expect(0, 66339, '\P{-	IS_Ital}', "");
Expect(1, 66339, '\P{^-	IS_Ital}', "");
Expect(0, 66340, '\p{-	IS_Ital}', "");
Expect(1, 66340, '\p{^-	IS_Ital}', "");
Expect(1, 66340, '\P{-	IS_Ital}', "");
Expect(0, 66340, '\P{^-	IS_Ital}', "");
Error('\p{/a/Old_Persian}');
Error('\P{/a/Old_Persian}');
Expect(1, 66517, '\p{oldpersian}', "");
Expect(0, 66517, '\p{^oldpersian}', "");
Expect(0, 66517, '\P{oldpersian}', "");
Expect(1, 66517, '\P{^oldpersian}', "");
Expect(0, 66518, '\p{oldpersian}', "");
Expect(1, 66518, '\p{^oldpersian}', "");
Expect(1, 66518, '\P{oldpersian}', "");
Expect(0, 66518, '\P{^oldpersian}', "");
Expect(1, 66517, '\p{_-OLD_Persian}', "");
Expect(0, 66517, '\p{^_-OLD_Persian}', "");
Expect(0, 66517, '\P{_-OLD_Persian}', "");
Expect(1, 66517, '\P{^_-OLD_Persian}', "");
Expect(0, 66518, '\p{_-OLD_Persian}', "");
Expect(1, 66518, '\p{^_-OLD_Persian}', "");
Expect(1, 66518, '\P{_-OLD_Persian}', "");
Expect(0, 66518, '\P{^_-OLD_Persian}', "");
Error('\p{/a/--Is_Old_persian}');
Error('\P{/a/--Is_Old_persian}');
Expect(1, 66517, '\p{isoldpersian}', "");
Expect(0, 66517, '\p{^isoldpersian}', "");
Expect(0, 66517, '\P{isoldpersian}', "");
Expect(1, 66517, '\P{^isoldpersian}', "");
Expect(0, 66518, '\p{isoldpersian}', "");
Expect(1, 66518, '\p{^isoldpersian}', "");
Expect(1, 66518, '\P{isoldpersian}', "");
Expect(0, 66518, '\P{^isoldpersian}', "");
Expect(1, 66517, '\p{ is_OLD_Persian}', "");
Expect(0, 66517, '\p{^ is_OLD_Persian}', "");
Expect(0, 66517, '\P{ is_OLD_Persian}', "");
Expect(1, 66517, '\P{^ is_OLD_Persian}', "");
Expect(0, 66518, '\p{ is_OLD_Persian}', "");
Expect(1, 66518, '\p{^ is_OLD_Persian}', "");
Expect(1, 66518, '\P{ is_OLD_Persian}', "");
Expect(0, 66518, '\P{^ is_OLD_Persian}', "");
Error('\p{-xpeo:=}');
Error('\P{-xpeo:=}');
Expect(1, 66517, '\p{xpeo}', "");
Expect(0, 66517, '\p{^xpeo}', "");
Expect(0, 66517, '\P{xpeo}', "");
Expect(1, 66517, '\P{^xpeo}', "");
Expect(0, 66518, '\p{xpeo}', "");
Expect(1, 66518, '\p{^xpeo}', "");
Expect(1, 66518, '\P{xpeo}', "");
Expect(0, 66518, '\P{^xpeo}', "");
Expect(1, 66517, '\p{  Xpeo}', "");
Expect(0, 66517, '\p{^  Xpeo}', "");
Expect(0, 66517, '\P{  Xpeo}', "");
Expect(1, 66517, '\P{^  Xpeo}', "");
Expect(0, 66518, '\p{  Xpeo}', "");
Expect(1, 66518, '\p{^  Xpeo}', "");
Expect(1, 66518, '\P{  Xpeo}', "");
Expect(0, 66518, '\P{^  Xpeo}', "");
Error('\p{/a/_	IS_xpeo}');
Error('\P{/a/_	IS_xpeo}');
Expect(1, 66517, '\p{isxpeo}', "");
Expect(0, 66517, '\p{^isxpeo}', "");
Expect(0, 66517, '\P{isxpeo}', "");
Expect(1, 66517, '\P{^isxpeo}', "");
Expect(0, 66518, '\p{isxpeo}', "");
Expect(1, 66518, '\p{^isxpeo}', "");
Expect(1, 66518, '\P{isxpeo}', "");
Expect(0, 66518, '\P{^isxpeo}', "");
Expect(1, 66517, '\p{ 	IS_Xpeo}', "");
Expect(0, 66517, '\p{^ 	IS_Xpeo}', "");
Expect(0, 66517, '\P{ 	IS_Xpeo}', "");
Expect(1, 66517, '\P{^ 	IS_Xpeo}', "");
Expect(0, 66518, '\p{ 	IS_Xpeo}', "");
Expect(1, 66518, '\p{^ 	IS_Xpeo}', "");
Expect(1, 66518, '\P{ 	IS_Xpeo}', "");
Expect(0, 66518, '\P{^ 	IS_Xpeo}', "");
Error('\p{/a/-	old_South_Arabian}');
Error('\P{/a/-	old_South_Arabian}');
Expect(1, 68223, '\p{oldsoutharabian}', "");
Expect(0, 68223, '\p{^oldsoutharabian}', "");
Expect(0, 68223, '\P{oldsoutharabian}', "");
Expect(1, 68223, '\P{^oldsoutharabian}', "");
Expect(0, 68224, '\p{oldsoutharabian}', "");
Expect(1, 68224, '\p{^oldsoutharabian}', "");
Expect(1, 68224, '\P{oldsoutharabian}', "");
Expect(0, 68224, '\P{^oldsoutharabian}', "");
Expect(1, 68223, '\p{--old_south_Arabian}', "");
Expect(0, 68223, '\p{^--old_south_Arabian}', "");
Expect(0, 68223, '\P{--old_south_Arabian}', "");
Expect(1, 68223, '\P{^--old_south_Arabian}', "");
Expect(0, 68224, '\p{--old_south_Arabian}', "");
Expect(1, 68224, '\p{^--old_south_Arabian}', "");
Expect(1, 68224, '\P{--old_south_Arabian}', "");
Expect(0, 68224, '\P{^--old_south_Arabian}', "");
Error('\p{	Is_Old_South_ARABIAN/a/}');
Error('\P{	Is_Old_South_ARABIAN/a/}');
Expect(1, 68223, '\p{isoldsoutharabian}', "");
Expect(0, 68223, '\p{^isoldsoutharabian}', "");
Expect(0, 68223, '\P{isoldsoutharabian}', "");
Expect(1, 68223, '\P{^isoldsoutharabian}', "");
Expect(0, 68224, '\p{isoldsoutharabian}', "");
Expect(1, 68224, '\p{^isoldsoutharabian}', "");
Expect(1, 68224, '\P{isoldsoutharabian}', "");
Expect(0, 68224, '\P{^isoldsoutharabian}', "");
Expect(1, 68223, '\p{-	Is_OLD_SOUTH_ARABIAN}', "");
Expect(0, 68223, '\p{^-	Is_OLD_SOUTH_ARABIAN}', "");
Expect(0, 68223, '\P{-	Is_OLD_SOUTH_ARABIAN}', "");
Expect(1, 68223, '\P{^-	Is_OLD_SOUTH_ARABIAN}', "");
Expect(0, 68224, '\p{-	Is_OLD_SOUTH_ARABIAN}', "");
Expect(1, 68224, '\p{^-	Is_OLD_SOUTH_ARABIAN}', "");
Expect(1, 68224, '\P{-	Is_OLD_SOUTH_ARABIAN}', "");
Expect(0, 68224, '\P{^-	Is_OLD_SOUTH_ARABIAN}', "");
Error('\p{	:=Sarb}');
Error('\P{	:=Sarb}');
Expect(1, 68223, '\p{sarb}', "");
Expect(0, 68223, '\p{^sarb}', "");
Expect(0, 68223, '\P{sarb}', "");
Expect(1, 68223, '\P{^sarb}', "");
Expect(0, 68224, '\p{sarb}', "");
Expect(1, 68224, '\p{^sarb}', "");
Expect(1, 68224, '\P{sarb}', "");
Expect(0, 68224, '\P{^sarb}', "");
Expect(1, 68223, '\p{		SARB}', "");
Expect(0, 68223, '\p{^		SARB}', "");
Expect(0, 68223, '\P{		SARB}', "");
Expect(1, 68223, '\P{^		SARB}', "");
Expect(0, 68224, '\p{		SARB}', "");
Expect(1, 68224, '\p{^		SARB}', "");
Expect(1, 68224, '\P{		SARB}', "");
Expect(0, 68224, '\P{^		SARB}', "");
Error('\p{:=--IS_SARB}');
Error('\P{:=--IS_SARB}');
Expect(1, 68223, '\p{issarb}', "");
Expect(0, 68223, '\p{^issarb}', "");
Expect(0, 68223, '\P{issarb}', "");
Expect(1, 68223, '\P{^issarb}', "");
Expect(0, 68224, '\p{issarb}', "");
Expect(1, 68224, '\p{^issarb}', "");
Expect(1, 68224, '\P{issarb}', "");
Expect(0, 68224, '\P{^issarb}', "");
Expect(1, 68223, '\p{-is_Sarb}', "");
Expect(0, 68223, '\p{^-is_Sarb}', "");
Expect(0, 68223, '\P{-is_Sarb}', "");
Expect(1, 68223, '\P{^-is_Sarb}', "");
Expect(0, 68224, '\p{-is_Sarb}', "");
Expect(1, 68224, '\p{^-is_Sarb}', "");
Expect(1, 68224, '\P{-is_Sarb}', "");
Expect(0, 68224, '\P{^-is_Sarb}', "");
Error('\p{--OLD_TURKIC/a/}');
Error('\P{--OLD_TURKIC/a/}');
Expect(1, 68680, '\p{oldturkic}', "");
Expect(0, 68680, '\p{^oldturkic}', "");
Expect(0, 68680, '\P{oldturkic}', "");
Expect(1, 68680, '\P{^oldturkic}', "");
Expect(0, 68681, '\p{oldturkic}', "");
Expect(1, 68681, '\p{^oldturkic}', "");
Expect(1, 68681, '\P{oldturkic}', "");
Expect(0, 68681, '\P{^oldturkic}', "");
Expect(1, 68680, '\p{- Old_Turkic}', "");
Expect(0, 68680, '\p{^- Old_Turkic}', "");
Expect(0, 68680, '\P{- Old_Turkic}', "");
Expect(1, 68680, '\P{^- Old_Turkic}', "");
Expect(0, 68681, '\p{- Old_Turkic}', "");
Expect(1, 68681, '\p{^- Old_Turkic}', "");
Expect(1, 68681, '\P{- Old_Turkic}', "");
Expect(0, 68681, '\P{^- Old_Turkic}', "");
Error('\p{/a/-	Is_Old_TURKIC}');
Error('\P{/a/-	Is_Old_TURKIC}');
Expect(1, 68680, '\p{isoldturkic}', "");
Expect(0, 68680, '\p{^isoldturkic}', "");
Expect(0, 68680, '\P{isoldturkic}', "");
Expect(1, 68680, '\P{^isoldturkic}', "");
Expect(0, 68681, '\p{isoldturkic}', "");
Expect(1, 68681, '\p{^isoldturkic}', "");
Expect(1, 68681, '\P{isoldturkic}', "");
Expect(0, 68681, '\P{^isoldturkic}', "");
Expect(1, 68680, '\p{- IS_old_Turkic}', "");
Expect(0, 68680, '\p{^- IS_old_Turkic}', "");
Expect(0, 68680, '\P{- IS_old_Turkic}', "");
Expect(1, 68680, '\P{^- IS_old_Turkic}', "");
Expect(0, 68681, '\p{- IS_old_Turkic}', "");
Expect(1, 68681, '\p{^- IS_old_Turkic}', "");
Expect(1, 68681, '\P{- IS_old_Turkic}', "");
Expect(0, 68681, '\P{^- IS_old_Turkic}', "");
Error('\p{/a/	-Orkh}');
Error('\P{/a/	-Orkh}');
Expect(1, 68680, '\p{orkh}', "");
Expect(0, 68680, '\p{^orkh}', "");
Expect(0, 68680, '\P{orkh}', "");
Expect(1, 68680, '\P{^orkh}', "");
Expect(0, 68681, '\p{orkh}', "");
Expect(1, 68681, '\p{^orkh}', "");
Expect(1, 68681, '\P{orkh}', "");
Expect(0, 68681, '\P{^orkh}', "");
Expect(1, 68680, '\p{ _orkh}', "");
Expect(0, 68680, '\p{^ _orkh}', "");
Expect(0, 68680, '\P{ _orkh}', "");
Expect(1, 68680, '\P{^ _orkh}', "");
Expect(0, 68681, '\p{ _orkh}', "");
Expect(1, 68681, '\p{^ _orkh}', "");
Expect(1, 68681, '\P{ _orkh}', "");
Expect(0, 68681, '\P{^ _orkh}', "");
Error('\p{:=_IS_Orkh}');
Error('\P{:=_IS_Orkh}');
Expect(1, 68680, '\p{isorkh}', "");
Expect(0, 68680, '\p{^isorkh}', "");
Expect(0, 68680, '\P{isorkh}', "");
Expect(1, 68680, '\P{^isorkh}', "");
Expect(0, 68681, '\p{isorkh}', "");
Expect(1, 68681, '\p{^isorkh}', "");
Expect(1, 68681, '\P{isorkh}', "");
Expect(0, 68681, '\P{^isorkh}', "");
Expect(1, 68680, '\p{-_Is_orkh}', "");
Expect(0, 68680, '\p{^-_Is_orkh}', "");
Expect(0, 68680, '\P{-_Is_orkh}', "");
Expect(1, 68680, '\P{^-_Is_orkh}', "");
Expect(0, 68681, '\p{-_Is_orkh}', "");
Expect(1, 68681, '\p{^-_Is_orkh}', "");
Expect(1, 68681, '\P{-_Is_orkh}', "");
Expect(0, 68681, '\P{^-_Is_orkh}', "");
Error('\p{/a/  open_punctuation}');
Error('\P{/a/  open_punctuation}');
Expect(1, 65378, '\p{openpunctuation}', "");
Expect(0, 65378, '\p{^openpunctuation}', "");
Expect(0, 65378, '\P{openpunctuation}', "");
Expect(1, 65378, '\P{^openpunctuation}', "");
Expect(0, 65379, '\p{openpunctuation}', "");
Expect(1, 65379, '\p{^openpunctuation}', "");
Expect(1, 65379, '\P{openpunctuation}', "");
Expect(0, 65379, '\P{^openpunctuation}', "");
Expect(1, 65378, '\p{_OPEN_PUNCTUATION}', "");
Expect(0, 65378, '\p{^_OPEN_PUNCTUATION}', "");
Expect(0, 65378, '\P{_OPEN_PUNCTUATION}', "");
Expect(1, 65378, '\P{^_OPEN_PUNCTUATION}', "");
Expect(0, 65379, '\p{_OPEN_PUNCTUATION}', "");
Expect(1, 65379, '\p{^_OPEN_PUNCTUATION}', "");
Expect(1, 65379, '\P{_OPEN_PUNCTUATION}', "");
Expect(0, 65379, '\P{^_OPEN_PUNCTUATION}', "");
Error('\p{	/a/IS_Open_punctuation}');
Error('\P{	/a/IS_Open_punctuation}');
Expect(1, 65378, '\p{isopenpunctuation}', "");
Expect(0, 65378, '\p{^isopenpunctuation}', "");
Expect(0, 65378, '\P{isopenpunctuation}', "");
Expect(1, 65378, '\P{^isopenpunctuation}', "");
Expect(0, 65379, '\p{isopenpunctuation}', "");
Expect(1, 65379, '\p{^isopenpunctuation}', "");
Expect(1, 65379, '\P{isopenpunctuation}', "");
Expect(0, 65379, '\P{^isopenpunctuation}', "");
Expect(1, 65378, '\p{  Is_open_Punctuation}', "");
Expect(0, 65378, '\p{^  Is_open_Punctuation}', "");
Expect(0, 65378, '\P{  Is_open_Punctuation}', "");
Expect(1, 65378, '\P{^  Is_open_Punctuation}', "");
Expect(0, 65379, '\p{  Is_open_Punctuation}', "");
Expect(1, 65379, '\p{^  Is_open_Punctuation}', "");
Expect(1, 65379, '\P{  Is_open_Punctuation}', "");
Expect(0, 65379, '\P{^  Is_open_Punctuation}', "");
Error('\p{	Ps/a/}');
Error('\P{	Ps/a/}');
Expect(1, 65378, '\p{ps}', "");
Expect(0, 65378, '\p{^ps}', "");
Expect(0, 65378, '\P{ps}', "");
Expect(1, 65378, '\P{^ps}', "");
Expect(0, 65379, '\p{ps}', "");
Expect(1, 65379, '\p{^ps}', "");
Expect(1, 65379, '\P{ps}', "");
Expect(0, 65379, '\P{^ps}', "");
Expect(1, 65378, '\p{ ps}', "");
Expect(0, 65378, '\p{^ ps}', "");
Expect(0, 65378, '\P{ ps}', "");
Expect(1, 65378, '\P{^ ps}', "");
Expect(0, 65379, '\p{ ps}', "");
Expect(1, 65379, '\p{^ ps}', "");
Expect(1, 65379, '\P{ ps}', "");
Expect(0, 65379, '\P{^ ps}', "");
Error('\p{:=_ Is_ps}');
Error('\P{:=_ Is_ps}');
Expect(1, 65378, '\p{isps}', "");
Expect(0, 65378, '\p{^isps}', "");
Expect(0, 65378, '\P{isps}', "");
Expect(1, 65378, '\P{^isps}', "");
Expect(0, 65379, '\p{isps}', "");
Expect(1, 65379, '\p{^isps}', "");
Expect(1, 65379, '\P{isps}', "");
Expect(0, 65379, '\P{^isps}', "");
Expect(1, 65378, '\p{ -is_ps}', "");
Expect(0, 65378, '\p{^ -is_ps}', "");
Expect(0, 65378, '\P{ -is_ps}', "");
Expect(1, 65378, '\P{^ -is_ps}', "");
Expect(0, 65379, '\p{ -is_ps}', "");
Expect(1, 65379, '\p{^ -is_ps}', "");
Expect(1, 65379, '\P{ -is_ps}', "");
Expect(0, 65379, '\P{^ -is_ps}', "");
Error('\p{ -optical_CHARACTER_Recognition/a/}');
Error('\P{ -optical_CHARACTER_Recognition/a/}');
Expect(1, 9311, '\p{opticalcharacterrecognition}', "");
Expect(0, 9311, '\p{^opticalcharacterrecognition}', "");
Expect(0, 9311, '\P{opticalcharacterrecognition}', "");
Expect(1, 9311, '\P{^opticalcharacterrecognition}', "");
Expect(0, 9312, '\p{opticalcharacterrecognition}', "");
Expect(1, 9312, '\p{^opticalcharacterrecognition}', "");
Expect(1, 9312, '\P{opticalcharacterrecognition}', "");
Expect(0, 9312, '\P{^opticalcharacterrecognition}', "");
Expect(1, 9311, '\p{-	Optical_Character_Recognition}', "");
Expect(0, 9311, '\p{^-	Optical_Character_Recognition}', "");
Expect(0, 9311, '\P{-	Optical_Character_Recognition}', "");
Expect(1, 9311, '\P{^-	Optical_Character_Recognition}', "");
Expect(0, 9312, '\p{-	Optical_Character_Recognition}', "");
Expect(1, 9312, '\p{^-	Optical_Character_Recognition}', "");
Expect(1, 9312, '\P{-	Optical_Character_Recognition}', "");
Expect(0, 9312, '\P{^-	Optical_Character_Recognition}', "");
Error('\p{:= 	Is_Optical_Character_recognition}');
Error('\P{:= 	Is_Optical_Character_recognition}');
Expect(1, 9311, '\p{isopticalcharacterrecognition}', "");
Expect(0, 9311, '\p{^isopticalcharacterrecognition}', "");
Expect(0, 9311, '\P{isopticalcharacterrecognition}', "");
Expect(1, 9311, '\P{^isopticalcharacterrecognition}', "");
Expect(0, 9312, '\p{isopticalcharacterrecognition}', "");
Expect(1, 9312, '\p{^isopticalcharacterrecognition}', "");
Expect(1, 9312, '\P{isopticalcharacterrecognition}', "");
Expect(0, 9312, '\P{^isopticalcharacterrecognition}', "");
Expect(1, 9311, '\p{_ IS_Optical_character_Recognition}', "");
Expect(0, 9311, '\p{^_ IS_Optical_character_Recognition}', "");
Expect(0, 9311, '\P{_ IS_Optical_character_Recognition}', "");
Expect(1, 9311, '\P{^_ IS_Optical_character_Recognition}', "");
Expect(0, 9312, '\p{_ IS_Optical_character_Recognition}', "");
Expect(1, 9312, '\p{^_ IS_Optical_character_Recognition}', "");
Expect(1, 9312, '\P{_ IS_Optical_character_Recognition}', "");
Expect(0, 9312, '\P{^_ IS_Optical_character_Recognition}', "");
Error('\p{/a/	In_OPTICAL_character_recognition}');
Error('\P{/a/	In_OPTICAL_character_recognition}');
Expect(1, 9311, '\p{inopticalcharacterrecognition}', "");
Expect(0, 9311, '\p{^inopticalcharacterrecognition}', "");
Expect(0, 9311, '\P{inopticalcharacterrecognition}', "");
Expect(1, 9311, '\P{^inopticalcharacterrecognition}', "");
Expect(0, 9312, '\p{inopticalcharacterrecognition}', "");
Expect(1, 9312, '\p{^inopticalcharacterrecognition}', "");
Expect(1, 9312, '\P{inopticalcharacterrecognition}', "");
Expect(0, 9312, '\P{^inopticalcharacterrecognition}', "");
Expect(1, 9311, '\p{ in_Optical_Character_Recognition}', "");
Expect(0, 9311, '\p{^ in_Optical_Character_Recognition}', "");
Expect(0, 9311, '\P{ in_Optical_Character_Recognition}', "");
Expect(1, 9311, '\P{^ in_Optical_Character_Recognition}', "");
Expect(0, 9312, '\p{ in_Optical_Character_Recognition}', "");
Expect(1, 9312, '\p{^ in_Optical_Character_Recognition}', "");
Expect(1, 9312, '\P{ in_Optical_Character_Recognition}', "");
Expect(0, 9312, '\P{^ in_Optical_Character_Recognition}', "");
Error('\p{	:=ocr}');
Error('\P{	:=ocr}');
Expect(1, 9311, '\p{ocr}', "");
Expect(0, 9311, '\p{^ocr}', "");
Expect(0, 9311, '\P{ocr}', "");
Expect(1, 9311, '\P{^ocr}', "");
Expect(0, 9312, '\p{ocr}', "");
Expect(1, 9312, '\p{^ocr}', "");
Expect(1, 9312, '\P{ocr}', "");
Expect(0, 9312, '\P{^ocr}', "");
Expect(1, 9311, '\p{- OCR}', "");
Expect(0, 9311, '\p{^- OCR}', "");
Expect(0, 9311, '\P{- OCR}', "");
Expect(1, 9311, '\P{^- OCR}', "");
Expect(0, 9312, '\p{- OCR}', "");
Expect(1, 9312, '\p{^- OCR}', "");
Expect(1, 9312, '\P{- OCR}', "");
Expect(0, 9312, '\P{^- OCR}', "");
Error('\p{/a/-	Is_OCR}');
Error('\P{/a/-	Is_OCR}');
Expect(1, 9311, '\p{isocr}', "");
Expect(0, 9311, '\p{^isocr}', "");
Expect(0, 9311, '\P{isocr}', "");
Expect(1, 9311, '\P{^isocr}', "");
Expect(0, 9312, '\p{isocr}', "");
Expect(1, 9312, '\p{^isocr}', "");
Expect(1, 9312, '\P{isocr}', "");
Expect(0, 9312, '\P{^isocr}', "");
Expect(1, 9311, '\p{	Is_OCR}', "");
Expect(0, 9311, '\p{^	Is_OCR}', "");
Expect(0, 9311, '\P{	Is_OCR}', "");
Expect(1, 9311, '\P{^	Is_OCR}', "");
Expect(0, 9312, '\p{	Is_OCR}', "");
Expect(1, 9312, '\p{^	Is_OCR}', "");
Expect(1, 9312, '\P{	Is_OCR}', "");
Expect(0, 9312, '\P{^	Is_OCR}', "");
Error('\p{-:=IN_ocr}');
Error('\P{-:=IN_ocr}');
Expect(1, 9311, '\p{inocr}', "");
Expect(0, 9311, '\p{^inocr}', "");
Expect(0, 9311, '\P{inocr}', "");
Expect(1, 9311, '\P{^inocr}', "");
Expect(0, 9312, '\p{inocr}', "");
Expect(1, 9312, '\p{^inocr}', "");
Expect(1, 9312, '\P{inocr}', "");
Expect(0, 9312, '\P{^inocr}', "");
Expect(1, 9311, '\p{_in_ocr}', "");
Expect(0, 9311, '\p{^_in_ocr}', "");
Expect(0, 9311, '\P{_in_ocr}', "");
Expect(1, 9311, '\P{^_in_ocr}', "");
Expect(0, 9312, '\p{_in_ocr}', "");
Expect(1, 9312, '\p{^_in_ocr}', "");
Expect(1, 9312, '\P{_in_ocr}', "");
Expect(0, 9312, '\P{^_in_ocr}', "");
Error('\p{  ORIYA:=}');
Error('\P{  ORIYA:=}');
Expect(1, 2935, '\p{oriya}', "");
Expect(0, 2935, '\p{^oriya}', "");
Expect(0, 2935, '\P{oriya}', "");
Expect(1, 2935, '\P{^oriya}', "");
Expect(0, 2936, '\p{oriya}', "");
Expect(1, 2936, '\p{^oriya}', "");
Expect(1, 2936, '\P{oriya}', "");
Expect(0, 2936, '\P{^oriya}', "");
Expect(1, 2935, '\p{ ORIYA}', "");
Expect(0, 2935, '\p{^ ORIYA}', "");
Expect(0, 2935, '\P{ ORIYA}', "");
Expect(1, 2935, '\P{^ ORIYA}', "");
Expect(0, 2936, '\p{ ORIYA}', "");
Expect(1, 2936, '\p{^ ORIYA}', "");
Expect(1, 2936, '\P{ ORIYA}', "");
Expect(0, 2936, '\P{^ ORIYA}', "");
Error('\p{-	is_Oriya/a/}');
Error('\P{-	is_Oriya/a/}');
Expect(1, 2935, '\p{isoriya}', "");
Expect(0, 2935, '\p{^isoriya}', "");
Expect(0, 2935, '\P{isoriya}', "");
Expect(1, 2935, '\P{^isoriya}', "");
Expect(0, 2936, '\p{isoriya}', "");
Expect(1, 2936, '\p{^isoriya}', "");
Expect(1, 2936, '\P{isoriya}', "");
Expect(0, 2936, '\P{^isoriya}', "");
Expect(1, 2935, '\p{--Is_oriya}', "");
Expect(0, 2935, '\p{^--Is_oriya}', "");
Expect(0, 2935, '\P{--Is_oriya}', "");
Expect(1, 2935, '\P{^--Is_oriya}', "");
Expect(0, 2936, '\p{--Is_oriya}', "");
Expect(1, 2936, '\p{^--Is_oriya}', "");
Expect(1, 2936, '\P{--Is_oriya}', "");
Expect(0, 2936, '\P{^--Is_oriya}', "");
Error('\p{- ORYA:=}');
Error('\P{- ORYA:=}');
Expect(1, 2935, '\p{orya}', "");
Expect(0, 2935, '\p{^orya}', "");
Expect(0, 2935, '\P{orya}', "");
Expect(1, 2935, '\P{^orya}', "");
Expect(0, 2936, '\p{orya}', "");
Expect(1, 2936, '\p{^orya}', "");
Expect(1, 2936, '\P{orya}', "");
Expect(0, 2936, '\P{^orya}', "");
Expect(1, 2935, '\p{	ORYA}', "");
Expect(0, 2935, '\p{^	ORYA}', "");
Expect(0, 2935, '\P{	ORYA}', "");
Expect(1, 2935, '\P{^	ORYA}', "");
Expect(0, 2936, '\p{	ORYA}', "");
Expect(1, 2936, '\p{^	ORYA}', "");
Expect(1, 2936, '\P{	ORYA}', "");
Expect(0, 2936, '\P{^	ORYA}', "");
Error('\p{_:=is_orya}');
Error('\P{_:=is_orya}');
Expect(1, 2935, '\p{isorya}', "");
Expect(0, 2935, '\p{^isorya}', "");
Expect(0, 2935, '\P{isorya}', "");
Expect(1, 2935, '\P{^isorya}', "");
Expect(0, 2936, '\p{isorya}', "");
Expect(1, 2936, '\p{^isorya}', "");
Expect(1, 2936, '\P{isorya}', "");
Expect(0, 2936, '\P{^isorya}', "");
Expect(1, 2935, '\p{--is_Orya}', "");
Expect(0, 2935, '\p{^--is_Orya}', "");
Expect(0, 2935, '\P{--is_Orya}', "");
Expect(1, 2935, '\P{^--is_Orya}', "");
Expect(0, 2936, '\p{--is_Orya}', "");
Expect(1, 2936, '\p{^--is_Orya}', "");
Expect(1, 2936, '\P{--is_Orya}', "");
Expect(0, 2936, '\P{^--is_Orya}', "");
Error('\p{	:=osmanya}');
Error('\P{	:=osmanya}');
Expect(1, 66729, '\p{osmanya}', "");
Expect(0, 66729, '\p{^osmanya}', "");
Expect(0, 66729, '\P{osmanya}', "");
Expect(1, 66729, '\P{^osmanya}', "");
Expect(0, 66730, '\p{osmanya}', "");
Expect(1, 66730, '\p{^osmanya}', "");
Expect(1, 66730, '\P{osmanya}', "");
Expect(0, 66730, '\P{^osmanya}', "");
Expect(1, 66729, '\p{ Osmanya}', "");
Expect(0, 66729, '\p{^ Osmanya}', "");
Expect(0, 66729, '\P{ Osmanya}', "");
Expect(1, 66729, '\P{^ Osmanya}', "");
Expect(0, 66730, '\p{ Osmanya}', "");
Expect(1, 66730, '\p{^ Osmanya}', "");
Expect(1, 66730, '\P{ Osmanya}', "");
Expect(0, 66730, '\P{^ Osmanya}', "");
Error('\p{/a/Is_Osmanya}');
Error('\P{/a/Is_Osmanya}');
Expect(1, 66729, '\p{isosmanya}', "");
Expect(0, 66729, '\p{^isosmanya}', "");
Expect(0, 66729, '\P{isosmanya}', "");
Expect(1, 66729, '\P{^isosmanya}', "");
Expect(0, 66730, '\p{isosmanya}', "");
Expect(1, 66730, '\p{^isosmanya}', "");
Expect(1, 66730, '\P{isosmanya}', "");
Expect(0, 66730, '\P{^isosmanya}', "");
Expect(1, 66729, '\p{	Is_Osmanya}', "");
Expect(0, 66729, '\p{^	Is_Osmanya}', "");
Expect(0, 66729, '\P{	Is_Osmanya}', "");
Expect(1, 66729, '\P{^	Is_Osmanya}', "");
Expect(0, 66730, '\p{	Is_Osmanya}', "");
Expect(1, 66730, '\p{^	Is_Osmanya}', "");
Expect(1, 66730, '\P{	Is_Osmanya}', "");
Expect(0, 66730, '\P{^	Is_Osmanya}', "");
Error('\p{_Osma:=}');
Error('\P{_Osma:=}');
Expect(1, 66729, '\p{osma}', "");
Expect(0, 66729, '\p{^osma}', "");
Expect(0, 66729, '\P{osma}', "");
Expect(1, 66729, '\P{^osma}', "");
Expect(0, 66730, '\p{osma}', "");
Expect(1, 66730, '\p{^osma}', "");
Expect(1, 66730, '\P{osma}', "");
Expect(0, 66730, '\P{^osma}', "");
Expect(1, 66729, '\p{- Osma}', "");
Expect(0, 66729, '\p{^- Osma}', "");
Expect(0, 66729, '\P{- Osma}', "");
Expect(1, 66729, '\P{^- Osma}', "");
Expect(0, 66730, '\p{- Osma}', "");
Expect(1, 66730, '\p{^- Osma}', "");
Expect(1, 66730, '\P{- Osma}', "");
Expect(0, 66730, '\P{^- Osma}', "");
Error('\p{ -IS_Osma/a/}');
Error('\P{ -IS_Osma/a/}');
Expect(1, 66729, '\p{isosma}', "");
Expect(0, 66729, '\p{^isosma}', "");
Expect(0, 66729, '\P{isosma}', "");
Expect(1, 66729, '\P{^isosma}', "");
Expect(0, 66730, '\p{isosma}', "");
Expect(1, 66730, '\p{^isosma}', "");
Expect(1, 66730, '\P{isosma}', "");
Expect(0, 66730, '\P{^isosma}', "");
Expect(1, 66729, '\p{	-is_Osma}', "");
Expect(0, 66729, '\p{^	-is_Osma}', "");
Expect(0, 66729, '\P{	-is_Osma}', "");
Expect(1, 66729, '\P{^	-is_Osma}', "");
Expect(0, 66730, '\p{	-is_Osma}', "");
Expect(1, 66730, '\p{^	-is_Osma}', "");
Expect(1, 66730, '\P{	-is_Osma}', "");
Expect(0, 66730, '\P{^	-is_Osma}', "");
Error('\p{-Other:=}');
Error('\P{-Other:=}');
Expect(1, 918000, '\p{other}', "");
Expect(0, 918000, '\p{^other}', "");
Expect(0, 918000, '\P{other}', "");
Expect(1, 918000, '\P{^other}', "");
Expect(0, 917999, '\p{other}', "");
Expect(1, 917999, '\p{^other}', "");
Expect(1, 917999, '\P{other}', "");
Expect(0, 917999, '\P{^other}', "");
Expect(1, 918000, '\p{ 	Other}', "");
Expect(0, 918000, '\p{^ 	Other}', "");
Expect(0, 918000, '\P{ 	Other}', "");
Expect(1, 918000, '\P{^ 	Other}', "");
Expect(0, 917999, '\p{ 	Other}', "");
Expect(1, 917999, '\p{^ 	Other}', "");
Expect(1, 917999, '\P{ 	Other}', "");
Expect(0, 917999, '\P{^ 	Other}', "");
Error('\p{	/a/is_Other}');
Error('\P{	/a/is_Other}');
Expect(1, 918000, '\p{isother}', "");
Expect(0, 918000, '\p{^isother}', "");
Expect(0, 918000, '\P{isother}', "");
Expect(1, 918000, '\P{^isother}', "");
Expect(0, 917999, '\p{isother}', "");
Expect(1, 917999, '\p{^isother}', "");
Expect(1, 917999, '\P{isother}', "");
Expect(0, 917999, '\P{^isother}', "");
Expect(1, 918000, '\p{		Is_Other}', "");
Expect(0, 918000, '\p{^		Is_Other}', "");
Expect(0, 918000, '\P{		Is_Other}', "");
Expect(1, 918000, '\P{^		Is_Other}', "");
Expect(0, 917999, '\p{		Is_Other}', "");
Expect(1, 917999, '\p{^		Is_Other}', "");
Expect(1, 917999, '\P{		Is_Other}', "");
Expect(0, 917999, '\P{^		Is_Other}', "");
Error('\p{-/a/c}');
Error('\P{-/a/c}');
Expect(1, 918000, '\p{c}', "");
Expect(0, 918000, '\p{^c}', "");
Expect(0, 918000, '\P{c}', "");
Expect(1, 918000, '\P{^c}', "");
Expect(0, 917999, '\p{c}', "");
Expect(1, 917999, '\p{^c}', "");
Expect(1, 917999, '\P{c}', "");
Expect(0, 917999, '\P{^c}', "");
Expect(1, 918000, '\p{-	C}', "");
Expect(0, 918000, '\p{^-	C}', "");
Expect(0, 918000, '\P{-	C}', "");
Expect(1, 918000, '\P{^-	C}', "");
Expect(0, 917999, '\p{-	C}', "");
Expect(1, 917999, '\p{^-	C}', "");
Expect(1, 917999, '\P{-	C}', "");
Expect(0, 917999, '\P{^-	C}', "");
Error('\p{-	Is_c:=}');
Error('\P{-	Is_c:=}');
Expect(1, 918000, '\p{isc}', "");
Expect(0, 918000, '\p{^isc}', "");
Expect(0, 918000, '\P{isc}', "");
Expect(1, 918000, '\P{^isc}', "");
Expect(0, 917999, '\p{isc}', "");
Expect(1, 917999, '\p{^isc}', "");
Expect(1, 917999, '\P{isc}', "");
Expect(0, 917999, '\P{^isc}', "");
Expect(1, 918000, '\p{is_c}', "");
Expect(0, 918000, '\p{^is_c}', "");
Expect(0, 918000, '\P{is_c}', "");
Expect(1, 918000, '\P{^is_c}', "");
Expect(0, 917999, '\p{is_c}', "");
Expect(1, 917999, '\p{^is_c}', "");
Expect(1, 917999, '\P{is_c}', "");
Expect(0, 917999, '\P{^is_c}', "");
Error('\p{Other_Alphabetic}');
Error('\P{Other_Alphabetic}');
Error('\p{Is_Other_Alphabetic}');
Error('\P{Is_Other_Alphabetic}');
Error('\p{OAlpha}');
Error('\P{OAlpha}');
Error('\p{Is_OAlpha}');
Error('\P{Is_OAlpha}');
Error('\p{Other_Default_Ignorable_Code_Point}');
Error('\P{Other_Default_Ignorable_Code_Point}');
Error('\p{Is_Other_Default_Ignorable_Code_Point}');
Error('\P{Is_Other_Default_Ignorable_Code_Point}');
Error('\p{ODI}');
Error('\P{ODI}');
Error('\p{Is_ODI}');
Error('\P{Is_ODI}');
Error('\p{Other_Grapheme_Extend}');
Error('\P{Other_Grapheme_Extend}');
Error('\p{Is_Other_Grapheme_Extend}');
Error('\P{Is_Other_Grapheme_Extend}');
Error('\p{OGr_Ext}');
Error('\P{OGr_Ext}');
Error('\p{Is_OGr_Ext}');
Error('\P{Is_OGr_Ext}');
Error('\p{Other_ID_Continue}');
Error('\P{Other_ID_Continue}');
Error('\p{Is_Other_ID_Continue}');
Error('\P{Is_Other_ID_Continue}');
Error('\p{OIDC}');
Error('\P{OIDC}');
Error('\p{Is_OIDC}');
Error('\P{Is_OIDC}');
Error('\p{Other_ID_Start}');
Error('\P{Other_ID_Start}');
Error('\p{Is_Other_ID_Start}');
Error('\P{Is_Other_ID_Start}');
Error('\p{OIDS}');
Error('\P{OIDS}');
Error('\p{Is_OIDS}');
Error('\P{Is_OIDS}');
Error('\p{/a/-other_LETTER}');
Error('\P{/a/-other_LETTER}');
Expect(1, 195101, '\p{otherletter}', "");
Expect(0, 195101, '\p{^otherletter}', "");
Expect(0, 195101, '\P{otherletter}', "");
Expect(1, 195101, '\P{^otherletter}', "");
Expect(0, 195102, '\p{otherletter}', "");
Expect(1, 195102, '\p{^otherletter}', "");
Expect(1, 195102, '\P{otherletter}', "");
Expect(0, 195102, '\P{^otherletter}', "");
Expect(1, 195101, '\p{ Other_letter}', "");
Expect(0, 195101, '\p{^ Other_letter}', "");
Expect(0, 195101, '\P{ Other_letter}', "");
Expect(1, 195101, '\P{^ Other_letter}', "");
Expect(0, 195102, '\p{ Other_letter}', "");
Expect(1, 195102, '\p{^ Other_letter}', "");
Expect(1, 195102, '\P{ Other_letter}', "");
Expect(0, 195102, '\P{^ Other_letter}', "");
Error('\p{		is_Other_Letter/a/}');
Error('\P{		is_Other_Letter/a/}');
Expect(1, 195101, '\p{isotherletter}', "");
Expect(0, 195101, '\p{^isotherletter}', "");
Expect(0, 195101, '\P{isotherletter}', "");
Expect(1, 195101, '\P{^isotherletter}', "");
Expect(0, 195102, '\p{isotherletter}', "");
Expect(1, 195102, '\p{^isotherletter}', "");
Expect(1, 195102, '\P{isotherletter}', "");
Expect(0, 195102, '\P{^isotherletter}', "");
Expect(1, 195101, '\p{_Is_OTHER_LETTER}', "");
Expect(0, 195101, '\p{^_Is_OTHER_LETTER}', "");
Expect(0, 195101, '\P{_Is_OTHER_LETTER}', "");
Expect(1, 195101, '\P{^_Is_OTHER_LETTER}', "");
Expect(0, 195102, '\p{_Is_OTHER_LETTER}', "");
Expect(1, 195102, '\p{^_Is_OTHER_LETTER}', "");
Expect(1, 195102, '\P{_Is_OTHER_LETTER}', "");
Expect(0, 195102, '\P{^_Is_OTHER_LETTER}', "");
Error('\p{ Lo:=}');
Error('\P{ Lo:=}');
Expect(1, 195101, '\p{lo}', "");
Expect(0, 195101, '\p{^lo}', "");
Expect(0, 195101, '\P{lo}', "");
Expect(1, 195101, '\P{^lo}', "");
Expect(0, 195102, '\p{lo}', "");
Expect(1, 195102, '\p{^lo}', "");
Expect(1, 195102, '\P{lo}', "");
Expect(0, 195102, '\P{^lo}', "");
Expect(1, 195101, '\p{_	LO}', "");
Expect(0, 195101, '\p{^_	LO}', "");
Expect(0, 195101, '\P{_	LO}', "");
Expect(1, 195101, '\P{^_	LO}', "");
Expect(0, 195102, '\p{_	LO}', "");
Expect(1, 195102, '\p{^_	LO}', "");
Expect(1, 195102, '\P{_	LO}', "");
Expect(0, 195102, '\P{^_	LO}', "");
Error('\p{-:=IS_Lo}');
Error('\P{-:=IS_Lo}');
Expect(1, 195101, '\p{islo}', "");
Expect(0, 195101, '\p{^islo}', "");
Expect(0, 195101, '\P{islo}', "");
Expect(1, 195101, '\P{^islo}', "");
Expect(0, 195102, '\p{islo}', "");
Expect(1, 195102, '\p{^islo}', "");
Expect(1, 195102, '\P{islo}', "");
Expect(0, 195102, '\P{^islo}', "");
Expect(1, 195101, '\p{_ Is_Lo}', "");
Expect(0, 195101, '\p{^_ Is_Lo}', "");
Expect(0, 195101, '\P{_ Is_Lo}', "");
Expect(1, 195101, '\P{^_ Is_Lo}', "");
Expect(0, 195102, '\p{_ Is_Lo}', "");
Expect(1, 195102, '\p{^_ Is_Lo}', "");
Expect(1, 195102, '\P{_ Is_Lo}', "");
Expect(0, 195102, '\P{^_ Is_Lo}', "");
Error('\p{Other_Lowercase}');
Error('\P{Other_Lowercase}');
Error('\p{Is_Other_Lowercase}');
Error('\P{Is_Other_Lowercase}');
Error('\p{OLower}');
Error('\P{OLower}');
Error('\p{Is_OLower}');
Error('\P{Is_OLower}');
Error('\p{Other_Math}');
Error('\P{Other_Math}');
Error('\p{Is_Other_Math}');
Error('\P{Is_Other_Math}');
Error('\p{OMath}');
Error('\P{OMath}');
Error('\p{Is_OMath}');
Error('\P{Is_OMath}');
Error('\p{	/a/OTHER_NUMBER}');
Error('\P{	/a/OTHER_NUMBER}');
Expect(1, 127242, '\p{othernumber}', "");
Expect(0, 127242, '\p{^othernumber}', "");
Expect(0, 127242, '\P{othernumber}', "");
Expect(1, 127242, '\P{^othernumber}', "");
Expect(0, 127243, '\p{othernumber}', "");
Expect(1, 127243, '\p{^othernumber}', "");
Expect(1, 127243, '\P{othernumber}', "");
Expect(0, 127243, '\P{^othernumber}', "");
Expect(1, 127242, '\p{__OTHER_Number}', "");
Expect(0, 127242, '\p{^__OTHER_Number}', "");
Expect(0, 127242, '\P{__OTHER_Number}', "");
Expect(1, 127242, '\P{^__OTHER_Number}', "");
Expect(0, 127243, '\p{__OTHER_Number}', "");
Expect(1, 127243, '\p{^__OTHER_Number}', "");
Expect(1, 127243, '\P{__OTHER_Number}', "");
Expect(0, 127243, '\P{^__OTHER_Number}', "");
Error('\p{:=-_IS_OTHER_Number}');
Error('\P{:=-_IS_OTHER_Number}');
Expect(1, 127242, '\p{isothernumber}', "");
Expect(0, 127242, '\p{^isothernumber}', "");
Expect(0, 127242, '\P{isothernumber}', "");
Expect(1, 127242, '\P{^isothernumber}', "");
Expect(0, 127243, '\p{isothernumber}', "");
Expect(1, 127243, '\p{^isothernumber}', "");
Expect(1, 127243, '\P{isothernumber}', "");
Expect(0, 127243, '\P{^isothernumber}', "");
Expect(1, 127242, '\p{ _IS_Other_Number}', "");
Expect(0, 127242, '\p{^ _IS_Other_Number}', "");
Expect(0, 127242, '\P{ _IS_Other_Number}', "");
Expect(1, 127242, '\P{^ _IS_Other_Number}', "");
Expect(0, 127243, '\p{ _IS_Other_Number}', "");
Expect(1, 127243, '\p{^ _IS_Other_Number}', "");
Expect(1, 127243, '\P{ _IS_Other_Number}', "");
Expect(0, 127243, '\P{^ _IS_Other_Number}', "");
Error('\p{_:=no}');
Error('\P{_:=no}');
Expect(1, 127242, '\p{no}', "");
Expect(0, 127242, '\p{^no}', "");
Expect(0, 127242, '\P{no}', "");
Expect(1, 127242, '\P{^no}', "");
Expect(0, 127243, '\p{no}', "");
Expect(1, 127243, '\p{^no}', "");
Expect(1, 127243, '\P{no}', "");
Expect(0, 127243, '\P{^no}', "");
Expect(1, 127242, '\p{-	No}', "");
Expect(0, 127242, '\p{^-	No}', "");
Expect(0, 127242, '\P{-	No}', "");
Expect(1, 127242, '\P{^-	No}', "");
Expect(0, 127243, '\p{-	No}', "");
Expect(1, 127243, '\p{^-	No}', "");
Expect(1, 127243, '\P{-	No}', "");
Expect(0, 127243, '\P{^-	No}', "");
Error('\p{:=_Is_no}');
Error('\P{:=_Is_no}');
Expect(1, 127242, '\p{isno}', "");
Expect(0, 127242, '\p{^isno}', "");
Expect(0, 127242, '\P{isno}', "");
Expect(1, 127242, '\P{^isno}', "");
Expect(0, 127243, '\p{isno}', "");
Expect(1, 127243, '\p{^isno}', "");
Expect(1, 127243, '\P{isno}', "");
Expect(0, 127243, '\P{^isno}', "");
Expect(1, 127242, '\p{_Is_No}', "");
Expect(0, 127242, '\p{^_Is_No}', "");
Expect(0, 127242, '\P{_Is_No}', "");
Expect(1, 127242, '\P{^_Is_No}', "");
Expect(0, 127243, '\p{_Is_No}', "");
Expect(1, 127243, '\p{^_Is_No}', "");
Expect(1, 127243, '\P{_Is_No}', "");
Expect(0, 127243, '\P{^_Is_No}', "");
Error('\p{/a/ Other_punctuation}');
Error('\P{/a/ Other_punctuation}');
Expect(1, 74867, '\p{otherpunctuation}', "");
Expect(0, 74867, '\p{^otherpunctuation}', "");
Expect(0, 74867, '\P{otherpunctuation}', "");
Expect(1, 74867, '\P{^otherpunctuation}', "");
Expect(0, 74868, '\p{otherpunctuation}', "");
Expect(1, 74868, '\p{^otherpunctuation}', "");
Expect(1, 74868, '\P{otherpunctuation}', "");
Expect(0, 74868, '\P{^otherpunctuation}', "");
Expect(1, 74867, '\p{ 	OTHER_PUNCTUATION}', "");
Expect(0, 74867, '\p{^ 	OTHER_PUNCTUATION}', "");
Expect(0, 74867, '\P{ 	OTHER_PUNCTUATION}', "");
Expect(1, 74867, '\P{^ 	OTHER_PUNCTUATION}', "");
Expect(0, 74868, '\p{ 	OTHER_PUNCTUATION}', "");
Expect(1, 74868, '\p{^ 	OTHER_PUNCTUATION}', "");
Expect(1, 74868, '\P{ 	OTHER_PUNCTUATION}', "");
Expect(0, 74868, '\P{^ 	OTHER_PUNCTUATION}', "");
Error('\p{/a/_Is_other_punctuation}');
Error('\P{/a/_Is_other_punctuation}');
Expect(1, 74867, '\p{isotherpunctuation}', "");
Expect(0, 74867, '\p{^isotherpunctuation}', "");
Expect(0, 74867, '\P{isotherpunctuation}', "");
Expect(1, 74867, '\P{^isotherpunctuation}', "");
Expect(0, 74868, '\p{isotherpunctuation}', "");
Expect(1, 74868, '\p{^isotherpunctuation}', "");
Expect(1, 74868, '\P{isotherpunctuation}', "");
Expect(0, 74868, '\P{^isotherpunctuation}', "");
Expect(1, 74867, '\p{-is_Other_PUNCTUATION}', "");
Expect(0, 74867, '\p{^-is_Other_PUNCTUATION}', "");
Expect(0, 74867, '\P{-is_Other_PUNCTUATION}', "");
Expect(1, 74867, '\P{^-is_Other_PUNCTUATION}', "");
Expect(0, 74868, '\p{-is_Other_PUNCTUATION}', "");
Expect(1, 74868, '\p{^-is_Other_PUNCTUATION}', "");
Expect(1, 74868, '\P{-is_Other_PUNCTUATION}', "");
Expect(0, 74868, '\P{^-is_Other_PUNCTUATION}', "");
Error('\p{/a/ 	Po}');
Error('\P{/a/ 	Po}');
Expect(1, 74867, '\p{po}', "");
Expect(0, 74867, '\p{^po}', "");
Expect(0, 74867, '\P{po}', "");
Expect(1, 74867, '\P{^po}', "");
Expect(0, 74868, '\p{po}', "");
Expect(1, 74868, '\p{^po}', "");
Expect(1, 74868, '\P{po}', "");
Expect(0, 74868, '\P{^po}', "");
Expect(1, 74867, '\p{	 Po}', "");
Expect(0, 74867, '\p{^	 Po}', "");
Expect(0, 74867, '\P{	 Po}', "");
Expect(1, 74867, '\P{^	 Po}', "");
Expect(0, 74868, '\p{	 Po}', "");
Expect(1, 74868, '\p{^	 Po}', "");
Expect(1, 74868, '\P{	 Po}', "");
Expect(0, 74868, '\P{^	 Po}', "");
Error('\p{- is_Po:=}');
Error('\P{- is_Po:=}');
Expect(1, 74867, '\p{ispo}', "");
Expect(0, 74867, '\p{^ispo}', "");
Expect(0, 74867, '\P{ispo}', "");
Expect(1, 74867, '\P{^ispo}', "");
Expect(0, 74868, '\p{ispo}', "");
Expect(1, 74868, '\p{^ispo}', "");
Expect(1, 74868, '\P{ispo}', "");
Expect(0, 74868, '\P{^ispo}', "");
Expect(1, 74867, '\p{_is_Po}', "");
Expect(0, 74867, '\p{^_is_Po}', "");
Expect(0, 74867, '\P{_is_Po}', "");
Expect(1, 74867, '\P{^_is_Po}', "");
Expect(0, 74868, '\p{_is_Po}', "");
Expect(1, 74868, '\p{^_is_Po}', "");
Expect(1, 74868, '\P{_is_Po}', "");
Expect(0, 74868, '\P{^_is_Po}', "");
Error('\p{:=- other_symbol}');
Error('\P{:=- other_symbol}');
Expect(1, 128883, '\p{othersymbol}', "");
Expect(0, 128883, '\p{^othersymbol}', "");
Expect(0, 128883, '\P{othersymbol}', "");
Expect(1, 128883, '\P{^othersymbol}', "");
Expect(0, 128884, '\p{othersymbol}', "");
Expect(1, 128884, '\p{^othersymbol}', "");
Expect(1, 128884, '\P{othersymbol}', "");
Expect(0, 128884, '\P{^othersymbol}', "");
Expect(1, 128883, '\p{-_Other_Symbol}', "");
Expect(0, 128883, '\p{^-_Other_Symbol}', "");
Expect(0, 128883, '\P{-_Other_Symbol}', "");
Expect(1, 128883, '\P{^-_Other_Symbol}', "");
Expect(0, 128884, '\p{-_Other_Symbol}', "");
Expect(1, 128884, '\p{^-_Other_Symbol}', "");
Expect(1, 128884, '\P{-_Other_Symbol}', "");
Expect(0, 128884, '\P{^-_Other_Symbol}', "");
Error('\p{:= -Is_Other_symbol}');
Error('\P{:= -Is_Other_symbol}');
Expect(1, 128883, '\p{isothersymbol}', "");
Expect(0, 128883, '\p{^isothersymbol}', "");
Expect(0, 128883, '\P{isothersymbol}', "");
Expect(1, 128883, '\P{^isothersymbol}', "");
Expect(0, 128884, '\p{isothersymbol}', "");
Expect(1, 128884, '\p{^isothersymbol}', "");
Expect(1, 128884, '\P{isothersymbol}', "");
Expect(0, 128884, '\P{^isothersymbol}', "");
Expect(1, 128883, '\p{	_Is_Other_Symbol}', "");
Expect(0, 128883, '\p{^	_Is_Other_Symbol}', "");
Expect(0, 128883, '\P{	_Is_Other_Symbol}', "");
Expect(1, 128883, '\P{^	_Is_Other_Symbol}', "");
Expect(0, 128884, '\p{	_Is_Other_Symbol}', "");
Expect(1, 128884, '\p{^	_Is_Other_Symbol}', "");
Expect(1, 128884, '\P{	_Is_Other_Symbol}', "");
Expect(0, 128884, '\P{^	_Is_Other_Symbol}', "");
Error('\p{/a/  So}');
Error('\P{/a/  So}');
Expect(1, 128883, '\p{so}', "");
Expect(0, 128883, '\p{^so}', "");
Expect(0, 128883, '\P{so}', "");
Expect(1, 128883, '\P{^so}', "");
Expect(0, 128884, '\p{so}', "");
Expect(1, 128884, '\p{^so}', "");
Expect(1, 128884, '\P{so}', "");
Expect(0, 128884, '\P{^so}', "");
Expect(1, 128883, '\p{	_So}', "");
Expect(0, 128883, '\p{^	_So}', "");
Expect(0, 128883, '\P{	_So}', "");
Expect(1, 128883, '\P{^	_So}', "");
Expect(0, 128884, '\p{	_So}', "");
Expect(1, 128884, '\p{^	_So}', "");
Expect(1, 128884, '\P{	_So}', "");
Expect(0, 128884, '\P{^	_So}', "");
Error('\p{/a/	_is_So}');
Error('\P{/a/	_is_So}');
Expect(1, 128883, '\p{isso}', "");
Expect(0, 128883, '\p{^isso}', "");
Expect(0, 128883, '\P{isso}', "");
Expect(1, 128883, '\P{^isso}', "");
Expect(0, 128884, '\p{isso}', "");
Expect(1, 128884, '\p{^isso}', "");
Expect(1, 128884, '\P{isso}', "");
Expect(0, 128884, '\P{^isso}', "");
Expect(1, 128883, '\p{-_Is_So}', "");
Expect(0, 128883, '\p{^-_Is_So}', "");
Expect(0, 128883, '\P{-_Is_So}', "");
Expect(1, 128883, '\P{^-_Is_So}', "");
Expect(0, 128884, '\p{-_Is_So}', "");
Expect(1, 128884, '\p{^-_Is_So}', "");
Expect(1, 128884, '\P{-_Is_So}', "");
Expect(0, 128884, '\P{^-_Is_So}', "");
Error('\p{Other_Uppercase}');
Error('\P{Other_Uppercase}');
Error('\p{Is_Other_Uppercase}');
Error('\P{Is_Other_Uppercase}');
Error('\p{OUpper}');
Error('\P{OUpper}');
Error('\p{Is_OUpper}');
Error('\P{Is_OUpper}');
Error('\p{_/a/PARAGRAPH_SEPARATOR}');
Error('\P{_/a/PARAGRAPH_SEPARATOR}');
Expect(1, 8233, '\p{paragraphseparator}', "");
Expect(0, 8233, '\p{^paragraphseparator}', "");
Expect(0, 8233, '\P{paragraphseparator}', "");
Expect(1, 8233, '\P{^paragraphseparator}', "");
Expect(0, 8234, '\p{paragraphseparator}', "");
Expect(1, 8234, '\p{^paragraphseparator}', "");
Expect(1, 8234, '\P{paragraphseparator}', "");
Expect(0, 8234, '\P{^paragraphseparator}', "");
Expect(1, 8233, '\p{ 	Paragraph_Separator}', "");
Expect(0, 8233, '\p{^ 	Paragraph_Separator}', "");
Expect(0, 8233, '\P{ 	Paragraph_Separator}', "");
Expect(1, 8233, '\P{^ 	Paragraph_Separator}', "");
Expect(0, 8234, '\p{ 	Paragraph_Separator}', "");
Expect(1, 8234, '\p{^ 	Paragraph_Separator}', "");
Expect(1, 8234, '\P{ 	Paragraph_Separator}', "");
Expect(0, 8234, '\P{^ 	Paragraph_Separator}', "");
Error('\p{:=	IS_Paragraph_separator}');
Error('\P{:=	IS_Paragraph_separator}');
Expect(1, 8233, '\p{isparagraphseparator}', "");
Expect(0, 8233, '\p{^isparagraphseparator}', "");
Expect(0, 8233, '\P{isparagraphseparator}', "");
Expect(1, 8233, '\P{^isparagraphseparator}', "");
Expect(0, 8234, '\p{isparagraphseparator}', "");
Expect(1, 8234, '\p{^isparagraphseparator}', "");
Expect(1, 8234, '\P{isparagraphseparator}', "");
Expect(0, 8234, '\P{^isparagraphseparator}', "");
Expect(1, 8233, '\p{	Is_PARAGRAPH_SEPARATOR}', "");
Expect(0, 8233, '\p{^	Is_PARAGRAPH_SEPARATOR}', "");
Expect(0, 8233, '\P{	Is_PARAGRAPH_SEPARATOR}', "");
Expect(1, 8233, '\P{^	Is_PARAGRAPH_SEPARATOR}', "");
Expect(0, 8234, '\p{	Is_PARAGRAPH_SEPARATOR}', "");
Expect(1, 8234, '\p{^	Is_PARAGRAPH_SEPARATOR}', "");
Expect(1, 8234, '\P{	Is_PARAGRAPH_SEPARATOR}', "");
Expect(0, 8234, '\P{^	Is_PARAGRAPH_SEPARATOR}', "");
Error('\p{-Zp:=}');
Error('\P{-Zp:=}');
Expect(1, 8233, '\p{zp}', "");
Expect(0, 8233, '\p{^zp}', "");
Expect(0, 8233, '\P{zp}', "");
Expect(1, 8233, '\P{^zp}', "");
Expect(0, 8234, '\p{zp}', "");
Expect(1, 8234, '\p{^zp}', "");
Expect(1, 8234, '\P{zp}', "");
Expect(0, 8234, '\P{^zp}', "");
Expect(1, 8233, '\p{-zp}', "");
Expect(0, 8233, '\p{^-zp}', "");
Expect(0, 8233, '\P{-zp}', "");
Expect(1, 8233, '\P{^-zp}', "");
Expect(0, 8234, '\p{-zp}', "");
Expect(1, 8234, '\p{^-zp}', "");
Expect(1, 8234, '\P{-zp}', "");
Expect(0, 8234, '\P{^-zp}', "");
Error('\p{/a/ IS_zp}');
Error('\P{/a/ IS_zp}');
Expect(1, 8233, '\p{iszp}', "");
Expect(0, 8233, '\p{^iszp}', "");
Expect(0, 8233, '\P{iszp}', "");
Expect(1, 8233, '\P{^iszp}', "");
Expect(0, 8234, '\p{iszp}', "");
Expect(1, 8234, '\p{^iszp}', "");
Expect(1, 8234, '\P{iszp}', "");
Expect(0, 8234, '\P{^iszp}', "");
Expect(1, 8233, '\p{ 	IS_zp}', "");
Expect(0, 8233, '\p{^ 	IS_zp}', "");
Expect(0, 8233, '\P{ 	IS_zp}', "");
Expect(1, 8233, '\P{^ 	IS_zp}', "");
Expect(0, 8234, '\p{ 	IS_zp}', "");
Expect(1, 8234, '\p{^ 	IS_zp}', "");
Expect(1, 8234, '\P{ 	IS_zp}', "");
Expect(0, 8234, '\P{^ 	IS_zp}', "");
Error('\p{_	pattern_syntax/a/}');
Error('\P{_	pattern_syntax/a/}');
Expect(1, 65094, '\p{patternsyntax}', "");
Expect(0, 65094, '\p{^patternsyntax}', "");
Expect(0, 65094, '\P{patternsyntax}', "");
Expect(1, 65094, '\P{^patternsyntax}', "");
Expect(0, 65095, '\p{patternsyntax}', "");
Expect(1, 65095, '\p{^patternsyntax}', "");
Expect(1, 65095, '\P{patternsyntax}', "");
Expect(0, 65095, '\P{^patternsyntax}', "");
Expect(1, 65094, '\p{_ Pattern_syntax}', "");
Expect(0, 65094, '\p{^_ Pattern_syntax}', "");
Expect(0, 65094, '\P{_ Pattern_syntax}', "");
Expect(1, 65094, '\P{^_ Pattern_syntax}', "");
Expect(0, 65095, '\p{_ Pattern_syntax}', "");
Expect(1, 65095, '\p{^_ Pattern_syntax}', "");
Expect(1, 65095, '\P{_ Pattern_syntax}', "");
Expect(0, 65095, '\P{^_ Pattern_syntax}', "");
Error('\p{_is_PATTERN_Syntax:=}');
Error('\P{_is_PATTERN_Syntax:=}');
Expect(1, 65094, '\p{ispatternsyntax}', "");
Expect(0, 65094, '\p{^ispatternsyntax}', "");
Expect(0, 65094, '\P{ispatternsyntax}', "");
Expect(1, 65094, '\P{^ispatternsyntax}', "");
Expect(0, 65095, '\p{ispatternsyntax}', "");
Expect(1, 65095, '\p{^ispatternsyntax}', "");
Expect(1, 65095, '\P{ispatternsyntax}', "");
Expect(0, 65095, '\P{^ispatternsyntax}', "");
Expect(1, 65094, '\p{ _IS_PATTERN_syntax}', "");
Expect(0, 65094, '\p{^ _IS_PATTERN_syntax}', "");
Expect(0, 65094, '\P{ _IS_PATTERN_syntax}', "");
Expect(1, 65094, '\P{^ _IS_PATTERN_syntax}', "");
Expect(0, 65095, '\p{ _IS_PATTERN_syntax}', "");
Expect(1, 65095, '\p{^ _IS_PATTERN_syntax}', "");
Expect(1, 65095, '\P{ _IS_PATTERN_syntax}', "");
Expect(0, 65095, '\P{^ _IS_PATTERN_syntax}', "");
Error('\p{--PAT_Syn/a/}');
Error('\P{--PAT_Syn/a/}');
Expect(1, 65094, '\p{patsyn}', "");
Expect(0, 65094, '\p{^patsyn}', "");
Expect(0, 65094, '\P{patsyn}', "");
Expect(1, 65094, '\P{^patsyn}', "");
Expect(0, 65095, '\p{patsyn}', "");
Expect(1, 65095, '\p{^patsyn}', "");
Expect(1, 65095, '\P{patsyn}', "");
Expect(0, 65095, '\P{^patsyn}', "");
Expect(1, 65094, '\p{ pat_SYN}', "");
Expect(0, 65094, '\p{^ pat_SYN}', "");
Expect(0, 65094, '\P{ pat_SYN}', "");
Expect(1, 65094, '\P{^ pat_SYN}', "");
Expect(0, 65095, '\p{ pat_SYN}', "");
Expect(1, 65095, '\p{^ pat_SYN}', "");
Expect(1, 65095, '\P{ pat_SYN}', "");
Expect(0, 65095, '\P{^ pat_SYN}', "");
Error('\p{/a/	 IS_PAT_SYN}');
Error('\P{/a/	 IS_PAT_SYN}');
Expect(1, 65094, '\p{ispatsyn}', "");
Expect(0, 65094, '\p{^ispatsyn}', "");
Expect(0, 65094, '\P{ispatsyn}', "");
Expect(1, 65094, '\P{^ispatsyn}', "");
Expect(0, 65095, '\p{ispatsyn}', "");
Expect(1, 65095, '\p{^ispatsyn}', "");
Expect(1, 65095, '\P{ispatsyn}', "");
Expect(0, 65095, '\P{^ispatsyn}', "");
Expect(1, 65094, '\p{ -Is_pat_syn}', "");
Expect(0, 65094, '\p{^ -Is_pat_syn}', "");
Expect(0, 65094, '\P{ -Is_pat_syn}', "");
Expect(1, 65094, '\P{^ -Is_pat_syn}', "");
Expect(0, 65095, '\p{ -Is_pat_syn}', "");
Expect(1, 65095, '\p{^ -Is_pat_syn}', "");
Expect(1, 65095, '\P{ -Is_pat_syn}', "");
Expect(0, 65095, '\P{^ -Is_pat_syn}', "");
Error('\p{	PATTERN_White_Space:=}');
Error('\P{	PATTERN_White_Space:=}');
Expect(1, 8233, '\p{patternwhitespace}', "");
Expect(0, 8233, '\p{^patternwhitespace}', "");
Expect(0, 8233, '\P{patternwhitespace}', "");
Expect(1, 8233, '\P{^patternwhitespace}', "");
Expect(0, 8234, '\p{patternwhitespace}', "");
Expect(1, 8234, '\p{^patternwhitespace}', "");
Expect(1, 8234, '\P{patternwhitespace}', "");
Expect(0, 8234, '\P{^patternwhitespace}', "");
Expect(1, 8233, '\p{-Pattern_White_space}', "");
Expect(0, 8233, '\p{^-Pattern_White_space}', "");
Expect(0, 8233, '\P{-Pattern_White_space}', "");
Expect(1, 8233, '\P{^-Pattern_White_space}', "");
Expect(0, 8234, '\p{-Pattern_White_space}', "");
Expect(1, 8234, '\p{^-Pattern_White_space}', "");
Expect(1, 8234, '\P{-Pattern_White_space}', "");
Expect(0, 8234, '\P{^-Pattern_White_space}', "");
Error('\p{Is_Pattern_White_SPACE/a/}');
Error('\P{Is_Pattern_White_SPACE/a/}');
Expect(1, 8233, '\p{ispatternwhitespace}', "");
Expect(0, 8233, '\p{^ispatternwhitespace}', "");
Expect(0, 8233, '\P{ispatternwhitespace}', "");
Expect(1, 8233, '\P{^ispatternwhitespace}', "");
Expect(0, 8234, '\p{ispatternwhitespace}', "");
Expect(1, 8234, '\p{^ispatternwhitespace}', "");
Expect(1, 8234, '\P{ispatternwhitespace}', "");
Expect(0, 8234, '\P{^ispatternwhitespace}', "");
Expect(1, 8233, '\p{ 	is_PATTERN_WHITE_space}', "");
Expect(0, 8233, '\p{^ 	is_PATTERN_WHITE_space}', "");
Expect(0, 8233, '\P{ 	is_PATTERN_WHITE_space}', "");
Expect(1, 8233, '\P{^ 	is_PATTERN_WHITE_space}', "");
Expect(0, 8234, '\p{ 	is_PATTERN_WHITE_space}', "");
Expect(1, 8234, '\p{^ 	is_PATTERN_WHITE_space}', "");
Expect(1, 8234, '\P{ 	is_PATTERN_WHITE_space}', "");
Expect(0, 8234, '\P{^ 	is_PATTERN_WHITE_space}', "");
Error('\p{_:=Pat_WS}');
Error('\P{_:=Pat_WS}');
Expect(1, 8233, '\p{patws}', "");
Expect(0, 8233, '\p{^patws}', "");
Expect(0, 8233, '\P{patws}', "");
Expect(1, 8233, '\P{^patws}', "");
Expect(0, 8234, '\p{patws}', "");
Expect(1, 8234, '\p{^patws}', "");
Expect(1, 8234, '\P{patws}', "");
Expect(0, 8234, '\P{^patws}', "");
Expect(1, 8233, '\p{- pat_WS}', "");
Expect(0, 8233, '\p{^- pat_WS}', "");
Expect(0, 8233, '\P{- pat_WS}', "");
Expect(1, 8233, '\P{^- pat_WS}', "");
Expect(0, 8234, '\p{- pat_WS}', "");
Expect(1, 8234, '\p{^- pat_WS}', "");
Expect(1, 8234, '\P{- pat_WS}', "");
Expect(0, 8234, '\P{^- pat_WS}', "");
Error('\p{_:=IS_PAT_WS}');
Error('\P{_:=IS_PAT_WS}');
Expect(1, 8233, '\p{ispatws}', "");
Expect(0, 8233, '\p{^ispatws}', "");
Expect(0, 8233, '\P{ispatws}', "");
Expect(1, 8233, '\P{^ispatws}', "");
Expect(0, 8234, '\p{ispatws}', "");
Expect(1, 8234, '\p{^ispatws}', "");
Expect(1, 8234, '\P{ispatws}', "");
Expect(0, 8234, '\P{^ispatws}', "");
Expect(1, 8233, '\p{_-is_Pat_WS}', "");
Expect(0, 8233, '\p{^_-is_Pat_WS}', "");
Expect(0, 8233, '\P{_-is_Pat_WS}', "");
Expect(1, 8233, '\P{^_-is_Pat_WS}', "");
Expect(0, 8234, '\p{_-is_Pat_WS}', "");
Expect(1, 8234, '\p{^_-is_Pat_WS}', "");
Expect(1, 8234, '\P{_-is_Pat_WS}', "");
Expect(0, 8234, '\P{^_-is_Pat_WS}', "");
Error('\p{:=-	perlspace}');
Error('\P{:=-	perlspace}');
Expect(1, 32, '\p{perlspace}', "");
Expect(0, 32, '\p{^perlspace}', "");
Expect(0, 32, '\P{perlspace}', "");
Expect(1, 32, '\P{^perlspace}', "");
Expect(0, 33, '\p{perlspace}', "");
Expect(1, 33, '\p{^perlspace}', "");
Expect(1, 33, '\P{perlspace}', "");
Expect(0, 33, '\P{^perlspace}', "");
Expect(1, 32, '\p{		PerlSpace}', "");
Expect(0, 32, '\p{^		PerlSpace}', "");
Expect(0, 32, '\P{		PerlSpace}', "");
Expect(1, 32, '\P{^		PerlSpace}', "");
Expect(0, 33, '\p{		PerlSpace}', "");
Expect(1, 33, '\p{^		PerlSpace}', "");
Expect(1, 33, '\P{		PerlSpace}', "");
Expect(0, 33, '\P{^		PerlSpace}', "");
Error('\p{ is_PerlSpace:=}');
Error('\P{ is_PerlSpace:=}');
Expect(1, 32, '\p{isperlspace}', "");
Expect(0, 32, '\p{^isperlspace}', "");
Expect(0, 32, '\P{isperlspace}', "");
Expect(1, 32, '\P{^isperlspace}', "");
Expect(0, 33, '\p{isperlspace}', "");
Expect(1, 33, '\p{^isperlspace}', "");
Expect(1, 33, '\P{isperlspace}', "");
Expect(0, 33, '\P{^isperlspace}', "");
Expect(1, 32, '\p{	is_PERLSPACE}', "");
Expect(0, 32, '\p{^	is_PERLSPACE}', "");
Expect(0, 32, '\P{	is_PERLSPACE}', "");
Expect(1, 32, '\P{^	is_PERLSPACE}', "");
Expect(0, 33, '\p{	is_PERLSPACE}', "");
Expect(1, 33, '\p{^	is_PERLSPACE}', "");
Expect(1, 33, '\P{	is_PERLSPACE}', "");
Expect(0, 33, '\P{^	is_PERLSPACE}', "");
Error('\p{_:=PerlWord}');
Error('\P{_:=PerlWord}');
Expect(1, 122, '\p{perlword}', "");
Expect(0, 122, '\p{^perlword}', "");
Expect(0, 122, '\P{perlword}', "");
Expect(1, 122, '\P{^perlword}', "");
Expect(0, 123, '\p{perlword}', "");
Expect(1, 123, '\p{^perlword}', "");
Expect(1, 123, '\P{perlword}', "");
Expect(0, 123, '\P{^perlword}', "");
Expect(1, 122, '\p{-PERLWORD}', "");
Expect(0, 122, '\p{^-PERLWORD}', "");
Expect(0, 122, '\P{-PERLWORD}', "");
Expect(1, 122, '\P{^-PERLWORD}', "");
Expect(0, 123, '\p{-PERLWORD}', "");
Expect(1, 123, '\p{^-PERLWORD}', "");
Expect(1, 123, '\P{-PERLWORD}', "");
Expect(0, 123, '\P{^-PERLWORD}', "");
Error('\p{ PosixWord:=}');
Error('\P{ PosixWord:=}');
Expect(1, 122, '\p{posixword}', "");
Expect(0, 122, '\p{^posixword}', "");
Expect(0, 122, '\P{posixword}', "");
Expect(1, 122, '\P{^posixword}', "");
Expect(0, 123, '\p{posixword}', "");
Expect(1, 123, '\p{^posixword}', "");
Expect(1, 123, '\P{posixword}', "");
Expect(0, 123, '\P{^posixword}', "");
Expect(1, 122, '\p{  posixword}', "");
Expect(0, 122, '\p{^  posixword}', "");
Expect(0, 122, '\P{  posixword}', "");
Expect(1, 122, '\P{^  posixword}', "");
Expect(0, 123, '\p{  posixword}', "");
Expect(1, 123, '\p{^  posixword}', "");
Expect(1, 123, '\P{  posixword}', "");
Expect(0, 123, '\P{^  posixword}', "");
Error('\p{/a/__Is_PerlWord}');
Error('\P{/a/__Is_PerlWord}');
Expect(1, 122, '\p{isperlword}', "");
Expect(0, 122, '\p{^isperlword}', "");
Expect(0, 122, '\P{isperlword}', "");
Expect(1, 122, '\P{^isperlword}', "");
Expect(0, 123, '\p{isperlword}', "");
Expect(1, 123, '\p{^isperlword}', "");
Expect(1, 123, '\P{isperlword}', "");
Expect(0, 123, '\P{^isperlword}', "");
Expect(1, 122, '\p{-Is_PerlWord}', "");
Expect(0, 122, '\p{^-Is_PerlWord}', "");
Expect(0, 122, '\P{-Is_PerlWord}', "");
Expect(1, 122, '\P{^-Is_PerlWord}', "");
Expect(0, 123, '\p{-Is_PerlWord}', "");
Expect(1, 123, '\p{^-Is_PerlWord}', "");
Expect(1, 123, '\P{-Is_PerlWord}', "");
Expect(0, 123, '\P{^-Is_PerlWord}', "");
Error('\p{_:=Is_PosixWord}');
Error('\P{_:=Is_PosixWord}');
Expect(1, 122, '\p{isposixword}', "");
Expect(0, 122, '\p{^isposixword}', "");
Expect(0, 122, '\P{isposixword}', "");
Expect(1, 122, '\P{^isposixword}', "");
Expect(0, 123, '\p{isposixword}', "");
Expect(1, 123, '\p{^isposixword}', "");
Expect(1, 123, '\P{isposixword}', "");
Expect(0, 123, '\P{^isposixword}', "");
Expect(1, 122, '\p{	is_posixword}', "");
Expect(0, 122, '\p{^	is_posixword}', "");
Expect(0, 122, '\P{	is_posixword}', "");
Expect(1, 122, '\P{^	is_posixword}', "");
Expect(0, 123, '\p{	is_posixword}', "");
Expect(1, 123, '\p{^	is_posixword}', "");
Expect(1, 123, '\P{	is_posixword}', "");
Expect(0, 123, '\P{^	is_posixword}', "");
Error('\p{:=	-phags_Pa}');
Error('\P{:=	-phags_Pa}');
Expect(1, 43127, '\p{phagspa}', "");
Expect(0, 43127, '\p{^phagspa}', "");
Expect(0, 43127, '\P{phagspa}', "");
Expect(1, 43127, '\P{^phagspa}', "");
Expect(0, 43128, '\p{phagspa}', "");
Expect(1, 43128, '\p{^phagspa}', "");
Expect(1, 43128, '\P{phagspa}', "");
Expect(0, 43128, '\P{^phagspa}', "");
Expect(1, 43127, '\p{_ Phags_pa}', "");
Expect(0, 43127, '\p{^_ Phags_pa}', "");
Expect(0, 43127, '\P{_ Phags_pa}', "");
Expect(1, 43127, '\P{^_ Phags_pa}', "");
Expect(0, 43128, '\p{_ Phags_pa}', "");
Expect(1, 43128, '\p{^_ Phags_pa}', "");
Expect(1, 43128, '\P{_ Phags_pa}', "");
Expect(0, 43128, '\P{^_ Phags_pa}', "");
Error('\p{/a/  Is_PHAGS_Pa}');
Error('\P{/a/  Is_PHAGS_Pa}');
Expect(1, 43127, '\p{isphagspa}', "");
Expect(0, 43127, '\p{^isphagspa}', "");
Expect(0, 43127, '\P{isphagspa}', "");
Expect(1, 43127, '\P{^isphagspa}', "");
Expect(0, 43128, '\p{isphagspa}', "");
Expect(1, 43128, '\p{^isphagspa}', "");
Expect(1, 43128, '\P{isphagspa}', "");
Expect(0, 43128, '\P{^isphagspa}', "");
Expect(1, 43127, '\p{-Is_phags_PA}', "");
Expect(0, 43127, '\p{^-Is_phags_PA}', "");
Expect(0, 43127, '\P{-Is_phags_PA}', "");
Expect(1, 43127, '\P{^-Is_phags_PA}', "");
Expect(0, 43128, '\p{-Is_phags_PA}', "");
Expect(1, 43128, '\p{^-Is_phags_PA}', "");
Expect(1, 43128, '\P{-Is_phags_PA}', "");
Expect(0, 43128, '\P{^-Is_phags_PA}', "");
Error('\p{Phag/a/}');
Error('\P{Phag/a/}');
Expect(1, 43127, '\p{phag}', "");
Expect(0, 43127, '\p{^phag}', "");
Expect(0, 43127, '\P{phag}', "");
Expect(1, 43127, '\P{^phag}', "");
Expect(0, 43128, '\p{phag}', "");
Expect(1, 43128, '\p{^phag}', "");
Expect(1, 43128, '\P{phag}', "");
Expect(0, 43128, '\P{^phag}', "");
Expect(1, 43127, '\p{-Phag}', "");
Expect(0, 43127, '\p{^-Phag}', "");
Expect(0, 43127, '\P{-Phag}', "");
Expect(1, 43127, '\P{^-Phag}', "");
Expect(0, 43128, '\p{-Phag}', "");
Expect(1, 43128, '\p{^-Phag}', "");
Expect(1, 43128, '\P{-Phag}', "");
Expect(0, 43128, '\P{^-Phag}', "");
Error('\p{ /a/Is_PHAG}');
Error('\P{ /a/Is_PHAG}');
Expect(1, 43127, '\p{isphag}', "");
Expect(0, 43127, '\p{^isphag}', "");
Expect(0, 43127, '\P{isphag}', "");
Expect(1, 43127, '\P{^isphag}', "");
Expect(0, 43128, '\p{isphag}', "");
Expect(1, 43128, '\p{^isphag}', "");
Expect(1, 43128, '\P{isphag}', "");
Expect(0, 43128, '\P{^isphag}', "");
Expect(1, 43127, '\p{--IS_Phag}', "");
Expect(0, 43127, '\p{^--IS_Phag}', "");
Expect(0, 43127, '\P{--IS_Phag}', "");
Expect(1, 43127, '\P{^--IS_Phag}', "");
Expect(0, 43128, '\p{--IS_Phag}', "");
Expect(1, 43128, '\p{^--IS_Phag}', "");
Expect(1, 43128, '\P{--IS_Phag}', "");
Expect(0, 43128, '\P{^--IS_Phag}', "");
Error('\p{-:=Phaistos_disc}');
Error('\P{-:=Phaistos_disc}');
Expect(1, 66047, '\p{phaistosdisc}', "");
Expect(0, 66047, '\p{^phaistosdisc}', "");
Expect(0, 66047, '\P{phaistosdisc}', "");
Expect(1, 66047, '\P{^phaistosdisc}', "");
Expect(0, 66048, '\p{phaistosdisc}', "");
Expect(1, 66048, '\p{^phaistosdisc}', "");
Expect(1, 66048, '\P{phaistosdisc}', "");
Expect(0, 66048, '\P{^phaistosdisc}', "");
Expect(1, 66047, '\p{_PHAISTOS_disc}', "");
Expect(0, 66047, '\p{^_PHAISTOS_disc}', "");
Expect(0, 66047, '\P{_PHAISTOS_disc}', "");
Expect(1, 66047, '\P{^_PHAISTOS_disc}', "");
Expect(0, 66048, '\p{_PHAISTOS_disc}', "");
Expect(1, 66048, '\p{^_PHAISTOS_disc}', "");
Expect(1, 66048, '\P{_PHAISTOS_disc}', "");
Expect(0, 66048, '\P{^_PHAISTOS_disc}', "");
Error('\p{/a/-	Is_Phaistos_disc}');
Error('\P{/a/-	Is_Phaistos_disc}');
Expect(1, 66047, '\p{isphaistosdisc}', "");
Expect(0, 66047, '\p{^isphaistosdisc}', "");
Expect(0, 66047, '\P{isphaistosdisc}', "");
Expect(1, 66047, '\P{^isphaistosdisc}', "");
Expect(0, 66048, '\p{isphaistosdisc}', "");
Expect(1, 66048, '\p{^isphaistosdisc}', "");
Expect(1, 66048, '\P{isphaistosdisc}', "");
Expect(0, 66048, '\P{^isphaistosdisc}', "");
Expect(1, 66047, '\p{-is_Phaistos_Disc}', "");
Expect(0, 66047, '\p{^-is_Phaistos_Disc}', "");
Expect(0, 66047, '\P{-is_Phaistos_Disc}', "");
Expect(1, 66047, '\P{^-is_Phaistos_Disc}', "");
Expect(0, 66048, '\p{-is_Phaistos_Disc}', "");
Expect(1, 66048, '\p{^-is_Phaistos_Disc}', "");
Expect(1, 66048, '\P{-is_Phaistos_Disc}', "");
Expect(0, 66048, '\P{^-is_Phaistos_Disc}', "");
Error('\p{-/a/IN_Phaistos_disc}');
Error('\P{-/a/IN_Phaistos_disc}');
Expect(1, 66047, '\p{inphaistosdisc}', "");
Expect(0, 66047, '\p{^inphaistosdisc}', "");
Expect(0, 66047, '\P{inphaistosdisc}', "");
Expect(1, 66047, '\P{^inphaistosdisc}', "");
Expect(0, 66048, '\p{inphaistosdisc}', "");
Expect(1, 66048, '\p{^inphaistosdisc}', "");
Expect(1, 66048, '\P{inphaistosdisc}', "");
Expect(0, 66048, '\P{^inphaistosdisc}', "");
Expect(1, 66047, '\p{	 in_Phaistos_Disc}', "");
Expect(0, 66047, '\p{^	 in_Phaistos_Disc}', "");
Expect(0, 66047, '\P{	 in_Phaistos_Disc}', "");
Expect(1, 66047, '\P{^	 in_Phaistos_Disc}', "");
Expect(0, 66048, '\p{	 in_Phaistos_Disc}', "");
Expect(1, 66048, '\p{^	 in_Phaistos_Disc}', "");
Expect(1, 66048, '\P{	 in_Phaistos_Disc}', "");
Expect(0, 66048, '\P{^	 in_Phaistos_Disc}', "");
Error('\p{/a/Phaistos}');
Error('\P{/a/Phaistos}');
Expect(1, 66047, '\p{phaistos}', "");
Expect(0, 66047, '\p{^phaistos}', "");
Expect(0, 66047, '\P{phaistos}', "");
Expect(1, 66047, '\P{^phaistos}', "");
Expect(0, 66048, '\p{phaistos}', "");
Expect(1, 66048, '\p{^phaistos}', "");
Expect(1, 66048, '\P{phaistos}', "");
Expect(0, 66048, '\P{^phaistos}', "");
Expect(1, 66047, '\p{	-phaistos}', "");
Expect(0, 66047, '\p{^	-phaistos}', "");
Expect(0, 66047, '\P{	-phaistos}', "");
Expect(1, 66047, '\P{^	-phaistos}', "");
Expect(0, 66048, '\p{	-phaistos}', "");
Expect(1, 66048, '\p{^	-phaistos}', "");
Expect(1, 66048, '\P{	-phaistos}', "");
Expect(0, 66048, '\P{^	-phaistos}', "");
Error('\p{/a/  Is_PHAISTOS}');
Error('\P{/a/  Is_PHAISTOS}');
Expect(1, 66047, '\p{isphaistos}', "");
Expect(0, 66047, '\p{^isphaistos}', "");
Expect(0, 66047, '\P{isphaistos}', "");
Expect(1, 66047, '\P{^isphaistos}', "");
Expect(0, 66048, '\p{isphaistos}', "");
Expect(1, 66048, '\p{^isphaistos}', "");
Expect(1, 66048, '\P{isphaistos}', "");
Expect(0, 66048, '\P{^isphaistos}', "");
Expect(1, 66047, '\p{-Is_phaistos}', "");
Expect(0, 66047, '\p{^-Is_phaistos}', "");
Expect(0, 66047, '\P{-Is_phaistos}', "");
Expect(1, 66047, '\P{^-Is_phaistos}', "");
Expect(0, 66048, '\p{-Is_phaistos}', "");
Expect(1, 66048, '\p{^-Is_phaistos}', "");
Expect(1, 66048, '\P{-Is_phaistos}', "");
Expect(0, 66048, '\P{^-Is_phaistos}', "");
Error('\p{ /a/In_Phaistos}');
Error('\P{ /a/In_Phaistos}');
Expect(1, 66047, '\p{inphaistos}', "");
Expect(0, 66047, '\p{^inphaistos}', "");
Expect(0, 66047, '\P{inphaistos}', "");
Expect(1, 66047, '\P{^inphaistos}', "");
Expect(0, 66048, '\p{inphaistos}', "");
Expect(1, 66048, '\p{^inphaistos}', "");
Expect(1, 66048, '\P{inphaistos}', "");
Expect(0, 66048, '\P{^inphaistos}', "");
Expect(1, 66047, '\p{	IN_Phaistos}', "");
Expect(0, 66047, '\p{^	IN_Phaistos}', "");
Expect(0, 66047, '\P{	IN_Phaistos}', "");
Expect(1, 66047, '\P{^	IN_Phaistos}', "");
Expect(0, 66048, '\p{	IN_Phaistos}', "");
Expect(1, 66048, '\p{^	IN_Phaistos}', "");
Expect(1, 66048, '\P{	IN_Phaistos}', "");
Expect(0, 66048, '\P{^	IN_Phaistos}', "");
Error('\p{	/a/Phoenician}');
Error('\P{	/a/Phoenician}');
Expect(1, 67871, '\p{phoenician}', "");
Expect(0, 67871, '\p{^phoenician}', "");
Expect(0, 67871, '\P{phoenician}', "");
Expect(1, 67871, '\P{^phoenician}', "");
Expect(0, 67872, '\p{phoenician}', "");
Expect(1, 67872, '\p{^phoenician}', "");
Expect(1, 67872, '\P{phoenician}', "");
Expect(0, 67872, '\P{^phoenician}', "");
Expect(1, 67871, '\p{_ Phoenician}', "");
Expect(0, 67871, '\p{^_ Phoenician}', "");
Expect(0, 67871, '\P{_ Phoenician}', "");
Expect(1, 67871, '\P{^_ Phoenician}', "");
Expect(0, 67872, '\p{_ Phoenician}', "");
Expect(1, 67872, '\p{^_ Phoenician}', "");
Expect(1, 67872, '\P{_ Phoenician}', "");
Expect(0, 67872, '\P{^_ Phoenician}', "");
Error('\p{/a/ -Is_Phoenician}');
Error('\P{/a/ -Is_Phoenician}');
Expect(1, 67871, '\p{isphoenician}', "");
Expect(0, 67871, '\p{^isphoenician}', "");
Expect(0, 67871, '\P{isphoenician}', "");
Expect(1, 67871, '\P{^isphoenician}', "");
Expect(0, 67872, '\p{isphoenician}', "");
Expect(1, 67872, '\p{^isphoenician}', "");
Expect(1, 67872, '\P{isphoenician}', "");
Expect(0, 67872, '\P{^isphoenician}', "");
Expect(1, 67871, '\p{  Is_phoenician}', "");
Expect(0, 67871, '\p{^  Is_phoenician}', "");
Expect(0, 67871, '\P{  Is_phoenician}', "");
Expect(1, 67871, '\P{^  Is_phoenician}', "");
Expect(0, 67872, '\p{  Is_phoenician}', "");
Expect(1, 67872, '\p{^  Is_phoenician}', "");
Expect(1, 67872, '\P{  Is_phoenician}', "");
Expect(0, 67872, '\P{^  Is_phoenician}', "");
Error('\p{  PHNX/a/}');
Error('\P{  PHNX/a/}');
Expect(1, 67871, '\p{phnx}', "");
Expect(0, 67871, '\p{^phnx}', "");
Expect(0, 67871, '\P{phnx}', "");
Expect(1, 67871, '\P{^phnx}', "");
Expect(0, 67872, '\p{phnx}', "");
Expect(1, 67872, '\p{^phnx}', "");
Expect(1, 67872, '\P{phnx}', "");
Expect(0, 67872, '\P{^phnx}', "");
Expect(1, 67871, '\p{_Phnx}', "");
Expect(0, 67871, '\p{^_Phnx}', "");
Expect(0, 67871, '\P{_Phnx}', "");
Expect(1, 67871, '\P{^_Phnx}', "");
Expect(0, 67872, '\p{_Phnx}', "");
Expect(1, 67872, '\p{^_Phnx}', "");
Expect(1, 67872, '\P{_Phnx}', "");
Expect(0, 67872, '\P{^_Phnx}', "");
Error('\p{_/a/Is_PHNX}');
Error('\P{_/a/Is_PHNX}');
Expect(1, 67871, '\p{isphnx}', "");
Expect(0, 67871, '\p{^isphnx}', "");
Expect(0, 67871, '\P{isphnx}', "");
Expect(1, 67871, '\P{^isphnx}', "");
Expect(0, 67872, '\p{isphnx}', "");
Expect(1, 67872, '\p{^isphnx}', "");
Expect(1, 67872, '\P{isphnx}', "");
Expect(0, 67872, '\P{^isphnx}', "");
Expect(1, 67871, '\p{_Is_PHNX}', "");
Expect(0, 67871, '\p{^_Is_PHNX}', "");
Expect(0, 67871, '\P{_Is_PHNX}', "");
Expect(1, 67871, '\P{^_Is_PHNX}', "");
Expect(0, 67872, '\p{_Is_PHNX}', "");
Expect(1, 67872, '\p{^_Is_PHNX}', "");
Expect(1, 67872, '\P{_Is_PHNX}', "");
Expect(0, 67872, '\P{^_Is_PHNX}', "");
Error('\p{	Phonetic_Extensions/a/}');
Error('\P{	Phonetic_Extensions/a/}');
Expect(1, 7551, '\p{phoneticextensions}', "");
Expect(0, 7551, '\p{^phoneticextensions}', "");
Expect(0, 7551, '\P{phoneticextensions}', "");
Expect(1, 7551, '\P{^phoneticextensions}', "");
Expect(0, 7552, '\p{phoneticextensions}', "");
Expect(1, 7552, '\p{^phoneticextensions}', "");
Expect(1, 7552, '\P{phoneticextensions}', "");
Expect(0, 7552, '\P{^phoneticextensions}', "");
Expect(1, 7551, '\p{ _Phonetic_Extensions}', "");
Expect(0, 7551, '\p{^ _Phonetic_Extensions}', "");
Expect(0, 7551, '\P{ _Phonetic_Extensions}', "");
Expect(1, 7551, '\P{^ _Phonetic_Extensions}', "");
Expect(0, 7552, '\p{ _Phonetic_Extensions}', "");
Expect(1, 7552, '\p{^ _Phonetic_Extensions}', "");
Expect(1, 7552, '\P{ _Phonetic_Extensions}', "");
Expect(0, 7552, '\P{^ _Phonetic_Extensions}', "");
Error('\p{- Is_Phonetic_extensions/a/}');
Error('\P{- Is_Phonetic_extensions/a/}');
Expect(1, 7551, '\p{isphoneticextensions}', "");
Expect(0, 7551, '\p{^isphoneticextensions}', "");
Expect(0, 7551, '\P{isphoneticextensions}', "");
Expect(1, 7551, '\P{^isphoneticextensions}', "");
Expect(0, 7552, '\p{isphoneticextensions}', "");
Expect(1, 7552, '\p{^isphoneticextensions}', "");
Expect(1, 7552, '\P{isphoneticextensions}', "");
Expect(0, 7552, '\P{^isphoneticextensions}', "");
Expect(1, 7551, '\p{-IS_Phonetic_Extensions}', "");
Expect(0, 7551, '\p{^-IS_Phonetic_Extensions}', "");
Expect(0, 7551, '\P{-IS_Phonetic_Extensions}', "");
Expect(1, 7551, '\P{^-IS_Phonetic_Extensions}', "");
Expect(0, 7552, '\p{-IS_Phonetic_Extensions}', "");
Expect(1, 7552, '\p{^-IS_Phonetic_Extensions}', "");
Expect(1, 7552, '\P{-IS_Phonetic_Extensions}', "");
Expect(0, 7552, '\P{^-IS_Phonetic_Extensions}', "");
Error('\p{_ In_phonetic_extensions:=}');
Error('\P{_ In_phonetic_extensions:=}');
Expect(1, 7551, '\p{inphoneticextensions}', "");
Expect(0, 7551, '\p{^inphoneticextensions}', "");
Expect(0, 7551, '\P{inphoneticextensions}', "");
Expect(1, 7551, '\P{^inphoneticextensions}', "");
Expect(0, 7552, '\p{inphoneticextensions}', "");
Expect(1, 7552, '\p{^inphoneticextensions}', "");
Expect(1, 7552, '\P{inphoneticextensions}', "");
Expect(0, 7552, '\P{^inphoneticextensions}', "");
Expect(1, 7551, '\p{	In_Phonetic_Extensions}', "");
Expect(0, 7551, '\p{^	In_Phonetic_Extensions}', "");
Expect(0, 7551, '\P{	In_Phonetic_Extensions}', "");
Expect(1, 7551, '\P{^	In_Phonetic_Extensions}', "");
Expect(0, 7552, '\p{	In_Phonetic_Extensions}', "");
Expect(1, 7552, '\p{^	In_Phonetic_Extensions}', "");
Expect(1, 7552, '\P{	In_Phonetic_Extensions}', "");
Expect(0, 7552, '\P{^	In_Phonetic_Extensions}', "");
Error('\p{_-Phonetic_Ext/a/}');
Error('\P{_-Phonetic_Ext/a/}');
Expect(1, 7551, '\p{phoneticext}', "");
Expect(0, 7551, '\p{^phoneticext}', "");
Expect(0, 7551, '\P{phoneticext}', "");
Expect(1, 7551, '\P{^phoneticext}', "");
Expect(0, 7552, '\p{phoneticext}', "");
Expect(1, 7552, '\p{^phoneticext}', "");
Expect(1, 7552, '\P{phoneticext}', "");
Expect(0, 7552, '\P{^phoneticext}', "");
Expect(1, 7551, '\p{	phonetic_Ext}', "");
Expect(0, 7551, '\p{^	phonetic_Ext}', "");
Expect(0, 7551, '\P{	phonetic_Ext}', "");
Expect(1, 7551, '\P{^	phonetic_Ext}', "");
Expect(0, 7552, '\p{	phonetic_Ext}', "");
Expect(1, 7552, '\p{^	phonetic_Ext}', "");
Expect(1, 7552, '\P{	phonetic_Ext}', "");
Expect(0, 7552, '\P{^	phonetic_Ext}', "");
Error('\p{-/a/Is_PHONETIC_EXT}');
Error('\P{-/a/Is_PHONETIC_EXT}');
Expect(1, 7551, '\p{isphoneticext}', "");
Expect(0, 7551, '\p{^isphoneticext}', "");
Expect(0, 7551, '\P{isphoneticext}', "");
Expect(1, 7551, '\P{^isphoneticext}', "");
Expect(0, 7552, '\p{isphoneticext}', "");
Expect(1, 7552, '\p{^isphoneticext}', "");
Expect(1, 7552, '\P{isphoneticext}', "");
Expect(0, 7552, '\P{^isphoneticext}', "");
Expect(1, 7551, '\p{IS_phonetic_Ext}', "");
Expect(0, 7551, '\p{^IS_phonetic_Ext}', "");
Expect(0, 7551, '\P{IS_phonetic_Ext}', "");
Expect(1, 7551, '\P{^IS_phonetic_Ext}', "");
Expect(0, 7552, '\p{IS_phonetic_Ext}', "");
Expect(1, 7552, '\p{^IS_phonetic_Ext}', "");
Expect(1, 7552, '\P{IS_phonetic_Ext}', "");
Expect(0, 7552, '\P{^IS_phonetic_Ext}', "");
Error('\p{-:=IN_phonetic_EXT}');
Error('\P{-:=IN_phonetic_EXT}');
Expect(1, 7551, '\p{inphoneticext}', "");
Expect(0, 7551, '\p{^inphoneticext}', "");
Expect(0, 7551, '\P{inphoneticext}', "");
Expect(1, 7551, '\P{^inphoneticext}', "");
Expect(0, 7552, '\p{inphoneticext}', "");
Expect(1, 7552, '\p{^inphoneticext}', "");
Expect(1, 7552, '\P{inphoneticext}', "");
Expect(0, 7552, '\P{^inphoneticext}', "");
Expect(1, 7551, '\p{	-IN_Phonetic_Ext}', "");
Expect(0, 7551, '\p{^	-IN_Phonetic_Ext}', "");
Expect(0, 7551, '\P{	-IN_Phonetic_Ext}', "");
Expect(1, 7551, '\P{^	-IN_Phonetic_Ext}', "");
Expect(0, 7552, '\p{	-IN_Phonetic_Ext}', "");
Expect(1, 7552, '\p{^	-IN_Phonetic_Ext}', "");
Expect(1, 7552, '\P{	-IN_Phonetic_Ext}', "");
Expect(0, 7552, '\P{^	-IN_Phonetic_Ext}', "");
Error('\p{:=_phonetic_Extensions_SUPPLEMENT}');
Error('\P{:=_phonetic_Extensions_SUPPLEMENT}');
Expect(1, 7615, '\p{phoneticextensionssupplement}', "");
Expect(0, 7615, '\p{^phoneticextensionssupplement}', "");
Expect(0, 7615, '\P{phoneticextensionssupplement}', "");
Expect(1, 7615, '\P{^phoneticextensionssupplement}', "");
Expect(0, 7616, '\p{phoneticextensionssupplement}', "");
Expect(1, 7616, '\p{^phoneticextensionssupplement}', "");
Expect(1, 7616, '\P{phoneticextensionssupplement}', "");
Expect(0, 7616, '\P{^phoneticextensionssupplement}', "");
Expect(1, 7615, '\p{		PHONETIC_Extensions_supplement}', "");
Expect(0, 7615, '\p{^		PHONETIC_Extensions_supplement}', "");
Expect(0, 7615, '\P{		PHONETIC_Extensions_supplement}', "");
Expect(1, 7615, '\P{^		PHONETIC_Extensions_supplement}', "");
Expect(0, 7616, '\p{		PHONETIC_Extensions_supplement}', "");
Expect(1, 7616, '\p{^		PHONETIC_Extensions_supplement}', "");
Expect(1, 7616, '\P{		PHONETIC_Extensions_supplement}', "");
Expect(0, 7616, '\P{^		PHONETIC_Extensions_supplement}', "");
Error('\p{ :=IS_Phonetic_Extensions_Supplement}');
Error('\P{ :=IS_Phonetic_Extensions_Supplement}');
Expect(1, 7615, '\p{isphoneticextensionssupplement}', "");
Expect(0, 7615, '\p{^isphoneticextensionssupplement}', "");
Expect(0, 7615, '\P{isphoneticextensionssupplement}', "");
Expect(1, 7615, '\P{^isphoneticextensionssupplement}', "");
Expect(0, 7616, '\p{isphoneticextensionssupplement}', "");
Expect(1, 7616, '\p{^isphoneticextensionssupplement}', "");
Expect(1, 7616, '\P{isphoneticextensionssupplement}', "");
Expect(0, 7616, '\P{^isphoneticextensionssupplement}', "");
Expect(1, 7615, '\p{- Is_Phonetic_EXTENSIONS_Supplement}', "");
Expect(0, 7615, '\p{^- Is_Phonetic_EXTENSIONS_Supplement}', "");
Expect(0, 7615, '\P{- Is_Phonetic_EXTENSIONS_Supplement}', "");
Expect(1, 7615, '\P{^- Is_Phonetic_EXTENSIONS_Supplement}', "");
Expect(0, 7616, '\p{- Is_Phonetic_EXTENSIONS_Supplement}', "");
Expect(1, 7616, '\p{^- Is_Phonetic_EXTENSIONS_Supplement}', "");
Expect(1, 7616, '\P{- Is_Phonetic_EXTENSIONS_Supplement}', "");
Expect(0, 7616, '\P{^- Is_Phonetic_EXTENSIONS_Supplement}', "");
Error('\p{:=	In_phonetic_Extensions_SUPPLEMENT}');
Error('\P{:=	In_phonetic_Extensions_SUPPLEMENT}');
Expect(1, 7615, '\p{inphoneticextensionssupplement}', "");
Expect(0, 7615, '\p{^inphoneticextensionssupplement}', "");
Expect(0, 7615, '\P{inphoneticextensionssupplement}', "");
Expect(1, 7615, '\P{^inphoneticextensionssupplement}', "");
Expect(0, 7616, '\p{inphoneticextensionssupplement}', "");
Expect(1, 7616, '\p{^inphoneticextensionssupplement}', "");
Expect(1, 7616, '\P{inphoneticextensionssupplement}', "");
Expect(0, 7616, '\P{^inphoneticextensionssupplement}', "");
Expect(1, 7615, '\p{--in_Phonetic_Extensions_Supplement}', "");
Expect(0, 7615, '\p{^--in_Phonetic_Extensions_Supplement}', "");
Expect(0, 7615, '\P{--in_Phonetic_Extensions_Supplement}', "");
Expect(1, 7615, '\P{^--in_Phonetic_Extensions_Supplement}', "");
Expect(0, 7616, '\p{--in_Phonetic_Extensions_Supplement}', "");
Expect(1, 7616, '\p{^--in_Phonetic_Extensions_Supplement}', "");
Expect(1, 7616, '\P{--in_Phonetic_Extensions_Supplement}', "");
Expect(0, 7616, '\P{^--in_Phonetic_Extensions_Supplement}', "");
Error('\p{--PHONETIC_ext_SUP:=}');
Error('\P{--PHONETIC_ext_SUP:=}');
Expect(1, 7615, '\p{phoneticextsup}', "");
Expect(0, 7615, '\p{^phoneticextsup}', "");
Expect(0, 7615, '\P{phoneticextsup}', "");
Expect(1, 7615, '\P{^phoneticextsup}', "");
Expect(0, 7616, '\p{phoneticextsup}', "");
Expect(1, 7616, '\p{^phoneticextsup}', "");
Expect(1, 7616, '\P{phoneticextsup}', "");
Expect(0, 7616, '\P{^phoneticextsup}', "");
Expect(1, 7615, '\p{	-PHONETIC_ext_sup}', "");
Expect(0, 7615, '\p{^	-PHONETIC_ext_sup}', "");
Expect(0, 7615, '\P{	-PHONETIC_ext_sup}', "");
Expect(1, 7615, '\P{^	-PHONETIC_ext_sup}', "");
Expect(0, 7616, '\p{	-PHONETIC_ext_sup}', "");
Expect(1, 7616, '\p{^	-PHONETIC_ext_sup}', "");
Expect(1, 7616, '\P{	-PHONETIC_ext_sup}', "");
Expect(0, 7616, '\P{^	-PHONETIC_ext_sup}', "");
Error('\p{- is_PHONETIC_Ext_SUP/a/}');
Error('\P{- is_PHONETIC_Ext_SUP/a/}');
Expect(1, 7615, '\p{isphoneticextsup}', "");
Expect(0, 7615, '\p{^isphoneticextsup}', "");
Expect(0, 7615, '\P{isphoneticextsup}', "");
Expect(1, 7615, '\P{^isphoneticextsup}', "");
Expect(0, 7616, '\p{isphoneticextsup}', "");
Expect(1, 7616, '\p{^isphoneticextsup}', "");
Expect(1, 7616, '\P{isphoneticextsup}', "");
Expect(0, 7616, '\P{^isphoneticextsup}', "");
Expect(1, 7615, '\p{-is_PHONETIC_ext_sup}', "");
Expect(0, 7615, '\p{^-is_PHONETIC_ext_sup}', "");
Expect(0, 7615, '\P{-is_PHONETIC_ext_sup}', "");
Expect(1, 7615, '\P{^-is_PHONETIC_ext_sup}', "");
Expect(0, 7616, '\p{-is_PHONETIC_ext_sup}', "");
Expect(1, 7616, '\p{^-is_PHONETIC_ext_sup}', "");
Expect(1, 7616, '\P{-is_PHONETIC_ext_sup}', "");
Expect(0, 7616, '\P{^-is_PHONETIC_ext_sup}', "");
Error('\p{:=In_Phonetic_EXT_Sup}');
Error('\P{:=In_Phonetic_EXT_Sup}');
Expect(1, 7615, '\p{inphoneticextsup}', "");
Expect(0, 7615, '\p{^inphoneticextsup}', "");
Expect(0, 7615, '\P{inphoneticextsup}', "");
Expect(1, 7615, '\P{^inphoneticextsup}', "");
Expect(0, 7616, '\p{inphoneticextsup}', "");
Expect(1, 7616, '\p{^inphoneticextsup}', "");
Expect(1, 7616, '\P{inphoneticextsup}', "");
Expect(0, 7616, '\P{^inphoneticextsup}', "");
Expect(1, 7615, '\p{	-In_Phonetic_ext_Sup}', "");
Expect(0, 7615, '\p{^	-In_Phonetic_ext_Sup}', "");
Expect(0, 7615, '\P{	-In_Phonetic_ext_Sup}', "");
Expect(1, 7615, '\P{^	-In_Phonetic_ext_Sup}', "");
Expect(0, 7616, '\p{	-In_Phonetic_ext_Sup}', "");
Expect(1, 7616, '\p{^	-In_Phonetic_ext_Sup}', "");
Expect(1, 7616, '\P{	-In_Phonetic_ext_Sup}', "");
Expect(0, 7616, '\P{^	-In_Phonetic_ext_Sup}', "");
Error('\p{- PLAYING_cards:=}');
Error('\P{- PLAYING_cards:=}');
Expect(1, 127231, '\p{playingcards}', "");
Expect(0, 127231, '\p{^playingcards}', "");
Expect(0, 127231, '\P{playingcards}', "");
Expect(1, 127231, '\P{^playingcards}', "");
Expect(0, 127232, '\p{playingcards}', "");
Expect(1, 127232, '\p{^playingcards}', "");
Expect(1, 127232, '\P{playingcards}', "");
Expect(0, 127232, '\P{^playingcards}', "");
Expect(1, 127231, '\p{-playing_CARDS}', "");
Expect(0, 127231, '\p{^-playing_CARDS}', "");
Expect(0, 127231, '\P{-playing_CARDS}', "");
Expect(1, 127231, '\P{^-playing_CARDS}', "");
Expect(0, 127232, '\p{-playing_CARDS}', "");
Expect(1, 127232, '\p{^-playing_CARDS}', "");
Expect(1, 127232, '\P{-playing_CARDS}', "");
Expect(0, 127232, '\P{^-playing_CARDS}', "");
Error('\p{/a/IS_PLAYING_CARDS}');
Error('\P{/a/IS_PLAYING_CARDS}');
Expect(1, 127231, '\p{isplayingcards}', "");
Expect(0, 127231, '\p{^isplayingcards}', "");
Expect(0, 127231, '\P{isplayingcards}', "");
Expect(1, 127231, '\P{^isplayingcards}', "");
Expect(0, 127232, '\p{isplayingcards}', "");
Expect(1, 127232, '\p{^isplayingcards}', "");
Expect(1, 127232, '\P{isplayingcards}', "");
Expect(0, 127232, '\P{^isplayingcards}', "");
Expect(1, 127231, '\p{	Is_playing_CARDS}', "");
Expect(0, 127231, '\p{^	Is_playing_CARDS}', "");
Expect(0, 127231, '\P{	Is_playing_CARDS}', "");
Expect(1, 127231, '\P{^	Is_playing_CARDS}', "");
Expect(0, 127232, '\p{	Is_playing_CARDS}', "");
Expect(1, 127232, '\p{^	Is_playing_CARDS}', "");
Expect(1, 127232, '\P{	Is_playing_CARDS}', "");
Expect(0, 127232, '\P{^	Is_playing_CARDS}', "");
Error('\p{-/a/In_Playing_Cards}');
Error('\P{-/a/In_Playing_Cards}');
Expect(1, 127231, '\p{inplayingcards}', "");
Expect(0, 127231, '\p{^inplayingcards}', "");
Expect(0, 127231, '\P{inplayingcards}', "");
Expect(1, 127231, '\P{^inplayingcards}', "");
Expect(0, 127232, '\p{inplayingcards}', "");
Expect(1, 127232, '\p{^inplayingcards}', "");
Expect(1, 127232, '\P{inplayingcards}', "");
Expect(0, 127232, '\P{^inplayingcards}', "");
Expect(1, 127231, '\p{ -In_playing_Cards}', "");
Expect(0, 127231, '\p{^ -In_playing_Cards}', "");
Expect(0, 127231, '\P{ -In_playing_Cards}', "");
Expect(1, 127231, '\P{^ -In_playing_Cards}', "");
Expect(0, 127232, '\p{ -In_playing_Cards}', "");
Expect(1, 127232, '\p{^ -In_playing_Cards}', "");
Expect(1, 127232, '\P{ -In_playing_Cards}', "");
Expect(0, 127232, '\P{^ -In_playing_Cards}', "");
Error('\p{:= PosixAlnum}');
Error('\P{:= PosixAlnum}');
Expect(1, 122, '\p{posixalnum}', "");
Expect(0, 122, '\p{^posixalnum}', "");
Expect(0, 122, '\P{posixalnum}', "");
Expect(1, 122, '\P{^posixalnum}', "");
Expect(0, 123, '\p{posixalnum}', "");
Expect(1, 123, '\p{^posixalnum}', "");
Expect(1, 123, '\P{posixalnum}', "");
Expect(0, 123, '\P{^posixalnum}', "");
Expect(1, 122, '\p{_-PosixAlnum}', "");
Expect(0, 122, '\p{^_-PosixAlnum}', "");
Expect(0, 122, '\P{_-PosixAlnum}', "");
Expect(1, 122, '\P{^_-PosixAlnum}', "");
Expect(0, 123, '\p{_-PosixAlnum}', "");
Expect(1, 123, '\p{^_-PosixAlnum}', "");
Expect(1, 123, '\P{_-PosixAlnum}', "");
Expect(0, 123, '\P{^_-PosixAlnum}', "");
Error('\p{-Is_POSIXALNUM/a/}');
Error('\P{-Is_POSIXALNUM/a/}');
Expect(1, 122, '\p{isposixalnum}', "");
Expect(0, 122, '\p{^isposixalnum}', "");
Expect(0, 122, '\P{isposixalnum}', "");
Expect(1, 122, '\P{^isposixalnum}', "");
Expect(0, 123, '\p{isposixalnum}', "");
Expect(1, 123, '\p{^isposixalnum}', "");
Expect(1, 123, '\P{isposixalnum}', "");
Expect(0, 123, '\P{^isposixalnum}', "");
Expect(1, 122, '\p{	IS_PosixAlnum}', "");
Expect(0, 122, '\p{^	IS_PosixAlnum}', "");
Expect(0, 122, '\P{	IS_PosixAlnum}', "");
Expect(1, 122, '\P{^	IS_PosixAlnum}', "");
Expect(0, 123, '\p{	IS_PosixAlnum}', "");
Expect(1, 123, '\p{^	IS_PosixAlnum}', "");
Expect(1, 123, '\P{	IS_PosixAlnum}', "");
Expect(0, 123, '\P{^	IS_PosixAlnum}', "");
Error('\p{:=  PosixAlpha}');
Error('\P{:=  PosixAlpha}');
Expect(1, 122, '\p{posixalpha}', "");
Expect(0, 122, '\p{^posixalpha}', "");
Expect(0, 122, '\P{posixalpha}', "");
Expect(1, 122, '\P{^posixalpha}', "");
Expect(0, 123, '\p{posixalpha}', "");
Expect(1, 123, '\p{^posixalpha}', "");
Expect(1, 123, '\P{posixalpha}', "");
Expect(0, 123, '\P{^posixalpha}', "");
Expect(1, 122, '\p{ -PosixAlpha}', "");
Expect(0, 122, '\p{^ -PosixAlpha}', "");
Expect(0, 122, '\P{ -PosixAlpha}', "");
Expect(1, 122, '\P{^ -PosixAlpha}', "");
Expect(0, 123, '\p{ -PosixAlpha}', "");
Expect(1, 123, '\p{^ -PosixAlpha}', "");
Expect(1, 123, '\P{ -PosixAlpha}', "");
Expect(0, 123, '\P{^ -PosixAlpha}', "");
Error('\p{:=	IS_PosixAlpha}');
Error('\P{:=	IS_PosixAlpha}');
Expect(1, 122, '\p{isposixalpha}', "");
Expect(0, 122, '\p{^isposixalpha}', "");
Expect(0, 122, '\P{isposixalpha}', "");
Expect(1, 122, '\P{^isposixalpha}', "");
Expect(0, 123, '\p{isposixalpha}', "");
Expect(1, 123, '\p{^isposixalpha}', "");
Expect(1, 123, '\P{isposixalpha}', "");
Expect(0, 123, '\P{^isposixalpha}', "");
Expect(1, 122, '\p{ -IS_PosixAlpha}', "");
Expect(0, 122, '\p{^ -IS_PosixAlpha}', "");
Expect(0, 122, '\P{ -IS_PosixAlpha}', "");
Expect(1, 122, '\P{^ -IS_PosixAlpha}', "");
Expect(0, 123, '\p{ -IS_PosixAlpha}', "");
Expect(1, 123, '\p{^ -IS_PosixAlpha}', "");
Expect(1, 123, '\P{ -IS_PosixAlpha}', "");
Expect(0, 123, '\P{^ -IS_PosixAlpha}', "");
Error('\p{/a/--POSIXBLANK}');
Error('\P{/a/--POSIXBLANK}');
Expect(1, 32, '\p{posixblank}', "");
Expect(0, 32, '\p{^posixblank}', "");
Expect(0, 32, '\P{posixblank}', "");
Expect(1, 32, '\P{^posixblank}', "");
Expect(0, 33, '\p{posixblank}', "");
Expect(1, 33, '\p{^posixblank}', "");
Expect(1, 33, '\P{posixblank}', "");
Expect(0, 33, '\P{^posixblank}', "");
Expect(1, 32, '\p{ -PosixBlank}', "");
Expect(0, 32, '\p{^ -PosixBlank}', "");
Expect(0, 32, '\P{ -PosixBlank}', "");
Expect(1, 32, '\P{^ -PosixBlank}', "");
Expect(0, 33, '\p{ -PosixBlank}', "");
Expect(1, 33, '\p{^ -PosixBlank}', "");
Expect(1, 33, '\P{ -PosixBlank}', "");
Expect(0, 33, '\P{^ -PosixBlank}', "");
Error('\p{/a/--Is_PosixBlank}');
Error('\P{/a/--Is_PosixBlank}');
Expect(1, 32, '\p{isposixblank}', "");
Expect(0, 32, '\p{^isposixblank}', "");
Expect(0, 32, '\P{isposixblank}', "");
Expect(1, 32, '\P{^isposixblank}', "");
Expect(0, 33, '\p{isposixblank}', "");
Expect(1, 33, '\p{^isposixblank}', "");
Expect(1, 33, '\P{isposixblank}', "");
Expect(0, 33, '\P{^isposixblank}', "");
Expect(1, 32, '\p{  is_PosixBlank}', "");
Expect(0, 32, '\p{^  is_PosixBlank}', "");
Expect(0, 32, '\P{  is_PosixBlank}', "");
Expect(1, 32, '\P{^  is_PosixBlank}', "");
Expect(0, 33, '\p{  is_PosixBlank}', "");
Expect(1, 33, '\p{^  is_PosixBlank}', "");
Expect(1, 33, '\P{  is_PosixBlank}', "");
Expect(0, 33, '\P{^  is_PosixBlank}', "");
Error('\p{:=_PosixCntrl}');
Error('\P{:=_PosixCntrl}');
Expect(1, 127, '\p{posixcntrl}', "");
Expect(0, 127, '\p{^posixcntrl}', "");
Expect(0, 127, '\P{posixcntrl}', "");
Expect(1, 127, '\P{^posixcntrl}', "");
Expect(0, 128, '\p{posixcntrl}', "");
Expect(1, 128, '\p{^posixcntrl}', "");
Expect(1, 128, '\P{posixcntrl}', "");
Expect(0, 128, '\P{^posixcntrl}', "");
Expect(1, 127, '\p{	 POSIXCNTRL}', "");
Expect(0, 127, '\p{^	 POSIXCNTRL}', "");
Expect(0, 127, '\P{	 POSIXCNTRL}', "");
Expect(1, 127, '\P{^	 POSIXCNTRL}', "");
Expect(0, 128, '\p{	 POSIXCNTRL}', "");
Expect(1, 128, '\p{^	 POSIXCNTRL}', "");
Expect(1, 128, '\P{	 POSIXCNTRL}', "");
Expect(0, 128, '\P{^	 POSIXCNTRL}', "");
Error('\p{:=	 is_posixcntrl}');
Error('\P{:=	 is_posixcntrl}');
Expect(1, 127, '\p{isposixcntrl}', "");
Expect(0, 127, '\p{^isposixcntrl}', "");
Expect(0, 127, '\P{isposixcntrl}', "");
Expect(1, 127, '\P{^isposixcntrl}', "");
Expect(0, 128, '\p{isposixcntrl}', "");
Expect(1, 128, '\p{^isposixcntrl}', "");
Expect(1, 128, '\P{isposixcntrl}', "");
Expect(0, 128, '\P{^isposixcntrl}', "");
Expect(1, 127, '\p{-is_POSIXCNTRL}', "");
Expect(0, 127, '\p{^-is_POSIXCNTRL}', "");
Expect(0, 127, '\P{-is_POSIXCNTRL}', "");
Expect(1, 127, '\P{^-is_POSIXCNTRL}', "");
Expect(0, 128, '\p{-is_POSIXCNTRL}', "");
Expect(1, 128, '\p{^-is_POSIXCNTRL}', "");
Expect(1, 128, '\P{-is_POSIXCNTRL}', "");
Expect(0, 128, '\P{^-is_POSIXCNTRL}', "");
Error('\p{:=_PosixDigit}');
Error('\P{:=_PosixDigit}');
Expect(1, 57, '\p{posixdigit}', "");
Expect(0, 57, '\p{^posixdigit}', "");
Expect(0, 57, '\P{posixdigit}', "");
Expect(1, 57, '\P{^posixdigit}', "");
Expect(0, 58, '\p{posixdigit}', "");
Expect(1, 58, '\p{^posixdigit}', "");
Expect(1, 58, '\P{posixdigit}', "");
Expect(0, 58, '\P{^posixdigit}', "");
Expect(1, 57, '\p{ POSIXDIGIT}', "");
Expect(0, 57, '\p{^ POSIXDIGIT}', "");
Expect(0, 57, '\P{ POSIXDIGIT}', "");
Expect(1, 57, '\P{^ POSIXDIGIT}', "");
Expect(0, 58, '\p{ POSIXDIGIT}', "");
Expect(1, 58, '\p{^ POSIXDIGIT}', "");
Expect(1, 58, '\P{ POSIXDIGIT}', "");
Expect(0, 58, '\P{^ POSIXDIGIT}', "");
Error('\p{ Is_posixdigit:=}');
Error('\P{ Is_posixdigit:=}');
Expect(1, 57, '\p{isposixdigit}', "");
Expect(0, 57, '\p{^isposixdigit}', "");
Expect(0, 57, '\P{isposixdigit}', "");
Expect(1, 57, '\P{^isposixdigit}', "");
Expect(0, 58, '\p{isposixdigit}', "");
Expect(1, 58, '\p{^isposixdigit}', "");
Expect(1, 58, '\P{isposixdigit}', "");
Expect(0, 58, '\P{^isposixdigit}', "");
Expect(1, 57, '\p{ Is_PosixDigit}', "");
Expect(0, 57, '\p{^ Is_PosixDigit}', "");
Expect(0, 57, '\P{ Is_PosixDigit}', "");
Expect(1, 57, '\P{^ Is_PosixDigit}', "");
Expect(0, 58, '\p{ Is_PosixDigit}', "");
Expect(1, 58, '\p{^ Is_PosixDigit}', "");
Expect(1, 58, '\P{ Is_PosixDigit}', "");
Expect(0, 58, '\P{^ Is_PosixDigit}', "");
Error('\p{:= _PosixGraph}');
Error('\P{:= _PosixGraph}');
Expect(1, 126, '\p{posixgraph}', "");
Expect(0, 126, '\p{^posixgraph}', "");
Expect(0, 126, '\P{posixgraph}', "");
Expect(1, 126, '\P{^posixgraph}', "");
Expect(0, 127, '\p{posixgraph}', "");
Expect(1, 127, '\p{^posixgraph}', "");
Expect(1, 127, '\P{posixgraph}', "");
Expect(0, 127, '\P{^posixgraph}', "");
Expect(1, 126, '\p{	 POSIXGRAPH}', "");
Expect(0, 126, '\p{^	 POSIXGRAPH}', "");
Expect(0, 126, '\P{	 POSIXGRAPH}', "");
Expect(1, 126, '\P{^	 POSIXGRAPH}', "");
Expect(0, 127, '\p{	 POSIXGRAPH}', "");
Expect(1, 127, '\p{^	 POSIXGRAPH}', "");
Expect(1, 127, '\P{	 POSIXGRAPH}', "");
Expect(0, 127, '\P{^	 POSIXGRAPH}', "");
Error('\p{ Is_PosixGraph:=}');
Error('\P{ Is_PosixGraph:=}');
Expect(1, 126, '\p{isposixgraph}', "");
Expect(0, 126, '\p{^isposixgraph}', "");
Expect(0, 126, '\P{isposixgraph}', "");
Expect(1, 126, '\P{^isposixgraph}', "");
Expect(0, 127, '\p{isposixgraph}', "");
Expect(1, 127, '\p{^isposixgraph}', "");
Expect(1, 127, '\P{isposixgraph}', "");
Expect(0, 127, '\P{^isposixgraph}', "");
Expect(1, 126, '\p{-	is_posixgraph}', "");
Expect(0, 126, '\p{^-	is_posixgraph}', "");
Expect(0, 126, '\P{-	is_posixgraph}', "");
Expect(1, 126, '\P{^-	is_posixgraph}', "");
Expect(0, 127, '\p{-	is_posixgraph}', "");
Expect(1, 127, '\p{^-	is_posixgraph}', "");
Expect(1, 127, '\P{-	is_posixgraph}', "");
Expect(0, 127, '\P{^-	is_posixgraph}', "");
Error('\p{/a/- PosixLower}');
Error('\P{/a/- PosixLower}');
Expect(1, 122, '\p{posixlower}', "");
Expect(0, 122, '\p{^posixlower}', "");
Expect(0, 122, '\P{posixlower}', "");
Expect(1, 122, '\P{^posixlower}', "");
Expect(0, 123, '\p{posixlower}', "");
Expect(1, 123, '\p{^posixlower}', "");
Expect(1, 123, '\P{posixlower}', "");
Expect(0, 123, '\P{^posixlower}', "");
Expect(1, 122, '\p{-PosixLower}', "");
Expect(0, 122, '\p{^-PosixLower}', "");
Expect(0, 122, '\P{-PosixLower}', "");
Expect(1, 122, '\P{^-PosixLower}', "");
Expect(0, 123, '\p{-PosixLower}', "");
Expect(1, 123, '\p{^-PosixLower}', "");
Expect(1, 123, '\P{-PosixLower}', "");
Expect(0, 123, '\P{^-PosixLower}', "");
Error('\p{:=IS_POSIXLOWER}');
Error('\P{:=IS_POSIXLOWER}');
Expect(1, 122, '\p{isposixlower}', "");
Expect(0, 122, '\p{^isposixlower}', "");
Expect(0, 122, '\P{isposixlower}', "");
Expect(1, 122, '\P{^isposixlower}', "");
Expect(0, 123, '\p{isposixlower}', "");
Expect(1, 123, '\p{^isposixlower}', "");
Expect(1, 123, '\P{isposixlower}', "");
Expect(0, 123, '\P{^isposixlower}', "");
Expect(1, 122, '\p{ Is_posixlower}', "");
Expect(0, 122, '\p{^ Is_posixlower}', "");
Expect(0, 122, '\P{ Is_posixlower}', "");
Expect(1, 122, '\P{^ Is_posixlower}', "");
Expect(0, 123, '\p{ Is_posixlower}', "");
Expect(1, 123, '\p{^ Is_posixlower}', "");
Expect(1, 123, '\P{ Is_posixlower}', "");
Expect(0, 123, '\P{^ Is_posixlower}', "");
Error('\p{/a/_POSIXPRINT}');
Error('\P{/a/_POSIXPRINT}');
Expect(1, 126, '\p{posixprint}', "");
Expect(0, 126, '\p{^posixprint}', "");
Expect(0, 126, '\P{posixprint}', "");
Expect(1, 126, '\P{^posixprint}', "");
Expect(0, 127, '\p{posixprint}', "");
Expect(1, 127, '\p{^posixprint}', "");
Expect(1, 127, '\P{posixprint}', "");
Expect(0, 127, '\P{^posixprint}', "");
Expect(1, 126, '\p{_	PosixPrint}', "");
Expect(0, 126, '\p{^_	PosixPrint}', "");
Expect(0, 126, '\P{_	PosixPrint}', "");
Expect(1, 126, '\P{^_	PosixPrint}', "");
Expect(0, 127, '\p{_	PosixPrint}', "");
Expect(1, 127, '\p{^_	PosixPrint}', "");
Expect(1, 127, '\P{_	PosixPrint}', "");
Expect(0, 127, '\P{^_	PosixPrint}', "");
Error('\p{/a/		Is_posixprint}');
Error('\P{/a/		Is_posixprint}');
Expect(1, 126, '\p{isposixprint}', "");
Expect(0, 126, '\p{^isposixprint}', "");
Expect(0, 126, '\P{isposixprint}', "");
Expect(1, 126, '\P{^isposixprint}', "");
Expect(0, 127, '\p{isposixprint}', "");
Expect(1, 127, '\p{^isposixprint}', "");
Expect(1, 127, '\P{isposixprint}', "");
Expect(0, 127, '\P{^isposixprint}', "");
Expect(1, 126, '\p{		Is_posixprint}', "");
Expect(0, 126, '\p{^		Is_posixprint}', "");
Expect(0, 126, '\P{		Is_posixprint}', "");
Expect(1, 126, '\P{^		Is_posixprint}', "");
Expect(0, 127, '\p{		Is_posixprint}', "");
Expect(1, 127, '\p{^		Is_posixprint}', "");
Expect(1, 127, '\P{		Is_posixprint}', "");
Expect(0, 127, '\P{^		Is_posixprint}', "");
Error('\p{-_PosixPunct/a/}');
Error('\P{-_PosixPunct/a/}');
Expect(1, 126, '\p{posixpunct}', "");
Expect(0, 126, '\p{^posixpunct}', "");
Expect(0, 126, '\P{posixpunct}', "");
Expect(1, 126, '\P{^posixpunct}', "");
Expect(0, 127, '\p{posixpunct}', "");
Expect(1, 127, '\p{^posixpunct}', "");
Expect(1, 127, '\P{posixpunct}', "");
Expect(0, 127, '\P{^posixpunct}', "");
Expect(1, 126, '\p{__PosixPunct}', "");
Expect(0, 126, '\p{^__PosixPunct}', "");
Expect(0, 126, '\P{__PosixPunct}', "");
Expect(1, 126, '\P{^__PosixPunct}', "");
Expect(0, 127, '\p{__PosixPunct}', "");
Expect(1, 127, '\p{^__PosixPunct}', "");
Expect(1, 127, '\P{__PosixPunct}', "");
Expect(0, 127, '\P{^__PosixPunct}', "");
Error('\p{	Is_POSIXPUNCT/a/}');
Error('\P{	Is_POSIXPUNCT/a/}');
Expect(1, 126, '\p{isposixpunct}', "");
Expect(0, 126, '\p{^isposixpunct}', "");
Expect(0, 126, '\P{isposixpunct}', "");
Expect(1, 126, '\P{^isposixpunct}', "");
Expect(0, 127, '\p{isposixpunct}', "");
Expect(1, 127, '\p{^isposixpunct}', "");
Expect(1, 127, '\P{isposixpunct}', "");
Expect(0, 127, '\P{^isposixpunct}', "");
Expect(1, 126, '\p{-	IS_posixpunct}', "");
Expect(0, 126, '\p{^-	IS_posixpunct}', "");
Expect(0, 126, '\P{-	IS_posixpunct}', "");
Expect(1, 126, '\P{^-	IS_posixpunct}', "");
Expect(0, 127, '\p{-	IS_posixpunct}', "");
Expect(1, 127, '\p{^-	IS_posixpunct}', "");
Expect(1, 127, '\P{-	IS_posixpunct}', "");
Expect(0, 127, '\P{^-	IS_posixpunct}', "");
Error('\p{ -POSIXSPACE/a/}');
Error('\P{ -POSIXSPACE/a/}');
Expect(1, 32, '\p{posixspace}', "");
Expect(0, 32, '\p{^posixspace}', "");
Expect(0, 32, '\P{posixspace}', "");
Expect(1, 32, '\P{^posixspace}', "");
Expect(0, 33, '\p{posixspace}', "");
Expect(1, 33, '\p{^posixspace}', "");
Expect(1, 33, '\P{posixspace}', "");
Expect(0, 33, '\P{^posixspace}', "");
Expect(1, 32, '\p{__POSIXSPACE}', "");
Expect(0, 32, '\p{^__POSIXSPACE}', "");
Expect(0, 32, '\P{__POSIXSPACE}', "");
Expect(1, 32, '\P{^__POSIXSPACE}', "");
Expect(0, 33, '\p{__POSIXSPACE}', "");
Expect(1, 33, '\p{^__POSIXSPACE}', "");
Expect(1, 33, '\P{__POSIXSPACE}', "");
Expect(0, 33, '\P{^__POSIXSPACE}', "");
Error('\p{ 	Is_PosixSpace:=}');
Error('\P{ 	Is_PosixSpace:=}');
Expect(1, 32, '\p{isposixspace}', "");
Expect(0, 32, '\p{^isposixspace}', "");
Expect(0, 32, '\P{isposixspace}', "");
Expect(1, 32, '\P{^isposixspace}', "");
Expect(0, 33, '\p{isposixspace}', "");
Expect(1, 33, '\p{^isposixspace}', "");
Expect(1, 33, '\P{isposixspace}', "");
Expect(0, 33, '\P{^isposixspace}', "");
Expect(1, 32, '\p{ -Is_posixspace}', "");
Expect(0, 32, '\p{^ -Is_posixspace}', "");
Expect(0, 32, '\P{ -Is_posixspace}', "");
Expect(1, 32, '\P{^ -Is_posixspace}', "");
Expect(0, 33, '\p{ -Is_posixspace}', "");
Expect(1, 33, '\p{^ -Is_posixspace}', "");
Expect(1, 33, '\P{ -Is_posixspace}', "");
Expect(0, 33, '\P{^ -Is_posixspace}', "");
Error('\p{ 	posixupper:=}');
Error('\P{ 	posixupper:=}');
Expect(1, 90, '\p{posixupper}', "");
Expect(0, 90, '\p{^posixupper}', "");
Expect(0, 90, '\P{posixupper}', "");
Expect(1, 90, '\P{^posixupper}', "");
Expect(0, 91, '\p{posixupper}', "");
Expect(1, 91, '\p{^posixupper}', "");
Expect(1, 91, '\P{posixupper}', "");
Expect(0, 91, '\P{^posixupper}', "");
Expect(1, 90, '\p{_PosixUpper}', "");
Expect(0, 90, '\p{^_PosixUpper}', "");
Expect(0, 90, '\P{_PosixUpper}', "");
Expect(1, 90, '\P{^_PosixUpper}', "");
Expect(0, 91, '\p{_PosixUpper}', "");
Expect(1, 91, '\p{^_PosixUpper}', "");
Expect(1, 91, '\P{_PosixUpper}', "");
Expect(0, 91, '\P{^_PosixUpper}', "");
Error('\p{:= IS_PosixUpper}');
Error('\P{:= IS_PosixUpper}');
Expect(1, 90, '\p{isposixupper}', "");
Expect(0, 90, '\p{^isposixupper}', "");
Expect(0, 90, '\P{isposixupper}', "");
Expect(1, 90, '\P{^isposixupper}', "");
Expect(0, 91, '\p{isposixupper}', "");
Expect(1, 91, '\p{^isposixupper}', "");
Expect(1, 91, '\P{isposixupper}', "");
Expect(0, 91, '\P{^isposixupper}', "");
Expect(1, 90, '\p{_	is_posixupper}', "");
Expect(0, 90, '\p{^_	is_posixupper}', "");
Expect(0, 90, '\P{_	is_posixupper}', "");
Expect(1, 90, '\P{^_	is_posixupper}', "");
Expect(0, 91, '\p{_	is_posixupper}', "");
Expect(1, 91, '\p{^_	is_posixupper}', "");
Expect(1, 91, '\P{_	is_posixupper}', "");
Expect(0, 91, '\P{^_	is_posixupper}', "");
Error('\p{:=PosixXDigit}');
Error('\P{:=PosixXDigit}');
Expect(1, 102, '\p{posixxdigit}', "");
Expect(0, 102, '\p{^posixxdigit}', "");
Expect(0, 102, '\P{posixxdigit}', "");
Expect(1, 102, '\P{^posixxdigit}', "");
Expect(0, 103, '\p{posixxdigit}', "");
Expect(1, 103, '\p{^posixxdigit}', "");
Expect(1, 103, '\P{posixxdigit}', "");
Expect(0, 103, '\P{^posixxdigit}', "");
Expect(1, 102, '\p{	-POSIXXDIGIT}', "");
Expect(0, 102, '\p{^	-POSIXXDIGIT}', "");
Expect(0, 102, '\P{	-POSIXXDIGIT}', "");
Expect(1, 102, '\P{^	-POSIXXDIGIT}', "");
Expect(0, 103, '\p{	-POSIXXDIGIT}', "");
Expect(1, 103, '\p{^	-POSIXXDIGIT}', "");
Expect(1, 103, '\P{	-POSIXXDIGIT}', "");
Expect(0, 103, '\P{^	-POSIXXDIGIT}', "");
Error('\p{ Is_PosixXDigit/a/}');
Error('\P{ Is_PosixXDigit/a/}');
Expect(1, 102, '\p{isposixxdigit}', "");
Expect(0, 102, '\p{^isposixxdigit}', "");
Expect(0, 102, '\P{isposixxdigit}', "");
Expect(1, 102, '\P{^isposixxdigit}', "");
Expect(0, 103, '\p{isposixxdigit}', "");
Expect(1, 103, '\p{^isposixxdigit}', "");
Expect(1, 103, '\P{isposixxdigit}', "");
Expect(0, 103, '\P{^isposixxdigit}', "");
Expect(1, 102, '\p{	is_POSIXXDIGIT}', "");
Expect(0, 102, '\p{^	is_POSIXXDIGIT}', "");
Expect(0, 102, '\P{	is_POSIXXDIGIT}', "");
Expect(1, 102, '\P{^	is_POSIXXDIGIT}', "");
Expect(0, 103, '\p{	is_POSIXXDIGIT}', "");
Expect(1, 103, '\p{^	is_POSIXXDIGIT}', "");
Expect(1, 103, '\P{	is_POSIXXDIGIT}', "");
Expect(0, 103, '\P{^	is_POSIXXDIGIT}', "");
Error('\p{	:=ascii_hex_Digit}');
Error('\P{	:=ascii_hex_Digit}');
Expect(1, 102, '\p{asciihexdigit}', "");
Expect(0, 102, '\p{^asciihexdigit}', "");
Expect(0, 102, '\P{asciihexdigit}', "");
Expect(1, 102, '\P{^asciihexdigit}', "");
Expect(0, 103, '\p{asciihexdigit}', "");
Expect(1, 103, '\p{^asciihexdigit}', "");
Expect(1, 103, '\P{asciihexdigit}', "");
Expect(0, 103, '\P{^asciihexdigit}', "");
Expect(1, 102, '\p{	ASCII_Hex_Digit}', "");
Expect(0, 102, '\p{^	ASCII_Hex_Digit}', "");
Expect(0, 102, '\P{	ASCII_Hex_Digit}', "");
Expect(1, 102, '\P{^	ASCII_Hex_Digit}', "");
Expect(0, 103, '\p{	ASCII_Hex_Digit}', "");
Expect(1, 103, '\p{^	ASCII_Hex_Digit}', "");
Expect(1, 103, '\P{	ASCII_Hex_Digit}', "");
Expect(0, 103, '\P{^	ASCII_Hex_Digit}', "");
Error('\p{	/a/Is_ASCII_Hex_Digit}');
Error('\P{	/a/Is_ASCII_Hex_Digit}');
Expect(1, 102, '\p{isasciihexdigit}', "");
Expect(0, 102, '\p{^isasciihexdigit}', "");
Expect(0, 102, '\P{isasciihexdigit}', "");
Expect(1, 102, '\P{^isasciihexdigit}', "");
Expect(0, 103, '\p{isasciihexdigit}', "");
Expect(1, 103, '\p{^isasciihexdigit}', "");
Expect(1, 103, '\P{isasciihexdigit}', "");
Expect(0, 103, '\P{^isasciihexdigit}', "");
Expect(1, 102, '\p{-is_ascii_HEX_Digit}', "");
Expect(0, 102, '\p{^-is_ascii_HEX_Digit}', "");
Expect(0, 102, '\P{-is_ascii_HEX_Digit}', "");
Expect(1, 102, '\P{^-is_ascii_HEX_Digit}', "");
Expect(0, 103, '\p{-is_ascii_HEX_Digit}', "");
Expect(1, 103, '\p{^-is_ascii_HEX_Digit}', "");
Expect(1, 103, '\P{-is_ascii_HEX_Digit}', "");
Expect(0, 103, '\P{^-is_ascii_HEX_Digit}', "");
Error('\p{/a/ AHex}');
Error('\P{/a/ AHex}');
Expect(1, 102, '\p{ahex}', "");
Expect(0, 102, '\p{^ahex}', "");
Expect(0, 102, '\P{ahex}', "");
Expect(1, 102, '\P{^ahex}', "");
Expect(0, 103, '\p{ahex}', "");
Expect(1, 103, '\p{^ahex}', "");
Expect(1, 103, '\P{ahex}', "");
Expect(0, 103, '\P{^ahex}', "");
Expect(1, 102, '\p{	AHex}', "");
Expect(0, 102, '\p{^	AHex}', "");
Expect(0, 102, '\P{	AHex}', "");
Expect(1, 102, '\P{^	AHex}', "");
Expect(0, 103, '\p{	AHex}', "");
Expect(1, 103, '\p{^	AHex}', "");
Expect(1, 103, '\P{	AHex}', "");
Expect(0, 103, '\P{^	AHex}', "");
Error('\p{:=-Is_ahex}');
Error('\P{:=-Is_ahex}');
Expect(1, 102, '\p{isahex}', "");
Expect(0, 102, '\p{^isahex}', "");
Expect(0, 102, '\P{isahex}', "");
Expect(1, 102, '\P{^isahex}', "");
Expect(0, 103, '\p{isahex}', "");
Expect(1, 103, '\p{^isahex}', "");
Expect(1, 103, '\P{isahex}', "");
Expect(0, 103, '\P{^isahex}', "");
Expect(1, 102, '\p{--IS_AHEX}', "");
Expect(0, 102, '\p{^--IS_AHEX}', "");
Expect(0, 102, '\P{--IS_AHEX}', "");
Expect(1, 102, '\P{^--IS_AHEX}', "");
Expect(0, 103, '\p{--IS_AHEX}', "");
Expect(1, 103, '\p{^--IS_AHEX}', "");
Expect(1, 103, '\P{--IS_AHEX}', "");
Expect(0, 103, '\P{^--IS_AHEX}', "");
Error('\p{/a/Print}');
Error('\P{/a/Print}');
Expect(1, 1114109, '\p{print}', "");
Expect(0, 1114109, '\p{^print}', "");
Expect(0, 1114109, '\P{print}', "");
Expect(1, 1114109, '\P{^print}', "");
Expect(0, 918000, '\p{print}', "");
Expect(1, 918000, '\p{^print}', "");
Expect(1, 918000, '\P{print}', "");
Expect(0, 918000, '\P{^print}', "");
Expect(1, 1114109, '\p{_Print}', "");
Expect(0, 1114109, '\p{^_Print}', "");
Expect(0, 1114109, '\P{_Print}', "");
Expect(1, 1114109, '\P{^_Print}', "");
Expect(0, 918000, '\p{_Print}', "");
Expect(1, 918000, '\p{^_Print}', "");
Expect(1, 918000, '\P{_Print}', "");
Expect(0, 918000, '\P{^_Print}', "");
Error('\p{-/a/XPosixPrint}');
Error('\P{-/a/XPosixPrint}');
Expect(1, 1114109, '\p{xposixprint}', "");
Expect(0, 1114109, '\p{^xposixprint}', "");
Expect(0, 1114109, '\P{xposixprint}', "");
Expect(1, 1114109, '\P{^xposixprint}', "");
Expect(0, 918000, '\p{xposixprint}', "");
Expect(1, 918000, '\p{^xposixprint}', "");
Expect(1, 918000, '\P{xposixprint}', "");
Expect(0, 918000, '\P{^xposixprint}', "");
Expect(1, 1114109, '\p{__XPOSIXPRINT}', "");
Expect(0, 1114109, '\p{^__XPOSIXPRINT}', "");
Expect(0, 1114109, '\P{__XPOSIXPRINT}', "");
Expect(1, 1114109, '\P{^__XPOSIXPRINT}', "");
Expect(0, 918000, '\p{__XPOSIXPRINT}', "");
Expect(1, 918000, '\p{^__XPOSIXPRINT}', "");
Expect(1, 918000, '\P{__XPOSIXPRINT}', "");
Expect(0, 918000, '\P{^__XPOSIXPRINT}', "");
Error('\p{	Is_PRINT:=}');
Error('\P{	Is_PRINT:=}');
Expect(1, 1114109, '\p{isprint}', "");
Expect(0, 1114109, '\p{^isprint}', "");
Expect(0, 1114109, '\P{isprint}', "");
Expect(1, 1114109, '\P{^isprint}', "");
Expect(0, 918000, '\p{isprint}', "");
Expect(1, 918000, '\p{^isprint}', "");
Expect(1, 918000, '\P{isprint}', "");
Expect(0, 918000, '\P{^isprint}', "");
Expect(1, 1114109, '\p{-	Is_Print}', "");
Expect(0, 1114109, '\p{^-	Is_Print}', "");
Expect(0, 1114109, '\P{-	Is_Print}', "");
Expect(1, 1114109, '\P{^-	Is_Print}', "");
Expect(0, 918000, '\p{-	Is_Print}', "");
Expect(1, 918000, '\p{^-	Is_Print}', "");
Expect(1, 918000, '\P{-	Is_Print}', "");
Expect(0, 918000, '\P{^-	Is_Print}', "");
Error('\p{:=	 IS_XPosixPrint}');
Error('\P{:=	 IS_XPosixPrint}');
Expect(1, 1114109, '\p{isxposixprint}', "");
Expect(0, 1114109, '\p{^isxposixprint}', "");
Expect(0, 1114109, '\P{isxposixprint}', "");
Expect(1, 1114109, '\P{^isxposixprint}', "");
Expect(0, 918000, '\p{isxposixprint}', "");
Expect(1, 918000, '\p{^isxposixprint}', "");
Expect(1, 918000, '\P{isxposixprint}', "");
Expect(0, 918000, '\P{^isxposixprint}', "");
Expect(1, 1114109, '\p{	Is_XPOSIXPRINT}', "");
Expect(0, 1114109, '\p{^	Is_XPOSIXPRINT}', "");
Expect(0, 1114109, '\P{	Is_XPOSIXPRINT}', "");
Expect(1, 1114109, '\P{^	Is_XPOSIXPRINT}', "");
Expect(0, 918000, '\p{	Is_XPOSIXPRINT}', "");
Expect(1, 918000, '\p{^	Is_XPOSIXPRINT}', "");
Expect(1, 918000, '\P{	Is_XPOSIXPRINT}', "");
Expect(0, 918000, '\P{^	Is_XPOSIXPRINT}', "");
Error('\p{ Private_use/a/}');
Error('\P{ Private_use/a/}');
Expect(1, 1114109, '\p{privateuse}', "");
Expect(0, 1114109, '\p{^privateuse}', "");
Expect(0, 1114109, '\P{privateuse}', "");
Expect(1, 1114109, '\P{^privateuse}', "");
Expect(0, 63744, '\p{privateuse}', "");
Expect(1, 63744, '\p{^privateuse}', "");
Expect(1, 63744, '\P{privateuse}', "");
Expect(0, 63744, '\P{^privateuse}', "");
Expect(1, 1114109, '\p{-Private_use}', "");
Expect(0, 1114109, '\p{^-Private_use}', "");
Expect(0, 1114109, '\P{-Private_use}', "");
Expect(1, 1114109, '\P{^-Private_use}', "");
Expect(0, 63744, '\p{-Private_use}', "");
Expect(1, 63744, '\p{^-Private_use}', "");
Expect(1, 63744, '\P{-Private_use}', "");
Expect(0, 63744, '\P{^-Private_use}', "");
Error('\p{/a/is_Private_Use}');
Error('\P{/a/is_Private_Use}');
Expect(1, 1114109, '\p{isprivateuse}', "");
Expect(0, 1114109, '\p{^isprivateuse}', "");
Expect(0, 1114109, '\P{isprivateuse}', "");
Expect(1, 1114109, '\P{^isprivateuse}', "");
Expect(0, 63744, '\p{isprivateuse}', "");
Expect(1, 63744, '\p{^isprivateuse}', "");
Expect(1, 63744, '\P{isprivateuse}', "");
Expect(0, 63744, '\P{^isprivateuse}', "");
Expect(1, 1114109, '\p{ -Is_Private_USE}', "");
Expect(0, 1114109, '\p{^ -Is_Private_USE}', "");
Expect(0, 1114109, '\P{ -Is_Private_USE}', "");
Expect(1, 1114109, '\P{^ -Is_Private_USE}', "");
Expect(0, 63744, '\p{ -Is_Private_USE}', "");
Expect(1, 63744, '\p{^ -Is_Private_USE}', "");
Expect(1, 63744, '\P{ -Is_Private_USE}', "");
Expect(0, 63744, '\P{^ -Is_Private_USE}', "");
Error('\p{:= _co}');
Error('\P{:= _co}');
Expect(1, 1114109, '\p{co}', "");
Expect(0, 1114109, '\p{^co}', "");
Expect(0, 1114109, '\P{co}', "");
Expect(1, 1114109, '\P{^co}', "");
Expect(0, 63744, '\p{co}', "");
Expect(1, 63744, '\p{^co}', "");
Expect(1, 63744, '\P{co}', "");
Expect(0, 63744, '\P{^co}', "");
Expect(1, 1114109, '\p{ 	Co}', "");
Expect(0, 1114109, '\p{^ 	Co}', "");
Expect(0, 1114109, '\P{ 	Co}', "");
Expect(1, 1114109, '\P{^ 	Co}', "");
Expect(0, 63744, '\p{ 	Co}', "");
Expect(1, 63744, '\p{^ 	Co}', "");
Expect(1, 63744, '\P{ 	Co}', "");
Expect(0, 63744, '\P{^ 	Co}', "");
Error('\p{/a/ _Is_co}');
Error('\P{/a/ _Is_co}');
Expect(1, 1114109, '\p{isco}', "");
Expect(0, 1114109, '\p{^isco}', "");
Expect(0, 1114109, '\P{isco}', "");
Expect(1, 1114109, '\P{^isco}', "");
Expect(0, 63744, '\p{isco}', "");
Expect(1, 63744, '\p{^isco}', "");
Expect(1, 63744, '\P{isco}', "");
Expect(0, 63744, '\P{^isco}', "");
Expect(1, 1114109, '\p{	Is_CO}', "");
Expect(0, 1114109, '\p{^	Is_CO}', "");
Expect(0, 1114109, '\P{	Is_CO}', "");
Expect(1, 1114109, '\P{^	Is_CO}', "");
Expect(0, 63744, '\p{	Is_CO}', "");
Expect(1, 63744, '\p{^	Is_CO}', "");
Expect(1, 63744, '\P{	Is_CO}', "");
Expect(0, 63744, '\P{^	Is_CO}', "");
Error('\p{_:=PRIVATE_use_area}');
Error('\P{_:=PRIVATE_use_area}');
Expect(1, 63743, '\p{privateusearea}', "");
Expect(0, 63743, '\p{^privateusearea}', "");
Expect(0, 63743, '\P{privateusearea}', "");
Expect(1, 63743, '\P{^privateusearea}', "");
Expect(0, 63744, '\p{privateusearea}', "");
Expect(1, 63744, '\p{^privateusearea}', "");
Expect(1, 63744, '\P{privateusearea}', "");
Expect(0, 63744, '\P{^privateusearea}', "");
Expect(1, 63743, '\p{_ private_Use_area}', "");
Expect(0, 63743, '\p{^_ private_Use_area}', "");
Expect(0, 63743, '\P{_ private_Use_area}', "");
Expect(1, 63743, '\P{^_ private_Use_area}', "");
Expect(0, 63744, '\p{_ private_Use_area}', "");
Expect(1, 63744, '\p{^_ private_Use_area}', "");
Expect(1, 63744, '\P{_ private_Use_area}', "");
Expect(0, 63744, '\P{^_ private_Use_area}', "");
Error('\p{-:=is_Private_Use_AREA}');
Error('\P{-:=is_Private_Use_AREA}');
Expect(1, 63743, '\p{isprivateusearea}', "");
Expect(0, 63743, '\p{^isprivateusearea}', "");
Expect(0, 63743, '\P{isprivateusearea}', "");
Expect(1, 63743, '\P{^isprivateusearea}', "");
Expect(0, 63744, '\p{isprivateusearea}', "");
Expect(1, 63744, '\p{^isprivateusearea}', "");
Expect(1, 63744, '\P{isprivateusearea}', "");
Expect(0, 63744, '\P{^isprivateusearea}', "");
Expect(1, 63743, '\p{-Is_private_use_area}', "");
Expect(0, 63743, '\p{^-Is_private_use_area}', "");
Expect(0, 63743, '\P{-Is_private_use_area}', "");
Expect(1, 63743, '\P{^-Is_private_use_area}', "");
Expect(0, 63744, '\p{-Is_private_use_area}', "");
Expect(1, 63744, '\p{^-Is_private_use_area}', "");
Expect(1, 63744, '\P{-Is_private_use_area}', "");
Expect(0, 63744, '\P{^-Is_private_use_area}', "");
Error('\p{-:=In_PRIVATE_Use_Area}');
Error('\P{-:=In_PRIVATE_Use_Area}');
Expect(1, 63743, '\p{inprivateusearea}', "");
Expect(0, 63743, '\p{^inprivateusearea}', "");
Expect(0, 63743, '\P{inprivateusearea}', "");
Expect(1, 63743, '\P{^inprivateusearea}', "");
Expect(0, 63744, '\p{inprivateusearea}', "");
Expect(1, 63744, '\p{^inprivateusearea}', "");
Expect(1, 63744, '\P{inprivateusearea}', "");
Expect(0, 63744, '\P{^inprivateusearea}', "");
Expect(1, 63743, '\p{		In_private_use_Area}', "");
Expect(0, 63743, '\p{^		In_private_use_Area}', "");
Expect(0, 63743, '\P{		In_private_use_Area}', "");
Expect(1, 63743, '\P{^		In_private_use_Area}', "");
Expect(0, 63744, '\p{		In_private_use_Area}', "");
Expect(1, 63744, '\p{^		In_private_use_Area}', "");
Expect(1, 63744, '\P{		In_private_use_Area}', "");
Expect(0, 63744, '\P{^		In_private_use_Area}', "");
Error('\p{	/a/PUA}');
Error('\P{	/a/PUA}');
Expect(1, 63743, '\p{pua}', "");
Expect(0, 63743, '\p{^pua}', "");
Expect(0, 63743, '\P{pua}', "");
Expect(1, 63743, '\P{^pua}', "");
Expect(0, 63744, '\p{pua}', "");
Expect(1, 63744, '\p{^pua}', "");
Expect(1, 63744, '\P{pua}', "");
Expect(0, 63744, '\P{^pua}', "");
Expect(1, 63743, '\p{_ pua}', "");
Expect(0, 63743, '\p{^_ pua}', "");
Expect(0, 63743, '\P{_ pua}', "");
Expect(1, 63743, '\P{^_ pua}', "");
Expect(0, 63744, '\p{_ pua}', "");
Expect(1, 63744, '\p{^_ pua}', "");
Expect(1, 63744, '\P{_ pua}', "");
Expect(0, 63744, '\P{^_ pua}', "");
Error('\p{/a/ IS_PUA}');
Error('\P{/a/ IS_PUA}');
Expect(1, 63743, '\p{ispua}', "");
Expect(0, 63743, '\p{^ispua}', "");
Expect(0, 63743, '\P{ispua}', "");
Expect(1, 63743, '\P{^ispua}', "");
Expect(0, 63744, '\p{ispua}', "");
Expect(1, 63744, '\p{^ispua}', "");
Expect(1, 63744, '\P{ispua}', "");
Expect(0, 63744, '\P{^ispua}', "");
Expect(1, 63743, '\p{ _IS_PUA}', "");
Expect(0, 63743, '\p{^ _IS_PUA}', "");
Expect(0, 63743, '\P{ _IS_PUA}', "");
Expect(1, 63743, '\P{^ _IS_PUA}', "");
Expect(0, 63744, '\p{ _IS_PUA}', "");
Expect(1, 63744, '\p{^ _IS_PUA}', "");
Expect(1, 63744, '\P{ _IS_PUA}', "");
Expect(0, 63744, '\P{^ _IS_PUA}', "");
Error('\p{_In_PUA/a/}');
Error('\P{_In_PUA/a/}');
Expect(1, 63743, '\p{inpua}', "");
Expect(0, 63743, '\p{^inpua}', "");
Expect(0, 63743, '\P{inpua}', "");
Expect(1, 63743, '\P{^inpua}', "");
Expect(0, 63744, '\p{inpua}', "");
Expect(1, 63744, '\p{^inpua}', "");
Expect(1, 63744, '\P{inpua}', "");
Expect(0, 63744, '\P{^inpua}', "");
Expect(1, 63743, '\p{  in_PUA}', "");
Expect(0, 63743, '\p{^  in_PUA}', "");
Expect(0, 63743, '\P{  in_PUA}', "");
Expect(1, 63743, '\P{^  in_PUA}', "");
Expect(0, 63744, '\p{  in_PUA}', "");
Expect(1, 63744, '\p{^  in_PUA}', "");
Expect(1, 63744, '\P{  in_PUA}', "");
Expect(0, 63744, '\P{^  in_PUA}', "");
Error('\p{-	In_Private_Use:=}');
Error('\P{-	In_Private_Use:=}');
Expect(1, 63743, '\p{inprivateuse}', "");
Expect(0, 63743, '\p{^inprivateuse}', "");
Expect(0, 63743, '\P{inprivateuse}', "");
Expect(1, 63743, '\P{^inprivateuse}', "");
Expect(0, 63744, '\p{inprivateuse}', "");
Expect(1, 63744, '\p{^inprivateuse}', "");
Expect(1, 63744, '\P{inprivateuse}', "");
Expect(0, 63744, '\P{^inprivateuse}', "");
Expect(1, 63743, '\p{_ In_private_Use}', "");
Expect(0, 63743, '\p{^_ In_private_Use}', "");
Expect(0, 63743, '\P{_ In_private_Use}', "");
Expect(1, 63743, '\P{^_ In_private_Use}', "");
Expect(0, 63744, '\p{_ In_private_Use}', "");
Expect(1, 63744, '\p{^_ In_private_Use}', "");
Expect(1, 63744, '\P{_ In_private_Use}', "");
Expect(0, 63744, '\P{^_ In_private_Use}', "");
Error('\p{:= PUNCT}');
Error('\P{:= PUNCT}');
Expect(1, 74867, '\p{punct}', "");
Expect(0, 74867, '\p{^punct}', "");
Expect(0, 74867, '\P{punct}', "");
Expect(1, 74867, '\P{^punct}', "");
Expect(0, 74868, '\p{punct}', "");
Expect(1, 74868, '\p{^punct}', "");
Expect(1, 74868, '\P{punct}', "");
Expect(0, 74868, '\P{^punct}', "");
Expect(1, 74867, '\p{	Punct}', "");
Expect(0, 74867, '\p{^	Punct}', "");
Expect(0, 74867, '\P{	Punct}', "");
Expect(1, 74867, '\P{^	Punct}', "");
Expect(0, 74868, '\p{	Punct}', "");
Expect(1, 74868, '\p{^	Punct}', "");
Expect(1, 74868, '\P{	Punct}', "");
Expect(0, 74868, '\P{^	Punct}', "");
Error('\p{ _IS_PUNCT:=}');
Error('\P{ _IS_PUNCT:=}');
Expect(1, 74867, '\p{ispunct}', "");
Expect(0, 74867, '\p{^ispunct}', "");
Expect(0, 74867, '\P{ispunct}', "");
Expect(1, 74867, '\P{^ispunct}', "");
Expect(0, 74868, '\p{ispunct}', "");
Expect(1, 74868, '\p{^ispunct}', "");
Expect(1, 74868, '\P{ispunct}', "");
Expect(0, 74868, '\P{^ispunct}', "");
Expect(1, 74867, '\p{  IS_Punct}', "");
Expect(0, 74867, '\p{^  IS_Punct}', "");
Expect(0, 74867, '\P{  IS_Punct}', "");
Expect(1, 74867, '\P{^  IS_Punct}', "");
Expect(0, 74868, '\p{  IS_Punct}', "");
Expect(1, 74868, '\p{^  IS_Punct}', "");
Expect(1, 74868, '\P{  IS_Punct}', "");
Expect(0, 74868, '\P{^  IS_Punct}', "");
Error('\p{-:=punctuation}');
Error('\P{-:=punctuation}');
Expect(1, 74867, '\p{punctuation}', "");
Expect(0, 74867, '\p{^punctuation}', "");
Expect(0, 74867, '\P{punctuation}', "");
Expect(1, 74867, '\P{^punctuation}', "");
Expect(0, 74868, '\p{punctuation}', "");
Expect(1, 74868, '\p{^punctuation}', "");
Expect(1, 74868, '\P{punctuation}', "");
Expect(0, 74868, '\P{^punctuation}', "");
Expect(1, 74867, '\p{_	Punctuation}', "");
Expect(0, 74867, '\p{^_	Punctuation}', "");
Expect(0, 74867, '\P{_	Punctuation}', "");
Expect(1, 74867, '\P{^_	Punctuation}', "");
Expect(0, 74868, '\p{_	Punctuation}', "");
Expect(1, 74868, '\p{^_	Punctuation}', "");
Expect(1, 74868, '\P{_	Punctuation}', "");
Expect(0, 74868, '\P{^_	Punctuation}', "");
Error('\p{	/a/IS_PUNCTUATION}');
Error('\P{	/a/IS_PUNCTUATION}');
Expect(1, 74867, '\p{ispunctuation}', "");
Expect(0, 74867, '\p{^ispunctuation}', "");
Expect(0, 74867, '\P{ispunctuation}', "");
Expect(1, 74867, '\P{^ispunctuation}', "");
Expect(0, 74868, '\p{ispunctuation}', "");
Expect(1, 74868, '\p{^ispunctuation}', "");
Expect(1, 74868, '\P{ispunctuation}', "");
Expect(0, 74868, '\P{^ispunctuation}', "");
Expect(1, 74867, '\p{_ Is_punctuation}', "");
Expect(0, 74867, '\p{^_ Is_punctuation}', "");
Expect(0, 74867, '\P{_ Is_punctuation}', "");
Expect(1, 74867, '\P{^_ Is_punctuation}', "");
Expect(0, 74868, '\p{_ Is_punctuation}', "");
Expect(1, 74868, '\p{^_ Is_punctuation}', "");
Expect(1, 74868, '\P{_ Is_punctuation}', "");
Expect(0, 74868, '\P{^_ Is_punctuation}', "");
Error('\p{:=	_P}');
Error('\P{:=	_P}');
Expect(1, 74867, '\p{p}', "");
Expect(0, 74867, '\p{^p}', "");
Expect(0, 74867, '\P{p}', "");
Expect(1, 74867, '\P{^p}', "");
Expect(0, 74868, '\p{p}', "");
Expect(1, 74868, '\p{^p}', "");
Expect(1, 74868, '\P{p}', "");
Expect(0, 74868, '\P{^p}', "");
Expect(1, 74867, '\p{_-P}', "");
Expect(0, 74867, '\p{^_-P}', "");
Expect(0, 74867, '\P{_-P}', "");
Expect(1, 74867, '\P{^_-P}', "");
Expect(0, 74868, '\p{_-P}', "");
Expect(1, 74868, '\p{^_-P}', "");
Expect(1, 74868, '\P{_-P}', "");
Expect(0, 74868, '\P{^_-P}', "");
Error('\p{_/a/is_P}');
Error('\P{_/a/is_P}');
Expect(1, 74867, '\p{isp}', "");
Expect(0, 74867, '\p{^isp}', "");
Expect(0, 74867, '\P{isp}', "");
Expect(1, 74867, '\P{^isp}', "");
Expect(0, 74868, '\p{isp}', "");
Expect(1, 74868, '\p{^isp}', "");
Expect(1, 74868, '\P{isp}', "");
Expect(0, 74868, '\P{^isp}', "");
Expect(1, 74867, '\p{- IS_P}', "");
Expect(0, 74867, '\p{^- IS_P}', "");
Expect(0, 74867, '\P{- IS_P}', "");
Expect(1, 74867, '\P{^- IS_P}', "");
Expect(0, 74868, '\p{- IS_P}', "");
Expect(1, 74868, '\p{^- IS_P}', "");
Expect(1, 74868, '\P{- IS_P}', "");
Expect(0, 74868, '\P{^- IS_P}', "");
Error('\p{-Quotation_Mark:=}');
Error('\P{-Quotation_Mark:=}');
Expect(1, 65379, '\p{quotationmark}', "");
Expect(0, 65379, '\p{^quotationmark}', "");
Expect(0, 65379, '\P{quotationmark}', "");
Expect(1, 65379, '\P{^quotationmark}', "");
Expect(0, 65380, '\p{quotationmark}', "");
Expect(1, 65380, '\p{^quotationmark}', "");
Expect(1, 65380, '\P{quotationmark}', "");
Expect(0, 65380, '\P{^quotationmark}', "");
Expect(1, 65379, '\p{__Quotation_Mark}', "");
Expect(0, 65379, '\p{^__Quotation_Mark}', "");
Expect(0, 65379, '\P{__Quotation_Mark}', "");
Expect(1, 65379, '\P{^__Quotation_Mark}', "");
Expect(0, 65380, '\p{__Quotation_Mark}', "");
Expect(1, 65380, '\p{^__Quotation_Mark}', "");
Expect(1, 65380, '\P{__Quotation_Mark}', "");
Expect(0, 65380, '\P{^__Quotation_Mark}', "");
Error('\p{is_Quotation_Mark:=}');
Error('\P{is_Quotation_Mark:=}');
Expect(1, 65379, '\p{isquotationmark}', "");
Expect(0, 65379, '\p{^isquotationmark}', "");
Expect(0, 65379, '\P{isquotationmark}', "");
Expect(1, 65379, '\P{^isquotationmark}', "");
Expect(0, 65380, '\p{isquotationmark}', "");
Expect(1, 65380, '\p{^isquotationmark}', "");
Expect(1, 65380, '\P{isquotationmark}', "");
Expect(0, 65380, '\P{^isquotationmark}', "");
Expect(1, 65379, '\p{_Is_quotation_Mark}', "");
Expect(0, 65379, '\p{^_Is_quotation_Mark}', "");
Expect(0, 65379, '\P{_Is_quotation_Mark}', "");
Expect(1, 65379, '\P{^_Is_quotation_Mark}', "");
Expect(0, 65380, '\p{_Is_quotation_Mark}', "");
Expect(1, 65380, '\p{^_Is_quotation_Mark}', "");
Expect(1, 65380, '\P{_Is_quotation_Mark}', "");
Expect(0, 65380, '\P{^_Is_quotation_Mark}', "");
Error('\p{:=		QMark}');
Error('\P{:=		QMark}');
Expect(1, 65379, '\p{qmark}', "");
Expect(0, 65379, '\p{^qmark}', "");
Expect(0, 65379, '\P{qmark}', "");
Expect(1, 65379, '\P{^qmark}', "");
Expect(0, 65380, '\p{qmark}', "");
Expect(1, 65380, '\p{^qmark}', "");
Expect(1, 65380, '\P{qmark}', "");
Expect(0, 65380, '\P{^qmark}', "");
Expect(1, 65379, '\p{-	QMark}', "");
Expect(0, 65379, '\p{^-	QMark}', "");
Expect(0, 65379, '\P{-	QMark}', "");
Expect(1, 65379, '\P{^-	QMark}', "");
Expect(0, 65380, '\p{-	QMark}', "");
Expect(1, 65380, '\p{^-	QMark}', "");
Expect(1, 65380, '\P{-	QMark}', "");
Expect(0, 65380, '\P{^-	QMark}', "");
Error('\p{:= is_QMARK}');
Error('\P{:= is_QMARK}');
Expect(1, 65379, '\p{isqmark}', "");
Expect(0, 65379, '\p{^isqmark}', "");
Expect(0, 65379, '\P{isqmark}', "");
Expect(1, 65379, '\P{^isqmark}', "");
Expect(0, 65380, '\p{isqmark}', "");
Expect(1, 65380, '\p{^isqmark}', "");
Expect(1, 65380, '\P{isqmark}', "");
Expect(0, 65380, '\P{^isqmark}', "");
Expect(1, 65379, '\p{_ IS_QMARK}', "");
Expect(0, 65379, '\p{^_ IS_QMARK}', "");
Expect(0, 65379, '\P{_ IS_QMARK}', "");
Expect(1, 65379, '\P{^_ IS_QMARK}', "");
Expect(0, 65380, '\p{_ IS_QMARK}', "");
Expect(1, 65380, '\p{^_ IS_QMARK}', "");
Expect(1, 65380, '\P{_ IS_QMARK}', "");
Expect(0, 65380, '\P{^_ IS_QMARK}', "");
Error('\p{:=_	radical}');
Error('\P{:=_	radical}');
Expect(1, 12245, '\p{radical}', "");
Expect(0, 12245, '\p{^radical}', "");
Expect(0, 12245, '\P{radical}', "");
Expect(1, 12245, '\P{^radical}', "");
Expect(0, 12246, '\p{radical}', "");
Expect(1, 12246, '\p{^radical}', "");
Expect(1, 12246, '\P{radical}', "");
Expect(0, 12246, '\P{^radical}', "");
Expect(1, 12245, '\p{ Radical}', "");
Expect(0, 12245, '\p{^ Radical}', "");
Expect(0, 12245, '\P{ Radical}', "");
Expect(1, 12245, '\P{^ Radical}', "");
Expect(0, 12246, '\p{ Radical}', "");
Expect(1, 12246, '\p{^ Radical}', "");
Expect(1, 12246, '\P{ Radical}', "");
Expect(0, 12246, '\P{^ Radical}', "");
Error('\p{:=--Is_Radical}');
Error('\P{:=--Is_Radical}');
Expect(1, 12245, '\p{isradical}', "");
Expect(0, 12245, '\p{^isradical}', "");
Expect(0, 12245, '\P{isradical}', "");
Expect(1, 12245, '\P{^isradical}', "");
Expect(0, 12246, '\p{isradical}', "");
Expect(1, 12246, '\p{^isradical}', "");
Expect(1, 12246, '\P{isradical}', "");
Expect(0, 12246, '\P{^isradical}', "");
Expect(1, 12245, '\p{-is_Radical}', "");
Expect(0, 12245, '\p{^-is_Radical}', "");
Expect(0, 12245, '\P{-is_Radical}', "");
Expect(1, 12245, '\P{^-is_Radical}', "");
Expect(0, 12246, '\p{-is_Radical}', "");
Expect(1, 12246, '\p{^-is_Radical}', "");
Expect(1, 12246, '\P{-is_Radical}', "");
Expect(0, 12246, '\P{^-is_Radical}', "");
Error('\p{_rejang:=}');
Error('\P{_rejang:=}');
Expect(1, 43359, '\p{rejang}', "");
Expect(0, 43359, '\p{^rejang}', "");
Expect(0, 43359, '\P{rejang}', "");
Expect(1, 43359, '\P{^rejang}', "");
Expect(0, 43360, '\p{rejang}', "");
Expect(1, 43360, '\p{^rejang}', "");
Expect(1, 43360, '\P{rejang}', "");
Expect(0, 43360, '\P{^rejang}', "");
Expect(1, 43359, '\p{- REJANG}', "");
Expect(0, 43359, '\p{^- REJANG}', "");
Expect(0, 43359, '\P{- REJANG}', "");
Expect(1, 43359, '\P{^- REJANG}', "");
Expect(0, 43360, '\p{- REJANG}', "");
Expect(1, 43360, '\p{^- REJANG}', "");
Expect(1, 43360, '\P{- REJANG}', "");
Expect(0, 43360, '\P{^- REJANG}', "");
Error('\p{is_Rejang/a/}');
Error('\P{is_Rejang/a/}');
Expect(1, 43359, '\p{isrejang}', "");
Expect(0, 43359, '\p{^isrejang}', "");
Expect(0, 43359, '\P{isrejang}', "");
Expect(1, 43359, '\P{^isrejang}', "");
Expect(0, 43360, '\p{isrejang}', "");
Expect(1, 43360, '\p{^isrejang}', "");
Expect(1, 43360, '\P{isrejang}', "");
Expect(0, 43360, '\P{^isrejang}', "");
Expect(1, 43359, '\p{	_IS_rejang}', "");
Expect(0, 43359, '\p{^	_IS_rejang}', "");
Expect(0, 43359, '\P{	_IS_rejang}', "");
Expect(1, 43359, '\P{^	_IS_rejang}', "");
Expect(0, 43360, '\p{	_IS_rejang}', "");
Expect(1, 43360, '\p{^	_IS_rejang}', "");
Expect(1, 43360, '\P{	_IS_rejang}', "");
Expect(0, 43360, '\P{^	_IS_rejang}', "");
Error('\p{- Rjng/a/}');
Error('\P{- Rjng/a/}');
Expect(1, 43359, '\p{rjng}', "");
Expect(0, 43359, '\p{^rjng}', "");
Expect(0, 43359, '\P{rjng}', "");
Expect(1, 43359, '\P{^rjng}', "");
Expect(0, 43360, '\p{rjng}', "");
Expect(1, 43360, '\p{^rjng}', "");
Expect(1, 43360, '\P{rjng}', "");
Expect(0, 43360, '\P{^rjng}', "");
Expect(1, 43359, '\p{		Rjng}', "");
Expect(0, 43359, '\p{^		Rjng}', "");
Expect(0, 43359, '\P{		Rjng}', "");
Expect(1, 43359, '\P{^		Rjng}', "");
Expect(0, 43360, '\p{		Rjng}', "");
Expect(1, 43360, '\p{^		Rjng}', "");
Expect(1, 43360, '\P{		Rjng}', "");
Expect(0, 43360, '\P{^		Rjng}', "");
Error('\p{  Is_Rjng/a/}');
Error('\P{  Is_Rjng/a/}');
Expect(1, 43359, '\p{isrjng}', "");
Expect(0, 43359, '\p{^isrjng}', "");
Expect(0, 43359, '\P{isrjng}', "");
Expect(1, 43359, '\P{^isrjng}', "");
Expect(0, 43360, '\p{isrjng}', "");
Expect(1, 43360, '\p{^isrjng}', "");
Expect(1, 43360, '\P{isrjng}', "");
Expect(0, 43360, '\P{^isrjng}', "");
Expect(1, 43359, '\p{-IS_Rjng}', "");
Expect(0, 43359, '\p{^-IS_Rjng}', "");
Expect(0, 43359, '\P{-IS_Rjng}', "");
Expect(1, 43359, '\P{^-IS_Rjng}', "");
Expect(0, 43360, '\p{-IS_Rjng}', "");
Expect(1, 43360, '\p{^-IS_Rjng}', "");
Expect(1, 43360, '\P{-IS_Rjng}', "");
Expect(0, 43360, '\P{^-IS_Rjng}', "");
Error('\p{/a/ 	Rumi_numeral_symbols}');
Error('\P{/a/ 	Rumi_numeral_symbols}');
Expect(1, 69247, '\p{ruminumeralsymbols}', "");
Expect(0, 69247, '\p{^ruminumeralsymbols}', "");
Expect(0, 69247, '\P{ruminumeralsymbols}', "");
Expect(1, 69247, '\P{^ruminumeralsymbols}', "");
Expect(0, 69248, '\p{ruminumeralsymbols}', "");
Expect(1, 69248, '\p{^ruminumeralsymbols}', "");
Expect(1, 69248, '\P{ruminumeralsymbols}', "");
Expect(0, 69248, '\P{^ruminumeralsymbols}', "");
Expect(1, 69247, '\p{	_rumi_Numeral_Symbols}', "");
Expect(0, 69247, '\p{^	_rumi_Numeral_Symbols}', "");
Expect(0, 69247, '\P{	_rumi_Numeral_Symbols}', "");
Expect(1, 69247, '\P{^	_rumi_Numeral_Symbols}', "");
Expect(0, 69248, '\p{	_rumi_Numeral_Symbols}', "");
Expect(1, 69248, '\p{^	_rumi_Numeral_Symbols}', "");
Expect(1, 69248, '\P{	_rumi_Numeral_Symbols}', "");
Expect(0, 69248, '\P{^	_rumi_Numeral_Symbols}', "");
Error('\p{ /a/IS_RUMI_Numeral_Symbols}');
Error('\P{ /a/IS_RUMI_Numeral_Symbols}');
Expect(1, 69247, '\p{isruminumeralsymbols}', "");
Expect(0, 69247, '\p{^isruminumeralsymbols}', "");
Expect(0, 69247, '\P{isruminumeralsymbols}', "");
Expect(1, 69247, '\P{^isruminumeralsymbols}', "");
Expect(0, 69248, '\p{isruminumeralsymbols}', "");
Expect(1, 69248, '\p{^isruminumeralsymbols}', "");
Expect(1, 69248, '\P{isruminumeralsymbols}', "");
Expect(0, 69248, '\P{^isruminumeralsymbols}', "");
Expect(1, 69247, '\p{ 	is_rumi_numeral_Symbols}', "");
Expect(0, 69247, '\p{^ 	is_rumi_numeral_Symbols}', "");
Expect(0, 69247, '\P{ 	is_rumi_numeral_Symbols}', "");
Expect(1, 69247, '\P{^ 	is_rumi_numeral_Symbols}', "");
Expect(0, 69248, '\p{ 	is_rumi_numeral_Symbols}', "");
Expect(1, 69248, '\p{^ 	is_rumi_numeral_Symbols}', "");
Expect(1, 69248, '\P{ 	is_rumi_numeral_Symbols}', "");
Expect(0, 69248, '\P{^ 	is_rumi_numeral_Symbols}', "");
Error('\p{/a/ 	IN_RUMI_Numeral_symbols}');
Error('\P{/a/ 	IN_RUMI_Numeral_symbols}');
Expect(1, 69247, '\p{inruminumeralsymbols}', "");
Expect(0, 69247, '\p{^inruminumeralsymbols}', "");
Expect(0, 69247, '\P{inruminumeralsymbols}', "");
Expect(1, 69247, '\P{^inruminumeralsymbols}', "");
Expect(0, 69248, '\p{inruminumeralsymbols}', "");
Expect(1, 69248, '\p{^inruminumeralsymbols}', "");
Expect(1, 69248, '\P{inruminumeralsymbols}', "");
Expect(0, 69248, '\P{^inruminumeralsymbols}', "");
Expect(1, 69247, '\p{	 IN_rumi_Numeral_symbols}', "");
Expect(0, 69247, '\p{^	 IN_rumi_Numeral_symbols}', "");
Expect(0, 69247, '\P{	 IN_rumi_Numeral_symbols}', "");
Expect(1, 69247, '\P{^	 IN_rumi_Numeral_symbols}', "");
Expect(0, 69248, '\p{	 IN_rumi_Numeral_symbols}', "");
Expect(1, 69248, '\p{^	 IN_rumi_Numeral_symbols}', "");
Expect(1, 69248, '\P{	 IN_rumi_Numeral_symbols}', "");
Expect(0, 69248, '\P{^	 IN_rumi_Numeral_symbols}', "");
Error('\p{	-Rumi/a/}');
Error('\P{	-Rumi/a/}');
Expect(1, 69247, '\p{rumi}', "");
Expect(0, 69247, '\p{^rumi}', "");
Expect(0, 69247, '\P{rumi}', "");
Expect(1, 69247, '\P{^rumi}', "");
Expect(0, 69248, '\p{rumi}', "");
Expect(1, 69248, '\p{^rumi}', "");
Expect(1, 69248, '\P{rumi}', "");
Expect(0, 69248, '\P{^rumi}', "");
Expect(1, 69247, '\p{	Rumi}', "");
Expect(0, 69247, '\p{^	Rumi}', "");
Expect(0, 69247, '\P{	Rumi}', "");
Expect(1, 69247, '\P{^	Rumi}', "");
Expect(0, 69248, '\p{	Rumi}', "");
Expect(1, 69248, '\p{^	Rumi}', "");
Expect(1, 69248, '\P{	Rumi}', "");
Expect(0, 69248, '\P{^	Rumi}', "");
Error('\p{_/a/is_RUMI}');
Error('\P{_/a/is_RUMI}');
Expect(1, 69247, '\p{isrumi}', "");
Expect(0, 69247, '\p{^isrumi}', "");
Expect(0, 69247, '\P{isrumi}', "");
Expect(1, 69247, '\P{^isrumi}', "");
Expect(0, 69248, '\p{isrumi}', "");
Expect(1, 69248, '\p{^isrumi}', "");
Expect(1, 69248, '\P{isrumi}', "");
Expect(0, 69248, '\P{^isrumi}', "");
Expect(1, 69247, '\p{	 Is_rumi}', "");
Expect(0, 69247, '\p{^	 Is_rumi}', "");
Expect(0, 69247, '\P{	 Is_rumi}', "");
Expect(1, 69247, '\P{^	 Is_rumi}', "");
Expect(0, 69248, '\p{	 Is_rumi}', "");
Expect(1, 69248, '\p{^	 Is_rumi}', "");
Expect(1, 69248, '\P{	 Is_rumi}', "");
Expect(0, 69248, '\P{^	 Is_rumi}', "");
Error('\p{/a/	In_Rumi}');
Error('\P{/a/	In_Rumi}');
Expect(1, 69247, '\p{inrumi}', "");
Expect(0, 69247, '\p{^inrumi}', "");
Expect(0, 69247, '\P{inrumi}', "");
Expect(1, 69247, '\P{^inrumi}', "");
Expect(0, 69248, '\p{inrumi}', "");
Expect(1, 69248, '\p{^inrumi}', "");
Expect(1, 69248, '\P{inrumi}', "");
Expect(0, 69248, '\P{^inrumi}', "");
Expect(1, 69247, '\p{--in_rumi}', "");
Expect(0, 69247, '\p{^--in_rumi}', "");
Expect(0, 69247, '\P{--in_rumi}', "");
Expect(1, 69247, '\P{^--in_rumi}', "");
Expect(0, 69248, '\p{--in_rumi}', "");
Expect(1, 69248, '\p{^--in_rumi}', "");
Expect(1, 69248, '\P{--in_rumi}', "");
Expect(0, 69248, '\P{^--in_rumi}', "");
Error('\p{	runic/a/}');
Error('\P{	runic/a/}');
Expect(1, 5872, '\p{runic}', "");
Expect(0, 5872, '\p{^runic}', "");
Expect(0, 5872, '\P{runic}', "");
Expect(1, 5872, '\P{^runic}', "");
Expect(0, 5873, '\p{runic}', "");
Expect(1, 5873, '\p{^runic}', "");
Expect(1, 5873, '\P{runic}', "");
Expect(0, 5873, '\P{^runic}', "");
Expect(1, 5872, '\p{	_runic}', "");
Expect(0, 5872, '\p{^	_runic}', "");
Expect(0, 5872, '\P{	_runic}', "");
Expect(1, 5872, '\P{^	_runic}', "");
Expect(0, 5873, '\p{	_runic}', "");
Expect(1, 5873, '\p{^	_runic}', "");
Expect(1, 5873, '\P{	_runic}', "");
Expect(0, 5873, '\P{^	_runic}', "");
Error('\p{:=IS_Runic}');
Error('\P{:=IS_Runic}');
Expect(1, 5872, '\p{isrunic}', "");
Expect(0, 5872, '\p{^isrunic}', "");
Expect(0, 5872, '\P{isrunic}', "");
Expect(1, 5872, '\P{^isrunic}', "");
Expect(0, 5873, '\p{isrunic}', "");
Expect(1, 5873, '\p{^isrunic}', "");
Expect(1, 5873, '\P{isrunic}', "");
Expect(0, 5873, '\P{^isrunic}', "");
Expect(1, 5872, '\p{ is_Runic}', "");
Expect(0, 5872, '\p{^ is_Runic}', "");
Expect(0, 5872, '\P{ is_Runic}', "");
Expect(1, 5872, '\P{^ is_Runic}', "");
Expect(0, 5873, '\p{ is_Runic}', "");
Expect(1, 5873, '\p{^ is_Runic}', "");
Expect(1, 5873, '\P{ is_Runic}', "");
Expect(0, 5873, '\P{^ is_Runic}', "");
Error('\p{	:=runr}');
Error('\P{	:=runr}');
Expect(1, 5872, '\p{runr}', "");
Expect(0, 5872, '\p{^runr}', "");
Expect(0, 5872, '\P{runr}', "");
Expect(1, 5872, '\P{^runr}', "");
Expect(0, 5873, '\p{runr}', "");
Expect(1, 5873, '\p{^runr}', "");
Expect(1, 5873, '\P{runr}', "");
Expect(0, 5873, '\P{^runr}', "");
Expect(1, 5872, '\p{	_runr}', "");
Expect(0, 5872, '\p{^	_runr}', "");
Expect(0, 5872, '\P{	_runr}', "");
Expect(1, 5872, '\P{^	_runr}', "");
Expect(0, 5873, '\p{	_runr}', "");
Expect(1, 5873, '\p{^	_runr}', "");
Expect(1, 5873, '\P{	_runr}', "");
Expect(0, 5873, '\P{^	_runr}', "");
Error('\p{_IS_RUNR/a/}');
Error('\P{_IS_RUNR/a/}');
Expect(1, 5872, '\p{isrunr}', "");
Expect(0, 5872, '\p{^isrunr}', "");
Expect(0, 5872, '\P{isrunr}', "");
Expect(1, 5872, '\P{^isrunr}', "");
Expect(0, 5873, '\p{isrunr}', "");
Expect(1, 5873, '\p{^isrunr}', "");
Expect(1, 5873, '\P{isrunr}', "");
Expect(0, 5873, '\P{^isrunr}', "");
Expect(1, 5872, '\p{_IS_Runr}', "");
Expect(0, 5872, '\p{^_IS_Runr}', "");
Expect(0, 5872, '\P{_IS_Runr}', "");
Expect(1, 5872, '\P{^_IS_Runr}', "");
Expect(0, 5873, '\p{_IS_Runr}', "");
Expect(1, 5873, '\p{^_IS_Runr}', "");
Expect(1, 5873, '\P{_IS_Runr}', "");
Expect(0, 5873, '\P{^_IS_Runr}', "");
Error('\p{_sterm:=}');
Error('\P{_sterm:=}');
Expect(1, 70086, '\p{sterm}', "");
Expect(0, 70086, '\p{^sterm}', "");
Expect(0, 70086, '\P{sterm}', "");
Expect(1, 70086, '\P{^sterm}', "");
Expect(0, 70087, '\p{sterm}', "");
Expect(1, 70087, '\p{^sterm}', "");
Expect(1, 70087, '\P{sterm}', "");
Expect(0, 70087, '\P{^sterm}', "");
Expect(1, 70086, '\p{		STerm}', "");
Expect(0, 70086, '\p{^		STerm}', "");
Expect(0, 70086, '\P{		STerm}', "");
Expect(1, 70086, '\P{^		STerm}', "");
Expect(0, 70087, '\p{		STerm}', "");
Expect(1, 70087, '\p{^		STerm}', "");
Expect(1, 70087, '\P{		STerm}', "");
Expect(0, 70087, '\P{^		STerm}', "");
Error('\p{ IS_STERM:=}');
Error('\P{ IS_STERM:=}');
Expect(1, 70086, '\p{issterm}', "");
Expect(0, 70086, '\p{^issterm}', "");
Expect(0, 70086, '\P{issterm}', "");
Expect(1, 70086, '\P{^issterm}', "");
Expect(0, 70087, '\p{issterm}', "");
Expect(1, 70087, '\p{^issterm}', "");
Expect(1, 70087, '\P{issterm}', "");
Expect(0, 70087, '\P{^issterm}', "");
Expect(1, 70086, '\p{-Is_STerm}', "");
Expect(0, 70086, '\p{^-Is_STerm}', "");
Expect(0, 70086, '\P{-Is_STerm}', "");
Expect(1, 70086, '\P{^-Is_STerm}', "");
Expect(0, 70087, '\p{-Is_STerm}', "");
Expect(1, 70087, '\p{^-Is_STerm}', "");
Expect(1, 70087, '\P{-Is_STerm}', "");
Expect(0, 70087, '\P{^-Is_STerm}', "");
Error('\p{	Samaritan:=}');
Error('\P{	Samaritan:=}');
Expect(1, 2110, '\p{samaritan}', "");
Expect(0, 2110, '\p{^samaritan}', "");
Expect(0, 2110, '\P{samaritan}', "");
Expect(1, 2110, '\P{^samaritan}', "");
Expect(0, 2111, '\p{samaritan}', "");
Expect(1, 2111, '\p{^samaritan}', "");
Expect(1, 2111, '\P{samaritan}', "");
Expect(0, 2111, '\P{^samaritan}', "");
Expect(1, 2110, '\p{		Samaritan}', "");
Expect(0, 2110, '\p{^		Samaritan}', "");
Expect(0, 2110, '\P{		Samaritan}', "");
Expect(1, 2110, '\P{^		Samaritan}', "");
Expect(0, 2111, '\p{		Samaritan}', "");
Expect(1, 2111, '\p{^		Samaritan}', "");
Expect(1, 2111, '\P{		Samaritan}', "");
Expect(0, 2111, '\P{^		Samaritan}', "");
Error('\p{/a/  is_samaritan}');
Error('\P{/a/  is_samaritan}');
Expect(1, 2110, '\p{issamaritan}', "");
Expect(0, 2110, '\p{^issamaritan}', "");
Expect(0, 2110, '\P{issamaritan}', "");
Expect(1, 2110, '\P{^issamaritan}', "");
Expect(0, 2111, '\p{issamaritan}', "");
Expect(1, 2111, '\p{^issamaritan}', "");
Expect(1, 2111, '\P{issamaritan}', "");
Expect(0, 2111, '\P{^issamaritan}', "");
Expect(1, 2110, '\p{	 IS_Samaritan}', "");
Expect(0, 2110, '\p{^	 IS_Samaritan}', "");
Expect(0, 2110, '\P{	 IS_Samaritan}', "");
Expect(1, 2110, '\P{^	 IS_Samaritan}', "");
Expect(0, 2111, '\p{	 IS_Samaritan}', "");
Expect(1, 2111, '\p{^	 IS_Samaritan}', "");
Expect(1, 2111, '\P{	 IS_Samaritan}', "");
Expect(0, 2111, '\P{^	 IS_Samaritan}', "");
Error('\p{_/a/SAMR}');
Error('\P{_/a/SAMR}');
Expect(1, 2110, '\p{samr}', "");
Expect(0, 2110, '\p{^samr}', "");
Expect(0, 2110, '\P{samr}', "");
Expect(1, 2110, '\P{^samr}', "");
Expect(0, 2111, '\p{samr}', "");
Expect(1, 2111, '\p{^samr}', "");
Expect(1, 2111, '\P{samr}', "");
Expect(0, 2111, '\P{^samr}', "");
Expect(1, 2110, '\p{ SAMR}', "");
Expect(0, 2110, '\p{^ SAMR}', "");
Expect(0, 2110, '\P{ SAMR}', "");
Expect(1, 2110, '\P{^ SAMR}', "");
Expect(0, 2111, '\p{ SAMR}', "");
Expect(1, 2111, '\p{^ SAMR}', "");
Expect(1, 2111, '\P{ SAMR}', "");
Expect(0, 2111, '\P{^ SAMR}', "");
Error('\p{/a/-_is_samr}');
Error('\P{/a/-_is_samr}');
Expect(1, 2110, '\p{issamr}', "");
Expect(0, 2110, '\p{^issamr}', "");
Expect(0, 2110, '\P{issamr}', "");
Expect(1, 2110, '\P{^issamr}', "");
Expect(0, 2111, '\p{issamr}', "");
Expect(1, 2111, '\p{^issamr}', "");
Expect(1, 2111, '\P{issamr}', "");
Expect(0, 2111, '\P{^issamr}', "");
Expect(1, 2110, '\p{_IS_samr}', "");
Expect(0, 2110, '\p{^_IS_samr}', "");
Expect(0, 2110, '\P{_IS_samr}', "");
Expect(1, 2110, '\P{^_IS_samr}', "");
Expect(0, 2111, '\p{_IS_samr}', "");
Expect(1, 2111, '\p{^_IS_samr}', "");
Expect(1, 2111, '\P{_IS_samr}', "");
Expect(0, 2111, '\P{^_IS_samr}', "");
Error('\p{:=__Saurashtra}');
Error('\P{:=__Saurashtra}');
Expect(1, 43225, '\p{saurashtra}', "");
Expect(0, 43225, '\p{^saurashtra}', "");
Expect(0, 43225, '\P{saurashtra}', "");
Expect(1, 43225, '\P{^saurashtra}', "");
Expect(0, 43226, '\p{saurashtra}', "");
Expect(1, 43226, '\p{^saurashtra}', "");
Expect(1, 43226, '\P{saurashtra}', "");
Expect(0, 43226, '\P{^saurashtra}', "");
Expect(1, 43225, '\p{		Saurashtra}', "");
Expect(0, 43225, '\p{^		Saurashtra}', "");
Expect(0, 43225, '\P{		Saurashtra}', "");
Expect(1, 43225, '\P{^		Saurashtra}', "");
Expect(0, 43226, '\p{		Saurashtra}', "");
Expect(1, 43226, '\p{^		Saurashtra}', "");
Expect(1, 43226, '\P{		Saurashtra}', "");
Expect(0, 43226, '\P{^		Saurashtra}', "");
Error('\p{:=_ is_Saurashtra}');
Error('\P{:=_ is_Saurashtra}');
Expect(1, 43225, '\p{issaurashtra}', "");
Expect(0, 43225, '\p{^issaurashtra}', "");
Expect(0, 43225, '\P{issaurashtra}', "");
Expect(1, 43225, '\P{^issaurashtra}', "");
Expect(0, 43226, '\p{issaurashtra}', "");
Expect(1, 43226, '\p{^issaurashtra}', "");
Expect(1, 43226, '\P{issaurashtra}', "");
Expect(0, 43226, '\P{^issaurashtra}', "");
Expect(1, 43225, '\p{_Is_saurashtra}', "");
Expect(0, 43225, '\p{^_Is_saurashtra}', "");
Expect(0, 43225, '\P{_Is_saurashtra}', "");
Expect(1, 43225, '\P{^_Is_saurashtra}', "");
Expect(0, 43226, '\p{_Is_saurashtra}', "");
Expect(1, 43226, '\p{^_Is_saurashtra}', "");
Expect(1, 43226, '\P{_Is_saurashtra}', "");
Expect(0, 43226, '\P{^_Is_saurashtra}', "");
Error('\p{	:=Saur}');
Error('\P{	:=Saur}');
Expect(1, 43225, '\p{saur}', "");
Expect(0, 43225, '\p{^saur}', "");
Expect(0, 43225, '\P{saur}', "");
Expect(1, 43225, '\P{^saur}', "");
Expect(0, 43226, '\p{saur}', "");
Expect(1, 43226, '\p{^saur}', "");
Expect(1, 43226, '\P{saur}', "");
Expect(0, 43226, '\P{^saur}', "");
Expect(1, 43225, '\p{ 	saur}', "");
Expect(0, 43225, '\p{^ 	saur}', "");
Expect(0, 43225, '\P{ 	saur}', "");
Expect(1, 43225, '\P{^ 	saur}', "");
Expect(0, 43226, '\p{ 	saur}', "");
Expect(1, 43226, '\p{^ 	saur}', "");
Expect(1, 43226, '\P{ 	saur}', "");
Expect(0, 43226, '\P{^ 	saur}', "");
Error('\p{- is_saur:=}');
Error('\P{- is_saur:=}');
Expect(1, 43225, '\p{issaur}', "");
Expect(0, 43225, '\p{^issaur}', "");
Expect(0, 43225, '\P{issaur}', "");
Expect(1, 43225, '\P{^issaur}', "");
Expect(0, 43226, '\p{issaur}', "");
Expect(1, 43226, '\p{^issaur}', "");
Expect(1, 43226, '\P{issaur}', "");
Expect(0, 43226, '\P{^issaur}', "");
Expect(1, 43225, '\p{- Is_SAUR}', "");
Expect(0, 43225, '\p{^- Is_SAUR}', "");
Expect(0, 43225, '\P{- Is_SAUR}', "");
Expect(1, 43225, '\P{^- Is_SAUR}', "");
Expect(0, 43226, '\p{- Is_SAUR}', "");
Expect(1, 43226, '\p{^- Is_SAUR}', "");
Expect(1, 43226, '\P{- Is_SAUR}', "");
Expect(0, 43226, '\P{^- Is_SAUR}', "");
Error('\p{:=		Separator}');
Error('\P{:=		Separator}');
Expect(1, 12288, '\p{separator}', "");
Expect(0, 12288, '\p{^separator}', "");
Expect(0, 12288, '\P{separator}', "");
Expect(1, 12288, '\P{^separator}', "");
Expect(0, 12289, '\p{separator}', "");
Expect(1, 12289, '\p{^separator}', "");
Expect(1, 12289, '\P{separator}', "");
Expect(0, 12289, '\P{^separator}', "");
Expect(1, 12288, '\p{	_separator}', "");
Expect(0, 12288, '\p{^	_separator}', "");
Expect(0, 12288, '\P{	_separator}', "");
Expect(1, 12288, '\P{^	_separator}', "");
Expect(0, 12289, '\p{	_separator}', "");
Expect(1, 12289, '\p{^	_separator}', "");
Expect(1, 12289, '\P{	_separator}', "");
Expect(0, 12289, '\P{^	_separator}', "");
Error('\p{	IS_Separator/a/}');
Error('\P{	IS_Separator/a/}');
Expect(1, 12288, '\p{isseparator}', "");
Expect(0, 12288, '\p{^isseparator}', "");
Expect(0, 12288, '\P{isseparator}', "");
Expect(1, 12288, '\P{^isseparator}', "");
Expect(0, 12289, '\p{isseparator}', "");
Expect(1, 12289, '\p{^isseparator}', "");
Expect(1, 12289, '\P{isseparator}', "");
Expect(0, 12289, '\P{^isseparator}', "");
Expect(1, 12288, '\p{		IS_Separator}', "");
Expect(0, 12288, '\p{^		IS_Separator}', "");
Expect(0, 12288, '\P{		IS_Separator}', "");
Expect(1, 12288, '\P{^		IS_Separator}', "");
Expect(0, 12289, '\p{		IS_Separator}', "");
Expect(1, 12289, '\p{^		IS_Separator}', "");
Expect(1, 12289, '\P{		IS_Separator}', "");
Expect(0, 12289, '\P{^		IS_Separator}', "");
Error('\p{:=_Z}');
Error('\P{:=_Z}');
Expect(1, 12288, '\p{z}', "");
Expect(0, 12288, '\p{^z}', "");
Expect(0, 12288, '\P{z}', "");
Expect(1, 12288, '\P{^z}', "");
Expect(0, 12289, '\p{z}', "");
Expect(1, 12289, '\p{^z}', "");
Expect(1, 12289, '\P{z}', "");
Expect(0, 12289, '\P{^z}', "");
Expect(1, 12288, '\p{ 	Z}', "");
Expect(0, 12288, '\p{^ 	Z}', "");
Expect(0, 12288, '\P{ 	Z}', "");
Expect(1, 12288, '\P{^ 	Z}', "");
Expect(0, 12289, '\p{ 	Z}', "");
Expect(1, 12289, '\p{^ 	Z}', "");
Expect(1, 12289, '\P{ 	Z}', "");
Expect(0, 12289, '\P{^ 	Z}', "");
Error('\p{		Is_Z:=}');
Error('\P{		Is_Z:=}');
Expect(1, 12288, '\p{isz}', "");
Expect(0, 12288, '\p{^isz}', "");
Expect(0, 12288, '\P{isz}', "");
Expect(1, 12288, '\P{^isz}', "");
Expect(0, 12289, '\p{isz}', "");
Expect(1, 12289, '\p{^isz}', "");
Expect(1, 12289, '\P{isz}', "");
Expect(0, 12289, '\P{^isz}', "");
Expect(1, 12288, '\p{ 	is_Z}', "");
Expect(0, 12288, '\p{^ 	is_Z}', "");
Expect(0, 12288, '\P{ 	is_Z}', "");
Expect(1, 12288, '\P{^ 	is_Z}', "");
Expect(0, 12289, '\p{ 	is_Z}', "");
Expect(1, 12289, '\p{^ 	is_Z}', "");
Expect(1, 12289, '\P{ 	is_Z}', "");
Expect(0, 12289, '\P{^ 	is_Z}', "");
Error('\p{-Sharada/a/}');
Error('\P{-Sharada/a/}');
Expect(1, 70105, '\p{sharada}', "");
Expect(0, 70105, '\p{^sharada}', "");
Expect(0, 70105, '\P{sharada}', "");
Expect(1, 70105, '\P{^sharada}', "");
Expect(0, 70106, '\p{sharada}', "");
Expect(1, 70106, '\p{^sharada}', "");
Expect(1, 70106, '\P{sharada}', "");
Expect(0, 70106, '\P{^sharada}', "");
Expect(1, 70105, '\p{-	Sharada}', "");
Expect(0, 70105, '\p{^-	Sharada}', "");
Expect(0, 70105, '\P{-	Sharada}', "");
Expect(1, 70105, '\P{^-	Sharada}', "");
Expect(0, 70106, '\p{-	Sharada}', "");
Expect(1, 70106, '\p{^-	Sharada}', "");
Expect(1, 70106, '\P{-	Sharada}', "");
Expect(0, 70106, '\P{^-	Sharada}', "");
Error('\p{/a/__Is_sharada}');
Error('\P{/a/__Is_sharada}');
Expect(1, 70105, '\p{issharada}', "");
Expect(0, 70105, '\p{^issharada}', "");
Expect(0, 70105, '\P{issharada}', "");
Expect(1, 70105, '\P{^issharada}', "");
Expect(0, 70106, '\p{issharada}', "");
Expect(1, 70106, '\p{^issharada}', "");
Expect(1, 70106, '\P{issharada}', "");
Expect(0, 70106, '\P{^issharada}', "");
Expect(1, 70105, '\p{	_IS_Sharada}', "");
Expect(0, 70105, '\p{^	_IS_Sharada}', "");
Expect(0, 70105, '\P{	_IS_Sharada}', "");
Expect(1, 70105, '\P{^	_IS_Sharada}', "");
Expect(0, 70106, '\p{	_IS_Sharada}', "");
Expect(1, 70106, '\p{^	_IS_Sharada}', "");
Expect(1, 70106, '\P{	_IS_Sharada}', "");
Expect(0, 70106, '\P{^	_IS_Sharada}', "");
Error('\p{-:=shrd}');
Error('\P{-:=shrd}');
Expect(1, 70105, '\p{shrd}', "");
Expect(0, 70105, '\p{^shrd}', "");
Expect(0, 70105, '\P{shrd}', "");
Expect(1, 70105, '\P{^shrd}', "");
Expect(0, 70106, '\p{shrd}', "");
Expect(1, 70106, '\p{^shrd}', "");
Expect(1, 70106, '\P{shrd}', "");
Expect(0, 70106, '\P{^shrd}', "");
Expect(1, 70105, '\p{-Shrd}', "");
Expect(0, 70105, '\p{^-Shrd}', "");
Expect(0, 70105, '\P{-Shrd}', "");
Expect(1, 70105, '\P{^-Shrd}', "");
Expect(0, 70106, '\p{-Shrd}', "");
Expect(1, 70106, '\p{^-Shrd}', "");
Expect(1, 70106, '\P{-Shrd}', "");
Expect(0, 70106, '\P{^-Shrd}', "");
Error('\p{/a/-IS_shrd}');
Error('\P{/a/-IS_shrd}');
Expect(1, 70105, '\p{isshrd}', "");
Expect(0, 70105, '\p{^isshrd}', "");
Expect(0, 70105, '\P{isshrd}', "");
Expect(1, 70105, '\P{^isshrd}', "");
Expect(0, 70106, '\p{isshrd}', "");
Expect(1, 70106, '\p{^isshrd}', "");
Expect(1, 70106, '\P{isshrd}', "");
Expect(0, 70106, '\P{^isshrd}', "");
Expect(1, 70105, '\p{		Is_shrd}', "");
Expect(0, 70105, '\p{^		Is_shrd}', "");
Expect(0, 70105, '\P{		Is_shrd}', "");
Expect(1, 70105, '\P{^		Is_shrd}', "");
Expect(0, 70106, '\p{		Is_shrd}', "");
Expect(1, 70106, '\p{^		Is_shrd}', "");
Expect(1, 70106, '\P{		Is_shrd}', "");
Expect(0, 70106, '\P{^		Is_shrd}', "");
Error('\p{ Shavian/a/}');
Error('\P{ Shavian/a/}');
Expect(1, 66687, '\p{shavian}', "");
Expect(0, 66687, '\p{^shavian}', "");
Expect(0, 66687, '\P{shavian}', "");
Expect(1, 66687, '\P{^shavian}', "");
Expect(0, 66688, '\p{shavian}', "");
Expect(1, 66688, '\p{^shavian}', "");
Expect(1, 66688, '\P{shavian}', "");
Expect(0, 66688, '\P{^shavian}', "");
Expect(1, 66687, '\p{Shavian}', "");
Expect(0, 66687, '\p{^Shavian}', "");
Expect(0, 66687, '\P{Shavian}', "");
Expect(1, 66687, '\P{^Shavian}', "");
Expect(0, 66688, '\p{Shavian}', "");
Expect(1, 66688, '\p{^Shavian}', "");
Expect(1, 66688, '\P{Shavian}', "");
Expect(0, 66688, '\P{^Shavian}', "");
Error('\p{-:=is_shavian}');
Error('\P{-:=is_shavian}');
Expect(1, 66687, '\p{isshavian}', "");
Expect(0, 66687, '\p{^isshavian}', "");
Expect(0, 66687, '\P{isshavian}', "");
Expect(1, 66687, '\P{^isshavian}', "");
Expect(0, 66688, '\p{isshavian}', "");
Expect(1, 66688, '\p{^isshavian}', "");
Expect(1, 66688, '\P{isshavian}', "");
Expect(0, 66688, '\P{^isshavian}', "");
Expect(1, 66687, '\p{ IS_Shavian}', "");
Expect(0, 66687, '\p{^ IS_Shavian}', "");
Expect(0, 66687, '\P{ IS_Shavian}', "");
Expect(1, 66687, '\P{^ IS_Shavian}', "");
Expect(0, 66688, '\p{ IS_Shavian}', "");
Expect(1, 66688, '\p{^ IS_Shavian}', "");
Expect(1, 66688, '\P{ IS_Shavian}', "");
Expect(0, 66688, '\P{^ IS_Shavian}', "");
Error('\p{:=_SHAW}');
Error('\P{:=_SHAW}');
Expect(1, 66687, '\p{shaw}', "");
Expect(0, 66687, '\p{^shaw}', "");
Expect(0, 66687, '\P{shaw}', "");
Expect(1, 66687, '\P{^shaw}', "");
Expect(0, 66688, '\p{shaw}', "");
Expect(1, 66688, '\p{^shaw}', "");
Expect(1, 66688, '\P{shaw}', "");
Expect(0, 66688, '\P{^shaw}', "");
Expect(1, 66687, '\p{  Shaw}', "");
Expect(0, 66687, '\p{^  Shaw}', "");
Expect(0, 66687, '\P{  Shaw}', "");
Expect(1, 66687, '\P{^  Shaw}', "");
Expect(0, 66688, '\p{  Shaw}', "");
Expect(1, 66688, '\p{^  Shaw}', "");
Expect(1, 66688, '\P{  Shaw}', "");
Expect(0, 66688, '\P{^  Shaw}', "");
Error('\p{:=	IS_shaw}');
Error('\P{:=	IS_shaw}');
Expect(1, 66687, '\p{isshaw}', "");
Expect(0, 66687, '\p{^isshaw}', "");
Expect(0, 66687, '\P{isshaw}', "");
Expect(1, 66687, '\P{^isshaw}', "");
Expect(0, 66688, '\p{isshaw}', "");
Expect(1, 66688, '\p{^isshaw}', "");
Expect(1, 66688, '\P{isshaw}', "");
Expect(0, 66688, '\P{^isshaw}', "");
Expect(1, 66687, '\p{_Is_shaw}', "");
Expect(0, 66687, '\p{^_Is_shaw}', "");
Expect(0, 66687, '\P{_Is_shaw}', "");
Expect(1, 66687, '\P{^_Is_shaw}', "");
Expect(0, 66688, '\p{_Is_shaw}', "");
Expect(1, 66688, '\p{^_Is_shaw}', "");
Expect(1, 66688, '\P{_Is_shaw}', "");
Expect(0, 66688, '\P{^_Is_shaw}', "");
Error('\p{:=-sinhala}');
Error('\P{:=-sinhala}');
Expect(1, 3572, '\p{sinhala}', "");
Expect(0, 3572, '\p{^sinhala}', "");
Expect(0, 3572, '\P{sinhala}', "");
Expect(1, 3572, '\P{^sinhala}', "");
Expect(0, 3573, '\p{sinhala}', "");
Expect(1, 3573, '\p{^sinhala}', "");
Expect(1, 3573, '\P{sinhala}', "");
Expect(0, 3573, '\P{^sinhala}', "");
Expect(1, 3572, '\p{Sinhala}', "");
Expect(0, 3572, '\p{^Sinhala}', "");
Expect(0, 3572, '\P{Sinhala}', "");
Expect(1, 3572, '\P{^Sinhala}', "");
Expect(0, 3573, '\p{Sinhala}', "");
Expect(1, 3573, '\p{^Sinhala}', "");
Expect(1, 3573, '\P{Sinhala}', "");
Expect(0, 3573, '\P{^Sinhala}', "");
Error('\p{Is_Sinhala:=}');
Error('\P{Is_Sinhala:=}');
Expect(1, 3572, '\p{issinhala}', "");
Expect(0, 3572, '\p{^issinhala}', "");
Expect(0, 3572, '\P{issinhala}', "");
Expect(1, 3572, '\P{^issinhala}', "");
Expect(0, 3573, '\p{issinhala}', "");
Expect(1, 3573, '\p{^issinhala}', "");
Expect(1, 3573, '\P{issinhala}', "");
Expect(0, 3573, '\P{^issinhala}', "");
Expect(1, 3572, '\p{ _Is_sinhala}', "");
Expect(0, 3572, '\p{^ _Is_sinhala}', "");
Expect(0, 3572, '\P{ _Is_sinhala}', "");
Expect(1, 3572, '\P{^ _Is_sinhala}', "");
Expect(0, 3573, '\p{ _Is_sinhala}', "");
Expect(1, 3573, '\p{^ _Is_sinhala}', "");
Expect(1, 3573, '\P{ _Is_sinhala}', "");
Expect(0, 3573, '\P{^ _Is_sinhala}', "");
Error('\p{/a/_-Sinh}');
Error('\P{/a/_-Sinh}');
Expect(1, 3572, '\p{sinh}', "");
Expect(0, 3572, '\p{^sinh}', "");
Expect(0, 3572, '\P{sinh}', "");
Expect(1, 3572, '\P{^sinh}', "");
Expect(0, 3573, '\p{sinh}', "");
Expect(1, 3573, '\p{^sinh}', "");
Expect(1, 3573, '\P{sinh}', "");
Expect(0, 3573, '\P{^sinh}', "");
Expect(1, 3572, '\p{	_Sinh}', "");
Expect(0, 3572, '\p{^	_Sinh}', "");
Expect(0, 3572, '\P{	_Sinh}', "");
Expect(1, 3572, '\P{^	_Sinh}', "");
Expect(0, 3573, '\p{	_Sinh}', "");
Expect(1, 3573, '\p{^	_Sinh}', "");
Expect(1, 3573, '\P{	_Sinh}', "");
Expect(0, 3573, '\P{^	_Sinh}', "");
Error('\p{/a/_ Is_SINH}');
Error('\P{/a/_ Is_SINH}');
Expect(1, 3572, '\p{issinh}', "");
Expect(0, 3572, '\p{^issinh}', "");
Expect(0, 3572, '\P{issinh}', "");
Expect(1, 3572, '\P{^issinh}', "");
Expect(0, 3573, '\p{issinh}', "");
Expect(1, 3573, '\p{^issinh}', "");
Expect(1, 3573, '\P{issinh}', "");
Expect(0, 3573, '\P{^issinh}', "");
Expect(1, 3572, '\p{-Is_Sinh}', "");
Expect(0, 3572, '\p{^-Is_Sinh}', "");
Expect(0, 3572, '\P{-Is_Sinh}', "");
Expect(1, 3572, '\P{^-Is_Sinh}', "");
Expect(0, 3573, '\p{-Is_Sinh}', "");
Expect(1, 3573, '\p{^-Is_Sinh}', "");
Expect(1, 3573, '\P{-Is_Sinh}', "");
Expect(0, 3573, '\P{^-Is_Sinh}', "");
Error('\p{/a/Small_Form_VARIANTS}');
Error('\P{/a/Small_Form_VARIANTS}');
Expect(1, 65135, '\p{smallformvariants}', "");
Expect(0, 65135, '\p{^smallformvariants}', "");
Expect(0, 65135, '\P{smallformvariants}', "");
Expect(1, 65135, '\P{^smallformvariants}', "");
Expect(0, 65136, '\p{smallformvariants}', "");
Expect(1, 65136, '\p{^smallformvariants}', "");
Expect(1, 65136, '\P{smallformvariants}', "");
Expect(0, 65136, '\P{^smallformvariants}', "");
Expect(1, 65135, '\p{--Small_Form_variants}', "");
Expect(0, 65135, '\p{^--Small_Form_variants}', "");
Expect(0, 65135, '\P{--Small_Form_variants}', "");
Expect(1, 65135, '\P{^--Small_Form_variants}', "");
Expect(0, 65136, '\p{--Small_Form_variants}', "");
Expect(1, 65136, '\p{^--Small_Form_variants}', "");
Expect(1, 65136, '\P{--Small_Form_variants}', "");
Expect(0, 65136, '\P{^--Small_Form_variants}', "");
Error('\p{_ IS_small_FORM_variants/a/}');
Error('\P{_ IS_small_FORM_variants/a/}');
Expect(1, 65135, '\p{issmallformvariants}', "");
Expect(0, 65135, '\p{^issmallformvariants}', "");
Expect(0, 65135, '\P{issmallformvariants}', "");
Expect(1, 65135, '\P{^issmallformvariants}', "");
Expect(0, 65136, '\p{issmallformvariants}', "");
Expect(1, 65136, '\p{^issmallformvariants}', "");
Expect(1, 65136, '\P{issmallformvariants}', "");
Expect(0, 65136, '\P{^issmallformvariants}', "");
Expect(1, 65135, '\p{_	Is_Small_Form_Variants}', "");
Expect(0, 65135, '\p{^_	Is_Small_Form_Variants}', "");
Expect(0, 65135, '\P{_	Is_Small_Form_Variants}', "");
Expect(1, 65135, '\P{^_	Is_Small_Form_Variants}', "");
Expect(0, 65136, '\p{_	Is_Small_Form_Variants}', "");
Expect(1, 65136, '\p{^_	Is_Small_Form_Variants}', "");
Expect(1, 65136, '\P{_	Is_Small_Form_Variants}', "");
Expect(0, 65136, '\P{^_	Is_Small_Form_Variants}', "");
Error('\p{-IN_small_form_Variants/a/}');
Error('\P{-IN_small_form_Variants/a/}');
Expect(1, 65135, '\p{insmallformvariants}', "");
Expect(0, 65135, '\p{^insmallformvariants}', "");
Expect(0, 65135, '\P{insmallformvariants}', "");
Expect(1, 65135, '\P{^insmallformvariants}', "");
Expect(0, 65136, '\p{insmallformvariants}', "");
Expect(1, 65136, '\p{^insmallformvariants}', "");
Expect(1, 65136, '\P{insmallformvariants}', "");
Expect(0, 65136, '\P{^insmallformvariants}', "");
Expect(1, 65135, '\p{-IN_SMALL_Form_Variants}', "");
Expect(0, 65135, '\p{^-IN_SMALL_Form_Variants}', "");
Expect(0, 65135, '\P{-IN_SMALL_Form_Variants}', "");
Expect(1, 65135, '\P{^-IN_SMALL_Form_Variants}', "");
Expect(0, 65136, '\p{-IN_SMALL_Form_Variants}', "");
Expect(1, 65136, '\p{^-IN_SMALL_Form_Variants}', "");
Expect(1, 65136, '\P{-IN_SMALL_Form_Variants}', "");
Expect(0, 65136, '\P{^-IN_SMALL_Form_Variants}', "");
Error('\p{:=-	SMALL_Forms}');
Error('\P{:=-	SMALL_Forms}');
Expect(1, 65135, '\p{smallforms}', "");
Expect(0, 65135, '\p{^smallforms}', "");
Expect(0, 65135, '\P{smallforms}', "");
Expect(1, 65135, '\P{^smallforms}', "");
Expect(0, 65136, '\p{smallforms}', "");
Expect(1, 65136, '\p{^smallforms}', "");
Expect(1, 65136, '\P{smallforms}', "");
Expect(0, 65136, '\P{^smallforms}', "");
Expect(1, 65135, '\p{ -SMALL_FORMS}', "");
Expect(0, 65135, '\p{^ -SMALL_FORMS}', "");
Expect(0, 65135, '\P{ -SMALL_FORMS}', "");
Expect(1, 65135, '\P{^ -SMALL_FORMS}', "");
Expect(0, 65136, '\p{ -SMALL_FORMS}', "");
Expect(1, 65136, '\p{^ -SMALL_FORMS}', "");
Expect(1, 65136, '\P{ -SMALL_FORMS}', "");
Expect(0, 65136, '\P{^ -SMALL_FORMS}', "");
Error('\p{:=  Is_SMALL_Forms}');
Error('\P{:=  Is_SMALL_Forms}');
Expect(1, 65135, '\p{issmallforms}', "");
Expect(0, 65135, '\p{^issmallforms}', "");
Expect(0, 65135, '\P{issmallforms}', "");
Expect(1, 65135, '\P{^issmallforms}', "");
Expect(0, 65136, '\p{issmallforms}', "");
Expect(1, 65136, '\p{^issmallforms}', "");
Expect(1, 65136, '\P{issmallforms}', "");
Expect(0, 65136, '\P{^issmallforms}', "");
Expect(1, 65135, '\p{_	Is_Small_FORMS}', "");
Expect(0, 65135, '\p{^_	Is_Small_FORMS}', "");
Expect(0, 65135, '\P{_	Is_Small_FORMS}', "");
Expect(1, 65135, '\P{^_	Is_Small_FORMS}', "");
Expect(0, 65136, '\p{_	Is_Small_FORMS}', "");
Expect(1, 65136, '\p{^_	Is_Small_FORMS}', "");
Expect(1, 65136, '\P{_	Is_Small_FORMS}', "");
Expect(0, 65136, '\P{^_	Is_Small_FORMS}', "");
Error('\p{:=_-in_SMALL_Forms}');
Error('\P{:=_-in_SMALL_Forms}');
Expect(1, 65135, '\p{insmallforms}', "");
Expect(0, 65135, '\p{^insmallforms}', "");
Expect(0, 65135, '\P{insmallforms}', "");
Expect(1, 65135, '\P{^insmallforms}', "");
Expect(0, 65136, '\p{insmallforms}', "");
Expect(1, 65136, '\p{^insmallforms}', "");
Expect(1, 65136, '\P{insmallforms}', "");
Expect(0, 65136, '\P{^insmallforms}', "");
Expect(1, 65135, '\p{	-In_Small_Forms}', "");
Expect(0, 65135, '\p{^	-In_Small_Forms}', "");
Expect(0, 65135, '\P{	-In_Small_Forms}', "");
Expect(1, 65135, '\P{^	-In_Small_Forms}', "");
Expect(0, 65136, '\p{	-In_Small_Forms}', "");
Expect(1, 65136, '\p{^	-In_Small_Forms}', "");
Expect(1, 65136, '\P{	-In_Small_Forms}', "");
Expect(0, 65136, '\P{^	-In_Small_Forms}', "");
Error('\p{/a/		Sora_Sompeng}');
Error('\P{/a/		Sora_Sompeng}');
Expect(1, 69881, '\p{sorasompeng}', "");
Expect(0, 69881, '\p{^sorasompeng}', "");
Expect(0, 69881, '\P{sorasompeng}', "");
Expect(1, 69881, '\P{^sorasompeng}', "");
Expect(0, 69882, '\p{sorasompeng}', "");
Expect(1, 69882, '\p{^sorasompeng}', "");
Expect(1, 69882, '\P{sorasompeng}', "");
Expect(0, 69882, '\P{^sorasompeng}', "");
Expect(1, 69881, '\p{-Sora_Sompeng}', "");
Expect(0, 69881, '\p{^-Sora_Sompeng}', "");
Expect(0, 69881, '\P{-Sora_Sompeng}', "");
Expect(1, 69881, '\P{^-Sora_Sompeng}', "");
Expect(0, 69882, '\p{-Sora_Sompeng}', "");
Expect(1, 69882, '\p{^-Sora_Sompeng}', "");
Expect(1, 69882, '\P{-Sora_Sompeng}', "");
Expect(0, 69882, '\P{^-Sora_Sompeng}', "");
Error('\p{:=Is_SORA_Sompeng}');
Error('\P{:=Is_SORA_Sompeng}');
Expect(1, 69881, '\p{issorasompeng}', "");
Expect(0, 69881, '\p{^issorasompeng}', "");
Expect(0, 69881, '\P{issorasompeng}', "");
Expect(1, 69881, '\P{^issorasompeng}', "");
Expect(0, 69882, '\p{issorasompeng}', "");
Expect(1, 69882, '\p{^issorasompeng}', "");
Expect(1, 69882, '\P{issorasompeng}', "");
Expect(0, 69882, '\P{^issorasompeng}', "");
Expect(1, 69881, '\p{-is_SORA_sompeng}', "");
Expect(0, 69881, '\p{^-is_SORA_sompeng}', "");
Expect(0, 69881, '\P{-is_SORA_sompeng}', "");
Expect(1, 69881, '\P{^-is_SORA_sompeng}', "");
Expect(0, 69882, '\p{-is_SORA_sompeng}', "");
Expect(1, 69882, '\p{^-is_SORA_sompeng}', "");
Expect(1, 69882, '\P{-is_SORA_sompeng}', "");
Expect(0, 69882, '\P{^-is_SORA_sompeng}', "");
Error('\p{:=_Sora}');
Error('\P{:=_Sora}');
Expect(1, 69881, '\p{sora}', "");
Expect(0, 69881, '\p{^sora}', "");
Expect(0, 69881, '\P{sora}', "");
Expect(1, 69881, '\P{^sora}', "");
Expect(0, 69882, '\p{sora}', "");
Expect(1, 69882, '\p{^sora}', "");
Expect(1, 69882, '\P{sora}', "");
Expect(0, 69882, '\P{^sora}', "");
Expect(1, 69881, '\p{_SORA}', "");
Expect(0, 69881, '\p{^_SORA}', "");
Expect(0, 69881, '\P{_SORA}', "");
Expect(1, 69881, '\P{^_SORA}', "");
Expect(0, 69882, '\p{_SORA}', "");
Expect(1, 69882, '\p{^_SORA}', "");
Expect(1, 69882, '\P{_SORA}', "");
Expect(0, 69882, '\P{^_SORA}', "");
Error('\p{/a/ _IS_SORA}');
Error('\P{/a/ _IS_SORA}');
Expect(1, 69881, '\p{issora}', "");
Expect(0, 69881, '\p{^issora}', "");
Expect(0, 69881, '\P{issora}', "");
Expect(1, 69881, '\P{^issora}', "");
Expect(0, 69882, '\p{issora}', "");
Expect(1, 69882, '\p{^issora}', "");
Expect(1, 69882, '\P{issora}', "");
Expect(0, 69882, '\P{^issora}', "");
Expect(1, 69881, '\p{ 	Is_Sora}', "");
Expect(0, 69881, '\p{^ 	Is_Sora}', "");
Expect(0, 69881, '\P{ 	Is_Sora}', "");
Expect(1, 69881, '\P{^ 	Is_Sora}', "");
Expect(0, 69882, '\p{ 	Is_Sora}', "");
Expect(1, 69882, '\p{^ 	Is_Sora}', "");
Expect(1, 69882, '\P{ 	Is_Sora}', "");
Expect(0, 69882, '\P{^ 	Is_Sora}', "");
Error('\p{ :=Space}');
Error('\P{ :=Space}');
Expect(1, 12288, '\p{space}', "");
Expect(0, 12288, '\p{^space}', "");
Expect(0, 12288, '\P{space}', "");
Expect(1, 12288, '\P{^space}', "");
Expect(0, 12289, '\p{space}', "");
Expect(1, 12289, '\p{^space}', "");
Expect(1, 12289, '\P{space}', "");
Expect(0, 12289, '\P{^space}', "");
Expect(1, 12288, '\p{--SPACE}', "");
Expect(0, 12288, '\p{^--SPACE}', "");
Expect(0, 12288, '\P{--SPACE}', "");
Expect(1, 12288, '\P{^--SPACE}', "");
Expect(0, 12289, '\p{--SPACE}', "");
Expect(1, 12289, '\p{^--SPACE}', "");
Expect(1, 12289, '\P{--SPACE}', "");
Expect(0, 12289, '\P{^--SPACE}', "");
Error('\p{ _XPOSIXSPACE:=}');
Error('\P{ _XPOSIXSPACE:=}');
Expect(1, 12288, '\p{xposixspace}', "");
Expect(0, 12288, '\p{^xposixspace}', "");
Expect(0, 12288, '\P{xposixspace}', "");
Expect(1, 12288, '\P{^xposixspace}', "");
Expect(0, 12289, '\p{xposixspace}', "");
Expect(1, 12289, '\p{^xposixspace}', "");
Expect(1, 12289, '\P{xposixspace}', "");
Expect(0, 12289, '\P{^xposixspace}', "");
Expect(1, 12288, '\p{-_XPosixSpace}', "");
Expect(0, 12288, '\p{^-_XPosixSpace}', "");
Expect(0, 12288, '\P{-_XPosixSpace}', "");
Expect(1, 12288, '\P{^-_XPosixSpace}', "");
Expect(0, 12289, '\p{-_XPosixSpace}', "");
Expect(1, 12289, '\p{^-_XPosixSpace}', "");
Expect(1, 12289, '\P{-_XPosixSpace}', "");
Expect(0, 12289, '\P{^-_XPosixSpace}', "");
Error('\p{ /a/Is_SPACE}');
Error('\P{ /a/Is_SPACE}');
Expect(1, 12288, '\p{isspace}', "");
Expect(0, 12288, '\p{^isspace}', "");
Expect(0, 12288, '\P{isspace}', "");
Expect(1, 12288, '\P{^isspace}', "");
Expect(0, 12289, '\p{isspace}', "");
Expect(1, 12289, '\p{^isspace}', "");
Expect(1, 12289, '\P{isspace}', "");
Expect(0, 12289, '\P{^isspace}', "");
Expect(1, 12288, '\p{	-Is_Space}', "");
Expect(0, 12288, '\p{^	-Is_Space}', "");
Expect(0, 12288, '\P{	-Is_Space}', "");
Expect(1, 12288, '\P{^	-Is_Space}', "");
Expect(0, 12289, '\p{	-Is_Space}', "");
Expect(1, 12289, '\p{^	-Is_Space}', "");
Expect(1, 12289, '\P{	-Is_Space}', "");
Expect(0, 12289, '\P{^	-Is_Space}', "");
Error('\p{/a/is_XPosixSpace}');
Error('\P{/a/is_XPosixSpace}');
Expect(1, 12288, '\p{isxposixspace}', "");
Expect(0, 12288, '\p{^isxposixspace}', "");
Expect(0, 12288, '\P{isxposixspace}', "");
Expect(1, 12288, '\P{^isxposixspace}', "");
Expect(0, 12289, '\p{isxposixspace}', "");
Expect(1, 12289, '\p{^isxposixspace}', "");
Expect(1, 12289, '\P{isxposixspace}', "");
Expect(0, 12289, '\P{^isxposixspace}', "");
Expect(1, 12288, '\p{	-is_XPOSIXSPACE}', "");
Expect(0, 12288, '\p{^	-is_XPOSIXSPACE}', "");
Expect(0, 12288, '\P{	-is_XPOSIXSPACE}', "");
Expect(1, 12288, '\P{^	-is_XPOSIXSPACE}', "");
Expect(0, 12289, '\p{	-is_XPOSIXSPACE}', "");
Expect(1, 12289, '\p{^	-is_XPOSIXSPACE}', "");
Expect(1, 12289, '\P{	-is_XPOSIXSPACE}', "");
Expect(0, 12289, '\P{^	-is_XPOSIXSPACE}', "");
Error('\p{- space_Separator:=}');
Error('\P{- space_Separator:=}');
Expect(1, 12288, '\p{spaceseparator}', "");
Expect(0, 12288, '\p{^spaceseparator}', "");
Expect(0, 12288, '\P{spaceseparator}', "");
Expect(1, 12288, '\P{^spaceseparator}', "");
Expect(0, 12289, '\p{spaceseparator}', "");
Expect(1, 12289, '\p{^spaceseparator}', "");
Expect(1, 12289, '\P{spaceseparator}', "");
Expect(0, 12289, '\P{^spaceseparator}', "");
Expect(1, 12288, '\p{ Space_Separator}', "");
Expect(0, 12288, '\p{^ Space_Separator}', "");
Expect(0, 12288, '\P{ Space_Separator}', "");
Expect(1, 12288, '\P{^ Space_Separator}', "");
Expect(0, 12289, '\p{ Space_Separator}', "");
Expect(1, 12289, '\p{^ Space_Separator}', "");
Expect(1, 12289, '\P{ Space_Separator}', "");
Expect(0, 12289, '\P{^ Space_Separator}', "");
Error('\p{:=is_space_SEPARATOR}');
Error('\P{:=is_space_SEPARATOR}');
Expect(1, 12288, '\p{isspaceseparator}', "");
Expect(0, 12288, '\p{^isspaceseparator}', "");
Expect(0, 12288, '\P{isspaceseparator}', "");
Expect(1, 12288, '\P{^isspaceseparator}', "");
Expect(0, 12289, '\p{isspaceseparator}', "");
Expect(1, 12289, '\p{^isspaceseparator}', "");
Expect(1, 12289, '\P{isspaceseparator}', "");
Expect(0, 12289, '\P{^isspaceseparator}', "");
Expect(1, 12288, '\p{Is_SPACE_Separator}', "");
Expect(0, 12288, '\p{^Is_SPACE_Separator}', "");
Expect(0, 12288, '\P{Is_SPACE_Separator}', "");
Expect(1, 12288, '\P{^Is_SPACE_Separator}', "");
Expect(0, 12289, '\p{Is_SPACE_Separator}', "");
Expect(1, 12289, '\p{^Is_SPACE_Separator}', "");
Expect(1, 12289, '\P{Is_SPACE_Separator}', "");
Expect(0, 12289, '\P{^Is_SPACE_Separator}', "");
Error('\p{	-Zs:=}');
Error('\P{	-Zs:=}');
Expect(1, 12288, '\p{zs}', "");
Expect(0, 12288, '\p{^zs}', "");
Expect(0, 12288, '\P{zs}', "");
Expect(1, 12288, '\P{^zs}', "");
Expect(0, 12289, '\p{zs}', "");
Expect(1, 12289, '\p{^zs}', "");
Expect(1, 12289, '\P{zs}', "");
Expect(0, 12289, '\P{^zs}', "");
Expect(1, 12288, '\p{_-ZS}', "");
Expect(0, 12288, '\p{^_-ZS}', "");
Expect(0, 12288, '\P{_-ZS}', "");
Expect(1, 12288, '\P{^_-ZS}', "");
Expect(0, 12289, '\p{_-ZS}', "");
Expect(1, 12289, '\p{^_-ZS}', "");
Expect(1, 12289, '\P{_-ZS}', "");
Expect(0, 12289, '\P{^_-ZS}', "");
Error('\p{/a/_-Is_Zs}');
Error('\P{/a/_-Is_Zs}');
Expect(1, 12288, '\p{iszs}', "");
Expect(0, 12288, '\p{^iszs}', "");
Expect(0, 12288, '\P{iszs}', "");
Expect(1, 12288, '\P{^iszs}', "");
Expect(0, 12289, '\p{iszs}', "");
Expect(1, 12289, '\p{^iszs}', "");
Expect(1, 12289, '\P{iszs}', "");
Expect(0, 12289, '\P{^iszs}', "");
Expect(1, 12288, '\p{-_Is_Zs}', "");
Expect(0, 12288, '\p{^-_Is_Zs}', "");
Expect(0, 12288, '\P{-_Is_Zs}', "");
Expect(1, 12288, '\P{^-_Is_Zs}', "");
Expect(0, 12289, '\p{-_Is_Zs}', "");
Expect(1, 12289, '\p{^-_Is_Zs}', "");
Expect(1, 12289, '\P{-_Is_Zs}', "");
Expect(0, 12289, '\P{^-_Is_Zs}', "");
Error('\p{/a/SPACING_Mark}');
Error('\P{/a/SPACING_Mark}');
Expect(1, 119154, '\p{spacingmark}', "");
Expect(0, 119154, '\p{^spacingmark}', "");
Expect(0, 119154, '\P{spacingmark}', "");
Expect(1, 119154, '\P{^spacingmark}', "");
Expect(0, 119155, '\p{spacingmark}', "");
Expect(1, 119155, '\p{^spacingmark}', "");
Expect(1, 119155, '\P{spacingmark}', "");
Expect(0, 119155, '\P{^spacingmark}', "");
Expect(1, 119154, '\p{	 Spacing_MARK}', "");
Expect(0, 119154, '\p{^	 Spacing_MARK}', "");
Expect(0, 119154, '\P{	 Spacing_MARK}', "");
Expect(1, 119154, '\P{^	 Spacing_MARK}', "");
Expect(0, 119155, '\p{	 Spacing_MARK}', "");
Expect(1, 119155, '\p{^	 Spacing_MARK}', "");
Expect(1, 119155, '\P{	 Spacing_MARK}', "");
Expect(0, 119155, '\P{^	 Spacing_MARK}', "");
Error('\p{/a/	-Is_Spacing_MARK}');
Error('\P{/a/	-Is_Spacing_MARK}');
Expect(1, 119154, '\p{isspacingmark}', "");
Expect(0, 119154, '\p{^isspacingmark}', "");
Expect(0, 119154, '\P{isspacingmark}', "");
Expect(1, 119154, '\P{^isspacingmark}', "");
Expect(0, 119155, '\p{isspacingmark}', "");
Expect(1, 119155, '\p{^isspacingmark}', "");
Expect(1, 119155, '\P{isspacingmark}', "");
Expect(0, 119155, '\P{^isspacingmark}', "");
Expect(1, 119154, '\p{		IS_spacing_Mark}', "");
Expect(0, 119154, '\p{^		IS_spacing_Mark}', "");
Expect(0, 119154, '\P{		IS_spacing_Mark}', "");
Expect(1, 119154, '\P{^		IS_spacing_Mark}', "");
Expect(0, 119155, '\p{		IS_spacing_Mark}', "");
Expect(1, 119155, '\p{^		IS_spacing_Mark}', "");
Expect(1, 119155, '\P{		IS_spacing_Mark}', "");
Expect(0, 119155, '\P{^		IS_spacing_Mark}', "");
Error('\p{_-Mc/a/}');
Error('\P{_-Mc/a/}');
Expect(1, 119154, '\p{mc}', "");
Expect(0, 119154, '\p{^mc}', "");
Expect(0, 119154, '\P{mc}', "");
Expect(1, 119154, '\P{^mc}', "");
Expect(0, 119155, '\p{mc}', "");
Expect(1, 119155, '\p{^mc}', "");
Expect(1, 119155, '\P{mc}', "");
Expect(0, 119155, '\P{^mc}', "");
Expect(1, 119154, '\p{__Mc}', "");
Expect(0, 119154, '\p{^__Mc}', "");
Expect(0, 119154, '\P{__Mc}', "");
Expect(1, 119154, '\P{^__Mc}', "");
Expect(0, 119155, '\p{__Mc}', "");
Expect(1, 119155, '\p{^__Mc}', "");
Expect(1, 119155, '\P{__Mc}', "");
Expect(0, 119155, '\P{^__Mc}', "");
Error('\p{:=_ Is_MC}');
Error('\P{:=_ Is_MC}');
Expect(1, 119154, '\p{ismc}', "");
Expect(0, 119154, '\p{^ismc}', "");
Expect(0, 119154, '\P{ismc}', "");
Expect(1, 119154, '\P{^ismc}', "");
Expect(0, 119155, '\p{ismc}', "");
Expect(1, 119155, '\p{^ismc}', "");
Expect(1, 119155, '\P{ismc}', "");
Expect(0, 119155, '\P{^ismc}', "");
Expect(1, 119154, '\p{__Is_MC}', "");
Expect(0, 119154, '\p{^__Is_MC}', "");
Expect(0, 119154, '\P{__Is_MC}', "");
Expect(1, 119154, '\P{^__Is_MC}', "");
Expect(0, 119155, '\p{__Is_MC}', "");
Expect(1, 119155, '\p{^__Is_MC}', "");
Expect(1, 119155, '\P{__Is_MC}', "");
Expect(0, 119155, '\P{^__Is_MC}', "");
Error('\p{_:=Spacing_Modifier_LETTERS}');
Error('\P{_:=Spacing_Modifier_LETTERS}');
Expect(1, 767, '\p{spacingmodifierletters}', "");
Expect(0, 767, '\p{^spacingmodifierletters}', "");
Expect(0, 767, '\P{spacingmodifierletters}', "");
Expect(1, 767, '\P{^spacingmodifierletters}', "");
Expect(0, 768, '\p{spacingmodifierletters}', "");
Expect(1, 768, '\p{^spacingmodifierletters}', "");
Expect(1, 768, '\P{spacingmodifierletters}', "");
Expect(0, 768, '\P{^spacingmodifierletters}', "");
Expect(1, 767, '\p{  Spacing_Modifier_Letters}', "");
Expect(0, 767, '\p{^  Spacing_Modifier_Letters}', "");
Expect(0, 767, '\P{  Spacing_Modifier_Letters}', "");
Expect(1, 767, '\P{^  Spacing_Modifier_Letters}', "");
Expect(0, 768, '\p{  Spacing_Modifier_Letters}', "");
Expect(1, 768, '\p{^  Spacing_Modifier_Letters}', "");
Expect(1, 768, '\P{  Spacing_Modifier_Letters}', "");
Expect(0, 768, '\P{^  Spacing_Modifier_Letters}', "");
Error('\p{:= 	Is_SPACING_Modifier_Letters}');
Error('\P{:= 	Is_SPACING_Modifier_Letters}');
Expect(1, 767, '\p{isspacingmodifierletters}', "");
Expect(0, 767, '\p{^isspacingmodifierletters}', "");
Expect(0, 767, '\P{isspacingmodifierletters}', "");
Expect(1, 767, '\P{^isspacingmodifierletters}', "");
Expect(0, 768, '\p{isspacingmodifierletters}', "");
Expect(1, 768, '\p{^isspacingmodifierletters}', "");
Expect(1, 768, '\P{isspacingmodifierletters}', "");
Expect(0, 768, '\P{^isspacingmodifierletters}', "");
Expect(1, 767, '\p{_	IS_spacing_Modifier_Letters}', "");
Expect(0, 767, '\p{^_	IS_spacing_Modifier_Letters}', "");
Expect(0, 767, '\P{_	IS_spacing_Modifier_Letters}', "");
Expect(1, 767, '\P{^_	IS_spacing_Modifier_Letters}', "");
Expect(0, 768, '\p{_	IS_spacing_Modifier_Letters}', "");
Expect(1, 768, '\p{^_	IS_spacing_Modifier_Letters}', "");
Expect(1, 768, '\P{_	IS_spacing_Modifier_Letters}', "");
Expect(0, 768, '\P{^_	IS_spacing_Modifier_Letters}', "");
Error('\p{ 	In_SPACING_modifier_LETTERS:=}');
Error('\P{ 	In_SPACING_modifier_LETTERS:=}');
Expect(1, 767, '\p{inspacingmodifierletters}', "");
Expect(0, 767, '\p{^inspacingmodifierletters}', "");
Expect(0, 767, '\P{inspacingmodifierletters}', "");
Expect(1, 767, '\P{^inspacingmodifierletters}', "");
Expect(0, 768, '\p{inspacingmodifierletters}', "");
Expect(1, 768, '\p{^inspacingmodifierletters}', "");
Expect(1, 768, '\P{inspacingmodifierletters}', "");
Expect(0, 768, '\P{^inspacingmodifierletters}', "");
Expect(1, 767, '\p{_	In_Spacing_modifier_LETTERS}', "");
Expect(0, 767, '\p{^_	In_Spacing_modifier_LETTERS}', "");
Expect(0, 767, '\P{_	In_Spacing_modifier_LETTERS}', "");
Expect(1, 767, '\P{^_	In_Spacing_modifier_LETTERS}', "");
Expect(0, 768, '\p{_	In_Spacing_modifier_LETTERS}', "");
Expect(1, 768, '\p{^_	In_Spacing_modifier_LETTERS}', "");
Expect(1, 768, '\P{_	In_Spacing_modifier_LETTERS}', "");
Expect(0, 768, '\P{^_	In_Spacing_modifier_LETTERS}', "");
Error('\p{	 MODIFIER_LETTERS:=}');
Error('\P{	 MODIFIER_LETTERS:=}');
Expect(1, 767, '\p{modifierletters}', "");
Expect(0, 767, '\p{^modifierletters}', "");
Expect(0, 767, '\P{modifierletters}', "");
Expect(1, 767, '\P{^modifierletters}', "");
Expect(0, 768, '\p{modifierletters}', "");
Expect(1, 768, '\p{^modifierletters}', "");
Expect(1, 768, '\P{modifierletters}', "");
Expect(0, 768, '\P{^modifierletters}', "");
Expect(1, 767, '\p{_	MODIFIER_Letters}', "");
Expect(0, 767, '\p{^_	MODIFIER_Letters}', "");
Expect(0, 767, '\P{_	MODIFIER_Letters}', "");
Expect(1, 767, '\P{^_	MODIFIER_Letters}', "");
Expect(0, 768, '\p{_	MODIFIER_Letters}', "");
Expect(1, 768, '\p{^_	MODIFIER_Letters}', "");
Expect(1, 768, '\P{_	MODIFIER_Letters}', "");
Expect(0, 768, '\P{^_	MODIFIER_Letters}', "");
Error('\p{_:=IS_Modifier_letters}');
Error('\P{_:=IS_Modifier_letters}');
Expect(1, 767, '\p{ismodifierletters}', "");
Expect(0, 767, '\p{^ismodifierletters}', "");
Expect(0, 767, '\P{ismodifierletters}', "");
Expect(1, 767, '\P{^ismodifierletters}', "");
Expect(0, 768, '\p{ismodifierletters}', "");
Expect(1, 768, '\p{^ismodifierletters}', "");
Expect(1, 768, '\P{ismodifierletters}', "");
Expect(0, 768, '\P{^ismodifierletters}', "");
Expect(1, 767, '\p{ 	Is_Modifier_LETTERS}', "");
Expect(0, 767, '\p{^ 	Is_Modifier_LETTERS}', "");
Expect(0, 767, '\P{ 	Is_Modifier_LETTERS}', "");
Expect(1, 767, '\P{^ 	Is_Modifier_LETTERS}', "");
Expect(0, 768, '\p{ 	Is_Modifier_LETTERS}', "");
Expect(1, 768, '\p{^ 	Is_Modifier_LETTERS}', "");
Expect(1, 768, '\P{ 	Is_Modifier_LETTERS}', "");
Expect(0, 768, '\P{^ 	Is_Modifier_LETTERS}', "");
Error('\p{- In_Modifier_LETTERS/a/}');
Error('\P{- In_Modifier_LETTERS/a/}');
Expect(1, 767, '\p{inmodifierletters}', "");
Expect(0, 767, '\p{^inmodifierletters}', "");
Expect(0, 767, '\P{inmodifierletters}', "");
Expect(1, 767, '\P{^inmodifierletters}', "");
Expect(0, 768, '\p{inmodifierletters}', "");
Expect(1, 768, '\p{^inmodifierletters}', "");
Expect(1, 768, '\P{inmodifierletters}', "");
Expect(0, 768, '\P{^inmodifierletters}', "");
Expect(1, 767, '\p{	In_Modifier_letters}', "");
Expect(0, 767, '\p{^	In_Modifier_letters}', "");
Expect(0, 767, '\P{	In_Modifier_letters}', "");
Expect(1, 767, '\P{^	In_Modifier_letters}', "");
Expect(0, 768, '\p{	In_Modifier_letters}', "");
Expect(1, 768, '\p{^	In_Modifier_letters}', "");
Expect(1, 768, '\P{	In_Modifier_letters}', "");
Expect(0, 768, '\P{^	In_Modifier_letters}', "");
Error('\p{ :=SPECIALS}');
Error('\P{ :=SPECIALS}');
Expect(1, 65520, '\p{specials}', "");
Expect(0, 65520, '\p{^specials}', "");
Expect(0, 65520, '\P{specials}', "");
Expect(1, 65520, '\P{^specials}', "");
Expect(0, 65536, '\p{specials}', "");
Expect(1, 65536, '\p{^specials}', "");
Expect(1, 65536, '\P{specials}', "");
Expect(0, 65536, '\P{^specials}', "");
Expect(1, 65520, '\p{_Specials}', "");
Expect(0, 65520, '\p{^_Specials}', "");
Expect(0, 65520, '\P{_Specials}', "");
Expect(1, 65520, '\P{^_Specials}', "");
Expect(0, 65536, '\p{_Specials}', "");
Expect(1, 65536, '\p{^_Specials}', "");
Expect(1, 65536, '\P{_Specials}', "");
Expect(0, 65536, '\P{^_Specials}', "");
Error('\p{-/a/IS_SPECIALS}');
Error('\P{-/a/IS_SPECIALS}');
Expect(1, 65520, '\p{isspecials}', "");
Expect(0, 65520, '\p{^isspecials}', "");
Expect(0, 65520, '\P{isspecials}', "");
Expect(1, 65520, '\P{^isspecials}', "");
Expect(0, 65536, '\p{isspecials}', "");
Expect(1, 65536, '\p{^isspecials}', "");
Expect(1, 65536, '\P{isspecials}', "");
Expect(0, 65536, '\P{^isspecials}', "");
Expect(1, 65520, '\p{		is_Specials}', "");
Expect(0, 65520, '\p{^		is_Specials}', "");
Expect(0, 65520, '\P{		is_Specials}', "");
Expect(1, 65520, '\P{^		is_Specials}', "");
Expect(0, 65536, '\p{		is_Specials}', "");
Expect(1, 65536, '\p{^		is_Specials}', "");
Expect(1, 65536, '\P{		is_Specials}', "");
Expect(0, 65536, '\P{^		is_Specials}', "");
Error('\p{	/a/in_specials}');
Error('\P{	/a/in_specials}');
Expect(1, 65520, '\p{inspecials}', "");
Expect(0, 65520, '\p{^inspecials}', "");
Expect(0, 65520, '\P{inspecials}', "");
Expect(1, 65520, '\P{^inspecials}', "");
Expect(0, 65536, '\p{inspecials}', "");
Expect(1, 65536, '\p{^inspecials}', "");
Expect(1, 65536, '\P{inspecials}', "");
Expect(0, 65536, '\P{^inspecials}', "");
Expect(1, 65520, '\p{-in_Specials}', "");
Expect(0, 65520, '\p{^-in_Specials}', "");
Expect(0, 65520, '\P{-in_Specials}', "");
Expect(1, 65520, '\P{^-in_Specials}', "");
Expect(0, 65536, '\p{-in_Specials}', "");
Expect(1, 65536, '\p{^-in_Specials}', "");
Expect(1, 65536, '\P{-in_Specials}', "");
Expect(0, 65536, '\P{^-in_Specials}', "");
Error('\p{-/a/sundanese}');
Error('\P{-/a/sundanese}');
Expect(1, 7367, '\p{sundanese}', "");
Expect(0, 7367, '\p{^sundanese}', "");
Expect(0, 7367, '\P{sundanese}', "");
Expect(1, 7367, '\P{^sundanese}', "");
Expect(0, 7368, '\p{sundanese}', "");
Expect(1, 7368, '\p{^sundanese}', "");
Expect(1, 7368, '\P{sundanese}', "");
Expect(0, 7368, '\P{^sundanese}', "");
Expect(1, 7367, '\p{ 	SUNDANESE}', "");
Expect(0, 7367, '\p{^ 	SUNDANESE}', "");
Expect(0, 7367, '\P{ 	SUNDANESE}', "");
Expect(1, 7367, '\P{^ 	SUNDANESE}', "");
Expect(0, 7368, '\p{ 	SUNDANESE}', "");
Expect(1, 7368, '\p{^ 	SUNDANESE}', "");
Expect(1, 7368, '\P{ 	SUNDANESE}', "");
Expect(0, 7368, '\P{^ 	SUNDANESE}', "");
Error('\p{	/a/IS_Sundanese}');
Error('\P{	/a/IS_Sundanese}');
Expect(1, 7367, '\p{issundanese}', "");
Expect(0, 7367, '\p{^issundanese}', "");
Expect(0, 7367, '\P{issundanese}', "");
Expect(1, 7367, '\P{^issundanese}', "");
Expect(0, 7368, '\p{issundanese}', "");
Expect(1, 7368, '\p{^issundanese}', "");
Expect(1, 7368, '\P{issundanese}', "");
Expect(0, 7368, '\P{^issundanese}', "");
Expect(1, 7367, '\p{-	Is_SUNDANESE}', "");
Expect(0, 7367, '\p{^-	Is_SUNDANESE}', "");
Expect(0, 7367, '\P{-	Is_SUNDANESE}', "");
Expect(1, 7367, '\P{^-	Is_SUNDANESE}', "");
Expect(0, 7368, '\p{-	Is_SUNDANESE}', "");
Expect(1, 7368, '\p{^-	Is_SUNDANESE}', "");
Expect(1, 7368, '\P{-	Is_SUNDANESE}', "");
Expect(0, 7368, '\P{^-	Is_SUNDANESE}', "");
Error('\p{--SUND:=}');
Error('\P{--SUND:=}');
Expect(1, 7367, '\p{sund}', "");
Expect(0, 7367, '\p{^sund}', "");
Expect(0, 7367, '\P{sund}', "");
Expect(1, 7367, '\P{^sund}', "");
Expect(0, 7368, '\p{sund}', "");
Expect(1, 7368, '\p{^sund}', "");
Expect(1, 7368, '\P{sund}', "");
Expect(0, 7368, '\P{^sund}', "");
Expect(1, 7367, '\p{_	SUND}', "");
Expect(0, 7367, '\p{^_	SUND}', "");
Expect(0, 7367, '\P{_	SUND}', "");
Expect(1, 7367, '\P{^_	SUND}', "");
Expect(0, 7368, '\p{_	SUND}', "");
Expect(1, 7368, '\p{^_	SUND}', "");
Expect(1, 7368, '\P{_	SUND}', "");
Expect(0, 7368, '\P{^_	SUND}', "");
Error('\p{-/a/Is_sund}');
Error('\P{-/a/Is_sund}');
Expect(1, 7367, '\p{issund}', "");
Expect(0, 7367, '\p{^issund}', "");
Expect(0, 7367, '\P{issund}', "");
Expect(1, 7367, '\P{^issund}', "");
Expect(0, 7368, '\p{issund}', "");
Expect(1, 7368, '\p{^issund}', "");
Expect(1, 7368, '\P{issund}', "");
Expect(0, 7368, '\P{^issund}', "");
Expect(1, 7367, '\p{_IS_sund}', "");
Expect(0, 7367, '\p{^_IS_sund}', "");
Expect(0, 7367, '\P{_IS_sund}', "");
Expect(1, 7367, '\P{^_IS_sund}', "");
Expect(0, 7368, '\p{_IS_sund}', "");
Expect(1, 7368, '\p{^_IS_sund}', "");
Expect(1, 7368, '\P{_IS_sund}', "");
Expect(0, 7368, '\P{^_IS_sund}', "");
Error('\p{		Sundanese_Supplement/a/}');
Error('\P{		Sundanese_Supplement/a/}');
Expect(1, 7375, '\p{sundanesesupplement}', "");
Expect(0, 7375, '\p{^sundanesesupplement}', "");
Expect(0, 7375, '\P{sundanesesupplement}', "");
Expect(1, 7375, '\P{^sundanesesupplement}', "");
Expect(0, 7376, '\p{sundanesesupplement}', "");
Expect(1, 7376, '\p{^sundanesesupplement}', "");
Expect(1, 7376, '\P{sundanesesupplement}', "");
Expect(0, 7376, '\P{^sundanesesupplement}', "");
Expect(1, 7375, '\p{  sundanese_Supplement}', "");
Expect(0, 7375, '\p{^  sundanese_Supplement}', "");
Expect(0, 7375, '\P{  sundanese_Supplement}', "");
Expect(1, 7375, '\P{^  sundanese_Supplement}', "");
Expect(0, 7376, '\p{  sundanese_Supplement}', "");
Expect(1, 7376, '\p{^  sundanese_Supplement}', "");
Expect(1, 7376, '\P{  sundanese_Supplement}', "");
Expect(0, 7376, '\P{^  sundanese_Supplement}', "");
Error('\p{/a/--is_sundanese_Supplement}');
Error('\P{/a/--is_sundanese_Supplement}');
Expect(1, 7375, '\p{issundanesesupplement}', "");
Expect(0, 7375, '\p{^issundanesesupplement}', "");
Expect(0, 7375, '\P{issundanesesupplement}', "");
Expect(1, 7375, '\P{^issundanesesupplement}', "");
Expect(0, 7376, '\p{issundanesesupplement}', "");
Expect(1, 7376, '\p{^issundanesesupplement}', "");
Expect(1, 7376, '\P{issundanesesupplement}', "");
Expect(0, 7376, '\P{^issundanesesupplement}', "");
Expect(1, 7375, '\p{__IS_SUNDANESE_SUPPLEMENT}', "");
Expect(0, 7375, '\p{^__IS_SUNDANESE_SUPPLEMENT}', "");
Expect(0, 7375, '\P{__IS_SUNDANESE_SUPPLEMENT}', "");
Expect(1, 7375, '\P{^__IS_SUNDANESE_SUPPLEMENT}', "");
Expect(0, 7376, '\p{__IS_SUNDANESE_SUPPLEMENT}', "");
Expect(1, 7376, '\p{^__IS_SUNDANESE_SUPPLEMENT}', "");
Expect(1, 7376, '\P{__IS_SUNDANESE_SUPPLEMENT}', "");
Expect(0, 7376, '\P{^__IS_SUNDANESE_SUPPLEMENT}', "");
Error('\p{/a/ -In_Sundanese_SUPPLEMENT}');
Error('\P{/a/ -In_Sundanese_SUPPLEMENT}');
Expect(1, 7375, '\p{insundanesesupplement}', "");
Expect(0, 7375, '\p{^insundanesesupplement}', "");
Expect(0, 7375, '\P{insundanesesupplement}', "");
Expect(1, 7375, '\P{^insundanesesupplement}', "");
Expect(0, 7376, '\p{insundanesesupplement}', "");
Expect(1, 7376, '\p{^insundanesesupplement}', "");
Expect(1, 7376, '\P{insundanesesupplement}', "");
Expect(0, 7376, '\P{^insundanesesupplement}', "");
Expect(1, 7375, '\p{-In_Sundanese_supplement}', "");
Expect(0, 7375, '\p{^-In_Sundanese_supplement}', "");
Expect(0, 7375, '\P{-In_Sundanese_supplement}', "");
Expect(1, 7375, '\P{^-In_Sundanese_supplement}', "");
Expect(0, 7376, '\p{-In_Sundanese_supplement}', "");
Expect(1, 7376, '\p{^-In_Sundanese_supplement}', "");
Expect(1, 7376, '\P{-In_Sundanese_supplement}', "");
Expect(0, 7376, '\P{^-In_Sundanese_supplement}', "");
Error('\p{:=	_Sundanese_sup}');
Error('\P{:=	_Sundanese_sup}');
Expect(1, 7375, '\p{sundanesesup}', "");
Expect(0, 7375, '\p{^sundanesesup}', "");
Expect(0, 7375, '\P{sundanesesup}', "");
Expect(1, 7375, '\P{^sundanesesup}', "");
Expect(0, 7376, '\p{sundanesesup}', "");
Expect(1, 7376, '\p{^sundanesesup}', "");
Expect(1, 7376, '\P{sundanesesup}', "");
Expect(0, 7376, '\P{^sundanesesup}', "");
Expect(1, 7375, '\p{	 Sundanese_Sup}', "");
Expect(0, 7375, '\p{^	 Sundanese_Sup}', "");
Expect(0, 7375, '\P{	 Sundanese_Sup}', "");
Expect(1, 7375, '\P{^	 Sundanese_Sup}', "");
Expect(0, 7376, '\p{	 Sundanese_Sup}', "");
Expect(1, 7376, '\p{^	 Sundanese_Sup}', "");
Expect(1, 7376, '\P{	 Sundanese_Sup}', "");
Expect(0, 7376, '\P{^	 Sundanese_Sup}', "");
Error('\p{_:=Is_SUNDANESE_sup}');
Error('\P{_:=Is_SUNDANESE_sup}');
Expect(1, 7375, '\p{issundanesesup}', "");
Expect(0, 7375, '\p{^issundanesesup}', "");
Expect(0, 7375, '\P{issundanesesup}', "");
Expect(1, 7375, '\P{^issundanesesup}', "");
Expect(0, 7376, '\p{issundanesesup}', "");
Expect(1, 7376, '\p{^issundanesesup}', "");
Expect(1, 7376, '\P{issundanesesup}', "");
Expect(0, 7376, '\P{^issundanesesup}', "");
Expect(1, 7375, '\p{_ is_Sundanese_Sup}', "");
Expect(0, 7375, '\p{^_ is_Sundanese_Sup}', "");
Expect(0, 7375, '\P{_ is_Sundanese_Sup}', "");
Expect(1, 7375, '\P{^_ is_Sundanese_Sup}', "");
Expect(0, 7376, '\p{_ is_Sundanese_Sup}', "");
Expect(1, 7376, '\p{^_ is_Sundanese_Sup}', "");
Expect(1, 7376, '\P{_ is_Sundanese_Sup}', "");
Expect(0, 7376, '\P{^_ is_Sundanese_Sup}', "");
Error('\p{:=In_SUNDANESE_sup}');
Error('\P{:=In_SUNDANESE_sup}');
Expect(1, 7375, '\p{insundanesesup}', "");
Expect(0, 7375, '\p{^insundanesesup}', "");
Expect(0, 7375, '\P{insundanesesup}', "");
Expect(1, 7375, '\P{^insundanesesup}', "");
Expect(0, 7376, '\p{insundanesesup}', "");
Expect(1, 7376, '\p{^insundanesesup}', "");
Expect(1, 7376, '\P{insundanesesup}', "");
Expect(0, 7376, '\P{^insundanesesup}', "");
Expect(1, 7375, '\p{ 	In_Sundanese_SUP}', "");
Expect(0, 7375, '\p{^ 	In_Sundanese_SUP}', "");
Expect(0, 7375, '\P{ 	In_Sundanese_SUP}', "");
Expect(1, 7375, '\P{^ 	In_Sundanese_SUP}', "");
Expect(0, 7376, '\p{ 	In_Sundanese_SUP}', "");
Expect(1, 7376, '\p{^ 	In_Sundanese_SUP}', "");
Expect(1, 7376, '\P{ 	In_Sundanese_SUP}', "");
Expect(0, 7376, '\P{^ 	In_Sundanese_SUP}', "");
Error('\p{	:=Superscripts_AND_SUBSCRIPTS}');
Error('\P{	:=Superscripts_AND_SUBSCRIPTS}');
Expect(1, 8351, '\p{superscriptsandsubscripts}', "");
Expect(0, 8351, '\p{^superscriptsandsubscripts}', "");
Expect(0, 8351, '\P{superscriptsandsubscripts}', "");
Expect(1, 8351, '\P{^superscriptsandsubscripts}', "");
Expect(0, 8352, '\p{superscriptsandsubscripts}', "");
Expect(1, 8352, '\p{^superscriptsandsubscripts}', "");
Expect(1, 8352, '\P{superscriptsandsubscripts}', "");
Expect(0, 8352, '\P{^superscriptsandsubscripts}', "");
Expect(1, 8351, '\p{ 	Superscripts_And_subscripts}', "");
Expect(0, 8351, '\p{^ 	Superscripts_And_subscripts}', "");
Expect(0, 8351, '\P{ 	Superscripts_And_subscripts}', "");
Expect(1, 8351, '\P{^ 	Superscripts_And_subscripts}', "");
Expect(0, 8352, '\p{ 	Superscripts_And_subscripts}', "");
Expect(1, 8352, '\p{^ 	Superscripts_And_subscripts}', "");
Expect(1, 8352, '\P{ 	Superscripts_And_subscripts}', "");
Expect(0, 8352, '\P{^ 	Superscripts_And_subscripts}', "");
Error('\p{	_Is_Superscripts_And_Subscripts:=}');
Error('\P{	_Is_Superscripts_And_Subscripts:=}');
Expect(1, 8351, '\p{issuperscriptsandsubscripts}', "");
Expect(0, 8351, '\p{^issuperscriptsandsubscripts}', "");
Expect(0, 8351, '\P{issuperscriptsandsubscripts}', "");
Expect(1, 8351, '\P{^issuperscriptsandsubscripts}', "");
Expect(0, 8352, '\p{issuperscriptsandsubscripts}', "");
Expect(1, 8352, '\p{^issuperscriptsandsubscripts}', "");
Expect(1, 8352, '\P{issuperscriptsandsubscripts}', "");
Expect(0, 8352, '\P{^issuperscriptsandsubscripts}', "");
Expect(1, 8351, '\p{_IS_SUPERSCRIPTS_And_Subscripts}', "");
Expect(0, 8351, '\p{^_IS_SUPERSCRIPTS_And_Subscripts}', "");
Expect(0, 8351, '\P{_IS_SUPERSCRIPTS_And_Subscripts}', "");
Expect(1, 8351, '\P{^_IS_SUPERSCRIPTS_And_Subscripts}', "");
Expect(0, 8352, '\p{_IS_SUPERSCRIPTS_And_Subscripts}', "");
Expect(1, 8352, '\p{^_IS_SUPERSCRIPTS_And_Subscripts}', "");
Expect(1, 8352, '\P{_IS_SUPERSCRIPTS_And_Subscripts}', "");
Expect(0, 8352, '\P{^_IS_SUPERSCRIPTS_And_Subscripts}', "");
Error('\p{	 in_Superscripts_And_Subscripts:=}');
Error('\P{	 in_Superscripts_And_Subscripts:=}');
Expect(1, 8351, '\p{insuperscriptsandsubscripts}', "");
Expect(0, 8351, '\p{^insuperscriptsandsubscripts}', "");
Expect(0, 8351, '\P{insuperscriptsandsubscripts}', "");
Expect(1, 8351, '\P{^insuperscriptsandsubscripts}', "");
Expect(0, 8352, '\p{insuperscriptsandsubscripts}', "");
Expect(1, 8352, '\p{^insuperscriptsandsubscripts}', "");
Expect(1, 8352, '\P{insuperscriptsandsubscripts}', "");
Expect(0, 8352, '\P{^insuperscriptsandsubscripts}', "");
Expect(1, 8351, '\p{ _in_Superscripts_AND_Subscripts}', "");
Expect(0, 8351, '\p{^ _in_Superscripts_AND_Subscripts}', "");
Expect(0, 8351, '\P{ _in_Superscripts_AND_Subscripts}', "");
Expect(1, 8351, '\P{^ _in_Superscripts_AND_Subscripts}', "");
Expect(0, 8352, '\p{ _in_Superscripts_AND_Subscripts}', "");
Expect(1, 8352, '\p{^ _in_Superscripts_AND_Subscripts}', "");
Expect(1, 8352, '\P{ _in_Superscripts_AND_Subscripts}', "");
Expect(0, 8352, '\P{^ _in_Superscripts_AND_Subscripts}', "");
Error('\p{/a/	-super_And_sub}');
Error('\P{/a/	-super_And_sub}');
Expect(1, 8351, '\p{superandsub}', "");
Expect(0, 8351, '\p{^superandsub}', "");
Expect(0, 8351, '\P{superandsub}', "");
Expect(1, 8351, '\P{^superandsub}', "");
Expect(0, 8352, '\p{superandsub}', "");
Expect(1, 8352, '\p{^superandsub}', "");
Expect(1, 8352, '\P{superandsub}', "");
Expect(0, 8352, '\P{^superandsub}', "");
Expect(1, 8351, '\p{  super_AND_SUB}', "");
Expect(0, 8351, '\p{^  super_AND_SUB}', "");
Expect(0, 8351, '\P{  super_AND_SUB}', "");
Expect(1, 8351, '\P{^  super_AND_SUB}', "");
Expect(0, 8352, '\p{  super_AND_SUB}', "");
Expect(1, 8352, '\p{^  super_AND_SUB}', "");
Expect(1, 8352, '\P{  super_AND_SUB}', "");
Expect(0, 8352, '\P{^  super_AND_SUB}', "");
Error('\p{ _IS_super_and_Sub:=}');
Error('\P{ _IS_super_and_Sub:=}');
Expect(1, 8351, '\p{issuperandsub}', "");
Expect(0, 8351, '\p{^issuperandsub}', "");
Expect(0, 8351, '\P{issuperandsub}', "");
Expect(1, 8351, '\P{^issuperandsub}', "");
Expect(0, 8352, '\p{issuperandsub}', "");
Expect(1, 8352, '\p{^issuperandsub}', "");
Expect(1, 8352, '\P{issuperandsub}', "");
Expect(0, 8352, '\P{^issuperandsub}', "");
Expect(1, 8351, '\p{  IS_Super_and_SUB}', "");
Expect(0, 8351, '\p{^  IS_Super_and_SUB}', "");
Expect(0, 8351, '\P{  IS_Super_and_SUB}', "");
Expect(1, 8351, '\P{^  IS_Super_and_SUB}', "");
Expect(0, 8352, '\p{  IS_Super_and_SUB}', "");
Expect(1, 8352, '\p{^  IS_Super_and_SUB}', "");
Expect(1, 8352, '\P{  IS_Super_and_SUB}', "");
Expect(0, 8352, '\P{^  IS_Super_and_SUB}', "");
Error('\p{ :=in_Super_And_Sub}');
Error('\P{ :=in_Super_And_Sub}');
Expect(1, 8351, '\p{insuperandsub}', "");
Expect(0, 8351, '\p{^insuperandsub}', "");
Expect(0, 8351, '\P{insuperandsub}', "");
Expect(1, 8351, '\P{^insuperandsub}', "");
Expect(0, 8352, '\p{insuperandsub}', "");
Expect(1, 8352, '\p{^insuperandsub}', "");
Expect(1, 8352, '\P{insuperandsub}', "");
Expect(0, 8352, '\P{^insuperandsub}', "");
Expect(1, 8351, '\p{_IN_Super_AND_SUB}', "");
Expect(0, 8351, '\p{^_IN_Super_AND_SUB}', "");
Expect(0, 8351, '\P{_IN_Super_AND_SUB}', "");
Expect(1, 8351, '\P{^_IN_Super_AND_SUB}', "");
Expect(0, 8352, '\p{_IN_Super_AND_SUB}', "");
Expect(1, 8352, '\p{^_IN_Super_AND_SUB}', "");
Expect(1, 8352, '\P{_IN_Super_AND_SUB}', "");
Expect(0, 8352, '\P{^_IN_Super_AND_SUB}', "");
Error('\p{/a/SUPPLEMENTAL_arrows_A}');
Error('\P{/a/SUPPLEMENTAL_arrows_A}');
Expect(1, 10239, '\p{supplementalarrowsa}', "");
Expect(0, 10239, '\p{^supplementalarrowsa}', "");
Expect(0, 10239, '\P{supplementalarrowsa}', "");
Expect(1, 10239, '\P{^supplementalarrowsa}', "");
Expect(0, 10240, '\p{supplementalarrowsa}', "");
Expect(1, 10240, '\p{^supplementalarrowsa}', "");
Expect(1, 10240, '\P{supplementalarrowsa}', "");
Expect(0, 10240, '\P{^supplementalarrowsa}', "");
Expect(1, 10239, '\p{	Supplemental_Arrows_A}', "");
Expect(0, 10239, '\p{^	Supplemental_Arrows_A}', "");
Expect(0, 10239, '\P{	Supplemental_Arrows_A}', "");
Expect(1, 10239, '\P{^	Supplemental_Arrows_A}', "");
Expect(0, 10240, '\p{	Supplemental_Arrows_A}', "");
Expect(1, 10240, '\p{^	Supplemental_Arrows_A}', "");
Expect(1, 10240, '\P{	Supplemental_Arrows_A}', "");
Expect(0, 10240, '\P{^	Supplemental_Arrows_A}', "");
Error('\p{/a/	Is_supplemental_Arrows_A}');
Error('\P{/a/	Is_supplemental_Arrows_A}');
Expect(1, 10239, '\p{issupplementalarrowsa}', "");
Expect(0, 10239, '\p{^issupplementalarrowsa}', "");
Expect(0, 10239, '\P{issupplementalarrowsa}', "");
Expect(1, 10239, '\P{^issupplementalarrowsa}', "");
Expect(0, 10240, '\p{issupplementalarrowsa}', "");
Expect(1, 10240, '\p{^issupplementalarrowsa}', "");
Expect(1, 10240, '\P{issupplementalarrowsa}', "");
Expect(0, 10240, '\P{^issupplementalarrowsa}', "");
Expect(1, 10239, '\p{	is_SUPPLEMENTAL_arrows_a}', "");
Expect(0, 10239, '\p{^	is_SUPPLEMENTAL_arrows_a}', "");
Expect(0, 10239, '\P{	is_SUPPLEMENTAL_arrows_a}', "");
Expect(1, 10239, '\P{^	is_SUPPLEMENTAL_arrows_a}', "");
Expect(0, 10240, '\p{	is_SUPPLEMENTAL_arrows_a}', "");
Expect(1, 10240, '\p{^	is_SUPPLEMENTAL_arrows_a}', "");
Expect(1, 10240, '\P{	is_SUPPLEMENTAL_arrows_a}', "");
Expect(0, 10240, '\P{^	is_SUPPLEMENTAL_arrows_a}', "");
Error('\p{-_in_Supplemental_Arrows_A:=}');
Error('\P{-_in_Supplemental_Arrows_A:=}');
Expect(1, 10239, '\p{insupplementalarrowsa}', "");
Expect(0, 10239, '\p{^insupplementalarrowsa}', "");
Expect(0, 10239, '\P{insupplementalarrowsa}', "");
Expect(1, 10239, '\P{^insupplementalarrowsa}', "");
Expect(0, 10240, '\p{insupplementalarrowsa}', "");
Expect(1, 10240, '\p{^insupplementalarrowsa}', "");
Expect(1, 10240, '\P{insupplementalarrowsa}', "");
Expect(0, 10240, '\P{^insupplementalarrowsa}', "");
Expect(1, 10239, '\p{_	In_SUPPLEMENTAL_Arrows_A}', "");
Expect(0, 10239, '\p{^_	In_SUPPLEMENTAL_Arrows_A}', "");
Expect(0, 10239, '\P{_	In_SUPPLEMENTAL_Arrows_A}', "");
Expect(1, 10239, '\P{^_	In_SUPPLEMENTAL_Arrows_A}', "");
Expect(0, 10240, '\p{_	In_SUPPLEMENTAL_Arrows_A}', "");
Expect(1, 10240, '\p{^_	In_SUPPLEMENTAL_Arrows_A}', "");
Expect(1, 10240, '\P{_	In_SUPPLEMENTAL_Arrows_A}', "");
Expect(0, 10240, '\P{^_	In_SUPPLEMENTAL_Arrows_A}', "");
Error('\p{:= SUP_Arrows_A}');
Error('\P{:= SUP_Arrows_A}');
Expect(1, 10239, '\p{suparrowsa}', "");
Expect(0, 10239, '\p{^suparrowsa}', "");
Expect(0, 10239, '\P{suparrowsa}', "");
Expect(1, 10239, '\P{^suparrowsa}', "");
Expect(0, 10240, '\p{suparrowsa}', "");
Expect(1, 10240, '\p{^suparrowsa}', "");
Expect(1, 10240, '\P{suparrowsa}', "");
Expect(0, 10240, '\P{^suparrowsa}', "");
Expect(1, 10239, '\p{_	Sup_arrows_A}', "");
Expect(0, 10239, '\p{^_	Sup_arrows_A}', "");
Expect(0, 10239, '\P{_	Sup_arrows_A}', "");
Expect(1, 10239, '\P{^_	Sup_arrows_A}', "");
Expect(0, 10240, '\p{_	Sup_arrows_A}', "");
Expect(1, 10240, '\p{^_	Sup_arrows_A}', "");
Expect(1, 10240, '\P{_	Sup_arrows_A}', "");
Expect(0, 10240, '\P{^_	Sup_arrows_A}', "");
Error('\p{:=_Is_sup_Arrows_A}');
Error('\P{:=_Is_sup_Arrows_A}');
Expect(1, 10239, '\p{issuparrowsa}', "");
Expect(0, 10239, '\p{^issuparrowsa}', "");
Expect(0, 10239, '\P{issuparrowsa}', "");
Expect(1, 10239, '\P{^issuparrowsa}', "");
Expect(0, 10240, '\p{issuparrowsa}', "");
Expect(1, 10240, '\p{^issuparrowsa}', "");
Expect(1, 10240, '\P{issuparrowsa}', "");
Expect(0, 10240, '\P{^issuparrowsa}', "");
Expect(1, 10239, '\p{  IS_Sup_ARROWS_A}', "");
Expect(0, 10239, '\p{^  IS_Sup_ARROWS_A}', "");
Expect(0, 10239, '\P{  IS_Sup_ARROWS_A}', "");
Expect(1, 10239, '\P{^  IS_Sup_ARROWS_A}', "");
Expect(0, 10240, '\p{  IS_Sup_ARROWS_A}', "");
Expect(1, 10240, '\p{^  IS_Sup_ARROWS_A}', "");
Expect(1, 10240, '\P{  IS_Sup_ARROWS_A}', "");
Expect(0, 10240, '\P{^  IS_Sup_ARROWS_A}', "");
Error('\p{	/a/In_SUP_Arrows_A}');
Error('\P{	/a/In_SUP_Arrows_A}');
Expect(1, 10239, '\p{insuparrowsa}', "");
Expect(0, 10239, '\p{^insuparrowsa}', "");
Expect(0, 10239, '\P{insuparrowsa}', "");
Expect(1, 10239, '\P{^insuparrowsa}', "");
Expect(0, 10240, '\p{insuparrowsa}', "");
Expect(1, 10240, '\p{^insuparrowsa}', "");
Expect(1, 10240, '\P{insuparrowsa}', "");
Expect(0, 10240, '\P{^insuparrowsa}', "");
Expect(1, 10239, '\p{_-In_sup_Arrows_A}', "");
Expect(0, 10239, '\p{^_-In_sup_Arrows_A}', "");
Expect(0, 10239, '\P{_-In_sup_Arrows_A}', "");
Expect(1, 10239, '\P{^_-In_sup_Arrows_A}', "");
Expect(0, 10240, '\p{_-In_sup_Arrows_A}', "");
Expect(1, 10240, '\p{^_-In_sup_Arrows_A}', "");
Expect(1, 10240, '\P{_-In_sup_Arrows_A}', "");
Expect(0, 10240, '\P{^_-In_sup_Arrows_A}', "");
Error('\p{/a/- Supplemental_arrows_B}');
Error('\P{/a/- Supplemental_arrows_B}');
Expect(1, 10623, '\p{supplementalarrowsb}', "");
Expect(0, 10623, '\p{^supplementalarrowsb}', "");
Expect(0, 10623, '\P{supplementalarrowsb}', "");
Expect(1, 10623, '\P{^supplementalarrowsb}', "");
Expect(0, 10624, '\p{supplementalarrowsb}', "");
Expect(1, 10624, '\p{^supplementalarrowsb}', "");
Expect(1, 10624, '\P{supplementalarrowsb}', "");
Expect(0, 10624, '\P{^supplementalarrowsb}', "");
Expect(1, 10623, '\p{_SUPPLEMENTAL_arrows_B}', "");
Expect(0, 10623, '\p{^_SUPPLEMENTAL_arrows_B}', "");
Expect(0, 10623, '\P{_SUPPLEMENTAL_arrows_B}', "");
Expect(1, 10623, '\P{^_SUPPLEMENTAL_arrows_B}', "");
Expect(0, 10624, '\p{_SUPPLEMENTAL_arrows_B}', "");
Expect(1, 10624, '\p{^_SUPPLEMENTAL_arrows_B}', "");
Expect(1, 10624, '\P{_SUPPLEMENTAL_arrows_B}', "");
Expect(0, 10624, '\P{^_SUPPLEMENTAL_arrows_B}', "");
Error('\p{/a/	Is_Supplemental_Arrows_B}');
Error('\P{/a/	Is_Supplemental_Arrows_B}');
Expect(1, 10623, '\p{issupplementalarrowsb}', "");
Expect(0, 10623, '\p{^issupplementalarrowsb}', "");
Expect(0, 10623, '\P{issupplementalarrowsb}', "");
Expect(1, 10623, '\P{^issupplementalarrowsb}', "");
Expect(0, 10624, '\p{issupplementalarrowsb}', "");
Expect(1, 10624, '\p{^issupplementalarrowsb}', "");
Expect(1, 10624, '\P{issupplementalarrowsb}', "");
Expect(0, 10624, '\P{^issupplementalarrowsb}', "");
Expect(1, 10623, '\p{-	IS_Supplemental_arrows_B}', "");
Expect(0, 10623, '\p{^-	IS_Supplemental_arrows_B}', "");
Expect(0, 10623, '\P{-	IS_Supplemental_arrows_B}', "");
Expect(1, 10623, '\P{^-	IS_Supplemental_arrows_B}', "");
Expect(0, 10624, '\p{-	IS_Supplemental_arrows_B}', "");
Expect(1, 10624, '\p{^-	IS_Supplemental_arrows_B}', "");
Expect(1, 10624, '\P{-	IS_Supplemental_arrows_B}', "");
Expect(0, 10624, '\P{^-	IS_Supplemental_arrows_B}', "");
Error('\p{/a/		in_SUPPLEMENTAL_ARROWS_b}');
Error('\P{/a/		in_SUPPLEMENTAL_ARROWS_b}');
Expect(1, 10623, '\p{insupplementalarrowsb}', "");
Expect(0, 10623, '\p{^insupplementalarrowsb}', "");
Expect(0, 10623, '\P{insupplementalarrowsb}', "");
Expect(1, 10623, '\P{^insupplementalarrowsb}', "");
Expect(0, 10624, '\p{insupplementalarrowsb}', "");
Expect(1, 10624, '\p{^insupplementalarrowsb}', "");
Expect(1, 10624, '\P{insupplementalarrowsb}', "");
Expect(0, 10624, '\P{^insupplementalarrowsb}', "");
Expect(1, 10623, '\p{  In_supplemental_arrows_B}', "");
Expect(0, 10623, '\p{^  In_supplemental_arrows_B}', "");
Expect(0, 10623, '\P{  In_supplemental_arrows_B}', "");
Expect(1, 10623, '\P{^  In_supplemental_arrows_B}', "");
Expect(0, 10624, '\p{  In_supplemental_arrows_B}', "");
Expect(1, 10624, '\p{^  In_supplemental_arrows_B}', "");
Expect(1, 10624, '\P{  In_supplemental_arrows_B}', "");
Expect(0, 10624, '\P{^  In_supplemental_arrows_B}', "");
Error('\p{/a/-Sup_Arrows_B}');
Error('\P{/a/-Sup_Arrows_B}');
Expect(1, 10623, '\p{suparrowsb}', "");
Expect(0, 10623, '\p{^suparrowsb}', "");
Expect(0, 10623, '\P{suparrowsb}', "");
Expect(1, 10623, '\P{^suparrowsb}', "");
Expect(0, 10624, '\p{suparrowsb}', "");
Expect(1, 10624, '\p{^suparrowsb}', "");
Expect(1, 10624, '\P{suparrowsb}', "");
Expect(0, 10624, '\P{^suparrowsb}', "");
Expect(1, 10623, '\p{	_SUP_ARROWS_b}', "");
Expect(0, 10623, '\p{^	_SUP_ARROWS_b}', "");
Expect(0, 10623, '\P{	_SUP_ARROWS_b}', "");
Expect(1, 10623, '\P{^	_SUP_ARROWS_b}', "");
Expect(0, 10624, '\p{	_SUP_ARROWS_b}', "");
Expect(1, 10624, '\p{^	_SUP_ARROWS_b}', "");
Expect(1, 10624, '\P{	_SUP_ARROWS_b}', "");
Expect(0, 10624, '\P{^	_SUP_ARROWS_b}', "");
Error('\p{_ IS_sup_arrows_B/a/}');
Error('\P{_ IS_sup_arrows_B/a/}');
Expect(1, 10623, '\p{issuparrowsb}', "");
Expect(0, 10623, '\p{^issuparrowsb}', "");
Expect(0, 10623, '\P{issuparrowsb}', "");
Expect(1, 10623, '\P{^issuparrowsb}', "");
Expect(0, 10624, '\p{issuparrowsb}', "");
Expect(1, 10624, '\p{^issuparrowsb}', "");
Expect(1, 10624, '\P{issuparrowsb}', "");
Expect(0, 10624, '\P{^issuparrowsb}', "");
Expect(1, 10623, '\p{ -Is_Sup_Arrows_b}', "");
Expect(0, 10623, '\p{^ -Is_Sup_Arrows_b}', "");
Expect(0, 10623, '\P{ -Is_Sup_Arrows_b}', "");
Expect(1, 10623, '\P{^ -Is_Sup_Arrows_b}', "");
Expect(0, 10624, '\p{ -Is_Sup_Arrows_b}', "");
Expect(1, 10624, '\p{^ -Is_Sup_Arrows_b}', "");
Expect(1, 10624, '\P{ -Is_Sup_Arrows_b}', "");
Expect(0, 10624, '\P{^ -Is_Sup_Arrows_b}', "");
Error('\p{ :=in_SUP_Arrows_B}');
Error('\P{ :=in_SUP_Arrows_B}');
Expect(1, 10623, '\p{insuparrowsb}', "");
Expect(0, 10623, '\p{^insuparrowsb}', "");
Expect(0, 10623, '\P{insuparrowsb}', "");
Expect(1, 10623, '\P{^insuparrowsb}', "");
Expect(0, 10624, '\p{insuparrowsb}', "");
Expect(1, 10624, '\p{^insuparrowsb}', "");
Expect(1, 10624, '\P{insuparrowsb}', "");
Expect(0, 10624, '\P{^insuparrowsb}', "");
Expect(1, 10623, '\p{-In_Sup_arrows_B}', "");
Expect(0, 10623, '\p{^-In_Sup_arrows_B}', "");
Expect(0, 10623, '\P{-In_Sup_arrows_B}', "");
Expect(1, 10623, '\P{^-In_Sup_arrows_B}', "");
Expect(0, 10624, '\p{-In_Sup_arrows_B}', "");
Expect(1, 10624, '\p{^-In_Sup_arrows_B}', "");
Expect(1, 10624, '\P{-In_Sup_arrows_B}', "");
Expect(0, 10624, '\P{^-In_Sup_arrows_B}', "");
Error('\p{/a/Supplemental_Mathematical_Operators}');
Error('\P{/a/Supplemental_Mathematical_Operators}');
Expect(1, 11007, '\p{supplementalmathematicaloperators}', "");
Expect(0, 11007, '\p{^supplementalmathematicaloperators}', "");
Expect(0, 11007, '\P{supplementalmathematicaloperators}', "");
Expect(1, 11007, '\P{^supplementalmathematicaloperators}', "");
Expect(0, 11008, '\p{supplementalmathematicaloperators}', "");
Expect(1, 11008, '\p{^supplementalmathematicaloperators}', "");
Expect(1, 11008, '\P{supplementalmathematicaloperators}', "");
Expect(0, 11008, '\P{^supplementalmathematicaloperators}', "");
Expect(1, 11007, '\p{	-SUPPLEMENTAL_Mathematical_Operators}', "");
Expect(0, 11007, '\p{^	-SUPPLEMENTAL_Mathematical_Operators}', "");
Expect(0, 11007, '\P{	-SUPPLEMENTAL_Mathematical_Operators}', "");
Expect(1, 11007, '\P{^	-SUPPLEMENTAL_Mathematical_Operators}', "");
Expect(0, 11008, '\p{	-SUPPLEMENTAL_Mathematical_Operators}', "");
Expect(1, 11008, '\p{^	-SUPPLEMENTAL_Mathematical_Operators}', "");
Expect(1, 11008, '\P{	-SUPPLEMENTAL_Mathematical_Operators}', "");
Expect(0, 11008, '\P{^	-SUPPLEMENTAL_Mathematical_Operators}', "");
Error('\p{/a/	-IS_supplemental_MATHEMATICAL_Operators}');
Error('\P{/a/	-IS_supplemental_MATHEMATICAL_Operators}');
Expect(1, 11007, '\p{issupplementalmathematicaloperators}', "");
Expect(0, 11007, '\p{^issupplementalmathematicaloperators}', "");
Expect(0, 11007, '\P{issupplementalmathematicaloperators}', "");
Expect(1, 11007, '\P{^issupplementalmathematicaloperators}', "");
Expect(0, 11008, '\p{issupplementalmathematicaloperators}', "");
Expect(1, 11008, '\p{^issupplementalmathematicaloperators}', "");
Expect(1, 11008, '\P{issupplementalmathematicaloperators}', "");
Expect(0, 11008, '\P{^issupplementalmathematicaloperators}', "");
Expect(1, 11007, '\p{-Is_supplemental_mathematical_operators}', "");
Expect(0, 11007, '\p{^-Is_supplemental_mathematical_operators}', "");
Expect(0, 11007, '\P{-Is_supplemental_mathematical_operators}', "");
Expect(1, 11007, '\P{^-Is_supplemental_mathematical_operators}', "");
Expect(0, 11008, '\p{-Is_supplemental_mathematical_operators}', "");
Expect(1, 11008, '\p{^-Is_supplemental_mathematical_operators}', "");
Expect(1, 11008, '\P{-Is_supplemental_mathematical_operators}', "");
Expect(0, 11008, '\P{^-Is_supplemental_mathematical_operators}', "");
Error('\p{:= _In_Supplemental_MATHEMATICAL_Operators}');
Error('\P{:= _In_Supplemental_MATHEMATICAL_Operators}');
Expect(1, 11007, '\p{insupplementalmathematicaloperators}', "");
Expect(0, 11007, '\p{^insupplementalmathematicaloperators}', "");
Expect(0, 11007, '\P{insupplementalmathematicaloperators}', "");
Expect(1, 11007, '\P{^insupplementalmathematicaloperators}', "");
Expect(0, 11008, '\p{insupplementalmathematicaloperators}', "");
Expect(1, 11008, '\p{^insupplementalmathematicaloperators}', "");
Expect(1, 11008, '\P{insupplementalmathematicaloperators}', "");
Expect(0, 11008, '\P{^insupplementalmathematicaloperators}', "");
Expect(1, 11007, '\p{	-In_supplemental_MATHEMATICAL_Operators}', "");
Expect(0, 11007, '\p{^	-In_supplemental_MATHEMATICAL_Operators}', "");
Expect(0, 11007, '\P{	-In_supplemental_MATHEMATICAL_Operators}', "");
Expect(1, 11007, '\P{^	-In_supplemental_MATHEMATICAL_Operators}', "");
Expect(0, 11008, '\p{	-In_supplemental_MATHEMATICAL_Operators}', "");
Expect(1, 11008, '\p{^	-In_supplemental_MATHEMATICAL_Operators}', "");
Expect(1, 11008, '\P{	-In_supplemental_MATHEMATICAL_Operators}', "");
Expect(0, 11008, '\P{^	-In_supplemental_MATHEMATICAL_Operators}', "");
Error('\p{:=  sup_math_Operators}');
Error('\P{:=  sup_math_Operators}');
Expect(1, 11007, '\p{supmathoperators}', "");
Expect(0, 11007, '\p{^supmathoperators}', "");
Expect(0, 11007, '\P{supmathoperators}', "");
Expect(1, 11007, '\P{^supmathoperators}', "");
Expect(0, 11008, '\p{supmathoperators}', "");
Expect(1, 11008, '\p{^supmathoperators}', "");
Expect(1, 11008, '\P{supmathoperators}', "");
Expect(0, 11008, '\P{^supmathoperators}', "");
Expect(1, 11007, '\p{__SUP_Math_Operators}', "");
Expect(0, 11007, '\p{^__SUP_Math_Operators}', "");
Expect(0, 11007, '\P{__SUP_Math_Operators}', "");
Expect(1, 11007, '\P{^__SUP_Math_Operators}', "");
Expect(0, 11008, '\p{__SUP_Math_Operators}', "");
Expect(1, 11008, '\p{^__SUP_Math_Operators}', "");
Expect(1, 11008, '\P{__SUP_Math_Operators}', "");
Expect(0, 11008, '\P{^__SUP_Math_Operators}', "");
Error('\p{/a/IS_sup_math_operators}');
Error('\P{/a/IS_sup_math_operators}');
Expect(1, 11007, '\p{issupmathoperators}', "");
Expect(0, 11007, '\p{^issupmathoperators}', "");
Expect(0, 11007, '\P{issupmathoperators}', "");
Expect(1, 11007, '\P{^issupmathoperators}', "");
Expect(0, 11008, '\p{issupmathoperators}', "");
Expect(1, 11008, '\p{^issupmathoperators}', "");
Expect(1, 11008, '\P{issupmathoperators}', "");
Expect(0, 11008, '\P{^issupmathoperators}', "");
Expect(1, 11007, '\p{	-Is_SUP_Math_OPERATORS}', "");
Expect(0, 11007, '\p{^	-Is_SUP_Math_OPERATORS}', "");
Expect(0, 11007, '\P{	-Is_SUP_Math_OPERATORS}', "");
Expect(1, 11007, '\P{^	-Is_SUP_Math_OPERATORS}', "");
Expect(0, 11008, '\p{	-Is_SUP_Math_OPERATORS}', "");
Expect(1, 11008, '\p{^	-Is_SUP_Math_OPERATORS}', "");
Expect(1, 11008, '\P{	-Is_SUP_Math_OPERATORS}', "");
Expect(0, 11008, '\P{^	-Is_SUP_Math_OPERATORS}', "");
Error('\p{	In_Sup_Math_OPERATORS/a/}');
Error('\P{	In_Sup_Math_OPERATORS/a/}');
Expect(1, 11007, '\p{insupmathoperators}', "");
Expect(0, 11007, '\p{^insupmathoperators}', "");
Expect(0, 11007, '\P{insupmathoperators}', "");
Expect(1, 11007, '\P{^insupmathoperators}', "");
Expect(0, 11008, '\p{insupmathoperators}', "");
Expect(1, 11008, '\p{^insupmathoperators}', "");
Expect(1, 11008, '\P{insupmathoperators}', "");
Expect(0, 11008, '\P{^insupmathoperators}', "");
Expect(1, 11007, '\p{ -In_Sup_math_Operators}', "");
Expect(0, 11007, '\p{^ -In_Sup_math_Operators}', "");
Expect(0, 11007, '\P{ -In_Sup_math_Operators}', "");
Expect(1, 11007, '\P{^ -In_Sup_math_Operators}', "");
Expect(0, 11008, '\p{ -In_Sup_math_Operators}', "");
Expect(1, 11008, '\p{^ -In_Sup_math_Operators}', "");
Expect(1, 11008, '\P{ -In_Sup_math_Operators}', "");
Expect(0, 11008, '\P{^ -In_Sup_math_Operators}', "");
Error('\p{/a/--Supplemental_Punctuation}');
Error('\P{/a/--Supplemental_Punctuation}');
Expect(1, 11903, '\p{supplementalpunctuation}', "");
Expect(0, 11903, '\p{^supplementalpunctuation}', "");
Expect(0, 11903, '\P{supplementalpunctuation}', "");
Expect(1, 11903, '\P{^supplementalpunctuation}', "");
Expect(0, 11904, '\p{supplementalpunctuation}', "");
Expect(1, 11904, '\p{^supplementalpunctuation}', "");
Expect(1, 11904, '\P{supplementalpunctuation}', "");
Expect(0, 11904, '\P{^supplementalpunctuation}', "");
Expect(1, 11903, '\p{_SUPPLEMENTAL_Punctuation}', "");
Expect(0, 11903, '\p{^_SUPPLEMENTAL_Punctuation}', "");
Expect(0, 11903, '\P{_SUPPLEMENTAL_Punctuation}', "");
Expect(1, 11903, '\P{^_SUPPLEMENTAL_Punctuation}', "");
Expect(0, 11904, '\p{_SUPPLEMENTAL_Punctuation}', "");
Expect(1, 11904, '\p{^_SUPPLEMENTAL_Punctuation}', "");
Expect(1, 11904, '\P{_SUPPLEMENTAL_Punctuation}', "");
Expect(0, 11904, '\P{^_SUPPLEMENTAL_Punctuation}', "");
Error('\p{ /a/Is_Supplemental_Punctuation}');
Error('\P{ /a/Is_Supplemental_Punctuation}');
Expect(1, 11903, '\p{issupplementalpunctuation}', "");
Expect(0, 11903, '\p{^issupplementalpunctuation}', "");
Expect(0, 11903, '\P{issupplementalpunctuation}', "");
Expect(1, 11903, '\P{^issupplementalpunctuation}', "");
Expect(0, 11904, '\p{issupplementalpunctuation}', "");
Expect(1, 11904, '\p{^issupplementalpunctuation}', "");
Expect(1, 11904, '\P{issupplementalpunctuation}', "");
Expect(0, 11904, '\P{^issupplementalpunctuation}', "");
Expect(1, 11903, '\p{--Is_Supplemental_PUNCTUATION}', "");
Expect(0, 11903, '\p{^--Is_Supplemental_PUNCTUATION}', "");
Expect(0, 11903, '\P{--Is_Supplemental_PUNCTUATION}', "");
Expect(1, 11903, '\P{^--Is_Supplemental_PUNCTUATION}', "");
Expect(0, 11904, '\p{--Is_Supplemental_PUNCTUATION}', "");
Expect(1, 11904, '\p{^--Is_Supplemental_PUNCTUATION}', "");
Expect(1, 11904, '\P{--Is_Supplemental_PUNCTUATION}', "");
Expect(0, 11904, '\P{^--Is_Supplemental_PUNCTUATION}', "");
Error('\p{_ in_SUPPLEMENTAL_PUNCTUATION:=}');
Error('\P{_ in_SUPPLEMENTAL_PUNCTUATION:=}');
Expect(1, 11903, '\p{insupplementalpunctuation}', "");
Expect(0, 11903, '\p{^insupplementalpunctuation}', "");
Expect(0, 11903, '\P{insupplementalpunctuation}', "");
Expect(1, 11903, '\P{^insupplementalpunctuation}', "");
Expect(0, 11904, '\p{insupplementalpunctuation}', "");
Expect(1, 11904, '\p{^insupplementalpunctuation}', "");
Expect(1, 11904, '\P{insupplementalpunctuation}', "");
Expect(0, 11904, '\P{^insupplementalpunctuation}', "");
Expect(1, 11903, '\p{_-in_SUPPLEMENTAL_PUNCTUATION}', "");
Expect(0, 11903, '\p{^_-in_SUPPLEMENTAL_PUNCTUATION}', "");
Expect(0, 11903, '\P{_-in_SUPPLEMENTAL_PUNCTUATION}', "");
Expect(1, 11903, '\P{^_-in_SUPPLEMENTAL_PUNCTUATION}', "");
Expect(0, 11904, '\p{_-in_SUPPLEMENTAL_PUNCTUATION}', "");
Expect(1, 11904, '\p{^_-in_SUPPLEMENTAL_PUNCTUATION}', "");
Expect(1, 11904, '\P{_-in_SUPPLEMENTAL_PUNCTUATION}', "");
Expect(0, 11904, '\P{^_-in_SUPPLEMENTAL_PUNCTUATION}', "");
Error('\p{-sup_PUNCTUATION/a/}');
Error('\P{-sup_PUNCTUATION/a/}');
Expect(1, 11903, '\p{suppunctuation}', "");
Expect(0, 11903, '\p{^suppunctuation}', "");
Expect(0, 11903, '\P{suppunctuation}', "");
Expect(1, 11903, '\P{^suppunctuation}', "");
Expect(0, 11904, '\p{suppunctuation}', "");
Expect(1, 11904, '\p{^suppunctuation}', "");
Expect(1, 11904, '\P{suppunctuation}', "");
Expect(0, 11904, '\P{^suppunctuation}', "");
Expect(1, 11903, '\p{	-Sup_Punctuation}', "");
Expect(0, 11903, '\p{^	-Sup_Punctuation}', "");
Expect(0, 11903, '\P{	-Sup_Punctuation}', "");
Expect(1, 11903, '\P{^	-Sup_Punctuation}', "");
Expect(0, 11904, '\p{	-Sup_Punctuation}', "");
Expect(1, 11904, '\p{^	-Sup_Punctuation}', "");
Expect(1, 11904, '\P{	-Sup_Punctuation}', "");
Expect(0, 11904, '\P{^	-Sup_Punctuation}', "");
Error('\p{_/a/Is_sup_punctuation}');
Error('\P{_/a/Is_sup_punctuation}');
Expect(1, 11903, '\p{issuppunctuation}', "");
Expect(0, 11903, '\p{^issuppunctuation}', "");
Expect(0, 11903, '\P{issuppunctuation}', "");
Expect(1, 11903, '\P{^issuppunctuation}', "");
Expect(0, 11904, '\p{issuppunctuation}', "");
Expect(1, 11904, '\p{^issuppunctuation}', "");
Expect(1, 11904, '\P{issuppunctuation}', "");
Expect(0, 11904, '\P{^issuppunctuation}', "");
Expect(1, 11903, '\p{	_Is_Sup_Punctuation}', "");
Expect(0, 11903, '\p{^	_Is_Sup_Punctuation}', "");
Expect(0, 11903, '\P{	_Is_Sup_Punctuation}', "");
Expect(1, 11903, '\P{^	_Is_Sup_Punctuation}', "");
Expect(0, 11904, '\p{	_Is_Sup_Punctuation}', "");
Expect(1, 11904, '\p{^	_Is_Sup_Punctuation}', "");
Expect(1, 11904, '\P{	_Is_Sup_Punctuation}', "");
Expect(0, 11904, '\P{^	_Is_Sup_Punctuation}', "");
Error('\p{:=	 in_Sup_Punctuation}');
Error('\P{:=	 in_Sup_Punctuation}');
Expect(1, 11903, '\p{insuppunctuation}', "");
Expect(0, 11903, '\p{^insuppunctuation}', "");
Expect(0, 11903, '\P{insuppunctuation}', "");
Expect(1, 11903, '\P{^insuppunctuation}', "");
Expect(0, 11904, '\p{insuppunctuation}', "");
Expect(1, 11904, '\p{^insuppunctuation}', "");
Expect(1, 11904, '\P{insuppunctuation}', "");
Expect(0, 11904, '\P{^insuppunctuation}', "");
Expect(1, 11903, '\p{_ In_SUP_punctuation}', "");
Expect(0, 11903, '\p{^_ In_SUP_punctuation}', "");
Expect(0, 11903, '\P{_ In_SUP_punctuation}', "");
Expect(1, 11903, '\P{^_ In_SUP_punctuation}', "");
Expect(0, 11904, '\p{_ In_SUP_punctuation}', "");
Expect(1, 11904, '\p{^_ In_SUP_punctuation}', "");
Expect(1, 11904, '\P{_ In_SUP_punctuation}', "");
Expect(0, 11904, '\P{^_ In_SUP_punctuation}', "");
Error('\p{--supplementary_Private_USE_Area_a/a/}');
Error('\P{--supplementary_Private_USE_Area_a/a/}');
Expect(1, 983040, '\p{supplementaryprivateuseareaa}', "");
Expect(0, 983040, '\p{^supplementaryprivateuseareaa}', "");
Expect(0, 983040, '\P{supplementaryprivateuseareaa}', "");
Expect(1, 983040, '\P{^supplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\p{supplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\p{^supplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\P{supplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\P{^supplementaryprivateuseareaa}', "");
Expect(1, 983040, '\p{Supplementary_Private_Use_Area_A}', "");
Expect(0, 983040, '\p{^Supplementary_Private_Use_Area_A}', "");
Expect(0, 983040, '\P{Supplementary_Private_Use_Area_A}', "");
Expect(1, 983040, '\P{^Supplementary_Private_Use_Area_A}', "");
Expect(0, 1048576, '\p{Supplementary_Private_Use_Area_A}', "");
Expect(1, 1048576, '\p{^Supplementary_Private_Use_Area_A}', "");
Expect(1, 1048576, '\P{Supplementary_Private_Use_Area_A}', "");
Expect(0, 1048576, '\P{^Supplementary_Private_Use_Area_A}', "");
Error('\p{--Is_Supplementary_private_Use_AREA_A:=}');
Error('\P{--Is_Supplementary_private_Use_AREA_A:=}');
Expect(1, 983040, '\p{issupplementaryprivateuseareaa}', "");
Expect(0, 983040, '\p{^issupplementaryprivateuseareaa}', "");
Expect(0, 983040, '\P{issupplementaryprivateuseareaa}', "");
Expect(1, 983040, '\P{^issupplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\p{issupplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\p{^issupplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\P{issupplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\P{^issupplementaryprivateuseareaa}', "");
Expect(1, 983040, '\p{		is_Supplementary_PRIVATE_Use_area_A}', "");
Expect(0, 983040, '\p{^		is_Supplementary_PRIVATE_Use_area_A}', "");
Expect(0, 983040, '\P{		is_Supplementary_PRIVATE_Use_area_A}', "");
Expect(1, 983040, '\P{^		is_Supplementary_PRIVATE_Use_area_A}', "");
Expect(0, 1048576, '\p{		is_Supplementary_PRIVATE_Use_area_A}', "");
Expect(1, 1048576, '\p{^		is_Supplementary_PRIVATE_Use_area_A}', "");
Expect(1, 1048576, '\P{		is_Supplementary_PRIVATE_Use_area_A}', "");
Expect(0, 1048576, '\P{^		is_Supplementary_PRIVATE_Use_area_A}', "");
Error('\p{In_supplementary_Private_Use_Area_a/a/}');
Error('\P{In_supplementary_Private_Use_Area_a/a/}');
Expect(1, 983040, '\p{insupplementaryprivateuseareaa}', "");
Expect(0, 983040, '\p{^insupplementaryprivateuseareaa}', "");
Expect(0, 983040, '\P{insupplementaryprivateuseareaa}', "");
Expect(1, 983040, '\P{^insupplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\p{insupplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\p{^insupplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\P{insupplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\P{^insupplementaryprivateuseareaa}', "");
Expect(1, 983040, '\p{-_In_SUPPLEMENTARY_PRIVATE_USE_area_A}', "");
Expect(0, 983040, '\p{^-_In_SUPPLEMENTARY_PRIVATE_USE_area_A}', "");
Expect(0, 983040, '\P{-_In_SUPPLEMENTARY_PRIVATE_USE_area_A}', "");
Expect(1, 983040, '\P{^-_In_SUPPLEMENTARY_PRIVATE_USE_area_A}', "");
Expect(0, 1048576, '\p{-_In_SUPPLEMENTARY_PRIVATE_USE_area_A}', "");
Expect(1, 1048576, '\p{^-_In_SUPPLEMENTARY_PRIVATE_USE_area_A}', "");
Expect(1, 1048576, '\P{-_In_SUPPLEMENTARY_PRIVATE_USE_area_A}', "");
Expect(0, 1048576, '\P{^-_In_SUPPLEMENTARY_PRIVATE_USE_area_A}', "");
Error('\p{:= _SUP_PUA_A}');
Error('\P{:= _SUP_PUA_A}');
Expect(1, 983040, '\p{suppuaa}', "");
Expect(0, 983040, '\p{^suppuaa}', "");
Expect(0, 983040, '\P{suppuaa}', "");
Expect(1, 983040, '\P{^suppuaa}', "");
Expect(0, 1048576, '\p{suppuaa}', "");
Expect(1, 1048576, '\p{^suppuaa}', "");
Expect(1, 1048576, '\P{suppuaa}', "");
Expect(0, 1048576, '\P{^suppuaa}', "");
Expect(1, 983040, '\p{-_Sup_PUA_a}', "");
Expect(0, 983040, '\p{^-_Sup_PUA_a}', "");
Expect(0, 983040, '\P{-_Sup_PUA_a}', "");
Expect(1, 983040, '\P{^-_Sup_PUA_a}', "");
Expect(0, 1048576, '\p{-_Sup_PUA_a}', "");
Expect(1, 1048576, '\p{^-_Sup_PUA_a}', "");
Expect(1, 1048576, '\P{-_Sup_PUA_a}', "");
Expect(0, 1048576, '\P{^-_Sup_PUA_a}', "");
Error('\p{	/a/Is_SUP_PUA_A}');
Error('\P{	/a/Is_SUP_PUA_A}');
Expect(1, 983040, '\p{issuppuaa}', "");
Expect(0, 983040, '\p{^issuppuaa}', "");
Expect(0, 983040, '\P{issuppuaa}', "");
Expect(1, 983040, '\P{^issuppuaa}', "");
Expect(0, 1048576, '\p{issuppuaa}', "");
Expect(1, 1048576, '\p{^issuppuaa}', "");
Expect(1, 1048576, '\P{issuppuaa}', "");
Expect(0, 1048576, '\P{^issuppuaa}', "");
Expect(1, 983040, '\p{ -is_Sup_PUA_a}', "");
Expect(0, 983040, '\p{^ -is_Sup_PUA_a}', "");
Expect(0, 983040, '\P{ -is_Sup_PUA_a}', "");
Expect(1, 983040, '\P{^ -is_Sup_PUA_a}', "");
Expect(0, 1048576, '\p{ -is_Sup_PUA_a}', "");
Expect(1, 1048576, '\p{^ -is_Sup_PUA_a}', "");
Expect(1, 1048576, '\P{ -is_Sup_PUA_a}', "");
Expect(0, 1048576, '\P{^ -is_Sup_PUA_a}', "");
Error('\p{:=	 In_Sup_pua_a}');
Error('\P{:=	 In_Sup_pua_a}');
Expect(1, 983040, '\p{insuppuaa}', "");
Expect(0, 983040, '\p{^insuppuaa}', "");
Expect(0, 983040, '\P{insuppuaa}', "");
Expect(1, 983040, '\P{^insuppuaa}', "");
Expect(0, 1048576, '\p{insuppuaa}', "");
Expect(1, 1048576, '\p{^insuppuaa}', "");
Expect(1, 1048576, '\P{insuppuaa}', "");
Expect(0, 1048576, '\P{^insuppuaa}', "");
Expect(1, 983040, '\p{-IN_sup_PUA_A}', "");
Expect(0, 983040, '\p{^-IN_sup_PUA_A}', "");
Expect(0, 983040, '\P{-IN_sup_PUA_A}', "");
Expect(1, 983040, '\P{^-IN_sup_PUA_A}', "");
Expect(0, 1048576, '\p{-IN_sup_PUA_A}', "");
Expect(1, 1048576, '\p{^-IN_sup_PUA_A}', "");
Expect(1, 1048576, '\P{-IN_sup_PUA_A}', "");
Expect(0, 1048576, '\P{^-IN_sup_PUA_A}', "");
Error('\p{ _SUPPLEMENTARY_Private_USE_Area_B/a/}');
Error('\P{ _SUPPLEMENTARY_Private_USE_Area_B/a/}');
Expect(1, 1048576, '\p{supplementaryprivateuseareab}', "");
Expect(0, 1048576, '\p{^supplementaryprivateuseareab}', "");
Expect(0, 1048576, '\P{supplementaryprivateuseareab}', "");
Expect(1, 1048576, '\P{^supplementaryprivateuseareab}', "");
Expect(0, 1, '\p{supplementaryprivateuseareab}', "");
Expect(1, 1, '\p{^supplementaryprivateuseareab}', "");
Expect(1, 1, '\P{supplementaryprivateuseareab}', "");
Expect(0, 1, '\P{^supplementaryprivateuseareab}', "");
Expect(1, 1048576, '\p{ SUPPLEMENTARY_Private_use_AREA_B}', "");
Expect(0, 1048576, '\p{^ SUPPLEMENTARY_Private_use_AREA_B}', "");
Expect(0, 1048576, '\P{ SUPPLEMENTARY_Private_use_AREA_B}', "");
Expect(1, 1048576, '\P{^ SUPPLEMENTARY_Private_use_AREA_B}', "");
Expect(0, 1, '\p{ SUPPLEMENTARY_Private_use_AREA_B}', "");
Expect(1, 1, '\p{^ SUPPLEMENTARY_Private_use_AREA_B}', "");
Expect(1, 1, '\P{ SUPPLEMENTARY_Private_use_AREA_B}', "");
Expect(0, 1, '\P{^ SUPPLEMENTARY_Private_use_AREA_B}', "");
Error('\p{:=Is_supplementary_PRIVATE_USE_Area_B}');
Error('\P{:=Is_supplementary_PRIVATE_USE_Area_B}');
Expect(1, 1048576, '\p{issupplementaryprivateuseareab}', "");
Expect(0, 1048576, '\p{^issupplementaryprivateuseareab}', "");
Expect(0, 1048576, '\P{issupplementaryprivateuseareab}', "");
Expect(1, 1048576, '\P{^issupplementaryprivateuseareab}', "");
Expect(0, 1, '\p{issupplementaryprivateuseareab}', "");
Expect(1, 1, '\p{^issupplementaryprivateuseareab}', "");
Expect(1, 1, '\P{issupplementaryprivateuseareab}', "");
Expect(0, 1, '\P{^issupplementaryprivateuseareab}', "");
Expect(1, 1048576, '\p{_Is_Supplementary_private_use_AREA_B}', "");
Expect(0, 1048576, '\p{^_Is_Supplementary_private_use_AREA_B}', "");
Expect(0, 1048576, '\P{_Is_Supplementary_private_use_AREA_B}', "");
Expect(1, 1048576, '\P{^_Is_Supplementary_private_use_AREA_B}', "");
Expect(0, 1, '\p{_Is_Supplementary_private_use_AREA_B}', "");
Expect(1, 1, '\p{^_Is_Supplementary_private_use_AREA_B}', "");
Expect(1, 1, '\P{_Is_Supplementary_private_use_AREA_B}', "");
Expect(0, 1, '\P{^_Is_Supplementary_private_use_AREA_B}', "");
Error('\p{/a/in_SUPPLEMENTARY_private_Use_area_b}');
Error('\P{/a/in_SUPPLEMENTARY_private_Use_area_b}');
Expect(1, 1048576, '\p{insupplementaryprivateuseareab}', "");
Expect(0, 1048576, '\p{^insupplementaryprivateuseareab}', "");
Expect(0, 1048576, '\P{insupplementaryprivateuseareab}', "");
Expect(1, 1048576, '\P{^insupplementaryprivateuseareab}', "");
Expect(0, 1, '\p{insupplementaryprivateuseareab}', "");
Expect(1, 1, '\p{^insupplementaryprivateuseareab}', "");
Expect(1, 1, '\P{insupplementaryprivateuseareab}', "");
Expect(0, 1, '\P{^insupplementaryprivateuseareab}', "");
Expect(1, 1048576, '\p{ in_Supplementary_private_Use_AREA_B}', "");
Expect(0, 1048576, '\p{^ in_Supplementary_private_Use_AREA_B}', "");
Expect(0, 1048576, '\P{ in_Supplementary_private_Use_AREA_B}', "");
Expect(1, 1048576, '\P{^ in_Supplementary_private_Use_AREA_B}', "");
Expect(0, 1, '\p{ in_Supplementary_private_Use_AREA_B}', "");
Expect(1, 1, '\p{^ in_Supplementary_private_Use_AREA_B}', "");
Expect(1, 1, '\P{ in_Supplementary_private_Use_AREA_B}', "");
Expect(0, 1, '\P{^ in_Supplementary_private_Use_AREA_B}', "");
Error('\p{/a/-sup_pua_B}');
Error('\P{/a/-sup_pua_B}');
Expect(1, 1048576, '\p{suppuab}', "");
Expect(0, 1048576, '\p{^suppuab}', "");
Expect(0, 1048576, '\P{suppuab}', "");
Expect(1, 1048576, '\P{^suppuab}', "");
Expect(0, 1, '\p{suppuab}', "");
Expect(1, 1, '\p{^suppuab}', "");
Expect(1, 1, '\P{suppuab}', "");
Expect(0, 1, '\P{^suppuab}', "");
Expect(1, 1048576, '\p{ Sup_PUA_B}', "");
Expect(0, 1048576, '\p{^ Sup_PUA_B}', "");
Expect(0, 1048576, '\P{ Sup_PUA_B}', "");
Expect(1, 1048576, '\P{^ Sup_PUA_B}', "");
Expect(0, 1, '\p{ Sup_PUA_B}', "");
Expect(1, 1, '\p{^ Sup_PUA_B}', "");
Expect(1, 1, '\P{ Sup_PUA_B}', "");
Expect(0, 1, '\P{^ Sup_PUA_B}', "");
Error('\p{_IS_SUP_PUA_b/a/}');
Error('\P{_IS_SUP_PUA_b/a/}');
Expect(1, 1048576, '\p{issuppuab}', "");
Expect(0, 1048576, '\p{^issuppuab}', "");
Expect(0, 1048576, '\P{issuppuab}', "");
Expect(1, 1048576, '\P{^issuppuab}', "");
Expect(0, 1, '\p{issuppuab}', "");
Expect(1, 1, '\p{^issuppuab}', "");
Expect(1, 1, '\P{issuppuab}', "");
Expect(0, 1, '\P{^issuppuab}', "");
Expect(1, 1048576, '\p{is_sup_PUA_B}', "");
Expect(0, 1048576, '\p{^is_sup_PUA_B}', "");
Expect(0, 1048576, '\P{is_sup_PUA_B}', "");
Expect(1, 1048576, '\P{^is_sup_PUA_B}', "");
Expect(0, 1, '\p{is_sup_PUA_B}', "");
Expect(1, 1, '\p{^is_sup_PUA_B}', "");
Expect(1, 1, '\P{is_sup_PUA_B}', "");
Expect(0, 1, '\P{^is_sup_PUA_B}', "");
Error('\p{/a/	-In_Sup_PUA_B}');
Error('\P{/a/	-In_Sup_PUA_B}');
Expect(1, 1048576, '\p{insuppuab}', "");
Expect(0, 1048576, '\p{^insuppuab}', "");
Expect(0, 1048576, '\P{insuppuab}', "");
Expect(1, 1048576, '\P{^insuppuab}', "");
Expect(0, 1, '\p{insuppuab}', "");
Expect(1, 1, '\p{^insuppuab}', "");
Expect(1, 1, '\P{insuppuab}', "");
Expect(0, 1, '\P{^insuppuab}', "");
Expect(1, 1048576, '\p{- in_sup_PUA_B}', "");
Expect(0, 1048576, '\p{^- in_sup_PUA_B}', "");
Expect(0, 1048576, '\P{- in_sup_PUA_B}', "");
Expect(1, 1048576, '\P{^- in_sup_PUA_B}', "");
Expect(0, 1, '\p{- in_sup_PUA_B}', "");
Expect(1, 1, '\p{^- in_sup_PUA_B}', "");
Expect(1, 1, '\P{- in_sup_PUA_B}', "");
Expect(0, 1, '\P{^- in_sup_PUA_B}', "");
Error('\p{_:=Surrogate}');
Error('\P{_:=Surrogate}');
Expect(1, 57343, '\p{surrogate}', "");
Expect(0, 57343, '\p{^surrogate}', "");
Expect(0, 57343, '\P{surrogate}', "");
Expect(1, 57343, '\P{^surrogate}', "");
Expect(0, 57344, '\p{surrogate}', "");
Expect(1, 57344, '\p{^surrogate}', "");
Expect(1, 57344, '\P{surrogate}', "");
Expect(0, 57344, '\P{^surrogate}', "");
Expect(1, 57343, '\p{ _Surrogate}', "");
Expect(0, 57343, '\p{^ _Surrogate}', "");
Expect(0, 57343, '\P{ _Surrogate}', "");
Expect(1, 57343, '\P{^ _Surrogate}', "");
Expect(0, 57344, '\p{ _Surrogate}', "");
Expect(1, 57344, '\p{^ _Surrogate}', "");
Expect(1, 57344, '\P{ _Surrogate}', "");
Expect(0, 57344, '\P{^ _Surrogate}', "");
Error('\p{-:=is_Surrogate}');
Error('\P{-:=is_Surrogate}');
Expect(1, 57343, '\p{issurrogate}', "");
Expect(0, 57343, '\p{^issurrogate}', "");
Expect(0, 57343, '\P{issurrogate}', "");
Expect(1, 57343, '\P{^issurrogate}', "");
Expect(0, 57344, '\p{issurrogate}', "");
Expect(1, 57344, '\p{^issurrogate}', "");
Expect(1, 57344, '\P{issurrogate}', "");
Expect(0, 57344, '\P{^issurrogate}', "");
Expect(1, 57343, '\p{ -is_Surrogate}', "");
Expect(0, 57343, '\p{^ -is_Surrogate}', "");
Expect(0, 57343, '\P{ -is_Surrogate}', "");
Expect(1, 57343, '\P{^ -is_Surrogate}', "");
Expect(0, 57344, '\p{ -is_Surrogate}', "");
Expect(1, 57344, '\p{^ -is_Surrogate}', "");
Expect(1, 57344, '\P{ -is_Surrogate}', "");
Expect(0, 57344, '\P{^ -is_Surrogate}', "");
Error('\p{_:=Cs}');
Error('\P{_:=Cs}');
Expect(1, 57343, '\p{cs}', "");
Expect(0, 57343, '\p{^cs}', "");
Expect(0, 57343, '\P{cs}', "");
Expect(1, 57343, '\P{^cs}', "");
Expect(0, 57344, '\p{cs}', "");
Expect(1, 57344, '\p{^cs}', "");
Expect(1, 57344, '\P{cs}', "");
Expect(0, 57344, '\P{^cs}', "");
Expect(1, 57343, '\p{	_cs}', "");
Expect(0, 57343, '\p{^	_cs}', "");
Expect(0, 57343, '\P{	_cs}', "");
Expect(1, 57343, '\P{^	_cs}', "");
Expect(0, 57344, '\p{	_cs}', "");
Expect(1, 57344, '\p{^	_cs}', "");
Expect(1, 57344, '\P{	_cs}', "");
Expect(0, 57344, '\P{^	_cs}', "");
Error('\p{/a/-	Is_Cs}');
Error('\P{/a/-	Is_Cs}');
Expect(1, 57343, '\p{iscs}', "");
Expect(0, 57343, '\p{^iscs}', "");
Expect(0, 57343, '\P{iscs}', "");
Expect(1, 57343, '\P{^iscs}', "");
Expect(0, 57344, '\p{iscs}', "");
Expect(1, 57344, '\p{^iscs}', "");
Expect(1, 57344, '\P{iscs}', "");
Expect(0, 57344, '\P{^iscs}', "");
Expect(1, 57343, '\p{ _Is_CS}', "");
Expect(0, 57343, '\p{^ _Is_CS}', "");
Expect(0, 57343, '\P{ _Is_CS}', "");
Expect(1, 57343, '\P{^ _Is_CS}', "");
Expect(0, 57344, '\p{ _Is_CS}', "");
Expect(1, 57344, '\p{^ _Is_CS}', "");
Expect(1, 57344, '\P{ _Is_CS}', "");
Expect(0, 57344, '\P{^ _Is_CS}', "");
Error('\p{ -syloti_Nagri/a/}');
Error('\P{ -syloti_Nagri/a/}');
Expect(1, 43051, '\p{sylotinagri}', "");
Expect(0, 43051, '\p{^sylotinagri}', "");
Expect(0, 43051, '\P{sylotinagri}', "");
Expect(1, 43051, '\P{^sylotinagri}', "");
Expect(0, 43052, '\p{sylotinagri}', "");
Expect(1, 43052, '\p{^sylotinagri}', "");
Expect(1, 43052, '\P{sylotinagri}', "");
Expect(0, 43052, '\P{^sylotinagri}', "");
Expect(1, 43051, '\p{  SYLOTI_nagri}', "");
Expect(0, 43051, '\p{^  SYLOTI_nagri}', "");
Expect(0, 43051, '\P{  SYLOTI_nagri}', "");
Expect(1, 43051, '\P{^  SYLOTI_nagri}', "");
Expect(0, 43052, '\p{  SYLOTI_nagri}', "");
Expect(1, 43052, '\p{^  SYLOTI_nagri}', "");
Expect(1, 43052, '\P{  SYLOTI_nagri}', "");
Expect(0, 43052, '\P{^  SYLOTI_nagri}', "");
Error('\p{ /a/Is_SYLOTI_NAGRI}');
Error('\P{ /a/Is_SYLOTI_NAGRI}');
Expect(1, 43051, '\p{issylotinagri}', "");
Expect(0, 43051, '\p{^issylotinagri}', "");
Expect(0, 43051, '\P{issylotinagri}', "");
Expect(1, 43051, '\P{^issylotinagri}', "");
Expect(0, 43052, '\p{issylotinagri}', "");
Expect(1, 43052, '\p{^issylotinagri}', "");
Expect(1, 43052, '\P{issylotinagri}', "");
Expect(0, 43052, '\P{^issylotinagri}', "");
Expect(1, 43051, '\p{-Is_SYLOTI_Nagri}', "");
Expect(0, 43051, '\p{^-Is_SYLOTI_Nagri}', "");
Expect(0, 43051, '\P{-Is_SYLOTI_Nagri}', "");
Expect(1, 43051, '\P{^-Is_SYLOTI_Nagri}', "");
Expect(0, 43052, '\p{-Is_SYLOTI_Nagri}', "");
Expect(1, 43052, '\p{^-Is_SYLOTI_Nagri}', "");
Expect(1, 43052, '\P{-Is_SYLOTI_Nagri}', "");
Expect(0, 43052, '\P{^-Is_SYLOTI_Nagri}', "");
Error('\p{_/a/Sylo}');
Error('\P{_/a/Sylo}');
Expect(1, 43051, '\p{sylo}', "");
Expect(0, 43051, '\p{^sylo}', "");
Expect(0, 43051, '\P{sylo}', "");
Expect(1, 43051, '\P{^sylo}', "");
Expect(0, 43052, '\p{sylo}', "");
Expect(1, 43052, '\p{^sylo}', "");
Expect(1, 43052, '\P{sylo}', "");
Expect(0, 43052, '\P{^sylo}', "");
Expect(1, 43051, '\p{- Sylo}', "");
Expect(0, 43051, '\p{^- Sylo}', "");
Expect(0, 43051, '\P{- Sylo}', "");
Expect(1, 43051, '\P{^- Sylo}', "");
Expect(0, 43052, '\p{- Sylo}', "");
Expect(1, 43052, '\p{^- Sylo}', "");
Expect(1, 43052, '\P{- Sylo}', "");
Expect(0, 43052, '\P{^- Sylo}', "");
Error('\p{/a/ 	Is_Sylo}');
Error('\P{/a/ 	Is_Sylo}');
Expect(1, 43051, '\p{issylo}', "");
Expect(0, 43051, '\p{^issylo}', "");
Expect(0, 43051, '\P{issylo}', "");
Expect(1, 43051, '\P{^issylo}', "");
Expect(0, 43052, '\p{issylo}', "");
Expect(1, 43052, '\p{^issylo}', "");
Expect(1, 43052, '\P{issylo}', "");
Expect(0, 43052, '\P{^issylo}', "");
Expect(1, 43051, '\p{_ Is_SYLO}', "");
Expect(0, 43051, '\p{^_ Is_SYLO}', "");
Expect(0, 43051, '\P{_ Is_SYLO}', "");
Expect(1, 43051, '\P{^_ Is_SYLO}', "");
Expect(0, 43052, '\p{_ Is_SYLO}', "");
Expect(1, 43052, '\p{^_ Is_SYLO}', "");
Expect(1, 43052, '\P{_ Is_SYLO}', "");
Expect(0, 43052, '\P{^_ Is_SYLO}', "");
Error('\p{:=_	symbol}');
Error('\P{:=_	symbol}');
Expect(1, 128883, '\p{symbol}', "");
Expect(0, 128883, '\p{^symbol}', "");
Expect(0, 128883, '\P{symbol}', "");
Expect(1, 128883, '\P{^symbol}', "");
Expect(0, 128884, '\p{symbol}', "");
Expect(1, 128884, '\p{^symbol}', "");
Expect(1, 128884, '\P{symbol}', "");
Expect(0, 128884, '\P{^symbol}', "");
Expect(1, 128883, '\p{- Symbol}', "");
Expect(0, 128883, '\p{^- Symbol}', "");
Expect(0, 128883, '\P{- Symbol}', "");
Expect(1, 128883, '\P{^- Symbol}', "");
Expect(0, 128884, '\p{- Symbol}', "");
Expect(1, 128884, '\p{^- Symbol}', "");
Expect(1, 128884, '\P{- Symbol}', "");
Expect(0, 128884, '\P{^- Symbol}', "");
Error('\p{_ Is_symbol:=}');
Error('\P{_ Is_symbol:=}');
Expect(1, 128883, '\p{issymbol}', "");
Expect(0, 128883, '\p{^issymbol}', "");
Expect(0, 128883, '\P{issymbol}', "");
Expect(1, 128883, '\P{^issymbol}', "");
Expect(0, 128884, '\p{issymbol}', "");
Expect(1, 128884, '\p{^issymbol}', "");
Expect(1, 128884, '\P{issymbol}', "");
Expect(0, 128884, '\P{^issymbol}', "");
Expect(1, 128883, '\p{	 is_Symbol}', "");
Expect(0, 128883, '\p{^	 is_Symbol}', "");
Expect(0, 128883, '\P{	 is_Symbol}', "");
Expect(1, 128883, '\P{^	 is_Symbol}', "");
Expect(0, 128884, '\p{	 is_Symbol}', "");
Expect(1, 128884, '\p{^	 is_Symbol}', "");
Expect(1, 128884, '\P{	 is_Symbol}', "");
Expect(0, 128884, '\P{^	 is_Symbol}', "");
Error('\p{ S/a/}');
Error('\P{ S/a/}');
Expect(1, 128883, '\p{s}', "");
Expect(0, 128883, '\p{^s}', "");
Expect(0, 128883, '\P{s}', "");
Expect(1, 128883, '\P{^s}', "");
Expect(0, 128884, '\p{s}', "");
Expect(1, 128884, '\p{^s}', "");
Expect(1, 128884, '\P{s}', "");
Expect(0, 128884, '\P{^s}', "");
Expect(1, 128883, '\p{ s}', "");
Expect(0, 128883, '\p{^ s}', "");
Expect(0, 128883, '\P{ s}', "");
Expect(1, 128883, '\P{^ s}', "");
Expect(0, 128884, '\p{ s}', "");
Expect(1, 128884, '\p{^ s}', "");
Expect(1, 128884, '\P{ s}', "");
Expect(0, 128884, '\P{^ s}', "");
Error('\p{-	IS_S/a/}');
Error('\P{-	IS_S/a/}');
Expect(1, 128883, '\p{iss}', "");
Expect(0, 128883, '\p{^iss}', "");
Expect(0, 128883, '\P{iss}', "");
Expect(1, 128883, '\P{^iss}', "");
Expect(0, 128884, '\p{iss}', "");
Expect(1, 128884, '\p{^iss}', "");
Expect(1, 128884, '\P{iss}', "");
Expect(0, 128884, '\P{^iss}', "");
Expect(1, 128883, '\p{-	is_S}', "");
Expect(0, 128883, '\p{^-	is_S}', "");
Expect(0, 128883, '\P{-	is_S}', "");
Expect(1, 128883, '\P{^-	is_S}', "");
Expect(0, 128884, '\p{-	is_S}', "");
Expect(1, 128884, '\p{^-	is_S}', "");
Expect(1, 128884, '\P{-	is_S}', "");
Expect(0, 128884, '\P{^-	is_S}', "");
Error('\p{ -Syriac/a/}');
Error('\P{ -Syriac/a/}');
Expect(1, 1871, '\p{syriac}', "");
Expect(0, 1871, '\p{^syriac}', "");
Expect(0, 1871, '\P{syriac}', "");
Expect(1, 1871, '\P{^syriac}', "");
Expect(0, 1872, '\p{syriac}', "");
Expect(1, 1872, '\p{^syriac}', "");
Expect(1, 1872, '\P{syriac}', "");
Expect(0, 1872, '\P{^syriac}', "");
Expect(1, 1871, '\p{_syriac}', "");
Expect(0, 1871, '\p{^_syriac}', "");
Expect(0, 1871, '\P{_syriac}', "");
Expect(1, 1871, '\P{^_syriac}', "");
Expect(0, 1872, '\p{_syriac}', "");
Expect(1, 1872, '\p{^_syriac}', "");
Expect(1, 1872, '\P{_syriac}', "");
Expect(0, 1872, '\P{^_syriac}', "");
Error('\p{_/a/Is_SYRIAC}');
Error('\P{_/a/Is_SYRIAC}');
Expect(1, 1871, '\p{issyriac}', "");
Expect(0, 1871, '\p{^issyriac}', "");
Expect(0, 1871, '\P{issyriac}', "");
Expect(1, 1871, '\P{^issyriac}', "");
Expect(0, 1872, '\p{issyriac}', "");
Expect(1, 1872, '\p{^issyriac}', "");
Expect(1, 1872, '\P{issyriac}', "");
Expect(0, 1872, '\P{^issyriac}', "");
Expect(1, 1871, '\p{_IS_Syriac}', "");
Expect(0, 1871, '\p{^_IS_Syriac}', "");
Expect(0, 1871, '\P{_IS_Syriac}', "");
Expect(1, 1871, '\P{^_IS_Syriac}', "");
Expect(0, 1872, '\p{_IS_Syriac}', "");
Expect(1, 1872, '\p{^_IS_Syriac}', "");
Expect(1, 1872, '\P{_IS_Syriac}', "");
Expect(0, 1872, '\P{^_IS_Syriac}', "");
Error('\p{/a/_	Syrc}');
Error('\P{/a/_	Syrc}');
Expect(1, 1871, '\p{syrc}', "");
Expect(0, 1871, '\p{^syrc}', "");
Expect(0, 1871, '\P{syrc}', "");
Expect(1, 1871, '\P{^syrc}', "");
Expect(0, 1872, '\p{syrc}', "");
Expect(1, 1872, '\p{^syrc}', "");
Expect(1, 1872, '\P{syrc}', "");
Expect(0, 1872, '\P{^syrc}', "");
Expect(1, 1871, '\p{ SYRC}', "");
Expect(0, 1871, '\p{^ SYRC}', "");
Expect(0, 1871, '\P{ SYRC}', "");
Expect(1, 1871, '\P{^ SYRC}', "");
Expect(0, 1872, '\p{ SYRC}', "");
Expect(1, 1872, '\p{^ SYRC}', "");
Expect(1, 1872, '\P{ SYRC}', "");
Expect(0, 1872, '\P{^ SYRC}', "");
Error('\p{/a/Is_syrc}');
Error('\P{/a/Is_syrc}');
Expect(1, 1871, '\p{issyrc}', "");
Expect(0, 1871, '\p{^issyrc}', "");
Expect(0, 1871, '\P{issyrc}', "");
Expect(1, 1871, '\P{^issyrc}', "");
Expect(0, 1872, '\p{issyrc}', "");
Expect(1, 1872, '\p{^issyrc}', "");
Expect(1, 1872, '\P{issyrc}', "");
Expect(0, 1872, '\P{^issyrc}', "");
Expect(1, 1871, '\p{	is_SYRC}', "");
Expect(0, 1871, '\p{^	is_SYRC}', "");
Expect(0, 1871, '\P{	is_SYRC}', "");
Expect(1, 1871, '\P{^	is_SYRC}', "");
Expect(0, 1872, '\p{	is_SYRC}', "");
Expect(1, 1872, '\p{^	is_SYRC}', "");
Expect(1, 1872, '\P{	is_SYRC}', "");
Expect(0, 1872, '\P{^	is_SYRC}', "");
Error('\p{ _Tagalog/a/}');
Error('\P{ _Tagalog/a/}');
Expect(1, 5908, '\p{tagalog}', "");
Expect(0, 5908, '\p{^tagalog}', "");
Expect(0, 5908, '\P{tagalog}', "");
Expect(1, 5908, '\P{^tagalog}', "");
Expect(0, 5909, '\p{tagalog}', "");
Expect(1, 5909, '\p{^tagalog}', "");
Expect(1, 5909, '\P{tagalog}', "");
Expect(0, 5909, '\P{^tagalog}', "");
Expect(1, 5908, '\p{ tagalog}', "");
Expect(0, 5908, '\p{^ tagalog}', "");
Expect(0, 5908, '\P{ tagalog}', "");
Expect(1, 5908, '\P{^ tagalog}', "");
Expect(0, 5909, '\p{ tagalog}', "");
Expect(1, 5909, '\p{^ tagalog}', "");
Expect(1, 5909, '\P{ tagalog}', "");
Expect(0, 5909, '\P{^ tagalog}', "");
Error('\p{/a/Is_Tagalog}');
Error('\P{/a/Is_Tagalog}');
Expect(1, 5908, '\p{istagalog}', "");
Expect(0, 5908, '\p{^istagalog}', "");
Expect(0, 5908, '\P{istagalog}', "");
Expect(1, 5908, '\P{^istagalog}', "");
Expect(0, 5909, '\p{istagalog}', "");
Expect(1, 5909, '\p{^istagalog}', "");
Expect(1, 5909, '\P{istagalog}', "");
Expect(0, 5909, '\P{^istagalog}', "");
Expect(1, 5908, '\p{_	is_TAGALOG}', "");
Expect(0, 5908, '\p{^_	is_TAGALOG}', "");
Expect(0, 5908, '\P{_	is_TAGALOG}', "");
Expect(1, 5908, '\P{^_	is_TAGALOG}', "");
Expect(0, 5909, '\p{_	is_TAGALOG}', "");
Expect(1, 5909, '\p{^_	is_TAGALOG}', "");
Expect(1, 5909, '\P{_	is_TAGALOG}', "");
Expect(0, 5909, '\P{^_	is_TAGALOG}', "");
Error('\p{_-TGLG/a/}');
Error('\P{_-TGLG/a/}');
Expect(1, 5908, '\p{tglg}', "");
Expect(0, 5908, '\p{^tglg}', "");
Expect(0, 5908, '\P{tglg}', "");
Expect(1, 5908, '\P{^tglg}', "");
Expect(0, 5909, '\p{tglg}', "");
Expect(1, 5909, '\p{^tglg}', "");
Expect(1, 5909, '\P{tglg}', "");
Expect(0, 5909, '\P{^tglg}', "");
Expect(1, 5908, '\p{_	Tglg}', "");
Expect(0, 5908, '\p{^_	Tglg}', "");
Expect(0, 5908, '\P{_	Tglg}', "");
Expect(1, 5908, '\P{^_	Tglg}', "");
Expect(0, 5909, '\p{_	Tglg}', "");
Expect(1, 5909, '\p{^_	Tglg}', "");
Expect(1, 5909, '\P{_	Tglg}', "");
Expect(0, 5909, '\P{^_	Tglg}', "");
Error('\p{_Is_Tglg:=}');
Error('\P{_Is_Tglg:=}');
Expect(1, 5908, '\p{istglg}', "");
Expect(0, 5908, '\p{^istglg}', "");
Expect(0, 5908, '\P{istglg}', "");
Expect(1, 5908, '\P{^istglg}', "");
Expect(0, 5909, '\p{istglg}', "");
Expect(1, 5909, '\p{^istglg}', "");
Expect(1, 5909, '\P{istglg}', "");
Expect(0, 5909, '\P{^istglg}', "");
Expect(1, 5908, '\p{ Is_tglg}', "");
Expect(0, 5908, '\p{^ Is_tglg}', "");
Expect(0, 5908, '\P{ Is_tglg}', "");
Expect(1, 5908, '\P{^ Is_tglg}', "");
Expect(0, 5909, '\p{ Is_tglg}', "");
Expect(1, 5909, '\p{^ Is_tglg}', "");
Expect(1, 5909, '\P{ Is_tglg}', "");
Expect(0, 5909, '\P{^ Is_tglg}', "");
Error('\p{:=  TAGBANWA}');
Error('\P{:=  TAGBANWA}');
Expect(1, 6003, '\p{tagbanwa}', "");
Expect(0, 6003, '\p{^tagbanwa}', "");
Expect(0, 6003, '\P{tagbanwa}', "");
Expect(1, 6003, '\P{^tagbanwa}', "");
Expect(0, 6004, '\p{tagbanwa}', "");
Expect(1, 6004, '\p{^tagbanwa}', "");
Expect(1, 6004, '\P{tagbanwa}', "");
Expect(0, 6004, '\P{^tagbanwa}', "");
Expect(1, 6003, '\p{  Tagbanwa}', "");
Expect(0, 6003, '\p{^  Tagbanwa}', "");
Expect(0, 6003, '\P{  Tagbanwa}', "");
Expect(1, 6003, '\P{^  Tagbanwa}', "");
Expect(0, 6004, '\p{  Tagbanwa}', "");
Expect(1, 6004, '\p{^  Tagbanwa}', "");
Expect(1, 6004, '\P{  Tagbanwa}', "");
Expect(0, 6004, '\P{^  Tagbanwa}', "");
Error('\p{:=-is_Tagbanwa}');
Error('\P{:=-is_Tagbanwa}');
Expect(1, 6003, '\p{istagbanwa}', "");
Expect(0, 6003, '\p{^istagbanwa}', "");
Expect(0, 6003, '\P{istagbanwa}', "");
Expect(1, 6003, '\P{^istagbanwa}', "");
Expect(0, 6004, '\p{istagbanwa}', "");
Expect(1, 6004, '\p{^istagbanwa}', "");
Expect(1, 6004, '\P{istagbanwa}', "");
Expect(0, 6004, '\P{^istagbanwa}', "");
Expect(1, 6003, '\p{ Is_Tagbanwa}', "");
Expect(0, 6003, '\p{^ Is_Tagbanwa}', "");
Expect(0, 6003, '\P{ Is_Tagbanwa}', "");
Expect(1, 6003, '\P{^ Is_Tagbanwa}', "");
Expect(0, 6004, '\p{ Is_Tagbanwa}', "");
Expect(1, 6004, '\p{^ Is_Tagbanwa}', "");
Expect(1, 6004, '\P{ Is_Tagbanwa}', "");
Expect(0, 6004, '\P{^ Is_Tagbanwa}', "");
Error('\p{:=	Tagb}');
Error('\P{:=	Tagb}');
Expect(1, 6003, '\p{tagb}', "");
Expect(0, 6003, '\p{^tagb}', "");
Expect(0, 6003, '\P{tagb}', "");
Expect(1, 6003, '\P{^tagb}', "");
Expect(0, 6004, '\p{tagb}', "");
Expect(1, 6004, '\p{^tagb}', "");
Expect(1, 6004, '\P{tagb}', "");
Expect(0, 6004, '\P{^tagb}', "");
Expect(1, 6003, '\p{ 	tagb}', "");
Expect(0, 6003, '\p{^ 	tagb}', "");
Expect(0, 6003, '\P{ 	tagb}', "");
Expect(1, 6003, '\P{^ 	tagb}', "");
Expect(0, 6004, '\p{ 	tagb}', "");
Expect(1, 6004, '\p{^ 	tagb}', "");
Expect(1, 6004, '\P{ 	tagb}', "");
Expect(0, 6004, '\P{^ 	tagb}', "");
Error('\p{/a/-_IS_Tagb}');
Error('\P{/a/-_IS_Tagb}');
Expect(1, 6003, '\p{istagb}', "");
Expect(0, 6003, '\p{^istagb}', "");
Expect(0, 6003, '\P{istagb}', "");
Expect(1, 6003, '\P{^istagb}', "");
Expect(0, 6004, '\p{istagb}', "");
Expect(1, 6004, '\p{^istagb}', "");
Expect(1, 6004, '\P{istagb}', "");
Expect(0, 6004, '\P{^istagb}', "");
Expect(1, 6003, '\p{	_IS_TAGB}', "");
Expect(0, 6003, '\p{^	_IS_TAGB}', "");
Expect(0, 6003, '\P{	_IS_TAGB}', "");
Expect(1, 6003, '\P{^	_IS_TAGB}', "");
Expect(0, 6004, '\p{	_IS_TAGB}', "");
Expect(1, 6004, '\p{^	_IS_TAGB}', "");
Expect(1, 6004, '\P{	_IS_TAGB}', "");
Expect(0, 6004, '\P{^	_IS_TAGB}', "");
Error('\p{-_Tags/a/}');
Error('\P{-_Tags/a/}');
Expect(1, 917631, '\p{tags}', "");
Expect(0, 917631, '\p{^tags}', "");
Expect(0, 917631, '\P{tags}', "");
Expect(1, 917631, '\P{^tags}', "");
Expect(0, 917632, '\p{tags}', "");
Expect(1, 917632, '\p{^tags}', "");
Expect(1, 917632, '\P{tags}', "");
Expect(0, 917632, '\P{^tags}', "");
Expect(1, 917631, '\p{ 	Tags}', "");
Expect(0, 917631, '\p{^ 	Tags}', "");
Expect(0, 917631, '\P{ 	Tags}', "");
Expect(1, 917631, '\P{^ 	Tags}', "");
Expect(0, 917632, '\p{ 	Tags}', "");
Expect(1, 917632, '\p{^ 	Tags}', "");
Expect(1, 917632, '\P{ 	Tags}', "");
Expect(0, 917632, '\P{^ 	Tags}', "");
Error('\p{_/a/Is_tags}');
Error('\P{_/a/Is_tags}');
Expect(1, 917631, '\p{istags}', "");
Expect(0, 917631, '\p{^istags}', "");
Expect(0, 917631, '\P{istags}', "");
Expect(1, 917631, '\P{^istags}', "");
Expect(0, 917632, '\p{istags}', "");
Expect(1, 917632, '\p{^istags}', "");
Expect(1, 917632, '\P{istags}', "");
Expect(0, 917632, '\P{^istags}', "");
Expect(1, 917631, '\p{	is_Tags}', "");
Expect(0, 917631, '\p{^	is_Tags}', "");
Expect(0, 917631, '\P{	is_Tags}', "");
Expect(1, 917631, '\P{^	is_Tags}', "");
Expect(0, 917632, '\p{	is_Tags}', "");
Expect(1, 917632, '\p{^	is_Tags}', "");
Expect(1, 917632, '\P{	is_Tags}', "");
Expect(0, 917632, '\P{^	is_Tags}', "");
Error('\p{-	In_TAGS/a/}');
Error('\P{-	In_TAGS/a/}');
Expect(1, 917631, '\p{intags}', "");
Expect(0, 917631, '\p{^intags}', "");
Expect(0, 917631, '\P{intags}', "");
Expect(1, 917631, '\P{^intags}', "");
Expect(0, 917632, '\p{intags}', "");
Expect(1, 917632, '\p{^intags}', "");
Expect(1, 917632, '\P{intags}', "");
Expect(0, 917632, '\P{^intags}', "");
Expect(1, 917631, '\p{ 	In_Tags}', "");
Expect(0, 917631, '\p{^ 	In_Tags}', "");
Expect(0, 917631, '\P{ 	In_Tags}', "");
Expect(1, 917631, '\P{^ 	In_Tags}', "");
Expect(0, 917632, '\p{ 	In_Tags}', "");
Expect(1, 917632, '\p{^ 	In_Tags}', "");
Expect(1, 917632, '\P{ 	In_Tags}', "");
Expect(0, 917632, '\P{^ 	In_Tags}', "");
Error('\p{_:=Tai_Le}');
Error('\P{_:=Tai_Le}');
Expect(1, 6516, '\p{taile}', "");
Expect(0, 6516, '\p{^taile}', "");
Expect(0, 6516, '\P{taile}', "");
Expect(1, 6516, '\P{^taile}', "");
Expect(0, 6517, '\p{taile}', "");
Expect(1, 6517, '\p{^taile}', "");
Expect(1, 6517, '\P{taile}', "");
Expect(0, 6517, '\P{^taile}', "");
Expect(1, 6516, '\p{ Tai_LE}', "");
Expect(0, 6516, '\p{^ Tai_LE}', "");
Expect(0, 6516, '\P{ Tai_LE}', "");
Expect(1, 6516, '\P{^ Tai_LE}', "");
Expect(0, 6517, '\p{ Tai_LE}', "");
Expect(1, 6517, '\p{^ Tai_LE}', "");
Expect(1, 6517, '\P{ Tai_LE}', "");
Expect(0, 6517, '\P{^ Tai_LE}', "");
Error('\p{_-IS_tai_Le/a/}');
Error('\P{_-IS_tai_Le/a/}');
Expect(1, 6516, '\p{istaile}', "");
Expect(0, 6516, '\p{^istaile}', "");
Expect(0, 6516, '\P{istaile}', "");
Expect(1, 6516, '\P{^istaile}', "");
Expect(0, 6517, '\p{istaile}', "");
Expect(1, 6517, '\p{^istaile}', "");
Expect(1, 6517, '\P{istaile}', "");
Expect(0, 6517, '\P{^istaile}', "");
Expect(1, 6516, '\p{-	Is_tai_Le}', "");
Expect(0, 6516, '\p{^-	Is_tai_Le}', "");
Expect(0, 6516, '\P{-	Is_tai_Le}', "");
Expect(1, 6516, '\P{^-	Is_tai_Le}', "");
Expect(0, 6517, '\p{-	Is_tai_Le}', "");
Expect(1, 6517, '\p{^-	Is_tai_Le}', "");
Expect(1, 6517, '\P{-	Is_tai_Le}', "");
Expect(0, 6517, '\P{^-	Is_tai_Le}', "");
Error('\p{:=tale}');
Error('\P{:=tale}');
Expect(1, 6516, '\p{tale}', "");
Expect(0, 6516, '\p{^tale}', "");
Expect(0, 6516, '\P{tale}', "");
Expect(1, 6516, '\P{^tale}', "");
Expect(0, 6517, '\p{tale}', "");
Expect(1, 6517, '\p{^tale}', "");
Expect(1, 6517, '\P{tale}', "");
Expect(0, 6517, '\P{^tale}', "");
Expect(1, 6516, '\p{--tale}', "");
Expect(0, 6516, '\p{^--tale}', "");
Expect(0, 6516, '\P{--tale}', "");
Expect(1, 6516, '\P{^--tale}', "");
Expect(0, 6517, '\p{--tale}', "");
Expect(1, 6517, '\p{^--tale}', "");
Expect(1, 6517, '\P{--tale}', "");
Expect(0, 6517, '\P{^--tale}', "");
Error('\p{-_IS_tale:=}');
Error('\P{-_IS_tale:=}');
Expect(1, 6516, '\p{istale}', "");
Expect(0, 6516, '\p{^istale}', "");
Expect(0, 6516, '\P{istale}', "");
Expect(1, 6516, '\P{^istale}', "");
Expect(0, 6517, '\p{istale}', "");
Expect(1, 6517, '\p{^istale}', "");
Expect(1, 6517, '\P{istale}', "");
Expect(0, 6517, '\P{^istale}', "");
Expect(1, 6516, '\p{ Is_tale}', "");
Expect(0, 6516, '\p{^ Is_tale}', "");
Expect(0, 6516, '\P{ Is_tale}', "");
Expect(1, 6516, '\P{^ Is_tale}', "");
Expect(0, 6517, '\p{ Is_tale}', "");
Expect(1, 6517, '\p{^ Is_tale}', "");
Expect(1, 6517, '\P{ Is_tale}', "");
Expect(0, 6517, '\P{^ Is_tale}', "");
Error('\p{_-Tai_tham:=}');
Error('\P{_-Tai_tham:=}');
Expect(1, 6829, '\p{taitham}', "");
Expect(0, 6829, '\p{^taitham}', "");
Expect(0, 6829, '\P{taitham}', "");
Expect(1, 6829, '\P{^taitham}', "");
Expect(0, 6830, '\p{taitham}', "");
Expect(1, 6830, '\p{^taitham}', "");
Expect(1, 6830, '\P{taitham}', "");
Expect(0, 6830, '\P{^taitham}', "");
Expect(1, 6829, '\p{		tai_Tham}', "");
Expect(0, 6829, '\p{^		tai_Tham}', "");
Expect(0, 6829, '\P{		tai_Tham}', "");
Expect(1, 6829, '\P{^		tai_Tham}', "");
Expect(0, 6830, '\p{		tai_Tham}', "");
Expect(1, 6830, '\p{^		tai_Tham}', "");
Expect(1, 6830, '\P{		tai_Tham}', "");
Expect(0, 6830, '\P{^		tai_Tham}', "");
Error('\p{		Is_Tai_THAM:=}');
Error('\P{		Is_Tai_THAM:=}');
Expect(1, 6829, '\p{istaitham}', "");
Expect(0, 6829, '\p{^istaitham}', "");
Expect(0, 6829, '\P{istaitham}', "");
Expect(1, 6829, '\P{^istaitham}', "");
Expect(0, 6830, '\p{istaitham}', "");
Expect(1, 6830, '\p{^istaitham}', "");
Expect(1, 6830, '\P{istaitham}', "");
Expect(0, 6830, '\P{^istaitham}', "");
Expect(1, 6829, '\p{-	Is_Tai_Tham}', "");
Expect(0, 6829, '\p{^-	Is_Tai_Tham}', "");
Expect(0, 6829, '\P{-	Is_Tai_Tham}', "");
Expect(1, 6829, '\P{^-	Is_Tai_Tham}', "");
Expect(0, 6830, '\p{-	Is_Tai_Tham}', "");
Expect(1, 6830, '\p{^-	Is_Tai_Tham}', "");
Expect(1, 6830, '\P{-	Is_Tai_Tham}', "");
Expect(0, 6830, '\P{^-	Is_Tai_Tham}', "");
Error('\p{:=		lana}');
Error('\P{:=		lana}');
Expect(1, 6829, '\p{lana}', "");
Expect(0, 6829, '\p{^lana}', "");
Expect(0, 6829, '\P{lana}', "");
Expect(1, 6829, '\P{^lana}', "");
Expect(0, 6830, '\p{lana}', "");
Expect(1, 6830, '\p{^lana}', "");
Expect(1, 6830, '\P{lana}', "");
Expect(0, 6830, '\P{^lana}', "");
Expect(1, 6829, '\p{  Lana}', "");
Expect(0, 6829, '\p{^  Lana}', "");
Expect(0, 6829, '\P{  Lana}', "");
Expect(1, 6829, '\P{^  Lana}', "");
Expect(0, 6830, '\p{  Lana}', "");
Expect(1, 6830, '\p{^  Lana}', "");
Expect(1, 6830, '\P{  Lana}', "");
Expect(0, 6830, '\P{^  Lana}', "");
Error('\p{	_Is_Lana:=}');
Error('\P{	_Is_Lana:=}');
Expect(1, 6829, '\p{islana}', "");
Expect(0, 6829, '\p{^islana}', "");
Expect(0, 6829, '\P{islana}', "");
Expect(1, 6829, '\P{^islana}', "");
Expect(0, 6830, '\p{islana}', "");
Expect(1, 6830, '\p{^islana}', "");
Expect(1, 6830, '\P{islana}', "");
Expect(0, 6830, '\P{^islana}', "");
Expect(1, 6829, '\p{	_Is_lana}', "");
Expect(0, 6829, '\p{^	_Is_lana}', "");
Expect(0, 6829, '\P{	_Is_lana}', "");
Expect(1, 6829, '\P{^	_Is_lana}', "");
Expect(0, 6830, '\p{	_Is_lana}', "");
Expect(1, 6830, '\p{^	_Is_lana}', "");
Expect(1, 6830, '\P{	_Is_lana}', "");
Expect(0, 6830, '\P{^	_Is_lana}', "");
Error('\p{ /a/TAI_Viet}');
Error('\P{ /a/TAI_Viet}');
Expect(1, 43743, '\p{taiviet}', "");
Expect(0, 43743, '\p{^taiviet}', "");
Expect(0, 43743, '\P{taiviet}', "");
Expect(1, 43743, '\P{^taiviet}', "");
Expect(0, 43744, '\p{taiviet}', "");
Expect(1, 43744, '\p{^taiviet}', "");
Expect(1, 43744, '\P{taiviet}', "");
Expect(0, 43744, '\P{^taiviet}', "");
Expect(1, 43743, '\p{--tai_Viet}', "");
Expect(0, 43743, '\p{^--tai_Viet}', "");
Expect(0, 43743, '\P{--tai_Viet}', "");
Expect(1, 43743, '\P{^--tai_Viet}', "");
Expect(0, 43744, '\p{--tai_Viet}', "");
Expect(1, 43744, '\p{^--tai_Viet}', "");
Expect(1, 43744, '\P{--tai_Viet}', "");
Expect(0, 43744, '\P{^--tai_Viet}', "");
Error('\p{-/a/is_tai_viet}');
Error('\P{-/a/is_tai_viet}');
Expect(1, 43743, '\p{istaiviet}', "");
Expect(0, 43743, '\p{^istaiviet}', "");
Expect(0, 43743, '\P{istaiviet}', "");
Expect(1, 43743, '\P{^istaiviet}', "");
Expect(0, 43744, '\p{istaiviet}', "");
Expect(1, 43744, '\p{^istaiviet}', "");
Expect(1, 43744, '\P{istaiviet}', "");
Expect(0, 43744, '\P{^istaiviet}', "");
Expect(1, 43743, '\p{ -is_TAI_Viet}', "");
Expect(0, 43743, '\p{^ -is_TAI_Viet}', "");
Expect(0, 43743, '\P{ -is_TAI_Viet}', "");
Expect(1, 43743, '\P{^ -is_TAI_Viet}', "");
Expect(0, 43744, '\p{ -is_TAI_Viet}', "");
Expect(1, 43744, '\p{^ -is_TAI_Viet}', "");
Expect(1, 43744, '\P{ -is_TAI_Viet}', "");
Expect(0, 43744, '\P{^ -is_TAI_Viet}', "");
Error('\p{ 	Tavt/a/}');
Error('\P{ 	Tavt/a/}');
Expect(1, 43743, '\p{tavt}', "");
Expect(0, 43743, '\p{^tavt}', "");
Expect(0, 43743, '\P{tavt}', "");
Expect(1, 43743, '\P{^tavt}', "");
Expect(0, 43744, '\p{tavt}', "");
Expect(1, 43744, '\p{^tavt}', "");
Expect(1, 43744, '\P{tavt}', "");
Expect(0, 43744, '\P{^tavt}', "");
Expect(1, 43743, '\p{-_Tavt}', "");
Expect(0, 43743, '\p{^-_Tavt}', "");
Expect(0, 43743, '\P{-_Tavt}', "");
Expect(1, 43743, '\P{^-_Tavt}', "");
Expect(0, 43744, '\p{-_Tavt}', "");
Expect(1, 43744, '\p{^-_Tavt}', "");
Expect(1, 43744, '\P{-_Tavt}', "");
Expect(0, 43744, '\P{^-_Tavt}', "");
Error('\p{ IS_Tavt:=}');
Error('\P{ IS_Tavt:=}');
Expect(1, 43743, '\p{istavt}', "");
Expect(0, 43743, '\p{^istavt}', "");
Expect(0, 43743, '\P{istavt}', "");
Expect(1, 43743, '\P{^istavt}', "");
Expect(0, 43744, '\p{istavt}', "");
Expect(1, 43744, '\p{^istavt}', "");
Expect(1, 43744, '\P{istavt}', "");
Expect(0, 43744, '\P{^istavt}', "");
Expect(1, 43743, '\p{	Is_TAVT}', "");
Expect(0, 43743, '\p{^	Is_TAVT}', "");
Expect(0, 43743, '\P{	Is_TAVT}', "");
Expect(1, 43743, '\P{^	Is_TAVT}', "");
Expect(0, 43744, '\p{	Is_TAVT}', "");
Expect(1, 43744, '\p{^	Is_TAVT}', "");
Expect(1, 43744, '\P{	Is_TAVT}', "");
Expect(0, 43744, '\P{^	Is_TAVT}', "");
Error('\p{_-Tai_xuan_JING_Symbols:=}');
Error('\P{_-Tai_xuan_JING_Symbols:=}');
Expect(1, 119647, '\p{taixuanjingsymbols}', "");
Expect(0, 119647, '\p{^taixuanjingsymbols}', "");
Expect(0, 119647, '\P{taixuanjingsymbols}', "");
Expect(1, 119647, '\P{^taixuanjingsymbols}', "");
Expect(0, 119648, '\p{taixuanjingsymbols}', "");
Expect(1, 119648, '\p{^taixuanjingsymbols}', "");
Expect(1, 119648, '\P{taixuanjingsymbols}', "");
Expect(0, 119648, '\P{^taixuanjingsymbols}', "");
Expect(1, 119647, '\p{_ TAI_XUAN_JING_Symbols}', "");
Expect(0, 119647, '\p{^_ TAI_XUAN_JING_Symbols}', "");
Expect(0, 119647, '\P{_ TAI_XUAN_JING_Symbols}', "");
Expect(1, 119647, '\P{^_ TAI_XUAN_JING_Symbols}', "");
Expect(0, 119648, '\p{_ TAI_XUAN_JING_Symbols}', "");
Expect(1, 119648, '\p{^_ TAI_XUAN_JING_Symbols}', "");
Expect(1, 119648, '\P{_ TAI_XUAN_JING_Symbols}', "");
Expect(0, 119648, '\P{^_ TAI_XUAN_JING_Symbols}', "");
Error('\p{:=  Is_Tai_Xuan_Jing_symbols}');
Error('\P{:=  Is_Tai_Xuan_Jing_symbols}');
Expect(1, 119647, '\p{istaixuanjingsymbols}', "");
Expect(0, 119647, '\p{^istaixuanjingsymbols}', "");
Expect(0, 119647, '\P{istaixuanjingsymbols}', "");
Expect(1, 119647, '\P{^istaixuanjingsymbols}', "");
Expect(0, 119648, '\p{istaixuanjingsymbols}', "");
Expect(1, 119648, '\p{^istaixuanjingsymbols}', "");
Expect(1, 119648, '\P{istaixuanjingsymbols}', "");
Expect(0, 119648, '\P{^istaixuanjingsymbols}', "");
Expect(1, 119647, '\p{ Is_TAI_Xuan_Jing_Symbols}', "");
Expect(0, 119647, '\p{^ Is_TAI_Xuan_Jing_Symbols}', "");
Expect(0, 119647, '\P{ Is_TAI_Xuan_Jing_Symbols}', "");
Expect(1, 119647, '\P{^ Is_TAI_Xuan_Jing_Symbols}', "");
Expect(0, 119648, '\p{ Is_TAI_Xuan_Jing_Symbols}', "");
Expect(1, 119648, '\p{^ Is_TAI_Xuan_Jing_Symbols}', "");
Expect(1, 119648, '\P{ Is_TAI_Xuan_Jing_Symbols}', "");
Expect(0, 119648, '\P{^ Is_TAI_Xuan_Jing_Symbols}', "");
Error('\p{:=  IN_Tai_Xuan_Jing_Symbols}');
Error('\P{:=  IN_Tai_Xuan_Jing_Symbols}');
Expect(1, 119647, '\p{intaixuanjingsymbols}', "");
Expect(0, 119647, '\p{^intaixuanjingsymbols}', "");
Expect(0, 119647, '\P{intaixuanjingsymbols}', "");
Expect(1, 119647, '\P{^intaixuanjingsymbols}', "");
Expect(0, 119648, '\p{intaixuanjingsymbols}', "");
Expect(1, 119648, '\p{^intaixuanjingsymbols}', "");
Expect(1, 119648, '\P{intaixuanjingsymbols}', "");
Expect(0, 119648, '\P{^intaixuanjingsymbols}', "");
Expect(1, 119647, '\p{ 	In_TAI_XUAN_Jing_Symbols}', "");
Expect(0, 119647, '\p{^ 	In_TAI_XUAN_Jing_Symbols}', "");
Expect(0, 119647, '\P{ 	In_TAI_XUAN_Jing_Symbols}', "");
Expect(1, 119647, '\P{^ 	In_TAI_XUAN_Jing_Symbols}', "");
Expect(0, 119648, '\p{ 	In_TAI_XUAN_Jing_Symbols}', "");
Expect(1, 119648, '\p{^ 	In_TAI_XUAN_Jing_Symbols}', "");
Expect(1, 119648, '\P{ 	In_TAI_XUAN_Jing_Symbols}', "");
Expect(0, 119648, '\P{^ 	In_TAI_XUAN_Jing_Symbols}', "");
Error('\p{		TAI_Xuan_jing/a/}');
Error('\P{		TAI_Xuan_jing/a/}');
Expect(1, 119647, '\p{taixuanjing}', "");
Expect(0, 119647, '\p{^taixuanjing}', "");
Expect(0, 119647, '\P{taixuanjing}', "");
Expect(1, 119647, '\P{^taixuanjing}', "");
Expect(0, 119648, '\p{taixuanjing}', "");
Expect(1, 119648, '\p{^taixuanjing}', "");
Expect(1, 119648, '\P{taixuanjing}', "");
Expect(0, 119648, '\P{^taixuanjing}', "");
Expect(1, 119647, '\p{	-Tai_XUAN_jing}', "");
Expect(0, 119647, '\p{^	-Tai_XUAN_jing}', "");
Expect(0, 119647, '\P{	-Tai_XUAN_jing}', "");
Expect(1, 119647, '\P{^	-Tai_XUAN_jing}', "");
Expect(0, 119648, '\p{	-Tai_XUAN_jing}', "");
Expect(1, 119648, '\p{^	-Tai_XUAN_jing}', "");
Expect(1, 119648, '\P{	-Tai_XUAN_jing}', "");
Expect(0, 119648, '\P{^	-Tai_XUAN_jing}', "");
Error('\p{:=-	is_Tai_XUAN_Jing}');
Error('\P{:=-	is_Tai_XUAN_Jing}');
Expect(1, 119647, '\p{istaixuanjing}', "");
Expect(0, 119647, '\p{^istaixuanjing}', "");
Expect(0, 119647, '\P{istaixuanjing}', "");
Expect(1, 119647, '\P{^istaixuanjing}', "");
Expect(0, 119648, '\p{istaixuanjing}', "");
Expect(1, 119648, '\p{^istaixuanjing}', "");
Expect(1, 119648, '\P{istaixuanjing}', "");
Expect(0, 119648, '\P{^istaixuanjing}', "");
Expect(1, 119647, '\p{-_is_Tai_Xuan_Jing}', "");
Expect(0, 119647, '\p{^-_is_Tai_Xuan_Jing}', "");
Expect(0, 119647, '\P{-_is_Tai_Xuan_Jing}', "");
Expect(1, 119647, '\P{^-_is_Tai_Xuan_Jing}', "");
Expect(0, 119648, '\p{-_is_Tai_Xuan_Jing}', "");
Expect(1, 119648, '\p{^-_is_Tai_Xuan_Jing}', "");
Expect(1, 119648, '\P{-_is_Tai_Xuan_Jing}', "");
Expect(0, 119648, '\P{^-_is_Tai_Xuan_Jing}', "");
Error('\p{ /a/IN_Tai_XUAN_Jing}');
Error('\P{ /a/IN_Tai_XUAN_Jing}');
Expect(1, 119647, '\p{intaixuanjing}', "");
Expect(0, 119647, '\p{^intaixuanjing}', "");
Expect(0, 119647, '\P{intaixuanjing}', "");
Expect(1, 119647, '\P{^intaixuanjing}', "");
Expect(0, 119648, '\p{intaixuanjing}', "");
Expect(1, 119648, '\p{^intaixuanjing}', "");
Expect(1, 119648, '\P{intaixuanjing}', "");
Expect(0, 119648, '\P{^intaixuanjing}', "");
Expect(1, 119647, '\p{--In_tai_Xuan_JING}', "");
Expect(0, 119647, '\p{^--In_tai_Xuan_JING}', "");
Expect(0, 119647, '\P{--In_tai_Xuan_JING}', "");
Expect(1, 119647, '\P{^--In_tai_Xuan_JING}', "");
Expect(0, 119648, '\p{--In_tai_Xuan_JING}', "");
Expect(1, 119648, '\p{^--In_tai_Xuan_JING}', "");
Expect(1, 119648, '\P{--In_tai_Xuan_JING}', "");
Expect(0, 119648, '\P{^--In_tai_Xuan_JING}', "");
Error('\p{ _Takri:=}');
Error('\P{ _Takri:=}');
Expect(1, 71369, '\p{takri}', "");
Expect(0, 71369, '\p{^takri}', "");
Expect(0, 71369, '\P{takri}', "");
Expect(1, 71369, '\P{^takri}', "");
Expect(0, 71370, '\p{takri}', "");
Expect(1, 71370, '\p{^takri}', "");
Expect(1, 71370, '\P{takri}', "");
Expect(0, 71370, '\P{^takri}', "");
Expect(1, 71369, '\p{_Takri}', "");
Expect(0, 71369, '\p{^_Takri}', "");
Expect(0, 71369, '\P{_Takri}', "");
Expect(1, 71369, '\P{^_Takri}', "");
Expect(0, 71370, '\p{_Takri}', "");
Expect(1, 71370, '\p{^_Takri}', "");
Expect(1, 71370, '\P{_Takri}', "");
Expect(0, 71370, '\P{^_Takri}', "");
Error('\p{:=	-is_Takri}');
Error('\P{:=	-is_Takri}');
Expect(1, 71369, '\p{istakri}', "");
Expect(0, 71369, '\p{^istakri}', "");
Expect(0, 71369, '\P{istakri}', "");
Expect(1, 71369, '\P{^istakri}', "");
Expect(0, 71370, '\p{istakri}', "");
Expect(1, 71370, '\p{^istakri}', "");
Expect(1, 71370, '\P{istakri}', "");
Expect(0, 71370, '\P{^istakri}', "");
Expect(1, 71369, '\p{__IS_Takri}', "");
Expect(0, 71369, '\p{^__IS_Takri}', "");
Expect(0, 71369, '\P{__IS_Takri}', "");
Expect(1, 71369, '\P{^__IS_Takri}', "");
Expect(0, 71370, '\p{__IS_Takri}', "");
Expect(1, 71370, '\p{^__IS_Takri}', "");
Expect(1, 71370, '\P{__IS_Takri}', "");
Expect(0, 71370, '\P{^__IS_Takri}', "");
Error('\p{ :=TAKR}');
Error('\P{ :=TAKR}');
Expect(1, 71369, '\p{takr}', "");
Expect(0, 71369, '\p{^takr}', "");
Expect(0, 71369, '\P{takr}', "");
Expect(1, 71369, '\P{^takr}', "");
Expect(0, 71370, '\p{takr}', "");
Expect(1, 71370, '\p{^takr}', "");
Expect(1, 71370, '\P{takr}', "");
Expect(0, 71370, '\P{^takr}', "");
Expect(1, 71369, '\p{_Takr}', "");
Expect(0, 71369, '\p{^_Takr}', "");
Expect(0, 71369, '\P{_Takr}', "");
Expect(1, 71369, '\P{^_Takr}', "");
Expect(0, 71370, '\p{_Takr}', "");
Expect(1, 71370, '\p{^_Takr}', "");
Expect(1, 71370, '\P{_Takr}', "");
Expect(0, 71370, '\P{^_Takr}', "");
Error('\p{:=__is_Takr}');
Error('\P{:=__is_Takr}');
Expect(1, 71369, '\p{istakr}', "");
Expect(0, 71369, '\p{^istakr}', "");
Expect(0, 71369, '\P{istakr}', "");
Expect(1, 71369, '\P{^istakr}', "");
Expect(0, 71370, '\p{istakr}', "");
Expect(1, 71370, '\p{^istakr}', "");
Expect(1, 71370, '\P{istakr}', "");
Expect(0, 71370, '\P{^istakr}', "");
Expect(1, 71369, '\p{-Is_TAKR}', "");
Expect(0, 71369, '\p{^-Is_TAKR}', "");
Expect(0, 71369, '\P{-Is_TAKR}', "");
Expect(1, 71369, '\P{^-Is_TAKR}', "");
Expect(0, 71370, '\p{-Is_TAKR}', "");
Expect(1, 71370, '\p{^-Is_TAKR}', "");
Expect(1, 71370, '\P{-Is_TAKR}', "");
Expect(0, 71370, '\P{^-Is_TAKR}', "");
Error('\p{:= _Tamil}');
Error('\P{:= _Tamil}');
Expect(1, 3066, '\p{tamil}', "");
Expect(0, 3066, '\p{^tamil}', "");
Expect(0, 3066, '\P{tamil}', "");
Expect(1, 3066, '\P{^tamil}', "");
Expect(0, 3067, '\p{tamil}', "");
Expect(1, 3067, '\p{^tamil}', "");
Expect(1, 3067, '\P{tamil}', "");
Expect(0, 3067, '\P{^tamil}', "");
Expect(1, 3066, '\p{ tamil}', "");
Expect(0, 3066, '\p{^ tamil}', "");
Expect(0, 3066, '\P{ tamil}', "");
Expect(1, 3066, '\P{^ tamil}', "");
Expect(0, 3067, '\p{ tamil}', "");
Expect(1, 3067, '\p{^ tamil}', "");
Expect(1, 3067, '\P{ tamil}', "");
Expect(0, 3067, '\P{^ tamil}', "");
Error('\p{/a/ 	IS_TAMIL}');
Error('\P{/a/ 	IS_TAMIL}');
Expect(1, 3066, '\p{istamil}', "");
Expect(0, 3066, '\p{^istamil}', "");
Expect(0, 3066, '\P{istamil}', "");
Expect(1, 3066, '\P{^istamil}', "");
Expect(0, 3067, '\p{istamil}', "");
Expect(1, 3067, '\p{^istamil}', "");
Expect(1, 3067, '\P{istamil}', "");
Expect(0, 3067, '\P{^istamil}', "");
Expect(1, 3066, '\p{ _Is_tamil}', "");
Expect(0, 3066, '\p{^ _Is_tamil}', "");
Expect(0, 3066, '\P{ _Is_tamil}', "");
Expect(1, 3066, '\P{^ _Is_tamil}', "");
Expect(0, 3067, '\p{ _Is_tamil}', "");
Expect(1, 3067, '\p{^ _Is_tamil}', "");
Expect(1, 3067, '\P{ _Is_tamil}', "");
Expect(0, 3067, '\P{^ _Is_tamil}', "");
Error('\p{-/a/Taml}');
Error('\P{-/a/Taml}');
Expect(1, 3066, '\p{taml}', "");
Expect(0, 3066, '\p{^taml}', "");
Expect(0, 3066, '\P{taml}', "");
Expect(1, 3066, '\P{^taml}', "");
Expect(0, 3067, '\p{taml}', "");
Expect(1, 3067, '\p{^taml}', "");
Expect(1, 3067, '\P{taml}', "");
Expect(0, 3067, '\P{^taml}', "");
Expect(1, 3066, '\p{__Taml}', "");
Expect(0, 3066, '\p{^__Taml}', "");
Expect(0, 3066, '\P{__Taml}', "");
Expect(1, 3066, '\P{^__Taml}', "");
Expect(0, 3067, '\p{__Taml}', "");
Expect(1, 3067, '\p{^__Taml}', "");
Expect(1, 3067, '\P{__Taml}', "");
Expect(0, 3067, '\P{^__Taml}', "");
Error('\p{	_IS_taml/a/}');
Error('\P{	_IS_taml/a/}');
Expect(1, 3066, '\p{istaml}', "");
Expect(0, 3066, '\p{^istaml}', "");
Expect(0, 3066, '\P{istaml}', "");
Expect(1, 3066, '\P{^istaml}', "");
Expect(0, 3067, '\p{istaml}', "");
Expect(1, 3067, '\p{^istaml}', "");
Expect(1, 3067, '\P{istaml}', "");
Expect(0, 3067, '\P{^istaml}', "");
Expect(1, 3066, '\p{- Is_taml}', "");
Expect(0, 3066, '\p{^- Is_taml}', "");
Expect(0, 3066, '\P{- Is_taml}', "");
Expect(1, 3066, '\P{^- Is_taml}', "");
Expect(0, 3067, '\p{- Is_taml}', "");
Expect(1, 3067, '\p{^- Is_taml}', "");
Expect(1, 3067, '\P{- Is_taml}', "");
Expect(0, 3067, '\P{^- Is_taml}', "");
Error('\p{:=_Telugu}');
Error('\P{:=_Telugu}');
Expect(1, 3199, '\p{telugu}', "");
Expect(0, 3199, '\p{^telugu}', "");
Expect(0, 3199, '\P{telugu}', "");
Expect(1, 3199, '\P{^telugu}', "");
Expect(0, 3200, '\p{telugu}', "");
Expect(1, 3200, '\p{^telugu}', "");
Expect(1, 3200, '\P{telugu}', "");
Expect(0, 3200, '\P{^telugu}', "");
Expect(1, 3199, '\p{Telugu}', "");
Expect(0, 3199, '\p{^Telugu}', "");
Expect(0, 3199, '\P{Telugu}', "");
Expect(1, 3199, '\P{^Telugu}', "");
Expect(0, 3200, '\p{Telugu}', "");
Expect(1, 3200, '\p{^Telugu}', "");
Expect(1, 3200, '\P{Telugu}', "");
Expect(0, 3200, '\P{^Telugu}', "");
Error('\p{-/a/Is_Telugu}');
Error('\P{-/a/Is_Telugu}');
Expect(1, 3199, '\p{istelugu}', "");
Expect(0, 3199, '\p{^istelugu}', "");
Expect(0, 3199, '\P{istelugu}', "");
Expect(1, 3199, '\P{^istelugu}', "");
Expect(0, 3200, '\p{istelugu}', "");
Expect(1, 3200, '\p{^istelugu}', "");
Expect(1, 3200, '\P{istelugu}', "");
Expect(0, 3200, '\P{^istelugu}', "");
Expect(1, 3199, '\p{	-is_TELUGU}', "");
Expect(0, 3199, '\p{^	-is_TELUGU}', "");
Expect(0, 3199, '\P{	-is_TELUGU}', "");
Expect(1, 3199, '\P{^	-is_TELUGU}', "");
Expect(0, 3200, '\p{	-is_TELUGU}', "");
Expect(1, 3200, '\p{^	-is_TELUGU}', "");
Expect(1, 3200, '\P{	-is_TELUGU}', "");
Expect(0, 3200, '\P{^	-is_TELUGU}', "");
Error('\p{_/a/TELU}');
Error('\P{_/a/TELU}');
Expect(1, 3199, '\p{telu}', "");
Expect(0, 3199, '\p{^telu}', "");
Expect(0, 3199, '\P{telu}', "");
Expect(1, 3199, '\P{^telu}', "");
Expect(0, 3200, '\p{telu}', "");
Expect(1, 3200, '\p{^telu}', "");
Expect(1, 3200, '\P{telu}', "");
Expect(0, 3200, '\P{^telu}', "");
Expect(1, 3199, '\p{-telu}', "");
Expect(0, 3199, '\p{^-telu}', "");
Expect(0, 3199, '\P{-telu}', "");
Expect(1, 3199, '\P{^-telu}', "");
Expect(0, 3200, '\p{-telu}', "");
Expect(1, 3200, '\p{^-telu}', "");
Expect(1, 3200, '\P{-telu}', "");
Expect(0, 3200, '\P{^-telu}', "");
Error('\p{:=_is_TELU}');
Error('\P{:=_is_TELU}');
Expect(1, 3199, '\p{istelu}', "");
Expect(0, 3199, '\p{^istelu}', "");
Expect(0, 3199, '\P{istelu}', "");
Expect(1, 3199, '\P{^istelu}', "");
Expect(0, 3200, '\p{istelu}', "");
Expect(1, 3200, '\p{^istelu}', "");
Expect(1, 3200, '\P{istelu}', "");
Expect(0, 3200, '\P{^istelu}', "");
Expect(1, 3199, '\p{	 Is_Telu}', "");
Expect(0, 3199, '\p{^	 Is_Telu}', "");
Expect(0, 3199, '\P{	 Is_Telu}', "");
Expect(1, 3199, '\P{^	 Is_Telu}', "");
Expect(0, 3200, '\p{	 Is_Telu}', "");
Expect(1, 3200, '\p{^	 Is_Telu}', "");
Expect(1, 3200, '\P{	 Is_Telu}', "");
Expect(0, 3200, '\P{^	 Is_Telu}', "");
Error('\p{_:=terminal_punctuation}');
Error('\P{_:=terminal_punctuation}');
Expect(1, 74867, '\p{terminalpunctuation}', "");
Expect(0, 74867, '\p{^terminalpunctuation}', "");
Expect(0, 74867, '\P{terminalpunctuation}', "");
Expect(1, 74867, '\P{^terminalpunctuation}', "");
Expect(0, 74868, '\p{terminalpunctuation}', "");
Expect(1, 74868, '\p{^terminalpunctuation}', "");
Expect(1, 74868, '\P{terminalpunctuation}', "");
Expect(0, 74868, '\P{^terminalpunctuation}', "");
Expect(1, 74867, '\p{ terminal_punctuation}', "");
Expect(0, 74867, '\p{^ terminal_punctuation}', "");
Expect(0, 74867, '\P{ terminal_punctuation}', "");
Expect(1, 74867, '\P{^ terminal_punctuation}', "");
Expect(0, 74868, '\p{ terminal_punctuation}', "");
Expect(1, 74868, '\p{^ terminal_punctuation}', "");
Expect(1, 74868, '\P{ terminal_punctuation}', "");
Expect(0, 74868, '\P{^ terminal_punctuation}', "");
Error('\p{	-Is_Terminal_Punctuation/a/}');
Error('\P{	-Is_Terminal_Punctuation/a/}');
Expect(1, 74867, '\p{isterminalpunctuation}', "");
Expect(0, 74867, '\p{^isterminalpunctuation}', "");
Expect(0, 74867, '\P{isterminalpunctuation}', "");
Expect(1, 74867, '\P{^isterminalpunctuation}', "");
Expect(0, 74868, '\p{isterminalpunctuation}', "");
Expect(1, 74868, '\p{^isterminalpunctuation}', "");
Expect(1, 74868, '\P{isterminalpunctuation}', "");
Expect(0, 74868, '\P{^isterminalpunctuation}', "");
Expect(1, 74867, '\p{_	Is_TERMINAL_punctuation}', "");
Expect(0, 74867, '\p{^_	Is_TERMINAL_punctuation}', "");
Expect(0, 74867, '\P{_	Is_TERMINAL_punctuation}', "");
Expect(1, 74867, '\P{^_	Is_TERMINAL_punctuation}', "");
Expect(0, 74868, '\p{_	Is_TERMINAL_punctuation}', "");
Expect(1, 74868, '\p{^_	Is_TERMINAL_punctuation}', "");
Expect(1, 74868, '\P{_	Is_TERMINAL_punctuation}', "");
Expect(0, 74868, '\P{^_	Is_TERMINAL_punctuation}', "");
Error('\p{- term/a/}');
Error('\P{- term/a/}');
Expect(1, 74867, '\p{term}', "");
Expect(0, 74867, '\p{^term}', "");
Expect(0, 74867, '\P{term}', "");
Expect(1, 74867, '\P{^term}', "");
Expect(0, 74868, '\p{term}', "");
Expect(1, 74868, '\p{^term}', "");
Expect(1, 74868, '\P{term}', "");
Expect(0, 74868, '\P{^term}', "");
Expect(1, 74867, '\p{_	TERM}', "");
Expect(0, 74867, '\p{^_	TERM}', "");
Expect(0, 74867, '\P{_	TERM}', "");
Expect(1, 74867, '\P{^_	TERM}', "");
Expect(0, 74868, '\p{_	TERM}', "");
Expect(1, 74868, '\p{^_	TERM}', "");
Expect(1, 74868, '\P{_	TERM}', "");
Expect(0, 74868, '\P{^_	TERM}', "");
Error('\p{	:=Is_Term}');
Error('\P{	:=Is_Term}');
Expect(1, 74867, '\p{isterm}', "");
Expect(0, 74867, '\p{^isterm}', "");
Expect(0, 74867, '\P{isterm}', "");
Expect(1, 74867, '\P{^isterm}', "");
Expect(0, 74868, '\p{isterm}', "");
Expect(1, 74868, '\p{^isterm}', "");
Expect(1, 74868, '\P{isterm}', "");
Expect(0, 74868, '\P{^isterm}', "");
Expect(1, 74867, '\p{ 	is_Term}', "");
Expect(0, 74867, '\p{^ 	is_Term}', "");
Expect(0, 74867, '\P{ 	is_Term}', "");
Expect(1, 74867, '\P{^ 	is_Term}', "");
Expect(0, 74868, '\p{ 	is_Term}', "");
Expect(1, 74868, '\p{^ 	is_Term}', "");
Expect(1, 74868, '\P{ 	is_Term}', "");
Expect(0, 74868, '\P{^ 	is_Term}', "");
Error('\p{/a/--thaana}');
Error('\P{/a/--thaana}');
Expect(1, 1969, '\p{thaana}', "");
Expect(0, 1969, '\p{^thaana}', "");
Expect(0, 1969, '\P{thaana}', "");
Expect(1, 1969, '\P{^thaana}', "");
Expect(0, 1970, '\p{thaana}', "");
Expect(1, 1970, '\p{^thaana}', "");
Expect(1, 1970, '\P{thaana}', "");
Expect(0, 1970, '\P{^thaana}', "");
Expect(1, 1969, '\p{-	thaana}', "");
Expect(0, 1969, '\p{^-	thaana}', "");
Expect(0, 1969, '\P{-	thaana}', "");
Expect(1, 1969, '\P{^-	thaana}', "");
Expect(0, 1970, '\p{-	thaana}', "");
Expect(1, 1970, '\p{^-	thaana}', "");
Expect(1, 1970, '\P{-	thaana}', "");
Expect(0, 1970, '\P{^-	thaana}', "");
Error('\p{:=IS_thaana}');
Error('\P{:=IS_thaana}');
Expect(1, 1969, '\p{isthaana}', "");
Expect(0, 1969, '\p{^isthaana}', "");
Expect(0, 1969, '\P{isthaana}', "");
Expect(1, 1969, '\P{^isthaana}', "");
Expect(0, 1970, '\p{isthaana}', "");
Expect(1, 1970, '\p{^isthaana}', "");
Expect(1, 1970, '\P{isthaana}', "");
Expect(0, 1970, '\P{^isthaana}', "");
Expect(1, 1969, '\p{_Is_THAANA}', "");
Expect(0, 1969, '\p{^_Is_THAANA}', "");
Expect(0, 1969, '\P{_Is_THAANA}', "");
Expect(1, 1969, '\P{^_Is_THAANA}', "");
Expect(0, 1970, '\p{_Is_THAANA}', "");
Expect(1, 1970, '\p{^_Is_THAANA}', "");
Expect(1, 1970, '\P{_Is_THAANA}', "");
Expect(0, 1970, '\P{^_Is_THAANA}', "");
Error('\p{	-Thaa:=}');
Error('\P{	-Thaa:=}');
Expect(1, 1969, '\p{thaa}', "");
Expect(0, 1969, '\p{^thaa}', "");
Expect(0, 1969, '\P{thaa}', "");
Expect(1, 1969, '\P{^thaa}', "");
Expect(0, 1970, '\p{thaa}', "");
Expect(1, 1970, '\p{^thaa}', "");
Expect(1, 1970, '\P{thaa}', "");
Expect(0, 1970, '\P{^thaa}', "");
Expect(1, 1969, '\p{ thaa}', "");
Expect(0, 1969, '\p{^ thaa}', "");
Expect(0, 1969, '\P{ thaa}', "");
Expect(1, 1969, '\P{^ thaa}', "");
Expect(0, 1970, '\p{ thaa}', "");
Expect(1, 1970, '\p{^ thaa}', "");
Expect(1, 1970, '\P{ thaa}', "");
Expect(0, 1970, '\P{^ thaa}', "");
Error('\p{-:=Is_Thaa}');
Error('\P{-:=Is_Thaa}');
Expect(1, 1969, '\p{isthaa}', "");
Expect(0, 1969, '\p{^isthaa}', "");
Expect(0, 1969, '\P{isthaa}', "");
Expect(1, 1969, '\P{^isthaa}', "");
Expect(0, 1970, '\p{isthaa}', "");
Expect(1, 1970, '\p{^isthaa}', "");
Expect(1, 1970, '\P{isthaa}', "");
Expect(0, 1970, '\P{^isthaa}', "");
Expect(1, 1969, '\p{--IS_Thaa}', "");
Expect(0, 1969, '\p{^--IS_Thaa}', "");
Expect(0, 1969, '\P{--IS_Thaa}', "");
Expect(1, 1969, '\P{^--IS_Thaa}', "");
Expect(0, 1970, '\p{--IS_Thaa}', "");
Expect(1, 1970, '\p{^--IS_Thaa}', "");
Expect(1, 1970, '\P{--IS_Thaa}', "");
Expect(0, 1970, '\P{^--IS_Thaa}', "");
Error('\p{/a/- Thai}');
Error('\P{/a/- Thai}');
Expect(1, 3675, '\p{thai}', "");
Expect(0, 3675, '\p{^thai}', "");
Expect(0, 3675, '\P{thai}', "");
Expect(1, 3675, '\P{^thai}', "");
Expect(0, 3676, '\p{thai}', "");
Expect(1, 3676, '\p{^thai}', "");
Expect(1, 3676, '\P{thai}', "");
Expect(0, 3676, '\P{^thai}', "");
Expect(1, 3675, '\p{ -Thai}', "");
Expect(0, 3675, '\p{^ -Thai}', "");
Expect(0, 3675, '\P{ -Thai}', "");
Expect(1, 3675, '\P{^ -Thai}', "");
Expect(0, 3676, '\p{ -Thai}', "");
Expect(1, 3676, '\p{^ -Thai}', "");
Expect(1, 3676, '\P{ -Thai}', "");
Expect(0, 3676, '\P{^ -Thai}', "");
Error('\p{-:=Is_thai}');
Error('\P{-:=Is_thai}');
Expect(1, 3675, '\p{isthai}', "");
Expect(0, 3675, '\p{^isthai}', "");
Expect(0, 3675, '\P{isthai}', "");
Expect(1, 3675, '\P{^isthai}', "");
Expect(0, 3676, '\p{isthai}', "");
Expect(1, 3676, '\p{^isthai}', "");
Expect(1, 3676, '\P{isthai}', "");
Expect(0, 3676, '\P{^isthai}', "");
Expect(1, 3675, '\p{ -Is_Thai}', "");
Expect(0, 3675, '\p{^ -Is_Thai}', "");
Expect(0, 3675, '\P{ -Is_Thai}', "");
Expect(1, 3675, '\P{^ -Is_Thai}', "");
Expect(0, 3676, '\p{ -Is_Thai}', "");
Expect(1, 3676, '\p{^ -Is_Thai}', "");
Expect(1, 3676, '\P{ -Is_Thai}', "");
Expect(0, 3676, '\P{^ -Is_Thai}', "");
Error('\p{-_tibetan/a/}');
Error('\P{-_tibetan/a/}');
Expect(1, 4058, '\p{tibetan}', "");
Expect(0, 4058, '\p{^tibetan}', "");
Expect(0, 4058, '\P{tibetan}', "");
Expect(1, 4058, '\P{^tibetan}', "");
Expect(0, 4059, '\p{tibetan}', "");
Expect(1, 4059, '\p{^tibetan}', "");
Expect(1, 4059, '\P{tibetan}', "");
Expect(0, 4059, '\P{^tibetan}', "");
Expect(1, 4058, '\p{ -tibetan}', "");
Expect(0, 4058, '\p{^ -tibetan}', "");
Expect(0, 4058, '\P{ -tibetan}', "");
Expect(1, 4058, '\P{^ -tibetan}', "");
Expect(0, 4059, '\p{ -tibetan}', "");
Expect(1, 4059, '\p{^ -tibetan}', "");
Expect(1, 4059, '\P{ -tibetan}', "");
Expect(0, 4059, '\P{^ -tibetan}', "");
Error('\p{_:=Is_tibetan}');
Error('\P{_:=Is_tibetan}');
Expect(1, 4058, '\p{istibetan}', "");
Expect(0, 4058, '\p{^istibetan}', "");
Expect(0, 4058, '\P{istibetan}', "");
Expect(1, 4058, '\P{^istibetan}', "");
Expect(0, 4059, '\p{istibetan}', "");
Expect(1, 4059, '\p{^istibetan}', "");
Expect(1, 4059, '\P{istibetan}', "");
Expect(0, 4059, '\P{^istibetan}', "");
Expect(1, 4058, '\p{_ is_tibetan}', "");
Expect(0, 4058, '\p{^_ is_tibetan}', "");
Expect(0, 4058, '\P{_ is_tibetan}', "");
Expect(1, 4058, '\P{^_ is_tibetan}', "");
Expect(0, 4059, '\p{_ is_tibetan}', "");
Expect(1, 4059, '\p{^_ is_tibetan}', "");
Expect(1, 4059, '\P{_ is_tibetan}', "");
Expect(0, 4059, '\P{^_ is_tibetan}', "");
Error('\p{:=	 tibt}');
Error('\P{:=	 tibt}');
Expect(1, 4058, '\p{tibt}', "");
Expect(0, 4058, '\p{^tibt}', "");
Expect(0, 4058, '\P{tibt}', "");
Expect(1, 4058, '\P{^tibt}', "");
Expect(0, 4059, '\p{tibt}', "");
Expect(1, 4059, '\p{^tibt}', "");
Expect(1, 4059, '\P{tibt}', "");
Expect(0, 4059, '\P{^tibt}', "");
Expect(1, 4058, '\p{_-TIBT}', "");
Expect(0, 4058, '\p{^_-TIBT}', "");
Expect(0, 4058, '\P{_-TIBT}', "");
Expect(1, 4058, '\P{^_-TIBT}', "");
Expect(0, 4059, '\p{_-TIBT}', "");
Expect(1, 4059, '\p{^_-TIBT}', "");
Expect(1, 4059, '\P{_-TIBT}', "");
Expect(0, 4059, '\P{^_-TIBT}', "");
Error('\p{:=__Is_tibt}');
Error('\P{:=__Is_tibt}');
Expect(1, 4058, '\p{istibt}', "");
Expect(0, 4058, '\p{^istibt}', "");
Expect(0, 4058, '\P{istibt}', "");
Expect(1, 4058, '\P{^istibt}', "");
Expect(0, 4059, '\p{istibt}', "");
Expect(1, 4059, '\p{^istibt}', "");
Expect(1, 4059, '\P{istibt}', "");
Expect(0, 4059, '\P{^istibt}', "");
Expect(1, 4058, '\p{ is_TIBT}', "");
Expect(0, 4058, '\p{^ is_TIBT}', "");
Expect(0, 4058, '\P{ is_TIBT}', "");
Expect(1, 4058, '\P{^ is_TIBT}', "");
Expect(0, 4059, '\p{ is_TIBT}', "");
Expect(1, 4059, '\p{^ is_TIBT}', "");
Expect(1, 4059, '\P{ is_TIBT}', "");
Expect(0, 4059, '\P{^ is_TIBT}', "");
Error('\p{/a/_ Tifinagh}');
Error('\P{/a/_ Tifinagh}');
Expect(1, 11647, '\p{tifinagh}', "");
Expect(0, 11647, '\p{^tifinagh}', "");
Expect(0, 11647, '\P{tifinagh}', "");
Expect(1, 11647, '\P{^tifinagh}', "");
Expect(0, 11648, '\p{tifinagh}', "");
Expect(1, 11648, '\p{^tifinagh}', "");
Expect(1, 11648, '\P{tifinagh}', "");
Expect(0, 11648, '\P{^tifinagh}', "");
Expect(1, 11647, '\p{-tifinagh}', "");
Expect(0, 11647, '\p{^-tifinagh}', "");
Expect(0, 11647, '\P{-tifinagh}', "");
Expect(1, 11647, '\P{^-tifinagh}', "");
Expect(0, 11648, '\p{-tifinagh}', "");
Expect(1, 11648, '\p{^-tifinagh}', "");
Expect(1, 11648, '\P{-tifinagh}', "");
Expect(0, 11648, '\P{^-tifinagh}', "");
Error('\p{:=	-Is_TIFINAGH}');
Error('\P{:=	-Is_TIFINAGH}');
Expect(1, 11647, '\p{istifinagh}', "");
Expect(0, 11647, '\p{^istifinagh}', "");
Expect(0, 11647, '\P{istifinagh}', "");
Expect(1, 11647, '\P{^istifinagh}', "");
Expect(0, 11648, '\p{istifinagh}', "");
Expect(1, 11648, '\p{^istifinagh}', "");
Expect(1, 11648, '\P{istifinagh}', "");
Expect(0, 11648, '\P{^istifinagh}', "");
Expect(1, 11647, '\p{-Is_Tifinagh}', "");
Expect(0, 11647, '\p{^-Is_Tifinagh}', "");
Expect(0, 11647, '\P{-Is_Tifinagh}', "");
Expect(1, 11647, '\P{^-Is_Tifinagh}', "");
Expect(0, 11648, '\p{-Is_Tifinagh}', "");
Expect(1, 11648, '\p{^-Is_Tifinagh}', "");
Expect(1, 11648, '\P{-Is_Tifinagh}', "");
Expect(0, 11648, '\P{^-Is_Tifinagh}', "");
Error('\p{-:=TFNG}');
Error('\P{-:=TFNG}');
Expect(1, 11647, '\p{tfng}', "");
Expect(0, 11647, '\p{^tfng}', "");
Expect(0, 11647, '\P{tfng}', "");
Expect(1, 11647, '\P{^tfng}', "");
Expect(0, 11648, '\p{tfng}', "");
Expect(1, 11648, '\p{^tfng}', "");
Expect(1, 11648, '\P{tfng}', "");
Expect(0, 11648, '\P{^tfng}', "");
Expect(1, 11647, '\p{  tfng}', "");
Expect(0, 11647, '\p{^  tfng}', "");
Expect(0, 11647, '\P{  tfng}', "");
Expect(1, 11647, '\P{^  tfng}', "");
Expect(0, 11648, '\p{  tfng}', "");
Expect(1, 11648, '\p{^  tfng}', "");
Expect(1, 11648, '\P{  tfng}', "");
Expect(0, 11648, '\P{^  tfng}', "");
Error('\p{/a/__Is_Tfng}');
Error('\P{/a/__Is_Tfng}');
Expect(1, 11647, '\p{istfng}', "");
Expect(0, 11647, '\p{^istfng}', "");
Expect(0, 11647, '\P{istfng}', "");
Expect(1, 11647, '\P{^istfng}', "");
Expect(0, 11648, '\p{istfng}', "");
Expect(1, 11648, '\p{^istfng}', "");
Expect(1, 11648, '\P{istfng}', "");
Expect(0, 11648, '\P{^istfng}', "");
Expect(1, 11647, '\p{  Is_TFNG}', "");
Expect(0, 11647, '\p{^  Is_TFNG}', "");
Expect(0, 11647, '\P{  Is_TFNG}', "");
Expect(1, 11647, '\P{^  Is_TFNG}', "");
Expect(0, 11648, '\p{  Is_TFNG}', "");
Expect(1, 11648, '\p{^  Is_TFNG}', "");
Expect(1, 11648, '\P{  Is_TFNG}', "");
Expect(0, 11648, '\P{^  Is_TFNG}', "");
Error('\p{-	Titlecase:=}');
Error('\P{-	Titlecase:=}');
Expect(1, 8188, '\p{titlecase}', "");
Expect(0, 8188, '\p{^titlecase}', "");
Expect(0, 8188, '\P{titlecase}', "");
Expect(1, 8188, '\P{^titlecase}', "");
Expect(0, 8189, '\p{titlecase}', "");
Expect(1, 8189, '\p{^titlecase}', "");
Expect(1, 8189, '\P{titlecase}', "");
Expect(0, 8189, '\P{^titlecase}', "");
Expect(1, 8188, '\p{_ TITLECASE}', "");
Expect(0, 8188, '\p{^_ TITLECASE}', "");
Expect(0, 8188, '\P{_ TITLECASE}', "");
Expect(1, 8188, '\P{^_ TITLECASE}', "");
Expect(0, 8189, '\p{_ TITLECASE}', "");
Expect(1, 8189, '\p{^_ TITLECASE}', "");
Expect(1, 8189, '\P{_ TITLECASE}', "");
Expect(0, 8189, '\P{^_ TITLECASE}', "");
Error('\p{/a/-TITLE}');
Error('\P{/a/-TITLE}');
Expect(1, 8188, '\p{title}', "");
Expect(0, 8188, '\p{^title}', "");
Expect(0, 8188, '\P{title}', "");
Expect(1, 8188, '\P{^title}', "");
Expect(0, 8189, '\p{title}', "");
Expect(1, 8189, '\p{^title}', "");
Expect(1, 8189, '\P{title}', "");
Expect(0, 8189, '\P{^title}', "");
Expect(1, 8188, '\p{	-Title}', "");
Expect(0, 8188, '\p{^	-Title}', "");
Expect(0, 8188, '\P{	-Title}', "");
Expect(1, 8188, '\P{^	-Title}', "");
Expect(0, 8189, '\p{	-Title}', "");
Expect(1, 8189, '\p{^	-Title}', "");
Expect(1, 8189, '\P{	-Title}', "");
Expect(0, 8189, '\P{^	-Title}', "");
Error('\p{_/a/Is_Titlecase}');
Error('\P{_/a/Is_Titlecase}');
Expect(1, 8188, '\p{istitlecase}', "");
Expect(0, 8188, '\p{^istitlecase}', "");
Expect(0, 8188, '\P{istitlecase}', "");
Expect(1, 8188, '\P{^istitlecase}', "");
Expect(0, 8189, '\p{istitlecase}', "");
Expect(1, 8189, '\p{^istitlecase}', "");
Expect(1, 8189, '\P{istitlecase}', "");
Expect(0, 8189, '\P{^istitlecase}', "");
Expect(1, 8188, '\p{ _Is_Titlecase}', "");
Expect(0, 8188, '\p{^ _Is_Titlecase}', "");
Expect(0, 8188, '\P{ _Is_Titlecase}', "");
Expect(1, 8188, '\P{^ _Is_Titlecase}', "");
Expect(0, 8189, '\p{ _Is_Titlecase}', "");
Expect(1, 8189, '\p{^ _Is_Titlecase}', "");
Expect(1, 8189, '\P{ _Is_Titlecase}', "");
Expect(0, 8189, '\P{^ _Is_Titlecase}', "");
Error('\p{-:=Is_Title}');
Error('\P{-:=Is_Title}');
Expect(1, 8188, '\p{istitle}', "");
Expect(0, 8188, '\p{^istitle}', "");
Expect(0, 8188, '\P{istitle}', "");
Expect(1, 8188, '\P{^istitle}', "");
Expect(0, 8189, '\p{istitle}', "");
Expect(1, 8189, '\p{^istitle}', "");
Expect(1, 8189, '\P{istitle}', "");
Expect(0, 8189, '\P{^istitle}', "");
Expect(1, 8188, '\p{	Is_Title}', "");
Expect(0, 8188, '\p{^	Is_Title}', "");
Expect(0, 8188, '\P{	Is_Title}', "");
Expect(1, 8188, '\P{^	Is_Title}', "");
Expect(0, 8189, '\p{	Is_Title}', "");
Expect(1, 8189, '\p{^	Is_Title}', "");
Expect(1, 8189, '\P{	Is_Title}', "");
Expect(0, 8189, '\P{^	Is_Title}', "");
Error('\p{_:=titlecase_LETTER}');
Error('\P{_:=titlecase_LETTER}');
Expect(1, 8188, '\p{titlecaseletter}', "");
Expect(0, 8188, '\p{^titlecaseletter}', "");
Expect(0, 8188, '\P{titlecaseletter}', "");
Expect(1, 8188, '\P{^titlecaseletter}', "");
Expect(0, 8189, '\p{titlecaseletter}', "");
Expect(1, 8189, '\p{^titlecaseletter}', "");
Expect(1, 8189, '\P{titlecaseletter}', "");
Expect(0, 8189, '\P{^titlecaseletter}', "");
Expect(1, 8188, '\p{	Titlecase_Letter}', "");
Expect(0, 8188, '\p{^	Titlecase_Letter}', "");
Expect(0, 8188, '\P{	Titlecase_Letter}', "");
Expect(1, 8188, '\P{^	Titlecase_Letter}', "");
Expect(0, 8189, '\p{	Titlecase_Letter}', "");
Expect(1, 8189, '\p{^	Titlecase_Letter}', "");
Expect(1, 8189, '\P{	Titlecase_Letter}', "");
Expect(0, 8189, '\P{^	Titlecase_Letter}', "");
Error('\p{/a/	_is_Titlecase_letter}');
Error('\P{/a/	_is_Titlecase_letter}');
Expect(1, 8188, '\p{istitlecaseletter}', "");
Expect(0, 8188, '\p{^istitlecaseletter}', "");
Expect(0, 8188, '\P{istitlecaseletter}', "");
Expect(1, 8188, '\P{^istitlecaseletter}', "");
Expect(0, 8189, '\p{istitlecaseletter}', "");
Expect(1, 8189, '\p{^istitlecaseletter}', "");
Expect(1, 8189, '\P{istitlecaseletter}', "");
Expect(0, 8189, '\P{^istitlecaseletter}', "");
Expect(1, 8188, '\p{_is_TITLECASE_Letter}', "");
Expect(0, 8188, '\p{^_is_TITLECASE_Letter}', "");
Expect(0, 8188, '\P{_is_TITLECASE_Letter}', "");
Expect(1, 8188, '\P{^_is_TITLECASE_Letter}', "");
Expect(0, 8189, '\p{_is_TITLECASE_Letter}', "");
Expect(1, 8189, '\p{^_is_TITLECASE_Letter}', "");
Expect(1, 8189, '\P{_is_TITLECASE_Letter}', "");
Expect(0, 8189, '\P{^_is_TITLECASE_Letter}', "");
Error('\p{:=_-Lt}');
Error('\P{:=_-Lt}');
Expect(1, 8188, '\p{lt}', "");
Expect(0, 8188, '\p{^lt}', "");
Expect(0, 8188, '\P{lt}', "");
Expect(1, 8188, '\P{^lt}', "");
Expect(0, 8189, '\p{lt}', "");
Expect(1, 8189, '\p{^lt}', "");
Expect(1, 8189, '\P{lt}', "");
Expect(0, 8189, '\P{^lt}', "");
Expect(1, 8188, '\p{_-Lt}', "");
Expect(0, 8188, '\p{^_-Lt}', "");
Expect(0, 8188, '\P{_-Lt}', "");
Expect(1, 8188, '\P{^_-Lt}', "");
Expect(0, 8189, '\p{_-Lt}', "");
Expect(1, 8189, '\p{^_-Lt}', "");
Expect(1, 8189, '\P{_-Lt}', "");
Expect(0, 8189, '\P{^_-Lt}', "");
Error('\p{ /a/IS_Lt}');
Error('\P{ /a/IS_Lt}');
Expect(1, 8188, '\p{islt}', "");
Expect(0, 8188, '\p{^islt}', "");
Expect(0, 8188, '\P{islt}', "");
Expect(1, 8188, '\P{^islt}', "");
Expect(0, 8189, '\p{islt}', "");
Expect(1, 8189, '\p{^islt}', "");
Expect(1, 8189, '\P{islt}', "");
Expect(0, 8189, '\P{^islt}', "");
Expect(1, 8188, '\p{-	is_Lt}', "");
Expect(0, 8188, '\p{^-	is_Lt}', "");
Expect(0, 8188, '\P{-	is_Lt}', "");
Expect(1, 8188, '\P{^-	is_Lt}', "");
Expect(0, 8189, '\p{-	is_Lt}', "");
Expect(1, 8189, '\p{^-	is_Lt}', "");
Expect(1, 8189, '\P{-	is_Lt}', "");
Expect(0, 8189, '\P{^-	is_Lt}', "");
Error('\p{	Transport_and_map_SYMBOLS/a/}');
Error('\P{	Transport_and_map_SYMBOLS/a/}');
Expect(1, 128767, '\p{transportandmapsymbols}', "");
Expect(0, 128767, '\p{^transportandmapsymbols}', "");
Expect(0, 128767, '\P{transportandmapsymbols}', "");
Expect(1, 128767, '\P{^transportandmapsymbols}', "");
Expect(0, 128768, '\p{transportandmapsymbols}', "");
Expect(1, 128768, '\p{^transportandmapsymbols}', "");
Expect(1, 128768, '\P{transportandmapsymbols}', "");
Expect(0, 128768, '\P{^transportandmapsymbols}', "");
Expect(1, 128767, '\p{	Transport_And_MAP_Symbols}', "");
Expect(0, 128767, '\p{^	Transport_And_MAP_Symbols}', "");
Expect(0, 128767, '\P{	Transport_And_MAP_Symbols}', "");
Expect(1, 128767, '\P{^	Transport_And_MAP_Symbols}', "");
Expect(0, 128768, '\p{	Transport_And_MAP_Symbols}', "");
Expect(1, 128768, '\p{^	Transport_And_MAP_Symbols}', "");
Expect(1, 128768, '\P{	Transport_And_MAP_Symbols}', "");
Expect(0, 128768, '\P{^	Transport_And_MAP_Symbols}', "");
Error('\p{	:=Is_TRANSPORT_and_Map_Symbols}');
Error('\P{	:=Is_TRANSPORT_and_Map_Symbols}');
Expect(1, 128767, '\p{istransportandmapsymbols}', "");
Expect(0, 128767, '\p{^istransportandmapsymbols}', "");
Expect(0, 128767, '\P{istransportandmapsymbols}', "");
Expect(1, 128767, '\P{^istransportandmapsymbols}', "");
Expect(0, 128768, '\p{istransportandmapsymbols}', "");
Expect(1, 128768, '\p{^istransportandmapsymbols}', "");
Expect(1, 128768, '\P{istransportandmapsymbols}', "");
Expect(0, 128768, '\P{^istransportandmapsymbols}', "");
Expect(1, 128767, '\p{	_is_Transport_And_map_symbols}', "");
Expect(0, 128767, '\p{^	_is_Transport_And_map_symbols}', "");
Expect(0, 128767, '\P{	_is_Transport_And_map_symbols}', "");
Expect(1, 128767, '\P{^	_is_Transport_And_map_symbols}', "");
Expect(0, 128768, '\p{	_is_Transport_And_map_symbols}', "");
Expect(1, 128768, '\p{^	_is_Transport_And_map_symbols}', "");
Expect(1, 128768, '\P{	_is_Transport_And_map_symbols}', "");
Expect(0, 128768, '\P{^	_is_Transport_And_map_symbols}', "");
Error('\p{-_In_Transport_And_Map_SYMBOLS/a/}');
Error('\P{-_In_Transport_And_Map_SYMBOLS/a/}');
Expect(1, 128767, '\p{intransportandmapsymbols}', "");
Expect(0, 128767, '\p{^intransportandmapsymbols}', "");
Expect(0, 128767, '\P{intransportandmapsymbols}', "");
Expect(1, 128767, '\P{^intransportandmapsymbols}', "");
Expect(0, 128768, '\p{intransportandmapsymbols}', "");
Expect(1, 128768, '\p{^intransportandmapsymbols}', "");
Expect(1, 128768, '\P{intransportandmapsymbols}', "");
Expect(0, 128768, '\P{^intransportandmapsymbols}', "");
Expect(1, 128767, '\p{_	in_TRANSPORT_AND_map_SYMBOLS}', "");
Expect(0, 128767, '\p{^_	in_TRANSPORT_AND_map_SYMBOLS}', "");
Expect(0, 128767, '\P{_	in_TRANSPORT_AND_map_SYMBOLS}', "");
Expect(1, 128767, '\P{^_	in_TRANSPORT_AND_map_SYMBOLS}', "");
Expect(0, 128768, '\p{_	in_TRANSPORT_AND_map_SYMBOLS}', "");
Expect(1, 128768, '\p{^_	in_TRANSPORT_AND_map_SYMBOLS}', "");
Expect(1, 128768, '\P{_	in_TRANSPORT_AND_map_SYMBOLS}', "");
Expect(0, 128768, '\P{^_	in_TRANSPORT_AND_map_SYMBOLS}', "");
Error('\p{		transport_And_Map:=}');
Error('\P{		transport_And_Map:=}');
Expect(1, 128767, '\p{transportandmap}', "");
Expect(0, 128767, '\p{^transportandmap}', "");
Expect(0, 128767, '\P{transportandmap}', "");
Expect(1, 128767, '\P{^transportandmap}', "");
Expect(0, 128768, '\p{transportandmap}', "");
Expect(1, 128768, '\p{^transportandmap}', "");
Expect(1, 128768, '\P{transportandmap}', "");
Expect(0, 128768, '\P{^transportandmap}', "");
Expect(1, 128767, '\p{transport_AND_map}', "");
Expect(0, 128767, '\p{^transport_AND_map}', "");
Expect(0, 128767, '\P{transport_AND_map}', "");
Expect(1, 128767, '\P{^transport_AND_map}', "");
Expect(0, 128768, '\p{transport_AND_map}', "");
Expect(1, 128768, '\p{^transport_AND_map}', "");
Expect(1, 128768, '\P{transport_AND_map}', "");
Expect(0, 128768, '\P{^transport_AND_map}', "");
Error('\p{/a/_-Is_Transport_AND_MAP}');
Error('\P{/a/_-Is_Transport_AND_MAP}');
Expect(1, 128767, '\p{istransportandmap}', "");
Expect(0, 128767, '\p{^istransportandmap}', "");
Expect(0, 128767, '\P{istransportandmap}', "");
Expect(1, 128767, '\P{^istransportandmap}', "");
Expect(0, 128768, '\p{istransportandmap}', "");
Expect(1, 128768, '\p{^istransportandmap}', "");
Expect(1, 128768, '\P{istransportandmap}', "");
Expect(0, 128768, '\P{^istransportandmap}', "");
Expect(1, 128767, '\p{__is_TRANSPORT_AND_map}', "");
Expect(0, 128767, '\p{^__is_TRANSPORT_AND_map}', "");
Expect(0, 128767, '\P{__is_TRANSPORT_AND_map}', "");
Expect(1, 128767, '\P{^__is_TRANSPORT_AND_map}', "");
Expect(0, 128768, '\p{__is_TRANSPORT_AND_map}', "");
Expect(1, 128768, '\p{^__is_TRANSPORT_AND_map}', "");
Expect(1, 128768, '\P{__is_TRANSPORT_AND_map}', "");
Expect(0, 128768, '\P{^__is_TRANSPORT_AND_map}', "");
Error('\p{	IN_TRANSPORT_And_map:=}');
Error('\P{	IN_TRANSPORT_And_map:=}');
Expect(1, 128767, '\p{intransportandmap}', "");
Expect(0, 128767, '\p{^intransportandmap}', "");
Expect(0, 128767, '\P{intransportandmap}', "");
Expect(1, 128767, '\P{^intransportandmap}', "");
Expect(0, 128768, '\p{intransportandmap}', "");
Expect(1, 128768, '\p{^intransportandmap}', "");
Expect(1, 128768, '\P{intransportandmap}', "");
Expect(0, 128768, '\P{^intransportandmap}', "");
Expect(1, 128767, '\p{_In_Transport_AND_Map}', "");
Expect(0, 128767, '\p{^_In_Transport_AND_Map}', "");
Expect(0, 128767, '\P{_In_Transport_AND_Map}', "");
Expect(1, 128767, '\P{^_In_Transport_AND_Map}', "");
Expect(0, 128768, '\p{_In_Transport_AND_Map}', "");
Expect(1, 128768, '\p{^_In_Transport_AND_Map}', "");
Expect(1, 128768, '\P{_In_Transport_AND_Map}', "");
Expect(0, 128768, '\P{^_In_Transport_AND_Map}', "");
Error('\p{:=	ugaritic}');
Error('\P{:=	ugaritic}');
Expect(1, 66463, '\p{ugaritic}', "");
Expect(0, 66463, '\p{^ugaritic}', "");
Expect(0, 66463, '\P{ugaritic}', "");
Expect(1, 66463, '\P{^ugaritic}', "");
Expect(0, 66464, '\p{ugaritic}', "");
Expect(1, 66464, '\p{^ugaritic}', "");
Expect(1, 66464, '\P{ugaritic}', "");
Expect(0, 66464, '\P{^ugaritic}', "");
Expect(1, 66463, '\p{	 ugaritic}', "");
Expect(0, 66463, '\p{^	 ugaritic}', "");
Expect(0, 66463, '\P{	 ugaritic}', "");
Expect(1, 66463, '\P{^	 ugaritic}', "");
Expect(0, 66464, '\p{	 ugaritic}', "");
Expect(1, 66464, '\p{^	 ugaritic}', "");
Expect(1, 66464, '\P{	 ugaritic}', "");
Expect(0, 66464, '\P{^	 ugaritic}', "");
Error('\p{-/a/IS_ugaritic}');
Error('\P{-/a/IS_ugaritic}');
Expect(1, 66463, '\p{isugaritic}', "");
Expect(0, 66463, '\p{^isugaritic}', "");
Expect(0, 66463, '\P{isugaritic}', "");
Expect(1, 66463, '\P{^isugaritic}', "");
Expect(0, 66464, '\p{isugaritic}', "");
Expect(1, 66464, '\p{^isugaritic}', "");
Expect(1, 66464, '\P{isugaritic}', "");
Expect(0, 66464, '\P{^isugaritic}', "");
Expect(1, 66463, '\p{IS_Ugaritic}', "");
Expect(0, 66463, '\p{^IS_Ugaritic}', "");
Expect(0, 66463, '\P{IS_Ugaritic}', "");
Expect(1, 66463, '\P{^IS_Ugaritic}', "");
Expect(0, 66464, '\p{IS_Ugaritic}', "");
Expect(1, 66464, '\p{^IS_Ugaritic}', "");
Expect(1, 66464, '\P{IS_Ugaritic}', "");
Expect(0, 66464, '\P{^IS_Ugaritic}', "");
Error('\p{ 	Ugar/a/}');
Error('\P{ 	Ugar/a/}');
Expect(1, 66463, '\p{ugar}', "");
Expect(0, 66463, '\p{^ugar}', "");
Expect(0, 66463, '\P{ugar}', "");
Expect(1, 66463, '\P{^ugar}', "");
Expect(0, 66464, '\p{ugar}', "");
Expect(1, 66464, '\p{^ugar}', "");
Expect(1, 66464, '\P{ugar}', "");
Expect(0, 66464, '\P{^ugar}', "");
Expect(1, 66463, '\p{_ Ugar}', "");
Expect(0, 66463, '\p{^_ Ugar}', "");
Expect(0, 66463, '\P{_ Ugar}', "");
Expect(1, 66463, '\P{^_ Ugar}', "");
Expect(0, 66464, '\p{_ Ugar}', "");
Expect(1, 66464, '\p{^_ Ugar}', "");
Expect(1, 66464, '\P{_ Ugar}', "");
Expect(0, 66464, '\P{^_ Ugar}', "");
Error('\p{ /a/Is_Ugar}');
Error('\P{ /a/Is_Ugar}');
Expect(1, 66463, '\p{isugar}', "");
Expect(0, 66463, '\p{^isugar}', "");
Expect(0, 66463, '\P{isugar}', "");
Expect(1, 66463, '\P{^isugar}', "");
Expect(0, 66464, '\p{isugar}', "");
Expect(1, 66464, '\p{^isugar}', "");
Expect(1, 66464, '\P{isugar}', "");
Expect(0, 66464, '\P{^isugar}', "");
Expect(1, 66463, '\p{__Is_Ugar}', "");
Expect(0, 66463, '\p{^__Is_Ugar}', "");
Expect(0, 66463, '\P{__Is_Ugar}', "");
Expect(1, 66463, '\P{^__Is_Ugar}', "");
Expect(0, 66464, '\p{__Is_Ugar}', "");
Expect(1, 66464, '\p{^__Is_Ugar}', "");
Expect(1, 66464, '\P{__Is_Ugar}', "");
Expect(0, 66464, '\P{^__Is_Ugar}', "");
Error('\p{:=Unassigned}');
Error('\P{:=Unassigned}');
Expect(1, 918000, '\p{unassigned}', "");
Expect(0, 918000, '\p{^unassigned}', "");
Expect(0, 918000, '\P{unassigned}', "");
Expect(1, 918000, '\P{^unassigned}', "");
Expect(0, 1114109, '\p{unassigned}', "");
Expect(1, 1114109, '\p{^unassigned}', "");
Expect(1, 1114109, '\P{unassigned}', "");
Expect(0, 1114109, '\P{^unassigned}', "");
Expect(1, 918000, '\p{_Unassigned}', "");
Expect(0, 918000, '\p{^_Unassigned}', "");
Expect(0, 918000, '\P{_Unassigned}', "");
Expect(1, 918000, '\P{^_Unassigned}', "");
Expect(0, 1114109, '\p{_Unassigned}', "");
Expect(1, 1114109, '\p{^_Unassigned}', "");
Expect(1, 1114109, '\P{_Unassigned}', "");
Expect(0, 1114109, '\P{^_Unassigned}', "");
Error('\p{	Is_Unassigned:=}');
Error('\P{	Is_Unassigned:=}');
Expect(1, 918000, '\p{isunassigned}', "");
Expect(0, 918000, '\p{^isunassigned}', "");
Expect(0, 918000, '\P{isunassigned}', "");
Expect(1, 918000, '\P{^isunassigned}', "");
Expect(0, 1114109, '\p{isunassigned}', "");
Expect(1, 1114109, '\p{^isunassigned}', "");
Expect(1, 1114109, '\P{isunassigned}', "");
Expect(0, 1114109, '\P{^isunassigned}', "");
Expect(1, 918000, '\p{ Is_Unassigned}', "");
Expect(0, 918000, '\p{^ Is_Unassigned}', "");
Expect(0, 918000, '\P{ Is_Unassigned}', "");
Expect(1, 918000, '\P{^ Is_Unassigned}', "");
Expect(0, 1114109, '\p{ Is_Unassigned}', "");
Expect(1, 1114109, '\p{^ Is_Unassigned}', "");
Expect(1, 1114109, '\P{ Is_Unassigned}', "");
Expect(0, 1114109, '\P{^ Is_Unassigned}', "");
Error('\p{  Cn:=}');
Error('\P{  Cn:=}');
Expect(1, 918000, '\p{cn}', "");
Expect(0, 918000, '\p{^cn}', "");
Expect(0, 918000, '\P{cn}', "");
Expect(1, 918000, '\P{^cn}', "");
Expect(0, 1114109, '\p{cn}', "");
Expect(1, 1114109, '\p{^cn}', "");
Expect(1, 1114109, '\P{cn}', "");
Expect(0, 1114109, '\P{^cn}', "");
Expect(1, 918000, '\p{ 	CN}', "");
Expect(0, 918000, '\p{^ 	CN}', "");
Expect(0, 918000, '\P{ 	CN}', "");
Expect(1, 918000, '\P{^ 	CN}', "");
Expect(0, 1114109, '\p{ 	CN}', "");
Expect(1, 1114109, '\p{^ 	CN}', "");
Expect(1, 1114109, '\P{ 	CN}', "");
Expect(0, 1114109, '\P{^ 	CN}', "");
Error('\p{:= 	Is_Cn}');
Error('\P{:= 	Is_Cn}');
Expect(1, 918000, '\p{iscn}', "");
Expect(0, 918000, '\p{^iscn}', "");
Expect(0, 918000, '\P{iscn}', "");
Expect(1, 918000, '\P{^iscn}', "");
Expect(0, 1114109, '\p{iscn}', "");
Expect(1, 1114109, '\p{^iscn}', "");
Expect(1, 1114109, '\P{iscn}', "");
Expect(0, 1114109, '\P{^iscn}', "");
Expect(1, 918000, '\p{- Is_CN}', "");
Expect(0, 918000, '\p{^- Is_CN}', "");
Expect(0, 918000, '\P{- Is_CN}', "");
Expect(1, 918000, '\P{^- Is_CN}', "");
Expect(0, 1114109, '\p{- Is_CN}', "");
Expect(1, 1114109, '\p{^- Is_CN}', "");
Expect(1, 1114109, '\P{- Is_CN}', "");
Expect(0, 1114109, '\P{^- Is_CN}', "");
Error('\p{/a/__unified_canadian_ABORIGINAL_Syllabics}');
Error('\P{/a/__unified_canadian_ABORIGINAL_Syllabics}');
Expect(1, 5759, '\p{unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\p{^unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\P{unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\P{^unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\p{unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\p{^unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\P{unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\P{^unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\p{- Unified_Canadian_aboriginal_Syllabics}', "");
Expect(0, 5759, '\p{^- Unified_Canadian_aboriginal_Syllabics}', "");
Expect(0, 5759, '\P{- Unified_Canadian_aboriginal_Syllabics}', "");
Expect(1, 5759, '\P{^- Unified_Canadian_aboriginal_Syllabics}', "");
Expect(0, 5760, '\p{- Unified_Canadian_aboriginal_Syllabics}', "");
Expect(1, 5760, '\p{^- Unified_Canadian_aboriginal_Syllabics}', "");
Expect(1, 5760, '\P{- Unified_Canadian_aboriginal_Syllabics}', "");
Expect(0, 5760, '\P{^- Unified_Canadian_aboriginal_Syllabics}', "");
Error('\p{:=		is_UNIFIED_Canadian_aboriginal_Syllabics}');
Error('\P{:=		is_UNIFIED_Canadian_aboriginal_Syllabics}');
Expect(1, 5759, '\p{isunifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\p{^isunifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\P{isunifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\P{^isunifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\p{isunifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\p{^isunifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\P{isunifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\P{^isunifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\p{_-Is_Unified_canadian_Aboriginal_SYLLABICS}', "");
Expect(0, 5759, '\p{^_-Is_Unified_canadian_Aboriginal_SYLLABICS}', "");
Expect(0, 5759, '\P{_-Is_Unified_canadian_Aboriginal_SYLLABICS}', "");
Expect(1, 5759, '\P{^_-Is_Unified_canadian_Aboriginal_SYLLABICS}', "");
Expect(0, 5760, '\p{_-Is_Unified_canadian_Aboriginal_SYLLABICS}', "");
Expect(1, 5760, '\p{^_-Is_Unified_canadian_Aboriginal_SYLLABICS}', "");
Expect(1, 5760, '\P{_-Is_Unified_canadian_Aboriginal_SYLLABICS}', "");
Expect(0, 5760, '\P{^_-Is_Unified_canadian_Aboriginal_SYLLABICS}', "");
Error('\p{/a/_ In_unified_canadian_Aboriginal_Syllabics}');
Error('\P{/a/_ In_unified_canadian_Aboriginal_Syllabics}');
Expect(1, 5759, '\p{inunifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\p{^inunifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\P{inunifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\P{^inunifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\p{inunifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\p{^inunifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\P{inunifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\P{^inunifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\p{_ in_Unified_Canadian_Aboriginal_syllabics}', "");
Expect(0, 5759, '\p{^_ in_Unified_Canadian_Aboriginal_syllabics}', "");
Expect(0, 5759, '\P{_ in_Unified_Canadian_Aboriginal_syllabics}', "");
Expect(1, 5759, '\P{^_ in_Unified_Canadian_Aboriginal_syllabics}', "");
Expect(0, 5760, '\p{_ in_Unified_Canadian_Aboriginal_syllabics}', "");
Expect(1, 5760, '\p{^_ in_Unified_Canadian_Aboriginal_syllabics}', "");
Expect(1, 5760, '\P{_ in_Unified_Canadian_Aboriginal_syllabics}', "");
Expect(0, 5760, '\P{^_ in_Unified_Canadian_Aboriginal_syllabics}', "");
Error('\p{:=_	UCAS}');
Error('\P{:=_	UCAS}');
Expect(1, 5759, '\p{ucas}', "");
Expect(0, 5759, '\p{^ucas}', "");
Expect(0, 5759, '\P{ucas}', "");
Expect(1, 5759, '\P{^ucas}', "");
Expect(0, 5760, '\p{ucas}', "");
Expect(1, 5760, '\p{^ucas}', "");
Expect(1, 5760, '\P{ucas}', "");
Expect(0, 5760, '\P{^ucas}', "");
Expect(1, 5759, '\p{-UCAS}', "");
Expect(0, 5759, '\p{^-UCAS}', "");
Expect(0, 5759, '\P{-UCAS}', "");
Expect(1, 5759, '\P{^-UCAS}', "");
Expect(0, 5760, '\p{-UCAS}', "");
Expect(1, 5760, '\p{^-UCAS}', "");
Expect(1, 5760, '\P{-UCAS}', "");
Expect(0, 5760, '\P{^-UCAS}', "");
Error('\p{-Is_ucas:=}');
Error('\P{-Is_ucas:=}');
Expect(1, 5759, '\p{isucas}', "");
Expect(0, 5759, '\p{^isucas}', "");
Expect(0, 5759, '\P{isucas}', "");
Expect(1, 5759, '\P{^isucas}', "");
Expect(0, 5760, '\p{isucas}', "");
Expect(1, 5760, '\p{^isucas}', "");
Expect(1, 5760, '\P{isucas}', "");
Expect(0, 5760, '\P{^isucas}', "");
Expect(1, 5759, '\p{		Is_UCAS}', "");
Expect(0, 5759, '\p{^		Is_UCAS}', "");
Expect(0, 5759, '\P{		Is_UCAS}', "");
Expect(1, 5759, '\P{^		Is_UCAS}', "");
Expect(0, 5760, '\p{		Is_UCAS}', "");
Expect(1, 5760, '\p{^		Is_UCAS}', "");
Expect(1, 5760, '\P{		Is_UCAS}', "");
Expect(0, 5760, '\P{^		Is_UCAS}', "");
Error('\p{:=IN_ucas}');
Error('\P{:=IN_ucas}');
Expect(1, 5759, '\p{inucas}', "");
Expect(0, 5759, '\p{^inucas}', "");
Expect(0, 5759, '\P{inucas}', "");
Expect(1, 5759, '\P{^inucas}', "");
Expect(0, 5760, '\p{inucas}', "");
Expect(1, 5760, '\p{^inucas}', "");
Expect(1, 5760, '\P{inucas}', "");
Expect(0, 5760, '\P{^inucas}', "");
Expect(1, 5759, '\p{  IN_ucas}', "");
Expect(0, 5759, '\p{^  IN_ucas}', "");
Expect(0, 5759, '\P{  IN_ucas}', "");
Expect(1, 5759, '\P{^  IN_ucas}', "");
Expect(0, 5760, '\p{  IN_ucas}', "");
Expect(1, 5760, '\p{^  IN_ucas}', "");
Expect(1, 5760, '\P{  IN_ucas}', "");
Expect(0, 5760, '\P{^  IN_ucas}', "");
Error('\p{  CANADIAN_Syllabics/a/}');
Error('\P{  CANADIAN_Syllabics/a/}');
Expect(1, 5759, '\p{canadiansyllabics}', "");
Expect(0, 5759, '\p{^canadiansyllabics}', "");
Expect(0, 5759, '\P{canadiansyllabics}', "");
Expect(1, 5759, '\P{^canadiansyllabics}', "");
Expect(0, 5760, '\p{canadiansyllabics}', "");
Expect(1, 5760, '\p{^canadiansyllabics}', "");
Expect(1, 5760, '\P{canadiansyllabics}', "");
Expect(0, 5760, '\P{^canadiansyllabics}', "");
Expect(1, 5759, '\p{_	Canadian_syllabics}', "");
Expect(0, 5759, '\p{^_	Canadian_syllabics}', "");
Expect(0, 5759, '\P{_	Canadian_syllabics}', "");
Expect(1, 5759, '\P{^_	Canadian_syllabics}', "");
Expect(0, 5760, '\p{_	Canadian_syllabics}', "");
Expect(1, 5760, '\p{^_	Canadian_syllabics}', "");
Expect(1, 5760, '\P{_	Canadian_syllabics}', "");
Expect(0, 5760, '\P{^_	Canadian_syllabics}', "");
Error('\p{- IS_Canadian_syllabics/a/}');
Error('\P{- IS_Canadian_syllabics/a/}');
Expect(1, 5759, '\p{iscanadiansyllabics}', "");
Expect(0, 5759, '\p{^iscanadiansyllabics}', "");
Expect(0, 5759, '\P{iscanadiansyllabics}', "");
Expect(1, 5759, '\P{^iscanadiansyllabics}', "");
Expect(0, 5760, '\p{iscanadiansyllabics}', "");
Expect(1, 5760, '\p{^iscanadiansyllabics}', "");
Expect(1, 5760, '\P{iscanadiansyllabics}', "");
Expect(0, 5760, '\P{^iscanadiansyllabics}', "");
Expect(1, 5759, '\p{	 Is_CANADIAN_Syllabics}', "");
Expect(0, 5759, '\p{^	 Is_CANADIAN_Syllabics}', "");
Expect(0, 5759, '\P{	 Is_CANADIAN_Syllabics}', "");
Expect(1, 5759, '\P{^	 Is_CANADIAN_Syllabics}', "");
Expect(0, 5760, '\p{	 Is_CANADIAN_Syllabics}', "");
Expect(1, 5760, '\p{^	 Is_CANADIAN_Syllabics}', "");
Expect(1, 5760, '\P{	 Is_CANADIAN_Syllabics}', "");
Expect(0, 5760, '\P{^	 Is_CANADIAN_Syllabics}', "");
Error('\p{-/a/In_Canadian_Syllabics}');
Error('\P{-/a/In_Canadian_Syllabics}');
Expect(1, 5759, '\p{incanadiansyllabics}', "");
Expect(0, 5759, '\p{^incanadiansyllabics}', "");
Expect(0, 5759, '\P{incanadiansyllabics}', "");
Expect(1, 5759, '\P{^incanadiansyllabics}', "");
Expect(0, 5760, '\p{incanadiansyllabics}', "");
Expect(1, 5760, '\p{^incanadiansyllabics}', "");
Expect(1, 5760, '\P{incanadiansyllabics}', "");
Expect(0, 5760, '\P{^incanadiansyllabics}', "");
Expect(1, 5759, '\p{ IN_canadian_SYLLABICS}', "");
Expect(0, 5759, '\p{^ IN_canadian_SYLLABICS}', "");
Expect(0, 5759, '\P{ IN_canadian_SYLLABICS}', "");
Expect(1, 5759, '\P{^ IN_canadian_SYLLABICS}', "");
Expect(0, 5760, '\p{ IN_canadian_SYLLABICS}', "");
Expect(1, 5760, '\p{^ IN_canadian_SYLLABICS}', "");
Expect(1, 5760, '\P{ IN_canadian_SYLLABICS}', "");
Expect(0, 5760, '\P{^ IN_canadian_SYLLABICS}', "");
Error('\p{/a/_Unified_Canadian_Aboriginal_SYLLABICS_EXTENDED}');
Error('\P{/a/_Unified_Canadian_Aboriginal_SYLLABICS_EXTENDED}');
Expect(1, 6399, '\p{unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\p{^unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\P{unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\P{^unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\p{unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\p{^unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\P{unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\P{^unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\p{ -Unified_Canadian_ABORIGINAL_syllabics_Extended}', "");
Expect(0, 6399, '\p{^ -Unified_Canadian_ABORIGINAL_syllabics_Extended}', "");
Expect(0, 6399, '\P{ -Unified_Canadian_ABORIGINAL_syllabics_Extended}', "");
Expect(1, 6399, '\P{^ -Unified_Canadian_ABORIGINAL_syllabics_Extended}', "");
Expect(0, 6400, '\p{ -Unified_Canadian_ABORIGINAL_syllabics_Extended}', "");
Expect(1, 6400, '\p{^ -Unified_Canadian_ABORIGINAL_syllabics_Extended}', "");
Expect(1, 6400, '\P{ -Unified_Canadian_ABORIGINAL_syllabics_Extended}', "");
Expect(0, 6400, '\P{^ -Unified_Canadian_ABORIGINAL_syllabics_Extended}', "");
Error('\p{	/a/Is_Unified_CANADIAN_Aboriginal_syllabics_EXTENDED}');
Error('\P{	/a/Is_Unified_CANADIAN_Aboriginal_syllabics_EXTENDED}');
Expect(1, 6399, '\p{isunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\p{^isunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\P{isunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\P{^isunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\p{isunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\p{^isunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\P{isunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\P{^isunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\p{_IS_UNIFIED_Canadian_aboriginal_SYLLABICS_extended}', "");
Expect(0, 6399, '\p{^_IS_UNIFIED_Canadian_aboriginal_SYLLABICS_extended}', "");
Expect(0, 6399, '\P{_IS_UNIFIED_Canadian_aboriginal_SYLLABICS_extended}', "");
Expect(1, 6399, '\P{^_IS_UNIFIED_Canadian_aboriginal_SYLLABICS_extended}', "");
Expect(0, 6400, '\p{_IS_UNIFIED_Canadian_aboriginal_SYLLABICS_extended}', "");
Expect(1, 6400, '\p{^_IS_UNIFIED_Canadian_aboriginal_SYLLABICS_extended}', "");
Expect(1, 6400, '\P{_IS_UNIFIED_Canadian_aboriginal_SYLLABICS_extended}', "");
Expect(0, 6400, '\P{^_IS_UNIFIED_Canadian_aboriginal_SYLLABICS_extended}', "");
Error('\p{:=-	In_unified_canadian_ABORIGINAL_Syllabics_EXTENDED}');
Error('\P{:=-	In_unified_canadian_ABORIGINAL_Syllabics_EXTENDED}');
Expect(1, 6399, '\p{inunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\p{^inunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\P{inunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\P{^inunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\p{inunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\p{^inunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\P{inunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\P{^inunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\p{_-IN_UNIFIED_Canadian_Aboriginal_syllabics_extended}', "");
Expect(0, 6399, '\p{^_-IN_UNIFIED_Canadian_Aboriginal_syllabics_extended}', "");
Expect(0, 6399, '\P{_-IN_UNIFIED_Canadian_Aboriginal_syllabics_extended}', "");
Expect(1, 6399, '\P{^_-IN_UNIFIED_Canadian_Aboriginal_syllabics_extended}', "");
Expect(0, 6400, '\p{_-IN_UNIFIED_Canadian_Aboriginal_syllabics_extended}', "");
Expect(1, 6400, '\p{^_-IN_UNIFIED_Canadian_Aboriginal_syllabics_extended}', "");
Expect(1, 6400, '\P{_-IN_UNIFIED_Canadian_Aboriginal_syllabics_extended}', "");
Expect(0, 6400, '\P{^_-IN_UNIFIED_Canadian_Aboriginal_syllabics_extended}', "");
Error('\p{/a/		ucas_Ext}');
Error('\P{/a/		ucas_Ext}');
Expect(1, 6399, '\p{ucasext}', "");
Expect(0, 6399, '\p{^ucasext}', "");
Expect(0, 6399, '\P{ucasext}', "");
Expect(1, 6399, '\P{^ucasext}', "");
Expect(0, 6400, '\p{ucasext}', "");
Expect(1, 6400, '\p{^ucasext}', "");
Expect(1, 6400, '\P{ucasext}', "");
Expect(0, 6400, '\P{^ucasext}', "");
Expect(1, 6399, '\p{- UCAS_ext}', "");
Expect(0, 6399, '\p{^- UCAS_ext}', "");
Expect(0, 6399, '\P{- UCAS_ext}', "");
Expect(1, 6399, '\P{^- UCAS_ext}', "");
Expect(0, 6400, '\p{- UCAS_ext}', "");
Expect(1, 6400, '\p{^- UCAS_ext}', "");
Expect(1, 6400, '\P{- UCAS_ext}', "");
Expect(0, 6400, '\P{^- UCAS_ext}', "");
Error('\p{/a/- Is_UCAS_Ext}');
Error('\P{/a/- Is_UCAS_Ext}');
Expect(1, 6399, '\p{isucasext}', "");
Expect(0, 6399, '\p{^isucasext}', "");
Expect(0, 6399, '\P{isucasext}', "");
Expect(1, 6399, '\P{^isucasext}', "");
Expect(0, 6400, '\p{isucasext}', "");
Expect(1, 6400, '\p{^isucasext}', "");
Expect(1, 6400, '\P{isucasext}', "");
Expect(0, 6400, '\P{^isucasext}', "");
Expect(1, 6399, '\p{_Is_ucas_ext}', "");
Expect(0, 6399, '\p{^_Is_ucas_ext}', "");
Expect(0, 6399, '\P{_Is_ucas_ext}', "");
Expect(1, 6399, '\P{^_Is_ucas_ext}', "");
Expect(0, 6400, '\p{_Is_ucas_ext}', "");
Expect(1, 6400, '\p{^_Is_ucas_ext}', "");
Expect(1, 6400, '\P{_Is_ucas_ext}', "");
Expect(0, 6400, '\P{^_Is_ucas_ext}', "");
Error('\p{/a/-in_UCAS_Ext}');
Error('\P{/a/-in_UCAS_Ext}');
Expect(1, 6399, '\p{inucasext}', "");
Expect(0, 6399, '\p{^inucasext}', "");
Expect(0, 6399, '\P{inucasext}', "");
Expect(1, 6399, '\P{^inucasext}', "");
Expect(0, 6400, '\p{inucasext}', "");
Expect(1, 6400, '\p{^inucasext}', "");
Expect(1, 6400, '\P{inucasext}', "");
Expect(0, 6400, '\P{^inucasext}', "");
Expect(1, 6399, '\p{	in_UCAS_EXT}', "");
Expect(0, 6399, '\p{^	in_UCAS_EXT}', "");
Expect(0, 6399, '\P{	in_UCAS_EXT}', "");
Expect(1, 6399, '\P{^	in_UCAS_EXT}', "");
Expect(0, 6400, '\p{	in_UCAS_EXT}', "");
Expect(1, 6400, '\p{^	in_UCAS_EXT}', "");
Expect(1, 6400, '\P{	in_UCAS_EXT}', "");
Expect(0, 6400, '\P{^	in_UCAS_EXT}', "");
Error('\p{ unified_Ideograph/a/}');
Error('\P{ unified_Ideograph/a/}');
Expect(1, 178205, '\p{unifiedideograph}', "");
Expect(0, 178205, '\p{^unifiedideograph}', "");
Expect(0, 178205, '\P{unifiedideograph}', "");
Expect(1, 178205, '\P{^unifiedideograph}', "");
Expect(0, 178206, '\p{unifiedideograph}', "");
Expect(1, 178206, '\p{^unifiedideograph}', "");
Expect(1, 178206, '\P{unifiedideograph}', "");
Expect(0, 178206, '\P{^unifiedideograph}', "");
Expect(1, 178205, '\p{	unified_ideograph}', "");
Expect(0, 178205, '\p{^	unified_ideograph}', "");
Expect(0, 178205, '\P{	unified_ideograph}', "");
Expect(1, 178205, '\P{^	unified_ideograph}', "");
Expect(0, 178206, '\p{	unified_ideograph}', "");
Expect(1, 178206, '\p{^	unified_ideograph}', "");
Expect(1, 178206, '\P{	unified_ideograph}', "");
Expect(0, 178206, '\P{^	unified_ideograph}', "");
Error('\p{/a/	 Is_unified_IDEOGRAPH}');
Error('\P{/a/	 Is_unified_IDEOGRAPH}');
Expect(1, 178205, '\p{isunifiedideograph}', "");
Expect(0, 178205, '\p{^isunifiedideograph}', "");
Expect(0, 178205, '\P{isunifiedideograph}', "");
Expect(1, 178205, '\P{^isunifiedideograph}', "");
Expect(0, 178206, '\p{isunifiedideograph}', "");
Expect(1, 178206, '\p{^isunifiedideograph}', "");
Expect(1, 178206, '\P{isunifiedideograph}', "");
Expect(0, 178206, '\P{^isunifiedideograph}', "");
Expect(1, 178205, '\p{Is_Unified_ideograph}', "");
Expect(0, 178205, '\p{^Is_Unified_ideograph}', "");
Expect(0, 178205, '\P{Is_Unified_ideograph}', "");
Expect(1, 178205, '\P{^Is_Unified_ideograph}', "");
Expect(0, 178206, '\p{Is_Unified_ideograph}', "");
Expect(1, 178206, '\p{^Is_Unified_ideograph}', "");
Expect(1, 178206, '\P{Is_Unified_ideograph}', "");
Expect(0, 178206, '\P{^Is_Unified_ideograph}', "");
Error('\p{/a/_UIDEO}');
Error('\P{/a/_UIDEO}');
Expect(1, 178205, '\p{uideo}', "");
Expect(0, 178205, '\p{^uideo}', "");
Expect(0, 178205, '\P{uideo}', "");
Expect(1, 178205, '\P{^uideo}', "");
Expect(0, 178206, '\p{uideo}', "");
Expect(1, 178206, '\p{^uideo}', "");
Expect(1, 178206, '\P{uideo}', "");
Expect(0, 178206, '\P{^uideo}', "");
Expect(1, 178205, '\p{_ UIDEO}', "");
Expect(0, 178205, '\p{^_ UIDEO}', "");
Expect(0, 178205, '\P{_ UIDEO}', "");
Expect(1, 178205, '\P{^_ UIDEO}', "");
Expect(0, 178206, '\p{_ UIDEO}', "");
Expect(1, 178206, '\p{^_ UIDEO}', "");
Expect(1, 178206, '\P{_ UIDEO}', "");
Expect(0, 178206, '\P{^_ UIDEO}', "");
Error('\p{-IS_UIdeo:=}');
Error('\P{-IS_UIdeo:=}');
Expect(1, 178205, '\p{isuideo}', "");
Expect(0, 178205, '\p{^isuideo}', "");
Expect(0, 178205, '\P{isuideo}', "");
Expect(1, 178205, '\P{^isuideo}', "");
Expect(0, 178206, '\p{isuideo}', "");
Expect(1, 178206, '\p{^isuideo}', "");
Expect(1, 178206, '\P{isuideo}', "");
Expect(0, 178206, '\P{^isuideo}', "");
Expect(1, 178205, '\p{_	Is_UIdeo}', "");
Expect(0, 178205, '\p{^_	Is_UIdeo}', "");
Expect(0, 178205, '\P{_	Is_UIdeo}', "");
Expect(1, 178205, '\P{^_	Is_UIdeo}', "");
Expect(0, 178206, '\p{_	Is_UIdeo}', "");
Expect(1, 178206, '\p{^_	Is_UIdeo}', "");
Expect(1, 178206, '\P{_	Is_UIdeo}', "");
Expect(0, 178206, '\P{^_	Is_UIdeo}', "");
Error('\p{	 Unknown:=}');
Error('\P{	 Unknown:=}');
Expect(1, 918000, '\p{unknown}', "");
Expect(0, 918000, '\p{^unknown}', "");
Expect(0, 918000, '\P{unknown}', "");
Expect(1, 918000, '\P{^unknown}', "");
Expect(0, 917999, '\p{unknown}', "");
Expect(1, 917999, '\p{^unknown}', "");
Expect(1, 917999, '\P{unknown}', "");
Expect(0, 917999, '\P{^unknown}', "");
Expect(1, 918000, '\p{-unknown}', "");
Expect(0, 918000, '\p{^-unknown}', "");
Expect(0, 918000, '\P{-unknown}', "");
Expect(1, 918000, '\P{^-unknown}', "");
Expect(0, 917999, '\p{-unknown}', "");
Expect(1, 917999, '\p{^-unknown}', "");
Expect(1, 917999, '\P{-unknown}', "");
Expect(0, 917999, '\P{^-unknown}', "");
Error('\p{:=Is_Unknown}');
Error('\P{:=Is_Unknown}');
Expect(1, 918000, '\p{isunknown}', "");
Expect(0, 918000, '\p{^isunknown}', "");
Expect(0, 918000, '\P{isunknown}', "");
Expect(1, 918000, '\P{^isunknown}', "");
Expect(0, 917999, '\p{isunknown}', "");
Expect(1, 917999, '\p{^isunknown}', "");
Expect(1, 917999, '\P{isunknown}', "");
Expect(0, 917999, '\P{^isunknown}', "");
Expect(1, 918000, '\p{	_Is_Unknown}', "");
Expect(0, 918000, '\p{^	_Is_Unknown}', "");
Expect(0, 918000, '\P{	_Is_Unknown}', "");
Expect(1, 918000, '\P{^	_Is_Unknown}', "");
Expect(0, 917999, '\p{	_Is_Unknown}', "");
Expect(1, 917999, '\p{^	_Is_Unknown}', "");
Expect(1, 917999, '\P{	_Is_Unknown}', "");
Expect(0, 917999, '\P{^	_Is_Unknown}', "");
Error('\p{	_ZZZZ/a/}');
Error('\P{	_ZZZZ/a/}');
Expect(1, 918000, '\p{zzzz}', "");
Expect(0, 918000, '\p{^zzzz}', "");
Expect(0, 918000, '\P{zzzz}', "");
Expect(1, 918000, '\P{^zzzz}', "");
Expect(0, 917999, '\p{zzzz}', "");
Expect(1, 917999, '\p{^zzzz}', "");
Expect(1, 917999, '\P{zzzz}', "");
Expect(0, 917999, '\P{^zzzz}', "");
Expect(1, 918000, '\p{ Zzzz}', "");
Expect(0, 918000, '\p{^ Zzzz}', "");
Expect(0, 918000, '\P{ Zzzz}', "");
Expect(1, 918000, '\P{^ Zzzz}', "");
Expect(0, 917999, '\p{ Zzzz}', "");
Expect(1, 917999, '\p{^ Zzzz}', "");
Expect(1, 917999, '\P{ Zzzz}', "");
Expect(0, 917999, '\P{^ Zzzz}', "");
Error('\p{-/a/IS_Zzzz}');
Error('\P{-/a/IS_Zzzz}');
Expect(1, 918000, '\p{iszzzz}', "");
Expect(0, 918000, '\p{^iszzzz}', "");
Expect(0, 918000, '\P{iszzzz}', "");
Expect(1, 918000, '\P{^iszzzz}', "");
Expect(0, 917999, '\p{iszzzz}', "");
Expect(1, 917999, '\p{^iszzzz}', "");
Expect(1, 917999, '\P{iszzzz}', "");
Expect(0, 917999, '\P{^iszzzz}', "");
Expect(1, 918000, '\p{	-is_zzzz}', "");
Expect(0, 918000, '\p{^	-is_zzzz}', "");
Expect(0, 918000, '\P{	-is_zzzz}', "");
Expect(1, 918000, '\P{^	-is_zzzz}', "");
Expect(0, 917999, '\p{	-is_zzzz}', "");
Expect(1, 917999, '\p{^	-is_zzzz}', "");
Expect(1, 917999, '\P{	-is_zzzz}', "");
Expect(0, 917999, '\P{^	-is_zzzz}', "");
Error('\p{	/a/upper}');
Error('\P{	/a/upper}');
Expect(1, 120778, '\p{upper}', "");
Expect(0, 120778, '\p{^upper}', "");
Expect(0, 120778, '\P{upper}', "");
Expect(1, 120778, '\P{^upper}', "");
Expect(0, 120779, '\p{upper}', "");
Expect(1, 120779, '\p{^upper}', "");
Expect(1, 120779, '\P{upper}', "");
Expect(0, 120779, '\P{^upper}', "");
Expect(1, 120778, '\p{--upper}', "");
Expect(0, 120778, '\p{^--upper}', "");
Expect(0, 120778, '\P{--upper}', "");
Expect(1, 120778, '\P{^--upper}', "");
Expect(0, 120779, '\p{--upper}', "");
Expect(1, 120779, '\p{^--upper}', "");
Expect(1, 120779, '\P{--upper}', "");
Expect(0, 120779, '\P{^--upper}', "");
Error('\p{	 xposixupper:=}');
Error('\P{	 xposixupper:=}');
Expect(1, 120778, '\p{xposixupper}', "");
Expect(0, 120778, '\p{^xposixupper}', "");
Expect(0, 120778, '\P{xposixupper}', "");
Expect(1, 120778, '\P{^xposixupper}', "");
Expect(0, 120779, '\p{xposixupper}', "");
Expect(1, 120779, '\p{^xposixupper}', "");
Expect(1, 120779, '\P{xposixupper}', "");
Expect(0, 120779, '\P{^xposixupper}', "");
Expect(1, 120778, '\p{	_xposixupper}', "");
Expect(0, 120778, '\p{^	_xposixupper}', "");
Expect(0, 120778, '\P{	_xposixupper}', "");
Expect(1, 120778, '\P{^	_xposixupper}', "");
Expect(0, 120779, '\p{	_xposixupper}', "");
Expect(1, 120779, '\p{^	_xposixupper}', "");
Expect(1, 120779, '\P{	_xposixupper}', "");
Expect(0, 120779, '\P{^	_xposixupper}', "");
Error('\p{-:=Is_UPPER}');
Error('\P{-:=Is_UPPER}');
Expect(1, 120778, '\p{isupper}', "");
Expect(0, 120778, '\p{^isupper}', "");
Expect(0, 120778, '\P{isupper}', "");
Expect(1, 120778, '\P{^isupper}', "");
Expect(0, 120779, '\p{isupper}', "");
Expect(1, 120779, '\p{^isupper}', "");
Expect(1, 120779, '\P{isupper}', "");
Expect(0, 120779, '\P{^isupper}', "");
Expect(1, 120778, '\p{ Is_upper}', "");
Expect(0, 120778, '\p{^ Is_upper}', "");
Expect(0, 120778, '\P{ Is_upper}', "");
Expect(1, 120778, '\P{^ Is_upper}', "");
Expect(0, 120779, '\p{ Is_upper}', "");
Expect(1, 120779, '\p{^ Is_upper}', "");
Expect(1, 120779, '\P{ Is_upper}', "");
Expect(0, 120779, '\P{^ Is_upper}', "");
Error('\p{/a/_-is_XPosixUpper}');
Error('\P{/a/_-is_XPosixUpper}');
Expect(1, 120778, '\p{isxposixupper}', "");
Expect(0, 120778, '\p{^isxposixupper}', "");
Expect(0, 120778, '\P{isxposixupper}', "");
Expect(1, 120778, '\P{^isxposixupper}', "");
Expect(0, 120779, '\p{isxposixupper}', "");
Expect(1, 120779, '\p{^isxposixupper}', "");
Expect(1, 120779, '\P{isxposixupper}', "");
Expect(0, 120779, '\P{^isxposixupper}', "");
Expect(1, 120778, '\p{_ is_XPOSIXUPPER}', "");
Expect(0, 120778, '\p{^_ is_XPOSIXUPPER}', "");
Expect(0, 120778, '\P{_ is_XPOSIXUPPER}', "");
Expect(1, 120778, '\P{^_ is_XPOSIXUPPER}', "");
Expect(0, 120779, '\p{_ is_XPOSIXUPPER}', "");
Expect(1, 120779, '\p{^_ is_XPOSIXUPPER}', "");
Expect(1, 120779, '\P{_ is_XPOSIXUPPER}', "");
Expect(0, 120779, '\P{^_ is_XPOSIXUPPER}', "");
Error('\p{/a/	-Uppercase}');
Error('\P{/a/	-Uppercase}');
Expect(1, 120778, '\p{uppercase}', "");
Expect(0, 120778, '\p{^uppercase}', "");
Expect(0, 120778, '\P{uppercase}', "");
Expect(1, 120778, '\P{^uppercase}', "");
Expect(0, 120779, '\p{uppercase}', "");
Expect(1, 120779, '\p{^uppercase}', "");
Expect(1, 120779, '\P{uppercase}', "");
Expect(0, 120779, '\P{^uppercase}', "");
Expect(1, 120778, '\p{ 	Uppercase}', "");
Expect(0, 120778, '\p{^ 	Uppercase}', "");
Expect(0, 120778, '\P{ 	Uppercase}', "");
Expect(1, 120778, '\P{^ 	Uppercase}', "");
Expect(0, 120779, '\p{ 	Uppercase}', "");
Expect(1, 120779, '\p{^ 	Uppercase}', "");
Expect(1, 120779, '\P{ 	Uppercase}', "");
Expect(0, 120779, '\P{^ 	Uppercase}', "");
Error('\p{:=	is_uppercase}');
Error('\P{:=	is_uppercase}');
Expect(1, 120778, '\p{isuppercase}', "");
Expect(0, 120778, '\p{^isuppercase}', "");
Expect(0, 120778, '\P{isuppercase}', "");
Expect(1, 120778, '\P{^isuppercase}', "");
Expect(0, 120779, '\p{isuppercase}', "");
Expect(1, 120779, '\p{^isuppercase}', "");
Expect(1, 120779, '\P{isuppercase}', "");
Expect(0, 120779, '\P{^isuppercase}', "");
Expect(1, 120778, '\p{-is_uppercase}', "");
Expect(0, 120778, '\p{^-is_uppercase}', "");
Expect(0, 120778, '\P{-is_uppercase}', "");
Expect(1, 120778, '\P{^-is_uppercase}', "");
Expect(0, 120779, '\p{-is_uppercase}', "");
Expect(1, 120779, '\p{^-is_uppercase}', "");
Expect(1, 120779, '\P{-is_uppercase}', "");
Expect(0, 120779, '\P{^-is_uppercase}', "");
Error('\p{ /a/Uppercase_letter}');
Error('\P{ /a/Uppercase_letter}');
Expect(1, 120778, '\p{uppercaseletter}', "");
Expect(0, 120778, '\p{^uppercaseletter}', "");
Expect(0, 120778, '\P{uppercaseletter}', "");
Expect(1, 120778, '\P{^uppercaseletter}', "");
Expect(0, 120779, '\p{uppercaseletter}', "");
Expect(1, 120779, '\p{^uppercaseletter}', "");
Expect(1, 120779, '\P{uppercaseletter}', "");
Expect(0, 120779, '\P{^uppercaseletter}', "");
Expect(1, 120778, '\p{ Uppercase_Letter}', "");
Expect(0, 120778, '\p{^ Uppercase_Letter}', "");
Expect(0, 120778, '\P{ Uppercase_Letter}', "");
Expect(1, 120778, '\P{^ Uppercase_Letter}', "");
Expect(0, 120779, '\p{ Uppercase_Letter}', "");
Expect(1, 120779, '\p{^ Uppercase_Letter}', "");
Expect(1, 120779, '\P{ Uppercase_Letter}', "");
Expect(0, 120779, '\P{^ Uppercase_Letter}', "");
Error('\p{/a/	Is_UPPERCASE_letter}');
Error('\P{/a/	Is_UPPERCASE_letter}');
Expect(1, 120778, '\p{isuppercaseletter}', "");
Expect(0, 120778, '\p{^isuppercaseletter}', "");
Expect(0, 120778, '\P{isuppercaseletter}', "");
Expect(1, 120778, '\P{^isuppercaseletter}', "");
Expect(0, 120779, '\p{isuppercaseletter}', "");
Expect(1, 120779, '\p{^isuppercaseletter}', "");
Expect(1, 120779, '\P{isuppercaseletter}', "");
Expect(0, 120779, '\P{^isuppercaseletter}', "");
Expect(1, 120778, '\p{		Is_Uppercase_Letter}', "");
Expect(0, 120778, '\p{^		Is_Uppercase_Letter}', "");
Expect(0, 120778, '\P{		Is_Uppercase_Letter}', "");
Expect(1, 120778, '\P{^		Is_Uppercase_Letter}', "");
Expect(0, 120779, '\p{		Is_Uppercase_Letter}', "");
Expect(1, 120779, '\p{^		Is_Uppercase_Letter}', "");
Expect(1, 120779, '\P{		Is_Uppercase_Letter}', "");
Expect(0, 120779, '\P{^		Is_Uppercase_Letter}', "");
Error('\p{/a/ _Lu}');
Error('\P{/a/ _Lu}');
Expect(1, 120778, '\p{lu}', "");
Expect(0, 120778, '\p{^lu}', "");
Expect(0, 120778, '\P{lu}', "");
Expect(1, 120778, '\P{^lu}', "");
Expect(0, 120779, '\p{lu}', "");
Expect(1, 120779, '\p{^lu}', "");
Expect(1, 120779, '\P{lu}', "");
Expect(0, 120779, '\P{^lu}', "");
Expect(1, 120778, '\p{-	Lu}', "");
Expect(0, 120778, '\p{^-	Lu}', "");
Expect(0, 120778, '\P{-	Lu}', "");
Expect(1, 120778, '\P{^-	Lu}', "");
Expect(0, 120779, '\p{-	Lu}', "");
Expect(1, 120779, '\p{^-	Lu}', "");
Expect(1, 120779, '\P{-	Lu}', "");
Expect(0, 120779, '\P{^-	Lu}', "");
Error('\p{-/a/is_LU}');
Error('\P{-/a/is_LU}');
Expect(1, 120778, '\p{islu}', "");
Expect(0, 120778, '\p{^islu}', "");
Expect(0, 120778, '\P{islu}', "");
Expect(1, 120778, '\P{^islu}', "");
Expect(0, 120779, '\p{islu}', "");
Expect(1, 120779, '\p{^islu}', "");
Expect(1, 120779, '\P{islu}', "");
Expect(0, 120779, '\P{^islu}', "");
Expect(1, 120778, '\p{-	is_Lu}', "");
Expect(0, 120778, '\p{^-	is_Lu}', "");
Expect(0, 120778, '\P{-	is_Lu}', "");
Expect(1, 120778, '\P{^-	is_Lu}', "");
Expect(0, 120779, '\p{-	is_Lu}', "");
Expect(1, 120779, '\p{^-	is_Lu}', "");
Expect(1, 120779, '\P{-	is_Lu}', "");
Expect(0, 120779, '\P{^-	is_Lu}', "");
Error('\p{:=--Vai}');
Error('\P{:=--Vai}');
Expect(1, 42539, '\p{vai}', "");
Expect(0, 42539, '\p{^vai}', "");
Expect(0, 42539, '\P{vai}', "");
Expect(1, 42539, '\P{^vai}', "");
Expect(0, 42540, '\p{vai}', "");
Expect(1, 42540, '\p{^vai}', "");
Expect(1, 42540, '\P{vai}', "");
Expect(0, 42540, '\P{^vai}', "");
Expect(1, 42539, '\p{	VAI}', "");
Expect(0, 42539, '\p{^	VAI}', "");
Expect(0, 42539, '\P{	VAI}', "");
Expect(1, 42539, '\P{^	VAI}', "");
Expect(0, 42540, '\p{	VAI}', "");
Expect(1, 42540, '\p{^	VAI}', "");
Expect(1, 42540, '\P{	VAI}', "");
Expect(0, 42540, '\P{^	VAI}', "");
Error('\p{/a/  Is_vai}');
Error('\P{/a/  Is_vai}');
Expect(1, 42539, '\p{isvai}', "");
Expect(0, 42539, '\p{^isvai}', "");
Expect(0, 42539, '\P{isvai}', "");
Expect(1, 42539, '\P{^isvai}', "");
Expect(0, 42540, '\p{isvai}', "");
Expect(1, 42540, '\p{^isvai}', "");
Expect(1, 42540, '\P{isvai}', "");
Expect(0, 42540, '\P{^isvai}', "");
Expect(1, 42539, '\p{ _is_VAI}', "");
Expect(0, 42539, '\p{^ _is_VAI}', "");
Expect(0, 42539, '\P{ _is_VAI}', "");
Expect(1, 42539, '\P{^ _is_VAI}', "");
Expect(0, 42540, '\p{ _is_VAI}', "");
Expect(1, 42540, '\p{^ _is_VAI}', "");
Expect(1, 42540, '\P{ _is_VAI}', "");
Expect(0, 42540, '\P{^ _is_VAI}', "");
Error('\p{/a/	_Vaii}');
Error('\P{/a/	_Vaii}');
Expect(1, 42539, '\p{vaii}', "");
Expect(0, 42539, '\p{^vaii}', "");
Expect(0, 42539, '\P{vaii}', "");
Expect(1, 42539, '\P{^vaii}', "");
Expect(0, 42540, '\p{vaii}', "");
Expect(1, 42540, '\p{^vaii}', "");
Expect(1, 42540, '\P{vaii}', "");
Expect(0, 42540, '\P{^vaii}', "");
Expect(1, 42539, '\p{ _vaii}', "");
Expect(0, 42539, '\p{^ _vaii}', "");
Expect(0, 42539, '\P{ _vaii}', "");
Expect(1, 42539, '\P{^ _vaii}', "");
Expect(0, 42540, '\p{ _vaii}', "");
Expect(1, 42540, '\p{^ _vaii}', "");
Expect(1, 42540, '\P{ _vaii}', "");
Expect(0, 42540, '\P{^ _vaii}', "");
Error('\p{	/a/is_Vaii}');
Error('\P{	/a/is_Vaii}');
Expect(1, 42539, '\p{isvaii}', "");
Expect(0, 42539, '\p{^isvaii}', "");
Expect(0, 42539, '\P{isvaii}', "");
Expect(1, 42539, '\P{^isvaii}', "");
Expect(0, 42540, '\p{isvaii}', "");
Expect(1, 42540, '\p{^isvaii}', "");
Expect(1, 42540, '\P{isvaii}', "");
Expect(0, 42540, '\P{^isvaii}', "");
Expect(1, 42539, '\p{_is_Vaii}', "");
Expect(0, 42539, '\p{^_is_Vaii}', "");
Expect(0, 42539, '\P{_is_Vaii}', "");
Expect(1, 42539, '\P{^_is_Vaii}', "");
Expect(0, 42540, '\p{_is_Vaii}', "");
Expect(1, 42540, '\p{^_is_Vaii}', "");
Expect(1, 42540, '\P{_is_Vaii}', "");
Expect(0, 42540, '\P{^_is_Vaii}', "");
Error('\p{-/a/Variation_selector}');
Error('\P{-/a/Variation_selector}');
Expect(1, 917999, '\p{variationselector}', "");
Expect(0, 917999, '\p{^variationselector}', "");
Expect(0, 917999, '\P{variationselector}', "");
Expect(1, 917999, '\P{^variationselector}', "");
Expect(0, 918000, '\p{variationselector}', "");
Expect(1, 918000, '\p{^variationselector}', "");
Expect(1, 918000, '\P{variationselector}', "");
Expect(0, 918000, '\P{^variationselector}', "");
Expect(1, 917999, '\p{Variation_SELECTOR}', "");
Expect(0, 917999, '\p{^Variation_SELECTOR}', "");
Expect(0, 917999, '\P{Variation_SELECTOR}', "");
Expect(1, 917999, '\P{^Variation_SELECTOR}', "");
Expect(0, 918000, '\p{Variation_SELECTOR}', "");
Expect(1, 918000, '\p{^Variation_SELECTOR}', "");
Expect(1, 918000, '\P{Variation_SELECTOR}', "");
Expect(0, 918000, '\P{^Variation_SELECTOR}', "");
Error('\p{_:=IS_variation_Selector}');
Error('\P{_:=IS_variation_Selector}');
Expect(1, 917999, '\p{isvariationselector}', "");
Expect(0, 917999, '\p{^isvariationselector}', "");
Expect(0, 917999, '\P{isvariationselector}', "");
Expect(1, 917999, '\P{^isvariationselector}', "");
Expect(0, 918000, '\p{isvariationselector}', "");
Expect(1, 918000, '\p{^isvariationselector}', "");
Expect(1, 918000, '\P{isvariationselector}', "");
Expect(0, 918000, '\P{^isvariationselector}', "");
Expect(1, 917999, '\p{  is_variation_Selector}', "");
Expect(0, 917999, '\p{^  is_variation_Selector}', "");
Expect(0, 917999, '\P{  is_variation_Selector}', "");
Expect(1, 917999, '\P{^  is_variation_Selector}', "");
Expect(0, 918000, '\p{  is_variation_Selector}', "");
Expect(1, 918000, '\p{^  is_variation_Selector}', "");
Expect(1, 918000, '\P{  is_variation_Selector}', "");
Expect(0, 918000, '\P{^  is_variation_Selector}', "");
Error('\p{	 VS:=}');
Error('\P{	 VS:=}');
Expect(1, 917999, '\p{vs}', "");
Expect(0, 917999, '\p{^vs}', "");
Expect(0, 917999, '\P{vs}', "");
Expect(1, 917999, '\P{^vs}', "");
Expect(0, 918000, '\p{vs}', "");
Expect(1, 918000, '\p{^vs}', "");
Expect(1, 918000, '\P{vs}', "");
Expect(0, 918000, '\P{^vs}', "");
Expect(1, 917999, '\p{--VS}', "");
Expect(0, 917999, '\p{^--VS}', "");
Expect(0, 917999, '\P{--VS}', "");
Expect(1, 917999, '\P{^--VS}', "");
Expect(0, 918000, '\p{--VS}', "");
Expect(1, 918000, '\p{^--VS}', "");
Expect(1, 918000, '\P{--VS}', "");
Expect(0, 918000, '\P{^--VS}', "");
Error('\p{	 IS_vs/a/}');
Error('\P{	 IS_vs/a/}');
Expect(1, 917999, '\p{isvs}', "");
Expect(0, 917999, '\p{^isvs}', "");
Expect(0, 917999, '\P{isvs}', "");
Expect(1, 917999, '\P{^isvs}', "");
Expect(0, 918000, '\p{isvs}', "");
Expect(1, 918000, '\p{^isvs}', "");
Expect(1, 918000, '\P{isvs}', "");
Expect(0, 918000, '\P{^isvs}', "");
Expect(1, 917999, '\p{ is_VS}', "");
Expect(0, 917999, '\p{^ is_VS}', "");
Expect(0, 917999, '\P{ is_VS}', "");
Expect(1, 917999, '\P{^ is_VS}', "");
Expect(0, 918000, '\p{ is_VS}', "");
Expect(1, 918000, '\p{^ is_VS}', "");
Expect(1, 918000, '\P{ is_VS}', "");
Expect(0, 918000, '\P{^ is_VS}', "");
Error('\p{/a/Variation_SELECTORS}');
Error('\P{/a/Variation_SELECTORS}');
Expect(1, 65039, '\p{variationselectors}', "");
Expect(0, 65039, '\p{^variationselectors}', "");
Expect(0, 65039, '\P{variationselectors}', "");
Expect(1, 65039, '\P{^variationselectors}', "");
Expect(0, 65040, '\p{variationselectors}', "");
Expect(1, 65040, '\p{^variationselectors}', "");
Expect(1, 65040, '\P{variationselectors}', "");
Expect(0, 65040, '\P{^variationselectors}', "");
Expect(1, 65039, '\p{  variation_SELECTORS}', "");
Expect(0, 65039, '\p{^  variation_SELECTORS}', "");
Expect(0, 65039, '\P{  variation_SELECTORS}', "");
Expect(1, 65039, '\P{^  variation_SELECTORS}', "");
Expect(0, 65040, '\p{  variation_SELECTORS}', "");
Expect(1, 65040, '\p{^  variation_SELECTORS}', "");
Expect(1, 65040, '\P{  variation_SELECTORS}', "");
Expect(0, 65040, '\P{^  variation_SELECTORS}', "");
Error('\p{-:=Is_Variation_Selectors}');
Error('\P{-:=Is_Variation_Selectors}');
Expect(1, 65039, '\p{isvariationselectors}', "");
Expect(0, 65039, '\p{^isvariationselectors}', "");
Expect(0, 65039, '\P{isvariationselectors}', "");
Expect(1, 65039, '\P{^isvariationselectors}', "");
Expect(0, 65040, '\p{isvariationselectors}', "");
Expect(1, 65040, '\p{^isvariationselectors}', "");
Expect(1, 65040, '\P{isvariationselectors}', "");
Expect(0, 65040, '\P{^isvariationselectors}', "");
Expect(1, 65039, '\p{_ Is_VARIATION_selectors}', "");
Expect(0, 65039, '\p{^_ Is_VARIATION_selectors}', "");
Expect(0, 65039, '\P{_ Is_VARIATION_selectors}', "");
Expect(1, 65039, '\P{^_ Is_VARIATION_selectors}', "");
Expect(0, 65040, '\p{_ Is_VARIATION_selectors}', "");
Expect(1, 65040, '\p{^_ Is_VARIATION_selectors}', "");
Expect(1, 65040, '\P{_ Is_VARIATION_selectors}', "");
Expect(0, 65040, '\P{^_ Is_VARIATION_selectors}', "");
Error('\p{__In_Variation_SELECTORS/a/}');
Error('\P{__In_Variation_SELECTORS/a/}');
Expect(1, 65039, '\p{invariationselectors}', "");
Expect(0, 65039, '\p{^invariationselectors}', "");
Expect(0, 65039, '\P{invariationselectors}', "");
Expect(1, 65039, '\P{^invariationselectors}', "");
Expect(0, 65040, '\p{invariationselectors}', "");
Expect(1, 65040, '\p{^invariationselectors}', "");
Expect(1, 65040, '\P{invariationselectors}', "");
Expect(0, 65040, '\P{^invariationselectors}', "");
Expect(1, 65039, '\p{_In_VARIATION_Selectors}', "");
Expect(0, 65039, '\p{^_In_VARIATION_Selectors}', "");
Expect(0, 65039, '\P{_In_VARIATION_Selectors}', "");
Expect(1, 65039, '\P{^_In_VARIATION_Selectors}', "");
Expect(0, 65040, '\p{_In_VARIATION_Selectors}', "");
Expect(1, 65040, '\p{^_In_VARIATION_Selectors}', "");
Expect(1, 65040, '\P{_In_VARIATION_Selectors}', "");
Expect(0, 65040, '\P{^_In_VARIATION_Selectors}', "");
Error('\p{/a/	_In_vs}');
Error('\P{/a/	_In_vs}');
Expect(1, 65039, '\p{invs}', "");
Expect(0, 65039, '\p{^invs}', "");
Expect(0, 65039, '\P{invs}', "");
Expect(1, 65039, '\P{^invs}', "");
Expect(0, 65040, '\p{invs}', "");
Expect(1, 65040, '\p{^invs}', "");
Expect(1, 65040, '\P{invs}', "");
Expect(0, 65040, '\P{^invs}', "");
Expect(1, 65039, '\p{-IN_vs}', "");
Expect(0, 65039, '\p{^-IN_vs}', "");
Expect(0, 65039, '\P{-IN_vs}', "");
Expect(1, 65039, '\P{^-IN_vs}', "");
Expect(0, 65040, '\p{-IN_vs}', "");
Expect(1, 65040, '\p{^-IN_vs}', "");
Expect(1, 65040, '\P{-IN_vs}', "");
Expect(0, 65040, '\P{^-IN_vs}', "");
Error('\p{	:=Variation_selectors_SUPPLEMENT}');
Error('\P{	:=Variation_selectors_SUPPLEMENT}');
Expect(1, 917999, '\p{variationselectorssupplement}', "");
Expect(0, 917999, '\p{^variationselectorssupplement}', "");
Expect(0, 917999, '\P{variationselectorssupplement}', "");
Expect(1, 917999, '\P{^variationselectorssupplement}', "");
Expect(0, 918000, '\p{variationselectorssupplement}', "");
Expect(1, 918000, '\p{^variationselectorssupplement}', "");
Expect(1, 918000, '\P{variationselectorssupplement}', "");
Expect(0, 918000, '\P{^variationselectorssupplement}', "");
Expect(1, 917999, '\p{	_Variation_SELECTORS_supplement}', "");
Expect(0, 917999, '\p{^	_Variation_SELECTORS_supplement}', "");
Expect(0, 917999, '\P{	_Variation_SELECTORS_supplement}', "");
Expect(1, 917999, '\P{^	_Variation_SELECTORS_supplement}', "");
Expect(0, 918000, '\p{	_Variation_SELECTORS_supplement}', "");
Expect(1, 918000, '\p{^	_Variation_SELECTORS_supplement}', "");
Expect(1, 918000, '\P{	_Variation_SELECTORS_supplement}', "");
Expect(0, 918000, '\P{^	_Variation_SELECTORS_supplement}', "");
Error('\p{ -Is_Variation_SELECTORS_Supplement:=}');
Error('\P{ -Is_Variation_SELECTORS_Supplement:=}');
Expect(1, 917999, '\p{isvariationselectorssupplement}', "");
Expect(0, 917999, '\p{^isvariationselectorssupplement}', "");
Expect(0, 917999, '\P{isvariationselectorssupplement}', "");
Expect(1, 917999, '\P{^isvariationselectorssupplement}', "");
Expect(0, 918000, '\p{isvariationselectorssupplement}', "");
Expect(1, 918000, '\p{^isvariationselectorssupplement}', "");
Expect(1, 918000, '\P{isvariationselectorssupplement}', "");
Expect(0, 918000, '\P{^isvariationselectorssupplement}', "");
Expect(1, 917999, '\p{- Is_Variation_Selectors_SUPPLEMENT}', "");
Expect(0, 917999, '\p{^- Is_Variation_Selectors_SUPPLEMENT}', "");
Expect(0, 917999, '\P{- Is_Variation_Selectors_SUPPLEMENT}', "");
Expect(1, 917999, '\P{^- Is_Variation_Selectors_SUPPLEMENT}', "");
Expect(0, 918000, '\p{- Is_Variation_Selectors_SUPPLEMENT}', "");
Expect(1, 918000, '\p{^- Is_Variation_Selectors_SUPPLEMENT}', "");
Expect(1, 918000, '\P{- Is_Variation_Selectors_SUPPLEMENT}', "");
Expect(0, 918000, '\P{^- Is_Variation_Selectors_SUPPLEMENT}', "");
Error('\p{ :=In_VARIATION_selectors_SUPPLEMENT}');
Error('\P{ :=In_VARIATION_selectors_SUPPLEMENT}');
Expect(1, 917999, '\p{invariationselectorssupplement}', "");
Expect(0, 917999, '\p{^invariationselectorssupplement}', "");
Expect(0, 917999, '\P{invariationselectorssupplement}', "");
Expect(1, 917999, '\P{^invariationselectorssupplement}', "");
Expect(0, 918000, '\p{invariationselectorssupplement}', "");
Expect(1, 918000, '\p{^invariationselectorssupplement}', "");
Expect(1, 918000, '\P{invariationselectorssupplement}', "");
Expect(0, 918000, '\P{^invariationselectorssupplement}', "");
Expect(1, 917999, '\p{	-in_VARIATION_Selectors_Supplement}', "");
Expect(0, 917999, '\p{^	-in_VARIATION_Selectors_Supplement}', "");
Expect(0, 917999, '\P{	-in_VARIATION_Selectors_Supplement}', "");
Expect(1, 917999, '\P{^	-in_VARIATION_Selectors_Supplement}', "");
Expect(0, 918000, '\p{	-in_VARIATION_Selectors_Supplement}', "");
Expect(1, 918000, '\p{^	-in_VARIATION_Selectors_Supplement}', "");
Expect(1, 918000, '\P{	-in_VARIATION_Selectors_Supplement}', "");
Expect(0, 918000, '\P{^	-in_VARIATION_Selectors_Supplement}', "");
Error('\p{:=	VS_sup}');
Error('\P{:=	VS_sup}');
Expect(1, 917999, '\p{vssup}', "");
Expect(0, 917999, '\p{^vssup}', "");
Expect(0, 917999, '\P{vssup}', "");
Expect(1, 917999, '\P{^vssup}', "");
Expect(0, 918000, '\p{vssup}', "");
Expect(1, 918000, '\p{^vssup}', "");
Expect(1, 918000, '\P{vssup}', "");
Expect(0, 918000, '\P{^vssup}', "");
Expect(1, 917999, '\p{	_VS_sup}', "");
Expect(0, 917999, '\p{^	_VS_sup}', "");
Expect(0, 917999, '\P{	_VS_sup}', "");
Expect(1, 917999, '\P{^	_VS_sup}', "");
Expect(0, 918000, '\p{	_VS_sup}', "");
Expect(1, 918000, '\p{^	_VS_sup}', "");
Expect(1, 918000, '\P{	_VS_sup}', "");
Expect(0, 918000, '\P{^	_VS_sup}', "");
Error('\p{__IS_VS_Sup/a/}');
Error('\P{__IS_VS_Sup/a/}');
Expect(1, 917999, '\p{isvssup}', "");
Expect(0, 917999, '\p{^isvssup}', "");
Expect(0, 917999, '\P{isvssup}', "");
Expect(1, 917999, '\P{^isvssup}', "");
Expect(0, 918000, '\p{isvssup}', "");
Expect(1, 918000, '\p{^isvssup}', "");
Expect(1, 918000, '\P{isvssup}', "");
Expect(0, 918000, '\P{^isvssup}', "");
Expect(1, 917999, '\p{ Is_VS_Sup}', "");
Expect(0, 917999, '\p{^ Is_VS_Sup}', "");
Expect(0, 917999, '\P{ Is_VS_Sup}', "");
Expect(1, 917999, '\P{^ Is_VS_Sup}', "");
Expect(0, 918000, '\p{ Is_VS_Sup}', "");
Expect(1, 918000, '\p{^ Is_VS_Sup}', "");
Expect(1, 918000, '\P{ Is_VS_Sup}', "");
Expect(0, 918000, '\P{^ Is_VS_Sup}', "");
Error('\p{/a/-In_VS_sup}');
Error('\P{/a/-In_VS_sup}');
Expect(1, 917999, '\p{invssup}', "");
Expect(0, 917999, '\p{^invssup}', "");
Expect(0, 917999, '\P{invssup}', "");
Expect(1, 917999, '\P{^invssup}', "");
Expect(0, 918000, '\p{invssup}', "");
Expect(1, 918000, '\p{^invssup}', "");
Expect(1, 918000, '\P{invssup}', "");
Expect(0, 918000, '\P{^invssup}', "");
Expect(1, 917999, '\p{	_In_VS_Sup}', "");
Expect(0, 917999, '\p{^	_In_VS_Sup}', "");
Expect(0, 917999, '\P{	_In_VS_Sup}', "");
Expect(1, 917999, '\P{^	_In_VS_Sup}', "");
Expect(0, 918000, '\p{	_In_VS_Sup}', "");
Expect(1, 918000, '\p{^	_In_VS_Sup}', "");
Expect(1, 918000, '\P{	_In_VS_Sup}', "");
Expect(0, 918000, '\P{^	_In_VS_Sup}', "");
Error('\p{:=-Vedic_Extensions}');
Error('\P{:=-Vedic_Extensions}');
Expect(1, 7423, '\p{vedicextensions}', "");
Expect(0, 7423, '\p{^vedicextensions}', "");
Expect(0, 7423, '\P{vedicextensions}', "");
Expect(1, 7423, '\P{^vedicextensions}', "");
Expect(0, 7424, '\p{vedicextensions}', "");
Expect(1, 7424, '\p{^vedicextensions}', "");
Expect(1, 7424, '\P{vedicextensions}', "");
Expect(0, 7424, '\P{^vedicextensions}', "");
Expect(1, 7423, '\p{	Vedic_Extensions}', "");
Expect(0, 7423, '\p{^	Vedic_Extensions}', "");
Expect(0, 7423, '\P{	Vedic_Extensions}', "");
Expect(1, 7423, '\P{^	Vedic_Extensions}', "");
Expect(0, 7424, '\p{	Vedic_Extensions}', "");
Expect(1, 7424, '\p{^	Vedic_Extensions}', "");
Expect(1, 7424, '\P{	Vedic_Extensions}', "");
Expect(0, 7424, '\P{^	Vedic_Extensions}', "");
Error('\p{-/a/is_vedic_Extensions}');
Error('\P{-/a/is_vedic_Extensions}');
Expect(1, 7423, '\p{isvedicextensions}', "");
Expect(0, 7423, '\p{^isvedicextensions}', "");
Expect(0, 7423, '\P{isvedicextensions}', "");
Expect(1, 7423, '\P{^isvedicextensions}', "");
Expect(0, 7424, '\p{isvedicextensions}', "");
Expect(1, 7424, '\p{^isvedicextensions}', "");
Expect(1, 7424, '\P{isvedicextensions}', "");
Expect(0, 7424, '\P{^isvedicextensions}', "");
Expect(1, 7423, '\p{_ Is_Vedic_extensions}', "");
Expect(0, 7423, '\p{^_ Is_Vedic_extensions}', "");
Expect(0, 7423, '\P{_ Is_Vedic_extensions}', "");
Expect(1, 7423, '\P{^_ Is_Vedic_extensions}', "");
Expect(0, 7424, '\p{_ Is_Vedic_extensions}', "");
Expect(1, 7424, '\p{^_ Is_Vedic_extensions}', "");
Expect(1, 7424, '\P{_ Is_Vedic_extensions}', "");
Expect(0, 7424, '\P{^_ Is_Vedic_extensions}', "");
Error('\p{:=In_vedic_Extensions}');
Error('\P{:=In_vedic_Extensions}');
Expect(1, 7423, '\p{invedicextensions}', "");
Expect(0, 7423, '\p{^invedicextensions}', "");
Expect(0, 7423, '\P{invedicextensions}', "");
Expect(1, 7423, '\P{^invedicextensions}', "");
Expect(0, 7424, '\p{invedicextensions}', "");
Expect(1, 7424, '\p{^invedicextensions}', "");
Expect(1, 7424, '\P{invedicextensions}', "");
Expect(0, 7424, '\P{^invedicextensions}', "");
Expect(1, 7423, '\p{--In_Vedic_EXTENSIONS}', "");
Expect(0, 7423, '\p{^--In_Vedic_EXTENSIONS}', "");
Expect(0, 7423, '\P{--In_Vedic_EXTENSIONS}', "");
Expect(1, 7423, '\P{^--In_Vedic_EXTENSIONS}', "");
Expect(0, 7424, '\p{--In_Vedic_EXTENSIONS}', "");
Expect(1, 7424, '\p{^--In_Vedic_EXTENSIONS}', "");
Expect(1, 7424, '\P{--In_Vedic_EXTENSIONS}', "");
Expect(0, 7424, '\P{^--In_Vedic_EXTENSIONS}', "");
Error('\p{:=	Vedic_Ext}');
Error('\P{:=	Vedic_Ext}');
Expect(1, 7423, '\p{vedicext}', "");
Expect(0, 7423, '\p{^vedicext}', "");
Expect(0, 7423, '\P{vedicext}', "");
Expect(1, 7423, '\P{^vedicext}', "");
Expect(0, 7424, '\p{vedicext}', "");
Expect(1, 7424, '\p{^vedicext}', "");
Expect(1, 7424, '\P{vedicext}', "");
Expect(0, 7424, '\P{^vedicext}', "");
Expect(1, 7423, '\p{	 VEDIC_Ext}', "");
Expect(0, 7423, '\p{^	 VEDIC_Ext}', "");
Expect(0, 7423, '\P{	 VEDIC_Ext}', "");
Expect(1, 7423, '\P{^	 VEDIC_Ext}', "");
Expect(0, 7424, '\p{	 VEDIC_Ext}', "");
Expect(1, 7424, '\p{^	 VEDIC_Ext}', "");
Expect(1, 7424, '\P{	 VEDIC_Ext}', "");
Expect(0, 7424, '\P{^	 VEDIC_Ext}', "");
Error('\p{-/a/is_Vedic_EXT}');
Error('\P{-/a/is_Vedic_EXT}');
Expect(1, 7423, '\p{isvedicext}', "");
Expect(0, 7423, '\p{^isvedicext}', "");
Expect(0, 7423, '\P{isvedicext}', "");
Expect(1, 7423, '\P{^isvedicext}', "");
Expect(0, 7424, '\p{isvedicext}', "");
Expect(1, 7424, '\p{^isvedicext}', "");
Expect(1, 7424, '\P{isvedicext}', "");
Expect(0, 7424, '\P{^isvedicext}', "");
Expect(1, 7423, '\p{	 Is_Vedic_EXT}', "");
Expect(0, 7423, '\p{^	 Is_Vedic_EXT}', "");
Expect(0, 7423, '\P{	 Is_Vedic_EXT}', "");
Expect(1, 7423, '\P{^	 Is_Vedic_EXT}', "");
Expect(0, 7424, '\p{	 Is_Vedic_EXT}', "");
Expect(1, 7424, '\p{^	 Is_Vedic_EXT}', "");
Expect(1, 7424, '\P{	 Is_Vedic_EXT}', "");
Expect(0, 7424, '\P{^	 Is_Vedic_EXT}', "");
Error('\p{__in_Vedic_ext:=}');
Error('\P{__in_Vedic_ext:=}');
Expect(1, 7423, '\p{invedicext}', "");
Expect(0, 7423, '\p{^invedicext}', "");
Expect(0, 7423, '\P{invedicext}', "");
Expect(1, 7423, '\P{^invedicext}', "");
Expect(0, 7424, '\p{invedicext}', "");
Expect(1, 7424, '\p{^invedicext}', "");
Expect(1, 7424, '\P{invedicext}', "");
Expect(0, 7424, '\P{^invedicext}', "");
Expect(1, 7423, '\p{ -In_Vedic_EXT}', "");
Expect(0, 7423, '\p{^ -In_Vedic_EXT}', "");
Expect(0, 7423, '\P{ -In_Vedic_EXT}', "");
Expect(1, 7423, '\P{^ -In_Vedic_EXT}', "");
Expect(0, 7424, '\p{ -In_Vedic_EXT}', "");
Expect(1, 7424, '\p{^ -In_Vedic_EXT}', "");
Expect(1, 7424, '\P{ -In_Vedic_EXT}', "");
Expect(0, 7424, '\P{^ -In_Vedic_EXT}', "");
Error('\p{_:=VERTSPACE}');
Error('\P{_:=VERTSPACE}');
Expect(1, 8233, '\p{vertspace}', "");
Expect(0, 8233, '\p{^vertspace}', "");
Expect(0, 8233, '\P{vertspace}', "");
Expect(1, 8233, '\P{^vertspace}', "");
Expect(0, 8234, '\p{vertspace}', "");
Expect(1, 8234, '\p{^vertspace}', "");
Expect(1, 8234, '\P{vertspace}', "");
Expect(0, 8234, '\P{^vertspace}', "");
Expect(1, 8233, '\p{_	vertspace}', "");
Expect(0, 8233, '\p{^_	vertspace}', "");
Expect(0, 8233, '\P{_	vertspace}', "");
Expect(1, 8233, '\P{^_	vertspace}', "");
Expect(0, 8234, '\p{_	vertspace}', "");
Expect(1, 8234, '\p{^_	vertspace}', "");
Expect(1, 8234, '\P{_	vertspace}', "");
Expect(0, 8234, '\P{^_	vertspace}', "");
Error('\p{-_Is_VERTSPACE:=}');
Error('\P{-_Is_VERTSPACE:=}');
Expect(1, 8233, '\p{isvertspace}', "");
Expect(0, 8233, '\p{^isvertspace}', "");
Expect(0, 8233, '\P{isvertspace}', "");
Expect(1, 8233, '\P{^isvertspace}', "");
Expect(0, 8234, '\p{isvertspace}', "");
Expect(1, 8234, '\p{^isvertspace}', "");
Expect(1, 8234, '\P{isvertspace}', "");
Expect(0, 8234, '\P{^isvertspace}', "");
Expect(1, 8233, '\p{ Is_VERTSPACE}', "");
Expect(0, 8233, '\p{^ Is_VERTSPACE}', "");
Expect(0, 8233, '\P{ Is_VERTSPACE}', "");
Expect(1, 8233, '\P{^ Is_VERTSPACE}', "");
Expect(0, 8234, '\p{ Is_VERTSPACE}', "");
Expect(1, 8234, '\p{^ Is_VERTSPACE}', "");
Expect(1, 8234, '\P{ Is_VERTSPACE}', "");
Expect(0, 8234, '\P{^ Is_VERTSPACE}', "");
Error('\p{VERTICAL_FORMS:=}');
Error('\P{VERTICAL_FORMS:=}');
Expect(1, 65055, '\p{verticalforms}', "");
Expect(0, 65055, '\p{^verticalforms}', "");
Expect(0, 65055, '\P{verticalforms}', "");
Expect(1, 65055, '\P{^verticalforms}', "");
Expect(0, 65056, '\p{verticalforms}', "");
Expect(1, 65056, '\p{^verticalforms}', "");
Expect(1, 65056, '\P{verticalforms}', "");
Expect(0, 65056, '\P{^verticalforms}', "");
Expect(1, 65055, '\p{	_Vertical_Forms}', "");
Expect(0, 65055, '\p{^	_Vertical_Forms}', "");
Expect(0, 65055, '\P{	_Vertical_Forms}', "");
Expect(1, 65055, '\P{^	_Vertical_Forms}', "");
Expect(0, 65056, '\p{	_Vertical_Forms}', "");
Expect(1, 65056, '\p{^	_Vertical_Forms}', "");
Expect(1, 65056, '\P{	_Vertical_Forms}', "");
Expect(0, 65056, '\P{^	_Vertical_Forms}', "");
Error('\p{/a/ _Is_vertical_FORMS}');
Error('\P{/a/ _Is_vertical_FORMS}');
Expect(1, 65055, '\p{isverticalforms}', "");
Expect(0, 65055, '\p{^isverticalforms}', "");
Expect(0, 65055, '\P{isverticalforms}', "");
Expect(1, 65055, '\P{^isverticalforms}', "");
Expect(0, 65056, '\p{isverticalforms}', "");
Expect(1, 65056, '\p{^isverticalforms}', "");
Expect(1, 65056, '\P{isverticalforms}', "");
Expect(0, 65056, '\P{^isverticalforms}', "");
Expect(1, 65055, '\p{-IS_VERTICAL_FORMS}', "");
Expect(0, 65055, '\p{^-IS_VERTICAL_FORMS}', "");
Expect(0, 65055, '\P{-IS_VERTICAL_FORMS}', "");
Expect(1, 65055, '\P{^-IS_VERTICAL_FORMS}', "");
Expect(0, 65056, '\p{-IS_VERTICAL_FORMS}', "");
Expect(1, 65056, '\p{^-IS_VERTICAL_FORMS}', "");
Expect(1, 65056, '\P{-IS_VERTICAL_FORMS}', "");
Expect(0, 65056, '\P{^-IS_VERTICAL_FORMS}', "");
Error('\p{:= _In_Vertical_Forms}');
Error('\P{:= _In_Vertical_Forms}');
Expect(1, 65055, '\p{inverticalforms}', "");
Expect(0, 65055, '\p{^inverticalforms}', "");
Expect(0, 65055, '\P{inverticalforms}', "");
Expect(1, 65055, '\P{^inverticalforms}', "");
Expect(0, 65056, '\p{inverticalforms}', "");
Expect(1, 65056, '\p{^inverticalforms}', "");
Expect(1, 65056, '\P{inverticalforms}', "");
Expect(0, 65056, '\P{^inverticalforms}', "");
Expect(1, 65055, '\p{--In_Vertical_FORMS}', "");
Expect(0, 65055, '\p{^--In_Vertical_FORMS}', "");
Expect(0, 65055, '\P{--In_Vertical_FORMS}', "");
Expect(1, 65055, '\P{^--In_Vertical_FORMS}', "");
Expect(0, 65056, '\p{--In_Vertical_FORMS}', "");
Expect(1, 65056, '\p{^--In_Vertical_FORMS}', "");
Expect(1, 65056, '\P{--In_Vertical_FORMS}', "");
Expect(0, 65056, '\P{^--In_Vertical_FORMS}', "");
Error('\p{_/a/White_Space}');
Error('\P{_/a/White_Space}');
Expect(1, 12288, '\p{whitespace}', "");
Expect(0, 12288, '\p{^whitespace}', "");
Expect(0, 12288, '\P{whitespace}', "");
Expect(1, 12288, '\P{^whitespace}', "");
Expect(0, 12289, '\p{whitespace}', "");
Expect(1, 12289, '\p{^whitespace}', "");
Expect(1, 12289, '\P{whitespace}', "");
Expect(0, 12289, '\P{^whitespace}', "");
Expect(1, 12288, '\p{	white_SPACE}', "");
Expect(0, 12288, '\p{^	white_SPACE}', "");
Expect(0, 12288, '\P{	white_SPACE}', "");
Expect(1, 12288, '\P{^	white_SPACE}', "");
Expect(0, 12289, '\p{	white_SPACE}', "");
Expect(1, 12289, '\p{^	white_SPACE}', "");
Expect(1, 12289, '\P{	white_SPACE}', "");
Expect(0, 12289, '\P{^	white_SPACE}', "");
Error('\p{ 	Is_white_space:=}');
Error('\P{ 	Is_white_space:=}');
Expect(1, 12288, '\p{iswhitespace}', "");
Expect(0, 12288, '\p{^iswhitespace}', "");
Expect(0, 12288, '\P{iswhitespace}', "");
Expect(1, 12288, '\P{^iswhitespace}', "");
Expect(0, 12289, '\p{iswhitespace}', "");
Expect(1, 12289, '\p{^iswhitespace}', "");
Expect(1, 12289, '\P{iswhitespace}', "");
Expect(0, 12289, '\P{^iswhitespace}', "");
Expect(1, 12288, '\p{--Is_WHITE_Space}', "");
Expect(0, 12288, '\p{^--Is_WHITE_Space}', "");
Expect(0, 12288, '\P{--Is_WHITE_Space}', "");
Expect(1, 12288, '\P{^--Is_WHITE_Space}', "");
Expect(0, 12289, '\p{--Is_WHITE_Space}', "");
Expect(1, 12289, '\p{^--Is_WHITE_Space}', "");
Expect(1, 12289, '\P{--Is_WHITE_Space}', "");
Expect(0, 12289, '\P{^--Is_WHITE_Space}', "");
Error('\p{/a/ wspace}');
Error('\P{/a/ wspace}');
Expect(1, 12288, '\p{wspace}', "");
Expect(0, 12288, '\p{^wspace}', "");
Expect(0, 12288, '\P{wspace}', "");
Expect(1, 12288, '\P{^wspace}', "");
Expect(0, 12289, '\p{wspace}', "");
Expect(1, 12289, '\p{^wspace}', "");
Expect(1, 12289, '\P{wspace}', "");
Expect(0, 12289, '\P{^wspace}', "");
Expect(1, 12288, '\p{_WSpace}', "");
Expect(0, 12288, '\p{^_WSpace}', "");
Expect(0, 12288, '\P{_WSpace}', "");
Expect(1, 12288, '\P{^_WSpace}', "");
Expect(0, 12289, '\p{_WSpace}', "");
Expect(1, 12289, '\p{^_WSpace}', "");
Expect(1, 12289, '\P{_WSpace}', "");
Expect(0, 12289, '\P{^_WSpace}', "");
Error('\p{_ Is_WSPACE/a/}');
Error('\P{_ Is_WSPACE/a/}');
Expect(1, 12288, '\p{iswspace}', "");
Expect(0, 12288, '\p{^iswspace}', "");
Expect(0, 12288, '\P{iswspace}', "");
Expect(1, 12288, '\P{^iswspace}', "");
Expect(0, 12289, '\p{iswspace}', "");
Expect(1, 12289, '\p{^iswspace}', "");
Expect(1, 12289, '\P{iswspace}', "");
Expect(0, 12289, '\P{^iswspace}', "");
Expect(1, 12288, '\p{_IS_WSpace}', "");
Expect(0, 12288, '\p{^_IS_WSpace}', "");
Expect(0, 12288, '\P{_IS_WSpace}', "");
Expect(1, 12288, '\P{^_IS_WSpace}', "");
Expect(0, 12289, '\p{_IS_WSpace}', "");
Expect(1, 12289, '\p{^_IS_WSpace}', "");
Expect(1, 12289, '\P{_IS_WSpace}', "");
Expect(0, 12289, '\P{^_IS_WSpace}', "");
Error('\p{-:=Word}');
Error('\P{-:=Word}');
Expect(1, 917999, '\p{word}', "");
Expect(0, 917999, '\p{^word}', "");
Expect(0, 917999, '\P{word}', "");
Expect(1, 917999, '\P{^word}', "");
Expect(0, 918000, '\p{word}', "");
Expect(1, 918000, '\p{^word}', "");
Expect(1, 918000, '\P{word}', "");
Expect(0, 918000, '\P{^word}', "");
Error('\p{/a/ xposixword}');
Error('\P{/a/ xposixword}');
Expect(1, 917999, '\p{xposixword}', "");
Expect(0, 917999, '\p{^xposixword}', "");
Expect(0, 917999, '\P{xposixword}', "");
Expect(1, 917999, '\P{^xposixword}', "");
Expect(0, 918000, '\p{xposixword}', "");
Expect(1, 918000, '\p{^xposixword}', "");
Expect(1, 918000, '\P{xposixword}', "");
Expect(0, 918000, '\P{^xposixword}', "");
Expect(1, 917999, '\p{ XPOSIXWORD}', "");
Expect(0, 917999, '\p{^ XPOSIXWORD}', "");
Expect(0, 917999, '\P{ XPOSIXWORD}', "");
Expect(1, 917999, '\P{^ XPOSIXWORD}', "");
Expect(0, 918000, '\p{ XPOSIXWORD}', "");
Expect(1, 918000, '\p{^ XPOSIXWORD}', "");
Expect(1, 918000, '\P{ XPOSIXWORD}', "");
Expect(0, 918000, '\P{^ XPOSIXWORD}', "");
Error('\p{_Is_WORD/a/}');
Error('\P{_Is_WORD/a/}');
Expect(1, 917999, '\p{isword}', "");
Expect(0, 917999, '\p{^isword}', "");
Expect(0, 917999, '\P{isword}', "");
Expect(1, 917999, '\P{^isword}', "");
Expect(0, 918000, '\p{isword}', "");
Expect(1, 918000, '\p{^isword}', "");
Expect(1, 918000, '\P{isword}', "");
Expect(0, 918000, '\P{^isword}', "");
Expect(1, 917999, '\p{ 	Is_word}', "");
Expect(0, 917999, '\p{^ 	Is_word}', "");
Expect(0, 917999, '\P{ 	Is_word}', "");
Expect(1, 917999, '\P{^ 	Is_word}', "");
Expect(0, 918000, '\p{ 	Is_word}', "");
Expect(1, 918000, '\p{^ 	Is_word}', "");
Expect(1, 918000, '\P{ 	Is_word}', "");
Expect(0, 918000, '\P{^ 	Is_word}', "");
Error('\p{:=		IS_XPOSIXWORD}');
Error('\P{:=		IS_XPOSIXWORD}');
Expect(1, 917999, '\p{isxposixword}', "");
Expect(0, 917999, '\p{^isxposixword}', "");
Expect(0, 917999, '\P{isxposixword}', "");
Expect(1, 917999, '\P{^isxposixword}', "");
Expect(0, 918000, '\p{isxposixword}', "");
Expect(1, 918000, '\p{^isxposixword}', "");
Expect(1, 918000, '\P{isxposixword}', "");
Expect(0, 918000, '\P{^isxposixword}', "");
Expect(1, 917999, '\p{ _Is_XPosixWord}', "");
Expect(0, 917999, '\p{^ _Is_XPosixWord}', "");
Expect(0, 917999, '\P{ _Is_XPosixWord}', "");
Expect(1, 917999, '\P{^ _Is_XPosixWord}', "");
Expect(0, 918000, '\p{ _Is_XPosixWord}', "");
Expect(1, 918000, '\p{^ _Is_XPosixWord}', "");
Expect(1, 918000, '\P{ _Is_XPosixWord}', "");
Expect(0, 918000, '\P{^ _Is_XPosixWord}', "");
Error('\p{/a/XDigit}');
Error('\P{/a/XDigit}');
Expect(1, 65350, '\p{xdigit}', "");
Expect(0, 65350, '\p{^xdigit}', "");
Expect(0, 65350, '\P{xdigit}', "");
Expect(1, 65350, '\P{^xdigit}', "");
Expect(0, 65351, '\p{xdigit}', "");
Expect(1, 65351, '\p{^xdigit}', "");
Expect(1, 65351, '\P{xdigit}', "");
Expect(0, 65351, '\P{^xdigit}', "");
Expect(1, 65350, '\p{--xdigit}', "");
Expect(0, 65350, '\p{^--xdigit}', "");
Expect(0, 65350, '\P{--xdigit}', "");
Expect(1, 65350, '\P{^--xdigit}', "");
Expect(0, 65351, '\p{--xdigit}', "");
Expect(1, 65351, '\p{^--xdigit}', "");
Expect(1, 65351, '\P{--xdigit}', "");
Expect(0, 65351, '\P{^--xdigit}', "");
Error('\p{ :=XPosixXDigit}');
Error('\P{ :=XPosixXDigit}');
Expect(1, 65350, '\p{xposixxdigit}', "");
Expect(0, 65350, '\p{^xposixxdigit}', "");
Expect(0, 65350, '\P{xposixxdigit}', "");
Expect(1, 65350, '\P{^xposixxdigit}', "");
Expect(0, 65351, '\p{xposixxdigit}', "");
Expect(1, 65351, '\p{^xposixxdigit}', "");
Expect(1, 65351, '\P{xposixxdigit}', "");
Expect(0, 65351, '\P{^xposixxdigit}', "");
Expect(1, 65350, '\p{-XPOSIXXDIGIT}', "");
Expect(0, 65350, '\p{^-XPOSIXXDIGIT}', "");
Expect(0, 65350, '\P{-XPOSIXXDIGIT}', "");
Expect(1, 65350, '\P{^-XPOSIXXDIGIT}', "");
Expect(0, 65351, '\p{-XPOSIXXDIGIT}', "");
Expect(1, 65351, '\p{^-XPOSIXXDIGIT}', "");
Expect(1, 65351, '\P{-XPOSIXXDIGIT}', "");
Expect(0, 65351, '\P{^-XPOSIXXDIGIT}', "");
Error('\p{:=Is_XDigit}');
Error('\P{:=Is_XDigit}');
Expect(1, 65350, '\p{isxdigit}', "");
Expect(0, 65350, '\p{^isxdigit}', "");
Expect(0, 65350, '\P{isxdigit}', "");
Expect(1, 65350, '\P{^isxdigit}', "");
Expect(0, 65351, '\p{isxdigit}', "");
Expect(1, 65351, '\p{^isxdigit}', "");
Expect(1, 65351, '\P{isxdigit}', "");
Expect(0, 65351, '\P{^isxdigit}', "");
Expect(1, 65350, '\p{	-Is_XDIGIT}', "");
Expect(0, 65350, '\p{^	-Is_XDIGIT}', "");
Expect(0, 65350, '\P{	-Is_XDIGIT}', "");
Expect(1, 65350, '\P{^	-Is_XDIGIT}', "");
Expect(0, 65351, '\p{	-Is_XDIGIT}', "");
Expect(1, 65351, '\p{^	-Is_XDIGIT}', "");
Expect(1, 65351, '\P{	-Is_XDIGIT}', "");
Expect(0, 65351, '\P{^	-Is_XDIGIT}', "");
Error('\p{	/a/IS_XPosixXDigit}');
Error('\P{	/a/IS_XPosixXDigit}');
Expect(1, 65350, '\p{isxposixxdigit}', "");
Expect(0, 65350, '\p{^isxposixxdigit}', "");
Expect(0, 65350, '\P{isxposixxdigit}', "");
Expect(1, 65350, '\P{^isxposixxdigit}', "");
Expect(0, 65351, '\p{isxposixxdigit}', "");
Expect(1, 65351, '\p{^isxposixxdigit}', "");
Expect(1, 65351, '\P{isxposixxdigit}', "");
Expect(0, 65351, '\P{^isxposixxdigit}', "");
Expect(1, 65350, '\p{ _is_XPosixXDigit}', "");
Expect(0, 65350, '\p{^ _is_XPosixXDigit}', "");
Expect(0, 65350, '\P{ _is_XPosixXDigit}', "");
Expect(1, 65350, '\P{^ _is_XPosixXDigit}', "");
Expect(0, 65351, '\p{ _is_XPosixXDigit}', "");
Expect(1, 65351, '\p{^ _is_XPosixXDigit}', "");
Expect(1, 65351, '\P{ _is_XPosixXDigit}', "");
Expect(0, 65351, '\P{^ _is_XPosixXDigit}', "");
Error('\p{/a/	_HEX_digit}');
Error('\P{/a/	_HEX_digit}');
Expect(1, 65350, '\p{hexdigit}', "");
Expect(0, 65350, '\p{^hexdigit}', "");
Expect(0, 65350, '\P{hexdigit}', "");
Expect(1, 65350, '\P{^hexdigit}', "");
Expect(0, 65351, '\p{hexdigit}', "");
Expect(1, 65351, '\p{^hexdigit}', "");
Expect(1, 65351, '\P{hexdigit}', "");
Expect(0, 65351, '\P{^hexdigit}', "");
Expect(1, 65350, '\p{		Hex_Digit}', "");
Expect(0, 65350, '\p{^		Hex_Digit}', "");
Expect(0, 65350, '\P{		Hex_Digit}', "");
Expect(1, 65350, '\P{^		Hex_Digit}', "");
Expect(0, 65351, '\p{		Hex_Digit}', "");
Expect(1, 65351, '\p{^		Hex_Digit}', "");
Expect(1, 65351, '\P{		Hex_Digit}', "");
Expect(0, 65351, '\P{^		Hex_Digit}', "");
Error('\p{/a/	 Is_HEX_DIGIT}');
Error('\P{/a/	 Is_HEX_DIGIT}');
Expect(1, 65350, '\p{ishexdigit}', "");
Expect(0, 65350, '\p{^ishexdigit}', "");
Expect(0, 65350, '\P{ishexdigit}', "");
Expect(1, 65350, '\P{^ishexdigit}', "");
Expect(0, 65351, '\p{ishexdigit}', "");
Expect(1, 65351, '\p{^ishexdigit}', "");
Expect(1, 65351, '\P{ishexdigit}', "");
Expect(0, 65351, '\P{^ishexdigit}', "");
Expect(1, 65350, '\p{ -IS_hex_Digit}', "");
Expect(0, 65350, '\p{^ -IS_hex_Digit}', "");
Expect(0, 65350, '\P{ -IS_hex_Digit}', "");
Expect(1, 65350, '\P{^ -IS_hex_Digit}', "");
Expect(0, 65351, '\p{ -IS_hex_Digit}', "");
Expect(1, 65351, '\p{^ -IS_hex_Digit}', "");
Expect(1, 65351, '\P{ -IS_hex_Digit}', "");
Expect(0, 65351, '\P{^ -IS_hex_Digit}', "");
Error('\p{/a/		hex}');
Error('\P{/a/		hex}');
Expect(1, 65350, '\p{hex}', "");
Expect(0, 65350, '\p{^hex}', "");
Expect(0, 65350, '\P{hex}', "");
Expect(1, 65350, '\P{^hex}', "");
Expect(0, 65351, '\p{hex}', "");
Expect(1, 65351, '\p{^hex}', "");
Expect(1, 65351, '\P{hex}', "");
Expect(0, 65351, '\P{^hex}', "");
Expect(1, 65350, '\p{ 	Hex}', "");
Expect(0, 65350, '\p{^ 	Hex}', "");
Expect(0, 65350, '\P{ 	Hex}', "");
Expect(1, 65350, '\P{^ 	Hex}', "");
Expect(0, 65351, '\p{ 	Hex}', "");
Expect(1, 65351, '\p{^ 	Hex}', "");
Expect(1, 65351, '\P{ 	Hex}', "");
Expect(0, 65351, '\P{^ 	Hex}', "");
Error('\p{_ IS_HEX/a/}');
Error('\P{_ IS_HEX/a/}');
Expect(1, 65350, '\p{ishex}', "");
Expect(0, 65350, '\p{^ishex}', "");
Expect(0, 65350, '\P{ishex}', "");
Expect(1, 65350, '\P{^ishex}', "");
Expect(0, 65351, '\p{ishex}', "");
Expect(1, 65351, '\p{^ishex}', "");
Expect(1, 65351, '\P{ishex}', "");
Expect(0, 65351, '\P{^ishex}', "");
Expect(1, 65350, '\p{_is_HEX}', "");
Expect(0, 65350, '\p{^_is_HEX}', "");
Expect(0, 65350, '\P{_is_HEX}', "");
Expect(1, 65350, '\P{^_is_HEX}', "");
Expect(0, 65351, '\p{_is_HEX}', "");
Expect(1, 65351, '\p{^_is_HEX}', "");
Expect(1, 65351, '\P{_is_HEX}', "");
Expect(0, 65351, '\P{^_is_HEX}', "");
Error('\p{__XID_CONTINUE:=}');
Error('\P{__XID_CONTINUE:=}');
Expect(1, 917999, '\p{xidcontinue}', "");
Expect(0, 917999, '\p{^xidcontinue}', "");
Expect(0, 917999, '\P{xidcontinue}', "");
Expect(1, 917999, '\P{^xidcontinue}', "");
Expect(0, 918000, '\p{xidcontinue}', "");
Expect(1, 918000, '\p{^xidcontinue}', "");
Expect(1, 918000, '\P{xidcontinue}', "");
Expect(0, 918000, '\P{^xidcontinue}', "");
Expect(1, 917999, '\p{_XID_continue}', "");
Expect(0, 917999, '\p{^_XID_continue}', "");
Expect(0, 917999, '\P{_XID_continue}', "");
Expect(1, 917999, '\P{^_XID_continue}', "");
Expect(0, 918000, '\p{_XID_continue}', "");
Expect(1, 918000, '\p{^_XID_continue}', "");
Expect(1, 918000, '\P{_XID_continue}', "");
Expect(0, 918000, '\P{^_XID_continue}', "");
Error('\p{ /a/is_XID_Continue}');
Error('\P{ /a/is_XID_Continue}');
Expect(1, 917999, '\p{isxidcontinue}', "");
Expect(0, 917999, '\p{^isxidcontinue}', "");
Expect(0, 917999, '\P{isxidcontinue}', "");
Expect(1, 917999, '\P{^isxidcontinue}', "");
Expect(0, 918000, '\p{isxidcontinue}', "");
Expect(1, 918000, '\p{^isxidcontinue}', "");
Expect(1, 918000, '\P{isxidcontinue}', "");
Expect(0, 918000, '\P{^isxidcontinue}', "");
Expect(1, 917999, '\p{--is_XID_Continue}', "");
Expect(0, 917999, '\p{^--is_XID_Continue}', "");
Expect(0, 917999, '\P{--is_XID_Continue}', "");
Expect(1, 917999, '\P{^--is_XID_Continue}', "");
Expect(0, 918000, '\p{--is_XID_Continue}', "");
Expect(1, 918000, '\p{^--is_XID_Continue}', "");
Expect(1, 918000, '\P{--is_XID_Continue}', "");
Expect(0, 918000, '\P{^--is_XID_Continue}', "");
Error('\p{/a/		XIDC}');
Error('\P{/a/		XIDC}');
Expect(1, 917999, '\p{xidc}', "");
Expect(0, 917999, '\p{^xidc}', "");
Expect(0, 917999, '\P{xidc}', "");
Expect(1, 917999, '\P{^xidc}', "");
Expect(0, 918000, '\p{xidc}', "");
Expect(1, 918000, '\p{^xidc}', "");
Expect(1, 918000, '\P{xidc}', "");
Expect(0, 918000, '\P{^xidc}', "");
Expect(1, 917999, '\p{_ XIDC}', "");
Expect(0, 917999, '\p{^_ XIDC}', "");
Expect(0, 917999, '\P{_ XIDC}', "");
Expect(1, 917999, '\P{^_ XIDC}', "");
Expect(0, 918000, '\p{_ XIDC}', "");
Expect(1, 918000, '\p{^_ XIDC}', "");
Expect(1, 918000, '\P{_ XIDC}', "");
Expect(0, 918000, '\P{^_ XIDC}', "");
Error('\p{ IS_XIDC/a/}');
Error('\P{ IS_XIDC/a/}');
Expect(1, 917999, '\p{isxidc}', "");
Expect(0, 917999, '\p{^isxidc}', "");
Expect(0, 917999, '\P{isxidc}', "");
Expect(1, 917999, '\P{^isxidc}', "");
Expect(0, 918000, '\p{isxidc}', "");
Expect(1, 918000, '\p{^isxidc}', "");
Expect(1, 918000, '\P{isxidc}', "");
Expect(0, 918000, '\P{^isxidc}', "");
Expect(1, 917999, '\p{ IS_XIDC}', "");
Expect(0, 917999, '\p{^ IS_XIDC}', "");
Expect(0, 917999, '\P{ IS_XIDC}', "");
Expect(1, 917999, '\P{^ IS_XIDC}', "");
Expect(0, 918000, '\p{ IS_XIDC}', "");
Expect(1, 918000, '\p{^ IS_XIDC}', "");
Expect(1, 918000, '\P{ IS_XIDC}', "");
Expect(0, 918000, '\P{^ IS_XIDC}', "");
Error('\p{/a/-XID_Start}');
Error('\P{/a/-XID_Start}');
Expect(1, 195101, '\p{xidstart}', "");
Expect(0, 195101, '\p{^xidstart}', "");
Expect(0, 195101, '\P{xidstart}', "");
Expect(1, 195101, '\P{^xidstart}', "");
Expect(0, 195102, '\p{xidstart}', "");
Expect(1, 195102, '\p{^xidstart}', "");
Expect(1, 195102, '\P{xidstart}', "");
Expect(0, 195102, '\P{^xidstart}', "");
Expect(1, 195101, '\p{_	XID_Start}', "");
Expect(0, 195101, '\p{^_	XID_Start}', "");
Expect(0, 195101, '\P{_	XID_Start}', "");
Expect(1, 195101, '\P{^_	XID_Start}', "");
Expect(0, 195102, '\p{_	XID_Start}', "");
Expect(1, 195102, '\p{^_	XID_Start}', "");
Expect(1, 195102, '\P{_	XID_Start}', "");
Expect(0, 195102, '\P{^_	XID_Start}', "");
Error('\p{  Is_XID_Start:=}');
Error('\P{  Is_XID_Start:=}');
Expect(1, 195101, '\p{isxidstart}', "");
Expect(0, 195101, '\p{^isxidstart}', "");
Expect(0, 195101, '\P{isxidstart}', "");
Expect(1, 195101, '\P{^isxidstart}', "");
Expect(0, 195102, '\p{isxidstart}', "");
Expect(1, 195102, '\p{^isxidstart}', "");
Expect(1, 195102, '\P{isxidstart}', "");
Expect(0, 195102, '\P{^isxidstart}', "");
Expect(1, 195101, '\p{_ IS_XID_START}', "");
Expect(0, 195101, '\p{^_ IS_XID_START}', "");
Expect(0, 195101, '\P{_ IS_XID_START}', "");
Expect(1, 195101, '\P{^_ IS_XID_START}', "");
Expect(0, 195102, '\p{_ IS_XID_START}', "");
Expect(1, 195102, '\p{^_ IS_XID_START}', "");
Expect(1, 195102, '\P{_ IS_XID_START}', "");
Expect(0, 195102, '\P{^_ IS_XID_START}', "");
Error('\p{_XIDS/a/}');
Error('\P{_XIDS/a/}');
Expect(1, 195101, '\p{xids}', "");
Expect(0, 195101, '\p{^xids}', "");
Expect(0, 195101, '\P{xids}', "");
Expect(1, 195101, '\P{^xids}', "");
Expect(0, 195102, '\p{xids}', "");
Expect(1, 195102, '\p{^xids}', "");
Expect(1, 195102, '\P{xids}', "");
Expect(0, 195102, '\P{^xids}', "");
Expect(1, 195101, '\p{_XIDS}', "");
Expect(0, 195101, '\p{^_XIDS}', "");
Expect(0, 195101, '\P{_XIDS}', "");
Expect(1, 195101, '\P{^_XIDS}', "");
Expect(0, 195102, '\p{_XIDS}', "");
Expect(1, 195102, '\p{^_XIDS}', "");
Expect(1, 195102, '\P{_XIDS}', "");
Expect(0, 195102, '\P{^_XIDS}', "");
Error('\p{	:=IS_XIDS}');
Error('\P{	:=IS_XIDS}');
Expect(1, 195101, '\p{isxids}', "");
Expect(0, 195101, '\p{^isxids}', "");
Expect(0, 195101, '\P{isxids}', "");
Expect(1, 195101, '\P{^isxids}', "");
Expect(0, 195102, '\p{isxids}', "");
Expect(1, 195102, '\p{^isxids}', "");
Expect(1, 195102, '\P{isxids}', "");
Expect(0, 195102, '\P{^isxids}', "");
Expect(1, 195101, '\p{-Is_XIDS}', "");
Expect(0, 195101, '\p{^-Is_XIDS}', "");
Expect(0, 195101, '\P{-Is_XIDS}', "");
Expect(1, 195101, '\P{^-Is_XIDS}', "");
Expect(0, 195102, '\p{-Is_XIDS}', "");
Expect(1, 195102, '\p{^-Is_XIDS}', "");
Expect(1, 195102, '\P{-Is_XIDS}', "");
Expect(0, 195102, '\P{^-Is_XIDS}', "");
Error('\p{-/a/XPerlSpace}');
Error('\P{-/a/XPerlSpace}');
Expect(1, 12288, '\p{xperlspace}', "");
Expect(0, 12288, '\p{^xperlspace}', "");
Expect(0, 12288, '\P{xperlspace}', "");
Expect(1, 12288, '\P{^xperlspace}', "");
Expect(0, 12289, '\p{xperlspace}', "");
Expect(1, 12289, '\p{^xperlspace}', "");
Expect(1, 12289, '\P{xperlspace}', "");
Expect(0, 12289, '\P{^xperlspace}', "");
Expect(1, 12288, '\p{ -XPerlSpace}', "");
Expect(0, 12288, '\p{^ -XPerlSpace}', "");
Expect(0, 12288, '\P{ -XPerlSpace}', "");
Expect(1, 12288, '\P{^ -XPerlSpace}', "");
Expect(0, 12289, '\p{ -XPerlSpace}', "");
Expect(1, 12289, '\p{^ -XPerlSpace}', "");
Expect(1, 12289, '\P{ -XPerlSpace}', "");
Expect(0, 12289, '\P{^ -XPerlSpace}', "");
Error('\p{ /a/SpacePerl}');
Error('\P{ /a/SpacePerl}');
Expect(1, 12288, '\p{spaceperl}', "");
Expect(0, 12288, '\p{^spaceperl}', "");
Expect(0, 12288, '\P{spaceperl}', "");
Expect(1, 12288, '\P{^spaceperl}', "");
Expect(0, 12289, '\p{spaceperl}', "");
Expect(1, 12289, '\p{^spaceperl}', "");
Expect(1, 12289, '\P{spaceperl}', "");
Expect(0, 12289, '\P{^spaceperl}', "");
Expect(1, 12288, '\p{-spaceperl}', "");
Expect(0, 12288, '\p{^-spaceperl}', "");
Expect(0, 12288, '\P{-spaceperl}', "");
Expect(1, 12288, '\P{^-spaceperl}', "");
Expect(0, 12289, '\p{-spaceperl}', "");
Expect(1, 12289, '\p{^-spaceperl}', "");
Expect(1, 12289, '\P{-spaceperl}', "");
Expect(0, 12289, '\P{^-spaceperl}', "");
Error('\p{/a/Is_xperlspace}');
Error('\P{/a/Is_xperlspace}');
Expect(1, 12288, '\p{isxperlspace}', "");
Expect(0, 12288, '\p{^isxperlspace}', "");
Expect(0, 12288, '\P{isxperlspace}', "");
Expect(1, 12288, '\P{^isxperlspace}', "");
Expect(0, 12289, '\p{isxperlspace}', "");
Expect(1, 12289, '\p{^isxperlspace}', "");
Expect(1, 12289, '\P{isxperlspace}', "");
Expect(0, 12289, '\P{^isxperlspace}', "");
Expect(1, 12288, '\p{ is_XPERLSPACE}', "");
Expect(0, 12288, '\p{^ is_XPERLSPACE}', "");
Expect(0, 12288, '\P{ is_XPERLSPACE}', "");
Expect(1, 12288, '\P{^ is_XPERLSPACE}', "");
Expect(0, 12289, '\p{ is_XPERLSPACE}', "");
Expect(1, 12289, '\p{^ is_XPERLSPACE}', "");
Expect(1, 12289, '\P{ is_XPERLSPACE}', "");
Expect(0, 12289, '\P{^ is_XPERLSPACE}', "");
Error('\p{/a/ _Is_SPACEPERL}');
Error('\P{/a/ _Is_SPACEPERL}');
Expect(1, 12288, '\p{isspaceperl}', "");
Expect(0, 12288, '\p{^isspaceperl}', "");
Expect(0, 12288, '\P{isspaceperl}', "");
Expect(1, 12288, '\P{^isspaceperl}', "");
Expect(0, 12289, '\p{isspaceperl}', "");
Expect(1, 12289, '\p{^isspaceperl}', "");
Expect(1, 12289, '\P{isspaceperl}', "");
Expect(0, 12289, '\P{^isspaceperl}', "");
Expect(1, 12288, '\p{ is_SpacePerl}', "");
Expect(0, 12288, '\p{^ is_SpacePerl}', "");
Expect(0, 12288, '\P{ is_SpacePerl}', "");
Expect(1, 12288, '\P{^ is_SpacePerl}', "");
Expect(0, 12289, '\p{ is_SpacePerl}', "");
Expect(1, 12289, '\p{^ is_SpacePerl}', "");
Expect(1, 12289, '\P{ is_SpacePerl}', "");
Expect(0, 12289, '\P{^ is_SpacePerl}', "");
Error('\p{:=XPOSIXPUNCT}');
Error('\P{:=XPOSIXPUNCT}');
Expect(1, 74867, '\p{xposixpunct}', "");
Expect(0, 74867, '\p{^xposixpunct}', "");
Expect(0, 74867, '\P{xposixpunct}', "");
Expect(1, 74867, '\P{^xposixpunct}', "");
Expect(0, 74868, '\p{xposixpunct}', "");
Expect(1, 74868, '\p{^xposixpunct}', "");
Expect(1, 74868, '\P{xposixpunct}', "");
Expect(0, 74868, '\P{^xposixpunct}', "");
Expect(1, 74867, '\p{  XPOSIXPUNCT}', "");
Expect(0, 74867, '\p{^  XPOSIXPUNCT}', "");
Expect(0, 74867, '\P{  XPOSIXPUNCT}', "");
Expect(1, 74867, '\P{^  XPOSIXPUNCT}', "");
Expect(0, 74868, '\p{  XPOSIXPUNCT}', "");
Expect(1, 74868, '\p{^  XPOSIXPUNCT}', "");
Expect(1, 74868, '\P{  XPOSIXPUNCT}', "");
Expect(0, 74868, '\P{^  XPOSIXPUNCT}', "");
Error('\p{_/a/Is_xposixpunct}');
Error('\P{_/a/Is_xposixpunct}');
Expect(1, 74867, '\p{isxposixpunct}', "");
Expect(0, 74867, '\p{^isxposixpunct}', "");
Expect(0, 74867, '\P{isxposixpunct}', "");
Expect(1, 74867, '\P{^isxposixpunct}', "");
Expect(0, 74868, '\p{isxposixpunct}', "");
Expect(1, 74868, '\p{^isxposixpunct}', "");
Expect(1, 74868, '\P{isxposixpunct}', "");
Expect(0, 74868, '\P{^isxposixpunct}', "");
Expect(1, 74867, '\p{	 is_XPosixPunct}', "");
Expect(0, 74867, '\p{^	 is_XPosixPunct}', "");
Expect(0, 74867, '\P{	 is_XPosixPunct}', "");
Expect(1, 74867, '\P{^	 is_XPosixPunct}', "");
Expect(0, 74868, '\p{	 is_XPosixPunct}', "");
Expect(1, 74868, '\p{^	 is_XPosixPunct}', "");
Expect(1, 74868, '\P{	 is_XPosixPunct}', "");
Expect(0, 74868, '\P{^	 is_XPosixPunct}', "");
Error('\p{ 	yi:=}');
Error('\P{ 	yi:=}');
Expect(1, 42182, '\p{yi}', "");
Expect(0, 42182, '\p{^yi}', "");
Expect(0, 42182, '\P{yi}', "");
Expect(1, 42182, '\P{^yi}', "");
Expect(0, 42183, '\p{yi}', "");
Expect(1, 42183, '\p{^yi}', "");
Expect(1, 42183, '\P{yi}', "");
Expect(0, 42183, '\P{^yi}', "");
Expect(1, 42182, '\p{_-YI}', "");
Expect(0, 42182, '\p{^_-YI}', "");
Expect(0, 42182, '\P{_-YI}', "");
Expect(1, 42182, '\P{^_-YI}', "");
Expect(0, 42183, '\p{_-YI}', "");
Expect(1, 42183, '\p{^_-YI}', "");
Expect(1, 42183, '\P{_-YI}', "");
Expect(0, 42183, '\P{^_-YI}', "");
Error('\p{_/a/IS_Yi}');
Error('\P{_/a/IS_Yi}');
Expect(1, 42182, '\p{isyi}', "");
Expect(0, 42182, '\p{^isyi}', "");
Expect(0, 42182, '\P{isyi}', "");
Expect(1, 42182, '\P{^isyi}', "");
Expect(0, 42183, '\p{isyi}', "");
Expect(1, 42183, '\p{^isyi}', "");
Expect(1, 42183, '\P{isyi}', "");
Expect(0, 42183, '\P{^isyi}', "");
Expect(1, 42182, '\p{	_IS_Yi}', "");
Expect(0, 42182, '\p{^	_IS_Yi}', "");
Expect(0, 42182, '\P{	_IS_Yi}', "");
Expect(1, 42182, '\P{^	_IS_Yi}', "");
Expect(0, 42183, '\p{	_IS_Yi}', "");
Expect(1, 42183, '\p{^	_IS_Yi}', "");
Expect(1, 42183, '\P{	_IS_Yi}', "");
Expect(0, 42183, '\P{^	_IS_Yi}', "");
Error('\p{:=Yiii}');
Error('\P{:=Yiii}');
Expect(1, 42182, '\p{yiii}', "");
Expect(0, 42182, '\p{^yiii}', "");
Expect(0, 42182, '\P{yiii}', "");
Expect(1, 42182, '\P{^yiii}', "");
Expect(0, 42183, '\p{yiii}', "");
Expect(1, 42183, '\p{^yiii}', "");
Expect(1, 42183, '\P{yiii}', "");
Expect(0, 42183, '\P{^yiii}', "");
Expect(1, 42182, '\p{YIII}', "");
Expect(0, 42182, '\p{^YIII}', "");
Expect(0, 42182, '\P{YIII}', "");
Expect(1, 42182, '\P{^YIII}', "");
Expect(0, 42183, '\p{YIII}', "");
Expect(1, 42183, '\p{^YIII}', "");
Expect(1, 42183, '\P{YIII}', "");
Expect(0, 42183, '\P{^YIII}', "");
Error('\p{-/a/Is_Yiii}');
Error('\P{-/a/Is_Yiii}');
Expect(1, 42182, '\p{isyiii}', "");
Expect(0, 42182, '\p{^isyiii}', "");
Expect(0, 42182, '\P{isyiii}', "");
Expect(1, 42182, '\P{^isyiii}', "");
Expect(0, 42183, '\p{isyiii}', "");
Expect(1, 42183, '\p{^isyiii}', "");
Expect(1, 42183, '\P{isyiii}', "");
Expect(0, 42183, '\P{^isyiii}', "");
Expect(1, 42182, '\p{-	Is_YIII}', "");
Expect(0, 42182, '\p{^-	Is_YIII}', "");
Expect(0, 42182, '\P{-	Is_YIII}', "");
Expect(1, 42182, '\P{^-	Is_YIII}', "");
Expect(0, 42183, '\p{-	Is_YIII}', "");
Expect(1, 42183, '\p{^-	Is_YIII}', "");
Expect(1, 42183, '\P{-	Is_YIII}', "");
Expect(0, 42183, '\P{^-	Is_YIII}', "");
Error('\p{	-yi_RADICALS:=}');
Error('\P{	-yi_RADICALS:=}');
Expect(1, 42191, '\p{yiradicals}', "");
Expect(0, 42191, '\p{^yiradicals}', "");
Expect(0, 42191, '\P{yiradicals}', "");
Expect(1, 42191, '\P{^yiradicals}', "");
Expect(0, 42192, '\p{yiradicals}', "");
Expect(1, 42192, '\p{^yiradicals}', "");
Expect(1, 42192, '\P{yiradicals}', "");
Expect(0, 42192, '\P{^yiradicals}', "");
Expect(1, 42191, '\p{-_Yi_Radicals}', "");
Expect(0, 42191, '\p{^-_Yi_Radicals}', "");
Expect(0, 42191, '\P{-_Yi_Radicals}', "");
Expect(1, 42191, '\P{^-_Yi_Radicals}', "");
Expect(0, 42192, '\p{-_Yi_Radicals}', "");
Expect(1, 42192, '\p{^-_Yi_Radicals}', "");
Expect(1, 42192, '\P{-_Yi_Radicals}', "");
Expect(0, 42192, '\P{^-_Yi_Radicals}', "");
Error('\p{	:=Is_yi_Radicals}');
Error('\P{	:=Is_yi_Radicals}');
Expect(1, 42191, '\p{isyiradicals}', "");
Expect(0, 42191, '\p{^isyiradicals}', "");
Expect(0, 42191, '\P{isyiradicals}', "");
Expect(1, 42191, '\P{^isyiradicals}', "");
Expect(0, 42192, '\p{isyiradicals}', "");
Expect(1, 42192, '\p{^isyiradicals}', "");
Expect(1, 42192, '\P{isyiradicals}', "");
Expect(0, 42192, '\P{^isyiradicals}', "");
Expect(1, 42191, '\p{-Is_yi_Radicals}', "");
Expect(0, 42191, '\p{^-Is_yi_Radicals}', "");
Expect(0, 42191, '\P{-Is_yi_Radicals}', "");
Expect(1, 42191, '\P{^-Is_yi_Radicals}', "");
Expect(0, 42192, '\p{-Is_yi_Radicals}', "");
Expect(1, 42192, '\p{^-Is_yi_Radicals}', "");
Expect(1, 42192, '\P{-Is_yi_Radicals}', "");
Expect(0, 42192, '\P{^-Is_yi_Radicals}', "");
Error('\p{_In_yi_Radicals/a/}');
Error('\P{_In_yi_Radicals/a/}');
Expect(1, 42191, '\p{inyiradicals}', "");
Expect(0, 42191, '\p{^inyiradicals}', "");
Expect(0, 42191, '\P{inyiradicals}', "");
Expect(1, 42191, '\P{^inyiradicals}', "");
Expect(0, 42192, '\p{inyiradicals}', "");
Expect(1, 42192, '\p{^inyiradicals}', "");
Expect(1, 42192, '\P{inyiradicals}', "");
Expect(0, 42192, '\P{^inyiradicals}', "");
Expect(1, 42191, '\p{	 In_Yi_radicals}', "");
Expect(0, 42191, '\p{^	 In_Yi_radicals}', "");
Expect(0, 42191, '\P{	 In_Yi_radicals}', "");
Expect(1, 42191, '\P{^	 In_Yi_radicals}', "");
Expect(0, 42192, '\p{	 In_Yi_radicals}', "");
Expect(1, 42192, '\p{^	 In_Yi_radicals}', "");
Expect(1, 42192, '\P{	 In_Yi_radicals}', "");
Expect(0, 42192, '\P{^	 In_Yi_radicals}', "");
Error('\p{/a/__Yi_Syllables}');
Error('\P{/a/__Yi_Syllables}');
Expect(1, 42127, '\p{yisyllables}', "");
Expect(0, 42127, '\p{^yisyllables}', "");
Expect(0, 42127, '\P{yisyllables}', "");
Expect(1, 42127, '\P{^yisyllables}', "");
Expect(0, 42128, '\p{yisyllables}', "");
Expect(1, 42128, '\p{^yisyllables}', "");
Expect(1, 42128, '\P{yisyllables}', "");
Expect(0, 42128, '\P{^yisyllables}', "");
Expect(1, 42127, '\p{__Yi_SYLLABLES}', "");
Expect(0, 42127, '\p{^__Yi_SYLLABLES}', "");
Expect(0, 42127, '\P{__Yi_SYLLABLES}', "");
Expect(1, 42127, '\P{^__Yi_SYLLABLES}', "");
Expect(0, 42128, '\p{__Yi_SYLLABLES}', "");
Expect(1, 42128, '\p{^__Yi_SYLLABLES}', "");
Expect(1, 42128, '\P{__Yi_SYLLABLES}', "");
Expect(0, 42128, '\P{^__Yi_SYLLABLES}', "");
Error('\p{-/a/is_Yi_Syllables}');
Error('\P{-/a/is_Yi_Syllables}');
Expect(1, 42127, '\p{isyisyllables}', "");
Expect(0, 42127, '\p{^isyisyllables}', "");
Expect(0, 42127, '\P{isyisyllables}', "");
Expect(1, 42127, '\P{^isyisyllables}', "");
Expect(0, 42128, '\p{isyisyllables}', "");
Expect(1, 42128, '\p{^isyisyllables}', "");
Expect(1, 42128, '\P{isyisyllables}', "");
Expect(0, 42128, '\P{^isyisyllables}', "");
Expect(1, 42127, '\p{-_is_YI_Syllables}', "");
Expect(0, 42127, '\p{^-_is_YI_Syllables}', "");
Expect(0, 42127, '\P{-_is_YI_Syllables}', "");
Expect(1, 42127, '\P{^-_is_YI_Syllables}', "");
Expect(0, 42128, '\p{-_is_YI_Syllables}', "");
Expect(1, 42128, '\p{^-_is_YI_Syllables}', "");
Expect(1, 42128, '\P{-_is_YI_Syllables}', "");
Expect(0, 42128, '\P{^-_is_YI_Syllables}', "");
Error('\p{-/a/in_yi_Syllables}');
Error('\P{-/a/in_yi_Syllables}');
Expect(1, 42127, '\p{inyisyllables}', "");
Expect(0, 42127, '\p{^inyisyllables}', "");
Expect(0, 42127, '\P{inyisyllables}', "");
Expect(1, 42127, '\P{^inyisyllables}', "");
Expect(0, 42128, '\p{inyisyllables}', "");
Expect(1, 42128, '\p{^inyisyllables}', "");
Expect(1, 42128, '\P{inyisyllables}', "");
Expect(0, 42128, '\P{^inyisyllables}', "");
Expect(1, 42127, '\p{ IN_Yi_syllables}', "");
Expect(0, 42127, '\p{^ IN_Yi_syllables}', "");
Expect(0, 42127, '\P{ IN_Yi_syllables}', "");
Expect(1, 42127, '\P{^ IN_Yi_syllables}', "");
Expect(0, 42128, '\p{ IN_Yi_syllables}', "");
Expect(1, 42128, '\p{^ IN_Yi_syllables}', "");
Expect(1, 42128, '\P{ IN_Yi_syllables}', "");
Expect(0, 42128, '\P{^ IN_Yi_syllables}', "");
Error('\p{		yijing_Hexagram_SYMBOLS:=}');
Error('\P{		yijing_Hexagram_SYMBOLS:=}');
Expect(1, 19967, '\p{yijinghexagramsymbols}', "");
Expect(0, 19967, '\p{^yijinghexagramsymbols}', "");
Expect(0, 19967, '\P{yijinghexagramsymbols}', "");
Expect(1, 19967, '\P{^yijinghexagramsymbols}', "");
Expect(0, 19968, '\p{yijinghexagramsymbols}', "");
Expect(1, 19968, '\p{^yijinghexagramsymbols}', "");
Expect(1, 19968, '\P{yijinghexagramsymbols}', "");
Expect(0, 19968, '\P{^yijinghexagramsymbols}', "");
Expect(1, 19967, '\p{--Yijing_HEXAGRAM_Symbols}', "");
Expect(0, 19967, '\p{^--Yijing_HEXAGRAM_Symbols}', "");
Expect(0, 19967, '\P{--Yijing_HEXAGRAM_Symbols}', "");
Expect(1, 19967, '\P{^--Yijing_HEXAGRAM_Symbols}', "");
Expect(0, 19968, '\p{--Yijing_HEXAGRAM_Symbols}', "");
Expect(1, 19968, '\p{^--Yijing_HEXAGRAM_Symbols}', "");
Expect(1, 19968, '\P{--Yijing_HEXAGRAM_Symbols}', "");
Expect(0, 19968, '\P{^--Yijing_HEXAGRAM_Symbols}', "");
Error('\p{:=		Is_Yijing_HEXAGRAM_Symbols}');
Error('\P{:=		Is_Yijing_HEXAGRAM_Symbols}');
Expect(1, 19967, '\p{isyijinghexagramsymbols}', "");
Expect(0, 19967, '\p{^isyijinghexagramsymbols}', "");
Expect(0, 19967, '\P{isyijinghexagramsymbols}', "");
Expect(1, 19967, '\P{^isyijinghexagramsymbols}', "");
Expect(0, 19968, '\p{isyijinghexagramsymbols}', "");
Expect(1, 19968, '\p{^isyijinghexagramsymbols}', "");
Expect(1, 19968, '\P{isyijinghexagramsymbols}', "");
Expect(0, 19968, '\P{^isyijinghexagramsymbols}', "");
Expect(1, 19967, '\p{- IS_yijing_hexagram_SYMBOLS}', "");
Expect(0, 19967, '\p{^- IS_yijing_hexagram_SYMBOLS}', "");
Expect(0, 19967, '\P{- IS_yijing_hexagram_SYMBOLS}', "");
Expect(1, 19967, '\P{^- IS_yijing_hexagram_SYMBOLS}', "");
Expect(0, 19968, '\p{- IS_yijing_hexagram_SYMBOLS}', "");
Expect(1, 19968, '\p{^- IS_yijing_hexagram_SYMBOLS}', "");
Expect(1, 19968, '\P{- IS_yijing_hexagram_SYMBOLS}', "");
Expect(0, 19968, '\P{^- IS_yijing_hexagram_SYMBOLS}', "");
Error('\p{:=-IN_YIJING_HEXAGRAM_symbols}');
Error('\P{:=-IN_YIJING_HEXAGRAM_symbols}');
Expect(1, 19967, '\p{inyijinghexagramsymbols}', "");
Expect(0, 19967, '\p{^inyijinghexagramsymbols}', "");
Expect(0, 19967, '\P{inyijinghexagramsymbols}', "");
Expect(1, 19967, '\P{^inyijinghexagramsymbols}', "");
Expect(0, 19968, '\p{inyijinghexagramsymbols}', "");
Expect(1, 19968, '\p{^inyijinghexagramsymbols}', "");
Expect(1, 19968, '\P{inyijinghexagramsymbols}', "");
Expect(0, 19968, '\P{^inyijinghexagramsymbols}', "");
Expect(1, 19967, '\p{	IN_Yijing_HEXAGRAM_Symbols}', "");
Expect(0, 19967, '\p{^	IN_Yijing_HEXAGRAM_Symbols}', "");
Expect(0, 19967, '\P{	IN_Yijing_HEXAGRAM_Symbols}', "");
Expect(1, 19967, '\P{^	IN_Yijing_HEXAGRAM_Symbols}', "");
Expect(0, 19968, '\p{	IN_Yijing_HEXAGRAM_Symbols}', "");
Expect(1, 19968, '\p{^	IN_Yijing_HEXAGRAM_Symbols}', "");
Expect(1, 19968, '\P{	IN_Yijing_HEXAGRAM_Symbols}', "");
Expect(0, 19968, '\P{^	IN_Yijing_HEXAGRAM_Symbols}', "");
Error('\p{:=_YIJING}');
Error('\P{:=_YIJING}');
Expect(1, 19967, '\p{yijing}', "");
Expect(0, 19967, '\p{^yijing}', "");
Expect(0, 19967, '\P{yijing}', "");
Expect(1, 19967, '\P{^yijing}', "");
Expect(0, 19968, '\p{yijing}', "");
Expect(1, 19968, '\p{^yijing}', "");
Expect(1, 19968, '\P{yijing}', "");
Expect(0, 19968, '\P{^yijing}', "");
Expect(1, 19967, '\p{_Yijing}', "");
Expect(0, 19967, '\p{^_Yijing}', "");
Expect(0, 19967, '\P{_Yijing}', "");
Expect(1, 19967, '\P{^_Yijing}', "");
Expect(0, 19968, '\p{_Yijing}', "");
Expect(1, 19968, '\p{^_Yijing}', "");
Expect(1, 19968, '\P{_Yijing}', "");
Expect(0, 19968, '\P{^_Yijing}', "");
Error('\p{_is_Yijing:=}');
Error('\P{_is_Yijing:=}');
Expect(1, 19967, '\p{isyijing}', "");
Expect(0, 19967, '\p{^isyijing}', "");
Expect(0, 19967, '\P{isyijing}', "");
Expect(1, 19967, '\P{^isyijing}', "");
Expect(0, 19968, '\p{isyijing}', "");
Expect(1, 19968, '\p{^isyijing}', "");
Expect(1, 19968, '\P{isyijing}', "");
Expect(0, 19968, '\P{^isyijing}', "");
Expect(1, 19967, '\p{ -IS_yijing}', "");
Expect(0, 19967, '\p{^ -IS_yijing}', "");
Expect(0, 19967, '\P{ -IS_yijing}', "");
Expect(1, 19967, '\P{^ -IS_yijing}', "");
Expect(0, 19968, '\p{ -IS_yijing}', "");
Expect(1, 19968, '\p{^ -IS_yijing}', "");
Expect(1, 19968, '\P{ -IS_yijing}', "");
Expect(0, 19968, '\P{^ -IS_yijing}', "");
Error('\p{	in_Yijing/a/}');
Error('\P{	in_Yijing/a/}');
Expect(1, 19967, '\p{inyijing}', "");
Expect(0, 19967, '\p{^inyijing}', "");
Expect(0, 19967, '\P{inyijing}', "");
Expect(1, 19967, '\P{^inyijing}', "");
Expect(0, 19968, '\p{inyijing}', "");
Expect(1, 19968, '\p{^inyijing}', "");
Expect(1, 19968, '\P{inyijing}', "");
Expect(0, 19968, '\P{^inyijing}', "");
Expect(1, 19967, '\p{_-in_Yijing}', "");
Expect(0, 19967, '\p{^_-in_Yijing}', "");
Expect(0, 19967, '\P{_-in_Yijing}', "");
Expect(1, 19967, '\P{^_-in_Yijing}', "");
Expect(0, 19968, '\p{_-in_Yijing}', "");
Expect(1, 19968, '\p{^_-in_Yijing}', "");
Expect(1, 19968, '\P{_-in_Yijing}', "");
Expect(0, 19968, '\P{^_-in_Yijing}', "");
Error('\p{	_SOFT_dotted:=}');
Error('\P{	_SOFT_dotted:=}');
Expect(1, 120467, '\p{softdotted}', "");
Expect(0, 120467, '\p{^softdotted}', "");
Expect(0, 120467, '\P{softdotted}', "");
Expect(1, 120467, '\P{^softdotted}', "");
Expect(0, 120468, '\p{softdotted}', "");
Expect(1, 120468, '\p{^softdotted}', "");
Expect(1, 120468, '\P{softdotted}', "");
Expect(0, 120468, '\P{^softdotted}', "");
Expect(1, 120467, '\p{ Soft_DOTTED}', "");
Expect(0, 120467, '\p{^ Soft_DOTTED}', "");
Expect(0, 120467, '\P{ Soft_DOTTED}', "");
Expect(1, 120467, '\P{^ Soft_DOTTED}', "");
Expect(0, 120468, '\p{ Soft_DOTTED}', "");
Expect(1, 120468, '\p{^ Soft_DOTTED}', "");
Expect(1, 120468, '\P{ Soft_DOTTED}', "");
Expect(0, 120468, '\P{^ Soft_DOTTED}', "");
Error('\p{	/a/Is_SOFT_dotted}');
Error('\P{	/a/Is_SOFT_dotted}');
Expect(1, 120467, '\p{issoftdotted}', "");
Expect(0, 120467, '\p{^issoftdotted}', "");
Expect(0, 120467, '\P{issoftdotted}', "");
Expect(1, 120467, '\P{^issoftdotted}', "");
Expect(0, 120468, '\p{issoftdotted}', "");
Expect(1, 120468, '\p{^issoftdotted}', "");
Expect(1, 120468, '\P{issoftdotted}', "");
Expect(0, 120468, '\P{^issoftdotted}', "");
Expect(1, 120467, '\p{  Is_Soft_DOTTED}', "");
Expect(0, 120467, '\p{^  Is_Soft_DOTTED}', "");
Expect(0, 120467, '\P{  Is_Soft_DOTTED}', "");
Expect(1, 120467, '\P{^  Is_Soft_DOTTED}', "");
Expect(0, 120468, '\p{  Is_Soft_DOTTED}', "");
Expect(1, 120468, '\p{^  Is_Soft_DOTTED}', "");
Expect(1, 120468, '\P{  Is_Soft_DOTTED}', "");
Expect(0, 120468, '\P{^  Is_Soft_DOTTED}', "");
Error('\p{	SD/a/}');
Error('\P{	SD/a/}');
Expect(1, 120467, '\p{sd}', "");
Expect(0, 120467, '\p{^sd}', "");
Expect(0, 120467, '\P{sd}', "");
Expect(1, 120467, '\P{^sd}', "");
Expect(0, 120468, '\p{sd}', "");
Expect(1, 120468, '\p{^sd}', "");
Expect(1, 120468, '\P{sd}', "");
Expect(0, 120468, '\P{^sd}', "");
Expect(1, 120467, '\p{ SD}', "");
Expect(0, 120467, '\p{^ SD}', "");
Expect(0, 120467, '\P{ SD}', "");
Expect(1, 120467, '\P{^ SD}', "");
Expect(0, 120468, '\p{ SD}', "");
Expect(1, 120468, '\p{^ SD}', "");
Expect(1, 120468, '\P{ SD}', "");
Expect(0, 120468, '\P{^ SD}', "");
Error('\p{:=_ Is_sd}');
Error('\P{:=_ Is_sd}');
Expect(1, 120467, '\p{issd}', "");
Expect(0, 120467, '\p{^issd}', "");
Expect(0, 120467, '\P{issd}', "");
Expect(1, 120467, '\P{^issd}', "");
Expect(0, 120468, '\p{issd}', "");
Expect(1, 120468, '\p{^issd}', "");
Expect(1, 120468, '\P{issd}', "");
Expect(0, 120468, '\P{^issd}', "");
Expect(1, 120467, '\p{__Is_SD}', "");
Expect(0, 120467, '\p{^__Is_SD}', "");
Expect(0, 120467, '\P{__Is_SD}', "");
Expect(1, 120467, '\P{^__Is_SD}', "");
Expect(0, 120468, '\p{__Is_SD}', "");
Expect(1, 120468, '\p{^__Is_SD}', "");
Expect(1, 120468, '\P{__Is_SD}', "");
Expect(0, 120468, '\P{^__Is_SD}', "");
Error('\p{__canondcij:=}');
Error('\P{__canondcij:=}');
Expect(1, 120467, '\p{_canondcij}', "");
Expect(0, 120467, '\p{^_canondcij}', "");
Expect(0, 120467, '\P{_canondcij}', "");
Expect(1, 120467, '\P{^_canondcij}', "");
Expect(0, 120468, '\p{_canondcij}', "");
Expect(1, 120468, '\p{^_canondcij}', "");
Expect(1, 120468, '\P{_canondcij}', "");
Expect(0, 120468, '\P{^_canondcij}', "");
Error('\p{/a/-_Case_Ignorable}');
Error('\P{/a/-_Case_Ignorable}');
Expect(1, 917999, '\p{caseignorable}', "");
Expect(0, 917999, '\p{^caseignorable}', "");
Expect(0, 917999, '\P{caseignorable}', "");
Expect(1, 917999, '\P{^caseignorable}', "");
Expect(0, 918000, '\p{caseignorable}', "");
Expect(1, 918000, '\p{^caseignorable}', "");
Expect(1, 918000, '\P{caseignorable}', "");
Expect(0, 918000, '\P{^caseignorable}', "");
Expect(1, 917999, '\p{  Case_Ignorable}', "");
Expect(0, 917999, '\p{^  Case_Ignorable}', "");
Expect(0, 917999, '\P{  Case_Ignorable}', "");
Expect(1, 917999, '\P{^  Case_Ignorable}', "");
Expect(0, 918000, '\p{  Case_Ignorable}', "");
Expect(1, 918000, '\p{^  Case_Ignorable}', "");
Expect(1, 918000, '\P{  Case_Ignorable}', "");
Expect(0, 918000, '\P{^  Case_Ignorable}', "");
Error('\p{ _IS_CASE_Ignorable/a/}');
Error('\P{ _IS_CASE_Ignorable/a/}');
Expect(1, 917999, '\p{iscaseignorable}', "");
Expect(0, 917999, '\p{^iscaseignorable}', "");
Expect(0, 917999, '\P{iscaseignorable}', "");
Expect(1, 917999, '\P{^iscaseignorable}', "");
Expect(0, 918000, '\p{iscaseignorable}', "");
Expect(1, 918000, '\p{^iscaseignorable}', "");
Expect(1, 918000, '\P{iscaseignorable}', "");
Expect(0, 918000, '\P{^iscaseignorable}', "");
Expect(1, 917999, '\p{_ Is_Case_Ignorable}', "");
Expect(0, 917999, '\p{^_ Is_Case_Ignorable}', "");
Expect(0, 917999, '\P{_ Is_Case_Ignorable}', "");
Expect(1, 917999, '\P{^_ Is_Case_Ignorable}', "");
Expect(0, 918000, '\p{_ Is_Case_Ignorable}', "");
Expect(1, 918000, '\p{^_ Is_Case_Ignorable}', "");
Expect(1, 918000, '\P{_ Is_Case_Ignorable}', "");
Expect(0, 918000, '\P{^_ Is_Case_Ignorable}', "");
Error('\p{-CI/a/}');
Error('\P{-CI/a/}');
Expect(1, 917999, '\p{ci}', "");
Expect(0, 917999, '\p{^ci}', "");
Expect(0, 917999, '\P{ci}', "");
Expect(1, 917999, '\P{^ci}', "");
Expect(0, 918000, '\p{ci}', "");
Expect(1, 918000, '\p{^ci}', "");
Expect(1, 918000, '\P{ci}', "");
Expect(0, 918000, '\P{^ci}', "");
Expect(1, 917999, '\p{		ci}', "");
Expect(0, 917999, '\p{^		ci}', "");
Expect(0, 917999, '\P{		ci}', "");
Expect(1, 917999, '\P{^		ci}', "");
Expect(0, 918000, '\p{		ci}', "");
Expect(1, 918000, '\p{^		ci}', "");
Expect(1, 918000, '\P{		ci}', "");
Expect(0, 918000, '\P{^		ci}', "");
Error('\p{/a/__is_CI}');
Error('\P{/a/__is_CI}');
Expect(1, 917999, '\p{isci}', "");
Expect(0, 917999, '\p{^isci}', "");
Expect(0, 917999, '\P{isci}', "");
Expect(1, 917999, '\P{^isci}', "");
Expect(0, 918000, '\p{isci}', "");
Expect(1, 918000, '\p{^isci}', "");
Expect(1, 918000, '\P{isci}', "");
Expect(0, 918000, '\P{^isci}', "");
Expect(1, 917999, '\p{ -IS_CI}', "");
Expect(0, 917999, '\p{^ -IS_CI}', "");
Expect(0, 917999, '\P{ -IS_CI}', "");
Expect(1, 917999, '\P{^ -IS_CI}', "");
Expect(0, 918000, '\p{ -IS_CI}', "");
Expect(1, 918000, '\p{^ -IS_CI}', "");
Expect(1, 918000, '\P{ -IS_CI}', "");
Expect(0, 918000, '\P{^ -IS_CI}', "");
Error('\p{/a/_Case_IGNORABLE}');
Error('\P{/a/_Case_IGNORABLE}');
Expect(1, 917999, '\p{_Case_Ignorable}', "");
Expect(0, 917999, '\p{^_Case_Ignorable}', "");
Expect(0, 917999, '\P{_Case_Ignorable}', "");
Expect(1, 917999, '\P{^_Case_Ignorable}', "");
Expect(0, 918000, '\p{_Case_Ignorable}', "");
Expect(1, 918000, '\p{^_Case_Ignorable}', "");
Expect(1, 918000, '\P{_Case_Ignorable}', "");
Expect(0, 918000, '\P{^_Case_Ignorable}', "");
Error('\p{	-_CombAbove:=}');
Error('\P{	-_CombAbove:=}');
Expect(1, 119364, '\p{_CombAbove}', "");
Expect(0, 119364, '\p{^_CombAbove}', "");
Expect(0, 119364, '\P{_CombAbove}', "");
Expect(1, 119364, '\P{^_CombAbove}', "");
Expect(0, 119365, '\p{_CombAbove}', "");
Expect(1, 119365, '\p{^_CombAbove}', "");
Expect(1, 119365, '\P{_CombAbove}', "");
Expect(0, 119365, '\P{^_CombAbove}', "");
Error('\p{--_COMBABOVE:=}');
Error('\P{--_COMBABOVE:=}');
Error('\p{--_PERL_any_folds/a/}');
Error('\P{--_PERL_any_folds/a/}');
Expect(1, 66639, '\p{_Perl_any_Folds}', "");
Expect(0, 66639, '\p{^_Perl_any_Folds}', "");
Expect(0, 66639, '\P{_Perl_any_Folds}', "");
Expect(1, 66639, '\P{^_Perl_any_Folds}', "");
Expect(0, 66640, '\p{_Perl_any_Folds}', "");
Expect(1, 66640, '\p{^_Perl_any_Folds}', "");
Expect(1, 66640, '\P{_Perl_any_Folds}', "");
Expect(0, 66640, '\P{^_Perl_any_Folds}', "");
Error('\p{ :=_PERL_Any_Folds}');
Error('\P{ :=_PERL_Any_Folds}');
Expect(1, 66639, '\p{_Perl_any_FOLDS}', "");
Expect(0, 66639, '\p{^_Perl_any_FOLDS}', "");
Expect(0, 66639, '\P{_Perl_any_FOLDS}', "");
Expect(1, 66639, '\P{^_Perl_any_FOLDS}', "");
Expect(0, 66640, '\p{_Perl_any_FOLDS}', "");
Expect(1, 66640, '\p{^_Perl_any_FOLDS}', "");
Expect(1, 66640, '\P{_Perl_any_FOLDS}', "");
Expect(0, 66640, '\P{^_Perl_any_FOLDS}', "");
Error('\p{/a/ __perl_Charname_Begin}');
Error('\P{/a/ __perl_Charname_Begin}');
Expect(1, 195101, '\p{_Perl_CHARNAME_Begin}', "");
Expect(0, 195101, '\p{^_Perl_CHARNAME_Begin}', "");
Expect(0, 195101, '\P{_Perl_CHARNAME_Begin}', "");
Expect(1, 195101, '\P{^_Perl_CHARNAME_Begin}', "");
Expect(0, 195102, '\p{_Perl_CHARNAME_Begin}', "");
Expect(1, 195102, '\p{^_Perl_CHARNAME_Begin}', "");
Expect(1, 195102, '\P{_Perl_CHARNAME_Begin}', "");
Expect(0, 195102, '\P{^_Perl_CHARNAME_Begin}', "");
Error('\p{:=__PERL_charname_begin}');
Error('\P{:=__PERL_charname_begin}');
Expect(1, 195101, '\p{_Perl_Charname_BEGIN}', "");
Expect(0, 195101, '\p{^_Perl_Charname_BEGIN}', "");
Expect(0, 195101, '\P{_Perl_Charname_BEGIN}', "");
Expect(1, 195101, '\P{^_Perl_Charname_BEGIN}', "");
Expect(0, 195102, '\p{_Perl_Charname_BEGIN}', "");
Expect(1, 195102, '\p{^_Perl_Charname_BEGIN}', "");
Expect(1, 195102, '\P{_Perl_Charname_BEGIN}', "");
Expect(0, 195102, '\P{^_Perl_Charname_BEGIN}', "");
Error('\p{:=	__Perl_Charname_CONTINUE}');
Error('\P{:=	__Perl_Charname_CONTINUE}');
Expect(1, 917999, '\p{_PERL_Charname_Continue}', "");
Expect(0, 917999, '\p{^_PERL_Charname_Continue}', "");
Expect(0, 917999, '\P{_PERL_Charname_Continue}', "");
Expect(1, 917999, '\P{^_PERL_Charname_Continue}', "");
Expect(0, 918000, '\p{_PERL_Charname_Continue}', "");
Expect(1, 918000, '\p{^_PERL_Charname_Continue}', "");
Expect(1, 918000, '\P{_PERL_Charname_Continue}', "");
Expect(0, 918000, '\P{^_PERL_Charname_Continue}', "");
Error('\p{__perl_Charname_continue:=}');
Error('\P{__perl_Charname_continue:=}');
Expect(1, 917999, '\p{_Perl_Charname_continue}', "");
Expect(0, 917999, '\p{^_Perl_Charname_continue}', "");
Expect(0, 917999, '\P{_Perl_Charname_continue}', "");
Expect(1, 917999, '\P{^_Perl_Charname_continue}', "");
Expect(0, 918000, '\p{_Perl_Charname_continue}', "");
Expect(1, 918000, '\p{^_Perl_Charname_continue}', "");
Expect(1, 918000, '\P{_Perl_Charname_continue}', "");
Expect(0, 918000, '\P{^_Perl_Charname_continue}', "");
Error('\p{:=	_PERL_Folds_to_MULTI_char}');
Error('\P{:=	_PERL_Folds_to_MULTI_char}');
Expect(1, 64279, '\p{_Perl_FOLDS_To_Multi_Char}', "");
Expect(0, 64279, '\p{^_Perl_FOLDS_To_Multi_Char}', "");
Expect(0, 64279, '\P{_Perl_FOLDS_To_Multi_Char}', "");
Expect(1, 64279, '\P{^_Perl_FOLDS_To_Multi_Char}', "");
Expect(0, 64280, '\p{_Perl_FOLDS_To_Multi_Char}', "");
Expect(1, 64280, '\p{^_Perl_FOLDS_To_Multi_Char}', "");
Expect(1, 64280, '\P{_Perl_FOLDS_To_Multi_Char}', "");
Expect(0, 64280, '\P{^_Perl_FOLDS_To_Multi_Char}', "");
Error('\p{-:=_Perl_Folds_To_multi_Char}');
Error('\P{-:=_Perl_Folds_To_multi_Char}');
Expect(1, 64279, '\p{_Perl_FOLDS_To_MULTI_Char}', "");
Expect(0, 64279, '\p{^_Perl_FOLDS_To_MULTI_Char}', "");
Expect(0, 64279, '\P{_Perl_FOLDS_To_MULTI_Char}', "");
Expect(1, 64279, '\P{^_Perl_FOLDS_To_MULTI_Char}', "");
Expect(0, 64280, '\p{_Perl_FOLDS_To_MULTI_Char}', "");
Expect(1, 64280, '\p{^_Perl_FOLDS_To_MULTI_Char}', "");
Expect(1, 64280, '\P{_Perl_FOLDS_To_MULTI_Char}', "");
Expect(0, 64280, '\P{^_Perl_FOLDS_To_MULTI_Char}', "");
Error('\p{/a/ _Perl_idcont}');
Error('\P{/a/ _Perl_idcont}');
Expect(1, 917999, '\p{_Perl_IDCONT}', "");
Expect(0, 917999, '\p{^_Perl_IDCONT}', "");
Expect(0, 917999, '\P{_Perl_IDCONT}', "");
Expect(1, 917999, '\P{^_Perl_IDCONT}', "");
Expect(0, 918000, '\p{_Perl_IDCONT}', "");
Expect(1, 918000, '\p{^_Perl_IDCONT}', "");
Expect(1, 918000, '\P{_Perl_IDCONT}', "");
Expect(0, 918000, '\P{^_Perl_IDCONT}', "");
Error('\p{	__perl_IDCont:=}');
Error('\P{	__perl_IDCont:=}');
Expect(1, 917999, '\p{_PERL_IDCONT}', "");
Expect(0, 917999, '\p{^_PERL_IDCONT}', "");
Expect(0, 917999, '\P{_PERL_IDCONT}', "");
Expect(1, 917999, '\P{^_PERL_IDCONT}', "");
Expect(0, 918000, '\p{_PERL_IDCONT}', "");
Expect(1, 918000, '\p{^_PERL_IDCONT}', "");
Expect(1, 918000, '\P{_PERL_IDCONT}', "");
Expect(0, 918000, '\P{^_PERL_IDCONT}', "");
Error('\p{_/a/_Perl_IDStart}');
Error('\P{_/a/_Perl_IDStart}');
Expect(1, 195101, '\p{_Perl_IDStart}', "");
Expect(0, 195101, '\p{^_Perl_IDStart}', "");
Expect(0, 195101, '\P{_Perl_IDStart}', "");
Expect(1, 195101, '\P{^_Perl_IDStart}', "");
Expect(0, 195102, '\p{_Perl_IDStart}', "");
Expect(1, 195102, '\p{^_Perl_IDStart}', "");
Expect(1, 195102, '\P{_Perl_IDStart}', "");
Expect(0, 195102, '\P{^_Perl_IDStart}', "");
Error('\p{__Perl_IDSTART:=}');
Error('\P{__Perl_IDSTART:=}');
Expect(1, 195101, '\p{_PERL_IDSTART}', "");
Expect(0, 195101, '\p{^_PERL_IDSTART}', "");
Expect(0, 195101, '\P{_PERL_IDSTART}', "");
Expect(1, 195101, '\P{^_PERL_IDSTART}', "");
Expect(0, 195102, '\p{_PERL_IDSTART}', "");
Expect(1, 195102, '\p{^_PERL_IDSTART}', "");
Expect(1, 195102, '\P{_PERL_IDSTART}', "");
Expect(0, 195102, '\P{^_PERL_IDSTART}', "");
Error('\p{_/a/_Perl_PROBLEMATIC_Locale_Foldeds_Start}');
Error('\P{_/a/_Perl_PROBLEMATIC_Locale_Foldeds_Start}');
Expect(1, 64262, '\p{_Perl_Problematic_Locale_Foldeds_Start}', "");
Expect(0, 64262, '\p{^_Perl_Problematic_Locale_Foldeds_Start}', "");
Expect(0, 64262, '\P{_Perl_Problematic_Locale_Foldeds_Start}', "");
Expect(1, 64262, '\P{^_Perl_Problematic_Locale_Foldeds_Start}', "");
Expect(0, 64263, '\p{_Perl_Problematic_Locale_Foldeds_Start}', "");
Expect(1, 64263, '\p{^_Perl_Problematic_Locale_Foldeds_Start}', "");
Expect(1, 64263, '\P{_Perl_Problematic_Locale_Foldeds_Start}', "");
Expect(0, 64263, '\P{^_Perl_Problematic_Locale_Foldeds_Start}', "");
Error('\p{/a/ 	_Perl_PROBLEMATIC_Locale_Foldeds_Start}');
Error('\P{/a/ 	_Perl_PROBLEMATIC_Locale_Foldeds_Start}');
Expect(1, 64262, '\p{_perl_PROBLEMATIC_LOCALE_Foldeds_Start}', "");
Expect(0, 64262, '\p{^_perl_PROBLEMATIC_LOCALE_Foldeds_Start}', "");
Expect(0, 64262, '\P{_perl_PROBLEMATIC_LOCALE_Foldeds_Start}', "");
Expect(1, 64262, '\P{^_perl_PROBLEMATIC_LOCALE_Foldeds_Start}', "");
Expect(0, 64263, '\p{_perl_PROBLEMATIC_LOCALE_Foldeds_Start}', "");
Expect(1, 64263, '\p{^_perl_PROBLEMATIC_LOCALE_Foldeds_Start}', "");
Expect(1, 64263, '\P{_perl_PROBLEMATIC_LOCALE_Foldeds_Start}', "");
Expect(0, 64263, '\P{^_perl_PROBLEMATIC_LOCALE_Foldeds_Start}', "");
Error('\p{ _Perl_PROBLEMATIC_locale_Folds/a/}');
Error('\P{ _Perl_PROBLEMATIC_locale_Folds/a/}');
Expect(1, 64262, '\p{_Perl_PROBLEMATIC_locale_Folds}', "");
Expect(0, 64262, '\p{^_Perl_PROBLEMATIC_locale_Folds}', "");
Expect(0, 64262, '\P{_Perl_PROBLEMATIC_locale_Folds}', "");
Expect(1, 64262, '\P{^_Perl_PROBLEMATIC_locale_Folds}', "");
Expect(0, 64263, '\p{_Perl_PROBLEMATIC_locale_Folds}', "");
Expect(1, 64263, '\p{^_Perl_PROBLEMATIC_locale_Folds}', "");
Expect(1, 64263, '\P{_Perl_PROBLEMATIC_locale_Folds}', "");
Expect(0, 64263, '\P{^_Perl_PROBLEMATIC_locale_Folds}', "");
Error('\p{/a/_perl_PROBLEMATIC_LOCALE_folds}');
Error('\P{/a/_perl_PROBLEMATIC_LOCALE_folds}');
Expect(1, 64262, '\p{_perl_problematic_locale_FOLDS}', "");
Expect(0, 64262, '\p{^_perl_problematic_locale_FOLDS}', "");
Expect(0, 64262, '\P{_perl_problematic_locale_FOLDS}', "");
Expect(1, 64262, '\P{^_perl_problematic_locale_FOLDS}', "");
Expect(0, 64263, '\p{_perl_problematic_locale_FOLDS}', "");
Expect(1, 64263, '\p{^_perl_problematic_locale_FOLDS}', "");
Expect(1, 64263, '\P{_perl_problematic_locale_FOLDS}', "");
Expect(0, 64263, '\P{^_perl_problematic_locale_FOLDS}', "");
Error('\p{	__PERL_Quotemeta:=}');
Error('\P{	__PERL_Quotemeta:=}');
Expect(1, 921599, '\p{_Perl_Quotemeta}', "");
Expect(0, 921599, '\p{^_Perl_Quotemeta}', "");
Expect(0, 921599, '\P{_Perl_Quotemeta}', "");
Expect(1, 921599, '\P{^_Perl_Quotemeta}', "");
Expect(0, 921600, '\p{_Perl_Quotemeta}', "");
Expect(1, 921600, '\p{^_Perl_Quotemeta}', "");
Expect(1, 921600, '\P{_Perl_Quotemeta}', "");
Expect(0, 921600, '\P{^_Perl_Quotemeta}', "");
Error('\p{/a/-__perl_Quotemeta}');
Error('\P{/a/-__perl_Quotemeta}');
Error('\p{/a/_X_Extend}');
Error('\P{/a/_X_Extend}');
Expect(1, 917999, '\p{_X_Extend}', "");
Expect(0, 917999, '\p{^_X_Extend}', "");
Expect(0, 917999, '\P{_X_Extend}', "");
Expect(1, 917999, '\P{^_X_Extend}', "");
Expect(0, 918000, '\p{_X_Extend}', "");
Expect(1, 918000, '\p{^_X_Extend}', "");
Expect(1, 918000, '\P{_X_Extend}', "");
Expect(0, 918000, '\P{^_X_Extend}', "");
Error('\p{-/a/_X_Extend}');
Error('\P{-/a/_X_Extend}');
Expect(1, 917999, '\p{_X_EXTEND}', "");
Expect(0, 917999, '\p{^_X_EXTEND}', "");
Expect(0, 917999, '\P{_X_EXTEND}', "");
Expect(1, 917999, '\P{^_X_EXTEND}', "");
Expect(0, 918000, '\p{_X_EXTEND}', "");
Expect(1, 918000, '\p{^_X_EXTEND}', "");
Expect(1, 918000, '\P{_X_EXTEND}', "");
Expect(0, 918000, '\P{^_X_EXTEND}', "");
Error('\p{ _X_gcb_L/a/}');
Error('\P{ _X_gcb_L/a/}');
Expect(1, 43388, '\p{_x_GCB_L}', "");
Expect(0, 43388, '\p{^_x_GCB_L}', "");
Expect(0, 43388, '\P{_x_GCB_L}', "");
Expect(1, 43388, '\P{^_x_GCB_L}', "");
Expect(0, 43389, '\p{_x_GCB_L}', "");
Expect(1, 43389, '\p{^_x_GCB_L}', "");
Expect(1, 43389, '\P{_x_GCB_L}', "");
Expect(0, 43389, '\P{^_x_GCB_L}', "");
Error('\p{-/a/_x_GCB_L}');
Error('\P{-/a/_x_GCB_L}');
Expect(1, 43388, '\p{_X_GCB_L}', "");
Expect(0, 43388, '\p{^_X_GCB_L}', "");
Expect(0, 43388, '\P{_X_GCB_L}', "");
Expect(1, 43388, '\P{^_X_GCB_L}', "");
Expect(0, 43389, '\p{_X_GCB_L}', "");
Expect(1, 43389, '\p{^_X_GCB_L}', "");
Expect(1, 43389, '\P{_X_GCB_L}', "");
Expect(0, 43389, '\P{^_X_GCB_L}', "");
Error('\p{___x_GCB_lv/a/}');
Error('\P{___x_GCB_lv/a/}');
Expect(1, 55176, '\p{_X_GCB_lv}', "");
Expect(0, 55176, '\p{^_X_GCB_lv}', "");
Expect(0, 55176, '\P{_X_GCB_lv}', "");
Expect(1, 55176, '\P{^_X_GCB_lv}', "");
Expect(0, 55177, '\p{_X_GCB_lv}', "");
Expect(1, 55177, '\p{^_X_GCB_lv}', "");
Expect(1, 55177, '\P{_X_GCB_lv}', "");
Expect(0, 55177, '\P{^_X_GCB_lv}', "");
Error('\p{__x_GCB_LV/a/}');
Error('\P{__x_GCB_LV/a/}');
Expect(1, 55176, '\p{_X_GCB_LV}', "");
Expect(0, 55176, '\p{^_X_GCB_LV}', "");
Expect(0, 55176, '\P{_X_GCB_LV}', "");
Expect(1, 55176, '\P{^_X_GCB_LV}', "");
Expect(0, 55177, '\p{_X_GCB_LV}', "");
Expect(1, 55177, '\p{^_X_GCB_LV}', "");
Expect(1, 55177, '\P{_X_GCB_LV}', "");
Expect(0, 55177, '\P{^_X_GCB_LV}', "");
Error('\p{_/a/_x_GCB_LVT}');
Error('\P{_/a/_x_GCB_LVT}');
Expect(1, 55203, '\p{_X_GCB_LVT}', "");
Expect(0, 55203, '\p{^_X_GCB_LVT}', "");
Expect(0, 55203, '\P{_X_GCB_LVT}', "");
Expect(1, 55203, '\P{^_X_GCB_LVT}', "");
Expect(0, 55204, '\p{_X_GCB_LVT}', "");
Expect(1, 55204, '\p{^_X_GCB_LVT}', "");
Expect(1, 55204, '\P{_X_GCB_LVT}', "");
Expect(0, 55204, '\P{^_X_GCB_LVT}', "");
Error('\p{	:=_X_gcb_LVT}');
Error('\P{	:=_X_gcb_LVT}');
Error('\p{ :=_X_GCB_Prepend}');
Error('\P{ :=_X_GCB_Prepend}');
Expect(0, 1, '\p{_X_GCB_PREPEND}', "");
Expect(1, 1, '\p{^_X_GCB_PREPEND}', "");
Expect(1, 1, '\P{_X_GCB_PREPEND}', "");
Expect(0, 1, '\P{^_X_GCB_PREPEND}', "");
Error('\p{___X_gcb_Prepend:=}');
Error('\P{___X_gcb_Prepend:=}');
Expect(0, 1, '\p{_X_gcb_prepend}', "");
Expect(1, 1, '\p{^_X_gcb_prepend}', "");
Expect(1, 1, '\P{_X_gcb_prepend}', "");
Expect(0, 1, '\P{^_X_gcb_prepend}', "");
Error('\p{_:=_X_GCB_T}');
Error('\P{_:=_X_GCB_T}');
Expect(1, 55291, '\p{_X_GCB_T}', "");
Expect(0, 55291, '\p{^_X_GCB_T}', "");
Expect(0, 55291, '\P{_X_GCB_T}', "");
Expect(1, 55291, '\P{^_X_GCB_T}', "");
Expect(0, 55292, '\p{_X_GCB_T}', "");
Expect(1, 55292, '\p{^_X_GCB_T}', "");
Expect(1, 55292, '\P{_X_GCB_T}', "");
Expect(0, 55292, '\P{^_X_GCB_T}', "");
Error('\p{	/a/_X_gcb_T}');
Error('\P{	/a/_X_gcb_T}');
Expect(1, 55291, '\p{_x_gcb_t}', "");
Expect(0, 55291, '\p{^_x_gcb_t}', "");
Expect(0, 55291, '\P{_x_gcb_t}', "");
Expect(1, 55291, '\P{^_x_gcb_t}', "");
Expect(0, 55292, '\p{_x_gcb_t}', "");
Expect(1, 55292, '\p{^_x_gcb_t}', "");
Expect(1, 55292, '\P{_x_gcb_t}', "");
Expect(0, 55292, '\P{^_x_gcb_t}', "");
Error('\p{_-_x_gcb_V:=}');
Error('\P{_-_x_gcb_V:=}');
Expect(1, 55238, '\p{_X_GCB_V}', "");
Expect(0, 55238, '\p{^_X_GCB_V}', "");
Expect(0, 55238, '\P{_X_GCB_V}', "");
Expect(1, 55238, '\P{^_X_GCB_V}', "");
Expect(0, 55239, '\p{_X_GCB_V}', "");
Expect(1, 55239, '\p{^_X_GCB_V}', "");
Expect(1, 55239, '\P{_X_GCB_V}', "");
Expect(0, 55239, '\P{^_X_GCB_V}', "");
Error('\p{_/a/_X_GCB_V}');
Error('\P{_/a/_X_GCB_V}');
Error('\p{_:=_X_LV_LVT_V}');
Error('\P{_:=_X_LV_LVT_V}');
Expect(1, 55238, '\p{_x_LV_lvt_v}', "");
Expect(0, 55238, '\p{^_x_LV_lvt_v}', "");
Expect(0, 55238, '\P{_x_LV_lvt_v}', "");
Expect(1, 55238, '\P{^_x_LV_lvt_v}', "");
Expect(0, 55239, '\p{_x_LV_lvt_v}', "");
Expect(1, 55239, '\p{^_x_LV_lvt_v}', "");
Expect(1, 55239, '\P{_x_LV_lvt_v}', "");
Expect(0, 55239, '\P{^_x_LV_lvt_v}', "");
Error('\p{__X_LV_LVT_V/a/}');
Error('\P{__X_LV_LVT_V/a/}');
Expect(1, 55238, '\p{_X_LV_LVT_v}', "");
Expect(0, 55238, '\p{^_X_LV_LVT_v}', "");
Expect(0, 55238, '\P{_X_LV_LVT_v}', "");
Expect(1, 55238, '\P{^_X_LV_LVT_v}', "");
Expect(0, 55239, '\p{_X_LV_LVT_v}', "");
Expect(1, 55239, '\p{^_X_LV_LVT_v}', "");
Expect(1, 55239, '\P{_X_LV_LVT_v}', "");
Expect(0, 55239, '\P{^_X_LV_LVT_v}', "");
Error('\p{-_X_RI:=}');
Error('\P{-_X_RI:=}');
Expect(1, 127487, '\p{_X_RI}', "");
Expect(0, 127487, '\p{^_X_RI}', "");
Expect(0, 127487, '\P{_X_RI}', "");
Expect(1, 127487, '\P{^_X_RI}', "");
Expect(0, 127488, '\p{_X_RI}', "");
Expect(1, 127488, '\p{^_X_RI}', "");
Expect(1, 127488, '\P{_X_RI}', "");
Expect(0, 127488, '\P{^_X_RI}', "");
Error('\p{/a/-__X_RI}');
Error('\P{/a/-__X_RI}');
Expect(1, 127487, '\p{_X_ri}', "");
Expect(0, 127487, '\p{^_X_ri}', "");
Expect(0, 127487, '\P{_X_ri}', "");
Expect(1, 127487, '\P{^_X_ri}', "");
Expect(0, 127488, '\p{_X_ri}', "");
Expect(1, 127488, '\p{^_X_ri}', "");
Expect(1, 127488, '\P{_X_ri}', "");
Expect(0, 127488, '\P{^_X_ri}', "");
Error('\p{/a/_X_regular_BEGIN}');
Error('\P{/a/_X_regular_BEGIN}');
Expect(1, 921600, '\p{_x_regular_Begin}', "");
Expect(0, 921600, '\p{^_x_regular_Begin}', "");
Expect(0, 921600, '\P{_x_regular_Begin}', "");
Expect(1, 921600, '\P{^_x_regular_Begin}', "");
Expect(0, 921599, '\p{_x_regular_Begin}', "");
Expect(1, 921599, '\p{^_x_regular_Begin}', "");
Expect(1, 921599, '\P{_x_regular_Begin}', "");
Expect(0, 921599, '\P{^_x_regular_Begin}', "");
Error('\p{ __X_Regular_Begin:=}');
Error('\P{ __X_Regular_Begin:=}');
Expect(1, 921600, '\p{_X_Regular_begin}', "");
Expect(0, 921600, '\p{^_X_Regular_begin}', "");
Expect(0, 921600, '\P{_X_Regular_begin}', "");
Expect(1, 921600, '\P{^_X_Regular_begin}', "");
Expect(0, 921599, '\p{_X_Regular_begin}', "");
Expect(1, 921599, '\p{^_X_Regular_begin}', "");
Expect(1, 921599, '\P{_X_Regular_begin}', "");
Expect(0, 921599, '\P{^_X_Regular_begin}', "");
Error('\p{:=_-_X_Special_BEGIN_start}');
Error('\P{:=_-_X_Special_BEGIN_start}');
Expect(1, 127487, '\p{_X_special_BEGIN_Start}', "");
Expect(0, 127487, '\p{^_X_special_BEGIN_Start}', "");
Expect(0, 127487, '\P{_X_special_BEGIN_Start}', "");
Expect(1, 127487, '\P{^_X_special_BEGIN_Start}', "");
Expect(0, 127488, '\p{_X_special_BEGIN_Start}', "");
Expect(1, 127488, '\p{^_X_special_BEGIN_Start}', "");
Expect(1, 127488, '\P{_X_special_BEGIN_Start}', "");
Expect(0, 127488, '\P{^_X_special_BEGIN_Start}', "");
Error('\p{:=	__x_Special_Begin_Start}');
Error('\P{:=	__x_Special_Begin_Start}');
Expect(1, 127487, '\p{_X_special_Begin_START}', "");
Expect(0, 127487, '\p{^_X_special_Begin_START}', "");
Expect(0, 127487, '\P{_X_special_Begin_START}', "");
Expect(1, 127487, '\P{^_X_special_Begin_START}', "");
Expect(0, 127488, '\p{_X_special_Begin_START}', "");
Expect(1, 127488, '\p{^_X_special_Begin_START}', "");
Expect(1, 127488, '\P{_X_special_Begin_START}', "");
Expect(0, 127488, '\P{^_X_special_Begin_START}', "");
Error('\p{Script=:=--arabic}');
Error('\P{Script=:=--arabic}');
Expect(1, 126705, '\p{Script=arabic}', "");
Expect(0, 126705, '\p{^Script=arabic}', "");
Expect(0, 126705, '\P{Script=arabic}', "");
Expect(1, 126705, '\P{^Script=arabic}', "");
Expect(0, 126706, '\p{Script=arabic}', "");
Expect(1, 126706, '\p{^Script=arabic}', "");
Expect(1, 126706, '\P{Script=arabic}', "");
Expect(0, 126706, '\P{^Script=arabic}', "");
Expect(1, 126705, '\p{Script=- arabic}', "");
Expect(0, 126705, '\p{^Script=- arabic}', "");
Expect(0, 126705, '\P{Script=- arabic}', "");
Expect(1, 126705, '\P{^Script=- arabic}', "");
Expect(0, 126706, '\p{Script=- arabic}', "");
Expect(1, 126706, '\p{^Script=- arabic}', "");
Expect(1, 126706, '\P{Script=- arabic}', "");
Expect(0, 126706, '\P{^Script=- arabic}', "");
Error('\p{Sc=	_ARAB/a/}');
Error('\P{Sc=	_ARAB/a/}');
Expect(1, 126705, '\p{Sc:	arab}', "");
Expect(0, 126705, '\p{^Sc:	arab}', "");
Expect(0, 126705, '\P{Sc:	arab}', "");
Expect(1, 126705, '\P{^Sc:	arab}', "");
Expect(0, 126706, '\p{Sc:	arab}', "");
Expect(1, 126706, '\p{^Sc:	arab}', "");
Expect(1, 126706, '\P{Sc:	arab}', "");
Expect(0, 126706, '\P{^Sc:	arab}', "");
Expect(1, 126705, '\p{Sc=-Arab}', "");
Expect(0, 126705, '\p{^Sc=-Arab}', "");
Expect(0, 126705, '\P{Sc=-Arab}', "");
Expect(1, 126705, '\P{^Sc=-Arab}', "");
Expect(0, 126706, '\p{Sc=-Arab}', "");
Expect(1, 126706, '\p{^Sc=-Arab}', "");
Expect(1, 126706, '\P{Sc=-Arab}', "");
Expect(0, 126706, '\P{^Sc=-Arab}', "");
Error('\p{Is_Script=	arabic/a/}');
Error('\P{Is_Script=	arabic/a/}');
Expect(1, 126705, '\p{Is_Script=arabic}', "");
Expect(0, 126705, '\p{^Is_Script=arabic}', "");
Expect(0, 126705, '\P{Is_Script=arabic}', "");
Expect(1, 126705, '\P{^Is_Script=arabic}', "");
Expect(0, 126706, '\p{Is_Script=arabic}', "");
Expect(1, 126706, '\p{^Is_Script=arabic}', "");
Expect(1, 126706, '\P{Is_Script=arabic}', "");
Expect(0, 126706, '\P{^Is_Script=arabic}', "");
Expect(1, 126705, '\p{Is_Script=-	ARABIC}', "");
Expect(0, 126705, '\p{^Is_Script=-	ARABIC}', "");
Expect(0, 126705, '\P{Is_Script=-	ARABIC}', "");
Expect(1, 126705, '\P{^Is_Script=-	ARABIC}', "");
Expect(0, 126706, '\p{Is_Script=-	ARABIC}', "");
Expect(1, 126706, '\p{^Is_Script=-	ARABIC}', "");
Expect(1, 126706, '\P{Is_Script=-	ARABIC}', "");
Expect(0, 126706, '\P{^Is_Script=-	ARABIC}', "");
Error('\p{Is_Sc:   	-Arab:=}');
Error('\P{Is_Sc:   	-Arab:=}');
Expect(1, 126705, '\p{Is_Sc: arab}', "");
Expect(0, 126705, '\p{^Is_Sc: arab}', "");
Expect(0, 126705, '\P{Is_Sc: arab}', "");
Expect(1, 126705, '\P{^Is_Sc: arab}', "");
Expect(0, 126706, '\p{Is_Sc: arab}', "");
Expect(1, 126706, '\p{^Is_Sc: arab}', "");
Expect(1, 126706, '\P{Is_Sc: arab}', "");
Expect(0, 126706, '\P{^Is_Sc: arab}', "");
Expect(1, 126705, '\p{Is_Sc=-_Arab}', "");
Expect(0, 126705, '\p{^Is_Sc=-_Arab}', "");
Expect(0, 126705, '\P{Is_Sc=-_Arab}', "");
Expect(1, 126705, '\P{^Is_Sc=-_Arab}', "");
Expect(0, 126706, '\p{Is_Sc=-_Arab}', "");
Expect(1, 126706, '\p{^Is_Sc=-_Arab}', "");
Expect(1, 126706, '\P{Is_Sc=-_Arab}', "");
Expect(0, 126706, '\P{^Is_Sc=-_Arab}', "");
Error('\p{Script=:=Imperial_Aramaic}');
Error('\P{Script=:=Imperial_Aramaic}');
Expect(1, 67679, '\p{Script:	imperialaramaic}', "");
Expect(0, 67679, '\p{^Script:	imperialaramaic}', "");
Expect(0, 67679, '\P{Script:	imperialaramaic}', "");
Expect(1, 67679, '\P{^Script:	imperialaramaic}', "");
Expect(0, 67680, '\p{Script:	imperialaramaic}', "");
Expect(1, 67680, '\p{^Script:	imperialaramaic}', "");
Expect(1, 67680, '\P{Script:	imperialaramaic}', "");
Expect(0, 67680, '\P{^Script:	imperialaramaic}', "");
Expect(1, 67679, '\p{Script=	Imperial_Aramaic}', "");
Expect(0, 67679, '\p{^Script=	Imperial_Aramaic}', "");
Expect(0, 67679, '\P{Script=	Imperial_Aramaic}', "");
Expect(1, 67679, '\P{^Script=	Imperial_Aramaic}', "");
Expect(0, 67680, '\p{Script=	Imperial_Aramaic}', "");
Expect(1, 67680, '\p{^Script=	Imperial_Aramaic}', "");
Expect(1, 67680, '\P{Script=	Imperial_Aramaic}', "");
Expect(0, 67680, '\P{^Script=	Imperial_Aramaic}', "");
Error('\p{Sc=	 ARMI:=}');
Error('\P{Sc=	 ARMI:=}');
Expect(1, 67679, '\p{Sc=armi}', "");
Expect(0, 67679, '\p{^Sc=armi}', "");
Expect(0, 67679, '\P{Sc=armi}', "");
Expect(1, 67679, '\P{^Sc=armi}', "");
Expect(0, 67680, '\p{Sc=armi}', "");
Expect(1, 67680, '\p{^Sc=armi}', "");
Expect(1, 67680, '\P{Sc=armi}', "");
Expect(0, 67680, '\P{^Sc=armi}', "");
Expect(1, 67679, '\p{Sc=-armi}', "");
Expect(0, 67679, '\p{^Sc=-armi}', "");
Expect(0, 67679, '\P{Sc=-armi}', "");
Expect(1, 67679, '\P{^Sc=-armi}', "");
Expect(0, 67680, '\p{Sc=-armi}', "");
Expect(1, 67680, '\p{^Sc=-armi}', "");
Expect(1, 67680, '\P{Sc=-armi}', "");
Expect(0, 67680, '\P{^Sc=-armi}', "");
Error('\p{Is_Script=:=		Imperial_Aramaic}');
Error('\P{Is_Script=:=		Imperial_Aramaic}');
Expect(1, 67679, '\p{Is_Script=imperialaramaic}', "");
Expect(0, 67679, '\p{^Is_Script=imperialaramaic}', "");
Expect(0, 67679, '\P{Is_Script=imperialaramaic}', "");
Expect(1, 67679, '\P{^Is_Script=imperialaramaic}', "");
Expect(0, 67680, '\p{Is_Script=imperialaramaic}', "");
Expect(1, 67680, '\p{^Is_Script=imperialaramaic}', "");
Expect(1, 67680, '\P{Is_Script=imperialaramaic}', "");
Expect(0, 67680, '\P{^Is_Script=imperialaramaic}', "");
Expect(1, 67679, '\p{Is_Script=--imperial_Aramaic}', "");
Expect(0, 67679, '\p{^Is_Script=--imperial_Aramaic}', "");
Expect(0, 67679, '\P{Is_Script=--imperial_Aramaic}', "");
Expect(1, 67679, '\P{^Is_Script=--imperial_Aramaic}', "");
Expect(0, 67680, '\p{Is_Script=--imperial_Aramaic}', "");
Expect(1, 67680, '\p{^Is_Script=--imperial_Aramaic}', "");
Expect(1, 67680, '\P{Is_Script=--imperial_Aramaic}', "");
Expect(0, 67680, '\P{^Is_Script=--imperial_Aramaic}', "");
Error('\p{Is_Sc= 	Armi:=}');
Error('\P{Is_Sc= 	Armi:=}');
Expect(1, 67679, '\p{Is_Sc=armi}', "");
Expect(0, 67679, '\p{^Is_Sc=armi}', "");
Expect(0, 67679, '\P{Is_Sc=armi}', "");
Expect(1, 67679, '\P{^Is_Sc=armi}', "");
Expect(0, 67680, '\p{Is_Sc=armi}', "");
Expect(1, 67680, '\p{^Is_Sc=armi}', "");
Expect(1, 67680, '\P{Is_Sc=armi}', "");
Expect(0, 67680, '\P{^Is_Sc=armi}', "");
Expect(1, 67679, '\p{Is_Sc=	 Armi}', "");
Expect(0, 67679, '\p{^Is_Sc=	 Armi}', "");
Expect(0, 67679, '\P{Is_Sc=	 Armi}', "");
Expect(1, 67679, '\P{^Is_Sc=	 Armi}', "");
Expect(0, 67680, '\p{Is_Sc=	 Armi}', "");
Expect(1, 67680, '\p{^Is_Sc=	 Armi}', "");
Expect(1, 67680, '\P{Is_Sc=	 Armi}', "");
Expect(0, 67680, '\P{^Is_Sc=	 Armi}', "");
Error('\p{Script:   :=		armenian}');
Error('\P{Script:   :=		armenian}');
Expect(1, 64279, '\p{Script: armenian}', "");
Expect(0, 64279, '\p{^Script: armenian}', "");
Expect(0, 64279, '\P{Script: armenian}', "");
Expect(1, 64279, '\P{^Script: armenian}', "");
Expect(0, 64280, '\p{Script: armenian}', "");
Expect(1, 64280, '\p{^Script: armenian}', "");
Expect(1, 64280, '\P{Script: armenian}', "");
Expect(0, 64280, '\P{^Script: armenian}', "");
Expect(1, 64279, '\p{Script=	-ARMENIAN}', "");
Expect(0, 64279, '\p{^Script=	-ARMENIAN}', "");
Expect(0, 64279, '\P{Script=	-ARMENIAN}', "");
Expect(1, 64279, '\P{^Script=	-ARMENIAN}', "");
Expect(0, 64280, '\p{Script=	-ARMENIAN}', "");
Expect(1, 64280, '\p{^Script=	-ARMENIAN}', "");
Expect(1, 64280, '\P{Script=	-ARMENIAN}', "");
Expect(0, 64280, '\P{^Script=	-ARMENIAN}', "");
Error('\p{Sc=/a/__Armn}');
Error('\P{Sc=/a/__Armn}');
Expect(1, 64279, '\p{Sc=armn}', "");
Expect(0, 64279, '\p{^Sc=armn}', "");
Expect(0, 64279, '\P{Sc=armn}', "");
Expect(1, 64279, '\P{^Sc=armn}', "");
Expect(0, 64280, '\p{Sc=armn}', "");
Expect(1, 64280, '\p{^Sc=armn}', "");
Expect(1, 64280, '\P{Sc=armn}', "");
Expect(0, 64280, '\P{^Sc=armn}', "");
Expect(1, 64279, '\p{Sc:	-	Armn}', "");
Expect(0, 64279, '\p{^Sc:	-	Armn}', "");
Expect(0, 64279, '\P{Sc:	-	Armn}', "");
Expect(1, 64279, '\P{^Sc:	-	Armn}', "");
Expect(0, 64280, '\p{Sc:	-	Armn}', "");
Expect(1, 64280, '\p{^Sc:	-	Armn}', "");
Expect(1, 64280, '\P{Sc:	-	Armn}', "");
Expect(0, 64280, '\P{^Sc:	-	Armn}', "");
Error('\p{Is_Script=_:=armenian}');
Error('\P{Is_Script=_:=armenian}');
Expect(1, 64279, '\p{Is_Script=armenian}', "");
Expect(0, 64279, '\p{^Is_Script=armenian}', "");
Expect(0, 64279, '\P{Is_Script=armenian}', "");
Expect(1, 64279, '\P{^Is_Script=armenian}', "");
Expect(0, 64280, '\p{Is_Script=armenian}', "");
Expect(1, 64280, '\p{^Is_Script=armenian}', "");
Expect(1, 64280, '\P{Is_Script=armenian}', "");
Expect(0, 64280, '\P{^Is_Script=armenian}', "");
Expect(1, 64279, '\p{Is_Script=_-Armenian}', "");
Expect(0, 64279, '\p{^Is_Script=_-Armenian}', "");
Expect(0, 64279, '\P{Is_Script=_-Armenian}', "");
Expect(1, 64279, '\P{^Is_Script=_-Armenian}', "");
Expect(0, 64280, '\p{Is_Script=_-Armenian}', "");
Expect(1, 64280, '\p{^Is_Script=_-Armenian}', "");
Expect(1, 64280, '\P{Is_Script=_-Armenian}', "");
Expect(0, 64280, '\P{^Is_Script=_-Armenian}', "");
Error('\p{Is_Sc=/a/	Armn}');
Error('\P{Is_Sc=/a/	Armn}');
Expect(1, 64279, '\p{Is_Sc=armn}', "");
Expect(0, 64279, '\p{^Is_Sc=armn}', "");
Expect(0, 64279, '\P{Is_Sc=armn}', "");
Expect(1, 64279, '\P{^Is_Sc=armn}', "");
Expect(0, 64280, '\p{Is_Sc=armn}', "");
Expect(1, 64280, '\p{^Is_Sc=armn}', "");
Expect(1, 64280, '\P{Is_Sc=armn}', "");
Expect(0, 64280, '\P{^Is_Sc=armn}', "");
Expect(1, 64279, '\p{Is_Sc:	Armn}', "");
Expect(0, 64279, '\p{^Is_Sc:	Armn}', "");
Expect(0, 64279, '\P{Is_Sc:	Armn}', "");
Expect(1, 64279, '\P{^Is_Sc:	Armn}', "");
Expect(0, 64280, '\p{Is_Sc:	Armn}', "");
Expect(1, 64280, '\p{^Is_Sc:	Armn}', "");
Expect(1, 64280, '\P{Is_Sc:	Armn}', "");
Expect(0, 64280, '\P{^Is_Sc:	Armn}', "");
Error('\p{Script= -Avestan/a/}');
Error('\P{Script= -Avestan/a/}');
Expect(1, 68415, '\p{Script=avestan}', "");
Expect(0, 68415, '\p{^Script=avestan}', "");
Expect(0, 68415, '\P{Script=avestan}', "");
Expect(1, 68415, '\P{^Script=avestan}', "");
Expect(0, 68416, '\p{Script=avestan}', "");
Expect(1, 68416, '\p{^Script=avestan}', "");
Expect(1, 68416, '\P{Script=avestan}', "");
Expect(0, 68416, '\P{^Script=avestan}', "");
Expect(1, 68415, '\p{Script=__avestan}', "");
Expect(0, 68415, '\p{^Script=__avestan}', "");
Expect(0, 68415, '\P{Script=__avestan}', "");
Expect(1, 68415, '\P{^Script=__avestan}', "");
Expect(0, 68416, '\p{Script=__avestan}', "");
Expect(1, 68416, '\p{^Script=__avestan}', "");
Expect(1, 68416, '\P{Script=__avestan}', "");
Expect(0, 68416, '\P{^Script=__avestan}', "");
Error('\p{Sc= Avst/a/}');
Error('\P{Sc= Avst/a/}');
Expect(1, 68415, '\p{Sc=avst}', "");
Expect(0, 68415, '\p{^Sc=avst}', "");
Expect(0, 68415, '\P{Sc=avst}', "");
Expect(1, 68415, '\P{^Sc=avst}', "");
Expect(0, 68416, '\p{Sc=avst}', "");
Expect(1, 68416, '\p{^Sc=avst}', "");
Expect(1, 68416, '\P{Sc=avst}', "");
Expect(0, 68416, '\P{^Sc=avst}', "");
Expect(1, 68415, '\p{Sc=_ Avst}', "");
Expect(0, 68415, '\p{^Sc=_ Avst}', "");
Expect(0, 68415, '\P{Sc=_ Avst}', "");
Expect(1, 68415, '\P{^Sc=_ Avst}', "");
Expect(0, 68416, '\p{Sc=_ Avst}', "");
Expect(1, 68416, '\p{^Sc=_ Avst}', "");
Expect(1, 68416, '\P{Sc=_ Avst}', "");
Expect(0, 68416, '\P{^Sc=_ Avst}', "");
Error('\p{Is_Script=-/a/AVESTAN}');
Error('\P{Is_Script=-/a/AVESTAN}');
Expect(1, 68415, '\p{Is_Script=avestan}', "");
Expect(0, 68415, '\p{^Is_Script=avestan}', "");
Expect(0, 68415, '\P{Is_Script=avestan}', "");
Expect(1, 68415, '\P{^Is_Script=avestan}', "");
Expect(0, 68416, '\p{Is_Script=avestan}', "");
Expect(1, 68416, '\p{^Is_Script=avestan}', "");
Expect(1, 68416, '\P{Is_Script=avestan}', "");
Expect(0, 68416, '\P{^Is_Script=avestan}', "");
Expect(1, 68415, '\p{Is_Script=-AVESTAN}', "");
Expect(0, 68415, '\p{^Is_Script=-AVESTAN}', "");
Expect(0, 68415, '\P{Is_Script=-AVESTAN}', "");
Expect(1, 68415, '\P{^Is_Script=-AVESTAN}', "");
Expect(0, 68416, '\p{Is_Script=-AVESTAN}', "");
Expect(1, 68416, '\p{^Is_Script=-AVESTAN}', "");
Expect(1, 68416, '\P{Is_Script=-AVESTAN}', "");
Expect(0, 68416, '\P{^Is_Script=-AVESTAN}', "");
Error('\p{Is_Sc=__avst/a/}');
Error('\P{Is_Sc=__avst/a/}');
Expect(1, 68415, '\p{Is_Sc=avst}', "");
Expect(0, 68415, '\p{^Is_Sc=avst}', "");
Expect(0, 68415, '\P{Is_Sc=avst}', "");
Expect(1, 68415, '\P{^Is_Sc=avst}', "");
Expect(0, 68416, '\p{Is_Sc=avst}', "");
Expect(1, 68416, '\p{^Is_Sc=avst}', "");
Expect(1, 68416, '\P{Is_Sc=avst}', "");
Expect(0, 68416, '\P{^Is_Sc=avst}', "");
Expect(1, 68415, '\p{Is_Sc=	-AVST}', "");
Expect(0, 68415, '\p{^Is_Sc=	-AVST}', "");
Expect(0, 68415, '\P{Is_Sc=	-AVST}', "");
Expect(1, 68415, '\P{^Is_Sc=	-AVST}', "");
Expect(0, 68416, '\p{Is_Sc=	-AVST}', "");
Expect(1, 68416, '\p{^Is_Sc=	-AVST}', "");
Expect(1, 68416, '\P{Is_Sc=	-AVST}', "");
Expect(0, 68416, '\P{^Is_Sc=	-AVST}', "");
Error('\p{Script:/a/- balinese}');
Error('\P{Script:/a/- balinese}');
Expect(1, 7036, '\p{Script=balinese}', "");
Expect(0, 7036, '\p{^Script=balinese}', "");
Expect(0, 7036, '\P{Script=balinese}', "");
Expect(1, 7036, '\P{^Script=balinese}', "");
Expect(0, 7037, '\p{Script=balinese}', "");
Expect(1, 7037, '\p{^Script=balinese}', "");
Expect(1, 7037, '\P{Script=balinese}', "");
Expect(0, 7037, '\P{^Script=balinese}', "");
Expect(1, 7036, '\p{Script:	 Balinese}', "");
Expect(0, 7036, '\p{^Script:	 Balinese}', "");
Expect(0, 7036, '\P{Script:	 Balinese}', "");
Expect(1, 7036, '\P{^Script:	 Balinese}', "");
Expect(0, 7037, '\p{Script:	 Balinese}', "");
Expect(1, 7037, '\p{^Script:	 Balinese}', "");
Expect(1, 7037, '\P{Script:	 Balinese}', "");
Expect(0, 7037, '\P{^Script:	 Balinese}', "");
Error('\p{Sc=	_Bali/a/}');
Error('\P{Sc=	_Bali/a/}');
Expect(1, 7036, '\p{Sc=bali}', "");
Expect(0, 7036, '\p{^Sc=bali}', "");
Expect(0, 7036, '\P{Sc=bali}', "");
Expect(1, 7036, '\P{^Sc=bali}', "");
Expect(0, 7037, '\p{Sc=bali}', "");
Expect(1, 7037, '\p{^Sc=bali}', "");
Expect(1, 7037, '\P{Sc=bali}', "");
Expect(0, 7037, '\P{^Sc=bali}', "");
Expect(1, 7036, '\p{Sc=  BALI}', "");
Expect(0, 7036, '\p{^Sc=  BALI}', "");
Expect(0, 7036, '\P{Sc=  BALI}', "");
Expect(1, 7036, '\P{^Sc=  BALI}', "");
Expect(0, 7037, '\p{Sc=  BALI}', "");
Expect(1, 7037, '\p{^Sc=  BALI}', "");
Expect(1, 7037, '\P{Sc=  BALI}', "");
Expect(0, 7037, '\P{^Sc=  BALI}', "");
Error('\p{Is_Script= :=balinese}');
Error('\P{Is_Script= :=balinese}');
Expect(1, 7036, '\p{Is_Script=balinese}', "");
Expect(0, 7036, '\p{^Is_Script=balinese}', "");
Expect(0, 7036, '\P{Is_Script=balinese}', "");
Expect(1, 7036, '\P{^Is_Script=balinese}', "");
Expect(0, 7037, '\p{Is_Script=balinese}', "");
Expect(1, 7037, '\p{^Is_Script=balinese}', "");
Expect(1, 7037, '\P{Is_Script=balinese}', "");
Expect(0, 7037, '\P{^Is_Script=balinese}', "");
Expect(1, 7036, '\p{Is_Script=	_balinese}', "");
Expect(0, 7036, '\p{^Is_Script=	_balinese}', "");
Expect(0, 7036, '\P{Is_Script=	_balinese}', "");
Expect(1, 7036, '\P{^Is_Script=	_balinese}', "");
Expect(0, 7037, '\p{Is_Script=	_balinese}', "");
Expect(1, 7037, '\p{^Is_Script=	_balinese}', "");
Expect(1, 7037, '\P{Is_Script=	_balinese}', "");
Expect(0, 7037, '\P{^Is_Script=	_balinese}', "");
Error('\p{Is_Sc=:=bali}');
Error('\P{Is_Sc=:=bali}');
Expect(1, 7036, '\p{Is_Sc=bali}', "");
Expect(0, 7036, '\p{^Is_Sc=bali}', "");
Expect(0, 7036, '\P{Is_Sc=bali}', "");
Expect(1, 7036, '\P{^Is_Sc=bali}', "");
Expect(0, 7037, '\p{Is_Sc=bali}', "");
Expect(1, 7037, '\p{^Is_Sc=bali}', "");
Expect(1, 7037, '\P{Is_Sc=bali}', "");
Expect(0, 7037, '\P{^Is_Sc=bali}', "");
Expect(1, 7036, '\p{Is_Sc= _Bali}', "");
Expect(0, 7036, '\p{^Is_Sc= _Bali}', "");
Expect(0, 7036, '\P{Is_Sc= _Bali}', "");
Expect(1, 7036, '\P{^Is_Sc= _Bali}', "");
Expect(0, 7037, '\p{Is_Sc= _Bali}', "");
Expect(1, 7037, '\p{^Is_Sc= _Bali}', "");
Expect(1, 7037, '\P{Is_Sc= _Bali}', "");
Expect(0, 7037, '\P{^Is_Sc= _Bali}', "");
Error('\p{Script= -BAMUM/a/}');
Error('\P{Script= -BAMUM/a/}');
Expect(1, 92728, '\p{Script:	bamum}', "");
Expect(0, 92728, '\p{^Script:	bamum}', "");
Expect(0, 92728, '\P{Script:	bamum}', "");
Expect(1, 92728, '\P{^Script:	bamum}', "");
Expect(0, 92729, '\p{Script:	bamum}', "");
Expect(1, 92729, '\p{^Script:	bamum}', "");
Expect(1, 92729, '\P{Script:	bamum}', "");
Expect(0, 92729, '\P{^Script:	bamum}', "");
Expect(1, 92728, '\p{Script=__Bamum}', "");
Expect(0, 92728, '\p{^Script=__Bamum}', "");
Expect(0, 92728, '\P{Script=__Bamum}', "");
Expect(1, 92728, '\P{^Script=__Bamum}', "");
Expect(0, 92729, '\p{Script=__Bamum}', "");
Expect(1, 92729, '\p{^Script=__Bamum}', "");
Expect(1, 92729, '\P{Script=__Bamum}', "");
Expect(0, 92729, '\P{^Script=__Bamum}', "");
Error('\p{Sc=:=		BAMU}');
Error('\P{Sc=:=		BAMU}');
Expect(1, 92728, '\p{Sc=bamu}', "");
Expect(0, 92728, '\p{^Sc=bamu}', "");
Expect(0, 92728, '\P{Sc=bamu}', "");
Expect(1, 92728, '\P{^Sc=bamu}', "");
Expect(0, 92729, '\p{Sc=bamu}', "");
Expect(1, 92729, '\p{^Sc=bamu}', "");
Expect(1, 92729, '\P{Sc=bamu}', "");
Expect(0, 92729, '\P{^Sc=bamu}', "");
Expect(1, 92728, '\p{Sc:	 -BAMU}', "");
Expect(0, 92728, '\p{^Sc:	 -BAMU}', "");
Expect(0, 92728, '\P{Sc:	 -BAMU}', "");
Expect(1, 92728, '\P{^Sc:	 -BAMU}', "");
Expect(0, 92729, '\p{Sc:	 -BAMU}', "");
Expect(1, 92729, '\p{^Sc:	 -BAMU}', "");
Expect(1, 92729, '\P{Sc:	 -BAMU}', "");
Expect(0, 92729, '\P{^Sc:	 -BAMU}', "");
Error('\p{Is_Script=_:=bamum}');
Error('\P{Is_Script=_:=bamum}');
Expect(1, 92728, '\p{Is_Script=bamum}', "");
Expect(0, 92728, '\p{^Is_Script=bamum}', "");
Expect(0, 92728, '\P{Is_Script=bamum}', "");
Expect(1, 92728, '\P{^Is_Script=bamum}', "");
Expect(0, 92729, '\p{Is_Script=bamum}', "");
Expect(1, 92729, '\p{^Is_Script=bamum}', "");
Expect(1, 92729, '\P{Is_Script=bamum}', "");
Expect(0, 92729, '\P{^Is_Script=bamum}', "");
Expect(1, 92728, '\p{Is_Script= Bamum}', "");
Expect(0, 92728, '\p{^Is_Script= Bamum}', "");
Expect(0, 92728, '\P{Is_Script= Bamum}', "");
Expect(1, 92728, '\P{^Is_Script= Bamum}', "");
Expect(0, 92729, '\p{Is_Script= Bamum}', "");
Expect(1, 92729, '\p{^Is_Script= Bamum}', "");
Expect(1, 92729, '\P{Is_Script= Bamum}', "");
Expect(0, 92729, '\P{^Is_Script= Bamum}', "");
Error('\p{Is_Sc=:=Bamu}');
Error('\P{Is_Sc=:=Bamu}');
Expect(1, 92728, '\p{Is_Sc=bamu}', "");
Expect(0, 92728, '\p{^Is_Sc=bamu}', "");
Expect(0, 92728, '\P{Is_Sc=bamu}', "");
Expect(1, 92728, '\P{^Is_Sc=bamu}', "");
Expect(0, 92729, '\p{Is_Sc=bamu}', "");
Expect(1, 92729, '\p{^Is_Sc=bamu}', "");
Expect(1, 92729, '\P{Is_Sc=bamu}', "");
Expect(0, 92729, '\P{^Is_Sc=bamu}', "");
Expect(1, 92728, '\p{Is_Sc= Bamu}', "");
Expect(0, 92728, '\p{^Is_Sc= Bamu}', "");
Expect(0, 92728, '\P{Is_Sc= Bamu}', "");
Expect(1, 92728, '\P{^Is_Sc= Bamu}', "");
Expect(0, 92729, '\p{Is_Sc= Bamu}', "");
Expect(1, 92729, '\p{^Is_Sc= Bamu}', "");
Expect(1, 92729, '\P{Is_Sc= Bamu}', "");
Expect(0, 92729, '\P{^Is_Sc= Bamu}', "");
Error('\p{Script=-_Batak:=}');
Error('\P{Script=-_Batak:=}');
Expect(1, 7167, '\p{Script:   batak}', "");
Expect(0, 7167, '\p{^Script:   batak}', "");
Expect(0, 7167, '\P{Script:   batak}', "");
Expect(1, 7167, '\P{^Script:   batak}', "");
Expect(0, 7168, '\p{Script:   batak}', "");
Expect(1, 7168, '\p{^Script:   batak}', "");
Expect(1, 7168, '\P{Script:   batak}', "");
Expect(0, 7168, '\P{^Script:   batak}', "");
Expect(1, 7167, '\p{Script=_	batak}', "");
Expect(0, 7167, '\p{^Script=_	batak}', "");
Expect(0, 7167, '\P{Script=_	batak}', "");
Expect(1, 7167, '\P{^Script=_	batak}', "");
Expect(0, 7168, '\p{Script=_	batak}', "");
Expect(1, 7168, '\p{^Script=_	batak}', "");
Expect(1, 7168, '\P{Script=_	batak}', "");
Expect(0, 7168, '\P{^Script=_	batak}', "");
Error('\p{Sc=_Batk/a/}');
Error('\P{Sc=_Batk/a/}');
Expect(1, 7167, '\p{Sc=batk}', "");
Expect(0, 7167, '\p{^Sc=batk}', "");
Expect(0, 7167, '\P{Sc=batk}', "");
Expect(1, 7167, '\P{^Sc=batk}', "");
Expect(0, 7168, '\p{Sc=batk}', "");
Expect(1, 7168, '\p{^Sc=batk}', "");
Expect(1, 7168, '\P{Sc=batk}', "");
Expect(0, 7168, '\P{^Sc=batk}', "");
Expect(1, 7167, '\p{Sc=--batk}', "");
Expect(0, 7167, '\p{^Sc=--batk}', "");
Expect(0, 7167, '\P{Sc=--batk}', "");
Expect(1, 7167, '\P{^Sc=--batk}', "");
Expect(0, 7168, '\p{Sc=--batk}', "");
Expect(1, 7168, '\p{^Sc=--batk}', "");
Expect(1, 7168, '\P{Sc=--batk}', "");
Expect(0, 7168, '\P{^Sc=--batk}', "");
Error('\p{Is_Script=-:=Batak}');
Error('\P{Is_Script=-:=Batak}');
Expect(1, 7167, '\p{Is_Script=batak}', "");
Expect(0, 7167, '\p{^Is_Script=batak}', "");
Expect(0, 7167, '\P{Is_Script=batak}', "");
Expect(1, 7167, '\P{^Is_Script=batak}', "");
Expect(0, 7168, '\p{Is_Script=batak}', "");
Expect(1, 7168, '\p{^Is_Script=batak}', "");
Expect(1, 7168, '\P{Is_Script=batak}', "");
Expect(0, 7168, '\P{^Is_Script=batak}', "");
Expect(1, 7167, '\p{Is_Script=_	BATAK}', "");
Expect(0, 7167, '\p{^Is_Script=_	BATAK}', "");
Expect(0, 7167, '\P{Is_Script=_	BATAK}', "");
Expect(1, 7167, '\P{^Is_Script=_	BATAK}', "");
Expect(0, 7168, '\p{Is_Script=_	BATAK}', "");
Expect(1, 7168, '\p{^Is_Script=_	BATAK}', "");
Expect(1, 7168, '\P{Is_Script=_	BATAK}', "");
Expect(0, 7168, '\P{^Is_Script=_	BATAK}', "");
Error('\p{Is_Sc=:=batk}');
Error('\P{Is_Sc=:=batk}');
Expect(1, 7167, '\p{Is_Sc=batk}', "");
Expect(0, 7167, '\p{^Is_Sc=batk}', "");
Expect(0, 7167, '\P{Is_Sc=batk}', "");
Expect(1, 7167, '\P{^Is_Sc=batk}', "");
Expect(0, 7168, '\p{Is_Sc=batk}', "");
Expect(1, 7168, '\p{^Is_Sc=batk}', "");
Expect(1, 7168, '\P{Is_Sc=batk}', "");
Expect(0, 7168, '\P{^Is_Sc=batk}', "");
Expect(1, 7167, '\p{Is_Sc: Batk}', "");
Expect(0, 7167, '\p{^Is_Sc: Batk}', "");
Expect(0, 7167, '\P{Is_Sc: Batk}', "");
Expect(1, 7167, '\P{^Is_Sc: Batk}', "");
Expect(0, 7168, '\p{Is_Sc: Batk}', "");
Expect(1, 7168, '\p{^Is_Sc: Batk}', "");
Expect(1, 7168, '\P{Is_Sc: Batk}', "");
Expect(0, 7168, '\P{^Is_Sc: Batk}', "");
Error('\p{Script=:=__BENGALI}');
Error('\P{Script=:=__BENGALI}');
Expect(1, 2555, '\p{Script=bengali}', "");
Expect(0, 2555, '\p{^Script=bengali}', "");
Expect(0, 2555, '\P{Script=bengali}', "");
Expect(1, 2555, '\P{^Script=bengali}', "");
Expect(0, 2556, '\p{Script=bengali}', "");
Expect(1, 2556, '\p{^Script=bengali}', "");
Expect(1, 2556, '\P{Script=bengali}', "");
Expect(0, 2556, '\P{^Script=bengali}', "");
Expect(1, 2555, '\p{Script:	 _bengali}', "");
Expect(0, 2555, '\p{^Script:	 _bengali}', "");
Expect(0, 2555, '\P{Script:	 _bengali}', "");
Expect(1, 2555, '\P{^Script:	 _bengali}', "");
Expect(0, 2556, '\p{Script:	 _bengali}', "");
Expect(1, 2556, '\p{^Script:	 _bengali}', "");
Expect(1, 2556, '\P{Script:	 _bengali}', "");
Expect(0, 2556, '\P{^Script:	 _bengali}', "");
Error('\p{Sc= -BENG:=}');
Error('\P{Sc= -BENG:=}');
Expect(1, 2555, '\p{Sc=beng}', "");
Expect(0, 2555, '\p{^Sc=beng}', "");
Expect(0, 2555, '\P{Sc=beng}', "");
Expect(1, 2555, '\P{^Sc=beng}', "");
Expect(0, 2556, '\p{Sc=beng}', "");
Expect(1, 2556, '\p{^Sc=beng}', "");
Expect(1, 2556, '\P{Sc=beng}', "");
Expect(0, 2556, '\P{^Sc=beng}', "");
Expect(1, 2555, '\p{Sc=	Beng}', "");
Expect(0, 2555, '\p{^Sc=	Beng}', "");
Expect(0, 2555, '\P{Sc=	Beng}', "");
Expect(1, 2555, '\P{^Sc=	Beng}', "");
Expect(0, 2556, '\p{Sc=	Beng}', "");
Expect(1, 2556, '\p{^Sc=	Beng}', "");
Expect(1, 2556, '\P{Sc=	Beng}', "");
Expect(0, 2556, '\P{^Sc=	Beng}', "");
Error('\p{Is_Script:		 BENGALI/a/}');
Error('\P{Is_Script:		 BENGALI/a/}');
Expect(1, 2555, '\p{Is_Script=bengali}', "");
Expect(0, 2555, '\p{^Is_Script=bengali}', "");
Expect(0, 2555, '\P{Is_Script=bengali}', "");
Expect(1, 2555, '\P{^Is_Script=bengali}', "");
Expect(0, 2556, '\p{Is_Script=bengali}', "");
Expect(1, 2556, '\p{^Is_Script=bengali}', "");
Expect(1, 2556, '\P{Is_Script=bengali}', "");
Expect(0, 2556, '\P{^Is_Script=bengali}', "");
Expect(1, 2555, '\p{Is_Script=	-BENGALI}', "");
Expect(0, 2555, '\p{^Is_Script=	-BENGALI}', "");
Expect(0, 2555, '\P{Is_Script=	-BENGALI}', "");
Expect(1, 2555, '\P{^Is_Script=	-BENGALI}', "");
Expect(0, 2556, '\p{Is_Script=	-BENGALI}', "");
Expect(1, 2556, '\p{^Is_Script=	-BENGALI}', "");
Expect(1, 2556, '\P{Is_Script=	-BENGALI}', "");
Expect(0, 2556, '\P{^Is_Script=	-BENGALI}', "");
Error('\p{Is_Sc=:= _Beng}');
Error('\P{Is_Sc=:= _Beng}');
Expect(1, 2555, '\p{Is_Sc=beng}', "");
Expect(0, 2555, '\p{^Is_Sc=beng}', "");
Expect(0, 2555, '\P{Is_Sc=beng}', "");
Expect(1, 2555, '\P{^Is_Sc=beng}', "");
Expect(0, 2556, '\p{Is_Sc=beng}', "");
Expect(1, 2556, '\p{^Is_Sc=beng}', "");
Expect(1, 2556, '\P{Is_Sc=beng}', "");
Expect(0, 2556, '\P{^Is_Sc=beng}', "");
Expect(1, 2555, '\p{Is_Sc= Beng}', "");
Expect(0, 2555, '\p{^Is_Sc= Beng}', "");
Expect(0, 2555, '\P{Is_Sc= Beng}', "");
Expect(1, 2555, '\P{^Is_Sc= Beng}', "");
Expect(0, 2556, '\p{Is_Sc= Beng}', "");
Expect(1, 2556, '\p{^Is_Sc= Beng}', "");
Expect(1, 2556, '\P{Is_Sc= Beng}', "");
Expect(0, 2556, '\P{^Is_Sc= Beng}', "");
Error('\p{Script=/a/ -bopomofo}');
Error('\P{Script=/a/ -bopomofo}');
Expect(1, 12730, '\p{Script=bopomofo}', "");
Expect(0, 12730, '\p{^Script=bopomofo}', "");
Expect(0, 12730, '\P{Script=bopomofo}', "");
Expect(1, 12730, '\P{^Script=bopomofo}', "");
Expect(0, 12731, '\p{Script=bopomofo}', "");
Expect(1, 12731, '\p{^Script=bopomofo}', "");
Expect(1, 12731, '\P{Script=bopomofo}', "");
Expect(0, 12731, '\P{^Script=bopomofo}', "");
Expect(1, 12730, '\p{Script=	 BOPOMOFO}', "");
Expect(0, 12730, '\p{^Script=	 BOPOMOFO}', "");
Expect(0, 12730, '\P{Script=	 BOPOMOFO}', "");
Expect(1, 12730, '\P{^Script=	 BOPOMOFO}', "");
Expect(0, 12731, '\p{Script=	 BOPOMOFO}', "");
Expect(1, 12731, '\p{^Script=	 BOPOMOFO}', "");
Expect(1, 12731, '\P{Script=	 BOPOMOFO}', "");
Expect(0, 12731, '\P{^Script=	 BOPOMOFO}', "");
Error('\p{Sc= /a/Bopo}');
Error('\P{Sc= /a/Bopo}');
Expect(1, 12730, '\p{Sc: bopo}', "");
Expect(0, 12730, '\p{^Sc: bopo}', "");
Expect(0, 12730, '\P{Sc: bopo}', "");
Expect(1, 12730, '\P{^Sc: bopo}', "");
Expect(0, 12731, '\p{Sc: bopo}', "");
Expect(1, 12731, '\p{^Sc: bopo}', "");
Expect(1, 12731, '\P{Sc: bopo}', "");
Expect(0, 12731, '\P{^Sc: bopo}', "");
Expect(1, 12730, '\p{Sc=_Bopo}', "");
Expect(0, 12730, '\p{^Sc=_Bopo}', "");
Expect(0, 12730, '\P{Sc=_Bopo}', "");
Expect(1, 12730, '\P{^Sc=_Bopo}', "");
Expect(0, 12731, '\p{Sc=_Bopo}', "");
Expect(1, 12731, '\p{^Sc=_Bopo}', "");
Expect(1, 12731, '\P{Sc=_Bopo}', "");
Expect(0, 12731, '\P{^Sc=_Bopo}', "");
Error('\p{Is_Script=	/a/Bopomofo}');
Error('\P{Is_Script=	/a/Bopomofo}');
Expect(1, 12730, '\p{Is_Script=bopomofo}', "");
Expect(0, 12730, '\p{^Is_Script=bopomofo}', "");
Expect(0, 12730, '\P{Is_Script=bopomofo}', "");
Expect(1, 12730, '\P{^Is_Script=bopomofo}', "");
Expect(0, 12731, '\p{Is_Script=bopomofo}', "");
Expect(1, 12731, '\p{^Is_Script=bopomofo}', "");
Expect(1, 12731, '\P{Is_Script=bopomofo}', "");
Expect(0, 12731, '\P{^Is_Script=bopomofo}', "");
Expect(1, 12730, '\p{Is_Script= -Bopomofo}', "");
Expect(0, 12730, '\p{^Is_Script= -Bopomofo}', "");
Expect(0, 12730, '\P{Is_Script= -Bopomofo}', "");
Expect(1, 12730, '\P{^Is_Script= -Bopomofo}', "");
Expect(0, 12731, '\p{Is_Script= -Bopomofo}', "");
Expect(1, 12731, '\p{^Is_Script= -Bopomofo}', "");
Expect(1, 12731, '\P{Is_Script= -Bopomofo}', "");
Expect(0, 12731, '\P{^Is_Script= -Bopomofo}', "");
Error('\p{Is_Sc=_Bopo:=}');
Error('\P{Is_Sc=_Bopo:=}');
Expect(1, 12730, '\p{Is_Sc=bopo}', "");
Expect(0, 12730, '\p{^Is_Sc=bopo}', "");
Expect(0, 12730, '\P{Is_Sc=bopo}', "");
Expect(1, 12730, '\P{^Is_Sc=bopo}', "");
Expect(0, 12731, '\p{Is_Sc=bopo}', "");
Expect(1, 12731, '\p{^Is_Sc=bopo}', "");
Expect(1, 12731, '\P{Is_Sc=bopo}', "");
Expect(0, 12731, '\P{^Is_Sc=bopo}', "");
Expect(1, 12730, '\p{Is_Sc:  Bopo}', "");
Expect(0, 12730, '\p{^Is_Sc:  Bopo}', "");
Expect(0, 12730, '\P{Is_Sc:  Bopo}', "");
Expect(1, 12730, '\P{^Is_Sc:  Bopo}', "");
Expect(0, 12731, '\p{Is_Sc:  Bopo}', "");
Expect(1, 12731, '\p{^Is_Sc:  Bopo}', "");
Expect(1, 12731, '\P{Is_Sc:  Bopo}', "");
Expect(0, 12731, '\P{^Is_Sc:  Bopo}', "");
Error('\p{Script:	/a/BRAHMI}');
Error('\P{Script:	/a/BRAHMI}');
Expect(1, 69743, '\p{Script=brahmi}', "");
Expect(0, 69743, '\p{^Script=brahmi}', "");
Expect(0, 69743, '\P{Script=brahmi}', "");
Expect(1, 69743, '\P{^Script=brahmi}', "");
Expect(0, 69744, '\p{Script=brahmi}', "");
Expect(1, 69744, '\p{^Script=brahmi}', "");
Expect(1, 69744, '\P{Script=brahmi}', "");
Expect(0, 69744, '\P{^Script=brahmi}', "");
Expect(1, 69743, '\p{Script=-_BRAHMI}', "");
Expect(0, 69743, '\p{^Script=-_BRAHMI}', "");
Expect(0, 69743, '\P{Script=-_BRAHMI}', "");
Expect(1, 69743, '\P{^Script=-_BRAHMI}', "");
Expect(0, 69744, '\p{Script=-_BRAHMI}', "");
Expect(1, 69744, '\p{^Script=-_BRAHMI}', "");
Expect(1, 69744, '\P{Script=-_BRAHMI}', "");
Expect(0, 69744, '\P{^Script=-_BRAHMI}', "");
Error('\p{Sc= :=brah}');
Error('\P{Sc= :=brah}');
Expect(1, 69743, '\p{Sc=brah}', "");
Expect(0, 69743, '\p{^Sc=brah}', "");
Expect(0, 69743, '\P{Sc=brah}', "");
Expect(1, 69743, '\P{^Sc=brah}', "");
Expect(0, 69744, '\p{Sc=brah}', "");
Expect(1, 69744, '\p{^Sc=brah}', "");
Expect(1, 69744, '\P{Sc=brah}', "");
Expect(0, 69744, '\P{^Sc=brah}', "");
Expect(1, 69743, '\p{Sc=	 Brah}', "");
Expect(0, 69743, '\p{^Sc=	 Brah}', "");
Expect(0, 69743, '\P{Sc=	 Brah}', "");
Expect(1, 69743, '\P{^Sc=	 Brah}', "");
Expect(0, 69744, '\p{Sc=	 Brah}', "");
Expect(1, 69744, '\p{^Sc=	 Brah}', "");
Expect(1, 69744, '\P{Sc=	 Brah}', "");
Expect(0, 69744, '\P{^Sc=	 Brah}', "");
Error('\p{Is_Script=_BRAHMI:=}');
Error('\P{Is_Script=_BRAHMI:=}');
Expect(1, 69743, '\p{Is_Script=brahmi}', "");
Expect(0, 69743, '\p{^Is_Script=brahmi}', "");
Expect(0, 69743, '\P{Is_Script=brahmi}', "");
Expect(1, 69743, '\P{^Is_Script=brahmi}', "");
Expect(0, 69744, '\p{Is_Script=brahmi}', "");
Expect(1, 69744, '\p{^Is_Script=brahmi}', "");
Expect(1, 69744, '\P{Is_Script=brahmi}', "");
Expect(0, 69744, '\P{^Is_Script=brahmi}', "");
Expect(1, 69743, '\p{Is_Script:   --BRAHMI}', "");
Expect(0, 69743, '\p{^Is_Script:   --BRAHMI}', "");
Expect(0, 69743, '\P{Is_Script:   --BRAHMI}', "");
Expect(1, 69743, '\P{^Is_Script:   --BRAHMI}', "");
Expect(0, 69744, '\p{Is_Script:   --BRAHMI}', "");
Expect(1, 69744, '\p{^Is_Script:   --BRAHMI}', "");
Expect(1, 69744, '\P{Is_Script:   --BRAHMI}', "");
Expect(0, 69744, '\P{^Is_Script:   --BRAHMI}', "");
Error('\p{Is_Sc=	_brah:=}');
Error('\P{Is_Sc=	_brah:=}');
Expect(1, 69743, '\p{Is_Sc=brah}', "");
Expect(0, 69743, '\p{^Is_Sc=brah}', "");
Expect(0, 69743, '\P{Is_Sc=brah}', "");
Expect(1, 69743, '\P{^Is_Sc=brah}', "");
Expect(0, 69744, '\p{Is_Sc=brah}', "");
Expect(1, 69744, '\p{^Is_Sc=brah}', "");
Expect(1, 69744, '\P{Is_Sc=brah}', "");
Expect(0, 69744, '\P{^Is_Sc=brah}', "");
Expect(1, 69743, '\p{Is_Sc= _Brah}', "");
Expect(0, 69743, '\p{^Is_Sc= _Brah}', "");
Expect(0, 69743, '\P{Is_Sc= _Brah}', "");
Expect(1, 69743, '\P{^Is_Sc= _Brah}', "");
Expect(0, 69744, '\p{Is_Sc= _Brah}', "");
Expect(1, 69744, '\p{^Is_Sc= _Brah}', "");
Expect(1, 69744, '\P{Is_Sc= _Brah}', "");
Expect(0, 69744, '\P{^Is_Sc= _Brah}', "");
Error('\p{Script= 	Braille:=}');
Error('\P{Script= 	Braille:=}');
Expect(1, 10495, '\p{Script=braille}', "");
Expect(0, 10495, '\p{^Script=braille}', "");
Expect(0, 10495, '\P{Script=braille}', "");
Expect(1, 10495, '\P{^Script=braille}', "");
Expect(0, 10496, '\p{Script=braille}', "");
Expect(1, 10496, '\p{^Script=braille}', "");
Expect(1, 10496, '\P{Script=braille}', "");
Expect(0, 10496, '\P{^Script=braille}', "");
Expect(1, 10495, '\p{Script=BRAILLE}', "");
Expect(0, 10495, '\p{^Script=BRAILLE}', "");
Expect(0, 10495, '\P{Script=BRAILLE}', "");
Expect(1, 10495, '\P{^Script=BRAILLE}', "");
Expect(0, 10496, '\p{Script=BRAILLE}', "");
Expect(1, 10496, '\p{^Script=BRAILLE}', "");
Expect(1, 10496, '\P{Script=BRAILLE}', "");
Expect(0, 10496, '\P{^Script=BRAILLE}', "");
Error('\p{Sc=:= Brai}');
Error('\P{Sc=:= Brai}');
Expect(1, 10495, '\p{Sc=brai}', "");
Expect(0, 10495, '\p{^Sc=brai}', "");
Expect(0, 10495, '\P{Sc=brai}', "");
Expect(1, 10495, '\P{^Sc=brai}', "");
Expect(0, 10496, '\p{Sc=brai}', "");
Expect(1, 10496, '\p{^Sc=brai}', "");
Expect(1, 10496, '\P{Sc=brai}', "");
Expect(0, 10496, '\P{^Sc=brai}', "");
Expect(1, 10495, '\p{Sc=_	brai}', "");
Expect(0, 10495, '\p{^Sc=_	brai}', "");
Expect(0, 10495, '\P{Sc=_	brai}', "");
Expect(1, 10495, '\P{^Sc=_	brai}', "");
Expect(0, 10496, '\p{Sc=_	brai}', "");
Expect(1, 10496, '\p{^Sc=_	brai}', "");
Expect(1, 10496, '\P{Sc=_	brai}', "");
Expect(0, 10496, '\P{^Sc=_	brai}', "");
Error('\p{Is_Script=-:=BRAILLE}');
Error('\P{Is_Script=-:=BRAILLE}');
Expect(1, 10495, '\p{Is_Script=braille}', "");
Expect(0, 10495, '\p{^Is_Script=braille}', "");
Expect(0, 10495, '\P{Is_Script=braille}', "");
Expect(1, 10495, '\P{^Is_Script=braille}', "");
Expect(0, 10496, '\p{Is_Script=braille}', "");
Expect(1, 10496, '\p{^Is_Script=braille}', "");
Expect(1, 10496, '\P{Is_Script=braille}', "");
Expect(0, 10496, '\P{^Is_Script=braille}', "");
Expect(1, 10495, '\p{Is_Script=  braille}', "");
Expect(0, 10495, '\p{^Is_Script=  braille}', "");
Expect(0, 10495, '\P{Is_Script=  braille}', "");
Expect(1, 10495, '\P{^Is_Script=  braille}', "");
Expect(0, 10496, '\p{Is_Script=  braille}', "");
Expect(1, 10496, '\p{^Is_Script=  braille}', "");
Expect(1, 10496, '\P{Is_Script=  braille}', "");
Expect(0, 10496, '\P{^Is_Script=  braille}', "");
Error('\p{Is_Sc=_	Brai:=}');
Error('\P{Is_Sc=_	Brai:=}');
Expect(1, 10495, '\p{Is_Sc=brai}', "");
Expect(0, 10495, '\p{^Is_Sc=brai}', "");
Expect(0, 10495, '\P{Is_Sc=brai}', "");
Expect(1, 10495, '\P{^Is_Sc=brai}', "");
Expect(0, 10496, '\p{Is_Sc=brai}', "");
Expect(1, 10496, '\p{^Is_Sc=brai}', "");
Expect(1, 10496, '\P{Is_Sc=brai}', "");
Expect(0, 10496, '\P{^Is_Sc=brai}', "");
Expect(1, 10495, '\p{Is_Sc: - BRAI}', "");
Expect(0, 10495, '\p{^Is_Sc: - BRAI}', "");
Expect(0, 10495, '\P{Is_Sc: - BRAI}', "");
Expect(1, 10495, '\P{^Is_Sc: - BRAI}', "");
Expect(0, 10496, '\p{Is_Sc: - BRAI}', "");
Expect(1, 10496, '\p{^Is_Sc: - BRAI}', "");
Expect(1, 10496, '\P{Is_Sc: - BRAI}', "");
Expect(0, 10496, '\P{^Is_Sc: - BRAI}', "");
Error('\p{Script=--buginese/a/}');
Error('\P{Script=--buginese/a/}');
Expect(1, 6687, '\p{Script=buginese}', "");
Expect(0, 6687, '\p{^Script=buginese}', "");
Expect(0, 6687, '\P{Script=buginese}', "");
Expect(1, 6687, '\P{^Script=buginese}', "");
Expect(0, 6688, '\p{Script=buginese}', "");
Expect(1, 6688, '\p{^Script=buginese}', "");
Expect(1, 6688, '\P{Script=buginese}', "");
Expect(0, 6688, '\P{^Script=buginese}', "");
Expect(1, 6687, '\p{Script=_buginese}', "");
Expect(0, 6687, '\p{^Script=_buginese}', "");
Expect(0, 6687, '\P{Script=_buginese}', "");
Expect(1, 6687, '\P{^Script=_buginese}', "");
Expect(0, 6688, '\p{Script=_buginese}', "");
Expect(1, 6688, '\p{^Script=_buginese}', "");
Expect(1, 6688, '\P{Script=_buginese}', "");
Expect(0, 6688, '\P{^Script=_buginese}', "");
Error('\p{Sc=_:=Bugi}');
Error('\P{Sc=_:=Bugi}');
Expect(1, 6687, '\p{Sc=bugi}', "");
Expect(0, 6687, '\p{^Sc=bugi}', "");
Expect(0, 6687, '\P{Sc=bugi}', "");
Expect(1, 6687, '\P{^Sc=bugi}', "");
Expect(0, 6688, '\p{Sc=bugi}', "");
Expect(1, 6688, '\p{^Sc=bugi}', "");
Expect(1, 6688, '\P{Sc=bugi}', "");
Expect(0, 6688, '\P{^Sc=bugi}', "");
Expect(1, 6687, '\p{Sc=_	bugi}', "");
Expect(0, 6687, '\p{^Sc=_	bugi}', "");
Expect(0, 6687, '\P{Sc=_	bugi}', "");
Expect(1, 6687, '\P{^Sc=_	bugi}', "");
Expect(0, 6688, '\p{Sc=_	bugi}', "");
Expect(1, 6688, '\p{^Sc=_	bugi}', "");
Expect(1, 6688, '\P{Sc=_	bugi}', "");
Expect(0, 6688, '\P{^Sc=_	bugi}', "");
Error('\p{Is_Script=_/a/Buginese}');
Error('\P{Is_Script=_/a/Buginese}');
Expect(1, 6687, '\p{Is_Script=buginese}', "");
Expect(0, 6687, '\p{^Is_Script=buginese}', "");
Expect(0, 6687, '\P{Is_Script=buginese}', "");
Expect(1, 6687, '\P{^Is_Script=buginese}', "");
Expect(0, 6688, '\p{Is_Script=buginese}', "");
Expect(1, 6688, '\p{^Is_Script=buginese}', "");
Expect(1, 6688, '\P{Is_Script=buginese}', "");
Expect(0, 6688, '\P{^Is_Script=buginese}', "");
Expect(1, 6687, '\p{Is_Script=_buginese}', "");
Expect(0, 6687, '\p{^Is_Script=_buginese}', "");
Expect(0, 6687, '\P{Is_Script=_buginese}', "");
Expect(1, 6687, '\P{^Is_Script=_buginese}', "");
Expect(0, 6688, '\p{Is_Script=_buginese}', "");
Expect(1, 6688, '\p{^Is_Script=_buginese}', "");
Expect(1, 6688, '\P{Is_Script=_buginese}', "");
Expect(0, 6688, '\P{^Is_Script=_buginese}', "");
Error('\p{Is_Sc=/a/Bugi}');
Error('\P{Is_Sc=/a/Bugi}');
Expect(1, 6687, '\p{Is_Sc=bugi}', "");
Expect(0, 6687, '\p{^Is_Sc=bugi}', "");
Expect(0, 6687, '\P{Is_Sc=bugi}', "");
Expect(1, 6687, '\P{^Is_Sc=bugi}', "");
Expect(0, 6688, '\p{Is_Sc=bugi}', "");
Expect(1, 6688, '\p{^Is_Sc=bugi}', "");
Expect(1, 6688, '\P{Is_Sc=bugi}', "");
Expect(0, 6688, '\P{^Is_Sc=bugi}', "");
Expect(1, 6687, '\p{Is_Sc=	bugi}', "");
Expect(0, 6687, '\p{^Is_Sc=	bugi}', "");
Expect(0, 6687, '\P{Is_Sc=	bugi}', "");
Expect(1, 6687, '\P{^Is_Sc=	bugi}', "");
Expect(0, 6688, '\p{Is_Sc=	bugi}', "");
Expect(1, 6688, '\p{^Is_Sc=	bugi}', "");
Expect(1, 6688, '\P{Is_Sc=	bugi}', "");
Expect(0, 6688, '\P{^Is_Sc=	bugi}', "");
Error('\p{Script= /a/buhid}');
Error('\P{Script= /a/buhid}');
Expect(1, 5971, '\p{Script=buhid}', "");
Expect(0, 5971, '\p{^Script=buhid}', "");
Expect(0, 5971, '\P{Script=buhid}', "");
Expect(1, 5971, '\P{^Script=buhid}', "");
Expect(0, 5972, '\p{Script=buhid}', "");
Expect(1, 5972, '\p{^Script=buhid}', "");
Expect(1, 5972, '\P{Script=buhid}', "");
Expect(0, 5972, '\P{^Script=buhid}', "");
Expect(1, 5971, '\p{Script=_-buhid}', "");
Expect(0, 5971, '\p{^Script=_-buhid}', "");
Expect(0, 5971, '\P{Script=_-buhid}', "");
Expect(1, 5971, '\P{^Script=_-buhid}', "");
Expect(0, 5972, '\p{Script=_-buhid}', "");
Expect(1, 5972, '\p{^Script=_-buhid}', "");
Expect(1, 5972, '\P{Script=_-buhid}', "");
Expect(0, 5972, '\P{^Script=_-buhid}', "");
Error('\p{Sc=	BUHD/a/}');
Error('\P{Sc=	BUHD/a/}');
Expect(1, 5971, '\p{Sc:buhd}', "");
Expect(0, 5971, '\p{^Sc:buhd}', "");
Expect(0, 5971, '\P{Sc:buhd}', "");
Expect(1, 5971, '\P{^Sc:buhd}', "");
Expect(0, 5972, '\p{Sc:buhd}', "");
Expect(1, 5972, '\p{^Sc:buhd}', "");
Expect(1, 5972, '\P{Sc:buhd}', "");
Expect(0, 5972, '\P{^Sc:buhd}', "");
Expect(1, 5971, '\p{Sc: --buhd}', "");
Expect(0, 5971, '\p{^Sc: --buhd}', "");
Expect(0, 5971, '\P{Sc: --buhd}', "");
Expect(1, 5971, '\P{^Sc: --buhd}', "");
Expect(0, 5972, '\p{Sc: --buhd}', "");
Expect(1, 5972, '\p{^Sc: --buhd}', "");
Expect(1, 5972, '\P{Sc: --buhd}', "");
Expect(0, 5972, '\P{^Sc: --buhd}', "");
Error('\p{Is_Script=	-Buhid:=}');
Error('\P{Is_Script=	-Buhid:=}');
Expect(1, 5971, '\p{Is_Script=buhid}', "");
Expect(0, 5971, '\p{^Is_Script=buhid}', "");
Expect(0, 5971, '\P{Is_Script=buhid}', "");
Expect(1, 5971, '\P{^Is_Script=buhid}', "");
Expect(0, 5972, '\p{Is_Script=buhid}', "");
Expect(1, 5972, '\p{^Is_Script=buhid}', "");
Expect(1, 5972, '\P{Is_Script=buhid}', "");
Expect(0, 5972, '\P{^Is_Script=buhid}', "");
Expect(1, 5971, '\p{Is_Script:- BUHID}', "");
Expect(0, 5971, '\p{^Is_Script:- BUHID}', "");
Expect(0, 5971, '\P{Is_Script:- BUHID}', "");
Expect(1, 5971, '\P{^Is_Script:- BUHID}', "");
Expect(0, 5972, '\p{Is_Script:- BUHID}', "");
Expect(1, 5972, '\p{^Is_Script:- BUHID}', "");
Expect(1, 5972, '\P{Is_Script:- BUHID}', "");
Expect(0, 5972, '\P{^Is_Script:- BUHID}', "");
Error('\p{Is_Sc=/a/		BUHD}');
Error('\P{Is_Sc=/a/		BUHD}');
Expect(1, 5971, '\p{Is_Sc=buhd}', "");
Expect(0, 5971, '\p{^Is_Sc=buhd}', "");
Expect(0, 5971, '\P{Is_Sc=buhd}', "");
Expect(1, 5971, '\P{^Is_Sc=buhd}', "");
Expect(0, 5972, '\p{Is_Sc=buhd}', "");
Expect(1, 5972, '\p{^Is_Sc=buhd}', "");
Expect(1, 5972, '\P{Is_Sc=buhd}', "");
Expect(0, 5972, '\P{^Is_Sc=buhd}', "");
Expect(1, 5971, '\p{Is_Sc= Buhd}', "");
Expect(0, 5971, '\p{^Is_Sc= Buhd}', "");
Expect(0, 5971, '\P{Is_Sc= Buhd}', "");
Expect(1, 5971, '\P{^Is_Sc= Buhd}', "");
Expect(0, 5972, '\p{Is_Sc= Buhd}', "");
Expect(1, 5972, '\p{^Is_Sc= Buhd}', "");
Expect(1, 5972, '\P{Is_Sc= Buhd}', "");
Expect(0, 5972, '\P{^Is_Sc= Buhd}', "");
Error('\p{Script=Chakma/a/}');
Error('\P{Script=Chakma/a/}');
Expect(1, 69955, '\p{Script=chakma}', "");
Expect(0, 69955, '\p{^Script=chakma}', "");
Expect(0, 69955, '\P{Script=chakma}', "");
Expect(1, 69955, '\P{^Script=chakma}', "");
Expect(0, 69956, '\p{Script=chakma}', "");
Expect(1, 69956, '\p{^Script=chakma}', "");
Expect(1, 69956, '\P{Script=chakma}', "");
Expect(0, 69956, '\P{^Script=chakma}', "");
Expect(1, 69955, '\p{Script=-	Chakma}', "");
Expect(0, 69955, '\p{^Script=-	Chakma}', "");
Expect(0, 69955, '\P{Script=-	Chakma}', "");
Expect(1, 69955, '\P{^Script=-	Chakma}', "");
Expect(0, 69956, '\p{Script=-	Chakma}', "");
Expect(1, 69956, '\p{^Script=-	Chakma}', "");
Expect(1, 69956, '\P{Script=-	Chakma}', "");
Expect(0, 69956, '\P{^Script=-	Chakma}', "");
Error('\p{Sc=-cakm/a/}');
Error('\P{Sc=-cakm/a/}');
Expect(1, 69955, '\p{Sc=cakm}', "");
Expect(0, 69955, '\p{^Sc=cakm}', "");
Expect(0, 69955, '\P{Sc=cakm}', "");
Expect(1, 69955, '\P{^Sc=cakm}', "");
Expect(0, 69956, '\p{Sc=cakm}', "");
Expect(1, 69956, '\p{^Sc=cakm}', "");
Expect(1, 69956, '\P{Sc=cakm}', "");
Expect(0, 69956, '\P{^Sc=cakm}', "");
Expect(1, 69955, '\p{Sc=	-cakm}', "");
Expect(0, 69955, '\p{^Sc=	-cakm}', "");
Expect(0, 69955, '\P{Sc=	-cakm}', "");
Expect(1, 69955, '\P{^Sc=	-cakm}', "");
Expect(0, 69956, '\p{Sc=	-cakm}', "");
Expect(1, 69956, '\p{^Sc=	-cakm}', "");
Expect(1, 69956, '\P{Sc=	-cakm}', "");
Expect(0, 69956, '\P{^Sc=	-cakm}', "");
Error('\p{Is_Script= /a/chakma}');
Error('\P{Is_Script= /a/chakma}');
Expect(1, 69955, '\p{Is_Script=chakma}', "");
Expect(0, 69955, '\p{^Is_Script=chakma}', "");
Expect(0, 69955, '\P{Is_Script=chakma}', "");
Expect(1, 69955, '\P{^Is_Script=chakma}', "");
Expect(0, 69956, '\p{Is_Script=chakma}', "");
Expect(1, 69956, '\p{^Is_Script=chakma}', "");
Expect(1, 69956, '\P{Is_Script=chakma}', "");
Expect(0, 69956, '\P{^Is_Script=chakma}', "");
Expect(1, 69955, '\p{Is_Script= -CHAKMA}', "");
Expect(0, 69955, '\p{^Is_Script= -CHAKMA}', "");
Expect(0, 69955, '\P{Is_Script= -CHAKMA}', "");
Expect(1, 69955, '\P{^Is_Script= -CHAKMA}', "");
Expect(0, 69956, '\p{Is_Script= -CHAKMA}', "");
Expect(1, 69956, '\p{^Is_Script= -CHAKMA}', "");
Expect(1, 69956, '\P{Is_Script= -CHAKMA}', "");
Expect(0, 69956, '\P{^Is_Script= -CHAKMA}', "");
Error('\p{Is_Sc=:=_ CAKM}');
Error('\P{Is_Sc=:=_ CAKM}');
Expect(1, 69955, '\p{Is_Sc=cakm}', "");
Expect(0, 69955, '\p{^Is_Sc=cakm}', "");
Expect(0, 69955, '\P{Is_Sc=cakm}', "");
Expect(1, 69955, '\P{^Is_Sc=cakm}', "");
Expect(0, 69956, '\p{Is_Sc=cakm}', "");
Expect(1, 69956, '\p{^Is_Sc=cakm}', "");
Expect(1, 69956, '\P{Is_Sc=cakm}', "");
Expect(0, 69956, '\P{^Is_Sc=cakm}', "");
Expect(1, 69955, '\p{Is_Sc:  	Cakm}', "");
Expect(0, 69955, '\p{^Is_Sc:  	Cakm}', "");
Expect(0, 69955, '\P{Is_Sc:  	Cakm}', "");
Expect(1, 69955, '\P{^Is_Sc:  	Cakm}', "");
Expect(0, 69956, '\p{Is_Sc:  	Cakm}', "");
Expect(1, 69956, '\p{^Is_Sc:  	Cakm}', "");
Expect(1, 69956, '\P{Is_Sc:  	Cakm}', "");
Expect(0, 69956, '\P{^Is_Sc:  	Cakm}', "");
Error('\p{Script:   _	CANADIAN_ABORIGINAL/a/}');
Error('\P{Script:   _	CANADIAN_ABORIGINAL/a/}');
Expect(1, 6389, '\p{Script=canadianaboriginal}', "");
Expect(0, 6389, '\p{^Script=canadianaboriginal}', "");
Expect(0, 6389, '\P{Script=canadianaboriginal}', "");
Expect(1, 6389, '\P{^Script=canadianaboriginal}', "");
Expect(0, 6390, '\p{Script=canadianaboriginal}', "");
Expect(1, 6390, '\p{^Script=canadianaboriginal}', "");
Expect(1, 6390, '\P{Script=canadianaboriginal}', "");
Expect(0, 6390, '\P{^Script=canadianaboriginal}', "");
Expect(1, 6389, '\p{Script: _ canadian_aboriginal}', "");
Expect(0, 6389, '\p{^Script: _ canadian_aboriginal}', "");
Expect(0, 6389, '\P{Script: _ canadian_aboriginal}', "");
Expect(1, 6389, '\P{^Script: _ canadian_aboriginal}', "");
Expect(0, 6390, '\p{Script: _ canadian_aboriginal}', "");
Expect(1, 6390, '\p{^Script: _ canadian_aboriginal}', "");
Expect(1, 6390, '\P{Script: _ canadian_aboriginal}', "");
Expect(0, 6390, '\P{^Script: _ canadian_aboriginal}', "");
Error('\p{Sc= -CANS:=}');
Error('\P{Sc= -CANS:=}');
Expect(1, 6389, '\p{Sc=cans}', "");
Expect(0, 6389, '\p{^Sc=cans}', "");
Expect(0, 6389, '\P{Sc=cans}', "");
Expect(1, 6389, '\P{^Sc=cans}', "");
Expect(0, 6390, '\p{Sc=cans}', "");
Expect(1, 6390, '\p{^Sc=cans}', "");
Expect(1, 6390, '\P{Sc=cans}', "");
Expect(0, 6390, '\P{^Sc=cans}', "");
Expect(1, 6389, '\p{Sc= _Cans}', "");
Expect(0, 6389, '\p{^Sc= _Cans}', "");
Expect(0, 6389, '\P{Sc= _Cans}', "");
Expect(1, 6389, '\P{^Sc= _Cans}', "");
Expect(0, 6390, '\p{Sc= _Cans}', "");
Expect(1, 6390, '\p{^Sc= _Cans}', "");
Expect(1, 6390, '\P{Sc= _Cans}', "");
Expect(0, 6390, '\P{^Sc= _Cans}', "");
Error('\p{Is_Script=_ Canadian_Aboriginal/a/}');
Error('\P{Is_Script=_ Canadian_Aboriginal/a/}');
Expect(1, 6389, '\p{Is_Script=canadianaboriginal}', "");
Expect(0, 6389, '\p{^Is_Script=canadianaboriginal}', "");
Expect(0, 6389, '\P{Is_Script=canadianaboriginal}', "");
Expect(1, 6389, '\P{^Is_Script=canadianaboriginal}', "");
Expect(0, 6390, '\p{Is_Script=canadianaboriginal}', "");
Expect(1, 6390, '\p{^Is_Script=canadianaboriginal}', "");
Expect(1, 6390, '\P{Is_Script=canadianaboriginal}', "");
Expect(0, 6390, '\P{^Is_Script=canadianaboriginal}', "");
Expect(1, 6389, '\p{Is_Script=	Canadian_ABORIGINAL}', "");
Expect(0, 6389, '\p{^Is_Script=	Canadian_ABORIGINAL}', "");
Expect(0, 6389, '\P{Is_Script=	Canadian_ABORIGINAL}', "");
Expect(1, 6389, '\P{^Is_Script=	Canadian_ABORIGINAL}', "");
Expect(0, 6390, '\p{Is_Script=	Canadian_ABORIGINAL}', "");
Expect(1, 6390, '\p{^Is_Script=	Canadian_ABORIGINAL}', "");
Expect(1, 6390, '\P{Is_Script=	Canadian_ABORIGINAL}', "");
Expect(0, 6390, '\P{^Is_Script=	Canadian_ABORIGINAL}', "");
Error('\p{Is_Sc=	Cans/a/}');
Error('\P{Is_Sc=	Cans/a/}');
Expect(1, 6389, '\p{Is_Sc:   cans}', "");
Expect(0, 6389, '\p{^Is_Sc:   cans}', "");
Expect(0, 6389, '\P{Is_Sc:   cans}', "");
Expect(1, 6389, '\P{^Is_Sc:   cans}', "");
Expect(0, 6390, '\p{Is_Sc:   cans}', "");
Expect(1, 6390, '\p{^Is_Sc:   cans}', "");
Expect(1, 6390, '\P{Is_Sc:   cans}', "");
Expect(0, 6390, '\P{^Is_Sc:   cans}', "");
Expect(1, 6389, '\p{Is_Sc=	 Cans}', "");
Expect(0, 6389, '\p{^Is_Sc=	 Cans}', "");
Expect(0, 6389, '\P{Is_Sc=	 Cans}', "");
Expect(1, 6389, '\P{^Is_Sc=	 Cans}', "");
Expect(0, 6390, '\p{Is_Sc=	 Cans}', "");
Expect(1, 6390, '\p{^Is_Sc=	 Cans}', "");
Expect(1, 6390, '\P{Is_Sc=	 Cans}', "");
Expect(0, 6390, '\P{^Is_Sc=	 Cans}', "");
Error('\p{Script:   /a/ _CARIAN}');
Error('\P{Script:   /a/ _CARIAN}');
Expect(1, 66256, '\p{Script=carian}', "");
Expect(0, 66256, '\p{^Script=carian}', "");
Expect(0, 66256, '\P{Script=carian}', "");
Expect(1, 66256, '\P{^Script=carian}', "");
Expect(0, 66257, '\p{Script=carian}', "");
Expect(1, 66257, '\p{^Script=carian}', "");
Expect(1, 66257, '\P{Script=carian}', "");
Expect(0, 66257, '\P{^Script=carian}', "");
Expect(1, 66256, '\p{Script= 	Carian}', "");
Expect(0, 66256, '\p{^Script= 	Carian}', "");
Expect(0, 66256, '\P{Script= 	Carian}', "");
Expect(1, 66256, '\P{^Script= 	Carian}', "");
Expect(0, 66257, '\p{Script= 	Carian}', "");
Expect(1, 66257, '\p{^Script= 	Carian}', "");
Expect(1, 66257, '\P{Script= 	Carian}', "");
Expect(0, 66257, '\P{^Script= 	Carian}', "");
Error('\p{Sc= :=cari}');
Error('\P{Sc= :=cari}');
Expect(1, 66256, '\p{Sc=cari}', "");
Expect(0, 66256, '\p{^Sc=cari}', "");
Expect(0, 66256, '\P{Sc=cari}', "");
Expect(1, 66256, '\P{^Sc=cari}', "");
Expect(0, 66257, '\p{Sc=cari}', "");
Expect(1, 66257, '\p{^Sc=cari}', "");
Expect(1, 66257, '\P{Sc=cari}', "");
Expect(0, 66257, '\P{^Sc=cari}', "");
Expect(1, 66256, '\p{Sc= _Cari}', "");
Expect(0, 66256, '\p{^Sc= _Cari}', "");
Expect(0, 66256, '\P{Sc= _Cari}', "");
Expect(1, 66256, '\P{^Sc= _Cari}', "");
Expect(0, 66257, '\p{Sc= _Cari}', "");
Expect(1, 66257, '\p{^Sc= _Cari}', "");
Expect(1, 66257, '\P{Sc= _Cari}', "");
Expect(0, 66257, '\P{^Sc= _Cari}', "");
Error('\p{Is_Script=/a/carian}');
Error('\P{Is_Script=/a/carian}');
Expect(1, 66256, '\p{Is_Script=carian}', "");
Expect(0, 66256, '\p{^Is_Script=carian}', "");
Expect(0, 66256, '\P{Is_Script=carian}', "");
Expect(1, 66256, '\P{^Is_Script=carian}', "");
Expect(0, 66257, '\p{Is_Script=carian}', "");
Expect(1, 66257, '\p{^Is_Script=carian}', "");
Expect(1, 66257, '\P{Is_Script=carian}', "");
Expect(0, 66257, '\P{^Is_Script=carian}', "");
Expect(1, 66256, '\p{Is_Script=--CARIAN}', "");
Expect(0, 66256, '\p{^Is_Script=--CARIAN}', "");
Expect(0, 66256, '\P{Is_Script=--CARIAN}', "");
Expect(1, 66256, '\P{^Is_Script=--CARIAN}', "");
Expect(0, 66257, '\p{Is_Script=--CARIAN}', "");
Expect(1, 66257, '\p{^Is_Script=--CARIAN}', "");
Expect(1, 66257, '\P{Is_Script=--CARIAN}', "");
Expect(0, 66257, '\P{^Is_Script=--CARIAN}', "");
Error('\p{Is_Sc:		/a/CARI}');
Error('\P{Is_Sc:		/a/CARI}');
Expect(1, 66256, '\p{Is_Sc: cari}', "");
Expect(0, 66256, '\p{^Is_Sc: cari}', "");
Expect(0, 66256, '\P{Is_Sc: cari}', "");
Expect(1, 66256, '\P{^Is_Sc: cari}', "");
Expect(0, 66257, '\p{Is_Sc: cari}', "");
Expect(1, 66257, '\p{^Is_Sc: cari}', "");
Expect(1, 66257, '\P{Is_Sc: cari}', "");
Expect(0, 66257, '\P{^Is_Sc: cari}', "");
Expect(1, 66256, '\p{Is_Sc=_Cari}', "");
Expect(0, 66256, '\p{^Is_Sc=_Cari}', "");
Expect(0, 66256, '\P{Is_Sc=_Cari}', "");
Expect(1, 66256, '\P{^Is_Sc=_Cari}', "");
Expect(0, 66257, '\p{Is_Sc=_Cari}', "");
Expect(1, 66257, '\p{^Is_Sc=_Cari}', "");
Expect(1, 66257, '\P{Is_Sc=_Cari}', "");
Expect(0, 66257, '\P{^Is_Sc=_Cari}', "");
Error('\p{Script=-_cham:=}');
Error('\P{Script=-_cham:=}');
Expect(1, 43615, '\p{Script=cham}', "");
Expect(0, 43615, '\p{^Script=cham}', "");
Expect(0, 43615, '\P{Script=cham}', "");
Expect(1, 43615, '\P{^Script=cham}', "");
Expect(0, 43616, '\p{Script=cham}', "");
Expect(1, 43616, '\p{^Script=cham}', "");
Expect(1, 43616, '\P{Script=cham}', "");
Expect(0, 43616, '\P{^Script=cham}', "");
Expect(1, 43615, '\p{Script= 	Cham}', "");
Expect(0, 43615, '\p{^Script= 	Cham}', "");
Expect(0, 43615, '\P{Script= 	Cham}', "");
Expect(1, 43615, '\P{^Script= 	Cham}', "");
Expect(0, 43616, '\p{Script= 	Cham}', "");
Expect(1, 43616, '\p{^Script= 	Cham}', "");
Expect(1, 43616, '\P{Script= 	Cham}', "");
Expect(0, 43616, '\P{^Script= 	Cham}', "");
Error('\p{Sc: - cham:=}');
Error('\P{Sc: - cham:=}');
Expect(1, 43615, '\p{Sc=cham}', "");
Expect(0, 43615, '\p{^Sc=cham}', "");
Expect(0, 43615, '\P{Sc=cham}', "");
Expect(1, 43615, '\P{^Sc=cham}', "");
Expect(0, 43616, '\p{Sc=cham}', "");
Expect(1, 43616, '\p{^Sc=cham}', "");
Expect(1, 43616, '\P{Sc=cham}', "");
Expect(0, 43616, '\P{^Sc=cham}', "");
Expect(1, 43615, '\p{Sc=-_Cham}', "");
Expect(0, 43615, '\p{^Sc=-_Cham}', "");
Expect(0, 43615, '\P{Sc=-_Cham}', "");
Expect(1, 43615, '\P{^Sc=-_Cham}', "");
Expect(0, 43616, '\p{Sc=-_Cham}', "");
Expect(1, 43616, '\p{^Sc=-_Cham}', "");
Expect(1, 43616, '\P{Sc=-_Cham}', "");
Expect(0, 43616, '\P{^Sc=-_Cham}', "");
Error('\p{Is_Script:_:=Cham}');
Error('\P{Is_Script:_:=Cham}');
Expect(1, 43615, '\p{Is_Script=cham}', "");
Expect(0, 43615, '\p{^Is_Script=cham}', "");
Expect(0, 43615, '\P{Is_Script=cham}', "");
Expect(1, 43615, '\P{^Is_Script=cham}', "");
Expect(0, 43616, '\p{Is_Script=cham}', "");
Expect(1, 43616, '\p{^Is_Script=cham}', "");
Expect(1, 43616, '\P{Is_Script=cham}', "");
Expect(0, 43616, '\P{^Is_Script=cham}', "");
Expect(1, 43615, '\p{Is_Script=--CHAM}', "");
Expect(0, 43615, '\p{^Is_Script=--CHAM}', "");
Expect(0, 43615, '\P{Is_Script=--CHAM}', "");
Expect(1, 43615, '\P{^Is_Script=--CHAM}', "");
Expect(0, 43616, '\p{Is_Script=--CHAM}', "");
Expect(1, 43616, '\p{^Is_Script=--CHAM}', "");
Expect(1, 43616, '\P{Is_Script=--CHAM}', "");
Expect(0, 43616, '\P{^Is_Script=--CHAM}', "");
Error('\p{Is_Sc=:=	CHAM}');
Error('\P{Is_Sc=:=	CHAM}');
Expect(1, 43615, '\p{Is_Sc: cham}', "");
Expect(0, 43615, '\p{^Is_Sc: cham}', "");
Expect(0, 43615, '\P{Is_Sc: cham}', "");
Expect(1, 43615, '\P{^Is_Sc: cham}', "");
Expect(0, 43616, '\p{Is_Sc: cham}', "");
Expect(1, 43616, '\p{^Is_Sc: cham}', "");
Expect(1, 43616, '\P{Is_Sc: cham}', "");
Expect(0, 43616, '\P{^Is_Sc: cham}', "");
Expect(1, 43615, '\p{Is_Sc= cham}', "");
Expect(0, 43615, '\p{^Is_Sc= cham}', "");
Expect(0, 43615, '\P{Is_Sc= cham}', "");
Expect(1, 43615, '\P{^Is_Sc= cham}', "");
Expect(0, 43616, '\p{Is_Sc= cham}', "");
Expect(1, 43616, '\p{^Is_Sc= cham}', "");
Expect(1, 43616, '\P{Is_Sc= cham}', "");
Expect(0, 43616, '\P{^Is_Sc= cham}', "");
Error('\p{Script=:=CHEROKEE}');
Error('\P{Script=:=CHEROKEE}');
Expect(1, 5108, '\p{Script=cherokee}', "");
Expect(0, 5108, '\p{^Script=cherokee}', "");
Expect(0, 5108, '\P{Script=cherokee}', "");
Expect(1, 5108, '\P{^Script=cherokee}', "");
Expect(0, 5109, '\p{Script=cherokee}', "");
Expect(1, 5109, '\p{^Script=cherokee}', "");
Expect(1, 5109, '\P{Script=cherokee}', "");
Expect(0, 5109, '\P{^Script=cherokee}', "");
Expect(1, 5108, '\p{Script= -cherokee}', "");
Expect(0, 5108, '\p{^Script= -cherokee}', "");
Expect(0, 5108, '\P{Script= -cherokee}', "");
Expect(1, 5108, '\P{^Script= -cherokee}', "");
Expect(0, 5109, '\p{Script= -cherokee}', "");
Expect(1, 5109, '\p{^Script= -cherokee}', "");
Expect(1, 5109, '\P{Script= -cherokee}', "");
Expect(0, 5109, '\P{^Script= -cherokee}', "");
Error('\p{Sc=	:=Cher}');
Error('\P{Sc=	:=Cher}');
Expect(1, 5108, '\p{Sc=cher}', "");
Expect(0, 5108, '\p{^Sc=cher}', "");
Expect(0, 5108, '\P{Sc=cher}', "");
Expect(1, 5108, '\P{^Sc=cher}', "");
Expect(0, 5109, '\p{Sc=cher}', "");
Expect(1, 5109, '\p{^Sc=cher}', "");
Expect(1, 5109, '\P{Sc=cher}', "");
Expect(0, 5109, '\P{^Sc=cher}', "");
Expect(1, 5108, '\p{Sc= _Cher}', "");
Expect(0, 5108, '\p{^Sc= _Cher}', "");
Expect(0, 5108, '\P{Sc= _Cher}', "");
Expect(1, 5108, '\P{^Sc= _Cher}', "");
Expect(0, 5109, '\p{Sc= _Cher}', "");
Expect(1, 5109, '\p{^Sc= _Cher}', "");
Expect(1, 5109, '\P{Sc= _Cher}', "");
Expect(0, 5109, '\P{^Sc= _Cher}', "");
Error('\p{Is_Script=_/a/Cherokee}');
Error('\P{Is_Script=_/a/Cherokee}');
Expect(1, 5108, '\p{Is_Script=cherokee}', "");
Expect(0, 5108, '\p{^Is_Script=cherokee}', "");
Expect(0, 5108, '\P{Is_Script=cherokee}', "");
Expect(1, 5108, '\P{^Is_Script=cherokee}', "");
Expect(0, 5109, '\p{Is_Script=cherokee}', "");
Expect(1, 5109, '\p{^Is_Script=cherokee}', "");
Expect(1, 5109, '\P{Is_Script=cherokee}', "");
Expect(0, 5109, '\P{^Is_Script=cherokee}', "");
Expect(1, 5108, '\p{Is_Script= -CHEROKEE}', "");
Expect(0, 5108, '\p{^Is_Script= -CHEROKEE}', "");
Expect(0, 5108, '\P{Is_Script= -CHEROKEE}', "");
Expect(1, 5108, '\P{^Is_Script= -CHEROKEE}', "");
Expect(0, 5109, '\p{Is_Script= -CHEROKEE}', "");
Expect(1, 5109, '\p{^Is_Script= -CHEROKEE}', "");
Expect(1, 5109, '\P{Is_Script= -CHEROKEE}', "");
Expect(0, 5109, '\P{^Is_Script= -CHEROKEE}', "");
Error('\p{Is_Sc=/a/	cher}');
Error('\P{Is_Sc=/a/	cher}');
Expect(1, 5108, '\p{Is_Sc=cher}', "");
Expect(0, 5108, '\p{^Is_Sc=cher}', "");
Expect(0, 5108, '\P{Is_Sc=cher}', "");
Expect(1, 5108, '\P{^Is_Sc=cher}', "");
Expect(0, 5109, '\p{Is_Sc=cher}', "");
Expect(1, 5109, '\p{^Is_Sc=cher}', "");
Expect(1, 5109, '\P{Is_Sc=cher}', "");
Expect(0, 5109, '\P{^Is_Sc=cher}', "");
Expect(1, 5108, '\p{Is_Sc=	_Cher}', "");
Expect(0, 5108, '\p{^Is_Sc=	_Cher}', "");
Expect(0, 5108, '\P{Is_Sc=	_Cher}', "");
Expect(1, 5108, '\P{^Is_Sc=	_Cher}', "");
Expect(0, 5109, '\p{Is_Sc=	_Cher}', "");
Expect(1, 5109, '\p{^Is_Sc=	_Cher}', "");
Expect(1, 5109, '\P{Is_Sc=	_Cher}', "");
Expect(0, 5109, '\P{^Is_Sc=	_Cher}', "");
Error('\p{Script=		Coptic/a/}');
Error('\P{Script=		Coptic/a/}');
Expect(1, 11519, '\p{Script=coptic}', "");
Expect(0, 11519, '\p{^Script=coptic}', "");
Expect(0, 11519, '\P{Script=coptic}', "");
Expect(1, 11519, '\P{^Script=coptic}', "");
Expect(0, 11520, '\p{Script=coptic}', "");
Expect(1, 11520, '\p{^Script=coptic}', "");
Expect(1, 11520, '\P{Script=coptic}', "");
Expect(0, 11520, '\P{^Script=coptic}', "");
Expect(1, 11519, '\p{Script= 	coptic}', "");
Expect(0, 11519, '\p{^Script= 	coptic}', "");
Expect(0, 11519, '\P{Script= 	coptic}', "");
Expect(1, 11519, '\P{^Script= 	coptic}', "");
Expect(0, 11520, '\p{Script= 	coptic}', "");
Expect(1, 11520, '\p{^Script= 	coptic}', "");
Expect(1, 11520, '\P{Script= 	coptic}', "");
Expect(0, 11520, '\P{^Script= 	coptic}', "");
Error('\p{Sc=-	Copt/a/}');
Error('\P{Sc=-	Copt/a/}');
Expect(1, 11519, '\p{Sc=copt}', "");
Expect(0, 11519, '\p{^Sc=copt}', "");
Expect(0, 11519, '\P{Sc=copt}', "");
Expect(1, 11519, '\P{^Sc=copt}', "");
Expect(0, 11520, '\p{Sc=copt}', "");
Expect(1, 11520, '\p{^Sc=copt}', "");
Expect(1, 11520, '\P{Sc=copt}', "");
Expect(0, 11520, '\P{^Sc=copt}', "");
Expect(1, 11519, '\p{Sc= copt}', "");
Expect(0, 11519, '\p{^Sc= copt}', "");
Expect(0, 11519, '\P{Sc= copt}', "");
Expect(1, 11519, '\P{^Sc= copt}', "");
Expect(0, 11520, '\p{Sc= copt}', "");
Expect(1, 11520, '\p{^Sc= copt}', "");
Expect(1, 11520, '\P{Sc= copt}', "");
Expect(0, 11520, '\P{^Sc= copt}', "");
Error('\p{Is_Script=-_QAAC/a/}');
Error('\P{Is_Script=-_QAAC/a/}');
Expect(1, 11519, '\p{Is_Script=qaac}', "");
Expect(0, 11519, '\p{^Is_Script=qaac}', "");
Expect(0, 11519, '\P{Is_Script=qaac}', "");
Expect(1, 11519, '\P{^Is_Script=qaac}', "");
Expect(0, 11520, '\p{Is_Script=qaac}', "");
Expect(1, 11520, '\p{^Is_Script=qaac}', "");
Expect(1, 11520, '\P{Is_Script=qaac}', "");
Expect(0, 11520, '\P{^Is_Script=qaac}', "");
Expect(1, 11519, '\p{Is_Script=_-qaac}', "");
Expect(0, 11519, '\p{^Is_Script=_-qaac}', "");
Expect(0, 11519, '\P{Is_Script=_-qaac}', "");
Expect(1, 11519, '\P{^Is_Script=_-qaac}', "");
Expect(0, 11520, '\p{Is_Script=_-qaac}', "");
Expect(1, 11520, '\p{^Is_Script=_-qaac}', "");
Expect(1, 11520, '\P{Is_Script=_-qaac}', "");
Expect(0, 11520, '\P{^Is_Script=_-qaac}', "");
Error('\p{Is_Sc= :=Coptic}');
Error('\P{Is_Sc= :=Coptic}');
Expect(1, 11519, '\p{Is_Sc=coptic}', "");
Expect(0, 11519, '\p{^Is_Sc=coptic}', "");
Expect(0, 11519, '\P{Is_Sc=coptic}', "");
Expect(1, 11519, '\P{^Is_Sc=coptic}', "");
Expect(0, 11520, '\p{Is_Sc=coptic}', "");
Expect(1, 11520, '\p{^Is_Sc=coptic}', "");
Expect(1, 11520, '\P{Is_Sc=coptic}', "");
Expect(0, 11520, '\P{^Is_Sc=coptic}', "");
Expect(1, 11519, '\p{Is_Sc=-_coptic}', "");
Expect(0, 11519, '\p{^Is_Sc=-_coptic}', "");
Expect(0, 11519, '\P{Is_Sc=-_coptic}', "");
Expect(1, 11519, '\P{^Is_Sc=-_coptic}', "");
Expect(0, 11520, '\p{Is_Sc=-_coptic}', "");
Expect(1, 11520, '\p{^Is_Sc=-_coptic}', "");
Expect(1, 11520, '\P{Is_Sc=-_coptic}', "");
Expect(0, 11520, '\P{^Is_Sc=-_coptic}', "");
Error('\p{Script=	_Cypriot:=}');
Error('\P{Script=	_Cypriot:=}');
Expect(1, 67647, '\p{Script=cypriot}', "");
Expect(0, 67647, '\p{^Script=cypriot}', "");
Expect(0, 67647, '\P{Script=cypriot}', "");
Expect(1, 67647, '\P{^Script=cypriot}', "");
Expect(0, 67648, '\p{Script=cypriot}', "");
Expect(1, 67648, '\p{^Script=cypriot}', "");
Expect(1, 67648, '\P{Script=cypriot}', "");
Expect(0, 67648, '\P{^Script=cypriot}', "");
Expect(1, 67647, '\p{Script=_Cypriot}', "");
Expect(0, 67647, '\p{^Script=_Cypriot}', "");
Expect(0, 67647, '\P{Script=_Cypriot}', "");
Expect(1, 67647, '\P{^Script=_Cypriot}', "");
Expect(0, 67648, '\p{Script=_Cypriot}', "");
Expect(1, 67648, '\p{^Script=_Cypriot}', "");
Expect(1, 67648, '\P{Script=_Cypriot}', "");
Expect(0, 67648, '\P{^Script=_Cypriot}', "");
Error('\p{Sc=:=-cprt}');
Error('\P{Sc=:=-cprt}');
Expect(1, 67647, '\p{Sc=cprt}', "");
Expect(0, 67647, '\p{^Sc=cprt}', "");
Expect(0, 67647, '\P{Sc=cprt}', "");
Expect(1, 67647, '\P{^Sc=cprt}', "");
Expect(0, 67648, '\p{Sc=cprt}', "");
Expect(1, 67648, '\p{^Sc=cprt}', "");
Expect(1, 67648, '\P{Sc=cprt}', "");
Expect(0, 67648, '\P{^Sc=cprt}', "");
Expect(1, 67647, '\p{Sc=--Cprt}', "");
Expect(0, 67647, '\p{^Sc=--Cprt}', "");
Expect(0, 67647, '\P{Sc=--Cprt}', "");
Expect(1, 67647, '\P{^Sc=--Cprt}', "");
Expect(0, 67648, '\p{Sc=--Cprt}', "");
Expect(1, 67648, '\p{^Sc=--Cprt}', "");
Expect(1, 67648, '\P{Sc=--Cprt}', "");
Expect(0, 67648, '\P{^Sc=--Cprt}', "");
Error('\p{Is_Script=:=Cypriot}');
Error('\P{Is_Script=:=Cypriot}');
Expect(1, 67647, '\p{Is_Script=cypriot}', "");
Expect(0, 67647, '\p{^Is_Script=cypriot}', "");
Expect(0, 67647, '\P{Is_Script=cypriot}', "");
Expect(1, 67647, '\P{^Is_Script=cypriot}', "");
Expect(0, 67648, '\p{Is_Script=cypriot}', "");
Expect(1, 67648, '\p{^Is_Script=cypriot}', "");
Expect(1, 67648, '\P{Is_Script=cypriot}', "");
Expect(0, 67648, '\P{^Is_Script=cypriot}', "");
Expect(1, 67647, '\p{Is_Script=_-cypriot}', "");
Expect(0, 67647, '\p{^Is_Script=_-cypriot}', "");
Expect(0, 67647, '\P{Is_Script=_-cypriot}', "");
Expect(1, 67647, '\P{^Is_Script=_-cypriot}', "");
Expect(0, 67648, '\p{Is_Script=_-cypriot}', "");
Expect(1, 67648, '\p{^Is_Script=_-cypriot}', "");
Expect(1, 67648, '\P{Is_Script=_-cypriot}', "");
Expect(0, 67648, '\P{^Is_Script=_-cypriot}', "");
Error('\p{Is_Sc=/a/ Cprt}');
Error('\P{Is_Sc=/a/ Cprt}');
Expect(1, 67647, '\p{Is_Sc=cprt}', "");
Expect(0, 67647, '\p{^Is_Sc=cprt}', "");
Expect(0, 67647, '\P{Is_Sc=cprt}', "");
Expect(1, 67647, '\P{^Is_Sc=cprt}', "");
Expect(0, 67648, '\p{Is_Sc=cprt}', "");
Expect(1, 67648, '\p{^Is_Sc=cprt}', "");
Expect(1, 67648, '\P{Is_Sc=cprt}', "");
Expect(0, 67648, '\P{^Is_Sc=cprt}', "");
Expect(1, 67647, '\p{Is_Sc= CPRT}', "");
Expect(0, 67647, '\p{^Is_Sc= CPRT}', "");
Expect(0, 67647, '\P{Is_Sc= CPRT}', "");
Expect(1, 67647, '\P{^Is_Sc= CPRT}', "");
Expect(0, 67648, '\p{Is_Sc= CPRT}', "");
Expect(1, 67648, '\p{^Is_Sc= CPRT}', "");
Expect(1, 67648, '\P{Is_Sc= CPRT}', "");
Expect(0, 67648, '\P{^Is_Sc= CPRT}', "");
Error('\p{Script=-:=Cyrillic}');
Error('\P{Script=-:=Cyrillic}');
Expect(1, 42655, '\p{Script=cyrillic}', "");
Expect(0, 42655, '\p{^Script=cyrillic}', "");
Expect(0, 42655, '\P{Script=cyrillic}', "");
Expect(1, 42655, '\P{^Script=cyrillic}', "");
Expect(0, 42656, '\p{Script=cyrillic}', "");
Expect(1, 42656, '\p{^Script=cyrillic}', "");
Expect(1, 42656, '\P{Script=cyrillic}', "");
Expect(0, 42656, '\P{^Script=cyrillic}', "");
Expect(1, 42655, '\p{Script=-Cyrillic}', "");
Expect(0, 42655, '\p{^Script=-Cyrillic}', "");
Expect(0, 42655, '\P{Script=-Cyrillic}', "");
Expect(1, 42655, '\P{^Script=-Cyrillic}', "");
Expect(0, 42656, '\p{Script=-Cyrillic}', "");
Expect(1, 42656, '\p{^Script=-Cyrillic}', "");
Expect(1, 42656, '\P{Script=-Cyrillic}', "");
Expect(0, 42656, '\P{^Script=-Cyrillic}', "");
Error('\p{Sc=_ Cyrl:=}');
Error('\P{Sc=_ Cyrl:=}');
Expect(1, 42655, '\p{Sc=cyrl}', "");
Expect(0, 42655, '\p{^Sc=cyrl}', "");
Expect(0, 42655, '\P{Sc=cyrl}', "");
Expect(1, 42655, '\P{^Sc=cyrl}', "");
Expect(0, 42656, '\p{Sc=cyrl}', "");
Expect(1, 42656, '\p{^Sc=cyrl}', "");
Expect(1, 42656, '\P{Sc=cyrl}', "");
Expect(0, 42656, '\P{^Sc=cyrl}', "");
Expect(1, 42655, '\p{Sc=_ Cyrl}', "");
Expect(0, 42655, '\p{^Sc=_ Cyrl}', "");
Expect(0, 42655, '\P{Sc=_ Cyrl}', "");
Expect(1, 42655, '\P{^Sc=_ Cyrl}', "");
Expect(0, 42656, '\p{Sc=_ Cyrl}', "");
Expect(1, 42656, '\p{^Sc=_ Cyrl}', "");
Expect(1, 42656, '\P{Sc=_ Cyrl}', "");
Expect(0, 42656, '\P{^Sc=_ Cyrl}', "");
Error('\p{Is_Script=_	CYRILLIC:=}');
Error('\P{Is_Script=_	CYRILLIC:=}');
Expect(1, 42655, '\p{Is_Script=cyrillic}', "");
Expect(0, 42655, '\p{^Is_Script=cyrillic}', "");
Expect(0, 42655, '\P{Is_Script=cyrillic}', "");
Expect(1, 42655, '\P{^Is_Script=cyrillic}', "");
Expect(0, 42656, '\p{Is_Script=cyrillic}', "");
Expect(1, 42656, '\p{^Is_Script=cyrillic}', "");
Expect(1, 42656, '\P{Is_Script=cyrillic}', "");
Expect(0, 42656, '\P{^Is_Script=cyrillic}', "");
Expect(1, 42655, '\p{Is_Script=	 cyrillic}', "");
Expect(0, 42655, '\p{^Is_Script=	 cyrillic}', "");
Expect(0, 42655, '\P{Is_Script=	 cyrillic}', "");
Expect(1, 42655, '\P{^Is_Script=	 cyrillic}', "");
Expect(0, 42656, '\p{Is_Script=	 cyrillic}', "");
Expect(1, 42656, '\p{^Is_Script=	 cyrillic}', "");
Expect(1, 42656, '\P{Is_Script=	 cyrillic}', "");
Expect(0, 42656, '\P{^Is_Script=	 cyrillic}', "");
Error('\p{Is_Sc=/a/_ CYRL}');
Error('\P{Is_Sc=/a/_ CYRL}');
Expect(1, 42655, '\p{Is_Sc=cyrl}', "");
Expect(0, 42655, '\p{^Is_Sc=cyrl}', "");
Expect(0, 42655, '\P{Is_Sc=cyrl}', "");
Expect(1, 42655, '\P{^Is_Sc=cyrl}', "");
Expect(0, 42656, '\p{Is_Sc=cyrl}', "");
Expect(1, 42656, '\p{^Is_Sc=cyrl}', "");
Expect(1, 42656, '\P{Is_Sc=cyrl}', "");
Expect(0, 42656, '\P{^Is_Sc=cyrl}', "");
Expect(1, 42655, '\p{Is_Sc:	 _cyrl}', "");
Expect(0, 42655, '\p{^Is_Sc:	 _cyrl}', "");
Expect(0, 42655, '\P{Is_Sc:	 _cyrl}', "");
Expect(1, 42655, '\P{^Is_Sc:	 _cyrl}', "");
Expect(0, 42656, '\p{Is_Sc:	 _cyrl}', "");
Expect(1, 42656, '\p{^Is_Sc:	 _cyrl}', "");
Expect(1, 42656, '\P{Is_Sc:	 _cyrl}', "");
Expect(0, 42656, '\P{^Is_Sc:	 _cyrl}', "");
Error('\p{Script=/a/ -devanagari}');
Error('\P{Script=/a/ -devanagari}');
Expect(1, 43259, '\p{Script=devanagari}', "");
Expect(0, 43259, '\p{^Script=devanagari}', "");
Expect(0, 43259, '\P{Script=devanagari}', "");
Expect(1, 43259, '\P{^Script=devanagari}', "");
Expect(0, 43260, '\p{Script=devanagari}', "");
Expect(1, 43260, '\p{^Script=devanagari}', "");
Expect(1, 43260, '\P{Script=devanagari}', "");
Expect(0, 43260, '\P{^Script=devanagari}', "");
Expect(1, 43259, '\p{Script=_devanagari}', "");
Expect(0, 43259, '\p{^Script=_devanagari}', "");
Expect(0, 43259, '\P{Script=_devanagari}', "");
Expect(1, 43259, '\P{^Script=_devanagari}', "");
Expect(0, 43260, '\p{Script=_devanagari}', "");
Expect(1, 43260, '\p{^Script=_devanagari}', "");
Expect(1, 43260, '\P{Script=_devanagari}', "");
Expect(0, 43260, '\P{^Script=_devanagari}', "");
Error('\p{Sc=/a/--Deva}');
Error('\P{Sc=/a/--Deva}');
Expect(1, 43259, '\p{Sc=deva}', "");
Expect(0, 43259, '\p{^Sc=deva}', "");
Expect(0, 43259, '\P{Sc=deva}', "");
Expect(1, 43259, '\P{^Sc=deva}', "");
Expect(0, 43260, '\p{Sc=deva}', "");
Expect(1, 43260, '\p{^Sc=deva}', "");
Expect(1, 43260, '\P{Sc=deva}', "");
Expect(0, 43260, '\P{^Sc=deva}', "");
Expect(1, 43259, '\p{Sc=_ DEVA}', "");
Expect(0, 43259, '\p{^Sc=_ DEVA}', "");
Expect(0, 43259, '\P{Sc=_ DEVA}', "");
Expect(1, 43259, '\P{^Sc=_ DEVA}', "");
Expect(0, 43260, '\p{Sc=_ DEVA}', "");
Expect(1, 43260, '\p{^Sc=_ DEVA}', "");
Expect(1, 43260, '\P{Sc=_ DEVA}', "");
Expect(0, 43260, '\P{^Sc=_ DEVA}', "");
Error('\p{Is_Script=_Devanagari/a/}');
Error('\P{Is_Script=_Devanagari/a/}');
Expect(1, 43259, '\p{Is_Script=devanagari}', "");
Expect(0, 43259, '\p{^Is_Script=devanagari}', "");
Expect(0, 43259, '\P{Is_Script=devanagari}', "");
Expect(1, 43259, '\P{^Is_Script=devanagari}', "");
Expect(0, 43260, '\p{Is_Script=devanagari}', "");
Expect(1, 43260, '\p{^Is_Script=devanagari}', "");
Expect(1, 43260, '\P{Is_Script=devanagari}', "");
Expect(0, 43260, '\P{^Is_Script=devanagari}', "");
Expect(1, 43259, '\p{Is_Script=	-Devanagari}', "");
Expect(0, 43259, '\p{^Is_Script=	-Devanagari}', "");
Expect(0, 43259, '\P{Is_Script=	-Devanagari}', "");
Expect(1, 43259, '\P{^Is_Script=	-Devanagari}', "");
Expect(0, 43260, '\p{Is_Script=	-Devanagari}', "");
Expect(1, 43260, '\p{^Is_Script=	-Devanagari}', "");
Expect(1, 43260, '\P{Is_Script=	-Devanagari}', "");
Expect(0, 43260, '\P{^Is_Script=	-Devanagari}', "");
Error('\p{Is_Sc=/a/-_deva}');
Error('\P{Is_Sc=/a/-_deva}');
Expect(1, 43259, '\p{Is_Sc=deva}', "");
Expect(0, 43259, '\p{^Is_Sc=deva}', "");
Expect(0, 43259, '\P{Is_Sc=deva}', "");
Expect(1, 43259, '\P{^Is_Sc=deva}', "");
Expect(0, 43260, '\p{Is_Sc=deva}', "");
Expect(1, 43260, '\p{^Is_Sc=deva}', "");
Expect(1, 43260, '\P{Is_Sc=deva}', "");
Expect(0, 43260, '\P{^Is_Sc=deva}', "");
Expect(1, 43259, '\p{Is_Sc=-DEVA}', "");
Expect(0, 43259, '\p{^Is_Sc=-DEVA}', "");
Expect(0, 43259, '\P{Is_Sc=-DEVA}', "");
Expect(1, 43259, '\P{^Is_Sc=-DEVA}', "");
Expect(0, 43260, '\p{Is_Sc=-DEVA}', "");
Expect(1, 43260, '\p{^Is_Sc=-DEVA}', "");
Expect(1, 43260, '\P{Is_Sc=-DEVA}', "");
Expect(0, 43260, '\P{^Is_Sc=-DEVA}', "");
Error('\p{Script=	_Deseret:=}');
Error('\P{Script=	_Deseret:=}');
Expect(1, 66639, '\p{Script=deseret}', "");
Expect(0, 66639, '\p{^Script=deseret}', "");
Expect(0, 66639, '\P{Script=deseret}', "");
Expect(1, 66639, '\P{^Script=deseret}', "");
Expect(0, 66640, '\p{Script=deseret}', "");
Expect(1, 66640, '\p{^Script=deseret}', "");
Expect(1, 66640, '\P{Script=deseret}', "");
Expect(0, 66640, '\P{^Script=deseret}', "");
Expect(1, 66639, '\p{Script= Deseret}', "");
Expect(0, 66639, '\p{^Script= Deseret}', "");
Expect(0, 66639, '\P{Script= Deseret}', "");
Expect(1, 66639, '\P{^Script= Deseret}', "");
Expect(0, 66640, '\p{Script= Deseret}', "");
Expect(1, 66640, '\p{^Script= Deseret}', "");
Expect(1, 66640, '\P{Script= Deseret}', "");
Expect(0, 66640, '\P{^Script= Deseret}', "");
Error('\p{Sc=_Dsrt/a/}');
Error('\P{Sc=_Dsrt/a/}');
Expect(1, 66639, '\p{Sc=dsrt}', "");
Expect(0, 66639, '\p{^Sc=dsrt}', "");
Expect(0, 66639, '\P{Sc=dsrt}', "");
Expect(1, 66639, '\P{^Sc=dsrt}', "");
Expect(0, 66640, '\p{Sc=dsrt}', "");
Expect(1, 66640, '\p{^Sc=dsrt}', "");
Expect(1, 66640, '\P{Sc=dsrt}', "");
Expect(0, 66640, '\P{^Sc=dsrt}', "");
Expect(1, 66639, '\p{Sc=_-Dsrt}', "");
Expect(0, 66639, '\p{^Sc=_-Dsrt}', "");
Expect(0, 66639, '\P{Sc=_-Dsrt}', "");
Expect(1, 66639, '\P{^Sc=_-Dsrt}', "");
Expect(0, 66640, '\p{Sc=_-Dsrt}', "");
Expect(1, 66640, '\p{^Sc=_-Dsrt}', "");
Expect(1, 66640, '\P{Sc=_-Dsrt}', "");
Expect(0, 66640, '\P{^Sc=_-Dsrt}', "");
Error('\p{Is_Script=:= deseret}');
Error('\P{Is_Script=:= deseret}');
Expect(1, 66639, '\p{Is_Script=deseret}', "");
Expect(0, 66639, '\p{^Is_Script=deseret}', "");
Expect(0, 66639, '\P{Is_Script=deseret}', "");
Expect(1, 66639, '\P{^Is_Script=deseret}', "");
Expect(0, 66640, '\p{Is_Script=deseret}', "");
Expect(1, 66640, '\p{^Is_Script=deseret}', "");
Expect(1, 66640, '\P{Is_Script=deseret}', "");
Expect(0, 66640, '\P{^Is_Script=deseret}', "");
Expect(1, 66639, '\p{Is_Script=  DESERET}', "");
Expect(0, 66639, '\p{^Is_Script=  DESERET}', "");
Expect(0, 66639, '\P{Is_Script=  DESERET}', "");
Expect(1, 66639, '\P{^Is_Script=  DESERET}', "");
Expect(0, 66640, '\p{Is_Script=  DESERET}', "");
Expect(1, 66640, '\p{^Is_Script=  DESERET}', "");
Expect(1, 66640, '\P{Is_Script=  DESERET}', "");
Expect(0, 66640, '\P{^Is_Script=  DESERET}', "");
Error('\p{Is_Sc=/a/ DSRT}');
Error('\P{Is_Sc=/a/ DSRT}');
Expect(1, 66639, '\p{Is_Sc=dsrt}', "");
Expect(0, 66639, '\p{^Is_Sc=dsrt}', "");
Expect(0, 66639, '\P{Is_Sc=dsrt}', "");
Expect(1, 66639, '\P{^Is_Sc=dsrt}', "");
Expect(0, 66640, '\p{Is_Sc=dsrt}', "");
Expect(1, 66640, '\p{^Is_Sc=dsrt}', "");
Expect(1, 66640, '\P{Is_Sc=dsrt}', "");
Expect(0, 66640, '\P{^Is_Sc=dsrt}', "");
Expect(1, 66639, '\p{Is_Sc=	_Dsrt}', "");
Expect(0, 66639, '\p{^Is_Sc=	_Dsrt}', "");
Expect(0, 66639, '\P{Is_Sc=	_Dsrt}', "");
Expect(1, 66639, '\P{^Is_Sc=	_Dsrt}', "");
Expect(0, 66640, '\p{Is_Sc=	_Dsrt}', "");
Expect(1, 66640, '\p{^Is_Sc=	_Dsrt}', "");
Expect(1, 66640, '\P{Is_Sc=	_Dsrt}', "");
Expect(0, 66640, '\P{^Is_Sc=	_Dsrt}', "");
Error('\p{Script=:=- Egyptian_Hieroglyphs}');
Error('\P{Script=:=- Egyptian_Hieroglyphs}');
Expect(1, 78894, '\p{Script:	egyptianhieroglyphs}', "");
Expect(0, 78894, '\p{^Script:	egyptianhieroglyphs}', "");
Expect(0, 78894, '\P{Script:	egyptianhieroglyphs}', "");
Expect(1, 78894, '\P{^Script:	egyptianhieroglyphs}', "");
Expect(0, 78895, '\p{Script:	egyptianhieroglyphs}', "");
Expect(1, 78895, '\p{^Script:	egyptianhieroglyphs}', "");
Expect(1, 78895, '\P{Script:	egyptianhieroglyphs}', "");
Expect(0, 78895, '\P{^Script:	egyptianhieroglyphs}', "");
Expect(1, 78894, '\p{Script=_-EGYPTIAN_Hieroglyphs}', "");
Expect(0, 78894, '\p{^Script=_-EGYPTIAN_Hieroglyphs}', "");
Expect(0, 78894, '\P{Script=_-EGYPTIAN_Hieroglyphs}', "");
Expect(1, 78894, '\P{^Script=_-EGYPTIAN_Hieroglyphs}', "");
Expect(0, 78895, '\p{Script=_-EGYPTIAN_Hieroglyphs}', "");
Expect(1, 78895, '\p{^Script=_-EGYPTIAN_Hieroglyphs}', "");
Expect(1, 78895, '\P{Script=_-EGYPTIAN_Hieroglyphs}', "");
Expect(0, 78895, '\P{^Script=_-EGYPTIAN_Hieroglyphs}', "");
Error('\p{Sc=	/a/egyp}');
Error('\P{Sc=	/a/egyp}');
Expect(1, 78894, '\p{Sc=egyp}', "");
Expect(0, 78894, '\p{^Sc=egyp}', "");
Expect(0, 78894, '\P{Sc=egyp}', "");
Expect(1, 78894, '\P{^Sc=egyp}', "");
Expect(0, 78895, '\p{Sc=egyp}', "");
Expect(1, 78895, '\p{^Sc=egyp}', "");
Expect(1, 78895, '\P{Sc=egyp}', "");
Expect(0, 78895, '\P{^Sc=egyp}', "");
Expect(1, 78894, '\p{Sc=	_Egyp}', "");
Expect(0, 78894, '\p{^Sc=	_Egyp}', "");
Expect(0, 78894, '\P{Sc=	_Egyp}', "");
Expect(1, 78894, '\P{^Sc=	_Egyp}', "");
Expect(0, 78895, '\p{Sc=	_Egyp}', "");
Expect(1, 78895, '\p{^Sc=	_Egyp}', "");
Expect(1, 78895, '\P{Sc=	_Egyp}', "");
Expect(0, 78895, '\P{^Sc=	_Egyp}', "");
Error('\p{Is_Script=:=  Egyptian_Hieroglyphs}');
Error('\P{Is_Script=:=  Egyptian_Hieroglyphs}');
Expect(1, 78894, '\p{Is_Script=egyptianhieroglyphs}', "");
Expect(0, 78894, '\p{^Is_Script=egyptianhieroglyphs}', "");
Expect(0, 78894, '\P{Is_Script=egyptianhieroglyphs}', "");
Expect(1, 78894, '\P{^Is_Script=egyptianhieroglyphs}', "");
Expect(0, 78895, '\p{Is_Script=egyptianhieroglyphs}', "");
Expect(1, 78895, '\p{^Is_Script=egyptianhieroglyphs}', "");
Expect(1, 78895, '\P{Is_Script=egyptianhieroglyphs}', "");
Expect(0, 78895, '\P{^Is_Script=egyptianhieroglyphs}', "");
Expect(1, 78894, '\p{Is_Script=- Egyptian_HIEROGLYPHS}', "");
Expect(0, 78894, '\p{^Is_Script=- Egyptian_HIEROGLYPHS}', "");
Expect(0, 78894, '\P{Is_Script=- Egyptian_HIEROGLYPHS}', "");
Expect(1, 78894, '\P{^Is_Script=- Egyptian_HIEROGLYPHS}', "");
Expect(0, 78895, '\p{Is_Script=- Egyptian_HIEROGLYPHS}', "");
Expect(1, 78895, '\p{^Is_Script=- Egyptian_HIEROGLYPHS}', "");
Expect(1, 78895, '\P{Is_Script=- Egyptian_HIEROGLYPHS}', "");
Expect(0, 78895, '\P{^Is_Script=- Egyptian_HIEROGLYPHS}', "");
Error('\p{Is_Sc: /a/	_Egyp}');
Error('\P{Is_Sc: /a/	_Egyp}');
Expect(1, 78894, '\p{Is_Sc=egyp}', "");
Expect(0, 78894, '\p{^Is_Sc=egyp}', "");
Expect(0, 78894, '\P{Is_Sc=egyp}', "");
Expect(1, 78894, '\P{^Is_Sc=egyp}', "");
Expect(0, 78895, '\p{Is_Sc=egyp}', "");
Expect(1, 78895, '\p{^Is_Sc=egyp}', "");
Expect(1, 78895, '\P{Is_Sc=egyp}', "");
Expect(0, 78895, '\P{^Is_Sc=egyp}', "");
Expect(1, 78894, '\p{Is_Sc=_Egyp}', "");
Expect(0, 78894, '\p{^Is_Sc=_Egyp}', "");
Expect(0, 78894, '\P{Is_Sc=_Egyp}', "");
Expect(1, 78894, '\P{^Is_Sc=_Egyp}', "");
Expect(0, 78895, '\p{Is_Sc=_Egyp}', "");
Expect(1, 78895, '\p{^Is_Sc=_Egyp}', "");
Expect(1, 78895, '\P{Is_Sc=_Egyp}', "");
Expect(0, 78895, '\P{^Is_Sc=_Egyp}', "");
Error('\p{Script=:=-ethiopic}');
Error('\P{Script=:=-ethiopic}');
Expect(1, 43822, '\p{Script=ethiopic}', "");
Expect(0, 43822, '\p{^Script=ethiopic}', "");
Expect(0, 43822, '\P{Script=ethiopic}', "");
Expect(1, 43822, '\P{^Script=ethiopic}', "");
Expect(0, 43823, '\p{Script=ethiopic}', "");
Expect(1, 43823, '\p{^Script=ethiopic}', "");
Expect(1, 43823, '\P{Script=ethiopic}', "");
Expect(0, 43823, '\P{^Script=ethiopic}', "");
Expect(1, 43822, '\p{Script=	Ethiopic}', "");
Expect(0, 43822, '\p{^Script=	Ethiopic}', "");
Expect(0, 43822, '\P{Script=	Ethiopic}', "");
Expect(1, 43822, '\P{^Script=	Ethiopic}', "");
Expect(0, 43823, '\p{Script=	Ethiopic}', "");
Expect(1, 43823, '\p{^Script=	Ethiopic}', "");
Expect(1, 43823, '\P{Script=	Ethiopic}', "");
Expect(0, 43823, '\P{^Script=	Ethiopic}', "");
Error('\p{Sc=Ethi/a/}');
Error('\P{Sc=Ethi/a/}');
Expect(1, 43822, '\p{Sc=ethi}', "");
Expect(0, 43822, '\p{^Sc=ethi}', "");
Expect(0, 43822, '\P{Sc=ethi}', "");
Expect(1, 43822, '\P{^Sc=ethi}', "");
Expect(0, 43823, '\p{Sc=ethi}', "");
Expect(1, 43823, '\p{^Sc=ethi}', "");
Expect(1, 43823, '\P{Sc=ethi}', "");
Expect(0, 43823, '\P{^Sc=ethi}', "");
Expect(1, 43822, '\p{Sc= 	Ethi}', "");
Expect(0, 43822, '\p{^Sc= 	Ethi}', "");
Expect(0, 43822, '\P{Sc= 	Ethi}', "");
Expect(1, 43822, '\P{^Sc= 	Ethi}', "");
Expect(0, 43823, '\p{Sc= 	Ethi}', "");
Expect(1, 43823, '\p{^Sc= 	Ethi}', "");
Expect(1, 43823, '\P{Sc= 	Ethi}', "");
Expect(0, 43823, '\P{^Sc= 	Ethi}', "");
Error('\p{Is_Script::= -ethiopic}');
Error('\P{Is_Script::= -ethiopic}');
Expect(1, 43822, '\p{Is_Script=ethiopic}', "");
Expect(0, 43822, '\p{^Is_Script=ethiopic}', "");
Expect(0, 43822, '\P{Is_Script=ethiopic}', "");
Expect(1, 43822, '\P{^Is_Script=ethiopic}', "");
Expect(0, 43823, '\p{Is_Script=ethiopic}', "");
Expect(1, 43823, '\p{^Is_Script=ethiopic}', "");
Expect(1, 43823, '\P{Is_Script=ethiopic}', "");
Expect(0, 43823, '\P{^Is_Script=ethiopic}', "");
Expect(1, 43822, '\p{Is_Script:   -Ethiopic}', "");
Expect(0, 43822, '\p{^Is_Script:   -Ethiopic}', "");
Expect(0, 43822, '\P{Is_Script:   -Ethiopic}', "");
Expect(1, 43822, '\P{^Is_Script:   -Ethiopic}', "");
Expect(0, 43823, '\p{Is_Script:   -Ethiopic}', "");
Expect(1, 43823, '\p{^Is_Script:   -Ethiopic}', "");
Expect(1, 43823, '\P{Is_Script:   -Ethiopic}', "");
Expect(0, 43823, '\P{^Is_Script:   -Ethiopic}', "");
Error('\p{Is_Sc= 	ETHI:=}');
Error('\P{Is_Sc= 	ETHI:=}');
Expect(1, 43822, '\p{Is_Sc=ethi}', "");
Expect(0, 43822, '\p{^Is_Sc=ethi}', "");
Expect(0, 43822, '\P{Is_Sc=ethi}', "");
Expect(1, 43822, '\P{^Is_Sc=ethi}', "");
Expect(0, 43823, '\p{Is_Sc=ethi}', "");
Expect(1, 43823, '\p{^Is_Sc=ethi}', "");
Expect(1, 43823, '\P{Is_Sc=ethi}', "");
Expect(0, 43823, '\P{^Is_Sc=ethi}', "");
Expect(1, 43822, '\p{Is_Sc=-Ethi}', "");
Expect(0, 43822, '\p{^Is_Sc=-Ethi}', "");
Expect(0, 43822, '\P{Is_Sc=-Ethi}', "");
Expect(1, 43822, '\P{^Is_Sc=-Ethi}', "");
Expect(0, 43823, '\p{Is_Sc=-Ethi}', "");
Expect(1, 43823, '\p{^Is_Sc=-Ethi}', "");
Expect(1, 43823, '\P{Is_Sc=-Ethi}', "");
Expect(0, 43823, '\P{^Is_Sc=-Ethi}', "");
Error('\p{Script=	/a/Georgian}');
Error('\P{Script=	/a/Georgian}');
Expect(1, 11565, '\p{Script=georgian}', "");
Expect(0, 11565, '\p{^Script=georgian}', "");
Expect(0, 11565, '\P{Script=georgian}', "");
Expect(1, 11565, '\P{^Script=georgian}', "");
Expect(0, 11566, '\p{Script=georgian}', "");
Expect(1, 11566, '\p{^Script=georgian}', "");
Expect(1, 11566, '\P{Script=georgian}', "");
Expect(0, 11566, '\P{^Script=georgian}', "");
Expect(1, 11565, '\p{Script=-Georgian}', "");
Expect(0, 11565, '\p{^Script=-Georgian}', "");
Expect(0, 11565, '\P{Script=-Georgian}', "");
Expect(1, 11565, '\P{^Script=-Georgian}', "");
Expect(0, 11566, '\p{Script=-Georgian}', "");
Expect(1, 11566, '\p{^Script=-Georgian}', "");
Expect(1, 11566, '\P{Script=-Georgian}', "");
Expect(0, 11566, '\P{^Script=-Georgian}', "");
Error('\p{Sc=:=-_Geor}');
Error('\P{Sc=:=-_Geor}');
Expect(1, 11565, '\p{Sc=geor}', "");
Expect(0, 11565, '\p{^Sc=geor}', "");
Expect(0, 11565, '\P{Sc=geor}', "");
Expect(1, 11565, '\P{^Sc=geor}', "");
Expect(0, 11566, '\p{Sc=geor}', "");
Expect(1, 11566, '\p{^Sc=geor}', "");
Expect(1, 11566, '\P{Sc=geor}', "");
Expect(0, 11566, '\P{^Sc=geor}', "");
Expect(1, 11565, '\p{Sc: _-GEOR}', "");
Expect(0, 11565, '\p{^Sc: _-GEOR}', "");
Expect(0, 11565, '\P{Sc: _-GEOR}', "");
Expect(1, 11565, '\P{^Sc: _-GEOR}', "");
Expect(0, 11566, '\p{Sc: _-GEOR}', "");
Expect(1, 11566, '\p{^Sc: _-GEOR}', "");
Expect(1, 11566, '\P{Sc: _-GEOR}', "");
Expect(0, 11566, '\P{^Sc: _-GEOR}', "");
Error('\p{Is_Script=	-GEORGIAN/a/}');
Error('\P{Is_Script=	-GEORGIAN/a/}');
Expect(1, 11565, '\p{Is_Script=georgian}', "");
Expect(0, 11565, '\p{^Is_Script=georgian}', "");
Expect(0, 11565, '\P{Is_Script=georgian}', "");
Expect(1, 11565, '\P{^Is_Script=georgian}', "");
Expect(0, 11566, '\p{Is_Script=georgian}', "");
Expect(1, 11566, '\p{^Is_Script=georgian}', "");
Expect(1, 11566, '\P{Is_Script=georgian}', "");
Expect(0, 11566, '\P{^Is_Script=georgian}', "");
Expect(1, 11565, '\p{Is_Script=-georgian}', "");
Expect(0, 11565, '\p{^Is_Script=-georgian}', "");
Expect(0, 11565, '\P{Is_Script=-georgian}', "");
Expect(1, 11565, '\P{^Is_Script=-georgian}', "");
Expect(0, 11566, '\p{Is_Script=-georgian}', "");
Expect(1, 11566, '\p{^Is_Script=-georgian}', "");
Expect(1, 11566, '\P{Is_Script=-georgian}', "");
Expect(0, 11566, '\P{^Is_Script=-georgian}', "");
Error('\p{Is_Sc:	/a/-	Geor}');
Error('\P{Is_Sc:	/a/-	Geor}');
Expect(1, 11565, '\p{Is_Sc:	geor}', "");
Expect(0, 11565, '\p{^Is_Sc:	geor}', "");
Expect(0, 11565, '\P{Is_Sc:	geor}', "");
Expect(1, 11565, '\P{^Is_Sc:	geor}', "");
Expect(0, 11566, '\p{Is_Sc:	geor}', "");
Expect(1, 11566, '\p{^Is_Sc:	geor}', "");
Expect(1, 11566, '\P{Is_Sc:	geor}', "");
Expect(0, 11566, '\P{^Is_Sc:	geor}', "");
Expect(1, 11565, '\p{Is_Sc=_Geor}', "");
Expect(0, 11565, '\p{^Is_Sc=_Geor}', "");
Expect(0, 11565, '\P{Is_Sc=_Geor}', "");
Expect(1, 11565, '\P{^Is_Sc=_Geor}', "");
Expect(0, 11566, '\p{Is_Sc=_Geor}', "");
Expect(1, 11566, '\p{^Is_Sc=_Geor}', "");
Expect(1, 11566, '\P{Is_Sc=_Geor}', "");
Expect(0, 11566, '\P{^Is_Sc=_Geor}', "");
Error('\p{Script=--glagolitic/a/}');
Error('\P{Script=--glagolitic/a/}');
Expect(1, 11358, '\p{Script=glagolitic}', "");
Expect(0, 11358, '\p{^Script=glagolitic}', "");
Expect(0, 11358, '\P{Script=glagolitic}', "");
Expect(1, 11358, '\P{^Script=glagolitic}', "");
Expect(0, 11359, '\p{Script=glagolitic}', "");
Expect(1, 11359, '\p{^Script=glagolitic}', "");
Expect(1, 11359, '\P{Script=glagolitic}', "");
Expect(0, 11359, '\P{^Script=glagolitic}', "");
Expect(1, 11358, '\p{Script=_	GLAGOLITIC}', "");
Expect(0, 11358, '\p{^Script=_	GLAGOLITIC}', "");
Expect(0, 11358, '\P{Script=_	GLAGOLITIC}', "");
Expect(1, 11358, '\P{^Script=_	GLAGOLITIC}', "");
Expect(0, 11359, '\p{Script=_	GLAGOLITIC}', "");
Expect(1, 11359, '\p{^Script=_	GLAGOLITIC}', "");
Expect(1, 11359, '\P{Script=_	GLAGOLITIC}', "");
Expect(0, 11359, '\P{^Script=_	GLAGOLITIC}', "");
Error('\p{Sc=:= 	GLAG}');
Error('\P{Sc=:= 	GLAG}');
Expect(1, 11358, '\p{Sc=glag}', "");
Expect(0, 11358, '\p{^Sc=glag}', "");
Expect(0, 11358, '\P{Sc=glag}', "");
Expect(1, 11358, '\P{^Sc=glag}', "");
Expect(0, 11359, '\p{Sc=glag}', "");
Expect(1, 11359, '\p{^Sc=glag}', "");
Expect(1, 11359, '\P{Sc=glag}', "");
Expect(0, 11359, '\P{^Sc=glag}', "");
Expect(1, 11358, '\p{Sc=_GLAG}', "");
Expect(0, 11358, '\p{^Sc=_GLAG}', "");
Expect(0, 11358, '\P{Sc=_GLAG}', "");
Expect(1, 11358, '\P{^Sc=_GLAG}', "");
Expect(0, 11359, '\p{Sc=_GLAG}', "");
Expect(1, 11359, '\p{^Sc=_GLAG}', "");
Expect(1, 11359, '\P{Sc=_GLAG}', "");
Expect(0, 11359, '\P{^Sc=_GLAG}', "");
Error('\p{Is_Script=	-glagolitic/a/}');
Error('\P{Is_Script=	-glagolitic/a/}');
Expect(1, 11358, '\p{Is_Script=glagolitic}', "");
Expect(0, 11358, '\p{^Is_Script=glagolitic}', "");
Expect(0, 11358, '\P{Is_Script=glagolitic}', "");
Expect(1, 11358, '\P{^Is_Script=glagolitic}', "");
Expect(0, 11359, '\p{Is_Script=glagolitic}', "");
Expect(1, 11359, '\p{^Is_Script=glagolitic}', "");
Expect(1, 11359, '\P{Is_Script=glagolitic}', "");
Expect(0, 11359, '\P{^Is_Script=glagolitic}', "");
Expect(1, 11358, '\p{Is_Script=	-Glagolitic}', "");
Expect(0, 11358, '\p{^Is_Script=	-Glagolitic}', "");
Expect(0, 11358, '\P{Is_Script=	-Glagolitic}', "");
Expect(1, 11358, '\P{^Is_Script=	-Glagolitic}', "");
Expect(0, 11359, '\p{Is_Script=	-Glagolitic}', "");
Expect(1, 11359, '\p{^Is_Script=	-Glagolitic}', "");
Expect(1, 11359, '\P{Is_Script=	-Glagolitic}', "");
Expect(0, 11359, '\P{^Is_Script=	-Glagolitic}', "");
Error('\p{Is_Sc=:=_Glag}');
Error('\P{Is_Sc=:=_Glag}');
Expect(1, 11358, '\p{Is_Sc=glag}', "");
Expect(0, 11358, '\p{^Is_Sc=glag}', "");
Expect(0, 11358, '\P{Is_Sc=glag}', "");
Expect(1, 11358, '\P{^Is_Sc=glag}', "");
Expect(0, 11359, '\p{Is_Sc=glag}', "");
Expect(1, 11359, '\p{^Is_Sc=glag}', "");
Expect(1, 11359, '\P{Is_Sc=glag}', "");
Expect(0, 11359, '\P{^Is_Sc=glag}', "");
Expect(1, 11358, '\p{Is_Sc=-Glag}', "");
Expect(0, 11358, '\p{^Is_Sc=-Glag}', "");
Expect(0, 11358, '\P{Is_Sc=-Glag}', "");
Expect(1, 11358, '\P{^Is_Sc=-Glag}', "");
Expect(0, 11359, '\p{Is_Sc=-Glag}', "");
Expect(1, 11359, '\p{^Is_Sc=-Glag}', "");
Expect(1, 11359, '\P{Is_Sc=-Glag}', "");
Expect(0, 11359, '\P{^Is_Sc=-Glag}', "");
Error('\p{Script= :=Gothic}');
Error('\P{Script= :=Gothic}');
Expect(1, 66378, '\p{Script=gothic}', "");
Expect(0, 66378, '\p{^Script=gothic}', "");
Expect(0, 66378, '\P{Script=gothic}', "");
Expect(1, 66378, '\P{^Script=gothic}', "");
Expect(0, 66379, '\p{Script=gothic}', "");
Expect(1, 66379, '\p{^Script=gothic}', "");
Expect(1, 66379, '\P{Script=gothic}', "");
Expect(0, 66379, '\P{^Script=gothic}', "");
Expect(1, 66378, '\p{Script=__GOTHIC}', "");
Expect(0, 66378, '\p{^Script=__GOTHIC}', "");
Expect(0, 66378, '\P{Script=__GOTHIC}', "");
Expect(1, 66378, '\P{^Script=__GOTHIC}', "");
Expect(0, 66379, '\p{Script=__GOTHIC}', "");
Expect(1, 66379, '\p{^Script=__GOTHIC}', "");
Expect(1, 66379, '\P{Script=__GOTHIC}', "");
Expect(0, 66379, '\P{^Script=__GOTHIC}', "");
Error('\p{Sc=:=_Goth}');
Error('\P{Sc=:=_Goth}');
Expect(1, 66378, '\p{Sc=goth}', "");
Expect(0, 66378, '\p{^Sc=goth}', "");
Expect(0, 66378, '\P{Sc=goth}', "");
Expect(1, 66378, '\P{^Sc=goth}', "");
Expect(0, 66379, '\p{Sc=goth}', "");
Expect(1, 66379, '\p{^Sc=goth}', "");
Expect(1, 66379, '\P{Sc=goth}', "");
Expect(0, 66379, '\P{^Sc=goth}', "");
Expect(1, 66378, '\p{Sc=	 goth}', "");
Expect(0, 66378, '\p{^Sc=	 goth}', "");
Expect(0, 66378, '\P{Sc=	 goth}', "");
Expect(1, 66378, '\P{^Sc=	 goth}', "");
Expect(0, 66379, '\p{Sc=	 goth}', "");
Expect(1, 66379, '\p{^Sc=	 goth}', "");
Expect(1, 66379, '\P{Sc=	 goth}', "");
Expect(0, 66379, '\P{^Sc=	 goth}', "");
Error('\p{Is_Script=-GOTHIC:=}');
Error('\P{Is_Script=-GOTHIC:=}');
Expect(1, 66378, '\p{Is_Script=gothic}', "");
Expect(0, 66378, '\p{^Is_Script=gothic}', "");
Expect(0, 66378, '\P{Is_Script=gothic}', "");
Expect(1, 66378, '\P{^Is_Script=gothic}', "");
Expect(0, 66379, '\p{Is_Script=gothic}', "");
Expect(1, 66379, '\p{^Is_Script=gothic}', "");
Expect(1, 66379, '\P{Is_Script=gothic}', "");
Expect(0, 66379, '\P{^Is_Script=gothic}', "");
Error('\p{Is_Sc:	--GOTH/a/}');
Error('\P{Is_Sc:	--GOTH/a/}');
Expect(1, 66378, '\p{Is_Sc=goth}', "");
Expect(0, 66378, '\p{^Is_Sc=goth}', "");
Expect(0, 66378, '\P{Is_Sc=goth}', "");
Expect(1, 66378, '\P{^Is_Sc=goth}', "");
Expect(0, 66379, '\p{Is_Sc=goth}', "");
Expect(1, 66379, '\p{^Is_Sc=goth}', "");
Expect(1, 66379, '\P{Is_Sc=goth}', "");
Expect(0, 66379, '\P{^Is_Sc=goth}', "");
Expect(1, 66378, '\p{Is_Sc=	GOTH}', "");
Expect(0, 66378, '\p{^Is_Sc=	GOTH}', "");
Expect(0, 66378, '\P{Is_Sc=	GOTH}', "");
Expect(1, 66378, '\P{^Is_Sc=	GOTH}', "");
Expect(0, 66379, '\p{Is_Sc=	GOTH}', "");
Expect(1, 66379, '\p{^Is_Sc=	GOTH}', "");
Expect(1, 66379, '\P{Is_Sc=	GOTH}', "");
Expect(0, 66379, '\P{^Is_Sc=	GOTH}', "");
Error('\p{Script=/a/	-greek}');
Error('\P{Script=/a/	-greek}');
Expect(1, 119365, '\p{Script=greek}', "");
Expect(0, 119365, '\p{^Script=greek}', "");
Expect(0, 119365, '\P{Script=greek}', "");
Expect(1, 119365, '\P{^Script=greek}', "");
Expect(0, 119366, '\p{Script=greek}', "");
Expect(1, 119366, '\p{^Script=greek}', "");
Expect(1, 119366, '\P{Script=greek}', "");
Expect(0, 119366, '\P{^Script=greek}', "");
Expect(1, 119365, '\p{Script:	_-GREEK}', "");
Expect(0, 119365, '\p{^Script:	_-GREEK}', "");
Expect(0, 119365, '\P{Script:	_-GREEK}', "");
Expect(1, 119365, '\P{^Script:	_-GREEK}', "");
Expect(0, 119366, '\p{Script:	_-GREEK}', "");
Expect(1, 119366, '\p{^Script:	_-GREEK}', "");
Expect(1, 119366, '\P{Script:	_-GREEK}', "");
Expect(0, 119366, '\P{^Script:	_-GREEK}', "");
Error('\p{Sc::= _GREK}');
Error('\P{Sc::= _GREK}');
Expect(1, 119365, '\p{Sc=grek}', "");
Expect(0, 119365, '\p{^Sc=grek}', "");
Expect(0, 119365, '\P{Sc=grek}', "");
Expect(1, 119365, '\P{^Sc=grek}', "");
Expect(0, 119366, '\p{Sc=grek}', "");
Expect(1, 119366, '\p{^Sc=grek}', "");
Expect(1, 119366, '\P{Sc=grek}', "");
Expect(0, 119366, '\P{^Sc=grek}', "");
Expect(1, 119365, '\p{Sc= 	GREK}', "");
Expect(0, 119365, '\p{^Sc= 	GREK}', "");
Expect(0, 119365, '\P{Sc= 	GREK}', "");
Expect(1, 119365, '\P{^Sc= 	GREK}', "");
Expect(0, 119366, '\p{Sc= 	GREK}', "");
Expect(1, 119366, '\p{^Sc= 	GREK}', "");
Expect(1, 119366, '\P{Sc= 	GREK}', "");
Expect(0, 119366, '\P{^Sc= 	GREK}', "");
Error('\p{Is_Script=:=_GREEK}');
Error('\P{Is_Script=:=_GREEK}');
Expect(1, 119365, '\p{Is_Script=greek}', "");
Expect(0, 119365, '\p{^Is_Script=greek}', "");
Expect(0, 119365, '\P{Is_Script=greek}', "");
Expect(1, 119365, '\P{^Is_Script=greek}', "");
Expect(0, 119366, '\p{Is_Script=greek}', "");
Expect(1, 119366, '\p{^Is_Script=greek}', "");
Expect(1, 119366, '\P{Is_Script=greek}', "");
Expect(0, 119366, '\P{^Is_Script=greek}', "");
Expect(1, 119365, '\p{Is_Script= 	Greek}', "");
Expect(0, 119365, '\p{^Is_Script= 	Greek}', "");
Expect(0, 119365, '\P{Is_Script= 	Greek}', "");
Expect(1, 119365, '\P{^Is_Script= 	Greek}', "");
Expect(0, 119366, '\p{Is_Script= 	Greek}', "");
Expect(1, 119366, '\p{^Is_Script= 	Greek}', "");
Expect(1, 119366, '\P{Is_Script= 	Greek}', "");
Expect(0, 119366, '\P{^Is_Script= 	Greek}', "");
Error('\p{Is_Sc= /a/Grek}');
Error('\P{Is_Sc= /a/Grek}');
Expect(1, 119365, '\p{Is_Sc=grek}', "");
Expect(0, 119365, '\p{^Is_Sc=grek}', "");
Expect(0, 119365, '\P{Is_Sc=grek}', "");
Expect(1, 119365, '\P{^Is_Sc=grek}', "");
Expect(0, 119366, '\p{Is_Sc=grek}', "");
Expect(1, 119366, '\p{^Is_Sc=grek}', "");
Expect(1, 119366, '\P{Is_Sc=grek}', "");
Expect(0, 119366, '\P{^Is_Sc=grek}', "");
Expect(1, 119365, '\p{Is_Sc=__grek}', "");
Expect(0, 119365, '\p{^Is_Sc=__grek}', "");
Expect(0, 119365, '\P{Is_Sc=__grek}', "");
Expect(1, 119365, '\P{^Is_Sc=__grek}', "");
Expect(0, 119366, '\p{Is_Sc=__grek}', "");
Expect(1, 119366, '\p{^Is_Sc=__grek}', "");
Expect(1, 119366, '\P{Is_Sc=__grek}', "");
Expect(0, 119366, '\P{^Is_Sc=__grek}', "");
Error('\p{Script= :=Gujarati}');
Error('\P{Script= :=Gujarati}');
Expect(1, 2801, '\p{Script:	gujarati}', "");
Expect(0, 2801, '\p{^Script:	gujarati}', "");
Expect(0, 2801, '\P{Script:	gujarati}', "");
Expect(1, 2801, '\P{^Script:	gujarati}', "");
Expect(0, 2802, '\p{Script:	gujarati}', "");
Expect(1, 2802, '\p{^Script:	gujarati}', "");
Expect(1, 2802, '\P{Script:	gujarati}', "");
Expect(0, 2802, '\P{^Script:	gujarati}', "");
Expect(1, 2801, '\p{Script=  gujarati}', "");
Expect(0, 2801, '\p{^Script=  gujarati}', "");
Expect(0, 2801, '\P{Script=  gujarati}', "");
Expect(1, 2801, '\P{^Script=  gujarati}', "");
Expect(0, 2802, '\p{Script=  gujarati}', "");
Expect(1, 2802, '\p{^Script=  gujarati}', "");
Expect(1, 2802, '\P{Script=  gujarati}', "");
Expect(0, 2802, '\P{^Script=  gujarati}', "");
Error('\p{Sc=/a/_Gujr}');
Error('\P{Sc=/a/_Gujr}');
Expect(1, 2801, '\p{Sc=gujr}', "");
Expect(0, 2801, '\p{^Sc=gujr}', "");
Expect(0, 2801, '\P{Sc=gujr}', "");
Expect(1, 2801, '\P{^Sc=gujr}', "");
Expect(0, 2802, '\p{Sc=gujr}', "");
Expect(1, 2802, '\p{^Sc=gujr}', "");
Expect(1, 2802, '\P{Sc=gujr}', "");
Expect(0, 2802, '\P{^Sc=gujr}', "");
Expect(1, 2801, '\p{Sc=	-Gujr}', "");
Expect(0, 2801, '\p{^Sc=	-Gujr}', "");
Expect(0, 2801, '\P{Sc=	-Gujr}', "");
Expect(1, 2801, '\P{^Sc=	-Gujr}', "");
Expect(0, 2802, '\p{Sc=	-Gujr}', "");
Expect(1, 2802, '\p{^Sc=	-Gujr}', "");
Expect(1, 2802, '\P{Sc=	-Gujr}', "");
Expect(0, 2802, '\P{^Sc=	-Gujr}', "");
Error('\p{Is_Script=-_Gujarati/a/}');
Error('\P{Is_Script=-_Gujarati/a/}');
Expect(1, 2801, '\p{Is_Script=gujarati}', "");
Expect(0, 2801, '\p{^Is_Script=gujarati}', "");
Expect(0, 2801, '\P{Is_Script=gujarati}', "");
Expect(1, 2801, '\P{^Is_Script=gujarati}', "");
Expect(0, 2802, '\p{Is_Script=gujarati}', "");
Expect(1, 2802, '\p{^Is_Script=gujarati}', "");
Expect(1, 2802, '\P{Is_Script=gujarati}', "");
Expect(0, 2802, '\P{^Is_Script=gujarati}', "");
Expect(1, 2801, '\p{Is_Script= 	gujarati}', "");
Expect(0, 2801, '\p{^Is_Script= 	gujarati}', "");
Expect(0, 2801, '\P{Is_Script= 	gujarati}', "");
Expect(1, 2801, '\P{^Is_Script= 	gujarati}', "");
Expect(0, 2802, '\p{Is_Script= 	gujarati}', "");
Expect(1, 2802, '\p{^Is_Script= 	gujarati}', "");
Expect(1, 2802, '\P{Is_Script= 	gujarati}', "");
Expect(0, 2802, '\P{^Is_Script= 	gujarati}', "");
Error('\p{Is_Sc=:=_gujr}');
Error('\P{Is_Sc=:=_gujr}');
Expect(1, 2801, '\p{Is_Sc=gujr}', "");
Expect(0, 2801, '\p{^Is_Sc=gujr}', "");
Expect(0, 2801, '\P{Is_Sc=gujr}', "");
Expect(1, 2801, '\P{^Is_Sc=gujr}', "");
Expect(0, 2802, '\p{Is_Sc=gujr}', "");
Expect(1, 2802, '\p{^Is_Sc=gujr}', "");
Expect(1, 2802, '\P{Is_Sc=gujr}', "");
Expect(0, 2802, '\P{^Is_Sc=gujr}', "");
Expect(1, 2801, '\p{Is_Sc=-Gujr}', "");
Expect(0, 2801, '\p{^Is_Sc=-Gujr}', "");
Expect(0, 2801, '\P{Is_Sc=-Gujr}', "");
Expect(1, 2801, '\P{^Is_Sc=-Gujr}', "");
Expect(0, 2802, '\p{Is_Sc=-Gujr}', "");
Expect(1, 2802, '\p{^Is_Sc=-Gujr}', "");
Expect(1, 2802, '\P{Is_Sc=-Gujr}', "");
Expect(0, 2802, '\P{^Is_Sc=-Gujr}', "");
Error('\p{Script:	_-gurmukhi:=}');
Error('\P{Script:	_-gurmukhi:=}');
Expect(1, 2677, '\p{Script:gurmukhi}', "");
Expect(0, 2677, '\p{^Script:gurmukhi}', "");
Expect(0, 2677, '\P{Script:gurmukhi}', "");
Expect(1, 2677, '\P{^Script:gurmukhi}', "");
Expect(0, 2678, '\p{Script:gurmukhi}', "");
Expect(1, 2678, '\p{^Script:gurmukhi}', "");
Expect(1, 2678, '\P{Script:gurmukhi}', "");
Expect(0, 2678, '\P{^Script:gurmukhi}', "");
Expect(1, 2677, '\p{Script=-	GURMUKHI}', "");
Expect(0, 2677, '\p{^Script=-	GURMUKHI}', "");
Expect(0, 2677, '\P{Script=-	GURMUKHI}', "");
Expect(1, 2677, '\P{^Script=-	GURMUKHI}', "");
Expect(0, 2678, '\p{Script=-	GURMUKHI}', "");
Expect(1, 2678, '\p{^Script=-	GURMUKHI}', "");
Expect(1, 2678, '\P{Script=-	GURMUKHI}', "");
Expect(0, 2678, '\P{^Script=-	GURMUKHI}', "");
Error('\p{Sc=	guru/a/}');
Error('\P{Sc=	guru/a/}');
Expect(1, 2677, '\p{Sc: guru}', "");
Expect(0, 2677, '\p{^Sc: guru}', "");
Expect(0, 2677, '\P{Sc: guru}', "");
Expect(1, 2677, '\P{^Sc: guru}', "");
Expect(0, 2678, '\p{Sc: guru}', "");
Expect(1, 2678, '\p{^Sc: guru}', "");
Expect(1, 2678, '\P{Sc: guru}', "");
Expect(0, 2678, '\P{^Sc: guru}', "");
Expect(1, 2677, '\p{Sc=		Guru}', "");
Expect(0, 2677, '\p{^Sc=		Guru}', "");
Expect(0, 2677, '\P{Sc=		Guru}', "");
Expect(1, 2677, '\P{^Sc=		Guru}', "");
Expect(0, 2678, '\p{Sc=		Guru}', "");
Expect(1, 2678, '\p{^Sc=		Guru}', "");
Expect(1, 2678, '\P{Sc=		Guru}', "");
Expect(0, 2678, '\P{^Sc=		Guru}', "");
Error('\p{Is_Script=	 Gurmukhi/a/}');
Error('\P{Is_Script=	 Gurmukhi/a/}');
Expect(1, 2677, '\p{Is_Script=gurmukhi}', "");
Expect(0, 2677, '\p{^Is_Script=gurmukhi}', "");
Expect(0, 2677, '\P{Is_Script=gurmukhi}', "");
Expect(1, 2677, '\P{^Is_Script=gurmukhi}', "");
Expect(0, 2678, '\p{Is_Script=gurmukhi}', "");
Expect(1, 2678, '\p{^Is_Script=gurmukhi}', "");
Expect(1, 2678, '\P{Is_Script=gurmukhi}', "");
Expect(0, 2678, '\P{^Is_Script=gurmukhi}', "");
Expect(1, 2677, '\p{Is_Script=_-Gurmukhi}', "");
Expect(0, 2677, '\p{^Is_Script=_-Gurmukhi}', "");
Expect(0, 2677, '\P{Is_Script=_-Gurmukhi}', "");
Expect(1, 2677, '\P{^Is_Script=_-Gurmukhi}', "");
Expect(0, 2678, '\p{Is_Script=_-Gurmukhi}', "");
Expect(1, 2678, '\p{^Is_Script=_-Gurmukhi}', "");
Expect(1, 2678, '\P{Is_Script=_-Gurmukhi}', "");
Expect(0, 2678, '\P{^Is_Script=_-Gurmukhi}', "");
Error('\p{Is_Sc=-/a/Guru}');
Error('\P{Is_Sc=-/a/Guru}');
Expect(1, 2677, '\p{Is_Sc=guru}', "");
Expect(0, 2677, '\p{^Is_Sc=guru}', "");
Expect(0, 2677, '\P{Is_Sc=guru}', "");
Expect(1, 2677, '\P{^Is_Sc=guru}', "");
Expect(0, 2678, '\p{Is_Sc=guru}', "");
Expect(1, 2678, '\p{^Is_Sc=guru}', "");
Expect(1, 2678, '\P{Is_Sc=guru}', "");
Expect(0, 2678, '\P{^Is_Sc=guru}', "");
Expect(1, 2677, '\p{Is_Sc=Guru}', "");
Expect(0, 2677, '\p{^Is_Sc=Guru}', "");
Expect(0, 2677, '\P{Is_Sc=Guru}', "");
Expect(1, 2677, '\P{^Is_Sc=Guru}', "");
Expect(0, 2678, '\p{Is_Sc=Guru}', "");
Expect(1, 2678, '\p{^Is_Sc=Guru}', "");
Expect(1, 2678, '\P{Is_Sc=Guru}', "");
Expect(0, 2678, '\P{^Is_Sc=Guru}', "");
Error('\p{Script=	:=Hangul}');
Error('\P{Script=	:=Hangul}');
Expect(1, 65500, '\p{Script=hangul}', "");
Expect(0, 65500, '\p{^Script=hangul}', "");
Expect(0, 65500, '\P{Script=hangul}', "");
Expect(1, 65500, '\P{^Script=hangul}', "");
Expect(0, 65501, '\p{Script=hangul}', "");
Expect(1, 65501, '\p{^Script=hangul}', "");
Expect(1, 65501, '\P{Script=hangul}', "");
Expect(0, 65501, '\P{^Script=hangul}', "");
Expect(1, 65500, '\p{Script=-	hangul}', "");
Expect(0, 65500, '\p{^Script=-	hangul}', "");
Expect(0, 65500, '\P{Script=-	hangul}', "");
Expect(1, 65500, '\P{^Script=-	hangul}', "");
Expect(0, 65501, '\p{Script=-	hangul}', "");
Expect(1, 65501, '\p{^Script=-	hangul}', "");
Expect(1, 65501, '\P{Script=-	hangul}', "");
Expect(0, 65501, '\P{^Script=-	hangul}', "");
Error('\p{Sc= /a/hang}');
Error('\P{Sc= /a/hang}');
Expect(1, 65500, '\p{Sc=hang}', "");
Expect(0, 65500, '\p{^Sc=hang}', "");
Expect(0, 65500, '\P{Sc=hang}', "");
Expect(1, 65500, '\P{^Sc=hang}', "");
Expect(0, 65501, '\p{Sc=hang}', "");
Expect(1, 65501, '\p{^Sc=hang}', "");
Expect(1, 65501, '\P{Sc=hang}', "");
Expect(0, 65501, '\P{^Sc=hang}', "");
Expect(1, 65500, '\p{Sc=-Hang}', "");
Expect(0, 65500, '\p{^Sc=-Hang}', "");
Expect(0, 65500, '\P{Sc=-Hang}', "");
Expect(1, 65500, '\P{^Sc=-Hang}', "");
Expect(0, 65501, '\p{Sc=-Hang}', "");
Expect(1, 65501, '\p{^Sc=-Hang}', "");
Expect(1, 65501, '\P{Sc=-Hang}', "");
Expect(0, 65501, '\P{^Sc=-Hang}', "");
Error('\p{Is_Script= /a/Hangul}');
Error('\P{Is_Script= /a/Hangul}');
Expect(1, 65500, '\p{Is_Script=hangul}', "");
Expect(0, 65500, '\p{^Is_Script=hangul}', "");
Expect(0, 65500, '\P{Is_Script=hangul}', "");
Expect(1, 65500, '\P{^Is_Script=hangul}', "");
Expect(0, 65501, '\p{Is_Script=hangul}', "");
Expect(1, 65501, '\p{^Is_Script=hangul}', "");
Expect(1, 65501, '\P{Is_Script=hangul}', "");
Expect(0, 65501, '\P{^Is_Script=hangul}', "");
Expect(1, 65500, '\p{Is_Script=- Hangul}', "");
Expect(0, 65500, '\p{^Is_Script=- Hangul}', "");
Expect(0, 65500, '\P{Is_Script=- Hangul}', "");
Expect(1, 65500, '\P{^Is_Script=- Hangul}', "");
Expect(0, 65501, '\p{Is_Script=- Hangul}', "");
Expect(1, 65501, '\p{^Is_Script=- Hangul}', "");
Expect(1, 65501, '\P{Is_Script=- Hangul}', "");
Expect(0, 65501, '\P{^Is_Script=- Hangul}', "");
Error('\p{Is_Sc=-:=Hang}');
Error('\P{Is_Sc=-:=Hang}');
Expect(1, 65500, '\p{Is_Sc=hang}', "");
Expect(0, 65500, '\p{^Is_Sc=hang}', "");
Expect(0, 65500, '\P{Is_Sc=hang}', "");
Expect(1, 65500, '\P{^Is_Sc=hang}', "");
Expect(0, 65501, '\p{Is_Sc=hang}', "");
Expect(1, 65501, '\p{^Is_Sc=hang}', "");
Expect(1, 65501, '\P{Is_Sc=hang}', "");
Expect(0, 65501, '\P{^Is_Sc=hang}', "");
Expect(1, 65500, '\p{Is_Sc=	HANG}', "");
Expect(0, 65500, '\p{^Is_Sc=	HANG}', "");
Expect(0, 65500, '\P{Is_Sc=	HANG}', "");
Expect(1, 65500, '\P{^Is_Sc=	HANG}', "");
Expect(0, 65501, '\p{Is_Sc=	HANG}', "");
Expect(1, 65501, '\p{^Is_Sc=	HANG}', "");
Expect(1, 65501, '\P{Is_Sc=	HANG}', "");
Expect(0, 65501, '\P{^Is_Sc=	HANG}', "");
Error('\p{Script=-/a/Han}');
Error('\P{Script=-/a/Han}');
Expect(1, 195101, '\p{Script=han}', "");
Expect(0, 195101, '\p{^Script=han}', "");
Expect(0, 195101, '\P{Script=han}', "");
Expect(1, 195101, '\P{^Script=han}', "");
Expect(0, 195102, '\p{Script=han}', "");
Expect(1, 195102, '\p{^Script=han}', "");
Expect(1, 195102, '\P{Script=han}', "");
Expect(0, 195102, '\P{^Script=han}', "");
Expect(1, 195101, '\p{Script= HAN}', "");
Expect(0, 195101, '\p{^Script= HAN}', "");
Expect(0, 195101, '\P{Script= HAN}', "");
Expect(1, 195101, '\P{^Script= HAN}', "");
Expect(0, 195102, '\p{Script= HAN}', "");
Expect(1, 195102, '\p{^Script= HAN}', "");
Expect(1, 195102, '\P{Script= HAN}', "");
Expect(0, 195102, '\P{^Script= HAN}', "");
Error('\p{Sc=:=-	Hani}');
Error('\P{Sc=:=-	Hani}');
Expect(1, 195101, '\p{Sc=hani}', "");
Expect(0, 195101, '\p{^Sc=hani}', "");
Expect(0, 195101, '\P{Sc=hani}', "");
Expect(1, 195101, '\P{^Sc=hani}', "");
Expect(0, 195102, '\p{Sc=hani}', "");
Expect(1, 195102, '\p{^Sc=hani}', "");
Expect(1, 195102, '\P{Sc=hani}', "");
Expect(0, 195102, '\P{^Sc=hani}', "");
Expect(1, 195101, '\p{Sc=	hani}', "");
Expect(0, 195101, '\p{^Sc=	hani}', "");
Expect(0, 195101, '\P{Sc=	hani}', "");
Expect(1, 195101, '\P{^Sc=	hani}', "");
Expect(0, 195102, '\p{Sc=	hani}', "");
Expect(1, 195102, '\p{^Sc=	hani}', "");
Expect(1, 195102, '\P{Sc=	hani}', "");
Expect(0, 195102, '\P{^Sc=	hani}', "");
Error('\p{Is_Script=_/a/Han}');
Error('\P{Is_Script=_/a/Han}');
Expect(1, 195101, '\p{Is_Script=han}', "");
Expect(0, 195101, '\p{^Is_Script=han}', "");
Expect(0, 195101, '\P{Is_Script=han}', "");
Expect(1, 195101, '\P{^Is_Script=han}', "");
Expect(0, 195102, '\p{Is_Script=han}', "");
Expect(1, 195102, '\p{^Is_Script=han}', "");
Expect(1, 195102, '\P{Is_Script=han}', "");
Expect(0, 195102, '\P{^Is_Script=han}', "");
Expect(1, 195101, '\p{Is_Script=_-HAN}', "");
Expect(0, 195101, '\p{^Is_Script=_-HAN}', "");
Expect(0, 195101, '\P{Is_Script=_-HAN}', "");
Expect(1, 195101, '\P{^Is_Script=_-HAN}', "");
Expect(0, 195102, '\p{Is_Script=_-HAN}', "");
Expect(1, 195102, '\p{^Is_Script=_-HAN}', "");
Expect(1, 195102, '\P{Is_Script=_-HAN}', "");
Expect(0, 195102, '\P{^Is_Script=_-HAN}', "");
Error('\p{Is_Sc=_:=Hani}');
Error('\P{Is_Sc=_:=Hani}');
Expect(1, 195101, '\p{Is_Sc=hani}', "");
Expect(0, 195101, '\p{^Is_Sc=hani}', "");
Expect(0, 195101, '\P{Is_Sc=hani}', "");
Expect(1, 195101, '\P{^Is_Sc=hani}', "");
Expect(0, 195102, '\p{Is_Sc=hani}', "");
Expect(1, 195102, '\p{^Is_Sc=hani}', "");
Expect(1, 195102, '\P{Is_Sc=hani}', "");
Expect(0, 195102, '\P{^Is_Sc=hani}', "");
Expect(1, 195101, '\p{Is_Sc=-_Hani}', "");
Expect(0, 195101, '\p{^Is_Sc=-_Hani}', "");
Expect(0, 195101, '\P{Is_Sc=-_Hani}', "");
Expect(1, 195101, '\P{^Is_Sc=-_Hani}', "");
Expect(0, 195102, '\p{Is_Sc=-_Hani}', "");
Expect(1, 195102, '\p{^Is_Sc=-_Hani}', "");
Expect(1, 195102, '\P{Is_Sc=-_Hani}', "");
Expect(0, 195102, '\P{^Is_Sc=-_Hani}', "");
Error('\p{Script=_:=Hanunoo}');
Error('\P{Script=_:=Hanunoo}');
Expect(1, 5940, '\p{Script=hanunoo}', "");
Expect(0, 5940, '\p{^Script=hanunoo}', "");
Expect(0, 5940, '\P{Script=hanunoo}', "");
Expect(1, 5940, '\P{^Script=hanunoo}', "");
Expect(0, 5941, '\p{Script=hanunoo}', "");
Expect(1, 5941, '\p{^Script=hanunoo}', "");
Expect(1, 5941, '\P{Script=hanunoo}', "");
Expect(0, 5941, '\P{^Script=hanunoo}', "");
Expect(1, 5940, '\p{Script=	_hanunoo}', "");
Expect(0, 5940, '\p{^Script=	_hanunoo}', "");
Expect(0, 5940, '\P{Script=	_hanunoo}', "");
Expect(1, 5940, '\P{^Script=	_hanunoo}', "");
Expect(0, 5941, '\p{Script=	_hanunoo}', "");
Expect(1, 5941, '\p{^Script=	_hanunoo}', "");
Expect(1, 5941, '\P{Script=	_hanunoo}', "");
Expect(0, 5941, '\P{^Script=	_hanunoo}', "");
Error('\p{Sc=/a/		Hano}');
Error('\P{Sc=/a/		Hano}');
Expect(1, 5940, '\p{Sc=hano}', "");
Expect(0, 5940, '\p{^Sc=hano}', "");
Expect(0, 5940, '\P{Sc=hano}', "");
Expect(1, 5940, '\P{^Sc=hano}', "");
Expect(0, 5941, '\p{Sc=hano}', "");
Expect(1, 5941, '\p{^Sc=hano}', "");
Expect(1, 5941, '\P{Sc=hano}', "");
Expect(0, 5941, '\P{^Sc=hano}', "");
Expect(1, 5940, '\p{Sc=--HANO}', "");
Expect(0, 5940, '\p{^Sc=--HANO}', "");
Expect(0, 5940, '\P{Sc=--HANO}', "");
Expect(1, 5940, '\P{^Sc=--HANO}', "");
Expect(0, 5941, '\p{Sc=--HANO}', "");
Expect(1, 5941, '\p{^Sc=--HANO}', "");
Expect(1, 5941, '\P{Sc=--HANO}', "");
Expect(0, 5941, '\P{^Sc=--HANO}', "");
Error('\p{Is_Script=/a/_	Hanunoo}');
Error('\P{Is_Script=/a/_	Hanunoo}');
Expect(1, 5940, '\p{Is_Script=hanunoo}', "");
Expect(0, 5940, '\p{^Is_Script=hanunoo}', "");
Expect(0, 5940, '\P{Is_Script=hanunoo}', "");
Expect(1, 5940, '\P{^Is_Script=hanunoo}', "");
Expect(0, 5941, '\p{Is_Script=hanunoo}', "");
Expect(1, 5941, '\p{^Is_Script=hanunoo}', "");
Expect(1, 5941, '\P{Is_Script=hanunoo}', "");
Expect(0, 5941, '\P{^Is_Script=hanunoo}', "");
Expect(1, 5940, '\p{Is_Script=	Hanunoo}', "");
Expect(0, 5940, '\p{^Is_Script=	Hanunoo}', "");
Expect(0, 5940, '\P{Is_Script=	Hanunoo}', "");
Expect(1, 5940, '\P{^Is_Script=	Hanunoo}', "");
Expect(0, 5941, '\p{Is_Script=	Hanunoo}', "");
Expect(1, 5941, '\p{^Is_Script=	Hanunoo}', "");
Expect(1, 5941, '\P{Is_Script=	Hanunoo}', "");
Expect(0, 5941, '\P{^Is_Script=	Hanunoo}', "");
Error('\p{Is_Sc=:= 	HANO}');
Error('\P{Is_Sc=:= 	HANO}');
Expect(1, 5940, '\p{Is_Sc=hano}', "");
Expect(0, 5940, '\p{^Is_Sc=hano}', "");
Expect(0, 5940, '\P{Is_Sc=hano}', "");
Expect(1, 5940, '\P{^Is_Sc=hano}', "");
Expect(0, 5941, '\p{Is_Sc=hano}', "");
Expect(1, 5941, '\p{^Is_Sc=hano}', "");
Expect(1, 5941, '\P{Is_Sc=hano}', "");
Expect(0, 5941, '\P{^Is_Sc=hano}', "");
Expect(1, 5940, '\p{Is_Sc=		Hano}', "");
Expect(0, 5940, '\p{^Is_Sc=		Hano}', "");
Expect(0, 5940, '\P{Is_Sc=		Hano}', "");
Expect(1, 5940, '\P{^Is_Sc=		Hano}', "");
Expect(0, 5941, '\p{Is_Sc=		Hano}', "");
Expect(1, 5941, '\p{^Is_Sc=		Hano}', "");
Expect(1, 5941, '\P{Is_Sc=		Hano}', "");
Expect(0, 5941, '\P{^Is_Sc=		Hano}', "");
Error('\p{Script=_/a/Hebrew}');
Error('\P{Script=_/a/Hebrew}');
Expect(1, 64335, '\p{Script=hebrew}', "");
Expect(0, 64335, '\p{^Script=hebrew}', "");
Expect(0, 64335, '\P{Script=hebrew}', "");
Expect(1, 64335, '\P{^Script=hebrew}', "");
Expect(0, 64336, '\p{Script=hebrew}', "");
Expect(1, 64336, '\p{^Script=hebrew}', "");
Expect(1, 64336, '\P{Script=hebrew}', "");
Expect(0, 64336, '\P{^Script=hebrew}', "");
Expect(1, 64335, '\p{Script= 	Hebrew}', "");
Expect(0, 64335, '\p{^Script= 	Hebrew}', "");
Expect(0, 64335, '\P{Script= 	Hebrew}', "");
Expect(1, 64335, '\P{^Script= 	Hebrew}', "");
Expect(0, 64336, '\p{Script= 	Hebrew}', "");
Expect(1, 64336, '\p{^Script= 	Hebrew}', "");
Expect(1, 64336, '\P{Script= 	Hebrew}', "");
Expect(0, 64336, '\P{^Script= 	Hebrew}', "");
Error('\p{Sc=__HEBR:=}');
Error('\P{Sc=__HEBR:=}');
Expect(1, 64335, '\p{Sc=hebr}', "");
Expect(0, 64335, '\p{^Sc=hebr}', "");
Expect(0, 64335, '\P{Sc=hebr}', "");
Expect(1, 64335, '\P{^Sc=hebr}', "");
Expect(0, 64336, '\p{Sc=hebr}', "");
Expect(1, 64336, '\p{^Sc=hebr}', "");
Expect(1, 64336, '\P{Sc=hebr}', "");
Expect(0, 64336, '\P{^Sc=hebr}', "");
Expect(1, 64335, '\p{Sc:		-HEBR}', "");
Expect(0, 64335, '\p{^Sc:		-HEBR}', "");
Expect(0, 64335, '\P{Sc:		-HEBR}', "");
Expect(1, 64335, '\P{^Sc:		-HEBR}', "");
Expect(0, 64336, '\p{Sc:		-HEBR}', "");
Expect(1, 64336, '\p{^Sc:		-HEBR}', "");
Expect(1, 64336, '\P{Sc:		-HEBR}', "");
Expect(0, 64336, '\P{^Sc:		-HEBR}', "");
Error('\p{Is_Script=-:=HEBREW}');
Error('\P{Is_Script=-:=HEBREW}');
Expect(1, 64335, '\p{Is_Script:   hebrew}', "");
Expect(0, 64335, '\p{^Is_Script:   hebrew}', "");
Expect(0, 64335, '\P{Is_Script:   hebrew}', "");
Expect(1, 64335, '\P{^Is_Script:   hebrew}', "");
Expect(0, 64336, '\p{Is_Script:   hebrew}', "");
Expect(1, 64336, '\p{^Is_Script:   hebrew}', "");
Expect(1, 64336, '\P{Is_Script:   hebrew}', "");
Expect(0, 64336, '\P{^Is_Script:   hebrew}', "");
Expect(1, 64335, '\p{Is_Script: _Hebrew}', "");
Expect(0, 64335, '\p{^Is_Script: _Hebrew}', "");
Expect(0, 64335, '\P{Is_Script: _Hebrew}', "");
Expect(1, 64335, '\P{^Is_Script: _Hebrew}', "");
Expect(0, 64336, '\p{Is_Script: _Hebrew}', "");
Expect(1, 64336, '\p{^Is_Script: _Hebrew}', "");
Expect(1, 64336, '\P{Is_Script: _Hebrew}', "");
Expect(0, 64336, '\P{^Is_Script: _Hebrew}', "");
Error('\p{Is_Sc=_Hebr:=}');
Error('\P{Is_Sc=_Hebr:=}');
Expect(1, 64335, '\p{Is_Sc=hebr}', "");
Expect(0, 64335, '\p{^Is_Sc=hebr}', "");
Expect(0, 64335, '\P{Is_Sc=hebr}', "");
Expect(1, 64335, '\P{^Is_Sc=hebr}', "");
Expect(0, 64336, '\p{Is_Sc=hebr}', "");
Expect(1, 64336, '\p{^Is_Sc=hebr}', "");
Expect(1, 64336, '\P{Is_Sc=hebr}', "");
Expect(0, 64336, '\P{^Is_Sc=hebr}', "");
Expect(1, 64335, '\p{Is_Sc=_ Hebr}', "");
Expect(0, 64335, '\p{^Is_Sc=_ Hebr}', "");
Expect(0, 64335, '\P{Is_Sc=_ Hebr}', "");
Expect(1, 64335, '\P{^Is_Sc=_ Hebr}', "");
Expect(0, 64336, '\p{Is_Sc=_ Hebr}', "");
Expect(1, 64336, '\p{^Is_Sc=_ Hebr}', "");
Expect(1, 64336, '\P{Is_Sc=_ Hebr}', "");
Expect(0, 64336, '\P{^Is_Sc=_ Hebr}', "");
Error('\p{Script=:=_Hiragana}');
Error('\P{Script=:=_Hiragana}');
Expect(1, 127488, '\p{Script=hiragana}', "");
Expect(0, 127488, '\p{^Script=hiragana}', "");
Expect(0, 127488, '\P{Script=hiragana}', "");
Expect(1, 127488, '\P{^Script=hiragana}', "");
Expect(0, 127489, '\p{Script=hiragana}', "");
Expect(1, 127489, '\p{^Script=hiragana}', "");
Expect(1, 127489, '\P{Script=hiragana}', "");
Expect(0, 127489, '\P{^Script=hiragana}', "");
Expect(1, 127488, '\p{Script:   	-hiragana}', "");
Expect(0, 127488, '\p{^Script:   	-hiragana}', "");
Expect(0, 127488, '\P{Script:   	-hiragana}', "");
Expect(1, 127488, '\P{^Script:   	-hiragana}', "");
Expect(0, 127489, '\p{Script:   	-hiragana}', "");
Expect(1, 127489, '\p{^Script:   	-hiragana}', "");
Expect(1, 127489, '\P{Script:   	-hiragana}', "");
Expect(0, 127489, '\P{^Script:   	-hiragana}', "");
Error('\p{Sc:_hira/a/}');
Error('\P{Sc:_hira/a/}');
Expect(1, 127488, '\p{Sc=hira}', "");
Expect(0, 127488, '\p{^Sc=hira}', "");
Expect(0, 127488, '\P{Sc=hira}', "");
Expect(1, 127488, '\P{^Sc=hira}', "");
Expect(0, 127489, '\p{Sc=hira}', "");
Expect(1, 127489, '\p{^Sc=hira}', "");
Expect(1, 127489, '\P{Sc=hira}', "");
Expect(0, 127489, '\P{^Sc=hira}', "");
Expect(1, 127488, '\p{Sc: 	-HIRA}', "");
Expect(0, 127488, '\p{^Sc: 	-HIRA}', "");
Expect(0, 127488, '\P{Sc: 	-HIRA}', "");
Expect(1, 127488, '\P{^Sc: 	-HIRA}', "");
Expect(0, 127489, '\p{Sc: 	-HIRA}', "");
Expect(1, 127489, '\p{^Sc: 	-HIRA}', "");
Expect(1, 127489, '\P{Sc: 	-HIRA}', "");
Expect(0, 127489, '\P{^Sc: 	-HIRA}', "");
Error('\p{Is_Script=:=	 Hiragana}');
Error('\P{Is_Script=:=	 Hiragana}');
Expect(1, 127488, '\p{Is_Script=hiragana}', "");
Expect(0, 127488, '\p{^Is_Script=hiragana}', "");
Expect(0, 127488, '\P{Is_Script=hiragana}', "");
Expect(1, 127488, '\P{^Is_Script=hiragana}', "");
Expect(0, 127489, '\p{Is_Script=hiragana}', "");
Expect(1, 127489, '\p{^Is_Script=hiragana}', "");
Expect(1, 127489, '\P{Is_Script=hiragana}', "");
Expect(0, 127489, '\P{^Is_Script=hiragana}', "");
Expect(1, 127488, '\p{Is_Script=_ hiragana}', "");
Expect(0, 127488, '\p{^Is_Script=_ hiragana}', "");
Expect(0, 127488, '\P{Is_Script=_ hiragana}', "");
Expect(1, 127488, '\P{^Is_Script=_ hiragana}', "");
Expect(0, 127489, '\p{Is_Script=_ hiragana}', "");
Expect(1, 127489, '\p{^Is_Script=_ hiragana}', "");
Expect(1, 127489, '\P{Is_Script=_ hiragana}', "");
Expect(0, 127489, '\P{^Is_Script=_ hiragana}', "");
Error('\p{Is_Sc=	/a/HIRA}');
Error('\P{Is_Sc=	/a/HIRA}');
Expect(1, 127488, '\p{Is_Sc=hira}', "");
Expect(0, 127488, '\p{^Is_Sc=hira}', "");
Expect(0, 127488, '\P{Is_Sc=hira}', "");
Expect(1, 127488, '\P{^Is_Sc=hira}', "");
Expect(0, 127489, '\p{Is_Sc=hira}', "");
Expect(1, 127489, '\p{^Is_Sc=hira}', "");
Expect(1, 127489, '\P{Is_Sc=hira}', "");
Expect(0, 127489, '\P{^Is_Sc=hira}', "");
Expect(1, 127488, '\p{Is_Sc=HIRA}', "");
Expect(0, 127488, '\p{^Is_Sc=HIRA}', "");
Expect(0, 127488, '\P{Is_Sc=HIRA}', "");
Expect(1, 127488, '\P{^Is_Sc=HIRA}', "");
Expect(0, 127489, '\p{Is_Sc=HIRA}', "");
Expect(1, 127489, '\p{^Is_Sc=HIRA}', "");
Expect(1, 127489, '\P{Is_Sc=HIRA}', "");
Expect(0, 127489, '\P{^Is_Sc=HIRA}', "");
Error('\p{Script=Katakana_Or_Hiragana}');
Error('\P{Script=Katakana_Or_Hiragana}');
Error('\p{Sc=Hrkt}');
Error('\P{Sc=Hrkt}');
Error('\p{Is_Script=Katakana_Or_Hiragana}');
Error('\P{Is_Script=Katakana_Or_Hiragana}');
Error('\p{Is_Sc=Hrkt}');
Error('\P{Is_Sc=Hrkt}');
Error('\p{Script=	/a/Old_italic}');
Error('\P{Script=	/a/Old_italic}');
Expect(1, 66339, '\p{Script=olditalic}', "");
Expect(0, 66339, '\p{^Script=olditalic}', "");
Expect(0, 66339, '\P{Script=olditalic}', "");
Expect(1, 66339, '\P{^Script=olditalic}', "");
Expect(0, 66340, '\p{Script=olditalic}', "");
Expect(1, 66340, '\p{^Script=olditalic}', "");
Expect(1, 66340, '\P{Script=olditalic}', "");
Expect(0, 66340, '\P{^Script=olditalic}', "");
Expect(1, 66339, '\p{Script=		Old_italic}', "");
Expect(0, 66339, '\p{^Script=		Old_italic}', "");
Expect(0, 66339, '\P{Script=		Old_italic}', "");
Expect(1, 66339, '\P{^Script=		Old_italic}', "");
Expect(0, 66340, '\p{Script=		Old_italic}', "");
Expect(1, 66340, '\p{^Script=		Old_italic}', "");
Expect(1, 66340, '\P{Script=		Old_italic}', "");
Expect(0, 66340, '\P{^Script=		Old_italic}', "");
Error('\p{Sc= /a/Ital}');
Error('\P{Sc= /a/Ital}');
Expect(1, 66339, '\p{Sc=ital}', "");
Expect(0, 66339, '\p{^Sc=ital}', "");
Expect(0, 66339, '\P{Sc=ital}', "");
Expect(1, 66339, '\P{^Sc=ital}', "");
Expect(0, 66340, '\p{Sc=ital}', "");
Expect(1, 66340, '\p{^Sc=ital}', "");
Expect(1, 66340, '\P{Sc=ital}', "");
Expect(0, 66340, '\P{^Sc=ital}', "");
Expect(1, 66339, '\p{Sc=-	Ital}', "");
Expect(0, 66339, '\p{^Sc=-	Ital}', "");
Expect(0, 66339, '\P{Sc=-	Ital}', "");
Expect(1, 66339, '\P{^Sc=-	Ital}', "");
Expect(0, 66340, '\p{Sc=-	Ital}', "");
Expect(1, 66340, '\p{^Sc=-	Ital}', "");
Expect(1, 66340, '\P{Sc=-	Ital}', "");
Expect(0, 66340, '\P{^Sc=-	Ital}', "");
Error('\p{Is_Script=-/a/Old_ITALIC}');
Error('\P{Is_Script=-/a/Old_ITALIC}');
Expect(1, 66339, '\p{Is_Script=olditalic}', "");
Expect(0, 66339, '\p{^Is_Script=olditalic}', "");
Expect(0, 66339, '\P{Is_Script=olditalic}', "");
Expect(1, 66339, '\P{^Is_Script=olditalic}', "");
Expect(0, 66340, '\p{Is_Script=olditalic}', "");
Expect(1, 66340, '\p{^Is_Script=olditalic}', "");
Expect(1, 66340, '\P{Is_Script=olditalic}', "");
Expect(0, 66340, '\P{^Is_Script=olditalic}', "");
Expect(1, 66339, '\p{Is_Script=_ Old_Italic}', "");
Expect(0, 66339, '\p{^Is_Script=_ Old_Italic}', "");
Expect(0, 66339, '\P{Is_Script=_ Old_Italic}', "");
Expect(1, 66339, '\P{^Is_Script=_ Old_Italic}', "");
Expect(0, 66340, '\p{Is_Script=_ Old_Italic}', "");
Expect(1, 66340, '\p{^Is_Script=_ Old_Italic}', "");
Expect(1, 66340, '\P{Is_Script=_ Old_Italic}', "");
Expect(0, 66340, '\P{^Is_Script=_ Old_Italic}', "");
Error('\p{Is_Sc=_Ital:=}');
Error('\P{Is_Sc=_Ital:=}');
Expect(1, 66339, '\p{Is_Sc=ital}', "");
Expect(0, 66339, '\p{^Is_Sc=ital}', "");
Expect(0, 66339, '\P{Is_Sc=ital}', "");
Expect(1, 66339, '\P{^Is_Sc=ital}', "");
Expect(0, 66340, '\p{Is_Sc=ital}', "");
Expect(1, 66340, '\p{^Is_Sc=ital}', "");
Expect(1, 66340, '\P{Is_Sc=ital}', "");
Expect(0, 66340, '\P{^Is_Sc=ital}', "");
Expect(1, 66339, '\p{Is_Sc:-	ITAL}', "");
Expect(0, 66339, '\p{^Is_Sc:-	ITAL}', "");
Expect(0, 66339, '\P{Is_Sc:-	ITAL}', "");
Expect(1, 66339, '\P{^Is_Sc:-	ITAL}', "");
Expect(0, 66340, '\p{Is_Sc:-	ITAL}', "");
Expect(1, 66340, '\p{^Is_Sc:-	ITAL}', "");
Expect(1, 66340, '\P{Is_Sc:-	ITAL}', "");
Expect(0, 66340, '\P{^Is_Sc:-	ITAL}', "");
Error('\p{Script=/a/		javanese}');
Error('\P{Script=/a/		javanese}');
Expect(1, 43487, '\p{Script=javanese}', "");
Expect(0, 43487, '\p{^Script=javanese}', "");
Expect(0, 43487, '\P{Script=javanese}', "");
Expect(1, 43487, '\P{^Script=javanese}', "");
Expect(0, 43488, '\p{Script=javanese}', "");
Expect(1, 43488, '\p{^Script=javanese}', "");
Expect(1, 43488, '\P{Script=javanese}', "");
Expect(0, 43488, '\P{^Script=javanese}', "");
Expect(1, 43487, '\p{Script=-_Javanese}', "");
Expect(0, 43487, '\p{^Script=-_Javanese}', "");
Expect(0, 43487, '\P{Script=-_Javanese}', "");
Expect(1, 43487, '\P{^Script=-_Javanese}', "");
Expect(0, 43488, '\p{Script=-_Javanese}', "");
Expect(1, 43488, '\p{^Script=-_Javanese}', "");
Expect(1, 43488, '\P{Script=-_Javanese}', "");
Expect(0, 43488, '\P{^Script=-_Javanese}', "");
Error('\p{Sc= :=Java}');
Error('\P{Sc= :=Java}');
Expect(1, 43487, '\p{Sc=java}', "");
Expect(0, 43487, '\p{^Sc=java}', "");
Expect(0, 43487, '\P{Sc=java}', "");
Expect(1, 43487, '\P{^Sc=java}', "");
Expect(0, 43488, '\p{Sc=java}', "");
Expect(1, 43488, '\p{^Sc=java}', "");
Expect(1, 43488, '\P{Sc=java}', "");
Expect(0, 43488, '\P{^Sc=java}', "");
Expect(1, 43487, '\p{Sc=_-JAVA}', "");
Expect(0, 43487, '\p{^Sc=_-JAVA}', "");
Expect(0, 43487, '\P{Sc=_-JAVA}', "");
Expect(1, 43487, '\P{^Sc=_-JAVA}', "");
Expect(0, 43488, '\p{Sc=_-JAVA}', "");
Expect(1, 43488, '\p{^Sc=_-JAVA}', "");
Expect(1, 43488, '\P{Sc=_-JAVA}', "");
Expect(0, 43488, '\P{^Sc=_-JAVA}', "");
Error('\p{Is_Script=	Javanese/a/}');
Error('\P{Is_Script=	Javanese/a/}');
Expect(1, 43487, '\p{Is_Script=javanese}', "");
Expect(0, 43487, '\p{^Is_Script=javanese}', "");
Expect(0, 43487, '\P{Is_Script=javanese}', "");
Expect(1, 43487, '\P{^Is_Script=javanese}', "");
Expect(0, 43488, '\p{Is_Script=javanese}', "");
Expect(1, 43488, '\p{^Is_Script=javanese}', "");
Expect(1, 43488, '\P{Is_Script=javanese}', "");
Expect(0, 43488, '\P{^Is_Script=javanese}', "");
Expect(1, 43487, '\p{Is_Script= 	javanese}', "");
Expect(0, 43487, '\p{^Is_Script= 	javanese}', "");
Expect(0, 43487, '\P{Is_Script= 	javanese}', "");
Expect(1, 43487, '\P{^Is_Script= 	javanese}', "");
Expect(0, 43488, '\p{Is_Script= 	javanese}', "");
Expect(1, 43488, '\p{^Is_Script= 	javanese}', "");
Expect(1, 43488, '\P{Is_Script= 	javanese}', "");
Expect(0, 43488, '\P{^Is_Script= 	javanese}', "");
Error('\p{Is_Sc=:=Java}');
Error('\P{Is_Sc=:=Java}');
Expect(1, 43487, '\p{Is_Sc=java}', "");
Expect(0, 43487, '\p{^Is_Sc=java}', "");
Expect(0, 43487, '\P{Is_Sc=java}', "");
Expect(1, 43487, '\P{^Is_Sc=java}', "");
Expect(0, 43488, '\p{Is_Sc=java}', "");
Expect(1, 43488, '\p{^Is_Sc=java}', "");
Expect(1, 43488, '\P{Is_Sc=java}', "");
Expect(0, 43488, '\P{^Is_Sc=java}', "");
Expect(1, 43487, '\p{Is_Sc=-Java}', "");
Expect(0, 43487, '\p{^Is_Sc=-Java}', "");
Expect(0, 43487, '\P{Is_Sc=-Java}', "");
Expect(1, 43487, '\P{^Is_Sc=-Java}', "");
Expect(0, 43488, '\p{Is_Sc=-Java}', "");
Expect(1, 43488, '\p{^Is_Sc=-Java}', "");
Expect(1, 43488, '\P{Is_Sc=-Java}', "");
Expect(0, 43488, '\P{^Is_Sc=-Java}', "");
Error('\p{Script=-	Kayah_li/a/}');
Error('\P{Script=-	Kayah_li/a/}');
Expect(1, 43311, '\p{Script=kayahli}', "");
Expect(0, 43311, '\p{^Script=kayahli}', "");
Expect(0, 43311, '\P{Script=kayahli}', "");
Expect(1, 43311, '\P{^Script=kayahli}', "");
Expect(0, 43312, '\p{Script=kayahli}', "");
Expect(1, 43312, '\p{^Script=kayahli}', "");
Expect(1, 43312, '\P{Script=kayahli}', "");
Expect(0, 43312, '\P{^Script=kayahli}', "");
Expect(1, 43311, '\p{Script:--kayah_LI}', "");
Expect(0, 43311, '\p{^Script:--kayah_LI}', "");
Expect(0, 43311, '\P{Script:--kayah_LI}', "");
Expect(1, 43311, '\P{^Script:--kayah_LI}', "");
Expect(0, 43312, '\p{Script:--kayah_LI}', "");
Expect(1, 43312, '\p{^Script:--kayah_LI}', "");
Expect(1, 43312, '\P{Script:--kayah_LI}', "");
Expect(0, 43312, '\P{^Script:--kayah_LI}', "");
Error('\p{Sc=/a/kali}');
Error('\P{Sc=/a/kali}');
Expect(1, 43311, '\p{Sc=kali}', "");
Expect(0, 43311, '\p{^Sc=kali}', "");
Expect(0, 43311, '\P{Sc=kali}', "");
Expect(1, 43311, '\P{^Sc=kali}', "");
Expect(0, 43312, '\p{Sc=kali}', "");
Expect(1, 43312, '\p{^Sc=kali}', "");
Expect(1, 43312, '\P{Sc=kali}', "");
Expect(0, 43312, '\P{^Sc=kali}', "");
Expect(1, 43311, '\p{Sc=	-KALI}', "");
Expect(0, 43311, '\p{^Sc=	-KALI}', "");
Expect(0, 43311, '\P{Sc=	-KALI}', "");
Expect(1, 43311, '\P{^Sc=	-KALI}', "");
Expect(0, 43312, '\p{Sc=	-KALI}', "");
Expect(1, 43312, '\p{^Sc=	-KALI}', "");
Expect(1, 43312, '\P{Sc=	-KALI}', "");
Expect(0, 43312, '\P{^Sc=	-KALI}', "");
Error('\p{Is_Script= kayah_li/a/}');
Error('\P{Is_Script= kayah_li/a/}');
Expect(1, 43311, '\p{Is_Script=kayahli}', "");
Expect(0, 43311, '\p{^Is_Script=kayahli}', "");
Expect(0, 43311, '\P{Is_Script=kayahli}', "");
Expect(1, 43311, '\P{^Is_Script=kayahli}', "");
Expect(0, 43312, '\p{Is_Script=kayahli}', "");
Expect(1, 43312, '\p{^Is_Script=kayahli}', "");
Expect(1, 43312, '\P{Is_Script=kayahli}', "");
Expect(0, 43312, '\P{^Is_Script=kayahli}', "");
Expect(1, 43311, '\p{Is_Script:_KAYAH_li}', "");
Expect(0, 43311, '\p{^Is_Script:_KAYAH_li}', "");
Expect(0, 43311, '\P{Is_Script:_KAYAH_li}', "");
Expect(1, 43311, '\P{^Is_Script:_KAYAH_li}', "");
Expect(0, 43312, '\p{Is_Script:_KAYAH_li}', "");
Expect(1, 43312, '\p{^Is_Script:_KAYAH_li}', "");
Expect(1, 43312, '\P{Is_Script:_KAYAH_li}', "");
Expect(0, 43312, '\P{^Is_Script:_KAYAH_li}', "");
Error('\p{Is_Sc=_ KALI/a/}');
Error('\P{Is_Sc=_ KALI/a/}');
Expect(1, 43311, '\p{Is_Sc=kali}', "");
Expect(0, 43311, '\p{^Is_Sc=kali}', "");
Expect(0, 43311, '\P{Is_Sc=kali}', "");
Expect(1, 43311, '\P{^Is_Sc=kali}', "");
Expect(0, 43312, '\p{Is_Sc=kali}', "");
Expect(1, 43312, '\p{^Is_Sc=kali}', "");
Expect(1, 43312, '\P{Is_Sc=kali}', "");
Expect(0, 43312, '\P{^Is_Sc=kali}', "");
Expect(1, 43311, '\p{Is_Sc=		Kali}', "");
Expect(0, 43311, '\p{^Is_Sc=		Kali}', "");
Expect(0, 43311, '\P{Is_Sc=		Kali}', "");
Expect(1, 43311, '\P{^Is_Sc=		Kali}', "");
Expect(0, 43312, '\p{Is_Sc=		Kali}', "");
Expect(1, 43312, '\p{^Is_Sc=		Kali}', "");
Expect(1, 43312, '\P{Is_Sc=		Kali}', "");
Expect(0, 43312, '\P{^Is_Sc=		Kali}', "");
Error('\p{Script:__KATAKANA/a/}');
Error('\P{Script:__KATAKANA/a/}');
Expect(1, 110592, '\p{Script=katakana}', "");
Expect(0, 110592, '\p{^Script=katakana}', "");
Expect(0, 110592, '\P{Script=katakana}', "");
Expect(1, 110592, '\P{^Script=katakana}', "");
Expect(0, 110593, '\p{Script=katakana}', "");
Expect(1, 110593, '\p{^Script=katakana}', "");
Expect(1, 110593, '\P{Script=katakana}', "");
Expect(0, 110593, '\P{^Script=katakana}', "");
Expect(1, 110592, '\p{Script=__katakana}', "");
Expect(0, 110592, '\p{^Script=__katakana}', "");
Expect(0, 110592, '\P{Script=__katakana}', "");
Expect(1, 110592, '\P{^Script=__katakana}', "");
Expect(0, 110593, '\p{Script=__katakana}', "");
Expect(1, 110593, '\p{^Script=__katakana}', "");
Expect(1, 110593, '\P{Script=__katakana}', "");
Expect(0, 110593, '\P{^Script=__katakana}', "");
Error('\p{Sc=:=-_kana}');
Error('\P{Sc=:=-_kana}');
Expect(1, 110592, '\p{Sc=kana}', "");
Expect(0, 110592, '\p{^Sc=kana}', "");
Expect(0, 110592, '\P{Sc=kana}', "");
Expect(1, 110592, '\P{^Sc=kana}', "");
Expect(0, 110593, '\p{Sc=kana}', "");
Expect(1, 110593, '\p{^Sc=kana}', "");
Expect(1, 110593, '\P{Sc=kana}', "");
Expect(0, 110593, '\P{^Sc=kana}', "");
Expect(1, 110592, '\p{Sc= Kana}', "");
Expect(0, 110592, '\p{^Sc= Kana}', "");
Expect(0, 110592, '\P{Sc= Kana}', "");
Expect(1, 110592, '\P{^Sc= Kana}', "");
Expect(0, 110593, '\p{Sc= Kana}', "");
Expect(1, 110593, '\p{^Sc= Kana}', "");
Expect(1, 110593, '\P{Sc= Kana}', "");
Expect(0, 110593, '\P{^Sc= Kana}', "");
Error('\p{Is_Script= Katakana:=}');
Error('\P{Is_Script= Katakana:=}');
Expect(1, 110592, '\p{Is_Script=katakana}', "");
Expect(0, 110592, '\p{^Is_Script=katakana}', "");
Expect(0, 110592, '\P{Is_Script=katakana}', "");
Expect(1, 110592, '\P{^Is_Script=katakana}', "");
Expect(0, 110593, '\p{Is_Script=katakana}', "");
Expect(1, 110593, '\p{^Is_Script=katakana}', "");
Expect(1, 110593, '\P{Is_Script=katakana}', "");
Expect(0, 110593, '\P{^Is_Script=katakana}', "");
Expect(1, 110592, '\p{Is_Script=	_KATAKANA}', "");
Expect(0, 110592, '\p{^Is_Script=	_KATAKANA}', "");
Expect(0, 110592, '\P{Is_Script=	_KATAKANA}', "");
Expect(1, 110592, '\P{^Is_Script=	_KATAKANA}', "");
Expect(0, 110593, '\p{Is_Script=	_KATAKANA}', "");
Expect(1, 110593, '\p{^Is_Script=	_KATAKANA}', "");
Expect(1, 110593, '\P{Is_Script=	_KATAKANA}', "");
Expect(0, 110593, '\P{^Is_Script=	_KATAKANA}', "");
Error('\p{Is_Sc=/a/-Kana}');
Error('\P{Is_Sc=/a/-Kana}');
Expect(1, 110592, '\p{Is_Sc=kana}', "");
Expect(0, 110592, '\p{^Is_Sc=kana}', "");
Expect(0, 110592, '\P{Is_Sc=kana}', "");
Expect(1, 110592, '\P{^Is_Sc=kana}', "");
Expect(0, 110593, '\p{Is_Sc=kana}', "");
Expect(1, 110593, '\p{^Is_Sc=kana}', "");
Expect(1, 110593, '\P{Is_Sc=kana}', "");
Expect(0, 110593, '\P{^Is_Sc=kana}', "");
Expect(1, 110592, '\p{Is_Sc=	_KANA}', "");
Expect(0, 110592, '\p{^Is_Sc=	_KANA}', "");
Expect(0, 110592, '\P{Is_Sc=	_KANA}', "");
Expect(1, 110592, '\P{^Is_Sc=	_KANA}', "");
Expect(0, 110593, '\p{Is_Sc=	_KANA}', "");
Expect(1, 110593, '\p{^Is_Sc=	_KANA}', "");
Expect(1, 110593, '\P{Is_Sc=	_KANA}', "");
Expect(0, 110593, '\P{^Is_Sc=	_KANA}', "");
Error('\p{Script= Kharoshthi:=}');
Error('\P{Script= Kharoshthi:=}');
Expect(1, 68184, '\p{Script=kharoshthi}', "");
Expect(0, 68184, '\p{^Script=kharoshthi}', "");
Expect(0, 68184, '\P{Script=kharoshthi}', "");
Expect(1, 68184, '\P{^Script=kharoshthi}', "");
Expect(0, 68185, '\p{Script=kharoshthi}', "");
Expect(1, 68185, '\p{^Script=kharoshthi}', "");
Expect(1, 68185, '\P{Script=kharoshthi}', "");
Expect(0, 68185, '\P{^Script=kharoshthi}', "");
Expect(1, 68184, '\p{Script: 	_Kharoshthi}', "");
Expect(0, 68184, '\p{^Script: 	_Kharoshthi}', "");
Expect(0, 68184, '\P{Script: 	_Kharoshthi}', "");
Expect(1, 68184, '\P{^Script: 	_Kharoshthi}', "");
Expect(0, 68185, '\p{Script: 	_Kharoshthi}', "");
Expect(1, 68185, '\p{^Script: 	_Kharoshthi}', "");
Expect(1, 68185, '\P{Script: 	_Kharoshthi}', "");
Expect(0, 68185, '\P{^Script: 	_Kharoshthi}', "");
Error('\p{Sc= _khar:=}');
Error('\P{Sc= _khar:=}');
Expect(1, 68184, '\p{Sc=khar}', "");
Expect(0, 68184, '\p{^Sc=khar}', "");
Expect(0, 68184, '\P{Sc=khar}', "");
Expect(1, 68184, '\P{^Sc=khar}', "");
Expect(0, 68185, '\p{Sc=khar}', "");
Expect(1, 68185, '\p{^Sc=khar}', "");
Expect(1, 68185, '\P{Sc=khar}', "");
Expect(0, 68185, '\P{^Sc=khar}', "");
Expect(1, 68184, '\p{Sc=__KHAR}', "");
Expect(0, 68184, '\p{^Sc=__KHAR}', "");
Expect(0, 68184, '\P{Sc=__KHAR}', "");
Expect(1, 68184, '\P{^Sc=__KHAR}', "");
Expect(0, 68185, '\p{Sc=__KHAR}', "");
Expect(1, 68185, '\p{^Sc=__KHAR}', "");
Expect(1, 68185, '\P{Sc=__KHAR}', "");
Expect(0, 68185, '\P{^Sc=__KHAR}', "");
Error('\p{Is_Script=/a/- kharoshthi}');
Error('\P{Is_Script=/a/- kharoshthi}');
Expect(1, 68184, '\p{Is_Script=kharoshthi}', "");
Expect(0, 68184, '\p{^Is_Script=kharoshthi}', "");
Expect(0, 68184, '\P{Is_Script=kharoshthi}', "");
Expect(1, 68184, '\P{^Is_Script=kharoshthi}', "");
Expect(0, 68185, '\p{Is_Script=kharoshthi}', "");
Expect(1, 68185, '\p{^Is_Script=kharoshthi}', "");
Expect(1, 68185, '\P{Is_Script=kharoshthi}', "");
Expect(0, 68185, '\P{^Is_Script=kharoshthi}', "");
Expect(1, 68184, '\p{Is_Script=--kharoshthi}', "");
Expect(0, 68184, '\p{^Is_Script=--kharoshthi}', "");
Expect(0, 68184, '\P{Is_Script=--kharoshthi}', "");
Expect(1, 68184, '\P{^Is_Script=--kharoshthi}', "");
Expect(0, 68185, '\p{Is_Script=--kharoshthi}', "");
Expect(1, 68185, '\p{^Is_Script=--kharoshthi}', "");
Expect(1, 68185, '\P{Is_Script=--kharoshthi}', "");
Expect(0, 68185, '\P{^Is_Script=--kharoshthi}', "");
Error('\p{Is_Sc=		khar:=}');
Error('\P{Is_Sc=		khar:=}');
Expect(1, 68184, '\p{Is_Sc=khar}', "");
Expect(0, 68184, '\p{^Is_Sc=khar}', "");
Expect(0, 68184, '\P{Is_Sc=khar}', "");
Expect(1, 68184, '\P{^Is_Sc=khar}', "");
Expect(0, 68185, '\p{Is_Sc=khar}', "");
Expect(1, 68185, '\p{^Is_Sc=khar}', "");
Expect(1, 68185, '\P{Is_Sc=khar}', "");
Expect(0, 68185, '\P{^Is_Sc=khar}', "");
Expect(1, 68184, '\p{Is_Sc=-	Khar}', "");
Expect(0, 68184, '\p{^Is_Sc=-	Khar}', "");
Expect(0, 68184, '\P{Is_Sc=-	Khar}', "");
Expect(1, 68184, '\P{^Is_Sc=-	Khar}', "");
Expect(0, 68185, '\p{Is_Sc=-	Khar}', "");
Expect(1, 68185, '\p{^Is_Sc=-	Khar}', "");
Expect(1, 68185, '\P{Is_Sc=-	Khar}', "");
Expect(0, 68185, '\P{^Is_Sc=-	Khar}', "");
Error('\p{Script=-:=KHMER}');
Error('\P{Script=-:=KHMER}');
Expect(1, 6655, '\p{Script=khmer}', "");
Expect(0, 6655, '\p{^Script=khmer}', "");
Expect(0, 6655, '\P{Script=khmer}', "");
Expect(1, 6655, '\P{^Script=khmer}', "");
Expect(0, 6656, '\p{Script=khmer}', "");
Expect(1, 6656, '\p{^Script=khmer}', "");
Expect(1, 6656, '\P{Script=khmer}', "");
Expect(0, 6656, '\P{^Script=khmer}', "");
Expect(1, 6655, '\p{Script:	 	Khmer}', "");
Expect(0, 6655, '\p{^Script:	 	Khmer}', "");
Expect(0, 6655, '\P{Script:	 	Khmer}', "");
Expect(1, 6655, '\P{^Script:	 	Khmer}', "");
Expect(0, 6656, '\p{Script:	 	Khmer}', "");
Expect(1, 6656, '\p{^Script:	 	Khmer}', "");
Expect(1, 6656, '\P{Script:	 	Khmer}', "");
Expect(0, 6656, '\P{^Script:	 	Khmer}', "");
Error('\p{Sc=:=-_Khmr}');
Error('\P{Sc=:=-_Khmr}');
Expect(1, 6655, '\p{Sc=khmr}', "");
Expect(0, 6655, '\p{^Sc=khmr}', "");
Expect(0, 6655, '\P{Sc=khmr}', "");
Expect(1, 6655, '\P{^Sc=khmr}', "");
Expect(0, 6656, '\p{Sc=khmr}', "");
Expect(1, 6656, '\p{^Sc=khmr}', "");
Expect(1, 6656, '\P{Sc=khmr}', "");
Expect(0, 6656, '\P{^Sc=khmr}', "");
Expect(1, 6655, '\p{Sc=-_khmr}', "");
Expect(0, 6655, '\p{^Sc=-_khmr}', "");
Expect(0, 6655, '\P{Sc=-_khmr}', "");
Expect(1, 6655, '\P{^Sc=-_khmr}', "");
Expect(0, 6656, '\p{Sc=-_khmr}', "");
Expect(1, 6656, '\p{^Sc=-_khmr}', "");
Expect(1, 6656, '\P{Sc=-_khmr}', "");
Expect(0, 6656, '\P{^Sc=-_khmr}', "");
Error('\p{Is_Script= Khmer:=}');
Error('\P{Is_Script= Khmer:=}');
Expect(1, 6655, '\p{Is_Script:khmer}', "");
Expect(0, 6655, '\p{^Is_Script:khmer}', "");
Expect(0, 6655, '\P{Is_Script:khmer}', "");
Expect(1, 6655, '\P{^Is_Script:khmer}', "");
Expect(0, 6656, '\p{Is_Script:khmer}', "");
Expect(1, 6656, '\p{^Is_Script:khmer}', "");
Expect(1, 6656, '\P{Is_Script:khmer}', "");
Expect(0, 6656, '\P{^Is_Script:khmer}', "");
Expect(1, 6655, '\p{Is_Script= _KHMER}', "");
Expect(0, 6655, '\p{^Is_Script= _KHMER}', "");
Expect(0, 6655, '\P{Is_Script= _KHMER}', "");
Expect(1, 6655, '\P{^Is_Script= _KHMER}', "");
Expect(0, 6656, '\p{Is_Script= _KHMER}', "");
Expect(1, 6656, '\p{^Is_Script= _KHMER}', "");
Expect(1, 6656, '\P{Is_Script= _KHMER}', "");
Expect(0, 6656, '\P{^Is_Script= _KHMER}', "");
Error('\p{Is_Sc:_/a/Khmr}');
Error('\P{Is_Sc:_/a/Khmr}');
Expect(1, 6655, '\p{Is_Sc=khmr}', "");
Expect(0, 6655, '\p{^Is_Sc=khmr}', "");
Expect(0, 6655, '\P{Is_Sc=khmr}', "");
Expect(1, 6655, '\P{^Is_Sc=khmr}', "");
Expect(0, 6656, '\p{Is_Sc=khmr}', "");
Expect(1, 6656, '\p{^Is_Sc=khmr}', "");
Expect(1, 6656, '\P{Is_Sc=khmr}', "");
Expect(0, 6656, '\P{^Is_Sc=khmr}', "");
Expect(1, 6655, '\p{Is_Sc=	_Khmr}', "");
Expect(0, 6655, '\p{^Is_Sc=	_Khmr}', "");
Expect(0, 6655, '\P{Is_Sc=	_Khmr}', "");
Expect(1, 6655, '\P{^Is_Sc=	_Khmr}', "");
Expect(0, 6656, '\p{Is_Sc=	_Khmr}', "");
Expect(1, 6656, '\p{^Is_Sc=	_Khmr}', "");
Expect(1, 6656, '\P{Is_Sc=	_Khmr}', "");
Expect(0, 6656, '\P{^Is_Sc=	_Khmr}', "");
Error('\p{Script= _kannada/a/}');
Error('\P{Script= _kannada/a/}');
Expect(1, 3314, '\p{Script=kannada}', "");
Expect(0, 3314, '\p{^Script=kannada}', "");
Expect(0, 3314, '\P{Script=kannada}', "");
Expect(1, 3314, '\P{^Script=kannada}', "");
Expect(0, 3315, '\p{Script=kannada}', "");
Expect(1, 3315, '\p{^Script=kannada}', "");
Expect(1, 3315, '\P{Script=kannada}', "");
Expect(0, 3315, '\P{^Script=kannada}', "");
Expect(1, 3314, '\p{Script=  Kannada}', "");
Expect(0, 3314, '\p{^Script=  Kannada}', "");
Expect(0, 3314, '\P{Script=  Kannada}', "");
Expect(1, 3314, '\P{^Script=  Kannada}', "");
Expect(0, 3315, '\p{Script=  Kannada}', "");
Expect(1, 3315, '\p{^Script=  Kannada}', "");
Expect(1, 3315, '\P{Script=  Kannada}', "");
Expect(0, 3315, '\P{^Script=  Kannada}', "");
Error('\p{Sc=-knda/a/}');
Error('\P{Sc=-knda/a/}');
Expect(1, 3314, '\p{Sc=knda}', "");
Expect(0, 3314, '\p{^Sc=knda}', "");
Expect(0, 3314, '\P{Sc=knda}', "");
Expect(1, 3314, '\P{^Sc=knda}', "");
Expect(0, 3315, '\p{Sc=knda}', "");
Expect(1, 3315, '\p{^Sc=knda}', "");
Expect(1, 3315, '\P{Sc=knda}', "");
Expect(0, 3315, '\P{^Sc=knda}', "");
Expect(1, 3314, '\p{Sc:-knda}', "");
Expect(0, 3314, '\p{^Sc:-knda}', "");
Expect(0, 3314, '\P{Sc:-knda}', "");
Expect(1, 3314, '\P{^Sc:-knda}', "");
Expect(0, 3315, '\p{Sc:-knda}', "");
Expect(1, 3315, '\p{^Sc:-knda}', "");
Expect(1, 3315, '\P{Sc:-knda}', "");
Expect(0, 3315, '\P{^Sc:-knda}', "");
Error('\p{Is_Script=_KANNADA:=}');
Error('\P{Is_Script=_KANNADA:=}');
Expect(1, 3314, '\p{Is_Script=kannada}', "");
Expect(0, 3314, '\p{^Is_Script=kannada}', "");
Expect(0, 3314, '\P{Is_Script=kannada}', "");
Expect(1, 3314, '\P{^Is_Script=kannada}', "");
Expect(0, 3315, '\p{Is_Script=kannada}', "");
Expect(1, 3315, '\p{^Is_Script=kannada}', "");
Expect(1, 3315, '\P{Is_Script=kannada}', "");
Expect(0, 3315, '\P{^Is_Script=kannada}', "");
Expect(1, 3314, '\p{Is_Script= Kannada}', "");
Expect(0, 3314, '\p{^Is_Script= Kannada}', "");
Expect(0, 3314, '\P{Is_Script= Kannada}', "");
Expect(1, 3314, '\P{^Is_Script= Kannada}', "");
Expect(0, 3315, '\p{Is_Script= Kannada}', "");
Expect(1, 3315, '\p{^Is_Script= Kannada}', "");
Expect(1, 3315, '\P{Is_Script= Kannada}', "");
Expect(0, 3315, '\P{^Is_Script= Kannada}', "");
Error('\p{Is_Sc=_:=knda}');
Error('\P{Is_Sc=_:=knda}');
Expect(1, 3314, '\p{Is_Sc=knda}', "");
Expect(0, 3314, '\p{^Is_Sc=knda}', "");
Expect(0, 3314, '\P{Is_Sc=knda}', "");
Expect(1, 3314, '\P{^Is_Sc=knda}', "");
Expect(0, 3315, '\p{Is_Sc=knda}', "");
Expect(1, 3315, '\p{^Is_Sc=knda}', "");
Expect(1, 3315, '\P{Is_Sc=knda}', "");
Expect(0, 3315, '\P{^Is_Sc=knda}', "");
Expect(1, 3314, '\p{Is_Sc=	KNDA}', "");
Expect(0, 3314, '\p{^Is_Sc=	KNDA}', "");
Expect(0, 3314, '\P{Is_Sc=	KNDA}', "");
Expect(1, 3314, '\P{^Is_Sc=	KNDA}', "");
Expect(0, 3315, '\p{Is_Sc=	KNDA}', "");
Expect(1, 3315, '\p{^Is_Sc=	KNDA}', "");
Expect(1, 3315, '\P{Is_Sc=	KNDA}', "");
Expect(0, 3315, '\P{^Is_Sc=	KNDA}', "");
Error('\p{Script=_/a/Kaithi}');
Error('\P{Script=_/a/Kaithi}');
Expect(1, 69825, '\p{Script=kaithi}', "");
Expect(0, 69825, '\p{^Script=kaithi}', "");
Expect(0, 69825, '\P{Script=kaithi}', "");
Expect(1, 69825, '\P{^Script=kaithi}', "");
Expect(0, 69826, '\p{Script=kaithi}', "");
Expect(1, 69826, '\p{^Script=kaithi}', "");
Expect(1, 69826, '\P{Script=kaithi}', "");
Expect(0, 69826, '\P{^Script=kaithi}', "");
Expect(1, 69825, '\p{Script=--Kaithi}', "");
Expect(0, 69825, '\p{^Script=--Kaithi}', "");
Expect(0, 69825, '\P{Script=--Kaithi}', "");
Expect(1, 69825, '\P{^Script=--Kaithi}', "");
Expect(0, 69826, '\p{Script=--Kaithi}', "");
Expect(1, 69826, '\p{^Script=--Kaithi}', "");
Expect(1, 69826, '\P{Script=--Kaithi}', "");
Expect(0, 69826, '\P{^Script=--Kaithi}', "");
Error('\p{Sc: _ Kthi:=}');
Error('\P{Sc: _ Kthi:=}');
Expect(1, 69825, '\p{Sc=kthi}', "");
Expect(0, 69825, '\p{^Sc=kthi}', "");
Expect(0, 69825, '\P{Sc=kthi}', "");
Expect(1, 69825, '\P{^Sc=kthi}', "");
Expect(0, 69826, '\p{Sc=kthi}', "");
Expect(1, 69826, '\p{^Sc=kthi}', "");
Expect(1, 69826, '\P{Sc=kthi}', "");
Expect(0, 69826, '\P{^Sc=kthi}', "");
Expect(1, 69825, '\p{Sc=-KTHI}', "");
Expect(0, 69825, '\p{^Sc=-KTHI}', "");
Expect(0, 69825, '\P{Sc=-KTHI}', "");
Expect(1, 69825, '\P{^Sc=-KTHI}', "");
Expect(0, 69826, '\p{Sc=-KTHI}', "");
Expect(1, 69826, '\p{^Sc=-KTHI}', "");
Expect(1, 69826, '\P{Sc=-KTHI}', "");
Expect(0, 69826, '\P{^Sc=-KTHI}', "");
Error('\p{Is_Script=_:=Kaithi}');
Error('\P{Is_Script=_:=Kaithi}');
Expect(1, 69825, '\p{Is_Script:	kaithi}', "");
Expect(0, 69825, '\p{^Is_Script:	kaithi}', "");
Expect(0, 69825, '\P{Is_Script:	kaithi}', "");
Expect(1, 69825, '\P{^Is_Script:	kaithi}', "");
Expect(0, 69826, '\p{Is_Script:	kaithi}', "");
Expect(1, 69826, '\p{^Is_Script:	kaithi}', "");
Expect(1, 69826, '\P{Is_Script:	kaithi}', "");
Expect(0, 69826, '\P{^Is_Script:	kaithi}', "");
Expect(1, 69825, '\p{Is_Script=_Kaithi}', "");
Expect(0, 69825, '\p{^Is_Script=_Kaithi}', "");
Expect(0, 69825, '\P{Is_Script=_Kaithi}', "");
Expect(1, 69825, '\P{^Is_Script=_Kaithi}', "");
Expect(0, 69826, '\p{Is_Script=_Kaithi}', "");
Expect(1, 69826, '\p{^Is_Script=_Kaithi}', "");
Expect(1, 69826, '\P{Is_Script=_Kaithi}', "");
Expect(0, 69826, '\P{^Is_Script=_Kaithi}', "");
Error('\p{Is_Sc= -kthi/a/}');
Error('\P{Is_Sc= -kthi/a/}');
Expect(1, 69825, '\p{Is_Sc=kthi}', "");
Expect(0, 69825, '\p{^Is_Sc=kthi}', "");
Expect(0, 69825, '\P{Is_Sc=kthi}', "");
Expect(1, 69825, '\P{^Is_Sc=kthi}', "");
Expect(0, 69826, '\p{Is_Sc=kthi}', "");
Expect(1, 69826, '\p{^Is_Sc=kthi}', "");
Expect(1, 69826, '\P{Is_Sc=kthi}', "");
Expect(0, 69826, '\P{^Is_Sc=kthi}', "");
Expect(1, 69825, '\p{Is_Sc=_kthi}', "");
Expect(0, 69825, '\p{^Is_Sc=_kthi}', "");
Expect(0, 69825, '\P{Is_Sc=_kthi}', "");
Expect(1, 69825, '\P{^Is_Sc=_kthi}', "");
Expect(0, 69826, '\p{Is_Sc=_kthi}', "");
Expect(1, 69826, '\p{^Is_Sc=_kthi}', "");
Expect(1, 69826, '\P{Is_Sc=_kthi}', "");
Expect(0, 69826, '\P{^Is_Sc=_kthi}', "");
Error('\p{Script=_TAI_THAM/a/}');
Error('\P{Script=_TAI_THAM/a/}');
Expect(1, 6829, '\p{Script=taitham}', "");
Expect(0, 6829, '\p{^Script=taitham}', "");
Expect(0, 6829, '\P{Script=taitham}', "");
Expect(1, 6829, '\P{^Script=taitham}', "");
Expect(0, 6830, '\p{Script=taitham}', "");
Expect(1, 6830, '\p{^Script=taitham}', "");
Expect(1, 6830, '\P{Script=taitham}', "");
Expect(0, 6830, '\P{^Script=taitham}', "");
Expect(1, 6829, '\p{Script=	Tai_tham}', "");
Expect(0, 6829, '\p{^Script=	Tai_tham}', "");
Expect(0, 6829, '\P{Script=	Tai_tham}', "");
Expect(1, 6829, '\P{^Script=	Tai_tham}', "");
Expect(0, 6830, '\p{Script=	Tai_tham}', "");
Expect(1, 6830, '\p{^Script=	Tai_tham}', "");
Expect(1, 6830, '\P{Script=	Tai_tham}', "");
Expect(0, 6830, '\P{^Script=	Tai_tham}', "");
Error('\p{Sc=	 Lana/a/}');
Error('\P{Sc=	 Lana/a/}');
Expect(1, 6829, '\p{Sc:	lana}', "");
Expect(0, 6829, '\p{^Sc:	lana}', "");
Expect(0, 6829, '\P{Sc:	lana}', "");
Expect(1, 6829, '\P{^Sc:	lana}', "");
Expect(0, 6830, '\p{Sc:	lana}', "");
Expect(1, 6830, '\p{^Sc:	lana}', "");
Expect(1, 6830, '\P{Sc:	lana}', "");
Expect(0, 6830, '\P{^Sc:	lana}', "");
Expect(1, 6829, '\p{Sc:-Lana}', "");
Expect(0, 6829, '\p{^Sc:-Lana}', "");
Expect(0, 6829, '\P{Sc:-Lana}', "");
Expect(1, 6829, '\P{^Sc:-Lana}', "");
Expect(0, 6830, '\p{Sc:-Lana}', "");
Expect(1, 6830, '\p{^Sc:-Lana}', "");
Expect(1, 6830, '\P{Sc:-Lana}', "");
Expect(0, 6830, '\P{^Sc:-Lana}', "");
Error('\p{Is_Script=:=-tai_Tham}');
Error('\P{Is_Script=:=-tai_Tham}');
Expect(1, 6829, '\p{Is_Script=taitham}', "");
Expect(0, 6829, '\p{^Is_Script=taitham}', "");
Expect(0, 6829, '\P{Is_Script=taitham}', "");
Expect(1, 6829, '\P{^Is_Script=taitham}', "");
Expect(0, 6830, '\p{Is_Script=taitham}', "");
Expect(1, 6830, '\p{^Is_Script=taitham}', "");
Expect(1, 6830, '\P{Is_Script=taitham}', "");
Expect(0, 6830, '\P{^Is_Script=taitham}', "");
Expect(1, 6829, '\p{Is_Script=	-TAI_Tham}', "");
Expect(0, 6829, '\p{^Is_Script=	-TAI_Tham}', "");
Expect(0, 6829, '\P{Is_Script=	-TAI_Tham}', "");
Expect(1, 6829, '\P{^Is_Script=	-TAI_Tham}', "");
Expect(0, 6830, '\p{Is_Script=	-TAI_Tham}', "");
Expect(1, 6830, '\p{^Is_Script=	-TAI_Tham}', "");
Expect(1, 6830, '\P{Is_Script=	-TAI_Tham}', "");
Expect(0, 6830, '\P{^Is_Script=	-TAI_Tham}', "");
Error('\p{Is_Sc=:=LANA}');
Error('\P{Is_Sc=:=LANA}');
Expect(1, 6829, '\p{Is_Sc=lana}', "");
Expect(0, 6829, '\p{^Is_Sc=lana}', "");
Expect(0, 6829, '\P{Is_Sc=lana}', "");
Expect(1, 6829, '\P{^Is_Sc=lana}', "");
Expect(0, 6830, '\p{Is_Sc=lana}', "");
Expect(1, 6830, '\p{^Is_Sc=lana}', "");
Expect(1, 6830, '\P{Is_Sc=lana}', "");
Expect(0, 6830, '\P{^Is_Sc=lana}', "");
Expect(1, 6829, '\p{Is_Sc=_	Lana}', "");
Expect(0, 6829, '\p{^Is_Sc=_	Lana}', "");
Expect(0, 6829, '\P{Is_Sc=_	Lana}', "");
Expect(1, 6829, '\P{^Is_Sc=_	Lana}', "");
Expect(0, 6830, '\p{Is_Sc=_	Lana}', "");
Expect(1, 6830, '\p{^Is_Sc=_	Lana}', "");
Expect(1, 6830, '\P{Is_Sc=_	Lana}', "");
Expect(0, 6830, '\P{^Is_Sc=_	Lana}', "");
Error('\p{Script=:=		Lao}');
Error('\P{Script=:=		Lao}');
Expect(1, 3807, '\p{Script=lao}', "");
Expect(0, 3807, '\p{^Script=lao}', "");
Expect(0, 3807, '\P{Script=lao}', "");
Expect(1, 3807, '\P{^Script=lao}', "");
Expect(0, 3808, '\p{Script=lao}', "");
Expect(1, 3808, '\p{^Script=lao}', "");
Expect(1, 3808, '\P{Script=lao}', "");
Expect(0, 3808, '\P{^Script=lao}', "");
Expect(1, 3807, '\p{Script= -lao}', "");
Expect(0, 3807, '\p{^Script= -lao}', "");
Expect(0, 3807, '\P{Script= -lao}', "");
Expect(1, 3807, '\P{^Script= -lao}', "");
Expect(0, 3808, '\p{Script= -lao}', "");
Expect(1, 3808, '\p{^Script= -lao}', "");
Expect(1, 3808, '\P{Script= -lao}', "");
Expect(0, 3808, '\P{^Script= -lao}', "");
Error('\p{Sc=-:=Laoo}');
Error('\P{Sc=-:=Laoo}');
Expect(1, 3807, '\p{Sc=laoo}', "");
Expect(0, 3807, '\p{^Sc=laoo}', "");
Expect(0, 3807, '\P{Sc=laoo}', "");
Expect(1, 3807, '\P{^Sc=laoo}', "");
Expect(0, 3808, '\p{Sc=laoo}', "");
Expect(1, 3808, '\p{^Sc=laoo}', "");
Expect(1, 3808, '\P{Sc=laoo}', "");
Expect(0, 3808, '\P{^Sc=laoo}', "");
Expect(1, 3807, '\p{Sc:-laoo}', "");
Expect(0, 3807, '\p{^Sc:-laoo}', "");
Expect(0, 3807, '\P{Sc:-laoo}', "");
Expect(1, 3807, '\P{^Sc:-laoo}', "");
Expect(0, 3808, '\p{Sc:-laoo}', "");
Expect(1, 3808, '\p{^Sc:-laoo}', "");
Expect(1, 3808, '\P{Sc:-laoo}', "");
Expect(0, 3808, '\P{^Sc:-laoo}', "");
Error('\p{Is_Script: :=LAO}');
Error('\P{Is_Script: :=LAO}');
Expect(1, 3807, '\p{Is_Script: lao}', "");
Expect(0, 3807, '\p{^Is_Script: lao}', "");
Expect(0, 3807, '\P{Is_Script: lao}', "");
Expect(1, 3807, '\P{^Is_Script: lao}', "");
Expect(0, 3808, '\p{Is_Script: lao}', "");
Expect(1, 3808, '\p{^Is_Script: lao}', "");
Expect(1, 3808, '\P{Is_Script: lao}', "");
Expect(0, 3808, '\P{^Is_Script: lao}', "");
Expect(1, 3807, '\p{Is_Script=	_Lao}', "");
Expect(0, 3807, '\p{^Is_Script=	_Lao}', "");
Expect(0, 3807, '\P{Is_Script=	_Lao}', "");
Expect(1, 3807, '\P{^Is_Script=	_Lao}', "");
Expect(0, 3808, '\p{Is_Script=	_Lao}', "");
Expect(1, 3808, '\p{^Is_Script=	_Lao}', "");
Expect(1, 3808, '\P{Is_Script=	_Lao}', "");
Expect(0, 3808, '\P{^Is_Script=	_Lao}', "");
Error('\p{Is_Sc=_/a/Laoo}');
Error('\P{Is_Sc=_/a/Laoo}');
Expect(1, 3807, '\p{Is_Sc:laoo}', "");
Expect(0, 3807, '\p{^Is_Sc:laoo}', "");
Expect(0, 3807, '\P{Is_Sc:laoo}', "");
Expect(1, 3807, '\P{^Is_Sc:laoo}', "");
Expect(0, 3808, '\p{Is_Sc:laoo}', "");
Expect(1, 3808, '\p{^Is_Sc:laoo}', "");
Expect(1, 3808, '\P{Is_Sc:laoo}', "");
Expect(0, 3808, '\P{^Is_Sc:laoo}', "");
Expect(1, 3807, '\p{Is_Sc=_LAOO}', "");
Expect(0, 3807, '\p{^Is_Sc=_LAOO}', "");
Expect(0, 3807, '\P{Is_Sc=_LAOO}', "");
Expect(1, 3807, '\P{^Is_Sc=_LAOO}', "");
Expect(0, 3808, '\p{Is_Sc=_LAOO}', "");
Expect(1, 3808, '\p{^Is_Sc=_LAOO}', "");
Expect(1, 3808, '\P{Is_Sc=_LAOO}', "");
Expect(0, 3808, '\P{^Is_Sc=_LAOO}', "");
Error('\p{Script=/a/	-Latin}');
Error('\P{Script=/a/	-Latin}');
Expect(1, 65370, '\p{Script=latin}', "");
Expect(0, 65370, '\p{^Script=latin}', "");
Expect(0, 65370, '\P{Script=latin}', "");
Expect(1, 65370, '\P{^Script=latin}', "");
Expect(0, 65371, '\p{Script=latin}', "");
Expect(1, 65371, '\p{^Script=latin}', "");
Expect(1, 65371, '\P{Script=latin}', "");
Expect(0, 65371, '\P{^Script=latin}', "");
Expect(1, 65370, '\p{Script:	-latin}', "");
Expect(0, 65370, '\p{^Script:	-latin}', "");
Expect(0, 65370, '\P{Script:	-latin}', "");
Expect(1, 65370, '\P{^Script:	-latin}', "");
Expect(0, 65371, '\p{Script:	-latin}', "");
Expect(1, 65371, '\p{^Script:	-latin}', "");
Expect(1, 65371, '\P{Script:	-latin}', "");
Expect(0, 65371, '\P{^Script:	-latin}', "");
Error('\p{Sc=/a/LATN}');
Error('\P{Sc=/a/LATN}');
Expect(1, 65370, '\p{Sc=latn}', "");
Expect(0, 65370, '\p{^Sc=latn}', "");
Expect(0, 65370, '\P{Sc=latn}', "");
Expect(1, 65370, '\P{^Sc=latn}', "");
Expect(0, 65371, '\p{Sc=latn}', "");
Expect(1, 65371, '\p{^Sc=latn}', "");
Expect(1, 65371, '\P{Sc=latn}', "");
Expect(0, 65371, '\P{^Sc=latn}', "");
Expect(1, 65370, '\p{Sc=_ LATN}', "");
Expect(0, 65370, '\p{^Sc=_ LATN}', "");
Expect(0, 65370, '\P{Sc=_ LATN}', "");
Expect(1, 65370, '\P{^Sc=_ LATN}', "");
Expect(0, 65371, '\p{Sc=_ LATN}', "");
Expect(1, 65371, '\p{^Sc=_ LATN}', "");
Expect(1, 65371, '\P{Sc=_ LATN}', "");
Expect(0, 65371, '\P{^Sc=_ LATN}', "");
Error('\p{Is_Script: _-latin:=}');
Error('\P{Is_Script: _-latin:=}');
Expect(1, 65370, '\p{Is_Script=latin}', "");
Expect(0, 65370, '\p{^Is_Script=latin}', "");
Expect(0, 65370, '\P{Is_Script=latin}', "");
Expect(1, 65370, '\P{^Is_Script=latin}', "");
Expect(0, 65371, '\p{Is_Script=latin}', "");
Expect(1, 65371, '\p{^Is_Script=latin}', "");
Expect(1, 65371, '\P{Is_Script=latin}', "");
Expect(0, 65371, '\P{^Is_Script=latin}', "");
Expect(1, 65370, '\p{Is_Script=_latin}', "");
Expect(0, 65370, '\p{^Is_Script=_latin}', "");
Expect(0, 65370, '\P{Is_Script=_latin}', "");
Expect(1, 65370, '\P{^Is_Script=_latin}', "");
Expect(0, 65371, '\p{Is_Script=_latin}', "");
Expect(1, 65371, '\p{^Is_Script=_latin}', "");
Expect(1, 65371, '\P{Is_Script=_latin}', "");
Expect(0, 65371, '\P{^Is_Script=_latin}', "");
Error('\p{Is_Sc=/a/	-Latn}');
Error('\P{Is_Sc=/a/	-Latn}');
Expect(1, 65370, '\p{Is_Sc=latn}', "");
Expect(0, 65370, '\p{^Is_Sc=latn}', "");
Expect(0, 65370, '\P{Is_Sc=latn}', "");
Expect(1, 65370, '\P{^Is_Sc=latn}', "");
Expect(0, 65371, '\p{Is_Sc=latn}', "");
Expect(1, 65371, '\p{^Is_Sc=latn}', "");
Expect(1, 65371, '\P{Is_Sc=latn}', "");
Expect(0, 65371, '\P{^Is_Sc=latn}', "");
Expect(1, 65370, '\p{Is_Sc=-	Latn}', "");
Expect(0, 65370, '\p{^Is_Sc=-	Latn}', "");
Expect(0, 65370, '\P{Is_Sc=-	Latn}', "");
Expect(1, 65370, '\P{^Is_Sc=-	Latn}', "");
Expect(0, 65371, '\p{Is_Sc=-	Latn}', "");
Expect(1, 65371, '\p{^Is_Sc=-	Latn}', "");
Expect(1, 65371, '\P{Is_Sc=-	Latn}', "");
Expect(0, 65371, '\P{^Is_Sc=-	Latn}', "");
Error('\p{Script= 	lepcha/a/}');
Error('\P{Script= 	lepcha/a/}');
Expect(1, 7247, '\p{Script: lepcha}', "");
Expect(0, 7247, '\p{^Script: lepcha}', "");
Expect(0, 7247, '\P{Script: lepcha}', "");
Expect(1, 7247, '\P{^Script: lepcha}', "");
Expect(0, 7248, '\p{Script: lepcha}', "");
Expect(1, 7248, '\p{^Script: lepcha}', "");
Expect(1, 7248, '\P{Script: lepcha}', "");
Expect(0, 7248, '\P{^Script: lepcha}', "");
Expect(1, 7247, '\p{Script=	-Lepcha}', "");
Expect(0, 7247, '\p{^Script=	-Lepcha}', "");
Expect(0, 7247, '\P{Script=	-Lepcha}', "");
Expect(1, 7247, '\P{^Script=	-Lepcha}', "");
Expect(0, 7248, '\p{Script=	-Lepcha}', "");
Expect(1, 7248, '\p{^Script=	-Lepcha}', "");
Expect(1, 7248, '\P{Script=	-Lepcha}', "");
Expect(0, 7248, '\P{^Script=	-Lepcha}', "");
Error('\p{Sc=	_LEPC/a/}');
Error('\P{Sc=	_LEPC/a/}');
Expect(1, 7247, '\p{Sc=lepc}', "");
Expect(0, 7247, '\p{^Sc=lepc}', "");
Expect(0, 7247, '\P{Sc=lepc}', "");
Expect(1, 7247, '\P{^Sc=lepc}', "");
Expect(0, 7248, '\p{Sc=lepc}', "");
Expect(1, 7248, '\p{^Sc=lepc}', "");
Expect(1, 7248, '\P{Sc=lepc}', "");
Expect(0, 7248, '\P{^Sc=lepc}', "");
Expect(1, 7247, '\p{Sc=--Lepc}', "");
Expect(0, 7247, '\p{^Sc=--Lepc}', "");
Expect(0, 7247, '\P{Sc=--Lepc}', "");
Expect(1, 7247, '\P{^Sc=--Lepc}', "");
Expect(0, 7248, '\p{Sc=--Lepc}', "");
Expect(1, 7248, '\p{^Sc=--Lepc}', "");
Expect(1, 7248, '\P{Sc=--Lepc}', "");
Expect(0, 7248, '\P{^Sc=--Lepc}', "");
Error('\p{Is_Script=-_lepcha:=}');
Error('\P{Is_Script=-_lepcha:=}');
Expect(1, 7247, '\p{Is_Script=lepcha}', "");
Expect(0, 7247, '\p{^Is_Script=lepcha}', "");
Expect(0, 7247, '\P{Is_Script=lepcha}', "");
Expect(1, 7247, '\P{^Is_Script=lepcha}', "");
Expect(0, 7248, '\p{Is_Script=lepcha}', "");
Expect(1, 7248, '\p{^Is_Script=lepcha}', "");
Expect(1, 7248, '\P{Is_Script=lepcha}', "");
Expect(0, 7248, '\P{^Is_Script=lepcha}', "");
Expect(1, 7247, '\p{Is_Script:   -_LEPCHA}', "");
Expect(0, 7247, '\p{^Is_Script:   -_LEPCHA}', "");
Expect(0, 7247, '\P{Is_Script:   -_LEPCHA}', "");
Expect(1, 7247, '\P{^Is_Script:   -_LEPCHA}', "");
Expect(0, 7248, '\p{Is_Script:   -_LEPCHA}', "");
Expect(1, 7248, '\p{^Is_Script:   -_LEPCHA}', "");
Expect(1, 7248, '\P{Is_Script:   -_LEPCHA}', "");
Expect(0, 7248, '\P{^Is_Script:   -_LEPCHA}', "");
Error('\p{Is_Sc=-/a/LEPC}');
Error('\P{Is_Sc=-/a/LEPC}');
Expect(1, 7247, '\p{Is_Sc=lepc}', "");
Expect(0, 7247, '\p{^Is_Sc=lepc}', "");
Expect(0, 7247, '\P{Is_Sc=lepc}', "");
Expect(1, 7247, '\P{^Is_Sc=lepc}', "");
Expect(0, 7248, '\p{Is_Sc=lepc}', "");
Expect(1, 7248, '\p{^Is_Sc=lepc}', "");
Expect(1, 7248, '\P{Is_Sc=lepc}', "");
Expect(0, 7248, '\P{^Is_Sc=lepc}', "");
Expect(1, 7247, '\p{Is_Sc= Lepc}', "");
Expect(0, 7247, '\p{^Is_Sc= Lepc}', "");
Expect(0, 7247, '\P{Is_Sc= Lepc}', "");
Expect(1, 7247, '\P{^Is_Sc= Lepc}', "");
Expect(0, 7248, '\p{Is_Sc= Lepc}', "");
Expect(1, 7248, '\p{^Is_Sc= Lepc}', "");
Expect(1, 7248, '\P{Is_Sc= Lepc}', "");
Expect(0, 7248, '\P{^Is_Sc= Lepc}', "");
Error('\p{Script=_:=LIMBU}');
Error('\P{Script=_:=LIMBU}');
Expect(1, 6479, '\p{Script=limbu}', "");
Expect(0, 6479, '\p{^Script=limbu}', "");
Expect(0, 6479, '\P{Script=limbu}', "");
Expect(1, 6479, '\P{^Script=limbu}', "");
Expect(0, 6480, '\p{Script=limbu}', "");
Expect(1, 6480, '\p{^Script=limbu}', "");
Expect(1, 6480, '\P{Script=limbu}', "");
Expect(0, 6480, '\P{^Script=limbu}', "");
Expect(1, 6479, '\p{Script=	-Limbu}', "");
Expect(0, 6479, '\p{^Script=	-Limbu}', "");
Expect(0, 6479, '\P{Script=	-Limbu}', "");
Expect(1, 6479, '\P{^Script=	-Limbu}', "");
Expect(0, 6480, '\p{Script=	-Limbu}', "");
Expect(1, 6480, '\p{^Script=	-Limbu}', "");
Expect(1, 6480, '\P{Script=	-Limbu}', "");
Expect(0, 6480, '\P{^Script=	-Limbu}', "");
Error('\p{Sc=/a/ 	Limb}');
Error('\P{Sc=/a/ 	Limb}');
Expect(1, 6479, '\p{Sc=limb}', "");
Expect(0, 6479, '\p{^Sc=limb}', "");
Expect(0, 6479, '\P{Sc=limb}', "");
Expect(1, 6479, '\P{^Sc=limb}', "");
Expect(0, 6480, '\p{Sc=limb}', "");
Expect(1, 6480, '\p{^Sc=limb}', "");
Expect(1, 6480, '\P{Sc=limb}', "");
Expect(0, 6480, '\P{^Sc=limb}', "");
Expect(1, 6479, '\p{Sc= 	Limb}', "");
Expect(0, 6479, '\p{^Sc= 	Limb}', "");
Expect(0, 6479, '\P{Sc= 	Limb}', "");
Expect(1, 6479, '\P{^Sc= 	Limb}', "");
Expect(0, 6480, '\p{Sc= 	Limb}', "");
Expect(1, 6480, '\p{^Sc= 	Limb}', "");
Expect(1, 6480, '\P{Sc= 	Limb}', "");
Expect(0, 6480, '\P{^Sc= 	Limb}', "");
Error('\p{Is_Script:/a/	_Limbu}');
Error('\P{Is_Script:/a/	_Limbu}');
Expect(1, 6479, '\p{Is_Script=limbu}', "");
Expect(0, 6479, '\p{^Is_Script=limbu}', "");
Expect(0, 6479, '\P{Is_Script=limbu}', "");
Expect(1, 6479, '\P{^Is_Script=limbu}', "");
Expect(0, 6480, '\p{Is_Script=limbu}', "");
Expect(1, 6480, '\p{^Is_Script=limbu}', "");
Expect(1, 6480, '\P{Is_Script=limbu}', "");
Expect(0, 6480, '\P{^Is_Script=limbu}', "");
Expect(1, 6479, '\p{Is_Script= 	limbu}', "");
Expect(0, 6479, '\p{^Is_Script= 	limbu}', "");
Expect(0, 6479, '\P{Is_Script= 	limbu}', "");
Expect(1, 6479, '\P{^Is_Script= 	limbu}', "");
Expect(0, 6480, '\p{Is_Script= 	limbu}', "");
Expect(1, 6480, '\p{^Is_Script= 	limbu}', "");
Expect(1, 6480, '\P{Is_Script= 	limbu}', "");
Expect(0, 6480, '\P{^Is_Script= 	limbu}', "");
Error('\p{Is_Sc=	_Limb:=}');
Error('\P{Is_Sc=	_Limb:=}');
Expect(1, 6479, '\p{Is_Sc=limb}', "");
Expect(0, 6479, '\p{^Is_Sc=limb}', "");
Expect(0, 6479, '\P{Is_Sc=limb}', "");
Expect(1, 6479, '\P{^Is_Sc=limb}', "");
Expect(0, 6480, '\p{Is_Sc=limb}', "");
Expect(1, 6480, '\p{^Is_Sc=limb}', "");
Expect(1, 6480, '\P{Is_Sc=limb}', "");
Expect(0, 6480, '\P{^Is_Sc=limb}', "");
Expect(1, 6479, '\p{Is_Sc=_	Limb}', "");
Expect(0, 6479, '\p{^Is_Sc=_	Limb}', "");
Expect(0, 6479, '\P{Is_Sc=_	Limb}', "");
Expect(1, 6479, '\P{^Is_Sc=_	Limb}', "");
Expect(0, 6480, '\p{Is_Sc=_	Limb}', "");
Expect(1, 6480, '\p{^Is_Sc=_	Limb}', "");
Expect(1, 6480, '\P{Is_Sc=_	Limb}', "");
Expect(0, 6480, '\P{^Is_Sc=_	Limb}', "");
Error('\p{Script=/a/ _Linear_B}');
Error('\P{Script=/a/ _Linear_B}');
Expect(1, 65786, '\p{Script=linearb}', "");
Expect(0, 65786, '\p{^Script=linearb}', "");
Expect(0, 65786, '\P{Script=linearb}', "");
Expect(1, 65786, '\P{^Script=linearb}', "");
Expect(0, 65787, '\p{Script=linearb}', "");
Expect(1, 65787, '\p{^Script=linearb}', "");
Expect(1, 65787, '\P{Script=linearb}', "");
Expect(0, 65787, '\P{^Script=linearb}', "");
Expect(1, 65786, '\p{Script=	_Linear_B}', "");
Expect(0, 65786, '\p{^Script=	_Linear_B}', "");
Expect(0, 65786, '\P{Script=	_Linear_B}', "");
Expect(1, 65786, '\P{^Script=	_Linear_B}', "");
Expect(0, 65787, '\p{Script=	_Linear_B}', "");
Expect(1, 65787, '\p{^Script=	_Linear_B}', "");
Expect(1, 65787, '\P{Script=	_Linear_B}', "");
Expect(0, 65787, '\P{^Script=	_Linear_B}', "");
Error('\p{Sc=	LINB:=}');
Error('\P{Sc=	LINB:=}');
Expect(1, 65786, '\p{Sc=linb}', "");
Expect(0, 65786, '\p{^Sc=linb}', "");
Expect(0, 65786, '\P{Sc=linb}', "");
Expect(1, 65786, '\P{^Sc=linb}', "");
Expect(0, 65787, '\p{Sc=linb}', "");
Expect(1, 65787, '\p{^Sc=linb}', "");
Expect(1, 65787, '\P{Sc=linb}', "");
Expect(0, 65787, '\P{^Sc=linb}', "");
Expect(1, 65786, '\p{Sc=-	linb}', "");
Expect(0, 65786, '\p{^Sc=-	linb}', "");
Expect(0, 65786, '\P{Sc=-	linb}', "");
Expect(1, 65786, '\P{^Sc=-	linb}', "");
Expect(0, 65787, '\p{Sc=-	linb}', "");
Expect(1, 65787, '\p{^Sc=-	linb}', "");
Expect(1, 65787, '\P{Sc=-	linb}', "");
Expect(0, 65787, '\P{^Sc=-	linb}', "");
Error('\p{Is_Script=-/a/Linear_B}');
Error('\P{Is_Script=-/a/Linear_B}');
Expect(1, 65786, '\p{Is_Script=linearb}', "");
Expect(0, 65786, '\p{^Is_Script=linearb}', "");
Expect(0, 65786, '\P{Is_Script=linearb}', "");
Expect(1, 65786, '\P{^Is_Script=linearb}', "");
Expect(0, 65787, '\p{Is_Script=linearb}', "");
Expect(1, 65787, '\p{^Is_Script=linearb}', "");
Expect(1, 65787, '\P{Is_Script=linearb}', "");
Expect(0, 65787, '\P{^Is_Script=linearb}', "");
Expect(1, 65786, '\p{Is_Script=-linear_b}', "");
Expect(0, 65786, '\p{^Is_Script=-linear_b}', "");
Expect(0, 65786, '\P{Is_Script=-linear_b}', "");
Expect(1, 65786, '\P{^Is_Script=-linear_b}', "");
Expect(0, 65787, '\p{Is_Script=-linear_b}', "");
Expect(1, 65787, '\p{^Is_Script=-linear_b}', "");
Expect(1, 65787, '\P{Is_Script=-linear_b}', "");
Expect(0, 65787, '\P{^Is_Script=-linear_b}', "");
Error('\p{Is_Sc=	:=LINB}');
Error('\P{Is_Sc=	:=LINB}');
Expect(1, 65786, '\p{Is_Sc=linb}', "");
Expect(0, 65786, '\p{^Is_Sc=linb}', "");
Expect(0, 65786, '\P{Is_Sc=linb}', "");
Expect(1, 65786, '\P{^Is_Sc=linb}', "");
Expect(0, 65787, '\p{Is_Sc=linb}', "");
Expect(1, 65787, '\p{^Is_Sc=linb}', "");
Expect(1, 65787, '\P{Is_Sc=linb}', "");
Expect(0, 65787, '\P{^Is_Sc=linb}', "");
Expect(1, 65786, '\p{Is_Sc= _LINB}', "");
Expect(0, 65786, '\p{^Is_Sc= _LINB}', "");
Expect(0, 65786, '\P{Is_Sc= _LINB}', "");
Expect(1, 65786, '\P{^Is_Sc= _LINB}', "");
Expect(0, 65787, '\p{Is_Sc= _LINB}', "");
Expect(1, 65787, '\p{^Is_Sc= _LINB}', "");
Expect(1, 65787, '\P{Is_Sc= _LINB}', "");
Expect(0, 65787, '\P{^Is_Sc= _LINB}', "");
Error('\p{Script=-	LISU:=}');
Error('\P{Script=-	LISU:=}');
Expect(1, 42239, '\p{Script=lisu}', "");
Expect(0, 42239, '\p{^Script=lisu}', "");
Expect(0, 42239, '\P{Script=lisu}', "");
Expect(1, 42239, '\P{^Script=lisu}', "");
Expect(0, 42240, '\p{Script=lisu}', "");
Expect(1, 42240, '\p{^Script=lisu}', "");
Expect(1, 42240, '\P{Script=lisu}', "");
Expect(0, 42240, '\P{^Script=lisu}', "");
Expect(1, 42239, '\p{Script=-lisu}', "");
Expect(0, 42239, '\p{^Script=-lisu}', "");
Expect(0, 42239, '\P{Script=-lisu}', "");
Expect(1, 42239, '\P{^Script=-lisu}', "");
Expect(0, 42240, '\p{Script=-lisu}', "");
Expect(1, 42240, '\p{^Script=-lisu}', "");
Expect(1, 42240, '\P{Script=-lisu}', "");
Expect(0, 42240, '\P{^Script=-lisu}', "");
Error('\p{Sc=:=_lisu}');
Error('\P{Sc=:=_lisu}');
Expect(1, 42239, '\p{Sc=lisu}', "");
Expect(0, 42239, '\p{^Sc=lisu}', "");
Expect(0, 42239, '\P{Sc=lisu}', "");
Expect(1, 42239, '\P{^Sc=lisu}', "");
Expect(0, 42240, '\p{Sc=lisu}', "");
Expect(1, 42240, '\p{^Sc=lisu}', "");
Expect(1, 42240, '\P{Sc=lisu}', "");
Expect(0, 42240, '\P{^Sc=lisu}', "");
Expect(1, 42239, '\p{Sc=_	Lisu}', "");
Expect(0, 42239, '\p{^Sc=_	Lisu}', "");
Expect(0, 42239, '\P{Sc=_	Lisu}', "");
Expect(1, 42239, '\P{^Sc=_	Lisu}', "");
Expect(0, 42240, '\p{Sc=_	Lisu}', "");
Expect(1, 42240, '\p{^Sc=_	Lisu}', "");
Expect(1, 42240, '\P{Sc=_	Lisu}', "");
Expect(0, 42240, '\P{^Sc=_	Lisu}', "");
Error('\p{Is_Script= :=Lisu}');
Error('\P{Is_Script= :=Lisu}');
Expect(1, 42239, '\p{Is_Script:   lisu}', "");
Expect(0, 42239, '\p{^Is_Script:   lisu}', "");
Expect(0, 42239, '\P{Is_Script:   lisu}', "");
Expect(1, 42239, '\P{^Is_Script:   lisu}', "");
Expect(0, 42240, '\p{Is_Script:   lisu}', "");
Expect(1, 42240, '\p{^Is_Script:   lisu}', "");
Expect(1, 42240, '\P{Is_Script:   lisu}', "");
Expect(0, 42240, '\P{^Is_Script:   lisu}', "");
Expect(1, 42239, '\p{Is_Script: _-LISU}', "");
Expect(0, 42239, '\p{^Is_Script: _-LISU}', "");
Expect(0, 42239, '\P{Is_Script: _-LISU}', "");
Expect(1, 42239, '\P{^Is_Script: _-LISU}', "");
Expect(0, 42240, '\p{Is_Script: _-LISU}', "");
Expect(1, 42240, '\p{^Is_Script: _-LISU}', "");
Expect(1, 42240, '\P{Is_Script: _-LISU}', "");
Expect(0, 42240, '\P{^Is_Script: _-LISU}', "");
Error('\p{Is_Sc:   		LISU/a/}');
Error('\P{Is_Sc:   		LISU/a/}');
Expect(1, 42239, '\p{Is_Sc=lisu}', "");
Expect(0, 42239, '\p{^Is_Sc=lisu}', "");
Expect(0, 42239, '\P{Is_Sc=lisu}', "");
Expect(1, 42239, '\P{^Is_Sc=lisu}', "");
Expect(0, 42240, '\p{Is_Sc=lisu}', "");
Expect(1, 42240, '\p{^Is_Sc=lisu}', "");
Expect(1, 42240, '\P{Is_Sc=lisu}', "");
Expect(0, 42240, '\P{^Is_Sc=lisu}', "");
Expect(1, 42239, '\p{Is_Sc=_	Lisu}', "");
Expect(0, 42239, '\p{^Is_Sc=_	Lisu}', "");
Expect(0, 42239, '\P{Is_Sc=_	Lisu}', "");
Expect(1, 42239, '\P{^Is_Sc=_	Lisu}', "");
Expect(0, 42240, '\p{Is_Sc=_	Lisu}', "");
Expect(1, 42240, '\p{^Is_Sc=_	Lisu}', "");
Expect(1, 42240, '\P{Is_Sc=_	Lisu}', "");
Expect(0, 42240, '\P{^Is_Sc=_	Lisu}', "");
Error('\p{Script= Lycian:=}');
Error('\P{Script= Lycian:=}');
Expect(1, 66204, '\p{Script=lycian}', "");
Expect(0, 66204, '\p{^Script=lycian}', "");
Expect(0, 66204, '\P{Script=lycian}', "");
Expect(1, 66204, '\P{^Script=lycian}', "");
Expect(0, 66205, '\p{Script=lycian}', "");
Expect(1, 66205, '\p{^Script=lycian}', "");
Expect(1, 66205, '\P{Script=lycian}', "");
Expect(0, 66205, '\P{^Script=lycian}', "");
Expect(1, 66204, '\p{Script:	-lycian}', "");
Expect(0, 66204, '\p{^Script:	-lycian}', "");
Expect(0, 66204, '\P{Script:	-lycian}', "");
Expect(1, 66204, '\P{^Script:	-lycian}', "");
Expect(0, 66205, '\p{Script:	-lycian}', "");
Expect(1, 66205, '\p{^Script:	-lycian}', "");
Expect(1, 66205, '\P{Script:	-lycian}', "");
Expect(0, 66205, '\P{^Script:	-lycian}', "");
Error('\p{Sc=-:=LYCI}');
Error('\P{Sc=-:=LYCI}');
Expect(1, 66204, '\p{Sc=lyci}', "");
Expect(0, 66204, '\p{^Sc=lyci}', "");
Expect(0, 66204, '\P{Sc=lyci}', "");
Expect(1, 66204, '\P{^Sc=lyci}', "");
Expect(0, 66205, '\p{Sc=lyci}', "");
Expect(1, 66205, '\p{^Sc=lyci}', "");
Expect(1, 66205, '\P{Sc=lyci}', "");
Expect(0, 66205, '\P{^Sc=lyci}', "");
Expect(1, 66204, '\p{Sc=  LYCI}', "");
Expect(0, 66204, '\p{^Sc=  LYCI}', "");
Expect(0, 66204, '\P{Sc=  LYCI}', "");
Expect(1, 66204, '\P{^Sc=  LYCI}', "");
Expect(0, 66205, '\p{Sc=  LYCI}', "");
Expect(1, 66205, '\p{^Sc=  LYCI}', "");
Expect(1, 66205, '\P{Sc=  LYCI}', "");
Expect(0, 66205, '\P{^Sc=  LYCI}', "");
Error('\p{Is_Script: --Lycian/a/}');
Error('\P{Is_Script: --Lycian/a/}');
Expect(1, 66204, '\p{Is_Script:   lycian}', "");
Expect(0, 66204, '\p{^Is_Script:   lycian}', "");
Expect(0, 66204, '\P{Is_Script:   lycian}', "");
Expect(1, 66204, '\P{^Is_Script:   lycian}', "");
Expect(0, 66205, '\p{Is_Script:   lycian}', "");
Expect(1, 66205, '\p{^Is_Script:   lycian}', "");
Expect(1, 66205, '\P{Is_Script:   lycian}', "");
Expect(0, 66205, '\P{^Is_Script:   lycian}', "");
Expect(1, 66204, '\p{Is_Script=_ Lycian}', "");
Expect(0, 66204, '\p{^Is_Script=_ Lycian}', "");
Expect(0, 66204, '\P{Is_Script=_ Lycian}', "");
Expect(1, 66204, '\P{^Is_Script=_ Lycian}', "");
Expect(0, 66205, '\p{Is_Script=_ Lycian}', "");
Expect(1, 66205, '\p{^Is_Script=_ Lycian}', "");
Expect(1, 66205, '\P{Is_Script=_ Lycian}', "");
Expect(0, 66205, '\P{^Is_Script=_ Lycian}', "");
Error('\p{Is_Sc=_/a/lyci}');
Error('\P{Is_Sc=_/a/lyci}');
Expect(1, 66204, '\p{Is_Sc=lyci}', "");
Expect(0, 66204, '\p{^Is_Sc=lyci}', "");
Expect(0, 66204, '\P{Is_Sc=lyci}', "");
Expect(1, 66204, '\P{^Is_Sc=lyci}', "");
Expect(0, 66205, '\p{Is_Sc=lyci}', "");
Expect(1, 66205, '\p{^Is_Sc=lyci}', "");
Expect(1, 66205, '\P{Is_Sc=lyci}', "");
Expect(0, 66205, '\P{^Is_Sc=lyci}', "");
Expect(1, 66204, '\p{Is_Sc: _Lyci}', "");
Expect(0, 66204, '\p{^Is_Sc: _Lyci}', "");
Expect(0, 66204, '\P{Is_Sc: _Lyci}', "");
Expect(1, 66204, '\P{^Is_Sc: _Lyci}', "");
Expect(0, 66205, '\p{Is_Sc: _Lyci}', "");
Expect(1, 66205, '\p{^Is_Sc: _Lyci}', "");
Expect(1, 66205, '\P{Is_Sc: _Lyci}', "");
Expect(0, 66205, '\P{^Is_Sc: _Lyci}', "");
Error('\p{Script: /a/  lydian}');
Error('\P{Script: /a/  lydian}');
Expect(1, 67903, '\p{Script=lydian}', "");
Expect(0, 67903, '\p{^Script=lydian}', "");
Expect(0, 67903, '\P{Script=lydian}', "");
Expect(1, 67903, '\P{^Script=lydian}', "");
Expect(0, 67904, '\p{Script=lydian}', "");
Expect(1, 67904, '\p{^Script=lydian}', "");
Expect(1, 67904, '\P{Script=lydian}', "");
Expect(0, 67904, '\P{^Script=lydian}', "");
Expect(1, 67903, '\p{Script=	lydian}', "");
Expect(0, 67903, '\p{^Script=	lydian}', "");
Expect(0, 67903, '\P{Script=	lydian}', "");
Expect(1, 67903, '\P{^Script=	lydian}', "");
Expect(0, 67904, '\p{Script=	lydian}', "");
Expect(1, 67904, '\p{^Script=	lydian}', "");
Expect(1, 67904, '\P{Script=	lydian}', "");
Expect(0, 67904, '\P{^Script=	lydian}', "");
Error('\p{Sc= /a/lydi}');
Error('\P{Sc= /a/lydi}');
Expect(1, 67903, '\p{Sc=lydi}', "");
Expect(0, 67903, '\p{^Sc=lydi}', "");
Expect(0, 67903, '\P{Sc=lydi}', "");
Expect(1, 67903, '\P{^Sc=lydi}', "");
Expect(0, 67904, '\p{Sc=lydi}', "");
Expect(1, 67904, '\p{^Sc=lydi}', "");
Expect(1, 67904, '\P{Sc=lydi}', "");
Expect(0, 67904, '\P{^Sc=lydi}', "");
Expect(1, 67903, '\p{Sc=  Lydi}', "");
Expect(0, 67903, '\p{^Sc=  Lydi}', "");
Expect(0, 67903, '\P{Sc=  Lydi}', "");
Expect(1, 67903, '\P{^Sc=  Lydi}', "");
Expect(0, 67904, '\p{Sc=  Lydi}', "");
Expect(1, 67904, '\p{^Sc=  Lydi}', "");
Expect(1, 67904, '\P{Sc=  Lydi}', "");
Expect(0, 67904, '\P{^Sc=  Lydi}', "");
Error('\p{Is_Script=lydian:=}');
Error('\P{Is_Script=lydian:=}');
Expect(1, 67903, '\p{Is_Script=lydian}', "");
Expect(0, 67903, '\p{^Is_Script=lydian}', "");
Expect(0, 67903, '\P{Is_Script=lydian}', "");
Expect(1, 67903, '\P{^Is_Script=lydian}', "");
Expect(0, 67904, '\p{Is_Script=lydian}', "");
Expect(1, 67904, '\p{^Is_Script=lydian}', "");
Expect(1, 67904, '\P{Is_Script=lydian}', "");
Expect(0, 67904, '\P{^Is_Script=lydian}', "");
Expect(1, 67903, '\p{Is_Script= -Lydian}', "");
Expect(0, 67903, '\p{^Is_Script= -Lydian}', "");
Expect(0, 67903, '\P{Is_Script= -Lydian}', "");
Expect(1, 67903, '\P{^Is_Script= -Lydian}', "");
Expect(0, 67904, '\p{Is_Script= -Lydian}', "");
Expect(1, 67904, '\p{^Is_Script= -Lydian}', "");
Expect(1, 67904, '\P{Is_Script= -Lydian}', "");
Expect(0, 67904, '\P{^Is_Script= -Lydian}', "");
Error('\p{Is_Sc: _:=LYDI}');
Error('\P{Is_Sc: _:=LYDI}');
Expect(1, 67903, '\p{Is_Sc=lydi}', "");
Expect(0, 67903, '\p{^Is_Sc=lydi}', "");
Expect(0, 67903, '\P{Is_Sc=lydi}', "");
Expect(1, 67903, '\P{^Is_Sc=lydi}', "");
Expect(0, 67904, '\p{Is_Sc=lydi}', "");
Expect(1, 67904, '\p{^Is_Sc=lydi}', "");
Expect(1, 67904, '\P{Is_Sc=lydi}', "");
Expect(0, 67904, '\P{^Is_Sc=lydi}', "");
Expect(1, 67903, '\p{Is_Sc=_	Lydi}', "");
Expect(0, 67903, '\p{^Is_Sc=_	Lydi}', "");
Expect(0, 67903, '\P{Is_Sc=_	Lydi}', "");
Expect(1, 67903, '\P{^Is_Sc=_	Lydi}', "");
Expect(0, 67904, '\p{Is_Sc=_	Lydi}', "");
Expect(1, 67904, '\p{^Is_Sc=_	Lydi}', "");
Expect(1, 67904, '\P{Is_Sc=_	Lydi}', "");
Expect(0, 67904, '\P{^Is_Sc=_	Lydi}', "");
Error('\p{Script=:=	-Mandaic}');
Error('\P{Script=:=	-Mandaic}');
Expect(1, 2142, '\p{Script=mandaic}', "");
Expect(0, 2142, '\p{^Script=mandaic}', "");
Expect(0, 2142, '\P{Script=mandaic}', "");
Expect(1, 2142, '\P{^Script=mandaic}', "");
Expect(0, 2143, '\p{Script=mandaic}', "");
Expect(1, 2143, '\p{^Script=mandaic}', "");
Expect(1, 2143, '\P{Script=mandaic}', "");
Expect(0, 2143, '\P{^Script=mandaic}', "");
Expect(1, 2142, '\p{Script=_MANDAIC}', "");
Expect(0, 2142, '\p{^Script=_MANDAIC}', "");
Expect(0, 2142, '\P{Script=_MANDAIC}', "");
Expect(1, 2142, '\P{^Script=_MANDAIC}', "");
Expect(0, 2143, '\p{Script=_MANDAIC}', "");
Expect(1, 2143, '\p{^Script=_MANDAIC}', "");
Expect(1, 2143, '\P{Script=_MANDAIC}', "");
Expect(0, 2143, '\P{^Script=_MANDAIC}', "");
Error('\p{Sc=:= -mand}');
Error('\P{Sc=:= -mand}');
Expect(1, 2142, '\p{Sc:   mand}', "");
Expect(0, 2142, '\p{^Sc:   mand}', "");
Expect(0, 2142, '\P{Sc:   mand}', "");
Expect(1, 2142, '\P{^Sc:   mand}', "");
Expect(0, 2143, '\p{Sc:   mand}', "");
Expect(1, 2143, '\p{^Sc:   mand}', "");
Expect(1, 2143, '\P{Sc:   mand}', "");
Expect(0, 2143, '\P{^Sc:   mand}', "");
Expect(1, 2142, '\p{Sc=- MAND}', "");
Expect(0, 2142, '\p{^Sc=- MAND}', "");
Expect(0, 2142, '\P{Sc=- MAND}', "");
Expect(1, 2142, '\P{^Sc=- MAND}', "");
Expect(0, 2143, '\p{Sc=- MAND}', "");
Expect(1, 2143, '\p{^Sc=- MAND}', "");
Expect(1, 2143, '\P{Sc=- MAND}', "");
Expect(0, 2143, '\P{^Sc=- MAND}', "");
Error('\p{Is_Script:   -Mandaic:=}');
Error('\P{Is_Script:   -Mandaic:=}');
Expect(1, 2142, '\p{Is_Script=mandaic}', "");
Expect(0, 2142, '\p{^Is_Script=mandaic}', "");
Expect(0, 2142, '\P{Is_Script=mandaic}', "");
Expect(1, 2142, '\P{^Is_Script=mandaic}', "");
Expect(0, 2143, '\p{Is_Script=mandaic}', "");
Expect(1, 2143, '\p{^Is_Script=mandaic}', "");
Expect(1, 2143, '\P{Is_Script=mandaic}', "");
Expect(0, 2143, '\P{^Is_Script=mandaic}', "");
Expect(1, 2142, '\p{Is_Script=-Mandaic}', "");
Expect(0, 2142, '\p{^Is_Script=-Mandaic}', "");
Expect(0, 2142, '\P{Is_Script=-Mandaic}', "");
Expect(1, 2142, '\P{^Is_Script=-Mandaic}', "");
Expect(0, 2143, '\p{Is_Script=-Mandaic}', "");
Expect(1, 2143, '\p{^Is_Script=-Mandaic}', "");
Expect(1, 2143, '\P{Is_Script=-Mandaic}', "");
Expect(0, 2143, '\P{^Is_Script=-Mandaic}', "");
Error('\p{Is_Sc=-/a/Mand}');
Error('\P{Is_Sc=-/a/Mand}');
Expect(1, 2142, '\p{Is_Sc=mand}', "");
Expect(0, 2142, '\p{^Is_Sc=mand}', "");
Expect(0, 2142, '\P{Is_Sc=mand}', "");
Expect(1, 2142, '\P{^Is_Sc=mand}', "");
Expect(0, 2143, '\p{Is_Sc=mand}', "");
Expect(1, 2143, '\p{^Is_Sc=mand}', "");
Expect(1, 2143, '\P{Is_Sc=mand}', "");
Expect(0, 2143, '\P{^Is_Sc=mand}', "");
Expect(1, 2142, '\p{Is_Sc=--mand}', "");
Expect(0, 2142, '\p{^Is_Sc=--mand}', "");
Expect(0, 2142, '\P{Is_Sc=--mand}', "");
Expect(1, 2142, '\P{^Is_Sc=--mand}', "");
Expect(0, 2143, '\p{Is_Sc=--mand}', "");
Expect(1, 2143, '\p{^Is_Sc=--mand}', "");
Expect(1, 2143, '\P{Is_Sc=--mand}', "");
Expect(0, 2143, '\P{^Is_Sc=--mand}', "");
Error('\p{Script=	Meroitic_cursive/a/}');
Error('\P{Script=	Meroitic_cursive/a/}');
Expect(1, 68031, '\p{Script=meroiticcursive}', "");
Expect(0, 68031, '\p{^Script=meroiticcursive}', "");
Expect(0, 68031, '\P{Script=meroiticcursive}', "");
Expect(1, 68031, '\P{^Script=meroiticcursive}', "");
Expect(0, 68032, '\p{Script=meroiticcursive}', "");
Expect(1, 68032, '\p{^Script=meroiticcursive}', "");
Expect(1, 68032, '\P{Script=meroiticcursive}', "");
Expect(0, 68032, '\P{^Script=meroiticcursive}', "");
Expect(1, 68031, '\p{Script=	MEROITIC_CURSIVE}', "");
Expect(0, 68031, '\p{^Script=	MEROITIC_CURSIVE}', "");
Expect(0, 68031, '\P{Script=	MEROITIC_CURSIVE}', "");
Expect(1, 68031, '\P{^Script=	MEROITIC_CURSIVE}', "");
Expect(0, 68032, '\p{Script=	MEROITIC_CURSIVE}', "");
Expect(1, 68032, '\p{^Script=	MEROITIC_CURSIVE}', "");
Expect(1, 68032, '\P{Script=	MEROITIC_CURSIVE}', "");
Expect(0, 68032, '\P{^Script=	MEROITIC_CURSIVE}', "");
Error('\p{Sc=_/a/Merc}');
Error('\P{Sc=_/a/Merc}');
Expect(1, 68031, '\p{Sc=merc}', "");
Expect(0, 68031, '\p{^Sc=merc}', "");
Expect(0, 68031, '\P{Sc=merc}', "");
Expect(1, 68031, '\P{^Sc=merc}', "");
Expect(0, 68032, '\p{Sc=merc}', "");
Expect(1, 68032, '\p{^Sc=merc}', "");
Expect(1, 68032, '\P{Sc=merc}', "");
Expect(0, 68032, '\P{^Sc=merc}', "");
Expect(1, 68031, '\p{Sc=		Merc}', "");
Expect(0, 68031, '\p{^Sc=		Merc}', "");
Expect(0, 68031, '\P{Sc=		Merc}', "");
Expect(1, 68031, '\P{^Sc=		Merc}', "");
Expect(0, 68032, '\p{Sc=		Merc}', "");
Expect(1, 68032, '\p{^Sc=		Merc}', "");
Expect(1, 68032, '\P{Sc=		Merc}', "");
Expect(0, 68032, '\P{^Sc=		Merc}', "");
Error('\p{Is_Script:	_:=meroitic_Cursive}');
Error('\P{Is_Script:	_:=meroitic_Cursive}');
Expect(1, 68031, '\p{Is_Script=meroiticcursive}', "");
Expect(0, 68031, '\p{^Is_Script=meroiticcursive}', "");
Expect(0, 68031, '\P{Is_Script=meroiticcursive}', "");
Expect(1, 68031, '\P{^Is_Script=meroiticcursive}', "");
Expect(0, 68032, '\p{Is_Script=meroiticcursive}', "");
Expect(1, 68032, '\p{^Is_Script=meroiticcursive}', "");
Expect(1, 68032, '\P{Is_Script=meroiticcursive}', "");
Expect(0, 68032, '\P{^Is_Script=meroiticcursive}', "");
Expect(1, 68031, '\p{Is_Script= MEROITIC_Cursive}', "");
Expect(0, 68031, '\p{^Is_Script= MEROITIC_Cursive}', "");
Expect(0, 68031, '\P{Is_Script= MEROITIC_Cursive}', "");
Expect(1, 68031, '\P{^Is_Script= MEROITIC_Cursive}', "");
Expect(0, 68032, '\p{Is_Script= MEROITIC_Cursive}', "");
Expect(1, 68032, '\p{^Is_Script= MEROITIC_Cursive}', "");
Expect(1, 68032, '\P{Is_Script= MEROITIC_Cursive}', "");
Expect(0, 68032, '\P{^Is_Script= MEROITIC_Cursive}', "");
Error('\p{Is_Sc: /a/_MERC}');
Error('\P{Is_Sc: /a/_MERC}');
Expect(1, 68031, '\p{Is_Sc=merc}', "");
Expect(0, 68031, '\p{^Is_Sc=merc}', "");
Expect(0, 68031, '\P{Is_Sc=merc}', "");
Expect(1, 68031, '\P{^Is_Sc=merc}', "");
Expect(0, 68032, '\p{Is_Sc=merc}', "");
Expect(1, 68032, '\p{^Is_Sc=merc}', "");
Expect(1, 68032, '\P{Is_Sc=merc}', "");
Expect(0, 68032, '\P{^Is_Sc=merc}', "");
Expect(1, 68031, '\p{Is_Sc= -Merc}', "");
Expect(0, 68031, '\p{^Is_Sc= -Merc}', "");
Expect(0, 68031, '\P{Is_Sc= -Merc}', "");
Expect(1, 68031, '\P{^Is_Sc= -Merc}', "");
Expect(0, 68032, '\p{Is_Sc= -Merc}', "");
Expect(1, 68032, '\p{^Is_Sc= -Merc}', "");
Expect(1, 68032, '\P{Is_Sc= -Merc}', "");
Expect(0, 68032, '\P{^Is_Sc= -Merc}', "");
Error('\p{Script=:=-	Meroitic_hieroglyphs}');
Error('\P{Script=:=-	Meroitic_hieroglyphs}');
Expect(1, 67999, '\p{Script=meroitichieroglyphs}', "");
Expect(0, 67999, '\p{^Script=meroitichieroglyphs}', "");
Expect(0, 67999, '\P{Script=meroitichieroglyphs}', "");
Expect(1, 67999, '\P{^Script=meroitichieroglyphs}', "");
Expect(0, 68000, '\p{Script=meroitichieroglyphs}', "");
Expect(1, 68000, '\p{^Script=meroitichieroglyphs}', "");
Expect(1, 68000, '\P{Script=meroitichieroglyphs}', "");
Expect(0, 68000, '\P{^Script=meroitichieroglyphs}', "");
Expect(1, 67999, '\p{Script= 	Meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\p{^Script= 	Meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\P{Script= 	Meroitic_Hieroglyphs}', "");
Expect(1, 67999, '\P{^Script= 	Meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\p{Script= 	Meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\p{^Script= 	Meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\P{Script= 	Meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\P{^Script= 	Meroitic_Hieroglyphs}', "");
Error('\p{Sc=/a/MERO}');
Error('\P{Sc=/a/MERO}');
Expect(1, 67999, '\p{Sc=mero}', "");
Expect(0, 67999, '\p{^Sc=mero}', "");
Expect(0, 67999, '\P{Sc=mero}', "");
Expect(1, 67999, '\P{^Sc=mero}', "");
Expect(0, 68000, '\p{Sc=mero}', "");
Expect(1, 68000, '\p{^Sc=mero}', "");
Expect(1, 68000, '\P{Sc=mero}', "");
Expect(0, 68000, '\P{^Sc=mero}', "");
Expect(1, 67999, '\p{Sc:   Mero}', "");
Expect(0, 67999, '\p{^Sc:   Mero}', "");
Expect(0, 67999, '\P{Sc:   Mero}', "");
Expect(1, 67999, '\P{^Sc:   Mero}', "");
Expect(0, 68000, '\p{Sc:   Mero}', "");
Expect(1, 68000, '\p{^Sc:   Mero}', "");
Expect(1, 68000, '\P{Sc:   Mero}', "");
Expect(0, 68000, '\P{^Sc:   Mero}', "");
Error('\p{Is_Script=/a/_Meroitic_Hieroglyphs}');
Error('\P{Is_Script=/a/_Meroitic_Hieroglyphs}');
Expect(1, 67999, '\p{Is_Script=meroitichieroglyphs}', "");
Expect(0, 67999, '\p{^Is_Script=meroitichieroglyphs}', "");
Expect(0, 67999, '\P{Is_Script=meroitichieroglyphs}', "");
Expect(1, 67999, '\P{^Is_Script=meroitichieroglyphs}', "");
Expect(0, 68000, '\p{Is_Script=meroitichieroglyphs}', "");
Expect(1, 68000, '\p{^Is_Script=meroitichieroglyphs}', "");
Expect(1, 68000, '\P{Is_Script=meroitichieroglyphs}', "");
Expect(0, 68000, '\P{^Is_Script=meroitichieroglyphs}', "");
Expect(1, 67999, '\p{Is_Script= 	MEROITIC_hieroglyphs}', "");
Expect(0, 67999, '\p{^Is_Script= 	MEROITIC_hieroglyphs}', "");
Expect(0, 67999, '\P{Is_Script= 	MEROITIC_hieroglyphs}', "");
Expect(1, 67999, '\P{^Is_Script= 	MEROITIC_hieroglyphs}', "");
Expect(0, 68000, '\p{Is_Script= 	MEROITIC_hieroglyphs}', "");
Expect(1, 68000, '\p{^Is_Script= 	MEROITIC_hieroglyphs}', "");
Expect(1, 68000, '\P{Is_Script= 	MEROITIC_hieroglyphs}', "");
Expect(0, 68000, '\P{^Is_Script= 	MEROITIC_hieroglyphs}', "");
Error('\p{Is_Sc=:=_-Mero}');
Error('\P{Is_Sc=:=_-Mero}');
Expect(1, 67999, '\p{Is_Sc:mero}', "");
Expect(0, 67999, '\p{^Is_Sc:mero}', "");
Expect(0, 67999, '\P{Is_Sc:mero}', "");
Expect(1, 67999, '\P{^Is_Sc:mero}', "");
Expect(0, 68000, '\p{Is_Sc:mero}', "");
Expect(1, 68000, '\p{^Is_Sc:mero}', "");
Expect(1, 68000, '\P{Is_Sc:mero}', "");
Expect(0, 68000, '\P{^Is_Sc:mero}', "");
Expect(1, 67999, '\p{Is_Sc=-mero}', "");
Expect(0, 67999, '\p{^Is_Sc=-mero}', "");
Expect(0, 67999, '\P{Is_Sc=-mero}', "");
Expect(1, 67999, '\P{^Is_Sc=-mero}', "");
Expect(0, 68000, '\p{Is_Sc=-mero}', "");
Expect(1, 68000, '\p{^Is_Sc=-mero}', "");
Expect(1, 68000, '\P{Is_Sc=-mero}', "");
Expect(0, 68000, '\P{^Is_Sc=-mero}', "");
Error('\p{Script=/a/	-Malayalam}');
Error('\P{Script=/a/	-Malayalam}');
Expect(1, 3455, '\p{Script=malayalam}', "");
Expect(0, 3455, '\p{^Script=malayalam}', "");
Expect(0, 3455, '\P{Script=malayalam}', "");
Expect(1, 3455, '\P{^Script=malayalam}', "");
Expect(0, 3456, '\p{Script=malayalam}', "");
Expect(1, 3456, '\p{^Script=malayalam}', "");
Expect(1, 3456, '\P{Script=malayalam}', "");
Expect(0, 3456, '\P{^Script=malayalam}', "");
Expect(1, 3455, '\p{Script=_-malayalam}', "");
Expect(0, 3455, '\p{^Script=_-malayalam}', "");
Expect(0, 3455, '\P{Script=_-malayalam}', "");
Expect(1, 3455, '\P{^Script=_-malayalam}', "");
Expect(0, 3456, '\p{Script=_-malayalam}', "");
Expect(1, 3456, '\p{^Script=_-malayalam}', "");
Expect(1, 3456, '\P{Script=_-malayalam}', "");
Expect(0, 3456, '\P{^Script=_-malayalam}', "");
Error('\p{Sc= Mlym/a/}');
Error('\P{Sc= Mlym/a/}');
Expect(1, 3455, '\p{Sc=mlym}', "");
Expect(0, 3455, '\p{^Sc=mlym}', "");
Expect(0, 3455, '\P{Sc=mlym}', "");
Expect(1, 3455, '\P{^Sc=mlym}', "");
Expect(0, 3456, '\p{Sc=mlym}', "");
Expect(1, 3456, '\p{^Sc=mlym}', "");
Expect(1, 3456, '\P{Sc=mlym}', "");
Expect(0, 3456, '\P{^Sc=mlym}', "");
Expect(1, 3455, '\p{Sc=  mlym}', "");
Expect(0, 3455, '\p{^Sc=  mlym}', "");
Expect(0, 3455, '\P{Sc=  mlym}', "");
Expect(1, 3455, '\P{^Sc=  mlym}', "");
Expect(0, 3456, '\p{Sc=  mlym}', "");
Expect(1, 3456, '\p{^Sc=  mlym}', "");
Expect(1, 3456, '\P{Sc=  mlym}', "");
Expect(0, 3456, '\P{^Sc=  mlym}', "");
Error('\p{Is_Script=	_MALAYALAM:=}');
Error('\P{Is_Script=	_MALAYALAM:=}');
Expect(1, 3455, '\p{Is_Script=malayalam}', "");
Expect(0, 3455, '\p{^Is_Script=malayalam}', "");
Expect(0, 3455, '\P{Is_Script=malayalam}', "");
Expect(1, 3455, '\P{^Is_Script=malayalam}', "");
Expect(0, 3456, '\p{Is_Script=malayalam}', "");
Expect(1, 3456, '\p{^Is_Script=malayalam}', "");
Expect(1, 3456, '\P{Is_Script=malayalam}', "");
Expect(0, 3456, '\P{^Is_Script=malayalam}', "");
Expect(1, 3455, '\p{Is_Script=- Malayalam}', "");
Expect(0, 3455, '\p{^Is_Script=- Malayalam}', "");
Expect(0, 3455, '\P{Is_Script=- Malayalam}', "");
Expect(1, 3455, '\P{^Is_Script=- Malayalam}', "");
Expect(0, 3456, '\p{Is_Script=- Malayalam}', "");
Expect(1, 3456, '\p{^Is_Script=- Malayalam}', "");
Expect(1, 3456, '\P{Is_Script=- Malayalam}', "");
Expect(0, 3456, '\P{^Is_Script=- Malayalam}', "");
Error('\p{Is_Sc=/a/ mlym}');
Error('\P{Is_Sc=/a/ mlym}');
Expect(1, 3455, '\p{Is_Sc=mlym}', "");
Expect(0, 3455, '\p{^Is_Sc=mlym}', "");
Expect(0, 3455, '\P{Is_Sc=mlym}', "");
Expect(1, 3455, '\P{^Is_Sc=mlym}', "");
Expect(0, 3456, '\p{Is_Sc=mlym}', "");
Expect(1, 3456, '\p{^Is_Sc=mlym}', "");
Expect(1, 3456, '\P{Is_Sc=mlym}', "");
Expect(0, 3456, '\P{^Is_Sc=mlym}', "");
Expect(1, 3455, '\p{Is_Sc=- Mlym}', "");
Expect(0, 3455, '\p{^Is_Sc=- Mlym}', "");
Expect(0, 3455, '\P{Is_Sc=- Mlym}', "");
Expect(1, 3455, '\P{^Is_Sc=- Mlym}', "");
Expect(0, 3456, '\p{Is_Sc=- Mlym}', "");
Expect(1, 3456, '\p{^Is_Sc=- Mlym}', "");
Expect(1, 3456, '\P{Is_Sc=- Mlym}', "");
Expect(0, 3456, '\P{^Is_Sc=- Mlym}', "");
Error('\p{Script= :=MONGOLIAN}');
Error('\P{Script= :=MONGOLIAN}');
Expect(1, 6314, '\p{Script: mongolian}', "");
Expect(0, 6314, '\p{^Script: mongolian}', "");
Expect(0, 6314, '\P{Script: mongolian}', "");
Expect(1, 6314, '\P{^Script: mongolian}', "");
Expect(0, 6315, '\p{Script: mongolian}', "");
Expect(1, 6315, '\p{^Script: mongolian}', "");
Expect(1, 6315, '\P{Script: mongolian}', "");
Expect(0, 6315, '\P{^Script: mongolian}', "");
Expect(1, 6314, '\p{Script=-_mongolian}', "");
Expect(0, 6314, '\p{^Script=-_mongolian}', "");
Expect(0, 6314, '\P{Script=-_mongolian}', "");
Expect(1, 6314, '\P{^Script=-_mongolian}', "");
Expect(0, 6315, '\p{Script=-_mongolian}', "");
Expect(1, 6315, '\p{^Script=-_mongolian}', "");
Expect(1, 6315, '\P{Script=-_mongolian}', "");
Expect(0, 6315, '\P{^Script=-_mongolian}', "");
Error('\p{Sc=_:=Mong}');
Error('\P{Sc=_:=Mong}');
Expect(1, 6314, '\p{Sc=mong}', "");
Expect(0, 6314, '\p{^Sc=mong}', "");
Expect(0, 6314, '\P{Sc=mong}', "");
Expect(1, 6314, '\P{^Sc=mong}', "");
Expect(0, 6315, '\p{Sc=mong}', "");
Expect(1, 6315, '\p{^Sc=mong}', "");
Expect(1, 6315, '\P{Sc=mong}', "");
Expect(0, 6315, '\P{^Sc=mong}', "");
Expect(1, 6314, '\p{Sc=__mong}', "");
Expect(0, 6314, '\p{^Sc=__mong}', "");
Expect(0, 6314, '\P{Sc=__mong}', "");
Expect(1, 6314, '\P{^Sc=__mong}', "");
Expect(0, 6315, '\p{Sc=__mong}', "");
Expect(1, 6315, '\p{^Sc=__mong}', "");
Expect(1, 6315, '\P{Sc=__mong}', "");
Expect(0, 6315, '\P{^Sc=__mong}', "");
Error('\p{Is_Script=__Mongolian/a/}');
Error('\P{Is_Script=__Mongolian/a/}');
Expect(1, 6314, '\p{Is_Script=mongolian}', "");
Expect(0, 6314, '\p{^Is_Script=mongolian}', "");
Expect(0, 6314, '\P{Is_Script=mongolian}', "");
Expect(1, 6314, '\P{^Is_Script=mongolian}', "");
Expect(0, 6315, '\p{Is_Script=mongolian}', "");
Expect(1, 6315, '\p{^Is_Script=mongolian}', "");
Expect(1, 6315, '\P{Is_Script=mongolian}', "");
Expect(0, 6315, '\P{^Is_Script=mongolian}', "");
Expect(1, 6314, '\p{Is_Script=	Mongolian}', "");
Expect(0, 6314, '\p{^Is_Script=	Mongolian}', "");
Expect(0, 6314, '\P{Is_Script=	Mongolian}', "");
Expect(1, 6314, '\P{^Is_Script=	Mongolian}', "");
Expect(0, 6315, '\p{Is_Script=	Mongolian}', "");
Expect(1, 6315, '\p{^Is_Script=	Mongolian}', "");
Expect(1, 6315, '\P{Is_Script=	Mongolian}', "");
Expect(0, 6315, '\P{^Is_Script=	Mongolian}', "");
Error('\p{Is_Sc:/a/_Mong}');
Error('\P{Is_Sc:/a/_Mong}');
Expect(1, 6314, '\p{Is_Sc=mong}', "");
Expect(0, 6314, '\p{^Is_Sc=mong}', "");
Expect(0, 6314, '\P{Is_Sc=mong}', "");
Expect(1, 6314, '\P{^Is_Sc=mong}', "");
Expect(0, 6315, '\p{Is_Sc=mong}', "");
Expect(1, 6315, '\p{^Is_Sc=mong}', "");
Expect(1, 6315, '\P{Is_Sc=mong}', "");
Expect(0, 6315, '\P{^Is_Sc=mong}', "");
Expect(1, 6314, '\p{Is_Sc= MONG}', "");
Expect(0, 6314, '\p{^Is_Sc= MONG}', "");
Expect(0, 6314, '\P{Is_Sc= MONG}', "");
Expect(1, 6314, '\P{^Is_Sc= MONG}', "");
Expect(0, 6315, '\p{Is_Sc= MONG}', "");
Expect(1, 6315, '\p{^Is_Sc= MONG}', "");
Expect(1, 6315, '\P{Is_Sc= MONG}', "");
Expect(0, 6315, '\P{^Is_Sc= MONG}', "");
Error('\p{Script=/a/ meetei_MAYEK}');
Error('\P{Script=/a/ meetei_MAYEK}');
Expect(1, 44025, '\p{Script:meeteimayek}', "");
Expect(0, 44025, '\p{^Script:meeteimayek}', "");
Expect(0, 44025, '\P{Script:meeteimayek}', "");
Expect(1, 44025, '\P{^Script:meeteimayek}', "");
Expect(0, 44026, '\p{Script:meeteimayek}', "");
Expect(1, 44026, '\p{^Script:meeteimayek}', "");
Expect(1, 44026, '\P{Script:meeteimayek}', "");
Expect(0, 44026, '\P{^Script:meeteimayek}', "");
Expect(1, 44025, '\p{Script=-Meetei_Mayek}', "");
Expect(0, 44025, '\p{^Script=-Meetei_Mayek}', "");
Expect(0, 44025, '\P{Script=-Meetei_Mayek}', "");
Expect(1, 44025, '\P{^Script=-Meetei_Mayek}', "");
Expect(0, 44026, '\p{Script=-Meetei_Mayek}', "");
Expect(1, 44026, '\p{^Script=-Meetei_Mayek}', "");
Expect(1, 44026, '\P{Script=-Meetei_Mayek}', "");
Expect(0, 44026, '\P{^Script=-Meetei_Mayek}', "");
Error('\p{Sc=:=	mtei}');
Error('\P{Sc=:=	mtei}');
Expect(1, 44025, '\p{Sc=mtei}', "");
Expect(0, 44025, '\p{^Sc=mtei}', "");
Expect(0, 44025, '\P{Sc=mtei}', "");
Expect(1, 44025, '\P{^Sc=mtei}', "");
Expect(0, 44026, '\p{Sc=mtei}', "");
Expect(1, 44026, '\p{^Sc=mtei}', "");
Expect(1, 44026, '\P{Sc=mtei}', "");
Expect(0, 44026, '\P{^Sc=mtei}', "");
Expect(1, 44025, '\p{Sc:   --Mtei}', "");
Expect(0, 44025, '\p{^Sc:   --Mtei}', "");
Expect(0, 44025, '\P{Sc:   --Mtei}', "");
Expect(1, 44025, '\P{^Sc:   --Mtei}', "");
Expect(0, 44026, '\p{Sc:   --Mtei}', "");
Expect(1, 44026, '\p{^Sc:   --Mtei}', "");
Expect(1, 44026, '\P{Sc:   --Mtei}', "");
Expect(0, 44026, '\P{^Sc:   --Mtei}', "");
Error('\p{Is_Script=	:=Meetei_Mayek}');
Error('\P{Is_Script=	:=Meetei_Mayek}');
Expect(1, 44025, '\p{Is_Script=meeteimayek}', "");
Expect(0, 44025, '\p{^Is_Script=meeteimayek}', "");
Expect(0, 44025, '\P{Is_Script=meeteimayek}', "");
Expect(1, 44025, '\P{^Is_Script=meeteimayek}', "");
Expect(0, 44026, '\p{Is_Script=meeteimayek}', "");
Expect(1, 44026, '\p{^Is_Script=meeteimayek}', "");
Expect(1, 44026, '\P{Is_Script=meeteimayek}', "");
Expect(0, 44026, '\P{^Is_Script=meeteimayek}', "");
Expect(1, 44025, '\p{Is_Script:	- Meetei_MAYEK}', "");
Expect(0, 44025, '\p{^Is_Script:	- Meetei_MAYEK}', "");
Expect(0, 44025, '\P{Is_Script:	- Meetei_MAYEK}', "");
Expect(1, 44025, '\P{^Is_Script:	- Meetei_MAYEK}', "");
Expect(0, 44026, '\p{Is_Script:	- Meetei_MAYEK}', "");
Expect(1, 44026, '\p{^Is_Script:	- Meetei_MAYEK}', "");
Expect(1, 44026, '\P{Is_Script:	- Meetei_MAYEK}', "");
Expect(0, 44026, '\P{^Is_Script:	- Meetei_MAYEK}', "");
Error('\p{Is_Sc=--Mtei/a/}');
Error('\P{Is_Sc=--Mtei/a/}');
Expect(1, 44025, '\p{Is_Sc=mtei}', "");
Expect(0, 44025, '\p{^Is_Sc=mtei}', "");
Expect(0, 44025, '\P{Is_Sc=mtei}', "");
Expect(1, 44025, '\P{^Is_Sc=mtei}', "");
Expect(0, 44026, '\p{Is_Sc=mtei}', "");
Expect(1, 44026, '\p{^Is_Sc=mtei}', "");
Expect(1, 44026, '\P{Is_Sc=mtei}', "");
Expect(0, 44026, '\P{^Is_Sc=mtei}', "");
Expect(1, 44025, '\p{Is_Sc= -mtei}', "");
Expect(0, 44025, '\p{^Is_Sc= -mtei}', "");
Expect(0, 44025, '\P{Is_Sc= -mtei}', "");
Expect(1, 44025, '\P{^Is_Sc= -mtei}', "");
Expect(0, 44026, '\p{Is_Sc= -mtei}', "");
Expect(1, 44026, '\p{^Is_Sc= -mtei}', "");
Expect(1, 44026, '\P{Is_Sc= -mtei}', "");
Expect(0, 44026, '\P{^Is_Sc= -mtei}', "");
Error('\p{Script=_MYANMAR/a/}');
Error('\P{Script=_MYANMAR/a/}');
Expect(1, 43643, '\p{Script=myanmar}', "");
Expect(0, 43643, '\p{^Script=myanmar}', "");
Expect(0, 43643, '\P{Script=myanmar}', "");
Expect(1, 43643, '\P{^Script=myanmar}', "");
Expect(0, 43644, '\p{Script=myanmar}', "");
Expect(1, 43644, '\p{^Script=myanmar}', "");
Expect(1, 43644, '\P{Script=myanmar}', "");
Expect(0, 43644, '\P{^Script=myanmar}', "");
Expect(1, 43643, '\p{Script:	_Myanmar}', "");
Expect(0, 43643, '\p{^Script:	_Myanmar}', "");
Expect(0, 43643, '\P{Script:	_Myanmar}', "");
Expect(1, 43643, '\P{^Script:	_Myanmar}', "");
Expect(0, 43644, '\p{Script:	_Myanmar}', "");
Expect(1, 43644, '\p{^Script:	_Myanmar}', "");
Expect(1, 43644, '\P{Script:	_Myanmar}', "");
Expect(0, 43644, '\P{^Script:	_Myanmar}', "");
Error('\p{Sc=-/a/MYMR}');
Error('\P{Sc=-/a/MYMR}');
Expect(1, 43643, '\p{Sc=mymr}', "");
Expect(0, 43643, '\p{^Sc=mymr}', "");
Expect(0, 43643, '\P{Sc=mymr}', "");
Expect(1, 43643, '\P{^Sc=mymr}', "");
Expect(0, 43644, '\p{Sc=mymr}', "");
Expect(1, 43644, '\p{^Sc=mymr}', "");
Expect(1, 43644, '\P{Sc=mymr}', "");
Expect(0, 43644, '\P{^Sc=mymr}', "");
Expect(1, 43643, '\p{Sc=-_MYMR}', "");
Expect(0, 43643, '\p{^Sc=-_MYMR}', "");
Expect(0, 43643, '\P{Sc=-_MYMR}', "");
Expect(1, 43643, '\P{^Sc=-_MYMR}', "");
Expect(0, 43644, '\p{Sc=-_MYMR}', "");
Expect(1, 43644, '\p{^Sc=-_MYMR}', "");
Expect(1, 43644, '\P{Sc=-_MYMR}', "");
Expect(0, 43644, '\P{^Sc=-_MYMR}', "");
Error('\p{Is_Script=Myanmar/a/}');
Error('\P{Is_Script=Myanmar/a/}');
Expect(1, 43643, '\p{Is_Script=myanmar}', "");
Expect(0, 43643, '\p{^Is_Script=myanmar}', "");
Expect(0, 43643, '\P{Is_Script=myanmar}', "");
Expect(1, 43643, '\P{^Is_Script=myanmar}', "");
Expect(0, 43644, '\p{Is_Script=myanmar}', "");
Expect(1, 43644, '\p{^Is_Script=myanmar}', "");
Expect(1, 43644, '\P{Is_Script=myanmar}', "");
Expect(0, 43644, '\P{^Is_Script=myanmar}', "");
Expect(1, 43643, '\p{Is_Script=_	MYANMAR}', "");
Expect(0, 43643, '\p{^Is_Script=_	MYANMAR}', "");
Expect(0, 43643, '\P{Is_Script=_	MYANMAR}', "");
Expect(1, 43643, '\P{^Is_Script=_	MYANMAR}', "");
Expect(0, 43644, '\p{Is_Script=_	MYANMAR}', "");
Expect(1, 43644, '\p{^Is_Script=_	MYANMAR}', "");
Expect(1, 43644, '\P{Is_Script=_	MYANMAR}', "");
Expect(0, 43644, '\P{^Is_Script=_	MYANMAR}', "");
Error('\p{Is_Sc=-_Mymr/a/}');
Error('\P{Is_Sc=-_Mymr/a/}');
Expect(1, 43643, '\p{Is_Sc=mymr}', "");
Expect(0, 43643, '\p{^Is_Sc=mymr}', "");
Expect(0, 43643, '\P{Is_Sc=mymr}', "");
Expect(1, 43643, '\P{^Is_Sc=mymr}', "");
Expect(0, 43644, '\p{Is_Sc=mymr}', "");
Expect(1, 43644, '\p{^Is_Sc=mymr}', "");
Expect(1, 43644, '\P{Is_Sc=mymr}', "");
Expect(0, 43644, '\P{^Is_Sc=mymr}', "");
Expect(1, 43643, '\p{Is_Sc= Mymr}', "");
Expect(0, 43643, '\p{^Is_Sc= Mymr}', "");
Expect(0, 43643, '\P{Is_Sc= Mymr}', "");
Expect(1, 43643, '\P{^Is_Sc= Mymr}', "");
Expect(0, 43644, '\p{Is_Sc= Mymr}', "");
Expect(1, 43644, '\p{^Is_Sc= Mymr}', "");
Expect(1, 43644, '\P{Is_Sc= Mymr}', "");
Expect(0, 43644, '\P{^Is_Sc= Mymr}', "");
Error('\p{Script=:=_-Nko}');
Error('\P{Script=:=_-Nko}');
Expect(1, 2042, '\p{Script=nko}', "");
Expect(0, 2042, '\p{^Script=nko}', "");
Expect(0, 2042, '\P{Script=nko}', "");
Expect(1, 2042, '\P{^Script=nko}', "");
Expect(0, 2043, '\p{Script=nko}', "");
Expect(1, 2043, '\p{^Script=nko}', "");
Expect(1, 2043, '\P{Script=nko}', "");
Expect(0, 2043, '\P{^Script=nko}', "");
Expect(1, 2042, '\p{Script=__NKO}', "");
Expect(0, 2042, '\p{^Script=__NKO}', "");
Expect(0, 2042, '\P{Script=__NKO}', "");
Expect(1, 2042, '\P{^Script=__NKO}', "");
Expect(0, 2043, '\p{Script=__NKO}', "");
Expect(1, 2043, '\p{^Script=__NKO}', "");
Expect(1, 2043, '\P{Script=__NKO}', "");
Expect(0, 2043, '\P{^Script=__NKO}', "");
Error('\p{Sc=/a/- Nkoo}');
Error('\P{Sc=/a/- Nkoo}');
Expect(1, 2042, '\p{Sc=nkoo}', "");
Expect(0, 2042, '\p{^Sc=nkoo}', "");
Expect(0, 2042, '\P{Sc=nkoo}', "");
Expect(1, 2042, '\P{^Sc=nkoo}', "");
Expect(0, 2043, '\p{Sc=nkoo}', "");
Expect(1, 2043, '\p{^Sc=nkoo}', "");
Expect(1, 2043, '\P{Sc=nkoo}', "");
Expect(0, 2043, '\P{^Sc=nkoo}', "");
Expect(1, 2042, '\p{Sc=-NKOO}', "");
Expect(0, 2042, '\p{^Sc=-NKOO}', "");
Expect(0, 2042, '\P{Sc=-NKOO}', "");
Expect(1, 2042, '\P{^Sc=-NKOO}', "");
Expect(0, 2043, '\p{Sc=-NKOO}', "");
Expect(1, 2043, '\p{^Sc=-NKOO}', "");
Expect(1, 2043, '\P{Sc=-NKOO}', "");
Expect(0, 2043, '\P{^Sc=-NKOO}', "");
Error('\p{Is_Script=/a/NKO}');
Error('\P{Is_Script=/a/NKO}');
Expect(1, 2042, '\p{Is_Script=nko}', "");
Expect(0, 2042, '\p{^Is_Script=nko}', "");
Expect(0, 2042, '\P{Is_Script=nko}', "");
Expect(1, 2042, '\P{^Is_Script=nko}', "");
Expect(0, 2043, '\p{Is_Script=nko}', "");
Expect(1, 2043, '\p{^Is_Script=nko}', "");
Expect(1, 2043, '\P{Is_Script=nko}', "");
Expect(0, 2043, '\P{^Is_Script=nko}', "");
Expect(1, 2042, '\p{Is_Script:   -Nko}', "");
Expect(0, 2042, '\p{^Is_Script:   -Nko}', "");
Expect(0, 2042, '\P{Is_Script:   -Nko}', "");
Expect(1, 2042, '\P{^Is_Script:   -Nko}', "");
Expect(0, 2043, '\p{Is_Script:   -Nko}', "");
Expect(1, 2043, '\p{^Is_Script:   -Nko}', "");
Expect(1, 2043, '\P{Is_Script:   -Nko}', "");
Expect(0, 2043, '\P{^Is_Script:   -Nko}', "");
Error('\p{Is_Sc=/a/ nkoo}');
Error('\P{Is_Sc=/a/ nkoo}');
Expect(1, 2042, '\p{Is_Sc=nkoo}', "");
Expect(0, 2042, '\p{^Is_Sc=nkoo}', "");
Expect(0, 2042, '\P{Is_Sc=nkoo}', "");
Expect(1, 2042, '\P{^Is_Sc=nkoo}', "");
Expect(0, 2043, '\p{Is_Sc=nkoo}', "");
Expect(1, 2043, '\p{^Is_Sc=nkoo}', "");
Expect(1, 2043, '\P{Is_Sc=nkoo}', "");
Expect(0, 2043, '\P{^Is_Sc=nkoo}', "");
Expect(1, 2042, '\p{Is_Sc=_Nkoo}', "");
Expect(0, 2042, '\p{^Is_Sc=_Nkoo}', "");
Expect(0, 2042, '\P{Is_Sc=_Nkoo}', "");
Expect(1, 2042, '\P{^Is_Sc=_Nkoo}', "");
Expect(0, 2043, '\p{Is_Sc=_Nkoo}', "");
Expect(1, 2043, '\p{^Is_Sc=_Nkoo}', "");
Expect(1, 2043, '\P{Is_Sc=_Nkoo}', "");
Expect(0, 2043, '\P{^Is_Sc=_Nkoo}', "");
Error('\p{Script= OGHAM/a/}');
Error('\P{Script= OGHAM/a/}');
Expect(1, 5788, '\p{Script=ogham}', "");
Expect(0, 5788, '\p{^Script=ogham}', "");
Expect(0, 5788, '\P{Script=ogham}', "");
Expect(1, 5788, '\P{^Script=ogham}', "");
Expect(0, 5789, '\p{Script=ogham}', "");
Expect(1, 5789, '\p{^Script=ogham}', "");
Expect(1, 5789, '\P{Script=ogham}', "");
Expect(0, 5789, '\P{^Script=ogham}', "");
Expect(1, 5788, '\p{Script: Ogham}', "");
Expect(0, 5788, '\p{^Script: Ogham}', "");
Expect(0, 5788, '\P{Script: Ogham}', "");
Expect(1, 5788, '\P{^Script: Ogham}', "");
Expect(0, 5789, '\p{Script: Ogham}', "");
Expect(1, 5789, '\p{^Script: Ogham}', "");
Expect(1, 5789, '\P{Script: Ogham}', "");
Expect(0, 5789, '\P{^Script: Ogham}', "");
Error('\p{Sc=/a/ _OGAM}');
Error('\P{Sc=/a/ _OGAM}');
Expect(1, 5788, '\p{Sc:   ogam}', "");
Expect(0, 5788, '\p{^Sc:   ogam}', "");
Expect(0, 5788, '\P{Sc:   ogam}', "");
Expect(1, 5788, '\P{^Sc:   ogam}', "");
Expect(0, 5789, '\p{Sc:   ogam}', "");
Expect(1, 5789, '\p{^Sc:   ogam}', "");
Expect(1, 5789, '\P{Sc:   ogam}', "");
Expect(0, 5789, '\P{^Sc:   ogam}', "");
Expect(1, 5788, '\p{Sc= 	Ogam}', "");
Expect(0, 5788, '\p{^Sc= 	Ogam}', "");
Expect(0, 5788, '\P{Sc= 	Ogam}', "");
Expect(1, 5788, '\P{^Sc= 	Ogam}', "");
Expect(0, 5789, '\p{Sc= 	Ogam}', "");
Expect(1, 5789, '\p{^Sc= 	Ogam}', "");
Expect(1, 5789, '\P{Sc= 	Ogam}', "");
Expect(0, 5789, '\P{^Sc= 	Ogam}', "");
Error('\p{Is_Script=	_Ogham/a/}');
Error('\P{Is_Script=	_Ogham/a/}');
Expect(1, 5788, '\p{Is_Script=ogham}', "");
Expect(0, 5788, '\p{^Is_Script=ogham}', "");
Expect(0, 5788, '\P{Is_Script=ogham}', "");
Expect(1, 5788, '\P{^Is_Script=ogham}', "");
Expect(0, 5789, '\p{Is_Script=ogham}', "");
Expect(1, 5789, '\p{^Is_Script=ogham}', "");
Expect(1, 5789, '\P{Is_Script=ogham}', "");
Expect(0, 5789, '\P{^Is_Script=ogham}', "");
Expect(1, 5788, '\p{Is_Script= -ogham}', "");
Expect(0, 5788, '\p{^Is_Script= -ogham}', "");
Expect(0, 5788, '\P{Is_Script= -ogham}', "");
Expect(1, 5788, '\P{^Is_Script= -ogham}', "");
Expect(0, 5789, '\p{Is_Script= -ogham}', "");
Expect(1, 5789, '\p{^Is_Script= -ogham}', "");
Expect(1, 5789, '\P{Is_Script= -ogham}', "");
Expect(0, 5789, '\P{^Is_Script= -ogham}', "");
Error('\p{Is_Sc= /a/Ogam}');
Error('\P{Is_Sc= /a/Ogam}');
Expect(1, 5788, '\p{Is_Sc:ogam}', "");
Expect(0, 5788, '\p{^Is_Sc:ogam}', "");
Expect(0, 5788, '\P{Is_Sc:ogam}', "");
Expect(1, 5788, '\P{^Is_Sc:ogam}', "");
Expect(0, 5789, '\p{Is_Sc:ogam}', "");
Expect(1, 5789, '\p{^Is_Sc:ogam}', "");
Expect(1, 5789, '\P{Is_Sc:ogam}', "");
Expect(0, 5789, '\P{^Is_Sc:ogam}', "");
Expect(1, 5788, '\p{Is_Sc=_ogam}', "");
Expect(0, 5788, '\p{^Is_Sc=_ogam}', "");
Expect(0, 5788, '\P{Is_Sc=_ogam}', "");
Expect(1, 5788, '\P{^Is_Sc=_ogam}', "");
Expect(0, 5789, '\p{Is_Sc=_ogam}', "");
Expect(1, 5789, '\p{^Is_Sc=_ogam}', "");
Expect(1, 5789, '\P{Is_Sc=_ogam}', "");
Expect(0, 5789, '\P{^Is_Sc=_ogam}', "");
Error('\p{Script=/a/_ol_CHIKI}');
Error('\P{Script=/a/_ol_CHIKI}');
Expect(1, 7295, '\p{Script=olchiki}', "");
Expect(0, 7295, '\p{^Script=olchiki}', "");
Expect(0, 7295, '\P{Script=olchiki}', "");
Expect(1, 7295, '\P{^Script=olchiki}', "");
Expect(0, 7296, '\p{Script=olchiki}', "");
Expect(1, 7296, '\p{^Script=olchiki}', "");
Expect(1, 7296, '\P{Script=olchiki}', "");
Expect(0, 7296, '\P{^Script=olchiki}', "");
Expect(1, 7295, '\p{Script=_OL_Chiki}', "");
Expect(0, 7295, '\p{^Script=_OL_Chiki}', "");
Expect(0, 7295, '\P{Script=_OL_Chiki}', "");
Expect(1, 7295, '\P{^Script=_OL_Chiki}', "");
Expect(0, 7296, '\p{Script=_OL_Chiki}', "");
Expect(1, 7296, '\p{^Script=_OL_Chiki}', "");
Expect(1, 7296, '\P{Script=_OL_Chiki}', "");
Expect(0, 7296, '\P{^Script=_OL_Chiki}', "");
Error('\p{Sc=/a/__Olck}');
Error('\P{Sc=/a/__Olck}');
Expect(1, 7295, '\p{Sc=olck}', "");
Expect(0, 7295, '\p{^Sc=olck}', "");
Expect(0, 7295, '\P{Sc=olck}', "");
Expect(1, 7295, '\P{^Sc=olck}', "");
Expect(0, 7296, '\p{Sc=olck}', "");
Expect(1, 7296, '\p{^Sc=olck}', "");
Expect(1, 7296, '\P{Sc=olck}', "");
Expect(0, 7296, '\P{^Sc=olck}', "");
Expect(1, 7295, '\p{Sc=	olck}', "");
Expect(0, 7295, '\p{^Sc=	olck}', "");
Expect(0, 7295, '\P{Sc=	olck}', "");
Expect(1, 7295, '\P{^Sc=	olck}', "");
Expect(0, 7296, '\p{Sc=	olck}', "");
Expect(1, 7296, '\p{^Sc=	olck}', "");
Expect(1, 7296, '\P{Sc=	olck}', "");
Expect(0, 7296, '\P{^Sc=	olck}', "");
Error('\p{Is_Script:   -/a/OL_Chiki}');
Error('\P{Is_Script:   -/a/OL_Chiki}');
Expect(1, 7295, '\p{Is_Script=olchiki}', "");
Expect(0, 7295, '\p{^Is_Script=olchiki}', "");
Expect(0, 7295, '\P{Is_Script=olchiki}', "");
Expect(1, 7295, '\P{^Is_Script=olchiki}', "");
Expect(0, 7296, '\p{Is_Script=olchiki}', "");
Expect(1, 7296, '\p{^Is_Script=olchiki}', "");
Expect(1, 7296, '\P{Is_Script=olchiki}', "");
Expect(0, 7296, '\P{^Is_Script=olchiki}', "");
Expect(1, 7295, '\p{Is_Script=Ol_chiki}', "");
Expect(0, 7295, '\p{^Is_Script=Ol_chiki}', "");
Expect(0, 7295, '\P{Is_Script=Ol_chiki}', "");
Expect(1, 7295, '\P{^Is_Script=Ol_chiki}', "");
Expect(0, 7296, '\p{Is_Script=Ol_chiki}', "");
Expect(1, 7296, '\p{^Is_Script=Ol_chiki}', "");
Expect(1, 7296, '\P{Is_Script=Ol_chiki}', "");
Expect(0, 7296, '\P{^Is_Script=Ol_chiki}', "");
Error('\p{Is_Sc=_-Olck:=}');
Error('\P{Is_Sc=_-Olck:=}');
Expect(1, 7295, '\p{Is_Sc=olck}', "");
Expect(0, 7295, '\p{^Is_Sc=olck}', "");
Expect(0, 7295, '\P{Is_Sc=olck}', "");
Expect(1, 7295, '\P{^Is_Sc=olck}', "");
Expect(0, 7296, '\p{Is_Sc=olck}', "");
Expect(1, 7296, '\p{^Is_Sc=olck}', "");
Expect(1, 7296, '\P{Is_Sc=olck}', "");
Expect(0, 7296, '\P{^Is_Sc=olck}', "");
Expect(1, 7295, '\p{Is_Sc=__Olck}', "");
Expect(0, 7295, '\p{^Is_Sc=__Olck}', "");
Expect(0, 7295, '\P{Is_Sc=__Olck}', "");
Expect(1, 7295, '\P{^Is_Sc=__Olck}', "");
Expect(0, 7296, '\p{Is_Sc=__Olck}', "");
Expect(1, 7296, '\p{^Is_Sc=__Olck}', "");
Expect(1, 7296, '\P{Is_Sc=__Olck}', "");
Expect(0, 7296, '\P{^Is_Sc=__Olck}', "");
Error('\p{Script=_/a/Old_Turkic}');
Error('\P{Script=_/a/Old_Turkic}');
Expect(1, 68680, '\p{Script: oldturkic}', "");
Expect(0, 68680, '\p{^Script: oldturkic}', "");
Expect(0, 68680, '\P{Script: oldturkic}', "");
Expect(1, 68680, '\P{^Script: oldturkic}', "");
Expect(0, 68681, '\p{Script: oldturkic}', "");
Expect(1, 68681, '\p{^Script: oldturkic}', "");
Expect(1, 68681, '\P{Script: oldturkic}', "");
Expect(0, 68681, '\P{^Script: oldturkic}', "");
Expect(1, 68680, '\p{Script=  old_turkic}', "");
Expect(0, 68680, '\p{^Script=  old_turkic}', "");
Expect(0, 68680, '\P{Script=  old_turkic}', "");
Expect(1, 68680, '\P{^Script=  old_turkic}', "");
Expect(0, 68681, '\p{Script=  old_turkic}', "");
Expect(1, 68681, '\p{^Script=  old_turkic}', "");
Expect(1, 68681, '\P{Script=  old_turkic}', "");
Expect(0, 68681, '\P{^Script=  old_turkic}', "");
Error('\p{Sc=/a/	Orkh}');
Error('\P{Sc=/a/	Orkh}');
Expect(1, 68680, '\p{Sc=orkh}', "");
Expect(0, 68680, '\p{^Sc=orkh}', "");
Expect(0, 68680, '\P{Sc=orkh}', "");
Expect(1, 68680, '\P{^Sc=orkh}', "");
Expect(0, 68681, '\p{Sc=orkh}', "");
Expect(1, 68681, '\p{^Sc=orkh}', "");
Expect(1, 68681, '\P{Sc=orkh}', "");
Expect(0, 68681, '\P{^Sc=orkh}', "");
Expect(1, 68680, '\p{Sc= ORKH}', "");
Expect(0, 68680, '\p{^Sc= ORKH}', "");
Expect(0, 68680, '\P{Sc= ORKH}', "");
Expect(1, 68680, '\P{^Sc= ORKH}', "");
Expect(0, 68681, '\p{Sc= ORKH}', "");
Expect(1, 68681, '\p{^Sc= ORKH}', "");
Expect(1, 68681, '\P{Sc= ORKH}', "");
Expect(0, 68681, '\P{^Sc= ORKH}', "");
Error('\p{Is_Script=	:=Old_turkic}');
Error('\P{Is_Script=	:=Old_turkic}');
Expect(1, 68680, '\p{Is_Script=oldturkic}', "");
Expect(0, 68680, '\p{^Is_Script=oldturkic}', "");
Expect(0, 68680, '\P{Is_Script=oldturkic}', "");
Expect(1, 68680, '\P{^Is_Script=oldturkic}', "");
Expect(0, 68681, '\p{Is_Script=oldturkic}', "");
Expect(1, 68681, '\p{^Is_Script=oldturkic}', "");
Expect(1, 68681, '\P{Is_Script=oldturkic}', "");
Expect(0, 68681, '\P{^Is_Script=oldturkic}', "");
Expect(1, 68680, '\p{Is_Script:    -old_turkic}', "");
Expect(0, 68680, '\p{^Is_Script:    -old_turkic}', "");
Expect(0, 68680, '\P{Is_Script:    -old_turkic}', "");
Expect(1, 68680, '\P{^Is_Script:    -old_turkic}', "");
Expect(0, 68681, '\p{Is_Script:    -old_turkic}', "");
Expect(1, 68681, '\p{^Is_Script:    -old_turkic}', "");
Expect(1, 68681, '\P{Is_Script:    -old_turkic}', "");
Expect(0, 68681, '\P{^Is_Script:    -old_turkic}', "");
Error('\p{Is_Sc=/a/-	Orkh}');
Error('\P{Is_Sc=/a/-	Orkh}');
Expect(1, 68680, '\p{Is_Sc=orkh}', "");
Expect(0, 68680, '\p{^Is_Sc=orkh}', "");
Expect(0, 68680, '\P{Is_Sc=orkh}', "");
Expect(1, 68680, '\P{^Is_Sc=orkh}', "");
Expect(0, 68681, '\p{Is_Sc=orkh}', "");
Expect(1, 68681, '\p{^Is_Sc=orkh}', "");
Expect(1, 68681, '\P{Is_Sc=orkh}', "");
Expect(0, 68681, '\P{^Is_Sc=orkh}', "");
Expect(1, 68680, '\p{Is_Sc=	Orkh}', "");
Expect(0, 68680, '\p{^Is_Sc=	Orkh}', "");
Expect(0, 68680, '\P{Is_Sc=	Orkh}', "");
Expect(1, 68680, '\P{^Is_Sc=	Orkh}', "");
Expect(0, 68681, '\p{Is_Sc=	Orkh}', "");
Expect(1, 68681, '\p{^Is_Sc=	Orkh}', "");
Expect(1, 68681, '\P{Is_Sc=	Orkh}', "");
Expect(0, 68681, '\P{^Is_Sc=	Orkh}', "");
Error('\p{Script=-	Oriya/a/}');
Error('\P{Script=-	Oriya/a/}');
Expect(1, 2935, '\p{Script=oriya}', "");
Expect(0, 2935, '\p{^Script=oriya}', "");
Expect(0, 2935, '\P{Script=oriya}', "");
Expect(1, 2935, '\P{^Script=oriya}', "");
Expect(0, 2936, '\p{Script=oriya}', "");
Expect(1, 2936, '\p{^Script=oriya}', "");
Expect(1, 2936, '\P{Script=oriya}', "");
Expect(0, 2936, '\P{^Script=oriya}', "");
Expect(1, 2935, '\p{Script=-_Oriya}', "");
Expect(0, 2935, '\p{^Script=-_Oriya}', "");
Expect(0, 2935, '\P{Script=-_Oriya}', "");
Expect(1, 2935, '\P{^Script=-_Oriya}', "");
Expect(0, 2936, '\p{Script=-_Oriya}', "");
Expect(1, 2936, '\p{^Script=-_Oriya}', "");
Expect(1, 2936, '\P{Script=-_Oriya}', "");
Expect(0, 2936, '\P{^Script=-_Oriya}', "");
Error('\p{Sc=:= _Orya}');
Error('\P{Sc=:= _Orya}');
Expect(1, 2935, '\p{Sc=orya}', "");
Expect(0, 2935, '\p{^Sc=orya}', "");
Expect(0, 2935, '\P{Sc=orya}', "");
Expect(1, 2935, '\P{^Sc=orya}', "");
Expect(0, 2936, '\p{Sc=orya}', "");
Expect(1, 2936, '\p{^Sc=orya}', "");
Expect(1, 2936, '\P{Sc=orya}', "");
Expect(0, 2936, '\P{^Sc=orya}', "");
Expect(1, 2935, '\p{Sc=_ORYA}', "");
Expect(0, 2935, '\p{^Sc=_ORYA}', "");
Expect(0, 2935, '\P{Sc=_ORYA}', "");
Expect(1, 2935, '\P{^Sc=_ORYA}', "");
Expect(0, 2936, '\p{Sc=_ORYA}', "");
Expect(1, 2936, '\p{^Sc=_ORYA}', "");
Expect(1, 2936, '\P{Sc=_ORYA}', "");
Expect(0, 2936, '\P{^Sc=_ORYA}', "");
Error('\p{Is_Script=_:=ORIYA}');
Error('\P{Is_Script=_:=ORIYA}');
Expect(1, 2935, '\p{Is_Script=oriya}', "");
Expect(0, 2935, '\p{^Is_Script=oriya}', "");
Expect(0, 2935, '\P{Is_Script=oriya}', "");
Expect(1, 2935, '\P{^Is_Script=oriya}', "");
Expect(0, 2936, '\p{Is_Script=oriya}', "");
Expect(1, 2936, '\p{^Is_Script=oriya}', "");
Expect(1, 2936, '\P{Is_Script=oriya}', "");
Expect(0, 2936, '\P{^Is_Script=oriya}', "");
Expect(1, 2935, '\p{Is_Script=-Oriya}', "");
Expect(0, 2935, '\p{^Is_Script=-Oriya}', "");
Expect(0, 2935, '\P{Is_Script=-Oriya}', "");
Expect(1, 2935, '\P{^Is_Script=-Oriya}', "");
Expect(0, 2936, '\p{Is_Script=-Oriya}', "");
Expect(1, 2936, '\p{^Is_Script=-Oriya}', "");
Expect(1, 2936, '\P{Is_Script=-Oriya}', "");
Expect(0, 2936, '\P{^Is_Script=-Oriya}', "");
Error('\p{Is_Sc=:=orya}');
Error('\P{Is_Sc=:=orya}');
Expect(1, 2935, '\p{Is_Sc=orya}', "");
Expect(0, 2935, '\p{^Is_Sc=orya}', "");
Expect(0, 2935, '\P{Is_Sc=orya}', "");
Expect(1, 2935, '\P{^Is_Sc=orya}', "");
Expect(0, 2936, '\p{Is_Sc=orya}', "");
Expect(1, 2936, '\p{^Is_Sc=orya}', "");
Expect(1, 2936, '\P{Is_Sc=orya}', "");
Expect(0, 2936, '\P{^Is_Sc=orya}', "");
Expect(1, 2935, '\p{Is_Sc:	 _orya}', "");
Expect(0, 2935, '\p{^Is_Sc:	 _orya}', "");
Expect(0, 2935, '\P{Is_Sc:	 _orya}', "");
Expect(1, 2935, '\P{^Is_Sc:	 _orya}', "");
Expect(0, 2936, '\p{Is_Sc:	 _orya}', "");
Expect(1, 2936, '\p{^Is_Sc:	 _orya}', "");
Expect(1, 2936, '\P{Is_Sc:	 _orya}', "");
Expect(0, 2936, '\P{^Is_Sc:	 _orya}', "");
Error('\p{Script=:=	Osmanya}');
Error('\P{Script=:=	Osmanya}');
Expect(1, 66729, '\p{Script=osmanya}', "");
Expect(0, 66729, '\p{^Script=osmanya}', "");
Expect(0, 66729, '\P{Script=osmanya}', "");
Expect(1, 66729, '\P{^Script=osmanya}', "");
Expect(0, 66730, '\p{Script=osmanya}', "");
Expect(1, 66730, '\p{^Script=osmanya}', "");
Expect(1, 66730, '\P{Script=osmanya}', "");
Expect(0, 66730, '\P{^Script=osmanya}', "");
Expect(1, 66729, '\p{Script=	OSMANYA}', "");
Expect(0, 66729, '\p{^Script=	OSMANYA}', "");
Expect(0, 66729, '\P{Script=	OSMANYA}', "");
Expect(1, 66729, '\P{^Script=	OSMANYA}', "");
Expect(0, 66730, '\p{Script=	OSMANYA}', "");
Expect(1, 66730, '\p{^Script=	OSMANYA}', "");
Expect(1, 66730, '\P{Script=	OSMANYA}', "");
Expect(0, 66730, '\P{^Script=	OSMANYA}', "");
Error('\p{Sc=-:=Osma}');
Error('\P{Sc=-:=Osma}');
Expect(1, 66729, '\p{Sc=osma}', "");
Expect(0, 66729, '\p{^Sc=osma}', "");
Expect(0, 66729, '\P{Sc=osma}', "");
Expect(1, 66729, '\P{^Sc=osma}', "");
Expect(0, 66730, '\p{Sc=osma}', "");
Expect(1, 66730, '\p{^Sc=osma}', "");
Expect(1, 66730, '\P{Sc=osma}', "");
Expect(0, 66730, '\P{^Sc=osma}', "");
Expect(1, 66729, '\p{Sc= OSMA}', "");
Expect(0, 66729, '\p{^Sc= OSMA}', "");
Expect(0, 66729, '\P{Sc= OSMA}', "");
Expect(1, 66729, '\P{^Sc= OSMA}', "");
Expect(0, 66730, '\p{Sc= OSMA}', "");
Expect(1, 66730, '\p{^Sc= OSMA}', "");
Expect(1, 66730, '\P{Sc= OSMA}', "");
Expect(0, 66730, '\P{^Sc= OSMA}', "");
Error('\p{Is_Script:-:=osmanya}');
Error('\P{Is_Script:-:=osmanya}');
Expect(1, 66729, '\p{Is_Script=osmanya}', "");
Expect(0, 66729, '\p{^Is_Script=osmanya}', "");
Expect(0, 66729, '\P{Is_Script=osmanya}', "");
Expect(1, 66729, '\P{^Is_Script=osmanya}', "");
Expect(0, 66730, '\p{Is_Script=osmanya}', "");
Expect(1, 66730, '\p{^Is_Script=osmanya}', "");
Expect(1, 66730, '\P{Is_Script=osmanya}', "");
Expect(0, 66730, '\P{^Is_Script=osmanya}', "");
Expect(1, 66729, '\p{Is_Script:	Osmanya}', "");
Expect(0, 66729, '\p{^Is_Script:	Osmanya}', "");
Expect(0, 66729, '\P{Is_Script:	Osmanya}', "");
Expect(1, 66729, '\P{^Is_Script:	Osmanya}', "");
Expect(0, 66730, '\p{Is_Script:	Osmanya}', "");
Expect(1, 66730, '\p{^Is_Script:	Osmanya}', "");
Expect(1, 66730, '\P{Is_Script:	Osmanya}', "");
Expect(0, 66730, '\P{^Is_Script:	Osmanya}', "");
Error('\p{Is_Sc:   	/a/Osma}');
Error('\P{Is_Sc:   	/a/Osma}');
Expect(1, 66729, '\p{Is_Sc:	osma}', "");
Expect(0, 66729, '\p{^Is_Sc:	osma}', "");
Expect(0, 66729, '\P{Is_Sc:	osma}', "");
Expect(1, 66729, '\P{^Is_Sc:	osma}', "");
Expect(0, 66730, '\p{Is_Sc:	osma}', "");
Expect(1, 66730, '\p{^Is_Sc:	osma}', "");
Expect(1, 66730, '\P{Is_Sc:	osma}', "");
Expect(0, 66730, '\P{^Is_Sc:	osma}', "");
Expect(1, 66729, '\p{Is_Sc=	_Osma}', "");
Expect(0, 66729, '\p{^Is_Sc=	_Osma}', "");
Expect(0, 66729, '\P{Is_Sc=	_Osma}', "");
Expect(1, 66729, '\P{^Is_Sc=	_Osma}', "");
Expect(0, 66730, '\p{Is_Sc=	_Osma}', "");
Expect(1, 66730, '\p{^Is_Sc=	_Osma}', "");
Expect(1, 66730, '\P{Is_Sc=	_Osma}', "");
Expect(0, 66730, '\P{^Is_Sc=	_Osma}', "");
Error('\p{Script:_PHAGS_pa/a/}');
Error('\P{Script:_PHAGS_pa/a/}');
Expect(1, 43127, '\p{Script=phagspa}', "");
Expect(0, 43127, '\p{^Script=phagspa}', "");
Expect(0, 43127, '\P{Script=phagspa}', "");
Expect(1, 43127, '\P{^Script=phagspa}', "");
Expect(0, 43128, '\p{Script=phagspa}', "");
Expect(1, 43128, '\p{^Script=phagspa}', "");
Expect(1, 43128, '\P{Script=phagspa}', "");
Expect(0, 43128, '\P{^Script=phagspa}', "");
Expect(1, 43127, '\p{Script=_ Phags_pa}', "");
Expect(0, 43127, '\p{^Script=_ Phags_pa}', "");
Expect(0, 43127, '\P{Script=_ Phags_pa}', "");
Expect(1, 43127, '\P{^Script=_ Phags_pa}', "");
Expect(0, 43128, '\p{Script=_ Phags_pa}', "");
Expect(1, 43128, '\p{^Script=_ Phags_pa}', "");
Expect(1, 43128, '\P{Script=_ Phags_pa}', "");
Expect(0, 43128, '\P{^Script=_ Phags_pa}', "");
Error('\p{Sc=/a/Phag}');
Error('\P{Sc=/a/Phag}');
Expect(1, 43127, '\p{Sc=phag}', "");
Expect(0, 43127, '\p{^Sc=phag}', "");
Expect(0, 43127, '\P{Sc=phag}', "");
Expect(1, 43127, '\P{^Sc=phag}', "");
Expect(0, 43128, '\p{Sc=phag}', "");
Expect(1, 43128, '\p{^Sc=phag}', "");
Expect(1, 43128, '\P{Sc=phag}', "");
Expect(0, 43128, '\P{^Sc=phag}', "");
Expect(1, 43127, '\p{Sc= _Phag}', "");
Expect(0, 43127, '\p{^Sc= _Phag}', "");
Expect(0, 43127, '\P{Sc= _Phag}', "");
Expect(1, 43127, '\P{^Sc= _Phag}', "");
Expect(0, 43128, '\p{Sc= _Phag}', "");
Expect(1, 43128, '\p{^Sc= _Phag}', "");
Expect(1, 43128, '\P{Sc= _Phag}', "");
Expect(0, 43128, '\P{^Sc= _Phag}', "");
Error('\p{Is_Script=	-Phags_Pa/a/}');
Error('\P{Is_Script=	-Phags_Pa/a/}');
Expect(1, 43127, '\p{Is_Script=phagspa}', "");
Expect(0, 43127, '\p{^Is_Script=phagspa}', "");
Expect(0, 43127, '\P{Is_Script=phagspa}', "");
Expect(1, 43127, '\P{^Is_Script=phagspa}', "");
Expect(0, 43128, '\p{Is_Script=phagspa}', "");
Expect(1, 43128, '\p{^Is_Script=phagspa}', "");
Expect(1, 43128, '\P{Is_Script=phagspa}', "");
Expect(0, 43128, '\P{^Is_Script=phagspa}', "");
Expect(1, 43127, '\p{Is_Script:			Phags_Pa}', "");
Expect(0, 43127, '\p{^Is_Script:			Phags_Pa}', "");
Expect(0, 43127, '\P{Is_Script:			Phags_Pa}', "");
Expect(1, 43127, '\P{^Is_Script:			Phags_Pa}', "");
Expect(0, 43128, '\p{Is_Script:			Phags_Pa}', "");
Expect(1, 43128, '\p{^Is_Script:			Phags_Pa}', "");
Expect(1, 43128, '\P{Is_Script:			Phags_Pa}', "");
Expect(0, 43128, '\P{^Is_Script:			Phags_Pa}', "");
Error('\p{Is_Sc=-	Phag:=}');
Error('\P{Is_Sc=-	Phag:=}');
Expect(1, 43127, '\p{Is_Sc=phag}', "");
Expect(0, 43127, '\p{^Is_Sc=phag}', "");
Expect(0, 43127, '\P{Is_Sc=phag}', "");
Expect(1, 43127, '\P{^Is_Sc=phag}', "");
Expect(0, 43128, '\p{Is_Sc=phag}', "");
Expect(1, 43128, '\p{^Is_Sc=phag}', "");
Expect(1, 43128, '\P{Is_Sc=phag}', "");
Expect(0, 43128, '\P{^Is_Sc=phag}', "");
Expect(1, 43127, '\p{Is_Sc=- PHAG}', "");
Expect(0, 43127, '\p{^Is_Sc=- PHAG}', "");
Expect(0, 43127, '\P{Is_Sc=- PHAG}', "");
Expect(1, 43127, '\P{^Is_Sc=- PHAG}', "");
Expect(0, 43128, '\p{Is_Sc=- PHAG}', "");
Expect(1, 43128, '\p{^Is_Sc=- PHAG}', "");
Expect(1, 43128, '\P{Is_Sc=- PHAG}', "");
Expect(0, 43128, '\P{^Is_Sc=- PHAG}', "");
Error('\p{Script:   :=_-INSCRIPTIONAL_pahlavi}');
Error('\P{Script:   :=_-INSCRIPTIONAL_pahlavi}');
Expect(1, 68479, '\p{Script=inscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^Script=inscriptionalpahlavi}', "");
Expect(0, 68479, '\P{Script=inscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^Script=inscriptionalpahlavi}', "");
Expect(0, 68480, '\p{Script=inscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^Script=inscriptionalpahlavi}', "");
Expect(1, 68480, '\P{Script=inscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^Script=inscriptionalpahlavi}', "");
Expect(1, 68479, '\p{Script=		Inscriptional_PAHLAVI}', "");
Expect(0, 68479, '\p{^Script=		Inscriptional_PAHLAVI}', "");
Expect(0, 68479, '\P{Script=		Inscriptional_PAHLAVI}', "");
Expect(1, 68479, '\P{^Script=		Inscriptional_PAHLAVI}', "");
Expect(0, 68480, '\p{Script=		Inscriptional_PAHLAVI}', "");
Expect(1, 68480, '\p{^Script=		Inscriptional_PAHLAVI}', "");
Expect(1, 68480, '\P{Script=		Inscriptional_PAHLAVI}', "");
Expect(0, 68480, '\P{^Script=		Inscriptional_PAHLAVI}', "");
Error('\p{Sc=	/a/Phli}');
Error('\P{Sc=	/a/Phli}');
Expect(1, 68479, '\p{Sc=phli}', "");
Expect(0, 68479, '\p{^Sc=phli}', "");
Expect(0, 68479, '\P{Sc=phli}', "");
Expect(1, 68479, '\P{^Sc=phli}', "");
Expect(0, 68480, '\p{Sc=phli}', "");
Expect(1, 68480, '\p{^Sc=phli}', "");
Expect(1, 68480, '\P{Sc=phli}', "");
Expect(0, 68480, '\P{^Sc=phli}', "");
Expect(1, 68479, '\p{Sc=__Phli}', "");
Expect(0, 68479, '\p{^Sc=__Phli}', "");
Expect(0, 68479, '\P{Sc=__Phli}', "");
Expect(1, 68479, '\P{^Sc=__Phli}', "");
Expect(0, 68480, '\p{Sc=__Phli}', "");
Expect(1, 68480, '\p{^Sc=__Phli}', "");
Expect(1, 68480, '\P{Sc=__Phli}', "");
Expect(0, 68480, '\P{^Sc=__Phli}', "");
Error('\p{Is_Script: /a/	_Inscriptional_PAHLAVI}');
Error('\P{Is_Script: /a/	_Inscriptional_PAHLAVI}');
Expect(1, 68479, '\p{Is_Script=inscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^Is_Script=inscriptionalpahlavi}', "");
Expect(0, 68479, '\P{Is_Script=inscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^Is_Script=inscriptionalpahlavi}', "");
Expect(0, 68480, '\p{Is_Script=inscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^Is_Script=inscriptionalpahlavi}', "");
Expect(1, 68480, '\P{Is_Script=inscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^Is_Script=inscriptionalpahlavi}', "");
Expect(1, 68479, '\p{Is_Script= inscriptional_Pahlavi}', "");
Expect(0, 68479, '\p{^Is_Script= inscriptional_Pahlavi}', "");
Expect(0, 68479, '\P{Is_Script= inscriptional_Pahlavi}', "");
Expect(1, 68479, '\P{^Is_Script= inscriptional_Pahlavi}', "");
Expect(0, 68480, '\p{Is_Script= inscriptional_Pahlavi}', "");
Expect(1, 68480, '\p{^Is_Script= inscriptional_Pahlavi}', "");
Expect(1, 68480, '\P{Is_Script= inscriptional_Pahlavi}', "");
Expect(0, 68480, '\P{^Is_Script= inscriptional_Pahlavi}', "");
Error('\p{Is_Sc=- phli/a/}');
Error('\P{Is_Sc=- phli/a/}');
Expect(1, 68479, '\p{Is_Sc=phli}', "");
Expect(0, 68479, '\p{^Is_Sc=phli}', "");
Expect(0, 68479, '\P{Is_Sc=phli}', "");
Expect(1, 68479, '\P{^Is_Sc=phli}', "");
Expect(0, 68480, '\p{Is_Sc=phli}', "");
Expect(1, 68480, '\p{^Is_Sc=phli}', "");
Expect(1, 68480, '\P{Is_Sc=phli}', "");
Expect(0, 68480, '\P{^Is_Sc=phli}', "");
Expect(1, 68479, '\p{Is_Sc=-phli}', "");
Expect(0, 68479, '\p{^Is_Sc=-phli}', "");
Expect(0, 68479, '\P{Is_Sc=-phli}', "");
Expect(1, 68479, '\P{^Is_Sc=-phli}', "");
Expect(0, 68480, '\p{Is_Sc=-phli}', "");
Expect(1, 68480, '\p{^Is_Sc=-phli}', "");
Expect(1, 68480, '\P{Is_Sc=-phli}', "");
Expect(0, 68480, '\P{^Is_Sc=-phli}', "");
Error('\p{Script=:= -Phoenician}');
Error('\P{Script=:= -Phoenician}');
Expect(1, 67871, '\p{Script:   phoenician}', "");
Expect(0, 67871, '\p{^Script:   phoenician}', "");
Expect(0, 67871, '\P{Script:   phoenician}', "");
Expect(1, 67871, '\P{^Script:   phoenician}', "");
Expect(0, 67872, '\p{Script:   phoenician}', "");
Expect(1, 67872, '\p{^Script:   phoenician}', "");
Expect(1, 67872, '\P{Script:   phoenician}', "");
Expect(0, 67872, '\P{^Script:   phoenician}', "");
Expect(1, 67871, '\p{Script= Phoenician}', "");
Expect(0, 67871, '\p{^Script= Phoenician}', "");
Expect(0, 67871, '\P{Script= Phoenician}', "");
Expect(1, 67871, '\P{^Script= Phoenician}', "");
Expect(0, 67872, '\p{Script= Phoenician}', "");
Expect(1, 67872, '\p{^Script= Phoenician}', "");
Expect(1, 67872, '\P{Script= Phoenician}', "");
Expect(0, 67872, '\P{^Script= Phoenician}', "");
Error('\p{Sc= :=PHNX}');
Error('\P{Sc= :=PHNX}');
Expect(1, 67871, '\p{Sc=phnx}', "");
Expect(0, 67871, '\p{^Sc=phnx}', "");
Expect(0, 67871, '\P{Sc=phnx}', "");
Expect(1, 67871, '\P{^Sc=phnx}', "");
Expect(0, 67872, '\p{Sc=phnx}', "");
Expect(1, 67872, '\p{^Sc=phnx}', "");
Expect(1, 67872, '\P{Sc=phnx}', "");
Expect(0, 67872, '\P{^Sc=phnx}', "");
Expect(1, 67871, '\p{Sc=PHNX}', "");
Expect(0, 67871, '\p{^Sc=PHNX}', "");
Expect(0, 67871, '\P{Sc=PHNX}', "");
Expect(1, 67871, '\P{^Sc=PHNX}', "");
Expect(0, 67872, '\p{Sc=PHNX}', "");
Expect(1, 67872, '\p{^Sc=PHNX}', "");
Expect(1, 67872, '\P{Sc=PHNX}', "");
Expect(0, 67872, '\P{^Sc=PHNX}', "");
Error('\p{Is_Script=:= PHOENICIAN}');
Error('\P{Is_Script=:= PHOENICIAN}');
Expect(1, 67871, '\p{Is_Script=phoenician}', "");
Expect(0, 67871, '\p{^Is_Script=phoenician}', "");
Expect(0, 67871, '\P{Is_Script=phoenician}', "");
Expect(1, 67871, '\P{^Is_Script=phoenician}', "");
Expect(0, 67872, '\p{Is_Script=phoenician}', "");
Expect(1, 67872, '\p{^Is_Script=phoenician}', "");
Expect(1, 67872, '\P{Is_Script=phoenician}', "");
Expect(0, 67872, '\P{^Is_Script=phoenician}', "");
Expect(1, 67871, '\p{Is_Script=	Phoenician}', "");
Expect(0, 67871, '\p{^Is_Script=	Phoenician}', "");
Expect(0, 67871, '\P{Is_Script=	Phoenician}', "");
Expect(1, 67871, '\P{^Is_Script=	Phoenician}', "");
Expect(0, 67872, '\p{Is_Script=	Phoenician}', "");
Expect(1, 67872, '\p{^Is_Script=	Phoenician}', "");
Expect(1, 67872, '\P{Is_Script=	Phoenician}', "");
Expect(0, 67872, '\P{^Is_Script=	Phoenician}', "");
Error('\p{Is_Sc=	_Phnx:=}');
Error('\P{Is_Sc=	_Phnx:=}');
Expect(1, 67871, '\p{Is_Sc=phnx}', "");
Expect(0, 67871, '\p{^Is_Sc=phnx}', "");
Expect(0, 67871, '\P{Is_Sc=phnx}', "");
Expect(1, 67871, '\P{^Is_Sc=phnx}', "");
Expect(0, 67872, '\p{Is_Sc=phnx}', "");
Expect(1, 67872, '\p{^Is_Sc=phnx}', "");
Expect(1, 67872, '\P{Is_Sc=phnx}', "");
Expect(0, 67872, '\P{^Is_Sc=phnx}', "");
Expect(1, 67871, '\p{Is_Sc=	 Phnx}', "");
Expect(0, 67871, '\p{^Is_Sc=	 Phnx}', "");
Expect(0, 67871, '\P{Is_Sc=	 Phnx}', "");
Expect(1, 67871, '\P{^Is_Sc=	 Phnx}', "");
Expect(0, 67872, '\p{Is_Sc=	 Phnx}', "");
Expect(1, 67872, '\p{^Is_Sc=	 Phnx}', "");
Expect(1, 67872, '\P{Is_Sc=	 Phnx}', "");
Expect(0, 67872, '\P{^Is_Sc=	 Phnx}', "");
Error('\p{Script=	/a/MIAO}');
Error('\P{Script=	/a/MIAO}');
Expect(1, 94111, '\p{Script=miao}', "");
Expect(0, 94111, '\p{^Script=miao}', "");
Expect(0, 94111, '\P{Script=miao}', "");
Expect(1, 94111, '\P{^Script=miao}', "");
Expect(0, 94112, '\p{Script=miao}', "");
Expect(1, 94112, '\p{^Script=miao}', "");
Expect(1, 94112, '\P{Script=miao}', "");
Expect(0, 94112, '\P{^Script=miao}', "");
Expect(1, 94111, '\p{Script=	_Miao}', "");
Expect(0, 94111, '\p{^Script=	_Miao}', "");
Expect(0, 94111, '\P{Script=	_Miao}', "");
Expect(1, 94111, '\P{^Script=	_Miao}', "");
Expect(0, 94112, '\p{Script=	_Miao}', "");
Expect(1, 94112, '\p{^Script=	_Miao}', "");
Expect(1, 94112, '\P{Script=	_Miao}', "");
Expect(0, 94112, '\P{^Script=	_Miao}', "");
Error('\p{Sc=/a/-Plrd}');
Error('\P{Sc=/a/-Plrd}');
Expect(1, 94111, '\p{Sc=plrd}', "");
Expect(0, 94111, '\p{^Sc=plrd}', "");
Expect(0, 94111, '\P{Sc=plrd}', "");
Expect(1, 94111, '\P{^Sc=plrd}', "");
Expect(0, 94112, '\p{Sc=plrd}', "");
Expect(1, 94112, '\p{^Sc=plrd}', "");
Expect(1, 94112, '\P{Sc=plrd}', "");
Expect(0, 94112, '\P{^Sc=plrd}', "");
Expect(1, 94111, '\p{Sc=_ PLRD}', "");
Expect(0, 94111, '\p{^Sc=_ PLRD}', "");
Expect(0, 94111, '\P{Sc=_ PLRD}', "");
Expect(1, 94111, '\P{^Sc=_ PLRD}', "");
Expect(0, 94112, '\p{Sc=_ PLRD}', "");
Expect(1, 94112, '\p{^Sc=_ PLRD}', "");
Expect(1, 94112, '\P{Sc=_ PLRD}', "");
Expect(0, 94112, '\P{^Sc=_ PLRD}', "");
Error('\p{Is_Script=-:=Miao}');
Error('\P{Is_Script=-:=Miao}');
Expect(1, 94111, '\p{Is_Script=miao}', "");
Expect(0, 94111, '\p{^Is_Script=miao}', "");
Expect(0, 94111, '\P{Is_Script=miao}', "");
Expect(1, 94111, '\P{^Is_Script=miao}', "");
Expect(0, 94112, '\p{Is_Script=miao}', "");
Expect(1, 94112, '\p{^Is_Script=miao}', "");
Expect(1, 94112, '\P{Is_Script=miao}', "");
Expect(0, 94112, '\P{^Is_Script=miao}', "");
Expect(1, 94111, '\p{Is_Script=		MIAO}', "");
Expect(0, 94111, '\p{^Is_Script=		MIAO}', "");
Expect(0, 94111, '\P{Is_Script=		MIAO}', "");
Expect(1, 94111, '\P{^Is_Script=		MIAO}', "");
Expect(0, 94112, '\p{Is_Script=		MIAO}', "");
Expect(1, 94112, '\p{^Is_Script=		MIAO}', "");
Expect(1, 94112, '\P{Is_Script=		MIAO}', "");
Expect(0, 94112, '\P{^Is_Script=		MIAO}', "");
Error('\p{Is_Sc=_/a/Plrd}');
Error('\P{Is_Sc=_/a/Plrd}');
Expect(1, 94111, '\p{Is_Sc:plrd}', "");
Expect(0, 94111, '\p{^Is_Sc:plrd}', "");
Expect(0, 94111, '\P{Is_Sc:plrd}', "");
Expect(1, 94111, '\P{^Is_Sc:plrd}', "");
Expect(0, 94112, '\p{Is_Sc:plrd}', "");
Expect(1, 94112, '\p{^Is_Sc:plrd}', "");
Expect(1, 94112, '\P{Is_Sc:plrd}', "");
Expect(0, 94112, '\P{^Is_Sc:plrd}', "");
Expect(1, 94111, '\p{Is_Sc=-Plrd}', "");
Expect(0, 94111, '\p{^Is_Sc=-Plrd}', "");
Expect(0, 94111, '\P{Is_Sc=-Plrd}', "");
Expect(1, 94111, '\P{^Is_Sc=-Plrd}', "");
Expect(0, 94112, '\p{Is_Sc=-Plrd}', "");
Expect(1, 94112, '\p{^Is_Sc=-Plrd}', "");
Expect(1, 94112, '\P{Is_Sc=-Plrd}', "");
Expect(0, 94112, '\P{^Is_Sc=-Plrd}', "");
Error('\p{Script=	-Inscriptional_parthian/a/}');
Error('\P{Script=	-Inscriptional_parthian/a/}');
Expect(1, 68447, '\p{Script=inscriptionalparthian}', "");
Expect(0, 68447, '\p{^Script=inscriptionalparthian}', "");
Expect(0, 68447, '\P{Script=inscriptionalparthian}', "");
Expect(1, 68447, '\P{^Script=inscriptionalparthian}', "");
Expect(0, 68448, '\p{Script=inscriptionalparthian}', "");
Expect(1, 68448, '\p{^Script=inscriptionalparthian}', "");
Expect(1, 68448, '\P{Script=inscriptionalparthian}', "");
Expect(0, 68448, '\P{^Script=inscriptionalparthian}', "");
Expect(1, 68447, '\p{Script=	_inscriptional_Parthian}', "");
Expect(0, 68447, '\p{^Script=	_inscriptional_Parthian}', "");
Expect(0, 68447, '\P{Script=	_inscriptional_Parthian}', "");
Expect(1, 68447, '\P{^Script=	_inscriptional_Parthian}', "");
Expect(0, 68448, '\p{Script=	_inscriptional_Parthian}', "");
Expect(1, 68448, '\p{^Script=	_inscriptional_Parthian}', "");
Expect(1, 68448, '\P{Script=	_inscriptional_Parthian}', "");
Expect(0, 68448, '\P{^Script=	_inscriptional_Parthian}', "");
Error('\p{Sc=	:=Prti}');
Error('\P{Sc=	:=Prti}');
Expect(1, 68447, '\p{Sc=prti}', "");
Expect(0, 68447, '\p{^Sc=prti}', "");
Expect(0, 68447, '\P{Sc=prti}', "");
Expect(1, 68447, '\P{^Sc=prti}', "");
Expect(0, 68448, '\p{Sc=prti}', "");
Expect(1, 68448, '\p{^Sc=prti}', "");
Expect(1, 68448, '\P{Sc=prti}', "");
Expect(0, 68448, '\P{^Sc=prti}', "");
Expect(1, 68447, '\p{Sc=_prti}', "");
Expect(0, 68447, '\p{^Sc=_prti}', "");
Expect(0, 68447, '\P{Sc=_prti}', "");
Expect(1, 68447, '\P{^Sc=_prti}', "");
Expect(0, 68448, '\p{Sc=_prti}', "");
Expect(1, 68448, '\p{^Sc=_prti}', "");
Expect(1, 68448, '\P{Sc=_prti}', "");
Expect(0, 68448, '\P{^Sc=_prti}', "");
Error('\p{Is_Script=		Inscriptional_PARTHIAN:=}');
Error('\P{Is_Script=		Inscriptional_PARTHIAN:=}');
Expect(1, 68447, '\p{Is_Script: inscriptionalparthian}', "");
Expect(0, 68447, '\p{^Is_Script: inscriptionalparthian}', "");
Expect(0, 68447, '\P{Is_Script: inscriptionalparthian}', "");
Expect(1, 68447, '\P{^Is_Script: inscriptionalparthian}', "");
Expect(0, 68448, '\p{Is_Script: inscriptionalparthian}', "");
Expect(1, 68448, '\p{^Is_Script: inscriptionalparthian}', "");
Expect(1, 68448, '\P{Is_Script: inscriptionalparthian}', "");
Expect(0, 68448, '\P{^Is_Script: inscriptionalparthian}', "");
Expect(1, 68447, '\p{Is_Script=__Inscriptional_PARTHIAN}', "");
Expect(0, 68447, '\p{^Is_Script=__Inscriptional_PARTHIAN}', "");
Expect(0, 68447, '\P{Is_Script=__Inscriptional_PARTHIAN}', "");
Expect(1, 68447, '\P{^Is_Script=__Inscriptional_PARTHIAN}', "");
Expect(0, 68448, '\p{Is_Script=__Inscriptional_PARTHIAN}', "");
Expect(1, 68448, '\p{^Is_Script=__Inscriptional_PARTHIAN}', "");
Expect(1, 68448, '\P{Is_Script=__Inscriptional_PARTHIAN}', "");
Expect(0, 68448, '\P{^Is_Script=__Inscriptional_PARTHIAN}', "");
Error('\p{Is_Sc= prti:=}');
Error('\P{Is_Sc= prti:=}');
Expect(1, 68447, '\p{Is_Sc=prti}', "");
Expect(0, 68447, '\p{^Is_Sc=prti}', "");
Expect(0, 68447, '\P{Is_Sc=prti}', "");
Expect(1, 68447, '\P{^Is_Sc=prti}', "");
Expect(0, 68448, '\p{Is_Sc=prti}', "");
Expect(1, 68448, '\p{^Is_Sc=prti}', "");
Expect(1, 68448, '\P{Is_Sc=prti}', "");
Expect(0, 68448, '\P{^Is_Sc=prti}', "");
Expect(1, 68447, '\p{Is_Sc=_-Prti}', "");
Expect(0, 68447, '\p{^Is_Sc=_-Prti}', "");
Expect(0, 68447, '\P{Is_Sc=_-Prti}', "");
Expect(1, 68447, '\P{^Is_Sc=_-Prti}', "");
Expect(0, 68448, '\p{Is_Sc=_-Prti}', "");
Expect(1, 68448, '\p{^Is_Sc=_-Prti}', "");
Expect(1, 68448, '\P{Is_Sc=_-Prti}', "");
Expect(0, 68448, '\P{^Is_Sc=_-Prti}', "");
Error('\p{Script=_-Rejang:=}');
Error('\P{Script=_-Rejang:=}');
Expect(1, 43359, '\p{Script=rejang}', "");
Expect(0, 43359, '\p{^Script=rejang}', "");
Expect(0, 43359, '\P{Script=rejang}', "");
Expect(1, 43359, '\P{^Script=rejang}', "");
Expect(0, 43360, '\p{Script=rejang}', "");
Expect(1, 43360, '\p{^Script=rejang}', "");
Expect(1, 43360, '\P{Script=rejang}', "");
Expect(0, 43360, '\P{^Script=rejang}', "");
Expect(1, 43359, '\p{Script=  Rejang}', "");
Expect(0, 43359, '\p{^Script=  Rejang}', "");
Expect(0, 43359, '\P{Script=  Rejang}', "");
Expect(1, 43359, '\P{^Script=  Rejang}', "");
Expect(0, 43360, '\p{Script=  Rejang}', "");
Expect(1, 43360, '\p{^Script=  Rejang}', "");
Expect(1, 43360, '\P{Script=  Rejang}', "");
Expect(0, 43360, '\P{^Script=  Rejang}', "");
Error('\p{Sc= -rjng:=}');
Error('\P{Sc= -rjng:=}');
Expect(1, 43359, '\p{Sc=rjng}', "");
Expect(0, 43359, '\p{^Sc=rjng}', "");
Expect(0, 43359, '\P{Sc=rjng}', "");
Expect(1, 43359, '\P{^Sc=rjng}', "");
Expect(0, 43360, '\p{Sc=rjng}', "");
Expect(1, 43360, '\p{^Sc=rjng}', "");
Expect(1, 43360, '\P{Sc=rjng}', "");
Expect(0, 43360, '\P{^Sc=rjng}', "");
Expect(1, 43359, '\p{Sc=  Rjng}', "");
Expect(0, 43359, '\p{^Sc=  Rjng}', "");
Expect(0, 43359, '\P{Sc=  Rjng}', "");
Expect(1, 43359, '\P{^Sc=  Rjng}', "");
Expect(0, 43360, '\p{Sc=  Rjng}', "");
Expect(1, 43360, '\p{^Sc=  Rjng}', "");
Expect(1, 43360, '\P{Sc=  Rjng}', "");
Expect(0, 43360, '\P{^Sc=  Rjng}', "");
Error('\p{Is_Script=	Rejang/a/}');
Error('\P{Is_Script=	Rejang/a/}');
Expect(1, 43359, '\p{Is_Script=rejang}', "");
Expect(0, 43359, '\p{^Is_Script=rejang}', "");
Expect(0, 43359, '\P{Is_Script=rejang}', "");
Expect(1, 43359, '\P{^Is_Script=rejang}', "");
Expect(0, 43360, '\p{Is_Script=rejang}', "");
Expect(1, 43360, '\p{^Is_Script=rejang}', "");
Expect(1, 43360, '\P{Is_Script=rejang}', "");
Expect(0, 43360, '\P{^Is_Script=rejang}', "");
Expect(1, 43359, '\p{Is_Script=_rejang}', "");
Expect(0, 43359, '\p{^Is_Script=_rejang}', "");
Expect(0, 43359, '\P{Is_Script=_rejang}', "");
Expect(1, 43359, '\P{^Is_Script=_rejang}', "");
Expect(0, 43360, '\p{Is_Script=_rejang}', "");
Expect(1, 43360, '\p{^Is_Script=_rejang}', "");
Expect(1, 43360, '\P{Is_Script=_rejang}', "");
Expect(0, 43360, '\P{^Is_Script=_rejang}', "");
Error('\p{Is_Sc=/a/ 	rjng}');
Error('\P{Is_Sc=/a/ 	rjng}');
Expect(1, 43359, '\p{Is_Sc=rjng}', "");
Expect(0, 43359, '\p{^Is_Sc=rjng}', "");
Expect(0, 43359, '\P{Is_Sc=rjng}', "");
Expect(1, 43359, '\P{^Is_Sc=rjng}', "");
Expect(0, 43360, '\p{Is_Sc=rjng}', "");
Expect(1, 43360, '\p{^Is_Sc=rjng}', "");
Expect(1, 43360, '\P{Is_Sc=rjng}', "");
Expect(0, 43360, '\P{^Is_Sc=rjng}', "");
Expect(1, 43359, '\p{Is_Sc=	RJNG}', "");
Expect(0, 43359, '\p{^Is_Sc=	RJNG}', "");
Expect(0, 43359, '\P{Is_Sc=	RJNG}', "");
Expect(1, 43359, '\P{^Is_Sc=	RJNG}', "");
Expect(0, 43360, '\p{Is_Sc=	RJNG}', "");
Expect(1, 43360, '\p{^Is_Sc=	RJNG}', "");
Expect(1, 43360, '\P{Is_Sc=	RJNG}', "");
Expect(0, 43360, '\P{^Is_Sc=	RJNG}', "");
Error('\p{Script=:=Runic}');
Error('\P{Script=:=Runic}');
Expect(1, 5872, '\p{Script:   runic}', "");
Expect(0, 5872, '\p{^Script:   runic}', "");
Expect(0, 5872, '\P{Script:   runic}', "");
Expect(1, 5872, '\P{^Script:   runic}', "");
Expect(0, 5873, '\p{Script:   runic}', "");
Expect(1, 5873, '\p{^Script:   runic}', "");
Expect(1, 5873, '\P{Script:   runic}', "");
Expect(0, 5873, '\P{^Script:   runic}', "");
Expect(1, 5872, '\p{Script=_RUNIC}', "");
Expect(0, 5872, '\p{^Script=_RUNIC}', "");
Expect(0, 5872, '\P{Script=_RUNIC}', "");
Expect(1, 5872, '\P{^Script=_RUNIC}', "");
Expect(0, 5873, '\p{Script=_RUNIC}', "");
Expect(1, 5873, '\p{^Script=_RUNIC}', "");
Expect(1, 5873, '\P{Script=_RUNIC}', "");
Expect(0, 5873, '\P{^Script=_RUNIC}', "");
Error('\p{Sc=/a/ 	RUNR}');
Error('\P{Sc=/a/ 	RUNR}');
Expect(1, 5872, '\p{Sc=runr}', "");
Expect(0, 5872, '\p{^Sc=runr}', "");
Expect(0, 5872, '\P{Sc=runr}', "");
Expect(1, 5872, '\P{^Sc=runr}', "");
Expect(0, 5873, '\p{Sc=runr}', "");
Expect(1, 5873, '\p{^Sc=runr}', "");
Expect(1, 5873, '\P{Sc=runr}', "");
Expect(0, 5873, '\P{^Sc=runr}', "");
Expect(1, 5872, '\p{Sc=-Runr}', "");
Expect(0, 5872, '\p{^Sc=-Runr}', "");
Expect(0, 5872, '\P{Sc=-Runr}', "");
Expect(1, 5872, '\P{^Sc=-Runr}', "");
Expect(0, 5873, '\p{Sc=-Runr}', "");
Expect(1, 5873, '\p{^Sc=-Runr}', "");
Expect(1, 5873, '\P{Sc=-Runr}', "");
Expect(0, 5873, '\P{^Sc=-Runr}', "");
Error('\p{Is_Script=	/a/RUNIC}');
Error('\P{Is_Script=	/a/RUNIC}');
Expect(1, 5872, '\p{Is_Script=runic}', "");
Expect(0, 5872, '\p{^Is_Script=runic}', "");
Expect(0, 5872, '\P{Is_Script=runic}', "");
Expect(1, 5872, '\P{^Is_Script=runic}', "");
Expect(0, 5873, '\p{Is_Script=runic}', "");
Expect(1, 5873, '\p{^Is_Script=runic}', "");
Expect(1, 5873, '\P{Is_Script=runic}', "");
Expect(0, 5873, '\P{^Is_Script=runic}', "");
Expect(1, 5872, '\p{Is_Script=	-Runic}', "");
Expect(0, 5872, '\p{^Is_Script=	-Runic}', "");
Expect(0, 5872, '\P{Is_Script=	-Runic}', "");
Expect(1, 5872, '\P{^Is_Script=	-Runic}', "");
Expect(0, 5873, '\p{Is_Script=	-Runic}', "");
Expect(1, 5873, '\p{^Is_Script=	-Runic}', "");
Expect(1, 5873, '\P{Is_Script=	-Runic}', "");
Expect(0, 5873, '\P{^Is_Script=	-Runic}', "");
Error('\p{Is_Sc=:=Runr}');
Error('\P{Is_Sc=:=Runr}');
Expect(1, 5872, '\p{Is_Sc=runr}', "");
Expect(0, 5872, '\p{^Is_Sc=runr}', "");
Expect(0, 5872, '\P{Is_Sc=runr}', "");
Expect(1, 5872, '\P{^Is_Sc=runr}', "");
Expect(0, 5873, '\p{Is_Sc=runr}', "");
Expect(1, 5873, '\p{^Is_Sc=runr}', "");
Expect(1, 5873, '\P{Is_Sc=runr}', "");
Expect(0, 5873, '\P{^Is_Sc=runr}', "");
Expect(1, 5872, '\p{Is_Sc=- Runr}', "");
Expect(0, 5872, '\p{^Is_Sc=- Runr}', "");
Expect(0, 5872, '\P{Is_Sc=- Runr}', "");
Expect(1, 5872, '\P{^Is_Sc=- Runr}', "");
Expect(0, 5873, '\p{Is_Sc=- Runr}', "");
Expect(1, 5873, '\p{^Is_Sc=- Runr}', "");
Expect(1, 5873, '\P{Is_Sc=- Runr}', "");
Expect(0, 5873, '\P{^Is_Sc=- Runr}', "");
Error('\p{Script: /a/ _Samaritan}');
Error('\P{Script: /a/ _Samaritan}');
Expect(1, 2110, '\p{Script=samaritan}', "");
Expect(0, 2110, '\p{^Script=samaritan}', "");
Expect(0, 2110, '\P{Script=samaritan}', "");
Expect(1, 2110, '\P{^Script=samaritan}', "");
Expect(0, 2111, '\p{Script=samaritan}', "");
Expect(1, 2111, '\p{^Script=samaritan}', "");
Expect(1, 2111, '\P{Script=samaritan}', "");
Expect(0, 2111, '\P{^Script=samaritan}', "");
Expect(1, 2110, '\p{Script=_Samaritan}', "");
Expect(0, 2110, '\p{^Script=_Samaritan}', "");
Expect(0, 2110, '\P{Script=_Samaritan}', "");
Expect(1, 2110, '\P{^Script=_Samaritan}', "");
Expect(0, 2111, '\p{Script=_Samaritan}', "");
Expect(1, 2111, '\p{^Script=_Samaritan}', "");
Expect(1, 2111, '\P{Script=_Samaritan}', "");
Expect(0, 2111, '\P{^Script=_Samaritan}', "");
Error('\p{Sc=_samr/a/}');
Error('\P{Sc=_samr/a/}');
Expect(1, 2110, '\p{Sc=samr}', "");
Expect(0, 2110, '\p{^Sc=samr}', "");
Expect(0, 2110, '\P{Sc=samr}', "");
Expect(1, 2110, '\P{^Sc=samr}', "");
Expect(0, 2111, '\p{Sc=samr}', "");
Expect(1, 2111, '\p{^Sc=samr}', "");
Expect(1, 2111, '\P{Sc=samr}', "");
Expect(0, 2111, '\P{^Sc=samr}', "");
Expect(1, 2110, '\p{Sc=_	Samr}', "");
Expect(0, 2110, '\p{^Sc=_	Samr}', "");
Expect(0, 2110, '\P{Sc=_	Samr}', "");
Expect(1, 2110, '\P{^Sc=_	Samr}', "");
Expect(0, 2111, '\p{Sc=_	Samr}', "");
Expect(1, 2111, '\p{^Sc=_	Samr}', "");
Expect(1, 2111, '\P{Sc=_	Samr}', "");
Expect(0, 2111, '\P{^Sc=_	Samr}', "");
Error('\p{Is_Script=:=samaritan}');
Error('\P{Is_Script=:=samaritan}');
Expect(1, 2110, '\p{Is_Script=samaritan}', "");
Expect(0, 2110, '\p{^Is_Script=samaritan}', "");
Expect(0, 2110, '\P{Is_Script=samaritan}', "");
Expect(1, 2110, '\P{^Is_Script=samaritan}', "");
Expect(0, 2111, '\p{Is_Script=samaritan}', "");
Expect(1, 2111, '\p{^Is_Script=samaritan}', "");
Expect(1, 2111, '\P{Is_Script=samaritan}', "");
Expect(0, 2111, '\P{^Is_Script=samaritan}', "");
Expect(1, 2110, '\p{Is_Script=	 SAMARITAN}', "");
Expect(0, 2110, '\p{^Is_Script=	 SAMARITAN}', "");
Expect(0, 2110, '\P{Is_Script=	 SAMARITAN}', "");
Expect(1, 2110, '\P{^Is_Script=	 SAMARITAN}', "");
Expect(0, 2111, '\p{Is_Script=	 SAMARITAN}', "");
Expect(1, 2111, '\p{^Is_Script=	 SAMARITAN}', "");
Expect(1, 2111, '\P{Is_Script=	 SAMARITAN}', "");
Expect(0, 2111, '\P{^Is_Script=	 SAMARITAN}', "");
Error('\p{Is_Sc=-Samr:=}');
Error('\P{Is_Sc=-Samr:=}');
Expect(1, 2110, '\p{Is_Sc=samr}', "");
Expect(0, 2110, '\p{^Is_Sc=samr}', "");
Expect(0, 2110, '\P{Is_Sc=samr}', "");
Expect(1, 2110, '\P{^Is_Sc=samr}', "");
Expect(0, 2111, '\p{Is_Sc=samr}', "");
Expect(1, 2111, '\p{^Is_Sc=samr}', "");
Expect(1, 2111, '\P{Is_Sc=samr}', "");
Expect(0, 2111, '\P{^Is_Sc=samr}', "");
Expect(1, 2110, '\p{Is_Sc=__samr}', "");
Expect(0, 2110, '\p{^Is_Sc=__samr}', "");
Expect(0, 2110, '\P{Is_Sc=__samr}', "");
Expect(1, 2110, '\P{^Is_Sc=__samr}', "");
Expect(0, 2111, '\p{Is_Sc=__samr}', "");
Expect(1, 2111, '\p{^Is_Sc=__samr}', "");
Expect(1, 2111, '\P{Is_Sc=__samr}', "");
Expect(0, 2111, '\P{^Is_Sc=__samr}', "");
Error('\p{Script: /a/Old_South_ARABIAN}');
Error('\P{Script: /a/Old_South_ARABIAN}');
Expect(1, 68223, '\p{Script=oldsoutharabian}', "");
Expect(0, 68223, '\p{^Script=oldsoutharabian}', "");
Expect(0, 68223, '\P{Script=oldsoutharabian}', "");
Expect(1, 68223, '\P{^Script=oldsoutharabian}', "");
Expect(0, 68224, '\p{Script=oldsoutharabian}', "");
Expect(1, 68224, '\p{^Script=oldsoutharabian}', "");
Expect(1, 68224, '\P{Script=oldsoutharabian}', "");
Expect(0, 68224, '\P{^Script=oldsoutharabian}', "");
Expect(1, 68223, '\p{Script:	__old_SOUTH_ARABIAN}', "");
Expect(0, 68223, '\p{^Script:	__old_SOUTH_ARABIAN}', "");
Expect(0, 68223, '\P{Script:	__old_SOUTH_ARABIAN}', "");
Expect(1, 68223, '\P{^Script:	__old_SOUTH_ARABIAN}', "");
Expect(0, 68224, '\p{Script:	__old_SOUTH_ARABIAN}', "");
Expect(1, 68224, '\p{^Script:	__old_SOUTH_ARABIAN}', "");
Expect(1, 68224, '\P{Script:	__old_SOUTH_ARABIAN}', "");
Expect(0, 68224, '\P{^Script:	__old_SOUTH_ARABIAN}', "");
Error('\p{Sc=/a/_-Sarb}');
Error('\P{Sc=/a/_-Sarb}');
Expect(1, 68223, '\p{Sc=sarb}', "");
Expect(0, 68223, '\p{^Sc=sarb}', "");
Expect(0, 68223, '\P{Sc=sarb}', "");
Expect(1, 68223, '\P{^Sc=sarb}', "");
Expect(0, 68224, '\p{Sc=sarb}', "");
Expect(1, 68224, '\p{^Sc=sarb}', "");
Expect(1, 68224, '\P{Sc=sarb}', "");
Expect(0, 68224, '\P{^Sc=sarb}', "");
Expect(1, 68223, '\p{Sc=-SARB}', "");
Expect(0, 68223, '\p{^Sc=-SARB}', "");
Expect(0, 68223, '\P{Sc=-SARB}', "");
Expect(1, 68223, '\P{^Sc=-SARB}', "");
Expect(0, 68224, '\p{Sc=-SARB}', "");
Expect(1, 68224, '\p{^Sc=-SARB}', "");
Expect(1, 68224, '\P{Sc=-SARB}', "");
Expect(0, 68224, '\P{^Sc=-SARB}', "");
Error('\p{Is_Script=:= -Old_SOUTH_ARABIAN}');
Error('\P{Is_Script=:= -Old_SOUTH_ARABIAN}');
Expect(1, 68223, '\p{Is_Script=oldsoutharabian}', "");
Expect(0, 68223, '\p{^Is_Script=oldsoutharabian}', "");
Expect(0, 68223, '\P{Is_Script=oldsoutharabian}', "");
Expect(1, 68223, '\P{^Is_Script=oldsoutharabian}', "");
Expect(0, 68224, '\p{Is_Script=oldsoutharabian}', "");
Expect(1, 68224, '\p{^Is_Script=oldsoutharabian}', "");
Expect(1, 68224, '\P{Is_Script=oldsoutharabian}', "");
Expect(0, 68224, '\P{^Is_Script=oldsoutharabian}', "");
Expect(1, 68223, '\p{Is_Script=-OLD_South_ARABIAN}', "");
Expect(0, 68223, '\p{^Is_Script=-OLD_South_ARABIAN}', "");
Expect(0, 68223, '\P{Is_Script=-OLD_South_ARABIAN}', "");
Expect(1, 68223, '\P{^Is_Script=-OLD_South_ARABIAN}', "");
Expect(0, 68224, '\p{Is_Script=-OLD_South_ARABIAN}', "");
Expect(1, 68224, '\p{^Is_Script=-OLD_South_ARABIAN}', "");
Expect(1, 68224, '\P{Is_Script=-OLD_South_ARABIAN}', "");
Expect(0, 68224, '\P{^Is_Script=-OLD_South_ARABIAN}', "");
Error('\p{Is_Sc=/a/ -sarb}');
Error('\P{Is_Sc=/a/ -sarb}');
Expect(1, 68223, '\p{Is_Sc=sarb}', "");
Expect(0, 68223, '\p{^Is_Sc=sarb}', "");
Expect(0, 68223, '\P{Is_Sc=sarb}', "");
Expect(1, 68223, '\P{^Is_Sc=sarb}', "");
Expect(0, 68224, '\p{Is_Sc=sarb}', "");
Expect(1, 68224, '\p{^Is_Sc=sarb}', "");
Expect(1, 68224, '\P{Is_Sc=sarb}', "");
Expect(0, 68224, '\P{^Is_Sc=sarb}', "");
Expect(1, 68223, '\p{Is_Sc=--sarb}', "");
Expect(0, 68223, '\p{^Is_Sc=--sarb}', "");
Expect(0, 68223, '\P{Is_Sc=--sarb}', "");
Expect(1, 68223, '\P{^Is_Sc=--sarb}', "");
Expect(0, 68224, '\p{Is_Sc=--sarb}', "");
Expect(1, 68224, '\p{^Is_Sc=--sarb}', "");
Expect(1, 68224, '\P{Is_Sc=--sarb}', "");
Expect(0, 68224, '\P{^Is_Sc=--sarb}', "");
Error('\p{Script=_/a/Saurashtra}');
Error('\P{Script=_/a/Saurashtra}');
Expect(1, 43225, '\p{Script=saurashtra}', "");
Expect(0, 43225, '\p{^Script=saurashtra}', "");
Expect(0, 43225, '\P{Script=saurashtra}', "");
Expect(1, 43225, '\P{^Script=saurashtra}', "");
Expect(0, 43226, '\p{Script=saurashtra}', "");
Expect(1, 43226, '\p{^Script=saurashtra}', "");
Expect(1, 43226, '\P{Script=saurashtra}', "");
Expect(0, 43226, '\P{^Script=saurashtra}', "");
Expect(1, 43225, '\p{Script=- Saurashtra}', "");
Expect(0, 43225, '\p{^Script=- Saurashtra}', "");
Expect(0, 43225, '\P{Script=- Saurashtra}', "");
Expect(1, 43225, '\P{^Script=- Saurashtra}', "");
Expect(0, 43226, '\p{Script=- Saurashtra}', "");
Expect(1, 43226, '\p{^Script=- Saurashtra}', "");
Expect(1, 43226, '\P{Script=- Saurashtra}', "");
Expect(0, 43226, '\P{^Script=- Saurashtra}', "");
Error('\p{Sc=:=- Saur}');
Error('\P{Sc=:=- Saur}');
Expect(1, 43225, '\p{Sc=saur}', "");
Expect(0, 43225, '\p{^Sc=saur}', "");
Expect(0, 43225, '\P{Sc=saur}', "");
Expect(1, 43225, '\P{^Sc=saur}', "");
Expect(0, 43226, '\p{Sc=saur}', "");
Expect(1, 43226, '\p{^Sc=saur}', "");
Expect(1, 43226, '\P{Sc=saur}', "");
Expect(0, 43226, '\P{^Sc=saur}', "");
Expect(1, 43225, '\p{Sc=_saur}', "");
Expect(0, 43225, '\p{^Sc=_saur}', "");
Expect(0, 43225, '\P{Sc=_saur}', "");
Expect(1, 43225, '\P{^Sc=_saur}', "");
Expect(0, 43226, '\p{Sc=_saur}', "");
Expect(1, 43226, '\p{^Sc=_saur}', "");
Expect(1, 43226, '\P{Sc=_saur}', "");
Expect(0, 43226, '\P{^Sc=_saur}', "");
Error('\p{Is_Script=:=-	Saurashtra}');
Error('\P{Is_Script=:=-	Saurashtra}');
Expect(1, 43225, '\p{Is_Script=saurashtra}', "");
Expect(0, 43225, '\p{^Is_Script=saurashtra}', "");
Expect(0, 43225, '\P{Is_Script=saurashtra}', "");
Expect(1, 43225, '\P{^Is_Script=saurashtra}', "");
Expect(0, 43226, '\p{Is_Script=saurashtra}', "");
Expect(1, 43226, '\p{^Is_Script=saurashtra}', "");
Expect(1, 43226, '\P{Is_Script=saurashtra}', "");
Expect(0, 43226, '\P{^Is_Script=saurashtra}', "");
Error('\p{Is_Sc=-/a/Saur}');
Error('\P{Is_Sc=-/a/Saur}');
Expect(1, 43225, '\p{Is_Sc=saur}', "");
Expect(0, 43225, '\p{^Is_Sc=saur}', "");
Expect(0, 43225, '\P{Is_Sc=saur}', "");
Expect(1, 43225, '\P{^Is_Sc=saur}', "");
Expect(0, 43226, '\p{Is_Sc=saur}', "");
Expect(1, 43226, '\p{^Is_Sc=saur}', "");
Expect(1, 43226, '\P{Is_Sc=saur}', "");
Expect(0, 43226, '\P{^Is_Sc=saur}', "");
Expect(1, 43225, '\p{Is_Sc=  Saur}', "");
Expect(0, 43225, '\p{^Is_Sc=  Saur}', "");
Expect(0, 43225, '\P{Is_Sc=  Saur}', "");
Expect(1, 43225, '\P{^Is_Sc=  Saur}', "");
Expect(0, 43226, '\p{Is_Sc=  Saur}', "");
Expect(1, 43226, '\p{^Is_Sc=  Saur}', "");
Expect(1, 43226, '\P{Is_Sc=  Saur}', "");
Expect(0, 43226, '\P{^Is_Sc=  Saur}', "");
Error('\p{Script=/a/ Shavian}');
Error('\P{Script=/a/ Shavian}');
Expect(1, 66687, '\p{Script=shavian}', "");
Expect(0, 66687, '\p{^Script=shavian}', "");
Expect(0, 66687, '\P{Script=shavian}', "");
Expect(1, 66687, '\P{^Script=shavian}', "");
Expect(0, 66688, '\p{Script=shavian}', "");
Expect(1, 66688, '\p{^Script=shavian}', "");
Expect(1, 66688, '\P{Script=shavian}', "");
Expect(0, 66688, '\P{^Script=shavian}', "");
Error('\p{Sc= Shaw/a/}');
Error('\P{Sc= Shaw/a/}');
Expect(1, 66687, '\p{Sc=shaw}', "");
Expect(0, 66687, '\p{^Sc=shaw}', "");
Expect(0, 66687, '\P{Sc=shaw}', "");
Expect(1, 66687, '\P{^Sc=shaw}', "");
Expect(0, 66688, '\p{Sc=shaw}', "");
Expect(1, 66688, '\p{^Sc=shaw}', "");
Expect(1, 66688, '\P{Sc=shaw}', "");
Expect(0, 66688, '\P{^Sc=shaw}', "");
Expect(1, 66687, '\p{Sc=_-SHAW}', "");
Expect(0, 66687, '\p{^Sc=_-SHAW}', "");
Expect(0, 66687, '\P{Sc=_-SHAW}', "");
Expect(1, 66687, '\P{^Sc=_-SHAW}', "");
Expect(0, 66688, '\p{Sc=_-SHAW}', "");
Expect(1, 66688, '\p{^Sc=_-SHAW}', "");
Expect(1, 66688, '\P{Sc=_-SHAW}', "");
Expect(0, 66688, '\P{^Sc=_-SHAW}', "");
Error('\p{Is_Script=		shavian/a/}');
Error('\P{Is_Script=		shavian/a/}');
Expect(1, 66687, '\p{Is_Script=shavian}', "");
Expect(0, 66687, '\p{^Is_Script=shavian}', "");
Expect(0, 66687, '\P{Is_Script=shavian}', "");
Expect(1, 66687, '\P{^Is_Script=shavian}', "");
Expect(0, 66688, '\p{Is_Script=shavian}', "");
Expect(1, 66688, '\p{^Is_Script=shavian}', "");
Expect(1, 66688, '\P{Is_Script=shavian}', "");
Expect(0, 66688, '\P{^Is_Script=shavian}', "");
Expect(1, 66687, '\p{Is_Script= _Shavian}', "");
Expect(0, 66687, '\p{^Is_Script= _Shavian}', "");
Expect(0, 66687, '\P{Is_Script= _Shavian}', "");
Expect(1, 66687, '\P{^Is_Script= _Shavian}', "");
Expect(0, 66688, '\p{Is_Script= _Shavian}', "");
Expect(1, 66688, '\p{^Is_Script= _Shavian}', "");
Expect(1, 66688, '\P{Is_Script= _Shavian}', "");
Expect(0, 66688, '\P{^Is_Script= _Shavian}', "");
Error('\p{Is_Sc=:=Shaw}');
Error('\P{Is_Sc=:=Shaw}');
Expect(1, 66687, '\p{Is_Sc=shaw}', "");
Expect(0, 66687, '\p{^Is_Sc=shaw}', "");
Expect(0, 66687, '\P{Is_Sc=shaw}', "");
Expect(1, 66687, '\P{^Is_Sc=shaw}', "");
Expect(0, 66688, '\p{Is_Sc=shaw}', "");
Expect(1, 66688, '\p{^Is_Sc=shaw}', "");
Expect(1, 66688, '\P{Is_Sc=shaw}', "");
Expect(0, 66688, '\P{^Is_Sc=shaw}', "");
Expect(1, 66687, '\p{Is_Sc=--shaw}', "");
Expect(0, 66687, '\p{^Is_Sc=--shaw}', "");
Expect(0, 66687, '\P{Is_Sc=--shaw}', "");
Expect(1, 66687, '\P{^Is_Sc=--shaw}', "");
Expect(0, 66688, '\p{Is_Sc=--shaw}', "");
Expect(1, 66688, '\p{^Is_Sc=--shaw}', "");
Expect(1, 66688, '\P{Is_Sc=--shaw}', "");
Expect(0, 66688, '\P{^Is_Sc=--shaw}', "");
Error('\p{Script=_	Sharada:=}');
Error('\P{Script=_	Sharada:=}');
Expect(1, 70105, '\p{Script=sharada}', "");
Expect(0, 70105, '\p{^Script=sharada}', "");
Expect(0, 70105, '\P{Script=sharada}', "");
Expect(1, 70105, '\P{^Script=sharada}', "");
Expect(0, 70106, '\p{Script=sharada}', "");
Expect(1, 70106, '\p{^Script=sharada}', "");
Expect(1, 70106, '\P{Script=sharada}', "");
Expect(0, 70106, '\P{^Script=sharada}', "");
Expect(1, 70105, '\p{Script=_-SHARADA}', "");
Expect(0, 70105, '\p{^Script=_-SHARADA}', "");
Expect(0, 70105, '\P{Script=_-SHARADA}', "");
Expect(1, 70105, '\P{^Script=_-SHARADA}', "");
Expect(0, 70106, '\p{Script=_-SHARADA}', "");
Expect(1, 70106, '\p{^Script=_-SHARADA}', "");
Expect(1, 70106, '\P{Script=_-SHARADA}', "");
Expect(0, 70106, '\P{^Script=_-SHARADA}', "");
Error('\p{Sc=-Shrd/a/}');
Error('\P{Sc=-Shrd/a/}');
Expect(1, 70105, '\p{Sc:	shrd}', "");
Expect(0, 70105, '\p{^Sc:	shrd}', "");
Expect(0, 70105, '\P{Sc:	shrd}', "");
Expect(1, 70105, '\P{^Sc:	shrd}', "");
Expect(0, 70106, '\p{Sc:	shrd}', "");
Expect(1, 70106, '\p{^Sc:	shrd}', "");
Expect(1, 70106, '\P{Sc:	shrd}', "");
Expect(0, 70106, '\P{^Sc:	shrd}', "");
Expect(1, 70105, '\p{Sc=_	shrd}', "");
Expect(0, 70105, '\p{^Sc=_	shrd}', "");
Expect(0, 70105, '\P{Sc=_	shrd}', "");
Expect(1, 70105, '\P{^Sc=_	shrd}', "");
Expect(0, 70106, '\p{Sc=_	shrd}', "");
Expect(1, 70106, '\p{^Sc=_	shrd}', "");
Expect(1, 70106, '\P{Sc=_	shrd}', "");
Expect(0, 70106, '\P{^Sc=_	shrd}', "");
Error('\p{Is_Script=-/a/Sharada}');
Error('\P{Is_Script=-/a/Sharada}');
Expect(1, 70105, '\p{Is_Script=sharada}', "");
Expect(0, 70105, '\p{^Is_Script=sharada}', "");
Expect(0, 70105, '\P{Is_Script=sharada}', "");
Expect(1, 70105, '\P{^Is_Script=sharada}', "");
Expect(0, 70106, '\p{Is_Script=sharada}', "");
Expect(1, 70106, '\p{^Is_Script=sharada}', "");
Expect(1, 70106, '\P{Is_Script=sharada}', "");
Expect(0, 70106, '\P{^Is_Script=sharada}', "");
Expect(1, 70105, '\p{Is_Script=Sharada}', "");
Expect(0, 70105, '\p{^Is_Script=Sharada}', "");
Expect(0, 70105, '\P{Is_Script=Sharada}', "");
Expect(1, 70105, '\P{^Is_Script=Sharada}', "");
Expect(0, 70106, '\p{Is_Script=Sharada}', "");
Expect(1, 70106, '\p{^Is_Script=Sharada}', "");
Expect(1, 70106, '\P{Is_Script=Sharada}', "");
Expect(0, 70106, '\P{^Is_Script=Sharada}', "");
Error('\p{Is_Sc=/a/shrd}');
Error('\P{Is_Sc=/a/shrd}');
Expect(1, 70105, '\p{Is_Sc=shrd}', "");
Expect(0, 70105, '\p{^Is_Sc=shrd}', "");
Expect(0, 70105, '\P{Is_Sc=shrd}', "");
Expect(1, 70105, '\P{^Is_Sc=shrd}', "");
Expect(0, 70106, '\p{Is_Sc=shrd}', "");
Expect(1, 70106, '\p{^Is_Sc=shrd}', "");
Expect(1, 70106, '\P{Is_Sc=shrd}', "");
Expect(0, 70106, '\P{^Is_Sc=shrd}', "");
Expect(1, 70105, '\p{Is_Sc=_Shrd}', "");
Expect(0, 70105, '\p{^Is_Sc=_Shrd}', "");
Expect(0, 70105, '\P{Is_Sc=_Shrd}', "");
Expect(1, 70105, '\P{^Is_Sc=_Shrd}', "");
Expect(0, 70106, '\p{Is_Sc=_Shrd}', "");
Expect(1, 70106, '\p{^Is_Sc=_Shrd}', "");
Expect(1, 70106, '\P{Is_Sc=_Shrd}', "");
Expect(0, 70106, '\P{^Is_Sc=_Shrd}', "");
Error('\p{Script=_/a/Sinhala}');
Error('\P{Script=_/a/Sinhala}');
Expect(1, 3572, '\p{Script=sinhala}', "");
Expect(0, 3572, '\p{^Script=sinhala}', "");
Expect(0, 3572, '\P{Script=sinhala}', "");
Expect(1, 3572, '\P{^Script=sinhala}', "");
Expect(0, 3573, '\p{Script=sinhala}', "");
Expect(1, 3573, '\p{^Script=sinhala}', "");
Expect(1, 3573, '\P{Script=sinhala}', "");
Expect(0, 3573, '\P{^Script=sinhala}', "");
Expect(1, 3572, '\p{Script=	_Sinhala}', "");
Expect(0, 3572, '\p{^Script=	_Sinhala}', "");
Expect(0, 3572, '\P{Script=	_Sinhala}', "");
Expect(1, 3572, '\P{^Script=	_Sinhala}', "");
Expect(0, 3573, '\p{Script=	_Sinhala}', "");
Expect(1, 3573, '\p{^Script=	_Sinhala}', "");
Expect(1, 3573, '\P{Script=	_Sinhala}', "");
Expect(0, 3573, '\P{^Script=	_Sinhala}', "");
Error('\p{Sc:	_/a/SINH}');
Error('\P{Sc:	_/a/SINH}');
Expect(1, 3572, '\p{Sc=sinh}', "");
Expect(0, 3572, '\p{^Sc=sinh}', "");
Expect(0, 3572, '\P{Sc=sinh}', "");
Expect(1, 3572, '\P{^Sc=sinh}', "");
Expect(0, 3573, '\p{Sc=sinh}', "");
Expect(1, 3573, '\p{^Sc=sinh}', "");
Expect(1, 3573, '\P{Sc=sinh}', "");
Expect(0, 3573, '\P{^Sc=sinh}', "");
Expect(1, 3572, '\p{Sc=-sinh}', "");
Expect(0, 3572, '\p{^Sc=-sinh}', "");
Expect(0, 3572, '\P{Sc=-sinh}', "");
Expect(1, 3572, '\P{^Sc=-sinh}', "");
Expect(0, 3573, '\p{Sc=-sinh}', "");
Expect(1, 3573, '\p{^Sc=-sinh}', "");
Expect(1, 3573, '\P{Sc=-sinh}', "");
Expect(0, 3573, '\P{^Sc=-sinh}', "");
Error('\p{Is_Script:	/a/	SINHALA}');
Error('\P{Is_Script:	/a/	SINHALA}');
Expect(1, 3572, '\p{Is_Script=sinhala}', "");
Expect(0, 3572, '\p{^Is_Script=sinhala}', "");
Expect(0, 3572, '\P{Is_Script=sinhala}', "");
Expect(1, 3572, '\P{^Is_Script=sinhala}', "");
Expect(0, 3573, '\p{Is_Script=sinhala}', "");
Expect(1, 3573, '\p{^Is_Script=sinhala}', "");
Expect(1, 3573, '\P{Is_Script=sinhala}', "");
Expect(0, 3573, '\P{^Is_Script=sinhala}', "");
Expect(1, 3572, '\p{Is_Script=	Sinhala}', "");
Expect(0, 3572, '\p{^Is_Script=	Sinhala}', "");
Expect(0, 3572, '\P{Is_Script=	Sinhala}', "");
Expect(1, 3572, '\P{^Is_Script=	Sinhala}', "");
Expect(0, 3573, '\p{Is_Script=	Sinhala}', "");
Expect(1, 3573, '\p{^Is_Script=	Sinhala}', "");
Expect(1, 3573, '\P{Is_Script=	Sinhala}', "");
Expect(0, 3573, '\P{^Is_Script=	Sinhala}', "");
Error('\p{Is_Sc=:=_	Sinh}');
Error('\P{Is_Sc=:=_	Sinh}');
Expect(1, 3572, '\p{Is_Sc=sinh}', "");
Expect(0, 3572, '\p{^Is_Sc=sinh}', "");
Expect(0, 3572, '\P{Is_Sc=sinh}', "");
Expect(1, 3572, '\P{^Is_Sc=sinh}', "");
Expect(0, 3573, '\p{Is_Sc=sinh}', "");
Expect(1, 3573, '\p{^Is_Sc=sinh}', "");
Expect(1, 3573, '\P{Is_Sc=sinh}', "");
Expect(0, 3573, '\P{^Is_Sc=sinh}', "");
Expect(1, 3572, '\p{Is_Sc=-	Sinh}', "");
Expect(0, 3572, '\p{^Is_Sc=-	Sinh}', "");
Expect(0, 3572, '\P{Is_Sc=-	Sinh}', "");
Expect(1, 3572, '\P{^Is_Sc=-	Sinh}', "");
Expect(0, 3573, '\p{Is_Sc=-	Sinh}', "");
Expect(1, 3573, '\p{^Is_Sc=-	Sinh}', "");
Expect(1, 3573, '\P{Is_Sc=-	Sinh}', "");
Expect(0, 3573, '\P{^Is_Sc=-	Sinh}', "");
Error('\p{Script=-:=sora_sompeng}');
Error('\P{Script=-:=sora_sompeng}');
Expect(1, 69881, '\p{Script=sorasompeng}', "");
Expect(0, 69881, '\p{^Script=sorasompeng}', "");
Expect(0, 69881, '\P{Script=sorasompeng}', "");
Expect(1, 69881, '\P{^Script=sorasompeng}', "");
Expect(0, 69882, '\p{Script=sorasompeng}', "");
Expect(1, 69882, '\p{^Script=sorasompeng}', "");
Expect(1, 69882, '\P{Script=sorasompeng}', "");
Expect(0, 69882, '\P{^Script=sorasompeng}', "");
Expect(1, 69881, '\p{Script= -Sora_SOMPENG}', "");
Expect(0, 69881, '\p{^Script= -Sora_SOMPENG}', "");
Expect(0, 69881, '\P{Script= -Sora_SOMPENG}', "");
Expect(1, 69881, '\P{^Script= -Sora_SOMPENG}', "");
Expect(0, 69882, '\p{Script= -Sora_SOMPENG}', "");
Expect(1, 69882, '\p{^Script= -Sora_SOMPENG}', "");
Expect(1, 69882, '\P{Script= -Sora_SOMPENG}', "");
Expect(0, 69882, '\P{^Script= -Sora_SOMPENG}', "");
Error('\p{Sc:_	sora:=}');
Error('\P{Sc:_	sora:=}');
Expect(1, 69881, '\p{Sc:	sora}', "");
Expect(0, 69881, '\p{^Sc:	sora}', "");
Expect(0, 69881, '\P{Sc:	sora}', "");
Expect(1, 69881, '\P{^Sc:	sora}', "");
Expect(0, 69882, '\p{Sc:	sora}', "");
Expect(1, 69882, '\p{^Sc:	sora}', "");
Expect(1, 69882, '\P{Sc:	sora}', "");
Expect(0, 69882, '\P{^Sc:	sora}', "");
Expect(1, 69881, '\p{Sc=	-Sora}', "");
Expect(0, 69881, '\p{^Sc=	-Sora}', "");
Expect(0, 69881, '\P{Sc=	-Sora}', "");
Expect(1, 69881, '\P{^Sc=	-Sora}', "");
Expect(0, 69882, '\p{Sc=	-Sora}', "");
Expect(1, 69882, '\p{^Sc=	-Sora}', "");
Expect(1, 69882, '\P{Sc=	-Sora}', "");
Expect(0, 69882, '\P{^Sc=	-Sora}', "");
Error('\p{Is_Script=/a/ sora_Sompeng}');
Error('\P{Is_Script=/a/ sora_Sompeng}');
Expect(1, 69881, '\p{Is_Script=sorasompeng}', "");
Expect(0, 69881, '\p{^Is_Script=sorasompeng}', "");
Expect(0, 69881, '\P{Is_Script=sorasompeng}', "");
Expect(1, 69881, '\P{^Is_Script=sorasompeng}', "");
Expect(0, 69882, '\p{Is_Script=sorasompeng}', "");
Expect(1, 69882, '\p{^Is_Script=sorasompeng}', "");
Expect(1, 69882, '\P{Is_Script=sorasompeng}', "");
Expect(0, 69882, '\P{^Is_Script=sorasompeng}', "");
Expect(1, 69881, '\p{Is_Script=--Sora_Sompeng}', "");
Expect(0, 69881, '\p{^Is_Script=--Sora_Sompeng}', "");
Expect(0, 69881, '\P{Is_Script=--Sora_Sompeng}', "");
Expect(1, 69881, '\P{^Is_Script=--Sora_Sompeng}', "");
Expect(0, 69882, '\p{Is_Script=--Sora_Sompeng}', "");
Expect(1, 69882, '\p{^Is_Script=--Sora_Sompeng}', "");
Expect(1, 69882, '\P{Is_Script=--Sora_Sompeng}', "");
Expect(0, 69882, '\P{^Is_Script=--Sora_Sompeng}', "");
Error('\p{Is_Sc=-	SORA:=}');
Error('\P{Is_Sc=-	SORA:=}');
Expect(1, 69881, '\p{Is_Sc=sora}', "");
Expect(0, 69881, '\p{^Is_Sc=sora}', "");
Expect(0, 69881, '\P{Is_Sc=sora}', "");
Expect(1, 69881, '\P{^Is_Sc=sora}', "");
Expect(0, 69882, '\p{Is_Sc=sora}', "");
Expect(1, 69882, '\p{^Is_Sc=sora}', "");
Expect(1, 69882, '\P{Is_Sc=sora}', "");
Expect(0, 69882, '\P{^Is_Sc=sora}', "");
Expect(1, 69881, '\p{Is_Sc: 	_SORA}', "");
Expect(0, 69881, '\p{^Is_Sc: 	_SORA}', "");
Expect(0, 69881, '\P{Is_Sc: 	_SORA}', "");
Expect(1, 69881, '\P{^Is_Sc: 	_SORA}', "");
Expect(0, 69882, '\p{Is_Sc: 	_SORA}', "");
Expect(1, 69882, '\p{^Is_Sc: 	_SORA}', "");
Expect(1, 69882, '\P{Is_Sc: 	_SORA}', "");
Expect(0, 69882, '\P{^Is_Sc: 	_SORA}', "");
Error('\p{Script=	Sundanese:=}');
Error('\P{Script=	Sundanese:=}');
Expect(1, 7367, '\p{Script=sundanese}', "");
Expect(0, 7367, '\p{^Script=sundanese}', "");
Expect(0, 7367, '\P{Script=sundanese}', "");
Expect(1, 7367, '\P{^Script=sundanese}', "");
Expect(0, 7368, '\p{Script=sundanese}', "");
Expect(1, 7368, '\p{^Script=sundanese}', "");
Expect(1, 7368, '\P{Script=sundanese}', "");
Expect(0, 7368, '\P{^Script=sundanese}', "");
Expect(1, 7367, '\p{Script= Sundanese}', "");
Expect(0, 7367, '\p{^Script= Sundanese}', "");
Expect(0, 7367, '\P{Script= Sundanese}', "");
Expect(1, 7367, '\P{^Script= Sundanese}', "");
Expect(0, 7368, '\p{Script= Sundanese}', "");
Expect(1, 7368, '\p{^Script= Sundanese}', "");
Expect(1, 7368, '\P{Script= Sundanese}', "");
Expect(0, 7368, '\P{^Script= Sundanese}', "");
Error('\p{Sc=_/a/SUND}');
Error('\P{Sc=_/a/SUND}');
Expect(1, 7367, '\p{Sc=sund}', "");
Expect(0, 7367, '\p{^Sc=sund}', "");
Expect(0, 7367, '\P{Sc=sund}', "");
Expect(1, 7367, '\P{^Sc=sund}', "");
Expect(0, 7368, '\p{Sc=sund}', "");
Expect(1, 7368, '\p{^Sc=sund}', "");
Expect(1, 7368, '\P{Sc=sund}', "");
Expect(0, 7368, '\P{^Sc=sund}', "");
Expect(1, 7367, '\p{Sc=  SUND}', "");
Expect(0, 7367, '\p{^Sc=  SUND}', "");
Expect(0, 7367, '\P{Sc=  SUND}', "");
Expect(1, 7367, '\P{^Sc=  SUND}', "");
Expect(0, 7368, '\p{Sc=  SUND}', "");
Expect(1, 7368, '\p{^Sc=  SUND}', "");
Expect(1, 7368, '\P{Sc=  SUND}', "");
Expect(0, 7368, '\P{^Sc=  SUND}', "");
Error('\p{Is_Script=/a/ Sundanese}');
Error('\P{Is_Script=/a/ Sundanese}');
Expect(1, 7367, '\p{Is_Script=sundanese}', "");
Expect(0, 7367, '\p{^Is_Script=sundanese}', "");
Expect(0, 7367, '\P{Is_Script=sundanese}', "");
Expect(1, 7367, '\P{^Is_Script=sundanese}', "");
Expect(0, 7368, '\p{Is_Script=sundanese}', "");
Expect(1, 7368, '\p{^Is_Script=sundanese}', "");
Expect(1, 7368, '\P{Is_Script=sundanese}', "");
Expect(0, 7368, '\P{^Is_Script=sundanese}', "");
Expect(1, 7367, '\p{Is_Script=_sundanese}', "");
Expect(0, 7367, '\p{^Is_Script=_sundanese}', "");
Expect(0, 7367, '\P{Is_Script=_sundanese}', "");
Expect(1, 7367, '\P{^Is_Script=_sundanese}', "");
Expect(0, 7368, '\p{Is_Script=_sundanese}', "");
Expect(1, 7368, '\p{^Is_Script=_sundanese}', "");
Expect(1, 7368, '\P{Is_Script=_sundanese}', "");
Expect(0, 7368, '\P{^Is_Script=_sundanese}', "");
Error('\p{Is_Sc=:=SUND}');
Error('\P{Is_Sc=:=SUND}');
Expect(1, 7367, '\p{Is_Sc=sund}', "");
Expect(0, 7367, '\p{^Is_Sc=sund}', "");
Expect(0, 7367, '\P{Is_Sc=sund}', "");
Expect(1, 7367, '\P{^Is_Sc=sund}', "");
Expect(0, 7368, '\p{Is_Sc=sund}', "");
Expect(1, 7368, '\p{^Is_Sc=sund}', "");
Expect(1, 7368, '\P{Is_Sc=sund}', "");
Expect(0, 7368, '\P{^Is_Sc=sund}', "");
Expect(1, 7367, '\p{Is_Sc=	Sund}', "");
Expect(0, 7367, '\p{^Is_Sc=	Sund}', "");
Expect(0, 7367, '\P{Is_Sc=	Sund}', "");
Expect(1, 7367, '\P{^Is_Sc=	Sund}', "");
Expect(0, 7368, '\p{Is_Sc=	Sund}', "");
Expect(1, 7368, '\p{^Is_Sc=	Sund}', "");
Expect(1, 7368, '\P{Is_Sc=	Sund}', "");
Expect(0, 7368, '\P{^Is_Sc=	Sund}', "");
Error('\p{Script:   --Syloti_NAGRI:=}');
Error('\P{Script:   --Syloti_NAGRI:=}');
Expect(1, 43051, '\p{Script=sylotinagri}', "");
Expect(0, 43051, '\p{^Script=sylotinagri}', "");
Expect(0, 43051, '\P{Script=sylotinagri}', "");
Expect(1, 43051, '\P{^Script=sylotinagri}', "");
Expect(0, 43052, '\p{Script=sylotinagri}', "");
Expect(1, 43052, '\p{^Script=sylotinagri}', "");
Expect(1, 43052, '\P{Script=sylotinagri}', "");
Expect(0, 43052, '\P{^Script=sylotinagri}', "");
Expect(1, 43051, '\p{Script=--Syloti_nagri}', "");
Expect(0, 43051, '\p{^Script=--Syloti_nagri}', "");
Expect(0, 43051, '\P{Script=--Syloti_nagri}', "");
Expect(1, 43051, '\P{^Script=--Syloti_nagri}', "");
Expect(0, 43052, '\p{Script=--Syloti_nagri}', "");
Expect(1, 43052, '\p{^Script=--Syloti_nagri}', "");
Expect(1, 43052, '\P{Script=--Syloti_nagri}', "");
Expect(0, 43052, '\P{^Script=--Syloti_nagri}', "");
Error('\p{Sc=	/a/Sylo}');
Error('\P{Sc=	/a/Sylo}');
Expect(1, 43051, '\p{Sc=sylo}', "");
Expect(0, 43051, '\p{^Sc=sylo}', "");
Expect(0, 43051, '\P{Sc=sylo}', "");
Expect(1, 43051, '\P{^Sc=sylo}', "");
Expect(0, 43052, '\p{Sc=sylo}', "");
Expect(1, 43052, '\p{^Sc=sylo}', "");
Expect(1, 43052, '\P{Sc=sylo}', "");
Expect(0, 43052, '\P{^Sc=sylo}', "");
Expect(1, 43051, '\p{Sc=-	sylo}', "");
Expect(0, 43051, '\p{^Sc=-	sylo}', "");
Expect(0, 43051, '\P{Sc=-	sylo}', "");
Expect(1, 43051, '\P{^Sc=-	sylo}', "");
Expect(0, 43052, '\p{Sc=-	sylo}', "");
Expect(1, 43052, '\p{^Sc=-	sylo}', "");
Expect(1, 43052, '\P{Sc=-	sylo}', "");
Expect(0, 43052, '\P{^Sc=-	sylo}', "");
Error('\p{Is_Script=/a/__Syloti_NAGRI}');
Error('\P{Is_Script=/a/__Syloti_NAGRI}');
Expect(1, 43051, '\p{Is_Script=sylotinagri}', "");
Expect(0, 43051, '\p{^Is_Script=sylotinagri}', "");
Expect(0, 43051, '\P{Is_Script=sylotinagri}', "");
Expect(1, 43051, '\P{^Is_Script=sylotinagri}', "");
Expect(0, 43052, '\p{Is_Script=sylotinagri}', "");
Expect(1, 43052, '\p{^Is_Script=sylotinagri}', "");
Expect(1, 43052, '\P{Is_Script=sylotinagri}', "");
Expect(0, 43052, '\P{^Is_Script=sylotinagri}', "");
Expect(1, 43051, '\p{Is_Script= 	SYLOTI_nagri}', "");
Expect(0, 43051, '\p{^Is_Script= 	SYLOTI_nagri}', "");
Expect(0, 43051, '\P{Is_Script= 	SYLOTI_nagri}', "");
Expect(1, 43051, '\P{^Is_Script= 	SYLOTI_nagri}', "");
Expect(0, 43052, '\p{Is_Script= 	SYLOTI_nagri}', "");
Expect(1, 43052, '\p{^Is_Script= 	SYLOTI_nagri}', "");
Expect(1, 43052, '\P{Is_Script= 	SYLOTI_nagri}', "");
Expect(0, 43052, '\P{^Is_Script= 	SYLOTI_nagri}', "");
Error('\p{Is_Sc=	_sylo:=}');
Error('\P{Is_Sc=	_sylo:=}');
Expect(1, 43051, '\p{Is_Sc=sylo}', "");
Expect(0, 43051, '\p{^Is_Sc=sylo}', "");
Expect(0, 43051, '\P{Is_Sc=sylo}', "");
Expect(1, 43051, '\P{^Is_Sc=sylo}', "");
Expect(0, 43052, '\p{Is_Sc=sylo}', "");
Expect(1, 43052, '\p{^Is_Sc=sylo}', "");
Expect(1, 43052, '\P{Is_Sc=sylo}', "");
Expect(0, 43052, '\P{^Is_Sc=sylo}', "");
Expect(1, 43051, '\p{Is_Sc:   	-SYLO}', "");
Expect(0, 43051, '\p{^Is_Sc:   	-SYLO}', "");
Expect(0, 43051, '\P{Is_Sc:   	-SYLO}', "");
Expect(1, 43051, '\P{^Is_Sc:   	-SYLO}', "");
Expect(0, 43052, '\p{Is_Sc:   	-SYLO}', "");
Expect(1, 43052, '\p{^Is_Sc:   	-SYLO}', "");
Expect(1, 43052, '\P{Is_Sc:   	-SYLO}', "");
Expect(0, 43052, '\P{^Is_Sc:   	-SYLO}', "");
Error('\p{Script=:=__Syriac}');
Error('\P{Script=:=__Syriac}');
Expect(1, 1871, '\p{Script=syriac}', "");
Expect(0, 1871, '\p{^Script=syriac}', "");
Expect(0, 1871, '\P{Script=syriac}', "");
Expect(1, 1871, '\P{^Script=syriac}', "");
Expect(0, 1872, '\p{Script=syriac}', "");
Expect(1, 1872, '\p{^Script=syriac}', "");
Expect(1, 1872, '\P{Script=syriac}', "");
Expect(0, 1872, '\P{^Script=syriac}', "");
Expect(1, 1871, '\p{Script=__SYRIAC}', "");
Expect(0, 1871, '\p{^Script=__SYRIAC}', "");
Expect(0, 1871, '\P{Script=__SYRIAC}', "");
Expect(1, 1871, '\P{^Script=__SYRIAC}', "");
Expect(0, 1872, '\p{Script=__SYRIAC}', "");
Expect(1, 1872, '\p{^Script=__SYRIAC}', "");
Expect(1, 1872, '\P{Script=__SYRIAC}', "");
Expect(0, 1872, '\P{^Script=__SYRIAC}', "");
Error('\p{Sc=-syrc/a/}');
Error('\P{Sc=-syrc/a/}');
Expect(1, 1871, '\p{Sc=syrc}', "");
Expect(0, 1871, '\p{^Sc=syrc}', "");
Expect(0, 1871, '\P{Sc=syrc}', "");
Expect(1, 1871, '\P{^Sc=syrc}', "");
Expect(0, 1872, '\p{Sc=syrc}', "");
Expect(1, 1872, '\p{^Sc=syrc}', "");
Expect(1, 1872, '\P{Sc=syrc}', "");
Expect(0, 1872, '\P{^Sc=syrc}', "");
Expect(1, 1871, '\p{Sc=	-Syrc}', "");
Expect(0, 1871, '\p{^Sc=	-Syrc}', "");
Expect(0, 1871, '\P{Sc=	-Syrc}', "");
Expect(1, 1871, '\P{^Sc=	-Syrc}', "");
Expect(0, 1872, '\p{Sc=	-Syrc}', "");
Expect(1, 1872, '\p{^Sc=	-Syrc}', "");
Expect(1, 1872, '\P{Sc=	-Syrc}', "");
Expect(0, 1872, '\P{^Sc=	-Syrc}', "");
Error('\p{Is_Script=		Syriac:=}');
Error('\P{Is_Script=		Syriac:=}');
Expect(1, 1871, '\p{Is_Script=syriac}', "");
Expect(0, 1871, '\p{^Is_Script=syriac}', "");
Expect(0, 1871, '\P{Is_Script=syriac}', "");
Expect(1, 1871, '\P{^Is_Script=syriac}', "");
Expect(0, 1872, '\p{Is_Script=syriac}', "");
Expect(1, 1872, '\p{^Is_Script=syriac}', "");
Expect(1, 1872, '\P{Is_Script=syriac}', "");
Expect(0, 1872, '\P{^Is_Script=syriac}', "");
Expect(1, 1871, '\p{Is_Script=-	Syriac}', "");
Expect(0, 1871, '\p{^Is_Script=-	Syriac}', "");
Expect(0, 1871, '\P{Is_Script=-	Syriac}', "");
Expect(1, 1871, '\P{^Is_Script=-	Syriac}', "");
Expect(0, 1872, '\p{Is_Script=-	Syriac}', "");
Expect(1, 1872, '\p{^Is_Script=-	Syriac}', "");
Expect(1, 1872, '\P{Is_Script=-	Syriac}', "");
Expect(0, 1872, '\P{^Is_Script=-	Syriac}', "");
Error('\p{Is_Sc=	/a/Syrc}');
Error('\P{Is_Sc=	/a/Syrc}');
Expect(1, 1871, '\p{Is_Sc=syrc}', "");
Expect(0, 1871, '\p{^Is_Sc=syrc}', "");
Expect(0, 1871, '\P{Is_Sc=syrc}', "");
Expect(1, 1871, '\P{^Is_Sc=syrc}', "");
Expect(0, 1872, '\p{Is_Sc=syrc}', "");
Expect(1, 1872, '\p{^Is_Sc=syrc}', "");
Expect(1, 1872, '\P{Is_Sc=syrc}', "");
Expect(0, 1872, '\P{^Is_Sc=syrc}', "");
Expect(1, 1871, '\p{Is_Sc= Syrc}', "");
Expect(0, 1871, '\p{^Is_Sc= Syrc}', "");
Expect(0, 1871, '\P{Is_Sc= Syrc}', "");
Expect(1, 1871, '\P{^Is_Sc= Syrc}', "");
Expect(0, 1872, '\p{Is_Sc= Syrc}', "");
Expect(1, 1872, '\p{^Is_Sc= Syrc}', "");
Expect(1, 1872, '\P{Is_Sc= Syrc}', "");
Expect(0, 1872, '\P{^Is_Sc= Syrc}', "");
Error('\p{Script=/a/-tagbanwa}');
Error('\P{Script=/a/-tagbanwa}');
Expect(1, 6003, '\p{Script=tagbanwa}', "");
Expect(0, 6003, '\p{^Script=tagbanwa}', "");
Expect(0, 6003, '\P{Script=tagbanwa}', "");
Expect(1, 6003, '\P{^Script=tagbanwa}', "");
Expect(0, 6004, '\p{Script=tagbanwa}', "");
Expect(1, 6004, '\p{^Script=tagbanwa}', "");
Expect(1, 6004, '\P{Script=tagbanwa}', "");
Expect(0, 6004, '\P{^Script=tagbanwa}', "");
Expect(1, 6003, '\p{Script= tagbanwa}', "");
Expect(0, 6003, '\p{^Script= tagbanwa}', "");
Expect(0, 6003, '\P{Script= tagbanwa}', "");
Expect(1, 6003, '\P{^Script= tagbanwa}', "");
Expect(0, 6004, '\p{Script= tagbanwa}', "");
Expect(1, 6004, '\p{^Script= tagbanwa}', "");
Expect(1, 6004, '\P{Script= tagbanwa}', "");
Expect(0, 6004, '\P{^Script= tagbanwa}', "");
Error('\p{Sc=_/a/Tagb}');
Error('\P{Sc=_/a/Tagb}');
Expect(1, 6003, '\p{Sc=tagb}', "");
Expect(0, 6003, '\p{^Sc=tagb}', "");
Expect(0, 6003, '\P{Sc=tagb}', "");
Expect(1, 6003, '\P{^Sc=tagb}', "");
Expect(0, 6004, '\p{Sc=tagb}', "");
Expect(1, 6004, '\p{^Sc=tagb}', "");
Expect(1, 6004, '\P{Sc=tagb}', "");
Expect(0, 6004, '\P{^Sc=tagb}', "");
Expect(1, 6003, '\p{Sc=TAGB}', "");
Expect(0, 6003, '\p{^Sc=TAGB}', "");
Expect(0, 6003, '\P{Sc=TAGB}', "");
Expect(1, 6003, '\P{^Sc=TAGB}', "");
Expect(0, 6004, '\p{Sc=TAGB}', "");
Expect(1, 6004, '\p{^Sc=TAGB}', "");
Expect(1, 6004, '\P{Sc=TAGB}', "");
Expect(0, 6004, '\P{^Sc=TAGB}', "");
Error('\p{Is_Script= /a/tagbanwa}');
Error('\P{Is_Script= /a/tagbanwa}');
Expect(1, 6003, '\p{Is_Script=tagbanwa}', "");
Expect(0, 6003, '\p{^Is_Script=tagbanwa}', "");
Expect(0, 6003, '\P{Is_Script=tagbanwa}', "");
Expect(1, 6003, '\P{^Is_Script=tagbanwa}', "");
Expect(0, 6004, '\p{Is_Script=tagbanwa}', "");
Expect(1, 6004, '\p{^Is_Script=tagbanwa}', "");
Expect(1, 6004, '\P{Is_Script=tagbanwa}', "");
Expect(0, 6004, '\P{^Is_Script=tagbanwa}', "");
Expect(1, 6003, '\p{Is_Script=_	tagbanwa}', "");
Expect(0, 6003, '\p{^Is_Script=_	tagbanwa}', "");
Expect(0, 6003, '\P{Is_Script=_	tagbanwa}', "");
Expect(1, 6003, '\P{^Is_Script=_	tagbanwa}', "");
Expect(0, 6004, '\p{Is_Script=_	tagbanwa}', "");
Expect(1, 6004, '\p{^Is_Script=_	tagbanwa}', "");
Expect(1, 6004, '\P{Is_Script=_	tagbanwa}', "");
Expect(0, 6004, '\P{^Is_Script=_	tagbanwa}', "");
Error('\p{Is_Sc=-/a/tagb}');
Error('\P{Is_Sc=-/a/tagb}');
Expect(1, 6003, '\p{Is_Sc=tagb}', "");
Expect(0, 6003, '\p{^Is_Sc=tagb}', "");
Expect(0, 6003, '\P{Is_Sc=tagb}', "");
Expect(1, 6003, '\P{^Is_Sc=tagb}', "");
Expect(0, 6004, '\p{Is_Sc=tagb}', "");
Expect(1, 6004, '\p{^Is_Sc=tagb}', "");
Expect(1, 6004, '\P{Is_Sc=tagb}', "");
Expect(0, 6004, '\P{^Is_Sc=tagb}', "");
Expect(1, 6003, '\p{Is_Sc=	-Tagb}', "");
Expect(0, 6003, '\p{^Is_Sc=	-Tagb}', "");
Expect(0, 6003, '\P{Is_Sc=	-Tagb}', "");
Expect(1, 6003, '\P{^Is_Sc=	-Tagb}', "");
Expect(0, 6004, '\p{Is_Sc=	-Tagb}', "");
Expect(1, 6004, '\p{^Is_Sc=	-Tagb}', "");
Expect(1, 6004, '\P{Is_Sc=	-Tagb}', "");
Expect(0, 6004, '\P{^Is_Sc=	-Tagb}', "");
Error('\p{Script= 	takri:=}');
Error('\P{Script= 	takri:=}');
Expect(1, 71369, '\p{Script=takri}', "");
Expect(0, 71369, '\p{^Script=takri}', "");
Expect(0, 71369, '\P{Script=takri}', "");
Expect(1, 71369, '\P{^Script=takri}', "");
Expect(0, 71370, '\p{Script=takri}', "");
Expect(1, 71370, '\p{^Script=takri}', "");
Expect(1, 71370, '\P{Script=takri}', "");
Expect(0, 71370, '\P{^Script=takri}', "");
Expect(1, 71369, '\p{Script=-Takri}', "");
Expect(0, 71369, '\p{^Script=-Takri}', "");
Expect(0, 71369, '\P{Script=-Takri}', "");
Expect(1, 71369, '\P{^Script=-Takri}', "");
Expect(0, 71370, '\p{Script=-Takri}', "");
Expect(1, 71370, '\p{^Script=-Takri}', "");
Expect(1, 71370, '\P{Script=-Takri}', "");
Expect(0, 71370, '\P{^Script=-Takri}', "");
Error('\p{Sc=	/a/Takr}');
Error('\P{Sc=	/a/Takr}');
Expect(1, 71369, '\p{Sc=takr}', "");
Expect(0, 71369, '\p{^Sc=takr}', "");
Expect(0, 71369, '\P{Sc=takr}', "");
Expect(1, 71369, '\P{^Sc=takr}', "");
Expect(0, 71370, '\p{Sc=takr}', "");
Expect(1, 71370, '\p{^Sc=takr}', "");
Expect(1, 71370, '\P{Sc=takr}', "");
Expect(0, 71370, '\P{^Sc=takr}', "");
Expect(1, 71369, '\p{Sc=	-Takr}', "");
Expect(0, 71369, '\p{^Sc=	-Takr}', "");
Expect(0, 71369, '\P{Sc=	-Takr}', "");
Expect(1, 71369, '\P{^Sc=	-Takr}', "");
Expect(0, 71370, '\p{Sc=	-Takr}', "");
Expect(1, 71370, '\p{^Sc=	-Takr}', "");
Expect(1, 71370, '\P{Sc=	-Takr}', "");
Expect(0, 71370, '\P{^Sc=	-Takr}', "");
Error('\p{Is_Script=/a/		TAKRI}');
Error('\P{Is_Script=/a/		TAKRI}');
Expect(1, 71369, '\p{Is_Script=takri}', "");
Expect(0, 71369, '\p{^Is_Script=takri}', "");
Expect(0, 71369, '\P{Is_Script=takri}', "");
Expect(1, 71369, '\P{^Is_Script=takri}', "");
Expect(0, 71370, '\p{Is_Script=takri}', "");
Expect(1, 71370, '\p{^Is_Script=takri}', "");
Expect(1, 71370, '\P{Is_Script=takri}', "");
Expect(0, 71370, '\P{^Is_Script=takri}', "");
Expect(1, 71369, '\p{Is_Script=	takri}', "");
Expect(0, 71369, '\p{^Is_Script=	takri}', "");
Expect(0, 71369, '\P{Is_Script=	takri}', "");
Expect(1, 71369, '\P{^Is_Script=	takri}', "");
Expect(0, 71370, '\p{Is_Script=	takri}', "");
Expect(1, 71370, '\p{^Is_Script=	takri}', "");
Expect(1, 71370, '\P{Is_Script=	takri}', "");
Expect(0, 71370, '\P{^Is_Script=	takri}', "");
Error('\p{Is_Sc=:= takr}');
Error('\P{Is_Sc=:= takr}');
Expect(1, 71369, '\p{Is_Sc:takr}', "");
Expect(0, 71369, '\p{^Is_Sc:takr}', "");
Expect(0, 71369, '\P{Is_Sc:takr}', "");
Expect(1, 71369, '\P{^Is_Sc:takr}', "");
Expect(0, 71370, '\p{Is_Sc:takr}', "");
Expect(1, 71370, '\p{^Is_Sc:takr}', "");
Expect(1, 71370, '\P{Is_Sc:takr}', "");
Expect(0, 71370, '\P{^Is_Sc:takr}', "");
Expect(1, 71369, '\p{Is_Sc= _takr}', "");
Expect(0, 71369, '\p{^Is_Sc= _takr}', "");
Expect(0, 71369, '\P{Is_Sc= _takr}', "");
Expect(1, 71369, '\P{^Is_Sc= _takr}', "");
Expect(0, 71370, '\p{Is_Sc= _takr}', "");
Expect(1, 71370, '\p{^Is_Sc= _takr}', "");
Expect(1, 71370, '\P{Is_Sc= _takr}', "");
Expect(0, 71370, '\P{^Is_Sc= _takr}', "");
Error('\p{Script= /a/tai_LE}');
Error('\P{Script= /a/tai_LE}');
Expect(1, 6516, '\p{Script=taile}', "");
Expect(0, 6516, '\p{^Script=taile}', "");
Expect(0, 6516, '\P{Script=taile}', "");
Expect(1, 6516, '\P{^Script=taile}', "");
Expect(0, 6517, '\p{Script=taile}', "");
Expect(1, 6517, '\p{^Script=taile}', "");
Expect(1, 6517, '\P{Script=taile}', "");
Expect(0, 6517, '\P{^Script=taile}', "");
Expect(1, 6516, '\p{Script=	-Tai_Le}', "");
Expect(0, 6516, '\p{^Script=	-Tai_Le}', "");
Expect(0, 6516, '\P{Script=	-Tai_Le}', "");
Expect(1, 6516, '\P{^Script=	-Tai_Le}', "");
Expect(0, 6517, '\p{Script=	-Tai_Le}', "");
Expect(1, 6517, '\p{^Script=	-Tai_Le}', "");
Expect(1, 6517, '\P{Script=	-Tai_Le}', "");
Expect(0, 6517, '\P{^Script=	-Tai_Le}', "");
Error('\p{Sc=-tale:=}');
Error('\P{Sc=-tale:=}');
Expect(1, 6516, '\p{Sc=tale}', "");
Expect(0, 6516, '\p{^Sc=tale}', "");
Expect(0, 6516, '\P{Sc=tale}', "");
Expect(1, 6516, '\P{^Sc=tale}', "");
Expect(0, 6517, '\p{Sc=tale}', "");
Expect(1, 6517, '\p{^Sc=tale}', "");
Expect(1, 6517, '\P{Sc=tale}', "");
Expect(0, 6517, '\P{^Sc=tale}', "");
Expect(1, 6516, '\p{Sc:  Tale}', "");
Expect(0, 6516, '\p{^Sc:  Tale}', "");
Expect(0, 6516, '\P{Sc:  Tale}', "");
Expect(1, 6516, '\P{^Sc:  Tale}', "");
Expect(0, 6517, '\p{Sc:  Tale}', "");
Expect(1, 6517, '\p{^Sc:  Tale}', "");
Expect(1, 6517, '\P{Sc:  Tale}', "");
Expect(0, 6517, '\P{^Sc:  Tale}', "");
Error('\p{Is_Script=:=__TAI_LE}');
Error('\P{Is_Script=:=__TAI_LE}');
Expect(1, 6516, '\p{Is_Script=taile}', "");
Expect(0, 6516, '\p{^Is_Script=taile}', "");
Expect(0, 6516, '\P{Is_Script=taile}', "");
Expect(1, 6516, '\P{^Is_Script=taile}', "");
Expect(0, 6517, '\p{Is_Script=taile}', "");
Expect(1, 6517, '\p{^Is_Script=taile}', "");
Expect(1, 6517, '\P{Is_Script=taile}', "");
Expect(0, 6517, '\P{^Is_Script=taile}', "");
Expect(1, 6516, '\p{Is_Script=_ Tai_le}', "");
Expect(0, 6516, '\p{^Is_Script=_ Tai_le}', "");
Expect(0, 6516, '\P{Is_Script=_ Tai_le}', "");
Expect(1, 6516, '\P{^Is_Script=_ Tai_le}', "");
Expect(0, 6517, '\p{Is_Script=_ Tai_le}', "");
Expect(1, 6517, '\p{^Is_Script=_ Tai_le}', "");
Expect(1, 6517, '\P{Is_Script=_ Tai_le}', "");
Expect(0, 6517, '\P{^Is_Script=_ Tai_le}', "");
Error('\p{Is_Sc=:=	 Tale}');
Error('\P{Is_Sc=:=	 Tale}');
Expect(1, 6516, '\p{Is_Sc:   tale}', "");
Expect(0, 6516, '\p{^Is_Sc:   tale}', "");
Expect(0, 6516, '\P{Is_Sc:   tale}', "");
Expect(1, 6516, '\P{^Is_Sc:   tale}', "");
Expect(0, 6517, '\p{Is_Sc:   tale}', "");
Expect(1, 6517, '\p{^Is_Sc:   tale}', "");
Expect(1, 6517, '\P{Is_Sc:   tale}', "");
Expect(0, 6517, '\P{^Is_Sc:   tale}', "");
Expect(1, 6516, '\p{Is_Sc= TALE}', "");
Expect(0, 6516, '\p{^Is_Sc= TALE}', "");
Expect(0, 6516, '\P{Is_Sc= TALE}', "");
Expect(1, 6516, '\P{^Is_Sc= TALE}', "");
Expect(0, 6517, '\p{Is_Sc= TALE}', "");
Expect(1, 6517, '\p{^Is_Sc= TALE}', "");
Expect(1, 6517, '\P{Is_Sc= TALE}', "");
Expect(0, 6517, '\P{^Is_Sc= TALE}', "");
Error('\p{Script=-:=New_Tai_Lue}');
Error('\P{Script=-:=New_Tai_Lue}');
Expect(1, 6623, '\p{Script=newtailue}', "");
Expect(0, 6623, '\p{^Script=newtailue}', "");
Expect(0, 6623, '\P{Script=newtailue}', "");
Expect(1, 6623, '\P{^Script=newtailue}', "");
Expect(0, 6624, '\p{Script=newtailue}', "");
Expect(1, 6624, '\p{^Script=newtailue}', "");
Expect(1, 6624, '\P{Script=newtailue}', "");
Expect(0, 6624, '\P{^Script=newtailue}', "");
Expect(1, 6623, '\p{Script=_	New_Tai_LUE}', "");
Expect(0, 6623, '\p{^Script=_	New_Tai_LUE}', "");
Expect(0, 6623, '\P{Script=_	New_Tai_LUE}', "");
Expect(1, 6623, '\P{^Script=_	New_Tai_LUE}', "");
Expect(0, 6624, '\p{Script=_	New_Tai_LUE}', "");
Expect(1, 6624, '\p{^Script=_	New_Tai_LUE}', "");
Expect(1, 6624, '\P{Script=_	New_Tai_LUE}', "");
Expect(0, 6624, '\P{^Script=_	New_Tai_LUE}', "");
Error('\p{Sc=_ Talu/a/}');
Error('\P{Sc=_ Talu/a/}');
Expect(1, 6623, '\p{Sc=talu}', "");
Expect(0, 6623, '\p{^Sc=talu}', "");
Expect(0, 6623, '\P{Sc=talu}', "");
Expect(1, 6623, '\P{^Sc=talu}', "");
Expect(0, 6624, '\p{Sc=talu}', "");
Expect(1, 6624, '\p{^Sc=talu}', "");
Expect(1, 6624, '\P{Sc=talu}', "");
Expect(0, 6624, '\P{^Sc=talu}', "");
Expect(1, 6623, '\p{Sc=_Talu}', "");
Expect(0, 6623, '\p{^Sc=_Talu}', "");
Expect(0, 6623, '\P{Sc=_Talu}', "");
Expect(1, 6623, '\P{^Sc=_Talu}', "");
Expect(0, 6624, '\p{Sc=_Talu}', "");
Expect(1, 6624, '\p{^Sc=_Talu}', "");
Expect(1, 6624, '\P{Sc=_Talu}', "");
Expect(0, 6624, '\P{^Sc=_Talu}', "");
Error('\p{Is_Script=	 NEW_Tai_LUE/a/}');
Error('\P{Is_Script=	 NEW_Tai_LUE/a/}');
Expect(1, 6623, '\p{Is_Script=newtailue}', "");
Expect(0, 6623, '\p{^Is_Script=newtailue}', "");
Expect(0, 6623, '\P{Is_Script=newtailue}', "");
Expect(1, 6623, '\P{^Is_Script=newtailue}', "");
Expect(0, 6624, '\p{Is_Script=newtailue}', "");
Expect(1, 6624, '\p{^Is_Script=newtailue}', "");
Expect(1, 6624, '\P{Is_Script=newtailue}', "");
Expect(0, 6624, '\P{^Is_Script=newtailue}', "");
Expect(1, 6623, '\p{Is_Script=_ new_tai_Lue}', "");
Expect(0, 6623, '\p{^Is_Script=_ new_tai_Lue}', "");
Expect(0, 6623, '\P{Is_Script=_ new_tai_Lue}', "");
Expect(1, 6623, '\P{^Is_Script=_ new_tai_Lue}', "");
Expect(0, 6624, '\p{Is_Script=_ new_tai_Lue}', "");
Expect(1, 6624, '\p{^Is_Script=_ new_tai_Lue}', "");
Expect(1, 6624, '\P{Is_Script=_ new_tai_Lue}', "");
Expect(0, 6624, '\P{^Is_Script=_ new_tai_Lue}', "");
Error('\p{Is_Sc=:=Talu}');
Error('\P{Is_Sc=:=Talu}');
Expect(1, 6623, '\p{Is_Sc=talu}', "");
Expect(0, 6623, '\p{^Is_Sc=talu}', "");
Expect(0, 6623, '\P{Is_Sc=talu}', "");
Expect(1, 6623, '\P{^Is_Sc=talu}', "");
Expect(0, 6624, '\p{Is_Sc=talu}', "");
Expect(1, 6624, '\p{^Is_Sc=talu}', "");
Expect(1, 6624, '\P{Is_Sc=talu}', "");
Expect(0, 6624, '\P{^Is_Sc=talu}', "");
Expect(1, 6623, '\p{Is_Sc= Talu}', "");
Expect(0, 6623, '\p{^Is_Sc= Talu}', "");
Expect(0, 6623, '\P{Is_Sc= Talu}', "");
Expect(1, 6623, '\P{^Is_Sc= Talu}', "");
Expect(0, 6624, '\p{Is_Sc= Talu}', "");
Expect(1, 6624, '\p{^Is_Sc= Talu}', "");
Expect(1, 6624, '\P{Is_Sc= Talu}', "");
Expect(0, 6624, '\P{^Is_Sc= Talu}', "");
Error('\p{Script=	/a/Tamil}');
Error('\P{Script=	/a/Tamil}');
Expect(1, 3066, '\p{Script=tamil}', "");
Expect(0, 3066, '\p{^Script=tamil}', "");
Expect(0, 3066, '\P{Script=tamil}', "");
Expect(1, 3066, '\P{^Script=tamil}', "");
Expect(0, 3067, '\p{Script=tamil}', "");
Expect(1, 3067, '\p{^Script=tamil}', "");
Expect(1, 3067, '\P{Script=tamil}', "");
Expect(0, 3067, '\P{^Script=tamil}', "");
Expect(1, 3066, '\p{Script=--TAMIL}', "");
Expect(0, 3066, '\p{^Script=--TAMIL}', "");
Expect(0, 3066, '\P{Script=--TAMIL}', "");
Expect(1, 3066, '\P{^Script=--TAMIL}', "");
Expect(0, 3067, '\p{Script=--TAMIL}', "");
Expect(1, 3067, '\p{^Script=--TAMIL}', "");
Expect(1, 3067, '\P{Script=--TAMIL}', "");
Expect(0, 3067, '\P{^Script=--TAMIL}', "");
Error('\p{Sc=:=-_taml}');
Error('\P{Sc=:=-_taml}');
Expect(1, 3066, '\p{Sc=taml}', "");
Expect(0, 3066, '\p{^Sc=taml}', "");
Expect(0, 3066, '\P{Sc=taml}', "");
Expect(1, 3066, '\P{^Sc=taml}', "");
Expect(0, 3067, '\p{Sc=taml}', "");
Expect(1, 3067, '\p{^Sc=taml}', "");
Expect(1, 3067, '\P{Sc=taml}', "");
Expect(0, 3067, '\P{^Sc=taml}', "");
Expect(1, 3066, '\p{Sc=_ Taml}', "");
Expect(0, 3066, '\p{^Sc=_ Taml}', "");
Expect(0, 3066, '\P{Sc=_ Taml}', "");
Expect(1, 3066, '\P{^Sc=_ Taml}', "");
Expect(0, 3067, '\p{Sc=_ Taml}', "");
Expect(1, 3067, '\p{^Sc=_ Taml}', "");
Expect(1, 3067, '\P{Sc=_ Taml}', "");
Expect(0, 3067, '\P{^Sc=_ Taml}', "");
Error('\p{Is_Script= :=Tamil}');
Error('\P{Is_Script= :=Tamil}');
Expect(1, 3066, '\p{Is_Script=tamil}', "");
Expect(0, 3066, '\p{^Is_Script=tamil}', "");
Expect(0, 3066, '\P{Is_Script=tamil}', "");
Expect(1, 3066, '\P{^Is_Script=tamil}', "");
Expect(0, 3067, '\p{Is_Script=tamil}', "");
Expect(1, 3067, '\p{^Is_Script=tamil}', "");
Expect(1, 3067, '\P{Is_Script=tamil}', "");
Expect(0, 3067, '\P{^Is_Script=tamil}', "");
Expect(1, 3066, '\p{Is_Script= -TAMIL}', "");
Expect(0, 3066, '\p{^Is_Script= -TAMIL}', "");
Expect(0, 3066, '\P{Is_Script= -TAMIL}', "");
Expect(1, 3066, '\P{^Is_Script= -TAMIL}', "");
Expect(0, 3067, '\p{Is_Script= -TAMIL}', "");
Expect(1, 3067, '\p{^Is_Script= -TAMIL}', "");
Expect(1, 3067, '\P{Is_Script= -TAMIL}', "");
Expect(0, 3067, '\P{^Is_Script= -TAMIL}', "");
Error('\p{Is_Sc=__TAML:=}');
Error('\P{Is_Sc=__TAML:=}');
Expect(1, 3066, '\p{Is_Sc=taml}', "");
Expect(0, 3066, '\p{^Is_Sc=taml}', "");
Expect(0, 3066, '\P{Is_Sc=taml}', "");
Expect(1, 3066, '\P{^Is_Sc=taml}', "");
Expect(0, 3067, '\p{Is_Sc=taml}', "");
Expect(1, 3067, '\p{^Is_Sc=taml}', "");
Expect(1, 3067, '\P{Is_Sc=taml}', "");
Expect(0, 3067, '\P{^Is_Sc=taml}', "");
Expect(1, 3066, '\p{Is_Sc:--Taml}', "");
Expect(0, 3066, '\p{^Is_Sc:--Taml}', "");
Expect(0, 3066, '\P{Is_Sc:--Taml}', "");
Expect(1, 3066, '\P{^Is_Sc:--Taml}', "");
Expect(0, 3067, '\p{Is_Sc:--Taml}', "");
Expect(1, 3067, '\p{^Is_Sc:--Taml}', "");
Expect(1, 3067, '\P{Is_Sc:--Taml}', "");
Expect(0, 3067, '\P{^Is_Sc:--Taml}', "");
Error('\p{Script=:=_Tai_Viet}');
Error('\P{Script=:=_Tai_Viet}');
Expect(1, 43743, '\p{Script=taiviet}', "");
Expect(0, 43743, '\p{^Script=taiviet}', "");
Expect(0, 43743, '\P{Script=taiviet}', "");
Expect(1, 43743, '\P{^Script=taiviet}', "");
Expect(0, 43744, '\p{Script=taiviet}', "");
Expect(1, 43744, '\p{^Script=taiviet}', "");
Expect(1, 43744, '\P{Script=taiviet}', "");
Expect(0, 43744, '\P{^Script=taiviet}', "");
Expect(1, 43743, '\p{Script=	 Tai_Viet}', "");
Expect(0, 43743, '\p{^Script=	 Tai_Viet}', "");
Expect(0, 43743, '\P{Script=	 Tai_Viet}', "");
Expect(1, 43743, '\P{^Script=	 Tai_Viet}', "");
Expect(0, 43744, '\p{Script=	 Tai_Viet}', "");
Expect(1, 43744, '\p{^Script=	 Tai_Viet}', "");
Expect(1, 43744, '\P{Script=	 Tai_Viet}', "");
Expect(0, 43744, '\P{^Script=	 Tai_Viet}', "");
Error('\p{Sc: /a/	TAVT}');
Error('\P{Sc: /a/	TAVT}');
Expect(1, 43743, '\p{Sc=tavt}', "");
Expect(0, 43743, '\p{^Sc=tavt}', "");
Expect(0, 43743, '\P{Sc=tavt}', "");
Expect(1, 43743, '\P{^Sc=tavt}', "");
Expect(0, 43744, '\p{Sc=tavt}', "");
Expect(1, 43744, '\p{^Sc=tavt}', "");
Expect(1, 43744, '\P{Sc=tavt}', "");
Expect(0, 43744, '\P{^Sc=tavt}', "");
Expect(1, 43743, '\p{Sc:    	tavt}', "");
Expect(0, 43743, '\p{^Sc:    	tavt}', "");
Expect(0, 43743, '\P{Sc:    	tavt}', "");
Expect(1, 43743, '\P{^Sc:    	tavt}', "");
Expect(0, 43744, '\p{Sc:    	tavt}', "");
Expect(1, 43744, '\p{^Sc:    	tavt}', "");
Expect(1, 43744, '\P{Sc:    	tavt}', "");
Expect(0, 43744, '\P{^Sc:    	tavt}', "");
Error('\p{Is_Script=-/a/Tai_viet}');
Error('\P{Is_Script=-/a/Tai_viet}');
Expect(1, 43743, '\p{Is_Script: taiviet}', "");
Expect(0, 43743, '\p{^Is_Script: taiviet}', "");
Expect(0, 43743, '\P{Is_Script: taiviet}', "");
Expect(1, 43743, '\P{^Is_Script: taiviet}', "");
Expect(0, 43744, '\p{Is_Script: taiviet}', "");
Expect(1, 43744, '\p{^Is_Script: taiviet}', "");
Expect(1, 43744, '\P{Is_Script: taiviet}', "");
Expect(0, 43744, '\P{^Is_Script: taiviet}', "");
Expect(1, 43743, '\p{Is_Script=_	Tai_Viet}', "");
Expect(0, 43743, '\p{^Is_Script=_	Tai_Viet}', "");
Expect(0, 43743, '\P{Is_Script=_	Tai_Viet}', "");
Expect(1, 43743, '\P{^Is_Script=_	Tai_Viet}', "");
Expect(0, 43744, '\p{Is_Script=_	Tai_Viet}', "");
Expect(1, 43744, '\p{^Is_Script=_	Tai_Viet}', "");
Expect(1, 43744, '\P{Is_Script=_	Tai_Viet}', "");
Expect(0, 43744, '\P{^Is_Script=_	Tai_Viet}', "");
Error('\p{Is_Sc=	/a/tavt}');
Error('\P{Is_Sc=	/a/tavt}');
Expect(1, 43743, '\p{Is_Sc=tavt}', "");
Expect(0, 43743, '\p{^Is_Sc=tavt}', "");
Expect(0, 43743, '\P{Is_Sc=tavt}', "");
Expect(1, 43743, '\P{^Is_Sc=tavt}', "");
Expect(0, 43744, '\p{Is_Sc=tavt}', "");
Expect(1, 43744, '\p{^Is_Sc=tavt}', "");
Expect(1, 43744, '\P{Is_Sc=tavt}', "");
Expect(0, 43744, '\P{^Is_Sc=tavt}', "");
Expect(1, 43743, '\p{Is_Sc=	Tavt}', "");
Expect(0, 43743, '\p{^Is_Sc=	Tavt}', "");
Expect(0, 43743, '\P{Is_Sc=	Tavt}', "");
Expect(1, 43743, '\P{^Is_Sc=	Tavt}', "");
Expect(0, 43744, '\p{Is_Sc=	Tavt}', "");
Expect(1, 43744, '\p{^Is_Sc=	Tavt}', "");
Expect(1, 43744, '\P{Is_Sc=	Tavt}', "");
Expect(0, 43744, '\P{^Is_Sc=	Tavt}', "");
Error('\p{Script=/a/TELUGU}');
Error('\P{Script=/a/TELUGU}');
Expect(1, 3199, '\p{Script=telugu}', "");
Expect(0, 3199, '\p{^Script=telugu}', "");
Expect(0, 3199, '\P{Script=telugu}', "");
Expect(1, 3199, '\P{^Script=telugu}', "");
Expect(0, 3200, '\p{Script=telugu}', "");
Expect(1, 3200, '\p{^Script=telugu}', "");
Expect(1, 3200, '\P{Script=telugu}', "");
Expect(0, 3200, '\P{^Script=telugu}', "");
Expect(1, 3199, '\p{Script=_Telugu}', "");
Expect(0, 3199, '\p{^Script=_Telugu}', "");
Expect(0, 3199, '\P{Script=_Telugu}', "");
Expect(1, 3199, '\P{^Script=_Telugu}', "");
Expect(0, 3200, '\p{Script=_Telugu}', "");
Expect(1, 3200, '\p{^Script=_Telugu}', "");
Expect(1, 3200, '\P{Script=_Telugu}', "");
Expect(0, 3200, '\P{^Script=_Telugu}', "");
Error('\p{Sc= /a/telu}');
Error('\P{Sc= /a/telu}');
Expect(1, 3199, '\p{Sc=telu}', "");
Expect(0, 3199, '\p{^Sc=telu}', "");
Expect(0, 3199, '\P{Sc=telu}', "");
Expect(1, 3199, '\P{^Sc=telu}', "");
Expect(0, 3200, '\p{Sc=telu}', "");
Expect(1, 3200, '\p{^Sc=telu}', "");
Expect(1, 3200, '\P{Sc=telu}', "");
Expect(0, 3200, '\P{^Sc=telu}', "");
Expect(1, 3199, '\p{Sc=  TELU}', "");
Expect(0, 3199, '\p{^Sc=  TELU}', "");
Expect(0, 3199, '\P{Sc=  TELU}', "");
Expect(1, 3199, '\P{^Sc=  TELU}', "");
Expect(0, 3200, '\p{Sc=  TELU}', "");
Expect(1, 3200, '\p{^Sc=  TELU}', "");
Expect(1, 3200, '\P{Sc=  TELU}', "");
Expect(0, 3200, '\P{^Sc=  TELU}', "");
Error('\p{Is_Script=/a/_Telugu}');
Error('\P{Is_Script=/a/_Telugu}');
Expect(1, 3199, '\p{Is_Script=telugu}', "");
Expect(0, 3199, '\p{^Is_Script=telugu}', "");
Expect(0, 3199, '\P{Is_Script=telugu}', "");
Expect(1, 3199, '\P{^Is_Script=telugu}', "");
Expect(0, 3200, '\p{Is_Script=telugu}', "");
Expect(1, 3200, '\p{^Is_Script=telugu}', "");
Expect(1, 3200, '\P{Is_Script=telugu}', "");
Expect(0, 3200, '\P{^Is_Script=telugu}', "");
Expect(1, 3199, '\p{Is_Script=	_Telugu}', "");
Expect(0, 3199, '\p{^Is_Script=	_Telugu}', "");
Expect(0, 3199, '\P{Is_Script=	_Telugu}', "");
Expect(1, 3199, '\P{^Is_Script=	_Telugu}', "");
Expect(0, 3200, '\p{Is_Script=	_Telugu}', "");
Expect(1, 3200, '\p{^Is_Script=	_Telugu}', "");
Expect(1, 3200, '\P{Is_Script=	_Telugu}', "");
Expect(0, 3200, '\P{^Is_Script=	_Telugu}', "");
Error('\p{Is_Sc=:=-telu}');
Error('\P{Is_Sc=:=-telu}');
Expect(1, 3199, '\p{Is_Sc=telu}', "");
Expect(0, 3199, '\p{^Is_Sc=telu}', "");
Expect(0, 3199, '\P{Is_Sc=telu}', "");
Expect(1, 3199, '\P{^Is_Sc=telu}', "");
Expect(0, 3200, '\p{Is_Sc=telu}', "");
Expect(1, 3200, '\p{^Is_Sc=telu}', "");
Expect(1, 3200, '\P{Is_Sc=telu}', "");
Expect(0, 3200, '\P{^Is_Sc=telu}', "");
Expect(1, 3199, '\p{Is_Sc: _	TELU}', "");
Expect(0, 3199, '\p{^Is_Sc: _	TELU}', "");
Expect(0, 3199, '\P{Is_Sc: _	TELU}', "");
Expect(1, 3199, '\P{^Is_Sc: _	TELU}', "");
Expect(0, 3200, '\p{Is_Sc: _	TELU}', "");
Expect(1, 3200, '\p{^Is_Sc: _	TELU}', "");
Expect(1, 3200, '\P{Is_Sc: _	TELU}', "");
Expect(0, 3200, '\P{^Is_Sc: _	TELU}', "");
Error('\p{Script:/a/_tifinagh}');
Error('\P{Script:/a/_tifinagh}');
Expect(1, 11647, '\p{Script=tifinagh}', "");
Expect(0, 11647, '\p{^Script=tifinagh}', "");
Expect(0, 11647, '\P{Script=tifinagh}', "");
Expect(1, 11647, '\P{^Script=tifinagh}', "");
Expect(0, 11648, '\p{Script=tifinagh}', "");
Expect(1, 11648, '\p{^Script=tifinagh}', "");
Expect(1, 11648, '\P{Script=tifinagh}', "");
Expect(0, 11648, '\P{^Script=tifinagh}', "");
Expect(1, 11647, '\p{Script=-tifinagh}', "");
Expect(0, 11647, '\p{^Script=-tifinagh}', "");
Expect(0, 11647, '\P{Script=-tifinagh}', "");
Expect(1, 11647, '\P{^Script=-tifinagh}', "");
Expect(0, 11648, '\p{Script=-tifinagh}', "");
Expect(1, 11648, '\p{^Script=-tifinagh}', "");
Expect(1, 11648, '\P{Script=-tifinagh}', "");
Expect(0, 11648, '\P{^Script=-tifinagh}', "");
Error('\p{Sc=/a/ Tfng}');
Error('\P{Sc=/a/ Tfng}');
Expect(1, 11647, '\p{Sc=tfng}', "");
Expect(0, 11647, '\p{^Sc=tfng}', "");
Expect(0, 11647, '\P{Sc=tfng}', "");
Expect(1, 11647, '\P{^Sc=tfng}', "");
Expect(0, 11648, '\p{Sc=tfng}', "");
Expect(1, 11648, '\p{^Sc=tfng}', "");
Expect(1, 11648, '\P{Sc=tfng}', "");
Expect(0, 11648, '\P{^Sc=tfng}', "");
Expect(1, 11647, '\p{Sc=	TFNG}', "");
Expect(0, 11647, '\p{^Sc=	TFNG}', "");
Expect(0, 11647, '\P{Sc=	TFNG}', "");
Expect(1, 11647, '\P{^Sc=	TFNG}', "");
Expect(0, 11648, '\p{Sc=	TFNG}', "");
Expect(1, 11648, '\p{^Sc=	TFNG}', "");
Expect(1, 11648, '\P{Sc=	TFNG}', "");
Expect(0, 11648, '\P{^Sc=	TFNG}', "");
Error('\p{Is_Script=-:=tifinagh}');
Error('\P{Is_Script=-:=tifinagh}');
Expect(1, 11647, '\p{Is_Script=tifinagh}', "");
Expect(0, 11647, '\p{^Is_Script=tifinagh}', "");
Expect(0, 11647, '\P{Is_Script=tifinagh}', "");
Expect(1, 11647, '\P{^Is_Script=tifinagh}', "");
Expect(0, 11648, '\p{Is_Script=tifinagh}', "");
Expect(1, 11648, '\p{^Is_Script=tifinagh}', "");
Expect(1, 11648, '\P{Is_Script=tifinagh}', "");
Expect(0, 11648, '\P{^Is_Script=tifinagh}', "");
Expect(1, 11647, '\p{Is_Script=-	TIFINAGH}', "");
Expect(0, 11647, '\p{^Is_Script=-	TIFINAGH}', "");
Expect(0, 11647, '\P{Is_Script=-	TIFINAGH}', "");
Expect(1, 11647, '\P{^Is_Script=-	TIFINAGH}', "");
Expect(0, 11648, '\p{Is_Script=-	TIFINAGH}', "");
Expect(1, 11648, '\p{^Is_Script=-	TIFINAGH}', "");
Expect(1, 11648, '\P{Is_Script=-	TIFINAGH}', "");
Expect(0, 11648, '\P{^Is_Script=-	TIFINAGH}', "");
Error('\p{Is_Sc=/a/  tfng}');
Error('\P{Is_Sc=/a/  tfng}');
Expect(1, 11647, '\p{Is_Sc=tfng}', "");
Expect(0, 11647, '\p{^Is_Sc=tfng}', "");
Expect(0, 11647, '\P{Is_Sc=tfng}', "");
Expect(1, 11647, '\P{^Is_Sc=tfng}', "");
Expect(0, 11648, '\p{Is_Sc=tfng}', "");
Expect(1, 11648, '\p{^Is_Sc=tfng}', "");
Expect(1, 11648, '\P{Is_Sc=tfng}', "");
Expect(0, 11648, '\P{^Is_Sc=tfng}', "");
Expect(1, 11647, '\p{Is_Sc=-Tfng}', "");
Expect(0, 11647, '\p{^Is_Sc=-Tfng}', "");
Expect(0, 11647, '\P{Is_Sc=-Tfng}', "");
Expect(1, 11647, '\P{^Is_Sc=-Tfng}', "");
Expect(0, 11648, '\p{Is_Sc=-Tfng}', "");
Expect(1, 11648, '\p{^Is_Sc=-Tfng}', "");
Expect(1, 11648, '\P{Is_Sc=-Tfng}', "");
Expect(0, 11648, '\P{^Is_Sc=-Tfng}', "");
Error('\p{Script=:=__Tagalog}');
Error('\P{Script=:=__Tagalog}');
Expect(1, 5908, '\p{Script=tagalog}', "");
Expect(0, 5908, '\p{^Script=tagalog}', "");
Expect(0, 5908, '\P{Script=tagalog}', "");
Expect(1, 5908, '\P{^Script=tagalog}', "");
Expect(0, 5909, '\p{Script=tagalog}', "");
Expect(1, 5909, '\p{^Script=tagalog}', "");
Expect(1, 5909, '\P{Script=tagalog}', "");
Expect(0, 5909, '\P{^Script=tagalog}', "");
Expect(1, 5908, '\p{Script=		tagalog}', "");
Expect(0, 5908, '\p{^Script=		tagalog}', "");
Expect(0, 5908, '\P{Script=		tagalog}', "");
Expect(1, 5908, '\P{^Script=		tagalog}', "");
Expect(0, 5909, '\p{Script=		tagalog}', "");
Expect(1, 5909, '\p{^Script=		tagalog}', "");
Expect(1, 5909, '\P{Script=		tagalog}', "");
Expect(0, 5909, '\P{^Script=		tagalog}', "");
Error('\p{Sc=:=_ Tglg}');
Error('\P{Sc=:=_ Tglg}');
Expect(1, 5908, '\p{Sc:tglg}', "");
Expect(0, 5908, '\p{^Sc:tglg}', "");
Expect(0, 5908, '\P{Sc:tglg}', "");
Expect(1, 5908, '\P{^Sc:tglg}', "");
Expect(0, 5909, '\p{Sc:tglg}', "");
Expect(1, 5909, '\p{^Sc:tglg}', "");
Expect(1, 5909, '\P{Sc:tglg}', "");
Expect(0, 5909, '\P{^Sc:tglg}', "");
Expect(1, 5908, '\p{Sc=-	Tglg}', "");
Expect(0, 5908, '\p{^Sc=-	Tglg}', "");
Expect(0, 5908, '\P{Sc=-	Tglg}', "");
Expect(1, 5908, '\P{^Sc=-	Tglg}', "");
Expect(0, 5909, '\p{Sc=-	Tglg}', "");
Expect(1, 5909, '\p{^Sc=-	Tglg}', "");
Expect(1, 5909, '\P{Sc=-	Tglg}', "");
Expect(0, 5909, '\P{^Sc=-	Tglg}', "");
Error('\p{Is_Script:   := -TAGALOG}');
Error('\P{Is_Script:   := -TAGALOG}');
Expect(1, 5908, '\p{Is_Script=tagalog}', "");
Expect(0, 5908, '\p{^Is_Script=tagalog}', "");
Expect(0, 5908, '\P{Is_Script=tagalog}', "");
Expect(1, 5908, '\P{^Is_Script=tagalog}', "");
Expect(0, 5909, '\p{Is_Script=tagalog}', "");
Expect(1, 5909, '\p{^Is_Script=tagalog}', "");
Expect(1, 5909, '\P{Is_Script=tagalog}', "");
Expect(0, 5909, '\P{^Is_Script=tagalog}', "");
Expect(1, 5908, '\p{Is_Script=	 tagalog}', "");
Expect(0, 5908, '\p{^Is_Script=	 tagalog}', "");
Expect(0, 5908, '\P{Is_Script=	 tagalog}', "");
Expect(1, 5908, '\P{^Is_Script=	 tagalog}', "");
Expect(0, 5909, '\p{Is_Script=	 tagalog}', "");
Expect(1, 5909, '\p{^Is_Script=	 tagalog}', "");
Expect(1, 5909, '\P{Is_Script=	 tagalog}', "");
Expect(0, 5909, '\P{^Is_Script=	 tagalog}', "");
Error('\p{Is_Sc=/a/-tglg}');
Error('\P{Is_Sc=/a/-tglg}');
Expect(1, 5908, '\p{Is_Sc=tglg}', "");
Expect(0, 5908, '\p{^Is_Sc=tglg}', "");
Expect(0, 5908, '\P{Is_Sc=tglg}', "");
Expect(1, 5908, '\P{^Is_Sc=tglg}', "");
Expect(0, 5909, '\p{Is_Sc=tglg}', "");
Expect(1, 5909, '\p{^Is_Sc=tglg}', "");
Expect(1, 5909, '\P{Is_Sc=tglg}', "");
Expect(0, 5909, '\P{^Is_Sc=tglg}', "");
Expect(1, 5908, '\p{Is_Sc=_-tglg}', "");
Expect(0, 5908, '\p{^Is_Sc=_-tglg}', "");
Expect(0, 5908, '\P{Is_Sc=_-tglg}', "");
Expect(1, 5908, '\P{^Is_Sc=_-tglg}', "");
Expect(0, 5909, '\p{Is_Sc=_-tglg}', "");
Expect(1, 5909, '\p{^Is_Sc=_-tglg}', "");
Expect(1, 5909, '\P{Is_Sc=_-tglg}', "");
Expect(0, 5909, '\P{^Is_Sc=_-tglg}', "");
Error('\p{Script=:=  THAANA}');
Error('\P{Script=:=  THAANA}');
Expect(1, 1969, '\p{Script=thaana}', "");
Expect(0, 1969, '\p{^Script=thaana}', "");
Expect(0, 1969, '\P{Script=thaana}', "");
Expect(1, 1969, '\P{^Script=thaana}', "");
Expect(0, 1970, '\p{Script=thaana}', "");
Expect(1, 1970, '\p{^Script=thaana}', "");
Expect(1, 1970, '\P{Script=thaana}', "");
Expect(0, 1970, '\P{^Script=thaana}', "");
Expect(1, 1969, '\p{Script= Thaana}', "");
Expect(0, 1969, '\p{^Script= Thaana}', "");
Expect(0, 1969, '\P{Script= Thaana}', "");
Expect(1, 1969, '\P{^Script= Thaana}', "");
Expect(0, 1970, '\p{Script= Thaana}', "");
Expect(1, 1970, '\p{^Script= Thaana}', "");
Expect(1, 1970, '\P{Script= Thaana}', "");
Expect(0, 1970, '\P{^Script= Thaana}', "");
Error('\p{Sc=/a/-thaa}');
Error('\P{Sc=/a/-thaa}');
Expect(1, 1969, '\p{Sc=thaa}', "");
Expect(0, 1969, '\p{^Sc=thaa}', "");
Expect(0, 1969, '\P{Sc=thaa}', "");
Expect(1, 1969, '\P{^Sc=thaa}', "");
Expect(0, 1970, '\p{Sc=thaa}', "");
Expect(1, 1970, '\p{^Sc=thaa}', "");
Expect(1, 1970, '\P{Sc=thaa}', "");
Expect(0, 1970, '\P{^Sc=thaa}', "");
Expect(1, 1969, '\p{Sc=_THAA}', "");
Expect(0, 1969, '\p{^Sc=_THAA}', "");
Expect(0, 1969, '\P{Sc=_THAA}', "");
Expect(1, 1969, '\P{^Sc=_THAA}', "");
Expect(0, 1970, '\p{Sc=_THAA}', "");
Expect(1, 1970, '\p{^Sc=_THAA}', "");
Expect(1, 1970, '\P{Sc=_THAA}', "");
Expect(0, 1970, '\P{^Sc=_THAA}', "");
Error('\p{Is_Script=_thaana/a/}');
Error('\P{Is_Script=_thaana/a/}');
Expect(1, 1969, '\p{Is_Script=thaana}', "");
Expect(0, 1969, '\p{^Is_Script=thaana}', "");
Expect(0, 1969, '\P{Is_Script=thaana}', "");
Expect(1, 1969, '\P{^Is_Script=thaana}', "");
Expect(0, 1970, '\p{Is_Script=thaana}', "");
Expect(1, 1970, '\p{^Is_Script=thaana}', "");
Expect(1, 1970, '\P{Is_Script=thaana}', "");
Expect(0, 1970, '\P{^Is_Script=thaana}', "");
Expect(1, 1969, '\p{Is_Script=		THAANA}', "");
Expect(0, 1969, '\p{^Is_Script=		THAANA}', "");
Expect(0, 1969, '\P{Is_Script=		THAANA}', "");
Expect(1, 1969, '\P{^Is_Script=		THAANA}', "");
Expect(0, 1970, '\p{Is_Script=		THAANA}', "");
Expect(1, 1970, '\p{^Is_Script=		THAANA}', "");
Expect(1, 1970, '\P{Is_Script=		THAANA}', "");
Expect(0, 1970, '\P{^Is_Script=		THAANA}', "");
Error('\p{Is_Sc= /a/thaa}');
Error('\P{Is_Sc= /a/thaa}');
Expect(1, 1969, '\p{Is_Sc=thaa}', "");
Expect(0, 1969, '\p{^Is_Sc=thaa}', "");
Expect(0, 1969, '\P{Is_Sc=thaa}', "");
Expect(1, 1969, '\P{^Is_Sc=thaa}', "");
Expect(0, 1970, '\p{Is_Sc=thaa}', "");
Expect(1, 1970, '\p{^Is_Sc=thaa}', "");
Expect(1, 1970, '\P{Is_Sc=thaa}', "");
Expect(0, 1970, '\P{^Is_Sc=thaa}', "");
Expect(1, 1969, '\p{Is_Sc:	  THAA}', "");
Expect(0, 1969, '\p{^Is_Sc:	  THAA}', "");
Expect(0, 1969, '\P{Is_Sc:	  THAA}', "");
Expect(1, 1969, '\P{^Is_Sc:	  THAA}', "");
Expect(0, 1970, '\p{Is_Sc:	  THAA}', "");
Expect(1, 1970, '\p{^Is_Sc:	  THAA}', "");
Expect(1, 1970, '\P{Is_Sc:	  THAA}', "");
Expect(0, 1970, '\P{^Is_Sc:	  THAA}', "");
Error('\p{Script=_ thai/a/}');
Error('\P{Script=_ thai/a/}');
Expect(1, 3675, '\p{Script=thai}', "");
Expect(0, 3675, '\p{^Script=thai}', "");
Expect(0, 3675, '\P{Script=thai}', "");
Expect(1, 3675, '\P{^Script=thai}', "");
Expect(0, 3676, '\p{Script=thai}', "");
Expect(1, 3676, '\p{^Script=thai}', "");
Expect(1, 3676, '\P{Script=thai}', "");
Expect(0, 3676, '\P{^Script=thai}', "");
Expect(1, 3675, '\p{Script= _THAI}', "");
Expect(0, 3675, '\p{^Script= _THAI}', "");
Expect(0, 3675, '\P{Script= _THAI}', "");
Expect(1, 3675, '\P{^Script= _THAI}', "");
Expect(0, 3676, '\p{Script= _THAI}', "");
Expect(1, 3676, '\p{^Script= _THAI}', "");
Expect(1, 3676, '\P{Script= _THAI}', "");
Expect(0, 3676, '\P{^Script= _THAI}', "");
Error('\p{Sc:   /a/_	Thai}');
Error('\P{Sc:   /a/_	Thai}');
Expect(1, 3675, '\p{Sc=thai}', "");
Expect(0, 3675, '\p{^Sc=thai}', "");
Expect(0, 3675, '\P{Sc=thai}', "");
Expect(1, 3675, '\P{^Sc=thai}', "");
Expect(0, 3676, '\p{Sc=thai}', "");
Expect(1, 3676, '\p{^Sc=thai}', "");
Expect(1, 3676, '\P{Sc=thai}', "");
Expect(0, 3676, '\P{^Sc=thai}', "");
Expect(1, 3675, '\p{Sc=THAI}', "");
Expect(0, 3675, '\p{^Sc=THAI}', "");
Expect(0, 3675, '\P{Sc=THAI}', "");
Expect(1, 3675, '\P{^Sc=THAI}', "");
Expect(0, 3676, '\p{Sc=THAI}', "");
Expect(1, 3676, '\p{^Sc=THAI}', "");
Expect(1, 3676, '\P{Sc=THAI}', "");
Expect(0, 3676, '\P{^Sc=THAI}', "");
Error('\p{Is_Script=_:=Thai}');
Error('\P{Is_Script=_:=Thai}');
Expect(1, 3675, '\p{Is_Script=thai}', "");
Expect(0, 3675, '\p{^Is_Script=thai}', "");
Expect(0, 3675, '\P{Is_Script=thai}', "");
Expect(1, 3675, '\P{^Is_Script=thai}', "");
Expect(0, 3676, '\p{Is_Script=thai}', "");
Expect(1, 3676, '\p{^Is_Script=thai}', "");
Expect(1, 3676, '\P{Is_Script=thai}', "");
Expect(0, 3676, '\P{^Is_Script=thai}', "");
Expect(1, 3675, '\p{Is_Script=_ Thai}', "");
Expect(0, 3675, '\p{^Is_Script=_ Thai}', "");
Expect(0, 3675, '\P{Is_Script=_ Thai}', "");
Expect(1, 3675, '\P{^Is_Script=_ Thai}', "");
Expect(0, 3676, '\p{Is_Script=_ Thai}', "");
Expect(1, 3676, '\p{^Is_Script=_ Thai}', "");
Expect(1, 3676, '\P{Is_Script=_ Thai}', "");
Expect(0, 3676, '\P{^Is_Script=_ Thai}', "");
Error('\p{Is_Sc=/a/ -Thai}');
Error('\P{Is_Sc=/a/ -Thai}');
Expect(1, 3675, '\p{Is_Sc=thai}', "");
Expect(0, 3675, '\p{^Is_Sc=thai}', "");
Expect(0, 3675, '\P{Is_Sc=thai}', "");
Expect(1, 3675, '\P{^Is_Sc=thai}', "");
Expect(0, 3676, '\p{Is_Sc=thai}', "");
Expect(1, 3676, '\p{^Is_Sc=thai}', "");
Expect(1, 3676, '\P{Is_Sc=thai}', "");
Expect(0, 3676, '\P{^Is_Sc=thai}', "");
Expect(1, 3675, '\p{Is_Sc=_Thai}', "");
Expect(0, 3675, '\p{^Is_Sc=_Thai}', "");
Expect(0, 3675, '\P{Is_Sc=_Thai}', "");
Expect(1, 3675, '\P{^Is_Sc=_Thai}', "");
Expect(0, 3676, '\p{Is_Sc=_Thai}', "");
Expect(1, 3676, '\p{^Is_Sc=_Thai}', "");
Expect(1, 3676, '\P{Is_Sc=_Thai}', "");
Expect(0, 3676, '\P{^Is_Sc=_Thai}', "");
Error('\p{Script:   :=_tibetan}');
Error('\P{Script:   :=_tibetan}');
Expect(1, 4058, '\p{Script: tibetan}', "");
Expect(0, 4058, '\p{^Script: tibetan}', "");
Expect(0, 4058, '\P{Script: tibetan}', "");
Expect(1, 4058, '\P{^Script: tibetan}', "");
Expect(0, 4059, '\p{Script: tibetan}', "");
Expect(1, 4059, '\p{^Script: tibetan}', "");
Expect(1, 4059, '\P{Script: tibetan}', "");
Expect(0, 4059, '\P{^Script: tibetan}', "");
Expect(1, 4058, '\p{Script=__Tibetan}', "");
Expect(0, 4058, '\p{^Script=__Tibetan}', "");
Expect(0, 4058, '\P{Script=__Tibetan}', "");
Expect(1, 4058, '\P{^Script=__Tibetan}', "");
Expect(0, 4059, '\p{Script=__Tibetan}', "");
Expect(1, 4059, '\p{^Script=__Tibetan}', "");
Expect(1, 4059, '\P{Script=__Tibetan}', "");
Expect(0, 4059, '\P{^Script=__Tibetan}', "");
Error('\p{Sc=__Tibt/a/}');
Error('\P{Sc=__Tibt/a/}');
Expect(1, 4058, '\p{Sc=tibt}', "");
Expect(0, 4058, '\p{^Sc=tibt}', "");
Expect(0, 4058, '\P{Sc=tibt}', "");
Expect(1, 4058, '\P{^Sc=tibt}', "");
Expect(0, 4059, '\p{Sc=tibt}', "");
Expect(1, 4059, '\p{^Sc=tibt}', "");
Expect(1, 4059, '\P{Sc=tibt}', "");
Expect(0, 4059, '\P{^Sc=tibt}', "");
Expect(1, 4058, '\p{Sc=	tibt}', "");
Expect(0, 4058, '\p{^Sc=	tibt}', "");
Expect(0, 4058, '\P{Sc=	tibt}', "");
Expect(1, 4058, '\P{^Sc=	tibt}', "");
Expect(0, 4059, '\p{Sc=	tibt}', "");
Expect(1, 4059, '\p{^Sc=	tibt}', "");
Expect(1, 4059, '\P{Sc=	tibt}', "");
Expect(0, 4059, '\P{^Sc=	tibt}', "");
Error('\p{Is_Script=-/a/Tibetan}');
Error('\P{Is_Script=-/a/Tibetan}');
Expect(1, 4058, '\p{Is_Script:	tibetan}', "");
Expect(0, 4058, '\p{^Is_Script:	tibetan}', "");
Expect(0, 4058, '\P{Is_Script:	tibetan}', "");
Expect(1, 4058, '\P{^Is_Script:	tibetan}', "");
Expect(0, 4059, '\p{Is_Script:	tibetan}', "");
Expect(1, 4059, '\p{^Is_Script:	tibetan}', "");
Expect(1, 4059, '\P{Is_Script:	tibetan}', "");
Expect(0, 4059, '\P{^Is_Script:	tibetan}', "");
Expect(1, 4058, '\p{Is_Script=	_Tibetan}', "");
Expect(0, 4058, '\p{^Is_Script=	_Tibetan}', "");
Expect(0, 4058, '\P{Is_Script=	_Tibetan}', "");
Expect(1, 4058, '\P{^Is_Script=	_Tibetan}', "");
Expect(0, 4059, '\p{Is_Script=	_Tibetan}', "");
Expect(1, 4059, '\p{^Is_Script=	_Tibetan}', "");
Expect(1, 4059, '\P{Is_Script=	_Tibetan}', "");
Expect(0, 4059, '\P{^Is_Script=	_Tibetan}', "");
Error('\p{Is_Sc=	_Tibt/a/}');
Error('\P{Is_Sc=	_Tibt/a/}');
Expect(1, 4058, '\p{Is_Sc=tibt}', "");
Expect(0, 4058, '\p{^Is_Sc=tibt}', "");
Expect(0, 4058, '\P{Is_Sc=tibt}', "");
Expect(1, 4058, '\P{^Is_Sc=tibt}', "");
Expect(0, 4059, '\p{Is_Sc=tibt}', "");
Expect(1, 4059, '\p{^Is_Sc=tibt}', "");
Expect(1, 4059, '\P{Is_Sc=tibt}', "");
Expect(0, 4059, '\P{^Is_Sc=tibt}', "");
Expect(1, 4058, '\p{Is_Sc=--Tibt}', "");
Expect(0, 4058, '\p{^Is_Sc=--Tibt}', "");
Expect(0, 4058, '\P{Is_Sc=--Tibt}', "");
Expect(1, 4058, '\P{^Is_Sc=--Tibt}', "");
Expect(0, 4059, '\p{Is_Sc=--Tibt}', "");
Expect(1, 4059, '\p{^Is_Sc=--Tibt}', "");
Expect(1, 4059, '\P{Is_Sc=--Tibt}', "");
Expect(0, 4059, '\P{^Is_Sc=--Tibt}', "");
Error('\p{Script=-/a/ugaritic}');
Error('\P{Script=-/a/ugaritic}');
Expect(1, 66463, '\p{Script=ugaritic}', "");
Expect(0, 66463, '\p{^Script=ugaritic}', "");
Expect(0, 66463, '\P{Script=ugaritic}', "");
Expect(1, 66463, '\P{^Script=ugaritic}', "");
Expect(0, 66464, '\p{Script=ugaritic}', "");
Expect(1, 66464, '\p{^Script=ugaritic}', "");
Expect(1, 66464, '\P{Script=ugaritic}', "");
Expect(0, 66464, '\P{^Script=ugaritic}', "");
Expect(1, 66463, '\p{Script= ugaritic}', "");
Expect(0, 66463, '\p{^Script= ugaritic}', "");
Expect(0, 66463, '\P{Script= ugaritic}', "");
Expect(1, 66463, '\P{^Script= ugaritic}', "");
Expect(0, 66464, '\p{Script= ugaritic}', "");
Expect(1, 66464, '\p{^Script= ugaritic}', "");
Expect(1, 66464, '\P{Script= ugaritic}', "");
Expect(0, 66464, '\P{^Script= ugaritic}', "");
Error('\p{Sc=:=ugar}');
Error('\P{Sc=:=ugar}');
Expect(1, 66463, '\p{Sc=ugar}', "");
Expect(0, 66463, '\p{^Sc=ugar}', "");
Expect(0, 66463, '\P{Sc=ugar}', "");
Expect(1, 66463, '\P{^Sc=ugar}', "");
Expect(0, 66464, '\p{Sc=ugar}', "");
Expect(1, 66464, '\p{^Sc=ugar}', "");
Expect(1, 66464, '\P{Sc=ugar}', "");
Expect(0, 66464, '\P{^Sc=ugar}', "");
Expect(1, 66463, '\p{Sc=	-Ugar}', "");
Expect(0, 66463, '\p{^Sc=	-Ugar}', "");
Expect(0, 66463, '\P{Sc=	-Ugar}', "");
Expect(1, 66463, '\P{^Sc=	-Ugar}', "");
Expect(0, 66464, '\p{Sc=	-Ugar}', "");
Expect(1, 66464, '\p{^Sc=	-Ugar}', "");
Expect(1, 66464, '\P{Sc=	-Ugar}', "");
Expect(0, 66464, '\P{^Sc=	-Ugar}', "");
Error('\p{Is_Script=_/a/ugaritic}');
Error('\P{Is_Script=_/a/ugaritic}');
Expect(1, 66463, '\p{Is_Script:	ugaritic}', "");
Expect(0, 66463, '\p{^Is_Script:	ugaritic}', "");
Expect(0, 66463, '\P{Is_Script:	ugaritic}', "");
Expect(1, 66463, '\P{^Is_Script:	ugaritic}', "");
Expect(0, 66464, '\p{Is_Script:	ugaritic}', "");
Expect(1, 66464, '\p{^Is_Script:	ugaritic}', "");
Expect(1, 66464, '\P{Is_Script:	ugaritic}', "");
Expect(0, 66464, '\P{^Is_Script:	ugaritic}', "");
Expect(1, 66463, '\p{Is_Script:   -UGARITIC}', "");
Expect(0, 66463, '\p{^Is_Script:   -UGARITIC}', "");
Expect(0, 66463, '\P{Is_Script:   -UGARITIC}', "");
Expect(1, 66463, '\P{^Is_Script:   -UGARITIC}', "");
Expect(0, 66464, '\p{Is_Script:   -UGARITIC}', "");
Expect(1, 66464, '\p{^Is_Script:   -UGARITIC}', "");
Expect(1, 66464, '\P{Is_Script:   -UGARITIC}', "");
Expect(0, 66464, '\P{^Is_Script:   -UGARITIC}', "");
Error('\p{Is_Sc::=Ugar}');
Error('\P{Is_Sc::=Ugar}');
Expect(1, 66463, '\p{Is_Sc: ugar}', "");
Expect(0, 66463, '\p{^Is_Sc: ugar}', "");
Expect(0, 66463, '\P{Is_Sc: ugar}', "");
Expect(1, 66463, '\P{^Is_Sc: ugar}', "");
Expect(0, 66464, '\p{Is_Sc: ugar}', "");
Expect(1, 66464, '\p{^Is_Sc: ugar}', "");
Expect(1, 66464, '\P{Is_Sc: ugar}', "");
Expect(0, 66464, '\P{^Is_Sc: ugar}', "");
Expect(1, 66463, '\p{Is_Sc=  UGAR}', "");
Expect(0, 66463, '\p{^Is_Sc=  UGAR}', "");
Expect(0, 66463, '\P{Is_Sc=  UGAR}', "");
Expect(1, 66463, '\P{^Is_Sc=  UGAR}', "");
Expect(0, 66464, '\p{Is_Sc=  UGAR}', "");
Expect(1, 66464, '\p{^Is_Sc=  UGAR}', "");
Expect(1, 66464, '\P{Is_Sc=  UGAR}', "");
Expect(0, 66464, '\P{^Is_Sc=  UGAR}', "");
Error('\p{Script=-/a/Vai}');
Error('\P{Script=-/a/Vai}');
Expect(1, 42539, '\p{Script=vai}', "");
Expect(0, 42539, '\p{^Script=vai}', "");
Expect(0, 42539, '\P{Script=vai}', "");
Expect(1, 42539, '\P{^Script=vai}', "");
Expect(0, 42540, '\p{Script=vai}', "");
Expect(1, 42540, '\p{^Script=vai}', "");
Expect(1, 42540, '\P{Script=vai}', "");
Expect(0, 42540, '\P{^Script=vai}', "");
Expect(1, 42539, '\p{Script= VAI}', "");
Expect(0, 42539, '\p{^Script= VAI}', "");
Expect(0, 42539, '\P{Script= VAI}', "");
Expect(1, 42539, '\P{^Script= VAI}', "");
Expect(0, 42540, '\p{Script= VAI}', "");
Expect(1, 42540, '\p{^Script= VAI}', "");
Expect(1, 42540, '\P{Script= VAI}', "");
Expect(0, 42540, '\P{^Script= VAI}', "");
Error('\p{Sc=vaii/a/}');
Error('\P{Sc=vaii/a/}');
Expect(1, 42539, '\p{Sc:   vaii}', "");
Expect(0, 42539, '\p{^Sc:   vaii}', "");
Expect(0, 42539, '\P{Sc:   vaii}', "");
Expect(1, 42539, '\P{^Sc:   vaii}', "");
Expect(0, 42540, '\p{Sc:   vaii}', "");
Expect(1, 42540, '\p{^Sc:   vaii}', "");
Expect(1, 42540, '\P{Sc:   vaii}', "");
Expect(0, 42540, '\P{^Sc:   vaii}', "");
Expect(1, 42539, '\p{Sc=	_Vaii}', "");
Expect(0, 42539, '\p{^Sc=	_Vaii}', "");
Expect(0, 42539, '\P{Sc=	_Vaii}', "");
Expect(1, 42539, '\P{^Sc=	_Vaii}', "");
Expect(0, 42540, '\p{Sc=	_Vaii}', "");
Expect(1, 42540, '\p{^Sc=	_Vaii}', "");
Expect(1, 42540, '\P{Sc=	_Vaii}', "");
Expect(0, 42540, '\P{^Sc=	_Vaii}', "");
Error('\p{Is_Script=/a/	VAI}');
Error('\P{Is_Script=/a/	VAI}');
Expect(1, 42539, '\p{Is_Script=vai}', "");
Expect(0, 42539, '\p{^Is_Script=vai}', "");
Expect(0, 42539, '\P{Is_Script=vai}', "");
Expect(1, 42539, '\P{^Is_Script=vai}', "");
Expect(0, 42540, '\p{Is_Script=vai}', "");
Expect(1, 42540, '\p{^Is_Script=vai}', "");
Expect(1, 42540, '\P{Is_Script=vai}', "");
Expect(0, 42540, '\P{^Is_Script=vai}', "");
Expect(1, 42539, '\p{Is_Script=-vai}', "");
Expect(0, 42539, '\p{^Is_Script=-vai}', "");
Expect(0, 42539, '\P{Is_Script=-vai}', "");
Expect(1, 42539, '\P{^Is_Script=-vai}', "");
Expect(0, 42540, '\p{Is_Script=-vai}', "");
Expect(1, 42540, '\p{^Is_Script=-vai}', "");
Expect(1, 42540, '\P{Is_Script=-vai}', "");
Expect(0, 42540, '\P{^Is_Script=-vai}', "");
Error('\p{Is_Sc=_:=VAII}');
Error('\P{Is_Sc=_:=VAII}');
Expect(1, 42539, '\p{Is_Sc=vaii}', "");
Expect(0, 42539, '\p{^Is_Sc=vaii}', "");
Expect(0, 42539, '\P{Is_Sc=vaii}', "");
Expect(1, 42539, '\P{^Is_Sc=vaii}', "");
Expect(0, 42540, '\p{Is_Sc=vaii}', "");
Expect(1, 42540, '\p{^Is_Sc=vaii}', "");
Expect(1, 42540, '\P{Is_Sc=vaii}', "");
Expect(0, 42540, '\P{^Is_Sc=vaii}', "");
Expect(1, 42539, '\p{Is_Sc=-	VAII}', "");
Expect(0, 42539, '\p{^Is_Sc=-	VAII}', "");
Expect(0, 42539, '\P{Is_Sc=-	VAII}', "");
Expect(1, 42539, '\P{^Is_Sc=-	VAII}', "");
Expect(0, 42540, '\p{Is_Sc=-	VAII}', "");
Expect(1, 42540, '\p{^Is_Sc=-	VAII}', "");
Expect(1, 42540, '\P{Is_Sc=-	VAII}', "");
Expect(0, 42540, '\P{^Is_Sc=-	VAII}', "");
Error('\p{Script=_:=old_Persian}');
Error('\P{Script=_:=old_Persian}');
Expect(1, 66517, '\p{Script=oldpersian}', "");
Expect(0, 66517, '\p{^Script=oldpersian}', "");
Expect(0, 66517, '\P{Script=oldpersian}', "");
Expect(1, 66517, '\P{^Script=oldpersian}', "");
Expect(0, 66518, '\p{Script=oldpersian}', "");
Expect(1, 66518, '\p{^Script=oldpersian}', "");
Expect(1, 66518, '\P{Script=oldpersian}', "");
Expect(0, 66518, '\P{^Script=oldpersian}', "");
Expect(1, 66517, '\p{Script=  Old_PERSIAN}', "");
Expect(0, 66517, '\p{^Script=  Old_PERSIAN}', "");
Expect(0, 66517, '\P{Script=  Old_PERSIAN}', "");
Expect(1, 66517, '\P{^Script=  Old_PERSIAN}', "");
Expect(0, 66518, '\p{Script=  Old_PERSIAN}', "");
Expect(1, 66518, '\p{^Script=  Old_PERSIAN}', "");
Expect(1, 66518, '\P{Script=  Old_PERSIAN}', "");
Expect(0, 66518, '\P{^Script=  Old_PERSIAN}', "");
Error('\p{Sc=  Xpeo:=}');
Error('\P{Sc=  Xpeo:=}');
Expect(1, 66517, '\p{Sc=xpeo}', "");
Expect(0, 66517, '\p{^Sc=xpeo}', "");
Expect(0, 66517, '\P{Sc=xpeo}', "");
Expect(1, 66517, '\P{^Sc=xpeo}', "");
Expect(0, 66518, '\p{Sc=xpeo}', "");
Expect(1, 66518, '\p{^Sc=xpeo}', "");
Expect(1, 66518, '\P{Sc=xpeo}', "");
Expect(0, 66518, '\P{^Sc=xpeo}', "");
Expect(1, 66517, '\p{Sc=--XPEO}', "");
Expect(0, 66517, '\p{^Sc=--XPEO}', "");
Expect(0, 66517, '\P{Sc=--XPEO}', "");
Expect(1, 66517, '\P{^Sc=--XPEO}', "");
Expect(0, 66518, '\p{Sc=--XPEO}', "");
Expect(1, 66518, '\p{^Sc=--XPEO}', "");
Expect(1, 66518, '\P{Sc=--XPEO}', "");
Expect(0, 66518, '\P{^Sc=--XPEO}', "");
Error('\p{Is_Script=/a/--Old_Persian}');
Error('\P{Is_Script=/a/--Old_Persian}');
Expect(1, 66517, '\p{Is_Script=oldpersian}', "");
Expect(0, 66517, '\p{^Is_Script=oldpersian}', "");
Expect(0, 66517, '\P{Is_Script=oldpersian}', "");
Expect(1, 66517, '\P{^Is_Script=oldpersian}', "");
Expect(0, 66518, '\p{Is_Script=oldpersian}', "");
Expect(1, 66518, '\p{^Is_Script=oldpersian}', "");
Expect(1, 66518, '\P{Is_Script=oldpersian}', "");
Expect(0, 66518, '\P{^Is_Script=oldpersian}', "");
Expect(1, 66517, '\p{Is_Script=_OLD_persian}', "");
Expect(0, 66517, '\p{^Is_Script=_OLD_persian}', "");
Expect(0, 66517, '\P{Is_Script=_OLD_persian}', "");
Expect(1, 66517, '\P{^Is_Script=_OLD_persian}', "");
Expect(0, 66518, '\p{Is_Script=_OLD_persian}', "");
Expect(1, 66518, '\p{^Is_Script=_OLD_persian}', "");
Expect(1, 66518, '\P{Is_Script=_OLD_persian}', "");
Expect(0, 66518, '\P{^Is_Script=_OLD_persian}', "");
Error('\p{Is_Sc=-	Xpeo/a/}');
Error('\P{Is_Sc=-	Xpeo/a/}');
Expect(1, 66517, '\p{Is_Sc=xpeo}', "");
Expect(0, 66517, '\p{^Is_Sc=xpeo}', "");
Expect(0, 66517, '\P{Is_Sc=xpeo}', "");
Expect(1, 66517, '\P{^Is_Sc=xpeo}', "");
Expect(0, 66518, '\p{Is_Sc=xpeo}', "");
Expect(1, 66518, '\p{^Is_Sc=xpeo}', "");
Expect(1, 66518, '\P{Is_Sc=xpeo}', "");
Expect(0, 66518, '\P{^Is_Sc=xpeo}', "");
Expect(1, 66517, '\p{Is_Sc=	Xpeo}', "");
Expect(0, 66517, '\p{^Is_Sc=	Xpeo}', "");
Expect(0, 66517, '\P{Is_Sc=	Xpeo}', "");
Expect(1, 66517, '\P{^Is_Sc=	Xpeo}', "");
Expect(0, 66518, '\p{Is_Sc=	Xpeo}', "");
Expect(1, 66518, '\p{^Is_Sc=	Xpeo}', "");
Expect(1, 66518, '\P{Is_Sc=	Xpeo}', "");
Expect(0, 66518, '\P{^Is_Sc=	Xpeo}', "");
Error('\p{Script=:=Cuneiform}');
Error('\P{Script=:=Cuneiform}');
Expect(1, 74867, '\p{Script=cuneiform}', "");
Expect(0, 74867, '\p{^Script=cuneiform}', "");
Expect(0, 74867, '\P{Script=cuneiform}', "");
Expect(1, 74867, '\P{^Script=cuneiform}', "");
Expect(0, 74868, '\p{Script=cuneiform}', "");
Expect(1, 74868, '\p{^Script=cuneiform}', "");
Expect(1, 74868, '\P{Script=cuneiform}', "");
Expect(0, 74868, '\P{^Script=cuneiform}', "");
Expect(1, 74867, '\p{Script=-	Cuneiform}', "");
Expect(0, 74867, '\p{^Script=-	Cuneiform}', "");
Expect(0, 74867, '\P{Script=-	Cuneiform}', "");
Expect(1, 74867, '\P{^Script=-	Cuneiform}', "");
Expect(0, 74868, '\p{Script=-	Cuneiform}', "");
Expect(1, 74868, '\p{^Script=-	Cuneiform}', "");
Expect(1, 74868, '\P{Script=-	Cuneiform}', "");
Expect(0, 74868, '\P{^Script=-	Cuneiform}', "");
Error('\p{Sc=/a/ XSUX}');
Error('\P{Sc=/a/ XSUX}');
Expect(1, 74867, '\p{Sc=xsux}', "");
Expect(0, 74867, '\p{^Sc=xsux}', "");
Expect(0, 74867, '\P{Sc=xsux}', "");
Expect(1, 74867, '\P{^Sc=xsux}', "");
Expect(0, 74868, '\p{Sc=xsux}', "");
Expect(1, 74868, '\p{^Sc=xsux}', "");
Expect(1, 74868, '\P{Sc=xsux}', "");
Expect(0, 74868, '\P{^Sc=xsux}', "");
Expect(1, 74867, '\p{Sc=		Xsux}', "");
Expect(0, 74867, '\p{^Sc=		Xsux}', "");
Expect(0, 74867, '\P{Sc=		Xsux}', "");
Expect(1, 74867, '\P{^Sc=		Xsux}', "");
Expect(0, 74868, '\p{Sc=		Xsux}', "");
Expect(1, 74868, '\p{^Sc=		Xsux}', "");
Expect(1, 74868, '\P{Sc=		Xsux}', "");
Expect(0, 74868, '\P{^Sc=		Xsux}', "");
Error('\p{Is_Script=	Cuneiform/a/}');
Error('\P{Is_Script=	Cuneiform/a/}');
Expect(1, 74867, '\p{Is_Script=cuneiform}', "");
Expect(0, 74867, '\p{^Is_Script=cuneiform}', "");
Expect(0, 74867, '\P{Is_Script=cuneiform}', "");
Expect(1, 74867, '\P{^Is_Script=cuneiform}', "");
Expect(0, 74868, '\p{Is_Script=cuneiform}', "");
Expect(1, 74868, '\p{^Is_Script=cuneiform}', "");
Expect(1, 74868, '\P{Is_Script=cuneiform}', "");
Expect(0, 74868, '\P{^Is_Script=cuneiform}', "");
Expect(1, 74867, '\p{Is_Script=_ CUNEIFORM}', "");
Expect(0, 74867, '\p{^Is_Script=_ CUNEIFORM}', "");
Expect(0, 74867, '\P{Is_Script=_ CUNEIFORM}', "");
Expect(1, 74867, '\P{^Is_Script=_ CUNEIFORM}', "");
Expect(0, 74868, '\p{Is_Script=_ CUNEIFORM}', "");
Expect(1, 74868, '\p{^Is_Script=_ CUNEIFORM}', "");
Expect(1, 74868, '\P{Is_Script=_ CUNEIFORM}', "");
Expect(0, 74868, '\P{^Is_Script=_ CUNEIFORM}', "");
Error('\p{Is_Sc=_:=XSUX}');
Error('\P{Is_Sc=_:=XSUX}');
Expect(1, 74867, '\p{Is_Sc=xsux}', "");
Expect(0, 74867, '\p{^Is_Sc=xsux}', "");
Expect(0, 74867, '\P{Is_Sc=xsux}', "");
Expect(1, 74867, '\P{^Is_Sc=xsux}', "");
Expect(0, 74868, '\p{Is_Sc=xsux}', "");
Expect(1, 74868, '\p{^Is_Sc=xsux}', "");
Expect(1, 74868, '\P{Is_Sc=xsux}', "");
Expect(0, 74868, '\P{^Is_Sc=xsux}', "");
Expect(1, 74867, '\p{Is_Sc=	Xsux}', "");
Expect(0, 74867, '\p{^Is_Sc=	Xsux}', "");
Expect(0, 74867, '\P{Is_Sc=	Xsux}', "");
Expect(1, 74867, '\P{^Is_Sc=	Xsux}', "");
Expect(0, 74868, '\p{Is_Sc=	Xsux}', "");
Expect(1, 74868, '\p{^Is_Sc=	Xsux}', "");
Expect(1, 74868, '\P{Is_Sc=	Xsux}', "");
Expect(0, 74868, '\P{^Is_Sc=	Xsux}', "");
Error('\p{Script= :=Yi}');
Error('\P{Script= :=Yi}');
Expect(1, 42182, '\p{Script=yi}', "");
Expect(0, 42182, '\p{^Script=yi}', "");
Expect(0, 42182, '\P{Script=yi}', "");
Expect(1, 42182, '\P{^Script=yi}', "");
Expect(0, 42183, '\p{Script=yi}', "");
Expect(1, 42183, '\p{^Script=yi}', "");
Expect(1, 42183, '\P{Script=yi}', "");
Expect(0, 42183, '\P{^Script=yi}', "");
Expect(1, 42182, '\p{Script= 	yi}', "");
Expect(0, 42182, '\p{^Script= 	yi}', "");
Expect(0, 42182, '\P{Script= 	yi}', "");
Expect(1, 42182, '\P{^Script= 	yi}', "");
Expect(0, 42183, '\p{Script= 	yi}', "");
Expect(1, 42183, '\p{^Script= 	yi}', "");
Expect(1, 42183, '\P{Script= 	yi}', "");
Expect(0, 42183, '\P{^Script= 	yi}', "");
Error('\p{Sc:  Yiii:=}');
Error('\P{Sc:  Yiii:=}');
Expect(1, 42182, '\p{Sc=yiii}', "");
Expect(0, 42182, '\p{^Sc=yiii}', "");
Expect(0, 42182, '\P{Sc=yiii}', "");
Expect(1, 42182, '\P{^Sc=yiii}', "");
Expect(0, 42183, '\p{Sc=yiii}', "");
Expect(1, 42183, '\p{^Sc=yiii}', "");
Expect(1, 42183, '\P{Sc=yiii}', "");
Expect(0, 42183, '\P{^Sc=yiii}', "");
Expect(1, 42182, '\p{Sc=_ yiii}', "");
Expect(0, 42182, '\p{^Sc=_ yiii}', "");
Expect(0, 42182, '\P{Sc=_ yiii}', "");
Expect(1, 42182, '\P{^Sc=_ yiii}', "");
Expect(0, 42183, '\p{Sc=_ yiii}', "");
Expect(1, 42183, '\p{^Sc=_ yiii}', "");
Expect(1, 42183, '\P{Sc=_ yiii}', "");
Expect(0, 42183, '\P{^Sc=_ yiii}', "");
Error('\p{Is_Script=	_Yi/a/}');
Error('\P{Is_Script=	_Yi/a/}');
Expect(1, 42182, '\p{Is_Script=yi}', "");
Expect(0, 42182, '\p{^Is_Script=yi}', "");
Expect(0, 42182, '\P{Is_Script=yi}', "");
Expect(1, 42182, '\P{^Is_Script=yi}', "");
Expect(0, 42183, '\p{Is_Script=yi}', "");
Expect(1, 42183, '\p{^Is_Script=yi}', "");
Expect(1, 42183, '\P{Is_Script=yi}', "");
Expect(0, 42183, '\P{^Is_Script=yi}', "");
Expect(1, 42182, '\p{Is_Script=		Yi}', "");
Expect(0, 42182, '\p{^Is_Script=		Yi}', "");
Expect(0, 42182, '\P{Is_Script=		Yi}', "");
Expect(1, 42182, '\P{^Is_Script=		Yi}', "");
Expect(0, 42183, '\p{Is_Script=		Yi}', "");
Expect(1, 42183, '\p{^Is_Script=		Yi}', "");
Expect(1, 42183, '\P{Is_Script=		Yi}', "");
Expect(0, 42183, '\P{^Is_Script=		Yi}', "");
Error('\p{Is_Sc=-	yiii/a/}');
Error('\P{Is_Sc=-	yiii/a/}');
Expect(1, 42182, '\p{Is_Sc=yiii}', "");
Expect(0, 42182, '\p{^Is_Sc=yiii}', "");
Expect(0, 42182, '\P{Is_Sc=yiii}', "");
Expect(1, 42182, '\P{^Is_Sc=yiii}', "");
Expect(0, 42183, '\p{Is_Sc=yiii}', "");
Expect(1, 42183, '\p{^Is_Sc=yiii}', "");
Expect(1, 42183, '\P{Is_Sc=yiii}', "");
Expect(0, 42183, '\P{^Is_Sc=yiii}', "");
Expect(1, 42182, '\p{Is_Sc=-YIII}', "");
Expect(0, 42182, '\p{^Is_Sc=-YIII}', "");
Expect(0, 42182, '\P{Is_Sc=-YIII}', "");
Expect(1, 42182, '\P{^Is_Sc=-YIII}', "");
Expect(0, 42183, '\p{Is_Sc=-YIII}', "");
Expect(1, 42183, '\p{^Is_Sc=-YIII}', "");
Expect(1, 42183, '\P{Is_Sc=-YIII}', "");
Expect(0, 42183, '\P{^Is_Sc=-YIII}', "");
Error('\p{Script=:= Inherited}');
Error('\P{Script=:= Inherited}');
Expect(1, 917999, '\p{Script=inherited}', "");
Expect(0, 917999, '\p{^Script=inherited}', "");
Expect(0, 917999, '\P{Script=inherited}', "");
Expect(1, 917999, '\P{^Script=inherited}', "");
Expect(0, 918000, '\p{Script=inherited}', "");
Expect(1, 918000, '\p{^Script=inherited}', "");
Expect(1, 918000, '\P{Script=inherited}', "");
Expect(0, 918000, '\P{^Script=inherited}', "");
Expect(1, 917999, '\p{Script=	inherited}', "");
Expect(0, 917999, '\p{^Script=	inherited}', "");
Expect(0, 917999, '\P{Script=	inherited}', "");
Expect(1, 917999, '\P{^Script=	inherited}', "");
Expect(0, 918000, '\p{Script=	inherited}', "");
Expect(1, 918000, '\p{^Script=	inherited}', "");
Expect(1, 918000, '\P{Script=	inherited}', "");
Expect(0, 918000, '\P{^Script=	inherited}', "");
Error('\p{Sc:	zinh:=}');
Error('\P{Sc:	zinh:=}');
Expect(1, 917999, '\p{Sc=zinh}', "");
Expect(0, 917999, '\p{^Sc=zinh}', "");
Expect(0, 917999, '\P{Sc=zinh}', "");
Expect(1, 917999, '\P{^Sc=zinh}', "");
Expect(0, 918000, '\p{Sc=zinh}', "");
Expect(1, 918000, '\p{^Sc=zinh}', "");
Expect(1, 918000, '\P{Sc=zinh}', "");
Expect(0, 918000, '\P{^Sc=zinh}', "");
Expect(1, 917999, '\p{Sc=	Zinh}', "");
Expect(0, 917999, '\p{^Sc=	Zinh}', "");
Expect(0, 917999, '\P{Sc=	Zinh}', "");
Expect(1, 917999, '\P{^Sc=	Zinh}', "");
Expect(0, 918000, '\p{Sc=	Zinh}', "");
Expect(1, 918000, '\p{^Sc=	Zinh}', "");
Expect(1, 918000, '\P{Sc=	Zinh}', "");
Expect(0, 918000, '\P{^Sc=	Zinh}', "");
Error('\p{Is_Script=_/a/QAAI}');
Error('\P{Is_Script=_/a/QAAI}');
Expect(1, 917999, '\p{Is_Script=qaai}', "");
Expect(0, 917999, '\p{^Is_Script=qaai}', "");
Expect(0, 917999, '\P{Is_Script=qaai}', "");
Expect(1, 917999, '\P{^Is_Script=qaai}', "");
Expect(0, 918000, '\p{Is_Script=qaai}', "");
Expect(1, 918000, '\p{^Is_Script=qaai}', "");
Expect(1, 918000, '\P{Is_Script=qaai}', "");
Expect(0, 918000, '\P{^Is_Script=qaai}', "");
Expect(1, 917999, '\p{Is_Script=-	qaai}', "");
Expect(0, 917999, '\p{^Is_Script=-	qaai}', "");
Expect(0, 917999, '\P{Is_Script=-	qaai}', "");
Expect(1, 917999, '\P{^Is_Script=-	qaai}', "");
Expect(0, 918000, '\p{Is_Script=-	qaai}', "");
Expect(1, 918000, '\p{^Is_Script=-	qaai}', "");
Expect(1, 918000, '\P{Is_Script=-	qaai}', "");
Expect(0, 918000, '\P{^Is_Script=-	qaai}', "");
Error('\p{Is_Sc:    	Inherited/a/}');
Error('\P{Is_Sc:    	Inherited/a/}');
Expect(1, 917999, '\p{Is_Sc=inherited}', "");
Expect(0, 917999, '\p{^Is_Sc=inherited}', "");
Expect(0, 917999, '\P{Is_Sc=inherited}', "");
Expect(1, 917999, '\P{^Is_Sc=inherited}', "");
Expect(0, 918000, '\p{Is_Sc=inherited}', "");
Expect(1, 918000, '\p{^Is_Sc=inherited}', "");
Expect(1, 918000, '\P{Is_Sc=inherited}', "");
Expect(0, 918000, '\P{^Is_Sc=inherited}', "");
Expect(1, 917999, '\p{Is_Sc=__Inherited}', "");
Expect(0, 917999, '\p{^Is_Sc=__Inherited}', "");
Expect(0, 917999, '\P{Is_Sc=__Inherited}', "");
Expect(1, 917999, '\P{^Is_Sc=__Inherited}', "");
Expect(0, 918000, '\p{Is_Sc=__Inherited}', "");
Expect(1, 918000, '\p{^Is_Sc=__Inherited}', "");
Expect(1, 918000, '\P{Is_Sc=__Inherited}', "");
Expect(0, 918000, '\P{^Is_Sc=__Inherited}', "");
Error('\p{Script= 	Common:=}');
Error('\P{Script= 	Common:=}');
Expect(1, 917631, '\p{Script=common}', "");
Expect(0, 917631, '\p{^Script=common}', "");
Expect(0, 917631, '\P{Script=common}', "");
Expect(1, 917631, '\P{^Script=common}', "");
Expect(0, 917632, '\p{Script=common}', "");
Expect(1, 917632, '\p{^Script=common}', "");
Expect(1, 917632, '\P{Script=common}', "");
Expect(0, 917632, '\P{^Script=common}', "");
Expect(1, 917631, '\p{Script=		COMMON}', "");
Expect(0, 917631, '\p{^Script=		COMMON}', "");
Expect(0, 917631, '\P{Script=		COMMON}', "");
Expect(1, 917631, '\P{^Script=		COMMON}', "");
Expect(0, 917632, '\p{Script=		COMMON}', "");
Expect(1, 917632, '\p{^Script=		COMMON}', "");
Expect(1, 917632, '\P{Script=		COMMON}', "");
Expect(0, 917632, '\P{^Script=		COMMON}', "");
Error('\p{Sc= _Zyyy:=}');
Error('\P{Sc= _Zyyy:=}');
Expect(1, 917631, '\p{Sc=zyyy}', "");
Expect(0, 917631, '\p{^Sc=zyyy}', "");
Expect(0, 917631, '\P{Sc=zyyy}', "");
Expect(1, 917631, '\P{^Sc=zyyy}', "");
Expect(0, 917632, '\p{Sc=zyyy}', "");
Expect(1, 917632, '\p{^Sc=zyyy}', "");
Expect(1, 917632, '\P{Sc=zyyy}', "");
Expect(0, 917632, '\P{^Sc=zyyy}', "");
Expect(1, 917631, '\p{Sc=_	zyyy}', "");
Expect(0, 917631, '\p{^Sc=_	zyyy}', "");
Expect(0, 917631, '\P{Sc=_	zyyy}', "");
Expect(1, 917631, '\P{^Sc=_	zyyy}', "");
Expect(0, 917632, '\p{Sc=_	zyyy}', "");
Expect(1, 917632, '\p{^Sc=_	zyyy}', "");
Expect(1, 917632, '\P{Sc=_	zyyy}', "");
Expect(0, 917632, '\P{^Sc=_	zyyy}', "");
Error('\p{Is_Script=:=__Common}');
Error('\P{Is_Script=:=__Common}');
Expect(1, 917631, '\p{Is_Script=common}', "");
Expect(0, 917631, '\p{^Is_Script=common}', "");
Expect(0, 917631, '\P{Is_Script=common}', "");
Expect(1, 917631, '\P{^Is_Script=common}', "");
Expect(0, 917632, '\p{Is_Script=common}', "");
Expect(1, 917632, '\p{^Is_Script=common}', "");
Expect(1, 917632, '\P{Is_Script=common}', "");
Expect(0, 917632, '\P{^Is_Script=common}', "");
Expect(1, 917631, '\p{Is_Script=  Common}', "");
Expect(0, 917631, '\p{^Is_Script=  Common}', "");
Expect(0, 917631, '\P{Is_Script=  Common}', "");
Expect(1, 917631, '\P{^Is_Script=  Common}', "");
Expect(0, 917632, '\p{Is_Script=  Common}', "");
Expect(1, 917632, '\p{^Is_Script=  Common}', "");
Expect(1, 917632, '\P{Is_Script=  Common}', "");
Expect(0, 917632, '\P{^Is_Script=  Common}', "");
Error('\p{Is_Sc=-zyyy/a/}');
Error('\P{Is_Sc=-zyyy/a/}');
Expect(1, 917631, '\p{Is_Sc=zyyy}', "");
Expect(0, 917631, '\p{^Is_Sc=zyyy}', "");
Expect(0, 917631, '\P{Is_Sc=zyyy}', "");
Expect(1, 917631, '\P{^Is_Sc=zyyy}', "");
Expect(0, 917632, '\p{Is_Sc=zyyy}', "");
Expect(1, 917632, '\p{^Is_Sc=zyyy}', "");
Expect(1, 917632, '\P{Is_Sc=zyyy}', "");
Expect(0, 917632, '\P{^Is_Sc=zyyy}', "");
Expect(1, 917631, '\p{Is_Sc=	-Zyyy}', "");
Expect(0, 917631, '\p{^Is_Sc=	-Zyyy}', "");
Expect(0, 917631, '\P{Is_Sc=	-Zyyy}', "");
Expect(1, 917631, '\P{^Is_Sc=	-Zyyy}', "");
Expect(0, 917632, '\p{Is_Sc=	-Zyyy}', "");
Expect(1, 917632, '\p{^Is_Sc=	-Zyyy}', "");
Expect(1, 917632, '\P{Is_Sc=	-Zyyy}', "");
Expect(0, 917632, '\P{^Is_Sc=	-Zyyy}', "");
Error('\p{Script=-/a/UNKNOWN}');
Error('\P{Script=-/a/UNKNOWN}');
Expect(1, 918000, '\p{Script=unknown}', "");
Expect(0, 918000, '\p{^Script=unknown}', "");
Expect(0, 918000, '\P{Script=unknown}', "");
Expect(1, 918000, '\P{^Script=unknown}', "");
Expect(0, 917999, '\p{Script=unknown}', "");
Expect(1, 917999, '\p{^Script=unknown}', "");
Expect(1, 917999, '\P{Script=unknown}', "");
Expect(0, 917999, '\P{^Script=unknown}', "");
Expect(1, 918000, '\p{Script= _Unknown}', "");
Expect(0, 918000, '\p{^Script= _Unknown}', "");
Expect(0, 918000, '\P{Script= _Unknown}', "");
Expect(1, 918000, '\P{^Script= _Unknown}', "");
Expect(0, 917999, '\p{Script= _Unknown}', "");
Expect(1, 917999, '\p{^Script= _Unknown}', "");
Expect(1, 917999, '\P{Script= _Unknown}', "");
Expect(0, 917999, '\P{^Script= _Unknown}', "");
Error('\p{Sc=/a/	_Zzzz}');
Error('\P{Sc=/a/	_Zzzz}');
Expect(1, 918000, '\p{Sc=zzzz}', "");
Expect(0, 918000, '\p{^Sc=zzzz}', "");
Expect(0, 918000, '\P{Sc=zzzz}', "");
Expect(1, 918000, '\P{^Sc=zzzz}', "");
Expect(0, 917999, '\p{Sc=zzzz}', "");
Expect(1, 917999, '\p{^Sc=zzzz}', "");
Expect(1, 917999, '\P{Sc=zzzz}', "");
Expect(0, 917999, '\P{^Sc=zzzz}', "");
Expect(1, 918000, '\p{Sc=-ZZZZ}', "");
Expect(0, 918000, '\p{^Sc=-ZZZZ}', "");
Expect(0, 918000, '\P{Sc=-ZZZZ}', "");
Expect(1, 918000, '\P{^Sc=-ZZZZ}', "");
Expect(0, 917999, '\p{Sc=-ZZZZ}', "");
Expect(1, 917999, '\p{^Sc=-ZZZZ}', "");
Expect(1, 917999, '\P{Sc=-ZZZZ}', "");
Expect(0, 917999, '\P{^Sc=-ZZZZ}', "");
Error('\p{Is_Script=-/a/unknown}');
Error('\P{Is_Script=-/a/unknown}');
Expect(1, 918000, '\p{Is_Script:	unknown}', "");
Expect(0, 918000, '\p{^Is_Script:	unknown}', "");
Expect(0, 918000, '\P{Is_Script:	unknown}', "");
Expect(1, 918000, '\P{^Is_Script:	unknown}', "");
Expect(0, 917999, '\p{Is_Script:	unknown}', "");
Expect(1, 917999, '\p{^Is_Script:	unknown}', "");
Expect(1, 917999, '\P{Is_Script:	unknown}', "");
Expect(0, 917999, '\P{^Is_Script:	unknown}', "");
Expect(1, 918000, '\p{Is_Script=- UNKNOWN}', "");
Expect(0, 918000, '\p{^Is_Script=- UNKNOWN}', "");
Expect(0, 918000, '\P{Is_Script=- UNKNOWN}', "");
Expect(1, 918000, '\P{^Is_Script=- UNKNOWN}', "");
Expect(0, 917999, '\p{Is_Script=- UNKNOWN}', "");
Expect(1, 917999, '\p{^Is_Script=- UNKNOWN}', "");
Expect(1, 917999, '\P{Is_Script=- UNKNOWN}', "");
Expect(0, 917999, '\P{^Is_Script=- UNKNOWN}', "");
Error('\p{Is_Sc=:=- Zzzz}');
Error('\P{Is_Sc=:=- Zzzz}');
Expect(1, 918000, '\p{Is_Sc=zzzz}', "");
Expect(0, 918000, '\p{^Is_Sc=zzzz}', "");
Expect(0, 918000, '\P{Is_Sc=zzzz}', "");
Expect(1, 918000, '\P{^Is_Sc=zzzz}', "");
Expect(0, 917999, '\p{Is_Sc=zzzz}', "");
Expect(1, 917999, '\p{^Is_Sc=zzzz}', "");
Expect(1, 917999, '\P{Is_Sc=zzzz}', "");
Expect(0, 917999, '\P{^Is_Sc=zzzz}', "");
Expect(1, 918000, '\p{Is_Sc=-zzzz}', "");
Expect(0, 918000, '\p{^Is_Sc=-zzzz}', "");
Expect(0, 918000, '\P{Is_Sc=-zzzz}', "");
Expect(1, 918000, '\P{^Is_Sc=-zzzz}', "");
Expect(0, 917999, '\p{Is_Sc=-zzzz}', "");
Expect(1, 917999, '\p{^Is_Sc=-zzzz}', "");
Expect(1, 917999, '\P{Is_Sc=-zzzz}', "");
Expect(0, 917999, '\P{^Is_Sc=-zzzz}', "");
Error('\p{Script_Extensions: __ARABIC:=}');
Error('\P{Script_Extensions: __ARABIC:=}');
Expect(1, 126705, '\p{Script_Extensions=arabic}', "");
Expect(0, 126705, '\p{^Script_Extensions=arabic}', "");
Expect(0, 126705, '\P{Script_Extensions=arabic}', "");
Expect(1, 126705, '\P{^Script_Extensions=arabic}', "");
Expect(0, 126706, '\p{Script_Extensions=arabic}', "");
Expect(1, 126706, '\p{^Script_Extensions=arabic}', "");
Expect(1, 126706, '\P{Script_Extensions=arabic}', "");
Expect(0, 126706, '\P{^Script_Extensions=arabic}', "");
Expect(1, 126705, '\p{Script_Extensions=_-ARABIC}', "");
Expect(0, 126705, '\p{^Script_Extensions=_-ARABIC}', "");
Expect(0, 126705, '\P{Script_Extensions=_-ARABIC}', "");
Expect(1, 126705, '\P{^Script_Extensions=_-ARABIC}', "");
Expect(0, 126706, '\p{Script_Extensions=_-ARABIC}', "");
Expect(1, 126706, '\p{^Script_Extensions=_-ARABIC}', "");
Expect(1, 126706, '\P{Script_Extensions=_-ARABIC}', "");
Expect(0, 126706, '\P{^Script_Extensions=_-ARABIC}', "");
Error('\p{Scx=:=-Arab}');
Error('\P{Scx=:=-Arab}');
Expect(1, 126705, '\p{Scx=arab}', "");
Expect(0, 126705, '\p{^Scx=arab}', "");
Expect(0, 126705, '\P{Scx=arab}', "");
Expect(1, 126705, '\P{^Scx=arab}', "");
Expect(0, 126706, '\p{Scx=arab}', "");
Expect(1, 126706, '\p{^Scx=arab}', "");
Expect(1, 126706, '\P{Scx=arab}', "");
Expect(0, 126706, '\P{^Scx=arab}', "");
Expect(1, 126705, '\p{Scx=	-Arab}', "");
Expect(0, 126705, '\p{^Scx=	-Arab}', "");
Expect(0, 126705, '\P{Scx=	-Arab}', "");
Expect(1, 126705, '\P{^Scx=	-Arab}', "");
Expect(0, 126706, '\p{Scx=	-Arab}', "");
Expect(1, 126706, '\p{^Scx=	-Arab}', "");
Expect(1, 126706, '\P{Scx=	-Arab}', "");
Expect(0, 126706, '\P{^Scx=	-Arab}', "");
Error('\p{Is_Script_Extensions=/a/ -Arabic}');
Error('\P{Is_Script_Extensions=/a/ -Arabic}');
Expect(1, 126705, '\p{Is_Script_Extensions=arabic}', "");
Expect(0, 126705, '\p{^Is_Script_Extensions=arabic}', "");
Expect(0, 126705, '\P{Is_Script_Extensions=arabic}', "");
Expect(1, 126705, '\P{^Is_Script_Extensions=arabic}', "");
Expect(0, 126706, '\p{Is_Script_Extensions=arabic}', "");
Expect(1, 126706, '\p{^Is_Script_Extensions=arabic}', "");
Expect(1, 126706, '\P{Is_Script_Extensions=arabic}', "");
Expect(0, 126706, '\P{^Is_Script_Extensions=arabic}', "");
Expect(1, 126705, '\p{Is_Script_Extensions=-ARABIC}', "");
Expect(0, 126705, '\p{^Is_Script_Extensions=-ARABIC}', "");
Expect(0, 126705, '\P{Is_Script_Extensions=-ARABIC}', "");
Expect(1, 126705, '\P{^Is_Script_Extensions=-ARABIC}', "");
Expect(0, 126706, '\p{Is_Script_Extensions=-ARABIC}', "");
Expect(1, 126706, '\p{^Is_Script_Extensions=-ARABIC}', "");
Expect(1, 126706, '\P{Is_Script_Extensions=-ARABIC}', "");
Expect(0, 126706, '\P{^Is_Script_Extensions=-ARABIC}', "");
Error('\p{Is_Scx:   -:=ARAB}');
Error('\P{Is_Scx:   -:=ARAB}');
Expect(1, 126705, '\p{Is_Scx=arab}', "");
Expect(0, 126705, '\p{^Is_Scx=arab}', "");
Expect(0, 126705, '\P{Is_Scx=arab}', "");
Expect(1, 126705, '\P{^Is_Scx=arab}', "");
Expect(0, 126706, '\p{Is_Scx=arab}', "");
Expect(1, 126706, '\p{^Is_Scx=arab}', "");
Expect(1, 126706, '\P{Is_Scx=arab}', "");
Expect(0, 126706, '\P{^Is_Scx=arab}', "");
Expect(1, 126705, '\p{Is_Scx= -Arab}', "");
Expect(0, 126705, '\p{^Is_Scx= -Arab}', "");
Expect(0, 126705, '\P{Is_Scx= -Arab}', "");
Expect(1, 126705, '\P{^Is_Scx= -Arab}', "");
Expect(0, 126706, '\p{Is_Scx= -Arab}', "");
Expect(1, 126706, '\p{^Is_Scx= -Arab}', "");
Expect(1, 126706, '\P{Is_Scx= -Arab}', "");
Expect(0, 126706, '\P{^Is_Scx= -Arab}', "");
Error('\p{Script_Extensions: /a/_-Imperial_ARAMAIC}');
Error('\P{Script_Extensions: /a/_-Imperial_ARAMAIC}');
Expect(1, 67679, '\p{Script_Extensions=imperialaramaic}', "");
Expect(0, 67679, '\p{^Script_Extensions=imperialaramaic}', "");
Expect(0, 67679, '\P{Script_Extensions=imperialaramaic}', "");
Expect(1, 67679, '\P{^Script_Extensions=imperialaramaic}', "");
Expect(0, 67680, '\p{Script_Extensions=imperialaramaic}', "");
Expect(1, 67680, '\p{^Script_Extensions=imperialaramaic}', "");
Expect(1, 67680, '\P{Script_Extensions=imperialaramaic}', "");
Expect(0, 67680, '\P{^Script_Extensions=imperialaramaic}', "");
Expect(1, 67679, '\p{Script_Extensions= _IMPERIAL_aramaic}', "");
Expect(0, 67679, '\p{^Script_Extensions= _IMPERIAL_aramaic}', "");
Expect(0, 67679, '\P{Script_Extensions= _IMPERIAL_aramaic}', "");
Expect(1, 67679, '\P{^Script_Extensions= _IMPERIAL_aramaic}', "");
Expect(0, 67680, '\p{Script_Extensions= _IMPERIAL_aramaic}', "");
Expect(1, 67680, '\p{^Script_Extensions= _IMPERIAL_aramaic}', "");
Expect(1, 67680, '\P{Script_Extensions= _IMPERIAL_aramaic}', "");
Expect(0, 67680, '\P{^Script_Extensions= _IMPERIAL_aramaic}', "");
Error('\p{Scx=:=Armi}');
Error('\P{Scx=:=Armi}');
Expect(1, 67679, '\p{Scx:armi}', "");
Expect(0, 67679, '\p{^Scx:armi}', "");
Expect(0, 67679, '\P{Scx:armi}', "");
Expect(1, 67679, '\P{^Scx:armi}', "");
Expect(0, 67680, '\p{Scx:armi}', "");
Expect(1, 67680, '\p{^Scx:armi}', "");
Expect(1, 67680, '\P{Scx:armi}', "");
Expect(0, 67680, '\P{^Scx:armi}', "");
Expect(1, 67679, '\p{Scx= Armi}', "");
Expect(0, 67679, '\p{^Scx= Armi}', "");
Expect(0, 67679, '\P{Scx= Armi}', "");
Expect(1, 67679, '\P{^Scx= Armi}', "");
Expect(0, 67680, '\p{Scx= Armi}', "");
Expect(1, 67680, '\p{^Scx= Armi}', "");
Expect(1, 67680, '\P{Scx= Armi}', "");
Expect(0, 67680, '\P{^Scx= Armi}', "");
Error('\p{Is_Script_Extensions=-:=Imperial_ARAMAIC}');
Error('\P{Is_Script_Extensions=-:=Imperial_ARAMAIC}');
Expect(1, 67679, '\p{Is_Script_Extensions=imperialaramaic}', "");
Expect(0, 67679, '\p{^Is_Script_Extensions=imperialaramaic}', "");
Expect(0, 67679, '\P{Is_Script_Extensions=imperialaramaic}', "");
Expect(1, 67679, '\P{^Is_Script_Extensions=imperialaramaic}', "");
Expect(0, 67680, '\p{Is_Script_Extensions=imperialaramaic}', "");
Expect(1, 67680, '\p{^Is_Script_Extensions=imperialaramaic}', "");
Expect(1, 67680, '\P{Is_Script_Extensions=imperialaramaic}', "");
Expect(0, 67680, '\P{^Is_Script_Extensions=imperialaramaic}', "");
Expect(1, 67679, '\p{Is_Script_Extensions:  Imperial_aramaic}', "");
Expect(0, 67679, '\p{^Is_Script_Extensions:  Imperial_aramaic}', "");
Expect(0, 67679, '\P{Is_Script_Extensions:  Imperial_aramaic}', "");
Expect(1, 67679, '\P{^Is_Script_Extensions:  Imperial_aramaic}', "");
Expect(0, 67680, '\p{Is_Script_Extensions:  Imperial_aramaic}', "");
Expect(1, 67680, '\p{^Is_Script_Extensions:  Imperial_aramaic}', "");
Expect(1, 67680, '\P{Is_Script_Extensions:  Imperial_aramaic}', "");
Expect(0, 67680, '\P{^Is_Script_Extensions:  Imperial_aramaic}', "");
Error('\p{Is_Scx=:= 	Armi}');
Error('\P{Is_Scx=:= 	Armi}');
Expect(1, 67679, '\p{Is_Scx:armi}', "");
Expect(0, 67679, '\p{^Is_Scx:armi}', "");
Expect(0, 67679, '\P{Is_Scx:armi}', "");
Expect(1, 67679, '\P{^Is_Scx:armi}', "");
Expect(0, 67680, '\p{Is_Scx:armi}', "");
Expect(1, 67680, '\p{^Is_Scx:armi}', "");
Expect(1, 67680, '\P{Is_Scx:armi}', "");
Expect(0, 67680, '\P{^Is_Scx:armi}', "");
Expect(1, 67679, '\p{Is_Scx=--ARMI}', "");
Expect(0, 67679, '\p{^Is_Scx=--ARMI}', "");
Expect(0, 67679, '\P{Is_Scx=--ARMI}', "");
Expect(1, 67679, '\P{^Is_Scx=--ARMI}', "");
Expect(0, 67680, '\p{Is_Scx=--ARMI}', "");
Expect(1, 67680, '\p{^Is_Scx=--ARMI}', "");
Expect(1, 67680, '\P{Is_Scx=--ARMI}', "");
Expect(0, 67680, '\P{^Is_Scx=--ARMI}', "");
Error('\p{Script_Extensions=_Armenian:=}');
Error('\P{Script_Extensions=_Armenian:=}');
Expect(1, 64279, '\p{Script_Extensions=armenian}', "");
Expect(0, 64279, '\p{^Script_Extensions=armenian}', "");
Expect(0, 64279, '\P{Script_Extensions=armenian}', "");
Expect(1, 64279, '\P{^Script_Extensions=armenian}', "");
Expect(0, 64280, '\p{Script_Extensions=armenian}', "");
Expect(1, 64280, '\p{^Script_Extensions=armenian}', "");
Expect(1, 64280, '\P{Script_Extensions=armenian}', "");
Expect(0, 64280, '\P{^Script_Extensions=armenian}', "");
Expect(1, 64279, '\p{Script_Extensions:   -Armenian}', "");
Expect(0, 64279, '\p{^Script_Extensions:   -Armenian}', "");
Expect(0, 64279, '\P{Script_Extensions:   -Armenian}', "");
Expect(1, 64279, '\P{^Script_Extensions:   -Armenian}', "");
Expect(0, 64280, '\p{Script_Extensions:   -Armenian}', "");
Expect(1, 64280, '\p{^Script_Extensions:   -Armenian}', "");
Expect(1, 64280, '\P{Script_Extensions:   -Armenian}', "");
Expect(0, 64280, '\P{^Script_Extensions:   -Armenian}', "");
Error('\p{Scx= /a/armn}');
Error('\P{Scx= /a/armn}');
Expect(1, 64279, '\p{Scx=armn}', "");
Expect(0, 64279, '\p{^Scx=armn}', "");
Expect(0, 64279, '\P{Scx=armn}', "");
Expect(1, 64279, '\P{^Scx=armn}', "");
Expect(0, 64280, '\p{Scx=armn}', "");
Expect(1, 64280, '\p{^Scx=armn}', "");
Expect(1, 64280, '\P{Scx=armn}', "");
Expect(0, 64280, '\P{^Scx=armn}', "");
Expect(1, 64279, '\p{Scx=__Armn}', "");
Expect(0, 64279, '\p{^Scx=__Armn}', "");
Expect(0, 64279, '\P{Scx=__Armn}', "");
Expect(1, 64279, '\P{^Scx=__Armn}', "");
Expect(0, 64280, '\p{Scx=__Armn}', "");
Expect(1, 64280, '\p{^Scx=__Armn}', "");
Expect(1, 64280, '\P{Scx=__Armn}', "");
Expect(0, 64280, '\P{^Scx=__Armn}', "");
Error('\p{Is_Script_Extensions=:=Armenian}');
Error('\P{Is_Script_Extensions=:=Armenian}');
Expect(1, 64279, '\p{Is_Script_Extensions=armenian}', "");
Expect(0, 64279, '\p{^Is_Script_Extensions=armenian}', "");
Expect(0, 64279, '\P{Is_Script_Extensions=armenian}', "");
Expect(1, 64279, '\P{^Is_Script_Extensions=armenian}', "");
Expect(0, 64280, '\p{Is_Script_Extensions=armenian}', "");
Expect(1, 64280, '\p{^Is_Script_Extensions=armenian}', "");
Expect(1, 64280, '\P{Is_Script_Extensions=armenian}', "");
Expect(0, 64280, '\P{^Is_Script_Extensions=armenian}', "");
Expect(1, 64279, '\p{Is_Script_Extensions=	Armenian}', "");
Expect(0, 64279, '\p{^Is_Script_Extensions=	Armenian}', "");
Expect(0, 64279, '\P{Is_Script_Extensions=	Armenian}', "");
Expect(1, 64279, '\P{^Is_Script_Extensions=	Armenian}', "");
Expect(0, 64280, '\p{Is_Script_Extensions=	Armenian}', "");
Expect(1, 64280, '\p{^Is_Script_Extensions=	Armenian}', "");
Expect(1, 64280, '\P{Is_Script_Extensions=	Armenian}', "");
Expect(0, 64280, '\P{^Is_Script_Extensions=	Armenian}', "");
Error('\p{Is_Scx=:=ARMN}');
Error('\P{Is_Scx=:=ARMN}');
Expect(1, 64279, '\p{Is_Scx=armn}', "");
Expect(0, 64279, '\p{^Is_Scx=armn}', "");
Expect(0, 64279, '\P{Is_Scx=armn}', "");
Expect(1, 64279, '\P{^Is_Scx=armn}', "");
Expect(0, 64280, '\p{Is_Scx=armn}', "");
Expect(1, 64280, '\p{^Is_Scx=armn}', "");
Expect(1, 64280, '\P{Is_Scx=armn}', "");
Expect(0, 64280, '\P{^Is_Scx=armn}', "");
Expect(1, 64279, '\p{Is_Scx=_Armn}', "");
Expect(0, 64279, '\p{^Is_Scx=_Armn}', "");
Expect(0, 64279, '\P{Is_Scx=_Armn}', "");
Expect(1, 64279, '\P{^Is_Scx=_Armn}', "");
Expect(0, 64280, '\p{Is_Scx=_Armn}', "");
Expect(1, 64280, '\p{^Is_Scx=_Armn}', "");
Expect(1, 64280, '\P{Is_Scx=_Armn}', "");
Expect(0, 64280, '\P{^Is_Scx=_Armn}', "");
Error('\p{Script_Extensions=		avestan/a/}');
Error('\P{Script_Extensions=		avestan/a/}');
Expect(1, 68415, '\p{Script_Extensions=avestan}', "");
Expect(0, 68415, '\p{^Script_Extensions=avestan}', "");
Expect(0, 68415, '\P{Script_Extensions=avestan}', "");
Expect(1, 68415, '\P{^Script_Extensions=avestan}', "");
Expect(0, 68416, '\p{Script_Extensions=avestan}', "");
Expect(1, 68416, '\p{^Script_Extensions=avestan}', "");
Expect(1, 68416, '\P{Script_Extensions=avestan}', "");
Expect(0, 68416, '\P{^Script_Extensions=avestan}', "");
Expect(1, 68415, '\p{Script_Extensions= AVESTAN}', "");
Expect(0, 68415, '\p{^Script_Extensions= AVESTAN}', "");
Expect(0, 68415, '\P{Script_Extensions= AVESTAN}', "");
Expect(1, 68415, '\P{^Script_Extensions= AVESTAN}', "");
Expect(0, 68416, '\p{Script_Extensions= AVESTAN}', "");
Expect(1, 68416, '\p{^Script_Extensions= AVESTAN}', "");
Expect(1, 68416, '\P{Script_Extensions= AVESTAN}', "");
Expect(0, 68416, '\P{^Script_Extensions= AVESTAN}', "");
Error('\p{Scx=	/a/AVST}');
Error('\P{Scx=	/a/AVST}');
Expect(1, 68415, '\p{Scx=avst}', "");
Expect(0, 68415, '\p{^Scx=avst}', "");
Expect(0, 68415, '\P{Scx=avst}', "");
Expect(1, 68415, '\P{^Scx=avst}', "");
Expect(0, 68416, '\p{Scx=avst}', "");
Expect(1, 68416, '\p{^Scx=avst}', "");
Expect(1, 68416, '\P{Scx=avst}', "");
Expect(0, 68416, '\P{^Scx=avst}', "");
Error('\p{Is_Script_Extensions=:=_avestan}');
Error('\P{Is_Script_Extensions=:=_avestan}');
Expect(1, 68415, '\p{Is_Script_Extensions=avestan}', "");
Expect(0, 68415, '\p{^Is_Script_Extensions=avestan}', "");
Expect(0, 68415, '\P{Is_Script_Extensions=avestan}', "");
Expect(1, 68415, '\P{^Is_Script_Extensions=avestan}', "");
Expect(0, 68416, '\p{Is_Script_Extensions=avestan}', "");
Expect(1, 68416, '\p{^Is_Script_Extensions=avestan}', "");
Expect(1, 68416, '\P{Is_Script_Extensions=avestan}', "");
Expect(0, 68416, '\P{^Is_Script_Extensions=avestan}', "");
Expect(1, 68415, '\p{Is_Script_Extensions=_AVESTAN}', "");
Expect(0, 68415, '\p{^Is_Script_Extensions=_AVESTAN}', "");
Expect(0, 68415, '\P{Is_Script_Extensions=_AVESTAN}', "");
Expect(1, 68415, '\P{^Is_Script_Extensions=_AVESTAN}', "");
Expect(0, 68416, '\p{Is_Script_Extensions=_AVESTAN}', "");
Expect(1, 68416, '\p{^Is_Script_Extensions=_AVESTAN}', "");
Expect(1, 68416, '\P{Is_Script_Extensions=_AVESTAN}', "");
Expect(0, 68416, '\P{^Is_Script_Extensions=_AVESTAN}', "");
Error('\p{Is_Scx=:=avst}');
Error('\P{Is_Scx=:=avst}');
Expect(1, 68415, '\p{Is_Scx: avst}', "");
Expect(0, 68415, '\p{^Is_Scx: avst}', "");
Expect(0, 68415, '\P{Is_Scx: avst}', "");
Expect(1, 68415, '\P{^Is_Scx: avst}', "");
Expect(0, 68416, '\p{Is_Scx: avst}', "");
Expect(1, 68416, '\p{^Is_Scx: avst}', "");
Expect(1, 68416, '\P{Is_Scx: avst}', "");
Expect(0, 68416, '\P{^Is_Scx: avst}', "");
Expect(1, 68415, '\p{Is_Scx:   	Avst}', "");
Expect(0, 68415, '\p{^Is_Scx:   	Avst}', "");
Expect(0, 68415, '\P{Is_Scx:   	Avst}', "");
Expect(1, 68415, '\P{^Is_Scx:   	Avst}', "");
Expect(0, 68416, '\p{Is_Scx:   	Avst}', "");
Expect(1, 68416, '\p{^Is_Scx:   	Avst}', "");
Expect(1, 68416, '\P{Is_Scx:   	Avst}', "");
Expect(0, 68416, '\P{^Is_Scx:   	Avst}', "");
Error('\p{Script_Extensions=_Balinese:=}');
Error('\P{Script_Extensions=_Balinese:=}');
Expect(1, 7036, '\p{Script_Extensions=balinese}', "");
Expect(0, 7036, '\p{^Script_Extensions=balinese}', "");
Expect(0, 7036, '\P{Script_Extensions=balinese}', "");
Expect(1, 7036, '\P{^Script_Extensions=balinese}', "");
Expect(0, 7037, '\p{Script_Extensions=balinese}', "");
Expect(1, 7037, '\p{^Script_Extensions=balinese}', "");
Expect(1, 7037, '\P{Script_Extensions=balinese}', "");
Expect(0, 7037, '\P{^Script_Extensions=balinese}', "");
Expect(1, 7036, '\p{Script_Extensions=	 Balinese}', "");
Expect(0, 7036, '\p{^Script_Extensions=	 Balinese}', "");
Expect(0, 7036, '\P{Script_Extensions=	 Balinese}', "");
Expect(1, 7036, '\P{^Script_Extensions=	 Balinese}', "");
Expect(0, 7037, '\p{Script_Extensions=	 Balinese}', "");
Expect(1, 7037, '\p{^Script_Extensions=	 Balinese}', "");
Expect(1, 7037, '\P{Script_Extensions=	 Balinese}', "");
Expect(0, 7037, '\P{^Script_Extensions=	 Balinese}', "");
Error('\p{Scx:		-Bali/a/}');
Error('\P{Scx:		-Bali/a/}');
Expect(1, 7036, '\p{Scx=bali}', "");
Expect(0, 7036, '\p{^Scx=bali}', "");
Expect(0, 7036, '\P{Scx=bali}', "");
Expect(1, 7036, '\P{^Scx=bali}', "");
Expect(0, 7037, '\p{Scx=bali}', "");
Expect(1, 7037, '\p{^Scx=bali}', "");
Expect(1, 7037, '\P{Scx=bali}', "");
Expect(0, 7037, '\P{^Scx=bali}', "");
Expect(1, 7036, '\p{Scx= -BALI}', "");
Expect(0, 7036, '\p{^Scx= -BALI}', "");
Expect(0, 7036, '\P{Scx= -BALI}', "");
Expect(1, 7036, '\P{^Scx= -BALI}', "");
Expect(0, 7037, '\p{Scx= -BALI}', "");
Expect(1, 7037, '\p{^Scx= -BALI}', "");
Expect(1, 7037, '\P{Scx= -BALI}', "");
Expect(0, 7037, '\P{^Scx= -BALI}', "");
Error('\p{Is_Script_Extensions=_ Balinese:=}');
Error('\P{Is_Script_Extensions=_ Balinese:=}');
Expect(1, 7036, '\p{Is_Script_Extensions=balinese}', "");
Expect(0, 7036, '\p{^Is_Script_Extensions=balinese}', "");
Expect(0, 7036, '\P{Is_Script_Extensions=balinese}', "");
Expect(1, 7036, '\P{^Is_Script_Extensions=balinese}', "");
Expect(0, 7037, '\p{Is_Script_Extensions=balinese}', "");
Expect(1, 7037, '\p{^Is_Script_Extensions=balinese}', "");
Expect(1, 7037, '\P{Is_Script_Extensions=balinese}', "");
Expect(0, 7037, '\P{^Is_Script_Extensions=balinese}', "");
Expect(1, 7036, '\p{Is_Script_Extensions=_balinese}', "");
Expect(0, 7036, '\p{^Is_Script_Extensions=_balinese}', "");
Expect(0, 7036, '\P{Is_Script_Extensions=_balinese}', "");
Expect(1, 7036, '\P{^Is_Script_Extensions=_balinese}', "");
Expect(0, 7037, '\p{Is_Script_Extensions=_balinese}', "");
Expect(1, 7037, '\p{^Is_Script_Extensions=_balinese}', "");
Expect(1, 7037, '\P{Is_Script_Extensions=_balinese}', "");
Expect(0, 7037, '\P{^Is_Script_Extensions=_balinese}', "");
Error('\p{Is_Scx=/a/Bali}');
Error('\P{Is_Scx=/a/Bali}');
Expect(1, 7036, '\p{Is_Scx=bali}', "");
Expect(0, 7036, '\p{^Is_Scx=bali}', "");
Expect(0, 7036, '\P{Is_Scx=bali}', "");
Expect(1, 7036, '\P{^Is_Scx=bali}', "");
Expect(0, 7037, '\p{Is_Scx=bali}', "");
Expect(1, 7037, '\p{^Is_Scx=bali}', "");
Expect(1, 7037, '\P{Is_Scx=bali}', "");
Expect(0, 7037, '\P{^Is_Scx=bali}', "");
Expect(1, 7036, '\p{Is_Scx:	_BALI}', "");
Expect(0, 7036, '\p{^Is_Scx:	_BALI}', "");
Expect(0, 7036, '\P{Is_Scx:	_BALI}', "");
Expect(1, 7036, '\P{^Is_Scx:	_BALI}', "");
Expect(0, 7037, '\p{Is_Scx:	_BALI}', "");
Expect(1, 7037, '\p{^Is_Scx:	_BALI}', "");
Expect(1, 7037, '\P{Is_Scx:	_BALI}', "");
Expect(0, 7037, '\P{^Is_Scx:	_BALI}', "");
Error('\p{Script_Extensions=_	bamum:=}');
Error('\P{Script_Extensions=_	bamum:=}');
Expect(1, 92728, '\p{Script_Extensions=bamum}', "");
Expect(0, 92728, '\p{^Script_Extensions=bamum}', "");
Expect(0, 92728, '\P{Script_Extensions=bamum}', "");
Expect(1, 92728, '\P{^Script_Extensions=bamum}', "");
Expect(0, 92729, '\p{Script_Extensions=bamum}', "");
Expect(1, 92729, '\p{^Script_Extensions=bamum}', "");
Expect(1, 92729, '\P{Script_Extensions=bamum}', "");
Expect(0, 92729, '\P{^Script_Extensions=bamum}', "");
Expect(1, 92728, '\p{Script_Extensions=_	BAMUM}', "");
Expect(0, 92728, '\p{^Script_Extensions=_	BAMUM}', "");
Expect(0, 92728, '\P{Script_Extensions=_	BAMUM}', "");
Expect(1, 92728, '\P{^Script_Extensions=_	BAMUM}', "");
Expect(0, 92729, '\p{Script_Extensions=_	BAMUM}', "");
Expect(1, 92729, '\p{^Script_Extensions=_	BAMUM}', "");
Expect(1, 92729, '\P{Script_Extensions=_	BAMUM}', "");
Expect(0, 92729, '\P{^Script_Extensions=_	BAMUM}', "");
Error('\p{Scx= Bamu:=}');
Error('\P{Scx= Bamu:=}');
Expect(1, 92728, '\p{Scx=bamu}', "");
Expect(0, 92728, '\p{^Scx=bamu}', "");
Expect(0, 92728, '\P{Scx=bamu}', "");
Expect(1, 92728, '\P{^Scx=bamu}', "");
Expect(0, 92729, '\p{Scx=bamu}', "");
Expect(1, 92729, '\p{^Scx=bamu}', "");
Expect(1, 92729, '\P{Scx=bamu}', "");
Expect(0, 92729, '\P{^Scx=bamu}', "");
Expect(1, 92728, '\p{Scx= _Bamu}', "");
Expect(0, 92728, '\p{^Scx= _Bamu}', "");
Expect(0, 92728, '\P{Scx= _Bamu}', "");
Expect(1, 92728, '\P{^Scx= _Bamu}', "");
Expect(0, 92729, '\p{Scx= _Bamu}', "");
Expect(1, 92729, '\p{^Scx= _Bamu}', "");
Expect(1, 92729, '\P{Scx= _Bamu}', "");
Expect(0, 92729, '\P{^Scx= _Bamu}', "");
Error('\p{Is_Script_Extensions:		:=BAMUM}');
Error('\P{Is_Script_Extensions:		:=BAMUM}');
Expect(1, 92728, '\p{Is_Script_Extensions=bamum}', "");
Expect(0, 92728, '\p{^Is_Script_Extensions=bamum}', "");
Expect(0, 92728, '\P{Is_Script_Extensions=bamum}', "");
Expect(1, 92728, '\P{^Is_Script_Extensions=bamum}', "");
Expect(0, 92729, '\p{Is_Script_Extensions=bamum}', "");
Expect(1, 92729, '\p{^Is_Script_Extensions=bamum}', "");
Expect(1, 92729, '\P{Is_Script_Extensions=bamum}', "");
Expect(0, 92729, '\P{^Is_Script_Extensions=bamum}', "");
Expect(1, 92728, '\p{Is_Script_Extensions=	Bamum}', "");
Expect(0, 92728, '\p{^Is_Script_Extensions=	Bamum}', "");
Expect(0, 92728, '\P{Is_Script_Extensions=	Bamum}', "");
Expect(1, 92728, '\P{^Is_Script_Extensions=	Bamum}', "");
Expect(0, 92729, '\p{Is_Script_Extensions=	Bamum}', "");
Expect(1, 92729, '\p{^Is_Script_Extensions=	Bamum}', "");
Expect(1, 92729, '\P{Is_Script_Extensions=	Bamum}', "");
Expect(0, 92729, '\P{^Is_Script_Extensions=	Bamum}', "");
Error('\p{Is_Scx=/a/_ Bamu}');
Error('\P{Is_Scx=/a/_ Bamu}');
Expect(1, 92728, '\p{Is_Scx=bamu}', "");
Expect(0, 92728, '\p{^Is_Scx=bamu}', "");
Expect(0, 92728, '\P{Is_Scx=bamu}', "");
Expect(1, 92728, '\P{^Is_Scx=bamu}', "");
Expect(0, 92729, '\p{Is_Scx=bamu}', "");
Expect(1, 92729, '\p{^Is_Scx=bamu}', "");
Expect(1, 92729, '\P{Is_Scx=bamu}', "");
Expect(0, 92729, '\P{^Is_Scx=bamu}', "");
Expect(1, 92728, '\p{Is_Scx=  Bamu}', "");
Expect(0, 92728, '\p{^Is_Scx=  Bamu}', "");
Expect(0, 92728, '\P{Is_Scx=  Bamu}', "");
Expect(1, 92728, '\P{^Is_Scx=  Bamu}', "");
Expect(0, 92729, '\p{Is_Scx=  Bamu}', "");
Expect(1, 92729, '\p{^Is_Scx=  Bamu}', "");
Expect(1, 92729, '\P{Is_Scx=  Bamu}', "");
Expect(0, 92729, '\P{^Is_Scx=  Bamu}', "");
Error('\p{Script_Extensions= Batak/a/}');
Error('\P{Script_Extensions= Batak/a/}');
Expect(1, 7167, '\p{Script_Extensions=batak}', "");
Expect(0, 7167, '\p{^Script_Extensions=batak}', "");
Expect(0, 7167, '\P{Script_Extensions=batak}', "");
Expect(1, 7167, '\P{^Script_Extensions=batak}', "");
Expect(0, 7168, '\p{Script_Extensions=batak}', "");
Expect(1, 7168, '\p{^Script_Extensions=batak}', "");
Expect(1, 7168, '\P{Script_Extensions=batak}', "");
Expect(0, 7168, '\P{^Script_Extensions=batak}', "");
Expect(1, 7167, '\p{Script_Extensions:	- Batak}', "");
Expect(0, 7167, '\p{^Script_Extensions:	- Batak}', "");
Expect(0, 7167, '\P{Script_Extensions:	- Batak}', "");
Expect(1, 7167, '\P{^Script_Extensions:	- Batak}', "");
Expect(0, 7168, '\p{Script_Extensions:	- Batak}', "");
Expect(1, 7168, '\p{^Script_Extensions:	- Batak}', "");
Expect(1, 7168, '\P{Script_Extensions:	- Batak}', "");
Expect(0, 7168, '\P{^Script_Extensions:	- Batak}', "");
Error('\p{Scx=	/a/Batk}');
Error('\P{Scx=	/a/Batk}');
Expect(1, 7167, '\p{Scx=batk}', "");
Expect(0, 7167, '\p{^Scx=batk}', "");
Expect(0, 7167, '\P{Scx=batk}', "");
Expect(1, 7167, '\P{^Scx=batk}', "");
Expect(0, 7168, '\p{Scx=batk}', "");
Expect(1, 7168, '\p{^Scx=batk}', "");
Expect(1, 7168, '\P{Scx=batk}', "");
Expect(0, 7168, '\P{^Scx=batk}', "");
Expect(1, 7167, '\p{Scx=_ Batk}', "");
Expect(0, 7167, '\p{^Scx=_ Batk}', "");
Expect(0, 7167, '\P{Scx=_ Batk}', "");
Expect(1, 7167, '\P{^Scx=_ Batk}', "");
Expect(0, 7168, '\p{Scx=_ Batk}', "");
Expect(1, 7168, '\p{^Scx=_ Batk}', "");
Expect(1, 7168, '\P{Scx=_ Batk}', "");
Expect(0, 7168, '\P{^Scx=_ Batk}', "");
Error('\p{Is_Script_Extensions:-:=Batak}');
Error('\P{Is_Script_Extensions:-:=Batak}');
Expect(1, 7167, '\p{Is_Script_Extensions=batak}', "");
Expect(0, 7167, '\p{^Is_Script_Extensions=batak}', "");
Expect(0, 7167, '\P{Is_Script_Extensions=batak}', "");
Expect(1, 7167, '\P{^Is_Script_Extensions=batak}', "");
Expect(0, 7168, '\p{Is_Script_Extensions=batak}', "");
Expect(1, 7168, '\p{^Is_Script_Extensions=batak}', "");
Expect(1, 7168, '\P{Is_Script_Extensions=batak}', "");
Expect(0, 7168, '\P{^Is_Script_Extensions=batak}', "");
Expect(1, 7167, '\p{Is_Script_Extensions=- batak}', "");
Expect(0, 7167, '\p{^Is_Script_Extensions=- batak}', "");
Expect(0, 7167, '\P{Is_Script_Extensions=- batak}', "");
Expect(1, 7167, '\P{^Is_Script_Extensions=- batak}', "");
Expect(0, 7168, '\p{Is_Script_Extensions=- batak}', "");
Expect(1, 7168, '\p{^Is_Script_Extensions=- batak}', "");
Expect(1, 7168, '\P{Is_Script_Extensions=- batak}', "");
Expect(0, 7168, '\P{^Is_Script_Extensions=- batak}', "");
Error('\p{Is_Scx= /a/Batk}');
Error('\P{Is_Scx= /a/Batk}');
Expect(1, 7167, '\p{Is_Scx=batk}', "");
Expect(0, 7167, '\p{^Is_Scx=batk}', "");
Expect(0, 7167, '\P{Is_Scx=batk}', "");
Expect(1, 7167, '\P{^Is_Scx=batk}', "");
Expect(0, 7168, '\p{Is_Scx=batk}', "");
Expect(1, 7168, '\p{^Is_Scx=batk}', "");
Expect(1, 7168, '\P{Is_Scx=batk}', "");
Expect(0, 7168, '\P{^Is_Scx=batk}', "");
Expect(1, 7167, '\p{Is_Scx=-Batk}', "");
Expect(0, 7167, '\p{^Is_Scx=-Batk}', "");
Expect(0, 7167, '\P{Is_Scx=-Batk}', "");
Expect(1, 7167, '\P{^Is_Scx=-Batk}', "");
Expect(0, 7168, '\p{Is_Scx=-Batk}', "");
Expect(1, 7168, '\p{^Is_Scx=-Batk}', "");
Expect(1, 7168, '\P{Is_Scx=-Batk}', "");
Expect(0, 7168, '\P{^Is_Scx=-Batk}', "");
Error('\p{Script_Extensions=:=		Bengali}');
Error('\P{Script_Extensions=:=		Bengali}');
Expect(1, 2555, '\p{Script_Extensions=bengali}', "");
Expect(0, 2555, '\p{^Script_Extensions=bengali}', "");
Expect(0, 2555, '\P{Script_Extensions=bengali}', "");
Expect(1, 2555, '\P{^Script_Extensions=bengali}', "");
Expect(0, 2556, '\p{Script_Extensions=bengali}', "");
Expect(1, 2556, '\p{^Script_Extensions=bengali}', "");
Expect(1, 2556, '\P{Script_Extensions=bengali}', "");
Expect(0, 2556, '\P{^Script_Extensions=bengali}', "");
Expect(1, 2555, '\p{Script_Extensions=--Bengali}', "");
Expect(0, 2555, '\p{^Script_Extensions=--Bengali}', "");
Expect(0, 2555, '\P{Script_Extensions=--Bengali}', "");
Expect(1, 2555, '\P{^Script_Extensions=--Bengali}', "");
Expect(0, 2556, '\p{Script_Extensions=--Bengali}', "");
Expect(1, 2556, '\p{^Script_Extensions=--Bengali}', "");
Expect(1, 2556, '\P{Script_Extensions=--Bengali}', "");
Expect(0, 2556, '\P{^Script_Extensions=--Bengali}', "");
Error('\p{Scx=Beng/a/}');
Error('\P{Scx=Beng/a/}');
Expect(1, 2555, '\p{Scx=beng}', "");
Expect(0, 2555, '\p{^Scx=beng}', "");
Expect(0, 2555, '\P{Scx=beng}', "");
Expect(1, 2555, '\P{^Scx=beng}', "");
Expect(0, 2556, '\p{Scx=beng}', "");
Expect(1, 2556, '\p{^Scx=beng}', "");
Expect(1, 2556, '\P{Scx=beng}', "");
Expect(0, 2556, '\P{^Scx=beng}', "");
Expect(1, 2555, '\p{Scx=_-Beng}', "");
Expect(0, 2555, '\p{^Scx=_-Beng}', "");
Expect(0, 2555, '\P{Scx=_-Beng}', "");
Expect(1, 2555, '\P{^Scx=_-Beng}', "");
Expect(0, 2556, '\p{Scx=_-Beng}', "");
Expect(1, 2556, '\p{^Scx=_-Beng}', "");
Expect(1, 2556, '\P{Scx=_-Beng}', "");
Expect(0, 2556, '\P{^Scx=_-Beng}', "");
Error('\p{Is_Script_Extensions:  _bengali/a/}');
Error('\P{Is_Script_Extensions:  _bengali/a/}');
Expect(1, 2555, '\p{Is_Script_Extensions=bengali}', "");
Expect(0, 2555, '\p{^Is_Script_Extensions=bengali}', "");
Expect(0, 2555, '\P{Is_Script_Extensions=bengali}', "");
Expect(1, 2555, '\P{^Is_Script_Extensions=bengali}', "");
Expect(0, 2556, '\p{Is_Script_Extensions=bengali}', "");
Expect(1, 2556, '\p{^Is_Script_Extensions=bengali}', "");
Expect(1, 2556, '\P{Is_Script_Extensions=bengali}', "");
Expect(0, 2556, '\P{^Is_Script_Extensions=bengali}', "");
Expect(1, 2555, '\p{Is_Script_Extensions=	-Bengali}', "");
Expect(0, 2555, '\p{^Is_Script_Extensions=	-Bengali}', "");
Expect(0, 2555, '\P{Is_Script_Extensions=	-Bengali}', "");
Expect(1, 2555, '\P{^Is_Script_Extensions=	-Bengali}', "");
Expect(0, 2556, '\p{Is_Script_Extensions=	-Bengali}', "");
Expect(1, 2556, '\p{^Is_Script_Extensions=	-Bengali}', "");
Expect(1, 2556, '\P{Is_Script_Extensions=	-Bengali}', "");
Expect(0, 2556, '\P{^Is_Script_Extensions=	-Bengali}', "");
Error('\p{Is_Scx=	Beng:=}');
Error('\P{Is_Scx=	Beng:=}');
Expect(1, 2555, '\p{Is_Scx=beng}', "");
Expect(0, 2555, '\p{^Is_Scx=beng}', "");
Expect(0, 2555, '\P{Is_Scx=beng}', "");
Expect(1, 2555, '\P{^Is_Scx=beng}', "");
Expect(0, 2556, '\p{Is_Scx=beng}', "");
Expect(1, 2556, '\p{^Is_Scx=beng}', "");
Expect(1, 2556, '\P{Is_Scx=beng}', "");
Expect(0, 2556, '\P{^Is_Scx=beng}', "");
Expect(1, 2555, '\p{Is_Scx= _Beng}', "");
Expect(0, 2555, '\p{^Is_Scx= _Beng}', "");
Expect(0, 2555, '\P{Is_Scx= _Beng}', "");
Expect(1, 2555, '\P{^Is_Scx= _Beng}', "");
Expect(0, 2556, '\p{Is_Scx= _Beng}', "");
Expect(1, 2556, '\p{^Is_Scx= _Beng}', "");
Expect(1, 2556, '\P{Is_Scx= _Beng}', "");
Expect(0, 2556, '\P{^Is_Scx= _Beng}', "");
Error('\p{Script_Extensions= -Bopomofo/a/}');
Error('\P{Script_Extensions= -Bopomofo/a/}');
Expect(1, 65381, '\p{Script_Extensions=bopomofo}', "");
Expect(0, 65381, '\p{^Script_Extensions=bopomofo}', "");
Expect(0, 65381, '\P{Script_Extensions=bopomofo}', "");
Expect(1, 65381, '\P{^Script_Extensions=bopomofo}', "");
Expect(0, 65382, '\p{Script_Extensions=bopomofo}', "");
Expect(1, 65382, '\p{^Script_Extensions=bopomofo}', "");
Expect(1, 65382, '\P{Script_Extensions=bopomofo}', "");
Expect(0, 65382, '\P{^Script_Extensions=bopomofo}', "");
Expect(1, 65381, '\p{Script_Extensions=_BOPOMOFO}', "");
Expect(0, 65381, '\p{^Script_Extensions=_BOPOMOFO}', "");
Expect(0, 65381, '\P{Script_Extensions=_BOPOMOFO}', "");
Expect(1, 65381, '\P{^Script_Extensions=_BOPOMOFO}', "");
Expect(0, 65382, '\p{Script_Extensions=_BOPOMOFO}', "");
Expect(1, 65382, '\p{^Script_Extensions=_BOPOMOFO}', "");
Expect(1, 65382, '\P{Script_Extensions=_BOPOMOFO}', "");
Expect(0, 65382, '\P{^Script_Extensions=_BOPOMOFO}', "");
Error('\p{Scx=/a/_bopo}');
Error('\P{Scx=/a/_bopo}');
Expect(1, 65381, '\p{Scx=bopo}', "");
Expect(0, 65381, '\p{^Scx=bopo}', "");
Expect(0, 65381, '\P{Scx=bopo}', "");
Expect(1, 65381, '\P{^Scx=bopo}', "");
Expect(0, 65382, '\p{Scx=bopo}', "");
Expect(1, 65382, '\p{^Scx=bopo}', "");
Expect(1, 65382, '\P{Scx=bopo}', "");
Expect(0, 65382, '\P{^Scx=bopo}', "");
Expect(1, 65381, '\p{Scx= BOPO}', "");
Expect(0, 65381, '\p{^Scx= BOPO}', "");
Expect(0, 65381, '\P{Scx= BOPO}', "");
Expect(1, 65381, '\P{^Scx= BOPO}', "");
Expect(0, 65382, '\p{Scx= BOPO}', "");
Expect(1, 65382, '\p{^Scx= BOPO}', "");
Expect(1, 65382, '\P{Scx= BOPO}', "");
Expect(0, 65382, '\P{^Scx= BOPO}', "");
Error('\p{Is_Script_Extensions=-/a/bopomofo}');
Error('\P{Is_Script_Extensions=-/a/bopomofo}');
Expect(1, 65381, '\p{Is_Script_Extensions=bopomofo}', "");
Expect(0, 65381, '\p{^Is_Script_Extensions=bopomofo}', "");
Expect(0, 65381, '\P{Is_Script_Extensions=bopomofo}', "");
Expect(1, 65381, '\P{^Is_Script_Extensions=bopomofo}', "");
Expect(0, 65382, '\p{Is_Script_Extensions=bopomofo}', "");
Expect(1, 65382, '\p{^Is_Script_Extensions=bopomofo}', "");
Expect(1, 65382, '\P{Is_Script_Extensions=bopomofo}', "");
Expect(0, 65382, '\P{^Is_Script_Extensions=bopomofo}', "");
Expect(1, 65381, '\p{Is_Script_Extensions= 	Bopomofo}', "");
Expect(0, 65381, '\p{^Is_Script_Extensions= 	Bopomofo}', "");
Expect(0, 65381, '\P{Is_Script_Extensions= 	Bopomofo}', "");
Expect(1, 65381, '\P{^Is_Script_Extensions= 	Bopomofo}', "");
Expect(0, 65382, '\p{Is_Script_Extensions= 	Bopomofo}', "");
Expect(1, 65382, '\p{^Is_Script_Extensions= 	Bopomofo}', "");
Expect(1, 65382, '\P{Is_Script_Extensions= 	Bopomofo}', "");
Expect(0, 65382, '\P{^Is_Script_Extensions= 	Bopomofo}', "");
Error('\p{Is_Scx=/a/_	BOPO}');
Error('\P{Is_Scx=/a/_	BOPO}');
Expect(1, 65381, '\p{Is_Scx=bopo}', "");
Expect(0, 65381, '\p{^Is_Scx=bopo}', "");
Expect(0, 65381, '\P{Is_Scx=bopo}', "");
Expect(1, 65381, '\P{^Is_Scx=bopo}', "");
Expect(0, 65382, '\p{Is_Scx=bopo}', "");
Expect(1, 65382, '\p{^Is_Scx=bopo}', "");
Expect(1, 65382, '\P{Is_Scx=bopo}', "");
Expect(0, 65382, '\P{^Is_Scx=bopo}', "");
Expect(1, 65381, '\p{Is_Scx=- Bopo}', "");
Expect(0, 65381, '\p{^Is_Scx=- Bopo}', "");
Expect(0, 65381, '\P{Is_Scx=- Bopo}', "");
Expect(1, 65381, '\P{^Is_Scx=- Bopo}', "");
Expect(0, 65382, '\p{Is_Scx=- Bopo}', "");
Expect(1, 65382, '\p{^Is_Scx=- Bopo}', "");
Expect(1, 65382, '\P{Is_Scx=- Bopo}', "");
Expect(0, 65382, '\P{^Is_Scx=- Bopo}', "");
Error('\p{Script_Extensions=:=_	Brahmi}');
Error('\P{Script_Extensions=:=_	Brahmi}');
Expect(1, 69743, '\p{Script_Extensions=brahmi}', "");
Expect(0, 69743, '\p{^Script_Extensions=brahmi}', "");
Expect(0, 69743, '\P{Script_Extensions=brahmi}', "");
Expect(1, 69743, '\P{^Script_Extensions=brahmi}', "");
Expect(0, 69744, '\p{Script_Extensions=brahmi}', "");
Expect(1, 69744, '\p{^Script_Extensions=brahmi}', "");
Expect(1, 69744, '\P{Script_Extensions=brahmi}', "");
Expect(0, 69744, '\P{^Script_Extensions=brahmi}', "");
Expect(1, 69743, '\p{Script_Extensions=	_brahmi}', "");
Expect(0, 69743, '\p{^Script_Extensions=	_brahmi}', "");
Expect(0, 69743, '\P{Script_Extensions=	_brahmi}', "");
Expect(1, 69743, '\P{^Script_Extensions=	_brahmi}', "");
Expect(0, 69744, '\p{Script_Extensions=	_brahmi}', "");
Expect(1, 69744, '\p{^Script_Extensions=	_brahmi}', "");
Expect(1, 69744, '\P{Script_Extensions=	_brahmi}', "");
Expect(0, 69744, '\P{^Script_Extensions=	_brahmi}', "");
Error('\p{Scx=--Brah/a/}');
Error('\P{Scx=--Brah/a/}');
Expect(1, 69743, '\p{Scx:   brah}', "");
Expect(0, 69743, '\p{^Scx:   brah}', "");
Expect(0, 69743, '\P{Scx:   brah}', "");
Expect(1, 69743, '\P{^Scx:   brah}', "");
Expect(0, 69744, '\p{Scx:   brah}', "");
Expect(1, 69744, '\p{^Scx:   brah}', "");
Expect(1, 69744, '\P{Scx:   brah}', "");
Expect(0, 69744, '\P{^Scx:   brah}', "");
Expect(1, 69743, '\p{Scx=_	BRAH}', "");
Expect(0, 69743, '\p{^Scx=_	BRAH}', "");
Expect(0, 69743, '\P{Scx=_	BRAH}', "");
Expect(1, 69743, '\P{^Scx=_	BRAH}', "");
Expect(0, 69744, '\p{Scx=_	BRAH}', "");
Expect(1, 69744, '\p{^Scx=_	BRAH}', "");
Expect(1, 69744, '\P{Scx=_	BRAH}', "");
Expect(0, 69744, '\P{^Scx=_	BRAH}', "");
Error('\p{Is_Script_Extensions=-BRAHMI/a/}');
Error('\P{Is_Script_Extensions=-BRAHMI/a/}');
Expect(1, 69743, '\p{Is_Script_Extensions=brahmi}', "");
Expect(0, 69743, '\p{^Is_Script_Extensions=brahmi}', "");
Expect(0, 69743, '\P{Is_Script_Extensions=brahmi}', "");
Expect(1, 69743, '\P{^Is_Script_Extensions=brahmi}', "");
Expect(0, 69744, '\p{Is_Script_Extensions=brahmi}', "");
Expect(1, 69744, '\p{^Is_Script_Extensions=brahmi}', "");
Expect(1, 69744, '\P{Is_Script_Extensions=brahmi}', "");
Expect(0, 69744, '\P{^Is_Script_Extensions=brahmi}', "");
Expect(1, 69743, '\p{Is_Script_Extensions:	_ Brahmi}', "");
Expect(0, 69743, '\p{^Is_Script_Extensions:	_ Brahmi}', "");
Expect(0, 69743, '\P{Is_Script_Extensions:	_ Brahmi}', "");
Expect(1, 69743, '\P{^Is_Script_Extensions:	_ Brahmi}', "");
Expect(0, 69744, '\p{Is_Script_Extensions:	_ Brahmi}', "");
Expect(1, 69744, '\p{^Is_Script_Extensions:	_ Brahmi}', "");
Expect(1, 69744, '\P{Is_Script_Extensions:	_ Brahmi}', "");
Expect(0, 69744, '\P{^Is_Script_Extensions:	_ Brahmi}', "");
Error('\p{Is_Scx=-:=brah}');
Error('\P{Is_Scx=-:=brah}');
Expect(1, 69743, '\p{Is_Scx=brah}', "");
Expect(0, 69743, '\p{^Is_Scx=brah}', "");
Expect(0, 69743, '\P{Is_Scx=brah}', "");
Expect(1, 69743, '\P{^Is_Scx=brah}', "");
Expect(0, 69744, '\p{Is_Scx=brah}', "");
Expect(1, 69744, '\p{^Is_Scx=brah}', "");
Expect(1, 69744, '\P{Is_Scx=brah}', "");
Expect(0, 69744, '\P{^Is_Scx=brah}', "");
Expect(1, 69743, '\p{Is_Scx=BRAH}', "");
Expect(0, 69743, '\p{^Is_Scx=BRAH}', "");
Expect(0, 69743, '\P{Is_Scx=BRAH}', "");
Expect(1, 69743, '\P{^Is_Scx=BRAH}', "");
Expect(0, 69744, '\p{Is_Scx=BRAH}', "");
Expect(1, 69744, '\p{^Is_Scx=BRAH}', "");
Expect(1, 69744, '\P{Is_Scx=BRAH}', "");
Expect(0, 69744, '\P{^Is_Scx=BRAH}', "");
Error('\p{Script_Extensions=:=_Braille}');
Error('\P{Script_Extensions=:=_Braille}');
Expect(1, 10495, '\p{Script_Extensions=braille}', "");
Expect(0, 10495, '\p{^Script_Extensions=braille}', "");
Expect(0, 10495, '\P{Script_Extensions=braille}', "");
Expect(1, 10495, '\P{^Script_Extensions=braille}', "");
Expect(0, 10496, '\p{Script_Extensions=braille}', "");
Expect(1, 10496, '\p{^Script_Extensions=braille}', "");
Expect(1, 10496, '\P{Script_Extensions=braille}', "");
Expect(0, 10496, '\P{^Script_Extensions=braille}', "");
Expect(1, 10495, '\p{Script_Extensions:		braille}', "");
Expect(0, 10495, '\p{^Script_Extensions:		braille}', "");
Expect(0, 10495, '\P{Script_Extensions:		braille}', "");
Expect(1, 10495, '\P{^Script_Extensions:		braille}', "");
Expect(0, 10496, '\p{Script_Extensions:		braille}', "");
Expect(1, 10496, '\p{^Script_Extensions:		braille}', "");
Expect(1, 10496, '\P{Script_Extensions:		braille}', "");
Expect(0, 10496, '\P{^Script_Extensions:		braille}', "");
Error('\p{Scx=:=Brai}');
Error('\P{Scx=:=Brai}');
Expect(1, 10495, '\p{Scx=brai}', "");
Expect(0, 10495, '\p{^Scx=brai}', "");
Expect(0, 10495, '\P{Scx=brai}', "");
Expect(1, 10495, '\P{^Scx=brai}', "");
Expect(0, 10496, '\p{Scx=brai}', "");
Expect(1, 10496, '\p{^Scx=brai}', "");
Expect(1, 10496, '\P{Scx=brai}', "");
Expect(0, 10496, '\P{^Scx=brai}', "");
Expect(1, 10495, '\p{Scx=-Brai}', "");
Expect(0, 10495, '\p{^Scx=-Brai}', "");
Expect(0, 10495, '\P{Scx=-Brai}', "");
Expect(1, 10495, '\P{^Scx=-Brai}', "");
Expect(0, 10496, '\p{Scx=-Brai}', "");
Expect(1, 10496, '\p{^Scx=-Brai}', "");
Expect(1, 10496, '\P{Scx=-Brai}', "");
Expect(0, 10496, '\P{^Scx=-Brai}', "");
Error('\p{Is_Script_Extensions=-:=braille}');
Error('\P{Is_Script_Extensions=-:=braille}');
Expect(1, 10495, '\p{Is_Script_Extensions=braille}', "");
Expect(0, 10495, '\p{^Is_Script_Extensions=braille}', "");
Expect(0, 10495, '\P{Is_Script_Extensions=braille}', "");
Expect(1, 10495, '\P{^Is_Script_Extensions=braille}', "");
Expect(0, 10496, '\p{Is_Script_Extensions=braille}', "");
Expect(1, 10496, '\p{^Is_Script_Extensions=braille}', "");
Expect(1, 10496, '\P{Is_Script_Extensions=braille}', "");
Expect(0, 10496, '\P{^Is_Script_Extensions=braille}', "");
Expect(1, 10495, '\p{Is_Script_Extensions=__Braille}', "");
Expect(0, 10495, '\p{^Is_Script_Extensions=__Braille}', "");
Expect(0, 10495, '\P{Is_Script_Extensions=__Braille}', "");
Expect(1, 10495, '\P{^Is_Script_Extensions=__Braille}', "");
Expect(0, 10496, '\p{Is_Script_Extensions=__Braille}', "");
Expect(1, 10496, '\p{^Is_Script_Extensions=__Braille}', "");
Expect(1, 10496, '\P{Is_Script_Extensions=__Braille}', "");
Expect(0, 10496, '\P{^Is_Script_Extensions=__Braille}', "");
Error('\p{Is_Scx=	brai/a/}');
Error('\P{Is_Scx=	brai/a/}');
Expect(1, 10495, '\p{Is_Scx=brai}', "");
Expect(0, 10495, '\p{^Is_Scx=brai}', "");
Expect(0, 10495, '\P{Is_Scx=brai}', "");
Expect(1, 10495, '\P{^Is_Scx=brai}', "");
Expect(0, 10496, '\p{Is_Scx=brai}', "");
Expect(1, 10496, '\p{^Is_Scx=brai}', "");
Expect(1, 10496, '\P{Is_Scx=brai}', "");
Expect(0, 10496, '\P{^Is_Scx=brai}', "");
Expect(1, 10495, '\p{Is_Scx=-_BRAI}', "");
Expect(0, 10495, '\p{^Is_Scx=-_BRAI}', "");
Expect(0, 10495, '\P{Is_Scx=-_BRAI}', "");
Expect(1, 10495, '\P{^Is_Scx=-_BRAI}', "");
Expect(0, 10496, '\p{Is_Scx=-_BRAI}', "");
Expect(1, 10496, '\p{^Is_Scx=-_BRAI}', "");
Expect(1, 10496, '\P{Is_Scx=-_BRAI}', "");
Expect(0, 10496, '\P{^Is_Scx=-_BRAI}', "");
Error('\p{Script_Extensions=	_BUGINESE:=}');
Error('\P{Script_Extensions=	_BUGINESE:=}');
Expect(1, 43471, '\p{Script_Extensions=buginese}', "");
Expect(0, 43471, '\p{^Script_Extensions=buginese}', "");
Expect(0, 43471, '\P{Script_Extensions=buginese}', "");
Expect(1, 43471, '\P{^Script_Extensions=buginese}', "");
Expect(0, 43472, '\p{Script_Extensions=buginese}', "");
Expect(1, 43472, '\p{^Script_Extensions=buginese}', "");
Expect(1, 43472, '\P{Script_Extensions=buginese}', "");
Expect(0, 43472, '\P{^Script_Extensions=buginese}', "");
Expect(1, 43471, '\p{Script_Extensions=- BUGINESE}', "");
Expect(0, 43471, '\p{^Script_Extensions=- BUGINESE}', "");
Expect(0, 43471, '\P{Script_Extensions=- BUGINESE}', "");
Expect(1, 43471, '\P{^Script_Extensions=- BUGINESE}', "");
Expect(0, 43472, '\p{Script_Extensions=- BUGINESE}', "");
Expect(1, 43472, '\p{^Script_Extensions=- BUGINESE}', "");
Expect(1, 43472, '\P{Script_Extensions=- BUGINESE}', "");
Expect(0, 43472, '\P{^Script_Extensions=- BUGINESE}', "");
Error('\p{Scx=_-Bugi:=}');
Error('\P{Scx=_-Bugi:=}');
Expect(1, 43471, '\p{Scx=bugi}', "");
Expect(0, 43471, '\p{^Scx=bugi}', "");
Expect(0, 43471, '\P{Scx=bugi}', "");
Expect(1, 43471, '\P{^Scx=bugi}', "");
Expect(0, 43472, '\p{Scx=bugi}', "");
Expect(1, 43472, '\p{^Scx=bugi}', "");
Expect(1, 43472, '\P{Scx=bugi}', "");
Expect(0, 43472, '\P{^Scx=bugi}', "");
Expect(1, 43471, '\p{Scx:		-Bugi}', "");
Expect(0, 43471, '\p{^Scx:		-Bugi}', "");
Expect(0, 43471, '\P{Scx:		-Bugi}', "");
Expect(1, 43471, '\P{^Scx:		-Bugi}', "");
Expect(0, 43472, '\p{Scx:		-Bugi}', "");
Expect(1, 43472, '\p{^Scx:		-Bugi}', "");
Expect(1, 43472, '\P{Scx:		-Bugi}', "");
Expect(0, 43472, '\P{^Scx:		-Bugi}', "");
Error('\p{Is_Script_Extensions= :=Buginese}');
Error('\P{Is_Script_Extensions= :=Buginese}');
Expect(1, 43471, '\p{Is_Script_Extensions=buginese}', "");
Expect(0, 43471, '\p{^Is_Script_Extensions=buginese}', "");
Expect(0, 43471, '\P{Is_Script_Extensions=buginese}', "");
Expect(1, 43471, '\P{^Is_Script_Extensions=buginese}', "");
Expect(0, 43472, '\p{Is_Script_Extensions=buginese}', "");
Expect(1, 43472, '\p{^Is_Script_Extensions=buginese}', "");
Expect(1, 43472, '\P{Is_Script_Extensions=buginese}', "");
Expect(0, 43472, '\P{^Is_Script_Extensions=buginese}', "");
Expect(1, 43471, '\p{Is_Script_Extensions:   __Buginese}', "");
Expect(0, 43471, '\p{^Is_Script_Extensions:   __Buginese}', "");
Expect(0, 43471, '\P{Is_Script_Extensions:   __Buginese}', "");
Expect(1, 43471, '\P{^Is_Script_Extensions:   __Buginese}', "");
Expect(0, 43472, '\p{Is_Script_Extensions:   __Buginese}', "");
Expect(1, 43472, '\p{^Is_Script_Extensions:   __Buginese}', "");
Expect(1, 43472, '\P{Is_Script_Extensions:   __Buginese}', "");
Expect(0, 43472, '\P{^Is_Script_Extensions:   __Buginese}', "");
Error('\p{Is_Scx=_:=Bugi}');
Error('\P{Is_Scx=_:=Bugi}');
Expect(1, 43471, '\p{Is_Scx=bugi}', "");
Expect(0, 43471, '\p{^Is_Scx=bugi}', "");
Expect(0, 43471, '\P{Is_Scx=bugi}', "");
Expect(1, 43471, '\P{^Is_Scx=bugi}', "");
Expect(0, 43472, '\p{Is_Scx=bugi}', "");
Expect(1, 43472, '\p{^Is_Scx=bugi}', "");
Expect(1, 43472, '\P{Is_Scx=bugi}', "");
Expect(0, 43472, '\P{^Is_Scx=bugi}', "");
Expect(1, 43471, '\p{Is_Scx:   -	BUGI}', "");
Expect(0, 43471, '\p{^Is_Scx:   -	BUGI}', "");
Expect(0, 43471, '\P{Is_Scx:   -	BUGI}', "");
Expect(1, 43471, '\P{^Is_Scx:   -	BUGI}', "");
Expect(0, 43472, '\p{Is_Scx:   -	BUGI}', "");
Expect(1, 43472, '\p{^Is_Scx:   -	BUGI}', "");
Expect(1, 43472, '\P{Is_Scx:   -	BUGI}', "");
Expect(0, 43472, '\P{^Is_Scx:   -	BUGI}', "");
Error('\p{Script_Extensions=/a/__buhid}');
Error('\P{Script_Extensions=/a/__buhid}');
Expect(1, 5971, '\p{Script_Extensions=buhid}', "");
Expect(0, 5971, '\p{^Script_Extensions=buhid}', "");
Expect(0, 5971, '\P{Script_Extensions=buhid}', "");
Expect(1, 5971, '\P{^Script_Extensions=buhid}', "");
Expect(0, 5972, '\p{Script_Extensions=buhid}', "");
Expect(1, 5972, '\p{^Script_Extensions=buhid}', "");
Expect(1, 5972, '\P{Script_Extensions=buhid}', "");
Expect(0, 5972, '\P{^Script_Extensions=buhid}', "");
Expect(1, 5971, '\p{Script_Extensions=  Buhid}', "");
Expect(0, 5971, '\p{^Script_Extensions=  Buhid}', "");
Expect(0, 5971, '\P{Script_Extensions=  Buhid}', "");
Expect(1, 5971, '\P{^Script_Extensions=  Buhid}', "");
Expect(0, 5972, '\p{Script_Extensions=  Buhid}', "");
Expect(1, 5972, '\p{^Script_Extensions=  Buhid}', "");
Expect(1, 5972, '\P{Script_Extensions=  Buhid}', "");
Expect(0, 5972, '\P{^Script_Extensions=  Buhid}', "");
Error('\p{Scx=:=BUHD}');
Error('\P{Scx=:=BUHD}');
Expect(1, 5971, '\p{Scx=buhd}', "");
Expect(0, 5971, '\p{^Scx=buhd}', "");
Expect(0, 5971, '\P{Scx=buhd}', "");
Expect(1, 5971, '\P{^Scx=buhd}', "");
Expect(0, 5972, '\p{Scx=buhd}', "");
Expect(1, 5972, '\p{^Scx=buhd}', "");
Expect(1, 5972, '\P{Scx=buhd}', "");
Expect(0, 5972, '\P{^Scx=buhd}', "");
Expect(1, 5971, '\p{Scx=-_Buhd}', "");
Expect(0, 5971, '\p{^Scx=-_Buhd}', "");
Expect(0, 5971, '\P{Scx=-_Buhd}', "");
Expect(1, 5971, '\P{^Scx=-_Buhd}', "");
Expect(0, 5972, '\p{Scx=-_Buhd}', "");
Expect(1, 5972, '\p{^Scx=-_Buhd}', "");
Expect(1, 5972, '\P{Scx=-_Buhd}', "");
Expect(0, 5972, '\P{^Scx=-_Buhd}', "");
Error('\p{Is_Script_Extensions=  Buhid/a/}');
Error('\P{Is_Script_Extensions=  Buhid/a/}');
Expect(1, 5971, '\p{Is_Script_Extensions=buhid}', "");
Expect(0, 5971, '\p{^Is_Script_Extensions=buhid}', "");
Expect(0, 5971, '\P{Is_Script_Extensions=buhid}', "");
Expect(1, 5971, '\P{^Is_Script_Extensions=buhid}', "");
Expect(0, 5972, '\p{Is_Script_Extensions=buhid}', "");
Expect(1, 5972, '\p{^Is_Script_Extensions=buhid}', "");
Expect(1, 5972, '\P{Is_Script_Extensions=buhid}', "");
Expect(0, 5972, '\P{^Is_Script_Extensions=buhid}', "");
Expect(1, 5971, '\p{Is_Script_Extensions=	_Buhid}', "");
Expect(0, 5971, '\p{^Is_Script_Extensions=	_Buhid}', "");
Expect(0, 5971, '\P{Is_Script_Extensions=	_Buhid}', "");
Expect(1, 5971, '\P{^Is_Script_Extensions=	_Buhid}', "");
Expect(0, 5972, '\p{Is_Script_Extensions=	_Buhid}', "");
Expect(1, 5972, '\p{^Is_Script_Extensions=	_Buhid}', "");
Expect(1, 5972, '\P{Is_Script_Extensions=	_Buhid}', "");
Expect(0, 5972, '\P{^Is_Script_Extensions=	_Buhid}', "");
Error('\p{Is_Scx=:= 	buhd}');
Error('\P{Is_Scx=:= 	buhd}');
Expect(1, 5971, '\p{Is_Scx=buhd}', "");
Expect(0, 5971, '\p{^Is_Scx=buhd}', "");
Expect(0, 5971, '\P{Is_Scx=buhd}', "");
Expect(1, 5971, '\P{^Is_Scx=buhd}', "");
Expect(0, 5972, '\p{Is_Scx=buhd}', "");
Expect(1, 5972, '\p{^Is_Scx=buhd}', "");
Expect(1, 5972, '\P{Is_Scx=buhd}', "");
Expect(0, 5972, '\P{^Is_Scx=buhd}', "");
Expect(1, 5971, '\p{Is_Scx=- Buhd}', "");
Expect(0, 5971, '\p{^Is_Scx=- Buhd}', "");
Expect(0, 5971, '\P{Is_Scx=- Buhd}', "");
Expect(1, 5971, '\P{^Is_Scx=- Buhd}', "");
Expect(0, 5972, '\p{Is_Scx=- Buhd}', "");
Expect(1, 5972, '\p{^Is_Scx=- Buhd}', "");
Expect(1, 5972, '\P{Is_Scx=- Buhd}', "");
Expect(0, 5972, '\P{^Is_Scx=- Buhd}', "");
Error('\p{Script_Extensions=:=Chakma}');
Error('\P{Script_Extensions=:=Chakma}');
Expect(1, 69955, '\p{Script_Extensions=chakma}', "");
Expect(0, 69955, '\p{^Script_Extensions=chakma}', "");
Expect(0, 69955, '\P{Script_Extensions=chakma}', "");
Expect(1, 69955, '\P{^Script_Extensions=chakma}', "");
Expect(0, 69956, '\p{Script_Extensions=chakma}', "");
Expect(1, 69956, '\p{^Script_Extensions=chakma}', "");
Expect(1, 69956, '\P{Script_Extensions=chakma}', "");
Expect(0, 69956, '\P{^Script_Extensions=chakma}', "");
Expect(1, 69955, '\p{Script_Extensions:		_CHAKMA}', "");
Expect(0, 69955, '\p{^Script_Extensions:		_CHAKMA}', "");
Expect(0, 69955, '\P{Script_Extensions:		_CHAKMA}', "");
Expect(1, 69955, '\P{^Script_Extensions:		_CHAKMA}', "");
Expect(0, 69956, '\p{Script_Extensions:		_CHAKMA}', "");
Expect(1, 69956, '\p{^Script_Extensions:		_CHAKMA}', "");
Expect(1, 69956, '\P{Script_Extensions:		_CHAKMA}', "");
Expect(0, 69956, '\P{^Script_Extensions:		_CHAKMA}', "");
Error('\p{Scx=	 CAKM/a/}');
Error('\P{Scx=	 CAKM/a/}');
Expect(1, 69955, '\p{Scx=cakm}', "");
Expect(0, 69955, '\p{^Scx=cakm}', "");
Expect(0, 69955, '\P{Scx=cakm}', "");
Expect(1, 69955, '\P{^Scx=cakm}', "");
Expect(0, 69956, '\p{Scx=cakm}', "");
Expect(1, 69956, '\p{^Scx=cakm}', "");
Expect(1, 69956, '\P{Scx=cakm}', "");
Expect(0, 69956, '\P{^Scx=cakm}', "");
Expect(1, 69955, '\p{Scx=-_Cakm}', "");
Expect(0, 69955, '\p{^Scx=-_Cakm}', "");
Expect(0, 69955, '\P{Scx=-_Cakm}', "");
Expect(1, 69955, '\P{^Scx=-_Cakm}', "");
Expect(0, 69956, '\p{Scx=-_Cakm}', "");
Expect(1, 69956, '\p{^Scx=-_Cakm}', "");
Expect(1, 69956, '\P{Scx=-_Cakm}', "");
Expect(0, 69956, '\P{^Scx=-_Cakm}', "");
Error('\p{Is_Script_Extensions=	_Chakma:=}');
Error('\P{Is_Script_Extensions=	_Chakma:=}');
Expect(1, 69955, '\p{Is_Script_Extensions=chakma}', "");
Expect(0, 69955, '\p{^Is_Script_Extensions=chakma}', "");
Expect(0, 69955, '\P{Is_Script_Extensions=chakma}', "");
Expect(1, 69955, '\P{^Is_Script_Extensions=chakma}', "");
Expect(0, 69956, '\p{Is_Script_Extensions=chakma}', "");
Expect(1, 69956, '\p{^Is_Script_Extensions=chakma}', "");
Expect(1, 69956, '\P{Is_Script_Extensions=chakma}', "");
Expect(0, 69956, '\P{^Is_Script_Extensions=chakma}', "");
Expect(1, 69955, '\p{Is_Script_Extensions=-_Chakma}', "");
Expect(0, 69955, '\p{^Is_Script_Extensions=-_Chakma}', "");
Expect(0, 69955, '\P{Is_Script_Extensions=-_Chakma}', "");
Expect(1, 69955, '\P{^Is_Script_Extensions=-_Chakma}', "");
Expect(0, 69956, '\p{Is_Script_Extensions=-_Chakma}', "");
Expect(1, 69956, '\p{^Is_Script_Extensions=-_Chakma}', "");
Expect(1, 69956, '\P{Is_Script_Extensions=-_Chakma}', "");
Expect(0, 69956, '\P{^Is_Script_Extensions=-_Chakma}', "");
Error('\p{Is_Scx=/a/	cakm}');
Error('\P{Is_Scx=/a/	cakm}');
Expect(1, 69955, '\p{Is_Scx=cakm}', "");
Expect(0, 69955, '\p{^Is_Scx=cakm}', "");
Expect(0, 69955, '\P{Is_Scx=cakm}', "");
Expect(1, 69955, '\P{^Is_Scx=cakm}', "");
Expect(0, 69956, '\p{Is_Scx=cakm}', "");
Expect(1, 69956, '\p{^Is_Scx=cakm}', "");
Expect(1, 69956, '\P{Is_Scx=cakm}', "");
Expect(0, 69956, '\P{^Is_Scx=cakm}', "");
Expect(1, 69955, '\p{Is_Scx=__cakm}', "");
Expect(0, 69955, '\p{^Is_Scx=__cakm}', "");
Expect(0, 69955, '\P{Is_Scx=__cakm}', "");
Expect(1, 69955, '\P{^Is_Scx=__cakm}', "");
Expect(0, 69956, '\p{Is_Scx=__cakm}', "");
Expect(1, 69956, '\p{^Is_Scx=__cakm}', "");
Expect(1, 69956, '\P{Is_Scx=__cakm}', "");
Expect(0, 69956, '\P{^Is_Scx=__cakm}', "");
Error('\p{Script_Extensions:   _:=Canadian_ABORIGINAL}');
Error('\P{Script_Extensions:   _:=Canadian_ABORIGINAL}');
Expect(1, 6389, '\p{Script_Extensions:   canadianaboriginal}', "");
Expect(0, 6389, '\p{^Script_Extensions:   canadianaboriginal}', "");
Expect(0, 6389, '\P{Script_Extensions:   canadianaboriginal}', "");
Expect(1, 6389, '\P{^Script_Extensions:   canadianaboriginal}', "");
Expect(0, 6390, '\p{Script_Extensions:   canadianaboriginal}', "");
Expect(1, 6390, '\p{^Script_Extensions:   canadianaboriginal}', "");
Expect(1, 6390, '\P{Script_Extensions:   canadianaboriginal}', "");
Expect(0, 6390, '\P{^Script_Extensions:   canadianaboriginal}', "");
Expect(1, 6389, '\p{Script_Extensions=_-canadian_ABORIGINAL}', "");
Expect(0, 6389, '\p{^Script_Extensions=_-canadian_ABORIGINAL}', "");
Expect(0, 6389, '\P{Script_Extensions=_-canadian_ABORIGINAL}', "");
Expect(1, 6389, '\P{^Script_Extensions=_-canadian_ABORIGINAL}', "");
Expect(0, 6390, '\p{Script_Extensions=_-canadian_ABORIGINAL}', "");
Expect(1, 6390, '\p{^Script_Extensions=_-canadian_ABORIGINAL}', "");
Expect(1, 6390, '\P{Script_Extensions=_-canadian_ABORIGINAL}', "");
Expect(0, 6390, '\P{^Script_Extensions=_-canadian_ABORIGINAL}', "");
Error('\p{Scx:/a/_	cans}');
Error('\P{Scx:/a/_	cans}');
Expect(1, 6389, '\p{Scx=cans}', "");
Expect(0, 6389, '\p{^Scx=cans}', "");
Expect(0, 6389, '\P{Scx=cans}', "");
Expect(1, 6389, '\P{^Scx=cans}', "");
Expect(0, 6390, '\p{Scx=cans}', "");
Expect(1, 6390, '\p{^Scx=cans}', "");
Expect(1, 6390, '\P{Scx=cans}', "");
Expect(0, 6390, '\P{^Scx=cans}', "");
Expect(1, 6389, '\p{Scx=	Cans}', "");
Expect(0, 6389, '\p{^Scx=	Cans}', "");
Expect(0, 6389, '\P{Scx=	Cans}', "");
Expect(1, 6389, '\P{^Scx=	Cans}', "");
Expect(0, 6390, '\p{Scx=	Cans}', "");
Expect(1, 6390, '\p{^Scx=	Cans}', "");
Expect(1, 6390, '\P{Scx=	Cans}', "");
Expect(0, 6390, '\P{^Scx=	Cans}', "");
Error('\p{Is_Script_Extensions=	-canadian_Aboriginal:=}');
Error('\P{Is_Script_Extensions=	-canadian_Aboriginal:=}');
Expect(1, 6389, '\p{Is_Script_Extensions=canadianaboriginal}', "");
Expect(0, 6389, '\p{^Is_Script_Extensions=canadianaboriginal}', "");
Expect(0, 6389, '\P{Is_Script_Extensions=canadianaboriginal}', "");
Expect(1, 6389, '\P{^Is_Script_Extensions=canadianaboriginal}', "");
Expect(0, 6390, '\p{Is_Script_Extensions=canadianaboriginal}', "");
Expect(1, 6390, '\p{^Is_Script_Extensions=canadianaboriginal}', "");
Expect(1, 6390, '\P{Is_Script_Extensions=canadianaboriginal}', "");
Expect(0, 6390, '\P{^Is_Script_Extensions=canadianaboriginal}', "");
Expect(1, 6389, '\p{Is_Script_Extensions=	canadian_Aboriginal}', "");
Expect(0, 6389, '\p{^Is_Script_Extensions=	canadian_Aboriginal}', "");
Expect(0, 6389, '\P{Is_Script_Extensions=	canadian_Aboriginal}', "");
Expect(1, 6389, '\P{^Is_Script_Extensions=	canadian_Aboriginal}', "");
Expect(0, 6390, '\p{Is_Script_Extensions=	canadian_Aboriginal}', "");
Expect(1, 6390, '\p{^Is_Script_Extensions=	canadian_Aboriginal}', "");
Expect(1, 6390, '\P{Is_Script_Extensions=	canadian_Aboriginal}', "");
Expect(0, 6390, '\P{^Is_Script_Extensions=	canadian_Aboriginal}', "");
Error('\p{Is_Scx=/a/	Cans}');
Error('\P{Is_Scx=/a/	Cans}');
Expect(1, 6389, '\p{Is_Scx=cans}', "");
Expect(0, 6389, '\p{^Is_Scx=cans}', "");
Expect(0, 6389, '\P{Is_Scx=cans}', "");
Expect(1, 6389, '\P{^Is_Scx=cans}', "");
Expect(0, 6390, '\p{Is_Scx=cans}', "");
Expect(1, 6390, '\p{^Is_Scx=cans}', "");
Expect(1, 6390, '\P{Is_Scx=cans}', "");
Expect(0, 6390, '\P{^Is_Scx=cans}', "");
Expect(1, 6389, '\p{Is_Scx=--Cans}', "");
Expect(0, 6389, '\p{^Is_Scx=--Cans}', "");
Expect(0, 6389, '\P{Is_Scx=--Cans}', "");
Expect(1, 6389, '\P{^Is_Scx=--Cans}', "");
Expect(0, 6390, '\p{Is_Scx=--Cans}', "");
Expect(1, 6390, '\p{^Is_Scx=--Cans}', "");
Expect(1, 6390, '\P{Is_Scx=--Cans}', "");
Expect(0, 6390, '\P{^Is_Scx=--Cans}', "");
Error('\p{Script_Extensions=-/a/Carian}');
Error('\P{Script_Extensions=-/a/Carian}');
Expect(1, 66256, '\p{Script_Extensions=carian}', "");
Expect(0, 66256, '\p{^Script_Extensions=carian}', "");
Expect(0, 66256, '\P{Script_Extensions=carian}', "");
Expect(1, 66256, '\P{^Script_Extensions=carian}', "");
Expect(0, 66257, '\p{Script_Extensions=carian}', "");
Expect(1, 66257, '\p{^Script_Extensions=carian}', "");
Expect(1, 66257, '\P{Script_Extensions=carian}', "");
Expect(0, 66257, '\P{^Script_Extensions=carian}', "");
Expect(1, 66256, '\p{Script_Extensions=-	carian}', "");
Expect(0, 66256, '\p{^Script_Extensions=-	carian}', "");
Expect(0, 66256, '\P{Script_Extensions=-	carian}', "");
Expect(1, 66256, '\P{^Script_Extensions=-	carian}', "");
Expect(0, 66257, '\p{Script_Extensions=-	carian}', "");
Expect(1, 66257, '\p{^Script_Extensions=-	carian}', "");
Expect(1, 66257, '\P{Script_Extensions=-	carian}', "");
Expect(0, 66257, '\P{^Script_Extensions=-	carian}', "");
Error('\p{Scx= /a/cari}');
Error('\P{Scx= /a/cari}');
Expect(1, 66256, '\p{Scx: cari}', "");
Expect(0, 66256, '\p{^Scx: cari}', "");
Expect(0, 66256, '\P{Scx: cari}', "");
Expect(1, 66256, '\P{^Scx: cari}', "");
Expect(0, 66257, '\p{Scx: cari}', "");
Expect(1, 66257, '\p{^Scx: cari}', "");
Expect(1, 66257, '\P{Scx: cari}', "");
Expect(0, 66257, '\P{^Scx: cari}', "");
Expect(1, 66256, '\p{Scx=	Cari}', "");
Expect(0, 66256, '\p{^Scx=	Cari}', "");
Expect(0, 66256, '\P{Scx=	Cari}', "");
Expect(1, 66256, '\P{^Scx=	Cari}', "");
Expect(0, 66257, '\p{Scx=	Cari}', "");
Expect(1, 66257, '\p{^Scx=	Cari}', "");
Expect(1, 66257, '\P{Scx=	Cari}', "");
Expect(0, 66257, '\P{^Scx=	Cari}', "");
Error('\p{Is_Script_Extensions=:=-	carian}');
Error('\P{Is_Script_Extensions=:=-	carian}');
Expect(1, 66256, '\p{Is_Script_Extensions=carian}', "");
Expect(0, 66256, '\p{^Is_Script_Extensions=carian}', "");
Expect(0, 66256, '\P{Is_Script_Extensions=carian}', "");
Expect(1, 66256, '\P{^Is_Script_Extensions=carian}', "");
Expect(0, 66257, '\p{Is_Script_Extensions=carian}', "");
Expect(1, 66257, '\p{^Is_Script_Extensions=carian}', "");
Expect(1, 66257, '\P{Is_Script_Extensions=carian}', "");
Expect(0, 66257, '\P{^Is_Script_Extensions=carian}', "");
Error('\p{Is_Scx: /a/-	Cari}');
Error('\P{Is_Scx: /a/-	Cari}');
Expect(1, 66256, '\p{Is_Scx=cari}', "");
Expect(0, 66256, '\p{^Is_Scx=cari}', "");
Expect(0, 66256, '\P{Is_Scx=cari}', "");
Expect(1, 66256, '\P{^Is_Scx=cari}', "");
Expect(0, 66257, '\p{Is_Scx=cari}', "");
Expect(1, 66257, '\p{^Is_Scx=cari}', "");
Expect(1, 66257, '\P{Is_Scx=cari}', "");
Expect(0, 66257, '\P{^Is_Scx=cari}', "");
Expect(1, 66256, '\p{Is_Scx= Cari}', "");
Expect(0, 66256, '\p{^Is_Scx= Cari}', "");
Expect(0, 66256, '\P{Is_Scx= Cari}', "");
Expect(1, 66256, '\P{^Is_Scx= Cari}', "");
Expect(0, 66257, '\p{Is_Scx= Cari}', "");
Expect(1, 66257, '\p{^Is_Scx= Cari}', "");
Expect(1, 66257, '\P{Is_Scx= Cari}', "");
Expect(0, 66257, '\P{^Is_Scx= Cari}', "");
Error('\p{Script_Extensions=cham:=}');
Error('\P{Script_Extensions=cham:=}');
Expect(1, 43615, '\p{Script_Extensions=cham}', "");
Expect(0, 43615, '\p{^Script_Extensions=cham}', "");
Expect(0, 43615, '\P{Script_Extensions=cham}', "");
Expect(1, 43615, '\P{^Script_Extensions=cham}', "");
Expect(0, 43616, '\p{Script_Extensions=cham}', "");
Expect(1, 43616, '\p{^Script_Extensions=cham}', "");
Expect(1, 43616, '\P{Script_Extensions=cham}', "");
Expect(0, 43616, '\P{^Script_Extensions=cham}', "");
Expect(1, 43615, '\p{Script_Extensions:	-Cham}', "");
Expect(0, 43615, '\p{^Script_Extensions:	-Cham}', "");
Expect(0, 43615, '\P{Script_Extensions:	-Cham}', "");
Expect(1, 43615, '\P{^Script_Extensions:	-Cham}', "");
Expect(0, 43616, '\p{Script_Extensions:	-Cham}', "");
Expect(1, 43616, '\p{^Script_Extensions:	-Cham}', "");
Expect(1, 43616, '\P{Script_Extensions:	-Cham}', "");
Expect(0, 43616, '\P{^Script_Extensions:	-Cham}', "");
Error('\p{Scx=:=Cham}');
Error('\P{Scx=:=Cham}');
Expect(1, 43615, '\p{Scx=cham}', "");
Expect(0, 43615, '\p{^Scx=cham}', "");
Expect(0, 43615, '\P{Scx=cham}', "");
Expect(1, 43615, '\P{^Scx=cham}', "");
Expect(0, 43616, '\p{Scx=cham}', "");
Expect(1, 43616, '\p{^Scx=cham}', "");
Expect(1, 43616, '\P{Scx=cham}', "");
Expect(0, 43616, '\P{^Scx=cham}', "");
Expect(1, 43615, '\p{Scx=	Cham}', "");
Expect(0, 43615, '\p{^Scx=	Cham}', "");
Expect(0, 43615, '\P{Scx=	Cham}', "");
Expect(1, 43615, '\P{^Scx=	Cham}', "");
Expect(0, 43616, '\p{Scx=	Cham}', "");
Expect(1, 43616, '\p{^Scx=	Cham}', "");
Expect(1, 43616, '\P{Scx=	Cham}', "");
Expect(0, 43616, '\P{^Scx=	Cham}', "");
Error('\p{Is_Script_Extensions= :=Cham}');
Error('\P{Is_Script_Extensions= :=Cham}');
Expect(1, 43615, '\p{Is_Script_Extensions=cham}', "");
Expect(0, 43615, '\p{^Is_Script_Extensions=cham}', "");
Expect(0, 43615, '\P{Is_Script_Extensions=cham}', "");
Expect(1, 43615, '\P{^Is_Script_Extensions=cham}', "");
Expect(0, 43616, '\p{Is_Script_Extensions=cham}', "");
Expect(1, 43616, '\p{^Is_Script_Extensions=cham}', "");
Expect(1, 43616, '\P{Is_Script_Extensions=cham}', "");
Expect(0, 43616, '\P{^Is_Script_Extensions=cham}', "");
Expect(1, 43615, '\p{Is_Script_Extensions=  Cham}', "");
Expect(0, 43615, '\p{^Is_Script_Extensions=  Cham}', "");
Expect(0, 43615, '\P{Is_Script_Extensions=  Cham}', "");
Expect(1, 43615, '\P{^Is_Script_Extensions=  Cham}', "");
Expect(0, 43616, '\p{Is_Script_Extensions=  Cham}', "");
Expect(1, 43616, '\p{^Is_Script_Extensions=  Cham}', "");
Expect(1, 43616, '\P{Is_Script_Extensions=  Cham}', "");
Expect(0, 43616, '\P{^Is_Script_Extensions=  Cham}', "");
Error('\p{Is_Scx=/a/__Cham}');
Error('\P{Is_Scx=/a/__Cham}');
Expect(1, 43615, '\p{Is_Scx=cham}', "");
Expect(0, 43615, '\p{^Is_Scx=cham}', "");
Expect(0, 43615, '\P{Is_Scx=cham}', "");
Expect(1, 43615, '\P{^Is_Scx=cham}', "");
Expect(0, 43616, '\p{Is_Scx=cham}', "");
Expect(1, 43616, '\p{^Is_Scx=cham}', "");
Expect(1, 43616, '\P{Is_Scx=cham}', "");
Expect(0, 43616, '\P{^Is_Scx=cham}', "");
Expect(1, 43615, '\p{Is_Scx= cham}', "");
Expect(0, 43615, '\p{^Is_Scx= cham}', "");
Expect(0, 43615, '\P{Is_Scx= cham}', "");
Expect(1, 43615, '\P{^Is_Scx= cham}', "");
Expect(0, 43616, '\p{Is_Scx= cham}', "");
Expect(1, 43616, '\p{^Is_Scx= cham}', "");
Expect(1, 43616, '\P{Is_Scx= cham}', "");
Expect(0, 43616, '\P{^Is_Scx= cham}', "");
Error('\p{Script_Extensions=__Cherokee:=}');
Error('\P{Script_Extensions=__Cherokee:=}');
Expect(1, 5108, '\p{Script_Extensions=cherokee}', "");
Expect(0, 5108, '\p{^Script_Extensions=cherokee}', "");
Expect(0, 5108, '\P{Script_Extensions=cherokee}', "");
Expect(1, 5108, '\P{^Script_Extensions=cherokee}', "");
Expect(0, 5109, '\p{Script_Extensions=cherokee}', "");
Expect(1, 5109, '\p{^Script_Extensions=cherokee}', "");
Expect(1, 5109, '\P{Script_Extensions=cherokee}', "");
Expect(0, 5109, '\P{^Script_Extensions=cherokee}', "");
Expect(1, 5108, '\p{Script_Extensions=_CHEROKEE}', "");
Expect(0, 5108, '\p{^Script_Extensions=_CHEROKEE}', "");
Expect(0, 5108, '\P{Script_Extensions=_CHEROKEE}', "");
Expect(1, 5108, '\P{^Script_Extensions=_CHEROKEE}', "");
Expect(0, 5109, '\p{Script_Extensions=_CHEROKEE}', "");
Expect(1, 5109, '\p{^Script_Extensions=_CHEROKEE}', "");
Expect(1, 5109, '\P{Script_Extensions=_CHEROKEE}', "");
Expect(0, 5109, '\P{^Script_Extensions=_CHEROKEE}', "");
Error('\p{Scx=:=-	CHER}');
Error('\P{Scx=:=-	CHER}');
Expect(1, 5108, '\p{Scx=cher}', "");
Expect(0, 5108, '\p{^Scx=cher}', "");
Expect(0, 5108, '\P{Scx=cher}', "");
Expect(1, 5108, '\P{^Scx=cher}', "");
Expect(0, 5109, '\p{Scx=cher}', "");
Expect(1, 5109, '\p{^Scx=cher}', "");
Expect(1, 5109, '\P{Scx=cher}', "");
Expect(0, 5109, '\P{^Scx=cher}', "");
Expect(1, 5108, '\p{Scx=	-cher}', "");
Expect(0, 5108, '\p{^Scx=	-cher}', "");
Expect(0, 5108, '\P{Scx=	-cher}', "");
Expect(1, 5108, '\P{^Scx=	-cher}', "");
Expect(0, 5109, '\p{Scx=	-cher}', "");
Expect(1, 5109, '\p{^Scx=	-cher}', "");
Expect(1, 5109, '\P{Scx=	-cher}', "");
Expect(0, 5109, '\P{^Scx=	-cher}', "");
Error('\p{Is_Script_Extensions=	CHEROKEE/a/}');
Error('\P{Is_Script_Extensions=	CHEROKEE/a/}');
Expect(1, 5108, '\p{Is_Script_Extensions=cherokee}', "");
Expect(0, 5108, '\p{^Is_Script_Extensions=cherokee}', "");
Expect(0, 5108, '\P{Is_Script_Extensions=cherokee}', "");
Expect(1, 5108, '\P{^Is_Script_Extensions=cherokee}', "");
Expect(0, 5109, '\p{Is_Script_Extensions=cherokee}', "");
Expect(1, 5109, '\p{^Is_Script_Extensions=cherokee}', "");
Expect(1, 5109, '\P{Is_Script_Extensions=cherokee}', "");
Expect(0, 5109, '\P{^Is_Script_Extensions=cherokee}', "");
Expect(1, 5108, '\p{Is_Script_Extensions= 	Cherokee}', "");
Expect(0, 5108, '\p{^Is_Script_Extensions= 	Cherokee}', "");
Expect(0, 5108, '\P{Is_Script_Extensions= 	Cherokee}', "");
Expect(1, 5108, '\P{^Is_Script_Extensions= 	Cherokee}', "");
Expect(0, 5109, '\p{Is_Script_Extensions= 	Cherokee}', "");
Expect(1, 5109, '\p{^Is_Script_Extensions= 	Cherokee}', "");
Expect(1, 5109, '\P{Is_Script_Extensions= 	Cherokee}', "");
Expect(0, 5109, '\P{^Is_Script_Extensions= 	Cherokee}', "");
Error('\p{Is_Scx=:= Cher}');
Error('\P{Is_Scx=:= Cher}');
Expect(1, 5108, '\p{Is_Scx=cher}', "");
Expect(0, 5108, '\p{^Is_Scx=cher}', "");
Expect(0, 5108, '\P{Is_Scx=cher}', "");
Expect(1, 5108, '\P{^Is_Scx=cher}', "");
Expect(0, 5109, '\p{Is_Scx=cher}', "");
Expect(1, 5109, '\p{^Is_Scx=cher}', "");
Expect(1, 5109, '\P{Is_Scx=cher}', "");
Expect(0, 5109, '\P{^Is_Scx=cher}', "");
Expect(1, 5108, '\p{Is_Scx=-_Cher}', "");
Expect(0, 5108, '\p{^Is_Scx=-_Cher}', "");
Expect(0, 5108, '\P{Is_Scx=-_Cher}', "");
Expect(1, 5108, '\P{^Is_Scx=-_Cher}', "");
Expect(0, 5109, '\p{Is_Scx=-_Cher}', "");
Expect(1, 5109, '\p{^Is_Scx=-_Cher}', "");
Expect(1, 5109, '\P{Is_Scx=-_Cher}', "");
Expect(0, 5109, '\P{^Is_Scx=-_Cher}', "");
Error('\p{Script_Extensions=-	COPTIC/a/}');
Error('\P{Script_Extensions=-	COPTIC/a/}');
Expect(1, 11519, '\p{Script_Extensions:   coptic}', "");
Expect(0, 11519, '\p{^Script_Extensions:   coptic}', "");
Expect(0, 11519, '\P{Script_Extensions:   coptic}', "");
Expect(1, 11519, '\P{^Script_Extensions:   coptic}', "");
Expect(0, 11520, '\p{Script_Extensions:   coptic}', "");
Expect(1, 11520, '\p{^Script_Extensions:   coptic}', "");
Expect(1, 11520, '\P{Script_Extensions:   coptic}', "");
Expect(0, 11520, '\P{^Script_Extensions:   coptic}', "");
Expect(1, 11519, '\p{Script_Extensions=	Coptic}', "");
Expect(0, 11519, '\p{^Script_Extensions=	Coptic}', "");
Expect(0, 11519, '\P{Script_Extensions=	Coptic}', "");
Expect(1, 11519, '\P{^Script_Extensions=	Coptic}', "");
Expect(0, 11520, '\p{Script_Extensions=	Coptic}', "");
Expect(1, 11520, '\p{^Script_Extensions=	Coptic}', "");
Expect(1, 11520, '\P{Script_Extensions=	Coptic}', "");
Expect(0, 11520, '\P{^Script_Extensions=	Coptic}', "");
Error('\p{Scx=	/a/copt}');
Error('\P{Scx=	/a/copt}');
Expect(1, 11519, '\p{Scx=copt}', "");
Expect(0, 11519, '\p{^Scx=copt}', "");
Expect(0, 11519, '\P{Scx=copt}', "");
Expect(1, 11519, '\P{^Scx=copt}', "");
Expect(0, 11520, '\p{Scx=copt}', "");
Expect(1, 11520, '\p{^Scx=copt}', "");
Expect(1, 11520, '\P{Scx=copt}', "");
Expect(0, 11520, '\P{^Scx=copt}', "");
Expect(1, 11519, '\p{Scx=_-COPT}', "");
Expect(0, 11519, '\p{^Scx=_-COPT}', "");
Expect(0, 11519, '\P{Scx=_-COPT}', "");
Expect(1, 11519, '\P{^Scx=_-COPT}', "");
Expect(0, 11520, '\p{Scx=_-COPT}', "");
Expect(1, 11520, '\p{^Scx=_-COPT}', "");
Expect(1, 11520, '\P{Scx=_-COPT}', "");
Expect(0, 11520, '\P{^Scx=_-COPT}', "");
Error('\p{Is_Script_Extensions= 	qaac/a/}');
Error('\P{Is_Script_Extensions= 	qaac/a/}');
Expect(1, 11519, '\p{Is_Script_Extensions=qaac}', "");
Expect(0, 11519, '\p{^Is_Script_Extensions=qaac}', "");
Expect(0, 11519, '\P{Is_Script_Extensions=qaac}', "");
Expect(1, 11519, '\P{^Is_Script_Extensions=qaac}', "");
Expect(0, 11520, '\p{Is_Script_Extensions=qaac}', "");
Expect(1, 11520, '\p{^Is_Script_Extensions=qaac}', "");
Expect(1, 11520, '\P{Is_Script_Extensions=qaac}', "");
Expect(0, 11520, '\P{^Is_Script_Extensions=qaac}', "");
Expect(1, 11519, '\p{Is_Script_Extensions=	-Qaac}', "");
Expect(0, 11519, '\p{^Is_Script_Extensions=	-Qaac}', "");
Expect(0, 11519, '\P{Is_Script_Extensions=	-Qaac}', "");
Expect(1, 11519, '\P{^Is_Script_Extensions=	-Qaac}', "");
Expect(0, 11520, '\p{Is_Script_Extensions=	-Qaac}', "");
Expect(1, 11520, '\p{^Is_Script_Extensions=	-Qaac}', "");
Expect(1, 11520, '\P{Is_Script_Extensions=	-Qaac}', "");
Expect(0, 11520, '\P{^Is_Script_Extensions=	-Qaac}', "");
Error('\p{Is_Scx=/a/	 Coptic}');
Error('\P{Is_Scx=/a/	 Coptic}');
Expect(1, 11519, '\p{Is_Scx=coptic}', "");
Expect(0, 11519, '\p{^Is_Scx=coptic}', "");
Expect(0, 11519, '\P{Is_Scx=coptic}', "");
Expect(1, 11519, '\P{^Is_Scx=coptic}', "");
Expect(0, 11520, '\p{Is_Scx=coptic}', "");
Expect(1, 11520, '\p{^Is_Scx=coptic}', "");
Expect(1, 11520, '\P{Is_Scx=coptic}', "");
Expect(0, 11520, '\P{^Is_Scx=coptic}', "");
Expect(1, 11519, '\p{Is_Scx= _COPTIC}', "");
Expect(0, 11519, '\p{^Is_Scx= _COPTIC}', "");
Expect(0, 11519, '\P{Is_Scx= _COPTIC}', "");
Expect(1, 11519, '\P{^Is_Scx= _COPTIC}', "");
Expect(0, 11520, '\p{Is_Scx= _COPTIC}', "");
Expect(1, 11520, '\p{^Is_Scx= _COPTIC}', "");
Expect(1, 11520, '\P{Is_Scx= _COPTIC}', "");
Expect(0, 11520, '\P{^Is_Scx= _COPTIC}', "");
Error('\p{Script_Extensions=:= -cypriot}');
Error('\P{Script_Extensions=:= -cypriot}');
Expect(1, 67647, '\p{Script_Extensions=cypriot}', "");
Expect(0, 67647, '\p{^Script_Extensions=cypriot}', "");
Expect(0, 67647, '\P{Script_Extensions=cypriot}', "");
Expect(1, 67647, '\P{^Script_Extensions=cypriot}', "");
Expect(0, 67648, '\p{Script_Extensions=cypriot}', "");
Expect(1, 67648, '\p{^Script_Extensions=cypriot}', "");
Expect(1, 67648, '\P{Script_Extensions=cypriot}', "");
Expect(0, 67648, '\P{^Script_Extensions=cypriot}', "");
Expect(1, 67647, '\p{Script_Extensions=  cypriot}', "");
Expect(0, 67647, '\p{^Script_Extensions=  cypriot}', "");
Expect(0, 67647, '\P{Script_Extensions=  cypriot}', "");
Expect(1, 67647, '\P{^Script_Extensions=  cypriot}', "");
Expect(0, 67648, '\p{Script_Extensions=  cypriot}', "");
Expect(1, 67648, '\p{^Script_Extensions=  cypriot}', "");
Expect(1, 67648, '\P{Script_Extensions=  cypriot}', "");
Expect(0, 67648, '\P{^Script_Extensions=  cypriot}', "");
Error('\p{Scx=:=-_Cprt}');
Error('\P{Scx=:=-_Cprt}');
Expect(1, 67647, '\p{Scx=cprt}', "");
Expect(0, 67647, '\p{^Scx=cprt}', "");
Expect(0, 67647, '\P{Scx=cprt}', "");
Expect(1, 67647, '\P{^Scx=cprt}', "");
Expect(0, 67648, '\p{Scx=cprt}', "");
Expect(1, 67648, '\p{^Scx=cprt}', "");
Expect(1, 67648, '\P{Scx=cprt}', "");
Expect(0, 67648, '\P{^Scx=cprt}', "");
Expect(1, 67647, '\p{Scx=-_cprt}', "");
Expect(0, 67647, '\p{^Scx=-_cprt}', "");
Expect(0, 67647, '\P{Scx=-_cprt}', "");
Expect(1, 67647, '\P{^Scx=-_cprt}', "");
Expect(0, 67648, '\p{Scx=-_cprt}', "");
Expect(1, 67648, '\p{^Scx=-_cprt}', "");
Expect(1, 67648, '\P{Scx=-_cprt}', "");
Expect(0, 67648, '\P{^Scx=-_cprt}', "");
Error('\p{Is_Script_Extensions=/a/_ Cypriot}');
Error('\P{Is_Script_Extensions=/a/_ Cypriot}');
Expect(1, 67647, '\p{Is_Script_Extensions=cypriot}', "");
Expect(0, 67647, '\p{^Is_Script_Extensions=cypriot}', "");
Expect(0, 67647, '\P{Is_Script_Extensions=cypriot}', "");
Expect(1, 67647, '\P{^Is_Script_Extensions=cypriot}', "");
Expect(0, 67648, '\p{Is_Script_Extensions=cypriot}', "");
Expect(1, 67648, '\p{^Is_Script_Extensions=cypriot}', "");
Expect(1, 67648, '\P{Is_Script_Extensions=cypriot}', "");
Expect(0, 67648, '\P{^Is_Script_Extensions=cypriot}', "");
Expect(1, 67647, '\p{Is_Script_Extensions= 	CYPRIOT}', "");
Expect(0, 67647, '\p{^Is_Script_Extensions= 	CYPRIOT}', "");
Expect(0, 67647, '\P{Is_Script_Extensions= 	CYPRIOT}', "");
Expect(1, 67647, '\P{^Is_Script_Extensions= 	CYPRIOT}', "");
Expect(0, 67648, '\p{Is_Script_Extensions= 	CYPRIOT}', "");
Expect(1, 67648, '\p{^Is_Script_Extensions= 	CYPRIOT}', "");
Expect(1, 67648, '\P{Is_Script_Extensions= 	CYPRIOT}', "");
Expect(0, 67648, '\P{^Is_Script_Extensions= 	CYPRIOT}', "");
Error('\p{Is_Scx= _Cprt:=}');
Error('\P{Is_Scx= _Cprt:=}');
Expect(1, 67647, '\p{Is_Scx=cprt}', "");
Expect(0, 67647, '\p{^Is_Scx=cprt}', "");
Expect(0, 67647, '\P{Is_Scx=cprt}', "");
Expect(1, 67647, '\P{^Is_Scx=cprt}', "");
Expect(0, 67648, '\p{Is_Scx=cprt}', "");
Expect(1, 67648, '\p{^Is_Scx=cprt}', "");
Expect(1, 67648, '\P{Is_Scx=cprt}', "");
Expect(0, 67648, '\P{^Is_Scx=cprt}', "");
Expect(1, 67647, '\p{Is_Scx= _Cprt}', "");
Expect(0, 67647, '\p{^Is_Scx= _Cprt}', "");
Expect(0, 67647, '\P{Is_Scx= _Cprt}', "");
Expect(1, 67647, '\P{^Is_Scx= _Cprt}', "");
Expect(0, 67648, '\p{Is_Scx= _Cprt}', "");
Expect(1, 67648, '\p{^Is_Scx= _Cprt}', "");
Expect(1, 67648, '\P{Is_Scx= _Cprt}', "");
Expect(0, 67648, '\P{^Is_Scx= _Cprt}', "");
Error('\p{Script_Extensions=-_CYRILLIC/a/}');
Error('\P{Script_Extensions=-_CYRILLIC/a/}');
Expect(1, 42655, '\p{Script_Extensions=cyrillic}', "");
Expect(0, 42655, '\p{^Script_Extensions=cyrillic}', "");
Expect(0, 42655, '\P{Script_Extensions=cyrillic}', "");
Expect(1, 42655, '\P{^Script_Extensions=cyrillic}', "");
Expect(0, 42656, '\p{Script_Extensions=cyrillic}', "");
Expect(1, 42656, '\p{^Script_Extensions=cyrillic}', "");
Expect(1, 42656, '\P{Script_Extensions=cyrillic}', "");
Expect(0, 42656, '\P{^Script_Extensions=cyrillic}', "");
Expect(1, 42655, '\p{Script_Extensions=__CYRILLIC}', "");
Expect(0, 42655, '\p{^Script_Extensions=__CYRILLIC}', "");
Expect(0, 42655, '\P{Script_Extensions=__CYRILLIC}', "");
Expect(1, 42655, '\P{^Script_Extensions=__CYRILLIC}', "");
Expect(0, 42656, '\p{Script_Extensions=__CYRILLIC}', "");
Expect(1, 42656, '\p{^Script_Extensions=__CYRILLIC}', "");
Expect(1, 42656, '\P{Script_Extensions=__CYRILLIC}', "");
Expect(0, 42656, '\P{^Script_Extensions=__CYRILLIC}', "");
Error('\p{Scx=_/a/Cyrl}');
Error('\P{Scx=_/a/Cyrl}');
Expect(1, 42655, '\p{Scx=cyrl}', "");
Expect(0, 42655, '\p{^Scx=cyrl}', "");
Expect(0, 42655, '\P{Scx=cyrl}', "");
Expect(1, 42655, '\P{^Scx=cyrl}', "");
Expect(0, 42656, '\p{Scx=cyrl}', "");
Expect(1, 42656, '\p{^Scx=cyrl}', "");
Expect(1, 42656, '\P{Scx=cyrl}', "");
Expect(0, 42656, '\P{^Scx=cyrl}', "");
Expect(1, 42655, '\p{Scx=	Cyrl}', "");
Expect(0, 42655, '\p{^Scx=	Cyrl}', "");
Expect(0, 42655, '\P{Scx=	Cyrl}', "");
Expect(1, 42655, '\P{^Scx=	Cyrl}', "");
Expect(0, 42656, '\p{Scx=	Cyrl}', "");
Expect(1, 42656, '\p{^Scx=	Cyrl}', "");
Expect(1, 42656, '\P{Scx=	Cyrl}', "");
Expect(0, 42656, '\P{^Scx=	Cyrl}', "");
Error('\p{Is_Script_Extensions=	/a/Cyrillic}');
Error('\P{Is_Script_Extensions=	/a/Cyrillic}');
Expect(1, 42655, '\p{Is_Script_Extensions=cyrillic}', "");
Expect(0, 42655, '\p{^Is_Script_Extensions=cyrillic}', "");
Expect(0, 42655, '\P{Is_Script_Extensions=cyrillic}', "");
Expect(1, 42655, '\P{^Is_Script_Extensions=cyrillic}', "");
Expect(0, 42656, '\p{Is_Script_Extensions=cyrillic}', "");
Expect(1, 42656, '\p{^Is_Script_Extensions=cyrillic}', "");
Expect(1, 42656, '\P{Is_Script_Extensions=cyrillic}', "");
Expect(0, 42656, '\P{^Is_Script_Extensions=cyrillic}', "");
Expect(1, 42655, '\p{Is_Script_Extensions= Cyrillic}', "");
Expect(0, 42655, '\p{^Is_Script_Extensions= Cyrillic}', "");
Expect(0, 42655, '\P{Is_Script_Extensions= Cyrillic}', "");
Expect(1, 42655, '\P{^Is_Script_Extensions= Cyrillic}', "");
Expect(0, 42656, '\p{Is_Script_Extensions= Cyrillic}', "");
Expect(1, 42656, '\p{^Is_Script_Extensions= Cyrillic}', "");
Expect(1, 42656, '\P{Is_Script_Extensions= Cyrillic}', "");
Expect(0, 42656, '\P{^Is_Script_Extensions= Cyrillic}', "");
Error('\p{Is_Scx=	Cyrl/a/}');
Error('\P{Is_Scx=	Cyrl/a/}');
Expect(1, 42655, '\p{Is_Scx=cyrl}', "");
Expect(0, 42655, '\p{^Is_Scx=cyrl}', "");
Expect(0, 42655, '\P{Is_Scx=cyrl}', "");
Expect(1, 42655, '\P{^Is_Scx=cyrl}', "");
Expect(0, 42656, '\p{Is_Scx=cyrl}', "");
Expect(1, 42656, '\p{^Is_Scx=cyrl}', "");
Expect(1, 42656, '\P{Is_Scx=cyrl}', "");
Expect(0, 42656, '\P{^Is_Scx=cyrl}', "");
Expect(1, 42655, '\p{Is_Scx=--Cyrl}', "");
Expect(0, 42655, '\p{^Is_Scx=--Cyrl}', "");
Expect(0, 42655, '\P{Is_Scx=--Cyrl}', "");
Expect(1, 42655, '\P{^Is_Scx=--Cyrl}', "");
Expect(0, 42656, '\p{Is_Scx=--Cyrl}', "");
Expect(1, 42656, '\p{^Is_Scx=--Cyrl}', "");
Expect(1, 42656, '\P{Is_Scx=--Cyrl}', "");
Expect(0, 42656, '\P{^Is_Scx=--Cyrl}', "");
Error('\p{Script_Extensions=	:=DEVANAGARI}');
Error('\P{Script_Extensions=	:=DEVANAGARI}');
Expect(1, 43259, '\p{Script_Extensions=devanagari}', "");
Expect(0, 43259, '\p{^Script_Extensions=devanagari}', "");
Expect(0, 43259, '\P{Script_Extensions=devanagari}', "");
Expect(1, 43259, '\P{^Script_Extensions=devanagari}', "");
Expect(0, 43260, '\p{Script_Extensions=devanagari}', "");
Expect(1, 43260, '\p{^Script_Extensions=devanagari}', "");
Expect(1, 43260, '\P{Script_Extensions=devanagari}', "");
Expect(0, 43260, '\P{^Script_Extensions=devanagari}', "");
Expect(1, 43259, '\p{Script_Extensions= -devanagari}', "");
Expect(0, 43259, '\p{^Script_Extensions= -devanagari}', "");
Expect(0, 43259, '\P{Script_Extensions= -devanagari}', "");
Expect(1, 43259, '\P{^Script_Extensions= -devanagari}', "");
Expect(0, 43260, '\p{Script_Extensions= -devanagari}', "");
Expect(1, 43260, '\p{^Script_Extensions= -devanagari}', "");
Expect(1, 43260, '\P{Script_Extensions= -devanagari}', "");
Expect(0, 43260, '\P{^Script_Extensions= -devanagari}', "");
Error('\p{Scx=/a/Deva}');
Error('\P{Scx=/a/Deva}');
Expect(1, 43259, '\p{Scx=deva}', "");
Expect(0, 43259, '\p{^Scx=deva}', "");
Expect(0, 43259, '\P{Scx=deva}', "");
Expect(1, 43259, '\P{^Scx=deva}', "");
Expect(0, 43260, '\p{Scx=deva}', "");
Expect(1, 43260, '\p{^Scx=deva}', "");
Expect(1, 43260, '\P{Scx=deva}', "");
Expect(0, 43260, '\P{^Scx=deva}', "");
Expect(1, 43259, '\p{Scx=	-deva}', "");
Expect(0, 43259, '\p{^Scx=	-deva}', "");
Expect(0, 43259, '\P{Scx=	-deva}', "");
Expect(1, 43259, '\P{^Scx=	-deva}', "");
Expect(0, 43260, '\p{Scx=	-deva}', "");
Expect(1, 43260, '\p{^Scx=	-deva}', "");
Expect(1, 43260, '\P{Scx=	-deva}', "");
Expect(0, 43260, '\P{^Scx=	-deva}', "");
Error('\p{Is_Script_Extensions:	:=Devanagari}');
Error('\P{Is_Script_Extensions:	:=Devanagari}');
Expect(1, 43259, '\p{Is_Script_Extensions=devanagari}', "");
Expect(0, 43259, '\p{^Is_Script_Extensions=devanagari}', "");
Expect(0, 43259, '\P{Is_Script_Extensions=devanagari}', "");
Expect(1, 43259, '\P{^Is_Script_Extensions=devanagari}', "");
Expect(0, 43260, '\p{Is_Script_Extensions=devanagari}', "");
Expect(1, 43260, '\p{^Is_Script_Extensions=devanagari}', "");
Expect(1, 43260, '\P{Is_Script_Extensions=devanagari}', "");
Expect(0, 43260, '\P{^Is_Script_Extensions=devanagari}', "");
Expect(1, 43259, '\p{Is_Script_Extensions=_ Devanagari}', "");
Expect(0, 43259, '\p{^Is_Script_Extensions=_ Devanagari}', "");
Expect(0, 43259, '\P{Is_Script_Extensions=_ Devanagari}', "");
Expect(1, 43259, '\P{^Is_Script_Extensions=_ Devanagari}', "");
Expect(0, 43260, '\p{Is_Script_Extensions=_ Devanagari}', "");
Expect(1, 43260, '\p{^Is_Script_Extensions=_ Devanagari}', "");
Expect(1, 43260, '\P{Is_Script_Extensions=_ Devanagari}', "");
Expect(0, 43260, '\P{^Is_Script_Extensions=_ Devanagari}', "");
Error('\p{Is_Scx=:=  DEVA}');
Error('\P{Is_Scx=:=  DEVA}');
Expect(1, 43259, '\p{Is_Scx=deva}', "");
Expect(0, 43259, '\p{^Is_Scx=deva}', "");
Expect(0, 43259, '\P{Is_Scx=deva}', "");
Expect(1, 43259, '\P{^Is_Scx=deva}', "");
Expect(0, 43260, '\p{Is_Scx=deva}', "");
Expect(1, 43260, '\p{^Is_Scx=deva}', "");
Expect(1, 43260, '\P{Is_Scx=deva}', "");
Expect(0, 43260, '\P{^Is_Scx=deva}', "");
Expect(1, 43259, '\p{Is_Scx=	Deva}', "");
Expect(0, 43259, '\p{^Is_Scx=	Deva}', "");
Expect(0, 43259, '\P{Is_Scx=	Deva}', "");
Expect(1, 43259, '\P{^Is_Scx=	Deva}', "");
Expect(0, 43260, '\p{Is_Scx=	Deva}', "");
Expect(1, 43260, '\p{^Is_Scx=	Deva}', "");
Expect(1, 43260, '\P{Is_Scx=	Deva}', "");
Expect(0, 43260, '\P{^Is_Scx=	Deva}', "");
Error('\p{Script_Extensions=	-Deseret:=}');
Error('\P{Script_Extensions=	-Deseret:=}');
Expect(1, 66639, '\p{Script_Extensions=deseret}', "");
Expect(0, 66639, '\p{^Script_Extensions=deseret}', "");
Expect(0, 66639, '\P{Script_Extensions=deseret}', "");
Expect(1, 66639, '\P{^Script_Extensions=deseret}', "");
Expect(0, 66640, '\p{Script_Extensions=deseret}', "");
Expect(1, 66640, '\p{^Script_Extensions=deseret}', "");
Expect(1, 66640, '\P{Script_Extensions=deseret}', "");
Expect(0, 66640, '\P{^Script_Extensions=deseret}', "");
Expect(1, 66639, '\p{Script_Extensions: Deseret}', "");
Expect(0, 66639, '\p{^Script_Extensions: Deseret}', "");
Expect(0, 66639, '\P{Script_Extensions: Deseret}', "");
Expect(1, 66639, '\P{^Script_Extensions: Deseret}', "");
Expect(0, 66640, '\p{Script_Extensions: Deseret}', "");
Expect(1, 66640, '\p{^Script_Extensions: Deseret}', "");
Expect(1, 66640, '\P{Script_Extensions: Deseret}', "");
Expect(0, 66640, '\P{^Script_Extensions: Deseret}', "");
Error('\p{Scx=	_DSRT:=}');
Error('\P{Scx=	_DSRT:=}');
Expect(1, 66639, '\p{Scx:dsrt}', "");
Expect(0, 66639, '\p{^Scx:dsrt}', "");
Expect(0, 66639, '\P{Scx:dsrt}', "");
Expect(1, 66639, '\P{^Scx:dsrt}', "");
Expect(0, 66640, '\p{Scx:dsrt}', "");
Expect(1, 66640, '\p{^Scx:dsrt}', "");
Expect(1, 66640, '\P{Scx:dsrt}', "");
Expect(0, 66640, '\P{^Scx:dsrt}', "");
Expect(1, 66639, '\p{Scx=		Dsrt}', "");
Expect(0, 66639, '\p{^Scx=		Dsrt}', "");
Expect(0, 66639, '\P{Scx=		Dsrt}', "");
Expect(1, 66639, '\P{^Scx=		Dsrt}', "");
Expect(0, 66640, '\p{Scx=		Dsrt}', "");
Expect(1, 66640, '\p{^Scx=		Dsrt}', "");
Expect(1, 66640, '\P{Scx=		Dsrt}', "");
Expect(0, 66640, '\P{^Scx=		Dsrt}', "");
Error('\p{Is_Script_Extensions=/a/	-Deseret}');
Error('\P{Is_Script_Extensions=/a/	-Deseret}');
Expect(1, 66639, '\p{Is_Script_Extensions=deseret}', "");
Expect(0, 66639, '\p{^Is_Script_Extensions=deseret}', "");
Expect(0, 66639, '\P{Is_Script_Extensions=deseret}', "");
Expect(1, 66639, '\P{^Is_Script_Extensions=deseret}', "");
Expect(0, 66640, '\p{Is_Script_Extensions=deseret}', "");
Expect(1, 66640, '\p{^Is_Script_Extensions=deseret}', "");
Expect(1, 66640, '\P{Is_Script_Extensions=deseret}', "");
Expect(0, 66640, '\P{^Is_Script_Extensions=deseret}', "");
Expect(1, 66639, '\p{Is_Script_Extensions= _DESERET}', "");
Expect(0, 66639, '\p{^Is_Script_Extensions= _DESERET}', "");
Expect(0, 66639, '\P{Is_Script_Extensions= _DESERET}', "");
Expect(1, 66639, '\P{^Is_Script_Extensions= _DESERET}', "");
Expect(0, 66640, '\p{Is_Script_Extensions= _DESERET}', "");
Expect(1, 66640, '\p{^Is_Script_Extensions= _DESERET}', "");
Expect(1, 66640, '\P{Is_Script_Extensions= _DESERET}', "");
Expect(0, 66640, '\P{^Is_Script_Extensions= _DESERET}', "");
Error('\p{Is_Scx:   _DSRT/a/}');
Error('\P{Is_Scx:   _DSRT/a/}');
Expect(1, 66639, '\p{Is_Scx=dsrt}', "");
Expect(0, 66639, '\p{^Is_Scx=dsrt}', "");
Expect(0, 66639, '\P{Is_Scx=dsrt}', "");
Expect(1, 66639, '\P{^Is_Scx=dsrt}', "");
Expect(0, 66640, '\p{Is_Scx=dsrt}', "");
Expect(1, 66640, '\p{^Is_Scx=dsrt}', "");
Expect(1, 66640, '\P{Is_Scx=dsrt}', "");
Expect(0, 66640, '\P{^Is_Scx=dsrt}', "");
Expect(1, 66639, '\p{Is_Scx= dsrt}', "");
Expect(0, 66639, '\p{^Is_Scx= dsrt}', "");
Expect(0, 66639, '\P{Is_Scx= dsrt}', "");
Expect(1, 66639, '\P{^Is_Scx= dsrt}', "");
Expect(0, 66640, '\p{Is_Scx= dsrt}', "");
Expect(1, 66640, '\p{^Is_Scx= dsrt}', "");
Expect(1, 66640, '\P{Is_Scx= dsrt}', "");
Expect(0, 66640, '\P{^Is_Scx= dsrt}', "");
Error('\p{Script_Extensions:	  Egyptian_Hieroglyphs/a/}');
Error('\P{Script_Extensions:	  Egyptian_Hieroglyphs/a/}');
Expect(1, 78894, '\p{Script_Extensions=egyptianhieroglyphs}', "");
Expect(0, 78894, '\p{^Script_Extensions=egyptianhieroglyphs}', "");
Expect(0, 78894, '\P{Script_Extensions=egyptianhieroglyphs}', "");
Expect(1, 78894, '\P{^Script_Extensions=egyptianhieroglyphs}', "");
Expect(0, 78895, '\p{Script_Extensions=egyptianhieroglyphs}', "");
Expect(1, 78895, '\p{^Script_Extensions=egyptianhieroglyphs}', "");
Expect(1, 78895, '\P{Script_Extensions=egyptianhieroglyphs}', "");
Expect(0, 78895, '\P{^Script_Extensions=egyptianhieroglyphs}', "");
Expect(1, 78894, '\p{Script_Extensions=	_egyptian_Hieroglyphs}', "");
Expect(0, 78894, '\p{^Script_Extensions=	_egyptian_Hieroglyphs}', "");
Expect(0, 78894, '\P{Script_Extensions=	_egyptian_Hieroglyphs}', "");
Expect(1, 78894, '\P{^Script_Extensions=	_egyptian_Hieroglyphs}', "");
Expect(0, 78895, '\p{Script_Extensions=	_egyptian_Hieroglyphs}', "");
Expect(1, 78895, '\p{^Script_Extensions=	_egyptian_Hieroglyphs}', "");
Expect(1, 78895, '\P{Script_Extensions=	_egyptian_Hieroglyphs}', "");
Expect(0, 78895, '\P{^Script_Extensions=	_egyptian_Hieroglyphs}', "");
Error('\p{Scx=:= Egyp}');
Error('\P{Scx=:= Egyp}');
Expect(1, 78894, '\p{Scx=egyp}', "");
Expect(0, 78894, '\p{^Scx=egyp}', "");
Expect(0, 78894, '\P{Scx=egyp}', "");
Expect(1, 78894, '\P{^Scx=egyp}', "");
Expect(0, 78895, '\p{Scx=egyp}', "");
Expect(1, 78895, '\p{^Scx=egyp}', "");
Expect(1, 78895, '\P{Scx=egyp}', "");
Expect(0, 78895, '\P{^Scx=egyp}', "");
Expect(1, 78894, '\p{Scx=--egyp}', "");
Expect(0, 78894, '\p{^Scx=--egyp}', "");
Expect(0, 78894, '\P{Scx=--egyp}', "");
Expect(1, 78894, '\P{^Scx=--egyp}', "");
Expect(0, 78895, '\p{Scx=--egyp}', "");
Expect(1, 78895, '\p{^Scx=--egyp}', "");
Expect(1, 78895, '\P{Scx=--egyp}', "");
Expect(0, 78895, '\P{^Scx=--egyp}', "");
Error('\p{Is_Script_Extensions= egyptian_hieroglyphs:=}');
Error('\P{Is_Script_Extensions= egyptian_hieroglyphs:=}');
Expect(1, 78894, '\p{Is_Script_Extensions=egyptianhieroglyphs}', "");
Expect(0, 78894, '\p{^Is_Script_Extensions=egyptianhieroglyphs}', "");
Expect(0, 78894, '\P{Is_Script_Extensions=egyptianhieroglyphs}', "");
Expect(1, 78894, '\P{^Is_Script_Extensions=egyptianhieroglyphs}', "");
Expect(0, 78895, '\p{Is_Script_Extensions=egyptianhieroglyphs}', "");
Expect(1, 78895, '\p{^Is_Script_Extensions=egyptianhieroglyphs}', "");
Expect(1, 78895, '\P{Is_Script_Extensions=egyptianhieroglyphs}', "");
Expect(0, 78895, '\P{^Is_Script_Extensions=egyptianhieroglyphs}', "");
Expect(1, 78894, '\p{Is_Script_Extensions=_-egyptian_Hieroglyphs}', "");
Expect(0, 78894, '\p{^Is_Script_Extensions=_-egyptian_Hieroglyphs}', "");
Expect(0, 78894, '\P{Is_Script_Extensions=_-egyptian_Hieroglyphs}', "");
Expect(1, 78894, '\P{^Is_Script_Extensions=_-egyptian_Hieroglyphs}', "");
Expect(0, 78895, '\p{Is_Script_Extensions=_-egyptian_Hieroglyphs}', "");
Expect(1, 78895, '\p{^Is_Script_Extensions=_-egyptian_Hieroglyphs}', "");
Expect(1, 78895, '\P{Is_Script_Extensions=_-egyptian_Hieroglyphs}', "");
Expect(0, 78895, '\P{^Is_Script_Extensions=_-egyptian_Hieroglyphs}', "");
Error('\p{Is_Scx=  egyp:=}');
Error('\P{Is_Scx=  egyp:=}');
Expect(1, 78894, '\p{Is_Scx=egyp}', "");
Expect(0, 78894, '\p{^Is_Scx=egyp}', "");
Expect(0, 78894, '\P{Is_Scx=egyp}', "");
Expect(1, 78894, '\P{^Is_Scx=egyp}', "");
Expect(0, 78895, '\p{Is_Scx=egyp}', "");
Expect(1, 78895, '\p{^Is_Scx=egyp}', "");
Expect(1, 78895, '\P{Is_Scx=egyp}', "");
Expect(0, 78895, '\P{^Is_Scx=egyp}', "");
Expect(1, 78894, '\p{Is_Scx=_ Egyp}', "");
Expect(0, 78894, '\p{^Is_Scx=_ Egyp}', "");
Expect(0, 78894, '\P{Is_Scx=_ Egyp}', "");
Expect(1, 78894, '\P{^Is_Scx=_ Egyp}', "");
Expect(0, 78895, '\p{Is_Scx=_ Egyp}', "");
Expect(1, 78895, '\p{^Is_Scx=_ Egyp}', "");
Expect(1, 78895, '\P{Is_Scx=_ Egyp}', "");
Expect(0, 78895, '\P{^Is_Scx=_ Egyp}', "");
Error('\p{Script_Extensions=	:=Ethiopic}');
Error('\P{Script_Extensions=	:=Ethiopic}');
Expect(1, 43822, '\p{Script_Extensions=ethiopic}', "");
Expect(0, 43822, '\p{^Script_Extensions=ethiopic}', "");
Expect(0, 43822, '\P{Script_Extensions=ethiopic}', "");
Expect(1, 43822, '\P{^Script_Extensions=ethiopic}', "");
Expect(0, 43823, '\p{Script_Extensions=ethiopic}', "");
Expect(1, 43823, '\p{^Script_Extensions=ethiopic}', "");
Expect(1, 43823, '\P{Script_Extensions=ethiopic}', "");
Expect(0, 43823, '\P{^Script_Extensions=ethiopic}', "");
Expect(1, 43822, '\p{Script_Extensions=-Ethiopic}', "");
Expect(0, 43822, '\p{^Script_Extensions=-Ethiopic}', "");
Expect(0, 43822, '\P{Script_Extensions=-Ethiopic}', "");
Expect(1, 43822, '\P{^Script_Extensions=-Ethiopic}', "");
Expect(0, 43823, '\p{Script_Extensions=-Ethiopic}', "");
Expect(1, 43823, '\p{^Script_Extensions=-Ethiopic}', "");
Expect(1, 43823, '\P{Script_Extensions=-Ethiopic}', "");
Expect(0, 43823, '\P{^Script_Extensions=-Ethiopic}', "");
Error('\p{Scx=_:=ETHI}');
Error('\P{Scx=_:=ETHI}');
Expect(1, 43822, '\p{Scx:   ethi}', "");
Expect(0, 43822, '\p{^Scx:   ethi}', "");
Expect(0, 43822, '\P{Scx:   ethi}', "");
Expect(1, 43822, '\P{^Scx:   ethi}', "");
Expect(0, 43823, '\p{Scx:   ethi}', "");
Expect(1, 43823, '\p{^Scx:   ethi}', "");
Expect(1, 43823, '\P{Scx:   ethi}', "");
Expect(0, 43823, '\P{^Scx:   ethi}', "");
Expect(1, 43822, '\p{Scx=_-ethi}', "");
Expect(0, 43822, '\p{^Scx=_-ethi}', "");
Expect(0, 43822, '\P{Scx=_-ethi}', "");
Expect(1, 43822, '\P{^Scx=_-ethi}', "");
Expect(0, 43823, '\p{Scx=_-ethi}', "");
Expect(1, 43823, '\p{^Scx=_-ethi}', "");
Expect(1, 43823, '\P{Scx=_-ethi}', "");
Expect(0, 43823, '\P{^Scx=_-ethi}', "");
Error('\p{Is_Script_Extensions:	-Ethiopic:=}');
Error('\P{Is_Script_Extensions:	-Ethiopic:=}');
Expect(1, 43822, '\p{Is_Script_Extensions=ethiopic}', "");
Expect(0, 43822, '\p{^Is_Script_Extensions=ethiopic}', "");
Expect(0, 43822, '\P{Is_Script_Extensions=ethiopic}', "");
Expect(1, 43822, '\P{^Is_Script_Extensions=ethiopic}', "");
Expect(0, 43823, '\p{Is_Script_Extensions=ethiopic}', "");
Expect(1, 43823, '\p{^Is_Script_Extensions=ethiopic}', "");
Expect(1, 43823, '\P{Is_Script_Extensions=ethiopic}', "");
Expect(0, 43823, '\P{^Is_Script_Extensions=ethiopic}', "");
Expect(1, 43822, '\p{Is_Script_Extensions=	-Ethiopic}', "");
Expect(0, 43822, '\p{^Is_Script_Extensions=	-Ethiopic}', "");
Expect(0, 43822, '\P{Is_Script_Extensions=	-Ethiopic}', "");
Expect(1, 43822, '\P{^Is_Script_Extensions=	-Ethiopic}', "");
Expect(0, 43823, '\p{Is_Script_Extensions=	-Ethiopic}', "");
Expect(1, 43823, '\p{^Is_Script_Extensions=	-Ethiopic}', "");
Expect(1, 43823, '\P{Is_Script_Extensions=	-Ethiopic}', "");
Expect(0, 43823, '\P{^Is_Script_Extensions=	-Ethiopic}', "");
Error('\p{Is_Scx: _:=ethi}');
Error('\P{Is_Scx: _:=ethi}');
Expect(1, 43822, '\p{Is_Scx:	ethi}', "");
Expect(0, 43822, '\p{^Is_Scx:	ethi}', "");
Expect(0, 43822, '\P{Is_Scx:	ethi}', "");
Expect(1, 43822, '\P{^Is_Scx:	ethi}', "");
Expect(0, 43823, '\p{Is_Scx:	ethi}', "");
Expect(1, 43823, '\p{^Is_Scx:	ethi}', "");
Expect(1, 43823, '\P{Is_Scx:	ethi}', "");
Expect(0, 43823, '\P{^Is_Scx:	ethi}', "");
Expect(1, 43822, '\p{Is_Scx=_-Ethi}', "");
Expect(0, 43822, '\p{^Is_Scx=_-Ethi}', "");
Expect(0, 43822, '\P{Is_Scx=_-Ethi}', "");
Expect(1, 43822, '\P{^Is_Scx=_-Ethi}', "");
Expect(0, 43823, '\p{Is_Scx=_-Ethi}', "");
Expect(1, 43823, '\p{^Is_Scx=_-Ethi}', "");
Expect(1, 43823, '\P{Is_Scx=_-Ethi}', "");
Expect(0, 43823, '\P{^Is_Scx=_-Ethi}', "");
Error('\p{Script_Extensions=/a/-	Georgian}');
Error('\P{Script_Extensions=/a/-	Georgian}');
Expect(1, 11565, '\p{Script_Extensions=georgian}', "");
Expect(0, 11565, '\p{^Script_Extensions=georgian}', "");
Expect(0, 11565, '\P{Script_Extensions=georgian}', "");
Expect(1, 11565, '\P{^Script_Extensions=georgian}', "");
Expect(0, 11566, '\p{Script_Extensions=georgian}', "");
Expect(1, 11566, '\p{^Script_Extensions=georgian}', "");
Expect(1, 11566, '\P{Script_Extensions=georgian}', "");
Expect(0, 11566, '\P{^Script_Extensions=georgian}', "");
Expect(1, 11565, '\p{Script_Extensions:	_georgian}', "");
Expect(0, 11565, '\p{^Script_Extensions:	_georgian}', "");
Expect(0, 11565, '\P{Script_Extensions:	_georgian}', "");
Expect(1, 11565, '\P{^Script_Extensions:	_georgian}', "");
Expect(0, 11566, '\p{Script_Extensions:	_georgian}', "");
Expect(1, 11566, '\p{^Script_Extensions:	_georgian}', "");
Expect(1, 11566, '\P{Script_Extensions:	_georgian}', "");
Expect(0, 11566, '\P{^Script_Extensions:	_georgian}', "");
Error('\p{Scx=/a/	Geor}');
Error('\P{Scx=/a/	Geor}');
Expect(1, 11565, '\p{Scx=geor}', "");
Expect(0, 11565, '\p{^Scx=geor}', "");
Expect(0, 11565, '\P{Scx=geor}', "");
Expect(1, 11565, '\P{^Scx=geor}', "");
Expect(0, 11566, '\p{Scx=geor}', "");
Expect(1, 11566, '\p{^Scx=geor}', "");
Expect(1, 11566, '\P{Scx=geor}', "");
Expect(0, 11566, '\P{^Scx=geor}', "");
Expect(1, 11565, '\p{Scx:-GEOR}', "");
Expect(0, 11565, '\p{^Scx:-GEOR}', "");
Expect(0, 11565, '\P{Scx:-GEOR}', "");
Expect(1, 11565, '\P{^Scx:-GEOR}', "");
Expect(0, 11566, '\p{Scx:-GEOR}', "");
Expect(1, 11566, '\p{^Scx:-GEOR}', "");
Expect(1, 11566, '\P{Scx:-GEOR}', "");
Expect(0, 11566, '\P{^Scx:-GEOR}', "");
Error('\p{Is_Script_Extensions=	/a/georgian}');
Error('\P{Is_Script_Extensions=	/a/georgian}');
Expect(1, 11565, '\p{Is_Script_Extensions=georgian}', "");
Expect(0, 11565, '\p{^Is_Script_Extensions=georgian}', "");
Expect(0, 11565, '\P{Is_Script_Extensions=georgian}', "");
Expect(1, 11565, '\P{^Is_Script_Extensions=georgian}', "");
Expect(0, 11566, '\p{Is_Script_Extensions=georgian}', "");
Expect(1, 11566, '\p{^Is_Script_Extensions=georgian}', "");
Expect(1, 11566, '\P{Is_Script_Extensions=georgian}', "");
Expect(0, 11566, '\P{^Is_Script_Extensions=georgian}', "");
Expect(1, 11565, '\p{Is_Script_Extensions= -GEORGIAN}', "");
Expect(0, 11565, '\p{^Is_Script_Extensions= -GEORGIAN}', "");
Expect(0, 11565, '\P{Is_Script_Extensions= -GEORGIAN}', "");
Expect(1, 11565, '\P{^Is_Script_Extensions= -GEORGIAN}', "");
Expect(0, 11566, '\p{Is_Script_Extensions= -GEORGIAN}', "");
Expect(1, 11566, '\p{^Is_Script_Extensions= -GEORGIAN}', "");
Expect(1, 11566, '\P{Is_Script_Extensions= -GEORGIAN}', "");
Expect(0, 11566, '\P{^Is_Script_Extensions= -GEORGIAN}', "");
Error('\p{Is_Scx=-/a/geor}');
Error('\P{Is_Scx=-/a/geor}');
Expect(1, 11565, '\p{Is_Scx=geor}', "");
Expect(0, 11565, '\p{^Is_Scx=geor}', "");
Expect(0, 11565, '\P{Is_Scx=geor}', "");
Expect(1, 11565, '\P{^Is_Scx=geor}', "");
Expect(0, 11566, '\p{Is_Scx=geor}', "");
Expect(1, 11566, '\p{^Is_Scx=geor}', "");
Expect(1, 11566, '\P{Is_Scx=geor}', "");
Expect(0, 11566, '\P{^Is_Scx=geor}', "");
Expect(1, 11565, '\p{Is_Scx=-_geor}', "");
Expect(0, 11565, '\p{^Is_Scx=-_geor}', "");
Expect(0, 11565, '\P{Is_Scx=-_geor}', "");
Expect(1, 11565, '\P{^Is_Scx=-_geor}', "");
Expect(0, 11566, '\p{Is_Scx=-_geor}', "");
Expect(1, 11566, '\p{^Is_Scx=-_geor}', "");
Expect(1, 11566, '\P{Is_Scx=-_geor}', "");
Expect(0, 11566, '\P{^Is_Scx=-_geor}', "");
Error('\p{Script_Extensions=:=	_GLAGOLITIC}');
Error('\P{Script_Extensions=:=	_GLAGOLITIC}');
Expect(1, 11358, '\p{Script_Extensions=glagolitic}', "");
Expect(0, 11358, '\p{^Script_Extensions=glagolitic}', "");
Expect(0, 11358, '\P{Script_Extensions=glagolitic}', "");
Expect(1, 11358, '\P{^Script_Extensions=glagolitic}', "");
Expect(0, 11359, '\p{Script_Extensions=glagolitic}', "");
Expect(1, 11359, '\p{^Script_Extensions=glagolitic}', "");
Expect(1, 11359, '\P{Script_Extensions=glagolitic}', "");
Expect(0, 11359, '\P{^Script_Extensions=glagolitic}', "");
Expect(1, 11358, '\p{Script_Extensions=_Glagolitic}', "");
Expect(0, 11358, '\p{^Script_Extensions=_Glagolitic}', "");
Expect(0, 11358, '\P{Script_Extensions=_Glagolitic}', "");
Expect(1, 11358, '\P{^Script_Extensions=_Glagolitic}', "");
Expect(0, 11359, '\p{Script_Extensions=_Glagolitic}', "");
Expect(1, 11359, '\p{^Script_Extensions=_Glagolitic}', "");
Expect(1, 11359, '\P{Script_Extensions=_Glagolitic}', "");
Expect(0, 11359, '\P{^Script_Extensions=_Glagolitic}', "");
Error('\p{Scx=:=_GLAG}');
Error('\P{Scx=:=_GLAG}');
Expect(1, 11358, '\p{Scx:	glag}', "");
Expect(0, 11358, '\p{^Scx:	glag}', "");
Expect(0, 11358, '\P{Scx:	glag}', "");
Expect(1, 11358, '\P{^Scx:	glag}', "");
Expect(0, 11359, '\p{Scx:	glag}', "");
Expect(1, 11359, '\p{^Scx:	glag}', "");
Expect(1, 11359, '\P{Scx:	glag}', "");
Expect(0, 11359, '\P{^Scx:	glag}', "");
Expect(1, 11358, '\p{Scx=-Glag}', "");
Expect(0, 11358, '\p{^Scx=-Glag}', "");
Expect(0, 11358, '\P{Scx=-Glag}', "");
Expect(1, 11358, '\P{^Scx=-Glag}', "");
Expect(0, 11359, '\p{Scx=-Glag}', "");
Expect(1, 11359, '\p{^Scx=-Glag}', "");
Expect(1, 11359, '\P{Scx=-Glag}', "");
Expect(0, 11359, '\P{^Scx=-Glag}', "");
Error('\p{Is_Script_Extensions=/a/ -Glagolitic}');
Error('\P{Is_Script_Extensions=/a/ -Glagolitic}');
Expect(1, 11358, '\p{Is_Script_Extensions=glagolitic}', "");
Expect(0, 11358, '\p{^Is_Script_Extensions=glagolitic}', "");
Expect(0, 11358, '\P{Is_Script_Extensions=glagolitic}', "");
Expect(1, 11358, '\P{^Is_Script_Extensions=glagolitic}', "");
Expect(0, 11359, '\p{Is_Script_Extensions=glagolitic}', "");
Expect(1, 11359, '\p{^Is_Script_Extensions=glagolitic}', "");
Expect(1, 11359, '\P{Is_Script_Extensions=glagolitic}', "");
Expect(0, 11359, '\P{^Is_Script_Extensions=glagolitic}', "");
Expect(1, 11358, '\p{Is_Script_Extensions=	GLAGOLITIC}', "");
Expect(0, 11358, '\p{^Is_Script_Extensions=	GLAGOLITIC}', "");
Expect(0, 11358, '\P{Is_Script_Extensions=	GLAGOLITIC}', "");
Expect(1, 11358, '\P{^Is_Script_Extensions=	GLAGOLITIC}', "");
Expect(0, 11359, '\p{Is_Script_Extensions=	GLAGOLITIC}', "");
Expect(1, 11359, '\p{^Is_Script_Extensions=	GLAGOLITIC}', "");
Expect(1, 11359, '\P{Is_Script_Extensions=	GLAGOLITIC}', "");
Expect(0, 11359, '\P{^Is_Script_Extensions=	GLAGOLITIC}', "");
Error('\p{Is_Scx=_/a/glag}');
Error('\P{Is_Scx=_/a/glag}');
Expect(1, 11358, '\p{Is_Scx=glag}', "");
Expect(0, 11358, '\p{^Is_Scx=glag}', "");
Expect(0, 11358, '\P{Is_Scx=glag}', "");
Expect(1, 11358, '\P{^Is_Scx=glag}', "");
Expect(0, 11359, '\p{Is_Scx=glag}', "");
Expect(1, 11359, '\p{^Is_Scx=glag}', "");
Expect(1, 11359, '\P{Is_Scx=glag}', "");
Expect(0, 11359, '\P{^Is_Scx=glag}', "");
Expect(1, 11358, '\p{Is_Scx=- glag}', "");
Expect(0, 11358, '\p{^Is_Scx=- glag}', "");
Expect(0, 11358, '\P{Is_Scx=- glag}', "");
Expect(1, 11358, '\P{^Is_Scx=- glag}', "");
Expect(0, 11359, '\p{Is_Scx=- glag}', "");
Expect(1, 11359, '\p{^Is_Scx=- glag}', "");
Expect(1, 11359, '\P{Is_Scx=- glag}', "");
Expect(0, 11359, '\P{^Is_Scx=- glag}', "");
Error('\p{Script_Extensions=/a/-Gothic}');
Error('\P{Script_Extensions=/a/-Gothic}');
Expect(1, 66378, '\p{Script_Extensions=gothic}', "");
Expect(0, 66378, '\p{^Script_Extensions=gothic}', "");
Expect(0, 66378, '\P{Script_Extensions=gothic}', "");
Expect(1, 66378, '\P{^Script_Extensions=gothic}', "");
Expect(0, 66379, '\p{Script_Extensions=gothic}', "");
Expect(1, 66379, '\p{^Script_Extensions=gothic}', "");
Expect(1, 66379, '\P{Script_Extensions=gothic}', "");
Expect(0, 66379, '\P{^Script_Extensions=gothic}', "");
Expect(1, 66378, '\p{Script_Extensions=--GOTHIC}', "");
Expect(0, 66378, '\p{^Script_Extensions=--GOTHIC}', "");
Expect(0, 66378, '\P{Script_Extensions=--GOTHIC}', "");
Expect(1, 66378, '\P{^Script_Extensions=--GOTHIC}', "");
Expect(0, 66379, '\p{Script_Extensions=--GOTHIC}', "");
Expect(1, 66379, '\p{^Script_Extensions=--GOTHIC}', "");
Expect(1, 66379, '\P{Script_Extensions=--GOTHIC}', "");
Expect(0, 66379, '\P{^Script_Extensions=--GOTHIC}', "");
Error('\p{Scx=-:=goth}');
Error('\P{Scx=-:=goth}');
Expect(1, 66378, '\p{Scx=goth}', "");
Expect(0, 66378, '\p{^Scx=goth}', "");
Expect(0, 66378, '\P{Scx=goth}', "");
Expect(1, 66378, '\P{^Scx=goth}', "");
Expect(0, 66379, '\p{Scx=goth}', "");
Expect(1, 66379, '\p{^Scx=goth}', "");
Expect(1, 66379, '\P{Scx=goth}', "");
Expect(0, 66379, '\P{^Scx=goth}', "");
Expect(1, 66378, '\p{Scx=-GOTH}', "");
Expect(0, 66378, '\p{^Scx=-GOTH}', "");
Expect(0, 66378, '\P{Scx=-GOTH}', "");
Expect(1, 66378, '\P{^Scx=-GOTH}', "");
Expect(0, 66379, '\p{Scx=-GOTH}', "");
Expect(1, 66379, '\p{^Scx=-GOTH}', "");
Expect(1, 66379, '\P{Scx=-GOTH}', "");
Expect(0, 66379, '\P{^Scx=-GOTH}', "");
Error('\p{Is_Script_Extensions=:=	-GOTHIC}');
Error('\P{Is_Script_Extensions=:=	-GOTHIC}');
Expect(1, 66378, '\p{Is_Script_Extensions=gothic}', "");
Expect(0, 66378, '\p{^Is_Script_Extensions=gothic}', "");
Expect(0, 66378, '\P{Is_Script_Extensions=gothic}', "");
Expect(1, 66378, '\P{^Is_Script_Extensions=gothic}', "");
Expect(0, 66379, '\p{Is_Script_Extensions=gothic}', "");
Expect(1, 66379, '\p{^Is_Script_Extensions=gothic}', "");
Expect(1, 66379, '\P{Is_Script_Extensions=gothic}', "");
Expect(0, 66379, '\P{^Is_Script_Extensions=gothic}', "");
Expect(1, 66378, '\p{Is_Script_Extensions=-_GOTHIC}', "");
Expect(0, 66378, '\p{^Is_Script_Extensions=-_GOTHIC}', "");
Expect(0, 66378, '\P{Is_Script_Extensions=-_GOTHIC}', "");
Expect(1, 66378, '\P{^Is_Script_Extensions=-_GOTHIC}', "");
Expect(0, 66379, '\p{Is_Script_Extensions=-_GOTHIC}', "");
Expect(1, 66379, '\p{^Is_Script_Extensions=-_GOTHIC}', "");
Expect(1, 66379, '\P{Is_Script_Extensions=-_GOTHIC}', "");
Expect(0, 66379, '\P{^Is_Script_Extensions=-_GOTHIC}', "");
Error('\p{Is_Scx=:=_	Goth}');
Error('\P{Is_Scx=:=_	Goth}');
Expect(1, 66378, '\p{Is_Scx=goth}', "");
Expect(0, 66378, '\p{^Is_Scx=goth}', "");
Expect(0, 66378, '\P{Is_Scx=goth}', "");
Expect(1, 66378, '\P{^Is_Scx=goth}', "");
Expect(0, 66379, '\p{Is_Scx=goth}', "");
Expect(1, 66379, '\p{^Is_Scx=goth}', "");
Expect(1, 66379, '\P{Is_Scx=goth}', "");
Expect(0, 66379, '\P{^Is_Scx=goth}', "");
Expect(1, 66378, '\p{Is_Scx=_ Goth}', "");
Expect(0, 66378, '\p{^Is_Scx=_ Goth}', "");
Expect(0, 66378, '\P{Is_Scx=_ Goth}', "");
Expect(1, 66378, '\P{^Is_Scx=_ Goth}', "");
Expect(0, 66379, '\p{Is_Scx=_ Goth}', "");
Expect(1, 66379, '\p{^Is_Scx=_ Goth}', "");
Expect(1, 66379, '\P{Is_Scx=_ Goth}', "");
Expect(0, 66379, '\P{^Is_Scx=_ Goth}', "");
Error('\p{Script_Extensions=/a/-Greek}');
Error('\P{Script_Extensions=/a/-Greek}');
Expect(1, 119365, '\p{Script_Extensions:greek}', "");
Expect(0, 119365, '\p{^Script_Extensions:greek}', "");
Expect(0, 119365, '\P{Script_Extensions:greek}', "");
Expect(1, 119365, '\P{^Script_Extensions:greek}', "");
Expect(0, 119366, '\p{Script_Extensions:greek}', "");
Expect(1, 119366, '\p{^Script_Extensions:greek}', "");
Expect(1, 119366, '\P{Script_Extensions:greek}', "");
Expect(0, 119366, '\P{^Script_Extensions:greek}', "");
Expect(1, 119365, '\p{Script_Extensions:__greek}', "");
Expect(0, 119365, '\p{^Script_Extensions:__greek}', "");
Expect(0, 119365, '\P{Script_Extensions:__greek}', "");
Expect(1, 119365, '\P{^Script_Extensions:__greek}', "");
Expect(0, 119366, '\p{Script_Extensions:__greek}', "");
Expect(1, 119366, '\p{^Script_Extensions:__greek}', "");
Expect(1, 119366, '\P{Script_Extensions:__greek}', "");
Expect(0, 119366, '\P{^Script_Extensions:__greek}', "");
Error('\p{Scx=	grek/a/}');
Error('\P{Scx=	grek/a/}');
Expect(1, 119365, '\p{Scx:grek}', "");
Expect(0, 119365, '\p{^Scx:grek}', "");
Expect(0, 119365, '\P{Scx:grek}', "");
Expect(1, 119365, '\P{^Scx:grek}', "");
Expect(0, 119366, '\p{Scx:grek}', "");
Expect(1, 119366, '\p{^Scx:grek}', "");
Expect(1, 119366, '\P{Scx:grek}', "");
Expect(0, 119366, '\P{^Scx:grek}', "");
Expect(1, 119365, '\p{Scx=-	GREK}', "");
Expect(0, 119365, '\p{^Scx=-	GREK}', "");
Expect(0, 119365, '\P{Scx=-	GREK}', "");
Expect(1, 119365, '\P{^Scx=-	GREK}', "");
Expect(0, 119366, '\p{Scx=-	GREK}', "");
Expect(1, 119366, '\p{^Scx=-	GREK}', "");
Expect(1, 119366, '\P{Scx=-	GREK}', "");
Expect(0, 119366, '\P{^Scx=-	GREK}', "");
Error('\p{Is_Script_Extensions:/a/-_GREEK}');
Error('\P{Is_Script_Extensions:/a/-_GREEK}');
Expect(1, 119365, '\p{Is_Script_Extensions=greek}', "");
Expect(0, 119365, '\p{^Is_Script_Extensions=greek}', "");
Expect(0, 119365, '\P{Is_Script_Extensions=greek}', "");
Expect(1, 119365, '\P{^Is_Script_Extensions=greek}', "");
Expect(0, 119366, '\p{Is_Script_Extensions=greek}', "");
Expect(1, 119366, '\p{^Is_Script_Extensions=greek}', "");
Expect(1, 119366, '\P{Is_Script_Extensions=greek}', "");
Expect(0, 119366, '\P{^Is_Script_Extensions=greek}', "");
Expect(1, 119365, '\p{Is_Script_Extensions=  greek}', "");
Expect(0, 119365, '\p{^Is_Script_Extensions=  greek}', "");
Expect(0, 119365, '\P{Is_Script_Extensions=  greek}', "");
Expect(1, 119365, '\P{^Is_Script_Extensions=  greek}', "");
Expect(0, 119366, '\p{Is_Script_Extensions=  greek}', "");
Expect(1, 119366, '\p{^Is_Script_Extensions=  greek}', "");
Expect(1, 119366, '\P{Is_Script_Extensions=  greek}', "");
Expect(0, 119366, '\P{^Is_Script_Extensions=  greek}', "");
Error('\p{Is_Scx=-_Grek/a/}');
Error('\P{Is_Scx=-_Grek/a/}');
Expect(1, 119365, '\p{Is_Scx=grek}', "");
Expect(0, 119365, '\p{^Is_Scx=grek}', "");
Expect(0, 119365, '\P{Is_Scx=grek}', "");
Expect(1, 119365, '\P{^Is_Scx=grek}', "");
Expect(0, 119366, '\p{Is_Scx=grek}', "");
Expect(1, 119366, '\p{^Is_Scx=grek}', "");
Expect(1, 119366, '\P{Is_Scx=grek}', "");
Expect(0, 119366, '\P{^Is_Scx=grek}', "");
Expect(1, 119365, '\p{Is_Scx= 	grek}', "");
Expect(0, 119365, '\p{^Is_Scx= 	grek}', "");
Expect(0, 119365, '\P{Is_Scx= 	grek}', "");
Expect(1, 119365, '\P{^Is_Scx= 	grek}', "");
Expect(0, 119366, '\p{Is_Scx= 	grek}', "");
Expect(1, 119366, '\p{^Is_Scx= 	grek}', "");
Expect(1, 119366, '\P{Is_Scx= 	grek}', "");
Expect(0, 119366, '\P{^Is_Scx= 	grek}', "");
Error('\p{Script_Extensions=:=- GUJARATI}');
Error('\P{Script_Extensions=:=- GUJARATI}');
Expect(1, 43065, '\p{Script_Extensions=gujarati}', "");
Expect(0, 43065, '\p{^Script_Extensions=gujarati}', "");
Expect(0, 43065, '\P{Script_Extensions=gujarati}', "");
Expect(1, 43065, '\P{^Script_Extensions=gujarati}', "");
Expect(0, 43066, '\p{Script_Extensions=gujarati}', "");
Expect(1, 43066, '\p{^Script_Extensions=gujarati}', "");
Expect(1, 43066, '\P{Script_Extensions=gujarati}', "");
Expect(0, 43066, '\P{^Script_Extensions=gujarati}', "");
Expect(1, 43065, '\p{Script_Extensions=	Gujarati}', "");
Expect(0, 43065, '\p{^Script_Extensions=	Gujarati}', "");
Expect(0, 43065, '\P{Script_Extensions=	Gujarati}', "");
Expect(1, 43065, '\P{^Script_Extensions=	Gujarati}', "");
Expect(0, 43066, '\p{Script_Extensions=	Gujarati}', "");
Expect(1, 43066, '\p{^Script_Extensions=	Gujarati}', "");
Expect(1, 43066, '\P{Script_Extensions=	Gujarati}', "");
Expect(0, 43066, '\P{^Script_Extensions=	Gujarati}', "");
Error('\p{Scx=/a/__GUJR}');
Error('\P{Scx=/a/__GUJR}');
Expect(1, 43065, '\p{Scx:   gujr}', "");
Expect(0, 43065, '\p{^Scx:   gujr}', "");
Expect(0, 43065, '\P{Scx:   gujr}', "");
Expect(1, 43065, '\P{^Scx:   gujr}', "");
Expect(0, 43066, '\p{Scx:   gujr}', "");
Expect(1, 43066, '\p{^Scx:   gujr}', "");
Expect(1, 43066, '\P{Scx:   gujr}', "");
Expect(0, 43066, '\P{^Scx:   gujr}', "");
Expect(1, 43065, '\p{Scx=		Gujr}', "");
Expect(0, 43065, '\p{^Scx=		Gujr}', "");
Expect(0, 43065, '\P{Scx=		Gujr}', "");
Expect(1, 43065, '\P{^Scx=		Gujr}', "");
Expect(0, 43066, '\p{Scx=		Gujr}', "");
Expect(1, 43066, '\p{^Scx=		Gujr}', "");
Expect(1, 43066, '\P{Scx=		Gujr}', "");
Expect(0, 43066, '\P{^Scx=		Gujr}', "");
Error('\p{Is_Script_Extensions=	gujarati/a/}');
Error('\P{Is_Script_Extensions=	gujarati/a/}');
Expect(1, 43065, '\p{Is_Script_Extensions=gujarati}', "");
Expect(0, 43065, '\p{^Is_Script_Extensions=gujarati}', "");
Expect(0, 43065, '\P{Is_Script_Extensions=gujarati}', "");
Expect(1, 43065, '\P{^Is_Script_Extensions=gujarati}', "");
Expect(0, 43066, '\p{Is_Script_Extensions=gujarati}', "");
Expect(1, 43066, '\p{^Is_Script_Extensions=gujarati}', "");
Expect(1, 43066, '\P{Is_Script_Extensions=gujarati}', "");
Expect(0, 43066, '\P{^Is_Script_Extensions=gujarati}', "");
Expect(1, 43065, '\p{Is_Script_Extensions=		gujarati}', "");
Expect(0, 43065, '\p{^Is_Script_Extensions=		gujarati}', "");
Expect(0, 43065, '\P{Is_Script_Extensions=		gujarati}', "");
Expect(1, 43065, '\P{^Is_Script_Extensions=		gujarati}', "");
Expect(0, 43066, '\p{Is_Script_Extensions=		gujarati}', "");
Expect(1, 43066, '\p{^Is_Script_Extensions=		gujarati}', "");
Expect(1, 43066, '\P{Is_Script_Extensions=		gujarati}', "");
Expect(0, 43066, '\P{^Is_Script_Extensions=		gujarati}', "");
Error('\p{Is_Scx=/a/gujr}');
Error('\P{Is_Scx=/a/gujr}');
Expect(1, 43065, '\p{Is_Scx=gujr}', "");
Expect(0, 43065, '\p{^Is_Scx=gujr}', "");
Expect(0, 43065, '\P{Is_Scx=gujr}', "");
Expect(1, 43065, '\P{^Is_Scx=gujr}', "");
Expect(0, 43066, '\p{Is_Scx=gujr}', "");
Expect(1, 43066, '\p{^Is_Scx=gujr}', "");
Expect(1, 43066, '\P{Is_Scx=gujr}', "");
Expect(0, 43066, '\P{^Is_Scx=gujr}', "");
Expect(1, 43065, '\p{Is_Scx=	GUJR}', "");
Expect(0, 43065, '\p{^Is_Scx=	GUJR}', "");
Expect(0, 43065, '\P{Is_Scx=	GUJR}', "");
Expect(1, 43065, '\P{^Is_Scx=	GUJR}', "");
Expect(0, 43066, '\p{Is_Scx=	GUJR}', "");
Expect(1, 43066, '\p{^Is_Scx=	GUJR}', "");
Expect(1, 43066, '\P{Is_Scx=	GUJR}', "");
Expect(0, 43066, '\P{^Is_Scx=	GUJR}', "");
Error('\p{Script_Extensions=/a/--gurmukhi}');
Error('\P{Script_Extensions=/a/--gurmukhi}');
Expect(1, 43065, '\p{Script_Extensions=gurmukhi}', "");
Expect(0, 43065, '\p{^Script_Extensions=gurmukhi}', "");
Expect(0, 43065, '\P{Script_Extensions=gurmukhi}', "");
Expect(1, 43065, '\P{^Script_Extensions=gurmukhi}', "");
Expect(0, 43066, '\p{Script_Extensions=gurmukhi}', "");
Expect(1, 43066, '\p{^Script_Extensions=gurmukhi}', "");
Expect(1, 43066, '\P{Script_Extensions=gurmukhi}', "");
Expect(0, 43066, '\P{^Script_Extensions=gurmukhi}', "");
Expect(1, 43065, '\p{Script_Extensions= -Gurmukhi}', "");
Expect(0, 43065, '\p{^Script_Extensions= -Gurmukhi}', "");
Expect(0, 43065, '\P{Script_Extensions= -Gurmukhi}', "");
Expect(1, 43065, '\P{^Script_Extensions= -Gurmukhi}', "");
Expect(0, 43066, '\p{Script_Extensions= -Gurmukhi}', "");
Expect(1, 43066, '\p{^Script_Extensions= -Gurmukhi}', "");
Expect(1, 43066, '\P{Script_Extensions= -Gurmukhi}', "");
Expect(0, 43066, '\P{^Script_Extensions= -Gurmukhi}', "");
Error('\p{Scx=/a/GURU}');
Error('\P{Scx=/a/GURU}');
Expect(1, 43065, '\p{Scx=guru}', "");
Expect(0, 43065, '\p{^Scx=guru}', "");
Expect(0, 43065, '\P{Scx=guru}', "");
Expect(1, 43065, '\P{^Scx=guru}', "");
Expect(0, 43066, '\p{Scx=guru}', "");
Expect(1, 43066, '\p{^Scx=guru}', "");
Expect(1, 43066, '\P{Scx=guru}', "");
Expect(0, 43066, '\P{^Scx=guru}', "");
Expect(1, 43065, '\p{Scx=-	GURU}', "");
Expect(0, 43065, '\p{^Scx=-	GURU}', "");
Expect(0, 43065, '\P{Scx=-	GURU}', "");
Expect(1, 43065, '\P{^Scx=-	GURU}', "");
Expect(0, 43066, '\p{Scx=-	GURU}', "");
Expect(1, 43066, '\p{^Scx=-	GURU}', "");
Expect(1, 43066, '\P{Scx=-	GURU}', "");
Expect(0, 43066, '\P{^Scx=-	GURU}', "");
Error('\p{Is_Script_Extensions:    -gurmukhi/a/}');
Error('\P{Is_Script_Extensions:    -gurmukhi/a/}');
Expect(1, 43065, '\p{Is_Script_Extensions=gurmukhi}', "");
Expect(0, 43065, '\p{^Is_Script_Extensions=gurmukhi}', "");
Expect(0, 43065, '\P{Is_Script_Extensions=gurmukhi}', "");
Expect(1, 43065, '\P{^Is_Script_Extensions=gurmukhi}', "");
Expect(0, 43066, '\p{Is_Script_Extensions=gurmukhi}', "");
Expect(1, 43066, '\p{^Is_Script_Extensions=gurmukhi}', "");
Expect(1, 43066, '\P{Is_Script_Extensions=gurmukhi}', "");
Expect(0, 43066, '\P{^Is_Script_Extensions=gurmukhi}', "");
Expect(1, 43065, '\p{Is_Script_Extensions= 	Gurmukhi}', "");
Expect(0, 43065, '\p{^Is_Script_Extensions= 	Gurmukhi}', "");
Expect(0, 43065, '\P{Is_Script_Extensions= 	Gurmukhi}', "");
Expect(1, 43065, '\P{^Is_Script_Extensions= 	Gurmukhi}', "");
Expect(0, 43066, '\p{Is_Script_Extensions= 	Gurmukhi}', "");
Expect(1, 43066, '\p{^Is_Script_Extensions= 	Gurmukhi}', "");
Expect(1, 43066, '\P{Is_Script_Extensions= 	Gurmukhi}', "");
Expect(0, 43066, '\P{^Is_Script_Extensions= 	Gurmukhi}', "");
Error('\p{Is_Scx=	Guru:=}');
Error('\P{Is_Scx=	Guru:=}');
Expect(1, 43065, '\p{Is_Scx=guru}', "");
Expect(0, 43065, '\p{^Is_Scx=guru}', "");
Expect(0, 43065, '\P{Is_Scx=guru}', "");
Expect(1, 43065, '\P{^Is_Scx=guru}', "");
Expect(0, 43066, '\p{Is_Scx=guru}', "");
Expect(1, 43066, '\p{^Is_Scx=guru}', "");
Expect(1, 43066, '\P{Is_Scx=guru}', "");
Expect(0, 43066, '\P{^Is_Scx=guru}', "");
Expect(1, 43065, '\p{Is_Scx=	_Guru}', "");
Expect(0, 43065, '\p{^Is_Scx=	_Guru}', "");
Expect(0, 43065, '\P{Is_Scx=	_Guru}', "");
Expect(1, 43065, '\P{^Is_Scx=	_Guru}', "");
Expect(0, 43066, '\p{Is_Scx=	_Guru}', "");
Expect(1, 43066, '\p{^Is_Scx=	_Guru}', "");
Expect(1, 43066, '\P{Is_Scx=	_Guru}', "");
Expect(0, 43066, '\P{^Is_Scx=	_Guru}', "");
Error('\p{Script_Extensions=/a/	_HANGUL}');
Error('\P{Script_Extensions=/a/	_HANGUL}');
Expect(1, 65500, '\p{Script_Extensions=hangul}', "");
Expect(0, 65500, '\p{^Script_Extensions=hangul}', "");
Expect(0, 65500, '\P{Script_Extensions=hangul}', "");
Expect(1, 65500, '\P{^Script_Extensions=hangul}', "");
Expect(0, 65501, '\p{Script_Extensions=hangul}', "");
Expect(1, 65501, '\p{^Script_Extensions=hangul}', "");
Expect(1, 65501, '\P{Script_Extensions=hangul}', "");
Expect(0, 65501, '\P{^Script_Extensions=hangul}', "");
Expect(1, 65500, '\p{Script_Extensions=- HANGUL}', "");
Expect(0, 65500, '\p{^Script_Extensions=- HANGUL}', "");
Expect(0, 65500, '\P{Script_Extensions=- HANGUL}', "");
Expect(1, 65500, '\P{^Script_Extensions=- HANGUL}', "");
Expect(0, 65501, '\p{Script_Extensions=- HANGUL}', "");
Expect(1, 65501, '\p{^Script_Extensions=- HANGUL}', "");
Expect(1, 65501, '\P{Script_Extensions=- HANGUL}', "");
Expect(0, 65501, '\P{^Script_Extensions=- HANGUL}', "");
Error('\p{Scx=- Hang:=}');
Error('\P{Scx=- Hang:=}');
Expect(1, 65500, '\p{Scx=hang}', "");
Expect(0, 65500, '\p{^Scx=hang}', "");
Expect(0, 65500, '\P{Scx=hang}', "");
Expect(1, 65500, '\P{^Scx=hang}', "");
Expect(0, 65501, '\p{Scx=hang}', "");
Expect(1, 65501, '\p{^Scx=hang}', "");
Expect(1, 65501, '\P{Scx=hang}', "");
Expect(0, 65501, '\P{^Scx=hang}', "");
Expect(1, 65500, '\p{Scx=-	HANG}', "");
Expect(0, 65500, '\p{^Scx=-	HANG}', "");
Expect(0, 65500, '\P{Scx=-	HANG}', "");
Expect(1, 65500, '\P{^Scx=-	HANG}', "");
Expect(0, 65501, '\p{Scx=-	HANG}', "");
Expect(1, 65501, '\p{^Scx=-	HANG}', "");
Expect(1, 65501, '\P{Scx=-	HANG}', "");
Expect(0, 65501, '\P{^Scx=-	HANG}', "");
Error('\p{Is_Script_Extensions=/a/-_Hangul}');
Error('\P{Is_Script_Extensions=/a/-_Hangul}');
Expect(1, 65500, '\p{Is_Script_Extensions=hangul}', "");
Expect(0, 65500, '\p{^Is_Script_Extensions=hangul}', "");
Expect(0, 65500, '\P{Is_Script_Extensions=hangul}', "");
Expect(1, 65500, '\P{^Is_Script_Extensions=hangul}', "");
Expect(0, 65501, '\p{Is_Script_Extensions=hangul}', "");
Expect(1, 65501, '\p{^Is_Script_Extensions=hangul}', "");
Expect(1, 65501, '\P{Is_Script_Extensions=hangul}', "");
Expect(0, 65501, '\P{^Is_Script_Extensions=hangul}', "");
Expect(1, 65500, '\p{Is_Script_Extensions=	-HANGUL}', "");
Expect(0, 65500, '\p{^Is_Script_Extensions=	-HANGUL}', "");
Expect(0, 65500, '\P{Is_Script_Extensions=	-HANGUL}', "");
Expect(1, 65500, '\P{^Is_Script_Extensions=	-HANGUL}', "");
Expect(0, 65501, '\p{Is_Script_Extensions=	-HANGUL}', "");
Expect(1, 65501, '\p{^Is_Script_Extensions=	-HANGUL}', "");
Expect(1, 65501, '\P{Is_Script_Extensions=	-HANGUL}', "");
Expect(0, 65501, '\P{^Is_Script_Extensions=	-HANGUL}', "");
Error('\p{Is_Scx= -hang/a/}');
Error('\P{Is_Scx= -hang/a/}');
Expect(1, 65500, '\p{Is_Scx=hang}', "");
Expect(0, 65500, '\p{^Is_Scx=hang}', "");
Expect(0, 65500, '\P{Is_Scx=hang}', "");
Expect(1, 65500, '\P{^Is_Scx=hang}', "");
Expect(0, 65501, '\p{Is_Scx=hang}', "");
Expect(1, 65501, '\p{^Is_Scx=hang}', "");
Expect(1, 65501, '\P{Is_Scx=hang}', "");
Expect(0, 65501, '\P{^Is_Scx=hang}', "");
Expect(1, 65500, '\p{Is_Scx:  Hang}', "");
Expect(0, 65500, '\p{^Is_Scx:  Hang}', "");
Expect(0, 65500, '\P{Is_Scx:  Hang}', "");
Expect(1, 65500, '\P{^Is_Scx:  Hang}', "");
Expect(0, 65501, '\p{Is_Scx:  Hang}', "");
Expect(1, 65501, '\p{^Is_Scx:  Hang}', "");
Expect(1, 65501, '\P{Is_Scx:  Hang}', "");
Expect(0, 65501, '\P{^Is_Scx:  Hang}', "");
Error('\p{Script_Extensions=/a/_-Han}');
Error('\P{Script_Extensions=/a/_-Han}');
Expect(1, 195101, '\p{Script_Extensions=han}', "");
Expect(0, 195101, '\p{^Script_Extensions=han}', "");
Expect(0, 195101, '\P{Script_Extensions=han}', "");
Expect(1, 195101, '\P{^Script_Extensions=han}', "");
Expect(0, 195102, '\p{Script_Extensions=han}', "");
Expect(1, 195102, '\p{^Script_Extensions=han}', "");
Expect(1, 195102, '\P{Script_Extensions=han}', "");
Expect(0, 195102, '\P{^Script_Extensions=han}', "");
Expect(1, 195101, '\p{Script_Extensions=_-Han}', "");
Expect(0, 195101, '\p{^Script_Extensions=_-Han}', "");
Expect(0, 195101, '\P{Script_Extensions=_-Han}', "");
Expect(1, 195101, '\P{^Script_Extensions=_-Han}', "");
Expect(0, 195102, '\p{Script_Extensions=_-Han}', "");
Expect(1, 195102, '\p{^Script_Extensions=_-Han}', "");
Expect(1, 195102, '\P{Script_Extensions=_-Han}', "");
Expect(0, 195102, '\P{^Script_Extensions=_-Han}', "");
Error('\p{Scx=	/a/hani}');
Error('\P{Scx=	/a/hani}');
Expect(1, 195101, '\p{Scx=hani}', "");
Expect(0, 195101, '\p{^Scx=hani}', "");
Expect(0, 195101, '\P{Scx=hani}', "");
Expect(1, 195101, '\P{^Scx=hani}', "");
Expect(0, 195102, '\p{Scx=hani}', "");
Expect(1, 195102, '\p{^Scx=hani}', "");
Expect(1, 195102, '\P{Scx=hani}', "");
Expect(0, 195102, '\P{^Scx=hani}', "");
Expect(1, 195101, '\p{Scx=_-hani}', "");
Expect(0, 195101, '\p{^Scx=_-hani}', "");
Expect(0, 195101, '\P{Scx=_-hani}', "");
Expect(1, 195101, '\P{^Scx=_-hani}', "");
Expect(0, 195102, '\p{Scx=_-hani}', "");
Expect(1, 195102, '\p{^Scx=_-hani}', "");
Expect(1, 195102, '\P{Scx=_-hani}', "");
Expect(0, 195102, '\P{^Scx=_-hani}', "");
Error('\p{Is_Script_Extensions=:=Han}');
Error('\P{Is_Script_Extensions=:=Han}');
Expect(1, 195101, '\p{Is_Script_Extensions=han}', "");
Expect(0, 195101, '\p{^Is_Script_Extensions=han}', "");
Expect(0, 195101, '\P{Is_Script_Extensions=han}', "");
Expect(1, 195101, '\P{^Is_Script_Extensions=han}', "");
Expect(0, 195102, '\p{Is_Script_Extensions=han}', "");
Expect(1, 195102, '\p{^Is_Script_Extensions=han}', "");
Expect(1, 195102, '\P{Is_Script_Extensions=han}', "");
Expect(0, 195102, '\P{^Is_Script_Extensions=han}', "");
Expect(1, 195101, '\p{Is_Script_Extensions= 	han}', "");
Expect(0, 195101, '\p{^Is_Script_Extensions= 	han}', "");
Expect(0, 195101, '\P{Is_Script_Extensions= 	han}', "");
Expect(1, 195101, '\P{^Is_Script_Extensions= 	han}', "");
Expect(0, 195102, '\p{Is_Script_Extensions= 	han}', "");
Expect(1, 195102, '\p{^Is_Script_Extensions= 	han}', "");
Expect(1, 195102, '\P{Is_Script_Extensions= 	han}', "");
Expect(0, 195102, '\P{^Is_Script_Extensions= 	han}', "");
Error('\p{Is_Scx=/a/	HANI}');
Error('\P{Is_Scx=/a/	HANI}');
Expect(1, 195101, '\p{Is_Scx=hani}', "");
Expect(0, 195101, '\p{^Is_Scx=hani}', "");
Expect(0, 195101, '\P{Is_Scx=hani}', "");
Expect(1, 195101, '\P{^Is_Scx=hani}', "");
Expect(0, 195102, '\p{Is_Scx=hani}', "");
Expect(1, 195102, '\p{^Is_Scx=hani}', "");
Expect(1, 195102, '\P{Is_Scx=hani}', "");
Expect(0, 195102, '\P{^Is_Scx=hani}', "");
Expect(1, 195101, '\p{Is_Scx=-_Hani}', "");
Expect(0, 195101, '\p{^Is_Scx=-_Hani}', "");
Expect(0, 195101, '\P{Is_Scx=-_Hani}', "");
Expect(1, 195101, '\P{^Is_Scx=-_Hani}', "");
Expect(0, 195102, '\p{Is_Scx=-_Hani}', "");
Expect(1, 195102, '\p{^Is_Scx=-_Hani}', "");
Expect(1, 195102, '\P{Is_Scx=-_Hani}', "");
Expect(0, 195102, '\P{^Is_Scx=-_Hani}', "");
Error('\p{Script_Extensions=/a/Hanunoo}');
Error('\P{Script_Extensions=/a/Hanunoo}');
Expect(1, 5942, '\p{Script_Extensions=hanunoo}', "");
Expect(0, 5942, '\p{^Script_Extensions=hanunoo}', "");
Expect(0, 5942, '\P{Script_Extensions=hanunoo}', "");
Expect(1, 5942, '\P{^Script_Extensions=hanunoo}', "");
Expect(0, 5943, '\p{Script_Extensions=hanunoo}', "");
Expect(1, 5943, '\p{^Script_Extensions=hanunoo}', "");
Expect(1, 5943, '\P{Script_Extensions=hanunoo}', "");
Expect(0, 5943, '\P{^Script_Extensions=hanunoo}', "");
Expect(1, 5942, '\p{Script_Extensions=_ hanunoo}', "");
Expect(0, 5942, '\p{^Script_Extensions=_ hanunoo}', "");
Expect(0, 5942, '\P{Script_Extensions=_ hanunoo}', "");
Expect(1, 5942, '\P{^Script_Extensions=_ hanunoo}', "");
Expect(0, 5943, '\p{Script_Extensions=_ hanunoo}', "");
Expect(1, 5943, '\p{^Script_Extensions=_ hanunoo}', "");
Expect(1, 5943, '\P{Script_Extensions=_ hanunoo}', "");
Expect(0, 5943, '\P{^Script_Extensions=_ hanunoo}', "");
Error('\p{Scx= 	hano/a/}');
Error('\P{Scx= 	hano/a/}');
Expect(1, 5942, '\p{Scx=hano}', "");
Expect(0, 5942, '\p{^Scx=hano}', "");
Expect(0, 5942, '\P{Scx=hano}', "");
Expect(1, 5942, '\P{^Scx=hano}', "");
Expect(0, 5943, '\p{Scx=hano}', "");
Expect(1, 5943, '\p{^Scx=hano}', "");
Expect(1, 5943, '\P{Scx=hano}', "");
Expect(0, 5943, '\P{^Scx=hano}', "");
Expect(1, 5942, '\p{Scx=-	hano}', "");
Expect(0, 5942, '\p{^Scx=-	hano}', "");
Expect(0, 5942, '\P{Scx=-	hano}', "");
Expect(1, 5942, '\P{^Scx=-	hano}', "");
Expect(0, 5943, '\p{Scx=-	hano}', "");
Expect(1, 5943, '\p{^Scx=-	hano}', "");
Expect(1, 5943, '\P{Scx=-	hano}', "");
Expect(0, 5943, '\P{^Scx=-	hano}', "");
Error('\p{Is_Script_Extensions=	Hanunoo:=}');
Error('\P{Is_Script_Extensions=	Hanunoo:=}');
Expect(1, 5942, '\p{Is_Script_Extensions=hanunoo}', "");
Expect(0, 5942, '\p{^Is_Script_Extensions=hanunoo}', "");
Expect(0, 5942, '\P{Is_Script_Extensions=hanunoo}', "");
Expect(1, 5942, '\P{^Is_Script_Extensions=hanunoo}', "");
Expect(0, 5943, '\p{Is_Script_Extensions=hanunoo}', "");
Expect(1, 5943, '\p{^Is_Script_Extensions=hanunoo}', "");
Expect(1, 5943, '\P{Is_Script_Extensions=hanunoo}', "");
Expect(0, 5943, '\P{^Is_Script_Extensions=hanunoo}', "");
Expect(1, 5942, '\p{Is_Script_Extensions= -HANUNOO}', "");
Expect(0, 5942, '\p{^Is_Script_Extensions= -HANUNOO}', "");
Expect(0, 5942, '\P{Is_Script_Extensions= -HANUNOO}', "");
Expect(1, 5942, '\P{^Is_Script_Extensions= -HANUNOO}', "");
Expect(0, 5943, '\p{Is_Script_Extensions= -HANUNOO}', "");
Expect(1, 5943, '\p{^Is_Script_Extensions= -HANUNOO}', "");
Expect(1, 5943, '\P{Is_Script_Extensions= -HANUNOO}', "");
Expect(0, 5943, '\P{^Is_Script_Extensions= -HANUNOO}', "");
Error('\p{Is_Scx=:=Hano}');
Error('\P{Is_Scx=:=Hano}');
Expect(1, 5942, '\p{Is_Scx=hano}', "");
Expect(0, 5942, '\p{^Is_Scx=hano}', "");
Expect(0, 5942, '\P{Is_Scx=hano}', "");
Expect(1, 5942, '\P{^Is_Scx=hano}', "");
Expect(0, 5943, '\p{Is_Scx=hano}', "");
Expect(1, 5943, '\p{^Is_Scx=hano}', "");
Expect(1, 5943, '\P{Is_Scx=hano}', "");
Expect(0, 5943, '\P{^Is_Scx=hano}', "");
Expect(1, 5942, '\p{Is_Scx=	-Hano}', "");
Expect(0, 5942, '\p{^Is_Scx=	-Hano}', "");
Expect(0, 5942, '\P{Is_Scx=	-Hano}', "");
Expect(1, 5942, '\P{^Is_Scx=	-Hano}', "");
Expect(0, 5943, '\p{Is_Scx=	-Hano}', "");
Expect(1, 5943, '\p{^Is_Scx=	-Hano}', "");
Expect(1, 5943, '\P{Is_Scx=	-Hano}', "");
Expect(0, 5943, '\P{^Is_Scx=	-Hano}', "");
Error('\p{Script_Extensions=_:=Hebrew}');
Error('\P{Script_Extensions=_:=Hebrew}');
Expect(1, 64335, '\p{Script_Extensions=hebrew}', "");
Expect(0, 64335, '\p{^Script_Extensions=hebrew}', "");
Expect(0, 64335, '\P{Script_Extensions=hebrew}', "");
Expect(1, 64335, '\P{^Script_Extensions=hebrew}', "");
Expect(0, 64336, '\p{Script_Extensions=hebrew}', "");
Expect(1, 64336, '\p{^Script_Extensions=hebrew}', "");
Expect(1, 64336, '\P{Script_Extensions=hebrew}', "");
Expect(0, 64336, '\P{^Script_Extensions=hebrew}', "");
Expect(1, 64335, '\p{Script_Extensions=	_hebrew}', "");
Expect(0, 64335, '\p{^Script_Extensions=	_hebrew}', "");
Expect(0, 64335, '\P{Script_Extensions=	_hebrew}', "");
Expect(1, 64335, '\P{^Script_Extensions=	_hebrew}', "");
Expect(0, 64336, '\p{Script_Extensions=	_hebrew}', "");
Expect(1, 64336, '\p{^Script_Extensions=	_hebrew}', "");
Expect(1, 64336, '\P{Script_Extensions=	_hebrew}', "");
Expect(0, 64336, '\P{^Script_Extensions=	_hebrew}', "");
Error('\p{Scx=	/a/hebr}');
Error('\P{Scx=	/a/hebr}');
Expect(1, 64335, '\p{Scx: hebr}', "");
Expect(0, 64335, '\p{^Scx: hebr}', "");
Expect(0, 64335, '\P{Scx: hebr}', "");
Expect(1, 64335, '\P{^Scx: hebr}', "");
Expect(0, 64336, '\p{Scx: hebr}', "");
Expect(1, 64336, '\p{^Scx: hebr}', "");
Expect(1, 64336, '\P{Scx: hebr}', "");
Expect(0, 64336, '\P{^Scx: hebr}', "");
Expect(1, 64335, '\p{Scx= HEBR}', "");
Expect(0, 64335, '\p{^Scx= HEBR}', "");
Expect(0, 64335, '\P{Scx= HEBR}', "");
Expect(1, 64335, '\P{^Scx= HEBR}', "");
Expect(0, 64336, '\p{Scx= HEBR}', "");
Expect(1, 64336, '\p{^Scx= HEBR}', "");
Expect(1, 64336, '\P{Scx= HEBR}', "");
Expect(0, 64336, '\P{^Scx= HEBR}', "");
Error('\p{Is_Script_Extensions= /a/HEBREW}');
Error('\P{Is_Script_Extensions= /a/HEBREW}');
Expect(1, 64335, '\p{Is_Script_Extensions=hebrew}', "");
Expect(0, 64335, '\p{^Is_Script_Extensions=hebrew}', "");
Expect(0, 64335, '\P{Is_Script_Extensions=hebrew}', "");
Expect(1, 64335, '\P{^Is_Script_Extensions=hebrew}', "");
Expect(0, 64336, '\p{Is_Script_Extensions=hebrew}', "");
Expect(1, 64336, '\p{^Is_Script_Extensions=hebrew}', "");
Expect(1, 64336, '\P{Is_Script_Extensions=hebrew}', "");
Expect(0, 64336, '\P{^Is_Script_Extensions=hebrew}', "");
Expect(1, 64335, '\p{Is_Script_Extensions=_HEBREW}', "");
Expect(0, 64335, '\p{^Is_Script_Extensions=_HEBREW}', "");
Expect(0, 64335, '\P{Is_Script_Extensions=_HEBREW}', "");
Expect(1, 64335, '\P{^Is_Script_Extensions=_HEBREW}', "");
Expect(0, 64336, '\p{Is_Script_Extensions=_HEBREW}', "");
Expect(1, 64336, '\p{^Is_Script_Extensions=_HEBREW}', "");
Expect(1, 64336, '\P{Is_Script_Extensions=_HEBREW}', "");
Expect(0, 64336, '\P{^Is_Script_Extensions=_HEBREW}', "");
Error('\p{Is_Scx: /a/hebr}');
Error('\P{Is_Scx: /a/hebr}');
Expect(1, 64335, '\p{Is_Scx=hebr}', "");
Expect(0, 64335, '\p{^Is_Scx=hebr}', "");
Expect(0, 64335, '\P{Is_Scx=hebr}', "");
Expect(1, 64335, '\P{^Is_Scx=hebr}', "");
Expect(0, 64336, '\p{Is_Scx=hebr}', "");
Expect(1, 64336, '\p{^Is_Scx=hebr}', "");
Expect(1, 64336, '\P{Is_Scx=hebr}', "");
Expect(0, 64336, '\P{^Is_Scx=hebr}', "");
Expect(1, 64335, '\p{Is_Scx=-_Hebr}', "");
Expect(0, 64335, '\p{^Is_Scx=-_Hebr}', "");
Expect(0, 64335, '\P{Is_Scx=-_Hebr}', "");
Expect(1, 64335, '\P{^Is_Scx=-_Hebr}', "");
Expect(0, 64336, '\p{Is_Scx=-_Hebr}', "");
Expect(1, 64336, '\p{^Is_Scx=-_Hebr}', "");
Expect(1, 64336, '\P{Is_Scx=-_Hebr}', "");
Expect(0, 64336, '\P{^Is_Scx=-_Hebr}', "");
Error('\p{Script_Extensions= :=hiragana}');
Error('\P{Script_Extensions= :=hiragana}');
Expect(1, 127488, '\p{Script_Extensions=hiragana}', "");
Expect(0, 127488, '\p{^Script_Extensions=hiragana}', "");
Expect(0, 127488, '\P{Script_Extensions=hiragana}', "");
Expect(1, 127488, '\P{^Script_Extensions=hiragana}', "");
Expect(0, 127489, '\p{Script_Extensions=hiragana}', "");
Expect(1, 127489, '\p{^Script_Extensions=hiragana}', "");
Expect(1, 127489, '\P{Script_Extensions=hiragana}', "");
Expect(0, 127489, '\P{^Script_Extensions=hiragana}', "");
Expect(1, 127488, '\p{Script_Extensions=-_HIRAGANA}', "");
Expect(0, 127488, '\p{^Script_Extensions=-_HIRAGANA}', "");
Expect(0, 127488, '\P{Script_Extensions=-_HIRAGANA}', "");
Expect(1, 127488, '\P{^Script_Extensions=-_HIRAGANA}', "");
Expect(0, 127489, '\p{Script_Extensions=-_HIRAGANA}', "");
Expect(1, 127489, '\p{^Script_Extensions=-_HIRAGANA}', "");
Expect(1, 127489, '\P{Script_Extensions=-_HIRAGANA}', "");
Expect(0, 127489, '\P{^Script_Extensions=-_HIRAGANA}', "");
Error('\p{Scx=  hira/a/}');
Error('\P{Scx=  hira/a/}');
Expect(1, 127488, '\p{Scx=hira}', "");
Expect(0, 127488, '\p{^Scx=hira}', "");
Expect(0, 127488, '\P{Scx=hira}', "");
Expect(1, 127488, '\P{^Scx=hira}', "");
Expect(0, 127489, '\p{Scx=hira}', "");
Expect(1, 127489, '\p{^Scx=hira}', "");
Expect(1, 127489, '\P{Scx=hira}', "");
Expect(0, 127489, '\P{^Scx=hira}', "");
Expect(1, 127488, '\p{Scx=	-HIRA}', "");
Expect(0, 127488, '\p{^Scx=	-HIRA}', "");
Expect(0, 127488, '\P{Scx=	-HIRA}', "");
Expect(1, 127488, '\P{^Scx=	-HIRA}', "");
Expect(0, 127489, '\p{Scx=	-HIRA}', "");
Expect(1, 127489, '\p{^Scx=	-HIRA}', "");
Expect(1, 127489, '\P{Scx=	-HIRA}', "");
Expect(0, 127489, '\P{^Scx=	-HIRA}', "");
Error('\p{Is_Script_Extensions=/a/__Hiragana}');
Error('\P{Is_Script_Extensions=/a/__Hiragana}');
Expect(1, 127488, '\p{Is_Script_Extensions=hiragana}', "");
Expect(0, 127488, '\p{^Is_Script_Extensions=hiragana}', "");
Expect(0, 127488, '\P{Is_Script_Extensions=hiragana}', "");
Expect(1, 127488, '\P{^Is_Script_Extensions=hiragana}', "");
Expect(0, 127489, '\p{Is_Script_Extensions=hiragana}', "");
Expect(1, 127489, '\p{^Is_Script_Extensions=hiragana}', "");
Expect(1, 127489, '\P{Is_Script_Extensions=hiragana}', "");
Expect(0, 127489, '\P{^Is_Script_Extensions=hiragana}', "");
Expect(1, 127488, '\p{Is_Script_Extensions=		hiragana}', "");
Expect(0, 127488, '\p{^Is_Script_Extensions=		hiragana}', "");
Expect(0, 127488, '\P{Is_Script_Extensions=		hiragana}', "");
Expect(1, 127488, '\P{^Is_Script_Extensions=		hiragana}', "");
Expect(0, 127489, '\p{Is_Script_Extensions=		hiragana}', "");
Expect(1, 127489, '\p{^Is_Script_Extensions=		hiragana}', "");
Expect(1, 127489, '\P{Is_Script_Extensions=		hiragana}', "");
Expect(0, 127489, '\P{^Is_Script_Extensions=		hiragana}', "");
Error('\p{Is_Scx=:=-Hira}');
Error('\P{Is_Scx=:=-Hira}');
Expect(1, 127488, '\p{Is_Scx=hira}', "");
Expect(0, 127488, '\p{^Is_Scx=hira}', "");
Expect(0, 127488, '\P{Is_Scx=hira}', "");
Expect(1, 127488, '\P{^Is_Scx=hira}', "");
Expect(0, 127489, '\p{Is_Scx=hira}', "");
Expect(1, 127489, '\p{^Is_Scx=hira}', "");
Expect(1, 127489, '\P{Is_Scx=hira}', "");
Expect(0, 127489, '\P{^Is_Scx=hira}', "");
Expect(1, 127488, '\p{Is_Scx=-	Hira}', "");
Expect(0, 127488, '\p{^Is_Scx=-	Hira}', "");
Expect(0, 127488, '\P{Is_Scx=-	Hira}', "");
Expect(1, 127488, '\P{^Is_Scx=-	Hira}', "");
Expect(0, 127489, '\p{Is_Scx=-	Hira}', "");
Expect(1, 127489, '\p{^Is_Scx=-	Hira}', "");
Expect(1, 127489, '\P{Is_Scx=-	Hira}', "");
Expect(0, 127489, '\P{^Is_Scx=-	Hira}', "");
Error('\p{Script_Extensions=Katakana_Or_Hiragana}');
Error('\P{Script_Extensions=Katakana_Or_Hiragana}');
Error('\p{Scx: Hrkt}');
Error('\P{Scx: Hrkt}');
Error('\p{Is_Script_Extensions=Katakana_Or_Hiragana}');
Error('\P{Is_Script_Extensions=Katakana_Or_Hiragana}');
Error('\p{Is_Scx=Hrkt}');
Error('\P{Is_Scx=Hrkt}');
Error('\p{Script_Extensions=:=Old_ITALIC}');
Error('\P{Script_Extensions=:=Old_ITALIC}');
Expect(1, 66339, '\p{Script_Extensions:   olditalic}', "");
Expect(0, 66339, '\p{^Script_Extensions:   olditalic}', "");
Expect(0, 66339, '\P{Script_Extensions:   olditalic}', "");
Expect(1, 66339, '\P{^Script_Extensions:   olditalic}', "");
Expect(0, 66340, '\p{Script_Extensions:   olditalic}', "");
Expect(1, 66340, '\p{^Script_Extensions:   olditalic}', "");
Expect(1, 66340, '\P{Script_Extensions:   olditalic}', "");
Expect(0, 66340, '\P{^Script_Extensions:   olditalic}', "");
Expect(1, 66339, '\p{Script_Extensions=_ old_ITALIC}', "");
Expect(0, 66339, '\p{^Script_Extensions=_ old_ITALIC}', "");
Expect(0, 66339, '\P{Script_Extensions=_ old_ITALIC}', "");
Expect(1, 66339, '\P{^Script_Extensions=_ old_ITALIC}', "");
Expect(0, 66340, '\p{Script_Extensions=_ old_ITALIC}', "");
Expect(1, 66340, '\p{^Script_Extensions=_ old_ITALIC}', "");
Expect(1, 66340, '\P{Script_Extensions=_ old_ITALIC}', "");
Expect(0, 66340, '\P{^Script_Extensions=_ old_ITALIC}', "");
Error('\p{Scx= :=ITAL}');
Error('\P{Scx= :=ITAL}');
Expect(1, 66339, '\p{Scx=ital}', "");
Expect(0, 66339, '\p{^Scx=ital}', "");
Expect(0, 66339, '\P{Scx=ital}', "");
Expect(1, 66339, '\P{^Scx=ital}', "");
Expect(0, 66340, '\p{Scx=ital}', "");
Expect(1, 66340, '\p{^Scx=ital}', "");
Expect(1, 66340, '\P{Scx=ital}', "");
Expect(0, 66340, '\P{^Scx=ital}', "");
Expect(1, 66339, '\p{Scx=	-Ital}', "");
Expect(0, 66339, '\p{^Scx=	-Ital}', "");
Expect(0, 66339, '\P{Scx=	-Ital}', "");
Expect(1, 66339, '\P{^Scx=	-Ital}', "");
Expect(0, 66340, '\p{Scx=	-Ital}', "");
Expect(1, 66340, '\p{^Scx=	-Ital}', "");
Expect(1, 66340, '\P{Scx=	-Ital}', "");
Expect(0, 66340, '\P{^Scx=	-Ital}', "");
Error('\p{Is_Script_Extensions=/a/ 	old_ITALIC}');
Error('\P{Is_Script_Extensions=/a/ 	old_ITALIC}');
Expect(1, 66339, '\p{Is_Script_Extensions=olditalic}', "");
Expect(0, 66339, '\p{^Is_Script_Extensions=olditalic}', "");
Expect(0, 66339, '\P{Is_Script_Extensions=olditalic}', "");
Expect(1, 66339, '\P{^Is_Script_Extensions=olditalic}', "");
Expect(0, 66340, '\p{Is_Script_Extensions=olditalic}', "");
Expect(1, 66340, '\p{^Is_Script_Extensions=olditalic}', "");
Expect(1, 66340, '\P{Is_Script_Extensions=olditalic}', "");
Expect(0, 66340, '\P{^Is_Script_Extensions=olditalic}', "");
Expect(1, 66339, '\p{Is_Script_Extensions:	--Old_ITALIC}', "");
Expect(0, 66339, '\p{^Is_Script_Extensions:	--Old_ITALIC}', "");
Expect(0, 66339, '\P{Is_Script_Extensions:	--Old_ITALIC}', "");
Expect(1, 66339, '\P{^Is_Script_Extensions:	--Old_ITALIC}', "");
Expect(0, 66340, '\p{Is_Script_Extensions:	--Old_ITALIC}', "");
Expect(1, 66340, '\p{^Is_Script_Extensions:	--Old_ITALIC}', "");
Expect(1, 66340, '\P{Is_Script_Extensions:	--Old_ITALIC}', "");
Expect(0, 66340, '\P{^Is_Script_Extensions:	--Old_ITALIC}', "");
Error('\p{Is_Scx=/a/ _ITAL}');
Error('\P{Is_Scx=/a/ _ITAL}');
Expect(1, 66339, '\p{Is_Scx=ital}', "");
Expect(0, 66339, '\p{^Is_Scx=ital}', "");
Expect(0, 66339, '\P{Is_Scx=ital}', "");
Expect(1, 66339, '\P{^Is_Scx=ital}', "");
Expect(0, 66340, '\p{Is_Scx=ital}', "");
Expect(1, 66340, '\p{^Is_Scx=ital}', "");
Expect(1, 66340, '\P{Is_Scx=ital}', "");
Expect(0, 66340, '\P{^Is_Scx=ital}', "");
Expect(1, 66339, '\p{Is_Scx= 	ital}', "");
Expect(0, 66339, '\p{^Is_Scx= 	ital}', "");
Expect(0, 66339, '\P{Is_Scx= 	ital}', "");
Expect(1, 66339, '\P{^Is_Scx= 	ital}', "");
Expect(0, 66340, '\p{Is_Scx= 	ital}', "");
Expect(1, 66340, '\p{^Is_Scx= 	ital}', "");
Expect(1, 66340, '\P{Is_Scx= 	ital}', "");
Expect(0, 66340, '\P{^Is_Scx= 	ital}', "");
Error('\p{Script_Extensions=	:=JAVANESE}');
Error('\P{Script_Extensions=	:=JAVANESE}');
Expect(1, 43487, '\p{Script_Extensions=javanese}', "");
Expect(0, 43487, '\p{^Script_Extensions=javanese}', "");
Expect(0, 43487, '\P{Script_Extensions=javanese}', "");
Expect(1, 43487, '\P{^Script_Extensions=javanese}', "");
Expect(0, 43488, '\p{Script_Extensions=javanese}', "");
Expect(1, 43488, '\p{^Script_Extensions=javanese}', "");
Expect(1, 43488, '\P{Script_Extensions=javanese}', "");
Expect(0, 43488, '\P{^Script_Extensions=javanese}', "");
Expect(1, 43487, '\p{Script_Extensions:	 JAVANESE}', "");
Expect(0, 43487, '\p{^Script_Extensions:	 JAVANESE}', "");
Expect(0, 43487, '\P{Script_Extensions:	 JAVANESE}', "");
Expect(1, 43487, '\P{^Script_Extensions:	 JAVANESE}', "");
Expect(0, 43488, '\p{Script_Extensions:	 JAVANESE}', "");
Expect(1, 43488, '\p{^Script_Extensions:	 JAVANESE}', "");
Expect(1, 43488, '\P{Script_Extensions:	 JAVANESE}', "");
Expect(0, 43488, '\P{^Script_Extensions:	 JAVANESE}', "");
Error('\p{Scx=_/a/JAVA}');
Error('\P{Scx=_/a/JAVA}');
Expect(1, 43487, '\p{Scx=java}', "");
Expect(0, 43487, '\p{^Scx=java}', "");
Expect(0, 43487, '\P{Scx=java}', "");
Expect(1, 43487, '\P{^Scx=java}', "");
Expect(0, 43488, '\p{Scx=java}', "");
Expect(1, 43488, '\p{^Scx=java}', "");
Expect(1, 43488, '\P{Scx=java}', "");
Expect(0, 43488, '\P{^Scx=java}', "");
Expect(1, 43487, '\p{Scx=-Java}', "");
Expect(0, 43487, '\p{^Scx=-Java}', "");
Expect(0, 43487, '\P{Scx=-Java}', "");
Expect(1, 43487, '\P{^Scx=-Java}', "");
Expect(0, 43488, '\p{Scx=-Java}', "");
Expect(1, 43488, '\p{^Scx=-Java}', "");
Expect(1, 43488, '\P{Scx=-Java}', "");
Expect(0, 43488, '\P{^Scx=-Java}', "");
Error('\p{Is_Script_Extensions=/a/-	Javanese}');
Error('\P{Is_Script_Extensions=/a/-	Javanese}');
Expect(1, 43487, '\p{Is_Script_Extensions=javanese}', "");
Expect(0, 43487, '\p{^Is_Script_Extensions=javanese}', "");
Expect(0, 43487, '\P{Is_Script_Extensions=javanese}', "");
Expect(1, 43487, '\P{^Is_Script_Extensions=javanese}', "");
Expect(0, 43488, '\p{Is_Script_Extensions=javanese}', "");
Expect(1, 43488, '\p{^Is_Script_Extensions=javanese}', "");
Expect(1, 43488, '\P{Is_Script_Extensions=javanese}', "");
Expect(0, 43488, '\P{^Is_Script_Extensions=javanese}', "");
Expect(1, 43487, '\p{Is_Script_Extensions=  javanese}', "");
Expect(0, 43487, '\p{^Is_Script_Extensions=  javanese}', "");
Expect(0, 43487, '\P{Is_Script_Extensions=  javanese}', "");
Expect(1, 43487, '\P{^Is_Script_Extensions=  javanese}', "");
Expect(0, 43488, '\p{Is_Script_Extensions=  javanese}', "");
Expect(1, 43488, '\p{^Is_Script_Extensions=  javanese}', "");
Expect(1, 43488, '\P{Is_Script_Extensions=  javanese}', "");
Expect(0, 43488, '\P{^Is_Script_Extensions=  javanese}', "");
Error('\p{Is_Scx=/a/ _java}');
Error('\P{Is_Scx=/a/ _java}');
Expect(1, 43487, '\p{Is_Scx: java}', "");
Expect(0, 43487, '\p{^Is_Scx: java}', "");
Expect(0, 43487, '\P{Is_Scx: java}', "");
Expect(1, 43487, '\P{^Is_Scx: java}', "");
Expect(0, 43488, '\p{Is_Scx: java}', "");
Expect(1, 43488, '\p{^Is_Scx: java}', "");
Expect(1, 43488, '\P{Is_Scx: java}', "");
Expect(0, 43488, '\P{^Is_Scx: java}', "");
Expect(1, 43487, '\p{Is_Scx= java}', "");
Expect(0, 43487, '\p{^Is_Scx= java}', "");
Expect(0, 43487, '\P{Is_Scx= java}', "");
Expect(1, 43487, '\P{^Is_Scx= java}', "");
Expect(0, 43488, '\p{Is_Scx= java}', "");
Expect(1, 43488, '\p{^Is_Scx= java}', "");
Expect(1, 43488, '\P{Is_Scx= java}', "");
Expect(0, 43488, '\P{^Is_Scx= java}', "");
Error('\p{Script_Extensions=-_kayah_li:=}');
Error('\P{Script_Extensions=-_kayah_li:=}');
Expect(1, 43311, '\p{Script_Extensions=kayahli}', "");
Expect(0, 43311, '\p{^Script_Extensions=kayahli}', "");
Expect(0, 43311, '\P{Script_Extensions=kayahli}', "");
Expect(1, 43311, '\P{^Script_Extensions=kayahli}', "");
Expect(0, 43312, '\p{Script_Extensions=kayahli}', "");
Expect(1, 43312, '\p{^Script_Extensions=kayahli}', "");
Expect(1, 43312, '\P{Script_Extensions=kayahli}', "");
Expect(0, 43312, '\P{^Script_Extensions=kayahli}', "");
Expect(1, 43311, '\p{Script_Extensions=	Kayah_li}', "");
Expect(0, 43311, '\p{^Script_Extensions=	Kayah_li}', "");
Expect(0, 43311, '\P{Script_Extensions=	Kayah_li}', "");
Expect(1, 43311, '\P{^Script_Extensions=	Kayah_li}', "");
Expect(0, 43312, '\p{Script_Extensions=	Kayah_li}', "");
Expect(1, 43312, '\p{^Script_Extensions=	Kayah_li}', "");
Expect(1, 43312, '\P{Script_Extensions=	Kayah_li}', "");
Expect(0, 43312, '\P{^Script_Extensions=	Kayah_li}', "");
Error('\p{Scx: /a/	-Kali}');
Error('\P{Scx: /a/	-Kali}');
Expect(1, 43311, '\p{Scx=kali}', "");
Expect(0, 43311, '\p{^Scx=kali}', "");
Expect(0, 43311, '\P{Scx=kali}', "");
Expect(1, 43311, '\P{^Scx=kali}', "");
Expect(0, 43312, '\p{Scx=kali}', "");
Expect(1, 43312, '\p{^Scx=kali}', "");
Expect(1, 43312, '\P{Scx=kali}', "");
Expect(0, 43312, '\P{^Scx=kali}', "");
Expect(1, 43311, '\p{Scx=_	Kali}', "");
Expect(0, 43311, '\p{^Scx=_	Kali}', "");
Expect(0, 43311, '\P{Scx=_	Kali}', "");
Expect(1, 43311, '\P{^Scx=_	Kali}', "");
Expect(0, 43312, '\p{Scx=_	Kali}', "");
Expect(1, 43312, '\p{^Scx=_	Kali}', "");
Expect(1, 43312, '\P{Scx=_	Kali}', "");
Expect(0, 43312, '\P{^Scx=_	Kali}', "");
Error('\p{Is_Script_Extensions=_ KAYAH_Li:=}');
Error('\P{Is_Script_Extensions=_ KAYAH_Li:=}');
Expect(1, 43311, '\p{Is_Script_Extensions=kayahli}', "");
Expect(0, 43311, '\p{^Is_Script_Extensions=kayahli}', "");
Expect(0, 43311, '\P{Is_Script_Extensions=kayahli}', "");
Expect(1, 43311, '\P{^Is_Script_Extensions=kayahli}', "");
Expect(0, 43312, '\p{Is_Script_Extensions=kayahli}', "");
Expect(1, 43312, '\p{^Is_Script_Extensions=kayahli}', "");
Expect(1, 43312, '\P{Is_Script_Extensions=kayahli}', "");
Expect(0, 43312, '\P{^Is_Script_Extensions=kayahli}', "");
Expect(1, 43311, '\p{Is_Script_Extensions= 	Kayah_Li}', "");
Expect(0, 43311, '\p{^Is_Script_Extensions= 	Kayah_Li}', "");
Expect(0, 43311, '\P{Is_Script_Extensions= 	Kayah_Li}', "");
Expect(1, 43311, '\P{^Is_Script_Extensions= 	Kayah_Li}', "");
Expect(0, 43312, '\p{Is_Script_Extensions= 	Kayah_Li}', "");
Expect(1, 43312, '\p{^Is_Script_Extensions= 	Kayah_Li}', "");
Expect(1, 43312, '\P{Is_Script_Extensions= 	Kayah_Li}', "");
Expect(0, 43312, '\P{^Is_Script_Extensions= 	Kayah_Li}', "");
Error('\p{Is_Scx=-_kali/a/}');
Error('\P{Is_Scx=-_kali/a/}');
Expect(1, 43311, '\p{Is_Scx=kali}', "");
Expect(0, 43311, '\p{^Is_Scx=kali}', "");
Expect(0, 43311, '\P{Is_Scx=kali}', "");
Expect(1, 43311, '\P{^Is_Scx=kali}', "");
Expect(0, 43312, '\p{Is_Scx=kali}', "");
Expect(1, 43312, '\p{^Is_Scx=kali}', "");
Expect(1, 43312, '\P{Is_Scx=kali}', "");
Expect(0, 43312, '\P{^Is_Scx=kali}', "");
Expect(1, 43311, '\p{Is_Scx=_-KALI}', "");
Expect(0, 43311, '\p{^Is_Scx=_-KALI}', "");
Expect(0, 43311, '\P{Is_Scx=_-KALI}', "");
Expect(1, 43311, '\P{^Is_Scx=_-KALI}', "");
Expect(0, 43312, '\p{Is_Scx=_-KALI}', "");
Expect(1, 43312, '\p{^Is_Scx=_-KALI}', "");
Expect(1, 43312, '\P{Is_Scx=_-KALI}', "");
Expect(0, 43312, '\P{^Is_Scx=_-KALI}', "");
Error('\p{Script_Extensions=/a/ -KATAKANA}');
Error('\P{Script_Extensions=/a/ -KATAKANA}');
Expect(1, 110592, '\p{Script_Extensions=katakana}', "");
Expect(0, 110592, '\p{^Script_Extensions=katakana}', "");
Expect(0, 110592, '\P{Script_Extensions=katakana}', "");
Expect(1, 110592, '\P{^Script_Extensions=katakana}', "");
Expect(0, 110593, '\p{Script_Extensions=katakana}', "");
Expect(1, 110593, '\p{^Script_Extensions=katakana}', "");
Expect(1, 110593, '\P{Script_Extensions=katakana}', "");
Expect(0, 110593, '\P{^Script_Extensions=katakana}', "");
Expect(1, 110592, '\p{Script_Extensions:   _	KATAKANA}', "");
Expect(0, 110592, '\p{^Script_Extensions:   _	KATAKANA}', "");
Expect(0, 110592, '\P{Script_Extensions:   _	KATAKANA}', "");
Expect(1, 110592, '\P{^Script_Extensions:   _	KATAKANA}', "");
Expect(0, 110593, '\p{Script_Extensions:   _	KATAKANA}', "");
Expect(1, 110593, '\p{^Script_Extensions:   _	KATAKANA}', "");
Expect(1, 110593, '\P{Script_Extensions:   _	KATAKANA}', "");
Expect(0, 110593, '\P{^Script_Extensions:   _	KATAKANA}', "");
Error('\p{Scx=	-kana/a/}');
Error('\P{Scx=	-kana/a/}');
Expect(1, 110592, '\p{Scx=kana}', "");
Expect(0, 110592, '\p{^Scx=kana}', "");
Expect(0, 110592, '\P{Scx=kana}', "");
Expect(1, 110592, '\P{^Scx=kana}', "");
Expect(0, 110593, '\p{Scx=kana}', "");
Expect(1, 110593, '\p{^Scx=kana}', "");
Expect(1, 110593, '\P{Scx=kana}', "");
Expect(0, 110593, '\P{^Scx=kana}', "");
Expect(1, 110592, '\p{Scx=--Kana}', "");
Expect(0, 110592, '\p{^Scx=--Kana}', "");
Expect(0, 110592, '\P{Scx=--Kana}', "");
Expect(1, 110592, '\P{^Scx=--Kana}', "");
Expect(0, 110593, '\p{Scx=--Kana}', "");
Expect(1, 110593, '\p{^Scx=--Kana}', "");
Expect(1, 110593, '\P{Scx=--Kana}', "");
Expect(0, 110593, '\P{^Scx=--Kana}', "");
Error('\p{Is_Script_Extensions=	/a/Katakana}');
Error('\P{Is_Script_Extensions=	/a/Katakana}');
Expect(1, 110592, '\p{Is_Script_Extensions=katakana}', "");
Expect(0, 110592, '\p{^Is_Script_Extensions=katakana}', "");
Expect(0, 110592, '\P{Is_Script_Extensions=katakana}', "");
Expect(1, 110592, '\P{^Is_Script_Extensions=katakana}', "");
Expect(0, 110593, '\p{Is_Script_Extensions=katakana}', "");
Expect(1, 110593, '\p{^Is_Script_Extensions=katakana}', "");
Expect(1, 110593, '\P{Is_Script_Extensions=katakana}', "");
Expect(0, 110593, '\P{^Is_Script_Extensions=katakana}', "");
Expect(1, 110592, '\p{Is_Script_Extensions=_-Katakana}', "");
Expect(0, 110592, '\p{^Is_Script_Extensions=_-Katakana}', "");
Expect(0, 110592, '\P{Is_Script_Extensions=_-Katakana}', "");
Expect(1, 110592, '\P{^Is_Script_Extensions=_-Katakana}', "");
Expect(0, 110593, '\p{Is_Script_Extensions=_-Katakana}', "");
Expect(1, 110593, '\p{^Is_Script_Extensions=_-Katakana}', "");
Expect(1, 110593, '\P{Is_Script_Extensions=_-Katakana}', "");
Expect(0, 110593, '\P{^Is_Script_Extensions=_-Katakana}', "");
Error('\p{Is_Scx= Kana/a/}');
Error('\P{Is_Scx= Kana/a/}');
Expect(1, 110592, '\p{Is_Scx=kana}', "");
Expect(0, 110592, '\p{^Is_Scx=kana}', "");
Expect(0, 110592, '\P{Is_Scx=kana}', "");
Expect(1, 110592, '\P{^Is_Scx=kana}', "");
Expect(0, 110593, '\p{Is_Scx=kana}', "");
Expect(1, 110593, '\p{^Is_Scx=kana}', "");
Expect(1, 110593, '\P{Is_Scx=kana}', "");
Expect(0, 110593, '\P{^Is_Scx=kana}', "");
Expect(1, 110592, '\p{Is_Scx=__Kana}', "");
Expect(0, 110592, '\p{^Is_Scx=__Kana}', "");
Expect(0, 110592, '\P{Is_Scx=__Kana}', "");
Expect(1, 110592, '\P{^Is_Scx=__Kana}', "");
Expect(0, 110593, '\p{Is_Scx=__Kana}', "");
Expect(1, 110593, '\p{^Is_Scx=__Kana}', "");
Expect(1, 110593, '\P{Is_Scx=__Kana}', "");
Expect(0, 110593, '\P{^Is_Scx=__Kana}', "");
Error('\p{Script_Extensions=/a/kharoshthi}');
Error('\P{Script_Extensions=/a/kharoshthi}');
Expect(1, 68184, '\p{Script_Extensions=kharoshthi}', "");
Expect(0, 68184, '\p{^Script_Extensions=kharoshthi}', "");
Expect(0, 68184, '\P{Script_Extensions=kharoshthi}', "");
Expect(1, 68184, '\P{^Script_Extensions=kharoshthi}', "");
Expect(0, 68185, '\p{Script_Extensions=kharoshthi}', "");
Expect(1, 68185, '\p{^Script_Extensions=kharoshthi}', "");
Expect(1, 68185, '\P{Script_Extensions=kharoshthi}', "");
Expect(0, 68185, '\P{^Script_Extensions=kharoshthi}', "");
Expect(1, 68184, '\p{Script_Extensions=Kharoshthi}', "");
Expect(0, 68184, '\p{^Script_Extensions=Kharoshthi}', "");
Expect(0, 68184, '\P{Script_Extensions=Kharoshthi}', "");
Expect(1, 68184, '\P{^Script_Extensions=Kharoshthi}', "");
Expect(0, 68185, '\p{Script_Extensions=Kharoshthi}', "");
Expect(1, 68185, '\p{^Script_Extensions=Kharoshthi}', "");
Expect(1, 68185, '\P{Script_Extensions=Kharoshthi}', "");
Expect(0, 68185, '\P{^Script_Extensions=Kharoshthi}', "");
Error('\p{Scx=_Khar:=}');
Error('\P{Scx=_Khar:=}');
Expect(1, 68184, '\p{Scx=khar}', "");
Expect(0, 68184, '\p{^Scx=khar}', "");
Expect(0, 68184, '\P{Scx=khar}', "");
Expect(1, 68184, '\P{^Scx=khar}', "");
Expect(0, 68185, '\p{Scx=khar}', "");
Expect(1, 68185, '\p{^Scx=khar}', "");
Expect(1, 68185, '\P{Scx=khar}', "");
Expect(0, 68185, '\P{^Scx=khar}', "");
Expect(1, 68184, '\p{Scx:   	khar}', "");
Expect(0, 68184, '\p{^Scx:   	khar}', "");
Expect(0, 68184, '\P{Scx:   	khar}', "");
Expect(1, 68184, '\P{^Scx:   	khar}', "");
Expect(0, 68185, '\p{Scx:   	khar}', "");
Expect(1, 68185, '\p{^Scx:   	khar}', "");
Expect(1, 68185, '\P{Scx:   	khar}', "");
Expect(0, 68185, '\P{^Scx:   	khar}', "");
Error('\p{Is_Script_Extensions=	/a/Kharoshthi}');
Error('\P{Is_Script_Extensions=	/a/Kharoshthi}');
Expect(1, 68184, '\p{Is_Script_Extensions=kharoshthi}', "");
Expect(0, 68184, '\p{^Is_Script_Extensions=kharoshthi}', "");
Expect(0, 68184, '\P{Is_Script_Extensions=kharoshthi}', "");
Expect(1, 68184, '\P{^Is_Script_Extensions=kharoshthi}', "");
Expect(0, 68185, '\p{Is_Script_Extensions=kharoshthi}', "");
Expect(1, 68185, '\p{^Is_Script_Extensions=kharoshthi}', "");
Expect(1, 68185, '\P{Is_Script_Extensions=kharoshthi}', "");
Expect(0, 68185, '\P{^Is_Script_Extensions=kharoshthi}', "");
Expect(1, 68184, '\p{Is_Script_Extensions:	 _Kharoshthi}', "");
Expect(0, 68184, '\p{^Is_Script_Extensions:	 _Kharoshthi}', "");
Expect(0, 68184, '\P{Is_Script_Extensions:	 _Kharoshthi}', "");
Expect(1, 68184, '\P{^Is_Script_Extensions:	 _Kharoshthi}', "");
Expect(0, 68185, '\p{Is_Script_Extensions:	 _Kharoshthi}', "");
Expect(1, 68185, '\p{^Is_Script_Extensions:	 _Kharoshthi}', "");
Expect(1, 68185, '\P{Is_Script_Extensions:	 _Kharoshthi}', "");
Expect(0, 68185, '\P{^Is_Script_Extensions:	 _Kharoshthi}', "");
Error('\p{Is_Scx=/a/	_Khar}');
Error('\P{Is_Scx=/a/	_Khar}');
Expect(1, 68184, '\p{Is_Scx=khar}', "");
Expect(0, 68184, '\p{^Is_Scx=khar}', "");
Expect(0, 68184, '\P{Is_Scx=khar}', "");
Expect(1, 68184, '\P{^Is_Scx=khar}', "");
Expect(0, 68185, '\p{Is_Scx=khar}', "");
Expect(1, 68185, '\p{^Is_Scx=khar}', "");
Expect(1, 68185, '\P{Is_Scx=khar}', "");
Expect(0, 68185, '\P{^Is_Scx=khar}', "");
Expect(1, 68184, '\p{Is_Scx=-	Khar}', "");
Expect(0, 68184, '\p{^Is_Scx=-	Khar}', "");
Expect(0, 68184, '\P{Is_Scx=-	Khar}', "");
Expect(1, 68184, '\P{^Is_Scx=-	Khar}', "");
Expect(0, 68185, '\p{Is_Scx=-	Khar}', "");
Expect(1, 68185, '\p{^Is_Scx=-	Khar}', "");
Expect(1, 68185, '\P{Is_Scx=-	Khar}', "");
Expect(0, 68185, '\P{^Is_Scx=-	Khar}', "");
Error('\p{Script_Extensions=:=KHMER}');
Error('\P{Script_Extensions=:=KHMER}');
Expect(1, 6655, '\p{Script_Extensions=khmer}', "");
Expect(0, 6655, '\p{^Script_Extensions=khmer}', "");
Expect(0, 6655, '\P{Script_Extensions=khmer}', "");
Expect(1, 6655, '\P{^Script_Extensions=khmer}', "");
Expect(0, 6656, '\p{Script_Extensions=khmer}', "");
Expect(1, 6656, '\p{^Script_Extensions=khmer}', "");
Expect(1, 6656, '\P{Script_Extensions=khmer}', "");
Expect(0, 6656, '\P{^Script_Extensions=khmer}', "");
Expect(1, 6655, '\p{Script_Extensions=	-Khmer}', "");
Expect(0, 6655, '\p{^Script_Extensions=	-Khmer}', "");
Expect(0, 6655, '\P{Script_Extensions=	-Khmer}', "");
Expect(1, 6655, '\P{^Script_Extensions=	-Khmer}', "");
Expect(0, 6656, '\p{Script_Extensions=	-Khmer}', "");
Expect(1, 6656, '\p{^Script_Extensions=	-Khmer}', "");
Expect(1, 6656, '\P{Script_Extensions=	-Khmer}', "");
Expect(0, 6656, '\P{^Script_Extensions=	-Khmer}', "");
Error('\p{Scx=-_KHMR/a/}');
Error('\P{Scx=-_KHMR/a/}');
Expect(1, 6655, '\p{Scx=khmr}', "");
Expect(0, 6655, '\p{^Scx=khmr}', "");
Expect(0, 6655, '\P{Scx=khmr}', "");
Expect(1, 6655, '\P{^Scx=khmr}', "");
Expect(0, 6656, '\p{Scx=khmr}', "");
Expect(1, 6656, '\p{^Scx=khmr}', "");
Expect(1, 6656, '\P{Scx=khmr}', "");
Expect(0, 6656, '\P{^Scx=khmr}', "");
Expect(1, 6655, '\p{Scx:   	Khmr}', "");
Expect(0, 6655, '\p{^Scx:   	Khmr}', "");
Expect(0, 6655, '\P{Scx:   	Khmr}', "");
Expect(1, 6655, '\P{^Scx:   	Khmr}', "");
Expect(0, 6656, '\p{Scx:   	Khmr}', "");
Expect(1, 6656, '\p{^Scx:   	Khmr}', "");
Expect(1, 6656, '\P{Scx:   	Khmr}', "");
Expect(0, 6656, '\P{^Scx:   	Khmr}', "");
Error('\p{Is_Script_Extensions=:=_-Khmer}');
Error('\P{Is_Script_Extensions=:=_-Khmer}');
Expect(1, 6655, '\p{Is_Script_Extensions=khmer}', "");
Expect(0, 6655, '\p{^Is_Script_Extensions=khmer}', "");
Expect(0, 6655, '\P{Is_Script_Extensions=khmer}', "");
Expect(1, 6655, '\P{^Is_Script_Extensions=khmer}', "");
Expect(0, 6656, '\p{Is_Script_Extensions=khmer}', "");
Expect(1, 6656, '\p{^Is_Script_Extensions=khmer}', "");
Expect(1, 6656, '\P{Is_Script_Extensions=khmer}', "");
Expect(0, 6656, '\P{^Is_Script_Extensions=khmer}', "");
Expect(1, 6655, '\p{Is_Script_Extensions: 		KHMER}', "");
Expect(0, 6655, '\p{^Is_Script_Extensions: 		KHMER}', "");
Expect(0, 6655, '\P{Is_Script_Extensions: 		KHMER}', "");
Expect(1, 6655, '\P{^Is_Script_Extensions: 		KHMER}', "");
Expect(0, 6656, '\p{Is_Script_Extensions: 		KHMER}', "");
Expect(1, 6656, '\p{^Is_Script_Extensions: 		KHMER}', "");
Expect(1, 6656, '\P{Is_Script_Extensions: 		KHMER}', "");
Expect(0, 6656, '\P{^Is_Script_Extensions: 		KHMER}', "");
Error('\p{Is_Scx=_/a/Khmr}');
Error('\P{Is_Scx=_/a/Khmr}');
Expect(1, 6655, '\p{Is_Scx=khmr}', "");
Expect(0, 6655, '\p{^Is_Scx=khmr}', "");
Expect(0, 6655, '\P{Is_Scx=khmr}', "");
Expect(1, 6655, '\P{^Is_Scx=khmr}', "");
Expect(0, 6656, '\p{Is_Scx=khmr}', "");
Expect(1, 6656, '\p{^Is_Scx=khmr}', "");
Expect(1, 6656, '\P{Is_Scx=khmr}', "");
Expect(0, 6656, '\P{^Is_Scx=khmr}', "");
Expect(1, 6655, '\p{Is_Scx=	_KHMR}', "");
Expect(0, 6655, '\p{^Is_Scx=	_KHMR}', "");
Expect(0, 6655, '\P{Is_Scx=	_KHMR}', "");
Expect(1, 6655, '\P{^Is_Scx=	_KHMR}', "");
Expect(0, 6656, '\p{Is_Scx=	_KHMR}', "");
Expect(1, 6656, '\p{^Is_Scx=	_KHMR}', "");
Expect(1, 6656, '\P{Is_Scx=	_KHMR}', "");
Expect(0, 6656, '\P{^Is_Scx=	_KHMR}', "");
Error('\p{Script_Extensions=:=_	Kannada}');
Error('\P{Script_Extensions=:=_	Kannada}');
Expect(1, 3314, '\p{Script_Extensions=kannada}', "");
Expect(0, 3314, '\p{^Script_Extensions=kannada}', "");
Expect(0, 3314, '\P{Script_Extensions=kannada}', "");
Expect(1, 3314, '\P{^Script_Extensions=kannada}', "");
Expect(0, 3315, '\p{Script_Extensions=kannada}', "");
Expect(1, 3315, '\p{^Script_Extensions=kannada}', "");
Expect(1, 3315, '\P{Script_Extensions=kannada}', "");
Expect(0, 3315, '\P{^Script_Extensions=kannada}', "");
Expect(1, 3314, '\p{Script_Extensions=	-kannada}', "");
Expect(0, 3314, '\p{^Script_Extensions=	-kannada}', "");
Expect(0, 3314, '\P{Script_Extensions=	-kannada}', "");
Expect(1, 3314, '\P{^Script_Extensions=	-kannada}', "");
Expect(0, 3315, '\p{Script_Extensions=	-kannada}', "");
Expect(1, 3315, '\p{^Script_Extensions=	-kannada}', "");
Expect(1, 3315, '\P{Script_Extensions=	-kannada}', "");
Expect(0, 3315, '\P{^Script_Extensions=	-kannada}', "");
Error('\p{Scx=/a/-Knda}');
Error('\P{Scx=/a/-Knda}');
Expect(1, 3314, '\p{Scx=knda}', "");
Expect(0, 3314, '\p{^Scx=knda}', "");
Expect(0, 3314, '\P{Scx=knda}', "");
Expect(1, 3314, '\P{^Scx=knda}', "");
Expect(0, 3315, '\p{Scx=knda}', "");
Expect(1, 3315, '\p{^Scx=knda}', "");
Expect(1, 3315, '\P{Scx=knda}', "");
Expect(0, 3315, '\P{^Scx=knda}', "");
Expect(1, 3314, '\p{Scx= KNDA}', "");
Expect(0, 3314, '\p{^Scx= KNDA}', "");
Expect(0, 3314, '\P{Scx= KNDA}', "");
Expect(1, 3314, '\P{^Scx= KNDA}', "");
Expect(0, 3315, '\p{Scx= KNDA}', "");
Expect(1, 3315, '\p{^Scx= KNDA}', "");
Expect(1, 3315, '\P{Scx= KNDA}', "");
Expect(0, 3315, '\P{^Scx= KNDA}', "");
Error('\p{Is_Script_Extensions=	:=kannada}');
Error('\P{Is_Script_Extensions=	:=kannada}');
Expect(1, 3314, '\p{Is_Script_Extensions=kannada}', "");
Expect(0, 3314, '\p{^Is_Script_Extensions=kannada}', "");
Expect(0, 3314, '\P{Is_Script_Extensions=kannada}', "");
Expect(1, 3314, '\P{^Is_Script_Extensions=kannada}', "");
Expect(0, 3315, '\p{Is_Script_Extensions=kannada}', "");
Expect(1, 3315, '\p{^Is_Script_Extensions=kannada}', "");
Expect(1, 3315, '\P{Is_Script_Extensions=kannada}', "");
Expect(0, 3315, '\P{^Is_Script_Extensions=kannada}', "");
Expect(1, 3314, '\p{Is_Script_Extensions=_Kannada}', "");
Expect(0, 3314, '\p{^Is_Script_Extensions=_Kannada}', "");
Expect(0, 3314, '\P{Is_Script_Extensions=_Kannada}', "");
Expect(1, 3314, '\P{^Is_Script_Extensions=_Kannada}', "");
Expect(0, 3315, '\p{Is_Script_Extensions=_Kannada}', "");
Expect(1, 3315, '\p{^Is_Script_Extensions=_Kannada}', "");
Expect(1, 3315, '\P{Is_Script_Extensions=_Kannada}', "");
Expect(0, 3315, '\P{^Is_Script_Extensions=_Kannada}', "");
Error('\p{Is_Scx=-:=knda}');
Error('\P{Is_Scx=-:=knda}');
Expect(1, 3314, '\p{Is_Scx=knda}', "");
Expect(0, 3314, '\p{^Is_Scx=knda}', "");
Expect(0, 3314, '\P{Is_Scx=knda}', "");
Expect(1, 3314, '\P{^Is_Scx=knda}', "");
Expect(0, 3315, '\p{Is_Scx=knda}', "");
Expect(1, 3315, '\p{^Is_Scx=knda}', "");
Expect(1, 3315, '\P{Is_Scx=knda}', "");
Expect(0, 3315, '\P{^Is_Scx=knda}', "");
Expect(1, 3314, '\p{Is_Scx=	knda}', "");
Expect(0, 3314, '\p{^Is_Scx=	knda}', "");
Expect(0, 3314, '\P{Is_Scx=	knda}', "");
Expect(1, 3314, '\P{^Is_Scx=	knda}', "");
Expect(0, 3315, '\p{Is_Scx=	knda}', "");
Expect(1, 3315, '\p{^Is_Scx=	knda}', "");
Expect(1, 3315, '\P{Is_Scx=	knda}', "");
Expect(0, 3315, '\P{^Is_Scx=	knda}', "");
Error('\p{Script_Extensions:	Kaithi:=}');
Error('\P{Script_Extensions:	Kaithi:=}');
Expect(1, 69825, '\p{Script_Extensions=kaithi}', "");
Expect(0, 69825, '\p{^Script_Extensions=kaithi}', "");
Expect(0, 69825, '\P{Script_Extensions=kaithi}', "");
Expect(1, 69825, '\P{^Script_Extensions=kaithi}', "");
Expect(0, 69826, '\p{Script_Extensions=kaithi}', "");
Expect(1, 69826, '\p{^Script_Extensions=kaithi}', "");
Expect(1, 69826, '\P{Script_Extensions=kaithi}', "");
Expect(0, 69826, '\P{^Script_Extensions=kaithi}', "");
Expect(1, 69825, '\p{Script_Extensions=-kaithi}', "");
Expect(0, 69825, '\p{^Script_Extensions=-kaithi}', "");
Expect(0, 69825, '\P{Script_Extensions=-kaithi}', "");
Expect(1, 69825, '\P{^Script_Extensions=-kaithi}', "");
Expect(0, 69826, '\p{Script_Extensions=-kaithi}', "");
Expect(1, 69826, '\p{^Script_Extensions=-kaithi}', "");
Expect(1, 69826, '\P{Script_Extensions=-kaithi}', "");
Expect(0, 69826, '\P{^Script_Extensions=-kaithi}', "");
Error('\p{Scx=	/a/Kthi}');
Error('\P{Scx=	/a/Kthi}');
Expect(1, 69825, '\p{Scx: kthi}', "");
Expect(0, 69825, '\p{^Scx: kthi}', "");
Expect(0, 69825, '\P{Scx: kthi}', "");
Expect(1, 69825, '\P{^Scx: kthi}', "");
Expect(0, 69826, '\p{Scx: kthi}', "");
Expect(1, 69826, '\p{^Scx: kthi}', "");
Expect(1, 69826, '\P{Scx: kthi}', "");
Expect(0, 69826, '\P{^Scx: kthi}', "");
Expect(1, 69825, '\p{Scx=_	Kthi}', "");
Expect(0, 69825, '\p{^Scx=_	Kthi}', "");
Expect(0, 69825, '\P{Scx=_	Kthi}', "");
Expect(1, 69825, '\P{^Scx=_	Kthi}', "");
Expect(0, 69826, '\p{Scx=_	Kthi}', "");
Expect(1, 69826, '\p{^Scx=_	Kthi}', "");
Expect(1, 69826, '\P{Scx=_	Kthi}', "");
Expect(0, 69826, '\P{^Scx=_	Kthi}', "");
Error('\p{Is_Script_Extensions= :=KAITHI}');
Error('\P{Is_Script_Extensions= :=KAITHI}');
Expect(1, 69825, '\p{Is_Script_Extensions=kaithi}', "");
Expect(0, 69825, '\p{^Is_Script_Extensions=kaithi}', "");
Expect(0, 69825, '\P{Is_Script_Extensions=kaithi}', "");
Expect(1, 69825, '\P{^Is_Script_Extensions=kaithi}', "");
Expect(0, 69826, '\p{Is_Script_Extensions=kaithi}', "");
Expect(1, 69826, '\p{^Is_Script_Extensions=kaithi}', "");
Expect(1, 69826, '\P{Is_Script_Extensions=kaithi}', "");
Expect(0, 69826, '\P{^Is_Script_Extensions=kaithi}', "");
Expect(1, 69825, '\p{Is_Script_Extensions=	kaithi}', "");
Expect(0, 69825, '\p{^Is_Script_Extensions=	kaithi}', "");
Expect(0, 69825, '\P{Is_Script_Extensions=	kaithi}', "");
Expect(1, 69825, '\P{^Is_Script_Extensions=	kaithi}', "");
Expect(0, 69826, '\p{Is_Script_Extensions=	kaithi}', "");
Expect(1, 69826, '\p{^Is_Script_Extensions=	kaithi}', "");
Expect(1, 69826, '\P{Is_Script_Extensions=	kaithi}', "");
Expect(0, 69826, '\P{^Is_Script_Extensions=	kaithi}', "");
Error('\p{Is_Scx=__kthi/a/}');
Error('\P{Is_Scx=__kthi/a/}');
Expect(1, 69825, '\p{Is_Scx=kthi}', "");
Expect(0, 69825, '\p{^Is_Scx=kthi}', "");
Expect(0, 69825, '\P{Is_Scx=kthi}', "");
Expect(1, 69825, '\P{^Is_Scx=kthi}', "");
Expect(0, 69826, '\p{Is_Scx=kthi}', "");
Expect(1, 69826, '\p{^Is_Scx=kthi}', "");
Expect(1, 69826, '\P{Is_Scx=kthi}', "");
Expect(0, 69826, '\P{^Is_Scx=kthi}', "");
Expect(1, 69825, '\p{Is_Scx:   -Kthi}', "");
Expect(0, 69825, '\p{^Is_Scx:   -Kthi}', "");
Expect(0, 69825, '\P{Is_Scx:   -Kthi}', "");
Expect(1, 69825, '\P{^Is_Scx:   -Kthi}', "");
Expect(0, 69826, '\p{Is_Scx:   -Kthi}', "");
Expect(1, 69826, '\p{^Is_Scx:   -Kthi}', "");
Expect(1, 69826, '\P{Is_Scx:   -Kthi}', "");
Expect(0, 69826, '\P{^Is_Scx:   -Kthi}', "");
Error('\p{Script_Extensions:    /a/Tai_Tham}');
Error('\P{Script_Extensions:    /a/Tai_Tham}');
Expect(1, 6829, '\p{Script_Extensions=taitham}', "");
Expect(0, 6829, '\p{^Script_Extensions=taitham}', "");
Expect(0, 6829, '\P{Script_Extensions=taitham}', "");
Expect(1, 6829, '\P{^Script_Extensions=taitham}', "");
Expect(0, 6830, '\p{Script_Extensions=taitham}', "");
Expect(1, 6830, '\p{^Script_Extensions=taitham}', "");
Expect(1, 6830, '\P{Script_Extensions=taitham}', "");
Expect(0, 6830, '\P{^Script_Extensions=taitham}', "");
Expect(1, 6829, '\p{Script_Extensions=		tai_Tham}', "");
Expect(0, 6829, '\p{^Script_Extensions=		tai_Tham}', "");
Expect(0, 6829, '\P{Script_Extensions=		tai_Tham}', "");
Expect(1, 6829, '\P{^Script_Extensions=		tai_Tham}', "");
Expect(0, 6830, '\p{Script_Extensions=		tai_Tham}', "");
Expect(1, 6830, '\p{^Script_Extensions=		tai_Tham}', "");
Expect(1, 6830, '\P{Script_Extensions=		tai_Tham}', "");
Expect(0, 6830, '\P{^Script_Extensions=		tai_Tham}', "");
Error('\p{Scx=-:=Lana}');
Error('\P{Scx=-:=Lana}');
Expect(1, 6829, '\p{Scx=lana}', "");
Expect(0, 6829, '\p{^Scx=lana}', "");
Expect(0, 6829, '\P{Scx=lana}', "");
Expect(1, 6829, '\P{^Scx=lana}', "");
Expect(0, 6830, '\p{Scx=lana}', "");
Expect(1, 6830, '\p{^Scx=lana}', "");
Expect(1, 6830, '\P{Scx=lana}', "");
Expect(0, 6830, '\P{^Scx=lana}', "");
Expect(1, 6829, '\p{Scx=LANA}', "");
Expect(0, 6829, '\p{^Scx=LANA}', "");
Expect(0, 6829, '\P{Scx=LANA}', "");
Expect(1, 6829, '\P{^Scx=LANA}', "");
Expect(0, 6830, '\p{Scx=LANA}', "");
Expect(1, 6830, '\p{^Scx=LANA}', "");
Expect(1, 6830, '\P{Scx=LANA}', "");
Expect(0, 6830, '\P{^Scx=LANA}', "");
Error('\p{Is_Script_Extensions=_TAI_Tham/a/}');
Error('\P{Is_Script_Extensions=_TAI_Tham/a/}');
Expect(1, 6829, '\p{Is_Script_Extensions=taitham}', "");
Expect(0, 6829, '\p{^Is_Script_Extensions=taitham}', "");
Expect(0, 6829, '\P{Is_Script_Extensions=taitham}', "");
Expect(1, 6829, '\P{^Is_Script_Extensions=taitham}', "");
Expect(0, 6830, '\p{Is_Script_Extensions=taitham}', "");
Expect(1, 6830, '\p{^Is_Script_Extensions=taitham}', "");
Expect(1, 6830, '\P{Is_Script_Extensions=taitham}', "");
Expect(0, 6830, '\P{^Is_Script_Extensions=taitham}', "");
Expect(1, 6829, '\p{Is_Script_Extensions= 	tai_tham}', "");
Expect(0, 6829, '\p{^Is_Script_Extensions= 	tai_tham}', "");
Expect(0, 6829, '\P{Is_Script_Extensions= 	tai_tham}', "");
Expect(1, 6829, '\P{^Is_Script_Extensions= 	tai_tham}', "");
Expect(0, 6830, '\p{Is_Script_Extensions= 	tai_tham}', "");
Expect(1, 6830, '\p{^Is_Script_Extensions= 	tai_tham}', "");
Expect(1, 6830, '\P{Is_Script_Extensions= 	tai_tham}', "");
Expect(0, 6830, '\P{^Is_Script_Extensions= 	tai_tham}', "");
Error('\p{Is_Scx=/a/lana}');
Error('\P{Is_Scx=/a/lana}');
Expect(1, 6829, '\p{Is_Scx=lana}', "");
Expect(0, 6829, '\p{^Is_Scx=lana}', "");
Expect(0, 6829, '\P{Is_Scx=lana}', "");
Expect(1, 6829, '\P{^Is_Scx=lana}', "");
Expect(0, 6830, '\p{Is_Scx=lana}', "");
Expect(1, 6830, '\p{^Is_Scx=lana}', "");
Expect(1, 6830, '\P{Is_Scx=lana}', "");
Expect(0, 6830, '\P{^Is_Scx=lana}', "");
Expect(1, 6829, '\p{Is_Scx= lana}', "");
Expect(0, 6829, '\p{^Is_Scx= lana}', "");
Expect(0, 6829, '\P{Is_Scx= lana}', "");
Expect(1, 6829, '\P{^Is_Scx= lana}', "");
Expect(0, 6830, '\p{Is_Scx= lana}', "");
Expect(1, 6830, '\p{^Is_Scx= lana}', "");
Expect(1, 6830, '\P{Is_Scx= lana}', "");
Expect(0, 6830, '\P{^Is_Scx= lana}', "");
Error('\p{Script_Extensions=-	LAO:=}');
Error('\P{Script_Extensions=-	LAO:=}');
Expect(1, 3807, '\p{Script_Extensions=lao}', "");
Expect(0, 3807, '\p{^Script_Extensions=lao}', "");
Expect(0, 3807, '\P{Script_Extensions=lao}', "");
Expect(1, 3807, '\P{^Script_Extensions=lao}', "");
Expect(0, 3808, '\p{Script_Extensions=lao}', "");
Expect(1, 3808, '\p{^Script_Extensions=lao}', "");
Expect(1, 3808, '\P{Script_Extensions=lao}', "");
Expect(0, 3808, '\P{^Script_Extensions=lao}', "");
Expect(1, 3807, '\p{Script_Extensions=		Lao}', "");
Expect(0, 3807, '\p{^Script_Extensions=		Lao}', "");
Expect(0, 3807, '\P{Script_Extensions=		Lao}', "");
Expect(1, 3807, '\P{^Script_Extensions=		Lao}', "");
Expect(0, 3808, '\p{Script_Extensions=		Lao}', "");
Expect(1, 3808, '\p{^Script_Extensions=		Lao}', "");
Expect(1, 3808, '\P{Script_Extensions=		Lao}', "");
Expect(0, 3808, '\P{^Script_Extensions=		Lao}', "");
Error('\p{Scx=-:=laoo}');
Error('\P{Scx=-:=laoo}');
Expect(1, 3807, '\p{Scx=laoo}', "");
Expect(0, 3807, '\p{^Scx=laoo}', "");
Expect(0, 3807, '\P{Scx=laoo}', "");
Expect(1, 3807, '\P{^Scx=laoo}', "");
Expect(0, 3808, '\p{Scx=laoo}', "");
Expect(1, 3808, '\p{^Scx=laoo}', "");
Expect(1, 3808, '\P{Scx=laoo}', "");
Expect(0, 3808, '\P{^Scx=laoo}', "");
Expect(1, 3807, '\p{Scx= laoo}', "");
Expect(0, 3807, '\p{^Scx= laoo}', "");
Expect(0, 3807, '\P{Scx= laoo}', "");
Expect(1, 3807, '\P{^Scx= laoo}', "");
Expect(0, 3808, '\p{Scx= laoo}', "");
Expect(1, 3808, '\p{^Scx= laoo}', "");
Expect(1, 3808, '\P{Scx= laoo}', "");
Expect(0, 3808, '\P{^Scx= laoo}', "");
Error('\p{Is_Script_Extensions=	:=Lao}');
Error('\P{Is_Script_Extensions=	:=Lao}');
Expect(1, 3807, '\p{Is_Script_Extensions=lao}', "");
Expect(0, 3807, '\p{^Is_Script_Extensions=lao}', "");
Expect(0, 3807, '\P{Is_Script_Extensions=lao}', "");
Expect(1, 3807, '\P{^Is_Script_Extensions=lao}', "");
Expect(0, 3808, '\p{Is_Script_Extensions=lao}', "");
Expect(1, 3808, '\p{^Is_Script_Extensions=lao}', "");
Expect(1, 3808, '\P{Is_Script_Extensions=lao}', "");
Expect(0, 3808, '\P{^Is_Script_Extensions=lao}', "");
Expect(1, 3807, '\p{Is_Script_Extensions=_ Lao}', "");
Expect(0, 3807, '\p{^Is_Script_Extensions=_ Lao}', "");
Expect(0, 3807, '\P{Is_Script_Extensions=_ Lao}', "");
Expect(1, 3807, '\P{^Is_Script_Extensions=_ Lao}', "");
Expect(0, 3808, '\p{Is_Script_Extensions=_ Lao}', "");
Expect(1, 3808, '\p{^Is_Script_Extensions=_ Lao}', "");
Expect(1, 3808, '\P{Is_Script_Extensions=_ Lao}', "");
Expect(0, 3808, '\P{^Is_Script_Extensions=_ Lao}', "");
Error('\p{Is_Scx=_/a/LAOO}');
Error('\P{Is_Scx=_/a/LAOO}');
Expect(1, 3807, '\p{Is_Scx=laoo}', "");
Expect(0, 3807, '\p{^Is_Scx=laoo}', "");
Expect(0, 3807, '\P{Is_Scx=laoo}', "");
Expect(1, 3807, '\P{^Is_Scx=laoo}', "");
Expect(0, 3808, '\p{Is_Scx=laoo}', "");
Expect(1, 3808, '\p{^Is_Scx=laoo}', "");
Expect(1, 3808, '\P{Is_Scx=laoo}', "");
Expect(0, 3808, '\P{^Is_Scx=laoo}', "");
Expect(1, 3807, '\p{Is_Scx=-Laoo}', "");
Expect(0, 3807, '\p{^Is_Scx=-Laoo}', "");
Expect(0, 3807, '\P{Is_Scx=-Laoo}', "");
Expect(1, 3807, '\P{^Is_Scx=-Laoo}', "");
Expect(0, 3808, '\p{Is_Scx=-Laoo}', "");
Expect(1, 3808, '\p{^Is_Scx=-Laoo}', "");
Expect(1, 3808, '\P{Is_Scx=-Laoo}', "");
Expect(0, 3808, '\P{^Is_Scx=-Laoo}', "");
Error('\p{Script_Extensions=/a/ Latin}');
Error('\P{Script_Extensions=/a/ Latin}');
Expect(1, 65370, '\p{Script_Extensions:	latin}', "");
Expect(0, 65370, '\p{^Script_Extensions:	latin}', "");
Expect(0, 65370, '\P{Script_Extensions:	latin}', "");
Expect(1, 65370, '\P{^Script_Extensions:	latin}', "");
Expect(0, 65371, '\p{Script_Extensions:	latin}', "");
Expect(1, 65371, '\p{^Script_Extensions:	latin}', "");
Expect(1, 65371, '\P{Script_Extensions:	latin}', "");
Expect(0, 65371, '\P{^Script_Extensions:	latin}', "");
Expect(1, 65370, '\p{Script_Extensions=-_latin}', "");
Expect(0, 65370, '\p{^Script_Extensions=-_latin}', "");
Expect(0, 65370, '\P{Script_Extensions=-_latin}', "");
Expect(1, 65370, '\P{^Script_Extensions=-_latin}', "");
Expect(0, 65371, '\p{Script_Extensions=-_latin}', "");
Expect(1, 65371, '\p{^Script_Extensions=-_latin}', "");
Expect(1, 65371, '\P{Script_Extensions=-_latin}', "");
Expect(0, 65371, '\P{^Script_Extensions=-_latin}', "");
Error('\p{Scx=__latn:=}');
Error('\P{Scx=__latn:=}');
Expect(1, 65370, '\p{Scx=latn}', "");
Expect(0, 65370, '\p{^Scx=latn}', "");
Expect(0, 65370, '\P{Scx=latn}', "");
Expect(1, 65370, '\P{^Scx=latn}', "");
Expect(0, 65371, '\p{Scx=latn}', "");
Expect(1, 65371, '\p{^Scx=latn}', "");
Expect(1, 65371, '\P{Scx=latn}', "");
Expect(0, 65371, '\P{^Scx=latn}', "");
Expect(1, 65370, '\p{Scx=- Latn}', "");
Expect(0, 65370, '\p{^Scx=- Latn}', "");
Expect(0, 65370, '\P{Scx=- Latn}', "");
Expect(1, 65370, '\P{^Scx=- Latn}', "");
Expect(0, 65371, '\p{Scx=- Latn}', "");
Expect(1, 65371, '\p{^Scx=- Latn}', "");
Expect(1, 65371, '\P{Scx=- Latn}', "");
Expect(0, 65371, '\P{^Scx=- Latn}', "");
Error('\p{Is_Script_Extensions=-	latin/a/}');
Error('\P{Is_Script_Extensions=-	latin/a/}');
Expect(1, 65370, '\p{Is_Script_Extensions=latin}', "");
Expect(0, 65370, '\p{^Is_Script_Extensions=latin}', "");
Expect(0, 65370, '\P{Is_Script_Extensions=latin}', "");
Expect(1, 65370, '\P{^Is_Script_Extensions=latin}', "");
Expect(0, 65371, '\p{Is_Script_Extensions=latin}', "");
Expect(1, 65371, '\p{^Is_Script_Extensions=latin}', "");
Expect(1, 65371, '\P{Is_Script_Extensions=latin}', "");
Expect(0, 65371, '\P{^Is_Script_Extensions=latin}', "");
Expect(1, 65370, '\p{Is_Script_Extensions=		Latin}', "");
Expect(0, 65370, '\p{^Is_Script_Extensions=		Latin}', "");
Expect(0, 65370, '\P{Is_Script_Extensions=		Latin}', "");
Expect(1, 65370, '\P{^Is_Script_Extensions=		Latin}', "");
Expect(0, 65371, '\p{Is_Script_Extensions=		Latin}', "");
Expect(1, 65371, '\p{^Is_Script_Extensions=		Latin}', "");
Expect(1, 65371, '\P{Is_Script_Extensions=		Latin}', "");
Expect(0, 65371, '\P{^Is_Script_Extensions=		Latin}', "");
Error('\p{Is_Scx=_ Latn/a/}');
Error('\P{Is_Scx=_ Latn/a/}');
Expect(1, 65370, '\p{Is_Scx=latn}', "");
Expect(0, 65370, '\p{^Is_Scx=latn}', "");
Expect(0, 65370, '\P{Is_Scx=latn}', "");
Expect(1, 65370, '\P{^Is_Scx=latn}', "");
Expect(0, 65371, '\p{Is_Scx=latn}', "");
Expect(1, 65371, '\p{^Is_Scx=latn}', "");
Expect(1, 65371, '\P{Is_Scx=latn}', "");
Expect(0, 65371, '\P{^Is_Scx=latn}', "");
Expect(1, 65370, '\p{Is_Scx=_latn}', "");
Expect(0, 65370, '\p{^Is_Scx=_latn}', "");
Expect(0, 65370, '\P{Is_Scx=_latn}', "");
Expect(1, 65370, '\P{^Is_Scx=_latn}', "");
Expect(0, 65371, '\p{Is_Scx=_latn}', "");
Expect(1, 65371, '\p{^Is_Scx=_latn}', "");
Expect(1, 65371, '\P{Is_Scx=_latn}', "");
Expect(0, 65371, '\P{^Is_Scx=_latn}', "");
Error('\p{Script_Extensions=	Lepcha/a/}');
Error('\P{Script_Extensions=	Lepcha/a/}');
Expect(1, 7247, '\p{Script_Extensions=lepcha}', "");
Expect(0, 7247, '\p{^Script_Extensions=lepcha}', "");
Expect(0, 7247, '\P{Script_Extensions=lepcha}', "");
Expect(1, 7247, '\P{^Script_Extensions=lepcha}', "");
Expect(0, 7248, '\p{Script_Extensions=lepcha}', "");
Expect(1, 7248, '\p{^Script_Extensions=lepcha}', "");
Expect(1, 7248, '\P{Script_Extensions=lepcha}', "");
Expect(0, 7248, '\P{^Script_Extensions=lepcha}', "");
Expect(1, 7247, '\p{Script_Extensions: -Lepcha}', "");
Expect(0, 7247, '\p{^Script_Extensions: -Lepcha}', "");
Expect(0, 7247, '\P{Script_Extensions: -Lepcha}', "");
Expect(1, 7247, '\P{^Script_Extensions: -Lepcha}', "");
Expect(0, 7248, '\p{Script_Extensions: -Lepcha}', "");
Expect(1, 7248, '\p{^Script_Extensions: -Lepcha}', "");
Expect(1, 7248, '\P{Script_Extensions: -Lepcha}', "");
Expect(0, 7248, '\P{^Script_Extensions: -Lepcha}', "");
Error('\p{Scx=	LEPC/a/}');
Error('\P{Scx=	LEPC/a/}');
Expect(1, 7247, '\p{Scx=lepc}', "");
Expect(0, 7247, '\p{^Scx=lepc}', "");
Expect(0, 7247, '\P{Scx=lepc}', "");
Expect(1, 7247, '\P{^Scx=lepc}', "");
Expect(0, 7248, '\p{Scx=lepc}', "");
Expect(1, 7248, '\p{^Scx=lepc}', "");
Expect(1, 7248, '\P{Scx=lepc}', "");
Expect(0, 7248, '\P{^Scx=lepc}', "");
Expect(1, 7247, '\p{Scx=-	LEPC}', "");
Expect(0, 7247, '\p{^Scx=-	LEPC}', "");
Expect(0, 7247, '\P{Scx=-	LEPC}', "");
Expect(1, 7247, '\P{^Scx=-	LEPC}', "");
Expect(0, 7248, '\p{Scx=-	LEPC}', "");
Expect(1, 7248, '\p{^Scx=-	LEPC}', "");
Expect(1, 7248, '\P{Scx=-	LEPC}', "");
Expect(0, 7248, '\P{^Scx=-	LEPC}', "");
Error('\p{Is_Script_Extensions=  LEPCHA:=}');
Error('\P{Is_Script_Extensions=  LEPCHA:=}');
Expect(1, 7247, '\p{Is_Script_Extensions=lepcha}', "");
Expect(0, 7247, '\p{^Is_Script_Extensions=lepcha}', "");
Expect(0, 7247, '\P{Is_Script_Extensions=lepcha}', "");
Expect(1, 7247, '\P{^Is_Script_Extensions=lepcha}', "");
Expect(0, 7248, '\p{Is_Script_Extensions=lepcha}', "");
Expect(1, 7248, '\p{^Is_Script_Extensions=lepcha}', "");
Expect(1, 7248, '\P{Is_Script_Extensions=lepcha}', "");
Expect(0, 7248, '\P{^Is_Script_Extensions=lepcha}', "");
Expect(1, 7247, '\p{Is_Script_Extensions= -lepcha}', "");
Expect(0, 7247, '\p{^Is_Script_Extensions= -lepcha}', "");
Expect(0, 7247, '\P{Is_Script_Extensions= -lepcha}', "");
Expect(1, 7247, '\P{^Is_Script_Extensions= -lepcha}', "");
Expect(0, 7248, '\p{Is_Script_Extensions= -lepcha}', "");
Expect(1, 7248, '\p{^Is_Script_Extensions= -lepcha}', "");
Expect(1, 7248, '\P{Is_Script_Extensions= -lepcha}', "");
Expect(0, 7248, '\P{^Is_Script_Extensions= -lepcha}', "");
Error('\p{Is_Scx=:=Lepc}');
Error('\P{Is_Scx=:=Lepc}');
Expect(1, 7247, '\p{Is_Scx=lepc}', "");
Expect(0, 7247, '\p{^Is_Scx=lepc}', "");
Expect(0, 7247, '\P{Is_Scx=lepc}', "");
Expect(1, 7247, '\P{^Is_Scx=lepc}', "");
Expect(0, 7248, '\p{Is_Scx=lepc}', "");
Expect(1, 7248, '\p{^Is_Scx=lepc}', "");
Expect(1, 7248, '\P{Is_Scx=lepc}', "");
Expect(0, 7248, '\P{^Is_Scx=lepc}', "");
Expect(1, 7247, '\p{Is_Scx=_	Lepc}', "");
Expect(0, 7247, '\p{^Is_Scx=_	Lepc}', "");
Expect(0, 7247, '\P{Is_Scx=_	Lepc}', "");
Expect(1, 7247, '\P{^Is_Scx=_	Lepc}', "");
Expect(0, 7248, '\p{Is_Scx=_	Lepc}', "");
Expect(1, 7248, '\p{^Is_Scx=_	Lepc}', "");
Expect(1, 7248, '\P{Is_Scx=_	Lepc}', "");
Expect(0, 7248, '\P{^Is_Scx=_	Lepc}', "");
Error('\p{Script_Extensions=/a/ _Limbu}');
Error('\P{Script_Extensions=/a/ _Limbu}');
Expect(1, 6479, '\p{Script_Extensions=limbu}', "");
Expect(0, 6479, '\p{^Script_Extensions=limbu}', "");
Expect(0, 6479, '\P{Script_Extensions=limbu}', "");
Expect(1, 6479, '\P{^Script_Extensions=limbu}', "");
Expect(0, 6480, '\p{Script_Extensions=limbu}', "");
Expect(1, 6480, '\p{^Script_Extensions=limbu}', "");
Expect(1, 6480, '\P{Script_Extensions=limbu}', "");
Expect(0, 6480, '\P{^Script_Extensions=limbu}', "");
Expect(1, 6479, '\p{Script_Extensions=-_Limbu}', "");
Expect(0, 6479, '\p{^Script_Extensions=-_Limbu}', "");
Expect(0, 6479, '\P{Script_Extensions=-_Limbu}', "");
Expect(1, 6479, '\P{^Script_Extensions=-_Limbu}', "");
Expect(0, 6480, '\p{Script_Extensions=-_Limbu}', "");
Expect(1, 6480, '\p{^Script_Extensions=-_Limbu}', "");
Expect(1, 6480, '\P{Script_Extensions=-_Limbu}', "");
Expect(0, 6480, '\P{^Script_Extensions=-_Limbu}', "");
Error('\p{Scx: _Limb/a/}');
Error('\P{Scx: _Limb/a/}');
Expect(1, 6479, '\p{Scx=limb}', "");
Expect(0, 6479, '\p{^Scx=limb}', "");
Expect(0, 6479, '\P{Scx=limb}', "");
Expect(1, 6479, '\P{^Scx=limb}', "");
Expect(0, 6480, '\p{Scx=limb}', "");
Expect(1, 6480, '\p{^Scx=limb}', "");
Expect(1, 6480, '\P{Scx=limb}', "");
Expect(0, 6480, '\P{^Scx=limb}', "");
Expect(1, 6479, '\p{Scx= _LIMB}', "");
Expect(0, 6479, '\p{^Scx= _LIMB}', "");
Expect(0, 6479, '\P{Scx= _LIMB}', "");
Expect(1, 6479, '\P{^Scx= _LIMB}', "");
Expect(0, 6480, '\p{Scx= _LIMB}', "");
Expect(1, 6480, '\p{^Scx= _LIMB}', "");
Expect(1, 6480, '\P{Scx= _LIMB}', "");
Expect(0, 6480, '\P{^Scx= _LIMB}', "");
Error('\p{Is_Script_Extensions=	/a/limbu}');
Error('\P{Is_Script_Extensions=	/a/limbu}');
Expect(1, 6479, '\p{Is_Script_Extensions=limbu}', "");
Expect(0, 6479, '\p{^Is_Script_Extensions=limbu}', "");
Expect(0, 6479, '\P{Is_Script_Extensions=limbu}', "");
Expect(1, 6479, '\P{^Is_Script_Extensions=limbu}', "");
Expect(0, 6480, '\p{Is_Script_Extensions=limbu}', "");
Expect(1, 6480, '\p{^Is_Script_Extensions=limbu}', "");
Expect(1, 6480, '\P{Is_Script_Extensions=limbu}', "");
Expect(0, 6480, '\P{^Is_Script_Extensions=limbu}', "");
Error('\p{Is_Scx=:=	_Limb}');
Error('\P{Is_Scx=:=	_Limb}');
Expect(1, 6479, '\p{Is_Scx=limb}', "");
Expect(0, 6479, '\p{^Is_Scx=limb}', "");
Expect(0, 6479, '\P{Is_Scx=limb}', "");
Expect(1, 6479, '\P{^Is_Scx=limb}', "");
Expect(0, 6480, '\p{Is_Scx=limb}', "");
Expect(1, 6480, '\p{^Is_Scx=limb}', "");
Expect(1, 6480, '\P{Is_Scx=limb}', "");
Expect(0, 6480, '\P{^Is_Scx=limb}', "");
Expect(1, 6479, '\p{Is_Scx= Limb}', "");
Expect(0, 6479, '\p{^Is_Scx= Limb}', "");
Expect(0, 6479, '\P{Is_Scx= Limb}', "");
Expect(1, 6479, '\P{^Is_Scx= Limb}', "");
Expect(0, 6480, '\p{Is_Scx= Limb}', "");
Expect(1, 6480, '\p{^Is_Scx= Limb}', "");
Expect(1, 6480, '\P{Is_Scx= Limb}', "");
Expect(0, 6480, '\P{^Is_Scx= Limb}', "");
Error('\p{Script_Extensions= Linear_B/a/}');
Error('\P{Script_Extensions= Linear_B/a/}');
Expect(1, 65855, '\p{Script_Extensions=linearb}', "");
Expect(0, 65855, '\p{^Script_Extensions=linearb}', "");
Expect(0, 65855, '\P{Script_Extensions=linearb}', "");
Expect(1, 65855, '\P{^Script_Extensions=linearb}', "");
Expect(0, 65856, '\p{Script_Extensions=linearb}', "");
Expect(1, 65856, '\p{^Script_Extensions=linearb}', "");
Expect(1, 65856, '\P{Script_Extensions=linearb}', "");
Expect(0, 65856, '\P{^Script_Extensions=linearb}', "");
Expect(1, 65855, '\p{Script_Extensions=	_LINEAR_B}', "");
Expect(0, 65855, '\p{^Script_Extensions=	_LINEAR_B}', "");
Expect(0, 65855, '\P{Script_Extensions=	_LINEAR_B}', "");
Expect(1, 65855, '\P{^Script_Extensions=	_LINEAR_B}', "");
Expect(0, 65856, '\p{Script_Extensions=	_LINEAR_B}', "");
Expect(1, 65856, '\p{^Script_Extensions=	_LINEAR_B}', "");
Expect(1, 65856, '\P{Script_Extensions=	_LINEAR_B}', "");
Expect(0, 65856, '\P{^Script_Extensions=	_LINEAR_B}', "");
Error('\p{Scx: /a/	 LINB}');
Error('\P{Scx: /a/	 LINB}');
Expect(1, 65855, '\p{Scx=linb}', "");
Expect(0, 65855, '\p{^Scx=linb}', "");
Expect(0, 65855, '\P{Scx=linb}', "");
Expect(1, 65855, '\P{^Scx=linb}', "");
Expect(0, 65856, '\p{Scx=linb}', "");
Expect(1, 65856, '\p{^Scx=linb}', "");
Expect(1, 65856, '\P{Scx=linb}', "");
Expect(0, 65856, '\P{^Scx=linb}', "");
Expect(1, 65855, '\p{Scx=_ LINB}', "");
Expect(0, 65855, '\p{^Scx=_ LINB}', "");
Expect(0, 65855, '\P{Scx=_ LINB}', "");
Expect(1, 65855, '\P{^Scx=_ LINB}', "");
Expect(0, 65856, '\p{Scx=_ LINB}', "");
Expect(1, 65856, '\p{^Scx=_ LINB}', "");
Expect(1, 65856, '\P{Scx=_ LINB}', "");
Expect(0, 65856, '\P{^Scx=_ LINB}', "");
Error('\p{Is_Script_Extensions=/a/- linear_B}');
Error('\P{Is_Script_Extensions=/a/- linear_B}');
Expect(1, 65855, '\p{Is_Script_Extensions=linearb}', "");
Expect(0, 65855, '\p{^Is_Script_Extensions=linearb}', "");
Expect(0, 65855, '\P{Is_Script_Extensions=linearb}', "");
Expect(1, 65855, '\P{^Is_Script_Extensions=linearb}', "");
Expect(0, 65856, '\p{Is_Script_Extensions=linearb}', "");
Expect(1, 65856, '\p{^Is_Script_Extensions=linearb}', "");
Expect(1, 65856, '\P{Is_Script_Extensions=linearb}', "");
Expect(0, 65856, '\P{^Is_Script_Extensions=linearb}', "");
Expect(1, 65855, '\p{Is_Script_Extensions=linear_B}', "");
Expect(0, 65855, '\p{^Is_Script_Extensions=linear_B}', "");
Expect(0, 65855, '\P{Is_Script_Extensions=linear_B}', "");
Expect(1, 65855, '\P{^Is_Script_Extensions=linear_B}', "");
Expect(0, 65856, '\p{Is_Script_Extensions=linear_B}', "");
Expect(1, 65856, '\p{^Is_Script_Extensions=linear_B}', "");
Expect(1, 65856, '\P{Is_Script_Extensions=linear_B}', "");
Expect(0, 65856, '\P{^Is_Script_Extensions=linear_B}', "");
Error('\p{Is_Scx= :=linb}');
Error('\P{Is_Scx= :=linb}');
Expect(1, 65855, '\p{Is_Scx=linb}', "");
Expect(0, 65855, '\p{^Is_Scx=linb}', "");
Expect(0, 65855, '\P{Is_Scx=linb}', "");
Expect(1, 65855, '\P{^Is_Scx=linb}', "");
Expect(0, 65856, '\p{Is_Scx=linb}', "");
Expect(1, 65856, '\p{^Is_Scx=linb}', "");
Expect(1, 65856, '\P{Is_Scx=linb}', "");
Expect(0, 65856, '\P{^Is_Scx=linb}', "");
Expect(1, 65855, '\p{Is_Scx=__Linb}', "");
Expect(0, 65855, '\p{^Is_Scx=__Linb}', "");
Expect(0, 65855, '\P{Is_Scx=__Linb}', "");
Expect(1, 65855, '\P{^Is_Scx=__Linb}', "");
Expect(0, 65856, '\p{Is_Scx=__Linb}', "");
Expect(1, 65856, '\p{^Is_Scx=__Linb}', "");
Expect(1, 65856, '\P{Is_Scx=__Linb}', "");
Expect(0, 65856, '\P{^Is_Scx=__Linb}', "");
Error('\p{Script_Extensions=_ LISU/a/}');
Error('\P{Script_Extensions=_ LISU/a/}');
Expect(1, 42239, '\p{Script_Extensions=lisu}', "");
Expect(0, 42239, '\p{^Script_Extensions=lisu}', "");
Expect(0, 42239, '\P{Script_Extensions=lisu}', "");
Expect(1, 42239, '\P{^Script_Extensions=lisu}', "");
Expect(0, 42240, '\p{Script_Extensions=lisu}', "");
Expect(1, 42240, '\p{^Script_Extensions=lisu}', "");
Expect(1, 42240, '\P{Script_Extensions=lisu}', "");
Expect(0, 42240, '\P{^Script_Extensions=lisu}', "");
Expect(1, 42239, '\p{Script_Extensions=	 lisu}', "");
Expect(0, 42239, '\p{^Script_Extensions=	 lisu}', "");
Expect(0, 42239, '\P{Script_Extensions=	 lisu}', "");
Expect(1, 42239, '\P{^Script_Extensions=	 lisu}', "");
Expect(0, 42240, '\p{Script_Extensions=	 lisu}', "");
Expect(1, 42240, '\p{^Script_Extensions=	 lisu}', "");
Expect(1, 42240, '\P{Script_Extensions=	 lisu}', "");
Expect(0, 42240, '\P{^Script_Extensions=	 lisu}', "");
Error('\p{Scx=_:=Lisu}');
Error('\P{Scx=_:=Lisu}');
Expect(1, 42239, '\p{Scx=lisu}', "");
Expect(0, 42239, '\p{^Scx=lisu}', "");
Expect(0, 42239, '\P{Scx=lisu}', "");
Expect(1, 42239, '\P{^Scx=lisu}', "");
Expect(0, 42240, '\p{Scx=lisu}', "");
Expect(1, 42240, '\p{^Scx=lisu}', "");
Expect(1, 42240, '\P{Scx=lisu}', "");
Expect(0, 42240, '\P{^Scx=lisu}', "");
Expect(1, 42239, '\p{Scx=--lisu}', "");
Expect(0, 42239, '\p{^Scx=--lisu}', "");
Expect(0, 42239, '\P{Scx=--lisu}', "");
Expect(1, 42239, '\P{^Scx=--lisu}', "");
Expect(0, 42240, '\p{Scx=--lisu}', "");
Expect(1, 42240, '\p{^Scx=--lisu}', "");
Expect(1, 42240, '\P{Scx=--lisu}', "");
Expect(0, 42240, '\P{^Scx=--lisu}', "");
Error('\p{Is_Script_Extensions: := Lisu}');
Error('\P{Is_Script_Extensions: := Lisu}');
Expect(1, 42239, '\p{Is_Script_Extensions=lisu}', "");
Expect(0, 42239, '\p{^Is_Script_Extensions=lisu}', "");
Expect(0, 42239, '\P{Is_Script_Extensions=lisu}', "");
Expect(1, 42239, '\P{^Is_Script_Extensions=lisu}', "");
Expect(0, 42240, '\p{Is_Script_Extensions=lisu}', "");
Expect(1, 42240, '\p{^Is_Script_Extensions=lisu}', "");
Expect(1, 42240, '\P{Is_Script_Extensions=lisu}', "");
Expect(0, 42240, '\P{^Is_Script_Extensions=lisu}', "");
Expect(1, 42239, '\p{Is_Script_Extensions=-_Lisu}', "");
Expect(0, 42239, '\p{^Is_Script_Extensions=-_Lisu}', "");
Expect(0, 42239, '\P{Is_Script_Extensions=-_Lisu}', "");
Expect(1, 42239, '\P{^Is_Script_Extensions=-_Lisu}', "");
Expect(0, 42240, '\p{Is_Script_Extensions=-_Lisu}', "");
Expect(1, 42240, '\p{^Is_Script_Extensions=-_Lisu}', "");
Expect(1, 42240, '\P{Is_Script_Extensions=-_Lisu}', "");
Expect(0, 42240, '\P{^Is_Script_Extensions=-_Lisu}', "");
Error('\p{Is_Scx= Lisu/a/}');
Error('\P{Is_Scx= Lisu/a/}');
Expect(1, 42239, '\p{Is_Scx=lisu}', "");
Expect(0, 42239, '\p{^Is_Scx=lisu}', "");
Expect(0, 42239, '\P{Is_Scx=lisu}', "");
Expect(1, 42239, '\P{^Is_Scx=lisu}', "");
Expect(0, 42240, '\p{Is_Scx=lisu}', "");
Expect(1, 42240, '\p{^Is_Scx=lisu}', "");
Expect(1, 42240, '\P{Is_Scx=lisu}', "");
Expect(0, 42240, '\P{^Is_Scx=lisu}', "");
Expect(1, 42239, '\p{Is_Scx=Lisu}', "");
Expect(0, 42239, '\p{^Is_Scx=Lisu}', "");
Expect(0, 42239, '\P{Is_Scx=Lisu}', "");
Expect(1, 42239, '\P{^Is_Scx=Lisu}', "");
Expect(0, 42240, '\p{Is_Scx=Lisu}', "");
Expect(1, 42240, '\p{^Is_Scx=Lisu}', "");
Expect(1, 42240, '\P{Is_Scx=Lisu}', "");
Expect(0, 42240, '\P{^Is_Scx=Lisu}', "");
Error('\p{Script_Extensions= Lycian:=}');
Error('\P{Script_Extensions= Lycian:=}');
Expect(1, 66204, '\p{Script_Extensions=lycian}', "");
Expect(0, 66204, '\p{^Script_Extensions=lycian}', "");
Expect(0, 66204, '\P{Script_Extensions=lycian}', "");
Expect(1, 66204, '\P{^Script_Extensions=lycian}', "");
Expect(0, 66205, '\p{Script_Extensions=lycian}', "");
Expect(1, 66205, '\p{^Script_Extensions=lycian}', "");
Expect(1, 66205, '\P{Script_Extensions=lycian}', "");
Expect(0, 66205, '\P{^Script_Extensions=lycian}', "");
Expect(1, 66204, '\p{Script_Extensions= Lycian}', "");
Expect(0, 66204, '\p{^Script_Extensions= Lycian}', "");
Expect(0, 66204, '\P{Script_Extensions= Lycian}', "");
Expect(1, 66204, '\P{^Script_Extensions= Lycian}', "");
Expect(0, 66205, '\p{Script_Extensions= Lycian}', "");
Expect(1, 66205, '\p{^Script_Extensions= Lycian}', "");
Expect(1, 66205, '\P{Script_Extensions= Lycian}', "");
Expect(0, 66205, '\P{^Script_Extensions= Lycian}', "");
Error('\p{Scx:	/a/lyci}');
Error('\P{Scx:	/a/lyci}');
Expect(1, 66204, '\p{Scx=lyci}', "");
Expect(0, 66204, '\p{^Scx=lyci}', "");
Expect(0, 66204, '\P{Scx=lyci}', "");
Expect(1, 66204, '\P{^Scx=lyci}', "");
Expect(0, 66205, '\p{Scx=lyci}', "");
Expect(1, 66205, '\p{^Scx=lyci}', "");
Expect(1, 66205, '\P{Scx=lyci}', "");
Expect(0, 66205, '\P{^Scx=lyci}', "");
Expect(1, 66204, '\p{Scx:	-Lyci}', "");
Expect(0, 66204, '\p{^Scx:	-Lyci}', "");
Expect(0, 66204, '\P{Scx:	-Lyci}', "");
Expect(1, 66204, '\P{^Scx:	-Lyci}', "");
Expect(0, 66205, '\p{Scx:	-Lyci}', "");
Expect(1, 66205, '\p{^Scx:	-Lyci}', "");
Expect(1, 66205, '\P{Scx:	-Lyci}', "");
Expect(0, 66205, '\P{^Scx:	-Lyci}', "");
Error('\p{Is_Script_Extensions=:=  Lycian}');
Error('\P{Is_Script_Extensions=:=  Lycian}');
Expect(1, 66204, '\p{Is_Script_Extensions=lycian}', "");
Expect(0, 66204, '\p{^Is_Script_Extensions=lycian}', "");
Expect(0, 66204, '\P{Is_Script_Extensions=lycian}', "");
Expect(1, 66204, '\P{^Is_Script_Extensions=lycian}', "");
Expect(0, 66205, '\p{Is_Script_Extensions=lycian}', "");
Expect(1, 66205, '\p{^Is_Script_Extensions=lycian}', "");
Expect(1, 66205, '\P{Is_Script_Extensions=lycian}', "");
Expect(0, 66205, '\P{^Is_Script_Extensions=lycian}', "");
Expect(1, 66204, '\p{Is_Script_Extensions=	 Lycian}', "");
Expect(0, 66204, '\p{^Is_Script_Extensions=	 Lycian}', "");
Expect(0, 66204, '\P{Is_Script_Extensions=	 Lycian}', "");
Expect(1, 66204, '\P{^Is_Script_Extensions=	 Lycian}', "");
Expect(0, 66205, '\p{Is_Script_Extensions=	 Lycian}', "");
Expect(1, 66205, '\p{^Is_Script_Extensions=	 Lycian}', "");
Expect(1, 66205, '\P{Is_Script_Extensions=	 Lycian}', "");
Expect(0, 66205, '\P{^Is_Script_Extensions=	 Lycian}', "");
Error('\p{Is_Scx=lyci/a/}');
Error('\P{Is_Scx=lyci/a/}');
Expect(1, 66204, '\p{Is_Scx=lyci}', "");
Expect(0, 66204, '\p{^Is_Scx=lyci}', "");
Expect(0, 66204, '\P{Is_Scx=lyci}', "");
Expect(1, 66204, '\P{^Is_Scx=lyci}', "");
Expect(0, 66205, '\p{Is_Scx=lyci}', "");
Expect(1, 66205, '\p{^Is_Scx=lyci}', "");
Expect(1, 66205, '\P{Is_Scx=lyci}', "");
Expect(0, 66205, '\P{^Is_Scx=lyci}', "");
Expect(1, 66204, '\p{Is_Scx=_ Lyci}', "");
Expect(0, 66204, '\p{^Is_Scx=_ Lyci}', "");
Expect(0, 66204, '\P{Is_Scx=_ Lyci}', "");
Expect(1, 66204, '\P{^Is_Scx=_ Lyci}', "");
Expect(0, 66205, '\p{Is_Scx=_ Lyci}', "");
Expect(1, 66205, '\p{^Is_Scx=_ Lyci}', "");
Expect(1, 66205, '\P{Is_Scx=_ Lyci}', "");
Expect(0, 66205, '\P{^Is_Scx=_ Lyci}', "");
Error('\p{Script_Extensions:	 /a/LYDIAN}');
Error('\P{Script_Extensions:	 /a/LYDIAN}');
Expect(1, 67903, '\p{Script_Extensions=lydian}', "");
Expect(0, 67903, '\p{^Script_Extensions=lydian}', "");
Expect(0, 67903, '\P{Script_Extensions=lydian}', "");
Expect(1, 67903, '\P{^Script_Extensions=lydian}', "");
Expect(0, 67904, '\p{Script_Extensions=lydian}', "");
Expect(1, 67904, '\p{^Script_Extensions=lydian}', "");
Expect(1, 67904, '\P{Script_Extensions=lydian}', "");
Expect(0, 67904, '\P{^Script_Extensions=lydian}', "");
Expect(1, 67903, '\p{Script_Extensions=-	lydian}', "");
Expect(0, 67903, '\p{^Script_Extensions=-	lydian}', "");
Expect(0, 67903, '\P{Script_Extensions=-	lydian}', "");
Expect(1, 67903, '\P{^Script_Extensions=-	lydian}', "");
Expect(0, 67904, '\p{Script_Extensions=-	lydian}', "");
Expect(1, 67904, '\p{^Script_Extensions=-	lydian}', "");
Expect(1, 67904, '\P{Script_Extensions=-	lydian}', "");
Expect(0, 67904, '\P{^Script_Extensions=-	lydian}', "");
Error('\p{Scx=	_Lydi:=}');
Error('\P{Scx=	_Lydi:=}');
Expect(1, 67903, '\p{Scx=lydi}', "");
Expect(0, 67903, '\p{^Scx=lydi}', "");
Expect(0, 67903, '\P{Scx=lydi}', "");
Expect(1, 67903, '\P{^Scx=lydi}', "");
Expect(0, 67904, '\p{Scx=lydi}', "");
Expect(1, 67904, '\p{^Scx=lydi}', "");
Expect(1, 67904, '\P{Scx=lydi}', "");
Expect(0, 67904, '\P{^Scx=lydi}', "");
Expect(1, 67903, '\p{Scx=--lydi}', "");
Expect(0, 67903, '\p{^Scx=--lydi}', "");
Expect(0, 67903, '\P{Scx=--lydi}', "");
Expect(1, 67903, '\P{^Scx=--lydi}', "");
Expect(0, 67904, '\p{Scx=--lydi}', "");
Expect(1, 67904, '\p{^Scx=--lydi}', "");
Expect(1, 67904, '\P{Scx=--lydi}', "");
Expect(0, 67904, '\P{^Scx=--lydi}', "");
Error('\p{Is_Script_Extensions=/a/		lydian}');
Error('\P{Is_Script_Extensions=/a/		lydian}');
Expect(1, 67903, '\p{Is_Script_Extensions=lydian}', "");
Expect(0, 67903, '\p{^Is_Script_Extensions=lydian}', "");
Expect(0, 67903, '\P{Is_Script_Extensions=lydian}', "");
Expect(1, 67903, '\P{^Is_Script_Extensions=lydian}', "");
Expect(0, 67904, '\p{Is_Script_Extensions=lydian}', "");
Expect(1, 67904, '\p{^Is_Script_Extensions=lydian}', "");
Expect(1, 67904, '\P{Is_Script_Extensions=lydian}', "");
Expect(0, 67904, '\P{^Is_Script_Extensions=lydian}', "");
Expect(1, 67903, '\p{Is_Script_Extensions=__Lydian}', "");
Expect(0, 67903, '\p{^Is_Script_Extensions=__Lydian}', "");
Expect(0, 67903, '\P{Is_Script_Extensions=__Lydian}', "");
Expect(1, 67903, '\P{^Is_Script_Extensions=__Lydian}', "");
Expect(0, 67904, '\p{Is_Script_Extensions=__Lydian}', "");
Expect(1, 67904, '\p{^Is_Script_Extensions=__Lydian}', "");
Expect(1, 67904, '\P{Is_Script_Extensions=__Lydian}', "");
Expect(0, 67904, '\P{^Is_Script_Extensions=__Lydian}', "");
Error('\p{Is_Scx::=_ Lydi}');
Error('\P{Is_Scx::=_ Lydi}');
Expect(1, 67903, '\p{Is_Scx=lydi}', "");
Expect(0, 67903, '\p{^Is_Scx=lydi}', "");
Expect(0, 67903, '\P{Is_Scx=lydi}', "");
Expect(1, 67903, '\P{^Is_Scx=lydi}', "");
Expect(0, 67904, '\p{Is_Scx=lydi}', "");
Expect(1, 67904, '\p{^Is_Scx=lydi}', "");
Expect(1, 67904, '\P{Is_Scx=lydi}', "");
Expect(0, 67904, '\P{^Is_Scx=lydi}', "");
Expect(1, 67903, '\p{Is_Scx=-lydi}', "");
Expect(0, 67903, '\p{^Is_Scx=-lydi}', "");
Expect(0, 67903, '\P{Is_Scx=-lydi}', "");
Expect(1, 67903, '\P{^Is_Scx=-lydi}', "");
Expect(0, 67904, '\p{Is_Scx=-lydi}', "");
Expect(1, 67904, '\p{^Is_Scx=-lydi}', "");
Expect(1, 67904, '\P{Is_Scx=-lydi}', "");
Expect(0, 67904, '\P{^Is_Scx=-lydi}', "");
Error('\p{Script_Extensions=:=-Mandaic}');
Error('\P{Script_Extensions=:=-Mandaic}');
Expect(1, 2142, '\p{Script_Extensions=mandaic}', "");
Expect(0, 2142, '\p{^Script_Extensions=mandaic}', "");
Expect(0, 2142, '\P{Script_Extensions=mandaic}', "");
Expect(1, 2142, '\P{^Script_Extensions=mandaic}', "");
Expect(0, 2143, '\p{Script_Extensions=mandaic}', "");
Expect(1, 2143, '\p{^Script_Extensions=mandaic}', "");
Expect(1, 2143, '\P{Script_Extensions=mandaic}', "");
Expect(0, 2143, '\P{^Script_Extensions=mandaic}', "");
Expect(1, 2142, '\p{Script_Extensions=_ Mandaic}', "");
Expect(0, 2142, '\p{^Script_Extensions=_ Mandaic}', "");
Expect(0, 2142, '\P{Script_Extensions=_ Mandaic}', "");
Expect(1, 2142, '\P{^Script_Extensions=_ Mandaic}', "");
Expect(0, 2143, '\p{Script_Extensions=_ Mandaic}', "");
Expect(1, 2143, '\p{^Script_Extensions=_ Mandaic}', "");
Expect(1, 2143, '\P{Script_Extensions=_ Mandaic}', "");
Expect(0, 2143, '\P{^Script_Extensions=_ Mandaic}', "");
Error('\p{Scx=:=MAND}');
Error('\P{Scx=:=MAND}');
Expect(1, 2142, '\p{Scx=mand}', "");
Expect(0, 2142, '\p{^Scx=mand}', "");
Expect(0, 2142, '\P{Scx=mand}', "");
Expect(1, 2142, '\P{^Scx=mand}', "");
Expect(0, 2143, '\p{Scx=mand}', "");
Expect(1, 2143, '\p{^Scx=mand}', "");
Expect(1, 2143, '\P{Scx=mand}', "");
Expect(0, 2143, '\P{^Scx=mand}', "");
Expect(1, 2142, '\p{Scx= MAND}', "");
Expect(0, 2142, '\p{^Scx= MAND}', "");
Expect(0, 2142, '\P{Scx= MAND}', "");
Expect(1, 2142, '\P{^Scx= MAND}', "");
Expect(0, 2143, '\p{Scx= MAND}', "");
Expect(1, 2143, '\p{^Scx= MAND}', "");
Expect(1, 2143, '\P{Scx= MAND}', "");
Expect(0, 2143, '\P{^Scx= MAND}', "");
Error('\p{Is_Script_Extensions=_-MANDAIC:=}');
Error('\P{Is_Script_Extensions=_-MANDAIC:=}');
Expect(1, 2142, '\p{Is_Script_Extensions=mandaic}', "");
Expect(0, 2142, '\p{^Is_Script_Extensions=mandaic}', "");
Expect(0, 2142, '\P{Is_Script_Extensions=mandaic}', "");
Expect(1, 2142, '\P{^Is_Script_Extensions=mandaic}', "");
Expect(0, 2143, '\p{Is_Script_Extensions=mandaic}', "");
Expect(1, 2143, '\p{^Is_Script_Extensions=mandaic}', "");
Expect(1, 2143, '\P{Is_Script_Extensions=mandaic}', "");
Expect(0, 2143, '\P{^Is_Script_Extensions=mandaic}', "");
Expect(1, 2142, '\p{Is_Script_Extensions:   	 mandaic}', "");
Expect(0, 2142, '\p{^Is_Script_Extensions:   	 mandaic}', "");
Expect(0, 2142, '\P{Is_Script_Extensions:   	 mandaic}', "");
Expect(1, 2142, '\P{^Is_Script_Extensions:   	 mandaic}', "");
Expect(0, 2143, '\p{Is_Script_Extensions:   	 mandaic}', "");
Expect(1, 2143, '\p{^Is_Script_Extensions:   	 mandaic}', "");
Expect(1, 2143, '\P{Is_Script_Extensions:   	 mandaic}', "");
Expect(0, 2143, '\P{^Is_Script_Extensions:   	 mandaic}', "");
Error('\p{Is_Scx=_ mand:=}');
Error('\P{Is_Scx=_ mand:=}');
Expect(1, 2142, '\p{Is_Scx=mand}', "");
Expect(0, 2142, '\p{^Is_Scx=mand}', "");
Expect(0, 2142, '\P{Is_Scx=mand}', "");
Expect(1, 2142, '\P{^Is_Scx=mand}', "");
Expect(0, 2143, '\p{Is_Scx=mand}', "");
Expect(1, 2143, '\p{^Is_Scx=mand}', "");
Expect(1, 2143, '\P{Is_Scx=mand}', "");
Expect(0, 2143, '\P{^Is_Scx=mand}', "");
Expect(1, 2142, '\p{Is_Scx=		Mand}', "");
Expect(0, 2142, '\p{^Is_Scx=		Mand}', "");
Expect(0, 2142, '\P{Is_Scx=		Mand}', "");
Expect(1, 2142, '\P{^Is_Scx=		Mand}', "");
Expect(0, 2143, '\p{Is_Scx=		Mand}', "");
Expect(1, 2143, '\p{^Is_Scx=		Mand}', "");
Expect(1, 2143, '\P{Is_Scx=		Mand}', "");
Expect(0, 2143, '\P{^Is_Scx=		Mand}', "");
Error('\p{Script_Extensions= :=meroitic_cursive}');
Error('\P{Script_Extensions= :=meroitic_cursive}');
Expect(1, 68031, '\p{Script_Extensions=meroiticcursive}', "");
Expect(0, 68031, '\p{^Script_Extensions=meroiticcursive}', "");
Expect(0, 68031, '\P{Script_Extensions=meroiticcursive}', "");
Expect(1, 68031, '\P{^Script_Extensions=meroiticcursive}', "");
Expect(0, 68032, '\p{Script_Extensions=meroiticcursive}', "");
Expect(1, 68032, '\p{^Script_Extensions=meroiticcursive}', "");
Expect(1, 68032, '\P{Script_Extensions=meroiticcursive}', "");
Expect(0, 68032, '\P{^Script_Extensions=meroiticcursive}', "");
Expect(1, 68031, '\p{Script_Extensions= MEROITIC_Cursive}', "");
Expect(0, 68031, '\p{^Script_Extensions= MEROITIC_Cursive}', "");
Expect(0, 68031, '\P{Script_Extensions= MEROITIC_Cursive}', "");
Expect(1, 68031, '\P{^Script_Extensions= MEROITIC_Cursive}', "");
Expect(0, 68032, '\p{Script_Extensions= MEROITIC_Cursive}', "");
Expect(1, 68032, '\p{^Script_Extensions= MEROITIC_Cursive}', "");
Expect(1, 68032, '\P{Script_Extensions= MEROITIC_Cursive}', "");
Expect(0, 68032, '\P{^Script_Extensions= MEROITIC_Cursive}', "");
Error('\p{Scx:	merc:=}');
Error('\P{Scx:	merc:=}');
Expect(1, 68031, '\p{Scx=merc}', "");
Expect(0, 68031, '\p{^Scx=merc}', "");
Expect(0, 68031, '\P{Scx=merc}', "");
Expect(1, 68031, '\P{^Scx=merc}', "");
Expect(0, 68032, '\p{Scx=merc}', "");
Expect(1, 68032, '\p{^Scx=merc}', "");
Expect(1, 68032, '\P{Scx=merc}', "");
Expect(0, 68032, '\P{^Scx=merc}', "");
Expect(1, 68031, '\p{Scx= _merc}', "");
Expect(0, 68031, '\p{^Scx= _merc}', "");
Expect(0, 68031, '\P{Scx= _merc}', "");
Expect(1, 68031, '\P{^Scx= _merc}', "");
Expect(0, 68032, '\p{Scx= _merc}', "");
Expect(1, 68032, '\p{^Scx= _merc}', "");
Expect(1, 68032, '\P{Scx= _merc}', "");
Expect(0, 68032, '\P{^Scx= _merc}', "");
Error('\p{Is_Script_Extensions:   _ Meroitic_Cursive:=}');
Error('\P{Is_Script_Extensions:   _ Meroitic_Cursive:=}');
Expect(1, 68031, '\p{Is_Script_Extensions=meroiticcursive}', "");
Expect(0, 68031, '\p{^Is_Script_Extensions=meroiticcursive}', "");
Expect(0, 68031, '\P{Is_Script_Extensions=meroiticcursive}', "");
Expect(1, 68031, '\P{^Is_Script_Extensions=meroiticcursive}', "");
Expect(0, 68032, '\p{Is_Script_Extensions=meroiticcursive}', "");
Expect(1, 68032, '\p{^Is_Script_Extensions=meroiticcursive}', "");
Expect(1, 68032, '\P{Is_Script_Extensions=meroiticcursive}', "");
Expect(0, 68032, '\P{^Is_Script_Extensions=meroiticcursive}', "");
Expect(1, 68031, '\p{Is_Script_Extensions=	 MEROITIC_CURSIVE}', "");
Expect(0, 68031, '\p{^Is_Script_Extensions=	 MEROITIC_CURSIVE}', "");
Expect(0, 68031, '\P{Is_Script_Extensions=	 MEROITIC_CURSIVE}', "");
Expect(1, 68031, '\P{^Is_Script_Extensions=	 MEROITIC_CURSIVE}', "");
Expect(0, 68032, '\p{Is_Script_Extensions=	 MEROITIC_CURSIVE}', "");
Expect(1, 68032, '\p{^Is_Script_Extensions=	 MEROITIC_CURSIVE}', "");
Expect(1, 68032, '\P{Is_Script_Extensions=	 MEROITIC_CURSIVE}', "");
Expect(0, 68032, '\P{^Is_Script_Extensions=	 MEROITIC_CURSIVE}', "");
Error('\p{Is_Scx=/a/-	Merc}');
Error('\P{Is_Scx=/a/-	Merc}');
Expect(1, 68031, '\p{Is_Scx=merc}', "");
Expect(0, 68031, '\p{^Is_Scx=merc}', "");
Expect(0, 68031, '\P{Is_Scx=merc}', "");
Expect(1, 68031, '\P{^Is_Scx=merc}', "");
Expect(0, 68032, '\p{Is_Scx=merc}', "");
Expect(1, 68032, '\p{^Is_Scx=merc}', "");
Expect(1, 68032, '\P{Is_Scx=merc}', "");
Expect(0, 68032, '\P{^Is_Scx=merc}', "");
Expect(1, 68031, '\p{Is_Scx=- merc}', "");
Expect(0, 68031, '\p{^Is_Scx=- merc}', "");
Expect(0, 68031, '\P{Is_Scx=- merc}', "");
Expect(1, 68031, '\P{^Is_Scx=- merc}', "");
Expect(0, 68032, '\p{Is_Scx=- merc}', "");
Expect(1, 68032, '\p{^Is_Scx=- merc}', "");
Expect(1, 68032, '\P{Is_Scx=- merc}', "");
Expect(0, 68032, '\P{^Is_Scx=- merc}', "");
Error('\p{Script_Extensions=__MEROITIC_HIEROGLYPHS:=}');
Error('\P{Script_Extensions=__MEROITIC_HIEROGLYPHS:=}');
Expect(1, 67999, '\p{Script_Extensions=meroitichieroglyphs}', "");
Expect(0, 67999, '\p{^Script_Extensions=meroitichieroglyphs}', "");
Expect(0, 67999, '\P{Script_Extensions=meroitichieroglyphs}', "");
Expect(1, 67999, '\P{^Script_Extensions=meroitichieroglyphs}', "");
Expect(0, 68000, '\p{Script_Extensions=meroitichieroglyphs}', "");
Expect(1, 68000, '\p{^Script_Extensions=meroitichieroglyphs}', "");
Expect(1, 68000, '\P{Script_Extensions=meroitichieroglyphs}', "");
Expect(0, 68000, '\P{^Script_Extensions=meroitichieroglyphs}', "");
Expect(1, 67999, '\p{Script_Extensions= MEROITIC_hieroglyphs}', "");
Expect(0, 67999, '\p{^Script_Extensions= MEROITIC_hieroglyphs}', "");
Expect(0, 67999, '\P{Script_Extensions= MEROITIC_hieroglyphs}', "");
Expect(1, 67999, '\P{^Script_Extensions= MEROITIC_hieroglyphs}', "");
Expect(0, 68000, '\p{Script_Extensions= MEROITIC_hieroglyphs}', "");
Expect(1, 68000, '\p{^Script_Extensions= MEROITIC_hieroglyphs}', "");
Expect(1, 68000, '\P{Script_Extensions= MEROITIC_hieroglyphs}', "");
Expect(0, 68000, '\P{^Script_Extensions= MEROITIC_hieroglyphs}', "");
Error('\p{Scx=--Mero:=}');
Error('\P{Scx=--Mero:=}');
Expect(1, 67999, '\p{Scx=mero}', "");
Expect(0, 67999, '\p{^Scx=mero}', "");
Expect(0, 67999, '\P{Scx=mero}', "");
Expect(1, 67999, '\P{^Scx=mero}', "");
Expect(0, 68000, '\p{Scx=mero}', "");
Expect(1, 68000, '\p{^Scx=mero}', "");
Expect(1, 68000, '\P{Scx=mero}', "");
Expect(0, 68000, '\P{^Scx=mero}', "");
Expect(1, 67999, '\p{Scx=-Mero}', "");
Expect(0, 67999, '\p{^Scx=-Mero}', "");
Expect(0, 67999, '\P{Scx=-Mero}', "");
Expect(1, 67999, '\P{^Scx=-Mero}', "");
Expect(0, 68000, '\p{Scx=-Mero}', "");
Expect(1, 68000, '\p{^Scx=-Mero}', "");
Expect(1, 68000, '\P{Scx=-Mero}', "");
Expect(0, 68000, '\P{^Scx=-Mero}', "");
Error('\p{Is_Script_Extensions=- Meroitic_HIEROGLYPHS/a/}');
Error('\P{Is_Script_Extensions=- Meroitic_HIEROGLYPHS/a/}');
Expect(1, 67999, '\p{Is_Script_Extensions=meroitichieroglyphs}', "");
Expect(0, 67999, '\p{^Is_Script_Extensions=meroitichieroglyphs}', "");
Expect(0, 67999, '\P{Is_Script_Extensions=meroitichieroglyphs}', "");
Expect(1, 67999, '\P{^Is_Script_Extensions=meroitichieroglyphs}', "");
Expect(0, 68000, '\p{Is_Script_Extensions=meroitichieroglyphs}', "");
Expect(1, 68000, '\p{^Is_Script_Extensions=meroitichieroglyphs}', "");
Expect(1, 68000, '\P{Is_Script_Extensions=meroitichieroglyphs}', "");
Expect(0, 68000, '\P{^Is_Script_Extensions=meroitichieroglyphs}', "");
Expect(1, 67999, '\p{Is_Script_Extensions:   _	meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\p{^Is_Script_Extensions:   _	meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\P{Is_Script_Extensions:   _	meroitic_Hieroglyphs}', "");
Expect(1, 67999, '\P{^Is_Script_Extensions:   _	meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\p{Is_Script_Extensions:   _	meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\p{^Is_Script_Extensions:   _	meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\P{Is_Script_Extensions:   _	meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\P{^Is_Script_Extensions:   _	meroitic_Hieroglyphs}', "");
Error('\p{Is_Scx=_:=mero}');
Error('\P{Is_Scx=_:=mero}');
Expect(1, 67999, '\p{Is_Scx=mero}', "");
Expect(0, 67999, '\p{^Is_Scx=mero}', "");
Expect(0, 67999, '\P{Is_Scx=mero}', "");
Expect(1, 67999, '\P{^Is_Scx=mero}', "");
Expect(0, 68000, '\p{Is_Scx=mero}', "");
Expect(1, 68000, '\p{^Is_Scx=mero}', "");
Expect(1, 68000, '\P{Is_Scx=mero}', "");
Expect(0, 68000, '\P{^Is_Scx=mero}', "");
Expect(1, 67999, '\p{Is_Scx=_	MERO}', "");
Expect(0, 67999, '\p{^Is_Scx=_	MERO}', "");
Expect(0, 67999, '\P{Is_Scx=_	MERO}', "");
Expect(1, 67999, '\P{^Is_Scx=_	MERO}', "");
Expect(0, 68000, '\p{Is_Scx=_	MERO}', "");
Expect(1, 68000, '\p{^Is_Scx=_	MERO}', "");
Expect(1, 68000, '\P{Is_Scx=_	MERO}', "");
Expect(0, 68000, '\P{^Is_Scx=_	MERO}', "");
Error('\p{Script_Extensions:-/a/Malayalam}');
Error('\P{Script_Extensions:-/a/Malayalam}');
Expect(1, 3455, '\p{Script_Extensions: malayalam}', "");
Expect(0, 3455, '\p{^Script_Extensions: malayalam}', "");
Expect(0, 3455, '\P{Script_Extensions: malayalam}', "");
Expect(1, 3455, '\P{^Script_Extensions: malayalam}', "");
Expect(0, 3456, '\p{Script_Extensions: malayalam}', "");
Expect(1, 3456, '\p{^Script_Extensions: malayalam}', "");
Expect(1, 3456, '\P{Script_Extensions: malayalam}', "");
Expect(0, 3456, '\P{^Script_Extensions: malayalam}', "");
Expect(1, 3455, '\p{Script_Extensions=	Malayalam}', "");
Expect(0, 3455, '\p{^Script_Extensions=	Malayalam}', "");
Expect(0, 3455, '\P{Script_Extensions=	Malayalam}', "");
Expect(1, 3455, '\P{^Script_Extensions=	Malayalam}', "");
Expect(0, 3456, '\p{Script_Extensions=	Malayalam}', "");
Expect(1, 3456, '\p{^Script_Extensions=	Malayalam}', "");
Expect(1, 3456, '\P{Script_Extensions=	Malayalam}', "");
Expect(0, 3456, '\P{^Script_Extensions=	Malayalam}', "");
Error('\p{Scx=_	MLYM/a/}');
Error('\P{Scx=_	MLYM/a/}');
Expect(1, 3455, '\p{Scx=mlym}', "");
Expect(0, 3455, '\p{^Scx=mlym}', "");
Expect(0, 3455, '\P{Scx=mlym}', "");
Expect(1, 3455, '\P{^Scx=mlym}', "");
Expect(0, 3456, '\p{Scx=mlym}', "");
Expect(1, 3456, '\p{^Scx=mlym}', "");
Expect(1, 3456, '\P{Scx=mlym}', "");
Expect(0, 3456, '\P{^Scx=mlym}', "");
Expect(1, 3455, '\p{Scx=-	Mlym}', "");
Expect(0, 3455, '\p{^Scx=-	Mlym}', "");
Expect(0, 3455, '\P{Scx=-	Mlym}', "");
Expect(1, 3455, '\P{^Scx=-	Mlym}', "");
Expect(0, 3456, '\p{Scx=-	Mlym}', "");
Expect(1, 3456, '\p{^Scx=-	Mlym}', "");
Expect(1, 3456, '\P{Scx=-	Mlym}', "");
Expect(0, 3456, '\P{^Scx=-	Mlym}', "");
Error('\p{Is_Script_Extensions=:=__MALAYALAM}');
Error('\P{Is_Script_Extensions=:=__MALAYALAM}');
Expect(1, 3455, '\p{Is_Script_Extensions=malayalam}', "");
Expect(0, 3455, '\p{^Is_Script_Extensions=malayalam}', "");
Expect(0, 3455, '\P{Is_Script_Extensions=malayalam}', "");
Expect(1, 3455, '\P{^Is_Script_Extensions=malayalam}', "");
Expect(0, 3456, '\p{Is_Script_Extensions=malayalam}', "");
Expect(1, 3456, '\p{^Is_Script_Extensions=malayalam}', "");
Expect(1, 3456, '\P{Is_Script_Extensions=malayalam}', "");
Expect(0, 3456, '\P{^Is_Script_Extensions=malayalam}', "");
Expect(1, 3455, '\p{Is_Script_Extensions=--Malayalam}', "");
Expect(0, 3455, '\p{^Is_Script_Extensions=--Malayalam}', "");
Expect(0, 3455, '\P{Is_Script_Extensions=--Malayalam}', "");
Expect(1, 3455, '\P{^Is_Script_Extensions=--Malayalam}', "");
Expect(0, 3456, '\p{Is_Script_Extensions=--Malayalam}', "");
Expect(1, 3456, '\p{^Is_Script_Extensions=--Malayalam}', "");
Expect(1, 3456, '\P{Is_Script_Extensions=--Malayalam}', "");
Expect(0, 3456, '\P{^Is_Script_Extensions=--Malayalam}', "");
Error('\p{Is_Scx::=	Mlym}');
Error('\P{Is_Scx::=	Mlym}');
Expect(1, 3455, '\p{Is_Scx:mlym}', "");
Expect(0, 3455, '\p{^Is_Scx:mlym}', "");
Expect(0, 3455, '\P{Is_Scx:mlym}', "");
Expect(1, 3455, '\P{^Is_Scx:mlym}', "");
Expect(0, 3456, '\p{Is_Scx:mlym}', "");
Expect(1, 3456, '\p{^Is_Scx:mlym}', "");
Expect(1, 3456, '\P{Is_Scx:mlym}', "");
Expect(0, 3456, '\P{^Is_Scx:mlym}', "");
Expect(1, 3455, '\p{Is_Scx=_-mlym}', "");
Expect(0, 3455, '\p{^Is_Scx=_-mlym}', "");
Expect(0, 3455, '\P{Is_Scx=_-mlym}', "");
Expect(1, 3455, '\P{^Is_Scx=_-mlym}', "");
Expect(0, 3456, '\p{Is_Scx=_-mlym}', "");
Expect(1, 3456, '\p{^Is_Scx=_-mlym}', "");
Expect(1, 3456, '\P{Is_Scx=_-mlym}', "");
Expect(0, 3456, '\P{^Is_Scx=_-mlym}', "");
Error('\p{Script_Extensions=- mongolian/a/}');
Error('\P{Script_Extensions=- mongolian/a/}');
Expect(1, 6314, '\p{Script_Extensions:	mongolian}', "");
Expect(0, 6314, '\p{^Script_Extensions:	mongolian}', "");
Expect(0, 6314, '\P{Script_Extensions:	mongolian}', "");
Expect(1, 6314, '\P{^Script_Extensions:	mongolian}', "");
Expect(0, 6315, '\p{Script_Extensions:	mongolian}', "");
Expect(1, 6315, '\p{^Script_Extensions:	mongolian}', "");
Expect(1, 6315, '\P{Script_Extensions:	mongolian}', "");
Expect(0, 6315, '\P{^Script_Extensions:	mongolian}', "");
Expect(1, 6314, '\p{Script_Extensions= _Mongolian}', "");
Expect(0, 6314, '\p{^Script_Extensions= _Mongolian}', "");
Expect(0, 6314, '\P{Script_Extensions= _Mongolian}', "");
Expect(1, 6314, '\P{^Script_Extensions= _Mongolian}', "");
Expect(0, 6315, '\p{Script_Extensions= _Mongolian}', "");
Expect(1, 6315, '\p{^Script_Extensions= _Mongolian}', "");
Expect(1, 6315, '\P{Script_Extensions= _Mongolian}', "");
Expect(0, 6315, '\P{^Script_Extensions= _Mongolian}', "");
Error('\p{Scx=/a/mong}');
Error('\P{Scx=/a/mong}');
Expect(1, 6314, '\p{Scx=mong}', "");
Expect(0, 6314, '\p{^Scx=mong}', "");
Expect(0, 6314, '\P{Scx=mong}', "");
Expect(1, 6314, '\P{^Scx=mong}', "");
Expect(0, 6315, '\p{Scx=mong}', "");
Expect(1, 6315, '\p{^Scx=mong}', "");
Expect(1, 6315, '\P{Scx=mong}', "");
Expect(0, 6315, '\P{^Scx=mong}', "");
Expect(1, 6314, '\p{Scx:_Mong}', "");
Expect(0, 6314, '\p{^Scx:_Mong}', "");
Expect(0, 6314, '\P{Scx:_Mong}', "");
Expect(1, 6314, '\P{^Scx:_Mong}', "");
Expect(0, 6315, '\p{Scx:_Mong}', "");
Expect(1, 6315, '\p{^Scx:_Mong}', "");
Expect(1, 6315, '\P{Scx:_Mong}', "");
Expect(0, 6315, '\P{^Scx:_Mong}', "");
Error('\p{Is_Script_Extensions=/a/	_MONGOLIAN}');
Error('\P{Is_Script_Extensions=/a/	_MONGOLIAN}');
Expect(1, 6314, '\p{Is_Script_Extensions=mongolian}', "");
Expect(0, 6314, '\p{^Is_Script_Extensions=mongolian}', "");
Expect(0, 6314, '\P{Is_Script_Extensions=mongolian}', "");
Expect(1, 6314, '\P{^Is_Script_Extensions=mongolian}', "");
Expect(0, 6315, '\p{Is_Script_Extensions=mongolian}', "");
Expect(1, 6315, '\p{^Is_Script_Extensions=mongolian}', "");
Expect(1, 6315, '\P{Is_Script_Extensions=mongolian}', "");
Expect(0, 6315, '\P{^Is_Script_Extensions=mongolian}', "");
Expect(1, 6314, '\p{Is_Script_Extensions=-	MONGOLIAN}', "");
Expect(0, 6314, '\p{^Is_Script_Extensions=-	MONGOLIAN}', "");
Expect(0, 6314, '\P{Is_Script_Extensions=-	MONGOLIAN}', "");
Expect(1, 6314, '\P{^Is_Script_Extensions=-	MONGOLIAN}', "");
Expect(0, 6315, '\p{Is_Script_Extensions=-	MONGOLIAN}', "");
Expect(1, 6315, '\p{^Is_Script_Extensions=-	MONGOLIAN}', "");
Expect(1, 6315, '\P{Is_Script_Extensions=-	MONGOLIAN}', "");
Expect(0, 6315, '\P{^Is_Script_Extensions=-	MONGOLIAN}', "");
Error('\p{Is_Scx=/a/	Mong}');
Error('\P{Is_Scx=/a/	Mong}');
Expect(1, 6314, '\p{Is_Scx=mong}', "");
Expect(0, 6314, '\p{^Is_Scx=mong}', "");
Expect(0, 6314, '\P{Is_Scx=mong}', "");
Expect(1, 6314, '\P{^Is_Scx=mong}', "");
Expect(0, 6315, '\p{Is_Scx=mong}', "");
Expect(1, 6315, '\p{^Is_Scx=mong}', "");
Expect(1, 6315, '\P{Is_Scx=mong}', "");
Expect(0, 6315, '\P{^Is_Scx=mong}', "");
Expect(1, 6314, '\p{Is_Scx=Mong}', "");
Expect(0, 6314, '\p{^Is_Scx=Mong}', "");
Expect(0, 6314, '\P{Is_Scx=Mong}', "");
Expect(1, 6314, '\P{^Is_Scx=Mong}', "");
Expect(0, 6315, '\p{Is_Scx=Mong}', "");
Expect(1, 6315, '\p{^Is_Scx=Mong}', "");
Expect(1, 6315, '\P{Is_Scx=Mong}', "");
Expect(0, 6315, '\P{^Is_Scx=Mong}', "");
Error('\p{Script_Extensions=	:=MEETEI_Mayek}');
Error('\P{Script_Extensions=	:=MEETEI_Mayek}');
Expect(1, 44025, '\p{Script_Extensions=meeteimayek}', "");
Expect(0, 44025, '\p{^Script_Extensions=meeteimayek}', "");
Expect(0, 44025, '\P{Script_Extensions=meeteimayek}', "");
Expect(1, 44025, '\P{^Script_Extensions=meeteimayek}', "");
Expect(0, 44026, '\p{Script_Extensions=meeteimayek}', "");
Expect(1, 44026, '\p{^Script_Extensions=meeteimayek}', "");
Expect(1, 44026, '\P{Script_Extensions=meeteimayek}', "");
Expect(0, 44026, '\P{^Script_Extensions=meeteimayek}', "");
Expect(1, 44025, '\p{Script_Extensions=_ meetei_Mayek}', "");
Expect(0, 44025, '\p{^Script_Extensions=_ meetei_Mayek}', "");
Expect(0, 44025, '\P{Script_Extensions=_ meetei_Mayek}', "");
Expect(1, 44025, '\P{^Script_Extensions=_ meetei_Mayek}', "");
Expect(0, 44026, '\p{Script_Extensions=_ meetei_Mayek}', "");
Expect(1, 44026, '\p{^Script_Extensions=_ meetei_Mayek}', "");
Expect(1, 44026, '\P{Script_Extensions=_ meetei_Mayek}', "");
Expect(0, 44026, '\P{^Script_Extensions=_ meetei_Mayek}', "");
Error('\p{Scx=/a/_	Mtei}');
Error('\P{Scx=/a/_	Mtei}');
Expect(1, 44025, '\p{Scx=mtei}', "");
Expect(0, 44025, '\p{^Scx=mtei}', "");
Expect(0, 44025, '\P{Scx=mtei}', "");
Expect(1, 44025, '\P{^Scx=mtei}', "");
Expect(0, 44026, '\p{Scx=mtei}', "");
Expect(1, 44026, '\p{^Scx=mtei}', "");
Expect(1, 44026, '\P{Scx=mtei}', "");
Expect(0, 44026, '\P{^Scx=mtei}', "");
Expect(1, 44025, '\p{Scx= _mtei}', "");
Expect(0, 44025, '\p{^Scx= _mtei}', "");
Expect(0, 44025, '\P{Scx= _mtei}', "");
Expect(1, 44025, '\P{^Scx= _mtei}', "");
Expect(0, 44026, '\p{Scx= _mtei}', "");
Expect(1, 44026, '\p{^Scx= _mtei}', "");
Expect(1, 44026, '\P{Scx= _mtei}', "");
Expect(0, 44026, '\P{^Scx= _mtei}', "");
Error('\p{Is_Script_Extensions=_MEETEI_Mayek:=}');
Error('\P{Is_Script_Extensions=_MEETEI_Mayek:=}');
Expect(1, 44025, '\p{Is_Script_Extensions=meeteimayek}', "");
Expect(0, 44025, '\p{^Is_Script_Extensions=meeteimayek}', "");
Expect(0, 44025, '\P{Is_Script_Extensions=meeteimayek}', "");
Expect(1, 44025, '\P{^Is_Script_Extensions=meeteimayek}', "");
Expect(0, 44026, '\p{Is_Script_Extensions=meeteimayek}', "");
Expect(1, 44026, '\p{^Is_Script_Extensions=meeteimayek}', "");
Expect(1, 44026, '\P{Is_Script_Extensions=meeteimayek}', "");
Expect(0, 44026, '\P{^Is_Script_Extensions=meeteimayek}', "");
Expect(1, 44025, '\p{Is_Script_Extensions=	meetei_Mayek}', "");
Expect(0, 44025, '\p{^Is_Script_Extensions=	meetei_Mayek}', "");
Expect(0, 44025, '\P{Is_Script_Extensions=	meetei_Mayek}', "");
Expect(1, 44025, '\P{^Is_Script_Extensions=	meetei_Mayek}', "");
Expect(0, 44026, '\p{Is_Script_Extensions=	meetei_Mayek}', "");
Expect(1, 44026, '\p{^Is_Script_Extensions=	meetei_Mayek}', "");
Expect(1, 44026, '\P{Is_Script_Extensions=	meetei_Mayek}', "");
Expect(0, 44026, '\P{^Is_Script_Extensions=	meetei_Mayek}', "");
Error('\p{Is_Scx= -Mtei:=}');
Error('\P{Is_Scx= -Mtei:=}');
Expect(1, 44025, '\p{Is_Scx:	mtei}', "");
Expect(0, 44025, '\p{^Is_Scx:	mtei}', "");
Expect(0, 44025, '\P{Is_Scx:	mtei}', "");
Expect(1, 44025, '\P{^Is_Scx:	mtei}', "");
Expect(0, 44026, '\p{Is_Scx:	mtei}', "");
Expect(1, 44026, '\p{^Is_Scx:	mtei}', "");
Expect(1, 44026, '\P{Is_Scx:	mtei}', "");
Expect(0, 44026, '\P{^Is_Scx:	mtei}', "");
Expect(1, 44025, '\p{Is_Scx=  Mtei}', "");
Expect(0, 44025, '\p{^Is_Scx=  Mtei}', "");
Expect(0, 44025, '\P{Is_Scx=  Mtei}', "");
Expect(1, 44025, '\P{^Is_Scx=  Mtei}', "");
Expect(0, 44026, '\p{Is_Scx=  Mtei}', "");
Expect(1, 44026, '\p{^Is_Scx=  Mtei}', "");
Expect(1, 44026, '\P{Is_Scx=  Mtei}', "");
Expect(0, 44026, '\P{^Is_Scx=  Mtei}', "");
Error('\p{Script_Extensions=- myanmar/a/}');
Error('\P{Script_Extensions=- myanmar/a/}');
Expect(1, 43643, '\p{Script_Extensions=myanmar}', "");
Expect(0, 43643, '\p{^Script_Extensions=myanmar}', "");
Expect(0, 43643, '\P{Script_Extensions=myanmar}', "");
Expect(1, 43643, '\P{^Script_Extensions=myanmar}', "");
Expect(0, 43644, '\p{Script_Extensions=myanmar}', "");
Expect(1, 43644, '\p{^Script_Extensions=myanmar}', "");
Expect(1, 43644, '\P{Script_Extensions=myanmar}', "");
Expect(0, 43644, '\P{^Script_Extensions=myanmar}', "");
Expect(1, 43643, '\p{Script_Extensions=	Myanmar}', "");
Expect(0, 43643, '\p{^Script_Extensions=	Myanmar}', "");
Expect(0, 43643, '\P{Script_Extensions=	Myanmar}', "");
Expect(1, 43643, '\P{^Script_Extensions=	Myanmar}', "");
Expect(0, 43644, '\p{Script_Extensions=	Myanmar}', "");
Expect(1, 43644, '\p{^Script_Extensions=	Myanmar}', "");
Expect(1, 43644, '\P{Script_Extensions=	Myanmar}', "");
Expect(0, 43644, '\P{^Script_Extensions=	Myanmar}', "");
Error('\p{Scx=-_Mymr:=}');
Error('\P{Scx=-_Mymr:=}');
Expect(1, 43643, '\p{Scx=mymr}', "");
Expect(0, 43643, '\p{^Scx=mymr}', "");
Expect(0, 43643, '\P{Scx=mymr}', "");
Expect(1, 43643, '\P{^Scx=mymr}', "");
Expect(0, 43644, '\p{Scx=mymr}', "");
Expect(1, 43644, '\p{^Scx=mymr}', "");
Expect(1, 43644, '\P{Scx=mymr}', "");
Expect(0, 43644, '\P{^Scx=mymr}', "");
Expect(1, 43643, '\p{Scx=_ MYMR}', "");
Expect(0, 43643, '\p{^Scx=_ MYMR}', "");
Expect(0, 43643, '\P{Scx=_ MYMR}', "");
Expect(1, 43643, '\P{^Scx=_ MYMR}', "");
Expect(0, 43644, '\p{Scx=_ MYMR}', "");
Expect(1, 43644, '\p{^Scx=_ MYMR}', "");
Expect(1, 43644, '\P{Scx=_ MYMR}', "");
Expect(0, 43644, '\P{^Scx=_ MYMR}', "");
Error('\p{Is_Script_Extensions=/a/_myanmar}');
Error('\P{Is_Script_Extensions=/a/_myanmar}');
Expect(1, 43643, '\p{Is_Script_Extensions=myanmar}', "");
Expect(0, 43643, '\p{^Is_Script_Extensions=myanmar}', "");
Expect(0, 43643, '\P{Is_Script_Extensions=myanmar}', "");
Expect(1, 43643, '\P{^Is_Script_Extensions=myanmar}', "");
Expect(0, 43644, '\p{Is_Script_Extensions=myanmar}', "");
Expect(1, 43644, '\p{^Is_Script_Extensions=myanmar}', "");
Expect(1, 43644, '\P{Is_Script_Extensions=myanmar}', "");
Expect(0, 43644, '\P{^Is_Script_Extensions=myanmar}', "");
Expect(1, 43643, '\p{Is_Script_Extensions= -myanmar}', "");
Expect(0, 43643, '\p{^Is_Script_Extensions= -myanmar}', "");
Expect(0, 43643, '\P{Is_Script_Extensions= -myanmar}', "");
Expect(1, 43643, '\P{^Is_Script_Extensions= -myanmar}', "");
Expect(0, 43644, '\p{Is_Script_Extensions= -myanmar}', "");
Expect(1, 43644, '\p{^Is_Script_Extensions= -myanmar}', "");
Expect(1, 43644, '\P{Is_Script_Extensions= -myanmar}', "");
Expect(0, 43644, '\P{^Is_Script_Extensions= -myanmar}', "");
Error('\p{Is_Scx=/a/mymr}');
Error('\P{Is_Scx=/a/mymr}');
Expect(1, 43643, '\p{Is_Scx=mymr}', "");
Expect(0, 43643, '\p{^Is_Scx=mymr}', "");
Expect(0, 43643, '\P{Is_Scx=mymr}', "");
Expect(1, 43643, '\P{^Is_Scx=mymr}', "");
Expect(0, 43644, '\p{Is_Scx=mymr}', "");
Expect(1, 43644, '\p{^Is_Scx=mymr}', "");
Expect(1, 43644, '\P{Is_Scx=mymr}', "");
Expect(0, 43644, '\P{^Is_Scx=mymr}', "");
Expect(1, 43643, '\p{Is_Scx:	-mymr}', "");
Expect(0, 43643, '\p{^Is_Scx:	-mymr}', "");
Expect(0, 43643, '\P{Is_Scx:	-mymr}', "");
Expect(1, 43643, '\P{^Is_Scx:	-mymr}', "");
Expect(0, 43644, '\p{Is_Scx:	-mymr}', "");
Expect(1, 43644, '\p{^Is_Scx:	-mymr}', "");
Expect(1, 43644, '\P{Is_Scx:	-mymr}', "");
Expect(0, 43644, '\P{^Is_Scx:	-mymr}', "");
Error('\p{Script_Extensions=/a/__Nko}');
Error('\P{Script_Extensions=/a/__Nko}');
Expect(1, 2042, '\p{Script_Extensions=nko}', "");
Expect(0, 2042, '\p{^Script_Extensions=nko}', "");
Expect(0, 2042, '\P{Script_Extensions=nko}', "");
Expect(1, 2042, '\P{^Script_Extensions=nko}', "");
Expect(0, 2043, '\p{Script_Extensions=nko}', "");
Expect(1, 2043, '\p{^Script_Extensions=nko}', "");
Expect(1, 2043, '\P{Script_Extensions=nko}', "");
Expect(0, 2043, '\P{^Script_Extensions=nko}', "");
Expect(1, 2042, '\p{Script_Extensions=_NKO}', "");
Expect(0, 2042, '\p{^Script_Extensions=_NKO}', "");
Expect(0, 2042, '\P{Script_Extensions=_NKO}', "");
Expect(1, 2042, '\P{^Script_Extensions=_NKO}', "");
Expect(0, 2043, '\p{Script_Extensions=_NKO}', "");
Expect(1, 2043, '\p{^Script_Extensions=_NKO}', "");
Expect(1, 2043, '\P{Script_Extensions=_NKO}', "");
Expect(0, 2043, '\P{^Script_Extensions=_NKO}', "");
Error('\p{Scx=/a/- Nkoo}');
Error('\P{Scx=/a/- Nkoo}');
Expect(1, 2042, '\p{Scx=nkoo}', "");
Expect(0, 2042, '\p{^Scx=nkoo}', "");
Expect(0, 2042, '\P{Scx=nkoo}', "");
Expect(1, 2042, '\P{^Scx=nkoo}', "");
Expect(0, 2043, '\p{Scx=nkoo}', "");
Expect(1, 2043, '\p{^Scx=nkoo}', "");
Expect(1, 2043, '\P{Scx=nkoo}', "");
Expect(0, 2043, '\P{^Scx=nkoo}', "");
Expect(1, 2042, '\p{Scx=  Nkoo}', "");
Expect(0, 2042, '\p{^Scx=  Nkoo}', "");
Expect(0, 2042, '\P{Scx=  Nkoo}', "");
Expect(1, 2042, '\P{^Scx=  Nkoo}', "");
Expect(0, 2043, '\p{Scx=  Nkoo}', "");
Expect(1, 2043, '\p{^Scx=  Nkoo}', "");
Expect(1, 2043, '\P{Scx=  Nkoo}', "");
Expect(0, 2043, '\P{^Scx=  Nkoo}', "");
Error('\p{Is_Script_Extensions=/a/Nko}');
Error('\P{Is_Script_Extensions=/a/Nko}');
Expect(1, 2042, '\p{Is_Script_Extensions=nko}', "");
Expect(0, 2042, '\p{^Is_Script_Extensions=nko}', "");
Expect(0, 2042, '\P{Is_Script_Extensions=nko}', "");
Expect(1, 2042, '\P{^Is_Script_Extensions=nko}', "");
Expect(0, 2043, '\p{Is_Script_Extensions=nko}', "");
Expect(1, 2043, '\p{^Is_Script_Extensions=nko}', "");
Expect(1, 2043, '\P{Is_Script_Extensions=nko}', "");
Expect(0, 2043, '\P{^Is_Script_Extensions=nko}', "");
Expect(1, 2042, '\p{Is_Script_Extensions=	Nko}', "");
Expect(0, 2042, '\p{^Is_Script_Extensions=	Nko}', "");
Expect(0, 2042, '\P{Is_Script_Extensions=	Nko}', "");
Expect(1, 2042, '\P{^Is_Script_Extensions=	Nko}', "");
Expect(0, 2043, '\p{Is_Script_Extensions=	Nko}', "");
Expect(1, 2043, '\p{^Is_Script_Extensions=	Nko}', "");
Expect(1, 2043, '\P{Is_Script_Extensions=	Nko}', "");
Expect(0, 2043, '\P{^Is_Script_Extensions=	Nko}', "");
Error('\p{Is_Scx= :=Nkoo}');
Error('\P{Is_Scx= :=Nkoo}');
Expect(1, 2042, '\p{Is_Scx=nkoo}', "");
Expect(0, 2042, '\p{^Is_Scx=nkoo}', "");
Expect(0, 2042, '\P{Is_Scx=nkoo}', "");
Expect(1, 2042, '\P{^Is_Scx=nkoo}', "");
Expect(0, 2043, '\p{Is_Scx=nkoo}', "");
Expect(1, 2043, '\p{^Is_Scx=nkoo}', "");
Expect(1, 2043, '\P{Is_Scx=nkoo}', "");
Expect(0, 2043, '\P{^Is_Scx=nkoo}', "");
Expect(1, 2042, '\p{Is_Scx=_-Nkoo}', "");
Expect(0, 2042, '\p{^Is_Scx=_-Nkoo}', "");
Expect(0, 2042, '\P{Is_Scx=_-Nkoo}', "");
Expect(1, 2042, '\P{^Is_Scx=_-Nkoo}', "");
Expect(0, 2043, '\p{Is_Scx=_-Nkoo}', "");
Expect(1, 2043, '\p{^Is_Scx=_-Nkoo}', "");
Expect(1, 2043, '\P{Is_Scx=_-Nkoo}', "");
Expect(0, 2043, '\P{^Is_Scx=_-Nkoo}', "");
Error('\p{Script_Extensions=-_OGHAM/a/}');
Error('\P{Script_Extensions=-_OGHAM/a/}');
Expect(1, 5788, '\p{Script_Extensions: ogham}', "");
Expect(0, 5788, '\p{^Script_Extensions: ogham}', "");
Expect(0, 5788, '\P{Script_Extensions: ogham}', "");
Expect(1, 5788, '\P{^Script_Extensions: ogham}', "");
Expect(0, 5789, '\p{Script_Extensions: ogham}', "");
Expect(1, 5789, '\p{^Script_Extensions: ogham}', "");
Expect(1, 5789, '\P{Script_Extensions: ogham}', "");
Expect(0, 5789, '\P{^Script_Extensions: ogham}', "");
Expect(1, 5788, '\p{Script_Extensions=_	OGHAM}', "");
Expect(0, 5788, '\p{^Script_Extensions=_	OGHAM}', "");
Expect(0, 5788, '\P{Script_Extensions=_	OGHAM}', "");
Expect(1, 5788, '\P{^Script_Extensions=_	OGHAM}', "");
Expect(0, 5789, '\p{Script_Extensions=_	OGHAM}', "");
Expect(1, 5789, '\p{^Script_Extensions=_	OGHAM}', "");
Expect(1, 5789, '\P{Script_Extensions=_	OGHAM}', "");
Expect(0, 5789, '\P{^Script_Extensions=_	OGHAM}', "");
Error('\p{Scx=:= -Ogam}');
Error('\P{Scx=:= -Ogam}');
Expect(1, 5788, '\p{Scx=ogam}', "");
Expect(0, 5788, '\p{^Scx=ogam}', "");
Expect(0, 5788, '\P{Scx=ogam}', "");
Expect(1, 5788, '\P{^Scx=ogam}', "");
Expect(0, 5789, '\p{Scx=ogam}', "");
Expect(1, 5789, '\p{^Scx=ogam}', "");
Expect(1, 5789, '\P{Scx=ogam}', "");
Expect(0, 5789, '\P{^Scx=ogam}', "");
Expect(1, 5788, '\p{Scx=	 OGAM}', "");
Expect(0, 5788, '\p{^Scx=	 OGAM}', "");
Expect(0, 5788, '\P{Scx=	 OGAM}', "");
Expect(1, 5788, '\P{^Scx=	 OGAM}', "");
Expect(0, 5789, '\p{Scx=	 OGAM}', "");
Expect(1, 5789, '\p{^Scx=	 OGAM}', "");
Expect(1, 5789, '\P{Scx=	 OGAM}', "");
Expect(0, 5789, '\P{^Scx=	 OGAM}', "");
Error('\p{Is_Script_Extensions=- Ogham:=}');
Error('\P{Is_Script_Extensions=- Ogham:=}');
Expect(1, 5788, '\p{Is_Script_Extensions=ogham}', "");
Expect(0, 5788, '\p{^Is_Script_Extensions=ogham}', "");
Expect(0, 5788, '\P{Is_Script_Extensions=ogham}', "");
Expect(1, 5788, '\P{^Is_Script_Extensions=ogham}', "");
Expect(0, 5789, '\p{Is_Script_Extensions=ogham}', "");
Expect(1, 5789, '\p{^Is_Script_Extensions=ogham}', "");
Expect(1, 5789, '\P{Is_Script_Extensions=ogham}', "");
Expect(0, 5789, '\P{^Is_Script_Extensions=ogham}', "");
Expect(1, 5788, '\p{Is_Script_Extensions=_OGHAM}', "");
Expect(0, 5788, '\p{^Is_Script_Extensions=_OGHAM}', "");
Expect(0, 5788, '\P{Is_Script_Extensions=_OGHAM}', "");
Expect(1, 5788, '\P{^Is_Script_Extensions=_OGHAM}', "");
Expect(0, 5789, '\p{Is_Script_Extensions=_OGHAM}', "");
Expect(1, 5789, '\p{^Is_Script_Extensions=_OGHAM}', "");
Expect(1, 5789, '\P{Is_Script_Extensions=_OGHAM}', "");
Expect(0, 5789, '\P{^Is_Script_Extensions=_OGHAM}', "");
Error('\p{Is_Scx=_:=Ogam}');
Error('\P{Is_Scx=_:=Ogam}');
Expect(1, 5788, '\p{Is_Scx=ogam}', "");
Expect(0, 5788, '\p{^Is_Scx=ogam}', "");
Expect(0, 5788, '\P{Is_Scx=ogam}', "");
Expect(1, 5788, '\P{^Is_Scx=ogam}', "");
Expect(0, 5789, '\p{Is_Scx=ogam}', "");
Expect(1, 5789, '\p{^Is_Scx=ogam}', "");
Expect(1, 5789, '\P{Is_Scx=ogam}', "");
Expect(0, 5789, '\P{^Is_Scx=ogam}', "");
Expect(1, 5788, '\p{Is_Scx= 	Ogam}', "");
Expect(0, 5788, '\p{^Is_Scx= 	Ogam}', "");
Expect(0, 5788, '\P{Is_Scx= 	Ogam}', "");
Expect(1, 5788, '\P{^Is_Scx= 	Ogam}', "");
Expect(0, 5789, '\p{Is_Scx= 	Ogam}', "");
Expect(1, 5789, '\p{^Is_Scx= 	Ogam}', "");
Expect(1, 5789, '\P{Is_Scx= 	Ogam}', "");
Expect(0, 5789, '\P{^Is_Scx= 	Ogam}', "");
Error('\p{Script_Extensions=_	Ol_CHIKI/a/}');
Error('\P{Script_Extensions=_	Ol_CHIKI/a/}');
Expect(1, 7295, '\p{Script_Extensions=olchiki}', "");
Expect(0, 7295, '\p{^Script_Extensions=olchiki}', "");
Expect(0, 7295, '\P{Script_Extensions=olchiki}', "");
Expect(1, 7295, '\P{^Script_Extensions=olchiki}', "");
Expect(0, 7296, '\p{Script_Extensions=olchiki}', "");
Expect(1, 7296, '\p{^Script_Extensions=olchiki}', "");
Expect(1, 7296, '\P{Script_Extensions=olchiki}', "");
Expect(0, 7296, '\P{^Script_Extensions=olchiki}', "");
Expect(1, 7295, '\p{Script_Extensions=_	ol_Chiki}', "");
Expect(0, 7295, '\p{^Script_Extensions=_	ol_Chiki}', "");
Expect(0, 7295, '\P{Script_Extensions=_	ol_Chiki}', "");
Expect(1, 7295, '\P{^Script_Extensions=_	ol_Chiki}', "");
Expect(0, 7296, '\p{Script_Extensions=_	ol_Chiki}', "");
Expect(1, 7296, '\p{^Script_Extensions=_	ol_Chiki}', "");
Expect(1, 7296, '\P{Script_Extensions=_	ol_Chiki}', "");
Expect(0, 7296, '\P{^Script_Extensions=_	ol_Chiki}', "");
Error('\p{Scx=	:=Olck}');
Error('\P{Scx=	:=Olck}');
Expect(1, 7295, '\p{Scx=olck}', "");
Expect(0, 7295, '\p{^Scx=olck}', "");
Expect(0, 7295, '\P{Scx=olck}', "");
Expect(1, 7295, '\P{^Scx=olck}', "");
Expect(0, 7296, '\p{Scx=olck}', "");
Expect(1, 7296, '\p{^Scx=olck}', "");
Expect(1, 7296, '\P{Scx=olck}', "");
Expect(0, 7296, '\P{^Scx=olck}', "");
Expect(1, 7295, '\p{Scx=  OLCK}', "");
Expect(0, 7295, '\p{^Scx=  OLCK}', "");
Expect(0, 7295, '\P{Scx=  OLCK}', "");
Expect(1, 7295, '\P{^Scx=  OLCK}', "");
Expect(0, 7296, '\p{Scx=  OLCK}', "");
Expect(1, 7296, '\p{^Scx=  OLCK}', "");
Expect(1, 7296, '\P{Scx=  OLCK}', "");
Expect(0, 7296, '\P{^Scx=  OLCK}', "");
Error('\p{Is_Script_Extensions:	:=  ol_chiki}');
Error('\P{Is_Script_Extensions:	:=  ol_chiki}');
Expect(1, 7295, '\p{Is_Script_Extensions=olchiki}', "");
Expect(0, 7295, '\p{^Is_Script_Extensions=olchiki}', "");
Expect(0, 7295, '\P{Is_Script_Extensions=olchiki}', "");
Expect(1, 7295, '\P{^Is_Script_Extensions=olchiki}', "");
Expect(0, 7296, '\p{Is_Script_Extensions=olchiki}', "");
Expect(1, 7296, '\p{^Is_Script_Extensions=olchiki}', "");
Expect(1, 7296, '\P{Is_Script_Extensions=olchiki}', "");
Expect(0, 7296, '\P{^Is_Script_Extensions=olchiki}', "");
Expect(1, 7295, '\p{Is_Script_Extensions=_Ol_Chiki}', "");
Expect(0, 7295, '\p{^Is_Script_Extensions=_Ol_Chiki}', "");
Expect(0, 7295, '\P{Is_Script_Extensions=_Ol_Chiki}', "");
Expect(1, 7295, '\P{^Is_Script_Extensions=_Ol_Chiki}', "");
Expect(0, 7296, '\p{Is_Script_Extensions=_Ol_Chiki}', "");
Expect(1, 7296, '\p{^Is_Script_Extensions=_Ol_Chiki}', "");
Expect(1, 7296, '\P{Is_Script_Extensions=_Ol_Chiki}', "");
Expect(0, 7296, '\P{^Is_Script_Extensions=_Ol_Chiki}', "");
Error('\p{Is_Scx= -Olck:=}');
Error('\P{Is_Scx= -Olck:=}');
Expect(1, 7295, '\p{Is_Scx=olck}', "");
Expect(0, 7295, '\p{^Is_Scx=olck}', "");
Expect(0, 7295, '\P{Is_Scx=olck}', "");
Expect(1, 7295, '\P{^Is_Scx=olck}', "");
Expect(0, 7296, '\p{Is_Scx=olck}', "");
Expect(1, 7296, '\p{^Is_Scx=olck}', "");
Expect(1, 7296, '\P{Is_Scx=olck}', "");
Expect(0, 7296, '\P{^Is_Scx=olck}', "");
Expect(1, 7295, '\p{Is_Scx=	_Olck}', "");
Expect(0, 7295, '\p{^Is_Scx=	_Olck}', "");
Expect(0, 7295, '\P{Is_Scx=	_Olck}', "");
Expect(1, 7295, '\P{^Is_Scx=	_Olck}', "");
Expect(0, 7296, '\p{Is_Scx=	_Olck}', "");
Expect(1, 7296, '\p{^Is_Scx=	_Olck}', "");
Expect(1, 7296, '\P{Is_Scx=	_Olck}', "");
Expect(0, 7296, '\P{^Is_Scx=	_Olck}', "");
Error('\p{Script_Extensions=/a/Old_Turkic}');
Error('\P{Script_Extensions=/a/Old_Turkic}');
Expect(1, 68680, '\p{Script_Extensions=oldturkic}', "");
Expect(0, 68680, '\p{^Script_Extensions=oldturkic}', "");
Expect(0, 68680, '\P{Script_Extensions=oldturkic}', "");
Expect(1, 68680, '\P{^Script_Extensions=oldturkic}', "");
Expect(0, 68681, '\p{Script_Extensions=oldturkic}', "");
Expect(1, 68681, '\p{^Script_Extensions=oldturkic}', "");
Expect(1, 68681, '\P{Script_Extensions=oldturkic}', "");
Expect(0, 68681, '\P{^Script_Extensions=oldturkic}', "");
Expect(1, 68680, '\p{Script_Extensions=-Old_Turkic}', "");
Expect(0, 68680, '\p{^Script_Extensions=-Old_Turkic}', "");
Expect(0, 68680, '\P{Script_Extensions=-Old_Turkic}', "");
Expect(1, 68680, '\P{^Script_Extensions=-Old_Turkic}', "");
Expect(0, 68681, '\p{Script_Extensions=-Old_Turkic}', "");
Expect(1, 68681, '\p{^Script_Extensions=-Old_Turkic}', "");
Expect(1, 68681, '\P{Script_Extensions=-Old_Turkic}', "");
Expect(0, 68681, '\P{^Script_Extensions=-Old_Turkic}', "");
Error('\p{Scx=-:=Orkh}');
Error('\P{Scx=-:=Orkh}');
Expect(1, 68680, '\p{Scx=orkh}', "");
Expect(0, 68680, '\p{^Scx=orkh}', "");
Expect(0, 68680, '\P{Scx=orkh}', "");
Expect(1, 68680, '\P{^Scx=orkh}', "");
Expect(0, 68681, '\p{Scx=orkh}', "");
Expect(1, 68681, '\p{^Scx=orkh}', "");
Expect(1, 68681, '\P{Scx=orkh}', "");
Expect(0, 68681, '\P{^Scx=orkh}', "");
Expect(1, 68680, '\p{Scx=  orkh}', "");
Expect(0, 68680, '\p{^Scx=  orkh}', "");
Expect(0, 68680, '\P{Scx=  orkh}', "");
Expect(1, 68680, '\P{^Scx=  orkh}', "");
Expect(0, 68681, '\p{Scx=  orkh}', "");
Expect(1, 68681, '\p{^Scx=  orkh}', "");
Expect(1, 68681, '\P{Scx=  orkh}', "");
Expect(0, 68681, '\P{^Scx=  orkh}', "");
Error('\p{Is_Script_Extensions=/a/ Old_turkic}');
Error('\P{Is_Script_Extensions=/a/ Old_turkic}');
Expect(1, 68680, '\p{Is_Script_Extensions=oldturkic}', "");
Expect(0, 68680, '\p{^Is_Script_Extensions=oldturkic}', "");
Expect(0, 68680, '\P{Is_Script_Extensions=oldturkic}', "");
Expect(1, 68680, '\P{^Is_Script_Extensions=oldturkic}', "");
Expect(0, 68681, '\p{Is_Script_Extensions=oldturkic}', "");
Expect(1, 68681, '\p{^Is_Script_Extensions=oldturkic}', "");
Expect(1, 68681, '\P{Is_Script_Extensions=oldturkic}', "");
Expect(0, 68681, '\P{^Is_Script_Extensions=oldturkic}', "");
Expect(1, 68680, '\p{Is_Script_Extensions:   _	OLD_Turkic}', "");
Expect(0, 68680, '\p{^Is_Script_Extensions:   _	OLD_Turkic}', "");
Expect(0, 68680, '\P{Is_Script_Extensions:   _	OLD_Turkic}', "");
Expect(1, 68680, '\P{^Is_Script_Extensions:   _	OLD_Turkic}', "");
Expect(0, 68681, '\p{Is_Script_Extensions:   _	OLD_Turkic}', "");
Expect(1, 68681, '\p{^Is_Script_Extensions:   _	OLD_Turkic}', "");
Expect(1, 68681, '\P{Is_Script_Extensions:   _	OLD_Turkic}', "");
Expect(0, 68681, '\P{^Is_Script_Extensions:   _	OLD_Turkic}', "");
Error('\p{Is_Scx=/a/_	orkh}');
Error('\P{Is_Scx=/a/_	orkh}');
Expect(1, 68680, '\p{Is_Scx:	orkh}', "");
Expect(0, 68680, '\p{^Is_Scx:	orkh}', "");
Expect(0, 68680, '\P{Is_Scx:	orkh}', "");
Expect(1, 68680, '\P{^Is_Scx:	orkh}', "");
Expect(0, 68681, '\p{Is_Scx:	orkh}', "");
Expect(1, 68681, '\p{^Is_Scx:	orkh}', "");
Expect(1, 68681, '\P{Is_Scx:	orkh}', "");
Expect(0, 68681, '\P{^Is_Scx:	orkh}', "");
Expect(1, 68680, '\p{Is_Scx=	-orkh}', "");
Expect(0, 68680, '\p{^Is_Scx=	-orkh}', "");
Expect(0, 68680, '\P{Is_Scx=	-orkh}', "");
Expect(1, 68680, '\P{^Is_Scx=	-orkh}', "");
Expect(0, 68681, '\p{Is_Scx=	-orkh}', "");
Expect(1, 68681, '\p{^Is_Scx=	-orkh}', "");
Expect(1, 68681, '\P{Is_Scx=	-orkh}', "");
Expect(0, 68681, '\P{^Is_Scx=	-orkh}', "");
Error('\p{Script_Extensions=/a/  oriya}');
Error('\P{Script_Extensions=/a/  oriya}');
Expect(1, 2935, '\p{Script_Extensions=oriya}', "");
Expect(0, 2935, '\p{^Script_Extensions=oriya}', "");
Expect(0, 2935, '\P{Script_Extensions=oriya}', "");
Expect(1, 2935, '\P{^Script_Extensions=oriya}', "");
Expect(0, 2936, '\p{Script_Extensions=oriya}', "");
Expect(1, 2936, '\p{^Script_Extensions=oriya}', "");
Expect(1, 2936, '\P{Script_Extensions=oriya}', "");
Expect(0, 2936, '\P{^Script_Extensions=oriya}', "");
Expect(1, 2935, '\p{Script_Extensions:_Oriya}', "");
Expect(0, 2935, '\p{^Script_Extensions:_Oriya}', "");
Expect(0, 2935, '\P{Script_Extensions:_Oriya}', "");
Expect(1, 2935, '\P{^Script_Extensions:_Oriya}', "");
Expect(0, 2936, '\p{Script_Extensions:_Oriya}', "");
Expect(1, 2936, '\p{^Script_Extensions:_Oriya}', "");
Expect(1, 2936, '\P{Script_Extensions:_Oriya}', "");
Expect(0, 2936, '\P{^Script_Extensions:_Oriya}', "");
Error('\p{Scx=_/a/ORYA}');
Error('\P{Scx=_/a/ORYA}');
Expect(1, 2935, '\p{Scx=orya}', "");
Expect(0, 2935, '\p{^Scx=orya}', "");
Expect(0, 2935, '\P{Scx=orya}', "");
Expect(1, 2935, '\P{^Scx=orya}', "");
Expect(0, 2936, '\p{Scx=orya}', "");
Expect(1, 2936, '\p{^Scx=orya}', "");
Expect(1, 2936, '\P{Scx=orya}', "");
Expect(0, 2936, '\P{^Scx=orya}', "");
Expect(1, 2935, '\p{Scx=-Orya}', "");
Expect(0, 2935, '\p{^Scx=-Orya}', "");
Expect(0, 2935, '\P{Scx=-Orya}', "");
Expect(1, 2935, '\P{^Scx=-Orya}', "");
Expect(0, 2936, '\p{Scx=-Orya}', "");
Expect(1, 2936, '\p{^Scx=-Orya}', "");
Expect(1, 2936, '\P{Scx=-Orya}', "");
Expect(0, 2936, '\P{^Scx=-Orya}', "");
Error('\p{Is_Script_Extensions:   --ORIYA:=}');
Error('\P{Is_Script_Extensions:   --ORIYA:=}');
Expect(1, 2935, '\p{Is_Script_Extensions:   oriya}', "");
Expect(0, 2935, '\p{^Is_Script_Extensions:   oriya}', "");
Expect(0, 2935, '\P{Is_Script_Extensions:   oriya}', "");
Expect(1, 2935, '\P{^Is_Script_Extensions:   oriya}', "");
Expect(0, 2936, '\p{Is_Script_Extensions:   oriya}', "");
Expect(1, 2936, '\p{^Is_Script_Extensions:   oriya}', "");
Expect(1, 2936, '\P{Is_Script_Extensions:   oriya}', "");
Expect(0, 2936, '\P{^Is_Script_Extensions:   oriya}', "");
Expect(1, 2935, '\p{Is_Script_Extensions=  oriya}', "");
Expect(0, 2935, '\p{^Is_Script_Extensions=  oriya}', "");
Expect(0, 2935, '\P{Is_Script_Extensions=  oriya}', "");
Expect(1, 2935, '\P{^Is_Script_Extensions=  oriya}', "");
Expect(0, 2936, '\p{Is_Script_Extensions=  oriya}', "");
Expect(1, 2936, '\p{^Is_Script_Extensions=  oriya}', "");
Expect(1, 2936, '\P{Is_Script_Extensions=  oriya}', "");
Expect(0, 2936, '\P{^Is_Script_Extensions=  oriya}', "");
Error('\p{Is_Scx=__Orya:=}');
Error('\P{Is_Scx=__Orya:=}');
Expect(1, 2935, '\p{Is_Scx=orya}', "");
Expect(0, 2935, '\p{^Is_Scx=orya}', "");
Expect(0, 2935, '\P{Is_Scx=orya}', "");
Expect(1, 2935, '\P{^Is_Scx=orya}', "");
Expect(0, 2936, '\p{Is_Scx=orya}', "");
Expect(1, 2936, '\p{^Is_Scx=orya}', "");
Expect(1, 2936, '\P{Is_Scx=orya}', "");
Expect(0, 2936, '\P{^Is_Scx=orya}', "");
Expect(1, 2935, '\p{Is_Scx=		orya}', "");
Expect(0, 2935, '\p{^Is_Scx=		orya}', "");
Expect(0, 2935, '\P{Is_Scx=		orya}', "");
Expect(1, 2935, '\P{^Is_Scx=		orya}', "");
Expect(0, 2936, '\p{Is_Scx=		orya}', "");
Expect(1, 2936, '\p{^Is_Scx=		orya}', "");
Expect(1, 2936, '\P{Is_Scx=		orya}', "");
Expect(0, 2936, '\P{^Is_Scx=		orya}', "");
Error('\p{Script_Extensions=_OSMANYA:=}');
Error('\P{Script_Extensions=_OSMANYA:=}');
Expect(1, 66729, '\p{Script_Extensions=osmanya}', "");
Expect(0, 66729, '\p{^Script_Extensions=osmanya}', "");
Expect(0, 66729, '\P{Script_Extensions=osmanya}', "");
Expect(1, 66729, '\P{^Script_Extensions=osmanya}', "");
Expect(0, 66730, '\p{Script_Extensions=osmanya}', "");
Expect(1, 66730, '\p{^Script_Extensions=osmanya}', "");
Expect(1, 66730, '\P{Script_Extensions=osmanya}', "");
Expect(0, 66730, '\P{^Script_Extensions=osmanya}', "");
Expect(1, 66729, '\p{Script_Extensions=_ osmanya}', "");
Expect(0, 66729, '\p{^Script_Extensions=_ osmanya}', "");
Expect(0, 66729, '\P{Script_Extensions=_ osmanya}', "");
Expect(1, 66729, '\P{^Script_Extensions=_ osmanya}', "");
Expect(0, 66730, '\p{Script_Extensions=_ osmanya}', "");
Expect(1, 66730, '\p{^Script_Extensions=_ osmanya}', "");
Expect(1, 66730, '\P{Script_Extensions=_ osmanya}', "");
Expect(0, 66730, '\P{^Script_Extensions=_ osmanya}', "");
Error('\p{Scx=:=OSMA}');
Error('\P{Scx=:=OSMA}');
Expect(1, 66729, '\p{Scx: osma}', "");
Expect(0, 66729, '\p{^Scx: osma}', "");
Expect(0, 66729, '\P{Scx: osma}', "");
Expect(1, 66729, '\P{^Scx: osma}', "");
Expect(0, 66730, '\p{Scx: osma}', "");
Expect(1, 66730, '\p{^Scx: osma}', "");
Expect(1, 66730, '\P{Scx: osma}', "");
Expect(0, 66730, '\P{^Scx: osma}', "");
Expect(1, 66729, '\p{Scx= _Osma}', "");
Expect(0, 66729, '\p{^Scx= _Osma}', "");
Expect(0, 66729, '\P{Scx= _Osma}', "");
Expect(1, 66729, '\P{^Scx= _Osma}', "");
Expect(0, 66730, '\p{Scx= _Osma}', "");
Expect(1, 66730, '\p{^Scx= _Osma}', "");
Expect(1, 66730, '\P{Scx= _Osma}', "");
Expect(0, 66730, '\P{^Scx= _Osma}', "");
Error('\p{Is_Script_Extensions: __Osmanya:=}');
Error('\P{Is_Script_Extensions: __Osmanya:=}');
Expect(1, 66729, '\p{Is_Script_Extensions=osmanya}', "");
Expect(0, 66729, '\p{^Is_Script_Extensions=osmanya}', "");
Expect(0, 66729, '\P{Is_Script_Extensions=osmanya}', "");
Expect(1, 66729, '\P{^Is_Script_Extensions=osmanya}', "");
Expect(0, 66730, '\p{Is_Script_Extensions=osmanya}', "");
Expect(1, 66730, '\p{^Is_Script_Extensions=osmanya}', "");
Expect(1, 66730, '\P{Is_Script_Extensions=osmanya}', "");
Expect(0, 66730, '\P{^Is_Script_Extensions=osmanya}', "");
Expect(1, 66729, '\p{Is_Script_Extensions=	-Osmanya}', "");
Expect(0, 66729, '\p{^Is_Script_Extensions=	-Osmanya}', "");
Expect(0, 66729, '\P{Is_Script_Extensions=	-Osmanya}', "");
Expect(1, 66729, '\P{^Is_Script_Extensions=	-Osmanya}', "");
Expect(0, 66730, '\p{Is_Script_Extensions=	-Osmanya}', "");
Expect(1, 66730, '\p{^Is_Script_Extensions=	-Osmanya}', "");
Expect(1, 66730, '\P{Is_Script_Extensions=	-Osmanya}', "");
Expect(0, 66730, '\P{^Is_Script_Extensions=	-Osmanya}', "");
Error('\p{Is_Scx=	Osma:=}');
Error('\P{Is_Scx=	Osma:=}');
Expect(1, 66729, '\p{Is_Scx=osma}', "");
Expect(0, 66729, '\p{^Is_Scx=osma}', "");
Expect(0, 66729, '\P{Is_Scx=osma}', "");
Expect(1, 66729, '\P{^Is_Scx=osma}', "");
Expect(0, 66730, '\p{Is_Scx=osma}', "");
Expect(1, 66730, '\p{^Is_Scx=osma}', "");
Expect(1, 66730, '\P{Is_Scx=osma}', "");
Expect(0, 66730, '\P{^Is_Scx=osma}', "");
Expect(1, 66729, '\p{Is_Scx=-_Osma}', "");
Expect(0, 66729, '\p{^Is_Scx=-_Osma}', "");
Expect(0, 66729, '\P{Is_Scx=-_Osma}', "");
Expect(1, 66729, '\P{^Is_Scx=-_Osma}', "");
Expect(0, 66730, '\p{Is_Scx=-_Osma}', "");
Expect(1, 66730, '\p{^Is_Scx=-_Osma}', "");
Expect(1, 66730, '\P{Is_Scx=-_Osma}', "");
Expect(0, 66730, '\P{^Is_Scx=-_Osma}', "");
Error('\p{Script_Extensions=_phags_Pa:=}');
Error('\P{Script_Extensions=_phags_Pa:=}');
Expect(1, 43127, '\p{Script_Extensions=phagspa}', "");
Expect(0, 43127, '\p{^Script_Extensions=phagspa}', "");
Expect(0, 43127, '\P{Script_Extensions=phagspa}', "");
Expect(1, 43127, '\P{^Script_Extensions=phagspa}', "");
Expect(0, 43128, '\p{Script_Extensions=phagspa}', "");
Expect(1, 43128, '\p{^Script_Extensions=phagspa}', "");
Expect(1, 43128, '\P{Script_Extensions=phagspa}', "");
Expect(0, 43128, '\P{^Script_Extensions=phagspa}', "");
Expect(1, 43127, '\p{Script_Extensions=- phags_pa}', "");
Expect(0, 43127, '\p{^Script_Extensions=- phags_pa}', "");
Expect(0, 43127, '\P{Script_Extensions=- phags_pa}', "");
Expect(1, 43127, '\P{^Script_Extensions=- phags_pa}', "");
Expect(0, 43128, '\p{Script_Extensions=- phags_pa}', "");
Expect(1, 43128, '\p{^Script_Extensions=- phags_pa}', "");
Expect(1, 43128, '\P{Script_Extensions=- phags_pa}', "");
Expect(0, 43128, '\P{^Script_Extensions=- phags_pa}', "");
Error('\p{Scx:   _:=Phag}');
Error('\P{Scx:   _:=Phag}');
Expect(1, 43127, '\p{Scx=phag}', "");
Expect(0, 43127, '\p{^Scx=phag}', "");
Expect(0, 43127, '\P{Scx=phag}', "");
Expect(1, 43127, '\P{^Scx=phag}', "");
Expect(0, 43128, '\p{Scx=phag}', "");
Expect(1, 43128, '\p{^Scx=phag}', "");
Expect(1, 43128, '\P{Scx=phag}', "");
Expect(0, 43128, '\P{^Scx=phag}', "");
Expect(1, 43127, '\p{Scx:   	Phag}', "");
Expect(0, 43127, '\p{^Scx:   	Phag}', "");
Expect(0, 43127, '\P{Scx:   	Phag}', "");
Expect(1, 43127, '\P{^Scx:   	Phag}', "");
Expect(0, 43128, '\p{Scx:   	Phag}', "");
Expect(1, 43128, '\p{^Scx:   	Phag}', "");
Expect(1, 43128, '\P{Scx:   	Phag}', "");
Expect(0, 43128, '\P{^Scx:   	Phag}', "");
Error('\p{Is_Script_Extensions=_/a/PHAGS_pa}');
Error('\P{Is_Script_Extensions=_/a/PHAGS_pa}');
Expect(1, 43127, '\p{Is_Script_Extensions=phagspa}', "");
Expect(0, 43127, '\p{^Is_Script_Extensions=phagspa}', "");
Expect(0, 43127, '\P{Is_Script_Extensions=phagspa}', "");
Expect(1, 43127, '\P{^Is_Script_Extensions=phagspa}', "");
Expect(0, 43128, '\p{Is_Script_Extensions=phagspa}', "");
Expect(1, 43128, '\p{^Is_Script_Extensions=phagspa}', "");
Expect(1, 43128, '\P{Is_Script_Extensions=phagspa}', "");
Expect(0, 43128, '\P{^Is_Script_Extensions=phagspa}', "");
Expect(1, 43127, '\p{Is_Script_Extensions=-Phags_Pa}', "");
Expect(0, 43127, '\p{^Is_Script_Extensions=-Phags_Pa}', "");
Expect(0, 43127, '\P{Is_Script_Extensions=-Phags_Pa}', "");
Expect(1, 43127, '\P{^Is_Script_Extensions=-Phags_Pa}', "");
Expect(0, 43128, '\p{Is_Script_Extensions=-Phags_Pa}', "");
Expect(1, 43128, '\p{^Is_Script_Extensions=-Phags_Pa}', "");
Expect(1, 43128, '\P{Is_Script_Extensions=-Phags_Pa}', "");
Expect(0, 43128, '\P{^Is_Script_Extensions=-Phags_Pa}', "");
Error('\p{Is_Scx=:= PHAG}');
Error('\P{Is_Scx=:= PHAG}');
Expect(1, 43127, '\p{Is_Scx=phag}', "");
Expect(0, 43127, '\p{^Is_Scx=phag}', "");
Expect(0, 43127, '\P{Is_Scx=phag}', "");
Expect(1, 43127, '\P{^Is_Scx=phag}', "");
Expect(0, 43128, '\p{Is_Scx=phag}', "");
Expect(1, 43128, '\p{^Is_Scx=phag}', "");
Expect(1, 43128, '\P{Is_Scx=phag}', "");
Expect(0, 43128, '\P{^Is_Scx=phag}', "");
Expect(1, 43127, '\p{Is_Scx=	 Phag}', "");
Expect(0, 43127, '\p{^Is_Scx=	 Phag}', "");
Expect(0, 43127, '\P{Is_Scx=	 Phag}', "");
Expect(1, 43127, '\P{^Is_Scx=	 Phag}', "");
Expect(0, 43128, '\p{Is_Scx=	 Phag}', "");
Expect(1, 43128, '\p{^Is_Scx=	 Phag}', "");
Expect(1, 43128, '\P{Is_Scx=	 Phag}', "");
Expect(0, 43128, '\P{^Is_Scx=	 Phag}', "");
Error('\p{Script_Extensions=:=--inscriptional_pahlavi}');
Error('\P{Script_Extensions=:=--inscriptional_pahlavi}');
Expect(1, 68479, '\p{Script_Extensions=inscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^Script_Extensions=inscriptionalpahlavi}', "");
Expect(0, 68479, '\P{Script_Extensions=inscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^Script_Extensions=inscriptionalpahlavi}', "");
Expect(0, 68480, '\p{Script_Extensions=inscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^Script_Extensions=inscriptionalpahlavi}', "");
Expect(1, 68480, '\P{Script_Extensions=inscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^Script_Extensions=inscriptionalpahlavi}', "");
Expect(1, 68479, '\p{Script_Extensions=Inscriptional_PAHLAVI}', "");
Expect(0, 68479, '\p{^Script_Extensions=Inscriptional_PAHLAVI}', "");
Expect(0, 68479, '\P{Script_Extensions=Inscriptional_PAHLAVI}', "");
Expect(1, 68479, '\P{^Script_Extensions=Inscriptional_PAHLAVI}', "");
Expect(0, 68480, '\p{Script_Extensions=Inscriptional_PAHLAVI}', "");
Expect(1, 68480, '\p{^Script_Extensions=Inscriptional_PAHLAVI}', "");
Expect(1, 68480, '\P{Script_Extensions=Inscriptional_PAHLAVI}', "");
Expect(0, 68480, '\P{^Script_Extensions=Inscriptional_PAHLAVI}', "");
Error('\p{Scx=	/a/Phli}');
Error('\P{Scx=	/a/Phli}');
Expect(1, 68479, '\p{Scx=phli}', "");
Expect(0, 68479, '\p{^Scx=phli}', "");
Expect(0, 68479, '\P{Scx=phli}', "");
Expect(1, 68479, '\P{^Scx=phli}', "");
Expect(0, 68480, '\p{Scx=phli}', "");
Expect(1, 68480, '\p{^Scx=phli}', "");
Expect(1, 68480, '\P{Scx=phli}', "");
Expect(0, 68480, '\P{^Scx=phli}', "");
Expect(1, 68479, '\p{Scx=-Phli}', "");
Expect(0, 68479, '\p{^Scx=-Phli}', "");
Expect(0, 68479, '\P{Scx=-Phli}', "");
Expect(1, 68479, '\P{^Scx=-Phli}', "");
Expect(0, 68480, '\p{Scx=-Phli}', "");
Expect(1, 68480, '\p{^Scx=-Phli}', "");
Expect(1, 68480, '\P{Scx=-Phli}', "");
Expect(0, 68480, '\P{^Scx=-Phli}', "");
Error('\p{Is_Script_Extensions= /a/inscriptional_PAHLAVI}');
Error('\P{Is_Script_Extensions= /a/inscriptional_PAHLAVI}');
Expect(1, 68479, '\p{Is_Script_Extensions=inscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^Is_Script_Extensions=inscriptionalpahlavi}', "");
Expect(0, 68479, '\P{Is_Script_Extensions=inscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^Is_Script_Extensions=inscriptionalpahlavi}', "");
Expect(0, 68480, '\p{Is_Script_Extensions=inscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^Is_Script_Extensions=inscriptionalpahlavi}', "");
Expect(1, 68480, '\P{Is_Script_Extensions=inscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^Is_Script_Extensions=inscriptionalpahlavi}', "");
Expect(1, 68479, '\p{Is_Script_Extensions=__Inscriptional_pahlavi}', "");
Expect(0, 68479, '\p{^Is_Script_Extensions=__Inscriptional_pahlavi}', "");
Expect(0, 68479, '\P{Is_Script_Extensions=__Inscriptional_pahlavi}', "");
Expect(1, 68479, '\P{^Is_Script_Extensions=__Inscriptional_pahlavi}', "");
Expect(0, 68480, '\p{Is_Script_Extensions=__Inscriptional_pahlavi}', "");
Expect(1, 68480, '\p{^Is_Script_Extensions=__Inscriptional_pahlavi}', "");
Expect(1, 68480, '\P{Is_Script_Extensions=__Inscriptional_pahlavi}', "");
Expect(0, 68480, '\P{^Is_Script_Extensions=__Inscriptional_pahlavi}', "");
Error('\p{Is_Scx=:=Phli}');
Error('\P{Is_Scx=:=Phli}');
Expect(1, 68479, '\p{Is_Scx:phli}', "");
Expect(0, 68479, '\p{^Is_Scx:phli}', "");
Expect(0, 68479, '\P{Is_Scx:phli}', "");
Expect(1, 68479, '\P{^Is_Scx:phli}', "");
Expect(0, 68480, '\p{Is_Scx:phli}', "");
Expect(1, 68480, '\p{^Is_Scx:phli}', "");
Expect(1, 68480, '\P{Is_Scx:phli}', "");
Expect(0, 68480, '\P{^Is_Scx:phli}', "");
Expect(1, 68479, '\p{Is_Scx=_ Phli}', "");
Expect(0, 68479, '\p{^Is_Scx=_ Phli}', "");
Expect(0, 68479, '\P{Is_Scx=_ Phli}', "");
Expect(1, 68479, '\P{^Is_Scx=_ Phli}', "");
Expect(0, 68480, '\p{Is_Scx=_ Phli}', "");
Expect(1, 68480, '\p{^Is_Scx=_ Phli}', "");
Expect(1, 68480, '\P{Is_Scx=_ Phli}', "");
Expect(0, 68480, '\P{^Is_Scx=_ Phli}', "");
Error('\p{Script_Extensions=_PHOENICIAN:=}');
Error('\P{Script_Extensions=_PHOENICIAN:=}');
Expect(1, 67871, '\p{Script_Extensions=phoenician}', "");
Expect(0, 67871, '\p{^Script_Extensions=phoenician}', "");
Expect(0, 67871, '\P{Script_Extensions=phoenician}', "");
Expect(1, 67871, '\P{^Script_Extensions=phoenician}', "");
Expect(0, 67872, '\p{Script_Extensions=phoenician}', "");
Expect(1, 67872, '\p{^Script_Extensions=phoenician}', "");
Expect(1, 67872, '\P{Script_Extensions=phoenician}', "");
Expect(0, 67872, '\P{^Script_Extensions=phoenician}', "");
Expect(1, 67871, '\p{Script_Extensions=-Phoenician}', "");
Expect(0, 67871, '\p{^Script_Extensions=-Phoenician}', "");
Expect(0, 67871, '\P{Script_Extensions=-Phoenician}', "");
Expect(1, 67871, '\P{^Script_Extensions=-Phoenician}', "");
Expect(0, 67872, '\p{Script_Extensions=-Phoenician}', "");
Expect(1, 67872, '\p{^Script_Extensions=-Phoenician}', "");
Expect(1, 67872, '\P{Script_Extensions=-Phoenician}', "");
Expect(0, 67872, '\P{^Script_Extensions=-Phoenician}', "");
Error('\p{Scx=_/a/phnx}');
Error('\P{Scx=_/a/phnx}');
Expect(1, 67871, '\p{Scx=phnx}', "");
Expect(0, 67871, '\p{^Scx=phnx}', "");
Expect(0, 67871, '\P{Scx=phnx}', "");
Expect(1, 67871, '\P{^Scx=phnx}', "");
Expect(0, 67872, '\p{Scx=phnx}', "");
Expect(1, 67872, '\p{^Scx=phnx}', "");
Expect(1, 67872, '\P{Scx=phnx}', "");
Expect(0, 67872, '\P{^Scx=phnx}', "");
Expect(1, 67871, '\p{Scx=_phnx}', "");
Expect(0, 67871, '\p{^Scx=_phnx}', "");
Expect(0, 67871, '\P{Scx=_phnx}', "");
Expect(1, 67871, '\P{^Scx=_phnx}', "");
Expect(0, 67872, '\p{Scx=_phnx}', "");
Expect(1, 67872, '\p{^Scx=_phnx}', "");
Expect(1, 67872, '\P{Scx=_phnx}', "");
Expect(0, 67872, '\P{^Scx=_phnx}', "");
Error('\p{Is_Script_Extensions=	phoenician:=}');
Error('\P{Is_Script_Extensions=	phoenician:=}');
Expect(1, 67871, '\p{Is_Script_Extensions=phoenician}', "");
Expect(0, 67871, '\p{^Is_Script_Extensions=phoenician}', "");
Expect(0, 67871, '\P{Is_Script_Extensions=phoenician}', "");
Expect(1, 67871, '\P{^Is_Script_Extensions=phoenician}', "");
Expect(0, 67872, '\p{Is_Script_Extensions=phoenician}', "");
Expect(1, 67872, '\p{^Is_Script_Extensions=phoenician}', "");
Expect(1, 67872, '\P{Is_Script_Extensions=phoenician}', "");
Expect(0, 67872, '\P{^Is_Script_Extensions=phoenician}', "");
Expect(1, 67871, '\p{Is_Script_Extensions= PHOENICIAN}', "");
Expect(0, 67871, '\p{^Is_Script_Extensions= PHOENICIAN}', "");
Expect(0, 67871, '\P{Is_Script_Extensions= PHOENICIAN}', "");
Expect(1, 67871, '\P{^Is_Script_Extensions= PHOENICIAN}', "");
Expect(0, 67872, '\p{Is_Script_Extensions= PHOENICIAN}', "");
Expect(1, 67872, '\p{^Is_Script_Extensions= PHOENICIAN}', "");
Expect(1, 67872, '\P{Is_Script_Extensions= PHOENICIAN}', "");
Expect(0, 67872, '\P{^Is_Script_Extensions= PHOENICIAN}', "");
Error('\p{Is_Scx=_Phnx:=}');
Error('\P{Is_Scx=_Phnx:=}');
Expect(1, 67871, '\p{Is_Scx:	phnx}', "");
Expect(0, 67871, '\p{^Is_Scx:	phnx}', "");
Expect(0, 67871, '\P{Is_Scx:	phnx}', "");
Expect(1, 67871, '\P{^Is_Scx:	phnx}', "");
Expect(0, 67872, '\p{Is_Scx:	phnx}', "");
Expect(1, 67872, '\p{^Is_Scx:	phnx}', "");
Expect(1, 67872, '\P{Is_Scx:	phnx}', "");
Expect(0, 67872, '\P{^Is_Scx:	phnx}', "");
Expect(1, 67871, '\p{Is_Scx=--Phnx}', "");
Expect(0, 67871, '\p{^Is_Scx=--Phnx}', "");
Expect(0, 67871, '\P{Is_Scx=--Phnx}', "");
Expect(1, 67871, '\P{^Is_Scx=--Phnx}', "");
Expect(0, 67872, '\p{Is_Scx=--Phnx}', "");
Expect(1, 67872, '\p{^Is_Scx=--Phnx}', "");
Expect(1, 67872, '\P{Is_Scx=--Phnx}', "");
Expect(0, 67872, '\P{^Is_Scx=--Phnx}', "");
Error('\p{Script_Extensions=  Miao:=}');
Error('\P{Script_Extensions=  Miao:=}');
Expect(1, 94111, '\p{Script_Extensions=miao}', "");
Expect(0, 94111, '\p{^Script_Extensions=miao}', "");
Expect(0, 94111, '\P{Script_Extensions=miao}', "");
Expect(1, 94111, '\P{^Script_Extensions=miao}', "");
Expect(0, 94112, '\p{Script_Extensions=miao}', "");
Expect(1, 94112, '\p{^Script_Extensions=miao}', "");
Expect(1, 94112, '\P{Script_Extensions=miao}', "");
Expect(0, 94112, '\P{^Script_Extensions=miao}', "");
Expect(1, 94111, '\p{Script_Extensions= Miao}', "");
Expect(0, 94111, '\p{^Script_Extensions= Miao}', "");
Expect(0, 94111, '\P{Script_Extensions= Miao}', "");
Expect(1, 94111, '\P{^Script_Extensions= Miao}', "");
Expect(0, 94112, '\p{Script_Extensions= Miao}', "");
Expect(1, 94112, '\p{^Script_Extensions= Miao}', "");
Expect(1, 94112, '\P{Script_Extensions= Miao}', "");
Expect(0, 94112, '\P{^Script_Extensions= Miao}', "");
Error('\p{Scx=		Plrd:=}');
Error('\P{Scx=		Plrd:=}');
Expect(1, 94111, '\p{Scx=plrd}', "");
Expect(0, 94111, '\p{^Scx=plrd}', "");
Expect(0, 94111, '\P{Scx=plrd}', "");
Expect(1, 94111, '\P{^Scx=plrd}', "");
Expect(0, 94112, '\p{Scx=plrd}', "");
Expect(1, 94112, '\p{^Scx=plrd}', "");
Expect(1, 94112, '\P{Scx=plrd}', "");
Expect(0, 94112, '\P{^Scx=plrd}', "");
Expect(1, 94111, '\p{Scx=_-plrd}', "");
Expect(0, 94111, '\p{^Scx=_-plrd}', "");
Expect(0, 94111, '\P{Scx=_-plrd}', "");
Expect(1, 94111, '\P{^Scx=_-plrd}', "");
Expect(0, 94112, '\p{Scx=_-plrd}', "");
Expect(1, 94112, '\p{^Scx=_-plrd}', "");
Expect(1, 94112, '\P{Scx=_-plrd}', "");
Expect(0, 94112, '\P{^Scx=_-plrd}', "");
Error('\p{Is_Script_Extensions= 	Miao:=}');
Error('\P{Is_Script_Extensions= 	Miao:=}');
Expect(1, 94111, '\p{Is_Script_Extensions=miao}', "");
Expect(0, 94111, '\p{^Is_Script_Extensions=miao}', "");
Expect(0, 94111, '\P{Is_Script_Extensions=miao}', "");
Expect(1, 94111, '\P{^Is_Script_Extensions=miao}', "");
Expect(0, 94112, '\p{Is_Script_Extensions=miao}', "");
Expect(1, 94112, '\p{^Is_Script_Extensions=miao}', "");
Expect(1, 94112, '\P{Is_Script_Extensions=miao}', "");
Expect(0, 94112, '\P{^Is_Script_Extensions=miao}', "");
Expect(1, 94111, '\p{Is_Script_Extensions=	Miao}', "");
Expect(0, 94111, '\p{^Is_Script_Extensions=	Miao}', "");
Expect(0, 94111, '\P{Is_Script_Extensions=	Miao}', "");
Expect(1, 94111, '\P{^Is_Script_Extensions=	Miao}', "");
Expect(0, 94112, '\p{Is_Script_Extensions=	Miao}', "");
Expect(1, 94112, '\p{^Is_Script_Extensions=	Miao}', "");
Expect(1, 94112, '\P{Is_Script_Extensions=	Miao}', "");
Expect(0, 94112, '\P{^Is_Script_Extensions=	Miao}', "");
Error('\p{Is_Scx= /a/PLRD}');
Error('\P{Is_Scx= /a/PLRD}');
Expect(1, 94111, '\p{Is_Scx=plrd}', "");
Expect(0, 94111, '\p{^Is_Scx=plrd}', "");
Expect(0, 94111, '\P{Is_Scx=plrd}', "");
Expect(1, 94111, '\P{^Is_Scx=plrd}', "");
Expect(0, 94112, '\p{Is_Scx=plrd}', "");
Expect(1, 94112, '\p{^Is_Scx=plrd}', "");
Expect(1, 94112, '\P{Is_Scx=plrd}', "");
Expect(0, 94112, '\P{^Is_Scx=plrd}', "");
Expect(1, 94111, '\p{Is_Scx:	-PLRD}', "");
Expect(0, 94111, '\p{^Is_Scx:	-PLRD}', "");
Expect(0, 94111, '\P{Is_Scx:	-PLRD}', "");
Expect(1, 94111, '\P{^Is_Scx:	-PLRD}', "");
Expect(0, 94112, '\p{Is_Scx:	-PLRD}', "");
Expect(1, 94112, '\p{^Is_Scx:	-PLRD}', "");
Expect(1, 94112, '\P{Is_Scx:	-PLRD}', "");
Expect(0, 94112, '\P{^Is_Scx:	-PLRD}', "");
Error('\p{Script_Extensions=_-inscriptional_parthian/a/}');
Error('\P{Script_Extensions=_-inscriptional_parthian/a/}');
Expect(1, 68447, '\p{Script_Extensions=inscriptionalparthian}', "");
Expect(0, 68447, '\p{^Script_Extensions=inscriptionalparthian}', "");
Expect(0, 68447, '\P{Script_Extensions=inscriptionalparthian}', "");
Expect(1, 68447, '\P{^Script_Extensions=inscriptionalparthian}', "");
Expect(0, 68448, '\p{Script_Extensions=inscriptionalparthian}', "");
Expect(1, 68448, '\p{^Script_Extensions=inscriptionalparthian}', "");
Expect(1, 68448, '\P{Script_Extensions=inscriptionalparthian}', "");
Expect(0, 68448, '\P{^Script_Extensions=inscriptionalparthian}', "");
Expect(1, 68447, '\p{Script_Extensions=_	Inscriptional_parthian}', "");
Expect(0, 68447, '\p{^Script_Extensions=_	Inscriptional_parthian}', "");
Expect(0, 68447, '\P{Script_Extensions=_	Inscriptional_parthian}', "");
Expect(1, 68447, '\P{^Script_Extensions=_	Inscriptional_parthian}', "");
Expect(0, 68448, '\p{Script_Extensions=_	Inscriptional_parthian}', "");
Expect(1, 68448, '\p{^Script_Extensions=_	Inscriptional_parthian}', "");
Expect(1, 68448, '\P{Script_Extensions=_	Inscriptional_parthian}', "");
Expect(0, 68448, '\P{^Script_Extensions=_	Inscriptional_parthian}', "");
Error('\p{Scx=--Prti/a/}');
Error('\P{Scx=--Prti/a/}');
Expect(1, 68447, '\p{Scx=prti}', "");
Expect(0, 68447, '\p{^Scx=prti}', "");
Expect(0, 68447, '\P{Scx=prti}', "");
Expect(1, 68447, '\P{^Scx=prti}', "");
Expect(0, 68448, '\p{Scx=prti}', "");
Expect(1, 68448, '\p{^Scx=prti}', "");
Expect(1, 68448, '\P{Scx=prti}', "");
Expect(0, 68448, '\P{^Scx=prti}', "");
Expect(1, 68447, '\p{Scx=	Prti}', "");
Expect(0, 68447, '\p{^Scx=	Prti}', "");
Expect(0, 68447, '\P{Scx=	Prti}', "");
Expect(1, 68447, '\P{^Scx=	Prti}', "");
Expect(0, 68448, '\p{Scx=	Prti}', "");
Expect(1, 68448, '\p{^Scx=	Prti}', "");
Expect(1, 68448, '\P{Scx=	Prti}', "");
Expect(0, 68448, '\P{^Scx=	Prti}', "");
Error('\p{Is_Script_Extensions=/a/ inscriptional_Parthian}');
Error('\P{Is_Script_Extensions=/a/ inscriptional_Parthian}');
Expect(1, 68447, '\p{Is_Script_Extensions=inscriptionalparthian}', "");
Expect(0, 68447, '\p{^Is_Script_Extensions=inscriptionalparthian}', "");
Expect(0, 68447, '\P{Is_Script_Extensions=inscriptionalparthian}', "");
Expect(1, 68447, '\P{^Is_Script_Extensions=inscriptionalparthian}', "");
Expect(0, 68448, '\p{Is_Script_Extensions=inscriptionalparthian}', "");
Expect(1, 68448, '\p{^Is_Script_Extensions=inscriptionalparthian}', "");
Expect(1, 68448, '\P{Is_Script_Extensions=inscriptionalparthian}', "");
Expect(0, 68448, '\P{^Is_Script_Extensions=inscriptionalparthian}', "");
Expect(1, 68447, '\p{Is_Script_Extensions=  INSCRIPTIONAL_PARTHIAN}', "");
Expect(0, 68447, '\p{^Is_Script_Extensions=  INSCRIPTIONAL_PARTHIAN}', "");
Expect(0, 68447, '\P{Is_Script_Extensions=  INSCRIPTIONAL_PARTHIAN}', "");
Expect(1, 68447, '\P{^Is_Script_Extensions=  INSCRIPTIONAL_PARTHIAN}', "");
Expect(0, 68448, '\p{Is_Script_Extensions=  INSCRIPTIONAL_PARTHIAN}', "");
Expect(1, 68448, '\p{^Is_Script_Extensions=  INSCRIPTIONAL_PARTHIAN}', "");
Expect(1, 68448, '\P{Is_Script_Extensions=  INSCRIPTIONAL_PARTHIAN}', "");
Expect(0, 68448, '\P{^Is_Script_Extensions=  INSCRIPTIONAL_PARTHIAN}', "");
Error('\p{Is_Scx=-Prti/a/}');
Error('\P{Is_Scx=-Prti/a/}');
Expect(1, 68447, '\p{Is_Scx=prti}', "");
Expect(0, 68447, '\p{^Is_Scx=prti}', "");
Expect(0, 68447, '\P{Is_Scx=prti}', "");
Expect(1, 68447, '\P{^Is_Scx=prti}', "");
Expect(0, 68448, '\p{Is_Scx=prti}', "");
Expect(1, 68448, '\p{^Is_Scx=prti}', "");
Expect(1, 68448, '\P{Is_Scx=prti}', "");
Expect(0, 68448, '\P{^Is_Scx=prti}', "");
Expect(1, 68447, '\p{Is_Scx=	Prti}', "");
Expect(0, 68447, '\p{^Is_Scx=	Prti}', "");
Expect(0, 68447, '\P{Is_Scx=	Prti}', "");
Expect(1, 68447, '\P{^Is_Scx=	Prti}', "");
Expect(0, 68448, '\p{Is_Scx=	Prti}', "");
Expect(1, 68448, '\p{^Is_Scx=	Prti}', "");
Expect(1, 68448, '\P{Is_Scx=	Prti}', "");
Expect(0, 68448, '\P{^Is_Scx=	Prti}', "");
Error('\p{Script_Extensions=		REJANG:=}');
Error('\P{Script_Extensions=		REJANG:=}');
Expect(1, 43359, '\p{Script_Extensions:   rejang}', "");
Expect(0, 43359, '\p{^Script_Extensions:   rejang}', "");
Expect(0, 43359, '\P{Script_Extensions:   rejang}', "");
Expect(1, 43359, '\P{^Script_Extensions:   rejang}', "");
Expect(0, 43360, '\p{Script_Extensions:   rejang}', "");
Expect(1, 43360, '\p{^Script_Extensions:   rejang}', "");
Expect(1, 43360, '\P{Script_Extensions:   rejang}', "");
Expect(0, 43360, '\P{^Script_Extensions:   rejang}', "");
Expect(1, 43359, '\p{Script_Extensions=-_Rejang}', "");
Expect(0, 43359, '\p{^Script_Extensions=-_Rejang}', "");
Expect(0, 43359, '\P{Script_Extensions=-_Rejang}', "");
Expect(1, 43359, '\P{^Script_Extensions=-_Rejang}', "");
Expect(0, 43360, '\p{Script_Extensions=-_Rejang}', "");
Expect(1, 43360, '\p{^Script_Extensions=-_Rejang}', "");
Expect(1, 43360, '\P{Script_Extensions=-_Rejang}', "");
Expect(0, 43360, '\P{^Script_Extensions=-_Rejang}', "");
Error('\p{Scx=/a/Rjng}');
Error('\P{Scx=/a/Rjng}');
Expect(1, 43359, '\p{Scx=rjng}', "");
Expect(0, 43359, '\p{^Scx=rjng}', "");
Expect(0, 43359, '\P{Scx=rjng}', "");
Expect(1, 43359, '\P{^Scx=rjng}', "");
Expect(0, 43360, '\p{Scx=rjng}', "");
Expect(1, 43360, '\p{^Scx=rjng}', "");
Expect(1, 43360, '\P{Scx=rjng}', "");
Expect(0, 43360, '\P{^Scx=rjng}', "");
Expect(1, 43359, '\p{Scx=_	RJNG}', "");
Expect(0, 43359, '\p{^Scx=_	RJNG}', "");
Expect(0, 43359, '\P{Scx=_	RJNG}', "");
Expect(1, 43359, '\P{^Scx=_	RJNG}', "");
Expect(0, 43360, '\p{Scx=_	RJNG}', "");
Expect(1, 43360, '\p{^Scx=_	RJNG}', "");
Expect(1, 43360, '\P{Scx=_	RJNG}', "");
Expect(0, 43360, '\P{^Scx=_	RJNG}', "");
Error('\p{Is_Script_Extensions= REJANG:=}');
Error('\P{Is_Script_Extensions= REJANG:=}');
Expect(1, 43359, '\p{Is_Script_Extensions=rejang}', "");
Expect(0, 43359, '\p{^Is_Script_Extensions=rejang}', "");
Expect(0, 43359, '\P{Is_Script_Extensions=rejang}', "");
Expect(1, 43359, '\P{^Is_Script_Extensions=rejang}', "");
Expect(0, 43360, '\p{Is_Script_Extensions=rejang}', "");
Expect(1, 43360, '\p{^Is_Script_Extensions=rejang}', "");
Expect(1, 43360, '\P{Is_Script_Extensions=rejang}', "");
Expect(0, 43360, '\P{^Is_Script_Extensions=rejang}', "");
Expect(1, 43359, '\p{Is_Script_Extensions=- Rejang}', "");
Expect(0, 43359, '\p{^Is_Script_Extensions=- Rejang}', "");
Expect(0, 43359, '\P{Is_Script_Extensions=- Rejang}', "");
Expect(1, 43359, '\P{^Is_Script_Extensions=- Rejang}', "");
Expect(0, 43360, '\p{Is_Script_Extensions=- Rejang}', "");
Expect(1, 43360, '\p{^Is_Script_Extensions=- Rejang}', "");
Expect(1, 43360, '\P{Is_Script_Extensions=- Rejang}', "");
Expect(0, 43360, '\P{^Is_Script_Extensions=- Rejang}', "");
Error('\p{Is_Scx:/a/--Rjng}');
Error('\P{Is_Scx:/a/--Rjng}');
Expect(1, 43359, '\p{Is_Scx: rjng}', "");
Expect(0, 43359, '\p{^Is_Scx: rjng}', "");
Expect(0, 43359, '\P{Is_Scx: rjng}', "");
Expect(1, 43359, '\P{^Is_Scx: rjng}', "");
Expect(0, 43360, '\p{Is_Scx: rjng}', "");
Expect(1, 43360, '\p{^Is_Scx: rjng}', "");
Expect(1, 43360, '\P{Is_Scx: rjng}', "");
Expect(0, 43360, '\P{^Is_Scx: rjng}', "");
Expect(1, 43359, '\p{Is_Scx= 	Rjng}', "");
Expect(0, 43359, '\p{^Is_Scx= 	Rjng}', "");
Expect(0, 43359, '\P{Is_Scx= 	Rjng}', "");
Expect(1, 43359, '\P{^Is_Scx= 	Rjng}', "");
Expect(0, 43360, '\p{Is_Scx= 	Rjng}', "");
Expect(1, 43360, '\p{^Is_Scx= 	Rjng}', "");
Expect(1, 43360, '\P{Is_Scx= 	Rjng}', "");
Expect(0, 43360, '\P{^Is_Scx= 	Rjng}', "");
Error('\p{Script_Extensions=_RUNIC/a/}');
Error('\P{Script_Extensions=_RUNIC/a/}');
Expect(1, 5872, '\p{Script_Extensions=runic}', "");
Expect(0, 5872, '\p{^Script_Extensions=runic}', "");
Expect(0, 5872, '\P{Script_Extensions=runic}', "");
Expect(1, 5872, '\P{^Script_Extensions=runic}', "");
Expect(0, 5873, '\p{Script_Extensions=runic}', "");
Expect(1, 5873, '\p{^Script_Extensions=runic}', "");
Expect(1, 5873, '\P{Script_Extensions=runic}', "");
Expect(0, 5873, '\P{^Script_Extensions=runic}', "");
Expect(1, 5872, '\p{Script_Extensions=	 runic}', "");
Expect(0, 5872, '\p{^Script_Extensions=	 runic}', "");
Expect(0, 5872, '\P{Script_Extensions=	 runic}', "");
Expect(1, 5872, '\P{^Script_Extensions=	 runic}', "");
Expect(0, 5873, '\p{Script_Extensions=	 runic}', "");
Expect(1, 5873, '\p{^Script_Extensions=	 runic}', "");
Expect(1, 5873, '\P{Script_Extensions=	 runic}', "");
Expect(0, 5873, '\P{^Script_Extensions=	 runic}', "");
Error('\p{Scx=_/a/RUNR}');
Error('\P{Scx=_/a/RUNR}');
Expect(1, 5872, '\p{Scx=runr}', "");
Expect(0, 5872, '\p{^Scx=runr}', "");
Expect(0, 5872, '\P{Scx=runr}', "");
Expect(1, 5872, '\P{^Scx=runr}', "");
Expect(0, 5873, '\p{Scx=runr}', "");
Expect(1, 5873, '\p{^Scx=runr}', "");
Expect(1, 5873, '\P{Scx=runr}', "");
Expect(0, 5873, '\P{^Scx=runr}', "");
Expect(1, 5872, '\p{Scx=	 runr}', "");
Expect(0, 5872, '\p{^Scx=	 runr}', "");
Expect(0, 5872, '\P{Scx=	 runr}', "");
Expect(1, 5872, '\P{^Scx=	 runr}', "");
Expect(0, 5873, '\p{Scx=	 runr}', "");
Expect(1, 5873, '\p{^Scx=	 runr}', "");
Expect(1, 5873, '\P{Scx=	 runr}', "");
Expect(0, 5873, '\P{^Scx=	 runr}', "");
Error('\p{Is_Script_Extensions= Runic/a/}');
Error('\P{Is_Script_Extensions= Runic/a/}');
Expect(1, 5872, '\p{Is_Script_Extensions=runic}', "");
Expect(0, 5872, '\p{^Is_Script_Extensions=runic}', "");
Expect(0, 5872, '\P{Is_Script_Extensions=runic}', "");
Expect(1, 5872, '\P{^Is_Script_Extensions=runic}', "");
Expect(0, 5873, '\p{Is_Script_Extensions=runic}', "");
Expect(1, 5873, '\p{^Is_Script_Extensions=runic}', "");
Expect(1, 5873, '\P{Is_Script_Extensions=runic}', "");
Expect(0, 5873, '\P{^Is_Script_Extensions=runic}', "");
Expect(1, 5872, '\p{Is_Script_Extensions=	Runic}', "");
Expect(0, 5872, '\p{^Is_Script_Extensions=	Runic}', "");
Expect(0, 5872, '\P{Is_Script_Extensions=	Runic}', "");
Expect(1, 5872, '\P{^Is_Script_Extensions=	Runic}', "");
Expect(0, 5873, '\p{Is_Script_Extensions=	Runic}', "");
Expect(1, 5873, '\p{^Is_Script_Extensions=	Runic}', "");
Expect(1, 5873, '\P{Is_Script_Extensions=	Runic}', "");
Expect(0, 5873, '\P{^Is_Script_Extensions=	Runic}', "");
Error('\p{Is_Scx=-/a/RUNR}');
Error('\P{Is_Scx=-/a/RUNR}');
Expect(1, 5872, '\p{Is_Scx=runr}', "");
Expect(0, 5872, '\p{^Is_Scx=runr}', "");
Expect(0, 5872, '\P{Is_Scx=runr}', "");
Expect(1, 5872, '\P{^Is_Scx=runr}', "");
Expect(0, 5873, '\p{Is_Scx=runr}', "");
Expect(1, 5873, '\p{^Is_Scx=runr}', "");
Expect(1, 5873, '\P{Is_Scx=runr}', "");
Expect(0, 5873, '\P{^Is_Scx=runr}', "");
Expect(1, 5872, '\p{Is_Scx=_RUNR}', "");
Expect(0, 5872, '\p{^Is_Scx=_RUNR}', "");
Expect(0, 5872, '\P{Is_Scx=_RUNR}', "");
Expect(1, 5872, '\P{^Is_Scx=_RUNR}', "");
Expect(0, 5873, '\p{Is_Scx=_RUNR}', "");
Expect(1, 5873, '\p{^Is_Scx=_RUNR}', "");
Expect(1, 5873, '\P{Is_Scx=_RUNR}', "");
Expect(0, 5873, '\P{^Is_Scx=_RUNR}', "");
Error('\p{Script_Extensions=/a/ Samaritan}');
Error('\P{Script_Extensions=/a/ Samaritan}');
Expect(1, 2110, '\p{Script_Extensions=samaritan}', "");
Expect(0, 2110, '\p{^Script_Extensions=samaritan}', "");
Expect(0, 2110, '\P{Script_Extensions=samaritan}', "");
Expect(1, 2110, '\P{^Script_Extensions=samaritan}', "");
Expect(0, 2111, '\p{Script_Extensions=samaritan}', "");
Expect(1, 2111, '\p{^Script_Extensions=samaritan}', "");
Expect(1, 2111, '\P{Script_Extensions=samaritan}', "");
Expect(0, 2111, '\P{^Script_Extensions=samaritan}', "");
Expect(1, 2110, '\p{Script_Extensions=	 samaritan}', "");
Expect(0, 2110, '\p{^Script_Extensions=	 samaritan}', "");
Expect(0, 2110, '\P{Script_Extensions=	 samaritan}', "");
Expect(1, 2110, '\P{^Script_Extensions=	 samaritan}', "");
Expect(0, 2111, '\p{Script_Extensions=	 samaritan}', "");
Expect(1, 2111, '\p{^Script_Extensions=	 samaritan}', "");
Expect(1, 2111, '\P{Script_Extensions=	 samaritan}', "");
Expect(0, 2111, '\P{^Script_Extensions=	 samaritan}', "");
Error('\p{Scx=-_Samr:=}');
Error('\P{Scx=-_Samr:=}');
Expect(1, 2110, '\p{Scx=samr}', "");
Expect(0, 2110, '\p{^Scx=samr}', "");
Expect(0, 2110, '\P{Scx=samr}', "");
Expect(1, 2110, '\P{^Scx=samr}', "");
Expect(0, 2111, '\p{Scx=samr}', "");
Expect(1, 2111, '\p{^Scx=samr}', "");
Expect(1, 2111, '\P{Scx=samr}', "");
Expect(0, 2111, '\P{^Scx=samr}', "");
Expect(1, 2110, '\p{Scx= 	Samr}', "");
Expect(0, 2110, '\p{^Scx= 	Samr}', "");
Expect(0, 2110, '\P{Scx= 	Samr}', "");
Expect(1, 2110, '\P{^Scx= 	Samr}', "");
Expect(0, 2111, '\p{Scx= 	Samr}', "");
Expect(1, 2111, '\p{^Scx= 	Samr}', "");
Expect(1, 2111, '\P{Scx= 	Samr}', "");
Expect(0, 2111, '\P{^Scx= 	Samr}', "");
Error('\p{Is_Script_Extensions=/a/__Samaritan}');
Error('\P{Is_Script_Extensions=/a/__Samaritan}');
Expect(1, 2110, '\p{Is_Script_Extensions=samaritan}', "");
Expect(0, 2110, '\p{^Is_Script_Extensions=samaritan}', "");
Expect(0, 2110, '\P{Is_Script_Extensions=samaritan}', "");
Expect(1, 2110, '\P{^Is_Script_Extensions=samaritan}', "");
Expect(0, 2111, '\p{Is_Script_Extensions=samaritan}', "");
Expect(1, 2111, '\p{^Is_Script_Extensions=samaritan}', "");
Expect(1, 2111, '\P{Is_Script_Extensions=samaritan}', "");
Expect(0, 2111, '\P{^Is_Script_Extensions=samaritan}', "");
Expect(1, 2110, '\p{Is_Script_Extensions: -	Samaritan}', "");
Expect(0, 2110, '\p{^Is_Script_Extensions: -	Samaritan}', "");
Expect(0, 2110, '\P{Is_Script_Extensions: -	Samaritan}', "");
Expect(1, 2110, '\P{^Is_Script_Extensions: -	Samaritan}', "");
Expect(0, 2111, '\p{Is_Script_Extensions: -	Samaritan}', "");
Expect(1, 2111, '\p{^Is_Script_Extensions: -	Samaritan}', "");
Expect(1, 2111, '\P{Is_Script_Extensions: -	Samaritan}', "");
Expect(0, 2111, '\P{^Is_Script_Extensions: -	Samaritan}', "");
Error('\p{Is_Scx=-	SAMR/a/}');
Error('\P{Is_Scx=-	SAMR/a/}');
Expect(1, 2110, '\p{Is_Scx=samr}', "");
Expect(0, 2110, '\p{^Is_Scx=samr}', "");
Expect(0, 2110, '\P{Is_Scx=samr}', "");
Expect(1, 2110, '\P{^Is_Scx=samr}', "");
Expect(0, 2111, '\p{Is_Scx=samr}', "");
Expect(1, 2111, '\p{^Is_Scx=samr}', "");
Expect(1, 2111, '\P{Is_Scx=samr}', "");
Expect(0, 2111, '\P{^Is_Scx=samr}', "");
Expect(1, 2110, '\p{Is_Scx=	_Samr}', "");
Expect(0, 2110, '\p{^Is_Scx=	_Samr}', "");
Expect(0, 2110, '\P{Is_Scx=	_Samr}', "");
Expect(1, 2110, '\P{^Is_Scx=	_Samr}', "");
Expect(0, 2111, '\p{Is_Scx=	_Samr}', "");
Expect(1, 2111, '\p{^Is_Scx=	_Samr}', "");
Expect(1, 2111, '\P{Is_Scx=	_Samr}', "");
Expect(0, 2111, '\P{^Is_Scx=	_Samr}', "");
Error('\p{Script_Extensions=/a/--Old_South_Arabian}');
Error('\P{Script_Extensions=/a/--Old_South_Arabian}');
Expect(1, 68223, '\p{Script_Extensions=oldsoutharabian}', "");
Expect(0, 68223, '\p{^Script_Extensions=oldsoutharabian}', "");
Expect(0, 68223, '\P{Script_Extensions=oldsoutharabian}', "");
Expect(1, 68223, '\P{^Script_Extensions=oldsoutharabian}', "");
Expect(0, 68224, '\p{Script_Extensions=oldsoutharabian}', "");
Expect(1, 68224, '\p{^Script_Extensions=oldsoutharabian}', "");
Expect(1, 68224, '\P{Script_Extensions=oldsoutharabian}', "");
Expect(0, 68224, '\P{^Script_Extensions=oldsoutharabian}', "");
Expect(1, 68223, '\p{Script_Extensions=	_OLD_South_ARABIAN}', "");
Expect(0, 68223, '\p{^Script_Extensions=	_OLD_South_ARABIAN}', "");
Expect(0, 68223, '\P{Script_Extensions=	_OLD_South_ARABIAN}', "");
Expect(1, 68223, '\P{^Script_Extensions=	_OLD_South_ARABIAN}', "");
Expect(0, 68224, '\p{Script_Extensions=	_OLD_South_ARABIAN}', "");
Expect(1, 68224, '\p{^Script_Extensions=	_OLD_South_ARABIAN}', "");
Expect(1, 68224, '\P{Script_Extensions=	_OLD_South_ARABIAN}', "");
Expect(0, 68224, '\P{^Script_Extensions=	_OLD_South_ARABIAN}', "");
Error('\p{Scx=-Sarb:=}');
Error('\P{Scx=-Sarb:=}');
Expect(1, 68223, '\p{Scx=sarb}', "");
Expect(0, 68223, '\p{^Scx=sarb}', "");
Expect(0, 68223, '\P{Scx=sarb}', "");
Expect(1, 68223, '\P{^Scx=sarb}', "");
Expect(0, 68224, '\p{Scx=sarb}', "");
Expect(1, 68224, '\p{^Scx=sarb}', "");
Expect(1, 68224, '\P{Scx=sarb}', "");
Expect(0, 68224, '\P{^Scx=sarb}', "");
Expect(1, 68223, '\p{Scx:   _ Sarb}', "");
Expect(0, 68223, '\p{^Scx:   _ Sarb}', "");
Expect(0, 68223, '\P{Scx:   _ Sarb}', "");
Expect(1, 68223, '\P{^Scx:   _ Sarb}', "");
Expect(0, 68224, '\p{Scx:   _ Sarb}', "");
Expect(1, 68224, '\p{^Scx:   _ Sarb}', "");
Expect(1, 68224, '\P{Scx:   _ Sarb}', "");
Expect(0, 68224, '\P{^Scx:   _ Sarb}', "");
Error('\p{Is_Script_Extensions=- old_south_Arabian:=}');
Error('\P{Is_Script_Extensions=- old_south_Arabian:=}');
Expect(1, 68223, '\p{Is_Script_Extensions=oldsoutharabian}', "");
Expect(0, 68223, '\p{^Is_Script_Extensions=oldsoutharabian}', "");
Expect(0, 68223, '\P{Is_Script_Extensions=oldsoutharabian}', "");
Expect(1, 68223, '\P{^Is_Script_Extensions=oldsoutharabian}', "");
Expect(0, 68224, '\p{Is_Script_Extensions=oldsoutharabian}', "");
Expect(1, 68224, '\p{^Is_Script_Extensions=oldsoutharabian}', "");
Expect(1, 68224, '\P{Is_Script_Extensions=oldsoutharabian}', "");
Expect(0, 68224, '\P{^Is_Script_Extensions=oldsoutharabian}', "");
Expect(1, 68223, '\p{Is_Script_Extensions=_-Old_south_ARABIAN}', "");
Expect(0, 68223, '\p{^Is_Script_Extensions=_-Old_south_ARABIAN}', "");
Expect(0, 68223, '\P{Is_Script_Extensions=_-Old_south_ARABIAN}', "");
Expect(1, 68223, '\P{^Is_Script_Extensions=_-Old_south_ARABIAN}', "");
Expect(0, 68224, '\p{Is_Script_Extensions=_-Old_south_ARABIAN}', "");
Expect(1, 68224, '\p{^Is_Script_Extensions=_-Old_south_ARABIAN}', "");
Expect(1, 68224, '\P{Is_Script_Extensions=_-Old_south_ARABIAN}', "");
Expect(0, 68224, '\P{^Is_Script_Extensions=_-Old_south_ARABIAN}', "");
Error('\p{Is_Scx= /a/SARB}');
Error('\P{Is_Scx= /a/SARB}');
Expect(1, 68223, '\p{Is_Scx=sarb}', "");
Expect(0, 68223, '\p{^Is_Scx=sarb}', "");
Expect(0, 68223, '\P{Is_Scx=sarb}', "");
Expect(1, 68223, '\P{^Is_Scx=sarb}', "");
Expect(0, 68224, '\p{Is_Scx=sarb}', "");
Expect(1, 68224, '\p{^Is_Scx=sarb}', "");
Expect(1, 68224, '\P{Is_Scx=sarb}', "");
Expect(0, 68224, '\P{^Is_Scx=sarb}', "");
Expect(1, 68223, '\p{Is_Scx=	_sarb}', "");
Expect(0, 68223, '\p{^Is_Scx=	_sarb}', "");
Expect(0, 68223, '\P{Is_Scx=	_sarb}', "");
Expect(1, 68223, '\P{^Is_Scx=	_sarb}', "");
Expect(0, 68224, '\p{Is_Scx=	_sarb}', "");
Expect(1, 68224, '\p{^Is_Scx=	_sarb}', "");
Expect(1, 68224, '\P{Is_Scx=	_sarb}', "");
Expect(0, 68224, '\P{^Is_Scx=	_sarb}', "");
Error('\p{Script_Extensions:   :=Saurashtra}');
Error('\P{Script_Extensions:   :=Saurashtra}');
Expect(1, 43225, '\p{Script_Extensions=saurashtra}', "");
Expect(0, 43225, '\p{^Script_Extensions=saurashtra}', "");
Expect(0, 43225, '\P{Script_Extensions=saurashtra}', "");
Expect(1, 43225, '\P{^Script_Extensions=saurashtra}', "");
Expect(0, 43226, '\p{Script_Extensions=saurashtra}', "");
Expect(1, 43226, '\p{^Script_Extensions=saurashtra}', "");
Expect(1, 43226, '\P{Script_Extensions=saurashtra}', "");
Expect(0, 43226, '\P{^Script_Extensions=saurashtra}', "");
Expect(1, 43225, '\p{Script_Extensions=	 Saurashtra}', "");
Expect(0, 43225, '\p{^Script_Extensions=	 Saurashtra}', "");
Expect(0, 43225, '\P{Script_Extensions=	 Saurashtra}', "");
Expect(1, 43225, '\P{^Script_Extensions=	 Saurashtra}', "");
Expect(0, 43226, '\p{Script_Extensions=	 Saurashtra}', "");
Expect(1, 43226, '\p{^Script_Extensions=	 Saurashtra}', "");
Expect(1, 43226, '\P{Script_Extensions=	 Saurashtra}', "");
Expect(0, 43226, '\P{^Script_Extensions=	 Saurashtra}', "");
Error('\p{Scx=/a/__SAUR}');
Error('\P{Scx=/a/__SAUR}');
Expect(1, 43225, '\p{Scx=saur}', "");
Expect(0, 43225, '\p{^Scx=saur}', "");
Expect(0, 43225, '\P{Scx=saur}', "");
Expect(1, 43225, '\P{^Scx=saur}', "");
Expect(0, 43226, '\p{Scx=saur}', "");
Expect(1, 43226, '\p{^Scx=saur}', "");
Expect(1, 43226, '\P{Scx=saur}', "");
Expect(0, 43226, '\P{^Scx=saur}', "");
Expect(1, 43225, '\p{Scx=_	saur}', "");
Expect(0, 43225, '\p{^Scx=_	saur}', "");
Expect(0, 43225, '\P{Scx=_	saur}', "");
Expect(1, 43225, '\P{^Scx=_	saur}', "");
Expect(0, 43226, '\p{Scx=_	saur}', "");
Expect(1, 43226, '\p{^Scx=_	saur}', "");
Expect(1, 43226, '\P{Scx=_	saur}', "");
Expect(0, 43226, '\P{^Scx=_	saur}', "");
Error('\p{Is_Script_Extensions=:= _Saurashtra}');
Error('\P{Is_Script_Extensions=:= _Saurashtra}');
Expect(1, 43225, '\p{Is_Script_Extensions=saurashtra}', "");
Expect(0, 43225, '\p{^Is_Script_Extensions=saurashtra}', "");
Expect(0, 43225, '\P{Is_Script_Extensions=saurashtra}', "");
Expect(1, 43225, '\P{^Is_Script_Extensions=saurashtra}', "");
Expect(0, 43226, '\p{Is_Script_Extensions=saurashtra}', "");
Expect(1, 43226, '\p{^Is_Script_Extensions=saurashtra}', "");
Expect(1, 43226, '\P{Is_Script_Extensions=saurashtra}', "");
Expect(0, 43226, '\P{^Is_Script_Extensions=saurashtra}', "");
Expect(1, 43225, '\p{Is_Script_Extensions=_	SAURASHTRA}', "");
Expect(0, 43225, '\p{^Is_Script_Extensions=_	SAURASHTRA}', "");
Expect(0, 43225, '\P{Is_Script_Extensions=_	SAURASHTRA}', "");
Expect(1, 43225, '\P{^Is_Script_Extensions=_	SAURASHTRA}', "");
Expect(0, 43226, '\p{Is_Script_Extensions=_	SAURASHTRA}', "");
Expect(1, 43226, '\p{^Is_Script_Extensions=_	SAURASHTRA}', "");
Expect(1, 43226, '\P{Is_Script_Extensions=_	SAURASHTRA}', "");
Expect(0, 43226, '\P{^Is_Script_Extensions=_	SAURASHTRA}', "");
Error('\p{Is_Scx:_-Saur:=}');
Error('\P{Is_Scx:_-Saur:=}');
Expect(1, 43225, '\p{Is_Scx=saur}', "");
Expect(0, 43225, '\p{^Is_Scx=saur}', "");
Expect(0, 43225, '\P{Is_Scx=saur}', "");
Expect(1, 43225, '\P{^Is_Scx=saur}', "");
Expect(0, 43226, '\p{Is_Scx=saur}', "");
Expect(1, 43226, '\p{^Is_Scx=saur}', "");
Expect(1, 43226, '\P{Is_Scx=saur}', "");
Expect(0, 43226, '\P{^Is_Scx=saur}', "");
Expect(1, 43225, '\p{Is_Scx=-SAUR}', "");
Expect(0, 43225, '\p{^Is_Scx=-SAUR}', "");
Expect(0, 43225, '\P{Is_Scx=-SAUR}', "");
Expect(1, 43225, '\P{^Is_Scx=-SAUR}', "");
Expect(0, 43226, '\p{Is_Scx=-SAUR}', "");
Expect(1, 43226, '\p{^Is_Scx=-SAUR}', "");
Expect(1, 43226, '\P{Is_Scx=-SAUR}', "");
Expect(0, 43226, '\P{^Is_Scx=-SAUR}', "");
Error('\p{Script_Extensions=:=	Shavian}');
Error('\P{Script_Extensions=:=	Shavian}');
Expect(1, 66687, '\p{Script_Extensions=shavian}', "");
Expect(0, 66687, '\p{^Script_Extensions=shavian}', "");
Expect(0, 66687, '\P{Script_Extensions=shavian}', "");
Expect(1, 66687, '\P{^Script_Extensions=shavian}', "");
Expect(0, 66688, '\p{Script_Extensions=shavian}', "");
Expect(1, 66688, '\p{^Script_Extensions=shavian}', "");
Expect(1, 66688, '\P{Script_Extensions=shavian}', "");
Expect(0, 66688, '\P{^Script_Extensions=shavian}', "");
Expect(1, 66687, '\p{Script_Extensions=_	Shavian}', "");
Expect(0, 66687, '\p{^Script_Extensions=_	Shavian}', "");
Expect(0, 66687, '\P{Script_Extensions=_	Shavian}', "");
Expect(1, 66687, '\P{^Script_Extensions=_	Shavian}', "");
Expect(0, 66688, '\p{Script_Extensions=_	Shavian}', "");
Expect(1, 66688, '\p{^Script_Extensions=_	Shavian}', "");
Expect(1, 66688, '\P{Script_Extensions=_	Shavian}', "");
Expect(0, 66688, '\P{^Script_Extensions=_	Shavian}', "");
Error('\p{Scx= _Shaw/a/}');
Error('\P{Scx= _Shaw/a/}');
Expect(1, 66687, '\p{Scx=shaw}', "");
Expect(0, 66687, '\p{^Scx=shaw}', "");
Expect(0, 66687, '\P{Scx=shaw}', "");
Expect(1, 66687, '\P{^Scx=shaw}', "");
Expect(0, 66688, '\p{Scx=shaw}', "");
Expect(1, 66688, '\p{^Scx=shaw}', "");
Expect(1, 66688, '\P{Scx=shaw}', "");
Expect(0, 66688, '\P{^Scx=shaw}', "");
Expect(1, 66687, '\p{Scx: -	shaw}', "");
Expect(0, 66687, '\p{^Scx: -	shaw}', "");
Expect(0, 66687, '\P{Scx: -	shaw}', "");
Expect(1, 66687, '\P{^Scx: -	shaw}', "");
Expect(0, 66688, '\p{Scx: -	shaw}', "");
Expect(1, 66688, '\p{^Scx: -	shaw}', "");
Expect(1, 66688, '\P{Scx: -	shaw}', "");
Expect(0, 66688, '\P{^Scx: -	shaw}', "");
Error('\p{Is_Script_Extensions:   	Shavian:=}');
Error('\P{Is_Script_Extensions:   	Shavian:=}');
Expect(1, 66687, '\p{Is_Script_Extensions=shavian}', "");
Expect(0, 66687, '\p{^Is_Script_Extensions=shavian}', "");
Expect(0, 66687, '\P{Is_Script_Extensions=shavian}', "");
Expect(1, 66687, '\P{^Is_Script_Extensions=shavian}', "");
Expect(0, 66688, '\p{Is_Script_Extensions=shavian}', "");
Expect(1, 66688, '\p{^Is_Script_Extensions=shavian}', "");
Expect(1, 66688, '\P{Is_Script_Extensions=shavian}', "");
Expect(0, 66688, '\P{^Is_Script_Extensions=shavian}', "");
Expect(1, 66687, '\p{Is_Script_Extensions:    shavian}', "");
Expect(0, 66687, '\p{^Is_Script_Extensions:    shavian}', "");
Expect(0, 66687, '\P{Is_Script_Extensions:    shavian}', "");
Expect(1, 66687, '\P{^Is_Script_Extensions:    shavian}', "");
Expect(0, 66688, '\p{Is_Script_Extensions:    shavian}', "");
Expect(1, 66688, '\p{^Is_Script_Extensions:    shavian}', "");
Expect(1, 66688, '\P{Is_Script_Extensions:    shavian}', "");
Expect(0, 66688, '\P{^Is_Script_Extensions:    shavian}', "");
Error('\p{Is_Scx:   /a/	 SHAW}');
Error('\P{Is_Scx:   /a/	 SHAW}');
Expect(1, 66687, '\p{Is_Scx:   shaw}', "");
Expect(0, 66687, '\p{^Is_Scx:   shaw}', "");
Expect(0, 66687, '\P{Is_Scx:   shaw}', "");
Expect(1, 66687, '\P{^Is_Scx:   shaw}', "");
Expect(0, 66688, '\p{Is_Scx:   shaw}', "");
Expect(1, 66688, '\p{^Is_Scx:   shaw}', "");
Expect(1, 66688, '\P{Is_Scx:   shaw}', "");
Expect(0, 66688, '\P{^Is_Scx:   shaw}', "");
Expect(1, 66687, '\p{Is_Scx=Shaw}', "");
Expect(0, 66687, '\p{^Is_Scx=Shaw}', "");
Expect(0, 66687, '\P{Is_Scx=Shaw}', "");
Expect(1, 66687, '\P{^Is_Scx=Shaw}', "");
Expect(0, 66688, '\p{Is_Scx=Shaw}', "");
Expect(1, 66688, '\p{^Is_Scx=Shaw}', "");
Expect(1, 66688, '\P{Is_Scx=Shaw}', "");
Expect(0, 66688, '\P{^Is_Scx=Shaw}', "");
Error('\p{Script_Extensions=/a/ sharada}');
Error('\P{Script_Extensions=/a/ sharada}');
Expect(1, 70105, '\p{Script_Extensions=sharada}', "");
Expect(0, 70105, '\p{^Script_Extensions=sharada}', "");
Expect(0, 70105, '\P{Script_Extensions=sharada}', "");
Expect(1, 70105, '\P{^Script_Extensions=sharada}', "");
Expect(0, 70106, '\p{Script_Extensions=sharada}', "");
Expect(1, 70106, '\p{^Script_Extensions=sharada}', "");
Expect(1, 70106, '\P{Script_Extensions=sharada}', "");
Expect(0, 70106, '\P{^Script_Extensions=sharada}', "");
Expect(1, 70105, '\p{Script_Extensions: 	_Sharada}', "");
Expect(0, 70105, '\p{^Script_Extensions: 	_Sharada}', "");
Expect(0, 70105, '\P{Script_Extensions: 	_Sharada}', "");
Expect(1, 70105, '\P{^Script_Extensions: 	_Sharada}', "");
Expect(0, 70106, '\p{Script_Extensions: 	_Sharada}', "");
Expect(1, 70106, '\p{^Script_Extensions: 	_Sharada}', "");
Expect(1, 70106, '\P{Script_Extensions: 	_Sharada}', "");
Expect(0, 70106, '\P{^Script_Extensions: 	_Sharada}', "");
Error('\p{Scx=/a/Shrd}');
Error('\P{Scx=/a/Shrd}');
Expect(1, 70105, '\p{Scx=shrd}', "");
Expect(0, 70105, '\p{^Scx=shrd}', "");
Expect(0, 70105, '\P{Scx=shrd}', "");
Expect(1, 70105, '\P{^Scx=shrd}', "");
Expect(0, 70106, '\p{Scx=shrd}', "");
Expect(1, 70106, '\p{^Scx=shrd}', "");
Expect(1, 70106, '\P{Scx=shrd}', "");
Expect(0, 70106, '\P{^Scx=shrd}', "");
Expect(1, 70105, '\p{Scx= 	Shrd}', "");
Expect(0, 70105, '\p{^Scx= 	Shrd}', "");
Expect(0, 70105, '\P{Scx= 	Shrd}', "");
Expect(1, 70105, '\P{^Scx= 	Shrd}', "");
Expect(0, 70106, '\p{Scx= 	Shrd}', "");
Expect(1, 70106, '\p{^Scx= 	Shrd}', "");
Expect(1, 70106, '\P{Scx= 	Shrd}', "");
Expect(0, 70106, '\P{^Scx= 	Shrd}', "");
Error('\p{Is_Script_Extensions=--Sharada/a/}');
Error('\P{Is_Script_Extensions=--Sharada/a/}');
Expect(1, 70105, '\p{Is_Script_Extensions=sharada}', "");
Expect(0, 70105, '\p{^Is_Script_Extensions=sharada}', "");
Expect(0, 70105, '\P{Is_Script_Extensions=sharada}', "");
Expect(1, 70105, '\P{^Is_Script_Extensions=sharada}', "");
Expect(0, 70106, '\p{Is_Script_Extensions=sharada}', "");
Expect(1, 70106, '\p{^Is_Script_Extensions=sharada}', "");
Expect(1, 70106, '\P{Is_Script_Extensions=sharada}', "");
Expect(0, 70106, '\P{^Is_Script_Extensions=sharada}', "");
Expect(1, 70105, '\p{Is_Script_Extensions=_-Sharada}', "");
Expect(0, 70105, '\p{^Is_Script_Extensions=_-Sharada}', "");
Expect(0, 70105, '\P{Is_Script_Extensions=_-Sharada}', "");
Expect(1, 70105, '\P{^Is_Script_Extensions=_-Sharada}', "");
Expect(0, 70106, '\p{Is_Script_Extensions=_-Sharada}', "");
Expect(1, 70106, '\p{^Is_Script_Extensions=_-Sharada}', "");
Expect(1, 70106, '\P{Is_Script_Extensions=_-Sharada}', "");
Expect(0, 70106, '\P{^Is_Script_Extensions=_-Sharada}', "");
Error('\p{Is_Scx:_SHRD:=}');
Error('\P{Is_Scx:_SHRD:=}');
Expect(1, 70105, '\p{Is_Scx=shrd}', "");
Expect(0, 70105, '\p{^Is_Scx=shrd}', "");
Expect(0, 70105, '\P{Is_Scx=shrd}', "");
Expect(1, 70105, '\P{^Is_Scx=shrd}', "");
Expect(0, 70106, '\p{Is_Scx=shrd}', "");
Expect(1, 70106, '\p{^Is_Scx=shrd}', "");
Expect(1, 70106, '\P{Is_Scx=shrd}', "");
Expect(0, 70106, '\P{^Is_Scx=shrd}', "");
Expect(1, 70105, '\p{Is_Scx=__shrd}', "");
Expect(0, 70105, '\p{^Is_Scx=__shrd}', "");
Expect(0, 70105, '\P{Is_Scx=__shrd}', "");
Expect(1, 70105, '\P{^Is_Scx=__shrd}', "");
Expect(0, 70106, '\p{Is_Scx=__shrd}', "");
Expect(1, 70106, '\p{^Is_Scx=__shrd}', "");
Expect(1, 70106, '\P{Is_Scx=__shrd}', "");
Expect(0, 70106, '\P{^Is_Scx=__shrd}', "");
Error('\p{Script_Extensions=_/a/SINHALA}');
Error('\P{Script_Extensions=_/a/SINHALA}');
Expect(1, 3572, '\p{Script_Extensions=sinhala}', "");
Expect(0, 3572, '\p{^Script_Extensions=sinhala}', "");
Expect(0, 3572, '\P{Script_Extensions=sinhala}', "");
Expect(1, 3572, '\P{^Script_Extensions=sinhala}', "");
Expect(0, 3573, '\p{Script_Extensions=sinhala}', "");
Expect(1, 3573, '\p{^Script_Extensions=sinhala}', "");
Expect(1, 3573, '\P{Script_Extensions=sinhala}', "");
Expect(0, 3573, '\P{^Script_Extensions=sinhala}', "");
Expect(1, 3572, '\p{Script_Extensions=	sinhala}', "");
Expect(0, 3572, '\p{^Script_Extensions=	sinhala}', "");
Expect(0, 3572, '\P{Script_Extensions=	sinhala}', "");
Expect(1, 3572, '\P{^Script_Extensions=	sinhala}', "");
Expect(0, 3573, '\p{Script_Extensions=	sinhala}', "");
Expect(1, 3573, '\p{^Script_Extensions=	sinhala}', "");
Expect(1, 3573, '\P{Script_Extensions=	sinhala}', "");
Expect(0, 3573, '\P{^Script_Extensions=	sinhala}', "");
Error('\p{Scx=:= 	sinh}');
Error('\P{Scx=:= 	sinh}');
Expect(1, 3572, '\p{Scx=sinh}', "");
Expect(0, 3572, '\p{^Scx=sinh}', "");
Expect(0, 3572, '\P{Scx=sinh}', "");
Expect(1, 3572, '\P{^Scx=sinh}', "");
Expect(0, 3573, '\p{Scx=sinh}', "");
Expect(1, 3573, '\p{^Scx=sinh}', "");
Expect(1, 3573, '\P{Scx=sinh}', "");
Expect(0, 3573, '\P{^Scx=sinh}', "");
Expect(1, 3572, '\p{Scx=	-Sinh}', "");
Expect(0, 3572, '\p{^Scx=	-Sinh}', "");
Expect(0, 3572, '\P{Scx=	-Sinh}', "");
Expect(1, 3572, '\P{^Scx=	-Sinh}', "");
Expect(0, 3573, '\p{Scx=	-Sinh}', "");
Expect(1, 3573, '\p{^Scx=	-Sinh}', "");
Expect(1, 3573, '\P{Scx=	-Sinh}', "");
Expect(0, 3573, '\P{^Scx=	-Sinh}', "");
Error('\p{Is_Script_Extensions= :=SINHALA}');
Error('\P{Is_Script_Extensions= :=SINHALA}');
Expect(1, 3572, '\p{Is_Script_Extensions=sinhala}', "");
Expect(0, 3572, '\p{^Is_Script_Extensions=sinhala}', "");
Expect(0, 3572, '\P{Is_Script_Extensions=sinhala}', "");
Expect(1, 3572, '\P{^Is_Script_Extensions=sinhala}', "");
Expect(0, 3573, '\p{Is_Script_Extensions=sinhala}', "");
Expect(1, 3573, '\p{^Is_Script_Extensions=sinhala}', "");
Expect(1, 3573, '\P{Is_Script_Extensions=sinhala}', "");
Expect(0, 3573, '\P{^Is_Script_Extensions=sinhala}', "");
Expect(1, 3572, '\p{Is_Script_Extensions=_ Sinhala}', "");
Expect(0, 3572, '\p{^Is_Script_Extensions=_ Sinhala}', "");
Expect(0, 3572, '\P{Is_Script_Extensions=_ Sinhala}', "");
Expect(1, 3572, '\P{^Is_Script_Extensions=_ Sinhala}', "");
Expect(0, 3573, '\p{Is_Script_Extensions=_ Sinhala}', "");
Expect(1, 3573, '\p{^Is_Script_Extensions=_ Sinhala}', "");
Expect(1, 3573, '\P{Is_Script_Extensions=_ Sinhala}', "");
Expect(0, 3573, '\P{^Is_Script_Extensions=_ Sinhala}', "");
Error('\p{Is_Scx: :=	sinh}');
Error('\P{Is_Scx: :=	sinh}');
Expect(1, 3572, '\p{Is_Scx=sinh}', "");
Expect(0, 3572, '\p{^Is_Scx=sinh}', "");
Expect(0, 3572, '\P{Is_Scx=sinh}', "");
Expect(1, 3572, '\P{^Is_Scx=sinh}', "");
Expect(0, 3573, '\p{Is_Scx=sinh}', "");
Expect(1, 3573, '\p{^Is_Scx=sinh}', "");
Expect(1, 3573, '\P{Is_Scx=sinh}', "");
Expect(0, 3573, '\P{^Is_Scx=sinh}', "");
Expect(1, 3572, '\p{Is_Scx=	sinh}', "");
Expect(0, 3572, '\p{^Is_Scx=	sinh}', "");
Expect(0, 3572, '\P{Is_Scx=	sinh}', "");
Expect(1, 3572, '\P{^Is_Scx=	sinh}', "");
Expect(0, 3573, '\p{Is_Scx=	sinh}', "");
Expect(1, 3573, '\p{^Is_Scx=	sinh}', "");
Expect(1, 3573, '\P{Is_Scx=	sinh}', "");
Expect(0, 3573, '\P{^Is_Scx=	sinh}', "");
Error('\p{Script_Extensions=/a/--Sora_Sompeng}');
Error('\P{Script_Extensions=/a/--Sora_Sompeng}');
Expect(1, 69881, '\p{Script_Extensions=sorasompeng}', "");
Expect(0, 69881, '\p{^Script_Extensions=sorasompeng}', "");
Expect(0, 69881, '\P{Script_Extensions=sorasompeng}', "");
Expect(1, 69881, '\P{^Script_Extensions=sorasompeng}', "");
Expect(0, 69882, '\p{Script_Extensions=sorasompeng}', "");
Expect(1, 69882, '\p{^Script_Extensions=sorasompeng}', "");
Expect(1, 69882, '\P{Script_Extensions=sorasompeng}', "");
Expect(0, 69882, '\P{^Script_Extensions=sorasompeng}', "");
Expect(1, 69881, '\p{Script_Extensions=-_SORA_sompeng}', "");
Expect(0, 69881, '\p{^Script_Extensions=-_SORA_sompeng}', "");
Expect(0, 69881, '\P{Script_Extensions=-_SORA_sompeng}', "");
Expect(1, 69881, '\P{^Script_Extensions=-_SORA_sompeng}', "");
Expect(0, 69882, '\p{Script_Extensions=-_SORA_sompeng}', "");
Expect(1, 69882, '\p{^Script_Extensions=-_SORA_sompeng}', "");
Expect(1, 69882, '\P{Script_Extensions=-_SORA_sompeng}', "");
Expect(0, 69882, '\P{^Script_Extensions=-_SORA_sompeng}', "");
Error('\p{Scx=:=sora}');
Error('\P{Scx=:=sora}');
Expect(1, 69881, '\p{Scx=sora}', "");
Expect(0, 69881, '\p{^Scx=sora}', "");
Expect(0, 69881, '\P{Scx=sora}', "");
Expect(1, 69881, '\P{^Scx=sora}', "");
Expect(0, 69882, '\p{Scx=sora}', "");
Expect(1, 69882, '\p{^Scx=sora}', "");
Expect(1, 69882, '\P{Scx=sora}', "");
Expect(0, 69882, '\P{^Scx=sora}', "");
Expect(1, 69881, '\p{Scx=	sora}', "");
Expect(0, 69881, '\p{^Scx=	sora}', "");
Expect(0, 69881, '\P{Scx=	sora}', "");
Expect(1, 69881, '\P{^Scx=	sora}', "");
Expect(0, 69882, '\p{Scx=	sora}', "");
Expect(1, 69882, '\p{^Scx=	sora}', "");
Expect(1, 69882, '\P{Scx=	sora}', "");
Expect(0, 69882, '\P{^Scx=	sora}', "");
Error('\p{Is_Script_Extensions=:=sora_SOMPENG}');
Error('\P{Is_Script_Extensions=:=sora_SOMPENG}');
Expect(1, 69881, '\p{Is_Script_Extensions=sorasompeng}', "");
Expect(0, 69881, '\p{^Is_Script_Extensions=sorasompeng}', "");
Expect(0, 69881, '\P{Is_Script_Extensions=sorasompeng}', "");
Expect(1, 69881, '\P{^Is_Script_Extensions=sorasompeng}', "");
Expect(0, 69882, '\p{Is_Script_Extensions=sorasompeng}', "");
Expect(1, 69882, '\p{^Is_Script_Extensions=sorasompeng}', "");
Expect(1, 69882, '\P{Is_Script_Extensions=sorasompeng}', "");
Expect(0, 69882, '\P{^Is_Script_Extensions=sorasompeng}', "");
Expect(1, 69881, '\p{Is_Script_Extensions=-sora_Sompeng}', "");
Expect(0, 69881, '\p{^Is_Script_Extensions=-sora_Sompeng}', "");
Expect(0, 69881, '\P{Is_Script_Extensions=-sora_Sompeng}', "");
Expect(1, 69881, '\P{^Is_Script_Extensions=-sora_Sompeng}', "");
Expect(0, 69882, '\p{Is_Script_Extensions=-sora_Sompeng}', "");
Expect(1, 69882, '\p{^Is_Script_Extensions=-sora_Sompeng}', "");
Expect(1, 69882, '\P{Is_Script_Extensions=-sora_Sompeng}', "");
Expect(0, 69882, '\P{^Is_Script_Extensions=-sora_Sompeng}', "");
Error('\p{Is_Scx=:=_	sora}');
Error('\P{Is_Scx=:=_	sora}');
Expect(1, 69881, '\p{Is_Scx=sora}', "");
Expect(0, 69881, '\p{^Is_Scx=sora}', "");
Expect(0, 69881, '\P{Is_Scx=sora}', "");
Expect(1, 69881, '\P{^Is_Scx=sora}', "");
Expect(0, 69882, '\p{Is_Scx=sora}', "");
Expect(1, 69882, '\p{^Is_Scx=sora}', "");
Expect(1, 69882, '\P{Is_Scx=sora}', "");
Expect(0, 69882, '\P{^Is_Scx=sora}', "");
Expect(1, 69881, '\p{Is_Scx= -Sora}', "");
Expect(0, 69881, '\p{^Is_Scx= -Sora}', "");
Expect(0, 69881, '\P{Is_Scx= -Sora}', "");
Expect(1, 69881, '\P{^Is_Scx= -Sora}', "");
Expect(0, 69882, '\p{Is_Scx= -Sora}', "");
Expect(1, 69882, '\p{^Is_Scx= -Sora}', "");
Expect(1, 69882, '\P{Is_Scx= -Sora}', "");
Expect(0, 69882, '\P{^Is_Scx= -Sora}', "");
Error('\p{Script_Extensions=/a/_ SUNDANESE}');
Error('\P{Script_Extensions=/a/_ SUNDANESE}');
Expect(1, 7367, '\p{Script_Extensions=sundanese}', "");
Expect(0, 7367, '\p{^Script_Extensions=sundanese}', "");
Expect(0, 7367, '\P{Script_Extensions=sundanese}', "");
Expect(1, 7367, '\P{^Script_Extensions=sundanese}', "");
Expect(0, 7368, '\p{Script_Extensions=sundanese}', "");
Expect(1, 7368, '\p{^Script_Extensions=sundanese}', "");
Expect(1, 7368, '\P{Script_Extensions=sundanese}', "");
Expect(0, 7368, '\P{^Script_Extensions=sundanese}', "");
Expect(1, 7367, '\p{Script_Extensions=	SUNDANESE}', "");
Expect(0, 7367, '\p{^Script_Extensions=	SUNDANESE}', "");
Expect(0, 7367, '\P{Script_Extensions=	SUNDANESE}', "");
Expect(1, 7367, '\P{^Script_Extensions=	SUNDANESE}', "");
Expect(0, 7368, '\p{Script_Extensions=	SUNDANESE}', "");
Expect(1, 7368, '\p{^Script_Extensions=	SUNDANESE}', "");
Expect(1, 7368, '\P{Script_Extensions=	SUNDANESE}', "");
Expect(0, 7368, '\P{^Script_Extensions=	SUNDANESE}', "");
Error('\p{Scx: _:=Sund}');
Error('\P{Scx: _:=Sund}');
Expect(1, 7367, '\p{Scx=sund}', "");
Expect(0, 7367, '\p{^Scx=sund}', "");
Expect(0, 7367, '\P{Scx=sund}', "");
Expect(1, 7367, '\P{^Scx=sund}', "");
Expect(0, 7368, '\p{Scx=sund}', "");
Expect(1, 7368, '\p{^Scx=sund}', "");
Expect(1, 7368, '\P{Scx=sund}', "");
Expect(0, 7368, '\P{^Scx=sund}', "");
Expect(1, 7367, '\p{Scx=-Sund}', "");
Expect(0, 7367, '\p{^Scx=-Sund}', "");
Expect(0, 7367, '\P{Scx=-Sund}', "");
Expect(1, 7367, '\P{^Scx=-Sund}', "");
Expect(0, 7368, '\p{Scx=-Sund}', "");
Expect(1, 7368, '\p{^Scx=-Sund}', "");
Expect(1, 7368, '\P{Scx=-Sund}', "");
Expect(0, 7368, '\P{^Scx=-Sund}', "");
Error('\p{Is_Script_Extensions=/a/ 	SUNDANESE}');
Error('\P{Is_Script_Extensions=/a/ 	SUNDANESE}');
Expect(1, 7367, '\p{Is_Script_Extensions=sundanese}', "");
Expect(0, 7367, '\p{^Is_Script_Extensions=sundanese}', "");
Expect(0, 7367, '\P{Is_Script_Extensions=sundanese}', "");
Expect(1, 7367, '\P{^Is_Script_Extensions=sundanese}', "");
Expect(0, 7368, '\p{Is_Script_Extensions=sundanese}', "");
Expect(1, 7368, '\p{^Is_Script_Extensions=sundanese}', "");
Expect(1, 7368, '\P{Is_Script_Extensions=sundanese}', "");
Expect(0, 7368, '\P{^Is_Script_Extensions=sundanese}', "");
Expect(1, 7367, '\p{Is_Script_Extensions: 	SUNDANESE}', "");
Expect(0, 7367, '\p{^Is_Script_Extensions: 	SUNDANESE}', "");
Expect(0, 7367, '\P{Is_Script_Extensions: 	SUNDANESE}', "");
Expect(1, 7367, '\P{^Is_Script_Extensions: 	SUNDANESE}', "");
Expect(0, 7368, '\p{Is_Script_Extensions: 	SUNDANESE}', "");
Expect(1, 7368, '\p{^Is_Script_Extensions: 	SUNDANESE}', "");
Expect(1, 7368, '\P{Is_Script_Extensions: 	SUNDANESE}', "");
Expect(0, 7368, '\P{^Is_Script_Extensions: 	SUNDANESE}', "");
Error('\p{Is_Scx=	:=Sund}');
Error('\P{Is_Scx=	:=Sund}');
Expect(1, 7367, '\p{Is_Scx=sund}', "");
Expect(0, 7367, '\p{^Is_Scx=sund}', "");
Expect(0, 7367, '\P{Is_Scx=sund}', "");
Expect(1, 7367, '\P{^Is_Scx=sund}', "");
Expect(0, 7368, '\p{Is_Scx=sund}', "");
Expect(1, 7368, '\p{^Is_Scx=sund}', "");
Expect(1, 7368, '\P{Is_Scx=sund}', "");
Expect(0, 7368, '\P{^Is_Scx=sund}', "");
Expect(1, 7367, '\p{Is_Scx=	-Sund}', "");
Expect(0, 7367, '\p{^Is_Scx=	-Sund}', "");
Expect(0, 7367, '\P{Is_Scx=	-Sund}', "");
Expect(1, 7367, '\P{^Is_Scx=	-Sund}', "");
Expect(0, 7368, '\p{Is_Scx=	-Sund}', "");
Expect(1, 7368, '\p{^Is_Scx=	-Sund}', "");
Expect(1, 7368, '\P{Is_Scx=	-Sund}', "");
Expect(0, 7368, '\P{^Is_Scx=	-Sund}', "");
Error('\p{Script_Extensions=:= 	syloti_Nagri}');
Error('\P{Script_Extensions=:= 	syloti_Nagri}');
Expect(1, 43051, '\p{Script_Extensions=sylotinagri}', "");
Expect(0, 43051, '\p{^Script_Extensions=sylotinagri}', "");
Expect(0, 43051, '\P{Script_Extensions=sylotinagri}', "");
Expect(1, 43051, '\P{^Script_Extensions=sylotinagri}', "");
Expect(0, 43052, '\p{Script_Extensions=sylotinagri}', "");
Expect(1, 43052, '\p{^Script_Extensions=sylotinagri}', "");
Expect(1, 43052, '\P{Script_Extensions=sylotinagri}', "");
Expect(0, 43052, '\P{^Script_Extensions=sylotinagri}', "");
Expect(1, 43051, '\p{Script_Extensions=	Syloti_Nagri}', "");
Expect(0, 43051, '\p{^Script_Extensions=	Syloti_Nagri}', "");
Expect(0, 43051, '\P{Script_Extensions=	Syloti_Nagri}', "");
Expect(1, 43051, '\P{^Script_Extensions=	Syloti_Nagri}', "");
Expect(0, 43052, '\p{Script_Extensions=	Syloti_Nagri}', "");
Expect(1, 43052, '\p{^Script_Extensions=	Syloti_Nagri}', "");
Expect(1, 43052, '\P{Script_Extensions=	Syloti_Nagri}', "");
Expect(0, 43052, '\P{^Script_Extensions=	Syloti_Nagri}', "");
Error('\p{Scx=-/a/Sylo}');
Error('\P{Scx=-/a/Sylo}');
Expect(1, 43051, '\p{Scx=sylo}', "");
Expect(0, 43051, '\p{^Scx=sylo}', "");
Expect(0, 43051, '\P{Scx=sylo}', "");
Expect(1, 43051, '\P{^Scx=sylo}', "");
Expect(0, 43052, '\p{Scx=sylo}', "");
Expect(1, 43052, '\p{^Scx=sylo}', "");
Expect(1, 43052, '\P{Scx=sylo}', "");
Expect(0, 43052, '\P{^Scx=sylo}', "");
Expect(1, 43051, '\p{Scx= _Sylo}', "");
Expect(0, 43051, '\p{^Scx= _Sylo}', "");
Expect(0, 43051, '\P{Scx= _Sylo}', "");
Expect(1, 43051, '\P{^Scx= _Sylo}', "");
Expect(0, 43052, '\p{Scx= _Sylo}', "");
Expect(1, 43052, '\p{^Scx= _Sylo}', "");
Expect(1, 43052, '\P{Scx= _Sylo}', "");
Expect(0, 43052, '\P{^Scx= _Sylo}', "");
Error('\p{Is_Script_Extensions= :=syloti_nagri}');
Error('\P{Is_Script_Extensions= :=syloti_nagri}');
Expect(1, 43051, '\p{Is_Script_Extensions=sylotinagri}', "");
Expect(0, 43051, '\p{^Is_Script_Extensions=sylotinagri}', "");
Expect(0, 43051, '\P{Is_Script_Extensions=sylotinagri}', "");
Expect(1, 43051, '\P{^Is_Script_Extensions=sylotinagri}', "");
Expect(0, 43052, '\p{Is_Script_Extensions=sylotinagri}', "");
Expect(1, 43052, '\p{^Is_Script_Extensions=sylotinagri}', "");
Expect(1, 43052, '\P{Is_Script_Extensions=sylotinagri}', "");
Expect(0, 43052, '\P{^Is_Script_Extensions=sylotinagri}', "");
Expect(1, 43051, '\p{Is_Script_Extensions=  SYLOTI_NAGRI}', "");
Expect(0, 43051, '\p{^Is_Script_Extensions=  SYLOTI_NAGRI}', "");
Expect(0, 43051, '\P{Is_Script_Extensions=  SYLOTI_NAGRI}', "");
Expect(1, 43051, '\P{^Is_Script_Extensions=  SYLOTI_NAGRI}', "");
Expect(0, 43052, '\p{Is_Script_Extensions=  SYLOTI_NAGRI}', "");
Expect(1, 43052, '\p{^Is_Script_Extensions=  SYLOTI_NAGRI}', "");
Expect(1, 43052, '\P{Is_Script_Extensions=  SYLOTI_NAGRI}', "");
Expect(0, 43052, '\P{^Is_Script_Extensions=  SYLOTI_NAGRI}', "");
Error('\p{Is_Scx=/a/-	Sylo}');
Error('\P{Is_Scx=/a/-	Sylo}');
Expect(1, 43051, '\p{Is_Scx=sylo}', "");
Expect(0, 43051, '\p{^Is_Scx=sylo}', "");
Expect(0, 43051, '\P{Is_Scx=sylo}', "");
Expect(1, 43051, '\P{^Is_Scx=sylo}', "");
Expect(0, 43052, '\p{Is_Scx=sylo}', "");
Expect(1, 43052, '\p{^Is_Scx=sylo}', "");
Expect(1, 43052, '\P{Is_Scx=sylo}', "");
Expect(0, 43052, '\P{^Is_Scx=sylo}', "");
Expect(1, 43051, '\p{Is_Scx=__Sylo}', "");
Expect(0, 43051, '\p{^Is_Scx=__Sylo}', "");
Expect(0, 43051, '\P{Is_Scx=__Sylo}', "");
Expect(1, 43051, '\P{^Is_Scx=__Sylo}', "");
Expect(0, 43052, '\p{Is_Scx=__Sylo}', "");
Expect(1, 43052, '\p{^Is_Scx=__Sylo}', "");
Expect(1, 43052, '\P{Is_Scx=__Sylo}', "");
Expect(0, 43052, '\P{^Is_Scx=__Sylo}', "");
Error('\p{Script_Extensions= :=Syriac}');
Error('\P{Script_Extensions= :=Syriac}');
Expect(1, 1871, '\p{Script_Extensions=syriac}', "");
Expect(0, 1871, '\p{^Script_Extensions=syriac}', "");
Expect(0, 1871, '\P{Script_Extensions=syriac}', "");
Expect(1, 1871, '\P{^Script_Extensions=syriac}', "");
Expect(0, 1872, '\p{Script_Extensions=syriac}', "");
Expect(1, 1872, '\p{^Script_Extensions=syriac}', "");
Expect(1, 1872, '\P{Script_Extensions=syriac}', "");
Expect(0, 1872, '\P{^Script_Extensions=syriac}', "");
Expect(1, 1871, '\p{Script_Extensions= Syriac}', "");
Expect(0, 1871, '\p{^Script_Extensions= Syriac}', "");
Expect(0, 1871, '\P{Script_Extensions= Syriac}', "");
Expect(1, 1871, '\P{^Script_Extensions= Syriac}', "");
Expect(0, 1872, '\p{Script_Extensions= Syriac}', "");
Expect(1, 1872, '\p{^Script_Extensions= Syriac}', "");
Expect(1, 1872, '\P{Script_Extensions= Syriac}', "");
Expect(0, 1872, '\P{^Script_Extensions= Syriac}', "");
Error('\p{Scx=:=_Syrc}');
Error('\P{Scx=:=_Syrc}');
Expect(1, 1871, '\p{Scx=syrc}', "");
Expect(0, 1871, '\p{^Scx=syrc}', "");
Expect(0, 1871, '\P{Scx=syrc}', "");
Expect(1, 1871, '\P{^Scx=syrc}', "");
Expect(0, 1872, '\p{Scx=syrc}', "");
Expect(1, 1872, '\p{^Scx=syrc}', "");
Expect(1, 1872, '\P{Scx=syrc}', "");
Expect(0, 1872, '\P{^Scx=syrc}', "");
Expect(1, 1871, '\p{Scx=_ Syrc}', "");
Expect(0, 1871, '\p{^Scx=_ Syrc}', "");
Expect(0, 1871, '\P{Scx=_ Syrc}', "");
Expect(1, 1871, '\P{^Scx=_ Syrc}', "");
Expect(0, 1872, '\p{Scx=_ Syrc}', "");
Expect(1, 1872, '\p{^Scx=_ Syrc}', "");
Expect(1, 1872, '\P{Scx=_ Syrc}', "");
Expect(0, 1872, '\P{^Scx=_ Syrc}', "");
Error('\p{Is_Script_Extensions=_/a/syriac}');
Error('\P{Is_Script_Extensions=_/a/syriac}');
Expect(1, 1871, '\p{Is_Script_Extensions=syriac}', "");
Expect(0, 1871, '\p{^Is_Script_Extensions=syriac}', "");
Expect(0, 1871, '\P{Is_Script_Extensions=syriac}', "");
Expect(1, 1871, '\P{^Is_Script_Extensions=syriac}', "");
Expect(0, 1872, '\p{Is_Script_Extensions=syriac}', "");
Expect(1, 1872, '\p{^Is_Script_Extensions=syriac}', "");
Expect(1, 1872, '\P{Is_Script_Extensions=syriac}', "");
Expect(0, 1872, '\P{^Is_Script_Extensions=syriac}', "");
Expect(1, 1871, '\p{Is_Script_Extensions=  Syriac}', "");
Expect(0, 1871, '\p{^Is_Script_Extensions=  Syriac}', "");
Expect(0, 1871, '\P{Is_Script_Extensions=  Syriac}', "");
Expect(1, 1871, '\P{^Is_Script_Extensions=  Syriac}', "");
Expect(0, 1872, '\p{Is_Script_Extensions=  Syriac}', "");
Expect(1, 1872, '\p{^Is_Script_Extensions=  Syriac}', "");
Expect(1, 1872, '\P{Is_Script_Extensions=  Syriac}', "");
Expect(0, 1872, '\P{^Is_Script_Extensions=  Syriac}', "");
Error('\p{Is_Scx:	/a/-SYRC}');
Error('\P{Is_Scx:	/a/-SYRC}');
Expect(1, 1871, '\p{Is_Scx=syrc}', "");
Expect(0, 1871, '\p{^Is_Scx=syrc}', "");
Expect(0, 1871, '\P{Is_Scx=syrc}', "");
Expect(1, 1871, '\P{^Is_Scx=syrc}', "");
Expect(0, 1872, '\p{Is_Scx=syrc}', "");
Expect(1, 1872, '\p{^Is_Scx=syrc}', "");
Expect(1, 1872, '\P{Is_Scx=syrc}', "");
Expect(0, 1872, '\P{^Is_Scx=syrc}', "");
Expect(1, 1871, '\p{Is_Scx= _syrc}', "");
Expect(0, 1871, '\p{^Is_Scx= _syrc}', "");
Expect(0, 1871, '\P{Is_Scx= _syrc}', "");
Expect(1, 1871, '\P{^Is_Scx= _syrc}', "");
Expect(0, 1872, '\p{Is_Scx= _syrc}', "");
Expect(1, 1872, '\p{^Is_Scx= _syrc}', "");
Expect(1, 1872, '\P{Is_Scx= _syrc}', "");
Expect(0, 1872, '\P{^Is_Scx= _syrc}', "");
Error('\p{Script_Extensions:   _tagbanwa:=}');
Error('\P{Script_Extensions:   _tagbanwa:=}');
Expect(1, 6003, '\p{Script_Extensions=tagbanwa}', "");
Expect(0, 6003, '\p{^Script_Extensions=tagbanwa}', "");
Expect(0, 6003, '\P{Script_Extensions=tagbanwa}', "");
Expect(1, 6003, '\P{^Script_Extensions=tagbanwa}', "");
Expect(0, 6004, '\p{Script_Extensions=tagbanwa}', "");
Expect(1, 6004, '\p{^Script_Extensions=tagbanwa}', "");
Expect(1, 6004, '\P{Script_Extensions=tagbanwa}', "");
Expect(0, 6004, '\P{^Script_Extensions=tagbanwa}', "");
Expect(1, 6003, '\p{Script_Extensions=_TAGBANWA}', "");
Expect(0, 6003, '\p{^Script_Extensions=_TAGBANWA}', "");
Expect(0, 6003, '\P{Script_Extensions=_TAGBANWA}', "");
Expect(1, 6003, '\P{^Script_Extensions=_TAGBANWA}', "");
Expect(0, 6004, '\p{Script_Extensions=_TAGBANWA}', "");
Expect(1, 6004, '\p{^Script_Extensions=_TAGBANWA}', "");
Expect(1, 6004, '\P{Script_Extensions=_TAGBANWA}', "");
Expect(0, 6004, '\P{^Script_Extensions=_TAGBANWA}', "");
Error('\p{Scx=:=-Tagb}');
Error('\P{Scx=:=-Tagb}');
Expect(1, 6003, '\p{Scx=tagb}', "");
Expect(0, 6003, '\p{^Scx=tagb}', "");
Expect(0, 6003, '\P{Scx=tagb}', "");
Expect(1, 6003, '\P{^Scx=tagb}', "");
Expect(0, 6004, '\p{Scx=tagb}', "");
Expect(1, 6004, '\p{^Scx=tagb}', "");
Expect(1, 6004, '\P{Scx=tagb}', "");
Expect(0, 6004, '\P{^Scx=tagb}', "");
Expect(1, 6003, '\p{Scx=_-Tagb}', "");
Expect(0, 6003, '\p{^Scx=_-Tagb}', "");
Expect(0, 6003, '\P{Scx=_-Tagb}', "");
Expect(1, 6003, '\P{^Scx=_-Tagb}', "");
Expect(0, 6004, '\p{Scx=_-Tagb}', "");
Expect(1, 6004, '\p{^Scx=_-Tagb}', "");
Expect(1, 6004, '\P{Scx=_-Tagb}', "");
Expect(0, 6004, '\P{^Scx=_-Tagb}', "");
Error('\p{Is_Script_Extensions:   	/a/tagbanwa}');
Error('\P{Is_Script_Extensions:   	/a/tagbanwa}');
Expect(1, 6003, '\p{Is_Script_Extensions: tagbanwa}', "");
Expect(0, 6003, '\p{^Is_Script_Extensions: tagbanwa}', "");
Expect(0, 6003, '\P{Is_Script_Extensions: tagbanwa}', "");
Expect(1, 6003, '\P{^Is_Script_Extensions: tagbanwa}', "");
Expect(0, 6004, '\p{Is_Script_Extensions: tagbanwa}', "");
Expect(1, 6004, '\p{^Is_Script_Extensions: tagbanwa}', "");
Expect(1, 6004, '\P{Is_Script_Extensions: tagbanwa}', "");
Expect(0, 6004, '\P{^Is_Script_Extensions: tagbanwa}', "");
Expect(1, 6003, '\p{Is_Script_Extensions=	 Tagbanwa}', "");
Expect(0, 6003, '\p{^Is_Script_Extensions=	 Tagbanwa}', "");
Expect(0, 6003, '\P{Is_Script_Extensions=	 Tagbanwa}', "");
Expect(1, 6003, '\P{^Is_Script_Extensions=	 Tagbanwa}', "");
Expect(0, 6004, '\p{Is_Script_Extensions=	 Tagbanwa}', "");
Expect(1, 6004, '\p{^Is_Script_Extensions=	 Tagbanwa}', "");
Expect(1, 6004, '\P{Is_Script_Extensions=	 Tagbanwa}', "");
Expect(0, 6004, '\P{^Is_Script_Extensions=	 Tagbanwa}', "");
Error('\p{Is_Scx=_-tagb:=}');
Error('\P{Is_Scx=_-tagb:=}');
Expect(1, 6003, '\p{Is_Scx=tagb}', "");
Expect(0, 6003, '\p{^Is_Scx=tagb}', "");
Expect(0, 6003, '\P{Is_Scx=tagb}', "");
Expect(1, 6003, '\P{^Is_Scx=tagb}', "");
Expect(0, 6004, '\p{Is_Scx=tagb}', "");
Expect(1, 6004, '\p{^Is_Scx=tagb}', "");
Expect(1, 6004, '\P{Is_Scx=tagb}', "");
Expect(0, 6004, '\P{^Is_Scx=tagb}', "");
Expect(1, 6003, '\p{Is_Scx=	-tagb}', "");
Expect(0, 6003, '\p{^Is_Scx=	-tagb}', "");
Expect(0, 6003, '\P{Is_Scx=	-tagb}', "");
Expect(1, 6003, '\P{^Is_Scx=	-tagb}', "");
Expect(0, 6004, '\p{Is_Scx=	-tagb}', "");
Expect(1, 6004, '\p{^Is_Scx=	-tagb}', "");
Expect(1, 6004, '\P{Is_Scx=	-tagb}', "");
Expect(0, 6004, '\P{^Is_Scx=	-tagb}', "");
Error('\p{Script_Extensions=-Takri/a/}');
Error('\P{Script_Extensions=-Takri/a/}');
Expect(1, 71369, '\p{Script_Extensions=takri}', "");
Expect(0, 71369, '\p{^Script_Extensions=takri}', "");
Expect(0, 71369, '\P{Script_Extensions=takri}', "");
Expect(1, 71369, '\P{^Script_Extensions=takri}', "");
Expect(0, 71370, '\p{Script_Extensions=takri}', "");
Expect(1, 71370, '\p{^Script_Extensions=takri}', "");
Expect(1, 71370, '\P{Script_Extensions=takri}', "");
Expect(0, 71370, '\P{^Script_Extensions=takri}', "");
Expect(1, 71369, '\p{Script_Extensions:   -TAKRI}', "");
Expect(0, 71369, '\p{^Script_Extensions:   -TAKRI}', "");
Expect(0, 71369, '\P{Script_Extensions:   -TAKRI}', "");
Expect(1, 71369, '\P{^Script_Extensions:   -TAKRI}', "");
Expect(0, 71370, '\p{Script_Extensions:   -TAKRI}', "");
Expect(1, 71370, '\p{^Script_Extensions:   -TAKRI}', "");
Expect(1, 71370, '\P{Script_Extensions:   -TAKRI}', "");
Expect(0, 71370, '\P{^Script_Extensions:   -TAKRI}', "");
Error('\p{Scx=_/a/TAKR}');
Error('\P{Scx=_/a/TAKR}');
Expect(1, 71369, '\p{Scx=takr}', "");
Expect(0, 71369, '\p{^Scx=takr}', "");
Expect(0, 71369, '\P{Scx=takr}', "");
Expect(1, 71369, '\P{^Scx=takr}', "");
Expect(0, 71370, '\p{Scx=takr}', "");
Expect(1, 71370, '\p{^Scx=takr}', "");
Expect(1, 71370, '\P{Scx=takr}', "");
Expect(0, 71370, '\P{^Scx=takr}', "");
Expect(1, 71369, '\p{Scx=-_TAKR}', "");
Expect(0, 71369, '\p{^Scx=-_TAKR}', "");
Expect(0, 71369, '\P{Scx=-_TAKR}', "");
Expect(1, 71369, '\P{^Scx=-_TAKR}', "");
Expect(0, 71370, '\p{Scx=-_TAKR}', "");
Expect(1, 71370, '\p{^Scx=-_TAKR}', "");
Expect(1, 71370, '\P{Scx=-_TAKR}', "");
Expect(0, 71370, '\P{^Scx=-_TAKR}', "");
Error('\p{Is_Script_Extensions=-:=Takri}');
Error('\P{Is_Script_Extensions=-:=Takri}');
Expect(1, 71369, '\p{Is_Script_Extensions=takri}', "");
Expect(0, 71369, '\p{^Is_Script_Extensions=takri}', "");
Expect(0, 71369, '\P{Is_Script_Extensions=takri}', "");
Expect(1, 71369, '\P{^Is_Script_Extensions=takri}', "");
Expect(0, 71370, '\p{Is_Script_Extensions=takri}', "");
Expect(1, 71370, '\p{^Is_Script_Extensions=takri}', "");
Expect(1, 71370, '\P{Is_Script_Extensions=takri}', "");
Expect(0, 71370, '\P{^Is_Script_Extensions=takri}', "");
Expect(1, 71369, '\p{Is_Script_Extensions=-Takri}', "");
Expect(0, 71369, '\p{^Is_Script_Extensions=-Takri}', "");
Expect(0, 71369, '\P{Is_Script_Extensions=-Takri}', "");
Expect(1, 71369, '\P{^Is_Script_Extensions=-Takri}', "");
Expect(0, 71370, '\p{Is_Script_Extensions=-Takri}', "");
Expect(1, 71370, '\p{^Is_Script_Extensions=-Takri}', "");
Expect(1, 71370, '\P{Is_Script_Extensions=-Takri}', "");
Expect(0, 71370, '\P{^Is_Script_Extensions=-Takri}', "");
Error('\p{Is_Scx=Takr:=}');
Error('\P{Is_Scx=Takr:=}');
Expect(1, 71369, '\p{Is_Scx=takr}', "");
Expect(0, 71369, '\p{^Is_Scx=takr}', "");
Expect(0, 71369, '\P{Is_Scx=takr}', "");
Expect(1, 71369, '\P{^Is_Scx=takr}', "");
Expect(0, 71370, '\p{Is_Scx=takr}', "");
Expect(1, 71370, '\p{^Is_Scx=takr}', "");
Expect(1, 71370, '\P{Is_Scx=takr}', "");
Expect(0, 71370, '\P{^Is_Scx=takr}', "");
Expect(1, 71369, '\p{Is_Scx=	TAKR}', "");
Expect(0, 71369, '\p{^Is_Scx=	TAKR}', "");
Expect(0, 71369, '\P{Is_Scx=	TAKR}', "");
Expect(1, 71369, '\P{^Is_Scx=	TAKR}', "");
Expect(0, 71370, '\p{Is_Scx=	TAKR}', "");
Expect(1, 71370, '\p{^Is_Scx=	TAKR}', "");
Expect(1, 71370, '\P{Is_Scx=	TAKR}', "");
Expect(0, 71370, '\P{^Is_Scx=	TAKR}', "");
Error('\p{Script_Extensions= /a/Tai_Le}');
Error('\P{Script_Extensions= /a/Tai_Le}');
Expect(1, 6516, '\p{Script_Extensions=taile}', "");
Expect(0, 6516, '\p{^Script_Extensions=taile}', "");
Expect(0, 6516, '\P{Script_Extensions=taile}', "");
Expect(1, 6516, '\P{^Script_Extensions=taile}', "");
Expect(0, 6517, '\p{Script_Extensions=taile}', "");
Expect(1, 6517, '\p{^Script_Extensions=taile}', "");
Expect(1, 6517, '\P{Script_Extensions=taile}', "");
Expect(0, 6517, '\P{^Script_Extensions=taile}', "");
Expect(1, 6516, '\p{Script_Extensions=-	Tai_Le}', "");
Expect(0, 6516, '\p{^Script_Extensions=-	Tai_Le}', "");
Expect(0, 6516, '\P{Script_Extensions=-	Tai_Le}', "");
Expect(1, 6516, '\P{^Script_Extensions=-	Tai_Le}', "");
Expect(0, 6517, '\p{Script_Extensions=-	Tai_Le}', "");
Expect(1, 6517, '\p{^Script_Extensions=-	Tai_Le}', "");
Expect(1, 6517, '\P{Script_Extensions=-	Tai_Le}', "");
Expect(0, 6517, '\P{^Script_Extensions=-	Tai_Le}', "");
Error('\p{Scx= :=TALE}');
Error('\P{Scx= :=TALE}');
Expect(1, 6516, '\p{Scx=tale}', "");
Expect(0, 6516, '\p{^Scx=tale}', "");
Expect(0, 6516, '\P{Scx=tale}', "");
Expect(1, 6516, '\P{^Scx=tale}', "");
Expect(0, 6517, '\p{Scx=tale}', "");
Expect(1, 6517, '\p{^Scx=tale}', "");
Expect(1, 6517, '\P{Scx=tale}', "");
Expect(0, 6517, '\P{^Scx=tale}', "");
Expect(1, 6516, '\p{Scx= -TALE}', "");
Expect(0, 6516, '\p{^Scx= -TALE}', "");
Expect(0, 6516, '\P{Scx= -TALE}', "");
Expect(1, 6516, '\P{^Scx= -TALE}', "");
Expect(0, 6517, '\p{Scx= -TALE}', "");
Expect(1, 6517, '\p{^Scx= -TALE}', "");
Expect(1, 6517, '\P{Scx= -TALE}', "");
Expect(0, 6517, '\P{^Scx= -TALE}', "");
Error('\p{Is_Script_Extensions=:=		tai_LE}');
Error('\P{Is_Script_Extensions=:=		tai_LE}');
Expect(1, 6516, '\p{Is_Script_Extensions=taile}', "");
Expect(0, 6516, '\p{^Is_Script_Extensions=taile}', "");
Expect(0, 6516, '\P{Is_Script_Extensions=taile}', "");
Expect(1, 6516, '\P{^Is_Script_Extensions=taile}', "");
Expect(0, 6517, '\p{Is_Script_Extensions=taile}', "");
Expect(1, 6517, '\p{^Is_Script_Extensions=taile}', "");
Expect(1, 6517, '\P{Is_Script_Extensions=taile}', "");
Expect(0, 6517, '\P{^Is_Script_Extensions=taile}', "");
Expect(1, 6516, '\p{Is_Script_Extensions=_tai_Le}', "");
Expect(0, 6516, '\p{^Is_Script_Extensions=_tai_Le}', "");
Expect(0, 6516, '\P{Is_Script_Extensions=_tai_Le}', "");
Expect(1, 6516, '\P{^Is_Script_Extensions=_tai_Le}', "");
Expect(0, 6517, '\p{Is_Script_Extensions=_tai_Le}', "");
Expect(1, 6517, '\p{^Is_Script_Extensions=_tai_Le}', "");
Expect(1, 6517, '\P{Is_Script_Extensions=_tai_Le}', "");
Expect(0, 6517, '\P{^Is_Script_Extensions=_tai_Le}', "");
Error('\p{Is_Scx=_	Tale/a/}');
Error('\P{Is_Scx=_	Tale/a/}');
Expect(1, 6516, '\p{Is_Scx=tale}', "");
Expect(0, 6516, '\p{^Is_Scx=tale}', "");
Expect(0, 6516, '\P{Is_Scx=tale}', "");
Expect(1, 6516, '\P{^Is_Scx=tale}', "");
Expect(0, 6517, '\p{Is_Scx=tale}', "");
Expect(1, 6517, '\p{^Is_Scx=tale}', "");
Expect(1, 6517, '\P{Is_Scx=tale}', "");
Expect(0, 6517, '\P{^Is_Scx=tale}', "");
Expect(1, 6516, '\p{Is_Scx:_	Tale}', "");
Expect(0, 6516, '\p{^Is_Scx:_	Tale}', "");
Expect(0, 6516, '\P{Is_Scx:_	Tale}', "");
Expect(1, 6516, '\P{^Is_Scx:_	Tale}', "");
Expect(0, 6517, '\p{Is_Scx:_	Tale}', "");
Expect(1, 6517, '\p{^Is_Scx:_	Tale}', "");
Expect(1, 6517, '\P{Is_Scx:_	Tale}', "");
Expect(0, 6517, '\P{^Is_Scx:_	Tale}', "");
Error('\p{Script_Extensions=_NEW_Tai_LUE/a/}');
Error('\P{Script_Extensions=_NEW_Tai_LUE/a/}');
Expect(1, 6623, '\p{Script_Extensions=newtailue}', "");
Expect(0, 6623, '\p{^Script_Extensions=newtailue}', "");
Expect(0, 6623, '\P{Script_Extensions=newtailue}', "");
Expect(1, 6623, '\P{^Script_Extensions=newtailue}', "");
Expect(0, 6624, '\p{Script_Extensions=newtailue}', "");
Expect(1, 6624, '\p{^Script_Extensions=newtailue}', "");
Expect(1, 6624, '\P{Script_Extensions=newtailue}', "");
Expect(0, 6624, '\P{^Script_Extensions=newtailue}', "");
Expect(1, 6623, '\p{Script_Extensions=		New_Tai_Lue}', "");
Expect(0, 6623, '\p{^Script_Extensions=		New_Tai_Lue}', "");
Expect(0, 6623, '\P{Script_Extensions=		New_Tai_Lue}', "");
Expect(1, 6623, '\P{^Script_Extensions=		New_Tai_Lue}', "");
Expect(0, 6624, '\p{Script_Extensions=		New_Tai_Lue}', "");
Expect(1, 6624, '\p{^Script_Extensions=		New_Tai_Lue}', "");
Expect(1, 6624, '\P{Script_Extensions=		New_Tai_Lue}', "");
Expect(0, 6624, '\P{^Script_Extensions=		New_Tai_Lue}', "");
Error('\p{Scx=- Talu/a/}');
Error('\P{Scx=- Talu/a/}');
Expect(1, 6623, '\p{Scx=talu}', "");
Expect(0, 6623, '\p{^Scx=talu}', "");
Expect(0, 6623, '\P{Scx=talu}', "");
Expect(1, 6623, '\P{^Scx=talu}', "");
Expect(0, 6624, '\p{Scx=talu}', "");
Expect(1, 6624, '\p{^Scx=talu}', "");
Expect(1, 6624, '\P{Scx=talu}', "");
Expect(0, 6624, '\P{^Scx=talu}', "");
Expect(1, 6623, '\p{Scx=-Talu}', "");
Expect(0, 6623, '\p{^Scx=-Talu}', "");
Expect(0, 6623, '\P{Scx=-Talu}', "");
Expect(1, 6623, '\P{^Scx=-Talu}', "");
Expect(0, 6624, '\p{Scx=-Talu}', "");
Expect(1, 6624, '\p{^Scx=-Talu}', "");
Expect(1, 6624, '\P{Scx=-Talu}', "");
Expect(0, 6624, '\P{^Scx=-Talu}', "");
Error('\p{Is_Script_Extensions=	new_TAI_Lue:=}');
Error('\P{Is_Script_Extensions=	new_TAI_Lue:=}');
Expect(1, 6623, '\p{Is_Script_Extensions:   newtailue}', "");
Expect(0, 6623, '\p{^Is_Script_Extensions:   newtailue}', "");
Expect(0, 6623, '\P{Is_Script_Extensions:   newtailue}', "");
Expect(1, 6623, '\P{^Is_Script_Extensions:   newtailue}', "");
Expect(0, 6624, '\p{Is_Script_Extensions:   newtailue}', "");
Expect(1, 6624, '\p{^Is_Script_Extensions:   newtailue}', "");
Expect(1, 6624, '\P{Is_Script_Extensions:   newtailue}', "");
Expect(0, 6624, '\P{^Is_Script_Extensions:   newtailue}', "");
Expect(1, 6623, '\p{Is_Script_Extensions=NEW_Tai_LUE}', "");
Expect(0, 6623, '\p{^Is_Script_Extensions=NEW_Tai_LUE}', "");
Expect(0, 6623, '\P{Is_Script_Extensions=NEW_Tai_LUE}', "");
Expect(1, 6623, '\P{^Is_Script_Extensions=NEW_Tai_LUE}', "");
Expect(0, 6624, '\p{Is_Script_Extensions=NEW_Tai_LUE}', "");
Expect(1, 6624, '\p{^Is_Script_Extensions=NEW_Tai_LUE}', "");
Expect(1, 6624, '\P{Is_Script_Extensions=NEW_Tai_LUE}', "");
Expect(0, 6624, '\P{^Is_Script_Extensions=NEW_Tai_LUE}', "");
Error('\p{Is_Scx=_:=Talu}');
Error('\P{Is_Scx=_:=Talu}');
Expect(1, 6623, '\p{Is_Scx=talu}', "");
Expect(0, 6623, '\p{^Is_Scx=talu}', "");
Expect(0, 6623, '\P{Is_Scx=talu}', "");
Expect(1, 6623, '\P{^Is_Scx=talu}', "");
Expect(0, 6624, '\p{Is_Scx=talu}', "");
Expect(1, 6624, '\p{^Is_Scx=talu}', "");
Expect(1, 6624, '\P{Is_Scx=talu}', "");
Expect(0, 6624, '\P{^Is_Scx=talu}', "");
Expect(1, 6623, '\p{Is_Scx=__Talu}', "");
Expect(0, 6623, '\p{^Is_Scx=__Talu}', "");
Expect(0, 6623, '\P{Is_Scx=__Talu}', "");
Expect(1, 6623, '\P{^Is_Scx=__Talu}', "");
Expect(0, 6624, '\p{Is_Scx=__Talu}', "");
Expect(1, 6624, '\p{^Is_Scx=__Talu}', "");
Expect(1, 6624, '\P{Is_Scx=__Talu}', "");
Expect(0, 6624, '\P{^Is_Scx=__Talu}', "");
Error('\p{Script_Extensions=	tamil:=}');
Error('\P{Script_Extensions=	tamil:=}');
Expect(1, 3066, '\p{Script_Extensions=tamil}', "");
Expect(0, 3066, '\p{^Script_Extensions=tamil}', "");
Expect(0, 3066, '\P{Script_Extensions=tamil}', "");
Expect(1, 3066, '\P{^Script_Extensions=tamil}', "");
Expect(0, 3067, '\p{Script_Extensions=tamil}', "");
Expect(1, 3067, '\p{^Script_Extensions=tamil}', "");
Expect(1, 3067, '\P{Script_Extensions=tamil}', "");
Expect(0, 3067, '\P{^Script_Extensions=tamil}', "");
Expect(1, 3066, '\p{Script_Extensions=  tamil}', "");
Expect(0, 3066, '\p{^Script_Extensions=  tamil}', "");
Expect(0, 3066, '\P{Script_Extensions=  tamil}', "");
Expect(1, 3066, '\P{^Script_Extensions=  tamil}', "");
Expect(0, 3067, '\p{Script_Extensions=  tamil}', "");
Expect(1, 3067, '\p{^Script_Extensions=  tamil}', "");
Expect(1, 3067, '\P{Script_Extensions=  tamil}', "");
Expect(0, 3067, '\P{^Script_Extensions=  tamil}', "");
Error('\p{Scx= :=TAML}');
Error('\P{Scx= :=TAML}');
Expect(1, 3066, '\p{Scx=taml}', "");
Expect(0, 3066, '\p{^Scx=taml}', "");
Expect(0, 3066, '\P{Scx=taml}', "");
Expect(1, 3066, '\P{^Scx=taml}', "");
Expect(0, 3067, '\p{Scx=taml}', "");
Expect(1, 3067, '\p{^Scx=taml}', "");
Expect(1, 3067, '\P{Scx=taml}', "");
Expect(0, 3067, '\P{^Scx=taml}', "");
Expect(1, 3066, '\p{Scx=-TAML}', "");
Expect(0, 3066, '\p{^Scx=-TAML}', "");
Expect(0, 3066, '\P{Scx=-TAML}', "");
Expect(1, 3066, '\P{^Scx=-TAML}', "");
Expect(0, 3067, '\p{Scx=-TAML}', "");
Expect(1, 3067, '\p{^Scx=-TAML}', "");
Expect(1, 3067, '\P{Scx=-TAML}', "");
Expect(0, 3067, '\P{^Scx=-TAML}', "");
Error('\p{Is_Script_Extensions= :=tamil}');
Error('\P{Is_Script_Extensions= :=tamil}');
Expect(1, 3066, '\p{Is_Script_Extensions=tamil}', "");
Expect(0, 3066, '\p{^Is_Script_Extensions=tamil}', "");
Expect(0, 3066, '\P{Is_Script_Extensions=tamil}', "");
Expect(1, 3066, '\P{^Is_Script_Extensions=tamil}', "");
Expect(0, 3067, '\p{Is_Script_Extensions=tamil}', "");
Expect(1, 3067, '\p{^Is_Script_Extensions=tamil}', "");
Expect(1, 3067, '\P{Is_Script_Extensions=tamil}', "");
Expect(0, 3067, '\P{^Is_Script_Extensions=tamil}', "");
Expect(1, 3066, '\p{Is_Script_Extensions= Tamil}', "");
Expect(0, 3066, '\p{^Is_Script_Extensions= Tamil}', "");
Expect(0, 3066, '\P{Is_Script_Extensions= Tamil}', "");
Expect(1, 3066, '\P{^Is_Script_Extensions= Tamil}', "");
Expect(0, 3067, '\p{Is_Script_Extensions= Tamil}', "");
Expect(1, 3067, '\p{^Is_Script_Extensions= Tamil}', "");
Expect(1, 3067, '\P{Is_Script_Extensions= Tamil}', "");
Expect(0, 3067, '\P{^Is_Script_Extensions= Tamil}', "");
Error('\p{Is_Scx=	taml:=}');
Error('\P{Is_Scx=	taml:=}');
Expect(1, 3066, '\p{Is_Scx=taml}', "");
Expect(0, 3066, '\p{^Is_Scx=taml}', "");
Expect(0, 3066, '\P{Is_Scx=taml}', "");
Expect(1, 3066, '\P{^Is_Scx=taml}', "");
Expect(0, 3067, '\p{Is_Scx=taml}', "");
Expect(1, 3067, '\p{^Is_Scx=taml}', "");
Expect(1, 3067, '\P{Is_Scx=taml}', "");
Expect(0, 3067, '\P{^Is_Scx=taml}', "");
Expect(1, 3066, '\p{Is_Scx= Taml}', "");
Expect(0, 3066, '\p{^Is_Scx= Taml}', "");
Expect(0, 3066, '\P{Is_Scx= Taml}', "");
Expect(1, 3066, '\P{^Is_Scx= Taml}', "");
Expect(0, 3067, '\p{Is_Scx= Taml}', "");
Expect(1, 3067, '\p{^Is_Scx= Taml}', "");
Expect(1, 3067, '\P{Is_Scx= Taml}', "");
Expect(0, 3067, '\P{^Is_Scx= Taml}', "");
Error('\p{Script_Extensions=	-Tai_Viet/a/}');
Error('\P{Script_Extensions=	-Tai_Viet/a/}');
Expect(1, 43743, '\p{Script_Extensions=taiviet}', "");
Expect(0, 43743, '\p{^Script_Extensions=taiviet}', "");
Expect(0, 43743, '\P{Script_Extensions=taiviet}', "");
Expect(1, 43743, '\P{^Script_Extensions=taiviet}', "");
Expect(0, 43744, '\p{Script_Extensions=taiviet}', "");
Expect(1, 43744, '\p{^Script_Extensions=taiviet}', "");
Expect(1, 43744, '\P{Script_Extensions=taiviet}', "");
Expect(0, 43744, '\P{^Script_Extensions=taiviet}', "");
Expect(1, 43743, '\p{Script_Extensions=		Tai_viet}', "");
Expect(0, 43743, '\p{^Script_Extensions=		Tai_viet}', "");
Expect(0, 43743, '\P{Script_Extensions=		Tai_viet}', "");
Expect(1, 43743, '\P{^Script_Extensions=		Tai_viet}', "");
Expect(0, 43744, '\p{Script_Extensions=		Tai_viet}', "");
Expect(1, 43744, '\p{^Script_Extensions=		Tai_viet}', "");
Expect(1, 43744, '\P{Script_Extensions=		Tai_viet}', "");
Expect(0, 43744, '\P{^Script_Extensions=		Tai_viet}', "");
Error('\p{Scx:		:=TAVT}');
Error('\P{Scx:		:=TAVT}');
Expect(1, 43743, '\p{Scx=tavt}', "");
Expect(0, 43743, '\p{^Scx=tavt}', "");
Expect(0, 43743, '\P{Scx=tavt}', "");
Expect(1, 43743, '\P{^Scx=tavt}', "");
Expect(0, 43744, '\p{Scx=tavt}', "");
Expect(1, 43744, '\p{^Scx=tavt}', "");
Expect(1, 43744, '\P{Scx=tavt}', "");
Expect(0, 43744, '\P{^Scx=tavt}', "");
Expect(1, 43743, '\p{Scx=-tavt}', "");
Expect(0, 43743, '\p{^Scx=-tavt}', "");
Expect(0, 43743, '\P{Scx=-tavt}', "");
Expect(1, 43743, '\P{^Scx=-tavt}', "");
Expect(0, 43744, '\p{Scx=-tavt}', "");
Expect(1, 43744, '\p{^Scx=-tavt}', "");
Expect(1, 43744, '\P{Scx=-tavt}', "");
Expect(0, 43744, '\P{^Scx=-tavt}', "");
Error('\p{Is_Script_Extensions=/a/_-Tai_viet}');
Error('\P{Is_Script_Extensions=/a/_-Tai_viet}');
Expect(1, 43743, '\p{Is_Script_Extensions=taiviet}', "");
Expect(0, 43743, '\p{^Is_Script_Extensions=taiviet}', "");
Expect(0, 43743, '\P{Is_Script_Extensions=taiviet}', "");
Expect(1, 43743, '\P{^Is_Script_Extensions=taiviet}', "");
Expect(0, 43744, '\p{Is_Script_Extensions=taiviet}', "");
Expect(1, 43744, '\p{^Is_Script_Extensions=taiviet}', "");
Expect(1, 43744, '\P{Is_Script_Extensions=taiviet}', "");
Expect(0, 43744, '\P{^Is_Script_Extensions=taiviet}', "");
Expect(1, 43743, '\p{Is_Script_Extensions= tai_Viet}', "");
Expect(0, 43743, '\p{^Is_Script_Extensions= tai_Viet}', "");
Expect(0, 43743, '\P{Is_Script_Extensions= tai_Viet}', "");
Expect(1, 43743, '\P{^Is_Script_Extensions= tai_Viet}', "");
Expect(0, 43744, '\p{Is_Script_Extensions= tai_Viet}', "");
Expect(1, 43744, '\p{^Is_Script_Extensions= tai_Viet}', "");
Expect(1, 43744, '\P{Is_Script_Extensions= tai_Viet}', "");
Expect(0, 43744, '\P{^Is_Script_Extensions= tai_Viet}', "");
Error('\p{Is_Scx=--Tavt/a/}');
Error('\P{Is_Scx=--Tavt/a/}');
Expect(1, 43743, '\p{Is_Scx=tavt}', "");
Expect(0, 43743, '\p{^Is_Scx=tavt}', "");
Expect(0, 43743, '\P{Is_Scx=tavt}', "");
Expect(1, 43743, '\P{^Is_Scx=tavt}', "");
Expect(0, 43744, '\p{Is_Scx=tavt}', "");
Expect(1, 43744, '\p{^Is_Scx=tavt}', "");
Expect(1, 43744, '\P{Is_Scx=tavt}', "");
Expect(0, 43744, '\P{^Is_Scx=tavt}', "");
Expect(1, 43743, '\p{Is_Scx= -Tavt}', "");
Expect(0, 43743, '\p{^Is_Scx= -Tavt}', "");
Expect(0, 43743, '\P{Is_Scx= -Tavt}', "");
Expect(1, 43743, '\P{^Is_Scx= -Tavt}', "");
Expect(0, 43744, '\p{Is_Scx= -Tavt}', "");
Expect(1, 43744, '\p{^Is_Scx= -Tavt}', "");
Expect(1, 43744, '\P{Is_Scx= -Tavt}', "");
Expect(0, 43744, '\P{^Is_Scx= -Tavt}', "");
Error('\p{Script_Extensions:-_telugu:=}');
Error('\P{Script_Extensions:-_telugu:=}');
Expect(1, 3199, '\p{Script_Extensions=telugu}', "");
Expect(0, 3199, '\p{^Script_Extensions=telugu}', "");
Expect(0, 3199, '\P{Script_Extensions=telugu}', "");
Expect(1, 3199, '\P{^Script_Extensions=telugu}', "");
Expect(0, 3200, '\p{Script_Extensions=telugu}', "");
Expect(1, 3200, '\p{^Script_Extensions=telugu}', "");
Expect(1, 3200, '\P{Script_Extensions=telugu}', "");
Expect(0, 3200, '\P{^Script_Extensions=telugu}', "");
Expect(1, 3199, '\p{Script_Extensions=	 telugu}', "");
Expect(0, 3199, '\p{^Script_Extensions=	 telugu}', "");
Expect(0, 3199, '\P{Script_Extensions=	 telugu}', "");
Expect(1, 3199, '\P{^Script_Extensions=	 telugu}', "");
Expect(0, 3200, '\p{Script_Extensions=	 telugu}', "");
Expect(1, 3200, '\p{^Script_Extensions=	 telugu}', "");
Expect(1, 3200, '\P{Script_Extensions=	 telugu}', "");
Expect(0, 3200, '\P{^Script_Extensions=	 telugu}', "");
Error('\p{Scx=/a/	TELU}');
Error('\P{Scx=/a/	TELU}');
Expect(1, 3199, '\p{Scx=telu}', "");
Expect(0, 3199, '\p{^Scx=telu}', "");
Expect(0, 3199, '\P{Scx=telu}', "");
Expect(1, 3199, '\P{^Scx=telu}', "");
Expect(0, 3200, '\p{Scx=telu}', "");
Expect(1, 3200, '\p{^Scx=telu}', "");
Expect(1, 3200, '\P{Scx=telu}', "");
Expect(0, 3200, '\P{^Scx=telu}', "");
Expect(1, 3199, '\p{Scx= 	TELU}', "");
Expect(0, 3199, '\p{^Scx= 	TELU}', "");
Expect(0, 3199, '\P{Scx= 	TELU}', "");
Expect(1, 3199, '\P{^Scx= 	TELU}', "");
Expect(0, 3200, '\p{Scx= 	TELU}', "");
Expect(1, 3200, '\p{^Scx= 	TELU}', "");
Expect(1, 3200, '\P{Scx= 	TELU}', "");
Expect(0, 3200, '\P{^Scx= 	TELU}', "");
Error('\p{Is_Script_Extensions= :=Telugu}');
Error('\P{Is_Script_Extensions= :=Telugu}');
Expect(1, 3199, '\p{Is_Script_Extensions=telugu}', "");
Expect(0, 3199, '\p{^Is_Script_Extensions=telugu}', "");
Expect(0, 3199, '\P{Is_Script_Extensions=telugu}', "");
Expect(1, 3199, '\P{^Is_Script_Extensions=telugu}', "");
Expect(0, 3200, '\p{Is_Script_Extensions=telugu}', "");
Expect(1, 3200, '\p{^Is_Script_Extensions=telugu}', "");
Expect(1, 3200, '\P{Is_Script_Extensions=telugu}', "");
Expect(0, 3200, '\P{^Is_Script_Extensions=telugu}', "");
Expect(1, 3199, '\p{Is_Script_Extensions=-	Telugu}', "");
Expect(0, 3199, '\p{^Is_Script_Extensions=-	Telugu}', "");
Expect(0, 3199, '\P{Is_Script_Extensions=-	Telugu}', "");
Expect(1, 3199, '\P{^Is_Script_Extensions=-	Telugu}', "");
Expect(0, 3200, '\p{Is_Script_Extensions=-	Telugu}', "");
Expect(1, 3200, '\p{^Is_Script_Extensions=-	Telugu}', "");
Expect(1, 3200, '\P{Is_Script_Extensions=-	Telugu}', "");
Expect(0, 3200, '\P{^Is_Script_Extensions=-	Telugu}', "");
Error('\p{Is_Scx=_/a/Telu}');
Error('\P{Is_Scx=_/a/Telu}');
Expect(1, 3199, '\p{Is_Scx=telu}', "");
Expect(0, 3199, '\p{^Is_Scx=telu}', "");
Expect(0, 3199, '\P{Is_Scx=telu}', "");
Expect(1, 3199, '\P{^Is_Scx=telu}', "");
Expect(0, 3200, '\p{Is_Scx=telu}', "");
Expect(1, 3200, '\p{^Is_Scx=telu}', "");
Expect(1, 3200, '\P{Is_Scx=telu}', "");
Expect(0, 3200, '\P{^Is_Scx=telu}', "");
Expect(1, 3199, '\p{Is_Scx=	-Telu}', "");
Expect(0, 3199, '\p{^Is_Scx=	-Telu}', "");
Expect(0, 3199, '\P{Is_Scx=	-Telu}', "");
Expect(1, 3199, '\P{^Is_Scx=	-Telu}', "");
Expect(0, 3200, '\p{Is_Scx=	-Telu}', "");
Expect(1, 3200, '\p{^Is_Scx=	-Telu}', "");
Expect(1, 3200, '\P{Is_Scx=	-Telu}', "");
Expect(0, 3200, '\P{^Is_Scx=	-Telu}', "");
Error('\p{Script_Extensions=-:=TIFINAGH}');
Error('\P{Script_Extensions=-:=TIFINAGH}');
Expect(1, 11647, '\p{Script_Extensions=tifinagh}', "");
Expect(0, 11647, '\p{^Script_Extensions=tifinagh}', "");
Expect(0, 11647, '\P{Script_Extensions=tifinagh}', "");
Expect(1, 11647, '\P{^Script_Extensions=tifinagh}', "");
Expect(0, 11648, '\p{Script_Extensions=tifinagh}', "");
Expect(1, 11648, '\p{^Script_Extensions=tifinagh}', "");
Expect(1, 11648, '\P{Script_Extensions=tifinagh}', "");
Expect(0, 11648, '\P{^Script_Extensions=tifinagh}', "");
Expect(1, 11647, '\p{Script_Extensions=  Tifinagh}', "");
Expect(0, 11647, '\p{^Script_Extensions=  Tifinagh}', "");
Expect(0, 11647, '\P{Script_Extensions=  Tifinagh}', "");
Expect(1, 11647, '\P{^Script_Extensions=  Tifinagh}', "");
Expect(0, 11648, '\p{Script_Extensions=  Tifinagh}', "");
Expect(1, 11648, '\p{^Script_Extensions=  Tifinagh}', "");
Expect(1, 11648, '\P{Script_Extensions=  Tifinagh}', "");
Expect(0, 11648, '\P{^Script_Extensions=  Tifinagh}', "");
Error('\p{Scx=	-tfng/a/}');
Error('\P{Scx=	-tfng/a/}');
Expect(1, 11647, '\p{Scx=tfng}', "");
Expect(0, 11647, '\p{^Scx=tfng}', "");
Expect(0, 11647, '\P{Scx=tfng}', "");
Expect(1, 11647, '\P{^Scx=tfng}', "");
Expect(0, 11648, '\p{Scx=tfng}', "");
Expect(1, 11648, '\p{^Scx=tfng}', "");
Expect(1, 11648, '\P{Scx=tfng}', "");
Expect(0, 11648, '\P{^Scx=tfng}', "");
Expect(1, 11647, '\p{Scx=__Tfng}', "");
Expect(0, 11647, '\p{^Scx=__Tfng}', "");
Expect(0, 11647, '\P{Scx=__Tfng}', "");
Expect(1, 11647, '\P{^Scx=__Tfng}', "");
Expect(0, 11648, '\p{Scx=__Tfng}', "");
Expect(1, 11648, '\p{^Scx=__Tfng}', "");
Expect(1, 11648, '\P{Scx=__Tfng}', "");
Expect(0, 11648, '\P{^Scx=__Tfng}', "");
Error('\p{Is_Script_Extensions=	tifinagh/a/}');
Error('\P{Is_Script_Extensions=	tifinagh/a/}');
Expect(1, 11647, '\p{Is_Script_Extensions=tifinagh}', "");
Expect(0, 11647, '\p{^Is_Script_Extensions=tifinagh}', "");
Expect(0, 11647, '\P{Is_Script_Extensions=tifinagh}', "");
Expect(1, 11647, '\P{^Is_Script_Extensions=tifinagh}', "");
Expect(0, 11648, '\p{Is_Script_Extensions=tifinagh}', "");
Expect(1, 11648, '\p{^Is_Script_Extensions=tifinagh}', "");
Expect(1, 11648, '\P{Is_Script_Extensions=tifinagh}', "");
Expect(0, 11648, '\P{^Is_Script_Extensions=tifinagh}', "");
Expect(1, 11647, '\p{Is_Script_Extensions=Tifinagh}', "");
Expect(0, 11647, '\p{^Is_Script_Extensions=Tifinagh}', "");
Expect(0, 11647, '\P{Is_Script_Extensions=Tifinagh}', "");
Expect(1, 11647, '\P{^Is_Script_Extensions=Tifinagh}', "");
Expect(0, 11648, '\p{Is_Script_Extensions=Tifinagh}', "");
Expect(1, 11648, '\p{^Is_Script_Extensions=Tifinagh}', "");
Expect(1, 11648, '\P{Is_Script_Extensions=Tifinagh}', "");
Expect(0, 11648, '\P{^Is_Script_Extensions=Tifinagh}', "");
Error('\p{Is_Scx=/a/- tfng}');
Error('\P{Is_Scx=/a/- tfng}');
Expect(1, 11647, '\p{Is_Scx=tfng}', "");
Expect(0, 11647, '\p{^Is_Scx=tfng}', "");
Expect(0, 11647, '\P{Is_Scx=tfng}', "");
Expect(1, 11647, '\P{^Is_Scx=tfng}', "");
Expect(0, 11648, '\p{Is_Scx=tfng}', "");
Expect(1, 11648, '\p{^Is_Scx=tfng}', "");
Expect(1, 11648, '\P{Is_Scx=tfng}', "");
Expect(0, 11648, '\P{^Is_Scx=tfng}', "");
Expect(1, 11647, '\p{Is_Scx=_	Tfng}', "");
Expect(0, 11647, '\p{^Is_Scx=_	Tfng}', "");
Expect(0, 11647, '\P{Is_Scx=_	Tfng}', "");
Expect(1, 11647, '\P{^Is_Scx=_	Tfng}', "");
Expect(0, 11648, '\p{Is_Scx=_	Tfng}', "");
Expect(1, 11648, '\p{^Is_Scx=_	Tfng}', "");
Expect(1, 11648, '\P{Is_Scx=_	Tfng}', "");
Expect(0, 11648, '\P{^Is_Scx=_	Tfng}', "");
Error('\p{Script_Extensions=_:=tagalog}');
Error('\P{Script_Extensions=_:=tagalog}');
Expect(1, 5942, '\p{Script_Extensions=tagalog}', "");
Expect(0, 5942, '\p{^Script_Extensions=tagalog}', "");
Expect(0, 5942, '\P{Script_Extensions=tagalog}', "");
Expect(1, 5942, '\P{^Script_Extensions=tagalog}', "");
Expect(0, 5943, '\p{Script_Extensions=tagalog}', "");
Expect(1, 5943, '\p{^Script_Extensions=tagalog}', "");
Expect(1, 5943, '\P{Script_Extensions=tagalog}', "");
Expect(0, 5943, '\P{^Script_Extensions=tagalog}', "");
Error('\p{Scx=/a/	tglg}');
Error('\P{Scx=/a/	tglg}');
Expect(1, 5942, '\p{Scx=tglg}', "");
Expect(0, 5942, '\p{^Scx=tglg}', "");
Expect(0, 5942, '\P{Scx=tglg}', "");
Expect(1, 5942, '\P{^Scx=tglg}', "");
Expect(0, 5943, '\p{Scx=tglg}', "");
Expect(1, 5943, '\p{^Scx=tglg}', "");
Expect(1, 5943, '\P{Scx=tglg}', "");
Expect(0, 5943, '\P{^Scx=tglg}', "");
Expect(1, 5942, '\p{Scx=	 tglg}', "");
Expect(0, 5942, '\p{^Scx=	 tglg}', "");
Expect(0, 5942, '\P{Scx=	 tglg}', "");
Expect(1, 5942, '\P{^Scx=	 tglg}', "");
Expect(0, 5943, '\p{Scx=	 tglg}', "");
Expect(1, 5943, '\p{^Scx=	 tglg}', "");
Expect(1, 5943, '\P{Scx=	 tglg}', "");
Expect(0, 5943, '\P{^Scx=	 tglg}', "");
Error('\p{Is_Script_Extensions=	tagalog/a/}');
Error('\P{Is_Script_Extensions=	tagalog/a/}');
Expect(1, 5942, '\p{Is_Script_Extensions=tagalog}', "");
Expect(0, 5942, '\p{^Is_Script_Extensions=tagalog}', "");
Expect(0, 5942, '\P{Is_Script_Extensions=tagalog}', "");
Expect(1, 5942, '\P{^Is_Script_Extensions=tagalog}', "");
Expect(0, 5943, '\p{Is_Script_Extensions=tagalog}', "");
Expect(1, 5943, '\p{^Is_Script_Extensions=tagalog}', "");
Expect(1, 5943, '\P{Is_Script_Extensions=tagalog}', "");
Expect(0, 5943, '\P{^Is_Script_Extensions=tagalog}', "");
Expect(1, 5942, '\p{Is_Script_Extensions=_-Tagalog}', "");
Expect(0, 5942, '\p{^Is_Script_Extensions=_-Tagalog}', "");
Expect(0, 5942, '\P{Is_Script_Extensions=_-Tagalog}', "");
Expect(1, 5942, '\P{^Is_Script_Extensions=_-Tagalog}', "");
Expect(0, 5943, '\p{Is_Script_Extensions=_-Tagalog}', "");
Expect(1, 5943, '\p{^Is_Script_Extensions=_-Tagalog}', "");
Expect(1, 5943, '\P{Is_Script_Extensions=_-Tagalog}', "");
Expect(0, 5943, '\P{^Is_Script_Extensions=_-Tagalog}', "");
Error('\p{Is_Scx:   --tglg/a/}');
Error('\P{Is_Scx:   --tglg/a/}');
Expect(1, 5942, '\p{Is_Scx=tglg}', "");
Expect(0, 5942, '\p{^Is_Scx=tglg}', "");
Expect(0, 5942, '\P{Is_Scx=tglg}', "");
Expect(1, 5942, '\P{^Is_Scx=tglg}', "");
Expect(0, 5943, '\p{Is_Scx=tglg}', "");
Expect(1, 5943, '\p{^Is_Scx=tglg}', "");
Expect(1, 5943, '\P{Is_Scx=tglg}', "");
Expect(0, 5943, '\P{^Is_Scx=tglg}', "");
Expect(1, 5942, '\p{Is_Scx= Tglg}', "");
Expect(0, 5942, '\p{^Is_Scx= Tglg}', "");
Expect(0, 5942, '\P{Is_Scx= Tglg}', "");
Expect(1, 5942, '\P{^Is_Scx= Tglg}', "");
Expect(0, 5943, '\p{Is_Scx= Tglg}', "");
Expect(1, 5943, '\p{^Is_Scx= Tglg}', "");
Expect(1, 5943, '\P{Is_Scx= Tglg}', "");
Expect(0, 5943, '\P{^Is_Scx= Tglg}', "");
Error('\p{Script_Extensions=-Thaana:=}');
Error('\P{Script_Extensions=-Thaana:=}');
Expect(1, 65021, '\p{Script_Extensions=thaana}', "");
Expect(0, 65021, '\p{^Script_Extensions=thaana}', "");
Expect(0, 65021, '\P{Script_Extensions=thaana}', "");
Expect(1, 65021, '\P{^Script_Extensions=thaana}', "");
Expect(0, 65022, '\p{Script_Extensions=thaana}', "");
Expect(1, 65022, '\p{^Script_Extensions=thaana}', "");
Expect(1, 65022, '\P{Script_Extensions=thaana}', "");
Expect(0, 65022, '\P{^Script_Extensions=thaana}', "");
Expect(1, 65021, '\p{Script_Extensions=_-Thaana}', "");
Expect(0, 65021, '\p{^Script_Extensions=_-Thaana}', "");
Expect(0, 65021, '\P{Script_Extensions=_-Thaana}', "");
Expect(1, 65021, '\P{^Script_Extensions=_-Thaana}', "");
Expect(0, 65022, '\p{Script_Extensions=_-Thaana}', "");
Expect(1, 65022, '\p{^Script_Extensions=_-Thaana}', "");
Expect(1, 65022, '\P{Script_Extensions=_-Thaana}', "");
Expect(0, 65022, '\P{^Script_Extensions=_-Thaana}', "");
Error('\p{Scx=thaa/a/}');
Error('\P{Scx=thaa/a/}');
Expect(1, 65021, '\p{Scx=thaa}', "");
Expect(0, 65021, '\p{^Scx=thaa}', "");
Expect(0, 65021, '\P{Scx=thaa}', "");
Expect(1, 65021, '\P{^Scx=thaa}', "");
Expect(0, 65022, '\p{Scx=thaa}', "");
Expect(1, 65022, '\p{^Scx=thaa}', "");
Expect(1, 65022, '\P{Scx=thaa}', "");
Expect(0, 65022, '\P{^Scx=thaa}', "");
Expect(1, 65021, '\p{Scx=  Thaa}', "");
Expect(0, 65021, '\p{^Scx=  Thaa}', "");
Expect(0, 65021, '\P{Scx=  Thaa}', "");
Expect(1, 65021, '\P{^Scx=  Thaa}', "");
Expect(0, 65022, '\p{Scx=  Thaa}', "");
Expect(1, 65022, '\p{^Scx=  Thaa}', "");
Expect(1, 65022, '\P{Scx=  Thaa}', "");
Expect(0, 65022, '\P{^Scx=  Thaa}', "");
Error('\p{Is_Script_Extensions=-/a/thaana}');
Error('\P{Is_Script_Extensions=-/a/thaana}');
Expect(1, 65021, '\p{Is_Script_Extensions=thaana}', "");
Expect(0, 65021, '\p{^Is_Script_Extensions=thaana}', "");
Expect(0, 65021, '\P{Is_Script_Extensions=thaana}', "");
Expect(1, 65021, '\P{^Is_Script_Extensions=thaana}', "");
Expect(0, 65022, '\p{Is_Script_Extensions=thaana}', "");
Expect(1, 65022, '\p{^Is_Script_Extensions=thaana}', "");
Expect(1, 65022, '\P{Is_Script_Extensions=thaana}', "");
Expect(0, 65022, '\P{^Is_Script_Extensions=thaana}', "");
Expect(1, 65021, '\p{Is_Script_Extensions=		THAANA}', "");
Expect(0, 65021, '\p{^Is_Script_Extensions=		THAANA}', "");
Expect(0, 65021, '\P{Is_Script_Extensions=		THAANA}', "");
Expect(1, 65021, '\P{^Is_Script_Extensions=		THAANA}', "");
Expect(0, 65022, '\p{Is_Script_Extensions=		THAANA}', "");
Expect(1, 65022, '\p{^Is_Script_Extensions=		THAANA}', "");
Expect(1, 65022, '\P{Is_Script_Extensions=		THAANA}', "");
Expect(0, 65022, '\P{^Is_Script_Extensions=		THAANA}', "");
Error('\p{Is_Scx=:=-	THAA}');
Error('\P{Is_Scx=:=-	THAA}');
Expect(1, 65021, '\p{Is_Scx=thaa}', "");
Expect(0, 65021, '\p{^Is_Scx=thaa}', "");
Expect(0, 65021, '\P{Is_Scx=thaa}', "");
Expect(1, 65021, '\P{^Is_Scx=thaa}', "");
Expect(0, 65022, '\p{Is_Scx=thaa}', "");
Expect(1, 65022, '\p{^Is_Scx=thaa}', "");
Expect(1, 65022, '\P{Is_Scx=thaa}', "");
Expect(0, 65022, '\P{^Is_Scx=thaa}', "");
Expect(1, 65021, '\p{Is_Scx=_	thaa}', "");
Expect(0, 65021, '\p{^Is_Scx=_	thaa}', "");
Expect(0, 65021, '\P{Is_Scx=_	thaa}', "");
Expect(1, 65021, '\P{^Is_Scx=_	thaa}', "");
Expect(0, 65022, '\p{Is_Scx=_	thaa}', "");
Expect(1, 65022, '\p{^Is_Scx=_	thaa}', "");
Expect(1, 65022, '\P{Is_Scx=_	thaa}', "");
Expect(0, 65022, '\P{^Is_Scx=_	thaa}', "");
Error('\p{Script_Extensions=  Thai/a/}');
Error('\P{Script_Extensions=  Thai/a/}');
Expect(1, 3675, '\p{Script_Extensions:   thai}', "");
Expect(0, 3675, '\p{^Script_Extensions:   thai}', "");
Expect(0, 3675, '\P{Script_Extensions:   thai}', "");
Expect(1, 3675, '\P{^Script_Extensions:   thai}', "");
Expect(0, 3676, '\p{Script_Extensions:   thai}', "");
Expect(1, 3676, '\p{^Script_Extensions:   thai}', "");
Expect(1, 3676, '\P{Script_Extensions:   thai}', "");
Expect(0, 3676, '\P{^Script_Extensions:   thai}', "");
Expect(1, 3675, '\p{Script_Extensions=  Thai}', "");
Expect(0, 3675, '\p{^Script_Extensions=  Thai}', "");
Expect(0, 3675, '\P{Script_Extensions=  Thai}', "");
Expect(1, 3675, '\P{^Script_Extensions=  Thai}', "");
Expect(0, 3676, '\p{Script_Extensions=  Thai}', "");
Expect(1, 3676, '\p{^Script_Extensions=  Thai}', "");
Expect(1, 3676, '\P{Script_Extensions=  Thai}', "");
Expect(0, 3676, '\P{^Script_Extensions=  Thai}', "");
Error('\p{Scx=/a/THAI}');
Error('\P{Scx=/a/THAI}');
Expect(1, 3675, '\p{Scx=thai}', "");
Expect(0, 3675, '\p{^Scx=thai}', "");
Expect(0, 3675, '\P{Scx=thai}', "");
Expect(1, 3675, '\P{^Scx=thai}', "");
Expect(0, 3676, '\p{Scx=thai}', "");
Expect(1, 3676, '\p{^Scx=thai}', "");
Expect(1, 3676, '\P{Scx=thai}', "");
Expect(0, 3676, '\P{^Scx=thai}', "");
Expect(1, 3675, '\p{Scx=_ Thai}', "");
Expect(0, 3675, '\p{^Scx=_ Thai}', "");
Expect(0, 3675, '\P{Scx=_ Thai}', "");
Expect(1, 3675, '\P{^Scx=_ Thai}', "");
Expect(0, 3676, '\p{Scx=_ Thai}', "");
Expect(1, 3676, '\p{^Scx=_ Thai}', "");
Expect(1, 3676, '\P{Scx=_ Thai}', "");
Expect(0, 3676, '\P{^Scx=_ Thai}', "");
Error('\p{Is_Script_Extensions= 	Thai/a/}');
Error('\P{Is_Script_Extensions= 	Thai/a/}');
Expect(1, 3675, '\p{Is_Script_Extensions:thai}', "");
Expect(0, 3675, '\p{^Is_Script_Extensions:thai}', "");
Expect(0, 3675, '\P{Is_Script_Extensions:thai}', "");
Expect(1, 3675, '\P{^Is_Script_Extensions:thai}', "");
Expect(0, 3676, '\p{Is_Script_Extensions:thai}', "");
Expect(1, 3676, '\p{^Is_Script_Extensions:thai}', "");
Expect(1, 3676, '\P{Is_Script_Extensions:thai}', "");
Expect(0, 3676, '\P{^Is_Script_Extensions:thai}', "");
Expect(1, 3675, '\p{Is_Script_Extensions=Thai}', "");
Expect(0, 3675, '\p{^Is_Script_Extensions=Thai}', "");
Expect(0, 3675, '\P{Is_Script_Extensions=Thai}', "");
Expect(1, 3675, '\P{^Is_Script_Extensions=Thai}', "");
Expect(0, 3676, '\p{Is_Script_Extensions=Thai}', "");
Expect(1, 3676, '\p{^Is_Script_Extensions=Thai}', "");
Expect(1, 3676, '\P{Is_Script_Extensions=Thai}', "");
Expect(0, 3676, '\P{^Is_Script_Extensions=Thai}', "");
Error('\p{Is_Scx:   :=-_Thai}');
Error('\P{Is_Scx:   :=-_Thai}');
Expect(1, 3675, '\p{Is_Scx=thai}', "");
Expect(0, 3675, '\p{^Is_Scx=thai}', "");
Expect(0, 3675, '\P{Is_Scx=thai}', "");
Expect(1, 3675, '\P{^Is_Scx=thai}', "");
Expect(0, 3676, '\p{Is_Scx=thai}', "");
Expect(1, 3676, '\p{^Is_Scx=thai}', "");
Expect(1, 3676, '\P{Is_Scx=thai}', "");
Expect(0, 3676, '\P{^Is_Scx=thai}', "");
Expect(1, 3675, '\p{Is_Scx= Thai}', "");
Expect(0, 3675, '\p{^Is_Scx= Thai}', "");
Expect(0, 3675, '\P{Is_Scx= Thai}', "");
Expect(1, 3675, '\P{^Is_Scx= Thai}', "");
Expect(0, 3676, '\p{Is_Scx= Thai}', "");
Expect(1, 3676, '\p{^Is_Scx= Thai}', "");
Expect(1, 3676, '\P{Is_Scx= Thai}', "");
Expect(0, 3676, '\P{^Is_Scx= Thai}', "");
Error('\p{Script_Extensions=	/a/tibetan}');
Error('\P{Script_Extensions=	/a/tibetan}');
Expect(1, 4058, '\p{Script_Extensions=tibetan}', "");
Expect(0, 4058, '\p{^Script_Extensions=tibetan}', "");
Expect(0, 4058, '\P{Script_Extensions=tibetan}', "");
Expect(1, 4058, '\P{^Script_Extensions=tibetan}', "");
Expect(0, 4059, '\p{Script_Extensions=tibetan}', "");
Expect(1, 4059, '\p{^Script_Extensions=tibetan}', "");
Expect(1, 4059, '\P{Script_Extensions=tibetan}', "");
Expect(0, 4059, '\P{^Script_Extensions=tibetan}', "");
Expect(1, 4058, '\p{Script_Extensions=-Tibetan}', "");
Expect(0, 4058, '\p{^Script_Extensions=-Tibetan}', "");
Expect(0, 4058, '\P{Script_Extensions=-Tibetan}', "");
Expect(1, 4058, '\P{^Script_Extensions=-Tibetan}', "");
Expect(0, 4059, '\p{Script_Extensions=-Tibetan}', "");
Expect(1, 4059, '\p{^Script_Extensions=-Tibetan}', "");
Expect(1, 4059, '\P{Script_Extensions=-Tibetan}', "");
Expect(0, 4059, '\P{^Script_Extensions=-Tibetan}', "");
Error('\p{Scx=/a/	 Tibt}');
Error('\P{Scx=/a/	 Tibt}');
Expect(1, 4058, '\p{Scx=tibt}', "");
Expect(0, 4058, '\p{^Scx=tibt}', "");
Expect(0, 4058, '\P{Scx=tibt}', "");
Expect(1, 4058, '\P{^Scx=tibt}', "");
Expect(0, 4059, '\p{Scx=tibt}', "");
Expect(1, 4059, '\p{^Scx=tibt}', "");
Expect(1, 4059, '\P{Scx=tibt}', "");
Expect(0, 4059, '\P{^Scx=tibt}', "");
Expect(1, 4058, '\p{Scx= _Tibt}', "");
Expect(0, 4058, '\p{^Scx= _Tibt}', "");
Expect(0, 4058, '\P{Scx= _Tibt}', "");
Expect(1, 4058, '\P{^Scx= _Tibt}', "");
Expect(0, 4059, '\p{Scx= _Tibt}', "");
Expect(1, 4059, '\p{^Scx= _Tibt}', "");
Expect(1, 4059, '\P{Scx= _Tibt}', "");
Expect(0, 4059, '\P{^Scx= _Tibt}', "");
Error('\p{Is_Script_Extensions=:=_ Tibetan}');
Error('\P{Is_Script_Extensions=:=_ Tibetan}');
Expect(1, 4058, '\p{Is_Script_Extensions:	tibetan}', "");
Expect(0, 4058, '\p{^Is_Script_Extensions:	tibetan}', "");
Expect(0, 4058, '\P{Is_Script_Extensions:	tibetan}', "");
Expect(1, 4058, '\P{^Is_Script_Extensions:	tibetan}', "");
Expect(0, 4059, '\p{Is_Script_Extensions:	tibetan}', "");
Expect(1, 4059, '\p{^Is_Script_Extensions:	tibetan}', "");
Expect(1, 4059, '\P{Is_Script_Extensions:	tibetan}', "");
Expect(0, 4059, '\P{^Is_Script_Extensions:	tibetan}', "");
Expect(1, 4058, '\p{Is_Script_Extensions=	_TIBETAN}', "");
Expect(0, 4058, '\p{^Is_Script_Extensions=	_TIBETAN}', "");
Expect(0, 4058, '\P{Is_Script_Extensions=	_TIBETAN}', "");
Expect(1, 4058, '\P{^Is_Script_Extensions=	_TIBETAN}', "");
Expect(0, 4059, '\p{Is_Script_Extensions=	_TIBETAN}', "");
Expect(1, 4059, '\p{^Is_Script_Extensions=	_TIBETAN}', "");
Expect(1, 4059, '\P{Is_Script_Extensions=	_TIBETAN}', "");
Expect(0, 4059, '\P{^Is_Script_Extensions=	_TIBETAN}', "");
Error('\p{Is_Scx=/a/- TIBT}');
Error('\P{Is_Scx=/a/- TIBT}');
Expect(1, 4058, '\p{Is_Scx=tibt}', "");
Expect(0, 4058, '\p{^Is_Scx=tibt}', "");
Expect(0, 4058, '\P{Is_Scx=tibt}', "");
Expect(1, 4058, '\P{^Is_Scx=tibt}', "");
Expect(0, 4059, '\p{Is_Scx=tibt}', "");
Expect(1, 4059, '\p{^Is_Scx=tibt}', "");
Expect(1, 4059, '\P{Is_Scx=tibt}', "");
Expect(0, 4059, '\P{^Is_Scx=tibt}', "");
Expect(1, 4058, '\p{Is_Scx:	-Tibt}', "");
Expect(0, 4058, '\p{^Is_Scx:	-Tibt}', "");
Expect(0, 4058, '\P{Is_Scx:	-Tibt}', "");
Expect(1, 4058, '\P{^Is_Scx:	-Tibt}', "");
Expect(0, 4059, '\p{Is_Scx:	-Tibt}', "");
Expect(1, 4059, '\p{^Is_Scx:	-Tibt}', "");
Expect(1, 4059, '\P{Is_Scx:	-Tibt}', "");
Expect(0, 4059, '\P{^Is_Scx:	-Tibt}', "");
Error('\p{Script_Extensions=_/a/UGARITIC}');
Error('\P{Script_Extensions=_/a/UGARITIC}');
Expect(1, 66463, '\p{Script_Extensions=ugaritic}', "");
Expect(0, 66463, '\p{^Script_Extensions=ugaritic}', "");
Expect(0, 66463, '\P{Script_Extensions=ugaritic}', "");
Expect(1, 66463, '\P{^Script_Extensions=ugaritic}', "");
Expect(0, 66464, '\p{Script_Extensions=ugaritic}', "");
Expect(1, 66464, '\p{^Script_Extensions=ugaritic}', "");
Expect(1, 66464, '\P{Script_Extensions=ugaritic}', "");
Expect(0, 66464, '\P{^Script_Extensions=ugaritic}', "");
Expect(1, 66463, '\p{Script_Extensions= -Ugaritic}', "");
Expect(0, 66463, '\p{^Script_Extensions= -Ugaritic}', "");
Expect(0, 66463, '\P{Script_Extensions= -Ugaritic}', "");
Expect(1, 66463, '\P{^Script_Extensions= -Ugaritic}', "");
Expect(0, 66464, '\p{Script_Extensions= -Ugaritic}', "");
Expect(1, 66464, '\p{^Script_Extensions= -Ugaritic}', "");
Expect(1, 66464, '\P{Script_Extensions= -Ugaritic}', "");
Expect(0, 66464, '\P{^Script_Extensions= -Ugaritic}', "");
Error('\p{Scx=  Ugar:=}');
Error('\P{Scx=  Ugar:=}');
Expect(1, 66463, '\p{Scx=ugar}', "");
Expect(0, 66463, '\p{^Scx=ugar}', "");
Expect(0, 66463, '\P{Scx=ugar}', "");
Expect(1, 66463, '\P{^Scx=ugar}', "");
Expect(0, 66464, '\p{Scx=ugar}', "");
Expect(1, 66464, '\p{^Scx=ugar}', "");
Expect(1, 66464, '\P{Scx=ugar}', "");
Expect(0, 66464, '\P{^Scx=ugar}', "");
Expect(1, 66463, '\p{Scx=	-Ugar}', "");
Expect(0, 66463, '\p{^Scx=	-Ugar}', "");
Expect(0, 66463, '\P{Scx=	-Ugar}', "");
Expect(1, 66463, '\P{^Scx=	-Ugar}', "");
Expect(0, 66464, '\p{Scx=	-Ugar}', "");
Expect(1, 66464, '\p{^Scx=	-Ugar}', "");
Expect(1, 66464, '\P{Scx=	-Ugar}', "");
Expect(0, 66464, '\P{^Scx=	-Ugar}', "");
Error('\p{Is_Script_Extensions=/a/	 UGARITIC}');
Error('\P{Is_Script_Extensions=/a/	 UGARITIC}');
Expect(1, 66463, '\p{Is_Script_Extensions=ugaritic}', "");
Expect(0, 66463, '\p{^Is_Script_Extensions=ugaritic}', "");
Expect(0, 66463, '\P{Is_Script_Extensions=ugaritic}', "");
Expect(1, 66463, '\P{^Is_Script_Extensions=ugaritic}', "");
Expect(0, 66464, '\p{Is_Script_Extensions=ugaritic}', "");
Expect(1, 66464, '\p{^Is_Script_Extensions=ugaritic}', "");
Expect(1, 66464, '\P{Is_Script_Extensions=ugaritic}', "");
Expect(0, 66464, '\P{^Is_Script_Extensions=ugaritic}', "");
Expect(1, 66463, '\p{Is_Script_Extensions=-_Ugaritic}', "");
Expect(0, 66463, '\p{^Is_Script_Extensions=-_Ugaritic}', "");
Expect(0, 66463, '\P{Is_Script_Extensions=-_Ugaritic}', "");
Expect(1, 66463, '\P{^Is_Script_Extensions=-_Ugaritic}', "");
Expect(0, 66464, '\p{Is_Script_Extensions=-_Ugaritic}', "");
Expect(1, 66464, '\p{^Is_Script_Extensions=-_Ugaritic}', "");
Expect(1, 66464, '\P{Is_Script_Extensions=-_Ugaritic}', "");
Expect(0, 66464, '\P{^Is_Script_Extensions=-_Ugaritic}', "");
Error('\p{Is_Scx=:=	-Ugar}');
Error('\P{Is_Scx=:=	-Ugar}');
Expect(1, 66463, '\p{Is_Scx=ugar}', "");
Expect(0, 66463, '\p{^Is_Scx=ugar}', "");
Expect(0, 66463, '\P{Is_Scx=ugar}', "");
Expect(1, 66463, '\P{^Is_Scx=ugar}', "");
Expect(0, 66464, '\p{Is_Scx=ugar}', "");
Expect(1, 66464, '\p{^Is_Scx=ugar}', "");
Expect(1, 66464, '\P{Is_Scx=ugar}', "");
Expect(0, 66464, '\P{^Is_Scx=ugar}', "");
Expect(1, 66463, '\p{Is_Scx=_ugar}', "");
Expect(0, 66463, '\p{^Is_Scx=_ugar}', "");
Expect(0, 66463, '\P{Is_Scx=_ugar}', "");
Expect(1, 66463, '\P{^Is_Scx=_ugar}', "");
Expect(0, 66464, '\p{Is_Scx=_ugar}', "");
Expect(1, 66464, '\p{^Is_Scx=_ugar}', "");
Expect(1, 66464, '\P{Is_Scx=_ugar}', "");
Expect(0, 66464, '\P{^Is_Scx=_ugar}', "");
Error('\p{Script_Extensions= :=VAI}');
Error('\P{Script_Extensions= :=VAI}');
Expect(1, 42539, '\p{Script_Extensions=vai}', "");
Expect(0, 42539, '\p{^Script_Extensions=vai}', "");
Expect(0, 42539, '\P{Script_Extensions=vai}', "");
Expect(1, 42539, '\P{^Script_Extensions=vai}', "");
Expect(0, 42540, '\p{Script_Extensions=vai}', "");
Expect(1, 42540, '\p{^Script_Extensions=vai}', "");
Expect(1, 42540, '\P{Script_Extensions=vai}', "");
Expect(0, 42540, '\P{^Script_Extensions=vai}', "");
Expect(1, 42539, '\p{Script_Extensions=_-Vai}', "");
Expect(0, 42539, '\p{^Script_Extensions=_-Vai}', "");
Expect(0, 42539, '\P{Script_Extensions=_-Vai}', "");
Expect(1, 42539, '\P{^Script_Extensions=_-Vai}', "");
Expect(0, 42540, '\p{Script_Extensions=_-Vai}', "");
Expect(1, 42540, '\p{^Script_Extensions=_-Vai}', "");
Expect(1, 42540, '\P{Script_Extensions=_-Vai}', "");
Expect(0, 42540, '\P{^Script_Extensions=_-Vai}', "");
Error('\p{Scx= 	VAII:=}');
Error('\P{Scx= 	VAII:=}');
Expect(1, 42539, '\p{Scx=vaii}', "");
Expect(0, 42539, '\p{^Scx=vaii}', "");
Expect(0, 42539, '\P{Scx=vaii}', "");
Expect(1, 42539, '\P{^Scx=vaii}', "");
Expect(0, 42540, '\p{Scx=vaii}', "");
Expect(1, 42540, '\p{^Scx=vaii}', "");
Expect(1, 42540, '\P{Scx=vaii}', "");
Expect(0, 42540, '\P{^Scx=vaii}', "");
Expect(1, 42539, '\p{Scx=  Vaii}', "");
Expect(0, 42539, '\p{^Scx=  Vaii}', "");
Expect(0, 42539, '\P{Scx=  Vaii}', "");
Expect(1, 42539, '\P{^Scx=  Vaii}', "");
Expect(0, 42540, '\p{Scx=  Vaii}', "");
Expect(1, 42540, '\p{^Scx=  Vaii}', "");
Expect(1, 42540, '\P{Scx=  Vaii}', "");
Expect(0, 42540, '\P{^Scx=  Vaii}', "");
Error('\p{Is_Script_Extensions=:=-	Vai}');
Error('\P{Is_Script_Extensions=:=-	Vai}');
Expect(1, 42539, '\p{Is_Script_Extensions=vai}', "");
Expect(0, 42539, '\p{^Is_Script_Extensions=vai}', "");
Expect(0, 42539, '\P{Is_Script_Extensions=vai}', "");
Expect(1, 42539, '\P{^Is_Script_Extensions=vai}', "");
Expect(0, 42540, '\p{Is_Script_Extensions=vai}', "");
Expect(1, 42540, '\p{^Is_Script_Extensions=vai}', "");
Expect(1, 42540, '\P{Is_Script_Extensions=vai}', "");
Expect(0, 42540, '\P{^Is_Script_Extensions=vai}', "");
Expect(1, 42539, '\p{Is_Script_Extensions=	vai}', "");
Expect(0, 42539, '\p{^Is_Script_Extensions=	vai}', "");
Expect(0, 42539, '\P{Is_Script_Extensions=	vai}', "");
Expect(1, 42539, '\P{^Is_Script_Extensions=	vai}', "");
Expect(0, 42540, '\p{Is_Script_Extensions=	vai}', "");
Expect(1, 42540, '\p{^Is_Script_Extensions=	vai}', "");
Expect(1, 42540, '\P{Is_Script_Extensions=	vai}', "");
Expect(0, 42540, '\P{^Is_Script_Extensions=	vai}', "");
Error('\p{Is_Scx=	-VAII/a/}');
Error('\P{Is_Scx=	-VAII/a/}');
Expect(1, 42539, '\p{Is_Scx=vaii}', "");
Expect(0, 42539, '\p{^Is_Scx=vaii}', "");
Expect(0, 42539, '\P{Is_Scx=vaii}', "");
Expect(1, 42539, '\P{^Is_Scx=vaii}', "");
Expect(0, 42540, '\p{Is_Scx=vaii}', "");
Expect(1, 42540, '\p{^Is_Scx=vaii}', "");
Expect(1, 42540, '\P{Is_Scx=vaii}', "");
Expect(0, 42540, '\P{^Is_Scx=vaii}', "");
Expect(1, 42539, '\p{Is_Scx:   --vaii}', "");
Expect(0, 42539, '\p{^Is_Scx:   --vaii}', "");
Expect(0, 42539, '\P{Is_Scx:   --vaii}', "");
Expect(1, 42539, '\P{^Is_Scx:   --vaii}', "");
Expect(0, 42540, '\p{Is_Scx:   --vaii}', "");
Expect(1, 42540, '\p{^Is_Scx:   --vaii}', "");
Expect(1, 42540, '\P{Is_Scx:   --vaii}', "");
Expect(0, 42540, '\P{^Is_Scx:   --vaii}', "");
Error('\p{Script_Extensions=:=- Old_persian}');
Error('\P{Script_Extensions=:=- Old_persian}');
Expect(1, 66517, '\p{Script_Extensions=oldpersian}', "");
Expect(0, 66517, '\p{^Script_Extensions=oldpersian}', "");
Expect(0, 66517, '\P{Script_Extensions=oldpersian}', "");
Expect(1, 66517, '\P{^Script_Extensions=oldpersian}', "");
Expect(0, 66518, '\p{Script_Extensions=oldpersian}', "");
Expect(1, 66518, '\p{^Script_Extensions=oldpersian}', "");
Expect(1, 66518, '\P{Script_Extensions=oldpersian}', "");
Expect(0, 66518, '\P{^Script_Extensions=oldpersian}', "");
Expect(1, 66517, '\p{Script_Extensions= -old_PERSIAN}', "");
Expect(0, 66517, '\p{^Script_Extensions= -old_PERSIAN}', "");
Expect(0, 66517, '\P{Script_Extensions= -old_PERSIAN}', "");
Expect(1, 66517, '\P{^Script_Extensions= -old_PERSIAN}', "");
Expect(0, 66518, '\p{Script_Extensions= -old_PERSIAN}', "");
Expect(1, 66518, '\p{^Script_Extensions= -old_PERSIAN}', "");
Expect(1, 66518, '\P{Script_Extensions= -old_PERSIAN}', "");
Expect(0, 66518, '\P{^Script_Extensions= -old_PERSIAN}', "");
Error('\p{Scx= :=Xpeo}');
Error('\P{Scx= :=Xpeo}');
Expect(1, 66517, '\p{Scx=xpeo}', "");
Expect(0, 66517, '\p{^Scx=xpeo}', "");
Expect(0, 66517, '\P{Scx=xpeo}', "");
Expect(1, 66517, '\P{^Scx=xpeo}', "");
Expect(0, 66518, '\p{Scx=xpeo}', "");
Expect(1, 66518, '\p{^Scx=xpeo}', "");
Expect(1, 66518, '\P{Scx=xpeo}', "");
Expect(0, 66518, '\P{^Scx=xpeo}', "");
Expect(1, 66517, '\p{Scx=_ Xpeo}', "");
Expect(0, 66517, '\p{^Scx=_ Xpeo}', "");
Expect(0, 66517, '\P{Scx=_ Xpeo}', "");
Expect(1, 66517, '\P{^Scx=_ Xpeo}', "");
Expect(0, 66518, '\p{Scx=_ Xpeo}', "");
Expect(1, 66518, '\p{^Scx=_ Xpeo}', "");
Expect(1, 66518, '\P{Scx=_ Xpeo}', "");
Expect(0, 66518, '\P{^Scx=_ Xpeo}', "");
Error('\p{Is_Script_Extensions: 	old_Persian/a/}');
Error('\P{Is_Script_Extensions: 	old_Persian/a/}');
Expect(1, 66517, '\p{Is_Script_Extensions=oldpersian}', "");
Expect(0, 66517, '\p{^Is_Script_Extensions=oldpersian}', "");
Expect(0, 66517, '\P{Is_Script_Extensions=oldpersian}', "");
Expect(1, 66517, '\P{^Is_Script_Extensions=oldpersian}', "");
Expect(0, 66518, '\p{Is_Script_Extensions=oldpersian}', "");
Expect(1, 66518, '\p{^Is_Script_Extensions=oldpersian}', "");
Expect(1, 66518, '\P{Is_Script_Extensions=oldpersian}', "");
Expect(0, 66518, '\P{^Is_Script_Extensions=oldpersian}', "");
Expect(1, 66517, '\p{Is_Script_Extensions=-Old_PERSIAN}', "");
Expect(0, 66517, '\p{^Is_Script_Extensions=-Old_PERSIAN}', "");
Expect(0, 66517, '\P{Is_Script_Extensions=-Old_PERSIAN}', "");
Expect(1, 66517, '\P{^Is_Script_Extensions=-Old_PERSIAN}', "");
Expect(0, 66518, '\p{Is_Script_Extensions=-Old_PERSIAN}', "");
Expect(1, 66518, '\p{^Is_Script_Extensions=-Old_PERSIAN}', "");
Expect(1, 66518, '\P{Is_Script_Extensions=-Old_PERSIAN}', "");
Expect(0, 66518, '\P{^Is_Script_Extensions=-Old_PERSIAN}', "");
Error('\p{Is_Scx:		 Xpeo/a/}');
Error('\P{Is_Scx:		 Xpeo/a/}');
Expect(1, 66517, '\p{Is_Scx=xpeo}', "");
Expect(0, 66517, '\p{^Is_Scx=xpeo}', "");
Expect(0, 66517, '\P{Is_Scx=xpeo}', "");
Expect(1, 66517, '\P{^Is_Scx=xpeo}', "");
Expect(0, 66518, '\p{Is_Scx=xpeo}', "");
Expect(1, 66518, '\p{^Is_Scx=xpeo}', "");
Expect(1, 66518, '\P{Is_Scx=xpeo}', "");
Expect(0, 66518, '\P{^Is_Scx=xpeo}', "");
Expect(1, 66517, '\p{Is_Scx: 	xpeo}', "");
Expect(0, 66517, '\p{^Is_Scx: 	xpeo}', "");
Expect(0, 66517, '\P{Is_Scx: 	xpeo}', "");
Expect(1, 66517, '\P{^Is_Scx: 	xpeo}', "");
Expect(0, 66518, '\p{Is_Scx: 	xpeo}', "");
Expect(1, 66518, '\p{^Is_Scx: 	xpeo}', "");
Expect(1, 66518, '\P{Is_Scx: 	xpeo}', "");
Expect(0, 66518, '\P{^Is_Scx: 	xpeo}', "");
Error('\p{Script_Extensions=_cuneiform/a/}');
Error('\P{Script_Extensions=_cuneiform/a/}');
Expect(1, 74867, '\p{Script_Extensions=cuneiform}', "");
Expect(0, 74867, '\p{^Script_Extensions=cuneiform}', "");
Expect(0, 74867, '\P{Script_Extensions=cuneiform}', "");
Expect(1, 74867, '\P{^Script_Extensions=cuneiform}', "");
Expect(0, 74868, '\p{Script_Extensions=cuneiform}', "");
Expect(1, 74868, '\p{^Script_Extensions=cuneiform}', "");
Expect(1, 74868, '\P{Script_Extensions=cuneiform}', "");
Expect(0, 74868, '\P{^Script_Extensions=cuneiform}', "");
Expect(1, 74867, '\p{Script_Extensions:__CUNEIFORM}', "");
Expect(0, 74867, '\p{^Script_Extensions:__CUNEIFORM}', "");
Expect(0, 74867, '\P{Script_Extensions:__CUNEIFORM}', "");
Expect(1, 74867, '\P{^Script_Extensions:__CUNEIFORM}', "");
Expect(0, 74868, '\p{Script_Extensions:__CUNEIFORM}', "");
Expect(1, 74868, '\p{^Script_Extensions:__CUNEIFORM}', "");
Expect(1, 74868, '\P{Script_Extensions:__CUNEIFORM}', "");
Expect(0, 74868, '\P{^Script_Extensions:__CUNEIFORM}', "");
Error('\p{Scx= :=xsux}');
Error('\P{Scx= :=xsux}');
Expect(1, 74867, '\p{Scx=xsux}', "");
Expect(0, 74867, '\p{^Scx=xsux}', "");
Expect(0, 74867, '\P{Scx=xsux}', "");
Expect(1, 74867, '\P{^Scx=xsux}', "");
Expect(0, 74868, '\p{Scx=xsux}', "");
Expect(1, 74868, '\p{^Scx=xsux}', "");
Expect(1, 74868, '\P{Scx=xsux}', "");
Expect(0, 74868, '\P{^Scx=xsux}', "");
Expect(1, 74867, '\p{Scx=_xsux}', "");
Expect(0, 74867, '\p{^Scx=_xsux}', "");
Expect(0, 74867, '\P{Scx=_xsux}', "");
Expect(1, 74867, '\P{^Scx=_xsux}', "");
Expect(0, 74868, '\p{Scx=_xsux}', "");
Expect(1, 74868, '\p{^Scx=_xsux}', "");
Expect(1, 74868, '\P{Scx=_xsux}', "");
Expect(0, 74868, '\P{^Scx=_xsux}', "");
Error('\p{Is_Script_Extensions:   	Cuneiform/a/}');
Error('\P{Is_Script_Extensions:   	Cuneiform/a/}');
Expect(1, 74867, '\p{Is_Script_Extensions=cuneiform}', "");
Expect(0, 74867, '\p{^Is_Script_Extensions=cuneiform}', "");
Expect(0, 74867, '\P{Is_Script_Extensions=cuneiform}', "");
Expect(1, 74867, '\P{^Is_Script_Extensions=cuneiform}', "");
Expect(0, 74868, '\p{Is_Script_Extensions=cuneiform}', "");
Expect(1, 74868, '\p{^Is_Script_Extensions=cuneiform}', "");
Expect(1, 74868, '\P{Is_Script_Extensions=cuneiform}', "");
Expect(0, 74868, '\P{^Is_Script_Extensions=cuneiform}', "");
Expect(1, 74867, '\p{Is_Script_Extensions=  Cuneiform}', "");
Expect(0, 74867, '\p{^Is_Script_Extensions=  Cuneiform}', "");
Expect(0, 74867, '\P{Is_Script_Extensions=  Cuneiform}', "");
Expect(1, 74867, '\P{^Is_Script_Extensions=  Cuneiform}', "");
Expect(0, 74868, '\p{Is_Script_Extensions=  Cuneiform}', "");
Expect(1, 74868, '\p{^Is_Script_Extensions=  Cuneiform}', "");
Expect(1, 74868, '\P{Is_Script_Extensions=  Cuneiform}', "");
Expect(0, 74868, '\P{^Is_Script_Extensions=  Cuneiform}', "");
Error('\p{Is_Scx=	/a/xsux}');
Error('\P{Is_Scx=	/a/xsux}');
Expect(1, 74867, '\p{Is_Scx=xsux}', "");
Expect(0, 74867, '\p{^Is_Scx=xsux}', "");
Expect(0, 74867, '\P{Is_Scx=xsux}', "");
Expect(1, 74867, '\P{^Is_Scx=xsux}', "");
Expect(0, 74868, '\p{Is_Scx=xsux}', "");
Expect(1, 74868, '\p{^Is_Scx=xsux}', "");
Expect(1, 74868, '\P{Is_Scx=xsux}', "");
Expect(0, 74868, '\P{^Is_Scx=xsux}', "");
Expect(1, 74867, '\p{Is_Scx=_Xsux}', "");
Expect(0, 74867, '\p{^Is_Scx=_Xsux}', "");
Expect(0, 74867, '\P{Is_Scx=_Xsux}', "");
Expect(1, 74867, '\P{^Is_Scx=_Xsux}', "");
Expect(0, 74868, '\p{Is_Scx=_Xsux}', "");
Expect(1, 74868, '\p{^Is_Scx=_Xsux}', "");
Expect(1, 74868, '\P{Is_Scx=_Xsux}', "");
Expect(0, 74868, '\P{^Is_Scx=_Xsux}', "");
Error('\p{Script_Extensions= Yi/a/}');
Error('\P{Script_Extensions= Yi/a/}');
Expect(1, 65381, '\p{Script_Extensions=yi}', "");
Expect(0, 65381, '\p{^Script_Extensions=yi}', "");
Expect(0, 65381, '\P{Script_Extensions=yi}', "");
Expect(1, 65381, '\P{^Script_Extensions=yi}', "");
Expect(0, 65382, '\p{Script_Extensions=yi}', "");
Expect(1, 65382, '\p{^Script_Extensions=yi}', "");
Expect(1, 65382, '\P{Script_Extensions=yi}', "");
Expect(0, 65382, '\P{^Script_Extensions=yi}', "");
Expect(1, 65381, '\p{Script_Extensions:  yi}', "");
Expect(0, 65381, '\p{^Script_Extensions:  yi}', "");
Expect(0, 65381, '\P{Script_Extensions:  yi}', "");
Expect(1, 65381, '\P{^Script_Extensions:  yi}', "");
Expect(0, 65382, '\p{Script_Extensions:  yi}', "");
Expect(1, 65382, '\p{^Script_Extensions:  yi}', "");
Expect(1, 65382, '\P{Script_Extensions:  yi}', "");
Expect(0, 65382, '\P{^Script_Extensions:  yi}', "");
Error('\p{Scx=-_yiii/a/}');
Error('\P{Scx=-_yiii/a/}');
Expect(1, 65381, '\p{Scx=yiii}', "");
Expect(0, 65381, '\p{^Scx=yiii}', "");
Expect(0, 65381, '\P{Scx=yiii}', "");
Expect(1, 65381, '\P{^Scx=yiii}', "");
Expect(0, 65382, '\p{Scx=yiii}', "");
Expect(1, 65382, '\p{^Scx=yiii}', "");
Expect(1, 65382, '\P{Scx=yiii}', "");
Expect(0, 65382, '\P{^Scx=yiii}', "");
Expect(1, 65381, '\p{Scx=	 yiii}', "");
Expect(0, 65381, '\p{^Scx=	 yiii}', "");
Expect(0, 65381, '\P{Scx=	 yiii}', "");
Expect(1, 65381, '\P{^Scx=	 yiii}', "");
Expect(0, 65382, '\p{Scx=	 yiii}', "");
Expect(1, 65382, '\p{^Scx=	 yiii}', "");
Expect(1, 65382, '\P{Scx=	 yiii}', "");
Expect(0, 65382, '\P{^Scx=	 yiii}', "");
Error('\p{Is_Script_Extensions=:=	Yi}');
Error('\P{Is_Script_Extensions=:=	Yi}');
Expect(1, 65381, '\p{Is_Script_Extensions=yi}', "");
Expect(0, 65381, '\p{^Is_Script_Extensions=yi}', "");
Expect(0, 65381, '\P{Is_Script_Extensions=yi}', "");
Expect(1, 65381, '\P{^Is_Script_Extensions=yi}', "");
Expect(0, 65382, '\p{Is_Script_Extensions=yi}', "");
Expect(1, 65382, '\p{^Is_Script_Extensions=yi}', "");
Expect(1, 65382, '\P{Is_Script_Extensions=yi}', "");
Expect(0, 65382, '\P{^Is_Script_Extensions=yi}', "");
Expect(1, 65381, '\p{Is_Script_Extensions=	 YI}', "");
Expect(0, 65381, '\p{^Is_Script_Extensions=	 YI}', "");
Expect(0, 65381, '\P{Is_Script_Extensions=	 YI}', "");
Expect(1, 65381, '\P{^Is_Script_Extensions=	 YI}', "");
Expect(0, 65382, '\p{Is_Script_Extensions=	 YI}', "");
Expect(1, 65382, '\p{^Is_Script_Extensions=	 YI}', "");
Expect(1, 65382, '\P{Is_Script_Extensions=	 YI}', "");
Expect(0, 65382, '\P{^Is_Script_Extensions=	 YI}', "");
Error('\p{Is_Scx=- YIII:=}');
Error('\P{Is_Scx=- YIII:=}');
Expect(1, 65381, '\p{Is_Scx=yiii}', "");
Expect(0, 65381, '\p{^Is_Scx=yiii}', "");
Expect(0, 65381, '\P{Is_Scx=yiii}', "");
Expect(1, 65381, '\P{^Is_Scx=yiii}', "");
Expect(0, 65382, '\p{Is_Scx=yiii}', "");
Expect(1, 65382, '\p{^Is_Scx=yiii}', "");
Expect(1, 65382, '\P{Is_Scx=yiii}', "");
Expect(0, 65382, '\P{^Is_Scx=yiii}', "");
Expect(1, 65381, '\p{Is_Scx=	Yiii}', "");
Expect(0, 65381, '\p{^Is_Scx=	Yiii}', "");
Expect(0, 65381, '\P{Is_Scx=	Yiii}', "");
Expect(1, 65381, '\P{^Is_Scx=	Yiii}', "");
Expect(0, 65382, '\p{Is_Scx=	Yiii}', "");
Expect(1, 65382, '\p{^Is_Scx=	Yiii}', "");
Expect(1, 65382, '\P{Is_Scx=	Yiii}', "");
Expect(0, 65382, '\P{^Is_Scx=	Yiii}', "");
Error('\p{Script_Extensions=/a/-_inherited}');
Error('\P{Script_Extensions=/a/-_inherited}');
Expect(1, 917999, '\p{Script_Extensions=inherited}', "");
Expect(0, 917999, '\p{^Script_Extensions=inherited}', "");
Expect(0, 917999, '\P{Script_Extensions=inherited}', "");
Expect(1, 917999, '\P{^Script_Extensions=inherited}', "");
Expect(0, 918000, '\p{Script_Extensions=inherited}', "");
Expect(1, 918000, '\p{^Script_Extensions=inherited}', "");
Expect(1, 918000, '\P{Script_Extensions=inherited}', "");
Expect(0, 918000, '\P{^Script_Extensions=inherited}', "");
Expect(1, 917999, '\p{Script_Extensions=  Inherited}', "");
Expect(0, 917999, '\p{^Script_Extensions=  Inherited}', "");
Expect(0, 917999, '\P{Script_Extensions=  Inherited}', "");
Expect(1, 917999, '\P{^Script_Extensions=  Inherited}', "");
Expect(0, 918000, '\p{Script_Extensions=  Inherited}', "");
Expect(1, 918000, '\p{^Script_Extensions=  Inherited}', "");
Expect(1, 918000, '\P{Script_Extensions=  Inherited}', "");
Expect(0, 918000, '\P{^Script_Extensions=  Inherited}', "");
Error('\p{Scx=__zinh:=}');
Error('\P{Scx=__zinh:=}');
Expect(1, 917999, '\p{Scx=zinh}', "");
Expect(0, 917999, '\p{^Scx=zinh}', "");
Expect(0, 917999, '\P{Scx=zinh}', "");
Expect(1, 917999, '\P{^Scx=zinh}', "");
Expect(0, 918000, '\p{Scx=zinh}', "");
Expect(1, 918000, '\p{^Scx=zinh}', "");
Expect(1, 918000, '\P{Scx=zinh}', "");
Expect(0, 918000, '\P{^Scx=zinh}', "");
Expect(1, 917999, '\p{Scx:   _Zinh}', "");
Expect(0, 917999, '\p{^Scx:   _Zinh}', "");
Expect(0, 917999, '\P{Scx:   _Zinh}', "");
Expect(1, 917999, '\P{^Scx:   _Zinh}', "");
Expect(0, 918000, '\p{Scx:   _Zinh}', "");
Expect(1, 918000, '\p{^Scx:   _Zinh}', "");
Expect(1, 918000, '\P{Scx:   _Zinh}', "");
Expect(0, 918000, '\P{^Scx:   _Zinh}', "");
Error('\p{Is_Script_Extensions:   :=qaai}');
Error('\P{Is_Script_Extensions:   :=qaai}');
Expect(1, 917999, '\p{Is_Script_Extensions=qaai}', "");
Expect(0, 917999, '\p{^Is_Script_Extensions=qaai}', "");
Expect(0, 917999, '\P{Is_Script_Extensions=qaai}', "");
Expect(1, 917999, '\P{^Is_Script_Extensions=qaai}', "");
Expect(0, 918000, '\p{Is_Script_Extensions=qaai}', "");
Expect(1, 918000, '\p{^Is_Script_Extensions=qaai}', "");
Expect(1, 918000, '\P{Is_Script_Extensions=qaai}', "");
Expect(0, 918000, '\P{^Is_Script_Extensions=qaai}', "");
Expect(1, 917999, '\p{Is_Script_Extensions=_QAAI}', "");
Expect(0, 917999, '\p{^Is_Script_Extensions=_QAAI}', "");
Expect(0, 917999, '\P{Is_Script_Extensions=_QAAI}', "");
Expect(1, 917999, '\P{^Is_Script_Extensions=_QAAI}', "");
Expect(0, 918000, '\p{Is_Script_Extensions=_QAAI}', "");
Expect(1, 918000, '\p{^Is_Script_Extensions=_QAAI}', "");
Expect(1, 918000, '\P{Is_Script_Extensions=_QAAI}', "");
Expect(0, 918000, '\P{^Is_Script_Extensions=_QAAI}', "");
Error('\p{Is_Scx::=	_inherited}');
Error('\P{Is_Scx::=	_inherited}');
Expect(1, 917999, '\p{Is_Scx: inherited}', "");
Expect(0, 917999, '\p{^Is_Scx: inherited}', "");
Expect(0, 917999, '\P{Is_Scx: inherited}', "");
Expect(1, 917999, '\P{^Is_Scx: inherited}', "");
Expect(0, 918000, '\p{Is_Scx: inherited}', "");
Expect(1, 918000, '\p{^Is_Scx: inherited}', "");
Expect(1, 918000, '\P{Is_Scx: inherited}', "");
Expect(0, 918000, '\P{^Is_Scx: inherited}', "");
Expect(1, 917999, '\p{Is_Scx=INHERITED}', "");
Expect(0, 917999, '\p{^Is_Scx=INHERITED}', "");
Expect(0, 917999, '\P{Is_Scx=INHERITED}', "");
Expect(1, 917999, '\P{^Is_Scx=INHERITED}', "");
Expect(0, 918000, '\p{Is_Scx=INHERITED}', "");
Expect(1, 918000, '\p{^Is_Scx=INHERITED}', "");
Expect(1, 918000, '\P{Is_Scx=INHERITED}', "");
Expect(0, 918000, '\P{^Is_Scx=INHERITED}', "");
Error('\p{Script_Extensions= -Common/a/}');
Error('\P{Script_Extensions= -Common/a/}');
Expect(1, 917631, '\p{Script_Extensions=common}', "");
Expect(0, 917631, '\p{^Script_Extensions=common}', "");
Expect(0, 917631, '\P{Script_Extensions=common}', "");
Expect(1, 917631, '\P{^Script_Extensions=common}', "");
Expect(0, 917632, '\p{Script_Extensions=common}', "");
Expect(1, 917632, '\p{^Script_Extensions=common}', "");
Expect(1, 917632, '\P{Script_Extensions=common}', "");
Expect(0, 917632, '\P{^Script_Extensions=common}', "");
Expect(1, 917631, '\p{Script_Extensions=- COMMON}', "");
Expect(0, 917631, '\p{^Script_Extensions=- COMMON}', "");
Expect(0, 917631, '\P{Script_Extensions=- COMMON}', "");
Expect(1, 917631, '\P{^Script_Extensions=- COMMON}', "");
Expect(0, 917632, '\p{Script_Extensions=- COMMON}', "");
Expect(1, 917632, '\p{^Script_Extensions=- COMMON}', "");
Expect(1, 917632, '\P{Script_Extensions=- COMMON}', "");
Expect(0, 917632, '\P{^Script_Extensions=- COMMON}', "");
Error('\p{Scx:	/a/ _ZYYY}');
Error('\P{Scx:	/a/ _ZYYY}');
Expect(1, 917631, '\p{Scx=zyyy}', "");
Expect(0, 917631, '\p{^Scx=zyyy}', "");
Expect(0, 917631, '\P{Scx=zyyy}', "");
Expect(1, 917631, '\P{^Scx=zyyy}', "");
Expect(0, 917632, '\p{Scx=zyyy}', "");
Expect(1, 917632, '\p{^Scx=zyyy}', "");
Expect(1, 917632, '\P{Scx=zyyy}', "");
Expect(0, 917632, '\P{^Scx=zyyy}', "");
Expect(1, 917631, '\p{Scx:__ZYYY}', "");
Expect(0, 917631, '\p{^Scx:__ZYYY}', "");
Expect(0, 917631, '\P{Scx:__ZYYY}', "");
Expect(1, 917631, '\P{^Scx:__ZYYY}', "");
Expect(0, 917632, '\p{Scx:__ZYYY}', "");
Expect(1, 917632, '\p{^Scx:__ZYYY}', "");
Expect(1, 917632, '\P{Scx:__ZYYY}', "");
Expect(0, 917632, '\P{^Scx:__ZYYY}', "");
Error('\p{Is_Script_Extensions::=__Common}');
Error('\P{Is_Script_Extensions::=__Common}');
Expect(1, 917631, '\p{Is_Script_Extensions=common}', "");
Expect(0, 917631, '\p{^Is_Script_Extensions=common}', "");
Expect(0, 917631, '\P{Is_Script_Extensions=common}', "");
Expect(1, 917631, '\P{^Is_Script_Extensions=common}', "");
Expect(0, 917632, '\p{Is_Script_Extensions=common}', "");
Expect(1, 917632, '\p{^Is_Script_Extensions=common}', "");
Expect(1, 917632, '\P{Is_Script_Extensions=common}', "");
Expect(0, 917632, '\P{^Is_Script_Extensions=common}', "");
Expect(1, 917631, '\p{Is_Script_Extensions=- COMMON}', "");
Expect(0, 917631, '\p{^Is_Script_Extensions=- COMMON}', "");
Expect(0, 917631, '\P{Is_Script_Extensions=- COMMON}', "");
Expect(1, 917631, '\P{^Is_Script_Extensions=- COMMON}', "");
Expect(0, 917632, '\p{Is_Script_Extensions=- COMMON}', "");
Expect(1, 917632, '\p{^Is_Script_Extensions=- COMMON}', "");
Expect(1, 917632, '\P{Is_Script_Extensions=- COMMON}', "");
Expect(0, 917632, '\P{^Is_Script_Extensions=- COMMON}', "");
Error('\p{Is_Scx=-Zyyy/a/}');
Error('\P{Is_Scx=-Zyyy/a/}');
Expect(1, 917631, '\p{Is_Scx=zyyy}', "");
Expect(0, 917631, '\p{^Is_Scx=zyyy}', "");
Expect(0, 917631, '\P{Is_Scx=zyyy}', "");
Expect(1, 917631, '\P{^Is_Scx=zyyy}', "");
Expect(0, 917632, '\p{Is_Scx=zyyy}', "");
Expect(1, 917632, '\p{^Is_Scx=zyyy}', "");
Expect(1, 917632, '\P{Is_Scx=zyyy}', "");
Expect(0, 917632, '\P{^Is_Scx=zyyy}', "");
Expect(1, 917631, '\p{Is_Scx=	zyyy}', "");
Expect(0, 917631, '\p{^Is_Scx=	zyyy}', "");
Expect(0, 917631, '\P{Is_Scx=	zyyy}', "");
Expect(1, 917631, '\P{^Is_Scx=	zyyy}', "");
Expect(0, 917632, '\p{Is_Scx=	zyyy}', "");
Expect(1, 917632, '\p{^Is_Scx=	zyyy}', "");
Expect(1, 917632, '\P{Is_Scx=	zyyy}', "");
Expect(0, 917632, '\P{^Is_Scx=	zyyy}', "");
Error('\p{Script_Extensions=:=		Unknown}');
Error('\P{Script_Extensions=:=		Unknown}');
Expect(1, 918000, '\p{Script_Extensions=unknown}', "");
Expect(0, 918000, '\p{^Script_Extensions=unknown}', "");
Expect(0, 918000, '\P{Script_Extensions=unknown}', "");
Expect(1, 918000, '\P{^Script_Extensions=unknown}', "");
Expect(0, 917999, '\p{Script_Extensions=unknown}', "");
Expect(1, 917999, '\p{^Script_Extensions=unknown}', "");
Expect(1, 917999, '\P{Script_Extensions=unknown}', "");
Expect(0, 917999, '\P{^Script_Extensions=unknown}', "");
Expect(1, 918000, '\p{Script_Extensions=_ Unknown}', "");
Expect(0, 918000, '\p{^Script_Extensions=_ Unknown}', "");
Expect(0, 918000, '\P{Script_Extensions=_ Unknown}', "");
Expect(1, 918000, '\P{^Script_Extensions=_ Unknown}', "");
Expect(0, 917999, '\p{Script_Extensions=_ Unknown}', "");
Expect(1, 917999, '\p{^Script_Extensions=_ Unknown}', "");
Expect(1, 917999, '\P{Script_Extensions=_ Unknown}', "");
Expect(0, 917999, '\P{^Script_Extensions=_ Unknown}', "");
Error('\p{Scx=	_Zzzz/a/}');
Error('\P{Scx=	_Zzzz/a/}');
Expect(1, 918000, '\p{Scx=zzzz}', "");
Expect(0, 918000, '\p{^Scx=zzzz}', "");
Expect(0, 918000, '\P{Scx=zzzz}', "");
Expect(1, 918000, '\P{^Scx=zzzz}', "");
Expect(0, 917999, '\p{Scx=zzzz}', "");
Expect(1, 917999, '\p{^Scx=zzzz}', "");
Expect(1, 917999, '\P{Scx=zzzz}', "");
Expect(0, 917999, '\P{^Scx=zzzz}', "");
Expect(1, 918000, '\p{Scx=	_Zzzz}', "");
Expect(0, 918000, '\p{^Scx=	_Zzzz}', "");
Expect(0, 918000, '\P{Scx=	_Zzzz}', "");
Expect(1, 918000, '\P{^Scx=	_Zzzz}', "");
Expect(0, 917999, '\p{Scx=	_Zzzz}', "");
Expect(1, 917999, '\p{^Scx=	_Zzzz}', "");
Expect(1, 917999, '\P{Scx=	_Zzzz}', "");
Expect(0, 917999, '\P{^Scx=	_Zzzz}', "");
Error('\p{Is_Script_Extensions=:=-_unknown}');
Error('\P{Is_Script_Extensions=:=-_unknown}');
Expect(1, 918000, '\p{Is_Script_Extensions=unknown}', "");
Expect(0, 918000, '\p{^Is_Script_Extensions=unknown}', "");
Expect(0, 918000, '\P{Is_Script_Extensions=unknown}', "");
Expect(1, 918000, '\P{^Is_Script_Extensions=unknown}', "");
Expect(0, 917999, '\p{Is_Script_Extensions=unknown}', "");
Expect(1, 917999, '\p{^Is_Script_Extensions=unknown}', "");
Expect(1, 917999, '\P{Is_Script_Extensions=unknown}', "");
Expect(0, 917999, '\P{^Is_Script_Extensions=unknown}', "");
Expect(1, 918000, '\p{Is_Script_Extensions=	_unknown}', "");
Expect(0, 918000, '\p{^Is_Script_Extensions=	_unknown}', "");
Expect(0, 918000, '\P{Is_Script_Extensions=	_unknown}', "");
Expect(1, 918000, '\P{^Is_Script_Extensions=	_unknown}', "");
Expect(0, 917999, '\p{Is_Script_Extensions=	_unknown}', "");
Expect(1, 917999, '\p{^Is_Script_Extensions=	_unknown}', "");
Expect(1, 917999, '\P{Is_Script_Extensions=	_unknown}', "");
Expect(0, 917999, '\P{^Is_Script_Extensions=	_unknown}', "");
Error('\p{Is_Scx=_/a/Zzzz}');
Error('\P{Is_Scx=_/a/Zzzz}');
Expect(1, 918000, '\p{Is_Scx:   zzzz}', "");
Expect(0, 918000, '\p{^Is_Scx:   zzzz}', "");
Expect(0, 918000, '\P{Is_Scx:   zzzz}', "");
Expect(1, 918000, '\P{^Is_Scx:   zzzz}', "");
Expect(0, 917999, '\p{Is_Scx:   zzzz}', "");
Expect(1, 917999, '\p{^Is_Scx:   zzzz}', "");
Expect(1, 917999, '\P{Is_Scx:   zzzz}', "");
Expect(0, 917999, '\P{^Is_Scx:   zzzz}', "");
Expect(1, 918000, '\p{Is_Scx=_-ZZZZ}', "");
Expect(0, 918000, '\p{^Is_Scx=_-ZZZZ}', "");
Expect(0, 918000, '\P{Is_Scx=_-ZZZZ}', "");
Expect(1, 918000, '\P{^Is_Scx=_-ZZZZ}', "");
Expect(0, 917999, '\p{Is_Scx=_-ZZZZ}', "");
Expect(1, 917999, '\p{^Is_Scx=_-ZZZZ}', "");
Expect(1, 917999, '\P{Is_Scx=_-ZZZZ}', "");
Expect(0, 917999, '\P{^Is_Scx=_-ZZZZ}', "");
Test_X('÷ 0020 ÷ 0020 ÷');
Test_X('÷ 0020 × 0308 ÷ 0020 ÷');
Test_X('÷ 0020 ÷ 000D ÷');
Test_X('÷ 0020 × 0308 ÷ 000D ÷');
Test_X('÷ 0020 ÷ 000A ÷');
Test_X('÷ 0020 × 0308 ÷ 000A ÷');
Test_X('÷ 0020 ÷ 0001 ÷');
Test_X('÷ 0020 × 0308 ÷ 0001 ÷');
Test_X('÷ 0020 × 0300 ÷');
Test_X('÷ 0020 × 0308 × 0300 ÷');
Test_X('÷ 0020 × 0903 ÷');
Test_X('÷ 0020 × 0308 × 0903 ÷');
Test_X('÷ 0020 ÷ 1100 ÷');
Test_X('÷ 0020 × 0308 ÷ 1100 ÷');
Test_X('÷ 0020 ÷ 1160 ÷');
Test_X('÷ 0020 × 0308 ÷ 1160 ÷');
Test_X('÷ 0020 ÷ 11A8 ÷');
Test_X('÷ 0020 × 0308 ÷ 11A8 ÷');
Test_X('÷ 0020 ÷ AC00 ÷');
Test_X('÷ 0020 × 0308 ÷ AC00 ÷');
Test_X('÷ 0020 ÷ AC01 ÷');
Test_X('÷ 0020 × 0308 ÷ AC01 ÷');
Test_X('÷ 0020 ÷ 1F1E6 ÷');
Test_X('÷ 0020 × 0308 ÷ 1F1E6 ÷');
Test_X('÷ 0020 ÷ 0378 ÷');
Test_X('÷ 0020 × 0308 ÷ 0378 ÷');
Test_X('÷ 0020 ÷ D800 ÷');
Test_X('÷ 0020 × 0308 ÷ D800 ÷');
Test_X('÷ 000D ÷ 0020 ÷');
Test_X('÷ 000D ÷ 0308 ÷ 0020 ÷');
Test_X('÷ 000D ÷ 000D ÷');
Test_X('÷ 000D ÷ 0308 ÷ 000D ÷');
Test_X('÷ 000D × 000A ÷');
Test_X('÷ 000D ÷ 0308 ÷ 000A ÷');
Test_X('÷ 000D ÷ 0001 ÷');
Test_X('÷ 000D ÷ 0308 ÷ 0001 ÷');
Test_X('÷ 000D ÷ 0300 ÷');
Test_X('÷ 000D ÷ 0308 × 0300 ÷');
Test_X('÷ 000D ÷ 0903 ÷');
Test_X('÷ 000D ÷ 0308 × 0903 ÷');
Test_X('÷ 000D ÷ 1100 ÷');
Test_X('÷ 000D ÷ 0308 ÷ 1100 ÷');
Test_X('÷ 000D ÷ 1160 ÷');
Test_X('÷ 000D ÷ 0308 ÷ 1160 ÷');
Test_X('÷ 000D ÷ 11A8 ÷');
Test_X('÷ 000D ÷ 0308 ÷ 11A8 ÷');
Test_X('÷ 000D ÷ AC00 ÷');
Test_X('÷ 000D ÷ 0308 ÷ AC00 ÷');
Test_X('÷ 000D ÷ AC01 ÷');
Test_X('÷ 000D ÷ 0308 ÷ AC01 ÷');
Test_X('÷ 000D ÷ 1F1E6 ÷');
Test_X('÷ 000D ÷ 0308 ÷ 1F1E6 ÷');
Test_X('÷ 000D ÷ 0378 ÷');
Test_X('÷ 000D ÷ 0308 ÷ 0378 ÷');
Test_X('÷ 000D ÷ D800 ÷');
Test_X('÷ 000D ÷ 0308 ÷ D800 ÷');
Test_X('÷ 000A ÷ 0020 ÷');
Test_X('÷ 000A ÷ 0308 ÷ 0020 ÷');
Test_X('÷ 000A ÷ 000D ÷');
Test_X('÷ 000A ÷ 0308 ÷ 000D ÷');
Test_X('÷ 000A ÷ 000A ÷');
Test_X('÷ 000A ÷ 0308 ÷ 000A ÷');
Test_X('÷ 000A ÷ 0001 ÷');
Test_X('÷ 000A ÷ 0308 ÷ 0001 ÷');
Test_X('÷ 000A ÷ 0300 ÷');
Test_X('÷ 000A ÷ 0308 × 0300 ÷');
Test_X('÷ 000A ÷ 0903 ÷');
Test_X('÷ 000A ÷ 0308 × 0903 ÷');
Test_X('÷ 000A ÷ 1100 ÷');
Test_X('÷ 000A ÷ 0308 ÷ 1100 ÷');
Test_X('÷ 000A ÷ 1160 ÷');
Test_X('÷ 000A ÷ 0308 ÷ 1160 ÷');
Test_X('÷ 000A ÷ 11A8 ÷');
Test_X('÷ 000A ÷ 0308 ÷ 11A8 ÷');
Test_X('÷ 000A ÷ AC00 ÷');
Test_X('÷ 000A ÷ 0308 ÷ AC00 ÷');
Test_X('÷ 000A ÷ AC01 ÷');
Test_X('÷ 000A ÷ 0308 ÷ AC01 ÷');
Test_X('÷ 000A ÷ 1F1E6 ÷');
Test_X('÷ 000A ÷ 0308 ÷ 1F1E6 ÷');
Test_X('÷ 000A ÷ 0378 ÷');
Test_X('÷ 000A ÷ 0308 ÷ 0378 ÷');
Test_X('÷ 000A ÷ D800 ÷');
Test_X('÷ 000A ÷ 0308 ÷ D800 ÷');
Test_X('÷ 0001 ÷ 0020 ÷');
Test_X('÷ 0001 ÷ 0308 ÷ 0020 ÷');
Test_X('÷ 0001 ÷ 000D ÷');
Test_X('÷ 0001 ÷ 0308 ÷ 000D ÷');
Test_X('÷ 0001 ÷ 000A ÷');
Test_X('÷ 0001 ÷ 0308 ÷ 000A ÷');
Test_X('÷ 0001 ÷ 0001 ÷');
Test_X('÷ 0001 ÷ 0308 ÷ 0001 ÷');
Test_X('÷ 0001 ÷ 0300 ÷');
Test_X('÷ 0001 ÷ 0308 × 0300 ÷');
Test_X('÷ 0001 ÷ 0903 ÷');
Test_X('÷ 0001 ÷ 0308 × 0903 ÷');
Test_X('÷ 0001 ÷ 1100 ÷');
Test_X('÷ 0001 ÷ 0308 ÷ 1100 ÷');
Test_X('÷ 0001 ÷ 1160 ÷');
Test_X('÷ 0001 ÷ 0308 ÷ 1160 ÷');
Test_X('÷ 0001 ÷ 11A8 ÷');
Test_X('÷ 0001 ÷ 0308 ÷ 11A8 ÷');
Test_X('÷ 0001 ÷ AC00 ÷');
Test_X('÷ 0001 ÷ 0308 ÷ AC00 ÷');
Test_X('÷ 0001 ÷ AC01 ÷');
Test_X('÷ 0001 ÷ 0308 ÷ AC01 ÷');
Test_X('÷ 0001 ÷ 1F1E6 ÷');
Test_X('÷ 0001 ÷ 0308 ÷ 1F1E6 ÷');
Test_X('÷ 0001 ÷ 0378 ÷');
Test_X('÷ 0001 ÷ 0308 ÷ 0378 ÷');
Test_X('÷ 0001 ÷ D800 ÷');
Test_X('÷ 0001 ÷ 0308 ÷ D800 ÷');
Test_X('÷ 0300 ÷ 0020 ÷');
Test_X('÷ 0300 × 0308 ÷ 0020 ÷');
Test_X('÷ 0300 ÷ 000D ÷');
Test_X('÷ 0300 × 0308 ÷ 000D ÷');
Test_X('÷ 0300 ÷ 000A ÷');
Test_X('÷ 0300 × 0308 ÷ 000A ÷');
Test_X('÷ 0300 ÷ 0001 ÷');
Test_X('÷ 0300 × 0308 ÷ 0001 ÷');
Test_X('÷ 0300 × 0300 ÷');
Test_X('÷ 0300 × 0308 × 0300 ÷');
Test_X('÷ 0300 × 0903 ÷');
Test_X('÷ 0300 × 0308 × 0903 ÷');
Test_X('÷ 0300 ÷ 1100 ÷');
Test_X('÷ 0300 × 0308 ÷ 1100 ÷');
Test_X('÷ 0300 ÷ 1160 ÷');
Test_X('÷ 0300 × 0308 ÷ 1160 ÷');
Test_X('÷ 0300 ÷ 11A8 ÷');
Test_X('÷ 0300 × 0308 ÷ 11A8 ÷');
Test_X('÷ 0300 ÷ AC00 ÷');
Test_X('÷ 0300 × 0308 ÷ AC00 ÷');
Test_X('÷ 0300 ÷ AC01 ÷');
Test_X('÷ 0300 × 0308 ÷ AC01 ÷');
Test_X('÷ 0300 ÷ 1F1E6 ÷');
Test_X('÷ 0300 × 0308 ÷ 1F1E6 ÷');
Test_X('÷ 0300 ÷ 0378 ÷');
Test_X('÷ 0300 × 0308 ÷ 0378 ÷');
Test_X('÷ 0300 ÷ D800 ÷');
Test_X('÷ 0300 × 0308 ÷ D800 ÷');
Test_X('÷ 0903 ÷ 0020 ÷');
Test_X('÷ 0903 × 0308 ÷ 0020 ÷');
Test_X('÷ 0903 ÷ 000D ÷');
Test_X('÷ 0903 × 0308 ÷ 000D ÷');
Test_X('÷ 0903 ÷ 000A ÷');
Test_X('÷ 0903 × 0308 ÷ 000A ÷');
Test_X('÷ 0903 ÷ 0001 ÷');
Test_X('÷ 0903 × 0308 ÷ 0001 ÷');
Test_X('÷ 0903 × 0300 ÷');
Test_X('÷ 0903 × 0308 × 0300 ÷');
Test_X('÷ 0903 × 0903 ÷');
Test_X('÷ 0903 × 0308 × 0903 ÷');
Test_X('÷ 0903 ÷ 1100 ÷');
Test_X('÷ 0903 × 0308 ÷ 1100 ÷');
Test_X('÷ 0903 ÷ 1160 ÷');
Test_X('÷ 0903 × 0308 ÷ 1160 ÷');
Test_X('÷ 0903 ÷ 11A8 ÷');
Test_X('÷ 0903 × 0308 ÷ 11A8 ÷');
Test_X('÷ 0903 ÷ AC00 ÷');
Test_X('÷ 0903 × 0308 ÷ AC00 ÷');
Test_X('÷ 0903 ÷ AC01 ÷');
Test_X('÷ 0903 × 0308 ÷ AC01 ÷');
Test_X('÷ 0903 ÷ 1F1E6 ÷');
Test_X('÷ 0903 × 0308 ÷ 1F1E6 ÷');
Test_X('÷ 0903 ÷ 0378 ÷');
Test_X('÷ 0903 × 0308 ÷ 0378 ÷');
Test_X('÷ 0903 ÷ D800 ÷');
Test_X('÷ 0903 × 0308 ÷ D800 ÷');
Test_X('÷ 1100 ÷ 0020 ÷');
Test_X('÷ 1100 × 0308 ÷ 0020 ÷');
Test_X('÷ 1100 ÷ 000D ÷');
Test_X('÷ 1100 × 0308 ÷ 000D ÷');
Test_X('÷ 1100 ÷ 000A ÷');
Test_X('÷ 1100 × 0308 ÷ 000A ÷');
Test_X('÷ 1100 ÷ 0001 ÷');
Test_X('÷ 1100 × 0308 ÷ 0001 ÷');
Test_X('÷ 1100 × 0300 ÷');
Test_X('÷ 1100 × 0308 × 0300 ÷');
Test_X('÷ 1100 × 0903 ÷');
Test_X('÷ 1100 × 0308 × 0903 ÷');
Test_X('÷ 1100 × 1100 ÷');
Test_X('÷ 1100 × 0308 ÷ 1100 ÷');
Test_X('÷ 1100 × 1160 ÷');
Test_X('÷ 1100 × 0308 ÷ 1160 ÷');
Test_X('÷ 1100 ÷ 11A8 ÷');
Test_X('÷ 1100 × 0308 ÷ 11A8 ÷');
Test_X('÷ 1100 × AC00 ÷');
Test_X('÷ 1100 × 0308 ÷ AC00 ÷');
Test_X('÷ 1100 × AC01 ÷');
Test_X('÷ 1100 × 0308 ÷ AC01 ÷');
Test_X('÷ 1100 ÷ 1F1E6 ÷');
Test_X('÷ 1100 × 0308 ÷ 1F1E6 ÷');
Test_X('÷ 1100 ÷ 0378 ÷');
Test_X('÷ 1100 × 0308 ÷ 0378 ÷');
Test_X('÷ 1100 ÷ D800 ÷');
Test_X('÷ 1100 × 0308 ÷ D800 ÷');
Test_X('÷ 1160 ÷ 0020 ÷');
Test_X('÷ 1160 × 0308 ÷ 0020 ÷');
Test_X('÷ 1160 ÷ 000D ÷');
Test_X('÷ 1160 × 0308 ÷ 000D ÷');
Test_X('÷ 1160 ÷ 000A ÷');
Test_X('÷ 1160 × 0308 ÷ 000A ÷');
Test_X('÷ 1160 ÷ 0001 ÷');
Test_X('÷ 1160 × 0308 ÷ 0001 ÷');
Test_X('÷ 1160 × 0300 ÷');
Test_X('÷ 1160 × 0308 × 0300 ÷');
Test_X('÷ 1160 × 0903 ÷');
Test_X('÷ 1160 × 0308 × 0903 ÷');
Test_X('÷ 1160 ÷ 1100 ÷');
Test_X('÷ 1160 × 0308 ÷ 1100 ÷');
Test_X('÷ 1160 × 1160 ÷');
Test_X('÷ 1160 × 0308 ÷ 1160 ÷');
Test_X('÷ 1160 × 11A8 ÷');
Test_X('÷ 1160 × 0308 ÷ 11A8 ÷');
Test_X('÷ 1160 ÷ AC00 ÷');
Test_X('÷ 1160 × 0308 ÷ AC00 ÷');
Test_X('÷ 1160 ÷ AC01 ÷');
Test_X('÷ 1160 × 0308 ÷ AC01 ÷');
Test_X('÷ 1160 ÷ 1F1E6 ÷');
Test_X('÷ 1160 × 0308 ÷ 1F1E6 ÷');
Test_X('÷ 1160 ÷ 0378 ÷');
Test_X('÷ 1160 × 0308 ÷ 0378 ÷');
Test_X('÷ 1160 ÷ D800 ÷');
Test_X('÷ 1160 × 0308 ÷ D800 ÷');
Test_X('÷ 11A8 ÷ 0020 ÷');
Test_X('÷ 11A8 × 0308 ÷ 0020 ÷');
Test_X('÷ 11A8 ÷ 000D ÷');
Test_X('÷ 11A8 × 0308 ÷ 000D ÷');
Test_X('÷ 11A8 ÷ 000A ÷');
Test_X('÷ 11A8 × 0308 ÷ 000A ÷');
Test_X('÷ 11A8 ÷ 0001 ÷');
Test_X('÷ 11A8 × 0308 ÷ 0001 ÷');
Test_X('÷ 11A8 × 0300 ÷');
Test_X('÷ 11A8 × 0308 × 0300 ÷');
Test_X('÷ 11A8 × 0903 ÷');
Test_X('÷ 11A8 × 0308 × 0903 ÷');
Test_X('÷ 11A8 ÷ 1100 ÷');
Test_X('÷ 11A8 × 0308 ÷ 1100 ÷');
Test_X('÷ 11A8 ÷ 1160 ÷');
Test_X('÷ 11A8 × 0308 ÷ 1160 ÷');
Test_X('÷ 11A8 × 11A8 ÷');
Test_X('÷ 11A8 × 0308 ÷ 11A8 ÷');
Test_X('÷ 11A8 ÷ AC00 ÷');
Test_X('÷ 11A8 × 0308 ÷ AC00 ÷');
Test_X('÷ 11A8 ÷ AC01 ÷');
Test_X('÷ 11A8 × 0308 ÷ AC01 ÷');
Test_X('÷ 11A8 ÷ 1F1E6 ÷');
Test_X('÷ 11A8 × 0308 ÷ 1F1E6 ÷');
Test_X('÷ 11A8 ÷ 0378 ÷');
Test_X('÷ 11A8 × 0308 ÷ 0378 ÷');
Test_X('÷ 11A8 ÷ D800 ÷');
Test_X('÷ 11A8 × 0308 ÷ D800 ÷');
Test_X('÷ AC00 ÷ 0020 ÷');
Test_X('÷ AC00 × 0308 ÷ 0020 ÷');
Test_X('÷ AC00 ÷ 000D ÷');
Test_X('÷ AC00 × 0308 ÷ 000D ÷');
Test_X('÷ AC00 ÷ 000A ÷');
Test_X('÷ AC00 × 0308 ÷ 000A ÷');
Test_X('÷ AC00 ÷ 0001 ÷');
Test_X('÷ AC00 × 0308 ÷ 0001 ÷');
Test_X('÷ AC00 × 0300 ÷');
Test_X('÷ AC00 × 0308 × 0300 ÷');
Test_X('÷ AC00 × 0903 ÷');
Test_X('÷ AC00 × 0308 × 0903 ÷');
Test_X('÷ AC00 ÷ 1100 ÷');
Test_X('÷ AC00 × 0308 ÷ 1100 ÷');
Test_X('÷ AC00 × 1160 ÷');
Test_X('÷ AC00 × 0308 ÷ 1160 ÷');
Test_X('÷ AC00 × 11A8 ÷');
Test_X('÷ AC00 × 0308 ÷ 11A8 ÷');
Test_X('÷ AC00 ÷ AC00 ÷');
Test_X('÷ AC00 × 0308 ÷ AC00 ÷');
Test_X('÷ AC00 ÷ AC01 ÷');
Test_X('÷ AC00 × 0308 ÷ AC01 ÷');
Test_X('÷ AC00 ÷ 1F1E6 ÷');
Test_X('÷ AC00 × 0308 ÷ 1F1E6 ÷');
Test_X('÷ AC00 ÷ 0378 ÷');
Test_X('÷ AC00 × 0308 ÷ 0378 ÷');
Test_X('÷ AC00 ÷ D800 ÷');
Test_X('÷ AC00 × 0308 ÷ D800 ÷');
Test_X('÷ AC01 ÷ 0020 ÷');
Test_X('÷ AC01 × 0308 ÷ 0020 ÷');
Test_X('÷ AC01 ÷ 000D ÷');
Test_X('÷ AC01 × 0308 ÷ 000D ÷');
Test_X('÷ AC01 ÷ 000A ÷');
Test_X('÷ AC01 × 0308 ÷ 000A ÷');
Test_X('÷ AC01 ÷ 0001 ÷');
Test_X('÷ AC01 × 0308 ÷ 0001 ÷');
Test_X('÷ AC01 × 0300 ÷');
Test_X('÷ AC01 × 0308 × 0300 ÷');
Test_X('÷ AC01 × 0903 ÷');
Test_X('÷ AC01 × 0308 × 0903 ÷');
Test_X('÷ AC01 ÷ 1100 ÷');
Test_X('÷ AC01 × 0308 ÷ 1100 ÷');
Test_X('÷ AC01 ÷ 1160 ÷');
Test_X('÷ AC01 × 0308 ÷ 1160 ÷');
Test_X('÷ AC01 × 11A8 ÷');
Test_X('÷ AC01 × 0308 ÷ 11A8 ÷');
Test_X('÷ AC01 ÷ AC00 ÷');
Test_X('÷ AC01 × 0308 ÷ AC00 ÷');
Test_X('÷ AC01 ÷ AC01 ÷');
Test_X('÷ AC01 × 0308 ÷ AC01 ÷');
Test_X('÷ AC01 ÷ 1F1E6 ÷');
Test_X('÷ AC01 × 0308 ÷ 1F1E6 ÷');
Test_X('÷ AC01 ÷ 0378 ÷');
Test_X('÷ AC01 × 0308 ÷ 0378 ÷');
Test_X('÷ AC01 ÷ D800 ÷');
Test_X('÷ AC01 × 0308 ÷ D800 ÷');
Test_X('÷ 1F1E6 ÷ 0020 ÷');
Test_X('÷ 1F1E6 × 0308 ÷ 0020 ÷');
Test_X('÷ 1F1E6 ÷ 000D ÷');
Test_X('÷ 1F1E6 × 0308 ÷ 000D ÷');
Test_X('÷ 1F1E6 ÷ 000A ÷');
Test_X('÷ 1F1E6 × 0308 ÷ 000A ÷');
Test_X('÷ 1F1E6 ÷ 0001 ÷');
Test_X('÷ 1F1E6 × 0308 ÷ 0001 ÷');
Test_X('÷ 1F1E6 × 0300 ÷');
Test_X('÷ 1F1E6 × 0308 × 0300 ÷');
Test_X('÷ 1F1E6 × 0903 ÷');
Test_X('÷ 1F1E6 × 0308 × 0903 ÷');
Test_X('÷ 1F1E6 ÷ 1100 ÷');
Test_X('÷ 1F1E6 × 0308 ÷ 1100 ÷');
Test_X('÷ 1F1E6 ÷ 1160 ÷');
Test_X('÷ 1F1E6 × 0308 ÷ 1160 ÷');
Test_X('÷ 1F1E6 ÷ 11A8 ÷');
Test_X('÷ 1F1E6 × 0308 ÷ 11A8 ÷');
Test_X('÷ 1F1E6 ÷ AC00 ÷');
Test_X('÷ 1F1E6 × 0308 ÷ AC00 ÷');
Test_X('÷ 1F1E6 ÷ AC01 ÷');
Test_X('÷ 1F1E6 × 0308 ÷ AC01 ÷');
Test_X('÷ 1F1E6 × 1F1E6 ÷');
Test_X('÷ 1F1E6 × 0308 ÷ 1F1E6 ÷');
Test_X('÷ 1F1E6 ÷ 0378 ÷');
Test_X('÷ 1F1E6 × 0308 ÷ 0378 ÷');
Test_X('÷ 1F1E6 ÷ D800 ÷');
Test_X('÷ 1F1E6 × 0308 ÷ D800 ÷');
Test_X('÷ 0378 ÷ 0020 ÷');
Test_X('÷ 0378 × 0308 ÷ 0020 ÷');
Test_X('÷ 0378 ÷ 000D ÷');
Test_X('÷ 0378 × 0308 ÷ 000D ÷');
Test_X('÷ 0378 ÷ 000A ÷');
Test_X('÷ 0378 × 0308 ÷ 000A ÷');
Test_X('÷ 0378 ÷ 0001 ÷');
Test_X('÷ 0378 × 0308 ÷ 0001 ÷');
Test_X('÷ 0378 × 0300 ÷');
Test_X('÷ 0378 × 0308 × 0300 ÷');
Test_X('÷ 0378 × 0903 ÷');
Test_X('÷ 0378 × 0308 × 0903 ÷');
Test_X('÷ 0378 ÷ 1100 ÷');
Test_X('÷ 0378 × 0308 ÷ 1100 ÷');
Test_X('÷ 0378 ÷ 1160 ÷');
Test_X('÷ 0378 × 0308 ÷ 1160 ÷');
Test_X('÷ 0378 ÷ 11A8 ÷');
Test_X('÷ 0378 × 0308 ÷ 11A8 ÷');
Test_X('÷ 0378 ÷ AC00 ÷');
Test_X('÷ 0378 × 0308 ÷ AC00 ÷');
Test_X('÷ 0378 ÷ AC01 ÷');
Test_X('÷ 0378 × 0308 ÷ AC01 ÷');
Test_X('÷ 0378 ÷ 1F1E6 ÷');
Test_X('÷ 0378 × 0308 ÷ 1F1E6 ÷');
Test_X('÷ 0378 ÷ 0378 ÷');
Test_X('÷ 0378 × 0308 ÷ 0378 ÷');
Test_X('÷ 0378 ÷ D800 ÷');
Test_X('÷ 0378 × 0308 ÷ D800 ÷');
Test_X('÷ D800 ÷ 0020 ÷');
Test_X('÷ D800 ÷ 0308 ÷ 0020 ÷');
Test_X('÷ D800 ÷ 000D ÷');
Test_X('÷ D800 ÷ 0308 ÷ 000D ÷');
Test_X('÷ D800 ÷ 000A ÷');
Test_X('÷ D800 ÷ 0308 ÷ 000A ÷');
Test_X('÷ D800 ÷ 0001 ÷');
Test_X('÷ D800 ÷ 0308 ÷ 0001 ÷');
Test_X('÷ D800 ÷ 0300 ÷');
Test_X('÷ D800 ÷ 0308 × 0300 ÷');
Test_X('÷ D800 ÷ 0903 ÷');
Test_X('÷ D800 ÷ 0308 × 0903 ÷');
Test_X('÷ D800 ÷ 1100 ÷');
Test_X('÷ D800 ÷ 0308 ÷ 1100 ÷');
Test_X('÷ D800 ÷ 1160 ÷');
Test_X('÷ D800 ÷ 0308 ÷ 1160 ÷');
Test_X('÷ D800 ÷ 11A8 ÷');
Test_X('÷ D800 ÷ 0308 ÷ 11A8 ÷');
Test_X('÷ D800 ÷ AC00 ÷');
Test_X('÷ D800 ÷ 0308 ÷ AC00 ÷');
Test_X('÷ D800 ÷ AC01 ÷');
Test_X('÷ D800 ÷ 0308 ÷ AC01 ÷');
Test_X('÷ D800 ÷ 1F1E6 ÷');
Test_X('÷ D800 ÷ 0308 ÷ 1F1E6 ÷');
Test_X('÷ D800 ÷ 0378 ÷');
Test_X('÷ D800 ÷ 0308 ÷ 0378 ÷');
Test_X('÷ D800 ÷ D800 ÷');
Test_X('÷ D800 ÷ 0308 ÷ D800 ÷');
Test_X('÷ 0061 ÷ 1F1E6 ÷ 0062 ÷');
Test_X('÷ 1F1F7 × 1F1FA ÷');
Test_X('÷ 1F1F7 × 1F1FA × 1F1F8 ÷');
Test_X('÷ 1F1F7 × 1F1FA × 1F1F8 × 1F1EA ÷');
Test_X('÷ 1F1F7 × 1F1FA ÷ 200B ÷ 1F1F8 × 1F1EA ÷');
Test_X('÷ 1F1E6 × 1F1E7 × 1F1E8 ÷');
Test_X('÷ 1F1E6 × 200D ÷ 1F1E7 × 1F1E8 ÷');
Test_X('÷ 1F1E6 × 1F1E7 × 200D ÷ 1F1E8 ÷');
Test_X('÷ 0020 × 200D ÷ 0646 ÷');
Test_X('÷ 0646 × 200D ÷ 0020 ÷');
Finished();
